#include <stdio.h>
#include <stdlib.h>

#include "F16Field.h"

/*
	Write a field element to a char array

	W : writer object
	a : field element to write
*/
void f16serialize_FELT(writer *W, f16FELT a) {
	serialize_uint64_t(W, a , 16);
}

/*
	Read a field element from a char array

	R : reader object

	returns : a field element
*/
f16FELT f16deserialize_FELT(reader *R) {
	return ((f16FELT) deserialize_uint64_t(R,16));
}

/*
	prints a field element, mainly used for debugging
*/
void f16printFELT(f16FELT a) {
	/* printf("%.5u ", a); */
	if(a==0){
		printf(".");
	}
	else{
		printf("X");
	}
}

/*
	Multiplies two field elements

	a,b : field element to multiply

	return : the product of a and b
*/
f16FELT f16multiply(f16FELT a, f16FELT b) {
	if (a == 0 || b == 0)
		return 0;
	return f16antilog((f16log(a) + f16log(b)));// % f16units);
}

/*
	Inverts a field element

	a : field element to invert

	return : the inverse of a, if a is nonzero
*/
f16FELT f16inverse(f16FELT a) {
	return f16antilog((f16units - f16log(a)) % f16units);
}

/*
 * Stuff concerning polynomials over F_2^16 , all polynomials are represented by arrays of length 6 because we will only use polynomials of degree up to 5
 */

/*
	The euclidean division algorithm. 
	Given polynomials a and b it calculates the quotient q and remainder r such that
	deg(r) < deg(b) and a = q*b + r 

	a,b : two polynomials over F_2^16
	q : quotient
	r : remainder
*/
void f16EucildeanDivision(f16FELT *a, f16FELT *b, f16FELT *q, f16FELT *r) {
	int i,j;
	f16FELT LCBinv;
	int degB = 0;
	for (i = 0; i < 6; i++) {
		if (b[i] != 0)
			degB = i;
		q[i] = 0;
		r[i] = a[i];
	}

	LCBinv = f16inverse(b[degB]);

	for (i = 5; i >= degB; i--) {
		if (r[i] == 0)
			continue;

		q[i - degB] = f16multiply(LCBinv, r[i]);
		r[i] = 0;
		for (j = 0; j < degB; j++) {
			r[j + i - degB] ^= f16multiply(q[i - degB], b[j]);
		}
	}

	return;
}

/*
	Checks if a is the zero polynomial
*/
int isArrayOfZeros(f16FELT *a) {
	int i;
	for (i = 0; i < 6; i++) {
		if (a[i] != 0)
			return 0;
	}
	return 1;
}

/*
	Multiplies two polynomials over F_2^16

	a,b : two polynomials
	ans : is set to a*b mod (X^6)
*/
void f16polyMult(f16FELT *a, f16FELT *b, f16FELT *ans) {
	int i,j;
	for (i = 0; i < 6; i++) {
		ans[i] = 0;
		for (j = 0; j <= i; j++) {
			ans[i] ^= f16multiply(a[j], b[i - j]);
		}
	}
}

/*
	Adds two polynomials over F_2^16

	a,b : two polynomials
	ans : is set to a+b
*/
void f16polyAdd(f16FELT *a, f16FELT *b, f16FELT *ans) {
	int i;
	for (i = 0; i < 6; i++) {
		ans[i] = f16add(a[i], b[i]);
	}
}

/*
	Sets a polynomial to the zero polynomial
*/
void f16polyZero(f16FELT *a) {
	int i;
	for (i = 0; i < 6; i++) {
		a[i] = 0;
	}
}

/*
	Sets a polynomial to the identity
*/
void f16polyOne(f16FELT *a) {
	int i;
	a[0] = 1;
	for (i = 1; i < 6; i++) {
		a[i] = 0;
	}
}

/*
	Copies a polynomial

	a : a polynomial to copy
	copy : is set to a 
*/
void f16polyCopy(f16FELT *a, f16FELT *copy) {
	int i;
	for (i = 0; i < 6; i++) {
		copy[i] = a[i];
	}
}

/*
Scalar multiplication

	a : an element in the field, which is muliplied by 2^b
	b : an integer
*/
void f16scalarMultiply(f16FELT *a, uint16_t b) {
	if (*a != 0) {
		*a = f16antilog((f16log(*a) + b) % f16units);
	}
}

/*
	Extended Euclidean Algorithm
	given two polynomials a and b it calculates the greatest common divisor, as well as polynomials x and y such that
	x*a + y*b = gcd
*/
void f16ExtendedEuclideanAlgorithm(f16FELT *a, f16FELT *b, f16FELT *x, f16FELT *y, f16FELT *gcd) {
	f16FELT q[6], r[6], A[6], B[6], u[6], v[6], m[6], n[6];

	f16polyCopy(a, A);
	f16polyCopy(b, B);

	f16polyZero(x);
	f16polyOne(y);

	f16polyOne(u);
	f16polyZero(v);

	while (!isArrayOfZeros(A))
	{
		f16FELT uq[6];
		f16FELT vq[6];

		f16EucildeanDivision(B, A, q, r);
		
		f16polyMult(u, q, uq);
		f16polyAdd(x, uq, m);
		
		f16polyMult(v, q, vq);
		f16polyAdd(y, vq, n);

		f16polyCopy(A, B);
		f16polyCopy(r, A);

		f16polyCopy(u, x);
		f16polyCopy(v, y);

		f16polyCopy(m, u);
		f16polyCopy(n, v);
	}
	f16polyCopy(B, gcd);
}

/*
	Calculates the logarithm of a field element with respect to the generator 2

	a : a field element

	return : x such that 2^x = a
*/
uint32_t f16log(f16FELT a)
{
	static const uint32_t f16LogTable[65536] =
	{
		0x0000U,0x0000U,0x0001U,0xc1baU,0x0002U,0x8375U,0xc1bbU,0xc0ceU,0x0003U,0x8289U,
		0x8376U,0x06f7U,0xc1bcU,0x1110U,0xc0cfU,0x4530U,0x0004U,0x06ebU,0x828aU,0x7b06U,
		0x8377U,0x819dU,0x06f8U,0xd2caU,0xc1bdU,0xa0ddU,0x1111U,0x4444U,0xc0d0U,0xc8b1U,
		0x4531U,0xeba4U,0x0005U,0xad5fU,0x06ecU,0xd1deU,0x828bU,0x9ba1U,0x7b07U,0x8a6cU,
		0x8378U,0x2b1dU,0x819eU,0x6298U,0x06f9U,0x05ffU,0xd2cbU,0x07f9U,0xc1beU,0xc7c5U,
		0xa0deU,0xc8a5U,0x1112U,0x3cc1U,0x4445U,0x61f7U,0xc0d1U,0x9485U,0xc8b2U,0x61c8U,
		0x4532U,0xba6eU,0xeba5U,0x4358U,0x0006U,0x0513U,0xad60U,0x8087U,0x06edU,0x0deeU,
		0xd1dfU,0x7c29U,0x828cU,0x90fdU,0x9ba2U,0x5640U,0x7b08U,0x2383U,0x8a6dU,0x61acU,
		0x8379U,0x2220U,0x2b1eU,0x8980U,0x819fU,0x8a60U,0x6299U,0x7fcdU,0x06faU,0x23b2U,
		0x0600U,0x1305U,0xd2ccU,0xac73U,0x07faU,0xfe7bU,0xc1bfU,0xb416U,0xc7c6U,0x6f1aU,
		0xa0dfU,0x9399U,0xc8a6U,0x77b0U,0x1113U,0x4c27U,0x3cc2U,0x426cU,0x4446U,0xb822U,
		0x61f8U,0x5d5cU,0xc0d2U,0xc9b3U,0x9486U,0x579cU,0xc8b3U,0x0468U,0x61c9U,0xc7b9U,
		0x4533U,0x3bd5U,0xba6fU,0xecd7U,0xeba6U,0x2453U,0x4359U,0x1807U,0x0007U,0xd9c1U,
		0x0514U,0x1a75U,0xad61U,0x22c6U,0x8088U,0xe60dU,0x06eeU,0x2e35U,0x0defU,0xfd8fU,
		0xd1e0U,0xae92U,0x7c2aU,0x1504U,0x828dU,0x844eU,0x90feU,0x8b6eU,0x9ba3U,0x1957U,
		0x5641U,0x8894U,0x7b09U,0x8974U,0x2384U,0xfcb4U,0x8a6eU,0x0159U,0x61adU,0xc622U,
		0x837aU,0x2297U,0x2221U,0x75d1U,0x2b1fU,0x30d5U,0x8981U,0xff3cU,0x81a0U,0x396bU,
		0x8a61U,0x067aU,0x629aU,0x81fdU,0x7fceU,0x5554U,0x06fbU,0x1f17U,0x23b3U,0x7b3dU,
		0x0601U,0xb1edU,0x1306U,0x79ddU,0xd2cdU,0x7e32U,0xac74U,0x0de2U,0x07fbU,0x0427U,
		0xfe7cU,0xfb13U,0xc1c0U,0x1eadU,0xb417U,0xc6cdU,0xc7c7U,0x4242U,0x6f1bU,0x8c16U,
		0xa0e0U,0x3de4U,0x939aU,0x7f44U,0xc8a7U,0xc8c7U,0x77b1U,0xcfa8U,0x1114U,0x2367U,
		0x4c28U,0x875aU,0x3cc3U,0x1fe3U,0x426dU,0xe53dU,0x4447U,0xf29bU,0xb823U,0x52b8U,
		0x61f9U,0x17fbU,0x5d5dU,0xebebU,0xc0d3U,0xc036U,0xc9b4U,0xa7d4U,0x9487U,0xf8cdU,
		0x579dU,0x6e2eU,0xc8b4U,0x92adU,0x0469U,0xe56cU,0x61caU,0xd4bfU,0xc7baU,0xd15aU,
		0x4534U,0xb0f9U,0x3bd6U,0xe3daU,0xba70U,0x4b3bU,0xecd8U,0xe778U,0xeba7U,0x4188U,
		0x2454U,0x5c70U,0x435aU,0x13bbU,0x1808U,0x4c1bU,0x0008U,0x0dd6U,0xd9c2U,0x78f1U,
		0x0515U,0xf60cU,0x1a76U,0xd575U,0xad62U,0x0ef0U,0x22c7U,0x0343U,0x8089U,0x1e2bU,
		0xe60eU,0x72d8U,0x06efU,0x033bU,0x2e36U,0x72b4U,0x0df0U,0xa595U,0xfd90U,0x8c82U,
		0xd1e1U,0xa933U,0xae93U,0x14c5U,0x7c2bU,0xefc3U,0x1505U,0x0cf6U,0x828eU,0xcb7eU,
		0x844fU,0x81f1U,0x90ffU,0x698fU,0x8b6fU,0x74e5U,0x9ba4U,0x2fe9U,0x1958U,0x58daU,
		0x5642U,0x4c77U,0x8895U,0xba88U,0x7b0aU,0x9315U,0x8975U,0x3214U,0x2385U,0x387fU,
		0xfcb5U,0x967aU,0x8a6fU,0xc07fU,0x015aU,0x5468U,0x61aeU,0xa727U,0xc623U,0xd70fU,
		0x837bU,0x41bbU,0x2298U,0xe067U,0x2222U,0x8888U,0x75d2U,0xa298U,0x2b20U,0x4dd1U,
		0x30d6U,0xc536U,0x8982U,0x0300U,0xff3dU,0x03fdU,0x81a1U,0x9163U,0x396cU,0x7307U,
		0x8a62U,0xd749U,0x067bU,0x8a82U,0x629bU,0x186bU,0x81feU,0xff9eU,0x7fcfU,0x40ffU,
		0x5555U,0xaaaaU,0x06fcU,0xada6U,0x1f18U,0x44f8U,0x23b4U,0xde70U,0x7b3eU,0xd9b5U,
		0x0602U,0xd306U,0xb1eeU,0xb456U,0x1307U,0x1473U,0x79deU,0xfca3U,0xd2ceU,0x2e9cU,
		0x7e33U,0xe521U,0xac75U,0x4915U,0x0de3U,0xcef5U,0x07fcU,0xa6f8U,0x0428U,0x4083U,
		0xfe7dU,0xd8d5U,0xfb14U,0xe19dU,0xc1c1U,0xd3d3U,0x1eaeU,0x9b7cU,0xb418U,0xdc2fU,
		0xc6ceU,0x9a56U,0xc7c8U,0xa7c8U,0x4243U,0x2ab0U,0x6f1cU,0x1e85U,0x8c17U,0xe480U,
		0xa0e1U,0xd6beU,0x3de5U,0x6d42U,0x939bU,0x68bfU,0x7f45U,0x704dU,0xc8a8U,0xacb1U,
		0xc8c8U,0xefefU,0x77b2U,0xbf4aU,0xcfa9U,0xd8aeU,0x1115U,0x87ddU,0x2368U,0x1c0aU,
		0x4c29U,0x409cU,0x875bU,0xc313U,0x3cc4U,0x32eeU,0x1fe4U,0x4b2fU,0x426eU,0xbe6fU,
		0xe53eU,0xc165U,0x4448U,0xf304U,0xf29cU,0x4609U,0xb824U,0x4d29U,0x52b9U,0xe2eeU,
		0x61faU,0x4a4fU,0x17fcU,0x1be4U,0x5d5eU,0xee25U,0xebecU,0xdb11U,0xc0d4U,0xbcceU,
		0xc037U,0x5696U,0xc9b5U,0x3c2dU,0xa7d5U,0xc5e1U,0x9488U,0x4156U,0xf8ceU,0x3fedU,
		0x579eU,0xcf9cU,0x6e2fU,0xa341U,0xc8b5U,0x66abU,0x92aeU,0xe0d1U,0x046aU,0x3cf8U,
		0xe56dU,0xa576U,0x61cbU,0x3b98U,0xd4c0U,0xcebcU,0xc7bbU,0xe5c5U,0xd15bU,0x73a8U,
		0x4535U,0x68eeU,0xb0faU,0xe451U,0x3bd7U,0x378cU,0xe3dbU,0x2bcdU,0xba71U,0xc0f7U,
		0x4b3cU,0x1909U,0xecd9U,0x227bU,0xe779U,0xf28fU,0xeba8U,0x170fU,0x4189U,0x5779U,
		0x2455U,0xb1d6U,0x5c71U,0x43b8U,0x435bU,0x6c83U,0x13bcU,0xfb25U,0x1809U,0xc834U,
		0x4c1cU,0x51ccU,0x0009U,0x1387U,0x0dd7U,0x6654U,0xd9c3U,0x5b4cU,0x78f2U,0x89efU,
		0x0516U,0x1578U,0xf60dU,0x2e3eU,0x1a77U,0xbce0U,0xd576U,0xb36aU,0xad63U,0xaa48U,
		0x0ef1U,0xd8c9U,0x22c8U,0x1934U,0x0344U,0x5969U,0x808aU,0x0573U,0x1e2cU,0x92c5U,
		0xe60fU,0xacbaU,0x72d9U,0x7391U,0x06f0U,0xfd88U,0x033cU,0x2aa9U,0x2e37U,0xa60cU,
		0x72b5U,0x68acU,0x0df1U,0xed87U,0xa596U,0xe0b1U,0xfd91U,0x28e5U,0x8c83U,0xf946U,
		0xd1e2U,0xb44aU,0xa934U,0xf536U,0xae94U,0x19fcU,0x14c6U,0xe435U,0x7c2cU,0x4829U,
		0xefc4U,0x82b2U,0x1506U,0xdac3U,0x0cf7U,0x16a8U,0x828fU,0x86c4U,0xcb7fU,0x7e89U,
		0x8450U,0x1851U,0x81f2U,0xdf7bU,0x9100U,0x879cU,0x6990U,0x1230U,0x8b70U,0x4d9cU,
		0x74e6U,0xfde7U,0x9ba5U,0x64fcU,0x2feaU,0x02a6U,0x1959U,0x4ce5U,0x58dbU,0x9157U,
		0x5643U,0xc661U,0x4c78U,0x0311U,0x8896U,0x01a8U,0xba89U,0xc75dU,0x7b0bU,0x3563U,
		0x9316U,0x8996U,0x8976U,0x75ecU,0x3215U,0xa780U,0x2386U,0x1038U,0x3880U,0xfd52U,
		0xfcb6U,0x9077U,0x967bU,0xbdf5U,0x8a70U,0x4013U,0xc080U,0x2866U,0x015bU,0xa28cU,
		0x5469U,0xeb14U,0x61afU,0x6731U,0xa728U,0x2917U,0xc624U,0xb9b0U,0xd710U,0xfeb2U,
		0x837cU,0x82ffU,0x41bcU,0x958eU,0x2299U,0x5d37U,0xe068U,0xca42U,0x2223U,0x5c11U,
		0x8889U,0xa613U,0x75d3U,0x9229U,0xa299U,0x9deaU,0x2b21U,0xa63bU,0x4dd2U,0x97f4U,
		0x30d7U,0xd393U,0xc537U,0xe03fU,0x8983U,0x8a39U,0x0301U,0x6983U,0xff3eU,0xec6aU,
		0x03feU,0x537cU,0x81a2U,0x9a69U,0x9164U,0x623aU,0x396dU,0x4df0U,0x7308U,0x8105U,
		0x8a63U,0x68a3U,0xd74aU,0x6e6cU,0x067cU,0xb1aaU,0x8a83U,0xc33aU,0x629cU,0xc932U,
		0x186cU,0x9879U,0x81ffU,0x2efdU,0xff9fU,0x50d0U,0x7fd0U,0x3208U,0x4100U,0xb99cU,
		0x5556U,0x9654U,0xaaabU,0x2a7aU,0x06fdU,0x9cccU,0xada7U,0x88c0U,0x1f19U,0xa847U,
		0x44f9U,0xafe0U,0x23b5U,0x4d6dU,0xde71U,0x0c0aU,0x7b3fU,0xdd9eU,0xd9b6U,0x6288U,
		0x0603U,0x2df9U,0xd307U,0xb4bfU,0xb1efU,0x07c4U,0xb457U,0x71c8U,0x1308U,0xa4a9U,
		0x1474U,0x5bf6U,0x79dfU,0x877eU,0xfca4U,0x0ee4U,0xd2cfU,0x1d3fU,0x2e9dU,0x4998U,
		0x7e34U,0xddc4U,0xe522U,0x14e5U,0xac76U,0x84ceU,0x4916U,0xa623U,0x0de4U,0xc526U,
		0xcef6U,0x0257U,0x07fdU,0x8320U,0xa6f9U,0xd489U,0x0429U,0xc6a9U,0x4084U,0x802aU,
		0xfe7eU,0xcacfU,0xd8d6U,0xf4a8U,0xfb15U,0x0ceaU,0xe19eU,0x02d5U,0xc1c2U,0x173cU,
		0xd3d4U,0xcf90U,0x1eafU,0x3aacU,0x9b7dU,0x16c2U,0xb419U,0x9730U,0xdc30U,0x72bcU,
		0xc6cfU,0x5e93U,0x9a57U,0xb7c7U,0xc7c9U,0x3493U,0xa7c9U,0xd6e6U,0x4244U,0xa2ebU,
		0x2ab1U,0xdfe5U,0x6f1dU,0x3c0cU,0x1e86U,0xd0aaU,0x8c18U,0xc4fdU,0xe481U,0xb404U,
		0xa0e2U,0xceb0U,0xd6bfU,0xe999U,0x3de6U,0xa20dU,0x6d43U,0xb17eU,0x939cU,0xdcfdU,
		0x68c0U,0x6aeeU,0x7f46U,0xd67fU,0x704eU,0x3f01U,0xc8a9U,0xefbbU,0xacb2U,0xc4f5U,
		0xc8c9U,0x346fU,0xeff0U,0x0b5fU,0x77b3U,0x4e3dU,0xbf4bU,0x7fe0U,0xcfaaU,0xbbe2U,
		0xd8afU,0x6750U,0x1116U,0x98caU,0x87deU,0x42ccU,0x2369U,0xf39dU,0x1c0bU,0x68e2U,
		0x4c2aU,0xce00U,0x409dU,0x823aU,0x875cU,0x1623U,0xc314U,0xf40cU,0x3cc5U,0xc748U,
		0x32efU,0x54d0U,0x1fe5U,0xf3ceU,0x4b30U,0x9197U,0x426fU,0x5835U,0xbe70U,0xa1c4U,
		0xe53fU,0x3c7fU,0xc166U,0xfa39U,0x4449U,0x45b6U,0xf305U,0x8d39U,0xf29dU,0x43acU,
		0x460aU,0xe365U,0xb825U,0x36a0U,0x4d2aU,0x4dbdU,0x52baU,0xf067U,0xe2efU,0x2b4aU,
		0x61fbU,0x7c43U,0x4a50U,0xb288U,0x17fdU,0xc00bU,0x1be5U,0x0e32U,0x5d5fU,0x1efeU,
		0xee26U,0xf1a3U,0xebedU,0x1a95U,0xdb12U,0xad41U,0xc0d5U,0xa358U,0xbccfU,0x8746U,
		0xc038U,0x1acaU,0x5697U,0x9a90U,0xc9b6U,0xdb43U,0x3c2eU,0x68b3U,0xa7d6U,0x023eU,
		0xc5e2U,0x8d7fU,0x9489U,0x7ea7U,0x4157U,0xf056U,0xf8cfU,0xa6dcU,0x3feeU,0x3575U,
		0x579fU,0x90b0U,0xcf9dU,0xe394U,0x6e30U,0x3330U,0xa342U,0x0ad0U,0xc8b6U,0x2889U,
		0x66acU,0x6f61U,0x92afU,0x06b3U,0xe0d2U,0xa8d6U,0x046bU,0x9b70U,0x3cf9U,0xbb0dU,
		0xe56eU,0xd5d2U,0xa577U,0xa02bU,0x61ccU,0xbe5eU,0x3b99U,0x90ddU,0xd4c1U,0xedccU,
		0xcebdU,0xd62dU,0xc7bcU,0xb441U,0xe5c6U,0x94c1U,0xd15cU,0x7611U,0x73a9U,0xef03U,
		0x4536U,0xbd83U,0x68efU,0x0376U,0xb0fbU,0xa222U,0xe452U,0x506fU,0x3bd8U,0x6453U,
		0x378dU,0x63e3U,0xe3dcU,0xb0e4U,0x2bceU,0x4a43U,0xba72U,0xc5b7U,0xc0f8U,0xc227U,
		0x4b3dU,0xfe5aU,0x190aU,0xc4baU,0xecdaU,0x3d9fU,0x227cU,0x0f8cU,0xe77aU,0x86f1U,
		0xf290U,0xbf19U,0xeba9U,0x6c65U,0x1710U,0xd524U,0x418aU,0x4963U,0x577aU,0x02baU,
		0x2456U,0xfa9bU,0xb1d7U,0xda25U,0x5c72U,0xc159U,0x43b9U,0x2415U,0x435cU,0xcbfaU,
		0x6c84U,0x531eU,0x13bdU,0x34c2U,0xfb26U,0x451dU,0x180aU,0x4c3dU,0xc835U,0xe499U,
		0x4c1dU,0xdcceU,0x51cdU,0x9904U,0x000aU,0x5abfU,0x1388U,0x8529U,0x0dd8U,0xa3bdU,
		0x6655U,0x9e89U,0xd9c4U,0xb708U,0x5b4dU,0x0df8U,0x78f3U,0xa654U,0x89f0U,0x3e30U,
		0x0517U,0x3743U,0x1579U,0x8db5U,0xf60eU,0x14d9U,0x2e3fU,0xb3d3U,0x1a78U,0x06d8U,
		0xbce1U,0x4768U,0xd577U,0xc8d1U,0xb36bU,0xf67cU,0xad64U,0xdcb2U,0xaa49U,0x2e20U,
		0x0ef2U,0x96dfU,0xd8caU,0xd753U,0x22c9U,0xc474U,0x1935U,0xe26aU,0x0345U,0x34d3U,
		0x596aU,0x0b1eU,0x808bU,0xe5cfU,0x0574U,0xaef4U,0x1e2dU,0x44abU,0x92c6U,0x8314U,
		0xe610U,0x020bU,0xacbbU,0xbc56U,0x72daU,0x9be0U,0x7392U,0x9fbbU,0x06f1U,0x563aU,
		0xfd89U,0x7f3eU,0x033dU,0xc530U,0x2aaaU,0x3fe7U,0x2e38U,0x122aU,0xa60dU,0x0c04U,
		0x72b6U,0x8234U,0x68adU,0x63ddU,0x0df2U,0x0bfeU,0xed88U,0x75f6U,0xa597U,0x0ff2U,
		0xe0b2U,0x729fU,0xfd92U,0x09deU,0x28e6U,0x260eU,0x8c84U,0x259eU,0xf947U,0xf3bcU,
		0xd1e3U,0x80d4U,0xb44bU,0x63afU,0xa935U,0x2673U,0xf537U,0x48acU,0xae95U,0xdcd8U,
		0x19fdU,0xff59U,0x14c7U,0xd146U,0xe436U,0xf58aU,0x7c2dU,0xc3bfU,0x482aU,0x8772U,
		0xefc5U,0x83e2U,0x82b3U,0xb8e4U,0x1507U,0x8675U,0xdac4U,0x016bU,0x0cf8U,0x6355U,
		0x16a9U,0xc015U,0x8290U,0x2176U,0x86c5U,0x6513U,0xcb80U,0x4901U,0x7e8aU,0x94a2U,
		0x8451U,0x5c4bU,0x1852U,0xed8eU,0x81f3U,0xa545U,0xdf7cU,0xdc84U,0x9101U,0x4f3aU,
		0x879dU,0xf888U,0x6991U,0x5b25U,0x1231U,0xc3f8U,0x8b71U,0xf325U,0x4d9dU,0x9cfeU,
		0x74e7U,0x2a6eU,0xfde8U,0x88f4U,0x9ba6U,0xcc8aU,0x64fdU,0xdf53U,0x2febU,0x0d71U,
		0x02a7U,0xf4eaU,0x195aU,0x7120U,0x4ce6U,0x526bU,0x58dcU,0xa54fU,0x9158U,0xea6fU,
		0x5644U,0xeb7eU,0xc662U,0x4062U,0x4c79U,0xb211U,0x0312U,0x3316U,0x8897U,0xf72fU,
		0x01a9U,0xa3c7U,0xba8aU,0xc209U,0xc75eU,0x6897U,0x7b0cU,0xb0beU,0x3564U,0x1c66U,
		0x9317U,0x4fd5U,0x8997U,0x37ccU,0x8977U,0x029dU,0x75edU,0x75fcU,0x3216U,0x567cU,
		0xa781U,0x6d80U,0x2387U,0x0634U,0x1039U,0x8019U,0x3881U,0x5298U,0xfd53U,0x2ffcU,
		0xfcb7U,0x97e8U,0x9078U,0xf879U,0x967cU,0x997dU,0xbdf6U,0xaf87U,0x8a71U,0x24cbU,
		0x4014U,0xea43U,0xc081U,0x311cU,0x2867U,0xcc3bU,0x015cU,0x6a91U,0xa28dU,0x298eU,
		0x546aU,0xb7afU,0xeb15U,0xc86dU,0x61b0U,0x61e6U,0x6732U,0x935aU,0xa729U,0xfb8dU,
		0x2918U,0x978dU,0xc625U,0x2e11U,0xb9b1U,0x5d2bU,0xd711U,0x7cc8U,0xfeb3U,0x33faU,
		0x837dU,0x8f8bU,0x8300U,0xd8f6U,0x41bdU,0x914bU,0x958fU,0x986aU,0x229aU,0xd87cU,
		0x5d38U,0xa59dU,0xe069U,0x601bU,0xca43U,0xfc66U,0x2224U,0x7982U,0x5c12U,0xa694U,
		0x888aU,0xc3eeU,0xa614U,0x204eU,0x75d4U,0x8b87U,0x922aU,0x58ebU,0xa29aU,0x3477U,
		0x9debU,0x1bc6U,0x2b22U,0x75bfU,0xa63cU,0xd908U,0x4dd3U,0x6645U,0x97f5U,0x86b8U,
		0x30d8U,0x30f3U,0xd394U,0xfdc6U,0xc538U,0x9265U,0xe040U,0x188eU,0x8984U,0x7682U,
		0x8a3aU,0xf64dU,0x0302U,0x98a1U,0x6984U,0x277aU,0xff3fU,0xa1a0U,0xec6bU,0x8b45U,
		0x03ffU,0x5dfaU,0x537dU,0x64a6U,0x81a3U,0x290bU,0x9a6aU,0xc218U,0x9165U,0xc390U,
		0x623bU,0x7d9dU,0x396eU,0x1765U,0x4df1U,0x0ff8U,0x7309U,0x419bU,0x8106U,0x0850U,
		0x8a64U,0x74ddU,0x68a4U,0xb176U,0xd74bU,0x86b0U,0x6e6dU,0x7c9aU,0x067dU,0xcd19U,
		0xb1abU,0xfcfbU,0x8a84U,0x8e0bU,0xc33bU,0xf629U,0x629dU,0x8d2aU,0xc933U,0x906bU,
		0x186dU,0xab54U,0x987aU,0x3014U,0x8200U,0x7339U,0x2efeU,0xb4a2U,0xffa0U,0x6410U,
		0x50d1U,0x63c8U,0x7fd1U,0x00bcU,0x3209U,0xa39aU,0x4101U,0x03acU,0xb99dU,0x983aU,
		0x5557U,0x54f4U,0x9655U,0x9eb8U,0xaaacU,0x2ca9U,0x2a7bU,0x0225U,0x06feU,0x6efcU,
		0x9ccdU,0xe349U,0xada8U,0xef53U,0x88c1U,0xdc4fU,0x1f1aU,0xfad3U,0xa848U,0xe0b8U,
		0x44faU,0xb35eU,0xafe1U,0x578dU,0x23b6U,0xd9d7U,0x4d6eU,0x3dfeU,0xde72U,0x7443U,
		0x0c0bU,0xbac0U,0x7b40U,0xcfecU,0xdd9fU,0x6867U,0xd9b7U,0x216cU,0x6289U,0x81c6U,
		0x0604U,0x3143U,0x2dfaU,0x0771U,0xd308U,0x4ef4U,0xb4c0U,0x29bdU,0xb1f0U,0xa520U,
		0x07c5U,0xdf8bU,0xb458U,0xc6beU,0x71c9U,0x0567U,0x1309U,0xed04U,0xa4aaU,0xf00cU,
		0x1475U,0xec9bU,0x5bf7U,0xb222U,0x79e0U,0x3fb6U,0x877fU,0xf85aU,0xfca5U,0x0f78U,
		0x0ee5U,0x9054U,0xd2d0U,0xa9cdU,0x1d40U,0x5a85U,0x2e9eU,0x0487U,0x4999U,0xb627U,
		0x7e35U,0x2a9dU,0xddc5U,0x72a5U,0xe523U,0x2fbdU,0x14e6U,0xb558U,0xac77U,0xb5c7U,
		0x84cfU,0x352cU,0x4917U,0x5900U,0xa624U,0xd7ddU,0x0de5U,0x1848U,0xc527U,0x8fbbU,
		0xcef7U,0x43f5U,0x0258U,0x1c37U,0x07feU,0xbbf4U,0x8321U,0x9d26U,0xa6faU,0xc94cU,
		0xd48aU,0xfe39U,0x042aU,0x1bfbU,0xc6aaU,0x19f0U,0x4085U,0x637fU,0x802bU,0x2d52U,
		0xfe7fU,0x4341U,0xcad0U,0x8903U,0xd8d7U,0x168bU,0xf4a9U,0x8b1bU,0xfb16U,0x5352U,
		0x0cebU,0x2a0eU,0xe19fU,0x129bU,0x02d6U,0xb589U,0xc1c3U,0x1269U,0x173dU,0xd541U,
		0xd3d5U,0x280fU,0xcf91U,0xffc4U,0x1eb0U,0x4baaU,0x3aadU,0xfd98U,0x9b7eU,0xbd72U,
		0x16c3U,0x1d07U,0xb41aU,0x7525U,0x9731U,0xcb19U,0xdc31U,0x85e6U,0x72bdU,0x7e9bU,
		0xc6d0U,0x69c1U,0x5e94U,0xd732U,0x9a58U,0xeff8U,0xb7c8U,0x93d5U,0xc7caU,0x354cU,
		0x3494U,0xe7b2U,0xa7caU,0xea39U,0xd6e7U,0x6e75U,0x4245U,0x05c7U,0xa2ecU,0xc72dU,
		0x2ab2U,0x5480U,0xdfe6U,0x0dc5U,0x6f1eU,0x94caU,0x3c0dU,0x6c03U,0x1e87U,0x9a84U,
		0xd0abU,0xf8eeU,0x8c19U,0x1b24U,0xc4feU,0x9f3fU,0xe482U,0xaecbU,0xb405U,0xdaeeU,
		0xa0e3U,0xd862U,0xceb1U,0x334bU,0xd6c0U,0x8fc4U,0xe99aU,0x9c7eU,0x3de7U,0xd851U,
		0xa20eU,0x09e4U,0x6d44U,0x446dU,0xb17fU,0x961fU,0x939dU,0x8e36U,0xdcfeU,0x7605U,
		0x68c1U,0xb6f1U,0x6aefU,0xef6aU,0x7f47U,0xa5f0U,0xd680U,0xb484U,0x704fU,0x955eU,
		0x3f02U,0xdbb6U,0xc8aaU,0x0152U,0xefbcU,0xbf43U,0xacb3U,0xec63U,0xc4f6U,0x3329U,
		0xc8caU,0x2a67U,0x3470U,0x2165U,0xeff1U,0x6b72U,0x0b60U,0x67c7U,0x77b4U,0xbb01U,
		0x4e3eU,0x99a4U,0xbf4cU,0xe58bU,0x7fe1U,0xea9fU,0xcfabU,0x5848U,0xbbe3U,0xaf42U,
		0xd8b0U,0xa26cU,0x6751U,0x8825U,0x1117U,0xc06dU,0x98cbU,0x8ce5U,0x87dfU,0x56f8U,
		0x42cdU,0x7b6bU,0x236aU,0x6a23U,0xf39eU,0x28ecU,0x1c0cU,0xead1U,0x68e3U,0xd939U,
		0x4c2bU,0x31dbU,0xce01U,0x01ceU,0x409eU,0xea20U,0x823bU,0x2b85U,0x875dU,0xaccfU,
		0x1624U,0xb7f0U,0xc315U,0x6b79U,0xf40dU,0x6447U,0x3cc6U,0xee6fU,0xc749U,0xf71dU,
		0x32f0U,0x4b51U,0x54d1U,0x4e55U,0x1fe6U,0x693bU,0xf3cfU,0x9818U,0x4b31U,0xfb83U,
		0x9198U,0x37a7U,0x4270U,0x7fb0U,0x5836U,0x2f1dU,0xbe71U,0x4d0fU,0xa1c5U,0x5232U,
		0xe540U,0x33d6U,0x3c80U,0xd1f2U,0xc167U,0xbf0dU,0xfa3aU,0x4eb9U,0x444aU,0xab20U,
		0x45b7U,0x487fU,0xf306U,0x4044U,0x8d3aU,0x028aU,0xf29eU,0xa136U,0x43adU,0x2614U,
		0x460bU,0xb78dU,0xe366U,0xda0bU,0xb826U,0xbfa2U,0x36a1U,0x9ef2U,0x4d2bU,0x6367U,
		0x4dbeU,0x0f57U,0x52bbU,0xf2b6U,0xf068U,0x4957U,0xe2f0U,0xd3eaU,0x2b4bU,0xe01bU,
		0x61fcU,0x8918U,0x7c44U,0xc3ceU,0x4a51U,0x3f45U,0xb289U,0xc362U,0x17feU,0xfea9U,
		0xc00cU,0x881cU,0x1be6U,0xc4cbU,0x0e33U,0x1ab2U,0x5d60U,0x8605U,0x1effU,0x26b7U,
		0xee27U,0xc460U,0xf1a4U,0xce66U,0xebeeU,0x5312U,0x1a96U,0xc267U,0xdb13U,0xa4ccU,
		0xad42U,0x0ea0U,0xc0d6U,0xc48fU,0xa359U,0x104dU,0xbcd0U,0x863bU,0x8747U,0xcea4U,
		0xc039U,0x39c0U,0x1acbU,0x8c8aU,0x5698U,0xb663U,0x9a91U,0x4572U,0xc9b7U,0xe1eaU,
		0xdb44U,0x44dbU,0x3c2fU,0x9644U,0x68b4U,0xb163U,0xa7d7U,0x41e5U,0x023fU,0xb6dbU,
		0xc5e3U,0x0b67U,0x8d80U,0x8864U,0x948aU,0x7f58U,0x7ea8U,0xdef9U,0x4158U,0x0b53U,
		0xf057U,0x7ca2U,0xf8d0U,0xd69fU,0xa6ddU,0x15a4U,0x3fefU,0x33a7U,0x3576U,0x9f7fU,
		0x57a0U,0xc411U,0x90b1U,0xb3e2U,0xcf9eU,0x6d76U,0xe395U,0x86e1U,0x6e31U,0x64beU,
		0x3331U,0x4689U,0xa343U,0x67deU,0x0ad1U,0xcfbeU,0xc8b7U,0xd593U,0x288aU,0x5e87U,
		0x66adU,0x4a7bU,0x6f62U,0x77c9U,0x92b0U,0x719bU,0x06b4U,0x25a4U,0xe0d3U,0x7487U,
		0xa8d7U,0x6a02U,0x046cU,0x2443U,0x9b71U,0xb092U,0x3cfaU,0x5d15U,0xbb0eU,0x9f59U,
		0xe56fU,0x21eaU,0xd5d3U,0x0f28U,0xa578U,0xcdc4U,0xa02cU,0x178aU,0x61cdU,0xd09eU,
		0xbe5fU,0xd261U,0x3b9aU,0x4d51U,0x90deU,0x4939U,0xd4c2U,0x5610U,0xedcdU,0x6664U,
		0xcebeU,0x1db1U,0xd62eU,0x8d14U,0xc7bdU,0x930dU,0xb442U,0xefb3U,0xe5c7U,0x767aU,
		0x94c2U,0xc409U,0xd15dU,0x3383U,0x7612U,0xdc64U,0x73aaU,0x8a20U,0xef04U,0xc97eU,
		0x4537U,0x80acU,0xbd84U,0x44baU,0x68f0U,0x5749U,0x0377U,0x971aU,0xb0fcU,0x8bfdU,
		0xa223U,0xf94dU,0xe453U,0xe463U,0x5070U,0x1ef2U,0x3bd9U,0x5fa5U,0x6454U,0xd658U,
		0x378eU,0x8f42U,0x63e4U,0x53e4U,0xe3ddU,0xf2e2U,0xb0e5U,0x1dccU,0x2bcfU,0x67ceU,
		0x4a44U,0x20abU,0xba73U,0x1537U,0xc5b8U,0xf992U,0xc0f9U,0xd2a2U,0xc228U,0xae25U,
		0x4b3eU,0xccbeU,0xfe5bU,0x4bf4U,0x190bU,0x2b3eU,0xc4bbU,0x814eU,0xecdbU,0x10b6U,
		0x3da0U,0x67f6U,0x227dU,0x59afU,0x0f8dU,0xfd69U,0xe77bU,0xa1faU,0x86f2U,0x0c24U,
		0xf291U,0x97deU,0xbf1aU,0x954eU,0xebaaU,0xec34U,0x6c66U,0x0d46U,0x1711U,0x6dc0U,
		0xd525U,0x580fU,0x418bU,0xc2fdU,0x4964U,0xf3c2U,0x577bU,0x7b57U,0x02bbU,0x7f6eU,
		0x2457U,0x19a9U,0xfa9cU,0x8aedU,0xb1d8U,0x5a34U,0xda26U,0x26daU,0x5c73U,0x128bU,
		0xc15aU,0x0f01U,0x43baU,0x69f0U,0x2416U,0xf0b7U,0x435dU,0x6c2cU,0xcbfbU,0x5c24U,
		0x6c85U,0x23f5U,0x531fU,0x8c4dU,0x13beU,0x42c0U,0x34c3U,0x5102U,0xfb27U,0x8e51U,
		0x451eU,0x0fabU,0x180bU,0x84f5U,0x4c3eU,0xba58U,0xc836U,0x35b4U,0xe49aU,0x7365U,
		0x4c1eU,0x102fU,0xdccfU,0x2a5eU,0x51ceU,0x3027U,0x9905U,0x6276U,0x000bU,0x2431U,
		0x5ac0U,0xb136U,0x1389U,0x1babU,0x852aU,0xf1e1U,0x0dd9U,0xb993U,0xa3beU,0xd1e9U,
		0x6656U,0xec18U,0x9e8aU,0xaaeaU,0xd9c5U,0x0e8cU,0xb709U,0x46b0U,0x5b4eU,0x7c13U,
		0x0df9U,0xa15bU,0x78f4U,0x3520U,0xa655U,0x3593U,0x89f1U,0x77bbU,0x3e31U,0xf76eU,
		0x0518U,0xd5ecU,0x3744U,0x2de7U,0x157aU,0x1ddfU,0x8db6U,0x0685U,0xf60fU,0x4e08U,
		0x14daU,0x570eU,0x2e40U,0x3c55U,0xb3d4U,0xe5afU,0x1a79U,0xd165U,0x06d9U,0xb686U,
		0xbce2U,0xa434U,0x4769U,0x500cU,0xd578U,0x1afeU,0xc8d2U,0x047bU,0xb36cU,0x12bdU,
		0xf67dU,0x8856U,0xad65U,0xf6d5U,0xdcb3U,0xadefU,0xaa4aU,0xcf00U,0x2e21U,0x350dU,
		0x0ef3U,0x19caU,0x96e0U,0x80daU,0xd8cbU,0x2c9fU,0xd754U,0x2f7bU,0x22caU,0x4129U,
		0xc475U,0x321fU,0x1936U,0x477aU,0xe26bU,0x3d12U,0x0346U,0x7357U,0x34d4U,0x84b8U,
		0x596bU,0xb57dU,0x0b1fU,0x0a52U,0x808cU,0xb272U,0xe5d0U,0xd5a2U,0x0575U,0x3bf9U,
		0xaef5U,0x2babU,0x1e2eU,0x51d7U,0x44acU,0xd445U,0x92c7U,0xd0bbU,0x8315U,0xdfccU,
		0xe611U,0x1362U,0x020cU,0xdb63U,0xacbcU,0x4ca8U,0xbc57U,0x7c5fU,0x72dbU,0xe894U,
		0x9be1U,0x8140U,0x7393U,0x6e10U,0x9fbcU,0x1befU,0x06f2U,0x6293U,0x563bU,0x4267U,
		0xfd8aU,0x0675U,0x7f3fU,0xe567U,0x033eU,0x58d5U,0xc531U,0xb451U,0x2aabU,0x4b2aU,
		0x3fe8U,0x1904U,0x2e39U,0xe0acU,0x122bU,0xfd4dU,0xa60eU,0x6e67U,0x0c05U,0xa61eU,
		0x72b7U,0x6ae9U,0x8235U,0x4db8U,0x68aeU,0xbb08U,0x63deU,0xda20U,0x0df3U,0xe265U,
		0x0bffU,0xff54U,0xed89U,0x5266U,0x75f7U,0x2989U,0xa598U,0xfdc1U,0x0ff3U,0xb49dU,
		0xe0b3U,0xdf86U,0x72a0U,0x19ebU,0xfd93U,0xc728U,0x09dfU,0x2160U,0x28e7U,0x9813U,
		0x260fU,0x8817U,0x8c85U,0x159fU,0x259fU,0x665fU,0xf948U,0x4befU,0xf3bdU,0x50fdU,
		0xd1e4U,0x5709U,0x80d5U,0xd440U,0xb44cU,0xb498U,0x63b0U,0x5999U,0xa936U,0x039bU,
		0x2674U,0x63b5U,0xf538U,0xcddeU,0x48adU,0xe39fU,0xae96U,0x2b77U,0xdcd9U,0xd270U,
		0x19feU,0x29b1U,0xff5aU,0x67a6U,0x14c8U,0xbf24U,0xd147U,0xb46cU,0xe437U,0x6e85U,
		0xf58bU,0x1b6aU,0x7c2eU,0xc88dU,0xc3c0U,0xd6f1U,0x482bU,0xbb4dU,0x8773U,0xcb93U,
		0xefc6U,0x6fe0U,0x83e3U,0x599eU,0x82b4U,0xb4dbU,0xb8e5U,0x945dU,0x1508U,0x4309U,
		0x8676U,0xd9fdU,0xdac5U,0x91a2U,0x016cU,0xecf8U,0x0cf9U,0x4d3dU,0x6356U,0x8e79U,
		0x16aaU,0x0dafU,0xc016U,0x8ecfU,0x8291U,0xd978U,0x2177U,0x864aU,0x86c6U,0xd207U,
		0x6514U,0xd80aU,0xcb81U,0x905fU,0x4902U,0xa93bU,0x7e8bU,0xdc41U,0x94a3U,0x47f6U,
		0x8452U,0x072dU,0x5c4cU,0xdfb7U,0x1853U,0xd790U,0xed8fU,0x781eU,0x81f4U,0xb1a1U,
		0xa546U,0xfb7aU,0xdf7dU,0x4e45U,0xdc85U,0x2a19U,0x9102U,0x4a1fU,0x4f3bU,0x1f62U,
		0x879eU,0x2785U,0xf889U,0xcd21U,0x6992U,0x0beaU,0x5b26U,0x03a0U,0x1232U,0x7896U,
		0xc3f9U,0x1c29U,0x8b72U,0x338bU,0xf326U,0xa3a5U,0x4d9eU,0x0696U,0x9cffU,0xa8aaU,
		0x74e8U,0x731eU,0x2a6fU,0x9c07U,0xfde9U,0xee87U,0x88f5U,0x57ffU,0x9ba7U,0x9179U,
		0xcc8bU,0x3cddU,0x64feU,0xed47U,0xdf54U,0x2999U,0x2fecU,0x276cU,0x0d72U,0x2679U,
		0x02a8U,0x0844U,0xf4ebU,0xfcdaU,0x195bU,0xea7aU,0x7121U,0x85ccU,0x4ce7U,0x759dU,
		0x526cU,0xde40U,0x58ddU,0x489cU,0xa550U,0x3c3eU,0x9159U,0x74d3U,0xea70U,0x2f31U,
		0x5645U,0x900aU,0xeb7fU,0x4113U,0xc663U,0xa0b4U,0x4063U,0x7238U,0x4c7aU,0x3e5dU,
		0xb212U,0x63baU,0x0313U,0x79feU,0x3317U,0xcd0dU,0x8898U,0x613aU,0xf730U,0x1672U,
		0x01aaU,0x3667U,0xa3c8U,0xf561U,0xba8bU,0x97b5U,0xc20aU,0x985aU,0xc75fU,0xd75eU,
		0x6898U,0xf4d8U,0x7b0dU,0x8b39U,0xb0bfU,0xb09dU,0x3565U,0xcc2dU,0x1c67U,0x4bdbU,
		0x9318U,0xc2e6U,0x4fd6U,0xf53dU,0x8998U,0x9e1fU,0x37cdU,0xa44eU,0x8978U,0x72ffU,
		0x029eU,0x54c8U,0x75eeU,0xb16eU,0x75fdU,0xb08aU,0x3217U,0x85c4U,0x567dU,0x15dfU,
		0xa782U,0x99abU,0x6d81U,0x3835U,0x2388U,0x40afU,0x0635U,0x9259U,0x103aU,0x941cU,
		0x801aU,0x6889U,0x3882U,0x0af4U,0x5299U,0xcde3U,0xfd54U,0x3179U,0x2ffdU,0x0f0cU,
		0xfcb8U,0x4ecfU,0x97e9U,0x7cb1U,0x9079U,0x957bU,0xf87aU,0xdf6bU,0x967dU,0x20b6U,
		0x997eU,0x17cbU,0xbdf7U,0x281fU,0xaf88U,0x09f7U,0x8a72U,0x7291U,0x24ccU,0x974eU,
		0x4015U,0x2042U,0xea44U,0xd37cU,0xc082U,0x3984U,0x311dU,0x48b2U,0x2868U,0xffcfU,
		0xcc3cU,0x0c36U,0x015dU,0x2bbdU,0x6a92U,0x0b88U,0xa28eU,0x7d93U,0x298fU,0x3642U,
		0x546bU,0x7069U,0xb7b0U,0x3356U,0xeb16U,0xe75eU,0xc86eU,0x9dccU,0x61b1U,0xd944U,
		0x61e7U,0x3e22U,0x6733U,0x78d2U,0x935bU,0x8f7fU,0xa72aU,0xaa68U,0xfb8eU,0xe3a4U,
		0x2919U,0xd0e2U,0x978eU,0x2c1bU,0xc626U,0xa98fU,0x2e12U,0xe5fdU,0xb9b2U,0x724dU,
		0x5d2cU,0x261fU,0xd712U,0x6114U,0x7cc9U,0x62dcU,0xfeb4U,0x87c6U,0x33fbU,0x1ed0U,
		0x837eU,0x4e66U,0x8f8cU,0xd423U,0x8301U,0x96fcU,0xd8f7U,0x03e2U,0x41beU,0xc17fU,
		0x914cU,0xae9bU,0x9590U,0xd31dU,0x986bU,0xe9c9U,0x229bU,0xdec1U,0xd87dU,0x5111U,
		0x5d39U,0xe2dbU,0xa59eU,0x7f2dU,0xe06aU,0x51acU,0x601cU,0x0d65U,0xca44U,0xbf53U,
		0xfc67U,0x536eU,0x2225U,0x5590U,0x7983U,0x2022U,0x5c13U,0x5fd1U,0xa695U,0xb1b3U,
		0x888bU,0xe036U,0xc3efU,0x2b7cU,0xa615U,0x98edU,0x204fU,0x7510U,0x75d5U,0x761aU,
		0x8b88U,0x36e0U,0x922bU,0x8cd4U,0x58ecU,0x2d1aU,0xa29bU,0x4056U,0x3478U,0xaba4U,
		0x9decU,0xafd2U,0x1bc7U,0x47a1U,0x2b23U,0x9ca9U,0x75c0U,0x96a1U,0xa63dU,0x43feU,
		0xd909U,0x7086U,0x4dd4U,0xee3aU,0x6646U,0xdcdeU,0x97f6U,0x60faU,0x86b9U,0x9c9eU,
		0x30d9U,0x7bdcU,0x30f4U,0x5685U,0xd395U,0x2dbeU,0xfdc7U,0xaea6U,0xc539U,0xbaa9U,
		0x9266U,0xf938U,0xe041U,0xb964U,0x188fU,0x5c3fU,0x8985U,0x6f4dU,0x7683U,0xf706U,
		0x8a3bU,0xa96dU,0xf64eU,0xe957U,0x0303U,0x3030U,0x98a2U,0xd275U,0x6985U,0x663bU,
		0x277bU,0xabf4U,0xff40U,0xcf7fU,0xa1a1U,0x1a1aU,0xec6cU,0x69a5U,0x8b46U,0x163bU,
		0x0400U,0xeb06U,0x5dfbU,0xc781U,0x537eU,0x88e8U,0x64a7U,0xc4d4U,0x81a4U,0x49e0U,
		0x290cU,0xc776U,0x9a6bU,0x262aU,0xc219U,0x6427U,0x9166U,0x4815U,0xc391U,0x1a03U,
		0x623cU,0x70fdU,0x7d9eU,0x2cf4U,0x396fU,0x21baU,0x1766U,0x7cbcU,0x4df2U,0x5b5fU,
		0x0ff9U,0x98f6U,0x730aU,0xac5aU,0x419cU,0xdb98U,0x8107U,0xe592U,0x0851U,0xa746U,
		0x8a65U,0xff35U,0x74deU,0xc30cU,0x68a5U,0x80feU,0xb177U,0xa8cfU,0xd74cU,0xf4e3U,
		0x86b1U,0x29b6U,0x6e6eU,0x4e4eU,0x7c9bU,0xae1eU,0x067eU,0x2982U,0xcd1aU,0x6882U,
		0xb1acU,0xa8c8U,0xfcfcU,0x2d2dU,0x8a85U,0xde06U,0x8e0cU,0xec21U,0xc33cU,0xe31fU,
		0xf62aU,0x9c12U,0x629eU,0xa463U,0x8d2bU,0x9a1dU,0xc934U,0xf505U,0x906cU,0x6630U,
		0x186eU,0x5e39U,0xab55U,0xff5fU,0x987bU,0x390bU,0x3015U,0x517fU,0x8201U,0x57daU,
		0x733aU,0xf3feU,0x2effU,0x8955U,0xb4a3U,0x0628U,0xffa1U,0xbf8aU,0x6411U,0x9a0cU,
		0x50d2U,0xcb9eU,0x63c9U,0x5de8U,0x7fd2U,0x9d71U,0x00bdU,0x7e1aU,0x320aU,0xf643U,
		0xa39bU,0x5719U,0x4102U,0xa3e4U,0x03adU,0x67abU,0xb99eU,0x763fU,0x983bU,0xcb61U,
		0x5558U,0xdfd7U,0x54f5U,0x4ff1U,0x9656U,0x37c0U,0x9eb9U,0x3f76U,0xaaadU,0xb125U,
		0x2caaU,0x5c82U,0x2a7cU,0x0415U,0x0226U,0x78acU,0x06ffU,0xd05aU,0x6efdU,0xb91cU,
		0x9cceU,0xbcffU,0xe34aU,0x6687U,0xada9U,0x76a4U,0xef54U,0x14cdU,0x88c2U,0x8422U,
		0xdc50U,0xfe6dU,0x1f1bU,0xbc69U,0xfad4U,0x47c0U,0xa849U,0xe871U,0xe0b9U,0x160eU,
		0x44fbU,0x9021U,0xb35fU,0xe9beU,0xafe2U,0x7fe8U,0x578eU,0x861bU,0x23b7U,0xbc1aU,
		0xd9d8U,0x4ad3U,0x4d6fU,0x8589U,0x3dffU,0xfd03U,0xde73U,0x851dU,0x7444U,0xbf29U,
		0x0c0cU,0xc0c0U,0xbac1U,0x0100U,0x7b41U,0xdc6cU,0xcfedU,0x6057U,0xdda0U,0xbdc1U,
		0x6868U,0x8686U,0xd9b8U,0x82f6U,0x216dU,0xc064U,0x628aU,0x49d7U,0x81c7U,0xfb5dU,
		0x0605U,0xef38U,0x3144U,0x6cdbU,0x2dfbU,0x0a3aU,0x0772U,0x7e52U,0xd309U,0xc444U,
		0x4ef5U,0xd14cU,0xb4c1U,0x8180U,0x29beU,0x01ffU,0xb1f1U,0x9bc6U,0xa521U,0x15e7U,
		0x07c6U,0x113eU,0xdf8cU,0x7948U,0xb459U,0x1e17U,0xc6bfU,0x62f1U,0x71caU,0xe7ceU,
		0x0568U,0x51a1U,0x130aU,0xa1d6U,0xed05U,0x71b3U,0xa4abU,0x7834U,0xf00dU,0x95a5U,
		0x1476U,0x79cfU,0xec9cU,0xb471U,0x5bf8U,0x0b12U,0xb223U,0xfa06U,0x79e1U,0x24b2U,
		0x3fb7U,0x815dU,0x8780U,0x60adU,0xf85bU,0x120bU,0xfca6U,0xd111U,0x0f79U,0x2838U,
		0x0ee6U,0xd872U,0x9055U,0x2522U,0xd2d1U,0x9379U,0xa9ceU,0x8228U,0x1d41U,0x4ea0U,
		0x5a86U,0xd701U,0x2e9fU,0x3d26U,0x0488U,0xe43cU,0x499aU,0xe4e3U,0xb628U,0x18b3U,
		0x7e36U,0x9af4U,0x2a9eU,0x2017U,0xddc6U,0xb754U,0x72a6U,0xac8cU,0xe524U,0x7a43U,
		0x2fbeU,0x2bdeU,0x14e7U,0xeaa6U,0xb559U,0xbb6aU,0xac78U,0x2602U,0xb5c8U,0xab66U,
		0x84d0U,0xd9a7U,0x352dU,0x2d34U,0x4918U,0x95f3U,0x5901U,0x6e8aU,0xa625U,0x79abU,
		0xd7deU,0x5bd0U,0x0de6U,0x8880U,0x1849U,0xf395U,0xc528U,0xc388U,0x8fbcU,0x4a73U,
		0xcef8U,0xed3fU,0x43f6U,0x0a32U,0x0259U,0xde18U,0x1c38U,0xabdbU,0x07ffU,0x1074U,
		0xbbf5U,0xe472U,0x8322U,0x05ecU,0x9d27U,0x80c8U,0xa6fbU,0xcd7eU,0xc94dU,0xf590U,
		0xd48bU,0x93adU,0xfe3aU,0xf6baU,0x042bU,0x2f95U,0x1bfcU,0x416bU,0xc6abU,0xf0d7U,
		0x19f1U,0x8cc9U,0x4086U,0x13edU,0x6380U,0x503dU,0x802cU,0x8ba4U,0x2d53U,0x0ecaU,
		0xfe80U,0xa568U,0x4342U,0xb02aU,0xcad1U,0xb8d8U,0x8904U,0xc0adU,0xd8d8U,0x1010U,
		0x168cU,0x1b6fU,0xf4aaU,0x7b82U,0x8b1cU,0x0d0cU,0xfb17U,0xf961U,0x5353U,0x9d9dU,
		0x0cecU,0x1bbcU,0x2a0fU,0xbd3eU,0xe1a0U,0xbed6U,0x129cU,0x2af6U,0x02d7U,0x59d3U,
		0xb58aU,0x3bc0U,0xc1c4U,0x1cc5U,0x126aU,0x1c7aU,0x173eU,0x46e4U,0xd542U,0xb5e1U,
		0xd3d6U,0x6044U,0x2810U,0x7c33U,0xcf92U,0x2fb3U,0xffc5U,0x6578U,0x1eb1U,0xffeaU,
		0x4babU,0x484dU,0x3aaeU,0xd18fU,0xfd99U,0x680fU,0x9b7fU,0x1437U,0xbd73U,0x78c3U,
		0x16c4U,0xcfb2U,0x1d08U,0x0e4aU,0xb41bU,0xb837U,0x7526U,0x3a3cU,0x9732U,0x3297U,
		0xcb1aU,0x8a8cU,0xdc32U,0xaa38U,0x85e7U,0xc892U,0x72beU,0x0923U,0x7e9cU,0x9ae9U,
		0xc6d1U,0x73b2U,0x69c2U,0xf8fdU,0x5e95U,0x4f70U,0xd733U,0xeec7U,0x9a59U,0x758eU,
		0xeff9U,0xe81fU,0xb7c9U,0xeae1U,0x93d6U,0xd693U,0xc7cbU,0x6176U,0x354dU,0x0c61U,
		0x3495U,0x0261U,0xe7b3U,0x5d9bU,0xa7cbU,0x498fU,0xea3aU,0xc3c5U,0xd6e8U,0x7e11U,
		0x6e76U,0xf2f4U,0x4246U,0xcf12U,0x05c8U,0xa78aU,0xa2edU,0x70afU,0xc72eU,0xc909U,
		0x2ab3U,0x44cfU,0x5481U,0x6916U,0xdfe7U,0x5d8dU,0x0dc6U,0x0666U,0x6f1fU,0x42e5U,
		0x94cbU,0x9e6dU,0x3c0eU,0xefdaU,0x6c04U,0x4324U,0x1e88U,0x990eU,0x9a85U,0xd6f6U,
		0xd0acU,0x2357U,0xf8efU,0x589aU,0x8c1aU,0xccd8U,0x1b25U,0xb93bU,0xc4ffU,0x91cfU,
		0x9f40U,0xf68dU,0xe483U,0x53c4U,0xaeccU,0x862fU,0xb406U,0xa425U,0xdaefU,0x0e3cU,
		0xa0e4U,0x81d0U,0xd863U,0x85f5U,0xceb2U,0x8fb1U,0x334cU,0x2510U,0xd6c1U,0x549eU,
		0x8fc5U,0x4830U,0xe99bU,0xc325U,0x9c7fU,0xecbcU,0x3de8U,0x66f2U,0xd852U,0x857aU,
		0xa20fU,0x492dU,0x09e5U,0x2058U,0x6d45U,0x7a9fU,0x446eU,0xe173U,0xb180U,0x584fU,
		0x9620U,0x459dU,0x939eU,0x5f95U,0x8e37U,0x428fU,0xdcffU,0x256aU,0x7606U,0xde0dU,
		0x68c2U,0x0a67U,0xb6f2U,0xbb52U,0x6af0U,0xdc8eU,0xef6bU,0xe82dU,0x7f48U,0xb745U,
		0xa5f1U,0x4afaU,0xd681U,0xd5b3U,0xb485U,0x9301U,0x7050U,0xf21eU,0x955fU,0x9e93U,
		0x3f03U,0xc114U,0xdbb7U,0x49bdU,0xc8abU,0xac6dU,0x0153U,0x17f5U,0xefbdU,0x40f9U,
		0xbf44U,0xe5bfU,0xacb4U,0x01a2U,0xec64U,0x8778U,0xc4f7U,0x3c79U,0x332aU,0x86ebU,
		0xc8cbU,0x2598U,0x2a68U,0x9977U,0x3471U,0x8e05U,0x2166U,0x43efU,0xeff2U,0x9558U,
		0x6b73U,0xd3e4U,0x0b61U,0xcdbeU,0x67c8U,0x69eaU,0x77b5U,0xb577U,0xbb02U,0x6e7fU,
		0x4e3fU,0x74cdU,0x99a5U,0xe758U,0xbf4dU,0xb95eU,0xe58cU,0xcb98U,0x7fe2U,0xe7c8U,
		0xeaa0U,0x8b9eU,0xcfacU,0x5d87U,0x5849U,0xcdb8U,0xbbe4U,0x37b1U,0xaf43U,0x1445U,
		0xd8b1U,0x345aU,0xa26dU,0x6da2U,0x6752U,0xc3a3U,0x8826U,0xd1acU,0x1118U,0xf5b4U,
		0xc06eU,0xed1dU,0x98ccU,0x119fU,0x8ce6U,0x3e83U,0x87e0U,0x252cU,0x56f9U,0xefcbU,
		0x42ceU,0x1ee6U,0x7b6cU,0x8041U,0x236bU,0x32c4U,0x6a24U,0x23a1U,0xf39fU,0x5515U,
		0x28edU,0x8a01U,0x1c0dU,0x5948U,0xead2U,0x1352U,0x68e4U,0xbbeaU,0xd93aU,0xbd48U,
		0x4c2cU,0x7a6bU,0x31dcU,0xe685U,0xce02U,0xabfeU,0x01cfU,0x8e13U,0x409fU,0x8df6U,
		0xea21U,0x6fe5U,0x823cU,0x442aU,0x2b86U,0xf2d6U,0x875eU,0x8a28U,0xacd0U,0x5723U,
		0x1625U,0xe6b8U,0xb7f1U,0x796aU,0xc316U,0x071dU,0x6b7aU,0x2c4cU,0xf40eU,0xeb48U,
		0x6448U,0xcd73U,0x3cc7U,0xab31U,0xee70U,0x7279U,0xc74aU,0xde20U,0xf71eU,0x364cU,
		0x32f1U,0xf986U,0x4b52U,0x83e8U,0x54d2U,0x8409U,0x4e56U,0x1190U,0x1fe7U,0x2f3bU,
		0x693cU,0x99b3U,0xf3d0U,0xf0a4U,0x9819U,0x1837U,0x4b32U,0xc6a0U,0xfb84U,0xc457U,
		0x9199U,0x37b7U,0x37a8U,0x2972U,0x4271U,0x7142U,0x7fb1U,0x1f96U,0x5837U,0x2000U,
		0x2f1eU,0x5b38U,0xbe72U,0x2309U,0x4d10U,0x59a3U,0xa1c6U,0xba34U,0x5233U,0xae48U,
		0xe541U,0xd7a7U,0x33d7U,0xc7eeU,0x3c81U,0x41d4U,0xd1f3U,0x69d7U,0xc168U,0xf1b6U,
		0xbf0eU,0x0ebfU,0xfa3bU,0x2f85U,0x4ebaU,0x1453U,0x444bU,0x00d4U,0xab21U,0xe330U,
		0x45b8U,0x26ceU,0x4880U,0x43ceU,0xf307U,0x565dU,0x4045U,0x82b9U,0x8d3bU,0x5e70U,
		0x028bU,0x0abcU,0xf29fU,0x9e3fU,0xa137U,0xac1aU,0x43aeU,0x8b11U,0x2615U,0x6700U,
		0x460cU,0xe27eU,0xb78eU,0x1e06U,0xe367U,0xaf49U,0xda0cU,0x4b98U,0xb827U,0x4aafU,
		0xbfa3U,0x52f8U,0x36a2U,0xa236U,0x9ef3U,0xec28U,0x4d2cU,0x77fcU,0x6368U,0xb4e0U,
		0x4dbfU,0x5eb9U,0x0f58U,0x25d5U,0x52bcU,0x66e2U,0xf2b7U,0x10f5U,0xf069U,0xba43U,
		0x4958U,0xb8cdU,0xe2f1U,0x85b3U,0xd3ebU,0xa0f2U,0x2b4cU,0xdb31U,0xe01cU,0x1ce0U,
		0x61fdU,0x2a52U,0x8919U,0x737aU,0x7c45U,0x4fb8U,0xc3cfU,0x83c4U,0x4a52U,0xeb63U,
		0x3f46U,0xb8eaU,0xb28aU,0x6582U,0xc363U,0x4b08U,0x17ffU,0xe195U,0xfeaaU,0xad39U,
		0xc00dU,0x021dU,0x881dU,0xc976U,0x1be7U,0xf4d0U,0xc4ccU,0x251aU,0x0e34U,0x144bU,
		0x1ab3U,0x73ccU,0x5d61U,0xbd52U,0x8606U,0x8e45U,0x1f00U,0xa10eU,0x26b8U,0xa2bfU,
		0xee28U,0xb6a5U,0xc461U,0x9462U,0xf1a5U,0xfa4aU,0xce67U,0xcf2bU,0xebefU,0xac2aU,
		0x5313U,0x9d92U,0x1a97U,0x7703U,0xc268U,0x670aU,0xdb14U,0x58c3U,0xa4cdU,0x32dbU,
		0xad43U,0x1426U,0x0ea1U,0x001fU,0xc0d7U,0x7744U,0xc490U,0x2c00U,0xa35aU,0x834bU,
		0x104eU,0xd455U,0xbcd1U,0x2723U,0x863cU,0x150dU,0x8748U,0xebc8U,0xcea5U,0xbc0fU,
		0xc03aU,0x1410U,0x39c1U,0x04fcU,0x1accU,0x4abeU,0x8c8bU,0x6fb6U,0x5699U,0x4cd6U,
		0xb664U,0x1c1bU,0x9a92U,0xd8b7U,0x4573U,0xd845U,0xc9b8U,0x24ecU,0xe1ebU,0x7dafU,
		0xdb45U,0x5ee1U,0x44dcU,0xc343U,0x3c30U,0xbff4U,0x9645U,0x430eU,0x68b5U,0x7439U,
		0xb164U,0x8b07U,0xa7d8U,0xef0cU,0x41e6U,0x55eeU,0x0240U,0xe1c4U,0xb6dcU,0x253aU,
		0xc5e4U,0xf553U,0x0b68U,0xddb5U,0x8d81U,0xdbaaU,0x8865U,0xd646U,0x948bU,0x5394U,
		0x7f59U,0x6b88U,0x7ea9U,0x1c40U,0xdefaU,0xcb3dU,0x4159U,0x77e2U,0x0b54U,0x867bU,
		0xf058U,0x52cbU,0x7ca3U,0xc641U,0xf8d1U,0x7713U,0xd6a0U,0x6d89U,0xa6deU,0x2695U,
		0x15a5U,0xf177U,0x3ff0U,0xdfa5U,0x33a8U,0xec57U,0x3577U,0x3460U,0x9f80U,0x3827U,
		0x57a1U,0xddf1U,0xc412U,0xb3a8U,0x90b2U,0x1a38U,0xb3e3U,0x05b0U,0xcf9fU,0x627fU,
		0x6d77U,0xda02U,0xe396U,0x5176U,0x86e2U,0x6ff3U,0x6e32U,0x3b09U,0x64bfU,0x7782U,
		0x3332U,0xf6e6U,0x468aU,0x6b17U,0xa344U,0x9998U,0x67dfU,0x8568U,0x0ad2U,0xd36eU,
		0xcfbfU,0x1abbU,0xc8b8U,0xfb66U,0xd594U,0x30b7U,0x288bU,0xa504U,0x5e88U,0x8417U,
		0x66aeU,0x9e0aU,0x4a7cU,0xdacaU,0x6f63U,0xe24fU,0x77caU,0xb10eU,0x92b1U,0x1948U,
		0x719cU,0x2f6dU,0x06b5U,0x4b6cU,0x25a5U,0x7519U,0xe0d4U,0xb605U,0x7488U,0xbc8eU,
		0xa8d8U,0xa273U,0x6a03U,0x825bU,0x046dU,0x4381U,0x2444U,0x2ebaU,0x9b72U,0xe33fU,
		0xb093U,0xe326U,0x3cfbU,0xe859U,0x5d16U,0x91a7U,0xbb0fU,0x2a22U,0x9f5aU,0x08f8U,
		0xe570U,0xa14dU,0x21ebU,0x9b92U,0xd5d4U,0xffb8U,0x0f29U,0x9bf1U,0xa579U,0x7c7bU,
		0xcdc5U,0xaaf3U,0xa02dU,0xd776U,0x178bU,0x35feU,0x61ceU,0x520fU,0xd09fU,0xe9b3U,
		0xbe60U,0x31a7U,0xd262U,0xd132U,0x3b9bU,0x1a4eU,0x4d52U,0x0171U,0x90dfU,0xba15U,
		0x493aU,0x6ffcU,0xd4c3U,0xa180U,0x5611U,0xaebfU,0xedceU,0xb1c7U,0x6665U,0x8ec1U,
		0xcebfU,0x73ddU,0x1db2U,0xbbbaU,0xd62fU,0x6da8U,0x8d15U,0xae56U,0xc7beU,0x1f10U,
		0x930eU,0xf2fdU,0xb443U,0xc92bU,0xefb4U,0xbe57U,0xe5c8U,0xeb77U,0x767bU,0xecfdU,
		0x94c3U,0x7fa9U,0xc40aU,0x10afU,0xd15eU,0xc721U,0x3384U,0x4ec8U,0x7613U,0x297bU,
		0xdc65U,0x8879U,0x73abU,0xb73eU,0x8a21U,0x66dbU,0xef05U,0xa146U,0xc97fU,0x257cU,
		0x4538U,0xa009U,0x80adU,0x5146U,0xbd85U,0x9ab1U,0x44bbU,0x3af6U,0x68f1U,0x5a25U,
		0x574aU,0x0cfeU,0x0378U,0x7ec4U,0x971bU,0x5306U,0xb0fdU,0xbe21U,0x8bfeU,0xe005U,
		0xa224U,0x22dbU,0xf94eU,0x21d6U,0xe454U,0xf7ddU,0xe464U,0x9a37U,0x5071U,0x6758U,
		0x1ef3U,0xd106U,0x3bdaU,0xdd80U,0x5fa6U,0xf2c5U,0x6455U,0x00a9U,0xd659U,0xf631U,
		0x378fU,0x4a65U,0x8f43U,0x4d42U,0x63e5U,0x1aa6U,0x53e5U,0xf51cU,0xe3deU,0xc986U,
		0xf2e3U,0x3b3dU,0xb0e6U,0x684fU,0x1dcdU,0xa6a8U,0x2bd0U,0xe208U,0x67cfU,0x36bbU,
		0x4a45U,0x3facU,0x20acU,0x85a9U,0xba74U,0x2661U,0x1538U,0xa757U,0xc5b9U,0xabe3U,
		0xf993U,0x1fb5U,0xc0faU,0xdf33U,0xd2a3U,0x635bU,0xc229U,0x4d00U,0xae26U,0x5193U,
		0x4b3fU,0x74fcU,0xccbfU,0x383dU,0xfe5cU,0xb808U,0x4bf5U,0x00e2U,0x190cU,0xe934U,
		0x2b3fU,0xf9fbU,0xc4bcU,0xc3a9U,0x814fU,0x5a5cU,0xecdcU,0xe9e5U,0x10b7U,0x377aU,
		0x3da1U,0x9ac3U,0x67f7U,0x19daU,0x227eU,0x4873U,0x59b0U,0x8e7eU,0x0f8eU,0x6679U,
		0xfd6aU,0x2800U,0xe77cU,0xda48U,0xa1fbU,0x7a7fU,0x86f3U,0xa11dU,0x0c25U,0x5420U,
		0xf292U,0xa4a0U,0x97dfU,0xf2adU,0xbf1bU,0xbf81U,0x954fU,0x73d4U,0xebabU,0xc3dfU,
		0xec35U,0x2ce6U,0x6c67U,0xe416U,0x0d47U,0xee63U,0x1712U,0xc85cU,0x6dc1U,0x16afU,
		0xd526U,0x6073U,0x5810U,0x8730U,0x418cU,0x65f9U,0xc2feU,0xc276U,0x4965U,0x6555U,
		0xf3c3U,0x1133U,0x577cU,0x59f5U,0x7b58U,0xc2a0U,0x02bcU,0x882cU,0x7f6fU,0xc566U,
		0x2458U,0x061aU,0x19aaU,0x4ee5U,0xfa9dU,0x5226U,0x8aeeU,0x9c19U,0xb1d9U,0xf1ceU,
		0x5a35U,0x0db4U,0xda27U,0x0c81U,0x26dbU,0x6d0fU,0x5c74U,0x2583U,0x128cU,0x39e5U,
		0xc15bU,0x055dU,0x0f02U,0x25cbU,0x43bbU,0xc374U,0x69f1U,0x34f4U,0x2417U,0x765dU,
		0xf0b8U,0xa2cdU,0x435eU,0x4c61U,0x6c2dU,0xeac5U,0xcbfcU,0x83d3U,0x5c25U,0x4793U,
		0x6c86U,0x3f58U,0x23f6U,0xc01bU,0x5320U,0xfc92U,0x8c4eU,0x854bU,0x13bfU,0xca0aU,
		0x42c1U,0x6cd0U,0x34c4U,0xa04aU,0x5103U,0x0356U,0xfb28U,0xef80U,0x8e52U,0xd91fU,
		0x451fU,0xd1b2U,0x0facU,0xf33dU,0x180cU,0xb7e4U,0x84f6U,0x0e6bU,0x4c3fU,0xe949U,
		0xba59U,0x4fc6U,0xc837U,0xf82fU,0x35b5U,0x8ed4U,0xe49bU,0xbeb6U,0x7366U,0xa01aU,
		0x4c1fU,0xd2feU,0x1030U,0x3698U,0xdcd0U,0x7331U,0x2a5fU,0x5608U,0x51cfU,0x3e55U,
		0x3028U,0x79c7U,0x9906U,0x2301U,0x6277U,0x486bU,0x000cU,0x0a26U,0x2432U,0xb9e6U,
		0x5ac1U,0x010fU,0xb137U,0xe4bbU,0x138aU,0x8a13U,0x1bacU,0x0010U,0x852bU,0x3b82U,
		0xf1e2U,0xb385U,0x0ddaU,0x1bdcU,0xb994U,0x94b9U,0xa3bfU,0xf852U,0xd1eaU,0x0c1cU,
		0x6657U,0x17c3U,0xec19U,0x0a2aU,0x9e8bU,0xa0eaU,0xaaebU,0x34ecU,0xd9c6U,0xe8d9U,
		0x0e8dU,0x799fU,0xb70aU,0xd025U,0x46b1U,0x81abU,0x5b4fU,0x1181U,0x7c14U,0x2436U,
		0x0dfaU,0x3999U,0xa15cU,0xab04U,0x78f5U,0x61d5U,0x3521U,0xad2eU,0xa656U,0xdc04U,
		0x3594U,0x8071U,0x89f2U,0x5fc1U,0x77bcU,0xb9eaU,0x3e32U,0x508fU,0xf76fU,0xaa10U,
		0x0519U,0x21fbU,0xd5edU,0x0e1cU,0x3745U,0xac80U,0x2de8U,0x6befU,0x157bU,0x094eU,
		0x1de0U,0x5ac5U,0x8db7U,0x0b46U,0x0686U,0x458eU,0xf610U,0x4706U,0x4e09U,0x7b15U,
		0x14dbU,0xc358U,0x570fU,0xbe4dU,0x2e41U,0x1655U,0x3c56U,0x0113U,0xb3d5U,0x81d6U,
		0xe5b0U,0xcc1dU,0x1a7aU,0xb4f8U,0xd166U,0x785cU,0x06daU,0xad1aU,0xb687U,0x936dU,
		0xbce3U,0x1814U,0xa435U,0xb13bU,0x476aU,0x9adaU,0x500dU,0x39b3U,0xd579U,0xe6e2U,
		0x1affU,0x8bc5U,0xc8d3U,0x2e8bU,0x047cU,0x4fadU,0xb36dU,0xc510U,0x12beU,0xe4bfU,
		0xf67eU,0xbb3dU,0x8857U,0x6205U,0xad66U,0x7b4aU,0xf6d6U,0x859aU,0xdcb4U,0xeea0U,
		0xadf0U,0x0fbfU,0xaa4bU,0xb01eU,0xcf01U,0x138eU,0x2e22U,0xdb80U,0x350eU,0xa5d1U,
		0x0ef4U,0x48ebU,0x19cbU,0x3197U,0x96e1U,0xc551U,0x80dbU,0x222eU,0xd8ccU,0x1a8cU,
		0x2ca0U,0x8a17U,0xd755U,0xd869U,0x2f7cU,0xbf78U,0x22cbU,0x8721U,0x412aU,0x6c1fU,
		0xc476U,0xfb3aU,0x3220U,0x49e7U,0x1937U,0x849dU,0x477bU,0x1bb0U,0xe26cU,0x845bU,
		0x3d13U,0xccaaU,0x0347U,0x5216U,0x7358U,0x27b4U,0x34d5U,0x8431U,0x84b9U,0xcda0U,
		0x596cU,0xd2edU,0xb57eU,0x0014U,0x0b20U,0xe922U,0x0a53U,0x2710U,0x808dU,0x6488U,
		0xb273U,0xd1c6U,0xe5d1U,0x8f35U,0xd5a3U,0x3818U,0x0576U,0xfedcU,0x3bfaU,0x852fU,
		0xaef6U,0xf6aeU,0x2bacU,0x57aaU,0x1e2fU,0x385bU,0x51d8U,0xe73dU,0x44adU,0xfb9fU,
		0xd446U,0x46d4U,0x92c8U,0xe785U,0xd0bcU,0x3b86U,0x8316U,0x85fbU,0xdfcdU,0xc717U,
		0xe612U,0x8336U,0x1363U,0xc7d4U,0x020dU,0x10a0U,0xdb64U,0xa1edU,0xacbdU,0x5dd4U,
		0x4ca9U,0xf1e6U,0xbc58U,0xaff1U,0x7c60U,0x13e1U,0x72dcU,0x2ecaU,0xe895U,0x659bU,
		0x9be2U,0x1ff0U,0x8141U,0xce3bU,0x7394U,0x09ccU,0x6e11U,0xb389U,0x9fbdU,0xcf6eU,
		0x1bf0U,0xb69aU,0x06f3U,0x4440U,0x6294U,0x61c4U,0x563cU,0x1301U,0x4268U,0xecd3U,
		0xfd8bU,0xfcb0U,0x0676U,0x0ddeU,0x7f40U,0x52b4U,0xe568U,0x5c6cU,0x033fU,0x14c1U,
		0x58d6U,0x5464U,0xc532U,0xff9aU,0xb452U,0x407fU,0x2aacU,0xefebU,0x4b2bU,0x1be0U,
		0x3fe9U,0xceb8U,0x1905U,0xfb21U,0x2e3aU,0x92c1U,0xe0adU,0x82aeU,0x122cU,0x030dU,
		0xfd4eU,0x2913U,0xa60fU,0x697fU,0x6e68U,0xb998U,0x0c06U,0x5bf2U,0xa61fU,0xf4a4U,
		0x72b8U,0xd0a6U,0x6aeaU,0x7fdcU,0x8236U,0xa1c0U,0x4db9U,0xf19fU,0x68afU,0xe390U,
		0xbb09U,0x94bdU,0x63dfU,0x0f88U,0xda21U,0xe495U,0x0df4U,0x4764U,0xe266U,0xbc52U,
		0x0c00U,0x260aU,0xff55U,0x0167U,0xed8aU,0x9cfaU,0x5267U,0xa3c3U,0x75f8U,0xf875U,
		0x298aU,0x5d27U,0xa599U,0x58e7U,0xfdc2U,0x8b41U,0x0ff4U,0xfcf7U,0xb49eU,0x9eb4U,
		0xe0b4U,0x6863U,0xdf87U,0xf856U,0x72a1U,0x8fb7U,0x19ecU,0x2a0aU,0xfd94U,0xd72eU,
		0xc729U,0x9f3bU,0x09e0U,0xb480U,0x2161U,0xaf3eU,0x28e8U,0xb7ecU,0x9814U,0xd1eeU,
		0x2610U,0x4953U,0x8818U,0xc263U,0x8c86U,0xb6d7U,0x15a0U,0x4685U,0x25a0U,0x0f24U,
		0x6660U,0xdc60U,0xf949U,0x1dc8U,0x4bf0U,0x0c20U,0xf3beU,0x0efdU,0x50feU,0x2a5aU,
		0xd1e5U,0x358fU,0x570aU,0x0477U,0x80d6U,0x84b4U,0xd441U,0x813cU,0xb44dU,0x4db4U,
		0xb499U,0x665bU,0x63b1U,0xb468U,0x599aU,0x8e75U,0xa937U,0xfb76U,0x039cU,0x9c03U,
		0x2675U,0x3c3aU,0x63b6U,0x9856U,0xf539U,0x15dbU,0xcddfU,0x17c7U,0x48aeU,0x3352U,
		0xe3a0U,0x62d8U,0xae97U,0x0d61U,0x2b78U,0xaba0U,0xdcdaU,0xf934U,0xd271U,0xc77dU,
		0x19ffU,0xdb94U,0x29b2U,0xec1dU,0xff5bU,0x9a08U,0x67a7U,0x5c7eU,0x14c9U,0xe9baU,
		0xbf25U,0xc060U,0xd148U,0x62edU,0xb46dU,0x2834U,0xe438U,0x2bdaU,0x6e86U,0x0a2eU,
		0xf58cU,0x5039U,0x1b6bU,0x2af2U,0x7c2fU,0x78bfU,0xc88eU,0xe81bU,0xc3c1U,0x6912U,
		0xd6f2U,0x862bU,0x482cU,0xe16fU,0xbb4eU,0x9e8fU,0x8774U,0xd3e0U,0xcb94U,0x6d9eU,
		0xefc7U,0x134eU,0x6fe1U,0x2c48U,0x83e4U,0xc453U,0x599fU,0x0ebbU,0x82b5U,0x1e02U,
		0xb4dcU,0xa0eeU,0xb8e6U,0x2516U,0x945eU,0x32d7U,0x1509U,0x1c17U,0x430aU,0xddb1U,
		0x8677U,0xec53U,0xd9feU,0x8564U,0xdac6U,0xbc8aU,0x91a3U,0xaaefU,0x016dU,0xbbb6U,
		0xecf9U,0x66d7U,0x0cfaU,0x9a33U,0x4d3eU,0x36b7U,0x6357U,0xf9f7U,0x8e7aU,0xf2a9U,
		0x16abU,0xc29cU,0x0db0U,0x34f0U,0xc017U,0xd91bU,0x8ed0U,0x79c3U,0x8292U,0x4bcaU,
		0xd979U,0x38ffU,0x2178U,0xedbaU,0x864bU,0xd337U,0x86c7U,0x9610U,0xd208U,0xd9caU,
		0x6515U,0xcb04U,0xd80bU,0x4506U,0xcb82U,0x7dcaU,0x9060U,0x1f8bU,0x4903U,0xc093U,
		0xa93cU,0xad83U,0x7e8cU,0x71d5U,0xdc42U,0xe8ddU,0x94a4U,0xd6c7U,0x47f7U,0x4654U,
		0x8453U,0x9a00U,0x072eU,0x7e41U,0x5c4dU,0xa326U,0xdfb8U,0x9a72U,0x1854U,0x0f6bU,
		0xd791U,0x0e91U,0xed90U,0xddd5U,0x781fU,0x9d4bU,0x81f5U,0xbe67U,0xb1a2U,0xd5caU,
		0xa547U,0xc6b6U,0xfb7bU,0x2b36U,0xdf7eU,0x3171U,0x4e46U,0x79a3U,0xdc86U,0x5eb1U,
		0x2a1aU,0x0c79U,0x9103U,0x9801U,0x4a20U,0x6698U,0x4f3cU,0xb5d0U,0x1f63U,0x9d65U,
		0x879fU,0x0789U,0x2786U,0xb70eU,0xf88aU,0x9f70U,0xcd22U,0x2a90U,0x6993U,0xa94fU,
		0x0bebU,0xb0c7U,0x5b27U,0x17a9U,0x03a1U,0x41c9U,0x1233U,0xe6f4U,0x7897U,0xd029U,
		0xc3faU,0x54a4U,0x1c2aU,0xa37fU,0x8b73U,0x57c7U,0x338cU,0xe6a6U,0xf327U,0x3f6aU,
		0xa3a6U,0xd121U,0x4d9fU,0x84feU,0x0697U,0x46b5U,0x9d00U,0xc93fU,0xa8abU,0x209cU,
		0x74e9U,0x4cc2U,0x731fU,0x6a38U,0x2a70U,0x4eafU,0x9c08U,0x35f4U,0xfdeaU,0x8be8U,
		0xee88U,0x81afU,0x88f6U,0x04c9U,0x5800U,0x8921U,0x9ba8U,0xdc75U,0x917aU,0x75b2U,
		0xcc8cU,0xb9bdU,0x3cdeU,0x9529U,0x64ffU,0xca31U,0xed48U,0x5b53U,0xdf55U,0x4723U,
		0x299aU,0xdea9U,0x2fedU,0xc1f4U,0x276dU,0xfcc3U,0x0d73U,0x393dU,0x267aU,0x5599U,
		0x02a9U,0x2cd2U,0x0845U,0x1185U,0xf4ecU,0x8fcbU,0xfcdbU,0xb8a1U,0x195cU,0x600eU,
		0xea7bU,0x9facU,0x7122U,0x7563U,0x85cdU,0x2631U,0x4ce8U,0xc76aU,0x759eU,0x7c18U,
		0x526dU,0x0736U,0xde41U,0xdbf2U,0x58deU,0x31aeU,0x489dU,0x6f9aU,0xa551U,0x53acU,
		0x3c3fU,0x1331U,0x915aU,0xc73fU,0x74d4U,0x243aU,0xea71U,0x9bbdU,0x2f32U,0x74f3U,
		0x5646U,0x3953U,0x900bU,0x154fU,0xeb80U,0x2d43U,0x4114U,0x54e7U,0xc664U,0x8cf8U,
		0xa0b5U,0x0dfeU,0x4064U,0xce93U,0x7239U,0xddfaU,0x4c7bU,0x87fcU,0x3e5eU,0x0bb3U,
		0xb213U,0xe10cU,0x63bbU,0x1486U,0x0314U,0xda51U,0x79ffU,0x399dU,0x3318U,0x4836U,
		0xcd0eU,0xc695U,0x8899U,0xf9e1U,0x613bU,0x617fU,0xf731U,0x86a3U,0x1673U,0xf61aU,
		0x01abU,0xa881U,0x3668U,0xa160U,0xa3c9U,0xae12U,0xf562U,0xb616U,0xba8cU,0x92d2U,
		0x97b6U,0x38ceU,0xc20bU,0x2f44U,0x985bU,0x6acaU,0xc760U,0xb332U,0xd75fU,0xab08U,
		0x6899U,0x6931U,0xf4d9U,0xe84fU,0x7b0eU,0xe736U,0x8b3aU,0x2c41U,0xb0c0U,0x0bacU,
		0xb09eU,0x6301U,0x3566U,0xe798U,0xcc2eU,0x78f9U,0x1c68U,0x2896U,0x4bdcU,0xbcbaU,
		0x9319U,0xe233U,0xc2e7U,0x88dcU,0x4fd7U,0x1083U,0xf53eU,0xda8aU,0x8999U,0x4a34U,
		0x9e20U,0x61d9U,0x37ceU,0xe9a1U,0xa44fU,0xeb35U,0x8979U,0x8753U,0x7300U,0xe0caU,
		0x029fU,0xb4b8U,0x54c9U,0xc220U,0x75efU,0x8012U,0xb16fU,0x3525U,0x75feU,0x9eebU,
		0xb08bU,0x8ae6U,0x3218U,0xd269U,0x85c5U,0x0b81U,0x567eU,0xf3f7U,0x15e0U,0x4164U,
		0xa783U,0x9970U,0x99acU,0xad32U,0x6d82U,0xaeb8U,0x3836U,0x6cc9U,0x2389U,0x07afU,
		0x40b0U,0x13caU,0x0636U,0xab6eU,0x925aU,0x441fU,0x103bU,0x92edU,0x941dU,0xa65aU,
		0x801bU,0x043aU,0x688aU,0xf361U,0x3883U,0x31b7U,0x0af5U,0xb0a5U,0x529aU,0x4970U,
		0xcde4U,0x7bd0U,0xfd55U,0x5a73U,0x317aU,0xdc08U,0x2ffeU,0xc32bU,0x0f0dU,0x7113U,
		0xfcb9U,0x7011U,0x4ed0U,0x9130U,0x97eaU,0x01c4U,0x7cb2U,0x2f63U,0x907aU,0x0e73U,
		0x957cU,0x3598U,0xf87bU,0x7d75U,0xdf6cU,0x1917U,0x967eU,0xe9f8U,0x20b7U,0x633bU,
		0x997fU,0x707aU,0x17ccU,0xe634U,0xbdf8U,0x507cU,0x2820U,0x8075U,0xaf89U,0x48f4U,
		0x09f8U,0x7382U,0x8a73U,0x7553U,0x7292U,0xbd21U,0x24cdU,0xf271U,0x974fU,0x1e93U,
		0x4016U,0x45d2U,0x2043U,0x89f6U,0xea45U,0xc9d6U,0xd37dU,0x66bfU,0xc083U,0x72c9U,
		0x3985U,0x562dU,0x311eU,0xb87eU,0x48b3U,0xa40aU,0x2869U,0x65cfU,0xffd0U,0x5fc5U,
		0xcc3dU,0x9c85U,0x0c37U,0xaf1cU,0x015eU,0x4416U,0x2bbeU,0xacfaU,0x6a93U,0xb039U,
		0x0b89U,0x642eU,0xa28fU,0x68d9U,0x7d94U,0x77c0U,0x2990U,0x7e49U,0x3643U,0x1facU,
		0x546cU,0xd139U,0x706aU,0xdb02U,0xb7b1U,0xf127U,0x3357U,0x4f05U,0xeb17U,0x36d4U,
		0xe75fU,0xb9eeU,0xc86fU,0xdd0aU,0x9dcdU,0x0d27U,0x61b2U,0xf7b8U,0xd945U,0x5b8eU,
		0x61e8U,0xad0aU,0x3e23U,0x9931U,0x6734U,0xf419U,0x78d3U,0x3e36U,0x935cU,0x6caeU,
		0x8f80U,0xf5a9U,0xa72bU,0x5b7cU,0xaa69U,0x6308U,0xfb8fU,0x5d4dU,0xe3a5U,0x06cbU,
		0x291aU,0x5dacU,0xd0e3U,0x5093U,0x978fU,0xecc2U,0x2c1cU,0xc173U,0xc627U,0xeddeU,
		0xa990U,0x053cU,0x2e13U,0xf074U,0xe5feU,0x12f1U,0xb9b3U,0xa4e1U,0x724eU,0xf773U,
		0x5d2dU,0x56eeU,0x2620U,0xa4faU,0xd713U,0xcab2U,0x6115U,0xebfaU,0x7ccaU,0x449eU,
		0x62ddU,0xebdcU,0xfeb5U,0xcf4cU,0x87c7U,0xaa14U,0x33fcU,0x5362U,0x1ed1U,0x9920U,
		0x837fU,0x682dU,0x4e67U,0xde7fU,0x8f8dU,0xde34U,0xd424U,0xf438U,0x8302U,0x779cU,
		0x96fdU,0x051dU,0xd8f8U,0x4de3U,0x03e3U,0x089fU,0x41bfU,0x2733U,0xc180U,0x4b7eU,
		0x914dU,0xef60U,0xae9cU,0xf16dU,0x9591U,0xce54U,0xd31eU,0x21ffU,0x986cU,0x3deeU,
		0xe9caU,0x16dfU,0x229cU,0xd004U,0xdec2U,0xcd58U,0xd87eU,0xdf10U,0x5112U,0x916dU,
		0x5d3aU,0x9330U,0xe2dcU,0xd5f1U,0xa59fU,0x3a7eU,0x7f2eU,0x5c04U,0xe06bU,0x3ba2U,
		0x51adU,0xc1a5U,0x601dU,0x0a08U,0x0d66U,0x255fU,0xca45U,0x29caU,0xbf54U,0x0e20U,
		0xfc68U,0xbb89U,0x536fU,0x934aU,0x2226U,0x984eU,0x5591U,0xa402U,0x7984U,0x84d8U,
		0x2023U,0xac9cU,0x5c14U,0x0cddU,0x5fd2U,0x3749U,0xa696U,0xa9daU,0xb1b4U,0x32b0U,
		0x888cU,0x7045U,0xe037U,0x356dU,0xc3f0U,0xbab8U,0x2b7dU,0x53dcU,0xa616U,0xb082U,
		0x98eeU,0xac84U,0x2050U,0x66f8U,0x7511U,0x112bU,0x75d6U,0x9d33U,0x761bU,0x79f2U,
		0x8b89U,0xfbf6U,0x36e1U,0x08d6U,0x922cU,0x4c47U,0x8cd5U,0x2decU,0x58edU,0x099dU,
		0x2d1bU,0xf451U,0xa29cU,0x5ca4U,0x4057U,0x4aefU,0x3479U,0x55c2U,0xaba5U,0xcaddU,
		0x9dedU,0x04aeU,0xafd3U,0x6bf3U,0x1bc8U,0x8a4dU,0x47a2U,0x7c4dU,0x2b24U,0xcff6U,
		0x9caaU,0xa9a2U,0x75c1U,0x1927U,0x96a2U,0x65e0U,0xa63eU,0x189bU,0x43ffU,0x157fU,
		0xd90aU,0x47eaU,0x7087U,0xe2b8U,0x4dd5U,0xce29U,0xee3bU,0x8e93U,0x6647U,0x7af6U,
		0xdcdfU,0x798cU,0x97f7U,0xb848U,0x60fbU,0x0952U,0x86baU,0xd858U,0x9c9fU,0x538aU,
		0x30daU,0x56a9U,0x7bddU,0x04a0U,0x30f5U,0x6028U,0x5686U,0x481cU,0xd396U,0x04dfU,
		0x2dbfU,0x1de4U,0xfdc8U,0x5c55U,0xaea7U,0xb195U,0xc53aU,0x1a55U,0xbaaaU,0xa2a7U,
		0x9267U,0x20d7U,0xf939U,0xe511U,0xe042U,0x7161U,0xb965U,0x5ac9U,0x1890U,0x98b1U,
		0x5c40U,0x5493U,0x8986U,0x4d8fU,0x6f4eU,0x2331U,0x7684U,0xce1bU,0xf707U,0x9662U,
		0x8a3cU,0x1f56U,0xa96eU,0x8dbbU,0xf64fU,0x1ca3U,0xe958U,0xc41bU,0x0304U,0xb4afU,
		0x3031U,0xe79fU,0x98a3U,0x3898U,0xd276U,0x3fccU,0x6986U,0xa204U,0x663cU,0x0b4aU,
		0x277cU,0x8580U,0xabf5U,0x00a0U,0xff41U,0xc820U,0xcf80U,0x3556U,0xa1a2U,0xc790U,
		0x1a1bU,0x1f48U,0xec6dU,0x49fcU,0x69a6U,0x068aU,0x8b47U,0x2ad1U,0x163cU,0xb2d4U,
		0x0401U,0xe672U,0xeb07U,0x90cdU,0x5dfcU,0x6945U,0xc782U,0x5a96U,0x537fU,0x8ac4U,
		0x88e9U,0x4592U,0x64a8U,0x1772U,0xc4d5U,0x326aU,0x81a5U,0x9367U,0x49e1U,0xa1e7U,
		0x290dU,0xaf38U,0xc777U,0x855eU,0x9a6cU,0xd11bU,0x262bU,0xf614U,0xc21aU,0x2f5dU,
		0x6428U,0x12ebU,0x9167U,0x08d0U,0x4816U,0x1f42U,0xc392U,0x8f73U,0x1a04U,0x4021U,
		0x623dU,0xd5ffU,0x70feU,0x470aU,0x7d9fU,0xa215U,0x2cf5U,0xf96bU,0x3970U,0xf20bU,
		0x21bbU,0x7732U,0x1767U,0x303aU,0x7cbdU,0xc398U,0x4df3U,0xa913U,0x5b60U,0x4e0dU,
		0x0ffaU,0x0641U,0x98f7U,0x3688U,0x730bU,0x4d59U,0xac5bU,0x9d7bU,0x419dU,0xbf62U,
		0xdb99U,0xa983U,0x8108U,0x0055U,0xe593U,0x7b19U,0x0852U,0x8d53U,0xa747U,0xfa8eU,
		0x8a66U,0x77aaU,0xff36U,0x6e28U,0x74dfU,0xd9afU,0xc30dU,0x2bc7U,0x68a6U,0xa77aU,
		0x80ffU,0x14dfU,0xb178U,0xe35fU,0xa8d0U,0x02b4U,0xd74dU,0x48a6U,0xf4e4U,0xcc35U,
		0x86b2U,0x300eU,0x29b7U,0xfe33U,0x6e6fU,0x3323U,0x4e4fU,0xc35cU,0x7c9cU,0x4933U,
		0xae1fU,0x8c47U,0x067fU,0x2ba5U,0x2983U,0xcb8dU,0xcd1bU,0x7232U,0x6883U,0x8f79U,
		0xb1adU,0xe951U,0xa8c9U,0x5713U,0xfcfdU,0x959fU,0x2d2eU,0xc0a7U,0x8a86U,0x431eU,
		0xde07U,0xe752U,0x8e0dU,0x5b32U,0xec22U,0xa2b9U,0xc33dU,0x05aaU,0xe320U,0xbe51U,
		0xf62bU,0x19d4U,0x9c13U,0x4fc0U,0x629fU,0x33bcU,0xa464U,0x438bU,0x8d2cU,0x47b0U,
		0x9a1eU,0x7ce3U,0xc935U,0xe6caU,0xf506U,0x2e45U,0x906dU,0x6b68U,0x6631U,0x516cU,
		0x186fU,0xae77U,0x5e3aU,0xb06cU,0xab56U,0xbffeU,0xff60U,0x84e0U,0x987cU,0x74b1U,
		0x390cU,0x1659U,0x3016U,0x09ebU,0x5180U,0x5ecdU,0x8202U,0x0758U,0x57dbU,0x96b5U,
		0x733bU,0x3075U,0xf3ffU,0x1a0aU,0x2f00U,0xf45fU,0x8956U,0x3c5aU,0xb4a4U,0xa32eU,
		0x0629U,0x258dU,0xffa2U,0x0178U,0xbf8bU,0x28adU,0x6412U,0x4735U,0x9a0dU,0x5828U,
		0x50d3U,0xe212U,0xcb9fU,0x0117U,0x63caU,0x908dU,0x5de9U,0x0ae8U,0x7fd3U,0x0b78U,
		0x9d72U,0xf475U,0x00beU,0xc8deU,0x7e1bU,0x82cfU,0x320bU,0xc4ecU,0xf644U,0xb3d9U,
		0xa39cU,0x604eU,0x571aU,0x3b34U,0x4103U,0xd672U,0xa3e5U,0x7900U,0x03aeU,0x0cb5U,
		0x67acU,0xca81U,0xb99fU,0x54bcU,0x7640U,0x81daU,0x983cU,0x205eU,0xcb62U,0x976eU,
		0x5559U,0x5952U,0xdfd8U,0x2150U,0x54f6U,0x044aU,0x4ff2U,0x4027U,0x9657U,0x9fc8U,
		0x37c1U,0xe5b4U,0x9ebaU,0x56c6U,0x3f77U,0xe724U,0xaaaeU,0xa9e8U,0xb126U,0x172fU,
		0x2cabU,0x3d71U,0x5c83U,0x9e49U,0x2a7dU,0xc6e2U,0x0416U,0xcc21U,0x0227U,0x7d0dU,
		0x78adU,0x5977U,0x0700U,0xc1d9U,0xd05bU,0xded7U,0x6efeU,0xbe38U,0xb91dU,0xd5e0U,
		0x9ccfU,0x9e57U,0xbd00U,0x1a7eU,0xe34bU,0xf495U,0x6688U,0x90f0U,0xadaaU,0x25b1U,
		0x76a5U,0x6de5U,0xef55U,0x5f4dU,0x14ceU,0xd184U,0x88c3U,0x28c5U,0x8423U,0xb4fcU,
		0xdc51U,0x6d4bU,0xfe6eU,0x38beU,0x1f1cU,0xccf2U,0xbc6aU,0x7f0dU,0xfad5U,0x5000U,
		0x47c1U,0x6243U,0xa84aU,0x647aU,0xe872U,0xd16aU,0xe0baU,0x938cU,0x160fU,0x62c9U,
		0x44fcU,0x90e6U,0x9022U,0x58adU,0xb360U,0x1cfdU,0xe9bfU,0xbc05U,0xafe3U,0xaf6bU,
		0x7fe9U,0x7860U,0x578fU,0x561dU,0x861cU,0x2c8bU,0x23b8U,0xb852U,0xbc1bU,0xec0cU,
		0xd9d9U,0x3535U,0x4ad4U,0xb9ffU,0x4d70U,0x457fU,0x858aU,0x06deU,0x3e00U,0x421dU,
		0xfd04U,0x1173U,0xde74U,0x0cc3U,0x851eU,0x1c6fU,0x7445U,0x29ddU,0xbf2aU,0x273dU,
		0x0c0dU,0x1c8bU,0xc0c1U,0xad1eU,0xbac2U,0x7aa5U,0x0101U,0x8d8dU,0x7b42U,0x3587U,
		0xdc6dU,0x754bU,0xcfeeU,0x33b4U,0x6058U,0xd605U,0xdda1U,0xba61U,0xbdc2U,0xb68bU,
		0x6869U,0xe6d4U,0x8687U,0x89b8U,0xd9b9U,0xd3cbU,0x82f7U,0xa350U,0x216eU,0x6ef4U,
		0xc065U,0x80a4U,0x628bU,0x8b31U,0x49d8U,0x9371U,0x81c8U,0x00ccU,0xfb5eU,0xc3d7U,
		0x0606U,0x6060U,0xef39U,0xc28eU,0x3145U,0xa4ebU,0x6cdcU,0x417bU,0x2dfcU,0x0589U,
		0x0a3bU,0xbce7U,0x0773U,0x6c39U,0x7e53U,0xe888U,0xd30aU,0xcc76U,0xc445U,0x701bU,
		0x4ef6U,0x57e9U,0xd14dU,0x202bU,0xb4c2U,0xb31dU,0x8181U,0x1818U,0x29bfU,0x4474U,
		0x0200U,0x7583U,0xb1f2U,0x0d53U,0x9bc7U,0xa532U,0xa522U,0x1fd6U,0x15e8U,0x7104U,
		0x07c7U,0xb33cU,0x113fU,0xa439U,0xdf8dU,0xdfc0U,0x7949U,0x5250U,0xb45aU,0xba1cU,
		0x1e18U,0x5ffaU,0xc6c0U,0x6dd5U,0x62f2U,0xfaa9U,0x71cbU,0x28bbU,0xe7cfU,0xb13fU,
		0x0569U,0x69b7U,0x51a2U,0x4cccU,0x130bU,0xde9aU,0xa1d7U,0x228aU,0xed06U,0xee0dU,
		0x71b4U,0x9cecU,0xa4acU,0x91f5U,0x7835U,0x476eU,0xf00eU,0x62adU,0x95a6U,0xb3b1U,
		0x1477U,0x52dbU,0x79d0U,0x289dU,0xec9dU,0xd2afU,0xb472U,0x362eU,0x5bf9U,0x7a88U,
		0x0b13U,0x9adeU,0xb224U,0xe179U,0xfa07U,0x7bfeU,0x79e2U,0xae85U,0x24b3U,0x0c6aU,
		0x3fb8U,0x14b3U,0x815eU,0x4710U,0x8781U,0xade3U,0x60aeU,0x5011U,0xf85cU,0x911bU,
		0x120cU,0x63f1U,0xfca7U,0xe78fU,0xd112U,0x0035U,0x0f7aU,0x99bcU,0x2839U,0x2074U,
		0x0ee7U,0x9727U,0xd873U,0x39b7U,0x9056U,0x769bU,0x2523U,0x5a1cU,0xd2d2U,0xcfcfU,
		0x937aU,0xb76fU,0xa9cfU,0xaed8U,0x8229U,0x7e06U,0x1d42U,0x003eU,0x4ea1U,0xd57dU,
		0x5a87U,0x284cU,0xd702U,0xd359U,0x2ea0U,0x41fcU,0x3d27U,0x2abfU,0x0489U,0xadc0U,
		0xe43dU,0xe0a0U,0x499bU,0x65b6U,0xe4e4U,0xe6e6U,0xb629U,0xb186U,0x18b4U,0x2fdcU,
		0x7e37U,0x7f03U,0x9af5U,0x27d2U,0x2a9fU,0xe7a8U,0x2018U,0x7da5U,0xddc7U,0x93e2U,
		0xb755U,0x1b03U,0x72a7U,0xd50cU,0xac8dU,0x10cbU,0xe525U,0x4941U,0x7a44U,0xf47eU,
		0x2fbfU,0xe55bU,0x2bdfU,0xda77U,0x14e8U,0x4bbcU,0xeaa7U,0x8bc9U,0xb55aU,0xf5ffU,
		0xbb6bU,0x16d0U,0xac79U,0x8f2eU,0x2603U,0x690bU,0xb5c9U,0x2d3cU,0xab67U,0xad03U,
		0x84d1U,0xce14U,0xd9a8U,0xc8d7U,0x352eU,0xee06U,0x2d35U,0x342dU,0x4919U,0x6fa3U,
		0x95f4U,0x4be3U,0x5902U,0x18deU,0x6e8bU,0xfe90U,0xa626U,0x3b25U,0x79acU,0x2e8fU,
		0xd7dfU,0x5855U,0x5bd1U,0xa873U,0x0de7U,0x423bU,0x8881U,0x3c26U,0x184aU,0xa840U,
		0xf396U,0xa21bU,0xc529U,0x4fceU,0xc389U,0x0480U,0x8fbdU,0x403dU,0x4a74U,0x6db9U,
		0xcef9U,0xb491U,0xed40U,0x203bU,0x43f7U,0xf4feU,0x0a33U,0x05e5U,0x025aU,0x40f2U,
		0xde19U,0x4fb1U,0x1c39U,0x31a0U,0xabdcU,0x83ccU,0x0800U,0xd60dU,0x1075U,0x4f20U,
		0xbbf6U,0x8f03U,0xe473U,0xf13fU,0x8323U,0x087bU,0x05edU,0xb371U,0x9d28U,0xd079U,
		0x80c9U,0x42daU,0xa6fcU,0x11d1U,0xcd7fU,0x9962U,0xc94eU,0x89a9U,0xf591U,0xaca4U,
		0xd48cU,0x2b92U,0x93aeU,0xc514U,0xfe3bU,0x9626U,0xf6bbU,0x038fU,0x042cU,0xe13bU,
		0x2f96U,0x32fdU,0x1bfdU,0xe150U,0x416cU,0x2cfbU,0xc6acU,0x1cf3U,0xf0d8U,0x12c2U,
		0x19f2U,0x9a7aU,0x8ccaU,0xe1baU,0x4087U,0x7003U,0x13eeU,0x5f11U,0x6381U,0x4d1cU,
		0x503eU,0x7befU,0x802dU,0x1d9fU,0x8ba5U,0xe4c3U,0x2d54U,0x1b5eU,0x0ecbU,0xf104U,
		0xfe81U,0x9effU,0xa569U,0x6cecU,0x4343U,0x3434U,0xb02bU,0xe7f6U,0xcad2U,0xf806U,
		0xb8d9U,0xf682U,0x8905U,0xfc2dU,0xc0aeU,0x9fdbU,0xd8d9U,0xd34aU,0x1011U,0xbcc1U,
		0x168dU,0x2c76U,0x1b70U,0x45c4U,0xf4abU,0x7b27U,0x7b83U,0xbb41U,0x8b1dU,0x45a3U,
		0x0d0dU,0xd0d0U,0xfb18U,0xeb2cU,0xf962U,0x2fd3U,0x5354U,0x8ee6U,0x9d9eU,0xf971U,
		0x0cedU,0xb3fbU,0x1bbdU,0x885bU,0x2a10U,0x8612U,0xbd3fU,0xd0fdU,0xe1a1U,0x226eU,
		0xbed7U,0xf0f5U,0x129dU,0x5b6eU,0x2af7U,0x20f3U,0x02d8U,0xd9f1U,0x59d4U,0x6209U,
		0xb58bU,0x2345U,0x3bc1U,0xb25fU,0xc1c5U,0x3c6cU,0x1cc6U,0xe5ebU,0x126bU,0x72f1U,
		0x1c7bU,0xdceeU,0x173fU,0xb39cU,0x46e5U,0xad6aU,0xd543U,0xbc27U,0xb5e2U,0xdd65U,
		0xd3d7U,0x6ca5U,0x6045U,0xfc24U,0x2811U,0xd953U,0x7c34U,0xadd3U,0xcf93U,0xb0dbU,
		0x2fb4U,0x7b4eU,0xffc6U,0x93a4U,0x6579U,0xfc89U,0x1eb2U,0xb929U,0xffebU,0xb56aU,
		0x4bacU,0x90a1U,0x484eU,0x3976U,0x3aafU,0x5cdcU,0xd190U,0xf6daU,0xfd9aU,0xf74dU,
		0x6810U,0x0085U,0x9b80U,0xd4caU,0x1438U,0xd046U,0xbd74U,0x086bU,0x78c4U,0xfae1U,
		0x16c5U,0x6316U,0xcfb3U,0x859eU,0x1d09U,0x5123U,0x0e4bU,0x3dceU,0xb41cU,0x4a11U,
		0xb838U,0x878dU,0x7527U,0x4920U,0x3a3dU,0xd477U,0x9733U,0x0820U,0x3298U,0xdcb8U,
		0xcb1bU,0xc635U,0x8a8dU,0x3a27U,0xdc33U,0xa05aU,0xaa39U,0x9320U,0x85e8U,0x7841U,
		0xc893U,0x4f99U,0x72bfU,0x11c7U,0x0924U,0xeea4U,0x7e9dU,0x5f9bU,0x9aeaU,0x65efU,
		0xc6d2U,0xcac0U,0x73b3U,0x97a7U,0x69c3U,0xefa1U,0xf8feU,0xf211U,0x5e96U,0xc83fU,
		0x4f71U,0xadf4U,0xd734U,0x544cU,0xeec8U,0xdf99U,0x9a5aU,0xa76aU,0x758fU,0xb5a2U,
		0xeffaU,0xea8bU,0xe820U,0xfe0fU,0xb7caU,0x7e5fU,0xeae2U,0x0fc3U,0x93d7U,0x18c9U,
		0xd694U,0x4a5aU,0xc7ccU,0xdda9U,0x6177U,0x0534U,0x354eU,0x2148U,0x0c62U,0x2fcbU,
		0x3496U,0xa418U,0x0262U,0xaa4fU,0xe7b4U,0x42fbU,0x5d9cU,0xc0e8U,0xa7ccU,0xe449U,
		0x4990U,0xd51cU,0xea3bU,0x9d1eU,0xc3c6U,0x5c1cU,0xd6e9U,0x3e1aU,0x7e12U,0xb022U,
		0x6e77U,0x8e3dU,0xf2f5U,0x0e63U,0x4247U,0x48d4U,0xcf13U,0x742dU,0x05c9U,0x975dU,
		0xa78bU,0x21c1U,0xa2eeU,0xed65U,0x70b0U,0xcf05U,0xc72fU,0x185cU,0xc90aU,0xfdb3U,
		0x2ab4U,0xa187U,0x44d0U,0xdffaU,0x5482U,0x3ed8U,0x6917U,0x9276U,0xdfe8U,0x0a87U,
		0x5d8eU,0x1392U,0x0dc7U,0x9600U,0x0667U,0xbba2U,0x6f20U,0x2936U,0x42e6U,0xb890U,
		0x94ccU,0x6faaU,0x9e6eU,0xe645U,0x3c0fU,0xf6c7U,0xefdbU,0x2e26U,0x6c05U,0x328aU,
		0x4325U,0x90bbU,0x1e89U,0xf135U,0x990fU,0xe23aU,0x9a86U,0x9710U,0xd6f7U,0xee59U,
		0xd0adU,0x86fcU,0x2358U,0xdb84U,0xf8f0U,0x4295U,0x589bU,0x2d60U,0x8c1bU,0xcc0cU,
		0xccd9U,0x349eU,0x1b26U,0x7b8fU,0xb93cU,0x7738U,0xc500U,0x9476U,0x91d0U,0x3512U,
		0x9f41U,0x4673U,0xf68eU,0x6037U,0xe484U,0x6397U,0x53c5U,0x02e4U,0xaecdU,0xf3d9U,
		0x8630U,0x9aa6U,0xb407U,0xfeccU,0xa426U,0xa5d5U,0xdaf0U,0x6aa1U,0x0e3dU,0x0935U,
		0xa0e5U,0x508aU,0x81d1U,0xbb38U,0xd864U,0xe91dU,0x85f6U,0xcf69U,0xceb3U,0x0f83U,
		0x8fb2U,0x0ef8U,0x334dU,0x5034U,0x2511U,0xd916U,0xd6c2U,0x5eacU,0x549fU,0x04c4U,
		0x8fc6U,0x9bb8U,0x4831U,0x692cU,0xe99cU,0xaeb3U,0xc326U,0x48efU,0x9c80U,0xdd05U,
		0xecbdU,0x535dU,0x3de9U,0xbb84U,0x66f3U,0x8a48U,0xd853U,0x98acU,0x857bU,0x176dU,
		0xa210U,0x8d4eU,0x492eU,0x19cfU,0x09e6U,0x9088U,0x2059U,0x7d08U,0x6d46U,0x5618U,
		0x7aa0U,0x00c7U,0x446fU,0x69b2U,0xe174U,0x7696U,0xb181U,0xf5faU,0x5850U,0x319bU,
		0x9621U,0x1b59U,0x459eU,0x2340U,0x939fU,0x511eU,0x5f96U,0x18c4U,0x8e38U,0x95fbU,
		0x4290U,0x6a9cU,0xdd00U,0x1b54U,0x256bU,0x96e5U,0x7607U,0x97d4U,0xde0eU,0x7653U,
		0x68c3U,0xa55aU,0x0a68U,0xc2eeU,0xb6f3U,0x76f6U,0xbb53U,0x8f99U,0x6af1U,0xdc16U,
		0xdc8fU,0xc555U,0xef6cU,0x2570U,0xe82eU,0xe404U,0x7f49U,0xdd24U,0xb746U,0xf08bU,
		0xa5f2U,0x6085U,0x4afbU,0x3040U,0xd682U,0xf837U,0xd5b4U,0x80dfU,0xb486U,0x7627U,
		0x9302U,0xda3dU,0x7051U,0x3374U,0xf21fU,0xed31U,0x9560U,0x942bU,0x9e94U,0xf4c3U,
		0x3f04U,0x2961U,0xc115U,0x2232U,0xdbb8U,0x96eaU,0x49beU,0xeb6bU,0xc8acU,0xba69U,
		0xac6eU,0x244eU,0x0154U,0x0422U,0x17f6U,0x13b6U,0xefbeU,0xa722U,0x40faU,0xd8d0U,
		0xbf45U,0xee20U,0xe5c0U,0xc82fU,0xacb5U,0xdabeU,0x01a3U,0xb9abU,0xec65U,0x964fU,
		0x8779U,0x0ce5U,0xc4f8U,0xbbddU,0x3c7aU,0x1a90U,0x332bU,0x760cU,0x86ecU,0xdcc9U,
		0xc8ccU,0x9bdbU,0x2599U,0x6350U,0x2a69U,0xc204U,0x9978U,0x7cc3U,0x3472U,0x5df5U,
		0x8e06U,0x2ca4U,0x2167U,0x0f73U,0x43f0U,0x1296U,0xeff3U,0xaec6U,0x9559U,0xa267U,
		0x6b74U,0xbf08U,0xd3e5U,0xa4c7U,0x0b62U,0x67d9U,0xcdbfU,0x8a1bU,0x67c9U,0x97d9U,
		0x69ebU,0x3022U,0x77b6U,0x12b8U,0xb578U,0x6e0bU,0xbb03U,0x4beaU,0x6e80U,0x0daaU,
		0x4e40U,0xee82U,0x74ceU,0xd759U,0x99a6U,0x281aU,0xe759U,0x87c1U,0xbf4eU,0xafcdU,
		0xb95fU,0x88e3U,0xe58dU,0xe31aU,0xcb99U,0x0410U,0x7fe3U,0x49d2U,0xe7c9U,0xd86dU,
		0xeaa1U,0xde13U,0x8b9fU,0x59ceU,0xcfadU,0xeadcU,0x5d88U,0xa420U,0x584aU,0xc10fU,
		0xcdb9U,0xc39eU,0xbbe5U,0xeb43U,0x37b2U,0x2f80U,0xaf44U,0xdb2cU,0x1446U,0x1421U,
		0xd8b2U,0xdba5U,0x345bU,0xd369U,0xa26eU,0xd771U,0x6da3U,0xa141U,0x6753U,0x3fa7U,
		0xc3a4U,0xbf7cU,0x8827U,0x7658U,0xd1adU,0x22fcU,0x1119U,0xe08aU,0xf5b5U,0x24dfU,
		0xc06fU,0x6adaU,0xed1eU,0x4981U,0x98cdU,0x2c28U,0x11a0U,0x22cfU,0x8ce7U,0x2497U,
		0x3e84U,0x902cU,0x87e1U,0x7571U,0x252dU,0x6b4aU,0x56faU,0xa7b8U,0xefccU,0x7764U,
		0x42cfU,0xe7d9U,0x1ee7U,0x8725U,0x7b6dU,0x68c8U,0x8042U,0x3408U,0x236cU,0xebbbU,
		0x32c5U,0x9affU,0x6a25U,0xffdcU,0x23a2U,0x4df9U,0xf3a0U,0x513aU,0x5516U,0x412eU,
		0x28eeU,0xa5b6U,0x8a02U,0x3a8dU,0x1c0eU,0xedd5U,0x5949U,0xeb23U,0xead3U,0xf0e2U,
		0x1353U,0x929dU,0x68e5U,0x6c5cU,0xbbebU,0x6c23U,0xd93bU,0xa55fU,0xbd49U,0xb7dbU,
		0x4c2dU,0x6105U,0x7a6cU,0x344cU,0x31ddU,0x5909U,0xe686U,0x998bU,0xce03U,0x9537U,
		0xabffU,0xc47aU,0x01d0U,0x2d93U,0x8e14U,0xe1fcU,0x40a0U,0xcdf0U,0x8df7U,0x4fdeU,
		0xea22U,0x2256U,0x6fe6U,0xc18aU,0x823dU,0x6469U,0x442bU,0xfb3eU,0x2b87U,0x0a6dU,
		0xf2d7U,0x59eaU,0x875fU,0x5f87U,0x8a29U,0x275dU,0xacd1U,0x2688U,0x5724U,0xa919U,
		0x1626U,0x35bdU,0xe6b9U,0x3224U,0xb7f2U,0xf510U,0x796bU,0x15c3U,0xc317U,0x5766U,
		0x071eU,0xed77U,0x6b7bU,0xcd42U,0x2c4dU,0xfc77U,0xf40fU,0xf7fcU,0xeb49U,0x49ebU,
		0x6449U,0xc2f3U,0xcd74U,0x3f4eU,0x3cc8U,0xbdcaU,0xab32U,0x4cf4U,0xee71U,0x7258U,
		0x727aU,0x1ae6U,0xc74bU,0x0d96U,0xde21U,0x193bU,0xf71fU,0xba9aU,0x364dU,0x8de8U,
		0x32f2U,0x6609U,0xf987U,0x4edaU,0x4b53U,0xed54U,0x83e9U,0x5fdaU,0x54d3U,0xc610U,
		0x840aU,0x84a1U,0x4e57U,0xb6f8U,0x1191U,0x08aeU,0x1fe8U,0xf9efU,0x2f3cU,0x4496U,
		0x693dU,0x3d69U,0x99b4U,0x5b66U,0xf3d1U,0xd769U,0xf0a5U,0x477fU,0x981aU,0xd799U,
		0x1838U,0xa3d5U,0x4b33U,0xb1ceU,0xc6a1U,0x34baU,0xfb85U,0x1683U,0xc458U,0x35acU,
		0x919aU,0x7245U,0x37b8U,0x1bb4U,0x37a9U,0x76fbU,0x2973U,0x7329U,0x4272U,0xcbb1U,
		0x7143U,0xf9b7U,0x7fb2U,0x18e5U,0x1f97U,0xe9d9U,0x5838U,0x82e7U,0x2001U,0xe270U,
		0x2f1fU,0xd985U,0x5b39U,0x1a41U,0xbe73U,0x246dU,0x230aU,0x108aU,0x4d11U,0x3e6cU,
		0x59a4U,0x0552U,0xa1c7U,0xa126U,0xba35U,0x845fU,0x5234U,0xbb58U,0xae49U,0x5736U,
		0xe542U,0x1f6fU,0xd7a8U,0x026aU,0x33d8U,0x5414U,0xc7efU,0x4e13U,0x3c82U,0x2a44U,
		0x41d5U,0x3d17U,0xd1f4U,0xc91eU,0x69d8U,0x55d7U,0xc169U,0xd0c6U,0xf1b7U,0x6a79U,
		0xbf0fU,0xf0adU,0x0ec0U,0xf333U,0xfa3cU,0x155dU,0x2f86U,0xccaeU,0x4ebbU,0x8f9eU,
		0x1454U,0xd833U,0x444cU,0x309dU,0x00d5U,0x9b33U,0xab22U,0x4805U,0xe331U,0x505aU,
		0x45b9U,0x99c5U,0x26cfU,0x034bU,0x4881U,0xa50fU,0x43cfU,0x93c2U,0xf308U,0x09b1U,
		0x565eU,0x2654U,0x4046U,0x5939U,0x82baU,0x9dfcU,0x8d3cU,0xe8b2U,0x5e71U,0x521aU,
		0x028cU,0x6af6U,0x0abdU,0x8eefU,0xf2a0U,0xebd3U,0x9e40U,0x20eaU,0xa138U,0x38a1U,
		0xac1bU,0x1000U,0x43afU,0x4514U,0x8b12U,0x735cU,0x2616U,0xbd35U,0x6701U,0x55ffU,
		0x460dU,0x666cU,0xe27fU,0xc8e7U,0xb78fU,0xa172U,0x1e07U,0x77edU,0xe368U,0x39d9U,
		0xaf4aU,0x27b8U,0xda0dU,0xdc1bU,0x4b99U,0x994bU,0xb828U,0x19baU,0x4ab0U,0xca15U,
		0xbfa4U,0x6e92U,0x52f9U,0xb042U,0x36a3U,0x2b0cU,0xa237U,0x34d9U,0x9ef4U,0x5dc2U,
		0xec29U,0x4c56U,0x4d2dU,0x53b5U,0x77fdU,0xf545U,0x6369U,0x6560U,0xb4e1U,0xf7d0U,
		0x4dc0U,0x6a65U,0x5ebaU,0x8435U,0x0f59U,0xdc94U,0x25d6U,0xc850U,0x52bdU,0x615dU,
		0x66e3U,0x0bdaU,0xf2b8U,0xe11cU,0x10f6U,0x0647U,0xf06aU,0x8edcU,0xba44U,0x84bdU,
		0x4959U,0x23ebU,0xb8ceU,0xe93fU,0xe2f2U,0xdde3U,0x85b4U,0x0193U,0xd3ecU,0x6842U,
		0xa0f3U,0x3a51U,0x2b4dU,0x6763U,0xdb32U,0xcda4U,0xe01dU,0xc55aU,0x1ce1U,0xb8f2U,
		0x61feU,0xb693U,0x2a53U,0x79bcU,0x891aU,0xe848U,0x737bU,0x9919U,0x7c46U,0x3263U,
		0x4fb9U,0x5970U,0xc3d0U,0x5a15U,0x83c5U,0xb258U,0x4a53U,0x092eU,0xeb64U,0x22f5U,
		0x3f47U,0xd82cU,0xb8ebU,0x3751U,0xb28bU,0xb71cU,0x6583U,0xd2f1U,0xc364U,0xef71U,
		0x4b09U,0xf821U,0x1800U,0x4c14U,0xe196U,0x51c5U,0xfeabU,0x02ceU,0xad3aU,0x98fdU,
		0xc00eU,0x33f3U,0x021eU,0xb582U,0x881eU,0x0e99U,0xc977U,0x626fU,0x1be8U,0x8ec8U,
		0xf4d1U,0x1ec9U,0xc4cdU,0x78a5U,0x251bU,0x3bb9U,0x0e35U,0xd1a5U,0x144cU,0x0018U,
		0x1ab4U,0x2575U,0x73cdU,0x4864U,0x5d62U,0x850cU,0xbd53U,0x5334U,0x8607U,0xfe97U,
		0x8e46U,0xbeabU,0x1f01U,0xeb53U,0xa10fU,0x0b24U,0x26b9U,0xd95cU,0xa2c0U,0xaf02U,
		0xee29U,0xbe95U,0xb6a6U,0xda91U,0xc462U,0xeb90U,0x9463U,0xc9feU,0xf1a6U,0x7c6cU,
		0xfa4bU,0xe926U,0xce68U,0xe833U,0xcf2cU,0xf1c1U,0xebf0U,0xf0ebU,0xac2bU,0xaa57U,
		0x5314U,0xba4eU,0x9d93U,0x368eU,0x1a98U,0x99d7U,0x7704U,0x0a57U,0xc269U,0xfdf8U,
		0x670bU,0xe3c5U,0xdb15U,0x39f5U,0x58c4U,0xac35U,0xa4ceU,0x4787U,0x32dcU,0x38aaU,
		0xad44U,0x9ffbU,0x1427U,0x2714U,0x0ea2U,0xe409U,0x0020U,0x3758U,0xc0d8U,0xfd7aU,
		0x7745U,0x0cceU,0xc491U,0xd783U,0x2c01U,0x1b8eU,0xa35bU,0x45f8U,0x834cU,0x8091U,
		0x104fU,0xecb0U,0xd456U,0xa9f7U,0xbcd2U,0x3dbeU,0x2724U,0xbb1cU,0x863dU,0x5f58U,
		0x150eU,0xbfd1U,0x8749U,0x7ef9U,0xebc9U,0x648cU,0xcea6U,0x7f4eU,0xbc10U,0xdd76U,
		0xc03bU,0x8ab5U,0x1411U,0x6723U,0x39c2U,0x71e5U,0x04fdU,0x7311U,0x1acdU,0x8268U,
		0x4abfU,0xb277U,0x8c8cU,0xfefeU,0x6fb7U,0x7a93U,0x569aU,0xcec6U,0x4cd7U,0xe3f4U,
		0xb665U,0x9b49U,0x1c1cU,0x3d3dU,0x9a93U,0x447fU,0xd8b8U,0xd1caU,0x4574U,0xdd29U,
		0xd846U,0x77d7U,0xc9b9U,0x0837U,0x24edU,0xd22eU,0xe1ecU,0xa62dU,0x7db0U,0xa904U,
		0xdb46U,0x4283U,0x5ee2U,0xe5d5U,0x44ddU,0xaee3U,0xc344U,0xc7a6U,0x3c31U,0xb875U,
		0xbff5U,0x89a0U,0x9646U,0xd823U,0x430fU,0x5568U,0x68b6U,0xa6d3U,0x743aU,0x8f39U,
		0xb165U,0xb74bU,0x8b08U,0x654cU,0xa7d9U,0xd084U,0xef0dU,0x8ffdU,0x41e7U,0x74a2U,
		0x55efU,0x4d5fU,0x0241U,0xe4a3U,0xe1c5U,0xd5a7U,0xb6ddU,0x11f8U,0x253bU,0x26a6U,
		0xc5e5U,0xbbc7U,0xf554U,0xf14fU,0x0b69U,0x0326U,0xddb6U,0xf811U,0x8d82U,0x4e84U,
		0xdbabU,0x381cU,0x8866U,0xf090U,0xd647U,0xb292U,0x948cU,0x6871U,0x5395U,0x5534U,
		0x7f5aU,0x43e3U,0x6b89U,0x308aU,0x7eaaU,0x98bcU,0x1c41U,0x057aU,0xdefbU,0x940fU,
		0xcb3eU,0x105bU,0x415aU,0xda6dU,0x77e3U,0x37f0U,0x0b55U,0x86d7U,0x867cU,0xa69eU,
		0xf059U,0x6fc3U,0x52ccU,0xfee0U,0x7ca4U,0xa5f7U,0xc642U,0x8acfU,0xf8d2U,0x7d25U,
		0x7714U,0xdb52U,0xd6a1U,0x0a13U,0x6d8aU,0xac61U,0xa6dfU,0xaf5cU,0x2696U,0x3bfeU,
		0x15a6U,0xed98U,0xf178U,0xc2d9U,0x3ff1U,0x73e4U,0xdfa6U,0x5cafU,0x33a9U,0xe1d1U,
		0xec58U,0x1a2dU,0x3578U,0x6e47U,0x3461U,0x8533U,0x9f81U,0x608aU,0x3828U,0x790fU,
		0x57a2U,0x6d96U,0xddf2U,0xf5a1U,0xc413U,0x3b2cU,0xb3a9U,0x9fd3U,0x90b3U,0x87b9U,
		0x1a39U,0xaefaU,0xb3e4U,0xcbecU,0x05b1U,0xb3ecU,0xcfa0U,0xa339U,0x6280U,0x4a3bU,
		0x6d78U,0xb550U,0xda03U,0x7f66U,0xe397U,0x0c2eU,0x5177U,0xf6b2U,0x86e3U,0x4b00U,
		0x6ff4U,0x8543U,0x6e33U,0x1256U,0x3b0aU,0xe7bcU,0x64c0U,0x6d21U,0x7783U,0x9d81U,
		0x3333U,0xeeeeU,0xf6e7U,0x2bb0U,0x468bU,0x3b16U,0x6b18U,0x9b62U,0xa345U,0x1d8bU,
		0x9999U,0x7777U,0x67e0U,0x9822U,0x8569U,0x3b66U,0x0ad3U,0x76ccU,0xd36fU,0x57aeU,
		0xcfc0U,0x3045U,0x1abcU,0xb723U,0xc8b9U,0x6cb9U,0xfb67U,0x9215U,0xd595U,0x7ad7U,
		0x30b8U,0x8e5fU,0x288cU,0x2842U,0xa505U,0x1e33U,0x5e89U,0xb659U,0x8418U,0x7ebaU,
		0x66afU,0xc028U,0x9e0bU,0xe160U,0x4a7dU,0x4c9bU,0xdacbU,0x45ddU,0x6f64U,0x5f00U,
		0xe250U,0x385fU,0x77cbU,0xd687U,0xb10fU,0x9da7U,0x92b2U,0x47d6U,0x1949U,0xa36fU,
		0x719dU,0xd27fU,0x2f6eU,0x41a3U,0x06b6U,0x247eU,0x4b6dU,0x51dcU,0x25a6U,0xab79U,
		0x751aU,0x1405U,0xe0d5U,0x1db9U,0xb606U,0x7e24U,0x7489U,0x097bU,0xbc8fU,0x734aU,
		0xa8d9U,0xd7c8U,0xa274U,0xe741U,0x6a04U,0xf83cU,0x825cU,0xaa2cU,0x046eU,0xbd18U,
		0x4382U,0x4f17U,0x2445U,0x79b3U,0x2ebbU,0x0b92U,0x9b73U,0x873dU,0xe340U,0x44b1U,
		0xb094U,0x821fU,0xe327U,0x2cddU,0x3cfcU,0x3c48U,0xe85aU,0x9e27U,0x5d17U,0x2212U,
		0x91a8U,0x848eU,0xbb10U,0x4841U,0x2a23U,0xfba3U,0x9f5bU,0xd5b9U,0x08f9U,0x7f7cU,
		0xe571U,0xc239U,0xa14eU,0x7dd5U,0x21ecU,0x0c8eU,0x9b93U,0xbf68U,0xd5d5U,0xbebeU,
		0xffb9U,0xd44aU,0x0f2aU,0x3ebdU,0x9bf2U,0x4744U,0xa57aU,0xc2baU,0x7c7cU,0x2424U,
		0xcdc6U,0x17b4U,0xaaf4U,0x827bU,0xa02eU,0xfc41U,0xd777U,0x46d8U,0x178cU,0x80e4U,
		0x35ffU,0x658aU,0x61cfU,0xe6dcU,0x5210U,0x2ec4U,0xd0a0U,0xb6d1U,0xe9b4U,0x9a2dU,
		0xbe61U,0x4cbcU,0x31a8U,0x92ccU,0xd263U,0xe9f2U,0xd133U,0xcaacU,0x3b9cU,0x5c9eU,
		0x1a4fU,0xe66cU,0x4d53U,0x4318U,0x0172U,0xa9e2U,0x90e0U,0xd3c5U,0xba16U,0xe789U,
		0x493bU,0xb48bU,0x6ffdU,0x2268U,0xd4c4U,0xa764U,0xa181U,0x6391U,0x5612U,0x336eU,
		0xaec0U,0xdb9fU,0xedcfU,0x5760U,0xb1c8U,0xd0c0U,0x6666U,0xddddU,0x8ec2U,0x39efU,
		0xcec0U,0xbbc1U,0x73deU,0x1d85U,0x1db3U,0xc2b4U,0xbbbbU,0xccccU,0xd630U,0x36c5U,
		0x6da9U,0x3b8aU,0x8d16U,0x762cU,0xae57U,0x6fd3U,0xc7bfU,0xc9adU,0x1f11U,0xb0f3U,
		0x930fU,0x2e96U,0xf2feU,0x1709U,0xb444U,0x400dU,0xc92cU,0x831aU,0xefb5U,0x7c3dU,
		0xbe58U,0xcbf4U,0xe5c9U,0xc3b9U,0xeb78U,0x61e0U,0x767cU,0x00b6U,0xecfeU,0x433bU,
		0x94c4U,0xbafbU,0x7faaU,0x85ffU,0xc40bU,0x9307U,0x10b0U,0x84efU,0xd15fU,0x135cU,
		0xc722U,0x4303U,0x3385U,0x6134U,0x4ec9U,0xa989U,0x7614U,0xcf79U,0x297cU,0xdfd1U,
		0xdc66U,0x24acU,0x887aU,0xf95bU,0x73acU,0xccd2U,0xb73fU,0x5d81U,0x8a22U,0xd7a1U,
		0x66dcU,0xac24U,0xef06U,0x3b03U,0xa147U,0xc71bU,0xc980U,0xda42U,0x257dU,0xd2f8U,
		0x4539U,0xb948U,0xa00aU,0x1021U,0x80aeU,0x95deU,0x5147U,0xbec9U,0xbd86U,0xc59cU,
		0x9ab2U,0xe616U,0x44bcU,0xb97fU,0x3af7U,0x58b7U,0x68f2U,0xab84U,0x5a26U,0x5a4cU,
		0x574bU,0x95bbU,0x0cffU,0x94d8U,0x0379U,0xb149U,0x7ec5U,0x833aU,0x971cU,0x7056U,
		0x5307U,0xb733U,0xb0feU,0x1529U,0xbe22U,0x27dcU,0x8bffU,0xf0caU,0xe006U,0x810eU,
		0xa225U,0x006aU,0x22dcU,0x1367U,0xf94fU,0xcf1fU,0x21d7U,0xc588U,0xe455U,0xd636U,
		0xf7deU,0xa07cU,0xe465U,0x12ccU,0x9a38U,0xc702U,0x5072U,0x40e8U,0x6759U,0xc7d8U,
		0x1ef4U,0x3379U,0xd107U,0xa496U,0x3bdbU,0x095cU,0xdd81U,0x25f5U,0x5fa7U,0xd7e6U,
		0xf2c6U,0x718dU,0x6456U,0xdd34U,0x00aaU,0x0211U,0xd65aU,0x6d5fU,0xf632U,0xcb27U,
		0x3790U,0x149bU,0x4a66U,0x37d5U,0x8f44U,0xf89aU,0x4d43U,0x4b88U,0x63e6U,0xeed4U,
		0x1aa7U,0x10a4U,0x53e6U,0xf224U,0xf51dU,0x4e8fU,0xe3dfU,0xe4d6U,0xc987U,0x174bU,
		0xf2e4U,0xe1dcU,0x3b3eU,0x005bU,0xb0e7U,0x736eU,0x6850U,0xdb68U,0x1dceU,0x2e4fU,
		0xa6a9U,0x218cU,0x2bd1U,0x36cbU,0xe209U,0x1d96U,0x67d0U,0xd19cU,0x36bcU,0x5aa8U,
		0x4a46U,0xb438U,0x3fadU,0xa1f1U,0x20adU,0xed36U,0x85aaU,0xc36bU,0xba75U,0x868fU,
		0x2662U,0x787eU,0x1539U,0xf77dU,0xa758U,0x4aa3U,0xc5baU,0xd7b9U,0xabe4U,0xacc1U,
		0xf994U,0x893cU,0x1fb6U,0xc437U,0xc0fbU,0xea5cU,0xdf34U,0x913aU,0xd2a4U,0xdbd4U,
		0x635cU,0xb1bcU,0xc22aU,0xd7f5U,0x4d01U,0x5dd8U,0xae27U,0x9565U,0x5194U,0x2b60U,
		0x4b40U,0x1e48U,0x74fdU,0x3108U,0xccc0U,0xb8c1U,0x383eU,0xe599U,0xfe5dU,0xab12U,
		0xb809U,0x4cadU,0x4bf6U,0x7827U,0x00e3U,0x6b28U,0x190dU,0x6dafU,0xe935U,0x6cfcU,
		0x2b40U,0x8d0aU,0xf9fcU,0x27f6U,0xc4bdU,0xabcbU,0xc3aaU,0xf1eaU,0x8150U,0x9430U,
		0x5a5dU,0x6a42U,0xecddU,0x8f53U,0xe9e6U,0x5e64U,0x10b8U,0x585cU,0x377bU,0xd4e3U,
		0x3da2U,0x3241U,0x9ac4U,0xbc5cU,0x67f8U,0xa6c4U,0x19dbU,0x05b9U,0x227fU,0x5e59U,
		0x4874U,0xe9a8U,0x59b1U,0xdbdfU,0x8e7fU,0x22b5U,0x0f8fU,0x5429U,0x667aU,0xaff5U,
		0xfd6bU,0x9e99U,0x2801U,0xf75eU,0xe77dU,0x1dfaU,0xda49U,0x5da4U,0xa1fcU,0x54b4U,
		0x7a80U,0x7b1fU,0x86f4U,0x49caU,0xa11eU,0x7c64U,0x0c26U,0xbaf3U,0x5421U,0xeedfU,
		0xf293U,0x3b90U,0xa4a1U,0x3d97U,0x97e0U,0x1840U,0xf2aeU,0x1283U,0xbf1cU,0x7061U,
		0xbf82U,0x13e5U,0x9550U,0xf4c8U,0x73d5U,0xef78U,0xebacU,0x3a67U,0xc3e0U,0xf3acU,
		0xec36U,0xf3e4U,0x2ce7U,0xc5cfU,0x6c68U,0x207dU,0xe417U,0x72e0U,0x0d48U,0x1e3dU,
		0xee64U,0x4376U,0x1713U,0xf7edU,0xc85dU,0xa192U,0x6dc2U,0x11acU,0x16b0U,0xac4dU,
		0xd527U,0x0131U,0x6074U,0x2eceU,0x5811U,0x3f09U,0x8731U,0xf97aU,0x418dU,0x6496U,
		0x65faU,0x5f2cU,0xc2ffU,0x3fd5U,0xc277U,0x0858U,0x4966U,0x18d4U,0x6556U,0xe899U,
		0xf3c4U,0x5d0bU,0x1134U,0xb7feU,0x577dU,0x8d1cU,0x59f6U,0x82d8U,0x7b59U,0xa8bbU,
		0xc2a1U,0x37faU,0x02bdU,0x8e20U,0x882dU,0x659fU,0x7f70U,0x2966U,0xc567U,0x6321U,
		0x2459U,0x988eU,0x061bU,0x661eU,0x19abU,0x5bd8U,0x4ee6U,0x6437U,0xfa9eU,0x27ebU,
		0x5227U,0x9be6U,0x8aefU,0xab3eU,0x9c1aU,0xb6c0U,0xb1daU,0x133aU,0xf1cfU,0xa456U,
		0x5a36U,0xe8a3U,0x0db5U,0xfac5U,0xda28U,0xe298U,0x0c82U,0x1ff4U,0x26dcU,0xc11aU,
		0x6d10U,0x31efU,0x5c75U,0x1fa3U,0x2584U,0xe1b1U,0x128dU,0x6266U,0x39e6U,0x8d59U,
		0xc15cU,0xa022U,0x055eU,0x8145U,0x0f03U,0x5bc7U,0x25ccU,0x6d06U,0x43bcU,0x7632U,
		0xc375U,0x1995U,0x69f2U,0xb5b9U,0x34f5U,0x33e4U,0x2418U,0xc7e2U,0x765eU,0xce3fU,
		0xf0b9U,0x2237U,0xa2ceU,0x4b10U,0x435fU,0x89c0U,0x4c62U,0x0b9bU,0x6c2eU,0x8931U,
		0xeac6U,0xe244U,0xcbfdU,0x25e2U,0x83d4U,0x7398U,0x5c26U,0xe0e9U,0x4794U,0xe7e4U,
		0x6c87U,0xeeaeU,0x3f59U,0xd4f7U,0x23f7U,0x5571U,0xc01cU,0x32b8U,0x5321U,0x0a78U,
		0xfc93U,0x09d0U,0x8c4fU,0xdbbdU,0x854cU,0xe029U,0x13c0U,0x6901U,0xca0bU,0x6614U,
		0x42c2U,0x5e7dU,0x6cd1U,0xa74dU,0x34c5U,0xfb03U,0xa04bU,0x6e15U,0x5104U,0x9d53U,
		0x0357U,0x8d62U,0xfb29U,0xae5dU,0xef81U,0xe374U,0x8e53U,0x3e77U,0xd920U,0x60ccU,
		0x4520U,0x5ab1U,0xd1b3U,0xb38dU,0x0fadU,0x96efU,0xf33eU,0x1d1aU,0x180dU,0x5dcdU,
		0xb7e5U,0xbc83U,0x84f7U,0xa87aU,0x0e6cU,0xa4daU,0x4c40U,0x49f5U,0xe94aU,0x9fc1U,
		0xba5aU,0xaddcU,0x4fc7U,0xb3f4U,0xc838U,0x946fU,0xf830U,0xeb3cU,0x35b6U,0x2a3dU,
		0x8ed5U,0x99d0U,0xe49cU,0xeee7U,0xbeb7U,0xcf72U,0x7367U,0x49c3U,0xa01bU,0x6a83U,
		0x4c20U,0x92a6U,0xd2ffU,0xc0f0U,0x1031U,0x84c7U,0x3699U,0xfa94U,0xdcd1U,0x6a8aU,
		0x7332U,0x1bf4U,0x2a60U,0xfea2U,0x5609U,0x42b9U,0x51d0U,0x6fd9U,0x3e56U,0xaa61U,
		0x3029U,0xa3ddU,0x79c8U,0x1009U,0x9907U,0xb957U,0x2302U,0xb69eU,0x6278U,0xeb70U,
		0x486cU,0xf828U,0x000dU,0xb9e3U,0x0a27U,0x94b6U,0x2433U,0x799cU,0xb9e7U,0xad2bU,
		0x5ac2U,0x0e19U,0x0110U,0x7b12U,0xb138U,0x7859U,0xe4bcU,0x8bc2U,0x138bU,0x8597U,
		0x8a14U,0x3194U,0x1badU,0x6c1cU,0x0011U,0x27b1U,0x852cU,0xd1c3U,0x3b83U,0xe73aU,
		0xf1e3U,0xc7d1U,0xb386U,0x6598U,0x0ddbU,0x61c1U,0x1bddU,0x5461U,0xb995U,0x82abU,
		0x94baU,0x7fd9U,0xa3c0U,0xbc4fU,0xf853U,0x8b3eU,0xd1ebU,0x9f38U,0x0c1dU,0x4682U,
		0x6658U,0x0474U,0x17c4U,0x9c00U,0xec1aU,0xab9dU,0x0a2bU,0xc05dU,0x9e8cU,0xe818U,
		0xa0ebU,0x2c45U,0xaaecU,0xddaeU,0x34edU,0x36b4U,0xd9c7U,0x38fcU,0xe8daU,0x1f88U,
		0x0e8eU,0x7e3eU,0x79a0U,0xd5c7U,0xb70bU,0x6695U,0xd026U,0xb0c4U,0x46b2U,0xe6a3U,
		0x81acU,0x6a35U,0x5b50U,0x75afU,0x1182U,0xfcc0U,0x7c15U,0x9fa9U,0x2437U,0x6f97U,
		0x0dfbU,0x154cU,0x399aU,0x0bb0U,0xa15dU,0x617cU,0xab05U,0x38cbU,0x78f6U,0x2c3eU,
		0x61d6U,0x88d9U,0x3522U,0xe0c7U,0xad2fU,0x0b7eU,0xa657U,0x13c7U,0xdc05U,0xb0a2U,
		0x3595U,0x912dU,0x8072U,0x6338U,0x89f3U,0xbd1eU,0x5fc2U,0x562aU,0x77bdU,0xacf7U,
		0xb9ebU,0xdaffU,0x3e33U,0x5b8bU,0x5090U,0x6305U,0xf770U,0x0539U,0xaa11U,0xebf7U,
		0x051aU,0xde7cU,0x21fcU,0x4b7bU,0xd5eeU,0xcd55U,0x0e1dU,0xc1a2U,0x3746U,0xa3ffU,
		0xac81U,0x356aU,0x2de9U,0x79efU,0x6bf0U,0x4aecU,0x157cU,0xa99fU,0x094fU,0x8e90U,
		0x1de1U,0x049dU,0x5ac6U,0xa2a4U,0x8db8U,0x232eU,0x0b47U,0xe79cU,0x0687U,0x3553U,
		0x458fU,0x90caU,0xf611U,0xa1e4U,0x4707U,0x1f3fU,0x4e0aU,0x772fU,0x7b16U,0x9d78U,
		0x14dcU,0x6e25U,0xc359U,0xcc32U,0x5710U,0xcb8aU,0xbe4eU,0xe74fU,0x2e42U,0x4388U,
		0x1656U,0xb069U,0x3c57U,0x96b2U,0x0114U,0x28aaU,0xb3d6U,0xf472U,0x81d7U,0x78fdU,
		0xe5b1U,0x214dU,0xcc1eU,0x172cU,0x1a7bU,0xded4U,0xb4f9U,0x6de2U,0xd167U,0x7f0aU,
		0x785dU,0x58aaU,0x06dbU,0xec09U,0xad1bU,0x1c6cU,0xb688U,0x7548U,0x936eU,0xa34dU,
		0xbce4U,0xc28bU,0x1815U,0x7018U,0xa436U,0xa52fU,0xb13cU,0x5ff7U,0x476bU,0x2287U,
		0x9adbU,0x289aU,0x500eU,0x0c67U,0x39b4U,0x0032U,0xd57aU,0xb76cU,0xe6e3U,0x2abcU,
		0x1b00U,0x27cfU,0x8bc6U,0xf47bU,0xc8d4U,0x6908U,0x2e8cU,0x4be0U,0x047dU,0x3c23U,
		0x4faeU,0x2038U,0xb36eU,0x4f1dU,0xc511U,0x995fU,0x12bfU,0x32faU,0xe4c0U,0x5f0eU,
		0xf67fU,0x6ce9U,0xbb3eU,0xbcbeU,0x8858U,0x2fd0U,0x6206U,0xf0f2U,0xad67U,0xe5e8U,
		0x7b4bU,0xfc21U,0xf6d7U,0xb567U,0x859bU,0xd043U,0xdcb5U,0x878aU,0xeea1U,0x931dU,
		0xadf1U,0x97a4U,0x0fc0U,0xb59fU,0xaa4cU,0x0531U,0xb01fU,0xd519U,0xcf02U,0x742aU,
		0x138fU,0xdff7U,0x2e23U,0xb88dU,0xdb81U,0xe237U,0x350fU,0x349bU,0xa5d2U,0x02e1U,
		0x0ef5U,0xbb35U,0x48ecU,0x04c1U,0x19ccU,0x8a45U,0x3198U,0x00c4U,0x96e2U,0x18c1U,
		0xc552U,0xc2ebU,0x80dcU,0xf088U,0x222fU,0xed2eU,0xd8cdU,0x244bU,0x1a8dU,0xb9a8U,
		0x2ca1U,0x634dU,0x8a18U,0xa264U,0xd756U,0x6e08U,0xd86aU,0x88e0U,0x2f7dU,0xa41dU,
		0xbf79U,0xd366U,0x22ccU,0x24dcU,0x8722U,0x6b47U,0x412bU,0x9afcU,0x6c20U,0xeb20U,
		0xc477U,0x3449U,0xfb3bU,0x4fdbU,0x3221U,0x275aU,0x49e8U,0xed74U,0x1938U,0x4cf1U,
		0x849eU,0x4ed7U,0x477cU,0x4493U,0x1bb1U,0x34b7U,0xe26dU,0xf9b4U,0x845cU,0x1087U,
		0x3d14U,0x0267U,0xccabU,0x6a76U,0x0348U,0x9b30U,0x5217U,0x2651U,0x7359U,0x20e7U,
		0x27b5U,0xc8e4U,0x34d6U,0xca12U,0x8432U,0xf542U,0x84baU,0x0bd7U,0xcda1U,0x0190U,
		0x596dU,0x79b9U,0xd2eeU,0x22f2U,0xb57fU,0x51c2U,0x0015U,0x1ec6U,0x0b21U,0x5331U,
		0xe923U,0xda8eU,0x0a54U,0xaa54U,0x2711U,0xac32U,0x808eU,0x0ccbU,0x6489U,0xbb19U,
		0xb274U,0x6720U,0xd1c7U,0xe3f1U,0xe5d2U,0xd22bU,0x8f36U,0x899dU,0xd5a4U,0x8ffaU,
		0x3819U,0xf14cU,0x0577U,0x5531U,0xfeddU,0x37edU,0x3bfbU,0xdb4fU,0x8530U,0x5cacU,
		0xaef7U,0xf59eU,0xf6afU,0x4a38U,0x2badU,0xe7b9U,0x57abU,0x7774U,0x1e30U,0x9212U,
		0x385cU,0xe15dU,0x51d9U,0xa36cU,0xe73eU,0x7e21U,0x44aeU,0x4f14U,0xfba0U,0x9e24U,
		0xd447U,0x7dd2U,0x46d5U,0x2421U,0x92c9U,0x2ec1U,0xe786U,0xe669U,0xd0bdU,0x638eU,
		0x3b87U,0x1d82U,0x8317U,0xb0f0U,0x85fcU,0x61ddU,0xdfceU,0x4300U,0xc718U,0x5d7eU,
		0xe613U,0x101eU,0x8337U,0x5a49U,0x1364U,0x27d9U,0xc7d5U,0xa079U,0x020eU,0x25f2U,
		0x10a1U,0x37d2U,0xdb65U,0x1748U,0xa1eeU,0x1d93U,0xacbeU,0x787bU,0x5dd5U,0x9137U,
		0x4caaU,0x3105U,0xf1e7U,0x6cf9U,0xbc59U,0x5e61U,0xaff2U,0xe9a5U,0x7c61U,0x5da1U,
		0x13e2U,0x3d94U,0x72ddU,0xf3a9U,0x2ecbU,0xa18fU,0xe896U,0x5f29U,0x659cU,0x82d5U,
		0x9be3U,0x661bU,0x1ff1U,0xa453U,0x8142U,0xe1aeU,0xce3cU,0x1992U,0x7395U,0x0b98U,
		0x09cdU,0xd4f4U,0x6e12U,0x6611U,0xb38aU,0xe371U,0x9fbeU,0xbc80U,0xcf6fU,0xeb39U,
		0x1bf1U,0xc0edU,0xb69bU,0xaa5eU,0x06f4U,0xc1b7U,0x4441U,0x7b03U,0x6295U,0xd1dbU,
		0x61c5U,0xc8a2U,0x563dU,0x8084U,0x1302U,0x897dU,0x4269U,0x6f17U,0xecd4U,0x5799U,
		0xfd8cU,0x1a72U,0xfcb1U,0x8b6bU,0x0677U,0x75ceU,0x0ddfU,0x7b3aU,0x7f41U,0xc6caU,
		0x52b5U,0x8757U,0xe569U,0xa7d1U,0x5c6dU,0xe3d7U,0x0340U,0x78eeU,0x14c2U,0x72b1U,
		0x58d7U,0x81eeU,0x5465U,0x3211U,0xc533U,0xe064U,0xff9bU,0x7304U,0xb453U,0x44f5U,
		0x4080U,0xe51eU,0x2aadU,0x9b79U,0xefecU,0x6d3fU,0x4b2cU,0x1c07U,0x1be1U,0x4606U,
		0x3feaU,0x5693U,0xceb9U,0xe0ceU,0x1906U,0xe44eU,0xfb22U,0x5776U,0x2e3bU,0x6651U,
		0x92c2U,0xd8c6U,0xe0aeU,0x2aa6U,0x82afU,0xf533U,0x122dU,0x7e86U,0x030eU,0x02a3U,
		0xfd4fU,0x8993U,0x2914U,0x2863U,0xa610U,0x958bU,0x6980U,0x97f1U,0x6e69U,0x6237U,
		0xb999U,0x9876U,0x0c07U,0x88bdU,0x5bf3U,0xb4bcU,0xa620U,0x4995U,0xf4a5U,0xd486U,
		0x72b9U,0xcf8dU,0xd0a7U,0xd6e3U,0x6aebU,0xe996U,0x7fddU,0xc4f2U,0x8237U,0x42c9U,
		0xa1c1U,0x54cdU,0x4dbaU,0x8d36U,0xf1a0U,0xb285U,0x68b0U,0x8743U,0xe391U,0xf053U,
		0xbb0aU,0x6f5eU,0x94beU,0x90daU,0x63e0U,0x0373U,0x0f89U,0xc224U,0xda22U,0xd521U,
		0xe496U,0x531bU,0x0df5U,0x8526U,0x4765U,0x8db2U,0xe267U,0x2e1dU,0xbc53U,0xaef1U,
		0x0c01U,0x7f3bU,0x260bU,0x75f3U,0xff56U,0x63acU,0x0168U,0x876fU,0xed8bU,0x6510U,
		0x9cfbU,0xf885U,0x5268U,0xdf50U,0xa3c4U,0x405fU,0x75f9U,0x1c63U,0xf876U,0x8016U,
		0x298bU,0xea40U,0x5d28U,0x9357U,0xa59aU,0xd8f3U,0x58e8U,0xa691U,0xfdc3U,0xd905U,
		0x8b42U,0xf64aU,0x0ff5U,0xc215U,0xfcf8U,0xb173U,0xb49fU,0x9068U,0x9eb5U,0xa397U,
		0xe0b5U,0xe346U,0x6864U,0x3dfbU,0xdf88U,0x076eU,0xf857U,0xf009U,0x72a2U,0x5a82U,
		0x8fb8U,0x3529U,0x19edU,0x9d23U,0x2a0bU,0x8900U,0xfd95U,0xd53eU,0xd72fU,0xcb16U,
		0xc72aU,0xe7afU,0x9f3cU,0x6c00U,0x09e1U,0x3348U,0xb481U,0x7602U,0x2162U,0xbf40U,
		0xaf3fU,0x99a1U,0x28e9U,0x8ce2U,0xb7edU,0x01cbU,0x9815U,0xf71aU,0xd1efU,0x2f1aU,
		0x2611U,0x487cU,0x4954U,0x9eefU,0x8819U,0xc3cbU,0xc264U,0x26b4U,0x8c87U,0x104aU,
		0xb6d8U,0x44d8U,0x15a1U,0xdef6U,0x4686U,0xb3dfU,0x25a1U,0x5e84U,0x0f25U,0xb08fU,
		0x6661U,0xd25eU,0xdc61U,0xefb0U,0xf94aU,0x44b7U,0x1dc9U,0xd655U,0x4bf1U,0xf98fU,
		0x0c21U,0x67f3U,0xf3bfU,0x0d43U,0x0efeU,0x8aeaU,0x50ffU,0x5c21U,0x2a5bU,0xba55U,
		0xd1e6U,0xb133U,0x3590U,0x46adU,0x570bU,0x2de4U,0x0478U,0xb683U,0x80d7U,0xadecU,
		0x84b5U,0x321cU,0xd442U,0xd59fU,0x813dU,0xdb60U,0xb44eU,0x4264U,0x4db5U,0xfd4aU,
		0xb49aU,0xff51U,0x665cU,0x215dU,0x63b2U,0xd43dU,0xb469U,0xd26dU,0x599bU,0xd6eeU,
		0x8e76U,0xd9faU,0xa938U,0x8647U,0xfb77U,0xdfb4U,0x039dU,0x1f5fU,0x9c04U,0xa3a2U,
		0x2676U,0x3cdaU,0x3c3bU,0x85c9U,0x63b7U,0x4110U,0x9857U,0x166fU,0xf53aU,0xb09aU,
		0x15dcU,0x54c5U,0xcde0U,0x9256U,0x17c8U,0x7caeU,0x48afU,0x974bU,0x3353U,0x0b85U,
		0xe3a1U,0x3e1fU,0x62d9U,0xe5faU,0xae98U,0xd420U,0x0d62U,0x510eU,0x2b79U,0x201fU,
		0xaba1U,0x36ddU,0xdcdbU,0x969eU,0xf935U,0x5682U,0xd272U,0xf703U,0xc77eU,0x1a17U,
		0x1a00U,0xc773U,0xdb95U,0x7cb9U,0x29b3U,0xc309U,0xec1eU,0x687fU,0xff5cU,0x9a1aU,
		0x9a09U,0xf3fbU,0x67a8U,0x7e17U,0x5c7fU,0x4feeU,0x14caU,0xb919U,0xe9bbU,0x47bdU,
		0xbf26U,0x4ad0U,0xc061U,0x6054U,0xd149U,0x6cd8U,0x62eeU,0x15e4U,0xb46eU,0x71b0U,
		0x2835U,0x815aU,0xe439U,0x8225U,0x2bdbU,0x2014U,0x6e87U,0xab63U,0x0a2fU,0xf392U,
		0xf58dU,0xe46fU,0x503aU,0x4168U,0x1b6cU,0xb027U,0x2af3U,0x9d9aU,0x7c30U,0x1c77U,
		0x78c0U,0x484aU,0xc88fU,0x3a39U,0xe81cU,0xf8faU,0xc3c2U,0x0c5eU,0x6913U,0xa787U,
		0xd6f3U,0x9e6aU,0x862cU,0xb938U,0x482dU,0x85f2U,0xe170U,0x8577U,0xbb4fU,0x428cU,
		0x9e90U,0x4af7U,0x8775U,0x17f2U,0xd3e1U,0x9974U,0xcb95U,0x6e7cU,0x6d9fU,0xcdb5U,
		0xefc8U,0xed1aU,0x134fU,0x239eU,0x6fe2U,0xe682U,0x2c49U,0x5720U,0x83e5U,0x7276U,
		0xc454U,0x99b0U,0x59a0U,0x1f93U,0x0ebcU,0xc7ebU,0x82b6U,0xe32dU,0x1e03U,0xac17U,
		0xb4ddU,0x52f5U,0xa0efU,0x10f2U,0xb8e7U,0x7377U,0x2517U,0xad36U,0x945fU,0x8e42U,
		0x32d8U,0x9d8fU,0x150aU,0x2bfdU,0x1c18U,0x04f9U,0x430bU,0x7dacU,0xddb2U,0x55ebU,
		0x8678U,0x6b85U,0xec54U,0x6d86U,0xd9ffU,0xb3a5U,0x8565U,0x777fU,0xdac7U,0x30b4U,
		0xbc8bU,0x2f6aU,0x91a4U,0x2eb7U,0xaaf0U,0x9b8fU,0x016eU,0xe9b0U,0xbbb7U,0xaebcU,
		0xecfaU,0xf2faU,0x66d8U,0x4ec5U,0x0cfbU,0x5143U,0x9a34U,0xe002U,0x4d3fU,0xf2c2U,
		0x36b8U,0x3b3aU,0x6358U,0xa754U,0xf9f8U,0x383aU,0x8e7bU,0x3777U,0xf2aaU,0x7a7cU,
		0x16acU,0x2ce3U,0xc29dU,0xc273U,0x0db1U,0x4ee2U,0x34f1U,0x39e2U,0xc018U,0xeac2U,
		0xd91cU,0x6ccdU,0x8ed1U,0x0e68U,0x79c4U,0x3695U,0x8293U,0xcbddU,0x4bcbU,0xbf35U,
		0xd97aU,0xce88U,0x3900U,0xdd93U,0x2179U,0xdd48U,0xedbbU,0x238dU,0x864cU,0xaa91U,
		0xd338U,0x993eU,0x86c8U,0x6bb2U,0x9611U,0xa89aU,0xd209U,0x76b0U,0xd9cbU,0xae6bU,
		0x6516U,0xe3b2U,0xcb05U,0x07b3U,0xd80cU,0x424cU,0x4507U,0x08beU,0xcb83U,0x9f31U,
		0x7dcbU,0xf081U,0x9061U,0x44eeU,0x1f8cU,0x4109U,0x4904U,0x2640U,0xc094U,0x40b4U,
		0xa93dU,0x2647U,0xad84U,0xfa12U,0x7e8dU,0x3d02U,0x71d6U,0xff78U,0xdc43U,0x7cd7U,
		0xe8deU,0x0aa3U,0x94a5U,0x818cU,0xd6c8U,0x13ceU,0x47f8U,0x48d9U,0x4655U,0x2113U,
		0x8454U,0x3992U,0x9a01U,0x5bebU,0x072fU,0xddceU,0x7e42U,0x9ee4U,0x5c4eU,0x3a77U,
		0xa327U,0x063aU,0xdfb9U,0x9385U,0x9a73U,0xd505U,0x1855U,0xf746U,0x0f6cU,0x9081U,
		0xd792U,0xa5afU,0x0e92U,0xbd2eU,0xed91U,0xfef7U,0xddd6U,0xab72U,0x7820U,0xcf18U,
		0x9d4cU,0x5d04U,0x81f6U,0x05f8U,0xbe68U,0x4c70U,0xb1a3U,0x28deU,0xd5cbU,0xd678U,
		0xa548U,0x34ccU,0xc6b7U,0x925eU,0xfb7cU,0x5479U,0x2b37U,0x33a0U,0xdf7fU,0x3c4eU,
		0x3172U,0x788fU,0x4e47U,0x98e6U,0x79a4U,0xc0b9U,0xdc87U,0x091cU,0x5eb2U,0x4423U,
		0x2a1bU,0x7432U,0x0c7aU,0x1a9fU,0x9104U,0x744dU,0x9802U,0x8c6aU,0x4a21U,0xaf96U,
		0x6699U,0xb24bU,0x4f3dU,0xb970U,0xb5d1U,0x103fU,0x1f64U,0xf9d6U,0x9d66U,0x1d34U,
		0x87a0U,0x370bU,0x078aU,0x7d82U,0x2787U,0xb30aU,0xb70fU,0x3481U,0xf88bU,0xb9ccU,
		0x9f71U,0x92f1U,0xcd23U,0x05ceU,0x2a91U,0xc4e0U,0x6994U,0xa8eeU,0xa950U,0x923fU,
		0x0becU,0x51b8U,0xb0c8U,0xa3ebU,0x5b28U,0x0f1aU,0x17aaU,0x9421U,0x03a2U,0x490bU,
		0x41caU,0x365dU,0x1234U,0xe860U,0xe6f5U,0xabb0U,0x7898U,0x31c9U,0xd02aU,0xd3b2U,
		0xc3fbU,0xc966U,0x54a5U,0xa65eU,0x1c2bU,0x9762U,0xa380U,0xf36eU,0x8b74U,0x0d82U,
		0x57c8U,0xc9f1U,0x338dU,0x93e9U,0xe6a7U,0x3f82U,0xf328U,0x6abfU,0x3f6bU,0x801fU,
		0xa3a7U,0x5bb8U,0xd122U,0x67e8U,0x4da0U,0x8961U,0x84ffU,0x0e7aU,0x0698U,0x197fU,
		0x46b6U,0x709eU,0x9d01U,0x59b9U,0xc940U,0x043eU,0xa8acU,0xa790U,0x209dU,0x6b07U,
		0x74eaU,0x2707U,0x4cc3U,0x548aU,0x7320U,0xbb99U,0x6a39U,0x7906U,0x2a71U,0xfb0aU,
		0x4eb0U,0x688eU,0x9c09U,0x50f4U,0x35f5U,0x49b4U,0xfdebU,0x9e2dU,0x8be9U,0x7a30U,
		0xee89U,0x4b5bU,0x81b0U,0x83acU,0x88f7U,0x1723U,0x04caU,0xf365U,0x5801U,0x21c6U,
		0x8922U,0x99deU,0x9ba9U,0x78deU,0xdc76U,0xe425U,0x917bU,0xace4U,0x75b3U,0xf1ffU,
		0xcc8dU,0x60b7U,0xb9beU,0x3887U,0x3cdfU,0x1969U,0x952aU,0x9502U,0x6500U,0x97c5U,
		0xca32U,0xdf45U,0xed49U,0x3932U,0x5b54U,0xd388U,0xdf56U,0xfd20U,0x4724U,0x31bbU,
		0x299bU,0xa2f3U,0xdeaaU,0x59ddU,0x2feeU,0x9187U,0xc1f5U,0xd410U,0x276eU,0xa977U,
		0xfcc4U,0x03b4U,0x0d74U,0x5f3cU,0x393eU,0x0af9U,0x267bU,0x40bbU,0x559aU,0x2edcU,
		0x02aaU,0x5d1dU,0x2cd3U,0x7649U,0x0846U,0x03f3U,0x1186U,0xfcd0U,0xf4edU,0x50e6U,
		0x8fccU,0xb0a9U,0xfcdcU,0xed6aU,0xb8a2U,0x315bU,0x195dU,0x8391U,0x600fU,0x2f51U,
		0xea7cU,0xb75cU,0x9fadU,0xe768U,0x7123U,0x618eU,0x7564U,0x529eU,0x85ceU,0x8eaaU,
		0x2632U,0xfce6U,0x4ce9U,0x75a7U,0xc76bU,0x9583U,0x759fU,0xf73eU,0x7c19U,0x8da7U,
		0x526eU,0x7a0aU,0x0737U,0x4974U,0xde42U,0x70b5U,0xdbf3U,0xce7aU,0x58dfU,0x46feU,
		0x31afU,0xa947U,0x489eU,0x703dU,0x6f9bU,0x0cbbU,0xa552U,0xa052U,0x53adU,0xcde8U,
		0x3c40U,0xb86dU,0x1332U,0x1493U,0x915bU,0x2218U,0xc740U,0x64f4U,0x74d5U,0x0bf6U,
		0x243bU,0x8e2eU,0xea72U,0x4121U,0x9bbeU,0x7bd4U,0x2f33U,0xcf0aU,0x74f4U,0x770bU,
		0x5647U,0x29e5U,0x3954U,0x2a2cU,0x900cU,0x16efU,0x1550U,0xe8eaU,0xeb81U,0xf244U,
		0x2d44U,0xfd59U,0x4115U,0x91bcU,0x54e8U,0x059eU,0xc665U,0xd215U,0x8cf9U,0x1e5cU,
		0xa0b6U,0x5eebU,0x0dffU,0x55caU,0x4065U,0x5a06U,0xce94U,0x5a77U,0x723aU,0xc734U,
		0xddfbU,0x31fdU,0x4c7cU,0x4c8aU,0x87fdU,0xa82aU,0x3e5fU,0x89d8U,0x0bb4U,0x67b2U,
		0xb214U,0x0f47U,0xe10dU,0x317eU,0x63bcU,0xc09bU,0x1487U,0x7667U,0x0315U,0x91aeU,
		0xda52U,0x9c28U,0x7a00U,0xf9aaU,0x399eU,0x8fe1U,0x3319U,0x6859U,0x4837U,0xdc0cU,
		0xcd0fU,0x1861U,0xc696U,0x4892U,0x889aU,0x3acaU,0xf9e2U,0x6ba0U,0x613cU,0x1b0aU,
		0x6180U,0x2245U,0xf732U,0xcd2fU,0x86a4U,0x3002U,0x1674U,0x46eeU,0xf61bU,0x982aU,
		0x01acU,0xf573U,0xa882U,0x359fU,0x3669U,0x1e6aU,0xa161U,0x702fU,0xa3caU,0xdbe7U,
		0xae13U,0xc32fU,0xf563U,0xc90fU,0xb617U,0xa38cU,0xba8dU,0x551fU,0x92d3U,0x3ee0U,
		0x97b7U,0x9d0dU,0x38cfU,0xca87U,0xc20cU,0x6e1cU,0x2f45U,0x0f11U,0x985cU,0xb3c3U,
		0x6acbU,0xcbcdU,0xc761U,0x8494U,0xb333U,0x04d6U,0xd760U,0xed5cU,0xab09U,0xaf53U,
		0x689aU,0x3962U,0x6932U,0x7117U,0xf4daU,0xfdb8U,0xe850U,0x0a5eU,0x7b0fU,0x94b3U,
		0xe737U,0x3191U,0x8b3bU,0x545eU,0x2c42U,0x9bfdU,0xb0c1U,0x1f85U,0x0badU,0xfcbdU,
		0xb09fU,0x88d6U,0x6302U,0x5627U,0x3567U,0x4b78U,0xe799U,0x8e8dU,0xcc2fU,0x1f3cU,
		0x78faU,0xb066U,0x1c69U,0x6ddfU,0x2897U,0x7015U,0x4bddU,0x2ab9U,0xbcbbU,0x995cU,
		0x931aU,0xfc1eU,0xe234U,0xd516U,0xc2e8U,0x04beU,0x88ddU,0xb9a5U,0x4fd8U,0x6b44U,
		0x1084U,0x4ed4U,0xf53fU,0x264eU,0xda8bU,0x22efU,0x899aU,0xbb16U,0x4a35U,0x37eaU,
		0x9e21U,0xe15aU,0x61daU,0xe666U,0x37cfU,0x5a46U,0xe9a2U,0x9134U,0xa450U,0xa18cU,
		0xeb36U,0xd4f1U,0x897aU,0x7b00U,0x8754U,0x8b68U,0x7301U,0x72aeU,0xe0cbU,0x6d3cU,
		0x02a0U,0xd8c3U,0xb4b9U,0x97eeU,0x54caU,0xd6e0U,0xc221U,0xf050U,0x75f0U,0x8dafU,
		0x8013U,0xf882U,0xb170U,0xa68eU,0x3526U,0x3df8U,0x75ffU,0xcb13U,0x9eecU,0x01c8U,
		0xb08cU,0x44d5U,0x8ae7U,0xd652U,0x3219U,0x46aaU,0xd26aU,0xfd47U,0x85c6U,0xdfb1U,
		0x0b82U,0x54c2U,0x567fU,0x510bU,0xf3f8U,0x7cb6U,0x15e1U,0x47baU,0x4165U,0x2011U,
		0xa784U,0x4847U,0x9971U,0x8574U,0x99adU,0x239bU,0xad33U,0xac14U,0x6d83U,0x04f6U,
		0xaeb9U,0x2f67U,0x3837U,0xdfffU,0x6ccaU,0xc270U,0x238aU,0xbf32U,0x07b0U,0xa897U,
		0x40b1U,0xf07eU,0x13cbU,0xff75U,0x0637U,0x5be8U,0xab6fU,0x907eU,0x925bU,0x4c6dU,
		0x4420U,0x788cU,0x103cU,0x8c67U,0x92eeU,0x7d7fU,0x941eU,0x923cU,0xa65bU,0xabadU,
		0x801cU,0xc9eeU,0x043bU,0x0e77U,0x688bU,0x5487U,0xf362U,0x7a2dU,0x3884U,0xe422U,
		0x31b8U,0xdf42U,0x0af6U,0xd40dU,0xb0a6U,0x7646U,0x529bU,0x2f4eU,0x4971U,0x9580U,
		0xcde5U,0xa944U,0x7bd1U,0x64f1U,0xfd56U,0x2a29U,0x5a74U,0x1e59U,0x317bU,0xa827U,
		0xdc09U,0x9c25U,0x2fffU,0x6b9dU,0xc32cU,0x359cU,0x0f0eU,0x3eddU,0x7114U,0x04d3U,
		0xfcbaU,0x318eU,0x7012U,0x8e8aU,0x4ed1U,0xd513U,0x9131U,0x37e7U,0x97ebU,0x8b65U,
		0x01c5U,0xf87fU,0x7cb3U,0xfd44U,0x2f64U,0x8571U,0x907bU,0xa894U,0x0e74U,0x7d7cU,
		0x957dU,0xdf3fU,0x3599U,0x1e56U,0xf87cU,0x8e87U,0x7d76U,0x7d79U,0xdf6dU,0x691cU,
		0x1918U,0x846fU,0x967fU,0x2e71U,0xe9f9U,0x691fU,0x20b8U,0x254cU,0x633cU,0x81e0U,
		0x9980U,0x9d5aU,0x707bU,0xdf70U,0x17cdU,0x098eU,0xe635U,0xf528U,0xbdf9U,0xfba9U,
		0x507dU,0x8472U,0x2821U,0x83f1U,0x8076U,0x9f98U,0xaf8aU,0x7ed0U,0x48f5U,0x191bU,
		0x09f9U,0x927bU,0x7383U,0xfdffU,0x8a74U,0xfe23U,0x7554U,0x2e74U,0x7293U,0x53d0U,
		0xbd22U,0x2e5dU,0x24ceU,0x501aU,0xf272U,0x9682U,0x9750U,0x7ae1U,0x1e94U,0x3c92U,
		0x4017U,0x4075U,0x45d3U,0x6922U,0x2044U,0x8c78U,0x89f7U,0x7814U,0xea46U,0x0bc9U,
		0xc9d7U,0xe9fcU,0xd37eU,0xdfedU,0x66c0U,0x6212U,0xc084U,0x5f62U,0x72caU,0x254fU,
		0x3986U,0x8dc4U,0x562eU,0x9842U,0x311fU,0xeaedU,0xb87fU,0x20bbU,0x48b4U,0xfa19U,
		0xa40bU,0xb2edU,0x286aU,0x9f61U,0x65d0U,0x81e3U,0xffd1U,0xa31bU,0x5fc6U,0xa601U,
		0xcc3eU,0x0798U,0x9c86U,0x633fU,0x0c38U,0x0a8cU,0xaf1dU,0x0e56U,0x015fU,0x6be7U,
		0x4417U,0x9d5dU,0x2bbfU,0xac94U,0xacfbU,0xb9f7U,0x6a94U,0xd46fU,0xb03aU,0x9983U,
		0x0b8aU,0xa8fcU,0x642fU,0x7185U,0xa290U,0x7c21U,0x68daU,0xdf73U,0x7d95U,0x3fdfU,
		0x77c1U,0x9c76U,0x2991U,0x3505U,0x7e4aU,0x707eU,0x3644U,0x5d93U,0x1fadU,0xcb35U,
		0x546dU,0x1242U,0xd13aU,0x0991U,0x706bU,0x652aU,0xdb03U,0x2064U,0xb7b2U,0x035eU,
		0xf128U,0x17d0U,0x3358U,0x7d3cU,0x4f06U,0x943aU,0xeb18U,0xd5bfU,0x36d5U,0xf52bU,
		0xe760U,0x9edcU,0xb9efU,0x6d34U,0xc870U,0x12d6U,0xdd0bU,0xe638U,0x9dceU,0x1397U,
		0x0d28U,0x6712U,0x61b3U,0x2745U,0xf7b9U,0xfbacU,0xd946U,0xb2a3U,0x5b8fU,0x429fU,
		0x61e9U,0x46c4U,0xad0bU,0xbdfcU,0x3e24U,0x0893U,0x9932U,0x804dU,0x6735U,0x784bU,
		0xf41aU,0x8475U,0x78d4U,0xe5deU,0x3e37U,0xcae5U,0x935dU,0x4436U,0x6cafU,0x5080U,
		0x8f81U,0x0dccU,0xf5aaU,0xd96eU,0xa72cU,0xf254U,0x5b7dU,0x83f4U,0xaa6aU,0x3f90U,
		0x6309U,0xcb68U,0xfb90U,0x8123U,0x5d4eU,0x2824U,0xe3a6U,0xad8bU,0x06ccU,0xce48U,
		0x291bU,0x08ffU,0x5dadU,0x9f9bU,0xd0e4U,0x2315U,0x5094U,0x0078U,0x9790U,0xdb71U,
		0xecc3U,0x8079U,0x2c1dU,0x9605U,0xc174U,0x156dU,0xc628U,0xa242U,0xeddfU,0x7ed3U,
		0xa991U,0x10d2U,0x053dU,0x3729U,0x2e14U,0xcd4cU,0xf075U,0xaf8dU,0xe5ffU,0xad73U,
		0x12f2U,0x3b6eU,0xb9b4U,0xee97U,0xa4e2U,0x191eU,0x724fU,0x213fU,0xf774U,0x43daU,
		0x5d2eU,0x22bdU,0x56efU,0x48f8U,0x2621U,0x066cU,0xa4fbU,0x8fa8U,0xd714U,0x4137U,
		0xcab3U,0x927eU,0x6116U,0xe126U,0xebfbU,0x9774U,0x7ccbU,0x8d69U,0x449fU,0x09fcU,
		0x62deU,0xbd5eU,0xebddU,0xe624U,0xfeb6U,0x7f82U,0xcf4dU,0xfe02U,0x87c8U,0x5fe2U,
		0xaa15U,0x27c1U,0x33fdU,0x4649U,0x5363U,0x7386U,0x1ed2U,0xbba7U,0x9921U,0xe3ccU,
		0x8380U,0x7c89U,0x682eU,0xfe26U,0x4e68U,0xa7a6U,0xde80U,0x36f8U,0x8f8eU,0x9ea9U,
		0xde35U,0x8a77U,0xd425U,0xb85eU,0xf439U,0x34acU,0x8303U,0x9f20U,0x779dU,0x2e77U,
		0x96feU,0xaabeU,0x051eU,0x7de2U,0xd8f9U,0x2d83U,0x4de4U,0x7557U,0x03e4U,0x6f25U,
		0x08a0U,0x4551U,0x41c0U,0xbe44U,0x2734U,0x53d3U,0xc181U,0x4f90U,0x4b7fU,0x555fU,
		0x914eU,0x7fc4U,0xef61U,0x7296U,0xae9dU,0x3d09U,0xf16eU,0xc900U,0x9592U,0xe577U,
		0xce55U,0x2e60U,0xd31fU,0xbddfU,0x2200U,0x624fU,0x986dU,0x6f8eU,0x3defU,0xbd25U,
		0xe9cbU,0x293bU,0x16e0U,0x9b0eU,0x229dU,0xff88U,0xd005U,0x501dU,0xdec3U,0xe52cU,
		0xcd59U,0x12deU,0xd87fU,0xaa82U,0xdf11U,0x24d1U,0x5113U,0x4759U,0x916eU,0x41b0U,
		0x5d3bU,0x7a12U,0x9331U,0x9685U,0xe2ddU,0x9201U,0xd5f2U,0x2944U,0xa5a0U,0xbc9cU,
		0x3a7fU,0xf275U,0x7f2fU,0x42ebU,0x5c05U,0xca25U,0xe06cU,0x3a0dU,0x3ba3U,0x7ae4U,
		0x51aeU,0x7725U,0xc1a6U,0x5958U,0x601eU,0xfb30U,0x0a09U,0x9753U,0x0d67U,0x30cbU,
		0x2560U,0x525cU,0xca46U,0xc23fU,0x29cbU,0x3c95U,0xbf55U,0x4a8aU,0x0e21U,0xb908U,
		0xfc69U,0x6c45U,0xbb8aU,0x1e97U,0x5370U,0xb895U,0x934bU,0xdb1cU,0x2227U,0x0c15U,
		0x984fU,0x4078U,0x5592U,0xad7cU,0xa403U,0xda83U,0x7985U,0xf166U,0x84d9U,0x401aU,
		0x2024U,0xd17dU,0xac9dU,0xe099U,0x5c15U,0xadccU,0x0cdeU,0x6925U,0x5fd3U,0x775dU,
		0x374aU,0x9df5U,0xa697U,0xbfcaU,0xa9dbU,0x45d6U,0xb1b5U,0x94d1U,0x32b1U,0xac46U,
		0x888dU,0xd2c3U,0x7046U,0x8c7bU,0xe038U,0x5962U,0x356eU,0xdfdeU,0xc3f1U,0xb3ccU,
		0xbab9U,0x2047U,0x2b7eU,0x7e94U,0x53ddU,0xb15cU,0xa617U,0xa154U,0xb083U,0x7817U,
		0x98efU,0x7f26U,0xac85U,0x1607U,0x2051U,0x6808U,0x66f9U,0x89faU,0x7512U,0x6fafU,
		0x112cU,0x21cfU,0x75d7U,0xfd35U,0x9d34U,0x0bccU,0x761cU,0x4948U,0x79f3U,0x1468U,
		0x8b8aU,0x9632U,0xfbf7U,0xea49U,0x36e2U,0x8cb8U,0x08d7U,0x0adbU,0x922dU,0xfbe1U,
		0x4c48U,0xe9ffU,0x8cd6U,0xb649U,0x2dedU,0x87f0U,0x58eeU,0x0f97U,0x099eU,0xc9daU,
		0x2d1cU,0x9e73U,0xf452U,0xf110U,0xa29dU,0x27aaU,0x5ca5U,0xdff0U,0x4058U,0x7b33U,
		0x4af0U,0x2156U,0x347aU,0xae64U,0x55c3U,0xd381U,0xaba6U,0xb05fU,0xcadeU,0x780dU,
		0x9deeU,0x7ddbU,0x04afU,0x6215U,0xafd4U,0x54dbU,0x6bf4U,0x11e8U,0x1bc9U,0x1154U,
		0x8a4eU,0x66c3U,0x47a3U,0xe64aU,0x7c4eU,0x39fcU,0x2b25U,0xcaefU,0xcff7U,0x5f65U,
		0x9cabU,0x5cc0U,0xa9a3U,0x2c5cU,0x75c2U,0xf865U,0x1928U,0xc087U,0x96a3U,0x6790U,
		0x65e1U,0xe2c4U,0xa63fU,0x7bbfU,0x189cU,0x2552U,0x4400U,0xc49eU,0x1580U,0x621cU,
		0xd90bU,0x5c61U,0x47ebU,0x72cdU,0x7088U,0x3c14U,0xe2b9U,0xb594U,0x4dd6U,0x6bc2U,
		0xce2aU,0x8dc7U,0xee3cU,0xf658U,0x8e94U,0x54fcU,0x6648U,0x38f3U,0x7af7U,0x3989U,
		0xdce0U,0xff7fU,0x798dU,0x3d4aU,0x97f8U,0x21f2U,0xb849U,0x9845U,0x60fcU,0x4a08U,
		0x0953U,0x082eU,0x86bbU,0x050aU,0xd859U,0x5631U,0x9ca0U,0xf6ccU,0x538bU,0x616dU,
		0x30dbU,0xdf1cU,0x56aaU,0xeaf0U,0x7bdeU,0x7a4bU,0x04a1U,0xc878U,0x30f6U,0xa800U,
		0x6029U,0x3122U,0x5687U,0xc9c6U,0x481dU,0x3165U,0xd397U,0x5276U,0x04e0U,0x20beU,
		0x2dc0U,0x361dU,0x1de5U,0xf444U,0xfdc9U,0x8febU,0x5c56U,0xb882U,0xaea8U,0xefe0U,
		0xb196U,0xa928U,0xc53bU,0xef1aU,0x1a56U,0xfa1cU,0xbaabU,0x26f2U,0xa2a8U,0x0450U,
		0x9268U,0xef88U,0x20d8U,0x48b7U,0xf93aU,0x9d3fU,0xe512U,0xae02U,0xe043U,0x0c94U,
		0x7162U,0xb2f0U,0xb966U,0xa3f5U,0x5acaU,0xf15cU,0x1891U,0xb014U,0x98b2U,0xa40eU,
		0x5c41U,0x2e2bU,0x5494U,0x58cbU,0x8987U,0x1c93U,0x4d90U,0x9f64U,0x6f4fU,0xc6eeU,
		0x2332U,0xbc35U,0x7685U,0xf185U,0xce1cU,0x286dU,0xf708U,0x212aU,0x9663U,0xe302U,
		0x8a3dU,0x82a3U,0x1f57U,0x81e6U,0xa96fU,0x44e6U,0x8dbcU,0x04b6U,0xf650U,0x4f88U,
		0x1ca4U,0x65d3U,0xe959U,0x6c0aU,0xc41cU,0x791cU,0x0305U,0xd01dU,0xb4b0U,0xa31eU,
		0x3032U,0xdf08U,0xe7a0U,0x4ff8U,0x98a4U,0x9099U,0x3899U,0xffd4U,0xd277U,0x71ddU,
		0x3fcdU,0xf0c2U,0x6987U,0x9b99U,0xa205U,0xa604U,0x663dU,0x96d7U,0x0b4bU,0xea31U,
		0x277dU,0x1dd7U,0x8581U,0x5fc9U,0xabf6U,0x328fU,0x00a1U,0x5ed9U,0xff42U,0x7d5dU,
		0xc821U,0x079bU,0xcf81U,0xf485U,0x3557U,0x57bbU,0xa1a3U,0x5e1fU,0xc791U,0xcc41U,
		0x1a1cU,0xd54cU,0x1f49U,0x76d4U,0xec6eU,0xf6f4U,0x49fdU,0x6342U,0x69a7U,0xa1b5U,
		0x068bU,0x3874U,0x8b48U,0x5431U,0x2ad2U,0x9c89U,0x163dU,0x432aU,0xb2d5U,0x0093U,
		0x0402U,0x28f7U,0xe673U,0x0a8fU,0xeb08U,0x35e8U,0x90ceU,0x402dU,0x5dfdU,0xe37bU,
		0x6946U,0x0c3bU,0xc783U,0x6ea8U,0x5a97U,0x5918U,0x5380U,0xbf6eU,0x8ac5U,0x0e59U,
		0x88eaU,0xc618U,0x4593U,0xda16U,0x64a9U,0xf03bU,0x1773U,0xaf20U,0xc4d6U,0x90c0U,
		0x326bU,0xac3cU,0x81a6U,0xe4b6U,0x9368U,0x6beaU,0x49e2U,0x0fbaU,0xa1e8U,0x3813U,
		0x290eU,0xecceU,0xaf39U,0x0162U,0xc778U,0x8137U,0x855fU,0x8626U,0x9a6dU,0xd332U,
		0xd11cU,0x9d60U,0x262cU,0x9524U,0xf615U,0x54e2U,0xc21bU,0x62fcU,0x2f5eU,0x441aU,
		0x6429U,0x1e8eU,0x12ecU,0x992cU,0x9168U,0xf433U,0x08d1U,0xac97U,0x4817U,0x65dbU,
		0x1f43U,0x965dU,0xc393U,0x8559U,0x8f74U,0x2bc2U,0x1a05U,0x7cdeU,0x4022U,0x82caU,
		0x623eU,0xd5dbU,0xd600U,0xb9faU,0x70ffU,0x4176U,0x470bU,0x9ce7U,0x7da0U,0x7e01U,
		0xa216U,0xacfeU,0x2cf6U,0xf13aU,0xf96cU,0xe7f1U,0x3971U,0xdce9U,0xf20cU,0xd472U,
		0x21bcU,0x2fc6U,0x7733U,0xe640U,0x1768U,0xcf64U,0x303bU,0x6a97U,0x7cbeU,0x13b1U,
		0xc399U,0x0da5U,0x4df4U,0x497cU,0xa914U,0x9986U,0x5b61U,0x1ae1U,0x4e0eU,0xe9d4U,
		0x0ffbU,0x5055U,0x0642U,0xb03dU,0x98f8U,0x9914U,0x3689U,0xbea6U,0x730cU,0x1b89U,
		0x4d5aU,0xa8ffU,0xac5cU,0x3085U,0x9d7cU,0x9fceU,0x419eU,0x8e5aU,0xbf63U,0x0b8dU,
		0xdb9aU,0x9a28U,0xa984U,0x1704U,0x8109U,0xbec4U,0x0056U,0x7188U,0xe594U,0x4a9eU,
		0x7b1aU,0xd4deU,0x0853U,0xc5caU,0x8d54U,0x6432U,0xa748U,0xe23fU,0xfa8fU,0xa4d5U,
		0x8a67U,0xc0c9U,0x77abU,0x7c24U,0xff37U,0xe608U,0x6e29U,0x8c11U,0x74e0U,0xd570U,
		0xd9b0U,0xa293U,0xc30eU,0x9a51U,0x2bc8U,0xc5dcU,0x68a7U,0x89eaU,0xa77bU,0xdf76U,
		0x8100U,0xca3dU,0x14e0U,0xafdbU,0xb179U,0x16bdU,0xe360U,0x68ddU,0xa8d1U,0x9a8bU,
		0x02b5U,0x506aU,0xd74eU,0x9e84U,0x48a7U,0x3fe2U,0xf4e5U,0x949dU,0xcc36U,0x37c7U,
		0x86b3U,0x9865U,0x300fU,0x7d98U,0x29b8U,0xdc4aU,0xfe34U,0xb622U,0x6e70U,0xffbfU,
		0x3324U,0x9c79U,0x4e50U,0x7b66U,0xc35dU,0x0285U,0x7c9dU,0xce9fU,0x4934U,0x77c4U,
		0xae20U,0x9715U,0x8c48U,0x580aU,0x0680U,0xf1dcU,0x2ba6U,0x3508U,0x2984U,0xe562U,
		0xcb8eU,0x5994U,0xcd1cU,0xd805U,0x7233U,0x2994U,0x6884U,0x4bd6U,0x8f7aU,0xd377U,
		0xb1aeU,0x03ddU,0xe952U,0x7081U,0xa8caU,0x6422U,0x5714U,0x662bU,0xfcfeU,0x6682U,
		0x95a0U,0x7e4dU,0x2d2fU,0xd6fcU,0xc0a8U,0x80c3U,0x8a87U,0xb5dcU,0x431fU,0x5d96U,
		0xde08U,0x250bU,0xe753U,0xe5baU,0x8e0eU,0x3e7eU,0x5b33U,0x3647U,0xec23U,0x43c9U,
		0xa2baU,0x83bfU,0xc33eU,0xd450U,0x05abU,0xcb38U,0xe321U,0x8412U,0xbe52U,0xd12dU,
		0xf62cU,0x3af1U,0x19d5U,0x1fb0U,0x9c14U,0xee5eU,0x4fc1U,0x478eU,0x62a0U,0x35d2U,
		0x33bdU,0x1245U,0xa465U,0x7cf3U,0x438cU,0xb002U,0x8d2dU,0x9124U,0x47b1U,0x5470U,
		0x9a1fU,0x30c2U,0x7ce4U,0xaad8U,0xc936U,0x9ad1U,0xe6cbU,0x0994U,0xf507U,0x5443U,
		0x2e46U,0x11efU,0x906eU,0xb819U,0x6b69U,0xd13dU,0x6632U,0xd0b2U,0x516dU,0x234eU,
		0x1870U,0x1518U,0xae78U,0x652dU,0x5e3bU,0x1cacU,0xb06dU,0x9ec0U,0xab57U,0x7477U,
		0xbfffU,0x706eU,0xff61U,0x0aaaU,0x84e1U,0xd994U,0x987dU,0x0f30U,0x74b2U,0x2067U,
		0x390dU,0xc67eU,0x165aU,0xc64cU,0x3017U,0x2ae7U,0x09ecU,0xdb06U,0x5181U,0x8701U,
		0x5eceU,0x5d73U,0x8203U,0xe4faU,0x0759U,0x0361U,0x57dcU,0x2be6U,0x96b6U,0xdd13U,
		0x733cU,0x1fc3U,0x3076U,0xb7b5U,0xf400U,0xf355U,0x1a0bU,0xb79fU,0x2f01U,0x073fU,
		0xf460U,0x17d3U,0x8957U,0xc281U,0x3c5bU,0xfbd7U,0xb4a5U,0x3851U,0xa32fU,0xf12bU,
		0x062aU,0x235dU,0x258eU,0x2b6dU,0xffa3U,0x4637U,0x0179U,0x7d3fU,0xbf8cU,0x4c03U,
		0x28aeU,0x56ccU,0x6413U,0xd927U,0x4736U,0x335bU,0x9a0eU,0xb34eU,0x5829U,0x5a67U,
		0x50d4U,0x3ec3U,0xe213U,0x943dU,0xcba0U,0x03c7U,0x0118U,0x5243U,0x63cbU,0x9a42U,
		0x908eU,0x4f09U,0x5deaU,0xdb89U,0x0ae9U,0x32e3U,0x7fd4U,0xad26U,0x0b79U,0xd5c2U,
		0x9d73U,0xc19dU,0xf476U,0x58a5U,0x00bfU,0xd03eU,0xc8dfU,0xeb1bU,0x7e1cU,0xe3ecU,
		0x82d0U,0xa074U,0x320cU,0xc89dU,0xc4edU,0xf52eU,0xf645U,0xaeecU,0xb3daU,0x6bfbU,
		0xa39dU,0xb67eU,0x604fU,0x36d8U,0x571bU,0xf8f5U,0x3b35U,0x55e6U,0x4104U,0xdd8eU,
		0xd673U,0x9edfU,0xa3e6U,0xb246U,0x7901U,0x3f7dU,0x03afU,0xf1faU,0x0cb6U,0xe763U,
		0x67adU,0xe8e5U,0xca82U,0x2240U,0xb9a0U,0x9bf8U,0x54bdU,0x6d37U,0x7641U,0xff70U,
		0x81dbU,0x37e2U,0x983dU,0x2e58U,0x205fU,0xb9f2U,0xcb63U,0x429aU,0x976fU,0x3724U,
		0x555aU,0x36f3U,0x5953U,0x12d9U,0xdfd9U,0xda7eU,0x2151U,0x1463U,0x54f7U,0x2c57U,
		0x044bU,0xc873U,0x4ff3U,0xbc30U,0x4028U,0x57b6U,0x9658U,0x380eU,0x9fc9U,0xe63bU,
		0x37c2U,0x8c0cU,0xe5b5U,0x598fU,0x9ebbU,0xaffdU,0x56c7U,0xdd0eU,0x3f78U,0x58a0U,
		0xe725U,0x145eU,0xaaafU,0xa5bfU,0xa9e9U,0x139aU,0xb127U,0x1100U,0x1730U,0xe72aU,
		0x2cacU,0x60d3U,0x3d72U,0x9dd1U,0x5c84U,0x8710U,0x9e4aU,0x94e5U,0x2a7eU,0x474aU,
		0xc6e3U,0x6715U,0x0417U,0x84a9U,0xcc22U,0xdc24U,0x0228U,0x5154U,0x7d0eU,0x0d2bU,
		0x78aeU,0x2d65U,0x5978U,0x38b1U,0x0701U,0xf912U,0xc1daU,0x2748U,0xd05cU,0xfc4dU,
		0xded8U,0xa5c4U,0x6effU,0xf69bU,0xbe39U,0x61b6U,0xb91eU,0xe8ceU,0xd5e1U,0xad54U,
		0x9cd0U,0x4f4dU,0x9e58U,0xfbafU,0xbd01U,0x6df0U,0x1a7fU,0x115bU,0xe34cU,0xfa64U,
		0xf496U,0xf7bcU,0x6689U,0x8c20U,0x90f1U,0x0942U,0xadabU,0x5ba6U,0x25b2U,0xb2a6U,
		0x76a6U,0x6c12U,0x6de6U,0xaab4U,0xef56U,0xf848U,0x5f4eU,0xd949U,0x14cfU,0x8193U,
		0xd185U,0x7c09U,0x88c4U,0xa580U,0x28c6U,0x42a2U,0x8424U,0xdaa8U,0xb4fdU,0xbfb3U,
		0xdc52U,0x1823U,0x6d4cU,0x5b92U,0xfe6fU,0xcc11U,0x38bfU,0x9db3U,0x1f1dU,0x3d53U,
		0xccf3U,0x46c7U,0xbc6bU,0x14efU,0x7f0eU,0x139fU,0xfad6U,0x8066U,0x5001U,0x61ecU,
		0x47c2U,0xb77aU,0x6244U,0xc052U,0xa84bU,0x70bdU,0x647bU,0xbdffU,0xe873U,0x6958U,
		0xd16bU,0x9b17U,0xe0bbU,0x3b4eU,0x938dU,0xad0eU,0x1610U,0xccdeU,0x62caU,0x903aU,
		0x44fdU,0xb37cU,0x90e7U,0x0896U,0x9023U,0xdd5cU,0x58aeU,0xa9eeU,0xb361U,0x4527U,
		0x1cfeU,0x3e27U,0xe9c0U,0xaae1U,0xbc06U,0x656fU,0xafe4U,0xc2c0U,0xaf6cU,0x8050U,
		0x7feaU,0x9c4cU,0x7861U,0x9fe6U,0x5790U,0x8bb9U,0x561eU,0x9935U,0x861dU,0x34a3U,
		0x2c8cU,0xad4bU,0x23b9U,0xbacaU,0xb853U,0x784eU,0xbc1cU,0x3b77U,0xec0dU,0x1105U,
		0xd9daU,0x5ad4U,0x3536U,0x6738U,0x4ad5U,0x1d5eU,0xba00U,0xaa03U,0x4d71U,0x7413U,
		0x4580U,0x8478U,0x858bU,0x7534U,0x06dfU,0x1bd0U,0x3e01U,0xff0aU,0x421eU,0xf41dU,
		0xfd05U,0x1b2bU,0x1174U,0x3275U,0xde75U,0xb9dcU,0x0cc4U,0xe5e1U,0x851fU,0xc1b0U,
		0x1c70U,0xb12cU,0x7446U,0xcbd6U,0x29deU,0x78d7U,0xbf2bU,0x94acU,0x273eU,0xfe1cU,
		0x0c0eU,0x7c82U,0x1c8cU,0xcae8U,0xc0c2U,0xe4afU,0xad1fU,0x35cbU,0xbac3U,0xf90bU,
		0x7aa6U,0x3e3aU,0x0102U,0x7b94U,0x8d8eU,0x99e7U,0x7b43U,0x0a1fU,0x3588U,0x4439U,
		0xdc6eU,0x4bc3U,0x754cU,0xe72fU,0xcfefU,0x6826U,0x33b5U,0x9360U,0x6059U,0xc1d2U,
		0xd606U,0xcfc8U,0xdda2U,0x3c65U,0xba62U,0x5083U,0xbdc3U,0xe083U,0xb68cU,0x3096U,
		0x686aU,0xfd73U,0xe6d5U,0x6cb2U,0x8688U,0xb941U,0x89b9U,0x3a60U,0xd9baU,0xfff8U,
		0xd3ccU,0x0dcfU,0x82f8U,0x1380U,0xa351U,0x1735U,0x216fU,0x5ab8U,0x6ef5U,0x8f84U,
		0xc066U,0x1262U,0x80a5U,0xc488U,0x628cU,0x242aU,0x8b32U,0xd971U,0x49d9U,0x4e5fU,
		0x9372U,0xd053U,0x81c9U,0x1cbeU,0x00cdU,0xf5adU,0xfb5fU,0x773dU,0xc3d8U,0xa002U,
		0x0607U,0x3e41U,0x6061U,0xf257U,0xef3aU,0x5ceeU,0xc28fU,0x60d8U,0x3146U,0x1215U,
		0xa4ecU,0xa72fU,0x6cddU,0x7ffcU,0x417cU,0x09c0U,0x2dfdU,0x557dU,0x058aU,0x83f7U,
		0x0a3cU,0xfbc4U,0xbce8U,0x66caU,0x0774U,0xc140U,0x6c3aU,0x5b80U,0x7e54U,0xc505U,
		0xe889U,0x3bcaU,0xd30bU,0x50aaU,0xcc77U,0x3f93U,0xc446U,0xe961U,0x701cU,0x2cb1U,
		0x4ef7U,0xbeedU,0x57eaU,0xaa6dU,0xd14eU,0x13d5U,0x202cU,0xbd99U,0xb4c3U,0xcdccU,
		0xb31eU,0xcb6bU,0x8182U,0xe80eU,0x1819U,0x6f84U,0x29c0U,0x5fb7U,0x4475U,0x630cU,
		0x0201U,0x947bU,0x7584U,0x1af4U,0xb1f3U,0x5b06U,0x0d54U,0x8126U,0x9bc8U,0xeaaeU,
		0xa533U,0x9dd6U,0xa523U,0x3f15U,0x1fd7U,0xfb93U,0x15e9U,0xe972U,0x7105U,0xb8acU,
		0x07c8U,0xde4aU,0xb33dU,0x2827U,0x1140U,0x8aa2U,0xa43aU,0x445bU,0xdf8eU,0x39a8U,
		0xdfc1U,0x5d51U,0x794aU,0x91d5U,0x5251U,0x632dU,0xb45bU,0x8ca7U,0xba1dU,0xad8eU,
		0x1e19U,0xe2a4U,0x5ffbU,0x3d77U,0xc6c1U,0xd1baU,0x6dd6U,0xe3a9U,0x62f3U,0x2d7aU,
		0xfaaaU,0xfa5bU,0x71ccU,0x17baU,0x28bcU,0xce4bU,0xe7d0U,0xb0d2U,0xb140U,0x7ef0U,
		0x056aU,0xa0d4U,0x69b8U,0x06cfU,0x51a3U,0x3517U,0x4ccdU,0x142eU,0x130cU,0x7aadU,
		0xde9bU,0x0902U,0xa1d8U,0x6777U,0x228bU,0x8715U,0xed07U,0xf8aaU,0xee0eU,0x291eU,
		0x71b5U,0x8f5fU,0x9cedU,0xf9c5U,0xa4adU,0x6edeU,0x91f6U,0x9f9eU,0x7836U,0xc34dU,
		0x476fU,0x8a55U,0xf00fU,0xfc0bU,0x62aeU,0x5db0U,0x95a7U,0x9f46U,0xb3b2U,0x29fdU,
		0x1478U,0xf789U,0x52dcU,0x2318U,0x79d1U,0xcd94U,0x289eU,0x5c89U,0xec9eU,0xc801U,
		0xd2b0U,0xd0e7U,0xb473U,0xd6cfU,0x362fU,0xa2d7U,0x5bfaU,0xaafaU,0x7a89U,0x007bU,
		0x0b14U,0x07efU,0x9adfU,0xe5a5U,0xb225U,0xa6b2U,0xe17aU,0x5097U,0xfa08U,0x4678U,
		0x7bffU,0xc8f6U,0x79e3U,0xa66eU,0xae86U,0xdb74U,0x24b4U,0x8bd0U,0x0c6bU,0x94eaU,
		0x3fb9U,0x71fdU,0x14b4U,0x9793U,0x815fU,0xb5ebU,0x4711U,0x304dU,0x8782U,0x0e11U,
		0xade4U,0x807cU,0x60afU,0xdd40U,0x5012U,0x1f7dU,0xf85dU,0x9ea1U,0x911cU,0xecc6U,
		0x120dU,0xf693U,0x63f2U,0x0b34U,0xfca8U,0x8a0bU,0xe790U,0x9608U,0xd113U,0x7794U,
		0x0036U,0x9e4fU,0x0f7bU,0xb394U,0x99bdU,0x2c20U,0x283aU,0x45f0U,0x2075U,0xc594U,
		0x0ee8U,0x8281U,0x9728U,0x1570U,0xd874U,0xb700U,0x39b8U,0x4ba2U,0x9057U,0xb98bU,
		0x769cU,0xc177U,0x2524U,0x603cU,0x5a1dU,0x271bU,0xd2d3U,0x7b9bU,0xcfd0U,0xa245U,
		0x937bU,0xe699U,0xb770U,0x474fU,0xa9d0U,0x0b3cU,0xaed9U,0xc62bU,0x822aU,0x2379U,
		0x7e07U,0x2c95U,0x1d43U,0x51e7U,0x003fU,0x7ed6U,0x4ea2U,0xedf5U,0xd57eU,0xe651U,
		0x5a88U,0xfab3U,0x284dU,0xede2U,0xd703U,0xe489U,0xd35aU,0x3767U,0x2ea1U,0xf5c2U,
		0x41fdU,0x10d5U,0x3d28U,0x7924U,0x2ac0U,0x2a83U,0x048aU,0xa08cU,0xadc1U,0xa994U,
		0xe43eU,0x48e0U,0xe0a1U,0x8443U,0x499cU,0xa034U,0x65b7U,0x372cU,0xe4e5U,0x2d05U,
		0xe6e7U,0x6b57U,0xb62aU,0x391fU,0xb187U,0x0540U,0x18b5U,0x639cU,0x2fddU,0x6973U,
		0x7e38U,0x7996U,0x7f04U,0xcd4fU,0x9af6U,0xb561U,0x27d3U,0x671aU,0x2aa0U,0xd1d5U,
		0xe7a9U,0x2e17U,0x2019U,0x2ddeU,0x7da6U,0x3a33U,0xddc8U,0xce82U,0x93e3U,0xaf90U,
		0xb756U,0xacdeU,0x1b04U,0x16e9U,0x72a8U,0x5458U,0xd50dU,0xf078U,0xac8eU,0x53caU,
		0x10ccU,0xb29dU,0xe526U,0xa7a0U,0x4942U,0xad76U,0x7a45U,0x5cbaU,0xf47fU,0xc6e8U,
		0x2fc0U,0x0fb4U,0xe55cU,0xe602U,0x2be0U,0x7cedU,0xda78U,0xc197U,0x14e9U,0xfc47U,
		0x4bbdU,0x3b71U,0xeaa8U,0x5ce8U,0x8bcaU,0x6771U,0xb55bU,0xe693U,0xf600U,0x12f5U,
		0xbb6cU,0x02e9U,0x16d1U,0x0ea9U,0xac7aU,0x0109U,0x8f2fU,0xee9aU,0x2604U,0x12fbU,
		0x690cU,0x84aeU,0xb5caU,0xedb4U,0x2d3dU,0xb9b7U,0xab68U,0x0ba6U,0xad04U,0xf26bU,
		0x84d2U,0xde2eU,0xce15U,0x1921U,0xd9a9U,0xaf32U,0xc8d8U,0x47aaU,0x352fU,0xbe32U,
		0xee07U,0xa4e5U,0x2d36U,0xaed2U,0x342eU,0x8efdU,0x491aU,0x72ebU,0x6fa4U,0x2142U,
		0x95f5U,0xe917U,0x4be4U,0x041cU,0x5903U,0x6ad4U,0x18dfU,0x7252U,0x6e8cU,0x47ffU,
		0xfe91U,0xe842U,0xa627U,0xd77dU,0x3b26U,0x43ddU,0x79adU,0x7ad1U,0x2e90U,0xb6cbU,
		0xd7e0U,0x95d8U,0x5856U,0xf777U,0x5bd2U,0xf3deU,0xa874U,0x892bU,0x0de8U,0x836fU,
		0x423cU,0x22c0U,0x8882U,0xf606U,0x3c27U,0xdc29U,0x184bU,0x5b46U,0xa841U,0x5d31U,
		0xf397U,0x3aa6U,0xa21cU,0x1ac4U,0xc52aU,0xa3b7U,0x4fcfU,0x48fbU,0xc38aU,0x9145U,
		0x0481U,0xef4dU,0x8fbeU,0x2809U,0x403eU,0x56f2U,0x4a75U,0x8635U,0x6dbaU,0x5743U,
		0xcefaU,0x1ba5U,0xb492U,0x066fU,0xed41U,0xd201U,0x203cU,0xcc27U,0x43f8U,0x96f6U,
		0xf4ffU,0x2624U,0x0a34U,0xbcf9U,0x05e6U,0x4e9aU,0x025bU,0x46deU,0x40f3U,0x8fabU,
		0xde1aU,0x1199U,0x4fb2U,0x26c8U,0x1c3aU,0x8345U,0x31a1U,0xa4feU,0xabddU,0x9aabU,
		0x83cdU,0xe410U,0x0801U,0x99eeU,0xd60eU,0x413aU,0x1076U,0x02efU,0x4f21U,0x5159U,
		0xbbf7U,0x63faU,0x8f04U,0xd717U,0xe474U,0x8e69U,0xf140U,0xb04fU,0x8324U,0xe8fcU,
		0x087cU,0x9281U,0x05eeU,0x2c34U,0xb372U,0x3a03U,0x9d29U,0xb4eeU,0xd07aU,0xcab6U,
		0x80caU,0xb40cU,0x42dbU,0xb268U,0xa6fdU,0xbfdbU,0x11d2U,0xe129U,0xcd80U,0xc424U,
		0x9963U,0x022dU,0xc94fU,0x0fceU,0x89aaU,0x6119U,0xf592U,0x211aU,0xaca5U,0x15cfU,
		0xd48dU,0x1792U,0x2b93U,0x9777U,0x93afU,0xd3f9U,0xc515U,0x8ad9U,0xfe3cU,0x2d9fU,
		0x9627U,0xebfeU,0xf6bcU,0xfed1U,0x0390U,0x3dd9U,0x042dU,0x566bU,0xe13cU,0x8d6cU,
		0x2f97U,0xbb72U,0x32feU,0x0d30U,0x1bfeU,0xab94U,0xe151U,0x7cceU,0x416dU,0xbdd6U,
		0x2cfcU,0xda9fU,0xc6adU,0xdbfbU,0x1cf4U,0x09ffU,0xf0d9U,0x0862U,0x12c3U,0x9b40U,
		0x19f3U,0x3cb8U,0x9a7bU,0x44a2U,0x8ccbU,0xa42bU,0xe1bbU,0x4f67U,0x4088U,0x18f1U,
		0x7004U,0xbd61U,0x13efU,0xf028U,0x5f12U,0x7d13U,0x6382U,0xf345U,0x4d1dU,0x62e1U,
		0x503fU,0x461aU,0x7bf0U,0x6a4cU,0x802eU,0x80eaU,0x1da0U,0xe627U,0x8ba6U,0xd244U,
		0xe4c4U,0x4e1fU,0x2d55U,0xc70cU,0x1b5fU,0xebe0U,0x0eccU,0xa5daU,0xf105U,0x0027U,
		0xfe82U,0x8d95U,0x9f00U,0x7f85U,0xa56aU,0x0eafU,0x6cedU,0x2d6aU,0x4344U,0xdca0U,
		0x3435U,0xfeb9U,0xb02cU,0x60e9U,0xe7f7U,0x3413U,0xcad3U,0x4fa3U,0xf807U,0xfe05U,
		0xb8daU,0xc7afU,0xf683U,0x7c55U,0x8906U,0xfb49U,0xfc2eU,0xcf50U,0xc0afU,0xdaf5U,
		0x9fdcU,0xb8feU,0xd8daU,0xca68U,0xd34bU,0x5fe5U,0x1012U,0xe8beU,0xbcc2U,0x78b3U,
		0x168eU,0xcf39U,0x2c77U,0x87cbU,0x1b71U,0x465cU,0x45c5U,0x4b19U,0xf4acU,0x3605U,
		0x7b28U,0x27c4U,0x7b84U,0x1095U,0xbb42U,0xf8c2U,0x8b1eU,0x2195U,0x45a4U,0xaa18U,
		0x0d0eU,0x6aa6U,0xd0d1U,0xaf0fU,0xfb19U,0x34e4U,0xeb2dU,0x464cU,0xf963U,0x16d7U,
		0x2fd4U,0x38b6U,0x5355U,0xfc81U,0x8ee7U,0x3400U,0x9d9fU,0xdd6eU,0xf972U,0xb72bU,
		0x0ceeU,0xeb9cU,0xb3fcU,0x7389U,0x1bbeU,0xf674U,0x885cU,0x93cdU,0x2a11U,0xf766U,
		0x8613U,0x5366U,0xbd40U,0x0e42U,0xd0feU,0xd83dU,0xe1a2U,0x3a96U,0x226fU,0xbbaaU,
		0xbed8U,0x0651U,0xf0f6U,0x597dU,0x129eU,0x1d21U,0x5b6fU,0x1ed5U,0x2af8U,0x533fU,
		0x20f4U,0xff23U,0x02d9U,0x6590U,0xd9f2U,0xe3cfU,0x59d5U,0x08b6U,0x620aU,0x9954U,
		0xb58cU,0x4549U,0x2346U,0x9924U,0x3bc2U,0x093aU,0xb260U,0x375fU,0xc1c6U,0xbad1U,
		0x3c6dU,0xcbe0U,0x1cc7U,0x7ba1U,0xe5ecU,0x35d7U,0x126cU,0xa676U,0x72f2U,0x8296U,
		0x1c7cU,0xbad5U,0xdcefU,0xc2c9U,0x1740U,0x7540U,0xb39dU,0xbf38U,0x46e6U,0x5bb0U,
		0xad6bU,0xfd3cU,0xd544U,0x8cb0U,0xbc28U,0x4bceU,0xb5e3U,0xc1caU,0xdd66U,0x3a9eU,
		0xd3d8U,0xf6a6U,0x6ca6U,0xce8bU,0x6046U,0x1c9bU,0xfc25U,0x62a5U,0x2812U,0x3282U,
		0xd954U,0xd97dU,0x7c35U,0xcbe4U,0xadd4U,0xa6bcU,0xcf94U,0xc8bfU,0xb0dcU,0xdd96U,
		0x2fb5U,0x5674U,0x7b4fU,0xb785U,0xffc7U,0xcdd6U,0x93a5U,0x3903U,0x657aU,0x3c71U,
		0xfc8aU,0xba0dU,0x1eb3U,0x6bcbU,0xb92aU,0xdd4bU,0xffecU,0xd2d9U,0xb56bU,0x124aU,
		0x4badU,0x8c2cU,0x90a2U,0x217cU,0x484fU,0x7ba5U,0x3977U,0x5e12U,0x3ab0U,0x8399U,
		0x5cddU,0x2390U,0xd191U,0x6ee9U,0xf6dbU,0xe580U,0xfd9bU,0x422cU,0xf74eU,0xedbeU,
		0x6811U,0x1ccbU,0x0086U,0x9dbfU,0x9b81U,0xe906U,0xd4cbU,0xaa94U,0x1439U,0x3b5aU,
		0xd047U,0x33c2U,0xbd75U,0x4366U,0x086cU,0x864fU,0x78c5U,0x35dbU,0xfae2U,0x91e0U,
		0x16c6U,0x6cbfU,0x6317U,0x9941U,0xcfb4U,0x6746U,0x859fU,0xfb53U,0x1d0aU,0xc14bU,
		0x5124U,0xd33bU,0x0e4cU,0xe5f0U,0x3dcfU,0x5d69U,0xb41dU,0x23c0U,0x4a12U,0x6bb5U,
		0xb839U,0xa24bU,0x878eU,0x7cf8U,0x7528U,0x0c9eU,0x4921U,0x86cbU,0x3a3eU,0xa67aU,
		0xd478U,0xdee5U,0x9734U,0xa097U,0x0821U,0xa89dU,0x3299U,0x4d80U,0xdcb9U,0x75deU,
		0xcb1cU,0x1168U,0xc636U,0x9614U,0x8a8eU,0x1270U,0x3a28U,0xf045U,0xdc34U,0x6783U,
		0xa05bU,0x76b3U,0xaa3aU,0x3a17U,0x9321U,0xa46aU,0x85e9U,0x5528U,0x7842U,0xd20cU,
		0xc894U,0x829aU,0x4f9aU,0x6ed5U,0x72c0U,0xfb6dU,0x11c8U,0xae6eU,0x0925U,0xdab5U,
		0xeea5U,0x5c95U,0x7e9eU,0xd6b5U,0x5f9cU,0xd9ceU,0x9aebU,0x72f6U,0x65f0U,0x9e36U,
		0xc6d3U,0xbe0cU,0xcac1U,0xe3b5U,0x73b4U,0xcfd6U,0x97a8U,0xb007U,0x69c4U,0x2daaU,
		0xefa2U,0x6519U,0xf8ffU,0xbad9U,0xf212U,0x6e3bU,0x5e97U,0x0749U,0xc840U,0x07b6U,
		0x4f72U,0x3f21U,0xadf5U,0xcd00U,0xd735U,0xece5U,0x544dU,0xcb08U,0xeec9U,0x1c80U,
		0xdf9aU,0xa5e5U,0x9a5bU,0x8dd8U,0xa76bU,0x424fU,0x7590U,0x3610U,0xb5a3U,0x4391U,
		0xeffbU,0x89c7U,0xea8cU,0xd80fU,0xe821U,0xc2cdU,0xfe10U,0xbda5U,0xb7cbU,0x921bU,
		0x7e60U,0x08c1U,0xeae3U,0x3cd0U,0x0fc4U,0x746dU,0x93d8U,0x8008U,0x18caU,0x450aU,
		0xd695U,0xdcf3U,0x4a5bU,0x8513U,0xc7cdU,0x7855U,0xddaaU,0x9f34U,0x6178U,0xe69fU,
		0x0535U,0x9129U,0x354fU,0x79ebU,0x2149U,0xcb86U,0x0c63U,0x7544U,0x2fccU,0x3c1fU,
		0x3497U,0x97a0U,0xa419U,0xf084U,0x0263U,0x2756U,0xaa50U,0x0bd3U,0xe7b5U,0x8ff6U,
		0x42fcU,0x7dceU,0x5d9dU,0x1744U,0xc0e9U,0xe1aaU,0xa7cdU,0x6f13U,0xe44aU,0x44f1U,
		0x4991U,0x898fU,0xd51dU,0x8d32U,0xea3cU,0x63a8U,0x9d1fU,0x9064U,0xc3c7U,0xbf3cU,
		0x5c1dU,0xd25aU,0xd6eaU,0xd59bU,0x3e1bU,0x410cU,0x7e13U,0xf6ffU,0xb023U,0x71acU,
		0x6e78U,0x9e66U,0x8e3eU,0x1f8fU,0xf2f6U,0xb3a1U,0x0e64U,0x3773U,0x4248U,0xaa8dU,
		0x48d5U,0x2643U,0xcf14U,0x9381U,0x742eU,0x5475U,0x05caU,0xf9d2U,0x975eU,0x4907U,
		0xa78cU,0x5bb4U,0x21c2U,0x50f0U,0xa2efU,0x1965U,0xed66U,0x40b7U,0x70b1U,0x8ea6U,
		0xcf06U,0xb869U,0xc730U,0x91b8U,0x185dU,0xc097U,0xc90bU,0x46eaU,0xfdb4U,0xb3bfU,
		0x2ab5U,0x88d2U,0xa188U,0x264aU,0x44d1U,0xd6dcU,0xdffbU,0x47b6U,0x5483U,0x4c69U,
		0x3ed9U,0xa940U,0x6918U,0xfd40U,0x9277U,0x098aU,0xdfe9U,0x7addU,0x0a88U,0xfa15U,
		0x5d8fU,0xa8f8U,0x1393U,0x7d38U,0x0dc8U,0x088fU,0x9601U,0xad87U,0x0668U,0xad6fU,
		0xbba3U,0xbd5aU,0x6f21U,0xb85aU,0x2937U,0x3d05U,0x42e7U,0x4755U,0xb891U,0x30c7U,
		0x94cdU,0xd179U,0x6fabU,0x7e90U,0x9e6fU,0x8cb4U,0xe646U,0xb05bU,0x3c10U,0x678cU,
		0xf6c8U,0xff7bU,0xefdcU,0xc9c2U,0x2e27U,0x9d3bU,0x6c06U,0x2126U,0x328bU,0x71d9U,
		0x4326U,0xd548U,0x90bcU,0x6ea4U,0x1e8aU,0x8133U,0xf136U,0x7cdaU,0x9910U,0x13adU,
		0xe23bU,0x9a24U,0x9a87U,0x9a4dU,0x9711U,0xdc46U,0xd6f8U,0x4bd2U,0xee5aU,0x43c5U,
		0xd0aeU,0x30beU,0x86fdU,0x0aa6U,0x2359U,0xf351U,0xdb85U,0xb34aU,0xf8f1U,0xe3e8U,
		0x4296U,0xe8e1U,0x589cU,0xbc2cU,0x2d61U,0x870cU,0x8c1cU,0xe8caU,0xcc0dU,0x818fU,
		0xccdaU,0xb776U,0x349fU,0xaaddU,0x1b27U,0x1d5aU,0x7b90U,0x94a8U,0xb93dU,0xc1ceU,
		0x7739U,0x125eU,0xc501U,0x7ff8U,0x9477U,0x13d1U,0x91d1U,0xe96eU,0x3513U,0x2d76U,
		0x9f42U,0x8f5bU,0x4674U,0xd6cbU,0xf68fU,0xb5e7U,0x6038U,0x45ecU,0xe485U,0x2375U,
		0x6398U,0x48dcU,0x53c6U,0x2ddaU,0x02e5U,0x7ce9U,0xaeceU,0x0ba2U,0xf3daU,0x47fbU,
		0x8631U,0x3aa2U,0x9aa7U,0xbcf5U,0xb408U,0x8e65U,0xfecdU,0x2116U,0xa427U,0xbdd2U,
		0xa5d6U,0x4616U,0xdaf1U,0x60e5U,0x6aa2U,0x4658U,0x0e3eU,0xdd6aU,0x0936U,0x533bU,
		0xa0e6U,0x3b7eU,0x508bU,0x3995U,0x81d2U,0x0b42U,0xbb39U,0x9ad6U,0xd865U,0xdb7cU,
		0xe91eU,0x8457U,0x85f7U,0xf6aaU,0xcf6aU,0xafedU,0xceb4U,0x52b0U,0x0f84U,0x5beeU,
		0x8fb3U,0xf871U,0x0ef9U,0x494fU,0x334eU,0xb464U,0x5035U,0x9a04U,0x2512U,0xd3dcU,
		0xd917U,0xbbb2U,0xd6c3U,0xcb00U,0x5eadU,0xddd1U,0x54a0U,0x9f6cU,0x04c5U,0xc93bU,
		0x8fc7U,0x471fU,0x9bb9U,0x0732U,0x4832U,0xce8fU,0x692dU,0xae0eU,0xe99dU,0x2892U,
		0xaeb4U,0x9ee7U,0xc327U,0x0436U,0x48f0U,0x7d71U,0x9c81U,0xc9d2U,0xdd06U,0x7e45U,
		0xecbeU,0x6caaU,0x535eU,0x56eaU,0x3deaU,0x4ddfU,0xbb85U,0x3a7aU,0x66f4U,0xa9d6U,
		0x8a49U,0x0999U,0xd854U,0x47e6U,0x98adU,0x5c51U,0x857cU,0x1c9fU,0x176eU,0x2acdU,
		0xa211U,0x2f59U,0x8d4fU,0x063dU,0x492fU,0xe35bU,0x19d0U,0x959bU,0x09e7U,0x6b64U,
		0x9089U,0xa32aU,0x205aU,0x604aU,0x7d09U,0x56c2U,0x6d47U,0xf491U,0x5619U,0x9388U,
		0x7aa1U,0x4219U,0x00c8U,0xe6d0U,0x4470U,0x6c35U,0x69b3U,0xdfbcU,0xe175U,0x62a9U,
		0x7697U,0x9117U,0xb182U,0x2848U,0xf5fbU,0xd508U,0x5851U,0xee02U,0x319cU,0x4039U,
		0x9622U,0xd075U,0x1b5aU,0x9a76U,0x459fU,0xfc29U,0x2341U,0x860eU,0x93a0U,0xbc23U,
		0x511fU,0xf749U,0x5f97U,0xc631U,0x18c5U,0x5448U,0x8e39U,0x42f7U,0x95fcU,0x1858U,
		0x4291U,0x3286U,0x6a9dU,0x466fU,0xdd01U,0x5030U,0x1b55U,0x9084U,0x256cU,0x97d0U,
		0x96e6U,0x7623U,0x7608U,0xee1cU,0x97d5U,0x0f6fU,0xde0fU,0x2816U,0x7654U,0xdb28U,
		0x68c4U,0x2493U,0xa55bU,0xa5b2U,0x0a69U,0x2d8fU,0xc2efU,0xf50cU,0xb6f4U,0xba96U,
		0x76f7U,0xd795U,0xbb54U,0xd981U,0x8f9aU,0xc91aU,0x6af2U,0xa50bU,0xdc17U,0xbd31U,
		0xdc90U,0x5dbeU,0xc556U,0x23e7U,0xef6dU,0x5a11U,0x2571U,0x0e95U,0xe82fU,0xd958U,
		0xe405U,0xfdf4U,0x7f4aU,0xecacU,0xdd25U,0xfefaU,0xb747U,0xaedfU,0xf08cU,0x11f4U,
		0xa5f3U,0x940bU,0x6086U,0xed94U,0x4afcU,0xcbe8U,0x3041U,0x3b12U,0xd683U,0xb655U,
		0xf838U,0xab75U,0xd5b5U,0x821bU,0x80e0U,0x3eb9U,0xb487U,0xe9eeU,0x7628U,0xddd9U,
		0x9303U,0x7c39U,0xda3eU,0x24a8U,0x7052U,0xb97bU,0x3375U,0xcf1bU,0xf220U,0x6d5bU,
		0xed32U,0x2e4bU,0x9561U,0x8938U,0x942cU,0x7823U,0x9e95U,0xa6c0U,0xf4c4U,0xbaefU,
		0x3f05U,0x1e39U,0x2962U,0x5d07U,0xc116U,0xab3aU,0x2233U,0x5bc3U,0xdbb9U,0xe0e5U,
		0x96ebU,0x9d4fU,0x49bfU,0xadd8U,0xeb6cU,0xfe9eU,0xc8adU,0x110cU,0xba6aU,0x05fbU,
		0xac6fU,0x237fU,0x244fU,0xb81eU,0x0155U,0xae8eU,0x0423U,0x81f9U,0x17f7U,0xc8c3U,
		0x13b7U,0xd4bbU,0xefbfU,0x1e27U,0xa723U,0x4c73U,0x40fbU,0x02fcU,0xd8d1U,0x146fU,
		0xbf46U,0x1e81U,0xee21U,0xbe6bU,0xe5c1U,0xcf98U,0xc830U,0x2277U,0xacb6U,0xbcdcU,
		0xdabfU,0x28e1U,0x01a4U,0x4d98U,0xb9acU,0x9073U,0xec66U,0x9225U,0x9650U,0xb1a6U,
		0x877aU,0xdd9aU,0x0ce6U,0xc522U,0xc4f9U,0x5e8fU,0xbbdeU,0xd67bU,0x3c7bU,0x161fU,
		0x1a91U,0xf063U,0x332cU,0x023aU,0x760dU,0xd5ceU,0x86edU,0xb0e0U,0xdccaU,0xc155U,
		0xc8cdU,0xa650U,0x9bdcU,0x34cfU,0x259aU,0x8230U,0x6351U,0xd142U,0x2a6aU,0xa541U,
		0xc205U,0xa54bU,0x9979U,0x5678U,0x7cc4U,0xb7abU,0x3473U,0x6017U,0x5df6U,0x9261U,
		0x8e07U,0x4197U,0x2ca5U,0x640cU,0x2168U,0xb35aU,0x0f74U,0xc6baU,0x43f1U,0x2fb9U,
		0x1297U,0x637bU,0xeff4U,0xbd6eU,0xaec7U,0x547cU,0x955aU,0x4469U,0xa268U,0x6b6eU,
		0x6b75U,0xeacdU,0xbf09U,0xfb7fU,0xd3e6U,0xb789U,0xa4c8U,0xc4c7U,0x0b63U,0xb65fU,
		0x67daU,0x33a3U,0xcdc0U,0x7483U,0x8a1cU,0x1dadU,0x67caU,0xe45fU,0x97daU,0x2b3aU,
		0x69ecU,0x7b53U,0x3023U,0x8e4dU,0x77b7U,0xec14U,0x12b9U,0x3c51U,0xb579U,0x2c9bU,
		0x6e0cU,0xd0b7U,0xbb04U,0x4b26U,0x4bebU,0xdf82U,0x6e81U,0xcddaU,0x0dabU,0xb4d7U,
		0x4e41U,0xdc3dU,0xee83U,0x7892U,0x74cfU,0x0840U,0xd75aU,0x79faU,0x99a7U,0x9e1bU,
		0x281bU,0x3175U,0xe75aU,0xffcbU,0x87c2U,0xd0deU,0xbf4fU,0xd319U,0xafceU,0x98e9U,
		0xb960U,0x60f6U,0x88e4U,0x6637U,0xe58eU,0x70f9U,0xe31bU,0x4e4aU,0xcb9aU,0x3907U,
		0x0411U,0x763bU,0x7fe4U,0x841eU,0x49d3U,0xc0bcU,0xe7caU,0x817cU,0xd86eU,0x0b0eU,
		0xeaa2U,0xe4dfU,0xde14U,0x79a7U,0x8ba0U,0x93a9U,0x59cfU,0x7b7eU,0xcfaeU,0x2fafU,
		0xeaddU,0x091fU,0x5d89U,0x7e0dU,0xa421U,0x2353U,0x584bU,0xc321U,0xc110U,0xdc8aU,
		0xcdbaU,0x3c75U,0xc39fU,0xe7c4U,0xbbe6U,0x1ee2U,0xeb44U,0x4426U,0x37b3U,0x8405U,
		0x2f81U,0xba30U,0xaf45U,0x5e6cU,0xdb2dU,0x5eb5U,0x1447U,0x657eU,0x1422U,0xfa46U,
		0xd8b3U,0xebc4U,0xdba6U,0x7435U,0x345cU,0x52c7U,0xd36aU,0x5172U,0xa26fU,0xe24bU,
		0xd772U,0x2a1eU,0x6da4U,0xba11U,0xa142U,0x7fa5U,0x6754U,0x7ec0U,0x3fa8U,0x1aa2U,
		0xc3a5U,0x4cfcU,0xbf7dU,0x6675U,0x8828U,0x606fU,0x7659U,0x0c7dU,0xd1aeU,0xfc8eU,
		0x22fdU,0xbeb2U,0x111aU,0x5adbU,0xe08bU,0x7450U,0xf5b6U,0x51edU,0x24e0U,0x151dU,
		0xc070U,0x8bd8U,0x6adbU,0x9107U,0xed1fU,0x6bcfU,0x4982U,0x8059U,0x98ceU,0xfa7dU,
		0x2c29U,0x8c6dU,0x11a1U,0xadb5U,0x22d0U,0x9639U,0x8ce8U,0xad97U,0x2498U,0x9805U,
		0x3e85U,0x1eb7U,0x902dU,0x0659U,0x87e2U,0x5f72U,0x7572U,0xaf99U,0x252eU,0xc6f6U,
		0x6b4bU,0x1875U,0x56fbU,0xd4abU,0xa7b9U,0x4a24U,0xefcdU,0xdd4fU,0x7765U,0xb22fU,
		0x42d0U,0x66b5U,0xe7daU,0xb24eU,0x1ee8U,0x8d75U,0x8726U,0x18a9U,0x7b6eU,0xb4cdU,
		0x68c9U,0x669cU,0x8043U,0xb92eU,0x3409U,0xa06aU,0x236dU,0x8dd0U,0xebbcU,0xb973U,
		0x32c6U,0x1d49U,0x9b00U,0x6532U,0x6a26U,0x5aecU,0xffddU,0x4f40U,0x23a3U,0xd2ddU,
		0x4dfaU,0x6ec5U,0xf3a1U,0xb764U,0x513bU,0x1042U,0x5517U,0x26ffU,0x412fU,0x2e69U,
		0x28efU,0x27a2U,0xa5b7U,0xb5d4U,0x8a03U,0xfff0U,0x3a8eU,0x1b9dU,0x1c0fU,0x832eU,
		0xedd6U,0xf9d9U,0x594aU,0xc818U,0xeb24U,0xae7dU,0xead4U,0xcc04U,0xf0e3U,0x1f67U,
		0x1354U,0x124eU,0x929eU,0x1df2U,0x68e6U,0xc02eU,0x6c5dU,0x1d37U,0xbbecU,0x24c3U,
		0x6c24U,0x8910U,0xd93cU,0xc885U,0xa560U,0x9d69U,0xbd4aU,0xb56fU,0xb7dcU,0x1f08U,
		0x4c2eU,0xd9e1U,0x6106U,0x370eU,0x7a6dU,0x7edcU,0x344dU,0x1cb1U,0x31deU,0xe04dU,
		0x590aU,0x87a3U,0xe687U,0x8c30U,0x998cU,0xb42cU,0xce04U,0x287aU,0x9538U,0x7d85U,
		0xac00U,0x10e3U,0xc47bU,0x8b91U,0x01d1U,0xa485U,0x2d94U,0x078dU,0x8e15U,0x4bb1U,
		0xe1fdU,0x64b3U,0x40a1U,0x70d7U,0xcdf1U,0xb30dU,0x8df8U,0xe076U,0x4fdfU,0x5e40U,
		0xea23U,0x3ee9U,0x2257U,0x278aU,0x6fe7U,0x2180U,0xc18bU,0x5878U,0x823eU,0x9e11U,
		0x646aU,0x3484U,0x442cU,0x17e8U,0xfb3fU,0xb674U,0x2b88U,0x65c5U,0x0a6eU,0xb712U,
		0xf2d8U,0x90a6U,0x59ebU,0x7a39U,0x8760U,0xcee1U,0x5f88U,0xb9cfU,0x8a2aU,0x0045U,
		0x275eU,0x9ec5U,0xacd2U,0x76e1U,0x2689U,0xf88eU,0x5725U,0x7ba9U,0xa91aU,0x64c8U,
		0x1627U,0x2f0bU,0x35beU,0x92f4U,0xe6baU,0x9cbdU,0x3225U,0xa0a6U,0xb7f3U,0x10c0U,
		0xf511U,0x9f74U,0x796cU,0x4853U,0x15c4U,0xb2e2U,0xc318U,0x1d51U,0x5767U,0x05d1U,
		0x071fU,0x6eb5U,0xed78U,0xb072U,0x6b7cU,0x25e9U,0xcd43U,0xcd26U,0x2c4eU,0x5e16U,
		0xfc78U,0x71f4U,0xf410U,0xe166U,0xf7fdU,0xc4e3U,0xeb4aU,0xee79U,0x49ecU,0x4004U,
		0x644aU,0x414dU,0xc2f4U,0x2a94U,0xcd75U,0x397bU,0x3f4fU,0xeb5aU,0x3cc9U,0x673fU,
		0xbdcbU,0xa8f1U,0xab33U,0xedfbU,0x4cf5U,0x747cU,0xee72U,0x24bcU,0x7259U,0x6997U,
		0x727bU,0x839dU,0x1ae7U,0xcd66U,0xc74cU,0x7e75U,0x0d97U,0x9242U,0xde22U,0x131bU,
		0x193cU,0xea50U,0xf720U,0x2951U,0xba9bU,0xa953U,0x364eU,0x3ab4U,0x8de9U,0xbee0U,
		0x32f3U,0xacf0U,0x660aU,0x51bbU,0xf988U,0x6f57U,0x4edbU,0xab5cU,0x4b54U,0x98dfU,
		0xed55U,0x0befU,0x83eaU,0x2394U,0x5fdbU,0x9ed5U,0x54d4U,0x4a83U,0xc611U,0xa3eeU,
		0x840bU,0x4a97U,0x84a2U,0x03c0U,0x4e58U,0x9c45U,0xb6f9U,0xb0cbU,0x1192U,0x5ce1U,
		0x08afU,0xd23dU,0x1fe9U,0x2e84U,0xf9f0U,0x0f1dU,0x2f3dU,0x4ea8U,0x4497U,0x7073U,
		0x693eU,0x55bbU,0x3d6aU,0x5b2bU,0x99b5U,0x6eedU,0x5b67U,0xf4f7U,0xf3d2U,0xea84U,
		0xd76aU,0x9424U,0xf0a6U,0xcd3bU,0x4780U,0x683bU,0x981bU,0x031fU,0xd79aU,0x17adU,
		0x1839U,0xd195U,0xa3d6U,0xb5b2U,0x4b34U,0x0461U,0xb1cfU,0x490eU,0xc6a2U,0xa285U,
		0x34bbU,0xc004U,0xfb86U,0x83dbU,0x1684U,0x03a5U,0xc459U,0xe584U,0x35adU,0x7673U,
		0x919bU,0x4ca1U,0x7246U,0x3660U,0x37b9U,0x699eU,0x1bb5U,0x60a6U,0x37aaU,0x91c8U,
		0x76fcU,0x41cdU,0x2974U,0xf6dfU,0x732aU,0xa116U,0x4273U,0x353dU,0xcbb2U,0xe863U,
		0x7144U,0xe657U,0xf9b8U,0x0aafU,0x7fb3U,0x8845U,0x18e6U,0x1237U,0x1f98U,0x4230U,
		0xe9daU,0xd093U,0x5839U,0x76bcU,0x82e8U,0xabb3U,0x2002U,0x24f6U,0xe271U,0xfbfeU,
		0x2f20U,0xa7efU,0xd986U,0xe6f8U,0x5b3aU,0xfd9fU,0x1a42U,0x3235U,0xbe74U,0x8172U,
		0x246eU,0x31ccU,0x230bU,0x52ebU,0x108bU,0xff66U,0x4d12U,0xf11dU,0x3e6dU,0x789bU,
		0x59a5U,0xedc2U,0x0553U,0xc37eU,0xa1c8U,0xdad1U,0xa127U,0xd3b5U,0xba36U,0x07e2U,
		0x8460U,0x3f30U,0x5235U,0xc990U,0xbb59U,0xd02dU,0xae4aU,0xf752U,0x5737U,0x08e4U,
		0xe543U,0x9c67U,0x1f70U,0xc969U,0xd7a9U,0xd584U,0x026bU,0xd999U,0x33d9U,0x05daU,
		0x5415U,0xc3feU,0xc7f0U,0x1ccfU,0x4e14U,0x6d29U,0x3c83U,0xb51cU,0x2a45U,0xa661U,
		0x41d6U,0x40d8U,0x3d18U,0xc5f4U,0xd1f5U,0x5864U,0xc91fU,0x54a8U,0x69d9U,0x6815U,
		0x55d8U,0xf37aU,0xc16aU,0x32ceU,0xd0c7U,0x9765U,0xf1b8U,0x59c5U,0x6a7aU,0x84e6U,
		0xbf10U,0x739fU,0xf0aeU,0x1c2eU,0x0ec1U,0x9dc3U,0xf334U,0x73c3U,0xfa3dU,0x45e3U,
		0x155eU,0xf371U,0x2f87U,0x7260U,0xccafU,0xe288U,0x4ebcU,0x3d8bU,0x8f9fU,0xa383U,
		0x1455U,0x008aU,0xd834U,0x0b2bU,0x444dU,0x1d65U,0x309eU,0x0d85U,0x00d6U,0xfab9U,
		0x9b34U,0x0f35U,0xab23U,0x94f2U,0x4806U,0x8b77U,0xe332U,0xe90aU,0x505bU,0xaf75U,
		0x45baU,0x06c1U,0x99c6U,0xc9f4U,0x26d0U,0xc4b0U,0x034cU,0x8cbfU,0x4882U,0xba26U,
		0xa510U,0x57cbU,0x43d0U,0x9b85U,0x93c3U,0x5985U,0xf309U,0x080fU,0x09b2U,0x93ecU,
		0x565fU,0xbc3dU,0x2655U,0x9882U,0x4047U,0x8665U,0x593aU,0x3390U,0x82bbU,0xaa98U,
		0x9dfdU,0xeda7U,0x8d3dU,0x6f6aU,0xe8b3U,0x3f85U,0x5e72U,0xe145U,0x521bU,0xd297U,
		0x028dU,0x6f3eU,0x6af7U,0xe6aaU,0x0abeU,0xd4cfU,0x8ef0U,0x824eU,0xf2a1U,0xce33U,
		0xebd4U,0x6ac2U,0x9e41U,0x5a8eU,0x20ebU,0x206cU,0xa139U,0x9a9eU,0x38a2U,0xf32bU,
		0xac1cU,0x3b5eU,0x1001U,0x127bU,0x43b0U,0xe770U,0x4515U,0x8022U,0x8b13U,0x9785U,
		0x735dU,0xce5eU,0x2617U,0xecf0U,0xbd36U,0x3f6eU,0x6702U,0x143dU,0x5600U,0x8871U,
		0x460eU,0x7465U,0x666dU,0x5bbbU,0xe280U,0x3ffcU,0xc8e8U,0x74b7U,0xb790U,0x5c2dU,
		0xa173U,0xa3aaU,0x1e08U,0x33c6U,0x77eeU,0x64e1U,0xe369U,0x5f06U,0x39daU,0x67ebU,
		0xaf4bU,0x83a4U,0x27b9U,0x9f90U,0xda0eU,0x11e0U,0xdc1cU,0xd125U,0x4b9aU,0xd04bU,
		0x994cU,0x26c0U,0xb829U,0x4adcU,0x19bbU,0x8964U,0x4ab1U,0xede8U,0xca16U,0xc683U,
		0xbfa5U,0xafaaU,0x6e93U,0x4da3U,0x52faU,0x436aU,0xb043U,0x70e7U,0x36a4U,0x2fa5U,
		0x2b0dU,0x0e7dU,0xa238U,0x2bf3U,0x34daU,0x36e9U,0x9ef5U,0xf7aeU,0x5dc3U,0x8502U,
		0xec2aU,0xbd79U,0x4c57U,0x106aU,0x4d2eU,0x6c73U,0x53b6U,0x1982U,0x77feU,0xa861U,
		0xf546U,0x3912U,0x636aU,0x92dcU,0x6561U,0x069bU,0xb4e2U,0x8653U,0xf7d1U,0x0125U,
		0x4dc1U,0xe256U,0x6a66U,0x70a1U,0x5ebbU,0x15f8U,0x8436U,0x8afaU,0x0f5aU,0x56d9U,
		0xdc95U,0x46b9U,0x25d7U,0x0870U,0xc851U,0x8bf2U,0x52beU,0x2dd1U,0x615eU,0x59bcU,
		0x66e4U,0x2853U,0x0bdbU,0xc651U,0xf2b9U,0x5f20U,0xe11dU,0x9d04U,0x10f7U,0x35dfU,
		0x0648U,0x778bU,0xf06bU,0xec4aU,0x8eddU,0x0441U,0xba45U,0xc106U,0x84beU,0x612bU,
		0x495aU,0x3783U,0x23ecU,0xc943U,0xb8cfU,0x78c9U,0xe940U,0xa105U,0xe2f3U,0x653aU,
		0xdde4U,0xa793U,0x85b5U,0x4e2dU,0x0194U,0x165fU,0xd3edU,0xe0f0U,0x6843U,0xa8afU,
		0xa0f4U,0x91e4U,0x3a52U,0x53f1U,0x2b4eU,0x3865U,0x6764U,0x6b0aU,0xdb33U,0x7282U,
		0xcda5U,0x74bfU,0xe01eU,0x0384U,0xc55bU,0x20a0U,0x1ce2U,0xfae6U,0xb8f3U,0xd963U,
		0x61ffU,0xaa0aU,0xb694U,0x270aU,0x2a54U,0xe48fU,0x79bdU,0x2aecU,0x891bU,0x0c73U,
		0xe849U,0x74edU,0x737cU,0x6cc3U,0x991aU,0x0d21U,0x7c47U,0x9344U,0x3264U,0x548dU,
		0x4fbaU,0xfa88U,0x5971U,0x0ae2U,0xc3d1U,0x2c85U,0x5a16U,0x4cc6U,0x83c6U,0x16caU,
		0xb259U,0xf0feU,0x4a54U,0x3dc8U,0x092fU,0xbb9cU,0xeb65U,0x233aU,0x22f6U,0x301cU,
		0x3f48U,0xb7d5U,0xd82dU,0x7323U,0xb8ecU,0x9945U,0x3752U,0x485eU,0xb28cU,0x77d1U,
		0xb71dU,0x7909U,0x6584U,0xaa26U,0xd2f2U,0x6fcdU,0xc365U,0xa490U,0xef72U,0x6a3cU,
		0x4b0aU,0x631bU,0xf822U,0x1d14U,0x1801U,0x4352U,0x4c15U,0xfb0dU,0xe197U,0xd709U,
		0x51c6U,0xdb0bU,0xfeacU,0x16a2U,0x02cfU,0x2a74U,0xad3bU,0x674aU,0x98feU,0xeefdU,
		0xc00fU,0x9fb5U,0x33f4U,0x6891U,0x021fU,0x64a0U,0xb583U,0x904eU,0x881fU,0xdae8U,
		0x0e9aU,0x4eb3U,0xc978U,0xcfb8U,0x6270U,0x9548U,0x1be9U,0x8850U,0x8ec9U,0x50f7U,
		0xf4d2U,0x57f9U,0x1ecaU,0x09f1U,0xc4ceU,0x479bU,0x78a6U,0x9c0cU,0x251cU,0xfb57U,
		0x3bbaU,0xabd5U,0x0e36U,0xd68dU,0xd1a6U,0x49b7U,0x144dU,0xcd6dU,0x0019U,0x1cdaU,
		0x1ab5U,0xd640U,0x2576U,0x35f8U,0x73ceU,0x85a3U,0x4865U,0xa2c7U,0x5d63U,0xba07U,
		0x850dU,0x9e30U,0xbd54U,0x376dU,0x5335U,0x8706U,0x8608U,0x56e4U,0xfe98U,0xfdeeU,
		0x8e47U,0xc14fU,0xbeacU,0x7b78U,0x1f02U,0xa064U,0xeb54U,0x7a33U,0xa110U,0xd237U,
		0x0b25U,0x08deU,0x26baU,0x8248U,0xd95dU,0x8becU,0xa2c1U,0x1d0eU,0xaf03U,0xfaf1U,
		0xee2aU,0x7a5cU,0xbe96U,0x4b5eU,0xb6a7U,0x6f75U,0xda92U,0x5186U,0xc463U,0x2102U,
		0xeb91U,0xee8cU,0x9464U,0xd33fU,0xc9ffU,0x199eU,0xf1a7U,0xb115U,0x7c6dU,0x83afU,
		0xfa4cU,0xbe7eU,0xe927U,0x3cabU,0xce69U,0x1754U,0xe834U,0x81b3U,0xcf2dU,0x5128U,
		0xf1c2U,0xfaf7U,0xebf1U,0x36aeU,0xf0ecU,0x1726U,0xac2cU,0xd360U,0xaa58U,0x5d78U,
		0x5315U,0x5770U,0xba4fU,0x88faU,0x9d94U,0xe5f4U,0x368fU,0x9d89U,0x1a99U,0x210dU,
		0x99d8U,0xf368U,0x7705U,0x3155U,0x0a58U,0x488cU,0xc26aU,0xd4ebU,0xfdf9U,0x04cdU,
		0x670cU,0x0e50U,0xe3c6U,0x1567U,0xdb16U,0x9b08U,0x39f6U,0x21c9U,0x58c5U,0x6167U,
		0xac36U,0x5ed3U,0xa4cfU,0xe7ebU,0x4788U,0x5804U,0x32ddU,0x5d6dU,0x38abU,0x371eU,
		0xad45U,0x9dadU,0x9ffcU,0x99e1U,0x1428U,0x1aeeU,0x2715U,0xc8f0U,0x0ea3U,0x696dU,
		0xe40aU,0x8925U,0x0021U,0x3dd3U,0x3759U,0xaf09U,0xc0d9U,0x741aU,0xfd7bU,0x78e1U,
		0x7746U,0xf5c8U,0x0ccfU,0xe4ffU,0xc492U,0x7205U,0xd784U,0x9bacU,0x2c02U,0x23c4U,
		0x1b8fU,0x9c55U,0xa35cU,0xee4aU,0x45f9U,0xe428U,0x834dU,0xb2b0U,0x8092U,0xfbe8U,
		0x1050U,0xe2adU,0xecb1U,0xdc79U,0xd457U,0xb421U,0xa9f8U,0x1d29U,0xbcd3U,0x6f0aU,
		0x3dbfU,0xace7U,0x2725U,0xf18dU,0xbb1dU,0x8208U,0x863eU,0xbb2cU,0x5f59U,0x917eU,
		0x150fU,0x6bb9U,0xbfd2U,0x50a1U,0x874aU,0x92b8U,0x7efaU,0xf202U,0xebcaU,0xbb7bU,
		0x648dU,0x47cdU,0xcea7U,0xcb75U,0x7f4fU,0x75b6U,0xbc11U,0x4a16U,0xdd77U,0x7a62U,
		0xc03cU,0xf661U,0x8ab6U,0x60baU,0x1412U,0x2ea7U,0x6724U,0x0366U,0x39c3U,0xa9b1U,
		0x71e6U,0xcc90U,0x04feU,0xa24fU,0x7312U,0xf5eeU,0x1aceU,0x6196U,0x8269U,0x388aU,
		0x4ac0U,0xa4b8U,0xb278U,0xbde8U,0x8c8dU,0xa47aU,0xfeffU,0xb9c1U,0x6fb8U,0xb83dU,
		0x7a94U,0xf31aU,0x569bU,0x928bU,0xcec7U,0x196cU,0x4cd8U,0x622aU,0xe3f5U,0x075eU,
		0xb666U,0x6c8eU,0x9b4aU,0x3ce2U,0x1c1dU,0x7cfcU,0x3d3eU,0x7955U,0x9a94U,0x47dcU,
		0x4480U,0x9505U,0xd8b9U,0x7e7cU,0xd1cbU,0xcf5aU,0x4575U,0x077fU,0xdd2aU,0x952dU,
		0xd847U,0x8792U,0x77d8U,0xee30U,0xc9baU,0x4d78U,0x0838U,0x97c8U,0x24eeU,0x10dbU,
		0xd22fU,0x2bebU,0xe1edU,0xb2faU,0xa62eU,0x6503U,0x7db1U,0x0ca2U,0xa905U,0x50beU,
		0xdb47U,0x0495U,0x4284U,0xdf48U,0x5ee3U,0xb302U,0xe5d6U,0x9234U,0x44deU,0x7755U,
		0xaee4U,0xca35U,0xc345U,0x752cU,0xc7a7U,0xaf2aU,0x3c32U,0xc549U,0xb876U,0x3935U,
		0xbff6U,0x7aeeU,0x89a1U,0x57e1U,0x9647U,0x9d16U,0xd824U,0xed4cU,0x4310U,0x86cfU,
		0x5569U,0xdbccU,0x68b7U,0x194fU,0xa6d4U,0xd38bU,0x743bU,0x5b1dU,0x8f3aU,0xea18U,
		0xb166U,0x6e5fU,0xb74cU,0x5b57U,0x8b09U,0x4925U,0x654dU,0x4b64U,0xa7daU,0xb2bfU,
		0xd085U,0xfd23U,0xef0eU,0x4203U,0x8ffeU,0xdd18U,0x41e8U,0xfe47U,0x74a3U,0xdf59U,
		0x55f0U,0xa67eU,0x4d60U,0x333bU,0x0242U,0x17ddU,0xe4a4U,0x31beU,0xe1c6U,0x20ccU,
		0xd5a8U,0xb206U,0xb6deU,0x3daaU,0x11f9U,0x4727U,0x253cU,0x3a42U,0x26a7U,0x0ed7U,
		0xc5e6U,0x5af4U,0xbbc8U,0xa2f6U,0xf555U,0xf4b7U,0xf150U,0x96bbU,0x0b6aU,0xeeb5U,
		0x0327U,0x299eU,0xddb7U,0xdee9U,0xf812U,0xa30bU,0x8d83U,0xa375U,0x4e85U,0x59e0U,
		0xdbacU,0xc753U,0x381dU,0x5c35U,0x8867U,0x56b8U,0xf091U,0xdeadU,0xd648U,0xd47cU,
		0xb293U,0xbe9cU,0x948dU,0x847fU,0x6872U,0x918aU,0x5396U,0x792aU,0x5535U,0x1fc8U,
		0x7f5bU,0x3fc1U,0x43e4U,0x2ff1U,0x6b8aU,0xa09bU,0x308bU,0x7093U,0x7eabU,0x9514U,
		0x98bdU,0xd413U,0x1c42U,0x7211U,0x057bU,0xea06U,0xdefcU,0xfdd4U,0x9410U,0xc1f8U,
		0xcb3fU,0x9738U,0x105cU,0x12a6U,0x415bU,0xf196U,0xda6eU,0xa97aU,0x77e4U,0x768dU,
		0x37f1U,0x7341U,0x0b56U,0x9671U,0x86d8U,0x2771U,0x867dU,0xa8a1U,0xa69fU,0x7961U,
		0xf05aU,0x71a3U,0x6fc4U,0x03b7U,0x52cdU,0xec79U,0xfee1U,0xec82U,0x7ca5U,0xa25bU,
		0xa5f8U,0xfcc7U,0xc643U,0x0825U,0x8ad0U,0x6f7bU,0xf8d3U,0xd88aU,0x7d26U,0x5f3fU,
		0x7715U,0x3d2eU,0xdb53U,0xb7baU,0xd6a2U,0xced6U,0x0a14U,0x0d77U,0x6d8bU,0x4d84U,
		0xac62U,0xb0b3U,0xa6e0U,0x712bU,0xaf5dU,0x0afcU,0x2697U,0xb639U,0x3bffU,0xd430U,
		0x15a7U,0x9c32U,0xed99U,0x3941U,0xf179U,0x329dU,0xc2daU,0xa716U,0x3ff2U,0x420fU,
		0x73e5U,0x40beU,0xdfa7U,0xe98cU,0x5cb0U,0x307bU,0x33aaU,0x3f60U,0xe1d2U,0x267eU,
		0xec59U,0x75e2U,0x1a2eU,0xa963U,0x3579U,0xd285U,0x6e48U,0x2edfU,0x3462U,0x9249U,
		0x8534U,0x96c7U,0x9f82U,0x8f1cU,0x608bU,0x559dU,0x3829U,0xdcbdU,0x7910U,0xb6adU,
		0x57a3U,0x4587U,0x6d97U,0x5d20U,0xddf3U,0x2a89U,0xf5a2U,0xf35aU,0xc414U,0x32a9U,
		0x3b2dU,0x02adU,0xb3aaU,0x116cU,0x9fd4U,0x3426U,0x90b4U,0x3a20U,0x87baU,0x764cU,
		0x1a3aU,0xe1f5U,0xaefbU,0x4c4fU,0xb3e5U,0xc79fU,0xcbedU,0x2cd6U,0x05b2U,0xcb20U,
		0xb3edU,0xb6b9U,0xcfa1U,0x61a5U,0xa33aU,0x03f6U,0x6281U,0xfde0U,0x4a3cU,0xf405U,
		0x6d79U,0x63d6U,0xb551U,0x0849U,0xda04U,0x9618U,0x7f67U,0x69fbU,0xe398U,0x2f74U,
		0x0c2fU,0xfcd3U,0x5178U,0x9c97U,0xf6b3U,0x01f8U,0x86e4U,0xf2edU,0x4b01U,0x1189U,
		0x6ff5U,0xc63aU,0x8544U,0x518cU,0x6e34U,0x5e0bU,0x1257U,0x50e9U,0x3b0bU,0x2ac6U,
		0xe7bdU,0xb7a4U,0x64c1U,0x6ebeU,0x6d22U,0xf4f0U,0x7784U,0x1274U,0x9d82U,0xeef6U,
		0x3334U,0xf5e7U,0xeeefU,0xb0acU,0xf6e8U,0xf5dbU,0x2bb1U,0xafc1U,0x468cU,0x3249U,
		0x3b17U,0x8fcfU,0x6b19U,0x8a92U,0x9b63U,0x7269U,0xa346U,0x6a2eU,0x1d8cU,0xed6dU,
		0x999aU,0x285cU,0x7778U,0x1a10U,0x67e1U,0xd4feU,0x9823U,0xfcdfU,0x856aU,0xf049U,
		0x3b67U,0x717eU,0x0ad4U,0x41a9U,0x76cdU,0x315eU,0xd370U,0x0d9eU,0x57afU,0xb798U,
		0xcfc1U,0xc04bU,0x3046U,0xb8a5U,0x1abdU,0x3a2cU,0xb724U,0xda98U,0xc8baU,0x753bU,
		0x6cbaU,0x8394U,0xfb68U,0xa092U,0x9216U,0x0744U,0xd596U,0x979bU,0x7ad8U,0x1960U,
		0x30b9U,0x6787U,0x8e60U,0x7ff3U,0x288dU,0x52abU,0x2843U,0x2f54U,0xa506U,0x502bU,
		0x1e34U,0xb650U,0x5e8aU,0x1e22U,0xb65aU,0x6012U,0x8419U,0xdc38U,0x7ebbU,0x1eddU,
		0x66b0U,0xfa78U,0xc029U,0xb75fU,0x9e0cU,0x2875U,0xe161U,0x2f06U,0x4a7eU,0x7e70U,
		0x4c9cU,0xea7fU,0xdaccU,0x76b7U,0x45deU,0xb517U,0x6f65U,0x06bcU,0x5f01U,0xe76bU,
		0xe251U,0x2fa0U,0x3860U,0xec45U,0x77ccU,0x933fU,0xd688U,0x9fb0U,0xb110U,0xa05fU,
		0x9da8U,0x2108U,0x92b3U,0xee45U,0x47d7U,0x6191U,0x194aU,0x0490U,0xa370U,0x17d8U,
		0x719eU,0x950fU,0xd280U,0x7126U,0x2f6fU,0x3a1bU,0x41a4U,0xf5e2U,0x06b7U,0x52a6U,
		0x247fU,0x52a1U,0x4b6eU,0x1a68U,0x51ddU,0xd328U,0x25a7U,0x7dc0U,0xab7aU,0x7567U,
		0x751bU,0xaa3eU,0x1406U,0xdeb7U,0xe0d6U,0xca54U,0x1dbaU,0x8eadU,0xb607U,0x40c8U,
		0x7e25U,0xf465U,0x748aU,0x23feU,0x097cU,0x85d1U,0xbc90U,0xa46eU,0x734bU,0x11bbU,
		0xa8daU,0x2484U,0xd7c9U,0xfce9U,0xa275U,0x1322U,0xe742U,0x5547U,0x6a05U,0xd4a0U,
		0xf83dU,0x2635U,0x825dU,0x9325U,0xaa2dU,0xc469U,0x046fU,0x8592U,0xbd19U,0x75aaU,
		0x4383U,0xa99aU,0x4f18U,0xc286U,0x2446U,0x052cU,0x79b4U,0x4cecU,0x2ebcU,0x552cU,
		0x0b93U,0x7876U,0x9b74U,0x1a6dU,0x873eU,0x9586U,0xe341U,0x650bU,0x44b2U,0x8cddU,
		0xb095U,0x425fU,0x8220U,0xc76eU,0xe328U,0x85edU,0x2cdeU,0x30afU,0x3cfdU,0x6badU,
		0x3c49U,0xf741U,0xe85bU,0x3706U,0x9e28U,0x895cU,0x5d18U,0x97c0U,0x2213U,0x75a2U,
		0x91a9U,0xd210U,0x848fU,0xf56eU,0xbb11U,0x4b73U,0x4842U,0x8daaU,0x2a24U,0x8c62U,
		0xfba4U,0xa88fU,0x9f5cU,0x4070U,0xd5baU,0x7c1cU,0x08faU,0x7846U,0x7f7dU,0xee92U,
		0xe572U,0x9f1bU,0xc23aU,0x7a0dU,0xa14fU,0xadc7U,0x7dd6U,0xfbdcU,0x21edU,0x7bbaU,
		0x0c8fU,0x5271U,0x9b94U,0x829eU,0xbf69U,0xf6efU,0xd5d6U,0xd32dU,0xbebfU,0x4977U,
		0xffbaU,0x89e5U,0xd44bU,0x03d8U,0x0f2bU,0x9accU,0x3ebeU,0x073aU,0x9bf3U,0xc898U,
		0x4745U,0x3809U,0xa57bU,0x4f48U,0xc2bbU,0x70b8U,0x7c7dU,0x740eU,0x2425U,0x3c60U,
		0xcdc7U,0x5578U,0x17b5U,0xde45U,0xaaf5U,0x6ed9U,0x827cU,0x0e0cU,0xa02fU,0x51e2U,
		0xfc42U,0xce7dU,0xd778U,0xde29U,0x46d9U,0xa3b2U,0x178dU,0xe8f7U,0x80e5U,0xdbf6U,
		0x3600U,0x4f9eU,0x658bU,0xeb97U,0x61d0U,0x1bd7U,0xe6ddU,0x4701U,0x5211U,0x48e6U,
		0x2ec5U,0x3856U,0xd0a1U,0x14bcU,0xb6d2U,0x58e2U,0xe9b5U,0xfb71U,0x9a2eU,0x1349U,
		0xbe62U,0x7dc5U,0x4cbdU,0xa94aU,0x31a9U,0xc1efU,0x92cdU,0x87f7U,0xd264U,0xe22eU,
		0xe9f3U,0x31b2U,0xd134U,0x72c4U,0xcaadU,0x5b77U,0x3b9dU,0x272eU,0x5c9fU,0x7040U,
		0x1a50U,0xce24U,0xe66dU,0xb4aaU,0x4d54U,0x08cbU,0x4319U,0x48a1U,0x0173U,0xae72U,
		0xa9e3U,0xd66dU,0x90e1U,0x25acU,0xd3c6U,0x0cbeU,0xba17U,0xcc71U,0xe78aU,0x52d6U,
		0x493cU,0x41f7U,0xb48cU,0x6f9eU,0x6ffeU,0x11ccU,0x2269U,0xd345U,0xd4c5U,0x6ca0U,
		0xa765U,0xa055U,0xa182U,0xe444U,0x6392U,0xf130U,0x5613U,0x5ea7U,0x336fU,0xa555U,
		0xaec1U,0xdab9U,0xdba0U,0xafc8U,0xedd0U,0x756cU,0x5761U,0xcdebU,0xb1c9U,0x6604U,
		0xd0c1U,0x2468U,0x6667U,0x09acU,0xdddeU,0x53b0U,0x8ec3U,0x0929U,0x39f0U,0xbe90U,
		0xcec1U,0x3db9U,0xbbc2U,0xb870U,0x73dfU,0xda68U,0x1d86U,0xa334U,0x1db4U,0xc023U,
		0xc2b5U,0x3c43U,0xbbbcU,0x5c99U,0xcccdU,0xc3b4U,0xd631U,0xab7fU,0x36c6U,0x1496U,
		0x6daaU,0xea57U,0x3b8bU,0x5e54U,0x8d17U,0xf7e8U,0x762dU,0x1335U,0xae58U,0xeea9U,
		0x6fd4U,0x946aU,0xc7c0U,0x06e6U,0xc9aeU,0x221bU,0x1f12U,0x8449U,0xb0f4U,0x2362U,
		0x9310U,0x0336U,0x2e97U,0x915eU,0xf2ffU,0xd6b9U,0x170aU,0x66a6U,0xb445U,0xaa43U,
		0x400eU,0x64f7U,0xc92dU,0xa636U,0x831bU,0x2df4U,0xefb6U,0x348eU,0x7c3eU,0xc743U,
		0xbe59U,0x7ea2U,0xcbf5U,0xc5b2U,0xe5caU,0x373eU,0xc3baU,0x0bf9U,0xeb79U,0x4f35U,
		0x61e1U,0x062fU,0x767dU,0x797dU,0x00b7U,0x74d8U,0xecffU,0xd9d2U,0x433cU,0xb5c2U,
		0x94c5U,0x7520U,0xbafcU,0x8e31U,0x7fabU,0x31d6U,0x8600U,0xbf9dU,0xc40cU,0xe1e5U,
		0x9308U,0x243eU,0x10b1U,0x5fa0U,0x84f0U,0x19a4U,0xd160U,0x0e87U,0x135dU,0x4124U,
		0xc723U,0xe0a7U,0x4304U,0x2b72U,0x3386U,0x0728U,0x6135U,0xea75U,0x4ecaU,0x72faU,
		0xa98aU,0x2bb8U,0x7615U,0xdebcU,0xcf7aU,0x7bd7U,0x297dU,0x21b5U,0xdfd2U,0x57d5U,
		0xdc67U,0xbc64U,0x24adU,0x9bc1U,0x887bU,0x9aefU,0xf95cU,0x2f90U,0x73adU,0xffe5U,
		0xccd3U,0xcf0dU,0xb740U,0x66edU,0x5d82U,0x2593U,0x8a23U,0x32bfU,0xd7a2U,0x2f36U,
		0x66ddU,0x9e3aU,0xac25U,0xe190U,0xef07U,0x140bU,0x3b04U,0x770eU,0xa148U,0x1943U,
		0xc71cU,0xa17bU,0xc981U,0xbe1cU,0xda43U,0x74f7U,0x257eU,0x65f4U,0xd2f9U,0xca05U,
		0x453aU,0xff11U,0xb949U,0x29e8U,0xa00bU,0xa03aU,0x1022U,0x463cU,0x80afU,0xb5f3U,
		0x95dfU,0x564aU,0x5148U,0xbe10U,0xbecaU,0x9fefU,0xbd87U,0xca59U,0xc59dU,0x2a2fU,
		0x9ab3U,0x25bcU,0xe617U,0x0f9eU,0x44bdU,0x3d80U,0xb980U,0x3957U,0x3af8U,0xc6d7U,
		0x58b8U,0x5347U,0x68f3U,0xd899U,0xab85U,0x16f2U,0x5a27U,0x2132U,0x5a4dU,0xffa8U,
		0x574cU,0x6a10U,0x95bcU,0x900fU,0x0d00U,0xe3b9U,0x94d9U,0xe184U,0x037aU,0xe0dbU,
		0xb14aU,0xe8edU,0x7ec6U,0x0d39U,0x833bU,0x3ae7U,0x971dU,0xb328U,0x7057U,0x1553U,
		0x5308U,0xcac5U,0xb734U,0xb11bU,0xb0ffU,0x5505U,0x152aU,0xf247U,0xbe23U,0x49a2U,
		0x27ddU,0x7d44U,0x8c00U,0x21a9U,0xf0cbU,0xeb84U,0xe007U,0xcfdaU,0x810fU,0x3250U,
		0xa226U,0x8eb2U,0x006bU,0xfd5cU,0x22ddU,0x8360U,0x1368U,0x6258U,0xf950U,0xe719U,
		0xcf20U,0x2d47U,0x21d8U,0x73b8U,0xc589U,0xcbc2U,0xe456U,0x0886U,0xd637U,0x91bfU,
		0xf7dfU,0x8f13U,0xa07dU,0x017eU,0xe466U,0x5328U,0x12cdU,0x4118U,0x9a39U,0xb00bU,
		0xc703U,0xa0cbU,0x5073U,0x1dbfU,0x40e9U,0x05a1U,0x675aU,0x2958U,0xc7d9U,0xfc38U,
		0x1ef5U,0xa6efU,0x337aU,0x54ebU,0xd108U,0x97acU,0xa497U,0xf1adU,0x3bdcU,0x3e08U,
		0x095dU,0xd218U,0xdd82U,0x3732U,0x25f6U,0x4c08U,0x5fa8U,0x716cU,0xd7e7U,0xc668U,
		0xf2c7U,0x2daeU,0x718eU,0xd462U,0x6457U,0x40cdU,0xdd35U,0x1e5fU,0x00abU,0x0caaU,
		0x0212U,0x58f5U,0xd65bU,0x8c98U,0x6d60U,0x8cfcU,0xf633U,0x69c8U,0xcb28U,0x177dU,
		0x3791U,0xd55dU,0x149cU,0x5eeeU,0x4a67U,0x3badU,0x37d6U,0xbf91U,0x8f45U,0xca90U,
		0xf89bU,0xa0b9U,0x4d44U,0x651dU,0x4b89U,0xbdb1U,0x63e7U,0xb60cU,0xeed5U,0x55cdU,
		0x1aa8U,0x024dU,0x10a5U,0xcb57U,0x53e7U,0xbae5U,0xf225U,0x0e02U,0xf51eU,0xefa6U,
		0x4e90U,0x83b5U,0xe3e0U,0xd6adU,0xe4d7U,0x5a09U,0xc988U,0x65bdU,0x174cU,0x56d1U,
		0xf2e5U,0x6e57U,0xe1ddU,0x4068U,0x3b3fU,0xbaddU,0x005cU,0x4693U,0xb0e8U,0xf46aU,
		0x736fU,0x5a7aU,0x6851U,0xc95eU,0xdb69U,0x6b95U,0x1dcfU,0x6800U,0x2e50U,0xce97U,
		0xa6aaU,0xf903U,0x218dU,0x95d0U,0x2bd2U,0xd2e5U,0x36ccU,0xc737U,0xe20aU,0x7159U,
		0x1d97U,0x28b3U,0x67d1U,0x0a7fU,0xd19dU,0x723dU,0x36bdU,0x6e3fU,0x5aa9U,0xabc3U,
		0x4a47U,0x7e2aU,0xb439U,0x3200U,0x3faeU,0xf727U,0xa1f2U,0x33ceU,0x20aeU,0x1597U,
		0xed37U,0xddfeU,0x85abU,0xf216U,0xc36cU,0x7c73U,0xba76U,0xf424U,0x8690U,0x4c8dU,
		0x2663U,0x2d0bU,0x787fU,0xd92cU,0x153aU,0x8167U,0xf77eU,0x4c7fU,0xa759U,0x074dU,
		0x4aa4U,0x3138U,0xc5bbU,0x2403U,0xd7baU,0xa82dU,0xabe5U,0x5288U,0xacc2U,0xc9e1U,
		0xf995U,0x1c51U,0x893dU,0x8800U,0x1fb7U,0x5e9bU,0xc438U,0x2b00U,0xc0fcU,0x8211U,
		0xea5dU,0x89dbU,0xdf35U,0xf710U,0x913bU,0x6418U,0xd2a5U,0xe102U,0xdbd5U,0x3e62U,
		0x635dU,0x07baU,0xb1bdU,0x894bU,0xc22bU,0x748fU,0xd7f6U,0x67b5U,0x4d02U,0xa81aU,
		0x5dd9U,0xec8bU,0xae28U,0xcc56U,0x9566U,0x0bb7U,0x5195U,0xc844U,0x2b61U,0xbe84U,
		0x4b41U,0x7939U,0x1e49U,0x0f4aU,0x74feU,0xe4ebU,0x3109U,0x3360U,0xccc1U,0x581dU,
		0xb8c2U,0xb217U,0x383fU,0x3f25U,0xe59aU,0x8fd6U,0xfe5eU,0x85d6U,0xab13U,0x3181U,
		0xb80aU,0x944bU,0x4caeU,0xde5dU,0x4bf7U,0xda5cU,0x7828U,0xe110U,0x00e4U,0x4f76U,
		0x6b29U,0xc573U,0x190eU,0xc25aU,0x6db0U,0xc09eU,0xe936U,0xda34U,0x6cfdU,0x473bU,
		0x2b41U,0xfc9aU,0x8d0bU,0x63bfU,0xf9fdU,0xcd04U,0x27f7U,0xae3fU,0xc4beU,0x0981U,
		0xabccU,0x766aU,0xc3abU,0xa95aU,0xf1ebU,0xae36U,0x8151U,0x0187U,0x9431U,0x148aU,
		0x5a5eU,0xadf9U,0x6a43U,0xfa52U,0xecdeU,0x4225U,0x8f54U,0x91b1U,0xe9e7U,0x6b5dU,
		0x5e65U,0xb353U,0x10b9U,0x279bU,0x585dU,0x0318U,0x377cU,0xece9U,0xd4e4U,0xdae1U,
		0x3da3U,0xa473U,0x3242U,0x9c2bU,0x9ac5U,0x7db9U,0xbc5dU,0x09a5U,0x67f9U,0xe712U,
		0xa6c5U,0xda55U,0x19dcU,0xd739U,0x05baU,0x6b37U,0x2280U,0x1545U,0x5e5aU,0xf9adU,
		0x4875U,0x88b6U,0xe9a9U,0x9a13U,0x59b2U,0xfef0U,0xdbe0U,0x7a03U,0x8e80U,0xcb0cU,
		0x22b6U,0x34feU,0x0f90U,0xbc95U,0x542aU,0x8fe4U,0x667bU,0x504eU,0xaff6U,0x384aU,
		0xfd6cU,0x3b47U,0x9e9aU,0x39a1U,0x2802U,0x5451U,0xf75fU,0x3cb1U,0xe77eU,0x164eU,
		0x1dfbU,0x685cU,0xda4aU,0xe6edU,0x5da5U,0x5a6cU,0xa1fdU,0xb07bU,0x54b5U,0x331cU,
		0x7a81U,0x1c84U,0x7b20U,0x3b1eU,0x86f5U,0x11c0U,0x49cbU,0xdc0fU,0xa11fU,0x6462U,
		0x7c65U,0x6a5eU,0x0c27U,0xa6ccU,0xbaf4U,0x483aU,0x5422U,0xeecdU,0xeee0U,0xe291U,
		0xf294U,0x23abU,0x3b91U,0x1864U,0xa4a2U,0xc65aU,0x3d98U,0x582eU,0x97e1U,0x09d7U,
		0x1841U,0xcd12U,0xf2afU,0xa5e9U,0x1284U,0x21e3U,0xbf1dU,0x7350U,0x7062U,0x4895U,
		0xbf83U,0xbaa2U,0x13e6U,0x1e10U,0x9551U,0x44c8U,0xf4c9U,0xc699U,0x73d6U,0xdf9eU,
		0xef79U,0xe92dU,0xebadU,0x1b32U,0x3a68U,0x3acdU,0xc3e1U,0x3925U,0xf3adU,0x3ec8U,
		0xec37U,0x3055U,0xf3e5U,0x889dU,0x2ce8U,0x8ddcU,0xc5d0U,0x786aU,0x6c69U,0x2489U,
		0x207eU,0x6ba3U,0xe418U,0xd8e9U,0x72e1U,0x9e7aU,0x0d49U,0x6004U,0x1e3eU,0xf9e5U,
		0xee65U,0x9a5fU,0x4377U,0xff2bU,0x1714U,0x51f9U,0xf7eeU,0x1b0dU,0xc85eU,0xe30aU,
		0xa193U,0x50d9U,0x6dc3U,0x01e2U,0x11adU,0x613fU,0x16b1U,0x4253U,0xac4eU,0xc604U,
		0xd528U,0xa8dfU,0x0132U,0x2248U,0x6075U,0x3307U,0x2ecfU,0x1881U,0x5812U,0x55b0U,
		0x3f0aU,0x6183U,0x8732U,0xa76fU,0xf97bU,0x175aU,0x418eU,0x8e9dU,0x6497U,0xcd32U,
		0x65fbU,0xb630U,0x5f2dU,0x9442U,0xc300U,0x448aU,0x3fd6U,0xf735U,0xc278U,0x3614U,
		0x0859U,0x8a99U,0x4967U,0xfceeU,0x18d5U,0x3005U,0x6557U,0x76edU,0xe89aU,0x2209U,
		0xf3c5U,0xd740U,0x5d0cU,0x86a7U,0x1135U,0x7594U,0xb7ffU,0xf09bU,0x577eU,0x5929U,
		0x8d1dU,0x46f1U,0x59f7U,0x73efU,0x82d9U,0xe218U,0x7b5aU,0x8c56U,0xa8bcU,0x1677U,
		0xc2a2U,0x4395U,0x37fbU,0x5884U,0x02beU,0xd7ceU,0x8e21U,0x982dU,0x882eU,0x3abbU,
		0x65a0U,0x5406U,0x7f71U,0x8c3cU,0x2967U,0xf61eU,0xc568U,0xb5a7U,0x6322U,0xce6fU,
		0x245aU,0xfd0cU,0x988fU,0xf576U,0x061cU,0x0546U,0x661fU,0x03ccU,0x19acU,0x3e9eU,
		0x5bd9U,0x01afU,0x4ee7U,0x89cbU,0x6438U,0x096bU,0xfa9fU,0x1327U,0x27ecU,0x35a2U,
		0x5228U,0x50c6U,0x9be7U,0x2d23U,0x8af0U,0x23ddU,0xab3fU,0xa885U,0x9c1bU,0xefffU,
		0xb6c1U,0x027bU,0xb1dbU,0x2088U,0x133bU,0x1e6dU,0xf1d0U,0x9f0fU,0xa457U,0xcba5U,
		0x5a37U,0x38d8U,0xe8a4U,0x366cU,0x0db6U,0xd813U,0xfac6U,0xf237U,0xda29U,0xa27aU,
		0xe299U,0x7032U,0x0c83U,0x7227U,0x1ff5U,0xab49U,0x26ddU,0x469bU,0xc11bU,0xa164U,
		0x6d11U,0xea90U,0x31f0U,0x81b9U,0x5c76U,0xcca2U,0x1fa4U,0xdbeaU,0x2585U,0xb18dU,
		0xe1b2U,0x5248U,0x128eU,0xfdabU,0x6267U,0xa3cdU,0x39e7U,0xc2d1U,0x8d5aU,0x6b20U,
		0xc15dU,0x554cU,0xa023U,0xc332U,0x055fU,0xea67U,0x8146U,0x379fU,0x0f04U,0x19e3U,
		0x5bc8U,0xae16U,0x25cdU,0xe825U,0x6d07U,0x08f0U,0x43bdU,0x6ecdU,0x7633U,0xc912U,
		0xc376U,0x5870U,0x1996U,0x011dU,0x69f3U,0xdbc4U,0xb5baU,0xf566U,0x34f6U,0xbda9U,
		0x33e5U,0xf22fU,0x2419U,0xe747U,0xc7e3U,0xa38fU,0x765fU,0x3655U,0xce40U,0x64e9U,
		0xf0baU,0xb154U,0x2238U,0xb61aU,0xa2cfU,0xfe14U,0x4b11U,0xe83aU,0x4360U,0x327cU,
		0x89c1U,0x5522U,0x4c63U,0x63a2U,0x0b9cU,0x9a47U,0x6c2fU,0x4719U,0x8932U,0xba90U,
		0xeac7U,0x921fU,0xe245U,0x70f3U,0xcbfeU,0xd4a5U,0x25e3U,0x3ee3U,0x83d5U,0x98d9U,
		0x7399U,0xf117U,0x5c27U,0x865fU,0xe0eaU,0x92d6U,0x4795U,0xb7cfU,0xe7e5U,0x20fcU,
		0x6c88U,0xbb26U,0xeeafU,0x9d10U,0x3f5aU,0x966bU,0xd4f8U,0x63d0U,0x23f8U,0x7e6aU,
		0x5572U,0x97baU,0xc01dU,0x08c5U,0x32b9U,0x7977U,0x5322U,0x6a0aU,0x0a79U,0xca8aU,
		0xfc94U,0xe0fcU,0x09d1U,0xfeeaU,0x8c50U,0x01dcU,0xdbbeU,0x38d2U,0x854dU,0x7e64U,
		0xe02aU,0x512eU,0x13c1U,0xbc49U,0x6902U,0x6e1fU,0xca0cU,0x18bbU,0x6615U,0x4f0eU,
		0x42c3U,0xe05eU,0x5e7eU,0xc20fU,0x6cd2U,0x3cd4U,0xa74eU,0x7270U,0x34c6U,0x263aU,
		0xfb04U,0x0f14U,0xa04cU,0x5f36U,0x6e16U,0x0f41U,0x5105U,0x6b3eU,0x9d54U,0x2f48U,
		0x0358U,0xeae7U,0x8d63U,0x811dU,0xfb2aU,0x7fbeU,0xae5eU,0xb3c6U,0xef82U,0x38edU,
		0xe375U,0x9093U,0x8e54U,0x8553U,0x3e78U,0x985fU,0xd921U,0x7471U,0x60cdU,0xf1f4U,
		0x4521U,0xf842U,0x5ab2U,0xcbd0U,0xd1b4U,0xbee7U,0xb38eU,0xc7fbU,0x0faeU,0xa086U,
		0x96f0U,0x6aceU,0xf33fU,0x0fc8U,0x1d1bU,0xcf33U,0x180eU,0x117bU,0x5dceU,0x8497U,
		0xb7e6U,0x6979U,0xbc84U,0xdb8eU,0x84f8U,0x0f65U,0xa87bU,0xc764U,0x0e6dU,0x800cU,
		0xa4dbU,0x68d3U,0x4c41U,0x932aU,0x49f6U,0x04d9U,0xe94bU,0xa90dU,0x9fc2U,0xf459U,
		0xba5bU,0x6474U,0xadddU,0xb336U,0x4fc8U,0x93dcU,0xb3f5U,0x1cedU,0xc839U,0x5cd6U,
		0x9470U,0xed5fU,0xf831U,0x8d48U,0xeb3dU,0x5defU,0x35b7U,0x5134U,0x2a3eU,0xd763U,
		0x8ed6U,0x450eU,0x99d1U,0x33edU,0xe49dU,0x8262U,0xeee8U,0xaf56U,0xbeb8U,0x2478U,
		0xcf73U,0x575aU,0x7368U,0x0064U,0x49c4U,0xab0cU,0xa01cU,0x18ceU,0x6a84U,0xfafdU,
		0x4c21U,0x2b17U,0x92a7U,0x3965U,0xd300U,0x2fe3U,0xc0f1U,0x32e8U,0x1032U,0xed81U,
		0x84c8U,0x689dU,0x369aU,0xdcf7U,0xfa95U,0x9b6aU,0xdcd2U,0xc46eU,0x6a8bU,0x711aU,
		0x7333U,0x30edU,0x1bf5U,0xa51aU,0x2a61U,0x05c1U,0xfea3U,0x6935U,0x560aU,0xd699U,
		0x42baU,0xccb8U,0x51d1U,0x4e02U,0x6fdaU,0xfdbbU,0x3e57U,0x0be4U,0xaa62U,0x0aeeU,
		0x302aU,0xe030U,0xa3deU,0xf4ddU,0x79c9U,0x8517U,0x100aU,0x95edU,0x9908U,0xaa32U,
		0xb958U,0x0a61U,0x2303U,0x8df0U,0xb69fU,0x77f6U,0x6279U,0xbfeeU,0xeb71U,0xe853U,
		0x486dU,0x4a5fU,0xf829U,0xf1c8U,0x000eU,0xb383U,0xb9e4U,0xe4b9U,0x0a28U,0x34eaU,
		0x94b7U,0x0c1aU,0x2434U,0xab02U,0x799dU,0x81a9U,0xb9e8U,0xaa0eU,0xad2cU,0x806fU,
		0x5ac3U,0x458cU,0x0e1aU,0x6bedU,0x0111U,0xcc1bU,0x7b13U,0xbe4bU,0xb139U,0x39b1U,
		0x785aU,0x936bU,0xe4bdU,0x6203U,0x8bc3U,0x4fabU,0x138cU,0xa5cfU,0x8598U,0x0fbdU,
		0x8a15U,0xbf76U,0x3195U,0x222cU,0x1baeU,0xcca8U,0x6c1dU,0x49e5U,0x0012U,0x270eU,
		0x27b2U,0xcd9eU,0x852dU,0x57a8U,0xd1c4U,0x3816U,0x3b84U,0xc715U,0xe73bU,0x46d2U,
		0xf1e4U,0x13dfU,0xc7d2U,0xa1ebU,0xb387U,0xb698U,0x6599U,0xce39U,0x0ddcU,0x5c6aU,
		0x61c2U,0xecd1U,0x1bdeU,0xfb1fU,0x5462U,0x407dU,0xb996U,0xf4a2U,0x82acU,0x2911U,
		0x94bbU,0xe493U,0x7fdaU,0xf19dU,0xa3c1U,0x5d25U,0xbc50U,0x0165U,0xf854U,0x2a08U,
		0x8b3fU,0x9eb2U,0xd1ecU,0xc261U,0x9f39U,0xaf3cU,0x0c1eU,0x2a58U,0x4683U,0xdc5eU,
		0x6659U,0x8e73U,0x0475U,0x813aU,0x17c5U,0x62d6U,0x9c01U,0x9854U,0xec1bU,0x5c7cU,
		0xab9eU,0xc77bU,0x0a2cU,0x2af0U,0xc05eU,0x2832U,0x9e8dU,0x6d9cU,0xe819U,0x8629U,
		0xa0ecU,0x32d5U,0x2c46U,0x0eb9U,0xaaedU,0x66d5U,0xddafU,0x8562U,0x34eeU,0x79c1U,
		0x36b5U,0xf2a7U,0xd9c8U,0x4504U,0x38fdU,0xd335U,0xe8dbU,0x4652U,0x1f89U,0xad81U,
		0x0e8fU,0x9d49U,0x7e3fU,0x9a70U,0x79a1U,0x0c77U,0xd5c8U,0x2b34U,0xb70cU,0x2a8eU,
		0x6696U,0x9d63U,0xd027U,0xa37dU,0xb0c5U,0x41c7U,0x46b3U,0x209aU,0xe6a4U,0xd11fU,
		0x81adU,0x891fU,0x6a36U,0x35f2U,0x5b51U,0xdea7U,0x75b0U,0x9527U,0x1183U,0xb89fU,
		0xfcc1U,0x5597U,0x7c16U,0xdbf0U,0x9faaU,0x262fU,0x2438U,0x74f1U,0x6f98U,0x132fU,
		0x0dfcU,0xddf8U,0x154dU,0x54e5U,0x399bU,0xc693U,0x0bb1U,0x1484U,0xa15eU,0xb614U,
		0x617dU,0xf618U,0xab06U,0xe84dU,0x38ccU,0x6ac8U,0x78f7U,0xbcb8U,0x2c3fU,0x62ffU,
		0x61d7U,0xeb33U,0x88daU,0xda88U,0x3523U,0x8ae4U,0xe0c8U,0xc21eU,0xad30U,0x6cc7U,
		0x0b7fU,0x4162U,0xa658U,0xf35fU,0x13c8U,0x441dU,0xdc06U,0x7111U,0xb0a3U,0x7bceU,
		0x3596U,0x1915U,0x912eU,0x2f61U,0x8073U,0x7380U,0x6339U,0xe632U,0x89f4U,0x66bdU,
		0xbd1fU,0x1e91U,0x5fc3U,0xaf1aU,0x562bU,0xa408U,0x77beU,0x1faaU,0xacf8U,0x642cU,
		0xb9ecU,0x0d25U,0xdb00U,0x4f03U,0x3e34U,0xf5a7U,0x5b8cU,0x992fU,0x5091U,0xc171U,
		0x6306U,0x06c9U,0xf771U,0xa4f8U,0x053aU,0x12efU,0xaa12U,0x991eU,0xebf8U,0xebdaU,
		0x051bU,0x089dU,0xde7dU,0xf436U,0x21fdU,0x16ddU,0x4b7cU,0xf16bU,0xd5efU,0x5c02U,
		0xcd56U,0x916bU,0x0e1eU,0x9348U,0xc1a3U,0x255dU,0x3747U,0x32aeU,0xa400U,0xac9aU,
		0xac82U,0x1129U,0x356bU,0x53daU,0x2deaU,0xf44fU,0x79f0U,0x08d4U,0x6bf1U,0x7c4bU,
		0x4aedU,0xcadbU,0x157dU,0xe2b6U,0xa9a0U,0x65deU,0x0950U,0x5388U,0x8e91U,0x798aU,
		0x1de2U,0xb193U,0x049eU,0x481aU,0x5ac7U,0x5491U,0xa2a5U,0xe50fU,0x8db9U,0xc419U,
		0x232fU,0x9660U,0x0b48U,0x009eU,0xe79dU,0x3fcaU,0x0688U,0xb2d2U,0x3554U,0x1f46U,
		0x4590U,0x3268U,0x90cbU,0x5a94U,0xf612U,0x12e9U,0xa1e5U,0x855cU,0x4708U,0xf969U,
		0x1f40U,0x401fU,0x4e0bU,0x3686U,0x7730U,0xc396U,0x7b17U,0xfa8cU,0x9d79U,0xa981U,
		0x14ddU,0x02b2U,0x6e26U,0x2bc5U,0xc35aU,0x8c45U,0xcc33U,0xfe31U,0x5711U,0xc0a5U,
		0xcb8bU,0x8f77U,0xbe4fU,0x4fbeU,0xe750U,0xa2b7U,0x2e43U,0x516aU,0x4389U,0x7ce1U,
		0x1657U,0x5ecbU,0xb06aU,0x84deU,0x3c58U,0x258bU,0x96b3U,0x1a08U,0x0115U,0x0ae6U,
		0x28abU,0x5826U,0xb3d7U,0x3b32U,0xf473U,0x82cdU,0x81d8U,0x976cU,0x78feU,0xca7fU,
		0xe5b2U,0xe722U,0x214eU,0x4025U,0xcc1fU,0x5975U,0x172dU,0x9e47U,0x1a7cU,0x90eeU,
		0xded5U,0xd5deU,0xb4faU,0x38bcU,0x6de3U,0xd182U,0xd168U,0x62c7U,0x7f0bU,0x6241U,
		0x785eU,0x2c89U,0x58abU,0xbc03U,0x06dcU,0x1171U,0xec0aU,0xb9fdU,0xad1cU,0x8d8bU,
		0x1c6dU,0x273bU,0xb689U,0x89b6U,0x7549U,0xd603U,0x936fU,0xc3d5U,0xa34eU,0x80a2U,
		0xbce5U,0xe886U,0xc28cU,0x4179U,0x1816U,0x7581U,0x7019U,0x2029U,0xa437U,0x524eU,
		0xa530U,0x7102U,0xb13dU,0x4ccaU,0x5ff8U,0xfaa7U,0x476cU,0xb3afU,0x2288U,0x9ceaU,
		0x9adcU,0x7bfcU,0x289bU,0x362cU,0x500fU,0x63efU,0x0c68U,0x470eU,0x39b5U,0x5a1aU,
		0x0033U,0x2072U,0xd57bU,0xd357U,0xb76dU,0x7e04U,0xe6e4U,0x2fdaU,0x2abdU,0xe09eU,
		0x1b01U,0x10c9U,0x27d0U,0x7da3U,0x8bc7U,0x16ceU,0xf47cU,0xda75U,0xc8d5U,0x342bU,
		0x6909U,0xad01U,0x2e8dU,0xa871U,0x4be1U,0xfe8eU,0x047eU,0x6db7U,0x3c24U,0xa219U,
		0x4fafU,0x83caU,0x2039U,0x05e3U,0xb36fU,0x42d8U,0x4f1eU,0xf13dU,0xc512U,0x038dU,
		0x9960U,0xaca2U,0x12c0U,0xe1b8U,0x32fbU,0x2cf9U,0xe4c1U,0xf102U,0x5f0fU,0x7bedU,
		0xf680U,0x9fd9U,0x6ceaU,0xe7f4U,0xbb3fU,0xd0ceU,0xbcbfU,0x45c2U,0x8859U,0xd0fbU,
		0x2fd1U,0xf96fU,0x6207U,0xb25dU,0xf0f3U,0x20f1U,0xad68U,0xdd63U,0xe5e9U,0xdcecU,
		0x7b4cU,0xfc87U,0xfc22U,0xadd1U,0xf6d8U,0x0083U,0xb568U,0x3974U,0x859cU,0x3dccU,
		0xd044U,0xfadfU,0xdcb6U,0x3a25U,0x878bU,0xd475U,0xeea2U,0x65edU,0x931eU,0x4f97U,
		0xadf2U,0xdf97U,0x97a5U,0xf20fU,0x0fc1U,0x4a58U,0xb5a0U,0xfe0dU,0xaa4dU,0xc0e6U,
		0x0532U,0x2fc9U,0xb020U,0x0e61U,0xd51aU,0x5c1aU,0xcf03U,0xfdb1U,0x742bU,0x21bfU,
		0x1390U,0xbba0U,0xdff8U,0x9274U,0x2e24U,0x90b9U,0xb88eU,0xe643U,0xdb82U,0x2d5eU,
		0xe238U,0xee57U,0x3510U,0x6035U,0x349cU,0x7736U,0xa5d3U,0x0933U,0x02e2U,0x9aa4U,
		0x0ef6U,0xd914U,0xbb36U,0xcf67U,0x48edU,0x535bU,0x04c2U,0x692aU,0x19cdU,0x7d06U,
		0x8a46U,0x176bU,0x3199U,0x233eU,0x00c5U,0x7694U,0x96e3U,0x7651U,0x18c2U,0x6a9aU,
		0xc553U,0xe402U,0xc2ecU,0x8f97U,0x80ddU,0xda3bU,0xf089U,0x303eU,0x2230U,0xeb69U,
		0xed2fU,0xf4c1U,0xd8ceU,0xc82dU,0x244cU,0x13b4U,0x1a8eU,0xdcc7U,0xb9a9U,0x0ce3U,
		0x2ca2U,0x1294U,0x634eU,0x7cc1U,0x8a19U,0x3020U,0xa265U,0xa4c5U,0xd757U,0x87bfU,
		0x6e09U,0x0da8U,0xd86bU,0x59ccU,0x88e1U,0x040eU,0x2f7eU,0x141fU,0xa41eU,0xc39cU,
		0xbf7aU,0x22faU,0xd367U,0xa13fU,0x22cdU,0x902aU,0x24ddU,0x497fU,0x8723U,0x3406U,
		0x6b48U,0x7762U,0x412cU,0x3a8bU,0x9afdU,0x4df7U,0x6c21U,0xb7d9U,0xeb21U,0x929bU,
		0xc478U,0xe1faU,0x344aU,0x9989U,0xfb3cU,0x59e8U,0x4fdcU,0xc188U,0x3222U,0x15c1U,
		0x275bU,0xa917U,0x49e9U,0x3f4cU,0xed75U,0xfc75U,0x1939U,0x8de6U,0x4cf2U,0x1ae4U,
		0x849fU,0x08acU,0x4ed8U,0x5fd8U,0x477dU,0xa3d3U,0x4494U,0x5b64U,0x1bb2U,0x7327U,
		0x34b8U,0x35aaU,0xe26eU,0x1a3fU,0xf9b5U,0xe9d7U,0x845dU,0x5734U,0x1088U,0x0550U,
		0x3d15U,0x55d5U,0x0268U,0x4e11U,0xccacU,0xd831U,0x6a77U,0xf331U,0x0349U,0x93c0U,
		0x9b31U,0x5058U,0x5218U,0x8eedU,0x2652U,0x9dfaU,0x735aU,0x55fdU,0x20e8U,0x0ffeU,
		0x27b6U,0x9949U,0xc8e5U,0x77ebU,0x34d7U,0x4c54U,0xca13U,0xb040U,0x8433U,0xc84eU,
		0xf543U,0xf7ceU,0x84bbU,0xe93dU,0x0bd8U,0x0645U,0xcda2U,0xb8f0U,0x0191U,0x3a4fU,
		0x596eU,0xb256U,0x79baU,0x9917U,0xd2efU,0xf81fU,0x22f3U,0x374fU,0xb580U,0x626dU,
		0x51c3U,0x98fbU,0x0016U,0x4862U,0x1ec7U,0x3bb7U,0x0b22U,0xaf00U,0x5332U,0xbea9U,
		0xe924U,0xf1bfU,0xda8fU,0xc9fcU,0x0a55U,0xe3c3U,0xaa55U,0x368cU,0x2712U,0x3756U,
		0xac33U,0x38a8U,0x808fU,0xa9f5U,0x0cccU,0x1b8cU,0x648aU,0xdd74U,0xbb1aU,0xbfcfU,
		0xb275U,0x7a91U,0x6721U,0x730fU,0xd1c8U,0x77d5U,0xe3f2U,0x3d3bU,0xe5d3U,0xc7a4U,
		0xd22cU,0xa902U,0x8f37U,0x654aU,0x899eU,0x5566U,0xd5a5U,0x26a4U,0x8ffbU,0x4d5dU,
		0x381aU,0xb290U,0xf14dU,0xf80fU,0x0578U,0x1059U,0x5532U,0x3088U,0xfedeU,0x8acdU,
		0x37eeU,0xa69cU,0x3bfcU,0xc2d7U,0xdb50U,0xac5fU,0x8531U,0x790dU,0x5cadU,0x1a2bU,
		0xaef8U,0xb3eaU,0xf59fU,0x9fd1U,0xf6b0U,0x8541U,0x4a39U,0x7f64U,0x2baeU,0x9b60U,
		0xe7baU,0x9d7fU,0x57acU,0xb721U,0x7775U,0x3b64U,0x1e31U,0x7eb8U,0x9213U,0x8e5dU,
		0x385dU,0x9da5U,0xe15eU,0x45dbU,0x51daU,0x1403U,0xa36dU,0x41a1U,0xe73fU,0xaa2aU,
		0x7e22U,0x7348U,0x44afU,0x2cdbU,0x4f15U,0x0b90U,0xfba1U,0x7f7aU,0x9e25U,0x848cU,
		0xd448U,0x4742U,0x7dd3U,0xbf66U,0x46d6U,0x6588U,0x2422U,0x8279U,0x92caU,0xcaaaU,
		0x2ec2U,0x9a2bU,0xe787U,0x2266U,0xe66aU,0xa9e0U,0xd0beU,0x39edU,0x638fU,0xdb9dU,
		0x3b88U,0x6fd1U,0x1d83U,0xcccaU,0x8318U,0xcbf2U,0xb0f1U,0x1707U,0x85fdU,0x84edU,
		0x61deU,0x4339U,0xdfcfU,0xf959U,0x4301U,0xa987U,0xc719U,0xd2f6U,0x5d7fU,0xac22U,
		0xe614U,0x58b5U,0x101fU,0xbec7U,0x8338U,0xb731U,0x5a4aU,0x94d6U,0x1365U,0xc586U,
		0x27daU,0x810cU,0xc7d6U,0xa494U,0xa07aU,0xc700U,0x020fU,0xcb25U,0x25f3U,0x718bU,
		0x10a2U,0x4e8dU,0x37d3U,0x4b86U,0xdb66U,0x218aU,0x1749U,0x0059U,0xa1efU,0xc369U,
		0x1d94U,0x5aa6U,0xacbfU,0xc435U,0x787cU,0x4aa1U,0x5dd6U,0x2b5eU,0x9138U,0xb1baU,
		0x4cabU,0x6b26U,0x3106U,0xe597U,0xf1e8U,0x6a40U,0x6cfaU,0x27f4U,0xbc5aU,0x05b7U,
		0x5e62U,0xd4e1U,0xaff3U,0xf75cU,0xe9a6U,0x22b3U,0x7c62U,0xeeddU,0x5da2U,0x7b1dU,
		0x13e3U,0xef76U,0x3d95U,0x1281U,0x72deU,0x4374U,0xf3aaU,0xc5cdU,0x2eccU,0xf978U,
		0xa190U,0xac4bU,0xe897U,0xb7fcU,0x5f2aU,0x0856U,0x659dU,0x631fU,0x82d6U,0x37f8U,
		0x9be4U,0xb6beU,0x661cU,0x6435U,0x1ff2U,0x31edU,0xa454U,0xfac3U,0x8143U,0x6d04U,
		0xe1afU,0x8d57U,0xce3dU,0x4b0eU,0x1993U,0x33e2U,0x7396U,0xe7e2U,0x0b99U,0xe242U,
		0x09ceU,0xe027U,0xd4f5U,0x32b6U,0x6e13U,0x8d60U,0x6612U,0xa74bU,0xb38bU,0x1d18U,
		0xe372U,0x60caU,0x9fbfU,0xb3f2U,0xbc81U,0xa4d8U,0xcf70U,0x6a81U,0xeb3aU,0x99ceU,
		0x1bf2U,0x42b7U,0xc0eeU,0xfa92U,0xb69cU,0xf826U,0xaa5fU,0x1007U,0x06f5U,0x452eU,
		0xc1b8U,0xc0ccU,0x4442U,0xeba2U,0x7b04U,0xd2c8U,0x6296U,0x07f7U,0xd1dcU,0x8a6aU,
		0x61c6U,0x4356U,0xc8a3U,0x61f5U,0x563eU,0x61aaU,0x8085U,0x7c27U,0x1303U,0xfe79U,
		0x897eU,0x7fcbU,0x426aU,0x5d5aU,0x6f18U,0x77aeU,0xecd5U,0x1805U,0x579aU,0xc7b7U,
		0xfd8dU,0x1502U,0x1a73U,0xe60bU,0xfcb2U,0xc620U,0x8b6cU,0x8892U,0x0678U,0x5552U,
		0x75cfU,0xff3aU,0x0de0U,0xfb11U,0x7b3bU,0x79dbU,0x7f42U,0xcfa6U,0xc6cbU,0x8c14U,
		0x52b6U,0xebe9U,0x8758U,0xe53bU,0xe56aU,0xd158U,0xa7d2U,0x6e2cU,0x5c6eU,0x4c19U,
		0xe3d8U,0xe776U,0x0341U,0x72d6U,0x78efU,0xd573U,0x14c3U,0x0cf4U,0x72b2U,0x8c80U,
		0x58d8U,0xba86U,0x81efU,0x74e3U,0x5466U,0xd70dU,0x3212U,0x9678U,0xc534U,0x03fbU,
		0xe065U,0xa296U,0xff9cU,0xaaa8U,0x7305U,0x8a80U,0xb454U,0xfca1U,0x44f6U,0xd9b3U,
		0x4081U,0xe19bU,0xe51fU,0xcef3U,0x2aaeU,0xe47eU,0x9b7aU,0x9a54U,0xefedU,0xd8acU,
		0x6d40U,0x704bU,0x4b2dU,0xc163U,0x1c08U,0xc311U,0x1be2U,0xdb0fU,0x4607U,0xe2ecU,
		0x3febU,0xa33fU,0x5694U,0xc5dfU,0xcebaU,0x73a6U,0xe0cfU,0xa574U,0x1907U,0xf28dU,
		0xe44fU,0x2bcbU,0xfb23U,0x51caU,0x5777U,0x43b6U,0x2e3cU,0xb368U,0x6652U,0x89edU,
		0x92c3U,0x738fU,0xd8c7U,0x5967U,0xe0afU,0xf944U,0x2aa7U,0x68aaU,0x82b0U,0x16a6U,
		0xf534U,0xe433U,0x122eU,0xfde5U,0x7e87U,0xdf79U,0x030fU,0xc75bU,0x02a4U,0x9155U,
		0xfd50U,0xbdf3U,0x8994U,0xa77eU,0x2915U,0xfeb0U,0x2864U,0xeb12U,0xa611U,0x9de8U,
		0x958cU,0xca40U,0x6981U,0x537aU,0x97f2U,0xe03dU,0x6e6aU,0xc338U,0x6238U,0x8103U,
		0xb99aU,0x2a78U,0x9877U,0x50ceU,0x0c08U,0x6286U,0x88beU,0xafdeU,0x5bf4U,0x0ee2U,
		0xb4bdU,0x71c6U,0xa621U,0x0255U,0x4996U,0x14e3U,0xf4a6U,0x02d3U,0xd487U,0x8028U,
		0x72baU,0xb7c5U,0xcf8eU,0x16c0U,0xd0a8U,0xb402U,0xd6e4U,0xdfe3U,0x6aecU,0x3effU,
		0xe997U,0xb17cU,0x7fdeU,0x674eU,0xc4f3U,0x0b5dU,0x8238U,0xf40aU,0x42caU,0x68e0U,
		0xa1c2U,0xfa37U,0x54ceU,0x9195U,0x4dbbU,0x2b48U,0x8d37U,0xe363U,0xf1a1U,0xad3fU,
		0xb286U,0x0e30U,0x68b1U,0x8d7dU,0x8744U,0x9a8eU,0xe392U,0x0aceU,0xf054U,0x3573U,
		0xbb0bU,0xa029U,0x6f5fU,0xa8d4U,0x94bfU,0xef01U,0x90dbU,0xd62bU,0x63e1U,0x4a41U,
		0x0374U,0x506dU,0x0f8aU,0xbf17U,0xc225U,0xc4b8U,0xda23U,0x2413U,0xd522U,0x02b8U,
		0xe497U,0x9902U,0x531cU,0x451bU,0x0df6U,0x3e2eU,0x8527U,0x9e87U,0x4766U,0xf67aU,
		0x8db3U,0xb3d1U,0xe268U,0x0b1cU,0x2e1eU,0xd751U,0xbc54U,0x9fb9U,0xaef2U,0x8312U,
		0x0c02U,0x63dbU,0x7f3cU,0x3fe5U,0x260cU,0xf3baU,0x75f4U,0x729dU,0xff57U,0xf588U,
		0x63adU,0x48aaU,0x0169U,0xc013U,0x8770U,0xb8e2U,0xed8cU,0xdc82U,0x6511U,0x94a0U,
		0x9cfcU,0x88f2U,0xf886U,0xc3f6U,0x5269U,0xea6dU,0xdf51U,0xf4e8U,0xa3c5U,0x6895U,
		0x4060U,0x3314U,0x75faU,0x6d7eU,0x1c64U,0x37caU,0xf877U,0xaf85U,0x8017U,0x2ffaU,
		0x298cU,0xc86bU,0xea41U,0xcc39U,0x5d29U,0x33f8U,0x9358U,0x978bU,0xa59bU,0xfc64U,
		0xd8f4U,0x9868U,0x58e9U,0x1bc4U,0xa692U,0x204cU,0xfdc4U,0x188cU,0xd906U,0x86b6U,
		0x8b43U,0x64a4U,0xf64bU,0x2778U,0x0ff6U,0x084eU,0xc216U,0x7d9bU,0xfcf9U,0xf627U,
		0xb174U,0x7c98U,0xb4a0U,0x63c6U,0x9069U,0x3012U,0x9eb6U,0x0223U,0xa398U,0x9838U,
		0xe0b6U,0x578bU,0xe347U,0xdc4dU,0x6865U,0x81c4U,0x3dfcU,0xbabeU,0xdf89U,0x0565U,
		0x076fU,0x29bbU,0xf858U,0x9052U,0xf00aU,0xb220U,0x72a3U,0xb556U,0x5a83U,0xb625U,
		0x8fb9U,0x1c35U,0x352aU,0xd7dbU,0x19eeU,0x2d50U,0x9d24U,0xfe37U,0x2a0cU,0xb587U,
		0x8901U,0x8b19U,0xfd96U,0x1d05U,0xd53fU,0xffc2U,0xd730U,0x93d3U,0xcb17U,0x7e99U,
		0xc72bU,0x0dc3U,0xe7b0U,0x6e73U,0x9f3dU,0xdaecU,0x6c01U,0xf8ecU,0x09e2U,0x961dU,
		0x3349U,0x9c7cU,0xb482U,0xdbb4U,0x7603U,0xef68U,0x2163U,0x67c5U,0xbf41U,0x3327U,
		0xaf40U,0x8823U,0x99a2U,0xea9dU,0x28eaU,0xd937U,0x8ce3U,0x7b69U,0xb7eeU,0x6445U,
		0x01ccU,0x2b83U,0x9816U,0x37a5U,0xf71bU,0x4e53U,0xd1f0U,0x4eb7U,0x2f1bU,0x5230U,
		0x2612U,0xda09U,0x487dU,0x0288U,0x4955U,0xe019U,0x9ef0U,0x0f55U,0x881aU,0x1ab0U,
		0xc3ccU,0xc360U,0xc265U,0x0e9eU,0x26b5U,0xce64U,0x8c88U,0x4570U,0x104bU,0xcea2U,
		0xb6d9U,0x8862U,0x44d9U,0xb161U,0x15a2U,0x9f7dU,0xdef7U,0x7ca0U,0x4687U,0xcfbcU,
		0xb3e0U,0x86dfU,0x25a2U,0x6a00U,0x5e85U,0x77c7U,0x0f26U,0x1788U,0xb090U,0x9f57U,
		0x6662U,0x8d12U,0xd25fU,0x4937U,0xdc62U,0xc97cU,0xefb1U,0xc407U,0xf94bU,0x1ef0U,
		0x44b8U,0x9718U,0x1dcaU,0x20a9U,0xd656U,0x53e2U,0x4bf2U,0x814cU,0xf990U,0xae23U,
		0x0c22U,0x954cU,0x67f4U,0xfd67U,0xf3c0U,0x7f6cU,0x0d44U,0x580dU,0x0effU,0xf0b5U,
		0x8aebU,0x26d8U,0x5100U,0x0fa9U,0x5c22U,0x8c4bU,0x2a5cU,0x6274U,0xba56U,0x7363U,
		0xd1e7U,0xaae8U,0xb134U,0xf1dfU,0x3591U,0xf76cU,0x46aeU,0xa159U,0x570cU,0xe5adU,
		0x2de5U,0x0683U,0x0479U,0x8854U,0xb684U,0x500aU,0x80d8U,0x2f79U,0xadedU,0x350bU,
		0x84b6U,0x0a50U,0x321dU,0x3d10U,0xd443U,0xdfcaU,0xd5a0U,0x2ba9U,0x813eU,0x1bedU,
		0xdb61U,0x7c5dU,0xb44fU,0x1902U,0x4265U,0xe565U,0x4db6U,0xda1eU,0xfd4bU,0xa61cU,
		0xb49bU,0x19e9U,0xff52U,0x2987U,0x665dU,0x50fbU,0x215eU,0x8815U,0x63b3U,0xe39dU,
		0xd43eU,0x5997U,0xb46aU,0x1b68U,0xd26eU,0x67a4U,0x599cU,0x945bU,0xd6efU,0xcb91U,
		0x8e77U,0x8ecdU,0xd9fbU,0xecf6U,0xa939U,0x47f4U,0x8648U,0xd808U,0xfb78U,0x2a17U,
		0xdfb5U,0x781cU,0x039eU,0x1c27U,0x1f60U,0xcd1fU,0x9c05U,0x57fdU,0xa3a3U,0xa8a8U,
		0x2677U,0xfcd8U,0x3cdbU,0x2997U,0x3c3cU,0x2f2fU,0x85caU,0xde3eU,0x63b8U,0xcd0bU,
		0x4111U,0x7236U,0x9858U,0xf4d6U,0x1670U,0xf55fU,0xf53bU,0xa44cU,0xb09bU,0x4bd9U,
		0x15ddU,0x3833U,0x54c6U,0xb088U,0xcde1U,0x0f0aU,0x9257U,0x6887U,0x17c9U,0x09f5U,
		0x7cafU,0xdf69U,0x48b0U,0x0c34U,0x974cU,0xd37aU,0x3354U,0x9dcaU,0x0b86U,0x3640U,
		0xe3a2U,0x2c19U,0x3e20U,0x8f7dU,0x62daU,0x1eceU,0xe5fbU,0x261dU,0xae99U,0xe9c7U,
		0xd421U,0x03e0U,0x0d63U,0x536cU,0x510fU,0x7f2bU,0x2b7aU,0x750eU,0x2020U,0xb1b1U,
		0xaba2U,0x479fU,0x36deU,0x2d18U,0xdcdcU,0x9c9cU,0x969fU,0x7084U,0xf936U,0x5c3dU,
		0x5683U,0xaea4U,0xd273U,0xabf2U,0xf704U,0xe955U,0xc77fU,0xc4d2U,0x1a18U,0x1639U,
		0x1a01U,0x2cf2U,0xc774U,0x6425U,0xdb96U,0xa744U,0x7cbaU,0x98f4U,0x29b4U,0xae1cU,
		0xc30aU,0xa8cdU,0xec1fU,0x9c10U,0x6880U,0x2d2bU,0xff5dU,0x517dU,0x9a1bU,0x662eU,
		0x9a0aU,0x5de6U,0xf3fcU,0x0626U,0x67a9U,0xcb5fU,0x7e18U,0x5717U,0x5c80U,0x78aaU,
		0x4fefU,0x3f74U,0x14cbU,0xfe6bU,0xb91aU,0x6685U,0xe9bcU,0x8619U,0x47beU,0x160cU,
		0xbf27U,0x00feU,0x4ad1U,0xfd01U,0xc062U,0xfb5bU,0x6055U,0x8684U,0xd14aU,0x01fdU,
		0x6cd9U,0x7e50U,0x62efU,0x519fU,0x15e5U,0x7946U,0xb46fU,0xfa04U,0x71b1U,0x95a3U,
		0x2836U,0x2520U,0x815bU,0x1209U,0xe43aU,0x18b1U,0x8226U,0xd6ffU,0x2bdcU,0xbb68U,
		0x2015U,0xac8aU,0x6e88U,0x5bceU,0xab64U,0x2d32U,0x0a30U,0xabd9U,0xf393U,0x4a71U,
		0xf58eU,0xf6b8U,0xe470U,0x80c6U,0x503bU,0x0ec8U,0x4169U,0x8cc7U,0x1b6dU,0x0d0aU,
		0xb028U,0xc0abU,0x2af4U,0x3bbeU,0x9d9bU,0xbd3cU,0x7c31U,0x6576U,0x1c78U,0xb5dfU,
		0x78c1U,0x0e48U,0x484bU,0x680dU,0xc890U,0x9ae7U,0x3a3aU,0x8a8aU,0xe81dU,0xd691U,
		0xf8fbU,0xeec5U,0xc3c3U,0xf2f2U,0x0c5fU,0x5d99U,0x6914U,0x0664U,0xa788U,0xc907U,
		0xd6f4U,0x5898U,0x9e6bU,0x4322U,0x862dU,0x0e3aU,0xb939U,0xf68bU,0x482eU,0xecbaU,
		0x85f3U,0x250eU,0xe171U,0x459bU,0x8578U,0x2056U,0xbb50U,0xe82bU,0x428dU,0xde0bU,
		0x9e91U,0x49bbU,0x4af8U,0x92ffU,0x8776U,0x86e9U,0x17f3U,0xe5bdU,0xd3e2U,0x69e8U,
		0x9975U,0x43edU,0xcb96U,0x8b9cU,0x6e7dU,0xe756U,0x6da0U,0xd1aaU,0xcdb6U,0x1443U,
		0xefc9U,0x803fU,0xed1bU,0x3e81U,0x1350U,0xbd46U,0x239fU,0x89ffU,0x6fe3U,0xf2d4U,
		0xe683U,0x8e11U,0x2c4aU,0xcd71U,0x5721U,0x7968U,0x83e6U,0x118eU,0x7277U,0x364aU,
		0xc455U,0x2970U,0x99b1U,0x1835U,0x59a1U,0xae46U,0x1f94U,0x5b36U,0x0ebdU,0x1451U,
		0xc7ecU,0x69d5U,0x82b7U,0x0abaU,0xe32eU,0x43ccU,0x1e04U,0x4b96U,0xac18U,0x66feU,
		0xb4deU,0x25d3U,0x52f6U,0xec26U,0xa0f0U,0x1cdeU,0x10f3U,0xb8cbU,0xb8e8U,0x4b06U,
		0x7378U,0x83c2U,0x2518U,0x73caU,0xad37U,0xc974U,0x9460U,0xcf29U,0x8e43U,0xa2bdU,
		0x32d9U,0x001dU,0x9d90U,0x6708U,0x150bU,0xbc0dU,0x2bfeU,0xd453U,0x1c19U,0xd843U,
		0x04faU,0x6fb4U,0x430cU,0x8b05U,0x7dadU,0xc341U,0xddb3U,0xd644U,0x55ecU,0x2538U,
		0x8679U,0xc63fU,0x6b86U,0xcb3bU,0xec55U,0x3825U,0x6d87U,0xf175U,0xda00U,0x6ff1U,
		0xb3a6U,0x05aeU,0x8566U,0x1ab9U,0x7780U,0x6b15U,0xdac8U,0xb10cU,0x30b5U,0x8415U,
		0xbc8cU,0x8259U,0x2f6bU,0x7517U,0x91a5U,0x08f6U,0x2eb8U,0xe324U,0xaaf1U,0x35fcU,
		0x9b90U,0x9befU,0x016fU,0x6ffaU,0xe9b1U,0xd130U,0xbbb8U,0xae54U,0xaebdU,0x8ebfU,
		0xecfbU,0x10adU,0xf2fbU,0xbe55U,0x66d9U,0x257aU,0x4ec6U,0x8877U,0x0cfcU,0x5304U,
		0x5144U,0x3af4U,0x9a35U,0xd104U,0xe003U,0x21d4U,0x4d40U,0xf51aU,0xf2c3U,0xf62fU,
		0x36b9U,0x85a7U,0x3b3bU,0xa6a6U,0x6359U,0x5191U,0xa755U,0x1fb3U,0xf9f9U,0x5a5aU,
		0x383bU,0x00e0U,0x8e7cU,0x27feU,0x3778U,0x19d8U,0xf2abU,0x73d2U,0x7a7dU,0x541eU,
		0x16adU,0x872eU,0x2ce4U,0xee61U,0xc29eU,0xc564U,0xc274U,0x1131U,0x0db2U,0x6d0dU,
		0x4ee3U,0x9c17U,0x34f2U,0xa2cbU,0x39e3U,0x25c9U,0xc019U,0x8549U,0xeac3U,0x4791U,
		0xd91dU,0xf33bU,0x6cceU,0x0354U,0x8ed2U,0xa018U,0x0e69U,0x4fc4U,0x79c5U,0x4869U,
		0x3696U,0x5606U,0x8294U,0xc2c7U,0xcbdeU,0x35d5U,0x4bccU,0x3a9cU,0xbf36U,0xfd3aU,
		0xd97bU,0xa6baU,0xce89U,0x62a3U,0x3901U,0xba0bU,0xdd94U,0xb783U,0x217aU,0x5e10U,
		0xdd49U,0x1248U,0xedbcU,0x9dbdU,0x238eU,0xe57eU,0x864dU,0x91deU,0xaa92U,0x33c0U,
		0xd339U,0x5d67U,0x993fU,0xfb51U,0x86c9U,0xdee3U,0x6bb3U,0x7cf6U,0x9612U,0xf043U,
		0xa89bU,0x75dcU,0xd20aU,0x6ed3U,0x76b1U,0xa468U,0xd9ccU,0x9e34U,0xae6cU,0x5c93U,
		0x6517U,0x6e39U,0xe3b3U,0xb005U,0xcb06U,0xa5e3U,0x07b4U,0xccfeU,0xd80dU,0xbda3U,
		0x424dU,0x438fU,0x4508U,0x8511U,0x08bfU,0x746bU,0xcb84U,0x3c1dU,0x9f32U,0x9127U,
		0x7dccU,0xe1a8U,0xf082U,0x0bd1U,0x9062U,0xd258U,0x44efU,0x8d30U,0x1f8dU,0x3771U,
		0x410aU,0x71aaU,0x4905U,0x50eeU,0x2641U,0x5473U,0xc095U,0xb3bdU,0x40b5U,0xb867U,
		0xa93eU,0x0988U,0x2648U,0x47b4U,0xad85U,0xbd58U,0xfa13U,0x7d36U,0x7e8eU,0xb059U,
		0x3d03U,0x30c5U,0x71d7U,0x6ea2U,0xff79U,0x9d39U,0xdc44U,0x43c3U,0x7cd8U,0x9a22U,
		0xe8dfU,0x870aU,0x0aa4U,0xb348U,0x94a6U,0x125cU,0x818dU,0xaadbU,0xd6c9U,0x45eaU,
		0x13cfU,0x2d74U,0x47f9U,0xbcf3U,0x48daU,0x7ce7U,0x4656U,0x5339U,0x2114U,0x4614U,
		0x8455U,0xafebU,0x3993U,0x9ad4U,0x9a02U,0xbbb0U,0x5becU,0x494dU,0x0730U,0xae0cU,
		0xddcfU,0xc939U,0x7e43U,0x56e8U,0x9ee5U,0x7d6fU,0x5c4fU,0x2acbU,0x3a78U,0x0997U,
		0xa328U,0x56c0U,0x063bU,0x9599U,0xdfbaU,0x9115U,0x9386U,0xe6ceU,0x9a74U,0x860cU,
		0xd506U,0x4037U,0x1856U,0x466dU,0xf747U,0x5446U,0x0f6dU,0xdb26U,0x9082U,0x7621U,
		0xd793U,0xc918U,0xa5b0U,0xf50aU,0x0e93U,0xfdf2U,0xbd2fU,0x23e5U,0xed92U,0x3b10U,
		0xfef8U,0x11f2U,0xddd7U,0x24a6U,0xab73U,0x3eb7U,0x7821U,0xbaedU,0xcf19U,0x2e49U,
		0x9d4dU,0xfe9cU,0x5d05U,0x5bc1U,0x81f7U,0xd4b9U,0x05f9U,0xb81cU,0xbe69U,0x2275U,
		0x4c71U,0x146dU,0xb1a4U,0xc520U,0x28dfU,0x9071U,0xd5ccU,0xc153U,0xd679U,0xf061U,
		0xa549U,0xb7a9U,0x34cdU,0xd140U,0xc6b8U,0x6379U,0x925fU,0x640aU,0xfb7dU,0xc4c5U,
		0x547aU,0x6b6cU,0x2b38U,0x8e4bU,0x33a1U,0x1dabU,0xdf80U,0xb4d5U,0x3c4fU,0xd0b5U,
		0x3173U,0xd0dcU,0x7890U,0x79f8U,0x4e48U,0x7639U,0x98e7U,0x6635U,0x79a5U,0x7b7cU,
		0xc0baU,0x0b0cU,0xdc88U,0xe7c2U,0x091dU,0x2351U,0x5eb3U,0xfa44U,0x4424U,0xba2eU,
		0x2a1cU,0x7fa3U,0x7433U,0x5170U,0x0c7bU,0xbeb0U,0x1aa0U,0x6673U,0x9105U,0x8057U,
		0x744eU,0x151bU,0x9803U,0x0657U,0x8c6bU,0x9637U,0x4a22U,0xb22dU,0xaf97U,0x1873U,
		0x669aU,0xa068U,0xb24cU,0x18a7U,0x4f3eU,0x6ec3U,0xb971U,0x6530U,0xb5d2U,0x1b9bU,
		0x1040U,0x2e67U,0x1f65U,0x1df0U,0xf9d7U,0xae7bU,0x9d67U,0x1f06U,0x1d35U,0x890eU,
		0x87a1U,0xb42aU,0x370cU,0x1cafU,0x078bU,0x64b1U,0x7d83U,0x8b8fU,0x2788U,0x5876U,
		0xb30bU,0x5e3eU,0xb710U,0x7a37U,0x3482U,0xb672U,0xf88cU,0x64c6U,0xb9cdU,0x9ec3U,
		0x9f72U,0xb2e0U,0x92f2U,0xa0a4U,0xcd24U,0x71f2U,0x05cfU,0xb070U,0x2a92U,0xeb58U,
		0xc4e1U,0x4002U,0x6995U,0xcd64U,0xa8efU,0x747aU,0xa951U,0xbedeU,0x9240U,0xea4eU,
		0x0bedU,0x9ed3U,0x51b9U,0xab5aU,0xb0c9U,0xd23bU,0xa3ecU,0x03beU,0x5b29U,0xf4f5U,
		0x0f1bU,0x7071U,0x17abU,0xb5b0U,0x9422U,0x6839U,0x03a3U,0x7671U,0x490cU,0xc002U,
		0x41cbU,0xa114U,0x365eU,0x60a4U,0x1235U,0xd091U,0xe861U,0x0aadU,0xe6f6U,0x3233U,
		0xabb1U,0xfbfcU,0x7899U,0xc37cU,0x31caU,0xff64U,0xd02bU,0x08e2U,0xd3b3U,0x3f2eU,
		0xc3fcU,0x6d27U,0xc967U,0xd997U,0x54a6U,0xf378U,0xa65fU,0xc5f2U,0x1c2cU,0x73c1U,
		0x9763U,0x84e4U,0xa381U,0x0b29U,0xf36fU,0xe286U,0x8b75U,0xaf73U,0x0d83U,0x0f33U,
		0x57c9U,0x5983U,0xc9f2U,0x8cbdU,0x338eU,0xeda5U,0x93eaU,0x9880U,0xe6a8U,0x824cU,
		0x3f83U,0xd295U,0xf329U,0x1279U,0x6ac0U,0x206aU,0x3f6cU,0x886fU,0x8020U,0xce5cU,
		0xa3a8U,0x64dfU,0x5bb9U,0x74b5U,0xd123U,0x26beU,0x67e9U,0x9f8eU,0x4da1U,0x70e5U,
		0x8962U,0xc681U,0x8500U,0x1068U,0x0e7bU,0x36e7U,0x0699U,0x0123U,0x1980U,0x3910U,
		0x46b7U,0x8bf0U,0x709fU,0x8af8U,0x9d02U,0x7789U,0x59baU,0xc64fU,0xc941U,0xa103U,
		0x043fU,0x6129U,0xa8adU,0x53efU,0xa791U,0x165dU,0x209eU,0xd961U,0x6b08U,0x74bdU,
		0x74ebU,0x0d1fU,0x2708U,0x2aeaU,0x4cc4U,0xf0fcU,0x548bU,0x0ae0U,0x7321U,0x485cU,
		0xbb9aU,0x301aU,0x6a3aU,0x1d12U,0x7907U,0x6fcbU,0x2a72U,0xeefbU,0xfb0bU,0xdb09U,
		0x4eb1U,0x9546U,0x688fU,0x904cU,0x9c0aU,0xabd3U,0x50f5U,0x09efU,0x35f6U,0xa2c5U,
		0x49b5U,0x1cd8U,0xfdecU,0x7b76U,0x9e2eU,0x8704U,0x8beaU,0xfaefU,0x7a31U,0x08dcU,
		0xee8aU,0x199cU,0x4b5cU,0x5184U,0x81b1U,0xfaf5U,0x83adU,0x3ca9U,0x88f8U,0x9d87U,
		0x1724U,0x5d76U,0x04cbU,0x1565U,0xf366U,0x488aU,0x5802U,0x371cU,0x21c7U,0x5ed1U,
		0x8923U,0xaf07U,0x99dfU,0xc8eeU,0x9baaU,0x9c53U,0x78dfU,0xe4fdU,0xdc77U,0x1d27U,
		0xe426U,0xfbe6U,0x917cU,0x509fU,0xace5U,0x8206U,0x75b4U,0x7a60U,0xf200U,0x47cbU,
		0xcc8eU,0xf5ecU,0x60b8U,0x0364U,0xb9bfU,0xf318U,0x3888U,0xbde6U,0x3ce0U,0x7953U,
		0x196aU,0x075cU,0x952bU,0xee2eU,0x9503U,0xcf58U,0x6501U,0x50bcU,0x97c6U,0x2be9U,
		0xca33U,0xaf28U,0xdf46U,0x9232U,0xed4aU,0xdbcaU,0x3933U,0x57dfU,0x5b55U,0x4b62U,
		0xd389U,0xea16U,0xdf57U,0x3339U,0xfd21U,0xdd16U,0x4725U,0x0ed5U,0x31bcU,0xb204U,
		0x299cU,0xa309U,0xa2f4U,0x96b9U,0xdeabU,0xbe9aU,0x59deU,0x5c33U,0x2fefU,0x7091U,
		0x9188U,0x1fc6U,0xc1f6U,0x12a4U,0xd411U,0xea04U,0x276fU,0x795fU,0xa978U,0x733fU,
		0xfcc5U,0x6f79U,0x03b5U,0xec80U,0x0d75U,0xb0b1U,0x5f3dU,0xb7b8U,0x393fU,0xa714U,
		0x0afaU,0xd42eU,0x267cU,0xa961U,0x40bcU,0x3079U,0x559bU,0xb6abU,0x2eddU,0x96c5U,
		0x02abU,0x3424U,0x5d1eU,0xf358U,0x2cd4U,0xb6b7U,0x764aU,0x4c4dU,0x0847U,0x69f9U,
		0x03f4U,0xf403U,0x1187U,0x518aU,0xfcd1U,0x01f6U,0xf4eeU,0xeef4U,0x50e7U,0xb7a2U,
		0x8fcdU,0x7267U,0xb0aaU,0xafbfU,0xfcddU,0x717cU,0xed6bU,0x1a0eU,0xb8a3U,0xda96U,
		0x315cU,0xb796U,0x195eU,0x7ff1U,0x8392U,0x0742U,0x6010U,0x1edbU,0x2f52U,0xb64eU,
		0xea7dU,0xb515U,0xb75dU,0x2f04U,0x9faeU,0x2106U,0xe769U,0xec43U,0x7124U,0xf5e0U,
		0x618fU,0x17d6U,0x7565U,0xdeb5U,0x529fU,0xd326U,0x85cfU,0x11b9U,0x8eabU,0xf463U,
		0x2633U,0xc467U,0xfce7U,0x5545U,0x4ceaU,0x7874U,0x75a8U,0xc284U,0xc76cU,0x30adU,
		0x9584U,0x8cdbU,0x75a0U,0xf56cU,0xf73fU,0x895aU,0x7c1aU,0xee90U,0x8da8U,0xa88dU,
		0x526fU,0xf6edU,0x7a0bU,0xfbdaU,0x0738U,0x3807U,0x4975U,0x03d6U,0xde43U,0x0e0aU,
		0x70b6U,0x3c5eU,0xdbf4U,0xeb95U,0xce7bU,0xa3b0U,0x58e0U,0x1347U,0x46ffU,0x3854U,
		0x31b0U,0x5b75U,0xa948U,0x87f5U,0x489fU,0xd66bU,0x703eU,0xb4a8U,0x6f9cU,0xd343U,
		0x0cbcU,0x52d4U,0xa553U,0xafc6U,0xa053U,0xf12eU,0x53aeU,0xbe8eU,0xcde9U,0x2466U,
		0x3c41U,0xc3b2U,0xb86eU,0xa332U,0x1333U,0x9468U,0x1494U,0x5e52U,0x915cU,0x66a4U,
		0x2219U,0x2360U,0xc741U,0xc5b0U,0x64f5U,0x2df2U,0x74d6U,0xb5c0U,0x0bf7U,0x062dU,
		0x243cU,0x19a2U,0x8e2fU,0xbf9bU,0xea73U,0x2bb6U,0x4122U,0x2b70U,0x9bbfU,0x2f8eU,
		0x7bd5U,0x57d3U,0x2f34U,0xe18eU,0xcf0bU,0x2591U,0x74f5U,0xca03U,0x770cU,0xa179U,
		0x5648U,0x9fedU,0x29e6U,0x463aU,0x3955U,0x5345U,0x2a2dU,0x0f9cU,0x900dU,0xe182U,
		0x16f0U,0xffa6U,0x1551U,0xb119U,0xe8ebU,0x3ae5U,0xeb82U,0x324eU,0xf245U,0x7d42U,
		0x2d45U,0xcbc0U,0xfd5aU,0x6256U,0x4116U,0xa0c9U,0x91bdU,0x017cU,0x54e9U,0xf1abU,
		0x059fU,0xfc36U,0xc666U,0xd460U,0xd216U,0x4c06U,0x8cfaU,0x177bU,0x1e5dU,0x58f3U,
		0xa0b7U,0xbdafU,0x5eecU,0xbf8fU,0x0e00U,0x83b3U,0x55cbU,0xcb55U,0x4066U,0x4691U,
		0x5a07U,0x56cfU,0xce95U,0x95ceU,0x5a78U,0x6b93U,0x723bU,0xabc1U,0xc735U,0x28b1U,
		0xddfcU,0x7c71U,0x31feU,0x33ccU,0x4c7dU,0x3136U,0x4c8bU,0xd92aU,0x87feU,0x2afeU,
		0xa82bU,0xc9dfU,0x3e60U,0x8949U,0x89d9U,0x6416U,0x0bb5U,0xbe82U,0x67b3U,0xec89U,
		0xb215U,0x8fd4U,0x0f48U,0x335eU,0xe10eU,0xc571U,0x317fU,0xde5bU,0x63bdU,0xae3dU,
		0xc09cU,0x4739U,0x1488U,0xfa50U,0x7668U,0xae34U,0x0316U,0xdadfU,0x91afU,0xb351U,
		0xda53U,0x6b35U,0x9c29U,0x09a3U,0x7a01U,0x34fcU,0xf9abU,0x9a11U,0x399fU,0x3cafU,
		0x8fe2U,0x3848U,0x331aU,0x3b1cU,0x685aU,0x5a6aU,0x4838U,0xe28fU,0xdc0dU,0x6a5cU,
		0xcd10U,0x21e1U,0x1862U,0x582cU,0xc697U,0xe92bU,0x4893U,0x1e0eU,0x889bU,0x7868U,
		0x3acbU,0x3ec6U,0xf9e3U,0xff29U,0x6ba1U,0x9e78U,0x613dU,0xc602U,0x1b0bU,0x50d7U,
		0x6181U,0x1758U,0x2246U,0x187fU,0xf733U,0x8a97U,0xcd30U,0x9440U,0x86a5U,0xf099U,
		0x3003U,0x2207U,0x1675U,0x5882U,0x46efU,0xe216U,0xf61cU,0xce6dU,0x982bU,0x5404U,
		0x01adU,0x0969U,0xf574U,0x03caU,0xa883U,0x0279U,0x35a0U,0x2d21U,0x366aU,0xf235U,
		0x1e6bU,0xcba3U,0xa162U,0x81b7U,0x7030U,0xab47U,0xa3cbU,0x6b1eU,0xdbe8U,0x5246U,
		0xae14U,0x08eeU,0xc330U,0x379dU,0xf564U,0xf22dU,0xc910U,0x011bU,0xb618U,0xe838U,
		0xa38dU,0x64e7U,0xba8eU,0x70f1U,0x5520U,0x9a45U,0x92d4U,0x20faU,0x3ee1U,0xf115U,
		0x97b8U,0x7975U,0x9d0eU,0x63ceU,0x38d0U,0x512cU,0xca88U,0xfee8U,0xc20dU,0x726eU,
		0x6e1dU,0x4f0cU,0x2f46U,0x811bU,0x0f12U,0x0f3fU,0x985dU,0xf1f2U,0xb3c4U,0x9091U,
		0x6accU,0xcf31U,0xcbceU,0xc7f9U,0xc762U,0x68d1U,0x8495U,0xdb8cU,0xb334U,0x1cebU,
		0x04d7U,0xf457U,0xd761U,0x33ebU,0xed5dU,0x5dedU,0xab0aU,0xfafbU,0xaf54U,0x5758U,
		0x689bU,0x9b68U,0x3963U,0x32e6U,0x6933U,0xccb6U,0x7118U,0xa518U,0xf4dbU,0x95ebU,
		0xfdb9U,0x0aecU,0xe851U,0xf1c6U,0x0a5fU,0x77f4U,0x7b10U,0x8bc0U,0x94b4U,0xad29U,
		0xe738U,0x6596U,0x3192U,0x27afU,0x8b3cU,0x4680U,0x545fU,0x7fd7U,0x2c43U,0x36b2U,
		0x9bfeU,0xc05bU,0xb0c2U,0x6a33U,0x1f86U,0xd5c5U,0x0baeU,0x38c9U,0xfcbeU,0x6f95U,
		0xb0a0U,0x6336U,0x88d7U,0x0b7cU,0x6303U,0xebf5U,0x5628U,0xdafdU,0x3568U,0x4aeaU,
		0x4b79U,0xc1a0U,0xe79aU,0x90c8U,0x8e8eU,0xa2a2U,0xcc30U,0xe74dU,0x1f3dU,0x9d76U,
		0x78fbU,0x172aU,0xb067U,0x28a8U,0x1c6aU,0xa34bU,0x6de0U,0x58a8U,0x2898U,0x0030U,
		0x7016U,0x5ff5U,0x4bdeU,0x2036U,0x2abaU,0xf479U,0xbcbcU,0xf0f0U,0x995dU,0x5f0cU,
		0x931bU,0xb59dU,0xfc1fU,0xd041U,0xe235U,0x02dfU,0xd517U,0xdff5U,0xc2e9U,0xed2cU,
		0x04bfU,0x00c2U,0x88deU,0xd364U,0xb9a6U,0xa262U,0x4fd9U,0xed72U,0x6b45U,0xeb1eU,
		0x1085U,0x6a74U,0x4ed5U,0x34b5U,0xf540U,0x018eU,0x264fU,0xc8e2U,0xda8cU,0xac30U,
		0x22f0U,0x1ec4U,0x899bU,0xf14aU,0xbb17U,0xe3efU,0x4a36U,0x7772U,0x37ebU,0x5caaU,
		0x9e22U,0x241fU,0xe15bU,0x7e1fU,0x61dbU,0x5d7cU,0xe667U,0x1d80U,0x37d0U,0x1d91U,
		0x5a47U,0xa077U,0xe9a3U,0x3d92U,0x9135U,0x6cf7U,0xa451U,0x1990U,0xa18dU,0x82d3U,
		0xeb37U,0xaa5cU,0xd4f2U,0xe36fU,0x897bU,0x5797U,0x7b01U,0xc8a0U,0x8755U,0xe3d5U,
		0x8b69U,0x7b38U,0x7302U,0xe51cU,0x72afU,0x320fU,0xe0ccU,0x5774U,0x6d3dU,0x4604U,
		0x02a1U,0x2861U,0xd8c4U,0xf531U,0xb4baU,0xd484U,0x97efU,0x9874U,0x54cbU,0xb283U,
		0xd6e1U,0xc4f0U,0xc222U,0x5319U,0xf051U,0x90d8U,0x75f1U,0x876dU,0x8db0U,0xaeefU,
		0x8014U,0x9355U,0xf883U,0x405dU,0xb171U,0xa395U,0xa68fU,0xf648U,0x3527U,0x88feU,
		0x3df9U,0xf007U,0x7600U,0x999fU,0xcb14U,0x6bfeU,0x9eedU,0x26b2U,0x01c9U,0x2f18U,
		0xb08dU,0xefaeU,0x44d6U,0xb3ddU,0x8ae8U,0xba53U,0xd653U,0x67f1U,0x321aU,0xdb5eU,
		0x46abU,0xb681U,0xd26bU,0xd9f8U,0xfd48U,0x215bU,0x85c7U,0x166dU,0xdfb2U,0xa3a0U,
		0x0b83U,0xe5f8U,0x54c3U,0x7cacU,0x5680U,0x1a15U,0x510cU,0x36dbU,0xf3f9U,0x4fecU,
		0x7cb7U,0x687dU,0x15e2U,0x8158U,0x47bbU,0x6052U,0x4166U,0x9d98U,0x2012U,0xf390U,
		0xa785U,0xb936U,0x4848U,0xf8f8U,0x9972U,0xcdb3U,0x8575U,0x4af5U,0x99aeU,0xc7e9U,
		0x239cU,0x571eU,0xad34U,0x9d8dU,0xac15U,0x10f0U,0x6d84U,0x777dU,0x04f7U,0x55e9U,
		0xaebaU,0x4ec3U,0x2f68U,0x9b8dU,0x3838U,0x7a7aU,0xe000U,0x3b38U,0x6ccbU,0x3693U,
		0xc271U,0x39e0U,0x238bU,0x993cU,0xbf33U,0xdd91U,0x07b1U,0x08bcU,0xa898U,0xae69U,
		0x40b2U,0xfa10U,0xf07fU,0x4107U,0x13ccU,0x2111U,0xff76U,0x0aa1U,0x0638U,0xd503U,
		0x5be9U,0x9ee2U,0xab70U,0x5d02U,0x907fU,0xbd2cU,0x925cU,0x339eU,0x4c6eU,0xd676U,
		0x4421U,0x1a9dU,0x788dU,0xc0b7U,0x103dU,0x1d32U,0x8c68U,0xb249U,0x92efU,0xc4deU,
		0x7d80U,0x347fU,0x941fU,0x365bU,0x923dU,0xa3e9U,0xa65cU,0xf36cU,0xabaeU,0xd3b0U,
		0x801dU,0x67e6U,0xc9efU,0x3f80U,0x043cU,0x6b05U,0x0e78U,0x709cU,0x688cU,0x49b2U,
		0x5488U,0x7904U,0xf363U,0x99dcU,0x7a2eU,0x83aaU,0x3885U,0x9500U,0xe423U,0xf1fdU,
		0x31b9U,0x59dbU,0xdf43U,0xd386U,0x0af7U,0x2edaU,0xd40eU,0x03b2U,0xb0a7U,0x3159U,
		0x7647U,0xfcceU,0x529cU,0xfce4U,0x2f4fU,0xe766U,0x4972U,0xce78U,0x9581U,0x8da5U,
		0xcde6U,0x1491U,0xa945U,0x0cb9U,0x7bd2U,0x7709U,0x64f2U,0x8e2cU,0xfd57U,0x059cU,
		0x2a2aU,0xe8e8U,0x5a75U,0x31fbU,0x1e5aU,0x55c8U,0x317cU,0x7665U,0xa828U,0x67b0U,
		0xdc0aU,0x4890U,0x9c26U,0x8fdfU,0x3000U,0x9828U,0x6b9eU,0x2243U,0xc32dU,0xa38aU,
		0x359dU,0x702dU,0x0f0fU,0xcbcbU,0x3edeU,0xca85U,0x7115U,0x0a5cU,0x04d4U,0xaf51U,
		0xfcbbU,0x5625U,0x318fU,0x9bfbU,0x7013U,0x995aU,0x8e8bU,0xb064U,0x4ed2U,0x22edU,
		0xd514U,0xb9a3U,0x9132U,0xd4efU,0x37e8U,0xe664U,0x97ecU,0xf04eU,0x8b66U,0x6d3aU,
		0x01c6U,0xd650U,0xf880U,0x3df6U,0x7cb4U,0x200fU,0xfd45U,0x54c0U,0x2f65U,0xc26eU,
		0x8572U,0xac12U,0x907cU,0x788aU,0xa895U,0xff73U,0x0e75U,0x7a2bU,0x7d7dU,0xababU,
		0x957eU,0x64efU,0xdf40U,0x7644U,0x359aU,0x04d1U,0x1e57U,0x9c23U,0xf87dU,0x856fU,
		0x8e88U,0x37e5U,0x7d77U,0x846dU,0x7d7aU,0x1e54U,0xdf6eU,0xf526U,0x691dU,0x81deU,
		0x1919U,0xfdfdU,0x8470U,0x9f96U,0x9680U,0x3c90U,0x2e72U,0x2e5bU,0xe9faU,0x6210U,
		0x6920U,0x7812U,0x20b9U,0xb2ebU,0x254dU,0x9840U,0x633dU,0x0e54U,0x81e1U,0xa5ffU,
		0x9981U,0x7183U,0x9d5bU,0xb9f5U,0x707cU,0xcb33U,0xdf71U,0x9c74U,0x17ceU,0x9438U,
		0x098fU,0x2062U,0xe636U,0x6710U,0xf529U,0x6d32U,0xbdfaU,0x804bU,0xfbaaU,0x429dU,
		0x507eU,0xd96cU,0x8473U,0xcae3U,0x2822U,0xce46U,0x83f2U,0xcb66U,0x8077U,0x156bU,
		0x9f99U,0x0076U,0xaf8bU,0x3b6cU,0x7ed1U,0x3727U,0x48f6U,0x8fa6U,0x191cU,0x43d8U,
		0x09faU,0xe622U,0x927cU,0x9772U,0x7384U,0xe3caU,0xfe00U,0x27bfU,0x8a75U,0x34aaU,
		0xfe24U,0x36f6U,0x7555U,0x454fU,0x2e75U,0x7de0U,0x7294U,0xc8feU,0x53d1U,0x555dU,
		0xbd23U,0x9b0cU,0x2e5eU,0x624dU,0x24cfU,0x41aeU,0x501bU,0x12dcU,0xf273U,0xca23U,
		0x9683U,0x2942U,0x9751U,0x525aU,0x7ae2U,0x5956U,0x1e95U,0xdb1aU,0x3c93U,0xb906U,
		0x4018U,0xe097U,0x4076U,0xda81U,0x45d4U,0xac44U,0x6923U,0x9df3U,0x2045U,0xb15aU,
		0x8c79U,0xdfdcU,0x89f8U,0x21cdU,0x7815U,0x1605U,0xea47U,0x0ad9U,0x0bcaU,0x1466U,
		0xc9d8U,0xf10eU,0xe9fdU,0x87eeU,0xd37fU,0x780bU,0xdfeeU,0x2154U,0x66c1U,0x39faU,
		0x6213U,0x11e6U,0xc085U,0xe2c2U,0x5f63U,0x2c5aU,0x72cbU,0xb592U,0x2550U,0x621aU,
		0x3987U,0x3d48U,0x8dc5U,0x54faU,0x562fU,0x616bU,0x9843U,0x082cU,0x3120U,0x3163U,
		0xeaeeU,0xc876U,0xb880U,0xa926U,0x20bcU,0xf442U,0x48b5U,0xae00U,0xfa1aU,0x044eU,
		0xa40cU,0x58c9U,0xb2eeU,0xf15aU,0x286bU,0xe300U,0x9f62U,0xbc33U,0x65d1U,0x791aU,
		0x81e4U,0x04b4U,0xffd2U,0xf0c0U,0xa31cU,0x4ff6U,0x5fc7U,0x5ed7U,0xa602U,0xea2fU,
		0xcc3fU,0x76d2U,0x0799U,0x57b9U,0x9c87U,0x0091U,0x6340U,0x3872U,0x0c39U,0x5916U,
		0x0a8dU,0x402bU,0xaf1eU,0xac3aU,0x0e57U,0xda14U,0x0160U,0x8624U,0x6be8U,0x3811U,
		0x4418U,0x992aU,0x9d5eU,0x54e0U,0x2bc0U,0x82c8U,0xac95U,0x965bU,0xacfcU,0xe7efU,
		0xb9f8U,0x9ce5U,0x6a95U,0x0da3U,0xd470U,0xe63eU,0xb03bU,0xbea4U,0x9984U,0xe9d2U,
		0x0b8bU,0x1702U,0xa8fdU,0x9fccU,0x6430U,0xa4d3U,0x7186U,0xd4dcU,0xa291U,0xc5daU,
		0x7c22U,0x8c0fU,0x68dbU,0x5068U,0xdf74U,0xafd9U,0x7d96U,0xb620U,0x3fe0U,0x37c5U,
		0x77c2U,0x5808U,0x9c77U,0x0283U,0x2992U,0xd375U,0x3506U,0x5992U,0x7e4bU,0x80c1U,
		0x707fU,0x6629U,0x3645U,0x83bdU,0x5d94U,0xe5b8U,0x1faeU,0x478cU,0xcb36U,0xd12bU,
		0x546eU,0xaad6U,0x1243U,0xb000U,0xd13bU,0x234cU,0x0992U,0x11edU,0x706cU,0xd992U,
		0x652bU,0x9ebeU,0xdb04U,0x5d71U,0x2065U,0xc64aU,0xb7b3U,0xb79dU,0x035fU,0xdd11U,
		0xf129U,0x2b6bU,0x17d1U,0xfbd5U,0x3359U,0x5a65U,0x7d3dU,0x56caU,0x4f07U,0x32e1U,
		0x943bU,0x5241U,0xeb19U,0xa072U,0xd5c0U,0x58a3U,0x36d6U,0x55e4U,0xf52cU,0x6bf9U,
		0xe761U,0x223eU,0x9eddU,0x3f7bU,0xb9f0U,0x3722U,0x6d35U,0x37e0U,0xc871U,0x57b4U,
		0x12d7U,0x1461U,0xdd0cU,0x145cU,0xe639U,0x598dU,0x9dcfU,0x94e3U,0x1398U,0xe728U,
		0x0d29U,0x38afU,0x6713U,0xdc22U,0x61b4U,0xad52U,0x2746U,0xa5c2U,0xf7baU,0x0940U,
		0xfbadU,0x1159U,0xd947U,0x7c07U,0xb2a4U,0xaab2U,0x5b90U,0x9db1U,0x42a0U,0xbfb1U,
		0x61eaU,0xc050U,0x46c5U,0x139dU,0xad0cU,0x9038U,0xbdfdU,0x9b15U,0x3e25U,0x656dU,
		0x0894U,0xa9ecU,0x9933U,0xad49U,0x804eU,0x9fe4U,0x6736U,0xaa01U,0x784cU,0x1103U,
		0xf41bU,0x3273U,0x8476U,0x1bceU,0x78d5U,0xfe1aU,0xe5dfU,0xb12aU,0x3e38U,0x99e5U,
		0xcae6U,0x35c9U,0x935eU,0xcfc6U,0x4437U,0xe72dU,0x6cb0U,0x3a5eU,0x5081U,0x3094U,
		0x8f82U,0xc486U,0x0dcdU,0x1733U,0xf5abU,0xa000U,0xd96fU,0xd051U,0xa72dU,0x09beU,
		0xf255U,0x60d6U,0x5b7eU,0x3bc8U,0x83f5U,0x66c8U,0xaa6bU,0xbd97U,0x3f91U,0x2cafU,
		0x630aU,0x1af2U,0xcb69U,0x6f82U,0xfb91U,0xb8aaU,0x8124U,0x9dd4U,0x5d4fU,0x632bU,
		0x2825U,0x4459U,0xe3a7U,0xfa59U,0xad8cU,0x3d75U,0x06cdU,0x142cU,0xce49U,0x7eeeU,
		0x291cU,0xf9c3U,0x0900U,0x8713U,0x5daeU,0x29fbU,0x9f9cU,0x8a53U,0xd0e5U,0xa2d5U,
		0x2316U,0x5c87U,0x5095U,0xc8f4U,0x0079U,0xe5a3U,0x9791U,0x304bU,0xdb72U,0x94e8U,
		0xecc4U,0x0b32U,0x807aU,0x1f7bU,0x2c1eU,0xc592U,0x9606U,0x9e4dU,0xc175U,0x2719U,
		0x156eU,0x4ba0U,0xc629U,0x2c93U,0xa243U,0x474dU,0xede0U,0x3765U,0x7ed4U,0xe64fU,
		0xa992U,0x8441U,0x10d3U,0x2a81U,0x053eU,0x6971U,0x372aU,0x6b55U,0x2e15U,0x3a31U,
		0xcd4dU,0x6718U,0xf076U,0xb29bU,0xaf8eU,0x16e7U,0xe600U,0xc195U,0xad74U,0xc6e6U,
		0x12f3U,0x0ea7U,0x3b6fU,0x676fU,0xb9b5U,0xf269U,0xee98U,0x84acU,0xa4e3U,0x8efbU,
		0x191fU,0x47a8U,0x7250U,0xe840U,0x2140U,0x041aU,0xf775U,0x8929U,0x43dbU,0xb6c9U,
		0x5d2fU,0x1ac2U,0x22beU,0xdc27U,0x56f0U,0x5741U,0x48f9U,0xef4bU,0x2622U,0x4e98U,
		0x066dU,0xcc25U,0xa4fcU,0xe40eU,0x8fa9U,0x26c6U,0xd715U,0xb04dU,0x4138U,0x5157U,
		0xcab4U,0xb266U,0x927fU,0x3a01U,0x6117U,0x15cdU,0xe127U,0x022bU,0xebfcU,0x3dd7U,
		0x9775U,0x8ad7U,0x7cccU,0xda9dU,0x8d6aU,0x0d2eU,0x44a0U,0x4f65U,0x09fdU,0x9b3eU,
		0x62dfU,0x6a4aU,0xbd5fU,0x7d11U,0xebdeU,0x0025U,0xe625U,0x4e1dU,0xfeb7U,0x3411U,
		0x7f83U,0x2d68U,0xcf4eU,0xb8fcU,0xfe03U,0x7c53U,0x87c9U,0x4b17U,0x5fe3U,0x78b1U,
		0xaa16U,0xaf0dU,0x27c2U,0xf8c0U,0x33feU,0xb729U,0x464aU,0x38b4U,0x5364U,0xd83bU,
		0x7387U,0x93cbU,0x1ed3U,0xff21U,0xbba8U,0x597bU,0x9922U,0x375dU,0xe3cdU,0x9952U,
		0x8381U,0x3d5aU,0x7c8aU,0x7c8cU,0x682fU,0x8d9bU,0xfe27U,0x7c8eU,0x4e69U,0xf791U,
		0xa7a7U,0x8383U,0xde81U,0x1d69U,0x36f9U,0x3d5cU,0x8f8fU,0x8484U,0x9eaaU,0xfe29U,
		0xde36U,0x182dU,0x8a78U,0x7c90U,0xd426U,0xde53U,0xb85fU,0x6831U,0xf43aU,0x4451U,
		0x34adU,0x8d9dU,0x8304U,0xfc58U,0x9f21U,0xa7a9U,0x779eU,0x28ffU,0x2e78U,0x8385U,
		0x96ffU,0xfd12U,0xaabfU,0x4e6bU,0x051fU,0x0d89U,0x7de3U,0xf793U,0xd8faU,0x9492U,
		0x2d84U,0x36fbU,0x4de5U,0x80f3U,0x7558U,0x3d5eU,0x03e5U,0xbef7U,0x6f26U,0xde83U,
		0x08a1U,0x30a2U,0x4552U,0x1d6bU,0x41c1U,0x7bc8U,0xbe45U,0x9eacU,0x2735U,0xfe88U,
		0x53d4U,0xfe2bU,0xc182U,0xf7c8U,0x4f91U,0x8f91U,0x4b80U,0xfabdU,0x5560U,0x8486U,
		0x914fU,0x918fU,0x7fc5U,0x8a7aU,0xef62U,0x9f51U,0x7297U,0x7c92U,0xae9eU,0x7940U,
		0x3d0aU,0xde38U,0xf16fU,0x00daU,0xc901U,0x182fU,0x9593U,0x6404U,0xe578U,0xb861U,
		0xce56U,0x9046U,0x2e61U,0x6833U,0xd320U,0x2460U,0xbde0U,0xd428U,0x2201U,0x0f39U,
		0x6250U,0xde55U,0x986eU,0x6877U,0x6f8fU,0x34afU,0x3df0U,0x9c6eU,0xbd26U,0x8d9fU,
		0xe9ccU,0xfbcfU,0x293cU,0xf43cU,0x16e1U,0x9b38U,0x9b0fU,0x4453U,0x229eU,0x1f24U,
		0xff89U,0x9f23U,0xd006U,0x7f8bU,0x501eU,0xa7abU,0xdec4U,0xef92U,0xe52dU,0x8306U,
		0xcd5aU,0x94f6U,0x12dfU,0xfc5aU,0xd880U,0x792fU,0xaa83U,0x2e7aU,0xdf12U,0x5afcU,
		0x24d2U,0x8387U,0x5114U,0xbd0eU,0x475aU,0x77a0U,0x916fU,0xab27U,0x41b1U,0x2901U,
		0x5d3cU,0x090eU,0x7a13U,0xaac1U,0x9332U,0x6c4fU,0x9686U,0x4e6dU,0xe2deU,0xf57cU,
		0x9202U,0x9701U,0xd5f3U,0x8b7bU,0x2945U,0xfd14U,0xa5a1U,0x539bU,0xbc9dU,0x7de5U,
		0x3a80U,0xb50aU,0xf276U,0xf795U,0x7f30U,0xd565U,0x42ecU,0x0521U,0x5c06U,0x480aU,
		0xca26U,0x0d8bU,0xe06dU,0xa858U,0x3a0eU,0x2d86U,0x3ba4U,0x9f06U,0x7ae5U,0x36fdU,
		0x51afU,0xd404U,0x7726U,0xd8fcU,0xc1a7U,0xe90eU,0x5959U,0x9494U,0x601fU,0x1fcdU,
		0xfb31U,0x755aU,0x0a0aU,0xb8b8U,0x9754U,0x3d60U,0x0d68U,0x4b48U,0x30ccU,0x4de7U,
		0x2561U,0xe336U,0x525dU,0x80f5U,0xca47U,0x5f7aU,0xc240U,0x6f28U,0x29ccU,0x6ab1U,
		0x3c96U,0xde85U,0xbf56U,0x9895U,0x4a8bU,0x03e7U,0x0e22U,0xaf79U,0xb909U,0xbef9U,
		0xfc6aU,0x553aU,0x6c46U,0x4554U,0xbb8bU,0xe54aU,0x1e98U,0x1d6dU,0x5371U,0xa73bU,
		0xb896U,0x08a3U,0x934cU,0x505fU,0xdb1dU,0x30a4U,0x2228U,0x46ceU,0x0c16U,0xbe47U,
		0x9850U,0x0eb5U,0x4079U,0x9eaeU,0x5593U,0x1480U,0xad7dU,0x41c3U,0xa404U,0x06c5U,
		0xda84U,0x7bcaU,0x7986U,0x3fc6U,0xf167U,0x53d6U,0x84daU,0xca7bU,0x401bU,0xfe2dU,
		0x2025U,0x3628U,0xd17eU,0x2737U,0xac9eU,0x45beU,0xe09aU,0xfe8aU,0x5c16U,0xee53U,
		0xadcdU,0x4f93U,0x0cdfU,0x040aU,0x6926U,0x8f93U,0x5fd4U,0x054cU,0x775eU,0xc184U,
		0x374bU,0xc9f8U,0x9df6U,0xf7caU,0xa698U,0x7f60U,0xbfcbU,0x5562U,0xa9dcU,0x4335U,
		0x45d7U,0x8488U,0xb1b6U,0x22afU,0x94d2U,0x4b82U,0x32b2U,0x99caU,0xac47U,0xfabfU,
		0x888eU,0xe537U,0xd2c4U,0x7fc7U,0x7047U,0xa570U,0x8c7cU,0x8a7cU,0xe039U,0x71c2U,
		0x5963U,0x9151U,0x356fU,0xc4b4U,0xdfdfU,0x9191U,0xc3f2U,0x2ff6U,0xb3cdU,0x7299U,
		0xbabaU,0xd7d7U,0x2048U,0x7c94U,0x2b7fU,0x0f51U,0x7e95U,0xef64U,0x53deU,0x26d4U,
		0xb15dU,0x9f53U,0xa618U,0x67a0U,0xa155U,0x3d0cU,0xb084U,0x363cU,0x7818U,0xde3aU,
		0x98f0U,0x0622U,0x7f27U,0xaea0U,0xac86U,0x8cc3U,0x1608U,0x7942U,0x2052U,0x43e9U,
		0x6809U,0xc903U,0x66faU,0xc970U,0x89fbU,0x1831U,0x7513U,0x8ebbU,0x6fb0U,0xf171U,
		0x112dU,0x0350U,0x21d0U,0x00dcU,0x75d8U,0xccfaU,0xfd36U,0xe57aU,0x9d35U,0x2d70U,
		0x0bcdU,0xb863U,0x761dU,0x3eb3U,0x4949U,0x9595U,0x79f4U,0xba2aU,0x1469U,0x6406U,
		0x8b8bU,0xa0a0U,0x9633U,0x2e63U,0xfbf8U,0xc5eeU,0xea4aU,0x6835U,0x36e3U,0x6125U,
		0x8cb9U,0xce58U,0x08d8U,0x4886U,0x0adcU,0x9048U,0x922eU,0xb200U,0xfbe2U,0xbde2U,
		0x4c49U,0xafbbU,0xea00U,0xd42aU,0x8cd7U,0x03d2U,0xb64aU,0xd322U,0x2deeU,0x57cfU,
		0x87f1U,0x2462U,0x58efU,0x6b8fU,0x0f98U,0x6252U,0x099fU,0x6a58U,0xc9dbU,0xde57U,
		0x2d1dU,0x3799U,0x9e74U,0x2203U,0xf453U,0xa514U,0xf111U,0x0f3bU,0xa29eU,0x5ff1U,
		0x27abU,0x6f91U,0x5ca6U,0x6cf3U,0xdff1U,0x34b1U,0x4059U,0x2f14U,0x7b34U,0x9870U,
		0x4af1U,0x9b89U,0x2157U,0x6879U,0x347bU,0x7098U,0xae65U,0xbd28U,0x55c4U,0x7029U,
		0xd382U,0x8da1U,0xaba7U,0x1e50U,0xb060U,0x3df2U,0xcadfU,0x43d4U,0x780eU,0x9c70U,
		0x9defU,0x87eaU,0x7ddcU,0x293eU,0x04b0U,0x386eU,0x6216U,0xf43eU,0xafd5U,0x6625U,
		0x54dcU,0xe9ceU,0x6bf5U,0x5989U,0x11e9U,0xfbd1U,0x1bcaU,0x3090U,0x1155U,0x9b11U,
		0x8a4fU,0x1f77U,0x66c4U,0x4455U,0x47a4U,0xef47U,0xe64bU,0x16e3U,0x7c4fU,0x93c7U,
		0x39fdU,0x9b3aU,0x2b26U,0x14f6U,0xcaf0U,0xff8bU,0xcff8U,0xdca6U,0x5f66U,0x9f25U,
		0x9cacU,0x2320U,0x5cc1U,0x22a0U,0xa9a4U,0x0813U,0x2c5dU,0x1f26U,0x75c3U,0x9519U,
		0xf866U,0x5020U,0x1929U,0xe2d0U,0xc088U,0xa7adU,0x96a4U,0x07d1U,0x6791U,0xd008U,
		0x65e2U,0xf30dU,0xe2c5U,0x7f8dU,0xa640U,0xf925U,0x7bc0U,0xe52fU,0x189dU,0x0a97U,
		0x2553U,0x8308U,0x4401U,0x3ea4U,0xc49fU,0xdec6U,0x1581U,0x93f0U,0x621dU,0xef94U,
		0xd90cU,0x7eb0U,0x5c62U,0x12e1U,0x47ecU,0x8037U,0x72ceU,0xfc5cU,0x7089U,0x312eU,
		0x3c15U,0xcd5cU,0xe2baU,0x09b6U,0xb595U,0x94f8U,0x4dd7U,0xa648U,0x6bc3U,0xaa85U,
		0xce2bU,0x434aU,0x8dc8U,0x2e7cU,0xee3dU,0x6c98U,0xf659U,0xd882U,0x8e95U,0xbc41U,
		0x54fdU,0x7931U,0x6649U,0xd418U,0x38f4U,0x24d4U,0x7af8U,0x6bdfU,0x398aU,0x8389U,
		0xdce1U,0xe4f2U,0xff80U,0xdf14U,0x798eU,0x5663U,0x3d4bU,0x5afeU,0x97f9U,0x07a7U,
		0x21f3U,0x475cU,0xb84aU,0x8f26U,0x9846U,0x77a2U,0x60fdU,0x19b2U,0x4a09U,0x5116U,
		0x0954U,0x9886U,0x082fU,0xbd10U,0x86bcU,0x98c2U,0x050bU,0x41b3U,0xd85aU,0xd58bU,
		0x5632U,0x2903U,0x9ca1U,0xef30U,0xf6cdU,0x9171U,0x538cU,0x2659U,0x616eU,0xab29U,
		0x30dcU,0xbc72U,0xdf1dU,0x7a15U,0x56abU,0xfebfU,0xeaf1U,0xaac3U,0x7bdfU,0xf281U,
		0x7a4cU,0x5d3eU,0x04a2U,0x8669U,0xc879U,0x0910U,0x30f7U,0x7216U,0xa801U,0x9688U,
		0x602aU,0xa2feU,0x3123U,0x4e6fU,0x5688U,0x7df6U,0xc9c7U,0x9334U,0x481eU,0x404bU,
		0x3166U,0x6c51U,0xd398U,0x305fU,0x5277U,0x9204U,0x04e1U,0x455dU,0x20bfU,0x9703U,
		0x2dc1U,0x01b5U,0x361eU,0xe2e0U,0x1de6U,0x3394U,0xf445U,0xf57eU,0xfdcaU,0x1c47U,
		0x8fecU,0x2947U,0x5c57U,0xc136U,0xb883U,0xfd16U,0xaea9U,0x5ef6U,0xefe1U,0xd5f5U,
		0xb197U,0x593eU,0xa929U,0x8b7dU,0xc53cU,0x48c7U,0xef1bU,0xbc9fU,0x1a57U,0x343bU,
		0xfa1dU,0x7de7U,0xbaacU,0x0fe6U,0x26f3U,0xa5a3U,0xa2a9U,0xaa9cU,0x0451U,0x539dU,
		0x9269U,0xea0bU,0xef89U,0xf278U,0x20d9U,0xcc60U,0x48b8U,0xf797U,0xf93bU,0x7505U,
		0x9d40U,0x3a82U,0xe513U,0x82bfU,0xae03U,0xb50cU,0xe044U,0xafa1U,0x0c95U,0x42eeU,
		0x7163U,0x3e95U,0xb2f1U,0x0523U,0xb967U,0x5bdfU,0xa3f6U,0x7f32U,0x5acbU,0xedabU,
		0xf15dU,0xd567U,0x1892U,0x0580U,0xb015U,0xca28U,0x98b3U,0xd7b0U,0xa40fU,0x0d8dU,
		0x5c42U,0x6a1aU,0x2e2cU,0x5c08U,0x5495U,0x9e01U,0x58ccU,0x480cU,0x8988U,0x13a6U,
		0x1c94U,0x3a10U,0x4d91U,0x60efU,0x9f65U,0x2d88U,0x6f50U,0x52e4U,0xc6efU,0xe06fU,
		0x2333U,0x6f6eU,0xbc36U,0xa85aU,0x7686U,0xfdd9U,0xf186U,0x7ae7U,0xce1dU,0x4f2eU,
		0x286eU,0x36ffU,0xf709U,0x88afU,0x212bU,0x3ba6U,0x9664U,0x8d41U,0xe303U,0x9f08U,
		0x8a3eU,0xa365U,0x82a4U,0x7728U,0x1f58U,0xe67bU,0x81e7U,0xd8feU,0xa970U,0x89d1U,
		0x44e7U,0x51b1U,0x8dbdU,0x3f89U,0x04b7U,0xd406U,0xf651U,0xdf01U,0x4f89U,0x595bU,
		0x1ca5U,0xb23fU,0x65d4U,0x9496U,0xe95aU,0xcd8dU,0x6c0bU,0xc1a9U,0xc41dU,0xe8b7U,
		0x791dU,0xe910U,0x0306U,0xf92dU,0xd01eU,0xfb33U,0xb4b1U,0xb032U,0xa31fU,0x755cU,
		0x3033U,0x306eU,0xdf09U,0x6021U,0xe7a1U,0xe149U,0x4ff9U,0x1fcfU,0x98a5U,0xc1fdU,
		0x909aU,0x9756U,0x389aU,0x02c7U,0xffd5U,0x3d62U,0xd278U,0x3367U,0x71deU,0x0a0cU,
		0x3fceU,0x5e76U,0xf0c3U,0xb8baU,0x6988U,0x4095U,0x9b9aU,0x30ceU,0xa206U,0x06acU,
		0xa605U,0x4de9U,0x663eU,0x4eedU,0x96d8U,0x0d6aU,0x0b4cU,0xd29bU,0xea32U,0x4b4aU,
		0x277eU,0x9415U,0x1dd8U,0x525fU,0x8582U,0xd9a0U,0x5fcaU,0x80f7U,0xabf7U,0xa22fU,
		0x3290U,0x2563U,0x00a2U,0x521fU,0x5edaU,0xe338U,0xff43U,0x7f15U,0x7d5eU,0xc242U,
		0xc822U,0x3419U,0x079cU,0x6f2aU,0xcf82U,0x1b7eU,0xf486U,0xca49U,0x3558U,0x6f42U,
		0x57bcU,0x5f7cU,0xa1a4U,0x973dU,0x5e20U,0x3c98U,0xc792U,0xbbd0U,0xcc42U,0xde87U,
		0x1a1dU,0x3ef3U,0xd54dU,0x29ceU,0x1f4aU,0x0291U,0x76d5U,0x6ab3U,0xec6fU,0xa810U,
		0xf6f5U,0x4a8dU,0x49feU,0xe804U,0x6343U,0x03e9U,0x69a8U,0x0971U,0xa1b6U,0xbf58U,
		0x068cU,0xe6aeU,0x3875U,0x9897U,0x8b49U,0xcb44U,0x5432U,0xb90bU,0x2ad3U,0xc99aU,
		0x9c8aU,0xbefbU,0x163eU,0x14a4U,0x432bU,0x0e24U,0xb2d6U,0x6afbU,0x0094U,0xaf7bU,
		0x0403U,0xafb4U,0x28f8U,0x6c48U,0xe674U,0xe7fdU,0x0a90U,0x4556U,0xeb09U,0x1630U,
		0x35e9U,0xfc6cU,0x90cfU,0xd4d3U,0x402eU,0x553cU,0x5dfeU,0x12abU,0xe37cU,0x1e9aU,
		0x6947U,0x0c50U,0x0c3cU,0x1d6fU,0xc784U,0x3110U,0x6ea9U,0xbb8dU,0x5a98U,0x0ac2U,
		0x5919U,0xe54cU,0x5381U,0x757aU,0xbf6fU,0xb898U,0x8ac6U,0x2b57U,0x0e5aU,0x08a5U,
		0x88ebU,0x643eU,0xc619U,0x5373U,0x4594U,0x8252U,0xda17U,0xa73dU,0x64aaU,0x1061U,
		0xf03cU,0xdb1fU,0x1774U,0x0272U,0xaf21U,0x30a6U,0xc4d7U,0x7a24U,0x90c1U,0x934eU,
		0x326cU,0x8ef4U,0xac3dU,0x5061U,0x81a7U,0x806dU,0xe4b7U,0x0c18U,0x9369U,0x4fa9U,
		0x6bebU,0xbe49U,0x49e3U,0xcd9cU,0x0fbbU,0x222aU,0xa1e9U,0xce37U,0x3814U,0x46d0U,
		0x290fU,0xf19bU,0xeccfU,0x407bU,0xaf3aU,0xdc5cU,0x0163U,0x9eb0U,0xc779U,0x2830U,
		0x8138U,0x9852U,0x8560U,0xf2a5U,0x8627U,0x0eb7U,0x9a6eU,0x2b32U,0xd333U,0xad7fU,
		0xd11dU,0x35f0U,0x9d61U,0x41c5U,0x262dU,0x132dU,0x9525U,0x5595U,0xf616U,0x6ac6U,
		0x54e3U,0x1482U,0xc21cU,0x4160U,0x62fdU,0xda86U,0x2f5fU,0xe630U,0x441bU,0x7bccU,
		0x642aU,0x4f01U,0x1e8fU,0xa406U,0x12edU,0xebd8U,0x992dU,0x06c7U,0x9169U,0x255bU,
		0xf434U,0xf169U,0x08d2U,0xcad9U,0xac98U,0x53d8U,0x4818U,0xe50dU,0x65dcU,0x7988U,
		0x1f44U,0x5a92U,0x965eU,0x3fc8U,0xc394U,0xa97fU,0x855aU,0x401dU,0x8f75U,0xa2b5U,
		0x2bc3U,0xfe2fU,0x1a06U,0x5824U,0x7cdfU,0x84dcU,0x4023U,0x9e45U,0x82cbU,0xca7dU,
		0x623fU,0xbc01U,0xd5dcU,0xd180U,0xd601U,0x80a0U,0xb9fbU,0x2739U,0x7100U,0xfaa5U,
		0x4177U,0x2027U,0x470cU,0x2070U,0x9ce8U,0x362aU,0x7da1U,0xda73U,0x7e02U,0xe09cU,
		0xa217U,0x05e1U,0xacffU,0xfe8cU,0x2cf7U,0x7bebU,0xf13bU,0xaca0U,0xf96dU,0x20efU,
		0xe7f2U,0x45c0U,0x3972U,0xfaddU,0xdceaU,0xadcfU,0xf20dU,0xfe0bU,0xd473U,0x4f95U,
		0x21bdU,0x9272U,0x2fc7U,0x5c18U,0x7734U,0x9aa2U,0xe641U,0xee55U,0x1769U,0x7692U,
		0xcf65U,0x6928U,0x303cU,0xf4bfU,0x6a98U,0x8f95U,0x7cbfU,0xa4c3U,0x13b2U,0x0ce1U,
		0xc39aU,0xa13dU,0x0da6U,0x040cU,0x4df5U,0x9299U,0x497dU,0x7760U,0xa915U,0xfc73U,
		0x9987U,0xc186U,0x5b62U,0x35a8U,0x1ae2U,0x5fd6U,0x4e0fU,0xf32fU,0xe9d5U,0x054eU,
		0x0ffcU,0x77e9U,0x5056U,0x9df8U,0x0643U,0x3a4dU,0xb03eU,0xf7ccU,0x98f9U,0x3bb5U,
		0x9915U,0x374dU,0x368aU,0x38a6U,0xbea7U,0xc9faU,0x730dU,0x3d39U,0x1b8aU,0xbfcdU,
		0x4d5bU,0xf80dU,0xa900U,0x5564U,0xac5dU,0x1a29U,0x3086U,0xa69aU,0x9d7dU,0x3b62U,
		0x9fcfU,0x7f62U,0x419fU,0x7346U,0x8e5bU,0x45d9U,0xbf64U,0x8277U,0x0b8eU,0x848aU,
		0xdb9bU,0xccc8U,0x9a29U,0xa9deU,0xa985U,0xac20U,0x1705U,0x4337U,0x810aU,0xc6feU,
		0xbec5U,0x94d4U,0x0057U,0x5aa4U,0x7189U,0x4b84U,0xe595U,0x27f2U,0x4a9fU,0xb1b8U,
		0x7b1bU,0x127fU,0xd4dfU,0x22b1U,0x0854U,0x37f6U,0xc5cbU,0xac49U,0x8d55U,0x33e0U,
		0x6433U,0xfac1U,0xa749U,0x60c8U,0xe240U,0x32b4U,0xfa90U,0x1005U,0xa4d6U,0x99ccU,
		0x8a68U,0x61f3U,0xc0caU,0xd2c6U,0x77acU,0xc7b5U,0x7c25U,0x7fc9U,0xff38U,0x79d9U,
		0xe609U,0x8890U,0x6e2aU,0xe774U,0x8c12U,0xe539U,0x74e1U,0x9676U,0xd571U,0x8c7eU,
		0xd9b1U,0xcef1U,0xa294U,0x8a7eU,0xc30fU,0xe2eaU,0x9a52U,0x7049U,0x2bc9U,0x43b4U,
		0xc5ddU,0xa572U,0x68a8U,0xe431U,0x89ebU,0x5965U,0xa77cU,0xeb10U,0xdf77U,0x9153U,
		0x8101U,0x50ccU,0xca3eU,0xe03bU,0x14e1U,0x8026U,0xafdcU,0x71c4U,0xb17aU,0x0b5bU,
		0x16beU,0xdfe1U,0xe361U,0x0e2eU,0x68deU,0x9193U,0xa8d2U,0xd629U,0x9a8cU,0x3571U,
		0x02b6U,0x4519U,0x506bU,0xc4b6U,0xd74fU,0x8310U,0x9e85U,0xb3cfU,0x48a8U,0xb8e0U,
		0x3fe3U,0x729bU,0xf4e6U,0x3312U,0x949eU,0xc3f4U,0xcc37U,0x9789U,0x37c8U,0x2ff8U,
		0x86b4U,0x2776U,0x9866U,0x204aU,0x3010U,0x9836U,0x7d99U,0x7c96U,0x29b9U,0xb21eU,
		0xdc4bU,0xbabcU,0xfe35U,0x8b17U,0xb623U,0xd7d9U,0x6e71U,0xf8eaU,0xffc0U,0x7e97U,
		0x3325U,0xea9bU,0x9c7aU,0xef66U,0x4e51U,0x522eU,0x7b67U,0x2b81U,0xc35eU,0xce62U,
		0x0286U,0x0f53U,0x7c9eU,0x86ddU,0xcea0U,0xb15fU,0x4935U,0xc405U,0x77c5U,0x9f55U,
		0xae21U,0xfd65U,0x9716U,0x53e0U,0x8c49U,0x7361U,0x580bU,0x26d6U,0x0681U,0x5008U,
		0xf1ddU,0xa157U,0x2ba7U,0x7c5bU,0x3509U,0x3d0eU,0x2985U,0x8813U,0xe563U,0xa61aU,
		0xcb8fU,0xecf4U,0x5995U,0x67a2U,0xcd1dU,0xa8a6U,0xd806U,0x781aU,0x7234U,0xf55dU,
		0x2995U,0xde3cU,0x6885U,0xdf67U,0x4bd7U,0xb086U,0x8f7bU,0x261bU,0xd378U,0x363eU,
		0xb1afU,0x2d16U,0x03deU,0x7f29U,0xe953U,0x1637U,0x7082U,0xaea2U,0xa8cbU,0x2d29U,
		0x6423U,0x98f2U,0x5715U,0x3f72U,0x662cU,0x0624U,0xfcffU,0x8682U,0x6683U,0x160aU,
		0x95a1U,0x1207U,0x7e4eU,0x7944U,0x2d30U,0x4a6fU,0xd6fdU,0xac88U,0xc0a9U,0xbd3aU,
		0x80c4U,0x8cc5U,0x8a88U,0xeec3U,0xb5ddU,0x680bU,0x4320U,0xf689U,0x5d97U,0xc905U,
		0xde09U,0x92fdU,0x250cU,0x2054U,0xe754U,0x1441U,0xe5bbU,0x43ebU,0x8e0fU,0x7966U,
		0x3e7fU,0x89fdU,0x5b34U,0x69d3U,0x3648U,0x1833U,0xec24U,0xb8c9U,0x43caU,0x66fcU,
		0xa2bbU,0x6706U,0x83c0U,0xc972U,0xc33fU,0x2536U,0xd451U,0x6fb2U,0x05acU,0x6b13U,
		0xcb39U,0xf173U,0xe322U,0x9bedU,0x8413U,0x7515U,0xbe53U,0x8875U,0xd12eU,0x8ebdU,
		0xf62dU,0xa6a4U,0x3af2U,0x21d2U,0x19d6U,0x541cU,0x1fb1U,0x00deU,0x9c15U,0x25c7U,
		0xee5fU,0x112fU,0x4fc2U,0x5604U,0x478fU,0x0352U,0x62a1U,0xb781U,0x35d3U,0xfd38U,
		0x33beU,0xfb4fU,0x1246U,0xe57cU,0xa466U,0x5c91U,0x7cf4U,0x75daU,0x438dU,0x7469U,
		0xb003U,0xccfcU,0x8d2eU,0x71a8U,0x9125U,0x0bcfU,0x47b2U,0x7d34U,0x5471U,0xb865U,
		0x9a20U,0xb346U,0x30c3U,0x9d37U,0x7ce5U,0x4612U,0xaad9U,0x2d72U,0xc937U,0x7d6dU,
		0x9ad2U,0x494bU,0xe6ccU,0x4035U,0x0995U,0x9597U,0xf508U,0x23e3U,0x5444U,0x761fU,
		0x2e47U,0x5bbfU,0x11f0U,0x3eb5U,0x906fU,0xf05fU,0xb81aU,0x146bU,0x6b6aU,0x1da9U,
		0xd13eU,0x6408U,0x6633U,0x0b0aU,0xd0b3U,0x79f6U,0x516eU,0x6671U,0x234fU,0xba2cU,
		0x1871U,0x18a5U,0x1519U,0x9635U,0xae79U,0x890cU,0x652eU,0x2e65U,0x5e3cU,0xb670U,
		0x1cadU,0x8b8dU,0xb06eU,0x4000U,0x9ec1U,0xa0a2U,0xab58U,0x03bcU,0x7478U,0xea4cU,
		0xc000U,0x60a2U,0x706fU,0x6837U,0xff62U,0x3f2cU,0x0aabU,0xfbfaU,0x84e2U,0xe284U,
		0xd995U,0xc5f0U,0x987eU,0xd293U,0x0f31U,0x8cbbU,0x74b3U,0x9f8cU,0x2068U,0xce5aU,
		0x390eU,0x8af6U,0xc67fU,0x36e5U,0x165bU,0x74bbU,0xc64dU,0x6127U,0x3018U,0x6fc9U,
		0x2ae8U,0x0adeU,0x09edU,0x1cd6U,0xdb07U,0x904aU,0x5182U,0x3ca7U,0x8702U,0x08daU,
		0x5ecfU,0xc8ecU,0x5d74U,0x4888U,0x8204U,0x47c9U,0xe4fbU,0xfbe4U,0x075aU,0xcf56U,
		0x0362U,0xbde4U,0x57ddU,0xea14U,0x2be7U,0x9230U,0x96b7U,0x5c31U,0xdd14U,0xb202U,
		0x733dU,0xec7eU,0x1fc4U,0xea02U,0x3077U,0x96c3U,0xb7b6U,0xd42cU,0xf401U,0x01f4U,
		0xf356U,0x4c4bU,0x1a0cU,0xb794U,0xb7a0U,0xafbdU,0x2f02U,0xec41U,0x0740U,0xb64cU,
		0xf461U,0x5543U,0x17d4U,0xd324U,0x8958U,0xa88bU,0xc282U,0x8cd9U,0x3c5cU,0xa3aeU,
		0xfbd8U,0x03d4U,0xb4a6U,0x52d2U,0x3852U,0x87f3U,0xa330U,0x5e50U,0xf12cU,0x2464U,
		0x062bU,0xbf99U,0x235eU,0x2df0U,0x258fU,0xa177U,0x2b6eU,0x57d1U,0xffa4U,0x3ae3U,
		0x4638U,0x0f9aU,0x017aU,0xfc34U,0x7d40U,0x6254U,0xbf8dU,0xcb53U,0x4c04U,0x58f1U,
		0x28afU,0x33caU,0x56cdU,0x6b91U,0x6414U,0xec87U,0xd928U,0xc9ddU,0x4737U,0xae32U,
		0x335cU,0xde59U,0x9a0fU,0x3846U,0xb34fU,0x09a1U,0x582aU,0x1e0cU,0x5a68U,0x6a5aU,
		0x50d5U,0x187dU,0x3ec4U,0x9e76U,0xe214U,0x5402U,0x943eU,0x2205U,0xcba1U,0xab45U,
		0x03c8U,0x2d1fU,0x0119U,0x64e5U,0x5244U,0x379bU,0x63ccU,0xfee6U,0x9a43U,0xf113U,
		0x908fU,0xc7f7U,0x4f0aU,0x0f3dU,0x5debU,0x5756U,0xdb8aU,0xf455U,0x0aeaU,0x77f2U,
		0x32e4U,0xa516U,0x7fd5U,0xc059U,0xad27U,0x27adU,0x0b7aU,0xdafbU,0xd5c3U,0x6f93U,
		0x9d74U,0x28a6U,0xc19eU,0xa2a0U,0xf477U,0x5f0aU,0x58a6U,0x5ff3U,0x00c0U,0xa260U,
		0xd03fU,0xdff3U,0xc8e0U,0x1ec2U,0xeb1cU,0x34b3U,0x7e1dU,0x1d7eU,0xe3edU,0x5ca8U,
		0x82d1U,0xe36dU,0xa075U,0x6cf5U,0x320dU,0x4602U,0xc89eU,0x7b36U,0xc4eeU,0x90d6U,
		0xf52fU,0x9872U,0xf646U,0xf005U,0xaeedU,0x405bU,0xb3dbU,0x67efU,0x6bfcU,0x2f16U,
		0xa39eU,0x7caaU,0xb67fU,0x2159U,0x6050U,0xf38eU,0x36d9U,0x687bU,0x571cU,0x10eeU,
		0xf8f6U,0x4af3U,0x3b36U,0x39deU,0x55e7U,0x9b8bU,0x4105U,0x0a9fU,0xdd8fU,0xae67U,
		0xd674U,0xc0b5U,0x9ee0U,0xbd2aU,0xa3e7U,0xd3aeU,0xb247U,0x347dU,0x7902U,0x83a8U,
		0x3f7eU,0x709aU,0x03b0U,0xfcccU,0xf1fbU,0xd384U,0x0cb7U,0x8e2aU,0xe764U,0x8da3U,
		0x67aeU,0x8fddU,0xe8e6U,0x55c6U,0xca83U,0xaf4fU,0x2241U,0x702bU,0xb9a1U,0xe662U,
		0x9bf9U,0xb062U,0x54beU,0xac10U,0x6d38U,0x3df4U,0x7642U,0x9c21U,0xff71U,0xaba9U,
		0x81dcU,0x9f94U,0x37e3U,0x1e52U,0x983eU,0xa5fdU,0x2e59U,0x7810U,0x2060U,0x6d30U,
		0xb9f3U,0x9c72U,0xcb64U,0x0074U,0x429bU,0xcae1U,0x9770U,0x27bdU,0x3725U,0x43d6U,
		0x555bU,0x624bU,0x36f4U,0x7ddeU,0x5954U,0xb904U,0x12daU,0x2940U,0xdfdaU,0x1603U,
		0xda7fU,0x9df1U,0x2152U,0x11e4U,0x1464U,0x87ecU,0x54f8U,0x082aU,0x2c58U,0x6218U,
		0x044cU,0xf158U,0xc874U,0xf440U,0x4ff4U,0xea2dU,0xbc31U,0x04b2U,0x4029U,0xda12U,
		0x57b7U,0x3870U,0x9659U,0x9ce3U,0x380fU,0x54deU,0x9fcaU,0xd4daU,0xe63cU,0xe9d0U,
		0x37c3U,0x0281U,0x8c0dU,0xafd7U,0xe5b6U,0xd129U,0x5990U,0x6627U,0x9ebcU,0xc648U,
		0xaffeU,0x11ebU,0x56c8U,0x523fU,0xdd0fU,0xfbd3U,0x3f79U,0x37deU,0x58a1U,0x6bf7U,
		0xe726U,0xdc20U,0x145fU,0x598bU,0xaab0U,0xbfafU,0xa5c0U,0x1157U,0xa9eaU,0x9fe2U,
		0x139bU,0x9b13U,0xb128U,0x35c7U,0x1101U,0x1bccU,0x1731U,0xd04fU,0xe72bU,0x3092U,
		0x2cadU,0x6f80U,0x60d4U,0x66c6U,0x3d73U,0x7eecU,0x9dd2U,0x4457U,0x5c85U,0xe5a1U,
		0x8711U,0x8a51U,0x9e4bU,0x4b9eU,0x94e6U,0x1f79U,0x2a7fU,0x6b53U,0x474bU,0xe64dU,
		0xc6e4U,0x676dU,0x6716U,0x16e5U,0x0418U,0xb6c7U,0x84aaU,0x47a6U,0xcc23U,0x26c4U,
		0xdc25U,0xef49U,0x0229U,0x8ad5U,0x5155U,0x39ffU,0x7d0fU,0x4e1bU,0x0d2cU,0x9b3cU,
		0x78afU,0xf8beU,0x2d66U,0x7c51U,0x5979U,0x9950U,0x38b2U,0x93c9U,0x0702U,0x70c4U,
		0xf913U,0xcaf2U,0xc1dbU,0xca6eU,0x2749U,0xff8dU,0xd05dU,0xc809U,0xfc4eU,0x2b28U,
		0xded9U,0x4ae0U,0xa5c5U,0x14f8U,0x6f00U,0xd88fU,0xf69cU,0x5f68U,0xbe3aU,0x5b9cU,
		0x61b7U,0x9f27U,0xb91fU,0x8aabU,0xe8cfU,0xcffaU,0xd5e2U,0xb82dU,0xad55U,0xdca8U,
		0x9cd1U,0xd067U,0x4f4eU,0x5cc3U,0x9e59U,0xe383U,0xfbb0U,0x22a2U,0xbd02U,0x208eU,
		0x6df1U,0x9caeU,0x1a80U,0x8968U,0x115cU,0x2322U,0xe34dU,0xf8d8U,0xfa65U,0x2c5fU,
		0xf497U,0xd24dU,0xf7bdU,0x1f28U,0x668aU,0xaa77U,0x8c21U,0xa9a6U,0x90f2U,0x19bfU,
		0x0943U,0x0815U,0xadacU,0xc4a7U,0x5ba7U,0xf868U,0x25b3U,0xd8e0U,0xb2a7U,0x5022U,
		0x76a7U,0x1f33U,0x6c13U,0x75c5U,0x6de7U,0xedecU,0xaab5U,0x951bU,0xef57U,0x5f44U,
		0xf849U,0xc08aU,0x5f4fU,0x95b2U,0xd94aU,0xa7afU,0x14d0U,0x85ddU,0x8194U,0x192bU,
		0xd186U,0x4ab5U,0x7c0aU,0xe2d2U,0x88c5U,0xd721U,0xa581U,0x6793U,0x28c7U,0x93fdU,
		0x42a3U,0xd00aU,0x8425U,0xb1e1U,0xdaa9U,0x96a6U,0xb4feU,0xc687U,0xbfb4U,0x07d3U,
		0xdc53U,0x7d2bU,0x1824U,0xe2c7U,0x6d4dU,0xb523U,0x5b93U,0x7f8fU,0xfe70U,0x0a47U,
		0xcc12U,0x65e4U,0x38c0U,0xca1aU,0x9db4U,0xf30fU,0x1f1eU,0xa852U,0x3d54U,0x7bc2U,
		0xccf4U,0x5febU,0x46c8U,0xe531U,0xbc6cU,0x48c1U,0x14f0U,0xa642U,0x7f0fU,0xafaeU,
		0x13a0U,0xf927U,0xfad7U,0x3d33U,0x8067U,0x2555U,0x5002U,0xeebdU,0x61edU,0x830aU,
		0x47c3U,0x3addU,0xb77bU,0x189fU,0x6245U,0xbfa9U,0xc053U,0x0a99U,0xa84cU,0xbcacU,
		0x70beU,0xc4a1U,0x647cU,0x1ea1U,0xbe00U,0xdec8U,0xe874U,0x1e73U,0x6959U,0x4403U,
		0xd16cU,0x4da7U,0x9b18U,0x3ea6U,0xe0bcU,0x771aU,0x3b4fU,0x621fU,0x938eU,0x3beeU,
		0xad0fU,0xef96U,0x1611U,0xca98U,0xccdfU,0x1583U,0x62cbU,0x6e97U,0x903bU,0x93f2U,
		0x44feU,0xbcb2U,0xb37dU,0x5c64U,0x90e8U,0xd351U,0x0897U,0x12e3U,0x9024U,0x93baU,
		0xdd5dU,0xd90eU,0x58afU,0x436eU,0xa9efU,0x7eb2U,0xb362U,0xb7bfU,0x4528U,0x72d0U,
		0x1cffU,0x456aU,0x3e28U,0xfc5eU,0xe9c1U,0xfe65U,0xaae2U,0x47eeU,0xbc07U,0x52feU,
		0x6570U,0x8039U,0xafe5U,0xd4b3U,0xc2c1U,0x3c17U,0xaf6dU,0x0d19U,0x8051U,0xcd5eU,
		0x7febU,0x1341U,0x9c4dU,0x708bU,0x7862U,0x70ebU,0x9fe7U,0x3130U,0x5791U,0xdb58U,
		0x8bbaU,0xb597U,0x561fU,0x3c8aU,0x9936U,0x94faU,0x861eU,0xaad0U,0x34a4U,0xe2bcU,
		0x2c8dU,0xb047U,0xad4cU,0x09b8U,0x23baU,0xbe06U,0xbacbU,0x6bc5U,0xb854U,0xe8c4U,
		0x784fU,0xaa87U,0xbc1dU,0xeca6U,0x3b78U,0x4dd9U,0xec0eU,0x2fa9U,0x1106U,0xa64aU,
		0xd9dbU,0xcedbU,0x5ad5U,0x8dcaU,0x3537U,0x9c61U,0x6739U,0x2e7eU,0x4ad6U,0x2dcbU,
		0x1d5fU,0xce2dU,0xba01U,0x36a8U,0xaa04U,0x434cU,0x4d72U,0xb2b9U,0x7414U,0xf65bU,
		0x4581U,0x5e05U,0x8479U,0xd884U,0x858cU,0x9f15U,0x7535U,0xee3fU,0x06e0U,0x0e81U,
		0x1bd1U,0x6c9aU,0x3e02U,0xd6a7U,0xff0bU,0x54ffU,0x421fU,0x1648U,0xf41eU,0x7933U,
		0xfd06U,0xcc9cU,0x1b2cU,0x8e97U,0x1175U,0x2b11U,0x3276U,0xbc43U,0xde76U,0xdeceU,
		0xb9ddU,0x38f6U,0x0cc5U,0x1018U,0xe5e2U,0x24d6U,0x8520U,0xd538U,0xc1b1U,0x664bU,
		0x1c71U,0x2bf7U,0xb12dU,0xd41aU,0x7447U,0x0d7cU,0xcbd7U,0x398cU,0x29dfU,0x3ac4U,
		0x78d8U,0x838bU,0xbf2cU,0x3188U,0x94adU,0x7afaU,0x273fU,0xa23cU,0xfe1dU,0x6be1U,
		0x0c0fU,0xfd2fU,0x7c83U,0xff82U,0x1c8dU,0x7d57U,0xcae9U,0xdf16U,0xc0c3U,0xf1d6U,
		0xe4b0U,0xdce3U,0xad20U,0x36edU,0x35ccU,0xe4f4U,0xbac4U,0x0a19U,0xf90cU,0x3d4dU,
		0x7aa7U,0xa668U,0x3e3bU,0x5b00U,0x0103U,0x8369U,0x7b95U,0x7990U,0x8d8fU,0x34deU,
		0x99e8U,0x5665U,0x7b44U,0x6482U,0x0a20U,0x21f5U,0x3589U,0x78b9U,0x443aU,0x475eU,
		0xdc6fU,0x394dU,0x4bc4U,0x97fbU,0x754dU,0xf7b2U,0xe730U,0x07a9U,0xcff0U,0x4d89U,
		0x6827U,0x9848U,0x33b6U,0x0b72U,0x9361U,0x77a4U,0x605aU,0xde94U,0xc1d3U,0xb84cU,
		0xd607U,0x9ef9U,0xcfc9U,0x8f28U,0xdda3U,0x2930U,0x3c66U,0x4a0bU,0xba63U,0x12b2U,
		0x5084U,0x5118U,0xbdc4U,0xcbabU,0xe084U,0x60ffU,0xb68dU,0x8506U,0x3097U,0x19b4U,
		0x686bU,0x6d90U,0xfd74U,0x0831U,0xe6d6U,0xc9a7U,0x6cb3U,0xbd12U,0x8689U,0x8f4dU,
		0xb942U,0x0956U,0x89baU,0x5dc7U,0x3a61U,0x9888U,0xd9bbU,0x1ea7U,0xfff9U,0x050dU,
		0xd3cdU,0xbcc8U,0x0dd0U,0x41b5U,0x82f9U,0x9cc6U,0x1381U,0x86beU,0xa352U,0xbd7dU,
		0x1736U,0x98c4U,0x2170U,0xb0b8U,0x5ab9U,0x5634U,0x6ef6U,0xa9c7U,0x8f85U,0x2905U,
		0xc067U,0xab1aU,0x1263U,0xd85cU,0x80a6U,0xec2eU,0xc489U,0xd58dU,0x628dU,0x5703U,
		0x242bU,0xf6cfU,0x8b33U,0x728bU,0xd972U,0x9173U,0x49daU,0xa45dU,0x4e60U,0x9ca3U,
		0x9373U,0x106eU,0xd054U,0xef32U,0x81caU,0xac67U,0x1cbfU,0x6170U,0x00ceU,0x2a4cU,
		0xf5aeU,0xab2bU,0xfb60U,0x5209U,0x773eU,0x538eU,0xc3d9U,0x4c5bU,0xa003U,0x265bU,
		0x0608U,0x695fU,0x3e42U,0xdf1fU,0x6062U,0xcf3fU,0xf258U,0x7a17U,0xef3bU,0xd0efU,
		0x5cefU,0x30deU,0xc290U,0x6c77U,0x60d9U,0xbc74U,0x3147U,0x7130U,0x1216U,0xeaf3U,
		0xa4edU,0x7f98U,0xa730U,0xaac5U,0x6cdeU,0x1149U,0x7ffdU,0x56adU,0x417dU,0x4d32U,
		0x09c1U,0xfec1U,0x2dfeU,0x1b40U,0x557eU,0x7a4eU,0x058bU,0x0c43U,0x83f8U,0x5d40U,
		0x0a3dU,0x38deU,0xfbc5U,0x7be1U,0xbce9U,0x1986U,0x66cbU,0xf283U,0x0775U,0xa6e5U,
		0xc141U,0xc87bU,0x6c3bU,0x8bafU,0x5b81U,0x0912U,0x7e55U,0x4e7aU,0xc506U,0x04a4U,
		0xe88aU,0x53baU,0x3bcbU,0x866bU,0xd30cU,0x4409U,0x50abU,0xa803U,0xcc78U,0x1694U,
		0x3f94U,0x968aU,0xc447U,0x3cecU,0xe962U,0x30f9U,0x701dU,0xa865U,0x2cb2U,0x7218U,
		0x4ef8U,0x0b01U,0xbeeeU,0x3125U,0x57ebU,0x0140U,0xaa6eU,0x4e71U,0xd14fU,0x9452U,
		0x13d6U,0x602cU,0x202dU,0x7802U,0xbd9aU,0xa300U,0xb4c4U,0x6f35U,0xcdcdU,0xc9c9U,
		0xb31fU,0x55a7U,0xcb6cU,0x9336U,0x8183U,0x5a3dU,0xe80fU,0x568aU,0x181aU,0x3916U,
		0x6f85U,0x7df8U,0x29c1U,0xaf62U,0x5fb8U,0x3168U,0x4476U,0x40dfU,0x630dU,0x6c53U,
		0x0202U,0x1b1bU,0x947cU,0x4820U,0x7585U,0xf54aU,0x1af5U,0x404dU,0xb1f4U,0xe87aU,
		0x5b07U,0x5279U,0x0d55U,0x87d1U,0x8127U,0x9206U,0x9bc9U,0xf7a0U,0xeaafU,0xd39aU,
		0xa534U,0x92e0U,0x9dd7U,0x3061U,0xa524U,0xb63eU,0x3f16U,0x20c1U,0x1fd8U,0x29a6U,
		0xfb94U,0x9705U,0x15eaU,0x614bU,0xe973U,0x04e3U,0x7106U,0x636eU,0xb8adU,0x455fU,
		0x07c9U,0x88a7U,0xde4bU,0x3620U,0xb33eU,0x1d76U,0x2828U,0xe2e2U,0x1141U,0x3672U,
		0x8aa3U,0x2dc3U,0xa43bU,0x069fU,0x445cU,0x01b7U,0xdf8fU,0x269cU,0x39a9U,0xf447U,
		0xdfc2U,0x5890U,0x5d52U,0xf580U,0x794bU,0xa0c1U,0x91d6U,0x1de8U,0x5252U,0x6565U,
		0x632eU,0x3396U,0xb45cU,0x1e79U,0x8ca8U,0x8feeU,0xba1eU,0x2c7dU,0xad8fU,0x2949U,
		0x1e1aU,0xe226U,0xe2a5U,0xfdccU,0x5ffcU,0x8657U,0x3d78U,0x1c49U,0xc6c2U,0xd435U,
		0xd1bbU,0xb885U,0x6dd7U,0x0bc1U,0xe3aaU,0xfd18U,0x62f4U,0xb811U,0x2d7bU,0x5c59U,
		0xfaabU,0xb4e6U,0xfa5cU,0xc138U,0x71cdU,0x4a2cU,0x17bbU,0xefe3U,0x28bdU,0x65aeU,
		0xce4cU,0xd5f7U,0xe7d1U,0xe8aaU,0xb0d3U,0xaeabU,0xb141U,0x0129U,0x7ef1U,0x5ef8U,
		0x056bU,0x3c04U,0xa0d5U,0xa92bU,0x69b9U,0x41ddU,0x06d0U,0x8b7fU,0x51a4U,0x9019U,
		0x3518U,0xb199U,0x4cceU,0xf7d5U,0x142fU,0x5940U,0x130dU,0x9b1eU,0x7aaeU,0xef1dU,
		0xde9cU,0x4662U,0x0903U,0xbca1U,0xa1d9U,0xd2b8U,0x6778U,0xc53eU,0x228cU,0xe25aU,
		0x8716U,0x48c9U,0xed08U,0x9c37U,0xf8abU,0xfa1fU,0xee0fU,0x2cc5U,0x291fU,0x7de9U,
		0x71b6U,0x367aU,0x8f60U,0x1a59U,0x9ceeU,0x4dc5U,0xf9c6U,0x343dU,0xa4aeU,0x8356U,
		0x6edfU,0x26f5U,0x91f7U,0x694eU,0x9f9fU,0xa5a5U,0x7837U,0xd819U,0xc34eU,0xbaaeU,
		0x4770U,0x70a5U,0x8a56U,0x0fe8U,0xf010U,0x15acU,0xfc0cU,0x0453U,0x62afU,0xc129U,
		0x5db1U,0x539fU,0x95a8U,0x73ffU,0x9f47U,0xa2abU,0xb3b3U,0x6a6aU,0x29feU,0xaa9eU,
		0x1479U,0x3eacU,0xf78aU,0xef8bU,0x52ddU,0x1b77U,0x2319U,0xf27aU,0x79d2U,0x880cU,
		0xcd95U,0x926bU,0x289fU,0x15fcU,0x5c8aU,0xea0dU,0xec9fU,0x3946U,0xc802U,0x48baU,
		0xd2b1U,0x8837U,0xd0e8U,0xf799U,0xb474U,0xde64U,0xd6d0U,0x20dbU,0x3630U,0x5ebfU,
		0xa2d8U,0xcc62U,0x5bfbU,0x62c0U,0xaafbU,0x9d42U,0x7a8aU,0xc57fU,0x007cU,0x3a84U,
		0x0b15U,0x0dbcU,0x07f0U,0xf93dU,0x9ae0U,0x8afeU,0xe5a6U,0x7507U,0xb226U,0xed9eU,
		0xa6b3U,0xae05U,0xe17bU,0xc5fbU,0x5098U,0xb50eU,0xfa09U,0x22e6U,0x4679U,0xe515U,
		0x7c00U,0x843aU,0xc8f7U,0x82c1U,0x79e4U,0xd172U,0xa66fU,0x0c97U,0xae87U,0x4b1fU,
		0xdb75U,0x42f0U,0x24b5U,0x883eU,0x8bd1U,0xe046U,0x0c6cU,0x56ddU,0x94ebU,0xafa3U,
		0x3fbaU,0x32a2U,0x71feU,0xb2f3U,0x14b5U,0x032fU,0x9794U,0x0525U,0x8160U,0x2794U,
		0xb5ecU,0x7165U,0x4712U,0x0f5eU,0x304eU,0x3e97U,0x8783U,0xd224U,0x0e12U,0xa3f8U,
		0xade5U,0x0c57U,0x807dU,0x7f34U,0x60b0U,0xf23dU,0xdd41U,0xb969U,0x5013U,0x46bdU,
		0x1f7eU,0x5be1U,0xf85eU,0xf17eU,0x9ea2U,0xf15fU,0x911dU,0xd037U,0xecc7U,0xd569U,
		0x120eU,0xf8a3U,0xf694U,0x5acdU,0x63f3U,0xdc99U,0x0b35U,0xedadU,0xfca9U,0x4dadU,
		0x8a0cU,0xb017U,0xe791U,0x45cbU,0x9609U,0xca2aU,0xd114U,0xe6c3U,0x7795U,0x1894U,
		0x0037U,0x0874U,0x9e50U,0x0582U,0x0f7cU,0xa71bU,0xb395U,0xa411U,0x99beU,0x325cU,
		0x2c21U,0x0d8fU,0x283bU,0x4cb5U,0x45f1U,0x98b5U,0x2076U,0x25dbU,0xc595U,0xd7b2U,
		0x0ee9U,0xa7c1U,0x8282U,0x2e2eU,0x9729U,0xdb3cU,0x1571U,0x5c0aU,0xd875U,0xfaccU,
		0xb701U,0x5c44U,0x39b9U,0x8bf6U,0x4ba3U,0x6a1cU,0x9058U,0xc2dfU,0xb98cU,0x58ceU,
		0x769dU,0x3d1fU,0xc178U,0x480eU,0x2525U,0x5656U,0x603dU,0x5497U,0x5a1eU,0xc855U,
		0x271cU,0x9e03U,0xd2d4U,0x3b55U,0x7b9cU,0x1c96U,0xcfd1U,0x360bU,0xa246U,0x3a12U,
		0x937cU,0xd6d7U,0xe69aU,0x898aU,0xb771U,0x2dd5U,0x4750U,0x13a8U,0xa9d1U,0x4214U,
		0x0b3dU,0x9f67U,0xaedaU,0x6d56U,0xc62cU,0x2d8aU,0x822bU,0x4464U,0x237aU,0x4d93U,
		0x7e08U,0x52c2U,0x2c96U,0x60f1U,0x1d44U,0xc813U,0x51e8U,0xc6f1U,0x0040U,0x6eb0U,
		0x7ed7U,0xe071U,0x4ea3U,0xa280U,0xedf6U,0x6f52U,0xd57fU,0x59c0U,0xe652U,0x52e6U,
		0x5a89U,0x3ff7U,0xfab4U,0xbc38U,0x284eU,0x4e28U,0xede3U,0xa85cU,0xd704U,0x57f4U,
		0xe48aU,0x2335U,0xd35bU,0x6162U,0x3768U,0x6f70U,0x2ea2U,0x6225U,0xf5c3U,0xf188U,
		0x41feU,0xf4b2U,0x10d6U,0x7ae9U,0x3d29U,0xe987U,0x7925U,0x7688U,0x2ac1U,0x2857U,
		0x2a84U,0xfddbU,0x048bU,0x40c3U,0xa08dU,0x2870U,0xadc2U,0x7409U,0xa995U,0x3701U,
		0xe43fU,0xda63U,0x48e1U,0xce1fU,0xe0a2U,0x66e8U,0x8444U,0x4f30U,0x499dU,0x8f0eU,
		0xa035U,0x212dU,0x65b8U,0x7154U,0x372dU,0x3ba8U,0xe4e6U,0xda2fU,0x2d06U,0xf70bU,
		0xe6e8U,0xc655U,0x6b58U,0x88b1U,0xb62bU,0x73eaU,0x3920U,0xe305U,0xb188U,0x586bU,
		0x0541U,0x9f0aU,0x18b6U,0x38e8U,0x639dU,0x9666U,0x2fdeU,0x0bdfU,0x6974U,0x8d43U,
		0x7e39U,0xe0c2U,0x7997U,0x82a6U,0x7f05U,0x27caU,0xcd50U,0x772aU,0x9af7U,0x20e2U,
		0xb562U,0x8a40U,0x27d4U,0x5f24U,0x671bU,0xa367U,0x2aa1U,0xe991U,0xd1d6U,0x81e9U,
		0xe7aaU,0xdef1U,0x2e18U,0xd900U,0x201aU,0x4acbU,0x2ddfU,0x1f5aU,0x7da7U,0xf2bdU,
		0x3a34U,0xe67dU,0xddc9U,0x28d9U,0xce83U,0x44e9U,0x93e4U,0xbb94U,0xaf91U,0x51b3U,
		0xb757U,0x7038U,0xacdfU,0xa972U,0x1b05U,0x9d08U,0x16eaU,0x89d3U,0x72a9U,0xdfacU,
		0x5459U,0x04b9U,0xd50eU,0x2547U,0xf079U,0xd408U,0xac8fU,0x6525U,0x53cbU,0x8dbfU,
		0x10cdU,0xe121U,0xb29eU,0x3f8bU,0xe527U,0x7720U,0xa7a1U,0x4f8bU,0x4943U,0x7b2eU,
		0xad77U,0x595dU,0x7a46U,0x26edU,0x5cbbU,0xf653U,0xf480U,0x35e3U,0xc6e9U,0xdf03U,
		0x2fc1U,0x3080U,0x0fb5U,0x65d6U,0xe55dU,0x2506U,0xe603U,0x9498U,0x2be1U,0x4bfeU,
		0x7ceeU,0x1ca7U,0xda79U,0x10fbU,0xc198U,0xb241U,0x14eaU,0xdd57U,0xfc48U,0x6c0dU,
		0x4bbeU,0x137bU,0x3b72U,0xc1abU,0xeaa9U,0xe29fU,0x5ce9U,0xe95cU,0x8bcbU,0x778fU,
		0x6772U,0xcd8fU,0xb55cU,0x5cb5U,0xe694U,0x791fU,0xf601U,0xd1fcU,0x12f6U,0xe912U,
		0xbb6dU,0xf023U,0x02eaU,0xc41fU,0x16d2U,0x064cU,0x0eaaU,0xe8b9U,0xac7bU,0xad15U,
		0x010aU,0xd020U,0x8f30U,0x109bU,0xee9bU,0xfb35U,0x2605U,0xb47bU,0x12fcU,0x0308U,
		0x690dU,0xec4eU,0x84afU,0xf92fU,0xb5cbU,0x3f65U,0xedb5U,0xa321U,0x2d3eU,0x869eU,
		0xb9b8U,0x755eU,0xab69U,0x01bfU,0x0ba7U,0xb4b3U,0xad05U,0xf06fU,0xf26cU,0xb034U,
		0x84d3U,0xfbf1U,0xde2fU,0xdf0bU,0xce16U,0xc78bU,0x1922U,0x6023U,0xd9aaU,0x722dU,
		0xaf33U,0x3035U,0xc8d9U,0x0445U,0x47abU,0x3070U,0x3530U,0x33afU,0xbe33U,0x4ffbU,
		0xee08U,0x14aeU,0xa4e6U,0x1fd1U,0x2d37U,0xa83bU,0xaed3U,0xe7a3U,0x342fU,0x8ee1U,
		0x8efeU,0xe14bU,0x491bU,0xef9cU,0x72ecU,0x909cU,0x6fa5U,0x7b8aU,0x2143U,0x9758U,
		0x95f6U,0x6080U,0xe918U,0x98a7U,0x4be5U,0xc10aU,0x041dU,0xc1ffU,0x5904U,0x2683U,
		0x6ad5U,0xffd7U,0x18e0U,0x540fU,0x7253U,0x3d64U,0x6e8dU,0xe117U,0x4800U,0x389cU,
		0xfe92U,0xba49U,0xe843U,0x02c9U,0xa628U,0x749dU,0xd77eU,0x71e0U,0x3b27U,0x6d1cU,
		0x43deU,0x0a0eU,0x79aeU,0x0c89U,0x7ad2U,0xd27aU,0x2e91U,0x612fU,0xb6ccU,0x3369U,
		0xd7e1U,0xe1d7U,0x95d9U,0xf0c5U,0x5857U,0x54afU,0xf778U,0xb8bcU,0x5bd3U,0x6261U,
		0xf3dfU,0x3fd0U,0xa875U,0x84c2U,0x892cU,0x5e78U,0x0de9U,0x9394U,0x8370U,0x9b9cU,
		0x423dU,0xf8c8U,0x22c1U,0x30d0U,0x8883U,0xde6bU,0xf607U,0x698aU,0x3c28U,0x3787U,
		0xdc2aU,0x4097U,0x184cU,0x75e7U,0x5b47U,0xa607U,0xa842U,0xddbfU,0x5d32U,0x4debU,
		0xf398U,0x43a7U,0x3aa7U,0xa208U,0xa21dU,0x495eU,0x1ac5U,0x06aeU,0xc52bU,0x266eU,
		0xa3b8U,0x96daU,0x4fd0U,0x3117U,0x48fcU,0x0d6cU,0xc38bU,0xab4fU,0x9146U,0x6640U,
		0x0482U,0xc947U,0xef4eU,0x4eefU,0x8fbfU,0xec5eU,0x280aU,0xea34U,0x403fU,0x3f40U,
		0x56f3U,0x4b4cU,0x4a76U,0x4d4cU,0x8636U,0x0b4eU,0x6dbbU,0x23f0U,0x5744U,0xd29dU,
		0xcefbU,0x3bf4U,0x1ba6U,0x1ddaU,0xb493U,0xbb48U,0x0670U,0x5261U,0xed42U,0xa0afU,
		0xd202U,0x2780U,0x203dU,0x78cdU,0xcc28U,0x9417U,0x43f9U,0xa968U,0x96f7U,0x5fccU,
		0xf500U,0xf63eU,0x2625U,0x80f9U,0x0a35U,0x782fU,0xbcfaU,0x8584U,0x05e7U,0xb8d3U,
		0x4e9bU,0xd9a2U,0x025cU,0xefd5U,0x46dfU,0x3292U,0x40f4U,0x74c8U,0x8facU,0x2565U,
		0xde1bU,0x1ffbU,0x119aU,0xabf9U,0x4fb3U,0xa109U,0x26c9U,0xa231U,0x1c3bU,0x1a33U,
		0x8346U,0x5edcU,0x31a2U,0xc926U,0xa4ffU,0xe33aU,0xabdeU,0x9abeU,0x9aacU,0x00a4U,
		0x83ceU,0xe944U,0xe411U,0x5221U,0x0802U,0xcce5U,0x99efU,0x7d60U,0xd60fU,0x219bU,
		0x413bU,0xc244U,0x1077U,0xa2dfU,0x02f0U,0xff45U,0x4f22U,0x653eU,0x515aU,0x7f17U,
		0xbbf8U,0xd28aU,0x63fbU,0x079eU,0x8f05U,0xb52cU,0xd718U,0x6f2cU,0xe475U,0xa443U,
		0x8e6aU,0xc824U,0xf141U,0xe2f7U,0xb050U,0x341bU,0x8325U,0x745cU,0xe8fdU,0xf488U,
		0x087dU,0x5920U,0x9282U,0xca4bU,0x05efU,0x46a1U,0x2c35U,0xcf84U,0xb373U,0xa797U,
		0x3a04U,0x1b80U,0x9d2aU,0x357eU,0xb4efU,0x57beU,0xd07bU,0xe4cdU,0xcab7U,0x5f7eU,
		0x80cbU,0x0149U,0xb40dU,0x355aU,0x42dcU,0xdde8U,0xb269U,0x6f44U,0xa6feU,0x1589U,
		0xbfdcU,0x5e22U,0x11d3U,0x8b24U,0xe12aU,0x3c9aU,0xcd81U,0x9b54U,0xc425U,0xa1a6U,
		0x9964U,0x4e31U,0x022eU,0x973fU,0xc950U,0x2ee4U,0x0fcfU,0xcc44U,0x89abU,0x64d4U,
		0x611aU,0xde89U,0xf593U,0x4f7dU,0x211bU,0xc794U,0xaca6U,0x85b9U,0x15d0U,0xbbd2U,
		0xd48eU,0xb535U,0x1793U,0xd54fU,0x2b94U,0x6095U,0x9778U,0x29d0U,0x93b0U,0x26e3U,
		0xd3faU,0x1a1fU,0xc516U,0x1663U,0x8adaU,0x3ef5U,0xfe3dU,0x6e4dU,0x2da0U,0x76d7U,
		0x9628U,0x681cU,0xebffU,0x6ab5U,0xf6bdU,0x87afU,0xfed2U,0x1f4cU,0x0391U,0x0198U,
		0x3ddaU,0x0293U,0x042eU,0x1617U,0x566cU,0xf6f7U,0xe13dU,0xaa1eU,0x8d6dU,0x4a8fU,
		0x2f98U,0xcc69U,0xbb73U,0xec71U,0x32ffU,0xe0f4U,0x0d31U,0xa812U,0x1bffU,0x924eU,
		0xab95U,0x6345U,0xe152U,0xa313U,0x7ccfU,0x03ebU,0x416eU,0xc676U,0xbdd7U,0x4a00U,
		0x2cfdU,0xd3f1U,0xdaa0U,0xe806U,0xc6aeU,0xf3efU,0xdbfcU,0xa1b8U,0x1cf5U,0xe553U,
		0x0a00U,0xbf5aU,0xf0daU,0xa16aU,0x0863U,0x69aaU,0x12c4U,0xa8b3U,0x9b41U,0x0973U,
		0x19f4U,0x3467U,0x3cb9U,0x3877U,0x9a7cU,0x6d6eU,0x44a3U,0x9899U,0x8cccU,0xbdb9U,
		0xa42cU,0x068eU,0xe1bcU,0x6847U,0x4f68U,0xe6b0U,0x4089U,0xca9eU,0x18f2U,0x5434U,
		0x7005U,0x45aaU,0xbd62U,0xb90dU,0x13f0U,0xe704U,0xf029U,0x8b4bU,0x5f13U,0x91e8U,
		0x7d14U,0xcb46U,0x6383U,0x96ccU,0xf346U,0x9c8cU,0x4d1eU,0x9570U,0x62e2U,0xbefdU,
		0x5040U,0x00ebU,0x461bU,0x2ad5U,0x7bf1U,0xa0f8U,0x6a4dU,0xc99cU,0x802fU,0xb237U,
		0x80ebU,0x432dU,0x1da1U,0xf386U,0xe628U,0x0e26U,0x8ba7U,0xc121U,0xd245U,0x1640U,
		0xe4c5U,0x53f5U,0x4e20U,0x14a6U,0x2d56U,0x8539U,0xc70dU,0x0096U,0x1b60U,0x69e0U,
		0xebe1U,0xaf7dU,0x0ecdU,0x95c6U,0xa5dbU,0xb2d8U,0xf106U,0x3a56U,0x0028U,0x6afdU,
		0xfe83U,0x9041U,0x8d96U,0x28faU,0x9f01U,0x6aacU,0x7f86U,0x6c4aU,0xa56bU,0x3637U,
		0x0eb0U,0x0405U,0x6ceeU,0x3869U,0x2d6bU,0xafb6U,0x4345U,0x8f21U,0xdca1U,0x0a92U,
		0x3436U,0x3e90U,0xfebaU,0x4558U,0xb02dU,0x06a7U,0x60eaU,0xe676U,0xe7f8U,0x2b52U,
		0x3414U,0xe7ffU,0xcad4U,0x809bU,0x4fa4U,0x35ebU,0xf808U,0x5a9fU,0xfe06U,0xfc6eU,
		0xb8dbU,0xea96U,0xc7b0U,0xeb0bU,0xf684U,0x6b0eU,0x7c56U,0x1632U,0x8907U,0x9f87U,
		0xfb4aU,0x4030U,0xfc2fU,0x53fdU,0xcf51U,0x553eU,0xc0b0U,0xac0bU,0xdaf6U,0x90d1U,
		0x9fddU,0x6768U,0xb8ffU,0xd4d5U,0xd8dbU,0x93f8U,0xca69U,0xe37eU,0xd34cU,0x0d14U,
		0x5fe6U,0x1e9cU,0x1013U,0x7d52U,0xe8bfU,0x5e00U,0xbcc3U,0x7286U,0x78b4U,0x12adU,
		0x168fU,0x55a2U,0xcf3aU,0x0c3eU,0x2c78U,0x65a9U,0x87ccU,0x1d71U,0x1b72U,0xc57aU,
		0x465dU,0x6949U,0x45c6U,0xdb37U,0x4b1aU,0x0c52U,0xf4adU,0x714fU,0x3606U,0x6eabU,
		0x7b29U,0x1376U,0x27c5U,0xbb8fU,0x7b85U,0x6d17U,0x1096U,0xc786U,0xbb43U,0x74c3U,
		0xf8c3U,0x3112U,0x8b1fU,0x6090U,0x2196U,0x591bU,0x45a5U,0xf381U,0xaa19U,0xe54eU,
		0x0d0fU,0x1371U,0x6aa7U,0x5a9aU,0xd0d2U,0xcda9U,0xaf10U,0x0ac4U,0xfb1aU,0x62d1U,
		0x34e5U,0xbf71U,0xeb2eU,0xaf15U,0x464dU,0xb89aU,0xf964U,0x5ec6U,0x16d8U,0x5383U,
		0x2fd5U,0x0388U,0x38b7U,0x757cU,0x5356U,0xdcc2U,0xfc82U,0x0e5cU,0x8ee8U,0xf81aU,
		0x3401U,0x08a7U,0x9da0U,0x2261U,0xdd6fU,0x8ac8U,0xf973U,0xe022U,0xb72cU,0x2b59U,
		0x0cefU,0xd8a7U,0xeb9dU,0xc61bU,0xb3fdU,0x0ac9U,0x738aU,0x5375U,0x1bbfU,0x81bfU,
		0xf675U,0x88edU,0x885dU,0x20a4U,0x93ceU,0x6440U,0x2a12U,0x382eU,0xf767U,0xda19U,
		0x8614U,0xbb63U,0x5367U,0xa73fU,0xbd41U,0x4b91U,0x0e43U,0x4596U,0xd0ffU,0xc55fU,
		0xd83eU,0x8254U,0xe1a3U,0x6e9dU,0x3a97U,0xf03eU,0x2270U,0xd0d7U,0xbbabU,0xdb21U,
		0xbed9U,0x322eU,0x0652U,0x64acU,0xf0f7U,0xfaeaU,0x597eU,0x1063U,0x129fU,0xb6b2U,
		0x1d22U,0xaf23U,0x5b70U,0xc5abU,0x1ed6U,0x30a8U,0x2af9U,0x6b30U,0x5340U,0x1776U,
		0x20f5U,0x1ce6U,0xff24U,0x0274U,0x02daU,0x776dU,0x6591U,0x90c3U,0xd9f3U,0xcdaeU,
		0xe3d0U,0x9350U,0x59d6U,0x31f6U,0x08b7U,0xc4d9U,0x620bU,0xd967U,0x9955U,0x7a26U,
		0xb58dU,0x7915U,0x454aU,0xac3fU,0x2347U,0x55dfU,0x9925U,0x5063U,0x3bc3U,0x29f6U,
		0x093bU,0x326eU,0xb261U,0xb8f7U,0x3760U,0x8ef6U,0xc1c7U,0x5badU,0xbad2U,0x7b9eU,
		0x3c6eU,0x5671U,0xcbe1U,0x1c98U,0x1cc8U,0x6ee6U,0x7ba2U,0xd2d6U,0xe5edU,0x6743U,
		0x35d8U,0x3b57U,0x126dU,0x4d7dU,0xa677U,0xa248U,0x72f3U,0xdab2U,0x8297U,0x3a14U,
		0x1c7dU,0x3f1eU,0xbad6U,0xcfd3U,0xdcf0U,0x3ccdU,0xc2caU,0x360dU,0x1741U,0x2753U,
		0x7541U,0xe69cU,0xb39eU,0xf6fcU,0xbf39U,0x898cU,0x46e7U,0x8ea3U,0x5bb1U,0x937eU,
		0xad6cU,0xa8f5U,0xfd3dU,0xd6d9U,0xd545U,0xc9bfU,0x8cb1U,0x4752U,0xbc29U,0xf34eU,
		0x4bcfU,0x13aaU,0xb5e4U,0xe96bU,0xc1cbU,0xb773U,0xdd67U,0xbdcfU,0x3a9fU,0x2dd7U,
		0xd3d9U,0xf86eU,0xf6a7U,0x0b3fU,0x6ca7U,0x0433U,0xce8cU,0x9f69U,0x6047U,0xe358U,
		0x1c9cU,0xa9d3U,0xfc26U,0xedffU,0x62a6U,0x4216U,0x2813U,0x97cdU,0x3283U,0xc62eU,
		0xd955U,0x5dbbU,0xd97eU,0x2d8cU,0x7c36U,0x8218U,0xcbe5U,0xaedcU,0xadd5U,0xab37U,
		0xa6bdU,0x6d58U,0xcf95U,0x02f9U,0xc8c0U,0x237cU,0xb0ddU,0x161cU,0xdd97U,0x4d95U,
		0x2fb6U,0x4194U,0x5675U,0x822dU,0x7b50U,0x7480U,0xb786U,0x4466U,0xffc8U,0x083dU,
		0xcdd7U,0x2c98U,0x93a6U,0x8179U,0x3904U,0x60f3U,0x657bU,0x8402U,0x3c72U,0x7e0aU,
		0xfc8bU,0x4cf9U,0xba0eU,0x52c4U,0x1eb4U,0xadb2U,0x6bccU,0x51eaU,0xb92bU,0x8d72U,
		0xdd4cU,0xc6f3U,0xffedU,0x26fcU,0xd2daU,0x1d46U,0xb56cU,0x24c0U,0x124bU,0xc815U,
		0x4baeU,0x10e0U,0x8c2dU,0x7ed9U,0x90a3U,0x17e5U,0x217dU,0xe073U,0x4850U,0x9cbaU,
		0x7ba6U,0x0042U,0x3978U,0xee76U,0x5e13U,0x6eb2U,0x3ab1U,0x1318U,0x839aU,0xedf8U,
		0x5cdeU,0x4a94U,0x2391U,0x6f54U,0xd192U,0xcd38U,0x6eeaU,0x4ea5U,0xf6dcU,0x699bU,
		0xe581U,0xa282U,0xfd9cU,0x24f3U,0x422dU,0xe654U,0xf74fU,0x07dfU,0xedbfU,0x52e8U,
		0x6812U,0x40d5U,0x1cccU,0xd581U,0x0087U,0x725dU,0x9dc0U,0x59c2U,0x9b82U,0xc4adU,
		0xe907U,0xfab6U,0xd4ccU,0xe142U,0xaa95U,0xbc3aU,0x143aU,0x9782U,0x3b5bU,0x5a8bU,
		0xd048U,0x83a1U,0x33c3U,0x3ff9U,0xbd76U,0x2bf0U,0x4367U,0xede5U,0x086dU,0x15f5U,
		0x8650U,0xa85eU,0x78c6U,0xc103U,0x35dcU,0x2850U,0xfae3U,0x727fU,0x91e1U,0x4e2aU,
		0x16c7U,0xfa85U,0x6cc0U,0xe48cU,0x6318U,0xaa23U,0x9942U,0x2337U,0xcfb5U,0x649dU,
		0x6747U,0xd706U,0x85a0U,0xcd6aU,0xfb54U,0x57f6U,0x1d0bU,0xd234U,0xc14cU,0x376aU,
		0x5125U,0xbe7bU,0xd33cU,0x6f72U,0x0e4dU,0x3152U,0xe5f1U,0xd35dU,0x3dd0U,0x1aebU,
		0x5d6aU,0x6164U,0xb41eU,0xb2adU,0x23c1U,0xf5c5U,0x4a13U,0xbb78U,0x6bb6U,0xf18aU,
		0xb83aU,0xa4b5U,0xa24cU,0x2ea4U,0x878fU,0x7e79U,0x7cf9U,0x6227U,0x7529U,0xb2ffU,
		0x0c9fU,0x10d8U,0x4922U,0x5b1aU,0x86ccU,0x7aebU,0x3a3fU,0x20c9U,0xa67bU,0x4200U,
		0xd479U,0xc750U,0xdee6U,0xf4b4U,0x9735U,0x720eU,0xa098U,0x7927U,0x0822U,0xec76U,
		0xa89eU,0x768aU,0x329aU,0xb636U,0x4d81U,0x3d2bU,0xdcbaU,0x9246U,0x75dfU,0xe989U,
		0xcb1dU,0xe1f2U,0x1169U,0x2a86U,0xc637U,0x9c94U,0x9615U,0xfdddU,0x8a8fU,0xf5d8U,
		0x1271U,0x2ac3U,0x3a29U,0x0d9bU,0xf046U,0x2859U,0xdc35U,0x5028U,0x6784U,0xa08fU,
		0xa05cU,0x2f9dU,0x76b4U,0x2872U,0xaa3bU,0x1a65U,0x3a18U,0x048dU,0x9322U,0x131fU,
		0xa46bU,0x40c5U,0x85eaU,0x6508U,0x5529U,0xa997U,0x7843U,0x8c5fU,0xd20dU,0x3703U,
		0xc895U,0x89e2U,0x829bU,0xadc4U,0x4f9bU,0xde26U,0x6ed6U,0x740bU,0x72c1U,0xc1ecU,
		0xfb6eU,0x48e3U,0x11c9U,0xcc6eU,0xae6fU,0xce21U,0x0926U,0x6601U,0xdab6U,0xe441U,
		0xeea6U,0xea54U,0x5c96U,0xda65U,0x7e9fU,0xa633U,0xd6b6U,0x8446U,0x5f9dU,0x31d3U,
		0xd9cfU,0x4f32U,0x9aecU,0x21b2U,0x72f7U,0xe0a4U,0x65f1U,0x1940U,0x9e37U,0x66eaU,
		0xc6d4U,0x25b9U,0xbe0dU,0xa037U,0xcac2U,0x0d36U,0xe3b6U,0x212fU,0x73b5U,0x835dU,
		0xcfd7U,0x499fU,0x97a9U,0x2955U,0xb008U,0x8f10U,0x69c5U,0x0ca7U,0x2dabU,0x372fU,
		0xefa3U,0x024aU,0x651aU,0x3baaU,0xf900U,0xc95bU,0xbadaU,0x65baU,0xf213U,0xf724U,
		0x6e3cU,0x7156U,0x5e98U,0x5285U,0x074aU,0x2d08U,0xc841U,0xa817U,0x07b7U,0xf70dU,
		0x4f73U,0x9448U,0x3f22U,0xe4e8U,0xadf6U,0xa957U,0xcd01U,0xda31U,0xd736U,0x7db6U,
		0xece6U,0x6b5aU,0x544eU,0x504bU,0xcb09U,0x88b3U,0xeecaU,0x645fU,0x1c81U,0xe6eaU,
		0xdf9bU,0xba9fU,0xa5e6U,0xc657U,0x9a5cU,0xd8e6U,0x8dd9U,0x3922U,0xa76cU,0x3304U,
		0x4250U,0xe307U,0x7591U,0x76eaU,0x3611U,0xb62dU,0xb5a4U,0x3ab8U,0x4392U,0x73ecU,
		0xeffcU,0x50c3U,0x89c8U,0x0543U,0xea8dU,0x7224U,0xd810U,0x9f0cU,0xe822U,0xea64U,
		0xc2ceU,0xb18aU,0xfe11U,0x3652U,0xbda6U,0x586dU,0xb7ccU,0x98d6U,0x921cU,0x639fU,
		0x7e61U,0xe0f9U,0x08c2U,0x9668U,0xeae4U,0x5f33U,0x3cd1U,0x18b8U,0x0fc5U,0xbee4U,
		0x746eU,0x38eaU,0x93d9U,0xa90aU,0x8009U,0x6976U,0x18cbU,0x2475U,0x450bU,0x8d45U,
		0xd696U,0x30eaU,0xdcf4U,0x2fe0U,0x4a5cU,0x8dedU,0x8514U,0x0be1U,0xc7ceU,0x6c19U,
		0x7856U,0x7999U,0xddabU,0xab9aU,0x9f35U,0x82a8U,0x6179U,0x9fa6U,0xe6a0U,0x7e3bU,
		0x0536U,0xacf4U,0x912aU,0xe0c4U,0x3550U,0x049aU,0x79ecU,0xcd52U,0x214aU,0x96afU,
		0xcb87U,0x772cU,0x0c64U,0xa52cU,0x7545U,0x7f07U,0x2fcdU,0x32f7U,0x3c20U,0x27ccU,
		0x3498U,0x7427U,0x97a1U,0xb564U,0xa41aU,0x634aU,0xf085U,0x8a42U,0x0264U,0x4490U,
		0x2757U,0x9af9U,0xaa51U,0x51bfU,0x0bd4U,0x20e4U,0xe7b6U,0xdb4cU,0x8ff7U,0x671dU,
		0x42fdU,0x638bU,0x7dcfU,0xa369U,0x5d9eU,0x3102U,0x1745U,0x27d6U,0xc0eaU,0x660eU,
		0xe1abU,0x5f26U,0xa7ceU,0x75cbU,0x6f14U,0xd1d8U,0xe44bU,0x1c04U,0x44f2U,0x81ebU,
		0x4992U,0x6234U,0x8990U,0x2aa3U,0xd51eU,0x6f5bU,0x8d33U,0xe993U,0xea3dU,0xdf4dU,
		0x63a9U,0x2e1aU,0x9d20U,0x076bU,0x9065U,0xd902U,0xc3c8U,0xf717U,0xbf3dU,0xe7acU,
		0x5c1eU,0xf98cU,0xd25bU,0xdef3U,0xd6ebU,0xff4eU,0xd59cU,0x2de1U,0x3e1cU,0x9253U,
		0x410dU,0x1f5cU,0x7e14U,0xc306U,0xf700U,0x201cU,0xb024U,0xab60U,0x71adU,0x4acdU,
		0x6e79U,0x4289U,0x9e67U,0x3a36U,0x8e3fU,0x52f2U,0x1f90U,0xe67fU,0xf2f7U,0x2eb4U,
		0xb3a2U,0x7da9U,0x0e65U,0x4edfU,0x3774U,0xf2bfU,0x4249U,0x76adU,0xaa8eU,0xce85U,
		0x48d6U,0x7cd4U,0x2644U,0x44ebU,0xcf15U,0xa5acU,0x9382U,0xddcbU,0x742fU,0x98e3U,
		0x5476U,0x28dbU,0x05cbU,0xb307U,0xf9d3U,0xaf93U,0x975fU,0x31c6U,0x4908U,0x51b5U,
		0xa78dU,0x197cU,0x5bb5U,0x93e6U,0x21c3U,0x4b58U,0x50f1U,0xbb96U,0xa2f0U,0x392fU,
		0x1966U,0xace1U,0xed67U,0x03f0U,0x40b8U,0xa974U,0x70b2U,0xf73bU,0x8ea7U,0xb759U,
		0xcf07U,0x0bf3U,0xb86aU,0x703aU,0xc731U,0x5ee8U,0x91b9U,0x16ecU,0x185eU,0xf9a7U,
		0xc098U,0x89d5U,0xc90cU,0x1e67U,0x46ebU,0x1b07U,0xfdb5U,0xed59U,0xb3c0U,0x9d0aU,
		0x2ab6U,0x1f39U,0x88d3U,0x545bU,0xa189U,0xe157U,0x264bU,0x04bbU,0x44d2U,0xa68bU,
		0xd6ddU,0x72abU,0xdffcU,0x2398U,0x47b7U,0xdfaeU,0x5484U,0x9239U,0x4c6aU,0xf07bU,
		0x3edaU,0xa824U,0xa941U,0xd40aU,0x6919U,0xdf3cU,0xfd41U,0xd510U,0x9278U,0x83eeU,
		0x098bU,0x2549U,0xdfeaU,0x8c75U,0x7adeU,0x53cdU,0x0a89U,0xa318U,0xfa16U,0x8dc1U,
		0x5d90U,0x3fdcU,0xa8f9U,0xac91U,0x1394U,0x9ed9U,0x7d39U,0x6527U,0x0dc9U,0xe5dbU,
		0x0890U,0xb2a0U,0x9602U,0x2312U,0xad88U,0x3f8dU,0x0669U,0x213cU,0xad70U,0x10cfU,
		0xbba4U,0x5fdfU,0xbd5bU,0xe123U,0x6f22U,0xaabbU,0xb85bU,0xa7a3U,0x2938U,0xbddcU,
		0x3d06U,0x4f8dU,0x42e8U,0x91feU,0x4756U,0xe529U,0xb892U,0x4a87U,0x30c8U,0x7722U,
		0x94ceU,0x775aU,0xd17aU,0xad79U,0x6facU,0x7f23U,0x7e91U,0x595fU,0x9e70U,0xb646U,
		0x8cb5U,0x4945U,0xe647U,0x54d8U,0xb05cU,0x7b30U,0x3c11U,0xc49bU,0x678dU,0x5cbdU,
		0xf6c9U,0x4a05U,0xff7cU,0xf655U,0xefddU,0x361aU,0xc9c3U,0x7a48U,0x2e28U,0xa3f2U,
		0x9d3cU,0x26efU,0x6c07U,0x44e3U,0x2127U,0xc6ebU,0x328cU,0x96d4U,0x71daU,0xdf05U,
		0x4327U,0xa1b2U,0xd549U,0xf482U,0x90bdU,0xc615U,0x6ea5U,0x35e5U,0x1e8bU,0x9521U,
		0x8134U,0x0fb7U,0xf137U,0x4173U,0x7cdbU,0x65d8U,0x9911U,0x1adeU,0x13aeU,0x2fc3U,
		0xe23cU,0x4a9bU,0x9a25U,0x3082U,0x9a88U,0xca3aU,0x9a4eU,0xe605U,0x9712U,0x7b63U,
		0xdc47U,0x949aU,0xd6f9U,0x641fU,0x4bd3U,0xe55fU,0xee5bU,0x840fU,0x43c6U,0x2508U,
		0xd0afU,0x5440U,0x30bfU,0x7cf0U,0x86feU,0xc67bU,0x0aa7U,0x1ca9U,0x235aU,0xc27eU,
		0xf352U,0x2be3U,0xdb86U,0x03c4U,0xb34bU,0x4c00U,0xf8f2U,0xaee9U,0xe3e9U,0xc19aU,
		0x4297U,0xff6dU,0xe8e2U,0xb243U,0x589dU,0x8c09U,0xbc2dU,0xda7bU,0x2d62U,0x84a6U,
		0x870dU,0x10fdU,0x8c1dU,0x6dedU,0xe8cbU,0xfc4aU,0xcc0eU,0xdaa5U,0x8190U,0x6c0fU,
		0xccdbU,0x6955U,0xb777U,0x14ecU,0x34a0U,0x9c49U,0xaadeU,0xdd59U,0x1b28U,0x7531U,
		0x1d5bU,0x3b74U,0x7b91U,0xe4acU,0x94a9U,0xc1adU,0xb93eU,0xe080U,0xc1cfU,0x4bc0U,
		0x773aU,0x4e5cU,0x125fU,0x137dU,0xc502U,0xfbc1U,0x7ff9U,0x5cebU,0x9478U,0xe80bU,
		0x13d2U,0xe95eU,0x91d2U,0x8a9fU,0xe96fU,0xeaabU,0x3514U,0xb0cfU,0x2d77U,0xe2a1U,
		0x9f43U,0xc34aU,0x8f5cU,0x6774U,0x4675U,0x07ecU,0xd6ccU,0xcd91U,0xf690U,0xdd3dU,
		0xb5e8U,0x8bcdU,0x6039U,0xb6fdU,0x45edU,0x7791U,0xe486U,0xedf2U,0x2376U,0xe696U,
		0x6399U,0x2d02U,0x48ddU,0x7921U,0x53c7U,0xacdbU,0x2ddbU,0xb55eU,0x02e6U,0x5ce5U,
		0x7ceaU,0x5cb7U,0xaecfU,0xaf2fU,0x0ba3U,0x12f8U,0xf3dbU,0x7aceU,0x47fcU,0xe914U,
		0x8632U,0x9142U,0x3aa3U,0xf603U,0x9aa8U,0x1196U,0xbcf6U,0xd1feU,0xb409U,0x2c31U,
		0x8e66U,0x02ecU,0xfeceU,0xd3f6U,0x2117U,0xc421U,0xa428U,0x085fU,0xbdd3U,0xbb6fU,
		0xa5d7U,0xd241U,0x4617U,0xf025U,0xdaf2U,0xc7acU,0x60e6U,0x0eacU,0x6aa3U,0x1092U,
		0x4659U,0xe8bbU,0x0e3fU,0xf671U,0xdd6bU,0x16d4U,0x0937U,0x08b3U,0x533cU,0x064eU,
		0xa0e7U,0xf84fU,0x3b7fU,0x010cU,0x508cU,0xdc01U,0x3996U,0xd022U,0x81d3U,0xc355U,
		0x0b43U,0xac7dU,0xbb3aU,0x2e88U,0x9ad7U,0xad17U,0xd866U,0xc54eU,0xdb7dU,0xee9dU,
		0xe91fU,0x842eU,0x8458U,0xfb37U,0x85f8U,0xfb9cU,0xf6abU,0x8f32U,0xcf6bU,0x1fedU,
		0xafeeU,0x109dU,0xceb5U,0xff97U,0x52b1U,0x12feU,0x0f85U,0xa1bdU,0x5befU,0x030aU,
		0x8fb4U,0xfcf4U,0xf872U,0x2607U,0x0efaU,0x0f21U,0x4950U,0xb47dU,0x334fU,0x3c37U,
		0xb465U,0x84b1U,0x5036U,0x62eaU,0x9a05U,0xf931U,0x2513U,0xc450U,0xd3ddU,0x690fU,
		0xd918U,0xf9f4U,0xbbb3U,0xec50U,0xd6c4U,0xc090U,0xcb01U,0xedb7U,0x5eaeU,0xc6b3U,
		0xddd2U,0xa323U,0x54a1U,0x17a6U,0x9f6dU,0xb5cdU,0x04c6U,0x4eacU,0xc93cU,0x3f67U,
		0x8fc8U,0x393aU,0x4720U,0xb9baU,0x9bbaU,0x53a9U,0x0733U,0x7560U,0x4833U,0xe109U,
		0xce90U,0x2d40U,0x692eU,0x2f41U,0xae0fU,0x86a0U,0xe99eU,0x1080U,0x2893U,0x0ba9U,
		0xaeb5U,0xf3f4U,0x9ee8U,0xb4b5U,0xc328U,0x496dU,0x0437U,0xab6bU,0x48f1U,0x7077U,
		0x7d72U,0x01c1U,0x9c82U,0xb87bU,0xc9d3U,0xf26eU,0xdd07U,0xf124U,0x7e46U,0xb036U,
		0xecbfU,0x5d4aU,0x6cabU,0xad07U,0x535fU,0x449bU,0x56ebU,0xf071U,0x3debU,0xef5dU,
		0x4de0U,0xde31U,0xbb86U,0x0a05U,0x3a7bU,0xdf0dU,0x66f5U,0xbab5U,0xa9d7U,0x84d5U,
		0x8a4aU,0x55bfU,0x099aU,0xfbf3U,0xd855U,0x7af3U,0x47e7U,0x1924U,0x98aeU,0x20d4U,
		0x5c52U,0x6025U,0x857dU,0x3895U,0x1ca0U,0xce18U,0x176fU,0x6942U,0x2aceU,0xc78dU,
		0xa212U,0x8f70U,0x2f5aU,0xaf35U,0x8d50U,0xbf5fU,0x063eU,0x3037U,0x4930U,0x300bU,
		0xe35cU,0xd9acU,0x19d1U,0x5b2fU,0x959cU,0x722fU,0x09e8U,0xbffbU,0x6b65U,0x47adU,
		0x908aU,0x4732U,0xa32bU,0x3072U,0x205bU,0x0cb2U,0x604bU,0xc8dbU,0x7d0aU,0x3d6eU,
		0x56c3U,0x0447U,0x6d48U,0x5f4aU,0xf492U,0xbe35U,0x561aU,0x1cfaU,0x9389U,0x4ffdU,
		0x7aa2U,0x29daU,0x421aU,0x3532U,0x00c9U,0x6ef1U,0xe6d1U,0x33b1U,0x4471U,0x57e6U,
		0x6c36U,0xa4e8U,0x69b4U,0x6dd2U,0xdfbdU,0x1fd3U,0xe176U,0xd2acU,0x62aaU,0xee0aU,
		0x7698U,0x99b9U,0x9118U,0x14b0U,0xb183U,0xadbdU,0x2849U,0xaed5U,0xf5fcU,0xe558U,
		0xd509U,0xe7a5U,0x5852U,0x18dbU,0xee03U,0x2d39U,0x319dU,0xf4fbU,0x403aU,0xa83dU,
		0x9623U,0x89a6U,0xd076U,0x8f00U,0x1b5bU,0x4d19U,0x9a77U,0xe14dU,0x45a0U,0x2c73U,
		0xfc2aU,0x3431U,0x2342U,0x5b6bU,0x860fU,0x8ee3U,0x93a1U,0xd950U,0xbc24U,0x72eeU,
		0x5120U,0x0868U,0xf74aU,0x909eU,0x5f98U,0x783eU,0xc632U,0x491dU,0x18c6U,0xea88U,
		0x5449U,0xef9eU,0x8e3aU,0x9d1bU,0x42f8U,0x2145U,0x95fdU,0x3ed5U,0x1859U,0x975aU,
		0x4292U,0x970dU,0x3287U,0x6fa7U,0x6a9eU,0xf3d6U,0x4670U,0x7b8cU,0xdd02U,0x9bb5U,
		0x5031U,0xe91aU,0x1b56U,0x69afU,0x9085U,0x98a9U,0x256dU,0x76f3U,0x97d1U,0x95f8U,
		0x96e7U,0x9428U,0x7624U,0x6082U,0x7609U,0x964cU,0xee1dU,0x041fU,0x97d6U,0xbf05U,
		0x0f70U,0xc201U,0xde10U,0xe317U,0x2817U,0x4be7U,0x7655U,0xd76eU,0xdb29U,0xc10cU,
		0x68c5U,0xa7b5U,0x2494U,0x6ad7U,0xa55cU,0xf0dfU,0xa5b3U,0xffd9U,0x0a6aU,0x2253U,
		0x2d90U,0x5906U,0xc2f0U,0xcd3fU,0xf50dU,0x2685U,0xb6f5U,0xed51U,0xba97U,0x7255U,
		0x76f8U,0x1680U,0xd796U,0x3d66U,0xbb55U,0x3e69U,0xd982U,0x18e2U,0x8f9bU,0xf0aaU,
		0xc91bU,0x5411U,0x6af3U,0x5936U,0xa50cU,0x4802U,0xdc18U,0xa16fU,0xbd32U,0x389eU,
		0xdc91U,0x655dU,0x5dbfU,0x6e8fU,0xc557U,0x683fU,0x23e8U,0xe119U,0xef6eU,0xd829U,
		0x5a12U,0xe845U,0x2572U,0x78a2U,0x0e96U,0x02cbU,0xe830U,0xeb8dU,0xd959U,0xfe94U,
		0xe406U,0x4784U,0xfdf5U,0xba4bU,0x7f4bU,0x5f55U,0xecadU,0xd780U,0xdd26U,0x9b46U,
		0xfefbU,0x71e2U,0xb748U,0xd820U,0xaee0U,0xa62aU,0xf08dU,0x0323U,0x11f5U,0x749fU,
		0xa5f4U,0x86d4U,0x940cU,0x43e0U,0x6087U,0xe1ceU,0xed95U,0x0a10U,0x4afdU,0xb54dU,
		0xcbe9U,0x3b29U,0x3042U,0x981fU,0x3b13U,0x6d1eU,0xd684U,0x4c98U,0xb656U,0x7ad4U,
		0xf839U,0x0978U,0xab76U,0xd27cU,0xd5b6U,0x220fU,0x821cU,0x79b0U,0x80e1U,0x17b1U,
		0x3ebaU,0x0c8bU,0xb488U,0x4315U,0xe9efU,0xb6ceU,0x7629U,0xc2b1U,0xdddaU,0x336bU,
		0x9304U,0x00b3U,0x7c3aU,0x2e93U,0xda3fU,0xd79eU,0x24a9U,0x6131U,0x7053U,0x95b8U,
		0xb97cU,0x95dbU,0x3376U,0x12c9U,0xcf1cU,0xf0c7U,0xf221U,0xf897U,0x6d5cU,0xd7e3U,
		0xed33U,0xd199U,0x2e4cU,0xe1d9U,0x9562U,0xdbd1U,0x8939U,0xf77aU,0x942dU,0x8d07U,
		0x7824U,0xb8beU,0x9e96U,0xdbdcU,0xa6c1U,0x5859U,0xf4c5U,0x183dU,0xbaf0U,0x54b1U,
		0x3f06U,0x11a9U,0x1e3aU,0xf3e1U,0x2963U,0xa8b8U,0x5d08U,0x3fd2U,0xc117U,0xe8a0U,
		0xab3bU,0x5bd5U,0x2234U,0xb5b6U,0x5bc4U,0x6263U,0xdbbaU,0x556eU,0xe0e6U,0x892eU,
		0x96ecU,0x3e74U,0x9d50U,0x5e7aU,0x49c0U,0x2a3aU,0xadd9U,0xa877U,0xeb6dU,0xa3daU,
		0xfe9fU,0x84c4U,0xc8aeU,0x819aU,0x110dU,0x8372U,0xba6bU,0x3cbeU,0x05fcU,0x9b9eU,
		0xac70U,0x8a5dU,0x2380U,0x0debU,0x2450U,0x0465U,0xb81fU,0x9396U,0x0156U,0x1954U,
		0xae8fU,0x22c3U,0x0424U,0xb1eaU,0x81faU,0x30d2U,0x17f8U,0x1fe0U,0xc8c4U,0x423fU,
		0x13b8U,0x4b38U,0xd4bcU,0xf8caU,0xefc0U,0xa592U,0x1e28U,0xf609U,0xa724U,0x387cU,
		0x4c74U,0x698cU,0x40fcU,0xd746U,0x02fdU,0x8885U,0xd8d2U,0x4912U,0x1470U,0xde6dU,
		0xbf47U,0x68bcU,0x1e82U,0xdc2cU,0xee22U,0x4d26U,0xbe6cU,0x4099U,0xe5c2U,0x3cf5U,
		0xcf99U,0x3c2aU,0xc831U,0xb1d3U,0x2278U,0x3789U,0xacb7U,0x1931U,0xbcddU,0x5b49U,
		0xdac0U,0x19f9U,0x28e2U,0xa609U,0x01a5U,0x4ce2U,0x4d99U,0x184eU,0xb9adU,0xa289U,
		0x9074U,0x75e9U,0xec67U,0xd390U,0x9226U,0x5d34U,0x9651U,0x2efaU,0xb1a7U,0x4dedU,
		0x877bU,0x07c1U,0xdd9bU,0xa844U,0x0ce7U,0xc6a6U,0xc523U,0xddc1U,0xc4faU,0xa2e8U,
		0x5e90U,0x3aa9U,0xbbdfU,0x346cU,0xd67cU,0xa20aU,0x3c7cU,0xf3cbU,0x1620U,0xf39aU,
		0x1a92U,0xc008U,0xf064U,0x43a9U,0x332dU,0xa6d9U,0x023bU,0x1ac7U,0x760eU,0xedc9U,
		0xd5cfU,0x06b0U,0x86eeU,0xfe57U,0xb0e1U,0xa21fU,0xdccbU,0x34bfU,0xc156U,0x4960U,
		0xc8ceU,0x14d6U,0xa651U,0xa3baU,0x9bddU,0x44a8U,0x34d0U,0x96dcU,0x259bU,0x0fefU,
		0x8231U,0xc52dU,0x6352U,0x83dfU,0xd143U,0x2670U,0x2a6bU,0x5b22U,0xa542U,0x48feU,
		0xc206U,0xb20eU,0xa54cU,0x0d6eU,0x997aU,0x5295U,0x5679U,0x4fd2U,0x7cc5U,0xfb8aU,
		0xb7acU,0x3119U,0x3474U,0xc3ebU,0x6018U,0x9148U,0x5df7U,0x989eU,0x9262U,0x6642U,
		0x8e08U,0x86adU,0x4198U,0xc38dU,0x2ca6U,0x03a9U,0x640dU,0xab51U,0x2169U,0x7440U,
		0xb35bU,0xef50U,0x0f75U,0xec98U,0xc6bbU,0x4ef1U,0x43f2U,0x58fdU,0x2fbaU,0x0484U,
		0x1298U,0x1688U,0x637cU,0xc949U,0xeff5U,0x85e3U,0xbd6fU,0x280cU,0xaec8U,0x9a81U,
		0x547dU,0xea36U,0x955bU,0xb6eeU,0x446aU,0x8fc1U,0xa269U,0xe588U,0x6b6fU,0xec60U,
		0x6b76U,0xea1dU,0xeaceU,0x56f5U,0xbf0aU,0x4d0cU,0xfb80U,0x4b4eU,0xd3e7U,0x6364U,
		0xb78aU,0x4041U,0xa4c9U,0xc45dU,0xc4c8U,0x3f42U,0x0b64U,0x9641U,0xb660U,0x8638U,
		0x67dbU,0x6d73U,0x33a4U,0x0b50U,0xcdc1U,0x5d12U,0x7484U,0x4a78U,0x8a1dU,0x7677U,
		0x1daeU,0x4d4eU,0x67cbU,0x8f3fU,0xe460U,0x5746U,0x97dbU,0x59acU,0x2b3bU,0xd29fU,
		0x69edU,0x5a31U,0x7b54U,0x6dbdU,0x3024U,0x35b1U,0x8e4eU,0x23f2U,0x77b8U,0x7c10U,
		0xec15U,0x1ba8U,0x12baU,0xa431U,0x3c52U,0x1ddcU,0xb57aU,0x4777U,0x2c9cU,0xcefdU,
		0x6e0dU,0x4ca5U,0xd0b8U,0x3bf6U,0xbb05U,0x6e64U,0x4b27U,0x0672U,0x4becU,0x9810U,
		0xdf83U,0x5263U,0x6e82U,0x29aeU,0xcddbU,0xb495U,0x0dacU,0x919fU,0xb4d8U,0xbb4aU,
		0x4e42U,0xd78dU,0xdc3eU,0xd204U,0xee84U,0x0693U,0x7893U,0x2782U,0x74d0U,0x759aU,
		0x0841U,0xed44U,0xd75bU,0x3664U,0x79fbU,0xa0b1U,0x99a8U,0xb16bU,0x9e1cU,0xcc2aU,
		0x281cU,0x9578U,0x3176U,0x9419U,0xe75bU,0x7d90U,0xffccU,0x203fU,0x87c3U,0x724aU,
		0xd0dfU,0x78cfU,0xbf50U,0xe2d8U,0xd31aU,0x96f9U,0xafcfU,0x8cd1U,0x98eaU,0x5fceU,
		0xb961U,0x2dbbU,0x60f7U,0x43fbU,0x88e5U,0x69a2U,0x6638U,0xa96aU,0xe58fU,0x5b5cU,
		0x70faU,0x2627U,0xe31cU,0xa8c5U,0x4e4bU,0x80fbU,0xcb9bU,0x8952U,0x3908U,0xf502U,
		0x0412U,0x37bdU,0x763cU,0xf640U,0x7fe5U,0xe86eU,0x841fU,0xbcfcU,0x49d4U,0xbdbeU,
		0xc0bdU,0x8586U,0xe7cbU,0x113bU,0x817dU,0x0a37U,0xd86fU,0x60aaU,0x0b0fU,0x7831U,
		0xeaa3U,0xb751U,0xe4e0U,0x4e9dU,0xde15U,0xc385U,0x79a8U,0xd9a4U,0x8ba1U,0xf0d4U,
		0x93aaU,0x05e9U,0x59d0U,0x1bb9U,0x7b7fU,0xb8d5U,0xcfafU,0xd18cU,0x2fb0U,0x46e1U,
		0xeadeU,0x4f6dU,0x0920U,0x3294U,0x5d8aU,0x70acU,0x7e0eU,0x025eU,0xa422U,0x91ccU,
		0x2354U,0xefd7U,0x584cU,0x492aU,0xc322U,0x8faeU,0xc111U,0xd5b0U,0xdc8bU,0x2567U,
		0xcdbbU,0x8e02U,0x3c76U,0x40f6U,0xc3a0U,0x37aeU,0xe7c5U,0x74caU,0xbbe7U,0x5512U,
		0x1ee3U,0x119cU,0xeb45U,0xe6b5U,0x4427U,0xabfbU,0x37b4U,0xf0a1U,0x8406U,0xde1dU,
		0x2f82U,0x41d1U,0xba31U,0x1ffdU,0xaf46U,0x8b0eU,0x5e6dU,0x26cbU,0xdb2eU,0xba40U,
		0x5eb6U,0xa233U,0x1448U,0x021aU,0x657fU,0x4fb5U,0x1423U,0x7700U,0xfa47U,0xa10bU,
		0xd8b4U,0x4abbU,0xebc5U,0x8348U,0xdba7U,0xe1c1U,0x7436U,0x5edeU,0x345dU,0x2692U,
		0x52c8U,0x1c3dU,0xd36bU,0xf6e3U,0x5173U,0x1a35U,0xa270U,0x4b69U,0xe24cU,0xa501U,
		0xd773U,0xffb5U,0x2a1fU,0xe33cU,0x6da5U,0xb1c4U,0xba12U,0x31a4U,0xa143U,0x2978U,
		0x7fa6U,0xc928U,0x6755U,0x22d8U,0x7ec1U,0x9aaeU,0x3fa9U,0x684cU,0x1aa3U,0x00a6U,
		0xc3a6U,0xb805U,0x4cfdU,0xabe0U,0xbf7eU,0xa11aU,0x6676U,0x9ac0U,0x8829U,0x6552U,
		0x6070U,0xe413U,0x765aU,0x055aU,0x0c7eU,0x5223U,0xd1afU,0xa047U,0xfc8fU,0x83d0U,
		0x22feU,0x732eU,0xbeb3U,0xe946U,0x111bU,0xa587U,0x5adcU,0x99f1U,0xe08cU,0x18f7U,
		0x7451U,0x7d62U,0xf5b7U,0xfc13U,0x51eeU,0x0804U,0x24e1U,0x3541U,0x151eU,0xcce7U,
		0xc071U,0xb2c4U,0x8bd9U,0x413dU,0x6adcU,0xbfbdU,0x9108U,0xc246U,0xed20U,0xe97bU,
		0x6bd0U,0xd611U,0x4983U,0x4277U,0x805aU,0x219dU,0x98cfU,0xc1e5U,0xfa7eU,0x02f2U,
		0x2c2aU,0x5439U,0x8c6eU,0xff47U,0x11a2U,0x592fU,0xadb6U,0x1079U,0x22d1U,0xe867U,
		0x963aU,0xa2e1U,0x8ce9U,0xa7dfU,0xad98U,0x515cU,0x2499U,0x4623U,0x9806U,0x7f19U,
		0x3e86U,0x2cbbU,0x1eb8U,0x4f24U,0x902eU,0xcbb6U,0x065aU,0x6540U,0x87e3U,0x6799U,
		0x5f73U,0x63fdU,0x7573U,0x408eU,0xaf9aU,0x07a0U,0x252fU,0xf8e3U,0xc6f7U,0xbbfaU,
		0x6b4cU,0xe65bU,0x1876U,0xd28cU,0x56fcU,0xfd28U,0xd4acU,0xd71aU,0xa7baU,0x62b9U,
		0x4a25U,0x6f2eU,0xefceU,0x7496U,0xdd50U,0x8f07U,0x7766U,0x7148U,0xb230U,0xb52eU,
		0x42d1U,0x5163U,0x66b6U,0x8e6cU,0xe7dbU,0xcaa3U,0xb24fU,0xc826U,0x1ee9U,0x5784U,
		0x8d76U,0xe477U,0x8727U,0x0ab3U,0x18aaU,0xa445U,0x7b6fU,0xd08aU,0xb4ceU,0xb052U,
		0x68caU,0xdad8U,0x669dU,0x341dU,0x8044U,0x0595U,0xb92fU,0xf143U,0x340aU,0xf9bcU,
		0xa06bU,0xe2f9U,0x236eU,0x88cbU,0x8dd1U,0xe8ffU,0xebbdU,0xbd67U,0xb974U,0xf48aU,
		0x32c7U,0x045aU,0x1d4aU,0x8327U,0x9b01U,0x8849U,0x6533U,0x745eU,0x6a27U,0x4208U,
		0x5aedU,0x9284U,0xffdeU,0x3db2U,0x4f41U,0xca4dU,0x23a4U,0xc253U,0xd2deU,0x087fU,
		0x4dfbU,0x7fb7U,0x6ec6U,0x5922U,0xf3a2U,0x9b29U,0xb765U,0x2c37U,0x513cU,0xb912U,
		0x1043U,0xcf86U,0x5518U,0x46f7U,0x2700U,0x05f1U,0x4130U,0x123bU,0x2e6aU,0x46a3U,
		0x28f0U,0xef13U,0x27a3U,0x3a06U,0xa5b8U,0x4630U,0xb5d5U,0x1b82U,0x8a04U,0x8ca0U,
		0xfff1U,0xb375U,0x3a8fU,0x18eaU,0x1b9eU,0xa799U,0x1c10U,0xd727U,0x832fU,0xb4f1U,
		0xedd7U,0x700aU,0xf9daU,0x57c0U,0x594bU,0x2b9eU,0xc819U,0x9d2cU,0xeb25U,0x4234U,
		0xae7eU,0x3580U,0xead5U,0xdd1dU,0xcc05U,0xcab9U,0xf0e4U,0x6156U,0x1f68U,0x5f80U,
		0x1355U,0xc232U,0x124fU,0xd07dU,0x929fU,0x1f9cU,0x1df3U,0xe4cfU,0x68e7U,0xad9fU,
		0xc02fU,0xb40fU,0x6c5eU,0x45afU,0x1d38U,0x355cU,0xbbedU,0x8d23U,0x24c4U,0x80cdU,
		0x6c25U,0xd097U,0x8911U,0x014bU,0xd93dU,0x9003U,0xc886U,0xb26bU,0xa561U,0xa1cfU,
		0x9d6aU,0x6f46U,0xbd4bU,0x713bU,0xb570U,0x42deU,0xb7ddU,0xe9deU,0x1f09U,0xddeaU,
		0x4c2fU,0x42a9U,0xd9e2U,0xbfdeU,0x6107U,0xf02eU,0x370fU,0x5e24U,0x7a6eU,0xf017U,
		0x7eddU,0xa700U,0x344eU,0x76c0U,0x1cb2U,0x158bU,0x31dfU,0xfe4cU,0xe04eU,0xe12cU,
		0x590bU,0x2c0eU,0x87a4U,0x3c9cU,0xe688U,0x04ebU,0x8c31U,0x11d5U,0x998dU,0x583dU,
		0xb42dU,0x8b26U,0xce05U,0x23cdU,0x287bU,0xc427U,0x9539U,0x8b50U,0x7d86U,0xa1a8U,
		0xac01U,0x73f5U,0x10e4U,0xcd83U,0xc47cU,0xabb7U,0x8b92U,0x9b56U,0x01d2U,0x41edU,
		0xa486U,0x0230U,0x2d95U,0x2addU,0x078eU,0x9741U,0x8e16U,0x39cfU,0x4bb2U,0x9966U,
		0xe1feU,0x82ecU,0x64b4U,0x4e33U,0x40a2U,0xd010U,0x70d8U,0x0fd1U,0xcdf2U,0x13f5U,
		0xb30eU,0xcc46U,0x8df9U,0xb544U,0xe077U,0xc952U,0x4fe0U,0x24faU,0x5e41U,0x2ee6U,
		0xea24U,0xdf5eU,0x3eeaU,0x611cU,0x2258U,0x439eU,0x278bU,0xde8bU,0x6fe8U,0x67bcU,
		0x2181U,0x89adU,0xc18cU,0x2006U,0x5879U,0x64d6U,0x823fU,0xa7e6U,0x9e12U,0x211dU,
		0x646bU,0xe709U,0x3485U,0xc796U,0x442dU,0x59fdU,0x17e9U,0xf595U,0xfb40U,0xfc02U,
		0xb675U,0x4f7fU,0x2b89U,0x74a8U,0x65c6U,0x15d2U,0x0a6fU,0xd3bcU,0xb713U,0xbbd4U,
		0xf2d9U,0xcfe3U,0x90a7U,0xaca8U,0x59ecU,0xe275U,0x7a3aU,0x85bbU,0x8761U,0x28cdU,
		0xcee2U,0x1795U,0x5f89U,0x7d19U,0xb9d0U,0xd551U,0x8a2bU,0x15b3U,0x0046U,0xd490U,
		0x275fU,0xa7f3U,0x9ec6U,0xb537U,0xacd3U,0xa683U,0x76e2U,0x977aU,0x268aU,0xb6e6U,
		0xf88fU,0x29d2U,0x5726U,0x2eefU,0x7baaU,0x2b96U,0xa91bU,0x2f24U,0x64c9U,0x6097U,
		0x1628U,0x0fdeU,0x2f0cU,0xd3fcU,0x35bfU,0xcb4bU,0x92f5U,0x1a21U,0xe6bbU,0xe21eU,
		0x9cbeU,0x93b2U,0x3226U,0xe6fcU,0xa0a7U,0x26e5U,0xb7f4U,0x55f5U,0x10c1U,0x8adcU,
		0xf512U,0x00f6U,0x9f75U,0x3ef7U,0x796dU,0xd663U,0x4854U,0xc518U,0x15c5U,0xd98aU,
		0xb2e3U,0x1665U,0xc319U,0x9403U,0x1d52U,0x2da2U,0x5768U,0x5f18U,0x05d2U,0x76d9U,
		0x0720U,0x7bb2U,0x6eb6U,0xfe3fU,0xed79U,0xfda3U,0xb073U,0x6e4fU,0x6b7dU,0x3340U,
		0x25eaU,0xec01U,0xcd44U,0x8b5dU,0xcd27U,0x6ab7U,0x2c4fU,0xd7fdU,0x5e17U,0x962aU,
		0xfc79U,0x5b3eU,0x71f5U,0x681eU,0xf411U,0x8cf0U,0xe167U,0xfed4U,0xf7feU,0x91edU,
		0xc4e4U,0x1f4eU,0xeb4bU,0x82dfU,0xee7aU,0xf6bfU,0x49edU,0x3239U,0x4005U,0x87b1U,
		0x644bU,0x4d65U,0x414eU,0x3ddcU,0xc2f5U,0xa12eU,0x2a95U,0x0295U,0xcd76U,0x5e31U,
		0x397cU,0x0393U,0x3f50U,0x1a46U,0xeb5bU,0x019aU,0x3ccaU,0xdaafU,0x6740U,0x566eU,
		0xbdccU,0xf34bU,0xa8f2U,0xf6f9U,0xab34U,0x5db8U,0xedfcU,0x0430U,0x4cf6U,0x8176U,
		0x747dU,0x1619U,0xee73U,0x17e2U,0x24bdU,0x8d6fU,0x725aU,0x07dcU,0x6998U,0x4a91U,
		0x727cU,0x15f2U,0x839eU,0xe13fU,0x1ae8U,0xbe78U,0xcd67U,0xaa20U,0xc74dU,0x5b17U,
		0x7e76U,0xbb75U,0x0d98U,0x9c91U,0x9243U,0xec73U,0xde23U,0x8c5cU,0x131cU,0x2f9aU,
		0x193dU,0x31d0U,0xea51U,0xcc6bU,0xf721U,0x0247U,0x2952U,0x0d33U,0xba9cU,0x5048U,
		0xa954U,0xa814U,0x364fU,0x7221U,0x3ab5U,0x3301U,0x8deaU,0x2472U,0xbee1U,0xe0f6U,
		0x32f4U,0x96acU,0xacf1U,0xab97U,0x660bU,0x6388U,0x51bcU,0x6347U,0xf989U,0x0768U,
		0x6f58U,0x1c01U,0x4edcU,0x52efU,0xab5dU,0x9250U,0x4b55U,0x31c3U,0x98e0U,0x7cd1U,
		0xed56U,0xf9a4U,0x0bf0U,0x03edU,0x83ebU,0xa821U,0x2395U,0xe154U,0x5fdcU,0x230fU,
		0x9ed6U,0xa315U,0x54d5U,0x7f20U,0x4a84U,0xbdd9U,0xc612U,0x96d1U,0xa3efU,0x4a02U,
		0x840cU,0x7b60U,0x4a98U,0x4170U,0x84a3U,0xff6aU,0x03c1U,0xc678U,0x4e59U,0xe4a9U,
		0x9c46U,0xdaa2U,0xb6faU,0x07e9U,0xb0ccU,0xe808U,0x1193U,0x7acbU,0x5ce2U,0x2cffU,
		0x08b0U,0x108fU,0xd23eU,0xd3f3U,0x1feaU,0x842bU,0x2e85U,0xdbfeU,0xf9f1U,0x62e7U,
		0x0f1eU,0xa1baU,0x2f3eU,0x53a6U,0x4ea9U,0xc6b0U,0x4498U,0xf121U,0x7074U,0xf3f1U,
		0x693fU,0x20d1U,0x55bcU,0x0a02U,0x3d6bU,0x472fU,0x5b2cU,0xbf5cU,0x99b6U,0x6dcfU,
		0x6eeeU,0x1cf7U,0x5b68U,0x4d16U,0xf4f8U,0xe555U,0xf3d3U,0x3ed2U,0xea85U,0x0865U,
		0xd76bU,0xbf02U,0x9425U,0x69acU,0xf0a7U,0x167dU,0xcd3cU,0xf0dcU,0x4781U,0x789fU,
		0x683cU,0xa16cU,0x981cU,0xe1cbU,0x0320U,0x9b43U,0xd79bU,0xc2aeU,0x17aeU,0x0975U,
		0x183aU,0x8d04U,0xd196U,0x12c6U,0xa3d7U,0x3e71U,0xb5b3U,0xa8b5U,0x4b35U,0xb1e7U,
		0x0462U,0x3cbbU,0xb1d0U,0x4d23U,0x490fU,0x3879U,0xc6a3U,0x2ef7U,0xa286U,0x19f6U,
		0x34bcU,0xedc6U,0xc005U,0x3469U,0xfb87U,0xb20bU,0x83dcU,0x44a5U,0x1685U,0xec95U,
		0x03a6U,0x989bU,0xc45aU,0x4d09U,0xe585U,0x9a7eU,0x35aeU,0x59a9U,0x7674U,0x6d70U,
		0x919cU,0x980dU,0x4ca2U,0xa42eU,0x7247U,0x9575U,0x3661U,0x0690U,0x37baU,0xa8c2U,
		0x699fU,0x8cceU,0x1bb6U,0xc382U,0x60a7U,0xbdbbU,0x37abU,0xd5adU,0x91c9U,0x4f6aU,
		0x76fdU,0xba3dU,0x41ceU,0xe6b2U,0x2975U,0xffb2U,0xf6e0U,0xe1beU,0x732bU,0x0557U,
		0xa117U,0x6849U,0x4274U,0xbfbaU,0x353eU,0x18f4U,0xcbb3U,0x4620U,0xe864U,0x5436U,
		0x7145U,0x62b6U,0xe658U,0x408bU,0xf9b9U,0xdad5U,0x0ab0U,0xcaa0U,0x7fb4U,0x3dafU,
		0x8846U,0xbd64U,0x18e7U,0x462dU,0x1238U,0xb90fU,0x1f99U,0x6153U,0x4231U,0x7007U,
		0xe9dbU,0xa1ccU,0xd094U,0x45acU,0x583aU,0x2c0bU,0x76bdU,0xf02bU,0x82e9U,0x2adaU,
		0xabb4U,0x8b4dU,0x2003U,0x439bU,0x24f7U,0x13f2U,0xe272U,0xd3b9U,0xfbffU,0xe706U,
		0x2f21U,0xb6e3U,0xa7f0U,0x7d16U,0xd987U,0x00f3U,0xe6f9U,0xcb48U,0x5b3bU,0x8b5aU,
		0xfda0U,0x5f15U,0x1a43U,0xa12bU,0x3236U,0x91eaU,0xbe75U,0x07d9U,0x8173U,0xf348U,
		0x246fU,0x5045U,0x31cdU,0x9c8eU,0x230cU,0xf9a1U,0x52ecU,0x6385U,0x108cU,0x07e6U,
		0xff67U,0x96ceU,0x4d13U,0x472cU,0xf11eU,0x62e4U,0x3e6eU,0xc2abU,0x789cU,0xbeffU,
		0x59a6U,0xec92U,0xedc3U,0x4d20U,0x0554U,0xba3aU,0xc37fU,0x9572U,0xa1c9U,0x462aU,
		0xdad2U,0x461dU,0xa128U,0x00f0U,0xd3b6U,0x2ad7U,0xba37U,0xc2a8U,0x07e3U,0x5042U,
		0x8461U,0x3f34U,0x3f31U,0x00edU,0x5236U,0x11feU,0xc991U,0x6a4fU,0xbb5aU,0x3f37U,
		0xd02eU,0xc99eU,0xae4bU,0xe010U,0xf753U,0x7bf3U,0x5738U,0x8464U,0x08e5U,0xa0faU,
		0xe544U,0xb504U,0x9c68U,0x80edU,0x1f71U,0x6a52U,0xc96aU,0x432fU,0xd7aaU,0xc130U,
		0xd585U,0x8031U,0x026cU,0xc994U,0xd99aU,0xb239U,0x33daU,0x3a47U,0x05dbU,0xe62aU,
		0x5416U,0x1201U,0xc3ffU,0x0e28U,0xc7f1U,0x5e4aU,0x1cd0U,0x1da3U,0x4e15U,0x5239U,
		0x6d2aU,0xf388U,0x3c84U,0x3be8U,0xb51dU,0xd247U,0x2a46U,0xc9a1U,0xa662U,0x1642U,
		0x41d7U,0x588aU,0x40d9U,0x8ba9U,0x3d19U,0xd031U,0xc5f5U,0xc123U,0xd1f6U,0x2541U,
		0x5865U,0x4e22U,0xc920U,0x3f3aU,0x54a9U,0x14a8U,0x69daU,0x6d68U,0x6816U,0xe4c7U,
		0x55d9U,0xbb5dU,0xf37bU,0x53f7U,0xc16bU,0xc68dU,0x32cfU,0xc70fU,0xd0c8U,0x7bf6U,
		0x9766U,0x0098U,0xf1b9U,0x572eU,0x59c6U,0x2d58U,0x6a7bU,0xf756U,0x84e7U,0x853bU,
		0xbf11U,0x0edcU,0x73a0U,0xebe3U,0xf0afU,0xe013U,0x1c2fU,0xaf7fU,0x0ec2U,0x5de0U,
		0x9dc4U,0x1b62U,0xf335U,0xae4eU,0x73c4U,0x69e2U,0xfa3eU,0x24a0U,0x45e4U,0xa5ddU,
		0x155fU,0xa0fdU,0xf372U,0xb2daU,0x2f88U,0x3801U,0x7261U,0x0ecfU,0xccb0U,0x08e8U,
		0xe289U,0x95c8U,0x4ebdU,0x26acU,0x3d8cU,0x002aU,0x8fa0U,0x8467U,0xa384U,0x6affU,
		0x1456U,0x80bbU,0x008bU,0xf108U,0xd835U,0x573bU,0x0b2cU,0x3a58U,0x444eU,0x182aU,
		0x1d66U,0x8d98U,0x309fU,0x80f0U,0x0d86U,0x28fcU,0x00d7U,0x9f4eU,0xfabaU,0xfe85U,
		0x9b35U,0x9c6bU,0x0f36U,0x9043U,0xab24U,0x5af9U,0x94f3U,0x7f88U,0x4807U,0xb507U,
		0x8b78U,0x6c4cU,0xe333U,0xb8b5U,0xe90bU,0x9f03U,0x505cU,0xe547U,0xaf76U,0x6aaeU,
		0x45bbU,0xca78U,0x06c2U,0x0eb2U,0x99c7U,0x4332U,0xc9f5U,0x0407U,0x26d1U,0xd7d4U,
		0xc4b1U,0xa56dU,0x034dU,0xc96dU,0x8cc0U,0x3639U,0x4883U,0xc5ebU,0xba27U,0x2d6dU,
		0xa511U,0x6a55U,0x57ccU,0xafb8U,0x43d1U,0x7026U,0x9b86U,0x6cf0U,0x93c4U,0x1f74U,
		0x5986U,0x386bU,0xf30aU,0xe2cdU,0x0810U,0xdca3U,0x09b3U,0x8034U,0x93edU,0x0a94U,
		0x5660U,0x6bdcU,0xbc3eU,0x4347U,0x2656U,0xd588U,0x9883U,0x8f23U,0x4048U,0xa2fbU,
		0x8666U,0xfebcU,0x593bU,0xc133U,0x3391U,0x455aU,0x82bcU,0xcc5dU,0xaa99U,0x3438U,
		0x9dfeU,0xd7adU,0xeda8U,0x3e92U,0x8d3eU,0x4f2bU,0x6f6bU,0x60ecU,0xe8b4U,0xb23cU,
		0x3f86U,0xe678U,0x5e73U,0x02c4U,0xe146U,0xb02fU,0x521cU,0xd99dU,0xd298U,0x06a9U,
		0x028eU,0xbbcdU,0x6f3fU,0x3416U,0x6af8U,0xc997U,0xe6abU,0xe801U,0x0abfU,0x0c4dU,
		0xd4d0U,0xe7faU,0x8ef1U,0x026fU,0x824fU,0x2b54U,0xf2a2U,0xdc59U,0xce34U,0x4fa6U,
		0xebd5U,0xe62dU,0x6ac3U,0x35edU,0x9e42U,0xa2b2U,0x5a8fU,0xcad6U,0x20ecU,0x05deU,
		0x206dU,0x809dU,0xa13aU,0xf4bcU,0x9a9fU,0xfe08U,0x38a3U,0x3a4aU,0xf32cU,0xfc70U,
		0xac1dU,0x8274U,0x3b5fU,0xf80aU,0x1002U,0x33ddU,0x127cU,0x5aa1U,0x43b1U,0xceeeU,
		0xe771U,0xc7b2U,0x4516U,0x0e2bU,0x8023U,0xeb0dU,0x8b14U,0x9833U,0x9786U,0xb8ddU,
		0x735eU,0xc402U,0xce5fU,0xea98U,0x2618U,0xf55aU,0xecf1U,0x7c58U,0xbd37U,0x1204U,
		0x3f6fU,0x1634U,0x6703U,0x69d0U,0x143eU,0xf686U,0x5601U,0x5419U,0x8872U,0x6b10U,
		0x460fU,0x7d31U,0x7466U,0xfb4cU,0x666eU,0x1da6U,0x5bbcU,0x4032U,0xe281U,0x609fU,
		0x3ffdU,0x8909U,0xc8e9U,0x1cd3U,0x74b8U,0x9f89U,0xb791U,0x96c0U,0x5c2eU,0xcf53U,
		0xa174U,0x5e4dU,0xa3abU,0x5540U,0x1e09U,0xae2fU,0x33c7U,0xfc31U,0x77efU,0xc7f4U,
		0x64e2U,0x53ffU,0xe36aU,0x1ebfU,0x5f07U,0xdaf8U,0x39dbU,0xf38bU,0x67ecU,0x90d3U,
		0xaf4cU,0x8e27U,0x83a5U,0xc0b2U,0x27baU,0x6d2dU,0x9f91U,0xac0dU,0xda0fU,0xf155U,
		0x11e1U,0xb901U,0xdc1dU,0x523cU,0xd126U,0xd4d7U,0x4b9bU,0x7ee9U,0xd04cU,0x9fdfU,
		0x994dU,0x4e18U,0x26c1U,0x676aU,0xb82aU,0x5b99U,0x4addU,0xca6bU,0x19bcU,0xd24aU,
		0x8965U,0xe380U,0x4ab2U,0x95afU,0xede9U,0xd8ddU,0xca17U,0xb520U,0xc684U,0x93faU,
		0xbfa6U,0xeebaU,0xafabU,0x5fe8U,0x6e94U,0x3bebU,0x4da4U,0x1e9eU,0x52fbU,0x4567U,
		0x436bU,0xd34eU,0xb044U,0x3c87U,0x70e8U,0x0d16U,0x36a5U,0x9c5eU,0x2fa6U,0xe8c1U,
		0x2b0eU,0x1645U,0x0e7eU,0x5e02U,0xa239U,0x3ac1U,0x2bf4U,0x1015U,0x34dbU,0xa665U,
		0x36eaU,0x7d54U,0x9ef6U,0x0b6fU,0xf7afU,0x78b6U,0x5dc4U,0xc9a4U,0x8503U,0x12afU,
		0xec2bU,0xa9c4U,0xbd7aU,0xbcc5U,0x4c58U,0x2a49U,0x106bU,0x7288U,0x4d2fU,0x7f95U,
		0x6c74U,0xcf3cU,0x53b7U,0x8bacU,0x1983U,0x0c40U,0x77ffU,0x013dU,0xa862U,0x1691U,
		0xf547U,0x40dcU,0x3913U,0x55a4U,0x636bU,0x29a3U,0x92ddU,0x87ceU,0x6562U,0x588dU,
		0x069cU,0x1d73U,0xb4e3U,0x0bbeU,0x8654U,0x2c7aU,0xf7d2U,0x41daU,0x0126U,0x65abU,
		0x4dc2U,0x2cc2U,0xe257U,0x465fU,0x6a67U,0xc126U,0x70a2U,0x694bU,0x5ebcU,0x8834U,
		0x15f9U,0x1b74U,0x8437U,0xc5f8U,0x8afbU,0xc57cU,0x0f5bU,0x032cU,0x56daU,0x4b1cU,
		0xdc96U,0xd034U,0x46baU,0x0c54U,0x25d8U,0x3259U,0x0871U,0x45c8U,0xc852U,0x3d1cU,
		0x8bf3U,0xdb39U,0x52bfU,0x6d53U,0x2dd2U,0x3608U,0x615fU,0x4e25U,0x59bdU,0x6eadU,
		0x66e5U,0x7406U,0x2854U,0xf4afU,0x0bdcU,0x5868U,0xc652U,0x7151U,0xf2baU,0xdeeeU,
		0x5f21U,0x27c7U,0xe11eU,0x2544U,0x9d05U,0xbb91U,0x10f8U,0x2503U,0x35e0U,0x7b2bU,
		0x0649U,0xd1f9U,0x778cU,0x1378U,0xf06cU,0x869bU,0xec4bU,0x1098U,0x8edeU,0x14abU,
		0x0442U,0xc788U,0xba46U,0x540cU,0xc107U,0x7b87U,0x84bfU,0x54acU,0x612cU,0x6d19U,
		0x495bU,0xddbcU,0x3784U,0xf8c5U,0x23edU,0x3f3dU,0xc944U,0x3114U,0xb8d0U,0xf63bU,
		0x78caU,0xbb45U,0xe941U,0xc923U,0xa106U,0x74c5U,0xe2f4U,0xb529U,0x653bU,0x2198U,
		0xdde5U,0xe4caU,0xa794U,0x591dU,0x85b6U,0x64d1U,0x4e2eU,0x8b21U,0x0195U,0x6819U,
		0x1660U,0x6092U,0xd3eeU,0xa310U,0xe0f1U,0xaa1bU,0x6844U,0x6d6bU,0xa8b0U,0xe550U,
		0xa0f5U,0x956dU,0x91e5U,0x45a7U,0x3a53U,0x69ddU,0x53f2U,0xf383U,0x2b4fU,0x3e8dU,
		0x3866U,0x6aa9U,0x6765U,0x53faU,0x6b0bU,0x5a9cU,0xdb34U,0x65a6U,0x7283U,0x0d11U,
		0xcda6U,0xf37eU,0x74c0U,0x1373U,0xe01fU,0xf817U,0x0385U,0xaf12U,0xc55cU,0xbb60U,
		0x20a1U,0x0ac6U,0x1ce3U,0xc5a8U,0xfae7U,0xd0d4U,0xb8f4U,0x55dcU,0xd964U,0xcdabU,
		0x6200U,0xcc18U,0xaa0bU,0x34e7U,0xb695U,0xc712U,0x270bU,0xbf73U,0x2a55U,0x2a05U,
		0xe490U,0xfb1cU,0x79beU,0x32d2U,0x2aedU,0x62d3U,0x891cU,0xa37aU,0x0c74U,0x464fU,
		0xe84aU,0xc690U,0x74eeU,0xb89cU,0x737dU,0x710eU,0x6cc4U,0xeb30U,0x991bU,0xc16eU,
		0x0d22U,0xaf17U,0x7c48U,0x1126U,0x9345U,0x16daU,0x3265U,0x009bU,0x548eU,0x5385U,
		0x4fbbU,0x8c42U,0xfa89U,0xf966U,0x5972U,0x9769U,0x0ae3U,0x5ec8U,0xc3d2U,0x8d88U,
		0x2c86U,0x38b9U,0x5a17U,0x7bf9U,0x4cc7U,0x757eU,0x83c7U,0xa86eU,0x16cbU,0x2fd7U,
		0xb25aU,0xd0cbU,0xf0ffU,0x038aU,0x4a55U,0x65eaU,0x3dc9U,0xfc84U,0x0930U,0x2d5bU,
		0xbb9dU,0x0e5eU,0xeb66U,0xe3ffU,0x233bU,0x5358U,0x22f7U,0x59c9U,0x301dU,0xdcc4U,
		0x3f49U,0x59e5U,0xb7d6U,0x3403U,0xd82eU,0x5731U,0x7324U,0x08a9U,0xb8edU,0xc84bU,
		0x9946U,0x8eeaU,0x3753U,0xf1bcU,0x485fU,0xf81cU,0xb28dU,0x6547U,0x77d2U,0xdd71U,
		0xb71eU,0x853eU,0x790aU,0x8acaU,0x6585U,0x7f77U,0xaa27U,0x9da2U,0xd2f3U,0x84eaU,
		0x6fceU,0x2263U,0xc366U,0x4e8aU,0xa491U,0xb72eU,0xef73U,0xf759U,0x6a3dU,0x2b5bU,
		0x4b0bU,0x31eaU,0x631cU,0xf975U,0xf823U,0x6a7eU,0x1d15U,0xe024U,0x1802U,0xfe76U,
		0x4353U,0xeb9fU,0x4c16U,0xebe6U,0xfb0eU,0xc61dU,0xe198U,0xaaa5U,0xd70aU,0x0cf1U,
		0x51c7U,0x73a3U,0xdb0cU,0xd8a9U,0xfeadU,0xc758U,0x16a3U,0x738cU,0x02d0U,0x0edfU,
		0x2a75U,0x5377U,0xad3cU,0xfa34U,0x674bU,0xb3ffU,0x98ffU,0xbf14U,0xeefeU,0x0acbU,
		0xc010U,0xf3b7U,0x9fb6U,0xf677U,0x33f5U,0xaf82U,0x6892U,0x88efU,0x0220U,0xf624U,
		0x64a1U,0x1bc1U,0xb584U,0x1c32U,0x904fU,0x81c1U,0x8820U,0xdbb1U,0xdae9U,0x93d0U,
		0x0e9bU,0xe016U,0x4eb4U,0x6442U,0xc979U,0x1785U,0xcfb9U,0x885fU,0x6271U,0xf0b2U,
		0x9549U,0x20a6U,0x1beaU,0x0a4dU,0x8851U,0xf769U,0x8ecaU,0x1b65U,0x50f8U,0xda1bU,
		0xf4d3U,0x2f2cU,0x57faU,0x2a14U,0x1ecbU,0x9dc7U,0x09f2U,0x3830U,0xc4cfU,0x5c3aU,
		0x479cU,0x5369U,0x78a7U,0x5de3U,0x9c0dU,0xa741U,0x251dU,0x519cU,0xfb58U,0x8616U,
		0x3bbbU,0x0ec5U,0xabd6U,0xbb65U,0x0e37U,0x0661U,0xd68eU,0x0e45U,0xd1a7U,0x69e5U,
		0x49b8U,0x4598U,0x144eU,0x296dU,0xcd6eU,0xbd43U,0x001aU,0x73c7U,0x1cdbU,0x4b93U,
		0x1ab6U,0x3822U,0xd641U,0xd840U,0x2577U,0xae51U,0x35f9U,0x8256U,0x73cfU,0x5a57U,
		0x85a4U,0xd101U,0x4866U,0xf338U,0xa2c8U,0xc561U,0x5d64U,0x9dbaU,0xba08U,0x3a99U,
		0x850eU,0xa5e0U,0x9e31U,0xf040U,0xbd55U,0xb3baU,0x376eU,0xe1a5U,0x5336U,0x45e7U,
		0x8707U,0x6e9fU,0x8609U,0x56bdU,0x56e5U,0xbbadU,0xfe99U,0x24a3U,0xfdefU,0xdb23U,
		0x8e48U,0x6376U,0xc150U,0x2272U,0xbeadU,0xfa41U,0x7b79U,0xd0d9U,0x1f03U,0x1b98U,
		0xa065U,0x0654U,0xeb55U,0xb2ddU,0x7a34U,0x64aeU,0xa111U,0xb5adU,0xd238U,0xbedbU,
		0x0b26U,0xf375U,0x08dfU,0x3230U,0x26bbU,0x886cU,0x8249U,0x5980U,0xd95eU,0xa100U,
		0x8bedU,0x1065U,0xa2c2U,0x9543U,0x1d0fU,0xf0f9U,0xaf04U,0x1562U,0xfaf2U,0xfaecU,
		0xee2bU,0xf315U,0x7a5dU,0x1d24U,0xbe97U,0x0ed2U,0x4b5fU,0xaf25U,0xb6a8U,0xa711U,
		0x6f76U,0x12a1U,0xda93U,0x7264U,0x5187U,0xb6b4U,0xc464U,0xdeb2U,0x2103U,0x1ed8U,
		0xeb92U,0x3804U,0xee8dU,0x30aaU,0x9465U,0xbe8bU,0xd340U,0x5b72U,0xca00U,0x2f8bU,
		0x199fU,0xc5adU,0xf1a8U,0xcbbdU,0xb116U,0x5342U,0x7c6eU,0x95cbU,0x83b0U,0x1778U,
		0xfa4dU,0xc56eU,0xbe7fU,0x2afbU,0xe928U,0xe28cU,0x3cacU,0x6b32U,0xce6aU,0xf096U,
		0x1755U,0xff26U,0xe835U,0x08ebU,0x81b4U,0x0276U,0xcf2eU,0x8118U,0x5129U,0x20f7U,
		0xf1c3U,0xccb3U,0xfaf8U,0x1ce8U,0xebf2U,0x38c6U,0x36afU,0x6593U,0xf0edU,0x002dU,
		0x1727U,0x90c5U,0xac2dU,0x6a71U,0xd361U,0x02dcU,0xaa59U,0x3d8fU,0x5d79U,0x776fU,
		0x5316U,0xd481U,0x5771U,0xe3d2U,0xba50U,0x26afU,0x88fbU,0x9352U,0x9d95U,0x4fe9U,
		0xe5f5U,0xd9f5U,0x3690U,0x4ec0U,0x9d8aU,0xcdb0U,0x1a9aU,0x5cffU,0x210eU,0x08b9U,
		0x99d9U,0x6b02U,0xf369U,0xc4dbU,0x7706U,0xce75U,0x3156U,0x59d8U,0x0a59U,0xa387U,
		0x488dU,0x31f8U,0xc26bU,0xd64dU,0xd4ecU,0x9957U,0xfdfaU,0x846aU,0x04ceU,0x7a28U,
		0x670dU,0xcb30U,0x0e51U,0x620dU,0xe3c7U,0x8fa3U,0x1568U,0xd969U,0xdb17U,0xca20U,
		0x9b09U,0x454cU,0x39f7U,0xf10bU,0x21caU,0xac41U,0x58c6U,0xa923U,0x6168U,0xb58fU,
		0xac37U,0x008eU,0x5ed4U,0x7917U,0xa4d0U,0xbea1U,0xe7ecU,0x9927U,0x4789U,0x80beU,
		0x5805U,0x5065U,0x32deU,0x2b68U,0x5d6eU,0x2349U,0x38acU,0x1459U,0x371fU,0x55e1U,
		0xad46U,0x9035U,0x9daeU,0x093dU,0x9ffdU,0x3a5bU,0x99e2U,0x3270U,0x1429U,0x6328U,
		0x1aefU,0x3bc5U,0x2716U,0x0b2fU,0xc8f1U,0x29f8U,0x0ea4U,0xb298U,0x696eU,0x3762U,
		0xe40bU,0x573eU,0x8926U,0x8ef8U,0x0022U,0x4f62U,0x3dd4U,0xb263U,0x375aU,0xd838U,
		0xaf0aU,0xb8f9U,0xc0daU,0xf919U,0x741bU,0xf915U,0xfd7cU,0x99f4U,0x78e2U,0xcaf4U,
		0x7747U,0x7ab5U,0xf5c9U,0x0704U,0x0cd0U,0x5adfU,0xe500U,0x70c6U,0xc493U,0x741fU,
		0x7206U,0x274bU,0xd785U,0xa58aU,0x9badU,0xff8fU,0x2c03U,0x5b0fU,0x23c5U,0xc1ddU,
		0x1b90U,0x111eU,0x9c56U,0xca70U,0xa35dU,0xf91dU,0xee4bU,0xfc50U,0x45faU,0x7d65U,
		0xe429U,0x2b2aU,0x834eU,0x1b38U,0xb2b1U,0xd05fU,0x8093U,0x7454U,0xfbe9U,0xc80bU,
		0x1051U,0xc0deU,0xe2aeU,0xa5c7U,0xecb2U,0x18faU,0xdc7aU,0x14faU,0xd458U,0x50b4U,
		0xb422U,0xdedbU,0xa9f9U,0xe08fU,0x1d2aU,0x4ae2U,0xbcd4U,0xcaf8U,0x6f0bU,0xf69eU,
		0x3dc0U,0x0807U,0xace8U,0x5f6aU,0x2726U,0xfa70U,0xf18eU,0x6f02U,0xbb1eU,0x51f1U,
		0x8209U,0xd891U,0x863fU,0x78e6U,0xbb2dU,0x61b9U,0x5f5aU,0xfc16U,0x917fU,0x9f29U,
		0x1510U,0xf42bU,0x6bbaU,0xbe3cU,0xbfd3U,0xf5baU,0x50a2U,0x5b9eU,0x874bU,0x99f8U,
		0x92b9U,0xe8d1U,0x7efbU,0xcceaU,0xf203U,0xcffcU,0xebcbU,0xebb3U,0xbb7cU,0xb921U,
		0x648eU,0x1521U,0x47ceU,0x8aadU,0xcea8U,0xfd80U,0xcb76U,0xad57U,0x7f50U,0x3544U,
		0x75b7U,0xdcaaU,0xbc12U,0x5588U,0x4a17U,0xd5e4U,0xdd78U,0x24e4U,0x7a63U,0xb82fU,
		0xc03dU,0x0708U,0xf662U,0x4f50U,0x8ab7U,0x4140U,0x60bbU,0x5cc5U,0x1413U,0xef24U,
		0x2ea8U,0x9cd3U,0x6725U,0x8bdcU,0x0367U,0xd069U,0x39c4U,0xf5cdU,0xa9b2U,0xfbb2U,
		0x71e7U,0xb2c7U,0xcc91U,0x22a4U,0x04ffU,0x6dfdU,0xa250U,0x9e5bU,0x7313U,0xc074U,
		0xf5efU,0xe385U,0x1acfU,0x7ab9U,0x6197U,0x6df3U,0x826aU,0xc249U,0x388bU,0x9cb0U,
		0x4ac1U,0x3ad3U,0xa4b9U,0xbd04U,0xb279U,0x910bU,0xbde9U,0x2090U,0x8c8eU,0x774bU,
		0xa47bU,0x115eU,0xff00U,0xbfc0U,0xb9c2U,0x2324U,0x6fb9U,0x3e10U,0xb83eU,0x1a82U,
		0x7a95U,0x6adfU,0xf31bU,0x896aU,0x569cU,0x70caU,0x928cU,0xfa67U,0xcec8U,0xd614U,
		0x196dU,0x2c61U,0x4cd9U,0x179dU,0x622bU,0xe34fU,0xe3f6U,0x6bd3U,0x075fU,0xf8daU,
		0xb667U,0xe504U,0x6c8fU,0xf7bfU,0x9b4bU,0xe97eU,0x3ce3U,0x1f2aU,0x1c1eU,0xba7dU,
		0x7cfdU,0xf499U,0x3d3fU,0xed23U,0x7956U,0xd24fU,0x9a95U,0x5ae3U,0x47ddU,0x8c23U,
		0x4481U,0x21a0U,0x9506U,0xa9a8U,0xd8baU,0x3a6eU,0x7e7dU,0x668cU,0xd1ccU,0x805dU,
		0xcf5bU,0xaa79U,0x4576U,0x0cd4U,0x0780U,0x0945U,0xdd2bU,0x427aU,0x952eU,0x0817U,
		0xd848U,0x1221U,0x8793U,0x90f4U,0x77d9U,0x4986U,0xee31U,0x19c1U,0xc9bbU,0x274fU,
		0x4d79U,0x5ba9U,0x0839U,0x02f5U,0x97c9U,0xf86aU,0x24efU,0x1314U,0x10dcU,0xadaeU,
		0xd230U,0xfa81U,0x2becU,0xc4a9U,0xe1eeU,0x720aU,0xb2fbU,0xb2a9U,0xa62fU,0xc1e8U,
		0x6504U,0x5024U,0x7db2U,0x5281U,0x0ca3U,0x25b5U,0xa906U,0x98d2U,0x50bfU,0xd8e2U,
		0xdb48U,0x7423U,0x0496U,0x6c15U,0x4285U,0xff4aU,0xdf49U,0x75c7U,0x5ee4U,0x392bU,
		0xb303U,0x76a9U,0xe5d7U,0x8c71U,0x9235U,0x1f35U,0x44dfU,0xc497U,0x7756U,0xaab7U,
		0xaee5U,0x543cU,0xca36U,0x951dU,0xc346U,0xfbbdU,0x752dU,0x6de9U,0xc7a8U,0x2c2dU,
		0xaf2bU,0xedeeU,0x3c33U,0xff93U,0xc54aU,0xf84bU,0xb877U,0x107cU,0x3936U,0xc08cU,
		0xbff7U,0x8f6cU,0x7aefU,0xef59U,0x89a2U,0xadb9U,0x57e2U,0x5f46U,0x9648U,0x9bb1U,
		0x9d17U,0xd94cU,0xd825U,0x5932U,0xed4dU,0xa7b1U,0x4311U,0x4c94U,0x86d0U,0x5f51U,
		0x556aU,0x11a5U,0xdbcdU,0x95b4U,0x68b8U,0xa58eU,0x1950U,0x8196U,0xa6d5U,0xa2e4U,
		0xd38cU,0x192dU,0x743cU,0xc3e7U,0x5b1eU,0x14d2U,0x8f3bU,0x963dU,0xea19U,0x85dfU,
		0xb167U,0xd789U,0x6e60U,0x7c0cU,0xb74dU,0xe86aU,0x5b58U,0xe2d4U,0x8b0aU,0x550eU,
		0x4926U,0xd188U,0x654eU,0x22d4U,0x4b65U,0x4ab7U,0xa7dbU,0xc1e1U,0xb2c0U,0xa583U,
		0xd086U,0x515fU,0xfd24U,0x6795U,0xef0fU,0x9b25U,0x4204U,0x88c7U,0x8fffU,0xad9bU,
		0xdd19U,0xd723U,0x41e9U,0x23c9U,0xfe48U,0x42a5U,0x74a4U,0xa7e2U,0xdf5aU,0xd00cU,
		0x55f1U,0x0fdaU,0xa67fU,0x28c9U,0x4d61U,0x8cecU,0x333cU,0x93ffU,0x0243U,0x5b13U,
		0x17deU,0xdaabU,0xe4a5U,0x7f1cU,0x31bfU,0x96a8U,0xe1c7U,0x3eceU,0x20cdU,0x8427U,
		0xd5a9U,0x9809U,0xb207U,0xb1e3U,0xb6dfU,0x2c07U,0x3dabU,0xbfb6U,0x11faU,0x4626U,
		0x4728U,0x07d5U,0x253dU,0x3be4U,0x3a43U,0xb500U,0x26a8U,0x249cU,0x0ed8U,0xc689U,
		0xc5e7U,0xca74U,0x5af5U,0x1826U,0xbbc9U,0x4f27U,0xa2f7U,0xe2c9U,0xf556U,0xceeaU,
		0xf4b8U,0xdc55U,0xf151U,0x1ebbU,0x96bcU,0x7d2dU,0x0b6bU,0x9c5aU,0xeeb6U,0x5b95U,
		0x0328U,0x2cbeU,0x299fU,0x7f91U,0xddb8U,0x8697U,0xdeeaU,0x6d4fU,0xf813U,0x3e89U,
		0xa30cU,0xb525U,0x8d84U,0x1122U,0xa376U,0xcc14U,0x4e86U,0x6543U,0x59e1U,0x65e6U,
		0xdbadU,0xf3b3U,0xc754U,0xfe72U,0x381eU,0x065dU,0x5c36U,0x0a49U,0x8868U,0x1b94U,
		0x56b9U,0x9db6U,0xf092U,0xcbb9U,0xdeaeU,0xf311U,0xd649U,0x5cfbU,0xd47dU,0x38c2U,
		0xb294U,0x9031U,0xbe9dU,0xca1cU,0x948eU,0xfc54U,0x8480U,0x3d56U,0x6873U,0x6400U,
		0x918bU,0x7bc4U,0x5397U,0x090aU,0x792bU,0x1f20U,0x5536U,0x5f76U,0x1fc9U,0xa854U,
		0x7f5cU,0xee4fU,0x3fc2U,0x46caU,0x43e5U,0x679cU,0x2ff2U,0xe533U,0x6b8bU,0xb1fcU,
		0xa09cU,0xccf6U,0x308cU,0x87e6U,0x7094U,0x5fedU,0x7eacU,0xf921U,0x9515U,0x14f2U,
		0x98beU,0x07a3U,0xd414U,0xa644U,0x1c43U,0x305bU,0x7212U,0xbc6eU,0x057cU,0xaf9dU,
		0xea07U,0x48c3U,0xdefdU,0xa361U,0xfdd5U,0x13a2U,0x9411U,0x4091U,0xc1f9U,0xf929U,
		0xcb40U,0xa80cU,0x9739U,0x7f11U,0x105dU,0x7576U,0x12a7U,0xafb0U,0x415cU,0x2b2eU,
		0xf197U,0x8069U,0xda6fU,0xbbfdU,0xa97bU,0x2557U,0x77e5U,0x9295U,0x768eU,0xfad9U,
		0x37f2U,0xc6faU,0x7342U,0x3d35U,0x0b57U,0xe42dU,0x9672U,0x61efU,0x86d9U,0xf8e6U,
		0x2772U,0x830cU,0x867eU,0x2d12U,0xa8a2U,0x5004U,0xa6a0U,0x2532U,0x7962U,0xeebfU,
		0xf05bU,0x7d69U,0x71a4U,0xb77dU,0x6fc5U,0xd28fU,0x03b8U,0x18a1U,0x52ceU,0xec3dU,
		0xec7aU,0x47c5U,0xfee2U,0x1879U,0xec83U,0x3adfU,0x7ca6U,0x45feU,0xa25cU,0xc055U,
		0xa5f9U,0xe65eU,0xfcc8U,0x0a9bU,0xc644U,0x9cdfU,0x0826U,0x6247U,0x8ad1U,0x6b4fU,
		0x6f7cU,0xbfabU,0xf8d4U,0xd063U,0xd88bU,0x70c0U,0x7d27U,0xd71dU,0x5f40U,0xc4a3U,
		0x7716U,0xbca8U,0x3d2fU,0xa84eU,0xdb54U,0xd4afU,0xb7bbU,0xbcaeU,0xd6a3U,0xb2b5U,
		0xced7U,0xbe02U,0x0a15U,0xfd2bU,0x0d78U,0xdecaU,0x6d8cU,0x292cU,0x4d85U,0x647eU,
		0xac63U,0x56ffU,0xb0b4U,0x1ea3U,0xa6e1U,0x1b3cU,0x712cU,0x695bU,0xaf5eU,0x6f31U,
		0x0afdU,0x4405U,0x2698U,0x88a3U,0xb63aU,0xe876U,0x3c00U,0x4a28U,0xd431U,0x1e75U,
		0x15a8U,0x8352U,0x9c33U,0x9b1aU,0xed9aU,0x62bcU,0x3942U,0x3ea8U,0xf17aU,0xd220U,
		0x329eU,0xd16eU,0xc2dbU,0xa7bdU,0xa717U,0x4da9U,0x3ff3U,0xc80fU,0x4210U,0x3b51U,
		0x73e6U,0x8f0aU,0x40bfU,0x6221U,0xdfa8U,0x28d5U,0xe98dU,0xe0beU,0x5cb1U,0xdd53U,
		0x307cU,0x771cU,0x33abU,0xfbedU,0x3f61U,0xad11U,0xe1d3U,0x7499U,0x267fU,0xef98U,
		0xec5aU,0x266aU,0x75e3U,0x9390U,0x1a2fU,0xefd1U,0xa964U,0x3bf0U,0x357aU,0x7458U,
		0xd286U,0xcce1U,0x6e49U,0xb531U,0x2ee0U,0x1585U,0x3463U,0xf3ebU,0x924aU,0x1613U,
		0x8535U,0xb233U,0x96c8U,0xca9aU,0x9f83U,0x8097U,0x8f1dU,0x903dU,0x608cU,0x714bU,
		0x559eU,0x93f4U,0x382aU,0xd8a3U,0xdcbeU,0x62cdU,0x7911U,0x7769U,0xb6aeU,0x6e99U,
		0x57a4U,0xa5cbU,0x4588U,0xb37fU,0x6d98U,0x8e6fU,0x5d21U,0x5c66U,0xddf4U,0xdea3U,
		0x2a8aU,0x4500U,0xf5a3U,0x66b9U,0xf35bU,0xbcb4U,0xc415U,0xe2b2U,0x32aaU,0x0899U,
		0x3b2eU,0x5166U,0x02aeU,0x12e5U,0xb3abU,0xe882U,0x116dU,0x90eaU,0x9fd5U,0x42d4U,
		0x3427U,0xd353U,0x90b5U,0xc0e2U,0x3a21U,0xdd5fU,0x87bbU,0xc829U,0x764dU,0xd910U,
		0x1a3bU,0x8de2U,0xe1f6U,0x9026U,0xaefcU,0xb252U,0x4c50U,0x93bcU,0xb3e6U,0x1055U,
		0xc7a0U,0xa9f1U,0xcbeeU,0xcaa6U,0x2cd7U,0x7eb4U,0x05b3U,0xc431U,0xcb21U,0x58b1U,
		0xb3eeU,0xe7deU,0xb6baU,0x4370U,0xcfa2U,0x14feU,0x61a6U,0x452aU,0xa33bU,0xe47aU,
		0x03f7U,0x72d2U,0x6282U,0x9de4U,0xfde1U,0xb364U,0x4a3dU,0x8d79U,0xf406U,0xb7c1U,
		0x6d7aU,0xdc7eU,0x63d7U,0x3e2aU,0xb552U,0x5787U,0x084aU,0xfc60U,0xda05U,0xd933U,
		0x9619U,0x1d01U,0x7f68U,0x1eecU,0x69fcU,0x456cU,0xe399U,0x18feU,0x2f75U,0xaae4U,
		0x0c30U,0xa448U,0xfcd4U,0x47f0U,0x5179U,0x2ceeU,0x9c98U,0xe9c3U,0xf6b4U,0x18adU,
		0x01f9U,0xfe67U,0x86e5U,0xecb6U,0xf2eeU,0x6572U,0x4b02U,0x0ab6U,0x118aU,0x803bU,
		0x6ff6U,0xb108U,0xc63bU,0xbc09U,0x8545U,0x872aU,0x518dU,0x5300U,0x6e35U,0xdedfU,
		0x5e0cU,0xc2c3U,0x1258U,0xb055U,0x50eaU,0x3c19U,0x3b0cU,0x4669U,0x2ac7U,0xafe7U,
		0xe7beU,0xb4d1U,0xb7a5U,0xd4b5U,0x64c2U,0xb426U,0x6ebfU,0x8053U,0x6d23U,0xd08dU,
		0xf4f1U,0xcd60U,0x7785U,0x70e1U,0x1275U,0xaf6fU,0x9d83U,0x7b72U,0xeef7U,0x0d1bU,
		0x3335U,0x50b8U,0xf5e8U,0x9c4fU,0xeef0U,0x3420U,0xb0adU,0x708dU,0xf6e9U,0x7870U,
		0xf5dcU,0x7fedU,0x2bb2U,0x66a0U,0xafc2U,0x1343U,0x468dU,0xd45cU,0x324aU,0x9fe9U,
		0x3b18U,0xdadbU,0x8fd0U,0x3132U,0x6b1aU,0x0965U,0x8a93U,0x7864U,0x9b64U,0x68cdU,
		0x726aU,0x70edU,0xa347U,0x4ae6U,0x6a2fU,0x8bbcU,0x1d8dU,0xf146U,0xed6eU,0xb599U,
		0x999bU,0x8769U,0x285dU,0x5793U,0x7779U,0xb932U,0x1a11U,0xdb5aU,0x67e2U,0x1d2eU,
		0xd4ffU,0x9938U,0x9824U,0x0598U,0xfce0U,0x94fcU,0x856bU,0x7886U,0xf04aU,0x5621U,
		0x3b68U,0x8047U,0x717fU,0x3c8cU,0x0ad5U,0xe093U,0x41aaU,0x34a6U,0x76ceU,0xe2fcU,
		0x315fU,0xe2beU,0xd371U,0xc5d6U,0x0d9fU,0x8620U,0x57b0U,0xa06eU,0xb799U,0xaad2U,
		0xcfc2U,0xa9fdU,0xc04cU,0xad4eU,0x3047U,0xf9bfU,0xb8a6U,0x09baU,0x1abeU,0xf265U,
		0x3a2dU,0x2c8fU,0xb725U,0x340dU,0xda99U,0xb049U,0xc8bbU,0xf6a2U,0x753cU,0xbacdU,
		0x6cbbU,0xe902U,0x8395U,0x6bc7U,0xfb69U,0x677fU,0xa093U,0x23bcU,0x9217U,0x8dd4U,
		0x0745U,0xbe08U,0xd597U,0x6f0fU,0x979cU,0x7851U,0x7ad9U,0x88ceU,0x1961U,0xaa89U,
		0x30baU,0x812fU,0x6788U,0xb856U,0x8e61U,0x2371U,0x7ff4U,0xe8c6U,0x288eU,0xcafcU,
		0x52acU,0x3b7aU,0x2844U,0xf48dU,0x2f55U,0x4ddbU,0xa507U,0x248fU,0x502cU,0xbc1fU,
		0x1e35U,0xb977U,0xb651U,0xeca8U,0x5e8bU,0xbcd8U,0x1e23U,0x1108U,0xb65bU,0xbd6aU,
		0x6013U,0xa64cU,0x841aU,0xd315U,0xdc39U,0xec10U,0x7ebcU,0xebc0U,0x1edeU,0x2fabU,
		0x66b1U,0x5f6eU,0xfa79U,0x5ad7U,0xc02aU,0x832aU,0xb760U,0x8dccU,0x9e0dU,0x70d3U,
		0x2876U,0xd9ddU,0xe162U,0x1d4dU,0x2f07U,0xceddU,0x4a7fU,0xacecU,0x7e71U,0x673bU,
		0x4c9dU,0x045dU,0xea80U,0x2e80U,0xdacdU,0x816eU,0x76b8U,0x3539U,0x45dfU,0x32caU,
		0xb518U,0x9c63U,0x6f66U,0x080bU,0x06bdU,0x1d61U,0x5f02U,0x7461U,0xe76cU,0xce2fU,
		0xe252U,0x6c6fU,0x2fa1U,0x4ad8U,0x3861U,0x6536U,0xec46U,0x2dcdU,0x77cdU,0x3dc4U,
		0x9340U,0xaa06U,0xd689U,0x884cU,0x9fb1U,0x434eU,0xb111U,0x7a58U,0xa060U,0xba03U,
		0x9da9U,0x9b04U,0x2109U,0x36aaU,0x92b4U,0x6f06U,0xee46U,0x7416U,0x47d8U,0x9287U,
		0x6192U,0xf65dU,0x194bU,0xc545U,0x0491U,0x4d74U,0xa371U,0x5af0U,0x17d9U,0xb2bbU,
		0x719fU,0xf192U,0x9510U,0x847bU,0xd281U,0x420bU,0x7127U,0xd886U,0x2f70U,0x61a1U,
		0x3a1cU,0x4583U,0x41a5U,0x6a2aU,0xf5e3U,0x5e07U,0x06b8U,0xfa74U,0x52a7U,0x7537U,
		0x2480U,0xca50U,0x52a2U,0xee41U,0x4b6fU,0x6ba9U,0x1a69U,0x858eU,0x51deU,0x4f44U,
		0xd329U,0x9f17U,0x25a8U,0x272aU,0x7dc1U,0x1bd3U,0xab7bU,0x3db5U,0x7568U,0x6c9cU,
		0x751cU,0x373aU,0xaa3fU,0x06e2U,0x1407U,0xffe1U,0xdeb8U,0x0e83U,0xe0d7U,0xd895U,
		0xca55U,0xff0dU,0x1dbbU,0x0882U,0x8eaeU,0x5501U,0xb608U,0xd559U,0x40c9U,0x3e04U,
		0x7e26U,0xd2e1U,0xf466U,0xd6a9U,0x748bU,0x820dU,0x23ffU,0xf420U,0x097dU,0xc256U,
		0x85d2U,0x7935U,0xbc91U,0x1541U,0xa46fU,0x4221U,0x734cU,0x23a7U,0x11bcU,0x164aU,
		0xa8dbU,0x51f5U,0x2485U,0x1b2eU,0xd7caU,0x5925U,0xfceaU,0x8e99U,0xa276U,0x2084U,
		0x1323U,0xfd08U,0xe743U,0x6ec9U,0x5548U,0xcc9eU,0x6a06U,0xbb22U,0xd4a1U,0x3278U,
		0xf83eU,0x7fbaU,0x2636U,0xbc45U,0x825eU,0x5cd2U,0x9326U,0x1177U,0xaa2eU,0x4dfeU,
		0xc46aU,0x2b13U,0x0470U,0x61bdU,0x8593U,0xb9dfU,0xbd1aU,0x2c3aU,0x75abU,0x38f8U,
		0x4384U,0xa1e0U,0xa99bU,0xde78U,0x4f19U,0xb768U,0xc287U,0xded0U,0x2447U,0xbb31U,
		0x052dU,0xe5e4U,0x79b5U,0x9b2cU,0x4cedU,0x24d8U,0x2ebdU,0x920eU,0x552dU,0x0cc7U,
		0x0b94U,0xf3a5U,0x7877U,0x101aU,0x9b75U,0x78eaU,0x1a6eU,0xc1b3U,0x873fU,0xcf89U,
		0x9587U,0x664dU,0xe342U,0xd8efU,0x650cU,0x8522U,0x44b3U,0x1046U,0x8cdeU,0xd53aU,
		0xb096U,0x8643U,0x4260U,0xb12fU,0x8221U,0xb915U,0xc76fU,0xd41cU,0xe329U,0xed16U,
		0x85eeU,0x1c73U,0x2cdfU,0x513fU,0x30b0U,0x2bf9U,0x3cfeU,0x9f2dU,0x6baeU,0xcbd9U,
		0x3c4aU,0x05f4U,0xf742U,0x398eU,0xe85cU,0xa8eaU,0x3707U,0x7449U,0x9e29U,0x2703U,
		0x895dU,0x0d7eU,0x5d19U,0x9183U,0x97c1U,0x78daU,0x2214U,0x46faU,0x75a3U,0x838dU,
		0x91aaU,0x4c86U,0xd211U,0x29e1U,0x8490U,0x551bU,0xf56fU,0x3ac6U,0xbb12U,0xfc1aU,
		0x4b74U,0x94afU,0x4843U,0x46a6U,0x8dabU,0x7afcU,0x2a25U,0xe41eU,0x8c63U,0xbf2eU,
		0xfba5U,0x2e6dU,0xa890U,0x318aU,0x9f5dU,0x5f5eU,0x4071U,0xfe1fU,0xd5bbU,0x123eU,
		0x7c1dU,0x6be3U,0x08fbU,0xf250U,0x7847U,0x2741U,0x7f7eU,0x4133U,0xee93U,0xa23eU,
		0xe573U,0xbe40U,0x9f1cU,0x7c85U,0xc23bU,0x3a09U,0x7a0eU,0xff84U,0xa150U,0xd2bfU,
		0xadc8U,0x0c11U,0x7dd7U,0x27a6U,0xfbddU,0xfd31U,0x21eeU,0x6bbeU,0x7bbbU,0xcaebU,
		0x0c90U,0xef16U,0x5272U,0xdf18U,0x9b95U,0xd019U,0x829fU,0x1c8fU,0xbf6aU,0x28f3U,
		0xf6f0U,0x7d59U,0xd5d7U,0xf42fU,0xd32eU,0xe4b2U,0xbec0U,0x1b85U,0x4978U,0xdce5U,
		0xffbbU,0x9e80U,0x89e6U,0xc0c5U,0xd44cU,0xb5d8U,0x03d9U,0xf1d8U,0x0f2cU,0x1514U,
		0x9acdU,0x35ceU,0x3ebfU,0x4633U,0x073bU,0xe4f6U,0x9bf4U,0xdd8aU,0xc899U,0xad22U,
		0x4746U,0xa5bbU,0x380aU,0x36efU,0xa57cU,0x5ba2U,0x4f49U,0xf90eU,0xc2bcU,0xb378U,
		0x70b9U,0x3d4fU,0x7c7eU,0xb9d8U,0x740fU,0xbac6U,0x2426U,0xfff4U,0x3c61U,0x0a1bU,
		0xcdc8U,0x50a6U,0x5579U,0x3e3dU,0x17b6U,0x8ca3U,0xde46U,0x5b02U,0xaaf6U,0xf785U,
		0x6edaU,0x7aa9U,0x827dU,0x8a07U,0x0e0dU,0xa66aU,0xa030U,0xf5beU,0x51e3U,0x7b97U,
		0xfc43U,0xa79cU,0xce7eU,0x7992U,0xd779U,0x72e7U,0xde2aU,0x0105U,0x46daU,0x1ba1U,
		0xa3b3U,0x836bU,0x178eU,0xbfd7U,0xe8f8U,0x99eaU,0x80e6U,0x18edU,0xdbf7U,0x5667U,
		0x3601U,0xca64U,0x4f9fU,0x8d91U,0x658cU,0x3a92U,0xeb98U,0x34e0U,0x61d1U,0xe8d5U,
		0x1bd8U,0x0a22U,0xe6deU,0xb4f4U,0x4702U,0x21f7U,0x5212U,0x871dU,0x48e7U,0x7b46U,
		0x2ec6U,0x8332U,0x3857U,0x6484U,0xd0a2U,0x92bdU,0x14bdU,0x443cU,0xb6d3U,0xd72aU,
		0x58e3U,0x4760U,0xe9b6U,0x0d5dU,0xfb72U,0x358bU,0x9a2fU,0x1c13U,0x134aU,0x78bbU,
		0xbe63U,0x99fcU,0x7dc6U,0x4bc6U,0x4cbeU,0x57c3U,0xa94bU,0x97fdU,0x31aaU,0x600aU,
		0xc1f0U,0xdc71U,0x92ceU,0xf9ddU,0x87f8U,0x394fU,0xd265U,0x874fU,0xe22fU,0xe732U,
		0xe9f4U,0x700dU,0x31b3U,0x07abU,0xd135U,0x4412U,0x72c5U,0x754fU,0xcaaeU,0xeddaU,
		0x5b78U,0xf7b4U,0x3b9eU,0xd000U,0x272fU,0x6829U,0x5ca0U,0x9d2fU,0x7041U,0x984aU,
		0x1a51U,0x56a5U,0xce25U,0xcff2U,0xe66eU,0xc81cU,0xb4abU,0x4d8bU,0x4d55U,0xf207U,
		0x08ccU,0x9363U,0x431aU,0x2ba1U,0x48a2U,0x77a6U,0x0174U,0x0754U,0xae73U,0x33b8U,
		0xa9e4U,0x594eU,0xd66eU,0x0b74U,0x90e2U,0xcceeU,0x25adU,0xc1d5U,0xd3c7U,0x3583U,
		0x0cbfU,0xb84eU,0xba18U,0x0d4fU,0xcc72U,0x605cU,0xe78bU,0xae81U,0x52d7U,0xde96U,
		0x493dU,0x7effU,0x41f8U,0xcfcbU,0xb48dU,0x4237U,0x6f9fU,0x8f2aU,0x6fffU,0xe137U,
		0x11cdU,0xd609U,0x226aU,0xeb28U,0xd346U,0x9efbU,0xd4c6U,0xb925U,0x6ca1U,0x3c68U,
		0xa766U,0xcabcU,0xa056U,0x4a0dU,0xa183U,0x48d0U,0xe445U,0xdda5U,0x6393U,0xcc08U,
		0xf131U,0x2932U,0x5614U,0xbb80U,0x5ea8U,0x5086U,0x3370U,0xdd20U,0xa556U,0x511aU,
		0xaec2U,0x9bd7U,0xdabaU,0xba65U,0xdba1U,0xead8U,0xafc9U,0x12b4U,0xedd1U,0xebb7U,
		0x756dU,0xe086U,0x5762U,0x5f83U,0xcdecU,0x6101U,0xb1caU,0xf9ebU,0x6605U,0xbdc6U,
		0xd0c2U,0x1f6bU,0x2469U,0xcbadU,0x6668U,0xebcfU,0x09adU,0x3099U,0xdddfU,0x6159U,
		0x53b1U,0x19b6U,0x8ec4U,0x4c10U,0x092aU,0xb68fU,0x39f1U,0xf0e7U,0xbe91U,0x8508U,
		0xcec2U,0x8ab1U,0x3dbaU,0xfd76U,0xbbc3U,0xd080U,0xb871U,0x0833U,0x73e0U,0x7d21U,
		0xda69U,0x686dU,0x1d87U,0x1252U,0xa335U,0x6d92U,0x1db5U,0x47d2U,0xc024U,0x6cb5U,
		0xc2b6U,0xc235U,0x3c44U,0xbd14U,0xbbbdU,0xa760U,0x5c9aU,0xe6d8U,0xccceU,0x1358U,
		0xc3b5U,0xc9a9U,0xd632U,0x1525U,0xab80U,0xb944U,0x36c7U,0xe4d2U,0x1497U,0x0958U,
		0x6dabU,0x1e44U,0xea58U,0x868bU,0x3b8cU,0x1df6U,0x5e55U,0x8f4fU,0x8d18U,0x6492U,
		0xf7e9U,0x3a63U,0x762eU,0x1f9fU,0x1336U,0x988aU,0xae59U,0x68fdU,0xeeaaU,0x89bcU,
		0x6fd5U,0x92a2U,0x946bU,0x5dc9U,0xc7c1U,0xad5bU,0x06e7U,0xfffbU,0xc9afU,0xb412U,
		0x221cU,0x050fU,0x1f13U,0x2293U,0x844aU,0xd9bdU,0xb0f5U,0xc032U,0x2363U,0x1ea9U,
		0x9311U,0xcb7aU,0x0337U,0x0dd2U,0x2e98U,0xada2U,0x915fU,0x41b7U,0xf300U,0x87d9U,
		0xd6baU,0xd3cfU,0x170bU,0x68eaU,0x66a7U,0xbccaU,0xb446U,0xfd84U,0xaa44U,0x1383U,
		0x400fU,0x355fU,0x64f8U,0x86c0U,0xc92eU,0x9a65U,0xa637U,0x82fbU,0x831cU,0x1d3bU,
		0x2df5U,0x9cc8U,0xefb7U,0xceacU,0x348fU,0x1738U,0x7c3fU,0x45b2U,0xc744U,0x98c6U,
		0xbe5aU,0x2885U,0x7ea3U,0xa354U,0xcbf6U,0x6c61U,0xc5b3U,0xbd7fU,0xe5cbU,0xdcaeU,
		0x373fU,0x5abbU,0xc3bbU,0x80d0U,0x0bfaU,0x5636U,0xeb7aU,0xcc86U,0x4f36U,0x2172U,
		0x61e2U,0x24c7U,0x0630U,0xb0baU,0x767eU,0x75bbU,0x797eU,0x8f87U,0x00b8U,0x8d26U,
		0x74d9U,0x2907U,0xed00U,0x313fU,0xd9d3U,0x6ef8U,0x433dU,0xbbf0U,0xb5c3U,0xa9c9U,
		0x94c6U,0x3548U,0x7521U,0x1265U,0xbafdU,0x014eU,0x8e32U,0xd85eU,0x7facU,0xee6bU,
		0x31d7U,0xc069U,0x8601U,0x8914U,0xbf9eU,0xab1cU,0xc40dU,0x7f54U,0xe1e6U,0xc48bU,
		0x9309U,0xd09aU,0x243fU,0xd58fU,0x10b2U,0x1533U,0x5fa1U,0x80a8U,0x84f1U,0x6c28U,
		0x19a5U,0xec30U,0xd161U,0xd5e8U,0x0e88U,0x242dU,0x135eU,0xb26eU,0x4125U,0xf6d1U,
		0xc724U,0xe261U,0xe0a8U,0x628fU,0x4305U,0xc889U,0x2b73U,0x5705U,0x3387U,0x4a1bU,
		0x0729U,0xd974U,0x6136U,0x9006U,0xea76U,0x9175U,0x4ecbU,0x40abU,0x72fbU,0x8b35U,
		0xa98bU,0xd940U,0x2bb9U,0x728dU,0x7616U,0x558cU,0xdebdU,0x4e62U,0xcf7bU,0x6f49U,
		0x7bd8U,0x9ca5U,0x297eU,0xff31U,0x21b6U,0x49dcU,0xdfd3U,0x9d6dU,0x57d6U,0xa45fU,
		0xdc68U,0xbc16U,0xbc65U,0xd056U,0x24aeU,0xa1d2U,0x9bc2U,0xef34U,0x887cU,0x25feU,
		0x9af0U,0x9375U,0xf95dU,0xa564U,0x2f91U,0x1070U,0x73aeU,0xb833U,0xffe6U,0x1cc1U,
		0xccd4U,0x42e1U,0xcf0eU,0x6172U,0xb741U,0x5f91U,0x66eeU,0x81ccU,0x5d83U,0xb573U,
		0x2594U,0xac69U,0x8a24U,0x7a67U,0x32c0U,0xf5b0U,0xd7a3U,0x713eU,0x2f37U,0xab2dU,
		0x66deU,0x4aabU,0x9e3bU,0x00d0U,0xac26U,0xbd4eU,0xe191U,0x2a4eU,0xef08U,0x24e8U,
		0x140cU,0x7740U,0x3b05U,0xddedU,0x770fU,0x5390U,0xa149U,0x437dU,0x1944U,0xfb62U,
		0xc71dU,0x1f0cU,0xa17cU,0x520bU,0xc982U,0xdd7cU,0xbe1dU,0xa005U,0xda44U,0xe9e1U,
		0x74f8U,0x265dU,0x257fU,0x0616U,0x65f5U,0xc3dbU,0xd2faU,0xb7e0U,0xca06U,0x4c5dU,
		0x453bU,0x4f54U,0xff12U,0x3e44U,0xb94aU,0xbfe1U,0x29e9U,0xdf21U,0xa00cU,0xf8b2U,
		0xa03bU,0x060aU,0x1023U,0xd9e5U,0x463dU,0x6961U,0x80b0U,0xf666U,0xb5f4U,0xf25aU,
		0x95e0U,0x42acU,0x564bU,0x7a19U,0x5149U,0xeab7U,0xbe11U,0x6064U,0xbecbU,0x4c32U,
		0x9ff0U,0xcf41U,0xbd88U,0x070cU,0xca5aU,0x5cf1U,0xc59eU,0x5e27U,0x2a30U,0x30e0U,
		0x9ab4U,0x51ffU,0x25bdU,0xef3dU,0xe618U,0x3712U,0x0f9fU,0xd0f1U,0x44beU,0xc041U,
		0x3d81U,0x60dbU,0xb981U,0xf031U,0x3958U,0xbc76U,0x3af9U,0x3f9dU,0xc6d8U,0xc292U,
		0x58b9U,0x610aU,0x5348U,0x6c79U,0x68f4U,0x5cc9U,0xd89aU,0x1218U,0xab86U,0xa703U,
		0x16f3U,0xeaf5U,0x5a28U,0x2c6aU,0x2133U,0x3149U,0x5a4eU,0x7ee0U,0xffa9U,0x7132U,
		0x574dU,0x60bfU,0x6a11U,0xa732U,0x95bdU,0xf01aU,0x9010U,0xaac7U,0x0d01U,0x240aU,
		0xe3baU,0xa4efU,0x94daU,0x7a71U,0xe185U,0x7f9aU,0x037bU,0x4144U,0xe0dcU,0x7fffU,
		0xb14bU,0x158eU,0xe8eeU,0x56afU,0x7ec7U,0x171aU,0x0d3aU,0x6ce0U,0x833cU,0x1cb5U,
		0x3ae8U,0x114bU,0x971eU,0x8abbU,0xb329U,0x09c3U,0x7058U,0x76c3U,0x1554U,0xfec3U,
		0x5309U,0x2e08U,0xcac6U,0x417fU,0xb735U,0x3451U,0xb11cU,0x4d34U,0xb100U,0x9cd7U,
		0x5506U,0x5580U,0x152bU,0xe12fU,0xf248U,0x7a50U,0xbe24U,0xfa26U,0x49a3U,0x2e00U,
		0x27deU,0xe051U,0x7d45U,0x1b42U,0x8c01U,0x2eacU,0x21aaU,0x83faU,0xf0ccU,0xfe4fU,
		0xeb85U,0x5d42U,0xe008U,0x7ac3U,0xcfdbU,0x058dU,0x8110U,0x31e2U,0x3251U,0x0c45U,
		0xa227U,0xef28U,0x8eb3U,0xfbc7U,0x006cU,0x3c9fU,0xfd5dU,0x7be3U,0x22deU,0x1b13U,
		0x8361U,0x0a3fU,0x1369U,0x87a7U,0x6259U,0x38e0U,0xf951U,0x1417U,0xe71aU,0x66cdU,
		0xcf21U,0x2c11U,0x2d48U,0xf285U,0x21d9U,0x7174U,0x73b9U,0xbcebU,0xc58aU,0x590eU,
		0xcbc3U,0x1988U,0xe457U,0xd06dU,0x0887U,0xc143U,0xd638U,0x11d8U,0x91c0U,0xc87dU,
		0xf7e0U,0xd498U,0x8f14U,0x0777U,0xa07eU,0x8c34U,0x017fU,0xa6e7U,0xe467U,0x036bU,
		0x5329U,0x5b83U,0x12ceU,0x04eeU,0x4119U,0x0914U,0x9a3aU,0xc5c2U,0xb00cU,0x6c3dU,
		0xc704U,0xe68bU,0xa0ccU,0x8bb1U,0x5074U,0x8be0U,0x1dc0U,0xc508U,0x40eaU,0x8b29U,
		0x05a2U,0x04a6U,0x675bU,0xf7f4U,0x2959U,0x7e57U,0xc7daU,0xb430U,0xfc39U,0x4e7cU,
		0x1ef6U,0x6729U,0xa6f0U,0x3bcdU,0x337bU,0x5840U,0x54ecU,0x866dU,0xd109U,0xeafeU,
		0x97adU,0xe88cU,0xa498U,0x9990U,0xf1aeU,0x53bcU,0x3bddU,0xfbb6U,0x3e09U,0x50adU,
		0x095eU,0xc42aU,0xd219U,0xa805U,0xdd83U,0xed0fU,0x3733U,0xd30eU,0x25f7U,0x287eU,
		0x4c09U,0x440bU,0x5fa9U,0xa9b6U,0x716dU,0x3f96U,0xd7e8U,0x23d0U,0xc669U,0x968cU,
		0xf2c8U,0xd3a2U,0x2dafU,0xcc7aU,0x718fU,0xce08U,0xd463U,0x1696U,0x6458U,0xf5d1U,
		0x40ceU,0xe964U,0xdd36U,0xa1abU,0x1e60U,0x30fbU,0x00acU,0xe310U,0x0cabU,0xc449U,
		0x0213U,0x7d89U,0x58f6U,0x3ceeU,0xd65cU,0x39c8U,0x8c99U,0x2cb4U,0x6d61U,0x8b53U,
		0x8cfdU,0x721aU,0xf634U,0xa9bdU,0x69c9U,0x701fU,0xcb29U,0x953cU,0x177eU,0xa867U,
		0x3792U,0x22a8U,0xd55eU,0xbef0U,0x149dU,0xcd86U,0x5eefU,0x3127U,0x4a68U,0xd622U,
		0x3baeU,0x4efaU,0x37d7U,0x10e7U,0xbf92U,0x0b03U,0x8f46U,0xcc95U,0xca91U,0xaa70U,
		0xf89cU,0x73f8U,0xa0baU,0x4e73U,0x4d45U,0xa834U,0x651eU,0x57edU,0x4b8aU,0xac04U,
		0xbdb2U,0x0142U,0x63e8U,0xb2cbU,0xb60dU,0x13d8U,0xeed6U,0x9b59U,0x55ceU,0x602eU,
		0x1aa9U,0xc864U,0x024eU,0xd151U,0x10a6U,0x8b95U,0xcb58U,0x9454U,0x53e8U,0x71ebU,
		0xbae6U,0xbd9cU,0xf226U,0xabbaU,0x0e03U,0xa302U,0xf51fU,0x49abU,0xefa7U,0x202fU,
		0x4e91U,0xc47fU,0x83b6U,0x7804U,0xe3e1U,0x9e5fU,0xd6aeU,0xcdcfU,0xe4d8U,0x0233U,
		0x5a0aU,0xc9cbU,0xc989U,0x9c3eU,0x65beU,0xb4c6U,0x174dU,0xa489U,0x56d2U,0x6f37U,
		0xf2e6U,0xa254U,0x6e58U,0xcb6eU,0xe1deU,0x41f0U,0x4069U,0x9338U,0x3b40U,0xcc4fU,
		0xbadeU,0xb321U,0x005dU,0x01d5U,0x4694U,0x55a9U,0xb0e9U,0x6e01U,0xf46bU,0xe811U,
		0x7370U,0x9744U,0x5a7bU,0x568cU,0x6852U,0x50dfU,0xc95fU,0x8185U,0xdb6aU,0x0791U,
		0x6b96U,0x5a3fU,0x1dd0U,0x0503U,0x6801U,0x6f87U,0x2e51U,0x2ae0U,0xce98U,0x7dfaU,
		0xa6abU,0x5fb0U,0xf904U,0x181cU,0x218eU,0x2d98U,0x95d1U,0x3918U,0x2bd3U,0xe389U,
		0xd2e6U,0x5fbaU,0x36cdU,0x9969U,0xc738U,0x316aU,0xe20bU,0x004eU,0x715aU,0x29c3U,
		0x1d98U,0x4bb5U,0x28b4U,0xaf64U,0x67d2U,0xf5f3U,0x0a80U,0x630fU,0xd19eU,0x39d2U,
		0x723eU,0x6c55U,0x36beU,0xd7c1U,0x6e40U,0x4478U,0x5aaaU,0x8e19U,0xabc4U,0x40e1U,
		0x4a48U,0xc078U,0x7e2bU,0x947eU,0xb43aU,0x4e36U,0x3201U,0x4822U,0x3fafU,0xa199U,
		0xf728U,0x0204U,0xa1f3U,0x64b7U,0x33cfU,0x1b1dU,0x20afU,0x7317U,0x1598U,0x1af7U,
		0xed38U,0x82efU,0xddffU,0x404fU,0x85acU,0x0716U,0xf217U,0x7587U,0xc36dU,0xe201U,
		0x7c74U,0xf54cU,0xba77U,0x6df7U,0xf425U,0x5b09U,0x8691U,0x0fd4U,0x4c8eU,0x527bU,
		0x2664U,0x2926U,0x2d0cU,0xb1f6U,0x7880U,0x70dbU,0xd92dU,0xe87cU,0x153bU,0x619bU,
		0x8168U,0x8129U,0xf77fU,0xd013U,0x4c80U,0x9208U,0xa75aU,0x9bd1U,0x074eU,0x0d57U,
		0x4aa5U,0x40a5U,0x3139U,0x87d3U,0xc5bcU,0x7abdU,0x2404U,0xeab1U,0xd7bbU,0xcc49U,
		0xa82eU,0xd39cU,0xabe6U,0x01e8U,0x5289U,0x9bcbU,0xacc3U,0xb311U,0xc9e2U,0xf7a2U,
		0xf996U,0x1ad3U,0x1c52U,0x9dd9U,0x893eU,0x13f8U,0x8801U,0x3063U,0x1fb8U,0x9693U,
		0x5e9cU,0xa536U,0xc439U,0xcdf5U,0x2b01U,0x92e2U,0xc0fdU,0x9cb4U,0x8212U,0x3f18U,
		0xea5eU,0xc955U,0x89dcU,0x20c3U,0xdf36U,0x1976U,0xf711U,0xa526U,0x913cU,0xe07aU,
		0x6419U,0xb640U,0xd2a6U,0x388fU,0xe103U,0xfb96U,0xdbd6U,0xb547U,0x3e63U,0x9707U,
		0x635eU,0x528fU,0x07bbU,0x1fdaU,0xb1beU,0x8dfcU,0x894cU,0x29a8U,0xc22cU,0xc24dU,
		0x7490U,0xe975U,0xd7f7U,0x2ee9U,0x67b6U,0x04e5U,0x4d03U,0x6dc9U,0xa81bU,0x15ecU,
		0x5ddaU,0x5e44U,0xec8cU,0x614dU,0xae29U,0x826eU,0xcc57U,0xb8afU,0x9567U,0x24fdU,
		0x0bb8U,0x4561U,0x5196U,0xfa2eU,0xc845U,0x7108U,0x2b62U,0x4fe3U,0xbe85U,0x6370U,
		0x4b42U,0xbd08U,0x793aU,0xde4dU,0x1e4aU,0x611fU,0x0f4bU,0x3622U,0x74ffU,0x7df0U,
		0xe4ecU,0x07cbU,0x310aU,0x3eedU,0x3361U,0x88a9U,0xccc2U,0xa4bdU,0x581eU,0x282aU,
		0xb8c3U,0xdf61U,0xb218U,0xe2e4U,0x3840U,0x01eeU,0x3f26U,0xb340U,0xe59bU,0xea27U,
		0x8fd7U,0x1d78U,0xfe5fU,0x3ad7U,0x85d7U,0x8aa5U,0xab14U,0xde8eU,0x3182U,0x2dc5U,
		0xb80bU,0x6145U,0x944cU,0x1143U,0x4cafU,0x278eU,0xde5eU,0x3674U,0x4bf8U,0x4ac5U,
		0xda5dU,0x445eU,0x7829U,0x43a1U,0xe111U,0x01b9U,0x00e5U,0xc670U,0x4f77U,0xa43dU,
		0x6b2aU,0x225bU,0xc574U,0x06a1U,0x190fU,0x2094U,0xc25bU,0x39abU,0x6db1U,0x89b0U,
		0xc09fU,0xf449U,0xe937U,0x15bbU,0xda35U,0xdf91U,0x6cfeU,0x2184U,0x473cU,0x269eU,
		0x2b42U,0xbdedU,0xfc9bU,0x5d54U,0x8d0cU,0x67bfU,0x63c0U,0xf582U,0xf9feU,0xabecU,
		0xcd05U,0xdfc4U,0x27f8U,0x6febU,0xae40U,0x5892U,0xc4bfU,0x910fU,0x0982U,0x91d8U,
		0xabcdU,0x64d9U,0x766bU,0x1deaU,0xc3acU,0x11b3U,0xa95bU,0x794dU,0xf1ecU,0x587cU,
		0xae37U,0xa0c3U,0x8152U,0xb27dU,0x0188U,0x6330U,0x9432U,0x2009U,0x148bU,0x3398U,
		0x5a5fU,0x16fcU,0xadfaU,0x5254U,0x6a44U,0xc18fU,0xfa53U,0x6567U,0xecdfU,0x1162U,
		0x4226U,0x8caaU,0x8f55U,0x2120U,0x91b2U,0x8ff0U,0xe9e8U,0xee16U,0x6b5eU,0xb45eU,
		0x5e66U,0x9e15U,0xb354U,0x1e7bU,0x10baU,0xa47fU,0x279cU,0xad91U,0x585eU,0xa7e9U,
		0x0319U,0x294bU,0x377dU,0xf7a8U,0xeceaU,0xba20U,0xd4e5U,0x8242U,0xdae2U,0x2c7fU,
		0x3da4U,0x774fU,0xa474U,0xe2a7U,0x3243U,0xc799U,0x9c2cU,0xfdceU,0x9ac6U,0x4259U,
		0x7dbaU,0x1e1cU,0xbc5eU,0x3488U,0x09a6U,0xe228U,0x67faU,0x8c92U,0xe713U,0x3d7aU,
		0xa6c6U,0xe70cU,0xda56U,0x1c4bU,0x19ddU,0x23d7U,0xd73aU,0x5ffeU,0x05bbU,0x646eU,
		0x6b38U,0x8659U,0x2281U,0x2328U,0x1546U,0xd1bdU,0x5e5bU,0xf598U,0xf9aeU,0xb887U,
		0x4876U,0x1c5dU,0x88b7U,0xc6c4U,0xe9aaU,0x17ecU,0x9a14U,0xd437U,0x59b3U,0xb9c6U,
		0xfef1U,0xe3acU,0xdbe1U,0x5a00U,0x7a04U,0xfd1aU,0x8e81U,0xc9e8U,0xcb0dU,0x6dd9U,
		0x22b7U,0x4430U,0x34ffU,0x0bc3U,0x0f91U,0xbfc4U,0xbc96U,0x2d7dU,0x542bU,0x4f82U,
		0x8fe5U,0x5c5bU,0x667cU,0x16b7U,0x504fU,0x62f6U,0xaff7U,0xb678U,0x384bU,0xb813U,
		0xfd6dU,0xff04U,0x3b48U,0xfa5eU,0x9e9bU,0xfc05U,0x39a2U,0xc13aU,0x2803U,0xbe2cU,
		0x5452U,0xfaadU,0xf760U,0xfb43U,0x3cb2U,0xb4e8U,0xe77fU,0x1a86U,0x164fU,0x17bdU,
		0x1dfcU,0x15d5U,0x685dU,0xefe5U,0xda4bU,0x2cccU,0xe6eeU,0x71cfU,0x5da6U,0x65c9U,
		0x5a6dU,0x4a2eU,0xa1feU,0xb842U,0xb07cU,0xce4eU,0x54b6U,0x74abU,0x331dU,0xd5f9U,
		0x7a82U,0xb317U,0x1c85U,0x28bfU,0x7b21U,0x2b8cU,0x3b1fU,0x65b0U,0x86f6U,0x3e14U,
		0x11c1U,0xb0d5U,0x49ccU,0xbbd7U,0xdc10U,0xaeadU,0xa120U,0xc60aU,0x6463U,0xe7d3U,
		0x7c66U,0xb716U,0x6a5fU,0xe8acU,0x0c28U,0x6fbdU,0xa6cdU,0x7ef3U,0xbaf5U,0xd3bfU,
		0x483bU,0x5efaU,0x5423U,0xd7efU,0xeeceU,0xb143U,0xeee1U,0x0a72U,0xe292U,0x012bU,
		0xf295U,0x896eU,0x23acU,0xa0d7U,0x3b92U,0xacabU,0x1865U,0xa92dU,0xa4a3U,0x8a33U,
		0xc65bU,0x056dU,0x3d99U,0x90aaU,0x582fU,0x3c06U,0x97e2U,0xf31fU,0x09d8U,0x06d2U,
		0x1842U,0xcfe6U,0xcd13U,0x8b81U,0xf2b0U,0xacc9U,0xa5eaU,0x69bbU,0x1285U,0xf2dcU,
		0x21e4U,0x41dfU,0xbf1eU,0x6ae3U,0x7351U,0x351aU,0x7063U,0x85beU,0x4896U,0xb19bU,
		0xbf84U,0xac54U,0xbaa3U,0x51a6U,0x13e7U,0x7a3dU,0x1e11U,0x901bU,0x9552U,0x7a99U,
		0x44c9U,0x1431U,0xf4caU,0xe278U,0xc69aU,0x5942U,0x73d7U,0xb5ffU,0xdf9fU,0x4cd0U,
		0xef7aU,0x59efU,0xe92eU,0xf7d7U,0xebaeU,0xfa6bU,0x1b33U,0x7ab0U,0x3a69U,0x1798U,
		0x3aceU,0xef1fU,0xc3e2U,0x8f67U,0x3926U,0x130fU,0xf3aeU,0xcee5U,0x3ec9U,0x9b20U,
		0xec38U,0x9290U,0x3056U,0x0905U,0xf3e6U,0x28d0U,0x889eU,0xbca3U,0x2ce9U,0x9ddfU,
		0x8dddU,0xde9eU,0xc5d1U,0x8764U,0x786bU,0x4664U,0x6c6aU,0x70ceU,0x248aU,0x677aU,
		0x207fU,0xd554U,0x6ba4U,0xc540U,0xe419U,0xa8e5U,0xd8eaU,0xa1dbU,0x72e2U,0xb9d3U,
		0x9e7bU,0xd2baU,0x0d4aU,0x56a0U,0x6005U,0x8718U,0x1e3fU,0x7d1cU,0xf9e6U,0x48cbU,
		0xee66U,0xcc81U,0x9a60U,0x228eU,0x4378U,0x5f8cU,0xff2cU,0xe25cU,0x1715U,0x2c65U,
		0x51faU,0xf8adU,0xf7efU,0xd493U,0x1b0eU,0xfa21U,0xc85fU,0xd61dU,0xe30bU,0xed0aU,
		0xa194U,0x0049U,0x50daU,0x9c39U,0x6dc4U,0x1971U,0x01e3U,0x2921U,0x11aeU,0x15b6U,
		0x6140U,0x7debU,0x16b2U,0x1c58U,0x4254U,0xee11U,0xac4fU,0x8a2eU,0xc605U,0x2cc7U,
		0xd529U,0xd618U,0xa8e0U,0x8f62U,0x0133U,0xb53aU,0x2249U,0x1a5bU,0x6076U,0xd52eU,
		0x3308U,0x71b8U,0x2ed0U,0x9ec9U,0x1882U,0x367cU,0x5813U,0xceccU,0x55b1U,0xf9c8U,
		0x3f0bU,0xa7f6U,0x6184U,0x343fU,0x8733U,0x1b4aU,0xa770U,0x9cf0U,0xf97cU,0x2762U,
		0x175bU,0x4dc7U,0x418fU,0xe353U,0x8e9eU,0x6ee1U,0x6498U,0x977dU,0xcd33U,0x26f7U,
		0x65fcU,0x1a60U,0xb631U,0xa4b0U,0x5f2eU,0x76e5U,0x9443U,0x8358U,0xc301U,0x622fU,
		0x448bU,0x9fa1U,0x3fd7U,0xa686U,0xf736U,0xa5a7U,0xc279U,0x1ad9U,0x3615U,0x91f9U,
		0x085aU,0xacd6U,0x8a9aU,0x6950U,0x4968U,0x17a1U,0xfcefU,0xc350U,0x18d6U,0x29d5U,
		0x3006U,0xbab0U,0x6558U,0x224eU,0x76eeU,0x7839U,0xe89bU,0xf892U,0x220aU,0xd81bU,
		0xf3c6U,0x4cddU,0xd741U,0x8a58U,0x5d0dU,0xb6e9U,0x86a8U,0x0feaU,0x1136U,0x2db6U,
		0x7595U,0x4772U,0xb800U,0x268dU,0xf09cU,0x70a7U,0x577fU,0xf8deU,0x592aU,0xfc0eU,
		0x8d1eU,0x2b99U,0x46f2U,0x0455U,0x59f8U,0xb53fU,0x73f0U,0xf012U,0x82daU,0x7badU,
		0xe219U,0x15aeU,0x7b5bU,0x0763U,0x8c57U,0x5db3U,0xa8bdU,0x2ef2U,0x1678U,0x53a1U,
		0xc2a3U,0xf99cU,0x4396U,0x62b1U,0x37fcU,0x5729U,0x5885U,0xc12bU,0x02bfU,0x6bd7U,
		0xd7cfU,0x9f49U,0x8e22U,0x609aU,0x982eU,0xa2adU,0x882fU,0x0138U,0x3abcU,0x95aaU,
		0x65a1U,0x64ccU,0x5407U,0x7401U,0x7f72U,0xe3faU,0x8c3dU,0x2a00U,0x2968U,0x2f27U,
		0xf61fU,0xaaa0U,0xc569U,0xa70cU,0xb5a8U,0xb3b5U,0x6323U,0xa91eU,0xce70U,0x6a6cU,
		0x245bU,0xf7c3U,0xfd0dU,0xf78cU,0x9890U,0xd3ffU,0xf577U,0xef8dU,0x061dU,0x71bdU,
		0x0547U,0x147bU,0x6620U,0x2f0fU,0x03cdU,0x3eaeU,0x19adU,0x6c93U,0x3e9fU,0x231bU,
		0x5bdaU,0x0fe1U,0x01b0U,0xf27cU,0x4ee8U,0x3069U,0x89ccU,0x52dfU,0x6439U,0x162bU,
		0x096cU,0x1b79U,0xfaa0U,0xe508U,0x1328U,0xcd97U,0x27edU,0x1a24U,0x35a3U,0x926dU,
		0x5229U,0x330dU,0x50c7U,0x79d4U,0x9be8U,0x92f8U,0x2d24U,0x880eU,0x8af1U,0xb66bU,
		0x23deU,0x5c8cU,0xab40U,0xcb4eU,0xa886U,0xea0fU,0x9c1cU,0xd3a9U,0xf000U,0x28a1U,
		0xb6c2U,0x35c2U,0x027cU,0x15feU,0xb1dcU,0x1f2eU,0x2089U,0xc804U,0x133cU,0x93b5U,
		0x1e6eU,0x48bcU,0xf1d1U,0xd533U,0x9f10U,0xeca1U,0xa458U,0x9cc1U,0xcba6U,0x3948U,
		0x5a38U,0x3ce7U,0x38d9U,0xd0eaU,0xe8a5U,0xe221U,0x366dU,0xf79bU,0x0db7U,0x8807U,
		0xd814U,0xd2b3U,0xfac7U,0xe6beU,0xf238U,0x8839U,0xda2aU,0xe982U,0xa27bU,0xd6d2U,
		0xe29aU,0x26e8U,0x7033U,0x20ddU,0x0c84U,0x607bU,0x7228U,0xb476U,0x1ff6U,0xa0aaU,
		0xab4aU,0xde66U,0x26deU,0x9b4fU,0x469cU,0xa2daU,0xc11cU,0xe6ffU,0xa165U,0xcc64U,
		0x6d12U,0x7d4dU,0xea91U,0x3632U,0x31f1U,0x3229U,0x81baU,0x5ec1U,0x5c77U,0xf49dU,
		0xcca3U,0xaafdU,0x1fa5U,0x8adfU,0xdbebU,0x9d44U,0x2586U,0x3681U,0xb18eU,0x5bfdU,
		0xe1b3U,0x10c4U,0x5249U,0x62c2U,0x128fU,0x7d01U,0xfdacU,0x007eU,0x6268U,0x55f8U,
		0xa3ceU,0x3a86U,0x39e8U,0x13feU,0xc2d2U,0x7a8cU,0x8d5bU,0xb7f7U,0x6b21U,0xc581U,
		0xc15eU,0xba81U,0x554dU,0x07f2U,0xa024U,0x3efaU,0xc333U,0xf93fU,0x0560U,0x1887U,
		0xea68U,0x0b17U,0x8147U,0x9f78U,0x37a0U,0x0dbeU,0x0f05U,0x1c22U,0x19e4U,0xe5a8U,
		0x5bc9U,0x00f9U,0xae17U,0x7509U,0x25ceU,0xf2cfU,0xe826U,0x9ae2U,0x6d08U,0xf515U,
		0x08f1U,0x8b00U,0x43beU,0xd253U,0x6eceU,0xa6b5U,0x7634U,0xc51bU,0xc913U,0xae07U,
		0xc377U,0x9eceU,0x5871U,0xb228U,0x1997U,0x4857U,0x011eU,0xeda0U,0x69f4U,0x795aU,
		0xdbc5U,0x509aU,0xb5bbU,0xd666U,0xf567U,0xb510U,0x34f7U,0x8944U,0xbdaaU,0xe17dU,
		0x33e6U,0x7970U,0xf230U,0xc5fdU,0x241aU,0xed27U,0xe748U,0x467bU,0xc7e4U,0x1668U,
		0xa390U,0xe517U,0x7660U,0x2ed5U,0x3656U,0xfa0bU,0xce41U,0xb2e6U,0x64eaU,0x22e8U,
		0xf0bbU,0x3d43U,0xb155U,0xc8f9U,0x2239U,0xd98dU,0xb61bU,0x82c3U,0xa2d0U,0xbd92U,
		0xfe15U,0x7c02U,0x4b12U,0x15c8U,0xe83bU,0x843cU,0x4361U,0x8c27U,0x327dU,0xa671U,
		0x89c2U,0x2da5U,0x5523U,0x0c99U,0x4c64U,0xf9cdU,0x63a3U,0x79e6U,0x0b9dU,0x1d55U,
		0x9a48U,0xd174U,0x6c30U,0x47e1U,0x471aU,0xdb77U,0x8933U,0x9406U,0xba91U,0x42f2U,
		0xeac8U,0xa53cU,0x9220U,0xae89U,0xe246U,0xc31cU,0x70f4U,0x4b21U,0xcbffU,0x5ae7U,
		0xd4a6U,0x8bd3U,0x25e4U,0x76dcU,0x3ee4U,0xe048U,0x83d6U,0x55b6U,0x98daU,0x24b7U,
		0x739aU,0x05d5U,0xf118U,0x8840U,0x5c28U,0x9a99U,0x8660U,0x94edU,0xe0ebU,0x5f1bU,
		0x92d7U,0xafa5U,0x4796U,0x169dU,0xb7d0U,0x0c6eU,0xe7e6U,0x576bU,0x20fdU,0x56dfU,
		0x6c89U,0xa9acU,0xbb27U,0x7200U,0xeeb0U,0xfe42U,0x9d11U,0xb2f5U,0x3f5bU,0xced1U,
		0x966cU,0x3fbcU,0xd4f9U,0x6eb9U,0x63d1U,0x32a4U,0x23f9U,0x950aU,0x7e6bU,0x9796U,
		0x5573U,0x7bb5U,0x97bbU,0x0527U,0xc01eU,0x5ea2U,0x08c6U,0x14b7U,0x32baU,0x0723U,
		0x7978U,0x0331U,0x5323U,0x21a4U,0x6a0bU,0xb5eeU,0x0a7aU,0x6e52U,0xca8bU,0x7167U,
		0xfc95U,0x5818U,0xe0fdU,0x8162U,0x09d2U,0xb076U,0xfeebU,0x2796U,0x8c51U,0x4485U,
		0x01ddU,0x3050U,0xdbbfU,0xfda6U,0x38d3U,0x3e99U,0x854eU,0xe059U,0x7e65U,0x4714U,
		0xe02bU,0xed7cU,0x512fU,0x0f60U,0x13c2U,0x6690U,0xbc4aU,0x0e14U,0x6903U,0xec04U,
		0x6e20U,0xa3faU,0xca0dU,0x3444U,0x18bcU,0x8785U,0x6616U,0x25edU,0x4f0fU,0xd226U,
		0x42c4U,0x7e81U,0xe05fU,0x807fU,0x5e7fU,0x3343U,0xc210U,0x7f36U,0x6cd3U,0x9699U,
		0x3cd5U,0xade7U,0xa74fU,0x6b80U,0x7271U,0x0c59U,0x34c7U,0x3a72U,0x263bU,0xdd43U,
		0xfb05U,0x6abaU,0x0f15U,0xb96bU,0xa04dU,0x6189U,0x5f37U,0x60b2U,0x6e17U,0xcd2aU,
		0x0f42U,0xf23fU,0x5106U,0xd8beU,0x6b3fU,0x1f80U,0x9d55U,0x8b60U,0x2f49U,0x5be3U,
		0x0359U,0xd46aU,0xeae8U,0x5015U,0x8d64U,0xcd47U,0x811eU,0x46bfU,0xfb2bU,0xaa7dU,
		0x7fbfU,0x9ea4U,0xae5fU,0x962dU,0xb3c7U,0xf161U,0xef83U,0xa7fbU,0x38eeU,0xf860U,
		0xe376U,0x5e1aU,0x9094U,0xf180U,0x8e55U,0xcf5fU,0x8554U,0xecc9U,0x3e79U,0xd800U,
		0x9860U,0xd56bU,0xd922U,0x1fbeU,0x7472U,0x911fU,0x60ceU,0x2c52U,0xf1f5U,0xd039U,
		0x4522U,0x8061U,0xf843U,0xf696U,0x5ab3U,0x6821U,0xcbd1U,0x5acfU,0xd1b5U,0x3f10U,
		0xbee8U,0x1210U,0xb38fU,0x71f8U,0xc7fcU,0xf8a5U,0x0fafU,0xd1d0U,0xa087U,0x0b37U,
		0x96f1U,0x5b41U,0x6acfU,0xedafU,0xf340U,0xab8fU,0x0fc9U,0x63f5U,0x1d1cU,0xfc7cU,
		0xcf34U,0xdc9bU,0x180fU,0x0949U,0x117cU,0x8a0eU,0x5dcfU,0xfed7U,0x8498U,0xb019U,
		0xb7e7U,0x9cf5U,0x697aU,0xfcabU,0xbc85U,0xe16aU,0xdb8fU,0x4dafU,0x84f9U,0x0784U,
		0x0f66U,0x960bU,0xa87cU,0x8cf3U,0xc765U,0xca2cU,0x0e6eU,0x92e8U,0x800dU,0xe793U,
		0xa4dcU,0xf414U,0x68d4U,0x45cdU,0x4c42U,0x0cd8U,0x932bU,0x7797U,0x49f7U,0x1f51U,
		0x04daU,0x1896U,0xe94cU,0xa775U,0xa90eU,0xd116U,0x9fc3U,0xc4e7U,0xf45aU,0xe6c5U,
		0xba5cU,0x457aU,0x6475U,0x9e52U,0xaddeU,0x91f0U,0xb337U,0x0584U,0x4fc9U,0xce0fU,
		0x93ddU,0x0039U,0xb3f6U,0xf801U,0x1ceeU,0x0876U,0xc83aU,0x081bU,0x5cd7U,0xb397U,
		0x9471U,0xf6c2U,0xed60U,0xa413U,0xf832U,0x1b4fU,0x8d49U,0x0f7eU,0xeb3eU,0xee7dU,
		0x5df0U,0xa71dU,0x35b8U,0x9532U,0x5135U,0x2c23U,0x2a3fU,0x82e2U,0xd764U,0x0d91U,
		0x8ed7U,0x2b07U,0x450fU,0x99c0U,0x99d2U,0xeb4eU,0x33eeU,0x325eU,0xe49eU,0x427eU,
		0x8263U,0x45f3U,0xeee9U,0x87b4U,0xaf57U,0x98b7U,0xbeb9U,0x8738U,0x2479U,0x283dU,
		0xcf74U,0x4008U,0x575bU,0x4cb7U,0x7369U,0xdd2fU,0x0065U,0xc597U,0x49c5U,0x323cU,
		0xab0dU,0xd7b4U,0xa01dU,0x27e6U,0x18cfU,0x2078U,0x6a85U,0x49f0U,0xfafeU,0x25ddU,
		0x4c22U,0x90f8U,0x2b18U,0x8284U,0x92a8U,0x3ddfU,0x3966U,0x2e30U,0xd301U,0x4dccU,
		0x2fe4U,0x0eebU,0xc0f2U,0x4151U,0x32e9U,0xa7c3U,0x1033U,0x8797U,0xed82U,0x1573U,
		0x84c9U,0x4d68U,0x689eU,0x5c0cU,0x369bU,0xcdfbU,0xdcf8U,0x972bU,0xfa96U,0x644eU,
		0x9b6bU,0xdb3eU,0xdcd3U,0x1225U,0xc46fU,0xb703U,0x6a8cU,0x0298U,0x711bU,0x5c46U,
		0x7334U,0x1760U,0x30eeU,0xd877U,0x1bf6U,0x2a98U,0xa51bU,0xfaceU,0x2a62U,0xd84cU,
		0x05c2U,0x4ba5U,0xfea4U,0xa131U,0x6936U,0x6a1eU,0x560bU,0x7196U,0xd69aU,0x39bbU,
		0x42bbU,0xc2f8U,0xccb9U,0x8bf8U,0x51d2U,0x19c5U,0x4e03U,0xb98eU,0x6fdbU,0x0396U,
		0xfdbcU,0x58d0U,0x3e58U,0x2767U,0x0be5U,0x905aU,0xaa63U,0x397fU,0x0aefU,0xc2e1U,
		0x302bU,0xee35U,0xe031U,0xc17aU,0xa3dfU,0x5e34U,0xf4deU,0x4810U,0x79caU,0xc43fU,
		0x8518U,0x769fU,0x100bU,0xcd79U,0x95eeU,0x3d21U,0x9909U,0x498aU,0xaa33U,0x603fU,
		0xb959U,0x019dU,0x0a62U,0x5499U,0x2304U,0xf981U,0x8df1U,0x2527U,0xb6a0U,0xeb5eU,
		0x77f7U,0x5658U,0x627aU,0x77ddU,0xbfefU,0x271eU,0xeb72U,0x1a49U,0xe854U,0x9e05U,
		0x486eU,0xdf2eU,0x4a60U,0x5a20U,0xf82aU,0x3f53U,0xf1c9U,0xc857U,0x000fU,0x8a12U,
		0xb384U,0x3b81U,0xb9e5U,0x0a25U,0xe4baU,0x010eU,0x0a29U,0x17c2U,0x34ebU,0xa0e9U,
		0x94b8U,0x1bdbU,0x0c1bU,0xf851U,0x2435U,0x1180U,0xab03U,0x3998U,0x799eU,0xe8d8U,
		0x81aaU,0xd024U,0xb9e9U,0x5fc0U,0xaa0fU,0x508eU,0xad2dU,0x61d4U,0x8070U,0xdc03U,
		0x5ac4U,0x094dU,0x458dU,0x0b45U,0x0e1bU,0x21faU,0x6beeU,0xac7fU,0x0112U,0x1654U,
		0xcc1cU,0x81d5U,0x7b14U,0x4705U,0xbe4cU,0xc357U,0xb13aU,0x1813U,0x39b2U,0x9ad9U,
		0x785bU,0xb4f7U,0x936cU,0xad19U,0xe4beU,0xc50fU,0x6204U,0xbb3cU,0x8bc4U,0xe6e1U,
		0x4facU,0x2e8aU,0x138dU,0xb01dU,0xa5d0U,0xdb7fU,0x8599U,0x7b49U,0x0fbeU,0xee9fU,
		0x8a16U,0x1a8bU,0xbf77U,0xd868U,0x3196U,0x48eaU,0x222dU,0xc550U,0x1bafU,0x849cU,
		0xcca9U,0x845aU,0x6c1eU,0x8720U,0x49e6U,0xfb39U,0x0013U,0xd2ecU,0x270fU,0xe921U,
		0x27b3U,0x5215U,0xcd9fU,0x8430U,0x852eU,0xfedbU,0x57a9U,0xf6adU,0xd1c5U,0x6487U,
		0x3817U,0x8f34U,0x3b85U,0xe784U,0xc716U,0x85faU,0xe73cU,0x385aU,0x46d3U,0xfb9eU,
		0xf1e5U,0x5dd3U,0x13e0U,0xaff0U,0xc7d3U,0x8335U,0xa1ecU,0x109fU,0xb388U,0x09cbU,
		0xb699U,0xcf6dU,0x659aU,0x2ec9U,0xce3aU,0x1fefU,0x0dddU,0xfcafU,0x5c6bU,0x52b3U,
		0x61c3U,0x443fU,0xecd2U,0x1300U,0x1bdfU,0xefeaU,0xfb20U,0xceb7U,0x5463U,0x14c0U,
		0x407eU,0xff99U,0xb997U,0x697eU,0xf4a3U,0x5bf1U,0x82adU,0x92c0U,0x2912U,0x030cU,
		0x94bcU,0xe38fU,0xe494U,0x0f87U,0x7fdbU,0xd0a5U,0xf19eU,0xa1bfU,0xa3c2U,0x9cf9U,
		0x5d26U,0xf874U,0xbc51U,0x4763U,0x0166U,0x2609U,0xf855U,0x6862U,0x2a09U,0x8fb6U,
		0x8b40U,0x58e6U,0x9eb3U,0xfcf6U,0xd1edU,0xb7ebU,0xc262U,0x4952U,0x9f3aU,0xd72dU,
		0xaf3dU,0xb47fU,0x0c1fU,0x1dc7U,0x2a59U,0x0efcU,0x4684U,0xb6d6U,0xdc5fU,0x0f23U,
		0x665aU,0x4db3U,0x8e74U,0xb467U,0x0476U,0x358eU,0x813bU,0x84b3U,0x17c6U,0x15daU,
		0x62d7U,0x3351U,0x9c02U,0xfb75U,0x9855U,0x3c39U,0xec1cU,0xdb93U,0x5c7dU,0x9a07U,
		0xab9fU,0x0d60U,0xc77cU,0xf933U,0x0a2dU,0x2bd9U,0x2af1U,0x5038U,0xc05fU,0xe9b9U,
		0x2833U,0x62ecU,0x9e8eU,0xe16eU,0x6d9dU,0xd3dfU,0xe81aU,0x78beU,0x862aU,0x6911U,
		0xa0edU,0x1e01U,0x32d6U,0x2515U,0x2c47U,0x134dU,0x0ebaU,0xc452U,0xaaeeU,0xbc89U,
		0x66d6U,0xbbb5U,0xddb0U,0x1c16U,0x8563U,0xec52U,0x34efU,0xc29bU,0x79c2U,0xd91aU,
		0x36b6U,0x9a32U,0xf2a8U,0xf9f6U,0xd9c9U,0x960fU,0x4505U,0xcb03U,0x38feU,0x4bc9U,
		0xd336U,0xedb9U,0xe8dcU,0x71d4U,0x4653U,0xd6c6U,0x1f8aU,0x7dc9U,0xad82U,0xc092U,
		0x0e90U,0x0f6aU,0x9d4aU,0xddd4U,0x7e40U,0x99ffU,0x9a71U,0xa325U,0x79a2U,0x3170U,
		0x0c78U,0x5eb0U,0xd5c9U,0xbe66U,0x2b35U,0xc6b5U,0xb70dU,0x0788U,0x2a8fU,0x9f6fU,
		0x6697U,0x9800U,0x9d64U,0xb5cfU,0xd028U,0xe6f3U,0xa37eU,0x54a3U,0xb0c6U,0xa94eU,
		0x41c8U,0x17a8U,0x46b4U,0x84fdU,0x209bU,0xc93eU,0xe6a5U,0x57c6U,0xd120U,0x3f69U,
		0x81aeU,0x8be7U,0x8920U,0x04c8U,0x6a37U,0x4cc1U,0x35f3U,0x4eaeU,0x5b52U,0xca30U,
		0xdea8U,0x4722U,0x75b1U,0xdc74U,0x9528U,0xb9bcU,0x1184U,0x2cd1U,0xb8a0U,0x8fcaU,
		0xfcc2U,0xc1f3U,0x5598U,0x393cU,0x7c17U,0xc769U,0xdbf1U,0x0735U,0x9fabU,0x600dU,
		0x2630U,0x7562U,0x2439U,0xc73eU,0x74f2U,0x9bbcU,0x6f99U,0x31adU,0x1330U,0x53abU,
		0x0dfdU,0x8cf7U,0xddf9U,0xce92U,0x154eU,0x3952U,0x54e6U,0x2d42U,0x399cU,0xda50U,
		0xc694U,0x4835U,0x0bb2U,0x87fbU,0x1485U,0xe10bU,0xa15fU,0xa880U,0xb615U,0xae11U,
		0x617eU,0xf9e0U,0xf619U,0x86a2U,0xab07U,0xb331U,0xe84eU,0x6930U,0x38cdU,0x92d1U,
		0x6ac9U,0x2f43U,0x78f8U,0xe797U,0xbcb9U,0x2895U,0x2c40U,0xe735U,0x6300U,0x0babU,
		0x61d8U,0x4a33U,0xeb34U,0xe9a0U,0x88dbU,0xe232U,0xda89U,0x1082U,0x3524U,0x8011U,
		0x8ae5U,0x9eeaU,0xe0c9U,0x8752U,0xc21fU,0xb4b7U,0xad31U,0x996fU,0x6cc8U,0xaeb7U,
		0x0b80U,0xd268U,0x4163U,0xf3f6U,0xa659U,0x92ecU,0xf360U,0x0439U,0x13c9U,0x07aeU,
		0x441eU,0xab6dU,0xdc07U,0x5a72U,0x7112U,0xc32aU,0xb0a4U,0x31b6U,0x7bcfU,0x496fU,
		0x3597U,0x0e72U,0x1916U,0x7d74U,0x912fU,0x7010U,0x2f62U,0x01c3U,0x8074U,0x507bU,
		0x7381U,0x48f3U,0x633aU,0xe9f7U,0xe633U,0x7079U,0x89f5U,0x45d1U,0x66beU,0xc9d5U,
		0xbd20U,0x7552U,0x1e92U,0xf270U,0x5fc4U,0x65ceU,0xaf1bU,0x9c84U,0x562cU,0x72c8U,
		0xa409U,0xb87dU,0x77bfU,0x68d8U,0x1fabU,0x7e48U,0xacf9U,0x4415U,0x642dU,0xb038U,
		0xb9edU,0x36d3U,0x0d26U,0xdd09U,0xdb01U,0xd138U,0x4f04U,0xf126U,0x3e35U,0xf418U,
		0xf5a8U,0x6cadU,0x5b8dU,0xf7b7U,0x9930U,0xad09U,0x5092U,0x5dabU,0xc172U,0xecc1U,
		0x6307U,0x5b7bU,0x06caU,0x5d4cU,0xf772U,0xa4e0U,0xa4f9U,0x56edU,0x053bU,0xedddU,
		0x12f0U,0xf073U,0xaa13U,0xcf4bU,0x991fU,0x5361U,0xebf9U,0xcab1U,0xebdbU,0x449dU,
		0x051cU,0x779bU,0x089eU,0x4de2U,0xde7eU,0x682cU,0xf437U,0xde33U,0x21feU,0xce53U,
		0x16deU,0x3dedU,0x4b7dU,0x2732U,0xf16cU,0xef5fU,0xd5f0U,0x932fU,0x5c03U,0x3a7dU,
		0xcd57U,0xd003U,0x916cU,0xdf0fU,0x0e1fU,0x29c9U,0x9349U,0xbb88U,0xc1a4U,0x3ba1U,
		0x255eU,0x0a07U,0x3748U,0x0cdcU,0x32afU,0xa9d9U,0xa401U,0x984dU,0xac9bU,0x84d7U,
		0xac83U,0xb081U,0x112aU,0x66f7U,0x356cU,0x7044U,0x53dbU,0xbab7U,0x2debU,0x4c46U,
		0xf450U,0x099cU,0x79f1U,0x9d32U,0x08d5U,0xfbf5U,0x6bf2U,0x04adU,0x7c4cU,0x8a4cU,
		0x4aeeU,0x5ca3U,0xcadcU,0x55c1U,0x157eU,0x189aU,0xe2b7U,0x47e9U,0xa9a1U,0xcff5U,
		0x65dfU,0x1926U,0x0951U,0xb847U,0x5389U,0xd857U,0x8e92U,0xce28U,0x798bU,0x7af5U,
		0x1de3U,0x04deU,0xb194U,0x5c54U,0x049fU,0x56a8U,0x481bU,0x6027U,0x5ac8U,0x7160U,
		0x5492U,0x98b0U,0xa2a6U,0x1a54U,0xe510U,0x20d6U,0x8dbaU,0x1f55U,0xc41aU,0x1ca2U,
		0x2330U,0x4d8eU,0x9661U,0xce1aU,0x0b49U,0xa203U,0x009fU,0x857fU,0xe79eU,0xb4aeU,
		0x3fcbU,0x3897U,0x0689U,0x49fbU,0xb2d3U,0x2ad0U,0x3555U,0xc81fU,0x1f47U,0xc78fU,
		0x4591U,0x8ac3U,0x3269U,0x1771U,0x90ccU,0xe671U,0x5a95U,0x6944U,0xf613U,0xd11aU,
		0x12eaU,0x2f5cU,0xa1e6U,0x9366U,0x855dU,0xaf37U,0x4709U,0xd5feU,0xf96aU,0xa214U,
		0x1f41U,0x08cfU,0x4020U,0x8f72U,0x4e0cU,0xa912U,0x3687U,0x0640U,0x7731U,0xf20aU,
		0xc397U,0x3039U,0x7b18U,0x0054U,0xfa8dU,0x8d52U,0x9d7aU,0x4d58U,0xa982U,0xbf61U,
		0x14deU,0xa779U,0x02b3U,0xe35eU,0x6e27U,0x77a9U,0x2bc6U,0xd9aeU,0xc35bU,0x3322U,
		0x8c46U,0x4932U,0xcc34U,0x48a5U,0xfe32U,0x300dU,0x5712U,0xe950U,0xc0a6U,0x959eU,
		0xcb8cU,0x2ba4U,0x8f78U,0x7231U,0xbe50U,0x05a9U,0x4fbfU,0x19d3U,0xe751U,0x431dU,
		0xa2b8U,0x5b31U,0x2e44U,0xe6c9U,0x516bU,0x6b67U,0x438aU,0x33bbU,0x7ce2U,0x47afU,
		0x1658U,0x74b0U,0x5eccU,0x09eaU,0xb06bU,0xae76U,0x84dfU,0xbffdU,0x3c59U,0xf45eU,
		0x258cU,0xa32dU,0x96b4U,0x0757U,0x1a09U,0x3074U,0x0116U,0xe211U,0x0ae7U,0x908cU,
		0x28acU,0x0177U,0x5827U,0x4734U,0xb3d8U,0xc4ebU,0x3b33U,0x604dU,0xf474U,0x0b77U,
		0x82ceU,0xc8ddU,0x81d9U,0x54bbU,0x976dU,0x205dU,0x78ffU,0xd671U,0xca80U,0x0cb4U,
		0xe5b3U,0x9fc7U,0xe723U,0x56c5U,0x214fU,0x5951U,0x4026U,0x0449U,0xcc20U,0xc6e1U,
		0x5976U,0x7d0cU,0x172eU,0xa9e7U,0x9e48U,0x3d70U,0x1a7dU,0x9e56U,0x90efU,0xf494U,
		0xded6U,0xc1d8U,0xd5dfU,0xbe37U,0xb4fbU,0x28c4U,0x38bdU,0x6d4aU,0x6de4U,0x25b0U,
		0xd183U,0x5f4cU,0xd169U,0x6479U,0x62c8U,0x938bU,0x7f0cU,0xccf1U,0x6242U,0x4fffU,
		0x785fU,0xaf6aU,0x2c8aU,0x561cU,0x58acU,0x90e5U,0xbc04U,0x1cfcU,0x06ddU,0x457eU,
		0x1172U,0x421cU,0xec0bU,0xb851U,0xb9feU,0x3534U,0xad1dU,0x1c8aU,0x8d8cU,0x7aa4U,
		0x1c6eU,0x0cc2U,0x273cU,0x29dcU,0xb68aU,0xba60U,0x89b7U,0xe6d3U,0x754aU,0x3586U,
		0xd604U,0x33b3U,0x9370U,0x8b30U,0xc3d6U,0x00cbU,0xa34fU,0xd3caU,0x80a3U,0x6ef3U,
		0xbce6U,0x0588U,0xe887U,0x6c38U,0xc28dU,0x605fU,0x417aU,0xa4eaU,0x1817U,0xb31cU,
		0x7582U,0x4473U,0x701aU,0xcc75U,0x202aU,0x57e8U,0xa438U,0xb33bU,0x524fU,0xdfbfU,
		0xa531U,0x0d52U,0x7103U,0x1fd5U,0xb13eU,0x28baU,0x4ccbU,0x69b6U,0x5ff9U,0xba1bU,
		0xfaa8U,0x6dd4U,0x476dU,0x91f4U,0xb3b0U,0x62acU,0x2289U,0xde99U,0x9cebU,0xee0cU,
		0x9addU,0x7a87U,0x7bfdU,0xe178U,0x289cU,0x52daU,0x362dU,0xd2aeU,0x5010U,0xade2U,
		0x63f0U,0x911aU,0x0c69U,0xae84U,0x470fU,0x14b2U,0x39b6U,0x9726U,0x5a1bU,0x769aU,
		0x0034U,0xe78eU,0x2073U,0x99bbU,0xd57cU,0x003dU,0xd358U,0x284bU,0xb76eU,0xcfceU,
		0x7e05U,0xaed7U,0xe6e5U,0x65b5U,0x2fdbU,0xb185U,0x2abeU,0x41fbU,0xe09fU,0xadbfU,
		0x1b02U,0x93e1U,0x10caU,0xd50bU,0x27d1U,0x7f02U,0x7da4U,0xe7a7U,0x8bc8U,0x4bbbU,
		0x16cfU,0xf5feU,0xf47dU,0x4940U,0xda76U,0xe55aU,0xc8d6U,0xce13U,0x342cU,0xee05U,
		0x690aU,0x8f2dU,0xad02U,0x2d3bU,0x2e8eU,0x3b24U,0xa872U,0x5854U,0x4be2U,0x6fa2U,
		0xfe8fU,0x18ddU,0x047fU,0x4fcdU,0x6db8U,0x403cU,0x3c25U,0x423aU,0xa21aU,0xa83fU,
		0x4fb0U,0x40f1U,0x83cbU,0x319fU,0x203aU,0xb490U,0x05e4U,0xf4fdU,0xb370U,0x087aU,
		0x42d9U,0xd078U,0x4f1fU,0xd60cU,0xf13eU,0x8f02U,0xc513U,0x2b91U,0x038eU,0x9625U,
		0x9961U,0x11d0U,0xaca3U,0x89a8U,0x12c1U,0x1cf2U,0xe1b9U,0x9a79U,0x32fcU,0xe13aU,
		0x2cfaU,0xe14fU,0xe4c2U,0x1d9eU,0xf103U,0x1b5dU,0x5f10U,0x7002U,0x7beeU,0x4d1bU,
		0xf681U,0xf805U,0x9fdaU,0xfc2cU,0x6cebU,0x9efeU,0xe7f5U,0x3433U,0xbb40U,0x7b26U,
		0xd0cfU,0x45a2U,0xbcc0U,0xd349U,0x45c3U,0x2c75U,0x885aU,0xb3faU,0xd0fcU,0x8611U,
		0x2fd2U,0xeb2bU,0xf970U,0x8ee5U,0x6208U,0xd9f0U,0xb25eU,0x2344U,0xf0f4U,0x226dU,
		0x20f2U,0x5b6dU,0xad69U,0xb39bU,0xdd64U,0xbc26U,0xe5eaU,0x3c6bU,0xdcedU,0x72f0U,
		0x7b4dU,0xb0daU,0xfc88U,0x93a3U,0xfc23U,0x6ca4U,0xadd2U,0xd952U,0xf6d9U,0x5cdbU,
		0x0084U,0xf74cU,0xb569U,0xb928U,0x3975U,0x90a0U,0x859dU,0x6315U,0x3dcdU,0x5122U,
		0xd045U,0xd4c9U,0xfae0U,0x086aU,0xdcb7U,0x081fU,0x3a26U,0xc634U,0x878cU,0x4a10U,
		0xd476U,0x491fU,0xeea3U,0x11c6U,0x65eeU,0x5f9aU,0x931fU,0xa059U,0x4f98U,0x7840U,
		0xadf3U,0xc83eU,0xdf98U,0x544bU,0x97a6U,0xcabfU,0xf210U,0xefa0U,0x0fc2U,0x7e5eU,
		0x4a59U,0x18c8U,0xb5a1U,0xa769U,0xfe0eU,0xea8aU,0xaa4eU,0xa417U,0xc0e7U,0x42faU,
		0x0533U,0xdda8U,0x2fcaU,0x2147U,0xb021U,0x3e19U,0x0e62U,0x8e3cU,0xd51bU,0xe448U,
		0x5c1bU,0x9d1dU,0xcf04U,0xed64U,0xfdb2U,0x185bU,0x742cU,0x48d3U,0x21c0U,0x975cU,
		0x1391U,0x0a86U,0xbba1U,0x95ffU,0xdff9U,0xa186U,0x9275U,0x3ed7U,0x2e25U,0xf6c6U,
		0x90baU,0x3289U,0xb88fU,0x2935U,0xe644U,0x6fa9U,0xdb83U,0x86fbU,0x2d5fU,0x4294U,
		0xe239U,0xf134U,0xee58U,0x970fU,0x3511U,0x9475U,0x6036U,0x4672U,0x349dU,0xcc0bU,
		0x7737U,0x7b8eU,0xa5d4U,0xfecbU,0x0934U,0x6aa0U,0x02e3U,0x6396U,0x9aa5U,0xf3d8U,
		0x0ef7U,0x0f82U,0xd915U,0x5033U,0xbb37U,0x5089U,0xcf68U,0xe91cU,0x48eeU,0xaeb2U,
		0x535cU,0xdd04U,0x04c3U,0x5eabU,0x692bU,0x9bb7U,0x19ceU,0x8d4dU,0x7d07U,0x9087U,
		0x8a47U,0xbb83U,0x176cU,0x98abU,0x319aU,0xf5f9U,0x233fU,0x1b58U,0x00c6U,0x5617U,
		0x7695U,0x69b1U,0x96e4U,0x1b53U,0x7652U,0x97d3U,0x18c3U,0x511dU,0x6a9bU,0x95faU,
		0xc554U,0xdc15U,0xe403U,0x256fU,0xc2edU,0xa559U,0x8f98U,0x76f5U,0x80deU,0xf836U,
		0xda3cU,0x7626U,0xf08aU,0xdd23U,0x303fU,0x6084U,0x2231U,0x2960U,0xeb6aU,0x96e9U,
		0xed30U,0x3373U,0xf4c2U,0x942aU,0xd8cfU,0xa721U,0xc82eU,0xee1fU,0x244dU,0xba68U,
		0x13b5U,0x0421U,0x1a8fU,0xbbdcU,0xdcc8U,0x760bU,0xb9aaU,0xdabdU,0x0ce4U,0x964eU,
		0x2ca3U,0x5df4U,0x1295U,0x0f72U,0x634fU,0x9bdaU,0x7cc2U,0xc203U,0x8a1aU,0x67d8U,
		0x3021U,0x97d8U,0xa266U,0xaec5U,0xa4c6U,0xbf07U,0xd758U,0xee81U,0x87c0U,0x2819U,
		0x6e0aU,0x12b7U,0x0da9U,0x4be9U,0xd86cU,0x49d1U,0x59cdU,0xde12U,0x88e2U,0xafccU,
		0x040fU,0xe319U,0x2f7fU,0xeb42U,0x1420U,0xdb2bU,0xa41fU,0xeadbU,0xc39dU,0xc10eU,
		0xbf7bU,0x3fa6U,0x22fbU,0x7657U,0xd368U,0xdba4U,0xa140U,0xd770U,0x22ceU,0x2c27U,
		0x902bU,0x2496U,0x24deU,0xe089U,0x4980U,0x6ad9U,0x8724U,0xe7d8U,0x3407U,0x68c7U,
		0x6b49U,0x7570U,0x7763U,0xa7b7U,0x412dU,0x5139U,0x3a8cU,0xa5b5U,0x9afeU,0xebbaU,
		0x4df8U,0xffdbU,0x6c22U,0x6c5bU,0xb7daU,0xa55eU,0xeb22U,0xedd4U,0x929cU,0xf0e1U,
		0xc479U,0x9536U,0xe1fbU,0x2d92U,0x344bU,0x6104U,0x998aU,0x5908U,0xfb3dU,0x6468U,
		0x59e9U,0x0a6cU,0x4fddU,0xcdefU,0xc189U,0x2255U,0x3223U,0x35bcU,0x15c2U,0xf50fU,
		0x275cU,0x5f86U,0xa918U,0x2687U,0x49eaU,0xf7fbU,0x3f4dU,0xc2f2U,0xed76U,0x5765U,
		0xfc76U,0xcd41U,0x193aU,0x0d95U,0x8de7U,0xba99U,0x4cf3U,0xbdc9U,0x1ae5U,0x7257U,
		0x84a0U,0xc60fU,0x08adU,0xb6f7U,0x4ed9U,0x6608U,0x5fd9U,0xed53U,0x477eU,0xd768U,
		0xa3d4U,0xd798U,0x4495U,0xf9eeU,0x5b65U,0x3d68U,0x1bb3U,0x7244U,0x7328U,0x76faU,
		0x34b9U,0xb1cdU,0x35abU,0x1682U,0xe26fU,0x82e6U,0x1a40U,0xd984U,0xf9b6U,0xcbb0U,
		0xe9d8U,0x18e4U,0x845eU,0xa125U,0x5735U,0xbb57U,0x1089U,0x246cU,0x0551U,0x3e6bU,
		0x3d16U,0x2a43U,0x55d6U,0xc91dU,0x0269U,0x1f6eU,0x4e12U,0x5413U,0xccadU,0x155cU,
		0xd832U,0x8f9dU,0x6a78U,0xd0c5U,0xf332U,0xf0acU,0x034aU,0x99c4U,0x93c1U,0xa50eU,
		0x9b32U,0x309cU,0x5059U,0x4804U,0x5219U,0xe8b1U,0x8eeeU,0x6af5U,0x2653U,0x09b0U,
		0x9dfbU,0x5938U,0x735bU,0x4513U,0x55feU,0xbd34U,0x20e9U,0xebd2U,0x0fffU,0x38a0U,
		0x27b7U,0x39d8U,0x994aU,0xdc1aU,0xc8e6U,0x666bU,0x77ecU,0xa171U,0x34d8U,0x2b0bU,
		0x4c55U,0x5dc1U,0xca14U,0x19b9U,0xb041U,0x6e91U,0x8434U,0x6a64U,0xc84fU,0xdc93U,
		0xf544U,0x53b4U,0xf7cfU,0x655fU,0x84bcU,0x8edbU,0xe93eU,0x23eaU,0x0bd9U,0x615cU,
		0x0646U,0xe11bU,0xcda3U,0x6762U,0xb8f1U,0xc559U,0x0192U,0xdde2U,0x3a50U,0x6841U,
		0x596fU,0x3262U,0xb257U,0x5a14U,0x79bbU,0xb692U,0x9918U,0xe847U,0xd2f0U,0xb71bU,
		0xf820U,0xef70U,0x22f4U,0x092dU,0x3750U,0xd82bU,0xb581U,0x33f2U,0x626eU,0x0e98U,
		0x51c4U,0x4c13U,0x98fcU,0x02cdU,0x0017U,0xd1a4U,0x4863U,0x2574U,0x1ec8U,0x8ec7U,
		0x3bb8U,0x78a4U,0x0b23U,0xeb52U,0xaf01U,0xd95bU,0x5333U,0x850bU,0xbeaaU,0xfe96U,
		0xe925U,0x7c6bU,0xf1c0U,0xe832U,0xda90U,0xbe94U,0xc9fdU,0xeb8fU,0x0a56U,0x99d6U,
		0xe3c4U,0xfdf7U,0xaa56U,0xf0eaU,0x368dU,0xba4dU,0x2713U,0x9ffaU,0x3757U,0xe408U,
		0xac34U,0x39f4U,0x38a9U,0x4786U,0x8090U,0x45f7U,0xa9f6U,0xecafU,0x0ccdU,0xfd79U,
		0x1b8dU,0xd782U,0x648bU,0x7ef8U,0xdd75U,0x7f4dU,0xbb1bU,0x3dbdU,0xbfd0U,0x5f57U,
		0xb276U,0x8267U,0x7a92U,0xfefdU,0x6722U,0x8ab4U,0x7310U,0x71e4U,0xd1c9U,0x447eU,
		0x77d6U,0xdd28U,0xe3f3U,0xcec5U,0x3d3cU,0x9b48U,0xe5d4U,0x4282U,0xc7a5U,0xaee2U,
		0xd22dU,0x0836U,0xa903U,0xa62cU,0x8f38U,0xa6d2U,0x654bU,0xb74aU,0x899fU,0xb874U,
		0x5567U,0xd822U,0xd5a6U,0xe4a2U,0x26a5U,0x11f7U,0x8ffcU,0xd083U,0x4d5eU,0x74a1U,
		0x381bU,0x4e83U,0xb291U,0xf08fU,0xf14eU,0xbbc6U,0xf810U,0x0325U,0x0579U,0x98bbU,
		0x105aU,0x940eU,0x5533U,0x6870U,0x3089U,0x43e2U,0xfedfU,0x6fc2U,0x8aceU,0xa5f6U,
		0x37efU,0xda6cU,0xa69dU,0x86d6U,0x3bfdU,0xaf5bU,0xc2d8U,0xed97U,0xdb51U,0x7d24U,
		0xac60U,0x0a12U,0x8532U,0x6e46U,0x790eU,0x6089U,0x5caeU,0x73e3U,0x1a2cU,0xe1d0U,
		0xaef9U,0x87b8U,0xb3ebU,0xcbebU,0xf5a0U,0x6d95U,0x9fd2U,0x3b2bU,0xf6b1U,0x0c2dU,
		0x8542U,0x4affU,0x4a3aU,0xa338U,0x7f65U,0xb54fU,0x2bafU,0xeeedU,0x9b61U,0x3b15U,
		0xe7bbU,0x1255U,0x9d80U,0x6d20U,0x57adU,0x76cbU,0xb722U,0x3044U,0x7776U,0x1d8aU,
		0x3b65U,0x9821U,0x1e32U,0x2841U,0x7eb9U,0xb658U,0x9214U,0x6cb8U,0x8e5eU,0x7ad6U,
		0x385eU,0x5effU,0x9da6U,0xd686U,0xe15fU,0xc027U,0x45dcU,0x4c9aU,0x51dbU,0x247dU,
		0x1404U,0xab78U,0xa36eU,0x47d5U,0x41a2U,0xd27eU,0xe740U,0xd7c7U,0xaa2bU,0xf83bU,
		0x7e23U,0x1db8U,0x7349U,0x097aU,0x44b0U,0x873cU,0x2cdcU,0x821eU,0x4f16U,0xbd17U,
		0x0b91U,0x79b2U,0xfba2U,0x4840U,0x7f7bU,0xd5b8U,0x9e26U,0x3c47U,0x848dU,0x2211U,
		0xd449U,0xbebdU,0x4743U,0x3ebcU,0x7dd4U,0xc238U,0xbf67U,0x0c8dU,0x46d7U,0xfc40U,
		0x6589U,0x80e3U,0x2423U,0xc2b9U,0x827aU,0x17b3U,0x92cbU,0x4cbbU,0xcaabU,0xe9f1U,
		0x2ec3U,0xe6dbU,0x9a2cU,0xb6d0U,0xe788U,0xd3c4U,0x2267U,0xb48aU,0xe66bU,0x5c9dU,
		0xa9e1U,0x4317U,0xd0bfU,0x575fU,0x39eeU,0xdddcU,0x6390U,0xa763U,0xdb9eU,0x336dU,
		0x3b89U,0x36c4U,0x6fd2U,0x762bU,0x1d84U,0xbbc0U,0xcccbU,0xc2b3U,0x8319U,0x400cU,
		0xcbf3U,0x7c3cU,0xb0f2U,0xc9acU,0x1708U,0x2e95U,0x85feU,0xbafaU,0x84eeU,0x9306U,
		0x61dfU,0xc3b8U,0x433aU,0x00b5U,0xdfd0U,0xcf78U,0xf95aU,0x24abU,0x4302U,0x135bU,
		0xa988U,0x6133U,0xc71aU,0x3b02U,0xd2f7U,0xda41U,0x5d80U,0xccd1U,0xac23U,0xd7a0U,
		0xe615U,0xc59bU,0x58b6U,0xb97eU,0x1020U,0xb947U,0xbec8U,0x95ddU,0x8339U,0xb148U,
		0xb732U,0x7055U,0x5a4bU,0xab83U,0x94d7U,0x95baU,0x1366U,0x0069U,0xc587U,0xcf1eU,
		0x27dbU,0x1528U,0x810dU,0xf0c9U,0xc7d7U,0x40e7U,0xa495U,0x3378U,0xa07bU,0xd635U,
		0xc701U,0x12cbU,0x0210U,0xdd33U,0xcb26U,0x6d5eU,0x25f4U,0x095bU,0x718cU,0xd7e5U,
		0x10a3U,0xeed3U,0x4e8eU,0xf223U,0x37d4U,0x149aU,0x4b87U,0xf899U,0xdb67U,0x736dU,
		0x218bU,0x2e4eU,0x174aU,0xe4d5U,0x005aU,0xe1dbU,0xa1f0U,0xb437U,0xc36aU,0xed35U,
		0x1d95U,0x36caU,0x5aa7U,0xd19bU,0xacc0U,0xd7b8U,0xc436U,0x893bU,0x787dU,0x868eU,
		0x4aa2U,0xf77cU,0x5dd7U,0xd7f4U,0x2b5fU,0x9564U,0x9139U,0xea5bU,0xb1bbU,0xdbd3U,
		0x4cacU,0xab11U,0x6b27U,0x7826U,0x3107U,0x1e47U,0xe598U,0xb8c0U,0xf1e9U,0xabcaU,
		0x6a41U,0x942fU,0x6cfbU,0x6daeU,0x27f5U,0x8d09U,0xbc5bU,0x3240U,0x05b8U,0xa6c3U,
		0x5e63U,0x8f52U,0xd4e2U,0x585bU,0xaff4U,0x5428U,0xf75dU,0x9e98U,0xe9a7U,0x5e58U,
		0x22b4U,0xdbdeU,0x7c63U,0x49c9U,0xeedeU,0xbaf2U,0x5da3U,0x1df9U,0x7b1eU,0x54b3U,
		0x13e4U,0x7060U,0xef77U,0xf4c7U,0x3d96U,0x3b8fU,0x1282U,0x183fU,0x72dfU,0x207cU,
		0x4375U,0x1e3cU,0xf3abU,0x3a66U,0xc5ceU,0xf3e3U,0x2ecdU,0x0130U,0xf979U,0x3f08U,
		0xa191U,0xf7ecU,0xac4cU,0x11abU,0xe898U,0x18d3U,0xb7fdU,0x5d0aU,0x5f2bU,0x6495U,
		0x0857U,0x3fd4U,0x659eU,0x8e1fU,0x6320U,0x2965U,0x82d7U,0x8d1bU,0x37f9U,0xa8baU,
		0x9be5U,0x27eaU,0xb6bfU,0xab3dU,0x661dU,0x988dU,0x6436U,0x5bd7U,0x1ff3U,0xe297U,
		0x31eeU,0xc119U,0xa455U,0x1339U,0xfac4U,0xe8a2U,0x8144U,0xa021U,0x6d05U,0x5bc6U,
		0xe1b0U,0x1fa2U,0x8d58U,0x6265U,0xce3eU,0xc7e1U,0x4b0fU,0x2236U,0x1994U,0x7631U,
		0x33e3U,0xb5b8U,0x7397U,0x25e1U,0xe7e3U,0xe0e8U,0x0b9aU,0x89bfU,0xe243U,0x8930U,
		0x09cfU,0x0a77U,0xe028U,0xdbbcU,0xd4f6U,0xeeadU,0x32b7U,0x5570U,0x6e14U,0xfb02U,
		0x8d61U,0x9d52U,0x6613U,0x6900U,0xa74cU,0x5e7cU,0xb38cU,0x5ab0U,0x1d19U,0x96eeU,
		0xe373U,0xae5cU,0x60cbU,0x3e76U,0x9fc0U,0x49f4U,0xb3f3U,0xaddbU,0xbc82U,0x5dccU,
		0xa4d9U,0xa879U,0xcf71U,0xeee6U,0x6a82U,0x49c2U,0xeb3bU,0x946eU,0x99cfU,0x2a3cU,
		0x1bf3U,0x6a89U,0x42b8U,0xfea1U,0xc0efU,0x92a5U,0xfa93U,0x84c6U,0xb69dU,0xb956U,
		0xf827U,0xeb6fU,0xaa60U,0x6fd8U,0x1008U,0xa3dcU,0x06f6U,0x8288U,0x452fU,0x110fU,
		0xc1b9U,0xfffeU,0xc0cdU,0x8374U,0x4443U,0xa0dcU,0xeba3U,0xc8b0U,0x7b05U,0x06eaU,
		0xd2c9U,0x819cU,0x6297U,0x2b1cU,0x07f8U,0x05feU,0xd1ddU,0xad5eU,0x8a6bU,0x9ba0U,
		0x61c7U,0x9484U,0x4357U,0xba6dU,0xc8a4U,0xc7c4U,0x61f6U,0x3cc0U,0x563fU,0x90fcU,
		0x61abU,0x2382U,0x8086U,0x0512U,0x7c28U,0x0dedU,0x1304U,0x23b1U,0xfe7aU,0xac72U,
		0x897fU,0x221fU,0x7fccU,0x8a5fU,0x426bU,0x4c26U,0x5d5bU,0xb821U,0x6f19U,0xb415U,
		0x77afU,0x9398U,0xecd6U,0x3bd4U,0x1806U,0x2452U,0x579bU,0xc9b2U,0xc7b8U,0x0467U,
		0xfd8eU,0x2e34U,0x1503U,0xae91U,0x1a74U,0xd9c0U,0xe60cU,0x22c5U,0xfcb3U,0x8973U,
		0xc621U,0x0158U,0x8b6dU,0x844dU,0x8893U,0x1956U,0x0679U,0x396aU,0x5553U,0x81fcU,
		0x75d0U,0x2296U,0xff3bU,0x30d4U,0x0de1U,0x7e31U,0xfb12U,0x0426U,0x7b3cU,0x1f16U,
		0x79dcU,0xb1ecU,0x7f43U,0x3de3U,0xcfa7U,0xc8c6U,0xc6ccU,0x1eacU,0x8c15U,0x4241U,
		0x52b7U,0xf29aU,0xebeaU,0x17faU,0x8759U,0x2366U,0xe53cU,0x1fe2U,0xe56bU,0x92acU,
		0xd159U,0xd4beU,0xa7d3U,0xc035U,0x6e2dU,0xf8ccU,0x5c6fU,0x4187U,0x4c1aU,0x13baU,
		0xe3d9U,0xb0f8U,0xe777U,0x4b3aU,0x0342U,0x0eefU,0x72d7U,0x1e2aU,0x78f0U,0x0dd5U,
		0xd574U,0xf60bU,0x14c4U,0xa932U,0x0cf5U,0xefc2U,0x72b3U,0x033aU,0x8c81U,0xa594U,
		0x58d9U,0x2fe8U,0xba87U,0x4c76U,0x81f0U,0xcb7dU,0x74e4U,0x698eU,0x5467U,0xc07eU,
		0xd70eU,0xa726U,0x3213U,0x9314U,0x9679U,0x387eU,0xc535U,0x4dd0U,0x03fcU,0x02ffU,
		0xe066U,0x41baU,0xa297U,0x8887U,0xff9dU,0x186aU,0xaaa9U,0x40feU,0x7306U,0x9162U,
		0x8a81U,0xd748U,0xb455U,0xd305U,0xfca2U,0x1472U,0x44f7U,0xada5U,0xd9b4U,0xde6fU,
		0x4082U,0xa6f7U,0xe19cU,0xd8d4U,0xe520U,0x2e9bU,0xcef4U,0x4914U,0x2aafU,0xa7c7U,
		0xe47fU,0x1e84U,0x9b7bU,0xd3d2U,0x9a55U,0xdc2eU,0xefeeU,0xacb0U,0xd8adU,0xbf49U,
		0x6d41U,0xd6bdU,0x704cU,0x68beU,0x4b2eU,0x32edU,0xc164U,0xbe6eU,0x1c09U,0x87dcU,
		0xc312U,0x409bU,0x1be3U,0x4a4eU,0xdb10U,0xee24U,0x4608U,0xf303U,0xe2edU,0x4d28U,
		0x3fecU,0x4155U,0xa340U,0xcf9bU,0x5695U,0xbccdU,0xc5e0U,0x3c2cU,0xcebbU,0x3b97U,
		0x73a7U,0xe5c4U,0xe0d0U,0x66aaU,0xa575U,0x3cf7U,0x1908U,0xc0f6U,0xf28eU,0x227aU,
		0xe450U,0x68edU,0x2bccU,0x378bU,0xfb24U,0x6c82U,0x51cbU,0xc833U,0x5778U,0x170eU,
		0x43b7U,0xb1d5U,0x2e3dU,0x1577U,0xb369U,0xbcdfU,0x6653U,0x1386U,0x89eeU,0x5b4bU,
		0x92c4U,0x0572U,0x7390U,0xacb9U,0xd8c8U,0xaa47U,0x5968U,0x1933U,0xe0b0U,0xed86U,
		0xf945U,0x28e4U,0x2aa8U,0xfd87U,0x68abU,0xa60bU,0x82b1U,0x4828U,0x16a7U,0xdac2U,
		0xf535U,0xb449U,0xe434U,0x19fbU,0x122fU,0x879bU,0xfde6U,0x4d9bU,0x7e88U,0x86c3U,
		0xdf7aU,0x1850U,0x0310U,0xc660U,0xc75cU,0x01a7U,0x02a5U,0x64fbU,0x9156U,0x4ce4U,
		0xfd51U,0x1037U,0xbdf4U,0x9076U,0x8995U,0x3562U,0xa77fU,0x75ebU,0x2916U,0x6730U,
		0xfeb1U,0xb9afU,0x2865U,0x4012U,0xeb13U,0xa28bU,0xa612U,0x5c10U,0x9de9U,0x9228U,
		0x958dU,0x82feU,0xca41U,0x5d36U,0x6982U,0x8a38U,0x537bU,0xec69U,0x97f3U,0xa63aU,
		0xe03eU,0xd392U,0x6e6bU,0x68a2U,0xc339U,0xb1a9U,0x6239U,0x9a68U,0x8104U,0x4defU,
		0xb99bU,0x3207U,0x2a79U,0x9653U,0x9878U,0xc931U,0x50cfU,0x2efcU,0x0c09U,0x4d6cU,
		0x6287U,0xdd9dU,0x88bfU,0x9ccbU,0xafdfU,0xa846U,0x5bf5U,0xa4a8U,0x0ee3U,0x877dU,
		0xb4beU,0x2df8U,0x71c7U,0x07c3U,0xa622U,0x84cdU,0x0256U,0xc525U,0x4997U,0x1d3eU,
		0x14e4U,0xddc3U,0xf4a7U,0xcaceU,0x02d4U,0x0ce9U,0xd488U,0x831fU,0x8029U,0xc6a8U,
		0x72bbU,0x972fU,0xb7c6U,0x5e92U,0xcf8fU,0x173bU,0x16c1U,0x3aabU,0xd0a9U,0x3c0bU,
		0xb403U,0xc4fcU,0xd6e5U,0x3492U,0xdfe4U,0xa2eaU,0x6aedU,0xdcfcU,0x3f00U,0xd67eU,
		0xe998U,0xceafU,0xb17dU,0xa20cU,0x7fdfU,0x4e3cU,0x674fU,0xbbe1U,0xc4f4U,0xefbaU,
		0x0b5eU,0x346eU,0x8239U,0xcdffU,0xf40bU,0x1622U,0x42cbU,0x98c9U,0x68e1U,0xf39cU,
		0xa1c3U,0x5834U,0xfa38U,0x3c7eU,0x54cfU,0xc747U,0x9196U,0xf3cdU,0x4dbcU,0x369fU,
		0x2b49U,0xf066U,0x8d38U,0x45b5U,0xe364U,0x43abU,0xf1a2U,0x1efdU,0xad40U,0x1a94U,
		0xb287U,0x7c42U,0x0e31U,0xc00aU,0x68b2U,0xdb42U,0x8d7eU,0x023dU,0x8745U,0xa357U,
		0x9a8fU,0x1ac9U,0xe393U,0x90afU,0x0acfU,0x332fU,0xf055U,0x7ea6U,0x3574U,0xa6dbU,
		0xbb0cU,0x9b6fU,0xa02aU,0xd5d1U,0x6f60U,0x2888U,0xa8d5U,0x06b2U,0x94c0U,0xb440U,
		0xef02U,0x7610U,0x90dcU,0xbe5dU,0xd62cU,0xedcbU,0x63e2U,0x6452U,0x4a42U,0xb0e3U,
		0x0375U,0xbd82U,0x506eU,0xa221U,0x0f8bU,0x3d9eU,0xbf18U,0x86f0U,0xc226U,0xc5b6U,
		0xc4b9U,0xfe59U,0xda24U,0xfa9aU,0x2414U,0xc158U,0xd523U,0x6c64U,0x02b9U,0x4962U,
		0xe498U,0x4c3cU,0x9903U,0xdccdU,0x531dU,0xcbf9U,0x451cU,0x34c1U,0x0df7U,0xb707U,
		0x3e2fU,0xa653U,0x8528U,0x5abeU,0x9e88U,0xa3bcU,0x4767U,0x06d7U,0xf67bU,0xc8d0U,
		0x8db4U,0x3742U,0xb3d2U,0x14d8U,0xe269U,0xc473U,0x0b1dU,0x34d2U,0x2e1fU,0xdcb1U,
		0xd752U,0x96deU,0xbc55U,0x020aU,0x9fbaU,0x9bdfU,0xaef3U,0xe5ceU,0x8313U,0x44aaU,
		0x0c03U,0x1229U,0x63dcU,0x8233U,0x7f3dU,0x5639U,0x3fe6U,0xc52fU,0x260dU,0x09ddU,
		0xf3bbU,0x259dU,0x75f5U,0x0bfdU,0x729eU,0x0ff1U,0xff58U,0xdcd7U,0xf589U,0xd145U,
		0x63aeU,0x80d3U,0x48abU,0x2672U,0x016aU,0x8674U,0xc014U,0x6354U,0x8771U,0xc3beU,
		0xb8e3U,0x83e1U,0xed8dU,0x5c4aU,0xdc83U,0xa544U,0x6512U,0x2175U,0x94a1U,0x4900U,
		0x9cfdU,0xf324U,0x88f3U,0x2a6dU,0xf887U,0x4f39U,0xc3f7U,0x5b24U,0x526aU,0x711fU,
		0xea6eU,0xa54eU,0xdf52U,0xcc89U,0xf4e9U,0x0d70U,0xa3c6U,0xf72eU,0x6896U,0xc208U,
		0x4061U,0xeb7dU,0x3315U,0xb210U,0x75fbU,0x029cU,0x6d7fU,0x567bU,0x1c65U,0xb0bdU,
		0x37cbU,0x4fd4U,0xf878U,0x97e7U,0xaf86U,0x997cU,0x8018U,0x0633U,0x2ffbU,0x5297U,
		0x298dU,0x6a90U,0xc86cU,0xb7aeU,0xea42U,0x24caU,0xcc3aU,0x311bU,0x5d2aU,0x2e10U,
		0x33f9U,0x7cc7U,0x9359U,0x61e5U,0x978cU,0xfb8cU,0xa59cU,0xd87bU,0xfc65U,0x601aU,
		0xd8f5U,0x8f8aU,0x9869U,0x914aU,0x58eaU,0x8b86U,0x1bc5U,0x3476U,0xa693U,0x7981U,
		0x204dU,0xc3edU,0xfdc5U,0x30f2U,0x188dU,0x9264U,0xd907U,0x75beU,0x86b7U,0x6644U,
		0x8b44U,0xa19fU,0x64a5U,0x5df9U,0xf64cU,0x7681U,0x2779U,0x98a0U,0x0ff7U,0x1764U,
		0x084fU,0x419aU,0xc217U,0x290aU,0x7d9cU,0xc38fU,0xfcfaU,0xcd18U,0xf628U,0x8e0aU,
		0xb175U,0x74dcU,0x7c99U,0x86afU,0xb4a1U,0x7338U,0x63c7U,0x640fU,0x906aU,0x8d29U,
		0x3013U,0xab53U,0x9eb7U,0x54f3U,0x0224U,0x2ca8U,0xa399U,0x00bbU,0x9839U,0x03abU,
		0xe0b7U,0xfad2U,0x578cU,0xb35dU,0xe348U,0x6efbU,0xdc4eU,0xef52U,0x6866U,0xcfebU,
		0x81c5U,0x216bU,0x3dfdU,0xd9d6U,0xbabfU,0x7442U,0xdf8aU,0xa51fU,0x0566U,0xc6bdU,
		0x0770U,0x3142U,0x29bcU,0x4ef3U,0xf859U,0x3fb5U,0x9053U,0x0f77U,0xf00bU,0xed03U,
		0xb221U,0xec9aU,0x72a4U,0x2a9cU,0xb557U,0x2fbcU,0x5a84U,0xa9ccU,0xb626U,0x0486U,
		0x8fbaU,0x1847U,0x1c36U,0x43f4U,0x352bU,0xb5c6U,0xd7dcU,0x58ffU,0x19efU,0x1bfaU,
		0x2d51U,0x637eU,0x9d25U,0xbbf3U,0xfe38U,0xc94bU,0x2a0dU,0x5351U,0xb588U,0x129aU,
		0x8902U,0x4340U,0x8b1aU,0x168aU,0xfd97U,0x4ba9U,0x1d06U,0xbd71U,0xd540U,0x1268U,
		0xffc3U,0x280eU,0xd731U,0x69c0U,0x93d4U,0xeff7U,0xcb18U,0x7524U,0x7e9aU,0x85e5U,
		0xc72cU,0x05c6U,0x0dc4U,0x547fU,0xe7b1U,0x354bU,0x6e74U,0xea38U,0x9f3eU,0x1b23U,
		0xdaedU,0xaecaU,0x6c02U,0x94c9U,0xf8edU,0x9a83U,0x09e3U,0xd850U,0x961eU,0x446cU,
		0x334aU,0xd861U,0x9c7dU,0x8fc3U,0xb483U,0xa5efU,0xdbb5U,0x955dU,0x7604U,0x8e35U,
		0xef69U,0xb6f0U,0x2164U,0x2a66U,0x67c6U,0x6b71U,0xbf42U,0x0151U,0x3328U,0xec62U,
		0xaf41U,0x5847U,0x8824U,0xa26bU,0x99a3U,0xbb00U,0xea9eU,0xe58aU,0x28ebU,0x6a22U,
		0xd938U,0xead0U,0x8ce4U,0xc06cU,0x7b6aU,0x56f7U,0xb7efU,0xacceU,0x6446U,0x6b78U,
		0x01cdU,0x31daU,0x2b84U,0xea1fU,0x9817U,0x693aU,0x37a6U,0xfb82U,0xf71cU,0xee6eU,
		0x4e54U,0x4b50U,0xd1f1U,0x33d5U,0x4eb8U,0xbf0cU,0x2f1cU,0x7fafU,0x5231U,0x4d0eU,
		0x2613U,0xa135U,0xda0aU,0xb78cU,0x487eU,0xab1fU,0x0289U,0x4043U,0x4956U,0xf2b5U,
		0xe01aU,0xd3e9U,0x9ef1U,0xbfa1U,0x0f56U,0x6366U,0x881bU,0xfea8U,0x1ab1U,0xc4caU,
		0xc3cdU,0x8917U,0xc361U,0x3f44U,0xc266U,0x5311U,0x0e9fU,0xa4cbU,0x26b6U,0x8604U,
		0xce65U,0xc45fU,0x8c89U,0x39bfU,0x4571U,0xb662U,0x104cU,0xc48eU,0xcea3U,0x863aU,
		0xb6daU,0x41e4U,0x8863U,0x0b66U,0x44daU,0xe1e9U,0xb162U,0x9643U,0x15a3U,0xd69eU,
		0x9f7eU,0x33a6U,0xdef8U,0x7f57U,0x7ca1U,0x0b52U,0x4688U,0x64bdU,0xcfbdU,0x67ddU,
		0xb3e1U,0xc410U,0x86e0U,0x6d75U,0x25a3U,0x719aU,0x6a01U,0x7486U,0x5e86U,0xd592U,
		0x77c8U,0x4a7aU,0x0f27U,0x21e9U,0x1789U,0xcdc3U,0xb091U,0x2442U,0x9f58U,0x5d14U,
		0x6663U,0x560fU,0x8d13U,0x1db0U,0xd260U,0xd09dU,0x4938U,0x4d50U,0xdc63U,0x3382U,
		0xc97dU,0x8a1fU,0xefb2U,0x930cU,0xc408U,0x7679U,0xf94cU,0x8bfcU,0x1ef1U,0xe462U,
		0x44b9U,0x80abU,0x9719U,0x5748U,0x1dcbU,0xf2e1U,0x20aaU,0x67cdU,0xd657U,0x5fa4U,
		0x53e3U,0x8f41U,0x4bf3U,0xccbdU,0x814dU,0x2b3dU,0xf991U,0x1536U,0xae24U,0xd2a1U,
		0x0c23U,0xa1f9U,0x954dU,0x97ddU,0x67f5U,0x10b5U,0xfd68U,0x59aeU,0xf3c1U,0xc2fcU,
		0x7f6dU,0x7b56U,0x0d45U,0xec33U,0x580eU,0x6dbfU,0x0f00U,0x128aU,0xf0b6U,0x69efU,
		0x8aecU,0x19a8U,0x26d9U,0x5a33U,0x5101U,0x42bfU,0x0faaU,0x8e50U,0x5c23U,0x6c2bU,
		0x8c4cU,0x23f4U,0x2a5dU,0x102eU,0x6275U,0x3026U,0xba57U,0x84f4U,0x7364U,0x35b3U,
		0xd1e8U,0xb992U,0xaae9U,0xec17U,0xb135U,0x2430U,0xf1e0U,0x1baaU,0x3592U,0x351fU,
		0xf76dU,0x77baU,0x46afU,0x0e8bU,0xa15aU,0x7c12U,0x570dU,0x4e07U,0xe5aeU,0x3c54U,
		0x2de6U,0xd5ebU,0x0684U,0x1ddeU,0x047aU,0x1afdU,0x8855U,0x12bcU,0xb685U,0xd164U,
		0x500bU,0xa433U,0x80d9U,0x19c9U,0x2f7aU,0x2c9eU,0xadeeU,0xf6d4U,0x350cU,0xceffU,
		0x84b7U,0x7356U,0x0a51U,0xb57cU,0x321eU,0x4128U,0x3d11U,0x4779U,0xd444U,0x51d6U,
		0xdfcbU,0xd0baU,0xd5a1U,0xb271U,0x2baaU,0x3bf8U,0x813fU,0xe893U,0x1beeU,0x6e0fU,
		0xdb62U,0x1361U,0x7c5eU,0x4ca7U,0xb450U,0x58d4U,0x1903U,0x4b29U,0x4266U,0x6292U,
		0xe566U,0x0674U,0x4db7U,0x6ae8U,0xda1fU,0xbb07U,0xfd4cU,0xe0abU,0xa61dU,0x6e66U,
		0xb49cU,0xfdc0U,0x19eaU,0xdf85U,0xff53U,0xe264U,0x2988U,0x5265U,0x665eU,0x159eU,
		0x50fcU,0x4beeU,0x215fU,0xc727U,0x8816U,0x9812U,0x63b4U,0x039aU,0xe39eU,0xcdddU,
		0xd43fU,0x5708U,0x5998U,0xb497U,0xb46bU,0xbf23U,0x1b69U,0x6e84U,0xd26fU,0x2b76U,
		0x67a5U,0x29b0U,0x599dU,0x6fdfU,0x945cU,0xb4daU,0xd6f0U,0xc88cU,0xcb92U,0xbb4cU,
		0x8e78U,0x4d3cU,0x8eceU,0x0daeU,0xd9fcU,0x4308U,0xecf7U,0x91a1U,0xa93aU,0x905eU,
		0x47f5U,0xdc40U,0x8649U,0xd977U,0xd809U,0xd206U,0xfb79U,0xb1a0U,0x2a18U,0x4e44U,
		0xdfb6U,0x072cU,0x781dU,0xd78fU,0x039fU,0x0be9U,0x1c28U,0x7895U,0x1f61U,0x4a1eU,
		0xcd20U,0x2784U,0x9c06U,0x731dU,0x57feU,0xee86U,0xa3a4U,0x338aU,0xa8a9U,0x0695U,
		0x2678U,0x276bU,0xfcd9U,0x0843U,0x3cdcU,0x9178U,0x2998U,0xed46U,0x3c3dU,0x489bU,
		0x2f30U,0x74d2U,0x85cbU,0xea79U,0xde3fU,0x759cU,0x63b9U,0x3e5cU,0xcd0cU,0x79fdU,
		0x4112U,0x9009U,0x7237U,0xa0b3U,0x9859U,0x97b4U,0xf4d7U,0xd75dU,0x1671U,0x6139U,
		0xf560U,0x3666U,0xf53cU,0xc2e5U,0xa44dU,0x9e1eU,0xb09cU,0x8b38U,0x4bdaU,0xcc2cU,
		0x15deU,0x85c3U,0x3834U,0x99aaU,0x54c7U,0x72feU,0xb089U,0xb16dU,0xcde2U,0x0af3U,
		0x0f0bU,0x3178U,0x9258U,0x40aeU,0x6888U,0x941bU,0x17caU,0x20b5U,0x09f6U,0x281eU,
		0x7cb0U,0x4eceU,0xdf6aU,0x957aU,0x48b1U,0x3983U,0x0c35U,0xffceU,0x974dU,0x7290U,
		0xd37bU,0x2041U,0x3355U,0x7068U,0x9dcbU,0xe75dU,0x0b87U,0x2bbcU,0x3641U,0x7d92U,
		0xe3a3U,0xaa67U,0x2c1aU,0xd0e1U,0x3e21U,0xd943U,0x8f7eU,0x78d1U,0x62dbU,0x6113U,
		0x1ecfU,0x87c5U,0xe5fcU,0xa98eU,0x261eU,0x724cU,0xae9aU,0xc17eU,0xe9c8U,0xd31cU,
		0xd422U,0x4e65U,0x03e1U,0x96fbU,0x0d64U,0x51abU,0x536dU,0xbf52U,0x5110U,0xdec0U,
		0x7f2cU,0xe2daU,0x2b7bU,0xe035U,0x750fU,0x98ecU,0x2021U,0x558fU,0xb1b2U,0x5fd0U,
		0xaba3U,0x4055U,0x47a0U,0xafd1U,0x36dfU,0x7619U,0x2d19U,0x8cd3U,0xdcddU,0xee39U,
		0x9c9dU,0x60f9U,0x96a0U,0x9ca8U,0x7085U,0x43fdU,0xf937U,0xbaa8U,0x5c3eU,0xb963U,
		0x5684U,0x7bdbU,0xaea5U,0x2dbdU,0xd274U,0x302fU,0xabf3U,0x663aU,0xf705U,0x6f4cU,
		0xe956U,0xa96cU,0xc780U,0xeb05U,0xc4d3U,0x88e7U,0x1a19U,0xcf7eU,0x163aU,0x69a4U,
		0x1a02U,0x4814U,0x2cf3U,0x70fcU,0xc775U,0x49dfU,0x6426U,0x2629U,0xdb97U,0xac59U,
		0xa745U,0xe591U,0x7cbbU,0x21b9U,0x98f5U,0x5b5eU,0x29b5U,0xf4e2U,0xae1dU,0x4e4dU,
		0xc30bU,0xff34U,0xa8ceU,0x80fdU,0xec20U,0xde05U,0x9c11U,0xe31eU,0x6881U,0x2981U,
		0x2d2cU,0xa8c7U,0xff5eU,0x5e38U,0x517eU,0x390aU,0x9a1cU,0xa462U,0x662fU,0xf504U,
		0x9a0bU,0xbf89U,0x5de7U,0xcb9dU,0xf3fdU,0x57d9U,0x0627U,0x8954U,0x67aaU,0xa3e3U,
		0xcb60U,0x763eU,0x7e19U,0x9d70U,0x5718U,0xf642U,0x5c81U,0xb124U,0x78abU,0x0414U,
		0x4ff0U,0xdfd6U,0x3f75U,0x37bfU,0x14ccU,0x76a3U,0xfe6cU,0x8421U,0xb91bU,0xd059U,
		0x6686U,0xbcfeU,0xe9bdU,0x9020U,0x861aU,0x7fe7U,0x47bfU,0xbc68U,0x160dU,0xe870U,
		0xbf28U,0x851cU,0x00ffU,0xc0bfU,0x4ad2U,0xbc19U,0xfd02U,0x8588U,0xc063U,0x82f5U,
		0xfb5cU,0x49d6U,0x6056U,0xdc6bU,0x8685U,0xbdc0U,0xd14bU,0xc443U,0x01feU,0x817fU,
		0x6cdaU,0xef37U,0x7e51U,0x0a39U,0x62f0U,0x1e16U,0x51a0U,0xe7cdU,0x15e6U,0x9bc5U,
		0x7947U,0x113dU,0xb470U,0x79ceU,0xfa05U,0x0b11U,0x71b2U,0xa1d5U,0x95a4U,0x7833U,
		0x2837U,0xd110U,0x2521U,0xd871U,0x815cU,0x24b1U,0x120aU,0x60acU,0xe43bU,0x3d25U,
		0x18b2U,0xe4e2U,0x8227U,0x9378U,0xd700U,0x4e9fU,0x2bddU,0x7a42U,0xbb69U,0xeaa5U,
		0x2016U,0x9af3U,0xac8bU,0xb753U,0x6e89U,0x95f2U,0x5bcfU,0x79aaU,0xab65U,0x2601U,
		0x2d33U,0xd9a6U,0x0a31U,0xed3eU,0xabdaU,0xde17U,0xf394U,0x887fU,0x4a72U,0xc387U,
		0xf58fU,0xcd7dU,0xf6b9U,0x93acU,0xe471U,0x1073U,0x80c7U,0x05ebU,0x503cU,0x13ecU,
		0x0ec9U,0x8ba3U,0x416aU,0x2f94U,0x8cc8U,0xf0d6U,0x1b6eU,0x100fU,0x0d0bU,0x7b81U,
		0xb029U,0xa567U,0xc0acU,0xb8d7U,0x2af5U,0xbed5U,0x3bbfU,0x59d2U,0x9d9cU,0xf960U,
		0xbd3dU,0x1bbbU,0x7c32U,0x6043U,0x6577U,0x2fb2U,0x1c79U,0x1cc4U,0xb5e0U,0x46e3U,
		0x78c2U,0x1436U,0x0e49U,0xcfb1U,0x484cU,0xffe9U,0x680eU,0xd18eU,0xc891U,0xaa37U,
		0x9ae8U,0x0922U,0x3a3bU,0xb836U,0x8a8bU,0x3296U,0xe81eU,0x758dU,0xd692U,0xeae0U,
		0xf8fcU,0x73b1U,0xeec6U,0x4f6fU,0xc3c4U,0x498eU,0xf2f3U,0x7e10U,0x0c60U,0x6175U,
		0x5d9aU,0x0260U,0x6915U,0x44ceU,0x0665U,0x5d8cU,0xa789U,0xcf11U,0xc908U,0x70aeU,
		0xd6f5U,0x990dU,0x5899U,0x2356U,0x9e6cU,0x42e4U,0x4323U,0xefd9U,0x862eU,0x53c3U,
		0x0e3bU,0xa424U,0xb93aU,0xccd7U,0xf68cU,0x91ceU,0x482fU,0x549dU,0xecbbU,0xc324U,
		0x85f4U,0x81cfU,0x250fU,0x8fb0U,0xe172U,0x7a9eU,0x459cU,0x584eU,0x8579U,0x66f1U,
		0x2057U,0x492cU,0xbb51U,0x0a66U,0xe82cU,0xdc8dU,0x428eU,0x5f94U,0xde0cU,0x2569U,
		0x9e92U,0xf21dU,0x49bcU,0xc113U,0x4af9U,0xb744U,0x9300U,0xd5b2U,0x8777U,0x01a1U,
		0x86eaU,0x3c78U,0x17f4U,0xac6cU,0xe5beU,0x40f8U,0xd3e3U,0x9557U,0x69e9U,0xcdbdU,
		0x9976U,0x2597U,0x43eeU,0x8e04U,0xcb97U,0xb95dU,0x8b9dU,0xe7c7U,0x6e7eU,0xb576U,
		0xe757U,0x74ccU,0x6da1U,0x3459U,0xd1abU,0xc3a2U,0xcdb7U,0x5d86U,0x1444U,0x37b0U,
		0xefcaU,0x252bU,0x8040U,0x1ee5U,0xed1cU,0xf5b3U,0x3e82U,0x119eU,0x1351U,0x5947U,
		0xbd47U,0xbbe9U,0x23a0U,0x32c3U,0x8a00U,0x5514U,0x6fe4U,0x8df5U,0xf2d5U,0x4429U,
		0xe684U,0x7a6aU,0x8e12U,0xabfdU,0x2c4bU,0x071cU,0xcd72U,0xeb47U,0x5722U,0x8a27U,
		0x7969U,0xe6b7U,0x83e7U,0xf985U,0x118fU,0x8408U,0x7278U,0xab30U,0x364bU,0xde1fU,
		0xc456U,0xc69fU,0x2971U,0x37b6U,0x99b2U,0x2f3aU,0x1836U,0xf0a3U,0x59a2U,0x2308U,
		0xae47U,0xba33U,0x1f95U,0x7141U,0x5b37U,0x1fffU,0x0ebeU,0xf1b5U,0x1452U,0x2f84U,
		0xc7edU,0xd7a6U,0x69d6U,0x41d3U,0x82b8U,0x565cU,0x0abbU,0x5e6fU,0xe32fU,0x00d3U,
		0x43cdU,0x26cdU,0x1e05U,0xe27dU,0x4b97U,0xaf48U,0xac19U,0x9e3eU,0x66ffU,0x8b10U,
		0xb4dfU,0x77fbU,0x25d4U,0x5eb8U,0x52f7U,0x4aaeU,0xec27U,0xa235U,0xa0f1U,0x85b2U,
		0x1cdfU,0xdb30U,0x10f4U,0x66e1U,0xb8ccU,0xba42U,0xb8e9U,0xeb62U,0x4b07U,0x6581U,
		0x7379U,0x2a51U,0x83c3U,0x4fb7U,0x2519U,0xf4cfU,0x73cbU,0x144aU,0xad38U,0xe194U,
		0xc975U,0x021cU,0x9461U,0xb6a4U,0xcf2aU,0xfa49U,0x8e44U,0xbd51U,0xa2beU,0xa10dU,
		0x32daU,0x58c2U,0x001eU,0x1425U,0x9d91U,0xac29U,0x6709U,0x7702U,0x150cU,0x2722U,
		0xbc0eU,0xebc7U,0x2bffU,0x7743U,0xd454U,0x834aU,0x1c1aU,0x4cd5U,0xd844U,0xd8b6U,
		0x04fbU,0x140fU,0x6fb5U,0x4abdU,0x430dU,0xbff3U,0x8b06U,0x7438U,0x7daeU,0x24ebU,
		0xc342U,0x5ee0U,0xddb4U,0xf552U,0xd645U,0xdba9U,0x55edU,0xef0bU,0x2539U,0xe1c3U,
		0x867aU,0x77e1U,0xc640U,0x52caU,0x6b87U,0x5393U,0xcb3cU,0x1c3fU,0xec56U,0xdfa4U,
		0x3826U,0x345fU,0x6d88U,0x7712U,0xf176U,0x2694U,0xda01U,0x627eU,0x6ff2U,0x5175U,
		0xb3a7U,0xddf0U,0x05afU,0x1a37U,0x8567U,0x9997U,0x1abaU,0xd36dU,0x7781U,0x3b08U,
		0x6b16U,0xf6e5U,0xdac9U,0x9e09U,0xb10dU,0xe24eU,0x30b6U,0xfb65U,0x8416U,0xa503U,
		0xbc8dU,0xb604U,0x825aU,0xa272U,0x2f6cU,0x1947U,0x7518U,0x4b6bU,0x91a6U,0xe858U,
		0x08f7U,0x2a21U,0x2eb9U,0x4380U,0xe325U,0xe33eU,0xaaf2U,0x7c7aU,0x35fdU,0xd775U,
		0x9b91U,0xa14cU,0x9bf0U,0xffb7U,0x0170U,0x1a4dU,0x6ffbU,0xba14U,0xe9b2U,0x520eU,
		0xd131U,0x31a6U,0xbbb9U,0x73dcU,0xae55U,0x6da7U,0xaebeU,0xa17fU,0x8ec0U,0xb1c6U,
		0xecfcU,0xeb76U,0x10aeU,0x7fa8U,0xf2fcU,0x1f0fU,0xbe56U,0xc92aU,0x66daU,0xb73dU,
		0x257bU,0xa145U,0x4ec7U,0xc720U,0x8878U,0x297aU,0x0cfdU,0x5a24U,0x5305U,0x7ec3U,
		0x5145U,0xa008U,0x3af5U,0x9ab0U,0x9a36U,0xf7dcU,0xd105U,0x6757U,0xe004U,0xbe20U,
		0x21d5U,0x22daU,0x4d41U,0x4a64U,0xf51bU,0x1aa5U,0xf2c4U,0xdd7fU,0xf630U,0x00a8U,
		0x36baU,0xe207U,0x85a8U,0x3fabU,0x3b3cU,0xc985U,0xa6a7U,0x684eU,0x635aU,0xdf32U,
		0x5192U,0x4cffU,0xa756U,0x2660U,0x1fb4U,0xabe2U,0xf9faU,0xe933U,0x5a5bU,0xc3a8U,
		0x383cU,0x74fbU,0x00e1U,0xb807U,0x8e7dU,0x4872U,0x27ffU,0x6678U,0x3779U,0xe9e4U,
		0x19d9U,0x9ac2U,0xf2acU,0xa49fU,0x73d3U,0xbf80U,0x7a7eU,0xda47U,0x541fU,0xa11cU,
		0x16aeU,0xc85bU,0x872fU,0x6072U,0x2ce5U,0xc3deU,0xee62U,0xe415U,0xc29fU,0x59f4U,
		0xc565U,0x882bU,0xc275U,0x65f8U,0x1132U,0x6554U,0x0db3U,0xf1cdU,0x6d0eU,0x0c80U,
		0x4ee4U,0x0619U,0x9c18U,0x5225U,0x34f3U,0xc373U,0xa2ccU,0x765cU,0x39e4U,0x2582U,
		0x25caU,0x055cU,0xc01aU,0x3f57U,0x854aU,0xfc91U,0xeac4U,0x4c60U,0x4792U,0x83d2U,
		0xd91eU,0xef7fU,0xf33cU,0xd1b1U,0x6ccfU,0xca09U,0x0355U,0xa049U,0x8ed3U,0xf82eU,
		0xa019U,0xbeb5U,0x0e6aU,0xb7e3U,0x4fc5U,0xe948U,0x79c6U,0x3e54U,0x486aU,0x2300U,
		0x3697U,0xd2fdU,0x5607U,0x7330U,0x8295U,0xa675U,0xc2c8U,0xbad4U,0xcbdfU,0xbad0U,
		0x35d6U,0x7ba0U,0x4bcdU,0x8cafU,0x3a9dU,0xc1c9U,0xbf37U,0x753fU,0xfd3bU,0x5bafU,
		0xd97cU,0x3281U,0xa6bbU,0xcbe3U,0xce8aU,0xf6a5U,0x62a4U,0x1c9aU,0x3902U,0xcdd5U,
		0xba0cU,0x3c70U,0xdd95U,0xc8beU,0xb784U,0x5673U,0x217bU,0x8c2bU,0x5e11U,0x7ba4U,
		0xdd4aU,0x6bcaU,0x1249U,0xd2d8U,0xedbdU,0x422bU,0x9dbeU,0x1ccaU,0x238fU,0x8398U,
		0xe57fU,0x6ee8U,0x864eU,0x4365U,0x91dfU,0x35daU,0xaa93U,0xe905U,0x33c1U,0x3b59U,
		0xd33aU,0xc14aU,0x5d68U,0xe5efU,0x9940U,0x6cbeU,0xfb52U,0x6745U,0x86caU,0x0c9dU,
		0xdee4U,0xa679U,0x6bb4U,0x23bfU,0x7cf7U,0xa24aU,0x9613U,0x1167U,0xf044U,0x126fU,
		0xa89cU,0xa096U,0x75ddU,0x4d7fU,0xd20bU,0x5527U,0x6ed4U,0x8299U,0x76b2U,0x6782U,
		0xa469U,0x3a16U,0xd9cdU,0xd6b4U,0x9e35U,0x72f5U,0xae6dU,0xfb6cU,0x5c94U,0xdab4U,
		0x6518U,0x2da9U,0x6e3aU,0xbad8U,0xe3b4U,0xbe0bU,0xb006U,0xcfd5U,0xcb07U,0xece4U,
		0xa5e4U,0x1c7fU,0x07b5U,0x0748U,0xccffU,0x3f20U,0xd80eU,0x89c6U,0xbda4U,0xc2ccU,
		0x424eU,0x8dd7U,0x4390U,0x360fU,0x4509U,0x8007U,0x8512U,0xdcf2U,0x08c0U,0x921aU,
		0x746cU,0x3ccfU,0xcb85U,0x79eaU,0x3c1eU,0x7543U,0x9f33U,0x7854U,0x9128U,0xe69eU,
		0x7dcdU,0x8ff5U,0xe1a9U,0x1743U,0xf083U,0x979fU,0x0bd2U,0x2755U,0x9063U,0x63a7U,
		0xd259U,0xbf3bU,0x44f0U,0x6f12U,0x8d31U,0x898eU,0x1f8eU,0x9e65U,0x3772U,0xb3a0U,
		0x410bU,0xd59aU,0x71abU,0xf6feU,0x4906U,0xf9d1U,0x50efU,0x5bb3U,0x2642U,0xaa8cU,
		0x5474U,0x9380U,0xc096U,0x91b7U,0xb3beU,0x46e9U,0x40b6U,0x1964U,0xb868U,0x8ea5U,
		0xa93fU,0x4c68U,0x0989U,0xfd3fU,0x2649U,0x88d1U,0x47b5U,0xd6dbU,0xad86U,0x088eU,
		0xbd59U,0xad6eU,0xfa14U,0x7adcU,0x7d37U,0xa8f7U,0x7e8fU,0xd178U,0xb05aU,0x8cb3U,
		0x3d04U,0xb859U,0x30c6U,0x4754U,0x71d8U,0x2125U,0x6ea3U,0xd547U,0xff7aU,0x678bU,
		0x9d3aU,0xc9c1U,0xdc45U,0x9a4cU,0x43c4U,0x4bd1U,0x7cd9U,0x8132U,0x9a23U,0x13acU,
		0xe8e0U,0xe3e7U,0x870bU,0xbc2bU,0x0aa5U,0x30bdU,0xb349U,0xf350U,0x94a7U,0x1d59U,
		0x125dU,0xc1cdU,0x818eU,0xe8c9U,0xaadcU,0xb775U,0xd6caU,0x8f5aU,0x45ebU,0xb5e6U,
		0x13d0U,0x7ff7U,0x2d75U,0xe96dU,0x47faU,0x0ba1U,0xbcf4U,0x3aa1U,0x48dbU,0x2374U,
		0x7ce8U,0x2dd9U,0x4657U,0x60e4U,0x533aU,0xdd69U,0x2115U,0x8e64U,0x4615U,0xbdd1U,
		0x8456U,0xdb7bU,0xafecU,0xf6a9U,0x3994U,0x3b7dU,0x9ad5U,0x0b41U,0x9a03U,0xb463U,
		0xbbb1U,0xd3dbU,0x5bedU,0x52afU,0x494eU,0xf870U,0x0731U,0x471eU,0xae0dU,0xce8eU,
		0xddd0U,0xcaffU,0xc93aU,0x9f6bU,0x7e44U,0xc9d1U,0x56e9U,0x6ca9U,0x9ee6U,0x2891U,
		0x7d70U,0x0435U,0x5c50U,0x47e5U,0x2accU,0x1c9eU,0x3a79U,0x4ddeU,0x0998U,0xa9d5U,
		0xa329U,0x6b63U,0x56c1U,0x6049U,0x063cU,0x2f58U,0x959aU,0xe35aU,0xdfbbU,0x6c34U,
		0x9116U,0x62a8U,0x9387U,0xf490U,0xe6cfU,0x4218U,0x9a75U,0xd074U,0x860dU,0xfc28U,
		0xd507U,0x2847U,0x4038U,0xee01U,0x1857U,0x42f6U,0x466eU,0x3285U,0xf748U,0xbc22U,
		0x5447U,0xc630U,0x0f6eU,0xee1bU,0xdb27U,0x2815U,0x9083U,0x502fU,0x7622U,0x97cfU,
		0xd794U,0xba95U,0xc919U,0xd980U,0xa5b1U,0x2492U,0xf50bU,0x2d8eU,0x0e94U,0x5a10U,
		0xfdf3U,0xd957U,0xbd30U,0xa50aU,0x23e6U,0x5dbdU,0xed93U,0x940aU,0x3b11U,0xcbe7U,
		0xfef9U,0xecabU,0x11f3U,0xaedeU,0xddd8U,0xe9edU,0x24a7U,0x7c38U,0xab74U,0xb654U,
		0x3eb8U,0x821aU,0x7822U,0x8937U,0xbaeeU,0xa6bfU,0xcf1aU,0xb97aU,0x2e4aU,0x6d5aU,
		0x9d4eU,0xe0e4U,0xfe9dU,0xadd7U,0x5d06U,0x1e38U,0x5bc2U,0xab39U,0x81f8U,0xae8dU,
		0xd4baU,0xc8c2U,0x05faU,0x110bU,0xb81dU,0x237eU,0xbe6aU,0x1e80U,0x2276U,0xcf97U,
		0x4c72U,0x1e26U,0x146eU,0x02fbU,0xb1a5U,0x9224U,0xc521U,0xdd99U,0x28e0U,0xbcdbU,
		0x9072U,0x4d97U,0xd5cdU,0x0239U,0xc154U,0xb0dfU,0xd67aU,0x5e8eU,0xf062U,0x161eU,
		0xa54aU,0xa540U,0xb7aaU,0x5677U,0x34ceU,0xa64fU,0xd141U,0x822fU,0xc6b9U,0xb359U,
		0x637aU,0x2fb8U,0x9260U,0x6016U,0x640bU,0x4196U,0xfb7eU,0xeaccU,0xc4c6U,0xb788U,
		0x547bU,0xbd6dU,0x6b6dU,0x4468U,0x2b39U,0xe45eU,0x8e4cU,0x7b52U,0x33a2U,0xb65eU,
		0x1dacU,0x7482U,0xdf81U,0x4b25U,0xb4d6U,0xcdd9U,0x3c50U,0xec13U,0xd0b6U,0x2c9aU,
		0x3174U,0x9e1aU,0xd0ddU,0xffcaU,0x7891U,0xdc3cU,0x79f9U,0x083fU,0x4e49U,0x70f8U,
		0x763aU,0x3906U,0x98e8U,0xd318U,0x6636U,0x60f5U,0x79a6U,0xe4deU,0x7b7dU,0x93a8U,
		0xc0bbU,0x841dU,0x0b0dU,0x817bU,0xdc89U,0xc320U,0xe7c3U,0x3c74U,0x091eU,0x2faeU,
		0x2352U,0x7e0cU,0x5eb4U,0x5e6bU,0xfa45U,0x657dU,0x4425U,0x1ee1U,0xba2fU,0x8404U,
		0x2a1dU,0xe24aU,0x7fa4U,0xba10U,0x7434U,0xebc3U,0x5171U,0x52c6U,0x0c7cU,0x606eU,
		0xbeb1U,0xfc8dU,0x1aa1U,0x7ebfU,0x6674U,0x4cfbU,0x9106U,0x8bd7U,0x8058U,0x6bceU,
		0x744fU,0x5adaU,0x151cU,0x51ecU,0x9804U,0xad96U,0x0658U,0x1eb6U,0x8c6cU,0xfa7cU,
		0x9638U,0xadb4U,0x4a23U,0xd4aaU,0xb22eU,0xdd4eU,0xaf98U,0x5f71U,0x1874U,0xc6f5U,
		0x669bU,0xb4ccU,0xa069U,0xb92dU,0xb24dU,0x66b4U,0x18a8U,0x8d74U,0x4f3fU,0x5aebU,
		0x6ec4U,0xd2dcU,0xb972U,0x8dcfU,0x6531U,0x1d48U,0xb5d3U,0x27a1U,0x1b9cU,0xffefU,
		0x1041U,0xb763U,0x2e68U,0x26feU,0x1f66U,0xcc03U,0x1df1U,0x124dU,0xf9d8U,0x832dU,
		0xae7cU,0xc817U,0x9d68U,0xc884U,0x1f07U,0xb56eU,0x1d36U,0xc02dU,0x890fU,0x24c2U,
		0x87a2U,0xe04cU,0xb42bU,0x8c2fU,0x370dU,0xd9e0U,0x1cb0U,0x7edbU,0x078cU,0xa484U,
		0x64b2U,0x4bb0U,0x7d84U,0x2879U,0x8b90U,0x10e2U,0x2789U,0x3ee8U,0x5877U,0x217fU,
		0xb30cU,0x70d6U,0x5e3fU,0xe075U,0xb711U,0x65c4U,0x7a38U,0x90a5U,0x3483U,0x9e10U,
		0xb673U,0x17e7U,0xf88dU,0x76e0U,0x64c7U,0x7ba8U,0xb9ceU,0xcee0U,0x9ec4U,0x0044U,
		0x9f73U,0x10bfU,0xb2e1U,0x4852U,0x92f3U,0x2f0aU,0xa0a5U,0x9cbcU,0xcd25U,0x25e8U,
		0x71f3U,0x5e15U,0x05d0U,0x1d50U,0xb071U,0x6eb4U,0x2a93U,0x414cU,0xeb59U,0x397aU,
		0xc4e2U,0xe165U,0x4003U,0xee78U,0x6996U,0x24bbU,0xcd65U,0x839cU,0xa8f0U,0x673eU,
		0x747bU,0xedfaU,0xa952U,0x2950U,0xbedfU,0x3ab3U,0x9241U,0x7e74U,0xea4fU,0x131aU,
		0x0beeU,0x98deU,0x9ed4U,0x2393U,0x51baU,0xacefU,0xab5bU,0x6f56U,0xb0caU,0x9c44U,
		0xd23cU,0x5ce0U,0xa3edU,0x4a82U,0x03bfU,0x4a96U,0x5b2aU,0x55baU,0xf4f6U,0x6eecU,
		0x0f1cU,0x2e83U,0x7072U,0x4ea7U,0x17acU,0x031eU,0xb5b1U,0xd194U,0x9423U,0xea83U,
		0x683aU,0xcd3aU,0x03a4U,0x83daU,0x7672U,0xe583U,0x490dU,0x0460U,0xc003U,0xa284U,
		0x41ccU,0x91c7U,0xa115U,0xf6deU,0x365fU,0x4ca0U,0x60a5U,0x699dU,0x1236U,0x8844U,
		0xd092U,0x422fU,0xe862U,0x353cU,0x0aaeU,0xe656U,0xe6f7U,0xa7eeU,0x3234U,0xfd9eU,
		0xabb2U,0x76bbU,0xfbfdU,0x24f5U,0x789aU,0xf11cU,0xc37dU,0xedc1U,0x31cbU,0x8171U,
		0xff65U,0x52eaU,0xd02cU,0xc98fU,0x08e3U,0xf751U,0xd3b4U,0xdad0U,0x3f2fU,0x07e1U,
		0xc3fdU,0x05d9U,0x6d28U,0x1cceU,0xc968U,0x9c66U,0xd998U,0xd583U,0x54a7U,0x5863U,
		0xf379U,0x6814U,0xa660U,0xb51bU,0xc5f3U,0x40d7U,0x1c2dU,0x739eU,0x73c2U,0x9dc2U,
		0x9764U,0x32cdU,0x84e5U,0x59c4U,0xa382U,0x3d8aU,0x0b2aU,0x0089U,0xf370U,0x45e2U,
		0xe287U,0x725fU,0x8b76U,0x94f1U,0xaf74U,0xe909U,0x0d84U,0x1d64U,0x0f34U,0xfab8U,
		0x57caU,0xba25U,0x5984U,0x9b84U,0xc9f3U,0x06c0U,0x8cbeU,0xc4afU,0x338fU,0x8664U,
		0xeda6U,0xaa97U,0x93ebU,0x080eU,0x9881U,0xbc3cU,0xe6a9U,0x6f3dU,0x824dU,0xd4ceU,
		0x3f84U,0x6f69U,0xd296U,0xe144U,0xf32aU,0x9a9dU,0x127aU,0x3b5dU,0x6ac1U,0xce32U,
		0x206bU,0x5a8dU,0x3f6dU,0xecefU,0x8870U,0x143cU,0x8021U,0xe76fU,0xce5dU,0x9784U,
		0xa3a9U,0x5c2cU,0x64e0U,0x33c5U,0x5bbaU,0x7464U,0x74b6U,0x3ffbU,0xd124U,0x11dfU,
		0x26bfU,0xd04aU,0x67eaU,0x5f05U,0x9f8fU,0x83a3U,0x4da2U,0xafa9U,0x70e6U,0x4369U,
		0x8963U,0x4adbU,0xc682U,0xede7U,0x8501U,0xf7adU,0x1069U,0xbd78U,0x0e7cU,0x2fa4U,
		0x36e8U,0x2bf2U,0x069aU,0x92dbU,0x0124U,0x8652U,0x1981U,0x6c72U,0x3911U,0xa860U,
		0x46b8U,0x56d8U,0x8bf1U,0x086fU,0x70a0U,0xe255U,0x8af9U,0x15f7U,0x9d03U,0x5f1fU,
		0x778aU,0x35deU,0x59bbU,0x2dd0U,0xc650U,0x2852U,0xc942U,0x3782U,0xa104U,0x78c8U,
		0x0440U,0xec49U,0x612aU,0xc105U,0xa8aeU,0xe0efU,0x53f0U,0x91e3U,0xa792U,0x6539U,
		0x165eU,0x4e2cU,0x209fU,0x0383U,0xd962U,0xfae5U,0x6b09U,0x3864U,0x74beU,0x7281U,
		0x74ecU,0x0c72U,0x0d20U,0x6cc2U,0x2709U,0xaa09U,0x2aebU,0xe48eU,0x4cc5U,0x2c84U,
		0xf0fdU,0x16c9U,0x548cU,0x9343U,0x0ae1U,0xfa87U,0x7322U,0xb7d4U,0x485dU,0x9944U,
		0xbb9bU,0x3dc7U,0x301bU,0x2339U,0x6a3bU,0xa48fU,0x1d13U,0x631aU,0x7908U,0x77d0U,
		0x6fccU,0xaa25U,0x2a73U,0x16a1U,0xeefcU,0x6749U,0xfb0cU,0x4351U,0xdb0aU,0xd708U,
		0x4eb2U,0xdae7U,0x9547U,0xcfb7U,0x6890U,0x9fb4U,0x904dU,0x649fU,0x9c0bU,0x479aU,
		0xabd4U,0xfb56U,0x50f6U,0x884fU,0x09f0U,0x57f8U,0x35f7U,0xd63fU,0xa2c6U,0x85a2U,
		0x49b6U,0xd68cU,0x1cd9U,0xcd6cU,0xfdedU,0x56e3U,0x7b77U,0xc14eU,0x9e2fU,0xba06U,
		0x8705U,0x376cU,0x8bebU,0x8247U,0xfaf0U,0x1d0dU,0x7a32U,0xa063U,0x08ddU,0xd236U,
		0xee8bU,0x2101U,0x199dU,0xd33eU,0x4b5dU,0x7a5bU,0x5185U,0x6f74U,0x81b2U,0x1753U,
		0xfaf6U,0x5127U,0x83aeU,0xb114U,0x3caaU,0xbe7dU,0x88f9U,0x576fU,0x9d88U,0xe5f3U,
		0x1725U,0x36adU,0x5d77U,0xd35fU,0x04ccU,0xd4eaU,0x1566U,0x0e4fU,0xf367U,0x210cU,
		0x488bU,0x3154U,0x5803U,0xe7eaU,0x371dU,0x5d6cU,0x21c8U,0x9b07U,0x5ed2U,0x6166U,
		0x8924U,0x696cU,0xaf08U,0x3dd2U,0x99e0U,0x9dacU,0xc8efU,0x1aedU,0x9babU,0x7204U,
		0x9c54U,0x23c3U,0x78e0U,0x7419U,0xe4feU,0xf5c7U,0xdc78U,0xe2acU,0x1d28U,0xb420U,
		0xe427U,0xee49U,0xfbe7U,0xb2afU,0x917dU,0xbb2bU,0x50a0U,0x6bb8U,0xace6U,0x6f09U,
		0x8207U,0xf18cU,0x75b5U,0xcb74U,0x7a61U,0x4a15U,0xf201U,0x92b7U,0x47ccU,0xbb7aU,
		0xcc8fU,0xa9b0U,0xf5edU,0xa24eU,0x60b9U,0xf660U,0x0365U,0x2ea6U,0xb9c0U,0xa479U,
		0xf319U,0xb83cU,0x3889U,0x6195U,0xbde7U,0xa4b7U,0x3ce1U,0x6c8dU,0x7954U,0x7cfbU,
		0x196bU,0x928aU,0x075dU,0x6229U,0x952cU,0x077eU,0xee2fU,0x8791U,0x9504U,0x47dbU,
		0xcf59U,0x7e7bU,0x6502U,0xb2f9U,0x50bdU,0x0ca1U,0x97c7U,0x4d77U,0x2beaU,0x10daU,
		0xca34U,0x7754U,0xaf29U,0x752bU,0xdf47U,0x0494U,0x9233U,0xb301U,0xed4bU,0x9d15U,
		0xdbcbU,0x86ceU,0x3934U,0xc548U,0x57e0U,0x7aedU,0x5b56U,0x6e5eU,0x4b63U,0x4924U,
		0xd38aU,0x194eU,0xea17U,0x5b1cU,0xdf58U,0xfe46U,0x333aU,0xa67dU,0xfd22U,0xb2beU,
		0xdd17U,0x4202U,0x4726U,0x3da9U,0x0ed6U,0x3a41U,0x31bdU,0x17dcU,0xb205U,0x20cbU,
		0x299dU,0xeeb4U,0xa30aU,0xdee8U,0xa2f5U,0x5af3U,0x96baU,0xf4b6U,0xdeacU,0x56b7U,
		0xbe9bU,0xd47bU,0x59dfU,0xa374U,0x5c34U,0xc752U,0x2ff0U,0x3fc0U,0x7092U,0xa09aU,
		0x9189U,0x847eU,0x1fc7U,0x7929U,0xc1f7U,0xfdd3U,0x12a5U,0x9737U,0xd412U,0x9513U,
		0xea05U,0x7210U,0x2770U,0x9670U,0x7960U,0xa8a0U,0xa979U,0xf195U,0x7340U,0x768cU,
		0xfcc6U,0xa25aU,0x6f7aU,0x0824U,0x03b6U,0x71a2U,0xec81U,0xec78U,0x0d76U,0xced5U,
		0xb0b2U,0x4d83U,0x5f3eU,0xd889U,0xb7b9U,0x3d2dU,0x3940U,0x9c31U,0xa715U,0x329cU,
		0x0afbU,0x712aU,0xd42fU,0xb638U,0x267dU,0x3f5fU,0xa962U,0x75e1U,0x40bdU,0x420eU,
		0x307aU,0xe98bU,0x559cU,0x8f1bU,0xb6acU,0xdcbcU,0x2edeU,0xd284U,0x96c6U,0x9248U,
		0x02acU,0x32a8U,0x3425U,0x116bU,0x5d1fU,0x4586U,0xf359U,0x2a88U,0x2cd5U,0xc79eU,
		0xb6b8U,0xcb1fU,0x764bU,0x3a1fU,0x4c4eU,0xe1f4U,0x0848U,0x63d5U,0x69faU,0x9617U,
		0x03f5U,0x61a4U,0xf404U,0xfddfU,0x1188U,0xf2ecU,0x518bU,0xc639U,0xfcd2U,0x2f73U,
		0x01f7U,0x9c96U,0xf4efU,0x6ebdU,0xeef5U,0x1273U,0x50e8U,0x5e0aU,0xb7a3U,0x2ac5U,
		0x8fceU,0x3248U,0x7268U,0x8a91U,0xb0abU,0xf5e6U,0xafc0U,0xf5daU,0xfcdeU,0xd4fdU,
		0x717dU,0xf048U,0xed6cU,0x6a2dU,0x1a0fU,0x285bU,0xb8a4U,0xc04aU,0xda97U,0x3a2bU,
		0x315dU,0x41a8U,0xb797U,0x0d9dU,0x195fU,0x979aU,0x7ff2U,0x6786U,0x8393U,0x753aU,
		0x0743U,0xa091U,0x6011U,0x1e21U,0x1edcU,0xdc37U,0x2f53U,0x52aaU,0xb64fU,0x502aU,
		0xea7eU,0x7e6fU,0xb516U,0x76b6U,0xb75eU,0xfa77U,0x2f05U,0x2874U,0x9fafU,0x933eU,
		0x2107U,0xa05eU,0xe76aU,0x06bbU,0xec44U,0x2f9fU,0x7125U,0x950eU,0xf5e1U,0x3a1aU,
		0x6190U,0xee44U,0x17d7U,0x048fU,0x7566U,0x7dbfU,0xdeb6U,0xaa3dU,0x52a0U,0x52a5U,
		0xd327U,0x1a67U,0x85d0U,0x23fdU,0x11baU,0xa46dU,0x8eacU,0xca53U,0xf464U,0x40c7U,
		0x2634U,0xd49fU,0xc468U,0x9324U,0xfce8U,0x2483U,0x5546U,0x1321U,0x4cebU,0x052bU,
		0x7875U,0x552bU,0x75a9U,0x8591U,0xc285U,0xa999U,0xc76dU,0x425eU,0x30aeU,0x85ecU,
		0x9585U,0x1a6cU,0x8cdcU,0x650aU,0x75a1U,0x97bfU,0xf56dU,0xd20fU,0xf740U,0x6bacU,
		0x895bU,0x3705U,0x7c1bU,0x406fU,0xee91U,0x7845U,0x8da9U,0x4b72U,0xa88eU,0x8c61U,
		0x5270U,0x7bb9U,0xf6eeU,0x829dU,0x7a0cU,0x9f1aU,0xfbdbU,0xadc6U,0x0739U,0x9acbU,
		0x3808U,0xc897U,0x4976U,0xd32cU,0x03d7U,0x89e4U,0xde44U,0x5577U,0x0e0bU,0x6ed8U,
		0x70b7U,0x4f47U,0x3c5fU,0x740dU,0xdbf5U,0xe8f6U,0xeb96U,0x4f9dU,0xce7cU,0x51e1U,
		0xa3b1U,0xde28U,0x58e1U,0x14bbU,0x1348U,0xfb70U,0x4700U,0x1bd6U,0x3855U,0x48e5U,
		0x31b1U,0xe22dU,0x5b76U,0x72c3U,0xa949U,0x7dc4U,0x87f6U,0xc1eeU,0x48a0U,0x08caU,
		0xd66cU,0xae71U,0x703fU,0x272dU,0xb4a9U,0xce23U,0x6f9dU,0x41f6U,0xd344U,0x11cbU,
		0x0cbdU,0x25abU,0x52d5U,0xcc70U,0xa554U,0x5ea6U,0xafc7U,0xdab8U,0xa054U,0x6c9fU,
		0xf12fU,0xe443U,0x53afU,0x09abU,0xbe8fU,0x0928U,0xcdeaU,0x756bU,0x2467U,0x6603U,
		0x3c42U,0xc022U,0xc3b3U,0x5c98U,0xb86fU,0x3db8U,0xa333U,0xda67U,0x1334U,0xf7e7U,
		0x9469U,0xeea8U,0x1495U,0xab7eU,0x5e53U,0xea56U,0x915dU,0x0335U,0x66a5U,0xd6b8U,
		0x221aU,0x06e5U,0x2361U,0x8448U,0xc742U,0x348dU,0xc5b1U,0x7ea1U,0x64f6U,0xaa42U,
		0x2df3U,0xa635U,0x74d7U,0x797cU,0xb5c1U,0xd9d1U,0x0bf8U,0x373dU,0x062eU,0x4f34U,
		0x243dU,0xe1e4U,0x19a3U,0x5f9fU,0x8e30U,0x751fU,0xbf9cU,0x31d5U,0xea74U,0x0727U,
		0x2bb7U,0x72f9U,0x4123U,0x0e86U,0x2b71U,0xe0a6U,0x9bc0U,0xbc63U,0x2f8fU,0x9aeeU,
		0x7bd6U,0xdebbU,0x57d4U,0x21b4U,0x2f35U,0x32beU,0xe18fU,0x9e39U,0xcf0cU,0xffe4U,
		0x2592U,0x66ecU,0x74f6U,0xbe1bU,0xca04U,0x65f3U,0x770dU,0x140aU,0xa17aU,0x1942U,
		0x5649U,0xb5f2U,0x9feeU,0xbe0fU,0x29e7U,0xff10U,0x463bU,0xa039U,0x3956U,0x3d7fU,
		0x5346U,0xc6d6U,0x2a2eU,0xca58U,0x0f9dU,0x25bbU,0x900eU,0x6a0fU,0xe183U,0xe3b8U,
		0x16f1U,0xd898U,0xffa7U,0x2131U,0x1552U,0xb327U,0xb11aU,0xcac4U,0xe8ecU,0xe0daU,
		0x3ae6U,0x0d38U,0xeb83U,0x21a8U,0x324fU,0xcfd9U,0xf246U,0x5504U,0x7d43U,0x49a1U,
		0x2d46U,0xe718U,0xcbc1U,0x73b7U,0xfd5bU,0x8eb1U,0x6257U,0x835fU,0x4117U,0x5327U,
		0xa0caU,0xb00aU,0x91beU,0x0885U,0x017dU,0x8f12U,0x54eaU,0xa6eeU,0xf1acU,0x97abU,
		0x05a0U,0x1dbeU,0xfc37U,0x2957U,0xc667U,0x716bU,0xd461U,0x2dadU,0xd217U,0x3e07U,
		0x4c07U,0x3731U,0x8cfbU,0x8c97U,0x177cU,0x69c7U,0x1e5eU,0x40ccU,0x58f4U,0x0ca9U,
		0xa0b8U,0xca8fU,0xbdb0U,0x651cU,0x5eedU,0xd55cU,0xbf90U,0x3bacU,0x0e01U,0xbae4U,
		0x83b4U,0xefa5U,0x55ccU,0xb60bU,0xcb56U,0x024cU,0x4067U,0x6e56U,0x4692U,0xbadcU,
		0x5a08U,0xd6acU,0x56d0U,0x65bcU,0xce96U,0x67ffU,0x95cfU,0xf902U,0x5a79U,0xf469U,
		0x6b94U,0xc95dU,0x723cU,0x0a7eU,0xabc2U,0x6e3eU,0xc736U,0xd2e4U,0x28b2U,0x7158U,
		0xddfdU,0x1596U,0x7c72U,0xf215U,0x31ffU,0x7e29U,0x33cdU,0xf726U,0x4c7eU,0x8166U,
		0x3137U,0x074cU,0x4c8cU,0xf423U,0xd92bU,0x2d0aU,0x87ffU,0x1c50U,0x2affU,0x5e9aU,
		0xa82cU,0x2402U,0xc9e0U,0x5287U,0x3e61U,0xe101U,0x894aU,0x07b9U,0x89daU,0x8210U,
		0x6417U,0xf70fU,0x0bb6U,0xcc55U,0xbe83U,0xc843U,0x67b4U,0x748eU,0xec8aU,0xa819U,
		0xb216U,0x581cU,0x8fd5U,0x3f24U,0x0f49U,0x7938U,0x335fU,0xe4eaU,0xe10fU,0xda5bU,
		0xc572U,0x4f75U,0x3180U,0x85d5U,0xde5cU,0x944aU,0x63beU,0xfc99U,0xae3eU,0xcd03U,
		0xc09dU,0xc259U,0x473aU,0xda33U,0x1489U,0x0186U,0xfa51U,0xadf8U,0x7669U,0x0980U,
		0xae35U,0xa959U,0x0317U,0x279aU,0xdae0U,0xece8U,0x91b0U,0x4224U,0xb352U,0x6b5cU,
		0xda54U,0xe711U,0x6b36U,0xd738U,0x9c2aU,0xa472U,0x09a4U,0x7db8U,0x7a02U,0xfeefU,
		0x34fdU,0xcb0bU,0xf9acU,0x1544U,0x9a12U,0x88b5U,0x39a0U,0x3b46U,0x3cb0U,0x5450U,
		0x8fe3U,0xbc94U,0x3849U,0x504dU,0x331bU,0xb07aU,0x3b1dU,0x1c83U,0x685bU,0x164dU,
		0x5a6bU,0xe6ecU,0x4839U,0xa6cbU,0xe290U,0xeeccU,0xdc0eU,0x11bfU,0x6a5dU,0x6461U,
		0xcd11U,0x09d6U,0x21e2U,0xa5e8U,0x1863U,0x23aaU,0x582dU,0xc659U,0xc698U,0x44c7U,
		0xe92cU,0xdf9dU,0x4894U,0x734fU,0x1e0fU,0xbaa1U,0x889cU,0x3054U,0x7869U,0x8ddbU,
		0x3accU,0x1b31U,0x3ec7U,0x3924U,0xf9e4U,0x6003U,0xff2aU,0x9a5eU,0x6ba2U,0x2488U,
		0x9e79U,0xd8e8U,0x613eU,0x01e1U,0xc603U,0x4252U,0x1b0cU,0x51f8U,0x50d8U,0xe309U,
		0x6182U,0x55afU,0x1759U,0xa76eU,0x2247U,0xa8deU,0x1880U,0x3306U,0xf734U,0x4489U,
		0x8a98U,0x3613U,0xcd31U,0x8e9cU,0x9441U,0xb62fU,0x86a6U,0xd73fU,0xf09aU,0x7593U,
		0x3004U,0xfcedU,0x2208U,0x76ecU,0x1676U,0x8c55U,0x5883U,0x4394U,0x46f0U,0x5928U,
		0xe217U,0x73eeU,0xf61dU,0x8c3bU,0xce6eU,0xb5a6U,0x982cU,0xd7cdU,0x5405U,0x3abaU,
		0x01aeU,0x3e9dU,0x096aU,0x89caU,0xf575U,0xfd0bU,0x03cbU,0x0545U,0xa884U,0x23dcU,
		0x027aU,0xeffeU,0x35a1U,0x1326U,0x2d22U,0x50c5U,0x366bU,0x38d7U,0xf236U,0xd812U,
		0x1e6cU,0x2087U,0xcba4U,0x9f0eU,0xa163U,0x469aU,0x81b8U,0xea8fU,0x7031U,0xa279U,
		0xab48U,0x7226U,0xa3ccU,0xfdaaU,0x6b1fU,0xc2d0U,0xdbe9U,0xcca1U,0x5247U,0xb18cU,
		0xae15U,0x19e2U,0x08efU,0xe824U,0xc331U,0x554bU,0x379eU,0xea66U,0xf565U,0xdbc3U,
		0xf22eU,0xbda8U,0xc911U,0x6eccU,0x011cU,0x586fU,0xb619U,0xb153U,0xe839U,0xfe13U,
		0xa38eU,0xe746U,0x64e8U,0x3654U,0xba8fU,0x4718U,0x70f2U,0x921eU,0x5521U,0x327bU,
		0x9a46U,0x63a1U,0x92d5U,0x865eU,0x20fbU,0xb7ceU,0x3ee2U,0xd4a4U,0xf116U,0x98d8U,
		0x97b9U,0x7e69U,0x7976U,0x08c4U,0x9d0fU,0xbb25U,0x63cfU,0x966aU,0x38d1U,0x01dbU,
		0x512dU,0x7e63U,0xca89U,0x6a09U,0xfee9U,0xe0fbU,0xc20eU,0xe05dU,0x726fU,0x3cd3U,
		0x6e1eU,0xbc48U,0x4f0dU,0x18baU,0x2f47U,0x6b3dU,0x811cU,0xeae6U,0x0f13U,0x2639U,
		0x0f40U,0x5f35U,0x985eU,0x8552U,0xf1f3U,0x7470U,0xb3c5U,0x7fbdU,0x9092U,0x38ecU,
		0x6acdU,0xa085U,0xcf32U,0x0fc7U,0xcbcfU,0xf841U,0xc7faU,0xbee6U,0xc763U,0x0f64U,
		0x68d2U,0x800bU,0x8496U,0x117aU,0xdb8dU,0x6978U,0xb335U,0x6473U,0x1cecU,0x93dbU,
		0x04d8U,0x9329U,0xf458U,0xa90cU,0xd762U,0x5133U,0x33ecU,0x450dU,0xed5eU,0x5cd5U,
		0x5deeU,0x8d47U,0xab0bU,0x0063U,0xfafcU,0x18cdU,0xaf55U,0x8261U,0x5759U,0x2477U,
		0x689cU,0xed80U,0x9b69U,0xdcf6U,0x3964U,0x2b16U,0x32e7U,0x2fe2U,0x6934U,0x05c0U,
		0xccb7U,0xd698U,0x7119U,0xc46dU,0xa519U,0x30ecU,0xf4dcU,0xe02fU,0x95ecU,0x8516U,
		0xfdbaU,0x4e01U,0x0aedU,0x0be3U,0xe852U,0xbfedU,0xf1c7U,0x4a5eU,0x0a60U,0xaa31U,
		0x77f5U,0x8defU,0x7b11U,0x0e18U,0x8bc1U,0x7858U,0x94b5U,0xb9e2U,0xad2aU,0x799bU,
		0xe739U,0xd1c2U,0x6597U,0xc7d0U,0x3193U,0x8596U,0x27b0U,0x6c1bU,0x8b3dU,0xbc4eU,
		0x4681U,0x9f37U,0x5460U,0x61c0U,0x7fd8U,0x82aaU,0x2c44U,0xe817U,0x36b3U,0xddadU,
		0x9bffU,0x0473U,0xc05cU,0xab9cU,0xb0c3U,0x6694U,0x6a34U,0xe6a2U,0x1f87U,0x38fbU,
		0xd5c6U,0x7e3dU,0x0bafU,0x154bU,0x38caU,0x617bU,0xfcbfU,0x75aeU,0x6f96U,0x9fa8U,
		0xb0a1U,0x13c6U,0x6337U,0x912cU,0x88d8U,0x2c3dU,0x0b7dU,0xe0c6U,0x6304U,0x5b8aU,
		0xebf6U,0x0538U,0x5629U,0xbd1dU,0xdafeU,0xacf6U,0x3569U,0xa3feU,0x4aebU,0x79eeU,
		0x4b7aU,0xde7bU,0xc1a1U,0xcd54U,0xe79bU,0x232dU,0x90c9U,0x3552U,0x8e8fU,0xa99eU,
		0xa2a3U,0x049cU,0xcc31U,0x6e24U,0xe74eU,0xcb89U,0x1f3eU,0xa1e3U,0x9d77U,0x772eU,
		0x78fcU,0xf471U,0x172bU,0x214cU,0xb068U,0x4387U,0x28a9U,0x96b1U,0x1c6bU,0xec08U,
		0xa34cU,0x7547U,0x6de1U,0xded3U,0x58a9U,0x7f09U,0x2899U,0x2286U,0x0031U,0x0c66U,
		0x7017U,0xc28aU,0x5ff6U,0xa52eU,0x4bdfU,0x6907U,0x2037U,0x3c22U,0x2abbU,0xb76bU,
		0xf47aU,0x27ceU,0xbcbdU,0x6ce8U,0xf0f1U,0x2fcfU,0x995eU,0x4f1cU,0x5f0dU,0x32f9U,
		0x931cU,0x8789U,0xb59eU,0x97a3U,0xfc20U,0xe5e7U,0xd042U,0xb566U,0xe236U,0xb88cU,
		0x02e0U,0x349aU,0xd518U,0x0530U,0xdff6U,0x7429U,0xc2eaU,0x18c0U,0xed2dU,0xf087U,
		0x04c0U,0xbb34U,0x00c3U,0x8a44U,0x88dfU,0x6e07U,0xd365U,0xa41cU,0xb9a7U,0x244aU,
		0xa263U,0x634cU,0x4fdaU,0x3448U,0xed73U,0x2759U,0x6b46U,0x24dbU,0xeb1fU,0x9afbU,
		0x1086U,0xf9b3U,0x6a75U,0x0266U,0x4ed6U,0x4cf0U,0x34b6U,0x4492U,0xf541U,0xca11U,
		0x018fU,0x0bd6U,0x2650U,0x9b2fU,0xc8e3U,0x20e6U,0xda8dU,0x5330U,0xac31U,0xaa53U,
		0x22f1U,0x79b8U,0x1ec5U,0x51c1U,0x899cU,0xd22aU,0xf14bU,0x8ff9U,0xbb18U,0x0ccaU,
		0xe3f0U,0x671fU,0x4a37U,0xf59dU,0x7773U,0xe7b8U,0x37ecU,0x5530U,0x5cabU,0xdb4eU,
		0x9e23U,0x4f13U,0x2420U,0x7dd1U,0xe15cU,0x9211U,0x7e20U,0xa36bU,0x61dcU,0xb0efU,
		0x5d7dU,0x42ffU,0xe668U,0x2ec0U,0x1d81U,0x638dU,0x37d1U,0x25f1U,0x1d92U,0x1747U,
		0x5a48U,0x101dU,0xa078U,0x27d8U,0xe9a4U,0x5e60U,0x3d93U,0x5da0U,0x9136U,0x787aU,
		0x6cf8U,0x3104U,0xa452U,0x661aU,0x1991U,0xe1adU,0xa18eU,0xf3a8U,0x82d4U,0x5f28U,
		0xeb38U,0xbc7fU,0xaa5dU,0xc0ecU,0xd4f3U,0x0b97U,0xe370U,0x6610U,0x897cU,0x8083U,
		0x5798U,0x6f16U,0x7b02U,0xc1b6U,0xc8a1U,0xd1daU,0x8756U,0xc6c9U,0xe3d6U,0xa7d0U,
		0x8b6aU,0x1a71U,0x7b39U,0x75cdU,0x7303U,0xe063U,0xe51dU,0x44f4U,0x72b0U,0x78edU,
		0x3210U,0x81edU,0xe0cdU,0x5692U,0x5775U,0xe44dU,0x6d3eU,0x9b78U,0x4605U,0x1c06U,
		0x02a2U,0x7e85U,0x2862U,0x8992U,0xd8c5U,0x6650U,0xf532U,0x2aa5U,0xb4bbU,0x88bcU,
		0xd485U,0x4994U,0x97f0U,0x958aU,0x9875U,0x6236U,0x54ccU,0x42c8U,0xb284U,0x8d35U,
		0xd6e2U,0xcf8cU,0xc4f1U,0xe995U,0xc223U,0x0372U,0x531aU,0xd520U,0xf052U,0x8742U,
		0x90d9U,0x6f5dU,0x75f2U,0x7f3aU,0x876eU,0x63abU,0x8db1U,0x8525U,0xaef0U,0x2e1cU,
		0x8015U,0x1c62U,0x9356U,0xea3fU,0xf884U,0x650fU,0x405eU,0xdf4fU,0xb172U,0xc214U,
		0xa396U,0x9067U,0xa690U,0xd8f2U,0xf649U,0xd904U,0x3528U,0x5a81U,0x88ffU,0x9d22U,
		0x3dfaU,0xe345U,0xf008U,0x076dU,0x7601U,0x3347U,0x99a0U,0xbf3fU,0xcb15U,0xd53dU,
		0x6bffU,0xe7aeU,0x9eeeU,0x487bU,0x26b3U,0xc3caU,0x01caU,0x8ce1U,0x2f19U,0xf719U,
		0xb08eU,0x5e83U,0xefafU,0xd25dU,0x44d7U,0x1049U,0xb3deU,0xdef5U,0x8ae9U,0x0d42U,
		0xba54U,0x5c20U,0xd654U,0x44b6U,0x67f2U,0xf98eU,0x321bU,0xadebU,0xdb5fU,0xd59eU,
		0x46acU,0xb132U,0xb682U,0x2de3U,0xd26cU,0xd43cU,0xd9f9U,0xd6edU,0xfd49U,0x4263U,
		0x215cU,0xff50U,0x85c8U,0x3cd9U,0x166eU,0x410fU,0xdfb3U,0x8646U,0xa3a1U,0x1f5eU,
		0x0b84U,0x974aU,0xe5f9U,0x3e1eU,0x54c4U,0xb099U,0x7cadU,0x9255U,0x5681U,0x969dU,
		0x1a16U,0xf702U,0x510dU,0xd41fU,0x36dcU,0x201eU,0xf3faU,0x9a19U,0x4fedU,0x7e16U,
		0x7cb8U,0xc772U,0x687eU,0xc308U,0x15e3U,0x6cd7U,0x8159U,0x71afU,0x47bcU,0xb918U,
		0x6053U,0x4acfU,0x4167U,0xe46eU,0x9d99U,0xb026U,0x2013U,0x8224U,0xf391U,0xab62U,
		0xa786U,0x0c5dU,0xb937U,0x9e69U,0x4849U,0x1c76U,0xf8f9U,0x3a38U,0x9973U,0x17f1U,
		0xcdb4U,0x6e7bU,0x8576U,0x85f1U,0x4af6U,0x428bU,0x99afU,0x7275U,0xc7eaU,0x1f92U,
		0x239dU,0xed19U,0x571fU,0xe681U,0xad35U,0x7376U,0x9d8eU,0x8e41U,0xac16U,0xe32cU,
		0x10f1U,0x52f4U,0x6d85U,0x6b84U,0x777eU,0xb3a4U,0x04f8U,0x2bfcU,0x55eaU,0x7dabU,
		0xaebbU,0xe9afU,0x4ec4U,0xf2f9U,0x2f69U,0x30b3U,0x9b8eU,0x2eb6U,0x3839U,0xa753U,
		0x7a7bU,0x3776U,0xe001U,0x5142U,0x3b39U,0xf2c1U,0x6cccU,0xeac1U,0x3694U,0x0e67U,
		0xc272U,0x2ce2U,0x39e1U,0x4ee1U,0x238cU,0xdd47U,0x993dU,0xaa90U,0xbf34U,0xcbdcU,
		0xdd92U,0xce87U,0x07b2U,0xe3b1U,0x08bdU,0x424bU,0xa899U,0x6bb1U,0xae6aU,0x76afU,
		0x40b3U,0x263fU,0xfa11U,0x2646U,0xf080U,0x9f30U,0x4108U,0x44edU,0x13cdU,0x818bU,
		0x2112U,0x48d8U,0xff77U,0x3d01U,0x0aa2U,0x7cd6U,0x0639U,0x3a76U,0xd504U,0x9384U,
		0x5beaU,0x3991U,0x9ee3U,0xddcdU,0xab71U,0xfef6U,0x5d03U,0xcf17U,0x9080U,0xf745U,
		0xbd2dU,0xa5aeU,0x925dU,0x34cbU,0x339fU,0x5478U,0x4c6fU,0x05f7U,0xd677U,0x28ddU,
		0x4422U,0x091bU,0x1a9eU,0x7431U,0x788eU,0x3c4dU,0xc0b8U,0x98e5U,0x103eU,0xb96fU,
		0x1d33U,0xf9d5U,0x8c69U,0x744cU,0xb24aU,0xaf95U,0x92f0U,0xb9cbU,0xc4dfU,0x05cdU,
		0x7d81U,0x370aU,0x3480U,0xb309U,0x9420U,0x0f19U,0x365cU,0x490aU,0x923eU,0xa8edU,
		0xa3eaU,0x51b7U,0xa65dU,0xc965U,0xf36dU,0x9761U,0xabafU,0xe85fU,0xd3b1U,0x31c8U,
		0x801eU,0x6abeU,0x67e7U,0x5bb7U,0xc9f0U,0x0d81U,0x3f81U,0x93e8U,0x043dU,0x59b8U,
		0x6b06U,0xa78fU,0x0e79U,0x8960U,0x709dU,0x197eU,0x688dU,0xfb09U,0x49b3U,0x50f3U,
		0x5489U,0x2706U,0x7905U,0xbb98U,0xf364U,0x1722U,0x99ddU,0x21c5U,0x7a2fU,0x9e2cU,
		0x83abU,0x4b5aU,0x3886U,0x60b6U,0x9501U,0x1968U,0xe424U,0x78ddU,0xf1feU,0xace3U,
		0x31baU,0xfd1fU,0x59dcU,0xa2f2U,0xdf44U,0x97c4U,0xd387U,0x3931U,0x0af8U,0x5f3bU,
		0x2edbU,0x40baU,0xd40fU,0x9186U,0x03b3U,0xa976U,0xb0a8U,0x50e5U,0x315aU,0xed69U,
		0x7648U,0x5d1cU,0xfccfU,0x03f2U,0x529dU,0x618dU,0xfce5U,0x8ea9U,0x2f50U,0x8390U,
		0xe767U,0xb75bU,0x4973U,0x7a09U,0xce79U,0x70b4U,0x9582U,0x75a6U,0x8da6U,0xf73dU,
		0xcde7U,0xa051U,0x1492U,0xb86cU,0xa946U,0x46fdU,0x0cbaU,0x703cU,0x7bd3U,0x4120U,
		0x770aU,0xcf09U,0x64f3U,0x2217U,0x8e2dU,0x0bf5U,0xfd58U,0xf243U,0x059dU,0x91bbU,
		0x2a2bU,0x29e4U,0xe8e9U,0x16eeU,0x5a76U,0x5a05U,0x31fcU,0xc733U,0x1e5bU,0xd214U,
		0x55c9U,0x5eeaU,0x317dU,0x0f46U,0x7666U,0xc09aU,0xa829U,0x4c89U,0x67b1U,0x89d7U,
		0xdc0bU,0x6858U,0x4891U,0x1860U,0x9c27U,0x91adU,0x8fe0U,0xf9a9U,0x3001U,0xcd2eU,
		0x9829U,0x46edU,0x6b9fU,0x3ac9U,0x2244U,0x1b09U,0xc32eU,0xdbe6U,0xa38bU,0xc90eU,
		0x359eU,0xf572U,0x702eU,0x1e69U,0x0f10U,0x6e1bU,0xcbccU,0xb3c2U,0x3edfU,0x551eU,
		0xca86U,0x9d0cU,0x7116U,0x3961U,0x0a5dU,0xfdb7U,0x04d5U,0x8493U,0xaf52U,0xed5bU,
		0xfcbcU,0x1f84U,0x5626U,0x88d5U,0x3190U,0x94b2U,0x9bfcU,0x545dU,0x7014U,0x6ddeU,
		0x995bU,0x2ab8U,0x8e8cU,0x4b77U,0xb065U,0x1f3bU,0x4ed3U,0x6b43U,0x22eeU,0x264dU,
		0xd515U,0xfc1dU,0xb9a4U,0x04bdU,0x9133U,0x5a45U,0xd4f0U,0xa18bU,0x37e9U,0xbb15U,
		0xe665U,0xe159U,0x97edU,0xd8c2U,0xf04fU,0xd6dfU,0x8b67U,0x7affU,0x6d3bU,0x72adU,
		0x01c7U,0xcb12U,0xd651U,0x44d4U,0xf881U,0x8daeU,0x3df7U,0xa68dU,0x7cb5U,0x510aU,
		0x2010U,0x47b9U,0xfd46U,0x46a9U,0x54c1U,0xdfb0U,0x2f66U,0x04f5U,0xc26fU,0xdffeU,
		0x8573U,0x4846U,0xac13U,0x239aU,0x907dU,0x5be7U,0x788bU,0x4c6cU,0xa896U,0xbf31U,
		0xff74U,0xf07dU,0x0e76U,0xc9edU,0x7a2cU,0x5486U,0x7d7eU,0x8c66U,0xabacU,0x923bU,
		0x957fU,0x2f4dU,0x64f0U,0xa943U,0xdf41U,0xe421U,0x7645U,0xd40cU,0x359bU,0x6b9cU,
		0x04d2U,0x3edcU,0x1e58U,0x2a28U,0x9c24U,0xa826U,0xf87eU,0x8b64U,0x8570U,0xfd43U,
		0x8e89U,0x318dU,0x37e6U,0xd512U,0x7d78U,0x8e86U,0x846eU,0x691bU,0x7d7bU,0xa893U,
		0x1e55U,0xdf3eU,0xdf6fU,0x9d59U,0xf527U,0x098dU,0x691eU,0x2e70U,0x81dfU,0x254bU,
		0x191aU,0x7ecfU,0xfdfeU,0x927aU,0x8471U,0xfba8U,0x9f97U,0x83f0U,0x9681U,0x5019U,
		0x3c91U,0x7ae0U,0x2e73U,0xfe22U,0x2e5cU,0x53cfU,0xe9fbU,0x0bc8U,0x6211U,0xdfecU,
		0x6921U,0x4074U,0x7813U,0x8c77U,0x20baU,0xeaecU,0xb2ecU,0xfa18U,0x254eU,0x5f61U,
		0x9841U,0x8dc3U,0x633eU,0x0797U,0x0e55U,0x0a8bU,0x81e2U,0x9f60U,0xa600U,0xa31aU,
		0x9982U,0xd46eU,0x7184U,0xa8fbU,0x9d5cU,0x6be6U,0xb9f6U,0xac93U,0x707dU,0x3504U,
		0xcb34U,0x5d92U,0xdf72U,0x7c20U,0x9c75U,0x3fdeU,0x17cfU,0x035dU,0x9439U,0x7d3bU,
		0x0990U,0x1241U,0x2063U,0x6529U,0xe637U,0x12d5U,0x6711U,0x1396U,0xf52aU,0xd5beU,
		0x6d33U,0x9edbU,0xbdfbU,0x46c3U,0x804cU,0x0892U,0xfbabU,0x2744U,0x429eU,0xb2a2U,
		0x507fU,0x4435U,0xd96dU,0x0dcbU,0x8474U,0x784aU,0xcae4U,0xe5ddU,0x2823U,0x8122U,
		0xce47U,0xad8aU,0x83f3U,0xf253U,0xcb67U,0x3f8fU,0x8078U,0xdb70U,0x156cU,0x9604U,
		0x9f9aU,0x08feU,0x0077U,0x2314U,0xaf8cU,0xcd4bU,0x3b6dU,0xad72U,0x7ed2U,0xa241U,
		0x3728U,0x10d1U,0x48f7U,0x22bcU,0x8fa7U,0x066bU,0x191dU,0xee96U,0x43d9U,0x213eU,
		0x09fbU,0x8d68U,0xe623U,0xbd5dU,0x927dU,0x4136U,0x9773U,0xe125U,0x7385U,0x4648U,
		0xe3cbU,0xbba6U,0xfe01U,0x7f81U,0x27c0U,0x5fe1U,0x8a76U,0x9ea8U,0x34abU,0xb85dU,
		0xfe25U,0x7c88U,0x36f7U,0xa7a5U,0x7556U,0x2d82U,0x4550U,0x6f24U,0x2e76U,0x9f1fU,
		0x7de1U,0xaabdU,0x7295U,0x7fc3U,0xc8ffU,0x3d08U,0x53d2U,0xbe43U,0x555eU,0x4f8fU,
		0xbd24U,0x6f8dU,0x9b0dU,0x293aU,0x2e5fU,0xe576U,0x624eU,0xbddeU,0x24d0U,0xaa81U,
		0x41afU,0x4758U,0x501cU,0xff87U,0x12ddU,0xe52bU,0xf274U,0xbc9bU,0xca24U,0x42eaU,
		0x9684U,0x7a11U,0x2943U,0x9200U,0x9752U,0xfb2fU,0x525bU,0x30caU,0x7ae3U,0x3a0cU,
		0x5957U,0x7724U,0x1e96U,0x6c44U,0xdb1bU,0xb894U,0x3c94U,0xc23eU,0xb907U,0x4a89U,
		0x4019U,0xf165U,0xe098U,0xd17cU,0x4077U,0x0c14U,0xda82U,0xad7bU,0x45d5U,0xbfc9U,
		0xac45U,0x94d0U,0x6924U,0xadcbU,0x9df4U,0x775cU,0x2046U,0xb3cbU,0xb15bU,0x7e93U,
		0x8c7aU,0xd2c2U,0xdfddU,0x5961U,0x89f9U,0x6807U,0x21ceU,0x6faeU,0x7816U,0xa153U,
		0x1606U,0x7f25U,0xea48U,0x9631U,0x0adaU,0x8cb7U,0x0bcbU,0xfd34U,0x1467U,0x4947U,
		0xc9d9U,0x0f96U,0xf10fU,0x9e72U,0xe9feU,0xfbe0U,0x87efU,0xb648U,0xd380U,0xae63U,
		0x780cU,0xb05eU,0xdfefU,0x27a9U,0x2155U,0x7b32U,0x66c2U,0x1153U,0x39fbU,0xe649U,
		0x6214U,0x7ddaU,0x11e7U,0x54daU,0xc086U,0xf864U,0xe2c3U,0x678fU,0x5f64U,0xcaeeU,
		0x2c5bU,0x5cbfU,0x72ccU,0x5c60U,0xb593U,0x3c13U,0x2551U,0x7bbeU,0x621bU,0xc49dU,
		0x3988U,0x38f2U,0x3d49U,0xff7eU,0x8dc6U,0x6bc1U,0x54fbU,0xf657U,0x5630U,0x0509U,
		0x616cU,0xf6cbU,0x9844U,0x21f1U,0x082dU,0x4a07U,0x3121U,0xa7ffU,0x3164U,0xc9c5U,
		0xeaefU,0xdf1bU,0xc877U,0x7a4aU,0xb881U,0x8feaU,0xa927U,0xefdfU,0x20bdU,0x5275U,
		0xf443U,0x361cU,0x48b6U,0xef87U,0xae01U,0x9d3eU,0xfa1bU,0xef19U,0x044fU,0x26f1U,
		0xa40dU,0xb013U,0x58caU,0x2e2aU,0xb2efU,0x0c93U,0xf15bU,0xa3f4U,0x286cU,0xf184U,
		0xe301U,0x2129U,0x9f63U,0x1c92U,0xbc34U,0xc6edU,0x65d2U,0x4f87U,0x791bU,0x6c09U,
		0x81e5U,0x82a2U,0x04b5U,0x44e5U,0xffd3U,0x9098U,0xf0c1U,0x71dcU,0xa31dU,0xd01cU,
		0x4ff7U,0xdf07U,0x5fc8U,0x1dd6U,0x5ed8U,0x328eU,0xa603U,0x9b98U,0xea30U,0x96d6U,
		0xcc40U,0x5e1eU,0x76d3U,0xd54bU,0x079aU,0x7d5cU,0x57baU,0xf484U,0x9c88U,0x5430U,
		0x0092U,0x4329U,0x6341U,0xf6f3U,0x3873U,0xa1b4U,0x0c3aU,0xe37aU,0x5917U,0x6ea7U,
		0x0a8eU,0x28f6U,0x402cU,0x35e7U,0xaf1fU,0xf03aU,0xac3bU,0x90bfU,0x0e58U,0xbf6dU,
		0xda15U,0xc617U,0x0161U,0xeccdU,0x8625U,0x8136U,0x6be9U,0xe4b5U,0x3812U,0x0fb9U,
		0x4419U,0x62fbU,0x992bU,0x1e8dU,0x9d5fU,0xd331U,0x54e1U,0x9523U,0x2bc1U,0x8558U,
		0x82c9U,0x7cddU,0xac96U,0xf432U,0x965cU,0x65daU,0xacfdU,0x7e00U,0xe7f0U,0xf139U,
		0xb9f9U,0xd5daU,0x9ce6U,0x4175U,0x6a96U,0xcf63U,0x0da4U,0x13b0U,0xd471U,0xdce8U,
		0xe63fU,0x2fc5U,0xb03cU,0x5054U,0xbea5U,0x9913U,0x9985U,0x497bU,0xe9d3U,0x1ae0U,
		0x0b8cU,0x8e59U,0x1703U,0x9a27U,0xa8feU,0x1b88U,0x9fcdU,0x3084U,0x6431U,0xc5c9U,
		0xa4d4U,0xe23eU,0x7187U,0xbec3U,0xd4ddU,0x4a9dU,0xa292U,0xd56fU,0xc5dbU,0x9a50U,
		0x7c23U,0xc0c8U,0x8c10U,0xe607U,0x68dcU,0x16bcU,0x5069U,0x9a8aU,0xdf75U,0x89e9U,
		0xafdaU,0xca3cU,0x7d97U,0x9864U,0xb621U,0xdc49U,0x3fe1U,0x9e83U,0x37c6U,0x949cU,
		0x77c3U,0xce9eU,0x5809U,0x9714U,0x9c78U,0xffbeU,0x0284U,0x7b65U,0x2993U,0xd804U,
		0xd376U,0x4bd5U,0x3507U,0xf1dbU,0x5993U,0xe561U,0x7e4cU,0x6681U,0x80c2U,0xd6fbU,
		0x7080U,0x03dcU,0x662aU,0x6421U,0x3646U,0x3e7dU,0x83beU,0x43c8U,0x5d95U,0xb5dbU,
		0xe5b9U,0x250aU,0x1fafU,0x3af0U,0x478dU,0xee5dU,0xcb37U,0xd44fU,0xd12cU,0x8411U,
		0x546fU,0x9123U,0xaad7U,0x30c1U,0x1244U,0x35d1U,0xb001U,0x7cf2U,0xd13cU,0xb818U,
		0x234dU,0xd0b1U,0x0993U,0x9ad0U,0x11eeU,0x5442U,0x706dU,0x7476U,0xd993U,0x0aa9U,
		0x652cU,0x1517U,0x9ebfU,0x1cabU,0xdb05U,0x2ae6U,0x5d72U,0x8700U,0x2066U,0x0f2fU,
		0xc64bU,0xc67dU,0xb7b4U,0x1fc2U,0xb79eU,0xf354U,0x0360U,0xe4f9U,0xdd12U,0x2be5U,
		0xf12aU,0x3850U,0x2b6cU,0x235cU,0x17d2U,0x073eU,0xfbd6U,0xc280U,0x335aU,0xd926U,
		0x5a66U,0xb34dU,0x7d3eU,0x4636U,0x56cbU,0x4c02U,0x4f08U,0x9a41U,0x32e2U,0xdb88U,
		0x943cU,0x3ec2U,0x5242U,0x03c6U,0xeb1aU,0xd03dU,0xa073U,0xe3ebU,0xd5c1U,0xad25U,
		0x58a4U,0xc19cU,0x36d7U,0xb67dU,0x55e5U,0xf8f4U,0xf52dU,0xc89cU,0x6bfaU,0xaeebU,
		0xe762U,0xf1f9U,0x223fU,0xe8e4U,0x9edeU,0xdd8dU,0x3f7cU,0xb245U,0xb9f1U,0x2e57U,
		0x3723U,0x4299U,0x6d36U,0x9bf7U,0x37e1U,0xff6fU,0xc872U,0x2c56U,0x57b5U,0xbc2fU,
		0x12d8U,0x36f2U,0x1462U,0xda7dU,0xdd0dU,0xaffcU,0x145dU,0x589fU,0xe63aU,0x380dU,
		0x598eU,0x8c0bU,0x9dd0U,0x60d2U,0x94e4U,0x870fU,0x1399U,0xa5beU,0xe729U,0x10ffU,
		0x0d2aU,0x5153U,0x38b0U,0x2d64U,0x6714U,0x4749U,0xdc23U,0x84a8U,0x61b5U,0xf69aU,
		0xad53U,0xe8cdU,0x2747U,0xf911U,0xa5c3U,0xfc4cU,0xf7bbU,0xfa63U,0x0941U,0x8c1fU,
		0xfbaeU,0x4f4cU,0x115aU,0x6defU,0xd948U,0xf847U,0x7c08U,0x8192U,0xb2a5U,0x5ba5U,
		0xaab3U,0x6c11U,0x5b91U,0x1822U,0x9db2U,0xcc10U,0x42a1U,0xa57fU,0xbfb2U,0xdaa7U,
		0x61ebU,0x8065U,0xc051U,0xb779U,0x46c6U,0x3d52U,0x139eU,0x14eeU,0xad0dU,0x3b4dU,
		0x9039U,0xccddU,0xbdfeU,0x70bcU,0x9b16U,0x6957U,0x3e26U,0x4526U,0x656eU,0xaae0U,
		0x0895U,0xb37bU,0xa9edU,0xdd5bU,0x9934U,0x8bb8U,0xad4aU,0x34a2U,0x804fU,0xc2bfU,
		0x9fe5U,0x9c4bU,0x6737U,0x5ad3U,0xaa02U,0x1d5dU,0x784dU,0xbac9U,0x1104U,0x3b76U,
		0xf41cU,0xff09U,0x3274U,0x1b2aU,0x8477U,0x7412U,0x1bcfU,0x7533U,0x78d6U,0xcbd5U,
		0xfe1bU,0x94abU,0xe5e0U,0xb9dbU,0xb12bU,0xc1afU,0x3e39U,0xf90aU,0x99e6U,0x7b93U,
		0xcae7U,0x7c81U,0x35caU,0xe4aeU,0x935fU,0x6825U,0xcfc7U,0xc1d1U,0x4438U,0x0a1eU,
		0xe72eU,0x4bc2U,0x6cb1U,0xfd72U,0x3a5fU,0xb940U,0x5082U,0x3c64U,0x3095U,0xe082U,
		0x8f83U,0x5ab7U,0xc487U,0x1261U,0x0dceU,0xfff7U,0x1734U,0x137fU,0xf5acU,0x1cbdU,
		0xa001U,0x773cU,0xd970U,0x2429U,0xd052U,0x4e5eU,0xa72eU,0x1214U,0x09bfU,0x7ffbU,
		0xf256U,0x3e40U,0x60d7U,0x5cedU,0x5b7fU,0xc13fU,0x3bc9U,0xc504U,0x83f6U,0x557cU,
		0x66c9U,0xfbc3U,0xaa6cU,0xbeecU,0xbd98U,0x13d4U,0x3f92U,0x50a9U,0x2cb0U,0xe960U,
		0x630bU,0x5fb6U,0x1af3U,0x947aU,0xcb6aU,0xcdcbU,0x6f83U,0xe80dU,0xfb92U,0x3f14U,
		0xb8abU,0xe971U,0x8125U,0x5b05U,0x9dd5U,0xeaadU,0x5d50U,0x39a7U,0x632cU,0x91d4U,
		0x2826U,0xde49U,0x445aU,0x8aa1U,0xe3a8U,0xd1b9U,0xfa5aU,0x2d79U,0xad8dU,0x8ca6U,
		0x3d76U,0xe2a3U,0x06ceU,0xa0d3U,0x142dU,0x3516U,0xce4aU,0x17b9U,0x7eefU,0xb0d1U,
		0x291dU,0xf8a9U,0xf9c4U,0x8f5eU,0x0901U,0x7aacU,0x8714U,0x6776U,0x5dafU,0xfc0aU,
		0x29fcU,0x9f45U,0x9f9dU,0x6eddU,0x8a54U,0xc34cU,0xd0e6U,0xc800U,0xa2d6U,0xd6ceU,
		0x2317U,0xf788U,0x5c88U,0xcd93U,0x5096U,0xa6b1U,0xc8f5U,0x4677U,0x007aU,0xaaf9U,
		0xe5a4U,0x07eeU,0x9792U,0x71fcU,0x304cU,0xb5eaU,0xdb73U,0xa66dU,0x94e9U,0x8bcfU,
		0xecc5U,0x9ea0U,0x0b33U,0xf692U,0x807bU,0x0e10U,0x1f7cU,0xdd3fU,0x2c1fU,0xb393U,
		0xc593U,0x45efU,0x9607U,0x8a0aU,0x9e4eU,0x7793U,0xc176U,0xb98aU,0x271aU,0x603bU,
		0x156fU,0x8280U,0x4ba1U,0xb6ffU,0xc62aU,0x0b3bU,0x2c94U,0x2378U,0xa244U,0x7b9aU,
		0x474eU,0xe698U,0xede1U,0xfab2U,0x3766U,0xe488U,0x7ed5U,0x51e6U,0xe650U,0xedf4U,
		0xa993U,0xa08bU,0x8442U,0x48dfU,0x10d4U,0xf5c1U,0x2a82U,0x7923U,0x053fU,0x391eU,
		0x6972U,0x639bU,0x372bU,0xa033U,0x6b56U,0x2d04U,0x2e16U,0xd1d4U,0x3a32U,0x2dddU,
		0xcd4eU,0x7995U,0x6719U,0xb560U,0xf077U,0x5457U,0xb29cU,0x53c9U,0xaf8fU,0xce81U,
		0x16e8U,0xacddU,0xe601U,0x0fb3U,0xc196U,0x7cecU,0xad75U,0xa79fU,0xc6e7U,0x5cb9U,
		0x12f4U,0xe692U,0x0ea8U,0x02e8U,0x3b70U,0xfc46U,0x6770U,0x5ce7U,0xb9b6U,0xedb3U,
		0xf26aU,0x0ba5U,0xee99U,0x0108U,0x84adU,0x12faU,0xa4e4U,0xbe31U,0x8efcU,0xaed1U,
		0x1920U,0xde2dU,0x47a9U,0xaf31U,0x7251U,0x6ad3U,0xe841U,0x47feU,0x2141U,0x72eaU,
		0x041bU,0xe916U,0xf776U,0x95d7U,0x892aU,0xf3ddU,0x43dcU,0xd77cU,0xb6caU,0x7ad0U,
		0x5d30U,0x5b45U,0x1ac3U,0x3aa5U,0x22bfU,0x836eU,0xdc28U,0xf605U,0x56f1U,0x2808U,
		0x5742U,0x8634U,0x48faU,0xa3b6U,0xef4cU,0x9144U,0x2623U,0x96f5U,0x4e99U,0xbcf8U,
		0x066eU,0x1ba4U,0xcc26U,0xd200U,0xa4fdU,0x8344U,0xe40fU,0x9aaaU,0x8faaU,0x46ddU,
		0x26c7U,0x1198U,0xd716U,0x63f9U,0xb04eU,0x8e68U,0x4139U,0x99edU,0x5158U,0x02eeU,
		0xcab5U,0xb4edU,0xb267U,0xb40bU,0x9280U,0xe8fbU,0x3a02U,0x2c33U,0x6118U,0x0fcdU,
		0x15ceU,0x2119U,0xe128U,0xbfdaU,0x022cU,0xc423U,0xebfdU,0x2d9eU,0x3dd8U,0xfed0U,
		0x9776U,0x1791U,0x8ad8U,0xd3f8U,0x7ccdU,0xab93U,0xda9eU,0xbdd5U,0x8d6bU,0x566aU,
		0x0d2fU,0xbb71U,0x44a1U,0x3cb7U,0x4f66U,0xa42aU,0x09feU,0xdbfaU,0x9b3fU,0x0861U,
		0x62e0U,0xf344U,0x6a4bU,0x4619U,0xbd60U,0x18f0U,0x7d12U,0xf027U,0xebdfU,0xc70bU,
		0x0026U,0xa5d9U,0xe626U,0x80e9U,0x4e1eU,0xd243U,0xfeb8U,0xdc9fU,0x3412U,0x60e8U,
		0x7f84U,0x8d94U,0x2d69U,0x0eaeU,0xcf4fU,0xfb48U,0xb8fdU,0xdaf4U,0xfe04U,0x4fa2U,
		0x7c54U,0xc7aeU,0x87caU,0xcf38U,0x4b18U,0x465bU,0x5fe4U,0xca67U,0x78b2U,0xe8bdU,
		0xaa17U,0x2194U,0xaf0eU,0x6aa5U,0x27c3U,0x3604U,0xf8c1U,0x1094U,0x33ffU,0xfc80U,
		0xb72aU,0xdd6dU,0x464bU,0x34e3U,0x38b5U,0x16d6U,0x5365U,0xf765U,0xd83cU,0x0e41U,
		0x7388U,0xeb9bU,0x93ccU,0xf673U,0x1ed4U,0x1d20U,0xff22U,0x533eU,0xbba9U,0x3a95U,
		0x597cU,0x0650U,0x9923U,0x4548U,0x375eU,0x0939U,0xe3ceU,0x658fU,0x9953U,0x08b5U,
		0x8382U,0xf790U,0x3d5bU,0x1d68U,0x7c8bU,0x3d59U,0x7c8dU,0x8d9aU,0x6830U,0xde52U,
		0x8d9cU,0x4450U,0xfe28U,0x8483U,0x7c8fU,0x182cU,0x4e6aU,0xfd11U,0xf792U,0x0d88U,
		0xa7a8U,0xfc57U,0x8384U,0x28feU,0xde82U,0xbef6U,0x1d6aU,0x30a1U,0x36faU,0x9491U,
		0x3d5dU,0x80f2U,0x8f90U,0xf7c7U,0x8485U,0xfabcU,0x9eabU,0x7bc7U,0xfe2aU,0xfe87U,
		0xde37U,0x793fU,0x182eU,0x00d9U,0x8a79U,0x918eU,0x7c91U,0x9f50U,0xd427U,0x245fU,
		0xde54U,0x0f38U,0xb860U,0x6403U,0x6832U,0x9045U,0xf43bU,0xfbceU,0x4452U,0x9b37U,
		0x34aeU,0x6876U,0x8d9eU,0x9c6dU,0x8305U,0xef91U,0xfc59U,0x94f5U,0x9f22U,0x1f23U,
		0xa7aaU,0x7f8aU,0x779fU,0xbd0dU,0x2900U,0xab26U,0x2e79U,0x792eU,0x8386U,0x5afbU,
		0x9700U,0xf57bU,0xfd13U,0x8b7aU,0xaac0U,0x090dU,0x4e6cU,0x6c4eU,0x0520U,0xd564U,
		0x0d8aU,0x4809U,0x7de4U,0x539aU,0xf794U,0xb509U,0xd8fbU,0xd403U,0x9493U,0xe90dU,
		0x2d85U,0xa857U,0x36fcU,0x9f05U,0x4de6U,0x4b47U,0x80f4U,0xe335U,0x7559U,0x1fccU,
		0x3d5fU,0xb8b7U,0x03e6U,0x9894U,0xbef8U,0xaf78U,0x6f27U,0x5f79U,0xde84U,0x6ab0U,
		0x08a2U,0xa73aU,0x30a3U,0x505eU,0x4553U,0x5539U,0x1d6cU,0xe549U,0x41c2U,0x147fU,
		0x7bc9U,0x06c4U,0xbe46U,0x46cdU,0x9eadU,0x0eb4U,0x2736U,0x3627U,0xfe89U,0x45bdU,
		0x53d5U,0x3fc5U,0xfe2cU,0xca7aU,0xc183U,0x054bU,0xf7c9U,0xc9f7U,0x4f92U,0xee52U,
		0x8f92U,0x0409U,0x4b81U,0x22aeU,0xfabeU,0x99c9U,0x5561U,0x7f5fU,0x8487U,0x4334U,
		0x9150U,0x71c1U,0x9190U,0xc4b3U,0x7fc6U,0xe536U,0x8a7bU,0xa56fU,0xef63U,0x0f50U,
		0x9f52U,0x26d3U,0x7298U,0x2ff5U,0x7c93U,0xd7d6U,0xae9fU,0x0621U,0x7941U,0x8cc2U,
		0x3d0bU,0x679fU,0xde39U,0x363bU,0xf170U,0x8ebaU,0x00dbU,0x034fU,0xc902U,0x43e8U,
		0x1830U,0xc96fU,0x9594U,0x3eb2U,0x6405U,0xba29U,0xe579U,0xccf9U,0xb862U,0x2d6fU,
		0xce57U,0x6124U,0x9047U,0x4885U,0x2e62U,0xa09fU,0x6834U,0xc5edU,0xd321U,0x03d1U,
		0x2461U,0x57ceU,0xbde1U,0xb1ffU,0xd429U,0xafbaU,0x2202U,0x3798U,0x0f3aU,0xa513U,
		0x6251U,0x6b8eU,0xde56U,0x6a57U,0x986fU,0x2f13U,0x6878U,0x9b88U,0x6f90U,0x5ff0U,
		0x34b0U,0x6cf2U,0x3df1U,0x1e4fU,0x9c6fU,0x43d3U,0xbd27U,0x7097U,0x8da0U,0x7028U,
		0xe9cdU,0x6624U,0xfbd0U,0x5988U,0x293dU,0x87e9U,0xf43dU,0x386dU,0x16e2U,0xef46U,
		0x9b39U,0x93c6U,0x9b10U,0x308fU,0x4454U,0x1f76U,0x229fU,0x231fU,0x1f25U,0x0812U,
		0xff8aU,0x14f5U,0x9f24U,0xdca5U,0xd007U,0x07d0U,0x7f8cU,0xf30cU,0x501fU,0x9518U,
		0xa7acU,0xe2cfU,0xdec5U,0x3ea3U,0xef93U,0x93efU,0xe52eU,0xf924U,0x8307U,0x0a96U,
		0xcd5bU,0x312dU,0x94f7U,0x09b5U,0x12e0U,0x7eafU,0xfc5bU,0x8036U,0xd881U,0x6c97U,
		0x7930U,0xbc40U,0xaa84U,0xa647U,0x2e7bU,0x4349U,0xdf13U,0xe4f1U,0x5afdU,0x5662U,
		0x24d3U,0xd417U,0x8388U,0x6bdeU,0x5115U,0x19b1U,0xbd0fU,0x9885U,0x475bU,0x07a6U,
		0x77a1U,0x8f25U,0x9170U,0xef2fU,0xab28U,0x2658U,0x41b2U,0x98c1U,0x2902U,0xd58aU,
		0x5d3dU,0xf280U,0x090fU,0x8668U,0x7a14U,0xbc71U,0xaac2U,0xfebeU,0x9333U,0x7df5U,
		0x6c50U,0x404aU,0x9687U,0x7215U,0x4e6eU,0xa2fdU,0xe2dfU,0x01b4U,0xf57dU,0x3393U,
		0x9203U,0x305eU,0x9702U,0x455cU,0xd5f4U,0x5ef5U,0x8b7cU,0x593dU,0x2946U,0x1c46U,
		0xfd15U,0xc135U,0xa5a2U,0x0fe5U,0x539cU,0xaa9bU,0xbc9eU,0x48c6U,0x7de6U,0x343aU,
		0x3a81U,0x7504U,0xb50bU,0x82beU,0xf277U,0xea0aU,0xf796U,0xcc5fU,0x7f31U,0x5bdeU,
		0xd566U,0xedaaU,0x42edU,0xafa0U,0x0522U,0x3e94U,0x5c07U,0x6a19U,0x480bU,0x9e00U,
		0xca27U,0x057fU,0x0d8cU,0xd7afU,0xe06eU,0x52e3U,0xa859U,0x6f6dU,0x3a0fU,0x13a5U,
		0x2d87U,0x60eeU,0x3ba5U,0x88aeU,0x9f07U,0x8d40U,0x7ae6U,0xfdd8U,0x36feU,0x4f2dU,
		0x51b0U,0x89d0U,0xd405U,0x3f88U,0x7727U,0xa364U,0xd8fdU,0xe67aU,0xc1a8U,0xcd8cU,
		0xe90fU,0xe8b6U,0x595aU,0xdf00U,0x9495U,0xb23eU,0x6020U,0x306dU,0x1fceU,0xe148U,
		0xfb32U,0xf92cU,0x755bU,0xb031U,0x0a0bU,0x3366U,0xb8b9U,0x5e75U,0x9755U,0xc1fcU,
		0x3d61U,0x02c6U,0x0d69U,0x4eecU,0x4b49U,0xd29aU,0x30cdU,0x4094U,0x4de8U,0x06abU,
		0x2562U,0xa22eU,0xe337U,0x521eU,0x525eU,0x9414U,0x80f6U,0xd99fU,0xca48U,0x1b7dU,
		0x5f7bU,0x6f41U,0xc241U,0x7f14U,0x6f29U,0x3418U,0x29cdU,0x3ef2U,0x6ab2U,0x0290U,
		0x3c97U,0x973cU,0xde86U,0xbbcfU,0xbf57U,0x0970U,0x9896U,0xe6adU,0x4a8cU,0xa80fU,
		0x03e8U,0xe803U,0x0e23U,0x14a3U,0xaf7aU,0x6afaU,0xb90aU,0xcb43U,0xbefaU,0xc999U,
		0xfc6bU,0x162fU,0x553bU,0xd4d2U,0x6c47U,0xafb3U,0x4555U,0xe7fcU,0xbb8cU,0x310fU,
		0xe54bU,0x0ac1U,0x1e99U,0x12aaU,0x1d6eU,0x0c4fU,0x5372U,0x643dU,0xa73cU,0x8251U,
		0xb897U,0x7579U,0x08a4U,0x2b56U,0x934dU,0x7a23U,0x5060U,0x8ef3U,0xdb1eU,0x1060U,
		0x30a5U,0x0271U,0x2229U,0xcd9bU,0x46cfU,0xce36U,0x0c17U,0x806cU,0xbe48U,0x4fa8U,
		0x9851U,0x282fU,0x0eb6U,0xf2a4U,0x407aU,0xf19aU,0x9eafU,0xdc5bU,0x5594U,0x132cU,
		0x1481U,0x6ac5U,0xad7eU,0x2b31U,0x41c4U,0x35efU,0xa405U,0x4f00U,0x06c6U,0xebd7U,
		0xda85U,0x415fU,0x7bcbU,0xe62fU,0x7987U,0xe50cU,0x3fc7U,0x5a91U,0xf168U,0x255aU,
		0x53d7U,0xcad8U,0x84dbU,0x5823U,0xca7cU,0x9e44U,0x401cU,0xa97eU,0xfe2eU,0xa2b4U,
		0x2026U,0xfaa4U,0x3629U,0x206fU,0xd17fU,0xbc00U,0x2738U,0x809fU,0xac9fU,0x7beaU,
		0x45bfU,0x20eeU,0xe09bU,0xda72U,0xfe8bU,0x05e0U,0x5c17U,0x9271U,0xee54U,0x9aa1U,
		0xadceU,0xfadcU,0x4f94U,0xfe0aU,0x0ce0U,0xa4c2U,0x040bU,0xa13cU,0x6927U,0x7691U,
		0x8f94U,0xf4beU,0x5fd5U,0x35a7U,0x054dU,0xf32eU,0x775fU,0x9298U,0xc185U,0xfc72U,
		0x374cU,0x3bb4U,0xc9f9U,0x38a5U,0x9df7U,0x77e8U,0xf7cbU,0x3a4cU,0xa699U,0x1a28U,
		0x7f61U,0x3b61U,0xbfccU,0x3d38U,0x5563U,0xf80cU,0xa9ddU,0xccc7U,0x4336U,0xac1fU,
		0x45d8U,0x7345U,0x8489U,0x8276U,0xb1b7U,0x27f1U,0x22b0U,0x127eU,0x94d3U,0xc6fdU,
		0x4b83U,0x5aa3U,0x32b3U,0x60c7U,0x99cbU,0x1004U,0xac48U,0x37f5U,0xfac0U,0x33dfU,
		0x888fU,0x79d8U,0xe538U,0xe773U,0xd2c5U,0x61f2U,0x7fc8U,0xc7b4U,0x7048U,0xe2e9U,
		0xa571U,0x43b3U,0x8c7dU,0x9675U,0x8a7dU,0xcef0U,0xe03aU,0x50cbU,0x71c3U,0x8025U,
		0x5964U,0xe430U,0x9152U,0xeb0fU,0x3570U,0xd628U,0xc4b5U,0x4518U,0xdfe0U,0x0b5aU,
		0x9192U,0x0e2dU,0xc3f3U,0x3311U,0x2ff7U,0x9788U,0xb3ceU,0x830fU,0x729aU,0xb8dfU,
		0xbabbU,0xb21dU,0xd7d8U,0x8b16U,0x2049U,0x2775U,0x7c95U,0x9835U,0x2b80U,0x522dU,
		0x0f52U,0xce61U,0x7e96U,0xf8e9U,0xef65U,0xea9aU,0x53dfU,0xfd64U,0x26d5U,0x7360U,
		0xb15eU,0x86dcU,0x9f54U,0xc404U,0xa619U,0x8812U,0x67a1U,0xecf3U,0xa156U,0x5007U,
		0x3d0dU,0x7c5aU,0xb085U,0xdf66U,0x363dU,0x261aU,0x7819U,0xa8a5U,0xde3bU,0xf55cU,
		0x98f1U,0x2d28U,0x0623U,0x3f71U,0x7f28U,0x2d15U,0xaea1U,0x1636U,0xac87U,0x4a6eU,
		0x8cc4U,0xbd39U,0x1609U,0x8681U,0x7943U,0x1206U,0x2053U,0x92fcU,0x43eaU,0x1440U,
		0x680aU,0xeec2U,0xc904U,0xf688U,0x66fbU,0xb8c8U,0xc971U,0x6705U,0x89fcU,0x7965U,
		0x1832U,0x69d2U,0x7514U,0x9becU,0x8ebcU,0x8874U,0x6fb1U,0x2535U,0xf172U,0x6b12U,
		0x112eU,0x25c6U,0x0351U,0x5603U,0x21d1U,0xa6a3U,0x00ddU,0x541bU,0x75d9U,0x5c90U,
		0xccfbU,0x7468U,0xfd37U,0xb780U,0xe57bU,0xfb4eU,0x9d36U,0xb345U,0x2d71U,0x4611U,
		0x0bceU,0x71a7U,0xb864U,0x7d33U,0x761eU,0x23e2U,0x3eb4U,0x5bbeU,0x494aU,0x7d6cU,
		0x9596U,0x4034U,0x79f5U,0x0b09U,0xba2bU,0x6670U,0x146aU,0xf05eU,0x6407U,0x1da8U,
		0x8b8cU,0xb66fU,0xa0a1U,0x3fffU,0x9634U,0x18a4U,0x2e64U,0x890bU,0xfbf9U,0x3f2bU,
		0xc5efU,0xe283U,0xea4bU,0x03bbU,0x6836U,0x60a1U,0x36e4U,0x8af5U,0x6126U,0x74baU,
		0x8cbaU,0xd292U,0xce59U,0x9f8bU,0x08d9U,0x3ca6U,0x4887U,0xc8ebU,0x0addU,0x6fc8U,
		0x9049U,0x1cd5U,0x922fU,0xea13U,0xb201U,0x5c30U,0xfbe3U,0x47c8U,0xbde3U,0xcf55U,
		0x4c4aU,0x01f3U,0xafbcU,0xb793U,0xea01U,0xec7dU,0xd42bU,0x96c2U,0x8cd8U,0xa88aU,
		0x03d3U,0xa3adU,0xb64bU,0xec40U,0xd323U,0x5542U,0x2defU,0xbf98U,0x57d0U,0xa176U,
		0x87f2U,0x52d1U,0x2463U,0x5e4fU,0x58f0U,0xcb52U,0x6b90U,0x33c9U,0x0f99U,0x3ae2U,
		0x6253U,0xfc33U,0x09a0U,0x3845U,0x6a59U,0x1e0bU,0xc9dcU,0xec86U,0xde58U,0xae31U,
		0x2d1eU,0xab44U,0x379aU,0x64e4U,0x9e75U,0x187cU,0x2204U,0x5401U,0xf454U,0x5755U,
		0xa515U,0x77f1U,0xf112U,0xfee5U,0x0f3cU,0xc7f6U,0xa29fU,0x28a5U,0x5ff2U,0x5f09U,
		0x27acU,0xc058U,0x6f92U,0xdafaU,0x5ca7U,0x1d7dU,0x6cf4U,0xe36cU,0xdff2U,0xa25fU,
		0x34b2U,0x1ec1U,0x405aU,0xf004U,0x2f15U,0x67eeU,0x7b35U,0x4601U,0x9871U,0x90d5U,
		0x4af2U,0x10edU,0x9b8aU,0x39ddU,0x2158U,0x7ca9U,0x687aU,0xf38dU,0x347cU,0xd3adU,
		0x7099U,0x83a7U,0xae66U,0x0a9eU,0xbd29U,0xc0b4U,0x55c5U,0x8fdcU,0x702aU,0xaf4eU,
		0xd383U,0xfccbU,0x8da2U,0x8e29U,0xaba8U,0x9c20U,0x1e51U,0x9f93U,0xb061U,0xe661U,
		0x3df3U,0xac0fU,0xcae0U,0x0073U,0x43d5U,0x27bcU,0x780fU,0xa5fcU,0x9c71U,0x6d2fU,
		0x9df0U,0x1602U,0x87ebU,0x11e3U,0x7dddU,0x624aU,0x293fU,0xb903U,0x04b1U,0xea2cU,
		0x386fU,0xda11U,0x6217U,0x0829U,0xf43fU,0xf157U,0xafd6U,0x0280U,0x6626U,0xd128U,
		0x54ddU,0x9ce2U,0xe9cfU,0xd4d9U,0x6bf6U,0x37ddU,0x598aU,0xdc1fU,0x11eaU,0xc647U,
		0xfbd2U,0x523eU,0x1bcbU,0x35c6U,0x3091U,0xd04eU,0x1156U,0xbfaeU,0x9b12U,0x9fe1U,
		0x8a50U,0xe5a0U,0x1f78U,0x4b9dU,0x66c5U,0x6f7fU,0x4456U,0x7eebU,0x47a5U,0xb6c6U,
		0xef48U,0x26c3U,0xe64cU,0x6b52U,0x16e4U,0x676cU,0x7c50U,0xf8bdU,0x93c8U,0x994fU,
		0x39feU,0x8ad4U,0x9b3bU,0x4e1aU,0x2b27U,0xc808U,0x14f7U,0x4adfU,0xcaf1U,0x70c3U,
		0xff8cU,0xca6dU,0xcff9U,0x8aaaU,0xdca7U,0xb82cU,0x5f67U,0xd88eU,0x9f26U,0x5b9bU,
		0x9cadU,0x208dU,0x2321U,0x8967U,0x5cc2U,0xd066U,0x22a1U,0xe382U,0xa9a5U,0xaa76U,
		0x0814U,0x19beU,0x2c5eU,0xf8d7U,0x1f27U,0xd24cU,0x75c4U,0x1f32U,0x951aU,0xedebU,
		0xf867U,0xc4a6U,0x5021U,0xd8dfU,0x192aU,0x85dcU,0xe2d1U,0x4ab4U,0xc089U,0x5f43U,
		0xa7aeU,0x95b1U,0x96a5U,0xb1e0U,0x07d2U,0xc686U,0x6792U,0xd720U,0xd009U,0x93fcU,
		0x65e3U,0x0a46U,0xf30eU,0xca19U,0xe2c6U,0x7d2aU,0x7f8eU,0xb522U,0xa641U,0x48c0U,
		0xf926U,0xafadU,0x7bc1U,0xa851U,0xe530U,0x5feaU,0x189eU,0x3adcU,0x0a98U,0xbfa8U,
		0x2554U,0x3d32U,0x8309U,0xeebcU,0x4402U,0x1e72U,0x3ea5U,0x4da6U,0xc4a0U,0xbcabU,
		0xdec7U,0x1ea0U,0x1582U,0xca97U,0x93f1U,0x6e96U,0x621eU,0x7719U,0xef95U,0x3bedU,
		0xd90dU,0x93b9U,0x7eb1U,0x436dU,0x5c63U,0xbcb1U,0x12e2U,0xd350U,0x47edU,0xfe64U,
		0x8038U,0x52fdU,0x72cfU,0xb7beU,0xfc5dU,0x4569U,0x708aU,0x1340U,0x312fU,0x70eaU,
		0x3c16U,0xd4b2U,0xcd5dU,0x0d18U,0xe2bbU,0xaacfU,0x09b7U,0xb046U,0xb596U,0xdb57U,
		0x94f9U,0x3c89U,0x4dd8U,0xeca5U,0xa649U,0x2fa8U,0x6bc4U,0xbe05U,0xaa86U,0xe8c3U,
		0xce2cU,0x2dcaU,0x434bU,0x36a7U,0x8dc9U,0xcedaU,0x2e7dU,0x9c60U,0xee3eU,0x9f14U,
		0x6c99U,0x0e80U,0xf65aU,0xb2b8U,0xd883U,0x5e04U,0x8e96U,0xcc9bU,0xbc42U,0x2b10U,
		0x54feU,0xd6a6U,0x7932U,0x1647U,0x664aU,0xd537U,0xd419U,0x2bf6U,0x38f5U,0xdecdU,
		0x24d5U,0x1017U,0x7af9U,0x3187U,0x6be0U,0xa23bU,0x398bU,0x0d7bU,0x838aU,0x3ac3U,
		0xdce2U,0xf1d5U,0xe4f3U,0x36ecU,0xff81U,0xfd2eU,0xdf15U,0x7d56U,0x798fU,0x8368U,
		0x5664U,0x34ddU,0x3d4cU,0x0a18U,0x5affU,0xa667U,0x97faU,0x394cU,0x07a8U,0xf7b1U,
		0x21f4U,0x6481U,0x475dU,0x78b8U,0xb84bU,0xde93U,0x8f27U,0x9ef8U,0x9847U,0x4d88U,
		0x77a3U,0x0b71U,0x60feU,0xcbaaU,0x19b3U,0x8505U,0x4a0aU,0x292fU,0x5117U,0x12b1U,
		0x0955U,0x8f4cU,0x9887U,0x5dc6U,0x0830U,0x6d8fU,0xbd11U,0xc9a6U,0x86bdU,0x9cc5U,
		0x98c3U,0xbd7cU,0x050cU,0x1ea6U,0x41b4U,0xbcc7U,0xd85bU,0xab19U,0xd58cU,0xec2dU,
		0x5633U,0xb0b7U,0x2904U,0xa9c6U,0x9ca2U,0xa45cU,0xef31U,0x106dU,0xf6ceU,0x5702U,
		0x9172U,0x728aU,0x538dU,0x5208U,0x265aU,0x4c5aU,0x616fU,0xac66U,0xab2aU,0x2a4bU,
		0x30ddU,0xd0eeU,0xbc73U,0x6c76U,0xdf1eU,0x695eU,0x7a16U,0xcf3eU,0x56acU,0x1148U,
		0xfec0U,0x4d31U,0xeaf2U,0x712fU,0xaac4U,0x7f97U,0x7be0U,0x38ddU,0xf282U,0x1985U,
		0x7a4dU,0x1b3fU,0x5d3fU,0x0c42U,0x04a3U,0x4e79U,0x866aU,0x53b9U,0xc87aU,0xa6e4U,
		0x0911U,0x8baeU,0x30f8U,0x3cebU,0x7217U,0xa864U,0xa802U,0x4408U,0x9689U,0x1693U,
		0x602bU,0x9451U,0xa2ffU,0x7801U,0x3124U,0x0b00U,0x4e70U,0x013fU,0x5689U,0x5a3cU,
		0x7df7U,0x3915U,0xc9c8U,0x6f34U,0x9335U,0x55a6U,0x481fU,0x1b1aU,0x404cU,0xf549U,
		0x3167U,0xaf61U,0x6c52U,0x40deU,0xd399U,0xf79fU,0x3060U,0x92dfU,0x5278U,0xe879U,
		0x9205U,0x87d0U,0x04e2U,0x614aU,0x455eU,0x636dU,0x20c0U,0xb63dU,0x9704U,0x29a5U,
		0x2dc2U,0x3671U,0x01b6U,0x069eU,0x361fU,0x88a6U,0xe2e1U,0x1d75U,0x1de7U,0xa0c0U,
		0x3395U,0x6564U,0xf446U,0x269bU,0xf57fU,0x588fU,0xfdcbU,0xe225U,0x1c48U,0x8656U,
		0x8fedU,0x1e78U,0x2948U,0x2c7cU,0x5c58U,0xb810U,0xc137U,0xb4e5U,0xb884U,0xd434U,
		0xfd17U,0x0bc0U,0xaeaaU,0xe8a9U,0x5ef7U,0x0128U,0xefe2U,0x4a2bU,0xd5f6U,0x65adU,
		0xb198U,0x9018U,0x593fU,0xf7d4U,0xa92aU,0x3c03U,0x8b7eU,0x41dcU,0xc53dU,0xd2b7U,
		0x48c8U,0xe259U,0xef1cU,0x9b1dU,0xbca0U,0x4661U,0x1a58U,0x3679U,0x343cU,0x4dc4U,
		0xfa1eU,0x9c36U,0x7de8U,0x2cc4U,0xbaadU,0xd818U,0x0fe7U,0x70a4U,0x26f4U,0x8355U,
		0xa5a4U,0x694dU,0xa2aaU,0x73feU,0xaa9dU,0x6a69U,0x0452U,0x15abU,0x539eU,0xc128U,
		0x926aU,0x880bU,0xea0cU,0x15fbU,0xef8aU,0x3eabU,0xf279U,0x1b76U,0x20daU,0xde63U,
		0xcc61U,0x5ebeU,0x48b9U,0x3945U,0xf798U,0x8836U,0xf93cU,0x0dbbU,0x7506U,0x8afdU,
		0x9d41U,0x62bfU,0x3a83U,0xc57eU,0xe514U,0x22e5U,0x82c0U,0x8439U,0xae04U,0xed9dU,
		0xb50dU,0xc5faU,0xe045U,0x883dU,0xafa2U,0x56dcU,0x0c96U,0xd171U,0x42efU,0x4b1eU,
		0x7164U,0x2793U,0x3e96U,0x0f5dU,0xb2f2U,0x32a1U,0x0524U,0x032eU,0xb968U,0xf23cU,
		0x5be0U,0x46bcU,0xa3f7U,0xd223U,0x7f33U,0x0c56U,0x5accU,0xf8a2U,0xedacU,0xdc98U,
		0xf15eU,0xf17dU,0xd568U,0xd036U,0x1893U,0xe6c2U,0x0581U,0x0873U,0xb016U,0x4dacU,
		0xca29U,0x45caU,0x98b4U,0x4cb4U,0xd7b1U,0x25daU,0xa410U,0xa71aU,0x0d8eU,0x325bU,
		0x5c43U,0xfacbU,0x6a1bU,0x8bf5U,0x2e2dU,0xa7c0U,0x5c09U,0xdb3bU,0x5496U,0x5655U,
		0x9e02U,0xc854U,0x58cdU,0xc2deU,0x480dU,0x3d1eU,0x8989U,0xd6d6U,0x13a7U,0x2dd4U,
		0x1c95U,0x3b54U,0x3a11U,0x360aU,0x4d92U,0x4463U,0x60f0U,0x52c1U,0x9f66U,0x4213U,
		0x2d89U,0x6d55U,0x6f51U,0xa27fU,0x52e5U,0x59bfU,0xc6f0U,0xc812U,0xe070U,0x6eafU,
		0x2334U,0x57f3U,0x6f6fU,0x6161U,0xbc37U,0x3ff6U,0xa85bU,0x4e27U,0x7687U,0xe986U,
		0xfddaU,0x2856U,0xf187U,0x6224U,0x7ae8U,0xf4b1U,0xce1eU,0xda62U,0x4f2fU,0x66e7U,
		0x286fU,0x40c2U,0x3700U,0x7408U,0xf70aU,0xda2eU,0x88b0U,0xc654U,0x212cU,0x8f0dU,
		0x3ba7U,0x7153U,0x9665U,0x38e7U,0x8d42U,0x0bdeU,0xe304U,0x73e9U,0x9f09U,0x586aU,
		0x8a3fU,0x20e1U,0xa366U,0x5f23U,0x82a5U,0xe0c1U,0x7729U,0x27c9U,0x1f59U,0x4acaU,
		0xe67cU,0xf2bcU,0x81e8U,0xe990U,0xd8ffU,0xdef0U,0xa971U,0x7037U,0x89d2U,0x9d07U,
		0x44e8U,0x28d8U,0x51b2U,0xbb93U,0x8dbeU,0x6524U,0x3f8aU,0xe120U,0x04b8U,0xdfabU,
		0xd407U,0x2546U,0xf652U,0x26ecU,0xdf02U,0x35e2U,0x4f8aU,0x771fU,0x595cU,0x7b2dU,
		0x1ca6U,0x4bfdU,0xb240U,0x10faU,0x65d5U,0x307fU,0x9497U,0x2505U,0xe95bU,0xe29eU,
		0xcd8eU,0x778eU,0x6c0cU,0xdd56U,0xc1aaU,0x137aU,0xc41eU,0xf022U,0xe8b8U,0x064bU,
		0x791eU,0x5cb4U,0xe911U,0xd1fbU,0x0307U,0xb47aU,0xf92eU,0xec4dU,0xd01fU,0xad14U,
		0xfb34U,0x109aU,0xb4b2U,0x01beU,0xb033U,0xf06eU,0xa320U,0x3f64U,0x755dU,0x869dU,
		0x3034U,0x722cU,0x306fU,0x0444U,0xdf0aU,0xfbf0U,0x6022U,0xc78aU,0xe7a2U,0xa83aU,
		0xe14aU,0x8ee0U,0x4ffaU,0x33aeU,0x1fd0U,0x14adU,0x98a6U,0x607fU,0xc1feU,0xc109U,
		0x909bU,0xef9bU,0x9757U,0x7b89U,0x389bU,0xe116U,0x02c8U,0xba48U,0xffd6U,0x2682U,
		0x3d63U,0x540eU,0xd279U,0x0c88U,0x3368U,0x612eU,0x71dfU,0x749cU,0x0a0dU,0x6d1bU,
		0x3fcfU,0x6260U,0x5e77U,0x84c1U,0xf0c4U,0xe1d6U,0xb8bbU,0x54aeU,0x6989U,0xde6aU,
		0x4096U,0x3786U,0x9b9bU,0x9393U,0x30cfU,0xf8c7U,0xa207U,0x43a6U,0x06adU,0x495dU,
		0xa606U,0x75e6U,0x4deaU,0xddbeU,0x663fU,0xab4eU,0x4eeeU,0xc946U,0x96d9U,0x266dU,
		0x0d6bU,0x3116U,0x0b4dU,0x4d4bU,0xd29cU,0x23efU,0xea33U,0xec5dU,0x4b4bU,0x3f3fU,
		0x277fU,0xa0aeU,0x9416U,0x78ccU,0x1dd9U,0x3bf3U,0x5260U,0xbb47U,0x8583U,0x782eU,
		0xd9a1U,0xb8d2U,0x5fcbU,0xa967U,0x80f8U,0xf63dU,0xabf8U,0x1ffaU,0xa230U,0xa108U,
		0x3291U,0xefd4U,0x2564U,0x74c7U,0x00a3U,0x9abdU,0x5220U,0xe943U,0x5edbU,0x1a32U,
		0xe339U,0xc925U,0xff44U,0xa2deU,0x7f16U,0x653dU,0x7d5fU,0xcce4U,0xc243U,0x219aU,
		0xc823U,0xa442U,0x341aU,0xe2f6U,0x079dU,0xd289U,0x6f2bU,0xb52bU,0xcf83U,0x46a0U,
		0x1b7fU,0xa796U,0xf487U,0x745bU,0xca4aU,0x591fU,0x3559U,0x0148U,0x6f43U,0xdde7U,
		0x57bdU,0x357dU,0x5f7dU,0xe4ccU,0xa1a5U,0x9b53U,0x973eU,0x4e30U,0x5e21U,0x1588U,
		0x3c99U,0x8b23U,0xc793U,0x4f7cU,0xbbd1U,0x85b8U,0xcc43U,0x2ee3U,0xde88U,0x64d3U,
		0x1a1eU,0x26e2U,0x3ef4U,0x1662U,0xd54eU,0xb534U,0x29cfU,0x6094U,0x1f4bU,0x87aeU,
		0x0292U,0x0197U,0x76d6U,0x6e4cU,0x6ab4U,0x681bU,0xec70U,0xcc68U,0xa811U,0xe0f3U,
		0xf6f6U,0x1616U,0x4a8eU,0xaa1dU,0x49ffU,0xc675U,0xe805U,0xd3f0U,0x6344U,0x924dU,
		0x03eaU,0xa312U,0x69a9U,0xa169U,0x0972U,0xa8b2U,0xa1b7U,0xf3eeU,0xbf59U,0xe552U,
		0x068dU,0xbdb8U,0xe6afU,0x6846U,0x3876U,0x3466U,0x9898U,0x6d6dU,0x8b4aU,0xe703U,
		0xcb45U,0x91e7U,0x5433U,0xca9dU,0xb90cU,0x45a9U,0x2ad4U,0x00eaU,0xc99bU,0xa0f7U,
		0x9c8bU,0x96cbU,0xbefcU,0x956fU,0x163fU,0xc120U,0x14a5U,0x53f4U,0x432cU,0xb236U,
		0x0e25U,0xf385U,0xb2d7U,0x95c5U,0x6afcU,0x3a55U,0x0095U,0x8538U,0xaf7cU,0x69dfU,
		0x0404U,0x3636U,0xafb5U,0x3868U,0x28f9U,0x9040U,0x6c49U,0x6aabU,0xe675U,0x06a6U,
		0xe7feU,0x2b51U,0x0a91U,0x8f20U,0x4557U,0x3e8fU,0xeb0aU,0xea95U,0x1631U,0x6b0dU,
		0x35eaU,0x809aU,0xfc6dU,0x5a9eU,0x90d0U,0xac0aU,0xd4d4U,0x6767U,0x402fU,0x9f86U,
		0x553dU,0x53fcU,0x5dffU,0x7d51U,0x12acU,0x7285U,0xe37dU,0x93f7U,0x1e9bU,0x0d13U,
		0x6948U,0xc579U,0x0c51U,0xdb36U,0x0c3dU,0x55a1U,0x1d70U,0x65a8U,0xc785U,0x6d16U,
		0x3111U,0x74c2U,0x6eaaU,0x714eU,0xbb8eU,0x1375U,0x5a99U,0x1370U,0x0ac3U,0xcda8U,
		0x591aU,0x608fU,0xe54dU,0xf380U,0x5382U,0x5ec5U,0x757bU,0x0387U,0xbf70U,0x62d0U,
		0xb899U,0xaf14U,0x8ac7U,0x2260U,0x2b58U,0xe021U,0x0e5bU,0xdcc1U,0x08a6U,0xf819U,
		0x88ecU,0x81beU,0x643fU,0x20a3U,0xc61aU,0xd8a6U,0x5374U,0x0ac8U,0x4595U,0x4b90U,
		0x8253U,0xc55eU,0xda18U,0x382dU,0xa73eU,0xbb62U,0x64abU,0x322dU,0x1062U,0xfae9U,
		0xf03dU,0x6e9cU,0xdb20U,0xd0d6U,0x1775U,0x6b2fU,0x0273U,0x1ce5U,0xaf22U,0xb6b1U,
		0x30a7U,0xc5aaU,0xc4d8U,0x31f5U,0x7a25U,0xd966U,0x90c2U,0x776cU,0x934fU,0xcdadU,
		0x326dU,0x29f5U,0x8ef5U,0xb8f6U,0xac3eU,0x7914U,0x5062U,0x55deU,0x81a8U,0xab01U,
		0x806eU,0xaa0dU,0xe4b8U,0xb382U,0x0c19U,0x34e9U,0x936aU,0x39b0U,0x4faaU,0x6202U,
		0x6becU,0x458bU,0xbe4aU,0xcc1aU,0x49e4U,0xcca7U,0xcd9dU,0x270dU,0x0fbcU,0xa5ceU,
		0x222bU,0xbf75U,0xa1eaU,0x13deU,0xce38U,0xb697U,0x3815U,0x57a7U,0x46d1U,0xc714U,
		0x2910U,0xf4a1U,0xf19cU,0xe492U,0xecd0U,0x5c69U,0x407cU,0xfb1eU,0xaf3bU,0xc260U,
		0xdc5dU,0x2a57U,0x0164U,0x5d24U,0x9eb1U,0x2a07U,0xc77aU,0x5c7bU,0x2831U,0x2aefU,
		0x8139U,0x8e72U,0x9853U,0x62d5U,0x8561U,0x66d4U,0xf2a6U,0x79c0U,0x8628U,0x6d9bU,
		0x0eb8U,0x32d4U,0x9a6fU,0x9d48U,0x2b33U,0x0c76U,0xd334U,0x4503U,0xad80U,0x4651U,
		0xd11eU,0x2099U,0x35f1U,0x891eU,0x9d62U,0x2a8dU,0x41c6U,0xa37cU,0x262eU,0xdbefU,
		0x132eU,0x74f0U,0x9526U,0xdea6U,0x5596U,0xb89eU,0xf617U,0xb613U,0x6ac7U,0xe84cU,
		0x54e4U,0xddf7U,0x1483U,0xc692U,0xc21dU,0x8ae3U,0x4161U,0x6cc6U,0x62feU,0xbcb7U,
		0xda87U,0xeb32U,0x2f60U,0x1914U,0xe631U,0x737fU,0x441cU,0xf35eU,0x7bcdU,0x7110U,
		0x642bU,0x1fa9U,0x4f02U,0x0d24U,0x1e90U,0x66bcU,0xa407U,0xaf19U,0x12eeU,0xa4f7U,
		0xebd9U,0x991dU,0x992eU,0xf5a6U,0x06c8U,0xc170U,0x916aU,0x5c01U,0x255cU,0x9347U,
		0xf435U,0x089cU,0xf16aU,0x16dcU,0x08d3U,0xf44eU,0xcadaU,0x7c4aU,0xac99U,0x32adU,
		0x53d9U,0x1128U,0x4819U,0xb192U,0xe50eU,0x5490U,0x65ddU,0xe2b5U,0x7989U,0x5387U,
		0x1f45U,0xb2d1U,0x5a93U,0x3267U,0x965fU,0xc418U,0x3fc9U,0x009dU,0xc395U,0x3685U,
		0xa980U,0xfa8bU,0x855bU,0x12e8U,0x401eU,0xf968U,0x8f76U,0xc0a4U,0xa2b6U,0x4fbdU,
		0x2bc4U,0x02b1U,0xfe30U,0x8c44U,0x1a07U,0x258aU,0x5825U,0x0ae5U,0x7ce0U,0x5169U,
		0x84ddU,0x5ecaU,0x4024U,0xe721U,0x9e46U,0x5974U,0x82ccU,0x3b31U,0xca7eU,0x976bU,
		0x6240U,0x62c6U,0xbc02U,0x2c88U,0xd5ddU,0x90edU,0xd181U,0x38bbU,0xd602U,0x89b5U,
		0x80a1U,0xc3d4U,0xb9fcU,0x1170U,0x273aU,0x8d8aU,0x7101U,0x524dU,0xfaa6U,0x4cc9U,
		0x4178U,0xe885U,0x2028U,0x7580U,0x470dU,0x63eeU,0x2071U,0x5a19U,0x9ce9U,0xb3aeU,
		0x362bU,0x7bfbU,0x7da2U,0x10c8U,0xda74U,0x16cdU,0x7e03U,0xd356U,0xe09dU,0x2fd9U,
		0xa218U,0x6db6U,0x05e2U,0x83c9U,0xad00U,0x342aU,0xfe8dU,0xa870U,0x2cf8U,0xe1b7U,
		0x7becU,0xf101U,0xf13cU,0x42d7U,0xaca1U,0x038cU,0xf96eU,0xd0faU,0x20f0U,0xb25cU,
		0xe7f3U,0x9fd8U,0x45c1U,0xd0cdU,0x3973U,0x0082U,0xfadeU,0x3dcbU,0xdcebU,0xdd62U,
		0xadd0U,0xfc86U,0xf20eU,0xdf96U,0xfe0cU,0x4a57U,0xd474U,0x3a24U,0x4f96U,0x65ecU,
		0x21beU,0xfdb0U,0x9273U,0xbb9fU,0x2fc8U,0xc0e5U,0x5c19U,0x0e60U,0x7735U,0x6034U,
		0x9aa3U,0x0932U,0xe642U,0x90b8U,0xee56U,0x2d5dU,0x176aU,0x7d05U,0x7693U,0x233dU,
		0xcf66U,0xd913U,0x6929U,0x535aU,0x303dU,0xda3aU,0xf4c0U,0xeb68U,0x6a99U,0x7650U,
		0x8f96U,0xe401U,0x7cc0U,0x1293U,0xa4c4U,0x301fU,0x13b3U,0xc82cU,0x0ce2U,0xdcc6U,
		0xc39bU,0x141eU,0xa13eU,0x22f9U,0x0da7U,0x87beU,0x040dU,0x59cbU,0x4df6U,0x3a8aU,
		0x929aU,0xb7d8U,0x497eU,0x9029U,0x7761U,0x3405U,0xa916U,0x15c0U,0xfc74U,0x3f4bU,
		0x9988U,0xe1f9U,0xc187U,0x59e7U,0x5b63U,0xa3d2U,0x35a9U,0x7326U,0x1ae3U,0x8de5U,
		0x5fd7U,0x08abU,0x4e10U,0x55d4U,0xf330U,0xd830U,0xe9d6U,0x1a3eU,0x054fU,0x5733U,
		0x0ffdU,0x55fcU,0x77eaU,0x9948U,0x5057U,0x93bfU,0x9df9U,0x8eecU,0x0644U,0xe93cU,
		0x3a4eU,0xb8efU,0xb03fU,0x4c53U,0xf7cdU,0xc84dU,0x98faU,0x626cU,0x3bb6U,0x4861U,
		0x9916U,0xb255U,0x374eU,0xf81eU,0x368bU,0xe3c2U,0x38a7U,0x3755U,0xbea8U,0xaeffU,
		0xc9fbU,0xf1beU,0x730eU,0x7a90U,0x3d3aU,0x77d4U,0x1b8bU,0xa9f4U,0xbfceU,0xdd73U,
		0x4d5cU,0x26a3U,0xf80eU,0xb28fU,0xa901U,0xc7a3U,0x5565U,0x6549U,0xac5eU,0xc2d6U,
		0x1a2aU,0x790cU,0x3087U,0x1058U,0xa69bU,0x8accU,0x9d7eU,0x9b5fU,0x3b63U,0xb720U,
		0x9fd0U,0xb3e9U,0x7f63U,0x8540U,0x41a0U,0x1402U,0x7347U,0xaa29U,0x8e5cU,0x7eb7U,
		0x45daU,0x9da4U,0xbf65U,0x4741U,0x8278U,0x6587U,0x0b8fU,0x2cdaU,0x848bU,0x7f79U,
		0xdb9cU,0x39ecU,0xccc9U,0x6fd0U,0x9a2aU,0xcaa9U,0xa9dfU,0x2265U,0xa986U,0xf958U,
		0xac21U,0xd2f5U,0x1706U,0xcbf1U,0x4338U,0x84ecU,0x810bU,0xc585U,0xc6ffU,0xa493U,
		0xbec6U,0x58b4U,0x94d5U,0xb730U,0x0058U,0x2189U,0x5aa5U,0xc368U,0x718aU,0xcb24U,
		0x4b85U,0x4e8cU,0xe596U,0x6b25U,0x27f3U,0x6a3fU,0x4aa0U,0xc434U,0xb1b9U,0x2b5dU,
		0x7b1cU,0xeedcU,0x1280U,0xef75U,0xd4e0U,0x05b6U,0x22b2U,0xf75bU,0x0855U,0xb7fbU,
		0x37f7U,0x631eU,0xc5ccU,0x4373U,0xac4aU,0xf977U,0x8d56U,0x6d03U,0x33e1U,0x4b0dU,
		0x6434U,0xb6bdU,0xfac2U,0x31ecU,0xa74aU,0x8d5fU,0x60c9U,0x1d17U,0xe241U,0xe7e1U,
		0x32b5U,0xe026U,0xfa91U,0x42b6U,0x1006U,0xf825U,0xa4d7U,0xb3f1U,0x99cdU,0x6a80U,
		0x8a69U,0x07f6U,0x61f4U,0x4355U,0xc0cbU,0x452dU,0xd2c7U,0xeba1U,0x77adU,0x5d59U,
		0xc7b6U,0x1804U,0x7c26U,0x61a9U,0x7fcaU,0xfe78U,0xff39U,0x5551U,0x79daU,0xfb10U,
		0xe60aU,0x1501U,0x8891U,0xc61fU,0x6e2bU,0xd157U,0xe775U,0x4c18U,0x8c13U,0xcfa5U,
		0xe53aU,0xebe8U,0x74e2U,0xba85U,0x9677U,0xd70cU,0xd572U,0x72d5U,0x8c7fU,0x0cf3U,
		0xd9b2U,0xfca0U,0xcef2U,0xe19aU,0xa295U,0x03faU,0x8a7fU,0xaaa7U,0xc310U,0xc162U,
		0xe2ebU,0xdb0eU,0x9a53U,0xe47dU,0x704aU,0xd8abU,0x2bcaU,0xf28cU,0x43b5U,0x51c9U,
		0xc5deU,0xa33eU,0xa573U,0x73a5U,0x68a9U,0xf943U,0xe432U,0x16a5U,0x89ecU,0xb367U,
		0x5966U,0x738eU,0xa77dU,0xbdf2U,0xeb11U,0xfeafU,0xdf78U,0xfde4U,0x9154U,0xc75aU,
		0x8102U,0xc337U,0x50cdU,0x2a77U,0xca3fU,0x9de7U,0xe03cU,0x5379U,0x14e2U,0x0254U,
		0x8027U,0x02d2U,0xafddU,0x6285U,0x71c5U,0x0ee1U,0xb17bU,0x3efeU,0x0b5cU,0x674dU,
		0x16bfU,0xb7c4U,0xdfe2U,0xb401U,0xe362U,0x2b47U,0x0e2fU,0xad3eU,0x68dfU,0xf409U,
		0x9194U,0xfa36U,0xa8d3U,0xa028U,0xd62aU,0xef00U,0x9a8dU,0x8d7cU,0x3572U,0x0acdU,
		0x02b7U,0x2412U,0x451aU,0x9901U,0x506cU,0x4a40U,0xc4b7U,0xbf16U,0xd750U,0x0b1bU,
		0x8311U,0x9fb8U,0x9e86U,0x3e2dU,0xb3d0U,0xf679U,0x48a9U,0xf587U,0xb8e1U,0xc012U,
		0x3fe4U,0x63daU,0x729cU,0xf3b9U,0xf4e7U,0xea6cU,0x3313U,0x6894U,0x949fU,0xdc81U,
		0xc3f5U,0x88f1U,0xcc38U,0xc86aU,0x978aU,0x33f7U,0x37c9U,0x6d7dU,0x2ff9U,0xaf84U,
		0x86b5U,0x188bU,0x2777U,0x64a3U,0x9867U,0xfc63U,0x204bU,0x1bc3U,0x3011U,0x63c5U,
		0x9837U,0x0222U,0x7d9aU,0x084dU,0x7c97U,0xf626U,0x29baU,0x0564U,0xb21fU,0x9051U,
		0xdc4cU,0x578aU,0xbabdU,0x81c3U,0xfe36U,0x2d4fU,0x8b18U,0xb586U,0xb624U,0xb555U,
		0xd7daU,0x1c34U,0x6e72U,0x0dc2U,0xf8ebU,0xdaebU,0xffc1U,0x1d04U,0x7e98U,0x93d2U,
		0x3326U,0x67c4U,0xea9cU,0x8822U,0x9c7bU,0x961cU,0xef67U,0xdbb3U,0x4e52U,0x37a4U,
		0x522fU,0x4eb6U,0x7b68U,0xd936U,0x2b82U,0x6444U,0xc35fU,0x1aafU,0xce63U,0x0e9dU,
		0x0287U,0xda08U,0x0f54U,0xe018U,0x7c9fU,0x9f7cU,0x86deU,0xcfbbU,0xcea1U,0x456fU,
		0xb160U,0x8861U,0x4936U,0x8d11U,0xc406U,0xc97bU,0x77c6U,0x69ffU,0x9f56U,0x1787U,
		0xae22U,0x814bU,0xfd66U,0x954bU,0x9717U,0x1eefU,0x53e1U,0x20a8U,0x8c4aU,0x0fa8U,
		0x7362U,0x6273U,0x580cU,0x7f6bU,0x26d7U,0xf0b4U,0x0682U,0xe5acU,0x5009U,0x8853U,
		0xf1deU,0xaae7U,0xa158U,0xf76bU,0x2ba8U,0xdfc9U,0x7c5cU,0x1becU,0x350aU,0x2f78U,
		0x3d0fU,0x0a4fU,0x2986U,0x19e8U,0x8814U,0x50faU,0xe564U,0x1901U,0xa61bU,0xda1dU,
		0xcb90U,0x945aU,0xecf5U,0x8eccU,0x5996U,0xe39cU,0x67a3U,0x1b67U,0xcd1eU,0x1c26U,
		0xa8a7U,0x57fcU,0xd807U,0x47f3U,0x781bU,0x2a16U,0x7235U,0xcd0aU,0xf55eU,0xf4d5U,
		0x2996U,0xfcd7U,0xde3dU,0x2f2eU,0x6886U,0x0f09U,0xdf68U,0x09f4U,0x4bd8U,0xa44bU,
		0xb087U,0x3832U,0x8f7cU,0x2c18U,0x261cU,0x1ecdU,0xd379U,0x0c33U,0x363fU,0x9dc9U,
		0xb1b0U,0x750dU,0x2d17U,0x479eU,0x03dfU,0xe9c6U,0x7f2aU,0x536bU,0xe954U,0xabf1U,
		0x1638U,0xc4d1U,0x7083U,0x9c9bU,0xaea3U,0x5c3cU,0xa8ccU,0xae1bU,0x2d2aU,0x9c0fU,
		0x6424U,0x2cf1U,0x98f3U,0xa743U,0x5716U,0xcb5eU,0x3f73U,0x78a9U,0x662dU,0x517cU,
		0x0625U,0x5de5U,0xfd00U,0x00fdU,0x8683U,0xfb5aU,0x6684U,0xfe6aU,0x160bU,0x8618U,
		0x95a2U,0xfa03U,0x1208U,0x251fU,0x7e4fU,0x01fcU,0x7945U,0x519eU,0x2d31U,0x5bcdU,
		0x4a70U,0xabd8U,0xd6feU,0x18b0U,0xac89U,0xbb67U,0xc0aaU,0x0d09U,0xbd3bU,0x3bbdU,
		0x80c5U,0xf6b7U,0x8cc6U,0x0ec7U,0x8a89U,0x9ae6U,0xeec4U,0xd690U,0xb5deU,0x6575U,
		0x680cU,0x0e47U,0x4321U,0x5897U,0xf68aU,0x0e39U,0x5d98U,0xf2f1U,0xc906U,0x0663U,
		0xde0aU,0xe82aU,0x92feU,0x49baU,0x250dU,0xecb9U,0x2055U,0x459aU,0xe755U,0x8b9bU,
		0x1442U,0xd1a9U,0xe5bcU,0x86e8U,0x43ecU,0x69e7U,0x8e10U,0xf2d3U,0x7967U,0xcd70U,
		0x3e80U,0x803eU,0x89feU,0xbd45U,0x5b35U,0xae45U,0x69d4U,0x1450U,0x3649U,0x118dU,
		0x1834U,0x296fU,0xec25U,0x25d2U,0xb8caU,0x1cddU,0x43cbU,0x0ab9U,0x66fdU,0x4b95U,
		0xa2bcU,0xcf28U,0x6707U,0x001cU,0x83c1U,0x4b05U,0xc973U,0x73c9U,0xc340U,0x8b04U,
		0x2537U,0xd643U,0xd452U,0xbc0cU,0x6fb3U,0xd842U,0x05adU,0x6ff0U,0x6b14U,0x1ab8U,
		0xcb3aU,0xc63eU,0xf174U,0x3824U,0xe323U,0x08f5U,0x9beeU,0x35fbU,0x8414U,0xb10bU,
		0x7516U,0x8258U,0xbe54U,0x10acU,0x8876U,0x2579U,0xd12fU,0x6ff9U,0x8ebeU,0xae53U,
		0xf62eU,0xf519U,0xa6a5U,0x85a6U,0x3af3U,0x5303U,0x21d3U,0xd103U,0x19d7U,0x27fdU,
		0x541dU,0x73d1U,0x1fb2U,0x5190U,0x00dfU,0x5a59U,0x9c16U,0x6d0cU,0x25c8U,0xa2caU,
		0xee60U,0x872dU,0x1130U,0xc563U,0x4fc3U,0xa017U,0x5605U,0x4868U,0x4790U,0x8548U,
		0x0353U,0xf33aU,0x62a2U,0xa6b9U,0xb782U,0xba0aU,0x35d4U,0xc2c6U,0xfd39U,0x3a9bU,
		0x33bfU,0x91ddU,0xfb50U,0x5d66U,0x1247U,0x5e0fU,0xe57dU,0x9dbcU,0xa467U,0x6ed2U,
		0x5c92U,0x9e33U,0x7cf5U,0xdee2U,0x75dbU,0xf042U,0x438eU,0xbda2U,0x746aU,0x8510U,
		0xb004U,0x6e38U,0xccfdU,0xa5e2U,0x8d2fU,0xd257U,0x71a9U,0x3770U,0x9126U,0x3c1cU,
		0x0bd0U,0xe1a7U,0x47b3U,0x0987U,0x7d35U,0xbd57U,0x5472U,0x50edU,0xb866U,0xb3bcU,
		0x9a21U,0x43c2U,0xb347U,0x8709U,0x30c4U,0xb058U,0x9d38U,0x6ea1U,0x7ce6U,0xbcf2U,
		0x4613U,0x5338U,0xaadaU,0x125bU,0x2d73U,0x45e9U,0xc938U,0xae0bU,0x7d6eU,0x56e7U,
		0x9ad3U,0xafeaU,0x494cU,0xbbafU,0xe6cdU,0x9114U,0x4036U,0x860bU,0x0996U,0x2acaU,
		0x9598U,0x56bfU,0xf509U,0xc917U,0x23e4U,0xfdf1U,0x5445U,0x466cU,0x7620U,0xdb25U,
		0x2e48U,0xbaecU,0x5bc0U,0xfe9bU,0x11f1U,0x3b0fU,0x3eb6U,0x24a5U,0x9070U,0xc51fU,
		0xf060U,0xc152U,0xb81bU,0xd4b8U,0x146cU,0x2274U,0x6b6bU,0xc4c4U,0x1daaU,0x8e4aU,
		0xd13fU,0xb7a8U,0x6409U,0x6378U,0x6634U,0x7638U,0x0b0bU,0x7b7bU,0xd0b4U,0xb4d4U,
		0x79f7U,0xd0dbU,0x516fU,0x7fa2U,0x6672U,0xbeafU,0x2350U,0xe7c1U,0xba2dU,0xfa43U,
		0x1872U,0xb22cU,0x18a6U,0xa067U,0x151aU,0x8056U,0x9636U,0x0656U,0xae7aU,0x1defU,
		0x890dU,0x1f05U,0x652fU,0x6ec2U,0x2e66U,0x1b9aU,0x5e3dU,0x5875U,0xb671U,0x7a36U,
		0x1caeU,0xb429U,0x8b8eU,0x64b0U,0xb06fU,0x71f1U,0x4001U,0xeb57U,0x9ec2U,0x64c5U,
		0xa0a3U,0xb2dfU,0xab59U,0x9ed2U,0x03bdU,0xd23aU,0x7479U,0xcd63U,0xea4dU,0xbeddU,
		0xc001U,0x7670U,0x60a3U,0xa113U,0x7070U,0xf4f4U,0x6838U,0xb5afU,0xff63U,0xc37bU,
		0x3f2dU,0x08e1U,0x0aacU,0xd090U,0xfbfbU,0x3232U,0x84e3U,0x73c0U,0xe285U,0x0b28U,
		0xd996U,0x6d26U,0xc5f1U,0xf377U,0x987fU,0xeda4U,0xd294U,0x824bU,0x0f32U,0xaf72U,
		0x8cbcU,0x5982U,0x74b4U,0x64deU,0x9f8dU,0x26bdU,0x2069U,0x1278U,0xce5bU,0x886eU,
		0x390fU,0x0122U,0x8af7U,0x8befU,0xc680U,0x70e4U,0x36e6U,0x1067U,0x165cU,0x53eeU,
		0x74bcU,0xd960U,0xc64eU,0x7788U,0x6128U,0xa102U,0x3019U,0x485bU,0x6fcaU,0x1d11U,
		0x2ae9U,0x0d1eU,0x0adfU,0xf0fbU,0x09eeU,0xabd2U,0x1cd7U,0xa2c4U,0xdb08U,0xeefaU,
		0x904bU,0x9545U,0x5183U,0x199bU,0x3ca8U,0xfaf4U,0x8703U,0x7b75U,0x08dbU,0xfaeeU,
		0x5ed0U,0x371bU,0xc8edU,0xaf06U,0x5d75U,0x9d86U,0x4889U,0x1564U,0x8205U,0x509eU,
		0x47caU,0x7a5fU,0xe4fcU,0x9c52U,0xfbe5U,0x1d26U,0x075bU,0x7952U,0xcf57U,0xee2dU,
		0x0363U,0xf5ebU,0xbde5U,0xf317U,0x57deU,0xdbc9U,0xea15U,0x4b61U,0x2be8U,0x50bbU,
		0x9231U,0xaf27U,0x96b8U,0xa308U,0x5c32U,0xbe99U,0xdd15U,0x3338U,0xb203U,0x0ed4U,
		0x733eU,0x795eU,0xec7fU,0x6f78U,0x1fc5U,0x7090U,0xea03U,0x12a3U,0x3078U,0xa960U,
		0x96c4U,0xb6aaU,0xb7b7U,0xb0b0U,0xd42dU,0xa713U,0xf402U,0x69f8U,0x01f5U,0x5189U,
		0xf357U,0x3423U,0x4c4cU,0xb6b6U,0x1a0dU,0x717bU,0xb795U,0xda95U,0xb7a1U,0xeef3U,
		0xafbeU,0x7266U,0x2f03U,0xb514U,0xec42U,0x2105U,0x0741U,0x7ff0U,0xb64dU,0x1edaU,
		0xf462U,0x11b8U,0x5544U,0xc466U,0x17d5U,0xf5dfU,0xd325U,0xdeb4U,0x8959U,0xf56bU,
		0xa88cU,0xee8fU,0xc283U,0x7873U,0x8cdaU,0x30acU,0x3c5dU,0x0e09U,0xa3afU,0xeb94U,
		0xfbd9U,0xf6ecU,0x03d5U,0x3806U,0xb4a7U,0xd66aU,0x52d3U,0xd342U,0x3853U,0x1346U,
		0x87f4U,0x5b74U,0xa331U,0xc3b1U,0x5e51U,0x9467U,0xf12dU,0xafc5U,0x2465U,0xbe8dU,
		0x062cU,0xb5bfU,0xbf9aU,0x19a1U,0x235fU,0x66a3U,0x2df1U,0xc5afU,0x2590U,0xe18dU,
		0xa178U,0xca02U,0x2b6fU,0x2bb5U,0x57d2U,0x2f8dU,0xffa5U,0xe181U,0x3ae4U,0xb118U,
		0x4639U,0x9fecU,0x0f9bU,0x5344U,0x017bU,0xa0c8U,0xfc35U,0xf1aaU,0x7d41U,0x324dU,
		0x6255U,0xcbbfU,0xbf8eU,0xbdaeU,0xcb54U,0x83b2U,0x4c05U,0xd45fU,0x58f2U,0x177aU,
		0x28b0U,0xabc0U,0x33cbU,0x7c70U,0x56ceU,0x4690U,0x6b92U,0x95cdU,0x6415U,0x8948U,
		0xec88U,0xbe81U,0xd929U,0x3135U,0xc9deU,0x2afdU,0x4738U,0xae3cU,0xae33U,0xfa4fU,
		0x335dU,0x8fd3U,0xde5aU,0xc570U,0x9a10U,0x34fbU,0x3847U,0x3caeU,0xb350U,0xdadeU,
		0x09a2U,0x6b34U,0x582bU,0x21e0U,0x1e0dU,0xe92aU,0x5a69U,0x3b1bU,0x6a5bU,0xe28eU,
		0x50d6U,0xc601U,0x187eU,0x1757U,0x3ec5U,0x7867U,0x9e77U,0xff28U,0xe215U,0x5881U,
		0x5403U,0xce6cU,0x943fU,0x8a96U,0x2206U,0xf098U,0xcba2U,0xf234U,0xab46U,0x81b6U,
		0x03c9U,0x0968U,0x2d20U,0x0278U,0x011aU,0xf22cU,0x64e6U,0xe837U,0x5245U,0x6b1dU,
		0x379cU,0x08edU,0x63cdU,0x7974U,0xfee7U,0x512bU,0x9a44U,0x70f0U,0xf114U,0x20f9U,
		0x9090U,0xf1f1U,0xc7f8U,0xcf30U,0x4f0bU,0x726dU,0x0f3eU,0x811aU,0x5decU,0x33eaU,
		0x5757U,0xfafaU,0xdb8bU,0x68d0U,0xf456U,0x1ceaU,0x0aebU,0x95eaU,0x77f3U,0xf1c5U,
		0x32e5U,0x9b67U,0xa517U,0xccb5U,0x7fd6U,0x467fU,0xc05aU,0x36b1U,0xad28U,0x8bbfU,
		0x27aeU,0x6595U,0x0b7bU,0x6335U,0xdafcU,0xebf4U,0xd5c4U,0x6a32U,0x6f94U,0x38c8U,
		0x9d75U,0xe74cU,0x28a7U,0x1729U,0xc19fU,0x4ae9U,0xa2a1U,0x90c7U,0xf478U,0x2035U,
		0x5f0bU,0xf0efU,0x58a7U,0xa34aU,0x5ff4U,0x002fU,0x00c1U,0xed2bU,0xa261U,0xd363U,
		0xd040U,0xb59cU,0xdff4U,0x02deU,0xc8e1U,0x018dU,0x1ec3U,0xac2fU,0xeb1dU,0xed71U,
		0x34b4U,0x6a73U,0x7e1eU,0x241eU,0x1d7fU,0x5d7bU,0xe3eeU,0xf149U,0x5ca9U,0x7771U,
		0x82d2U,0x198fU,0xe36eU,0xaa5bU,0xa076U,0x1d90U,0x6cf6U,0x3d91U,0x320eU,0xe51bU,
		0x4603U,0x5773U,0xc89fU,0x5796U,0x7b37U,0xe3d4U,0xc4efU,0xb282U,0x90d7U,0x5318U,
		0xf530U,0x2860U,0x9873U,0xd483U,0xf647U,0xa394U,0xf006U,0x88fdU,0xaeeeU,0x876cU,
		0x405cU,0x9354U,0xb3dcU,0xefadU,0x67f0U,0xba52U,0x6bfdU,0x999eU,0x2f17U,0x26b1U,
		0xa39fU,0x166cU,0x7cabU,0xe5f7U,0xb680U,0xdb5dU,0x215aU,0xd9f7U,0x6051U,0x8157U,
		0xf38fU,0x9d97U,0x36daU,0x1a14U,0x687cU,0x4febU,0x571dU,0xc7e8U,0x10efU,0x9d8cU,
		0xf8f7U,0xb935U,0x4af4U,0xcdb2U,0x3b37U,0x7a79U,0x39dfU,0x3692U,0x55e8U,0x777cU,
		0x9b8cU,0x4ec2U,0x4106U,0xfa0fU,0x0aa0U,0x2110U,0xdd90U,0x993bU,0xae68U,0x08bbU,
		0xd675U,0x339dU,0xc0b6U,0x1a9cU,0x9ee1U,0xd502U,0xbd2bU,0x5d01U,0xa3e8U,0x365aU,
		0xd3afU,0xf36bU,0xb248U,0x1d31U,0x347eU,0xc4ddU,0x7903U,0x49b1U,0x83a9U,0x99dbU,
		0x3f7fU,0x67e5U,0x709bU,0x6b04U,0x03b1U,0x2ed9U,0xfccdU,0x3158U,0xf1fcU,0x94ffU,
		0xd385U,0x59daU,0x0cb8U,0x1490U,0x8e2bU,0x7708U,0xe765U,0xfce3U,0x8da4U,0xce77U,
		0x67afU,0x7664U,0x8fdeU,0x488fU,0xe8e7U,0x059bU,0x55c7U,0x31faU,0xca84U,0xcbcaU,
		0xaf50U,0x0a5bU,0x2242U,0x9827U,0x702cU,0xa389U,0xb9a2U,0x22ecU,0xe663U,0xd4eeU,
		0x9bfaU,0x5624U,0xb063U,0x9959U,0x54bfU,0x200eU,0xac11U,0xc26dU,0x6d39U,0xf04dU,
		0x3df5U,0xd64fU,0x7643U,0x64eeU,0x9c22U,0x04d0U,0xff72U,0x7889U,0xabaaU,0x7a2aU,
		0x81ddU,0xf525U,0x9f95U,0xfdfcU,0x37e4U,0x856eU,0x1e53U,0x846cU,0x983fU,0xb2eaU,
		0xa5feU,0x0e53U,0x2e5aU,0x3c8fU,0x7811U,0x620fU,0x2061U,0x9437U,0x6d31U,0x670fU,
		0xb9f4U,0x7182U,0x9c73U,0xcb32U,0xcb65U,0xce45U,0x0075U,0x156aU,0x429cU,0x804aU,
		0xcae2U,0xd96bU,0x9771U,0xe621U,0x27beU,0xe3c9U,0x3726U,0x3b6bU,0x43d7U,0x8fa5U,
		0x555cU,0xc8fdU,0x624cU,0x9b0bU,0x36f5U,0x34a9U,0x7ddfU,0x454eU,0x5955U,0x5259U,
		0xb905U,0xdb19U,0x12dbU,0x41adU,0x2941U,0xca22U,0xdfdbU,0xb159U,0x1604U,0x21ccU,
		0xda80U,0xe096U,0x9df2U,0xac43U,0x2153U,0x780aU,0x11e5U,0x39f9U,0x1465U,0x0ad8U,
		0x87edU,0xf10dU,0x54f9U,0x3d47U,0x082bU,0x616aU,0x2c59U,0xe2c1U,0x6219U,0xb591U,
		0x044dU,0xadffU,0xf159U,0x58c8U,0xc875U,0x3162U,0xf441U,0xa925U,0x4ff5U,0xf0bfU,
		0xea2eU,0x5ed6U,0xbc32U,0xe2ffU,0x04b3U,0x7919U,0x402aU,0x5915U,0xda13U,0xac39U,
		0x57b8U,0x76d1U,0x3871U,0x0090U,0x965aU,0x82c7U,0x9ce4U,0xe7eeU,0x3810U,0x8623U,
		0x54dfU,0x9929U,0x9fcbU,0x1701U,0xd4dbU,0xa4d2U,0xe63dU,0x0da2U,0xe9d1U,0xbea3U,
		0x37c4U,0xb61fU,0x0282U,0x5807U,0x8c0eU,0xc5d9U,0xafd8U,0x5067U,0xe5b7U,0x83bcU,
		0xd12aU,0x478bU,0x5991U,0xd374U,0x6628U,0x80c0U,0x9ebdU,0xd991U,0xc649U,0x5d70U,
		0xafffU,0xaad5U,0x11ecU,0x234bU,0x56c9U,0x5a64U,0x5240U,0x32e0U,0xdd10U,0xb79cU,
		0xfbd4U,0x2b6aU,0x3f7aU,0x223dU,0x37dfU,0x3721U,0x58a2U,0xa071U,0x6bf8U,0x55e3U,
		0xe727U,0x94e2U,0xdc21U,0x38aeU,0x1460U,0x57b3U,0x598cU,0x145bU,0xaab1U,0x7c06U,
		0xbfb0U,0x9db0U,0xa5c1U,0xad51U,0x1158U,0x093fU,0xa9ebU,0x656cU,0x9fe3U,0xad48U,
		0x139cU,0xc04fU,0x9b14U,0x9037U,0xb129U,0xfe19U,0x35c8U,0x99e4U,0x1102U,0xaa00U,
		0x1bcdU,0x3272U,0x1732U,0xc485U,0xd050U,0x9fffU,0xe72cU,0xcfc5U,0x3093U,0x3a5dU,
		0x2caeU,0xbd96U,0x6f81U,0x1af1U,0x60d5U,0x09bdU,0x66c7U,0x3bc7U,0x3d74U,0xfa58U,
		0x7eedU,0x142bU,0x9dd3U,0xb8a9U,0x4458U,0x632aU,0x5c86U,0xa2d4U,0xe5a2U,0xc8f3U,
		0x8712U,0xf9c2U,0x8a52U,0x29faU,0x9e4cU,0xc591U,0x4b9fU,0x2718U,0x94e7U,0x304aU,
		0x1f7aU,0x0b31U,0x2a80U,0x8440U,0x6b54U,0x6970U,0x474cU,0x2c92U,0xe64eU,0x3764U,
		0xc6e5U,0xc194U,0x676eU,0x0ea6U,0x6717U,0x3a30U,0x16e6U,0xb29aU,0x0419U,0xe83fU,
		0xb6c8U,0x8928U,0x84abU,0xf268U,0x47a7U,0x8efaU,0xcc24U,0x4e97U,0x26c5U,0xe40dU,
		0xdc26U,0x1ac1U,0xef4aU,0x5740U,0x022aU,0x15ccU,0x8ad6U,0x3dd6U,0x5156U,0xb04cU,
		0x3a00U,0xb265U,0x7d10U,0x6a49U,0x4e1cU,0x0024U,0x0d2dU,0xda9cU,0x9b3dU,0x4f64U,
		0x78b0U,0x4b16U,0xf8bfU,0xaf0cU,0x2d67U,0x3410U,0x7c52U,0xb8fbU,0x597aU,0xff20U,
		0x9951U,0x375cU,0x38b3U,0xb728U,0x93caU,0xd83aU,0x0703U,0x7ab4U,0x70c5U,0x5adeU,
		0xf914U,0xf918U,0xcaf3U,0x99f3U,0xc1dcU,0x5b0eU,0xca6fU,0x111dU,0x274aU,0x741eU,
		0xff8eU,0xa589U,0xd05eU,0x1b37U,0xc80aU,0x7453U,0xfc4fU,0xf91cU,0x2b29U,0x7d64U,
		0xdedaU,0x50b3U,0x4ae1U,0xe08eU,0xa5c6U,0xc0ddU,0x14f9U,0x18f9U,0x6f01U,0xfa6fU,
		0xd890U,0x51f0U,0xf69dU,0xcaf7U,0x5f69U,0x0806U,0xbe3bU,0xf42aU,0x5b9dU,0xf5b9U,
		0x61b8U,0x78e5U,0x9f28U,0xfc15U,0xb920U,0xebb2U,0x8aacU,0x1520U,0xe8d0U,0x99f7U,
		0xcffbU,0xcce9U,0xd5e3U,0x5587U,0xb82eU,0x24e3U,0xad56U,0xfd7fU,0xdca9U,0x3543U,
		0x9cd2U,0xef23U,0xd068U,0x8bdbU,0x4f4fU,0x0707U,0x5cc4U,0x413fU,0x9e5aU,0x6dfcU,
		0xe384U,0xc073U,0xfbb1U,0xf5ccU,0x22a3U,0xb2c6U,0xbd03U,0x3ad2U,0x208fU,0x910aU,
		0x6df2U,0x7ab8U,0x9cafU,0xc248U,0x1a81U,0x3e0fU,0x8969U,0x6adeU,0x115dU,0x774aU,
		0x2323U,0xbfbfU,0xe34eU,0x179cU,0xf8d9U,0x6bd2U,0xfa66U,0x70c9U,0x2c60U,0xd613U,
		0xf498U,0xba7cU,0xd24eU,0xed22U,0xf7beU,0xe503U,0x1f29U,0xe97dU,0x668bU,0x3a6dU,
		0xaa78U,0x805cU,0x8c22U,0x5ae2U,0xa9a7U,0x219fU,0x90f3U,0x1220U,0x19c0U,0x4985U,
		0x0944U,0x0cd3U,0x0816U,0x4279U,0xadadU,0x1313U,0xc4a8U,0xfa80U,0x5ba8U,0x274eU,
		0xf869U,0x02f4U,0x25b4U,0x5280U,0xd8e1U,0x98d1U,0xb2a8U,0x7209U,0x5023U,0xc1e7U,
		0x76a8U,0x392aU,0x1f34U,0x8c70U,0x6c14U,0x7422U,0x75c6U,0xff49U,0x6de8U,0xfbbcU,
		0xededU,0x2c2cU,0xaab6U,0xc496U,0x951cU,0x543bU,0xef58U,0x8f6bU,0x5f45U,0xadb8U,
		0xf84aU,0xff92U,0xc08bU,0x107bU,0x5f50U,0x4c93U,0x95b3U,0x11a4U,0xd94bU,0x9bb0U,
		0xa7b0U,0x5931U,0x14d1U,0xc3e6U,0x85deU,0x963cU,0x8195U,0xa58dU,0x192cU,0xa2e3U,
		0xd187U,0x550dU,0x4ab6U,0x22d3U,0x7c0bU,0xd788U,0xe2d3U,0xe869U,0x88c6U,0x9b24U,
		0xd722U,0xad9aU,0xa582U,0xc1e0U,0x6794U,0x515eU,0x28c8U,0x0fd9U,0x93feU,0x8cebU,
		0x42a4U,0x23c8U,0xd00bU,0xa7e1U,0x8426U,0x3ecdU,0xb1e2U,0x9808U,0xdaaaU,0x5b12U,
		0x96a7U,0x7f1bU,0xb4ffU,0x3be3U,0xc688U,0x249bU,0xbfb5U,0x2c06U,0x07d4U,0x4625U,
		0xdc54U,0xcee9U,0x7d2cU,0x1ebaU,0x1825U,0xca73U,0xe2c8U,0x4f26U,0x6d4eU,0x8696U,
		0xb524U,0x3e88U,0x5b94U,0x9c59U,0x7f90U,0x2cbdU,0xfe71U,0xf3b2U,0x0a48U,0x065cU,
		0xcc13U,0x1121U,0x65e5U,0x6542U,0x38c1U,0x5cfaU,0xca1bU,0x9030U,0x9db5U,0x1b93U,
		0xf310U,0xcbb8U,0x1f1fU,0x0909U,0xa853U,0x5f75U,0x3d55U,0xfc53U,0x7bc3U,0x63ffU,
		0xccf5U,0xb1fbU,0x5fecU,0x87e5U,0x46c9U,0xee4eU,0xe532U,0x679bU,0xbc6dU,0x305aU,
		0x48c2U,0xaf9cU,0x14f1U,0xf920U,0xa643U,0x07a2U,0x7f10U,0xa80bU,0xafafU,0x7575U,
		0x13a1U,0xa360U,0xf928U,0x4090U,0xfad8U,0x9294U,0x3d34U,0xc6f9U,0x8068U,0x2b2dU,
		0x2556U,0xbbfcU,0x5003U,0x2d11U,0xeebeU,0x2531U,0x61eeU,0xe42cU,0x830bU,0xf8e5U,
		0x47c4U,0xec3cU,0x3adeU,0x1878U,0xb77cU,0x7d68U,0x18a0U,0xd28eU,0x6246U,0x9cdeU,
		0xbfaaU,0x6b4eU,0xc054U,0x45fdU,0x0a9aU,0xe65dU,0xa84dU,0xbca7U,0xbcadU,0xd4aeU,
		0x70bfU,0xd062U,0xc4a2U,0xd71cU,0x647dU,0x292bU,0x1ea2U,0x56feU,0xbe01U,0xb2b4U,
		0xdec9U,0xfd2aU,0xe875U,0x88a2U,0x1e74U,0x4a27U,0x695aU,0x1b3bU,0x4404U,0x6f30U,
		0xd16dU,0xd21fU,0x4da8U,0xa7bcU,0x9b19U,0x8351U,0x3ea7U,0x62bbU,0xe0bdU,0x28d4U,
		0x771bU,0xdd52U,0x3b50U,0xc80eU,0x6220U,0x8f09U,0x938fU,0x2669U,0x3befU,0xefd0U,
		0xad10U,0xfbecU,0xef97U,0x7498U,0x1612U,0xf3eaU,0xca99U,0xb232U,0xcce0U,0x7457U,
		0x1584U,0xb530U,0x62ccU,0xd8a2U,0x6e98U,0x7768U,0x903cU,0x8096U,0x93f3U,0x714aU,
		0x44ffU,0xdea2U,0xbcb3U,0x66b8U,0xb37eU,0xa5caU,0x5c65U,0x8e6eU,0x90e9U,0xe881U,
		0xd352U,0x42d3U,0x0898U,0xe2b1U,0x12e4U,0x5165U,0x9025U,0x8de1U,0x93bbU,0xb251U,
		0xdd5eU,0xc0e1U,0xd90fU,0xc828U,0x58b0U,0xc430U,0x436fU,0xe7ddU,0xa9f0U,0x1054U,
		0x7eb3U,0xcaa5U,0xb363U,0x9de3U,0xb7c0U,0x8d78U,0x4529U,0x14fdU,0x72d1U,0xe479U,
		0x1d00U,0xd932U,0x456bU,0x1eebU,0x3e29U,0xdc7dU,0xfc5fU,0x5786U,0xe9c2U,0x2cedU,
		0xfe66U,0x18acU,0xaae3U,0x18fdU,0x47efU,0xa447U,0xbc08U,0xb107U,0x52ffU,0x8729U,
		0x6571U,0xecb5U,0x803aU,0x0ab5U,0xafe6U,0x4668U,0xd4b4U,0xb4d0U,0xc2c2U,0xdedeU,
		0x3c18U,0xb054U,0xaf6eU,0x70e0U,0x0d1aU,0x7b71U,0x8052U,0xb425U,0xcd5fU,0xd08cU,
		0x7fecU,0x786fU,0x1342U,0x669fU,0x9c4eU,0x50b7U,0x708cU,0x341fU,0x7863U,0x0964U,
		0x70ecU,0x68ccU,0x9fe8U,0xd45bU,0x3131U,0xdadaU,0x5792U,0x8768U,0xdb59U,0xb931U,
		0x8bbbU,0x4ae5U,0xb598U,0xf145U,0x5620U,0x7885U,0x3c8bU,0x8046U,0x9937U,0x1d2dU,
		0x94fbU,0x0597U,0x861fU,0xc5d5U,0xaad1U,0xa06dU,0x34a5U,0xe092U,0xe2bdU,0xe2fbU,
		0x2c8eU,0xf264U,0xb048U,0x340cU,0xad4dU,0xa9fcU,0x09b9U,0xf9beU,0x23bbU,0x677eU,
		0xbe07U,0x8dd3U,0xbaccU,0xf6a1U,0x6bc6U,0xe901U,0xb855U,0x812eU,0xe8c5U,0x2370U,
		0x7850U,0x6f0eU,0xaa88U,0x88cdU,0xbc1eU,0x248eU,0xeca7U,0xb976U,0x3b79U,0xcafbU,
		0x4ddaU,0xf48cU,0xec0fU,0xd314U,0x2faaU,0xebbfU,0x1107U,0xbcd7U,0xa64bU,0xbd69U,
		0xd9dcU,0x70d2U,0xcedcU,0x1d4cU,0x5ad6U,0x5f6dU,0x8dcbU,0x8329U,0x3538U,0x816dU,
		0x9c62U,0x32c9U,0x673aU,0xacebU,0x2e7fU,0x045cU,0x4ad7U,0x6c6eU,0x2dccU,0x6535U,
		0x1d60U,0x080aU,0xce2eU,0x7460U,0xba02U,0x7a57U,0x36a9U,0x9b03U,0xaa05U,0x3dc3U,
		0x434dU,0x884bU,0x4d73U,0xc544U,0xb2baU,0x5aefU,0x7415U,0x6f05U,0xf65cU,0x9286U,
		0x4582U,0x61a0U,0x5e06U,0x6a29U,0x847aU,0xf191U,0xd885U,0x420aU,0x858dU,0x6ba8U,
		0x9f16U,0x4f43U,0x7536U,0xfa73U,0xee40U,0xca4fU,0x06e1U,0x3739U,0x0e82U,0xffe0U,
		0x1bd2U,0x2729U,0x6c9bU,0x3db4U,0x3e03U,0xd558U,0xd6a8U,0xd2e0U,0xff0cU,0xd894U,
		0x5500U,0x0881U,0x4220U,0x1540U,0x1649U,0x23a6U,0xf41fU,0x820cU,0x7934U,0xc255U,
		0xfd07U,0x2083U,0xcc9dU,0x6ec8U,0x1b2dU,0x51f4U,0x8e98U,0x5924U,0x1176U,0x5cd1U,
		0x2b12U,0x4dfdU,0x3277U,0xbb21U,0xbc44U,0x7fb9U,0xde77U,0xa1dfU,0xdecfU,0xb767U,
		0xb9deU,0x61bcU,0x38f7U,0x2c39U,0x0cc6U,0x920dU,0x1019U,0xf3a4U,0xe5e3U,0xbb30U,
		0x24d7U,0x9b2bU,0x8521U,0xd8eeU,0xd539U,0x1045U,0xc1b2U,0x78e9U,0x664cU,0xcf88U,
		0x1c72U,0xed15U,0x2bf8U,0x513eU,0xb12eU,0x8642U,0xd41bU,0xb914U,0x7448U,0xa8e9U,
		0x0d7dU,0x2702U,0xcbd8U,0x9f2cU,0x398dU,0x05f3U,0x29e0U,0x4c85U,0x3ac5U,0x551aU,
		0x78d9U,0x9182U,0x838cU,0x46f9U,0xbf2dU,0xe41dU,0x3189U,0x2e6cU,0x94aeU,0xfc19U,
		0x7afbU,0x46a5U,0x2740U,0xf24fU,0xa23dU,0x4132U,0xfe1eU,0x5f5dU,0x6be2U,0x123dU,
		0x0c10U,0xd2beU,0xfd30U,0x27a5U,0x7c84U,0xbe3fU,0xff83U,0x3a08U,0x1c8eU,0xd018U,
		0x7d58U,0x28f2U,0xcaeaU,0x6bbdU,0xdf17U,0xef15U,0xc0c4U,0x9e7fU,0xf1d7U,0xb5d7U,
		0xe4b1U,0xf42eU,0xdce4U,0x1b84U,0xad21U,0xdd89U,0x36eeU,0xa5baU,0x35cdU,0x1513U,
		0xe4f5U,0x4632U,0xbac5U,0xb9d7U,0x0a1aU,0xfff3U,0xf90dU,0x5ba1U,0x3d4eU,0xb377U,
		0x7aa8U,0xf784U,0xa669U,0x8a06U,0x3e3cU,0x50a5U,0x5b01U,0x8ca2U,0x0104U,0x72e6U,
		0x836aU,0x1ba0U,0x7b96U,0xf5bdU,0x7991U,0xa79bU,0x8d90U,0xca63U,0x34dfU,0x3a91U,
		0x99e9U,0xbfd6U,0x5666U,0x18ecU,0x7b45U,0x871cU,0x6483U,0x8331U,0x0a21U,0xe8d4U,
		0x21f6U,0xb4f3U,0x358aU,0x0d5cU,0x78baU,0x1c12U,0x443bU,0x92bcU,0x475fU,0xd729U,
		0xdc70U,0x6009U,0x394eU,0xf9dcU,0x4bc5U,0x99fbU,0x97fcU,0x57c2U,0x754eU,0x4411U,
		0xf7b3U,0xedd9U,0xe731U,0x874eU,0x07aaU,0x700cU,0xcff1U,0x56a4U,0x4d8aU,0xc81bU,
		0x6828U,0xcfffU,0x9849U,0x9d2eU,0x33b7U,0x0753U,0x0b73U,0x594dU,0x9362U,0xf206U,
		0x77a5U,0x2ba0U,0x605bU,0x0d4eU,0xde95U,0xae80U,0xc1d4U,0xccedU,0xb84dU,0x3582U,
		0xd608U,0xe136U,0x9efaU,0xeb27U,0xcfcaU,0x7efeU,0x8f29U,0x4236U,0xdda4U,0x48cfU,
		0x2931U,0xcc07U,0x3c67U,0xb924U,0x4a0cU,0xcabbU,0xba64U,0x9bd6U,0x12b3U,0xead7U,
		0x5085U,0xbb7fU,0x5119U,0xdd1fU,0xbdc5U,0xf9eaU,0xcbacU,0x1f6aU,0xe085U,0xebb6U,
		0x6100U,0x5f82U,0xb68eU,0x4c0fU,0x8507U,0xf0e6U,0x3098U,0xebceU,0x19b5U,0x6158U,
		0x686cU,0x7d20U,0x6d91U,0x1251U,0xfd75U,0x8ab0U,0x0832U,0xd07fU,0xe6d7U,0xa75fU,
		0xc9a8U,0x1357U,0x6cb4U,0x47d1U,0xbd13U,0xc234U,0x868aU,0x1e43U,0x8f4eU,0x1df5U,
		0xb943U,0x1524U,0x0957U,0xe4d1U,0x89bbU,0x68fcU,0x5dc8U,0x92a1U,0x3a62U,0x6491U,
		0x9889U,0x1f9eU,0xd9bcU,0x2292U,0x1ea8U,0xc031U,0xfffaU,0xad5aU,0x050eU,0xb411U,
		0xd3ceU,0x87d8U,0xbcc9U,0x68e9U,0x0dd1U,0xcb79U,0x41b6U,0xada1U,0x82faU,0x9a64U,
		0x9cc7U,0x1d3aU,0x1382U,0xfd83U,0x86bfU,0x355eU,0xa353U,0x2884U,0xbd7eU,0x6c60U,
		0x1737U,0xceabU,0x98c5U,0x45b1U,0x2171U,0xcc85U,0xb0b9U,0x24c6U,0x5abaU,0xdcadU,
		0x5635U,0x80cfU,0x6ef7U,0x313eU,0xa9c8U,0xbbefU,0x8f86U,0x75baU,0x2906U,0x8d25U,
		0xc068U,0xee6aU,0xab1bU,0x8913U,0x1264U,0x3547U,0xd85dU,0x014dU,0x80a7U,0x1532U,
		0xec2fU,0x6c27U,0xc48aU,0x7f53U,0xd58eU,0xd099U,0x628eU,0xe260U,0x5704U,0xc888U,
		0x242cU,0xd5e7U,0xf6d0U,0xb26dU,0x8b34U,0x40aaU,0x728cU,0xd93fU,0xd973U,0x4a1aU,
		0x9174U,0x9005U,0x49dbU,0xff30U,0xa45eU,0x9d6cU,0x4e61U,0x558bU,0x9ca4U,0x6f48U,
		0x9374U,0x25fdU,0x106fU,0xa563U,0xd055U,0xbc15U,0xef33U,0xa1d1U,0x81cbU,0x5f90U,
		0xac68U,0xb572U,0x1cc0U,0xb832U,0x6171U,0x42e0U,0x00cfU,0x4aaaU,0x2a4dU,0xbd4dU,
		0xf5afU,0x7a66U,0xab2cU,0x713dU,0xfb61U,0x437cU,0x520aU,0x1f0bU,0x773fU,0x24e7U,
		0x538fU,0xddecU,0xc3daU,0x0615U,0x4c5cU,0xb7dfU,0xa004U,0xdd7bU,0x265cU,0xe9e0U,
		0x0609U,0xf8b1U,0x6960U,0xd9e4U,0x3e43U,0x4f53U,0xdf20U,0xbfe0U,0x6063U,0xeab6U,
		0xcf40U,0x4c31U,0xf259U,0xf665U,0x7a18U,0x42abU,0xef3cU,0x51feU,0xd0f0U,0x3711U,
		0x5cf0U,0x070bU,0x30dfU,0x5e26U,0xc291U,0x3f9cU,0x6c78U,0x6109U,0x60daU,0xc040U,
		0xbc75U,0xf030U,0x3148U,0x2c69U,0x7131U,0x7edfU,0x1217U,0x5cc8U,0xeaf4U,0xa702U,
		0xa4eeU,0x2409U,0x7f99U,0x7a70U,0xa731U,0x60beU,0xaac6U,0xf019U,0x6cdfU,0x1719U,
		0x114aU,0x1cb4U,0x7ffeU,0x4143U,0x56aeU,0x158dU,0x417eU,0x2e07U,0x4d33U,0x3450U,
		0x09c2U,0x8abaU,0xfec2U,0x76c2U,0x2dffU,0xfa25U,0x1b41U,0xe050U,0x557fU,0x9cd6U,
		0x7a4fU,0xe12eU,0x058cU,0x7ac2U,0x0c44U,0x31e1U,0x83f9U,0x2eabU,0x5d41U,0xfe4eU,
		0x0a3eU,0x1b12U,0x38dfU,0x87a6U,0xfbc6U,0xef27U,0x7be2U,0x3c9eU,0xbceaU,0x7173U,
		0x1987U,0x590dU,0x66ccU,0x1416U,0xf284U,0x2c10U,0x0776U,0xd497U,0xa6e6U,0x8c33U,
		0xc142U,0xd06cU,0xc87cU,0x11d7U,0x6c3cU,0xc5c1U,0x8bb0U,0xe68aU,0x5b82U,0x036aU,
		0x0913U,0x04edU,0x7e56U,0xf7f3U,0x4e7bU,0xb42fU,0xc507U,0x8bdfU,0x04a5U,0x8b28U,
		0xe88bU,0xeafdU,0x53bbU,0x998fU,0x3bccU,0x6728U,0x866cU,0x583fU,0xd30dU,0xed0eU,
		0x440aU,0x287dU,0x50acU,0xfbb5U,0xa804U,0xc429U,0xcc79U,0xd3a1U,0x1695U,0xce07U,
		0x3f95U,0xa9b5U,0x968bU,0x23cfU,0xc448U,0xe30fU,0x3cedU,0x7d88U,0xe963U,0xf5d0U,
		0x30faU,0xa1aaU,0x701eU,0xa9bcU,0xa866U,0x953bU,0x2cb3U,0x39c7U,0x7219U,0x8b52U,
		0x4ef9U,0xd621U,0x0b02U,0x10e6U,0xbeefU,0x22a7U,0x3126U,0xcd85U,0x57ecU,0xa833U,
		0x0141U,0xac03U,0xaa6fU,0xcc94U,0x4e72U,0x73f7U,0xd150U,0xc863U,0x9453U,0x8b94U,
		0x13d7U,0xb2caU,0x602dU,0x9b58U,0x202eU,0x49aaU,0x7803U,0xc47eU,0xbd9bU,0x71eaU,
		0xa301U,0xabb9U,0xb4c5U,0x9c3dU,0x6f36U,0xa488U,0xcdceU,0x9e5eU,0xc9caU,0x0232U,
		0xb320U,0xcc4eU,0x55a8U,0x01d4U,0xcb6dU,0xa253U,0x9337U,0x41efU,0x8184U,0x50deU,
		0x5a3eU,0x0790U,0xe810U,0x6e00U,0x568bU,0x9743U,0x181bU,0x5fafU,0x3917U,0x2d97U,
		0x6f86U,0x0502U,0x7df9U,0x2adfU,0x29c2U,0x004dU,0xaf63U,0x4bb4U,0x5fb9U,0xe388U,
		0x3169U,0x9968U,0x4477U,0xd7c0U,0x40e0U,0x8e18U,0x630eU,0xf5f2U,0x6c54U,0x39d1U,
		0x0203U,0xa198U,0x1b1cU,0x64b6U,0x947dU,0xc077U,0x4821U,0x4e35U,0x7586U,0x0715U,
		0xf54bU,0xe200U,0x1af6U,0x7316U,0x404eU,0x82eeU,0xb1f5U,0x2925U,0xe87bU,0x70daU,
		0x5b08U,0x6df6U,0x527aU,0x0fd3U,0x0d56U,0x9bd0U,0x87d2U,0x40a4U,0x8128U,0x619aU,
		0x9207U,0xd012U,0x9bcaU,0x01e7U,0xf7a1U,0xb310U,0xeab0U,0x7abcU,0xd39bU,0xcc48U,
		0xa535U,0x9692U,0x92e1U,0xcdf4U,0x9dd8U,0x1ad2U,0x3062U,0x13f7U,0xa525U,0x1975U,
		0xb63fU,0xe079U,0x3f17U,0x9cb3U,0x20c2U,0xc954U,0x1fd9U,0x528eU,0x29a7U,0x8dfbU,
		0xfb95U,0x388eU,0x9706U,0xb546U,0x15ebU,0x6dc8U,0x614cU,0x5e43U,0xe974U,0xc24cU,
		0x04e4U,0x2ee8U,0x7107U,0xfa2dU,0x636fU,0x4fe2U,0xb8aeU,0x826dU,0x4560U,0x24fcU,
		0x07caU,0x7defU,0x88a8U,0x3eecU,0xde4cU,0xbd07U,0x3621U,0x611eU,0xb33fU,0x01edU,
		0x1d77U,0xea26U,0x2829U,0xa4bcU,0xe2e3U,0xdf60U,0x1142U,0x6144U,0x3673U,0x278dU,
		0x8aa4U,0x3ad6U,0x2dc4U,0xde8dU,0xa43cU,0xc66fU,0x06a0U,0x225aU,0x445dU,0x4ac4U,
		0x01b8U,0x43a0U,0xdf90U,0x15baU,0x269dU,0x2183U,0x39aaU,0x2093U,0xf448U,0x89afU,
		0xdfc3U,0xabebU,0x5891U,0x6feaU,0x5d53U,0xbdecU,0xf581U,0x67beU,0x794cU,0x11b2U,
		0xa0c2U,0x587bU,0x91d7U,0x910eU,0x1de9U,0x64d8U,0x5253U,0x16fbU,0x6566U,0xc18eU,
		0x632fU,0xb27cU,0x3397U,0x2008U,0xb45dU,0xee15U,0x1e7aU,0x9e14U,0x8ca9U,0x1161U,
		0x8fefU,0x211fU,0xba1fU,0xf7a7U,0x2c7eU,0x8241U,0xad90U,0xa47eU,0x294aU,0xa7e8U,
		0x1e1bU,0x4258U,0xe227U,0x3487U,0xe2a6U,0x774eU,0xfdcdU,0xc798U,0x5ffdU,0x23d6U,
		0x8658U,0x646dU,0x3d79U,0x8c91U,0x1c4aU,0xe70bU,0xc6c3U,0x1c5cU,0xd436U,0x17ebU,
		0xd1bcU,0x2327U,0xb886U,0xf597U,0x6dd8U,0xc9e7U,0x0bc2U,0x442fU,0xe3abU,0xb9c5U,
		0xfd19U,0x59ffU,0x62f5U,0x16b6U,0xb812U,0xb677U,0x2d7cU,0xbfc3U,0x5c5aU,0x4f81U,
		0xfaacU,0xbe2bU,0xb4e7U,0xfb42U,0xfa5dU,0xff03U,0xc139U,0xfc04U,0x71ceU,0x2ccbU,
		0x4a2dU,0x65c8U,0x17bcU,0x1a85U,0xefe4U,0x15d4U,0x28beU,0xb316U,0x65afU,0x2b8bU,
		0xce4dU,0xb841U,0xd5f8U,0x74aaU,0xe7d2U,0xc609U,0xe8abU,0xb715U,0xb0d4U,0x3e13U,
		0xaeacU,0xbbd6U,0xb142U,0xd7eeU,0x012aU,0x0a71U,0x7ef2U,0x6fbcU,0x5ef9U,0xd3beU,
		0x056cU,0x8a32U,0x3c05U,0x90a9U,0xa0d6U,0x896dU,0xa92cU,0xacaaU,0x69baU,0xacc8U,
		0x41deU,0xf2dbU,0x06d1U,0xf31eU,0x8b80U,0xcfe5U,0x51a5U,0xac53U,0x901aU,0x7a3cU,
		0x3519U,0x6ae2U,0xb19aU,0x85bdU,0x4ccfU,0xb5feU,0xf7d6U,0x59eeU,0x1430U,0x7a98U,
		0x5941U,0xe277U,0x130eU,0x8f66U,0x9b1fU,0xcee4U,0x7aafU,0xfa6aU,0xef1eU,0x1797U,
		0xde9dU,0x9ddeU,0x4663U,0x8763U,0x0904U,0x928fU,0xbca2U,0x28cfU,0xa1daU,0xa8e4U,
		0xd2b9U,0xb9d2U,0x6779U,0x70cdU,0xc53fU,0xd553U,0x228dU,0xcc80U,0xe25bU,0x5f8bU,
		0x8717U,0x569fU,0x48caU,0x7d1bU,0xed09U,0xd61cU,0x9c38U,0x0048U,0xf8acU,0x2c64U,
		0xfa20U,0xd492U,0xee10U,0x1c57U,0x2cc6U,0x8a2dU,0x2920U,0x1970U,0x7deaU,0x15b5U,
		0x71b7U,0xd52dU,0x367bU,0x9ec8U,0x8f61U,0xd617U,0x1a5aU,0xb539U,0x9cefU,0x1b49U,
		0x4dc6U,0x2761U,0xf9c7U,0xcecbU,0x343eU,0xa7f5U,0xa4afU,0x1a5fU,0x8357U,0x76e4U,
		0x6ee0U,0xe352U,0x26f6U,0x977cU,0x91f8U,0x1ad8U,0x694fU,0xacd5U,0x9fa0U,0x622eU,
		0xa5a6U,0xa685U,0x7838U,0x224dU,0xd81aU,0xf891U,0xc34fU,0x17a0U,0xbaafU,0x29d4U,
		0x4771U,0x2db5U,0x70a6U,0x268cU,0x8a57U,0x4cdcU,0x0fe9U,0xb6e8U,0xf011U,0xb53eU,
		0x15adU,0x7bacU,0xfc0dU,0xf8ddU,0x0454U,0x2b98U,0x62b0U,0xf99bU,0xc12aU,0x5728U,
		0x5db2U,0x0762U,0x53a0U,0x2ef1U,0x95a9U,0x0137U,0x7400U,0x64cbU,0x9f48U,0x6bd6U,
		0xa2acU,0x6099U,0xb3b4U,0xa70bU,0x6a6bU,0xa91dU,0x29ffU,0xe3f9U,0xaa9fU,0x2f26U,
		0x147aU,0x71bcU,0x3eadU,0x2f0eU,0xf78bU,0xf7c2U,0xef8cU,0xd3feU,0x52deU,0x3068U,
		0x1b78U,0x162aU,0x231aU,0x6c92U,0xf27bU,0x0fe0U,0x79d3U,0x330cU,0x880dU,0x92f7U,
		0xcd96U,0xe507U,0x926cU,0x1a23U,0x28a0U,0xd3a8U,0x15fdU,0x35c1U,0x5c8bU,0xb66aU,
		0xea0eU,0xcb4dU,0xeca0U,0xd532U,0x3947U,0x9cc0U,0xc803U,0x1f2dU,0x48bbU,0x93b4U,
		0xd2b2U,0x8806U,0x8838U,0xe6bdU,0xd0e9U,0x3ce6U,0xf79aU,0xe220U,0xb475U,0x607aU,
		0xde65U,0xa0a9U,0xd6d1U,0xe981U,0x20dcU,0x26e7U,0x3631U,0x7d4cU,0x5ec0U,0x3228U,
		0xa2d9U,0x9b4eU,0xcc63U,0xe6feU,0x5bfcU,0x3680U,0x62c1U,0x10c3U,0xaafcU,0xf49cU,
		0x9d43U,0x8adeU,0x7a8bU,0x13fdU,0xc580U,0xb7f6U,0x007dU,0x7d00U,0x3a85U,0x55f7U,
		0x0b16U,0x1886U,0x0dbdU,0x9f77U,0x07f1U,0xba80U,0xf93eU,0x3ef9U,0x9ae1U,0xf2ceU,
		0x8affU,0xf514U,0xe5a7U,0x1c21U,0x7508U,0x00f8U,0xb227U,0x9ecdU,0xed9fU,0x4856U,
		0xa6b4U,0xd252U,0xae06U,0xc51aU,0xe17cU,0x8943U,0xc5fcU,0x796fU,0x5099U,0x7959U,
		0xb50fU,0xd665U,0xfa0aU,0x2ed4U,0x22e7U,0xb2e5U,0x467aU,0xed26U,0xe516U,0x1667U,
		0x7c01U,0xbd91U,0x843bU,0x15c7U,0xc8f8U,0x3d42U,0x82c2U,0xd98cU,0x79e5U,0xf9ccU,
		0xd173U,0x1d54U,0xa670U,0x8c26U,0x0c98U,0x2da4U,0xae88U,0xa53bU,0x4b20U,0xc31bU,
		0xdb76U,0x47e0U,0x42f1U,0x9405U,0x24b6U,0x55b5U,0x883fU,0x05d4U,0x8bd2U,0x5ae6U,
		0xe047U,0x76dbU,0x0c6dU,0x169cU,0x56deU,0x576aU,0x94ecU,0x9a98U,0xafa4U,0x5f1aU,
		0x3fbbU,0xced0U,0x32a3U,0x6eb8U,0x71ffU,0xa9abU,0xb2f4U,0xfe41U,0x14b6U,0x5ea1U,
		0x0330U,0x0722U,0x9795U,0x9509U,0x0526U,0x7bb4U,0x8161U,0x5817U,0x2795U,0xb075U,
		0xb5edU,0x21a3U,0x7166U,0x6e51U,0x4713U,0xe058U,0x0f5fU,0xed7bU,0x304fU,0x4484U,
		0x3e98U,0xfda5U,0x8784U,0x3443U,0xd225U,0x25ecU,0x0e13U,0x668fU,0xa3f9U,0xec03U,
		0xade6U,0x9698U,0x0c58U,0x6b7fU,0x807eU,0x7e80U,0x7f35U,0x3342U,0x60b1U,0x6188U,
		0xf23eU,0xcd29U,0xdd42U,0x3a71U,0xb96aU,0x6ab9U,0x5014U,0xd469U,0x46beU,0xcd46U,
		0x1f7fU,0xd8bdU,0x5be2U,0x8b5fU,0xf85fU,0xa7faU,0xf17fU,0x5e19U,0x9ea3U,0xaa7cU,
		0xf160U,0x962cU,0x911eU,0x1fbdU,0xd038U,0x2c51U,0xecc8U,0xcf5eU,0xd56aU,0xd7ffU,
		0x120fU,0x3f0fU,0xf8a4U,0x71f7U,0xf695U,0x8060U,0x5aceU,0x6820U,0x63f4U,0xab8eU,
		0xdc9aU,0xfc7bU,0x0b36U,0xd1cfU,0xedaeU,0x5b40U,0xfcaaU,0x9cf4U,0x4daeU,0xe169U,
		0x8a0dU,0x0948U,0xb018U,0xfed6U,0xe792U,0x92e7U,0x45ccU,0xf413U,0x960aU,0x0783U,
		0xca2bU,0x8cf2U,0xd115U,0xa774U,0xe6c4U,0xc4e6U,0x7796U,0x0cd7U,0x1895U,0x1f50U,
		0x0038U,0xce0eU,0x0875U,0xf800U,0x9e51U,0x4579U,0x0583U,0x91efU,0x0f7dU,0x1b4eU,
		0xa71cU,0xee7cU,0xb396U,0x081aU,0xa412U,0xf6c1U,0x99bfU,0x2b06U,0x325dU,0xeb4dU,
		0x2c22U,0x9531U,0x0d90U,0x82e1U,0x283cU,0x8737U,0x4cb6U,0x4007U,0x45f2U,0x427dU,
		0x98b6U,0x87b3U,0x2077U,0x27e5U,0x25dcU,0x49efU,0xc596U,0xdd2eU,0xd7b3U,0x323bU,
		0x0eeaU,0x4dcbU,0xa7c2U,0x4150U,0x8283U,0x90f7U,0x2e2fU,0x3ddeU,0x972aU,0xcdfaU,
		0xdb3dU,0x644dU,0x1572U,0x8796U,0x5c0bU,0x4d67U,0xd876U,0x175fU,0xfacdU,0x2a97U,
		0xb702U,0x1224U,0x5c45U,0x0297U,0x39baU,0x7195U,0x8bf7U,0xc2f7U,0x4ba4U,0xd84bU,
		0x6a1dU,0xa130U,0x9059U,0x2766U,0xc2e0U,0x397eU,0xb98dU,0x19c4U,0x58cfU,0x0395U,
		0x769eU,0xc43eU,0x3d20U,0xcd78U,0xc179U,0xee34U,0x480fU,0x5e33U,0x2526U,0xf980U,
		0x5657U,0xeb5dU,0x603eU,0x4989U,0x5498U,0x019cU,0x5a1fU,0xdf2dU,0xc856U,0x3f52U,
		0x271dU,0x77dcU,0x9e04U,0x1a48U,0xd2d5U,0x6ee5U,0x3b56U,0x6742U,0x7b9dU,0x5bacU,
		0x1c97U,0x5670U,0xcfd2U,0x3f1dU,0x360cU,0x3cccU,0xa247U,0x4d7cU,0x3a13U,0xdab1U,
		0x937dU,0x8ea2U,0xd6d8U,0xa8f4U,0xe69bU,0x2752U,0x898bU,0xf6fbU,0xb772U,0xe96aU,
		0x2dd6U,0xbdceU,0x4751U,0xc9beU,0x13a9U,0xf34dU,0xa9d2U,0xe357U,0x4215U,0xedfeU,
		0x0b3eU,0xf86dU,0x9f68U,0x0432U,0xaedbU,0x8217U,0x6d57U,0xab36U,0xc62dU,0x97ccU,
		0x2d8bU,0x5dbaU,0x822cU,0x4193U,0x4465U,0x747fU,0x237bU,0x02f8U,0x4d94U,0x161bU,
		0x7e09U,0x8401U,0x52c3U,0x4cf8U,0x2c97U,0x083cU,0x60f2U,0x8178U,0x1d45U,0x26fbU,
		0xc814U,0x24bfU,0x51e9U,0xadb1U,0xc6f2U,0x8d71U,0x0041U,0x9cb9U,0x6eb1U,0xee75U,
		0x7ed8U,0x10dfU,0xe072U,0x17e4U,0x4ea4U,0xcd37U,0xa281U,0x699aU,0xedf7U,0x1317U,
		0x6f53U,0x4a93U,0xd580U,0x40d4U,0x59c1U,0x725cU,0xe653U,0x24f2U,0x52e7U,0x07deU,
		0x5a8aU,0x9781U,0x3ff8U,0x83a0U,0xfab5U,0xc4acU,0xbc39U,0xe141U,0x284fU,0xc102U,
		0x4e29U,0x727eU,0xede4U,0x2befU,0xa85dU,0x15f4U,0xd705U,0x649cU,0x57f5U,0xcd69U,
		0xe48bU,0xfa84U,0x2336U,0xaa22U,0xd35cU,0x3151U,0x6163U,0x1aeaU,0x3769U,0xd233U,
		0x6f71U,0xbe7aU,0x2ea3U,0xa4b4U,0x6226U,0x7e78U,0xf5c4U,0xb2acU,0xf189U,0xbb77U,
		0x41ffU,0x20c8U,0xf4b3U,0xc74fU,0x10d7U,0xb2feU,0x7aeaU,0x5b19U,0x3d2aU,0xb635U,
		0xe988U,0x9245U,0x7926U,0x720dU,0x7689U,0xec75U,0x2ac2U,0xf5d7U,0x2858U,0x0d9aU,
		0x2a85U,0xe1f1U,0xfddcU,0x9c93U,0x048cU,0x1a64U,0x40c4U,0x131eU,0xa08eU,0x5027U,
		0x2871U,0x2f9cU,0xadc3U,0x89e1U,0x740aU,0xde25U,0xa996U,0x6507U,0x3702U,0x8c5eU,
		0xe440U,0x6600U,0xda64U,0xea53U,0x48e2U,0xc1ebU,0xce20U,0xcc6dU,0xe0a3U,0x21b1U,
		0x66e9U,0x193fU,0x8445U,0xa632U,0x4f31U,0x31d2U,0x499eU,0x835cU,0x8f0fU,0x2954U,
		0xa036U,0x25b8U,0x212eU,0x0d35U,0x65b9U,0xc95aU,0x7155U,0xf723U,0x372eU,0x0ca6U,
		0x3ba9U,0x0249U,0xe4e7U,0x9447U,0xda30U,0xa956U,0x2d07U,0x5284U,0xf70cU,0xa816U,
		0xe6e9U,0x645eU,0xc656U,0xba9eU,0x6b59U,0x7db5U,0x88b2U,0x504aU,0xb62cU,0x76e9U,
		0x73ebU,0x3ab7U,0x3921U,0xd8e5U,0xe306U,0x3303U,0xb189U,0xea63U,0x586cU,0x3651U,
		0x0542U,0x50c2U,0x9f0bU,0x7223U,0x18b7U,0x5f32U,0x38e9U,0xbee3U,0x639eU,0x98d5U,
		0x9667U,0xe0f8U,0x2fdfU,0x30e9U,0x0be0U,0x8decU,0x6975U,0xa909U,0x8d44U,0x2474U,
		0x7e3aU,0x9fa5U,0xe0c3U,0xacf3U,0x7998U,0x6c18U,0x82a7U,0xab99U,0x7f06U,0xa52bU,
		0x27cbU,0x32f6U,0xcd51U,0x0499U,0x772bU,0x96aeU,0x9af8U,0x448fU,0x20e3U,0x51beU,
		0xb563U,0x7426U,0x8a41U,0x6349U,0x27d5U,0x3101U,0x5f25U,0x660dU,0x671cU,0xdb4bU,
		0xa368U,0x638aU,0x2aa2U,0x6233U,0xe992U,0x6f5aU,0xd1d7U,0x75caU,0x81eaU,0x1c03U,
		0xe7abU,0xf716U,0xdef2U,0xf98bU,0x2e19U,0xdf4cU,0xd901U,0x076aU,0x201bU,0xc305U,
		0x4accU,0xab5fU,0x2de0U,0xff4dU,0x1f5bU,0x9252U,0x7da8U,0x2eb3U,0xf2beU,0x4edeU,
		0x3a35U,0x4288U,0xe67eU,0x52f1U,0xddcaU,0xa5abU,0x28daU,0x98e2U,0xce84U,0x76acU,
		0x44eaU,0x7cd3U,0x93e5U,0x197bU,0xbb95U,0x4b57U,0xaf92U,0xb306U,0x51b4U,0x31c5U,
		0xb758U,0xf73aU,0x7039U,0x0bf2U,0xace0U,0x392eU,0xa973U,0x03efU,0x1b06U,0x1e66U,
		0x9d09U,0xed58U,0x16ebU,0x5ee7U,0x89d4U,0xf9a6U,0x72aaU,0xa68aU,0xdfadU,0x2397U,
		0x545aU,0x1f38U,0x04baU,0xe156U,0xd50fU,0xdf3bU,0x2548U,0x83edU,0xf07aU,0x9238U,
		0xd409U,0xa823U,0xac90U,0x3fdbU,0x6526U,0x9ed8U,0x53ccU,0x8c74U,0x8dc0U,0xa317U,
		0x10ceU,0x213bU,0xe122U,0x5fdeU,0xb29fU,0xe5daU,0x3f8cU,0x2311U,0xe528U,0x91fdU,
		0x7721U,0x4a86U,0xa7a2U,0xaabaU,0x4f8cU,0xbddbU,0x4944U,0xb645U,0x7b2fU,0x54d7U,
		0xad78U,0x7759U,0x595eU,0x7f22U,0x7a47U,0x3619U,0x26eeU,0xa3f1U,0x5cbcU,0xc49aU,
		0xf654U,0x4a04U,0xf481U,0xa1b1U,0x35e4U,0xc614U,0xc6eaU,0x44e2U,0xdf04U,0x96d3U,
		0x2fc2U,0x1addU,0x3081U,0x4a9aU,0x0fb6U,0x9520U,0x65d7U,0x4172U,0xe55eU,0x641eU,
		0x2507U,0x840eU,0xe604U,0xca39U,0x9499U,0x7b62U,0x2be2U,0xc27dU,0x4bffU,0x03c3U,
		0x7cefU,0x543fU,0x1ca8U,0xc67aU,0xda7aU,0x8c08U,0x10fcU,0x84a5U,0xc199U,0xaee8U,
		0xb242U,0xff6cU,0x14ebU,0x6954U,0xdd58U,0x9c48U,0xfc49U,0x6decU,0x6c0eU,0xdaa4U,
		0x4bbfU,0xe07fU,0x137cU,0x4e5bU,0x3b73U,0x7530U,0xc1acU,0xe4abU,0xeaaaU,0x8a9eU,
		0xe2a0U,0xb0ceU,0x5ceaU,0xfbc0U,0xe95dU,0xe80aU,0x8bccU,0xdd3cU,0x7790U,0xb6fcU,
		0x6773U,0xc349U,0xcd90U,0x07ebU,0xb55dU,0xacdaU,0x5cb6U,0x5ce4U,0xe695U,0xedf1U,
		0x7920U,0x2d01U,0xf602U,0x9141U,0xd1fdU,0x1195U,0x12f7U,0xaf2eU,0xe913U,0x7acdU,
		0xbb6eU,0x085eU,0xf024U,0xd240U,0x02ebU,0x2c30U,0xc420U,0xd3f5U,0x16d3U,0xf670U,
		0x064dU,0x08b2U,0x0eabU,0xc7abU,0xe8baU,0x1091U,0xac7cU,0xc354U,0xad16U,0x2e87U,
		0x010bU,0xf84eU,0xd021U,0xdc00U,0x8f31U,0xfb9bU,0x109cU,0x1fecU,0xee9cU,0xc54dU,
		0xfb36U,0x842dU,0x2606U,0xfcf3U,0xb47cU,0x0f20U,0x12fdU,0xff96U,0x0309U,0xa1bcU,
		0x690eU,0xc44fU,0xec4fU,0xf9f3U,0x84b0U,0x3c36U,0xf930U,0x62e9U,0xb5ccU,0x17a5U,
		0x3f66U,0x4eabU,0xedb6U,0xc08fU,0xa322U,0xc6b2U,0x2d3fU,0xe108U,0x869fU,0x2f40U,
		0xb9b9U,0x3939U,0x755fU,0x53a8U,0xab6aU,0x496cU,0x01c0U,0x7076U,0x0ba8U,0x107fU,
		0xb4b4U,0xf3f3U,0xad06U,0x5d49U,0xf070U,0x449aU,0xf26dU,0xb87aU,0xb035U,0xf123U,
		0x84d4U,0xbab4U,0xfbf2U,0x55beU,0xde30U,0xef5cU,0xdf0cU,0x0a04U,0xce17U,0x3894U,
		0xc78cU,0x6941U,0x1923U,0x7af2U,0x6024U,0x20d3U,0xd9abU,0x300aU,0x722eU,0x5b2eU,
		0xaf34U,0x8f6fU,0x3036U,0xbf5eU,0xc8daU,0x0cb1U,0x0446U,0x3d6dU,0x47acU,0xbffaU,
		0x3071U,0x4731U,0x3531U,0x29d9U,0x33b0U,0x6ef0U,0xbe34U,0x5f49U,0x4ffcU,0x1cf9U,
		0xee09U,0xd2abU,0x14afU,0x99b8U,0xa4e7U,0x57e5U,0x1fd2U,0x6dd1U,0x2d38U,0x18daU,
		0xa83cU,0xf4faU,0xaed4U,0xadbcU,0xe7a4U,0xe557U,0x3430U,0x2c72U,0x8ee2U,0x5b6aU,
		0x8effU,0x89a5U,0xe14cU,0x4d18U,0x491cU,0x783dU,0xef9dU,0xea87U,0x72edU,0xd94fU,
		0x909dU,0x0867U,0x6fa6U,0x970cU,0x7b8bU,0xf3d5U,0x2144U,0x9d1aU,0x9759U,0x3ed4U,
		0x95f7U,0x76f2U,0x6081U,0x9427U,0xe919U,0x9bb4U,0x98a8U,0x69aeU,0x4be6U,0xe316U,
		0xc10bU,0xd76dU,0x041eU,0x964bU,0xc200U,0xbf04U,0x5905U,0x2252U,0x2684U,0xcd3eU,
		0x6ad6U,0xa7b4U,0xffd8U,0xf0deU,0x18e1U,0x3e68U,0x5410U,0xf0a9U,0x7254U,0xed50U,
		0x3d65U,0x167fU,0x6e8eU,0x655cU,0xe118U,0x683eU,0x4801U,0x5935U,0x389dU,0xa16eU,
		0xfe93U,0xeb8cU,0xba4aU,0x4783U,0xe844U,0xd828U,0x02caU,0x78a1U,0xa629U,0xd81fU,
		0x749eU,0x0322U,0xd77fU,0x5f54U,0x71e1U,0x9b45U,0x3b28U,0xb54cU,0x6d1dU,0x981eU,
		0x43dfU,0x86d3U,0x0a0fU,0xe1cdU,0x79afU,0x220eU,0x0c8aU,0x17b0U,0x7ad3U,0x4c97U,
		0xd27bU,0x0977U,0x2e92U,0x00b2U,0x6130U,0xd79dU,0xb6cdU,0x4314U,0x336aU,0xc2b0U,
		0xd7e2U,0xf896U,0xe1d8U,0xd198U,0x95daU,0x95b7U,0xf0c6U,0x12c8U,0x5858U,0xdbdbU,
		0x54b0U,0x183cU,0xf779U,0xdbd0U,0xb8bdU,0x8d06U,0x5bd4U,0xe89fU,0x6262U,0xb5b5U,
		0xf3e0U,0x11a8U,0x3fd1U,0xa8b7U,0xa876U,0x2a39U,0x84c3U,0xa3d9U,0x892dU,0x556dU,
		0x5e79U,0x3e73U,0x0deaU,0x8a5cU,0x9395U,0x0464U,0x8371U,0x8199U,0x9b9dU,0x3cbdU,
		0x423eU,0x1fdfU,0xf8c9U,0x4b37U,0x22c2U,0x1953U,0x30d1U,0xb1e9U,0x8884U,0xd745U,
		0xde6cU,0x4911U,0xf608U,0xa591U,0x698bU,0x387bU,0x3c29U,0x3cf4U,0x3788U,0xb1d2U,
		0xdc2bU,0x68bbU,0x4098U,0x4d25U,0x184dU,0x4ce1U,0x75e8U,0xa288U,0x5b48U,0x1930U,
		0xa608U,0x19f8U,0xa843U,0x07c0U,0xddc0U,0xc6a5U,0x5d33U,0xd38fU,0x4decU,0x2ef9U,
		0xf399U,0xf3caU,0x43a8U,0xc007U,0x3aa8U,0xa2e7U,0xa209U,0x346bU,0xa21eU,0xfe56U,
		0x495fU,0x34beU,0x1ac6U,0xa6d8U,0x06afU,0xedc8U,0xc52cU,0x0feeU,0x266fU,0x83deU,
		0xa3b9U,0x14d5U,0x96dbU,0x44a7U,0x4fd1U,0x5294U,0x3118U,0xfb89U,0x48fdU,0x5b21U,
		0x0d6dU,0xb20dU,0xc38cU,0x86acU,0xab50U,0x03a8U,0x9147U,0xc3eaU,0x6641U,0x989dU,
		0x0483U,0x58fcU,0xc948U,0x1687U,0xef4fU,0x743fU,0x4ef0U,0xec97U,0x8fc0U,0xb6edU,
		0xec5fU,0xe587U,0x280bU,0x85e2U,0xea35U,0x9a80U,0x4040U,0x6363U,0x3f41U,0xc45cU,
		0x56f4U,0xea1cU,0x4b4dU,0x4d0bU,0x4a77U,0x5d11U,0x4d4dU,0x7676U,0x8637U,0x9640U,
		0x0b4fU,0x6d72U,0x6dbcU,0x5a30U,0x23f1U,0x35b0U,0x5745U,0x8f3eU,0xd29eU,0x59abU,
		0xcefcU,0x4776U,0x3bf5U,0x4ca4U,0x1ba7U,0x7c0fU,0x1ddbU,0xa430U,0xb494U,0x29adU,
		0xbb49U,0x919eU,0x0671U,0x6e63U,0x5262U,0x980fU,0xed43U,0x7599U,0xa0b0U,0x3663U,
		0xd203U,0xd78cU,0x2781U,0x0692U,0x203eU,0x7d8fU,0x78ceU,0x7249U,0xcc29U,0xb16aU,
		0x9418U,0x9577U,0x43faU,0x2dbaU,0xa969U,0x69a1U,0x96f8U,0xe2d7U,0x5fcdU,0x8cd0U,
		0xf501U,0x8951U,0xf63fU,0x37bcU,0x2626U,0x5b5bU,0x80faU,0xa8c4U,0x0a36U,0x113aU,
		0x7830U,0x60a9U,0xbcfbU,0xe86dU,0x8585U,0xbdbdU,0x05e8U,0xf0d3U,0xb8d4U,0x1bb8U,
		0x4e9cU,0xb750U,0xd9a3U,0xc384U,0x025dU,0x70abU,0xefd6U,0x91cbU,0x46e0U,0xd18bU,
		0x3293U,0x4f6cU,0x40f5U,0x8e01U,0x74c9U,0x37adU,0x8fadU,0x4929U,0x2566U,0xd5afU,
		0xde1cU,0xf0a0U,0x1ffcU,0x41d0U,0x119bU,0x5511U,0xabfaU,0xe6b4U,0x4fb4U,0x0219U,
		0xa10aU,0x76ffU,0x26caU,0x8b0dU,0xa232U,0xba3fU,0x1c3cU,0x2691U,0x1a34U,0xf6e2U,
		0x8347U,0x4abaU,0x5eddU,0xe1c0U,0x31a3U,0xb1c3U,0xc927U,0x2977U,0xa500U,0x4b68U,
		0xe33bU,0xffb4U,0xabdfU,0xb804U,0x9abfU,0xa119U,0x9aadU,0x22d7U,0x00a5U,0x684bU,
		0x83cfU,0xa046U,0xe945U,0x732dU,0xe412U,0x6551U,0x5222U,0x0559U,0x0803U,0xfc12U,
		0xcce6U,0x3540U,0x99f0U,0xa586U,0x7d61U,0x18f6U,0xd610U,0xe97aU,0x219cU,0x4276U,
		0x413cU,0xb2c3U,0xc245U,0xbfbcU,0x1078U,0x592eU,0xa2e0U,0xe866U,0x02f1U,0xc1e4U,
		0xff46U,0x5438U,0x4f23U,0x2cbaU,0x653fU,0xcbb5U,0x515bU,0xa7deU,0x7f18U,0x4622U,
		0xbbf9U,0xf8e2U,0xd28bU,0xe65aU,0x63fcU,0x6798U,0x079fU,0x408dU,0x8f06U,0x7495U,
		0xb52dU,0x7147U,0xd719U,0xfd27U,0x6f2dU,0x62b8U,0xe476U,0x5783U,0xa444U,0x0ab2U,
		0x8e6bU,0x5162U,0xc825U,0xcaa2U,0xf142U,0x0594U,0xe2f8U,0xf9bbU,0xb051U,0xd089U,
		0x341cU,0xdad7U,0x8326U,0x0459U,0x745dU,0x8848U,0xe8feU,0x88caU,0xf489U,0xbd66U,
		0x087eU,0xc252U,0x5921U,0x7fb6U,0x9283U,0x4207U,0xca4cU,0x3db1U,0x05f0U,0x46f6U,
		0x46a2U,0x123aU,0x2c36U,0x9b28U,0xcf85U,0xb911U,0xb374U,0x8c9fU,0xa798U,0x18e9U,
		0x3a05U,0xef12U,0x1b81U,0x462fU,0x9d2bU,0x2b9dU,0x357fU,0x4233U,0xb4f0U,0xd726U,
		0x57bfU,0x7009U,0xd07cU,0xc231U,0xe4ceU,0x1f9bU,0xcab8U,0xdd1cU,0x5f7fU,0x6155U,
		0x80ccU,0x8d22U,0x014aU,0xd096U,0xb40eU,0xad9eU,0x355bU,0x45aeU,0x42ddU,0x713aU,
		0xdde9U,0xe9ddU,0xb26aU,0x9002U,0x6f45U,0xa1ceU,0xa6ffU,0xf016U,0x158aU,0x76bfU,
		0xbfddU,0x42a8U,0x5e23U,0xf02dU,0x11d4U,0x04eaU,0x8b25U,0x583cU,0xe12bU,0xfe4bU,
		0x3c9bU,0x2c0dU,0xcd82U,0x73f4U,0x9b55U,0xabb6U,0xc426U,0x23ccU,0xa1a7U,0x8b4fU,
		0x9965U,0x39ceU,0x4e32U,0x82ebU,0x022fU,0x41ecU,0x9740U,0x2adcU,0xc951U,0xb543U,
		0x2ee5U,0x24f9U,0x0fd0U,0xd00fU,0xcc45U,0x13f4U,0x89acU,0x67bbU,0x64d5U,0x2005U,
		0x611bU,0xdf5dU,0xde8aU,0x439dU,0xf594U,0x59fcU,0x4f7eU,0xfc01U,0x211cU,0xa7e5U,
		0xc795U,0xe708U,0xaca7U,0xcfe2U,0x85baU,0xe274U,0x15d1U,0x74a7U,0xbbd3U,0xd3bbU,
		0xd48fU,0x15b2U,0xb536U,0xa7f2U,0x1794U,0x28ccU,0xd550U,0x7d18U,0x2b95U,0x2eeeU,
		0x6096U,0x2f23U,0x9779U,0xa682U,0x29d1U,0xb6e5U,0x93b1U,0xe21dU,0x26e4U,0xe6fbU,
		0xd3fbU,0x0fddU,0x1a20U,0xcb4aU,0xc517U,0xd662U,0x1664U,0xd989U,0x8adbU,0x55f4U,
		0x3ef6U,0x00f5U,0xfe3eU,0x7bb1U,0x6e4eU,0xfda2U,0x2da1U,0x9402U,0x76d8U,0x5f17U,
		0x9629U,0xd7fcU,0x681dU,0x5b3dU,0xec00U,0x333fU,0x6ab6U,0x8b5cU,0xf6beU,0x82deU,
		0x87b0U,0x3238U,0xfed3U,0x8cefU,0x1f4dU,0x91ecU,0x0392U,0x5e30U,0x0199U,0x1a45U,
		0x3ddbU,0x4d64U,0x0294U,0xa12dU,0x042fU,0x5db7U,0x1618U,0x8175U,0x566dU,0xdaaeU,
		0xf6f8U,0xf34aU,0xe13eU,0x15f1U,0xaa1fU,0xbe77U,0x8d6eU,0x17e1U,0x4a90U,0x07dbU,
		0x2f99U,0x8c5bU,0xcc6aU,0x31cfU,0xbb74U,0x5b16U,0xec72U,0x9c90U,0x3300U,0x7220U,
		0xe0f5U,0x2471U,0x0d32U,0x0246U,0xa813U,0x5047U,0x1c00U,0x0767U,0x924fU,0x52eeU,
		0xab96U,0x96abU,0x6346U,0x6387U,0xe153U,0xa820U,0xa314U,0x230eU,0x7cd0U,0x31c2U,
		0x03ecU,0xf9a3U,0x416fU,0x7b5fU,0xc677U,0xff69U,0xbdd8U,0x7f1fU,0x4a01U,0x96d0U,
		0x2cfeU,0x7acaU,0xd3f2U,0x108eU,0xdaa1U,0xe4a8U,0xe807U,0x07e8U,0xc6afU,0x53a5U,
		0xf3f0U,0xf120U,0xdbfdU,0x842aU,0xa1b9U,0x62e6U,0x1cf6U,0x6dceU,0xe554U,0x4d15U,
		0x0a01U,0x20d0U,0xbf5bU,0x472eU,0xf0dbU,0x167cU,0xa16bU,0x789eU,0x0864U,0x3ed1U,
		0x69abU,0xbf01U,0x12c5U,0x8d03U,0xa8b4U,0x3e70U,0x9b42U,0xe1caU,0x0974U,0xc2adU,
		0x19f5U,0x2ef6U,0x3468U,0xedc5U,0x3cbaU,0xb1e6U,0x3878U,0x4d22U,0x9a7dU,0x4d08U,
		0x6d6fU,0x59a8U,0x44a4U,0xb20aU,0x989aU,0xec94U,0x8ccdU,0xa8c1U,0xbdbaU,0xc381U,
		0xa42dU,0x980cU,0x068fU,0x9574U,0xe1bdU,0xffb1U,0x6848U,0x0556U,0x4f69U,0xd5acU,
		0xe6b1U,0xba3cU,0x408aU,0x62b5U,0xca9fU,0xdad4U,0x18f3U,0xbfb9U,0x5435U,0x461fU,
		0x7006U,0x6152U,0x45abU,0xa1cbU,0xbd63U,0x3daeU,0xb90eU,0x462cU,0x13f1U,0x439aU,
		0xe705U,0xd3b8U,0xf02aU,0x2c0aU,0x8b4cU,0x2ad9U,0x5f14U,0x8b59U,0x91e9U,0xa12aU,
		0x7d15U,0xb6e2U,0xcb47U,0x00f2U,0x6384U,0xf9a0U,0x96cdU,0x07e5U,0xf347U,0x07d8U,
		0x9c8dU,0x5044U,0x4d1fU,0xec91U,0x9571U,0xba39U,0x62e3U,0x472bU,0xbefeU,0xc2aaU,
		0x5041U,0xc2a7U,0x00ecU,0x3f33U,0x461cU,0x4629U,0x2ad6U,0x00efU,0x7bf2U,0xe00fU,
		0xa0f9U,0x8463U,0x6a4eU,0x11fdU,0xc99dU,0x3f36U,0x8030U,0xc12fU,0xb238U,0xc993U,
		0x80ecU,0xb503U,0x432eU,0x6a51U,0x1da2U,0x5e49U,0xf387U,0x5238U,0xe629U,0x3a46U,
		0x0e27U,0x1200U,0x8ba8U,0x5889U,0xc122U,0xd030U,0xd246U,0x3be7U,0x1641U,0xc9a0U,
		0xe4c6U,0x6d67U,0x53f6U,0xbb5cU,0x4e21U,0x2540U,0x14a7U,0x3f39U,0x2d57U,0x572dU,
		0x853aU,0xf755U,0xc70eU,0xc68cU,0x0097U,0x7bf5U,0x1b61U,0x5ddfU,0x69e1U,0xae4dU,
		0xebe2U,0x0edbU,0xaf7eU,0xe012U,0x0eceU,0x3800U,0x95c7U,0x08e7U,0xa5dcU,0x249fU,
		0xb2d9U,0xa0fcU,0xf107U,0x80baU,0x3a57U,0x573aU,0x0029U,0x26abU,0x6afeU,0x8466U,
		0xfe84U,0x9f4dU,0x9042U,0x9c6aU,0x8d97U,0x1829U,0x28fbU,0x80efU,0x9f02U,0xb8b4U,
		0x6aadU,0xe546U,0x7f87U,0x5af8U,0x6c4bU,0xb506U,0xa56cU,0xd7d3U,0x3638U,0xc96cU,
		0x0eb1U,0xca77U,0x0406U,0x4331U,0x6cefU,0x7025U,0x386aU,0x1f73U,0x2d6cU,0xc5eaU,
		0xafb7U,0x6a54U,0x4346U,0x6bdbU,0x8f22U,0xd587U,0xdca2U,0xe2ccU,0x0a93U,0x8033U,
		0x3437U,0xcc5cU,0x3e91U,0xd7acU,0xfebbU,0xa2faU,0x4559U,0xc132U,0xb02eU,0x02c3U,
		0x06a8U,0xd99cU,0x60ebU,0x4f2aU,0xe677U,0xb23bU,0xe7f9U,0x0c4cU,0x2b53U,0x026eU,
		0x3415U,0xbbccU,0xe800U,0xc996U,0xcad5U,0xa2b1U,0x809cU,0x05ddU,0x4fa5U,0xdc58U,
		0x35ecU,0xe62cU,0xf809U,0x8273U,0x5aa0U,0x33dcU,0xfe07U,0xf4bbU,0xfc6fU,0x3a49U,
		0xb8dcU,0x9832U,0xea97U,0xc401U,0xc7b1U,0xceedU,0xeb0cU,0x0e2aU,0xf685U,0x69cfU,
		0x6b0fU,0x5418U,0x7c57U,0xf559U,0x1633U,0x1203U,0x8908U,0x609eU,0x9f88U,0x1cd2U,
		0xfb4bU,0x7d30U,0x4031U,0x1da5U,0xfc30U,0xae2eU,0x53feU,0xc7f3U,0xcf52U,0x96bfU,
		0x553fU,0x5e4cU,0xc0b1U,0x8e26U,0xac0cU,0x6d2cU,0xdaf7U,0x1ebeU,0x90d2U,0xf38aU,
		0x9fdeU,0x7ee8U,0x6769U,0x4e17U,0xb900U,0xf154U,0xd4d6U,0x523bU,0xd8dcU,0x95aeU,
		0x93f9U,0xb51fU,0xca6aU,0x5b98U,0xe37fU,0xd249U,0xd34dU,0x4566U,0x0d15U,0x3c86U,
		0x5fe7U,0xeeb9U,0x1e9dU,0x3beaU,0x1014U,0x3ac0U,0x7d53U,0xa664U,0xe8c0U,0x9c5dU,
		0x5e01U,0x1644U,0xbcc4U,0xa9c3U,0x7287U,0x2a48U,0x78b5U,0x0b6eU,0x12aeU,0xc9a3U,
		0x1690U,0x013cU,0x55a3U,0x40dbU,0xcf3bU,0x7f94U,0x0c3fU,0x8babU,0x2c79U,0x0bbdU,
		0x65aaU,0x41d9U,0x87cdU,0x29a2U,0x1d72U,0x588cU,0x1b73U,0x8833U,0xc57bU,0xc5f7U,
		0x465eU,0x2cc1U,0x694aU,0xc125U,0x45c7U,0x3258U,0xdb38U,0x3d1bU,0x4b1bU,0x032bU,
		0x0c53U,0xd033U,0xf4aeU,0x7405U,0x7150U,0x5867U,0x3607U,0x6d52U,0x6eacU,0x4e24U,
		0x7b2aU,0x2502U,0x1377U,0xd1f8U,0x27c6U,0xdeedU,0xbb90U,0x2543U,0x7b86U,0x540bU,
		0x6d18U,0x54abU,0x1097U,0x869aU,0xc787U,0x14aaU,0xbb44U,0xf63aU,0x74c4U,0xc922U,
		0xf8c4U,0xddbbU,0x3113U,0x3f3cU,0x8b20U,0x64d0U,0x6091U,0x6818U,0x2197U,0xb528U,
		0x591cU,0xe4c9U,0x45a6U,0x956cU,0xf382U,0x69dcU,0xaa1aU,0xa30fU,0xe54fU,0x6d6aU,
		0x0d10U,0x65a5U,0x1372U,0xf37dU,0x6aa8U,0x3e8cU,0x5a9bU,0x53f9U,0xd0d3U,0xc5a7U,
		0xcdaaU,0x55dbU,0xaf11U,0xf816U,0x0ac5U,0xbb5fU,0xfb1bU,0x2a04U,0x62d2U,0x32d1U,
		0x34e6U,0xcc17U,0xbf72U,0xc711U,0xeb2fU,0x710dU,0xaf16U,0xc16dU,0x464eU,0xa379U,
		0xb89bU,0xc68fU,0xf965U,0x8c41U,0x5ec7U,0x9768U,0x16d9U,0x1125U,0x5384U,0x009aU,
		0x2fd6U,0xa86dU,0x0389U,0xd0caU,0x38b8U,0x8d87U,0x757dU,0x7bf8U,0x5357U,0xe3feU,
		0xdcc3U,0x59c8U,0xfc83U,0x65e9U,0x0e5dU,0x2d5aU,0x8ee9U,0xc84aU,0xf81bU,0xf1bbU,
		0x3402U,0x59e4U,0x08a8U,0x5730U,0x9da1U,0x7f76U,0x2262U,0x84e9U,0xdd70U,0x6546U,
		0x8ac9U,0x853dU,0xf974U,0x31e9U,0xe023U,0x6a7dU,0xb72dU,0x4e89U,0x2b5aU,0xf758U,
		0x0cf0U,0xaaa4U,0xd8a8U,0x73a2U,0xeb9eU,0xfe75U,0xc61cU,0xebe5U,0xb3feU,0xfa33U,
		0x0acaU,0xbf13U,0x738bU,0xc757U,0x5376U,0x0edeU,0x1bc0U,0xf623U,0x81c0U,0x1c31U,
		0xf676U,0xf3b6U,0x88eeU,0xaf81U,0x885eU,0x1784U,0x20a5U,0xf0b1U,0x93cfU,0xdbb0U,
		0x6441U,0xe015U,0x2a13U,0x2f2bU,0x382fU,0x9dc6U,0xf768U,0x0a4cU,0xda1aU,0x1b64U,
		0x8615U,0x519bU,0xbb64U,0x0ec4U,0x5368U,0x5c39U,0xa740U,0x5de2U,0xbd42U,0x296cU,
		0x4b92U,0x73c6U,0x0e44U,0x0660U,0x4597U,0x69e4U,0xd100U,0x5a56U,0xc560U,0xf337U,
		0xd83fU,0x3821U,0x8255U,0xae50U,0xe1a4U,0xb3b9U,0x6e9eU,0x45e6U,0x3a98U,0x9db9U,
		0xf03fU,0xa5dfU,0x2271U,0x6375U,0xd0d8U,0xfa40U,0xbbacU,0x56bcU,0xdb22U,0x24a2U,
		0xbedaU,0xb5acU,0x322fU,0xf374U,0x0653U,0x1b97U,0x64adU,0xb2dcU,0xf0f8U,0x9542U,
		0xfaebU,0x1561U,0x597fU,0x886bU,0x1064U,0xa0ffU,0x12a0U,0xa710U,0xb6b3U,0x7263U,
		0x1d23U,0xf314U,0xaf24U,0x0ed1U,0x5b71U,0xbe8aU,0xc5acU,0x2f8aU,0x1ed7U,0xdeb1U,
		0x30a9U,0x3803U,0x2afaU,0xc56dU,0x6b31U,0xe28bU,0x5341U,0xcbbcU,0x1777U,0x95caU,
		0x20f6U,0x8117U,0x1ce7U,0xccb2U,0xff25U,0xf095U,0x0275U,0x08eaU,0x02dbU,0x6a70U,
		0x776eU,0x3d8eU,0x6592U,0x38c5U,0x90c4U,0x002cU,0xd9f4U,0x4fe8U,0xcdafU,0x4ebfU,
		0xe3d1U,0xd480U,0x9351U,0x26aeU,0x59d7U,0xce74U,0x31f7U,0xa386U,0x08b8U,0x5cfeU,
		0xc4daU,0x6b01U,0x620cU,0xcb2fU,0xd968U,0x8fa2U,0x9956U,0xd64cU,0x7a27U,0x8469U,
		0xb58eU,0xa922U,0x7916U,0x008dU,0x454bU,0xca1fU,0xac40U,0xf10aU,0x2348U,0x2b67U,
		0x55e0U,0x1458U,0x9926U,0xbea0U,0x5064U,0x80bdU,0x3bc4U,0x6327U,0x29f7U,0x0b2eU,
		0x093cU,0x9034U,0x326fU,0x3a5aU,0xb262U,0x4f61U,0xb8f8U,0xd837U,0x3761U,0xb297U,
		0x8ef7U,0x573dU,0xc1c8U,0x8caeU,0x5baeU,0x753eU,0xbad3U,0xa674U,0x7b9fU,0xbacfU,
		0x3c6fU,0xcdd4U,0x5672U,0xc8bdU,0xcbe2U,0x3280U,0x1c99U,0xf6a4U,0x1cc9U,0x422aU,
		0x6ee7U,0x8397U,0x7ba3U,0x8c2aU,0xd2d7U,0x6bc9U,0xe5eeU,0xc149U,0x6744U,0x6cbdU,
		0x35d9U,0x4364U,0x3b58U,0xe904U,0x126eU,0x1166U,0x4d7eU,0xa095U,0xa678U,0x0c9cU,
		0xa249U,0x23beU,0x72f4U,0xd6b3U,0xdab3U,0xfb6bU,0x8298U,0x5526U,0x3a15U,0x6781U,
		0x1c7eU,0xece3U,0x3f1fU,0x0747U,0xbad7U,0x2da8U,0xcfd4U,0xbe0aU,0xdcf1U,0x8006U,
		0x3cceU,0x9219U,0xc2cbU,0x89c5U,0x360eU,0x8dd6U,0x1742U,0x8ff4U,0x2754U,0x979eU,
		0x7542U,0x79e9U,0xe69dU,0x7853U,0xb39fU,0x9e64U,0xf6fdU,0xd599U,0xbf3aU,0x63a6U,
		0x898dU,0x6f11U,0x46e8U,0x91b6U,0x8ea4U,0x1963U,0x5bb2U,0xf9d0U,0x937fU,0xaa8bU,
		0xad6dU,0x088dU,0xa8f6U,0x7adbU,0xfd3eU,0x4c67U,0xd6daU,0x88d0U,0xd546U,0x2124U,
		0xc9c0U,0x678aU,0x8cb2U,0xd177U,0x4753U,0xb858U,0xbc2aU,0xe3e6U,0xf34fU,0x30bcU,
		0x4bd0U,0x9a4bU,0x13abU,0x8131U,0xb5e5U,0x8f59U,0xe96cU,0x7ff6U,0xc1ccU,0x1d58U,
		0xb774U,0xe8c8U,0xdd68U,0x60e3U,0xbdd0U,0x8e63U,0x3aa0U,0x0ba0U,0x2dd8U,0x2373U,
		0xd3daU,0xb462U,0xf86fU,0x52aeU,0xf6a8U,0xdb7aU,0x0b40U,0x3b7cU,0x6ca8U,0xc9d0U,
		0x0434U,0x2890U,0xce8dU,0x471dU,0x9f6aU,0xcafeU,0x6048U,0x6b62U,0xe359U,0x2f57U,
		0x1c9dU,0x47e4U,0xa9d4U,0x4dddU,0xfc27U,0xd073U,0xee00U,0x2846U,0x62a7U,0x6c33U,
		0x4217U,0xf48fU,0x2814U,0xee1aU,0x97ceU,0x502eU,0x3284U,0x42f5U,0xc62fU,0xbc21U,
		0xd956U,0x5a0fU,0x5dbcU,0xa509U,0xd97fU,0xba94U,0x2d8dU,0x2491U,0x7c37U,0xe9ecU,
		0x8219U,0xb653U,0xcbe6U,0x9409U,0xaeddU,0xecaaU,0xadd6U,0xe0e3U,0xab38U,0x1e37U,
		0xa6beU,0x8936U,0x6d59U,0xb979U,0xcf96U,0x1e7fU,0x02faU,0x1e25U,0xc8c1U,0xae8cU,
		0x237dU,0x110aU,0xb0deU,0x0238U,0x161dU,0x5e8dU,0xdd98U,0x9223U,0x4d96U,0xbcdaU,
		0x2fb7U,0xb358U,0x4195U,0x6015U,0x5676U,0xa53fU,0x822eU,0xa64eU,0x7b51U,0xe45dU,
		0x7481U,0xb65dU,0xb787U,0xeacbU,0x4467U,0xbd6cU,0xffc9U,0x9e19U,0x083eU,0xdc3bU,
		0xcdd8U,0x4b24U,0x2c99U,0xec12U,0x93a7U,0xe4ddU,0x817aU,0x841cU,0x3905U,0x70f7U,
		0x60f4U,0xd317U,0x657cU,0x5e6aU,0x8403U,0x1ee0U,0x3c73U,0xc31fU,0x7e0bU,0x2fadU,
		0xfc8cU,0x606dU,0x4cfaU,0x7ebeU,0xba0fU,0xe249U,0x52c5U,0xebc2U,0x1eb5U,0xad95U,
		0xadb3U,0xfa7bU,0x6bcdU,0x8bd6U,0x51ebU,0x5ad9U,0xb92cU,0xb4cbU,0x8d73U,0x66b3U,
		0xdd4dU,0xd4a9U,0xc6f4U,0x5f70U,0xffeeU,0x27a0U,0x26fdU,0xb762U,0xd2dbU,0x5aeaU,
		0x1d47U,0x8dceU,0xb56dU,0xc883U,0x24c1U,0xc02cU,0x124cU,0xcc02U,0xc816U,0x832cU,
		0x4bafU,0xa483U,0x10e1U,0x2878U,0x8c2eU,0xe04bU,0x7edaU,0xd9dfU,0x90a4U,0x65c3U,
		0x17e6U,0x9e0fU,0x217eU,0x3ee7U,0xe074U,0x70d5U,0x4851U,0x10beU,0x9cbbU,0x2f09U,
		0x7ba7U,0x76dfU,0x0043U,0xcedfU,0x3979U,0x414bU,0xee77U,0xe164U,0x5e14U,0x25e7U,
		0x6eb3U,0x1d4fU,0x3ab2U,0x294fU,0x1319U,0x7e73U,0x839bU,0x24baU,0xedf9U,0x673dU,
		0x5cdfU,0x9c43U,0x4a95U,0x4a81U,0x2392U,0x98ddU,0x6f55U,0xaceeU,0xd193U,0x031dU,
		0xcd39U,0xea82U,0x6eebU,0x55b9U,0x4ea6U,0x2e82U,0xf6ddU,0x91c6U,0x699cU,0x4c9fU,
		0xe582U,0x83d9U,0xa283U,0x045fU,0xfd9dU,0xa7edU,0x24f4U,0x76baU,0x422eU,0x8843U,
		0xe655U,0x353bU,0xf750U,0xc98eU,0x07e0U,0xdacfU,0xedc0U,0xf11bU,0x52e9U,0x8170U,
		0x6813U,0x5862U,0x40d6U,0xb51aU,0x1ccdU,0x05d8U,0xd582U,0x9c65U,0x0088U,0x3d89U,
		0x725eU,0x45e1U,0x9dc1U,0x739dU,0x59c3U,0x32ccU,0x9b83U,0xba24U,0xc4aeU,0x06bfU,
		0xe908U,0x94f0U,0xfab7U,0x1d63U,0xd4cdU,0x6f3cU,0xe143U,0x6f68U,0xaa96U,0x8663U,
		0xbc3bU,0x080dU,0x143bU,0xeceeU,0x9783U,0xe76eU,0x3b5cU,0x9a9cU,0x5a8cU,0xce31U,
		0xd049U,0x11deU,0x83a2U,0x5f04U,0x33c4U,0x5c2bU,0x3ffaU,0x7463U,0xbd77U,0xf7acU,
		0x2bf1U,0x2fa3U,0x4368U,0xafa8U,0xede6U,0x4adaU,0x086eU,0x56d7U,0x15f6U,0xe254U,
		0x8651U,0x92daU,0xa85fU,0x6c71U,0x78c7U,0x3781U,0xc104U,0xec48U,0x35ddU,0x5f1eU,
		0x2851U,0x2dcfU,0xfae4U,0x0382U,0x7280U,0x3863U,0x91e2U,0xe0eeU,0x4e2bU,0x6538U,
		0x16c8U,0x2c83U,0xfa86U,0x9342U,0x6cc1U,0x0c71U,0xe48dU,0xaa08U,0x6319U,0xa48eU,
		0xaa24U,0x77cfU,0x9943U,0xb7d3U,0x2338U,0x3dc6U,0xcfb6U,0xdae6U,0x649eU,0x9fb3U,
		0x6748U,0x16a0U,0xd707U,0x4350U,0x85a1U,0xd63eU,0xcd6bU,0xd68bU,0xfb55U,0x4799U,
		0x57f7U,0x884eU,0x1d0cU,0x8246U,0xd235U,0xa062U,0xc14dU,0x56e2U,0x376bU,0xba05U,
		0x5126U,0x1752U,0xbe7cU,0xb113U,0xd33dU,0x2100U,0x6f73U,0x7a5aU,0x0e4eU,0xd4e9U,
		0x3153U,0x210bU,0xe5f2U,0x576eU,0xd35eU,0x36acU,0x3dd1U,0x696bU,0x1aecU,0x9dabU,
		0x5d6bU,0xe7e9U,0x6165U,0x9b06U,0xb41fU,0xe2abU,0xb2aeU,0xee48U,0x23c2U,0x7203U,
		0xf5c6U,0x7418U,0x4a14U,0xcb73U,0xbb79U,0x92b6U,0x6bb7U,0xbb2aU,0xf18bU,0x6f08U,
		0xb83bU,0xa478U,0xa4b6U,0x6194U,0xa24dU,0xa9afU,0x2ea5U,0xf65fU,0x8790U,0x077dU,
		0x7e7aU,0x47daU,0x7cfaU,0x6c8cU,0x6228U,0x9289U,0x752aU,0x7753U,0xb300U,0x0493U,
		0x0ca0U,0xb2f8U,0x10d9U,0x4d76U,0x4923U,0x6e5dU,0x5b1bU,0x194dU,0x86cdU,0x9d14U,
		0x7aecU,0xc547U,0x3a40U,0x3da8U,0x20caU,0x17dbU,0xa67cU,0xfe45U,0x4201U,0xb2bdU,
		0xd47aU,0x56b6U,0xc751U,0xa373U,0xdee7U,0xeeb3U,0xf4b5U,0x5af2U,0x9736U,0xfdd2U,
		0x720fU,0x9512U,0xa099U,0x3fbfU,0x7928U,0x847dU,0x0823U,0xa259U,0xec77U,0x71a1U,
		0xa89fU,0x966fU,0x768bU,0xf194U,0x329bU,0x9c30U,0xb637U,0x7129U,0x4d82U,0xced4U,
		0x3d2cU,0xd888U,0xdcbbU,0x8f1aU,0x9247U,0xd283U,0x75e0U,0x3f5eU,0xe98aU,0x420dU,
		0xcb1eU,0xc79dU,0xe1f3U,0x3a1eU,0x116aU,0x32a7U,0x2a87U,0x4585U,0xc638U,0xf2ebU,
		0x9c95U,0x2f72U,0x9616U,0x63d4U,0xfddeU,0x61a3U,0x8a90U,0x3247U,0xf5d9U,0xf5e5U,
		0x1272U,0x6ebcU,0x2ac4U,0x5e09U,0x3a2aU,0xc049U,0x0d9cU,0x41a7U,0xf047U,0xd4fcU,
		0x285aU,0x6a2cU,0xdc36U,0x1e20U,0x5029U,0x52a9U,0x6785U,0x9799U,0xa090U,0x7539U,
		0xa05dU,0x933dU,0x2f9eU,0x06baU,0x76b5U,0x7e6eU,0x2873U,0xfa76U,0xaa3cU,0x7dbeU,
		0x1a66U,0x52a4U,0x3a19U,0x950dU,0x048eU,0xee43U,0x9323U,0xd49eU,0x1320U,0x2482U,
		0xa46cU,0x23fcU,0x40c6U,0xca52U,0x85ebU,0x425dU,0x6509U,0x1a6bU,0x552aU,0x052aU,
		0xa998U,0x8590U,0x7844U,0x406eU,0x8c60U,0x4b71U,0xd20eU,0x97beU,0x3704U,0x6babU,
		0xc896U,0x9acaU,0x89e3U,0xd32bU,0x829cU,0x7bb8U,0xadc5U,0x9f19U,0x4f9cU,0xe8f5U,
		0xde27U,0x51e0U,0x6ed7U,0x5576U,0x740cU,0x4f46U,0x72c2U,0xe22cU,0xc1edU,0x7dc3U,
		0xfb6fU,0x14baU,0x48e4U,0x1bd5U,0x11caU,0x41f5U,0xcc6fU,0x25aaU,0xae70U,0x08c9U,
		0xce22U,0x272cU,0x0927U,0x09aaU,0x6602U,0x756aU,0xdab7U,0x5ea5U,0xe442U,0x6c9eU,
		0xeea7U,0xf7e6U,0xea55U,0xab7dU,0x5c97U,0xc021U,0xda66U,0x3db7U,0x7ea0U,0x348cU,
		0xa634U,0xaa41U,0xd6b7U,0x0334U,0x8447U,0x06e4U,0x5f9eU,0xe1e3U,0x31d4U,0x751eU,
		0xd9d0U,0x797bU,0x4f33U,0x373cU,0x9aedU,0xbc62U,0x21b3U,0xdebaU,0x72f8U,0x0726U,
		0xe0a5U,0x0e85U,0x65f2U,0xbe1aU,0x1941U,0x1409U,0x9e38U,0x32bdU,0x66ebU,0xffe3U,
		0xc6d5U,0x3d7eU,0x25baU,0xca57U,0xbe0eU,0xb5f1U,0xa038U,0xff0fU,0xcac3U,0xb326U,
		0x0d37U,0xe0d9U,0xe3b7U,0x6a0eU,0x2130U,0xd897U,0x73b6U,0xe717U,0x835eU,0x8eb0U,
		0xcfd8U,0x21a7U,0x49a0U,0x5503U,0x97aaU,0xa6edU,0x2956U,0x1dbdU,0xb009U,0x5326U,
		0x8f11U,0x0884U,0x69c6U,0x8c96U,0x0ca8U,0x40cbU,0x2dacU,0x716aU,0x3730U,0x3e06U,
		0xefa4U,0xbae3U,0x024bU,0xb60aU,0x651bU,0xca8eU,0x3babU,0xd55bU,0xf901U,0x67feU,
		0xc95cU,0xf468U,0xbadbU,0x6e55U,0x65bbU,0xd6abU,0xf214U,0x1595U,0xf725U,0x7e28U,
		0x6e3dU,0x0a7dU,0x7157U,0xd2e3U,0x5e99U,0x1c4fU,0x5286U,0x2401U,0x074bU,0x8165U,
		0x2d09U,0xf422U,0xc842U,0xcc54U,0xa818U,0x748dU,0x07b8U,0xe100U,0xf70eU,0x820fU,
		0x4f74U,0xda5aU,0x9449U,0x85d4U,0x3f23U,0x581bU,0xe4e9U,0x7937U,0xadf7U,0x0185U,
		0xa958U,0x097fU,0xcd02U,0xfc98U,0xda32U,0xc258U,0xd737U,0xe710U,0x7db7U,0xa471U,
		0xece7U,0x2799U,0x6b5bU,0x4223U,0x544fU,0x3b45U,0x504cU,0xbc93U,0xcb0aU,0xfeeeU,
		0x88b4U,0x1543U,0xeecbU,0xa6caU,0x6460U,0x11beU,0x1c82U,0xb079U,0xe6ebU,0x164cU,
		0xdf9cU,0x44c6U,0xbaa0U,0x734eU,0xa5e7U,0x09d5U,0xc658U,0x23a9U,0x9a5dU,0x6002U,
		0xd8e7U,0x2487U,0x8ddaU,0x3053U,0x3923U,0x1b30U,0xa76dU,0x55aeU,0x3305U,0xa8ddU,
		0x4251U,0x01e0U,0xe308U,0x51f7U,0x7592U,0xd73eU,0x76ebU,0xfcecU,0x3612U,0x4488U,
		0xb62eU,0x8e9bU,0xb5a5U,0x8c3aU,0x3ab9U,0xd7ccU,0x4393U,0x8c54U,0x73edU,0x5927U,
		0xeffdU,0x23dbU,0x50c4U,0x1325U,0x89c9U,0x3e9cU,0x0544U,0xfd0aU,0xea8eU,0x4699U,
		0x7225U,0xa278U,0xd811U,0x38d6U,0x9f0dU,0x2086U,0xe823U,0x19e1U,0xea65U,0x554aU,
		0xc2cfU,0xfda9U,0xb18bU,0xcca0U,0xfe12U,0xb152U,0x3653U,0xe745U,0xbda7U,0xdbc2U,
		0x586eU,0x6ecbU,0xb7cdU,0x865dU,0x98d7U,0xd4a3U,0x921dU,0x4717U,0x63a0U,0x327aU,
		0x7e62U,0x01daU,0xe0faU,0x6a08U,0x08c3U,0x7e68U,0x9669U,0xbb24U,0xeae5U,0x6b3cU,
		0x5f34U,0x2638U,0x3cd2U,0xe05cU,0x18b9U,0xbc47U,0x0fc6U,0xa084U,0xbee5U,0xf840U,
		0x746fU,0x8551U,0x38ebU,0x7fbcU,0x93daU,0x6472U,0xa90bU,0x9328U,0x800aU,0x0f63U,
		0x6977U,0x1179U,0x18ccU,0x0062U,0x2476U,0x8260U,0x450cU,0x5132U,0x8d46U,0x5cd4U,
		0xd697U,0x05bfU,0x30ebU,0xc46cU,0xdcf5U,0xed7fU,0x2fe1U,0x2b15U,0x4a5dU,0xbfecU,
		0x8deeU,0xaa30U,0x8515U,0xe02eU,0x0be2U,0x4e00U,0xc7cfU,0xd1c1U,0x6c1aU,0x8595U,
		0x7857U,0x0e17U,0x799aU,0xb9e1U,0xddacU,0xe816U,0xab9bU,0x0472U,0x9f36U,0xbc4dU,
		0x82a9U,0x61bfU,0x617aU,0x154aU,0x9fa7U,0x75adU,0xe6a1U,0x6693U,0x7e3cU,0x38faU,
		0x0537U,0x5b89U,0xacf5U,0xbd1cU,0x912bU,0x13c5U,0xe0c5U,0x2c3cU,0x3551U,0x232cU,
		0x049bU,0xa99dU,0x79edU,0xa3fdU,0xcd53U,0xde7aU,0x214bU,0xf470U,0x96b0U,0x4386U,
		0xcb88U,0x6e23U,0x772dU,0xa1e2U,0x0c65U,0x2285U,0xa52dU,0xc289U,0x7546U,0xec07U,
		0x7f08U,0xded2U,0x2fceU,0x6ce7U,0x32f8U,0x4f1bU,0x3c21U,0x6906U,0x27cdU,0xb76aU,
		0x3499U,0xb88bU,0x7428U,0x052fU,0x97a2U,0x8788U,0xb565U,0xe5e6U,0xa41bU,0x6e06U,
		0x634bU,0x2449U,0xf086U,0x18bfU,0x8a43U,0xbb33U,0x0265U,0xf9b2U,0x4491U,0x4cefU,
		0x2758U,0x3447U,0x9afaU,0x24daU,0xaa52U,0x532fU,0x51c0U,0x79b7U,0x0bd5U,0xca10U,
		0x20e5U,0x9b2eU,0xe7b7U,0xf59cU,0xdb4dU,0x552fU,0x8ff8U,0xd229U,0x671eU,0x0cc9U,
		0x42feU,0xb0eeU,0x638cU,0x2ebfU,0x7dd0U,0x4f12U,0xa36aU,0x9210U,0x5d9fU,0x5e5fU,
		0x3103U,0x7879U,0x1746U,0x25f0U,0x27d7U,0x101cU,0xc0ebU,0xbc7eU,0x660fU,0x0b96U,
		0xe1acU,0x6619U,0x5f27U,0xf3a7U,0xa7cfU,0xc6c8U,0x75ccU,0x1a70U,0x6f15U,0x8082U,
		0xd1d9U,0xc1b5U,0xe44cU,0x5691U,0x1c05U,0x9b77U,0x44f3U,0xe062U,0x81ecU,0x78ecU,
		0x4993U,0x88bbU,0x6235U,0x9589U,0x8991U,0x7e84U,0x2aa4U,0x664fU,0xd51fU,0x0371U,
		0x6f5cU,0x8741U,0x8d34U,0x42c7U,0xe994U,0xcf8bU,0xea3eU,0x1c61U,0xdf4eU,0x650eU,
		0x63aaU,0x7f39U,0x2e1bU,0x8524U,0x9d21U,0x5a80U,0x076cU,0xe344U,0x9066U,0xc213U,
		0xd903U,0xd8f1U,0xc3c9U,0x487aU,0xf718U,0x8ce0U,0xbf3eU,0x3346U,0xe7adU,0xd53cU,
		0x5c1fU,0x0d41U,0xf98dU,0x44b5U,0xd25cU,0x5e82U,0xdef4U,0x1048U,0xd6ecU,0xd43bU,
		0xff4fU,0x4262U,0xd59dU,0xadeaU,0x2de2U,0xb131U,0x3e1dU,0x9749U,0x9254U,0xb098U,
		0x410eU,0x3cd8U,0x1f5dU,0x8645U,0x7e15U,0x9a18U,0xc307U,0xc771U,0xf701U,0x969cU,
		0x201dU,0xd41eU,0xb025U,0xe46dU,0xab61U,0x8223U,0x71aeU,0x6cd6U,0x4aceU,0xb917U,
		0x6e7aU,0x17f0U,0x428aU,0x85f0U,0x9e68U,0x0c5cU,0x3a37U,0x1c75U,0x8e40U,0x7375U,
		0x52f3U,0xe32bU,0x1f91U,0x7274U,0xe680U,0xed18U,0xf2f8U,0xe9aeU,0x2eb5U,0x30b2U,
		0xb3a3U,0x6b83U,0x7daaU,0x2bfbU,0x0e66U,0xeac0U,0x4ee0U,0x2ce1U,0x3775U,0xa752U,
		0xf2c0U,0x5141U,0x424aU,0xe3b0U,0x76aeU,0x6bb0U,0xaa8fU,0xdd46U,0xce86U,0xcbdbU,
		0x48d7U,0x818aU,0x7cd5U,0x3d00U,0x2645U,0x263eU,0x44ecU,0x9f2fU,0xcf16U,0xfef5U,
		0xa5adU,0xf744U,0x9383U,0x3a75U,0xddccU,0x3990U,0x7430U,0x091aU,0x98e4U,0x3c4cU,
		0x5477U,0x34caU,0x28dcU,0x05f6U,0x05ccU,0xb9caU,0xb308U,0x3709U,0xf9d4U,0xb96eU,
		0xaf94U,0x744bU,0x9760U,0xc964U,0x31c7U,0xe85eU,0x4909U,0x0f18U,0x51b6U,0xa8ecU,
		0xa78eU,0x59b7U,0x197dU,0x895fU,0x5bb6U,0x6abdU,0x93e7U,0x0d80U,0x21c4U,0x1721U,
		0x4b59U,0x9e2bU,0x50f2U,0xfb08U,0xbb97U,0x2705U,0xa2f1U,0xfd1eU,0x3930U,0x97c3U,
		0x1967U,0x60b5U,0xace2U,0x78dcU,0xed68U,0x50e4U,0x03f1U,0x5d1bU,0x40b9U,0x5f3aU,
		0xa975U,0x9185U,0x70b3U,0x7a08U,0xf73cU,0x75a5U,0x8ea8U,0x618cU,0xb75aU,0x838fU,
		0xcf08U,0x411fU,0x0bf4U,0x2216U,0xb86bU,0xa050U,0x703bU,0x46fcU,0xc732U,0x5a04U,
		0x5ee9U,0xd213U,0x91baU,0xf242U,0x16edU,0x29e3U,0x185fU,0x6857U,0xf9a8U,0x91acU,
		0xc099U,0x0f45U,0x89d6U,0x4c88U,0xc90dU,0xdbe5U,0x1e68U,0xf571U,0x46ecU,0xcd2dU,
		0x1b08U,0x3ac8U,0xfdb6U,0x3960U,0xed5aU,0x8492U,0xb3c1U,0x6e1aU,0x9d0bU,0x551dU,
		0x2ab7U,0x6dddU,0x1f3aU,0x4b76U,0x88d4U,0x1f83U,0x545cU,0x94b1U,0xa18aU,0x5a44U,
		0xe158U,0xbb14U,0x264cU,0x6b42U,0x04bcU,0xfc1cU,0x44d3U,0xcb11U,0xa68cU,0x8dadU,
		0xd6deU,0xd8c1U,0x72acU,0x7afeU,0xdffdU,0x04f4U,0x2399U,0x4845U,0x47b8U,0x5109U,
		0xdfafU,0x46a8U,0x5485U,0xc9ecU,0x923aU,0x8c65U,0x4c6bU,0x5be6U,0xf07cU,0xbf30U,
		0x3edbU,0x6b9bU,0xa825U,0x2a27U,0xa942U,0x2f4cU,0xd40bU,0xe420U,0x691aU,0x8e85U,
		0xdf3dU,0xa892U,0xfd42U,0x8b63U,0xd511U,0x318cU,0x9279U,0x7eceU,0x83efU,0xfba7U,
		0x098cU,0x9d58U,0x254aU,0x2e6fU,0xdfebU,0x0bc7U,0x8c76U,0x4073U,0x7adfU,0x5018U,
		0x53ceU,0xfe21U,0x0a8aU,0x0796U,0xa319U,0x9f5fU,0xfa17U,0xeaebU,0x8dc2U,0x5f60U,
		0x5d91U,0x3503U,0x3fddU,0x7c1fU,0xa8faU,0xd46dU,0xac92U,0x6be5U,0x1395U,0x12d4U,
		0x9edaU,0xd5bdU,0x7d3aU,0x035cU,0x6528U,0x1240U,0x0dcaU,0x4434U,0xe5dcU,0x7849U,
		0x0891U,0x46c2U,0xb2a1U,0x2743U,0x9603U,0xdb6fU,0x2313U,0x08fdU,0xad89U,0x8121U,
		0x3f8eU,0xf252U,0x066aU,0x22bbU,0x213dU,0xee95U,0xad71U,0xcd4aU,0x10d0U,0xa240U,
		0xbba5U,0x4647U,0x5fe0U,0x7f80U,0xbd5cU,0x8d67U,0xe124U,0x4135U,0x6f23U,0x2d81U,
		0xaabcU,0x9f1eU,0xb85cU,0x9ea7U,0xa7a4U,0x7c87U,0x2939U,0x6f8cU,0xbdddU,0xe575U,
		0x3d07U,0x7fc2U,0x4f8eU,0xbe42U,0x42e9U,0xbc9aU,0x91ffU,0x7a10U,0x4757U,0xaa80U,
		0xe52aU,0xff86U,0xb893U,0x6c43U,0x4a88U,0xc23dU,0x30c9U,0xfb2eU,0x7723U,0x3a0bU,
		0x94cfU,0xbfc8U,0x775bU,0xadcaU,0xd17bU,0xf164U,0xad7aU,0x0c13U,0x6fadU,0x6806U,
		0x7f24U,0xa152U,0x7e92U,0xb3caU,0x5960U,0xd2c1U,0x9e71U,0x0f95U,0xb647U,0xfbdfU,
		0x8cb6U,0x9630U,0x4946U,0xfd33U,0xe648U,0x1152U,0x54d9U,0x7dd9U,0xb05dU,0xae62U,
		0x7b31U,0x27a8U,0x3c12U,0x5c5fU,0xc49cU,0x7bbdU,0x678eU,0xf863U,0x5cbeU,0xcaedU,
		0xf6caU,0x0508U,0x4a06U,0x21f0U,0xff7dU,0x38f1U,0xf656U,0x6bc0U,0xefdeU,0x8fe9U,
		0x361bU,0x5274U,0xc9c4U,0xa7feU,0x7a49U,0xdf1aU,0x2e29U,0xb012U,0xa3f3U,0x0c92U,
		0x9d3dU,0xef86U,0x26f0U,0xef18U,0x6c08U,0x4f86U,0x44e4U,0x82a1U,0x2128U,0xf183U,
		0xc6ecU,0x1c91U,0x328dU,0x1dd5U,0x96d5U,0x9b97U,0x71dbU,0x9097U,0xdf06U,0xd01bU,
		0x4328U,0x542fU,0xa1b3U,0xf6f2U,0xd54aU,0x5e1dU,0xf483U,0x7d5bU,0x90beU,0xf039U,
		0xc616U,0xbf6cU,0x6ea6U,0xe379U,0x35e6U,0x28f5U,0x1e8cU,0x62faU,0x9522U,0xd330U,
		0x8135U,0xecccU,0x0fb8U,0xe4b4U,0xf138U,0x7dffU,0x4174U,0xd5d9U,0x7cdcU,0x8557U,
		0x65d9U,0xf431U,0x9912U,0x5053U,0x1adfU,0x497aU,0x13afU,0xcf62U,0x2fc4U,0xdce7U,
		0xe23dU,0xc5c8U,0x4a9cU,0xbec2U,0x9a26U,0x8e58U,0x3083U,0x1b87U,0x9a89U,0x16bbU,
		0xca3bU,0x89e8U,0x9a4fU,0xd56eU,0xe606U,0xc0c7U,0x9713U,0xce9dU,0x7b64U,0xffbdU,
		0xdc48U,0x9863U,0x949bU,0x9e82U,0xd6faU,0x6680U,0x6420U,0x03dbU,0x4bd4U,0xd803U,
		0xe560U,0xf1daU,0xee5cU,0x3aefU,0x8410U,0xd44eU,0x43c7U,0x3e7cU,0x2509U,0xb5daU,
		0xd0b0U,0xb817U,0x5441U,0x9acfU,0x30c0U,0x9122U,0x7cf1U,0x35d0U,0x86ffU,0x2ae5U,
		0xc67cU,0x0f2eU,0x0aa8U,0x7475U,0x1caaU,0x1516U,0x235bU,0x384fU,0xc27fU,0x073dU,
		0xf353U,0x1fc1U,0x2be4U,0xe4f8U,0xdb87U,0x9a40U,0x03c5U,0x3ec1U,0xb34cU,0xd925U,
		0x4c01U,0x4635U,0xf8f3U,0xb67cU,0xaeeaU,0xc89bU,0xe3eaU,0xd03cU,0xc19bU,0xad24U,
		0x4298U,0x2e56U,0xff6eU,0x9bf6U,0xe8e3U,0xf1f8U,0xb244U,0xdd8cU,0x589eU,0xaffbU,
		0x8c0aU,0x380cU,0xbc2eU,0x2c55U,0xda7cU,0x36f1U,0x2d63U,0x5152U,0x84a7U,0x4748U,
		0x870eU,0x60d1U,0x10feU,0xa5bdU,0x8c1eU,0xfa62U,0x6deeU,0x4f4bU,0xe8ccU,0xf699U,
		0xfc4bU,0xf910U,0xcc0fU,0x1821U,0xdaa6U,0xa57eU,0x8191U,0xf846U,0x6c10U,0x5ba4U,
		0xccdcU,0x3b4cU,0x6956U,0x70bbU,0xb778U,0x8064U,0x14edU,0x3d51U,0x34a1U,0x8bb7U,
		0x9c4aU,0xc2beU,0xaadfU,0x4525U,0xdd5aU,0xb37aU,0x1b29U,0xff08U,0x7532U,0x7411U,
		0x1d5cU,0x5ad2U,0x3b75U,0xbac8U,0x7b92U,0xf909U,0xe4adU,0x7c80U,0x94aaU,0xcbd4U,
		0xc1aeU,0xb9daU,0xb93fU,0xfd71U,0xe081U,0x3c63U,0xc1d0U,0x6824U,0x4bc1U,0x0a1dU,
		0x773bU,0x1cbcU,0x4e5dU,0x2428U,0x1260U,0x5ab6U,0x137eU,0xfff6U,0xc503U,0xc13eU,
		0xfbc2U,0x557bU,0x7ffaU,0x1213U,0x5cecU,0x3e3fU,0x9479U,0x5fb5U,0xe80cU,0xcdcaU,
		0x13d3U,0xbeebU,0xe95fU,0x50a8U,0x91d3U,0x39a6U,0x8aa0U,0xde48U,0xe970U,0x3f13U,
		0xeaacU,0x5b04U,0x3515U,0xa0d2U,0xb0d0U,0x17b8U,0x2d78U,0xd1b8U,0xe2a2U,0x8ca5U,
		0x9f44U,0xfc09U,0xc34bU,0x6edcU,0x8f5dU,0xf8a8U,0x6775U,0x7aabU,0x4676U,0xa6b0U,
		0x07edU,0xaaf8U,0xd6cdU,0xc7ffU,0xcd92U,0xf787U,0xf691U,0x9e9fU,0xdd3eU,0x0e0fU,
		0xb5e9U,0x71fbU,0x8bceU,0xa66cU,0x603aU,0xb989U,0xb6feU,0x827fU,0x45eeU,0xb392U,
		0x7792U,0x8a09U,0xe487U,0xfab1U,0xedf3U,0x51e5U,0x2377U,0x0b3aU,0xe697U,0x7b99U,
		0x639aU,0x391dU,0x2d03U,0xa032U,0x48deU,0xa08aU,0x7922U,0xf5c0U,0x53c8U,0x5456U,
		0xacdcU,0xce80U,0x2ddcU,0xd1d3U,0xb55fU,0x7994U,0x02e7U,0xe691U,0x5ce6U,0xfc45U,
		0x7cebU,0x0fb2U,0x5cb8U,0xa79eU,0xaed0U,0xbe30U,0xaf30U,0xde2cU,0x0ba4U,0xedb2U,
		0x12f9U,0x0107U,0xf3dcU,0x95d6U,0x7acfU,0xd77bU,0x47fdU,0x6ad2U,0xe915U,0x72e9U,
		0x8633U,0x2807U,0x9143U,0xa3b5U,0x3aa4U,0x5b44U,0xf604U,0x836dU,0x9aa9U,0x8343U,
		0x1197U,0x46dcU,0xbcf7U,0x96f4U,0xd1ffU,0x1ba3U,0xb40aU,0xb4ecU,0x2c32U,0xe8faU,
		0x8e67U,0x63f8U,0x02edU,0x99ecU,0xfecfU,0x2d9dU,0xd3f7U,0x1790U,0x2118U,0x0fccU,
		0xc422U,0xbfd9U,0xa429U,0x3cb6U,0x0860U,0xdbf9U,0xbdd4U,0xab92U,0xbb70U,0x5669U,
		0xa5d8U,0xc70aU,0xd242U,0x80e8U,0x4618U,0xf343U,0xf026U,0x18efU,0xdaf3U,0xfb47U,
		0xc7adU,0x4fa1U,0x60e7U,0xdc9eU,0x0eadU,0x8d93U,0x6aa4U,0x2193U,0x1093U,0x3603U,
		0x465aU,0xcf37U,0xe8bcU,0xca66U,0x0e40U,0xf764U,0xf672U,0xeb9aU,0xdd6cU,0xfc7fU,
		0x16d5U,0x34e2U,0x0938U,0x4547U,0x08b4U,0x658eU,0x533dU,0x1d1fU,0x064fU,0x3a94U,
		0xa0e8U,0x17c1U,0xf850U,0x1bdaU,0x3b80U,0x8a11U,0x010dU,0x0a24U,0x508dU,0x5fbfU,
		0xdc02U,0x61d3U,0x3997U,0x117fU,0xd023U,0xe8d7U,0x81d4U,0x1653U,0xc356U,0x4704U,
		0x0b44U,0x094cU,0xac7eU,0x21f9U,0xbb3bU,0xc50eU,0x2e89U,0xe6e0U,0x9ad8U,0x1812U,
		0xad18U,0xb4f6U,0xd867U,0x1a8aU,0xc54fU,0x48e9U,0xdb7eU,0xb01cU,0xee9eU,0x7b48U,
		0xe920U,0xd2ebU,0x842fU,0x5214U,0x8459U,0x849bU,0xfb38U,0x871fU,0x85f9U,0xe783U,
		0xfb9dU,0x3859U,0xf6acU,0xfedaU,0x8f33U,0x6486U,0xcf6cU,0x09caU,0x1feeU,0x2ec8U,
		0xafefU,0x5dd2U,0x109eU,0x8334U,0xceb6U,0xefe9U,0xff98U,0x14bfU,0x52b2U,0xfcaeU,
		0x12ffU,0x443eU,0x0f86U,0xe38eU,0xa1beU,0xd0a4U,0x5bf0U,0x697dU,0x030bU,0x92bfU,
		0x8fb5U,0x6861U,0xfcf5U,0x58e5U,0xf873U,0x9cf8U,0x2608U,0x4762U,0x0efbU,0x1dc6U,
		0x0f22U,0xb6d5U,0x4951U,0xb7eaU,0xb47eU,0xd72cU,0x3350U,0x15d9U,0x3c38U,0xfb74U,
		0xb466U,0x4db2U,0x84b2U,0x358dU,0x5037U,0x2bd8U,0x62ebU,0xe9b8U,0x9a06U,0xdb92U,
		0xf932U,0x0d5fU,0x2514U,0x1e00U,0xc451U,0x134cU,0xd3deU,0xe16dU,0x6910U,0x78bdU,
		0xd919U,0xc29aU,0xf9f5U,0x9a31U,0xbbb4U,0xbc88U,0xec51U,0x1c15U,0xd6c5U,0x71d3U,
		0xc091U,0x7dc8U,0xcb02U,0x960eU,0xedb8U,0x4bc8U,0x5eafU,0x316fU,0xc6b4U,0xbe65U,
		0xddd3U,0x0f69U,0xa324U,0x99feU,0x54a2U,0xe6f2U,0x17a7U,0xa94dU,0x9f6eU,0x0787U,
		0xb5ceU,0x97ffU,0x04c7U,0x8be6U,0x4eadU,0x4cc0U,0xc93dU,0x84fcU,0x3f68U,0x57c5U,
		0x8fc9U,0x2cd0U,0x393bU,0xc1f2U,0x4721U,0xca2fU,0xb9bbU,0xdc73U,0x9bbbU,0xc73dU,
		0x53aaU,0x31acU,0x0734U,0xc768U,0x7561U,0x600cU,0x4834U,0xda4fU,0xe10aU,0x87faU,
		0xce91U,0x8cf6U,0x2d41U,0x3951U,0x692fU,0xb330U,0x2f42U,0x92d0U,0xae10U,0xa87fU,
		0x86a1U,0xf9dfU,0xe99fU,0x4a32U,0x1081U,0xe231U,0x2894U,0xe796U,0x0baaU,0xe734U,
		0xaeb6U,0x996eU,0xf3f5U,0xd267U,0x9ee9U,0x8010U,0xb4b6U,0x8751U,0xc329U,0x5a71U,
		0x496eU,0x31b5U,0x0438U,0x92ebU,0xab6cU,0x07adU,0x48f2U,0x507aU,0x7078U,0xe9f6U,
		0x7d73U,0x0e71U,0x01c2U,0x700fU,0x9c83U,0x65cdU,0xb87cU,0x72c7U,0xc9d4U,0x45d0U,
		0xf26fU,0x7551U,0xdd08U,0x36d2U,0xf125U,0xd137U,0x7e47U,0x68d7U,0xb037U,0x4414U,
		0xecc0U,0x5daaU,0x5d4bU,0x5b7aU,0x6cacU,0xf417U,0xad08U,0xf7b6U,0x5360U,0xcf4aU,
		0x449cU,0xcab0U,0x56ecU,0xa4dfU,0xf072U,0xeddcU,0x3decU,0xce52U,0xef5eU,0x2731U,
		0x4de1U,0x779aU,0xde32U,0x682bU,0xbb87U,0x29c8U,0x0a06U,0x3ba0U,0x3a7cU,0x932eU,
		0xdf0eU,0xd002U,0x66f6U,0xb080U,0xbab6U,0x7043U,0xa9d8U,0x0cdbU,0x84d6U,0x984cU,
		0x8a4bU,0x04acU,0x55c0U,0x5ca2U,0x099bU,0x4c45U,0xfbf4U,0x9d31U,0xd856U,0xb846U,
		0x7af4U,0xce27U,0x47e8U,0x1899U,0x1925U,0xcff4U,0x98afU,0x715fU,0x20d5U,0x1a53U,
		0x5c53U,0x04ddU,0x6026U,0x56a7U,0x857eU,0xa202U,0x3896U,0xb4adU,0x1ca1U,0x1f54U,
		0xce19U,0x4d8dU,0x1770U,0x8ac2U,0x6943U,0xe670U,0x2acfU,0x49faU,0xc78eU,0xc81eU,
		0xa213U,0xd5fdU,0x8f71U,0x08ceU,0x2f5bU,0xd119U,0xaf36U,0x9365U,0x8d51U,0x0053U,
		0xbf60U,0x4d57U,0x063fU,0xa911U,0x3038U,0xf209U,0x4931U,0x3321U,0x300cU,0x48a4U,
		0xe35dU,0xa778U,0xd9adU,0x77a8U,0x19d2U,0x05a8U,0x5b30U,0x431cU,0x959dU,0xe94fU,
		0x7230U,0x2ba3U,0x09e9U,0x74afU,0xbffcU,0xae75U,0x6b66U,0xe6c8U,0x47aeU,0x33baU,
		0x908bU,0xe210U,0x4733U,0x0176U,0xa32cU,0xf45dU,0x3073U,0x0756U,0x205cU,0x54baU,
		0x0cb3U,0xd670U,0x604cU,0xc4eaU,0xc8dcU,0x0b76U,0x7d0bU,0xc6e0U,0x3d6fU,0xa9e6U,
		0x56c4U,0x9fc6U,0x0448U,0x5950U,0x6d49U,0x28c3U,0x5f4bU,0x25afU,0xf493U,0x9e55U,
		0xbe36U,0xc1d7U,0x561bU,0xaf69U,0x1cfbU,0x90e4U,0x938aU,0x6478U,0x4ffeU,0xccf0U,
		0x7aa3U,0x1c89U,0x29dbU,0x0cc1U,0x421bU,0x457dU,0x3533U,0xb850U,0x00caU,0x8b2fU,
		0x6ef2U,0xd3c9U,0xe6d2U,0xba5fU,0x33b2U,0x3585U,0x4472U,0xb31bU,0x57e7U,0xcc74U,
		0x6c37U,0x0587U,0xa4e9U,0x605eU,0x69b5U,0x28b9U,0x6dd3U,0xba1aU,0xdfbeU,0xb33aU,
		0x1fd4U,0x0d51U,0xe177U,0x7a86U,0xd2adU,0x52d9U,0x62abU,0x91f3U,0xee0bU,0xde98U,
		0x7699U,0x9725U,0x99baU,0xe78dU,0x9119U,0xade1U,0x14b1U,0xae83U,0xb184U,0x65b4U,
		0xadbeU,0x41faU,0x284aU,0x003cU,0xaed6U,0xcfcdU,0xf5fdU,0x4bbaU,0xe559U,0x493fU,
		0xd50aU,0x93e0U,0xe7a6U,0x7f01U,0x5853U,0x3b23U,0x18dcU,0x6fa1U,0xee04U,0xce12U,
		0x2d3aU,0x8f2cU,0x319eU,0x40f0U,0xf4fcU,0xb48fU,0x403bU,0x4fccU,0xa83eU,0x4239U,
		0x9624U,0x2b90U,0x89a7U,0x11cfU,0xd077U,0x0879U,0x8f01U,0xd60bU,0x1b5cU,0x1d9dU,
		0x4d1aU,0x7001U,0x9a78U,0x1cf1U,0xe14eU,0xe139U,0x45a1U,0x7b25U,0x2c74U,0xd348U,
		0xfc2bU,0xf804U,0x3432U,0x9efdU,0x2343U,0xd9efU,0x5b6cU,0x226cU,0x8610U,0xb3f9U,
		0x8ee4U,0xeb2aU,0x93a2U,0xb0d9U,0xd951U,0x6ca3U,0xbc25U,0xb39aU,0x72efU,0x3c6aU,
		0x5121U,0x6314U,0x0869U,0xd4c8U,0xf74bU,0x5cdaU,0x909fU,0xb927U,0x5f99U,0x11c5U,
		0x783fU,0xa058U,0xc633U,0x081eU,0x491eU,0x4a0fU,0x18c7U,0x7e5dU,0xea89U,0xa768U,
		0x544aU,0xc83dU,0xef9fU,0xcabeU,0x8e3bU,0x3e18U,0x9d1cU,0xe447U,0x42f9U,0xa416U,
		0x2146U,0xdda7U,0x95feU,0x0a85U,0x3ed6U,0xa185U,0x185aU,0xed63U,0x975bU,0x48d2U,
		0x4293U,0x86faU,0x970eU,0xf133U,0x3288U,0xf6c5U,0x6fa8U,0x2934U,0x6a9fU,0xfecaU,
		0xf3d7U,0x6395U,0x4671U,0x9474U,0x7b8dU,0xcc0aU,0xdd03U,0xaeb1U,0x9bb6U,0x5eaaU,
		0x5032U,0x0f81U,0xe91bU,0x5088U,0x1b57U,0xf5f8U,0x69b0U,0x5616U,0x9086U,0x8d4cU,
		0x98aaU,0xbb82U,0x256eU,0xdc14U,0x76f4U,0xa558U,0x97d2U,0x1b52U,0x95f9U,0x511cU,
		0x96e8U,0x295fU,0x9429U,0x3372U,0x7625U,0xf835U,0x6083U,0xdd22U,0x760aU,0xbbdbU,
		0x964dU,0xdabcU,0xee1eU,0xa720U,0x0420U,0xba67U,0x97d7U,0x67d7U,0xbf06U,0xaec4U,
		0x0f71U,0x5df3U,0xc202U,0x9bd9U,0xde11U,0x49d0U,0xe318U,0xafcbU,0x2818U,0xee80U,
		0x4be8U,0x12b6U,0x7656U,0x3fa5U,0xd76fU,0xdba3U,0xdb2aU,0xeb41U,0xc10dU,0xeadaU,
		0x68c6U,0xe7d7U,0xa7b6U,0x756fU,0x2495U,0x2c26U,0x6ad8U,0xe088U,0xa55dU,0x6c5aU,
		0xf0e0U,0xedd3U,0xa5b4U,0x5138U,0xffdaU,0xebb9U,0x0a6bU,0x6467U,0x2254U,0xcdeeU,
		0x2d91U,0x9535U,0x5907U,0x6103U,0xc2f1U,0xf7faU,0xcd40U,0x5764U,0xf50eU,0x35bbU,
		0x2686U,0x5f85U,0xb6f6U,0xc60eU,0xed52U,0x6607U,0xba98U,0x0d94U,0x7256U,0xbdc8U,
		0x76f9U,0x7243U,0x1681U,0xb1ccU,0xd797U,0xd767U,0x3d67U,0xf9edU,0xbb56U,0xa124U,
		0x3e6aU,0x246bU,0xd983U,0x82e5U,0x18e3U,0xcbafU,0x8f9cU,0x155bU,0xf0abU,0xd0c4U,
		0xc91cU,0x2a42U,0x5412U,0x1f6dU,0x6af4U,0xe8b0U,0x5937U,0x09afU,0xa50dU,0x99c3U,
		0x4803U,0x309bU,0xdc19U,0x39d7U,0xa170U,0x666aU,0xbd33U,0x4512U,0x389fU,0xebd1U,
		0xdc92U,0x6a63U,0x655eU,0x53b3U,0x5dc0U,0x2b0aU,0x6e90U,0x19b8U,0xc558U,0x6761U,
		0x6840U,0xdde1U,0x23e9U,0x8edaU,0xe11aU,0x615bU,0xef6fU,0xb71aU,0xd82aU,0x092cU,
		0x5a13U,0x3261U,0xe846U,0xb691U,0x2573U,0xd1a3U,0x78a3U,0x8ec6U,0x0e97U,0x33f1U,
		0x02ccU,0x4c12U,0xe831U,0x7c6aU,0xeb8eU,0xbe93U,0xd95aU,0xeb51U,0xfe95U,0x850aU,
		0xe407U,0x9ff9U,0x4785U,0x39f3U,0xfdf6U,0x99d5U,0xba4cU,0xf0e9U,0x7f4cU,0x7ef7U,
		0x5f56U,0x3dbcU,0xecaeU,0x45f6U,0xd781U,0xfd78U,0xdd27U,0x447dU,0x9b47U,0xcec4U,
		0xfefcU,0x8266U,0x71e3U,0x8ab3U,0xb749U,0xa6d1U,0xd821U,0xb873U,0xaee1U,0x4281U,
		0xa62bU,0x0835U,0xf08eU,0x4e82U,0x0324U,0xbbc5U,0x11f6U,0xe4a1U,0x74a0U,0xd082U,
		0xa5f5U,0x6fc1U,0x86d5U,0xda6bU,0x940dU,0x98baU,0x43e1U,0x686fU,0x6088U,0x6e45U,
		0xe1cfU,0x73e2U,0xed96U,0xaf5aU,0x0a11U,0x7d23U,0x4afeU,0x0c2cU,0xb54eU,0xa337U,
		0xcbeaU,0x87b7U,0x3b2aU,0x6d94U,0x3043U,0x76caU,0x9820U,0x1d89U,0x3b14U,0xeeecU,
		0x6d1fU,0x1254U,0xd685U,0x5efeU,0x4c99U,0xc026U,0xb657U,0x2840U,0x7ad5U,0x6cb7U,
		0xf83aU,0xd7c6U,0x0979U,0x1db7U,0xab77U,0x247cU,0xd27dU,0x47d4U,0xd5b7U,0x483fU,
		0x2210U,0x3c46U,0x821dU,0x873bU,0x79b1U,0xbd16U,0x80e2U,0xfc3fU,0x17b2U,0xc2b8U,
		0x3ebbU,0xbebcU,0x0c8cU,0xc237U,0xb489U,0xd3c3U,0x4316U,0x5c9cU,0xe9f0U,0x4cbaU,
		0xb6cfU,0xe6daU,0x762aU,0x36c3U,0xc2b2U,0xbbbfU,0xdddbU,0x575eU,0x336cU,0xa762U,
		0x9305U,0xbaf9U,0x00b4U,0xc3b7U,0x7c3bU,0x400bU,0x2e94U,0xc9abU,0xda40U,0x3b01U,
		0xd79fU,0xccd0U,0x24aaU,0xcf77U,0x6132U,0x135aU,0x7054U,0xb147U,0x95b9U,0xab82U,
		0xb97dU,0xc59aU,0x95dcU,0xb946U,0x3377U,0x40e6U,0x12caU,0xd634U,0xcf1dU,0x0068U,
		0xf0c8U,0x1527U,0xf222U,0xeed2U,0xf898U,0x1499U,0x6d5dU,0xdd32U,0xd7e4U,0x095aU,
		0xed34U,0xb436U,0xd19aU,0x36c9U,0x2e4dU,0x736cU,0xe1daU,0xe4d4U,0x9563U,0xd7f3U,
		0xdbd2U,0xea5aU,0x893aU,0xd7b7U,0xf77bU,0x868dU,0x942eU,0xabc9U,0x8d08U,0x6dadU,
		0x7825U,0xab10U,0xb8bfU,0x1e46U,0x9e97U,0x5427U,0xdbddU,0x5e57U,0xa6c2U,0x323fU,
		0x585aU,0x8f51U,0xf4c6U,0x705fU,0x183eU,0x3b8eU,0xbaf1U,0x49c8U,0x54b2U,0x1df8U,
		0x3f07U,0x012fU,0x11aaU,0xf7ebU,0x1e3bU,0x207bU,0xf3e2U,0x3a65U,0x2964U,0x8e1eU,
		0xa8b9U,0x8d1aU,0x5d09U,0x18d2U,0x3fd3U,0x6494U,0xc118U,0xe296U,0xe8a1U,0x1338U,
		0xab3cU,0x27e9U,0x5bd6U,0x988cU,0x2235U,0xc7e0U,0xb5b7U,0x7630U,0x5bc5U,0xa020U,
		0x6264U,0x1fa1U,0xdbbbU,0x0a76U,0x556fU,0xeeacU,0xe0e7U,0x25e0U,0x892fU,0x89beU,
		0x96edU,0x5aafU,0x3e75U,0xae5bU,0x9d51U,0xfb01U,0x5e7bU,0x68ffU,0x49c1U,0xeee5U,
		0x2a3bU,0x946dU,0xaddaU,0x49f3U,0xa878U,0x5dcbU,0xeb6eU,0xb955U,0xa3dbU,0x6fd7U,
		0xfea0U,0x6a88U,0x84c5U,0x92a4U,0xc8afU,0xa0dbU,0x819bU,0x06e9U,0x110eU,0x8287U,
		0x8373U,0xfffdU,0xba6cU,0x9483U,0x3cbfU,0xc7c3U,0x05fdU,0x2b1bU,0x9b9fU,0xad5dU,
		0xac71U,0x23b0U,0x8a5eU,0x221eU,0x2381U,0x90fbU,0x0decU,0x0511U,0x2451U,0x3bd3U,
		0x0466U,0xc9b1U,0xb820U,0x4c25U,0x9397U,0xb414U,0x0157U,0x8972U,0x1955U,0x844cU,
		0xae90U,0x2e33U,0x22c4U,0xd9bfU,0x0425U,0x7e30U,0xb1ebU,0x1f15U,0x81fbU,0x3969U,
		0x30d3U,0x2295U,0x17f9U,0xf299U,0x1fe1U,0x2365U,0xc8c5U,0x3de2U,0x4240U,0x1eabU,
		0x13b9U,0x4186U,0x4b39U,0xb0f7U,0xd4bdU,0x92abU,0xf8cbU,0xc034U,0xefc1U,0xa931U,
		0xa593U,0x0339U,0x1e29U,0x0eeeU,0xf60aU,0x0dd4U,0xa725U,0xc07dU,0x387dU,0x9313U,
		0x4c75U,0x2fe7U,0x698dU,0xcb7cU,0x40fdU,0x1869U,0xd747U,0x9161U,0x02feU,0x4dcfU,
		0x8886U,0x41b9U,0xd8d3U,0xa6f6U,0x4913U,0x2e9aU,0x1471U,0xd304U,0xde6eU,0xada4U,
		0xbf48U,0xacafU,0x68bdU,0xd6bcU,0x1e83U,0xa7c6U,0xdc2dU,0xd3d1U,0xee23U,0x4a4dU,
		0x4d27U,0xf302U,0xbe6dU,0x32ecU,0x409aU,0x87dbU,0xe5c3U,0x3b96U,0x3cf6U,0x66a9U,
		0xcf9aU,0x4154U,0x3c2bU,0xbcccU,0xc832U,0x6c81U,0xb1d4U,0x170dU,0x2279U,0xc0f5U,
		0x378aU,0x68ecU,0xacb8U,0x0571U,0x1932U,0xaa46U,0xbcdeU,0x1576U,0x5b4aU,0x1385U,
		0xdac1U,0x4827U,0x19faU,0xb448U,0x28e3U,0xed85U,0xa60aU,0xfd86U,0x01a6U,0xc65fU,
		0x4ce3U,0x64faU,0x4d9aU,0x879aU,0x184fU,0x86c2U,0xb9aeU,0x672fU,0xa28aU,0x4011U,
		0x9075U,0x1036U,0x75eaU,0x3561U,0xec68U,0x8a37U,0xd391U,0xa639U,0x9227U,0x5c0fU,
		0x5d35U,0x82fdU,0x9652U,0x3206U,0x2efbU,0xc930U,0xb1a8U,0x68a1U,0x4deeU,0x9a67U,
		0x877cU,0xa4a7U,0x07c2U,0x2df7U,0xdd9cU,0x4d6bU,0xa845U,0x9ccaU,0x0ce8U,0xcacdU,
		0xc6a7U,0x831eU,0xc524U,0x84ccU,0xddc2U,0x1d3dU,0xc4fbU,0x3c0aU,0xa2e9U,0x3491U,
		0x5e91U,0x972eU,0x3aaaU,0x173aU,0xbbe0U,0x4e3bU,0x346dU,0xefb9U,0xd67dU,0xdcfbU,
		0xa20bU,0xceaeU,0x3c7dU,0x5833U,0xf3ccU,0xc746U,0x1621U,0xcdfeU,0xf39bU,0x98c8U,
		0x1a93U,0x1efcU,0xc009U,0x7c41U,0xf065U,0x369eU,0x43aaU,0x45b4U,0x332eU,0x90aeU,
		0xa6daU,0x7ea5U,0x023cU,0xdb41U,0x1ac8U,0xa356U,0x760fU,0xb43fU,0xedcaU,0xbe5cU,
		0xd5d0U,0x9b6eU,0x06b1U,0x2887U,0x86efU,0x3d9dU,0xfe58U,0xc5b5U,0xb0e2U,0x6451U,
		0xa220U,0xbd81U,0xdcccU,0x4c3bU,0x34c0U,0xcbf8U,0xc157U,0xfa99U,0x4961U,0x6c63U,
		0xc8cfU,0x06d6U,0x14d7U,0x3741U,0xa652U,0xb706U,0xa3bbU,0x5abdU,0x9bdeU,0x0209U,
		0x44a9U,0xe5cdU,0x34d1U,0xc472U,0x96ddU,0xdcb0U,0x259cU,0x09dcU,0x0ff0U,0x0bfcU,
		0x8232U,0x1228U,0xc52eU,0x5638U,0x6353U,0x8673U,0x83e0U,0xc3bdU,0xd144U,0xdcd6U,
		0x2671U,0x80d2U,0x2a6cU,0xf323U,0x5b23U,0x4f38U,0xa543U,0x5c49U,0x48ffU,0x2174U,
		0xc207U,0xf72dU,0xb20fU,0xeb7cU,0xa54dU,0x711eU,0x0d6fU,0xcc88U,0x997bU,0x97e6U,
		0x5296U,0x0632U,0x567aU,0x029bU,0x4fd3U,0xb0bcU,0x7cc6U,0x2e0fU,0xfb8bU,0x61e4U,
		0xb7adU,0x6a8fU,0x311aU,0x24c9U,0x3475U,0x8b85U,0xc3ecU,0x7980U,0x6019U,0xd87aU,
		0x9149U,0x8f89U,0x5df8U,0xa19eU,0x989fU,0x7680U,0x9263U,0x30f1U,0x6643U,0x75bdU,
		0x8e09U,0xcd17U,0x86aeU,0x74dbU,0x4199U,0x1763U,0xc38eU,0x2909U,0x2ca7U,0x54f2U,
		0x03aaU,0x00baU,0x640eU,0x7337U,0xab52U,0x8d28U,0x216aU,0xcfeaU,0x7441U,0xd9d5U,
		0xb35cU,0xfad1U,0xef51U,0x6efaU,0x0f76U,0x3fb4U,0xec99U,0xed02U,0xc6bcU,0xa51eU,
		0x4ef2U,0x3141U,0x43f3U,0x1846U,0x58feU,0xb5c5U,0x2fbbU,0x2a9bU,0x0485U,0xa9cbU,
		0x1299U,0x5350U,0x1689U,0x433fU,0x637dU,0x1bf9U,0xc94aU,0xbbf2U,0xeff6U,0x69bfU,
		0x85e4U,0x7523U,0xbd70U,0x4ba8U,0x280dU,0x1267U,0xaec9U,0x1b22U,0x9a82U,0x94c8U,
		0x547eU,0x05c5U,0xea37U,0x354aU,0x955cU,0xa5eeU,0xb6efU,0x8e34U,0x446bU,0xd84fU,
		0x8fc2U,0xd860U,0xa26aU,0x5846U,0xe589U,0xbaffU,0x6b70U,0x2a65U,0xec61U,0x0150U,
		0x6b77U,0xaccdU,0xea1eU,0x31d9U,0xeacfU,0x6a21U,0x56f6U,0xc06bU,0xbf0bU,0x33d4U,
		0x4d0dU,0x7faeU,0xfb81U,0x6939U,0x4b4fU,0xee6dU,0xd3e8U,0xf2b4U,0x6365U,0xbfa0U,
		0xb78bU,0xa134U,0x4042U,0xab1eU,0xa4caU,0x5310U,0xc45eU,0x8603U,0xc4c9U,0xfea7U,
		0x3f43U,0x8916U,0x0b65U,0x41e3U,0x9642U,0xe1e8U,0xb661U,0x39beU,0x8639U,0xc48dU,
		0x67dcU,0x64bcU,0x6d74U,0xc40fU,0x33a5U,0xd69dU,0x0b51U,0x7f56U,0xcdc2U,0x21e8U,
		0x5d13U,0x2441U,0x7485U,0x7199U,0x4a79U,0xd591U,0x8a1eU,0x3381U,0x7678U,0x930bU,
		0x1dafU,0x560eU,0x4d4fU,0xd09cU,0x67ccU,0xf2e0U,0x8f40U,0x5fa3U,0xe461U,0x8bfbU,
		0x5747U,0x80aaU,0x97dcU,0xa1f8U,0x59adU,0x10b4U,0x2b3cU,0xccbcU,0xd2a0U,0x1535U,
		0x69eeU,0x1289U,0x5a32U,0x19a7U,0x7b55U,0xc2fbU,0x6dbeU,0xec32U,0x3025U,0x102dU,
		0x35b2U,0x84f3U,0x8e4fU,0x42beU,0x23f3U,0x6c2aU,0x77b9U,0x351eU,0x7c11U,0x0e8aU,
		0xec16U,0xb991U,0x1ba9U,0x242fU,0x12bbU,0x1afcU,0xa432U,0xd163U,0x3c53U,0x4e06U,
		0x1dddU,0xd5eaU,0xb57bU,0x7355U,0x4778U,0x4127U,0x2c9dU,0x19c8U,0xcefeU,0xf6d3U,
		0x6e0eU,0xe892U,0x4ca6U,0x1360U,0xd0b9U,0x51d5U,0x3bf7U,0xb270U,0xbb06U,0x6ae7U,
		0x6e65U,0xe0aaU,0x4b28U,0x58d3U,0x0673U,0x6291U,0x4bedU,0x159dU,0x9811U,0xc726U,
		0xdf84U,0xfdbfU,0x5264U,0xe263U,0x6e83U,0xbf22U,0x29afU,0x2b75U,0xcddcU,0x0399U,
		0xb496U,0x5707U,0x0dadU,0x4d3bU,0x91a0U,0x4307U,0xb4d9U,0x6fdeU,0xbb4bU,0xc88bU,
		0x4e43U,0xb19fU,0xd78eU,0x072bU,0xdc3fU,0x905dU,0xd205U,0xd976U,0xee85U,0x731cU,
		0x0694U,0x3389U,0x7894U,0x0be8U,0x2783U,0x4a1dU,0x74d1U,0x489aU,0x759bU,0xea78U,
		0x0842U,0x276aU,0xed45U,0x9177U,0xd75cU,0x97b3U,0x3665U,0x6138U,0x79fcU,0x3e5bU,
		0xa0b2U,0x9008U,0x99a9U,0x85c2U,0xb16cU,0x72fdU,0x9e1dU,0xc2e4U,0xcc2bU,0x8b37U,
		0x281dU,0x20b4U,0x9579U,0x4ecdU,0x3177U,0x0af2U,0x941aU,0x40adU,0xe75cU,0x7067U,
		0x7d91U,0x2bbbU,0xffcdU,0x3982U,0x2040U,0x728fU,0x87c4U,0x6112U,0x724bU,0xa98dU,
		0xd0e0U,0xaa66U,0x78d0U,0xd942U,0xbf51U,0x51aaU,0xe2d9U,0xdebfU,0xd31bU,0xc17dU,
		0x96faU,0x4e64U,0xafd0U,0x4054U,0x8cd2U,0x7618U,0x98ebU,0xe034U,0x5fcfU,0x558eU,
		0xb962U,0xbaa7U,0x2dbcU,0x7bdaU,0x60f8U,0xee38U,0x43fcU,0x9ca7U,0x88e6U,0xeb04U,
		0x69a3U,0xcf7dU,0x6639U,0x302eU,0xa96bU,0x6f4bU,0xe590U,0xac58U,0x5b5dU,0x21b8U,
		0x70fbU,0x4813U,0x2628U,0x49deU,0xe31dU,0xde04U,0xa8c6U,0x2980U,0x4e4cU,0xf4e1U,
		0x80fcU,0xff33U,0xcb9cU,0xbf88U,0x8953U,0x57d8U,0x3909U,0x5e37U,0xf503U,0xa461U,
		0x0413U,0xb123U,0x37beU,0xdfd5U,0x763dU,0xa3e2U,0xf641U,0x9d6fU,0x7fe6U,0x901fU,
		0xe86fU,0xbc67U,0x8420U,0x76a2U,0xbcfdU,0xd058U,0x49d5U,0x82f4U,0xbdbfU,0xdc6aU,
		0xc0beU,0x851bU,0x8587U,0xbc18U,0xe7ccU,0x1e15U,0x113cU,0x9bc4U,0x817eU,0xc442U,
		0x0a38U,0xef36U,0xd870U,0xd10fU,0x60abU,0x24b0U,0x0b10U,0x79cdU,0x7832U,0xa1d4U,
		0xeaa4U,0x7a41U,0xb752U,0x9af2U,0xe4e1U,0x3d24U,0x4e9eU,0x9377U,0xde16U,0xed3dU,
		0xc386U,0x887eU,0x79a9U,0x95f1U,0xd9a5U,0x2600U,0x8ba2U,0x13ebU,0xf0d5U,0x2f93U,
		0x93abU,0xcd7cU,0x05eaU,0x1072U,0x59d1U,0xbed4U,0x1bbaU,0xf95fU,0x7b80U,0x100eU,
		0xb8d6U,0xa566U,0xcfb0U,0x1435U,0xd18dU,0xffe8U,0x2fb1U,0x6042U,0x46e2U,0x1cc3U,
		0xeadfU,0x758cU,0x4f6eU,0x73b0U,0x0921U,0xaa36U,0x3295U,0xb835U,0x5d8bU,0x44cdU,
		0x70adU,0xcf10U,0x7e0fU,0x498dU,0x025fU,0x6174U,0xa423U,0x53c2U,0x91cdU,0xccd6U,
		0x2355U,0x990cU,0xefd8U,0x42e3U,0x584dU,0x7a9dU,0x492bU,0x66f0U,0xc323U,0x549cU,
		0x8fafU,0x81ceU,0xc112U,0xf21cU,0xd5b1U,0xb743U,0xdc8cU,0x0a65U,0x2568U,0x5f93U,
		0xcdbcU,0x9556U,0x8e03U,0x2596U,0x3c77U,0x01a0U,0x40f7U,0xac6bU,0xc3a1U,0x3458U,
		0x37afU,0x5d85U,0xe7c6U,0xb95cU,0x74cbU,0xb575U,0xbbe8U,0x5946U,0x5513U,0x32c2U,
		0x1ee4U,0x252aU,0x119dU,0xf5b2U,0xeb46U,0x071bU,0xe6b6U,0x8a26U,0x4428U,0x8df4U,
		0xabfcU,0x7a69U,0x37b5U,0xc69eU,0xf0a2U,0x2f39U,0x8407U,0xf984U,0xde1eU,0xab2fU,
		0x2f83U,0xf1b4U,0x41d2U,0xd7a5U,0xba32U,0x2307U,0x1ffeU,0x7140U,0xaf47U,0xe27cU,
		0x8b0fU,0x9e3dU,0x5e6eU,0x565bU,0x26ccU,0x00d2U,0xdb2fU,0x85b1U,0xba41U,0x66e0U,
		0x5eb7U,0x77faU,0xa234U,0x4aadU,0x1449U,0xf4ceU,0x021bU,0xe193U,0x6580U,0xeb61U,
		0x4fb6U,0x2a50U,0x1424U,0x58c1U,0x7701U,0xac28U,0xfa48U,0xb6a3U,0xa10cU,0xbd50U,
		0xd8b5U,0x4cd4U,0x4abcU,0x140eU,0xebc6U,0x2721U,0x8349U,0x7742U,0xdba8U,0xf551U,
		0xe1c2U,0xef0aU,0x7437U,0xbff2U,0x5edfU,0x24eaU,0x345eU,0xdfa3U,0x2693U,0x7711U,
		0x52c9U,0x77e0U,0x1c3eU,0x5392U,0xd36cU,0x9996U,0xf6e4U,0x3b07U,0x5174U,0x627dU,
		0x1a36U,0xddefU,0xa271U,0xb603U,0x4b6aU,0x1946U,0xe24dU,0x9e08U,0xa502U,0xfb64U,
		0xd774U,0x7c79U,0xffb6U,0xa14bU,0x2a20U,0xe857U,0xe33dU,0x437fU,0x6da6U,0x73dbU,
		0xb1c5U,0xa17eU,0xba13U,0x1a4cU,0x31a5U,0x520dU,0xa144U,0xb73cU,0x2979U,0xc71fU,
		0x7fa7U,0xeb75U,0xc929U,0x1f0eU,0x6756U,0xf7dbU,0x22d9U,0xbe1fU,0x7ec2U,0x5a23U,
		0x9aafU,0xa007U,0x3faaU,0xe206U,0x684dU,0xc984U,0x1aa4U,0x4a63U,0x00a7U,0xdd7eU,
		0xc3a7U,0xe932U,0xb806U,0x74faU,0x4cfeU,0xdf31U,0xabe1U,0x265fU,0xbf7fU,0xa49eU,
		0xa11bU,0xda46U,0x6677U,0x4871U,0x9ac1U,0xe9e3U,0x882aU,0x59f3U,0x6553U,0x65f7U,
		0x6071U,0xc85aU,0xe414U,0xc3ddU,0x765bU,0xc372U,0x055bU,0x2581U,0x0c7fU,0xf1ccU,
		0x5224U,0x0618U,0xd1b0U,0xef7eU,0xa048U,0xca08U,0xfc90U,0x3f56U,0x83d1U,0x4c5fU,
		0x22ffU,0x3e53U,0x732fU,0xd2fcU,0xbeb4U,0xf82dU,0xe947U,0xb7e2U,0x111cU,0x5b0dU,
		0xa588U,0x741dU,0x5addU,0x7ab3U,0x99f2U,0xf917U,0xe08dU,0x50b2U,0x18f8U,0xc0dcU,
		0x7452U,0x1b36U,0x7d63U,0xf91bU,0xf5b8U,0xf429U,0xfc14U,0x78e4U,0x51efU,0xfa6eU,
		0x0805U,0xcaf6U,0x24e2U,0x5586U,0x3542U,0xfd7eU,0x151fU,0xebb1U,0xcce8U,0x99f6U,
		0xc072U,0x6dfbU,0xb2c5U,0xf5cbU,0x8bdaU,0xef22U,0x413eU,0x0706U,0x6addU,0x3e0eU,
		0xbfbeU,0x7749U,0x9109U,0x3ad1U,0xc247U,0x7ab7U,0xed21U,0xba7bU,0xe97cU,0xe502U,
		0x6bd1U,0x179bU,0xd612U,0x70c8U,0x4984U,0x121fU,0x4278U,0x0cd2U,0x805bU,0x3a6cU,
		0x219eU,0x5ae1U,0x98d0U,0x527fU,0xc1e6U,0x7208U,0xfa7fU,0x1312U,0x02f3U,0x274dU,
		0x2c2bU,0xfbbbU,0x543aU,0xc495U,0x8c6fU,0x3929U,0xff48U,0x7421U,0x11a3U,0x4c92U,
		0x5930U,0x9bafU,0xadb7U,0x8f6aU,0x107aU,0xff91U,0x22d2U,0x550cU,0xe868U,0xd787U,
		0x963bU,0xc3e5U,0xa2e2U,0xa58cU,0x8ceaU,0x0fd8U,0xa7e0U,0x23c7U,0xad99U,0x9b23U,
		0x515dU,0xc1dfU,0x249aU,0x3be2U,0x4624U,0x2c05U,0x9807U,0x3eccU,0x7f1aU,0x5b11U,
		0x3e87U,0x8695U,0x2cbcU,0x9c58U,0x1eb9U,0xcee8U,0x4f25U,0xca72U,0x902fU,0x5cf9U,
		0xcbb7U,0x1b92U,0x065bU,0xf3b1U,0x6541U,0x1120U,0x87e4U,0xb1faU,0x679aU,0xee4dU,
		0x5f74U,0x0908U,0x63feU,0xfc52U,0x7574U,0xa80aU,0x408fU,0xa35fU,0xaf9bU,0x3059U,
		0x07a1U,0xf91fU,0x2530U,0x2d10U,0xf8e4U,0xe42bU,0xc6f8U,0x9293U,0xbbfbU,0x2b2cU,
		0x6b4dU,0x9cddU,0xe65cU,0x45fcU,0x1877U,0xec3bU,0xd28dU,0x7d67U,0x56fdU,0x292aU,
		0xfd29U,0xb2b3U,0xd4adU,0xbca6U,0xd71bU,0xd061U,0xa7bbU,0xd21eU,0x62baU,0x8350U,
		0x4a26U,0x88a1U,0x6f2fU,0x1b3aU,0xefcfU,0x2668U,0x7497U,0xfbebU,0xdd51U,0x28d3U,
		0x8f08U,0xc80dU,0x7767U,0xd8a1U,0x7149U,0x8095U,0xb231U,0xf3e9U,0xb52fU,0x7456U,
		0x42d2U,0xe880U,0x5164U,0xe2b0U,0x66b7U,0xdea1U,0x8e6dU,0xa5c9U,0xe7dcU,0xc42fU,
		0xcaa4U,0x1053U,0xb250U,0x8de0U,0xc827U,0xc0e0U,0x1eeaU,0xd931U,0x5785U,0xdc7cU,
		0x8d77U,0x9de2U,0xe478U,0x14fcU,0x8728U,0xb106U,0x0ab4U,0xecb4U,0x18abU,0x2cecU,
		0xa446U,0x18fcU,0x7b70U,0x70dfU,0xd08bU,0xb424U,0xb4cfU,0x4667U,0xb053U,0xdeddU,
		0x68cbU,0x0963U,0xdad9U,0xd45aU,0x669eU,0x786eU,0x341eU,0x50b6U,0x8045U,0x7884U,
		0x0596U,0x1d2cU,0xb930U,0x8767U,0xf144U,0x4ae4U,0x340bU,0xf263U,0xf9bdU,0xa9fbU,
		0xa06cU,0xc5d4U,0xe2faU,0xe091U,0x236fU,0x812dU,0x88ccU,0x6f0dU,0x8dd2U,0x677dU,
		0xe900U,0xf6a0U,0xebbeU,0xd313U,0xbd68U,0xbcd6U,0xb975U,0x248dU,0xf48bU,0xcafaU,
		0x32c8U,0x816cU,0x045bU,0xaceaU,0x1d4bU,0x70d1U,0x8328U,0x5f6cU,0x9b02U,0x7a56U,
		0x884aU,0x3dc2U,0x6534U,0x6c6dU,0x745fU,0x0809U,0x6a28U,0x619fU,0x4209U,0xf190U,
		0x5aeeU,0xc543U,0x9285U,0x6f04U,0xffdfU,0x3738U,0x3db3U,0x2728U,0x4f42U,0x6ba7U,
		0xca4eU,0xfa72U,0x23a5U,0x153fU,0xc254U,0x820bU,0xd2dfU,0xd557U,0x0880U,0xd893U,
		0x4dfcU,0x5cd0U,0x7fb8U,0xbb20U,0x6ec7U,0x2082U,0x5923U,0x51f3U,0xf3a3U,0x920cU,
		0x9b2aU,0xbb2fU,0xb766U,0xa1deU,0x2c38U,0x61bbU,0x513dU,0xed14U,0xb913U,0x8641U,
		0x1044U,0xd8edU,0xcf87U,0x78e8U,0x5519U,0x4c84U,0x46f8U,0x9181U,0x2701U,0xa8e8U,
		0x05f2U,0x9f2bU,0x4131U,0xf24eU,0x123cU,0x5f5cU,0x2e6bU,0xe41cU,0x46a4U,0xfc18U,
		0x28f1U,0xd017U,0xef14U,0x6bbcU,0x27a4U,0xd2bdU,0x3a07U,0xbe3eU,0xa5b9U,0xdd88U,
		0x4631U,0x1512U,0xb5d6U,0x9e7eU,0x1b83U,0xf42dU,0x8a05U,0xf783U,0x8ca1U,0x50a4U,
		0xfff2U,0xb9d6U,0xb376U,0x5ba0U,0x3a90U,0xca62U,0x18ebU,0xbfd5U,0x1b9fU,0x72e5U,
		0xa79aU,0xf5bcU,0x1c11U,0x0d5bU,0xd728U,0x92bbU,0x8330U,0x871bU,0xb4f2U,0xe8d3U,
		0xedd8U,0x4410U,0x700bU,0x874dU,0xf9dbU,0x6008U,0x57c1U,0x99faU,0x594cU,0x0752U,
		0x2b9fU,0xf205U,0xc81aU,0x56a3U,0x9d2dU,0xcffeU,0xeb26U,0xe135U,0x4235U,0x7efdU,
		0xae7fU,0x0d4dU,0x3581U,0xccecU,0xead6U,0x9bd5U,0xdd1eU,0xbb7eU,0xcc06U,0x48ceU,
		0xcabaU,0xb923U,0xf0e5U,0x4c0eU,0x6157U,0xebcdU,0x1f69U,0xf9e9U,0x5f81U,0xebb5U,
		0x1356U,0xa75eU,0xc233U,0x47d0U,0x1250U,0x7d1fU,0xd07eU,0x8aafU,0x92a0U,0x68fbU,
		0x1f9dU,0x6490U,0x1df4U,0x1e42U,0xe4d0U,0x1523U,0x68e8U,0x87d7U,0xada0U,0xcb78U,
		0xc030U,0x2291U,0xb410U,0xad59U,0x6c5fU,0x2883U,0x45b0U,0xceaaU,0x1d39U,0x9a63U,
		0x355dU,0xfd82U,0xbbeeU,0x313dU,0x8d24U,0x75b9U,0x24c5U,0xcc84U,0x80ceU,0xdcacU,
		0x6c26U,0x1531U,0xd098U,0x7f52U,0x8912U,0xee69U,0x014cU,0x3546U,0xd93eU,0x40a9U,
		0x9004U,0x4a19U,0xc887U,0xe25fU,0xb26cU,0xd5e6U,0xa562U,0x25fcU,0xa1d0U,0xbc14U,
		0x9d6bU,0xff2fU,0x6f47U,0x558aU,0xbd4cU,0x4aa9U,0x713cU,0x7a65U,0xb571U,0x5f8fU,
		0x42dfU,0xb831U,0xb7deU,0x0614U,0xe9dfU,0xdd7aU,0x1f0aU,0x437bU,0xddebU,0x24e6U,
		0x4c30U,0xeab5U,0x42aaU,0xf664U,0xd9e3U,0xf8b0U,0xbfdfU,0x4f52U,0x6108U,0x3f9bU,
		0xf02fU,0xc03fU,0x3710U,0x51fdU,0x5e25U,0x070aU,0x7a6fU,0x2408U,0xf018U,0x60bdU,
		0x7edeU,0x2c68U,0xa701U,0x5cc7U,0x344fU,0x2e06U,0x76c1U,0x8ab9U,0x1cb3U,0x1718U,
		0x158cU,0x4142U,0x31e0U,0x7ac1U,0xfe4dU,0x2eaaU,0xe04fU,0xfa24U,0xe12dU,0x9cd5U,
		0x590cU,0x7172U,0x2c0fU,0x1415U,0x87a5U,0x1b11U,0x3c9dU,0xef26U,0xe689U,0xc5c0U,
		0x04ecU,0x0369U,0x8c32U,0xd496U,0x11d6U,0xd06bU,0x998eU,0xeafcU,0x583eU,0x6727U,
		0xb42eU,0xf7f2U,0x8b27U,0x8bdeU,0xce06U,0xd3a0U,0x23ceU,0xa9b4U,0x287cU,0xed0dU,
		0xc428U,0xfbb4U,0x953aU,0xa9bbU,0x8b51U,0x39c6U,0x7d87U,0xe30eU,0xa1a9U,0xf5cfU,
		0xac02U,0xa832U,0x73f6U,0xcc93U,0x10e5U,0xd620U,0xcd84U,0x22a6U,0xc47dU,0x49a9U,
		0xabb8U,0x71e9U,0x8b93U,0xc862U,0x9b57U,0xb2c9U,0x01d3U,0xcc4dU,0x41eeU,0xa252U,
		0xa487U,0x9c3cU,0x0231U,0x9e5dU,0x2d96U,0x5faeU,0x2adeU,0x0501U,0x078fU,0x50ddU,
		0x9742U,0x6dffU,0x8e17U,0xd7bfU,0x39d0U,0xf5f1U,0x4bb3U,0x004cU,0x9967U,0xe387U,
		0xe1ffU,0x0714U,0x82edU,0x7315U,0x64b5U,0xa197U,0x4e34U,0xc076U,0x40a3U,0x9bcfU,
		0xd011U,0x6199U,0x70d9U,0x2924U,0x0fd2U,0x6df5U,0xcdf3U,0x9691U,0x13f6U,0x1ad1U,
		0xb30fU,0x01e6U,0xcc47U,0x7abbU,0x8dfaU,0x528dU,0xb545U,0x388dU,0xe078U,0x1974U,
		0xc953U,0x9cb2U,0x4fe1U,0xfa2cU,0x24fbU,0x826cU,0x5e42U,0x6dc7U,0x2ee7U,0xc24bU,
		0xea25U,0x01ecU,0xdf5fU,0xa4bbU,0x3eebU,0x7deeU,0x611dU,0xbd06U,0x2259U,0xc66eU,
		0x439fU,0x4ac3U,0x278cU,0x6143U,0xde8cU,0x3ad5U,0x6fe9U,0xabeaU,0x67bdU,0xbdebU,
		0x2182U,0x15b9U,0x89aeU,0x2092U,0xc18dU,0x16faU,0x2007U,0xb27bU,0x587aU,0x11b1U,
		0x64d7U,0x910dU,0x8240U,0xf7a6U,0xa7e7U,0xa47dU,0x9e13U,0xee14U,0x211eU,0x1160U,
		0x646cU,0x23d5U,0xe70aU,0x8c90U,0x3486U,0x4257U,0xc797U,0x774dU,0x442eU,0xc9e6U,
		0x59feU,0xb9c4U,0x17eaU,0x1c5bU,0xf596U,0x2326U,0xfb41U,0xbe2aU,0xfc03U,0xff02U,
		0xb676U,0x16b5U,0x4f80U,0xbfc2U,0x2b8aU,0xb315U,0x74a9U,0xb840U,0x65c7U,0x2ccaU,
		0x15d3U,0x1a84U,0x0a70U,0xd7edU,0xd3bdU,0x6fbbU,0xb714U,0xc608U,0xbbd5U,0x3e12U,
		0xf2daU,0xacc7U,0xcfe4U,0xf31dU,0x90a8U,0x8a31U,0xaca9U,0x896cU,0x59edU,0xb5fdU,
		0xe276U,0x7a97U,0x7a3bU,0xac52U,0x85bcU,0x6ae1U,0x8762U,0x9dddU,0x28ceU,0x928eU,
		0xcee3U,0x8f65U,0x1796U,0xfa69U,0x5f8aU,0xcc7fU,0x7d1aU,0x569eU,0xb9d1U,0xa8e3U,
		0xd552U,0x70ccU,0x8a2cU,0x1c56U,0x15b4U,0x196fU,0x0047U,0xd61bU,0xd491U,0x2c63U,
		0x2760U,0x1b48U,0xa7f4U,0xcecaU,0x9ec7U,0xd52cU,0xb538U,0xd616U,0xacd4U,0x1ad7U,
		0xa684U,0x622dU,0x76e3U,0x1a5eU,0x977bU,0xe351U,0x268bU,0x2db4U,0xb6e7U,0x4cdbU,
		0xf890U,0x224cU,0x29d3U,0x179fU,0x5727U,0xf99aU,0x2ef0U,0x0761U,0x7babU,0xb53dU,
		0x2b97U,0xf8dcU,0xa91cU,0xa70aU,0x2f25U,0xe3f8U,0x64caU,0x0136U,0x6098U,0x6bd5U,
		0x1629U,0x3067U,0x0fdfU,0x6c91U,0x2f0dU,0x71bbU,0xd3fdU,0xf7c1U,0x35c0U,0xd3a7U,
		0xcb4cU,0xb669U,0x92f6U,0x330bU,0x1a22U,0xe506U,0xe6bcU,0x8805U,0xe21fU,0x3ce5U,
		0x9cbfU,0xd531U,0x93b3U,0x1f2cU,0x3227U,0x7d4bU,0xe6fdU,0x9b4dU,0xa0a8U,0x6079U,
		0x26e6U,0xe980U,0xb7f5U,0x13fcU,0x55f6U,0x7cffU,0x10c2U,0x367fU,0x8addU,0xf49bU,
		0xf513U,0xf2cdU,0x00f7U,0x1c20U,0x9f76U,0x1885U,0x3ef8U,0xba7fU,0x796eU,0x8942U,
		0xd664U,0x7958U,0x4855U,0x9eccU,0xc519U,0xd251U,0x15c6U,0xbd90U,0xd98bU,0x3d41U,
		0xb2e4U,0x2ed3U,0x1666U,0xed25U,0xc31aU,0xa53aU,0x9404U,0x47dfU,0x1d53U,0xf9cbU,
		0x2da3U,0x8c25U,0x5769U,0x169bU,0x5f19U,0x9a97U,0x05d3U,0x55b4U,0x76daU,0x5ae5U,
		0x0721U,0x5ea0U,0x7bb3U,0x9508U,0x6eb7U,0xcecfU,0xfe40U,0xa9aaU,0xed7aU,0xe057U,
		0xfda4U,0x4483U,0xb074U,0x5816U,0x6e50U,0x21a2U,0x6b7eU,0x9697U,0x3341U,0x7e7fU,
		0x25ebU,0x3442U,0xec02U,0x668eU,0xcd45U,0xd468U,0x8b5eU,0xd8bcU,0xcd28U,0x6187U,
		0x6ab8U,0x3a70U,0x2c50U,0x1fbcU,0xd7feU,0xcf5dU,0x5e18U,0xa7f9U,0x962bU,0xaa7bU,
		0xfc7aU,0xab8dU,0x5b3fU,0xd1ceU,0x71f6U,0x3f0eU,0x681fU,0x805fU,0xf412U,0x92e6U,
		0x8cf1U,0x0782U,0xe168U,0x9cf3U,0xfed5U,0x0947U,0xf7ffU,0xce0dU,0x91eeU,0x4578U,
		0xc4e5U,0xa773U,0x1f4fU,0x0cd6U,0xeb4cU,0x2b05U,0x82e0U,0x9530U,0xee7bU,0x1b4dU,
		0xf6c0U,0x0819U,0x49eeU,0x27e4U,0x323aU,0xdd2dU,0x4006U,0x8736U,0x87b2U,0x427cU,
		0x644cU,0xcdf9U,0x4d66U,0x8795U,0x414fU,0x4dcaU,0x3dddU,0x90f6U,0xc2f6U,0x7194U,
		0xa12fU,0xd84aU,0x2a96U,0x175eU,0x0296U,0x1223U,0xcd77U,0xc43dU,0x5e32U,0xee33U,
		0x397dU,0x2765U,0x0394U,0x19c3U,0x3f51U,0xdf2cU,0x1a47U,0x77dbU,0xeb5cU,0xf97fU,
		0x019bU,0x4988U,0x3ccbU,0x3f1cU,0xdab0U,0x4d7bU,0x6741U,0x6ee4U,0x566fU,0x5babU,
		0xbdcdU,0xe969U,0xf34cU,0xc9bdU,0xa8f3U,0x8ea1U,0xf6faU,0x2751U,0xab35U,0x8216U,
		0x5db9U,0x97cbU,0xedfdU,0xe356U,0x0431U,0xf86cU,0x4cf7U,0x8400U,0x8177U,0x083bU,
		0x747eU,0x4192U,0x161aU,0x02f7U,0xee74U,0x9cb8U,0x17e3U,0x10deU,0x24beU,0x26faU,
		0x8d70U,0xadb0U,0x725bU,0x40d3U,0x07ddU,0x24f1U,0x6999U,0xcd36U,0x4a92U,0x1316U,
		0x727dU,0xc101U,0x15f3U,0x2beeU,0x839fU,0x9780U,0xe140U,0xc4abU,0x1ae9U,0x3150U,
		0xbe79U,0xd232U,0xcd68U,0x649bU,0xaa21U,0xfa83U,0xc74eU,0x20c7U,0x5b18U,0xb2fdU,
		0x7e77U,0xa4b3U,0xbb76U,0xb2abU,0x0d99U,0xf5d6U,0x9c92U,0xe1f0U,0x9244U,0xb634U,
		0xec74U,0x720cU,0xde24U,0x89e0U,0x8c5dU,0x6506U,0x131dU,0x1a63U,0x2f9bU,0x5026U,
		0x193eU,0x21b0U,0x31d1U,0xa631U,0xea52U,0x65ffU,0xcc6cU,0xc1eaU,0xf722U,0xc959U,
		0x0248U,0x0ca5U,0x2953U,0x835bU,0x0d34U,0x25b7U,0xba9dU,0x645dU,0x5049U,0x7db4U,
		0xa955U,0x9446U,0xa815U,0x5283U,0x3650U,0xea62U,0x7222U,0x50c1U,0x3ab6U,0x76e8U,
		0x3302U,0xd8e4U,0x8debU,0x30e8U,0x2473U,0xa908U,0xbee2U,0x5f31U,0xe0f7U,0x98d4U,
		0x32f5U,0xa52aU,0x96adU,0x0498U,0xacf2U,0x9fa4U,0xab98U,0x6c17U,0x660cU,0x3100U,
		0x6389U,0xdb4aU,0x51bdU,0x448eU,0x6348U,0x7425U,0xf98aU,0xf715U,0x0769U,0xdf4bU,
		0x6f59U,0x6232U,0x1c02U,0x75c9U,0x4eddU,0x2eb2U,0x52f0U,0x4287U,0xab5eU,0xc304U,
		0x9251U,0xff4cU,0x4b56U,0x197aU,0x31c4U,0xb305U,0x98e1U,0xa5aaU,0x7cd2U,0x76abU,
		0xed57U,0x1e65U,0xf9a5U,0x5ee6U,0x0bf1U,0xf739U,0x03eeU,0x392dU,0x83ecU,0xdf3aU,
		0xa822U,0x9237U,0x2396U,0xa689U,0xe155U,0x1f37U,0x5fddU,0x213aU,0x2310U,0xe5d9U,
		0x9ed7U,0x3fdaU,0xa316U,0x8c73U,0x54d6U,0xb644U,0x7f21U,0x7758U,0x4a85U,0x91fcU,
		0xbddaU,0xaab9U,0xc613U,0xa1b0U,0x96d2U,0x44e1U,0xa3f0U,0x3618U,0x4a03U,0xc499U,
		0x840dU,0x641dU,0x7b61U,0xca38U,0x4a99U,0x1adcU,0x4171U,0x951fU,0x84a4U,0x8c07U,
		0xff6bU,0xaee7U,0x03c2U,0xc27cU,0xc679U,0x543eU,0x4e5aU,0xe07eU,0xe4aaU,0x752fU,
		0x9c47U,0x6953U,0xdaa3U,0x6debU,0xb6fbU,0xdd3bU,0x07eaU,0xc348U,0xb0cdU,0x8a9dU,
		0xe809U,0xfbbfU,0x1194U,0x9140U,0x7accU,0xaf2dU,0x5ce3U,0xacd9U,0x2d00U,0xedf0U,
		0x08b1U,0xf66fU,0x1090U,0xc7aaU,0xd23fU,0x085dU,0xd3f4U,0x2c2fU,0x1febU,0xfb9aU,
		0x842cU,0xc54cU,0x2e86U,0xc353U,0xdbffU,0xf84dU,0xf9f2U,0xc44eU,0x62e8U,0x3c35U,
		0x0f1fU,0xfcf2U,0xa1bbU,0xff95U,0x2f3fU,0xe107U,0x53a7U,0x3938U,0x4eaaU,0x17a4U,
		0xc6b1U,0xc08eU,0x4499U,0x5d48U,0xf122U,0xb879U,0x7075U,0x496bU,0xf3f2U,0x107eU,
		0x6940U,0x3893U,0x20d2U,0x7af1U,0x55bdU,0xbab3U,0x0a03U,0xef5bU,0x3d6cU,0x0cb0U,
		0x4730U,0xbff9U,0x5b2dU,0x3009U,0xbf5dU,0x8f6eU,0x99b7U,0xd2aaU,0x6dd0U,0x57e4U,
		0x6eefU,0x29d8U,0x1cf8U,0x5f48U,0x5b69U,0x2c71U,0x4d17U,0x89a4U,0xf4f9U,0x18d9U,
		0xe556U,0xadbbU,0xf3d4U,0x970bU,0x3ed3U,0x9d19U,0xea86U,0x783cU,0x0866U,0xd94eU,
		0xd76cU,0xe315U,0xbf03U,0x964aU,0x9426U,0x76f1U,0x69adU,0x9bb3U,0xf0a8U,0x3e67U,
		0x167eU,0xed4fU,0xcd3dU,0x2251U,0xf0ddU,0xa7b3U,0x4782U,0xeb8bU,0x78a0U,0xd827U,
		0x683dU,0x655bU,0xa16dU,0x5934U,0x981dU,0xb54bU,0xe1ccU,0x86d2U,0x0321U,0xd81eU,
		0x9b44U,0x5f53U,0xd79cU,0x00b1U,0xc2afU,0x4313U,0x17afU,0x220dU,0x0976U,0x4c96U,
		0x183bU,0xdbdaU,0x8d05U,0xdbcfU,0xd197U,0xf895U,0x12c7U,0x95b6U,0xa3d8U,0x2a38U,
		0x3e72U,0x556cU,0xb5b4U,0xe89eU,0xa8b6U,0x11a7U,0x4b36U,0x1fdeU,0xb1e8U,0x1952U,
		0x0463U,0x8a5bU,0x3cbcU,0x8198U,0xb1d1U,0x3cf3U,0x4d24U,0x68baU,0x4910U,0xd744U,
		0x387aU,0xa590U,0xc6a4U,0x07bfU,0x2ef8U,0xd38eU,0xa287U,0x4ce0U,0x19f7U,0x192fU,
		0x34bdU,0xfe55U,0xedc7U,0xa6d7U,0xc006U,0xf3c9U,0x346aU,0xa2e6U,0xfb88U,0x5293U,
		0xb20cU,0x5b20U,0x83ddU,0x0fedU,0x44a6U,0x14d4U,0x1686U,0x58fbU,0xec96U,0x743eU,
		0x03a7U,0x86abU,0x989cU,0xc3e9U,0xc45bU,0x6362U,0x4d0aU,0xea1bU,0xe586U,0xb6ecU,
		0x9a7fU,0x85e1U,0x35afU,0x5a2fU,0x59aaU,0x8f3dU,0x7675U,0x5d10U,0x6d71U,0x963fU,
		0x919dU,0x29acU,0x980eU,0x6e62U,0x4ca3U,0x4775U,0xa42fU,0x7c0eU,0x7248U,0x7d8eU,
		0x9576U,0xb169U,0x3662U,0x7598U,0x0691U,0xd78bU,0x37bbU,0x8950U,0xa8c3U,0x5b5aU,
		0x69a0U,0x2db9U,0x8ccfU,0xe2d6U,0x1bb7U,0xf0d2U,0xc383U,0xb74fU,0x60a8U,0x1139U,
		0xbdbcU,0xe86cU,0x37acU,0x8e00U,0xd5aeU,0x4928U,0x91caU,0x70aaU,0x4f6bU,0xd18aU,
		0x76feU,0x0218U,0xba3eU,0x8b0cU,0x41cfU,0xf09fU,0xe6b3U,0x5510U,0x2976U,0xb1c2U,
		0xffb3U,0x4b67U,0xf6e1U,0x2690U,0xe1bfU,0x4ab9U,0x732cU,0xa045U,0x0558U,0x6550U,
		0xa118U,0xb803U,0x684aU,0x22d6U,0x4275U,0xe979U,0xbfbbU,0xb2c2U,0x353fU,0xfc11U,
		0x18f5U,0xa585U,0xcbb4U,0x2cb9U,0x4621U,0xa7ddU,0xe865U,0x592dU,0x5437U,0xc1e3U,
		0x7146U,0x7494U,0x62b7U,0xfd26U,0xe659U,0xf8e1U,0x408cU,0x6797U,0xf9baU,0x0593U,
		0xdad6U,0xd088U,0x0ab1U,0x5782U,0xcaa1U,0x5161U,0x7fb5U,0xc251U,0x3db0U,0x4206U,
		0x8847U,0x0458U,0xbd65U,0x88c9U,0x18e8U,0x8c9eU,0x462eU,0xef11U,0x1239U,0x46f5U,
		0xb910U,0x9b27U,0x1f9aU,0xc230U,0x6154U,0xdd1bU,0x4232U,0x2b9cU,0x7008U,0xd725U,
		0xe9dcU,0x7139U,0xa1cdU,0x9001U,0xd095U,0x8d21U,0x45adU,0xad9dU,0x583bU,0x04e9U,
		0x2c0cU,0xfe4aU,0x76beU,0xf015U,0xf02cU,0x42a7U,0x82eaU,0x39cdU,0x2adbU,0x41ebU,
		0xabb5U,0x73f3U,0x8b4eU,0x23cbU,0x2004U,0x67baU,0x439cU,0xdf5cU,0x24f8U,0xb542U,
		0x13f3U,0xd00eU,0xe273U,0xcfe1U,0xd3baU,0x74a6U,0xfc00U,0x59fbU,0xe707U,0xa7e4U,
		0x2f22U,0x2eedU,0xb6e4U,0xa681U,0xa7f1U,0x15b1U,0x7d17U,0x28cbU,0xd988U,0xd661U,
		0x00f4U,0x55f3U,0xe6faU,0xe21cU,0xcb49U,0x0fdcU,0x5b3cU,0xd7fbU,0x8b5bU,0x333eU,
		0xfda1U,0x7bb0U,0x5f16U,0x9401U,0x1a44U,0x5e2fU,0xa12cU,0x4d63U,0x3237U,0x82ddU,
		0x91ebU,0x8ceeU,0xbe76U,0x15f0U,0x07daU,0x17e0U,0x8174U,0x5db6U,0xf349U,0xdaadU,
		0x2470U,0x721fU,0x5046U,0x0245U,0x31ceU,0x8c5aU,0x9c8fU,0x5b15U,0x230dU,0xa81fU,
		0xf9a2U,0x31c1U,0x52edU,0x0766U,0x6386U,0x96aaU,0x108dU,0x7ac9U,0x07e7U,0xe4a7U,
		0xff68U,0x7b5eU,0x96cfU,0x7f1eU,0x4d14U,0x6dcdU,0x472dU,0x20cfU,0xf11fU,0x53a4U,
		0x62e5U,0x8429U,0x3e6fU,0x8d02U,0xc2acU,0xe1c9U,0x789dU,0x167bU,0xbf00U,0x3ed0U,
		0x59a7U,0x4d07U,0xec93U,0xb209U,0xedc4U,0x2ef5U,0x4d21U,0xb1e5U,0x0555U,0xffb0U,
		0xba3bU,0xd5abU,0xc380U,0xa8c0U,0x9573U,0x980bU,0xa1caU,0x6151U,0x462bU,0x3dadU,
		0xdad3U,0x62b4U,0x461eU,0xbfb8U,0xa129U,0x8b58U,0x00f1U,0xb6e1U,0xd3b7U,0x4399U,
		0x2ad8U,0x2c09U,0xba38U,0xec90U,0xc2a9U,0x472aU,0x07e4U,0xf99fU,0x5043U,0x07d7U,
		0x8462U,0xe00eU,0x3f35U,0x11fcU,0x3f32U,0xc2a6U,0x00eeU,0x4628U,0x5237U,0x5e48U,
		0x11ffU,0x3a45U,0xc992U,0xc12eU,0x6a50U,0xb502U,0xbb5bU,0x6d66U,0x3f38U,0x253fU,
		0xd02fU,0x5888U,0xc99fU,0x3be6U,0xae4cU,0x5ddeU,0xe011U,0x0edaU,0xf754U,0x572cU,
		0x7bf4U,0xc68bU,0x5739U,0x80b9U,0x8465U,0x26aaU,0x08e6U,0x37ffU,0xa0fbU,0x249eU,
		0xe545U,0xb8b3U,0xb505U,0x5af7U,0x9c69U,0x9f4cU,0x80eeU,0x1828U,0x1f72U,0x7024U,
		0x6a53U,0xc5e9U,0xc96bU,0xd7d2U,0x4330U,0xca76U,0xd7abU,0xcc5bU,0xc131U,0xa2f9U,
		0xd586U,0x6bdaU,0x8032U,0xe2cbU,0x026dU,0x0c4bU,0xc995U,0xbbcbU,0xd99bU,0x02c2U,
		0xb23aU,0x4f29U,0x33dbU,0x8272U,0x3a48U,0xf4baU,0x05dcU,0xa2b0U,0xe62bU,0xdc57U,
		0x5417U,0x69ceU,0x1202U,0xf558U,0xc400U,0x9831U,0x0e29U,0xceecU,0xc7f2U,0xae2dU,
		0x5e4bU,0x96beU,0x1cd1U,0x609dU,0x1da4U,0x7d2fU,0x4e16U,0x7ee7U,0x523aU,0xf153U,
		0x6d2bU,0x8e25U,0xf389U,0x1ebdU,0x3c85U,0x4565U,0x3be9U,0xeeb8U,0xb51eU,0x95adU,
		0xd248U,0x5b97U,0x2a47U,0xa9c2U,0xc9a2U,0x0b6dU,0xa663U,0x3abfU,0x1643U,0x9c5cU,
		0x41d8U,0x0bbcU,0x588bU,0x29a1U,0x40daU,0x013bU,0x8baaU,0x7f93U,0x3d1aU,0x3257U,
		0xd032U,0x032aU,0xc5f6U,0x8832U,0xc124U,0x2cc0U,0xd1f7U,0x2501U,0x2542U,0xdeecU,
		0x5866U,0x7404U,0x4e23U,0x6d51U,0xc921U,0xf639U,0x3f3bU,0xddbaU,0x54aaU,0x540aU,
		0x14a9U,0x8699U,0x69dbU,0x956bU,0x6d69U,0xa30eU,0x6817U,0x64cfU,0xe4c8U,0xb527U,
		0x55daU,0xc5a6U,0xbb5eU,0xf815U,0xf37cU,0x65a4U,0x53f8U,0x3e8bU,0xc16cU,0x710cU,
		0xc68eU,0xa378U,0x32d0U,0x2a03U,0xc710U,0xcc16U,0xd0c9U,0xa86cU,0x7bf7U,0x8d86U,
		0x9767U,0x8c40U,0x0099U,0x1124U,0xf1baU,0xc849U,0x572fU,0x59e3U,0x59c7U,0xe3fdU,
		0x2d59U,0x65e8U,0x6a7cU,0x31e8U,0xf757U,0x4e88U,0x84e8U,0x7f75U,0x853cU,0x6545U,
		0xbf12U,0xfa32U,0x0eddU,0xc756U,0x73a1U,0xaaa3U,0xebe4U,0xfe74U,0xf0b0U,0x1783U,
		0xe014U,0xdbafU,0x1c30U,0xf622U,0xaf80U,0xf3b5U,0x0ec3U,0x519aU,0x5de1U,0x5c38U,
		0x9dc5U,0x2f2aU,0x1b63U,0x0a4bU,0xf336U,0x5a55U,0xae4fU,0x3820U,0x73c5U,0x296bU,
		0x69e3U,0x065fU,0xfa3fU,0x6374U,0x24a1U,0x56bbU,0x45e5U,0xb3b8U,0xa5deU,0x9db8U,
		0x1560U,0x9541U,0xa0feU,0x886aU,0xf373U,0xb5abU,0xb2dbU,0x1b96U,0x2f89U,0xbe89U,
		0x3802U,0xdeb0U,0x7262U,0xa70fU,0x0ed0U,0xf313U,0xccb1U,0x8116U,0x08e9U,0xf094U,
		0xe28aU,0xc56cU,0x95c9U,0xcbbbU,0x4ebeU,0x4fe7U,0x26adU,0xd47fU,0x3d8dU,0x6a6fU,
		0x002bU,0x38c4U,0x8fa1U,0xcb2eU,0x8468U,0xd64bU,0xa385U,0xce73U,0x6b00U,0x5cfdU,
		0x1457U,0x2b66U,0x80bcU,0xbe9fU,0x008cU,0xa921U,0xf109U,0xca1eU,0xd836U,0x4f60U,
		0x573cU,0xb296U,0x0b2dU,0x6326U,0x3a59U,0x9033U,0x444fU,0xde51U,0x182bU,0x8482U,
		0x1d67U,0xf78fU,0x8d99U,0x3d58U,0x30a0U,0xbef5U,0x80f1U,0x9490U,0x0d87U,0xfd10U,
		0x28fdU,0xfc56U,0x00d8U,0x793eU,0x9f4fU,0x918dU,0xfabbU,0xf7c6U,0xfe86U,0x7bc6U,
		0x9b36U,0xfbcdU,0x9c6cU,0x6875U,0x0f37U,0x245eU,0x9044U,0x6402U,0xab25U,0xbd0cU,
		0x5afaU,0x792dU,0x94f4U,0xef90U,0x7f89U,0x1f22U,0x4808U,0xd563U,0xb508U,0x5399U,
		0x8b79U,0xf57aU,0x6c4dU,0x090cU,0xe334U,0x4b46U,0xb8b6U,0x1fcbU,0xe90cU,0xd402U,
		0x9f04U,0xa856U,0x505dU,0xa739U,0xe548U,0x5538U,0xaf77U,0x9893U,0x6aafU,0x5f78U,
		0x45bcU,0x3626U,0xca79U,0x3fc4U,0x06c3U,0x147eU,0x0eb3U,0x46ccU,0x99c8U,0x22adU,
		0x4333U,0x7f5eU,0xc9f6U,0x054aU,0x0408U,0xee51U,0x26d2U,0x0f4fU,0xd7d5U,0x2ff4U,
		0xc4b2U,0x71c0U,0xa56eU,0xe535U,0x034eU,0x8eb9U,0xc96eU,0x43e7U,0x8cc1U,0x0620U,
		0x363aU,0x679eU,0x4884U,0x6123U,0xc5ecU,0xa09eU,0xba28U,0x3eb1U,0x2d6eU,0xccf8U,
		0xa512U,0x3797U,0x6a56U,0x6b8dU,0x57cdU,0x03d0U,0xafb9U,0xb1feU,0x43d2U,0x1e4eU,
		0x7027U,0x7096U,0x9b87U,0x2f12U,0x6cf1U,0x5fefU,0x93c5U,0xef45U,0x1f75U,0x308eU,
		0x5987U,0x6623U,0x386cU,0x87e8U,0xf30bU,0x07cfU,0xe2ceU,0x9517U,0x0811U,0x231eU,
		0xdca4U,0x14f4U,0x09b4U,0x312cU,0x8035U,0x7eaeU,0x93eeU,0x3ea2U,0x0a95U,0xf923U,
		0x5661U,0xe4f0U,0x6bddU,0xd416U,0xbc3fU,0x6c96U,0x4348U,0xa646U,0x2657U,0xef2eU,
		0xd589U,0x98c0U,0x9884U,0x19b0U,0x8f24U,0x07a5U,0x4049U,0x7df4U,0xa2fcU,0x7214U,
		0x8667U,0xf27fU,0xfebdU,0xbc70U,0x593cU,0x5ef4U,0xc134U,0x1c45U,0x3392U,0x01b3U,
		0x455bU,0x305dU,0x82bdU,0x7503U,0xcc5eU,0xea09U,0xaa9aU,0x0fe4U,0x3439U,0x48c5U,
		0x9dffU,0x6a18U,0xd7aeU,0x057eU,0xeda9U,0x5bddU,0x3e93U,0xaf9fU,0x8d3fU,0x88adU,
		0x4f2cU,0xfdd7U,0x6f6cU,0x52e2U,0x60edU,0x13a4U,0xe8b5U,0xcd8bU,0xb23dU,0xdeffU,
		0x3f87U,0x89cfU,0xe679U,0xa363U,0x5e74U,0x3365U,0x02c5U,0xc1fbU,0xe147U,0x306cU,
		0xb030U,0xf92bU,0x521dU,0xa22dU,0xd99eU,0x9413U,0xd299U,0x4eebU,0x06aaU,0x4093U,
		0x028fU,0x3ef1U,0xbbceU,0x973bU,0x6f40U,0x1b7cU,0x3417U,0x7f13U,0x6af9U,0x14a2U,
		0xc998U,0xcb42U,0xe6acU,0x096fU,0xe802U,0xa80eU,0x0ac0U,0x310eU,0x0c4eU,0x12a9U,
		0xd4d1U,0x162eU,0xe7fbU,0xafb2U,0x8ef2U,0x7a22U,0x0270U,0x105fU,0x8250U,0x643cU,
		0x2b55U,0x7578U,0xf2a3U,0x282eU,0xdc5aU,0xf199U,0xce35U,0xcd9aU,0x4fa7U,0x806bU,
		0xebd6U,0x4effU,0xe62eU,0x415eU,0x6ac4U,0x132bU,0x35eeU,0x2b30U,0x9e43U,0x5822U,
		0xa2b3U,0xa97dU,0x5a90U,0xe50bU,0xcad7U,0x2559U,0x20edU,0x7be9U,0x05dfU,0xda71U,
		0x206eU,0xfaa3U,0x809eU,0xbbffU,0xa13bU,0xa4c1U,0xf4bdU,0x7690U,0x9aa0U,0x9270U,
		0xfe09U,0xfadbU,0x38a4U,0x3bb3U,0x3a4bU,0x77e7U,0xf32dU,0x35a6U,0xfc71U,0x9297U,
		0xac1eU,0xccc6U,0x8275U,0x7344U,0x3b60U,0x1a27U,0xf80bU,0x3d37U,0x1003U,0x60c6U,
		0x33deU,0x37f4U,0x127dU,0x27f0U,0x5aa2U,0xc6fcU,0x43b2U,0xe2e8U,0xceefU,0x9674U,
		0xe772U,0x79d7U,0xc7b3U,0x61f1U,0x4517U,0xd627U,0x0e2cU,0x0b59U,0x8024U,0x50caU,
		0xeb0eU,0xe42fU,0x8b15U,0xb21cU,0x9834U,0x2774U,0x9787U,0x3310U,0xb8deU,0x830eU,
		0x735fU,0xfd63U,0xc403U,0x86dbU,0xce60U,0x522cU,0xea99U,0xf8e8U,0x2619U,0xdf65U,
		0xf55bU,0xa8a4U,0xecf2U,0x8811U,0x7c59U,0x5006U,0xbd38U,0x4a6dU,0x1205U,0x8680U,
		0x3f70U,0x2d27U,0x1635U,0x2d14U,0x6704U,0xb8c7U,0x69d1U,0x7964U,0x143fU,0x92fbU,
		0xf687U,0xeec1U,0x5602U,0x25c5U,0x541aU,0xa6a2U,0x8873U,0x9bebU,0x6b11U,0x2534U,
		0x4610U,0xb344U,0x7d32U,0x71a6U,0x7467U,0x5c8fU,0xfb4dU,0xb77fU,0x666fU,0x0b08U,
		0x1da7U,0xf05dU,0x5bbdU,0x23e1U,0x4033U,0x7d6bU,0xe282U,0x3f2aU,0x60a0U,0x03baU,
		0x3ffeU,0xb66eU,0x890aU,0x18a3U,0xc8eaU,0x3ca5U,0x1cd4U,0x6fc7U,0x74b9U,0x8af4U,
		0x9f8aU,0xd291U,0xb792U,0x01f2U,0x96c1U,0xec7cU,0x5c2fU,0xea12U,0xcf54U,0x47c7U,
		0xa175U,0xbf97U,0x5e4eU,0x52d0U,0xa3acU,0xa889U,0x5541U,0xec3fU,0x1e0aU,0x3844U,
		0xae30U,0xec85U,0x33c8U,0xcb51U,0xfc32U,0x3ae1U,0x77f0U,0x5754U,0xc7f5U,0xfee4U,
		0x64e3U,0xab43U,0x5400U,0x187bU,0xe36bU,0x1d7cU,0x1ec0U,0xa25eU,0x5f08U,0x28a4U,
		0xdaf9U,0xc057U,0x39dcU,0x10ecU,0xf38cU,0x7ca8U,0x67edU,0xf003U,0x90d4U,0x4600U,
		0xaf4dU,0x8fdbU,0x8e28U,0xfccaU,0x83a6U,0xd3acU,0xc0b3U,0x0a9dU,0x27bbU,0x0072U,
		0x6d2eU,0xa5fbU,0x9f92U,0x9c1fU,0xac0eU,0xe660U,0xda10U,0xea2bU,0xf156U,0x0828U,
		0x11e2U,0x1601U,0xb902U,0x6249U,0xdc1eU,0x37dcU,0x523dU,0xc646U,0xd127U,0x027fU,
		0xd4d8U,0x9ce1U,0x4b9cU,0xe59fU,0x7eeaU,0x6f7eU,0xd04dU,0x35c5U,0x9fe0U,0xbfadU,
		0x994eU,0xf8bcU,0x4e19U,0x8ad3U,0x26c2U,0xb6c5U,0x676bU,0x6b51U,0xb82bU,0x8aa9U,
		0x5b9aU,0xd88dU,0x4adeU,0xc807U,0xca6cU,0x70c2U,0x19bdU,0xaa75U,0xd24bU,0xf8d6U,
		0x8966U,0x208cU,0xe381U,0xd065U,0x4ab3U,0x85dbU,0x95b0U,0x5f42U,0xedeaU,0x1f31U,
		0xd8deU,0xc4a5U,0xca18U,0x0a45U,0xb521U,0x7d29U,0xc685U,0xb1dfU,0x93fbU,0xd71fU,
		0xbfa7U,0x3adbU,0xeebbU,0x3d31U,0xafacU,0x48bfU,0x5fe9U,0xa850U,0x6e95U,0xca96U,
		0x3becU,0x7718U,0x4da5U,0x1e71U,0x1e9fU,0xbcaaU,0x52fcU,0xfe63U,0x4568U,0xb7bdU,
		0x436cU,0x93b8U,0xd34fU,0xbcb0U,0xb045U,0xaaceU,0x3c88U,0xdb56U,0x70e9U,0x133fU,
		0x0d17U,0xd4b1U,0x36a6U,0x2dc9U,0x9c5fU,0xced9U,0x2fa7U,0xeca4U,0xe8c2U,0xbe04U,
		0x2b0fU,0xcc9aU,0x1646U,0xd6a5U,0x0e7fU,0x9f13U,0x5e03U,0xb2b7U,0xa23aU,0x3186U,
		0x3ac2U,0x0d7aU,0x2bf5U,0xd536U,0x1016U,0xdeccU,0x34dcU,0x8367U,0xa666U,0x0a17U,
		0x36ebU,0xf1d4U,0x7d55U,0xfd2dU,0x9ef7U,0xde92U,0x0b70U,0x4d87U,0xf7b0U,0x394bU,
		0x78b7U,0x6480U,0x5dc5U,0x8f4bU,0xc9a5U,0x6d8eU,0x8504U,0xcba9U,0x12b0U,0x292eU,
		0xec2cU,0xab18U,0xa9c5U,0xb0b6U,0xbd7bU,0x9cc4U,0xbcc6U,0x1ea5U,0x4c59U,0x5207U,
		0x2a4aU,0xac65U,0x106cU,0xa45bU,0x7289U,0x5701U,0x4d30U,0x1147U,0x7f96U,0x712eU,
		0x6c75U,0xd0edU,0xcf3dU,0x695dU,0x53b8U,0x4e78U,0x8badU,0xa6e3U,0x1984U,0x38dcU,
		0x0c41U,0x1b3eU,0x7800U,0x9450U,0x013eU,0x0affU,0xa863U,0x3ceaU,0x1692U,0x4407U,
		0xf548U,0x1b19U,0x40ddU,0xaf60U,0x3914U,0x5a3bU,0x55a5U,0x6f33U,0x636cU,0x6149U,
		0x29a4U,0xb63cU,0x92deU,0xf79eU,0x87cfU,0xe878U,0x6563U,0xa0bfU,0x588eU,0x269aU,
		0x069dU,0x3670U,0x1d74U,0x88a5U,0xb4e4U,0xb80fU,0x0bbfU,0xd433U,0x8655U,0xe224U,
		0x2c7bU,0x1e77U,0xf7d3U,0x9017U,0x41dbU,0x3c02U,0x0127U,0xe8a8U,0x65acU,0x4a2aU,
		0x4dc3U,0x3678U,0x2cc3U,0x9c35U,0xe258U,0xd2b6U,0x4660U,0x9b1cU,0x6a68U,0x73fdU,
		0xc127U,0x15aaU,0x70a3U,0xd817U,0x694cU,0x8354U,0x5ebdU,0xde62U,0x8835U,0x3944U,
		0x15faU,0x880aU,0x1b75U,0x3eaaU,0x8438U,0x22e4U,0xc5f9U,0xed9cU,0x8afcU,0x0dbaU,
		0xc57dU,0x62beU,0x0f5cU,0x2792U,0x032dU,0x32a0U,0x56dbU,0x883cU,0x4b1dU,0xd170U,
		0xdc97U,0xf8a1U,0xd035U,0xf17cU,0x46bbU,0xf23bU,0x0c55U,0xd222U,0x25d9U,0x4cb3U,
		0x325aU,0xa719U,0x0872U,0xe6c1U,0x45c9U,0x4dabU,0xc853U,0x5654U,0x3d1dU,0xc2ddU,
		0x8bf4U,0xfacaU,0xdb3aU,0xa7bfU,0x52c0U,0x4462U,0x6d54U,0x4212U,0x2dd3U,0xd6d5U,
		0x3609U,0x3b53U,0x6160U,0x57f2U,0x4e26U,0x3ff5U,0x59beU,0xa27eU,0x6eaeU,0xc811U,
		0x66e6U,0xda61U,0x7407U,0x40c1U,0x2855U,0xe985U,0xf4b0U,0x6223U,0x0bddU,0x38e6U,
		0x5869U,0x73e8U,0xc653U,0xda2dU,0x7152U,0x8f0cU,0xf2bbU,0x4ac9U,0xdeefU,0xe98fU,
		0x5f22U,0x20e0U,0x27c8U,0xe0c0U,0xe11fU,0x6523U,0x2545U,0xdfaaU,0x9d06U,0x7036U,
		0xbb92U,0x28d7U,0x10f9U,0x4bfcU,0x2504U,0x307eU,0x35e1U,0x26ebU,0x7b2cU,0x771eU,
		0x064aU,0xf021U,0xd1faU,0x5cb3U,0x778dU,0xe29dU,0x1379U,0xdd55U,0xf06dU,0x01bdU,
		0x869cU,0x3f63U,0xec4cU,0xb479U,0x1099U,0xad13U,0x8edfU,0xa839U,0x14acU,0x33adU,
		0x0443U,0x722bU,0xc789U,0xfbefU,0xba47U,0xe115U,0x540dU,0x2681U,0xc108U,0x607eU,
		0x7b88U,0xef9aU,0x84c0U,0x625fU,0x54adU,0xe1d5U,0x612dU,0x0c87U,0x6d1aU,0x749bU,
		0x495cU,0x43a5U,0xddbdU,0x75e5U,0x3785U,0xde69U,0xf8c6U,0x9392U,0x23eeU,0x4d4aU,
		0x3f3eU,0xec5cU,0xc945U,0xab4dU,0x3115U,0x266cU,0xb8d1U,0x782dU,0xf63cU,0xa966U,
		0x78cbU,0xa0adU,0xbb46U,0x3bf2U,0xe942U,0x9abcU,0xc924U,0x1a31U,0xa107U,0x1ff9U,
		0x74c6U,0xefd3U,0xe2f5U,0xa441U,0xb52aU,0xd288U,0x653cU,0xa2ddU,0x2199U,0xcce3U,
		0xdde6U,0x0147U,0xe4cbU,0x357cU,0xa795U,0x469fU,0x591eU,0x745aU,0x85b7U,0x4f7bU,
		0x64d2U,0x2ee2U,0x4e2fU,0x9b52U,0x8b22U,0x1587U,0x0196U,0x87adU,0x681aU,0x6e4bU,
		0x1661U,0x26e1U,0x6093U,0xb533U,0xd3efU,0xc674U,0xa311U,0x924cU,0xe0f2U,0xcc67U,
		0xaa1cU,0x1615U,0x6845U,0xbdb7U,0x6d6cU,0x3465U,0xa8b1U,0xa168U,0xe551U,0xf3edU,
		0xa0f6U,0x00e9U,0x956eU,0x96caU,0x91e6U,0xe702U,0x45a8U,0xca9cU,0x3a54U,0x95c4U,
		0x69deU,0x8537U,0x53f3U,0xc11fU,0xf384U,0xb235U,0x2b50U,0x06a5U,0x3e8eU,0x8f1fU,
		0x3867U,0x3635U,0x6aaaU,0x903fU,0x6766U,0xac09U,0x53fbU,0x9f85U,0x6b0cU,0xea94U,
		0x5a9dU,0x8099U,0xdb35U,0xc578U,0x65a7U,0x55a0U,0x7284U,0x7d50U,0x0d12U,0x93f6U,
		0xcda7U,0x136fU,0xf37fU,0x608eU,0x74c1U,0x6d15U,0x1374U,0x714dU,0xe020U,0x225fU,
		0xf818U,0xdcc0U,0x0386U,0x5ec4U,0xaf13U,0x62cfU,0xc55dU,0x4b8fU,0xbb61U,0x382cU,
		0x20a2U,0x81bdU,0x0ac7U,0xd8a5U,0x1ce4U,0x6b2eU,0xc5a9U,0xb6b0U,0xfae8U,0x322cU,
		0xd0d5U,0x6e9bU,0xb8f5U,0x29f4U,0x55ddU,0x7913U,0xd965U,0x31f4U,0xcdacU,0x776bU,
		0x6201U,0x39afU,0xcc19U,0x458aU,0xaa0cU,0xab00U,0x34e8U,0xb381U,0xb696U,0x13ddU,
		0xc713U,0x57a6U,0x270cU,0xcca6U,0xbf74U,0xa5cdU,0x2a56U,0xc25fU,0x2a06U,0x5d23U,
		0xe491U,0xf4a0U,0xfb1dU,0x5c68U,0x79bfU,0x66d3U,0x32d3U,0x6d9aU,0x2aeeU,0x5c7aU,
		0x62d4U,0x8e71U,0x891dU,0x2098U,0xa37bU,0x2a8cU,0x0c75U,0x9d47U,0x4650U,0x4502U,
		0xe84bU,0xb612U,0xc691U,0xddf6U,0x74efU,0xdbeeU,0xb89dU,0xdea5U,0x737eU,0x1913U,
		0x710fU,0xf35dU,0x6cc5U,0x8ae2U,0xeb31U,0xbcb6U,0x991cU,0xa4f6U,0xc16fU,0xf5a5U,
		0x0d23U,0x1fa8U,0xaf18U,0x66bbU,0x7c49U,0xf44dU,0x1127U,0x32acU,0x9346U,0x5c00U,
		0x16dbU,0x089bU,0x3266U,0xb2d0U,0x009cU,0xc417U,0x548fU,0xb191U,0x5386U,0xe2b4U,
		0x4fbcU,0xc0a3U,0x8c43U,0x02b0U,0xfa8aU,0x3684U,0xf967U,0x12e7U,0x5973U,0xe720U,
		0x976aU,0x3b30U,0x0ae4U,0x2589U,0x5ec9U,0x5168U,0xc3d3U,0x89b4U,0x8d89U,0x116fU,
		0x2c87U,0x62c5U,0x38baU,0x90ecU,0x5a18U,0x63edU,0x7bfaU,0xb3adU,0x4cc8U,0x524cU,
		0x757fU,0xe884U,0x83c8U,0x6db5U,0xa86fU,0x3429U,0x16ccU,0x10c7U,0x2fd8U,0xd355U,
		0xb25bU,0xd0f9U,0xd0ccU,0x9fd7U,0xf100U,0xe1b6U,0x038bU,0x42d6U,0x4a56U,0xdf95U,
		0x65ebU,0x3a23U,0x3dcaU,0x0081U,0xfc85U,0xdd61U,0x0931U,0x6033U,0x2d5cU,0x90b7U,
		0xbb9eU,0xfdafU,0x0e5fU,0xc0e4U,0xeb67U,0xda39U,0xe400U,0x764fU,0x233cU,0x7d04U,
		0x5359U,0xd912U,0x22f8U,0x141dU,0x59caU,0x87bdU,0x301eU,0x1292U,0xdcc5U,0xc82bU,
		0x3f4aU,0x15bfU,0x59e6U,0xe1f8U,0xb7d7U,0x3a89U,0x3404U,0x9028U,0xd82fU,0x55d3U,
		0x5732U,0x1a3dU,0x7325U,0xa3d1U,0x08aaU,0x8de4U,0xb8eeU,0xe93bU,0xc84cU,0x4c52U,
		0x9947U,0x55fbU,0x8eebU,0x93beU,0x3754U,0xe3c1U,0xf1bdU,0xaefeU,0x4860U,0x626bU,
		0xf81dU,0xb254U,0xb28eU,0x26a2U,0x6548U,0xc7a2U,0x77d3U,0x7a8fU,0xdd72U,0xa9f3U,
		0xb71fU,0x9b5eU,0x853fU,0xb3e8U,0x790bU,0xc2d5U,0x8acbU,0x1057U,0x6586U,0x4740U,
		0x7f78U,0x2cd9U,0xaa28U,0x1401U,0x9da3U,0x7eb6U,0xd2f4U,0xf957U,0x84ebU,0xcbf0U,
		0x6fcfU,0x39ebU,0x2264U,0xcaa8U,0xc367U,0x2188U,0x4e8bU,0xcb23U,0xa492U,0xc584U,
		0xb72fU,0x58b3U,0xef74U,0xeedbU,0xf75aU,0x05b5U,0x6a3eU,0x6b24U,0x2b5cU,0xc433U,
		0x4b0cU,0x6d02U,0x31ebU,0xb6bcU,0x631dU,0xb7faU,0xf976U,0x4372U,0xf824U,0x42b5U,
		0x6a7fU,0xb3f0U,0x1d16U,0x8d5eU,0xe025U,0xe7e0U,0x1803U,0x5d58U,0xfe77U,0x61a8U,
		0x4354U,0x07f5U,0xeba0U,0x452cU,0x4c17U,0xd156U,0xebe7U,0xcfa4U,0xfb0fU,0x5550U,
		0xc61eU,0x1500U,0xe199U,0xfc9fU,0xaaa6U,0x03f9U,0xd70bU,0xba84U,0x0cf2U,0x72d4U,
		0x51c8U,0xf28bU,0x73a4U,0xa33dU,0xdb0dU,0xc161U,0xd8aaU,0xe47cU,0xfeaeU,0xbdf1U,
		0xc759U,0xfde3U,0x16a4U,0xf942U,0x738dU,0xb366U,0x02d1U,0x0253U,0x0ee0U,0x6284U,
		0x2a76U,0xc336U,0x5378U,0x9de6U,0xad3dU,0x2b46U,0xfa35U,0xf408U,0x674cU,0x3efdU,
		0xb400U,0xb7c3U,0x9900U,0x2411U,0xbf15U,0x4a3fU,0xeeffU,0xa027U,0x0accU,0x8d7bU,
		0xc011U,0xf586U,0xf3b8U,0x63d9U,0x9fb7U,0x0b1aU,0xf678U,0x3e2cU,0x33f6U,0xc869U,
		0xaf83U,0x6d7cU,0x6893U,0xea6bU,0x88f0U,0xdc80U,0x0221U,0x63c4U,0xf625U,0x084cU,
		0x64a2U,0x188aU,0x1bc2U,0xfc62U,0xb585U,0x2d4eU,0x1c33U,0xb554U,0x9050U,0x0563U,
		0x81c2U,0x5789U,0x8821U,0x67c3U,0xdbb2U,0x961bU,0xdaeaU,0x0dc1U,0x93d1U,0x1d03U,
		0x0e9cU,0x1aaeU,0xe017U,0xda07U,0x4eb5U,0x37a3U,0x6443U,0xd935U,0xc97aU,0x8d10U,
		0x1786U,0x69feU,0xcfbaU,0x9f7bU,0x8860U,0x456eU,0x6272U,0x0fa7U,0xf0b3U,0x7f6aU,
		0x954aU,0x814aU,0x20a7U,0x1eeeU,0x1bebU,0xdfc8U,0x0a4eU,0x2f77U,0x8852U,0xe5abU,
		0xf76aU,0xaae6U,0x8ecbU,0x9459U,0x1b66U,0xe39bU,0x50f9U,0x19e7U,0xda1cU,0x1900U,
		0xf4d4U,0xcd09U,0x2f2dU,0xfcd6U,0x57fbU,0x1c25U,0x2a15U,0x47f2U,0x1eccU,0x2c17U,
		0x9dc8U,0x0c32U,0x09f3U,0x0f08U,0x3831U,0xa44aU,0xc4d0U,0xabf0U,0x5c3bU,0x9c9aU,
		0x479dU,0x750cU,0x536aU,0xe9c5U,0x78a8U,0xcb5dU,0x5de4U,0x517bU,0x9c0eU,0xae1aU,
		0xa742U,0x2cf0U,0x251eU,0xfa02U,0x519dU,0x01fbU,0xfb59U,0x00fcU,0x8617U,0xfe69U,
		0x3bbcU,0x0d08U,0x0ec6U,0xf6b6U,0xabd7U,0x5bccU,0xbb66U,0x18afU,0x0e38U,0x5896U,
		0x0662U,0xf2f0U,0xd68fU,0x9ae5U,0x0e46U,0x6574U,0xd1a8U,0x8b9aU,0x69e6U,0x86e7U,
		0x49b9U,0xe829U,0x4599U,0xecb8U,0x144fU,0xae44U,0x296eU,0x118cU,0xcd6fU,0xf2d2U,
		0xbd44U,0x803dU,0x001bU,0xcf27U,0x73c8U,0x4b04U,0x1cdcU,0x25d1U,0x4b94U,0x0ab8U,
		0x1ab7U,0x6fefU,0x3823U,0xc63dU,0xd642U,0x8b03U,0xd841U,0xbc0bU,0x2578U,0x10abU,
		0xae52U,0x6ff8U,0x35faU,0x08f4U,0x8257U,0xb10aU,0x73d0U,0x27fcU,0x5a58U,0x518fU,
		0x85a5U,0xf518U,0xd102U,0x5302U,0x4867U,0xa016U,0xf339U,0x8547U,0xa2c9U,0x6d0bU,
		0xc562U,0x872cU,0x5d65U,0x91dcU,0x9dbbU,0x5e0eU,0xba09U,0xa6b8U,0x3a9aU,0xc2c5U,
		0x850fU,0xbda1U,0xa5e1U,0x6e37U,0x9e32U,0x6ed1U,0xf041U,0xdee1U,0xbd56U,0x0986U,
		0xb3bbU,0x50ecU,0x376fU,0xd256U,0xe1a6U,0x3c1bU,0x5337U,0xbcf1U,0x45e8U,0x125aU,
		0x8708U,0x43c1U,0x6ea0U,0xb057U,0x860aU,0x9113U,0x56beU,0x2ac9U,0x56e6U,0xae0aU,
		0xbbaeU,0xafe9U,0xfe9aU,0xbaebU,0x24a4U,0x3b0eU,0xfdf0U,0xc916U,0xdb24U,0x466bU,
		0x8e49U,0xc4c3U,0x6377U,0xb7a7U,0xc151U,0xc51eU,0x2273U,0xd4b7U,0xbeaeU,0x7fa1U,
		0xfa42U,0xe7c0U,0x7b7aU,0x7637U,0xd0daU,0xb4d3U,0x1f04U,0x1deeU,0x1b99U,0x6ec1U,
		0xa066U,0xb22bU,0x0655U,0x8055U,0xeb56U,0x71f0U,0xb2deU,0x64c4U,0x7a35U,0x5874U,
		0x64afU,0xb428U,0xa112U,0x766fU,0xb5aeU,0xf4f3U,0xd239U,0x9ed1U,0xbedcU,0xcd62U,
		0x0b27U,0x73bfU,0xf376U,0x6d25U,0x08e0U,0xc37aU,0x3231U,0xd08fU,0x26bcU,0x64ddU,
		0x886dU,0x1277U,0x824aU,0xeda3U,0x5981U,0xaf71U,0xd95fU,0x53edU,0xa101U,0x7787U,
		0x8beeU,0x0121U,0x1066U,0x70e3U,0xa2c3U,0xabd1U,0x9544U,0xeef9U,0x1d10U,0x485aU,
		0xf0faU,0x0d1dU,0xaf05U,0x371aU,0x1563U,0x9d85U,0xfaf3U,0x199aU,0xfaedU,0x7b74U,
		0xee2cU,0x7951U,0xf316U,0xf5eaU,0x7a5eU,0x509dU,0x1d25U,0x9c51U,0xbe98U,0xa307U,
		0x0ed3U,0x3337U,0x4b60U,0xdbc8U,0xaf26U,0x50baU,0xb6a9U,0xa95fU,0xa712U,0xb0afU,
		0x6f77U,0x795dU,0x12a2U,0x708fU,0xda94U,0x717aU,0x7265U,0xeef2U,0x5188U,0x69f7U,
		0xb6b5U,0x3422U,0xc465U,0x11b7U,0xdeb3U,0xf5deU,0x2104U,0xb513U,0x1ed9U,0x7fefU,
		0xeb93U,0x0e08U,0x3805U,0xf6ebU,0xee8eU,0xf56aU,0x30abU,0x7872U,0x9466U,0xc3b0U,
		0xbe8cU,0xafc4U,0xd341U,0xd669U,0x5b73U,0x1345U,0xca01U,0xe18cU,0x2f8cU,0x2bb4U,
		0x19a0U,0xb5beU,0xc5aeU,0x66a2U,0xf1a9U,0xa0c7U,0xcbbeU,0x324cU,0xb117U,0xe180U,
		0x5343U,0x9febU,0x7c6fU,0xabbfU,0x95ccU,0x468fU,0x83b1U,0xbdadU,0x1779U,0xd45eU,
		0xfa4eU,0xae3bU,0xc56fU,0x8fd2U,0xbe80U,0x8947U,0x2afcU,0x3134U,0xe929U,0x21dfU,
		0xe28dU,0x3b1aU,0x3cadU,0x34faU,0x6b33U,0xdaddU,0xce6bU,0x5880U,0xf097U,0x8a95U,
		0x1756U,0xc600U,0xff27U,0x7866U,0xe836U,0xf22bU,0x08ecU,0x6b1cU,0x81b5U,0xf233U,
		0x0277U,0x0967U,0xcf2fU,0xf1f0U,0x8119U,0x726cU,0x512aU,0x7973U,0x20f8U,0x70efU,
		0xf1c4U,0x95e9U,0xccb4U,0x9b66U,0xfaf9U,0x33e9U,0x1ce9U,0x68cfU,0xebf3U,0x6334U,
		0x38c7U,0x6a31U,0x36b0U,0x467eU,0x6594U,0x8bbeU,0xf0eeU,0x2034U,0x002eU,0xa349U,
		0x1728U,0xe74bU,0x90c6U,0x4ae8U,0xac2eU,0x018cU,0x6a72U,0xed70U,0xd362U,0xed2aU,
		0x02ddU,0xb59bU,0xaa5aU,0x198eU,0x3d90U,0x1d8fU,0x5d7aU,0x241dU,0x7770U,0xf148U,
		0x5317U,0xb281U,0xd482U,0x285fU,0x5772U,0xe51aU,0xe3d3U,0x5795U,0xba51U,0xefacU,
		0x26b0U,0x999dU,0x88fcU,0xa393U,0x9353U,0x876bU,0x9d96U,0x8156U,0x4feaU,0x1a13U,
		0xe5f6U,0x166bU,0xd9f6U,0xdb5cU,0x3691U,0x7a78U,0x4ec1U,0x777bU,0x9d8bU,0xc7e7U,
		0xcdb1U,0xb934U,0x1a9bU,0x339cU,0x5d00U,0xd501U,0x210fU,0xfa0eU,0x08baU,0x993aU,
		0x99daU,0x49b0U,0x6b03U,0x67e4U,0xf36aU,0x3659U,0xc4dcU,0x1d30U,0x7707U,0x148fU,
		0xce76U,0xfce2U,0x3157U,0x2ed8U,0x59d9U,0x94feU,0x0a5aU,0xcbc9U,0xa388U,0x9826U,
		0x488eU,0x7663U,0x31f9U,0x059aU,0xc26cU,0x200dU,0xd64eU,0xf04cU,0xd4edU,0x22ebU,
		0x9958U,0x5623U,0xfdfbU,0xf524U,0x846bU,0x856dU,0x04cfU,0x64edU,0x7a29U,0x7888U,
		0x670eU,0x9436U,0xcb31U,0x7181U,0x0e52U,0xb2e9U,0x620eU,0x3c8eU,0xe3c8U,0xe620U,
		0x8fa4U,0x3b6aU,0x1569U,0xce44U,0xd96aU,0x8049U,0xdb18U,0x5258U,0xca21U,0x41acU,
		0x9b0aU,0xc8fcU,0x454dU,0x34a8U,0x39f8U,0x7809U,0xf10cU,0x0ad7U,0x21cbU,0xb158U,
		0xac42U,0xe095U,0x58c7U,0xadfeU,0xa924U,0x3161U,0x6169U,0x3d46U,0xb590U,0xe2c0U,
		0xac38U,0x5914U,0x008fU,0x76d0U,0x5ed5U,0xf0beU,0x7918U,0xe2feU,0xa4d1U,0x1700U,
		0xbea2U,0x0da1U,0xe7edU,0x82c6U,0x9928U,0x8622U,0x478aU,0x83bbU,0x80bfU,0xd373U,
		0x5806U,0xb61eU,0x5066U,0xc5d8U,0x32dfU,0x5a63U,0x2b69U,0xb79bU,0x5d6fU,0xd990U,
		0x234aU,0xaad4U,0x38adU,0x94e1U,0x145aU,0x57b2U,0x3720U,0x223cU,0x55e2U,0xa070U,
		0xad47U,0x656bU,0x9036U,0xc04eU,0x9dafU,0x7c05U,0x093eU,0xad50U,0x9ffeU,0xc484U,
		0x3a5cU,0xcfc4U,0x99e3U,0xfe18U,0x3271U,0xa9ffU,0x142aU,0xfa57U,0x6329U,0xb8a8U,
		0x1af0U,0xbd95U,0x3bc6U,0x09bcU,0x2717U,0xc590U,0x0b30U,0x3049U,0xc8f2U,0xa2d3U,
		0x29f9U,0xf9c1U,0x0ea5U,0xc193U,0xb299U,0x3a2fU,0x696fU,0x843fU,0x3763U,0x2c91U,
		0xe40cU,0x4e96U,0x573fU,0x1ac0U,0x8927U,0xe83eU,0x8ef9U,0xf267U,0x0023U,0x6a48U,
		0x4f63U,0xda9bU,0x3dd5U,0x15cbU,0xb264U,0xb04bU,0x375bU,0xff1fU,0xd839U,0xb727U,
		0xaf0bU,0x4b15U,0xb8faU,0x340fU,0xc0dbU,0x50b1U,0xf91aU,0x1b35U,0x741cU,0x5b0cU,
		0xf916U,0x7ab2U,0xfd7dU,0x5585U,0x99f5U,0xebb0U,0x78e3U,0xf428U,0xcaf5U,0xfa6dU,
		0x7748U,0x3e0dU,0x7ab6U,0x3ad0U,0xf5caU,0x6dfaU,0x0705U,0xef21U,0x0cd1U,0x121eU,
		0x5ae0U,0x3a6bU,0xe501U,0xba7aU,0x70c7U,0x179aU,0xc494U,0xfbbaU,0x7420U,0x3928U,
		0x7207U,0x527eU,0x274cU,0x1311U,0xd786U,0x550bU,0xa58bU,0xc3e4U,0x9baeU,0x4c91U,
		0xff90U,0x8f69U,0x2c04U,0x3be1U,0x5b10U,0x3ecbU,0x23c6U,0x0fd7U,0xc1deU,0x9b22U,
		0x1b91U,0x5cf8U,0x111fU,0xf3b0U,0x9c57U,0x8694U,0xca71U,0xcee7U,0xa35eU,0xa809U,
		0xf91eU,0x3058U,0xee4cU,0xb1f9U,0xfc51U,0x0907U,0x45fbU,0x9cdcU,0x7d66U,0xec3aU,
		0xe42aU,0x2d0fU,0x2b2bU,0x9292U,0x834fU,0xd21dU,0x1b39U,0x88a0U,0xb2b2U,0x2929U,
		0xd060U,0xbca5U,0x8094U,0xd8a0U,0x7455U,0xf3e8U,0xfbeaU,0x2667U,0xc80cU,0x28d2U,
		0x1052U,0xc42eU,0xc0dfU,0x8ddfU,0xe2afU,0xe87fU,0xa5c8U,0xdea0U,0xecb3U,0xb105U,
		0x18fbU,0x2cebU,0xdc7bU,0xd930U,0x14fbU,0x9de1U,0xd459U,0x0962U,0x50b5U,0x786dU,
		0xb423U,0x70deU,0xdedcU,0x4666U,0xa9faU,0xf262U,0xe090U,0xc5d3U,0x1d2bU,0x7883U,
		0x4ae3U,0x8766U,0xbcd5U,0xd312U,0xcaf9U,0x248cU,0x6f0cU,0x812cU,0xf69fU,0x677cU,
		0x3dc1U,0x7a55U,0x0808U,0x6c6cU,0xace9U,0x816bU,0x5f6bU,0x70d0U,0x2727U,0x3737U,
		0xfa71U,0x6ba6U,0xf18fU,0x619eU,0x6f03U,0xc542U,0xbb1fU,0x5ccfU,0x51f2U,0x2081U,
		0x820aU,0x153eU,0xd892U,0xd556U,0x8640U,0xed13U,0x78e7U,0xd8ecU,0xbb2eU,0x920bU,
		0x61baU,0xa1ddU,0x5f5bU,0xf24dU,0xfc17U,0xe41bU,0x9180U,0x4c83U,0x9f2aU,0xa8e7U,
		0x1511U,0xdd87U,0xf42cU,0x9e7dU,0x6bbbU,0xd016U,0xbe3dU,0xd2bcU,0xbfd4U,0xca61U,
		0xf5bbU,0x72e4U,0x50a3U,0xf782U,0x5b9fU,0xb9d5U,0x874cU,0x440fU,0x99f9U,0x6007U,
		0x92baU,0x0d5aU,0xe8d2U,0x871aU,0x7efcU,0xe134U,0xccebU,0x0d4cU,0xf204U,0x0751U,
		0xcffdU,0x56a2U,0xebccU,0x4c0dU,0xebb4U,0xf9e8U,0xbb7dU,0x9bd4U,0xb922U,0x48cdU,
		0x648fU,0x68faU,0x1522U,0x1e41U,0x47cfU,0xa75dU,0x8aaeU,0x7d1eU,0xcea9U,0x2882U,
		0xfd81U,0x9a62U,0xcb77U,0x87d6U,0xad58U,0x2290U,0x7f51U,0x1530U,0x3545U,0xee68U,
		0x75b8U,0x313cU,0xdcabU,0xcc83U,0xbc13U,0x25fbU,0x5589U,0xff2eU,0x4a18U,0x40a8U,
		0xd5e5U,0xe25eU,0xdd79U,0x0613U,0x24e5U,0x437aU,0x7a64U,0x4aa8U,0xb830U,0x5f8eU,
		0xc03eU,0x3f9aU,0x0709U,0x51fcU,0xf663U,0xeab4U,0x4f51U,0xf8afU,0x8ab8U,0x2e05U,
		0x4141U,0x1717U,0x60bcU,0x2407U,0x5cc6U,0x2c67U,0x1414U,0x7171U,0xef25U,0x1b10U,
		0x2ea9U,0x7ac0U,0x9cd4U,0xfa23U,0x6726U,0xeafbU,0x8bddU,0xf7f1U,0x0368U,0xc5bfU,
		0xd06aU,0xd495U,0x39c5U,0xa9baU,0xf5ceU,0xe30dU,0xa9b3U,0xd39fU,0xfbb3U,0xed0cU,
		0x71e8U,0x49a8U,0xb2c8U,0xc861U,0xcc92U,0xa831U,0x22a5U,0xd61fU,0x0500U,0x5fadU,
		0x6dfeU,0x50dcU,0xa251U,0xcc4cU,0x9e5cU,0x9c3bU,0x7314U,0x0713U,0xc075U,0xa196U,
		0xf5f0U,0xd7beU,0xe386U,0x004bU,0x1ad0U,0x9690U,0x7abaU,0x01e5U,0x6198U,0x9bceU,
		0x6df4U,0x2923U,0x826bU,0xfa2bU,0xc24aU,0x6dc6U,0x388cU,0x528cU,0x9cb1U,0x1973U,
		0x4ac2U,0xc66dU,0x3ad4U,0x6142U,0xa4baU,0x01ebU,0xbd05U,0x7dedU,0xb27aU,0x16f9U,
		0x910cU,0x11b0U,0xbdeaU,0xabe9U,0x2091U,0x15b8U,0x8c8fU,0x23d4U,0x774cU,0x4256U,
		0xa47cU,0xf7a5U,0x115fU,0xee13U,0xff01U,0xbe29U,0xbfc1U,0x16b4U,0xb9c3U,0xc9e5U,
		0x2325U,0x1c5aU,0x6fbaU,0xd7ecU,0x3e11U,0xc607U,0xb83fU,0xb314U,0x1a83U,0x2cc9U,
		0x7a96U,0xb5fcU,0x6ae0U,0xac51U,0xf31cU,0xacc6U,0x896bU,0x8a30U,0x569dU,0xcc7eU,
		0x70cbU,0xa8e2U,0x928dU,0x9ddcU,0xfa68U,0x8f64U,0xcec9U,0x1b47U,0xd615U,0xd52bU,
		0x196eU,0x1c55U,0x2c62U,0xd61aU,0x4cdaU,0x2db3U,0x179eU,0x224bU,0x622cU,0x1ad6U,
		0xe350U,0x1a5dU,0xe3f7U,0xa709U,0x6bd4U,0x0135U,0x0760U,0xf999U,0xf8dbU,0xb53cU,
		0xb668U,0xd3a6U,0xe505U,0x330aU,0x6c90U,0x3066U,0xf7c0U,0x71baU,0x9b4cU,0x7d4aU,
		0xe97fU,0x6078U,0x3ce4U,0x8804U,0x1f2bU,0xd530U,0x1c1fU,0xf2ccU,0xba7eU,0x1884U,
		0x7cfeU,0x13fbU,0xf49aU,0x367eU,0x3d40U,0xbd8fU,0xed24U,0x2ed2U,0x7957U,0x8941U,
		0xd250U,0x9ecbU,0x9a96U,0x169aU,0x5ae4U,0x55b3U,0x47deU,0xa539U,0x8c24U,0xf9caU,
		0x4482U,0xe056U,0x21a1U,0x5815U,0x9507U,0x5e9fU,0xa9a9U,0xceceU,0xd8bbU,0xd467U,
		0x3a6fU,0x6186U,0x7e7eU,0x9696U,0x668dU,0x3441U,0xd1cdU,0xab8cU,0x805eU,0x3f0dU,
		0xcf5cU,0x1fbbU,0xaa7aU,0xa7f8U,0x4577U,0xce0cU,0x0cd5U,0xa772U,0x0781U,0x92e5U,
		0x0946U,0x9cf2U,0xdd2cU,0x27e3U,0x427bU,0x8735U,0x952fU,0x2b04U,0x0818U,0x1b4cU,
		0xd849U,0x7193U,0x1222U,0x175dU,0x8794U,0xcdf8U,0x90f5U,0x4dc9U,0x77daU,0xdf2bU,
		0x4987U,0xf97eU,0xee32U,0xc43cU,0x19c2U,0x2764U,0xc9bcU,0xe968U,0x2750U,0x8ea0U,
		0x4d7aU,0x3f1bU,0x5baaU,0x6ee3U,0x083aU,0x83ffU,0x02f6U,0x4191U,0x97caU,0x8215U,
		0xf86bU,0xe355U,0x24f0U,0x40d2U,0x1315U,0xcd35U,0x10ddU,0x9cb7U,0xadafU,0x26f9U,
		0xd231U,0x314fU,0xfa82U,0x649aU,0x2bedU,0xc100U,0xc4aaU,0x977fU,0xe1efU,0xf5d5U,
		0x720bU,0xb633U,0xb2fcU,0x20c6U,0xb2aaU,0xa4b2U,0xa630U,0x21afU,0xc1e9U,0x65feU,
		0x6505U,0x89dfU,0x5025U,0x1a62U,0x7db3U,0x645cU,0x5282U,0x9445U,0x0ca4U,0xc958U,
		0x25b6U,0x835aU,0xa907U,0x30e7U,0x98d3U,0x5f30U,0x50c0U,0xea61U,0xd8e3U,0x76e7U,
		0xdb49U,0x30ffU,0x7424U,0x448dU,0x0497U,0xa529U,0x6c16U,0x9fa3U,0x4286U,0x2eb1U,
		0xff4bU,0xc303U,0xdf4aU,0xf714U,0x75c8U,0x6231U,0x5ee5U,0x1e64U,0x392cU,0xf738U,
		0xb304U,0x1979U,0x76aaU,0xa5a9U,0xe5d8U,0x2139U,0x8c72U,0x3fd9U,0x9236U,0xdf39U,
		0x1f36U,0xa688U,0x44e0U,0xa1afU,0xc498U,0x3617U,0x7757U,0xb643U,0xaab8U,0x91fbU,
		0xaee6U,0x8c06U,0x543dU,0xc27bU,0xca37U,0x641cU,0x951eU,0x1adbU,0xc347U,0xdd3aU,
		0xfbbeU,0x8a9cU,0x752eU,0xe07dU,0x6deaU,0x6952U,0xc7a9U,0xf66eU,0x2c2eU,0x085cU,
		0xaf2cU,0x913fU,0xedefU,0xacd8U,0x3c34U,0xc44dU,0xff94U,0xfcf1U,0xc54bU,0xfb99U,
		0xf84cU,0xc352U,0xb878U,0x5d47U,0x107dU,0x496aU,0x3937U,0xe106U,0xc08dU,0x17a3U,
		0xbff8U,0x0cafU,0x8f6dU,0x3008U,0x7af0U,0x3892U,0xef5aU,0xbab2U,0x89a3U,0x2c70U,
		0xadbaU,0x18d8U,0x57e3U,0xd2a9U,0x5f47U,0x29d7U,0x9649U,0xe314U,0x9bb2U,0x76f0U,
		0x9d18U,0x970aU,0xd94dU,0x783bU,0xd826U,0xeb8aU,0x5933U,0x655aU,0xed4eU,0x3e66U,
		0xa7b2U,0x2250U,0x4312U,0x00b0U,0x4c95U,0x220cU,0x86d1U,0xb54aU,0x5f52U,0xd81dU,
		0x556bU,0x2a37U,0x11a6U,0xe89dU,0xdbceU,0xdbd9U,0x95b5U,0xf894U,0x68b9U,0x3cf2U,
		0xa58fU,0xd743U,0x1951U,0x1fddU,0x8197U,0x8a5aU,0xa6d6U,0xfe54U,0xa2e5U,0xf3c8U,
		0xd38dU,0x07beU,0x192eU,0x4cdfU,0x743dU,0x58faU,0xc3e8U,0x86aaU,0x5b1fU,0x5292U,
		0x14d3U,0x0fecU,0x8f3cU,0x5a2eU,0x963eU,0x5d0fU,0xea1aU,0x6361U,0x85e0U,0xb6ebU,
		0xb168U,0x7d8dU,0xd78aU,0x7597U,0x6e61U,0x29abU,0x7c0dU,0x4774U,0xb74eU,0xf0d1U,
		0xe86bU,0x1138U,0x5b59U,0x894fU,0xe2d5U,0x2db8U,0x8b0bU,0x0217U,0x550fU,0xf09eU,
		0x4927U,0x8dffU,0xd189U,0x70a9U,0x654fU,0xa044U,0x22d5U,0xb802U,0x4b66U,0xb1c1U,
		0x4ab8U,0x268fU,0xa7dcU,0x2cb8U,0xc1e2U,0x592cU,0xb2c1U,0xe978U,0xa584U,0xfc10U,
		0xd087U,0x0592U,0x5160U,0x5781U,0xfd25U,0x7493U,0x6796U,0xf8e0U,0xef10U,0x8c9dU,
		0x9b26U,0x46f4U,0x4205U,0xc250U,0x88c8U,0x0457U,0x9000U,0x7138U,0xad9cU,0x8d20U,
		0xdd1aU,0xc22fU,0xd724U,0x2b9bU,0x41eaU,0x39ccU,0x23caU,0x73f2U,0xfe49U,0x04e8U,
		0x42a6U,0xf014U,0x74a5U,0xcfe0U,0xa7e3U,0x59faU,0xdf5bU,0x67b9U,0xd00dU,0xb541U,
		0x55f2U,0xd660U,0x0fdbU,0xe21bU,0xa680U,0x2eecU,0x28caU,0x15b0U,0x4d62U,0x5e2eU,
		0x8cedU,0x82dcU,0x333dU,0xd7faU,0x9400U,0x7bafU,0x0244U,0x721eU,0x5b14U,0x8c59U,
		0x17dfU,0x15efU,0xdaacU,0x5db5U,0xe4a6U,0x7ac8U,0x7f1dU,0x7b5dU,0x31c0U,0xa81eU,
		0x96a9U,0x0765U,0xe1c8U,0x8d01U,0x3ecfU,0x167aU,0x20ceU,0x6dccU,0x8428U,0x53a3U,
		0xd5aaU,0xffafU,0x980aU,0xa8bfU,0xb208U,0x4d06U,0xb1e4U,0x2ef4U,0xb6e0U,0x8b57U,
		0x2c08U,0x4398U,0x3dacU,0x6150U,0xbfb7U,0x62b3U,0x11fbU,0xe00dU,0x4627U,0xc2a5U,
		0x4729U,0xec8fU,0x07d6U,0xf99eU,0x253eU,0x6d65U,0x3be5U,0x5887U,0x3a44U,0x5e47U,
		0xb501U,0xc12dU,0x26a9U,0x80b8U,0x249dU,0x37feU,0x0ed9U,0x5dddU,0xc68aU,0x572bU,
		0xc5e8U,0x7023U,0xca75U,0xd7d1U,0x5af6U,0xb8b2U,0x1827U,0x9f4bU,0xbbcaU,0x0c4aU,
		0x4f28U,0x02c1U,0xa2f8U,0xcc5aU,0xe2caU,0x6bd9U,0xf557U,0x69cdU,0xceebU,0x9830U,
		0xf4b9U,0x8271U,0xdc56U,0xa2afU,0xf152U,0x7ee6U,0x1ebcU,0x8e24U,0x96bdU,0xae2cU,
		0x7d2eU,0x609cU,0x0b6cU,0xa9c1U,0x9c5bU,0x3abeU,0xeeb7U,0x4564U,0x5b96U,0x95acU,
		0x0329U,0x3256U,0x2cbfU,0x8831U,0x29a0U,0x0bbbU,0x7f92U,0x013aU,0xddb9U,0xf638U,
		0x8698U,0x5409U,0xdeebU,0x2500U,0x6d50U,0x7403U,0xf814U,0xc5a5U,0x3e8aU,0x65a3U,
		0xa30dU,0x956aU,0xb526U,0x64ceU,0x8d85U,0xa86bU,0x1123U,0x8c3fU,0xa377U,0x710bU,
		0xcc15U,0x2a02U,0x4e87U,0x31e7U,0x6544U,0x7f74U,0x59e2U,0xc848U,0x65e7U,0xe3fcU,
		0xdbaeU,0x1782U,0xf3b4U,0xf621U,0xc755U,0xfa31U,0xfe73U,0xaaa2U,0x381fU,0x5a54U,
		0x065eU,0x296aU,0x5c37U,0x5199U,0x0a4aU,0x2f29U,0x8869U,0x9540U,0x1b95U,0xb5aaU,
		0x56baU,0x6373U,0x9db7U,0xb3b7U,0xf093U,0x8115U,0xcbbaU,0xc56bU,0xdeafU,0xbe88U,
		0xf312U,0xa70eU,0xd64aU,0xcb2dU,0x5cfcU,0xce72U,0xd47eU,0x4fe6U,0x38c3U,0x6a6eU,
		0xb295U,0x4f5fU,0x9032U,0x6325U,0xbe9eU,0x2b65U,0xca1dU,0xa920U,0x948fU,0xbef4U,
		0xfc55U,0xfd0fU,0x8481U,0xde50U,0x3d57U,0xf78eU,0x6874U,0xfbccU,0x6401U,0x245dU,
		0x918cU,0x793dU,0x7bc5U,0xf7c5U,0x5398U,0xd562U,0x090bU,0xf579U,0x792cU,0xbd0bU,
		0x1f21U,0xef8fU,0x5537U,0xa738U,0x5f77U,0x9892U,0x1fcaU,0x4b45U,0xa855U,0xd401U,
		0x7f5dU,0x22acU,0xee50U,0x0549U,0x3fc3U,0x3625U,0x46cbU,0x147dU,0x43e6U,0x8eb8U,
		0x679dU,0x061fU,0x2ff3U,0x0f4eU,0xe534U,0x71bfU,0x6b8cU,0x3796U,0xb1fdU,0x03cfU,
		0xa09dU,0x6122U,0xccf7U,0x3eb0U,0x308dU,0xef44U,0x87e7U,0x6622U,0x7095U,0x1e4dU,
		0x5feeU,0x2f11U,0x7eadU,0x312bU,0xf922U,0x3ea1U,0x9516U,0x07ceU,0x14f3U,0x231dU,
		0x98bfU,0xef2dU,0x07a4U,0x19afU,0xd415U,0xe4efU,0xa645U,0x6c95U,0x1c44U,0x5ef3U,
		0x305cU,0x01b2U,0x7213U,0x7df3U,0xbc6fU,0xf27eU,0x057dU,0x6a17U,0xaf9eU,0x5bdcU,
		0xea08U,0x7502U,0x48c4U,0x0fe3U,0xdefeU,0xcd8aU,0xa362U,0x89ceU,0xfdd6U,0x88acU,
		0x13a3U,0x52e1U,0x9412U,0xa22cU,0x4092U,0x4eeaU,0xc1faU,0x3364U,0xf92aU,0x306bU,
		0xcb41U,0x14a1U,0xa80dU,0x096eU,0x973aU,0x3ef0U,0x7f12U,0x1b7bU,0x105eU,0x7a21U,
		0x7577U,0x643bU,0x12a8U,0x310dU,0xafb1U,0x162dU,0x415dU,0x4efeU,0x2b2fU,0x132aU,
		0xf198U,0x282dU,0x806aU,0xcd99U,0xda70U,0x7be8U,0xbbfeU,0xfaa2U,0xa97cU,0x5821U,
		0x2558U,0xe50aU,0x77e6U,0x3bb2U,0x9296U,0x35a5U,0x768fU,0xa4c0U,0xfadaU,0x926fU,
		0x37f3U,0x60c5U,0xc6fbU,0x27efU,0x7343U,0xccc5U,0x3d36U,0x1a26U,0x0b58U,0xd626U,
		0xe42eU,0x50c9U,0x9673U,0xe2e7U,0x61f0U,0x79d6U,0x86daU,0xfd62U,0xf8e7U,0x522bU,
		0x2773U,0xb21bU,0x830dU,0x330fU,0x867fU,0x4a6cU,0x2d13U,0x2d26U,0xa8a3U,0xdf64U,
		0x5005U,0x8810U,0xa6a1U,0x25c4U,0x2533U,0x9beaU,0x7963U,0xb8c6U,0xeec0U,0x92faU,
		0xf05cU,0x0b07U,0x7d6aU,0x23e0U,0x71a5U,0xb343U,0xb77eU,0x5c8eU,0x6fc6U,0x3ca4U,
		0xd290U,0x8af3U,0x03b9U,0x3f29U,0x18a2U,0xb66dU,0x52cfU,0xbf96U,0xec3eU,0xa888U,
		0xec7bU,0x01f1U,0x47c6U,0xea11U,0xfee3U,0x5753U,0x187aU,0xab42U,0xec84U,0x3843U,
		0x3ae0U,0xcb50U,0x7ca7U,0x10ebU,0x45ffU,0xf002U,0xa25dU,0x1d7bU,0xc056U,0x28a3U,
		0xa5faU,0x0071U,0xe65fU,0x9c1eU,0xfcc9U,0x8fdaU,0x0a9cU,0xd3abU,0xc645U,0x37dbU,
		0x9ce0U,0x027eU,0x0827U,0xea2aU,0x6248U,0x1600U,0x8ad2U,0xf8bbU,0x6b50U,0xb6c4U,
		0x6f7dU,0xe59eU,0xbfacU,0x35c4U,0xf8d5U,0xaa74U,0xd064U,0x208bU,0xd88cU,0x8aa8U,
		0x70c1U,0xc806U,0x7d28U,0x0a44U,0xd71eU,0xb1deU,0x5f41U,0x85daU,0xc4a4U,0x1f30U,
		0x7717U,0xca95U,0xbca9U,0x1e70U,0x3d30U,0x3adaU,0xa84fU,0x48beU,0xdb55U,0xaacdU,
		0xd4b0U,0x133eU,0xb7bcU,0xfe62U,0xbcafU,0x93b7U,0xd6a4U,0xcc99U,0xb2b6U,0x9f12U,
		0xced8U,0x2dc8U,0xbe03U,0xeca3U,0x0a16U,0x8366U,0xfd2cU,0xf1d3U,0x0d79U,0x3185U,
		0xdecbU,0xd535U,0x6d8dU,0x8f4aU,0x292dU,0xcba8U,0x4d86U,0xde91U,0x647fU,0x394aU,
		0xac64U,0x5206U,0x5700U,0xa45aU,0xb0b5U,0xab17U,0x1ea4U,0x9cc3U,0xa6e2U,0x4e77U,
		0x1b3dU,0x38dbU,0x712dU,0x1146U,0x695cU,0xd0ecU,0xaf5fU,0x1b18U,0x6f32U,0x5a3aU,
		0x0afeU,0x944fU,0x4406U,0x3ce9U,0x2699U,0xa0beU,0x88a4U,0x366fU,0xb63bU,0x6148U,
		0xe877U,0xf79dU,0x3c01U,0x9016U,0x4a29U,0xe8a7U,0xd432U,0xb80eU,0x1e76U,0xe223U,
		0x15a9U,0x73fcU,0x8353U,0xd816U,0x9c34U,0x3677U,0x9b1bU,0xd2b5U,0xed9bU,0x22e3U,
		0x62bdU,0x0db9U,0x3943U,0xde61U,0x3ea9U,0x8809U,0xf17bU,0xf8a0U,0xd221U,0xf23aU,
		0x329fU,0x2791U,0xd16fU,0x883bU,0xc2dcU,0x5653U,0xa7beU,0xfac9U,0xa718U,0x4cb2U,
		0x4daaU,0xe6c0U,0x3ff4U,0x57f1U,0xc810U,0xa27dU,0x4211U,0x4461U,0x3b52U,0xd6d4U,
		0x73e7U,0x38e5U,0x8f0bU,0xda2cU,0x40c0U,0xda60U,0x6222U,0xe984U,0xdfa9U,0x6522U,
		0x28d6U,0x7035U,0xe98eU,0x4ac8U,0xe0bfU,0x20dfU,0x5cb2U,0xf020U,0xdd54U,0xe29cU,
		0x307dU,0x4bfbU,0x771dU,0x26eaU,0x33acU,0xa838U,0xfbeeU,0x722aU,0x3f62U,0x01bcU,
		0xad12U,0xb478U,0xe1d4U,0x625eU,0x749aU,0x0c86U,0x2680U,0xe114U,0xef99U,0x607dU,
		0xec5bU,0x4d49U,0x266bU,0xab4cU,0x75e4U,0x43a4U,0x9391U,0xde68U,0x1a30U,0x9abbU,
		0xefd2U,0x1ff8U,0xa965U,0x782cU,0x3bf1U,0xa0acU,0x357bU,0x0146U,0x7459U,0x469eU,
		0xd287U,0xa440U,0xcce2U,0xa2dcU,0x6e4aU,0x87acU,0xb532U,0x26e0U,0x2ee1U,0x4f7aU,
		0x1586U,0x9b51U,0x3464U,0xbdb6U,0xf3ecU,0xa167U,0x924bU,0xc673U,0x1614U,0xcc66U,
		0x8536U,0x95c3U,0xb234U,0xc11eU,0x96c9U,0x00e8U,0xca9bU,0xe701U,0x9f84U,0xac08U,
		0x8098U,0xea93U,0x8f1eU,0x06a4U,0x903eU,0x3634U,0x608dU,0x136eU,0x714cU,0x6d14U,
		0x559fU,0xc577U,0x93f5U,0x7d4fU,0x382bU,0x4b8eU,0xd8a4U,0x81bcU,0xdcbfU,0x225eU,
		0x62ceU,0x5ec3U,0x7912U,0x29f3U,0x776aU,0x31f3U,0xb6afU,0x6b2dU,0x6e9aU,0x322bU,
		0x57a5U,0x13dcU,0xa5ccU,0xcca5U,0x4589U,0x39aeU,0xb380U,0xaaffU,0x6d99U,0x66d2U,
		0x8e70U,0x5c79U,0x5d22U,0xc25eU,0x5c67U,0xf49fU,0xddf5U,0xb611U,0xdea4U,0xdbedU,
		0x2a8bU,0x2097U,0x4501U,0x9d46U,0xf5a4U,0xa4f5U,0x66baU,0x1fa7U,0xf35cU,0x1912U,
		0xbcb5U,0x8ae1U,0xc416U,0xb2cfU,0xe2b3U,0xb190U,0x32abU,0xf44cU,0x089aU,0x5bffU,
		0x3b2fU,0xe71fU,0x5167U,0x2588U,0x02afU,0xc0a2U,0x12e6U,0x3683U,0xb3acU,0x63ecU,
		0xe883U,0x524bU,0x116eU,0x89b3U,0x90ebU,0x62c4U,0x9fd6U,0xd0f8U,0x42d5U,0xe1b5U,
		0x3428U,0x6db4U,0xd354U,0x10c6U,0x90b6U,0x6032U,0xc0e3U,0xfdaeU,0x3a22U,0xdf94U,
		0xdd60U,0x0080U,0x87bcU,0x141cU,0xc82aU,0x1291U,0x764eU,0xda38U,0xd911U,0x7d03U,
		0x1a3cU,0x55d2U,0x8de3U,0xa3d0U,0xe1f7U,0x15beU,0x9027U,0x3a88U,0xaefdU,0xe3c0U,
		0xb253U,0x626aU,0x4c51U,0xe93aU,0x93bdU,0x55faU,0xb3e7U,0x9b5dU,0x1056U,0xc2d4U,
		0xc7a1U,0x26a1U,0xa9f2U,0x7a8eU,0xcbefU,0xf956U,0xcaa7U,0x39eaU,0x2cd8U,0x473fU,
		0x7eb5U,0x1400U,0x05b4U,0xeedaU,0xc432U,0x6b23U,0xcb22U,0x2187U,0x58b2U,0xc583U,
		0xb3efU,0x42b4U,0xe7dfU,0x8d5dU,0xb6bbU,0x6d01U,0x4371U,0xb7f9U,0xcfa3U,0xd155U,
		0x14ffU,0x554fU,0x61a7U,0x5d57U,0x452bU,0x07f4U,0xa33cU,0xf28aU,0xe47bU,0xc160U,
		0x03f8U,0xfc9eU,0x72d3U,0xba83U,0x6283U,0x0252U,0x9de5U,0xc335U,0xfde2U,0xbdf0U,
		0xb365U,0xf941U,0x4a3eU,0x2410U,0x8d7aU,0xa026U,0xf407U,0x2b45U,0xb7c2U,0x3efcU,
		0x6d7bU,0xc868U,0xdc7fU,0xea6aU,0x63d8U,0xf585U,0x3e2bU,0x0b19U,0xb553U,0x2d4dU,
		0x5788U,0x0562U,0x084bU,0x63c3U,0xfc61U,0x1889U,0xda06U,0x1aadU,0xd934U,0x37a2U,
		0x961aU,0x67c2U,0x1d02U,0x0dc0U,0x7f69U,0x0fa6U,0x1eedU,0x8149U,0x69fdU,0x8d0fU,
		0x456dU,0x9f7aU,0xe39aU,0x9458U,0x18ffU,0x19e6U,0x2f76U,0xdfc7U,0xaae5U,0xe5aaU,
		0x0c31U,0x2c16U,0xa449U,0x0f07U,0xfcd5U,0xcd08U,0x47f1U,0x1c24U,0x517aU,0xcb5cU,
		0x2cefU,0xae19U,0x9c99U,0xabefU,0xe9c4U,0x750bU,0xf6b5U,0x0d07U,0x18aeU,0x5bcbU,
		0x01faU,0xfa01U,0xfe68U,0x00fbU,0x86e6U,0x8b99U,0xecb7U,0xe828U,0xf2efU,0x5895U,
		0x6573U,0x9ae4U,0x4b03U,0xcf26U,0x0ab7U,0x25d0U,0x118bU,0xae43U,0x803cU,0xf2d1U,
		0x6ff7U,0x10aaU,0xb109U,0x08f3U,0xc63cU,0x6feeU,0xbc0aU,0x8b02U,0x8546U,0xa015U,
		0x872bU,0x6d0aU,0x518eU,0x27fbU,0x5301U,0xf517U,0x6e36U,0xbda0U,0xdee0U,0x6ed0U,
		0x5e0dU,0x91dbU,0xc2c4U,0xa6b7U,0x1259U,0xbcf0U,0xb056U,0x43c0U,0x50ebU,0x0985U,
		0x3c1aU,0xd255U,0x3b0dU,0xbaeaU,0x466aU,0xc915U,0x2ac8U,0x9112U,0xafe8U,0xae09U,
		0xe7bfU,0x7fa0U,0xb4d2U,0x7636U,0xb7a6U,0xc4c2U,0xd4b6U,0xc51dU,0x64c3U,0x71efU,
		0xb427U,0x5873U,0x6ec0U,0x1dedU,0x8054U,0xb22aU,0x6d24U,0x73beU,0xd08eU,0xc379U,
		0xf4f2U,0x766eU,0xcd61U,0x9ed0U,0x7786U,0x53ecU,0x70e2U,0x0120U,0x1276U,0x64dcU,
		0xaf70U,0xeda2U,0x9d84U,0x3719U,0x7b73U,0x1999U,0xeef8U,0xabd0U,0x0d1cU,0x4859U,
		0x3336U,0xa306U,0x50b9U,0xdbc7U,0xf5e9U,0x7950U,0x9c50U,0x509cU,0xeef1U,0x7179U,
		0x3421U,0x69f6U,0xb0aeU,0xa95eU,0x708eU,0x795cU,0xf6eaU,0x0e07U,0x7871U,0xf569U,
		0xf5ddU,0x11b6U,0x7feeU,0xb512U,0x2bb3U,0xe18bU,0x66a1U,0xb5bdU,0xafc3U,0xc3afU,
		0x1344U,0xd668U,0x468eU,0xabbeU,0xd45dU,0xbdacU,0x324bU,0xa0c6U,0x9feaU,0xe17fU,
		0x3b19U,0x21deU,0xdadcU,0x34f9U,0x8fd1U,0xae3aU,0x3133U,0x8946U,0x6b1bU,0xf22aU,
		0x0966U,0xf232U,0x8a94U,0x587fU,0x7865U,0xc5ffU,0x9b65U,0x95e8U,0x68ceU,0x33e8U,
		0x726bU,0xf1efU,0x70eeU,0x7972U,0xa348U,0x2033U,0x4ae7U,0xe74aU,0x6a30U,0x6333U,
		0x8bbdU,0x467dU,0x1d8eU,0x198dU,0xf147U,0x241cU,0xed6fU,0x018bU,0xb59aU,0xed29U,
		0x999cU,0xefabU,0x876aU,0xa392U,0x285eU,0xb280U,0x5794U,0xe519U,0x777aU,0x7a77U,
		0xb933U,0xc7e6U,0x1a12U,0x8155U,0xdb5bU,0x166aU,0x67e3U,0x49afU,0x1d2fU,0x3658U,
		0xd500U,0x339bU,0x9939U,0xfa0dU,0x9825U,0xcbc8U,0x0599U,0x7662U,0xfce1U,0x148eU,
		0x94fdU,0x2ed7U,0x856cU,0xf523U,0x7887U,0x64ecU,0xf04bU,0x200cU,0x5622U,0x22eaU,
		0x3b69U,0xe61fU,0x8048U,0xce43U,0x7180U,0x9435U,0x3c8dU,0xb2e8U,0x0ad6U,0x7808U,
		0xe094U,0xb157U,0x41abU,0x5257U,0x34a7U,0xc8fbU,0x76cfU,0x5913U,0xe2fdU,0xf0bdU,
		0x3160U,0xadfdU,0xe2bfU,0x3d45U,0xd372U,0x83baU,0xc5d7U,0xb61dU,0x0da0U,0x16ffU,
		0x8621U,0x82c5U,0x57b1U,0x94e0U,0xa06fU,0x223bU,0xb79aU,0x5a62U,0xaad3U,0xd98fU,
		0xcfc3U,0xc483U,0xa9feU,0xfe17U,0xc04dU,0x656aU,0xad4fU,0x7c04U,0x3048U,0xc58fU,
		0xf9c0U,0xa2d2U,0xb8a7U,0xfa56U,0x09bbU,0xbd94U,0x1abfU,0x4e95U,0xf266U,0xe83dU,
		0x3a2eU,0xc192U,0x2c90U,0x843eU,0xb726U,0xff1eU,0x340eU,0x4b14U,0xda9aU,0x6a47U,
		0xb04aU,0x15caU,0xc8bcU,0xcdd3U,0xf6a3U,0x327fU,0x753dU,0x8cadU,0xbaceU,0xa673U,
		0x6cbcU,0xc148U,0xe903U,0x4363U,0x8396U,0x4229U,0x6bc8U,0x8c29U,0xfb6aU,0xd6b2U,
		0x6780U,0x5525U,0xa094U,0x1165U,0x23bdU,0x0c9bU,0x9218U,0x8005U,0x8dd5U,0x89c4U,
		0x0746U,0xece2U,0xbe09U,0x2da7U,0xd598U,0x9e63U,0x6f10U,0x63a5U,0x979dU,0x8ff3U,
		0x7852U,0x79e8U,0x7adaU,0x088cU,0x88cfU,0x4c66U,0x1962U,0x91b5U,0xaa8aU,0xf9cfU,
		0x30bbU,0xe3e5U,0x8130U,0x9a4aU,0x6789U,0x2123U,0xb857U,0xd176U,0x8e62U,0x60e2U,
		0x2372U,0x0b9fU,0x7ff5U,0x8f58U,0xe8c7U,0x1d57U,0x288fU,0xc9cfU,0xcafdU,0x471cU,
		0x52adU,0xb461U,0x3b7bU,0xdb79U,0x2845U,0xd072U,0xf48eU,0x6c32U,0x2f56U,0x6b61U,
		0x4ddcU,0x47e3U,0xa508U,0x5a0eU,0x2490U,0xba93U,0x502dU,0xee19U,0xbc20U,0x42f4U,
		0x1e36U,0xe0e2U,0xb978U,0x8935U,0xb652U,0xe9ebU,0xeca9U,0x9408U,0x5e8cU,0x0237U,
		0xbcd9U,0x9222U,0x1e24U,0x1e7eU,0x1109U,0xae8bU,0xb65cU,0xe45cU,0xbd6bU,0xeacaU,
		0x6014U,0xb357U,0xa64dU,0xa53eU,0x841bU,0xe4dcU,0xd316U,0x70f6U,0xdc3aU,0x9e18U,
		0xec11U,0x4b23U,0x7ebdU,0x606cU,0xebc1U,0xe248U,0x1edfU,0x5e69U,0x2facU,0xc31eU,
		0x66b2U,0xb4caU,0x5f6fU,0xd4a8U,0xfa7aU,0xad94U,0x5ad8U,0x8bd5U,0xc02bU,0xc882U,
		0x832bU,0xcc01U,0xb761U,0x279fU,0x8dcdU,0x5ae9U,0x9e0eU,0x65c2U,0x70d4U,0x3ee6U,
		0x2877U,0xa482U,0xd9deU,0xe04aU,0xe163U,0x414aU,0x1d4eU,0x25e6U,0x2f08U,0x10bdU,
		0xcedeU,0x76deU,0x4a80U,0x9c42U,0xacedU,0x98dcU,0x7e72U,0x294eU,0x673cU,0x24b9U,
		0x4c9eU,0x91c5U,0x045eU,0x83d8U,0xea81U,0x031cU,0x2e81U,0x55b8U,0xdaceU,0xc98dU,
		0x816fU,0xf11aU,0x76b9U,0xa7ecU,0x353aU,0x8842U,0x45e0U,0x3d88U,0x32cbU,0x739cU,
		0xb519U,0x5861U,0x9c64U,0x05d7U,0x6f67U,0x6f3bU,0x080cU,0x8662U,0x06beU,0xba23U,
		0x1d62U,0x94efU,0x5f03U,0x11ddU,0x7462U,0x5c2aU,0xe76dU,0xecedU,0xce30U,0x9a9bU,
		0xe253U,0x56d6U,0x6c70U,0x92d9U,0x2fa2U,0xf7abU,0x4ad9U,0xafa7U,0x3862U,0x0381U,
		0x6537U,0xe0edU,0xec47U,0x3780U,0x2dceU,0x5f1dU,0x77ceU,0xa48dU,0x3dc5U,0xb7d2U,
		0x9341U,0x2c82U,0xaa07U,0x0c70U,0xd68aU,0xd63dU,0x884dU,0x4798U,0x9fb2U,0xdae5U,
		0x434fU,0x169fU,0xb112U,0x1751U,0x7a59U,0x20ffU,0xa061U,0x8245U,0xba04U,0x56e1U,
		0x9daaU,0x696aU,0x9b05U,0xe7e8U,0x210aU,0xd4e8U,0x36abU,0x576dU,0x92b5U,0xcb72U,
		0x6f07U,0xbb29U,0xee47U,0xe2aaU,0x7417U,0x7202U,0x47d9U,0x077cU,0x9288U,0x6c8bU,
		0x6193U,0xa477U,0xf65eU,0xa9aeU,0x194cU,0x6e5cU,0xc546U,0x9d13U,0x0492U,0x7752U,
		0x4d75U,0xb2f7U,0xa372U,0x56b5U,0x5af1U,0xeeb2U,0x17daU,0x3da7U,0xb2bcU,0xfe44U,
		0x71a0U,0xa258U,0xf193U,0x966eU,0x9511U,0xfdd1U,0x847cU,0x3fbeU,0xd282U,0x8f19U,
		0x420cU,0x3f5dU,0x7128U,0x9c2fU,0xd887U,0xced3U,0x2f71U,0xf2eaU,0x61a2U,0x63d3U,
		0x3a1dU,0xc79cU,0x4584U,0x32a6U,0x41a6U,0xc048U,0x6a2bU,0xd4fbU,0xf5e4U,0x3246U,
		0x5e08U,0x6ebbU,0x06b9U,0x933cU,0xfa75U,0x7e6dU,0x52a8U,0x1e1fU,0x7538U,0x9798U,
		0x2481U,0xd49dU,0xca51U,0x23fbU,0x52a3U,0x7dbdU,0xee42U,0x950cU,0x4b70U,0x406dU,
		0x6baaU,0x97bdU,0x1a6aU,0x425cU,0x858fU,0x0529U,0x51dfU,0xe8f4U,0x4f45U,0x5575U,
		0xd32aU,0x9ac9U,0x9f18U,0x7bb7U,0x25a9U,0x41f4U,0x272bU,0x08c8U,0x7dc2U,0xe22bU,
		0x1bd4U,0x14b9U,0xab7cU,0xf7e5U,0x3db6U,0xc020U,0x7569U,0x09a9U,0x6c9dU,0x5ea4U,
		0x751dU,0xe1e2U,0x373bU,0x797aU,0xaa40U,0x348bU,0x06e3U,0x0333U,0x1408U,0xbe19U,
		0xffe2U,0x32bcU,0xdeb9U,0xbc61U,0x0e84U,0x0725U,0xe0d8U,0xb325U,0xd896U,0x6a0dU,
		0xca56U,0x3d7dU,0xff0eU,0xb5f0U,0x1dbcU,0xa6ecU,0x0883U,0x5325U,0x8eafU,0xe716U,
		0x5502U,0x21a6U,0xb609U,0xbae2U,0xd55aU,0xca8dU,0x40caU,0x8c95U,0x3e05U,0x7169U,
		0x7e27U,0x1594U,0xd2e2U,0x0a7cU,0xf467U,0x67fdU,0xd6aaU,0x6e54U,0x748cU,0xcc53U,
		0x820eU,0xe0ffU,0x2400U,0x1c4eU,0xf421U,0x8164U,0x097eU,0x0184U,0xc257U,0xfc97U,
		0x85d3U,0xda59U,0x7936U,0x581aU,0xbc92U,0x3b44U,0x1542U,0xfeedU,0xa470U,0xe70fU,
		0x4222U,0x2798U,0x734dU,0x44c5U,0x23a8U,0x09d4U,0x11bdU,0xa6c9U,0x164bU,0xb078U,
		0xa8dcU,0x55adU,0x51f6U,0x01dfU,0x2486U,0x6001U,0x1b2fU,0x3052U,0xd7cbU,0x8c39U,
		0x5926U,0x8c53U,0xfcebU,0xd73dU,0x8e9aU,0x4487U,0xa277U,0x4698U,0x2085U,0x38d5U,
		0x1324U,0x23daU,0xfd09U,0x3e9bU,0xe744U,0xb151U,0x6ecaU,0xdbc1U,0x5549U,0x19e0U,
		0xcc9fU,0xfda8U,0x6a07U,0x01d9U,0xbb23U,0x7e67U,0xd4a2U,0x865cU,0x3279U,0x4716U,
		0xf83fU,0xa083U,0x7fbbU,0x8550U,0x2637U,0x6b3bU,0xbc46U,0xe05bU,0x825fU,0x0061U,
		0x5cd3U,0x5131U,0x9327U,0x6471U,0x1178U,0x0f62U,0xaa2fU,0xbfebU,0x4dffU,0xe02dU,
		0xc46bU,0x05beU,0x2b14U,0xed7eU,0x0471U,0xe815U,0x61beU,0xbc4cU,0x8594U,0xd1c0U,
		0xb9e0U,0x0e16U,0xbd1bU,0x5b88U,0x2c3bU,0x13c4U,0x75acU,0x1549U,0x38f9U,0x6692U,
		0x4385U,0xf46fU,0xa1e1U,0x6e22U,0xa99cU,0x232bU,0xde79U,0xa3fcU,0x4f1aU,0x6ce6U,
		0xb769U,0x6905U,0xc288U,0x2284U,0xded1U,0xec06U,0x2448U,0x6e05U,0xbb32U,0x18beU,
		0x052eU,0xb88aU,0xe5e5U,0x8787U,0x79b6U,0x532eU,0x9b2dU,0xca0fU,0x4ceeU,0xf9b1U,
		0x24d9U,0x3446U,0x2ebeU,0xb0edU,0x920fU,0x4f11U,0x552eU,0xf59bU,0x0cc8U,0xd228U,
		0x0b95U,0xbc7dU,0xf3a6U,0x6618U,0x7878U,0x5e5eU,0x101bU,0x25efU,0x9b76U,0x5690U,
		0x78ebU,0xe061U,0x1a6fU,0xc6c7U,0xc1b4U,0x8081U,0x8740U,0x0370U,0xcf8aU,0x42c6U,
		0x9588U,0x88baU,0x664eU,0x7e83U,0xe343U,0x5a7fU,0xd8f0U,0xc212U,0x650dU,0x1c60U,
		0x8523U,0x7f38U,0x44b4U,0x0d40U,0x1047U,0x5e81U,0x8cdfU,0x4879U,0xd53bU,0x3345U,
		0xb097U,0x9748U,0x8644U,0x3cd7U,0x4261U,0xd43aU,0xb130U,0xade9U,0x8222U,0xe46cU,
		0xb916U,0x6cd5U,0xc770U,0x9a17U,0xd41dU,0x969bU,0xe32aU,0x7374U,0xed17U,0x7273U,
		0x85efU,0x17efU,0x1c74U,0x0c5bU,0x2ce0U,0xeabfU,0x5140U,0xa751U,0x30b1U,0xe9adU,
		0x2bfaU,0x6b82U,0x3cffU,0x8189U,0x9f2eU,0x263dU,0x6bafU,0xe3afU,0xcbdaU,0xdd45U,
		0x3c4bU,0x0919U,0x05f5U,0x34c9U,0xf743U,0xfef4U,0x398fU,0x3a74U,0xe85dU,0xc963U,
		0xa8ebU,0x0f17U,0x3708U,0xb9c9U,0x744aU,0xb96dU,0x9e2aU,0x1720U,0x2704U,0xfb07U,
		0x895eU,0x59b6U,0x0d7fU,0x6abcU,0x5d1aU,0x50e3U,0x9184U,0x5f39U,0x97c2U,0xfd1dU,
		0x78dbU,0x60b4U,0x2215U,0x411eU,0x46fbU,0xa04fU,0x75a4U,0x7a07U,0x838eU,0x618bU,
		0x91abU,0x6856U,0x4c87U,0x0f44U,0xd212U,0x5a03U,0x29e2U,0xf241U,0x8491U,0x395fU,
		0x551cU,0x6e19U,0xf570U,0xdbe4U,0x3ac7U,0xcd2cU,0xbb13U,0x5a43U,0xfc1bU,0x6b41U,
		0x4b75U,0x6ddcU,0x94b0U,0x1f82U,0x4844U,0x04f3U,0x46a7U,0x5108U,0x8dacU,0xcb10U,
		0x7afdU,0xd8c0U,0x2a26U,0x6b9aU,0xe41fU,0x2f4bU,0x8c64U,0xc9ebU,0xbf2fU,0x5be5U,
		0xfba6U,0x7ecdU,0x2e6eU,0x9d57U,0xa891U,0x8e84U,0x318bU,0x8b62U,0x9f5eU,0x0795U,
		0x5f5fU,0xeaeaU,0x4072U,0x0bc6U,0xfe20U,0x5017U,0xd5bcU,0x12d3U,0x123fU,0x035bU,
		0x7c1eU,0x3502U,0x6be4U,0xd46cU,0x08fcU,0xdb6eU,0xf251U,0x8120U,0x7848U,0x4433U,
		0x2742U,0x46c1U,0x7f7fU,0x4646U,0x4134U,0x8d66U,0xee94U,0x22baU,0xa23fU,0xcd49U,
		0xe574U,0x6f8bU,0xbe41U,0x7fc1U,0x9f1dU,0x2d80U,0x7c86U,0x9ea6U,0xc23cU,0x6c42U,
		0x3a0aU,0xfb2dU,0x7a0fU,0xbc99U,0xff85U,0xaa7fU,0xa151U,0x6805U,0xd2c0U,0xb3c9U,
		0xadc9U,0xbfc7U,0x0c12U,0xf163U,0x7dd8U,0x1151U,0x27a7U,0xae61U,0xfbdeU,0x0f94U,
		0xfd32U,0x962fU,0x21efU,0x0507U,0x6bbfU,0x38f0U,0x7bbcU,0x5c5eU,0xcaecU,0xf862U,
		0x0c91U,0xb011U,0xef17U,0xef85U,0x5273U,0x8fe8U,0xdf19U,0xa7fdU,0x9b96U,0x1dd4U,
		0xd01aU,0x9096U,0x82a0U,0x4f85U,0x1c90U,0xf182U,0xbf6bU,0xf038U,0x28f4U,0xe378U,
		0xf6f1U,0x542eU,0x7d5aU,0x5e1cU,0xd5d8U,0x7dfeU,0xf430U,0x8556U,0xd32fU,0x62f9U,
		0xe4b3U,0xeccbU,0xbec1U,0xc5c7U,0x1b86U,0x8e57U,0x4979U,0x5052U,0xdce6U,0xcf61U,
		0xffbcU,0xce9cU,0x9e81U,0x9862U,0x89e7U,0x16baU,0xc0c6U,0xd56dU,0xd44dU,0x3aeeU,
		0xb5d9U,0x3e7bU,0x03daU,0x667fU,0xf1d9U,0xd802U,0x0f2dU,0x2ae4U,0x1515U,0x7474U,
		0x9aceU,0xb816U,0x35cfU,0x9121U,0x3ec0U,0x9a3fU,0x4634U,0xd924U,0x073cU,0x384eU,
		0xe4f7U,0x1fc0U,0x9bf5U,0x2e55U,0xdd8bU,0xf1f7U,0xc89aU,0xb67bU,0xad23U,0xd03bU,
		0x4747U,0x5151U,0xa5bcU,0x60d0U,0x380bU,0xaffaU,0x36f0U,0x2c54U,0xa57dU,0x1820U,
		0x5ba3U,0xf845U,0x4f4aU,0xfa61U,0xf90fU,0xf698U,0xc2bdU,0x8bb6U,0xb379U,0x4524U,
		0x70baU,0x3b4bU,0x3d50U,0x8063U,0x7c7fU,0xf908U,0xb9d9U,0xcbd3U,0x7410U,0xff07U,
		0xbac7U,0x5ad1U,0x2427U,0x1cbbU,0xfff5U,0x5ab5U,0x3c62U,0xfd70U,0x0a1cU,0x6823U,
		0xcdc9U,0x5fb4U,0x50a7U,0xbeeaU,0x557aU,0xc13dU,0x3e3eU,0x1212U,0x17b7U,0xa0d1U,
		0x8ca4U,0xd1b7U,0xde47U,0x39a5U,0x5b03U,0x3f12U,0xaaf7U,0xa6afU,0xf786U,0xc7feU,
		0x6edbU,0xfc08U,0x7aaaU,0xf8a7U,0x827eU,0xb988U,0x8a08U,0xb391U,0x0e0eU,0x9e9eU,
		0xa66bU,0x71faU,0xa031U,0x391cU,0xf5bfU,0xa089U,0x51e4U,0xfab0U,0x7b98U,0x0b39U,
		0xfc44U,0xe690U,0xa79dU,0x0fb1U,0xce7fU,0x5455U,0x7993U,0xd1d2U,0xd77aU,0x95d5U,
		0x72e8U,0x6ad1U,0xde2bU,0xbe2fU,0x0106U,0xedb1U,0x46dbU,0x8342U,0x1ba2U,0x96f3U,
		0xa3b4U,0x2806U,0x836cU,0x5b43U,0x178fU,0x2d9cU,0xbfd8U,0x0fcbU,0xe8f9U,0xb4ebU,
		0x99ebU,0x63f7U,0x80e7U,0xc709U,0x18eeU,0xf342U,0xdbf8U,0x3cb5U,0x5668U,0xab91U,
		0x3602U,0x2192U,0xca65U,0xcf36U,0x4fa0U,0xfb46U,0x8d92U,0xdc9dU,0x658dU,0x4546U,
		0x3a93U,0x1d1eU,0xeb99U,0xf763U,0x34e1U,0xfc7eU,0x61d2U,0x5fbeU,0xe8d6U,0x117eU,
		0x1bd9U,0x17c0U,0x0a23U,0x8a10U,0xe6dfU,0xc50dU,0xb4f5U,0x1811U,0x4703U,0x1652U,
		0x21f8U,0x094bU,0x5213U,0xd2eaU,0x871eU,0x849aU,0x48e8U,0x1a89U,0x7b47U,0xb01bU,
		0x2ec7U,0x09c9U,0x8333U,0x5dd1U,0x3858U,0xe782U,0x6485U,0xfed9U,0xd0a3U,0xe38dU,
		0x92beU,0x697cU,0x14beU,0xefe8U,0x443dU,0xfcadU,0xb6d4U,0x1dc5U,0xd72bU,0xb7e9U,
		0x58e4U,0x6860U,0x4761U,0x9cf7U,0xe9b7U,0x2bd7U,0x0d5eU,0xdb91U,0xfb73U,0x15d8U,
		0x358cU,0x4db1U,0x9a30U,0xc299U,0x1c14U,0xbc87U,0x134bU,0x1dffU,0x78bcU,0xe16cU,
		0xbe64U,0x316eU,0x99fdU,0x0f68U,0x7dc7U,0x71d2U,0x4bc7U,0x960dU,0x4cbfU,0x8be5U,
		0x57c4U,0x84fbU,0xa94cU,0xe6f1U,0x97feU,0x0786U,0x31abU,0xc73cU,0x600bU,0xc767U,
		0xc1f1U,0x2ccfU,0xdc72U,0xca2eU,0x92cfU,0xb32fU,0xf9deU,0xa87eU,0x87f9U,0xda4eU,
		0x3950U,0x8cf5U,0xd266U,0x996dU,0x8750U,0x800fU,0xe230U,0x4a31U,0xe733U,0xe795U,
		0xe9f5U,0x5079U,0x700eU,0x0e70U,0x31b4U,0x5a70U,0x07acU,0x92eaU,0xd136U,0x36d1U,
		0x4413U,0x68d6U,0x72c6U,0x65ccU,0x7550U,0x45cfU,0xcaafU,0xcf49U,0xeddbU,0xa4deU,
		0x5b79U,0x5da9U,0xf7b5U,0xf416U,0x3b9fU,0x29c7U,0xd001U,0x932dU,0x2730U,0xce51U,
		0x682aU,0x7799U,0x5ca1U,0x04abU,0x9d30U,0x4c44U,0x7042U,0xb07fU,0x984bU,0x0cdaU,
		0x1a52U,0x715eU,0x56a6U,0x04dcU,0xce26U,0xb845U,0xcff3U,0x1898U,0xe66fU,0x8ac1U,
		0xc81dU,0x49f9U,0xb4acU,0xa201U,0x4d8cU,0x1f53U,0x4d56U,0x0052U,0xf208U,0xa910U,
		0x08cdU,0xd5fcU,0x9364U,0xd118U,0x431bU,0x05a7U,0x2ba2U,0xe94eU,0x48a3U,0x3320U,
		0x77a7U,0xa777U,0x0175U,0xe20fU,0x0755U,0xf45cU,0xae74U,0x74aeU,0x33b9U,0xe6c7U,
		0xa9e5U,0xc6dfU,0x594fU,0x9fc5U,0xd66fU,0x54b9U,0x0b75U,0xc4e9U,0x90e3U,0xaf68U,
		0xccefU,0x6477U,0x25aeU,0x28c2U,0xc1d6U,0x9e54U,0xd3c8U,0x8b2eU,0x3584U,0xba5eU,
		0x0cc0U,0x1c88U,0xb84fU,0x457cU,0xba19U,0x28b8U,0x0d50U,0xb339U,0xcc73U,0xb31aU,
		0x605dU,0x0586U,0xe78cU,0x9724U,0xae82U,0xade0U,0x52d8U,0x7a85U,0xde97U,0x91f2U,
		0x493eU,0x4bb9U,0x7f00U,0x93dfU,0x41f9U,0x65b3U,0xcfccU,0x003bU,0xb48eU,0x40efU,
		0x4238U,0x4fcbU,0x6fa0U,0x3b22U,0x8f2bU,0xce11U,0x7000U,0x1d9cU,0xe138U,0x1cf0U,
		0x11ceU,0x2b8fU,0xd60aU,0x0878U,0x226bU,0xd9eeU,0xeb29U,0xb3f8U,0xd347U,0x7b24U,
		0x9efcU,0xf803U,0xd4c7U,0x6313U,0xb926U,0x5cd9U,0x6ca2U,0xb0d8U,0x3c69U,0xb399U,
		0xa767U,0x7e5cU,0xcabdU,0xc83cU,0xa057U,0x11c4U,0x4a0eU,0x081dU,0xa184U,0x0a84U,
		0x48d1U,0xed62U,0xe446U,0x3e17U,0xdda6U,0xa415U,0x6394U,0xfec9U,0xcc09U,0x9473U,
		0xf132U,0x86f9U,0x2933U,0xf6c4U,0x5615U,0xf5f7U,0xbb81U,0x8d4bU,0x5ea9U,0xaeb0U,
		0x5087U,0x0f80U,0x3371U,0x295eU,0xdd21U,0xf834U,0xa557U,0xdc13U,0x511bU,0x1b51U,
		0xaec3U,0x67d6U,0x9bd8U,0x5df2U,0xdabbU,0xbbdaU,0xba66U,0xa71fU,0xdba2U,0x3fa4U,
		0xead9U,0xeb40U,0xafcaU,0x49cfU,0x12b5U,0xee7fU,0xedd2U,0x6c59U,0xebb8U,0x5137U,
		0x756eU,0xe7d6U,0xe087U,0x2c25U,0x5763U,0xf7f9U,0x5f84U,0x35baU,0xcdedU,0x6466U,
		0x6102U,0x9534U,0xb1cbU,0x7242U,0xf9ecU,0xd766U,0x6606U,0xc60dU,0xbdc7U,0x0d93U,
		0xd0c3U,0x155aU,0x1f6cU,0x2a41U,0x246aU,0xa123U,0xcbaeU,0x82e4U,0x6669U,0x39d6U,
		0xebd0U,0x4511U,0x09aeU,0xe8afU,0x309aU,0x99c2U,0xdde0U,0x6760U,0x615aU,0x8ed9U,
		0x53b2U,0x6a62U,0x19b7U,0x2b09U,0x8ec5U,0xd1a2U,0x4c11U,0x33f0U,0x092bU,0xb719U,
		0xb690U,0x3260U,0x39f2U,0x9ff8U,0xf0e8U,0x99d4U,0xbe92U,0x7c69U,0x8509U,0xeb50U,
		0xcec3U,0x447cU,0x8ab2U,0x8265U,0x3dbbU,0x7ef6U,0xfd77U,0x45f5U,0xbbc4U,0x4e81U,
		0xd081U,0xe4a0U,0xb872U,0xa6d0U,0x0834U,0x4280U,0x73e1U,0x6e44U,0x7d22U,0xaf59U,
		0xda6aU,0x6fc0U,0x686eU,0x98b9U,0x1d88U,0x76c9U,0x1253U,0xeeebU,0xa336U,0x0c2bU,
		0x6d93U,0x87b6U,0x1db6U,0xd7c5U,0x47d3U,0x247bU,0xc025U,0x5efdU,0x6cb6U,0x283fU,
		0xc2b7U,0xfc3eU,0xc236U,0xbebbU,0x3c45U,0x483eU,0xbd15U,0x873aU,0xbbbeU,0x36c2U,
		0xa761U,0x575dU,0x5c9bU,0xd3c2U,0xe6d9U,0x4cb9U,0xcccfU,0x3b00U,0x1359U,0xcf76U,
		0xc3b6U,0xbaf8U,0xc9aaU,0x400aU,0xd633U,0x40e5U,0x1526U,0x0067U,0xab81U,0xb146U,
		0xb945U,0xc599U,0x36c8U,0xb435U,0xe4d3U,0x736bU,0x1498U,0xeed1U,0x0959U,0xdd31U,
		0x6dacU,0xabc8U,0x1e45U,0xab0fU,0xea59U,0xd7f2U,0x868cU,0xd7b6U,0x3b8dU,0x705eU,
		0x1df7U,0x49c7U,0x5e56U,0x5426U,0x8f50U,0x323eU,0x8d19U,0x8e1dU,0x6493U,0x18d1U,
		0xf7eaU,0x012eU,0x3a64U,0x207aU,0x762fU,0xc7dfU,0x1fa0U,0xa01fU,0x1337U,0xe295U,
		0x988bU,0x27e8U,0xae5aU,0x5aaeU,0x68feU,0xfb00U,0xeeabU,0x0a75U,0x89bdU,0x25dfU,
		0x6fd6U,0xb954U,0x92a3U,0x6a87U,0x946cU,0xeee4U,0x5dcaU,0x49f2U,0xc7c2U,0x9482U,
		0xad5cU,0x2b1aU,0x06e8U,0xa0daU,0xfffcU,0x8286U,0xc9b0U,0x3bd2U,0xb413U,0x4c24U,
		0x221dU,0x23afU,0x0510U,0x90faU,0x1f14U,0x7e2fU,0x2294U,0x3968U,0x844bU,0x8971U,
		0xd9beU,0x2e32U,0xb0f6U,0x4185U,0xc033U,0x92aaU,0x2364U,0xf298U,0x1eaaU,0x3de1U,
		0x9312U,0xc07cU,0xcb7bU,0x2fe6U,0x0338U,0xa930U,0x0dd3U,0x0eedU,0x2e99U,0xa6f5U,
		0xada3U,0xd303U,0x9160U,0x1868U,0x41b8U,0x4dceU,0xf301U,0x4a4cU,0x87daU,0x32ebU,
		0xd6bbU,0xacaeU,0xd3d0U,0xa7c5U,0x170cU,0x6c80U,0x68ebU,0xc0f4U,0x66a8U,0x3b95U,
		0xbccbU,0x4153U,0xb447U,0x4826U,0xfd85U,0xed84U,0xaa45U,0x0570U,0x1384U,0x1575U,
		0x4010U,0x672eU,0x3560U,0x1035U,0x64f9U,0xc65eU,0x86c1U,0x8799U,0xc92fU,0x3205U,
		0x9a66U,0x68a0U,0xa638U,0x8a36U,0x82fcU,0x5c0eU,0x831dU,0xcaccU,0x1d3cU,0x84cbU,
		0x2df6U,0xa4a6U,0x9cc9U,0x4d6aU,0xefb8U,0x4e3aU,0xceadU,0xdcfaU,0x3490U,0x3c09U,
		0x1739U,0x972dU,0x7c40U,0x1efbU,0x45b3U,0x369dU,0xc745U,0x5832U,0x98c7U,0xcdfdU,
		0xbe5bU,0xb43eU,0x2886U,0x9b6dU,0x7ea4U,0x90adU,0xa355U,0xdb40U,0xcbf7U,0x4c3aU,
		0x6c62U,0xfa98U,0xc5b4U,0x3d9cU,0xbd80U,0x6450U,0xe5ccU,0x0208U,0xdcafU,0xc471U,
		0x3740U,0x06d5U,0x5abcU,0xb705U,0xc3bcU,0x8672U,0x80d1U,0xdcd5U,0x0bfbU,0x09dbU,
		0x5637U,0x1227U,0xeb7bU,0xf72cU,0xcc87U,0x711dU,0x4f37U,0xf322U,0x2173U,0x5c48U,
		0x61e3U,0x2e0eU,0x24c8U,0x6a8eU,0x0631U,0x97e5U,0xb0bbU,0x029aU,0x767fU,0xa19dU,
		0x75bcU,0x30f0U,0x797fU,0x8b84U,0x8f88U,0xd879U,0x00b9U,0x54f1U,0x8d27U,0x7336U,
		0x74daU,0xcd16U,0x2908U,0x1762U,0xed01U,0x3fb3U,0x3140U,0xa51dU,0xd9d4U,0xcfe9U,
		0x6ef9U,0xfad0U,0x433eU,0x534fU,0xbbf1U,0x1bf8U,0xb5c4U,0x1845U,0xa9caU,0x2a9aU,
		0x94c7U,0x1b21U,0x3549U,0x05c4U,0x7522U,0x69beU,0x1266U,0x4ba7U,0xbafeU,0x5845U,
		0x014fU,0x2a64U,0x8e33U,0xa5edU,0xd85fU,0xd84eU,0x7fadU,0x33d3U,0xee6cU,0x6938U,
		0x31d8U,0xacccU,0xc06aU,0x6a20U,0x8602U,0x530fU,0x8915U,0xfea6U,0xbf9fU,0xf2b3U,
		0xab1dU,0xa133U,0xc40eU,0x64bbU,0x7f55U,0xd69cU,0xe1e7U,0x41e2U,0xc48cU,0x39bdU,
		0x930aU,0x3380U,0xd09bU,0x560dU,0x2440U,0x21e7U,0xd590U,0x7198U,0x10b3U,0xa1f7U,
		0x1534U,0xccbbU,0x5fa2U,0xf2dfU,0x80a9U,0x8bfaU,0x84f2U,0x102cU,0x6c29U,0x42bdU,
		0x19a6U,0x1288U,0xec31U,0xc2faU,0xd162U,0x1afbU,0xd5e9U,0x4e05U,0x0e89U,0x351dU,
		0x242eU,0xb990U,0x135fU,0xe891U,0xb26fU,0x51d4U,0x4126U,0x7354U,0xf6d2U,0x19c7U,
		0xc725U,0x159cU,0xe262U,0xfdbeU,0xe0a9U,0x6ae6U,0x6290U,0x58d2U,0x4306U,0x4d3aU,
		0xc88aU,0x6fddU,0x2b74U,0xbf21U,0x5706U,0x0398U,0x3388U,0x731bU,0x4a1cU,0x0be7U,
		0x072aU,0xb19eU,0xd975U,0x905cU,0x6137U,0x97b2U,0x9007U,0x3e5aU,0xea77U,0x4899U,
		0x9176U,0x2769U,0x4eccU,0x20b3U,0x40acU,0x0af1U,0x72fcU,0x85c1U,0x8b36U,0xc2e3U,
		0xa98cU,0x6111U,0xd941U,0xaa65U,0x2bbaU,0x7066U,0x728eU,0x3981U,0x7617U,0x4053U,
		0x558dU,0xe033U,0xdebeU,0x51a9U,0x4e63U,0xc17cU,0xcf7cU,0xeb03U,0x6f4aU,0x302dU,
		0x7bd9U,0xbaa6U,0x9ca6U,0xee37U,0x297fU,0xde03U,0xff32U,0xf4e0U,0x21b7U,0xac57U,
		0x49ddU,0x4812U,0xdfd4U,0xb122U,0x9d6eU,0xa3e1U,0x57d7U,0xbf87U,0xa460U,0x5e36U,
		0xdc69U,0x82f3U,0xbc17U,0x851aU,0xbc66U,0x901eU,0xd057U,0x76a1U,0x24afU,0xd10eU,
		0xa1d3U,0x79ccU,0x9bc3U,0x1e14U,0xef35U,0xc441U,0x887dU,0xed3cU,0x25ffU,0x95f0U,
		0x9af1U,0x7a40U,0x9376U,0x3d23U,0xf95eU,0xbed3U,0xa565U,0x100dU,0x2f92U,0x13eaU,
		0x1071U,0xcd7bU,0x73afU,0x758bU,0xb834U,0xaa35U,0xffe7U,0x1434U,0x1cc2U,0x6041U,
		0xccd5U,0x53c1U,0x42e2U,0x990bU,0xcf0fU,0x44ccU,0x6173U,0x498cU,0xb742U,0xf21bU,
		0x5f92U,0x0a64U,0x66efU,0x7a9cU,0x81cdU,0x549bU,0x5d84U,0x3457U,0xb574U,0xb95bU,
		0x2595U,0x9555U,0xac6aU,0x019fU,0x8a25U,0x071aU,0x7a68U,0x8df3U,0x32c1U,0x5945U,
		0xf5b1U,0x2529U,0xd7a4U,0xf1b3U,0x713fU,0x2306U,0x2f38U,0xc69dU,0xab2eU,0xf983U,
		0x66dfU,0x85b0U,0x4aacU,0x77f9U,0x9e3cU,0xe27bU,0x00d1U,0x565aU,0xac27U,0x58c0U,
		0xbd4fU,0xb6a2U,0xe192U,0xf4cdU,0x2a4fU,0xeb60U,0xef09U,0xf550U,0x24e9U,0xbff1U,
		0x140dU,0x4cd3U,0x7741U,0x2720U,0x3b06U,0x9995U,0xddeeU,0x627cU,0x7710U,0xdfa2U,
		0x5391U,0x77dfU,0xa14aU,0x7c78U,0x437eU,0xe856U,0x1945U,0xb602U,0xfb63U,0x9e07U,
		0xc71eU,0xb73bU,0x1f0dU,0xeb74U,0xa17dU,0x73daU,0x520cU,0x1a4bU,0xc983U,0xe205U,
		0xdd7dU,0x4a62U,0xbe1eU,0xf7daU,0xa006U,0x5a22U,0xda45U,0xa49dU,0xe9e2U,0x4870U,
		0x74f9U,0xe931U,0x265eU,0xdf30U,0x2580U,0xc371U,0x0617U,0xf1cbU,0x65f6U,0x59f2U,
		0xc3dcU,0xc859U,0xd2fbU,0x3e52U,0xb7e1U,0xf82cU,0xca07U,0xef7dU,0x4c5eU,0x3f55U,
		0x453cU,0x3e48U,0x4f55U,0xff14U,0xff13U,0x3e47U,0x3e45U,0x3e46U,0xb94bU,0x3e49U,
		0xbfe2U,0x453dU,0x29eaU,0xff15U,0xdf22U,0x4f56U,0xa00dU,0xff16U,0xf8b3U,0x29ebU,
		0xa03cU,0x4f57U,0x060bU,0xdf23U,0x1024U,0x3e4aU,0xd9e6U,0xb94cU,0x463eU,0x453eU,
		0x6962U,0xbfe3U,0x80b1U,0x4f58U,0xf667U,0xa03dU,0xb5f5U,0xdf24U,0xf25bU,0x060cU,
		0x95e1U,0xff17U,0x42adU,0xa00eU,0x564cU,0x29ecU,0x7a1aU,0xf8b4U,0x514aU,0x453fU,
		0xeab8U,0x463fU,0xbe12U,0xbfe4U,0x6065U,0x6963U,0xbeccU,0x3e4bU,0x4c33U,0x1025U,
		0x9ff1U,0xb94dU,0xcf42U,0xd9e7U,0xbd89U,0xdf25U,0x070dU,0xb5f6U,0xca5bU,0x060dU,
		0x5cf2U,0xf25cU,0xc59fU,0x4f59U,0x5e28U,0x80b2U,0x2a31U,0xa03eU,0x30e1U,0xf668U,
		0x9ab5U,0x29edU,0x5200U,0x564dU,0x25beU,0xf8b5U,0xef3eU,0x7a1bU,0xe619U,0xff18U,
		0x3713U,0x95e2U,0x0fa0U,0xa00fU,0xd0f2U,0x42aeU,0x44bfU,0xbfe5U,0xc042U,0xbe13U,
		0x3d82U,0x6964U,0x60dcU,0x6066U,0xb982U,0x4540U,0xf032U,0x514bU,0x3959U,0x4640U,
		0xbc77U,0xeab9U,0x3afaU,0xb94eU,0x3f9eU,0x9ff2U,0xc6d9U,0xd9e8U,0xc293U,0xcf43U,
		0x58baU,0x3e4cU,0x610bU,0xbecdU,0x5349U,0x1026U,0x6c7aU,0x4c34U,0x68f5U,0x060eU,
		0x5ccaU,0xca5cU,0xd89bU,0xf25dU,0x1219U,0x5cf3U,0xab87U,0xdf26U,0xa704U,0xbd8aU,
		0x16f4U,0xb5f7U,0xeaf6U,0x070eU,0x5a29U,0xa03fU,0x2c6bU,0x2a32U,0x2134U,0xf669U,
		0x314aU,0x30e2U,0x5a4fU,0x4f5aU,0x7ee1U,0xc5a0U,0xffaaU,0x80b3U,0x7133U,0x5e29U,
		0x574eU,0xf8b6U,0x60c0U,0x25bfU,0x6a12U,0x7a1cU,0xa733U,0xef3fU,0x95beU,0x29eeU,
		0xf01bU,0x9ab6U,0x9011U,0x564eU,0xaac8U,0x5201U,0x0d02U,0xa010U,0x240bU,0x0fa1U,
		0xe3bbU,0x42afU,0xa4f0U,0xd0f3U,0x94dbU,0xff19U,0x7a72U,0xe61aU,0xe186U,0x95e3U,
		0x7f9bU,0x3714U,0x037cU,0x6965U,0x4145U,0x3d83U,0xe0ddU,0x6067U,0x8000U,0x60ddU,
		0xb14cU,0xbfe6U,0x158fU,0x44c0U,0xe8efU,0xbe14U,0x56b0U,0xc043U,0x7ec8U,0x4641U,
		0x171bU,0x395aU,0x0d3bU,0xeabaU,0x6ce1U,0xbc78U,0x833dU,0x4541U,0x1cb6U,0xb983U,
		0x3ae9U,0x514cU,0x114cU,0xf033U,0x971fU,0xd9e9U,0x8abcU,0xc6daU,0xb32aU,0xcf44U,
		0x09c4U,0xc294U,0x7059U,0xb94fU,0x76c4U,0x3afbU,0x1555U,0x9ff3U,0xfec4U,0x3f9fU,
		0x530aU,0x1027U,0x2e09U,0x534aU,0xcac7U,0x4c35U,0x4180U,0x6c7bU,0xb736U,0x3e4dU,
		0x3452U,0x58bbU,0xb11dU,0xbeceU,0x4d35U,0x610cU,0xb101U,0xf25eU,0x9cd8U,0xd89cU,
		0x5507U,0x5cf4U,0x5581U,0x121aU,0x152cU,0x060fU,0xe130U,0x68f6U,0xf249U,0xca5dU,
		0x7a51U,0x5ccbU,0xbe25U,0xb5f8U,0xfa27U,0x16f5U,0x49a4U,0x070fU,0x2e01U,0xeaf7U,
		0x27dfU,0xdf27U,0xe052U,0xab88U,0x7d46U,0xbd8bU,0x1b43U,0xa705U,0x8c02U,0xf66aU,
		0x2eadU,0x2135U,0x21abU,0x30e3U,0x83fbU,0x314bU,0xf0cdU,0xa040U,0xfe50U,0x5a2aU,
		0xeb86U,0x2a33U,0x5d43U,0x2c6cU,0xe009U,0x80b4U,0x7ac4U,0xffabU,0xcfdcU,0x5e2aU,
		0x058eU,0x7134U,0x8111U,0x4f5bU,0x31e3U,0x5a50U,0x3252U,0xc5a1U,0x0c46U,0x7ee2U,
		0xa228U,0x7a1dU,0xef29U,0x6a13U,0x8eb4U,0xef40U,0xfbc8U,0xa734U,0x006dU,0xf8b7U,
		0x3ca0U,0x574fU,0xfd5eU,0x25c0U,0x7be4U,0x60c1U,0x22dfU,0x564fU,0x1b14U,0x9012U,
		0x8362U,0x5202U,0x0a40U,0xaac9U,0x136aU,0x29efU,0x87a8U,0x95bfU,0x625aU,0x9ab7U,
		0x38e1U,0xf01cU,0xf952U,0x42b0U,0x1418U,0xe3bcU,0xe71bU,0xd0f4U,0x66ceU,0xa4f1U,
		0xcf22U,0xa011U,0x2c12U,0x0d03U,0x2d49U,0x0fa2U,0xf286U,0x240cU,0x21daU,0x95e4U,
		0x7175U,0xe187U,0x73baU,0x3715U,0xbcecU,0x7f9cU,0xc58bU,0xff1aU,0x590fU,0x94dcU,
		0xcbc4U,0xe61bU,0x1989U,0x7a73U,0xe458U,0x6068U,0xd06eU,0xe0deU,0x0888U,0x60deU,
		0xc144U,0x8001U,0xd639U,0x6966U,0x11d9U,0x037dU,0x91c1U,0x3d84U,0xc87eU,0x4146U,
		0xf7e1U,0xbe15U,0xd499U,0xe8f0U,0x8f15U,0xc044U,0x0778U,0x56b1U,0xa07fU,0xbfe7U,
		0x8c35U,0xb14dU,0x0180U,0x44c1U,0xa6e8U,0x1590U,0xe468U,0xeabbU,0x036cU,0x0d3cU,
		0x532aU,0xbc79U,0x5b84U,0x6ce2U,0x12cfU,0x4642U,0x04efU,0x7ec9U,0x411aU,0x395bU,
		0x0915U,0x171cU,0x9a3bU,0x514dU,0xc5c3U,0x3aeaU,0xb00dU,0xf034U,0x6c3eU,0x114dU,
		0xc705U,0x4542U,0xe68cU,0x833eU,0xa0cdU,0xb984U,0x8bb2U,0x1cb7U,0x5075U,0xcf45U,
		0x8be1U,0xb32bU,0x1dc1U,0xc295U,0xc509U,0x09c5U,0x40ebU,0xd9eaU,0x8b2aU,0x9720U,
		0x05a3U,0xc6dbU,0x04a7U,0x8abdU,0x675cU,0x9ff4U,0xf7f5U,0x1556U,0x295aU,0x3fa0U,
		0x7e58U,0xfec5U,0xc7dbU,0xb950U,0xb431U,0x705aU,0xfc3aU,0x3afcU,0x4e7dU,0x76c5U,
		0x1ef7U,0x4c36U,0x672aU,0xcac8U,0xa6f1U,0x6c7cU,0x3bceU,0x4181U,0x337cU,0x1028U,
		0x5841U,0x530bU,0x54edU,0x534bU,0x866eU,0x2e0aU,0xd10aU,0xbecfU,0xeaffU,0xb11eU,
		0x97aeU,0x610dU,0xe88dU,0x4d36U,0xa499U,0x3e4eU,0x9991U,0xb737U,0xf1afU,0x58bcU,
		0x53bdU,0x3453U,0x3bdeU,0x5cf5U,0xfbb7U,0x5508U,0x3e0aU,0x121bU,0x50aeU,0x5582U,
		0x095fU,0xf25fU,0xc42bU,0xb102U,0xd21aU,0xd89dU,0xa806U,0x9cd9U,0xdd84U,0xca5eU,
		0xed10U,0xf24aU,0x3734U,0x5cccU,0xd30fU,0x7a52U,0x25f8U,0x0610U,0x287fU,0x152dU,
		0x4c0aU,0x68f7U,0x440cU,0xe131U,0x5faaU,0x0710U,0xa9b7U,0x49a5U,0x716eU,0xeaf8U,
		0x3f97U,0x2e02U,0xd7e9U,0xb5f9U,0x23d1U,0xbe26U,0xc66aU,0x16f6U,0x968dU,0xfa28U,
		0xf2c9U,0xbd8cU,0xd3a3U,0x7d47U,0x2db0U,0xa706U,0xcc7bU,0x1b44U,0x7190U,0xdf28U,
		0xce09U,0x27e0U,0xd464U,0xab89U,0x1697U,0xe053U,0x6459U,0x30e4U,0xf5d2U,0x21acU,
		0x40cfU,0x314cU,0xe965U,0x83fcU,0xdd37U,0xf66bU,0xa1acU,0x8c03U,0x1e61U,0x2136U,
		0x30fcU,0x2eaeU,0x00adU,0x2a34U,0xe311U,0xeb87U,0x0cacU,0x2c6dU,0xc44aU,0x5d44U,
		0x0214U,0xa041U,0x7d8aU,0xf0ceU,0x58f7U,0x5a2bU,0x3cefU,0xfe51U,0xd65dU,0x5e2bU,
		0x39c9U,0xcfddU,0x8c9aU,0x7135U,0x2cb5U,0x058fU,0x6d62U,0x80b5U,0x8b54U,0xe00aU,
		0x8cfeU,0xffacU,0x721bU,0x7ac5U,0xf635U,0xc5a2U,0xa9beU,0x3253U,0x69caU,0x7ee3U,
		0x7020U,0x0c47U,0xcb2aU,0x4f5cU,0x953dU,0x8112U,0x177fU,0x5a51U,0xa868U,0x31e4U,
		0x3793U,0xef41U,0x22a9U,0x8eb5U,0xd55fU,0xa735U,0xbef1U,0xfbc9U,0x149eU,0x7a1eU,
		0xcd87U,0xa229U,0x5ef0U,0x6a14U,0x3128U,0xef2aU,0x4a69U,0x25c1U,0xd623U,0xfd5fU,
		0x3bafU,0x60c2U,0x4efbU,0x7be5U,0x37d8U,0xf8b8U,0x10e8U,0x006eU,0xbf93U,0x5750U,
		0x0b04U,0x3ca1U,0x8f47U,0x5203U,0xcc96U,0x8363U,0xca92U,0xaacaU,0xaa71U,0x0a41U,
		0xf89dU,0x5650U,0x73f9U,0x22e0U,0xa0bbU,0x9013U,0x4e74U,0x1b15U,0x4d46U,0x9ab8U,
		0xa835U,0x625bU,0x651fU,0xf01dU,0x57eeU,0x38e2U,0x4b8bU,0x29f0U,0xac05U,0x136bU,
		0xbdb3U,0x95c0U,0x0143U,0x87a9U,0x63e9U,0xd0f5U,0xb2ccU,0xe71cU,0xb60eU,0xa4f2U,
		0x13d9U,0x66cfU,0xeed7U,0x42b1U,0x9b5aU,0xf953U,0x55cfU,0xe3bdU,0x602fU,0x1419U,
		0x1aaaU,0x0fa3U,0xc865U,0x2d4aU,0x024fU,0x240dU,0xd152U,0xf287U,0x10a7U,0xa012U,
		0x8b96U,0xcf23U,0xcb59U,0x0d04U,0x9455U,0x2c13U,0x53e9U,0x3716U,0x71ecU,0x73bbU,
		0xbae7U,0x7f9dU,0xbd9dU,0xbcedU,0xf227U,0x95e5U,0xabbbU,0x21dbU,0x0e04U,0xe188U,
		0xa303U,0x7176U,0xf520U,0xe61cU,0x49acU,0xcbc5U,0xefa8U,0x7a74U,0x2030U,0x198aU,
		0x4e92U,0xff1bU,0xc480U,0xc58cU,0x83b7U,0x94ddU,0x7805U,0x5910U,0xe3e2U,0x60dfU,
		0x9e60U,0x0889U,0xd6afU,0x8002U,0xcdd0U,0xc145U,0xe4d9U,0x6069U,0x0234U,0xe459U,
		0x5a0bU,0xe0dfU,0xc9ccU,0xd06fU,0xc98aU,0x3d85U,0x9c3fU,0x91c2U,0x65bfU,0x4147U,
		0xb4c7U,0xc87fU,0x174eU,0x6967U,0xa48aU,0xd63aU,0x56d3U,0x037eU,0x6f38U,0x11daU,
		0xf2e7U,0xc045U,0xa255U,0x8f16U,0x6e59U,0x56b2U,0xcb6fU,0x0779U,0xe1dfU,0xbe16U,
		0x41f1U,0xf7e2U,0x406aU,0xe8f1U,0x9339U,0xd49aU,0x3b41U,0x44c2U,0xcc50U,0x0181U,
		0xbadfU,0x1591U,0xb322U,0xa6e9U,0x005eU,0xbfe8U,0x01d6U,0xa080U,0x4695U,0xb14eU,
		0x55aaU,0x8c36U,0xb0eaU,0xbc7aU,0x6e02U,0x532bU,0xf46cU,0x6ce3U,0xe812U,0x5b85U,
		0x7371U,0xeabcU,0x9745U,0xe469U,0x5a7cU,0x0d3dU,0x568dU,0x036dU,0x6853U,0x395cU,
		0x50e0U,0x411bU,0xc960U,0x171dU,0x8186U,0x0916U,0xdb6bU,0x4643U,0x0792U,0x12d0U,
		0x6b97U,0x7ecaU,0x5a40U,0x04f0U,0x1dd1U,0xf035U,0x0504U,0xb00eU,0x6802U,0x114eU,
		0x6f88U,0x6c3fU,0x2e52U,0x514eU,0x2ae1U,0x9a3cU,0xce99U,0x3aebU,0x7dfbU,0xc5c4U,
		0xa6acU,0xb985U,0x5fb1U,0xa0ceU,0xf905U,0x1cb8U,0x181dU,0x8bb3U,0x218fU,0x4543U,
		0x2d99U,0xc706U,0x95d2U,0x833fU,0x3919U,0xe68dU,0x2bd4U,0xc296U,0xe38aU,0x1dc2U,
		0xd2e7U,0x09c6U,0x5fbbU,0xc50aU,0x36ceU,0xcf46U,0x996aU,0x5076U,0xc739U,0xb32cU,
		0x316bU,0x8be2U,0xe20cU,0xc6dcU,0x004fU,0x05a4U,0x715bU,0x8abeU,0x29c4U,0x04a8U,
		0x1d99U,0xd9ebU,0x4bb6U,0x40ecU,0x28b5U,0x9721U,0xaf65U,0x8b2bU,0x67d3U,0x3fa1U,
		0xf5f4U,0x295bU,0x0a81U,0xfec6U,0x6310U,0x7e59U,0xd19fU,0x9ff5U,0x39d3U,0x675dU,
		0x723fU,0x1557U,0x6c56U,0xf7f6U,0x36bfU,0x3afdU,0xd7c2U,0xfc3bU,0x6e41U,0x76c6U,
		0x4479U,0x4e7eU,0x5aabU,0xb951U,0x8e1aU,0xc7dcU,0xabc5U,0x705bU,0x40e2U,0xb432U,
		0x4a49U,0x6c7dU,0xc079U,0xa6f2U,0x7e2cU,0x4182U,0x947fU,0x3bcfU,0xb43bU,0x4c37U,
		0x4e37U,0x1ef8U,0x3202U,0xcac9U,0x4823U,0x672bU,0x3fb0U,0x534cU,0xa19aU,0x54eeU,
		0xf729U,0x2e0bU,0x0205U,0x866fU,0xa1f4U,0x1029U,0x64b8U,0x337dU,0x33d0U,0x530cU,
		0x1b1eU,0x5842U,0x20b0U,0x610eU,0x7318U,0x97afU,0x1599U,0x4d37U,0x1af8U,0xe88eU,
		0xed39U,0xbed0U,0x82f0U,0xd10bU,0xde00U,0xb11fU,0x4050U,0xeb00U,0x85adU,0x58bdU,
		0x0717U,0xf1b0U,0xf218U,0x3454U,0x7588U,0x53beU,0xc36eU,0x3e4fU,0xe202U,0xa49aU,
		0x7c75U,0xb738U,0xf54dU,0x9992U,0xba78U,0x121cU,0x6df8U,0x3e0bU,0xf426U,0x5583U,
		0x5b0aU,0x50afU,0x8692U,0x5cf6U,0x0fd5U,0x3bdfU,0x4c8fU,0x5509U,0x527cU,0xfbb8U,
		0x2665U,0xd89eU,0x2927U,0xd21bU,0x2d0dU,0x9cdaU,0xb1f7U,0xa807U,0x7881U,0xf260U,
		0x70dcU,0x0960U,0xd92eU,0xb103U,0xe87dU,0xc42cU,0x153cU,0x5ccdU,0x619cU,0x3735U,
		0x8169U,0x7a53U,0x812aU,0xd310U,0xf780U,0xca5fU,0xd014U,0xdd85U,0x4c81U,0xf24bU,
		0x9209U,0xed11U,0xa75bU,0x68f8U,0x9bd2U,0x4c0bU,0x074fU,0xe132U,0x0d58U,0x440dU,
		0x4aa6U,0x0611U,0x40a6U,0x25f9U,0x313aU,0x152eU,0x87d4U,0x2880U,0xc5bdU,0xeaf9U,
		0x7abeU,0x716fU,0x2405U,0x2e03U,0xeab2U,0x3f98U,0xd7bcU,0x0711U,0xcc4aU,0x5fabU,
		0xa82fU,0x49a6U,0xd39dU,0xa9b8U,0xabe7U,0x16f7U,0x01e9U,0xc66bU,0x528aU,0xfa29U,
		0x9bccU,0x968eU,0xacc4U,0xb5faU,0xb312U,0xd7eaU,0xc9e3U,0xbe27U,0xf7a3U,0x23d2U,
		0xf997U,0xa707U,0x1ad4U,0x2db1U,0x1c53U,0x1b45U,0x9ddaU,0xcc7cU,0x893fU,0xbd8dU,
		0x13f9U,0xf2caU,0x8802U,0x7d48U,0x3064U,0xd3a4U,0x1fb9U,0xab8aU,0x9694U,0xd465U,
		0x5e9dU,0xe054U,0xa537U,0x1698U,0xc43aU,0xdf29U,0xcdf6U,0x7191U,0x2b02U,0x27e1U,
		0x92e3U,0xce0aU,0xc0feU,0x314dU,0x9cb5U,0x40d0U,0x8213U,0x83fdU,0x3f19U,0xe966U,
		0xea5fU,0x30e5U,0xc956U,0x645aU,0x89ddU,0x21adU,0x20c4U,0xf5d3U,0xdf37U,0x2137U,
		0x1977U,0x1e62U,0xf712U,0x2eafU,0xa527U,0x30fdU,0x913dU,0xf66cU,0xe07bU,0xdd38U,
		0x641aU,0x8c04U,0xb641U,0xa1adU,0xd2a7U,0x2c6eU,0x3890U,0x0cadU,0xe104U,0x5d45U,
		0xfb97U,0xc44bU,0xdbd7U,0x2a35U,0xb548U,0x00aeU,0x3e64U,0xeb88U,0x9708U,0xe312U,
		0x635fU,0x5a2cU,0x5290U,0x58f8U,0x07bcU,0xfe52U,0x1fdbU,0x3cf0U,0xb1bfU,0xa042U,
		0x8dfdU,0x0215U,0x894dU,0xf0cfU,0x29a9U,0x7d8bU,0xc22dU,0x7136U,0xc24eU,0x8c9bU,
		0x7491U,0x0590U,0xe976U,0x2cb6U,0xd7f8U,0x5e2cU,0x2eeaU,0xd65eU,0x67b7U,0xcfdeU,
		0x04e6U,0x39caU,0x4d04U,0xffadU,0x6dcaU,0x8cffU,0xa81cU,0x7ac6U,0x15edU,0x721cU,
		0x5ddbU,0x80b6U,0x5e45U,0x6d63U,0xec8dU,0xe00bU,0x614eU,0x8b55U,0xae2aU,0x7ee4U,
		0x826fU,0x69cbU,0xcc58U,0x0c48U,0xb8b0U,0x7021U,0x9568U,0xc5a3U,0x24feU,0xf636U,
		0x0bb9U,0x3254U,0x4562U,0xa9bfU,0x5197U,0x5a52U,0xfa2fU,0x1780U,0xc846U,0x31e5U,
		0x7109U,0xa869U,0x2b63U,0x4f5dU,0x4fe4U,0xcb2bU,0xbe86U,0x8113U,0x6371U,0x953eU,
		0x4b43U,0xa736U,0xbd09U,0xd560U,0x793bU,0xfbcaU,0xde4eU,0xbef2U,0x1e4bU,0xef42U,
		0x6120U,0x3794U,0x0f4cU,0x8eb6U,0x3623U,0x22aaU,0x7500U,0x6a15U,0x7df1U,0x5ef1U,
		0xe4edU,0xef2bU,0x07ccU,0x3129U,0x310bU,0x7a1fU,0x3eeeU,0x149fU,0x3362U,0xa22aU,
		0x88aaU,0xcd88U,0xccc3U,0x60c3U,0xa4beU,0x3bb0U,0x581fU,0x7be6U,0x282bU,0x4efcU,
		0xb8c4U,0x25c2U,0xdf62U,0x4a6aU,0xb219U,0xfd60U,0xe2e5U,0xd624U,0x3841U,0x5751U,
		0x01efU,0xbf94U,0x3f27U,0x3ca2U,0xb341U,0x0b05U,0xe59cU,0xf8b9U,0xea28U,0x37d9U,
		0x8fd8U,0x006fU,0x1d79U,0x10e9U,0xfe60U,0xaacbU,0x3ad8U,0xca93U,0x85d8U,0x0a42U,
		0x8aa6U,0xaa72U,0xab15U,0x5204U,0xde8fU,0x8f48U,0x3183U,0x8364U,0x2dc6U,0xcc97U,
		0xb80cU,0x9014U,0x6146U,0xa0bcU,0x944dU,0x1b16U,0x1144U,0x4e75U,0x4cb0U,0x5651U,
		0x278fU,0xf89eU,0xde5fU,0x22e1U,0x3675U,0x73faU,0x4bf9U,0xf01eU,0x4ac6U,0x6520U,
		0xda5eU,0x38e3U,0x445fU,0x57efU,0x782aU,0x9ab9U,0x43a2U,0x4d47U,0xe112U,0x625cU,
		0x01baU,0xa836U,0x00e6U,0x95c1U,0xc671U,0xbdb4U,0x4f78U,0x87aaU,0xa43eU,0x0144U,
		0x6b2bU,0x29f1U,0x225cU,0x4b8cU,0xc575U,0x136cU,0x06a2U,0xac06U,0x1910U,0xa4f3U,
		0x2095U,0xb60fU,0xc25cU,0x66d0U,0x39acU,0x13daU,0x6db2U,0xd0f6U,0x89b1U,0x63eaU,
		0xc0a0U,0xe71dU,0xf44aU,0xb2cdU,0xe938U,0xe3beU,0x15bcU,0x55d0U,0xda36U,0x141aU,
		0xdf92U,0x6030U,0x6cffU,0x42b2U,0x2185U,0xeed8U,0x473dU,0xf954U,0x269fU,0x9b5bU,
		0x2b43U,0x240eU,0xbdeeU,0x0250U,0xfc9cU,0xf288U,0x5d55U,0xd153U,0x8d0dU,0x0fa4U,
		0x67c0U,0x1aabU,0x63c1U,0x2d4bU,0xf583U,0xc866U,0xf9ffU,0x0d05U,0xabedU,0xcb5aU,
		0xcd06U,0x2c14U,0xdfc5U,0x9456U,0x27f9U,0xa013U,0x6fecU,0x10a8U,0xae41U,0xcf24U,
		0x5893U,0x8b97U,0xc4c0U,0x7f9eU,0x9110U,0xbae8U,0x0983U,0xbceeU,0x91d9U,0xbd9eU,
		0xabceU,0x3717U,0x64daU,0x53eaU,0x766cU,0x73bcU,0x1debU,0x71edU,0xc3adU,0xe189U,
		0x11b4U,0x0e05U,0xa95cU,0x7177U,0x794eU,0xa304U,0xf1edU,0x95e6U,0x587dU,0xf228U,
		0xae38U,0x21dcU,0xa0c4U,0xabbcU,0x8153U,0x7a75U,0xb27eU,0xefa9U,0x0189U,0x198bU,
		0x6331U,0x2031U,0x9433U,0xe61dU,0x200aU,0xf521U,0x148cU,0xcbc6U,0x3399U,0x49adU,
		0x5a60U,0x94deU,0x16fdU,0x83b8U,0xadfbU,0x5911U,0x5255U,0x7806U,0x6a45U,0xff1cU,
		0xc190U,0x4e93U,0xfa54U,0xc58dU,0x6568U,0xc481U,0xece0U,0x8003U,0x1163U,0xd6b0U,
		0x4227U,0xc146U,0x8cabU,0xcdd1U,0x8f56U,0x60e0U,0x2121U,0xe3e3U,0x91b3U,0x088aU,
		0x8ff1U,0x9e61U,0xe9e9U,0xe0e0U,0xee17U,0x5a0cU,0x6b5fU,0xd070U,0xb45fU,0xc9cdU,
		0x5e67U,0x606aU,0x9e16U,0xe4daU,0xb355U,0xe45aU,0x1e7cU,0x0235U,0x10bbU,0x4148U,
		0xa480U,0x65c0U,0x279dU,0xc880U,0xad92U,0xb4c8U,0x585fU,0x3d86U,0xa7eaU,0xc98bU,
		0x031aU,0x91c3U,0x294cU,0x9c40U,0x377eU,0x037fU,0xf7a9U,0x56d4U,0xecebU,0x11dbU,
		0xba21U,0x6f39U,0xd4e6U,0x6968U,0x8243U,0x174fU,0xdae3U,0xd63bU,0x2c80U,0xa48bU,
		0x3da5U,0x56b3U,0x7750U,0x6e5aU,0xa475U,0x077aU,0xe2a8U,0xcb70U,0x3244U,0xc046U,
		0xc79aU,0xf2e8U,0x9c2dU,0x8f17U,0xfdcfU,0xa256U,0x9ac7U,0xe8f2U,0x425aU,0x406bU,
		0x7dbbU,0xd49bU,0x1e1dU,0x933aU,0xbc5fU,0xbe17U,0x3489U,0xe1e0U,0x09a7U,0xf7e3U,
		0xe229U,0x41f2U,0x67fbU,0x1592U,0x8c93U,0xbae0U,0xe714U,0xa6eaU,0x3d7bU,0xb323U,
		0xa6c7U,0x44c3U,0xe70dU,0x3b42U,0xda57U,0x0182U,0x1c4cU,0xcc51U,0x19deU,0xb14fU,
		0x23d8U,0x4696U,0xd73bU,0x8c37U,0x5fffU,0x55abU,0x05bcU,0xbfe9U,0x646fU,0x005fU,
		0x6b39U,0xa081U,0x865aU,0x01d7U,0x2282U,0x6ce4U,0x2329U,0xf46dU,0x1547U,0x5b86U,
		0xd1beU,0xe813U,0x5e5cU,0xbc7bU,0xf599U,0xb0ebU,0xf9afU,0x532cU,0xb888U,0x6e03U,
		0x4877U,0x0d3eU,0x1c5eU,0x5a7dU,0x88b8U,0x036eU,0xc6c5U,0x568eU,0xe9abU,0xeabdU,
		0x17edU,0x7372U,0x9a15U,0xe46aU,0xd438U,0x9746U,0x59b4U,0x171eU,0xb9c7U,0xc961U,
		0xfef2U,0x0917U,0xe3adU,0x8187U,0xdbe2U,0x395dU,0x5a01U,0x6854U,0x7a05U,0x411cU,
		0xfd1bU,0x50e1U,0x8e82U,0x7ecbU,0xc9e9U,0x6b98U,0xcb0eU,0x04f1U,0x6ddaU,0x5a41U,
		0x22b8U,0x4644U,0x4431U,0xdb6cU,0x3500U,0x12d1U,0x0bc4U,0x0793U,0x0f92U,0x114fU,
		0xbfc5U,0x6803U,0xbc97U,0x6c40U,0x2d7eU,0x6f89U,0x542cU,0xf036U,0x4f83U,0x1dd2U,
		0x8fe6U,0xb00fU,0x5c5cU,0x0505U,0x667dU,0x3aecU,0x16b8U,0xce9aU,0x5050U,0xc5c5U,
		0x62f7U,0x7dfcU,0xaff8U,0x514fU,0xb679U,0x2e53U,0x384cU,0x9a3dU,0xb814U,0x2ae2U,
		0xfd6eU,0x1cb9U,0xff05U,0xf906U,0x3b49U,0x8bb4U,0xfa5fU,0x181eU,0x9e9cU,0xb986U,
		0xfc06U,0xa6adU,0x39a3U,0xa0cfU,0xc13bU,0x5fb2U,0x2804U,0x8340U,0xbe2dU,0x95d3U,
		0x5453U,0xe68eU,0xfaaeU,0x391aU,0xf761U,0x4544U,0xfb44U,0x2190U,0x3cb3U,0xc707U,
		0xb4e9U,0x2d9aU,0xe780U,0x09c7U,0x1a87U,0xd2e8U,0x1650U,0xc50bU,0x17beU,0x5fbcU,
		0x1dfdU,0xc297U,0x15d6U,0x2bd5U,0x685eU,0x1dc3U,0xefe6U,0xe38bU,0xda4cU,0xb32dU,
		0x2ccdU,0xc73aU,0xe6efU,0x8be3U,0x71d0U,0x316cU,0x5da7U,0xcf47U,0x65caU,0x36cfU,
		0x5a6eU,0x5077U,0x4a2fU,0x996bU,0xa1ffU,0x8abfU,0xb843U,0x715cU,0xb07dU,0x04a9U,
		0xce4fU,0x29c5U,0x54b7U,0xc6ddU,0x74acU,0xe20dU,0x331eU,0x05a5U,0xd5faU,0x0050U,
		0x7a83U,0x9722U,0xb318U,0x28b6U,0x1c86U,0x8b2cU,0x28c0U,0xaf66U,0x7b22U,0xd9ecU,
		0x2b8dU,0x1d9aU,0x3b20U,0x40edU,0x65b1U,0x4bb7U,0x86f7U,0xfec7U,0x3e15U,0x0a82U,
		0x11c2U,0x7e5aU,0xb0d6U,0x6311U,0x49cdU,0x3fa2U,0xbbd8U,0x67d4U,0xdc11U,0x295cU,
		0xaeaeU,0xf5f5U,0xa121U,0x1558U,0xc60bU,0x7240U,0x6464U,0xf7f7U,0xe7d4U,0x6c57U,
		0x7c67U,0x9ff6U,0xb717U,0xd1a0U,0x6a60U,0x675eU,0xe8adU,0x39d4U,0x0c29U,0x76c7U,
		0x6fbeU,0x6e42U,0xa6ceU,0x4e7fU,0x7ef4U,0x447aU,0xbaf6U,0x3afeU,0xd3c0U,0x36c0U,
		0x483cU,0xfc3cU,0x5efbU,0xd7c3U,0x5424U,0x705cU,0xd7f0U,0xabc6U,0xeecfU,0xb433U,
		0xb144U,0x40e3U,0xeee2U,0xb952U,0x0a73U,0x5aacU,0xe293U,0xc7ddU,0x012cU,0x8e1bU,
		0xf296U,0x4183U,0x896fU,0x7e2dU,0x23adU,0x3bd0U,0xa0d8U,0x9480U,0x3b93U,0x6c7eU,
		0xacacU,0x4a4aU,0x1866U,0xa6f3U,0xa92eU,0xc07aU,0xa4a4U,0xcacaU,0x8a34U,0x3203U,
		0xc65cU,0x672cU,0x056eU,0x4824U,0x3d9aU,0x4c38U,0x90abU,0xb43cU,0x5830U,0x1ef9U,
		0x3c07U,0x4e38U,0x97e3U,0x2e0cU,0xf320U,0xf72aU,0x09d9U,0x8670U,0x06d3U,0x0206U,
		0x1843U,0x534dU,0xcfe7U,0x3fb1U,0xcd14U,0x54efU,0x8b82U,0xa19bU,0xf2b1U,0x530dU,
		0xaccaU,0x33d1U,0xa5ebU,0x5843U,0x69bcU,0x1b1fU,0x1286U,0x102aU,0xf2ddU,0xa1f5U,
		0x21e5U,0x337eU,0x41e0U,0x64b9U,0xbf1fU,0x4d38U,0x6ae4U,0x159aU,0x7352U,0xe88fU,
		0x351bU,0x1af9U,0x7064U,0x610fU,0x85bfU,0x20b1U,0x4897U,0x97b0U,0xb19cU,0x7319U,
		0xbf85U,0xb120U,0xac55U,0xde01U,0xbaa4U,0xeb01U,0x51a7U,0x4051U,0x13e8U,0xbed1U,
		0x7a3eU,0xed3aU,0x1e12U,0xd10cU,0x901cU,0x82f1U,0x9553U,0x3455U,0x7a9aU,0xf219U,
		0x44caU,0x53bfU,0x1432U,0x7589U,0xf4cbU,0x58beU,0xe279U,0x85aeU,0xc69bU,0xf1b1U,
		0x5943U,0x0718U,0x73d8U,0xb739U,0xb600U,0x7c76U,0xdfa0U,0x9993U,0x4cd1U,0xf54eU,
		0xef7bU,0x3e50U,0x59f0U,0xc36fU,0xe92fU,0xa49bU,0xf7d8U,0xe203U,0xebafU,0x5584U,
		0xfa6cU,0xf427U,0x1b34U,0x50b0U,0x7ab1U,0x5b0bU,0x3a6aU,0x121dU,0x1799U,0xba79U,
		0x3acfU,0x3e0cU,0xef20U,0x6df9U,0xc3e3U,0x550aU,0x8f68U,0x4c90U,0x3927U,0xfbb9U,
		0x1310U,0x527dU,0xf3afU,0x5cf7U,0xcee6U,0x8693U,0x3ecaU,0x3be0U,0x9b21U,0x0fd6U,
		0xec39U,0x9cdbU,0x9291U,0x2d0eU,0x3057U,0xa808U,0x0906U,0xb1f8U,0xf3e7U,0xd89fU,
		0x28d1U,0x2666U,0x889fU,0xd21cU,0xbca4U,0x2928U,0x2ceaU,0xb104U,0x9de0U,0xd92fU,
		0x8ddeU,0xc42dU,0xde9fU,0xe87eU,0xc5d2U,0xf261U,0x8765U,0x7882U,0x786cU,0x0961U,
		0x4665U,0x70ddU,0x6c6bU,0x7a54U,0x70cfU,0x816aU,0x248bU,0xd311U,0x677bU,0x812bU,
		0x2080U,0x5cceU,0xd555U,0x153dU,0x6ba5U,0x3736U,0xc541U,0x619dU,0xe41aU,0xf24cU,
		0xa8e6U,0x4c82U,0xd8ebU,0xed12U,0xa1dcU,0x920aU,0x72e3U,0xca60U,0xb9d4U,0xf781U,
		0x9e7cU,0xdd86U,0xd2bbU,0xd015U,0x0d4bU,0xe133U,0x56a1U,0x0750U,0x6006U,0x440eU,
		0x8719U,0x0d59U,0x1e40U,0x68f9U,0x7d1dU,0xa75cU,0xf9e7U,0x4c0cU,0x48ccU,0x9bd3U,
		0xee67U,0x152fU,0xcc82U,0x313bU,0x9a61U,0x2881U,0x228fU,0x87d5U,0x4379U,0x0612U,
		0x5f8dU,0x4aa7U,0xff2dU,0x25faU,0xe25dU,0x40a7U,0x1716U,0x2e04U,0x2c66U,0x2406U,
		0x51fbU,0x3f99U,0xf8aeU,0xeab3U,0xf7f0U,0xeafaU,0xd494U,0xc5beU,0x1b0fU,0x7170U,
		0xfa22U,0x7abfU,0xc860U,0x49a7U,0xd61eU,0xa830U,0xe30cU,0xa9b9U,0xed0bU,0xd39eU,
		0xa195U,0x0712U,0x004aU,0xd7bdU,0x50dbU,0x5facU,0x9c3aU,0xcc4bU,0x6dc5U,0xfa2aU,
		0x1972U,0x528bU,0x01e4U,0x968fU,0x2922U,0x9bcdU,0x11afU,0x16f8U,0x15b7U,0xabe8U,
		0x6141U,0xc66cU,0x7decU,0x01eaU,0x16b3U,0xbe28U,0x1c59U,0xc9e4U,0x4255U,0x23d3U,
		0xee12U,0xf7a4U,0xac50U,0xb5fbU,0x8a2fU,0xacc5U,0xc606U,0xd7ebU,0x2cc8U,0xb313U,
		0xd52aU,0x1b46U,0xd619U,0x1c54U,0xa8e1U,0xcc7dU,0x8f63U,0x9ddbU,0x0134U,0xa708U,
		0xb53bU,0xf998U,0x224aU,0x2db2U,0x1a5cU,0x1ad5U,0x6077U,0x7d49U,0xd52fU,0x8803U,
		0x3309U,0xd3a5U,0x71b9U,0x3065U,0x2ed1U,0xbd8eU,0x9ecaU,0x8940U,0x1883U,0xf2cbU,
		0x367dU,0x13faU,0x5814U,0xe055U,0xcecdU,0x5e9eU,0x55b2U,0x1699U,0xf9c9U,0xa538U,
		0x3f0cU,0xab8bU,0xa7f7U,0x1fbaU,0x6185U,0xd466U,0x3440U,0x9695U,0x8734U,0x27e2U,
		0x1b4bU,0x2b03U,0xa771U,0xce0bU,0x9cf1U,0x92e4U,0xf97dU,0xdf2aU,0x2763U,0xc43bU,
		0x175cU,0x7192U,0x4dc8U,0xcdf7U,0x4190U,0x83feU,0xe354U,0x8214U,0x8e9fU,0xe967U,
		0x6ee2U,0x3f1aU,0x6499U,0x314eU,0x977eU,0xc0ffU,0xcd34U,0x40d1U,0x26f8U,0x9cb6U,
		0x65fdU,0x21aeU,0x1a61U,0x89deU,0xb632U,0xf5d4U,0xa4b1U,0x20c5U,0x5f2fU,0x30e6U,
		0x76e6U,0xea60U,0x9444U,0x645bU,0x8359U,0xc957U,0xc302U,0x2eb0U,0x6230U,0xf713U,
		0x448cU,0x30feU,0x9fa2U,0xa528U,0x3fd8U,0x2138U,0xa687U,0xdf38U,0xf737U,0x1e63U,
		0xa5a8U,0x1978U,0xc27aU,0x8c05U,0x1adaU,0x641bU,0x3616U,0xa1aeU,0x91faU,0xb642U,
		0x085bU,0xf66dU,0xacd7U,0x913eU,0x8a9bU,0xdd39U,0x6951U,0xe07cU,0x4969U,0x5d46U,
		0x17a2U,0xe105U,0xfcf0U,0xc44cU,0xc351U,0xfb98U,0x18d7U,0x2c6fU,0x29d6U,0xd2a8U,
		0x3007U,0x0caeU,0xbab1U,0x3891U,0x6559U,0xeb89U,0x224fU,0x3e65U,0x76efU,0xe313U,
		0x783aU,0x9709U,0xe89cU,0x2a36U,0xf893U,0xdbd8U,0x220bU,0x00afU,0xd81cU,0xb549U,
		0xf3c7U,0xfe53U,0x4cdeU,0x07bdU,0xd742U,0x3cf1U,0x8a59U,0x1fdcU,0x5d0eU,0x5a2dU,
		0xb6eaU,0x6360U,0x86a9U,0x58f9U,0x0febU,0x5291U,0x1137U,0xf0d0U,0x2db7U,0x894eU,
		0x7596U,0x7d8cU,0x4773U,0x29aaU,0xb801U,0xa043U,0x268eU,0xb1c0U,0xf09dU,0x0216U,
		0x70a8U,0x8dfeU,0x5780U,0x0591U,0xf8dfU,0x7492U,0x592bU,0x2cb7U,0xfc0fU,0xe977U,
		0x8d1fU,0x7137U,0x2b9aU,0xc22eU,0x46f3U,0x8c9cU,0x0456U,0xc24fU,0x59f9U,0xcfdfU,
		0xb540U,0x67b8U,0x73f1U,0x39cbU,0xf013U,0x04e7U,0x82dbU,0x5e2dU,0x7baeU,0xd7f9U,
		0xe21aU,0xd65fU,0x15afU,0x2eebU,0x7b5cU,0x7ac7U,0x0764U,0xa81dU,0x8c58U,0x721dU,
		0x5db4U,0x15eeU,0xa8beU,0xffaeU,0x2ef3U,0x4d05U,0x1679U,0x8d00U,0x53a2U,0x6dcbU,
		0xc2a4U,0xe00cU,0xf99dU,0xec8eU,0x4397U,0x8b56U,0x62b2U,0x614fU,0x37fdU,0x80b7U,
		0x572aU,0x5ddcU,0x5886U,0x6d64U,0xc12cU,0x5e46U,0x02c0U,0x0c49U,0x6bd8U,0xcc59U,
		0xd7d0U,0x7022U,0x9f4aU,0xb8b1U,0x8e23U,0x7ee5U,0x609bU,0xae2bU,0x982fU,0x69ccU,
		0xa2aeU,0x8270U,0x8830U,0x3255U,0x0139U,0x0bbaU,0x3abdU,0xa9c0U,0x95abU,0x4563U,
		0x65a2U,0xc5a4U,0x64cdU,0x9569U,0x5408U,0xf637U,0x7402U,0x24ffU,0x7f73U,0x31e6U,
		0xe3fbU,0xc847U,0x8c3eU,0xa86aU,0x2a01U,0x710aU,0x2969U,0x5a53U,0x2f28U,0x5198U,
		0xf620U,0x1781U,0xaaa1U,0xfa30U,0xc56aU,0x8114U,0xa70dU,0xbe87U,0xb5a9U,0x953fU,
		0xb3b6U,0x6372U,0x6324U,0x4f5eU,0xa91fU,0x2b64U,0xce71U,0xcb2cU,0x6a6dU,0x4fe5U,
		0x245cU,0xfbcbU,0xf7c4U,0x793cU,0xfd0eU,0xbef3U,0xf78dU,0xde4fU,0x9891U,0xa737U,
		0xd400U,0x4b44U,0xf578U,0xd561U,0xef8eU,0xbd0aU,0x061eU,0x8eb7U,0x71beU,0x0f4dU,
		0x0548U,0x22abU,0x147cU,0x3624U,0x6621U,0xef43U,0x2f10U,0x1e4cU,0x03ceU,0x3795U,
		0x3eafU,0x6121U,0x19aeU,0xef2cU,0x6c94U,0xe4eeU,0x3ea0U,0x312aU,0x231cU,0x07cdU,
		0x5bdbU,0x6a16U,0x0fe2U,0x7501U,0x01b1U,0x5ef2U,0xf27dU,0x7df2U,0x4ee9U,0xa22bU,
		0x306aU,0x3363U,0x89cdU,0xcd89U,0x52e0U,0x88abU,0x643aU,0x7a20U,0x162cU,0x310cU,
		0x096dU,0x14a0U,0x1b7aU,0x3eefU,0xfaa1U,0x7be7U,0xe509U,0x5820U,0x1329U,0x4efdU,
		0xcd98U,0x282cU,0x27eeU,0x60c4U,0x1a25U,0xccc4U,0x35a4U,0x3bb1U,0x926eU,0xa4bfU,
		0x522aU,0xfd61U,0x330eU,0xb21aU,0x50c8U,0xd625U,0x79d5U,0xe2e6U,0x9be9U,0x25c3U,
		0x92f9U,0xb8c5U,0x2d25U,0x4a6bU,0x880fU,0xdf63U,0x8af2U,0x3ca3U,0xb66cU,0x3f28U,
		0x23dfU,0x0b06U,0x5c8dU,0xb342U,0xab41U,0x5752U,0xcb4fU,0x3842U,0xa887U,0xbf95U,
		0xea10U,0x01f0U,0x9c1dU,0x0070U,0xd3aaU,0x8fd9U,0xf001U,0x10eaU,0x28a2U,0x1d7aU,
		0xb6c3U,0xf8baU,0x35c3U,0xe59dU,0x027dU,0x37daU,0x15ffU,0xea29U,0xb1ddU,0x0a43U,
		0x1f2fU,0x85d9U,0x208aU,0xaa73U,0xc805U,0x8aa7U,0x133dU,0xaaccU,0x93b6U,0xfe61U,
		0x1e6fU,0xca94U,0x48bdU,0x3ad9U,0xf1d2U,0x8365U,0xd534U,0x3184U,0x9f11U,0xcc98U,
		0xeca2U,0x2dc7U,0xa459U,0x5205U,0x9cc2U,0xab16U,0xcba7U,0x8f49U,0x3949U,0xde90U,
		0x5a39U,0x1b17U,0x3ce8U,0x944eU,0x38daU,0x4e76U,0xd0ebU,0x1145U,0xe8a6U,0x9015U,
		0xe222U,0xb80dU,0x366eU,0xa0bdU,0xf79cU,0x6147U,0x0db8U,0x22e2U,0x8808U,0xde60U,
		0xd815U,0x73fbU,0xd2b4U,0x3676U,0xfac8U,0x5652U,0xe6bfU,0x4cb1U,0xf239U,0xf89fU,
		0x883aU,0x2790U,0xda2bU,0x38e4U,0xe983U,0xda5fU,0xa27cU,0x57f0U,0xd6d3U,0x4460U,
		0xe29bU,0xf01fU,0x26e9U,0x4bfaU,0x7034U,0x6521U,0x20deU,0x4ac7U,0x0c85U,0x625dU,
		0x607cU,0xe113U,0x7229U,0xa837U,0xb477U,0x01bbU,0x1ff7U,0x9abaU,0xa0abU,0x782bU,
		0xab4bU,0x4d48U,0xde67U,0x43a3U,0x26dfU,0x87abU,0x9b50U,0x4f79U,0x469dU,0x0145U,
		0xa2dbU,0xa43fU,0xc11dU,0x95c2U,0xe700U,0x00e7U,0xa166U,0xbdb5U,0xcc65U,0xc672U,
		0x6d13U,0x136dU,0x7d4eU,0xc576U,0xea92U,0xac07U,0x3633U,0x06a3U,0x31f2U,0x29f2U,
		0x322aU,0x6b2cU,0x81bbU,0x4b8dU,0x5ec2U,0x225dU,0x5c78U,0x66d1U,0xf49eU,0xc25dU,
		0xcca4U,0x13dbU,0xaafeU,0x39adU,0x1fa6U,0xa4f4U,0x8ae0U,0x1911U,0xdbecU,0xb610U,
		0x9d45U,0x2096U,0x2587U,0xe71eU,0x3682U,0xc0a1U,0xb18fU,0xb2ceU,0x5bfeU,0xf44bU,
		0xe1b4U,0xd0f7U,0x10c5U,0x6db3U,0x524aU,0x63ebU,0x62c3U,0x89b2U,0x1290U,0x141bU,
		0x7d02U,0xda37U,0xfdadU,0x6031U,0x007fU,0xdf93U,0x6269U,0xe3bfU,0x55f9U,0xe939U,
		0xa3cfU,0x55d1U,0x3a87U,0x15bdU,0x39e9U,0xf955U,0x13ffU,0x473eU,0xc2d3U,0x9b5cU,
		0x7a8dU,0x26a0U,0x8d5cU,0x42b3U,0xb7f8U,0x6d00U,0x6b22U,0xeed9U,0xc582U,0x2186U,
		0xc15fU,0xf289U,0xba82U,0xfc9dU,0x554eU,0xd154U,0x07f3U,0x5d56U,0xa025U,0x240fU,
		0x3efbU,0x2b44U,0xc334U,0x0251U,0xf940U,0xbdefU,0x0561U,0x2d4cU,0x1888U,0x63c2U,
		0xea69U,0xc867U,0x0b18U,0xf584U,0x8148U,0x0fa5U,0x9f79U,0x8d0eU,0x37a1U,0x1aacU,
		0x0dbfU,0x67c1U,0x0f06U,0x2c15U,0x1c23U,0xcd07U,0x19e5U,0x9457U,0xe5a9U,0xdfc6U,
		0x5bcaU,0x0d06U,0x00faU,0xfa00U,0xae18U,0xcb5bU,0x750aU,0xabeeU,0x25cfU,0xcf25U,
		0xf2d0U,0xae42U,0xe827U,0x8b98U,0x9ae3U,0x5894U,0x6d09U,0xa014U,0xf516U,0x27faU,
		0x08f2U,0x10a9U,0x8b01U,0x6fedU,0x43bfU,0xbcefU,0xd254U,0x0984U,0x6ecfU,0xbd9fU,
		0xa6b6U,0x91daU,0x7635U,0x7f9fU,0xc51cU,0xc4c1U,0xc914U,0xbae9U,0xae08U,0x9111U,
		0xc378U,0x73bdU,0x9ecfU,0x766dU,0x5872U,0x71eeU,0xb229U,0x1decU,0x1998U,0x3718U,
		0x4858U,0xabcfU,0x011fU,0x53ebU,0xeda1U,0x64dbU,0x69f5U,0x7178U,0x795bU,0xa95dU,
		0xdbc6U,0xa305U,0x509bU,0x794fU,0xb5bcU,0xe18aU,0xd667U,0xc3aeU,0xf568U,0x0e06U,
		0xb511U,0x11b5U,0x34f8U,0x21ddU,0x8945U,0xae39U,0xbdabU,0xabbdU,0xe17eU,0xa0c5U,
		0x33e7U,0x95e7U,0x7971U,0xf1eeU,0xf231U,0xf229U,0xc5feU,0x587eU,0x241bU,0x198cU,
		0xed28U,0x018aU,0xe749U,0x2032U,0x467cU,0x6332U,0xc7e5U,0x7a76U,0x1669U,0x8154U,
		0xa391U,0xefaaU,0xe518U,0xb27fU,0x7661U,0xcbc7U,0x2ed6U,0x148dU,0x3657U,0x49aeU,
		0xfa0cU,0x339aU,0xce42U,0xe61eU,0xb2e7U,0x9434U,0x64ebU,0xf522U,0x22e9U,0x200bU,
		0xf0bcU,0x5912U,0x3d44U,0xadfcU,0xb156U,0x7807U,0xc8faU,0x5256U,0x223aU,0x94dfU,
		0xd98eU,0x5a61U,0xb61cU,0x83b9U,0x82c4U,0x16feU,0xa2d1U,0xc58eU,0xbd93U,0xfa55U,
		0xfe16U,0xc482U,0x7c03U,0x6569U,0x4b13U,0xff1dU,0x15c9U,0x6a46U,0xe83cU,0x4e94U,
		0x843dU,0xc191U,0x4362U,0xc147U,0x8c28U,0x4228U,0x327eU,0xcdd2U,0xa672U,0x8cacU,
		0x89c3U,0x8004U,0x2da6U,0xece1U,0x5524U,0xd6b1U,0x0c9aU,0x1164U,0x4c65U,0x088bU,
		0xf9ceU,0x91b4U,0x63a4U,0x9e62U,0x79e7U,0x8ff2U,0x0b9eU,0x60e1U,0x1d56U,0x8f57U,
		0x9a49U,0xe3e4U,0xd175U,0x2122U,0x6c31U,0xd071U,0x47e2U,0x6b60U,0x471bU,0xc9ceU,
		0xdb78U,0xb460U,0x8934U,0xe0e1U,0x9407U,0xe9eaU,0xba92U,0x5a0dU,0x42f3U,0xee18U,
		0xeac9U,0xe45bU,0xa53dU,0xb356U,0x9221U,0x0236U,0xae8aU,0x1e7dU,0xe247U,0x606bU,
		0xc31dU,0x5e68U,0x70f5U,0xe4dbU,0x4b22U,0x9e17U,0xcc00U,0xc881U,0x5ae8U,0x279eU,
		0xd4a7U,0xb4c9U,0x8bd4U,0xad93U,0x25e5U,0x4149U,0x76ddU,0x10bcU,0x3ee5U,0x65c1U,
		0xe049U,0xa481U,0x83d7U,0x91c4U,0x55b7U,0x031bU,0x98dbU,0x9c41U,0x24b8U,0x294dU,
		0x739bU,0x3d87U,0x05d6U,0x5860U,0xf119U,0xc98cU,0x8841U,0xa7ebU,0x5c29U,0x11dcU,
		0x9a9aU,0xececU,0x8661U,0x6f3aU,0x94eeU,0xba22U,0xe0ecU,0x0380U,0x5f1cU,0x377fU,
		0x92d8U,0x56d5U,0xafa6U,0xf7aaU,0x4797U,0xd63cU,0x169eU,0xdae4U,0xb7d1U,0xa48cU,
		0x0c6fU,0x2c81U,0xe7e7U,0x6969U,0x576cU,0xd4e7U,0x20feU,0x1750U,0x56e0U,0x8244U,
		0x6c8aU,0x077bU,0xa9adU,0xa476U,0xbb28U,0xcb71U,0x7201U,0xe2a9U,0xeeb1U,0x56b4U,
		0xfe43U,0x3da6U,0x9d12U,0x6e5bU,0xb2f6U,0x7751U,0x3f5cU,0x8f18U,0xced2U,0x9c2eU,
		0x966dU,0xa257U,0x3fbdU,0xfdd0U,0xd4faU,0xc047U,0x6ebaU,0x3245U,0x63d2U,0xf2e9U,
		0x32a5U,0xc79bU,0x23faU,0xd49cU,0x950bU,0x7dbcU,0x7e6cU,0x933bU,0x9797U,0x1e1eU,
		0x5574U,0xe8f3U,0x7bb6U,0x9ac8U,0x97bcU,0x406cU,0x0528U,0x425bU,0xc01fU,0xf7e4U,
		0x5ea3U,0x09a8U,0x08c7U,0x41f3U,0x14b8U,0xe22aU,0x32bbU,0xbe18U,0x0724U,0xbc60U,
		0x7979U,0xe1e1U,0x0332U,0x348aU,0x5324U,0xa6ebU,0x21a5U,0xe715U,0x6a0cU,0xb324U,
		0xb5efU,0x3d7cU,0x0a7bU,0x1593U,0x6e53U,0x67fcU,0xca8cU,0xbae1U,0x7168U,0x8c94U,
		0xfc96U,0x0183U,0x5819U,0xda58U,0xe0feU,0xcc52U,0x8163U,0x1c4dU,0x09d3U,0x44c4U,
		0xb077U,0xa6c8U,0xfeecU,0x3b43U,0x2797U,0xe70eU,0x8c52U,0x8c38U,0x4486U,0xd73cU,
		0x01deU,0x55acU,0x3051U,0x6000U,0xdbc0U,0xb150U,0xfda7U,0x19dfU,0x38d4U,0x4697U,
		0x3e9aU,0x23d9U,0x854fU,0xa082U,0xe05aU,0x6b3aU,0x7e66U,0x01d8U,0x4715U,0x865bU,
		0xe02cU,0xbfeaU,0xed7dU,0x05bdU,0x5130U,0x0060U,0x0f61U,0x6470U,0x13c3U,0x5b87U,
		0x6691U,0x1548U,0xbc4bU,0xe814U,0x0e15U,0xd1bfU,0x6904U,0x6ce5U,0xec05U,0x2283U,
		0x6e21U,0xf46eU,0xa3fbU,0x232aU,0xca0eU,0x532dU,0x3445U,0xf9b0U,0x18bdU,0x6e04U,
		0x8786U,0xb889U,0x6617U,0xbc7cU,0x25eeU,0x5e5dU,0x4f10U,0xb0ecU,0xd227U,0xf59aU,
		0x42c5U,0x036fU,0x7e82U,0x88b9U,0xe060U,0x568fU,0x8080U,0xc6c6U,0x5e80U,0x0d3fU,
		0x3344U,0x4878U,0xc211U,0x5a7eU,0x7f37U,0x1c5fU,0x6cd4U,0xe46bU,0x969aU,0x9a16U,
		0x3cd6U,0x9747U,0xade8U,0xd439U,0xa750U,0xeabeU,0x6b81U,0xe9acU,0x7272U,0x7373U,
		0x0c5aU,0x17eeU,0x34c8U,0x0918U,0x3a73U,0xfef3U,0x263cU,0x8188U,0xdd44U,0xe3aeU,
		0xfb06U,0x171fU,0x6abbU,0x59b5U,0x0f16U,0xc962U,0xb96cU,0xb9c8U,0xa04eU,0x411dU,
		0x618aU,0x7a06U,0x5f38U,0x50e2U,0x60b3U,0xfd1cU,0x6e18U,0x395eU,0xcd2bU,0xdbe3U,
		0x0f43U,0x6855U,0xf240U,0x5a02U,0x5107U,0x04f2U,0xd8bfU,0xcb0fU,0x6b40U,0x5a42U,
		0x1f81U,0x6ddbU,0x9d56U,0x7eccU,0x8b61U,0x8e83U,0x2f4aU,0x6b99U,0x5be4U,0xc9eaU,
		0x035aU,0x12d2U,0xd46bU,0x3501U,0xeae9U,0x0794U,0x5016U,0x0bc5U,0x8d65U,0x4645U,
		0xcd48U,0x22b9U,0x811fU,0xdb6dU,0x46c0U,0x4432U,0xfb2cU,0x6c41U,0xaa7eU,0xbc98U,
		0x7fc0U,0x6f8aU,0x9ea5U,0x2d7fU,0xae60U,0x1150U,0x962eU,0x0f93U,0xb3c8U,0x6804U,
		0xf162U,0xbfc6U,0xef84U,0xb010U,0xa7fcU,0x8fe7U,0x38efU,0x0506U,0xf861U,0x5c5dU,
		0xe377U,0xf037U,0x5e1bU,0x542dU,0x9095U,0x1dd3U,0xf181U,0x4f84U,0x8e56U,0xc5c6U,
		0xcf60U,0x5051U,0x8555U,0x7dfdU,0xeccaU,0x62f8U,0x3e7aU,0x3aedU,0xd801U,0x667eU,
		0x9861U,0xce9bU,0xd56cU,0x16b9U,0xd923U,0x9a3eU,0x1fbfU,0x384dU,0x7473U,0x2ae3U,
		0x9120U,0xb815U,0x60cfU,0x5150U,0x2c53U,0xaff9U,0xf1f6U,0x2e54U,0xd03aU,0xb67aU,
		0x4523U,0x8bb5U,0x8062U,0x3b4aU,0xf844U,0x181fU,0xf697U,0xfa60U,0x5ab4U,0x1cbaU,
		0x6822U,0xfd6fU,0xcbd2U,0xf907U,0x5ad0U,0xff06U,0xd1b6U,0xa0d0U,0x3f11U,0x39a4U,
		0xbee9U,0x5fb3U,0x1211U,0xc13cU,0xb390U,0xb987U,0x71f9U,0x9e9dU,0xc7fdU,0xa6aeU,
		0xf8a6U,0xfc07U,0x0fb0U,0xe68fU,0xd1d1U,0x5454U,0xa088U,0x391bU,0x0b38U,0xfaafU,
		0x96f2U,0x8341U,0x5b42U,0x2805U,0x6ad0U,0x95d4U,0xedb0U,0xbe2eU,0xf341U,0xc708U,
		0xab90U,0x3cb4U,0x0fcaU,0x2d9bU,0x63f6U,0xb4eaU,0x1d1dU,0x4545U,0xfc7dU,0xf762U,
		0xcf35U,0x2191U,0xdc9cU,0xfb45U,0x1810U,0xc50cU,0x094aU,0x1651U,0x117dU,0x5fbdU,
		0x8a0fU,0x17bfU,0x5dd0U,0x09c8U,0xfed8U,0xe781U,0x8499U,0xd2e9U,0xb01aU,0x1a88U,
		0xb7e8U,0x1dc4U,0x9cf6U,0x685fU,0x697bU,0xe38cU,0xfcacU,0xefe7U,0xbc86U,0xc298U,
		0xe16bU,0x1dfeU,0xdb90U,0x2bd6U,0x4db0U,0x15d7U,0x84faU,0x8be4U,0x0785U,0xe6f0U,
		0x0f67U,0x316dU,0x960cU,0x71d1U,0xa87dU,0xb32eU,0x8cf4U,0xda4dU,0xc766U,0xc73bU,
		0xca2dU,0x2cceU,0x0e6fU,0x5078U,0x92e9U,0x5a6fU,0x800eU,0x996cU,0xe794U,0x4a30U,
		0xa4ddU,0xcf48U,0xf415U,0x5da8U,0x68d5U,0x36d0U,0x45ceU,0x65cbU,0x4c43U,0x04aaU,
		0x0cd9U,0xb07eU,0x932cU,0x29c6U,0x7798U,0xce50U,0x49f8U,0x8ac0U,0x1f52U,0xa200U,
		0x04dbU,0x715dU,0x1897U,0xb844U,0xe94dU,0x05a6U,0xa776U,0x331fU,0xa90fU,0x0051U,
		0xd117U,0xd5fbU,0x9fc4U,0xc6deU,0xc4e8U,0x54b8U,0xf45bU,0xe20eU,0xe6c6U,0x74adU,
		0xba5dU,0x8b2dU,0x457bU,0x1c87U,0x6476U,0xaf67U,0x9e53U,0x28c1U,0xaddfU,0x9723U,
		0x91f1U,0x7a84U,0xb338U,0x28b7U,0x0585U,0xb319U,0x4fcaU,0x40eeU,0xce10U,0x3b21U,
		0x93deU,0x4bb8U,0x003aU,0x65b2U,0xb3f7U,0xd9edU,0xf802U,0x7b23U,0x1cefU,0x1d9bU,
		0x0877U,0x2b8eU,0xc83bU,0x7e5bU,0x081cU,0x11c3U,0x5cd8U,0x6312U,0xb398U,0xb0d7U,
		0x9472U,0xfec8U,0xf6c3U,0x86f8U,0xed61U,0x0a83U,0xa414U,0x3e16U,0xf833U,0x295dU,
		0x1b50U,0xdc12U,0x8d4aU,0xf5f6U,0x0f7fU,0xaeafU,0xeb3fU,0x3fa3U,0xee7eU,0x49ceU,
		0x5df1U,0x67d5U,0xa71eU,0xbbd9U,0x35b9U,0xf7f8U,0x9533U,0x6465U,0x5136U,0x6c58U,
		0x2c24U,0xe7d5U,0x2a40U,0x1559U,0x82e3U,0xa122U,0xd765U,0x7241U,0x0d92U,0xc60cU,
		0x8ed8U,0x675fU,0x2b08U,0x6a61U,0x4510U,0x39d5U,0x99c1U,0xe8aeU,0x99d3U,0x9ff7U,
		0xeb4fU,0x7c68U,0x33efU,0xd1a1U,0x325fU,0xb718U,0xe49fU,0x4e80U,0x427fU,0xa6cfU,
		0x8264U,0x447bU,0x45f4U,0x7ef5U,0xeeeaU,0x76c8U,0x87b5U,0x0c2aU,0xaf58U,0x6e43U,
		0x98b8U,0x6fbfU,0xbebaU,0xfc3dU,0x8739U,0x483dU,0x247aU,0xd7c4U,0x283eU,0x5efcU,
		0xcf75U,0x3affU,0x4009U,0xbaf7U,0x575cU,0x36c1U,0x4cb8U,0xd3c1U,0x736aU,0xb434U,
		0xdd30U,0xeed0U,0x0066U,0x40e4U,0xc598U,0xb145U,0x49c6U,0x705dU,0x323dU,0x5425U,
		0xab0eU,0xabc7U,0xd7b5U,0xd7f1U,0xa01eU,0xc7deU,0x27e7U,0xe294U,0x18d0U,0x8e1cU,
		0x2079U,0x012dU,0x6a86U,0xb953U,0x49f1U,0xeee3U,0xfaffU,0x5aadU,0x25deU,0x0a74U,
		0x4c23U,0x3bd1U,0x90f9U,0x23aeU,0x2b19U,0x9481U,0x8285U,0xa0d9U,0x92a9U,0x4184U,
		0x3de0U,0xf297U,0x3967U,0x7e2eU,0x2e31U,0x8970U,0xd302U,0xa6f4U,0x4dcdU,0x1867U,
		0x2fe5U,0xc07bU,0x0eecU,0xa92fU,0xc0f3U,0x6c7fU,0x4152U,0x3b94U,0x32eaU,0x4a4bU,
		0xa7c4U,0xacadU,0x1034U,0x672dU,0x8798U,0xc65dU,0xed83U,0x4825U,0x1574U,0x056fU,
		0x84caU,0xcacbU,0x4d69U,0xa4a5U,0x689fU,0x3204U,0x5c0dU,0x8a35U,0x369cU,0x1efaU,
		0xcdfcU,0x5831U,0xdcf9U,0x4e39U,0x972cU,0x3c08U,0xfa97U,0x4c39U,0x644fU,0x3d9bU,
		0x9b6cU,0xb43dU,0xdb3fU,0x90acU,0xdcd4U,0x8671U,0x1226U,0x09daU,0xc470U,0x0207U,
		0xb704U,0x06d4U,0x6a8dU,0x2e0dU,0x0299U,0x97e4U,0x711cU,0xf72bU,0x5c47U,0xf321U,
		0x7335U,0x54f0U,0x1761U,0xcd15U,0x30efU,0xa19cU,0xd878U,0x8b83U,0x1bf7U,0x534eU,
		0x2a99U,0x1844U,0xa51cU,0x3fb2U,0xfacfU,0xcfe8U,0x2a63U,0x5844U,0xd84dU,0xa5ecU,
		0x05c3U,0x1b20U,0x4ba6U,0x69bdU,0xfea5U,0x530eU,0xa132U,0xf2b2U,0x6937U,0x33d2U,
		0x6a1fU,0xaccbU,0x560cU,0x337fU,0x7197U,0x21e6U,0xd69bU,0x64baU,0x39bcU,0x41e1U,
		0x42bcU,0x102bU,0xc2f9U,0x1287U,0xccbaU,0xa1f6U,0x8bf9U,0xf2deU,0x51d3U,0xe890U,
		0x19c6U,0x7353U,0x4e04U,0x1afaU,0xb98fU,0x351cU,0x6fdcU,0x4d39U,0x0397U,0xbf20U,
		0xfdbdU,0x159bU,0x58d1U,0x6ae5U,0x3e59U,0x97b1U,0x2768U,0x4898U,0x0be6U,0x731aU,
		0x905bU,0xb19dU,0xaa64U,0x6110U,0x3980U,0x7065U,0x0af0U,0x20b2U,0xc2e2U,0x85c0U,
		0x302cU,0xeb02U,0xee36U,0xbaa5U,0xe032U,0x4052U,0xc17bU,0x51a8U,0xa3e0U,0xb121U,
		0x5e35U,0xbf86U,0xf4dfU,0xde02U,0x4811U,0xac56U,0x79cbU,0xd10dU,0xc440U,0x1e13U,
		0x8519U,0x82f2U,0x76a0U,0x901dU,0x100cU,0xbed2U,0xcd7aU,0x13e9U,0x95efU,0xed3bU,
		0x3d22U,0x7a3fU,0x990aU,0x53c0U,0x498bU,0x44cbU,0xaa34U,0x758aU,0x6040U,0x1433U,
		0xb95aU,0x3456U,0x019eU,0x9554U,0x0a63U,0xf21aU,0x549aU,0x7a9bU,0x2305U,0xf1b2U,
		0xf982U,0xc69cU,0x8df2U,0x0719U,0x2528U,0x5944U,0xb6a1U,0x58bfU,0xeb5fU,0xf4ccU,
		0x77f8U,0x85afU,0x5659U,0xe27aU,0x627bU,0x9994U,0x77deU,0xdfa1U,0xbff0U,0xf54fU,
		0x271fU,0x4cd2U,0xeb73U,0xb73aU,0x1a4aU,0x73d9U,0xe855U,0x7c77U,0x9e06U,0xb601U,
		0x486fU,0xa49cU,0xdf2fU,0xe930U,0x4a61U,0xe204U,0x5a21U,0xf7d9U,0xf82bU,0x3e51U,
		0x3f54U,0xef7cU,0xf1caU,0xc370U,0xc858U,0x59f1U
	};
	return f16LogTable[a];
}

/*
	Calculates the inverse of the logarithm with respect to the generator 2  

	a : an integer

	return : the field element 2^a
*/
f16FELT f16antilog(uint32_t a)
{
	static const f16FELT f16AntilogTable[65535*2] =
	{
		0x0001U,0x0002U,0x0004U,0x0008U,0x0010U,0x0020U,0x0040U,0x0080U,0x0100U,0x0200U,
		0x0400U,0x0800U,0x1000U,0x2000U,0x4000U,0x8000U,0x100bU,0x2016U,0x402cU,0x8058U,
		0x10bbU,0x2176U,0x42ecU,0x85d8U,0x1bbbU,0x3776U,0x6eecU,0xddd8U,0xabbbU,0x477dU,
		0x8efaU,0x0dffU,0x1bfeU,0x37fcU,0x6ff8U,0xdff0U,0xafebU,0x4fddU,0x9fbaU,0x2f7fU,
		0x5efeU,0xbdfcU,0x6bf3U,0xd7e6U,0xbfc7U,0x6f85U,0xdf0aU,0xae1fU,0x4c35U,0x986aU,
		0x20dfU,0x41beU,0x837cU,0x16f3U,0x2de6U,0x5bccU,0xb798U,0x7f3bU,0xfe76U,0xece7U,
		0xc9c5U,0x8381U,0x1709U,0x2e12U,0x5c24U,0xb848U,0x609bU,0xc136U,0x9267U,0x34c5U,
		0x698aU,0xd314U,0xb623U,0x7c4dU,0xf89aU,0xe13fU,0xd275U,0xb4e1U,0x79c9U,0xf392U,
		0xf72fU,0xfe55U,0xeca1U,0xc949U,0x8299U,0x1539U,0x2a72U,0x54e4U,0xa9c8U,0x439bU,
		0x8736U,0x1e67U,0x3cceU,0x799cU,0xf338U,0xf67bU,0xfcfdU,0xe9f1U,0xc3e9U,0x97d9U,
		0x3fb9U,0x7f72U,0xfee4U,0xedc3U,0xcb8dU,0x8711U,0x1e29U,0x3c52U,0x78a4U,0xf148U,
		0xf29bU,0xf53dU,0xfa71U,0xe4e9U,0xd9d9U,0xa3b9U,0x5779U,0xaef2U,0x4defU,0x9bdeU,
		0x27b7U,0x4f6eU,0x9edcU,0x2db3U,0x5b66U,0xb6ccU,0x7d93U,0xfb26U,0xe647U,0xdc85U,
		0xa901U,0x4209U,0x8412U,0x182fU,0x305eU,0x60bcU,0xc178U,0x92fbU,0x35fdU,0x6bfaU,
		0xd7f4U,0xbfe3U,0x6fcdU,0xdf9aU,0xaf3fU,0x4e75U,0x9ceaU,0x29dfU,0x53beU,0xa77cU,
		0x5ef3U,0xbde6U,0x6bc7U,0xd78eU,0xbf17U,0x6e25U,0xdc4aU,0xa89fU,0x4135U,0x826aU,
		0x14dfU,0x29beU,0x537cU,0xa6f8U,0x5dfbU,0xbbf6U,0x67e7U,0xcfceU,0x8f97U,0x0f25U,
		0x1e4aU,0x3c94U,0x7928U,0xf250U,0xf4abU,0xf95dU,0xe2b1U,0xd569U,0xbad9U,0x65b9U,
		0xcb72U,0x86efU,0x1dd5U,0x3baaU,0x7754U,0xeea8U,0xcd5bU,0x8abdU,0x0571U,0x0ae2U,
		0x15c4U,0x2b88U,0x5710U,0xae20U,0x4c4bU,0x9896U,0x2127U,0x424eU,0x849cU,0x1933U,
		0x3266U,0x64ccU,0xc998U,0x833bU,0x167dU,0x2cfaU,0x59f4U,0xb3e8U,0x77dbU,0xefb6U,
		0xcf67U,0x8ec5U,0x0d81U,0x1b02U,0x3604U,0x6c08U,0xd810U,0xa02bU,0x505dU,0xa0baU,
		0x517fU,0xa2feU,0x55f7U,0xabeeU,0x47d7U,0x8faeU,0x0f57U,0x1eaeU,0x3d5cU,0x7ab8U,
		0xf570U,0xfaebU,0xe5ddU,0xdbb1U,0xa769U,0x5ed9U,0xbdb2U,0x6b6fU,0xd6deU,0xbdb7U,
		0x6b65U,0xd6caU,0xbd9fU,0x6b35U,0xd66aU,0xbcdfU,0x69b5U,0xd36aU,0xb6dfU,0x7db5U,
		0xfb6aU,0xe6dfU,0xddb5U,0xab61U,0x46c9U,0x8d92U,0x0b2fU,0x165eU,0x2cbcU,0x5978U,
		0xb2f0U,0x75ebU,0xebd6U,0xc7a7U,0x9f45U,0x2e81U,0x5d02U,0xba04U,0x6403U,0xc806U,
		0x8007U,0x1005U,0x200aU,0x4014U,0x8028U,0x105bU,0x20b6U,0x416cU,0x82d8U,0x15bbU,
		0x2b76U,0x56ecU,0xadd8U,0x4bbbU,0x9776U,0x3ee7U,0x7dceU,0xfb9cU,0xe733U,0xde6dU,
		0xacd1U,0x49a9U,0x9352U,0x36afU,0x6d5eU,0xdabcU,0xa573U,0x5aedU,0xb5daU,0x7bbfU,
		0xf77eU,0xfef7U,0xede5U,0xcbc1U,0x8789U,0x1f19U,0x3e32U,0x7c64U,0xf8c8U,0xe19bU,
		0xd33dU,0xb671U,0x7ce9U,0xf9d2U,0xe3afU,0xd755U,0xbea1U,0x6d49U,0xda92U,0xa52fU,
		0x5a55U,0xb4aaU,0x795fU,0xf2beU,0xf577U,0xfae5U,0xe5c1U,0xdb89U,0xa719U,0x5e39U,
		0xbc72U,0x68efU,0xd1deU,0xb3b7U,0x7765U,0xeecaU,0xcd9fU,0x8b35U,0x0661U,0x0cc2U,
		0x1984U,0x3308U,0x6610U,0xcc20U,0x884bU,0x009dU,0x013aU,0x0274U,0x04e8U,0x09d0U,
		0x13a0U,0x2740U,0x4e80U,0x9d00U,0x2a0bU,0x5416U,0xa82cU,0x4053U,0x80a6U,0x1147U,
		0x228eU,0x451cU,0x8a38U,0x047bU,0x08f6U,0x11ecU,0x23d8U,0x47b0U,0x8f60U,0x0ecbU,
		0x1d96U,0x3b2cU,0x7658U,0xecb0U,0xc96bU,0x82ddU,0x15b1U,0x2b62U,0x56c4U,0xad88U,
		0x4b1bU,0x9636U,0x3c67U,0x78ceU,0xf19cU,0xf333U,0xf66dU,0xfcd1U,0xe9a9U,0xc359U,
		0x96b9U,0x3d79U,0x7af2U,0xf5e4U,0xfbc3U,0xe78dU,0xdf11U,0xae29U,0x4c59U,0x98b2U,
		0x216fU,0x42deU,0x85bcU,0x1b73U,0x36e6U,0x6dccU,0xdb98U,0xa73bU,0x5e7dU,0xbcfaU,
		0x69ffU,0xd3feU,0xb7f7U,0x7fe5U,0xffcaU,0xef9fU,0xcf35U,0x8e61U,0x0cc9U,0x1992U,
		0x3324U,0x6648U,0xcc90U,0x892bU,0x025dU,0x04baU,0x0974U,0x12e8U,0x25d0U,0x4ba0U,
		0x9740U,0x3e8bU,0x7d16U,0xfa2cU,0xe453U,0xd8adU,0xa151U,0x52a9U,0xa552U,0x5aafU,
		0xb55eU,0x7ab7U,0xf56eU,0xfad7U,0xe5a5U,0xdb41U,0xa689U,0x5d19U,0xba32U,0x646fU,
		0xc8deU,0x81b7U,0x1365U,0x26caU,0x4d94U,0x9b28U,0x265bU,0x4cb6U,0x996cU,0x22d3U,
		0x45a6U,0x8b4cU,0x0693U,0x0d26U,0x1a4cU,0x3498U,0x6930U,0xd260U,0xb4cbU,0x799dU,
		0xf33aU,0xf67fU,0xfcf5U,0xe9e1U,0xc3c9U,0x9799U,0x3f39U,0x7e72U,0xfce4U,0xe9c3U,
		0xc38dU,0x9711U,0x3e29U,0x7c52U,0xf8a4U,0xe143U,0xd28dU,0xb511U,0x7a29U,0xf452U,
		0xf8afU,0xe155U,0xd2a1U,0xb549U,0x7a99U,0xf532U,0xfa6fU,0xe4d5U,0xd9a1U,0xa349U,
		0x5699U,0xad32U,0x4a6fU,0x94deU,0x39b7U,0x736eU,0xe6dcU,0xddb3U,0xab6dU,0x46d1U,
		0x8da2U,0x0b4fU,0x169eU,0x2d3cU,0x5a78U,0xb4f0U,0x79ebU,0xf3d6U,0xf7a7U,0xff45U,
		0xee81U,0xcd09U,0x8a19U,0x0439U,0x0872U,0x10e4U,0x21c8U,0x4390U,0x8720U,0x1e4bU,
		0x3c96U,0x792cU,0xf258U,0xf4bbU,0xf97dU,0xe2f1U,0xd5e9U,0xbbd9U,0x67b9U,0xcf72U,
		0x8eefU,0x0dd5U,0x1baaU,0x3754U,0x6ea8U,0xdd50U,0xaaabU,0x455dU,0x8abaU,0x057fU,
		0x0afeU,0x15fcU,0x2bf8U,0x57f0U,0xafe0U,0x4fcbU,0x9f96U,0x2f27U,0x5e4eU,0xbc9cU,
		0x6933U,0xd266U,0xb4c7U,0x7985U,0xf30aU,0xf61fU,0xfc35U,0xe861U,0xc0c9U,0x9199U,
		0x3339U,0x6672U,0xcce4U,0x89c3U,0x038dU,0x071aU,0x0e34U,0x1c68U,0x38d0U,0x71a0U,
		0xe340U,0xd68bU,0xbd1dU,0x6a31U,0xd462U,0xb8cfU,0x6195U,0xc32aU,0x965fU,0x3cb5U,
		0x796aU,0xf2d4U,0xf5a3U,0xfb4dU,0xe691U,0xdd29U,0xaa59U,0x44b9U,0x8972U,0x02efU,
		0x05deU,0x0bbcU,0x1778U,0x2ef0U,0x5de0U,0xbbc0U,0x678bU,0xcf16U,0x8e27U,0x0c45U,
		0x188aU,0x3114U,0x6228U,0xc450U,0x98abU,0x215dU,0x42baU,0x8574U,0x1ae3U,0x35c6U,
		0x6b8cU,0xd718U,0xbe3bU,0x6c7dU,0xd8faU,0xa1ffU,0x53f5U,0xa7eaU,0x5fdfU,0xbfbeU,
		0x6f77U,0xdeeeU,0xadd7U,0x4ba5U,0x974aU,0x3e9fU,0x7d3eU,0xfa7cU,0xe4f3U,0xd9edU,
		0xa3d1U,0x57a9U,0xaf52U,0x4eafU,0x9d5eU,0x2ab7U,0x556eU,0xaadcU,0x45b3U,0x8b66U,
		0x06c7U,0x0d8eU,0x1b1cU,0x3638U,0x6c70U,0xd8e0U,0xa1cbU,0x539dU,0xa73aU,0x5e7fU,
		0xbcfeU,0x69f7U,0xd3eeU,0xb7d7U,0x7fa5U,0xff4aU,0xee9fU,0xcd35U,0x8a61U,0x04c9U,
		0x0992U,0x1324U,0x2648U,0x4c90U,0x9920U,0x224bU,0x4496U,0x892cU,0x0253U,0x04a6U,
		0x094cU,0x1298U,0x2530U,0x4a60U,0x94c0U,0x398bU,0x7316U,0xe62cU,0xdc53U,0xa8adU,
		0x4151U,0x82a2U,0x154fU,0x2a9eU,0x553cU,0xaa78U,0x44fbU,0x89f6U,0x03e7U,0x07ceU,
		0x0f9cU,0x1f38U,0x3e70U,0x7ce0U,0xf9c0U,0xe38bU,0xd71dU,0xbe31U,0x6c69U,0xd8d2U,
		0xa1afU,0x5355U,0xa6aaU,0x5d5fU,0xbabeU,0x6577U,0xcaeeU,0x85d7U,0x1ba5U,0x374aU,
		0x6e94U,0xdd28U,0xaa5bU,0x44bdU,0x897aU,0x02ffU,0x05feU,0x0bfcU,0x17f8U,0x2ff0U,
		0x5fe0U,0xbfc0U,0x6f8bU,0xdf16U,0xae27U,0x4c45U,0x988aU,0x211fU,0x423eU,0x847cU,
		0x18f3U,0x31e6U,0x63ccU,0xc798U,0x9f3bU,0x2e7dU,0x5cfaU,0xb9f4U,0x63e3U,0xc7c6U,
		0x9f87U,0x2f05U,0x5e0aU,0xbc14U,0x6823U,0xd046U,0xb087U,0x7105U,0xe20aU,0xd41fU,
		0xb835U,0x6061U,0xc0c2U,0x918fU,0x3315U,0x662aU,0xcc54U,0x88a3U,0x014dU,0x029aU,
		0x0534U,0x0a68U,0x14d0U,0x29a0U,0x5340U,0xa680U,0x5d0bU,0xba16U,0x6427U,0xc84eU,
		0x8097U,0x1125U,0x224aU,0x4494U,0x8928U,0x025bU,0x04b6U,0x096cU,0x12d8U,0x25b0U,
		0x4b60U,0x96c0U,0x3d8bU,0x7b16U,0xf62cU,0xfc53U,0xe8adU,0xc151U,0x92a9U,0x3559U,
		0x6ab2U,0xd564U,0xbac3U,0x658dU,0xcb1aU,0x863fU,0x1c75U,0x38eaU,0x71d4U,0xe3a8U,
		0xd75bU,0xbebdU,0x6d71U,0xdae2U,0xa5cfU,0x5b95U,0xb72aU,0x7e5fU,0xfcbeU,0xe977U,
		0xc2e5U,0x95c1U,0x3b89U,0x7712U,0xee24U,0xcc43U,0x888dU,0x0111U,0x0222U,0x0444U,
		0x0888U,0x1110U,0x2220U,0x4440U,0x8880U,0x010bU,0x0216U,0x042cU,0x0858U,0x10b0U,
		0x2160U,0x42c0U,0x8580U,0x1b0bU,0x3616U,0x6c2cU,0xd858U,0xa0bbU,0x517dU,0xa2faU,
		0x55ffU,0xabfeU,0x47f7U,0x8feeU,0x0fd7U,0x1faeU,0x3f5cU,0x7eb8U,0xfd70U,0xeaebU,
		0xc5ddU,0x9bb1U,0x2769U,0x4ed2U,0x9da4U,0x2b43U,0x5686U,0xad0cU,0x4a13U,0x9426U,
		0x3847U,0x708eU,0xe11cU,0xd233U,0xb46dU,0x78d1U,0xf1a2U,0xf34fU,0xf695U,0xfd21U,
		0xea49U,0xc499U,0x9939U,0x2279U,0x44f2U,0x89e4U,0x03c3U,0x0786U,0x0f0cU,0x1e18U,
		0x3c30U,0x7860U,0xf0c0U,0xf18bU,0xf31dU,0xf631U,0xfc69U,0xe8d9U,0xc1b9U,0x9379U,
		0x36f9U,0x6df2U,0xdbe4U,0xa7c3U,0x5f8dU,0xbf1aU,0x6e3fU,0xdc7eU,0xa8f7U,0x41e5U,
		0x83caU,0x179fU,0x2f3eU,0x5e7cU,0xbcf8U,0x69fbU,0xd3f6U,0xb7e7U,0x7fc5U,0xff8aU,
		0xef1fU,0xce35U,0x8c61U,0x08c9U,0x1192U,0x2324U,0x4648U,0x8c90U,0x092bU,0x1256U,
		0x24acU,0x4958U,0x92b0U,0x356bU,0x6ad6U,0xd5acU,0xbb53U,0x66adU,0xcd5aU,0x8abfU,
		0x0575U,0x0aeaU,0x15d4U,0x2ba8U,0x5750U,0xaea0U,0x4d4bU,0x9a96U,0x2527U,0x4a4eU,
		0x949cU,0x3933U,0x7266U,0xe4ccU,0xd993U,0xa32dU,0x5651U,0xaca2U,0x494fU,0x929eU,
		0x3537U,0x6a6eU,0xd4dcU,0xb9b3U,0x636dU,0xc6daU,0x9dbfU,0x2b75U,0x56eaU,0xadd4U,
		0x4ba3U,0x9746U,0x3e87U,0x7d0eU,0xfa1cU,0xe433U,0xd86dU,0xa0d1U,0x51a9U,0xa352U,
		0x56afU,0xad5eU,0x4ab7U,0x956eU,0x3ad7U,0x75aeU,0xeb5cU,0xc6b3U,0x9d6dU,0x2ad1U,
		0x55a2U,0xab44U,0x4683U,0x8d06U,0x0a07U,0x140eU,0x281cU,0x5038U,0xa070U,0x50ebU,
		0xa1d6U,0x53a7U,0xa74eU,0x5e97U,0xbd2eU,0x6a57U,0xd4aeU,0xb957U,0x62a5U,0xc54aU,
		0x9a9fU,0x2535U,0x4a6aU,0x94d4U,0x39a3U,0x7346U,0xe68cU,0xdd13U,0xaa2dU,0x4451U,
		0x88a2U,0x014fU,0x029eU,0x053cU,0x0a78U,0x14f0U,0x29e0U,0x53c0U,0xa780U,0x5f0bU,
		0xbe16U,0x6c27U,0xd84eU,0xa097U,0x5125U,0xa24aU,0x549fU,0xa93eU,0x4277U,0x84eeU,
		0x19d7U,0x33aeU,0x675cU,0xceb8U,0x8d7bU,0x0afdU,0x15faU,0x2bf4U,0x57e8U,0xafd0U,
		0x4fabU,0x9f56U,0x2ea7U,0x5d4eU,0xba9cU,0x6533U,0xca66U,0x84c7U,0x1985U,0x330aU,
		0x6614U,0xcc28U,0x885bU,0x00bdU,0x017aU,0x02f4U,0x05e8U,0x0bd0U,0x17a0U,0x2f40U,
		0x5e80U,0xbd00U,0x6a0bU,0xd416U,0xb827U,0x6045U,0xc08aU,0x911fU,0x3235U,0x646aU,
		0xc8d4U,0x81a3U,0x134dU,0x269aU,0x4d34U,0x9a68U,0x24dbU,0x49b6U,0x936cU,0x36d3U,
		0x6da6U,0xdb4cU,0xa693U,0x5d2dU,0xba5aU,0x64bfU,0xc97eU,0x82f7U,0x15e5U,0x2bcaU,
		0x5794U,0xaf28U,0x4e5bU,0x9cb6U,0x2967U,0x52ceU,0xa59cU,0x5b33U,0xb666U,0x7cc7U,
		0xf98eU,0xe317U,0xd625U,0xbc41U,0x6889U,0xd112U,0xb22fU,0x7455U,0xe8aaU,0xc15fU,
		0x92b5U,0x3561U,0x6ac2U,0xd584U,0xbb03U,0x660dU,0xcc1aU,0x883fU,0x0075U,0x00eaU,
		0x01d4U,0x03a8U,0x0750U,0x0ea0U,0x1d40U,0x3a80U,0x7500U,0xea00U,0xc40bU,0x981dU,
		0x2031U,0x4062U,0x80c4U,0x1183U,0x2306U,0x460cU,0x8c18U,0x083bU,0x1076U,0x20ecU,
		0x41d8U,0x83b0U,0x176bU,0x2ed6U,0x5dacU,0xbb58U,0x66bbU,0xcd76U,0x8ae7U,0x05c5U,
		0x0b8aU,0x1714U,0x2e28U,0x5c50U,0xb8a0U,0x614bU,0xc296U,0x9527U,0x3a45U,0x748aU,
		0xe914U,0xc223U,0x944dU,0x3891U,0x7122U,0xe244U,0xd483U,0xb90dU,0x6211U,0xc422U,
		0x984fU,0x2095U,0x412aU,0x8254U,0x14a3U,0x2946U,0x528cU,0xa518U,0x5a3bU,0xb476U,
		0x78e7U,0xf1ceU,0xf397U,0xf725U,0xfe41U,0xec89U,0xc919U,0x8239U,0x1479U,0x28f2U,
		0x51e4U,0xa3c8U,0x579bU,0xaf36U,0x4e67U,0x9cceU,0x2997U,0x532eU,0xa65cU,0x5cb3U,
		0xb966U,0x62c7U,0xc58eU,0x9b17U,0x2625U,0x4c4aU,0x9894U,0x2123U,0x4246U,0x848cU,
		0x1913U,0x3226U,0x644cU,0xc898U,0x813bU,0x127dU,0x24faU,0x49f4U,0x93e8U,0x37dbU,
		0x6fb6U,0xdf6cU,0xaed3U,0x4dadU,0x9b5aU,0x26bfU,0x4d7eU,0x9afcU,0x25f3U,0x4be6U,
		0x97ccU,0x3f93U,0x7f26U,0xfe4cU,0xec93U,0xc92dU,0x8251U,0x14a9U,0x2952U,0x52a4U,
		0xa548U,0x5a9bU,0xb536U,0x7a67U,0xf4ceU,0xf997U,0xe325U,0xd641U,0xbc89U,0x6919U,
		0xd232U,0xb46fU,0x78d5U,0xf1aaU,0xf35fU,0xf6b5U,0xfd61U,0xeac9U,0xc599U,0x9b39U,
		0x2679U,0x4cf2U,0x99e4U,0x23c3U,0x4786U,0x8f0cU,0x0e13U,0x1c26U,0x384cU,0x7098U,
		0xe130U,0xd26bU,0xb4ddU,0x79b1U,0xf362U,0xf6cfU,0xfd95U,0xeb21U,0xc649U,0x9c99U,
		0x2939U,0x5272U,0xa4e4U,0x59c3U,0xb386U,0x7707U,0xee0eU,0xcc17U,0x8825U,0x0041U,
		0x0082U,0x0104U,0x0208U,0x0410U,0x0820U,0x1040U,0x2080U,0x4100U,0x8200U,0x140bU,
		0x2816U,0x502cU,0xa058U,0x50bbU,0xa176U,0x52e7U,0xa5ceU,0x5b97U,0xb72eU,0x7e57U,
		0xfcaeU,0xe957U,0xc2a5U,0x9541U,0x3a89U,0x7512U,0xea24U,0xc443U,0x988dU,0x2111U,
		0x4222U,0x8444U,0x1883U,0x3106U,0x620cU,0xc418U,0x983bU,0x207dU,0x40faU,0x81f4U,
		0x13e3U,0x27c6U,0x4f8cU,0x9f18U,0x2e3bU,0x5c76U,0xb8ecU,0x61d3U,0xc3a6U,0x9747U,
		0x3e85U,0x7d0aU,0xfa14U,0xe423U,0xd84dU,0xa091U,0x5129U,0xa252U,0x54afU,0xa95eU,
		0x42b7U,0x856eU,0x1ad7U,0x35aeU,0x6b5cU,0xd6b8U,0xbd7bU,0x6afdU,0xd5faU,0xbbffU,
		0x67f5U,0xcfeaU,0x8fdfU,0x0fb5U,0x1f6aU,0x3ed4U,0x7da8U,0xfb50U,0xe6abU,0xdd5dU,
		0xaab1U,0x4569U,0x8ad2U,0x05afU,0x0b5eU,0x16bcU,0x2d78U,0x5af0U,0xb5e0U,0x7bcbU,
		0xf796U,0xff27U,0xee45U,0xcc81U,0x8909U,0x0219U,0x0432U,0x0864U,0x10c8U,0x2190U,
		0x4320U,0x8640U,0x1c8bU,0x3916U,0x722cU,0xe458U,0xd8bbU,0xa17dU,0x52f1U,0xa5e2U,
		0x5bcfU,0xb79eU,0x7f37U,0xfe6eU,0xecd7U,0xc9a5U,0x8341U,0x1689U,0x2d12U,0x5a24U,
		0xb448U,0x789bU,0xf136U,0xf267U,0xf4c5U,0xf981U,0xe309U,0xd619U,0xbc39U,0x6879U,
		0xd0f2U,0xb1efU,0x73d5U,0xe7aaU,0xdf5fU,0xaeb5U,0x4d61U,0x9ac2U,0x258fU,0x4b1eU,
		0x963cU,0x3c73U,0x78e6U,0xf1ccU,0xf393U,0xf72dU,0xfe51U,0xeca9U,0xc959U,0x82b9U,
		0x1579U,0x2af2U,0x55e4U,0xabc8U,0x479bU,0x8f36U,0x0e67U,0x1cceU,0x399cU,0x7338U,
		0xe670U,0xdcebU,0xa9ddU,0x43b1U,0x8762U,0x1ecfU,0x3d9eU,0x7b3cU,0xf678U,0xfcfbU,
		0xe9fdU,0xc3f1U,0x97e9U,0x3fd9U,0x7fb2U,0xff64U,0xeec3U,0xcd8dU,0x8b11U,0x0629U,
		0x0c52U,0x18a4U,0x3148U,0x6290U,0xc520U,0x9a4bU,0x249dU,0x493aU,0x9274U,0x34e3U,
		0x69c6U,0xd38cU,0xb713U,0x7e2dU,0xfc5aU,0xe8bfU,0xc175U,0x92e1U,0x35c9U,0x6b92U,
		0xd724U,0xbe43U,0x6c8dU,0xd91aU,0xa23fU,0x5475U,0xa8eaU,0x41dfU,0x83beU,0x1777U,
		0x2eeeU,0x5ddcU,0xbbb8U,0x677bU,0xcef6U,0x8de7U,0x0bc5U,0x178aU,0x2f14U,0x5e28U,
		0xbc50U,0x68abU,0xd156U,0xb2a7U,0x7545U,0xea8aU,0xc51fU,0x9a35U,0x2461U,0x48c2U,
		0x9184U,0x3303U,0x6606U,0xcc0cU,0x8813U,0x002dU,0x005aU,0x00b4U,0x0168U,0x02d0U,
		0x05a0U,0x0b40U,0x1680U,0x2d00U,0x5a00U,0xb400U,0x780bU,0xf016U,0xf027U,0xf045U,
		0xf081U,0xf109U,0xf219U,0xf439U,0xf879U,0xe0f9U,0xd1f9U,0xb3f9U,0x77f9U,0xeff2U,
		0xcfefU,0x8fd5U,0x0fa1U,0x1f42U,0x3e84U,0x7d08U,0xfa10U,0xe42bU,0xd85dU,0xa0b1U,
		0x5169U,0xa2d2U,0x55afU,0xab5eU,0x46b7U,0x8d6eU,0x0ad7U,0x15aeU,0x2b5cU,0x56b8U,
		0xad70U,0x4aebU,0x95d6U,0x3ba7U,0x774eU,0xee9cU,0xcd33U,0x8a6dU,0x04d1U,0x09a2U,
		0x1344U,0x2688U,0x4d10U,0x9a20U,0x244bU,0x4896U,0x912cU,0x3253U,0x64a6U,0xc94cU,
		0x8293U,0x152dU,0x2a5aU,0x54b4U,0xa968U,0x42dbU,0x85b6U,0x1b67U,0x36ceU,0x6d9cU,
		0xdb38U,0xa67bU,0x5cfdU,0xb9faU,0x63ffU,0xc7feU,0x9ff7U,0x2fe5U,0x5fcaU,0xbf94U,
		0x6f23U,0xde46U,0xac87U,0x4905U,0x920aU,0x341fU,0x683eU,0xd07cU,0xb0f3U,0x71edU,
		0xe3daU,0xd7bfU,0xbf75U,0x6ee1U,0xddc2U,0xab8fU,0x4715U,0x8e2aU,0x0c5fU,0x18beU,
		0x317cU,0x62f8U,0xc5f0U,0x9bebU,0x27ddU,0x4fbaU,0x9f74U,0x2ee3U,0x5dc6U,0xbb8cU,
		0x6713U,0xce26U,0x8c47U,0x0885U,0x110aU,0x2214U,0x4428U,0x8850U,0x00abU,0x0156U,
		0x02acU,0x0558U,0x0ab0U,0x1560U,0x2ac0U,0x5580U,0xab00U,0x460bU,0x8c16U,0x0827U,
		0x104eU,0x209cU,0x4138U,0x8270U,0x14ebU,0x29d6U,0x53acU,0xa758U,0x5ebbU,0xbd76U,
		0x6ae7U,0xd5ceU,0xbb97U,0x6725U,0xce4aU,0x8c9fU,0x0935U,0x126aU,0x24d4U,0x49a8U,
		0x9350U,0x36abU,0x6d56U,0xdaacU,0xa553U,0x5aadU,0xb55aU,0x7abfU,0xf57eU,0xfaf7U,
		0xe5e5U,0xdbc1U,0xa789U,0x5f19U,0xbe32U,0x6c6fU,0xd8deU,0xa1b7U,0x5365U,0xa6caU,
		0x5d9fU,0xbb3eU,0x6677U,0xcceeU,0x89d7U,0x03a5U,0x074aU,0x0e94U,0x1d28U,0x3a50U,
		0x74a0U,0xe940U,0xc28bU,0x951dU,0x3a31U,0x7462U,0xe8c4U,0xc183U,0x930dU,0x3611U,
		0x6c22U,0xd844U,0xa083U,0x510dU,0xa21aU,0x543fU,0xa87eU,0x40f7U,0x81eeU,0x13d7U,
		0x27aeU,0x4f5cU,0x9eb8U,0x2d7bU,0x5af6U,0xb5ecU,0x7bd3U,0xf7a6U,0xff47U,0xee85U,
		0xcd01U,0x8a09U,0x0419U,0x0832U,0x1064U,0x20c8U,0x4190U,0x8320U,0x164bU,0x2c96U,
		0x592cU,0xb258U,0x74bbU,0xe976U,0xc2e7U,0x95c5U,0x3b81U,0x7702U,0xee04U,0xcc03U,
		0x880dU,0x0011U,0x0022U,0x0044U,0x0088U,0x0110U,0x0220U,0x0440U,0x0880U,0x1100U,
		0x2200U,0x4400U,0x8800U,0x000bU,0x0016U,0x002cU,0x0058U,0x00b0U,0x0160U,0x02c0U,
		0x0580U,0x0b00U,0x1600U,0x2c00U,0x5800U,0xb000U,0x700bU,0xe016U,0xd027U,0xb045U,
		0x7081U,0xe102U,0xd20fU,0xb415U,0x7821U,0xf042U,0xf08fU,0xf115U,0xf221U,0xf449U,
		0xf899U,0xe139U,0xd279U,0xb4f9U,0x79f9U,0xf3f2U,0xf7efU,0xffd5U,0xefa1U,0xcf49U,
		0x8e99U,0x0d39U,0x1a72U,0x34e4U,0x69c8U,0xd390U,0xb72bU,0x7e5dU,0xfcbaU,0xe97fU,
		0xc2f5U,0x95e1U,0x3bc9U,0x7792U,0xef24U,0xce43U,0x8c8dU,0x0911U,0x1222U,0x2444U,
		0x4888U,0x9110U,0x322bU,0x6456U,0xc8acU,0x8153U,0x12adU,0x255aU,0x4ab4U,0x9568U,
		0x3adbU,0x75b6U,0xeb6cU,0xc6d3U,0x9dadU,0x2b51U,0x56a2U,0xad44U,0x4a83U,0x9506U,
		0x3a07U,0x740eU,0xe81cU,0xc033U,0x906dU,0x30d1U,0x61a2U,0xc344U,0x9683U,0x3d0dU,
		0x7a1aU,0xf434U,0xf863U,0xe0cdU,0xd191U,0xb329U,0x7659U,0xecb2U,0xc96fU,0x82d5U,
		0x15a1U,0x2b42U,0x5684U,0xad08U,0x4a1bU,0x9436U,0x3867U,0x70ceU,0xe19cU,0xd333U,
		0xb66dU,0x7cd1U,0xf9a2U,0xe34fU,0xd695U,0xbd21U,0x6a49U,0xd492U,0xb92fU,0x6255U,
		0xc4aaU,0x995fU,0x22b5U,0x456aU,0x8ad4U,0x05a3U,0x0b46U,0x168cU,0x2d18U,0x5a30U,
		0xb460U,0x78cbU,0xf196U,0xf327U,0xf645U,0xfc81U,0xe909U,0xc219U,0x9439U,0x3879U,
		0x70f2U,0xe1e4U,0xd3c3U,0xb78dU,0x7f11U,0xfe22U,0xec4fU,0xc895U,0x8121U,0x1249U,
		0x2492U,0x4924U,0x9248U,0x349bU,0x6936U,0xd26cU,0xb4d3U,0x79adU,0xf35aU,0xf6bfU,
		0xfd75U,0xeae1U,0xc5c9U,0x9b99U,0x2739U,0x4e72U,0x9ce4U,0x29c3U,0x5386U,0xa70cU,
		0x5e13U,0xbc26U,0x6847U,0xd08eU,0xb117U,0x7225U,0xe44aU,0xd89fU,0xa135U,0x5261U,
		0xa4c2U,0x598fU,0xb31eU,0x7637U,0xec6eU,0xc8d7U,0x81a5U,0x1341U,0x2682U,0x4d04U,
		0x9a08U,0x241bU,0x4836U,0x906cU,0x30d3U,0x61a6U,0xc34cU,0x9693U,0x3d2dU,0x7a5aU,
		0xf4b4U,0xf963U,0xe2cdU,0xd591U,0xbb29U,0x6659U,0xccb2U,0x896fU,0x02d5U,0x05aaU,
		0x0b54U,0x16a8U,0x2d50U,0x5aa0U,0xb540U,0x7a8bU,0xf516U,0xfa27U,0xe445U,0xd881U,
		0xa109U,0x5219U,0xa432U,0x586fU,0xb0deU,0x71b7U,0xe36eU,0xd6d7U,0xbda5U,0x6b41U,
		0xd682U,0xbd0fU,0x6a15U,0xd42aU,0xb85fU,0x60b5U,0xc16aU,0x92dfU,0x35b5U,0x6b6aU,
		0xd6d4U,0xbda3U,0x6b4dU,0xd69aU,0xbd3fU,0x6a75U,0xd4eaU,0xb9dfU,0x63b5U,0xc76aU,
		0x9edfU,0x2db5U,0x5b6aU,0xb6d4U,0x7da3U,0xfb46U,0xe687U,0xdd05U,0xaa01U,0x4409U,
		0x8812U,0x002fU,0x005eU,0x00bcU,0x0178U,0x02f0U,0x05e0U,0x0bc0U,0x1780U,0x2f00U,
		0x5e00U,0xbc00U,0x680bU,0xd016U,0xb027U,0x7045U,0xe08aU,0xd11fU,0xb235U,0x7461U,
		0xe8c2U,0xc18fU,0x9315U,0x3621U,0x6c42U,0xd884U,0xa103U,0x520dU,0xa41aU,0x583fU,
		0xb07eU,0x70f7U,0xe1eeU,0xd3d7U,0xb7a5U,0x7f41U,0xfe82U,0xed0fU,0xca15U,0x8421U,
		0x1849U,0x3092U,0x6124U,0xc248U,0x949bU,0x393dU,0x727aU,0xe4f4U,0xd9e3U,0xa3cdU,
		0x5791U,0xaf22U,0x4e4fU,0x9c9eU,0x2937U,0x526eU,0xa4dcU,0x59b3U,0xb366U,0x76c7U,
		0xed8eU,0xcb17U,0x8625U,0x1c41U,0x3882U,0x7104U,0xe208U,0xd41bU,0xb83dU,0x6071U,
		0xc0e2U,0x91cfU,0x3395U,0x672aU,0xce54U,0x8ca3U,0x094dU,0x129aU,0x2534U,0x4a68U,
		0x94d0U,0x39abU,0x7356U,0xe6acU,0xdd53U,0xaaadU,0x4551U,0x8aa2U,0x054fU,0x0a9eU,
		0x153cU,0x2a78U,0x54f0U,0xa9e0U,0x43cbU,0x8796U,0x1f27U,0x3e4eU,0x7c9cU,0xf938U,
		0xe27bU,0xd4fdU,0xb9f1U,0x63e9U,0xc7d2U,0x9fafU,0x2f55U,0x5eaaU,0xbd54U,0x6aa3U,
		0xd546U,0xba87U,0x6505U,0xca0aU,0x841fU,0x1835U,0x306aU,0x60d4U,0xc1a8U,0x935bU,
		0x36bdU,0x6d7aU,0xdaf4U,0xa5e3U,0x5bcdU,0xb79aU,0x7f3fU,0xfe7eU,0xecf7U,0xc9e5U,
		0x83c1U,0x1789U,0x2f12U,0x5e24U,0xbc48U,0x689bU,0xd136U,0xb267U,0x74c5U,0xe98aU,
		0xc31fU,0x9635U,0x3c61U,0x78c2U,0xf184U,0xf303U,0xf60dU,0xfc11U,0xe829U,0xc059U,
		0x90b9U,0x3179U,0x62f2U,0xc5e4U,0x9bc3U,0x278dU,0x4f1aU,0x9e34U,0x2c63U,0x58c6U,
		0xb18cU,0x7313U,0xe626U,0xdc47U,0xa885U,0x4101U,0x8202U,0x140fU,0x281eU,0x503cU,
		0xa078U,0x50fbU,0xa1f6U,0x53e7U,0xa7ceU,0x5f97U,0xbf2eU,0x6e57U,0xdcaeU,0xa957U,
		0x42a5U,0x854aU,0x1a9fU,0x353eU,0x6a7cU,0xd4f8U,0xb9fbU,0x63fdU,0xc7faU,0x9fffU,
		0x2ff5U,0x5feaU,0xbfd4U,0x6fa3U,0xdf46U,0xae87U,0x4d05U,0x9a0aU,0x241fU,0x483eU,
		0x907cU,0x30f3U,0x61e6U,0xc3ccU,0x9793U,0x3f2dU,0x7e5aU,0xfcb4U,0xe963U,0xc2cdU,
		0x9591U,0x3b29U,0x7652U,0xeca4U,0xc943U,0x828dU,0x1511U,0x2a22U,0x5444U,0xa888U,
		0x411bU,0x8236U,0x1467U,0x28ceU,0x519cU,0xa338U,0x567bU,0xacf6U,0x49e7U,0x93ceU,
		0x3797U,0x6f2eU,0xde5cU,0xacb3U,0x496dU,0x92daU,0x35bfU,0x6b7eU,0xd6fcU,0xbdf3U,
		0x6bedU,0xd7daU,0xbfbfU,0x6f75U,0xdeeaU,0xaddfU,0x4bb5U,0x976aU,0x3edfU,0x7dbeU,
		0xfb7cU,0xe6f3U,0xddedU,0xabd1U,0x47a9U,0x8f52U,0x0eafU,0x1d5eU,0x3abcU,0x7578U,
		0xeaf0U,0xc5ebU,0x9bddU,0x27b1U,0x4f62U,0x9ec4U,0x2d83U,0x5b06U,0xb60cU,0x7c13U,
		0xf826U,0xe047U,0xd085U,0xb101U,0x7209U,0xe412U,0xd82fU,0xa055U,0x50a1U,0xa142U,
		0x528fU,0xa51eU,0x5a37U,0xb46eU,0x78d7U,0xf1aeU,0xf357U,0xf6a5U,0xfd41U,0xea89U,
		0xc519U,0x9a39U,0x2479U,0x48f2U,0x91e4U,0x33c3U,0x6786U,0xcf0cU,0x8e13U,0x0c2dU,
		0x185aU,0x30b4U,0x6168U,0xc2d0U,0x95abU,0x3b5dU,0x76baU,0xed74U,0xcae3U,0x85cdU,
		0x1b91U,0x3722U,0x6e44U,0xdc88U,0xa91bU,0x423dU,0x847aU,0x18ffU,0x31feU,0x63fcU,
		0xc7f8U,0x9ffbU,0x2ffdU,0x5ffaU,0xbff4U,0x6fe3U,0xdfc6U,0xaf87U,0x4f05U,0x9e0aU,
		0x2c1fU,0x583eU,0xb07cU,0x70f3U,0xe1e6U,0xd3c7U,0xb785U,0x7f01U,0xfe02U,0xec0fU,
		0xc815U,0x8021U,0x1049U,0x2092U,0x4124U,0x8248U,0x149bU,0x2936U,0x526cU,0xa4d8U,
		0x59bbU,0xb376U,0x76e7U,0xedceU,0xcb97U,0x8725U,0x1e41U,0x3c82U,0x7904U,0xf208U,
		0xf41bU,0xf83dU,0xe071U,0xd0e9U,0xb1d9U,0x73b9U,0xe772U,0xdeefU,0xadd5U,0x4ba1U,
		0x9742U,0x3e8fU,0x7d1eU,0xfa3cU,0xe473U,0xd8edU,0xa1d1U,0x53a9U,0xa752U,0x5eafU,
		0xbd5eU,0x6ab7U,0xd56eU,0xbad7U,0x65a5U,0xcb4aU,0x869fU,0x1d35U,0x3a6aU,0x74d4U,
		0xe9a8U,0xc35bU,0x96bdU,0x3d71U,0x7ae2U,0xf5c4U,0xfb83U,0xe70dU,0xde11U,0xac29U,
		0x4859U,0x90b2U,0x316fU,0x62deU,0xc5bcU,0x9b73U,0x26edU,0x4ddaU,0x9bb4U,0x2763U,
		0x4ec6U,0x9d8cU,0x2b13U,0x5626U,0xac4cU,0x4893U,0x9126U,0x3247U,0x648eU,0xc91cU,
		0x8233U,0x146dU,0x28daU,0x51b4U,0xa368U,0x56dbU,0xadb6U,0x4b67U,0x96ceU,0x3d97U,
		0x7b2eU,0xf65cU,0xfcb3U,0xe96dU,0xc2d1U,0x95a9U,0x3b59U,0x76b2U,0xed64U,0xcac3U,
		0x858dU,0x1b11U,0x3622U,0x6c44U,0xd888U,0xa11bU,0x523dU,0xa47aU,0x58ffU,0xb1feU,
		0x73f7U,0xe7eeU,0xdfd7U,0xafa5U,0x4f41U,0x9e82U,0x2d0fU,0x5a1eU,0xb43cU,0x7873U,
		0xf0e6U,0xf1c7U,0xf385U,0xf701U,0xfe09U,0xec19U,0xc839U,0x8079U,0x10f9U,0x21f2U,
		0x43e4U,0x87c8U,0x1f9bU,0x3f36U,0x7e6cU,0xfcd8U,0xe9bbU,0xc37dU,0x96f1U,0x3de9U,
		0x7bd2U,0xf7a4U,0xff43U,0xee8dU,0xcd11U,0x8a29U,0x0459U,0x08b2U,0x1164U,0x22c8U,
		0x4590U,0x8b20U,0x064bU,0x0c96U,0x192cU,0x3258U,0x64b0U,0xc960U,0x82cbU,0x159dU,
		0x2b3aU,0x5674U,0xace8U,0x49dbU,0x93b6U,0x3767U,0x6eceU,0xdd9cU,0xab33U,0x466dU,
		0x8cdaU,0x09bfU,0x137eU,0x26fcU,0x4df8U,0x9bf0U,0x27ebU,0x4fd6U,0x9facU,0x2f53U,
		0x5ea6U,0xbd4cU,0x6a93U,0xd526U,0xba47U,0x6485U,0xc90aU,0x821fU,0x1435U,0x286aU,
		0x50d4U,0xa1a8U,0x535bU,0xa6b6U,0x5d67U,0xbaceU,0x6597U,0xcb2eU,0x8657U,0x1ca5U,
		0x394aU,0x7294U,0xe528U,0xda5bU,0xa4bdU,0x5971U,0xb2e2U,0x75cfU,0xeb9eU,0xc737U,
		0x9e65U,0x2cc1U,0x5982U,0xb304U,0x7603U,0xec06U,0xc807U,0x8005U,0x1001U,0x2002U,
		0x4004U,0x8008U,0x101bU,0x2036U,0x406cU,0x80d8U,0x11bbU,0x2376U,0x46ecU,0x8dd8U,
		0x0bbbU,0x1776U,0x2eecU,0x5dd8U,0xbbb0U,0x676bU,0xced6U,0x8da7U,0x0b45U,0x168aU,
		0x2d14U,0x5a28U,0xb450U,0x78abU,0xf156U,0xf2a7U,0xf545U,0xfa81U,0xe509U,0xda19U,
		0xa439U,0x5879U,0xb0f2U,0x71efU,0xe3deU,0xd7b7U,0xbf65U,0x6ec1U,0xdd82U,0xab0fU,
		0x4615U,0x8c2aU,0x085fU,0x10beU,0x217cU,0x42f8U,0x85f0U,0x1bebU,0x37d6U,0x6facU,
		0xdf58U,0xaebbU,0x4d7dU,0x9afaU,0x25ffU,0x4bfeU,0x97fcU,0x3ff3U,0x7fe6U,0xffccU,
		0xef93U,0xcf2dU,0x8e51U,0x0ca9U,0x1952U,0x32a4U,0x6548U,0xca90U,0x852bU,0x1a5dU,
		0x34baU,0x6974U,0xd2e8U,0xb5dbU,0x7bbdU,0xf77aU,0xfeffU,0xedf5U,0xcbe1U,0x87c9U,
		0x1f99U,0x3f32U,0x7e64U,0xfcc8U,0xe99bU,0xc33dU,0x9671U,0x3ce9U,0x79d2U,0xf3a4U,
		0xf743U,0xfe8dU,0xed11U,0xca29U,0x8459U,0x18b9U,0x3172U,0x62e4U,0xc5c8U,0x9b9bU,
		0x273dU,0x4e7aU,0x9cf4U,0x29e3U,0x53c6U,0xa78cU,0x5f13U,0xbe26U,0x6c47U,0xd88eU,
		0xa117U,0x5225U,0xa44aU,0x589fU,0xb13eU,0x7277U,0xe4eeU,0xd9d7U,0xa3a5U,0x5741U,
		0xae82U,0x4d0fU,0x9a1eU,0x2437U,0x486eU,0x90dcU,0x31b3U,0x6366U,0xc6ccU,0x9d93U,
		0x2b2dU,0x565aU,0xacb4U,0x4963U,0x92c6U,0x3587U,0x6b0eU,0xd61cU,0xbc33U,0x686dU,
		0xd0daU,0xb1bfU,0x7375U,0xe6eaU,0xdddfU,0xabb5U,0x4761U,0x8ec2U,0x0d8fU,0x1b1eU,
		0x363cU,0x6c78U,0xd8f0U,0xa1ebU,0x53ddU,0xa7baU,0x5f7fU,0xbefeU,0x6df7U,0xdbeeU,
		0xa7d7U,0x5fa5U,0xbf4aU,0x6e9fU,0xdd3eU,0xaa77U,0x44e5U,0x89caU,0x039fU,0x073eU,
		0x0e7cU,0x1cf8U,0x39f0U,0x73e0U,0xe7c0U,0xdf8bU,0xaf1dU,0x4e31U,0x9c62U,0x28cfU,
		0x519eU,0xa33cU,0x5673U,0xace6U,0x49c7U,0x938eU,0x3717U,0x6e2eU,0xdc5cU,0xa8b3U,
		0x416dU,0x82daU,0x15bfU,0x2b7eU,0x56fcU,0xadf8U,0x4bfbU,0x97f6U,0x3fe7U,0x7fceU,
		0xff9cU,0xef33U,0xce6dU,0x8cd1U,0x09a9U,0x1352U,0x26a4U,0x4d48U,0x9a90U,0x252bU,
		0x4a56U,0x94acU,0x3953U,0x72a6U,0xe54cU,0xda93U,0xa52dU,0x5a51U,0xb4a2U,0x794fU,
		0xf29eU,0xf537U,0xfa65U,0xe4c1U,0xd989U,0xa319U,0x5639U,0xac72U,0x48efU,0x91deU,
		0x33b7U,0x676eU,0xcedcU,0x8db3U,0x0b6dU,0x16daU,0x2db4U,0x5b68U,0xb6d0U,0x7dabU,
		0xfb56U,0xe6a7U,0xdd45U,0xaa81U,0x4509U,0x8a12U,0x042fU,0x085eU,0x10bcU,0x2178U,
		0x42f0U,0x85e0U,0x1bcbU,0x3796U,0x6f2cU,0xde58U,0xacbbU,0x497dU,0x92faU,0x35ffU,
		0x6bfeU,0xd7fcU,0xbff3U,0x6fedU,0xdfdaU,0xafbfU,0x4f75U,0x9eeaU,0x2ddfU,0x5bbeU,
		0xb77cU,0x7ef3U,0xfde6U,0xebc7U,0xc785U,0x9f01U,0x2e09U,0x5c12U,0xb824U,0x6043U,
		0xc086U,0x9107U,0x3205U,0x640aU,0xc814U,0x8023U,0x104dU,0x209aU,0x4134U,0x8268U,
		0x14dbU,0x29b6U,0x536cU,0xa6d8U,0x5dbbU,0xbb76U,0x66e7U,0xcdceU,0x8b97U,0x0725U,
		0x0e4aU,0x1c94U,0x3928U,0x7250U,0xe4a0U,0xd94bU,0xa29dU,0x5531U,0xaa62U,0x44cfU,
		0x899eU,0x0337U,0x066eU,0x0cdcU,0x19b8U,0x3370U,0x66e0U,0xcdc0U,0x8b8bU,0x071dU,
		0x0e3aU,0x1c74U,0x38e8U,0x71d0U,0xe3a0U,0xd74bU,0xbe9dU,0x6d31U,0xda62U,0xa4cfU,
		0x5995U,0xb32aU,0x765fU,0xecbeU,0xc977U,0x82e5U,0x15c1U,0x2b82U,0x5704U,0xae08U,
		0x4c1bU,0x9836U,0x2067U,0x40ceU,0x819cU,0x1333U,0x2666U,0x4cccU,0x9998U,0x233bU,
		0x4676U,0x8cecU,0x09d3U,0x13a6U,0x274cU,0x4e98U,0x9d30U,0x2a6bU,0x54d6U,0xa9acU,
		0x4353U,0x86a6U,0x1d47U,0x3a8eU,0x751cU,0xea38U,0xc47bU,0x98fdU,0x21f1U,0x43e2U,
		0x87c4U,0x1f83U,0x3f06U,0x7e0cU,0xfc18U,0xe83bU,0xc07dU,0x90f1U,0x31e9U,0x63d2U,
		0xc7a4U,0x9f43U,0x2e8dU,0x5d1aU,0xba34U,0x6463U,0xc8c6U,0x8187U,0x1305U,0x260aU,
		0x4c14U,0x9828U,0x205bU,0x40b6U,0x816cU,0x12d3U,0x25a6U,0x4b4cU,0x9698U,0x3d3bU,
		0x7a76U,0xf4ecU,0xf9d3U,0xe3adU,0xd751U,0xbea9U,0x6d59U,0xdab2U,0xa56fU,0x5ad5U,
		0xb5aaU,0x7b5fU,0xf6beU,0xfd77U,0xeae5U,0xc5c1U,0x9b89U,0x2719U,0x4e32U,0x9c64U,
		0x28c3U,0x5186U,0xa30cU,0x5613U,0xac26U,0x4847U,0x908eU,0x3117U,0x622eU,0xc45cU,
		0x98b3U,0x216dU,0x42daU,0x85b4U,0x1b63U,0x36c6U,0x6d8cU,0xdb18U,0xa63bU,0x5c7dU,
		0xb8faU,0x61ffU,0xc3feU,0x97f7U,0x3fe5U,0x7fcaU,0xff94U,0xef23U,0xce4dU,0x8c91U,
		0x0929U,0x1252U,0x24a4U,0x4948U,0x9290U,0x352bU,0x6a56U,0xd4acU,0xb953U,0x62adU,
		0xc55aU,0x9abfU,0x2575U,0x4aeaU,0x95d4U,0x3ba3U,0x7746U,0xee8cU,0xcd13U,0x8a2dU,
		0x0451U,0x08a2U,0x1144U,0x2288U,0x4510U,0x8a20U,0x044bU,0x0896U,0x112cU,0x2258U,
		0x44b0U,0x8960U,0x02cbU,0x0596U,0x0b2cU,0x1658U,0x2cb0U,0x5960U,0xb2c0U,0x758bU,
		0xeb16U,0xc627U,0x9c45U,0x2881U,0x5102U,0xa204U,0x5403U,0xa806U,0x4007U,0x800eU,
		0x1017U,0x202eU,0x405cU,0x80b8U,0x117bU,0x22f6U,0x45ecU,0x8bd8U,0x07bbU,0x0f76U,
		0x1eecU,0x3dd8U,0x7bb0U,0xf760U,0xfecbU,0xed9dU,0xcb31U,0x8669U,0x1cd9U,0x39b2U,
		0x7364U,0xe6c8U,0xdd9bU,0xab3dU,0x4671U,0x8ce2U,0x09cfU,0x139eU,0x273cU,0x4e78U,
		0x9cf0U,0x29ebU,0x53d6U,0xa7acU,0x5f53U,0xbea6U,0x6d47U,0xda8eU,0xa517U,0x5a25U,
		0xb44aU,0x789fU,0xf13eU,0xf277U,0xf4e5U,0xf9c1U,0xe389U,0xd719U,0xbe39U,0x6c79U,
		0xd8f2U,0xa1efU,0x53d5U,0xa7aaU,0x5f5fU,0xbebeU,0x6d77U,0xdaeeU,0xa5d7U,0x5ba5U,
		0xb74aU,0x7e9fU,0xfd3eU,0xea77U,0xc4e5U,0x99c1U,0x2389U,0x4712U,0x8e24U,0x0c43U,
		0x1886U,0x310cU,0x6218U,0xc430U,0x986bU,0x20ddU,0x41baU,0x8374U,0x16e3U,0x2dc6U,
		0x5b8cU,0xb718U,0x7e3bU,0xfc76U,0xe8e7U,0xc1c5U,0x9381U,0x3709U,0x6e12U,0xdc24U,
		0xa843U,0x408dU,0x811aU,0x123fU,0x247eU,0x48fcU,0x91f8U,0x33fbU,0x67f6U,0xcfecU,
		0x8fd3U,0x0fadU,0x1f5aU,0x3eb4U,0x7d68U,0xfad0U,0xe5abU,0xdb5dU,0xa6b1U,0x5d69U,
		0xbad2U,0x65afU,0xcb5eU,0x86b7U,0x1d65U,0x3acaU,0x7594U,0xeb28U,0xc65bU,0x9cbdU,
		0x2971U,0x52e2U,0xa5c4U,0x5b83U,0xb706U,0x7e07U,0xfc0eU,0xe817U,0xc025U,0x9041U,
		0x3089U,0x6112U,0xc224U,0x9443U,0x388dU,0x711aU,0xe234U,0xd463U,0xb8cdU,0x6191U,
		0xc322U,0x964fU,0x3c95U,0x792aU,0xf254U,0xf4a3U,0xf94dU,0xe291U,0xd529U,0xba59U,
		0x64b9U,0xc972U,0x82efU,0x15d5U,0x2baaU,0x5754U,0xaea8U,0x4d5bU,0x9ab6U,0x2567U,
		0x4aceU,0x959cU,0x3b33U,0x7666U,0xecccU,0xc993U,0x832dU,0x1651U,0x2ca2U,0x5944U,
		0xb288U,0x751bU,0xea36U,0xc467U,0x98c5U,0x2181U,0x4302U,0x8604U,0x1c03U,0x3806U,
		0x700cU,0xe018U,0xd03bU,0xb07dU,0x70f1U,0xe1e2U,0xd3cfU,0xb795U,0x7f21U,0xfe42U,
		0xec8fU,0xc915U,0x8221U,0x1449U,0x2892U,0x5124U,0xa248U,0x549bU,0xa936U,0x4267U,
		0x84ceU,0x1997U,0x332eU,0x665cU,0xccb8U,0x897bU,0x02fdU,0x05faU,0x0bf4U,0x17e8U,
		0x2fd0U,0x5fa0U,0xbf40U,0x6e8bU,0xdd16U,0xaa27U,0x4445U,0x888aU,0x011fU,0x023eU,
		0x047cU,0x08f8U,0x11f0U,0x23e0U,0x47c0U,0x8f80U,0x0f0bU,0x1e16U,0x3c2cU,0x7858U,
		0xf0b0U,0xf16bU,0xf2ddU,0xf5b1U,0xfb69U,0xe6d9U,0xddb9U,0xab79U,0x46f9U,0x8df2U,
		0x0befU,0x17deU,0x2fbcU,0x5f78U,0xbef0U,0x6debU,0xdbd6U,0xa7a7U,0x5f45U,0xbe8aU,
		0x6d1fU,0xda3eU,0xa477U,0x58e5U,0xb1caU,0x739fU,0xe73eU,0xde77U,0xace5U,0x49c1U,
		0x9382U,0x370fU,0x6e1eU,0xdc3cU,0xa873U,0x40edU,0x81daU,0x13bfU,0x277eU,0x4efcU,
		0x9df8U,0x2bfbU,0x57f6U,0xafecU,0x4fd3U,0x9fa6U,0x2f47U,0x5e8eU,0xbd1cU,0x6a33U,
		0xd466U,0xb8c7U,0x6185U,0xc30aU,0x961fU,0x3c35U,0x786aU,0xf0d4U,0xf1a3U,0xf34dU,
		0xf691U,0xfd29U,0xea59U,0xc4b9U,0x9979U,0x22f9U,0x45f2U,0x8be4U,0x07c3U,0x0f86U,
		0x1f0cU,0x3e18U,0x7c30U,0xf860U,0xe0cbU,0xd19dU,0xb331U,0x7669U,0xecd2U,0xc9afU,
		0x8355U,0x16a1U,0x2d42U,0x5a84U,0xb508U,0x7a1bU,0xf436U,0xf867U,0xe0c5U,0xd181U,
		0xb309U,0x7619U,0xec32U,0xc86fU,0x80d5U,0x11a1U,0x2342U,0x4684U,0x8d08U,0x0a1bU,
		0x1436U,0x286cU,0x50d8U,0xa1b0U,0x536bU,0xa6d6U,0x5da7U,0xbb4eU,0x6697U,0xcd2eU,
		0x8a57U,0x04a5U,0x094aU,0x1294U,0x2528U,0x4a50U,0x94a0U,0x394bU,0x7296U,0xe52cU,
		0xda53U,0xa4adU,0x5951U,0xb2a2U,0x754fU,0xea9eU,0xc537U,0x9a65U,0x24c1U,0x4982U,
		0x9304U,0x3603U,0x6c06U,0xd80cU,0xa013U,0x502dU,0xa05aU,0x50bfU,0xa17eU,0x52f7U,
		0xa5eeU,0x5bd7U,0xb7aeU,0x7f57U,0xfeaeU,0xed57U,0xcaa5U,0x8541U,0x1a89U,0x3512U,
		0x6a24U,0xd448U,0xb89bU,0x613dU,0xc27aU,0x94ffU,0x39f5U,0x73eaU,0xe7d4U,0xdfa3U,
		0xaf4dU,0x4e91U,0x9d22U,0x2a4fU,0x549eU,0xa93cU,0x4273U,0x84e6U,0x19c7U,0x338eU,
		0x671cU,0xce38U,0x8c7bU,0x08fdU,0x11faU,0x23f4U,0x47e8U,0x8fd0U,0x0fabU,0x1f56U,
		0x3eacU,0x7d58U,0xfab0U,0xe56bU,0xdaddU,0xa5b1U,0x5b69U,0xb6d2U,0x7dafU,0xfb5eU,
		0xe6b7U,0xdd65U,0xaac1U,0x4589U,0x8b12U,0x062fU,0x0c5eU,0x18bcU,0x3178U,0x62f0U,
		0xc5e0U,0x9bcbU,0x279dU,0x4f3aU,0x9e74U,0x2ce3U,0x59c6U,0xb38cU,0x7713U,0xee26U,
		0xcc47U,0x8885U,0x0101U,0x0202U,0x0404U,0x0808U,0x1010U,0x2020U,0x4040U,0x8080U,
		0x110bU,0x2216U,0x442cU,0x8858U,0x00bbU,0x0176U,0x02ecU,0x05d8U,0x0bb0U,0x1760U,
		0x2ec0U,0x5d80U,0xbb00U,0x660bU,0xcc16U,0x8827U,0x0045U,0x008aU,0x0114U,0x0228U,
		0x0450U,0x08a0U,0x1140U,0x2280U,0x4500U,0x8a00U,0x040bU,0x0816U,0x102cU,0x2058U,
		0x40b0U,0x8160U,0x12cbU,0x2596U,0x4b2cU,0x9658U,0x3cbbU,0x7976U,0xf2ecU,0xf5d3U,
		0xfbadU,0xe751U,0xdea9U,0xad59U,0x4ab9U,0x9572U,0x3aefU,0x75deU,0xebbcU,0xc773U,
		0x9eedU,0x2dd1U,0x5ba2U,0xb744U,0x7e83U,0xfd06U,0xea07U,0xc405U,0x9801U,0x2009U,
		0x4012U,0x8024U,0x1043U,0x2086U,0x410cU,0x8218U,0x143bU,0x2876U,0x50ecU,0xa1d8U,
		0x53bbU,0xa776U,0x5ee7U,0xbdceU,0x6b97U,0xd72eU,0xbe57U,0x6ca5U,0xd94aU,0xa29fU,
		0x5535U,0xaa6aU,0x44dfU,0x89beU,0x0377U,0x06eeU,0x0ddcU,0x1bb8U,0x3770U,0x6ee0U,
		0xddc0U,0xab8bU,0x471dU,0x8e3aU,0x0c7fU,0x18feU,0x31fcU,0x63f8U,0xc7f0U,0x9febU,
		0x2fddU,0x5fbaU,0xbf74U,0x6ee3U,0xddc6U,0xab87U,0x4705U,0x8e0aU,0x0c1fU,0x183eU,
		0x307cU,0x60f8U,0xc1f0U,0x93ebU,0x37ddU,0x6fbaU,0xdf74U,0xaee3U,0x4dcdU,0x9b9aU,
		0x273fU,0x4e7eU,0x9cfcU,0x29f3U,0x53e6U,0xa7ccU,0x5f93U,0xbf26U,0x6e47U,0xdc8eU,
		0xa917U,0x4225U,0x844aU,0x189fU,0x313eU,0x627cU,0xc4f8U,0x99fbU,0x23fdU,0x47faU,
		0x8ff4U,0x0fe3U,0x1fc6U,0x3f8cU,0x7f18U,0xfe30U,0xec6bU,0xc8ddU,0x81b1U,0x1369U,
		0x26d2U,0x4da4U,0x9b48U,0x269bU,0x4d36U,0x9a6cU,0x24d3U,0x49a6U,0x934cU,0x3693U,
		0x6d26U,0xda4cU,0xa493U,0x592dU,0xb25aU,0x74bfU,0xe97eU,0xc2f7U,0x95e5U,0x3bc1U,
		0x7782U,0xef04U,0xce03U,0x8c0dU,0x0811U,0x1022U,0x2044U,0x4088U,0x8110U,0x122bU,
		0x2456U,0x48acU,0x9158U,0x32bbU,0x6576U,0xcaecU,0x85d3U,0x1badU,0x375aU,0x6eb4U,
		0xdd68U,0xaadbU,0x45bdU,0x8b7aU,0x06ffU,0x0dfeU,0x1bfcU,0x37f8U,0x6ff0U,0xdfe0U,
		0xafcbU,0x4f9dU,0x9f3aU,0x2e7fU,0x5cfeU,0xb9fcU,0x63f3U,0xc7e6U,0x9fc7U,0x2f85U,
		0x5f0aU,0xbe14U,0x6c23U,0xd846U,0xa087U,0x5105U,0xa20aU,0x541fU,0xa83eU,0x4077U,
		0x80eeU,0x11d7U,0x23aeU,0x475cU,0x8eb8U,0x0d7bU,0x1af6U,0x35ecU,0x6bd8U,0xd7b0U,
		0xbf6bU,0x6eddU,0xddbaU,0xab7fU,0x46f5U,0x8deaU,0x0bdfU,0x17beU,0x2f7cU,0x5ef8U,
		0xbdf0U,0x6bebU,0xd7d6U,0xbfa7U,0x6f45U,0xde8aU,0xad1fU,0x4a35U,0x946aU,0x38dfU,
		0x71beU,0xe37cU,0xd6f3U,0xbdedU,0x6bd1U,0xd7a2U,0xbf4fU,0x6e95U,0xdd2aU,0xaa5fU,
		0x44b5U,0x896aU,0x02dfU,0x05beU,0x0b7cU,0x16f8U,0x2df0U,0x5be0U,0xb7c0U,0x7f8bU,
		0xff16U,0xee27U,0xcc45U,0x8881U,0x0109U,0x0212U,0x0424U,0x0848U,0x1090U,0x2120U,
		0x4240U,0x8480U,0x190bU,0x3216U,0x642cU,0xc858U,0x80bbU,0x117dU,0x22faU,0x45f4U,
		0x8be8U,0x07dbU,0x0fb6U,0x1f6cU,0x3ed8U,0x7db0U,0xfb60U,0xe6cbU,0xdd9dU,0xab31U,
		0x4669U,0x8cd2U,0x09afU,0x135eU,0x26bcU,0x4d78U,0x9af0U,0x25ebU,0x4bd6U,0x97acU,
		0x3f53U,0x7ea6U,0xfd4cU,0xea93U,0xc52dU,0x9a51U,0x24a9U,0x4952U,0x92a4U,0x3543U,
		0x6a86U,0xd50cU,0xba13U,0x642dU,0xc85aU,0x80bfU,0x1175U,0x22eaU,0x45d4U,0x8ba8U,
		0x075bU,0x0eb6U,0x1d6cU,0x3ad8U,0x75b0U,0xeb60U,0xc6cbU,0x9d9dU,0x2b31U,0x5662U,
		0xacc4U,0x4983U,0x9306U,0x3607U,0x6c0eU,0xd81cU,0xa033U,0x506dU,0xa0daU,0x51bfU,
		0xa37eU,0x56f7U,0xadeeU,0x4bd7U,0x97aeU,0x3f57U,0x7eaeU,0xfd5cU,0xeab3U,0xc56dU,
		0x9ad1U,0x25a9U,0x4b52U,0x96a4U,0x3d43U,0x7a86U,0xf50cU,0xfa13U,0xe42dU,0xd851U,
		0xa0a9U,0x5159U,0xa2b2U,0x556fU,0xaadeU,0x45b7U,0x8b6eU,0x06d7U,0x0daeU,0x1b5cU,
		0x36b8U,0x6d70U,0xdae0U,0xa5cbU,0x5b9dU,0xb73aU,0x7e7fU,0xfcfeU,0xe9f7U,0xc3e5U,
		0x97c1U,0x3f89U,0x7f12U,0xfe24U,0xec43U,0xc88dU,0x8111U,0x1229U,0x2452U,0x48a4U,
		0x9148U,0x329bU,0x6536U,0xca6cU,0x84d3U,0x19adU,0x335aU,0x66b4U,0xcd68U,0x8adbU,
		0x05bdU,0x0b7aU,0x16f4U,0x2de8U,0x5bd0U,0xb7a0U,0x7f4bU,0xfe96U,0xed27U,0xca45U,
		0x8481U,0x1909U,0x3212U,0x6424U,0xc848U,0x809bU,0x113dU,0x227aU,0x44f4U,0x89e8U,
		0x03dbU,0x07b6U,0x0f6cU,0x1ed8U,0x3db0U,0x7b60U,0xf6c0U,0xfd8bU,0xeb1dU,0xc631U,
		0x9c69U,0x28d9U,0x51b2U,0xa364U,0x56c3U,0xad86U,0x4b07U,0x960eU,0x3c17U,0x782eU,
		0xf05cU,0xf0b3U,0xf16dU,0xf2d1U,0xf5a9U,0xfb59U,0xe6b9U,0xdd79U,0xaaf9U,0x45f9U,
		0x8bf2U,0x07efU,0x0fdeU,0x1fbcU,0x3f78U,0x7ef0U,0xfde0U,0xebcbU,0xc79dU,0x9f31U,
		0x2e69U,0x5cd2U,0xb9a4U,0x6343U,0xc686U,0x9d07U,0x2a05U,0x540aU,0xa814U,0x4023U,
		0x8046U,0x1087U,0x210eU,0x421cU,0x8438U,0x187bU,0x30f6U,0x61ecU,0xc3d8U,0x97bbU,
		0x3f7dU,0x7efaU,0xfdf4U,0xebe3U,0xc7cdU,0x9f91U,0x2f29U,0x5e52U,0xbca4U,0x6943U,
		0xd286U,0xb507U,0x7a05U,0xf40aU,0xf81fU,0xe035U,0xd061U,0xb0c9U,0x7199U,0xe332U,
		0xd66fU,0xbcd5U,0x69a1U,0xd342U,0xb68fU,0x7d15U,0xfa2aU,0xe45fU,0xd8b5U,0xa161U,
		0x52c9U,0xa592U,0x5b2fU,0xb65eU,0x7cb7U,0xf96eU,0xe2d7U,0xd5a5U,0xbb41U,0x6689U,
		0xcd12U,0x8a2fU,0x0455U,0x08aaU,0x1154U,0x22a8U,0x4550U,0x8aa0U,0x054bU,0x0a96U,
		0x152cU,0x2a58U,0x54b0U,0xa960U,0x42cbU,0x8596U,0x1b27U,0x364eU,0x6c9cU,0xd938U,
		0xa27bU,0x54fdU,0xa9faU,0x43ffU,0x87feU,0x1ff7U,0x3feeU,0x7fdcU,0xffb8U,0xef7bU,
		0xcefdU,0x8df1U,0x0be9U,0x17d2U,0x2fa4U,0x5f48U,0xbe90U,0x6d2bU,0xda56U,0xa4a7U,
		0x5945U,0xb28aU,0x751fU,0xea3eU,0xc477U,0x98e5U,0x21c1U,0x4382U,0x8704U,0x1e03U,
		0x3c06U,0x780cU,0xf018U,0xf03bU,0xf07dU,0xf0f1U,0xf1e9U,0xf3d9U,0xf7b9U,0xff79U,
		0xeef9U,0xcdf9U,0x8bf9U,0x07f9U,0x0ff2U,0x1fe4U,0x3fc8U,0x7f90U,0xff20U,0xee4bU,
		0xcc9dU,0x8931U,0x0269U,0x04d2U,0x09a4U,0x1348U,0x2690U,0x4d20U,0x9a40U,0x248bU,
		0x4916U,0x922cU,0x3453U,0x68a6U,0xd14cU,0xb293U,0x752dU,0xea5aU,0xc4bfU,0x9975U,
		0x22e1U,0x45c2U,0x8b84U,0x0703U,0x0e06U,0x1c0cU,0x3818U,0x7030U,0xe060U,0xd0cbU,
		0xb19dU,0x7331U,0xe662U,0xdccfU,0xa995U,0x4321U,0x8642U,0x1c8fU,0x391eU,0x723cU,
		0xe478U,0xd8fbU,0xa1fdU,0x53f1U,0xa7e2U,0x5fcfU,0xbf9eU,0x6f37U,0xde6eU,0xacd7U,
		0x49a5U,0x934aU,0x369fU,0x6d3eU,0xda7cU,0xa4f3U,0x59edU,0xb3daU,0x77bfU,0xef7eU,
		0xcef7U,0x8de5U,0x0bc1U,0x1782U,0x2f04U,0x5e08U,0xbc10U,0x682bU,0xd056U,0xb0a7U,
		0x7145U,0xe28aU,0xd51fU,0xba35U,0x6461U,0xc8c2U,0x818fU,0x1315U,0x262aU,0x4c54U,
		0x98a8U,0x215bU,0x42b6U,0x856cU,0x1ad3U,0x35a6U,0x6b4cU,0xd698U,0xbd3bU,0x6a7dU,
		0xd4faU,0xb9ffU,0x63f5U,0xc7eaU,0x9fdfU,0x2fb5U,0x5f6aU,0xbed4U,0x6da3U,0xdb46U,
		0xa687U,0x5d05U,0xba0aU,0x641fU,0xc83eU,0x8077U,0x10e5U,0x21caU,0x4394U,0x8728U,
		0x1e5bU,0x3cb6U,0x796cU,0xf2d8U,0xf5bbU,0xfb7dU,0xe6f1U,0xdde9U,0xabd9U,0x47b9U,
		0x8f72U,0x0eefU,0x1ddeU,0x3bbcU,0x7778U,0xeef0U,0xcdebU,0x8bddU,0x07b1U,0x0f62U,
		0x1ec4U,0x3d88U,0x7b10U,0xf620U,0xfc4bU,0xe89dU,0xc131U,0x9269U,0x34d9U,0x69b2U,
		0xd364U,0xb6c3U,0x7d8dU,0xfb1aU,0xe63fU,0xdc75U,0xa8e1U,0x41c9U,0x8392U,0x172fU,
		0x2e5eU,0x5cbcU,0xb978U,0x62fbU,0xc5f6U,0x9be7U,0x27c5U,0x4f8aU,0x9f14U,0x2e23U,
		0x5c46U,0xb88cU,0x6113U,0xc226U,0x9447U,0x3885U,0x710aU,0xe214U,0xd423U,0xb84dU,
		0x6091U,0xc122U,0x924fU,0x3495U,0x692aU,0xd254U,0xb4a3U,0x794dU,0xf29aU,0xf53fU,
		0xfa75U,0xe4e1U,0xd9c9U,0xa399U,0x5739U,0xae72U,0x4cefU,0x99deU,0x23b7U,0x476eU,
		0x8edcU,0x0db3U,0x1b66U,0x36ccU,0x6d98U,0xdb30U,0xa66bU,0x5cddU,0xb9baU,0x637fU,
		0xc6feU,0x9df7U,0x2be5U,0x57caU,0xaf94U,0x4f23U,0x9e46U,0x2c87U,0x590eU,0xb21cU,
		0x7433U,0xe866U,0xc0c7U,0x9185U,0x3301U,0x6602U,0xcc04U,0x8803U,0x000dU,0x001aU,
		0x0034U,0x0068U,0x00d0U,0x01a0U,0x0340U,0x0680U,0x0d00U,0x1a00U,0x3400U,0x6800U,
		0xd000U,0xb00bU,0x701dU,0xe03aU,0xd07fU,0xb0f5U,0x71e1U,0xe3c2U,0xd78fU,0xbf15U,
		0x6e21U,0xdc42U,0xa88fU,0x4115U,0x822aU,0x145fU,0x28beU,0x517cU,0xa2f8U,0x55fbU,
		0xabf6U,0x47e7U,0x8fceU,0x0f97U,0x1f2eU,0x3e5cU,0x7cb8U,0xf970U,0xe2ebU,0xd5ddU,
		0xbbb1U,0x6769U,0xced2U,0x8dafU,0x0b55U,0x16aaU,0x2d54U,0x5aa8U,0xb550U,0x7aabU,
		0xf556U,0xfaa7U,0xe545U,0xda81U,0xa509U,0x5a19U,0xb432U,0x786fU,0xf0deU,0xf1b7U,
		0xf365U,0xf6c1U,0xfd89U,0xeb19U,0xc639U,0x9c79U,0x28f9U,0x51f2U,0xa3e4U,0x57c3U,
		0xaf86U,0x4f07U,0x9e0eU,0x2c17U,0x582eU,0xb05cU,0x70b3U,0xe166U,0xd2c7U,0xb585U,
		0x7b01U,0xf602U,0xfc0fU,0xe815U,0xc021U,0x9049U,0x3099U,0x6132U,0xc264U,0x94c3U,
		0x398dU,0x731aU,0xe634U,0xdc63U,0xa8cdU,0x4191U,0x8322U,0x164fU,0x2c9eU,0x593cU,
		0xb278U,0x74fbU,0xe9f6U,0xc3e7U,0x97c5U,0x3f81U,0x7f02U,0xfe04U,0xec03U,0xc80dU,
		0x8011U,0x1029U,0x2052U,0x40a4U,0x8148U,0x129bU,0x2536U,0x4a6cU,0x94d8U,0x39bbU,
		0x7376U,0xe6ecU,0xddd3U,0xabadU,0x4751U,0x8ea2U,0x0d4fU,0x1a9eU,0x353cU,0x6a78U,
		0xd4f0U,0xb9ebU,0x63ddU,0xc7baU,0x9f7fU,0x2ef5U,0x5deaU,0xbbd4U,0x67a3U,0xcf46U,
		0x8e87U,0x0d05U,0x1a0aU,0x3414U,0x6828U,0xd050U,0xb0abU,0x715dU,0xe2baU,0xd57fU,
		0xbaf5U,0x65e1U,0xcbc2U,0x878fU,0x1f15U,0x3e2aU,0x7c54U,0xf8a8U,0xe15bU,0xd2bdU,
		0xb571U,0x7ae9U,0xf5d2U,0xfbafU,0xe755U,0xdea1U,0xad49U,0x4a99U,0x9532U,0x3a6fU,
		0x74deU,0xe9bcU,0xc373U,0x96edU,0x3dd1U,0x7ba2U,0xf744U,0xfe83U,0xed0dU,0xca11U,
		0x8429U,0x1859U,0x30b2U,0x6164U,0xc2c8U,0x959bU,0x3b3dU,0x767aU,0xecf4U,0xc9e3U,
		0x83cdU,0x1791U,0x2f22U,0x5e44U,0xbc88U,0x691bU,0xd236U,0xb467U,0x78c5U,0xf18aU,
		0xf31fU,0xf635U,0xfc61U,0xe8c9U,0xc199U,0x9339U,0x3679U,0x6cf2U,0xd9e4U,0xa3c3U,
		0x578dU,0xaf1aU,0x4e3fU,0x9c7eU,0x28f7U,0x51eeU,0xa3dcU,0x57b3U,0xaf66U,0x4ec7U,
		0x9d8eU,0x2b17U,0x562eU,0xac5cU,0x48b3U,0x9166U,0x32c7U,0x658eU,0xcb1cU,0x8633U,
		0x1c6dU,0x38daU,0x71b4U,0xe368U,0xd6dbU,0xbdbdU,0x6b71U,0xd6e2U,0xbdcfU,0x6b95U,
		0xd72aU,0xbe5fU,0x6cb5U,0xd96aU,0xa2dfU,0x55b5U,0xab6aU,0x46dfU,0x8dbeU,0x0b77U,
		0x16eeU,0x2ddcU,0x5bb8U,0xb770U,0x7eebU,0xfdd6U,0xeba7U,0xc745U,0x9e81U,0x2d09U,
		0x5a12U,0xb424U,0x7843U,0xf086U,0xf107U,0xf205U,0xf401U,0xf809U,0xe019U,0xd039U,
		0xb079U,0x70f9U,0xe1f2U,0xd3efU,0xb7d5U,0x7fa1U,0xff42U,0xee8fU,0xcd15U,0x8a21U,
		0x0449U,0x0892U,0x1124U,0x2248U,0x4490U,0x8920U,0x024bU,0x0496U,0x092cU,0x1258U,
		0x24b0U,0x4960U,0x92c0U,0x358bU,0x6b16U,0xd62cU,0xbc53U,0x68adU,0xd15aU,0xb2bfU,
		0x7575U,0xeaeaU,0xc5dfU,0x9bb5U,0x2761U,0x4ec2U,0x9d84U,0x2b03U,0x5606U,0xac0cU,
		0x4813U,0x9026U,0x3047U,0x608eU,0xc11cU,0x9233U,0x346dU,0x68daU,0xd1b4U,0xb363U,
		0x76cdU,0xed9aU,0xcb3fU,0x8675U,0x1ce1U,0x39c2U,0x7384U,0xe708U,0xde1bU,0xac3dU,
		0x4871U,0x90e2U,0x31cfU,0x639eU,0xc73cU,0x9e73U,0x2cedU,0x59daU,0xb3b4U,0x7763U,
		0xeec6U,0xcd87U,0x8b05U,0x0601U,0x0c02U,0x1804U,0x3008U,0x6010U,0xc020U,0x904bU,
		0x309dU,0x613aU,0xc274U,0x94e3U,0x39cdU,0x739aU,0xe734U,0xde63U,0xaccdU,0x4991U,
		0x9322U,0x364fU,0x6c9eU,0xd93cU,0xa273U,0x54edU,0xa9daU,0x43bfU,0x877eU,0x1ef7U,
		0x3deeU,0x7bdcU,0xf7b8U,0xff7bU,0xeefdU,0xcdf1U,0x8be9U,0x07d9U,0x0fb2U,0x1f64U,
		0x3ec8U,0x7d90U,0xfb20U,0xe64bU,0xdc9dU,0xa931U,0x4269U,0x84d2U,0x19afU,0x335eU,
		0x66bcU,0xcd78U,0x8afbU,0x05fdU,0x0bfaU,0x17f4U,0x2fe8U,0x5fd0U,0xbfa0U,0x6f4bU,
		0xde96U,0xad27U,0x4a45U,0x948aU,0x391fU,0x723eU,0xe47cU,0xd8f3U,0xa1edU,0x53d1U,
		0xa7a2U,0x5f4fU,0xbe9eU,0x6d37U,0xda6eU,0xa4d7U,0x59a5U,0xb34aU,0x769fU,0xed3eU,
		0xca77U,0x84e5U,0x19c1U,0x3382U,0x6704U,0xce08U,0x8c1bU,0x083dU,0x107aU,0x20f4U,
		0x41e8U,0x83d0U,0x17abU,0x2f56U,0x5eacU,0xbd58U,0x6abbU,0xd576U,0xbae7U,0x65c5U,
		0xcb8aU,0x871fU,0x1e35U,0x3c6aU,0x78d4U,0xf1a8U,0xf35bU,0xf6bdU,0xfd71U,0xeae9U,
		0xc5d9U,0x9bb9U,0x2779U,0x4ef2U,0x9de4U,0x2bc3U,0x5786U,0xaf0cU,0x4e13U,0x9c26U,
		0x2847U,0x508eU,0xa11cU,0x5233U,0xa466U,0x58c7U,0xb18eU,0x7317U,0xe62eU,0xdc57U,
		0xa8a5U,0x4141U,0x8282U,0x150fU,0x2a1eU,0x543cU,0xa878U,0x40fbU,0x81f6U,0x13e7U,
		0x27ceU,0x4f9cU,0x9f38U,0x2e7bU,0x5cf6U,0xb9ecU,0x63d3U,0xc7a6U,0x9f47U,0x2e85U,
		0x5d0aU,0xba14U,0x6423U,0xc846U,0x8087U,0x1105U,0x220aU,0x4414U,0x8828U,0x005bU,
		0x00b6U,0x016cU,0x02d8U,0x05b0U,0x0b60U,0x16c0U,0x2d80U,0x5b00U,0xb600U,0x7c0bU,
		0xf816U,0xe027U,0xd045U,0xb081U,0x7109U,0xe212U,0xd42fU,0xb855U,0x60a1U,0xc142U,
		0x928fU,0x3515U,0x6a2aU,0xd454U,0xb8a3U,0x614dU,0xc29aU,0x953fU,0x3a75U,0x74eaU,
		0xe9d4U,0xc3a3U,0x974dU,0x3e91U,0x7d22U,0xfa44U,0xe483U,0xd90dU,0xa211U,0x5429U,
		0xa852U,0x40afU,0x815eU,0x12b7U,0x256eU,0x4adcU,0x95b8U,0x3b7bU,0x76f6U,0xedecU,
		0xcbd3U,0x87adU,0x1f51U,0x3ea2U,0x7d44U,0xfa88U,0xe51bU,0xda3dU,0xa471U,0x58e9U,
		0xb1d2U,0x73afU,0xe75eU,0xdeb7U,0xad65U,0x4ac1U,0x9582U,0x3b0fU,0x761eU,0xec3cU,
		0xc873U,0x80edU,0x11d1U,0x23a2U,0x4744U,0x8e88U,0x0d1bU,0x1a36U,0x346cU,0x68d8U,
		0xd1b0U,0xb36bU,0x76ddU,0xedbaU,0xcb7fU,0x86f5U,0x1de1U,0x3bc2U,0x7784U,0xef08U,
		0xce1bU,0x8c3dU,0x0871U,0x10e2U,0x21c4U,0x4388U,0x8710U,0x1e2bU,0x3c56U,0x78acU,
		0xf158U,0xf2bbU,0xf57dU,0xfaf1U,0xe5e9U,0xdbd9U,0xa7b9U,0x5f79U,0xbef2U,0x6defU,
		0xdbdeU,0xa7b7U,0x5f65U,0xbecaU,0x6d9fU,0xdb3eU,0xa677U,0x5ce5U,0xb9caU,0x639fU,
		0xc73eU,0x9e77U,0x2ce5U,0x59caU,0xb394U,0x7723U,0xee46U,0xcc87U,0x8905U,0x0201U,
		0x0402U,0x0804U,0x1008U,0x2010U,0x4020U,0x8040U,0x108bU,0x2116U,0x422cU,0x8458U,
		0x18bbU,0x3176U,0x62ecU,0xc5d8U,0x9bbbU,0x277dU,0x4efaU,0x9df4U,0x2be3U,0x57c6U,
		0xaf8cU,0x4f13U,0x9e26U,0x2c47U,0x588eU,0xb11cU,0x7233U,0xe466U,0xd8c7U,0xa185U,
		0x5301U,0xa602U,0x5c0fU,0xb81eU,0x6037U,0xc06eU,0x90d7U,0x31a5U,0x634aU,0xc694U,
		0x9d23U,0x2a4dU,0x549aU,0xa934U,0x4263U,0x84c6U,0x1987U,0x330eU,0x661cU,0xcc38U,
		0x887bU,0x00fdU,0x01faU,0x03f4U,0x07e8U,0x0fd0U,0x1fa0U,0x3f40U,0x7e80U,0xfd00U,
		0xea0bU,0xc41dU,0x9831U,0x2069U,0x40d2U,0x81a4U,0x1343U,0x2686U,0x4d0cU,0x9a18U,
		0x243bU,0x4876U,0x90ecU,0x31d3U,0x63a6U,0xc74cU,0x9e93U,0x2d2dU,0x5a5aU,0xb4b4U,
		0x7963U,0xf2c6U,0xf587U,0xfb05U,0xe601U,0xdc09U,0xa819U,0x4039U,0x8072U,0x10efU,
		0x21deU,0x43bcU,0x8778U,0x1efbU,0x3df6U,0x7becU,0xf7d8U,0xffbbU,0xef7dU,0xcef1U,
		0x8de9U,0x0bd9U,0x17b2U,0x2f64U,0x5ec8U,0xbd90U,0x6b2bU,0xd656U,0xbca7U,0x6945U,
		0xd28aU,0xb51fU,0x7a35U,0xf46aU,0xf8dfU,0xe1b5U,0xd361U,0xb6c9U,0x7d99U,0xfb32U,
		0xe66fU,0xdcd5U,0xa9a1U,0x4349U,0x8692U,0x1d2fU,0x3a5eU,0x74bcU,0xe978U,0xc2fbU,
		0x95fdU,0x3bf1U,0x77e2U,0xefc4U,0xcf83U,0x8f0dU,0x0e11U,0x1c22U,0x3844U,0x7088U,
		0xe110U,0xd22bU,0xb45dU,0x78b1U,0xf162U,0xf2cfU,0xf595U,0xfb21U,0xe649U,0xdc99U,
		0xa939U,0x4279U,0x84f2U,0x19efU,0x33deU,0x67bcU,0xcf78U,0x8efbU,0x0dfdU,0x1bfaU,
		0x37f4U,0x6fe8U,0xdfd0U,0xafabU,0x4f5dU,0x9ebaU,0x2d7fU,0x5afeU,0xb5fcU,0x7bf3U,
		0xf7e6U,0xffc7U,0xef85U,0xcf01U,0x8e09U,0x0c19U,0x1832U,0x3064U,0x60c8U,0xc190U,
		0x932bU,0x365dU,0x6cbaU,0xd974U,0xa2e3U,0x55cdU,0xab9aU,0x473fU,0x8e7eU,0x0cf7U,
		0x19eeU,0x33dcU,0x67b8U,0xcf70U,0x8eebU,0x0dddU,0x1bbaU,0x3774U,0x6ee8U,0xddd0U,
		0xababU,0x475dU,0x8ebaU,0x0d7fU,0x1afeU,0x35fcU,0x6bf8U,0xd7f0U,0xbfebU,0x6fddU,
		0xdfbaU,0xaf7fU,0x4ef5U,0x9deaU,0x2bdfU,0x57beU,0xaf7cU,0x4ef3U,0x9de6U,0x2bc7U,
		0x578eU,0xaf1cU,0x4e33U,0x9c66U,0x28c7U,0x518eU,0xa31cU,0x5633U,0xac66U,0x48c7U,
		0x918eU,0x3317U,0x662eU,0xcc5cU,0x88b3U,0x016dU,0x02daU,0x05b4U,0x0b68U,0x16d0U,
		0x2da0U,0x5b40U,0xb680U,0x7d0bU,0xfa16U,0xe427U,0xd845U,0xa081U,0x5109U,0xa212U,
		0x542fU,0xa85eU,0x40b7U,0x816eU,0x12d7U,0x25aeU,0x4b5cU,0x96b8U,0x3d7bU,0x7af6U,
		0xf5ecU,0xfbd3U,0xe7adU,0xdf51U,0xaea9U,0x4d59U,0x9ab2U,0x256fU,0x4adeU,0x95bcU,
		0x3b73U,0x76e6U,0xedccU,0xcb93U,0x872dU,0x1e51U,0x3ca2U,0x7944U,0xf288U,0xf51bU,
		0xfa3dU,0xe471U,0xd8e9U,0xa1d9U,0x53b9U,0xa772U,0x5eefU,0xbddeU,0x6bb7U,0xd76eU,
		0xbed7U,0x6da5U,0xdb4aU,0xa69fU,0x5d35U,0xba6aU,0x64dfU,0xc9beU,0x8377U,0x16e5U,
		0x2dcaU,0x5b94U,0xb728U,0x7e5bU,0xfcb6U,0xe967U,0xc2c5U,0x9581U,0x3b09U,0x7612U,
		0xec24U,0xc843U,0x808dU,0x1111U,0x2222U,0x4444U,0x8888U,0x011bU,0x0236U,0x046cU,
		0x08d8U,0x11b0U,0x2360U,0x46c0U,0x8d80U,0x0b0bU,0x1616U,0x2c2cU,0x5858U,0xb0b0U,
		0x716bU,0xe2d6U,0xd5a7U,0xbb45U,0x6681U,0xcd02U,0x8a0fU,0x0415U,0x082aU,0x1054U,
		0x20a8U,0x4150U,0x82a0U,0x154bU,0x2a96U,0x552cU,0xaa58U,0x44bbU,0x8976U,0x02e7U,
		0x05ceU,0x0b9cU,0x1738U,0x2e70U,0x5ce0U,0xb9c0U,0x638bU,0xc716U,0x9e27U,0x2c45U,
		0x588aU,0xb114U,0x7223U,0xe446U,0xd887U,0xa105U,0x5201U,0xa402U,0x580fU,0xb01eU,
		0x7037U,0xe06eU,0xd0d7U,0xb1a5U,0x7341U,0xe682U,0xdd0fU,0xaa15U,0x4421U,0x8842U,
		0x008fU,0x011eU,0x023cU,0x0478U,0x08f0U,0x11e0U,0x23c0U,0x4780U,0x8f00U,0x0e0bU,
		0x1c16U,0x382cU,0x7058U,0xe0b0U,0xd16bU,0xb2ddU,0x75b1U,0xeb62U,0xc6cfU,0x9d95U,
		0x2b21U,0x5642U,0xac84U,0x4903U,0x9206U,0x3407U,0x680eU,0xd01cU,0xb033U,0x706dU,
		0xe0daU,0xd1bfU,0xb375U,0x76e1U,0xedc2U,0xcb8fU,0x8715U,0x1e21U,0x3c42U,0x7884U,
		0xf108U,0xf21bU,0xf43dU,0xf871U,0xe0e9U,0xd1d9U,0xb3b9U,0x7779U,0xeef2U,0xcdefU,
		0x8bd5U,0x07a1U,0x0f42U,0x1e84U,0x3d08U,0x7a10U,0xf420U,0xf84bU,0xe09dU,0xd131U,
		0xb269U,0x74d9U,0xe9b2U,0xc36fU,0x96d5U,0x3da1U,0x7b42U,0xf684U,0xfd03U,0xea0dU,
		0xc411U,0x9829U,0x2059U,0x40b2U,0x8164U,0x12c3U,0x2586U,0x4b0cU,0x9618U,0x3c3bU,
		0x7876U,0xf0ecU,0xf1d3U,0xf3adU,0xf751U,0xfea9U,0xed59U,0xcab9U,0x8579U,0x1af9U,
		0x35f2U,0x6be4U,0xd7c8U,0xbf9bU,0x6f3dU,0xde7aU,0xacffU,0x49f5U,0x93eaU,0x37dfU,
		0x6fbeU,0xdf7cU,0xaef3U,0x4dedU,0x9bdaU,0x27bfU,0x4f7eU,0x9efcU,0x2df3U,0x5be6U,
		0xb7ccU,0x7f93U,0xff26U,0xee47U,0xcc85U,0x8901U,0x0209U,0x0412U,0x0824U,0x1048U,
		0x2090U,0x4120U,0x8240U,0x148bU,0x2916U,0x522cU,0xa458U,0x58bbU,0xb176U,0x72e7U,
		0xe5ceU,0xdb97U,0xa725U,0x5e41U,0xbc82U,0x690fU,0xd21eU,0xb437U,0x7865U,0xf0caU,
		0xf19fU,0xf335U,0xf661U,0xfcc9U,0xe999U,0xc339U,0x9679U,0x3cf9U,0x79f2U,0xf3e4U,
		0xf7c3U,0xff8dU,0xef11U,0xce29U,0x8c59U,0x08b9U,0x1172U,0x22e4U,0x45c8U,0x8b90U,
		0x072bU,0x0e56U,0x1cacU,0x3958U,0x72b0U,0xe560U,0xdacbU,0xa59dU,0x5b31U,0xb662U,
		0x7ccfU,0xf99eU,0xe337U,0xd665U,0xbcc1U,0x6989U,0xd312U,0xb62fU,0x7c55U,0xf8aaU,
		0xe15fU,0xd2b5U,0xb561U,0x7ac9U,0xf592U,0xfb2fU,0xe655U,0xdca1U,0xa949U,0x4299U,
		0x8532U,0x1a6fU,0x34deU,0x69bcU,0xd378U,0xb6fbU,0x7dfdU,0xfbfaU,0xe7ffU,0xdff5U,
		0xafe1U,0x4fc9U,0x9f92U,0x2f2fU,0x5e5eU,0xbcbcU,0x6973U,0xd2e6U,0xb5c7U,0x7b85U,
		0xf70aU,0xfe1fU,0xec35U,0xc861U,0x80c9U,0x1199U,0x2332U,0x4664U,0x8cc8U,0x099bU,
		0x1336U,0x266cU,0x4cd8U,0x99b0U,0x236bU,0x46d6U,0x8dacU,0x0b53U,0x16a6U,0x2d4cU,
		0x5a98U,0xb530U,0x7a6bU,0xf4d6U,0xf9a7U,0xe345U,0xd681U,0xbd09U,0x6a19U,0xd432U,
		0xb86fU,0x60d5U,0xc1aaU,0x935fU,0x36b5U,0x6d6aU,0xdad4U,0xa5a3U,0x5b4dU,0xb69aU,
		0x7d3fU,0xfa7eU,0xe4f7U,0xd9e5U,0xa3c1U,0x5789U,0xaf12U,0x4e2fU,0x9c5eU,0x28b7U,
		0x516eU,0xa2dcU,0x55b3U,0xab66U,0x46c7U,0x8d8eU,0x0b17U,0x162eU,0x2c5cU,0x58b8U,
		0xb170U,0x72ebU,0xe5d6U,0xdba7U,0xa745U,0x5e81U,0xbd02U,0x6a0fU,0xd41eU,0xb837U,
		0x6065U,0xc0caU,0x919fU,0x3335U,0x666aU,0xccd4U,0x89a3U,0x034dU,0x069aU,0x0d34U,
		0x1a68U,0x34d0U,0x69a0U,0xd340U,0xb68bU,0x7d1dU,0xfa3aU,0xe47fU,0xd8f5U,0xa1e1U,
		0x53c9U,0xa792U,0x5f2fU,0xbe5eU,0x6cb7U,0xd96eU,0xa2d7U,0x55a5U,0xab4aU,0x469fU,
		0x8d3eU,0x0a77U,0x14eeU,0x29dcU,0x53b8U,0xa770U,0x5eebU,0xbdd6U,0x6ba7U,0xd74eU,
		0xbe97U,0x6d25U,0xda4aU,0xa49fU,0x5935U,0xb26aU,0x74dfU,0xe9beU,0xc377U,0x96e5U,
		0x3dc1U,0x7b82U,0xf704U,0xfe03U,0xec0dU,0xc811U,0x8029U,0x1059U,0x20b2U,0x4164U,
		0x82c8U,0x159bU,0x2b36U,0x566cU,0xacd8U,0x49bbU,0x9376U,0x36e7U,0x6dceU,0xdb9cU,
		0xa733U,0x5e6dU,0xbcdaU,0x69bfU,0xd37eU,0xb6f7U,0x7de5U,0xfbcaU,0xe79fU,0xdf35U,
		0xae61U,0x4cc9U,0x9992U,0x232fU,0x465eU,0x8cbcU,0x0973U,0x12e6U,0x25ccU,0x4b98U,
		0x9730U,0x3e6bU,0x7cd6U,0xf9acU,0xe353U,0xd6adU,0xbd51U,0x6aa9U,0xd552U,0xbaafU,
		0x6555U,0xcaaaU,0x855fU,0x1ab5U,0x356aU,0x6ad4U,0xd5a8U,0xbb5bU,0x66bdU,0xcd7aU,
		0x8affU,0x05f5U,0x0beaU,0x17d4U,0x2fa8U,0x5f50U,0xbea0U,0x6d4bU,0xda96U,0xa527U,
		0x5a45U,0xb48aU,0x791fU,0xf23eU,0xf477U,0xf8e5U,0xe1c1U,0xd389U,0xb719U,0x7e39U,
		0xfc72U,0xe8efU,0xc1d5U,0x93a1U,0x3749U,0x6e92U,0xdd24U,0xaa43U,0x448dU,0x891aU,
		0x023fU,0x047eU,0x08fcU,0x11f8U,0x23f0U,0x47e0U,0x8fc0U,0x0f8bU,0x1f16U,0x3e2cU,
		0x7c58U,0xf8b0U,0xe16bU,0xd2ddU,0xb5b1U,0x7b69U,0xf6d2U,0xfdafU,0xeb55U,0xc6a1U,
		0x9d49U,0x2a99U,0x5532U,0xaa64U,0x44c3U,0x8986U,0x0307U,0x060eU,0x0c1cU,0x1838U,
		0x3070U,0x60e0U,0xc1c0U,0x938bU,0x371dU,0x6e3aU,0xdc74U,0xa8e3U,0x41cdU,0x839aU,
		0x173fU,0x2e7eU,0x5cfcU,0xb9f8U,0x63fbU,0xc7f6U,0x9fe7U,0x2fc5U,0x5f8aU,0xbf14U,
		0x6e23U,0xdc46U,0xa887U,0x4105U,0x820aU,0x141fU,0x283eU,0x507cU,0xa0f8U,0x51fbU,
		0xa3f6U,0x57e7U,0xafceU,0x4f97U,0x9f2eU,0x2e57U,0x5caeU,0xb95cU,0x62b3U,0xc566U,
		0x9ac7U,0x2585U,0x4b0aU,0x9614U,0x3c23U,0x7846U,0xf08cU,0xf113U,0xf22dU,0xf451U,
		0xf8a9U,0xe159U,0xd2b9U,0xb579U,0x7af9U,0xf5f2U,0xfbefU,0xe7d5U,0xdfa1U,0xaf49U,
		0x4e99U,0x9d32U,0x2a6fU,0x54deU,0xa9bcU,0x4373U,0x86e6U,0x1dc7U,0x3b8eU,0x771cU,
		0xee38U,0xcc7bU,0x88fdU,0x01f1U,0x03e2U,0x07c4U,0x0f88U,0x1f10U,0x3e20U,0x7c40U,
		0xf880U,0xe10bU,0xd21dU,0xb431U,0x7869U,0xf0d2U,0xf1afU,0xf355U,0xf6a1U,0xfd49U,
		0xea99U,0xc539U,0x9a79U,0x24f9U,0x49f2U,0x93e4U,0x37c3U,0x6f86U,0xdf0cU,0xae13U,
		0x4c2dU,0x985aU,0x20bfU,0x417eU,0x82fcU,0x15f3U,0x2be6U,0x57ccU,0xaf98U,0x4f3bU,
		0x9e76U,0x2ce7U,0x59ceU,0xb39cU,0x7733U,0xee66U,0xccc7U,0x8985U,0x0301U,0x0602U,
		0x0c04U,0x1808U,0x3010U,0x6020U,0xc040U,0x908bU,0x311dU,0x623aU,0xc474U,0x98e3U,
		0x21cdU,0x439aU,0x8734U,0x1e63U,0x3cc6U,0x798cU,0xf318U,0xf63bU,0xfc7dU,0xe8f1U,
		0xc1e9U,0x93d9U,0x37b9U,0x6f72U,0xdee4U,0xadc3U,0x4b8dU,0x971aU,0x3e3fU,0x7c7eU,
		0xf8fcU,0xe1f3U,0xd3edU,0xb7d1U,0x7fa9U,0xff52U,0xeeafU,0xcd55U,0x8aa1U,0x0549U,
		0x0a92U,0x1524U,0x2a48U,0x5490U,0xa920U,0x424bU,0x8496U,0x1927U,0x324eU,0x649cU,
		0xc938U,0x827bU,0x14fdU,0x29faU,0x53f4U,0xa7e8U,0x5fdbU,0xbfb6U,0x6f67U,0xdeceU,
		0xad97U,0x4b25U,0x964aU,0x3c9fU,0x793eU,0xf27cU,0xf4f3U,0xf9edU,0xe3d1U,0xd7a9U,
		0xbf59U,0x6eb9U,0xdd72U,0xaaefU,0x45d5U,0x8baaU,0x075fU,0x0ebeU,0x1d7cU,0x3af8U,
		0x75f0U,0xebe0U,0xc7cbU,0x9f9dU,0x2f31U,0x5e62U,0xbcc4U,0x6983U,0xd306U,0xb607U,
		0x7c05U,0xf80aU,0xe01fU,0xd035U,0xb061U,0x70c9U,0xe192U,0xd32fU,0xb655U,0x7ca1U,
		0xf942U,0xe28fU,0xd515U,0xba21U,0x6449U,0xc892U,0x812fU,0x1255U,0x24aaU,0x4954U,
		0x92a8U,0x355bU,0x6ab6U,0xd56cU,0xbad3U,0x65adU,0xcb5aU,0x86bfU,0x1d75U,0x3aeaU,
		0x75d4U,0xeba8U,0xc75bU,0x9ebdU,0x2d71U,0x5ae2U,0xb5c4U,0x7b83U,0xf706U,0xfe07U,
		0xec05U,0xc801U,0x8009U,0x1019U,0x2032U,0x4064U,0x80c8U,0x119bU,0x2336U,0x466cU,
		0x8cd8U,0x09bbU,0x1376U,0x26ecU,0x4dd8U,0x9bb0U,0x276bU,0x4ed6U,0x9dacU,0x2b53U,
		0x56a6U,0xad4cU,0x4a93U,0x9526U,0x3a47U,0x748eU,0xe91cU,0xc233U,0x946dU,0x38d1U,
		0x71a2U,0xe344U,0xd683U,0xbd0dU,0x6a11U,0xd422U,0xb84fU,0x6095U,0xc12aU,0x925fU,
		0x34b5U,0x696aU,0xd2d4U,0xb5a3U,0x7b4dU,0xf69aU,0xfd3fU,0xea75U,0xc4e1U,0x99c9U,
		0x2399U,0x4732U,0x8e64U,0x0cc3U,0x1986U,0x330cU,0x6618U,0xcc30U,0x886bU,0x00ddU,
		0x01baU,0x0374U,0x06e8U,0x0dd0U,0x1ba0U,0x3740U,0x6e80U,0xdd00U,0xaa0bU,0x441dU,
		0x883aU,0x007fU,0x00feU,0x01fcU,0x03f8U,0x07f0U,0x0fe0U,0x1fc0U,0x3f80U,0x7f00U,
		0xfe00U,0xec0bU,0xc81dU,0x8031U,0x1069U,0x20d2U,0x41a4U,0x8348U,0x169bU,0x2d36U,
		0x5a6cU,0xb4d8U,0x79bbU,0xf376U,0xf6e7U,0xfdc5U,0xeb81U,0xc709U,0x9e19U,0x2c39U,
		0x5872U,0xb0e4U,0x71c3U,0xe386U,0xd707U,0xbe05U,0x6c01U,0xd802U,0xa00fU,0x5015U,
		0xa02aU,0x505fU,0xa0beU,0x5177U,0xa2eeU,0x55d7U,0xabaeU,0x4757U,0x8eaeU,0x0d57U,
		0x1aaeU,0x355cU,0x6ab8U,0xd570U,0xbaebU,0x65ddU,0xcbbaU,0x877fU,0x1ef5U,0x3deaU,
		0x7bd4U,0xf7a8U,0xff5bU,0xeebdU,0xcd71U,0x8ae9U,0x05d9U,0x0bb2U,0x1764U,0x2ec8U,
		0x5d90U,0xbb20U,0x664bU,0xcc96U,0x8927U,0x0245U,0x048aU,0x0914U,0x1228U,0x2450U,
		0x48a0U,0x9140U,0x328bU,0x6516U,0xca2cU,0x8453U,0x18adU,0x315aU,0x62b4U,0xc568U,
		0x9adbU,0x25bdU,0x4b7aU,0x96f4U,0x3de3U,0x7bc6U,0xf78cU,0xff13U,0xee2dU,0xcc51U,
		0x88a9U,0x0159U,0x02b2U,0x0564U,0x0ac8U,0x1590U,0x2b20U,0x5640U,0xac80U,0x490bU,
		0x9216U,0x3427U,0x684eU,0xd09cU,0xb133U,0x726dU,0xe4daU,0xd9bfU,0xa375U,0x56e1U,
		0xadc2U,0x4b8fU,0x971eU,0x3e37U,0x7c6eU,0xf8dcU,0xe1b3U,0xd36dU,0xb6d1U,0x7da9U,
		0xfb52U,0xe6afU,0xdd55U,0xaaa1U,0x4549U,0x8a92U,0x052fU,0x0a5eU,0x14bcU,0x2978U,
		0x52f0U,0xa5e0U,0x5bcbU,0xb796U,0x7f27U,0xfe4eU,0xec97U,0xc925U,0x8241U,0x1489U,
		0x2912U,0x5224U,0xa448U,0x589bU,0xb136U,0x7267U,0xe4ceU,0xd997U,0xa325U,0x5641U,
		0xac82U,0x490fU,0x921eU,0x3437U,0x686eU,0xd0dcU,0xb1b3U,0x736dU,0xe6daU,0xddbfU,
		0xab75U,0x46e1U,0x8dc2U,0x0b8fU,0x171eU,0x2e3cU,0x5c78U,0xb8f0U,0x61ebU,0xc3d6U,
		0x97a7U,0x3f45U,0x7e8aU,0xfd14U,0xea23U,0xc44dU,0x9891U,0x2129U,0x4252U,0x84a4U,
		0x1943U,0x3286U,0x650cU,0xca18U,0x843bU,0x187dU,0x30faU,0x61f4U,0xc3e8U,0x97dbU,
		0x3fbdU,0x7f7aU,0xfef4U,0xede3U,0xcbcdU,0x8791U,0x1f29U,0x3e52U,0x7ca4U,0xf948U,
		0xe29bU,0xd53dU,0xba71U,0x64e9U,0xc9d2U,0x83afU,0x1755U,0x2eaaU,0x5d54U,0xbaa8U,
		0x655bU,0xcab6U,0x8567U,0x1ac5U,0x358aU,0x6b14U,0xd628U,0xbc5bU,0x68bdU,0xd17aU,
		0xb2ffU,0x75f5U,0xebeaU,0xc7dfU,0x9fb5U,0x2f61U,0x5ec2U,0xbd84U,0x6b03U,0xd606U,
		0xbc07U,0x6805U,0xd00aU,0xb01fU,0x7035U,0xe06aU,0xd0dfU,0xb1b5U,0x7361U,0xe6c2U,
		0xdd8fU,0xab15U,0x4621U,0x8c42U,0x088fU,0x111eU,0x223cU,0x4478U,0x88f0U,0x01ebU,
		0x03d6U,0x07acU,0x0f58U,0x1eb0U,0x3d60U,0x7ac0U,0xf580U,0xfb0bU,0xe61dU,0xdc31U,
		0xa869U,0x40d9U,0x81b2U,0x136fU,0x26deU,0x4dbcU,0x9b78U,0x26fbU,0x4df6U,0x9becU,
		0x27d3U,0x4fa6U,0x9f4cU,0x2e93U,0x5d26U,0xba4cU,0x6493U,0xc926U,0x8247U,0x1485U,
		0x290aU,0x5214U,0xa428U,0x585bU,0xb0b6U,0x7167U,0xe2ceU,0xd597U,0xbb25U,0x6641U,
		0xcc82U,0x890fU,0x0215U,0x042aU,0x0854U,0x10a8U,0x2150U,0x42a0U,0x8540U,0x1a8bU,
		0x3516U,0x6a2cU,0xd458U,0xb8bbU,0x617dU,0xc2faU,0x95ffU,0x3bf5U,0x77eaU,0xefd4U,
		0xcfa3U,0x8f4dU,0x0e91U,0x1d22U,0x3a44U,0x7488U,0xe910U,0xc22bU,0x945dU,0x38b1U,
		0x7162U,0xe2c4U,0xd583U,0xbb0dU,0x6611U,0xcc22U,0x884fU,0x0095U,0x012aU,0x0254U,
		0x04a8U,0x0950U,0x12a0U,0x2540U,0x4a80U,0x9500U,0x3a0bU,0x7416U,0xe82cU,0xc053U,
		0x90adU,0x3151U,0x62a2U,0xc544U,0x9a83U,0x250dU,0x4a1aU,0x9434U,0x3863U,0x70c6U,
		0xe18cU,0xd313U,0xb62dU,0x7c51U,0xf8a2U,0xe14fU,0xd295U,0xb521U,0x7a49U,0xf492U,
		0xf92fU,0xe255U,0xd4a1U,0xb949U,0x6299U,0xc532U,0x9a6fU,0x24d5U,0x49aaU,0x9354U,
		0x36a3U,0x6d46U,0xda8cU,0xa513U,0x5a2dU,0xb45aU,0x78bfU,0xf17eU,0xf2f7U,0xf5e5U,
		0xfbc1U,0xe789U,0xdf19U,0xae39U,0x4c79U,0x98f2U,0x21efU,0x43deU,0x87bcU,0x1f73U,
		0x3ee6U,0x7dccU,0xfb98U,0xe73bU,0xde7dU,0xacf1U,0x49e9U,0x93d2U,0x37afU,0x6f5eU,
		0xdebcU,0xad73U,0x4aedU,0x95daU,0x3bbfU,0x777eU,0xeefcU,0xcdf3U,0x8bedU,0x07d1U,
		0x0fa2U,0x1f44U,0x3e88U,0x7d10U,0xfa20U,0xe44bU,0xd89dU,0xa131U,0x5269U,0xa4d2U,
		0x59afU,0xb35eU,0x76b7U,0xed6eU,0xcad7U,0x85a5U,0x1b41U,0x3682U,0x6d04U,0xda08U,
		0xa41bU,0x583dU,0xb07aU,0x70ffU,0xe1feU,0xd3f7U,0xb7e5U,0x7fc1U,0xff82U,0xef0fU,
		0xce15U,0x8c21U,0x0849U,0x1092U,0x2124U,0x4248U,0x8490U,0x192bU,0x3256U,0x64acU,
		0xc958U,0x82bbU,0x157dU,0x2afaU,0x55f4U,0xabe8U,0x47dbU,0x8fb6U,0x0f67U,0x1eceU,
		0x3d9cU,0x7b38U,0xf670U,0xfcebU,0xe9ddU,0xc3b1U,0x9769U,0x3ed9U,0x7db2U,0xfb64U,
		0xe6c3U,0xdd8dU,0xab11U,0x4629U,0x8c52U,0x08afU,0x115eU,0x22bcU,0x4578U,0x8af0U,
		0x05ebU,0x0bd6U,0x17acU,0x2f58U,0x5eb0U,0xbd60U,0x6acbU,0xd596U,0xbb27U,0x6645U,
		0xcc8aU,0x891fU,0x0235U,0x046aU,0x08d4U,0x11a8U,0x2350U,0x46a0U,0x8d40U,0x0a8bU,
		0x1516U,0x2a2cU,0x5458U,0xa8b0U,0x416bU,0x82d6U,0x15a7U,0x2b4eU,0x569cU,0xad38U,
		0x4a7bU,0x94f6U,0x39e7U,0x73ceU,0xe79cU,0xdf33U,0xae6dU,0x4cd1U,0x99a2U,0x234fU,
		0x469eU,0x8d3cU,0x0a73U,0x14e6U,0x29ccU,0x5398U,0xa730U,0x5e6bU,0xbcd6U,0x69a7U,
		0xd34eU,0xb697U,0x7d25U,0xfa4aU,0xe49fU,0xd935U,0xa261U,0x54c9U,0xa992U,0x432fU,
		0x865eU,0x1cb7U,0x396eU,0x72dcU,0xe5b8U,0xdb7bU,0xa6fdU,0x5df1U,0xbbe2U,0x67cfU,
		0xcf9eU,0x8f37U,0x0e65U,0x1ccaU,0x3994U,0x7328U,0xe650U,0xdcabU,0xa95dU,0x42b1U,
		0x8562U,0x1acfU,0x359eU,0x6b3cU,0xd678U,0xbcfbU,0x69fdU,0xd3faU,0xb7ffU,0x7ff5U,
		0xffeaU,0xefdfU,0xcfb5U,0x8f61U,0x0ec9U,0x1d92U,0x3b24U,0x7648U,0xec90U,0xc92bU,
		0x825dU,0x14b1U,0x2962U,0x52c4U,0xa588U,0x5b1bU,0xb636U,0x7c67U,0xf8ceU,0xe197U,
		0xd325U,0xb641U,0x7c89U,0xf912U,0xe22fU,0xd455U,0xb8a1U,0x6149U,0xc292U,0x952fU,
		0x3a55U,0x74aaU,0xe954U,0xc2a3U,0x954dU,0x3a91U,0x7522U,0xea44U,0xc483U,0x990dU,
		0x2211U,0x4422U,0x8844U,0x0083U,0x0106U,0x020cU,0x0418U,0x0830U,0x1060U,0x20c0U,
		0x4180U,0x8300U,0x160bU,0x2c16U,0x582cU,0xb058U,0x70bbU,0xe176U,0xd2e7U,0xb5c5U,
		0x7b81U,0xf702U,0xfe0fU,0xec15U,0xc821U,0x8049U,0x1099U,0x2132U,0x4264U,0x84c8U,
		0x199bU,0x3336U,0x666cU,0xccd8U,0x89bbU,0x037dU,0x06faU,0x0df4U,0x1be8U,0x37d0U,
		0x6fa0U,0xdf40U,0xae8bU,0x4d1dU,0x9a3aU,0x247fU,0x48feU,0x91fcU,0x33f3U,0x67e6U,
		0xcfccU,0x8f93U,0x0f2dU,0x1e5aU,0x3cb4U,0x7968U,0xf2d0U,0xf5abU,0xfb5dU,0xe6b1U,
		0xdd69U,0xaad9U,0x45b9U,0x8b72U,0x06efU,0x0ddeU,0x1bbcU,0x3778U,0x6ef0U,0xdde0U,
		0xabcbU,0x479dU,0x8f3aU,0x0e7fU,0x1cfeU,0x39fcU,0x73f8U,0xe7f0U,0xdfebU,0xafddU,
		0x4fb1U,0x9f62U,0x2ecfU,0x5d9eU,0xbb3cU,0x6673U,0xcce6U,0x89c7U,0x0385U,0x070aU,
		0x0e14U,0x1c28U,0x3850U,0x70a0U,0xe140U,0xd28bU,0xb51dU,0x7a31U,0xf462U,0xf8cfU,
		0xe195U,0xd321U,0xb649U,0x7c99U,0xf932U,0xe26fU,0xd4d5U,0xb9a1U,0x6349U,0xc692U,
		0x9d2fU,0x2a55U,0x54aaU,0xa954U,0x42a3U,0x8546U,0x1a87U,0x350eU,0x6a1cU,0xd438U,
		0xb87bU,0x60fdU,0xc1faU,0x93ffU,0x37f5U,0x6feaU,0xdfd4U,0xafa3U,0x4f4dU,0x9e9aU,
		0x2d3fU,0x5a7eU,0xb4fcU,0x79f3U,0xf3e6U,0xf7c7U,0xff85U,0xef01U,0xce09U,0x8c19U,
		0x0839U,0x1072U,0x20e4U,0x41c8U,0x8390U,0x172bU,0x2e56U,0x5cacU,0xb958U,0x62bbU,
		0xc576U,0x9ae7U,0x25c5U,0x4b8aU,0x9714U,0x3e23U,0x7c46U,0xf88cU,0xe113U,0xd22dU,
		0xb451U,0x78a9U,0xf152U,0xf2afU,0xf555U,0xfaa1U,0xe549U,0xda99U,0xa539U,0x5a79U,
		0xb4f2U,0x79efU,0xf3deU,0xf7b7U,0xff65U,0xeec1U,0xcd89U,0x8b19U,0x0639U,0x0c72U,
		0x18e4U,0x31c8U,0x6390U,0xc720U,0x9e4bU,0x2c9dU,0x593aU,0xb274U,0x74e3U,0xe9c6U,
		0xc387U,0x9705U,0x3e01U,0x7c02U,0xf804U,0xe003U,0xd00dU,0xb011U,0x7029U,0xe052U,
		0xd0afU,0xb155U,0x72a1U,0xe542U,0xda8fU,0xa515U,0x5a21U,0xb442U,0x788fU,0xf11eU,
		0xf237U,0xf465U,0xf8c1U,0xe189U,0xd319U,0xb639U,0x7c79U,0xf8f2U,0xe1efU,0xd3d5U,
		0xb7a1U,0x7f49U,0xfe92U,0xed2fU,0xca55U,0x84a1U,0x1949U,0x3292U,0x6524U,0xca48U,
		0x849bU,0x193dU,0x327aU,0x64f4U,0xc9e8U,0x83dbU,0x17bdU,0x2f7aU,0x5ef4U,0xbde8U,
		0x6bdbU,0xd7b6U,0xbf67U,0x6ec5U,0xdd8aU,0xab1fU,0x4635U,0x8c6aU,0x08dfU,0x11beU,
		0x237cU,0x46f8U,0x8df0U,0x0bebU,0x17d6U,0x2facU,0x5f58U,0xbeb0U,0x6d6bU,0xdad6U,
		0xa5a7U,0x5b45U,0xb68aU,0x7d1fU,0xfa3eU,0xe477U,0xd8e5U,0xa1c1U,0x5389U,0xa712U,
		0x5e2fU,0xbc5eU,0x68b7U,0xd16eU,0xb2d7U,0x75a5U,0xeb4aU,0xc69fU,0x9d35U,0x2a61U,
		0x54c2U,0xa984U,0x4303U,0x8606U,0x1c07U,0x380eU,0x701cU,0xe038U,0xd07bU,0xb0fdU,
		0x71f1U,0xe3e2U,0xd7cfU,0xbf95U,0x6f21U,0xde42U,0xac8fU,0x4915U,0x922aU,0x345fU,
		0x68beU,0xd17cU,0xb2f3U,0x75edU,0xebdaU,0xc7bfU,0x9f75U,0x2ee1U,0x5dc2U,0xbb84U,
		0x6703U,0xce06U,0x8c07U,0x0805U,0x100aU,0x2014U,0x4028U,0x8050U,0x10abU,0x2156U,
		0x42acU,0x8558U,0x1abbU,0x3576U,0x6aecU,0xd5d8U,0xbbbbU,0x677dU,0xcefaU,0x8dffU,
		0x0bf5U,0x17eaU,0x2fd4U,0x5fa8U,0xbf50U,0x6eabU,0xdd56U,0xaaa7U,0x4545U,0x8a8aU,
		0x051fU,0x0a3eU,0x147cU,0x28f8U,0x51f0U,0xa3e0U,0x57cbU,0xaf96U,0x4f27U,0x9e4eU,
		0x2c97U,0x592eU,0xb25cU,0x74b3U,0xe966U,0xc2c7U,0x9585U,0x3b01U,0x7602U,0xec04U,
		0xc803U,0x800dU,0x1011U,0x2022U,0x4044U,0x8088U,0x111bU,0x2236U,0x446cU,0x88d8U,
		0x01bbU,0x0376U,0x06ecU,0x0dd8U,0x1bb0U,0x3760U,0x6ec0U,0xdd80U,0xab0bU,0x461dU,
		0x8c3aU,0x087fU,0x10feU,0x21fcU,0x43f8U,0x87f0U,0x1febU,0x3fd6U,0x7facU,0xff58U,
		0xeebbU,0xcd7dU,0x8af1U,0x05e9U,0x0bd2U,0x17a4U,0x2f48U,0x5e90U,0xbd20U,0x6a4bU,
		0xd496U,0xb927U,0x6245U,0xc48aU,0x991fU,0x2235U,0x446aU,0x88d4U,0x01a3U,0x0346U,
		0x068cU,0x0d18U,0x1a30U,0x3460U,0x68c0U,0xd180U,0xb30bU,0x761dU,0xec3aU,0xc87fU,
		0x80f5U,0x11e1U,0x23c2U,0x4784U,0x8f08U,0x0e1bU,0x1c36U,0x386cU,0x70d8U,0xe1b0U,
		0xd36bU,0xb6ddU,0x7db1U,0xfb62U,0xe6cfU,0xdd95U,0xab21U,0x4649U,0x8c92U,0x092fU,
		0x125eU,0x24bcU,0x4978U,0x92f0U,0x35ebU,0x6bd6U,0xd7acU,0xbf53U,0x6eadU,0xdd5aU,
		0xaabfU,0x4575U,0x8aeaU,0x05dfU,0x0bbeU,0x177cU,0x2ef8U,0x5df0U,0xbbe0U,0x67cbU,
		0xcf96U,0x8f27U,0x0e45U,0x1c8aU,0x3914U,0x7228U,0xe450U,0xd8abU,0xa15dU,0x52b1U,
		0xa562U,0x5acfU,0xb59eU,0x7b37U,0xf66eU,0xfcd7U,0xe9a5U,0xc341U,0x9689U,0x3d19U,
		0x7a32U,0xf464U,0xf8c3U,0xe18dU,0xd311U,0xb629U,0x7c59U,0xf8b2U,0xe16fU,0xd2d5U,
		0xb5a1U,0x7b49U,0xf692U,0xfd2fU,0xea55U,0xc4a1U,0x9949U,0x2299U,0x4532U,0x8a64U,
		0x04c3U,0x0986U,0x130cU,0x2618U,0x4c30U,0x9860U,0x20cbU,0x4196U,0x832cU,0x1653U,
		0x2ca6U,0x594cU,0xb298U,0x753bU,0xea76U,0xc4e7U,0x99c5U,0x2381U,0x4702U,0x8e04U,
		0x0c03U,0x1806U,0x300cU,0x6018U,0xc030U,0x906bU,0x30ddU,0x61baU,0xc374U,0x96e3U,
		0x3dcdU,0x7b9aU,0xf734U,0xfe63U,0xeccdU,0xc991U,0x8329U,0x1659U,0x2cb2U,0x5964U,
		0xb2c8U,0x759bU,0xeb36U,0xc667U,0x9cc5U,0x2981U,0x5302U,0xa604U,0x5c03U,0xb806U,
		0x6007U,0xc00eU,0x9017U,0x3025U,0x604aU,0xc094U,0x9123U,0x324dU,0x649aU,0xc934U,
		0x8263U,0x14cdU,0x299aU,0x5334U,0xa668U,0x5cdbU,0xb9b6U,0x6367U,0xc6ceU,0x9d97U,
		0x2b25U,0x564aU,0xac94U,0x4923U,0x9246U,0x3487U,0x690eU,0xd21cU,0xb433U,0x786dU,
		0xf0daU,0xf1bfU,0xf375U,0xf6e1U,0xfdc9U,0xeb99U,0xc739U,0x9e79U,0x2cf9U,0x59f2U,
		0xb3e4U,0x77c3U,0xef86U,0xcf07U,0x8e05U,0x0c01U,0x1802U,0x3004U,0x6008U,0xc010U,
		0x902bU,0x305dU,0x60baU,0xc174U,0x92e3U,0x35cdU,0x6b9aU,0xd734U,0xbe63U,0x6ccdU,
		0xd99aU,0xa33fU,0x5675U,0xaceaU,0x49dfU,0x93beU,0x3777U,0x6eeeU,0xdddcU,0xabb3U,
		0x476dU,0x8edaU,0x0dbfU,0x1b7eU,0x36fcU,0x6df8U,0xdbf0U,0xa7ebU,0x5fddU,0xbfbaU,
		0x6f7fU,0xdefeU,0xadf7U,0x4be5U,0x97caU,0x3f9fU,0x7f3eU,0xfe7cU,0xecf3U,0xc9edU,
		0x83d1U,0x17a9U,0x2f52U,0x5ea4U,0xbd48U,0x6a9bU,0xd536U,0xba67U,0x64c5U,0xc98aU,
		0x831fU,0x1635U,0x2c6aU,0x58d4U,0xb1a8U,0x735bU,0xe6b6U,0xdd67U,0xaac5U,0x4581U,
		0x8b02U,0x060fU,0x0c1eU,0x183cU,0x3078U,0x60f0U,0xc1e0U,0x93cbU,0x379dU,0x6f3aU,
		0xde74U,0xace3U,0x49cdU,0x939aU,0x373fU,0x6e7eU,0xdcfcU,0xa9f3U,0x43edU,0x87daU,
		0x1fbfU,0x3f7eU,0x7efcU,0xfdf8U,0xebfbU,0xc7fdU,0x9ff1U,0x2fe9U,0x5fd2U,0xbfa4U,
		0x6f43U,0xde86U,0xad07U,0x4a05U,0x940aU,0x381fU,0x703eU,0xe07cU,0xd0f3U,0xb1edU,
		0x73d1U,0xe7a2U,0xdf4fU,0xae95U,0x4d21U,0x9a42U,0x248fU,0x491eU,0x923cU,0x3473U,
		0x68e6U,0xd1ccU,0xb393U,0x772dU,0xee5aU,0xccbfU,0x8975U,0x02e1U,0x05c2U,0x0b84U,
		0x1708U,0x2e10U,0x5c20U,0xb840U,0x608bU,0xc116U,0x9227U,0x3445U,0x688aU,0xd114U,
		0xb223U,0x744dU,0xe89aU,0xc13fU,0x9275U,0x34e1U,0x69c2U,0xd384U,0xb703U,0x7e0dU,
		0xfc1aU,0xe83fU,0xc075U,0x90e1U,0x31c9U,0x6392U,0xc724U,0x9e43U,0x2c8dU,0x591aU,
		0xb234U,0x7463U,0xe8c6U,0xc187U,0x9305U,0x3601U,0x6c02U,0xd804U,0xa003U,0x500dU,
		0xa01aU,0x503fU,0xa07eU,0x50f7U,0xa1eeU,0x53d7U,0xa7aeU,0x5f57U,0xbeaeU,0x6d57U,
		0xdaaeU,0xa557U,0x5aa5U,0xb54aU,0x7a9fU,0xf53eU,0xfa77U,0xe4e5U,0xd9c1U,0xa389U,
		0x5719U,0xae32U,0x4c6fU,0x98deU,0x21b7U,0x436eU,0x86dcU,0x1db3U,0x3b66U,0x76ccU,
		0xed98U,0xcb3bU,0x867dU,0x1cf1U,0x39e2U,0x73c4U,0xe788U,0xdf1bU,0xae3dU,0x4c71U,
		0x98e2U,0x21cfU,0x439eU,0x873cU,0x1e73U,0x3ce6U,0x79ccU,0xf398U,0xf73bU,0xfe7dU,
		0xecf1U,0xc9e9U,0x83d9U,0x17b9U,0x2f72U,0x5ee4U,0xbdc8U,0x6b9bU,0xd736U,0xbe67U,
		0x6cc5U,0xd98aU,0xa31fU,0x5635U,0xac6aU,0x48dfU,0x91beU,0x3377U,0x66eeU,0xcddcU,
		0x8bb3U,0x076dU,0x0edaU,0x1db4U,0x3b68U,0x76d0U,0xeda0U,0xcb4bU,0x869dU,0x1d31U,
		0x3a62U,0x74c4U,0xe988U,0xc31bU,0x963dU,0x3c71U,0x78e2U,0xf1c4U,0xf383U,0xf70dU,
		0xfe11U,0xec29U,0xc859U,0x80b9U,0x1179U,0x22f2U,0x45e4U,0x8bc8U,0x079bU,0x0f36U,
		0x1e6cU,0x3cd8U,0x79b0U,0xf360U,0xf6cbU,0xfd9dU,0xeb31U,0xc669U,0x9cd9U,0x29b9U,
		0x5372U,0xa6e4U,0x5dc3U,0xbb86U,0x6707U,0xce0eU,0x8c17U,0x0825U,0x104aU,0x2094U,
		0x4128U,0x8250U,0x14abU,0x2956U,0x52acU,0xa558U,0x5abbU,0xb576U,0x7ae7U,0xf5ceU,
		0xfb97U,0xe725U,0xde41U,0xac89U,0x4919U,0x9232U,0x346fU,0x68deU,0xd1bcU,0xb373U,
		0x76edU,0xeddaU,0xcbbfU,0x8775U,0x1ee1U,0x3dc2U,0x7b84U,0xf708U,0xfe1bU,0xec3dU,
		0xc871U,0x80e9U,0x11d9U,0x23b2U,0x4764U,0x8ec8U,0x0d9bU,0x1b36U,0x366cU,0x6cd8U,
		0xd9b0U,0xa36bU,0x56ddU,0xadbaU,0x4b7fU,0x96feU,0x3df7U,0x7beeU,0xf7dcU,0xffb3U,
		0xef6dU,0xced1U,0x8da9U,0x0b59U,0x16b2U,0x2d64U,0x5ac8U,0xb590U,0x7b2bU,0xf656U,
		0xfca7U,0xe945U,0xc281U,0x9509U,0x3a19U,0x7432U,0xe864U,0xc0c3U,0x918dU,0x3311U,
		0x6622U,0xcc44U,0x8883U,0x010dU,0x021aU,0x0434U,0x0868U,0x10d0U,0x21a0U,0x4340U,
		0x8680U,0x1d0bU,0x3a16U,0x742cU,0xe858U,0xc0bbU,0x917dU,0x32f1U,0x65e2U,0xcbc4U,
		0x8783U,0x1f0dU,0x3e1aU,0x7c34U,0xf868U,0xe0dbU,0xd1bdU,0xb371U,0x76e9U,0xedd2U,
		0xcbafU,0x8755U,0x1ea1U,0x3d42U,0x7a84U,0xf508U,0xfa1bU,0xe43dU,0xd871U,0xa0e9U,
		0x51d9U,0xa3b2U,0x576fU,0xaedeU,0x4db7U,0x9b6eU,0x26d7U,0x4daeU,0x9b5cU,0x26b3U,
		0x4d66U,0x9accU,0x2593U,0x4b26U,0x964cU,0x3c93U,0x7926U,0xf24cU,0xf493U,0xf92dU,
		0xe251U,0xd4a9U,0xb959U,0x62b9U,0xc572U,0x9aefU,0x25d5U,0x4baaU,0x9754U,0x3ea3U,
		0x7d46U,0xfa8cU,0xe513U,0xda2dU,0xa451U,0x58a9U,0xb152U,0x72afU,0xe55eU,0xdab7U,
		0xa565U,0x5ac1U,0xb582U,0x7b0fU,0xf61eU,0xfc37U,0xe865U,0xc0c1U,0x9189U,0x3319U,
		0x6632U,0xcc64U,0x88c3U,0x018dU,0x031aU,0x0634U,0x0c68U,0x18d0U,0x31a0U,0x6340U,
		0xc680U,0x9d0bU,0x2a1dU,0x543aU,0xa874U,0x40e3U,0x81c6U,0x1387U,0x270eU,0x4e1cU,
		0x9c38U,0x287bU,0x50f6U,0xa1ecU,0x53d3U,0xa7a6U,0x5f47U,0xbe8eU,0x6d17U,0xda2eU,
		0xa457U,0x58a5U,0xb14aU,0x729fU,0xe53eU,0xda77U,0xa4e5U,0x59c1U,0xb382U,0x770fU,
		0xee1eU,0xcc37U,0x8865U,0x00c1U,0x0182U,0x0304U,0x0608U,0x0c10U,0x1820U,0x3040U,
		0x6080U,0xc100U,0x920bU,0x341dU,0x683aU,0xd074U,0xb0e3U,0x71cdU,0xe39aU,0xd73fU,
		0xbe75U,0x6ce1U,0xd9c2U,0xa38fU,0x5715U,0xae2aU,0x4c5fU,0x98beU,0x2177U,0x42eeU,
		0x85dcU,0x1bb3U,0x3766U,0x6eccU,0xdd98U,0xab3bU,0x467dU,0x8cfaU,0x09ffU,0x13feU,
		0x27fcU,0x4ff8U,0x9ff0U,0x2febU,0x5fd6U,0xbfacU,0x6f53U,0xdea6U,0xad47U,0x4a85U,
		0x950aU,0x3a1fU,0x743eU,0xe87cU,0xc0f3U,0x91edU,0x33d1U,0x67a2U,0xcf44U,0x8e83U,
		0x0d0dU,0x1a1aU,0x3434U,0x6868U,0xd0d0U,0xb1abU,0x735dU,0xe6baU,0xdd7fU,0xaaf5U,
		0x45e1U,0x8bc2U,0x078fU,0x0f1eU,0x1e3cU,0x3c78U,0x78f0U,0xf1e0U,0xf3cbU,0xf79dU,
		0xff31U,0xee69U,0xccd9U,0x89b9U,0x0379U,0x06f2U,0x0de4U,0x1bc8U,0x3790U,0x6f20U,
		0xde40U,0xac8bU,0x491dU,0x923aU,0x347fU,0x68feU,0xd1fcU,0xb3f3U,0x77edU,0xefdaU,
		0xcfbfU,0x8f75U,0x0ee1U,0x1dc2U,0x3b84U,0x7708U,0xee10U,0xcc2bU,0x885dU,0x00b1U,
		0x0162U,0x02c4U,0x0588U,0x0b10U,0x1620U,0x2c40U,0x5880U,0xb100U,0x720bU,0xe416U,
		0xd827U,0xa045U,0x5081U,0xa102U,0x520fU,0xa41eU,0x5837U,0xb06eU,0x70d7U,0xe1aeU,
		0xd357U,0xb6a5U,0x7d41U,0xfa82U,0xe50fU,0xda15U,0xa421U,0x5849U,0xb092U,0x712fU,
		0xe25eU,0xd4b7U,0xb965U,0x62c1U,0xc582U,0x9b0fU,0x2615U,0x4c2aU,0x9854U,0x20a3U,
		0x4146U,0x828cU,0x1513U,0x2a26U,0x544cU,0xa898U,0x413bU,0x8276U,0x14e7U,0x29ceU,
		0x539cU,0xa738U,0x5e7bU,0xbcf6U,0x69e7U,0xd3ceU,0xb797U,0x7f25U,0xfe4aU,0xec9fU,
		0xc935U,0x8261U,0x14c9U,0x2992U,0x5324U,0xa648U,0x5c9bU,0xb936U,0x6267U,0xc4ceU,
		0x9997U,0x2325U,0x464aU,0x8c94U,0x0923U,0x1246U,0x248cU,0x4918U,0x9230U,0x346bU,
		0x68d6U,0xd1acU,0xb353U,0x76adU,0xed5aU,0xcabfU,0x8575U,0x1ae1U,0x35c2U,0x6b84U,
		0xd708U,0xbe1bU,0x6c3dU,0xd87aU,0xa0ffU,0x51f5U,0xa3eaU,0x57dfU,0xafbeU,0x4f77U,
		0x9eeeU,0x2dd7U,0x5baeU,0xb75cU,0x7eb3U,0xfd66U,0xeac7U,0xc585U,0x9b01U,0x2609U,
		0x4c12U,0x9824U,0x2043U,0x4086U,0x810cU,0x1213U,0x2426U,0x484cU,0x9098U,0x313bU,
		0x6276U,0xc4ecU,0x99d3U,0x23adU,0x475aU,0x8eb4U,0x0d63U,0x1ac6U,0x358cU,0x6b18U,
		0xd630U,0xbc6bU,0x68ddU,0xd1baU,0xb37fU,0x76f5U,0xedeaU,0xcbdfU,0x87b5U,0x1f61U,
		0x3ec2U,0x7d84U,0xfb08U,0xe61bU,0xdc3dU,0xa871U,0x40e9U,0x81d2U,0x13afU,0x275eU,
		0x4ebcU,0x9d78U,0x2afbU,0x55f6U,0xabecU,0x47d3U,0x8fa6U,0x0f47U,0x1e8eU,0x3d1cU,
		0x7a38U,0xf470U,0xf8ebU,0xe1ddU,0xd3b1U,0xb769U,0x7ed9U,0xfdb2U,0xeb6fU,0xc6d5U,
		0x9da1U,0x2b49U,0x5692U,0xad24U,0x4a43U,0x9486U,0x3907U,0x720eU,0xe41cU,0xd833U,
		0xa06dU,0x50d1U,0xa1a2U,0x534fU,0xa69eU,0x5d37U,0xba6eU,0x64d7U,0xc9aeU,0x8357U,
		0x16a5U,0x2d4aU,0x5a94U,0xb528U,0x7a5bU,0xf4b6U,0xf967U,0xe2c5U,0xd581U,0xbb09U,
		0x6619U,0xcc32U,0x886fU,0x00d5U,0x01aaU,0x0354U,0x06a8U,0x0d50U,0x1aa0U,0x3540U,
		0x6a80U,0xd500U,0xba0bU,0x641dU,0xc83aU,0x807fU,0x10f5U,0x21eaU,0x43d4U,0x87a8U,
		0x1f5bU,0x3eb6U,0x7d6cU,0xfad8U,0xe5bbU,0xdb7dU,0xa6f1U,0x5de9U,0xbbd2U,0x67afU,
		0xcf5eU,0x8eb7U,0x0d65U,0x1acaU,0x3594U,0x6b28U,0xd650U,0xbcabU,0x695dU,0xd2baU,
		0xb57fU,0x7af5U,0xf5eaU,0xfbdfU,0xe7b5U,0xdf61U,0xaec9U,0x4d99U,0x9b32U,0x266fU,
		0x4cdeU,0x99bcU,0x2373U,0x46e6U,0x8dccU,0x0b93U,0x1726U,0x2e4cU,0x5c98U,0xb930U,
		0x626bU,0xc4d6U,0x99a7U,0x2345U,0x468aU,0x8d14U,0x0a23U,0x1446U,0x288cU,0x5118U,
		0xa230U,0x546bU,0xa8d6U,0x41a7U,0x834eU,0x1697U,0x2d2eU,0x5a5cU,0xb4b8U,0x797bU,
		0xf2f6U,0xf5e7U,0xfbc5U,0xe781U,0xdf09U,0xae19U,0x4c39U,0x9872U,0x20efU,0x41deU,
		0x83bcU,0x1773U,0x2ee6U,0x5dccU,0xbb98U,0x673bU,0xce76U,0x8ce7U,0x09c5U,0x138aU,
		0x2714U,0x4e28U,0x9c50U,0x28abU,0x5156U,0xa2acU,0x5553U,0xaaa6U,0x4547U,0x8a8eU,
		0x0517U,0x0a2eU,0x145cU,0x28b8U,0x5170U,0xa2e0U,0x55cbU,0xab96U,0x4727U,0x8e4eU,
		0x0c97U,0x192eU,0x325cU,0x64b8U,0xc970U,0x82ebU,0x15ddU,0x2bbaU,0x5774U,0xaee8U,
		0x4ddbU,0x9bb6U,0x2767U,0x4eceU,0x9d9cU,0x2b33U,0x5666U,0xacccU,0x4993U,0x9326U,
		0x3647U,0x6c8eU,0xd91cU,0xa233U,0x546dU,0xa8daU,0x41bfU,0x837eU,0x16f7U,0x2deeU,
		0x5bdcU,0xb7b8U,0x7f7bU,0xfef6U,0xede7U,0xcbc5U,0x8781U,0x1f09U,0x3e12U,0x7c24U,
		0xf848U,0xe09bU,0xd13dU,0xb271U,0x74e9U,0xe9d2U,0xc3afU,0x9755U,0x3ea1U,0x7d42U,
		0xfa84U,0xe503U,0xda0dU,0xa411U,0x5829U,0xb052U,0x70afU,0xe15eU,0xd2b7U,0xb565U,
		0x7ac1U,0xf582U,0xfb0fU,0xe615U,0xdc21U,0xa849U,0x4099U,0x8132U,0x126fU,0x24deU,
		0x49bcU,0x9378U,0x36fbU,0x6df6U,0xdbecU,0xa7d3U,0x5fadU,0xbf5aU,0x6ebfU,0xdd7eU,
		0xaaf7U,0x45e5U,0x8bcaU,0x079fU,0x0f3eU,0x1e7cU,0x3cf8U,0x79f0U,0xf3e0U,0xf7cbU,
		0xff9dU,0xef31U,0xce69U,0x8cd9U,0x09b9U,0x1372U,0x26e4U,0x4dc8U,0x9b90U,0x272bU,
		0x4e56U,0x9cacU,0x2953U,0x52a6U,0xa54cU,0x5a93U,0xb526U,0x7a47U,0xf48eU,0xf917U,
		0xe225U,0xd441U,0xb889U,0x6119U,0xc232U,0x946fU,0x38d5U,0x71aaU,0xe354U,0xd6a3U,
		0xbd4dU,0x6a91U,0xd522U,0xba4fU,0x6495U,0xc92aU,0x825fU,0x14b5U,0x296aU,0x52d4U,
		0xa5a8U,0x5b5bU,0xb6b6U,0x7d67U,0xfaceU,0xe597U,0xdb25U,0xa641U,0x5c89U,0xb912U,
		0x622fU,0xc45eU,0x98b7U,0x2165U,0x42caU,0x8594U,0x1b23U,0x3646U,0x6c8cU,0xd918U,
		0xa23bU,0x547dU,0xa8faU,0x41ffU,0x83feU,0x17f7U,0x2feeU,0x5fdcU,0xbfb8U,0x6f7bU,
		0xdef6U,0xade7U,0x4bc5U,0x978aU,0x3f1fU,0x7e3eU,0xfc7cU,0xe8f3U,0xc1edU,0x93d1U,
		0x37a9U,0x6f52U,0xdea4U,0xad43U,0x4a8dU,0x951aU,0x3a3fU,0x747eU,0xe8fcU,0xc1f3U,
		0x93edU,0x37d1U,0x6fa2U,0xdf44U,0xae83U,0x4d0dU,0x9a1aU,0x243fU,0x487eU,0x90fcU,
		0x31f3U,0x63e6U,0xc7ccU,0x9f93U,0x2f2dU,0x5e5aU,0xbcb4U,0x6963U,0xd2c6U,0xb587U,
		0x7b05U,0xf60aU,0xfc1fU,0xe835U,0xc061U,0x90c9U,0x3199U,0x6332U,0xc664U,0x9cc3U,
		0x298dU,0x531aU,0xa634U,0x5c63U,0xb8c6U,0x6187U,0xc30eU,0x9617U,0x3c25U,0x784aU,
		0xf094U,0xf123U,0xf24dU,0xf491U,0xf929U,0xe259U,0xd4b9U,0xb979U,0x62f9U,0xc5f2U,
		0x9befU,0x27d5U,0x4faaU,0x9f54U,0x2ea3U,0x5d46U,0xba8cU,0x6513U,0xca26U,0x8447U,
		0x1885U,0x310aU,0x6214U,0xc428U,0x985bU,0x20bdU,0x417aU,0x82f4U,0x15e3U,0x2bc6U,
		0x578cU,0xaf18U,0x4e3bU,0x9c76U,0x28e7U,0x51ceU,0xa39cU,0x5733U,0xae66U,0x4cc7U,
		0x998eU,0x2317U,0x462eU,0x8c5cU,0x08b3U,0x1166U,0x22ccU,0x4598U,0x8b30U,0x066bU,
		0x0cd6U,0x19acU,0x3358U,0x66b0U,0xcd60U,0x8acbU,0x059dU,0x0b3aU,0x1674U,0x2ce8U,
		0x59d0U,0xb3a0U,0x774bU,0xee96U,0xcd27U,0x8a45U,0x0481U,0x0902U,0x1204U,0x2408U,
		0x4810U,0x9020U,0x304bU,0x6096U,0xc12cU,0x9253U,0x34adU,0x695aU,0xd2b4U,0xb563U,
		0x7acdU,0xf59aU,0xfb3fU,0xe675U,0xdce1U,0xa9c9U,0x4399U,0x8732U,0x1e6fU,0x3cdeU,
		0x79bcU,0xf378U,0xf6fbU,0xfdfdU,0xebf1U,0xc7e9U,0x9fd9U,0x2fb9U,0x5f72U,0xbee4U,
		0x6dc3U,0xdb86U,0xa707U,0x5e05U,0xbc0aU,0x681fU,0xd03eU,0xb077U,0x70e5U,0xe1caU,
		0xd39fU,0xb735U,0x7e61U,0xfcc2U,0xe98fU,0xc315U,0x9621U,0x3c49U,0x7892U,0xf124U,
		0xf243U,0xf48dU,0xf911U,0xe229U,0xd459U,0xb8b9U,0x6179U,0xc2f2U,0x95efU,0x3bd5U,
		0x77aaU,0xef54U,0xcea3U,0x8d4dU,0x0a91U,0x1522U,0x2a44U,0x5488U,0xa910U,0x422bU,
		0x8456U,0x18a7U,0x314eU,0x629cU,0xc538U,0x9a7bU,0x24fdU,0x49faU,0x93f4U,0x37e3U,
		0x6fc6U,0xdf8cU,0xaf13U,0x4e2dU,0x9c5aU,0x28bfU,0x517eU,0xa2fcU,0x55f3U,0xabe6U,
		0x47c7U,0x8f8eU,0x0f17U,0x1e2eU,0x3c5cU,0x78b8U,0xf170U,0xf2ebU,0xf5ddU,0xfbb1U,
		0xe769U,0xded9U,0xadb9U,0x4b79U,0x96f2U,0x3defU,0x7bdeU,0xf7bcU,0xff73U,0xeeedU,
		0xcdd1U,0x8ba9U,0x0759U,0x0eb2U,0x1d64U,0x3ac8U,0x7590U,0xeb20U,0xc64bU,0x9c9dU,
		0x2931U,0x5262U,0xa4c4U,0x5983U,0xb306U,0x7607U,0xec0eU,0xc817U,0x8025U,0x1041U,
		0x2082U,0x4104U,0x8208U,0x141bU,0x2836U,0x506cU,0xa0d8U,0x51bbU,0xa376U,0x56e7U,
		0xadceU,0x4b97U,0x972eU,0x3e57U,0x7caeU,0xf95cU,0xe2b3U,0xd56dU,0xbad1U,0x65a9U,
		0xcb52U,0x86afU,0x1d55U,0x3aaaU,0x7554U,0xeaa8U,0xc55bU,0x9abdU,0x2571U,0x4ae2U,
		0x95c4U,0x3b83U,0x7706U,0xee0cU,0xcc13U,0x882dU,0x0051U,0x00a2U,0x0144U,0x0288U,
		0x0510U,0x0a20U,0x1440U,0x2880U,0x5100U,0xa200U,0x540bU,0xa816U,0x4027U,0x804eU,
		0x1097U,0x212eU,0x425cU,0x84b8U,0x197bU,0x32f6U,0x65ecU,0xcbd8U,0x87bbU,0x1f7dU,
		0x3efaU,0x7df4U,0xfbe8U,0xe7dbU,0xdfbdU,0xaf71U,0x4ee9U,0x9dd2U,0x2bafU,0x575eU,
		0xaebcU,0x4d73U,0x9ae6U,0x25c7U,0x4b8eU,0x971cU,0x3e33U,0x7c66U,0xf8ccU,0xe193U,
		0xd32dU,0xb651U,0x7ca9U,0xf952U,0xe2afU,0xd555U,0xbaa1U,0x6549U,0xca92U,0x852fU,
		0x1a55U,0x34aaU,0x6954U,0xd2a8U,0xb55bU,0x7abdU,0xf57aU,0xfaffU,0xe5f5U,0xdbe1U,
		0xa7c9U,0x5f99U,0xbf32U,0x6e6fU,0xdcdeU,0xa9b7U,0x4365U,0x86caU,0x1d9fU,0x3b3eU,
		0x767cU,0xecf8U,0xc9fbU,0x83fdU,0x17f1U,0x2fe2U,0x5fc4U,0xbf88U,0x6f1bU,0xde36U,
		0xac67U,0x48c5U,0x918aU,0x331fU,0x663eU,0xcc7cU,0x88f3U,0x01edU,0x03daU,0x07b4U,
		0x0f68U,0x1ed0U,0x3da0U,0x7b40U,0xf680U,0xfd0bU,0xea1dU,0xc431U,0x9869U,0x20d9U,
		0x41b2U,0x8364U,0x16c3U,0x2d86U,0x5b0cU,0xb618U,0x7c3bU,0xf876U,0xe0e7U,0xd1c5U,
		0xb381U,0x7709U,0xee12U,0xcc2fU,0x8855U,0x00a1U,0x0142U,0x0284U,0x0508U,0x0a10U,
		0x1420U,0x2840U,0x5080U,0xa100U,0x520bU,0xa416U,0x5827U,0xb04eU,0x7097U,0xe12eU,
		0xd257U,0xb4a5U,0x7941U,0xf282U,0xf50fU,0xfa15U,0xe421U,0xd849U,0xa099U,0x5139U,
		0xa272U,0x54efU,0xa9deU,0x43b7U,0x876eU,0x1ed7U,0x3daeU,0x7b5cU,0xf6b8U,0xfd7bU,
		0xeafdU,0xc5f1U,0x9be9U,0x27d9U,0x4fb2U,0x9f64U,0x2ec3U,0x5d86U,0xbb0cU,0x6613U,
		0xcc26U,0x8847U,0x0085U,0x010aU,0x0214U,0x0428U,0x0850U,0x10a0U,0x2140U,0x4280U,
		0x8500U,0x1a0bU,0x3416U,0x682cU,0xd058U,0xb0bbU,0x717dU,0xe2faU,0xd5ffU,0xbbf5U,
		0x67e1U,0xcfc2U,0x8f8fU,0x0f15U,0x1e2aU,0x3c54U,0x78a8U,0xf150U,0xf2abU,0xf55dU,
		0xfab1U,0xe569U,0xdad9U,0xa5b9U,0x5b79U,0xb6f2U,0x7defU,0xfbdeU,0xe7b7U,0xdf65U,
		0xaec1U,0x4d89U,0x9b12U,0x262fU,0x4c5eU,0x98bcU,0x2173U,0x42e6U,0x85ccU,0x1b93U,
		0x3726U,0x6e4cU,0xdc98U,0xa93bU,0x427dU,0x84faU,0x19ffU,0x33feU,0x67fcU,0xcff8U,
		0x8ffbU,0x0ffdU,0x1ffaU,0x3ff4U,0x7fe8U,0xffd0U,0xefabU,0xcf5dU,0x8eb1U,0x0d69U,
		0x1ad2U,0x35a4U,0x6b48U,0xd690U,0xbd2bU,0x6a5dU,0xd4baU,0xb97fU,0x62f5U,0xc5eaU,
		0x9bdfU,0x27b5U,0x4f6aU,0x9ed4U,0x2da3U,0x5b46U,0xb68cU,0x7d13U,0xfa26U,0xe447U,
		0xd885U,0xa101U,0x5209U,0xa412U,0x582fU,0xb05eU,0x70b7U,0xe16eU,0xd2d7U,0xb5a5U,
		0x7b41U,0xf682U,0xfd0fU,0xea15U,0xc421U,0x9849U,0x2099U,0x4132U,0x8264U,0x14c3U,
		0x2986U,0x530cU,0xa618U,0x5c3bU,0xb876U,0x60e7U,0xc1ceU,0x9397U,0x3725U,0x6e4aU,
		0xdc94U,0xa923U,0x424dU,0x849aU,0x193fU,0x327eU,0x64fcU,0xc9f8U,0x83fbU,0x17fdU,
		0x2ffaU,0x5ff4U,0xbfe8U,0x6fdbU,0xdfb6U,0xaf67U,0x4ec5U,0x9d8aU,0x2b1fU,0x563eU,
		0xac7cU,0x48f3U,0x91e6U,0x33c7U,0x678eU,0xcf1cU,0x8e33U,0x0c6dU,0x18daU,0x31b4U,
		0x6368U,0xc6d0U,0x9dabU,0x2b5dU,0x56baU,0xad74U,0x4ae3U,0x95c6U,0x3b87U,0x770eU,
		0xee1cU,0xcc33U,0x886dU,0x00d1U,0x01a2U,0x0344U,0x0688U,0x0d10U,0x1a20U,0x3440U,
		0x6880U,0xd100U,0xb20bU,0x741dU,0xe83aU,0xc07fU,0x90f5U,0x31e1U,0x63c2U,0xc784U,
		0x9f03U,0x2e0dU,0x5c1aU,0xb834U,0x6063U,0xc0c6U,0x9187U,0x3305U,0x660aU,0xcc14U,
		0x8823U,0x004dU,0x009aU,0x0134U,0x0268U,0x04d0U,0x09a0U,0x1340U,0x2680U,0x4d00U,
		0x9a00U,0x240bU,0x4816U,0x902cU,0x3053U,0x60a6U,0xc14cU,0x9293U,0x352dU,0x6a5aU,
		0xd4b4U,0xb963U,0x62cdU,0xc59aU,0x9b3fU,0x2675U,0x4ceaU,0x99d4U,0x23a3U,0x4746U,
		0x8e8cU,0x0d13U,0x1a26U,0x344cU,0x6898U,0xd130U,0xb26bU,0x74ddU,0xe9baU,0xc37fU,
		0x96f5U,0x3de1U,0x7bc2U,0xf784U,0xff03U,0xee0dU,0xcc11U,0x8829U,0x0059U,0x00b2U,
		0x0164U,0x02c8U,0x0590U,0x0b20U,0x1640U,0x2c80U,0x5900U,0xb200U,0x740bU,0xe816U,
		0xc027U,0x9045U,0x3081U,0x6102U,0xc204U,0x9403U,0x380dU,0x701aU,0xe034U,0xd063U,
		0xb0cdU,0x7191U,0xe322U,0xd64fU,0xbc95U,0x6921U,0xd242U,0xb48fU,0x7915U,0xf22aU,
		0xf45fU,0xf8b5U,0xe161U,0xd2c9U,0xb599U,0x7b39U,0xf672U,0xfcefU,0xe9d5U,0xc3a1U,
		0x9749U,0x3e99U,0x7d32U,0xfa64U,0xe4c3U,0xd98dU,0xa311U,0x5629U,0xac52U,0x48afU,
		0x915eU,0x32b7U,0x656eU,0xcadcU,0x85b3U,0x1b6dU,0x36daU,0x6db4U,0xdb68U,0xa6dbU,
		0x5dbdU,0xbb7aU,0x66ffU,0xcdfeU,0x8bf7U,0x07e5U,0x0fcaU,0x1f94U,0x3f28U,0x7e50U,
		0xfca0U,0xe94bU,0xc29dU,0x9531U,0x3a69U,0x74d2U,0xe9a4U,0xc343U,0x968dU,0x3d11U,
		0x7a22U,0xf444U,0xf883U,0xe10dU,0xd211U,0xb429U,0x7859U,0xf0b2U,0xf16fU,0xf2d5U,
		0xf5a1U,0xfb49U,0xe699U,0xdd39U,0xaa79U,0x44f9U,0x89f2U,0x03efU,0x07deU,0x0fbcU,
		0x1f78U,0x3ef0U,0x7de0U,0xfbc0U,0xe78bU,0xdf1dU,0xae31U,0x4c69U,0x98d2U,0x21afU,
		0x435eU,0x86bcU,0x1d73U,0x3ae6U,0x75ccU,0xeb98U,0xc73bU,0x9e7dU,0x2cf1U,0x59e2U,
		0xb3c4U,0x7783U,0xef06U,0xce07U,0x8c05U,0x0801U,0x1002U,0x2004U,0x4008U,0x8010U,
		0x102bU,0x2056U,0x40acU,0x8158U,0x12bbU,0x2576U,0x4aecU,0x95d8U,0x3bbbU,0x7776U,
		0xeeecU,0xcdd3U,0x8badU,0x0751U,0x0ea2U,0x1d44U,0x3a88U,0x7510U,0xea20U,0xc44bU,
		0x989dU,0x2131U,0x4262U,0x84c4U,0x1983U,0x3306U,0x660cU,0xcc18U,0x883bU,0x007dU,
		0x00faU,0x01f4U,0x03e8U,0x07d0U,0x0fa0U,0x1f40U,0x3e80U,0x7d00U,0xfa00U,0xe40bU,
		0xd81dU,0xa031U,0x5069U,0xa0d2U,0x51afU,0xa35eU,0x56b7U,0xad6eU,0x4ad7U,0x95aeU,
		0x3b57U,0x76aeU,0xed5cU,0xcab3U,0x856dU,0x1ad1U,0x35a2U,0x6b44U,0xd688U,0xbd1bU,
		0x6a3dU,0xd47aU,0xb8ffU,0x61f5U,0xc3eaU,0x97dfU,0x3fb5U,0x7f6aU,0xfed4U,0xeda3U,
		0xcb4dU,0x8691U,0x1d29U,0x3a52U,0x74a4U,0xe948U,0xc29bU,0x953dU,0x3a71U,0x74e2U,
		0xe9c4U,0xc383U,0x970dU,0x3e11U,0x7c22U,0xf844U,0xe083U,0xd10dU,0xb211U,0x7429U,
		0xe852U,0xc0afU,0x9155U,0x32a1U,0x6542U,0xca84U,0x8503U,0x1a0dU,0x341aU,0x6834U,
		0xd068U,0xb0dbU,0x71bdU,0xe37aU,0xd6ffU,0xbdf5U,0x6be1U,0xd7c2U,0xbf8fU,0x6f15U,
		0xde2aU,0xac5fU,0x48b5U,0x916aU,0x32dfU,0x65beU,0xcb7cU,0x86f3U,0x1dedU,0x3bdaU,
		0x77b4U,0xef68U,0xcedbU,0x8dbdU,0x0b71U,0x16e2U,0x2dc4U,0x5b88U,0xb710U,0x7e2bU,
		0xfc56U,0xe8a7U,0xc145U,0x9281U,0x3509U,0x6a12U,0xd424U,0xb843U,0x608dU,0xc11aU,
		0x923fU,0x3475U,0x68eaU,0xd1d4U,0xb3a3U,0x774dU,0xee9aU,0xcd3fU,0x8a75U,0x04e1U,
		0x09c2U,0x1384U,0x2708U,0x4e10U,0x9c20U,0x284bU,0x5096U,0xa12cU,0x5253U,0xa4a6U,
		0x5947U,0xb28eU,0x7517U,0xea2eU,0xc457U,0x98a5U,0x2141U,0x4282U,0x8504U,0x1a03U,
		0x3406U,0x680cU,0xd018U,0xb03bU,0x707dU,0xe0faU,0xd1ffU,0xb3f5U,0x77e1U,0xefc2U,
		0xcf8fU,0x8f15U,0x0e21U,0x1c42U,0x3884U,0x7108U,0xe210U,0xd42bU,0xb85dU,0x60b1U,
		0xc162U,0x92cfU,0x3595U,0x6b2aU,0xd654U,0xbca3U,0x694dU,0xd29aU,0xb53fU,0x7a75U,
		0xf4eaU,0xf9dfU,0xe3b5U,0xd761U,0xbec9U,0x6d99U,0xdb32U,0xa66fU,0x5cd5U,0xb9aaU,
		0x635fU,0xc6beU,0x9d77U,0x2ae5U,0x55caU,0xab94U,0x4723U,0x8e46U,0x0c87U,0x190eU,
		0x321cU,0x6438U,0xc870U,0x80ebU,0x11ddU,0x23baU,0x4774U,0x8ee8U,0x0ddbU,0x1bb6U,
		0x376cU,0x6ed8U,0xddb0U,0xab6bU,0x46ddU,0x8dbaU,0x0b7fU,0x16feU,0x2dfcU,0x5bf8U,
		0xb7f0U,0x7febU,0xffd6U,0xefa7U,0xcf45U,0x8e81U,0x0d09U,0x1a12U,0x3424U,0x6848U,
		0xd090U,0xb12bU,0x725dU,0xe4baU,0xd97fU,0xa2f5U,0x55e1U,0xabc2U,0x478fU,0x8f1eU,
		0x0e37U,0x1c6eU,0x38dcU,0x71b8U,0xe370U,0xd6ebU,0xbdddU,0x6bb1U,0xd762U,0xbecfU,
		0x6d95U,0xdb2aU,0xa65fU,0x5cb5U,0xb96aU,0x62dfU,0xc5beU,0x9b77U,0x26e5U,0x4dcaU,
		0x9b94U,0x2723U,0x4e46U,0x9c8cU,0x2913U,0x5226U,0xa44cU,0x5893U,0xb126U,0x7247U,
		0xe48eU,0xd917U,0xa225U,0x5441U,0xa882U,0x410fU,0x821eU,0x1437U,0x286eU,0x50dcU,
		0xa1b8U,0x537bU,0xa6f6U,0x5de7U,0xbbceU,0x6797U,0xcf2eU,0x8e57U,0x0ca5U,0x194aU,
		0x3294U,0x6528U,0xca50U,0x84abU,0x195dU,0x32baU,0x6574U,0xcae8U,0x85dbU,0x1bbdU,
		0x377aU,0x6ef4U,0xdde8U,0xabdbU,0x47bdU,0x8f7aU,0x0effU,0x1dfeU,0x3bfcU,0x77f8U,
		0xeff0U,0xcfebU,0x8fddU,0x0fb1U,0x1f62U,0x3ec4U,0x7d88U,0xfb10U,0xe62bU,0xdc5dU,
		0xa8b1U,0x4169U,0x82d2U,0x15afU,0x2b5eU,0x56bcU,0xad78U,0x4afbU,0x95f6U,0x3be7U,
		0x77ceU,0xef9cU,0xcf33U,0x8e6dU,0x0cd1U,0x19a2U,0x3344U,0x6688U,0xcd10U,0x8a2bU,
		0x045dU,0x08baU,0x1174U,0x22e8U,0x45d0U,0x8ba0U,0x074bU,0x0e96U,0x1d2cU,0x3a58U,
		0x74b0U,0xe960U,0xc2cbU,0x959dU,0x3b31U,0x7662U,0xecc4U,0xc983U,0x830dU,0x1611U,
		0x2c22U,0x5844U,0xb088U,0x711bU,0xe236U,0xd467U,0xb8c5U,0x6181U,0xc302U,0x960fU,
		0x3c15U,0x782aU,0xf054U,0xf0a3U,0xf14dU,0xf291U,0xf529U,0xfa59U,0xe4b9U,0xd979U,
		0xa2f9U,0x55f9U,0xabf2U,0x47efU,0x8fdeU,0x0fb7U,0x1f6eU,0x3edcU,0x7db8U,0xfb70U,
		0xe6ebU,0xddddU,0xabb1U,0x4769U,0x8ed2U,0x0dafU,0x1b5eU,0x36bcU,0x6d78U,0xdaf0U,
		0xa5ebU,0x5bddU,0xb7baU,0x7f7fU,0xfefeU,0xedf7U,0xcbe5U,0x87c1U,0x1f89U,0x3f12U,
		0x7e24U,0xfc48U,0xe89bU,0xc13dU,0x9271U,0x34e9U,0x69d2U,0xd3a4U,0xb743U,0x7e8dU,
		0xfd1aU,0xea3fU,0xc475U,0x98e1U,0x21c9U,0x4392U,0x8724U,0x1e43U,0x3c86U,0x790cU,
		0xf218U,0xf43bU,0xf87dU,0xe0f1U,0xd1e9U,0xb3d9U,0x77b9U,0xef72U,0xceefU,0x8dd5U,
		0x0ba1U,0x1742U,0x2e84U,0x5d08U,0xba10U,0x642bU,0xc856U,0x80a7U,0x1145U,0x228aU,
		0x4514U,0x8a28U,0x045bU,0x08b6U,0x116cU,0x22d8U,0x45b0U,0x8b60U,0x06cbU,0x0d96U,
		0x1b2cU,0x3658U,0x6cb0U,0xd960U,0xa2cbU,0x559dU,0xab3aU,0x467fU,0x8cfeU,0x09f7U,
		0x13eeU,0x27dcU,0x4fb8U,0x9f70U,0x2eebU,0x5dd6U,0xbbacU,0x6753U,0xcea6U,0x8d47U,
		0x0a85U,0x150aU,0x2a14U,0x5428U,0xa850U,0x40abU,0x8156U,0x12a7U,0x254eU,0x4a9cU,
		0x9538U,0x3a7bU,0x74f6U,0xe9ecU,0xc3d3U,0x97adU,0x3f51U,0x7ea2U,0xfd44U,0xea83U,
		0xc50dU,0x9a11U,0x2429U,0x4852U,0x90a4U,0x3143U,0x6286U,0xc50cU,0x9a13U,0x242dU,
		0x485aU,0x90b4U,0x3163U,0x62c6U,0xc58cU,0x9b13U,0x262dU,0x4c5aU,0x98b4U,0x2163U,
		0x42c6U,0x858cU,0x1b13U,0x3626U,0x6c4cU,0xd898U,0xa13bU,0x527dU,0xa4faU,0x59ffU,
		0xb3feU,0x77f7U,0xefeeU,0xcfd7U,0x8fa5U,0x0f41U,0x1e82U,0x3d04U,0x7a08U,0xf410U,
		0xf82bU,0xe05dU,0xd0b1U,0xb169U,0x72d9U,0xe5b2U,0xdb6fU,0xa6d5U,0x5da1U,0xbb42U,
		0x668fU,0xcd1eU,0x8a37U,0x0465U,0x08caU,0x1194U,0x2328U,0x4650U,0x8ca0U,0x094bU,
		0x1296U,0x252cU,0x4a58U,0x94b0U,0x396bU,0x72d6U,0xe5acU,0xdb53U,0xa6adU,0x5d51U,
		0xbaa2U,0x654fU,0xca9eU,0x8537U,0x1a65U,0x34caU,0x6994U,0xd328U,0xb65bU,0x7cbdU,
		0xf97aU,0xe2ffU,0xd5f5U,0xbbe1U,0x67c9U,0xcf92U,0x8f2fU,0x0e55U,0x1caaU,0x3954U,
		0x72a8U,0xe550U,0xdaabU,0xa55dU,0x5ab1U,0xb562U,0x7acfU,0xf59eU,0xfb37U,0xe665U,
		0xdcc1U,0xa989U,0x4319U,0x8632U,0x1c6fU,0x38deU,0x71bcU,0xe378U,0xd6fbU,0xbdfdU,
		0x6bf1U,0xd7e2U,0xbfcfU,0x6f95U,0xdf2aU,0xae5fU,0x4cb5U,0x996aU,0x22dfU,0x45beU,
		0x8b7cU,0x06f3U,0x0de6U,0x1bccU,0x3798U,0x6f30U,0xde60U,0xaccbU,0x499dU,0x933aU,
		0x367fU,0x6cfeU,0xd9fcU,0xa3f3U,0x57edU,0xafdaU,0x4fbfU,0x9f7eU,0x2ef7U,0x5deeU,
		0xbbdcU,0x67b3U,0xcf66U,0x8ec7U,0x0d85U,0x1b0aU,0x3614U,0x6c28U,0xd850U,0xa0abU,
		0x515dU,0xa2baU,0x557fU,0xaafeU,0x45f7U,0x8beeU,0x07d7U,0x0faeU,0x1f5cU,0x3eb8U,
		0x7d70U,0xfae0U,0xe5cbU,0xdb9dU,0xa731U,0x5e69U,0xbcd2U,0x69afU,0xd35eU,0xb6b7U,
		0x7d65U,0xfacaU,0xe59fU,0xdb35U,0xa661U,0x5cc9U,0xb992U,0x632fU,0xc65eU,0x9cb7U,
		0x2965U,0x52caU,0xa594U,0x5b23U,0xb646U,0x7c87U,0xf90eU,0xe217U,0xd425U,0xb841U,
		0x6089U,0xc112U,0x922fU,0x3455U,0x68aaU,0xd154U,0xb2a3U,0x754dU,0xea9aU,0xc53fU,
		0x9a75U,0x24e1U,0x49c2U,0x9384U,0x3703U,0x6e06U,0xdc0cU,0xa813U,0x402dU,0x805aU,
		0x10bfU,0x217eU,0x42fcU,0x85f8U,0x1bfbU,0x37f6U,0x6fecU,0xdfd8U,0xafbbU,0x4f7dU,
		0x9efaU,0x2dffU,0x5bfeU,0xb7fcU,0x7ff3U,0xffe6U,0xefc7U,0xcf85U,0x8f01U,0x0e09U,
		0x1c12U,0x3824U,0x7048U,0xe090U,0xd12bU,0xb25dU,0x74b1U,0xe962U,0xc2cfU,0x9595U,
		0x3b21U,0x7642U,0xec84U,0xc903U,0x820dU,0x1411U,0x2822U,0x5044U,0xa088U,0x511bU,
		0xa236U,0x5467U,0xa8ceU,0x4197U,0x832eU,0x1657U,0x2caeU,0x595cU,0xb2b8U,0x757bU,
		0xeaf6U,0xc5e7U,0x9bc5U,0x2781U,0x4f02U,0x9e04U,0x2c03U,0x5806U,0xb00cU,0x7013U,
		0xe026U,0xd047U,0xb085U,0x7101U,0xe202U,0xd40fU,0xb815U,0x6021U,0xc042U,0x908fU,
		0x3115U,0x622aU,0xc454U,0x98a3U,0x214dU,0x429aU,0x8534U,0x1a63U,0x34c6U,0x698cU,
		0xd318U,0xb63bU,0x7c7dU,0xf8faU,0xe1ffU,0xd3f5U,0xb7e1U,0x7fc9U,0xff92U,0xef2fU,
		0xce55U,0x8ca1U,0x0949U,0x1292U,0x2524U,0x4a48U,0x9490U,0x392bU,0x7256U,0xe4acU,
		0xd953U,0xa2adU,0x5551U,0xaaa2U,0x454fU,0x8a9eU,0x0537U,0x0a6eU,0x14dcU,0x29b8U,
		0x5370U,0xa6e0U,0x5dcbU,0xbb96U,0x6727U,0xce4eU,0x8c97U,0x0925U,0x124aU,0x2494U,
		0x4928U,0x9250U,0x34abU,0x6956U,0xd2acU,0xb553U,0x7aadU,0xf55aU,0xfabfU,0xe575U,
		0xdae1U,0xa5c9U,0x5b99U,0xb732U,0x7e6fU,0xfcdeU,0xe9b7U,0xc365U,0x96c1U,0x3d89U,
		0x7b12U,0xf624U,0xfc43U,0xe88dU,0xc111U,0x9229U,0x3459U,0x68b2U,0xd164U,0xb2c3U,
		0x758dU,0xeb1aU,0xc63fU,0x9c75U,0x28e1U,0x51c2U,0xa384U,0x5703U,0xae06U,0x4c07U,
		0x980eU,0x2017U,0x402eU,0x805cU,0x10b3U,0x2166U,0x42ccU,0x8598U,0x1b3bU,0x3676U,
		0x6cecU,0xd9d8U,0xa3bbU,0x577dU,0xaefaU,0x4dffU,0x9bfeU,0x27f7U,0x4feeU,0x9fdcU,
		0x2fb3U,0x5f66U,0xbeccU,0x6d93U,0xdb26U,0xa647U,0x5c85U,0xb90aU,0x621fU,0xc43eU,
		0x9877U,0x20e5U,0x41caU,0x8394U,0x1723U,0x2e46U,0x5c8cU,0xb918U,0x623bU,0xc476U,
		0x98e7U,0x21c5U,0x438aU,0x8714U,0x1e23U,0x3c46U,0x788cU,0xf118U,0xf23bU,0xf47dU,
		0xf8f1U,0xe1e9U,0xd3d9U,0xb7b9U,0x7f79U,0xfef2U,0xedefU,0xcbd5U,0x87a1U,0x1f49U,
		0x3e92U,0x7d24U,0xfa48U,0xe49bU,0xd93dU,0xa271U,0x54e9U,0xa9d2U,0x43afU,0x875eU,
		0x1eb7U,0x3d6eU,0x7adcU,0xf5b8U,0xfb7bU,0xe6fdU,0xddf1U,0xabe9U,0x47d9U,0x8fb2U,
		0x0f6fU,0x1edeU,0x3dbcU,0x7b78U,0xf6f0U,0xfdebU,0xebddU,0xc7b1U,0x9f69U,0x2ed9U,
		0x5db2U,0xbb64U,0x66c3U,0xcd86U,0x8b07U,0x0605U,0x0c0aU,0x1814U,0x3028U,0x6050U,
		0xc0a0U,0x914bU,0x329dU,0x653aU,0xca74U,0x84e3U,0x19cdU,0x339aU,0x6734U,0xce68U,
		0x8cdbU,0x09bdU,0x137aU,0x26f4U,0x4de8U,0x9bd0U,0x27abU,0x4f56U,0x9eacU,0x2d53U,
		0x5aa6U,0xb54cU,0x7a93U,0xf526U,0xfa47U,0xe485U,0xd901U,0xa209U,0x5419U,0xa832U,
		0x406fU,0x80deU,0x11b7U,0x236eU,0x46dcU,0x8db8U,0x0b7bU,0x16f6U,0x2decU,0x5bd8U,
		0xb7b0U,0x7f6bU,0xfed6U,0xeda7U,0xcb45U,0x8681U,0x1d09U,0x3a12U,0x7424U,0xe848U,
		0xc09bU,0x913dU,0x3271U,0x64e2U,0xc9c4U,0x8383U,0x170dU,0x2e1aU,0x5c34U,0xb868U,
		0x60dbU,0xc1b6U,0x9367U,0x36c5U,0x6d8aU,0xdb14U,0xa623U,0x5c4dU,0xb89aU,0x613fU,
		0xc27eU,0x94f7U,0x39e5U,0x73caU,0xe794U,0xdf23U,0xae4dU,0x4c91U,0x9922U,0x224fU,
		0x449eU,0x893cU,0x0273U,0x04e6U,0x09ccU,0x1398U,0x2730U,0x4e60U,0x9cc0U,0x298bU,
		0x5316U,0xa62cU,0x5c53U,0xb8a6U,0x6147U,0xc28eU,0x9517U,0x3a25U,0x744aU,0xe894U,
		0xc123U,0x924dU,0x3491U,0x6922U,0xd244U,0xb483U,0x790dU,0xf21aU,0xf43fU,0xf875U,
		0xe0e1U,0xd1c9U,0xb399U,0x7739U,0xee72U,0xccefU,0x89d5U,0x03a1U,0x0742U,0x0e84U,
		0x1d08U,0x3a10U,0x7420U,0xe840U,0xc08bU,0x911dU,0x3231U,0x6462U,0xc8c4U,0x8183U,
		0x130dU,0x261aU,0x4c34U,0x9868U,0x20dbU,0x41b6U,0x836cU,0x16d3U,0x2da6U,0x5b4cU,
		0xb698U,0x7d3bU,0xfa76U,0xe4e7U,0xd9c5U,0xa381U,0x5709U,0xae12U,0x4c2fU,0x985eU,
		0x20b7U,0x416eU,0x82dcU,0x15b3U,0x2b66U,0x56ccU,0xad98U,0x4b3bU,0x9676U,0x3ce7U,
		0x79ceU,0xf39cU,0xf733U,0xfe6dU,0xecd1U,0xc9a9U,0x8359U,0x16b9U,0x2d72U,0x5ae4U,
		0xb5c8U,0x7b9bU,0xf736U,0xfe67U,0xecc5U,0xc981U,0x8309U,0x1619U,0x2c32U,0x5864U,
		0xb0c8U,0x719bU,0xe336U,0xd667U,0xbcc5U,0x6981U,0xd302U,0xb60fU,0x7c15U,0xf82aU,
		0xe05fU,0xd0b5U,0xb161U,0x72c9U,0xe592U,0xdb2fU,0xa655U,0x5ca1U,0xb942U,0x628fU,
		0xc51eU,0x9a37U,0x2465U,0x48caU,0x9194U,0x3323U,0x6646U,0xcc8cU,0x8913U,0x022dU,
		0x045aU,0x08b4U,0x1168U,0x22d0U,0x45a0U,0x8b40U,0x068bU,0x0d16U,0x1a2cU,0x3458U,
		0x68b0U,0xd160U,0xb2cbU,0x759dU,0xeb3aU,0xc67fU,0x9cf5U,0x29e1U,0x53c2U,0xa784U,
		0x5f03U,0xbe06U,0x6c07U,0xd80eU,0xa017U,0x5025U,0xa04aU,0x509fU,0xa13eU,0x5277U,
		0xa4eeU,0x59d7U,0xb3aeU,0x7757U,0xeeaeU,0xcd57U,0x8aa5U,0x0541U,0x0a82U,0x1504U,
		0x2a08U,0x5410U,0xa820U,0x404bU,0x8096U,0x1127U,0x224eU,0x449cU,0x8938U,0x027bU,
		0x04f6U,0x09ecU,0x13d8U,0x27b0U,0x4f60U,0x9ec0U,0x2d8bU,0x5b16U,0xb62cU,0x7c53U,
		0xf8a6U,0xe147U,0xd285U,0xb501U,0x7a09U,0xf412U,0xf82fU,0xe055U,0xd0a1U,0xb149U,
		0x7299U,0xe532U,0xda6fU,0xa4d5U,0x59a1U,0xb342U,0x768fU,0xed1eU,0xca37U,0x8465U,
		0x18c1U,0x3182U,0x6304U,0xc608U,0x9c1bU,0x283dU,0x507aU,0xa0f4U,0x51e3U,0xa3c6U,
		0x5787U,0xaf0eU,0x4e17U,0x9c2eU,0x2857U,0x50aeU,0xa15cU,0x52b3U,0xa566U,0x5ac7U,
		0xb58eU,0x7b17U,0xf62eU,0xfc57U,0xe8a5U,0xc141U,0x9289U,0x3519U,0x6a32U,0xd464U,
		0xb8c3U,0x618dU,0xc31aU,0x963fU,0x3c75U,0x78eaU,0xf1d4U,0xf3a3U,0xf74dU,0xfe91U,
		0xed29U,0xca59U,0x84b9U,0x1979U,0x32f2U,0x65e4U,0xcbc8U,0x879bU,0x1f3dU,0x3e7aU,
		0x7cf4U,0xf9e8U,0xe3dbU,0xd7bdU,0xbf71U,0x6ee9U,0xddd2U,0xabafU,0x4755U,0x8eaaU,
		0x0d5fU,0x1abeU,0x357cU,0x6af8U,0xd5f0U,0xbbebU,0x67ddU,0xcfbaU,0x8f7fU,0x0ef5U,
		0x1deaU,0x3bd4U,0x77a8U,0xef50U,0xceabU,0x8d5dU,0x0ab1U,0x1562U,0x2ac4U,0x5588U,
		0xab10U,0x462bU,0x8c56U,0x08a7U,0x114eU,0x229cU,0x4538U,0x8a70U,0x04ebU,0x09d6U,
		0x13acU,0x2758U,0x4eb0U,0x9d60U,0x2acbU,0x5596U,0xab2cU,0x4653U,0x8ca6U,0x0947U,
		0x128eU,0x251cU,0x4a38U,0x9470U,0x38ebU,0x71d6U,0xe3acU,0xd753U,0xbeadU,0x6d51U,
		0xdaa2U,0xa54fU,0x5a95U,0xb52aU,0x7a5fU,0xf4beU,0xf977U,0xe2e5U,0xd5c1U,0xbb89U,
		0x6719U,0xce32U,0x8c6fU,0x08d5U,0x11aaU,0x2354U,0x46a8U,0x8d50U,0x0aabU,0x1556U,
		0x2aacU,0x5558U,0xaab0U,0x456bU,0x8ad6U,0x05a7U,0x0b4eU,0x169cU,0x2d38U,0x5a70U,
		0xb4e0U,0x79cbU,0xf396U,0xf727U,0xfe45U,0xec81U,0xc909U,0x8219U,0x1439U,0x2872U,
		0x50e4U,0xa1c8U,0x539bU,0xa736U,0x5e67U,0xbcceU,0x6997U,0xd32eU,0xb657U,0x7ca5U,
		0xf94aU,0xe29fU,0xd535U,0xba61U,0x64c9U,0xc992U,0x832fU,0x1655U,0x2caaU,0x5954U,
		0xb2a8U,0x755bU,0xeab6U,0xc567U,0x9ac5U,0x2581U,0x4b02U,0x9604U,0x3c03U,0x7806U,
		0xf00cU,0xf013U,0xf02dU,0xf051U,0xf0a9U,0xf159U,0xf2b9U,0xf579U,0xfaf9U,0xe5f9U,
		0xdbf9U,0xa7f9U,0x5ff9U,0xbff2U,0x6fefU,0xdfdeU,0xafb7U,0x4f65U,0x9ecaU,0x2d9fU,
		0x5b3eU,0xb67cU,0x7cf3U,0xf9e6U,0xe3c7U,0xd785U,0xbf01U,0x6e09U,0xdc12U,0xa82fU,
		0x4055U,0x80aaU,0x115fU,0x22beU,0x457cU,0x8af8U,0x05fbU,0x0bf6U,0x17ecU,0x2fd8U,
		0x5fb0U,0xbf60U,0x6ecbU,0xdd96U,0xab27U,0x4645U,0x8c8aU,0x091fU,0x123eU,0x247cU,
		0x48f8U,0x91f0U,0x33ebU,0x67d6U,0xcfacU,0x8f53U,0x0eadU,0x1d5aU,0x3ab4U,0x7568U,
		0xead0U,0xc5abU,0x9b5dU,0x26b1U,0x4d62U,0x9ac4U,0x2583U,0x4b06U,0x960cU,0x3c13U,
		0x7826U,0xf04cU,0xf093U,0xf12dU,0xf251U,0xf4a9U,0xf959U,0xe2b9U,0xd579U,0xbaf9U,
		0x65f9U,0xcbf2U,0x87efU,0x1fd5U,0x3faaU,0x7f54U,0xfea8U,0xed5bU,0xcabdU,0x8571U,
		0x1ae9U,0x35d2U,0x6ba4U,0xd748U,0xbe9bU,0x6d3dU,0xda7aU,0xa4ffU,0x59f5U,0xb3eaU,
		0x77dfU,0xefbeU,0xcf77U,0x8ee5U,0x0dc1U,0x1b82U,0x3704U,0x6e08U,0xdc10U,0xa82bU,
		0x405dU,0x80baU,0x117fU,0x22feU,0x45fcU,0x8bf8U,0x07fbU,0x0ff6U,0x1fecU,0x3fd8U,
		0x7fb0U,0xff60U,0xeecbU,0xcd9dU,0x8b31U,0x0669U,0x0cd2U,0x19a4U,0x3348U,0x6690U,
		0xcd20U,0x8a4bU,0x049dU,0x093aU,0x1274U,0x24e8U,0x49d0U,0x93a0U,0x374bU,0x6e96U,
		0xdd2cU,0xaa53U,0x44adU,0x895aU,0x02bfU,0x057eU,0x0afcU,0x15f8U,0x2bf0U,0x57e0U,
		0xafc0U,0x4f8bU,0x9f16U,0x2e27U,0x5c4eU,0xb89cU,0x6133U,0xc266U,0x94c7U,0x3985U,
		0x730aU,0xe614U,0xdc23U,0xa84dU,0x4091U,0x8122U,0x124fU,0x249eU,0x493cU,0x9278U,
		0x34fbU,0x69f6U,0xd3ecU,0xb7d3U,0x7fadU,0xff5aU,0xeebfU,0xcd75U,0x8ae1U,0x05c9U,
		0x0b92U,0x1724U,0x2e48U,0x5c90U,0xb920U,0x624bU,0xc496U,0x9927U,0x2245U,0x448aU,
		0x8914U,0x0223U,0x0446U,0x088cU,0x1118U,0x2230U,0x4460U,0x88c0U,0x018bU,0x0316U,
		0x062cU,0x0c58U,0x18b0U,0x3160U,0x62c0U,0xc580U,0x9b0bU,0x261dU,0x4c3aU,0x9874U,
		0x20e3U,0x41c6U,0x838cU,0x1713U,0x2e26U,0x5c4cU,0xb898U,0x613bU,0xc276U,0x94e7U,
		0x39c5U,0x738aU,0xe714U,0xde23U,0xac4dU,0x4891U,0x9122U,0x324fU,0x649eU,0xc93cU,
		0x8273U,0x14edU,0x29daU,0x53b4U,0xa768U,0x5edbU,0xbdb6U,0x6b67U,0xd6ceU,0xbd97U,
		0x6b25U,0xd64aU,0xbc9fU,0x6935U,0xd26aU,0xb4dfU,0x79b5U,0xf36aU,0xf6dfU,0xfdb5U,
		0xeb61U,0xc6c9U,0x9d99U,0x2b39U,0x5672U,0xace4U,0x49c3U,0x9386U,0x3707U,0x6e0eU,
		0xdc1cU,0xa833U,0x406dU,0x80daU,0x11bfU,0x237eU,0x46fcU,0x8df8U,0x0bfbU,0x17f6U,
		0x2fecU,0x5fd8U,0xbfb0U,0x6f6bU,0xded6U,0xada7U,0x4b45U,0x968aU,0x3d1fU,0x7a3eU,
		0xf47cU,0xf8f3U,0xe1edU,0xd3d1U,0xb7a9U,0x7f59U,0xfeb2U,0xed6fU,0xcad5U,0x85a1U,
		0x1b49U,0x3692U,0x6d24U,0xda48U,0xa49bU,0x593dU,0xb27aU,0x74ffU,0xe9feU,0xc3f7U,
		0x97e5U,0x3fc1U,0x7f82U,0xff04U,0xee03U,0xcc0dU,0x8811U,0x0029U,0x0052U,0x00a4U,
		0x0148U,0x0290U,0x0520U,0x0a40U,0x1480U,0x2900U,0x5200U,0xa400U,0x580bU,0xb016U,
		0x7027U,0xe04eU,0xd097U,0xb125U,0x7241U,0xe482U,0xd90fU,0xa215U,0x5421U,0xa842U,
		0x408fU,0x811eU,0x1237U,0x246eU,0x48dcU,0x91b8U,0x337bU,0x66f6U,0xcdecU,0x8bd3U,
		0x07adU,0x0f5aU,0x1eb4U,0x3d68U,0x7ad0U,0xf5a0U,0xfb4bU,0xe69dU,0xdd31U,0xaa69U,
		0x44d9U,0x89b2U,0x036fU,0x06deU,0x0dbcU,0x1b78U,0x36f0U,0x6de0U,0xdbc0U,0xa78bU,
		0x5f1dU,0xbe3aU,0x6c7fU,0xd8feU,0xa1f7U,0x53e5U,0xa7caU,0x5f9fU,0xbf3eU,0x6e77U,
		0xdceeU,0xa9d7U,0x43a5U,0x874aU,0x1e9fU,0x3d3eU,0x7a7cU,0xf4f8U,0xf9fbU,0xe3fdU,
		0xd7f1U,0xbfe9U,0x6fd9U,0xdfb2U,0xaf6fU,0x4ed5U,0x9daaU,0x2b5fU,0x56beU,0xad7cU,
		0x4af3U,0x95e6U,0x3bc7U,0x778eU,0xef1cU,0xce33U,0x8c6dU,0x08d1U,0x11a2U,0x2344U,
		0x4688U,0x8d10U,0x0a2bU,0x1456U,0x28acU,0x5158U,0xa2b0U,0x556bU,0xaad6U,0x45a7U,
		0x8b4eU,0x0697U,0x0d2eU,0x1a5cU,0x34b8U,0x6970U,0xd2e0U,0xb5cbU,0x7b9dU,0xf73aU,
		0xfe7fU,0xecf5U,0xc9e1U,0x83c9U,0x1799U,0x2f32U,0x5e64U,0xbcc8U,0x699bU,0xd336U,
		0xb667U,0x7cc5U,0xf98aU,0xe31fU,0xd635U,0xbc61U,0x68c9U,0xd192U,0xb32fU,0x7655U,
		0xecaaU,0xc95fU,0x82b5U,0x1561U,0x2ac2U,0x5584U,0xab08U,0x461bU,0x8c36U,0x0867U,
		0x10ceU,0x219cU,0x4338U,0x8670U,0x1cebU,0x39d6U,0x73acU,0xe758U,0xdebbU,0xad7dU,
		0x4af1U,0x95e2U,0x3bcfU,0x779eU,0xef3cU,0xce73U,0x8cedU,0x09d1U,0x13a2U,0x2744U,
		0x4e88U,0x9d10U,0x2a2bU,0x5456U,0xa8acU,0x4153U,0x82a6U,0x1547U,0x2a8eU,0x551cU,
		0xaa38U,0x447bU,0x88f6U,0x01e7U,0x03ceU,0x079cU,0x0f38U,0x1e70U,0x3ce0U,0x79c0U,
		0xf380U,0xf70bU,0xfe1dU,0xec31U,0xc869U,0x80d9U,0x11b9U,0x2372U,0x46e4U,0x8dc8U,
		0x0b9bU,0x1736U,0x2e6cU,0x5cd8U,0xb9b0U,0x636bU,0xc6d6U,0x9da7U,0x2b45U,0x568aU,
		0xad14U,0x4a23U,0x9446U,0x3887U,0x710eU,0xe21cU,0xd433U,0xb86dU,0x60d1U,0xc1a2U,
		0x934fU,0x3695U,0x6d2aU,0xda54U,0xa4a3U,0x594dU,0xb29aU,0x753fU,0xea7eU,0xc4f7U,
		0x99e5U,0x23c1U,0x4782U,0x8f04U,0x0e03U,0x1c06U,0x380cU,0x7018U,0xe030U,0xd06bU,
		0xb0ddU,0x71b1U,0xe362U,0xd6cfU,0xbd95U,0x6b21U,0xd642U,0xbc8fU,0x6915U,0xd22aU,
		0xb45fU,0x78b5U,0xf16aU,0xf2dfU,0xf5b5U,0xfb61U,0xe6c9U,0xdd99U,0xab39U,0x4679U,
		0x8cf2U,0x09efU,0x13deU,0x27bcU,0x4f78U,0x9ef0U,0x2debU,0x5bd6U,0xb7acU,0x7f53U,
		0xfea6U,0xed47U,0xca85U,0x8501U,0x1a09U,0x3412U,0x6824U,0xd048U,0xb09bU,0x713dU,
		0xe27aU,0xd4ffU,0xb9f5U,0x63e1U,0xc7c2U,0x9f8fU,0x2f15U,0x5e2aU,0xbc54U,0x68a3U,
		0xd146U,0xb287U,0x7505U,0xea0aU,0xc41fU,0x9835U,0x2061U,0x40c2U,0x8184U,0x1303U,
		0x2606U,0x4c0cU,0x9818U,0x203bU,0x4076U,0x80ecU,0x11d3U,0x23a6U,0x474cU,0x8e98U,
		0x0d3bU,0x1a76U,0x34ecU,0x69d8U,0xd3b0U,0xb76bU,0x7eddU,0xfdbaU,0xeb7fU,0xc6f5U,
		0x9de1U,0x2bc9U,0x5792U,0xaf24U,0x4e43U,0x9c86U,0x2907U,0x520eU,0xa41cU,0x5833U,
		0xb066U,0x70c7U,0xe18eU,0xd317U,0xb625U,0x7c41U,0xf882U,0xe10fU,0xd215U,0xb421U,
		0x7849U,0xf092U,0xf12fU,0xf255U,0xf4a1U,0xf949U,0xe299U,0xd539U,0xba79U,0x64f9U,
		0xc9f2U,0x83efU,0x17d5U,0x2faaU,0x5f54U,0xbea8U,0x6d5bU,0xdab6U,0xa567U,0x5ac5U,
		0xb58aU,0x7b1fU,0xf63eU,0xfc77U,0xe8e5U,0xc1c1U,0x9389U,0x3719U,0x6e32U,0xdc64U,
		0xa8c3U,0x418dU,0x831aU,0x163fU,0x2c7eU,0x58fcU,0xb1f8U,0x73fbU,0xe7f6U,0xdfe7U,
		0xafc5U,0x4f81U,0x9f02U,0x2e0fU,0x5c1eU,0xb83cU,0x6073U,0xc0e6U,0x91c7U,0x3385U,
		0x670aU,0xce14U,0x8c23U,0x084dU,0x109aU,0x2134U,0x4268U,0x84d0U,0x19abU,0x3356U,
		0x66acU,0xcd58U,0x8abbU,0x057dU,0x0afaU,0x15f4U,0x2be8U,0x57d0U,0xafa0U,0x4f4bU,
		0x9e96U,0x2d27U,0x5a4eU,0xb49cU,0x7933U,0xf266U,0xf4c7U,0xf985U,0xe301U,0xd609U,
		0xbc19U,0x6839U,0xd072U,0xb0efU,0x71d5U,0xe3aaU,0xd75fU,0xbeb5U,0x6d61U,0xdac2U,
		0xa58fU,0x5b15U,0xb62aU,0x7c5fU,0xf8beU,0xe177U,0xd2e5U,0xb5c1U,0x7b89U,0xf712U,
		0xfe2fU,0xec55U,0xc8a1U,0x8149U,0x1299U,0x2532U,0x4a64U,0x94c8U,0x399bU,0x7336U,
		0xe66cU,0xdcd3U,0xa9adU,0x4351U,0x86a2U,0x1d4fU,0x3a9eU,0x753cU,0xea78U,0xc4fbU,
		0x99fdU,0x23f1U,0x47e2U,0x8fc4U,0x0f83U,0x1f06U,0x3e0cU,0x7c18U,0xf830U,0xe06bU,
		0xd0ddU,0xb1b1U,0x7369U,0xe6d2U,0xddafU,0xab55U,0x46a1U,0x8d42U,0x0a8fU,0x151eU,
		0x2a3cU,0x5478U,0xa8f0U,0x41ebU,0x83d6U,0x17a7U,0x2f4eU,0x5e9cU,0xbd38U,0x6a7bU,
		0xd4f6U,0xb9e7U,0x63c5U,0xc78aU,0x9f1fU,0x2e35U,0x5c6aU,0xb8d4U,0x61a3U,0xc346U,
		0x9687U,0x3d05U,0x7a0aU,0xf414U,0xf823U,0xe04dU,0xd091U,0xb129U,0x7259U,0xe4b2U,
		0xd96fU,0xa2d5U,0x55a1U,0xab42U,0x468fU,0x8d1eU,0x0a37U,0x146eU,0x28dcU,0x51b8U,
		0xa370U,0x56ebU,0xadd6U,0x4ba7U,0x974eU,0x3e97U,0x7d2eU,0xfa5cU,0xe4b3U,0xd96dU,
		0xa2d1U,0x55a9U,0xab52U,0x46afU,0x8d5eU,0x0ab7U,0x156eU,0x2adcU,0x55b8U,0xab70U,
		0x46ebU,0x8dd6U,0x0ba7U,0x174eU,0x2e9cU,0x5d38U,0xba70U,0x64ebU,0xc9d6U,0x83a7U,
		0x1745U,0x2e8aU,0x5d14U,0xba28U,0x645bU,0xc8b6U,0x8167U,0x12c5U,0x258aU,0x4b14U,
		0x9628U,0x3c5bU,0x78b6U,0xf16cU,0xf2d3U,0xf5adU,0xfb51U,0xe6a9U,0xdd59U,0xaab9U,
		0x4579U,0x8af2U,0x05efU,0x0bdeU,0x17bcU,0x2f78U,0x5ef0U,0xbde0U,0x6bcbU,0xd796U,
		0xbf27U,0x6e45U,0xdc8aU,0xa91fU,0x4235U,0x846aU,0x18dfU,0x31beU,0x637cU,0xc6f8U,
		0x9dfbU,0x2bfdU,0x57faU,0xaff4U,0x4fe3U,0x9fc6U,0x2f87U,0x5f0eU,0xbe1cU,0x6c33U,
		0xd866U,0xa0c7U,0x5185U,0xa30aU,0x561fU,0xac3eU,0x4877U,0x90eeU,0x31d7U,0x63aeU,
		0xc75cU,0x9eb3U,0x2d6dU,0x5adaU,0xb5b4U,0x7b63U,0xf6c6U,0xfd87U,0xeb05U,0xc601U,
		0x9c09U,0x2819U,0x5032U,0xa064U,0x50c3U,0xa186U,0x5307U,0xa60eU,0x5c17U,0xb82eU,
		0x6057U,0xc0aeU,0x9157U,0x32a5U,0x654aU,0xca94U,0x8523U,0x1a4dU,0x349aU,0x6934U,
		0xd268U,0xb4dbU,0x79bdU,0xf37aU,0xf6ffU,0xfdf5U,0xebe1U,0xc7c9U,0x9f99U,0x2f39U,
		0x5e72U,0xbce4U,0x69c3U,0xd386U,0xb707U,0x7e05U,0xfc0aU,0xe81fU,0xc035U,0x9061U,
		0x30c9U,0x6192U,0xc324U,0x9643U,0x3c8dU,0x791aU,0xf234U,0xf463U,0xf8cdU,0xe191U,
		0xd329U,0xb659U,0x7cb9U,0xf972U,0xe2efU,0xd5d5U,0xbba1U,0x6749U,0xce92U,0x8d2fU,
		0x0a55U,0x14aaU,0x2954U,0x52a8U,0xa550U,0x5aabU,0xb556U,0x7aa7U,0xf54eU,0xfa97U,
		0xe525U,0xda41U,0xa489U,0x5919U,0xb232U,0x746fU,0xe8deU,0xc1b7U,0x9365U,0x36c1U,
		0x6d82U,0xdb04U,0xa603U,0x5c0dU,0xb81aU,0x603fU,0xc07eU,0x90f7U,0x31e5U,0x63caU,
		0xc794U,0x9f23U,0x2e4dU,0x5c9aU,0xb934U,0x6263U,0xc4c6U,0x9987U,0x2305U,0x460aU,
		0x8c14U,0x0823U,0x1046U,0x208cU,0x4118U,0x8230U,0x146bU,0x28d6U,0x51acU,0xa358U,
		0x56bbU,0xad76U,0x4ae7U,0x95ceU,0x3b97U,0x772eU,0xee5cU,0xccb3U,0x896dU,0x02d1U,
		0x05a2U,0x0b44U,0x1688U,0x2d10U,0x5a20U,0xb440U,0x788bU,0xf116U,0xf227U,0xf445U,
		0xf881U,0xe109U,0xd219U,0xb439U,0x7879U,0xf0f2U,0xf1efU,0xf3d5U,0xf7a1U,0xff49U,
		0xee99U,0xcd39U,0x8a79U,0x04f9U,0x09f2U,0x13e4U,0x27c8U,0x4f90U,0x9f20U,0x2e4bU,
		0x5c96U,0xb92cU,0x6253U,0xc4a6U,0x9947U,0x2285U,0x450aU,0x8a14U,0x0423U,0x0846U,
		0x108cU,0x2118U,0x4230U,0x8460U,0x18cbU,0x3196U,0x632cU,0xc658U,0x9cbbU,0x297dU,
		0x52faU,0xa5f4U,0x5be3U,0xb7c6U,0x7f87U,0xff0eU,0xee17U,0xcc25U,0x8841U,0x0089U,
		0x0112U,0x0224U,0x0448U,0x0890U,0x1120U,0x2240U,0x4480U,0x8900U,0x020bU,0x0416U,
		0x082cU,0x1058U,0x20b0U,0x4160U,0x82c0U,0x158bU,0x2b16U,0x562cU,0xac58U,0x48bbU,
		0x9176U,0x32e7U,0x65ceU,0xcb9cU,0x8733U,0x1e6dU,0x3cdaU,0x79b4U,0xf368U,0xf6dbU,
		0xfdbdU,0xeb71U,0xc6e9U,0x9dd9U,0x2bb9U,0x5772U,0xaee4U,0x4dc3U,0x9b86U,0x2707U,
		0x4e0eU,0x9c1cU,0x2833U,0x5066U,0xa0ccU,0x5193U,0xa326U,0x5647U,0xac8eU,0x4917U,
		0x922eU,0x3457U,0x68aeU,0xd15cU,0xb2b3U,0x756dU,0xeadaU,0xc5bfU,0x9b75U,0x26e1U,
		0x4dc2U,0x9b84U,0x2703U,0x4e06U,0x9c0cU,0x2813U,0x5026U,0xa04cU,0x5093U,0xa126U,
		0x5247U,0xa48eU,0x5917U,0xb22eU,0x7457U,0xe8aeU,0xc157U,0x92a5U,0x3541U,0x6a82U,
		0xd504U,0xba03U,0x640dU,0xc81aU,0x803fU,0x1075U,0x20eaU,0x41d4U,0x83a8U,0x175bU,
		0x2eb6U,0x5d6cU,0xbad8U,0x65bbU,0xcb76U,0x86e7U,0x1dc5U,0x3b8aU,0x7714U,0xee28U,
		0xcc5bU,0x88bdU,0x0171U,0x02e2U,0x05c4U,0x0b88U,0x1710U,0x2e20U,0x5c40U,0xb880U,
		0x610bU,0xc216U,0x9427U,0x3845U,0x708aU,0xe114U,0xd223U,0xb44dU,0x7891U,0xf122U,
		0xf24fU,0xf495U,0xf921U,0xe249U,0xd499U,0xb939U,0x6279U,0xc4f2U,0x99efU,0x23d5U,
		0x47aaU,0x8f54U,0x0ea3U,0x1d46U,0x3a8cU,0x7518U,0xea30U,0xc46bU,0x98ddU,0x21b1U,
		0x4362U,0x86c4U,0x1d83U,0x3b06U,0x760cU,0xec18U,0xc83bU,0x807dU,0x10f1U,0x21e2U,
		0x43c4U,0x8788U,0x1f1bU,0x3e36U,0x7c6cU,0xf8d8U,0xe1bbU,0xd37dU,0xb6f1U,0x7de9U,
		0xfbd2U,0xe7afU,0xdf55U,0xaea1U,0x4d49U,0x9a92U,0x252fU,0x4a5eU,0x94bcU,0x3973U,
		0x72e6U,0xe5ccU,0xdb93U,0xa72dU,0x5e51U,0xbca2U,0x694fU,0xd29eU,0xb537U,0x7a65U,
		0xf4caU,0xf99fU,0xe335U,0xd661U,0xbcc9U,0x6999U,0xd332U,0xb66fU,0x7cd5U,0xf9aaU,
		0xe35fU,0xd6b5U,0xbd61U,0x6ac9U,0xd592U,0xbb2fU,0x6655U,0xccaaU,0x895fU,0x02b5U,
		0x056aU,0x0ad4U,0x15a8U,0x2b50U,0x56a0U,0xad40U,0x4a8bU,0x9516U,0x3a27U,0x744eU,
		0xe89cU,0xc133U,0x926dU,0x34d1U,0x69a2U,0xd344U,0xb683U,0x7d0dU,0xfa1aU,0xe43fU,
		0xd875U,0xa0e1U,0x51c9U,0xa392U,0x572fU,0xae5eU,0x4cb7U,0x996eU,0x22d7U,0x45aeU,
		0x8b5cU,0x06b3U,0x0d66U,0x1accU,0x3598U,0x6b30U,0xd660U,0xbccbU,0x699dU,0xd33aU,
		0xb67fU,0x7cf5U,0xf9eaU,0xe3dfU,0xd7b5U,0xbf61U,0x6ec9U,0xdd92U,0xab2fU,0x4655U,
		0x8caaU,0x095fU,0x12beU,0x257cU,0x4af8U,0x95f0U,0x3bebU,0x77d6U,0xefacU,0xcf53U,
		0x8eadU,0x0d51U,0x1aa2U,0x3544U,0x6a88U,0xd510U,0xba2bU,0x645dU,0xc8baU,0x817fU,
		0x12f5U,0x25eaU,0x4bd4U,0x97a8U,0x3f5bU,0x7eb6U,0xfd6cU,0xead3U,0xc5adU,0x9b51U,
		0x26a9U,0x4d52U,0x9aa4U,0x2543U,0x4a86U,0x950cU,0x3a13U,0x7426U,0xe84cU,0xc093U,
		0x912dU,0x3251U,0x64a2U,0xc944U,0x8283U,0x150dU,0x2a1aU,0x5434U,0xa868U,0x40dbU,
		0x81b6U,0x1367U,0x26ceU,0x4d9cU,0x9b38U,0x267bU,0x4cf6U,0x99ecU,0x23d3U,0x47a6U,
		0x8f4cU,0x0e93U,0x1d26U,0x3a4cU,0x7498U,0xe930U,0xc26bU,0x94ddU,0x39b1U,0x7362U,
		0xe6c4U,0xdd83U,0xab0dU,0x4611U,0x8c22U,0x084fU,0x109eU,0x213cU,0x4278U,0x84f0U,
		0x19ebU,0x33d6U,0x67acU,0xcf58U,0x8ebbU,0x0d7dU,0x1afaU,0x35f4U,0x6be8U,0xd7d0U,
		0xbfabU,0x6f5dU,0xdebaU,0xad7fU,0x4af5U,0x95eaU,0x3bdfU,0x77beU,0xef7cU,0xcef3U,
		0x8dedU,0x0bd1U,0x17a2U,0x2f44U,0x5e88U,0xbd10U,0x6a2bU,0xd456U,0xb8a7U,0x6145U,
		0xc28aU,0x951fU,0x3a35U,0x746aU,0xe8d4U,0xc1a3U,0x934dU,0x3691U,0x6d22U,0xda44U,
		0xa483U,0x590dU,0xb21aU,0x743fU,0xe87eU,0xc0f7U,0x91e5U,0x33c1U,0x6782U,0xcf04U,
		0x8e03U,0x0c0dU,0x181aU,0x3034U,0x6068U,0xc0d0U,0x91abU,0x335dU,0x66baU,0xcd74U,
		0x8ae3U,0x05cdU,0x0b9aU,0x1734U,0x2e68U,0x5cd0U,0xb9a0U,0x634bU,0xc696U,0x9d27U,
		0x2a45U,0x548aU,0xa914U,0x4223U,0x8446U,0x1887U,0x310eU,0x621cU,0xc438U,0x987bU,
		0x20fdU,0x41faU,0x83f4U,0x17e3U,0x2fc6U,0x5f8cU,0xbf18U,0x6e3bU,0xdc76U,0xa8e7U,
		0x41c5U,0x838aU,0x171fU,0x2e3eU,0x5c7cU,0xb8f8U,0x61fbU,0xc3f6U,0x97e7U,0x3fc5U,
		0x7f8aU,0xff14U,0xee23U,0xcc4dU,0x8891U,0x0129U,0x0252U,0x04a4U,0x0948U,0x1290U,
		0x2520U,0x4a40U,0x9480U,0x390bU,0x7216U,0xe42cU,0xd853U,0xa0adU,0x5151U,0xa2a2U,
		0x554fU,0xaa9eU,0x4537U,0x8a6eU,0x04d7U,0x09aeU,0x135cU,0x26b8U,0x4d70U,0x9ae0U,
		0x25cbU,0x4b96U,0x972cU,0x3e53U,0x7ca6U,0xf94cU,0xe293U,0xd52dU,0xba51U,0x64a9U,
		0xc952U,0x82afU,0x1555U,0x2aaaU,0x5554U,0xaaa8U,0x455bU,0x8ab6U,0x0567U,0x0aceU,
		0x159cU,0x2b38U,0x5670U,0xace0U,0x49cbU,0x9396U,0x3727U,0x6e4eU,0xdc9cU,0xa933U,
		0x426dU,0x84daU,0x19bfU,0x337eU,0x66fcU,0xcdf8U,0x8bfbU,0x07fdU,0x0ffaU,0x1ff4U,
		0x3fe8U,0x7fd0U,0xffa0U,0xef4bU,0xce9dU,0x8d31U,0x0a69U,0x14d2U,0x29a4U,0x5348U,
		0xa690U,0x5d2bU,0xba56U,0x64a7U,0xc94eU,0x8297U,0x1525U,0x2a4aU,0x5494U,0xa928U,
		0x425bU,0x84b6U,0x1967U,0x32ceU,0x659cU,0xcb38U,0x867bU,0x1cfdU,0x39faU,0x73f4U,
		0xe7e8U,0xdfdbU,0xafbdU,0x4f71U,0x9ee2U,0x2dcfU,0x5b9eU,0xb73cU,0x7e73U,0xfce6U,
		0xe9c7U,0xc385U,0x9701U,0x3e09U,0x7c12U,0xf824U,0xe043U,0xd08dU,0xb111U,0x7229U,
		0xe452U,0xd8afU,0xa155U,0x52a1U,0xa542U,0x5a8fU,0xb51eU,0x7a37U,0xf46eU,0xf8d7U,
		0xe1a5U,0xd341U,0xb689U,0x7d19U,0xfa32U,0xe46fU,0xd8d5U,0xa1a1U,0x5349U,0xa692U,
		0x5d2fU,0xba5eU,0x64b7U,0xc96eU,0x82d7U,0x15a5U,0x2b4aU,0x5694U,0xad28U,0x4a5bU,
		0x94b6U,0x3967U,0x72ceU,0xe59cU,0xdb33U,0xa66dU,0x5cd1U,0xb9a2U,0x634fU,0xc69eU,
		0x9d37U,0x2a65U,0x54caU,0xa994U,0x4323U,0x8646U,0x1c87U,0x390eU,0x721cU,0xe438U,
		0xd87bU,0xa0fdU,0x51f1U,0xa3e2U,0x57cfU,0xaf9eU,0x4f37U,0x9e6eU,0x2cd7U,0x59aeU,
		0xb35cU,0x76b3U,0xed66U,0xcac7U,0x8585U,0x1b01U,0x3602U,0x6c04U,0xd808U,0xa01bU,
		0x503dU,0xa07aU,0x50ffU,0xa1feU,0x53f7U,0xa7eeU,0x5fd7U,0xbfaeU,0x6f57U,0xdeaeU,
		0xad57U,0x4aa5U,0x954aU,0x3a9fU,0x753eU,0xea7cU,0xc4f3U,0x99edU,0x23d1U,0x47a2U,
		0x8f44U,0x0e83U,0x1d06U,0x3a0cU,0x7418U,0xe830U,0xc06bU,0x90ddU,0x31b1U,0x6362U,
		0xc6c4U,0x9d83U,0x2b0dU,0x561aU,0xac34U,0x4863U,0x90c6U,0x3187U,0x630eU,0xc61cU,
		0x9c33U,0x286dU,0x50daU,0xa1b4U,0x5363U,0xa6c6U,0x5d87U,0xbb0eU,0x6617U,0xcc2eU,
		0x8857U,0x00a5U,0x014aU,0x0294U,0x0528U,0x0a50U,0x14a0U,0x2940U,0x5280U,0xa500U,
		0x5a0bU,0xb416U,0x7827U,0xf04eU,0xf097U,0xf125U,0xf241U,0xf489U,0xf919U,0xe239U,
		0xd479U,0xb8f9U,0x61f9U,0xc3f2U,0x97efU,0x3fd5U,0x7faaU,0xff54U,0xeea3U,0xcd4dU,
		0x8a91U,0x0529U,0x0a52U,0x14a4U,0x2948U,0x5290U,0xa520U,0x5a4bU,0xb496U,0x7927U,
		0xf24eU,0xf497U,0xf925U,0xe241U,0xd489U,0xb919U,0x6239U,0xc472U,0x98efU,0x21d5U,
		0x43aaU,0x8754U,0x1ea3U,0x3d46U,0x7a8cU,0xf518U,0xfa3bU,0xe47dU,0xd8f1U,0xa1e9U,
		0x53d9U,0xa7b2U,0x5f6fU,0xbedeU,0x6db7U,0xdb6eU,0xa6d7U,0x5da5U,0xbb4aU,0x669fU,
		0xcd3eU,0x8a77U,0x04e5U,0x09caU,0x1394U,0x2728U,0x4e50U,0x9ca0U,0x294bU,0x5296U,
		0xa52cU,0x5a53U,0xb4a6U,0x7947U,0xf28eU,0xf517U,0xfa25U,0xe441U,0xd889U,0xa119U,
		0x5239U,0xa472U,0x58efU,0xb1deU,0x73b7U,0xe76eU,0xded7U,0xada5U,0x4b41U,0x9682U,
		0x3d0fU,0x7a1eU,0xf43cU,0xf873U,0xe0edU,0xd1d1U,0xb3a9U,0x7759U,0xeeb2U,0xcd6fU,
		0x8ad5U,0x05a1U,0x0b42U,0x1684U,0x2d08U,0x5a10U,0xb420U,0x784bU,0xf096U,0xf127U,
		0xf245U,0xf481U,0xf909U,0xe219U,0xd439U,0xb879U,0x60f9U,0xc1f2U,0x93efU,0x37d5U,
		0x6faaU,0xdf54U,0xaea3U,0x4d4dU,0x9a9aU,0x253fU,0x4a7eU,0x94fcU,0x39f3U,0x73e6U,
		0xe7ccU,0xdf93U,0xaf2dU,0x4e51U,0x9ca2U,0x294fU,0x529eU,0xa53cU,0x5a73U,0xb4e6U,
		0x79c7U,0xf38eU,0xf717U,0xfe25U,0xec41U,0xc889U,0x8119U,0x1239U,0x2472U,0x48e4U,
		0x91c8U,0x339bU,0x6736U,0xce6cU,0x8cd3U,0x09adU,0x135aU,0x26b4U,0x4d68U,0x9ad0U,
		0x25abU,0x4b56U,0x96acU,0x3d53U,0x7aa6U,0xf54cU,0xfa93U,0xe52dU,0xda51U,0xa4a9U,
		0x5959U,0xb2b2U,0x756fU,0xeadeU,0xc5b7U,0x9b65U,0x26c1U,0x4d82U,0x9b04U,0x2603U,
		0x4c06U,0x980cU,0x2013U,0x4026U,0x804cU,0x1093U,0x2126U,0x424cU,0x8498U,0x193bU,
		0x3276U,0x64ecU,0xc9d8U,0x83bbU,0x177dU,0x2efaU,0x5df4U,0xbbe8U,0x67dbU,0xcfb6U,
		0x8f67U,0x0ec5U,0x1d8aU,0x3b14U,0x7628U,0xec50U,0xc8abU,0x815dU,0x12b1U,0x2562U,
		0x4ac4U,0x9588U,0x3b1bU,0x7636U,0xec6cU,0xc8d3U,0x81adU,0x1351U,0x26a2U,0x4d44U,
		0x9a88U,0x251bU,0x4a36U,0x946cU,0x38d3U,0x71a6U,0xe34cU,0xd693U,0xbd2dU,0x6a51U,
		0xd4a2U,0xb94fU,0x6295U,0xc52aU,0x9a5fU,0x24b5U,0x496aU,0x92d4U,0x35a3U,0x6b46U,
		0xd68cU,0xbd13U,0x6a2dU,0xd45aU,0xb8bfU,0x6175U,0xc2eaU,0x95dfU,0x3bb5U,0x776aU,
		0xeed4U,0xcda3U,0x8b4dU,0x0691U,0x0d22U,0x1a44U,0x3488U,0x6910U,0xd220U,0xb44bU,
		0x789dU,0xf13aU,0xf27fU,0xf4f5U,0xf9e1U,0xe3c9U,0xd799U,0xbf39U,0x6e79U,0xdcf2U,
		0xa9efU,0x43d5U,0x87aaU,0x1f5fU,0x3ebeU,0x7d7cU,0xfaf8U,0xe5fbU,0xdbfdU,0xa7f1U,
		0x5fe9U,0xbfd2U,0x6fafU,0xdf5eU,0xaeb7U,0x4d65U,0x9acaU,0x259fU,0x4b3eU,0x967cU,
		0x3cf3U,0x79e6U,0xf3ccU,0xf793U,0xff2dU,0xee51U,0xcca9U,0x8959U,0x02b9U,0x0572U,
		0x0ae4U,0x15c8U,0x2b90U,0x5720U,0xae40U,0x4c8bU,0x9916U,0x2227U,0x444eU,0x889cU,
		0x0133U,0x0266U,0x04ccU,0x0998U,0x1330U,0x2660U,0x4cc0U,0x9980U,0x230bU,0x4616U,
		0x8c2cU,0x0853U,0x10a6U,0x214cU,0x4298U,0x8530U,0x1a6bU,0x34d6U,0x69acU,0xd358U,
		0xb6bbU,0x7d7dU,0xfafaU,0xe5ffU,0xdbf5U,0xa7e1U,0x5fc9U,0xbf92U,0x6f2fU,0xde5eU,
		0xacb7U,0x4965U,0x92caU,0x359fU,0x6b3eU,0xd67cU,0xbcf3U,0x69edU,0xd3daU,0xb7bfU,
		0x7f75U,0xfeeaU,0xeddfU,0xcbb5U,0x8761U,0x1ec9U,0x3d92U,0x7b24U,0xf648U,0xfc9bU,
		0xe93dU,0xc271U,0x94e9U,0x39d9U,0x73b2U,0xe764U,0xdec3U,0xad8dU,0x4b11U,0x9622U,
		0x3c4fU,0x789eU,0xf13cU,0xf273U,0xf4edU,0xf9d1U,0xe3a9U,0xd759U,0xbeb9U,0x6d79U,
		0xdaf2U,0xa5efU,0x5bd5U,0xb7aaU,0x7f5fU,0xfebeU,0xed77U,0xcae5U,0x85c1U,0x1b89U,
		0x3712U,0x6e24U,0xdc48U,0xa89bU,0x413dU,0x827aU,0x14ffU,0x29feU,0x53fcU,0xa7f8U,
		0x5ffbU,0xbff6U,0x6fe7U,0xdfceU,0xaf97U,0x4f25U,0x9e4aU,0x2c9fU,0x593eU,0xb27cU,
		0x74f3U,0xe9e6U,0xc3c7U,0x9785U,0x3f01U,0x7e02U,0xfc04U,0xe803U,0xc00dU,0x9011U,
		0x3029U,0x6052U,0xc0a4U,0x9143U,0x328dU,0x651aU,0xca34U,0x8463U,0x18cdU,0x319aU,
		0x6334U,0xc668U,0x9cdbU,0x29bdU,0x537aU,0xa6f4U,0x5de3U,0xbbc6U,0x6787U,0xcf0eU,
		0x8e17U,0x0c25U,0x184aU,0x3094U,0x6128U,0xc250U,0x94abU,0x395dU,0x72baU,0xe574U,
		0xdae3U,0xa5cdU,0x5b91U,0xb722U,0x7e4fU,0xfc9eU,0xe937U,0xc265U,0x94c1U,0x3989U,
		0x7312U,0xe624U,0xdc43U,0xa88dU,0x4111U,0x8222U,0x144fU,0x289eU,0x513cU,0xa278U,
		0x54fbU,0xa9f6U,0x43e7U,0x87ceU,0x1f97U,0x3f2eU,0x7e5cU,0xfcb8U,0xe97bU,0xc2fdU,
		0x95f1U,0x3be9U,0x77d2U,0xefa4U,0xcf43U,0x8e8dU,0x0d11U,0x1a22U,0x3444U,0x6888U,
		0xd110U,0xb22bU,0x745dU,0xe8baU,0xc17fU,0x92f5U,0x35e1U,0x6bc2U,0xd784U,0xbf03U,
		0x6e0dU,0xdc1aU,0xa83fU,0x4075U,0x80eaU,0x11dfU,0x23beU,0x477cU,0x8ef8U,0x0dfbU,
		0x1bf6U,0x37ecU,0x6fd8U,0xdfb0U,0xaf6bU,0x4eddU,0x9dbaU,0x2b7fU,0x56feU,0xadfcU,
		0x4bf3U,0x97e6U,0x3fc7U,0x7f8eU,0xff1cU,0xee33U,0xcc6dU,0x88d1U,0x01a9U,0x0352U,
		0x06a4U,0x0d48U,0x1a90U,0x3520U,0x6a40U,0xd480U,0xb90bU,0x621dU,0xc43aU,0x987fU,
		0x20f5U,0x41eaU,0x83d4U,0x17a3U,0x2f46U,0x5e8cU,0xbd18U,0x6a3bU,0xd476U,0xb8e7U,
		0x61c5U,0xc38aU,0x971fU,0x3e35U,0x7c6aU,0xf8d4U,0xe1a3U,0xd34dU,0xb691U,0x7d29U,
		0xfa52U,0xe4afU,0xd955U,0xa2a1U,0x5549U,0xaa92U,0x452fU,0x8a5eU,0x04b7U,0x096eU,
		0x12dcU,0x25b8U,0x4b70U,0x96e0U,0x3dcbU,0x7b96U,0xf72cU,0xfe53U,0xecadU,0xc951U,
		0x82a9U,0x1559U,0x2ab2U,0x5564U,0xaac8U,0x459bU,0x8b36U,0x0667U,0x0cceU,0x199cU,
		0x3338U,0x6670U,0xcce0U,0x89cbU,0x039dU,0x073aU,0x0e74U,0x1ce8U,0x39d0U,0x73a0U,
		0xe740U,0xde8bU,0xad1dU,0x4a31U,0x9462U,0x38cfU,0x719eU,0xe33cU,0xd673U,0xbcedU,
		0x69d1U,0xd3a2U,0xb74fU,0x7e95U,0xfd2aU,0xea5fU,0xc4b5U,0x9961U,0x22c9U,0x4592U,
		0x8b24U,0x0643U,0x0c86U,0x190cU,0x3218U,0x6430U,0xc860U,0x80cbU,0x119dU,0x233aU,
		0x4674U,0x8ce8U,0x09dbU,0x13b6U,0x276cU,0x4ed8U,0x9db0U,0x2b6bU,0x56d6U,0xadacU,
		0x4b53U,0x96a6U,0x3d47U,0x7a8eU,0xf51cU,0xfa33U,0xe46dU,0xd8d1U,0xa1a9U,0x5359U,
		0xa6b2U,0x5d6fU,0xbadeU,0x65b7U,0xcb6eU,0x86d7U,0x1da5U,0x3b4aU,0x7694U,0xed28U,
		0xca5bU,0x84bdU,0x1971U,0x32e2U,0x65c4U,0xcb88U,0x871bU,0x1e3dU,0x3c7aU,0x78f4U,
		0xf1e8U,0xf3dbU,0xf7bdU,0xff71U,0xeee9U,0xcdd9U,0x8bb9U,0x0779U,0x0ef2U,0x1de4U,
		0x3bc8U,0x7790U,0xef20U,0xce4bU,0x8c9dU,0x0931U,0x1262U,0x24c4U,0x4988U,0x9310U,
		0x362bU,0x6c56U,0xd8acU,0xa153U,0x52adU,0xa55aU,0x5abfU,0xb57eU,0x7af7U,0xf5eeU,
		0xfbd7U,0xe7a5U,0xdf41U,0xae89U,0x4d19U,0x9a32U,0x246fU,0x48deU,0x91bcU,0x3373U,
		0x66e6U,0xcdccU,0x8b93U,0x072dU,0x0e5aU,0x1cb4U,0x3968U,0x72d0U,0xe5a0U,0xdb4bU,
		0xa69dU,0x5d31U,0xba62U,0x64cfU,0xc99eU,0x8337U,0x1665U,0x2ccaU,0x5994U,0xb328U,
		0x765bU,0xecb6U,0xc967U,0x82c5U,0x1581U,0x2b02U,0x5604U,0xac08U,0x481bU,0x9036U,
		0x3067U,0x60ceU,0xc19cU,0x9333U,0x366dU,0x6cdaU,0xd9b4U,0xa363U,0x56cdU,0xad9aU,
		0x4b3fU,0x967eU,0x3cf7U,0x79eeU,0xf3dcU,0xf7b3U,0xff6dU,0xeed1U,0xcda9U,0x8b59U,
		0x06b9U,0x0d72U,0x1ae4U,0x35c8U,0x6b90U,0xd720U,0xbe4bU,0x6c9dU,0xd93aU,0xa27fU,
		0x54f5U,0xa9eaU,0x43dfU,0x87beU,0x1f77U,0x3eeeU,0x7ddcU,0xfbb8U,0xe77bU,0xdefdU,
		0xadf1U,0x4be9U,0x97d2U,0x3fafU,0x7f5eU,0xfebcU,0xed73U,0xcaedU,0x85d1U,0x1ba9U,
		0x3752U,0x6ea4U,0xdd48U,0xaa9bU,0x453dU,0x8a7aU,0x04ffU,0x09feU,0x13fcU,0x27f8U,
		0x4ff0U,0x9fe0U,0x2fcbU,0x5f96U,0xbf2cU,0x6e53U,0xdca6U,0xa947U,0x4285U,0x850aU,
		0x1a1fU,0x343eU,0x687cU,0xd0f8U,0xb1fbU,0x73fdU,0xe7faU,0xdfffU,0xaff5U,0x4fe1U,
		0x9fc2U,0x2f8fU,0x5f1eU,0xbe3cU,0x6c73U,0xd8e6U,0xa1c7U,0x5385U,0xa70aU,0x5e1fU,
		0xbc3eU,0x6877U,0xd0eeU,0xb1d7U,0x73a5U,0xe74aU,0xde9fU,0xad35U,0x4a61U,0x94c2U,
		0x398fU,0x731eU,0xe63cU,0xdc73U,0xa8edU,0x41d1U,0x83a2U,0x174fU,0x2e9eU,0x5d3cU,
		0xba78U,0x64fbU,0xc9f6U,0x83e7U,0x17c5U,0x2f8aU,0x5f14U,0xbe28U,0x6c5bU,0xd8b6U,
		0xa167U,0x52c5U,0xa58aU,0x5b1fU,0xb63eU,0x7c77U,0xf8eeU,0xe1d7U,0xd3a5U,0xb741U,
		0x7e89U,0xfd12U,0xea2fU,0xc455U,0x98a1U,0x2149U,0x4292U,0x8524U,0x1a43U,0x3486U,
		0x690cU,0xd218U,0xb43bU,0x787dU,0xf0faU,0xf1ffU,0xf3f5U,0xf7e1U,0xffc9U,0xef99U,
		0xcf39U,0x8e79U,0x0cf9U,0x19f2U,0x33e4U,0x67c8U,0xcf90U,0x8f2bU,0x0e5dU,0x1cbaU,
		0x3974U,0x72e8U,0xe5d0U,0xdbabU,0xa75dU,0x5eb1U,0xbd62U,0x6acfU,0xd59eU,0xbb37U,
		0x6665U,0xcccaU,0x899fU,0x0335U,0x066aU,0x0cd4U,0x19a8U,0x3350U,0x66a0U,0xcd40U,
		0x8a8bU,0x051dU,0x0a3aU,0x1474U,0x28e8U,0x51d0U,0xa3a0U,0x574bU,0xae96U,0x4d27U,
		0x9a4eU,0x2497U,0x492eU,0x925cU,0x34b3U,0x6966U,0xd2ccU,0xb593U,0x7b2dU,0xf65aU,
		0xfcbfU,0xe975U,0xc2e1U,0x95c9U,0x3b99U,0x7732U,0xee64U,0xccc3U,0x898dU,0x0311U,
		0x0622U,0x0c44U,0x1888U,0x3110U,0x6220U,0xc440U,0x988bU,0x211dU,0x423aU,0x8474U,
		0x18e3U,0x31c6U,0x638cU,0xc718U,0x9e3bU,0x2c7dU,0x58faU,0xb1f4U,0x73e3U,0xe7c6U,
		0xdf87U,0xaf05U,0x4e01U,0x9c02U,0x280fU,0x501eU,0xa03cU,0x5073U,0xa0e6U,0x51c7U,
		0xa38eU,0x5717U,0xae2eU,0x4c57U,0x98aeU,0x2157U,0x42aeU,0x855cU,0x1ab3U,0x3566U,
		0x6accU,0xd598U,0xbb3bU,0x667dU,0xccfaU,0x89ffU,0x03f5U,0x07eaU,0x0fd4U,0x1fa8U,
		0x3f50U,0x7ea0U,0xfd40U,0xea8bU,0xc51dU,0x9a31U,0x2469U,0x48d2U,0x91a4U,0x3343U,
		0x6686U,0xcd0cU,0x8a13U,0x042dU,0x085aU,0x10b4U,0x2168U,0x42d0U,0x85a0U,0x1b4bU,
		0x3696U,0x6d2cU,0xda58U,0xa4bbU,0x597dU,0xb2faU,0x75ffU,0xebfeU,0xc7f7U,0x9fe5U,
		0x2fc1U,0x5f82U,0xbf04U,0x6e03U,0xdc06U,0xa807U,0x4005U,0x800aU,0x101fU,0x203eU,
		0x407cU,0x80f8U,0x11fbU,0x23f6U,0x47ecU,0x8fd8U,0x0fbbU,0x1f76U,0x3eecU,0x7dd8U,
		0xfbb0U,0xe76bU,0xdeddU,0xadb1U,0x4b69U,0x96d2U,0x3dafU,0x7b5eU,0xf6bcU,0xfd73U,
		0xeaedU,0xc5d1U,0x9ba9U,0x2759U,0x4eb2U,0x9d64U,0x2ac3U,0x5586U,0xab0cU,0x4613U,
		0x8c26U,0x0847U,0x108eU,0x211cU,0x4238U,0x8470U,0x18ebU,0x31d6U,0x63acU,0xc758U,
		0x9ebbU,0x2d7dU,0x5afaU,0xb5f4U,0x7be3U,0xf7c6U,0xff87U,0xef05U,0xce01U,0x8c09U,
		0x0819U,0x1032U,0x2064U,0x40c8U,0x8190U,0x132bU,0x2656U,0x4cacU,0x9958U,0x22bbU,
		0x4576U,0x8aecU,0x05d3U,0x0ba6U,0x174cU,0x2e98U,0x5d30U,0xba60U,0x64cbU,0xc996U,
		0x8327U,0x1645U,0x2c8aU,0x5914U,0xb228U,0x745bU,0xe8b6U,0xc167U,0x92c5U,0x3581U,
		0x6b02U,0xd604U,0xbc03U,0x680dU,0xd01aU,0xb03fU,0x7075U,0xe0eaU,0xd1dfU,0xb3b5U,
		0x7761U,0xeec2U,0xcd8fU,0x8b15U,0x0621U,0x0c42U,0x1884U,0x3108U,0x6210U,0xc420U,
		0x984bU,0x209dU,0x413aU,0x8274U,0x14e3U,0x29c6U,0x538cU,0xa718U,0x5e3bU,0xbc76U,
		0x68e7U,0xd1ceU,0xb397U,0x7725U,0xee4aU,0xcc9fU,0x8935U,0x0261U,0x04c2U,0x0984U,
		0x1308U,0x2610U,0x4c20U,0x9840U,0x208bU,0x4116U,0x822cU,0x1453U,0x28a6U,0x514cU,
		0xa298U,0x553bU,0xaa76U,0x44e7U,0x89ceU,0x0397U,0x072eU,0x0e5cU,0x1cb8U,0x3970U,
		0x72e0U,0xe5c0U,0xdb8bU,0xa71dU,0x5e31U,0xbc62U,0x68cfU,0xd19eU,0xb337U,0x7665U,
		0xeccaU,0xc99fU,0x8335U,0x1661U,0x2cc2U,0x5984U,0xb308U,0x761bU,0xec36U,0xc867U,
		0x80c5U,0x1181U,0x2302U,0x4604U,0x8c08U,0x081bU,0x1036U,0x206cU,0x40d8U,0x81b0U,
		0x136bU,0x26d6U,0x4dacU,0x9b58U,0x26bbU,0x4d76U,0x9aecU,0x25d3U,0x4ba6U,0x974cU,
		0x3e93U,0x7d26U,0xfa4cU,0xe493U,0xd92dU,0xa251U,0x54a9U,0xa952U,0x42afU,0x855eU,
		0x1ab7U,0x356eU,0x6adcU,0xd5b8U,0xbb7bU,0x66fdU,0xcdfaU,0x8bffU,0x07f5U,0x0feaU,
		0x1fd4U,0x3fa8U,0x7f50U,0xfea0U,0xed4bU,0xca9dU,0x8531U,0x1a69U,0x34d2U,0x69a4U,
		0xd348U,0xb69bU,0x7d3dU,0xfa7aU,0xe4ffU,0xd9f5U,0xa3e1U,0x57c9U,0xaf92U,0x4f2fU,
		0x9e5eU,0x2cb7U,0x596eU,0xb2dcU,0x75b3U,0xeb66U,0xc6c7U,0x9d85U,0x2b01U,0x5602U,
		0xac04U,0x4803U,0x9006U,0x3007U,0x600eU,0xc01cU,0x9033U,0x306dU,0x60daU,0xc1b4U,
		0x9363U,0x36cdU,0x6d9aU,0xdb34U,0xa663U,0x5ccdU,0xb99aU,0x633fU,0xc67eU,0x9cf7U,
		0x29e5U,0x53caU,0xa794U,0x5f23U,0xbe46U,0x6c87U,0xd90eU,0xa217U,0x5425U,0xa84aU,
		0x409fU,0x813eU,0x1277U,0x24eeU,0x49dcU,0x93b8U,0x377bU,0x6ef6U,0xddecU,0xabd3U,
		0x47adU,0x8f5aU,0x0ebfU,0x1d7eU,0x3afcU,0x75f8U,0xebf0U,0xc7ebU,0x9fddU,0x2fb1U,
		0x5f62U,0xbec4U,0x6d83U,0xdb06U,0xa607U,0x5c05U,0xb80aU,0x601fU,0xc03eU,0x9077U,
		0x30e5U,0x61caU,0xc394U,0x9723U,0x3e4dU,0x7c9aU,0xf934U,0xe263U,0xd4cdU,0xb991U,
		0x6329U,0xc652U,0x9cafU,0x2955U,0x52aaU,0xa554U,0x5aa3U,0xb546U,0x7a87U,0xf50eU,
		0xfa17U,0xe425U,0xd841U,0xa089U,0x5119U,0xa232U,0x546fU,0xa8deU,0x41b7U,0x836eU,
		0x16d7U,0x2daeU,0x5b5cU,0xb6b8U,0x7d7bU,0xfaf6U,0xe5e7U,0xdbc5U,0xa781U,0x5f09U,
		0xbe12U,0x6c2fU,0xd85eU,0xa0b7U,0x5165U,0xa2caU,0x559fU,0xab3eU,0x4677U,0x8ceeU,
		0x09d7U,0x13aeU,0x275cU,0x4eb8U,0x9d70U,0x2aebU,0x55d6U,0xabacU,0x4753U,0x8ea6U,
		0x0d47U,0x1a8eU,0x351cU,0x6a38U,0xd470U,0xb8ebU,0x61ddU,0xc3baU,0x977fU,0x3ef5U,
		0x7deaU,0xfbd4U,0xe7a3U,0xdf4dU,0xae91U,0x4d29U,0x9a52U,0x24afU,0x495eU,0x92bcU,
		0x3573U,0x6ae6U,0xd5ccU,0xbb93U,0x672dU,0xce5aU,0x8cbfU,0x0975U,0x12eaU,0x25d4U,
		0x4ba8U,0x9750U,0x3eabU,0x7d56U,0xfaacU,0xe553U,0xdaadU,0xa551U,0x5aa9U,0xb552U,
		0x7aafU,0xf55eU,0xfab7U,0xe565U,0xdac1U,0xa589U,0x5b19U,0xb632U,0x7c6fU,0xf8deU,
		0xe1b7U,0xd365U,0xb6c1U,0x7d89U,0xfb12U,0xe62fU,0xdc55U,0xa8a1U,0x4149U,0x8292U,
		0x152fU,0x2a5eU,0x54bcU,0xa978U,0x42fbU,0x85f6U,0x1be7U,0x37ceU,0x6f9cU,0xdf38U,
		0xae7bU,0x4cfdU,0x99faU,0x23ffU,0x47feU,0x8ffcU,0x0ff3U,0x1fe6U,0x3fccU,0x7f98U,
		0xff30U,0xee6bU,0xccddU,0x89b1U,0x0369U,0x06d2U,0x0da4U,0x1b48U,0x3690U,0x6d20U,
		0xda40U,0xa48bU,0x591dU,0xb23aU,0x747fU,0xe8feU,0xc1f7U,0x93e5U,0x37c1U,0x6f82U,
		0xdf04U,0xae03U,0x4c0dU,0x981aU,0x203fU,0x407eU,0x80fcU,0x11f3U,0x23e6U,0x47ccU,
		0x8f98U,0x0f3bU,0x1e76U,0x3cecU,0x79d8U,0xf3b0U,0xf76bU,0xfeddU,0xedb1U,0xcb69U,
		0x86d9U,0x1db9U,0x3b72U,0x76e4U,0xedc8U,0xcb9bU,0x873dU,0x1e71U,0x3ce2U,0x79c4U,
		0xf388U,0xf71bU,0xfe3dU,0xec71U,0xc8e9U,0x81d9U,0x13b9U,0x2772U,0x4ee4U,0x9dc8U,
		0x2b9bU,0x5736U,0xae6cU,0x4cd3U,0x99a6U,0x2347U,0x468eU,0x8d1cU,0x0a33U,0x1466U,
		0x28ccU,0x5198U,0xa330U,0x566bU,0xacd6U,0x49a7U,0x934eU,0x3697U,0x6d2eU,0xda5cU,
		0xa4b3U,0x596dU,0xb2daU,0x75bfU,0xeb7eU,0xc6f7U,0x9de5U,0x2bc1U,0x5782U,0xaf04U,
		0x4e03U,0x9c06U,0x2807U,0x500eU,0xa01cU,0x5033U,0xa066U,0x50c7U,0xa18eU,0x5317U,
		0xa62eU,0x5c57U,0xb8aeU,0x6157U,0xc2aeU,0x9557U,0x3aa5U,0x754aU,0xea94U,0xc523U,
		0x9a4dU,0x2491U,0x4922U,0x9244U,0x3483U,0x6906U,0xd20cU,0xb413U,0x782dU,0xf05aU,
		0xf0bfU,0xf175U,0xf2e1U,0xf5c9U,0xfb99U,0xe739U,0xde79U,0xacf9U,0x49f9U,0x93f2U,
		0x37efU,0x6fdeU,0xdfbcU,0xaf73U,0x4eedU,0x9ddaU,0x2bbfU,0x577eU,0xaefcU,0x4df3U,
		0x9be6U,0x27c7U,0x4f8eU,0x9f1cU,0x2e33U,0x5c66U,0xb8ccU,0x6193U,0xc326U,0x9647U,
		0x3c85U,0x790aU,0xf214U,0xf423U,0xf84dU,0xe091U,0xd129U,0xb259U,0x74b9U,0xe972U,
		0xc2efU,0x95d5U,0x3ba1U,0x7742U,0xee84U,0xcd03U,0x8a0dU,0x0411U,0x0822U,0x1044U,
		0x2088U,0x4110U,0x8220U,0x144bU,0x2896U,0x512cU,0xa258U,0x54bbU,0xa976U,0x42e7U,
		0x85ceU,0x1b97U,0x372eU,0x6e5cU,0xdcb8U,0xa97bU,0x42fdU,0x85faU,0x1bffU,0x37feU,
		0x6ffcU,0xdff8U,0xaffbU,0x4ffdU,0x9ffaU,0x2fffU,0x5ffeU,0xbffcU,0x6ff3U,0xdfe6U,
		0xafc7U,0x4f85U,0x9f0aU,0x2e1fU,0x5c3eU,0xb87cU,0x60f3U,0xc1e6U,0x93c7U,0x3785U,
		0x6f0aU,0xde14U,0xac23U,0x484dU,0x909aU,0x313fU,0x627eU,0xc4fcU,0x99f3U,0x23edU,
		0x47daU,0x8fb4U,0x0f63U,0x1ec6U,0x3d8cU,0x7b18U,0xf630U,0xfc6bU,0xe8ddU,0xc1b1U,
		0x9369U,0x36d9U,0x6db2U,0xdb64U,0xa6c3U,0x5d8dU,0xbb1aU,0x663fU,0xcc7eU,0x88f7U,
		0x01e5U,0x03caU,0x0794U,0x0f28U,0x1e50U,0x3ca0U,0x7940U,0xf280U,0xf50bU,0xfa1dU,
		0xe431U,0xd869U,0xa0d9U,0x51b9U,0xa372U,0x56efU,0xaddeU,0x4bb7U,0x976eU,0x3ed7U,
		0x7daeU,0xfb5cU,0xe6b3U,0xdd6dU,0xaad1U,0x45a9U,0x8b52U,0x06afU,0x0d5eU,0x1abcU,
		0x3578U,0x6af0U,0xd5e0U,0xbbcbU,0x679dU,0xcf3aU,0x8e7fU,0x0cf5U,0x19eaU,0x33d4U,
		0x67a8U,0xcf50U,0x8eabU,0x0d5dU,0x1abaU,0x3574U,0x6ae8U,0xd5d0U,0xbbabU,0x675dU,
		0xcebaU,0x8d7fU,0x0af5U,0x15eaU,0x2bd4U,0x57a8U,0xaf50U,0x4eabU,0x9d56U,0x2aa7U,
		0x554eU,0xaa9cU,0x4533U,0x8a66U,0x04c7U,0x098eU,0x131cU,0x2638U,0x4c70U,0x98e0U,
		0x21cbU,0x4396U,0x872cU,0x1e53U,0x3ca6U,0x794cU,0xf298U,0xf53bU,0xfa7dU,0xe4f1U,
		0xd9e9U,0xa3d9U,0x57b9U,0xaf72U,0x4eefU,0x9ddeU,0x2bb7U,0x576eU,0xaedcU,0x4db3U,
		0x9b66U,0x26c7U,0x4d8eU,0x9b1cU,0x2633U,0x4c66U,0x98ccU,0x2193U,0x4326U,0x864cU,
		0x1c93U,0x3926U,0x724cU,0xe498U,0xd93bU,0xa27dU,0x54f1U,0xa9e2U,0x43cfU,0x879eU,
		0x1f37U,0x3e6eU,0x7cdcU,0xf9b8U,0xe37bU,0xd6fdU,0xbdf1U,0x6be9U,0xd7d2U,0xbfafU,
		0x6f55U,0xdeaaU,0xad5fU,0x4ab5U,0x956aU,0x3adfU,0x75beU,0xeb7cU,0xc6f3U,0x9dedU,
		0x2bd1U,0x57a2U,0xaf44U,0x4e83U,0x9d06U,0x2a07U,0x540eU,0xa81cU,0x4033U,0x8066U,
		0x10c7U,0x218eU,0x431cU,0x8638U,0x1c7bU,0x38f6U,0x71ecU,0xe3d8U,0xd7bbU,0xbf7dU,
		0x6ef1U,0xdde2U,0xabcfU,0x4795U,0x8f2aU,0x0e5fU,0x1cbeU,0x397cU,0x72f8U,0xe5f0U,
		0xdbebU,0xa7ddU,0x5fb1U,0xbf62U,0x6ecfU,0xdd9eU,0xab37U,0x4665U,0x8ccaU,0x099fU,
		0x133eU,0x267cU,0x4cf8U,0x99f0U,0x23ebU,0x47d6U,0x8facU,0x0f53U,0x1ea6U,0x3d4cU,
		0x7a98U,0xf530U,0xfa6bU,0xe4ddU,0xd9b1U,0xa369U,0x56d9U,0xadb2U,0x4b6fU,0x96deU,
		0x3db7U,0x7b6eU,0xf6dcU,0xfdb3U,0xeb6dU,0xc6d1U,0x9da9U,0x2b59U,0x56b2U,0xad64U,
		0x4ac3U,0x9586U,0x3b07U,0x760eU,0xec1cU,0xc833U,0x806dU,0x10d1U,0x21a2U,0x4344U,
		0x8688U,0x1d1bU,0x3a36U,0x746cU,0xe8d8U,0xc1bbU,0x937dU,0x36f1U,0x6de2U,0xdbc4U,
		0xa783U,0x5f0dU,0xbe1aU,0x6c3fU,0xd87eU,0xa0f7U,0x51e5U,0xa3caU,0x579fU,0xaf3eU,
		0x4e77U,0x9ceeU,0x29d7U,0x53aeU,0xa75cU,0x5eb3U,0xbd66U,0x6ac7U,0xd58eU,0xbb17U,
		0x6625U,0xcc4aU,0x889fU,0x0135U,0x026aU,0x04d4U,0x09a8U,0x1350U,0x26a0U,0x4d40U,
		0x9a80U,0x250bU,0x4a16U,0x942cU,0x3853U,0x70a6U,0xe14cU,0xd293U,0xb52dU,0x7a51U,
		0xf4a2U,0xf94fU,0xe295U,0xd521U,0xba49U,0x6499U,0xc932U,0x826fU,0x14d5U,0x29aaU,
		0x5354U,0xa6a8U,0x5d5bU,0xbab6U,0x6567U,0xcaceU,0x8597U,0x1b25U,0x364aU,0x6c94U,
		0xd928U,0xa25bU,0x54bdU,0xa97aU,0x42ffU,0x85feU,0x1bf7U,0x37eeU,0x6fdcU,0xdfb8U,
		0xaf7bU,0x4efdU,0x9dfaU,0x2bffU,0x57feU,0xaffcU,0x4ff3U,0x9fe6U,0x2fc7U,0x5f8eU,
		0xbf1cU,0x6e33U,0xdc66U,0xa8c7U,0x4185U,0x830aU,0x161fU,0x2c3eU,0x587cU,0xb0f8U,
		0x71fbU,0xe3f6U,0xd7e7U,0xbfc5U,0x6f81U,0xdf02U,0xae0fU,0x4c15U,0x982aU,0x205fU,
		0x40beU,0x817cU,0x12f3U,0x25e6U,0x4bccU,0x9798U,0x3f3bU,0x7e76U,0xfcecU,0xe9d3U,
		0xc3adU,0x9751U,0x3ea9U,0x7d52U,0xfaa4U,0xe543U,0xda8dU,0xa511U,0x5a29U,0xb452U,
		0x78afU,0xf15eU,0xf2b7U,0xf565U,0xfac1U,0xe589U,0xdb19U,0xa639U,0x5c79U,0xb8f2U,
		0x61efU,0xc3deU,0x97b7U,0x3f65U,0x7ecaU,0xfd94U,0xeb23U,0xc64dU,0x9c91U,0x2929U,
		0x5252U,0xa4a4U,0x5943U,0xb286U,0x7507U,0xea0eU,0xc417U,0x9825U,0x2041U,0x4082U,
		0x8104U,0x1203U,0x2406U,0x480cU,0x9018U,0x303bU,0x6076U,0xc0ecU,0x91d3U,0x33adU,
		0x675aU,0xceb4U,0x8d63U,0x0acdU,0x159aU,0x2b34U,0x5668U,0xacd0U,0x49abU,0x9356U,
		0x36a7U,0x6d4eU,0xda9cU,0xa533U,0x5a6dU,0xb4daU,0x79bfU,0xf37eU,0xf6f7U,0xfde5U,
		0xebc1U,0xc789U,0x9f19U,0x2e39U,0x5c72U,0xb8e4U,0x61c3U,0xc386U,0x9707U,0x3e05U,
		0x7c0aU,0xf814U,0xe023U,0xd04dU,0xb091U,0x7129U,0xe252U,0xd4afU,0xb955U,0x62a1U,
		0xc542U,0x9a8fU,0x2515U,0x4a2aU,0x9454U,0x38a3U,0x7146U,0xe28cU,0xd513U,0xba2dU,
		0x6451U,0xc8a2U,0x814fU,0x1295U,0x252aU,0x4a54U,0x94a8U,0x395bU,0x72b6U,0xe56cU,
		0xdad3U,0xa5adU,0x5b51U,0xb6a2U,0x7d4fU,0xfa9eU,0xe537U,0xda65U,0xa4c1U,0x5989U,
		0xb312U,0x762fU,0xec5eU,0xc8b7U,0x8165U,0x12c1U,0x2582U,0x4b04U,0x9608U,0x3c1bU,
		0x7836U,0xf06cU,0xf0d3U,0xf1adU,0xf351U,0xf6a9U,0xfd59U,0xeab9U,0xc579U,0x9af9U,
		0x25f9U,0x4bf2U,0x97e4U,0x3fc3U,0x7f86U,0xff0cU,0xee13U,0xcc2dU,0x8851U,0x00a9U,
		0x0152U,0x02a4U,0x0548U,0x0a90U,0x1520U,0x2a40U,0x5480U,0xa900U,0x420bU,0x8416U,
		0x1827U,0x304eU,0x609cU,0xc138U,0x927bU,0x34fdU,0x69faU,0xd3f4U,0xb7e3U,0x7fcdU,
		0xff9aU,0xef3fU,0xce75U,0x8ce1U,0x09c9U,0x1392U,0x2724U,0x4e48U,0x9c90U,0x292bU,
		0x5256U,0xa4acU,0x5953U,0xb2a6U,0x7547U,0xea8eU,0xc517U,0x9a25U,0x2441U,0x4882U,
		0x9104U,0x3203U,0x6406U,0xc80cU,0x8013U,0x102dU,0x205aU,0x40b4U,0x8168U,0x12dbU,
		0x25b6U,0x4b6cU,0x96d8U,0x3dbbU,0x7b76U,0xf6ecU,0xfdd3U,0xebadU,0xc751U,0x9ea9U,
		0x2d59U,0x5ab2U,0xb564U,0x7ac3U,0xf586U,0xfb07U,0xe605U,0xdc01U,0xa809U,0x4019U,
		0x8032U,0x106fU,0x20deU,0x41bcU,0x8378U,0x16fbU,0x2df6U,0x5becU,0xb7d8U,0x7fbbU,
		0xff76U,0xeee7U,0xcdc5U,0x8b81U,0x0709U,0x0e12U,0x1c24U,0x3848U,0x7090U,0xe120U,
		0xd24bU,0xb49dU,0x7931U,0xf262U,0xf4cfU,0xf995U,0xe321U,0xd649U,0xbc99U,0x6939U,
		0xd272U,0xb4efU,0x79d5U,0xf3aaU,0xf75fU,0xfeb5U,0xed61U,0xcac9U,0x8599U,0x1b39U,
		0x3672U,0x6ce4U,0xd9c8U,0xa39bU,0x573dU,0xae7aU,0x4cffU,0x99feU,0x23f7U,0x47eeU,
		0x8fdcU,0x0fb3U,0x1f66U,0x3eccU,0x7d98U,0xfb30U,0xe66bU,0xdcddU,0xa9b1U,0x4369U,
		0x86d2U,0x1dafU,0x3b5eU,0x76bcU,0xed78U,0xcafbU,0x85fdU,0x1bf1U,0x37e2U,0x6fc4U,
		0xdf88U,0xaf1bU,0x4e3dU,0x9c7aU,0x28ffU,0x51feU,0xa3fcU,0x57f3U,0xafe6U,0x4fc7U,
		0x9f8eU,0x2f17U,0x5e2eU,0xbc5cU,0x68b3U,0xd166U,0xb2c7U,0x7585U,0xeb0aU,0xc61fU,
		0x9c35U,0x2861U,0x50c2U,0xa184U,0x5303U,0xa606U,0x5c07U,0xb80eU,0x6017U,0xc02eU,
		0x9057U,0x30a5U,0x614aU,0xc294U,0x9523U,0x3a4dU,0x749aU,0xe934U,0xc263U,0x94cdU,
		0x3991U,0x7322U,0xe644U,0xdc83U,0xa90dU,0x4211U,0x8422U,0x184fU,0x309eU,0x613cU,
		0xc278U,0x94fbU,0x39fdU,0x73faU,0xe7f4U,0xdfe3U,0xafcdU,0x4f91U,0x9f22U,0x2e4fU,
		0x5c9eU,0xb93cU,0x6273U,0xc4e6U,0x99c7U,0x2385U,0x470aU,0x8e14U,0x0c23U,0x1846U,
		0x308cU,0x6118U,0xc230U,0x946bU,0x38ddU,0x71baU,0xe374U,0xd6e3U,0xbdcdU,0x6b91U,
		0xd722U,0xbe4fU,0x6c95U,0xd92aU,0xa25fU,0x54b5U,0xa96aU,0x42dfU,0x85beU,0x1b77U,
		0x36eeU,0x6ddcU,0xdbb8U,0xa77bU,0x5efdU,0xbdfaU,0x6bffU,0xd7feU,0xbff7U,0x6fe5U,
		0xdfcaU,0xaf9fU,0x4f35U,0x9e6aU,0x2cdfU,0x59beU,0xb37cU,0x76f3U,0xede6U,0xcbc7U,
		0x8785U,0x1f01U,0x3e02U,0x7c04U,0xf808U,0xe01bU,0xd03dU,0xb071U,0x70e9U,0xe1d2U,
		0xd3afU,0xb755U,0x7ea1U,0xfd42U,0xea8fU,0xc515U,0x9a21U,0x2449U,0x4892U,0x9124U,
		0x3243U,0x6486U,0xc90cU,0x8213U,0x142dU,0x285aU,0x50b4U,0xa168U,0x52dbU,0xa5b6U,
		0x5b67U,0xb6ceU,0x7d97U,0xfb2eU,0xe657U,0xdca5U,0xa941U,0x4289U,0x8512U,0x1a2fU,
		0x345eU,0x68bcU,0xd178U,0xb2fbU,0x75fdU,0xebfaU,0xc7ffU,0x9ff5U,0x2fe1U,0x5fc2U,
		0xbf84U,0x6f03U,0xde06U,0xac07U,0x4805U,0x900aU,0x301fU,0x603eU,0xc07cU,0x90f3U,
		0x31edU,0x63daU,0xc7b4U,0x9f63U,0x2ecdU,0x5d9aU,0xbb34U,0x6663U,0xccc6U,0x8987U,
		0x0305U,0x060aU,0x0c14U,0x1828U,0x3050U,0x60a0U,0xc140U,0x928bU,0x351dU,0x6a3aU,
		0xd474U,0xb8e3U,0x61cdU,0xc39aU,0x973fU,0x3e75U,0x7ceaU,0xf9d4U,0xe3a3U,0xd74dU,
		0xbe91U,0x6d29U,0xda52U,0xa4afU,0x5955U,0xb2aaU,0x755fU,0xeabeU,0xc577U,0x9ae5U,
		0x25c1U,0x4b82U,0x9704U,0x3e03U,0x7c06U,0xf80cU,0xe013U,0xd02dU,0xb051U,0x70a9U,
		0xe152U,0xd2afU,0xb555U,0x7aa1U,0xf542U,0xfa8fU,0xe515U,0xda21U,0xa449U,0x5899U,
		0xb132U,0x726fU,0xe4deU,0xd9b7U,0xa365U,0x56c1U,0xad82U,0x4b0fU,0x961eU,0x3c37U,
		0x786eU,0xf0dcU,0xf1b3U,0xf36dU,0xf6d1U,0xfda9U,0xeb59U,0xc6b9U,0x9d79U,0x2af9U,
		0x55f2U,0xabe4U,0x47c3U,0x8f86U,0x0f07U,0x1e0eU,0x3c1cU,0x7838U,0xf070U,0xf0ebU,
		0xf1ddU,0xf3b1U,0xf769U,0xfed9U,0xedb9U,0xcb79U,0x86f9U,0x1df9U,0x3bf2U,0x77e4U,
		0xefc8U,0xcf9bU,0x8f3dU,0x0e71U,0x1ce2U,0x39c4U,0x7388U,0xe710U,0xde2bU,0xac5dU,
		0x48b1U,0x9162U,0x32cfU,0x659eU,0xcb3cU,0x8673U,0x1cedU,0x39daU,0x73b4U,0xe768U,
		0xdedbU,0xadbdU,0x4b71U,0x96e2U,0x3dcfU,0x7b9eU,0xf73cU,0xfe73U,0xecedU,0xc9d1U,
		0x83a9U,0x1759U,0x2eb2U,0x5d64U,0xbac8U,0x659bU,0xcb36U,0x8667U,0x1cc5U,0x398aU,
		0x7314U,0xe628U,0xdc5bU,0xa8bdU,0x4171U,0x82e2U,0x15cfU,0x2b9eU,0x573cU,0xae78U,
		0x4cfbU,0x99f6U,0x23e7U,0x47ceU,0x8f9cU,0x0f33U,0x1e66U,0x3cccU,0x7998U,0xf330U,
		0xf66bU,0xfcddU,0xe9b1U,0xc369U,0x96d9U,0x3db9U,0x7b72U,0xf6e4U,0xfdc3U,0xeb8dU,
		0xc711U,0x9e29U,0x2c59U,0x58b2U,0xb164U,0x72c3U,0xe586U,0xdb07U,0xa605U,0x5c01U,
		0xb802U,0x600fU,0xc01eU,0x9037U,0x3065U,0x60caU,0xc194U,0x9323U,0x364dU,0x6c9aU,
		0xd934U,0xa263U,0x54cdU,0xa99aU,0x433fU,0x867eU,0x1cf7U,0x39eeU,0x73dcU,0xe7b8U,
		0xdf7bU,0xaefdU,0x4df1U,0x9be2U,0x27cfU,0x4f9eU,0x9f3cU,0x2e73U,0x5ce6U,0xb9ccU,
		0x6393U,0xc726U,0x9e47U,0x2c85U,0x590aU,0xb214U,0x7423U,0xe846U,0xc087U,0x9105U,
		0x3201U,0x6402U,0xc804U,0x8003U,0x100dU,0x201aU,0x4034U,0x8068U,0x10dbU,0x21b6U,
		0x436cU,0x86d8U,0x1dbbU,0x3b76U,0x76ecU,0xedd8U,0xcbbbU,0x877dU,0x1ef1U,0x3de2U,
		0x7bc4U,0xf788U,0xff1bU,0xee3dU,0xcc71U,0x88e9U,0x01d9U,0x03b2U,0x0764U,0x0ec8U,
		0x1d90U,0x3b20U,0x7640U,0xec80U,0xc90bU,0x821dU,0x1431U,0x2862U,0x50c4U,0xa188U,
		0x531bU,0xa636U,0x5c67U,0xb8ceU,0x6197U,0xc32eU,0x9657U,0x3ca5U,0x794aU,0xf294U,
		0xf523U,0xfa4dU,0xe491U,0xd929U,0xa259U,0x54b9U,0xa972U,0x42efU,0x85deU,0x1bb7U,
		0x376eU,0x6edcU,0xddb8U,0xab7bU,0x46fdU,0x8dfaU,0x0bffU,0x17feU,0x2ffcU,0x5ff8U,
		0xbff0U,0x6febU,0xdfd6U,0xafa7U,0x4f45U,0x9e8aU,0x2d1fU,0x5a3eU,0xb47cU,0x78f3U,
		0xf1e6U,0xf3c7U,0xf785U,0xff01U,0xee09U,0xcc19U,0x8839U,0x0079U,0x00f2U,0x01e4U,
		0x03c8U,0x0790U,0x0f20U,0x1e40U,0x3c80U,0x7900U,0xf200U,0xf40bU,0xf81dU,0xe031U,
		0xd069U,0xb0d9U,0x71b9U,0xe372U,0xd6efU,0xbdd5U,0x6ba1U,0xd742U,0xbe8fU,0x6d15U,
		0xda2aU,0xa45fU,0x58b5U,0xb16aU,0x72dfU,0xe5beU,0xdb77U,0xa6e5U,0x5dc1U,0xbb82U,
		0x670fU,0xce1eU,0x8c37U,0x0865U,0x10caU,0x2194U,0x4328U,0x8650U,0x1cabU,0x3956U,
		0x72acU,0xe558U,0xdabbU,0xa57dU,0x5af1U,0xb5e2U,0x7bcfU,0xf79eU,0xff37U,0xee65U,
		0xccc1U,0x8989U,0x0319U,0x0632U,0x0c64U,0x18c8U,0x3190U,0x6320U,0xc640U,0x9c8bU,
		0x291dU,0x523aU,0xa474U,0x58e3U,0xb1c6U,0x7387U,0xe70eU,0xde17U,0xac25U,0x4841U,
		0x9082U,0x310fU,0x621eU,0xc43cU,0x9873U,0x20edU,0x41daU,0x83b4U,0x1763U,0x2ec6U,
		0x5d8cU,0xbb18U,0x663bU,0xcc76U,0x88e7U,0x01c5U,0x038aU,0x0714U,0x0e28U,0x1c50U,
		0x38a0U,0x7140U,0xe280U,0xd50bU,0xba1dU,0x6431U,0xc862U,0x80cfU,0x1195U,0x232aU,
		0x4654U,0x8ca8U,0x095bU,0x12b6U,0x256cU,0x4ad8U,0x95b0U,0x3b6bU,0x76d6U,0xedacU,
		0xcb53U,0x86adU,0x1d51U,0x3aa2U,0x7544U,0xea88U,0xc51bU,0x9a3dU,0x2471U,0x48e2U,
		0x91c4U,0x3383U,0x6706U,0xce0cU,0x8c13U,0x082dU,0x105aU,0x20b4U,0x4168U,0x82d0U,
		0x15abU,0x2b56U,0x56acU,0xad58U,0x4abbU,0x9576U,0x3ae7U,0x75ceU,0xeb9cU,0xc733U,
		0x9e6dU,0x2cd1U,0x59a2U,0xb344U,0x7683U,0xed06U,0xca07U,0x8405U,0x1801U,0x3002U,
		0x6004U,0xc008U,0x901bU,0x303dU,0x607aU,0xc0f4U,0x91e3U,0x33cdU,0x679aU,0xcf34U,
		0x8e63U,0x0ccdU,0x199aU,0x3334U,0x6668U,0xccd0U,0x89abU,0x035dU,0x06baU,0x0d74U,
		0x1ae8U,0x35d0U,0x6ba0U,0xd740U,0xbe8bU,0x6d1dU,0xda3aU,0xa47fU,0x58f5U,0xb1eaU,
		0x73dfU,0xe7beU,0xdf77U,0xaee5U,0x4dc1U,0x9b82U,0x270fU,0x4e1eU,0x9c3cU,0x2873U,
		0x50e6U,0xa1ccU,0x5393U,0xa726U,0x5e47U,0xbc8eU,0x6917U,0xd22eU,0xb457U,0x78a5U,
		0xf14aU,0xf29fU,0xf535U,0xfa61U,0xe4c9U,0xd999U,0xa339U,0x5679U,0xacf2U,0x49efU,
		0x93deU,0x37b7U,0x6f6eU,0xdedcU,0xadb3U,0x4b6dU,0x96daU,0x3dbfU,0x7b7eU,0xf6fcU,
		0xfdf3U,0xebedU,0xc7d1U,0x9fa9U,0x2f59U,0x5eb2U,0xbd64U,0x6ac3U,0xd586U,0xbb07U,
		0x6605U,0xcc0aU,0x881fU,0x0035U,0x006aU,0x00d4U,0x01a8U,0x0350U,0x06a0U,0x0d40U,
		0x1a80U,0x3500U,0x6a00U,0xd400U,0xb80bU,0x601dU,0xc03aU,0x907fU,0x30f5U,0x61eaU,
		0xc3d4U,0x97a3U,0x3f4dU,0x7e9aU,0xfd34U,0xea63U,0xc4cdU,0x9991U,0x2329U,0x4652U,
		0x8ca4U,0x0943U,0x1286U,0x250cU,0x4a18U,0x9430U,0x386bU,0x70d6U,0xe1acU,0xd353U,
		0xb6adU,0x7d51U,0xfaa2U,0xe54fU,0xda95U,0xa521U,0x5a49U,0xb492U,0x792fU,0xf25eU,
		0xf4b7U,0xf965U,0xe2c1U,0xd589U,0xbb19U,0x6639U,0xcc72U,0x88efU,0x01d5U,0x03aaU,
		0x0754U,0x0ea8U,0x1d50U,0x3aa0U,0x7540U,0xea80U,0xc50bU,0x9a1dU,0x2431U,0x4862U,
		0x90c4U,0x3183U,0x6306U,0xc60cU,0x9c13U,0x282dU,0x505aU,0xa0b4U,0x5163U,0xa2c6U,
		0x5587U,0xab0eU,0x4617U,0x8c2eU,0x0857U,0x10aeU,0x215cU,0x42b8U,0x8570U,0x1aebU,
		0x35d6U,0x6bacU,0xd758U,0xbebbU,0x6d7dU,0xdafaU,0xa5ffU,0x5bf5U,0xb7eaU,0x7fdfU,
		0xffbeU,0xef77U,0xcee5U,0x8dc1U,0x0b89U,0x1712U,0x2e24U,0x5c48U,0xb890U,0x612bU,
		0xc256U,0x94a7U,0x3945U,0x728aU,0xe514U,0xda23U,0xa44dU,0x5891U,0xb122U,0x724fU,
		0xe49eU,0xd937U,0xa265U,0x54c1U,0xa982U,0x430fU,0x861eU,0x1c37U,0x386eU,0x70dcU,
		0xe1b8U,0xd37bU,0xb6fdU,0x7df1U,0xfbe2U,0xe7cfU,0xdf95U,0xaf21U,0x4e49U,0x9c92U,
		0x292fU,0x525eU,0xa4bcU,0x5973U,0xb2e6U,0x75c7U,0xeb8eU,0xc717U,0x9e25U,0x2c41U,
		0x5882U,0xb104U,0x7203U,0xe406U,0xd807U,0xa005U,0x5001U,0xa002U,0x500fU,0xa01eU,
		0x5037U,0xa06eU,0x50d7U,0xa1aeU,0x5357U,0xa6aeU,0x5d57U,0xbaaeU,0x6557U,0xcaaeU,
		0x8557U,0x1aa5U,0x354aU,0x6a94U,0xd528U,0xba5bU,0x64bdU,0xc97aU,0x82ffU,0x15f5U,
		0x2beaU,0x57d4U,0xafa8U,0x4f5bU,0x9eb6U,0x2d67U,0x5aceU,0xb59cU,0x7b33U,0xf666U,
		0xfcc7U,0xe985U,0xc301U,0x9609U,0x3c19U,0x7832U,0xf064U,0xf0c3U,0xf18dU,0xf311U,
		0xf629U,0xfc59U,0xe8b9U,0xc179U,0x92f9U,0x35f9U,0x6bf2U,0xd7e4U,0xbfc3U,0x6f8dU,
		0xdf1aU,0xae3fU,0x4c75U,0x98eaU,0x21dfU,0x43beU,0x877cU,0x1ef3U,0x3de6U,0x7bccU,
		0xf798U,0xff3bU,0xee7dU,0xccf1U,0x89e9U,0x03d9U,0x07b2U,0x0f64U,0x1ec8U,0x3d90U,
		0x7b20U,0xf640U,0xfc8bU,0xe91dU,0xc231U,0x9469U,0x38d9U,0x71b2U,0xe364U,0xd6c3U,
		0xbd8dU,0x6b11U,0xd622U,0xbc4fU,0x6895U,0xd12aU,0xb25fU,0x74b5U,0xe96aU,0xc2dfU,
		0x95b5U,0x3b61U,0x76c2U,0xed84U,0xcb03U,0x860dU,0x1c11U,0x3822U,0x7044U,0xe088U,
		0xd11bU,0xb23dU,0x7471U,0xe8e2U,0xc1cfU,0x9395U,0x3721U,0x6e42U,0xdc84U,0xa903U,
		0x420dU,0x841aU,0x183fU,0x307eU,0x60fcU,0xc1f8U,0x93fbU,0x37fdU,0x6ffaU,0xdff4U,
		0xafe3U,0x4fcdU,0x9f9aU,0x2f3fU,0x5e7eU,0xbcfcU,0x69f3U,0xd3e6U,0xb7c7U,0x7f85U,
		0xff0aU,0xee1fU,0xcc35U,0x8861U,0x00c9U,0x0192U,0x0324U,0x0648U,0x0c90U,0x1920U,
		0x3240U,0x6480U,0xc900U,0x820bU,0x141dU,0x283aU,0x5074U,0xa0e8U,0x51dbU,0xa3b6U,
		0x5767U,0xaeceU,0x4d97U,0x9b2eU,0x2657U,0x4caeU,0x995cU,0x22b3U,0x4566U,0x8accU,
		0x0593U,0x0b26U,0x164cU,0x2c98U,0x5930U,0xb260U,0x74cbU,0xe996U,0xc327U,0x9645U,
		0x3c81U,0x7902U,0xf204U,0xf403U,0xf80dU,0xe011U,0xd029U,0xb059U,0x70b9U,0xe172U,
		0xd2efU,0xb5d5U,0x7ba1U,0xf742U,0xfe8fU,0xed15U,0xca21U,0x8449U,0x1899U,0x3132U,
		0x6264U,0xc4c8U,0x999bU,0x233dU,0x467aU,0x8cf4U,0x09e3U,0x13c6U,0x278cU,0x4f18U,
		0x9e30U,0x2c6bU,0x58d6U,0xb1acU,0x7353U,0xe6a6U,0xdd47U,0xaa85U,0x4501U,0x8a02U,
		0x040fU,0x081eU,0x103cU,0x2078U,0x40f0U,0x81e0U,0x13cbU,0x2796U,0x4f2cU,0x9e58U,
		0x2cbbU,0x5976U,0xb2ecU,0x75d3U,0xeba6U,0xc747U,0x9e85U,0x2d01U,0x5a02U,0xb404U,
		0x7803U,0xf006U,0xf007U,0xf005U,0xf001U,0xf009U,0xf019U,0xf039U,0xf079U,0xf0f9U,
		0xf1f9U,0xf3f9U,0xf7f9U,0xfff9U,0xeff9U,0xcff9U,0x8ff9U,0x0ff9U,0x1ff2U,0x3fe4U,
		0x7fc8U,0xff90U,0xef2bU,0xce5dU,0x8cb1U,0x0969U,0x12d2U,0x25a4U,0x4b48U,0x9690U,
		0x3d2bU,0x7a56U,0xf4acU,0xf953U,0xe2adU,0xd551U,0xbaa9U,0x6559U,0xcab2U,0x856fU,
		0x1ad5U,0x35aaU,0x6b54U,0xd6a8U,0xbd5bU,0x6abdU,0xd57aU,0xbaffU,0x65f5U,0xcbeaU,
		0x87dfU,0x1fb5U,0x3f6aU,0x7ed4U,0xfda8U,0xeb5bU,0xc6bdU,0x9d71U,0x2ae9U,0x55d2U,
		0xaba4U,0x4743U,0x8e86U,0x0d07U,0x1a0eU,0x341cU,0x6838U,0xd070U,0xb0ebU,0x71ddU,
		0xe3baU,0xd77fU,0xbef5U,0x6de1U,0xdbc2U,0xa78fU,0x5f15U,0xbe2aU,0x6c5fU,0xd8beU,
		0xa177U,0x52e5U,0xa5caU,0x5b9fU,0xb73eU,0x7e77U,0xfceeU,0xe9d7U,0xc3a5U,0x9741U,
		0x3e89U,0x7d12U,0xfa24U,0xe443U,0xd88dU,0xa111U,0x5229U,0xa452U,0x58afU,0xb15eU,
		0x72b7U,0xe56eU,0xdad7U,0xa5a5U,0x5b41U,0xb682U,0x7d0fU,0xfa1eU,0xe437U,0xd865U,
		0xa0c1U,0x5189U,0xa312U,0x562fU,0xac5eU,0x48b7U,0x916eU,0x32d7U,0x65aeU,0xcb5cU,
		0x86b3U,0x1d6dU,0x3adaU,0x75b4U,0xeb68U,0xc6dbU,0x9dbdU,0x2b71U,0x56e2U,0xadc4U,
		0x4b83U,0x9706U,0x3e07U,0x7c0eU,0xf81cU,0xe033U,0xd06dU,0xb0d1U,0x71a9U,0xe352U,
		0xd6afU,0xbd55U,0x6aa1U,0xd542U,0xba8fU,0x6515U,0xca2aU,0x845fU,0x18b5U,0x316aU,
		0x62d4U,0xc5a8U,0x9b5bU,0x26bdU,0x4d7aU,0x9af4U,0x25e3U,0x4bc6U,0x978cU,0x3f13U,
		0x7e26U,0xfc4cU,0xe893U,0xc12dU,0x9251U,0x34a9U,0x6952U,0xd2a4U,0xb543U,0x7a8dU,
		0xf51aU,0xfa3fU,0xe475U,0xd8e1U,0xa1c9U,0x5399U,0xa732U,0x5e6fU,0xbcdeU,0x69b7U,
		0xd36eU,0xb6d7U,0x7da5U,0xfb4aU,0xe69fU,0xdd35U,0xaa61U,0x44c9U,0x8992U,0x032fU,
		0x065eU,0x0cbcU,0x1978U,0x32f0U,0x65e0U,0xcbc0U,0x878bU,0x1f1dU,0x3e3aU,0x7c74U,
		0xf8e8U,0xe1dbU,0xd3bdU,0xb771U,0x7ee9U,0xfdd2U,0xebafU,0xc755U,0x9ea1U,0x2d49U,
		0x5a92U,0xb524U,0x7a43U,0xf486U,0xf907U,0xe205U,0xd401U,0xb809U,0x6019U,0xc032U,
		0x906fU,0x30d5U,0x61aaU,0xc354U,0x96a3U,0x3d4dU,0x7a9aU,0xf534U,0xfa63U,0xe4cdU,
		0xd991U,0xa329U,0x5659U,0xacb2U,0x496fU,0x92deU,0x35b7U,0x6b6eU,0xd6dcU,0xbdb3U,
		0x6b6dU,0xd6daU,0xbdbfU,0x6b75U,0xd6eaU,0xbddfU,0x6bb5U,0xd76aU,0xbedfU,0x6db5U,
		0xdb6aU,0xa6dfU,0x5db5U,0xbb6aU,0x66dfU,0xcdbeU,0x8b77U,0x06e5U,0x0dcaU,0x1b94U,
		0x3728U,0x6e50U,0xdca0U,0xa94bU,0x429dU,0x853aU,0x1a7fU,0x34feU,0x69fcU,0xd3f8U,
		0xb7fbU,0x7ffdU,0xfffaU,0xefffU,0xcff5U,0x8fe1U,0x0fc9U,0x1f92U,0x3f24U,0x7e48U,
		0xfc90U,0xe92bU,0xc25dU,0x94b1U,0x3969U,0x72d2U,0xe5a4U,0xdb43U,0xa68dU,0x5d11U,
		0xba22U,0x644fU,0xc89eU,0x8137U,0x1265U,0x24caU,0x4994U,0x9328U,0x365bU,0x6cb6U,
		0xd96cU,0xa2d3U,0x55adU,0xab5aU,0x46bfU,0x8d7eU,0x0af7U,0x15eeU,0x2bdcU,0x57b8U,
		0xaf70U,0x4eebU,0x9dd6U,0x2ba7U,0x574eU,0xae9cU,0x4d33U,0x9a66U,0x24c7U,0x498eU,
		0x931cU,0x3633U,0x6c66U,0xd8ccU,0xa193U,0x532dU,0xa65aU,0x5cbfU,0xb97eU,0x62f7U,
		0xc5eeU,0x9bd7U,0x27a5U,0x4f4aU,0x9e94U,0x2d23U,0x5a46U,0xb48cU,0x7913U,0xf226U,
		0xf447U,0xf885U,0xe101U,0xd209U,0xb419U,0x7839U,0xf072U,0xf0efU,0xf1d5U,0xf3a1U,
		0xf749U,0xfe99U,0xed39U,0xca79U,0x84f9U,0x19f9U,0x33f2U,0x67e4U,0xcfc8U,0x8f9bU,
		0x0f3dU,0x1e7aU,0x3cf4U,0x79e8U,0xf3d0U,0xf7abU,0xff5dU,0xeeb1U,0xcd69U,0x8ad9U,
		0x05b9U,0x0b72U,0x16e4U,0x2dc8U,0x5b90U,0xb720U,0x7e4bU,0xfc96U,0xe927U,0xc245U,
		0x9481U,0x3909U,0x7212U,0xe424U,0xd843U,0xa08dU,0x5111U,0xa222U,0x544fU,0xa89eU,
		0x4137U,0x826eU,0x14d7U,0x29aeU,0x535cU,0xa6b8U,0x5d7bU,0xbaf6U,0x65e7U,0xcbceU,
		0x8797U,0x1f25U,0x3e4aU,0x7c94U,0xf928U,0xe25bU,0xd4bdU,0xb971U,0x62e9U,0xc5d2U,
		0x9bafU,0x2755U,0x4eaaU,0x9d54U,0x2aa3U,0x5546U,0xaa8cU,0x4513U,0x8a26U,0x0447U,
		0x088eU,0x111cU,0x2238U,0x4470U,0x88e0U,0x01cbU,0x0396U,0x072cU,0x0e58U,0x1cb0U,
		0x3960U,0x72c0U,0xe580U,0xdb0bU,0xa61dU,0x5c31U,0xb862U,0x60cfU,0xc19eU,0x9337U,
		0x3665U,0x6ccaU,0xd994U,0xa323U,0x564dU,0xac9aU,0x493fU,0x927eU,0x34f7U,0x69eeU,
		0xd3dcU,0xb7b3U,0x7f6dU,0xfedaU,0xedbfU,0xcb75U,0x86e1U,0x1dc9U,0x3b92U,0x7724U,
		0xee48U,0xcc9bU,0x893dU,0x0271U,0x04e2U,0x09c4U,0x1388U,0x2710U,0x4e20U,0x9c40U,
		0x288bU,0x5116U,0xa22cU,0x5453U,0xa8a6U,0x4147U,0x828eU,0x1517U,0x2a2eU,0x545cU,
		0xa8b8U,0x417bU,0x82f6U,0x15e7U,0x2bceU,0x579cU,0xaf38U,0x4e7bU,0x9cf6U,0x29e7U,
		0x53ceU,0xa79cU,0x5f33U,0xbe66U,0x6cc7U,0xd98eU,0xa317U,0x5625U,0xac4aU,0x489fU,
		0x913eU,0x3277U,0x64eeU,0xc9dcU,0x83b3U,0x176dU,0x2edaU,0x5db4U,0xbb68U,0x66dbU,
		0xcdb6U,0x8b67U,0x06c5U,0x0d8aU,0x1b14U,0x3628U,0x6c50U,0xd8a0U,0xa14bU,0x529dU,
		0xa53aU,0x5a7fU,0xb4feU,0x79f7U,0xf3eeU,0xf7d7U,0xffa5U,0xef41U,0xce89U,0x8d19U,
		0x0a39U,0x1472U,0x28e4U,0x51c8U,0xa390U,0x572bU,0xae56U,0x4ca7U,0x994eU,0x2297U,
		0x452eU,0x8a5cU,0x04b3U,0x0966U,0x12ccU,0x2598U,0x4b30U,0x9660U,0x3ccbU,0x7996U,
		0xf32cU,0xf653U,0xfcadU,0xe951U,0xc2a9U,0x9559U,0x3ab9U,0x7572U,0xeae4U,0xc5c3U,
		0x9b8dU,0x2711U,0x4e22U,0x9c44U,0x2883U,0x5106U,0xa20cU,0x5413U,0xa826U,0x4047U,
		0x808eU,0x1117U,0x222eU,0x445cU,0x88b8U,0x017bU,0x02f6U,0x05ecU,0x0bd8U,0x17b0U,
		0x2f60U,0x5ec0U,0xbd80U,0x6b0bU,0xd616U,0xbc27U,0x6845U,0xd08aU,0xb11fU,0x7235U,
		0xe46aU,0xd8dfU,0xa1b5U,0x5361U,0xa6c2U,0x5d8fU,0xbb1eU,0x6637U,0xcc6eU,0x88d7U,
		0x01a5U,0x034aU,0x0694U,0x0d28U,0x1a50U,0x34a0U,0x6940U,0xd280U,0xb50bU,0x7a1dU,
		0xf43aU,0xf87fU,0xe0f5U,0xd1e1U,0xb3c9U,0x7799U,0xef32U,0xce6fU,0x8cd5U,0x09a1U,
		0x1342U,0x2684U,0x4d08U,0x9a10U,0x242bU,0x4856U,0x90acU,0x3153U,0x62a6U,0xc54cU,
		0x9a93U,0x252dU,0x4a5aU,0x94b4U,0x3963U,0x72c6U,0xe58cU,0xdb13U,0xa62dU,0x5c51U,
		0xb8a2U,0x614fU,0xc29eU,0x9537U,0x3a65U,0x74caU,0xe994U,0xc323U,0x964dU,0x3c91U,
		0x7922U,0xf244U,0xf483U,0xf90dU,0xe211U,0xd429U,0xb859U,0x60b9U,0xc172U,0x92efU,
		0x35d5U,0x6baaU,0xd754U,0xbea3U,0x6d4dU,0xda9aU,0xa53fU,0x5a75U,0xb4eaU,0x79dfU,
		0xf3beU,0xf777U,0xfee5U,0xedc1U,0xcb89U,0x8719U,0x1e39U,0x3c72U,0x78e4U,0xf1c8U,
		0xf39bU,0xf73dU,0xfe71U,0xece9U,0xc9d9U,0x83b9U,0x1779U,0x2ef2U,0x5de4U,0xbbc8U,
		0x679bU,0xcf36U,0x8e67U,0x0cc5U,0x198aU,0x3314U,0x6628U,0xcc50U,0x88abU,0x015dU,
		0x02baU,0x0574U,0x0ae8U,0x15d0U,0x2ba0U,0x5740U,0xae80U,0x4d0bU,0x9a16U,0x2427U,
		0x484eU,0x909cU,0x3133U,0x6266U,0xc4ccU,0x9993U,0x232dU,0x465aU,0x8cb4U,0x0963U,
		0x12c6U,0x258cU,0x4b18U,0x9630U,0x3c6bU,0x78d6U,0xf1acU,0xf353U,0xf6adU,0xfd51U,
		0xeaa9U,0xc559U,0x9ab9U,0x2579U,0x4af2U,0x95e4U,0x3bc3U,0x7786U,0xef0cU,0xce13U,
		0x8c2dU,0x0851U,0x10a2U,0x2144U,0x4288U,0x8510U,0x1a2bU,0x3456U,0x68acU,0xd158U,
		0xb2bbU,0x757dU,0xeafaU,0xc5ffU,0x9bf5U,0x27e1U,0x4fc2U,0x9f84U,0x2f03U,0x5e06U,
		0xbc0cU,0x6813U,0xd026U,0xb047U,0x7085U,0xe10aU,0xd21fU,0xb435U,0x7861U,0xf0c2U,
		0xf18fU,0xf315U,0xf621U,0xfc49U,0xe899U,0xc139U,0x9279U,0x34f9U,0x69f2U,0xd3e4U,
		0xb7c3U,0x7f8dU,0xff1aU,0xee3fU,0xcc75U,0x88e1U,0x01c9U,0x0392U,0x0724U,0x0e48U,
		0x1c90U,0x3920U,0x7240U,0xe480U,0xd90bU,0xa21dU,0x5431U,0xa862U,0x40cfU,0x819eU,
		0x1337U,0x266eU,0x4cdcU,0x99b8U,0x237bU,0x46f6U,0x8decU,0x0bd3U,0x17a6U,0x2f4cU,
		0x5e98U,0xbd30U,0x6a6bU,0xd4d6U,0xb9a7U,0x6345U,0xc68aU,0x9d1fU,0x2a35U,0x546aU,
		0xa8d4U,0x41a3U,0x8346U,0x1687U,0x2d0eU,0x5a1cU,0xb438U,0x787bU,0xf0f6U,0xf1e7U,
		0xf3c5U,0xf781U,0xff09U,0xee19U,0xcc39U,0x8879U,0x00f9U,0x01f2U,0x03e4U,0x07c8U,
		0x0f90U,0x1f20U,0x3e40U,0x7c80U,0xf900U,0xe20bU,0xd41dU,0xb831U,0x6069U,0xc0d2U,
		0x91afU,0x3355U,0x66aaU,0xcd54U,0x8aa3U,0x054dU,0x0a9aU,0x1534U,0x2a68U,0x54d0U,
		0xa9a0U,0x434bU,0x8696U,0x1d27U,0x3a4eU,0x749cU,0xe938U,0xc27bU,0x94fdU,0x39f1U,
		0x73e2U,0xe7c4U,0xdf83U,0xaf0dU,0x4e11U,0x9c22U,0x284fU,0x509eU,0xa13cU,0x5273U,
		0xa4e6U,0x59c7U,0xb38eU,0x7717U,0xee2eU,0xcc57U,0x88a5U,0x0141U,0x0282U,0x0504U,
		0x0a08U,0x1410U,0x2820U,0x5040U,0xa080U,0x510bU,0xa216U,0x5427U,0xa84eU,0x4097U,
		0x812eU,0x1257U,0x24aeU,0x495cU,0x92b8U,0x357bU,0x6af6U,0xd5ecU,0xbbd3U,0x67adU,
		0xcf5aU,0x8ebfU,0x0d75U,0x1aeaU,0x35d4U,0x6ba8U,0xd750U,0xbeabU,0x6d5dU,0xdabaU,
		0xa57fU,0x5af5U,0xb5eaU,0x7bdfU,0xf7beU,0xff77U,0xeee5U,0xcdc1U,0x8b89U,0x0719U,
		0x0e32U,0x1c64U,0x38c8U,0x7190U,0xe320U,0xd64bU,0xbc9dU,0x6931U,0xd262U,0xb4cfU,
		0x7995U,0xf32aU,0xf65fU,0xfcb5U,0xe961U,0xc2c9U,0x9599U,0x3b39U,0x7672U,0xece4U,
		0xc9c3U,0x838dU,0x1711U,0x2e22U,0x5c44U,0xb888U,0x611bU,0xc236U,0x9467U,0x38c5U,
		0x718aU,0xe314U,0xd623U,0xbc4dU,0x6891U,0xd122U,0xb24fU,0x7495U,0xe92aU,0xc25fU,
		0x94b5U,0x3961U,0x72c2U,0xe584U,0xdb03U,0xa60dU,0x5c11U,0xb822U,0x604fU,0xc09eU,
		0x9137U,0x3265U,0x64caU,0xc994U,0x8323U,0x164dU,0x2c9aU,0x5934U,0xb268U,0x74dbU,
		0xe9b6U,0xc367U,0x96c5U,0x3d81U,0x7b02U,0xf604U,0xfc03U,0xe80dU,0xc011U,0x9029U,
		0x3059U,0x60b2U,0xc164U,0x92c3U,0x358dU,0x6b1aU,0xd634U,0xbc63U,0x68cdU,0xd19aU,
		0xb33fU,0x7675U,0xeceaU,0xc9dfU,0x83b5U,0x1761U,0x2ec2U,0x5d84U,0xbb08U,0x661bU,
		0xcc36U,0x8867U,0x00c5U,0x018aU,0x0314U,0x0628U,0x0c50U,0x18a0U,0x3140U,0x6280U,
		0xc500U,0x9a0bU,0x241dU,0x483aU,0x9074U,0x30e3U,0x61c6U,0xc38cU,0x9713U,0x3e2dU,
		0x7c5aU,0xf8b4U,0xe163U,0xd2cdU,0xb591U,0x7b29U,0xf652U,0xfcafU,0xe955U,0xc2a1U,
		0x9549U,0x3a99U,0x7532U,0xea64U,0xc4c3U,0x998dU,0x2311U,0x4622U,0x8c44U,0x0883U,
		0x1106U,0x220cU,0x4418U,0x8830U,0x006bU,0x00d6U,0x01acU,0x0358U,0x06b0U,0x0d60U,
		0x1ac0U,0x3580U,0x6b00U,0xd600U,0xbc0bU,0x681dU,0xd03aU,0xb07fU,0x70f5U,0xe1eaU,
		0xd3dfU,0xb7b5U,0x7f61U,0xfec2U,0xed8fU,0xcb15U,0x8621U,0x1c49U,0x3892U,0x7124U,
		0xe248U,0xd49bU,0xb93dU,0x6271U,0xc4e2U,0x99cfU,0x2395U,0x472aU,0x8e54U,0x0ca3U,
		0x1946U,0x328cU,0x6518U,0xca30U,0x846bU,0x18ddU,0x31baU,0x6374U,0xc6e8U,0x9ddbU,
		0x2bbdU,0x577aU,0xaef4U,0x4de3U,0x9bc6U,0x2787U,0x4f0eU,0x9e1cU,0x2c33U,0x5866U,
		0xb0ccU,0x7193U,0xe326U,0xd647U,0xbc85U,0x6901U,0xd202U,0xb40fU,0x7815U,0xf02aU,
		0xf05fU,0xf0b5U,0xf161U,0xf2c9U,0xf599U,0xfb39U,0xe679U,0xdcf9U,0xa9f9U,0x43f9U,
		0x87f2U,0x1fefU,0x3fdeU,0x7fbcU,0xff78U,0xeefbU,0xcdfdU,0x8bf1U,0x07e9U,0x0fd2U,
		0x1fa4U,0x3f48U,0x7e90U,0xfd20U,0xea4bU,0xc49dU,0x9931U,0x2269U,0x44d2U,0x89a4U,
		0x0343U,0x0686U,0x0d0cU,0x1a18U,0x3430U,0x6860U,0xd0c0U,0xb18bU,0x731dU,0xe63aU,
		0xdc7fU,0xa8f5U,0x41e1U,0x83c2U,0x178fU,0x2f1eU,0x5e3cU,0xbc78U,0x68fbU,0xd1f6U,
		0xb3e7U,0x77c5U,0xef8aU,0xcf1fU,0x8e35U,0x0c61U,0x18c2U,0x3184U,0x6308U,0xc610U,
		0x9c2bU,0x285dU,0x50baU,0xa174U,0x52e3U,0xa5c6U,0x5b87U,0xb70eU,0x7e17U,0xfc2eU,
		0xe857U,0xc0a5U,0x9141U,0x3289U,0x6512U,0xca24U,0x8443U,0x188dU,0x311aU,0x6234U,
		0xc468U,0x98dbU,0x21bdU,0x437aU,0x86f4U,0x1de3U,0x3bc6U,0x778cU,0xef18U,0xce3bU,
		0x8c7dU,0x08f1U,0x11e2U,0x23c4U,0x4788U,0x8f10U,0x0e2bU,0x1c56U,0x38acU,0x7158U,
		0xe2b0U,0xd56bU,0xbaddU,0x65b1U,0xcb62U,0x86cfU,0x1d95U,0x3b2aU,0x7654U,0xeca8U,
		0xc95bU,0x82bdU,0x1571U,0x2ae2U,0x55c4U,0xab88U,0x471bU,0x8e36U,0x0c67U,0x18ceU,
		0x319cU,0x6338U,0xc670U,0x9cebU,0x29ddU,0x53baU,0xa774U,0x5ee3U,0xbdc6U,0x6b87U,
		0xd70eU,0xbe17U,0x6c25U,0xd84aU,0xa09fU,0x5135U,0xa26aU,0x54dfU,0xa9beU,0x4377U,
		0x86eeU,0x1dd7U,0x3baeU,0x775cU,0xeeb8U,0xcd7bU,0x8afdU,0x05f1U,0x0be2U,0x17c4U,
		0x2f88U,0x5f10U,0xbe20U,0x6c4bU,0xd896U,0xa127U,0x5245U,0xa48aU,0x591fU,0xb23eU,
		0x7477U,0xe8eeU,0xc1d7U,0x93a5U,0x3741U,0x6e82U,0xdd04U,0xaa03U,0x440dU,0x881aU,
		0x003fU,0x007eU,0x00fcU,0x01f8U,0x03f0U,0x07e0U,0x0fc0U,0x1f80U,0x3f00U,0x7e00U,
		0xfc00U,0xe80bU,0xc01dU,0x9031U,0x3069U,0x60d2U,0xc1a4U,0x9343U,0x368dU,0x6d1aU,
		0xda34U,0xa463U,0x58cdU,0xb19aU,0x733fU,0xe67eU,0xdcf7U,0xa9e5U,0x43c1U,0x8782U,
		0x1f0fU,0x3e1eU,0x7c3cU,0xf878U,0xe0fbU,0xd1fdU,0xb3f1U,0x77e9U,0xefd2U,0xcfafU,
		0x8f55U,0x0ea1U,0x1d42U,0x3a84U,0x7508U,0xea10U,0xc42bU,0x985dU,0x20b1U,0x4162U,
		0x82c4U,0x1583U,0x2b06U,0x560cU,0xac18U,0x483bU,0x9076U,0x30e7U,0x61ceU,0xc39cU,
		0x9733U,0x3e6dU,0x7cdaU,0xf9b4U,0xe363U,0xd6cdU,0xbd91U,0x6b29U,0xd652U,0xbcafU,
		0x6955U,0xd2aaU,0xb55fU,0x7ab5U,0xf56aU,0xfadfU,0xe5b5U,0xdb61U,0xa6c9U,0x5d99U,
		0xbb32U,0x666fU,0xccdeU,0x89b7U,0x0365U,0x06caU,0x0d94U,0x1b28U,0x3650U,0x6ca0U,
		0xd940U,0xa28bU,0x551dU,0xaa3aU,0x447fU,0x88feU,0x01f7U,0x03eeU,0x07dcU,0x0fb8U,
		0x1f70U,0x3ee0U,0x7dc0U,0xfb80U,0xe70bU,0xde1dU,0xac31U,0x4869U,0x90d2U,0x31afU,
		0x635eU,0xc6bcU,0x9d73U,0x2aedU,0x55daU,0xabb4U,0x4763U,0x8ec6U,0x0d87U,0x1b0eU,
		0x361cU,0x6c38U,0xd870U,0xa0ebU,0x51ddU,0xa3baU,0x577fU,0xaefeU,0x4df7U,0x9beeU,
		0x27d7U,0x4faeU,0x9f5cU,0x2eb3U,0x5d66U,0xbaccU,0x6593U,0xcb26U,0x8647U,0x1c85U,
		0x390aU,0x7214U,0xe428U,0xd85bU,0xa0bdU,0x5171U,0xa2e2U,0x55cfU,0xab9eU,0x4737U,
		0x8e6eU,0x0cd7U,0x19aeU,0x335cU,0x66b8U,0xcd70U,0x8aebU,0x05ddU,0x0bbaU,0x1774U,
		0x2ee8U,0x5dd0U,0xbba0U,0x674bU,0xce96U,0x8d27U,0x0a45U,0x148aU,0x2914U,0x5228U,
		0xa450U,0x58abU,0xb156U,0x72a7U,0xe54eU,0xda97U,0xa525U,0x5a41U,0xb482U,0x790fU,
		0xf21eU,0xf437U,0xf865U,0xe0c1U,0xd189U,0xb319U,0x7639U,0xec72U,0xc8efU,0x81d5U,
		0x13a1U,0x2742U,0x4e84U,0x9d08U,0x2a1bU,0x5436U,0xa86cU,0x40d3U,0x81a6U,0x1347U,
		0x268eU,0x4d1cU,0x9a38U,0x247bU,0x48f6U,0x91ecU,0x33d3U,0x67a6U,0xcf4cU,0x8e93U,
		0x0d2dU,0x1a5aU,0x34b4U,0x6968U,0xd2d0U,0xb5abU,0x7b5dU,0xf6baU,0xfd7fU,0xeaf5U,
		0xc5e1U,0x9bc9U,0x2799U,0x4f32U,0x9e64U,0x2cc3U,0x5986U,0xb30cU,0x7613U,0xec26U,
		0xc847U,0x8085U,0x1101U,0x2202U,0x4404U,0x8808U,0x001bU,0x0036U,0x006cU,0x00d8U,
		0x01b0U,0x0360U,0x06c0U,0x0d80U,0x1b00U,0x3600U,0x6c00U,0xd800U,0xa00bU,0x501dU,
		0xa03aU,0x507fU,0xa0feU,0x51f7U,0xa3eeU,0x57d7U,0xafaeU,0x4f57U,0x9eaeU,0x2d57U,
		0x5aaeU,0xb55cU,0x7ab3U,0xf566U,0xfac7U,0xe585U,0xdb01U,0xa609U,0x5c19U,0xb832U,
		0x606fU,0xc0deU,0x91b7U,0x3365U,0x66caU,0xcd94U,0x8b23U,0x064dU,0x0c9aU,0x1934U,
		0x3268U,0x64d0U,0xc9a0U,0x834bU,0x169dU,0x2d3aU,0x5a74U,0xb4e8U,0x79dbU,0xf3b6U,
		0xf767U,0xfec5U,0xed81U,0xcb09U,0x8619U,0x1c39U,0x3872U,0x70e4U,0xe1c8U,0xd39bU,
		0xb73dU,0x7e71U,0xfce2U,0xe9cfU,0xc395U,0x9721U,0x3e49U,0x7c92U,0xf924U,0xe243U,
		0xd48dU,0xb911U,0x6229U,0xc452U,0x98afU,0x2155U,0x42aaU,0x8554U,0x1aa3U,0x3546U,
		0x6a8cU,0xd518U,0xba3bU,0x647dU,0xc8faU,0x81ffU,0x13f5U,0x27eaU,0x4fd4U,0x9fa8U,
		0x2f5bU,0x5eb6U,0xbd6cU,0x6ad3U,0xd5a6U,0xbb47U,0x6685U,0xcd0aU,0x8a1fU,0x0435U,
		0x086aU,0x10d4U,0x21a8U,0x4350U,0x86a0U,0x1d4bU,0x3a96U,0x752cU,0xea58U,0xc4bbU,
		0x997dU,0x22f1U,0x45e2U,0x8bc4U,0x0783U,0x0f06U,0x1e0cU,0x3c18U,0x7830U,0xf060U,
		0xf0cbU,0xf19dU,0xf331U,0xf669U,0xfcd9U,0xe9b9U,0xc379U,0x96f9U,0x3df9U,0x7bf2U,
		0xf7e4U,0xffc3U,0xef8dU,0xcf11U,0x8e29U,0x0c59U,0x18b2U,0x3164U,0x62c8U,0xc590U,
		0x9b2bU,0x265dU,0x4cbaU,0x9974U,0x22e3U,0x45c6U,0x8b8cU,0x0713U,0x0e26U,0x1c4cU,
		0x3898U,0x7130U,0xe260U,0xd4cbU,0xb99dU,0x6331U,0xc662U,0x9ccfU,0x2995U,0x532aU,
		0xa654U,0x5ca3U,0xb946U,0x6287U,0xc50eU,0x9a17U,0x2425U,0x484aU,0x9094U,0x3123U,
		0x6246U,0xc48cU,0x9913U,0x222dU,0x445aU,0x88b4U,0x0163U,0x02c6U,0x058cU,0x0b18U,
		0x1630U,0x2c60U,0x58c0U,0xb180U,0x730bU,0xe616U,0xdc27U,0xa845U,0x4081U,0x8102U,
		0x120fU,0x241eU,0x483cU,0x9078U,0x30fbU,0x61f6U,0xc3ecU,0x97d3U,0x3fadU,0x7f5aU,
		0xfeb4U,0xed63U,0xcacdU,0x8591U,0x1b29U,0x3652U,0x6ca4U,0xd948U,0xa29bU,0x553dU,
		0xaa7aU,0x44ffU,0x89feU,0x03f7U,0x07eeU,0x0fdcU,0x1fb8U,0x3f70U,0x7ee0U,0xfdc0U,
		0xeb8bU,0xc71dU,0x9e31U,0x2c69U,0x58d2U,0xb1a4U,0x7343U,0xe686U,0xdd07U,0xaa05U,
		0x4401U,0x8802U,0x000fU,0x001eU,0x003cU,0x0078U,0x00f0U,0x01e0U,0x03c0U,0x0780U,
		0x0f00U,0x1e00U,0x3c00U,0x7800U,0xf000U,0xf00bU,0xf01dU,0xf031U,0xf069U,0xf0d9U,
		0xf1b9U,0xf379U,0xf6f9U,0xfdf9U,0xebf9U,0xc7f9U,0x9ff9U,0x2ff9U,0x5ff2U,0xbfe4U,
		0x6fc3U,0xdf86U,0xaf07U,0x4e05U,0x9c0aU,0x281fU,0x503eU,0xa07cU,0x50f3U,0xa1e6U,
		0x53c7U,0xa78eU,0x5f17U,0xbe2eU,0x6c57U,0xd8aeU,0xa157U,0x52a5U,0xa54aU,0x5a9fU,
		0xb53eU,0x7a77U,0xf4eeU,0xf9d7U,0xe3a5U,0xd741U,0xbe89U,0x6d19U,0xda32U,0xa46fU,
		0x58d5U,0xb1aaU,0x735fU,0xe6beU,0xdd77U,0xaae5U,0x45c1U,0x8b82U,0x070fU,0x0e1eU,
		0x1c3cU,0x3878U,0x70f0U,0xe1e0U,0xd3cbU,0xb79dU,0x7f31U,0xfe62U,0xeccfU,0xc995U,
		0x8321U,0x1649U,0x2c92U,0x5924U,0xb248U,0x749bU,0xe936U,0xc267U,0x94c5U,0x3981U,
		0x7302U,0xe604U,0xdc03U,0xa80dU,0x4011U,0x8022U,0x104fU,0x209eU,0x413cU,0x8278U,
		0x14fbU,0x29f6U,0x53ecU,0xa7d8U,0x5fbbU,0xbf76U,0x6ee7U,0xddceU,0xab97U,0x4725U,
		0x8e4aU,0x0c9fU,0x193eU,0x327cU,0x64f8U,0xc9f0U,0x83ebU,0x17ddU,0x2fbaU,0x5f74U,
		0xbee8U,0x6ddbU,0xdbb6U,0xa767U,0x5ec5U,0xbd8aU,0x6b1fU,0xd63eU,0xbc77U,0x68e5U,
		0xd1caU,0xb39fU,0x7735U,0xee6aU,0xccdfU,0x89b5U,0x0361U,0x06c2U,0x0d84U,0x1b08U,
		0x3610U,0x6c20U,0xd840U,0xa08bU,0x511dU,0xa23aU,0x547fU,0xa8feU,0x41f7U,0x83eeU,
		0x17d7U,0x2faeU,0x5f5cU,0xbeb8U,0x6d7bU,0xdaf6U,0xa5e7U,0x5bc5U,0xb78aU,0x7f1fU,
		0xfe3eU,0xec77U,0xc8e5U,0x81c1U,0x1389U,0x2712U,0x4e24U,0x9c48U,0x289bU,0x5136U,
		0xa26cU,0x54d3U,0xa9a6U,0x4347U,0x868eU,0x1d17U,0x3a2eU,0x745cU,0xe8b8U,0xc17bU,
		0x92fdU,0x35f1U,0x6be2U,0xd7c4U,0xbf83U,0x6f0dU,0xde1aU,0xac3fU,0x4875U,0x90eaU,
		0x31dfU,0x63beU,0xc77cU,0x9ef3U,0x2dedU,0x5bdaU,0xb7b4U,0x7f63U,0xfec6U,0xed87U,
		0xcb05U,0x8601U,0x1c09U,0x3812U,0x7024U,0xe048U,0xd09bU,0xb13dU,0x7271U,0xe4e2U,
		0xd9cfU,0xa395U,0x5721U,0xae42U,0x4c8fU,0x991eU,0x2237U,0x446eU,0x88dcU,0x01b3U,
		0x0366U,0x06ccU,0x0d98U,0x1b30U,0x3660U,0x6cc0U,0xd980U,0xa30bU,0x561dU,0xac3aU,
		0x487fU,0x90feU,0x31f7U,0x63eeU,0xc7dcU,0x9fb3U,0x2f6dU,0x5edaU,0xbdb4U,0x6b63U,
		0xd6c6U,0xbd87U,0x6b05U,0xd60aU,0xbc1fU,0x6835U,0xd06aU,0xb0dfU,0x71b5U,0xe36aU,
		0xd6dfU,0xbdb5U,0x6b61U,0xd6c2U,0xbd8fU,0x6b15U,0xd62aU,0xbc5fU,0x68b5U,0xd16aU,
		0xb2dfU,0x75b5U,0xeb6aU,0xc6dfU,0x9db5U,0x2b61U,0x56c2U,0xad84U,0x4b03U,0x9606U,
		0x3c07U,0x780eU,0xf01cU,0xf033U,0xf06dU,0xf0d1U,0xf1a9U,0xf359U,0xf6b9U,0xfd79U,
		0xeaf9U,0xc5f9U,0x9bf9U,0x27f9U,0x4ff2U,0x9fe4U,0x2fc3U,0x5f86U,0xbf0cU,0x6e13U,
		0xdc26U,0xa847U,0x4085U,0x810aU,0x121fU,0x243eU,0x487cU,0x90f8U,0x31fbU,0x63f6U,
		0xc7ecU,0x9fd3U,0x2fadU,0x5f5aU,0xbeb4U,0x6d63U,0xdac6U,0xa587U,0x5b05U,0xb60aU,
		0x7c1fU,0xf83eU,0xe077U,0xd0e5U,0xb1c1U,0x7389U,0xe712U,0xde2fU,0xac55U,0x48a1U,
		0x9142U,0x328fU,0x651eU,0xca3cU,0x8473U,0x18edU,0x31daU,0x63b4U,0xc768U,0x9edbU,
		0x2dbdU,0x5b7aU,0xb6f4U,0x7de3U,0xfbc6U,0xe787U,0xdf05U,0xae01U,0x4c09U,0x9812U,
		0x202fU,0x405eU,0x80bcU,0x1173U,0x22e6U,0x45ccU,0x8b98U,0x073bU,0x0e76U,0x1cecU,
		0x39d8U,0x73b0U,0xe760U,0xdecbU,0xad9dU,0x4b31U,0x9662U,0x3ccfU,0x799eU,0xf33cU,
		0xf673U,0xfcedU,0xe9d1U,0xc3a9U,0x9759U,0x3eb9U,0x7d72U,0xfae4U,0xe5c3U,0xdb8dU,
		0xa711U,0x5e29U,0xbc52U,0x68afU,0xd15eU,0xb2b7U,0x7565U,0xeacaU,0xc59fU,0x9b35U,
		0x2661U,0x4cc2U,0x9984U,0x2303U,0x4606U,0x8c0cU,0x0813U,0x1026U,0x204cU,0x4098U,
		0x8130U,0x126bU,0x24d6U,0x49acU,0x9358U,0x36bbU,0x6d76U,0xdaecU,0xa5d3U,0x5badU,
		0xb75aU,0x7ebfU,0xfd7eU,0xeaf7U,0xc5e5U,0x9bc1U,0x2789U,0x4f12U,0x9e24U,0x2c43U,
		0x5886U,0xb10cU,0x7213U,0xe426U,0xd847U,0xa085U,0x5101U,0xa202U,0x540fU,0xa81eU,
		0x4037U,0x806eU,0x10d7U,0x21aeU,0x435cU,0x86b8U,0x1d7bU,0x3af6U,0x75ecU,0xebd8U,
		0xc7bbU,0x9f7dU,0x2ef1U,0x5de2U,0xbbc4U,0x6783U,0xcf06U,0x8e07U,0x0c05U,0x180aU,
		0x3014U,0x6028U,0xc050U,0x90abU,0x315dU,0x62baU,0xc574U,0x9ae3U,0x25cdU,0x4b9aU,
		0x9734U,0x3e63U,0x7cc6U,0xf98cU,0xe313U,0xd62dU,0xbc51U,0x68a9U,0xd152U,0xb2afU,
		0x7555U,0xeaaaU,0xc55fU,0x9ab5U,0x2561U,0x4ac2U,0x9584U,0x3b03U,0x7606U,0xec0cU,
		0xc813U,0x802dU,0x1051U,0x20a2U,0x4144U,0x8288U,0x151bU,0x2a36U,0x546cU,0xa8d8U,
		0x41bbU,0x8376U,0x16e7U,0x2dceU,0x5b9cU,0xb738U,0x7e7bU,0xfcf6U,0xe9e7U,0xc3c5U,
		0x9781U,0x3f09U,0x7e12U,0xfc24U,0xe843U,0xc08dU,0x9111U,0x3229U,0x6452U,0xc8a4U,
		0x8143U,0x128dU,0x251aU,0x4a34U,0x9468U,0x38dbU,0x71b6U,0xe36cU,0xd6d3U,0xbdadU,
		0x6b51U,0xd6a2U,0xbd4fU,0x6a95U,0xd52aU,0xba5fU,0x64b5U,0xc96aU,0x82dfU,0x15b5U,
		0x2b6aU,0x56d4U,0xada8U,0x4b5bU,0x96b6U,0x3d67U,0x7aceU,0xf59cU,0xfb33U,0xe66dU,
		0xdcd1U,0xa9a9U,0x4359U,0x86b2U,0x1d6fU,0x3adeU,0x75bcU,0xeb78U,0xc6fbU,0x9dfdU,
		0x2bf1U,0x57e2U,0xafc4U,0x4f83U,0x9f06U,0x2e07U,0x5c0eU,0xb81cU,0x6033U,0xc066U,
		0x90c7U,0x3185U,0x630aU,0xc614U,0x9c23U,0x284dU,0x509aU,0xa134U,0x5263U,0xa4c6U,
		0x5987U,0xb30eU,0x7617U,0xec2eU,0xc857U,0x80a5U,0x1141U,0x2282U,0x4504U,0x8a08U,
		0x041bU,0x0836U,0x106cU,0x20d8U,0x41b0U,0x8360U,0x16cbU,0x2d96U,0x5b2cU,0xb658U,
		0x7cbbU,0xf976U,0xe2e7U,0xd5c5U,0xbb81U,0x6709U,0xce12U,0x8c2fU,0x0855U,0x10aaU,
		0x2154U,0x42a8U,0x8550U,0x1aabU,0x3556U,0x6aacU,0xd558U,0xbabbU,0x657dU,0xcafaU,
		0x85ffU,0x1bf5U,0x37eaU,0x6fd4U,0xdfa8U,0xaf5bU,0x4ebdU,0x9d7aU,0x2affU,0x55feU,
		0xabfcU,0x47f3U,0x8fe6U,0x0fc7U,0x1f8eU,0x3f1cU,0x7e38U,0xfc70U,0xe8ebU,0xc1ddU,
		0x93b1U,0x3769U,0x6ed2U,0xdda4U,0xab43U,0x468dU,0x8d1aU,0x0a3fU,0x147eU,0x28fcU,
		0x51f8U,0xa3f0U,0x57ebU,0xafd6U,0x4fa7U,0x9f4eU,0x2e97U,0x5d2eU,0xba5cU,0x64b3U,
		0xc966U,0x82c7U,0x1585U,0x2b0aU,0x5614U,0xac28U,0x485bU,0x90b6U,0x3167U,0x62ceU,
		0xc59cU,0x9b33U,0x266dU,0x4cdaU,0x99b4U,0x2363U,0x46c6U,0x8d8cU,0x0b13U,0x1626U,
		0x2c4cU,0x5898U,0xb130U,0x726bU,0xe4d6U,0xd9a7U,0xa345U,0x5681U,0xad02U,0x4a0fU,
		0x941eU,0x3837U,0x706eU,0xe0dcU,0xd1b3U,0xb36dU,0x76d1U,0xeda2U,0xcb4fU,0x8695U,
		0x1d21U,0x3a42U,0x7484U,0xe908U,0xc21bU,0x943dU,0x3871U,0x70e2U,0xe1c4U,0xd383U,
		0xb70dU,0x7e11U,0xfc22U,0xe84fU,0xc095U,0x9121U,0x3249U,0x6492U,0xc924U,0x8243U,
		0x148dU,0x291aU,0x5234U,0xa468U,0x58dbU,0xb1b6U,0x7367U,0xe6ceU,0xdd97U,0xab25U,
		0x4641U,0x8c82U,0x090fU,0x121eU,0x243cU,0x4878U,0x90f0U,0x31ebU,0x63d6U,0xc7acU,
		0x9f53U,0x2eadU,0x5d5aU,0xbab4U,0x6563U,0xcac6U,0x8587U,0x1b05U,0x360aU,0x6c14U,
		0xd828U,0xa05bU,0x50bdU,0xa17aU,0x52ffU,0xa5feU,0x5bf7U,0xb7eeU,0x7fd7U,0xffaeU,
		0xef57U,0xcea5U,0x8d41U,0x0a89U,0x1512U,0x2a24U,0x5448U,0xa890U,0x412bU,0x8256U,
		0x14a7U,0x294eU,0x529cU,0xa538U,0x5a7bU,0xb4f6U,0x79e7U,0xf3ceU,0xf797U,0xff25U,
		0xee41U,0xcc89U,0x8919U,0x0239U,0x0472U,0x08e4U,0x11c8U,0x2390U,0x4720U,0x8e40U,
		0x0c8bU,0x1916U,0x322cU,0x6458U,0xc8b0U,0x816bU,0x12ddU,0x25baU,0x4b74U,0x96e8U,
		0x3ddbU,0x7bb6U,0xf76cU,0xfed3U,0xedadU,0xcb51U,0x86a9U,0x1d59U,0x3ab2U,0x7564U,
		0xeac8U,0xc59bU,0x9b3dU,0x2671U,0x4ce2U,0x99c4U,0x2383U,0x4706U,0x8e0cU,0x0c13U,
		0x1826U,0x304cU,0x6098U,0xc130U,0x926bU,0x34ddU,0x69baU,0xd374U,0xb6e3U,0x7dcdU,
		0xfb9aU,0xe73fU,0xde75U,0xace1U,0x49c9U,0x9392U,0x372fU,0x6e5eU,0xdcbcU,0xa973U,
		0x42edU,0x85daU,0x1bbfU,0x377eU,0x6efcU,0xddf8U,0xabfbU,0x47fdU,0x8ffaU,0x0fffU,
		0x1ffeU,0x3ffcU,0x7ff8U,0xfff0U,0xefebU,0xcfddU,0x8fb1U,0x0f69U,0x1ed2U,0x3da4U,
		0x7b48U,0xf690U,0xfd2bU,0xea5dU,0xc4b1U,0x9969U,0x22d9U,0x45b2U,0x8b64U,0x06c3U,
		0x0d86U,0x1b0cU,0x3618U,0x6c30U,0xd860U,0xa0cbU,0x519dU,0xa33aU,0x567fU,0xacfeU,
		0x49f7U,0x93eeU,0x37d7U,0x6faeU,0xdf5cU,0xaeb3U,0x4d6dU,0x9adaU,0x25bfU,0x4b7eU,
		0x96fcU,0x3df3U,0x7be6U,0xf7ccU,0xff93U,0xef2dU,0xce51U,0x8ca9U,0x0959U,0x12b2U,
		0x2564U,0x4ac8U,0x9590U,0x3b2bU,0x7656U,0xecacU,0xc953U,0x82adU,0x1551U,0x2aa2U,
		0x5544U,0xaa88U,0x451bU,0x8a36U,0x0467U,0x08ceU,0x119cU,0x2338U,0x4670U,0x8ce0U,
		0x09cbU,0x1396U,0x272cU,0x4e58U,0x9cb0U,0x296bU,0x52d6U,0xa5acU,0x5b53U,0xb6a6U,
		0x7d47U,0xfa8eU,0xe517U,0xda25U,0xa441U,0x5889U,0xb112U,0x722fU,0xe45eU,0xd8b7U,
		0xa165U,0x52c1U,0xa582U,0x5b0fU,0xb61eU,0x7c37U,0xf86eU,0xe0d7U,0xd1a5U,0xb341U,
		0x7689U,0xed12U,0xca2fU,0x8455U,0x18a1U,0x3142U,0x6284U,0xc508U,0x9a1bU,0x243dU,
		0x487aU,0x90f4U,0x31e3U,0x63c6U,0xc78cU,0x9f13U,0x2e2dU,0x5c5aU,0xb8b4U,0x6163U,
		0xc2c6U,0x9587U,0x3b05U,0x760aU,0xec14U,0xc823U,0x804dU,0x1091U,0x2122U,0x4244U,
		0x8488U,0x191bU,0x3236U,0x646cU,0xc8d8U,0x81bbU,0x137dU,0x26faU,0x4df4U,0x9be8U,
		0x27dbU,0x4fb6U,0x9f6cU,0x2ed3U,0x5da6U,0xbb4cU,0x6693U,0xcd26U,0x8a47U,0x0485U,
		0x090aU,0x1214U,0x2428U,0x4850U,0x90a0U,0x314bU,0x6296U,0xc52cU,0x9a53U,0x24adU,
		0x495aU,0x92b4U,0x3563U,0x6ac6U,0xd58cU,0xbb13U,0x662dU,0xcc5aU,0x88bfU,0x0175U,
		0x02eaU,0x05d4U,0x0ba8U,0x1750U,0x2ea0U,0x5d40U,0xba80U,0x650bU,0xca16U,0x8427U,
		0x1845U,0x308aU,0x6114U,0xc228U,0x945bU,0x38bdU,0x717aU,0xe2f4U,0xd5e3U,0xbbcdU,
		0x6791U,0xcf22U,0x8e4fU,0x0c95U,0x192aU,0x3254U,0x64a8U,0xc950U,0x82abU,0x155dU,
		0x2abaU,0x5574U,0xaae8U,0x45dbU,0x8bb6U,0x0767U,0x0eceU,0x1d9cU,0x3b38U,0x7670U,
		0xece0U,0xc9cbU,0x839dU,0x1731U,0x2e62U,0x5cc4U,0xb988U,0x631bU,0xc636U,0x9c67U,
		0x28c5U,0x518aU,0xa314U,0x5623U,0xac46U,0x4887U,0x910eU,0x3217U,0x642eU,0xc85cU,
		0x80b3U,0x116dU,0x22daU,0x45b4U,0x8b68U,0x06dbU,0x0db6U,0x1b6cU,0x36d8U,0x6db0U,
		0xdb60U,0xa6cbU,0x5d9dU,0xbb3aU,0x667fU,0xccfeU,0x89f7U,0x03e5U,0x07caU,0x0f94U,
		0x1f28U,0x3e50U,0x7ca0U,0xf940U,0xe28bU,0xd51dU,0xba31U,0x6469U,0xc8d2U,0x81afU,
		0x1355U,0x26aaU,0x4d54U,0x9aa8U,0x255bU,0x4ab6U,0x956cU,0x3ad3U,0x75a6U,0xeb4cU,
		0xc693U,0x9d2dU,0x2a51U,0x54a2U,0xa944U,0x4283U,0x8506U,0x1a07U,0x340eU,0x681cU,
		0xd038U,0xb07bU,0x70fdU,0xe1faU,0xd3ffU,0xb7f5U,0x7fe1U,0xffc2U,0xef8fU,0xcf15U,
		0x8e21U,0x0c49U,0x1892U,0x3124U,0x6248U,0xc490U,0x992bU,0x225dU,0x44baU,0x8974U,
		0x02e3U,0x05c6U,0x0b8cU,0x1718U,0x2e30U,0x5c60U,0xb8c0U,0x618bU,0xc316U,0x9627U,
		0x3c45U,0x788aU,0xf114U,0xf223U,0xf44dU,0xf891U,0xe129U,0xd259U,0xb4b9U,0x7979U,
		0xf2f2U,0xf5efU,0xfbd5U,0xe7a1U,0xdf49U,0xae99U,0x4d39U,0x9a72U,0x24efU,0x49deU,
		0x93bcU,0x3773U,0x6ee6U,0xddccU,0xab93U,0x472dU,0x8e5aU,0x0cbfU,0x197eU,0x32fcU,
		0x65f8U,0xcbf0U,0x87ebU,0x1fddU,0x3fbaU,0x7f74U,0xfee8U,0xeddbU,0xcbbdU,0x8771U,
		0x1ee9U,0x3dd2U,0x7ba4U,0xf748U,0xfe9bU,0xed3dU,0xca71U,0x84e9U,0x19d9U,0x33b2U,
		0x6764U,0xcec8U,0x8d9bU,0x0b3dU,0x167aU,0x2cf4U,0x59e8U,0xb3d0U,0x77abU,0xef56U,
		0xcea7U,0x8d45U,0x0a81U,0x1502U,0x2a04U,0x5408U,0xa810U,0x402bU,0x8056U,0x10a7U,
		0x214eU,0x429cU,0x8538U,0x1a7bU,0x34f6U,0x69ecU,0xd3d8U,0xb7bbU,0x7f7dU,0xfefaU,
		0xedffU,0xcbf5U,0x87e1U,0x1fc9U,0x3f92U,0x7f24U,0xfe48U,0xec9bU,0xc93dU,0x8271U,
		0x14e9U,0x29d2U,0x53a4U,0xa748U,0x5e9bU,0xbd36U,0x6a67U,0xd4ceU,0xb997U,0x6325U,
		0xc64aU,0x9c9fU,0x2935U,0x526aU,0xa4d4U,0x59a3U,0xb346U,0x7687U,0xed0eU,0xca17U,
		0x8425U,0x1841U,0x3082U,0x6104U,0xc208U,0x941bU,0x383dU,0x707aU,0xe0f4U,0xd1e3U,
		0xb3cdU,0x7791U,0xef22U,0xce4fU,0x8c95U,0x0921U,0x1242U,0x2484U,0x4908U,0x9210U,
		0x342bU,0x6856U,0xd0acU,0xb153U,0x72adU,0xe55aU,0xdabfU,0xa575U,0x5ae1U,0xb5c2U,
		0x7b8fU,0xf71eU,0xfe37U,0xec65U,0xc8c1U,0x8189U,0x1319U,0x2632U,0x4c64U,0x98c8U,
		0x219bU,0x4336U,0x866cU,0x1cd3U,0x39a6U,0x734cU,0xe698U,0xdd3bU,0xaa7dU,0x44f1U,
		0x89e2U,0x03cfU,0x079eU,0x0f3cU,0x1e78U,0x3cf0U,0x79e0U,0xf3c0U,0xf78bU,0xff1dU,
		0xee31U,0xcc69U,0x88d9U,0x01b9U,0x0372U,0x06e4U,0x0dc8U,0x1b90U,0x3720U,0x6e40U,
		0xdc80U,0xa90bU,0x421dU,0x843aU,0x187fU,0x30feU,0x61fcU,0xc3f8U,0x97fbU,0x3ffdU,
		0x7ffaU,0xfff4U,0xefe3U,0xcfcdU,0x8f91U,0x0f29U,0x1e52U,0x3ca4U,0x7948U,0xf290U,
		0xf52bU,0xfa5dU,0xe4b1U,0xd969U,0xa2d9U,0x55b9U,0xab72U,0x46efU,0x8ddeU,0x0bb7U,
		0x176eU,0x2edcU,0x5db8U,0xbb70U,0x66ebU,0xcdd6U,0x8ba7U,0x0745U,0x0e8aU,0x1d14U,
		0x3a28U,0x7450U,0xe8a0U,0xc14bU,0x929dU,0x3531U,0x6a62U,0xd4c4U,0xb983U,0x630dU,
		0xc61aU,0x9c3fU,0x2875U,0x50eaU,0xa1d4U,0x53a3U,0xa746U,0x5e87U,0xbd0eU,0x6a17U,
		0xd42eU,0xb857U,0x60a5U,0xc14aU,0x929fU,0x3535U,0x6a6aU,0xd4d4U,0xb9a3U,0x634dU,
		0xc69aU,0x9d3fU,0x2a75U,0x54eaU,0xa9d4U,0x43a3U,0x8746U,0x1e87U,0x3d0eU,0x7a1cU,
		0xf438U,0xf87bU,0xe0fdU,0xd1f1U,0xb3e9U,0x77d9U,0xefb2U,0xcf6fU,0x8ed5U,0x0da1U,
		0x1b42U,0x3684U,0x6d08U,0xda10U,0xa42bU,0x585dU,0xb0baU,0x717fU,0xe2feU,0xd5f7U,
		0xbbe5U,0x67c1U,0xcf82U,0x8f0fU,0x0e15U,0x1c2aU,0x3854U,0x70a8U,0xe150U,0xd2abU,
		0xb55dU,0x7ab1U,0xf562U,0xfacfU,0xe595U,0xdb21U,0xa649U,0x5c99U,0xb932U,0x626fU,
		0xc4deU,0x99b7U,0x2365U,0x46caU,0x8d94U,0x0b23U,0x1646U,0x2c8cU,0x5918U,0xb230U,
		0x746bU,0xe8d6U,0xc1a7U,0x9345U,0x3681U,0x6d02U,0xda04U,0xa403U,0x580dU,0xb01aU,
		0x703fU,0xe07eU,0xd0f7U,0xb1e5U,0x73c1U,0xe782U,0xdf0fU,0xae15U,0x4c21U,0x9842U,
		0x208fU,0x411eU,0x823cU,0x1473U,0x28e6U,0x51ccU,0xa398U,0x573bU,0xae76U,0x4ce7U,
		0x99ceU,0x2397U,0x472eU,0x8e5cU,0x0cb3U,0x1966U,0x32ccU,0x6598U,0xcb30U,0x866bU,
		0x1cddU,0x39baU,0x7374U,0xe6e8U,0xdddbU,0xabbdU,0x4771U,0x8ee2U,0x0dcfU,0x1b9eU,
		0x373cU,0x6e78U,0xdcf0U,0xa9ebU,0x43ddU,0x87baU,0x1f7fU,0x3efeU,0x7dfcU,0xfbf8U,
		0xe7fbU,0xdffdU,0xaff1U,0x4fe9U,0x9fd2U,0x2fafU,0x5f5eU,0xbebcU,0x6d73U,0xdae6U,
		0xa5c7U,0x5b85U,0xb70aU,0x7e1fU,0xfc3eU,0xe877U,0xc0e5U,0x91c1U,0x3389U,0x6712U,
		0xce24U,0x8c43U,0x088dU,0x111aU,0x2234U,0x4468U,0x88d0U,0x01abU,0x0356U,0x06acU,
		0x0d58U,0x1ab0U,0x3560U,0x6ac0U,0xd580U,0xbb0bU,0x661dU,0xcc3aU,0x887fU,0x00f5U,
		0x01eaU,0x03d4U,0x07a8U,0x0f50U,0x1ea0U,0x3d40U,0x7a80U,0xf500U,0xfa0bU,0xe41dU,
		0xd831U,0xa069U,0x50d9U,0xa1b2U,0x536fU,0xa6deU,0x5db7U,0xbb6eU,0x66d7U,0xcdaeU,
		0x8b57U,0x06a5U,0x0d4aU,0x1a94U,0x3528U,0x6a50U,0xd4a0U,0xb94bU,0x629dU,0xc53aU,
		0x9a7fU,0x24f5U,0x49eaU,0x93d4U,0x37a3U,0x6f46U,0xde8cU,0xad13U,0x4a2dU,0x945aU,
		0x38bfU,0x717eU,0xe2fcU,0xd5f3U,0xbbedU,0x67d1U,0xcfa2U,0x8f4fU,0x0e95U,0x1d2aU,
		0x3a54U,0x74a8U,0xe950U,0xc2abU,0x955dU,0x3ab1U,0x7562U,0xeac4U,0xc583U,0x9b0dU,
		0x2611U,0x4c22U,0x9844U,0x2083U,0x4106U,0x820cU,0x1413U,0x2826U,0x504cU,0xa098U,
		0x513bU,0xa276U,0x54e7U,0xa9ceU,0x4397U,0x872eU,0x1e57U,0x3caeU,0x795cU,0xf2b8U,
		0xf57bU,0xfafdU,0xe5f1U,0xdbe9U,0xa7d9U,0x5fb9U,0xbf72U,0x6eefU,0xdddeU,0xabb7U,
		0x4765U,0x8ecaU,0x0d9fU,0x1b3eU,0x367cU,0x6cf8U,0xd9f0U,0xa3ebU,0x57ddU,0xafbaU,
		0x4f7fU,0x9efeU,0x2df7U,0x5beeU,0xb7dcU,0x7fb3U,0xff66U,0xeec7U,0xcd85U,0x8b01U,
		0x0609U,0x0c12U,0x1824U,0x3048U,0x6090U,0xc120U,0x924bU,0x349dU,0x693aU,0xd274U,
		0xb4e3U,0x79cdU,0xf39aU,0xf73fU,0xfe75U,0xece1U,0xc9c9U,0x8399U,0x1739U,0x2e72U,
		0x5ce4U,0xb9c8U,0x639bU,0xc736U,0x9e67U,0x2cc5U,0x598aU,0xb314U,0x7623U,0xec46U,
		0xc887U,0x8105U,0x1201U,0x2402U,0x4804U,0x9008U,0x301bU,0x6036U,0xc06cU,0x90d3U,
		0x31adU,0x635aU,0xc6b4U,0x9d63U,0x2acdU,0x559aU,0xab34U,0x4663U,0x8cc6U,0x0987U,
		0x130eU,0x261cU,0x4c38U,0x9870U,0x20ebU,0x41d6U,0x83acU,0x1753U,0x2ea6U,0x5d4cU,
		0xba98U,0x653bU,0xca76U,0x84e7U,0x19c5U,0x338aU,0x6714U,0xce28U,0x8c5bU,0x08bdU,
		0x117aU,0x22f4U,0x45e8U,0x8bd0U,0x07abU,0x0f56U,0x1eacU,0x3d58U,0x7ab0U,0xf560U,
		0xfacbU,0xe59dU,0xdb31U,0xa669U,0x5cd9U,0xb9b2U,0x636fU,0xc6deU,0x9db7U,0x2b65U,
		0x56caU,0xad94U,0x4b23U,0x9646U,0x3c87U,0x790eU,0xf21cU,0xf433U,0xf86dU,0xe0d1U,
		0xd1a9U,0xb359U,0x76b9U,0xed72U,0xcaefU,0x85d5U,0x1ba1U,0x3742U,0x6e84U,0xdd08U,
		0xaa1bU,0x443dU,0x887aU,0x00ffU,0x01feU,0x03fcU,0x07f8U,0x0ff0U,0x1fe0U,0x3fc0U,
		0x7f80U,0xff00U,0xee0bU,0xcc1dU,0x8831U,0x0069U,0x00d2U,0x01a4U,0x0348U,0x0690U,
		0x0d20U,0x1a40U,0x3480U,0x6900U,0xd200U,0xb40bU,0x781dU,0xf03aU,0xf07fU,0xf0f5U,
		0xf1e1U,0xf3c9U,0xf799U,0xff39U,0xee79U,0xccf9U,0x89f9U,0x03f9U,0x07f2U,0x0fe4U,
		0x1fc8U,0x3f90U,0x7f20U,0xfe40U,0xec8bU,0xc91dU,0x8231U,0x1469U,0x28d2U,0x51a4U,
		0xa348U,0x569bU,0xad36U,0x4a67U,0x94ceU,0x3997U,0x732eU,0xe65cU,0xdcb3U,0xa96dU,
		0x42d1U,0x85a2U,0x1b4fU,0x369eU,0x6d3cU,0xda78U,0xa4fbU,0x59fdU,0xb3faU,0x77ffU,
		0xeffeU,0xcff7U,0x8fe5U,0x0fc1U,0x1f82U,0x3f04U,0x7e08U,0xfc10U,0xe82bU,0xc05dU,
		0x90b1U,0x3169U,0x62d2U,0xc5a4U,0x9b43U,0x268dU,0x4d1aU,0x9a34U,0x2463U,0x48c6U,
		0x918cU,0x3313U,0x6626U,0xcc4cU,0x8893U,0x012dU,0x025aU,0x04b4U,0x0968U,0x12d0U,
		0x25a0U,0x4b40U,0x9680U,0x3d0bU,0x7a16U,0xf42cU,0xf853U,0xe0adU,0xd151U,0xb2a9U,
		0x7559U,0xeab2U,0xc56fU,0x9ad5U,0x25a1U,0x4b42U,0x9684U,0x3d03U,0x7a06U,0xf40cU,
		0xf813U,0xe02dU,0xd051U,0xb0a9U,0x7159U,0xe2b2U,0xd56fU,0xbad5U,0x65a1U,0xcb42U,
		0x868fU,0x1d15U,0x3a2aU,0x7454U,0xe8a8U,0xc15bU,0x92bdU,0x3571U,0x6ae2U,0xd5c4U,
		0xbb83U,0x670dU,0xce1aU,0x8c3fU,0x0875U,0x10eaU,0x21d4U,0x43a8U,0x8750U,0x1eabU,
		0x3d56U,0x7aacU,0xf558U,0xfabbU,0xe57dU,0xdaf1U,0xa5e9U,0x5bd9U,0xb7b2U,0x7f6fU,
		0xfedeU,0xedb7U,0xcb65U,0x86c1U,0x1d89U,0x3b12U,0x7624U,0xec48U,0xc89bU,0x813dU,
		0x1271U,0x24e2U,0x49c4U,0x9388U,0x371bU,0x6e36U,0xdc6cU,0xa8d3U,0x41adU,0x835aU,
		0x16bfU,0x2d7eU,0x5afcU,0xb5f8U,0x7bfbU,0xf7f6U,0xffe7U,0xefc5U,0xcf81U,0x8f09U,
		0x0e19U,0x1c32U,0x3864U,0x70c8U,0xe190U,0xd32bU,0xb65dU,0x7cb1U,0xf962U,0xe2cfU,
		0xd595U,0xbb21U,0x6649U,0xcc92U,0x892fU,0x0255U,0x04aaU,0x0954U,0x12a8U,0x2550U,
		0x4aa0U,0x9540U,0x3a8bU,0x7516U,0xea2cU,0xc453U,0x98adU,0x2151U,0x42a2U,0x8544U,
		0x1a83U,0x3506U,0x6a0cU,0xd418U,0xb83bU,0x607dU,0xc0faU,0x91ffU,0x33f5U,0x67eaU,
		0xcfd4U,0x8fa3U,0x0f4dU,0x1e9aU,0x3d34U,0x7a68U,0xf4d0U,0xf9abU,0xe35dU,0xd6b1U,
		0xbd69U,0x6ad9U,0xd5b2U,0xbb6fU,0x66d5U,0xcdaaU,0x8b5fU,0x06b5U,0x0d6aU,0x1ad4U,
		0x35a8U,0x6b50U,0xd6a0U,0xbd4bU,0x6a9dU,0xd53aU,0xba7fU,0x64f5U,0xc9eaU,0x83dfU,
		0x17b5U,0x2f6aU,0x5ed4U,0xbda8U,0x6b5bU,0xd6b6U,0xbd67U,0x6ac5U,0xd58aU,0xbb1fU,
		0x6635U,0xcc6aU,0x88dfU,0x01b5U,0x036aU,0x06d4U,0x0da8U,0x1b50U,0x36a0U,0x6d40U,
		0xda80U,0xa50bU,0x5a1dU,0xb43aU,0x787fU,0xf0feU,0xf1f7U,0xf3e5U,0xf7c1U,0xff89U,
		0xef19U,0xce39U,0x8c79U,0x08f9U,0x11f2U,0x23e4U,0x47c8U,0x8f90U,0x0f2bU,0x1e56U,
		0x3cacU,0x7958U,0xf2b0U,0xf56bU,0xfaddU,0xe5b1U,0xdb69U,0xa6d9U,0x5db9U,0xbb72U,
		0x66efU,0xcddeU,0x8bb7U,0x0765U,0x0ecaU,0x1d94U,0x3b28U,0x7650U,0xeca0U,0xc94bU,
		0x829dU,0x1531U,0x2a62U,0x54c4U,0xa988U,0x431bU,0x8636U,0x1c67U,0x38ceU,0x719cU,
		0xe338U,0xd67bU,0xbcfdU,0x69f1U,0xd3e2U,0xb7cfU,0x7f95U,0xff2aU,0xee5fU,0xccb5U,
		0x8961U,0x02c9U,0x0592U,0x0b24U,0x1648U,0x2c90U,0x5920U,0xb240U,0x748bU,0xe916U,
		0xc227U,0x9445U,0x3881U,0x7102U,0xe204U,0xd403U,0xb80dU,0x6011U,0xc022U,0x904fU,
		0x3095U,0x612aU,0xc254U,0x94a3U,0x394dU,0x729aU,0xe534U,0xda63U,0xa4cdU,0x5991U,
		0xb322U,0x764fU,0xec9eU,0xc937U,0x8265U,0x14c1U,0x2982U,0x5304U,0xa608U,0x5c1bU,
		0xb836U,0x6067U,0xc0ceU,0x9197U,0x3325U,0x664aU,0xcc94U,0x8923U,0x024dU,0x049aU,
		0x0934U,0x1268U,0x24d0U,0x49a0U,0x9340U,0x368bU,0x6d16U,0xda2cU,0xa453U,0x58adU,
		0xb15aU,0x72bfU,0xe57eU,0xdaf7U,0xa5e5U,0x5bc1U,0xb782U,0x7f0fU,0xfe1eU,0xec37U,
		0xc865U,0x80c1U,0x1189U,0x2312U,0x4624U,0x8c48U,0x089bU,0x1136U,0x226cU,0x44d8U,
		0x89b0U,0x036bU,0x06d6U,0x0dacU,0x1b58U,0x36b0U,0x6d60U,0xdac0U,0xa58bU,0x5b1dU,
		0xb63aU,0x7c7fU,0xf8feU,0xe1f7U,0xd3e5U,0xb7c1U,0x7f89U,0xff12U,0xee2fU,0xcc55U,
		0x88a1U,0x0149U,0x0292U,0x0524U,0x0a48U,0x1490U,0x2920U,0x5240U,0xa480U,0x590bU,
		0xb216U,0x7427U,0xe84eU,0xc097U,0x9125U,0x3241U,0x6482U,0xc904U,0x8203U,0x140dU,
		0x281aU,0x5034U,0xa068U,0x50dbU,0xa1b6U,0x5367U,0xa6ceU,0x5d97U,0xbb2eU,0x6657U,
		0xccaeU,0x8957U,0x02a5U,0x054aU,0x0a94U,0x1528U,0x2a50U,0x54a0U,0xa940U,0x428bU,
		0x8516U,0x1a27U,0x344eU,0x689cU,0xd138U,0xb27bU,0x74fdU,0xe9faU,0xc3ffU,0x97f5U,
		0x3fe1U,0x7fc2U,0xff84U,0xef03U,0xce0dU,0x8c11U,0x0829U,0x1052U,0x20a4U,0x4148U,
		0x8290U,0x152bU,0x2a56U,0x54acU,0xa958U,0x42bbU,0x8576U,0x1ae7U,0x35ceU,0x6b9cU,
		0xd738U,0xbe7bU,0x6cfdU,0xd9faU,0xa3ffU,0x57f5U,0xafeaU,0x4fdfU,0x9fbeU,0x2f77U,
		0x5eeeU,0xbddcU,0x6bb3U,0xd766U,0xbec7U,0x6d85U,0xdb0aU,0xa61fU,0x5c35U,0xb86aU,
		0x60dfU,0xc1beU,0x9377U,0x36e5U,0x6dcaU,0xdb94U,0xa723U,0x5e4dU,0xbc9aU,0x693fU,
		0xd27eU,0xb4f7U,0x79e5U,0xf3caU,0xf79fU,0xff35U,0xee61U,0xccc9U,0x8999U,0x0339U,
		0x0672U,0x0ce4U,0x19c8U,0x3390U,0x6720U,0xce40U,0x8c8bU,0x091dU,0x123aU,0x2474U,
		0x48e8U,0x91d0U,0x33abU,0x6756U,0xceacU,0x8d53U,0x0aadU,0x155aU,0x2ab4U,0x5568U,
		0xaad0U,0x45abU,0x8b56U,0x06a7U,0x0d4eU,0x1a9cU,0x3538U,0x6a70U,0xd4e0U,0xb9cbU,
		0x639dU,0xc73aU,0x9e7fU,0x2cf5U,0x59eaU,0xb3d4U,0x77a3U,0xef46U,0xce87U,0x8d05U,
		0x0a01U,0x1402U,0x2804U,0x5008U,0xa010U,0x502bU,0xa056U,0x50a7U,0xa14eU,0x5297U,
		0xa52eU,0x5a57U,0xb4aeU,0x7957U,0xf2aeU,0xf557U,0xfaa5U,0xe541U,0xda89U,0xa519U,
		0x5a39U,0xb472U,0x78efU,0xf1deU,0xf3b7U,0xf765U,0xfec1U,0xed89U,0xcb19U,0x8639U,
		0x1c79U,0x38f2U,0x71e4U,0xe3c8U,0xd79bU,0xbf3dU,0x6e71U,0xdce2U,0xa9cfU,0x4395U,
		0x872aU,0x1e5fU,0x3cbeU,0x797cU,0xf2f8U,0xf5fbU,0xfbfdU,0xe7f1U,0xdfe9U,0xafd9U,
		0x4fb9U,0x9f72U,0x2eefU,0x5ddeU,0xbbbcU,0x6773U,0xcee6U,0x8dc7U,0x0b85U,0x170aU,
		0x2e14U,0x5c28U,0xb850U,0x60abU,0xc156U,0x92a7U,0x3545U,0x6a8aU,0xd514U,0xba23U,
		0x644dU,0xc89aU,0x813fU,0x1275U,0x24eaU,0x49d4U,0x93a8U,0x375bU,0x6eb6U,0xdd6cU,
		0xaad3U,0x45adU,0x8b5aU,0x06bfU,0x0d7eU,0x1afcU,0x35f8U,0x6bf0U,0xd7e0U,0xbfcbU,
		0x6f9dU,0xdf3aU,0xae7fU,0x4cf5U,0x99eaU,0x23dfU,0x47beU,0x8f7cU,0x0ef3U,0x1de6U,
		0x3bccU,0x7798U,0xef30U,0xce6bU,0x8cddU,0x09b1U,0x1362U,0x26c4U,0x4d88U,0x9b10U,
		0x262bU,0x4c56U,0x98acU,0x2153U,0x42a6U,0x854cU,0x1a93U,0x3526U,0x6a4cU,0xd498U,
		0xb93bU,0x627dU,0xc4faU,0x99ffU,0x23f5U,0x47eaU,0x8fd4U,0x0fa3U,0x1f46U,0x3e8cU,
		0x7d18U,0xfa30U,0xe46bU,0xd8ddU,0xa1b1U,0x5369U,0xa6d2U,0x5dafU,0xbb5eU,0x66b7U,
		0xcd6eU,0x8ad7U,0x05a5U,0x0b4aU,0x1694U,0x2d28U,0x5a50U,0xb4a0U,0x794bU,0xf296U,
		0xf527U,0xfa45U,0xe481U,0xd909U,0xa219U,0x5439U,0xa872U,0x40efU,0x81deU,0x13b7U,
		0x276eU,0x4edcU,0x9db8U,0x2b7bU,0x56f6U,0xadecU,0x4bd3U,0x97a6U,0x3f47U,0x7e8eU,
		0xfd1cU,0xea33U,0xc46dU,0x98d1U,0x21a9U,0x4352U,0x86a4U,0x1d43U,0x3a86U,0x750cU,
		0xea18U,0xc43bU,0x987dU,0x20f1U,0x41e2U,0x83c4U,0x1783U,0x2f06U,0x5e0cU,0xbc18U,
		0x683bU,0xd076U,0xb0e7U,0x71c5U,0xe38aU,0xd71fU,0xbe35U,0x6c61U,0xd8c2U,0xa18fU,
		0x5315U,0xa62aU,0x5c5fU,0xb8beU,0x6177U,0xc2eeU,0x95d7U,0x3ba5U,0x774aU,0xee94U,
		0xcd23U,0x8a4dU,0x0491U,0x0922U,0x1244U,0x2488U,0x4910U,0x9220U,0x344bU,0x6896U,
		0xd12cU,0xb253U,0x74adU,0xe95aU,0xc2bfU,0x9575U,0x3ae1U,0x75c2U,0xeb84U,0xc703U,
		0x9e0dU,0x2c11U,0x5822U,0xb044U,0x7083U,0xe106U,0xd207U,0xb405U,0x7801U,0xf002U,
		0xf00fU,0xf015U,0xf021U,0xf049U,0xf099U,0xf139U,0xf279U,0xf4f9U,0xf9f9U,0xe3f9U,
		0xd7f9U,0xbff9U,0x6ff9U,0xdff2U,0xafefU,0x4fd5U,0x9faaU,0x2f5fU,0x5ebeU,0xbd7cU,
		0x6af3U,0xd5e6U,0xbbc7U,0x6785U,0xcf0aU,0x8e1fU,0x0c35U,0x186aU,0x30d4U,0x61a8U,
		0xc350U,0x96abU,0x3d5dU,0x7abaU,0xf574U,0xfae3U,0xe5cdU,0xdb91U,0xa729U,0x5e59U,
		0xbcb2U,0x696fU,0xd2deU,0xb5b7U,0x7b65U,0xf6caU,0xfd9fU,0xeb35U,0xc661U,0x9cc9U,
		0x2999U,0x5332U,0xa664U,0x5cc3U,0xb986U,0x6307U,0xc60eU,0x9c17U,0x2825U,0x504aU,
		0xa094U,0x5123U,0xa246U,0x5487U,0xa90eU,0x4217U,0x842eU,0x1857U,0x30aeU,0x615cU,
		0xc2b8U,0x957bU,0x3afdU,0x75faU,0xebf4U,0xc7e3U,0x9fcdU,0x2f91U,0x5f22U,0xbe44U,
		0x6c83U,0xd906U,0xa207U,0x5405U,0xa80aU,0x401fU,0x803eU,0x1077U,0x20eeU,0x41dcU,
		0x83b8U,0x177bU,0x2ef6U,0x5decU,0xbbd8U,0x67bbU,0xcf76U,0x8ee7U,0x0dc5U,0x1b8aU,
		0x3714U,0x6e28U,0xdc50U,0xa8abU,0x415dU,0x82baU,0x157fU,0x2afeU,0x55fcU,0xabf8U,
		0x47fbU,0x8ff6U,0x0fe7U,0x1fceU,0x3f9cU,0x7f38U,0xfe70U,0xecebU,0xc9ddU,0x83b1U,
		0x1769U,0x2ed2U,0x5da4U,0xbb48U,0x669bU,0xcd36U,0x8a67U,0x04c5U,0x098aU,0x1314U,
		0x2628U,0x4c50U,0x98a0U,0x214bU,0x4296U,0x852cU,0x1a53U,0x34a6U,0x694cU,0xd298U,
		0xb53bU,0x7a7dU,0xf4faU,0xf9ffU,0xe3f5U,0xd7e1U,0xbfc9U,0x6f99U,0xdf32U,0xae6fU,
		0x4cd5U,0x99aaU,0x235fU,0x46beU,0x8d7cU,0x0af3U,0x15e6U,0x2bccU,0x5798U,0xaf30U,
		0x4e6bU,0x9cd6U,0x29a7U,0x534eU,0xa69cU,0x5d33U,0xba66U,0x64c7U,0xc98eU,0x8317U,
		0x1625U,0x2c4aU,0x5894U,0xb128U,0x725bU,0xe4b6U,0xd967U,0xa2c5U,0x5581U,0xab02U,
		0x460fU,0x8c1eU,0x0837U,0x106eU,0x20dcU,0x41b8U,0x8370U,0x16ebU,0x2dd6U,0x5bacU,
		0xb758U,0x7ebbU,0xfd76U,0xeae7U,0xc5c5U,0x9b81U,0x2709U,0x4e12U,0x9c24U,0x2843U,
		0x5086U,0xa10cU,0x5213U,0xa426U,0x5847U,0xb08eU,0x7117U,0xe22eU,0xd457U,0xb8a5U,
		0x6141U,0xc282U,0x950fU,0x3a15U,0x742aU,0xe854U,0xc0a3U,0x914dU,0x3291U,0x6522U,
		0xca44U,0x8483U,0x190dU,0x321aU,0x6434U,0xc868U,0x80dbU,0x11bdU,0x237aU,0x46f4U,
		0x8de8U,0x0bdbU,0x17b6U,0x2f6cU,0x5ed8U,0xbdb0U,0x6b6bU,0xd6d6U,0xbda7U,0x6b45U,
		0xd68aU,0xbd1fU,0x6a35U,0xd46aU,0xb8dfU,0x61b5U,0xc36aU,0x96dfU,0x3db5U,0x7b6aU,
		0xf6d4U,0xfda3U,0xeb4dU,0xc691U,0x9d29U,0x2a59U,0x54b2U,0xa964U,0x42c3U,0x8586U,
		0x1b07U,0x360eU,0x6c1cU,0xd838U,0xa07bU,0x50fdU,0xa1faU,0x53ffU,0xa7feU,0x5ff7U,
		0xbfeeU,0x6fd7U,0xdfaeU,0xaf57U,0x4ea5U,0x9d4aU,0x2a9fU,0x553eU,0xaa7cU,0x44f3U,
		0x89e6U,0x03c7U,0x078eU,0x0f1cU,0x1e38U,0x3c70U,0x78e0U,0xf1c0U,0xf38bU,0xf71dU,
		0xfe31U,0xec69U,0xc8d9U,0x81b9U,0x1379U,0x26f2U,0x4de4U,0x9bc8U,0x279bU,0x4f36U,
		0x9e6cU,0x2cd3U,0x59a6U,0xb34cU,0x7693U,0xed26U,0xca47U,0x8485U,0x1901U,0x3202U,
		0x6404U,0xc808U,0x801bU,0x103dU,0x207aU,0x40f4U,0x81e8U,0x13dbU,0x27b6U,0x4f6cU,
		0x9ed8U,0x2dbbU,0x5b76U,0xb6ecU,0x7dd3U,0xfba6U,0xe747U,0xde85U,0xad01U,0x4a09U,
		0x9412U,0x382fU,0x705eU,0xe0bcU,0xd173U,0xb2edU,0x75d1U,0xeba2U,0xc74fU,0x9e95U,
		0x2d21U,0x5a42U,0xb484U,0x7903U,0xf206U,0xf407U,0xf805U,0xe001U,0xd009U,0xb019U,
		0x7039U,0xe072U,0xd0efU,0xb1d5U,0x73a1U,0xe742U,0xde8fU,0xad15U,0x4a21U,0x9442U,
		0x388fU,0x711eU,0xe23cU,0xd473U,0xb8edU,0x61d1U,0xc3a2U,0x974fU,0x3e95U,0x7d2aU,
		0xfa54U,0xe4a3U,0xd94dU,0xa291U,0x5529U,0xaa52U,0x44afU,0x895eU,0x02b7U,0x056eU,
		0x0adcU,0x15b8U,0x2b70U,0x56e0U,0xadc0U,0x4b8bU,0x9716U,0x3e27U,0x7c4eU,0xf89cU,
		0xe133U,0xd26dU,0xb4d1U,0x79a9U,0xf352U,0xf6afU,0xfd55U,0xeaa1U,0xc549U,0x9a99U,
		0x2539U,0x4a72U,0x94e4U,0x39c3U,0x7386U,0xe70cU,0xde13U,0xac2dU,0x4851U,0x90a2U,
		0x314fU,0x629eU,0xc53cU,0x9a73U,0x24edU,0x49daU,0x93b4U,0x3763U,0x6ec6U,0xdd8cU,
		0xab13U,0x462dU,0x8c5aU,0x08bfU,0x117eU,0x22fcU,0x45f8U,0x8bf0U,0x07ebU,0x0fd6U,
		0x1facU,0x3f58U,0x7eb0U,0xfd60U,0xeacbU,0xc59dU,0x9b31U,0x2669U,0x4cd2U,0x99a4U,
		0x2343U,0x4686U,0x8d0cU,0x0a13U,0x1426U,0x284cU,0x5098U,0xa130U,0x526bU,0xa4d6U,
		0x59a7U,0xb34eU,0x7697U,0xed2eU,0xca57U,0x84a5U,0x1941U,0x3282U,0x6504U,0xca08U,
		0x841bU,0x183dU,0x307aU,0x60f4U,0xc1e8U,0x93dbU,0x37bdU,0x6f7aU,0xdef4U,0xade3U,
		0x4bcdU,0x979aU,0x3f3fU,0x7e7eU,0xfcfcU,0xe9f3U,0xc3edU,0x97d1U,0x3fa9U,0x7f52U,
		0xfea4U,0xed43U,0xca8dU,0x8511U,0x1a29U,0x3452U,0x68a4U,0xd148U,0xb29bU,0x753dU,
		0xea7aU,0xc4ffU,0x99f5U,0x23e1U,0x47c2U,0x8f84U,0x0f03U,0x1e06U,0x3c0cU,0x7818U,
		0xf030U,0xf06bU,0xf0ddU,0xf1b1U,0xf369U,0xf6d9U,0xfdb9U,0xeb79U,0xc6f9U,0x9df9U,
		0x2bf9U,0x57f2U,0xafe4U,0x4fc3U,0x9f86U,0x2f07U,0x5e0eU,0xbc1cU,0x6833U,0xd066U,
		0xb0c7U,0x7185U,0xe30aU,0xd61fU,0xbc35U,0x6861U,0xd0c2U,0xb18fU,0x7315U,0xe62aU,
		0xdc5fU,0xa8b5U,0x4161U,0x82c2U,0x158fU,0x2b1eU,0x563cU,0xac78U,0x48fbU,0x91f6U,
		0x33e7U,0x67ceU,0xcf9cU,0x8f33U,0x0e6dU,0x1cdaU,0x39b4U,0x7368U,0xe6d0U,0xddabU,
		0xab5dU,0x46b1U,0x8d62U,0x0acfU,0x159eU,0x2b3cU,0x5678U,0xacf0U,0x49ebU,0x93d6U,
		0x37a7U,0x6f4eU,0xde9cU,0xad33U,0x4a6dU,0x94daU,0x39bfU,0x737eU,0xe6fcU,0xddf3U,
		0xabedU,0x47d1U,0x8fa2U,0x0f4fU,0x1e9eU,0x3d3cU,0x7a78U,0xf4f0U,0xf9ebU,0xe3ddU,
		0xd7b1U,0xbf69U,0x6ed9U,0xddb2U,0xab6fU,0x46d5U,0x8daaU,0x0b5fU,0x16beU,0x2d7cU,
		0x5af8U,0xb5f0U,0x7bebU,0xf7d6U,0xffa7U,0xef45U,0xce81U,0x8d09U,0x0a19U,0x1432U,
		0x2864U,0x50c8U,0xa190U,0x532bU,0xa656U,0x5ca7U,0xb94eU,0x6297U,0xc52eU,0x9a57U,
		0x24a5U,0x494aU,0x9294U,0x3523U,0x6a46U,0xd48cU,0xb913U,0x622dU,0xc45aU,0x98bfU,
		0x2175U,0x42eaU,0x85d4U,0x1ba3U,0x3746U,0x6e8cU,0xdd18U,0xaa3bU,0x447dU,0x88faU,
		0x01ffU,0x03feU,0x07fcU,0x0ff8U,0x1ff0U,0x3fe0U,0x7fc0U,0xff80U,0xef0bU,0xce1dU,
		0x8c31U,0x0869U,0x10d2U,0x21a4U,0x4348U,0x8690U,0x1d2bU,0x3a56U,0x74acU,0xe958U,
		0xc2bbU,0x957dU,0x3af1U,0x75e2U,0xebc4U,0xc783U,0x9f0dU,0x2e11U,0x5c22U,0xb844U,
		0x6083U,0xc106U,0x9207U,0x3405U,0x680aU,0xd014U,0xb023U,0x704dU,0xe09aU,0xd13fU,
		0xb275U,0x74e1U,0xe9c2U,0xc38fU,0x9715U,0x3e21U,0x7c42U,0xf884U,0xe103U,0xd20dU,
		0xb411U,0x7829U,0xf052U,0xf0afU,0xf155U,0xf2a1U,0xf549U,0xfa99U,0xe539U,0xda79U,
		0xa4f9U,0x59f9U,0xb3f2U,0x77efU,0xefdeU,0xcfb7U,0x8f65U,0x0ec1U,0x1d82U,0x3b04U,
		0x7608U,0xec10U,0xc82bU,0x805dU,0x10b1U,0x2162U,0x42c4U,0x8588U,0x1b1bU,0x3636U,
		0x6c6cU,0xd8d8U,0xa1bbU,0x537dU,0xa6faU,0x5dffU,0xbbfeU,0x67f7U,0xcfeeU,0x8fd7U,
		0x0fa5U,0x1f4aU,0x3e94U,0x7d28U,0xfa50U,0xe4abU,0xd95dU,0xa2b1U,0x5569U,0xaad2U,
		0x45afU,0x8b5eU,0x06b7U,0x0d6eU,0x1adcU,0x35b8U,0x6b70U,0xd6e0U,0xbdcbU,0x6b9dU,
		0xd73aU,0xbe7fU,0x6cf5U,0xd9eaU,0xa3dfU,0x57b5U,0xaf6aU,0x4edfU,0x9dbeU,0x2b77U,
		0x56eeU,0xaddcU,0x4bb3U,0x9766U,0x3ec7U,0x7d8eU,0xfb1cU,0xe633U,0xdc6dU,0xa8d1U,
		0x41a9U,0x8352U,0x16afU,0x2d5eU,0x5abcU,0xb578U,0x7afbU,0xf5f6U,0xfbe7U,0xe7c5U,
		0xdf81U,0xaf09U,0x4e19U,0x9c32U,0x286fU,0x50deU,0xa1bcU,0x5373U,0xa6e6U,0x5dc7U,
		0xbb8eU,0x6717U,0xce2eU,0x8c57U,0x08a5U,0x114aU,0x2294U,0x4528U,0x8a50U,0x04abU,
		0x0956U,0x12acU,0x2558U,0x4ab0U,0x9560U,0x3acbU,0x7596U,0xeb2cU,0xc653U,0x9cadU,
		0x2951U,0x52a2U,0xa544U,0x5a83U,0xb506U,0x7a07U,0xf40eU,0xf817U,0xe025U,0xd041U,
		0xb089U,0x7119U,0xe232U,0xd46fU,0xb8d5U,0x61a1U,0xc342U,0x968fU,0x3d15U,0x7a2aU,
		0xf454U,0xf8a3U,0xe14dU,0xd291U,0xb529U,0x7a59U,0xf4b2U,0xf96fU,0xe2d5U,0xd5a1U,
		0xbb49U,0x6699U,0xcd32U,0x8a6fU,0x04d5U,0x09aaU,0x1354U,0x26a8U,0x4d50U,0x9aa0U,
		0x254bU,0x4a96U,0x952cU,0x3a53U,0x74a6U,0xe94cU,0xc293U,0x952dU,0x3a51U,0x74a2U,
		0xe944U,0xc283U,0x950dU,0x3a11U,0x7422U,0xe844U,0xc083U,0x910dU,0x3211U,0x6422U,
		0xc844U,0x8083U,0x110dU,0x221aU,0x4434U,0x8868U,0x00dbU,0x01b6U,0x036cU,0x06d8U,
		0x0db0U,0x1b60U,0x36c0U,0x6d80U,0xdb00U,0xa60bU,0x5c1dU,0xb83aU,0x607fU,0xc0feU,
		0x91f7U,0x33e5U,0x67caU,0xcf94U,0x8f23U,0x0e4dU,0x1c9aU,0x3934U,0x7268U,0xe4d0U,
		0xd9abU,0xa35dU,0x56b1U,0xad62U,0x4acfU,0x959eU,0x3b37U,0x766eU,0xecdcU,0xc9b3U,
		0x836dU,0x16d1U,0x2da2U,0x5b44U,0xb688U,0x7d1bU,0xfa36U,0xe467U,0xd8c5U,0xa181U,
		0x5309U,0xa612U,0x5c2fU,0xb85eU,0x60b7U,0xc16eU,0x92d7U,0x35a5U,0x6b4aU,0xd694U,
		0xbd23U,0x6a4dU,0xd49aU,0xb93fU,0x6275U,0xc4eaU,0x99dfU,0x23b5U,0x476aU,0x8ed4U,
		0x0da3U,0x1b46U,0x368cU,0x6d18U,0xda30U,0xa46bU,0x58ddU,0xb1baU,0x737fU,0xe6feU,
		0xddf7U,0xabe5U,0x47c1U,0x8f82U,0x0f0fU,0x1e1eU,0x3c3cU,0x7878U,0xf0f0U,0xf1ebU,
		0xf3ddU,0xf7b1U,0xff69U,0xeed9U,0xcdb9U,0x8b79U,0x06f9U,0x0df2U,0x1be4U,0x37c8U,
		0x6f90U,0xdf20U,0xae4bU,0x4c9dU,0x993aU,0x227fU,0x44feU,0x89fcU,0x03f3U,0x07e6U,
		0x0fccU,0x1f98U,0x3f30U,0x7e60U,0xfcc0U,0xe98bU,0xc31dU,0x9631U,0x3c69U,0x78d2U,
		0xf1a4U,0xf343U,0xf68dU,0xfd11U,0xea29U,0xc459U,0x98b9U,0x2179U,0x42f2U,0x85e4U,
		0x1bc3U,0x3786U,0x6f0cU,0xde18U,0xac3bU,0x487dU,0x90faU,0x31ffU,0x63feU,0xc7fcU,
		0x9ff3U,0x2fedU,0x5fdaU,0xbfb4U,0x6f63U,0xdec6U,0xad87U,0x4b05U,0x960aU,0x3c1fU,
		0x783eU,0xf07cU,0xf0f3U,0xf1edU,0xf3d1U,0xf7a9U,0xff59U,0xeeb9U,0xcd79U,0x8af9U,
		0x05f9U,0x0bf2U,0x17e4U,0x2fc8U,0x5f90U,0xbf20U,0x6e4bU,0xdc96U,0xa927U,0x4245U,
		0x848aU,0x191fU,0x323eU,0x647cU,0xc8f8U,0x81fbU,0x13fdU,0x27faU,0x4ff4U,0x9fe8U,
		0x2fdbU,0x5fb6U,0xbf6cU,0x6ed3U,0xdda6U,0xab47U,0x4685U,0x8d0aU,0x0a1fU,0x143eU,
		0x287cU,0x50f8U,0xa1f0U,0x53ebU,0xa7d6U,0x5fa7U,0xbf4eU,0x6e97U,0xdd2eU,0xaa57U,
		0x44a5U,0x894aU,0x029fU,0x053eU,0x0a7cU,0x14f8U,0x29f0U,0x53e0U,0xa7c0U,0x5f8bU,
		0xbf16U,0x6e27U,0xdc4eU,0xa897U,0x4125U,0x824aU,0x149fU,0x293eU,0x527cU,0xa4f8U,
		0x59fbU,0xb3f6U,0x77e7U,0xefceU,0xcf97U,0x8f25U,0x0e41U,0x1c82U,0x3904U,0x7208U,
		0xe410U,0xd82bU,0xa05dU,0x50b1U,0xa162U,0x52cfU,0xa59eU,0x5b37U,0xb66eU,0x7cd7U,
		0xf9aeU,0xe357U,0xd6a5U,0xbd41U,0x6a89U,0xd512U,0xba2fU,0x6455U,0xc8aaU,0x815fU,
		0x12b5U,0x256aU,0x4ad4U,0x95a8U,0x3b5bU,0x76b6U,0xed6cU,0xcad3U,0x85adU,0x1b51U,
		0x36a2U,0x6d44U,0xda88U,0xa51bU,0x5a3dU,0xb47aU,0x78ffU,0xf1feU,0xf3f7U,0xf7e5U,
		0xffc1U,0xef89U,0xcf19U,0x8e39U,0x0c79U,0x18f2U,0x31e4U,0x63c8U,0xc790U,0x9f2bU,
		0x2e5dU,0x5cbaU,0xb974U,0x62e3U,0xc5c6U,0x9b87U,0x2705U,0x4e0aU,0x9c14U,0x2823U,
		0x5046U,0xa08cU,0x5113U,0xa226U,0x5447U,0xa88eU,0x4117U,0x822eU,0x1457U,0x28aeU,
		0x515cU,0xa2b8U,0x557bU,0xaaf6U,0x45e7U,0x8bceU,0x0797U,0x0f2eU,0x1e5cU,0x3cb8U,
		0x7970U,0xf2e0U,0xf5cbU,0xfb9dU,0xe731U,0xde69U,0xacd9U,0x49b9U,0x9372U,0x36efU,
		0x6ddeU,0xdbbcU,0xa773U,0x5eedU,0xbddaU,0x6bbfU,0xd77eU,0xbef7U,0x6de5U,0xdbcaU,
		0xa79fU,0x5f35U,0xbe6aU,0x6cdfU,0xd9beU,0xa377U,0x56e5U,0xadcaU,0x4b9fU,0x973eU,
		0x3e77U,0x7ceeU,0xf9dcU,0xe3b3U,0xd76dU,0xbed1U,0x6da9U,0xdb52U,0xa6afU,0x5d55U,
		0xbaaaU,0x655fU,0xcabeU,0x8577U,0x1ae5U,0x35caU,0x6b94U,0xd728U,0xbe5bU,0x6cbdU,
		0xd97aU,0xa2ffU,0x55f5U,0xabeaU,0x47dfU,0x8fbeU,0x0f77U,0x1eeeU,0x3ddcU,0x7bb8U,
		0xf770U,0xfeebU,0xedddU,0xcbb1U,0x8769U,0x1ed9U,0x3db2U,0x7b64U,0xf6c8U,0xfd9bU,
		0xeb3dU,0xc671U,0x9ce9U,0x29d9U,0x53b2U,0xa764U,0x5ec3U,0xbd86U,0x6b07U,0xd60eU,
		0xbc17U,0x6825U,0xd04aU,0xb09fU,0x7135U,0xe26aU,0xd4dfU,0xb9b5U,0x6361U,0xc6c2U,
		0x9d8fU,0x2b15U,0x562aU,0xac54U,0x48a3U,0x9146U,0x3287U,0x650eU,0xca1cU,0x8433U,
		0x186dU,0x30daU,0x61b4U,0xc368U,0x96dbU,0x3dbdU,0x7b7aU,0xf6f4U,0xfde3U,0xebcdU,
		0xc791U,0x9f29U,0x2e59U,0x5cb2U,0xb964U,0x62c3U,0xc586U,0x9b07U,0x2605U,0x4c0aU,
		0x9814U,0x2023U,0x4046U,0x808cU,0x1113U,0x2226U,0x444cU,0x8898U,0x013bU,0x0276U,
		0x04ecU,0x09d8U,0x13b0U,0x2760U,0x4ec0U,0x9d80U,0x2b0bU,0x5616U,0xac2cU,0x4853U,
		0x90a6U,0x3147U,0x628eU,0xc51cU,0x9a33U,0x246dU,0x48daU,0x91b4U,0x3363U,0x66c6U,
		0xcd8cU,0x8b13U,0x062dU,0x0c5aU,0x18b4U,0x3168U,0x62d0U,0xc5a0U,0x9b4bU,0x269dU,
		0x4d3aU,0x9a74U,0x24e3U,0x49c6U,0x938cU,0x3713U,0x6e26U,0xdc4cU,0xa893U,0x412dU,
		0x825aU,0x14bfU,0x297eU,0x52fcU,0xa5f8U,0x5bfbU,0xb7f6U,0x7fe7U,0xffceU,0xef97U,
		0xcf25U,0x8e41U,0x0c89U,0x1912U,0x3224U,0x6448U,0xc890U,0x812bU,0x125dU,0x24baU,
		0x4974U,0x92e8U,0x35dbU,0x6bb6U,0xd76cU,0xbed3U,0x6dadU,0xdb5aU,0xa6bfU,0x5d75U,
		0xbaeaU,0x65dfU,0xcbbeU,0x8777U,0x1ee5U,0x3dcaU,0x7b94U,0xf728U,0xfe5bU,0xecbdU,
		0xc971U,0x82e9U,0x15d9U,0x2bb2U,0x5764U,0xaec8U,0x4d9bU,0x9b36U,0x2667U,0x4cceU,
		0x999cU,0x2333U,0x4666U,0x8cccU,0x0993U,0x1326U,0x264cU,0x4c98U,0x9930U,0x226bU,
		0x44d6U,0x89acU,0x0353U,0x06a6U,0x0d4cU,0x1a98U,0x3530U,0x6a60U,0xd4c0U,0xb98bU,
		0x631dU,0xc63aU,0x9c7fU,0x28f5U,0x51eaU,0xa3d4U,0x57a3U,0xaf46U,0x4e87U,0x9d0eU,
		0x2a17U,0x542eU,0xa85cU,0x40b3U,0x8166U,0x12c7U,0x258eU,0x4b1cU,0x9638U,0x3c7bU,
		0x78f6U,0xf1ecU,0xf3d3U,0xf7adU,0xff51U,0xeea9U,0xcd59U,0x8ab9U,0x0579U,0x0af2U,
		0x15e4U,0x2bc8U,0x5790U,0xaf20U,0x4e4bU,0x9c96U,0x2927U,0x524eU,0xa49cU,0x5933U,
		0xb266U,0x74c7U,0xe98eU,0xc317U,0x9625U,0x3c41U,0x7882U,0xf104U,0xf203U,0xf40dU,
		0xf811U,0xe029U,0xd059U,0xb0b9U,0x7179U,0xe2f2U,0xd5efU,0xbbd5U,0x67a1U,0xcf42U,
		0x8e8fU,0x0d15U,0x1a2aU,0x3454U,0x68a8U,0xd150U,0xb2abU,0x755dU,0xeabaU,0xc57fU,
		0x9af5U,0x25e1U,0x4bc2U,0x9784U,0x3f03U,0x7e06U,0xfc0cU,0xe813U,0xc02dU,0x9051U,
		0x30a9U,0x6152U,0xc2a4U,0x9543U,0x3a8dU,0x751aU,0xea34U,0xc463U,0x98cdU,0x2191U,
		0x4322U,0x8644U,0x1c83U,0x3906U,0x720cU,0xe418U,0xd83bU,0xa07dU,0x50f1U,0xa1e2U,
		0x53cfU,0xa79eU,0x5f37U,0xbe6eU,0x6cd7U,0xd9aeU,0xa357U,0x56a5U,0xad4aU,0x4a9fU,
		0x953eU,0x3a77U,0x74eeU,0xe9dcU,0xc3b3U,0x976dU,0x3ed1U,0x7da2U,0xfb44U,0xe683U,
		0xdd0dU,0xaa11U,0x4429U,0x8852U,0x00afU,0x015eU,0x02bcU,0x0578U,0x0af0U,0x15e0U,
		0x2bc0U,0x5780U,0xaf00U,0x4e0bU,0x9c16U,0x2827U,0x504eU,0xa09cU,0x5133U,0xa266U,
		0x54c7U,0xa98eU,0x4317U,0x862eU,0x1c57U,0x38aeU,0x715cU,0xe2b8U,0xd57bU,0xbafdU,
		0x65f1U,0xcbe2U,0x87cfU,0x1f95U,0x3f2aU,0x7e54U,0xfca8U,0xe95bU,0xc2bdU,0x9571U,
		0x3ae9U,0x75d2U,0xeba4U,0xc743U,0x9e8dU,0x2d11U,0x5a22U,0xb444U,0x7883U,0xf106U,
		0xf207U,0xf405U,0xf801U,0xe009U,0xd019U,0xb039U,0x7079U,0xe0f2U,0xd1efU,0xb3d5U,
		0x77a1U,0xef42U,0xce8fU,0x8d15U,0x0a21U,0x1442U,0x2884U,0x5108U,0xa210U,0x542bU,
		0xa856U,0x40a7U,0x814eU,0x1297U,0x252eU,0x4a5cU,0x94b8U,0x397bU,0x72f6U,0xe5ecU,
		0xdbd3U,0xa7adU,0x5f51U,0xbea2U,0x6d4fU,0xda9eU,0xa537U,0x5a65U,0xb4caU,0x799fU,
		0xf33eU,0xf677U,0xfce5U,0xe9c1U,0xc389U,0x9719U,0x3e39U,0x7c72U,0xf8e4U,0xe1c3U,
		0xd38dU,0xb711U,0x7e29U,0xfc52U,0xe8afU,0xc155U,0x92a1U,0x3549U,0x6a92U,0xd524U,
		0xba43U,0x648dU,0xc91aU,0x823fU,0x1475U,0x28eaU,0x51d4U,0xa3a8U,0x575bU,0xaeb6U,
		0x4d67U,0x9aceU,0x2597U,0x4b2eU,0x965cU,0x3cb3U,0x7966U,0xf2ccU,0xf593U,0xfb2dU,
		0xe651U,0xdca9U,0xa959U,0x42b9U,0x8572U,0x1aefU,0x35deU,0x6bbcU,0xd778U,0xbefbU,
		0x6dfdU,0xdbfaU,0xa7ffU,0x5ff5U,0xbfeaU,0x6fdfU,0xdfbeU,0xaf77U,0x4ee5U,0x9dcaU,
		0x2b9fU,0x573eU,0xae7cU,0x4cf3U,0x99e6U,0x23c7U,0x478eU,0x8f1cU,0x0e33U,0x1c66U,
		0x38ccU,0x7198U,0xe330U,0xd66bU,0xbcddU,0x69b1U,0xd362U,0xb6cfU,0x7d95U,0xfb2aU,
		0xe65fU,0xdcb5U,0xa961U,0x42c9U,0x8592U,0x1b2fU,0x365eU,0x6cbcU,0xd978U,0xa2fbU,
		0x55fdU,0xabfaU,0x47ffU,0x8ffeU,0x0ff7U,0x1feeU,0x3fdcU,0x7fb8U,0xff70U,0xeeebU,
		0xcdddU,0x8bb1U,0x0769U,0x0ed2U,0x1da4U,0x3b48U,0x7690U,0xed20U,0xca4bU,0x849dU,
		0x1931U,0x3262U,0x64c4U,0xc988U,0x831bU,0x163dU,0x2c7aU,0x58f4U,0xb1e8U,0x73dbU,
		0xe7b6U,0xdf67U,0xaec5U,0x4d81U,0x9b02U,0x260fU,0x4c1eU,0x983cU,0x2073U,0x40e6U,
		0x81ccU,0x1393U,0x2726U,0x4e4cU,0x9c98U,0x293bU,0x5276U,0xa4ecU,0x59d3U,0xb3a6U,
		0x7747U,0xee8eU,0xcd17U,0x8a25U,0x0441U,0x0882U,0x1104U,0x2208U,0x4410U,0x8820U,
		0x004bU,0x0096U,0x012cU,0x0258U,0x04b0U,0x0960U,0x12c0U,0x2580U,0x4b00U,0x9600U,
		0x3c0bU,0x7816U,0xf02cU,0xf053U,0xf0adU,0xf151U,0xf2a9U,0xf559U,0xfab9U,0xe579U,
		0xdaf9U,0xa5f9U,0x5bf9U,0xb7f2U,0x7fefU,0xffdeU,0xefb7U,0xcf65U,0x8ec1U,0x0d89U,
		0x1b12U,0x3624U,0x6c48U,0xd890U,0xa12bU,0x525dU,0xa4baU,0x597fU,0xb2feU,0x75f7U,
		0xebeeU,0xc7d7U,0x9fa5U,0x2f41U,0x5e82U,0xbd04U,0x6a03U,0xd406U,0xb807U,0x6005U,
		0xc00aU,0x901fU,0x3035U,0x606aU,0xc0d4U,0x91a3U,0x334dU,0x669aU,0xcd34U,0x8a63U,
		0x04cdU,0x099aU,0x1334U,0x2668U,0x4cd0U,0x99a0U,0x234bU,0x4696U,0x8d2cU,0x0a53U,
		0x14a6U,0x294cU,0x5298U,0xa530U,0x5a6bU,0xb4d6U,0x79a7U,0xf34eU,0xf697U,0xfd25U,
		0xea41U,0xc489U,0x9919U,0x2239U,0x4472U,0x88e4U,0x01c3U,0x0386U,0x070cU,0x0e18U,
		0x1c30U,0x3860U,0x70c0U,0xe180U,0xd30bU,0xb61dU,0x7c31U,0xf862U,0xe0cfU,0xd195U,
		0xb321U,0x7649U,0xec92U,0xc92fU,0x8255U,0x14a1U,0x2942U,0x5284U,0xa508U,0x5a1bU,
		0xb436U,0x7867U,0xf0ceU,0xf197U,0xf325U,0xf641U,0xfc89U,0xe919U,0xc239U,0x9479U,
		0x38f9U,0x71f2U,0xe3e4U,0xd7c3U,0xbf8dU,0x6f11U,0xde22U,0xac4fU,0x4895U,0x912aU,
		0x325fU,0x64beU,0xc97cU,0x82f3U,0x15edU,0x2bdaU,0x57b4U,0xaf68U,0x4edbU,0x9db6U,
		0x2b67U,0x56ceU,0xad9cU,0x4b33U,0x9666U,0x3cc7U,0x798eU,0xf31cU,0xf633U,0xfc6dU,
		0xe8d1U,0xc1a9U,0x9359U,0x36b9U,0x6d72U,0xdae4U,0xa5c3U,0x5b8dU,0xb71aU,0x7e3fU,
		0xfc7eU,0xe8f7U,0xc1e5U,0x93c1U,0x3789U,0x6f12U,0xde24U,0xac43U,0x488dU,0x911aU,
		0x323fU,0x647eU,0xc8fcU,0x81f3U,0x13edU,0x27daU,0x4fb4U,0x9f68U,0x2edbU,0x5db6U,
		0xbb6cU,0x66d3U,0xcda6U,0x8b47U,0x0685U,0x0d0aU,0x1a14U,0x3428U,0x6850U,0xd0a0U,
		0xb14bU,0x729dU,0xe53aU,0xda7fU,0xa4f5U,0x59e1U,0xb3c2U,0x778fU,0xef1eU,0xce37U,
		0x8c65U,0x08c1U,0x1182U,0x2304U,0x4608U,0x8c10U,0x082bU,0x1056U,0x20acU,0x4158U,
		0x82b0U,0x156bU,0x2ad6U,0x55acU,0xab58U,0x46bbU,0x8d76U,0x0ae7U,0x15ceU,0x2b9cU,
		0x5738U,0xae70U,0x4cebU,0x99d6U,0x23a7U,0x474eU,0x8e9cU,0x0d33U,0x1a66U,0x34ccU,
		0x6998U,0xd330U,0xb66bU,0x7cddU,0xf9baU,0xe37fU,0xd6f5U,0xbde1U,0x6bc9U,0xd792U,
		0xbf2fU,0x6e55U,0xdcaaU,0xa95fU,0x42b5U,0x856aU,0x1adfU,0x35beU,0x6b7cU,0xd6f8U,
		0xbdfbU,0x6bfdU,0xd7faU,0xbfffU,0x6ff5U,0xdfeaU,0xafdfU,0x4fb5U,0x9f6aU,0x2edfU,
		0x5dbeU,0xbb7cU,0x66f3U,0xcde6U,0x8bc7U,0x0785U,0x0f0aU,0x1e14U,0x3c28U,0x7850U,
		0xf0a0U,0xf14bU,0xf29dU,0xf531U,0xfa69U,0xe4d9U,0xd9b9U,0xa379U,0x56f9U,0xadf2U,
		0x4befU,0x97deU,0x3fb7U,0x7f6eU,0xfedcU,0xedb3U,0xcb6dU,0x86d1U,0x1da9U,0x3b52U,
		0x76a4U,0xed48U,0xca9bU,0x853dU,0x1a71U,0x34e2U,0x69c4U,0xd388U,0xb71bU,0x7e3dU,
		0xfc7aU,0xe8ffU,0xc1f5U,0x93e1U,0x37c9U,0x6f92U,0xdf24U,0xae43U,0x4c8dU,0x991aU,
		0x223fU,0x447eU,0x88fcU,0x01f3U,0x03e6U,0x07ccU,0x0f98U,0x1f30U,0x3e60U,0x7cc0U,
		0xf980U,0xe30bU,0xd61dU,0xbc31U,0x6869U,0xd0d2U,0xb1afU,0x7355U,0xe6aaU,0xdd5fU,
		0xaab5U,0x4561U,0x8ac2U,0x058fU,0x0b1eU,0x163cU,0x2c78U,0x58f0U,0xb1e0U,0x73cbU,
		0xe796U,0xdf27U,0xae45U,0x4c81U,0x9902U,0x220fU,0x441eU,0x883cU,0x0073U,0x00e6U,
		0x01ccU,0x0398U,0x0730U,0x0e60U,0x1cc0U,0x3980U,0x7300U,0xe600U,0xdc0bU,0xa81dU,
		0x4031U,0x8062U,0x10cfU,0x219eU,0x433cU,0x8678U,0x1cfbU,0x39f6U,0x73ecU,0xe7d8U,
		0xdfbbU,0xaf7dU,0x4ef1U,0x9de2U,0x2bcfU,0x579eU,0xaf3cU,0x4e73U,0x9ce6U,0x29c7U,
		0x538eU,0xa71cU,0x5e33U,0xbc66U,0x68c7U,0xd18eU,0xb317U,0x7625U,0xec4aU,0xc89fU,
		0x8135U,0x1261U,0x24c2U,0x4984U,0x9308U,0x361bU,0x6c36U,0xd86cU,0xa0d3U,0x51adU,
		0xa35aU,0x56bfU,0xad7eU,0x4af7U,0x95eeU,0x3bd7U,0x77aeU,0xef5cU,0xceb3U,0x8d6dU,
		0x0ad1U,0x15a2U,0x2b44U,0x5688U,0xad10U,0x4a2bU,0x9456U,0x38a7U,0x714eU,0xe29cU,
		0xd533U,0xba6dU,0x64d1U,0xc9a2U,0x834fU,0x1695U,0x2d2aU,0x5a54U,0xb4a8U,0x795bU,
		0xf2b6U,0xf567U,0xfac5U,0xe581U,0xdb09U,0xa619U,0x5c39U,0xb872U,0x60efU,0xc1deU,
		0x93b7U,0x3765U,0x6ecaU,0xdd94U,0xab23U,0x464dU,0x8c9aU,0x093fU,0x127eU,0x24fcU,
		0x49f8U,0x93f0U,0x37ebU,0x6fd6U,0xdfacU,0xaf53U,0x4eadU,0x9d5aU,0x2abfU,0x557eU,
		0xaafcU,0x45f3U,0x8be6U,0x07c7U,0x0f8eU,0x1f1cU,0x3e38U,0x7c70U,0xf8e0U,0xe1cbU,
		0xd39dU,0xb731U,0x7e69U,0xfcd2U,0xe9afU,0xc355U,0x96a1U,0x3d49U,0x7a92U,0xf524U,
		0xfa43U,0xe48dU,0xd911U,0xa229U,0x5459U,0xa8b2U,0x416fU,0x82deU,0x15b7U,0x2b6eU,
		0x56dcU,0xadb8U,0x4b7bU,0x96f6U,0x3de7U,0x7bceU,0xf79cU,0xff33U,0xee6dU,0xccd1U,
		0x89a9U,0x0359U,0x06b2U,0x0d64U,0x1ac8U,0x3590U,0x6b20U,0xd640U,0xbc8bU,0x691dU,
		0xd23aU,0xb47fU,0x78f5U,0xf1eaU,0xf3dfU,0xf7b5U,0xff61U,0xeec9U,0xcd99U,0x8b39U,
		0x0679U,0x0cf2U,0x19e4U,0x33c8U,0x6790U,0xcf20U,0x8e4bU,0x0c9dU,0x193aU,0x3274U,
		0x64e8U,0xc9d0U,0x83abU,0x175dU,0x2ebaU,0x5d74U,0xbae8U,0x65dbU,0xcbb6U,0x8767U,
		0x1ec5U,0x3d8aU,0x7b14U,0xf628U,0xfc5bU,0xe8bdU,0xc171U,0x92e9U,0x35d9U,0x6bb2U,
		0xd764U,0xbec3U,0x6d8dU,0xdb1aU,0xa63fU,0x5c75U,0xb8eaU,0x61dfU,0xc3beU,0x9777U,
		0x3ee5U,0x7dcaU,0xfb94U,0xe723U,0xde4dU,0xac91U,0x4929U,0x9252U,0x34afU,0x695eU,
		0xd2bcU,0xb573U,0x7aedU,0xf5daU,0xfbbfU,0xe775U,0xdee1U,0xadc9U,0x4b99U,0x9732U,
		0x3e6fU,0x7cdeU,0xf9bcU,0xe373U,0xd6edU,0xbdd1U,0x6ba9U,0xd752U,0xbeafU,0x6d55U,
		0xdaaaU,0xa55fU,0x5ab5U,0xb56aU,0x7adfU,0xf5beU,0xfb77U,0xe6e5U,0xddc1U,0xab89U,
		0x4719U,0x8e32U,0x0c6fU,0x18deU,0x31bcU,0x6378U,0xc6f0U,0x9debU,0x2bddU,0x57baU,
		0xaf74U,0x4ee3U,0x9dc6U,0x2b87U,0x570eU,0xae1cU,0x4c33U,0x9866U,0x20c7U,0x418eU,
		0x831cU,0x1633U,0x2c66U,0x58ccU,0xb198U,0x733bU,0xe676U,0xdce7U,0xa9c5U,0x4381U,
		0x8702U,0x1e0fU,0x3c1eU,0x783cU,0xf078U,0xf0fbU,0xf1fdU,0xf3f1U,0xf7e9U,0xffd9U,
		0xefb9U,0xcf79U,0x8ef9U,0x0df9U,0x1bf2U,0x37e4U,0x6fc8U,0xdf90U,0xaf2bU,0x4e5dU,
		0x9cbaU,0x297fU,0x52feU,0xa5fcU,0x5bf3U,0xb7e6U,0x7fc7U,0xff8eU,0xef17U,0xce25U,
		0x8c41U,0x0889U,0x1112U,0x2224U,0x4448U,0x8890U,0x012bU,0x0256U,0x04acU,0x0958U,
		0x12b0U,0x2560U,0x4ac0U,0x9580U,0x3b0bU,0x7616U,0xec2cU,0xc853U,0x80adU,0x1151U,
		0x22a2U,0x4544U,0x8a88U,0x051bU,0x0a36U,0x146cU,0x28d8U,0x51b0U,0xa360U,0x56cbU,
		0xad96U,0x4b27U,0x964eU,0x3c97U,0x792eU,0xf25cU,0xf4b3U,0xf96dU,0xe2d1U,0xd5a9U,
		0xbb59U,0x66b9U,0xcd72U,0x8aefU,0x05d5U,0x0baaU,0x1754U,0x2ea8U,0x5d50U,0xbaa0U,
		0x654bU,0xca96U,0x8527U,0x1a45U,0x348aU,0x6914U,0xd228U,0xb45bU,0x78bdU,0xf17aU,
		0xf2ffU,0xf5f5U,0xfbe1U,0xe7c9U,0xdf99U,0xaf39U,0x4e79U,0x9cf2U,0x29efU,0x53deU,
		0xa7bcU,0x5f73U,0xbee6U,0x6dc7U,0xdb8eU,0xa717U,0x5e25U,0xbc4aU,0x689fU,0xd13eU,
		0xb277U,0x74e5U,0xe9caU,0xc39fU,0x9735U,0x3e61U,0x7cc2U,0xf984U,0xe303U,0xd60dU,
		0xbc11U,0x6829U,0xd052U,0xb0afU,0x7155U,0xe2aaU,0xd55fU,0xbab5U,0x6561U,0xcac2U,
		0x858fU,0x1b15U,0x362aU,0x6c54U,0xd8a8U,0xa15bU,0x52bdU,0xa57aU,0x5affU,0xb5feU,
		0x7bf7U,0xf7eeU,0xffd7U,0xefa5U,0xcf41U,0x8e89U,0x0d19U,0x1a32U,0x3464U,0x68c8U,
		0xd190U,0xb32bU,0x765dU,0xecbaU,0xc97fU,0x82f5U,0x15e1U,0x2bc2U,0x5784U,0xaf08U,
		0x4e1bU,0x9c36U,0x2867U,0x50ceU,0xa19cU,0x5333U,0xa666U,0x5cc7U,0xb98eU,0x6317U,
		0xc62eU,0x9c57U,0x28a5U,0x514aU,0xa294U,0x5523U,0xaa46U,0x4487U,0x890eU,0x0217U,
		0x042eU,0x085cU,0x10b8U,0x2170U,0x42e0U,0x85c0U,0x1b8bU,0x3716U,0x6e2cU,0xdc58U,
		0xa8bbU,0x417dU,0x82faU,0x15ffU,0x2bfeU,0x57fcU,0xaff8U,0x4ffbU,0x9ff6U,0x2fe7U,
		0x5fceU,0xbf9cU,0x6f33U,0xde66U,0xacc7U,0x4985U,0x930aU,0x361fU,0x6c3eU,0xd87cU,
		0xa0f3U,0x51edU,0xa3daU,0x57bfU,0xaf7eU,0x4ef7U,0x9deeU,0x2bd7U,0x57aeU,0xaf5cU,
		0x4eb3U,0x9d66U,0x2ac7U,0x558eU,0xab1cU,0x4633U,0x8c66U,0x08c7U,0x118eU,0x231cU,
		0x4638U,0x8c70U,0x08ebU,0x11d6U,0x23acU,0x4758U,0x8eb0U,0x0d6bU,0x1ad6U,0x35acU,
		0x6b58U,0xd6b0U,0xbd6bU,0x6addU,0xd5baU,0xbb7fU,0x66f5U,0xcdeaU,0x8bdfU,0x07b5U,
		0x0f6aU,0x1ed4U,0x3da8U,0x7b50U,0xf6a0U,0xfd4bU,0xea9dU,0xc531U,0x9a69U,0x24d9U,
		0x49b2U,0x9364U,0x36c3U,0x6d86U,0xdb0cU,0xa613U,0x5c2dU,0xb85aU,0x60bfU,0xc17eU,
		0x92f7U,0x35e5U,0x6bcaU,0xd794U,0xbf23U,0x6e4dU,0xdc9aU,0xa93fU,0x4275U,0x84eaU,
		0x19dfU,0x33beU,0x677cU,0xcef8U,0x8dfbU,0x0bfdU,0x17faU,0x2ff4U,0x5fe8U,0xbfd0U,
		0x6fabU,0xdf56U,0xaea7U,0x4d45U,0x9a8aU,0x251fU,0x4a3eU,0x947cU,0x38f3U,0x71e6U,
		0xe3ccU,0xd793U,0xbf2dU,0x6e51U,0xdca2U,0xa94fU,0x4295U,0x852aU,0x1a5fU,0x34beU,
		0x697cU,0xd2f8U,0xb5fbU,0x7bfdU,0xf7faU,0xffffU,0xeff5U,0xcfe1U,0x8fc9U,0x0f99U,
		0x1f32U,0x3e64U,0x7cc8U,0xf990U,0xe32bU,0xd65dU,0xbcb1U,0x6969U,0xd2d2U,0xb5afU,
		0x7b55U,0xf6aaU,0xfd5fU,0xeab5U,0xc561U,0x9ac9U,0x2599U,0x4b32U,0x9664U,0x3cc3U,
		0x7986U,0xf30cU,0xf613U,0xfc2dU,0xe851U,0xc0a9U,0x9159U,0x32b9U,0x6572U,0xcae4U,
		0x85c3U,0x1b8dU,0x371aU,0x6e34U,0xdc68U,0xa8dbU,0x41bdU,0x837aU,0x16ffU,0x2dfeU,
		0x5bfcU,0xb7f8U,0x7ffbU,0xfff6U,0xefe7U,0xcfc5U,0x8f81U,0x0f09U,0x1e12U,0x3c24U,
		0x7848U,0xf090U,0xf12bU,0xf25dU,0xf4b1U,0xf969U,0xe2d9U,0xd5b9U,0xbb79U,0x66f9U,
		0xcdf2U,0x8befU,0x07d5U,0x0faaU,0x1f54U,0x3ea8U,0x7d50U,0xfaa0U,0xe54bU,0xda9dU,
		0xa531U,0x5a69U,0xb4d2U,0x79afU,0xf35eU,0xf6b7U,0xfd65U,0xeac1U,0xc589U,0x9b19U,
		0x2639U,0x4c72U,0x98e4U,0x21c3U,0x4386U,0x870cU,0x1e13U,0x3c26U,0x784cU,0xf098U,
		0xf13bU,0xf27dU,0xf4f1U,0xf9e9U,0xe3d9U,0xd7b9U,0xbf79U,0x6ef9U,0xddf2U,0xabefU,
		0x47d5U,0x8faaU,0x0f5fU,0x1ebeU,0x3d7cU,0x7af8U,0xf5f0U,0xfbebU,0xe7ddU,0xdfb1U,
		0xaf69U,0x4ed9U,0x9db2U,0x2b6fU,0x56deU,0xadbcU,0x4b73U,0x96e6U,0x3dc7U,0x7b8eU,
		0xf71cU,0xfe33U,0xec6dU,0xc8d1U,0x81a9U,0x1359U,0x26b2U,0x4d64U,0x9ac8U,0x259bU,
		0x4b36U,0x966cU,0x3cd3U,0x79a6U,0xf34cU,0xf693U,0xfd2dU,0xea51U,0xc4a9U,0x9959U,
		0x22b9U,0x4572U,0x8ae4U,0x05c3U,0x0b86U,0x170cU,0x2e18U,0x5c30U,0xb860U,0x60cbU,
		0xc196U,0x9327U,0x3645U,0x6c8aU,0xd914U,0xa223U,0x544dU,0xa89aU,0x413fU,0x827eU,
		0x14f7U,0x29eeU,0x53dcU,0xa7b8U,0x5f7bU,0xbef6U,0x6de7U,0xdbceU,0xa797U,0x5f25U,
		0xbe4aU,0x6c9fU,0xd93eU,0xa277U,0x54e5U,0xa9caU,0x439fU,0x873eU,0x1e77U,0x3ceeU,
		0x79dcU,0xf3b8U,0xf77bU,0xfefdU,0xedf1U,0xcbe9U,0x87d9U,0x1fb9U,0x3f72U,0x7ee4U,
		0xfdc8U,0xeb9bU,0xc73dU,0x9e71U,0x2ce9U,0x59d2U,0xb3a4U,0x7743U,0xee86U,0xcd07U,
		0x8a05U,0x0401U,0x0802U,0x1004U,0x2008U,0x4010U,0x8020U,0x104bU,0x2096U,0x412cU,
		0x8258U,0x14bbU,0x2976U,0x52ecU,0xa5d8U,0x5bbbU,0xb776U,0x7ee7U,0xfdceU,0xeb97U,
		0xc725U,0x9e41U,0x2c89U,0x5912U,0xb224U,0x7443U,0xe886U,0xc107U,0x9205U,0x3401U,
		0x6802U,0xd004U,0xb003U,0x700dU,0xe01aU,0xd03fU,0xb075U,0x70e1U,0xe1c2U,0xd38fU,
		0xb715U,0x7e21U,0xfc42U,0xe88fU,0xc115U,0x9221U,0x3449U,0x6892U,0xd124U,0xb243U,
		0x748dU,0xe91aU,0xc23fU,0x9475U,0x38e1U,0x71c2U,0xe384U,0xd703U,0xbe0dU,0x6c11U,
		0xd822U,0xa04fU,0x5095U,0xa12aU,0x525fU,0xa4beU,0x5977U,0xb2eeU,0x75d7U,0xebaeU,
		0xc757U,0x9ea5U,0x2d41U,0x5a82U,0xb504U,0x7a03U,0xf406U,0xf807U,0xe005U,0xd001U,
		0xb009U,0x7019U,0xe032U,0xd06fU,0xb0d5U,0x71a1U,0xe342U,0xd68fU,0xbd15U,0x6a21U,
		0xd442U,0xb88fU,0x6115U,0xc22aU,0x945fU,0x38b5U,0x716aU,0xe2d4U,0xd5a3U,0xbb4dU,
		0x6691U,0xcd22U,0x8a4fU,0x0495U,0x092aU,0x1254U,0x24a8U,0x4950U,0x92a0U,0x354bU,
		0x6a96U,0xd52cU,0xba53U,0x64adU,0xc95aU,0x82bfU,0x1575U,0x2aeaU,0x55d4U,0xaba8U,
		0x475bU,0x8eb6U,0x0d67U,0x1aceU,0x359cU,0x6b38U,0xd670U,0xbcebU,0x69ddU,0xd3baU,
		0xb77fU,0x7ef5U,0xfdeaU,0xebdfU,0xc7b5U,0x9f61U,0x2ec9U,0x5d92U,0xbb24U,0x6643U,
		0xcc86U,0x8907U,0x0205U,0x040aU,0x0814U,0x1028U,0x2050U,0x40a0U,0x8140U,0x128bU,
		0x2516U,0x4a2cU,0x9458U,0x38bbU,0x7176U,0xe2ecU,0xd5d3U,0xbbadU,0x6751U,0xcea2U,
		0x8d4fU,0x0a95U,0x152aU,0x2a54U,0x54a8U,0xa950U,0x42abU,0x8556U,0x1aa7U,0x354eU,
		0x6a9cU,0xd538U,0xba7bU,0x64fdU,0xc9faU,0x83ffU,0x17f5U,0x2feaU,0x5fd4U,0xbfa8U,
		0x6f5bU,0xdeb6U,0xad67U,0x4ac5U,0x958aU,0x3b1fU,0x763eU,0xec7cU,0xc8f3U,0x81edU,
		0x13d1U,0x27a2U,0x4f44U,0x9e88U,0x2d1bU,0x5a36U,0xb46cU,0x78d3U,0xf1a6U,0xf347U,
		0xf685U,0xfd01U,0xea09U,0xc419U,0x9839U,0x2079U,0x40f2U,0x81e4U,0x13c3U,0x2786U,
		0x4f0cU,0x9e18U,0x2c3bU,0x5876U,0xb0ecU,0x71d3U,0xe3a6U,0xd747U,0xbe85U,0x6d01U,
		0xda02U,0xa40fU,0x5815U,0xb02aU,0x705fU,0xe0beU,0xd177U,0xb2e5U,0x75c1U,0xeb82U,
		0xc70fU,0x9e15U,0x2c21U,0x5842U,0xb084U,0x7103U,0xe206U,0xd407U,0xb805U,0x6001U,
		0xc002U,0x900fU,0x3015U,0x602aU,0xc054U,0x90a3U,0x314dU,0x629aU,0xc534U,0x9a63U,
		0x24cdU,0x499aU,0x9334U,0x3663U,0x6cc6U,0xd98cU,0xa313U,0x562dU,0xac5aU,0x48bfU,
		0x917eU,0x32f7U,0x65eeU,0xcbdcU,0x87b3U,0x1f6dU,0x3edaU,0x7db4U,0xfb68U,0xe6dbU,
		0xddbdU,0xab71U,0x46e9U,0x8dd2U,0x0bafU,0x175eU,0x2ebcU,0x5d78U,0xbaf0U,0x65ebU,
		0xcbd6U,0x87a7U,0x1f45U,0x3e8aU,0x7d14U,0xfa28U,0xe45bU,0xd8bdU,0xa171U,0x52e9U,
		0xa5d2U,0x5bafU,0xb75eU,0x7eb7U,0xfd6eU,0xead7U,0xc5a5U,0x9b41U,0x2689U,0x4d12U,
		0x9a24U,0x2443U,0x4886U,0x910cU,0x3213U,0x6426U,0xc84cU,0x8093U,0x112dU,0x225aU,
		0x44b4U,0x8968U,0x02dbU,0x05b6U,0x0b6cU,0x16d8U,0x2db0U,0x5b60U,0xb6c0U,0x7d8bU,
		0xfb16U,0xe627U,0xdc45U,0xa881U,0x4109U,0x8212U,0x142fU,0x285eU,0x50bcU,0xa178U,
		0x52fbU,0xa5f6U,0x5be7U,0xb7ceU,0x7f97U,0xff2eU,0xee57U,0xcca5U,0x8941U,0x0289U,
		0x0512U,0x0a24U,0x1448U,0x2890U,0x5120U,0xa240U,0x548bU,0xa916U,0x4227U,0x844eU,
		0x1897U,0x312eU,0x625cU,0xc4b8U,0x997bU,0x22fdU,0x45faU,0x8bf4U,0x07e3U,0x0fc6U,
		0x1f8cU,0x3f18U,0x7e30U,0xfc60U,0xe8cbU,0xc19dU,0x9331U,0x3669U,0x6cd2U,0xd9a4U,
		0xa343U,0x568dU,0xad1aU,0x4a3fU,0x947eU,0x38f7U,0x71eeU,0xe3dcU,0xd7b3U,0xbf6dU,
		0x6ed1U,0xdda2U,0xab4fU,0x4695U,0x8d2aU,0x0a5fU,0x14beU,0x297cU,0x52f8U,0xa5f0U,
		0x5bebU,0xb7d6U,0x7fa7U,0xff4eU,0xee97U,0xcd25U,0x8a41U,0x0489U,0x0912U,0x1224U,
		0x2448U,0x4890U,0x9120U,0x324bU,0x6496U,0xc92cU,0x8253U,0x14adU,0x295aU,0x52b4U,
		0xa568U,0x5adbU,0xb5b6U,0x7b67U,0xf6ceU,0xfd97U,0xeb25U,0xc641U,0x9c89U,0x2919U,
		0x5232U,0xa464U,0x58c3U,0xb186U,0x7307U,0xe60eU,0xdc17U,0xa825U,0x4041U,0x8082U,
		0x110fU,0x221eU,0x443cU,0x8878U,0x00fbU,0x01f6U,0x03ecU,0x07d8U,0x0fb0U,0x1f60U,
		0x3ec0U,0x7d80U,0xfb00U,0xe60bU,0xdc1dU,0xa831U,0x4069U,0x80d2U,0x11afU,0x235eU,
		0x46bcU,0x8d78U,0x0afbU,0x15f6U,0x2becU,0x57d8U,0xafb0U,0x4f6bU,0x9ed6U,0x2da7U,
		0x5b4eU,0xb69cU,0x7d33U,0xfa66U,0xe4c7U,0xd985U,0xa301U,0x5609U,0xac12U,0x482fU,
		0x905eU,0x30b7U,0x616eU,0xc2dcU,0x95b3U,0x3b6dU,0x76daU,0xedb4U,0xcb63U,0x86cdU,
		0x1d91U,0x3b22U,0x7644U,0xec88U,0xc91bU,0x823dU,0x1471U,0x28e2U,0x51c4U,0xa388U,
		0x571bU,0xae36U,0x4c67U,0x98ceU,0x2197U,0x432eU,0x865cU,0x1cb3U,0x3966U,0x72ccU,
		0xe598U,0xdb3bU,0xa67dU,0x5cf1U,0xb9e2U,0x63cfU,0xc79eU,0x9f37U,0x2e65U,0x5ccaU,
		0xb994U,0x6323U,0xc646U,0x9c87U,0x2905U,0x520aU,0xa414U,0x5823U,0xb046U,0x7087U,
		0xe10eU,0xd217U,0xb425U,0x7841U,0xf082U,0xf10fU,0xf215U,0xf421U,0xf849U,0xe099U,
		0xd139U,0xb279U,0x74f9U,0xe9f2U,0xc3efU,0x97d5U,0x3fa1U,0x7f42U,0xfe84U,0xed03U,
		0xca0dU,0x8411U,0x1829U,0x3052U,0x60a4U,0xc148U,0x929bU,0x353dU,0x6a7aU,0xd4f4U,
		0xb9e3U,0x63cdU,0xc79aU,0x9f3fU,0x2e75U,0x5ceaU,0xb9d4U,0x63a3U,0xc746U,0x9e87U,
		0x2d05U,0x5a0aU,0xb414U,0x7823U,0xf046U,0xf087U,0xf105U,0xf201U,0xf409U,0xf819U,
		0xe039U,0xd079U,0xb0f9U,0x71f9U,0xe3f2U,0xd7efU,0xbfd5U,0x6fa1U,0xdf42U,0xae8fU,
		0x4d15U,0x9a2aU,0x245fU,0x48beU,0x917cU,0x32f3U,0x65e6U,0xcbccU,0x8793U,0x1f2dU,
		0x3e5aU,0x7cb4U,0xf968U,0xe2dbU,0xd5bdU,0xbb71U,0x66e9U,0xcdd2U,0x8bafU,0x0755U,
		0x0eaaU,0x1d54U,0x3aa8U,0x7550U,0xeaa0U,0xc54bU,0x9a9dU,0x2531U,0x4a62U,0x94c4U,
		0x3983U,0x7306U,0xe60cU,0xdc13U,0xa82dU,0x4051U,0x80a2U,0x114fU,0x229eU,0x453cU,
		0x8a78U,0x04fbU,0x09f6U,0x13ecU,0x27d8U,0x4fb0U,0x9f60U,0x2ecbU,0x5d96U,0xbb2cU,
		0x6653U,0xcca6U,0x8947U,0x0285U,0x050aU,0x0a14U,0x1428U,0x2850U,0x50a0U,0xa140U,
		0x528bU,0xa516U,0x5a27U,0xb44eU,0x7897U,0xf12eU,0xf257U,0xf4a5U,0xf941U,0xe289U,
		0xd519U,0xba39U,0x6479U,0xc8f2U,0x81efU,0x13d5U,0x27aaU,0x4f54U,0x9ea8U,0x2d5bU,
		0x5ab6U,0xb56cU,0x7ad3U,0xf5a6U,0xfb47U,0xe685U,0xdd01U,0xaa09U,0x4419U,0x8832U,
		0x006fU,0x00deU,0x01bcU,0x0378U,0x06f0U,0x0de0U,0x1bc0U,0x3780U,0x6f00U,0xde00U,
		0xac0bU,0x481dU,0x903aU,0x307fU,0x60feU,0xc1fcU,0x93f3U,0x37edU,0x6fdaU,0xdfb4U,
		0xaf63U,0x4ecdU,0x9d9aU,0x2b3fU,0x567eU,0xacfcU,0x49f3U,0x93e6U,0x37c7U,0x6f8eU,
		0xdf1cU,0xae33U,0x4c6dU,0x98daU,0x21bfU,0x437eU,0x86fcU,0x1df3U,0x3be6U,0x77ccU,
		0xef98U,0xcf3bU,0x8e7dU,0x0cf1U,0x19e2U,0x33c4U,0x6788U,0xcf10U,0x8e2bU,0x0c5dU,
		0x18baU,0x3174U,0x62e8U,0xc5d0U,0x9babU,0x275dU,0x4ebaU,0x9d74U,0x2ae3U,0x55c6U,
		0xab8cU,0x4713U,0x8e26U,0x0c47U,0x188eU,0x311cU,0x6238U,0xc470U,0x98ebU,0x21ddU,
		0x43baU,0x8774U,0x1ee3U,0x3dc6U,0x7b8cU,0xf718U,0xfe3bU,0xec7dU,0xc8f1U,0x81e9U,
		0x13d9U,0x27b2U,0x4f64U,0x9ec8U,0x2d9bU,0x5b36U,0xb66cU,0x7cd3U,0xf9a6U,0xe347U,
		0xd685U,0xbd01U,0x6a09U,0xd412U,0xb82fU,0x6055U,0xc0aaU,0x915fU,0x32b5U,0x656aU,
		0xcad4U,0x85a3U,0x1b4dU,0x369aU,0x6d34U,0xda68U,0xa4dbU,0x59bdU,0xb37aU,0x76ffU,
		0xedfeU,0xcbf7U,0x87e5U,0x1fc1U,0x3f82U,0x7f04U,0xfe08U,0xec1bU,0xc83dU,0x8071U,
		0x10e9U,0x21d2U,0x43a4U,0x8748U,0x1e9bU,0x3d36U,0x7a6cU,0xf4d8U,0xf9bbU,0xe37dU,
		0xd6f1U,0xbde9U,0x6bd9U,0xd7b2U,0xbf6fU,0x6ed5U,0xddaaU,0xab5fU,0x46b5U,0x8d6aU,
		0x0adfU,0x15beU,0x2b7cU,0x56f8U,0xadf0U,0x4bebU,0x97d6U,0x3fa7U,0x7f4eU,0xfe9cU,
		0xed33U,0xca6dU,0x84d1U,0x19a9U,0x3352U,0x66a4U,0xcd48U,0x8a9bU,0x053dU,0x0a7aU,
		0x14f4U,0x29e8U,0x53d0U,0xa7a0U,0x5f4bU,0xbe96U,0x6d27U,0xda4eU,0xa497U,0x5925U,
		0xb24aU,0x749fU,0xe93eU,0xc277U,0x94e5U,0x39c1U,0x7382U,0xe704U,0xde03U,0xac0dU,
		0x4811U,0x9022U,0x304fU,0x609eU,0xc13cU,0x9273U,0x34edU,0x69daU,0xd3b4U,0xb763U,
		0x7ecdU,0xfd9aU,0xeb3fU,0xc675U,0x9ce1U,0x29c9U,0x5392U,0xa724U,0x5e43U,0xbc86U,
		0x6907U,0xd20eU,0xb417U,0x7825U,0xf04aU,0xf09fU,0xf135U,0xf261U,0xf4c9U,0xf999U,
		0xe339U,0xd679U,0xbcf9U,0x69f9U,0xd3f2U,0xb7efU,0x7fd5U,0xffaaU,0xef5fU,0xceb5U,
		0x8d61U,0x0ac9U,0x1592U,0x2b24U,0x5648U,0xac90U,0x492bU,0x9256U,0x34a7U,0x694eU,
		0xd29cU,0xb533U,0x7a6dU,0xf4daU,0xf9bfU,0xe375U,0xd6e1U,0xbdc9U,0x6b99U,0xd732U,
		0xbe6fU,0x6cd5U,0xd9aaU,0xa35fU,0x56b5U,0xad6aU,0x4adfU,0x95beU,0x3b77U,0x76eeU,
		0xeddcU,0xcbb3U,0x876dU,0x1ed1U,0x3da2U,0x7b44U,0xf688U,0xfd1bU,0xea3dU,0xc471U,
		0x98e9U,0x21d9U,0x43b2U,0x8764U,0x1ec3U,0x3d86U,0x7b0cU,0xf618U,0xfc3bU,0xe87dU,
		0xc0f1U,0x91e9U,0x33d9U,0x67b2U,0xcf64U,0x8ec3U,0x0d8dU,0x1b1aU,0x3634U,0x6c68U,
		0xd8d0U,0xa1abU,0x535dU,0xa6baU,0x5d7fU,0xbafeU,0x65f7U,0xcbeeU,0x87d7U,0x1fa5U,
		0x3f4aU,0x7e94U,0xfd28U,0xea5bU,0xc4bdU,0x9971U,0x22e9U,0x45d2U,0x8ba4U,0x0743U,
		0x0e86U,0x1d0cU,0x3a18U,0x7430U,0xe860U,0xc0cbU,0x919dU,0x3331U,0x6662U,0xccc4U,
		0x8983U,0x030dU,0x061aU,0x0c34U,0x1868U,0x30d0U,0x61a0U,0xc340U,0x968bU,0x3d1dU,
		0x7a3aU,0xf474U,0xf8e3U,0xe1cdU,0xd391U,0xb729U,0x7e59U,0xfcb2U,0xe96fU,0xc2d5U,
		0x95a1U,0x3b49U,0x7692U,0xed24U,0xca43U,0x848dU,0x1911U,0x3222U,0x6444U,0xc888U,
		0x811bU,0x123dU,0x247aU,0x48f4U,0x91e8U,0x33dbU,0x67b6U,0xcf6cU,0x8ed3U,0x0dadU,
		0x1b5aU,0x36b4U,0x6d68U,0xdad0U,0xa5abU,0x5b5dU,0xb6baU,0x7d7fU,0xfafeU,0xe5f7U,
		0xdbe5U,0xa7c1U,0x5f89U,0xbf12U,0x6e2fU,0xdc5eU,0xa8b7U,0x4165U,0x82caU,0x159fU,
		0x2b3eU,0x567cU,0xacf8U,0x49fbU,0x93f6U,0x37e7U,0x6fceU,0xdf9cU,0xaf33U,0x4e6dU,
		0x9cdaU,0x29bfU,0x537eU,0xa6fcU,0x5df3U,0xbbe6U,0x67c7U,0xcf8eU,0x8f17U,0x0e25U,
		0x1c4aU,0x3894U,0x7128U,0xe250U,0xd4abU,0xb95dU,0x62b1U,0xc562U,0x9acfU,0x2595U,
		0x4b2aU,0x9654U,0x3ca3U,0x7946U,0xf28cU,0xf513U,0xfa2dU,0xe451U,0xd8a9U,0xa159U,
		0x52b9U,0xa572U,0x5aefU,0xb5deU,0x7bb7U,0xf76eU,0xfed7U,0xeda5U,0xcb41U,0x8689U,
		0x1d19U,0x3a32U,0x7464U,0xe8c8U,0xc19bU,0x933dU,0x3671U,0x6ce2U,0xd9c4U,0xa383U,
		0x570dU,0xae1aU,0x4c3fU,0x987eU,0x20f7U,0x41eeU,0x83dcU,0x17b3U,0x2f66U,0x5eccU,
		0xbd98U,0x6b3bU,0xd676U,0xbce7U,0x69c5U,0xd38aU,0xb71fU,0x7e35U,0xfc6aU,0xe8dfU,
		0xc1b5U,0x9361U,0x36c9U,0x6d92U,0xdb24U,0xa643U,0x5c8dU,0xb91aU,0x623fU,0xc47eU,
		0x98f7U,0x21e5U,0x43caU,0x8794U,0x1f23U,0x3e46U,0x7c8cU,0xf918U,0xe23bU,0xd47dU,
		0xb8f1U,0x61e9U,0xc3d2U,0x97afU,0x3f55U,0x7eaaU,0xfd54U,0xeaa3U,0xc54dU,0x9a91U,
		0x2529U,0x4a52U,0x94a4U,0x3943U,0x7286U,0xe50cU,0xda13U,0xa42dU,0x5851U,0xb0a2U,
		0x714fU,0xe29eU,0xd537U,0xba65U,0x64c1U,0xc982U,0x830fU,0x1615U,0x2c2aU,0x5854U,
		0xb0a8U,0x715bU,0xe2b6U,0xd567U,0xbac5U,0x6581U,0xcb02U,0x860fU,0x1c15U,0x382aU,
		0x7054U,0xe0a8U,0xd15bU,0xb2bdU,0x7571U,0xeae2U,0xc5cfU,0x9b95U,0x2721U,0x4e42U,
		0x9c84U,0x2903U,0x5206U,0xa40cU,0x5813U,0xb026U,0x7047U,0xe08eU,0xd117U,0xb225U,
		0x7441U,0xe882U,0xc10fU,0x9215U,0x3421U,0x6842U,0xd084U,0xb103U,0x720dU,0xe41aU,
		0xd83fU,0xa075U,0x50e1U,0xa1c2U,0x538fU,0xa71eU,0x5e37U,0xbc6eU,0x68d7U,0xd1aeU,
		0xb357U,0x76a5U,0xed4aU,0xca9fU,0x8535U,0x1a61U,0x34c2U,0x6984U,0xd308U,0xb61bU,
		0x7c3dU,0xf87aU,0xe0ffU,0xd1f5U,0xb3e1U,0x77c9U,0xef92U,0xcf2fU,0x8e55U,0x0ca1U,
		0x1942U,0x3284U,0x6508U,0xca10U,0x842bU,0x185dU,0x30baU,0x6174U,0xc2e8U,0x95dbU,
		0x3bbdU,0x777aU,0xeef4U,0xcde3U,0x8bcdU,0x0791U,0x0f22U,0x1e44U,0x3c88U,0x7910U,
		0xf220U,0xf44bU,0xf89dU,0xe131U,0xd269U,0xb4d9U,0x79b9U,0xf372U,0xf6efU,0xfdd5U,
		0xeba1U,0xc749U,0x9e99U,0x2d39U,0x5a72U,0xb4e4U,0x79c3U,0xf386U,0xf707U,0xfe05U,
		0xec01U,0xc809U,0x8019U,0x1039U,0x2072U,0x40e4U,0x81c8U,0x139bU,0x2736U,0x4e6cU,
		0x9cd8U,0x29bbU,0x5376U,0xa6ecU,0x5dd3U,0xbba6U,0x6747U,0xce8eU,0x8d17U,0x0a25U,
		0x144aU,0x2894U,0x5128U,0xa250U,0x54abU,0xa956U,0x42a7U,0x854eU,0x1a97U,0x352eU,
		0x6a5cU,0xd4b8U,0xb97bU,0x62fdU,0xc5faU,0x9bffU,0x27f5U,0x4feaU,0x9fd4U,0x2fa3U,
		0x5f46U,0xbe8cU,0x6d13U,0xda26U,0xa447U,0x5885U,0xb10aU,0x721fU,0xe43eU,0xd877U,
		0xa0e5U,0x51c1U,0xa382U,0x570fU,0xae1eU,0x4c37U,0x986eU,0x20d7U,0x41aeU,0x835cU,
		0x16b3U,0x2d66U,0x5accU,0xb598U,0x7b3bU,0xf676U,0xfce7U,0xe9c5U,0xc381U,0x9709U,
		0x3e19U,0x7c32U,0xf864U,0xe0c3U,0xd18dU,0xb311U,0x7629U,0xec52U,0xc8afU,0x8155U,
		0x12a1U,0x2542U,0x4a84U,0x9508U,0x3a1bU,0x7436U,0xe86cU,0xc0d3U,0x91adU,0x3351U,
		0x66a2U,0xcd44U,0x8a83U,0x050dU,0x0a1aU,0x1434U,0x2868U,0x50d0U,0xa1a0U,0x534bU,
		0xa696U,0x5d27U,0xba4eU,0x6497U,0xc92eU,0x8257U,0x14a5U,0x294aU,0x5294U,0xa528U,
		0x5a5bU,0xb4b6U,0x7967U,0xf2ceU,0xf597U,0xfb25U,0xe641U,0xdc89U,0xa919U,0x4239U,
		0x8472U,0x18efU,0x31deU,0x63bcU,0xc778U,0x9efbU,0x2dfdU,0x5bfaU,0xb7f4U,0x7fe3U,
		0xffc6U,0xef87U,0xcf05U,0x8e01U,0x0c09U,0x1812U,0x3024U,0x6048U,0xc090U,0x912bU,
		0x325dU,0x64baU,0xc974U,0x82e3U,0x15cdU,0x2b9aU,0x5734U,0xae68U,0x4cdbU,0x99b6U,
		0x2367U,0x46ceU,0x8d9cU,0x0b33U,0x1666U,0x2cccU,0x5998U,0xb330U,0x766bU,0xecd6U,
		0xc9a7U,0x8345U,0x1681U,0x2d02U,0x5a04U,0xb408U,0x781bU,0xf036U,0xf067U,0xf0c5U,
		0xf181U,0xf309U,0xf619U,0xfc39U,0xe879U,0xc0f9U,0x91f9U,0x33f9U,0x67f2U,0xcfe4U,
		0x8fc3U,0x0f8dU,0x1f1aU,0x3e34U,0x7c68U,0xf8d0U,0xe1abU,0xd35dU,0xb6b1U,0x7d69U,
		0xfad2U,0xe5afU,0xdb55U,0xa6a1U,0x5d49U,0xba92U,0x652fU,0xca5eU,0x84b7U,0x1965U,
		0x32caU,0x6594U,0xcb28U,0x865bU,0x1cbdU,0x397aU,0x72f4U,0xe5e8U,0xdbdbU,0xa7bdU,
		0x5f71U,0xbee2U,0x6dcfU,0xdb9eU,0xa737U,0x5e65U,0xbccaU,0x699fU,0xd33eU,0xb677U,
		0x7ce5U,0xf9caU,0xe39fU,0xd735U,0xbe61U,0x6cc9U,0xd992U,0xa32fU,0x5655U,0xacaaU,
		0x495fU,0x92beU,0x3577U,0x6aeeU,0xd5dcU,0xbbb3U,0x676dU,0xcedaU,0x8dbfU,0x0b75U,
		0x16eaU,0x2dd4U,0x5ba8U,0xb750U,0x7eabU,0xfd56U,0xeaa7U,0xc545U,0x9a81U,0x2509U,
		0x4a12U,0x9424U,0x3843U,0x7086U,0xe10cU,0xd213U,0xb42dU,0x7851U,0xf0a2U,0xf14fU,
		0xf295U,0xf521U,0xfa49U,0xe499U,0xd939U,0xa279U,0x54f9U,0xa9f2U,0x43efU,0x87deU,
		0x1fb7U,0x3f6eU,0x7edcU,0xfdb8U,0xeb7bU,0xc6fdU,0x9df1U,0x2be9U,0x57d2U,0xafa4U,
		0x4f43U,0x9e86U,0x2d07U,0x5a0eU,0xb41cU,0x7833U,0xf066U,0xf0c7U,0xf185U,0xf301U,
		0xf609U,0xfc19U,0xe839U,0xc079U,0x90f9U,0x31f9U,0x63f2U,0xc7e4U,0x9fc3U,0x2f8dU,
		0x5f1aU,0xbe34U,0x6c63U,0xd8c6U,0xa187U,0x5305U,0xa60aU,0x5c1fU,0xb83eU,0x6077U,
		0xc0eeU,0x91d7U,0x33a5U,0x674aU,0xce94U,0x8d23U,0x0a4dU,0x149aU,0x2934U,0x5268U,
		0xa4d0U,0x59abU,0xb356U,0x76a7U,0xed4eU,0xca97U,0x8525U,0x1a41U,0x3482U,0x6904U,
		0xd208U,0xb41bU,0x783dU,0xf07aU,0xf0ffU,0xf1f5U,0xf3e1U,0xf7c9U,0xff99U,0xef39U,
		0xce79U,0x8cf9U,0x09f9U,0x13f2U,0x27e4U,0x4fc8U,0x9f90U,0x2f2bU,0x5e56U,0xbcacU,
		0x6953U,0xd2a6U,0xb547U,0x7a85U,0xf50aU,0xfa1fU,0xe435U,0xd861U,0xa0c9U,0x5199U,
		0xa332U,0x566fU,0xacdeU,0x49b7U,0x936eU,0x36d7U,0x6daeU,0xdb5cU,0xa6b3U,0x5d6dU,
		0xbadaU,0x65bfU,0xcb7eU,0x86f7U,0x1de5U,0x3bcaU,0x7794U,0xef28U,0xce5bU,0x8cbdU,
		0x0971U,0x12e2U,0x25c4U,0x4b88U,0x9710U,0x3e2bU,0x7c56U,0xf8acU,0xe153U,0xd2adU,
		0xb551U,0x7aa9U,0xf552U,0xfaafU,0xe555U,0xdaa1U,0xa549U,0x5a99U,0xb532U,0x7a6fU,
		0xf4deU,0xf9b7U,0xe365U,0xd6c1U,0xbd89U,0x6b19U,0xd632U,0xbc6fU,0x68d5U,0xd1aaU,
		0xb35fU,0x76b5U,0xed6aU,0xcadfU,0x85b5U,0x1b61U,0x36c2U,0x6d84U,0xdb08U,0xa61bU,
		0x5c3dU,0xb87aU,0x60ffU,0xc1feU,0x93f7U,0x37e5U,0x6fcaU,0xdf94U,0xaf23U,0x4e4dU,
		0x9c9aU,0x293fU,0x527eU,0xa4fcU,0x59f3U,0xb3e6U,0x77c7U,0xef8eU,0xcf17U,0x8e25U,
		0x0c41U,0x1882U,0x3104U,0x6208U,0xc410U,0x982bU,0x205dU,0x40baU,0x8174U,0x12e3U,
		0x25c6U,0x4b8cU,0x9718U,0x3e3bU,0x7c76U,0xf8ecU,0xe1d3U,0xd3adU,0xb751U,0x7ea9U,
		0xfd52U,0xeaafU,0xc555U,0x9aa1U,0x2549U,0x4a92U,0x9524U,0x3a43U,0x7486U,0xe90cU,
		0xc213U,0x942dU,0x3851U,0x70a2U,0xe144U,0xd283U,0xb50dU,0x7a11U,0xf422U,0xf84fU,
		0xe095U,0xd121U,0xb249U,0x7499U,0xe932U,0xc26fU,0x94d5U,0x39a1U,0x7342U,0xe684U,
		0xdd03U,0xaa0dU,0x4411U,0x8822U,0x004fU,0x009eU,0x013cU,0x0278U,0x04f0U,0x09e0U,
		0x13c0U,0x2780U,0x4f00U,0x9e00U,0x2c0bU,0x5816U,0xb02cU,0x7053U,0xe0a6U,0xd147U,
		0xb285U,0x7501U,0xea02U,0xc40fU,0x9815U,0x2021U,0x4042U,0x8084U,0x1103U,0x2206U,
		0x440cU,0x8818U,0x003bU,0x0076U,0x00ecU,0x01d8U,0x03b0U,0x0760U,0x0ec0U,0x1d80U,
		0x3b00U,0x7600U,0xec00U,0xc80bU,0x801dU,0x1031U,0x2062U,0x40c4U,0x8188U,0x131bU,
		0x2636U,0x4c6cU,0x98d8U,0x21bbU,0x4376U,0x86ecU,0x1dd3U,0x3ba6U,0x774cU,0xee98U,
		0xcd3bU,0x8a7dU,0x04f1U,0x09e2U,0x13c4U,0x2788U,0x4f10U,0x9e20U,0x2c4bU,0x5896U,
		0xb12cU,0x7253U,0xe4a6U,0xd947U,0xa285U,0x5501U,0xaa02U,0x440fU,0x881eU,0x0037U,
		0x006eU,0x00dcU,0x01b8U,0x0370U,0x06e0U,0x0dc0U,0x1b80U,0x3700U,0x6e00U,0xdc00U,
		0xa80bU,0x401dU,0x803aU,0x107fU,0x20feU,0x41fcU,0x83f8U,0x17fbU,0x2ff6U,0x5fecU,
		0xbfd8U,0x6fbbU,0xdf76U,0xaee7U,0x4dc5U,0x9b8aU,0x271fU,0x4e3eU,0x9c7cU,0x28f3U,
		0x51e6U,0xa3ccU,0x5793U,0xaf26U,0x4e47U,0x9c8eU,0x2917U,0x522eU,0xa45cU,0x58b3U,
		0xb166U,0x72c7U,0xe58eU,0xdb17U,0xa625U,0x5c41U,0xb882U,0x610fU,0xc21eU,0x9437U,
		0x3865U,0x70caU,0xe194U,0xd323U,0xb64dU,0x7c91U,0xf922U,0xe24fU,0xd495U,0xb921U,
		0x6249U,0xc492U,0x992fU,0x2255U,0x44aaU,0x8954U,0x02a3U,0x0546U,0x0a8cU,0x1518U,
		0x2a30U,0x5460U,0xa8c0U,0x418bU,0x8316U,0x1627U,0x2c4eU,0x589cU,0xb138U,0x727bU,
		0xe4f6U,0xd9e7U,0xa3c5U,0x5781U,0xaf02U,0x4e0fU,0x9c1eU,0x2837U,0x506eU,0xa0dcU,
		0x51b3U,0xa366U,0x56c7U,0xad8eU,0x4b17U,0x962eU,0x3c57U,0x78aeU,0xf15cU,0xf2b3U,
		0xf56dU,0xfad1U,0xe5a9U,0xdb59U,0xa6b9U,0x5d79U,0xbaf2U,0x65efU,0xcbdeU,0x87b7U,
		0x1f65U,0x3ecaU,0x7d94U,0xfb28U,0xe65bU,0xdcbdU,0xa971U,0x42e9U,0x85d2U,0x1bafU,
		0x375eU,0x6ebcU,0xdd78U,0xaafbU,0x45fdU,0x8bfaU,0x07ffU,0x0ffeU,0x1ffcU,0x3ff8U,
		0x7ff0U,0xffe0U,0xefcbU,0xcf9dU,0x8f31U,0x0e69U,0x1cd2U,0x39a4U,0x7348U,0xe690U,
		0xdd2bU,0xaa5dU,0x44b1U,0x8962U,0x02cfU,0x059eU,0x0b3cU,0x1678U,0x2cf0U,0x59e0U,
		0xb3c0U,0x778bU,0xef16U,0xce27U,0x8c45U,0x0881U,0x1102U,0x2204U,0x4408U,0x8810U,
		0x002bU,0x0056U,0x00acU,0x0158U,0x02b0U,0x0560U,0x0ac0U,0x1580U,0x2b00U,0x5600U,
		0xac00U,0x480bU,0x9016U,0x3027U,0x604eU,0xc09cU,0x9133U,0x326dU,0x64daU,0xc9b4U,
		0x8363U,0x16cdU,0x2d9aU,0x5b34U,0xb668U,0x7cdbU,0xf9b6U,0xe367U,0xd6c5U,0xbd81U,
		0x6b09U,0xd612U,0xbc2fU,0x6855U,0xd0aaU,0xb15fU,0x72b5U,0xe56aU,0xdadfU,0xa5b5U,
		0x5b61U,0xb6c2U,0x7d8fU,0xfb1eU,0xe637U,0xdc65U,0xa8c1U,0x4189U,0x8312U,0x162fU,
		0x2c5eU,0x58bcU,0xb178U,0x72fbU,0xe5f6U,0xdbe7U,0xa7c5U,0x5f81U,0xbf02U,0x6e0fU,
		0xdc1eU,0xa837U,0x4065U,0x80caU,0x119fU,0x233eU,0x467cU,0x8cf8U,0x09fbU,0x13f6U,
		0x27ecU,0x4fd8U,0x9fb0U,0x2f6bU,0x5ed6U,0xbdacU,0x6b53U,0xd6a6U,0xbd47U,0x6a85U,
		0xd50aU,0xba1fU,0x6435U,0xc86aU,0x80dfU,0x11b5U,0x236aU,0x46d4U,0x8da8U,0x0b5bU,
		0x16b6U,0x2d6cU,0x5ad8U,0xb5b0U,0x7b6bU,0xf6d6U,0xfda7U,0xeb45U,0xc681U,0x9d09U,
		0x2a19U,0x5432U,0xa864U,0x40c3U,0x8186U,0x1307U,0x260eU,0x4c1cU,0x9838U,0x207bU,
		0x40f6U,0x81ecU,0x13d3U,0x27a6U,0x4f4cU,0x9e98U,0x2d3bU,0x5a76U,0xb4ecU,0x79d3U,
		0xf3a6U,0xf747U,0xfe85U,0xed01U,0xca09U,0x8419U,0x1839U,0x3072U,0x60e4U,0xc1c8U,
		0x939bU,0x373dU,0x6e7aU,0xdcf4U,0xa9e3U,0x43cdU,0x879aU,0x1f3fU,0x3e7eU,0x7cfcU,
		0xf9f8U,0xe3fbU,0xd7fdU,0xbff1U,0x6fe9U,0xdfd2U,0xafafU,0x4f55U,0x9eaaU,0x2d5fU,
		0x5abeU,0xb57cU,0x7af3U,0xf5e6U,0xfbc7U,0xe785U,0xdf01U,0xae09U,0x4c19U,0x9832U,
		0x206fU,0x40deU,0x81bcU,0x1373U,0x26e6U,0x4dccU,0x9b98U,0x273bU,0x4e76U,0x9cecU,
		0x29d3U,0x53a6U,0xa74cU,0x5e93U,0xbd26U,0x6a47U,0xd48eU,0xb917U,0x6225U,0xc44aU,
		0x989fU,0x2135U,0x426aU,0x84d4U,0x19a3U,0x3346U,0x668cU,0xcd18U,0x8a3bU,0x047dU,
		0x08faU,0x11f4U,0x23e8U,0x47d0U,0x8fa0U,0x0f4bU,0x1e96U,0x3d2cU,0x7a58U,0xf4b0U,
		0xf96bU,0xe2ddU,0xd5b1U,0xbb69U,0x66d9U,0xcdb2U,0x8b6fU,0x06d5U,0x0daaU,0x1b54U,
		0x36a8U,0x6d50U,0xdaa0U,0xa54bU,0x5a9dU,0xb53aU,0x7a7fU,0xf4feU,0xf9f7U,0xe3e5U,
		0xd7c1U,0xbf89U,0x6f19U,0xde32U,0xac6fU,0x48d5U,0x91aaU,0x335fU,0x66beU,0xcd7cU,
		0x8af3U,0x05edU,0x0bdaU,0x17b4U,0x2f68U,0x5ed0U,0xbda0U,0x6b4bU,0xd696U,0xbd27U,
		0x6a45U,0xd48aU,0xb91fU,0x6235U,0xc46aU,0x98dfU,0x21b5U,0x436aU,0x86d4U,0x1da3U,
		0x3b46U,0x768cU,0xed18U,0xca3bU,0x847dU,0x18f1U,0x31e2U,0x63c4U,0xc788U,0x9f1bU,
		0x2e3dU,0x5c7aU,0xb8f4U,0x61e3U,0xc3c6U,0x9787U,0x3f05U,0x7e0aU,0xfc14U,0xe823U,
		0xc04dU,0x9091U,0x3129U,0x6252U,0xc4a4U,0x9943U,0x228dU,0x451aU,0x8a34U,0x0463U,
		0x08c6U,0x118cU,0x2318U,0x4630U,0x8c60U,0x08cbU,0x1196U,0x232cU,0x4658U,0x8cb0U,
		0x096bU,0x12d6U,0x25acU,0x4b58U,0x96b0U,0x3d6bU,0x7ad6U,0xf5acU,0xfb53U,0xe6adU,
		0xdd51U,0xaaa9U,0x4559U,0x8ab2U,0x056fU,0x0adeU,0x15bcU,0x2b78U,0x56f0U,0xade0U,
		0x4bcbU,0x9796U,0x3f27U,0x7e4eU,0xfc9cU,0xe933U,0xc26dU,0x94d1U,0x39a9U,0x7352U,
		0xe6a4U,0xdd43U,0xaa8dU,0x4511U,0x8a22U,0x044fU,0x089eU,0x113cU,0x2278U,0x44f0U,
		0x89e0U,0x03cbU,0x0796U,0x0f2cU,0x1e58U,0x3cb0U,0x7960U,0xf2c0U,0xf58bU,0xfb1dU,
		0xe631U,0xdc69U,0xa8d9U,0x41b9U,0x8372U,0x16efU,0x2ddeU,0x5bbcU,0xb778U,0x7efbU,
		0xfdf6U,0xebe7U,0xc7c5U,0x9f81U,0x2f09U,0x5e12U,0xbc24U,0x6843U,0xd086U,0xb107U,
		0x7205U,0xe40aU,0xd81fU,0xa035U,0x5061U,0xa0c2U,0x518fU,0xa31eU,0x5637U,0xac6eU,
		0x48d7U,0x91aeU,0x3357U,0x66aeU,0xcd5cU,0x8ab3U,0x056dU,0x0adaU,0x15b4U,0x2b68U,
		0x56d0U,0xada0U,0x4b4bU,0x9696U,0x3d27U,0x7a4eU,0xf49cU,0xf933U,0xe26dU,0xd4d1U,
		0xb9a9U,0x6359U,0xc6b2U,0x9d6fU,0x2ad5U,0x55aaU,0xab54U,0x46a3U,0x8d46U,0x0a87U,
		0x150eU,0x2a1cU,0x5438U,0xa870U,0x40ebU,0x81d6U,0x13a7U,0x274eU,0x4e9cU,0x9d38U,
		0x2a7bU,0x54f6U,0xa9ecU,0x43d3U,0x87a6U,0x1f47U,0x3e8eU,0x7d1cU,0xfa38U,0xe47bU,
		0xd8fdU,0xa1f1U,0x53e9U,0xa7d2U,0x5fafU,0xbf5eU,0x6eb7U,0xdd6eU,0xaad7U,0x45a5U,
		0x8b4aU,0x069fU,0x0d3eU,0x1a7cU,0x34f8U,0x69f0U,0xd3e0U,0xb7cbU,0x7f9dU,0xff3aU,
		0xee7fU,0xccf5U,0x89e1U,0x03c9U,0x0792U,0x0f24U,0x1e48U,0x3c90U,0x7920U,0xf240U,
		0xf48bU,0xf91dU,0xe231U,0xd469U,0xb8d9U,0x61b9U,0xc372U,0x96efU,0x3dd5U,0x7baaU,
		0xf754U,0xfea3U,0xed4dU,0xca91U,0x8529U,0x1a59U,0x34b2U,0x6964U,0xd2c8U,0xb59bU,
		0x7b3dU,0xf67aU,0xfcffU,0xe9f5U,0xc3e1U,0x97c9U,0x3f99U,0x7f32U,0xfe64U,0xecc3U,
		0xc98dU,0x8311U,0x1629U,0x2c52U,0x58a4U,0xb148U,0x729bU,0xe536U,0xda67U,0xa4c5U,
		0x5981U,0xb302U,0x760fU,0xec1eU,0xc837U,0x8065U,0x10c1U,0x2182U,0x4304U,0x8608U,
		0x1c1bU,0x3836U,0x706cU,0xe0d8U,0xd1bbU,0xb37dU,0x76f1U,0xede2U,0xcbcfU,0x8795U,
		0x1f21U,0x3e42U,0x7c84U,0xf908U,0xe21bU,0xd43dU,0xb871U,0x60e9U,0xc1d2U,0x93afU,
		0x3755U,0x6eaaU,0xdd54U,0xaaa3U,0x454dU,0x8a9aU,0x053fU,0x0a7eU,0x14fcU,0x29f8U,
		0x53f0U,0xa7e0U,0x5fcbU,0xbf96U,0x6f27U,0xde4eU,0xac97U,0x4925U,0x924aU,0x349fU,
		0x693eU,0xd27cU,0xb4f3U,0x79edU,0xf3daU,0xf7bfU,0xff75U,0xeee1U,0xcdc9U,0x8b99U,
		0x0739U,0x0e72U,0x1ce4U,0x39c8U,0x7390U,0xe720U,0xde4bU,0xac9dU,0x4931U,0x9262U,
		0x34cfU,0x699eU,0xd33cU,0xb673U,0x7cedU,0xf9daU,0xe3bfU,0xd775U,0xbee1U,0x6dc9U,
		0xdb92U,0xa72fU,0x5e55U,0xbcaaU,0x695fU,0xd2beU,0xb577U,0x7ae5U,0xf5caU,0xfb9fU,
		0xe735U,0xde61U,0xacc9U,0x4999U,0x9332U,0x366fU,0x6cdeU,0xd9bcU,0xa373U,0x56edU,
		0xaddaU,0x4bbfU,0x977eU,0x3ef7U,0x7deeU,0xfbdcU,0xe7b3U,0xdf6dU,0xaed1U,0x4da9U,
		0x9b52U,0x26afU,0x4d5eU,0x9abcU,0x2573U,0x4ae6U,0x95ccU,0x3b93U,0x7726U,0xee4cU,
		0xcc93U,0x892dU,0x0251U,0x04a2U,0x0944U,0x1288U,0x2510U,0x4a20U,0x9440U,0x388bU,
		0x7116U,0xe22cU,0xd453U,0xb8adU,0x6151U,0xc2a2U,0x954fU,0x3a95U,0x752aU,0xea54U,
		0xc4a3U,0x994dU,0x2291U,0x4522U,0x8a44U,0x0483U,0x0906U,0x120cU,0x2418U,0x4830U,
		0x9060U,0x30cbU,0x6196U,0xc32cU,0x9653U,0x3cadU,0x795aU,0xf2b4U,0xf563U,0xfacdU,
		0xe591U,0xdb29U,0xa659U,0x5cb9U,0xb972U,0x62efU,0xc5deU,0x9bb7U,0x2765U,0x4ecaU,
		0x9d94U,0x2b23U,0x5646U,0xac8cU,0x4913U,0x9226U,0x3447U,0x688eU,0xd11cU,0xb233U,
		0x746dU,0xe8daU,0xc1bfU,0x9375U,0x36e1U,0x6dc2U,0xdb84U,0xa703U,0x5e0dU,0xbc1aU,
		0x683fU,0xd07eU,0xb0f7U,0x71e5U,0xe3caU,0xd79fU,0xbf35U,0x6e61U,0xdcc2U,0xa98fU,
		0x4315U,0x862aU,0x1c5fU,0x38beU,0x717cU,0xe2f8U,0xd5fbU,0xbbfdU,0x67f1U,0xcfe2U,
		0x8fcfU,0x0f95U,0x1f2aU,0x3e54U,0x7ca8U,0xf950U,0xe2abU,0xd55dU,0xbab1U,0x6569U,
		0xcad2U,0x85afU,0x1b55U,0x36aaU,0x6d54U,0xdaa8U,0xa55bU,0x5abdU,0xb57aU,0x7affU,
		0xf5feU,0xfbf7U,0xe7e5U,0xdfc1U,0xaf89U,0x4f19U,0x9e32U,0x2c6fU,0x58deU,0xb1bcU,
		0x7373U,0xe6e6U,0xddc7U,0xab85U,0x4701U,0x8e02U,0x0c0fU,0x181eU,0x303cU,0x6078U,
		0xc0f0U,0x91ebU,0x33ddU,0x67baU,0xcf74U,0x8ee3U,0x0dcdU,0x1b9aU,0x3734U,0x6e68U,
		0xdcd0U,0xa9abU,0x435dU,0x86baU,0x1d7fU,0x3afeU,0x75fcU,0xebf8U,0xc7fbU,0x9ffdU,
		0x2ff1U,0x5fe2U,0xbfc4U,0x6f83U,0xdf06U,0xae07U,0x4c05U,0x980aU,0x201fU,0x403eU,
		0x807cU,0x10f3U,0x21e6U,0x43ccU,0x8798U,0x1f3bU,0x3e76U,0x7cecU,0xf9d8U,0xe3bbU,
		0xd77dU,0xbef1U,0x6de9U,0xdbd2U,0xa7afU,0x5f55U,0xbeaaU,0x6d5fU,0xdabeU,0xa577U,
		0x5ae5U,0xb5caU,0x7b9fU,0xf73eU,0xfe77U,0xece5U,0xc9c1U,0x8389U,0x1719U,0x2e32U,
		0x5c64U,0xb8c8U,0x619bU,0xc336U,0x9667U,0x3cc5U,0x798aU,0xf314U,0xf623U,0xfc4dU,
		0xe891U,0xc129U,0x9259U,0x34b9U,0x6972U,0xd2e4U,0xb5c3U,0x7b8dU,0xf71aU,0xfe3fU,
		0xec75U,0xc8e1U,0x81c9U,0x1399U,0x2732U,0x4e64U,0x9cc8U,0x299bU,0x5336U,0xa66cU,
		0x5cd3U,0xb9a6U,0x6347U,0xc68eU,0x9d17U,0x2a25U,0x544aU,0xa894U,0x4123U,0x8246U,
		0x1487U,0x290eU,0x521cU,0xa438U,0x587bU,0xb0f6U,0x71e7U,0xe3ceU,0xd797U,0xbf25U,
		0x6e41U,0xdc82U,0xa90fU,0x4215U,0x842aU,0x185fU,0x30beU,0x617cU,0xc2f8U,0x95fbU,
		0x3bfdU,0x77faU,0xeff4U,0xcfe3U,0x8fcdU,0x0f91U,0x1f22U,0x3e44U,0x7c88U,0xf910U,
		0xe22bU,0xd45dU,0xb8b1U,0x6169U,0xc2d2U,0x95afU,0x3b55U,0x76aaU,0xed54U,0xcaa3U,
		0x854dU,0x1a91U,0x3522U,0x6a44U,0xd488U,0xb91bU,0x623dU,0xc47aU,0x98ffU,0x21f5U,
		0x43eaU,0x87d4U,0x1fa3U,0x3f46U,0x7e8cU,0xfd18U,0xea3bU,0xc47dU,0x98f1U,0x21e9U,
		0x43d2U,0x87a4U,0x1f43U,0x3e86U,0x7d0cU,0xfa18U,0xe43bU,0xd87dU,0xa0f1U,0x51e9U,
		0xa3d2U,0x57afU,0xaf5eU,0x4eb7U,0x9d6eU,0x2ad7U,0x55aeU,0xab5cU,0x46b3U,0x8d66U,
		0x0ac7U,0x158eU,0x2b1cU,0x5638U,0xac70U,0x48ebU,0x91d6U,0x33a7U,0x674eU,0xce9cU,
		0x8d33U,0x0a6dU,0x14daU,0x29b4U,0x5368U,0xa6d0U,0x5dabU,0xbb56U,0x66a7U,0xcd4eU,
		0x8a97U,0x0525U,0x0a4aU,0x1494U,0x2928U,0x5250U,0xa4a0U,0x594bU,0xb296U,0x7527U,
		0xea4eU,0xc497U,0x9925U,0x2241U,0x4482U,0x8904U,0x0203U,0x0406U,0x080cU,0x1018U,
		0x2030U,0x4060U,0x80c0U,0x118bU,0x2316U,0x462cU,0x8c58U,0x08bbU,0x1176U,0x22ecU,
		0x45d8U,0x8bb0U,0x076bU,0x0ed6U,0x1dacU,0x3b58U,0x76b0U,0xed60U,0xcacbU,0x859dU,
		0x1b31U,0x3662U,0x6cc4U,0xd988U,0xa31bU,0x563dU,0xac7aU,0x48ffU,0x91feU,0x33f7U,
		0x67eeU,0xcfdcU,0x8fb3U,0x0f6dU,0x1edaU,0x3db4U,0x7b68U,0xf6d0U,0xfdabU,0xeb5dU,
		0xc6b1U,0x9d69U,0x2ad9U,0x55b2U,0xab64U,0x46c3U,0x8d86U,0x0b07U,0x160eU,0x2c1cU,
		0x5838U,0xb070U,0x70ebU,0xe1d6U,0xd3a7U,0xb745U,0x7e81U,0xfd02U,0xea0fU,0xc415U,
		0x9821U,0x2049U,0x4092U,0x8124U,0x1243U,0x2486U,0x490cU,0x9218U,0x343bU,0x6876U,
		0xd0ecU,0xb1d3U,0x73adU,0xe75aU,0xdebfU,0xad75U,0x4ae1U,0x95c2U,0x3b8fU,0x771eU,
		0xee3cU,0xcc73U,0x88edU,0x01d1U,0x03a2U,0x0744U,0x0e88U,0x1d10U,0x3a20U,0x7440U,
		0xe880U,0xc10bU,0x921dU,0x3431U,0x6862U,0xd0c4U,0xb183U,0x730dU,0xe61aU,0xdc3fU,
		0xa875U,0x40e1U,0x81c2U,0x138fU,0x271eU,0x4e3cU,0x9c78U,0x28fbU,0x51f6U,0xa3ecU,
		0x57d3U,0xafa6U,0x4f47U,0x9e8eU,0x2d17U,0x5a2eU,0xb45cU,0x78b3U,0xf166U,0xf2c7U,
		0xf585U,0xfb01U,0xe609U,0xdc19U,0xa839U,0x4079U,0x80f2U,0x11efU,0x23deU,0x47bcU,
		0x8f78U,0x0efbU,0x1df6U,0x3becU,0x77d8U,0xefb0U,0xcf6bU,0x8eddU,0x0db1U,0x1b62U,
		0x36c4U,0x6d88U,0xdb10U,0xa62bU,0x5c5dU,0xb8baU,0x617fU,0xc2feU,0x95f7U,0x3be5U,
		0x77caU,0xef94U,0xcf23U,0x8e4dU,0x0c91U,0x1922U,0x3244U,0x6488U,0xc910U,0x822bU,
		0x145dU,0x28baU,0x5174U,0xa2e8U,0x55dbU,0xabb6U,0x4767U,0x8eceU,0x0d97U,0x1b2eU,
		0x365cU,0x6cb8U,0xd970U,0xa2ebU,0x55ddU,0xabbaU,0x477fU,0x8efeU,0x0df7U,0x1beeU,
		0x37dcU,0x6fb8U,0xdf70U,0xaeebU,0x4dddU,0x9bbaU,0x277fU,0x4efeU,0x9dfcU,0x2bf3U,
		0x57e6U,0xafccU,0x4f93U,0x9f26U,0x2e47U,0x5c8eU,0xb91cU,0x6233U,0xc466U,0x98c7U,
		0x2185U,0x430aU,0x8614U,0x1c23U,0x3846U,0x708cU,0xe118U,0xd23bU,0xb47dU,0x78f1U,
		0xf1e2U,0xf3cfU,0xf795U,0xff21U,0xee49U,0xcc99U,0x8939U,0x0279U,0x04f2U,0x09e4U,
		0x13c8U,0x2790U,0x4f20U,0x9e40U,0x2c8bU,0x5916U,0xb22cU,0x7453U,0xe8a6U,0xc147U,
		0x9285U,0x3501U,0x6a02U,0xd404U,0xb803U,0x600dU,0xc01aU,0x903fU,0x3075U,0x60eaU,
		0xc1d4U,0x93a3U,0x374dU,0x6e9aU,0xdd34U,0xaa63U,0x44cdU,0x899aU,0x033fU,0x067eU,
		0x0cfcU,0x19f8U,0x33f0U,0x67e0U,0xcfc0U,0x8f8bU,0x0f1dU,0x1e3aU,0x3c74U,0x78e8U,
		0xf1d0U,0xf3abU,0xf75dU,0xfeb1U,0xed69U,0xcad9U,0x85b9U,0x1b79U,0x36f2U,0x6de4U,
		0xdbc8U,0xa79bU,0x5f3dU,0xbe7aU,0x6cffU,0xd9feU,0xa3f7U,0x57e5U,0xafcaU,0x4f9fU,
		0x9f3eU,0x2e77U,0x5ceeU,0xb9dcU,0x63b3U,0xc766U,0x9ec7U,0x2d85U,0x5b0aU,0xb614U,
		0x7c23U,0xf846U,0xe087U,0xd105U,0xb201U,0x7409U,0xe812U,0xc02fU,0x9055U,0x30a1U,
		0x6142U,0xc284U,0x9503U,0x3a0dU,0x741aU,0xe834U,0xc063U,0x90cdU,0x3191U,0x6322U,
		0xc644U,0x9c83U,0x290dU,0x521aU,0xa434U,0x5863U,0xb0c6U,0x7187U,0xe30eU,0xd617U,
		0xbc25U,0x6841U,0xd082U,0xb10fU,0x7215U,0xe42aU,0xd85fU,0xa0b5U,0x5161U,0xa2c2U,
		0x558fU,0xab1eU,0x4637U,0x8c6eU,0x08d7U,0x11aeU,0x235cU,0x46b8U,0x8d70U,0x0aebU,
		0x15d6U,0x2bacU,0x5758U,0xaeb0U,0x4d6bU,0x9ad6U,0x25a7U,0x4b4eU,0x969cU,0x3d33U,
		0x7a66U,0xf4ccU,0xf993U,0xe32dU,0xd651U,0xbca9U,0x6959U,0xd2b2U,0xb56fU,0x7ad5U,
		0xf5aaU,0xfb5fU,0xe6b5U,0xdd61U,0xaac9U,0x4599U,0x8b32U,0x066fU,0x0cdeU,0x19bcU,
		0x3378U,0x66f0U,0xcde0U,0x8bcbU,0x079dU,0x0f3aU,0x1e74U,0x3ce8U,0x79d0U,0xf3a0U,
		0xf74bU,0xfe9dU,0xed31U,0xca69U,0x84d9U,0x19b9U,0x3372U,0x66e4U,0xcdc8U,0x8b9bU,
		0x073dU,0x0e7aU,0x1cf4U,0x39e8U,0x73d0U,0xe7a0U,0xdf4bU,0xae9dU,0x4d31U,0x9a62U,
		0x24cfU,0x499eU,0x933cU,0x3673U,0x6ce6U,0xd9ccU,0xa393U,0x572dU,0xae5aU,0x4cbfU,
		0x997eU,0x22f7U,0x45eeU,0x8bdcU,0x07b3U,0x0f66U,0x1eccU,0x3d98U,0x7b30U,0xf660U,
		0xfccbU,0xe99dU,0xc331U,0x9669U,0x3cd9U,0x79b2U,0xf364U,0xf6c3U,0xfd8dU,0xeb11U,
		0xc629U,0x9c59U,0x28b9U,0x5172U,0xa2e4U,0x55c3U,0xab86U,0x4707U,0x8e0eU,0x0c17U,
		0x182eU,0x305cU,0x60b8U,0xc170U,0x92ebU,0x35ddU,0x6bbaU,0xd774U,0xbee3U,0x6dcdU,
		0xdb9aU,0xa73fU,0x5e75U,0xbceaU,0x69dfU,0xd3beU,0xb777U,0x7ee5U,0xfdcaU,0xeb9fU,
		0xc735U,0x9e61U,0x2cc9U,0x5992U,0xb324U,0x7643U,0xec86U,0xc907U,0x8205U,0x1401U,
		0x2802U,0x5004U,0xa008U,0x501bU,0xa036U,0x5067U,0xa0ceU,0x5197U,0xa32eU,0x5657U,
		0xacaeU,0x4957U,0x92aeU,0x3557U,0x6aaeU,0xd55cU,0xbab3U,0x656dU,0xcadaU,0x85bfU,
		0x1b75U,0x36eaU,0x6dd4U,0xdba8U,0xa75bU,0x5ebdU,0xbd7aU,0x6affU,0xd5feU,0xbbf7U,
		0x67e5U,0xcfcaU,0x8f9fU,0x0f35U,0x1e6aU,0x3cd4U,0x79a8U,0xf350U,0xf6abU,0xfd5dU,
		0xeab1U,0xc569U,0x9ad9U,0x25b9U,0x4b72U,0x96e4U,0x3dc3U,0x7b86U,0xf70cU,0xfe13U,
		0xec2dU,0xc851U,0x80a9U,0x1159U,0x22b2U,0x4564U,0x8ac8U,0x059bU,0x0b36U,0x166cU,
		0x2cd8U,0x59b0U,0xb360U,0x76cbU,0xed96U,0xcb27U,0x8645U,0x1c81U,0x3902U,0x7204U,
		0xe408U,0xd81bU,0xa03dU,0x5071U,0xa0e2U,0x51cfU,0xa39eU,0x5737U,0xae6eU,0x4cd7U,
		0x99aeU,0x2357U,0x46aeU,0x8d5cU,0x0ab3U,0x1566U,0x2accU,0x5598U,0xab30U,0x466bU,
		0x8cd6U,0x09a7U,0x134eU,0x269cU,0x4d38U,0x9a70U,0x24ebU,0x49d6U,0x93acU,0x3753U,
		0x6ea6U,0xdd4cU,0xaa93U,0x452dU,0x8a5aU,0x04bfU,0x097eU,0x12fcU,0x25f8U,0x4bf0U,
		0x97e0U,0x3fcbU,0x7f96U,0xff2cU,0xee53U,0xccadU,0x8951U,0x02a9U,0x0552U,0x0aa4U,
		0x1548U,0x2a90U,0x5520U,0xaa40U,0x448bU,0x8916U,0x0227U,0x044eU,0x089cU,0x1138U,
		0x2270U,0x44e0U,0x89c0U,0x038bU,0x0716U,0x0e2cU,0x1c58U,0x38b0U,0x7160U,0xe2c0U,
		0xd58bU,0xbb1dU,0x6631U,0xcc62U,0x88cfU,0x0195U,0x032aU,0x0654U,0x0ca8U,0x1950U,
		0x32a0U,0x6540U,0xca80U,0x850bU,0x1a1dU,0x343aU,0x6874U,0xd0e8U,0xb1dbU,0x73bdU,
		0xe77aU,0xdeffU,0xadf5U,0x4be1U,0x97c2U,0x3f8fU,0x7f1eU,0xfe3cU,0xec73U,0xc8edU,
		0x81d1U,0x13a9U,0x2752U,0x4ea4U,0x9d48U,0x2a9bU,0x5536U,0xaa6cU,0x44d3U,0x89a6U,
		0x0347U,0x068eU,0x0d1cU,0x1a38U,0x3470U,0x68e0U,0xd1c0U,0xb38bU,0x771dU,0xee3aU,
		0xcc7fU,0x88f5U,0x01e1U,0x03c2U,0x0784U,0x0f08U,0x1e10U,0x3c20U,0x7840U,0xf080U,
		0xf10bU,0xf21dU,0xf431U,0xf869U,0xe0d9U,0xd1b9U,0xb379U,0x76f9U,0xedf2U,0xcbefU,
		0x87d5U,0x1fa1U,0x3f42U,0x7e84U,0xfd08U,0xea1bU,0xc43dU,0x9871U,0x20e9U,0x41d2U,
		0x83a4U,0x1743U,0x2e86U,0x5d0cU,0xba18U,0x643bU,0xc876U,0x80e7U,0x11c5U,0x238aU,
		0x4714U,0x8e28U,0x0c5bU,0x18b6U,0x316cU,0x62d8U,0xc5b0U,0x9b6bU,0x26ddU,0x4dbaU,
		0x9b74U,0x26e3U,0x4dc6U,0x9b8cU,0x2713U,0x4e26U,0x9c4cU,0x2893U,0x5126U,0xa24cU,
		0x5493U,0xa926U,0x4247U,0x848eU,0x1917U,0x322eU,0x645cU,0xc8b8U,0x817bU,0x12fdU,
		0x25faU,0x4bf4U,0x97e8U,0x3fdbU,0x7fb6U,0xff6cU,0xeed3U,0xcdadU,0x8b51U,0x06a9U,
		0x0d52U,0x1aa4U,0x3548U,0x6a90U,0xd520U,0xba4bU,0x649dU,0xc93aU,0x827fU,0x14f5U,
		0x29eaU,0x53d4U,0xa7a8U,0x5f5bU,0xbeb6U,0x6d67U,0xdaceU,0xa597U,0x5b25U,0xb64aU,
		0x7c9fU,0xf93eU,0xe277U,0xd4e5U,0xb9c1U,0x6389U,0xc712U,0x9e2fU,0x2c55U,0x58aaU,
		0xb154U,0x72a3U,0xe546U,0xda87U,0xa505U,0x5a01U,0xb402U,0x780fU,0xf01eU,0xf037U,
		0xf065U,0xf0c1U,0xf189U,0xf319U,0xf639U,0xfc79U,0xe8f9U,0xc1f9U,0x93f9U,0x37f9U,
		0x6ff2U,0xdfe4U,0xafc3U,0x4f8dU,0x9f1aU,0x2e3fU,0x5c7eU,0xb8fcU,0x61f3U,0xc3e6U,
		0x97c7U,0x3f85U,0x7f0aU,0xfe14U,0xec23U,0xc84dU,0x8091U,0x1129U,0x2252U,0x44a4U,
		0x8948U,0x029bU,0x0536U,0x0a6cU,0x14d8U,0x29b0U,0x5360U,0xa6c0U,0x5d8bU,0xbb16U,
		0x6627U,0xcc4eU,0x8897U,0x0125U,0x024aU,0x0494U,0x0928U,0x1250U,0x24a0U,0x4940U,
		0x9280U,0x350bU,0x6a16U,0xd42cU,0xb853U,0x60adU,0xc15aU,0x92bfU,0x3575U,0x6aeaU,
		0xd5d4U,0xbba3U,0x674dU,0xce9aU,0x8d3fU,0x0a75U,0x14eaU,0x29d4U,0x53a8U,0xa750U,
		0x5eabU,0xbd56U,0x6aa7U,0xd54eU,0xba97U,0x6525U,0xca4aU,0x849fU,0x1935U,0x326aU,
		0x64d4U,0xc9a8U,0x835bU,0x16bdU,0x2d7aU,0x5af4U,0xb5e8U,0x7bdbU,0xf7b6U,0xff67U,
		0xeec5U,0xcd81U,0x8b09U,0x0619U,0x0c32U,0x1864U,0x30c8U,0x6190U,0xc320U,0x964bU,
		0x3c9dU,0x793aU,0xf274U,0xf4e3U,0xf9cdU,0xe391U,0xd729U,0xbe59U,0x6cb9U,0xd972U,
		0xa2efU,0x55d5U,0xabaaU,0x475fU,0x8ebeU,0x0d77U,0x1aeeU,0x35dcU,0x6bb8U,0xd770U,
		0xbeebU,0x6dddU,0xdbbaU,0xa77fU,0x5ef5U,0xbdeaU,0x6bdfU,0xd7beU,0xbf77U,0x6ee5U,
		0xddcaU,0xab9fU,0x4735U,0x8e6aU,0x0cdfU,0x19beU,0x337cU,0x66f8U,0xcdf0U,0x8bebU,
		0x07ddU,0x0fbaU,0x1f74U,0x3ee8U,0x7dd0U,0xfba0U,0xe74bU,0xde9dU,0xad31U,0x4a69U,
		0x94d2U,0x39afU,0x735eU,0xe6bcU,0xdd73U,0xaaedU,0x45d1U,0x8ba2U,0x074fU,0x0e9eU,
		0x1d3cU,0x3a78U,0x74f0U,0xe9e0U,0xc3cbU,0x979dU,0x3f31U,0x7e62U,0xfcc4U,0xe983U,
		0xc30dU,0x9611U,0x3c29U,0x7852U,0xf0a4U,0xf143U,0xf28dU,0xf511U,0xfa29U,0xe459U,
		0xd8b9U,0xa179U,0x52f9U,0xa5f2U,0x5befU,0xb7deU,0x7fb7U,0xff6eU,0xeed7U,0xcda5U,
		0x8b41U,0x0689U,0x0d12U,0x1a24U,0x3448U,0x6890U,0xd120U,0xb24bU,0x749dU,0xe93aU,
		0xc27fU,0x94f5U,0x39e1U,0x73c2U,0xe784U,0xdf03U,0xae0dU,0x4c11U,0x9822U,0x204fU,
		0x409eU,0x813cU,0x1273U,0x24e6U,0x49ccU,0x9398U,0x373bU,0x6e76U,0xdcecU,0xa9d3U,
		0x43adU,0x875aU,0x1ebfU,0x3d7eU,0x7afcU,0xf5f8U,0xfbfbU,0xe7fdU,0xdff1U,0xafe9U,
		0x4fd9U,0x9fb2U,0x2f6fU,0x5edeU,0xbdbcU,0x6b73U,0xd6e6U,0xbdc7U,0x6b85U,0xd70aU,
		0xbe1fU,0x6c35U,0xd86aU,0xa0dfU,0x51b5U,0xa36aU,0x56dfU,0xadbeU,0x4b77U,0x96eeU,
		0x3dd7U,0x7baeU,0xf75cU,0xfeb3U,0xed6dU,0xcad1U,0x85a9U,0x1b59U,0x36b2U,0x6d64U,
		0xdac8U,0xa59bU,0x5b3dU,0xb67aU,0x7cffU,0xf9feU,0xe3f7U,0xd7e5U,0xbfc1U,0x6f89U,
		0xdf12U,0xae2fU,0x4c55U,0x98aaU,0x215fU,0x42beU,0x857cU,0x1af3U,0x35e6U,0x6bccU,
		0xd798U,0xbf3bU,0x6e7dU,0xdcfaU,0xa9ffU,0x43f5U,0x87eaU,0x1fdfU,0x3fbeU,0x7f7cU,
		0xfef8U,0xedfbU,0xcbfdU,0x87f1U,0x1fe9U,0x3fd2U,0x7fa4U,0xff48U,0xee9bU,0xcd3dU,
		0x8a71U,0x04e9U,0x09d2U,0x13a4U,0x2748U,0x4e90U,0x9d20U,0x2a4bU,0x5496U,0xa92cU,
		0x4253U,0x84a6U,0x1947U,0x328eU,0x651cU,0xca38U,0x847bU,0x18fdU,0x31faU,0x63f4U,
		0xc7e8U,0x9fdbU,0x2fbdU,0x5f7aU,0xbef4U,0x6de3U,0xdbc6U,0xa787U,0x5f05U,0xbe0aU,
		0x6c1fU,0xd83eU,0xa077U,0x50e5U,0xa1caU,0x539fU,0xa73eU,0x5e77U,0xbceeU,0x69d7U,
		0xd3aeU,0xb757U,0x7ea5U,0xfd4aU,0xea9fU,0xc535U,0x9a61U,0x24c9U,0x4992U,0x9324U,
		0x3643U,0x6c86U,0xd90cU,0xa213U,0x542dU,0xa85aU,0x40bfU,0x817eU,0x12f7U,0x25eeU,
		0x4bdcU,0x97b8U,0x3f7bU,0x7ef6U,0xfdecU,0xebd3U,0xc7adU,0x9f51U,0x2ea9U,0x5d52U,
		0xbaa4U,0x6543U,0xca86U,0x8507U,0x1a05U,0x340aU,0x6814U,0xd028U,0xb05bU,0x70bdU,
		0xe17aU,0xd2ffU,0xb5f5U,0x7be1U,0xf7c2U,0xff8fU,0xef15U,0xce21U,0x8c49U,0x0899U,
		0x1132U,0x2264U,0x44c8U,0x8990U,0x032bU,0x0656U,0x0cacU,0x1958U,0x32b0U,0x6560U,
		0xcac0U,0x858bU,0x1b1dU,0x363aU,0x6c74U,0xd8e8U,0xa1dbU,0x53bdU,0xa77aU,0x5effU,
		0xbdfeU,0x6bf7U,0xd7eeU,0xbfd7U,0x6fa5U,0xdf4aU,0xae9fU,0x4d35U,0x9a6aU,0x24dfU,
		0x49beU,0x937cU,0x36f3U,0x6de6U,0xdbccU,0xa793U,0x5f2dU,0xbe5aU,0x6cbfU,0xd97eU,
		0xa2f7U,0x55e5U,0xabcaU,0x479fU,0x8f3eU,0x0e77U,0x1ceeU,0x39dcU,0x73b8U,0xe770U,
		0xdeebU,0xadddU,0x4bb1U,0x9762U,0x3ecfU,0x7d9eU,0xfb3cU,0xe673U,0xdcedU,0xa9d1U,
		0x43a9U,0x8752U,0x1eafU,0x3d5eU,0x7abcU,0xf578U,0xfafbU,0xe5fdU,0xdbf1U,0xa7e9U,
		0x5fd9U,0xbfb2U,0x6f6fU,0xdedeU,0xadb7U,0x4b65U,0x96caU,0x3d9fU,0x7b3eU,0xf67cU,
		0xfcf3U,0xe9edU,0xc3d1U,0x97a9U,0x3f59U,0x7eb2U,0xfd64U,0xeac3U,0xc58dU,0x9b11U,
		0x2629U,0x4c52U,0x98a4U,0x2143U,0x4286U,0x850cU,0x1a13U,0x3426U,0x684cU,0xd098U,
		0xb13bU,0x727dU,0xe4faU,0xd9ffU,0xa3f5U,0x57e1U,0xafc2U,0x4f8fU,0x9f1eU,0x2e37U,
		0x5c6eU,0xb8dcU,0x61b3U,0xc366U,0x96c7U,0x3d85U,0x7b0aU,0xf614U,0xfc23U,0xe84dU,
		0xc091U,0x9129U,0x3259U,0x64b2U,0xc964U,0x82c3U,0x158dU,0x2b1aU,0x5634U,0xac68U,
		0x48dbU,0x91b6U,0x3367U,0x66ceU,0xcd9cU,0x8b33U,0x066dU,0x0cdaU,0x19b4U,0x3368U,
		0x66d0U,0xcda0U,0x8b4bU,0x069dU,0x0d3aU,0x1a74U,0x34e8U,0x69d0U,0xd3a0U,0xb74bU,
		0x7e9dU,0xfd3aU,0xea7fU,0xc4f5U,0x99e1U,0x23c9U,0x4792U,0x8f24U,0x0e43U,0x1c86U,
		0x390cU,0x7218U,0xe430U,0xd86bU,0xa0ddU,0x51b1U,0xa362U,0x56cfU,0xad9eU,0x4b37U,
		0x966eU,0x3cd7U,0x79aeU,0xf35cU,0xf6b3U,0xfd6dU,0xead1U,0xc5a9U,0x9b59U,0x26b9U,
		0x4d72U,0x9ae4U,0x25c3U,0x4b86U,0x970cU,0x3e13U,0x7c26U,0xf84cU,0xe093U,0xd12dU,
		0xb251U,0x74a9U,0xe952U,0xc2afU,0x9555U,0x3aa1U,0x7542U,0xea84U,0xc503U,0x9a0dU,
		0x2411U,0x4822U,0x9044U,0x3083U,0x6106U,0xc20cU,0x9413U,0x382dU,0x705aU,0xe0b4U,
		0xd163U,0xb2cdU,0x7591U,0xeb22U,0xc64fU,0x9c95U,0x2921U,0x5242U,0xa484U,0x5903U,
		0xb206U,0x7407U,0xe80eU,0xc017U,0x9025U,0x3041U,0x6082U,0xc104U,0x9203U,0x340dU,
		0x681aU,0xd034U,0xb063U,0x70cdU,0xe19aU,0xd33fU,0xb675U,0x7ce1U,0xf9c2U,0xe38fU,
		0xd715U,0xbe21U,0x6c49U,0xd892U,0xa12fU,0x5255U,0xa4aaU,0x595fU,0xb2beU,0x7577U,
		0xeaeeU,0xc5d7U,0x9ba5U,0x2741U,0x4e82U,0x9d04U,0x2a03U,0x5406U,0xa80cU,0x4013U,
		0x8026U,0x1047U,0x208eU,0x411cU,0x8238U,0x147bU,0x28f6U,0x51ecU,0xa3d8U,0x57bbU,
		0xaf76U,0x4ee7U,0x9dceU,0x2b97U,0x572eU,0xae5cU,0x4cb3U,0x9966U,0x22c7U,0x458eU,
		0x8b1cU,0x0633U,0x0c66U,0x18ccU,0x3198U,0x6330U,0xc660U,0x9ccbU,0x299dU,0x533aU,
		0xa674U,0x5ce3U,0xb9c6U,0x6387U,0xc70eU,0x9e17U,0x2c25U,0x584aU,0xb094U,0x7123U,
		0xe246U,0xd487U,0xb905U,0x6201U,0xc402U,0x980fU,0x2015U,0x402aU,0x8054U,0x10a3U,
		0x2146U,0x428cU,0x8518U,0x1a3bU,0x3476U,0x68ecU,0xd1d8U,0xb3bbU,0x777dU,0xeefaU,
		0xcdffU,0x8bf5U,0x07e1U,0x0fc2U,0x1f84U,0x3f08U,0x7e10U,0xfc20U,0xe84bU,0xc09dU,
		0x9131U,0x3269U,0x64d2U,0xc9a4U,0x8343U,0x168dU,0x2d1aU,0x5a34U,0xb468U,0x78dbU,
		0xf1b6U,0xf367U,0xf6c5U,0xfd81U,0xeb09U,0xc619U,0x9c39U,0x2879U,0x50f2U,0xa1e4U,
		0x53c3U,0xa786U,0x5f07U,0xbe0eU,0x6c17U,0xd82eU,0xa057U,0x50a5U,0xa14aU,0x529fU,
		0xa53eU,0x5a77U,0xb4eeU,0x79d7U,0xf3aeU,0xf757U,0xfea5U,0xed41U,0xca89U,0x8519U,
		0x1a39U,0x3472U,0x68e4U,0xd1c8U,0xb39bU,0x773dU,0xee7aU,0xccffU,0x89f5U,0x03e1U,
		0x07c2U,0x0f84U,0x1f08U,0x3e10U,0x7c20U,0xf840U,0xe08bU,0xd11dU,0xb231U,0x7469U,
		0xe8d2U,0xc1afU,0x9355U,0x36a1U,0x6d42U,0xda84U,0xa503U,0x5a0dU,0xb41aU,0x783fU,
		0xf07eU,0xf0f7U,0xf1e5U,0xf3c1U,0xf789U,0xff19U,0xee39U,0xcc79U,0x88f9U,0x01f9U,
		0x03f2U,0x07e4U,0x0fc8U,0x1f90U,0x3f20U,0x7e40U,0xfc80U,0xe90bU,0xc21dU,0x9431U,
		0x3869U,0x70d2U,0xe1a4U,0xd343U,0xb68dU,0x7d11U,0xfa22U,0xe44fU,0xd895U,0xa121U,
		0x5249U,0xa492U,0x592fU,0xb25eU,0x74b7U,0xe96eU,0xc2d7U,0x95a5U,0x3b41U,0x7682U,
		0xed04U,0xca03U,0x840dU,0x1811U,0x3022U,0x6044U,0xc088U,0x911bU,0x323dU,0x647aU,
		0xc8f4U,0x81e3U,0x13cdU,0x279aU,0x4f34U,0x9e68U,0x2cdbU,0x59b6U,0xb36cU,0x76d3U,
		0xeda6U,0xcb47U,0x8685U,0x1d01U,0x3a02U,0x7404U,0xe808U,0xc01bU,0x903dU,0x3071U,
		0x60e2U,0xc1c4U,0x9383U,0x370dU,0x6e1aU,0xdc34U,0xa863U,0x40cdU,0x819aU,0x133fU,
		0x267eU,0x4cfcU,0x99f8U,0x23fbU,0x47f6U,0x8fecU,0x0fd3U,0x1fa6U,0x3f4cU,0x7e98U,
		0xfd30U,0xea6bU,0xc4ddU,0x99b1U,0x2369U,0x46d2U,0x8da4U,0x0b43U,0x1686U,0x2d0cU,
		0x5a18U,0xb430U,0x786bU,0xf0d6U,0xf1a7U,0xf345U,0xf681U,0xfd09U,0xea19U,0xc439U,
		0x9879U,0x20f9U,0x41f2U,0x83e4U,0x17c3U,0x2f86U,0x5f0cU,0xbe18U,0x6c3bU,0xd876U,
		0xa0e7U,0x51c5U,0xa38aU,0x571fU,0xae3eU,0x4c77U,0x98eeU,0x21d7U,0x43aeU,0x875cU,
		0x1eb3U,0x3d66U,0x7accU,0xf598U,0xfb3bU,0xe67dU,0xdcf1U,0xa9e9U,0x43d9U,0x87b2U,
		0x1f6fU,0x3edeU,0x7dbcU,0xfb78U,0xe6fbU,0xddfdU,0xabf1U,0x47e9U,0x8fd2U,0x0fafU,
		0x1f5eU,0x3ebcU,0x7d78U,0xfaf0U,0xe5ebU,0xdbddU,0xa7b1U,0x5f69U,0xbed2U,0x6dafU,
		0xdb5eU,0xa6b7U,0x5d65U,0xbacaU,0x659fU,0xcb3eU,0x8677U,0x1ce5U,0x39caU,0x7394U,
		0xe728U,0xde5bU,0xacbdU,0x4971U,0x92e2U,0x35cfU,0x6b9eU,0xd73cU,0xbe73U,0x6cedU,
		0xd9daU,0xa3bfU,0x5775U,0xaeeaU,0x4ddfU,0x9bbeU,0x2777U,0x4eeeU,0x9ddcU,0x2bb3U,
		0x5766U,0xaeccU,0x4d93U,0x9b26U,0x2647U,0x4c8eU,0x991cU,0x2233U,0x4466U,0x88ccU,
		0x0193U,0x0326U,0x064cU,0x0c98U,0x1930U,0x3260U,0x64c0U,0xc980U,0x830bU,0x161dU,
		0x2c3aU,0x5874U,0xb0e8U,0x71dbU,0xe3b6U,0xd767U,0xbec5U,0x6d81U,0xdb02U,0xa60fU,
		0x5c15U,0xb82aU,0x605fU,0xc0beU,0x9177U,0x32e5U,0x65caU,0xcb94U,0x8723U,0x1e4dU,
		0x3c9aU,0x7934U,0xf268U,0xf4dbU,0xf9bdU,0xe371U,0xd6e9U,0xbdd9U,0x6bb9U,0xd772U,
		0xbeefU,0x6dd5U,0xdbaaU,0xa75fU,0x5eb5U,0xbd6aU,0x6adfU,0xd5beU,0xbb77U,0x66e5U,
		0xcdcaU,0x8b9fU,0x0735U,0x0e6aU,0x1cd4U,0x39a8U,0x7350U,0xe6a0U,0xdd4bU,0xaa9dU,
		0x4531U,0x8a62U,0x04cfU,0x099eU,0x133cU,0x2678U,0x4cf0U,0x99e0U,0x23cbU,0x4796U,
		0x8f2cU,0x0e53U,0x1ca6U,0x394cU,0x7298U,0xe530U,0xda6bU,0xa4ddU,0x59b1U,0xb362U,
		0x76cfU,0xed9eU,0xcb37U,0x8665U,0x1cc1U,0x3982U,0x7304U,0xe608U,0xdc1bU,0xa83dU,
		0x4071U,0x80e2U,0x11cfU,0x239eU,0x473cU,0x8e78U,0x0cfbU,0x19f6U,0x33ecU,0x67d8U,
		0xcfb0U,0x8f6bU,0x0eddU,0x1dbaU,0x3b74U,0x76e8U,0xedd0U,0xcbabU,0x875dU,0x1eb1U,
		0x3d62U,0x7ac4U,0xf588U,0xfb1bU,0xe63dU,0xdc71U,0xa8e9U,0x41d9U,0x83b2U,0x176fU,
		0x2edeU,0x5dbcU,0xbb78U,0x66fbU,0xcdf6U,0x8be7U,0x07c5U,0x0f8aU,0x1f14U,0x3e28U,
		0x7c50U,0xf8a0U,0xe14bU,0xd29dU,0xb531U,0x7a69U,0xf4d2U,0xf9afU,0xe355U,0xd6a1U,
		0xbd49U,0x6a99U,0xd532U,0xba6fU,0x64d5U,0xc9aaU,0x835fU,0x16b5U,0x2d6aU,0x5ad4U,
		0xb5a8U,0x7b5bU,0xf6b6U,0xfd67U,0xeac5U,0xc581U,0x9b09U,0x2619U,0x4c32U,0x9864U,
		0x20c3U,0x4186U,0x830cU,0x1613U,0x2c26U,0x584cU,0xb098U,0x713bU,0xe276U,0xd4e7U,
		0xb9c5U,0x6381U,0xc702U,0x9e0fU,0x2c15U,0x582aU,0xb054U,0x70a3U,0xe146U,0xd287U,
		0xb505U,0x7a01U,0xf402U,0xf80fU,0xe015U,0xd021U,0xb049U,0x7099U,0xe132U,0xd26fU,
		0xb4d5U,0x79a1U,0xf342U,0xf68fU,0xfd15U,0xea21U,0xc449U,0x9899U,0x2139U,0x4272U,
		0x84e4U,0x19c3U,0x3386U,0x670cU,0xce18U,0x8c3bU,0x087dU,0x10faU,0x21f4U,0x43e8U,
		0x87d0U,0x1fabU,0x3f56U,0x7eacU,0xfd58U,0xeabbU,0xc57dU,0x9af1U,0x25e9U,0x4bd2U,
		0x97a4U,0x3f43U,0x7e86U,0xfd0cU,0xea13U,0xc42dU,0x9851U,0x20a9U,0x4152U,0x82a4U,
		0x1543U,0x2a86U,0x550cU,0xaa18U,0x443bU,0x8876U,0x00e7U,0x01ceU,0x039cU,0x0738U,
		0x0e70U,0x1ce0U,0x39c0U,0x7380U,0xe700U,0xde0bU,0xac1dU,0x4831U,0x9062U,0x30cfU,
		0x619eU,0xc33cU,0x9673U,0x3cedU,0x79daU,0xf3b4U,0xf763U,0xfecdU,0xed91U,0xcb29U,
		0x8659U,0x1cb9U,0x3972U,0x72e4U,0xe5c8U,0xdb9bU,0xa73dU,0x5e71U,0xbce2U,0x69cfU,
		0xd39eU,0xb737U,0x7e65U,0xfccaU,0xe99fU,0xc335U,0x9661U,0x3cc9U,0x7992U,0xf324U,
		0xf643U,0xfc8dU,0xe911U,0xc229U,0x9459U,0x38b9U,0x7172U,0xe2e4U,0xd5c3U,0xbb8dU,
		0x6711U,0xce22U,0x8c4fU,0x0895U,0x112aU,0x2254U,0x44a8U,0x8950U,0x02abU,0x0556U,
		0x0aacU,0x1558U,0x2ab0U,0x5560U,0xaac0U,0x458bU,0x8b16U,0x0627U,0x0c4eU,0x189cU,
		0x3138U,0x6270U,0xc4e0U,0x99cbU,0x239dU,0x473aU,0x8e74U,0x0ce3U,0x19c6U,0x338cU,
		0x6718U,0xce30U,0x8c6bU,0x08ddU,0x11baU,0x2374U,0x46e8U,0x8dd0U,0x0babU,0x1756U,
		0x2eacU,0x5d58U,0xbab0U,0x656bU,0xcad6U,0x85a7U,0x1b45U,0x368aU,0x6d14U,0xda28U,
		0xa45bU,0x58bdU,0xb17aU,0x72ffU,0xe5feU,0xdbf7U,0xa7e5U,0x5fc1U,0xbf82U,0x6f0fU,
		0xde1eU,0xac37U,0x4865U,0x90caU,0x319fU,0x633eU,0xc67cU,0x9cf3U,0x29edU,0x53daU,
		0xa7b4U,0x5f63U,0xbec6U,0x6d87U,0xdb0eU,0xa617U,0x5c25U,0xb84aU,0x609fU,0xc13eU,
		0x9277U,0x34e5U,0x69caU,0xd394U,0xb723U,0x7e4dU,0xfc9aU,0xe93fU,0xc275U,0x94e1U,
		0x39c9U,0x7392U,0xe724U,0xde43U,0xac8dU,0x4911U,0x9222U,0x344fU,0x689eU,0xd13cU,
		0xb273U,0x74edU,0xe9daU,0xc3bfU,0x9775U,0x3ee1U,0x7dc2U,0xfb84U,0xe703U,0xde0dU,
		0xac11U,0x4829U,0x9052U,0x30afU,0x615eU,0xc2bcU,0x9573U,0x3aedU,0x75daU,0xebb4U,
		0xc763U,0x9ecdU,0x2d91U,0x5b22U,0xb644U,0x7c83U,0xf906U,0xe207U,0xd405U,0xb801U,
		0x6009U,0xc012U,0x902fU,0x3055U,0x60aaU,0xc154U,0x92a3U,0x354dU,0x6a9aU,0xd534U,
		0xba63U,0x64cdU,0xc99aU,0x833fU,0x1675U,0x2ceaU,0x59d4U,0xb3a8U,0x775bU,0xeeb6U,
		0xcd67U,0x8ac5U,0x0581U,0x0b02U,0x1604U,0x2c08U,0x5810U,0xb020U,0x704bU,0xe096U,
		0xd127U,0xb245U,0x7481U,0xe902U,0xc20fU,0x9415U,0x3821U,0x7042U,0xe084U,0xd103U,
		0xb20dU,0x7411U,0xe822U,0xc04fU,0x9095U,0x3121U,0x6242U,0xc484U,0x9903U,0x220dU,
		0x441aU,0x8834U,0x0063U,0x00c6U,0x018cU,0x0318U,0x0630U,0x0c60U,0x18c0U,0x3180U,
		0x6300U,0xc600U,0x9c0bU,0x281dU,0x503aU,0xa074U,0x50e3U,0xa1c6U,0x5387U,0xa70eU,
		0x5e17U,0xbc2eU,0x6857U,0xd0aeU,0xb157U,0x72a5U,0xe54aU,0xda9fU,0xa535U,0x5a61U,
		0xb4c2U,0x798fU,0xf31eU,0xf637U,0xfc65U,0xe8c1U,0xc189U,0x9319U,0x3639U,0x6c72U,
		0xd8e4U,0xa1c3U,0x538dU,0xa71aU,0x5e3fU,0xbc7eU,0x68f7U,0xd1eeU,0xb3d7U,0x77a5U,
		0xef4aU,0xce9fU,0x8d35U,0x0a61U,0x14c2U,0x2984U,0x5308U,0xa610U,0x5c2bU,0xb856U,
		0x60a7U,0xc14eU,0x9297U,0x3525U,0x6a4aU,0xd494U,0xb923U,0x624dU,0xc49aU,0x993fU,
		0x2275U,0x44eaU,0x89d4U,0x03a3U,0x0746U,0x0e8cU,0x1d18U,0x3a30U,0x7460U,0xe8c0U,
		0xc18bU,0x931dU,0x3631U,0x6c62U,0xd8c4U,0xa183U,0x530dU,0xa61aU,0x5c3fU,0xb87eU,
		0x60f7U,0xc1eeU,0x93d7U,0x37a5U,0x6f4aU,0xde94U,0xad23U,0x4a4dU,0x949aU,0x393fU,
		0x727eU,0xe4fcU,0xd9f3U,0xa3edU,0x57d1U,0xafa2U,0x4f4fU,0x9e9eU,0x2d37U,0x5a6eU,
		0xb4dcU,0x79b3U,0xf366U,0xf6c7U,0xfd85U,0xeb01U,0xc609U,0x9c19U,0x2839U,0x5072U,
		0xa0e4U,0x51c3U,0xa386U,0x5707U,0xae0eU,0x4c17U,0x982eU,0x2057U,0x40aeU,0x815cU,
		0x12b3U,0x2566U,0x4accU,0x9598U,0x3b3bU,0x7676U,0xececU,0xc9d3U,0x83adU,0x1751U,
		0x2ea2U,0x5d44U,0xba88U,0x651bU,0xca36U,0x8467U,0x18c5U,0x318aU,0x6314U,0xc628U,
		0x9c5bU,0x28bdU,0x517aU,0xa2f4U,0x55e3U,0xabc6U,0x4787U,0x8f0eU,0x0e17U,0x1c2eU,
		0x385cU,0x70b8U,0xe170U,0xd2ebU,0xb5ddU,0x7bb1U,0xf762U,0xfecfU,0xed95U,0xcb21U,
		0x8649U,0x1c99U,0x3932U,0x7264U,0xe4c8U,0xd99bU,0xa33dU,0x5671U,0xace2U,0x49cfU,
		0x939eU,0x3737U,0x6e6eU,0xdcdcU,0xa9b3U,0x436dU,0x86daU,0x1dbfU,0x3b7eU,0x76fcU,
		0xedf8U,0xcbfbU,0x87fdU,0x1ff1U,0x3fe2U,0x7fc4U,0xff88U,0xef1bU,0xce3dU,0x8c71U,
		0x08e9U,0x11d2U,0x23a4U,0x4748U,0x8e90U,0x0d2bU,0x1a56U,0x34acU,0x6958U,0xd2b0U,
		0xb56bU,0x7addU,0xf5baU,0xfb7fU,0xe6f5U,0xdde1U,0xabc9U,0x4799U,0x8f32U,0x0e6fU,
		0x1cdeU,0x39bcU,0x7378U,0xe6f0U,0xddebU,0xabddU,0x47b1U,0x8f62U,0x0ecfU,0x1d9eU,
		0x3b3cU,0x7678U,0xecf0U,0xc9ebU,0x83ddU,0x17b1U,0x2f62U,0x5ec4U,0xbd88U,0x6b1bU,
		0xd636U,0xbc67U,0x68c5U,0xd18aU,0xb31fU,0x7635U,0xec6aU,0xc8dfU,0x81b5U,0x1361U,
		0x26c2U,0x4d84U,0x9b08U,0x261bU,0x4c36U,0x986cU,0x20d3U,0x41a6U,0x834cU,0x1693U,
		0x2d26U,0x5a4cU,0xb498U,0x793bU,0xf276U,0xf4e7U,0xf9c5U,0xe381U,0xd709U,0xbe19U,
		0x6c39U,0xd872U,0xa0efU,0x51d5U,0xa3aaU,0x575fU,0xaebeU,0x4d77U,0x9aeeU,0x25d7U,
		0x4baeU,0x975cU,0x3eb3U,0x7d66U,0xfaccU,0xe593U,0xdb2dU,0xa651U,0x5ca9U,0xb952U,
		0x62afU,0xc55eU,0x9ab7U,0x2565U,0x4acaU,0x9594U,0x3b23U,0x7646U,0xec8cU,0xc913U,
		0x822dU,0x1451U,0x28a2U,0x5144U,0xa288U,0x551bU,0xaa36U,0x4467U,0x88ceU,0x0197U,
		0x032eU,0x065cU,0x0cb8U,0x1970U,0x32e0U,0x65c0U,0xcb80U,0x870bU,0x1e1dU,0x3c3aU,
		0x7874U,0xf0e8U,0xf1dbU,0xf3bdU,0xf771U,0xfee9U,0xedd9U,0xcbb9U,0x8779U,0x1ef9U,
		0x3df2U,0x7be4U,0xf7c8U,0xff9bU,0xef3dU,0xce71U,0x8ce9U,0x09d9U,0x13b2U,0x2764U,
		0x4ec8U,0x9d90U,0x2b2bU,0x5656U,0xacacU,0x4953U,0x92a6U,0x3547U,0x6a8eU,0xd51cU,
		0xba33U,0x646dU,0xc8daU,0x81bfU,0x1375U,0x26eaU,0x4dd4U,0x9ba8U,0x275bU,0x4eb6U,
		0x9d6cU,0x2ad3U,0x55a6U,0xab4cU,0x4693U,0x8d26U,0x0a47U,0x148eU,0x291cU,0x5238U,
		0xa470U,0x58ebU,0xb1d6U,0x73a7U,0xe74eU,0xde97U,0xad25U,0x4a41U,0x9482U,0x390fU,
		0x721eU,0xe43cU,0xd873U,0xa0edU,0x51d1U,0xa3a2U,0x574fU,0xae9eU,0x4d37U,0x9a6eU,
		0x24d7U,0x49aeU,0x935cU,0x36b3U,0x6d66U,0xdaccU,0xa593U,0x5b2dU,0xb65aU,0x7cbfU,
		0xf97eU,0xe2f7U,0xd5e5U,0xbbc1U,0x6789U,0xcf12U,0x8e2fU,0x0c55U,0x18aaU,0x3154U,
		0x62a8U,0xc550U,0x9aabU,0x255dU,0x4abaU,0x9574U,0x3ae3U,0x75c6U,0xeb8cU,0xc713U,
		0x9e2dU,0x2c51U,0x58a2U,0xb144U,0x7283U,0xe506U,0xda07U,0xa405U,0x5801U,0xb002U,
		0x700fU,0xe01eU,0xd037U,0xb065U,0x70c1U,0xe182U,0xd30fU,0xb615U,0x7c21U,0xf842U,
		0xe08fU,0xd115U,0xb221U,0x7449U,0xe892U,0xc12fU,0x9255U,0x34a1U,0x6942U,0xd284U,
		0xb503U,0x7a0dU,0xf41aU,0xf83fU,0xe075U,0xd0e1U,0xb1c9U,0x7399U,0xe732U,0xde6fU,
		0xacd5U,0x49a1U,0x9342U,0x368fU,0x6d1eU,0xda3cU,0xa473U,0x58edU,0xb1daU,0x73bfU,
		0xe77eU,0xdef7U,0xade5U,0x4bc1U,0x9782U,0x3f0fU,0x7e1eU,0xfc3cU,0xe873U,0xc0edU,
		0x91d1U,0x33a9U,0x6752U,0xcea4U,0x8d43U,0x0a8dU,0x151aU,0x2a34U,0x5468U,0xa8d0U,
		0x41abU,0x8356U,0x16a7U,0x2d4eU,0x5a9cU,0xb538U,0x7a7bU,0xf4f6U,0xf9e7U,0xe3c5U,
		0xd781U,0xbf09U,0x6e19U,0xdc32U,0xa86fU,0x40d5U,0x81aaU,0x135fU,0x26beU,0x4d7cU,
		0x9af8U,0x25fbU,0x4bf6U,0x97ecU,0x3fd3U,0x7fa6U,0xff4cU,0xee93U,0xcd2dU,0x8a51U,
		0x04a9U,0x0952U,0x12a4U,0x2548U,0x4a90U,0x9520U,0x3a4bU,0x7496U,0xe92cU,0xc253U,
		0x94adU,0x3951U,0x72a2U,0xe544U,0xda83U,0xa50dU,0x5a11U,0xb422U,0x784fU,0xf09eU,
		0xf137U,0xf265U,0xf4c1U,0xf989U,0xe319U,0xd639U,0xbc79U,0x68f9U,0xd1f2U,0xb3efU,
		0x77d5U,0xefaaU,0xcf5fU,0x8eb5U,0x0d61U,0x1ac2U,0x3584U,0x6b08U,0xd610U,0xbc2bU,
		0x685dU,0xd0baU,0xb17fU,0x72f5U,0xe5eaU,0xdbdfU,0xa7b5U,0x5f61U,0xbec2U,0x6d8fU,
		0xdb1eU,0xa637U,0x5c65U,0xb8caU,0x619fU,0xc33eU,0x9677U,0x3ce5U,0x79caU,0xf394U,
		0xf723U,0xfe4dU,0xec91U,0xc929U,0x8259U,0x14b9U,0x2972U,0x52e4U,0xa5c8U,0x5b9bU,
		0xb736U,0x7e67U,0xfcceU,0xe997U,0xc325U,0x9641U,0x3c89U,0x7912U,0xf224U,0xf443U,
		0xf88dU,0xe111U,0xd229U,0xb459U,0x78b9U,0xf172U,0xf2efU,0xf5d5U,0xfba1U,0xe749U,
		0xde99U,0xad39U,0x4a79U,0x94f2U,0x39efU,0x73deU,0xe7bcU,0xdf73U,0xaeedU,0x4dd1U,
		0x9ba2U,0x274fU,0x4e9eU,0x9d3cU,0x2a73U,0x54e6U,0xa9ccU,0x4393U,0x8726U,0x1e47U,
		0x3c8eU,0x791cU,0xf238U,0xf47bU,0xf8fdU,0xe1f1U,0xd3e9U,0xb7d9U,0x7fb9U,0xff72U,
		0xeeefU,0xcdd5U,0x8ba1U,0x0749U,0x0e92U,0x1d24U,0x3a48U,0x7490U,0xe920U,0xc24bU,
		0x949dU,0x3931U,0x7262U,0xe4c4U,0xd983U,0xa30dU,0x5611U,0xac22U,0x484fU,0x909eU,
		0x3137U,0x626eU,0xc4dcU,0x99b3U,0x236dU,0x46daU,0x8db4U,0x0b63U,0x16c6U,0x2d8cU,
		0x5b18U,0xb630U,0x7c6bU,0xf8d6U,0xe1a7U,0xd345U,0xb681U,0x7d09U,0xfa12U,0xe42fU,
		0xd855U,0xa0a1U,0x5149U,0xa292U,0x552fU,0xaa5eU,0x44b7U,0x896eU,0x02d7U,0x05aeU,
		0x0b5cU,0x16b8U,0x2d70U,0x5ae0U,0xb5c0U,0x7b8bU,0xf716U,0xfe27U,0xec45U,0xc881U,
		0x8109U,0x1219U,0x2432U,0x4864U,0x90c8U,0x319bU,0x6336U,0xc66cU,0x9cd3U,0x29adU,
		0x535aU,0xa6b4U,0x5d63U,0xbac6U,0x6587U,0xcb0eU,0x8617U,0x1c25U,0x384aU,0x7094U,
		0xe128U,0xd25bU,0xb4bdU,0x7971U,0xf2e2U,0xf5cfU,0xfb95U,0xe721U,0xde49U,0xac99U,
		0x4939U,0x9272U,0x34efU,0x69deU,0xd3bcU,0xb773U,0x7eedU,0xfddaU,0xebbfU,0xc775U,
		0x9ee1U,0x2dc9U,0x5b92U,0xb724U,0x7e43U,0xfc86U,0xe907U,0xc205U,0x9401U,0x3809U,
		0x7012U,0xe024U,0xd043U,0xb08dU,0x7111U,0xe222U,0xd44fU,0xb895U,0x6121U,0xc242U,
		0x948fU,0x3915U,0x722aU,0xe454U,0xd8a3U,0xa14dU,0x5291U,0xa522U,0x5a4fU,0xb49eU,
		0x7937U,0xf26eU,0xf4d7U,0xf9a5U,0xe341U,0xd689U,0xbd19U,0x6a39U,0xd472U,0xb8efU,
		0x61d5U,0xc3aaU,0x975fU,0x3eb5U,0x7d6aU,0xfad4U,0xe5a3U,0xdb4dU,0xa691U,0x5d29U,
		0xba52U,0x64afU,0xc95eU,0x82b7U,0x1565U,0x2acaU,0x5594U,0xab28U,0x465bU,0x8cb6U,
		0x0967U,0x12ceU,0x259cU,0x4b38U,0x9670U,0x3cebU,0x79d6U,0xf3acU,0xf753U,0xfeadU,
		0xed51U,0xcaa9U,0x8559U,0x1ab9U,0x3572U,0x6ae4U,0xd5c8U,0xbb9bU,0x673dU,0xce7aU,
		0x8cffU,0x09f5U,0x13eaU,0x27d4U,0x4fa8U,0x9f50U,0x2eabU,0x5d56U,0xbaacU,0x6553U,
		0xcaa6U,0x8547U,0x1a85U,0x350aU,0x6a14U,0xd428U,0xb85bU,0x60bdU,0xc17aU,0x92ffU,
		0x35f5U,0x6beaU,0xd7d4U,0xbfa3U,0x6f4dU,0xde9aU,0xad3fU,0x4a75U,0x94eaU,0x39dfU,
		0x73beU,0xe77cU,0xdef3U,0xadedU,0x4bd1U,0x97a2U,0x3f4fU,0x7e9eU,0xfd3cU,0xea73U,
		0xc4edU,0x99d1U,0x23a9U,0x4752U,0x8ea4U,0x0d43U,0x1a86U,0x350cU,0x6a18U,0xd430U,
		0xb86bU,0x60ddU,0xc1baU,0x937fU,0x36f5U,0x6deaU,0xdbd4U,0xa7a3U,0x5f4dU,0xbe9aU,
		0x6d3fU,0xda7eU,0xa4f7U,0x59e5U,0xb3caU,0x779fU,0xef3eU,0xce77U,0x8ce5U,0x09c1U,
		0x1382U,0x2704U,0x4e08U,0x9c10U,0x282bU,0x5056U,0xa0acU,0x5153U,0xa2a6U,0x5547U,
		0xaa8eU,0x4517U,0x8a2eU,0x0457U,0x08aeU,0x115cU,0x22b8U,0x4570U,0x8ae0U,0x05cbU,
		0x0b96U,0x172cU,0x2e58U,0x5cb0U,0xb960U,0x62cbU,0xc596U,0x9b27U,0x2645U,0x4c8aU,
		0x9914U,0x2223U,0x4446U,0x888cU,0x0113U,0x0226U,0x044cU,0x0898U,0x1130U,0x2260U,
		0x44c0U,0x8980U,0x030bU,0x0616U,0x0c2cU,0x1858U,0x30b0U,0x6160U,0xc2c0U,0x958bU,
		0x3b1dU,0x763aU,0xec74U,0xc8e3U,0x81cdU,0x1391U,0x2722U,0x4e44U,0x9c88U,0x291bU,
		0x5236U,0xa46cU,0x58d3U,0xb1a6U,0x7347U,0xe68eU,0xdd17U,0xaa25U,0x4441U,0x8882U,
		0x010fU,0x021eU,0x043cU,0x0878U,0x10f0U,0x21e0U,0x43c0U,0x8780U,0x1f0bU,0x3e16U,
		0x7c2cU,0xf858U,0xe0bbU,0xd17dU,0xb2f1U,0x75e9U,0xebd2U,0xc7afU,0x9f55U,0x2ea1U,
		0x5d42U,0xba84U,0x6503U,0xca06U,0x8407U,0x1805U,0x300aU,0x6014U,0xc028U,0x905bU,
		0x30bdU,0x617aU,0xc2f4U,0x95e3U,0x3bcdU,0x779aU,0xef34U,0xce63U,0x8ccdU,0x0991U,
		0x1322U,0x2644U,0x4c88U,0x9910U,0x222bU,0x4456U,0x88acU,0x0153U,0x02a6U,0x054cU,
		0x0a98U,0x1530U,0x2a60U,0x54c0U,0xa980U,0x430bU,0x8616U,0x1c27U,0x384eU,0x709cU,
		0xe138U,0xd27bU,0xb4fdU,0x79f1U,0xf3e2U,0xf7cfU,0xff95U,0xef21U,0xce49U,0x8c99U,
		0x0939U,0x1272U,0x24e4U,0x49c8U,0x9390U,0x372bU,0x6e56U,0xdcacU,0xa953U,0x42adU,
		0x855aU,0x1abfU,0x357eU,0x6afcU,0xd5f8U,0xbbfbU,0x67fdU,0xcffaU,0x8fffU,0x0ff5U,
		0x1feaU,0x3fd4U,0x7fa8U,0xff50U,0xeeabU,0xcd5dU,0x8ab1U,0x0569U,0x0ad2U,0x15a4U,
		0x2b48U,0x5690U,0xad20U,0x4a4bU,0x9496U,0x3927U,0x724eU,0xe49cU,0xd933U,0xa26dU,
		0x54d1U,0xa9a2U,0x434fU,0x869eU,0x1d37U,0x3a6eU,0x74dcU,0xe9b8U,0xc37bU,0x96fdU,
		0x3df1U,0x7be2U,0xf7c4U,0xff83U,0xef0dU,0xce11U,0x8c29U,0x0859U,0x10b2U,0x2164U,
		0x42c8U,0x8590U,0x1b2bU,0x3656U,0x6cacU,0xd958U,0xa2bbU,0x557dU,0xaafaU,0x45ffU,
		0x8bfeU,0x07f7U,0x0feeU,0x1fdcU,0x3fb8U,0x7f70U,0xfee0U,0xedcbU,0xcb9dU,0x8731U,
		0x1e69U,0x3cd2U,0x79a4U,0xf348U,0xf69bU,0xfd3dU,0xea71U,0xc4e9U,0x99d9U,0x23b9U,
		0x4772U,0x8ee4U,0x0dc3U,0x1b86U,0x370cU,0x6e18U,0xdc30U,0xa86bU,0x40ddU,0x81baU,
		0x137fU,0x26feU,0x4dfcU,0x9bf8U,0x27fbU,0x4ff6U,0x9fecU,0x2fd3U,0x5fa6U,0xbf4cU,
		0x6e93U,0xdd26U,0xaa47U,0x4485U,0x890aU,0x021fU,0x043eU,0x087cU,0x10f8U,0x21f0U,
		0x43e0U,0x87c0U,0x1f8bU,0x3f16U,0x7e2cU,0xfc58U,0xe8bbU,0xc17dU,0x92f1U,0x35e9U,
		0x6bd2U,0xd7a4U,0xbf43U,0x6e8dU,0xdd1aU,0xaa3fU,0x4475U,0x88eaU,0x01dfU,0x03beU,
		0x077cU,0x0ef8U,0x1df0U,0x3be0U,0x77c0U,0xef80U,0xcf0bU,0x8e1dU,0x0c31U,0x1862U,
		0x30c4U,0x6188U,0xc310U,0x962bU,0x3c5dU,0x78baU,0xf174U,0xf2e3U,0xf5cdU,0xfb91U,
		0xe729U,0xde59U,0xacb9U,0x4979U,0x92f2U,0x35efU,0x6bdeU,0xd7bcU,0xbf73U,0x6eedU,
		0xdddaU,0xabbfU,0x4775U,0x8eeaU,0x0ddfU,0x1bbeU,0x377cU,0x6ef8U,0xddf0U,0xabebU,
		0x47ddU,0x8fbaU,0x0f7fU,0x1efeU,0x3dfcU,0x7bf8U,0xf7f0U,0xffebU,0xefddU,0xcfb1U,
		0x8f69U,0x0ed9U,0x1db2U,0x3b64U,0x76c8U,0xed90U,0xcb2bU,0x865dU,0x1cb1U,0x3962U,
		0x72c4U,0xe588U,0xdb1bU,0xa63dU,0x5c71U,0xb8e2U,0x61cfU,0xc39eU,0x9737U,0x3e65U,
		0x7ccaU,0xf994U,0xe323U,0xd64dU,0xbc91U,0x6929U,0xd252U,0xb4afU,0x7955U,0xf2aaU,
		0xf55fU,0xfab5U,0xe561U,0xdac9U,0xa599U,0x5b39U,0xb672U,0x7cefU,0xf9deU,0xe3b7U,
		0xd765U,0xbec1U,0x6d89U,0xdb12U,0xa62fU,0x5c55U,0xb8aaU,0x615fU,0xc2beU,0x9577U,
		0x3ae5U,0x75caU,0xeb94U,0xc723U,0x9e4dU,0x2c91U,0x5922U,0xb244U,0x7483U,0xe906U,
		0xc207U,0x9405U,0x3801U,0x7002U,0xe004U,0xd003U,0xb00dU,0x7011U,0xe022U,0xd04fU,
		0xb095U,0x7121U,0xe242U,0xd48fU,0xb915U,0x6221U,0xc442U,0x988fU,0x2115U,0x422aU,
		0x8454U,0x18a3U,0x3146U,0x628cU,0xc518U,0x9a3bU,0x247dU,0x48faU,0x91f4U,0x33e3U,
		0x67c6U,0xcf8cU,0x8f13U,0x0e2dU,0x1c5aU,0x38b4U,0x7168U,0xe2d0U,0xd5abU,0xbb5dU,
		0x66b1U,0xcd62U,0x8acfU,0x0595U,0x0b2aU,0x1654U,0x2ca8U,0x5950U,0xb2a0U,0x754bU,
		0xea96U,0xc527U,0x9a45U,0x2481U,0x4902U,0x9204U,0x3403U,0x6806U,0xd00cU,0xb013U,
		0x702dU,0xe05aU,0xd0bfU,0xb175U,0x72e1U,0xe5c2U,0xdb8fU,0xa715U,0x5e21U,0xbc42U,
		0x688fU,0xd11eU,0xb237U,0x7465U,0xe8caU,0xc19fU,0x9335U,0x3661U,0x6cc2U,0xd984U,
		0xa303U,0x560dU,0xac1aU,0x483fU,0x907eU,0x30f7U,0x61eeU,0xc3dcU,0x97b3U,0x3f6dU,
		0x7edaU,0xfdb4U,0xeb63U,0xc6cdU,0x9d91U,0x2b29U,0x5652U,0xaca4U,0x4943U,0x9286U,
		0x3507U,0x6a0eU,0xd41cU,0xb833U,0x606dU,0xc0daU,0x91bfU,0x3375U,0x66eaU,0xcdd4U,
		0x8ba3U,0x074dU,0x0e9aU,0x1d34U,0x3a68U,0x74d0U,0xe9a0U,0xc34bU,0x969dU,0x3d31U,
		0x7a62U,0xf4c4U,0xf983U,0xe30dU,0xd611U,0xbc29U,0x6859U,0xd0b2U,0xb16fU,0x72d5U,
		0xe5aaU,0xdb5fU,0xa6b5U,0x5d61U,0xbac2U,0x658fU,0xcb1eU,0x8637U,0x1c65U,0x38caU,
		0x7194U,0xe328U,0xd65bU,0xbcbdU,0x6971U,0xd2e2U,0xb5cfU,0x7b95U,0xf72aU,0xfe5fU,
		0xecb5U,0xc961U,0x82c9U,0x1599U,0x2b32U,0x5664U,0xacc8U,0x499bU,0x9336U,0x3667U,
		0x6cceU,0xd99cU,0xa333U,0x566dU,0xacdaU,0x49bfU,0x937eU,0x36f7U,0x6deeU,0xdbdcU,
		0xa7b3U,0x5f6dU,0xbedaU,0x6dbfU,0xdb7eU,0xa6f7U,0x5de5U,0xbbcaU,0x679fU,0xcf3eU,
		0x8e77U,0x0ce5U,0x19caU,0x3394U,0x6728U,0xce50U,0x8cabU,0x095dU,0x12baU,0x2574U,
		0x4ae8U,0x95d0U,0x3babU,0x7756U,0xeeacU,0xcd53U,0x8aadU,0x0551U,0x0aa2U,0x1544U,
		0x2a88U,0x5510U,0xaa20U,0x444bU,0x8896U,0x0127U,0x024eU,0x049cU,0x0938U,0x1270U,
		0x24e0U,0x49c0U,0x9380U,0x370bU,0x6e16U,0xdc2cU,0xa853U,0x40adU,0x815aU,0x12bfU,
		0x257eU,0x4afcU,0x95f8U,0x3bfbU,0x77f6U,0xefecU,0xcfd3U,0x8fadU,0x0f51U,0x1ea2U,
		0x3d44U,0x7a88U,0xf510U,0xfa2bU,0xe45dU,0xd8b1U,0xa169U,0x52d9U,0xa5b2U,0x5b6fU,
		0xb6deU,0x7db7U,0xfb6eU,0xe6d7U,0xdda5U,0xab41U,0x4689U,0x8d12U,0x0a2fU,0x145eU,
		0x28bcU,0x5178U,0xa2f0U,0x55ebU,0xabd6U,0x47a7U,0x8f4eU,0x0e97U,0x1d2eU,0x3a5cU,
		0x74b8U,0xe970U,0xc2ebU,0x95ddU,0x3bb1U,0x7762U,0xeec4U,0xcd83U,0x8b0dU,0x0611U,
		0x0c22U,0x1844U,0x3088U,0x6110U,0xc220U,0x944bU,0x389dU,0x713aU,0xe274U,0xd4e3U,
		0xb9cdU,0x6391U,0xc722U,0x9e4fU,0x2c95U,0x592aU,0xb254U,0x74a3U,0xe946U,0xc287U,
		0x9505U,0x3a01U,0x7402U,0xe804U,0xc003U,0x900dU,0x3011U,0x6022U,0xc044U,0x9083U,
		0x310dU,0x621aU,0xc434U,0x9863U,0x20cdU,0x419aU,0x8334U,0x1663U,0x2cc6U,0x598cU,
		0xb318U,0x763bU,0xec76U,0xc8e7U,0x81c5U,0x1381U,0x2702U,0x4e04U,0x9c08U,0x281bU,
		0x5036U,0xa06cU,0x50d3U,0xa1a6U,0x5347U,0xa68eU,0x5d17U,0xba2eU,0x6457U,0xc8aeU,
		0x8157U,0x12a5U,0x254aU,0x4a94U,0x9528U,0x3a5bU,0x74b6U,0xe96cU,0xc2d3U,0x95adU,
		0x3b51U,0x76a2U,0xed44U,0xca83U,0x850dU,0x1a11U,0x3422U,0x6844U,0xd088U,0xb11bU,
		0x723dU,0xe47aU,0xd8ffU,0xa1f5U,0x53e1U,0xa7c2U,0x5f8fU,0xbf1eU,0x6e37U,0xdc6eU,
		0xa8d7U,0x41a5U,0x834aU,0x169fU,0x2d3eU,0x5a7cU,0xb4f8U,0x79fbU,0xf3f6U,0xf7e7U,
		0xffc5U,0xef81U,0xcf09U,0x8e19U,0x0c39U,0x1872U,0x30e4U,0x61c8U,0xc390U,0x972bU,
		0x3e5dU,0x7cbaU,0xf974U,0xe2e3U,0xd5cdU,0xbb91U,0x6729U,0xce52U,0x8cafU,0x0955U,
		0x12aaU,0x2554U,0x4aa8U,0x9550U,0x3aabU,0x7556U,0xeaacU,0xc553U,0x9aadU,0x2551U,
		0x4aa2U,0x9544U,0x3a83U,0x7506U,0xea0cU,0xc413U,0x982dU,0x2051U,0x40a2U,0x8144U,
		0x1283U,0x2506U,0x4a0cU,0x9418U,0x383bU,0x7076U,0xe0ecU,0xd1d3U,0xb3adU,0x7751U,
		0xeea2U,0xcd4fU,0x8a95U,0x0521U,0x0a42U,0x1484U,0x2908U,0x5210U,0xa420U,0x584bU,
		0xb096U,0x7127U,0xe24eU,0xd497U,0xb925U,0x6241U,0xc482U,0x990fU,0x2215U,0x442aU,
		0x8854U,0x00a3U,0x0146U,0x028cU,0x0518U,0x0a30U,0x1460U,0x28c0U,0x5180U,0xa300U,
		0x560bU,0xac16U,0x4827U,0x904eU,0x3097U,0x612eU,0xc25cU,0x94b3U,0x396dU,0x72daU,
		0xe5b4U,0xdb63U,0xa6cdU,0x5d91U,0xbb22U,0x664fU,0xcc9eU,0x8937U,0x0265U,0x04caU,
		0x0994U,0x1328U,0x2650U,0x4ca0U,0x9940U,0x228bU,0x4516U,0x8a2cU,0x0453U,0x08a6U,
		0x114cU,0x2298U,0x4530U,0x8a60U,0x04cbU,0x0996U,0x132cU,0x2658U,0x4cb0U,0x9960U,
		0x22cbU,0x4596U,0x8b2cU,0x0653U,0x0ca6U,0x194cU,0x3298U,0x6530U,0xca60U,0x84cbU,
		0x199dU,0x333aU,0x6674U,0xcce8U,0x89dbU,0x03bdU,0x077aU,0x0ef4U,0x1de8U,0x3bd0U,
		0x77a0U,0xef40U,0xce8bU,0x8d1dU,0x0a31U,0x1462U,0x28c4U,0x5188U,0xa310U,0x562bU,
		0xac56U,0x48a7U,0x914eU,0x3297U,0x652eU,0xca5cU,0x84b3U,0x196dU,0x32daU,0x65b4U,
		0xcb68U,0x86dbU,0x1dbdU,0x3b7aU,0x76f4U,0xede8U,0xcbdbU,0x87bdU,0x1f71U,0x3ee2U,
		0x7dc4U,0xfb88U,0xe71bU,0xde3dU,0xac71U,0x48e9U,0x91d2U,0x33afU,0x675eU,0xcebcU,
		0x8d73U,0x0aedU,0x15daU,0x2bb4U,0x5768U,0xaed0U,0x4dabU,0x9b56U,0x26a7U,0x4d4eU,
		0x9a9cU,0x2533U,0x4a66U,0x94ccU,0x3993U,0x7326U,0xe64cU,0xdc93U,0xa92dU,0x4251U,
		0x84a2U,0x194fU,0x329eU,0x653cU,0xca78U,0x84fbU,0x19fdU,0x33faU,0x67f4U,0xcfe8U,
		0x8fdbU,0x0fbdU,0x1f7aU,0x3ef4U,0x7de8U,0xfbd0U,0xe7abU,0xdf5dU,0xaeb1U,0x4d69U,
		0x9ad2U,0x25afU,0x4b5eU,0x96bcU,0x3d73U,0x7ae6U,0xf5ccU,0xfb93U,0xe72dU,0xde51U,
		0xaca9U,0x4959U,0x92b2U,0x356fU,0x6adeU,0xd5bcU,0xbb73U,0x66edU,0xcddaU,0x8bbfU,
		0x0775U,0x0eeaU,0x1dd4U,0x3ba8U,0x7750U,0xeea0U,0xcd4bU,0x8a9dU,0x0531U,0x0a62U,
		0x14c4U,0x2988U,0x5310U,0xa620U,0x5c4bU,0xb896U,0x6127U,0xc24eU,0x9497U,0x3925U,
		0x724aU,0xe494U,0xd923U,0xa24dU,0x5491U,0xa922U,0x424fU,0x849eU,0x1937U,0x326eU,
		0x64dcU,0xc9b8U,0x837bU,0x16fdU,0x2dfaU,0x5bf4U,0xb7e8U,0x7fdbU,0xffb6U,0xef67U,
		0xcec5U,0x8d81U,0x0b09U,0x1612U,0x2c24U,0x5848U,0xb090U,0x712bU,0xe256U,0xd4a7U,
		0xb945U,0x6281U,0xc502U,0x9a0fU,0x2415U,0x482aU,0x9054U,0x30a3U,0x6146U,0xc28cU,
		0x9513U,0x3a2dU,0x745aU,0xe8b4U,0xc163U,0x92cdU,0x3591U,0x6b22U,0xd644U,0xbc83U,
		0x690dU,0xd21aU,0xb43fU,0x7875U,0xf0eaU,0xf1dfU,0xf3b5U,0xf761U,0xfec9U,0xed99U,
		0xcb39U,0x8679U,0x1cf9U,0x39f2U,0x73e4U,0xe7c8U,0xdf9bU,0xaf3dU,0x4e71U,0x9ce2U,
		0x29cfU,0x539eU,0xa73cU,0x5e73U,0xbce6U,0x69c7U,0xd38eU,0xb717U,0x7e25U,0xfc4aU,
		0xe89fU,0xc135U,0x9261U,0x34c9U,0x6992U,0xd324U,0xb643U,0x7c8dU,0xf91aU,0xe23fU,
		0xd475U,0xb8e1U,0x61c9U,0xc392U,0x972fU,0x3e55U,0x7caaU,0xf954U,0xe2a3U,0xd54dU,
		0xba91U,0x6529U,0xca52U,0x84afU,0x1955U,0x32aaU,0x6554U,0xcaa8U,0x855bU,0x1abdU,
		0x357aU,0x6af4U,0xd5e8U,0xbbdbU,0x67bdU,0xcf7aU,0x8effU,0x0df5U,0x1beaU,0x37d4U,
		0x6fa8U,0xdf50U,0xaeabU,0x4d5dU,0x9abaU,0x257fU,0x4afeU,0x95fcU,0x3bf3U,0x77e6U,
		0xefccU,0xcf93U,0x8f2dU,0x0e51U,0x1ca2U,0x3944U,0x7288U,0xe510U,0xda2bU,0xa45dU,
		0x58b1U,0xb162U,0x72cfU,0xe59eU,0xdb37U,0xa665U,0x5cc1U,0xb982U,0x630fU,0xc61eU,
		0x9c37U,0x2865U,0x50caU,0xa194U,0x5323U,0xa646U,0x5c87U,0xb90eU,0x6217U,0xc42eU,
		0x9857U,0x20a5U,0x414aU,0x8294U,0x1523U,0x2a46U,0x548cU,0xa918U,0x423bU,0x8476U,
		0x18e7U,0x31ceU,0x639cU,0xc738U,0x9e7bU,0x2cfdU,0x59faU,0xb3f4U,0x77e3U,0xefc6U,
		0xcf87U,0x8f05U,0x0e01U,0x1c02U,0x3804U,0x7008U,0xe010U,0xd02bU,0xb05dU,0x70b1U,
		0xe162U,0xd2cfU,0xb595U,0x7b21U,0xf642U,0xfc8fU,0xe915U,0xc221U,0x9449U,0x3899U,
		0x7132U,0xe264U,0xd4c3U,0xb98dU,0x6311U,0xc622U,0x9c4fU,0x2895U,0x512aU,0xa254U,
		0x54a3U,0xa946U,0x4287U,0x850eU,0x1a17U,0x342eU,0x685cU,0xd0b8U,0xb17bU,0x72fdU,
		0xe5faU,0xdbffU,0xa7f5U,0x5fe1U,0xbfc2U,0x6f8fU,0xdf1eU,0xae37U,0x4c65U,0x98caU,
		0x219fU,0x433eU,0x867cU,0x1cf3U,0x39e6U,0x73ccU,0xe798U,0xdf3bU,0xae7dU,0x4cf1U,
		0x99e2U,0x23cfU,0x479eU,0x8f3cU,0x0e73U,0x1ce6U,0x39ccU,0x7398U,0xe730U,0xde6bU,
		0xacddU,0x49b1U,0x9362U,0x36cfU,0x6d9eU,0xdb3cU,0xa673U,0x5cedU,0xb9daU,0x63bfU,
		0xc77eU,0x9ef7U,0x2de5U,0x5bcaU,0xb794U,0x7f23U,0xfe46U,0xec87U,0xc905U,0x8201U,
		0x1409U,0x2812U,0x5024U,0xa048U,0x509bU,0xa136U,0x5267U,0xa4ceU,0x5997U,0xb32eU,
		0x7657U,0xecaeU,0xc957U,0x82a5U,0x1541U,0x2a82U,0x5504U,0xaa08U,0x441bU,0x8836U,
		0x0067U,0x00ceU,0x019cU,0x0338U,0x0670U,0x0ce0U,0x19c0U,0x3380U,0x6700U,0xce00U,
		0x8c0bU,0x081dU,0x103aU,0x2074U,0x40e8U,0x81d0U,0x13abU,0x2756U,0x4eacU,0x9d58U,
		0x2abbU,0x5576U,0xaaecU,0x45d3U,0x8ba6U,0x0747U,0x0e8eU,0x1d1cU,0x3a38U,0x7470U,
		0xe8e0U,0xc1cbU,0x939dU,0x3731U,0x6e62U,0xdcc4U,0xa983U,0x430dU,0x861aU,0x1c3fU,
		0x387eU,0x70fcU,0xe1f8U,0xd3fbU,0xb7fdU,0x7ff1U,0xffe2U,0xefcfU,0xcf95U,0x8f21U,
		0x0e49U,0x1c92U,0x3924U,0x7248U,0xe490U,0xd92bU,0xa25dU,0x54b1U,0xa962U,0x42cfU,
		0x859eU,0x1b37U,0x366eU,0x6cdcU,0xd9b8U,0xa37bU,0x56fdU,0xadfaU,0x4bffU,0x97feU,
		0x3ff7U,0x7feeU,0xffdcU,0xefb3U,0xcf6dU,0x8ed1U,0x0da9U,0x1b52U,0x36a4U,0x6d48U,
		0xda90U,0xa52bU,0x5a5dU,0xb4baU,0x797fU,0xf2feU,0xf5f7U,0xfbe5U,0xe7c1U,0xdf89U,
		0xaf19U,0x4e39U,0x9c72U,0x28efU,0x51deU,0xa3bcU,0x5773U,0xaee6U,0x4dc7U,0x9b8eU,
		0x2717U,0x4e2eU,0x9c5cU,0x28b3U,0x5166U,0xa2ccU,0x5593U,0xab26U,0x4647U,0x8c8eU,
		0x0917U,0x122eU,0x245cU,0x48b8U,0x9170U,0x32ebU,0x65d6U,0xcbacU,0x8753U,0x1eadU,
		0x3d5aU,0x7ab4U,0xf568U,0xfadbU,0xe5bdU,0xdb71U,0xa6e9U,0x5dd9U,0xbbb2U,0x676fU,
		0xcedeU,0x8db7U,0x0b65U,0x16caU,0x2d94U,0x5b28U,0xb650U,0x7cabU,0xf956U,0xe2a7U,
		0xd545U,0xba81U,0x6509U,0xca12U,0x842fU,0x1855U,0x30aaU,0x6154U,0xc2a8U,0x955bU,
		0x3abdU,0x757aU,0xeaf4U,0xc5e3U,0x9bcdU,0x2791U,0x4f22U,0x9e44U,0x2c83U,0x5906U,
		0xb20cU,0x7413U,0xe826U,0xc047U,0x9085U,0x3101U,0x6202U,0xc404U,0x9803U,0x200dU,
		0x401aU,0x8034U,0x1063U,0x20c6U,0x418cU,0x8318U,0x163bU,0x2c76U,0x58ecU,0xb1d8U,
		0x73bbU,0xe776U,0xdee7U,0xadc5U,0x4b81U,0x9702U,0x3e0fU,0x7c1eU,0xf83cU,0xe073U,
		0xd0edU,0xb1d1U,0x73a9U,0xe752U,0xdeafU,0xad55U,0x4aa1U,0x9542U,0x3a8fU,0x751eU,
		0xea3cU,0xc473U,0x98edU,0x21d1U,0x43a2U,0x8744U,0x1e83U,0x3d06U,0x7a0cU,0xf418U,
		0xf83bU,0xe07dU,0xd0f1U,0xb1e9U,0x73d9U,0xe7b2U,0xdf6fU,0xaed5U,0x4da1U,0x9b42U,
		0x268fU,0x4d1eU,0x9a3cU,0x2473U,0x48e6U,0x91ccU,0x3393U,0x6726U,0xce4cU,0x8c93U,
		0x092dU,0x125aU,0x24b4U,0x4968U,0x92d0U,0x35abU,0x6b56U,0xd6acU,0xbd53U,0x6aadU,
		0xd55aU,0xbabfU,0x6575U,0xcaeaU,0x85dfU,0x1bb5U,0x376aU,0x6ed4U,0xdda8U,0xab5bU,
		0x46bdU,0x8d7aU,0x0affU,0x15feU,0x2bfcU,0x57f8U,0xaff0U,0x4febU,0x9fd6U,0x2fa7U,
		0x5f4eU,0xbe9cU,0x6d33U,0xda66U,0xa4c7U,0x5985U,0xb30aU,0x761fU,0xec3eU,0xc877U,
		0x80e5U,0x11c1U,0x2382U,0x4704U,0x8e08U,0x0c1bU,0x1836U,0x306cU,0x60d8U,0xc1b0U,
		0x936bU,0x36ddU,0x6dbaU,0xdb74U,0xa6e3U,0x5dcdU,0xbb9aU,0x673fU,0xce7eU,0x8cf7U,
		0x09e5U,0x13caU,0x2794U,0x4f28U,0x9e50U,0x2cabU,0x5956U,0xb2acU,0x7553U,0xeaa6U,
		0xc547U,0x9a85U,0x2501U,0x4a02U,0x9404U,0x3803U,0x7006U,0xe00cU,0xd013U,0xb02dU,
		0x7051U,0xe0a2U,0xd14fU,0xb295U,0x7521U,0xea42U,0xc48fU,0x9915U,0x2221U,0x4442U,
		0x8884U,0x0103U,0x0206U,0x040cU,0x0818U,0x1030U,0x2060U,0x40c0U,0x8180U,0x130bU,
		0x2616U,0x4c2cU,0x9858U,0x20bbU,0x4176U,0x82ecU,0x15d3U,0x2ba6U,0x574cU,0xae98U,
		0x4d3bU,0x9a76U,0x24e7U,0x49ceU,0x939cU,0x3733U,0x6e66U,0xdcccU,0xa993U,0x432dU,
		0x865aU,0x1cbfU,0x397eU,0x72fcU,0xe5f8U,0xdbfbU,0xa7fdU,0x5ff1U,0xbfe2U,0x6fcfU,
		0xdf9eU,0xaf37U,0x4e65U,0x9ccaU,0x299fU,0x533eU,0xa67cU,0x5cf3U,0xb9e6U,0x63c7U,
		0xc78eU,0x9f17U,0x2e25U,0x5c4aU,0xb894U,0x6123U,0xc246U,0x9487U,0x3905U,0x720aU,
		0xe414U,0xd823U,0xa04dU,0x5091U,0xa122U,0x524fU,0xa49eU,0x5937U,0xb26eU,0x74d7U,
		0xe9aeU,0xc357U,0x96a5U,0x3d41U,0x7a82U,0xf504U,0xfa03U,0xe40dU,0xd811U,0xa029U,
		0x5059U,0xa0b2U,0x516fU,0xa2deU,0x55b7U,0xab6eU,0x46d7U,0x8daeU,0x0b57U,0x16aeU,
		0x2d5cU,0x5ab8U,0xb570U,0x7aebU,0xf5d6U,0xfba7U,0xe745U,0xde81U,0xad09U,0x4a19U,
		0x9432U,0x386fU,0x70deU,0xe1bcU,0xd373U,0xb6edU,0x7dd1U,0xfba2U,0xe74fU,0xde95U,
		0xad21U,0x4a49U,0x9492U,0x392fU,0x725eU,0xe4bcU,0xd973U,0xa2edU,0x55d1U,0xaba2U,
		0x474fU,0x8e9eU,0x0d37U,0x1a6eU,0x34dcU,0x69b8U,0xd370U,0xb6ebU,0x7dddU,0xfbbaU,
		0xe77fU,0xdef5U,0xade1U,0x4bc9U,0x9792U,0x3f2fU,0x7e5eU,0xfcbcU,0xe973U,0xc2edU,
		0x95d1U,0x3ba9U,0x7752U,0xeea4U,0xcd43U,0x8a8dU,0x0511U,0x0a22U,0x1444U,0x2888U,
		0x5110U,0xa220U,0x544bU,0xa896U,0x4127U,0x824eU,0x1497U,0x292eU,0x525cU,0xa4b8U,
		0x597bU,0xb2f6U,0x75e7U,0xebceU,0xc797U,0x9f25U,0x2e41U,0x5c82U,0xb904U,0x6203U,
		0xc406U,0x9807U,0x2005U,0x400aU,0x8014U,0x1023U,0x2046U,0x408cU,0x8118U,0x123bU,
		0x2476U,0x48ecU,0x91d8U,0x33bbU,0x6776U,0xceecU,0x8dd3U,0x0badU,0x175aU,0x2eb4U,
		0x5d68U,0xbad0U,0x65abU,0xcb56U,0x86a7U,0x1d45U,0x3a8aU,0x7514U,0xea28U,0xc45bU,
		0x98bdU,0x2171U,0x42e2U,0x85c4U,0x1b83U,0x3706U,0x6e0cU,0xdc18U,0xa83bU,0x407dU,
		0x80faU,0x11ffU,0x23feU,0x47fcU,0x8ff8U,0x0ffbU,0x1ff6U,0x3fecU,0x7fd8U,0xffb0U,
		0xef6bU,0xceddU,0x8db1U,0x0b69U,0x16d2U,0x2da4U,0x5b48U,0xb690U,0x7d2bU,0xfa56U,
		0xe4a7U,0xd945U,0xa281U,0x5509U,0xaa12U,0x442fU,0x885eU,0x00b7U,0x016eU,0x02dcU,
		0x05b8U,0x0b70U,0x16e0U,0x2dc0U,0x5b80U,0xb700U,0x7e0bU,0xfc16U,0xe827U,0xc045U,
		0x9081U,0x3109U,0x6212U,0xc424U,0x9843U,0x208dU,0x411aU,0x8234U,0x1463U,0x28c6U,
		0x518cU,0xa318U,0x563bU,0xac76U,0x48e7U,0x91ceU,0x3397U,0x672eU,0xce5cU,0x8cb3U,
		0x096dU,0x12daU,0x25b4U,0x4b68U,0x96d0U,0x3dabU,0x7b56U,0xf6acU,0xfd53U,0xeaadU,
		0xc551U,0x9aa9U,0x2559U,0x4ab2U,0x9564U,0x3ac3U,0x7586U,0xeb0cU,0xc613U,0x9c2dU,
		0x2851U,0x50a2U,0xa144U,0x5283U,0xa506U,0x5a07U,0xb40eU,0x7817U,0xf02eU,0xf057U,
		0xf0a5U,0xf141U,0xf289U,0xf519U,0xfa39U,0xe479U,0xd8f9U,0xa1f9U,0x53f9U,0xa7f2U,
		0x5fefU,0xbfdeU,0x6fb7U,0xdf6eU,0xaed7U,0x4da5U,0x9b4aU,0x269fU,0x4d3eU,0x9a7cU,
		0x24f3U,0x49e6U,0x93ccU,0x3793U,0x6f26U,0xde4cU,0xac93U,0x492dU,0x925aU,0x34bfU,
		0x697eU,0xd2fcU,0xb5f3U,0x7bedU,0xf7daU,0xffbfU,0xef75U,0xcee1U,0x8dc9U,0x0b99U,
		0x1732U,0x2e64U,0x5cc8U,0xb990U,0x632bU,0xc656U,0x9ca7U,0x2945U,0x528aU,0xa514U,
		0x5a23U,0xb446U,0x7887U,0xf10eU,0xf217U,0xf425U,0xf841U,0xe089U,0xd119U,0xb239U,
		0x7479U,0xe8f2U,0xc1efU,0x93d5U,0x37a1U,0x6f42U,0xde84U,0xad03U,0x4a0dU,0x941aU,
		0x383fU,0x707eU,0xe0fcU,0xd1f3U,0xb3edU,0x77d1U,0xefa2U,0xcf4fU,0x8e95U,0x0d21U,
		0x1a42U,0x3484U,0x6908U,0xd210U,0xb42bU,0x785dU,0xf0baU,0xf17fU,0xf2f5U,0xf5e1U,
		0xfbc9U,0xe799U,0xdf39U,0xae79U,0x4cf9U,0x99f2U,0x23efU,0x47deU,0x8fbcU,0x0f73U,
		0x1ee6U,0x3dccU,0x7b98U,0xf730U,0xfe6bU,0xecddU,0xc9b1U,0x8369U,0x16d9U,0x2db2U,
		0x5b64U,0xb6c8U,0x7d9bU,0xfb36U,0xe667U,0xdcc5U,0xa981U,0x4309U,0x8612U,0x1c2fU,
		0x385eU,0x70bcU,0xe178U,0xd2fbU,0xb5fdU,0x7bf1U,0xf7e2U,0xffcfU,0xef95U,0xcf21U,
		0x8e49U,0x0c99U,0x1932U,0x3264U,0x64c8U,0xc990U,0x832bU,0x165dU,0x2cbaU,0x5974U,
		0xb2e8U,0x75dbU,0xebb6U,0xc767U,0x9ec5U,0x2d81U,0x5b02U,0xb604U,0x7c03U,0xf806U,
		0xe007U,0xd005U,0xb001U,0x7009U,0xe012U,0xd02fU,0xb055U,0x70a1U,0xe142U,0xd28fU,
		0xb515U,0x7a21U,0xf442U,0xf88fU,0xe115U,0xd221U,0xb449U,0x7899U,0xf132U,0xf26fU,
		0xf4d5U,0xf9a1U,0xe349U,0xd699U,0xbd39U,0x6a79U,0xd4f2U,0xb9efU,0x63d5U,0xc7aaU,
		0x9f5fU,0x2eb5U,0x5d6aU,0xbad4U,0x65a3U,0xcb46U,0x8687U,0x1d05U,0x3a0aU,0x7414U,
		0xe828U,0xc05bU,0x90bdU,0x3171U,0x62e2U,0xc5c4U,0x9b83U,0x270dU,0x4e1aU,0x9c34U,
		0x2863U,0x50c6U,0xa18cU,0x5313U,0xa626U,0x5c47U,0xb88eU,0x6117U,0xc22eU,0x9457U,
		0x38a5U,0x714aU,0xe294U,0xd523U,0xba4dU,0x6491U,0xc922U,0x824fU,0x1495U,0x292aU,
		0x5254U,0xa4a8U,0x595bU,0xb2b6U,0x7567U,0xeaceU,0xc597U,0x9b25U,0x2641U,0x4c82U,
		0x9904U,0x2203U,0x4406U,0x880cU,0x0013U,0x0026U,0x004cU,0x0098U,0x0130U,0x0260U,
		0x04c0U,0x0980U,0x1300U,0x2600U,0x4c00U,0x9800U,0x200bU,0x4016U,0x802cU,0x1053U,
		0x20a6U,0x414cU,0x8298U,0x153bU,0x2a76U,0x54ecU,0xa9d8U,0x43bbU,0x8776U,0x1ee7U,
		0x3dceU,0x7b9cU,0xf738U,0xfe7bU,0xecfdU,0xc9f1U,0x83e9U,0x17d9U,0x2fb2U,0x5f64U,
		0xbec8U,0x6d9bU,0xdb36U,0xa667U,0x5cc5U,0xb98aU,0x631fU,0xc63eU,0x9c77U,0x28e5U,
		0x51caU,0xa394U,0x5723U,0xae46U,0x4c87U,0x990eU,0x2217U,0x442eU,0x885cU,0x00b3U,
		0x0166U,0x02ccU,0x0598U,0x0b30U,0x1660U,0x2cc0U,0x5980U,0xb300U,0x760bU,0xec16U,
		0xc827U,0x8045U,0x1081U,0x2102U,0x4204U,0x8408U,0x181bU,0x3036U,0x606cU,0xc0d8U,
		0x91bbU,0x337dU,0x66faU,0xcdf4U,0x8be3U,0x07cdU,0x0f9aU,0x1f34U,0x3e68U,0x7cd0U,
		0xf9a0U,0xe34bU,0xd69dU,0xbd31U,0x6a69U,0xd4d2U,0xb9afU,0x6355U,0xc6aaU,0x9d5fU,
		0x2ab5U,0x556aU,0xaad4U,0x45a3U,0x8b46U,0x0687U,0x0d0eU,0x1a1cU,0x3438U,0x6870U,
		0xd0e0U,0xb1cbU,0x739dU,0xe73aU,0xde7fU,0xacf5U,0x49e1U,0x93c2U,0x378fU,0x6f1eU,
		0xde3cU,0xac73U,0x48edU,0x91daU,0x33bfU,0x677eU,0xcefcU,0x8df3U,0x0bedU,0x17daU,
		0x2fb4U,0x5f68U,0xbed0U,0x6dabU,0xdb56U,0xa6a7U,0x5d45U,0xba8aU,0x651fU,0xca3eU,
		0x8477U,0x18e5U,0x31caU,0x6394U,0xc728U,0x9e5bU,0x2cbdU,0x597aU,0xb2f4U,0x75e3U,
		0xebc6U,0xc787U,0x9f05U,0x2e01U,0x5c02U,0xb804U,0x6003U,0xc006U,0x9007U,0x3005U,
		0x600aU,0xc014U,0x9023U,0x304dU,0x609aU,0xc134U,0x9263U,0x34cdU,0x699aU,0xd334U,
		0xb663U,0x7ccdU,0xf99aU,0xe33fU,0xd675U,0xbce1U,0x69c9U,0xd392U,0xb72fU,0x7e55U,
		0xfcaaU,0xe95fU,0xc2b5U,0x9561U,0x3ac9U,0x7592U,0xeb24U,0xc643U,0x9c8dU,0x2911U,
		0x5222U,0xa444U,0x5883U,0xb106U,0x7207U,0xe40eU,0xd817U,0xa025U,0x5041U,0xa082U,
		0x510fU,0xa21eU,0x5437U,0xa86eU,0x40d7U,0x81aeU,0x1357U,0x26aeU,0x4d5cU,0x9ab8U,
		0x257bU,0x4af6U,0x95ecU,0x3bd3U,0x77a6U,0xef4cU,0xce93U,0x8d2dU,0x0a51U,0x14a2U,
		0x2944U,0x5288U,0xa510U,0x5a2bU,0xb456U,0x78a7U,0xf14eU,0xf297U,0xf525U,0xfa41U,
		0xe489U,0xd919U,0xa239U,0x5479U,0xa8f2U,0x41efU,0x83deU,0x17b7U,0x2f6eU,0x5edcU,
		0xbdb8U,0x6b7bU,0xd6f6U,0xbde7U,0x6bc5U,0xd78aU,0xbf1fU,0x6e35U,0xdc6aU,0xa8dfU,
		0x41b5U,0x836aU,0x16dfU,0x2dbeU,0x5b7cU,0xb6f8U,0x7dfbU,0xfbf6U,0xe7e7U,0xdfc5U,
		0xaf81U,0x4f09U,0x9e12U,0x2c2fU,0x585eU,0xb0bcU,0x7173U,0xe2e6U,0xd5c7U,0xbb85U,
		0x6701U,0xce02U,0x8c0fU,0x0815U,0x102aU,0x2054U,0x40a8U,0x8150U,0x12abU,0x2556U,
		0x4aacU,0x9558U,0x3abbU,0x7576U,0xeaecU,0xc5d3U,0x9badU,0x2751U,0x4ea2U,0x9d44U,
		0x2a83U,0x5506U,0xaa0cU,0x4413U,0x8826U,0x0047U,0x008eU,0x011cU,0x0238U,0x0470U,
		0x08e0U,0x11c0U,0x2380U,0x4700U,0x8e00U,0x0c0bU,0x1816U,0x302cU,0x6058U,0xc0b0U,
		0x916bU,0x32ddU,0x65baU,0xcb74U,0x86e3U,0x1dcdU,0x3b9aU,0x7734U,0xee68U,0xccdbU,
		0x89bdU,0x0371U,0x06e2U,0x0dc4U,0x1b88U,0x3710U,0x6e20U,0xdc40U,0xa88bU,0x411dU,
		0x823aU,0x147fU,0x28feU,0x51fcU,0xa3f8U,0x57fbU,0xaff6U,0x4fe7U,0x9fceU,0x2f97U,
		0x5f2eU,0xbe5cU,0x6cb3U,0xd966U,0xa2c7U,0x5585U,0xab0aU,0x461fU,0x8c3eU,0x0877U,
		0x10eeU,0x21dcU,0x43b8U,0x8770U,0x1eebU,0x3dd6U,0x7bacU,0xf758U,0xfebbU,0xed7dU,
		0xcaf1U,0x85e9U,0x1bd9U,0x37b2U,0x6f64U,0xdec8U,0xad9bU,0x4b3dU,0x967aU,0x3cffU,
		0x79feU,0xf3fcU,0xf7f3U,0xffedU,0xefd1U,0xcfa9U,0x8f59U,0x0eb9U,0x1d72U,0x3ae4U,
		0x75c8U,0xeb90U,0xc72bU,0x9e5dU,0x2cb1U,0x5962U,0xb2c4U,0x7583U,0xeb06U,0xc607U,
		0x9c05U,0x2801U,0x5002U,0xa004U,0x5003U,0xa006U,0x5007U,0xa00eU,0x5017U,0xa02eU,
		0x5057U,0xa0aeU,0x5157U,0xa2aeU,0x5557U,0xaaaeU,0x4557U,0x8aaeU,0x0557U,0x0aaeU,
		0x155cU,0x2ab8U,0x5570U,0xaae0U,0x45cbU,0x8b96U,0x0727U,0x0e4eU,0x1c9cU,0x3938U,
		0x7270U,0xe4e0U,0xd9cbU,0xa39dU,0x5731U,0xae62U,0x4ccfU,0x999eU,0x2337U,0x466eU,
		0x8cdcU,0x09b3U,0x1366U,0x26ccU,0x4d98U,0x9b30U,0x266bU,0x4cd6U,0x99acU,0x2353U,
		0x46a6U,0x8d4cU,0x0a93U,0x1526U,0x2a4cU,0x5498U,0xa930U,0x426bU,0x84d6U,0x19a7U,
		0x334eU,0x669cU,0xcd38U,0x8a7bU,0x04fdU,0x09faU,0x13f4U,0x27e8U,0x4fd0U,0x9fa0U,
		0x2f4bU,0x5e96U,0xbd2cU,0x6a53U,0xd4a6U,0xb947U,0x6285U,0xc50aU,0x9a1fU,0x2435U,
		0x486aU,0x90d4U,0x31a3U,0x6346U,0xc68cU,0x9d13U,0x2a2dU,0x545aU,0xa8b4U,0x4163U,
		0x82c6U,0x1587U,0x2b0eU,0x561cU,0xac38U,0x487bU,0x90f6U,0x31e7U,0x63ceU,0xc79cU,
		0x9f33U,0x2e6dU,0x5cdaU,0xb9b4U,0x6363U,0xc6c6U,0x9d87U,0x2b05U,0x560aU,0xac14U,
		0x4823U,0x9046U,0x3087U,0x610eU,0xc21cU,0x9433U,0x386dU,0x70daU,0xe1b4U,0xd363U,
		0xb6cdU,0x7d91U,0xfb22U,0xe64fU,0xdc95U,0xa921U,0x4249U,0x8492U,0x192fU,0x325eU,
		0x64bcU,0xc978U,0x82fbU,0x15fdU,0x2bfaU,0x57f4U,0xafe8U,0x4fdbU,0x9fb6U,0x2f67U,
		0x5eceU,0xbd9cU,0x6b33U,0xd666U,0xbcc7U,0x6985U,0xd30aU,0xb61fU,0x7c35U,0xf86aU,
		0xe0dfU,0xd1b5U,0xb361U,0x76c9U,0xed92U,0xcb2fU,0x8655U,0x1ca1U,0x3942U,0x7284U,
		0xe508U,0xda1bU,0xa43dU,0x5871U,0xb0e2U,0x71cfU,0xe39eU,0xd737U,0xbe65U,0x6cc1U,
		0xd982U,0xa30fU,0x5615U,0xac2aU,0x485fU,0x90beU,0x3177U,0x62eeU,0xc5dcU,0x9bb3U,
		0x276dU,0x4edaU,0x9db4U,0x2b63U,0x56c6U,0xad8cU,0x4b13U,0x9626U,0x3c47U,0x788eU,
		0xf11cU,0xf233U,0xf46dU,0xf8d1U,0xe1a9U,0xd359U,0xb6b9U,0x7d79U,0xfaf2U,0xe5efU,
		0xdbd5U,0xa7a1U,0x5f49U,0xbe92U,0x6d2fU,0xda5eU,0xa4b7U,0x5965U,0xb2caU,0x759fU,
		0xeb3eU,0xc677U,0x9ce5U,0x29c1U,0x5382U,0xa704U,0x5e03U,0xbc06U,0x6807U,0xd00eU,
		0xb017U,0x7025U,0xe04aU,0xd09fU,0xb135U,0x7261U,0xe4c2U,0xd98fU,0xa315U,0x5621U,
		0xac42U,0x488fU,0x911eU,0x3237U,0x646eU,0xc8dcU,0x81b3U,0x136dU,0x26daU,0x4db4U,
		0x9b68U,0x26dbU,0x4db6U,0x9b6cU,0x26d3U,0x4da6U,0x9b4cU,0x2693U,0x4d26U,0x9a4cU,
		0x2493U,0x4926U,0x924cU,0x3493U,0x6926U,0xd24cU,0xb493U,0x792dU,0xf25aU,0xf4bfU,
		0xf975U,0xe2e1U,0xd5c9U,0xbb99U,0x6739U,0xce72U,0x8cefU,0x09d5U,0x13aaU,0x2754U,
		0x4ea8U,0x9d50U,0x2aabU,0x5556U,0xaaacU,0x4553U,0x8aa6U,0x0547U,0x0a8eU,0x151cU,
		0x2a38U,0x5470U,0xa8e0U,0x41cbU,0x8396U,0x1727U,0x2e4eU,0x5c9cU,0xb938U,0x627bU,
		0xc4f6U,0x99e7U,0x23c5U,0x478aU,0x8f14U,0x0e23U,0x1c46U,0x388cU,0x7118U,0xe230U,
		0xd46bU,0xb8ddU,0x61b1U,0xc362U,0x96cfU,0x3d95U,0x7b2aU,0xf654U,0xfca3U,0xe94dU,
		0xc291U,0x9529U,0x3a59U,0x74b2U,0xe964U,0xc2c3U,0x958dU,0x3b11U,0x7622U,0xec44U,
		0xc883U,0x810dU,0x1211U,0x2422U,0x4844U,0x9088U,0x311bU,0x6236U,0xc46cU,0x98d3U,
		0x21adU,0x435aU,0x86b4U,0x1d63U,0x3ac6U,0x758cU,0xeb18U,0xc63bU,0x9c7dU,0x28f1U,
		0x51e2U,0xa3c4U,0x5783U,0xaf06U,0x4e07U,0x9c0eU,0x2817U,0x502eU,0xa05cU,0x50b3U,
		0xa166U,0x52c7U,0xa58eU,0x5b17U,0xb62eU,0x7c57U,0xf8aeU,0xe157U,0xd2a5U,0xb541U,
		0x7a89U,0xf512U,0xfa2fU,0xe455U,0xd8a1U,0xa149U,0x5299U,0xa532U,0x5a6fU,0xb4deU,
		0x79b7U,0xf36eU,0xf6d7U,0xfda5U,0xeb41U,0xc689U,0x9d19U,0x2a39U,0x5472U,0xa8e4U,
		0x41c3U,0x8386U,0x1707U,0x2e0eU,0x5c1cU,0xb838U,0x607bU,0xc0f6U,0x91e7U,0x33c5U,
		0x678aU,0xcf14U,0x8e23U,0x0c4dU,0x189aU,0x3134U,0x6268U,0xc4d0U,0x99abU,0x235dU,
		0x46baU,0x8d74U,0x0ae3U,0x15c6U,0x2b8cU,0x5718U,0xae30U,0x4c6bU,0x98d6U,0x21a7U,
		0x434eU,0x869cU,0x1d33U,0x3a66U,0x74ccU,0xe998U,0xc33bU,0x967dU,0x3cf1U,0x79e2U,
		0xf3c4U,0xf783U,0xff0dU,0xee11U,0xcc29U,0x8859U,0x00b9U,0x0172U,0x02e4U,0x05c8U,
		0x0b90U,0x1720U,0x2e40U,0x5c80U,0xb900U,0x620bU,0xc416U,0x9827U,0x2045U,0x408aU,
		0x8114U,0x1223U,0x2446U,0x488cU,0x9118U,0x323bU,0x6476U,0xc8ecU,0x81d3U,0x13adU,
		0x275aU,0x4eb4U,0x9d68U,0x2adbU,0x55b6U,0xab6cU,0x46d3U,0x8da6U,0x0b47U,0x168eU,
		0x2d1cU,0x5a38U,0xb470U,0x78ebU,0xf1d6U,0xf3a7U,0xf745U,0xfe81U,0xed09U,0xca19U,
		0x8439U,0x1879U,0x30f2U,0x61e4U,0xc3c8U,0x979bU,0x3f3dU,0x7e7aU,0xfcf4U,0xe9e3U,
		0xc3cdU,0x9791U,0x3f29U,0x7e52U,0xfca4U,0xe943U,0xc28dU,0x9511U,0x3a29U,0x7452U,
		0xe8a4U,0xc143U,0x928dU,0x3511U,0x6a22U,0xd444U,0xb883U,0x610dU,0xc21aU,0x943fU,
		0x3875U,0x70eaU,0xe1d4U,0xd3a3U,0xb74dU,0x7e91U,0xfd22U,0xea4fU,0xc495U,0x9921U,
		0x2249U,0x4492U,0x8924U,0x0243U,0x0486U,0x090cU,0x1218U,0x2430U,0x4860U,0x90c0U,
		0x318bU,0x6316U,0xc62cU,0x9c53U,0x28adU,0x515aU,0xa2b4U,0x5563U,0xaac6U,0x4587U,
		0x8b0eU,0x0617U,0x0c2eU,0x185cU,0x30b8U,0x6170U,0xc2e0U,0x95cbU,0x3b9dU,0x773aU,
		0xee74U,0xcce3U,0x89cdU,0x0391U,0x0722U,0x0e44U,0x1c88U,0x3910U,0x7220U,0xe440U,
		0xd88bU,0xa11dU,0x5231U,0xa462U,0x58cfU,0xb19eU,0x7337U,0xe66eU,0xdcd7U,0xa9a5U,
		0x4341U,0x8682U,0x1d0fU,0x3a1eU,0x743cU,0xe878U,0xc0fbU,0x91fdU,0x33f1U,0x67e2U,
		0xcfc4U,0x8f83U,0x0f0dU,0x1e1aU,0x3c34U,0x7868U,0xf0d0U,0xf1abU,0xf35dU,0xf6b1U,
		0xfd69U,0xead9U,0xc5b9U,0x9b79U,0x26f9U,0x4df2U,0x9be4U,0x27c3U,0x4f86U,0x9f0cU,
		0x2e13U,0x5c26U,0xb84cU,0x6093U,0xc126U,0x9247U,0x3485U,0x690aU,0xd214U,0xb423U,
		0x784dU,0xf09aU,0xf13fU,0xf275U,0xf4e1U,0xf9c9U,0xe399U,0xd739U,0xbe79U,0x6cf9U,
		0xd9f2U,0xa3efU,0x57d5U,0xafaaU,0x4f5fU,0x9ebeU,0x2d77U,0x5aeeU,0xb5dcU,0x7bb3U,
		0xf766U,0xfec7U,0xed85U,0xcb01U,0x8609U,0x1c19U,0x3832U,0x7064U,0xe0c8U,0xd19bU,
		0xb33dU,0x7671U,0xece2U,0xc9cfU,0x8395U,0x1721U,0x2e42U,0x5c84U,0xb908U,0x621bU,
		0xc436U,0x9867U,0x20c5U,0x418aU,0x8314U,0x1623U,0x2c46U,0x588cU,0xb118U,0x723bU,
		0xe476U,0xd8e7U,0xa1c5U,0x5381U,0xa702U,0x5e0fU,0xbc1eU,0x6837U,0xd06eU,0xb0d7U,
		0x71a5U,0xe34aU,0xd69fU,0xbd35U,0x6a61U,0xd4c2U,0xb98fU,0x6315U,0xc62aU,0x9c5fU,
		0x28b5U,0x516aU,0xa2d4U,0x55a3U,0xab46U,0x4687U,0x8d0eU,0x0a17U,0x142eU,0x285cU,
		0x50b8U,0xa170U,0x52ebU,0xa5d6U,0x5ba7U,0xb74eU,0x7e97U,0xfd2eU,0xea57U,0xc4a5U,
		0x9941U,0x2289U,0x4512U,0x8a24U,0x0443U,0x0886U,0x110cU,0x2218U,0x4430U,0x8860U,
		0x00cbU,0x0196U,0x032cU,0x0658U,0x0cb0U,0x1960U,0x32c0U,0x6580U,0xcb00U,0x860bU,
		0x1c1dU,0x383aU,0x7074U,0xe0e8U,0xd1dbU,0xb3bdU,0x7771U,0xeee2U,0xcdcfU,0x8b95U,
		0x0721U,0x0e42U,0x1c84U,0x3908U,0x7210U,0xe420U,0xd84bU,0xa09dU,0x5131U,0xa262U,
		0x54cfU,0xa99eU,0x4337U,0x866eU,0x1cd7U,0x39aeU,0x735cU,0xe6b8U,0xdd7bU,0xaafdU,
		0x45f1U,0x8be2U,0x07cfU,0x0f9eU,0x1f3cU,0x3e78U,0x7cf0U,0xf9e0U,0xe3cbU,0xd79dU,
		0xbf31U,0x6e69U,0xdcd2U,0xa9afU,0x4355U,0x86aaU,0x1d5fU,0x3abeU,0x757cU,0xeaf8U,
		0xc5fbU,0x9bfdU,0x27f1U,0x4fe2U,0x9fc4U,0x2f83U,0x5f06U,0xbe0cU,0x6c13U,0xd826U,
		0xa047U,0x5085U,0xa10aU,0x521fU,0xa43eU,0x5877U,0xb0eeU,0x71d7U,0xe3aeU,0xd757U,
		0xbea5U,0x6d41U,0xda82U,0xa50fU,0x5a15U,0xb42aU,0x785fU,0xf0beU,0xf177U,0xf2e5U,
		0xf5c1U,0xfb89U,0xe719U,0xde39U,0xac79U,0x48f9U,0x91f2U,0x33efU,0x67deU,0xcfbcU,
		0x8f73U,0x0eedU,0x1ddaU,0x3bb4U,0x7768U,0xeed0U,0xcdabU,0x8b5dU,0x06b1U,0x0d62U,
		0x1ac4U,0x3588U,0x6b10U,0xd620U,0xbc4bU,0x689dU,0xd13aU,0xb27fU,0x74f5U,0xe9eaU,
		0xc3dfU,0x97b5U,0x3f61U,0x7ec2U,0xfd84U,0xeb03U,0xc60dU,0x9c11U,0x2829U,0x5052U,
		0xa0a4U,0x5143U,0xa286U,0x5507U,0xaa0eU,0x4417U,0x882eU,0x0057U,0x00aeU,0x015cU,
		0x02b8U,0x0570U,0x0ae0U,0x15c0U,0x2b80U,0x5700U,0xae00U,0x4c0bU,0x9816U,0x2027U,
		0x404eU,0x809cU,0x1133U,0x2266U,0x44ccU,0x8998U,0x033bU,0x0676U,0x0cecU,0x19d8U,
		0x33b0U,0x6760U,0xcec0U,0x8d8bU,0x0b1dU,0x163aU,0x2c74U,0x58e8U,0xb1d0U,0x73abU,
		0xe756U,0xdea7U,0xad45U,0x4a81U,0x9502U,0x3a0fU,0x741eU,0xe83cU,0xc073U,0x90edU,
		0x31d1U,0x63a2U,0xc744U,0x9e83U,0x2d0dU,0x5a1aU,0xb434U,0x7863U,0xf0c6U,0xf187U,
		0xf305U,0xf601U,0xfc09U,0xe819U,0xc039U,0x9079U,0x30f9U,0x61f2U,0xc3e4U,0x97c3U,
		0x3f8dU,0x7f1aU,0xfe34U,0xec63U,0xc8cdU,0x8191U,0x1329U,0x2652U,0x4ca4U,0x9948U,
		0x229bU,0x4536U,0x8a6cU,0x04d3U,0x09a6U,0x134cU,0x2698U,0x4d30U,0x9a60U,0x24cbU,
		0x4996U,0x932cU,0x3653U,0x6ca6U,0xd94cU,0xa293U,0x552dU,0xaa5aU,0x44bfU,0x897eU,
		0x02f7U,0x05eeU,0x0bdcU,0x17b8U,0x2f70U,0x5ee0U,0xbdc0U,0x6b8bU,0xd716U,0xbe27U,
		0x6c45U,0xd88aU,0xa11fU,0x5235U,0xa46aU,0x58dfU,0xb1beU,0x7377U,0xe6eeU,0xddd7U,
		0xaba5U,0x4741U,0x8e82U,0x0d0fU,0x1a1eU,0x343cU,0x6878U,0xd0f0U,0xb1ebU,0x73ddU,
		0xe7baU,0xdf7fU,0xaef5U,0x4de1U,0x9bc2U,0x278fU,0x4f1eU,0x9e3cU,0x2c73U,0x58e6U,
		0xb1ccU,0x7393U,0xe726U,0xde47U,0xac85U,0x4901U,0x9202U,0x340fU,0x681eU,0xd03cU,
		0xb073U,0x70edU,0xe1daU,0xd3bfU,0xb775U,0x7ee1U,0xfdc2U,0xeb8fU,0xc715U,0x9e21U,
		0x2c49U,0x5892U,0xb124U,0x7243U,0xe486U,0xd907U,0xa205U,0x5401U,0xa802U,0x400fU,
		0x801eU,0x1037U,0x206eU,0x40dcU,0x81b8U,0x137bU,0x26f6U,0x4decU,0x9bd8U,0x27bbU,
		0x4f76U,0x9eecU,0x2dd3U,0x5ba6U,0xb74cU,0x7e93U,0xfd26U,0xea47U,0xc485U,0x9901U,
		0x2209U,0x4412U,0x8824U,0x0043U,0x0086U,0x010cU,0x0218U,0x0430U,0x0860U,0x10c0U,
		0x2180U,0x4300U,0x8600U,0x1c0bU,0x3816U,0x702cU,0xe058U,0xd0bbU,0xb17dU,0x72f1U,
		0xe5e2U,0xdbcfU,0xa795U,0x5f21U,0xbe42U,0x6c8fU,0xd91eU,0xa237U,0x5465U,0xa8caU,
		0x419fU,0x833eU,0x1677U,0x2ceeU,0x59dcU,0xb3b8U,0x777bU,0xeef6U,0xcde7U,0x8bc5U,
		0x0781U,0x0f02U,0x1e04U,0x3c08U,0x7810U,0xf020U,0xf04bU,0xf09dU,0xf131U,0xf269U,
		0xf4d9U,0xf9b9U,0xe379U,0xd6f9U,0xbdf9U,0x6bf9U,0xd7f2U,0xbfefU,0x6fd5U,0xdfaaU,
		0xaf5fU,0x4eb5U,0x9d6aU,0x2adfU,0x55beU,0xab7cU,0x46f3U,0x8de6U,0x0bc7U,0x178eU,
		0x2f1cU,0x5e38U,0xbc70U,0x68ebU,0xd1d6U,0xb3a7U,0x7745U,0xee8aU,0xcd1fU,0x8a35U,
		0x0461U,0x08c2U,0x1184U,0x2308U,0x4610U,0x8c20U,0x084bU,0x1096U,0x212cU,0x4258U,
		0x84b0U,0x196bU,0x32d6U,0x65acU,0xcb58U,0x86bbU,0x1d7dU,0x3afaU,0x75f4U,0xebe8U,
		0xc7dbU,0x9fbdU,0x2f71U,0x5ee2U,0xbdc4U,0x6b83U,0xd706U,0xbe07U,0x6c05U,0xd80aU,
		0xa01fU,0x5035U,0xa06aU,0x50dfU,0xa1beU,0x5377U,0xa6eeU,0x5dd7U,0xbbaeU,0x6757U,
		0xceaeU,0x8d57U,0x0aa5U,0x154aU,0x2a94U,0x5528U,0xaa50U,0x44abU,0x8956U,0x02a7U,
		0x054eU,0x0a9cU,0x1538U,0x2a70U,0x54e0U,0xa9c0U,0x438bU,0x8716U,0x1e27U,0x3c4eU,
		0x789cU,0xf138U,0xf27bU,0xf4fdU,0xf9f1U,0xe3e9U,0xd7d9U,0xbfb9U,0x6f79U,0xdef2U,
		0xadefU,0x4bd5U,0x97aaU,0x3f5fU,0x7ebeU,0xfd7cU,0xeaf3U,0xc5edU,0x9bd1U,0x27a9U,
		0x4f52U,0x9ea4U,0x2d43U,0x5a86U,0xb50cU,0x7a13U,0xf426U,0xf847U,0xe085U,0xd101U,
		0xb209U,0x7419U,0xe832U,0xc06fU,0x90d5U,0x31a1U,0x6342U,0xc684U,0x9d03U,0x2a0dU,
		0x541aU,0xa834U,0x4063U,0x80c6U,0x1187U,0x230eU,0x461cU,0x8c38U,0x087bU,0x10f6U,
		0x21ecU,0x43d8U,0x87b0U,0x1f6bU,0x3ed6U,0x7dacU,0xfb58U,0xe6bbU,0xdd7dU,0xaaf1U,
		0x45e9U,0x8bd2U,0x07afU,0x0f5eU,0x1ebcU,0x3d78U,0x7af0U,0xf5e0U,0xfbcbU,0xe79dU,
		0xdf31U,0xae69U,0x4cd9U,0x99b2U,0x236fU,0x46deU,0x8dbcU,0x0b73U,0x16e6U,0x2dccU,
		0x5b98U,0xb730U,0x7e6bU,0xfcd6U,0xe9a7U,0xc345U,0x9681U,0x3d09U,0x7a12U,0xf424U,
		0xf843U,0xe08dU,0xd111U,0xb229U,0x7459U,0xe8b2U,0xc16fU,0x92d5U,0x35a1U,0x6b42U,
		0xd684U,0xbd03U,0x6a0dU,0xd41aU,0xb83fU,0x6075U,0xc0eaU,0x91dfU,0x33b5U,0x676aU,
		0xced4U,0x8da3U,0x0b4dU,0x169aU,0x2d34U,0x5a68U,0xb4d0U,0x79abU,0xf356U,0xf6a7U,
		0xfd45U,0xea81U,0xc509U,0x9a19U,0x2439U,0x4872U,0x90e4U,0x31c3U,0x6386U,0xc70cU,
		0x9e13U,0x2c2dU,0x585aU,0xb0b4U,0x7163U,0xe2c6U,0xd587U,0xbb05U,0x6601U,0xcc02U,
		0x880fU,0x0015U,0x002aU,0x0054U,0x00a8U,0x0150U,0x02a0U,0x0540U,0x0a80U,0x1500U,
		0x2a00U,0x5400U,0xa800U,0x400bU,0x8016U,0x1027U,0x204eU,0x409cU,0x8138U,0x127bU,
		0x24f6U,0x49ecU,0x93d8U,0x37bbU,0x6f76U,0xdeecU,0xadd3U,0x4badU,0x975aU,0x3ebfU,
		0x7d7eU,0xfafcU,0xe5f3U,0xdbedU,0xa7d1U,0x5fa9U,0xbf52U,0x6eafU,0xdd5eU,0xaab7U,
		0x4565U,0x8acaU,0x059fU,0x0b3eU,0x167cU,0x2cf8U,0x59f0U,0xb3e0U,0x77cbU,0xef96U,
		0xcf27U,0x8e45U,0x0c81U,0x1902U,0x3204U,0x6408U,0xc810U,0x802bU,0x105dU,0x20baU,
		0x4174U,0x82e8U,0x15dbU,0x2bb6U,0x576cU,0xaed8U,0x4dbbU,0x9b76U,0x26e7U,0x4dceU,
		0x9b9cU,0x2733U,0x4e66U,0x9cccU,0x2993U,0x5326U,0xa64cU,0x5c93U,0xb926U,0x6247U,
		0xc48eU,0x9917U,0x2225U,0x444aU,0x8894U,0x0123U,0x0246U,0x048cU,0x0918U,0x1230U,
		0x2460U,0x48c0U,0x9180U,0x330bU,0x6616U,0xcc2cU,0x8853U,0x00adU,0x015aU,0x02b4U,
		0x0568U,0x0ad0U,0x15a0U,0x2b40U,0x5680U,0xad00U,0x4a0bU,0x9416U,0x3827U,0x704eU,
		0xe09cU,0xd133U,0xb26dU,0x74d1U,0xe9a2U,0xc34fU,0x9695U,0x3d21U,0x7a42U,0xf484U,
		0xf903U,0xe20dU,0xd411U,0xb829U,0x6059U,0xc0b2U,0x916fU,0x32d5U,0x65aaU,0xcb54U,
		0x86a3U,0x1d4dU,0x3a9aU,0x7534U,0xea68U,0xc4dbU,0x99bdU,0x2371U,0x46e2U,0x8dc4U,
		0x0b83U,0x1706U,0x2e0cU,0x5c18U,0xb830U,0x606bU,0xc0d6U,0x91a7U,0x3345U,0x668aU,
		0xcd14U,0x8a23U,0x044dU,0x089aU,0x1134U,0x2268U,0x44d0U,0x89a0U,0x034bU,0x0696U,
		0x0d2cU,0x1a58U,0x34b0U,0x6960U,0xd2c0U,0xb58bU,0x7b1dU,0xf63aU,0xfc7fU,0xe8f5U,
		0xc1e1U,0x93c9U,0x3799U,0x6f32U,0xde64U,0xacc3U,0x498dU,0x931aU,0x363fU,0x6c7eU,
		0xd8fcU,0xa1f3U,0x53edU,0xa7daU,0x5fbfU,0xbf7eU,0x6ef7U,0xddeeU,0xabd7U,0x47a5U,
		0x8f4aU,0x0e9fU,0x1d3eU,0x3a7cU,0x74f8U,0xe9f0U,0xc3ebU,0x97ddU,0x3fb1U,0x7f62U,
		0xfec4U,0xed83U,0xcb0dU,0x8611U,0x1c29U,0x3852U,0x70a4U,0xe148U,0xd29bU,0xb53dU,
		0x7a71U,0xf4e2U,0xf9cfU,0xe395U,0xd721U,0xbe49U,0x6c99U,0xd932U,0xa26fU,0x54d5U,
		0xa9aaU,0x435fU,0x86beU,0x1d77U,0x3aeeU,0x75dcU,0xebb8U,0xc77bU,0x9efdU,0x2df1U,
		0x5be2U,0xb7c4U,0x7f83U,0xff06U,0xee07U,0xcc05U,0x8801U,0x0009U,0x0012U,0x0024U,
		0x0048U,0x0090U,0x0120U,0x0240U,0x0480U,0x0900U,0x1200U,0x2400U,0x4800U,0x9000U,
		0x300bU,0x6016U,0xc02cU,0x9053U,0x30adU,0x615aU,0xc2b4U,0x9563U,0x3acdU,0x759aU,
		0xeb34U,0xc663U,0x9ccdU,0x2991U,0x5322U,0xa644U,0x5c83U,0xb906U,0x6207U,0xc40eU,
		0x9817U,0x2025U,0x404aU,0x8094U,0x1123U,0x2246U,0x448cU,0x8918U,0x023bU,0x0476U,
		0x08ecU,0x11d8U,0x23b0U,0x4760U,0x8ec0U,0x0d8bU,0x1b16U,0x362cU,0x6c58U,0xd8b0U,
		0xa16bU,0x52ddU,0xa5baU,0x5b7fU,0xb6feU,0x7df7U,0xfbeeU,0xe7d7U,0xdfa5U,0xaf41U,
		0x4e89U,0x9d12U,0x2a2fU,0x545eU,0xa8bcU,0x4173U,0x82e6U,0x15c7U,0x2b8eU,0x571cU,
		0xae38U,0x4c7bU,0x98f6U,0x21e7U,0x43ceU,0x879cU,0x1f33U,0x3e66U,0x7cccU,0xf998U,
		0xe33bU,0xd67dU,0xbcf1U,0x69e9U,0xd3d2U,0xb7afU,0x7f55U,0xfeaaU,0xed5fU,0xcab5U,
		0x8561U,0x1ac9U,0x3592U,0x6b24U,0xd648U,0xbc9bU,0x693dU,0xd27aU,0xb4ffU,0x79f5U,
		0xf3eaU,0xf7dfU,0xffb5U,0xef61U,0xcec9U,0x8d99U,0x0b39U,0x1672U,0x2ce4U,0x59c8U,
		0xb390U,0x772bU,0xee56U,0xcca7U,0x8945U,0x0281U,0x0502U,0x0a04U,0x1408U,0x2810U,
		0x5020U,0xa040U,0x508bU,0xa116U,0x5227U,0xa44eU,0x5897U,0xb12eU,0x7257U,0xe4aeU,
		0xd957U,0xa2a5U,0x5541U,0xaa82U,0x450fU,0x8a1eU,0x0437U,0x086eU,0x10dcU,0x21b8U,
		0x4370U,0x86e0U,0x1dcbU,0x3b96U,0x772cU,0xee58U,0xccbbU,0x897dU,0x02f1U,0x05e2U,
		0x0bc4U,0x1788U,0x2f10U,0x5e20U,0xbc40U,0x688bU,0xd116U,0xb227U,0x7445U,0xe88aU,
		0xc11fU,0x9235U,0x3461U,0x68c2U,0xd184U,0xb303U,0x760dU,0xec1aU,0xc83fU,0x8075U,
		0x10e1U,0x21c2U,0x4384U,0x8708U,0x1e1bU,0x3c36U,0x786cU,0xf0d8U,0xf1bbU,0xf37dU,
		0xf6f1U,0xfde9U,0xebd9U,0xc7b9U,0x9f79U,0x2ef9U,0x5df2U,0xbbe4U,0x67c3U,0xcf86U,
		0x8f07U,0x0e05U,0x1c0aU,0x3814U,0x7028U,0xe050U,0xd0abU,0xb15dU,0x72b1U,0xe562U,
		0xdacfU,0xa595U,0x5b21U,0xb642U,0x7c8fU,0xf91eU,0xe237U,0xd465U,0xb8c1U,0x6189U,
		0xc312U,0x962fU,0x3c55U,0x78aaU,0xf154U,0xf2a3U,0xf54dU,0xfa91U,0xe529U,0xda59U,
		0xa4b9U,0x5979U,0xb2f2U,0x75efU,0xebdeU,0xc7b7U,0x9f65U,0x2ec1U,0x5d82U,0xbb04U,
		0x6603U,0xcc06U,0x8807U,0x0005U,0x000aU,0x0014U,0x0028U,0x0050U,0x00a0U,0x0140U,
		0x0280U,0x0500U,0x0a00U,0x1400U,0x2800U,0x5000U,0xa000U,0x500bU,0xa016U,0x5027U,
		0xa04eU,0x5097U,0xa12eU,0x5257U,0xa4aeU,0x5957U,0xb2aeU,0x7557U,0xeaaeU,0xc557U,
		0x9aa5U,0x2541U,0x4a82U,0x9504U,0x3a03U,0x7406U,0xe80cU,0xc013U,0x902dU,0x3051U,
		0x60a2U,0xc144U,0x9283U,0x350dU,0x6a1aU,0xd434U,0xb863U,0x60cdU,0xc19aU,0x933fU,
		0x3675U,0x6ceaU,0xd9d4U,0xa3a3U,0x574dU,0xae9aU,0x4d3fU,0x9a7eU,0x24f7U,0x49eeU,
		0x93dcU,0x37b3U,0x6f66U,0xdeccU,0xad93U,0x4b2dU,0x965aU,0x3cbfU,0x797eU,0xf2fcU,
		0xf5f3U,0xfbedU,0xe7d1U,0xdfa9U,0xaf59U,0x4eb9U,0x9d72U,0x2aefU,0x55deU,0xabbcU,
		0x4773U,0x8ee6U,0x0dc7U,0x1b8eU,0x371cU,0x6e38U,0xdc70U,0xa8ebU,0x41ddU,0x83baU,
		0x177fU,0x2efeU,0x5dfcU,0xbbf8U,0x67fbU,0xcff6U,0x8fe7U,0x0fc5U,0x1f8aU,0x3f14U,
		0x7e28U,0xfc50U,0xe8abU,0xc15dU,0x92b1U,0x3569U,0x6ad2U,0xd5a4U,0xbb43U,0x668dU,
		0xcd1aU,0x8a3fU,0x0475U,0x08eaU,0x11d4U,0x23a8U,0x4750U,0x8ea0U,0x0d4bU,0x1a96U,
		0x352cU,0x6a58U,0xd4b0U,0xb96bU,0x62ddU,0xc5baU,0x9b7fU,0x26f5U,0x4deaU,0x9bd4U,
		0x27a3U,0x4f46U,0x9e8cU,0x2d13U,0x5a26U,0xb44cU,0x7893U,0xf126U,0xf247U,0xf485U,
		0xf901U,0xe209U,0xd419U,0xb839U,0x6079U,0xc0f2U,0x91efU,0x33d5U,0x67aaU,0xcf54U,
		0x8ea3U,0x0d4dU,0x1a9aU,0x3534U,0x6a68U,0xd4d0U,0xb9abU,0x635dU,0xc6baU,0x9d7fU,
		0x2af5U,0x55eaU,0xabd4U,0x47a3U,0x8f46U,0x0e87U,0x1d0eU,0x3a1cU,0x7438U,0xe870U,
		0xc0ebU,0x91ddU,0x33b1U,0x6762U,0xcec4U,0x8d83U,0x0b0dU,0x161aU,0x2c34U,0x5868U,
		0xb0d0U,0x71abU,0xe356U,0xd6a7U,0xbd45U,0x6a81U,0xd502U,0xba0fU,0x6415U,0xc82aU,
		0x805fU,0x10b5U,0x216aU,0x42d4U,0x85a8U,0x1b5bU,0x36b6U,0x6d6cU,0xdad8U,0xa5bbU,
		0x5b7dU,0xb6faU,0x7dffU,0xfbfeU,0xe7f7U,0xdfe5U,0xafc1U,0x4f89U,0x9f12U,0x2e2fU,
		0x5c5eU,0xb8bcU,0x6173U,0xc2e6U,0x95c7U,0x3b85U,0x770aU,0xee14U,0xcc23U,0x884dU,
		0x0091U,0x0122U,0x0244U,0x0488U,0x0910U,0x1220U,0x2440U,0x4880U,0x9100U,0x320bU,
		0x6416U,0xc82cU,0x8053U,0x10adU,0x215aU,0x42b4U,0x8568U,0x1adbU,0x35b6U,0x6b6cU,
		0xd6d8U,0xbdbbU,0x6b7dU,0xd6faU,0xbdffU,0x6bf5U,0xd7eaU,0xbfdfU,0x6fb5U,0xdf6aU,
		0xaedfU,0x4db5U,0x9b6aU,0x26dfU,0x4dbeU,0x9b7cU,0x26f3U,0x4de6U,0x9bccU,0x2793U,
		0x4f26U,0x9e4cU,0x2c93U,0x5926U,0xb24cU,0x7493U,0xe926U,0xc247U,0x9485U,0x3901U,
		0x7202U,0xe404U,0xd803U,0xa00dU,0x5011U,0xa022U,0x504fU,0xa09eU,0x5137U,0xa26eU,
		0x54d7U,0xa9aeU,0x4357U,0x86aeU,0x1d57U,0x3aaeU,0x755cU,0xeab8U,0xc57bU,0x9afdU,
		0x25f1U,0x4be2U,0x97c4U,0x3f83U,0x7f06U,0xfe0cU,0xec13U,0xc82dU,0x8051U,0x10a9U,
		0x2152U,0x42a4U,0x8548U,0x1a9bU,0x3536U,0x6a6cU,0xd4d8U,0xb9bbU,0x637dU,0xc6faU,
		0x9dffU,0x2bf5U,0x57eaU,0xafd4U,0x4fa3U,0x9f46U,0x2e87U,0x5d0eU,0xba1cU,0x6433U,
		0xc866U,0x80c7U,0x1185U,0x230aU,0x4614U,0x8c28U,0x085bU,0x10b6U,0x216cU,0x42d8U,
		0x85b0U,0x1b6bU,0x36d6U,0x6dacU,0xdb58U,0xa6bbU,0x5d7dU,0xbafaU,0x65ffU,0xcbfeU,
		0x87f7U,0x1fe5U,0x3fcaU,0x7f94U,0xff28U,0xee5bU,0xccbdU,0x8971U,0x02e9U,0x05d2U,
		0x0ba4U,0x1748U,0x2e90U,0x5d20U,0xba40U,0x648bU,0xc916U,0x8227U,0x1445U,0x288aU,
		0x5114U,0xa228U,0x545bU,0xa8b6U,0x4167U,0x82ceU,0x1597U,0x2b2eU,0x565cU,0xacb8U,
		0x497bU,0x92f6U,0x35e7U,0x6bceU,0xd79cU,0xbf33U,0x6e6dU,0xdcdaU,0xa9bfU,0x4375U,
		0x86eaU,0x1ddfU,0x3bbeU,0x777cU,0xeef8U,0xcdfbU,0x8bfdU,0x07f1U,0x0fe2U,0x1fc4U,
		0x3f88U,0x7f10U,0xfe20U,0xec4bU,0xc89dU,0x8131U,0x1269U,0x24d2U,0x49a4U,0x9348U,
		0x369bU,0x6d36U,0xda6cU,0xa4d3U,0x59adU,0xb35aU,0x76bfU,0xed7eU,0xcaf7U,0x85e5U,
		0x1bc1U,0x3782U,0x6f04U,0xde08U,0xac1bU,0x483dU,0x907aU,0x30ffU,0x61feU,0xc3fcU,
		0x97f3U,0x3fedU,0x7fdaU,0xffb4U,0xef63U,0xcecdU,0x8d91U,0x0b29U,0x1652U,0x2ca4U,
		0x5948U,0xb290U,0x752bU,0xea56U,0xc4a7U,0x9945U,0x2281U,0x4502U,0x8a04U,0x0403U,
		0x0806U,0x100cU,0x2018U,0x4030U,0x8060U,0x10cbU,0x2196U,0x432cU,0x8658U,0x1cbbU,
		0x3976U,0x72ecU,0xe5d8U,0xdbbbU,0xa77dU,0x5ef1U,0xbde2U,0x6bcfU,0xd79eU,0xbf37U,
		0x6e65U,0xdccaU,0xa99fU,0x4335U,0x866aU,0x1cdfU,0x39beU,0x737cU,0xe6f8U,0xddfbU,
		0xabfdU,0x47f1U,0x8fe2U,0x0fcfU,0x1f9eU,0x3f3cU,0x7e78U,0xfcf0U,0xe9ebU,0xc3ddU,
		0x97b1U,0x3f69U,0x7ed2U,0xfda4U,0xeb43U,0xc68dU,0x9d11U,0x2a29U,0x5452U,0xa8a4U,
		0x4143U,0x8286U,0x1507U,0x2a0eU,0x541cU,0xa838U,0x407bU,0x80f6U,0x11e7U,0x23ceU,
		0x479cU,0x8f38U,0x0e7bU,0x1cf6U,0x39ecU,0x73d8U,0xe7b0U,0xdf6bU,0xaeddU,0x4db1U,
		0x9b62U,0x26cfU,0x4d9eU,0x9b3cU,0x2673U,0x4ce6U,0x99ccU,0x2393U,0x4726U,0x8e4cU,
		0x0c93U,0x1926U,0x324cU,0x6498U,0xc930U,0x826bU,0x14ddU,0x29baU,0x5374U,0xa6e8U,
		0x5ddbU,0xbbb6U,0x6767U,0xceceU,0x8d97U,0x0b25U,0x164aU,0x2c94U,0x5928U,0xb250U,
		0x74abU,0xe956U,0xc2a7U,0x9545U,0x3a81U,0x7502U,0xea04U,0xc403U,0x980dU,0x2011U,
		0x4022U,0x8044U,0x1083U,0x2106U,0x420cU,0x8418U,0x183bU,0x3076U,0x60ecU,0xc1d8U,
		0x93bbU,0x377dU,0x6efaU,0xddf4U,0xabe3U,0x47cdU,0x8f9aU,0x0f3fU,0x1e7eU,0x3cfcU,
		0x79f8U,0xf3f0U,0xf7ebU,0xffddU,0xefb1U,0xcf69U,0x8ed9U,0x0db9U,0x1b72U,0x36e4U,
		0x6dc8U,0xdb90U,0xa72bU,0x5e5dU,0xbcbaU,0x697fU,0xd2feU,0xb5f7U,0x7be5U,0xf7caU,
		0xff9fU,0xef35U,0xce61U,0x8cc9U,0x0999U,0x1332U,0x2664U,0x4cc8U,0x9990U,0x232bU,
		0x4656U,0x8cacU,0x0953U,0x12a6U,0x254cU,0x4a98U,0x9530U,0x3a6bU,0x74d6U,0xe9acU,
		0xc353U,0x96adU,0x3d51U,0x7aa2U,0xf544U,0xfa83U,0xe50dU,0xda11U,0xa429U,0x5859U,
		0xb0b2U,0x716fU,0xe2deU,0xd5b7U,0xbb65U,0x66c1U,0xcd82U,0x8b0fU,0x0615U,0x0c2aU,
		0x1854U,0x30a8U,0x6150U,0xc2a0U,0x954bU,0x3a9dU,0x753aU,0xea74U,0xc4e3U,0x99cdU,
		0x2391U,0x4722U,0x8e44U,0x0c83U,0x1906U,0x320cU,0x6418U,0xc830U,0x806bU,0x10ddU,
		0x21baU,0x4374U,0x86e8U,0x1ddbU,0x3bb6U,0x776cU,0xeed8U,0xcdbbU,0x8b7dU,0x06f1U,
		0x0de2U,0x1bc4U,0x3788U,0x6f10U,0xde20U,0xac4bU,0x489dU,0x913aU,0x327fU,0x64feU,
		0xc9fcU,0x83f3U,0x17edU,0x2fdaU,0x5fb4U,0xbf68U,0x6edbU,0xddb6U,0xab67U,0x46c5U,
		0x8d8aU,0x0b1fU,0x163eU,0x2c7cU,0x58f8U,0xb1f0U,0x73ebU,0xe7d6U,0xdfa7U,0xaf45U,
		0x4e81U,0x9d02U,0x2a0fU,0x541eU,0xa83cU,0x4073U,0x80e6U,0x11c7U,0x238eU,0x471cU,
		0x8e38U,0x0c7bU,0x18f6U,0x31ecU,0x63d8U,0xc7b0U,0x9f6bU,0x2eddU,0x5dbaU,0xbb74U,
		0x66e3U,0xcdc6U,0x8b87U,0x0705U,0x0e0aU,0x1c14U,0x3828U,0x7050U,0xe0a0U,0xd14bU,
		0xb29dU,0x7531U,0xea62U,0xc4cfU,0x9995U,0x2321U,0x4642U,0x8c84U,0x0903U,0x1206U,
		0x240cU,0x4818U,0x9030U,0x306bU,0x60d6U,0xc1acU,0x9353U,0x36adU,0x6d5aU,0xdab4U,
		0xa563U,0x5acdU,0xb59aU,0x7b3fU,0xf67eU,0xfcf7U,0xe9e5U,0xc3c1U,0x9789U,0x3f19U,
		0x7e32U,0xfc64U,0xe8c3U,0xc18dU,0x9311U,0x3629U,0x6c52U,0xd8a4U,0xa143U,0x528dU,
		0xa51aU,0x5a3fU,0xb47eU,0x78f7U,0xf1eeU,0xf3d7U,0xf7a5U,0xff41U,0xee89U,0xcd19U,
		0x8a39U,0x0479U,0x08f2U,0x11e4U,0x23c8U,0x4790U,0x8f20U,0x0e4bU,0x1c96U,0x392cU,
		0x7258U,0xe4b0U,0xd96bU,0xa2ddU,0x55b1U,0xab62U,0x46cfU,0x8d9eU,0x0b37U,0x166eU,
		0x2cdcU,0x59b8U,0xb370U,0x76ebU,0xedd6U,0xcba7U,0x8745U,0x1e81U,0x3d02U,0x7a04U,
		0xf408U,0xf81bU,0xe03dU,0xd071U,0xb0e9U,0x71d9U,0xe3b2U,0xd76fU,0xbed5U,0x6da1U,
		0xdb42U,0xa68fU,0x5d15U,0xba2aU,0x645fU,0xc8beU,0x8177U,0x12e5U,0x25caU,0x4b94U,
		0x9728U,0x3e5bU,0x7cb6U,0xf96cU,0xe2d3U,0xd5adU,0xbb51U,0x66a9U,0xcd52U,0x8aafU,
		0x0555U,0x0aaaU,0x1554U,0x2aa8U,0x5550U,0xaaa0U,0x454bU,0x8a96U,0x0527U,0x0a4eU,
		0x149cU,0x2938U,0x5270U,0xa4e0U,0x59cbU,0xb396U,0x7727U,0xee4eU,0xcc97U,0x8925U,
		0x0241U,0x0482U,0x0904U,0x1208U,0x2410U,0x4820U,0x9040U,0x308bU,0x6116U,0xc22cU,
		0x9453U,0x38adU,0x715aU,0xe2b4U,0xd563U,0xbacdU,0x6591U,0xcb22U,0x864fU,0x1c95U,
		0x392aU,0x7254U,0xe4a8U,0xd95bU,0xa2bdU,0x5571U,0xaae2U,0x45cfU,0x8b9eU,0x0737U,
		0x0e6eU,0x1cdcU,0x39b8U,0x7370U,0xe6e0U,0xddcbU,0xab9dU,0x4731U,0x8e62U,0x0ccfU,
		0x199eU,0x333cU,0x6678U,0xccf0U,0x89ebU,0x03ddU,0x07baU,0x0f74U,0x1ee8U,0x3dd0U,
		0x7ba0U,0xf740U,0xfe8bU,0xed1dU,0xca31U,0x8469U,0x18d9U,0x31b2U,0x6364U,0xc6c8U,
		0x9d9bU,0x2b3dU,0x567aU,0xacf4U,0x49e3U,0x93c6U,0x3787U,0x6f0eU,0xde1cU,0xac33U,
		0x486dU,0x90daU,0x31bfU,0x637eU,0xc6fcU,0x9df3U,0x2bedU,0x57daU,0xafb4U,0x4f63U,
		0x9ec6U,0x2d87U,0x5b0eU,0xb61cU,0x7c33U,0xf866U,0xe0c7U,0xd185U,0xb301U,0x7609U,
		0xec12U,0xc82fU,0x8055U,0x10a1U,0x2142U,0x4284U,0x8508U,0x1a1bU,0x3436U,0x686cU,
		0xd0d8U,0xb1bbU,0x737dU,0xe6faU,0xddffU,0xabf5U,0x47e1U,0x8fc2U,0x0f8fU,0x1f1eU,
		0x3e3cU,0x7c78U,0xf8f0U,0xe1ebU,0xd3ddU,0xb7b1U,0x7f69U,0xfed2U,0xedafU,0xcb55U,
		0x86a1U,0x1d49U,0x3a92U,0x7524U,0xea48U,0xc49bU,0x993dU,0x2271U,0x44e2U,0x89c4U,
		0x0383U,0x0706U,0x0e0cU,0x1c18U,0x3830U,0x7060U,0xe0c0U,0xd18bU,0xb31dU,0x7631U,
		0xec62U,0xc8cfU,0x8195U,0x1321U,0x2642U,0x4c84U,0x9908U,0x221bU,0x4436U,0x886cU,
		0x00d3U,0x01a6U,0x034cU,0x0698U,0x0d30U,0x1a60U,0x34c0U,0x6980U,0xd300U,0xb60bU,
		0x7c1dU,0xf83aU,0xe07fU,0xd0f5U,0xb1e1U,0x73c9U,0xe792U,0xdf2fU,0xae55U,0x4ca1U,
		0x9942U,0x228fU,0x451eU,0x8a3cU,0x0473U,0x08e6U,0x11ccU,0x2398U,0x4730U,0x8e60U,
		0x0ccbU,0x1996U,0x332cU,0x6658U,0xccb0U,0x896bU,0x02ddU,0x05baU,0x0b74U,0x16e8U,
		0x2dd0U,0x5ba0U,0xb740U,0x7e8bU,0xfd16U,0xea27U,0xc445U,0x9881U,0x2109U,0x4212U,
		0x8424U,0x1843U,0x3086U,0x610cU,0xc218U,0x943bU,0x387dU,0x70faU,0xe1f4U,0xd3e3U,
		0xb7cdU,0x7f91U,0xff22U,0xee4fU,0xcc95U,0x8921U,0x0249U,0x0492U,0x0924U,0x1248U,
		0x2490U,0x4920U,0x9240U,0x348bU,0x6916U,0xd22cU,0xb453U,0x78adU,0xf15aU,0xf2bfU,
		0xf575U,0xfae1U,0xe5c9U,0xdb99U,0xa739U,0x5e79U,0xbcf2U,0x69efU,0xd3deU,0xb7b7U,
		0x7f65U,0xfecaU,0xed9fU,0xcb35U,0x8661U,0x1cc9U,0x3992U,0x7324U,0xe648U,0xdc9bU,
		0xa93dU,0x4271U,0x84e2U,0x19cfU,0x339eU,0x673cU,0xce78U,0x8cfbU,0x09fdU,0x13faU,
		0x27f4U,0x4fe8U,0x9fd0U,0x2fabU,0x5f56U,0xbeacU,0x6d53U,0xdaa6U,0xa547U,0x5a85U,
		0xb50aU,0x7a1fU,0xf43eU,0xf877U,0xe0e5U,0xd1c1U,0xb389U,0x7719U,0xee32U,0xcc6fU,
		0x88d5U,0x01a1U,0x0342U,0x0684U,0x0d08U,0x1a10U,0x3420U,0x6840U,0xd080U,0xb10bU,
		0x721dU,0xe43aU,0xd87fU,0xa0f5U,0x51e1U,0xa3c2U,0x578fU,0xaf1eU,0x4e37U,0x9c6eU,
		0x28d7U,0x51aeU,0xa35cU,0x56b3U,0xad66U,0x4ac7U,0x958eU,0x3b17U,0x762eU,0xec5cU,
		0xc8b3U,0x816dU,0x12d1U,0x25a2U,0x4b44U,0x9688U,0x3d1bU,0x7a36U,0xf46cU,0xf8d3U,
		0xe1adU,0xd351U,0xb6a9U,0x7d59U,0xfab2U,0xe56fU,0xdad5U,0xa5a1U,0x5b49U,0xb692U,
		0x7d2fU,0xfa5eU,0xe4b7U,0xd965U,0xa2c1U,0x5589U,0xab12U,0x462fU,0x8c5eU,0x08b7U,
		0x116eU,0x22dcU,0x45b8U,0x8b70U,0x06ebU,0x0dd6U,0x1bacU,0x3758U,0x6eb0U,0xdd60U,
		0xaacbU,0x459dU,0x8b3aU,0x067fU,0x0cfeU,0x19fcU,0x33f8U,0x67f0U,0xcfe0U,0x8fcbU,
		0x0f9dU,0x1f3aU,0x3e74U,0x7ce8U,0xf9d0U,0xe3abU,0xd75dU,0xbeb1U,0x6d69U,0xdad2U,
		0xa5afU,0x5b55U,0xb6aaU,0x7d5fU,0xfabeU,0xe577U,0xdae5U,0xa5c1U,0x5b89U,0xb712U,
		0x7e2fU,0xfc5eU,0xe8b7U,0xc165U,0x92c1U,0x3589U,0x6b12U,0xd624U,0xbc43U,0x688dU,
		0xd11aU,0xb23fU,0x7475U,0xe8eaU,0xc1dfU,0x93b5U,0x3761U,0x6ec2U,0xdd84U,0xab03U,
		0x460dU,0x8c1aU,0x083fU,0x107eU,0x20fcU,0x41f8U,0x83f0U,0x17ebU,0x2fd6U,0x5facU,
		0xbf58U,0x6ebbU,0xdd76U,0xaae7U,0x45c5U,0x8b8aU,0x071fU,0x0e3eU,0x1c7cU,0x38f8U,
		0x71f0U,0xe3e0U,0xd7cbU,0xbf9dU,0x6f31U,0xde62U,0xaccfU,0x4995U,0x932aU,0x365fU,
		0x6cbeU,0xd97cU,0xa2f3U,0x55edU,0xabdaU,0x47bfU,0x8f7eU,0x0ef7U,0x1deeU,0x3bdcU,
		0x77b8U,0xef70U,0xceebU,0x8dddU,0x0bb1U,0x1762U,0x2ec4U,0x5d88U,0xbb10U,0x662bU,
		0xcc56U,0x88a7U,0x0145U,0x028aU,0x0514U,0x0a28U,0x1450U,0x28a0U,0x5140U,0xa280U,
		0x550bU,0xaa16U,0x4427U,0x884eU,0x0097U,0x012eU,0x025cU,0x04b8U,0x0970U,0x12e0U,
		0x25c0U,0x4b80U,0x9700U,0x3e0bU,0x7c16U,0xf82cU,0xe053U,0xd0adU,0xb151U,0x72a9U,
		0xe552U,0xdaafU,0xa555U,0x5aa1U,0xb542U,0x7a8fU,0xf51eU,0xfa37U,0xe465U,0xd8c1U,
		0xa189U,0x5319U,0xa632U,0x5c6fU,0xb8deU,0x61b7U,0xc36eU,0x96d7U,0x3da5U,0x7b4aU,
		0xf694U,0xfd23U,0xea4dU,0xc491U,0x9929U,0x2259U,0x44b2U,0x8964U,0x02c3U,0x0586U,
		0x0b0cU,0x1618U,0x2c30U,0x5860U,0xb0c0U,0x718bU,0xe316U,0xd627U,0xbc45U,0x6881U,
		0xd102U,0xb20fU,0x7415U,0xe82aU,0xc05fU,0x90b5U,0x3161U,0x62c2U,0xc584U,0x9b03U,
		0x260dU,0x4c1aU,0x9834U,0x2063U,0x40c6U,0x818cU,0x1313U,0x2626U,0x4c4cU,0x9898U,
		0x213bU,0x4276U,0x84ecU,0x19d3U,0x33a6U,0x674cU,0xce98U,0x8d3bU,0x0a7dU,0x14faU,
		0x29f4U,0x53e8U,0xa7d0U,0x5fabU,0xbf56U,0x6ea7U,0xdd4eU,0xaa97U,0x4525U,0x8a4aU,
		0x049fU,0x093eU,0x127cU,0x24f8U,0x49f0U,0x93e0U,0x37cbU,0x6f96U,0xdf2cU,0xae53U,
		0x4cadU,0x995aU,0x22bfU,0x457eU,0x8afcU,0x05f3U,0x0be6U,0x17ccU,0x2f98U,0x5f30U,
		0xbe60U,0x6ccbU,0xd996U,0xa327U,0x5645U,0xac8aU,0x491fU,0x923eU,0x3477U,0x68eeU,
		0xd1dcU,0xb3b3U,0x776dU,0xeedaU,0xcdbfU,0x8b75U,0x06e1U,0x0dc2U,0x1b84U,0x3708U,
		0x6e10U,0xdc20U,0xa84bU,0x409dU,0x813aU,0x127fU,0x24feU,0x49fcU,0x93f8U,0x37fbU,
		0x6ff6U,0xdfecU,0xafd3U,0x4fadU,0x9f5aU,0x2ebfU,0x5d7eU,0xbafcU,0x65f3U,0xcbe6U,
		0x87c7U,0x1f85U,0x3f0aU,0x7e14U,0xfc28U,0xe85bU,0xc0bdU,0x9171U,0x32e9U,0x65d2U,
		0xcba4U,0x8743U,0x1e8dU,0x3d1aU,0x7a34U,0xf468U,0xf8dbU,0xe1bdU,0xd371U,0xb6e9U,
		0x7dd9U,0xfbb2U,0xe76fU,0xded5U,0xada1U,0x4b49U,0x9692U,0x3d2fU,0x7a5eU,0xf4bcU,
		0xf973U,0xe2edU,0xd5d1U,0xbba9U,0x6759U,0xceb2U,0x8d6fU,0x0ad5U,0x15aaU,0x2b54U,
		0x56a8U,0xad50U,0x4aabU,0x9556U,0x3aa7U,0x754eU,0xea9cU,0xc533U,0x9a6dU,0x24d1U,
		0x49a2U,0x9344U,0x3683U,0x6d06U,0xda0cU,0xa413U,0x582dU,0xb05aU,0x70bfU,0xe17eU,
		0xd2f7U,0xb5e5U,0x7bc1U,0xf782U,0xff0fU,0xee15U,0xcc21U,0x8849U,0x0099U,0x0132U,
		0x0264U,0x04c8U,0x0990U,0x1320U,0x2640U,0x4c80U,0x9900U,0x220bU,0x4416U,0x882cU,
		0x0053U,0x00a6U,0x014cU,0x0298U,0x0530U,0x0a60U,0x14c0U,0x2980U,0x5300U,0xa600U,
		0x5c0bU,0xb816U,0x6027U,0xc04eU,0x9097U,0x3125U,0x624aU,0xc494U,0x9923U,0x224dU,
		0x449aU,0x8934U,0x0263U,0x04c6U,0x098cU,0x1318U,0x2630U,0x4c60U,0x98c0U,0x218bU,
		0x4316U,0x862cU,0x1c53U,0x38a6U,0x714cU,0xe298U,0xd53bU,0xba7dU,0x64f1U,0xc9e2U,
		0x83cfU,0x1795U,0x2f2aU,0x5e54U,0xbca8U,0x695bU,0xd2b6U,0xb567U,0x7ac5U,0xf58aU,
		0xfb1fU,0xe635U,0xdc61U,0xa8c9U,0x4199U,0x8332U,0x166fU,0x2cdeU,0x59bcU,0xb378U,
		0x76fbU,0xedf6U,0xcbe7U,0x87c5U,0x1f81U,0x3f02U,0x7e04U,0xfc08U,0xe81bU,0xc03dU,
		0x9071U,0x30e9U,0x61d2U,0xc3a4U,0x9743U,0x3e8dU,0x7d1aU,0xfa34U,0xe463U,0xd8cdU,
		0xa191U,0x5329U,0xa652U,0x5cafU,0xb95eU,0x62b7U,0xc56eU,0x9ad7U,0x25a5U,0x4b4aU,
		0x9694U,0x3d23U,0x7a46U,0xf48cU,0xf913U,0xe22dU,0xd451U,0xb8a9U,0x6159U,0xc2b2U,
		0x956fU,0x3ad5U,0x75aaU,0xeb54U,0xc6a3U,0x9d4dU,0x2a91U,0x5522U,0xaa44U,0x4483U,
		0x8906U,0x0207U,0x040eU,0x081cU,0x1038U,0x2070U,0x40e0U,0x81c0U,0x138bU,0x2716U,
		0x4e2cU,0x9c58U,0x28bbU,0x5176U,0xa2ecU,0x55d3U,0xaba6U,0x4747U,0x8e8eU,0x0d17U,
		0x1a2eU,0x345cU,0x68b8U,0xd170U,0xb2ebU,0x75ddU,0xebbaU,0xc77fU,0x9ef5U,0x2de1U,
		0x5bc2U,0xb784U,0x7f03U,0xfe06U,0xec07U,0xc805U,0x8001U,0x1009U,0x2012U,0x4024U,
		0x8048U,0x109bU,0x2136U,0x426cU,0x84d8U,0x19bbU,0x3376U,0x66ecU,0xcdd8U,0x8bbbU,
		0x077dU,0x0efaU,0x1df4U,0x3be8U,0x77d0U,0xefa0U,0xcf4bU,0x8e9dU,0x0d31U,0x1a62U,
		0x34c4U,0x6988U,0xd310U,0xb62bU,0x7c5dU,0xf8baU,0xe17fU,0xd2f5U,0xb5e1U,0x7bc9U,
		0xf792U,0xff2fU,0xee55U,0xcca1U,0x8949U,0x0299U,0x0532U,0x0a64U,0x14c8U,0x2990U,
		0x5320U,0xa640U,0x5c8bU,0xb916U,0x6227U,0xc44eU,0x9897U,0x2125U,0x424aU,0x8494U,
		0x1923U,0x3246U,0x648cU,0xc918U,0x823bU,0x147dU,0x28faU,0x51f4U,0xa3e8U,0x57dbU,
		0xafb6U,0x4f67U,0x9eceU,0x2d97U,0x5b2eU,0xb65cU,0x7cb3U,0xf966U,0xe2c7U,0xd585U,
		0xbb01U,0x6609U,0xcc12U,0x882fU,0x0055U,0x00aaU,0x0154U,0x02a8U,0x0550U,0x0aa0U,
		0x1540U,0x2a80U,0x5500U,0xaa00U,0x440bU,0x8816U,0x0027U,0x004eU,0x009cU,0x0138U,
		0x0270U,0x04e0U,0x09c0U,0x1380U,0x2700U,0x4e00U,0x9c00U,0x280bU,0x5016U,0xa02cU,
		0x5053U,0xa0a6U,0x5147U,0xa28eU,0x5517U,0xaa2eU,0x4457U,0x88aeU,0x0157U,0x02aeU,
		0x055cU,0x0ab8U,0x1570U,0x2ae0U,0x55c0U,0xab80U,0x470bU,0x8e16U,0x0c27U,0x184eU,
		0x309cU,0x6138U,0xc270U,0x94ebU,0x39ddU,0x73baU,0xe774U,0xdee3U,0xadcdU,0x4b91U,
		0x9722U,0x3e4fU,0x7c9eU,0xf93cU,0xe273U,0xd4edU,0xb9d1U,0x63a9U,0xc752U,0x9eafU,
		0x2d55U,0x5aaaU,0xb554U,0x7aa3U,0xf546U,0xfa87U,0xe505U,0xda01U,0xa409U,0x5819U,
		0xb032U,0x706fU,0xe0deU,0xd1b7U,0xb365U,0x76c1U,0xed82U,0xcb0fU,0x8615U,0x1c21U,
		0x3842U,0x7084U,0xe108U,0xd21bU,0xb43dU,0x7871U,0xf0e2U,0xf1cfU,0xf395U,0xf721U,
		0xfe49U,0xec99U,0xc939U,0x8279U,0x14f9U,0x29f2U,0x53e4U,0xa7c8U,0x5f9bU,0xbf36U,
		0x6e67U,0xdcceU,0xa997U,0x4325U,0x864aU,0x1c9fU,0x393eU,0x727cU,0xe4f8U,0xd9fbU,
		0xa3fdU,0x57f1U,0xafe2U,0x4fcfU,0x9f9eU,0x2f37U,0x5e6eU,0xbcdcU,0x69b3U,0xd366U,
		0xb6c7U,0x7d85U,0xfb0aU,0xe61fU,0xdc35U,0xa861U,0x40c9U,0x8192U,0x132fU,0x265eU,
		0x4cbcU,0x9978U,0x22fbU,0x45f6U,0x8becU,0x07d3U,0x0fa6U,0x1f4cU,0x3e98U,0x7d30U,
		0xfa60U,0xe4cbU,0xd99dU,0xa331U,0x5669U,0xacd2U,0x49afU,0x935eU,0x36b7U,0x6d6eU,
		0xdadcU,0xa5b3U,0x5b6dU,0xb6daU,0x7dbfU,0xfb7eU,0xe6f7U,0xdde5U,0xabc1U,0x4789U,
		0x8f12U,0x0e2fU,0x1c5eU,0x38bcU,0x7178U,0xe2f0U,0xd5ebU,0xbbddU,0x67b1U,0xcf62U,
		0x8ecfU,0x0d95U,0x1b2aU,0x3654U,0x6ca8U,0xd950U,0xa2abU,0x555dU,0xaabaU,0x457fU,
		0x8afeU,0x05f7U,0x0beeU,0x17dcU,0x2fb8U,0x5f70U,0xbee0U,0x6dcbU,0xdb96U,0xa727U,
		0x5e45U,0xbc8aU,0x691fU,0xd23eU,0xb477U,0x78e5U,0xf1caU,0xf39fU,0xf735U,0xfe61U,
		0xecc9U,0xc999U,0x8339U,0x1679U,0x2cf2U,0x59e4U,0xb3c8U,0x779bU,0xef36U,0xce67U,
		0x8cc5U,0x0981U,0x1302U,0x2604U,0x4c08U,0x9810U,0x202bU,0x4056U,0x80acU,0x1153U,
		0x22a6U,0x454cU,0x8a98U,0x053bU,0x0a76U,0x14ecU,0x29d8U,0x53b0U,0xa760U,0x5ecbU,
		0xbd96U,0x6b27U,0xd64eU,0xbc97U,0x6925U,0xd24aU,0xb49fU,0x7935U,0xf26aU,0xf4dfU,
		0xf9b5U,0xe361U,0xd6c9U,0xbd99U,0x6b39U,0xd672U,0xbcefU,0x69d5U,0xd3aaU,0xb75fU,
		0x7eb5U,0xfd6aU,0xeadfU,0xc5b5U,0x9b61U,0x26c9U,0x4d92U,0x9b24U,0x2643U,0x4c86U,
		0x990cU,0x2213U,0x4426U,0x884cU,0x0093U,0x0126U,0x024cU,0x0498U,0x0930U,0x1260U,
		0x24c0U,0x4980U,0x9300U,0x360bU,0x6c16U,0xd82cU,0xa053U,0x50adU,0xa15aU,0x52bfU,
		0xa57eU,0x5af7U,0xb5eeU,0x7bd7U,0xf7aeU,0xff57U,0xeea5U,0xcd41U,0x8a89U,0x0519U,
		0x0a32U,0x1464U,0x28c8U,0x5190U,0xa320U,0x564bU,0xac96U,0x4927U,0x924eU,0x3497U,
		0x692eU,0xd25cU,0xb4b3U,0x796dU,0xf2daU,0xf5bfU,0xfb75U,0xe6e1U,0xddc9U,0xab99U,
		0x4739U,0x8e72U,0x0cefU,0x19deU,0x33bcU,0x6778U,0xcef0U,0x8debU,0x0bddU,0x17baU,
		0x2f74U,0x5ee8U,0xbdd0U,0x6babU,0xd756U,0xbea7U,0x6d45U,0xda8aU,0xa51fU,0x5a35U,
		0xb46aU,0x78dfU,0xf1beU,0xf377U,0xf6e5U,0xfdc1U,0xeb89U,0xc719U,0x9e39U,0x2c79U,
		0x58f2U,0xb1e4U,0x73c3U,0xe786U,0xdf07U,0xae05U,0x4c01U,0x9802U,0x200fU,0x401eU,
		0x803cU,0x1073U,0x20e6U,0x41ccU,0x8398U,0x173bU,0x2e76U,0x5cecU,0xb9d8U,0x63bbU,
		0xc776U,0x9ee7U,0x2dc5U,0x5b8aU,0xb714U,0x7e23U,0xfc46U,0xe887U,0xc105U,0x9201U,
		0x3409U,0x6812U,0xd024U,0xb043U,0x708dU,0xe11aU,0xd23fU,0xb475U,0x78e1U,0xf1c2U,
		0xf38fU,0xf715U,0xfe21U,0xec49U,0xc899U,0x8139U,0x1279U,0x24f2U,0x49e4U,0x93c8U,
		0x379bU,0x6f36U,0xde6cU,0xacd3U,0x49adU,0x935aU,0x36bfU,0x6d7eU,0xdafcU,0xa5f3U,
		0x5bedU,0xb7daU,0x7fbfU,0xff7eU,0xeef7U,0xcde5U,0x8bc1U,0x0789U,0x0f12U,0x1e24U,
		0x3c48U,0x7890U,0xf120U,0xf24bU,0xf49dU,0xf931U,0xe269U,0xd4d9U,0xb9b9U,0x6379U,
		0xc6f2U,0x9defU,0x2bd5U,0x57aaU,0xaf54U,0x4ea3U,0x9d46U,0x2a87U,0x550eU,0xaa1cU,
		0x4433U,0x8866U,0x00c7U,0x018eU,0x031cU,0x0638U,0x0c70U,0x18e0U,0x31c0U,0x6380U,
		0xc700U,0x9e0bU,0x2c1dU,0x583aU,0xb074U,0x70e3U,0xe1c6U,0xd387U,0xb705U,0x7e01U,
		0xfc02U,0xe80fU,0xc015U,0x9021U,0x3049U,0x6092U,0xc124U,0x9243U,0x348dU,0x691aU,
		0xd234U,0xb463U,0x78cdU,0xf19aU,0xf33fU,0xf675U,0xfce1U,0xe9c9U,0xc399U,0x9739U,
		0x3e79U,0x7cf2U,0xf9e4U,0xe3c3U,0xd78dU,0xbf11U,0x6e29U,0xdc52U,0xa8afU,0x4155U,
		0x82aaU,0x155fU,0x2abeU,0x557cU,0xaaf8U,0x45fbU,0x8bf6U,0x07e7U,0x0fceU,0x1f9cU,
		0x3f38U,0x7e70U,0xfce0U,0xe9cbU,0xc39dU,0x9731U,0x3e69U,0x7cd2U,0xf9a4U,0xe343U,
		0xd68dU,0xbd11U,0x6a29U,0xd452U,0xb8afU,0x6155U,0xc2aaU,0x955fU,0x3ab5U,0x756aU,
		0xead4U,0xc5a3U,0x9b4dU,0x2691U,0x4d22U,0x9a44U,0x2483U,0x4906U,0x920cU,0x3413U,
		0x6826U,0xd04cU,0xb093U,0x712dU,0xe25aU,0xd4bfU,0xb975U,0x62e1U,0xc5c2U,0x9b8fU,
		0x2715U,0x4e2aU,0x9c54U,0x28a3U,0x5146U,0xa28cU,0x5513U,0xaa26U,0x4447U,0x888eU,
		0x0117U,0x022eU,0x045cU,0x08b8U,0x1170U,0x22e0U,0x45c0U,0x8b80U,0x070bU,0x0e16U,
		0x1c2cU,0x3858U,0x70b0U,0xe160U,0xd2cbU,0xb59dU,0x7b31U,0xf662U,0xfccfU,0xe995U,
		0xc321U,0x9649U,0x3c99U,0x7932U,0xf264U,0xf4c3U,0xf98dU,0xe311U,0xd629U,0xbc59U,
		0x68b9U,0xd172U,0xb2efU,0x75d5U,0xebaaU,0xc75fU,0x9eb5U,0x2d61U,0x5ac2U,0xb584U,
		0x7b03U,0xf606U,0xfc07U,0xe805U,0xc001U,0x9009U,0x3019U,0x6032U,0xc064U,0x90c3U,
		0x318dU,0x631aU,0xc634U,0x9c63U,0x28cdU,0x519aU,0xa334U,0x5663U,0xacc6U,0x4987U,
		0x930eU,0x3617U,0x6c2eU,0xd85cU,0xa0b3U,0x516dU,0xa2daU,0x55bfU,0xab7eU,0x46f7U,
		0x8deeU,0x0bd7U,0x17aeU,0x2f5cU,0x5eb8U,0xbd70U,0x6aebU,0xd5d6U,0xbba7U,0x6745U,
		0xce8aU,0x8d1fU,0x0a35U,0x146aU,0x28d4U,0x51a8U,0xa350U,0x56abU,0xad56U,0x4aa7U,
		0x954eU,0x3a97U,0x752eU,0xea5cU,0xc4b3U,0x996dU,0x22d1U,0x45a2U,0x8b44U,0x0683U,
		0x0d06U,0x1a0cU,0x3418U,0x6830U,0xd060U,0xb0cbU,0x719dU,0xe33aU,0xd67fU,0xbcf5U,
		0x69e1U,0xd3c2U,0xb78fU,0x7f15U,0xfe2aU,0xec5fU,0xc8b5U,0x8161U,0x12c9U,0x2592U,
		0x4b24U,0x9648U,0x3c9bU,0x7936U,0xf26cU,0xf4d3U,0xf9adU,0xe351U,0xd6a9U,0xbd59U,
		0x6ab9U,0xd572U,0xbaefU,0x65d5U,0xcbaaU,0x875fU,0x1eb5U,0x3d6aU,0x7ad4U,0xf5a8U,
		0xfb5bU,0xe6bdU,0xdd71U,0xaae9U,0x45d9U,0x8bb2U,0x076fU,0x0edeU,0x1dbcU,0x3b78U,
		0x76f0U,0xede0U,0xcbcbU,0x879dU,0x1f31U,0x3e62U,0x7cc4U,0xf988U,0xe31bU,0xd63dU,
		0xbc71U,0x68e9U,0xd1d2U,0xb3afU,0x7755U,0xeeaaU,0xcd5fU,0x8ab5U,0x0561U,0x0ac2U,
		0x1584U,0x2b08U,0x5610U,0xac20U,0x484bU,0x9096U,0x3127U,0x624eU,0xc49cU,0x9933U,
		0x226dU,0x44daU,0x89b4U,0x0363U,0x06c6U,0x0d8cU,0x1b18U,0x3630U,0x6c60U,0xd8c0U,
		0xa18bU,0x531dU,0xa63aU,0x5c7fU,0xb8feU,0x61f7U,0xc3eeU,0x97d7U,0x3fa5U,0x7f4aU,
		0xfe94U,0xed23U,0xca4dU,0x8491U,0x1929U,0x3252U,0x64a4U,0xc948U,0x829bU,0x153dU,
		0x2a7aU,0x54f4U,0xa9e8U,0x43dbU,0x87b6U,0x1f67U,0x3eceU,0x7d9cU,0xfb38U,0xe67bU,
		0xdcfdU,0xa9f1U,0x43e9U,0x87d2U,0x1fafU,0x3f5eU,0x7ebcU,0xfd78U,0xeafbU,0xc5fdU,
		0x9bf1U,0x27e9U,0x4fd2U,0x9fa4U,0x2f43U,0x5e86U,0xbd0cU,0x6a13U,0xd426U,0xb847U,
		0x6085U,0xc10aU,0x921fU,0x3435U,0x686aU,0xd0d4U,0xb1a3U,0x734dU,0xe69aU,0xdd3fU,
		0xaa75U,0x44e1U,0x89c2U,0x038fU,0x071eU,0x0e3cU,0x1c78U,0x38f0U,0x71e0U,0xe3c0U,
		0xd78bU,0xbf1dU,0x6e31U,0xdc62U,0xa8cfU,0x4195U,0x832aU,0x165fU,0x2cbeU,0x597cU,
		0xb2f8U,0x75fbU,0xebf6U,0xc7e7U,0x9fc5U,0x2f81U,0x5f02U,0xbe04U,0x6c03U,0xd806U,
		0xa007U,0x5005U,0xa00aU,0x501fU,0xa03eU,0x5077U,0xa0eeU,0x51d7U,0xa3aeU,0x5757U,
		0xaeaeU,0x4d57U,0x9aaeU,0x2557U,0x4aaeU,0x955cU,0x3ab3U,0x7566U,0xeaccU,0xc593U,
		0x9b2dU,0x2651U,0x4ca2U,0x9944U,0x2283U,0x4506U,0x8a0cU,0x0413U,0x0826U,0x104cU,
		0x2098U,0x4130U,0x8260U,0x14cbU,0x2996U,0x532cU,0xa658U,0x5cbbU,0xb976U,0x62e7U,
		0xc5ceU,0x9b97U,0x2725U,0x4e4aU,0x9c94U,0x2923U,0x5246U,0xa48cU,0x5913U,0xb226U,
		0x7447U,0xe88eU,0xc117U,0x9225U,0x3441U,0x6882U,0xd104U,0xb203U,0x740dU,0xe81aU,
		0xc03fU,0x9075U,0x30e1U,0x61c2U,0xc384U,0x9703U,0x3e0dU,0x7c1aU,0xf834U,0xe063U,
		0xd0cdU,0xb191U,0x7329U,0xe652U,0xdcafU,0xa955U,0x42a1U,0x8542U,0x1a8fU,0x351eU,
		0x6a3cU,0xd478U,0xb8fbU,0x61fdU,0xc3faU,0x97ffU,0x3ff5U,0x7feaU,0xffd4U,0xefa3U,
		0xcf4dU,0x8e91U,0x0d29U,0x1a52U,0x34a4U,0x6948U,0xd290U,0xb52bU,0x7a5dU,0xf4baU,
		0xf97fU,0xe2f5U,0xd5e1U,0xbbc9U,0x6799U,0xcf32U,0x8e6fU,0x0cd5U,0x19aaU,0x3354U,
		0x66a8U,0xcd50U,0x8aabU,0x055dU,0x0abaU,0x1574U,0x2ae8U,0x55d0U,0xaba0U,0x474bU,
		0x8e96U,0x0d27U,0x1a4eU,0x349cU,0x6938U,0xd270U,0xb4ebU,0x79ddU,0xf3baU,0xf77fU,
		0xfef5U,0xede1U,0xcbc9U,0x8799U,0x1f39U,0x3e72U,0x7ce4U,0xf9c8U,0xe39bU,0xd73dU,
		0xbe71U,0x6ce9U,0xd9d2U,0xa3afU,0x5755U,0xaeaaU,0x4d5fU,0x9abeU,0x2577U,0x4aeeU,
		0x95dcU,0x3bb3U,0x7766U,0xeeccU,0xcd93U,0x8b2dU,0x0651U,0x0ca2U,0x1944U,0x3288U,
		0x6510U,0xca20U,0x844bU,0x189dU,0x313aU,0x6274U,0xc4e8U,0x99dbU,0x23bdU,0x477aU,
		0x8ef4U,0x0de3U,0x1bc6U,0x378cU,0x6f18U,0xde30U,0xac6bU,0x48ddU,0x91baU,0x337fU,
		0x66feU,0xcdfcU,0x8bf3U,0x07edU,0x0fdaU,0x1fb4U,0x3f68U,0x7ed0U,0xfda0U,0xeb4bU,
		0xc69dU,0x9d31U,0x2a69U,0x54d2U,0xa9a4U,0x4343U,0x8686U,0x1d07U,0x3a0eU,0x741cU,
		0xe838U,0xc07bU,0x90fdU,0x31f1U,0x63e2U,0xc7c4U,0x9f83U,0x2f0dU,0x5e1aU,0xbc34U,
		0x6863U,0xd0c6U,0xb187U,0x7305U,0xe60aU,0xdc1fU,0xa835U,0x4061U,0x80c2U,0x118fU,
		0x231eU,0x463cU,0x8c78U,0x08fbU,0x11f6U,0x23ecU,0x47d8U,0x8fb0U,0x0f6bU,0x1ed6U,
		0x3dacU,0x7b58U,0xf6b0U,0xfd6bU,0xeaddU,0xc5b1U,0x9b69U,0x26d9U,0x4db2U,0x9b64U,
		0x26c3U,0x4d86U,0x9b0cU,0x2613U,0x4c26U,0x984cU,0x2093U,0x4126U,0x824cU,0x1493U,
		0x2926U,0x524cU,0xa498U,0x593bU,0xb276U,0x74e7U,0xe9ceU,0xc397U,0x9725U,0x3e41U,
		0x7c82U,0xf904U,0xe203U,0xd40dU,0xb811U,0x6029U,0xc052U,0x90afU,0x3155U,0x62aaU,
		0xc554U,0x9aa3U,0x254dU,0x4a9aU,0x9534U,0x3a63U,0x74c6U,0xe98cU,0xc313U,0x962dU,
		0x3c51U,0x78a2U,0xf144U,0xf283U,0xf50dU,0xfa11U,0xe429U,0xd859U,0xa0b9U,0x5179U,
		0xa2f2U,0x55efU,0xabdeU,0x47b7U,0x8f6eU,0x0ed7U,0x1daeU,0x3b5cU,0x76b8U,0xed70U,
		0xcaebU,0x85ddU,0x1bb1U,0x3762U,0x6ec4U,0xdd88U,0xab1bU,0x463dU,0x8c7aU,0x08ffU,
		0x11feU,0x23fcU,0x47f8U,0x8ff0U,0x0febU,0x1fd6U,0x3facU,0x7f58U,0xfeb0U,0xed6bU,
		0xcaddU,0x85b1U,0x1b69U,0x36d2U,0x6da4U,0xdb48U,0xa69bU,0x5d3dU,0xba7aU,0x64ffU,
		0xc9feU,0x83f7U,0x17e5U,0x2fcaU,0x5f94U,0xbf28U,0x6e5bU,0xdcb6U,0xa967U,0x42c5U,
		0x858aU,0x1b1fU,0x363eU,0x6c7cU,0xd8f8U,0xa1fbU,0x53fdU,0xa7faU,0x5fffU,0xbffeU,
		0x6ff7U,0xdfeeU,0xafd7U,0x4fa5U,0x9f4aU,0x2e9fU,0x5d3eU,0xba7cU,0x64f3U,0xc9e6U,
		0x83c7U,0x1785U,0x2f0aU,0x5e14U,0xbc28U,0x685bU,0xd0b6U,0xb167U,0x72c5U,0xe58aU,
		0xdb1fU,0xa635U,0x5c61U,0xb8c2U,0x618fU,0xc31eU,0x9637U,0x3c65U,0x78caU,0xf194U,
		0xf323U,0xf64dU,0xfc91U,0xe929U,0xc259U,0x94b9U,0x3979U,0x72f2U,0xe5e4U,0xdbc3U,
		0xa78dU,0x5f11U,0xbe22U,0x6c4fU,0xd89eU,0xa137U,0x5265U,0xa4caU,0x599fU,0xb33eU,
		0x7677U,0xeceeU,0xc9d7U,0x83a5U,0x1741U,0x2e82U,0x5d04U,0xba08U,0x641bU,0xc836U,
		0x8067U,0x10c5U,0x218aU,0x4314U,0x8628U,0x1c5bU,0x38b6U,0x716cU,0xe2d8U,0xd5bbU,
		0xbb7dU,0x66f1U,0xcde2U,0x8bcfU,0x0795U,0x0f2aU,0x1e54U,0x3ca8U,0x7950U,0xf2a0U,
		0xf54bU,0xfa9dU,0xe531U,0xda69U,0xa4d9U,0x59b9U,0xb372U,0x76efU,0xeddeU,0xcbb7U,
		0x8765U,0x1ec1U,0x3d82U,0x7b04U,0xf608U,0xfc1bU,0xe83dU,0xc071U,0x90e9U,0x31d9U,
		0x63b2U,0xc764U,0x9ec3U,0x2d8dU,0x5b1aU,0xb634U,0x7c63U,0xf8c6U,0xe187U,0xd305U,
		0xb601U,0x7c09U,0xf812U,0xe02fU,0xd055U,0xb0a1U,0x7149U,0xe292U,0xd52fU,0xba55U,
		0x64a1U,0xc942U,0x828fU,0x1515U,0x2a2aU,0x5454U,0xa8a8U,0x415bU,0x82b6U,0x1567U,
		0x2aceU,0x559cU,0xab38U,0x467bU,0x8cf6U,0x09e7U,0x13ceU,0x279cU,0x4f38U,0x9e70U,
		0x2cebU,0x59d6U,0xb3acU,0x7753U,0xeea6U,0xcd47U,0x8a85U,0x0501U,0x0a02U,0x1404U,
		0x2808U,0x5010U,0xa020U,0x504bU,0xa096U,0x5127U,0xa24eU,0x5497U,0xa92eU,0x4257U,
		0x84aeU,0x1957U,0x32aeU,0x655cU,0xcab8U,0x857bU,0x1afdU,0x35faU,0x6bf4U,0xd7e8U,
		0xbfdbU,0x6fbdU,0xdf7aU,0xaeffU,0x4df5U,0x9beaU,0x27dfU,0x4fbeU,0x9f7cU,0x2ef3U,
		0x5de6U,0xbbccU,0x6793U,0xcf26U,0x8e47U,0x0c85U,0x190aU,0x3214U,0x6428U,0xc850U,
		0x80abU,0x115dU,0x22baU,0x4574U,0x8ae8U,0x05dbU,0x0bb6U,0x176cU,0x2ed8U,0x5db0U,
		0xbb60U,0x66cbU,0xcd96U,0x8b27U,0x0645U,0x0c8aU,0x1914U,0x3228U,0x6450U,0xc8a0U,
		0x814bU,0x129dU,0x253aU,0x4a74U,0x94e8U,0x39dbU,0x73b6U,0xe76cU,0xded3U,0xadadU,
		0x4b51U,0x96a2U,0x3d4fU,0x7a9eU,0xf53cU,0xfa73U,0xe4edU,0xd9d1U,0xa3a9U,0x5759U,
		0xaeb2U,0x4d6fU,0x9adeU,0x25b7U,0x4b6eU,0x96dcU,0x3db3U,0x7b66U,0xf6ccU,0xfd93U,
		0xeb2dU,0xc651U,0x9ca9U,0x2959U,0x52b2U,0xa564U,0x5ac3U,0xb586U,0x7b07U,0xf60eU,
		0xfc17U,0xe825U,0xc041U,0x9089U,0x3119U,0x6232U,0xc464U,0x98c3U,0x218dU,0x431aU,
		0x8634U,0x1c63U,0x38c6U,0x718cU,0xe318U,0xd63bU,0xbc7dU,0x68f1U,0xd1e2U,0xb3cfU,
		0x7795U,0xef2aU,0xce5fU,0x8cb5U,0x0961U,0x12c2U,0x2584U,0x4b08U,0x9610U,0x3c2bU,
		0x7856U,0xf0acU,0xf153U,0xf2adU,0xf551U,0xfaa9U,0xe559U,0xdab9U,0xa579U,0x5af9U,
		0xb5f2U,0x7befU,0xf7deU,0xffb7U,0xef65U,0xcec1U,0x8d89U,0x0b19U,0x1632U,0x2c64U,
		0x58c8U,0xb190U,0x732bU,0xe656U,0xdca7U,0xa945U,0x4281U,0x8502U,0x1a0fU,0x341eU,
		0x683cU,0xd078U,0xb0fbU,0x71fdU,0xe3faU,0xd7ffU,0xbff5U,0x6fe1U,0xdfc2U,0xaf8fU,
		0x4f15U,0x9e2aU,0x2c5fU,0x58beU,0xb17cU,0x72f3U,0xe5e6U,0xdbc7U,0xa785U,0x5f01U,
		0xbe02U,0x6c0fU,0xd81eU,0xa037U,0x5065U,0xa0caU,0x519fU,0xa33eU,0x5677U,0xaceeU,
		0x49d7U,0x93aeU,0x3757U,0x6eaeU,0xdd5cU,0xaab3U,0x456dU,0x8adaU,0x05bfU,0x0b7eU,
		0x16fcU,0x2df8U,0x5bf0U,0xb7e0U,0x7fcbU,0xff96U,0xef27U,0xce45U,0x8c81U,0x0909U,
		0x1212U,0x2424U,0x4848U,0x9090U,0x312bU,0x6256U,0xc4acU,0x9953U,0x22adU,0x455aU,
		0x8ab4U,0x0563U,0x0ac6U,0x158cU,0x2b18U,0x5630U,0xac60U,0x48cbU,0x9196U,0x3327U,
		0x664eU,0xcc9cU,0x8933U,0x026dU,0x04daU,0x09b4U,0x1368U,0x26d0U,0x4da0U,0x9b40U,
		0x268bU,0x4d16U,0x9a2cU,0x2453U,0x48a6U,0x914cU,0x3293U,0x6526U,0xca4cU,0x8493U,
		0x192dU,0x325aU,0x64b4U,0xc968U,0x82dbU,0x15bdU,0x2b7aU,0x56f4U,0xade8U,0x4bdbU,
		0x97b6U,0x3f67U,0x7eceU,0xfd9cU,0xeb33U,0xc66dU,0x9cd1U,0x29a9U,0x5352U,0xa6a4U,
		0x5d43U,0xba86U,0x6507U,0xca0eU,0x8417U,0x1825U,0x304aU,0x6094U,0xc128U,0x925bU,
		0x34bdU,0x697aU,0xd2f4U,0xb5e3U,0x7bcdU,0xf79aU,0xff3fU,0xee75U,0xcce1U,0x89c9U,
		0x0399U,0x0732U,0x0e64U,0x1cc8U,0x3990U,0x7320U,0xe640U,0xdc8bU,0xa91dU,0x4231U,
		0x8462U,0x18cfU,0x319eU,0x633cU,0xc678U,0x9cfbU,0x29fdU,0x53faU,0xa7f4U,0x5fe3U,
		0xbfc6U,0x6f87U,0xdf0eU,0xae17U,0x4c25U,0x984aU,0x209fU,0x413eU,0x827cU,0x14f3U,
		0x29e6U,0x53ccU,0xa798U,0x5f3bU,0xbe76U,0x6ce7U,0xd9ceU,0xa397U,0x5725U,0xae4aU,
		0x4c9fU,0x993eU,0x2277U,0x44eeU,0x89dcU,0x03b3U,0x0766U,0x0eccU,0x1d98U,0x3b30U,
		0x7660U,0xecc0U,0xc98bU,0x831dU,0x1631U,0x2c62U,0x58c4U,0xb188U,0x731bU,0xe636U,
		0xdc67U,0xa8c5U,0x4181U,0x8302U,0x160fU,0x2c1eU,0x583cU,0xb078U,0x70fbU,0xe1f6U,
		0xd3e7U,0xb7c5U,0x7f81U,0xff02U,0xee0fU,0xcc15U,0x8821U,0x0049U,0x0092U,0x0124U,
		0x0248U,0x0490U,0x0920U,0x1240U,0x2480U,0x4900U,0x9200U,0x340bU,0x6816U,0xd02cU,
		0xb053U,0x70adU,0xe15aU,0xd2bfU,0xb575U,0x7ae1U,0xf5c2U,0xfb8fU,0xe715U,0xde21U,
		0xac49U,0x4899U,0x9132U,0x326fU,0x64deU,0xc9bcU,0x8373U,0x16edU,0x2ddaU,0x5bb4U,
		0xb768U,0x7edbU,0xfdb6U,0xeb67U,0xc6c5U,0x9d81U,0x2b09U,0x5612U,0xac24U,0x4843U,
		0x9086U,0x3107U,0x620eU,0xc41cU,0x9833U,0x206dU,0x40daU,0x81b4U,0x1363U,0x26c6U,
		0x4d8cU,0x9b18U,0x263bU,0x4c76U,0x98ecU,0x21d3U,0x43a6U,0x874cU,0x1e93U,0x3d26U,
		0x7a4cU,0xf498U,0xf93bU,0xe27dU,0xd4f1U,0xb9e9U,0x63d9U,0xc7b2U,0x9f6fU,0x2ed5U,
		0x5daaU,0xbb54U,0x66a3U,0xcd46U,0x8a87U,0x0505U,0x0a0aU,0x1414U,0x2828U,0x5050U,
		0xa0a0U,0x514bU,0xa296U,0x5527U,0xaa4eU,0x4497U,0x892eU,0x0257U,0x04aeU,0x095cU,
		0x12b8U,0x2570U,0x4ae0U,0x95c0U,0x3b8bU,0x7716U,0xee2cU,0xcc53U,0x88adU,0x0151U,
		0x02a2U,0x0544U,0x0a88U,0x1510U,0x2a20U,0x5440U,0xa880U,0x410bU,0x8216U,0x1427U,
		0x284eU,0x509cU,0xa138U,0x527bU,0xa4f6U,0x59e7U,0xb3ceU,0x7797U,0xef2eU,0xce57U,
		0x8ca5U,0x0941U,0x1282U,0x2504U,0x4a08U,0x9410U,0x382bU,0x7056U,0xe0acU,0xd153U,
		0xb2adU,0x7551U,0xeaa2U,0xc54fU,0x9a95U,0x2521U,0x4a42U,0x9484U,0x3903U,0x7206U,
		0xe40cU,0xd813U,0xa02dU,0x5051U,0xa0a2U,0x514fU,0xa29eU,0x5537U,0xaa6eU,0x44d7U,
		0x89aeU,0x0357U,0x06aeU,0x0d5cU,0x1ab8U,0x3570U,0x6ae0U,0xd5c0U,0xbb8bU,0x671dU,
		0xce3aU,0x8c7fU,0x08f5U,0x11eaU,0x23d4U,0x47a8U,0x8f50U,0x0eabU,0x1d56U,0x3aacU,
		0x7558U,0xeab0U,0xc56bU,0x9addU,0x25b1U,0x4b62U,0x96c4U,0x3d83U,0x7b06U,0xf60cU,
		0xfc13U,0xe82dU,0xc051U,0x90a9U,0x3159U,0x62b2U,0xc564U,0x9ac3U,0x258dU,0x4b1aU,
		0x9634U,0x3c63U,0x78c6U,0xf18cU,0xf313U,0xf62dU,0xfc51U,0xe8a9U,0xc159U,0x92b9U,
		0x3579U,0x6af2U,0xd5e4U,0xbbc3U,0x678dU,0xcf1aU,0x8e3fU,0x0c75U,0x18eaU,0x31d4U,
		0x63a8U,0xc750U,0x9eabU,0x2d5dU,0x5abaU,0xb574U,0x7ae3U,0xf5c6U,0xfb87U,0xe705U,
		0xde01U,0xac09U,0x4819U,0x9032U,0x306fU,0x60deU,0xc1bcU,0x9373U,0x36edU,0x6ddaU,
		0xdbb4U,0xa763U,0x5ecdU,0xbd9aU,0x6b3fU,0xd67eU,0xbcf7U,0x69e5U,0xd3caU,0xb79fU,
		0x7f35U,0xfe6aU,0xecdfU,0xc9b5U,0x8361U,0x16c9U,0x2d92U,0x5b24U,0xb648U,0x7c9bU,
		0xf936U,0xe267U,0xd4c5U,0xb981U,0x6309U,0xc612U,0x9c2fU,0x2855U,0x50aaU,0xa154U,
		0x52a3U,0xa546U,0x5a87U,0xb50eU,0x7a17U,0xf42eU,0xf857U,0xe0a5U,0xd141U,0xb289U,
		0x7519U,0xea32U,0xc46fU,0x98d5U,0x21a1U,0x4342U,0x8684U,0x1d03U,0x3a06U,0x740cU,
		0xe818U,0xc03bU,0x907dU,0x30f1U,0x61e2U,0xc3c4U,0x9783U,0x3f0dU,0x7e1aU,0xfc34U,
		0xe863U,0xc0cdU,0x9191U,0x3329U,0x6652U,0xcca4U,0x8943U,0x028dU,0x051aU,0x0a34U,
		0x1468U,0x28d0U,0x51a0U,0xa340U,0x568bU,0xad16U,0x4a27U,0x944eU,0x3897U,0x712eU,
		0xe25cU,0xd4b3U,0xb96dU,0x62d1U,0xc5a2U,0x9b4fU,0x2695U,0x4d2aU,0x9a54U,0x24a3U,
		0x4946U,0x928cU,0x3513U,0x6a26U,0xd44cU,0xb893U,0x612dU,0xc25aU,0x94bfU,0x3975U,
		0x72eaU,0xe5d4U,0xdba3U,0xa74dU,0x5e91U,0xbd22U,0x6a4fU,0xd49eU,0xb937U,0x6265U,
		0xc4caU,0x999fU,0x2335U,0x466aU,0x8cd4U,0x09a3U,0x1346U,0x268cU,0x4d18U,0x9a30U,
		0x246bU,0x48d6U,0x91acU,0x3353U,0x66a6U,0xcd4cU,0x8a93U,0x052dU,0x0a5aU,0x14b4U,
		0x2968U,0x52d0U,0xa5a0U,0x5b4bU,0xb696U,0x7d27U,0xfa4eU,0xe497U,0xd925U,0xa241U,
		0x5489U,0xa912U,0x422fU,0x845eU,0x18b7U,0x316eU,0x62dcU,0xc5b8U,0x9b7bU,0x26fdU,
		0x4dfaU,0x9bf4U,0x27e3U,0x4fc6U,0x9f8cU,0x2f13U,0x5e26U,0xbc4cU,0x6893U,0xd126U,
		0xb247U,0x7485U,0xe90aU,0xc21fU,0x9435U,0x3861U,0x70c2U,0xe184U,0xd303U,0xb60dU,
		0x7c11U,0xf822U,0xe04fU,0xd095U,0xb121U,0x7249U,0xe492U,0xd92fU,0xa255U,0x54a1U,
		0xa942U,0x428fU,0x851eU,0x1a37U,0x346eU,0x68dcU,0xd1b8U,0xb37bU,0x76fdU,0xedfaU,
		0xcbffU,0x87f5U,0x1fe1U,0x3fc2U,0x7f84U,0xff08U,0xee1bU,0xcc3dU,0x8871U,0x00e9U,
		0x01d2U,0x03a4U,0x0748U,0x0e90U,0x1d20U,0x3a40U,0x7480U,0xe900U,0xc20bU,0x941dU,
		0x3831U,0x7062U,0xe0c4U,0xd183U,0xb30dU,0x7611U,0xec22U,0xc84fU,0x8095U,0x1121U,
		0x2242U,0x4484U,0x8908U,0x021bU,0x0436U,0x086cU,0x10d8U,0x21b0U,0x4360U,0x86c0U,
		0x1d8bU,0x3b16U,0x762cU,0xec58U,0xc8bbU,0x817dU,0x12f1U,0x25e2U,0x4bc4U,0x9788U,
		0x3f1bU,0x7e36U,0xfc6cU,0xe8d3U,0xc1adU,0x9351U,0x36a9U,0x6d52U,0xdaa4U,0xa543U,
		0x5a8dU,0xb51aU,0x7a3fU,0xf47eU,0xf8f7U,0xe1e5U,0xd3c1U,0xb789U,0x7f19U,0xfe32U,
		0xec6fU,0xc8d5U,0x81a1U,0x1349U,0x2692U,0x4d24U,0x9a48U,0x249bU,0x4936U,0x926cU,
		0x34d3U,0x69a6U,0xd34cU,0xb693U,0x7d2dU,0xfa5aU,0xe4bfU,0xd975U,0xa2e1U,0x55c9U,
		0xab92U,0x472fU,0x8e5eU,0x0cb7U,0x196eU,0x32dcU,0x65b8U,0xcb70U,0x86ebU,0x1dddU,
		0x3bbaU,0x7774U,0xeee8U,0xcddbU,0x8bbdU,0x0771U,0x0ee2U,0x1dc4U,0x3b88U,0x7710U,
		0xee20U,0xcc4bU,0x889dU,0x0131U,0x0262U,0x04c4U,0x0988U,0x1310U,0x2620U,0x4c40U,
		0x9880U,0x210bU,0x4216U,0x842cU,0x1853U,0x30a6U,0x614cU,0xc298U,0x953bU,0x3a7dU,
		0x74faU,0xe9f4U,0xc3e3U,0x97cdU,0x3f91U,0x7f22U,0xfe44U,0xec83U,0xc90dU,0x8211U,
		0x1429U,0x2852U,0x50a4U,0xa148U,0x529bU,0xa536U,0x5a67U,0xb4ceU,0x7997U,0xf32eU,
		0xf657U,0xfca5U,0xe941U,0xc289U,0x9519U,0x3a39U,0x7472U,0xe8e4U,0xc1c3U,0x938dU,
		0x3711U,0x6e22U,0xdc44U,0xa883U,0x410dU,0x821aU,0x143fU,0x287eU,0x50fcU,0xa1f8U,
		0x53fbU,0xa7f6U,0x5fe7U,0xbfceU,0x6f97U,0xdf2eU,0xae57U,0x4ca5U,0x994aU,0x229fU,
		0x453eU,0x8a7cU,0x04f3U,0x09e6U,0x13ccU,0x2798U,0x4f30U,0x9e60U,0x2ccbU,0x5996U,
		0xb32cU,0x7653U,0xeca6U,0xc947U,0x8285U,0x1501U,0x2a02U,0x5404U,0xa808U,0x401bU,
		0x8036U,0x1067U,0x20ceU,0x419cU,0x8338U,0x167bU,0x2cf6U,0x59ecU,0xb3d8U,0x77bbU,
		0xef76U,0xcee7U,0x8dc5U,0x0b81U,0x1702U,0x2e04U,0x5c08U,0xb810U,0x602bU,0xc056U,
		0x90a7U,0x3145U,0x628aU,0xc514U,0x9a23U,0x244dU,0x489aU,0x9134U,0x3263U,0x64c6U,
		0xc98cU,0x8313U,0x162dU,0x2c5aU,0x58b4U,0xb168U,0x72dbU,0xe5b6U,0xdb67U,0xa6c5U,
		0x5d81U,0xbb02U,0x660fU,0xcc1eU,0x8837U,0x0065U,0x00caU,0x0194U,0x0328U,0x0650U,
		0x0ca0U,0x1940U,0x3280U,0x6500U,0xca00U,0x840bU,0x181dU,0x303aU,0x6074U,0xc0e8U,
		0x91dbU,0x33bdU,0x677aU,0xcef4U,0x8de3U,0x0bcdU,0x179aU,0x2f34U,0x5e68U,0xbcd0U,
		0x69abU,0xd356U,0xb6a7U,0x7d45U,0xfa8aU,0xe51fU,0xda35U,0xa461U,0x58c9U,0xb192U,
		0x732fU,0xe65eU,0xdcb7U,0xa965U,0x42c1U,0x8582U,0x1b0fU,0x361eU,0x6c3cU,0xd878U,
		0xa0fbU,0x51fdU,0xa3faU,0x57ffU,0xaffeU,0x4ff7U,0x9feeU,0x2fd7U,0x5faeU,0xbf5cU,
		0x6eb3U,0xdd66U,0xaac7U,0x4585U,0x8b0aU,0x061fU,0x0c3eU,0x187cU,0x30f8U,0x61f0U,
		0xc3e0U,0x97cbU,0x3f9dU,0x7f3aU,0xfe74U,0xece3U,0xc9cdU,0x8391U,0x1729U,0x2e52U,
		0x5ca4U,0xb948U,0x629bU,0xc536U,0x9a67U,0x24c5U,0x498aU,0x9314U,0x3623U,0x6c46U,
		0xd88cU,0xa113U,0x522dU,0xa45aU,0x58bfU,0xb17eU,0x72f7U,0xe5eeU,0xdbd7U,0xa7a5U,
		0x5f41U,0xbe82U,0x6d0fU,0xda1eU,0xa437U,0x5865U,0xb0caU,0x719fU,0xe33eU,0xd677U,
		0xbce5U,0x69c1U,0xd382U,0xb70fU,0x7e15U,0xfc2aU,0xe85fU,0xc0b5U,0x9161U,0x32c9U,
		0x6592U,0xcb24U,0x8643U,0x1c8dU,0x391aU,0x7234U,0xe468U,0xd8dbU,0xa1bdU,0x5371U,
		0xa6e2U,0x5dcfU,0xbb9eU,0x6737U,0xce6eU,0x8cd7U,0x09a5U,0x134aU,0x2694U,0x4d28U,
		0x9a50U,0x24abU,0x4956U,0x92acU,0x3553U,0x6aa6U,0xd54cU,0xba93U,0x652dU,0xca5aU,
		0x84bfU,0x1975U,0x32eaU,0x65d4U,0xcba8U,0x875bU,0x1ebdU,0x3d7aU,0x7af4U,0xf5e8U,
		0xfbdbU,0xe7bdU,0xdf71U,0xaee9U,0x4dd9U,0x9bb2U,0x276fU,0x4edeU,0x9dbcU,0x2b73U,
		0x56e6U,0xadccU,0x4b93U,0x9726U,0x3e47U,0x7c8eU,0xf91cU,0xe233U,0xd46dU,0xb8d1U,
		0x61a9U,0xc352U,0x96afU,0x3d55U,0x7aaaU,0xf554U,0xfaa3U,0xe54dU,0xda91U,0xa529U,
		0x5a59U,0xb4b2U,0x796fU,0xf2deU,0xf5b7U,0xfb65U,0xe6c1U,0xdd89U,0xab19U,0x4639U,
		0x8c72U,0x08efU,0x11deU,0x23bcU,0x4778U,0x8ef0U,0x0debU,0x1bd6U,0x37acU,0x6f58U,
		0xdeb0U,0xad6bU,0x4addU,0x95baU,0x3b7fU,0x76feU,0xedfcU,0xcbf3U,0x87edU,0x1fd1U,
		0x3fa2U,0x7f44U,0xfe88U,0xed1bU,0xca3dU,0x8471U,0x18e9U,0x31d2U,0x63a4U,0xc748U,
		0x9e9bU,0x2d3dU,0x5a7aU,0xb4f4U,0x79e3U,0xf3c6U,0xf787U,0xff05U,0xee01U,0xcc09U,
		0x8819U,0x0039U,0x0072U,0x00e4U,0x01c8U,0x0390U,0x0720U,0x0e40U,0x1c80U,0x3900U,
		0x7200U,0xe400U,0xd80bU,0xa01dU,0x5031U,0xa062U,0x50cfU,0xa19eU,0x5337U,0xa66eU,
		0x5cd7U,0xb9aeU,0x6357U,0xc6aeU,0x9d57U,0x2aa5U,0x554aU,0xaa94U,0x4523U,0x8a46U,
		0x0487U,0x090eU,0x121cU,0x2438U,0x4870U,0x90e0U,0x31cbU,0x6396U,0xc72cU,0x9e53U,
		0x2cadU,0x595aU,0xb2b4U,0x7563U,0xeac6U,0xc587U,0x9b05U,0x2601U,0x4c02U,0x9804U,
		0x2003U,0x4006U,0x800cU,0x1013U,0x2026U,0x404cU,0x8098U,0x113bU,0x2276U,0x44ecU,
		0x89d8U,0x03bbU,0x0776U,0x0eecU,0x1dd8U,0x3bb0U,0x7760U,0xeec0U,0xcd8bU,0x8b1dU,
		0x0631U,0x0c62U,0x18c4U,0x3188U,0x6310U,0xc620U,0x9c4bU,0x289dU,0x513aU,0xa274U,
		0x54e3U,0xa9c6U,0x4387U,0x870eU,0x1e17U,0x3c2eU,0x785cU,0xf0b8U,0xf17bU,0xf2fdU,
		0xf5f1U,0xfbe9U,0xe7d9U,0xdfb9U,0xaf79U,0x4ef9U,0x9df2U,0x2befU,0x57deU,0xafbcU,
		0x4f73U,0x9ee6U,0x2dc7U,0x5b8eU,0xb71cU,0x7e33U,0xfc66U,0xe8c7U,0xc185U,0x9301U,
		0x3609U,0x6c12U,0xd824U,0xa043U,0x508dU,0xa11aU,0x523fU,0xa47eU,0x58f7U,0xb1eeU,
		0x73d7U,0xe7aeU,0xdf57U,0xaea5U,0x4d41U,0x9a82U,0x250fU,0x4a1eU,0x943cU,0x3873U,
		0x70e6U,0xe1ccU,0xd393U,0xb72dU,0x7e51U,0xfca2U,0xe94fU,0xc295U,0x9521U,0x3a49U,
		0x7492U,0xe924U,0xc243U,0x948dU,0x3911U,0x7222U,0xe444U,0xd883U,0xa10dU,0x5211U,
		0xa422U,0x584fU,0xb09eU,0x7137U,0xe26eU,0xd4d7U,0xb9a5U,0x6341U,0xc682U,0x9d0fU,
		0x2a15U,0x542aU,0xa854U,0x40a3U,0x8146U,0x1287U,0x250eU,0x4a1cU,0x9438U,0x387bU,
		0x70f6U,0xe1ecU,0xd3d3U,0xb7adU,0x7f51U,0xfea2U,0xed4fU,0xca95U,0x8521U,0x1a49U,
		0x3492U,0x6924U,0xd248U,0xb49bU,0x793dU,0xf27aU,0xf4ffU,0xf9f5U,0xe3e1U,0xd7c9U,
		0xbf99U,0x6f39U,0xde72U,0xacefU,0x49d5U,0x93aaU,0x375fU,0x6ebeU,0xdd7cU,0xaaf3U,
		0x45edU,0x8bdaU,0x07bfU,0x0f7eU,0x1efcU,0x3df8U,0x7bf0U,0xf7e0U,0xffcbU,0xef9dU,
		0xcf31U,0x8e69U,0x0cd9U,0x19b2U,0x3364U,0x66c8U,0xcd90U,0x8b2bU,0x065dU,0x0cbaU,
		0x1974U,0x32e8U,0x65d0U,0xcba0U,0x874bU,0x1e9dU,0x3d3aU,0x7a74U,0xf4e8U,0xf9dbU,
		0xe3bdU,0xd771U,0xbee9U,0x6dd9U,0xdbb2U,0xa76fU,0x5ed5U,0xbdaaU,0x6b5fU,0xd6beU,
		0xbd77U,0x6ae5U,0xd5caU,0xbb9fU,0x6735U,0xce6aU,0x8cdfU,0x09b5U,0x136aU,0x26d4U,
		0x4da8U,0x9b50U,0x26abU,0x4d56U,0x9aacU,0x2553U,0x4aa6U,0x954cU,0x3a93U,0x7526U,
		0xea4cU,0xc493U,0x992dU,0x2251U,0x44a2U,0x8944U,0x0283U,0x0506U,0x0a0cU,0x1418U,
		0x2830U,0x5060U,0xa0c0U,0x518bU,0xa316U,0x5627U,0xac4eU,0x4897U,0x912eU,0x3257U,
		0x64aeU,0xc95cU,0x82b3U,0x156dU,0x2adaU,0x55b4U,0xab68U,0x46dbU,0x8db6U,0x0b67U,
		0x16ceU,0x2d9cU,0x5b38U,0xb670U,0x7cebU,0xf9d6U,0xe3a7U,0xd745U,0xbe81U,0x6d09U,
		0xda12U,0xa42fU,0x5855U,0xb0aaU,0x715fU,0xe2beU,0xd577U,0xbae5U,0x65c1U,0xcb82U,
		0x870fU,0x1e15U,0x3c2aU,0x7854U,0xf0a8U,0xf15bU,0xf2bdU,0xf571U,0xfae9U,0xe5d9U,
		0xdbb9U,0xa779U,0x5ef9U,0xbdf2U,0x6befU,0xd7deU,0xbfb7U,0x6f65U,0xdecaU,0xad9fU,
		0x4b35U,0x966aU,0x3cdfU,0x79beU,0xf37cU,0xf6f3U,0xfdedU,0xebd1U,0xc7a9U,0x9f59U,
		0x2eb9U,0x5d72U,0xbae4U,0x65c3U,0xcb86U,0x8707U,0x1e05U,0x3c0aU,0x7814U,0xf028U,
		0xf05bU,0xf0bdU,0xf171U,0xf2e9U,0xf5d9U,0xfbb9U,0xe779U,0xdef9U,0xadf9U,0x4bf9U,
		0x97f2U,0x3fefU,0x7fdeU,0xffbcU,0xef73U,0xceedU,0x8dd1U,0x0ba9U,0x1752U,0x2ea4U,
		0x5d48U,0xba90U,0x652bU,0xca56U,0x84a7U,0x1945U,0x328aU,0x6514U,0xca28U,0x845bU,
		0x18bdU,0x317aU,0x62f4U,0xc5e8U,0x9bdbU,0x27bdU,0x4f7aU,0x9ef4U,0x2de3U,0x5bc6U,
		0xb78cU,0x7f13U,0xfe26U,0xec47U,0xc885U,0x8101U,0x1209U,0x2412U,0x4824U,0x9048U,
		0x309bU,0x6136U,0xc26cU,0x94d3U,0x39adU,0x735aU,0xe6b4U,0xdd63U,0xaacdU,0x4591U,
		0x8b22U,0x064fU,0x0c9eU,0x193cU,0x3278U,0x64f0U,0xc9e0U,0x83cbU,0x179dU,0x2f3aU,
		0x5e74U,0xbce8U,0x69dbU,0xd3b6U,0xb767U,0x7ec5U,0xfd8aU,0xeb1fU,0xc635U,0x9c61U,
		0x28c9U,0x5192U,0xa324U,0x5643U,0xac86U,0x4907U,0x920eU,0x3417U,0x682eU,0xd05cU,
		0xb0b3U,0x716dU,0xe2daU,0xd5bfU,0xbb75U,0x66e1U,0xcdc2U,0x8b8fU,0x0715U,0x0e2aU,
		0x1c54U,0x38a8U,0x7150U,0xe2a0U,0xd54bU,0xba9dU,0x6531U,0xca62U,0x84cfU,0x1995U,
		0x332aU,0x6654U,0xcca8U,0x895bU,0x02bdU,0x057aU,0x0af4U,0x15e8U,0x2bd0U,0x57a0U,
		0xaf40U,0x4e8bU,0x9d16U,0x2a27U,0x544eU,0xa89cU,0x4133U,0x8266U,0x14c7U,0x298eU,
		0x531cU,0xa638U,0x5c7bU,0xb8f6U,0x61e7U,0xc3ceU,0x9797U,0x3f25U,0x7e4aU,0xfc94U,
		0xe923U,0xc24dU,0x9491U,0x3929U,0x7252U,0xe4a4U,0xd943U,0xa28dU,0x5511U,0xaa22U,
		0x444fU,0x889eU,0x0137U,0x026eU,0x04dcU,0x09b8U,0x1370U,0x26e0U,0x4dc0U,0x9b80U,
		0x270bU,0x4e16U,0x9c2cU,0x2853U,0x50a6U,0xa14cU,0x5293U,0xa526U,0x5a47U,0xb48eU,
		0x7917U,0xf22eU,0xf457U,0xf8a5U,0xe141U,0xd289U,0xb519U,0x7a39U,0xf472U,0xf8efU,
		0xe1d5U,0xd3a1U,0xb749U,0x7e99U,0xfd32U,0xea6fU,0xc4d5U,0x99a1U,0x2349U,0x4692U,
		0x8d24U,0x0a43U,0x1486U,0x290cU,0x5218U,0xa430U,0x586bU,0xb0d6U,0x71a7U,0xe34eU,
		0xd697U,0xbd25U,0x6a41U,0xd482U,0xb90fU,0x6215U,0xc42aU,0x985fU,0x20b5U,0x416aU,
		0x82d4U,0x15a3U,0x2b46U,0x568cU,0xad18U,0x4a3bU,0x9476U,0x38e7U,0x71ceU,0xe39cU,
		0xd733U,0xbe6dU,0x6cd1U,0xd9a2U,0xa34fU,0x5695U,0xad2aU,0x4a5fU,0x94beU,0x3977U,
		0x72eeU,0xe5dcU,0xdbb3U,0xa76dU,0x5ed1U,0xbda2U,0x6b4fU,0xd69eU,0xbd37U,0x6a65U,
		0xd4caU,0xb99fU,0x6335U,0xc66aU,0x9cdfU,0x29b5U,0x536aU,0xa6d4U,0x5da3U,0xbb46U,
		0x6687U,0xcd0eU,0x8a17U,0x0425U,0x084aU,0x1094U,0x2128U,0x4250U,0x84a0U,0x194bU,
		0x3296U,0x652cU,0xca58U,0x84bbU,0x197dU,0x32faU,0x65f4U,0xcbe8U,0x87dbU,0x1fbdU,
		0x3f7aU,0x7ef4U,0xfde8U,0xebdbU,0xc7bdU,0x9f71U,0x2ee9U,0x5dd2U,0xbba4U,0x6743U,
		0xce86U,0x8d07U,0x0a05U,0x140aU,0x2814U,0x5028U,0xa050U,0x50abU,0xa156U,0x52a7U,
		0xa54eU,0x5a97U,0xb52eU,0x7a57U,0xf4aeU,0xf957U,0xe2a5U,0xd541U,0xba89U,0x6519U,
		0xca32U,0x846fU,0x18d5U,0x31aaU,0x6354U,0xc6a8U,0x9d5bU,0x2abdU,0x557aU,0xaaf4U,
		0x45e3U,0x8bc6U,0x0787U,0x0f0eU,0x1e1cU,0x3c38U,0x7870U,0xf0e0U,0xf1cbU,0xf39dU,
		0xf731U,0xfe69U,0xecd9U,0xc9b9U,0x8379U,0x16f9U,0x2df2U,0x5be4U,0xb7c8U,0x7f9bU,
		0xff36U,0xee67U,0xccc5U,0x8981U,0x0309U,0x0612U,0x0c24U,0x1848U,0x3090U,0x6120U,
		0xc240U,0x948bU,0x391dU,0x723aU,0xe474U,0xd8e3U,0xa1cdU,0x5391U,0xa722U,0x5e4fU,
		0xbc9eU,0x6937U,0xd26eU,0xb4d7U,0x79a5U,0xf34aU,0xf69fU,0xfd35U,0xea61U,0xc4c9U,
		0x9999U,0x2339U,0x4672U,0x8ce4U,0x09c3U,0x1386U,0x270cU,0x4e18U,0x9c30U,0x286bU,
		0x50d6U,0xa1acU,0x5353U,0xa6a6U,0x5d47U,0xba8eU,0x6517U,0xca2eU,0x8457U,0x18a5U,
		0x314aU,0x6294U,0xc528U,0x9a5bU,0x24bdU,0x497aU,0x92f4U,0x35e3U,0x6bc6U,0xd78cU,
		0xbf13U,0x6e2dU,0xdc5aU,0xa8bfU,0x4175U,0x82eaU,0x15dfU,0x2bbeU,0x577cU,0xaef8U,
		0x4dfbU,0x9bf6U,0x27e7U,0x4fceU,0x9f9cU,0x2f33U,0x5e66U,0xbcccU,0x6993U,0xd326U,
		0xb647U,0x7c85U,0xf90aU,0xe21fU,0xd435U,0xb861U,0x60c9U,0xc192U,0x932fU,0x3655U,
		0x6caaU,0xd954U,0xa2a3U,0x554dU,0xaa9aU,0x453fU,0x8a7eU,0x04f7U,0x09eeU,0x13dcU,
		0x27b8U,0x4f70U,0x9ee0U,0x2dcbU,0x5b96U,0xb72cU,0x7e53U,0xfca6U,0xe947U,0xc285U,
		0x9501U,0x3a09U,0x7412U,0xe824U,0xc043U,0x908dU,0x3111U,0x6222U,0xc444U,0x9883U,
		0x210dU,0x421aU,0x8434U,0x1863U,0x30c6U,0x618cU,0xc318U,0x963bU,0x3c7dU,0x78faU,
		0xf1f4U,0xf3e3U,0xf7cdU,0xff91U,0xef29U,0xce59U,0x8cb9U,0x0979U,0x12f2U,0x25e4U,
		0x4bc8U,0x9790U,0x3f2bU,0x7e56U,0xfcacU,0xe953U,0xc2adU,0x9551U,0x3aa9U,0x7552U,
		0xeaa4U,0xc543U,0x9a8dU,0x2511U,0x4a22U,0x9444U,0x3883U,0x7106U,0xe20cU,0xd413U,
		0xb82dU,0x6051U,0xc0a2U,0x914fU,0x3295U,0x652aU,0xca54U,0x84a3U,0x194dU,0x329aU,
		0x6534U,0xca68U,0x84dbU,0x19bdU,0x337aU,0x66f4U,0xcde8U,0x8bdbU,0x07bdU,0x0f7aU,
		0x1ef4U,0x3de8U,0x7bd0U,0xf7a0U,0xff4bU,0xee9dU,0xcd31U,0x8a69U,0x04d9U,0x09b2U,
		0x1364U,0x26c8U,0x4d90U,0x9b20U,0x264bU,0x4c96U,0x992cU,0x2253U,0x44a6U,0x894cU,
		0x0293U,0x0526U,0x0a4cU,0x1498U,0x2930U,0x5260U,0xa4c0U,0x598bU,0xb316U,0x7627U,
		0xec4eU,0xc897U,0x8125U,0x1241U,0x2482U,0x4904U,0x9208U,0x341bU,0x6836U,0xd06cU,
		0xb0d3U,0x71adU,0xe35aU,0xd6bfU,0xbd75U,0x6ae1U,0xd5c2U,0xbb8fU,0x6715U,0xce2aU,
		0x8c5fU,0x08b5U,0x116aU,0x22d4U,0x45a8U,0x8b50U,0x06abU,0x0d56U,0x1aacU,0x3558U,
		0x6ab0U,0xd560U,0xbacbU,0x659dU,0xcb3aU,0x867fU,0x1cf5U,0x39eaU,0x73d4U,0xe7a8U,
		0xdf5bU,0xaebdU,0x4d71U,0x9ae2U,0x25cfU,0x4b9eU,0x973cU,0x3e73U,0x7ce6U,0xf9ccU,
		0xe393U,0xd72dU,0xbe51U,0x6ca9U,0xd952U,0xa2afU,0x5555U,0xaaaaU,0x455fU,0x8abeU,
		0x0577U,0x0aeeU,0x15dcU,0x2bb8U,0x5770U,0xaee0U,0x4dcbU,0x9b96U,0x2727U,0x4e4eU,
		0x9c9cU,0x2933U,0x5266U,0xa4ccU,0x5993U,0xb326U,0x7647U,0xec8eU,0xc917U,0x8225U,
		0x1441U,0x2882U,0x5104U,0xa208U,0x541bU,0xa836U,0x4067U,0x80ceU,0x1197U,0x232eU,
		0x465cU,0x8cb8U,0x097bU,0x12f6U,0x25ecU,0x4bd8U,0x97b0U,0x3f6bU,0x7ed6U,0xfdacU,
		0xeb53U,0xc6adU,0x9d51U,0x2aa9U,0x5552U,0xaaa4U,0x4543U,0x8a86U,0x0507U,0x0a0eU,
		0x141cU,0x2838U,0x5070U,0xa0e0U,0x51cbU,0xa396U,0x5727U,0xae4eU,0x4c97U,0x992eU,
		0x2257U,0x44aeU,0x895cU,0x02b3U,0x0566U,0x0accU,0x1598U,0x2b30U,0x5660U,0xacc0U,
		0x498bU,0x9316U,0x3627U,0x6c4eU,0xd89cU,0xa133U,0x526dU,0xa4daU,0x59bfU,0xb37eU,
		0x76f7U,0xedeeU,0xcbd7U,0x87a5U,0x1f41U,0x3e82U,0x7d04U,0xfa08U,0xe41bU,0xd83dU,
		0xa071U,0x50e9U,0xa1d2U,0x53afU,0xa75eU,0x5eb7U,0xbd6eU,0x6ad7U,0xd5aeU,0xbb57U,
		0x66a5U,0xcd4aU,0x8a9fU,0x0535U,0x0a6aU,0x14d4U,0x29a8U,0x5350U,0xa6a0U,0x5d4bU,
		0xba96U,0x6527U,0xca4eU,0x8497U,0x1925U,0x324aU,0x6494U,0xc928U,0x825bU,0x14bdU,
		0x297aU,0x52f4U,0xa5e8U,0x5bdbU,0xb7b6U,0x7f67U,0xfeceU,0xed97U,0xcb25U,0x8641U,
		0x1c89U,0x3912U,0x7224U,0xe448U,0xd89bU,0xa13dU,0x5271U,0xa4e2U,0x59cfU,0xb39eU,
		0x7737U,0xee6eU,0xccd7U,0x89a5U,0x0341U,0x0682U,0x0d04U,0x1a08U,0x3410U,0x6820U,
		0xd040U,0xb08bU,0x711dU,0xe23aU,0xd47fU,0xb8f5U,0x61e1U,0xc3c2U,0x978fU,0x3f15U,
		0x7e2aU,0xfc54U,0xe8a3U,0xc14dU,0x9291U,0x3529U,0x6a52U,0xd4a4U,0xb943U,0x628dU,
		0xc51aU,0x9a3fU,0x2475U,0x48eaU,0x91d4U,0x33a3U,0x6746U,0xce8cU,0x8d13U,0x0a2dU,
		0x145aU,0x28b4U,0x5168U,0xa2d0U,0x55abU,0xab56U,0x46a7U,0x8d4eU,0x0a97U,0x152eU,
		0x2a5cU,0x54b8U,0xa970U,0x42ebU,0x85d6U,0x1ba7U,0x374eU,0x6e9cU,0xdd38U,0xaa7bU,
		0x44fdU,0x89faU,0x03ffU,0x07feU,0x0ffcU,0x1ff8U,0x3ff0U,0x7fe0U,0xffc0U,0xef8bU,
		0xcf1dU,0x8e31U,0x0c69U,0x18d2U,0x31a4U,0x6348U,0xc690U,0x9d2bU,0x2a5dU,0x54baU,
		0xa974U,0x42e3U,0x85c6U,0x1b87U,0x370eU,0x6e1cU,0xdc38U,0xa87bU,0x40fdU,0x81faU,
		0x13ffU,0x27feU,0x4ffcU,0x9ff8U,0x2ffbU,0x5ff6U,0xbfecU,0x6fd3U,0xdfa6U,0xaf47U,
		0x4e85U,0x9d0aU,0x2a1fU,0x543eU,0xa87cU,0x40f3U,0x81e6U,0x13c7U,0x278eU,0x4f1cU,
		0x9e38U,0x2c7bU,0x58f6U,0xb1ecU,0x73d3U,0xe7a6U,0xdf47U,0xae85U,0x4d01U,0x9a02U,
		0x240fU,0x481eU,0x903cU,0x3073U,0x60e6U,0xc1ccU,0x9393U,0x372dU,0x6e5aU,0xdcb4U,
		0xa963U,0x42cdU,0x859aU,0x1b3fU,0x367eU,0x6cfcU,0xd9f8U,0xa3fbU,0x57fdU,0xaffaU,
		0x4fffU,0x9ffeU,0x2ff7U,0x5feeU,0xbfdcU,0x6fb3U,0xdf66U,0xaec7U,0x4d85U,0x9b0aU,
		0x261fU,0x4c3eU,0x987cU,0x20f3U,0x41e6U,0x83ccU,0x1793U,0x2f26U,0x5e4cU,0xbc98U,
		0x693bU,0xd276U,0xb4e7U,0x79c5U,0xf38aU,0xf71fU,0xfe35U,0xec61U,0xc8c9U,0x8199U,
		0x1339U,0x2672U,0x4ce4U,0x99c8U,0x239bU,0x4736U,0x8e6cU,0x0cd3U,0x19a6U,0x334cU,
		0x6698U,0xcd30U,0x8a6bU,0x04ddU,0x09baU,0x1374U,0x26e8U,0x4dd0U,0x9ba0U,0x274bU,
		0x4e96U,0x9d2cU,0x2a53U,0x54a6U,0xa94cU,0x4293U,0x8526U,0x1a47U,0x348eU,0x691cU,
		0xd238U,0xb47bU,0x78fdU,0xf1faU,0xf3ffU,0xf7f5U,0xffe1U,0xefc9U,0xcf99U,0x8f39U,
		0x0e79U,0x1cf2U,0x39e4U,0x73c8U,0xe790U,0xdf2bU,0xae5dU,0x4cb1U,0x9962U,0x22cfU,
		0x459eU,0x8b3cU,0x0673U,0x0ce6U,0x19ccU,0x3398U,0x6730U,0xce60U,0x8ccbU,0x099dU,
		0x133aU,0x2674U,0x4ce8U,0x99d0U,0x23abU,0x4756U,0x8eacU,0x0d53U,0x1aa6U,0x354cU,
		0x6a98U,0xd530U,0xba6bU,0x64ddU,0xc9baU,0x837fU,0x16f5U,0x2deaU,0x5bd4U,0xb7a8U,
		0x7f5bU,0xfeb6U,0xed67U,0xcac5U,0x8581U,0x1b09U,0x3612U,0x6c24U,0xd848U,0xa09bU,
		0x513dU,0xa27aU,0x54ffU,0xa9feU,0x43f7U,0x87eeU,0x1fd7U,0x3faeU,0x7f5cU,0xfeb8U,
		0xed7bU,0xcafdU,0x85f1U,0x1be9U,0x37d2U,0x6fa4U,0xdf48U,0xae9bU,0x4d3dU,0x9a7aU,
		0x24ffU,0x49feU,0x93fcU,0x37f3U,0x6fe6U,0xdfccU,0xaf93U,0x4f2dU,0x9e5aU,0x2cbfU,
		0x597eU,0xb2fcU,0x75f3U,0xebe6U,0xc7c7U,0x9f85U,0x2f01U,0x5e02U,0xbc04U,0x6803U,
		0xd006U,0xb007U,0x7005U,0xe00aU,0xd01fU,0xb035U,0x7061U,0xe0c2U,0xd18fU,0xb315U,
		0x7621U,0xec42U,0xc88fU,0x8115U,0x1221U,0x2442U,0x4884U,0x9108U,0x321bU,0x6436U,
		0xc86cU,0x80d3U,0x11adU,0x235aU,0x46b4U,0x8d68U,0x0adbU,0x15b6U,0x2b6cU,0x56d8U,
		0xadb0U,0x4b6bU,0x96d6U,0x3da7U,0x7b4eU,0xf69cU,0xfd33U,0xea6dU,0xc4d1U,0x99a9U,
		0x2359U,0x46b2U,0x8d64U,0x0ac3U,0x1586U,0x2b0cU,0x5618U,0xac30U,0x486bU,0x90d6U,
		0x31a7U,0x634eU,0xc69cU,0x9d33U,0x2a6dU,0x54daU,0xa9b4U,0x4363U,0x86c6U,0x1d87U,
		0x3b0eU,0x761cU,0xec38U,0xc87bU,0x80fdU,0x11f1U,0x23e2U,0x47c4U,0x8f88U,0x0f1bU,
		0x1e36U,0x3c6cU,0x78d8U,0xf1b0U,0xf36bU,0xf6ddU,0xfdb1U,0xeb69U,0xc6d9U,0x9db9U,
		0x2b79U,0x56f2U,0xade4U,0x4bc3U,0x9786U,0x3f07U,0x7e0eU,0xfc1cU,0xe833U,0xc06dU,
		0x90d1U,0x31a9U,0x6352U,0xc6a4U,0x9d43U,0x2a8dU,0x551aU,0xaa34U,0x4463U,0x88c6U,
		0x0187U,0x030eU,0x061cU,0x0c38U,0x1870U,0x30e0U,0x61c0U,0xc380U,0x970bU,0x3e1dU,
		0x7c3aU,0xf874U,0xe0e3U,0xd1cdU,0xb391U,0x7729U,0xee52U,0xccafU,0x8955U,0x02a1U,
		0x0542U,0x0a84U,0x1508U,0x2a10U,0x5420U,0xa840U,0x408bU,0x8116U,0x1227U,0x244eU,
		0x489cU,0x9138U,0x327bU,0x64f6U,0xc9ecU,0x83d3U,0x17adU,0x2f5aU,0x5eb4U,0xbd68U,
		0x6adbU,0xd5b6U,0xbb67U,0x66c5U,0xcd8aU,0x8b1fU,0x0635U,0x0c6aU,0x18d4U,0x31a8U,
		0x6350U,0xc6a0U,0x9d4bU,0x2a9dU,0x553aU,0xaa74U,0x44e3U,0x89c6U,0x0387U,0x070eU,
		0x0e1cU,0x1c38U,0x3870U,0x70e0U,0xe1c0U,0xd38bU,0xb71dU,0x7e31U,0xfc62U,0xe8cfU,
		0xc195U,0x9321U,0x3649U,0x6c92U,0xd924U,0xa243U,0x548dU,0xa91aU,0x423fU,0x847eU,
		0x18f7U,0x31eeU,0x63dcU,0xc7b8U,0x9f7bU,0x2efdU,0x5dfaU,0xbbf4U,0x67e3U,0xcfc6U,
		0x8f87U,0x0f05U,0x1e0aU,0x3c14U,0x7828U,0xf050U,0xf0abU,0xf15dU,0xf2b1U,0xf569U,
		0xfad9U,0xe5b9U,0xdb79U,0xa6f9U,0x5df9U,0xbbf2U,0x67efU,0xcfdeU,0x8fb7U,0x0f65U,
		0x1ecaU,0x3d94U,0x7b28U,0xf650U,0xfcabU,0xe95dU,0xc2b1U,0x9569U,0x3ad9U,0x75b2U,
		0xeb64U,0xc6c3U,0x9d8dU,0x2b11U,0x5622U,0xac44U,0x4883U,0x9106U,0x3207U,0x640eU,
		0xc81cU,0x8033U,0x106dU,0x20daU,0x41b4U,0x8368U,0x16dbU,0x2db6U,0x5b6cU,0xb6d8U,
		0x7dbbU,0xfb76U,0xe6e7U,0xddc5U,0xab81U,0x4709U,0x8e12U,0x0c2fU,0x185eU,0x30bcU,
		0x6178U,0xc2f0U,0x95ebU,0x3bddU,0x77baU,0xef74U,0xcee3U,0x8dcdU,0x0b91U,0x1722U,
		0x2e44U,0x5c88U,0xb910U,0x622bU,0xc456U,0x98a7U,0x2145U,0x428aU,0x8514U,0x1a23U,
		0x3446U,0x688cU,0xd118U,0xb23bU,0x747dU,0xe8faU,0xc1ffU,0x93f5U,0x37e1U,0x6fc2U,
		0xdf84U,0xaf03U,0x4e0dU,0x9c1aU,0x283fU,0x507eU,0xa0fcU,0x51f3U,0xa3e6U,0x57c7U,
		0xaf8eU,0x4f17U,0x9e2eU,0x2c57U,0x58aeU,0xb15cU,0x72b3U,0xe566U,0xdac7U,0xa585U,
		0x5b01U,0xb602U,0x7c0fU,0xf81eU,0xe037U,0xd065U,0xb0c1U,0x7189U,0xe312U,0xd62fU,
		0xbc55U,0x68a1U,0xd142U,0xb28fU,0x7515U,0xea2aU,0xc45fU,0x98b5U,0x2161U,0x42c2U,
		0x8584U,0x1b03U,0x3606U,0x6c0cU,0xd818U,0xa03bU,0x507dU,0xa0faU,0x51ffU,0xa3feU,
		0x57f7U,0xafeeU,0x4fd7U,0x9faeU,0x2f57U,0x5eaeU,0xbd5cU,0x6ab3U,0xd566U,0xbac7U,
		0x6585U,0xcb0aU,0x861fU,0x1c35U,0x386aU,0x70d4U,0xe1a8U,0xd35bU,0xb6bdU,0x7d71U,
		0xfae2U,0xe5cfU,0xdb95U,0xa721U,0x5e49U,0xbc92U,0x692fU,0xd25eU,0xb4b7U,0x7965U,
		0xf2caU,0xf59fU,0xfb35U,0xe661U,0xdcc9U,0xa999U,0x4339U,0x8672U,0x1cefU,0x39deU,
		0x73bcU,0xe778U,0xdefbU,0xadfdU,0x4bf1U,0x97e2U,0x3fcfU,0x7f9eU,0xff3cU,0xee73U,
		0xccedU,0x89d1U,0x03a9U,0x0752U,0x0ea4U,0x1d48U,0x3a90U,0x7520U,0xea40U,0xc48bU,
		0x991dU,0x2231U,0x4462U,0x88c4U,0x0183U,0x0306U,0x060cU,0x0c18U,0x1830U,0x3060U,
		0x60c0U,0xc180U,0x930bU,0x361dU,0x6c3aU,0xd874U,0xa0e3U,0x51cdU,0xa39aU,0x573fU,
		0xae7eU,0x4cf7U,0x99eeU,0x23d7U,0x47aeU,0x8f5cU,0x0eb3U,0x1d66U,0x3accU,0x7598U,
		0xeb30U,0xc66bU,0x9cddU,0x29b1U,0x5362U,0xa6c4U,0x5d83U,0xbb06U,0x6607U,0xcc0eU,
		0x8817U,0x0025U,0x004aU,0x0094U,0x0128U,0x0250U,0x04a0U,0x0940U,0x1280U,0x2500U,
		0x4a00U,0x9400U,0x380bU,0x7016U,0xe02cU,0xd053U,0xb0adU,0x7151U,0xe2a2U,0xd54fU,
		0xba95U,0x6521U,0xca42U,0x848fU,0x1915U,0x322aU,0x6454U,0xc8a8U,0x815bU,0x12bdU,
		0x257aU,0x4af4U,0x95e8U,0x3bdbU,0x77b6U,0xef6cU,0xced3U,0x8dadU,0x0b51U,0x16a2U,
		0x2d44U,0x5a88U,0xb510U,0x7a2bU,0xf456U,0xf8a7U,0xe145U,0xd281U,0xb509U,0x7a19U,
		0xf432U,0xf86fU,0xe0d5U,0xd1a1U,0xb349U,0x7699U,0xed32U,0xca6fU,0x84d5U,0x19a1U,
		0x3342U,0x6684U,0xcd08U,0x8a1bU,0x043dU,0x087aU,0x10f4U,0x21e8U,0x43d0U,0x87a0U,
		0x1f4bU,0x3e96U,0x7d2cU,0xfa58U,0xe4bbU,0xd97dU,0xa2f1U,0x55e9U,0xabd2U,0x47afU,
		0x8f5eU,0x0eb7U,0x1d6eU,0x3adcU,0x75b8U,0xeb70U,0xc6ebU,0x9dddU,0x2bb1U,0x5762U,
		0xaec4U,0x4d83U,0x9b06U,0x2607U,0x4c0eU,0x981cU,0x2033U,0x4066U,0x80ccU,0x1193U,
		0x2326U,0x464cU,0x8c98U,0x093bU,0x1276U,0x24ecU,0x49d8U,0x93b0U,0x376bU,0x6ed6U,
		0xddacU,0xab53U,0x46adU,0x8d5aU,0x0abfU,0x157eU,0x2afcU,0x55f8U,0xabf0U,0x47ebU,
		0x8fd6U,0x0fa7U,0x1f4eU,0x3e9cU,0x7d38U,0xfa70U,0xe4ebU,0xd9ddU,0xa3b1U,0x5769U,
		0xaed2U,0x4dafU,0x9b5eU,0x26b7U,0x4d6eU,0x9adcU,0x25b3U,0x4b66U,0x96ccU,0x3d93U,
		0x7b26U,0xf64cU,0xfc93U,0xe92dU,0xc251U,0x94a9U,0x3959U,0x72b2U,0xe564U,0xdac3U,
		0xa58dU,0x5b11U,0xb622U,0x7c4fU,0xf89eU,0xe137U,0xd265U,0xb4c1U,0x7989U,0xf312U,
		0xf62fU,0xfc55U,0xe8a1U,0xc149U,0x9299U,0x3539U,0x6a72U,0xd4e4U,0xb9c3U,0x638dU,
		0xc71aU,0x9e3fU,0x2c75U,0x58eaU,0xb1d4U,0x73a3U,0xe746U,0xde87U,0xad05U,0x4a01U,
		0x9402U,0x380fU,0x701eU,0xe03cU,0xd073U,0xb0edU,0x71d1U,0xe3a2U,0xd74fU,0xbe95U,
		0x6d21U,0xda42U,0xa48fU,0x5915U,0xb22aU,0x745fU,0xe8beU,0xc177U,0x92e5U,0x35c1U,
		0x6b82U,0xd704U,0xbe03U,0x6c0dU,0xd81aU,0xa03fU,0x5075U,0xa0eaU,0x51dfU,0xa3beU,
		0x5777U,0xaeeeU,0x4dd7U,0x9baeU,0x2757U,0x4eaeU,0x9d5cU,0x2ab3U,0x5566U,0xaaccU,
		0x4593U,0x8b26U,0x0647U,0x0c8eU,0x191cU,0x3238U,0x6470U,0xc8e0U,0x81cbU,0x139dU,
		0x273aU,0x4e74U,0x9ce8U,0x29dbU,0x53b6U,0xa76cU,0x5ed3U,0xbda6U,0x6b47U,0xd68eU,
		0xbd17U,0x6a25U,0xd44aU,0xb89fU,0x6135U,0xc26aU,0x94dfU,0x39b5U,0x736aU,0xe6d4U,
		0xdda3U,0xab4dU,0x4691U,0x8d22U,0x0a4fU,0x149eU,0x293cU,0x5278U,0xa4f0U,0x59ebU,
		0xb3d6U,0x77a7U,0xef4eU,0xce97U,0x8d25U,0x0a41U,0x1482U,0x2904U,0x5208U,0xa410U,
		0x582bU,0xb056U,0x70a7U,0xe14eU,0xd297U,0xb525U,0x7a41U,0xf482U,0xf90fU,0xe215U,
		0xd421U,0xb849U,0x6099U,0xc132U,0x926fU,0x34d5U,0x69aaU,0xd354U,0xb6a3U,0x7d4dU,
		0xfa9aU,0xe53fU,0xda75U,0xa4e1U,0x59c9U,0xb392U,0x772fU,0xee5eU,0xccb7U,0x8965U,
		0x02c1U,0x0582U,0x0b04U,0x1608U,0x2c10U,0x5820U,0xb040U,0x708bU,0xe116U,0xd227U,
		0xb445U,0x7881U,0xf102U,0xf20fU,0xf415U,0xf821U,0xe049U,0xd099U,0xb139U,0x7279U,
		0xe4f2U,0xd9efU,0xa3d5U,0x57a1U,0xaf42U,0x4e8fU,0x9d1eU,0x2a37U,0x546eU,0xa8dcU,
		0x41b3U,0x8366U,0x16c7U,0x2d8eU,0x5b1cU,0xb638U,0x7c7bU,0xf8f6U,0xe1e7U,0xd3c5U,
		0xb781U,0x7f09U,0xfe12U,0xec2fU,0xc855U,0x80a1U,0x1149U,0x2292U,0x4524U,0x8a48U,
		0x049bU,0x0936U,0x126cU,0x24d8U,0x49b0U,0x9360U,0x36cbU,0x6d96U,0xdb2cU,0xa653U,
		0x5cadU,0xb95aU,0x62bfU,0xc57eU,0x9af7U,0x25e5U,0x4bcaU,0x9794U,0x3f23U,0x7e46U,
		0xfc8cU,0xe913U,0xc22dU,0x9451U,0x38a9U,0x7152U,0xe2a4U,0xd543U,0xba8dU,0x6511U,
		0xca22U,0x844fU,0x1895U,0x312aU,0x6254U,0xc4a8U,0x995bU,0x22bdU,0x457aU,0x8af4U,
		0x05e3U,0x0bc6U,0x178cU,0x2f18U,0x5e30U,0xbc60U,0x68cbU,0xd196U,0xb327U,0x7645U,
		0xec8aU,0xc91fU,0x8235U,0x1461U,0x28c2U,0x5184U,0xa308U,0x561bU,0xac36U,0x4867U,
		0x90ceU,0x3197U,0x632eU,0xc65cU,0x9cb3U,0x296dU,0x52daU,0xa5b4U,0x5b63U,0xb6c6U,
		0x7d87U,0xfb0eU,0xe617U,0xdc25U,0xa841U,0x4089U,0x8112U,0x122fU,0x245eU,0x48bcU,
		0x9178U,0x32fbU,0x65f6U,0xcbecU,0x87d3U,0x1fadU,0x3f5aU,0x7eb4U,0xfd68U,0xeadbU,
		0xc5bdU,0x9b71U,0x26e9U,0x4dd2U,0x9ba4U,0x2743U,0x4e86U,0x9d0cU,0x2a13U,0x5426U,
		0xa84cU,0x4093U,0x8126U,0x1247U,0x248eU,0x491cU,0x9238U,0x347bU,0x68f6U,0xd1ecU,
		0xb3d3U,0x77adU,0xef5aU,0xcebfU,0x8d75U,0x0ae1U,0x15c2U,0x2b84U,0x5708U,0xae10U,
		0x4c2bU,0x9856U,0x20a7U,0x414eU,0x829cU,0x1533U,0x2a66U,0x54ccU,0xa998U,0x433bU,
		0x8676U,0x1ce7U,0x39ceU,0x739cU,0xe738U,0xde7bU,0xacfdU,0x49f1U,0x93e2U,0x37cfU,
		0x6f9eU,0xdf3cU,0xae73U,0x4cedU,0x99daU,0x23bfU,0x477eU,0x8efcU,0x0df3U,0x1be6U,
		0x37ccU,0x6f98U,0xdf30U,0xae6bU,0x4cddU,0x99baU,0x237fU,0x46feU,0x8dfcU,0x0bf3U,
		0x17e6U,0x2fccU,0x5f98U,0xbf30U,0x6e6bU,0xdcd6U,0xa9a7U,0x4345U,0x868aU,0x1d1fU,
		0x3a3eU,0x747cU,0xe8f8U,0xc1fbU,0x93fdU,0x37f1U,0x6fe2U,0xdfc4U,0xaf83U,0x4f0dU,
		0x9e1aU,0x2c3fU,0x587eU,0xb0fcU,0x71f3U,0xe3e6U,0xd7c7U,0xbf85U,0x6f01U,0xde02U,
		0xac0fU,0x4815U,0x902aU,0x305fU,0x60beU,0xc17cU,0x92f3U,0x35edU,0x6bdaU,0xd7b4U,
		0xbf63U,0x6ecdU,0xdd9aU,0xab3fU,0x4675U,0x8ceaU,0x09dfU,0x13beU,0x277cU,0x4ef8U,
		0x9df0U,0x2bebU,0x57d6U,0xafacU,0x4f53U,0x9ea6U,0x2d47U,0x5a8eU,0xb51cU,0x7a33U,
		0xf466U,0xf8c7U,0xe185U,0xd301U,0xb609U,0x7c19U,0xf832U,0xe06fU,0xd0d5U,0xb1a1U,
		0x7349U,0xe692U,0xdd2fU,0xaa55U,0x44a1U,0x8942U,0x028fU,0x051eU,0x0a3cU,0x1478U,
		0x28f0U,0x51e0U,0xa3c0U,0x578bU,0xaf16U,0x4e27U,0x9c4eU,0x2897U,0x512eU,0xa25cU,
		0x54b3U,0xa966U,0x42c7U,0x858eU,0x1b17U,0x362eU,0x6c5cU,0xd8b8U,0xa17bU,0x52fdU,
		0xa5faU,0x5bffU,0xb7feU,0x7ff7U,0xffeeU,0xefd7U,0xcfa5U,0x8f41U,0x0e89U,0x1d12U,
		0x3a24U,0x7448U,0xe890U,0xc12bU,0x925dU,0x34b1U,0x6962U,0xd2c4U,0xb583U,0x7b0dU,
		0xf61aU,0xfc3fU,0xe875U,0xc0e1U,0x91c9U,0x3399U,0x6732U,0xce64U,0x8cc3U,0x098dU,
		0x131aU,0x2634U,0x4c68U,0x98d0U,0x21abU,0x4356U,0x86acU,0x1d53U,0x3aa6U,0x754cU,
		0xea98U,0xc53bU,0x9a7dU,0x24f1U,0x49e2U,0x93c4U,0x3783U,0x6f06U,0xde0cU,0xac13U,
		0x482dU,0x905aU,0x30bfU,0x617eU,0xc2fcU,0x95f3U,0x3bedU,0x77daU,0xefb4U,0xcf63U,
		0x8ecdU,0x0d91U,0x1b22U,0x3644U,0x6c88U,0xd910U,0xa22bU,0x545dU,0xa8baU,0x417fU,
		0x82feU,0x15f7U,0x2beeU,0x57dcU,0xafb8U,0x4f7bU,0x9ef6U,0x2de7U,0x5bceU,0xb79cU,
		0x7f33U,0xfe66U,0xecc7U,0xc985U,0x8301U,0x1609U,0x2c12U,0x5824U,0xb048U,0x709bU,
		0xe136U,0xd267U,0xb4c5U,0x7981U,0xf302U,0xf60fU,0xfc15U,0xe821U,0xc049U,0x9099U,
		0x3139U,0x6272U,0xc4e4U,0x99c3U,0x238dU,0x471aU,0x8e34U,0x0c63U,0x18c6U,0x318cU,
		0x6318U,0xc630U,0x9c6bU,0x28ddU,0x51baU,0xa374U,0x56e3U,0xadc6U,0x4b87U,0x970eU,
		0x3e17U,0x7c2eU,0xf85cU,0xe0b3U,0xd16dU,0xb2d1U,0x75a9U,0xeb52U,0xc6afU,0x9d55U,
		0x2aa1U,0x5542U,0xaa84U,0x4503U,0x8a06U,0x0407U,0x080eU,0x101cU,0x2038U,0x4070U,
		0x80e0U,0x11cbU,0x2396U,0x472cU,0x8e58U,0x0cbbU,0x1976U,0x32ecU,0x65d8U,0xcbb0U,
		0x876bU,0x1eddU,0x3dbaU,0x7b74U,0xf6e8U,0xfddbU,0xebbdU,0xc771U,0x9ee9U,0x2dd9U,
		0x5bb2U,0xb764U,0x7ec3U,0xfd86U,0xeb07U,0xc605U,0x9c01U,0x2809U,0x5012U,0xa024U,
		0x5043U,0xa086U,0x5107U,0xa20eU,0x5417U,0xa82eU,0x4057U,0x80aeU,0x1157U,0x22aeU,
		0x455cU,0x8ab8U,0x057bU,0x0af6U,0x15ecU,0x2bd8U,0x57b0U,0xaf60U,0x4ecbU,0x9d96U,
		0x2b27U,0x564eU,0xac9cU,0x4933U,0x9266U,0x34c7U,0x698eU,0xd31cU,0xb633U,0x7c6dU,
		0xf8daU,0xe1bfU,0xd375U,0xb6e1U,0x7dc9U,0xfb92U,0xe72fU,0xde55U,0xaca1U,0x4949U,
		0x9292U,0x352fU,0x6a5eU,0xd4bcU,0xb973U,0x62edU,0xc5daU,0x9bbfU,0x2775U,0x4eeaU,
		0x9dd4U,0x2ba3U,0x5746U,0xae8cU,0x4d13U,0x9a26U,0x2447U,0x488eU,0x911cU,0x3233U,
		0x6466U,0xc8ccU,0x8193U,0x132dU,0x265aU,0x4cb4U,0x9968U,0x22dbU,0x45b6U,0x8b6cU,
		0x06d3U,0x0da6U,0x1b4cU,0x3698U,0x6d30U,0xda60U,0xa4cbU,0x599dU,0xb33aU,0x767fU,
		0xecfeU,0xc9f7U,0x83e5U,0x17c1U,0x2f82U,0x5f04U,0xbe08U,0x6c1bU,0xd836U,0xa067U,
		0x50c5U,0xa18aU,0x531fU,0xa63eU,0x5c77U,0xb8eeU,0x61d7U,0xc3aeU,0x9757U,0x3ea5U,
		0x7d4aU,0xfa94U,0xe523U,0xda4dU,0xa491U,0x5929U,0xb252U,0x74afU,0xe95eU,0xc2b7U,
		0x9565U,0x3ac1U,0x7582U,0xeb04U,0xc603U,0x9c0dU,0x2811U,0x5022U,0xa044U,0x5083U,
		0xa106U,0x5207U,0xa40eU,0x5817U,0xb02eU,0x7057U,0xe0aeU,0xd157U,0xb2a5U,0x7541U,
		0xea82U,0xc50fU,0x9a15U,0x2421U,0x4842U,0x9084U,0x3103U,0x6206U,0xc40cU,0x9813U,
		0x202dU,0x405aU,0x80b4U,0x1163U,0x22c6U,0x458cU,0x8b18U,0x063bU,0x0c76U,0x18ecU,
		0x31d8U,0x63b0U,0xc760U,0x9ecbU,0x2d9dU,0x5b3aU,0xb674U,0x7ce3U,0xf9c6U,0xe387U,
		0xd705U,0xbe01U,0x6c09U,0xd812U,0xa02fU,0x5055U,0xa0aaU,0x515fU,0xa2beU,0x5577U,
		0xaaeeU,0x45d7U,0x8baeU,0x0757U,0x0eaeU,0x1d5cU,0x3ab8U,0x7570U,0xeae0U,0xc5cbU,
		0x9b9dU,0x2731U,0x4e62U,0x9cc4U,0x2983U,0x5306U,0xa60cU,0x5c13U,0xb826U,0x6047U,
		0xc08eU,0x9117U,0x3225U,0x644aU,0xc894U,0x8123U,0x124dU,0x249aU,0x4934U,0x9268U,
		0x34dbU,0x69b6U,0xd36cU,0xb6d3U,0x7dadU,0xfb5aU,0xe6bfU,0xdd75U,0xaae1U,0x45c9U,
		0x8b92U,0x072fU,0x0e5eU,0x1cbcU,0x3978U,0x72f0U,0xe5e0U,0xdbcbU,0xa79dU,0x5f31U,
		0xbe62U,0x6ccfU,0xd99eU,0xa337U,0x5665U,0xaccaU,0x499fU,0x933eU,0x3677U,0x6ceeU,
		0xd9dcU,0xa3b3U,0x576dU,0xaedaU,0x4dbfU,0x9b7eU,0x26f7U,0x4deeU,0x9bdcU,0x27b3U,
		0x4f66U,0x9eccU,0x2d93U,0x5b26U,0xb64cU,0x7c93U,0xf926U,0xe247U,0xd485U,0xb901U,
		0x6209U,0xc412U,0x982fU,0x2055U,0x40aaU,0x8154U,0x12a3U,0x2546U,0x4a8cU,0x9518U,
		0x3a3bU,0x7476U,0xe8ecU,0xc1d3U,0x93adU,0x3751U,0x6ea2U,0xdd44U,0xaa83U,0x450dU,
		0x8a1aU,0x043fU,0x087eU,0x10fcU,0x21f8U,0x43f0U,0x87e0U,0x1fcbU,0x3f96U,0x7f2cU,
		0xfe58U,0xecbbU,0xc97dU,0x82f1U,0x15e9U,0x2bd2U,0x57a4U,0xaf48U,0x4e9bU,0x9d36U,
		0x2a67U,0x54ceU,0xa99cU,0x4333U,0x8666U,0x1cc7U,0x398eU,0x731cU,0xe638U,0xdc7bU,
		0xa8fdU,0x41f1U,0x83e2U,0x17cfU,0x2f9eU,0x5f3cU,0xbe78U,0x6cfbU,0xd9f6U,0xa3e7U,
		0x57c5U,0xaf8aU,0x4f1fU,0x9e3eU,0x2c77U,0x58eeU,0xb1dcU,0x73b3U,0xe766U,0xdec7U,
		0xad85U,0x4b01U,0x9602U,0x3c0fU,0x781eU,0xf03cU,0xf073U,0xf0edU,0xf1d1U,0xf3a9U,
		0xf759U,0xfeb9U,0xed79U,0xcaf9U,0x85f9U,0x1bf9U,0x37f2U,0x6fe4U,0xdfc8U,0xaf9bU,
		0x4f3dU,0x9e7aU,0x2cffU,0x59feU,0xb3fcU,0x77f3U,0xefe6U,0xcfc7U,0x8f85U,0x0f01U,
		0x1e02U,0x3c04U,0x7808U,0xf010U,0xf02bU,0xf05dU,0xf0b1U,0xf169U,0xf2d9U,0xf5b9U,
		0xfb79U,0xe6f9U,0xddf9U,0xabf9U,0x47f9U,0x8ff2U,0x0fefU,0x1fdeU,0x3fbcU,0x7f78U,
		0xfef0U,0xedebU,0xcbddU,0x87b1U,0x1f69U,0x3ed2U,0x7da4U,0xfb48U,0xe69bU,0xdd3dU,
		0xaa71U,0x44e9U,0x89d2U,0x03afU,0x075eU,0x0ebcU,0x1d78U,0x3af0U,0x75e0U,0xebc0U,
		0xc78bU,0x9f1dU,0x2e31U,0x5c62U,0xb8c4U,0x6183U,0xc306U,0x9607U,0x3c05U,0x780aU,
		0xf014U,0xf023U,0xf04dU,0xf091U,0xf129U,0xf259U,0xf4b9U,0xf979U,0xe2f9U,0xd5f9U,
		0xbbf9U,0x67f9U,0xcff2U,0x8fefU,0x0fd5U,0x1faaU,0x3f54U,0x7ea8U,0xfd50U,0xeaabU,
		0xc55dU,0x9ab1U,0x2569U,0x4ad2U,0x95a4U,0x3b43U,0x7686U,0xed0cU,0xca13U,0x842dU,
		0x1851U,0x30a2U,0x6144U,0xc288U,0x951bU,0x3a3dU,0x747aU,0xe8f4U,0xc1e3U,0x93cdU,
		0x3791U,0x6f22U,0xde44U,0xac83U,0x490dU,0x921aU,0x343fU,0x687eU,0xd0fcU,0xb1f3U,
		0x73edU,0xe7daU,0xdfbfU,0xaf75U,0x4ee1U,0x9dc2U,0x2b8fU,0x571eU,0xae3cU,0x4c73U,
		0x98e6U,0x21c7U,0x438eU,0x871cU,0x1e33U,0x3c66U,0x78ccU,0xf198U,0xf33bU,0xf67dU,
		0xfcf1U,0xe9e9U,0xc3d9U,0x97b9U,0x3f79U,0x7ef2U,0xfde4U,0xebc3U,0xc78dU,0x9f11U,
		0x2e29U,0x5c52U,0xb8a4U,0x6143U,0xc286U,0x9507U,0x3a05U,0x740aU,0xe814U,0xc023U,
		0x904dU,0x3091U,0x6122U,0xc244U,0x9483U,0x390dU,0x721aU,0xe434U,0xd863U,0xa0cdU,
		0x5191U,0xa322U,0x564fU,0xac9eU,0x4937U,0x926eU,0x34d7U,0x69aeU,0xd35cU,0xb6b3U,
		0x7d6dU,0xfadaU,0xe5bfU,0xdb75U,0xa6e1U,0x5dc9U,0xbb92U,0x672fU,0xce5eU,0x8cb7U,
		0x0965U,0x12caU,0x2594U,0x4b28U,0x9650U,0x3cabU,0x7956U,0xf2acU,0xf553U,0xfaadU,
		0xe551U,0xdaa9U,0xa559U,0x5ab9U,0xb572U,0x7aefU,0xf5deU,0xfbb7U,0xe765U,0xdec1U,
		0xad89U,0x4b19U,0x9632U,0x3c6fU,0x78deU,0xf1bcU,0xf373U,0xf6edU,0xfdd1U,0xeba9U,
		0xc759U,0x9eb9U,0x2d79U,0x5af2U,0xb5e4U,0x7bc3U,0xf786U,0xff07U,0xee05U,0xcc01U,
		0x8809U,0x0019U,0x0032U,0x0064U,0x00c8U,0x0190U,0x0320U,0x0640U,0x0c80U,0x1900U,
		0x3200U,0x6400U,0xc800U,0x800bU,0x101dU,0x203aU,0x4074U,0x80e8U,0x11dbU,0x23b6U,
		0x476cU,0x8ed8U,0x0dbbU,0x1b76U,0x36ecU,0x6dd8U,0xdbb0U,0xa76bU,0x5eddU,0xbdbaU,
		0x6b7fU,0xd6feU,0xbdf7U,0x6be5U,0xd7caU,0xbf9fU,0x6f35U,0xde6aU,0xacdfU,0x49b5U,
		0x936aU,0x36dfU,0x6dbeU,0xdb7cU,0xa6f3U,0x5dedU,0xbbdaU,0x67bfU,0xcf7eU,0x8ef7U,
		0x0de5U,0x1bcaU,0x3794U,0x6f28U,0xde50U,0xacabU,0x495dU,0x92baU,0x357fU,0x6afeU,
		0xd5fcU,0xbbf3U,0x67edU,0xcfdaU,0x8fbfU,0x0f75U,0x1eeaU,0x3dd4U,0x7ba8U,0xf750U,
		0xfeabU,0xed5dU,0xcab1U,0x8569U,0x1ad9U,0x35b2U,0x6b64U,0xd6c8U,0xbd9bU,0x6b3dU,
		0xd67aU,0xbcffU,0x69f5U,0xd3eaU,0xb7dfU,0x7fb5U,0xff6aU,0xeedfU,0xcdb5U,0x8b61U,
		0x06c9U,0x0d92U,0x1b24U,0x3648U,0x6c90U,0xd920U,0xa24bU,0x549dU,0xa93aU,0x427fU,
		0x84feU,0x19f7U,0x33eeU,0x67dcU,0xcfb8U,0x8f7bU,0x0efdU,0x1dfaU,0x3bf4U,0x77e8U,
		0xefd0U,0xcfabU,0x8f5dU,0x0eb1U,0x1d62U,0x3ac4U,0x7588U,0xeb10U,0xc62bU,0x9c5dU,
		0x28b1U,0x5162U,0xa2c4U,0x5583U,0xab06U,0x4607U,0x8c0eU,0x0817U,0x102eU,0x205cU,
		0x40b8U,0x8170U,0x12ebU,0x25d6U,0x4bacU,0x9758U,0x3ebbU,0x7d76U,0xfaecU,0xe5d3U,
		0xdbadU,0xa751U,0x5ea9U,0xbd52U,0x6aafU,0xd55eU,0xbab7U,0x6565U,0xcacaU,0x859fU,
		0x1b35U,0x366aU,0x6cd4U,0xd9a8U,0xa35bU,0x56bdU,0xad7aU,0x4affU,0x95feU,0x3bf7U,
		0x77eeU,0xefdcU,0xcfb3U,0x8f6dU,0x0ed1U,0x1da2U,0x3b44U,0x7688U,0xed10U,0xca2bU,
		0x845dU,0x18b1U,0x3162U,0x62c4U,0xc588U,0x9b1bU,0x263dU,0x4c7aU,0x98f4U,0x21e3U,
		0x43c6U,0x878cU,0x1f13U,0x3e26U,0x7c4cU,0xf898U,0xe13bU,0xd27dU,0xb4f1U,0x79e9U,
		0xf3d2U,0xf7afU,0xff55U,0xeea1U,0xcd49U,0x8a99U,0x0539U,0x0a72U,0x14e4U,0x29c8U,
		0x5390U,0xa720U,0x5e4bU,0xbc96U,0x6927U,0xd24eU,0xb497U,0x7925U,0xf24aU,0xf49fU,
		0xf935U,0xe261U,0xd4c9U,0xb999U,0x6339U,0xc672U,0x9cefU,0x29d5U,0x53aaU,0xa754U,
		0x5ea3U,0xbd46U,0x6a87U,0xd50eU,0xba17U,0x6425U,0xc84aU,0x809fU,0x1135U,0x226aU,
		0x44d4U,0x89a8U,0x035bU,0x06b6U,0x0d6cU,0x1ad8U,0x35b0U,0x6b60U,0xd6c0U,0xbd8bU,
		0x6b1dU,0xd63aU,0xbc7fU,0x68f5U,0xd1eaU,0xb3dfU,0x77b5U,0xef6aU,0xcedfU,0x8db5U,
		0x0b61U,0x16c2U,0x2d84U,0x5b08U,0xb610U,0x7c2bU,0xf856U,0xe0a7U,0xd145U,0xb281U,
		0x7509U,0xea12U,0xc42fU,0x9855U,0x20a1U,0x4142U,0x8284U,0x1503U,0x2a06U,0x540cU,
		0xa818U,0x403bU,0x8076U,0x10e7U,0x21ceU,0x439cU,0x8738U,0x1e7bU,0x3cf6U,0x79ecU,
		0xf3d8U,0xf7bbU,0xff7dU,0xeef1U,0xcde9U,0x8bd9U,0x07b9U,0x0f72U,0x1ee4U,0x3dc8U,
		0x7b90U,0xf720U,0xfe4bU,0xec9dU,0xc931U,0x8269U,0x14d9U,0x29b2U,0x5364U,0xa6c8U,
		0x5d9bU,0xbb36U,0x6667U,0xccceU,0x8997U,0x0325U,0x064aU,0x0c94U,0x1928U,0x3250U,
		0x64a0U,0xc940U,0x828bU,0x151dU,0x2a3aU,0x5474U,0xa8e8U,0x41dbU,0x83b6U,0x1767U,
		0x2eceU,0x5d9cU,0xbb38U,0x667bU,0xccf6U,0x89e7U,0x03c5U,0x078aU,0x0f14U,0x1e28U,
		0x3c50U,0x78a0U,0xf140U,0xf28bU,0xf51dU,0xfa31U,0xe469U,0xd8d9U,0xa1b9U,0x5379U,
		0xa6f2U,0x5defU,0xbbdeU,0x67b7U,0xcf6eU,0x8ed7U,0x0da5U,0x1b4aU,0x3694U,0x6d28U,
		0xda50U,0xa4abU,0x595dU,0xb2baU,0x757fU,0xeafeU,0xc5f7U,0x9be5U,0x27c1U,0x4f82U,
		0x9f04U,0x2e03U,0x5c06U,0xb80cU,0x6013U,0xc026U,0x9047U,0x3085U,0x610aU,0xc214U,
		0x9423U,0x384dU,0x709aU,0xe134U,0xd263U,0xb4cdU,0x7991U,0xf322U,0xf64fU,0xfc95U,
		0xe921U,0xc249U,0x9499U,0x3939U,0x7272U,0xe4e4U,0xd9c3U,0xa38dU,0x5711U,0xae22U,
		0x4c4fU,0x989eU,0x2137U,0x426eU,0x84dcU,0x19b3U,0x3366U,0x66ccU,0xcd98U,0x8b3bU,
		0x067dU,0x0cfaU,0x19f4U,0x33e8U,0x67d0U,0xcfa0U,0x8f4bU,0x0e9dU,0x1d3aU,0x3a74U,
		0x74e8U,0xe9d0U,0xc3abU,0x975dU,0x3eb1U,0x7d62U,0xfac4U,0xe583U,0xdb0dU,0xa611U,
		0x5c29U,0xb852U,0x60afU,0xc15eU,0x92b7U,0x3565U,0x6acaU,0xd594U,0xbb23U,0x664dU,
		0xcc9aU,0x893fU,0x0275U,0x04eaU,0x09d4U,0x13a8U,0x2750U,0x4ea0U,0x9d40U,0x2a8bU,
		0x5516U,0xaa2cU,0x4453U,0x88a6U,0x0147U,0x028eU,0x051cU,0x0a38U,0x1470U,0x28e0U,
		0x51c0U,0xa380U,0x570bU,0xae16U,0x4c27U,0x984eU,0x2097U,0x412eU,0x825cU,0x14b3U,
		0x2966U,0x52ccU,0xa598U,0x5b3bU,0xb676U,0x7ce7U,0xf9ceU,0xe397U,0xd725U,0xbe41U,
		0x6c89U,0xd912U,0xa22fU,0x5455U,0xa8aaU,0x415fU,0x82beU,0x1577U,0x2aeeU,0x55dcU,
		0xabb8U,0x477bU,0x8ef6U,0x0de7U,0x1bceU,0x379cU,0x6f38U,0xde70U,0xacebU,0x49ddU,
		0x93baU,0x377fU,0x6efeU,0xddfcU,0xabf3U,0x47edU,0x8fdaU,0x0fbfU,0x1f7eU,0x3efcU,
		0x7df8U,0xfbf0U,0xe7ebU,0xdfddU,0xafb1U,0x4f69U,0x9ed2U,0x2dafU,0x5b5eU,0xb6bcU,
		0x7d73U,0xfae6U,0xe5c7U,0xdb85U,0xa701U,0x5e09U,0xbc12U,0x682fU,0xd05eU,0xb0b7U,
		0x7165U,0xe2caU,0xd59fU,0xbb35U,0x6661U,0xccc2U,0x898fU,0x0315U,0x062aU,0x0c54U,
		0x18a8U,0x3150U,0x62a0U,0xc540U,0x9a8bU,0x251dU,0x4a3aU,0x9474U,0x38e3U,0x71c6U,
		0xe38cU,0xd713U,0xbe2dU,0x6c51U,0xd8a2U,0xa14fU,0x5295U,0xa52aU,0x5a5fU,0xb4beU,
		0x7977U,0xf2eeU,0xf5d7U,0xfba5U,0xe741U,0xde89U,0xad19U,0x4a39U,0x9472U,0x38efU,
		0x71deU,0xe3bcU,0xd773U,0xbeedU,0x6dd1U,0xdba2U,0xa74fU,0x5e95U,0xbd2aU,0x6a5fU,
		0xd4beU,0xb977U,0x62e5U,0xc5caU,0x9b9fU,0x2735U,0x4e6aU,0x9cd4U,0x29a3U,0x5346U,
		0xa68cU,0x5d13U,0xba26U,0x6447U,0xc88eU,0x8117U,0x1225U,0x244aU,0x4894U,0x9128U,
		0x325bU,0x64b6U,0xc96cU,0x82d3U,0x15adU,0x2b5aU,0x56b4U,0xad68U,0x4adbU,0x95b6U,
		0x3b67U,0x76ceU,0xed9cU,0xcb33U,0x866dU,0x1cd1U,0x39a2U,0x7344U,0xe688U,0xdd1bU,
		0xaa3dU,0x4471U,0x88e2U,0x01cfU,0x039eU,0x073cU,0x0e78U,0x1cf0U,0x39e0U,0x73c0U,
		0xe780U,0xdf0bU,0xae1dU,0x4c31U,0x9862U,0x20cfU,0x419eU,0x833cU,0x1673U,0x2ce6U,
		0x59ccU,0xb398U,0x773bU,0xee76U,0xcce7U,0x89c5U,0x0381U,0x0702U,0x0e04U,0x1c08U,
		0x3810U,0x7020U,0xe040U,0xd08bU,0xb11dU,0x7231U,0xe462U,0xd8cfU,0xa195U,0x5321U,
		0xa642U,0x5c8fU,0xb91eU,0x6237U,0xc46eU,0x98d7U,0x21a5U,0x434aU,0x8694U,0x1d23U,
		0x3a46U,0x748cU,0xe918U,0xc23bU,0x947dU,0x38f1U,0x71e2U,0xe3c4U,0xd783U,0xbf0dU,
		0x6e11U,0xdc22U,0xa84fU,0x4095U,0x812aU,0x125fU,0x24beU,0x497cU,0x92f8U,0x35fbU,
		0x6bf6U,0xd7ecU,0xbfd3U,0x6fadU,0xdf5aU,0xaebfU,0x4d75U,0x9aeaU,0x25dfU,0x4bbeU,
		0x977cU,0x3ef3U,0x7de6U,0xfbccU,0xe793U,0xdf2dU,0xae51U,0x4ca9U,0x9952U,0x22afU,
		0x455eU,0x8abcU,0x0573U,0x0ae6U,0x15ccU,0x2b98U,0x5730U,0xae60U,0x4ccbU,0x9996U,
		0x2327U,0x464eU,0x8c9cU,0x0933U,0x1266U,0x24ccU,0x4998U,0x9330U,0x366bU,0x6cd6U,
		0xd9acU,0xa353U,0x56adU,0xad5aU,0x4abfU,0x957eU,0x3af7U,0x75eeU,0xebdcU,0xc7b3U,
		0x9f6dU,0x2ed1U,0x5da2U,0xbb44U,0x6683U,0xcd06U,0x8a07U,0x0405U,0x080aU,0x1014U,
		0x2028U,0x4050U,0x80a0U,0x114bU,0x2296U,0x452cU,0x8a58U,0x04bbU,0x0976U,0x12ecU,
		0x25d8U,0x4bb0U,0x9760U,0x3ecbU,0x7d96U,0xfb2cU,0xe653U,0xdcadU,0xa951U,0x42a9U,
		0x8552U,0x1aafU,0x355eU,0x6abcU,0xd578U,0xbafbU,0x65fdU,0xcbfaU,0x87ffU,0x1ff5U,
		0x3feaU,0x7fd4U,0xffa8U,0xef5bU,0xcebdU,0x8d71U,0x0ae9U,0x15d2U,0x2ba4U,0x5748U,
		0xae90U,0x4d2bU,0x9a56U,0x24a7U,0x494eU,0x929cU,0x3533U,0x6a66U,0xd4ccU,0xb993U,
		0x632dU,0xc65aU,0x9cbfU,0x2975U,0x52eaU,0xa5d4U,0x5ba3U,0xb746U,0x7e87U,0xfd0eU,
		0xea17U,0xc425U,0x9841U,0x2089U,0x4112U,0x8224U,0x1443U,0x2886U,0x510cU,0xa218U,
		0x543bU,0xa876U,0x40e7U,0x81ceU,0x1397U,0x272eU,0x4e5cU,0x9cb8U,0x297bU,0x52f6U,
		0xa5ecU,0x5bd3U,0xb7a6U,0x7f47U,0xfe8eU,0xed17U,0xca25U,0x8441U,0x1889U,0x3112U,
		0x6224U,0xc448U,0x989bU,0x213dU,0x427aU,0x84f4U,0x19e3U,0x33c6U,0x678cU,0xcf18U,
		0x8e3bU,0x0c7dU,0x18faU,0x31f4U,0x63e8U,0xc7d0U,0x9fabU,0x2f5dU,0x5ebaU,0xbd74U,
		0x6ae3U,0xd5c6U,0xbb87U,0x6705U,0xce0aU,0x8c1fU,0x0835U,0x106aU,0x20d4U,0x41a8U,
		0x8350U,0x16abU,0x2d56U,0x5aacU,0xb558U,0x7abbU,0xf576U,0xfae7U,0xe5c5U,0xdb81U,
		0xa709U,0x5e19U,0xbc32U,0x686fU,0xd0deU,0xb1b7U,0x7365U,0xe6caU,0xdd9fU,0xab35U,
		0x4661U,0x8cc2U,0x098fU,0x131eU,0x263cU,0x4c78U,0x98f0U,0x21ebU,0x43d6U,0x87acU,
		0x1f53U,0x3ea6U,0x7d4cU,0xfa98U,0xe53bU,0xda7dU,0xa4f1U,0x59e9U,0xb3d2U,0x77afU,
		0xef5eU,0xceb7U,0x8d65U,0x0ac1U,0x1582U,0x2b04U,0x5608U,0xac10U,0x482bU,0x9056U,
		0x30a7U,0x614eU,0xc29cU,0x9533U,0x3a6dU,0x74daU,0xe9b4U,0xc363U,0x96cdU,0x3d91U,
		0x7b22U,0xf644U,0xfc83U,0xe90dU,0xc211U,0x9429U,0x3859U,0x70b2U,0xe164U,0xd2c3U,
		0xb58dU,0x7b11U,0xf622U,0xfc4fU,0xe895U,0xc121U,0x9249U,0x3499U,0x6932U,0xd264U,
		0xb4c3U,0x798dU,0xf31aU,0xf63fU,0xfc75U,0xe8e1U,0xc1c9U,0x9399U,0x3739U,0x6e72U,
		0xdce4U,0xa9c3U,0x438dU,0x871aU,0x1e3fU,0x3c7eU,0x78fcU,0xf1f8U,0xf3fbU,0xf7fdU,
		0xfff1U,0xefe9U,0xcfd9U,0x8fb9U,0x0f79U,0x1ef2U,0x3de4U,0x7bc8U,0xf790U,0xff2bU,
		0xee5dU,0xccb1U,0x8969U,0x02d9U,0x05b2U,0x0b64U,0x16c8U,0x2d90U,0x5b20U,0xb640U,
		0x7c8bU,0xf916U,0xe227U,0xd445U,0xb881U,0x6109U,0xc212U,0x942fU,0x3855U,0x70aaU,
		0xe154U,0xd2a3U,0xb54dU,0x7a91U,0xf522U,0xfa4fU,0xe495U,0xd921U,0xa249U,0x5499U,
		0xa932U,0x426fU,0x84deU,0x19b7U,0x336eU,0x66dcU,0xcdb8U,0x8b7bU,0x06fdU,0x0dfaU,
		0x1bf4U,0x37e8U,0x6fd0U,0xdfa0U,0xaf4bU,0x4e9dU,0x9d3aU,0x2a7fU,0x54feU,0xa9fcU,
		0x43f3U,0x87e6U,0x1fc7U,0x3f8eU,0x7f1cU,0xfe38U,0xec7bU,0xc8fdU,0x81f1U,0x13e9U,
		0x27d2U,0x4fa4U,0x9f48U,0x2e9bU,0x5d36U,0xba6cU,0x64d3U,0xc9a6U,0x8347U,0x1685U,
		0x2d0aU,0x5a14U,0xb428U,0x785bU,0xf0b6U,0xf167U,0xf2c5U,0xf581U,0xfb09U,0xe619U,
		0xdc39U,0xa879U,0x40f9U,0x81f2U,0x13efU,0x27deU,0x4fbcU,0x9f78U,0x2efbU,0x5df6U,
		0xbbecU,0x67d3U,0xcfa6U,0x8f47U,0x0e85U,0x1d0aU,0x3a14U,0x7428U,0xe850U,0xc0abU,
		0x915dU,0x32b1U,0x6562U,0xcac4U,0x8583U,0x1b0dU,0x361aU,0x6c34U,0xd868U,0xa0dbU,
		0x51bdU,0xa37aU,0x56ffU,0xadfeU,0x4bf7U,0x97eeU,0x3fd7U,0x7faeU,0xff5cU,0xeeb3U,
		0xcd6dU,0x8ad1U,0x05a9U,0x0b52U,0x16a4U,0x2d48U,0x5a90U,0xb520U,0x7a4bU,0xf496U,
		0xf927U,0xe245U,0xd481U,0xb909U,0x6219U,0xc432U,0x986fU,0x20d5U,0x41aaU,0x8354U,
		0x16a3U,0x2d46U,0x5a8cU,0xb518U,0x7a3bU,0xf476U,0xf8e7U,0xe1c5U,0xd381U,0xb709U,
		0x7e19U,0xfc32U,0xe86fU,0xc0d5U,0x91a1U,0x3349U,0x6692U,0xcd24U,0x8a43U,0x048dU,
		0x091aU,0x1234U,0x2468U,0x48d0U,0x91a0U,0x334bU,0x6696U,0xcd2cU,0x8a53U,0x04adU,
		0x095aU,0x12b4U,0x2568U,0x4ad0U,0x95a0U,0x3b4bU,0x7696U,0xed2cU,0xca53U,0x84adU,
		0x1951U,0x32a2U,0x6544U,0xca88U,0x851bU,0x1a3dU,0x347aU,0x68f4U,0xd1e8U,0xb3dbU,
		0x77bdU,0xef7aU,0xceffU,0x8df5U,0x0be1U,0x17c2U,0x2f84U,0x5f08U,0xbe10U,0x6c2bU,
		0xd856U,0xa0a7U,0x5145U,0xa28aU,0x551fU,0xaa3eU,0x4477U,0x88eeU,0x01d7U,0x03aeU,
		0x075cU,0x0eb8U,0x1d70U,0x3ae0U,0x75c0U,0xeb80U,0xc70bU,0x9e1dU,0x2c31U,0x5862U,
		0xb0c4U,0x7183U,0xe306U,0xd607U,0xbc05U,0x6801U,0xd002U,0xb00fU,0x7015U,0xe02aU,
		0xd05fU,0xb0b5U,0x7161U,0xe2c2U,0xd58fU,0xbb15U,0x6621U,0xcc42U,0x888fU,0x0115U,
		0x022aU,0x0454U,0x08a8U,0x1150U,0x22a0U,0x4540U,0x8a80U,0x050bU,0x0a16U,0x142cU,
		0x2858U,0x50b0U,0xa160U,0x52cbU,0xa596U,0x5b27U,0xb64eU,0x7c97U,0xf92eU,0xe257U,
		0xd4a5U,0xb941U,0x6289U,0xc512U,0x9a2fU,0x2455U,0x48aaU,0x9154U,0x32a3U,0x6546U,
		0xca8cU,0x8513U,0x1a2dU,0x345aU,0x68b4U,0xd168U,0xb2dbU,0x75bdU,0xeb7aU,0xc6ffU,
		0x9df5U,0x2be1U,0x57c2U,0xaf84U,0x4f03U,0x9e06U,0x2c07U,0x580eU,0xb01cU,0x7033U,
		0xe066U,0xd0c7U,0xb185U,0x7301U,0xe602U,0xdc0fU,0xa815U,0x4021U,0x8042U,0x108fU,
		0x211eU,0x423cU,0x8478U,0x18fbU,0x31f6U,0x63ecU,0xc7d8U,0x9fbbU,0x2f7dU,0x5efaU,
		0xbdf4U,0x6be3U,0xd7c6U,0xbf87U,0x6f05U,0xde0aU,0xac1fU,0x4835U,0x906aU,0x30dfU,
		0x61beU,0xc37cU,0x96f3U,0x3dedU,0x7bdaU,0xf7b4U,0xff63U,0xeecdU,0xcd91U,0x8b29U,
		0x0659U,0x0cb2U,0x1964U,0x32c8U,0x6590U,0xcb20U,0x864bU,0x1c9dU,0x393aU,0x7274U,
		0xe4e8U,0xd9dbU,0xa3bdU,0x5771U,0xaee2U,0x4dcfU,0x9b9eU,0x2737U,0x4e6eU,0x9cdcU,
		0x29b3U,0x5366U,0xa6ccU,0x5d93U,0xbb26U,0x6647U,0xcc8eU,0x8917U,0x0225U,0x044aU,
		0x0894U,0x1128U,0x2250U,0x44a0U,0x8940U,0x028bU,0x0516U,0x0a2cU,0x1458U,0x28b0U,
		0x5160U,0xa2c0U,0x558bU,0xab16U,0x4627U,0x8c4eU,0x0897U,0x112eU,0x225cU,0x44b8U,
		0x8970U,0x02ebU,0x05d6U,0x0bacU,0x1758U,0x2eb0U,0x5d60U,0xbac0U,0x658bU,0xcb16U,
		0x8627U,0x1c45U,0x388aU,0x7114U,0xe228U,0xd45bU,0xb8bdU,0x6171U,0xc2e2U,0x95cfU,
		0x3b95U,0x772aU,0xee54U,0xcca3U,0x894dU,0x0291U,0x0522U,0x0a44U,0x1488U,0x2910U,
		0x5220U,0xa440U,0x588bU,0xb116U,0x7227U,0xe44eU,0xd897U,0xa125U,0x5241U,0xa482U,
		0x590fU,0xb21eU,0x7437U,0xe86eU,0xc0d7U,0x91a5U,0x3341U,0x6682U,0xcd04U,0x8a03U,
		0x040dU,0x081aU,0x1034U,0x2068U,0x40d0U,0x81a0U,0x134bU,0x2696U,0x4d2cU,0x9a58U,
		0x24bbU,0x4976U,0x92ecU,0x35d3U,0x6ba6U,0xd74cU,0xbe93U,0x6d2dU,0xda5aU,0xa4bfU,
		0x5975U,0xb2eaU,0x75dfU,0xebbeU,0xc777U,0x9ee5U,0x2dc1U,0x5b82U,0xb704U,0x7e03U,
		0xfc06U,0xe807U,0xc005U,0x9001U,0x3009U,0x6012U,0xc024U,0x9043U,0x308dU,0x611aU,
		0xc234U,0x9463U,0x38cdU,0x719aU,0xe334U,0xd663U,0xbccdU,0x6991U,0xd322U,0xb64fU,
		0x7c95U,0xf92aU,0xe25fU,0xd4b5U,0xb961U,0x62c9U,0xc592U,0x9b2fU,0x2655U,0x4caaU,
		0x9954U,0x22a3U,0x4546U,0x8a8cU,0x0513U,0x0a26U,0x144cU,0x2898U,0x5130U,0xa260U,
		0x54cbU,0xa996U,0x4327U,0x864eU,0x1c97U,0x392eU,0x725cU,0xe4b8U,0xd97bU,0xa2fdU,
		0x55f1U,0xabe2U,0x47cfU,0x8f9eU,0x0f37U,0x1e6eU,0x3cdcU,0x79b8U,0xf370U,0xf6ebU,
		0xfdddU,0xebb1U,0xc769U,0x9ed9U,0x2db9U,0x5b72U,0xb6e4U,0x7dc3U,0xfb86U,0xe707U,
		0xde05U,0xac01U,0x4809U,0x9012U,0x302fU,0x605eU,0xc0bcU,0x9173U,0x32edU,0x65daU,
		0xcbb4U,0x8763U,0x1ecdU,0x3d9aU,0x7b34U,0xf668U,0xfcdbU,0xe9bdU,0xc371U,0x96e9U,
		0x3dd9U,0x7bb2U,0xf764U,0xfec3U,0xed8dU,0xcb11U,0x8629U,0x1c59U,0x38b2U,0x7164U,
		0xe2c8U,0xd59bU,0xbb3dU,0x6671U,0xcce2U,0x89cfU,0x0395U,0x072aU,0x0e54U,0x1ca8U,
		0x3950U,0x72a0U,0xe540U,0xda8bU,0xa51dU,0x5a31U,0xb462U,0x78cfU,0xf19eU,0xf337U,
		0xf665U,0xfcc1U,0xe989U,0xc319U,0x9639U,0x3c79U,0x78f2U,0xf1e4U,0xf3c3U,0xf78dU,
		0xff11U,0xee29U,0xcc59U,0x88b9U,0x0179U,0x02f2U,0x05e4U,0x0bc8U,0x1790U,0x2f20U,
		0x5e40U,0xbc80U,0x690bU,0xd216U,0xb427U,0x7845U,0xf08aU,0xf11fU,0xf235U,0xf461U,
		0xf8c9U,0xe199U,0xd339U,0xb679U,0x7cf9U,0xf9f2U,0xe3efU,0xd7d5U,0xbfa1U,0x6f49U,
		0xde92U,0xad2fU,0x4a55U,0x94aaU,0x395fU,0x72beU,0xe57cU,0xdaf3U,0xa5edU,0x5bd1U,
		0xb7a2U,0x7f4fU,0xfe9eU,0xed37U,0xca65U,0x84c1U,0x1989U,0x3312U,0x6624U,0xcc48U,
		0x889bU,0x013dU,0x027aU,0x04f4U,0x09e8U,0x13d0U,0x27a0U,0x4f40U,0x9e80U,0x2d0bU,
		0x5a16U,0xb42cU,0x7853U,0xf0a6U,0xf147U,0xf285U,0xf501U,0xfa09U,0xe419U,0xd839U,
		0xa079U,0x50f9U,0xa1f2U,0x53efU,0xa7deU,0x5fb7U,0xbf6eU,0x6ed7U,0xddaeU,0xab57U,
		0x46a5U,0x8d4aU,0x0a9fU,0x153eU,0x2a7cU,0x54f8U,0xa9f0U,0x43ebU,0x87d6U,0x1fa7U,
		0x3f4eU,0x7e9cU,0xfd38U,0xea7bU,0xc4fdU,0x99f1U,0x23e9U,0x47d2U,0x8fa4U,0x0f43U,
		0x1e86U,0x3d0cU,0x7a18U,0xf430U,0xf86bU,0xe0ddU,0xd1b1U,0xb369U,0x76d9U,0xedb2U,
		0xcb6fU,0x86d5U,0x1da1U,0x3b42U,0x7684U,0xed08U,0xca1bU,0x843dU,0x1871U,0x30e2U,
		0x61c4U,0xc388U,0x971bU,0x3e3dU,0x7c7aU,0xf8f4U,0xe1e3U,0xd3cdU,0xb791U,0x7f29U,
		0xfe52U,0xecafU,0xc955U,0x82a1U,0x1549U,0x2a92U,0x5524U,0xaa48U,0x449bU,0x8936U,
		0x0267U,0x04ceU,0x099cU,0x1338U,0x2670U,0x4ce0U,0x99c0U,0x238bU,0x4716U,0x8e2cU,
		0x0c53U,0x18a6U,0x314cU,0x6298U,0xc530U,0x9a6bU,0x24ddU,0x49baU,0x9374U,0x36e3U,
		0x6dc6U,0xdb8cU,0xa713U,0x5e2dU,0xbc5aU,0x68bfU,0xd17eU,0xb2f7U,0x75e5U,0xebcaU,
		0xc79fU,0x9f35U,0x2e61U,0x5cc2U,0xb984U,0x6303U,0xc606U,0x9c07U,0x2805U,0x500aU,
		0xa014U,0x5023U,0xa046U,0x5087U,0xa10eU,0x5217U,0xa42eU,0x5857U,0xb0aeU,0x7157U,
		0xe2aeU,0xd557U,0xbaa5U,0x6541U,0xca82U,0x850fU,0x1a15U,0x342aU,0x6854U,0xd0a8U,
		0xb15bU,0x72bdU,0xe57aU,0xdaffU,0xa5f5U,0x5be1U,0xb7c2U,0x7f8fU,0xff1eU,0xee37U,
		0xcc65U,0x88c1U,0x0189U,0x0312U,0x0624U,0x0c48U,0x1890U,0x3120U,0x6240U,0xc480U,
		0x990bU,0x221dU,0x443aU,0x8874U,0x00e3U,0x01c6U,0x038cU,0x0718U,0x0e30U,0x1c60U,
		0x38c0U,0x7180U,0xe300U,0xd60bU,0xbc1dU,0x6831U,0xd062U,0xb0cfU,0x7195U,0xe32aU,
		0xd65fU,0xbcb5U,0x6961U,0xd2c2U,0xb58fU,0x7b15U,0xf62aU,0xfc5fU,0xe8b5U,0xc161U,
		0x92c9U,0x3599U,0x6b32U,0xd664U,0xbcc3U,0x698dU,0xd31aU,0xb63fU,0x7c75U,0xf8eaU,
		0xe1dfU,0xd3b5U,0xb761U,0x7ec9U,0xfd92U,0xeb2fU,0xc655U,0x9ca1U,0x2949U,0x5292U,
		0xa524U,0x5a43U,0xb486U,0x7907U,0xf20eU,0xf417U,0xf825U,0xe041U,0xd089U,0xb119U,
		0x7239U,0xe472U,0xd8efU,0xa1d5U,0x53a1U,0xa742U,0x5e8fU,0xbd1eU,0x6a37U,0xd46eU,
		0xb8d7U,0x61a5U,0xc34aU,0x969fU,0x3d35U,0x7a6aU,0xf4d4U,0xf9a3U,0xe34dU,0xd691U,
		0xbd29U,0x6a59U,0xd4b2U,0xb96fU,0x62d5U,0xc5aaU,0x9b5fU,0x26b5U,0x4d6aU,0x9ad4U,
		0x25a3U,0x4b46U,0x968cU,0x3d13U,0x7a26U,0xf44cU,0xf893U,0xe12dU,0xd251U,0xb4a9U,
		0x7959U,0xf2b2U,0xf56fU,0xfad5U,0xe5a1U,0xdb49U,0xa699U,0x5d39U,0xba72U,0x64efU,
		0xc9deU,0x83b7U,0x1765U,0x2ecaU,0x5d94U,0xbb28U,0x665bU,0xccb6U,0x8967U,0x02c5U,
		0x058aU,0x0b14U,0x1628U,0x2c50U,0x58a0U,0xb140U,0x728bU,0xe516U,0xda27U,0xa445U,
		0x5881U,0xb102U,0x720fU,0xe41eU,0xd837U,0xa065U,0x50c1U,0xa182U,0x530fU,0xa61eU,
		0x5c37U,0xb86eU,0x60d7U,0xc1aeU,0x9357U,0x36a5U,0x6d4aU,0xda94U,0xa523U,0x5a4dU,
		0xb49aU,0x793fU,0xf27eU,0xf4f7U,0xf9e5U,0xe3c1U,0xd789U,0xbf19U,0x6e39U,0xdc72U,
		0xa8efU,0x41d5U,0x83aaU,0x175fU,0x2ebeU,0x5d7cU,0xbaf8U,0x65fbU,0xcbf6U,0x87e7U,
		0x1fc5U,0x3f8aU,0x7f14U,0xfe28U,0xec5bU,0xc8bdU,0x8171U,0x12e9U,0x25d2U,0x4ba4U,
		0x9748U,0x3e9bU,0x7d36U,0xfa6cU,0xe4d3U,0xd9adU,0xa351U,0x56a9U,0xad52U,0x4aafU,
		0x955eU,0x3ab7U,0x756eU,0xeadcU,0xc5b3U,0x9b6dU,0x26d1U,0x4da2U,0x9b44U,0x2683U,
		0x4d06U,0x9a0cU,0x2413U,0x4826U,0x904cU,0x3093U,0x6126U,0xc24cU,0x9493U,0x392dU,
		0x725aU,0xe4b4U,0xd963U,0xa2cdU,0x5591U,0xab22U,0x464fU,0x8c9eU,0x0937U,0x126eU,
		0x24dcU,0x49b8U,0x9370U,0x36ebU,0x6dd6U,0xdbacU,0xa753U,0x5eadU,0xbd5aU,0x6abfU,
		0xd57eU,0xbaf7U,0x65e5U,0xcbcaU,0x879fU,0x1f35U,0x3e6aU,0x7cd4U,0xf9a8U,0xe35bU,
		0xd6bdU,0xbd71U,0x6ae9U,0xd5d2U,0xbbafU,0x6755U,0xceaaU,0x8d5fU,0x0ab5U,0x156aU,
		0x2ad4U,0x55a8U,0xab50U,0x46abU,0x8d56U,0x0aa7U,0x154eU,0x2a9cU,0x5538U,0xaa70U,
		0x44ebU,0x89d6U,0x03a7U,0x074eU,0x0e9cU,0x1d38U,0x3a70U,0x74e0U,0xe9c0U,0xc38bU,
		0x971dU,0x3e31U,0x7c62U,0xf8c4U,0xe183U,0xd30dU,0xb611U,0x7c29U,0xf852U,0xe0afU,
		0xd155U,0xb2a1U,0x7549U,0xea92U,0xc52fU,0x9a55U,0x24a1U,0x4942U,0x9284U,0x3503U,
		0x6a06U,0xd40cU,0xb813U,0x602dU,0xc05aU,0x90bfU,0x3175U,0x62eaU,0xc5d4U,0x9ba3U,
		0x274dU,0x4e9aU,0x9d34U,0x2a63U,0x54c6U,0xa98cU,0x4313U,0x8626U,0x1c47U,0x388eU,
		0x711cU,0xe238U,0xd47bU,0xb8fdU,0x61f1U,0xc3e2U,0x97cfU,0x3f95U,0x7f2aU,0xfe54U,
		0xeca3U,0xc94dU,0x8291U,0x1529U,0x2a52U,0x54a4U,0xa948U,0x429bU,0x8536U,0x1a67U,
		0x34ceU,0x699cU,0xd338U,0xb67bU,0x7cfdU,0xf9faU,0xe3ffU,0xd7f5U,0xbfe1U,0x6fc9U,
		0xdf92U,0xaf2fU,0x4e55U,0x9caaU,0x295fU,0x52beU,0xa57cU,0x5af3U,0xb5e6U,0x7bc7U,
		0xf78eU,0xff17U,0xee25U,0xcc41U,0x8889U,0x0119U,0x0232U,0x0464U,0x08c8U,0x1190U,
		0x2320U,0x4640U,0x8c80U,0x090bU,0x1216U,0x242cU,0x4858U,0x90b0U,0x316bU,0x62d6U,
		0xc5acU,0x9b53U,0x26adU,0x4d5aU,0x9ab4U,0x2563U,0x4ac6U,0x958cU,0x3b13U,0x7626U,
		0xec4cU,0xc893U,0x812dU,0x1251U,0x24a2U,0x4944U,0x9288U,0x351bU,0x6a36U,0xd46cU,
		0xb8d3U,0x61adU,0xc35aU,0x96bfU,0x3d75U,0x7aeaU,0xf5d4U,0xfba3U,0xe74dU,0xde91U,
		0xad29U,0x4a59U,0x94b2U,0x396fU,0x72deU,0xe5bcU,0xdb73U,0xa6edU,0x5dd1U,0xbba2U,
		0x674fU,0xce9eU,0x8d37U,0x0a65U,0x14caU,0x2994U,0x5328U,0xa650U,0x5cabU,0xb956U,
		0x62a7U,0xc54eU,0x9a97U,0x2525U,0x4a4aU,0x9494U,0x3923U,0x7246U,0xe48cU,0xd913U,
		0xa22dU,0x5451U,0xa8a2U,0x414fU,0x829eU,0x1537U,0x2a6eU,0x54dcU,0xa9b8U,0x437bU,
		0x86f6U,0x1de7U,0x3bceU,0x779cU,0xef38U,0xce7bU,0x8cfdU,0x09f1U,0x13e2U,0x27c4U,
		0x4f88U,0x9f10U,0x2e2bU,0x5c56U,0xb8acU,0x6153U,0xc2a6U,0x9547U,0x3a85U,0x750aU,
		0xea14U,0xc423U,0x984dU,0x2091U,0x4122U,0x8244U,0x1483U,0x2906U,0x520cU,0xa418U,
		0x583bU,0xb076U,0x70e7U,0xe1ceU,0xd397U,0xb725U,0x7e41U,0xfc82U,0xe90fU,0xc215U,
		0x9421U,0x3849U,0x7092U,0xe124U,0xd243U,0xb48dU,0x7911U,0xf222U,0xf44fU,0xf895U,
		0xe121U,0xd249U,0xb499U,0x7939U,0xf272U,0xf4efU,0xf9d5U,0xe3a1U,0xd749U,0xbe99U,
		0x6d39U,0xda72U,0xa4efU,0x59d5U,0xb3aaU,0x775fU,0xeebeU,0xcd77U,0x8ae5U,0x05c1U,
		0x0b82U,0x1704U,0x2e08U,0x5c10U,0xb820U,0x604bU,0xc096U,0x9127U,0x3245U,0x648aU,
		0xc914U,0x8223U,0x144dU,0x289aU,0x5134U,0xa268U,0x54dbU,0xa9b6U,0x4367U,0x86ceU,
		0x1d97U,0x3b2eU,0x765cU,0xecb8U,0xc97bU,0x82fdU,0x15f1U,0x2be2U,0x57c4U,0xaf88U,
		0x4f1bU,0x9e36U,0x2c67U,0x58ceU,0xb19cU,0x7333U,0xe666U,0xdcc7U,0xa985U,0x4301U,
		0x8602U,0x1c0fU,0x381eU,0x703cU,0xe078U,0xd0fbU,0xb1fdU,0x73f1U,0xe7e2U,0xdfcfU,
		0xaf95U,0x4f21U,0x9e42U,0x2c8fU,0x591eU,0xb23cU,0x7473U,0xe8e6U,0xc1c7U,0x9385U,
		0x3701U,0x6e02U,0xdc04U,0xa803U,0x400dU,0x801aU,0x103fU,0x207eU,0x40fcU,0x81f8U,
		0x13fbU,0x27f6U,0x4fecU,0x9fd8U,0x2fbbU,0x5f76U,0xbeecU,0x6dd3U,0xdba6U,0xa747U,
		0x5e85U,0xbd0aU,0x6a1fU,0xd43eU,0xb877U,0x60e5U,0xc1caU,0x939fU,0x3735U,0x6e6aU,
		0xdcd4U,0xa9a3U,0x434dU,0x869aU,0x1d3fU,0x3a7eU,0x74fcU,0xe9f8U,0xc3fbU,0x97fdU,
		0x3ff1U,0x7fe2U,0xffc4U,0xef83U,0xcf0dU,0x8e11U,0x0c29U,0x1852U,0x30a4U,0x6148U,
		0xc290U,0x952bU,0x3a5dU,0x74baU,0xe974U,0xc2e3U,0x95cdU,0x3b91U,0x7722U,0xee44U,
		0xcc83U,0x890dU,0x0211U,0x0422U,0x0844U,0x1088U,0x2110U,0x4220U,0x8440U,0x188bU,
		0x3116U,0x622cU,0xc458U,0x98bbU,0x217dU,0x42faU,0x85f4U,0x1be3U,0x37c6U,0x6f8cU,
		0xdf18U,0xae3bU,0x4c7dU,0x98faU,0x21ffU,0x43feU,0x87fcU,0x1ff3U,0x3fe6U,0x7fccU,
		0xff98U,0xef3bU,0xce7dU,0x8cf1U,0x09e9U,0x13d2U,0x27a4U,0x4f48U,0x9e90U,0x2d2bU,
		0x5a56U,0xb4acU,0x7953U,0xf2a6U,0xf547U,0xfa85U,0xe501U,0xda09U,0xa419U,0x5839U,
		0xb072U,0x70efU,0xe1deU,0xd3b7U,0xb765U,0x7ec1U,0xfd82U,0xeb0fU,0xc615U,0x9c21U,
		0x2849U,0x5092U,0xa124U,0x5243U,0xa486U,0x5907U,0xb20eU,0x7417U,0xe82eU,0xc057U,
		0x90a5U,0x3141U,0x6282U,0xc504U,0x9a03U,0x240dU,0x481aU,0x9034U,0x3063U,0x60c6U,
		0xc18cU,0x9313U,0x362dU,0x6c5aU,0xd8b4U,0xa163U,0x52cdU,0xa59aU,0x5b3fU,0xb67eU,
		0x7cf7U,0xf9eeU,0xe3d7U,0xd7a5U,0xbf41U,0x6e89U,0xdd12U,0xaa2fU,0x4455U,0x88aaU,
		0x015fU,0x02beU,0x057cU,0x0af8U,0x15f0U,0x2be0U,0x57c0U,0xaf80U,0x4f0bU,0x9e16U,
		0x2c27U,0x584eU,0xb09cU,0x7133U,0xe266U,0xd4c7U,0xb985U,0x6301U,0xc602U,0x9c0fU,
		0x2815U,0x502aU,0xa054U,0x50a3U,0xa146U,0x5287U,0xa50eU,0x5a17U,0xb42eU,0x7857U,
		0xf0aeU,0xf157U,0xf2a5U,0xf541U,0xfa89U,0xe519U,0xda39U,0xa479U,0x58f9U,0xb1f2U,
		0x73efU,0xe7deU,0xdfb7U,0xaf65U,0x4ec1U,0x9d82U,0x2b0fU,0x561eU,0xac3cU,0x4873U,
		0x90e6U,0x31c7U,0x638eU,0xc71cU,0x9e33U,0x2c6dU,0x58daU,0xb1b4U,0x7363U,0xe6c6U,
		0xdd87U,0xab05U,0x4601U,0x8c02U,0x080fU,0x101eU,0x203cU,0x4078U,0x80f0U,0x11ebU,
		0x23d6U,0x47acU,0x8f58U,0x0ebbU,0x1d76U,0x3aecU,0x75d8U,0xebb0U,0xc76bU,0x9eddU,
		0x2db1U,0x5b62U,0xb6c4U,0x7d83U,0xfb06U,0xe607U,0xdc05U,0xa801U,0x4009U,0x8012U,
		0x102fU,0x205eU,0x40bcU,0x8178U,0x12fbU,0x25f6U,0x4becU,0x97d8U,0x3fbbU,0x7f76U,
		0xfeecU,0xedd3U,0xcbadU,0x8751U,0x1ea9U,0x3d52U,0x7aa4U,0xf548U,0xfa9bU,0xe53dU,
		0xda71U,0xa4e9U,0x59d9U,0xb3b2U,0x776fU,0xeedeU,0xcdb7U,0x8b65U,0x06c1U,0x0d82U,
		0x1b04U,0x3608U,0x6c10U,0xd820U,0xa04bU,0x509dU,0xa13aU,0x527fU,0xa4feU,0x59f7U,
		0xb3eeU,0x77d7U,0xefaeU,0xcf57U,0x8ea5U,0x0d41U,0x1a82U,0x3504U,0x6a08U,0xd410U,
		0xb82bU,0x605dU,0xc0baU,0x917fU,0x32f5U,0x65eaU,0xcbd4U,0x87a3U,0x1f4dU,0x3e9aU,
		0x7d34U,0xfa68U,0xe4dbU,0xd9bdU,0xa371U,0x56e9U,0xadd2U,0x4bafU,0x975eU,0x3eb7U,
		0x7d6eU,0xfadcU,0xe5b3U,0xdb6dU,0xa6d1U,0x5da9U,0xbb52U,0x66afU,0xcd5eU,0x8ab7U,
		0x0565U,0x0acaU,0x1594U,0x2b28U,0x5650U,0xaca0U,0x494bU,0x9296U,0x3527U,0x6a4eU,
		0xd49cU,0xb933U,0x626dU,0xc4daU,0x99bfU,0x2375U,0x46eaU,0x8dd4U,0x0ba3U,0x1746U,
		0x2e8cU,0x5d18U,0xba30U,0x646bU,0xc8d6U,0x81a7U,0x1345U,0x268aU,0x4d14U,0x9a28U,
		0x245bU,0x48b6U,0x916cU,0x32d3U,0x65a6U,0xcb4cU,0x8693U,0x1d2dU,0x3a5aU,0x74b4U,
		0xe968U,0xc2dbU,0x95bdU,0x3b71U,0x76e2U,0xedc4U,0xcb83U,0x870dU,0x1e11U,0x3c22U,
		0x7844U,0xf088U,0xf11bU,0xf23dU,0xf471U,0xf8e9U,0xe1d9U,0xd3b9U,0xb779U,0x7ef9U,
		0xfdf2U,0xebefU,0xc7d5U,0x9fa1U,0x2f49U,0x5e92U,0xbd24U,0x6a43U,0xd486U,0xb907U,
		0x6205U,0xc40aU,0x981fU,0x2035U,0x406aU,0x80d4U,0x11a3U,0x2346U,0x468cU,0x8d18U,
		0x0a3bU,0x1476U,0x28ecU,0x51d8U,0xa3b0U,0x576bU,0xaed6U,0x4da7U,0x9b4eU,0x2697U,
		0x4d2eU,0x9a5cU,0x24b3U,0x4966U,0x92ccU,0x3593U,0x6b26U,0xd64cU,0xbc93U,0x692dU,
		0xd25aU,0xb4bfU,0x7975U,0xf2eaU,0xf5dfU,0xfbb5U,0xe761U,0xdec9U,0xad99U,0x4b39U,
		0x9672U,0x3cefU,0x79deU,0xf3bcU,0xf773U,0xfeedU,0xedd1U,0xcba9U,0x8759U,0x1eb9U,
		0x3d72U,0x7ae4U,0xf5c8U,0xfb9bU,0xe73dU,0xde71U,0xace9U,0x49d9U,0x93b2U,0x376fU,
		0x6edeU,0xddbcU,0xab73U,0x46edU,0x8ddaU,0x0bbfU,0x177eU,0x2efcU,0x5df8U,0xbbf0U,
		0x67ebU,0xcfd6U,0x8fa7U,0x0f45U,0x1e8aU,0x3d14U,0x7a28U,0xf450U,0xf8abU,0xe15dU,
		0xd2b1U,0xb569U,0x7ad9U,0xf5b2U,0xfb6fU,0xe6d5U,0xdda1U,0xab49U,0x4699U,0x8d32U,
		0x0a6fU,0x14deU,0x29bcU,0x5378U,0xa6f0U,0x5debU,0xbbd6U,0x67a7U,0xcf4eU,0x8e97U,
		0x0d25U,0x1a4aU,0x3494U,0x6928U,0xd250U,0xb4abU,0x795dU,0xf2baU,0xf57fU,0xfaf5U,
		0xe5e1U,0xdbc9U,0xa799U,0x5f39U,0xbe72U,0x6cefU,0xd9deU,0xa3b7U,0x5765U,0xaecaU,
		0x4d9fU,0x9b3eU,0x2677U,0x4ceeU,0x99dcU,0x23b3U,0x4766U,0x8eccU,0x0d93U,0x1b26U,
		0x364cU,0x6c98U,0xd930U,0xa26bU,0x54ddU,0xa9baU,0x437fU,0x86feU,0x1df7U,0x3beeU,
		0x77dcU,0xefb8U,0xcf7bU,0x8efdU,0x0df1U,0x1be2U,0x37c4U,0x6f88U,0xdf10U,0xae2bU,
		0x4c5dU,0x98baU,0x217fU,0x42feU,0x85fcU,0x1bf3U,0x37e6U,0x6fccU,0xdf98U,0xaf3bU,
		0x4e7dU,0x9cfaU,0x29ffU,0x53feU,0xa7fcU,0x5ff3U,0xbfe6U,0x6fc7U,0xdf8eU,0xaf17U,
		0x4e25U,0x9c4aU,0x289fU,0x513eU,0xa27cU,0x54f3U,0xa9e6U,0x43c7U,0x878eU,0x1f17U,
		0x3e2eU,0x7c5cU,0xf8b8U,0xe17bU,0xd2fdU,0xb5f1U,0x7be9U,0xf7d2U,0xffafU,0xef55U,
		0xcea1U,0x8d49U,0x0a99U,0x1532U,0x2a64U,0x54c8U,0xa990U,0x432bU,0x8656U,0x1ca7U,
		0x394eU,0x729cU,0xe538U,0xda7bU,0xa4fdU,0x59f1U,0xb3e2U,0x77cfU,0xef9eU,0xcf37U,
		0x8e65U,0x0cc1U,0x1982U,0x3304U,0x6608U,0xcc10U,0x882bU,0x005dU,0x00baU,0x0174U,
		0x02e8U,0x05d0U,0x0ba0U,0x1740U,0x2e80U,0x5d00U,0xba00U,0x640bU,0xc816U,0x8027U,
		0x1045U,0x208aU,0x4114U,0x8228U,0x145bU,0x28b6U,0x516cU,0xa2d8U,0x55bbU,0xab76U,
		0x46e7U,0x8dceU,0x0b97U,0x172eU,0x2e5cU,0x5cb8U,0xb970U,0x62ebU,0xc5d6U,0x9ba7U,
		0x2745U,0x4e8aU,0x9d14U,0x2a23U,0x5446U,0xa88cU,0x4113U,0x8226U,0x1447U,0x288eU,
		0x511cU,0xa238U,0x547bU,0xa8f6U,0x41e7U,0x83ceU,0x1797U,0x2f2eU,0x5e5cU,0xbcb8U,
		0x697bU,0xd2f6U,0xb5e7U,0x7bc5U,0xf78aU,0xff1fU,0xee35U,0xcc61U,0x88c9U,0x0199U,
		0x0332U,0x0664U,0x0cc8U,0x1990U,0x3320U,0x6640U,0xcc80U,0x890bU,0x021dU,0x043aU,
		0x0874U,0x10e8U,0x21d0U,0x43a0U,0x8740U,0x1e8bU,0x3d16U,0x7a2cU,0xf458U,0xf8bbU,
		0xe17dU,0xd2f1U,0xb5e9U,0x7bd9U,0xf7b2U,0xff6fU,0xeed5U,0xcda1U,0x8b49U,0x0699U,
		0x0d32U,0x1a64U,0x34c8U,0x6990U,0xd320U,0xb64bU,0x7c9dU,0xf93aU,0xe27fU,0xd4f5U,
		0xb9e1U,0x63c9U,0xc792U,0x9f2fU,0x2e55U,0x5caaU,0xb954U,0x62a3U,0xc546U,0x9a87U,
		0x2505U,0x4a0aU,0x9414U,0x3823U,0x7046U,0xe08cU,0xd113U,0xb22dU,0x7451U,0xe8a2U,
		0xc14fU,0x9295U,0x3521U,0x6a42U,0xd484U,0xb903U,0x620dU,0xc41aU,0x983fU,0x2075U,
		0x40eaU,0x81d4U,0x13a3U,0x2746U,0x4e8cU,0x9d18U,0x2a3bU,0x5476U,0xa8ecU,0x41d3U,
		0x83a6U,0x1747U,0x2e8eU,0x5d1cU,0xba38U,0x647bU,0xc8f6U,0x81e7U,0x13c5U,0x278aU,
		0x4f14U,0x9e28U,0x2c5bU,0x58b6U,0xb16cU,0x72d3U,0xe5a6U,0xdb47U,0xa685U,0x5d01U,
		0xba02U,0x640fU,0xc81eU,0x8037U,0x1065U,0x20caU,0x4194U,0x8328U,0x165bU,0x2cb6U,
		0x596cU,0xb2d8U,0x75bbU,0xeb76U,0xc6e7U,0x9dc5U,0x2b81U,0x5702U,0xae04U,0x4c03U,
		0x9806U,0x2007U,0x400eU,0x801cU,0x1033U,0x2066U,0x40ccU,0x8198U,0x133bU,0x2676U,
		0x4cecU,0x99d8U,0x23bbU,0x4776U,0x8eecU,0x0dd3U,0x1ba6U,0x374cU,0x6e98U,0xdd30U,
		0xaa6bU,0x44ddU,0x89baU,0x037fU,0x06feU,0x0dfcU,0x1bf8U,0x37f0U,0x6fe0U,0xdfc0U,
		0xaf8bU,0x4f1dU,0x9e3aU,0x2c7fU,0x58feU,0xb1fcU,0x73f3U,0xe7e6U,0xdfc7U,0xaf85U,
		0x4f01U,0x9e02U,0x2c0fU,0x581eU,0xb03cU,0x7073U,0xe0e6U,0xd1c7U,0xb385U,0x7701U,
		0xee02U,0xcc0fU,0x8815U,0x0021U,0x0042U,0x0084U,0x0108U,0x0210U,0x0420U,0x0840U,
		0x1080U,0x2100U,0x4200U,0x8400U,0x180bU,0x3016U,0x602cU,0xc058U,0x90bbU,0x317dU,
		0x62faU,0xc5f4U,0x9be3U,0x27cdU,0x4f9aU,0x9f34U,0x2e63U,0x5cc6U,0xb98cU,0x6313U,
		0xc626U,0x9c47U,0x2885U,0x510aU,0xa214U,0x5423U,0xa846U,0x4087U,0x810eU,0x1217U,
		0x242eU,0x485cU,0x90b8U,0x317bU,0x62f6U,0xc5ecU,0x9bd3U,0x27adU,0x4f5aU,0x9eb4U,
		0x2d63U,0x5ac6U,0xb58cU,0x7b13U,0xf626U,0xfc47U,0xe885U,0xc101U,0x9209U,0x3419U,
		0x6832U,0xd064U,0xb0c3U,0x718dU,0xe31aU,0xd63fU,0xbc75U,0x68e1U,0xd1c2U,0xb38fU,
		0x7715U,0xee2aU,0xcc5fU,0x88b5U,0x0161U,0x02c2U,0x0584U,0x0b08U,0x1610U,0x2c20U,
		0x5840U,0xb080U,0x710bU,0xe216U,0xd427U,0xb845U,0x6081U,0xc102U,0x920fU,0x3415U,
		0x682aU,0xd054U,0xb0a3U,0x714dU,0xe29aU,0xd53fU,0xba75U,0x64e1U,0xc9c2U,0x838fU,
		0x1715U,0x2e2aU,0x5c54U,0xb8a8U,0x615bU,0xc2b6U,0x9567U,0x3ac5U,0x758aU,0xeb14U,
		0xc623U,0x9c4dU,0x2891U,0x5122U,0xa244U,0x5483U,0xa906U,0x4207U,0x840eU,0x1817U,
		0x302eU,0x605cU,0xc0b8U,0x917bU,0x32fdU,0x65faU,0xcbf4U,0x87e3U,0x1fcdU,0x3f9aU,
		0x7f34U,0xfe68U,0xecdbU,0xc9bdU,0x8371U,0x16e9U,0x2dd2U,0x5ba4U,0xb748U,0x7e9bU,
		0xfd36U,0xea67U,0xc4c5U,0x9981U,0x2309U,0x4612U,0x8c24U,0x0843U,0x1086U,0x210cU,
		0x4218U,0x8430U,0x186bU,0x30d6U,0x61acU,0xc358U,0x96bbU,0x3d7dU,0x7afaU,0xf5f4U,
		0xfbe3U,0xe7cdU,0xdf91U,0xaf29U,0x4e59U,0x9cb2U,0x296fU,0x52deU,0xa5bcU,0x5b73U,
		0xb6e6U,0x7dc7U,0xfb8eU,0xe717U,0xde25U,0xac41U,0x4889U,0x9112U,0x322fU,0x645eU,
		0xc8bcU,0x8173U,0x12edU,0x25daU,0x4bb4U,0x9768U,0x3edbU,0x7db6U,0xfb6cU,0xe6d3U,
		0xddadU,0xab51U,0x46a9U,0x8d52U,0x0aafU,0x155eU,0x2abcU,0x5578U,0xaaf0U,0x45ebU,
		0x8bd6U,0x07a7U,0x0f4eU,0x1e9cU,0x3d38U,0x7a70U,0xf4e0U,0xf9cbU,0xe39dU,0xd731U,
		0xbe69U,0x6cd9U,0xd9b2U,0xa36fU,0x56d5U,0xadaaU,0x4b5fU,0x96beU,0x3d77U,0x7aeeU,
		0xf5dcU,0xfbb3U,0xe76dU,0xded1U,0xada9U,0x4b59U,0x96b2U,0x3d6fU,0x7adeU,0xf5bcU,
		0xfb73U,0xe6edU,0xddd1U,0xaba9U,0x4759U,0x8eb2U,0x0d6fU,0x1adeU,0x35bcU,0x6b78U,
		0xd6f0U,0xbdebU,0x6bddU,0xd7baU,0xbf7fU,0x6ef5U,0xddeaU,0xabdfU,0x47b5U,0x8f6aU,
		0x0edfU,0x1dbeU,0x3b7cU,0x76f8U,0xedf0U,0xcbebU,0x87ddU,0x1fb1U,0x3f62U,0x7ec4U,
		0xfd88U,0xeb1bU,0xc63dU,0x9c71U,0x28e9U,0x51d2U,0xa3a4U,0x5743U,0xae86U,0x4d07U,
		0x9a0eU,0x2417U,0x482eU,0x905cU,0x30b3U,0x6166U,0xc2ccU,0x9593U,0x3b2dU,0x765aU,
		0xecb4U,0xc963U,0x82cdU,0x1591U,0x2b22U,0x5644U,0xac88U,0x491bU,0x9236U,0x3467U,
		0x68ceU,0xd19cU,0xb333U,0x766dU,0xecdaU,0xc9bfU,0x8375U,0x16e1U,0x2dc2U,0x5b84U,
		0xb708U,0x7e1bU,0xfc36U,0xe867U,0xc0c5U,0x9181U,0x3309U,0x6612U,0xcc24U,0x8843U,
		0x008dU,0x011aU,0x0234U,0x0468U,0x08d0U,0x11a0U,0x2340U,0x4680U,0x8d00U,0x0a0bU,
		0x1416U,0x282cU,0x5058U,0xa0b0U,0x516bU,0xa2d6U,0x55a7U,0xab4eU,0x4697U,0x8d2eU,
		0x0a57U,0x14aeU,0x295cU,0x52b8U,0xa570U,0x5aebU,0xb5d6U,0x7ba7U,0xf74eU,0xfe97U,
		0xed25U,0xca41U,0x8489U,0x1919U,0x3232U,0x6464U,0xc8c8U,0x819bU,0x133dU,0x267aU,
		0x4cf4U,0x99e8U,0x23dbU,0x47b6U,0x8f6cU,0x0ed3U,0x1da6U,0x3b4cU,0x7698U,0xed30U,
		0xca6bU,0x84ddU,0x19b1U,0x3362U,0x66c4U,0xcd88U,0x8b1bU,0x063dU,0x0c7aU,0x18f4U,
		0x31e8U,0x63d0U,0xc7a0U,0x9f4bU,0x2e9dU,0x5d3aU,0xba74U,0x64e3U,0xc9c6U,0x8387U,
		0x1705U,0x2e0aU,0x5c14U,0xb828U,0x605bU,0xc0b6U,0x9167U,0x32c5U,0x658aU,0xcb14U,
		0x8623U,0x1c4dU,0x389aU,0x7134U,0xe268U,0xd4dbU,0xb9bdU,0x6371U,0xc6e2U,0x9dcfU,
		0x2b95U,0x572aU,0xae54U,0x4ca3U,0x9946U,0x2287U,0x450eU,0x8a1cU,0x0433U,0x0866U,
		0x10ccU,0x2198U,0x4330U,0x8660U,0x1ccbU,0x3996U,0x732cU,0xe658U,0xdcbbU,0xa97dU,
		0x42f1U,0x85e2U,0x1bcfU,0x379eU,0x6f3cU,0xde78U,0xacfbU,0x49fdU,0x93faU,0x37ffU,
		0x6ffeU,0xdffcU,0xaff3U,0x4fedU,0x9fdaU,0x2fbfU,0x5f7eU,0xbefcU,0x6df3U,0xdbe6U,
		0xa7c7U,0x5f85U,0xbf0aU,0x6e1fU,0xdc3eU,0xa877U,0x40e5U,0x81caU,0x139fU,0x273eU,
		0x4e7cU,0x9cf8U,0x29fbU,0x53f6U,0xa7ecU,0x5fd3U,0xbfa6U,0x6f47U,0xde8eU,0xad17U,
		0x4a25U,0x944aU,0x389fU,0x713eU,0xe27cU,0xd4f3U,0xb9edU,0x63d1U,0xc7a2U,0x9f4fU,
		0x2e95U,0x5d2aU,0xba54U,0x64a3U,0xc946U,0x8287U,0x1505U,0x2a0aU,0x5414U,0xa828U,
		0x405bU,0x80b6U,0x1167U,0x22ceU,0x459cU,0x8b38U,0x067bU,0x0cf6U,0x19ecU,0x33d8U,
		0x67b0U,0xcf60U,0x8ecbU,0x0d9dU,0x1b3aU,0x3674U,0x6ce8U,0xd9d0U,0xa3abU,0x575dU,
		0xaebaU,0x4d7fU,0x9afeU,0x25f7U,0x4beeU,0x97dcU,0x3fb3U,0x7f66U,0xfeccU,0xed93U,
		0xcb2dU,0x8651U,0x1ca9U,0x3952U,0x72a4U,0xe548U,0xda9bU,0xa53dU,0x5a71U,0xb4e2U,
		0x79cfU,0xf39eU,0xf737U,0xfe65U,0xecc1U,0xc989U,0x8319U,0x1639U,0x2c72U,0x58e4U,
		0xb1c8U,0x739bU,0xe736U,0xde67U,0xacc5U,0x4981U,0x9302U,0x360fU,0x6c1eU,0xd83cU,
		0xa073U,0x50edU,0xa1daU,0x53bfU,0xa77eU,0x5ef7U,0xbdeeU,0x6bd7U,0xd7aeU,0xbf57U,
		0x6ea5U,0xdd4aU,0xaa9fU,0x4535U,0x8a6aU,0x04dfU,0x09beU,0x137cU,0x26f8U,0x4df0U,
		0x9be0U,0x27cbU,0x4f96U,0x9f2cU,0x2e53U,0x5ca6U,0xb94cU,0x6293U,0xc526U,0x9a47U,
		0x2485U,0x490aU,0x9214U,0x3423U,0x6846U,0xd08cU,0xb113U,0x722dU,0xe45aU,0xd8bfU,
		0xa175U,0x52e1U,0xa5c2U,0x5b8fU,0xb71eU,0x7e37U,0xfc6eU,0xe8d7U,0xc1a5U,0x9341U,
		0x3689U,0x6d12U,0xda24U,0xa443U,0x588dU,0xb11aU,0x723fU,0xe47eU,0xd8f7U,0xa1e5U,
		0x53c1U,0xa782U,0x5f0fU,0xbe1eU,0x6c37U,0xd86eU,0xa0d7U,0x51a5U,0xa34aU,0x569fU,
		0xad3eU,0x4a77U,0x94eeU,0x39d7U,0x73aeU,0xe75cU,0xdeb3U,0xad6dU,0x4ad1U,0x95a2U,
		0x3b4fU,0x769eU,0xed3cU,0xca73U,0x84edU,0x19d1U,0x33a2U,0x6744U,0xce88U,0x8d1bU,
		0x0a3dU,0x147aU,0x28f4U,0x51e8U,0xa3d0U,0x57abU,0xaf56U,0x4ea7U,0x9d4eU,0x2a97U,
		0x552eU,0xaa5cU,0x44b3U,0x8966U,0x02c7U,0x058eU,0x0b1cU,0x1638U,0x2c70U,0x58e0U,
		0xb1c0U,0x738bU,0xe716U,0xde27U,0xac45U,0x4881U,0x9102U,0x320fU,0x641eU,0xc83cU,
		0x8073U,0x10edU,0x21daU,0x43b4U,0x8768U,0x1edbU,0x3db6U,0x7b6cU,0xf6d8U,0xfdbbU,
		0xeb7dU,0xc6f1U,0x9de9U,0x2bd9U,0x57b2U,0xaf64U,0x4ec3U,0x9d86U,0x2b07U,0x560eU,
		0xac1cU,0x4833U,0x9066U,0x30c7U,0x618eU,0xc31cU,0x9633U,0x3c6dU,0x78daU,0xf1b4U,
		0xf363U,0xf6cdU,0xfd91U,0xeb29U,0xc659U,0x9cb9U,0x2979U,0x52f2U,0xa5e4U,0x5bc3U,
		0xb786U,0x7f07U,0xfe0eU,0xec17U,0xc825U,0x8041U,0x1089U,0x2112U,0x4224U,0x8448U,
		0x189bU,0x3136U,0x626cU,0xc4d8U,0x99bbU,0x237dU,0x46faU,0x8df4U,0x0be3U,0x17c6U,
		0x2f8cU,0x5f18U,0xbe30U,0x6c6bU,0xd8d6U,0xa1a7U,0x5345U,0xa68aU,0x5d1fU,0xba3eU,
		0x6477U,0xc8eeU,0x81d7U,0x13a5U,0x274aU,0x4e94U,0x9d28U,0x2a5bU,0x54b6U,0xa96cU,
		0x42d3U,0x85a6U,0x1b47U,0x368eU,0x6d1cU,0xda38U,0xa47bU,0x58fdU,0xb1faU,0x73ffU,
		0xe7feU,0xdff7U,0xafe5U,0x4fc1U,0x9f82U,0x2f0fU,0x5e1eU,0xbc3cU,0x6873U,0xd0e6U,
		0xb1c7U,0x7385U,0xe70aU,0xde1fU,0xac35U,0x4861U,0x90c2U,0x318fU,0x631eU,0xc63cU,
		0x9c73U,0x28edU,0x51daU,0xa3b4U,0x5763U,0xaec6U,0x4d87U,0x9b0eU,0x2617U,0x4c2eU,
		0x985cU,0x20b3U,0x4166U,0x82ccU,0x1593U,0x2b26U,0x564cU,0xac98U,0x493bU,0x9276U,
		0x34e7U,0x69ceU,0xd39cU,0xb733U,0x7e6dU,0xfcdaU,0xe9bfU,0xc375U,0x96e1U,0x3dc9U,
		0x7b92U,0xf724U,0xfe43U,0xec8dU,0xc911U,0x8229U,0x1459U,0x28b2U,0x5164U,0xa2c8U,
		0x559bU,0xab36U,0x4667U,0x8cceU,0x0997U,0x132eU,0x265cU,0x4cb8U,0x9970U,0x22ebU,
		0x45d6U,0x8bacU,0x0753U,0x0ea6U,0x1d4cU,0x3a98U,0x7530U,0xea60U,0xc4cbU,0x999dU,
		0x2331U,0x4662U,0x8cc4U,0x0983U,0x1306U,0x260cU,0x4c18U,0x9830U,0x206bU,0x40d6U,
		0x81acU,0x1353U,0x26a6U,0x4d4cU,0x9a98U,0x253bU,0x4a76U,0x94ecU,0x39d3U,0x73a6U,
		0xe74cU,0xde93U,0xad2dU,0x4a51U,0x94a2U,0x394fU,0x729eU,0xe53cU,0xda73U,0xa4edU,
		0x59d1U,0xb3a2U,0x774fU,0xee9eU,0xcd37U,0x8a65U,0x04c1U,0x0982U,0x1304U,0x2608U,
		0x4c10U,0x9820U,0x204bU,0x4096U,0x812cU,0x1253U,0x24a6U,0x494cU,0x9298U,0x353bU,
		0x6a76U,0xd4ecU,0xb9d3U,0x63adU,0xc75aU,0x9ebfU,0x2d75U,0x5aeaU,0xb5d4U,0x7ba3U,
		0xf746U,0xfe87U,0xed05U,0xca01U,0x8409U,0x1819U,0x3032U,0x6064U,0xc0c8U,0x919bU,
		0x333dU,0x667aU,0xccf4U,0x89e3U,0x03cdU,0x079aU,0x0f34U,0x1e68U,0x3cd0U,0x79a0U,
		0xf340U,0xf68bU,0xfd1dU,0xea31U,0xc469U,0x98d9U,0x21b9U,0x4372U,0x86e4U,0x1dc3U,
		0x3b86U,0x770cU,0xee18U,0xcc3bU,0x887dU,0x00f1U,0x01e2U,0x03c4U,0x0788U,0x0f10U,
		0x1e20U,0x3c40U,0x7880U,0xf100U,0xf20bU,0xf41dU,0xf831U,0xe069U,0xd0d9U,0xb1b9U,
		0x7379U,0xe6f2U,0xddefU,0xabd5U,0x47a1U,0x8f42U,0x0e8fU,0x1d1eU,0x3a3cU,0x7478U,
		0xe8f0U,0xc1ebU,0x93ddU,0x37b1U,0x6f62U,0xdec4U,0xad83U,0x4b0dU,0x961aU,0x3c3fU,
		0x787eU,0xf0fcU,0xf1f3U,0xf3edU,0xf7d1U,0xffa9U,0xef59U,0xceb9U,0x8d79U,0x0af9U,
		0x15f2U,0x2be4U,0x57c8U,0xaf90U,0x4f2bU,0x9e56U,0x2ca7U,0x594eU,0xb29cU,0x7533U,
		0xea66U,0xc4c7U,0x9985U,0x2301U,0x4602U,0x8c04U,0x0803U,0x1006U,0x200cU,0x4018U,
		0x8030U,0x106bU,0x20d6U,0x41acU,0x8358U,0x16bbU,0x2d76U,0x5aecU,0xb5d8U,0x7bbbU,
		0xf776U,0xfee7U,0xedc5U,0xcb81U,0x8709U,0x1e19U,0x3c32U,0x7864U,0xf0c8U,0xf19bU,
		0xf33dU,0xf671U,0xfce9U,0xe9d9U,0xc3b9U,0x9779U,0x3ef9U,0x7df2U,0xfbe4U,0xe7c3U,
		0xdf8dU,0xaf11U,0x4e29U,0x9c52U,0x28afU,0x515eU,0xa2bcU,0x5573U,0xaae6U,0x45c7U,
		0x8b8eU,0x0717U,0x0e2eU,0x1c5cU,0x38b8U,0x7170U,0xe2e0U,0xd5cbU,0xbb9dU,0x6731U,
		0xce62U,0x8ccfU,0x0995U,0x132aU,0x2654U,0x4ca8U,0x9950U,0x22abU,0x4556U,0x8aacU,
		0x0553U,0x0aa6U,0x154cU,0x2a98U,0x5530U,0xaa60U,0x44cbU,0x8996U,0x0327U,0x064eU,
		0x0c9cU,0x1938U,0x3270U,0x64e0U,0xc9c0U,0x838bU,0x171dU,0x2e3aU,0x5c74U,0xb8e8U,
		0x61dbU,0xc3b6U,0x9767U,0x3ec5U,0x7d8aU,0xfb14U,0xe623U,0xdc4dU,0xa891U,0x4129U,
		0x8252U,0x14afU,0x295eU,0x52bcU,0xa578U,0x5afbU,0xb5f6U,0x7be7U,0xf7ceU,0xff97U,
		0xef25U,0xce41U,0x8c89U,0x0919U,0x1232U,0x2464U,0x48c8U,0x9190U,0x332bU,0x6656U,
		0xccacU,0x8953U,0x02adU,0x055aU,0x0ab4U,0x1568U,0x2ad0U,0x55a0U,0xab40U,0x468bU,
		0x8d16U,0x0a27U,0x144eU,0x289cU,0x5138U,0xa270U,0x54ebU,0xa9d6U,0x43a7U,0x874eU,
		0x1e97U,0x3d2eU,0x7a5cU,0xf4b8U,0xf97bU,0xe2fdU,0xd5f1U,0xbbe9U,0x67d9U,0xcfb2U,
		0x8f6fU,0x0ed5U,0x1daaU,0x3b54U,0x76a8U,0xed50U,0xcaabU,0x855dU,0x1ab1U,0x3562U,
		0x6ac4U,0xd588U,0xbb1bU,0x663dU,0xcc7aU,0x88ffU,0x01f5U,0x03eaU,0x07d4U,0x0fa8U,
		0x1f50U,0x3ea0U,0x7d40U,0xfa80U,0xe50bU,0xda1dU,0xa431U,0x5869U,0xb0d2U,0x71afU,
		0xe35eU,0xd6b7U,0xbd65U,0x6ac1U,0xd582U,0xbb0fU,0x6615U,0xcc2aU,0x885fU,0x00b5U,
		0x016aU,0x02d4U,0x05a8U,0x0b50U,0x16a0U,0x2d40U,0x5a80U,0xb500U,0x7a0bU,0xf416U,
		0xf827U,0xe045U,0xd081U,0xb109U,0x7219U,0xe432U,0xd86fU,0xa0d5U,0x51a1U,0xa342U,
		0x568fU,0xad1eU,0x4a37U,0x946eU,0x38d7U,0x71aeU,0xe35cU,0xd6b3U,0xbd6dU,0x6ad1U,
		0xd5a2U,0xbb4fU,0x6695U,0xcd2aU,0x8a5fU,0x04b5U,0x096aU,0x12d4U,0x25a8U,0x4b50U,
		0x96a0U,0x3d4bU,0x7a96U,0xf52cU,0xfa53U,0xe4adU,0xd951U,0xa2a9U,0x5559U,0xaab2U,
		0x456fU,0x8adeU,0x05b7U,0x0b6eU,0x16dcU,0x2db8U,0x5b70U,0xb6e0U,0x7dcbU,0xfb96U,
		0xe727U,0xde45U,0xac81U,0x4909U,0x9212U,0x342fU,0x685eU,0xd0bcU,0xb173U,0x72edU,
		0xe5daU,0xdbbfU,0xa775U,0x5ee1U,0xbdc2U,0x6b8fU,0xd71eU,0xbe37U,0x6c65U,0xd8caU,
		0xa19fU,0x5335U,0xa66aU,0x5cdfU,0xb9beU,0x6377U,0xc6eeU,0x9dd7U,0x2ba5U,0x574aU,
		0xae94U,0x4d23U,0x9a46U,0x2487U,0x490eU,0x921cU,0x3433U,0x6866U,0xd0ccU,0xb193U,
		0x732dU,0xe65aU,0xdcbfU,0xa975U,0x42e1U,0x85c2U,0x1b8fU,0x371eU,0x6e3cU,0xdc78U,
		0xa8fbU,0x41fdU,0x83faU,0x17ffU,0x2ffeU,0x5ffcU,0xbff8U,0x6ffbU,0xdff6U,0xafe7U,
		0x4fc5U,0x9f8aU,0x2f1fU,0x5e3eU,0xbc7cU,0x68f3U,0xd1e6U,0xb3c7U,0x7785U,0xef0aU,
		0xce1fU,0x8c35U,0x0861U,0x10c2U,0x2184U,0x4308U,0x8610U,0x1c2bU,0x3856U,0x70acU,
		0xe158U,0xd2bbU,0xb57dU,0x7af1U,0xf5e2U,0xfbcfU,0xe795U,0xdf21U,0xae49U,0x4c99U,
		0x9932U,0x226fU,0x44deU,0x89bcU,0x0373U,0x06e6U,0x0dccU,0x1b98U,0x3730U,0x6e60U,
		0xdcc0U,0xa98bU,0x431dU,0x863aU,0x1c7fU,0x38feU,0x71fcU,0xe3f8U,0xd7fbU,0xbffdU,
		0x6ff1U,0xdfe2U,0xafcfU,0x4f95U,0x9f2aU,0x2e5fU,0x5cbeU,0xb97cU,0x62f3U,0xc5e6U,
		0x9bc7U,0x2785U,0x4f0aU,0x9e14U,0x2c23U,0x5846U,0xb08cU,0x7113U,0xe226U,0xd447U,
		0xb885U,0x6101U,0xc202U,0x940fU,0x3815U,0x702aU,0xe054U,0xd0a3U,0xb14dU,0x7291U,
		0xe522U,0xda4fU,0xa495U,0x5921U,0xb242U,0x748fU,0xe91eU,0xc237U,0x9465U,0x38c1U,
		0x7182U,0xe304U,0xd603U,0xbc0dU,0x6811U,0xd022U,0xb04fU,0x7095U,0xe12aU,0xd25fU,
		0xb4b5U,0x7961U,0xf2c2U,0xf58fU,0xfb15U,0xe621U,0xdc49U,0xa899U,0x4139U,0x8272U,
		0x14efU,0x29deU,0x53bcU,0xa778U,0x5efbU,0xbdf6U,0x6be7U,0xd7ceU,0xbf97U,0x6f25U,
		0xde4aU,0xac9fU,0x4935U,0x926aU,0x34dfU,0x69beU,0xd37cU,0xb6f3U,0x7dedU,0xfbdaU,
		0xe7bfU,0xdf75U,0xaee1U,0x4dc9U,0x9b92U,0x272fU,0x4e5eU,0x9cbcU,0x2973U,0x52e6U,
		0xa5ccU,0x5b93U,0xb726U,0x7e47U,0xfc8eU,0xe917U,0xc225U,0x9441U,0x3889U,0x7112U,
		0xe224U,0xd443U,0xb88dU,0x6111U,0xc222U,0x944fU,0x3895U,0x712aU,0xe254U,0xd4a3U,
		0xb94dU,0x6291U,0xc522U,0x9a4fU,0x2495U,0x492aU,0x9254U,0x34a3U,0x6946U,0xd28cU,
		0xb513U,0x7a2dU,0xf45aU,0xf8bfU,0xe175U,0xd2e1U,0xb5c9U,0x7b99U,0xf732U,0xfe6fU,
		0xecd5U,0xc9a1U,0x8349U,0x1699U,0x2d32U,0x5a64U,0xb4c8U,0x799bU,0xf336U,0xf667U,
		0xfcc5U,0xe981U,0xc309U,0x9619U,0x3c39U,0x7872U,0xf0e4U,0xf1c3U,0xf38dU,0xf711U,
		0xfe29U,0xec59U,0xc8b9U,0x8179U,0x12f9U,0x25f2U,0x4be4U,0x97c8U,0x3f9bU,0x7f36U,
		0xfe6cU,0xecd3U,0xc9adU,0x8351U,0x16a9U,0x2d52U,0x5aa4U,0xb548U,0x7a9bU,0xf536U,
		0xfa67U,0xe4c5U,0xd981U,0xa309U,0x5619U,0xac32U,0x486fU,0x90deU,0x31b7U,0x636eU,
		0xc6dcU,0x9db3U,0x2b6dU,0x56daU,0xadb4U,0x4b63U,0x96c6U,0x3d87U,0x7b0eU,0xf61cU,
		0xfc33U,0xe86dU,0xc0d1U,0x91a9U,0x3359U,0x66b2U,0xcd64U,0x8ac3U,0x058dU,0x0b1aU,
		0x1634U,0x2c68U,0x58d0U,0xb1a0U,0x734bU,0xe696U,0xdd27U,0xaa45U,0x4481U,0x8902U,
		0x020fU,0x041eU,0x083cU,0x1078U,0x20f0U,0x41e0U,0x83c0U,0x178bU,0x2f16U,0x5e2cU,
		0xbc58U,0x68bbU,0xd176U,0xb2e7U,0x75c5U,0xeb8aU,0xc71fU,0x9e35U,0x2c61U,0x58c2U,
		0xb184U,0x7303U,0xe606U,0xdc07U,0xa805U,0x4001U,0x8002U,0x100fU,0x201eU,0x403cU,
		0x8078U,0x10fbU,0x21f6U,0x43ecU,0x87d8U,0x1fbbU,0x3f76U,0x7eecU,0xfdd8U,0xebbbU,
		0xc77dU,0x9ef1U,0x2de9U,0x5bd2U,0xb7a4U,0x7f43U,0xfe86U,0xed07U,0xca05U,0x8401U,
		0x1809U,0x3012U,0x6024U,0xc048U,0x909bU,0x313dU,0x627aU,0xc4f4U,0x99e3U,0x23cdU,
		0x479aU,0x8f34U,0x0e63U,0x1cc6U,0x398cU,0x7318U,0xe630U,0xdc6bU,0xa8ddU,0x41b1U,
		0x8362U,0x16cfU,0x2d9eU,0x5b3cU,0xb678U,0x7cfbU,0xf9f6U,0xe3e7U,0xd7c5U,0xbf81U,
		0x6f09U,0xde12U,0xac2fU,0x4855U,0x90aaU,0x315fU,0x62beU,0xc57cU,0x9af3U,0x25edU,
		0x4bdaU,0x97b4U,0x3f63U,0x7ec6U,0xfd8cU,0xeb13U,0xc62dU,0x9c51U,0x28a9U,0x5152U,
		0xa2a4U,0x5543U,0xaa86U,0x4507U,0x8a0eU,0x0417U,0x082eU,0x105cU,0x20b8U,0x4170U,
		0x82e0U,0x15cbU,0x2b96U,0x572cU,0xae58U,0x4cbbU,0x9976U,0x22e7U,0x45ceU,0x8b9cU,
		0x0733U,0x0e66U,0x1cccU,0x3998U,0x7330U,0xe660U,0xdccbU,0xa99dU,0x4331U,0x8662U,
		0x1ccfU,0x399eU,0x733cU,0xe678U,0xdcfbU,0xa9fdU,0x43f1U,0x87e2U,0x1fcfU,0x3f9eU,
		0x7f3cU,0xfe78U,0xecfbU,0xc9fdU,0x83f1U,0x17e9U,0x2fd2U,0x5fa4U,0xbf48U,0x6e9bU,
		0xdd36U,0xaa67U,0x44c5U,0x898aU,0x031fU,0x063eU,0x0c7cU,0x18f8U,0x31f0U,0x63e0U,
		0xc7c0U,0x9f8bU,0x2f1dU,0x5e3aU,0xbc74U,0x68e3U,0xd1c6U,0xb387U,0x7705U,0xee0aU,
		0xcc1fU,0x8835U,0x0061U,0x00c2U,0x0184U,0x0308U,0x0610U,0x0c20U,0x1840U,0x3080U,
		0x6100U,0xc200U,0x940bU,0x381dU,0x703aU,0xe074U,0xd0e3U,0xb1cdU,0x7391U,0xe722U,
		0xde4fU,0xac95U,0x4921U,0x9242U,0x348fU,0x691eU,0xd23cU,0xb473U,0x78edU,0xf1daU,
		0xf3bfU,0xf775U,0xfee1U,0xedc9U,0xcb99U,0x8739U,0x1e79U,0x3cf2U,0x79e4U,0xf3c8U,
		0xf79bU,0xff3dU,0xee71U,0xcce9U,0x89d9U,0x03b9U,0x0772U,0x0ee4U,0x1dc8U,0x3b90U,
		0x7720U,0xee40U,0xcc8bU,0x891dU,0x0231U,0x0462U,0x08c4U,0x1188U,0x2310U,0x4620U,
		0x8c40U,0x088bU,0x1116U,0x222cU,0x4458U,0x88b0U,0x016bU,0x02d6U,0x05acU,0x0b58U,
		0x16b0U,0x2d60U,0x5ac0U,0xb580U,0x7b0bU,0xf616U,0xfc27U,0xe845U,0xc081U,0x9109U,
		0x3219U,0x6432U,0xc864U,0x80c3U,0x118dU,0x231aU,0x4634U,0x8c68U,0x08dbU,0x11b6U,
		0x236cU,0x46d8U,0x8db0U,0x0b6bU,0x16d6U,0x2dacU,0x5b58U,0xb6b0U,0x7d6bU,0xfad6U,
		0xe5a7U,0xdb45U,0xa681U,0x5d09U,0xba12U,0x642fU,0xc85eU,0x80b7U,0x1165U,0x22caU,
		0x4594U,0x8b28U,0x065bU,0x0cb6U,0x196cU,0x32d8U,0x65b0U,0xcb60U,0x86cbU,0x1d9dU,
		0x3b3aU,0x7674U,0xece8U,0xc9dbU,0x83bdU,0x1771U,0x2ee2U,0x5dc4U,0xbb88U,0x671bU,
		0xce36U,0x8c67U,0x08c5U,0x118aU,0x2314U,0x4628U,0x8c50U,0x08abU,0x1156U,0x22acU,
		0x4558U,0x8ab0U,0x056bU,0x0ad6U,0x15acU,0x2b58U,0x56b0U,0xad60U,0x4acbU,0x9596U,
		0x3b27U,0x764eU,0xec9cU,0xc933U,0x826dU,0x14d1U,0x29a2U,0x5344U,0xa688U,0x5d1bU,
		0xba36U,0x6467U,0xc8ceU,0x8197U,0x1325U,0x264aU,0x4c94U,0x9928U,0x225bU,0x44b6U,
		0x896cU,0x02d3U,0x05a6U,0x0b4cU,0x1698U,0x2d30U,0x5a60U,0xb4c0U,0x798bU,0xf316U,
		0xf627U,0xfc45U,0xe881U,0xc109U,0x9219U,0x3439U,0x6872U,0xd0e4U,0xb1c3U,0x738dU,
		0xe71aU,0xde3fU,0xac75U,0x48e1U,0x91c2U,0x338fU,0x671eU,0xce3cU,0x8c73U,0x08edU,
		0x11daU,0x23b4U,0x4768U,0x8ed0U,0x0dabU,0x1b56U,0x36acU,0x6d58U,0xdab0U,0xa56bU,
		0x5addU,0xb5baU,0x7b7fU,0xf6feU,0xfdf7U,0xebe5U,0xc7c1U,0x9f89U,0x2f19U,0x5e32U,
		0xbc64U,0x68c3U,0xd186U,0xb307U,0x7605U,0xec0aU,0xc81fU,0x8035U,0x1061U,0x20c2U,
		0x4184U,0x8308U,0x161bU,0x2c36U,0x586cU,0xb0d8U,0x71bbU,0xe376U,0xd6e7U,0xbdc5U,
		0x6b81U,0xd702U,0xbe0fU,0x6c15U,0xd82aU,0xa05fU,0x50b5U,0xa16aU,0x52dfU,0xa5beU,
		0x5b77U,0xb6eeU,0x7dd7U,0xfbaeU,0xe757U,0xdea5U,0xad41U,0x4a89U,0x9512U,0x3a2fU,
		0x745eU,0xe8bcU,0xc173U,0x92edU,0x35d1U,0x6ba2U,0xd744U,0xbe83U,0x6d0dU,0xda1aU,
		0xa43fU,0x5875U,0xb0eaU,0x71dfU,0xe3beU,0xd777U,0xbee5U,0x6dc1U,0xdb82U,0xa70fU,
		0x5e15U,0xbc2aU,0x685fU,0xd0beU,0xb177U,0x72e5U,0xe5caU,0xdb9fU,0xa735U,0x5e61U,
		0xbcc2U,0x698fU,0xd31eU,0xb637U,0x7c65U,0xf8caU,0xe19fU,0xd335U,0xb661U,0x7cc9U,
		0xf992U,0xe32fU,0xd655U,0xbca1U,0x6949U,0xd292U,0xb52fU,0x7a55U,0xf4aaU,0xf95fU,
		0xe2b5U,0xd561U,0xbac9U,0x6599U,0xcb32U,0x866fU,0x1cd5U,0x39aaU,0x7354U,0xe6a8U,
		0xdd5bU,0xaabdU,0x4571U,0x8ae2U,0x05cfU,0x0b9eU,0x173cU,0x2e78U,0x5cf0U,0xb9e0U,
		0x63cbU,0xc796U,0x9f27U,0x2e45U,0x5c8aU,0xb914U,0x6223U,0xc446U,0x9887U,0x2105U,
		0x420aU,0x8414U,0x1823U,0x3046U,0x608cU,0xc118U,0x923bU,0x347dU,0x68faU,0xd1f4U,
		0xb3e3U,0x77cdU,0xef9aU,0xcf3fU,0x8e75U,0x0ce1U,0x19c2U,0x3384U,0x6708U,0xce10U,
		0x8c2bU,0x085dU,0x10baU,0x2174U,0x42e8U,0x85d0U,0x1babU,0x3756U,0x6eacU,0xdd58U,
		0xaabbU,0x457dU,0x8afaU,0x05ffU,0x0bfeU,0x17fcU,0x2ff8U,0x5ff0U,0xbfe0U,0x6fcbU,
		0xdf96U,0xaf27U,0x4e45U,0x9c8aU,0x291fU,0x523eU,0xa47cU,0x58f3U,0xb1e6U,0x73c7U,
		0xe78eU,0xdf17U,0xae25U,0x4c41U,0x9882U,0x210fU,0x421eU,0x843cU,0x1873U,0x30e6U,
		0x61ccU,0xc398U,0x973bU,0x3e7dU,0x7cfaU,0xf9f4U,0xe3e3U,0xd7cdU,0xbf91U,0x6f29U,
		0xde52U,0xacafU,0x4955U,0x92aaU,0x355fU,0x6abeU,0xd57cU,0xbaf3U,0x65edU,0xcbdaU,
		0x87bfU,0x1f75U,0x3eeaU,0x7dd4U,0xfba8U,0xe75bU,0xdebdU,0xad71U,0x4ae9U,0x95d2U,
		0x3bafU,0x775eU,0xeebcU,0xcd73U,0x8aedU,0x05d1U,0x0ba2U,0x1744U,0x2e88U,0x5d10U,
		0xba20U,0x644bU,0xc896U,0x8127U,0x1245U,0x248aU,0x4914U,0x9228U,0x345bU,0x68b6U,
		0xd16cU,0xb2d3U,0x75adU,0xeb5aU,0xc6bfU,0x9d75U,0x2ae1U,0x55c2U,0xab84U,0x4703U,
		0x8e06U,0x0c07U,0x180eU,0x301cU,0x6038U,0xc070U,0x90ebU,0x31ddU,0x63baU,0xc774U,
		0x9ee3U,0x2dcdU,0x5b9aU,0xb734U,0x7e63U,0xfcc6U,0xe987U,0xc305U,0x9601U,0x3c09U,
		0x7812U,0xf024U,0xf043U,0xf08dU,0xf111U,0xf229U,0xf459U,0xf8b9U,0xe179U,0xd2f9U,
		0xb5f9U,0x7bf9U,0xf7f2U,0xffefU,0xefd5U,0xcfa1U,0x8f49U,0x0e99U,0x1d32U,0x3a64U,
		0x74c8U,0xe990U,0xc32bU,0x965dU,0x3cb1U,0x7962U,0xf2c4U,0xf583U,0xfb0dU,0xe611U,
		0xdc29U,0xa859U,0x40b9U,0x8172U,0x12efU,0x25deU,0x4bbcU,0x9778U,0x3efbU,0x7df6U,
		0xfbecU,0xe7d3U,0xdfadU,0xaf51U,0x4ea9U,0x9d52U,0x2aafU,0x555eU,0xaabcU,0x4573U,
		0x8ae6U,0x05c7U,0x0b8eU,0x171cU,0x2e38U,0x5c70U,0xb8e0U,0x61cbU,0xc396U,0x9727U,
		0x3e45U,0x7c8aU,0xf914U,0xe223U,0xd44dU,0xb891U,0x6129U,0xc252U,0x94afU,0x3955U,
		0x72aaU,0xe554U,0xdaa3U,0xa54dU,0x5a91U,0xb522U,0x7a4fU,0xf49eU,0xf937U,0xe265U,
		0xd4c1U,0xb989U,0x6319U,0xc632U,0x9c6fU,0x28d5U,0x51aaU,0xa354U,0x56a3U,0xad46U,
		0x4a87U,0x950eU,0x3a17U,0x742eU,0xe85cU,0xc0b3U,0x916dU,0x32d1U,0x65a2U,0xcb44U,
		0x8683U,0x1d0dU,0x3a1aU,0x7434U,0xe868U,0xc0dbU,0x91bdU,0x3371U,0x66e2U,0xcdc4U,
		0x8b83U,0x070dU,0x0e1aU,0x1c34U,0x3868U,0x70d0U,0xe1a0U,0xd34bU,0xb69dU,0x7d31U,
		0xfa62U,0xe4cfU,0xd995U,0xa321U,0x5649U,0xac92U,0x492fU,0x925eU,0x34b7U,0x696eU,
		0xd2dcU,0xb5b3U,0x7b6dU,0xf6daU,0xfdbfU,0xeb75U,0xc6e1U,0x9dc9U,0x2b99U,0x5732U,
		0xae64U,0x4cc3U,0x9986U,0x2307U,0x460eU,0x8c1cU,0x0833U,0x1066U,0x20ccU,0x4198U,
		0x8330U,0x166bU,0x2cd6U,0x59acU,0xb358U,0x76bbU,0xed76U,0xcae7U,0x85c5U,0x1b81U,
		0x3702U,0x6e04U,0xdc08U,0xa81bU,0x403dU,0x807aU,0x10ffU,0x21feU,0x43fcU,0x87f8U,
		0x1ffbU,0x3ff6U,0x7fecU,0xffd8U,0xefbbU,0xcf7dU,0x8ef1U,0x0de9U,0x1bd2U,0x37a4U,
		0x6f48U,0xde90U,0xad2bU,0x4a5dU,0x94baU,0x397fU,0x72feU,0xe5fcU,0xdbf3U,0xa7edU,
		0x5fd1U,0xbfa2U,0x6f4fU,0xde9eU,0xad37U,0x4a65U,0x94caU,0x399fU,0x733eU,0xe67cU,
		0xdcf3U,0xa9edU,0x43d1U,0x87a2U,0x1f4fU,0x3e9eU,0x7d3cU,0xfa78U,0xe4fbU,0xd9fdU,
		0xa3f1U,0x57e9U,0xafd2U,0x4fafU,0x9f5eU,0x2eb7U,0x5d6eU,0xbadcU,0x65b3U,0xcb66U,
		0x86c7U,0x1d85U,0x3b0aU,0x7614U,0xec28U,0xc85bU,0x80bdU,0x1171U,0x22e2U,0x45c4U,
		0x8b88U,0x071bU,0x0e36U,0x1c6cU,0x38d8U,0x71b0U,0xe360U,0xd6cbU,0xbd9dU,0x6b31U,
		0xd662U,0xbccfU,0x6995U,0xd32aU,0xb65fU,0x7cb5U,0xf96aU,0xe2dfU,0xd5b5U,0xbb61U,
		0x66c9U,0xcd92U,0x8b2fU,0x0655U,0x0caaU,0x1954U,0x32a8U,0x6550U,0xcaa0U,0x854bU,
		0x1a9dU,0x353aU,0x6a74U,0xd4e8U,0xb9dbU,0x63bdU,0xc77aU,0x9effU,0x2df5U,0x5beaU,
		0xb7d4U,0x7fa3U,0xff46U,0xee87U,0xcd05U,0x8a01U,0x0409U,0x0812U,0x1024U,0x2048U,
		0x4090U,0x8120U,0x124bU,0x2496U,0x492cU,0x9258U,0x34bbU,0x6976U,0xd2ecU,0xb5d3U,
		0x7badU,0xf75aU,0xfebfU,0xed75U,0xcae1U,0x85c9U,0x1b99U,0x3732U,0x6e64U,0xdcc8U,
		0xa99bU,0x433dU,0x867aU,0x1cffU,0x39feU,0x73fcU,0xe7f8U,0xdffbU,0xaffdU,0x4ff1U,
		0x9fe2U,0x2fcfU,0x5f9eU,0xbf3cU,0x6e73U,0xdce6U,0xa9c7U,0x4385U,0x870aU,0x1e1fU,
		0x3c3eU,0x787cU,0xf0f8U,0xf1fbU,0xf3fdU,0xf7f1U,0xffe9U,0xefd9U,0xcfb9U,0x8f79U,
		0x0ef9U,0x1df2U,0x3be4U,0x77c8U,0xef90U,0xcf2bU,0x8e5dU,0x0cb1U,0x1962U,0x32c4U,
		0x6588U,0xcb10U,0x862bU,0x1c5dU,0x38baU,0x7174U,0xe2e8U,0xd5dbU,0xbbbdU,0x6771U,
		0xcee2U,0x8dcfU,0x0b95U,0x172aU,0x2e54U,0x5ca8U,0xb950U,0x62abU,0xc556U,0x9aa7U,
		0x2545U,0x4a8aU,0x9514U,0x3a23U,0x7446U,0xe88cU,0xc113U,0x922dU,0x3451U,0x68a2U,
		0xd144U,0xb283U,0x750dU,0xea1aU,0xc43fU,0x9875U,0x20e1U,0x41c2U,0x8384U,0x1703U,
		0x2e06U,0x5c0cU,0xb818U,0x603bU,0xc076U,0x90e7U,0x31c5U,0x638aU,0xc714U,0x9e23U,
		0x2c4dU,0x589aU,0xb134U,0x7263U,0xe4c6U,0xd987U,0xa305U,0x5601U,0xac02U,0x480fU,
		0x901eU,0x3037U,0x606eU,0xc0dcU,0x91b3U,0x336dU,0x66daU,0xcdb4U,0x8b63U,0x06cdU,
		0x0d9aU,0x1b34U,0x3668U,0x6cd0U,0xd9a0U,0xa34bU,0x569dU,0xad3aU,0x4a7fU,0x94feU,
		0x39f7U,0x73eeU,0xe7dcU,0xdfb3U,0xaf6dU,0x4ed1U,0x9da2U,0x2b4fU,0x569eU,0xad3cU,
		0x4a73U,0x94e6U,0x39c7U,0x738eU,0xe71cU,0xde33U,0xac6dU,0x48d1U,0x91a2U,0x334fU,
		0x669eU,0xcd3cU,0x8a73U,0x04edU,0x09daU,0x13b4U,0x2768U,0x4ed0U,0x9da0U,0x2b4bU,
		0x5696U,0xad2cU,0x4a53U,0x94a6U,0x3947U,0x728eU,0xe51cU,0xda33U,0xa46dU,0x58d1U,
		0xb1a2U,0x734fU,0xe69eU,0xdd37U,0xaa65U,0x44c1U,0x8982U,0x030fU,0x061eU,0x0c3cU,
		0x1878U,0x30f0U,0x61e0U,0xc3c0U,0x978bU,0x3f1dU,0x7e3aU,0xfc74U,0xe8e3U,0xc1cdU,
		0x9391U,0x3729U,0x6e52U,0xdca4U,0xa943U,0x428dU,0x851aU,0x1a3fU,0x347eU,0x68fcU,
		0xd1f8U,0xb3fbU,0x77fdU,0xeffaU,0xcfffU,0x8ff5U,0x0fe1U,0x1fc2U,0x3f84U,0x7f08U,
		0xfe10U,0xec2bU,0xc85dU,0x80b1U,0x1169U,0x22d2U,0x45a4U,0x8b48U,0x069bU,0x0d36U,
		0x1a6cU,0x34d8U,0x69b0U,0xd360U,0xb6cbU,0x7d9dU,0xfb3aU,0xe67fU,0xdcf5U,0xa9e1U,
		0x43c9U,0x8792U,0x1f2fU,0x3e5eU,0x7cbcU,0xf978U,0xe2fbU,0xd5fdU,0xbbf1U,0x67e9U,
		0xcfd2U,0x8fafU,0x0f55U,0x1eaaU,0x3d54U,0x7aa8U,0xf550U,0xfaabU,0xe55dU,0xdab1U,
		0xa569U,0x5ad9U,0xb5b2U,0x7b6fU,0xf6deU,0xfdb7U,0xeb65U,0xc6c1U,0x9d89U,0x2b19U,
		0x5632U,0xac64U,0x48c3U,0x9186U,0x3307U,0x660eU,0xcc1cU,0x8833U,0x006dU,0x00daU,
		0x01b4U,0x0368U,0x06d0U,0x0da0U,0x1b40U,0x3680U,0x6d00U,0xda00U,0xa40bU,0x581dU,
		0xb03aU,0x707fU,0xe0feU,0xd1f7U,0xb3e5U,0x77c1U,0xef82U,0xcf0fU,0x8e15U,0x0c21U,
		0x1842U,0x3084U,0x6108U,0xc210U,0x942bU,0x385dU,0x70baU,0xe174U,0xd2e3U,0xb5cdU,
		0x7b91U,0xf722U,0xfe4fU,0xec95U,0xc921U,0x8249U,0x1499U,0x2932U,0x5264U,0xa4c8U,
		0x599bU,0xb336U,0x7667U,0xecceU,0xc997U,0x8325U,0x1641U,0x2c82U,0x5904U,0xb208U,
		0x741bU,0xe836U,0xc067U,0x90c5U,0x3181U,0x6302U,0xc604U,0x9c03U,0x280dU,0x501aU,
		0xa034U,0x5063U,0xa0c6U,0x5187U,0xa30eU,0x5617U,0xac2eU,0x4857U,0x90aeU,0x3157U,
		0x62aeU,0xc55cU,0x9ab3U,0x256dU,0x4adaU,0x95b4U,0x3b63U,0x76c6U,0xed8cU,0xcb13U,
		0x862dU,0x1c51U,0x38a2U,0x7144U,0xe288U,0xd51bU,0xba3dU,0x6471U,0xc8e2U,0x81cfU,
		0x1395U,0x272aU,0x4e54U,0x9ca8U,0x295bU,0x52b6U,0xa56cU,0x5ad3U,0xb5a6U,0x7b47U,
		0xf68eU,0xfd17U,0xea25U,0xc441U,0x9889U,0x2119U,0x4232U,0x8464U,0x18c3U,0x3186U,
		0x630cU,0xc618U,0x9c3bU,0x287dU,0x50faU,0xa1f4U,0x53e3U,0xa7c6U,0x5f87U,0xbf0eU,
		0x6e17U,0xdc2eU,0xa857U,0x40a5U,0x814aU,0x129fU,0x253eU,0x4a7cU,0x94f8U,0x39fbU,
		0x73f6U,0xe7ecU,0xdfd3U,0xafadU,0x4f51U,0x9ea2U,0x2d4fU,0x5a9eU,0xb53cU,0x7a73U,
		0xf4e6U,0xf9c7U,0xe385U,0xd701U,0xbe09U,0x6c19U,0xd832U,0xa06fU,0x50d5U,0xa1aaU,
		0x535fU,0xa6beU,0x5d77U,0xbaeeU,0x65d7U,0xcbaeU,0x8757U,0x1ea5U,0x3d4aU,0x7a94U,
		0xf528U,0xfa5bU,0xe4bdU,0xd971U,0xa2e9U,0x55d9U,0xabb2U,0x476fU,0x8edeU,0x0db7U,
		0x1b6eU,0x36dcU,0x6db8U,0xdb70U,0xa6ebU,0x5dddU,0xbbbaU,0x677fU,0xcefeU,0x8df7U,
		0x0be5U,0x17caU,0x2f94U,0x5f28U,0xbe50U,0x6cabU,0xd956U,0xa2a7U,0x5545U,0xaa8aU,
		0x451fU,0x8a3eU,0x0477U,0x08eeU,0x11dcU,0x23b8U,0x4770U,0x8ee0U,0x0dcbU,0x1b96U,
		0x372cU,0x6e58U,0xdcb0U,0xa96bU,0x42ddU,0x85baU,0x1b7fU,0x36feU,0x6dfcU,0xdbf8U,
		0xa7fbU,0x5ffdU,0xbffaU,0x6fffU,0xdffeU,0xaff7U,0x4fe5U,0x9fcaU,0x2f9fU,0x5f3eU,
		0xbe7cU,0x6cf3U,0xd9e6U,0xa3c7U,0x5785U,0xaf0aU,0x4e1fU,0x9c3eU,0x2877U,0x50eeU,
		0xa1dcU,0x53b3U,0xa766U,0x5ec7U,0xbd8eU,0x6b17U,0xd62eU,0xbc57U,0x68a5U,0xd14aU,
		0xb29fU,0x7535U,0xea6aU,0xc4dfU,0x99b5U,0x2361U,0x46c2U,0x8d84U,0x0b03U,0x1606U,
		0x2c0cU,0x5818U,0xb030U,0x706bU,0xe0d6U,0xd1a7U,0xb345U,0x7681U,0xed02U,0xca0fU,
		0x8415U,0x1821U,0x3042U,0x6084U,0xc108U,0x921bU,0x343dU,0x687aU,0xd0f4U,0xb1e3U,
		0x73cdU,0xe79aU,0xdf3fU,0xae75U,0x4ce1U,0x99c2U,0x238fU,0x471eU,0x8e3cU,0x0c73U,
		0x18e6U,0x31ccU,0x6398U,0xc730U,0x9e6bU,0x2cddU,0x59baU,0xb374U,0x76e3U,0xedc6U,
		0xcb87U,0x8705U,0x1e01U,0x3c02U,0x7804U,0xf008U,0xf01bU,0xf03dU,0xf071U,0xf0e9U,
		0xf1d9U,0xf3b9U,0xf779U,0xfef9U,0xedf9U,0xcbf9U,0x87f9U,0x1ff9U,0x3ff2U,0x7fe4U,
		0xffc8U,0xef9bU,0xcf3dU,0x8e71U,0x0ce9U,0x19d2U,0x33a4U,0x6748U,0xce90U,0x8d2bU,
		0x0a5dU,0x14baU,0x2974U,0x52e8U,0xa5d0U,0x5babU,0xb756U,0x7ea7U,0xfd4eU,0xea97U,
		0xc525U,0x9a41U,0x2489U,0x4912U,0x9224U,0x3443U,0x6886U,0xd10cU,0xb213U,0x742dU,
		0xe85aU,0xc0bfU,0x9175U,0x32e1U,0x65c2U,0xcb84U,0x8703U,0x1e0dU,0x3c1aU,0x7834U,
		0xf068U,0xf0dbU,0xf1bdU,0xf371U,0xf6e9U,0xfdd9U,0xebb9U,0xc779U,0x9ef9U,0x2df9U,
		0x5bf2U,0xb7e4U,0x7fc3U,0xff86U,0xef07U,0xce05U,0x8c01U,0x0809U,0x1012U,0x2024U,
		0x4048U,0x8090U,0x112bU,0x2256U,0x44acU,0x8958U,0x02bbU,0x0576U,0x0aecU,0x15d8U,
		0x2bb0U,0x5760U,0xaec0U,0x4d8bU,0x9b16U,0x2627U,0x4c4eU,0x989cU,0x2133U,0x4266U,
		0x84ccU,0x1993U,0x3326U,0x664cU,0xcc98U,0x893bU,0x027dU,0x04faU,0x09f4U,0x13e8U,
		0x27d0U,0x4fa0U,0x9f40U,0x2e8bU,0x5d16U,0xba2cU,0x6453U,0xc8a6U,0x8147U,0x1285U,
		0x250aU,0x4a14U,0x9428U,0x385bU,0x70b6U,0xe16cU,0xd2d3U,0xb5adU,0x7b51U,0xf6a2U,
		0xfd4fU,0xea95U,0xc521U,0x9a49U,0x2499U,0x4932U,0x9264U,0x34c3U,0x6986U,0xd30cU,
		0xb613U,0x7c2dU,0xf85aU,0xe0bfU,0xd175U,0xb2e1U,0x75c9U,0xeb92U,0xc72fU,0x9e55U,
		0x2ca1U,0x5942U,0xb284U,0x7503U,0xea06U,0xc407U,0x9805U,0x2001U,0x4002U,0x8004U,
		0x1003U,0x2006U,0x400cU,0x8018U,0x103bU,0x2076U,0x40ecU,0x81d8U,0x13bbU,0x2776U,
		0x4eecU,0x9dd8U,0x2bbbU,0x5776U,0xaeecU,0x4dd3U,0x9ba6U,0x2747U,0x4e8eU,0x9d1cU,
		0x2a33U,0x5466U,0xa8ccU,0x4193U,0x8326U,0x1647U,0x2c8eU,0x591cU,0xb238U,0x747bU,
		0xe8f6U,0xc1e7U,0x93c5U,0x3781U,0x6f02U,0xde04U,0xac03U,0x480dU,0x901aU,0x303fU,
		0x607eU,0xc0fcU,0x91f3U,0x33edU,0x67daU,0xcfb4U,0x8f63U,0x0ecdU,0x1d9aU,0x3b34U,
		0x7668U,0xecd0U,0xc9abU,0x835dU,0x16b1U,0x2d62U,0x5ac4U,0xb588U,0x7b1bU,0xf636U,
		0xfc67U,0xe8c5U,0xc181U,0x9309U,0x3619U,0x6c32U,0xd864U,0xa0c3U,0x518dU,0xa31aU,
		0x563fU,0xac7eU,0x48f7U,0x91eeU,0x33d7U,0x67aeU,0xcf5cU,0x8eb3U,0x0d6dU,0x1adaU,
		0x35b4U,0x6b68U,0xd6d0U,0xbdabU,0x6b5dU,0xd6baU,0xbd7fU,0x6af5U,0xd5eaU,0xbbdfU,
		0x67b5U,0xcf6aU,0x8edfU,0x0db5U,0x1b6aU,0x36d4U,0x6da8U,0xdb50U,0xa6abU,0x5d5dU,
		0xbabaU,0x657fU,0xcafeU,0x85f7U,0x1be5U,0x37caU,0x6f94U,0xdf28U,0xae5bU,0x4cbdU,
		0x997aU,0x22ffU,0x45feU,0x8bfcU,0x07f3U,0x0fe6U,0x1fccU,0x3f98U,0x7f30U,0xfe60U,
		0xeccbU,0xc99dU,0x8331U,0x1669U,0x2cd2U,0x59a4U,0xb348U,0x769bU,0xed36U,0xca67U,
		0x84c5U,0x1981U,0x3302U,0x6604U,0xcc08U,0x881bU,0x003dU,0x007aU,0x00f4U,0x01e8U,
		0x03d0U,0x07a0U,0x0f40U,0x1e80U,0x3d00U,0x7a00U,0xf400U,0xf80bU,0xe01dU,0xd031U,
		0xb069U,0x70d9U,0xe1b2U,0xd36fU,0xb6d5U,0x7da1U,0xfb42U,0xe68fU,0xdd15U,0xaa21U,
		0x4449U,0x8892U,0x012fU,0x025eU,0x04bcU,0x0978U,0x12f0U,0x25e0U,0x4bc0U,0x9780U,
		0x3f0bU,0x7e16U,0xfc2cU,0xe853U,0xc0adU,0x9151U,0x32a9U,0x6552U,0xcaa4U,0x8543U,
		0x1a8dU,0x351aU,0x6a34U,0xd468U,0xb8dbU,0x61bdU,0xc37aU,0x96ffU,0x3df5U,0x7beaU,
		0xf7d4U,0xffa3U,0xef4dU,0xce91U,0x8d29U,0x0a59U,0x14b2U,0x2964U,0x52c8U,0xa590U,
		0x5b2bU,0xb656U,0x7ca7U,0xf94eU,0xe297U,0xd525U,0xba41U,0x6489U,0xc912U,0x822fU,
		0x1455U,0x28aaU,0x5154U,0xa2a8U,0x555bU,0xaab6U,0x4567U,0x8aceU,0x0597U,0x0b2eU,
		0x165cU,0x2cb8U,0x5970U,0xb2e0U,0x75cbU,0xeb96U,0xc727U,0x9e45U,0x2c81U,0x5902U,
		0xb204U,0x7403U,0xe806U,0xc007U,0x9005U,0x3001U,0x6002U,0xc004U,0x9003U,0x300dU,
		0x601aU,0xc034U,0x9063U,0x30cdU,0x619aU,0xc334U,0x9663U,0x3ccdU,0x799aU,0xf334U,
		0xf663U,0xfccdU,0xe991U,0xc329U,0x9659U,0x3cb9U,0x7972U,0xf2e4U,0xf5c3U,0xfb8dU,
		0xe711U,0xde29U,0xac59U,0x48b9U,0x9172U,0x32efU,0x65deU,0xcbbcU,0x8773U,0x1eedU,
		0x3ddaU,0x7bb4U,0xf768U,0xfedbU,0xedbdU,0xcb71U,0x86e9U,0x1dd9U,0x3bb2U,0x7764U,
		0xeec8U,0xcd9bU,0x8b3dU,0x0671U,0x0ce2U,0x19c4U,0x3388U,0x6710U,0xce20U,0x8c4bU,
		0x089dU,0x113aU,0x2274U,0x44e8U,0x89d0U,0x03abU,0x0756U,0x0eacU,0x1d58U,0x3ab0U,
		0x7560U,0xeac0U,0xc58bU,0x9b1dU,0x2631U,0x4c62U,0x98c4U,0x2183U,0x4306U,0x860cU,
		0x1c13U,0x3826U,0x704cU,0xe098U,0xd13bU,0xb27dU,0x74f1U,0xe9e2U,0xc3cfU,0x9795U,
		0x3f21U,0x7e42U,0xfc84U,0xe903U,0xc20dU,0x9411U,0x3829U,0x7052U,0xe0a4U,0xd143U,
		0xb28dU,0x7511U,0xea22U,0xc44fU,0x9895U,0x2121U,0x4242U,0x8484U,0x1903U,0x3206U,
		0x640cU,0xc818U,0x803bU,0x107dU,0x20faU,0x41f4U,0x83e8U,0x17dbU,0x2fb6U,0x5f6cU,
		0xbed8U,0x6dbbU,0xdb76U,0xa6e7U,0x5dc5U,0xbb8aU,0x671fU,0xce3eU,0x8c77U,0x08e5U,
		0x11caU,0x2394U,0x4728U,0x8e50U,0x0cabU,0x1956U,0x32acU,0x6558U,0xcab0U,0x856bU,
		0x1addU,0x35baU,0x6b74U,0xd6e8U,0xbddbU,0x6bbdU,0xd77aU,0xbeffU,0x6df5U,0xdbeaU,
		0xa7dfU,0x5fb5U,0xbf6aU,0x6edfU,0xddbeU,0xab77U,0x46e5U,0x8dcaU,0x0b9fU,0x173eU,
		0x2e7cU,0x5cf8U,0xb9f0U,0x63ebU,0xc7d6U,0x9fa7U,0x2f45U,0x5e8aU,0xbd14U,0x6a23U,
		0xd446U,0xb887U,0x6105U,0xc20aU,0x941fU,0x3835U,0x706aU,0xe0d4U,0xd1a3U,0xb34dU,
		0x7691U,0xed22U,0xca4fU,0x8495U,0x1921U,0x3242U,0x6484U,0xc908U,0x821bU,0x143dU,
		0x287aU,0x50f4U,0xa1e8U,0x53dbU,0xa7b6U,0x5f67U,0xbeceU,0x6d97U,0xdb2eU,0xa657U,
		0x5ca5U,0xb94aU,0x629fU,0xc53eU,0x9a77U,0x24e5U,0x49caU,0x9394U,0x3723U,0x6e46U,
		0xdc8cU,0xa913U,0x422dU,0x845aU,0x18bfU,0x317eU,0x62fcU,0xc5f8U,0x9bfbU,0x27fdU,
		0x4ffaU,0x9ff4U,0x2fe3U,0x5fc6U,0xbf8cU,0x6f13U,0xde26U,0xac47U,0x4885U,0x910aU,
		0x321fU,0x643eU,0xc87cU,0x80f3U,0x11edU,0x23daU,0x47b4U,0x8f68U,0x0edbU,0x1db6U,
		0x3b6cU,0x76d8U,0xedb0U,0xcb6bU,0x86ddU,0x1db1U,0x3b62U,0x76c4U,0xed88U,0xcb1bU,
		0x863dU,0x1c71U,0x38e2U,0x71c4U,0xe388U,0xd71bU,0xbe3dU,0x6c71U,0xd8e2U,0xa1cfU,
		0x5395U,0xa72aU,0x5e5fU,0xbcbeU,0x6977U,0xd2eeU,0xb5d7U,0x7ba5U,0xf74aU,0xfe9fU,
		0xed35U,0xca61U,0x84c9U,0x1999U,0x3332U,0x6664U,0xccc8U,0x899bU,0x033dU,0x067aU,
		0x0cf4U,0x19e8U,0x33d0U,0x67a0U,0xcf40U,0x8e8bU,0x0d1dU,0x1a3aU,0x3474U,0x68e8U,
		0xd1d0U,0xb3abU,0x775dU,0xeebaU,0xcd7fU,0x8af5U,0x05e1U,0x0bc2U,0x1784U,0x2f08U,
		0x5e10U,0xbc20U,0x684bU,0xd096U,0xb127U,0x7245U,0xe48aU,0xd91fU,0xa235U,0x5461U,
		0xa8c2U,0x418fU,0x831eU,0x1637U,0x2c6eU,0x58dcU,0xb1b8U,0x737bU,0xe6f6U,0xdde7U,
		0xabc5U,0x4781U,0x8f02U,0x0e0fU,0x1c1eU,0x383cU,0x7078U,0xe0f0U,0xd1ebU,0xb3ddU,
		0x77b1U,0xef62U,0xcecfU,0x8d95U,0x0b21U,0x1642U,0x2c84U,0x5908U,0xb210U,0x742bU,
		0xe856U,0xc0a7U,0x9145U,0x3281U,0x6502U,0xca04U,0x8403U,0x180dU,0x301aU,0x6034U,
		0xc068U,0x90dbU,0x31bdU,0x637aU,0xc6f4U,0x9de3U,0x2bcdU,0x579aU,0xaf34U,0x4e63U,
		0x9cc6U,0x2987U,0x530eU,0xa61cU,0x5c33U,0xb866U,0x60c7U,0xc18eU,0x9317U,0x3625U,
		0x6c4aU,0xd894U,0xa123U,0x524dU,0xa49aU,0x593fU,0xb27eU,0x74f7U,0xe9eeU,0xc3d7U,
		0x97a5U,0x3f41U,0x7e82U,0xfd04U,0xea03U,0xc40dU,0x9811U,0x2029U,0x4052U,0x80a4U,
		0x1143U,0x2286U,0x450cU,0x8a18U,0x043bU,0x0876U,0x10ecU,0x21d8U,0x43b0U,0x8760U,
		0x1ecbU,0x3d96U,0x7b2cU,0xf658U,0xfcbbU,0xe97dU,0xc2f1U,0x95e9U,0x3bd9U,0x77b2U,
		0xef64U,0xcec3U,0x8d8dU,0x0b11U,0x1622U,0x2c44U,0x5888U,0xb110U,0x722bU,0xe456U,
		0xd8a7U,0xa145U,0x5281U,0xa502U,0x5a0fU,0xb41eU,0x7837U,0xf06eU,0xf0d7U,0xf1a5U,
		0xf341U,0xf689U,0xfd19U,0xea39U,0xc479U,0x98f9U,0x21f9U,0x43f2U,0x87e4U,0x1fc3U,
		0x3f86U,0x7f0cU,0xfe18U,0xec3bU,0xc87dU,0x80f1U,0x11e9U,0x23d2U,0x47a4U,0x8f48U,
		0x0e9bU,0x1d36U,0x3a6cU,0x74d8U,0xe9b0U,0xc36bU,0x96ddU,0x3db1U,0x7b62U,0xf6c4U,
		0xfd83U,0xeb0dU,0xc611U,0x9c29U,0x2859U,0x50b2U,0xa164U,0x52c3U,0xa586U,0x5b07U,
		0xb60eU,0x7c17U,0xf82eU,0xe057U,0xd0a5U,0xb141U,0x7289U,0xe512U,0xda2fU,0xa455U,
		0x58a1U,0xb142U,0x728fU,0xe51eU,0xda37U,0xa465U,0x58c1U,0xb182U,0x730fU,0xe61eU,
		0xdc37U,0xa865U,0x40c1U,0x8182U,0x130fU,0x261eU,0x4c3cU,0x9878U,0x20fbU,0x41f6U,
		0x83ecU,0x17d3U,0x2fa6U,0x5f4cU,0xbe98U,0x6d3bU,0xda76U,0xa4e7U,0x59c5U,0xb38aU,
		0x771fU,0xee3eU,0xcc77U,0x88e5U,0x01c1U,0x0382U,0x0704U,0x0e08U,0x1c10U,0x3820U,
		0x7040U,0xe080U,0xd10bU,0xb21dU,0x7431U,0xe862U,0xc0cfU,0x9195U,0x3321U,0x6642U,
		0xcc84U,0x8903U,0x020dU,0x041aU,0x0834U,0x1068U,0x20d0U,0x41a0U,0x8340U,0x168bU,
		0x2d16U,0x5a2cU,0xb458U,0x78bbU,0xf176U,0xf2e7U,0xf5c5U,0xfb81U,0xe709U,0xde19U,
		0xac39U,0x4879U,0x90f2U,0x31efU,0x63deU,0xc7bcU,0x9f73U,0x2eedU,0x5ddaU,0xbbb4U,
		0x6763U,0xcec6U,0x8d87U,0x0b05U,0x160aU,0x2c14U,0x5828U,0xb050U,0x70abU,0xe156U,
		0xd2a7U,0xb545U,0x7a81U,0xf502U,0xfa0fU,0xe415U,0xd821U,0xa049U,0x5099U,0xa132U,
		0x526fU,0xa4deU,0x59b7U,0xb36eU,0x76d7U,0xedaeU,0xcb57U,0x86a5U,0x1d41U,0x3a82U,
		0x7504U,0xea08U,0xc41bU,0x983dU,0x2071U,0x40e2U,0x81c4U,0x1383U,0x2706U,0x4e0cU,
		0x9c18U,0x283bU,0x5076U,0xa0ecU,0x51d3U,0xa3a6U,0x5747U,0xae8eU,0x4d17U,0x9a2eU,
		0x2457U,0x48aeU,0x915cU,0x32b3U,0x6566U,0xcaccU,0x8593U,0x1b2dU,0x365aU,0x6cb4U,
		0xd968U,0xa2dbU,0x55bdU,0xab7aU,0x46ffU,0x8dfeU,0x0bf7U,0x17eeU,0x2fdcU,0x5fb8U,
		0xbf70U,0x6eebU,0xddd6U,0xaba7U,0x4745U,0x8e8aU,0x0d1fU,0x1a3eU,0x347cU,0x68f8U,
		0xd1f0U,0xb3ebU,0x77ddU,0xefbaU,0xcf7fU,0x8ef5U,0x0de1U,0x1bc2U,0x3784U,0x6f08U,
		0xde10U,0xac2bU,0x485dU,0x90baU,0x317fU,0x62feU,0xc5fcU,0x9bf3U,0x27edU,0x4fdaU,
		0x9fb4U,0x2f63U,0x5ec6U,0xbd8cU,0x6b13U,0xd626U,0xbc47U,0x6885U,0xd10aU,0xb21fU,
		0x7435U,0xe86aU,0xc0dfU,0x91b5U,0x3361U,0x66c2U,0xcd84U,0x8b03U,0x060dU,0x0c1aU,
		0x1834U,0x3068U,0x60d0U,0xc1a0U,0x934bU,0x369dU,0x6d3aU,0xda74U,0xa4e3U,0x59cdU,
		0xb39aU,0x773fU,0xee7eU,0xccf7U,0x89e5U,0x03c1U,0x0782U,0x0f04U,0x1e08U,0x3c10U,
		0x7820U,0xf040U,0xf08bU,0xf11dU,0xf231U,0xf469U,0xf8d9U,0xe1b9U,0xd379U,0xb6f9U,
		0x7df9U,0xfbf2U,0xe7efU,0xdfd5U,0xafa1U,0x4f49U,0x9e92U,0x2d2fU,0x5a5eU,0xb4bcU,
		0x7973U,0xf2e6U,0xf5c7U,0xfb85U,0xe701U,0xde09U,0xac19U,0x4839U,0x9072U,0x30efU,
		0x61deU,0xc3bcU,0x9773U,0x3eedU,0x7ddaU,0xfbb4U,0xe763U,0xdecdU,0xad91U,0x4b29U,
		0x9652U,0x3cafU,0x795eU,0xf2bcU,0xf573U,0xfaedU,0xe5d1U,0xdba9U,0xa759U,0x5eb9U,
		0xbd72U,0x6aefU,0xd5deU,0xbbb7U,0x6765U,0xcecaU,0x8d9fU,0x0b35U,0x166aU,0x2cd4U,
		0x59a8U,0xb350U,0x76abU,0xed56U,0xcaa7U,0x8545U,0x1a81U,0x3502U,0x6a04U,0xd408U,
		0xb81bU,0x603dU,0xc07aU,0x90ffU,0x31f5U,0x63eaU,0xc7d4U,0x9fa3U,0x2f4dU,0x5e9aU,
		0xbd34U,0x6a63U,0xd4c6U,0xb987U,0x6305U,0xc60aU,0x9c1fU,0x2835U,0x506aU,0xa0d4U,
		0x51a3U,0xa346U,0x5687U,0xad0eU,0x4a17U,0x942eU,0x3857U,0x70aeU,0xe15cU,0xd2b3U,
		0xb56dU,0x7ad1U,0xf5a2U,0xfb4fU,0xe695U,0xdd21U,0xaa49U,0x4499U,0x8932U,0x026fU,
		0x04deU,0x09bcU,0x1378U,0x26f0U,0x4de0U,0x9bc0U,0x278bU,0x4f16U,0x9e2cU,0x2c53U,
		0x58a6U,0xb14cU,0x7293U,0xe526U,0xda47U,0xa485U,0x5901U,0xb202U,0x740fU,0xe81eU,
		0xc037U,0x9065U,0x30c1U,0x6182U,0xc304U,0x9603U,0x3c0dU,0x781aU,0xf034U,0xf063U,
		0xf0cdU,0xf191U,0xf329U,0xf659U,0xfcb9U,0xe979U,0xc2f9U,0x95f9U,0x3bf9U,0x77f2U,
		0xefe4U,0xcfc3U,0x8f8dU,0x0f11U,0x1e22U,0x3c44U,0x7888U,0xf110U,0xf22bU,0xf45dU,
		0xf8b1U,0xe169U,0xd2d9U,0xb5b9U,0x7b79U,0xf6f2U,0xfdefU,0xebd5U,0xc7a1U,0x9f49U,
		0x2e99U,0x5d32U,0xba64U,0x64c3U,0xc986U,0x8307U,0x1605U,0x2c0aU,0x5814U,0xb028U,
		0x705bU,0xe0b6U,0xd167U,0xb2c5U,0x7581U,0xeb02U,0xc60fU,0x9c15U,0x2821U,0x5042U,
		0xa084U,0x5103U,0xa206U,0x5407U,0xa80eU,0x4017U,0x802eU,0x1057U,0x20aeU,0x415cU,
		0x82b8U,0x157bU,0x2af6U,0x55ecU,0xabd8U,0x47bbU,0x8f76U,0x0ee7U,0x1dceU,0x3b9cU,
		0x7738U,0xee70U,0xccebU,0x89ddU,0x03b1U,0x0762U,0x0ec4U,0x1d88U,0x3b10U,0x7620U,
		0xec40U,0xc88bU,0x811dU,0x1231U,0x2462U,0x48c4U,0x9188U,0x331bU,0x6636U,0xcc6cU,
		0x88d3U,0x01adU,0x035aU,0x06b4U,0x0d68U,0x1ad0U,0x35a0U,0x6b40U,0xd680U,0xbd0bU,
		0x6a1dU,0xd43aU,0xb87fU,0x60f5U,0xc1eaU,0x93dfU,0x37b5U,0x6f6aU,0xded4U,0xada3U,
		0x4b4dU,0x969aU,0x3d3fU,0x7a7eU,0xf4fcU,0xf9f3U,0xe3edU,0xd7d1U,0xbfa9U,0x6f59U,
		0xdeb2U,0xad6fU,0x4ad5U,0x95aaU,0x3b5fU,0x76beU,0xed7cU,0xcaf3U,0x85edU,0x1bd1U,
		0x37a2U,0x6f44U,0xde88U,0xad1bU,0x4a3dU,0x947aU,0x38ffU,0x71feU,0xe3fcU,0xd7f3U,
		0xbfedU,0x6fd1U,0xdfa2U,0xaf4fU,0x4e95U,0x9d2aU,0x2a5fU,0x54beU,0xa97cU,0x42f3U,
		0x85e6U,0x1bc7U,0x378eU,0x6f1cU,0xde38U,0xac7bU,0x48fdU,0x91faU,0x33ffU,0x67feU,
		0xcffcU,0x8ff3U,0x0fedU,0x1fdaU,0x3fb4U,0x7f68U,0xfed0U,0xedabU,0xcb5dU,0x86b1U,
		0x1d69U,0x3ad2U,0x75a4U,0xeb48U,0xc69bU,0x9d3dU,0x2a71U,0x54e2U,0xa9c4U,0x4383U,
		0x8706U,0x1e07U,0x3c0eU,0x781cU,0xf038U,0xf07bU,0xf0fdU,0xf1f1U,0xf3e9U,0xf7d9U,
		0xffb9U,0xef79U,0xcef9U,0x8df9U,0x0bf9U,0x17f2U,0x2fe4U,0x5fc8U,0xbf90U,0x6f2bU,
		0xde56U,0xaca7U,0x4945U,0x928aU,0x351fU,0x6a3eU,0xd47cU,0xb8f3U,0x61edU,0xc3daU,
		0x97bfU,0x3f75U,0x7eeaU,0xfdd4U,0xeba3U,0xc74dU,0x9e91U,0x2d29U,0x5a52U,0xb4a4U,
		0x7943U,0xf286U,0xf507U,0xfa05U,0xe401U,0xd809U,0xa019U,0x5039U,0xa072U,0x50efU,
		0xa1deU,0x53b7U,0xa76eU,0x5ed7U,0xbdaeU,0x6b57U,0xd6aeU,0xbd57U,0x6aa5U,0xd54aU,
		0xba9fU,0x6535U,0xca6aU,0x84dfU,0x19b5U,0x336aU,0x66d4U,0xcda8U,0x8b5bU,0x06bdU,
		0x0d7aU,0x1af4U,0x35e8U,0x6bd0U,0xd7a0U,0xbf4bU,0x6e9dU,0xdd3aU,0xaa7fU,0x44f5U,
		0x89eaU,0x03dfU,0x07beU,0x0f7cU,0x1ef8U,0x3df0U,0x7be0U,0xf7c0U,0xff8bU,0xef1dU,
		0xce31U,0x8c69U,0x08d9U,0x11b2U,0x2364U,0x46c8U,0x8d90U,0x0b2bU,0x1656U,0x2cacU,
		0x5958U,0xb2b0U,0x756bU,0xead6U,0xc5a7U,0x9b45U,0x2681U,0x4d02U,0x9a04U,0x2403U,
		0x4806U,0x900cU,0x3013U,0x6026U,0xc04cU,0x9093U,0x312dU,0x625aU,0xc4b4U,0x9963U,
		0x22cdU,0x459aU,0x8b34U,0x0663U,0x0cc6U,0x198cU,0x3318U,0x6630U,0xcc60U,0x88cbU,
		0x019dU,0x033aU,0x0674U,0x0ce8U,0x19d0U,0x33a0U,0x6740U,0xce80U,0x8d0bU,0x0a1dU,
		0x143aU,0x2874U,0x50e8U,0xa1d0U,0x53abU,0xa756U,0x5ea7U,0xbd4eU,0x6a97U,0xd52eU,
		0xba57U,0x64a5U,0xc94aU,0x829fU,0x1535U,0x2a6aU,0x54d4U,0xa9a8U,0x435bU,0x86b6U,
		0x1d67U,0x3aceU,0x759cU,0xeb38U,0xc67bU,0x9cfdU,0x29f1U,0x53e2U,0xa7c4U,0x5f83U,
		0xbf06U,0x6e07U,0xdc0eU,0xa817U,0x4025U,0x804aU,0x109fU,0x213eU,0x427cU,0x84f8U,
		0x19fbU,0x33f6U,0x67ecU,0xcfd8U,0x8fbbU,0x0f7dU,0x1efaU,0x3df4U,0x7be8U,0xf7d0U,
		0xffabU,0xef5dU,0xceb1U,0x8d69U,0x0ad9U,0x15b2U,0x2b64U,0x56c8U,0xad90U,0x4b2bU,
		0x9656U,0x3ca7U,0x794eU,0xf29cU,0xf533U,0xfa6dU,0xe4d1U,0xd9a9U,0xa359U,0x56b9U,
		0xad72U,0x4aefU,0x95deU,0x3bb7U,0x776eU,0xeedcU,0xcdb3U,0x8b6dU,0x06d1U,0x0da2U,
		0x1b44U,0x3688U,0x6d10U,0xda20U,0xa44bU,0x589dU,0xb13aU,0x727fU,0xe4feU,0xd9f7U,
		0xa3e5U,0x57c1U,0xaf82U,0x4f0fU,0x9e1eU,0x2c37U,0x586eU,0xb0dcU,0x71b3U,0xe366U,
		0xd6c7U,0xbd85U,0x6b01U,0xd602U,0xbc0fU,0x6815U,0xd02aU,0xb05fU,0x70b5U,0xe16aU,
		0xd2dfU,0xb5b5U,0x7b61U,0xf6c2U,0xfd8fU,0xeb15U,0xc621U,0x9c49U,0x2899U,0x5132U,
		0xa264U,0x54c3U,0xa986U,0x4307U,0x860eU,0x1c17U,0x382eU,0x705cU,0xe0b8U,0xd17bU,
		0xb2fdU,0x75f1U,0xebe2U,0xc7cfU,0x9f95U,0x2f21U,0x5e42U,0xbc84U,0x6903U,0xd206U,
		0xb407U,0x7805U,0xf00aU,0xf01fU,0xf035U,0xf061U,0xf0c9U,0xf199U,0xf339U,0xf679U,
		0xfcf9U,0xe9f9U,0xc3f9U,0x97f9U,0x3ff9U,0x7ff2U,0xffe4U,0xefc3U,0xcf8dU,0x8f11U,
		0x0e29U,0x1c52U,0x38a4U,0x7148U,0xe290U,0xd52bU,0xba5dU,0x64b1U,0xc962U,0x82cfU,
		0x1595U,0x2b2aU,0x5654U,0xaca8U,0x495bU,0x92b6U,0x3567U,0x6aceU,0xd59cU,0xbb33U,
		0x666dU,0xccdaU,0x89bfU,0x0375U,0x06eaU,0x0dd4U,0x1ba8U,0x3750U,0x6ea0U,0xdd40U,
		0xaa8bU,0x451dU,0x8a3aU,0x047fU,0x08feU,0x11fcU,0x23f8U,0x47f0U,0x8fe0U,0x0fcbU,
		0x1f96U,0x3f2cU,0x7e58U,0xfcb0U,0xe96bU,0xc2ddU,0x95b1U,0x3b69U,0x76d2U,0xeda4U,
		0xcb43U,0x868dU,0x1d11U,0x3a22U,0x7444U,0xe888U,0xc11bU,0x923dU,0x3471U,0x68e2U,
		0xd1c4U,0xb383U,0x770dU,0xee1aU,0xcc3fU,0x8875U,0x00e1U,0x01c2U,0x0384U,0x0708U,
		0x0e10U,0x1c20U,0x3840U,0x7080U,0xe100U,0xd20bU,0xb41dU,0x7831U,0xf062U,0xf0cfU,
		0xf195U,0xf321U,0xf649U,0xfc99U,0xe939U,0xc279U,0x94f9U,0x39f9U,0x73f2U,0xe7e4U,
		0xdfc3U,0xaf8dU,0x4f11U,0x9e22U,0x2c4fU,0x589eU,0xb13cU,0x7273U,0xe4e6U,0xd9c7U,
		0xa385U,0x5701U,0xae02U,0x4c0fU,0x981eU,0x2037U,0x406eU,0x80dcU,0x11b3U,0x2366U,
		0x46ccU,0x8d98U,0x0b3bU,0x1676U,0x2cecU,0x59d8U,0xb3b0U,0x776bU,0xeed6U,0xcda7U,
		0x8b45U,0x0681U,0x0d02U,0x1a04U,0x3408U,0x6810U,0xd020U,0xb04bU,0x709dU,0xe13aU,
		0xd27fU,0xb4f5U,0x79e1U,0xf3c2U,0xf78fU,0xff15U,0xee21U,0xcc49U,0x8899U,0x0139U,
		0x0272U,0x04e4U,0x09c8U,0x1390U,0x2720U,0x4e40U,0x9c80U,0x290bU,0x5216U,0xa42cU,
		0x5853U,0xb0a6U,0x7147U,0xe28eU,0xd517U,0xba25U,0x6441U,0xc882U,0x810fU,0x1215U,
		0x242aU,0x4854U,0x90a8U,0x315bU,0x62b6U,0xc56cU,0x9ad3U,0x25adU,0x4b5aU,0x96b4U,
		0x3d63U,0x7ac6U,0xf58cU,0xfb13U,0xe62dU,0xdc51U,0xa8a9U,0x4159U,0x82b2U,0x156fU,
		0x2adeU,0x55bcU,0xab78U,0x46fbU,0x8df6U,0x0be7U,0x17ceU,0x2f9cU,0x5f38U,0xbe70U,
		0x6cebU,0xd9d6U,0xa3a7U,0x5745U,0xae8aU,0x4d1fU,0x9a3eU,0x2477U,0x48eeU,0x91dcU,
		0x33b3U,0x6766U,0xceccU,0x8d93U,0x0b2dU,0x165aU,0x2cb4U,0x5968U,0xb2d0U,0x75abU,
		0xeb56U,0xc6a7U,0x9d45U,0x2a81U,0x5502U,0xaa04U,0x4403U,0x8806U,0x0007U,0x000eU,
		0x001cU,0x0038U,0x0070U,0x00e0U,0x01c0U,0x0380U,0x0700U,0x0e00U,0x1c00U,0x3800U,
		0x7000U,0xe000U,0xd00bU,0xb01dU,0x7031U,0xe062U,0xd0cfU,0xb195U,0x7321U,0xe642U,
		0xdc8fU,0xa915U,0x4221U,0x8442U,0x188fU,0x311eU,0x623cU,0xc478U,0x98fbU,0x21fdU,
		0x43faU,0x87f4U,0x1fe3U,0x3fc6U,0x7f8cU,0xff18U,0xee3bU,0xcc7dU,0x88f1U,0x01e9U,
		0x03d2U,0x07a4U,0x0f48U,0x1e90U,0x3d20U,0x7a40U,0xf480U,0xf90bU,0xe21dU,0xd431U,
		0xb869U,0x60d9U,0xc1b2U,0x936fU,0x36d5U,0x6daaU,0xdb54U,0xa6a3U,0x5d4dU,0xba9aU,
		0x653fU,0xca7eU,0x84f7U,0x19e5U,0x33caU,0x6794U,0xcf28U,0x8e5bU,0x0cbdU,0x197aU,
		0x32f4U,0x65e8U,0xcbd0U,0x87abU,0x1f5dU,0x3ebaU,0x7d74U,0xfae8U,0xe5dbU,0xdbbdU,
		0xa771U,0x5ee9U,0xbdd2U,0x6bafU,0xd75eU,0xbeb7U,0x6d65U,0xdacaU,0xa59fU,0x5b35U,
		0xb66aU,0x7cdfU,0xf9beU,0xe377U,0xd6e5U,0xbdc1U,0x6b89U,0xd712U,0xbe2fU,0x6c55U,
		0xd8aaU,0xa15fU,0x52b5U,0xa56aU,0x5adfU,0xb5beU,0x7b77U,0xf6eeU,0xfdd7U,0xeba5U,
		0xc741U,0x9e89U,0x2d19U,0x5a32U,0xb464U,0x78c3U,0xf186U,0xf307U,0xf605U,0xfc01U,
		0xe809U,0xc019U,0x9039U,0x3079U,0x60f2U,0xc1e4U,0x93c3U,0x378dU,0x6f1aU,0xde34U,
		0xac63U,0x48cdU,0x919aU,0x333fU,0x667eU,0xccfcU,0x89f3U,0x03edU,0x07daU,0x0fb4U,
		0x1f68U,0x3ed0U,0x7da0U,0xfb40U,0xe68bU,0xdd1dU,0xaa31U,0x4469U,0x88d2U,0x01afU,
		0x035eU,0x06bcU,0x0d78U,0x1af0U,0x35e0U,0x6bc0U,0xd780U,0xbf0bU,0x6e1dU,0xdc3aU,
		0xa87fU,0x40f5U,0x81eaU,0x13dfU,0x27beU,0x4f7cU,0x9ef8U,0x2dfbU,0x5bf6U,0xb7ecU,
		0x7fd3U,0xffa6U,0xef47U,0xce85U,0x8d01U,0x0a09U,0x1412U,0x2824U,0x5048U,0xa090U,
		0x512bU,0xa256U,0x54a7U,0xa94eU,0x4297U,0x852eU,0x1a57U,0x34aeU,0x695cU,0xd2b8U,
		0xb57bU,0x7afdU,0xf5faU,0xfbffU,0xe7f5U,0xdfe1U,0xafc9U,0x4f99U,0x9f32U,0x2e6fU,
		0x5cdeU,0xb9bcU,0x6373U,0xc6e6U,0x9dc7U,0x2b85U,0x570aU,0xae14U,0x4c23U,0x9846U,
		0x2087U,0x410eU,0x821cU,0x1433U,0x2866U,0x50ccU,0xa198U,0x533bU,0xa676U,0x5ce7U,
		0xb9ceU,0x6397U,0xc72eU,0x9e57U,0x2ca5U,0x594aU,0xb294U,0x7523U,0xea46U,0xc487U,
		0x9905U,0x2201U,0x4402U,0x8804U,0x0003U,0x0006U,0x000cU,0x0018U,0x0030U,0x0060U,
		0x00c0U,0x0180U,0x0300U,0x0600U,0x0c00U,0x1800U,0x3000U,0x6000U,0xc000U,0x900bU,
		0x301dU,0x603aU,0xc074U,0x90e3U,0x31cdU,0x639aU,0xc734U,0x9e63U,0x2ccdU,0x599aU,
		0xb334U,0x7663U,0xecc6U,0xc987U,0x8305U,0x1601U,0x2c02U,0x5804U,0xb008U,0x701bU,
		0xe036U,0xd067U,0xb0c5U,0x7181U,0xe302U,0xd60fU,0xbc15U,0x6821U,0xd042U,0xb08fU,
		0x7115U,0xe22aU,0xd45fU,0xb8b5U,0x6161U,0xc2c2U,0x958fU,0x3b15U,0x762aU,0xec54U,
		0xc8a3U,0x814dU,0x1291U,0x2522U,0x4a44U,0x9488U,0x391bU,0x7236U,0xe46cU,0xd8d3U,
		0xa1adU,0x5351U,0xa6a2U,0x5d4fU,0xba9eU,0x6537U,0xca6eU,0x84d7U,0x19a5U,0x334aU,
		0x6694U,0xcd28U,0x8a5bU,0x04bdU,0x097aU,0x12f4U,0x25e8U,0x4bd0U,0x97a0U,0x3f4bU,
		0x7e96U,0xfd2cU,0xea53U,0xc4adU,0x9951U,0x22a9U,0x4552U,0x8aa4U,0x0543U,0x0a86U,
		0x150cU,0x2a18U,0x5430U,0xa860U,0x40cbU,0x8196U,0x1327U,0x264eU,0x4c9cU,0x9938U,
		0x227bU,0x44f6U,0x89ecU,0x03d3U,0x07a6U,0x0f4cU,0x1e98U,0x3d30U,0x7a60U,0xf4c0U,
		0xf98bU,0xe31dU,0xd631U,0xbc69U,0x68d9U,0xd1b2U,0xb36fU,0x76d5U,0xedaaU,0xcb5fU,
		0x86b5U,0x1d61U,0x3ac2U,0x7584U,0xeb08U,0xc61bU,0x9c3dU,0x2871U,0x50e2U,0xa1c4U,
		0x5383U,0xa706U,0x5e07U,0xbc0eU,0x6817U,0xd02eU,0xb057U,0x70a5U,0xe14aU,0xd29fU,
		0xb535U,0x7a61U,0xf4c2U,0xf98fU,0xe315U,0xd621U,0xbc49U,0x6899U,0xd132U,0xb26fU,
		0x74d5U,0xe9aaU,0xc35fU,0x96b5U,0x3d61U,0x7ac2U,0xf584U,0xfb03U,0xe60dU,0xdc11U,
		0xa829U,0x4059U,0x80b2U,0x116fU,0x22deU,0x45bcU,0x8b78U,0x06fbU,0x0df6U,0x1becU,
		0x37d8U,0x6fb0U,0xdf60U,0xaecbU,0x4d9dU,0x9b3aU,0x267fU,0x4cfeU,0x99fcU,0x23f3U,
		0x47e6U,0x8fccU,0x0f93U,0x1f26U,0x3e4cU,0x7c98U,0xf930U,0xe26bU,0xd4ddU,0xb9b1U,
		0x6369U,0xc6d2U,0x9dafU,0x2b55U,0x56aaU,0xad54U,0x4aa3U,0x9546U,0x3a87U,0x750eU,
		0xea1cU,0xc433U,0x986dU,0x20d1U,0x41a2U,0x8344U,0x1683U,0x2d06U,0x5a0cU,0xb418U,
		0x783bU,0xf076U,0xf0e7U,0xf1c5U,0xf381U,0xf709U,0xfe19U,0xec39U,0xc879U,0x80f9U,
		0x11f9U,0x23f2U,0x47e4U,0x8fc8U,0x0f9bU,0x1f36U,0x3e6cU,0x7cd8U,0xf9b0U,0xe36bU,
		0xd6ddU,0xbdb1U,0x6b69U,0xd6d2U,0xbdafU,0x6b55U,0xd6aaU,0xbd5fU,0x6ab5U,0xd56aU,
		0xbadfU,0x65b5U,0xcb6aU,0x86dfU,0x1db5U,0x3b6aU,0x76d4U,0xeda8U,0xcb5bU,0x86bdU,
		0x1d71U,0x3ae2U,0x75c4U,0xeb88U,0xc71bU,0x9e3dU,0x2c71U,0x58e2U,0xb1c4U,0x7383U,
		0xe706U,0xde07U,0xac05U,0x4801U,0x9002U,0x300fU,0x601eU,0xc03cU,0x9073U,0x30edU,
		0x61daU,0xc3b4U,0x9763U,0x3ecdU,0x7d9aU,0xfb34U,0xe663U,0xdccdU,0xa991U,0x4329U,
		0x8652U,0x1cafU,0x395eU,0x72bcU,0xe578U,0xdafbU,0xa5fdU,0x5bf1U,0xb7e2U,0x7fcfU,
		0xff9eU,0xef37U,0xce65U,0x8cc1U,0x0989U,0x1312U,0x2624U,0x4c48U,0x9890U,0x212bU,
		0x4256U,0x84acU,0x1953U,0x32a6U,0x654cU,0xca98U,0x853bU,0x1a7dU,0x34faU,0x69f4U,
		0xd3e8U,0xb7dbU,0x7fbdU,0xff7aU,0xeeffU,0xcdf5U,0x8be1U,0x07c9U,0x0f92U,0x1f24U,
		0x3e48U,0x7c90U,0xf920U,0xe24bU,0xd49dU,0xb931U,0x6269U,0xc4d2U,0x99afU,0x2355U,
		0x46aaU,0x8d54U,0x0aa3U,0x1546U,0x2a8cU,0x5518U,0xaa30U,0x446bU,0x88d6U,0x01a7U,
		0x034eU,0x069cU,0x0d38U,0x1a70U,0x34e0U,0x69c0U,0xd380U,0xb70bU,0x7e1dU,0xfc3aU,
		0xe87fU,0xc0f5U,0x91e1U,0x33c9U,0x6792U,0xcf24U,0x8e43U,0x0c8dU,0x191aU,0x3234U,
		0x6468U,0xc8d0U,0x81abU,0x135dU,0x26baU,0x4d74U,0x9ae8U,0x25dbU,0x4bb6U,0x976cU,
		0x3ed3U,0x7da6U,0xfb4cU,0xe693U,0xdd2dU,0xaa51U,0x44a9U,0x8952U,0x02afU,0x055eU,
		0x0abcU,0x1578U,0x2af0U,0x55e0U,0xabc0U,0x478bU,0x8f16U,0x0e27U,0x1c4eU,0x389cU,
		0x7138U,0xe270U,0xd4ebU,0xb9ddU,0x63b1U,0xc762U,0x9ecfU,0x2d95U,0x5b2aU,0xb654U,
		0x7ca3U,0xf946U,0xe287U,0xd505U,0xba01U,0x6409U,0xc812U,0x802fU,0x1055U,0x20aaU,
		0x4154U,0x82a8U,0x155bU,0x2ab6U,0x556cU,0xaad8U,0x45bbU,0x8b76U,0x06e7U,0x0dceU,
		0x1b9cU,0x3738U,0x6e70U,0xdce0U,0xa9cbU,0x439dU,0x873aU,0x1e7fU,0x3cfeU,0x79fcU,
		0xf3f8U,0xf7fbU,0xfffdU,0xeff1U,0xcfe9U,0x8fd9U,0x0fb9U,0x1f72U,0x3ee4U,0x7dc8U,
		0xfb90U,0xe72bU,0xde5dU,0xacb1U,0x4969U,0x92d2U,0x35afU,0x6b5eU,0xd6bcU,0xbd73U,
		0x6aedU,0xd5daU,0xbbbfU,0x6775U,0xceeaU,0x8ddfU,0x0bb5U,0x176aU,0x2ed4U,0x5da8U,
		0xbb50U,0x66abU,0xcd56U,0x8aa7U,0x0545U,0x0a8aU,0x1514U,0x2a28U,0x5450U,0xa8a0U,
		0x414bU,0x8296U,0x1527U,0x2a4eU,0x549cU,0xa938U,0x427bU,0x84f6U,0x19e7U,0x33ceU,
		0x679cU,0xcf38U,0x8e7bU,0x0cfdU,0x19faU,0x33f4U,0x67e8U,0xcfd0U,0x8fabU,0x0f5dU,
		0x1ebaU,0x3d74U,0x7ae8U,0xf5d0U,0xfbabU,0xe75dU,0xdeb1U,0xad69U,0x4ad9U,0x95b2U,
		0x3b6fU,0x76deU,0xedbcU,0xcb73U,0x86edU,0x1dd1U,0x3ba2U,0x7744U,0xee88U,0xcd1bU,
		0x8a3dU,0x0471U,0x08e2U,0x11c4U,0x2388U,0x4710U,0x8e20U,0x0c4bU,0x1896U,0x312cU,
		0x6258U,0xc4b0U,0x996bU,0x22ddU,0x45baU,0x8b74U,0x06e3U,0x0dc6U,0x1b8cU,0x3718U,
		0x6e30U,0xdc60U,0xa8cbU,0x419dU,0x833aU,0x167fU,0x2cfeU,0x59fcU,0xb3f8U,0x77fbU,
		0xeff6U,0xcfe7U,0x8fc5U,0x0f81U,0x1f02U,0x3e04U,0x7c08U,0xf810U,0xe02bU,0xd05dU,
		0xb0b1U,0x7169U,0xe2d2U,0xd5afU,0xbb55U,0x66a1U,0xcd42U,0x8a8fU,0x0515U,0x0a2aU,
		0x1454U,0x28a8U,0x5150U,0xa2a0U,0x554bU,0xaa96U,0x4527U,0x8a4eU,0x0497U,0x092eU,
		0x125cU,0x24b8U,0x4970U,0x92e0U,0x35cbU,0x6b96U,0xd72cU,0xbe53U,0x6cadU,0xd95aU,
		0xa2bfU,0x5575U,0xaaeaU,0x45dfU,0x8bbeU,0x0777U,0x0eeeU,0x1ddcU,0x3bb8U,0x7770U,
		0xeee0U,0xcdcbU,0x8b9dU,0x0731U,0x0e62U,0x1cc4U,0x3988U,0x7310U,0xe620U,0xdc4bU,
		0xa89dU,0x4131U,0x8262U,0x14cfU,0x299eU,0x533cU,0xa678U,0x5cfbU,0xb9f6U,0x63e7U,
		0xc7ceU,0x9f97U,0x2f25U,0x5e4aU,0xbc94U,0x6923U,0xd246U,0xb487U,0x7905U,0xf20aU,
		0xf41fU,0xf835U,0xe061U,0xd0c9U,0xb199U,0x7339U,0xe672U,0xdcefU,0xa9d5U,0x43a1U,
		0x8742U,0x1e8fU,0x3d1eU,0x7a3cU,0xf478U,0xf8fbU,0xe1fdU,0xd3f1U,0xb7e9U,0x7fd9U,
		0xffb2U,0xef6fU,0xced5U,0x8da1U,0x0b49U,0x1692U,0x2d24U,0x5a48U,0xb490U,0x792bU,
		0xf256U,0xf4a7U,0xf945U,0xe281U,0xd509U,0xba19U,0x6439U,0xc872U,0x80efU,0x11d5U,
		0x23aaU,0x4754U,0x8ea8U,0x0d5bU,0x1ab6U,0x356cU,0x6ad8U,0xd5b0U,0xbb6bU,0x66ddU,
		0xcdbaU,0x8b7fU,0x06f5U,0x0deaU,0x1bd4U,0x37a8U,0x6f50U,0xdea0U,0xad4bU,0x4a9dU,
		0x953aU,0x3a7fU,0x74feU,0xe9fcU,0xc3f3U,0x97edU,0x3fd1U,0x7fa2U,0xff44U,0xee83U,
		0xcd0dU,0x8a11U,0x0429U,0x0852U,0x10a4U,0x2148U,0x4290U,0x8520U,0x1a4bU,0x3496U,
		0x692cU,0xd258U,0xb4bbU,0x797dU,0xf2faU,0xf5ffU,0xfbf5U,0xe7e1U,0xdfc9U,0xaf99U,
		0x4f39U,0x9e72U,0x2cefU,0x59deU,0xb3bcU,0x7773U,0xeee6U,0xcdc7U,0x8b85U,0x0701U,
		0x0e02U,0x1c04U,0x3808U,0x7010U,0xe020U,0xd04bU,0xb09dU,0x7131U,0xe262U,0xd4cfU,
		0xb995U,0x6321U,0xc642U,0x9c8fU,0x2915U,0x522aU,0xa454U,0x58a3U,0xb146U,0x7287U,
		0xe50eU,0xda17U,0xa425U,0x5841U,0xb082U,0x710fU,0xe21eU,0xd437U,0xb865U,0x60c1U,
		0xc182U,0x930fU,0x3615U,0x6c2aU,0xd854U,0xa0a3U,0x514dU,0xa29aU,0x553fU,0xaa7eU,
		0x44f7U,0x89eeU,0x03d7U,0x07aeU,0x0f5cU,0x1eb8U,0x3d70U,0x7ae0U,0xf5c0U,0xfb8bU,
		0xe71dU,0xde31U,0xac69U,0x48d9U,0x91b2U,0x336fU,0x66deU,0xcdbcU,0x8b73U,0x06edU,
		0x0ddaU,0x1bb4U,0x3768U,0x6ed0U,0xdda0U,0xab4bU,0x469dU,0x8d3aU,0x0a7fU,0x14feU,
		0x29fcU,0x53f8U,0xa7f0U,0x5febU,0xbfd6U,0x6fa7U,0xdf4eU,0xae97U,0x4d25U,0x9a4aU,
		0x249fU,0x493eU,0x927cU,0x34f3U,0x69e6U,0xd3ccU,0xb793U,0x7f2dU,0xfe5aU,0xecbfU,
		0xc975U,0x82e1U,0x15c9U,0x2b92U,0x5724U,0xae48U,0x4c9bU,0x9936U,0x2267U,0x44ceU,
		0x899cU,0x0333U,0x0666U,0x0cccU,0x1998U,0x3330U,0x6660U,0xccc0U,0x898bU,0x031dU,
		0x063aU,0x0c74U,0x18e8U,0x31d0U,0x63a0U,0xc740U,0x9e8bU,0x2d1dU,0x5a3aU,0xb474U,
		0x78e3U,0xf1c6U,0xf387U,0xf705U,0xfe01U,0xec09U,0xc819U,0x8039U,0x1079U,0x20f2U,
		0x41e4U,0x83c8U,0x179bU,0x2f36U,0x5e6cU,0xbcd8U,0x69bbU,0xd376U,0xb6e7U,0x7dc5U,
		0xfb8aU,0xe71fU,0xde35U,0xac61U,0x48c9U,0x9192U,0x332fU,0x665eU,0xccbcU,0x8973U,
		0x02edU,0x05daU,0x0bb4U,0x1768U,0x2ed0U,0x5da0U,0xbb40U,0x668bU,0xcd16U,0x8a27U,
		0x0445U,0x088aU,0x1114U,0x2228U,0x4450U,0x88a0U,0x014bU,0x0296U,0x052cU,0x0a58U,
		0x14b0U,0x2960U,0x52c0U,0xa580U,0x5b0bU,0xb616U,0x7c27U,0xf84eU,0xe097U,0xd125U,
		0xb241U,0x7489U,0xe912U,0xc22fU,0x9455U,0x38a1U,0x7142U,0xe284U,0xd503U,0xba0dU,
		0x6411U,0xc822U,0x804fU,0x1095U,0x212aU,0x4254U,0x84a8U,0x195bU,0x32b6U,0x656cU,
		0xcad8U,0x85bbU,0x1b7dU,0x36faU,0x6df4U,0xdbe8U,0xa7dbU,0x5fbdU,0xbf7aU,0x6effU,
		0xddfeU,0xabf7U,0x47e5U,0x8fcaU,0x0f9fU,0x1f3eU,0x3e7cU,0x7cf8U,0xf9f0U,0xe3ebU,
		0xd7ddU,0xbfb1U,0x6f69U,0xded2U,0xadafU,0x4b55U,0x96aaU,0x3d5fU,0x7abeU,0xf57cU,
		0xfaf3U,0xe5edU,0xdbd1U,0xa7a9U,0x5f59U,0xbeb2U,0x6d6fU,0xdadeU,0xa5b7U,0x5b65U,
		0xb6caU,0x7d9fU,0xfb3eU,0xe677U,0xdce5U,0xa9c1U,0x4389U,0x8712U,0x1e2fU,0x3c5eU,
		0x78bcU,0xf178U,0xf2fbU,0xf5fdU,0xfbf1U,0xe7e9U,0xdfd9U,0xafb9U,0x4f79U,0x9ef2U,
		0x2defU,0x5bdeU,0xb7bcU,0x7f73U,0xfee6U,0xedc7U,0xcb85U,0x8701U,0x1e09U,0x3c12U,
		0x7824U,0xf048U,0xf09bU,0xf13dU,0xf271U,0xf4e9U,0xf9d9U,0xe3b9U,0xd779U,0xbef9U,
		0x6df9U,0xdbf2U,0xa7efU,0x5fd5U,0xbfaaU,0x6f5fU,0xdebeU,0xad77U,0x4ae5U,0x95caU,
		0x3b9fU,0x773eU,0xee7cU,0xccf3U,0x89edU,0x03d1U,0x07a2U,0x0f44U,0x1e88U,0x3d10U,
		0x7a20U,0xf440U,0xf88bU,0xe11dU,0xd231U,0xb469U,0x78d9U,0xf1b2U,0xf36fU,0xf6d5U,
		0xfda1U,0xeb49U,0xc699U,0x9d39U,0x2a79U,0x54f2U,0xa9e4U,0x43c3U,0x8786U,0x1f07U,
		0x3e0eU,0x7c1cU,0xf838U,0xe07bU,0xd0fdU,0xb1f1U,0x73e9U,0xe7d2U,0xdfafU,0xaf55U,
		0x4ea1U,0x9d42U,0x2a8fU,0x551eU,0xaa3cU,0x4473U,0x88e6U,0x01c7U,0x038eU,0x071cU,
		0x0e38U,0x1c70U,0x38e0U,0x71c0U,0xe380U,0xd70bU,0xbe1dU,0x6c31U,0xd862U,0xa0cfU,
		0x5195U,0xa32aU,0x565fU,0xacbeU,0x4977U,0x92eeU,0x35d7U,0x6baeU,0xd75cU,0xbeb3U,
		0x6d6dU,0xdadaU,0xa5bfU,0x5b75U,0xb6eaU,0x7ddfU,0xfbbeU,0xe777U,0xdee5U,0xadc1U,
		0x4b89U,0x9712U,0x3e2fU,0x7c5eU,0xf8bcU,0xe173U,0xd2edU,0xb5d1U,0x7ba9U,0xf752U,
		0xfeafU,0xed55U,0xcaa1U,0x8549U,0x1a99U,0x3532U,0x6a64U,0xd4c8U,0xb99bU,0x633dU,
		0xc67aU,0x9cffU,0x29f5U,0x53eaU,0xa7d4U,0x5fa3U,0xbf46U,0x6e87U,0xdd0eU,0xaa17U,
		0x4425U,0x884aU,0x009fU,0x013eU,0x027cU,0x04f8U,0x09f0U,0x13e0U,0x27c0U,0x4f80U,
		0x9f00U,0x2e0bU,0x5c16U,0xb82cU,0x6053U,0xc0a6U,0x9147U,0x3285U,0x650aU,0xca14U,
		0x8423U,0x184dU,0x309aU,0x6134U,0xc268U,0x94dbU,0x39bdU,0x737aU,0xe6f4U,0xdde3U,
		0xabcdU,0x4791U,0x8f22U,0x0e4fU,0x1c9eU,0x393cU,0x7278U,0xe4f0U,0xd9ebU,0xa3ddU,
		0x57b1U,0xaf62U,0x4ecfU,0x9d9eU,0x2b37U,0x566eU,0xacdcU,0x49b3U,0x9366U,0x36c7U,
		0x6d8eU,0xdb1cU,0xa633U,0x5c6dU,0xb8daU,0x61bfU,0xc37eU,0x96f7U,0x3de5U,0x7bcaU,
		0xf794U,0xff23U,0xee4dU,0xcc91U,0x8929U,0x0259U,0x04b2U,0x0964U,0x12c8U,0x2590U,
		0x4b20U,0x9640U,0x3c8bU,0x7916U,0xf22cU,0xf453U,0xf8adU,0xe151U,0xd2a9U,0xb559U,
		0x7ab9U,0xf572U,0xfaefU,0xe5d5U,0xdba1U,0xa749U,0x5e99U,0xbd32U,0x6a6fU,0xd4deU,
		0xb9b7U,0x6365U,0xc6caU,0x9d9fU,0x2b35U,0x566aU,0xacd4U,0x49a3U,0x9346U,0x3687U,
		0x6d0eU,0xda1cU,0xa433U,0x586dU,0xb0daU,0x71bfU,0xe37eU,0xd6f7U,0xbde5U,0x6bc1U,
		0xd782U,0xbf0fU,0x6e15U,0xdc2aU,0xa85fU,0x40b5U,0x816aU,0x12dfU,0x25beU,0x4b7cU,
		0x96f8U,0x3dfbU,0x7bf6U,0xf7ecU,0xffd3U,0xefadU,0xcf51U,0x8ea9U,0x0d59U,0x1ab2U,
		0x3564U,0x6ac8U,0xd590U,0xbb2bU,0x665dU,0xccbaU,0x897fU,0x02f5U,0x05eaU,0x0bd4U,
		0x17a8U,0x2f50U,0x5ea0U,0xbd40U,0x6a8bU,0xd516U,0xba27U,0x6445U,0xc88aU,0x811fU,
		0x1235U,0x246aU,0x48d4U,0x91a8U,0x335bU,0x66b6U,0xcd6cU,0x8ad3U,0x05adU,0x0b5aU,
		0x16b4U,0x2d68U,0x5ad0U,0xb5a0U,0x7b4bU,0xf696U,0xfd27U,0xea45U,0xc481U,0x9909U,
		0x2219U,0x4432U,0x8864U,0x00c3U,0x0186U,0x030cU,0x0618U,0x0c30U,0x1860U,0x30c0U,
		0x6180U,0xc300U,0x960bU,0x3c1dU,0x783aU,0xf074U,0xf0e3U,0xf1cdU,0xf391U,0xf729U,
		0xfe59U,0xecb9U,0xc979U,0x82f9U,0x15f9U,0x2bf2U,0x57e4U,0xafc8U,0x4f9bU,0x9f36U,
		0x2e67U,0x5cceU,0xb99cU,0x6333U,0xc666U,0x9cc7U,0x2985U,0x530aU,0xa614U,0x5c23U,
		0xb846U,0x6087U,0xc10eU,0x9217U,0x3425U,0x684aU,0xd094U,0xb123U,0x724dU,0xe49aU,
		0xd93fU,0xa275U,0x54e1U,0xa9c2U,0x438fU,0x871eU,0x1e37U,0x3c6eU,0x78dcU,0xf1b8U,
		0xf37bU,0xf6fdU,0xfdf1U,0xebe9U,0xc7d9U,0x9fb9U,0x2f79U,0x5ef2U,0xbde4U,0x6bc3U,
		0xd786U,0xbf07U,0x6e05U,0xdc0aU,0xa81fU,0x4035U,0x806aU,0x10dfU,0x21beU,0x437cU,
		0x86f8U,0x1dfbU,0x3bf6U,0x77ecU,0xefd8U,0xcfbbU,0x8f7dU,0x0ef1U,0x1de2U,0x3bc4U,
		0x7788U,0xef10U,0xce2bU,0x8c5dU,0x08b1U,0x1162U,0x22c4U,0x4588U,0x8b10U,0x062bU,
		0x0c56U,0x18acU,0x3158U,0x62b0U,0xc560U,0x9acbU,0x259dU,0x4b3aU,0x9674U,0x3ce3U,
		0x79c6U,0xf38cU,0xf713U,0xfe2dU,0xec51U,0xc8a9U,0x8159U,0x12b9U,0x2572U,0x4ae4U,
		0x95c8U,0x3b9bU,0x7736U,0xee6cU,0xccd3U,0x89adU,0x0351U,0x06a2U,0x0d44U,0x1a88U,
		0x3510U,0x6a20U,0xd440U,0xb88bU,0x611dU,0xc23aU,0x947fU,0x38f5U,0x71eaU,0xe3d4U,
		0xd7a3U,0xbf4dU,0x6e91U,0xdd22U,0xaa4fU,0x4495U,0x892aU,0x025fU,0x04beU,0x097cU,
		0x12f8U,0x25f0U,0x4be0U,0x97c0U,0x3f8bU,0x7f16U,0xfe2cU,0xec53U,0xc8adU,0x8151U,
		0x12a9U,0x2552U,0x4aa4U,0x9548U,0x3a9bU,0x7536U,0xea6cU,0xc4d3U,0x99adU,0x2351U,
		0x46a2U,0x8d44U,0x0a83U,0x1506U,0x2a0cU,0x5418U,0xa830U,0x406bU,0x80d6U,0x11a7U,
		0x234eU,0x469cU,0x8d38U,0x0a7bU,0x14f6U,0x29ecU,0x53d8U,0xa7b0U,0x5f6bU,0xbed6U,
		0x6da7U,0xdb4eU,0xa697U,0x5d25U,0xba4aU,0x649fU,0xc93eU,0x8277U,0x14e5U,0x29caU,
		0x5394U,0xa728U,0x5e5bU,0xbcb6U,0x6967U,0xd2ceU,0xb597U,0x7b25U,0xf64aU,0xfc9fU,
		0xe935U,0xc261U,0x94c9U,0x3999U,0x7332U,0xe664U,0xdcc3U,0xa98dU,0x4311U,0x8622U,
		0x1c4fU,0x389eU,0x713cU,0xe278U,0xd4fbU,0xb9fdU,0x63f1U,0xc7e2U,0x9fcfU,0x2f95U,
		0x5f2aU,0xbe54U,0x6ca3U,0xd946U,0xa287U,0x5505U,0xaa0aU,0x441fU,0x883eU,0x0077U,
		0x00eeU,0x01dcU,0x03b8U,0x0770U,0x0ee0U,0x1dc0U,0x3b80U,0x7700U,0xee00U,0xcc0bU,
		0x881dU,0x0031U,0x0062U,0x00c4U,0x0188U,0x0310U,0x0620U,0x0c40U,0x1880U,0x3100U,
		0x6200U,0xc400U,0x980bU,0x201dU,0x403aU,0x8074U,0x10e3U,0x21c6U,0x438cU,0x8718U,
		0x1e3bU,0x3c76U,0x78ecU,0xf1d8U,0xf3bbU,0xf77dU,0xfef1U,0xede9U,0xcbd9U,0x87b9U,
		0x1f79U,0x3ef2U,0x7de4U,0xfbc8U,0xe79bU,0xdf3dU,0xae71U,0x4ce9U,0x99d2U,0x23afU,
		0x475eU,0x8ebcU,0x0d73U,0x1ae6U,0x35ccU,0x6b98U,0xd730U,0xbe6bU,0x6cddU,0xd9baU,
		0xa37fU,0x56f5U,0xadeaU,0x4bdfU,0x97beU,0x3f77U,0x7eeeU,0xfddcU,0xebb3U,0xc76dU,
		0x9ed1U,0x2da9U,0x5b52U,0xb6a4U,0x7d43U,0xfa86U,0xe507U,0xda05U,0xa401U,0x5809U,
		0xb012U,0x702fU,0xe05eU,0xd0b7U,0xb165U,0x72c1U,0xe582U,0xdb0fU,0xa615U,0x5c21U,
		0xb842U,0x608fU,0xc11eU,0x9237U,0x3465U,0x68caU,0xd194U,0xb323U,0x764dU,0xec9aU,
		0xc93fU,0x8275U,0x14e1U,0x29c2U,0x5384U,0xa708U,0x5e1bU,0xbc36U,0x6867U,0xd0ceU,
		0xb197U,0x7325U,0xe64aU,0xdc9fU,0xa935U,0x4261U,0x84c2U,0x198fU,0x331eU,0x663cU,
		0xcc78U,0x88fbU,0x01fdU,0x03faU,0x07f4U,0x0fe8U,0x1fd0U,0x3fa0U,0x7f40U,0xfe80U,
		0xed0bU,0xca1dU,0x8431U,0x1869U,0x30d2U,0x61a4U,0xc348U,0x969bU,0x3d3dU,0x7a7aU,
		0xf4f4U,0xf9e3U,0xe3cdU,0xd791U,0xbf29U,0x6e59U,0xdcb2U,0xa96fU,0x42d5U,0x85aaU,
		0x1b5fU,0x36beU,0x6d7cU,0xdaf8U,0xa5fbU,0x5bfdU,0xb7faU,0x7fffU,0xfffeU,0xeff7U,
		0xcfe5U,0x8fc1U,0x0f89U,0x1f12U,0x3e24U,0x7c48U,0xf890U,0xe12bU,0xd25dU,0xb4b1U,
		0x7969U,0xf2d2U,0xf5afU,0xfb55U,0xe6a1U,0xdd49U,0xaa99U,0x4539U,0x8a72U,0x04efU,
		0x09deU,0x13bcU,0x2778U,0x4ef0U,0x9de0U,0x2bcbU,0x5796U,0xaf2cU,0x4e53U,0x9ca6U,
		0x2947U,0x528eU,0xa51cU,0x5a33U,0xb466U,0x78c7U,0xf18eU,0xf317U,0xf625U,0xfc41U,
		0xe889U,0xc119U,0x9239U,0x3479U,0x68f2U,0xd1e4U,0xb3c3U,0x778dU,0xef1aU,0xce3fU,
		0x8c75U,0x08e1U,0x11c2U,0x2384U,0x4708U,0x8e10U,0x0c2bU,0x1856U,0x30acU,0x6158U,
		0xc2b0U,0x956bU,0x3addU,0x75baU,0xeb74U,0xc6e3U,0x9dcdU,0x2b91U,0x5722U,0xae44U,
		0x4c83U,0x9906U,0x2207U,0x440eU,0x881cU,0x0033U,0x0066U,0x00ccU,0x0198U,0x0330U,
		0x0660U,0x0cc0U,0x1980U,0x3300U,0x6600U,0xcc00U,0x880bU,0x001dU,0x003aU,0x0074U,
		0x00e8U,0x01d0U,0x03a0U,0x0740U,0x0e80U,0x1d00U,0x3a00U,0x7400U,0xe800U,0xc00bU,
		0x901dU,0x3031U,0x6062U,0xc0c4U,0x9183U,0x330dU,0x661aU,0xcc34U,0x8863U,0x00cdU,
		0x019aU,0x0334U,0x0668U,0x0cd0U,0x19a0U,0x3340U,0x6680U,0xcd00U,0x8a0bU,0x041dU,
		0x083aU,0x1074U,0x20e8U,0x41d0U,0x83a0U,0x174bU,0x2e96U,0x5d2cU,0xba58U,0x64bbU,
		0xc976U,0x82e7U,0x15c5U,0x2b8aU,0x5714U,0xae28U,0x4c5bU,0x98b6U,0x2167U,0x42ceU,
		0x859cU,0x1b33U,0x3666U,0x6cccU,0xd998U,0xa33bU,0x567dU,0xacfaU,0x49ffU,0x93feU,
		0x37f7U,0x6feeU,0xdfdcU,0xafb3U,0x4f6dU,0x9edaU,0x2dbfU,0x5b7eU,0xb6fcU,0x7df3U,
		0xfbe6U,0xe7c7U,0xdf85U,0xaf01U,0x4e09U,0x9c12U,0x282fU,0x505eU,0xa0bcU,0x5173U,
		0xa2e6U,0x55c7U,0xab8eU,0x4717U,0x8e2eU,0x0c57U,0x18aeU,0x315cU,0x62b8U,0xc570U,
		0x9aebU,0x25ddU,0x4bbaU,0x9774U,0x3ee3U,0x7dc6U,0xfb8cU,0xe713U,0xde2dU,0xac51U,
		0x48a9U,0x9152U,0x32afU,0x655eU,0xcabcU,0x8573U,0x1aedU,0x35daU,0x6bb4U,0xd768U,
		0xbedbU,0x6dbdU,0xdb7aU,0xa6ffU,0x5df5U,0xbbeaU,0x67dfU,0xcfbeU,0x8f77U,0x0ee5U,
		0x1dcaU,0x3b94U,0x7728U,0xee50U,0xccabU,0x895dU,0x02b1U,0x0562U,0x0ac4U,0x1588U,
		0x2b10U,0x5620U,0xac40U,0x488bU,0x9116U,0x3227U,0x644eU,0xc89cU,0x8133U,0x126dU,
		0x24daU,0x49b4U,0x9368U,0x36dbU,0x6db6U,0xdb6cU,0xa6d3U,0x5dadU,0xbb5aU,0x66bfU,
		0xcd7eU,0x8af7U,0x05e5U,0x0bcaU,0x1794U,0x2f28U,0x5e50U,0xbca0U,0x694bU,0xd296U,
		0xb527U,0x7a45U,0xf48aU,0xf91fU,0xe235U,0xd461U,0xb8c9U,0x6199U,0xc332U,0x966fU,
		0x3cd5U,0x79aaU,0xf354U,0xf6a3U,0xfd4dU,0xea91U,0xc529U,0x9a59U,0x24b9U,0x4972U,
		0x92e4U,0x35c3U,0x6b86U,0xd70cU,0xbe13U,0x6c2dU,0xd85aU,0xa0bfU,0x5175U,0xa2eaU,
		0x55dfU,0xabbeU,0x4777U,0x8eeeU,0x0dd7U,0x1baeU,0x375cU,0x6eb8U,0xdd70U,0xaaebU,
		0x45ddU,0x8bbaU,0x077fU,0x0efeU,0x1dfcU,0x3bf8U,0x77f0U,0xefe0U,0xcfcbU,0x8f9dU,
		0x0f31U,0x1e62U,0x3cc4U,0x7988U,0xf310U,0xf62bU,0xfc5dU,0xe8b1U,0xc169U,0x92d9U,
		0x35b9U,0x6b72U,0xd6e4U,0xbdc3U,0x6b8dU,0xd71aU,0xbe3fU,0x6c75U,0xd8eaU,0xa1dfU,
		0x53b5U,0xa76aU,0x5edfU,0xbdbeU,0x6b77U,0xd6eeU,0xbdd7U,0x6ba5U,0xd74aU,0xbe9fU,
		0x6d35U,0xda6aU,0xa4dfU,0x59b5U,0xb36aU,0x76dfU,0xedbeU,0xcb77U,0x86e5U,0x1dc1U,
		0x3b82U,0x7704U,0xee08U,0xcc1bU,0x883dU,0x0071U,0x00e2U,0x01c4U,0x0388U,0x0710U,
		0x0e20U,0x1c40U,0x3880U,0x7100U,0xe200U,0xd40bU,0xb81dU,0x6031U,0xc062U,0x90cfU,
		0x3195U,0x632aU,0xc654U,0x9ca3U,0x294dU,0x529aU,0xa534U,0x5a63U,0xb4c6U,0x7987U,
		0xf30eU,0xf617U,0xfc25U,0xe841U,0xc089U,0x9119U,0x3239U,0x6472U,0xc8e4U,0x81c3U,
		0x138dU,0x271aU,0x4e34U,0x9c68U,0x28dbU,0x51b6U,0xa36cU,0x56d3U,0xada6U,0x4b47U,
		0x968eU,0x3d17U,0x7a2eU,0xf45cU,0xf8b3U,0xe16dU,0xd2d1U,0xb5a9U,0x7b59U,0xf6b2U,
		0xfd6fU,0xead5U,0xc5a1U,0x9b49U,0x2699U,0x4d32U,0x9a64U,0x24c3U,0x4986U,0x930cU,
		0x3613U,0x6c26U,0xd84cU,0xa093U,0x512dU,0xa25aU,0x54bfU,0xa97eU,0x42f7U,0x85eeU,
		0x1bd7U,0x37aeU,0x6f5cU,0xdeb8U,0xad7bU,0x4afdU,0x95faU,0x3bffU,0x77feU,0xeffcU,
		0xcff3U,0x8fedU,0x0fd1U,0x1fa2U,0x3f44U,0x7e88U,0xfd10U,0xea2bU,0xc45dU,0x98b1U,
		0x2169U,0x42d2U,0x85a4U,0x1b43U,0x3686U,0x6d0cU,0xda18U,0xa43bU,0x587dU,0xb0faU,
		0x71ffU,0xe3feU,0xd7f7U,0xbfe5U,0x6fc1U,0xdf82U,0xaf0fU,0x4e15U,0x9c2aU,0x285fU,
		0x50beU,0xa17cU,0x52f3U,0xa5e6U,0x5bc7U,0xb78eU,0x7f17U,0xfe2eU,0xec57U,0xc8a5U,
		0x8141U,0x1289U,0x2512U,0x4a24U,0x9448U,0x389bU,0x7136U,0xe26cU,0xd4d3U,0xb9adU,
		0x6351U,0xc6a2U,0x9d4fU,0x2a95U,0x552aU,0xaa54U,0x44a3U,0x8946U,0x0287U,0x050eU,
		0x0a1cU,0x1438U,0x2870U,0x50e0U,0xa1c0U,0x538bU,0xa716U,0x5e27U,0xbc4eU,0x6897U,
		0xd12eU,0xb257U,0x74a5U,0xe94aU,0xc29fU,0x9535U,0x3a61U,0x74c2U,0xe984U,0xc303U,
		0x960dU,0x3c11U,0x7822U,0xf044U,0xf083U,0xf10dU,0xf211U,0xf429U,0xf859U,0xe0b9U,
		0xd179U,0xb2f9U,0x75f9U,0xebf2U,0xc7efU,0x9fd5U,0x2fa1U,0x5f42U,0xbe84U,0x6d03U,
		0xda06U,0xa407U,0x5805U,0xb00aU,0x701fU,0xe03eU,0xd077U,0xb0e5U,0x71c1U,0xe382U,
		0xd70fU,0xbe15U,0x6c21U,0xd842U,0xa08fU,0x5115U,0xa22aU,0x545fU,0xa8beU,0x4177U,
		0x82eeU,0x15d7U,0x2baeU,0x575cU,0xaeb8U,0x4d7bU,0x9af6U,0x25e7U,0x4bceU,0x979cU,
		0x3f33U,0x7e66U,0xfcccU,0xe993U,0xc32dU,0x9651U,0x3ca9U,0x7952U,0xf2a4U,0xf543U,
		0xfa8dU,0xe511U,0xda29U,0xa459U,0x58b9U,0xb172U,0x72efU,0xe5deU,0xdbb7U,0xa765U,
		0x5ec1U,0xbd82U,0x6b0fU,0xd61eU,0xbc37U,0x6865U,0xd0caU,0xb19fU,0x7335U,0xe66aU,
		0xdcdfU,0xa9b5U,0x4361U,0x86c2U,0x1d8fU,0x3b1eU,0x763cU,0xec78U,0xc8fbU,0x81fdU,
		0x13f1U,0x27e2U,0x4fc4U,0x9f88U,0x2f1bU,0x5e36U,0xbc6cU,0x68d3U,0xd1a6U,0xb347U,
		0x7685U,0xed0aU,0xca1fU,0x8435U,0x1861U,0x30c2U,0x6184U,0xc308U,0x961bU,0x3c3dU,
		0x787aU,0xf0f4U,0xf1e3U,0xf3cdU,0xf791U,0xff29U,0xee59U,0xccb9U,0x8979U,0x02f9U,
		0x05f2U,0x0be4U,0x17c8U,0x2f90U,0x5f20U,0xbe40U,0x6c8bU,0xd916U,0xa227U,0x5445U,
		0xa88aU,0x411fU,0x823eU,0x1477U,0x28eeU,0x51dcU,0xa3b8U,0x577bU,0xaef6U,0x4de7U,
		0x9bceU,0x2797U,0x4f2eU,0x9e5cU,0x2cb3U,0x5966U,0xb2ccU,0x7593U,0xeb26U,0xc647U,
		0x9c85U,0x2901U,0x5202U,0xa404U,0x5803U,0xb006U,0x7007U,0xe00eU,0xd017U,0xb025U,
		0x7041U,0xe082U,0xd10fU,0xb215U,0x7421U,0xe842U,0xc08fU,0x9115U,0x3221U,0x6442U,
		0xc884U,0x8103U,0x120dU,0x241aU,0x4834U,0x9068U,0x30dbU,0x61b6U,0xc36cU,0x96d3U,
		0x3dadU,0x7b5aU,0xf6b4U,0xfd63U,0xeacdU,0xc591U,0x9b29U,0x2659U,0x4cb2U,0x9964U,
		0x22c3U,0x4586U,0x8b0cU,0x0613U,0x0c26U,0x184cU,0x3098U,0x6130U,0xc260U,0x94cbU,
		0x399dU,0x733aU,0xe674U,0xdce3U,0xa9cdU,0x4391U,0x8722U,0x1e4fU,0x3c9eU,0x793cU,
		0xf278U,0xf4fbU,0xf9fdU,0xe3f1U,0xd7e9U,0xbfd9U,0x6fb9U,0xdf72U,0xaeefU,0x4dd5U,
		0x9baaU,0x275fU,0x4ebeU,0x9d7cU,0x2af3U,0x55e6U,0xabccU,0x4793U,0x8f26U,0x0e47U,
		0x1c8eU,0x391cU,0x7238U,0xe470U,0xd8ebU,0xa1ddU,0x53b1U,0xa762U,0x5ecfU,0xbd9eU,
		0x6b37U,0xd66eU,0xbcd7U,0x69a5U,0xd34aU,0xb69fU,0x7d35U,0xfa6aU,0xe4dfU,0xd9b5U,
		0xa361U,0x56c9U,0xad92U,0x4b2fU,0x965eU,0x3cb7U,0x796eU,0xf2dcU,0xf5b3U,0xfb6dU,
		0xe6d1U,0xdda9U,0xab59U,0x46b9U,0x8d72U,0x0aefU,0x15deU,0x2bbcU,0x5778U,0xaef0U,
		0x4debU,0x9bd6U,0x27a7U,0x4f4eU,0x9e9cU,0x2d33U,0x5a66U,0xb4ccU,0x7993U,0xf326U,
		0xf647U,0xfc85U,0xe901U,0xc209U,0x9419U,0x3839U,0x7072U,0xe0e4U,0xd1c3U,0xb38dU,
		0x7711U,0xee22U,0xcc4fU,0x8895U,0x0121U,0x0242U,0x0484U,0x0908U,0x1210U,0x2420U,
		0x4840U,0x9080U,0x310bU,0x6216U,0xc42cU,0x9853U,0x20adU,0x415aU,0x82b4U,0x1563U,
		0x2ac6U,0x558cU,0xab18U,0x463bU,0x8c76U,0x08e7U,0x11ceU,0x239cU,0x4738U,0x8e70U,
		0x0cebU,0x19d6U,0x33acU,0x6758U,0xceb0U,0x8d6bU,0x0addU,0x15baU,0x2b74U,0x56e8U,
		0xadd0U,0x4babU,0x9756U,0x3ea7U,0x7d4eU,0xfa9cU,0xe533U,0xda6dU,0xa4d1U,0x59a9U,
		0xb352U,0x76afU,0xed5eU,0xcab7U,0x8565U,0x1ac1U,0x3582U,0x6b04U,0xd608U,0xbc1bU,
		0x683dU,0xd07aU,0xb0ffU,0x71f5U,0xe3eaU,0xd7dfU,0xbfb5U,0x6f61U,0xdec2U,0xad8fU,
		0x4b15U,0x962aU,0x3c5fU,0x78beU,0xf17cU,0xf2f3U,0xf5edU,0xfbd1U,0xe7a9U,0xdf59U,
		0xaeb9U,0x4d79U,0x9af2U,0x25efU,0x4bdeU,0x97bcU,0x3f73U,0x7ee6U,0xfdccU,0xeb93U,
		0xc72dU,0x9e51U,0x2ca9U,0x5952U,0xb2a4U,0x7543U,0xea86U,0xc507U,0x9a05U,0x2401U,
		0x4802U,0x9004U,0x3003U,0x6006U,0xc00cU,0x9013U,0x302dU,0x605aU,0xc0b4U,0x9163U,
		0x32cdU,0x659aU,0xcb34U,0x8663U,0x1ccdU,0x399aU,0x7334U,0xe668U,0xdcdbU,0xa9bdU,
		0x4371U,0x86e2U,0x1dcfU,0x3b9eU,0x773cU,0xee78U,0xccfbU,0x89fdU,0x03f1U,0x07e2U,
		0x0fc4U,0x1f88U,0x3f10U,0x7e20U,0xfc40U,0xe88bU,0xc11dU,0x9231U,0x3469U,0x68d2U,
		0xd1a4U,0xb343U,0x768dU,0xed1aU,0xca3fU,0x8475U,0x18e1U,0x31c2U,0x6384U,0xc708U,
		0x9e1bU,0x2c3dU,0x587aU,0xb0f4U,0x71e3U,0xe3c6U,0xd787U,0xbf05U,0x6e01U,0xdc02U,
		0xa80fU,0x4015U,0x802aU,0x105fU,0x20beU,0x417cU,0x82f8U,0x15fbU,0x2bf6U,0x57ecU,
		0xafd8U,0x4fbbU,0x9f76U,0x2ee7U,0x5dceU,0xbb9cU,0x6733U,0xce66U,0x8cc7U,0x0985U,
		0x130aU,0x2614U,0x4c28U,0x9850U,0x20abU,0x4156U,0x82acU,0x1553U,0x2aa6U,0x554cU,
		0xaa98U,0x453bU,0x8a76U,0x04e7U,0x09ceU,0x139cU,0x2738U,0x4e70U,0x9ce0U,0x29cbU,
		0x5396U,0xa72cU,0x5e53U,0xbca6U,0x6947U,0xd28eU,0xb517U,0x7a25U,0xf44aU,0xf89fU,
		0xe135U,0xd261U,0xb4c9U,0x7999U,0xf332U,0xf66fU,0xfcd5U,0xe9a1U,0xc349U,0x9699U,
		0x3d39U,0x7a72U,0xf4e4U,0xf9c3U,0xe38dU,0xd711U,0xbe29U,0x6c59U,0xd8b2U,0xa16fU,
		0x52d5U,0xa5aaU,0x5b5fU,0xb6beU,0x7d77U,0xfaeeU,0xe5d7U,0xdba5U,0xa741U,0x5e89U,
		0xbd12U,0x6a2fU,0xd45eU,0xb8b7U,0x6165U,0xc2caU,0x959fU,0x3b35U,0x766aU,0xecd4U,
		0xc9a3U,0x834dU,0x1691U,0x2d22U,0x5a44U,0xb488U,0x791bU,0xf236U,0xf467U,0xf8c5U,
		0xe181U,0xd309U,0xb619U,0x7c39U,0xf872U,0xe0efU,0xd1d5U,0xb3a1U,0x7749U,0xee92U,
		0xcd2fU,0x8a55U,0x04a1U,0x0942U,0x1284U,0x2508U,0x4a10U,0x9420U,0x384bU,0x7096U,
		0xe12cU,0xd253U,0xb4adU,0x7951U,0xf2a2U,0xf54fU,0xfa95U,0xe521U,0xda49U,0xa499U,
		0x5939U,0xb272U,0x74efU,0xe9deU,0xc3b7U,0x9765U,0x3ec1U,0x7d82U,0xfb04U,0xe603U,
		0xdc0dU,0xa811U,0x4029U,0x8052U,0x10afU,0x215eU,0x42bcU,0x8578U,0x1afbU,0x35f6U,
		0x6becU,0xd7d8U,0xbfbbU,0x6f7dU,0xdefaU,0xadffU,0x4bf5U,0x97eaU,0x3fdfU,0x7fbeU,
		0xff7cU,0xeef3U,0xcdedU,0x8bd1U,0x07a9U,0x0f52U,0x1ea4U,0x3d48U,0x7a90U,0xf520U,
		0xfa4bU,0xe49dU,0xd931U,0xa269U,0x54d9U,0xa9b2U,0x436fU,0x86deU,0x1db7U,0x3b6eU,
		0x76dcU,0xedb8U,0xcb7bU,0x86fdU,0x1df1U,0x3be2U,0x77c4U,0xef88U,0xcf1bU,0x8e3dU,
		0x0c71U,0x18e2U,0x31c4U,0x6388U,0xc710U,0x9e2bU,0x2c5dU,0x58baU,0xb174U,0x72e3U,
		0xe5c6U,0xdb87U,0xa705U,0x5e01U,0xbc02U,0x680fU,0xd01eU,0xb037U,0x7065U,0xe0caU,
		0xd19fU,0xb335U,0x7661U,0xecc2U,0xc98fU,0x8315U,0x1621U,0x2c42U,0x5884U,0xb108U,
		0x721bU,0xe436U,0xd867U,0xa0c5U,0x5181U,0xa302U,0x560fU,0xac1eU,0x4837U,0x906eU,
		0x30d7U,0x61aeU,0xc35cU,0x96b3U,0x3d6dU,0x7adaU,0xf5b4U,0xfb63U,0xe6cdU,0xdd91U,
		0xab29U,0x4659U,0x8cb2U,0x096fU,0x12deU,0x25bcU,0x4b78U,0x96f0U,0x3debU,0x7bd6U,
		0xf7acU,0xff53U,0xeeadU,0xcd51U,0x8aa9U,0x0559U,0x0ab2U,0x1564U,0x2ac8U,0x5590U,
		0xab20U,0x464bU,0x8c96U,0x0927U,0x124eU,0x249cU,0x4938U,0x9270U,0x34ebU,0x69d6U,
		0xd3acU,0xb753U,0x7eadU,0xfd5aU,0xeabfU,0xc575U,0x9ae1U,0x25c9U,0x4b92U,0x9724U,
		0x3e43U,0x7c86U,0xf90cU,0xe213U,0xd42dU,0xb851U,0x60a9U,0xc152U,0x92afU,0x3555U,
		0x6aaaU,0xd554U,0xbaa3U,0x654dU,0xca9aU,0x853fU,0x1a75U,0x34eaU,0x69d4U,0xd3a8U,
		0xb75bU,0x7ebdU,0xfd7aU,0xeaffU,0xc5f5U,0x9be1U,0x27c9U,0x4f92U,0x9f24U,0x2e43U,
		0x5c86U,0xb90cU,0x6213U,0xc426U,0x9847U,0x2085U,0x410aU,0x8214U,0x1423U,0x2846U,
		0x508cU,0xa118U,0x523bU,0xa476U,0x58e7U,0xb1ceU,0x7397U,0xe72eU,0xde57U,0xaca5U,
		0x4941U,0x9282U,0x350fU,0x6a1eU,0xd43cU,0xb873U,0x60edU,0xc1daU,0x93bfU,0x3775U,
		0x6eeaU,0xddd4U,0xaba3U,0x474dU,0x8e9aU,0x0d3fU,0x1a7eU,0x34fcU,0x69f8U,0xd3f0U,
		0xb7ebU,0x7fddU,0xffbaU,0xef7fU,0xcef5U,0x8de1U,0x0bc9U,0x1792U,0x2f24U,0x5e48U,
		0xbc90U,0x692bU,0xd256U,0xb4a7U,0x7945U,0xf28aU,0xf51fU,0xfa35U,0xe461U,0xd8c9U,
		0xa199U,0x5339U,0xa672U,0x5cefU,0xb9deU,0x63b7U,0xc76eU,0x9ed7U,0x2da5U,0x5b4aU,
		0xb694U,0x7d23U,0xfa46U,0xe487U,0xd905U,0xa201U,0x5409U,0xa812U,0x402fU,0x805eU,
		0x10b7U,0x216eU,0x42dcU,0x85b8U,0x1b7bU,0x36f6U,0x6decU,0xdbd8U,0xa7bbU,0x5f7dU,
		0xbefaU,0x6dffU,0xdbfeU,0xa7f7U,0x5fe5U,0xbfcaU,0x6f9fU,0xdf3eU,0xae77U,0x4ce5U,
		0x99caU,0x239fU,0x473eU,0x8e7cU,0x0cf3U,0x19e6U,0x33ccU,0x6798U,0xcf30U,0x8e6bU,
		0x0cddU,0x19baU,0x3374U,0x66e8U,0xcdd0U,0x8babU,0x075dU,0x0ebaU,0x1d74U,0x3ae8U,
		0x75d0U,0xeba0U,0xc74bU,0x9e9dU,0x2d31U,0x5a62U,0xb4c4U,0x7983U,0xf306U,0xf607U,
		0xfc05U,0xe801U,0xc009U,0x9019U,0x3039U,0x6072U,0xc0e4U,0x91c3U,0x338dU,0x671aU,
		0xce34U,0x8c63U,0x08cdU,0x119aU,0x2334U,0x4668U,0x8cd0U,0x09abU,0x1356U,0x26acU,
		0x4d58U,0x9ab0U,0x256bU,0x4ad6U,0x95acU,0x3b53U,0x76a6U,0xed4cU,0xca93U,0x852dU,
		0x1a51U,0x34a2U,0x6944U,0xd288U,0xb51bU,0x7a3dU,0xf47aU,0xf8ffU,0xe1f5U,0xd3e1U,
		0xb7c9U,0x7f99U,0xff32U,0xee6fU,0xccd5U,0x89a1U,0x0349U,0x0692U,0x0d24U,0x1a48U,
		0x3490U,0x6920U,0xd240U,0xb48bU,0x791dU,0xf23aU,0xf47fU,0xf8f5U,0xe1e1U,0xd3c9U,
		0xb799U,0x7f39U,0xfe72U,0xecefU,0xc9d5U,0x83a1U,0x1749U,0x2e92U,0x5d24U,0xba48U,
		0x649bU,0xc936U,0x8267U,0x14c5U,0x298aU,0x5314U,0xa628U,0x5c5bU,0xb8b6U,0x6167U,
		0xc2ceU,0x9597U,0x3b25U,0x764aU,0xec94U,0xc923U,0x824dU,0x1491U,0x2922U,0x5244U,
		0xa488U,0x591bU,0xb236U,0x7467U,0xe8ceU,0xc197U,0x9325U,0x3641U,0x6c82U,0xd904U,
		0xa203U,0x540dU,0xa81aU,0x403fU,0x807eU,0x10f7U,0x21eeU,0x43dcU,0x87b8U,0x1f7bU,
		0x3ef6U,0x7decU,0xfbd8U,0xe7bbU,0xdf7dU,0xaef1U,0x4de9U,0x9bd2U,0x27afU,0x4f5eU,
		0x9ebcU,0x2d73U,0x5ae6U,0xb5ccU,0x7b93U,0xf726U,0xfe47U,0xec85U,0xc901U,0x8209U,
		0x1419U,0x2832U,0x5064U,0xa0c8U,0x519bU,0xa336U,0x5667U,0xacceU,0x4997U,0x932eU,
		0x3657U,0x6caeU,0xd95cU,0xa2b3U,0x556dU,0xaadaU,0x45bfU,0x8b7eU,0x06f7U,0x0deeU,
		0x1bdcU,0x37b8U,0x6f70U,0xdee0U,0xadcbU,0x4b9dU,0x973aU,0x3e7fU,0x7cfeU,0xf9fcU,
		0xe3f3U,0xd7edU,0xbfd1U,0x6fa9U,0xdf52U,0xaeafU,0x4d55U,0x9aaaU,0x255fU,0x4abeU,
		0x957cU,0x3af3U,0x75e6U,0xebccU,0xc793U,0x9f2dU,0x2e51U,0x5ca2U,0xb944U,0x6283U,
		0xc506U,0x9a07U,0x2405U,0x480aU,0x9014U,0x3023U,0x6046U,0xc08cU,0x9113U,0x322dU,
		0x645aU,0xc8b4U,0x8163U,0x12cdU,0x259aU,0x4b34U,0x9668U,0x3cdbU,0x79b6U,0xf36cU,
		0xf6d3U,0xfdadU,0xeb51U,0xc6a9U,0x9d59U,0x2ab9U,0x5572U,0xaae4U,0x45c3U,0x8b86U,
		0x0707U,0x0e0eU,0x1c1cU,0x3838U,0x7070U,0xe0e0U,0xd1cbU,0xb39dU,0x7731U,0xee62U,
		0xcccfU,0x8995U,0x0321U,0x0642U,0x0c84U,0x1908U,0x3210U,0x6420U,0xc840U,0x808bU,
		0x111dU,0x223aU,0x4474U,0x88e8U,0x01dbU,0x03b6U,0x076cU,0x0ed8U,0x1db0U,0x3b60U,
		0x76c0U,0xed80U,0xcb0bU,0x861dU,0x1c31U,0x3862U,0x70c4U,0xe188U,0xd31bU,0xb63dU,
		0x7c71U,0xf8e2U,0xe1cfU,0xd395U,0xb721U,0x7e49U,0xfc92U,0xe92fU,0xc255U,0x94a1U,
		0x3949U,0x7292U,0xe524U,0xda43U,0xa48dU,0x5911U,0xb222U,0x744fU,0xe89eU,0xc137U,
		0x9265U,0x34c1U,0x6982U,0xd304U,0xb603U,0x7c0dU,0xf81aU,0xe03fU,0xd075U,0xb0e1U,
		0x71c9U,0xe392U,0xd72fU,0xbe55U,0x6ca1U,0xd942U,0xa28fU,0x5515U,0xaa2aU,0x445fU,
		0x88beU,0x0177U,0x02eeU,0x05dcU,0x0bb8U,0x1770U,0x2ee0U,0x5dc0U,0xbb80U,0x670bU,
		0xce16U,0x8c27U,0x0845U,0x108aU,0x2114U,0x4228U,0x8450U,0x18abU,0x3156U,0x62acU,
		0xc558U,0x9abbU,0x257dU,0x4afaU,0x95f4U,0x3be3U,0x77c6U,0xef8cU,0xcf13U,0x8e2dU,
		0x0c51U,0x18a2U,0x3144U,0x6288U,0xc510U,0x9a2bU,0x245dU,0x48baU,0x9174U,0x32e3U,
		0x65c6U,0xcb8cU,0x8713U,0x1e2dU,0x3c5aU,0x78b4U,0xf168U,0xf2dbU,0xf5bdU,0xfb71U,
		0xe6e9U,0xddd9U,0xabb9U,0x4779U,0x8ef2U,0x0defU,0x1bdeU,0x37bcU,0x6f78U,0xdef0U,
		0xadebU,0x4bddU,0x97baU,0x3f7fU,0x7efeU,0xfdfcU,0xebf3U,0xc7edU,0x9fd1U,0x2fa9U,
		0x5f52U,0xbea4U,0x6d43U,0xda86U,0xa507U,0x5a05U,0xb40aU,0x781fU,0xf03eU,0xf077U,
		0xf0e5U,0xf1c1U,0xf389U,0xf719U,0xfe39U,0xec79U,0xc8f9U,0x81f9U,0x13f9U,0x27f2U,
		0x4fe4U,0x9fc8U,0x2f9bU,0x5f36U,0xbe6cU,0x6cd3U,0xd9a6U,0xa347U,0x5685U,0xad0aU,
		0x4a1fU,0x943eU,0x3877U,0x70eeU,0xe1dcU,0xd3b3U,0xb76dU,0x7ed1U,0xfda2U,0xeb4fU,
		0xc695U,0x9d21U,0x2a49U,0x5492U,0xa924U,0x4243U,0x8486U,0x1907U,0x320eU,0x641cU,
		0xc838U,0x807bU,0x10fdU,0x21faU,0x43f4U,0x87e8U,0x1fdbU,0x3fb6U,0x7f6cU,0xfed8U,
		0xedbbU,0xcb7dU,0x86f1U,0x1de9U,0x3bd2U,0x77a4U,0xef48U,0xce9bU,0x8d3dU,0x0a71U,
		0x14e2U,0x29c4U,0x5388U,0xa710U,0x5e2bU,0xbc56U,0x68a7U,0xd14eU,0xb297U,0x7525U,
		0xea4aU,0xc49fU,0x9935U,0x2261U,0x44c2U,0x8984U,0x0303U,0x0606U,0x0c0cU,0x1818U,
		0x3030U,0x6060U,0xc0c0U,0x918bU,0x331dU,0x663aU,0xcc74U,0x88e3U,0x01cdU,0x039aU,
		0x0734U,0x0e68U,0x1cd0U,0x39a0U,0x7340U,0xe680U,0xdd0bU,0xaa1dU,0x4431U,0x8862U,
		0x00cfU,0x019eU,0x033cU,0x0678U,0x0cf0U,0x19e0U,0x33c0U,0x6780U,0xcf00U,0x8e0bU,
		0x0c1dU,0x183aU,0x3074U,0x60e8U,0xc1d0U,0x93abU,0x375dU,0x6ebaU,0xdd74U,0xaae3U,
		0x45cdU,0x8b9aU,0x073fU,0x0e7eU,0x1cfcU,0x39f8U,0x73f0U,0xe7e0U,0xdfcbU,0xaf9dU,
		0x4f31U,0x9e62U,0x2ccfU,0x599eU,0xb33cU,0x7673U,0xece6U,0xc9c7U,0x8385U,0x1701U,
		0x2e02U,0x5c04U,0xb808U,0x601bU,0xc036U,0x9067U,0x30c5U,0x618aU,0xc314U,0x9623U,
		0x3c4dU,0x789aU,0xf134U,0xf263U,0xf4cdU,0xf991U,0xe329U,0xd659U,0xbcb9U,0x6979U,
		0xd2f2U,0xb5efU,0x7bd5U,0xf7aaU,0xff5fU,0xeeb5U,0xcd61U,0x8ac9U,0x0599U,0x0b32U,
		0x1664U,0x2cc8U,0x5990U,0xb320U,0x764bU,0xec96U,0xc927U,0x8245U,0x1481U,0x2902U,
		0x5204U,0xa408U,0x581bU,0xb036U,0x7067U,0xe0ceU,0xd197U,0xb325U,0x7641U,0xec82U,
		0xc90fU,0x8215U,0x1421U,0x2842U,0x5084U,0xa108U,0x521bU,0xa436U,0x5867U,0xb0ceU,
		0x7197U,0xe32eU,0xd657U,0xbca5U,0x6941U,0xd282U,0xb50fU,0x7a15U,0xf42aU,0xf85fU,
		0xe0b5U,0xd161U,0xb2c9U,0x7599U,0xeb32U,0xc66fU,0x9cd5U,0x29a1U,0x5342U,0xa684U,
		0x5d03U,0xba06U,0x6407U,0xc80eU,0x8017U,0x1025U,0x204aU,0x4094U,0x8128U,0x125bU,
		0x24b6U,0x496cU,0x92d8U,0x35bbU,0x6b76U,0xd6ecU,0xbdd3U,0x6badU,0xd75aU,0xbebfU,
		0x6d75U,0xdaeaU,0xa5dfU,0x5bb5U,0xb76aU,0x7edfU,0xfdbeU,0xeb77U,0xc6e5U,0x9dc1U,
		0x2b89U,0x5712U,0xae24U,0x4c43U,0x9886U,0x2107U,0x420eU,0x841cU,0x1833U,0x3066U,
		0x60ccU,0xc198U,0x933bU,0x367dU,0x6cfaU,0xd9f4U,0xa3e3U,0x57cdU,0xaf9aU,0x4f3fU,
		0x9e7eU,0x2cf7U,0x59eeU,0xb3dcU,0x77b3U,0xef66U,0xcec7U,0x8d85U,0x0b01U,0x1602U,
		0x2c04U,0x5808U,0xb010U,0x702bU,0xe056U,0xd0a7U,0xb145U,0x7281U,0xe502U,0xda0fU,
		0xa415U,0x5821U,0xb042U,0x708fU,0xe11eU,0xd237U,0xb465U,0x78c1U,0xf182U,0xf30fU,
		0xf615U,0xfc21U,0xe849U,0xc099U,0x9139U,0x3279U,0x64f2U,0xc9e4U,0x83c3U,0x178dU,
		0x2f1aU,0x5e34U,0xbc68U,0x68dbU,0xd1b6U,0xb367U,0x76c5U,0xed8aU,0xcb1fU,0x8635U,
		0x1c61U,0x38c2U,0x7184U,0xe308U,0xd61bU,0xbc3dU,0x6871U,0xd0e2U,0xb1cfU,0x7395U,
		0xe72aU,0xde5fU,0xacb5U,0x4961U,0x92c2U,0x358fU,0x6b1eU,0xd63cU,0xbc73U,0x68edU,
		0xd1daU,0xb3bfU,0x7775U,0xeeeaU,0xcddfU,0x8bb5U,0x0761U,0x0ec2U,0x1d84U,0x3b08U,
		0x7610U,0xec20U,0xc84bU,0x809dU,0x1131U,0x2262U,0x44c4U,0x8988U,0x031bU,0x0636U,
		0x0c6cU,0x18d8U,0x31b0U,0x6360U,0xc6c0U,0x9d8bU,0x2b1dU,0x563aU,0xac74U,0x48e3U,
		0x91c6U,0x3387U,0x670eU,0xce1cU,0x8c33U,0x086dU,0x10daU,0x21b4U,0x4368U,0x86d0U,
		0x1dabU,0x3b56U,0x76acU,0xed58U,0xcabbU,0x857dU,0x1af1U,0x35e2U,0x6bc4U,0xd788U,
		0xbf1bU,0x6e3dU,0xdc7aU,0xa8ffU,0x41f5U,0x83eaU,0x17dfU,0x2fbeU,0x5f7cU,0xbef8U,
		0x6dfbU,0xdbf6U,0xa7e7U,0x5fc5U,0xbf8aU,0x6f1fU,0xde3eU,0xac77U,0x48e5U,0x91caU,
		0x339fU,0x673eU,0xce7cU,0x8cf3U,0x09edU,0x13daU,0x27b4U,0x4f68U,0x9ed0U,0x2dabU,
		0x5b56U,0xb6acU,0x7d53U,0xfaa6U,0xe547U,0xda85U,0xa501U,0x5a09U,0xb412U,0x782fU,
		0xf05eU,0xf0b7U,0xf165U,0xf2c1U,0xf589U,0xfb19U,0xe639U,0xdc79U,0xa8f9U,0x41f9U,
		0x83f2U,0x17efU,0x2fdeU,0x5fbcU,0xbf78U,0x6efbU,0xddf6U,0xabe7U,0x47c5U,0x8f8aU,
		0x0f1fU,0x1e3eU,0x3c7cU,0x78f8U,0xf1f0U,0xf3ebU,0xf7ddU,0xffb1U,0xef69U,0xced9U,
		0x8db9U,0x0b79U,0x16f2U,0x2de4U,0x5bc8U,0xb790U,0x7f2bU,0xfe56U,0xeca7U,0xc945U,
		0x8281U,0x1509U,0x2a12U,0x5424U,0xa848U,0x409bU,0x8136U,0x1267U,0x24ceU,0x499cU,
		0x9338U,0x367bU,0x6cf6U,0xd9ecU,0xa3d3U,0x57adU,0xaf5aU,0x4ebfU,0x9d7eU,0x2af7U,
		0x55eeU,0xabdcU,0x47b3U,0x8f66U,0x0ec7U,0x1d8eU,0x3b1cU,0x7638U,0xec70U,0xc8ebU,
		0x81ddU,0x13b1U,0x2762U,0x4ec4U,0x9d88U,0x2b1bU,0x5636U,0xac6cU,0x48d3U,0x91a6U,
		0x3347U,0x668eU,0xcd1cU,0x8a33U,0x046dU,0x08daU,0x11b4U,0x2368U,0x46d0U,0x8da0U,
		0x0b4bU,0x1696U,0x2d2cU,0x5a58U,0xb4b0U,0x796bU,0xf2d6U,0xf5a7U,0xfb45U,0xe681U,
		0xdd09U,0xaa19U,0x4439U,0x8872U,0x00efU,0x01deU,0x03bcU,0x0778U,0x0ef0U,0x1de0U,
		0x3bc0U,0x7780U,0xef00U,0xce0bU,0x8c1dU,0x0831U,0x1062U,0x20c4U,0x4188U,0x8310U,
		0x162bU,0x2c56U,0x58acU,0xb158U,0x72bbU,0xe576U,0xdae7U,0xa5c5U,0x5b81U,0xb702U,
		0x7e0fU,0xfc1eU,0xe837U,0xc065U,0x90c1U,0x3189U,0x6312U,0xc624U,0x9c43U,0x288dU,
		0x511aU,0xa234U,0x5463U,0xa8c6U,0x4187U,0x830eU,0x1617U,0x2c2eU,0x585cU,0xb0b8U,
		0x717bU,0xe2f6U,0xd5e7U,0xbbc5U,0x6781U,0xcf02U,0x8e0fU,0x0c15U,0x182aU,0x3054U,
		0x60a8U,0xc150U,0x92abU,0x355dU,0x6abaU,0xd574U,0xbae3U,0x65cdU,0xcb9aU,0x873fU,
		0x1e75U,0x3ceaU,0x79d4U,0xf3a8U,0xf75bU,0xfebdU,0xed71U,0xcae9U,0x85d9U,0x1bb9U,
		0x3772U,0x6ee4U,0xddc8U,0xab9bU,0x473dU,0x8e7aU,0x0cffU,0x19feU,0x33fcU,0x67f8U,
		0xcff0U,0x8febU,0x0fddU,0x1fbaU,0x3f74U,0x7ee8U,0xfdd0U,0xebabU,0xc75dU,0x9eb1U,
		0x2d69U,0x5ad2U,0xb5a4U,0x7b43U,0xf686U,0xfd07U,0xea05U,0xc401U,0x9809U,0x2019U,
		0x4032U,0x8064U,0x10c3U,0x2186U,0x430cU,0x8618U,0x1c3bU,0x3876U,0x70ecU,0xe1d8U,
		0xd3bbU,0xb77dU,0x7ef1U,0xfde2U,0xebcfU,0xc795U,0x9f21U,0x2e49U,0x5c92U,0xb924U,
		0x6243U,0xc486U,0x9907U,0x2205U,0x440aU,0x8814U,0x0023U,0x0046U,0x008cU,0x0118U,
		0x0230U,0x0460U,0x08c0U,0x1180U,0x2300U,0x4600U,0x8c00U,0x080bU,0x1016U,0x202cU,
		0x4058U,0x80b0U,0x116bU,0x22d6U,0x45acU,0x8b58U,0x06bbU,0x0d76U,0x1aecU,0x35d8U,
		0x6bb0U,0xd760U,0xbecbU,0x6d9dU,0xdb3aU,0xa67fU,0x5cf5U,0xb9eaU,0x63dfU,0xc7beU,
		0x9f77U,0x2ee5U,0x5dcaU,0xbb94U,0x6723U,0xce46U,0x8c87U,0x0905U,0x120aU,0x2414U,
		0x4828U,0x9050U,0x30abU,0x6156U,0xc2acU,0x9553U,0x3aadU,0x755aU,0xeab4U,0xc563U,
		0x9acdU,0x2591U,0x4b22U,0x9644U,0x3c83U,0x7906U,0xf20cU,0xf413U,0xf82dU,0xe051U,
		0xd0a9U,0xb159U,0x72b9U,0xe572U,0xdaefU,0xa5d5U,0x5ba1U,0xb742U,0x7e8fU,0xfd1eU,
		0xea37U,0xc465U,0x98c1U,0x2189U,0x4312U,0x8624U,0x1c43U,0x3886U,0x710cU,0xe218U,
		0xd43bU,0xb87dU,0x60f1U,0xc1e2U,0x93cfU,0x3795U,0x6f2aU,0xde54U,0xaca3U,0x494dU,
		0x929aU,0x353fU,0x6a7eU,0xd4fcU,0xb9f3U,0x63edU,0xc7daU,0x9fbfU,0x2f75U,0x5eeaU,
		0xbdd4U,0x6ba3U,0xd746U,0xbe87U,0x6d05U,0xda0aU,0xa41fU,0x5835U,0xb06aU,0x70dfU,
		0xe1beU,0xd377U,0xb6e5U,0x7dc1U,0xfb82U,0xe70fU,0xde15U,0xac21U,0x4849U,0x9092U,
		0x312fU,0x625eU,0xc4bcU,0x9973U,0x22edU,0x45daU,0x8bb4U,0x0763U,0x0ec6U,0x1d8cU,
		0x3b18U,0x7630U,0xec60U,0xc8cbU,0x819dU,0x1331U,0x2662U,0x4cc4U,0x9988U,0x231bU,
		0x4636U,0x8c6cU,0x08d3U,0x11a6U,0x234cU,0x4698U,0x8d30U,0x0a6bU,0x14d6U,0x29acU,
		0x5358U,0xa6b0U,0x5d6bU,0xbad6U,0x65a7U,0xcb4eU,0x8697U,0x1d25U,0x3a4aU,0x7494U,
		0xe928U,0xc25bU,0x94bdU,0x3971U,0x72e2U,0xe5c4U,0xdb83U,0xa70dU,0x5e11U,0xbc22U,
		0x684fU,0xd09eU,0xb137U,0x7265U,0xe4caU,0xd99fU,0xa335U,0x5661U,0xacc2U,0x498fU,
		0x931eU,0x3637U,0x6c6eU,0xd8dcU,0xa1b3U,0x536dU,0xa6daU,0x5dbfU,0xbb7eU,0x66f7U,
		0xcdeeU,0x8bd7U,0x07a5U,0x0f4aU,0x1e94U,0x3d28U,0x7a50U,0xf4a0U,0xf94bU,0xe29dU,
		0xd531U,0xba69U,0x64d9U,0xc9b2U,0x836fU,0x16d5U,0x2daaU,0x5b54U,0xb6a8U,0x7d5bU,
		0xfab6U,0xe567U,0xdac5U,0xa581U,0x5b09U,0xb612U,0x7c2fU,0xf85eU,0xe0b7U,0xd165U,
		0xb2c1U,0x7589U,0xeb12U,0xc62fU,0x9c55U,0x28a1U,0x5142U,0xa284U,0x5503U,0xaa06U,
		0x4407U,0x880eU,0x0017U,0x002eU,0x005cU,0x00b8U,0x0170U,0x02e0U,0x05c0U,0x0b80U,
		0x1700U,0x2e00U,0x5c00U,0xb800U,0x600bU,0xc016U,0x9027U,0x3045U,0x608aU,0xc114U,
		0x9223U,0x344dU,0x689aU,0xd134U,0xb263U,0x74cdU,0xe99aU,0xc33fU,0x9675U,0x3ce1U,
		0x79c2U,0xf384U,0xf703U,0xfe0dU,0xec11U,0xc829U,0x8059U,0x10b9U,0x2172U,0x42e4U,
		0x85c8U,0x1b9bU,0x3736U,0x6e6cU,0xdcd8U,0xa9bbU,0x437dU,0x86faU,0x1dffU,0x3bfeU,
		0x77fcU,0xeff8U,0xcffbU,0x8ffdU,0x0ff1U,0x1fe2U,0x3fc4U,0x7f88U,0xff10U,0xee2bU,
		0xcc5dU,0x88b1U,0x0169U,0x02d2U,0x05a4U,0x0b48U,0x1690U,0x2d20U,0x5a40U,0xb480U,
		0x790bU,0xf216U,0xf427U,0xf845U,0xe081U,0xd109U,0xb219U,0x7439U,0xe872U,0xc0efU,
		0x91d5U,0x33a1U,0x6742U,0xce84U,0x8d03U,0x0a0dU,0x141aU,0x2834U,0x5068U,0xa0d0U,
		0x51abU,0xa356U,0x56a7U,0xad4eU,0x4a97U,0x952eU,0x3a57U,0x74aeU,0xe95cU,0xc2b3U,
		0x956dU,0x3ad1U,0x75a2U,0xeb44U,0xc683U,0x9d0dU,0x2a11U,0x5422U,0xa844U,0x4083U,
		0x8106U,0x1207U,0x240eU,0x481cU,0x9038U,0x307bU,0x60f6U,0xc1ecU,0x93d3U,0x37adU,
		0x6f5aU,0xdeb4U,0xad63U,0x4acdU,0x959aU,0x3b3fU,0x767eU,0xecfcU,0xc9f3U,0x83edU,
		0x17d1U,0x2fa2U,0x5f44U,0xbe88U,0x6d1bU,0xda36U,0xa467U,0x58c5U,0xb18aU,0x731fU,
		0xe63eU,0xdc77U,0xa8e5U,0x41c1U,0x8382U,0x170fU,0x2e1eU,0x5c3cU,0xb878U,0x60fbU,
		0xc1f6U,0x93e7U,0x37c5U,0x6f8aU,0xdf14U,0xae23U,0x4c4dU,0x989aU,0x213fU,0x427eU,
		0x84fcU,0x19f3U,0x33e6U,0x67ccU,0xcf98U,0x8f3bU,0x0e7dU,0x1cfaU,0x39f4U,0x73e8U,
		0xe7d0U,0xdfabU,0xaf5dU,0x4eb1U,0x9d62U,0x2acfU,0x559eU,0xab3cU,0x4673U,0x8ce6U,
		0x09c7U,0x138eU,0x271cU,0x4e38U,0x9c70U,0x28ebU,0x51d6U,0xa3acU,0x5753U,0xaea6U,
		0x4d47U,0x9a8eU,0x2517U,0x4a2eU,0x945cU,0x38b3U,0x7166U,0xe2ccU,0xd593U,0xbb2dU,
		0x6651U,0xcca2U,0x894fU,0x0295U,0x052aU,0x0a54U,0x14a8U,0x2950U,0x52a0U,0xa540U,
		0x5a8bU,0xb516U,0x7a27U,0xf44eU,0xf897U,0xe125U,0xd241U,0xb489U,0x7919U,0xf232U,
		0xf46fU,0xf8d5U,0xe1a1U,0xd349U,0xb699U,0x7d39U,0xfa72U,0xe4efU,0xd9d5U,0xa3a1U,
		0x5749U,0xae92U,0x4d2fU,0x9a5eU,0x24b7U,0x496eU,0x92dcU,0x35b3U,0x6b66U,0xd6ccU,
		0xbd93U,0x6b2dU,0xd65aU,0xbcbfU,0x6975U,0xd2eaU,0xb5dfU,0x7bb5U,0xf76aU,0xfedfU,
		0xedb5U,0xcb61U,0x86c9U,0x1d99U,0x3b32U,0x7664U,0xecc8U,0xc99bU,0x833dU,0x1671U,
		0x2ce2U,0x59c4U,0xb388U,0x771bU,0xee36U,0xcc67U,0x88c5U,0x0181U,0x0302U,0x0604U,
		0x0c08U,0x1810U,0x3020U,0x6040U,0xc080U,0x910bU,0x321dU,0x643aU,0xc874U,0x80e3U,
		0x11cdU,0x239aU,0x4734U,0x8e68U,0x0cdbU,0x19b6U,0x336cU,0x66d8U,0xcdb0U,0x8b6bU,
		0x06ddU,0x0dbaU,0x1b74U,0x36e8U,0x6dd0U,0xdba0U,0xa74bU,0x5e9dU,0xbd3aU,0x6a7fU,
		0xd4feU,0xb9f7U,0x63e5U,0xc7caU,0x9f9fU,0x2f35U,0x5e6aU,0xbcd4U,0x69a3U,0xd346U,
		0xb687U,0x7d05U,0xfa0aU,0xe41fU,0xd835U,0xa061U,0x50c9U,0xa192U,0x532fU,0xa65eU,
		0x5cb7U,0xb96eU,0x62d7U,0xc5aeU,0x9b57U,0x26a5U,0x4d4aU,0x9a94U,0x2523U,0x4a46U,
		0x948cU,0x3913U,0x7226U,0xe44cU,0xd893U,0xa12dU,0x5251U,0xa4a2U,0x594fU,0xb29eU,
		0x7537U,0xea6eU,0xc4d7U,0x99a5U,0x2341U,0x4682U,0x8d04U,0x0a03U,0x1406U,0x280cU,
		0x5018U,0xa030U,0x506bU,0xa0d6U,0x51a7U,0xa34eU,0x5697U,0xad2eU,0x4a57U,0x94aeU,
		0x3957U,0x72aeU,0xe55cU,0xdab3U,0xa56dU,0x5ad1U,0xb5a2U,0x7b4fU,0xf69eU,0xfd37U,
		0xea65U,0xc4c1U,0x9989U,0x2319U,0x4632U,0x8c64U,0x08c3U,0x1186U,0x230cU,0x4618U,
		0x8c30U,0x086bU,0x10d6U,0x21acU,0x4358U,0x86b0U,0x1d6bU,0x3ad6U,0x75acU,0xeb58U,
		0xc6bbU,0x9d7dU,0x2af1U,0x55e2U,0xabc4U,0x4783U,0x8f06U,0x0e07U,0x1c0eU,0x381cU,
		0x7038U,0xe070U,0xd0ebU,0xb1ddU,0x73b1U,0xe762U,0xdecfU,0xad95U,0x4b21U,0x9642U,
		0x3c8fU,0x791eU,0xf23cU,0xf473U,0xf8edU,0xe1d1U,0xd3a9U,0xb759U,0x7eb9U,0xfd72U,
		0xeaefU,0xc5d5U,0x9ba1U,0x2749U,0x4e92U,0x9d24U,0x2a43U,0x5486U,0xa90cU,0x4213U,
		0x8426U,0x1847U,0x308eU,0x611cU,0xc238U,0x947bU,0x38fdU,0x71faU,0xe3f4U,0xd7e3U,
		0xbfcdU,0x6f91U,0xdf22U,0xae4fU,0x4c95U,0x992aU,0x225fU,0x44beU,0x897cU,0x02f3U,
		0x05e6U,0x0bccU,0x1798U,0x2f30U,0x5e60U,0xbcc0U,0x698bU,0xd316U,0xb627U,0x7c45U,
		0xf88aU,0xe11fU,0xd235U,0xb461U,0x78c9U,0xf192U,0xf32fU,0xf655U,0xfca1U,0xe949U,
		0xc299U,0x9539U,0x3a79U,0x74f2U,0xe9e4U,0xc3c3U,0x978dU,0x3f11U,0x7e22U,0xfc44U,
		0xe883U,0xc10dU,0x9211U,0x3429U,0x6852U,0xd0a4U,0xb143U,0x728dU,0xe51aU,0xda3fU,
		0xa475U,0x58e1U,0xb1c2U,0x738fU,0xe71eU,0xde37U,0xac65U,0x48c1U,0x9182U,0x330fU,
		0x661eU,0xcc3cU,0x8873U,0x00edU,0x01daU,0x03b4U,0x0768U,0x0ed0U,0x1da0U,0x3b40U,
		0x7680U,0xed00U,0xca0bU,0x841dU,0x1831U,0x3062U,0x60c4U,0xc188U,0x931bU,0x363dU,
		0x6c7aU,0xd8f4U,0xa1e3U,0x53cdU,0xa79aU,0x5f3fU,0xbe7eU,0x6cf7U,0xd9eeU,0xa3d7U,
		0x57a5U,0xaf4aU,0x4e9fU,0x9d3eU,0x2a77U,0x54eeU,0xa9dcU,0x43b3U,0x8766U,0x1ec7U,
		0x3d8eU,0x7b1cU,0xf638U,0xfc7bU,0xe8fdU,0xc1f1U,0x93e9U,0x37d9U,0x6fb2U,0xdf64U,
		0xaec3U,0x4d8dU,0x9b1aU,0x263fU,0x4c7eU,0x98fcU,0x21f3U,0x43e6U,0x87ccU,0x1f93U,
		0x3f26U,0x7e4cU,0xfc98U,0xe93bU,0xc27dU,0x94f1U,0x39e9U,0x73d2U,0xe7a4U,0xdf43U,
		0xae8dU,0x4d11U,0x9a22U,0x244fU,0x489eU,0x913cU,0x3273U,0x64e6U,0xc9ccU,0x8393U,
		0x172dU,0x2e5aU,0x5cb4U,0xb968U,0x62dbU,0xc5b6U,0x9b67U,0x26c5U,0x4d8aU,0x9b14U,
		0x2623U,0x4c46U,0x988cU,0x2113U,0x4226U,0x844cU,0x1893U,0x3126U,0x624cU,0xc498U,
		0x993bU,0x227dU,0x44faU,0x89f4U,0x03e3U,0x07c6U,0x0f8cU,0x1f18U,0x3e30U,0x7c60U,
		0xf8c0U,0xe18bU,0xd31dU,0xb631U,0x7c69U,0xf8d2U,0xe1afU,0xd355U,0xb6a1U,0x7d49U,
		0xfa92U,0xe52fU,0xda55U,0xa4a1U,0x5949U,0xb292U,0x752fU,0xea5eU,0xc4b7U,0x9965U,
		0x22c1U,0x4582U,0x8b04U,0x0603U,0x0c06U,0x180cU,0x3018U,0x6030U,0xc060U,0x90cbU,
		0x319dU,0x633aU,0xc674U,0x9ce3U,0x29cdU,0x539aU,0xa734U,0x5e63U,0xbcc6U,0x6987U,
		0xd30eU,0xb617U,0x7c25U,0xf84aU,0xe09fU,0xd135U,0xb261U,0x74c9U,0xe992U,0xc32fU,
		0x9655U,0x3ca1U,0x7942U,0xf284U,0xf503U,0xfa0dU,0xe411U,0xd829U,0xa059U,0x50b9U,
		0xa172U,0x52efU,0xa5deU,0x5bb7U,0xb76eU,0x7ed7U,0xfdaeU,0xeb57U,0xc6a5U,0x9d41U,
		0x2a89U,0x5512U,0xaa24U,0x4443U,0x8886U,0x0107U,0x020eU,0x041cU,0x0838U,0x1070U,
		0x20e0U,0x41c0U,0x8380U,0x170bU,0x2e16U,0x5c2cU,0xb858U,0x60bbU,0xc176U,0x92e7U,
		0x35c5U,0x6b8aU,0xd714U,0xbe23U,0x6c4dU,0xd89aU,0xa13fU,0x5275U,0xa4eaU,0x59dfU,
		0xb3beU,0x7777U,0xeeeeU,0xcdd7U,0x8ba5U,0x0741U,0x0e82U,0x1d04U,0x3a08U,0x7410U,
		0xe820U,0xc04bU,0x909dU,0x3131U,0x6262U,0xc4c4U,0x9983U,0x230dU,0x461aU,0x8c34U,
		0x0863U,0x10c6U,0x218cU,0x4318U,0x8630U,0x1c6bU,0x38d6U,0x71acU,0xe358U,0xd6bbU,
		0xbd7dU,0x6af1U,0xd5e2U,0xbbcfU,0x6795U,0xcf2aU,0x8e5fU,0x0cb5U,0x196aU,0x32d4U,
		0x65a8U,0xcb50U,0x86abU,0x1d5dU,0x3abaU,0x7574U,0xeae8U,0xc5dbU,0x9bbdU,0x2771U,
		0x4ee2U,0x9dc4U,0x2b83U,0x5706U,0xae0cU,0x4c13U,0x9826U,0x2047U,0x408eU,0x811cU,
		0x1233U,0x2466U,0x48ccU,0x9198U,0x333bU,0x6676U,0xccecU,0x89d3U,0x03adU,0x075aU,
		0x0eb4U,0x1d68U,0x3ad0U,0x75a0U,0xeb40U,0xc68bU,0x9d1dU,0x2a31U,0x5462U,0xa8c4U,
		0x4183U,0x8306U,0x1607U,0x2c0eU,0x581cU,0xb038U,0x707bU,0xe0f6U,0xd1e7U,0xb3c5U,
		0x7781U,0xef02U,0xce0fU,0x8c15U,0x0821U,0x1042U,0x2084U,0x4108U,0x8210U,0x142bU,
		0x2856U,0x50acU,0xa158U,0x52bbU,0xa576U,0x5ae7U,0xb5ceU,0x7b97U,0xf72eU,0xfe57U,
		0xeca5U,0xc941U,0x8289U,0x1519U,0x2a32U,0x5464U,0xa8c8U,0x419bU,0x8336U,0x1667U,
		0x2cceU,0x599cU,0xb338U,0x767bU,0xecf6U,0xc9e7U,0x83c5U,0x1781U,0x2f02U,0x5e04U,
		0xbc08U,0x681bU,0xd036U,0xb067U,0x70c5U,0xe18aU,0xd31fU,0xb635U,0x7c61U,0xf8c2U,
		0xe18fU,0xd315U,0xb621U,0x7c49U,0xf892U,0xe12fU,0xd255U,0xb4a1U,0x7949U,0xf292U,
		0xf52fU,0xfa55U,0xe4a1U,0xd949U,0xa299U,0x5539U,0xaa72U,0x44efU,0x89deU,0x03b7U,
		0x076eU,0x0edcU,0x1db8U,0x3b70U,0x76e0U,0xedc0U,0xcb8bU,0x871dU,0x1e31U,0x3c62U,
		0x78c4U,0xf188U,0xf31bU,0xf63dU,0xfc71U,0xe8e9U,0xc1d9U,0x93b9U,0x3779U,0x6ef2U,
		0xdde4U,0xabc3U,0x478dU,0x8f1aU,0x0e3fU,0x1c7eU,0x38fcU,0x71f8U,0xe3f0U,0xd7ebU,
		0xbfddU,0x6fb1U,0xdf62U,0xaecfU,0x4d95U,0x9b2aU,0x265fU,0x4cbeU,0x997cU,0x22f3U,
		0x45e6U,0x8bccU,0x0793U,0x0f26U,0x1e4cU,0x3c98U,0x7930U,0xf260U,0xf4cbU,0xf99dU,
		0xe331U,0xd669U,0xbcd9U,0x69b9U,0xd372U,0xb6efU,0x7dd5U,0xfbaaU,0xe75fU,0xdeb5U,
		0xad61U,0x4ac9U,0x9592U,0x3b2fU,0x765eU,0xecbcU,0xc973U,0x82edU,0x15d1U,0x2ba2U,
		0x5744U,0xae88U,0x4d1bU,0x9a36U,0x2467U,0x48ceU,0x919cU,0x3333U,0x6666U,0xccccU,
		0x8993U,0x032dU,0x065aU,0x0cb4U,0x1968U,0x32d0U,0x65a0U,0xcb40U,0x868bU,0x1d1dU,
		0x3a3aU,0x7474U,0xe8e8U,0xc1dbU,0x93bdU,0x3771U,0x6ee2U,0xddc4U,0xab83U,0x470dU,
		0x8e1aU,0x0c3fU,0x187eU,0x30fcU,0x61f8U,0xc3f0U,0x97ebU,0x3fddU,0x7fbaU,0xff74U,
		0xeee3U,0xcdcdU,0x8b91U,0x0729U,0x0e52U,0x1ca4U,0x3948U,0x7290U,0xe520U,0xda4bU,
		0xa49dU,0x5931U,0xb262U,0x74cfU,0xe99eU,0xc337U,0x9665U,0x3cc1U,0x7982U,0xf304U,
		0xf603U,0xfc0dU,0xe811U,0xc029U,0x9059U,0x30b9U,0x6172U,0xc2e4U,0x95c3U,0x3b8dU,
		0x771aU,0xee34U,0xcc63U,0x88cdU,0x0191U,0x0322U,0x0644U,0x0c88U,0x1910U,0x3220U,
		0x6440U,0xc880U,0x810bU,0x121dU,0x243aU,0x4874U,0x90e8U,0x31dbU,0x63b6U,0xc76cU,
		0x9ed3U,0x2dadU,0x5b5aU,0xb6b4U,0x7d63U,0xfac6U,0xe587U,0xdb05U,0xa601U,0x5c09U,
		0xb812U,0x602fU,0xc05eU,0x90b7U,0x3165U,0x62caU,0xc594U,0x9b23U,0x264dU,0x4c9aU,
		0x9934U,0x2263U,0x44c6U,0x898cU,0x0313U,0x0626U,0x0c4cU,0x1898U,0x3130U,0x6260U,
		0xc4c0U,0x998bU,0x231dU,0x463aU,0x8c74U,0x08e3U,0x11c6U,0x238cU,0x4718U,0x8e30U,
		0x0c6bU,0x18d6U,0x31acU,0x6358U,0xc6b0U,0x9d6bU,0x2addU,0x55baU,0xab74U,0x46e3U,
		0x8dc6U,0x0b87U,0x170eU,0x2e1cU,0x5c38U,0xb870U,0x60ebU,0xc1d6U,0x93a7U,0x3745U,
		0x6e8aU,0xdd14U,0xaa23U,0x444dU,0x889aU,0x013fU,0x027eU,0x04fcU,0x09f8U,0x13f0U,
		0x27e0U,0x4fc0U,0x9f80U,0x2f0bU,0x5e16U,0xbc2cU,0x6853U,0xd0a6U,0xb147U,0x7285U,
		0xe50aU,0xda1fU,0xa435U,0x5861U,0xb0c2U,0x718fU,0xe31eU,0xd637U,0xbc65U,0x68c1U,
		0xd182U,0xb30fU,0x7615U,0xec2aU,0xc85fU,0x80b5U,0x1161U,0x22c2U,0x4584U,0x8b08U,
		0x061bU,0x0c36U,0x186cU,0x30d8U,0x61b0U,0xc360U,0x96cbU,0x3d9dU,0x7b3aU,0xf674U,
		0xfce3U,0xe9cdU,0xc391U,0x9729U,0x3e59U,0x7cb2U,0xf964U,0xe2c3U,0xd58dU,0xbb11U,
		0x6629U,0xcc52U,0x88afU,0x0155U,0x02aaU,0x0554U,0x0aa8U,0x1550U,0x2aa0U,0x5540U,
		0xaa80U,0x450bU,0x8a16U,0x0427U,0x084eU,0x109cU,0x2138U,0x4270U,0x84e0U,0x19cbU,
		0x3396U,0x672cU,0xce58U,0x8cbbU,0x097dU,0x12faU,0x25f4U,0x4be8U,0x97d0U,0x3fabU,
		0x7f56U,0xfeacU,0xed53U,0xcaadU,0x8551U,0x1aa9U,0x3552U,0x6aa4U,0xd548U,0xba9bU,
		0x653dU,0xca7aU,0x84ffU,0x19f5U,0x33eaU,0x67d4U,0xcfa8U,0x8f5bU,0x0ebdU,0x1d7aU,
		0x3af4U,0x75e8U,0xebd0U,0xc7abU,0x9f5dU,0x2eb1U,0x5d62U,0xbac4U,0x6583U,0xcb06U,
		0x8607U,0x1c05U,0x380aU,0x7014U,0xe028U,0xd05bU,0xb0bdU,0x7171U,0xe2e2U,0xd5cfU,
		0xbb95U,0x6721U,0xce42U,0x8c8fU,0x0915U,0x122aU,0x2454U,0x48a8U,0x9150U,0x32abU,
		0x6556U,0xcaacU,0x8553U,0x1aadU,0x355aU,0x6ab4U,0xd568U,0xbadbU,0x65bdU,0xcb7aU,
		0x86ffU,0x1df5U,0x3beaU,0x77d4U,0xefa8U,0xcf5bU,0x8ebdU,0x0d71U,0x1ae2U,0x35c4U,
		0x6b88U,0xd710U,0xbe2bU,0x6c5dU,0xd8baU,0xa17fU,0x52f5U,0xa5eaU,0x5bdfU,0xb7beU,
		0x7f77U,0xfeeeU,0xedd7U,0xcba5U,0x8741U,0x1e89U,0x3d12U,0x7a24U,0xf448U,0xf89bU,
		0xe13dU,0xd271U,0xb4e9U,0x79d9U,0xf3b2U,0xf76fU,0xfed5U,0xeda1U,0xcb49U,0x8699U,
		0x1d39U,0x3a72U,0x74e4U,0xe9c8U,0xc39bU,0x973dU,0x3e71U,0x7ce2U,0xf9c4U,0xe383U,
		0xd70dU,0xbe11U,0x6c29U,0xd852U,0xa0afU,0x5155U,0xa2aaU,0x555fU,0xaabeU,0x4577U,
		0x8aeeU,0x05d7U,0x0baeU,0x175cU,0x2eb8U,0x5d70U,0xbae0U,0x65cbU,0xcb96U,0x8727U,
		0x1e45U,0x3c8aU,0x7914U,0xf228U,0xf45bU,0xf8bdU,0xe171U,0xd2e9U,0xb5d9U,0x7bb9U,
		0xf772U,0xfeefU,0xedd5U,0xcba1U,0x8749U,0x1e99U,0x3d32U,0x7a64U,0xf4c8U,0xf99bU,
		0xe33dU,0xd671U,0xbce9U,0x69d9U,0xd3b2U,0xb76fU,0x7ed5U,0xfdaaU,0xeb5fU,0xc6b5U,
		0x9d61U,0x2ac9U,0x5592U,0xab24U,0x4643U,0x8c86U,0x0907U,0x120eU,0x241cU,0x4838U,
		0x9070U,0x30ebU,0x61d6U,0xc3acU,0x9753U,0x3eadU,0x7d5aU,0xfab4U,0xe563U,0xdacdU,
		0xa591U,0x5b29U,0xb652U,0x7cafU,0xf95eU,0xe2b7U,0xd565U,0xbac1U,0x6589U,0xcb12U,
		0x862fU,0x1c55U,0x38aaU,0x7154U,0xe2a8U,0xd55bU,0xbabdU,0x6571U,0xcae2U,0x85cfU,
		0x1b95U,0x372aU,0x6e54U,0xdca8U,0xa95bU,0x42bdU,0x857aU,0x1affU,0x35feU,0x6bfcU,
		0xd7f8U,0xbffbU,0x6ffdU,0xdffaU,0xafffU,0x4ff5U,0x9feaU,0x2fdfU,0x5fbeU,0xbf7cU,
		0x6ef3U,0xdde6U,0xabc7U,0x4785U,0x8f0aU,0x0e1fU,0x1c3eU,0x387cU,0x70f8U,0xe1f0U,
		0xd3ebU,0xb7ddU,0x7fb1U,0xff62U,0xeecfU,0xcd95U,0x8b21U,0x0649U,0x0c92U,0x1924U,
		0x3248U,0x6490U,0xc920U,0x824bU,0x149dU,0x293aU,0x5274U,0xa4e8U,0x59dbU,0xb3b6U,
		0x7767U,0xeeceU,0xcd97U,0x8b25U,0x0641U,0x0c82U,0x1904U,0x3208U,0x6410U,0xc820U,
		0x804bU,0x109dU,0x213aU,0x4274U,0x84e8U,0x19dbU,0x33b6U,0x676cU,0xced8U,0x8dbbU,
		0x0b7dU,0x16faU,0x2df4U,0x5be8U,0xb7d0U,0x7fabU,0xff56U,0xeea7U,0xcd45U,0x8a81U,
		0x0509U,0x0a12U,0x1424U,0x2848U,0x5090U,0xa120U,0x524bU,0xa496U,0x5927U,0xb24eU,
		0x7497U,0xe92eU,0xc257U,0x94a5U,0x3941U,0x7282U,0xe504U,0xda03U,0xa40dU,0x5811U,
		0xb022U,0x704fU,0xe09eU,0xd137U,0xb265U,0x74c1U,0xe982U,0xc30fU,0x9615U,0x3c21U,
		0x7842U,0xf084U,0xf103U,0xf20dU,0xf411U,0xf829U,0xe059U,0xd0b9U,0xb179U,0x72f9U,
		0xe5f2U,0xdbefU,0xa7d5U,0x5fa1U,0xbf42U,0x6e8fU,0xdd1eU,0xaa37U,0x4465U,0x88caU,
		0x019fU,0x033eU,0x067cU,0x0cf8U,0x19f0U,0x33e0U,0x67c0U,0xcf80U,0x8f0bU,0x0e1dU,
		0x1c3aU,0x3874U,0x70e8U,0xe1d0U,0xd3abU,0xb75dU,0x7eb1U,0xfd62U,0xeacfU,0xc595U,
		0x9b21U,0x2649U,0x4c92U,0x9924U,0x2243U,0x4486U,0x890cU,0x0213U,0x0426U,0x084cU,
		0x1098U,0x2130U,0x4260U,0x84c0U,0x198bU,0x3316U,0x662cU,0xcc58U,0x88bbU,0x017dU,
		0x02faU,0x05f4U,0x0be8U,0x17d0U,0x2fa0U,0x5f40U,0xbe80U,0x6d0bU,0xda16U,0xa427U,
		0x5845U,0xb08aU,0x711fU,0xe23eU,0xd477U,0xb8e5U,0x61c1U,0xc382U,0x970fU,0x3e15U,
		0x7c2aU,0xf854U,0xe0a3U,0xd14dU,0xb291U,0x7529U,0xea52U,0xc4afU,0x9955U,0x22a1U,
		0x4542U,0x8a84U,0x0503U,0x0a06U,0x140cU,0x2818U,0x5030U,0xa060U,0x50cbU,0xa196U,
		0x5327U,0xa64eU,0x5c97U,0xb92eU,0x6257U,0xc4aeU,0x9957U,0x22a5U,0x454aU,0x8a94U,
		0x0523U,0x0a46U,0x148cU,0x2918U,0x5230U,0xa460U,0x58cbU,0xb196U,0x7327U,0xe64eU,
		0xdc97U,0xa925U,0x4241U,0x8482U,0x190fU,0x321eU,0x643cU,0xc878U,0x80fbU,0x11fdU,
		0x23faU,0x47f4U,0x8fe8U,0x0fdbU,0x1fb6U,0x3f6cU,0x7ed8U,0xfdb0U,0xeb6bU,0xc6ddU,
		0x9db1U,0x2b69U,0x56d2U,0xada4U,0x4b43U,0x9686U,0x3d07U,0x7a0eU,0xf41cU,0xf833U,
		0xe06dU,0xd0d1U,0xb1a9U,0x7359U,0xe6b2U,0xdd6fU,0xaad5U,0x45a1U,0x8b42U,0x068fU,
		0x0d1eU,0x1a3cU,0x3478U,0x68f0U,0xd1e0U,0xb3cbU,0x779dU,0xef3aU,0xce7fU,0x8cf5U,
		0x09e1U,0x13c2U,0x2784U,0x4f08U,0x9e10U,0x2c2bU,0x5856U,0xb0acU,0x7153U,0xe2a6U,
		0xd547U,0xba85U,0x6501U,0xca02U,0x840fU,0x1815U,0x302aU,0x6054U,0xc0a8U,0x915bU,
		0x32bdU,0x657aU,0xcaf4U,0x85e3U,0x1bcdU,0x379aU,0x6f34U,0xde68U,0xacdbU,0x49bdU,
		0x937aU,0x36ffU,0x6dfeU,0xdbfcU,0xa7f3U,0x5fedU,0xbfdaU,0x6fbfU,0xdf7eU,0xaef7U,
		0x4de5U,0x9bcaU,0x279fU,0x4f3eU,0x9e7cU,0x2cf3U,0x59e6U,0xb3ccU,0x7793U,0xef26U,
		0xce47U,0x8c85U,0x0901U,0x1202U,0x2404U,0x4808U,0x9010U,0x302bU,0x6056U,0xc0acU,
		0x9153U,0x32adU,0x655aU,0xcab4U,0x8563U,0x1acdU,0x359aU,0x6b34U,0xd668U,0xbcdbU,
		0x69bdU,0xd37aU,0xb6ffU,0x7df5U,0xfbeaU,0xe7dfU,0xdfb5U,0xaf61U,0x4ec9U,0x9d92U,
		0x2b2fU,0x565eU,0xacbcU,0x4973U,0x92e6U,0x35c7U,0x6b8eU,0xd71cU,0xbe33U,0x6c6dU,
		0xd8daU,0xa1bfU,0x5375U,0xa6eaU,0x5ddfU,0xbbbeU,0x6777U,0xceeeU,0x8dd7U,0x0ba5U,
		0x174aU,0x2e94U,0x5d28U,0xba50U,0x64abU,0xc956U,0x82a7U,0x1545U,0x2a8aU,0x5514U,
		0xaa28U,0x445bU,0x88b6U,0x0167U,0x02ceU,0x059cU,0x0b38U,0x1670U,0x2ce0U,0x59c0U,
		0xb380U,0x770bU,0xee16U,0xcc27U,0x8845U,0x0081U,0x0102U,0x0204U,0x0408U,0x0810U,
		0x1020U,0x2040U,0x4080U,0x8100U,0x120bU,0x2416U,0x482cU,0x9058U,0x30bbU,0x6176U,
		0xc2ecU,0x95d3U,0x3badU,0x775aU,0xeeb4U,0xcd63U,0x8acdU,0x0591U,0x0b22U,0x1644U,
		0x2c88U,0x5910U,0xb220U,0x744bU,0xe896U,0xc127U,0x9245U,0x3481U,0x6902U,0xd204U,
		0xb403U,0x780dU,0xf01aU,0xf03fU,0xf075U,0xf0e1U,0xf1c9U,0xf399U,0xf739U,0xfe79U,
		0xecf9U,0xc9f9U,0x83f9U,0x17f9U,0x2ff2U,0x5fe4U,0xbfc8U,0x6f9bU,0xdf36U,0xae67U,
		0x4cc5U,0x998aU,0x231fU,0x463eU,0x8c7cU,0x08f3U,0x11e6U,0x23ccU,0x4798U,0x8f30U,
		0x0e6bU,0x1cd6U,0x39acU,0x7358U,0xe6b0U,0xdd6bU,0xaaddU,0x45b1U,0x8b62U,0x06cfU,
		0x0d9eU,0x1b3cU,0x3678U,0x6cf0U,0xd9e0U,0xa3cbU,0x579dU,0xaf3aU,0x4e7fU,0x9cfeU,
		0x29f7U,0x53eeU,0xa7dcU,0x5fb3U,0xbf66U,0x6ec7U,0xdd8eU,0xab17U,0x4625U,0x8c4aU,
		0x089fU,0x113eU,0x227cU,0x44f8U,0x89f0U,0x03ebU,0x07d6U,0x0facU,0x1f58U,0x3eb0U,
		0x7d60U,0xfac0U,0xe58bU,0xdb1dU,0xa631U,0x5c69U,0xb8d2U,0x61afU,0xc35eU,0x96b7U,
		0x3d65U,0x7acaU,0xf594U,0xfb23U,0xe64dU,0xdc91U,0xa929U,0x4259U,0x84b2U,0x196fU,
		0x32deU,0x65bcU,0xcb78U,0x86fbU,0x1dfdU,0x3bfaU,0x77f4U,0xefe8U,0xcfdbU,0x8fbdU,
		0x0f71U,0x1ee2U,0x3dc4U,0x7b88U,0xf710U,0xfe2bU,0xec5dU,0xc8b1U,0x8169U,0x12d9U,
		0x25b2U,0x4b64U,0x96c8U,0x3d9bU,0x7b36U,0xf66cU,0xfcd3U,0xe9adU,0xc351U,0x96a9U,
		0x3d59U,0x7ab2U,0xf564U,0xfac3U,0xe58dU,0xdb11U,0xa629U,0x5c59U,0xb8b2U,0x616fU,
		0xc2deU,0x95b7U,0x3b65U,0x76caU,0xed94U,0xcb23U,0x864dU,0x1c91U,0x3922U,0x7244U,
		0xe488U,0xd91bU,0xa23dU,0x5471U,0xa8e2U,0x41cfU,0x839eU,0x1737U,0x2e6eU,0x5cdcU,
		0xb9b8U,0x637bU,0xc6f6U,0x9de7U,0x2bc5U,0x578aU,0xaf14U,0x4e23U,0x9c46U,0x2887U,
		0x510eU,0xa21cU,0x5433U,0xa866U,0x40c7U,0x818eU,0x1317U,0x262eU,0x4c5cU,0x98b8U,
		0x217bU,0x42f6U,0x85ecU,0x1bd3U,0x37a6U,0x6f4cU,0xde98U,0xad3bU,0x4a7dU,0x94faU,
		0x39ffU,0x73feU,0xe7fcU,0xdff3U,0xafedU,0x4fd1U,0x9fa2U,0x2f4fU,0x5e9eU,0xbd3cU,
		0x6a73U,0xd4e6U,0xb9c7U,0x6385U,0xc70aU,0x9e1fU,0x2c35U,0x586aU,0xb0d4U,0x71a3U,
		0xe346U,0xd687U,0xbd05U,0x6a01U,0xd402U,0xb80fU,0x6015U,0xc02aU,0x905fU,0x30b5U,
		0x616aU,0xc2d4U,0x95a3U,0x3b4dU,0x769aU,0xed34U,0xca63U,0x84cdU,0x1991U,0x3322U,
		0x6644U,0xcc88U,0x891bU,0x023dU,0x047aU,0x08f4U,0x11e8U,0x23d0U,0x47a0U,0x8f40U,
		0x0e8bU,0x1d16U,0x3a2cU,0x7458U,0xe8b0U,0xc16bU,0x92ddU,0x35b1U,0x6b62U,0xd6c4U,
		0xbd83U,0x6b0dU,0xd61aU,0xbc3fU,0x6875U,0xd0eaU,0xb1dfU,0x73b5U,0xe76aU,0xdedfU,
		0xadb5U,0x4b61U,0x96c2U,0x3d8fU,0x7b1eU,0xf63cU,0xfc73U,0xe8edU,0xc1d1U,0x93a9U,
		0x3759U,0x6eb2U,0xdd64U,0xaac3U,0x458dU,0x8b1aU,0x063fU,0x0c7eU,0x18fcU,0x31f8U,
		0x63f0U,0xc7e0U,0x9fcbU,0x2f9dU,0x5f3aU,0xbe74U,0x6ce3U,0xd9c6U,0xa387U,0x5705U,
		0xae0aU,0x4c1fU,0x983eU,0x2077U,0x40eeU,0x81dcU,0x13b3U,0x2766U,0x4eccU,0x9d98U,
		0x2b3bU,0x5676U,0xacecU,0x49d3U,0x93a6U,0x3747U,0x6e8eU,0xdd1cU,0xaa33U,0x446dU,
		0x88daU,0x01bfU,0x037eU,0x06fcU,0x0df8U,0x1bf0U,0x37e0U,0x6fc0U,0xdf80U,0xaf0bU,
		0x4e1dU,0x9c3aU,0x287fU,0x50feU,0xa1fcU,0x53f3U,0xa7e6U,0x5fc7U,0xbf8eU,0x6f17U,
		0xde2eU,0xac57U,0x48a5U,0x914aU,0x329fU,0x653eU,0xca7cU,0x84f3U,0x19edU,0x33daU,
		0x67b4U,0xcf68U,0x8edbU,0x0dbdU,0x1b7aU,0x36f4U,0x6de8U,0xdbd0U,0xa7abU,0x5f5dU,
		0xbebaU,0x6d7fU,0xdafeU,0xa5f7U,0x5be5U,0xb7caU,0x7f9fU,0xff3eU,0xee77U,0xcce5U,
		0x89c1U,0x0389U,0x0712U,0x0e24U,0x1c48U,0x3890U,0x7120U,0xe240U,0xd48bU,0xb91dU,
		0x6231U,0xc462U,0x98cfU,0x2195U,0x432aU,0x8654U,0x1ca3U,0x3946U,0x728cU,0xe518U,
		0xda3bU,0xa47dU,0x58f1U,0xb1e2U,0x73cfU,0xe79eU,0xdf37U,0xae65U,0x4cc1U,0x9982U,
		0x230fU,0x461eU,0x8c3cU,0x0873U,0x10e6U,0x21ccU,0x4398U,0x8730U,0x1e6bU,0x3cd6U,
		0x79acU,0xf358U,0xf6bbU,0xfd7dU,0xeaf1U,0xc5e9U,0x9bd9U,0x27b9U,0x4f72U,0x9ee4U,
		0x2dc3U,0x5b86U,0xb70cU,0x7e13U,0xfc26U,0xe847U,0xc085U,0x9101U,0x3209U,0x6412U,
		0xc824U,0x8043U,0x108dU,0x211aU,0x4234U,0x8468U,0x18dbU,0x31b6U,0x636cU,0xc6d8U,
		0x9dbbU,0x2b7dU,0x56faU,0xadf4U,0x4be3U,0x97c6U,0x3f87U,0x7f0eU,0xfe1cU,0xec33U,
		0xc86dU,0x80d1U,0x11a9U,0x2352U,0x46a4U,0x8d48U,0x0a9bU,0x1536U,0x2a6cU,0x54d8U,
		0xa9b0U,0x436bU,0x86d6U,0x1da7U,0x3b4eU,0x769cU,0xed38U,0xca7bU,0x84fdU,0x19f1U,
		0x33e2U,0x67c4U,0xcf88U,0x8f1bU,0x0e3dU,0x1c7aU,0x38f4U,0x71e8U,0xe3d0U,0xd7abU,
		0xbf5dU,0x6eb1U,0xdd62U,0xaacfU,0x4595U,0x8b2aU,0x065fU,0x0cbeU,0x197cU,0x32f8U,
		0x65f0U,0xcbe0U,0x87cbU,0x1f9dU,0x3f3aU,0x7e74U,0xfce8U,0xe9dbU,0xc3bdU,0x9771U,
		0x3ee9U,0x7dd2U,0xfba4U,0xe743U,0xde8dU,0xad11U,0x4a29U,0x9452U,0x38afU,0x715eU,
		0xe2bcU,0xd573U,0xbaedU,0x65d1U,0xcba2U,0x874fU,0x1e95U,0x3d2aU,0x7a54U,0xf4a8U,
		0xf95bU,0xe2bdU,0xd571U,0xbae9U,0x65d9U,0xcbb2U,0x876fU,0x1ed5U,0x3daaU,0x7b54U,
		0xf6a8U,0xfd5bU,0xeabdU,0xc571U,0x9ae9U,0x25d9U,0x4bb2U,0x9764U,0x3ec3U,0x7d86U,
		0xfb0cU,0xe613U,0xdc2dU,0xa851U,0x40a9U,0x8152U,0x12afU,0x255eU,0x4abcU,0x9578U,
		0x3afbU,0x75f6U,0xebecU,0xc7d3U,0x9fadU,0x2f51U,0x5ea2U,0xbd44U,0x6a83U,0xd506U,
		0xba07U,0x6405U,0xc80aU,0x801fU,0x1035U,0x206aU,0x40d4U,0x81a8U,0x135bU,0x26b6U,
		0x4d6cU,0x9ad8U,0x25bbU,0x4b76U,0x96ecU,0x3dd3U,0x7ba6U,0xf74cU,0xfe93U,0xed2dU,
		0xca51U,0x84a9U,0x1959U,0x32b2U,0x6564U,0xcac8U,0x859bU,0x1b3dU,0x367aU,0x6cf4U,
		0xd9e8U,0xa3dbU,0x57bdU,0xaf7aU,0x4effU,0x9dfeU,0x2bf7U,0x57eeU,0xafdcU,0x4fb3U,
		0x9f66U,0x2ec7U,0x5d8eU,0xbb1cU,0x6633U,0xcc66U,0x88c7U,0x0185U,0x030aU,0x0614U,
		0x0c28U,0x1850U,0x30a0U,0x6140U,0xc280U,0x950bU,0x3a1dU,0x743aU,0xe874U,0xc0e3U,
		0x91cdU,0x3391U,0x6722U,0xce44U,0x8c83U,0x090dU,0x121aU,0x2434U,0x4868U,0x90d0U,
		0x31abU,0x6356U,0xc6acU,0x9d53U,0x2aadU,0x555aU,0xaab4U,0x4563U,0x8ac6U,0x0587U,
		0x0b0eU,0x161cU,0x2c38U,0x5870U,0xb0e0U,0x71cbU,0xe396U,0xd727U,0xbe45U,0x6c81U,
		0xd902U,0xa20fU,0x5415U,0xa82aU,0x405fU,0x80beU,0x1177U,0x22eeU,0x45dcU,0x8bb8U,
		0x077bU,0x0ef6U,0x1decU,0x3bd8U,0x77b0U,0xef60U,0xcecbU,0x8d9dU,0x0b31U,0x1662U,
		0x2cc4U,0x5988U,0xb310U,0x762bU,0xec56U,0xc8a7U,0x8145U,0x1281U,0x2502U,0x4a04U,
		0x9408U,0x381bU,0x7036U,0xe06cU,0xd0d3U,0xb1adU,0x7351U,0xe6a2U,0xdd4fU,0xaa95U,
		0x4521U,0x8a42U,0x048fU,0x091eU,0x123cU,0x2478U,0x48f0U,0x91e0U,0x33cbU,0x6796U,
		0xcf2cU,0x8e53U,0x0cadU,0x195aU,0x32b4U,0x6568U,0xcad0U,0x85abU,0x1b5dU,0x36baU,
		0x6d74U,0xdae8U,0xa5dbU,0x5bbdU,0xb77aU,0x7effU,0xfdfeU,0xebf7U,0xc7e5U,0x9fc1U,
		0x2f89U,0x5f12U,0xbe24U,0x6c43U,0xd886U,0xa107U,0x5205U,0xa40aU,0x581fU,0xb03eU,
		0x7077U,0xe0eeU,0xd1d7U,0xb3a5U,0x7741U,0xee82U,0xcd0fU,0x8a15U,0x0421U,0x0842U,
		0x1084U,0x2108U,0x4210U,0x8420U,0x184bU,0x3096U,0x612cU,0xc258U,0x94bbU,0x397dU,
		0x72faU,0xe5f4U,0xdbe3U,0xa7cdU,0x5f91U,0xbf22U,0x6e4fU,0xdc9eU,0xa937U,0x4265U,
		0x84caU,0x199fU,0x333eU,0x667cU,0xccf8U,0x89fbU,0x03fdU,0x07faU,0x0ff4U,0x1fe8U,
		0x3fd0U,0x7fa0U,0xff40U,0xee8bU,0xcd1dU,0x8a31U,0x0469U,0x08d2U,0x11a4U,0x2348U,
		0x4690U,0x8d20U,0x0a4bU,0x1496U,0x292cU,0x5258U,0xa4b0U,0x596bU,0xb2d6U,0x75a7U,
		0xeb4eU,0xc697U,0x9d25U,0x2a41U,0x5482U,0xa904U,0x4203U,0x8406U,0x1807U,0x300eU,
		0x601cU,0xc038U,0x907bU,0x30fdU,0x61faU,0xc3f4U,0x97e3U,0x3fcdU,0x7f9aU,0xff34U,
		0xee63U,0xcccdU,0x8991U,0x0329U,0x0652U,0x0ca4U,0x1948U,0x3290U,0x6520U,0xca40U,
		0x848bU,0x191dU,0x323aU,0x6474U,0xc8e8U,0x81dbU,0x13bdU,0x277aU,0x4ef4U,0x9de8U,
		0x2bdbU,0x57b6U,0xaf6cU,0x4ed3U,0x9da6U,0x2b47U,0x568eU,0xad1cU,0x4a33U,0x9466U,
		0x38c7U,0x718eU,0xe31cU,0xd633U,0xbc6dU,0x68d1U,0xd1a2U,0xb34fU,0x7695U,0xed2aU,
		0xca5fU,0x84b5U,0x1961U,0x32c2U,0x6584U,0xcb08U,0x861bU,0x1c3dU,0x387aU,0x70f4U,
		0xe1e8U,0xd3dbU,0xb7bdU,0x7f71U,0xfee2U,0xedcfU,0xcb95U,0x8721U,0x1e49U,0x3c92U,
		0x7924U,0xf248U,0xf49bU,0xf93dU,0xe271U,0xd4e9U,0xb9d9U,0x63b9U,0xc772U,0x9eefU,
		0x2dd5U,0x5baaU,0xb754U,0x7ea3U,0xfd46U,0xea87U,0xc505U,0x9a01U,0x2409U,0x4812U,
		0x9024U,0x3043U,0x6086U,0xc10cU,0x9213U,0x342dU,0x685aU,0xd0b4U,0xb163U,0x72cdU,
		0xe59aU,0xdb3fU,0xa675U,0x5ce1U,0xb9c2U,0x638fU,0xc71eU,0x9e37U,0x2c65U,0x58caU,
		0xb194U,0x7323U,0xe646U,0xdc87U,0xa905U,0x4201U,0x8402U,0x180fU,0x301eU,0x603cU,
		0xc078U,0x90fbU,0x31fdU,0x63faU,0xc7f4U,0x9fe3U,0x2fcdU,0x5f9aU,0xbf34U,0x6e63U,
		0xdcc6U,0xa987U,0x4305U,0x860aU,0x1c1fU,0x383eU,0x707cU,0xe0f8U,0xd1fbU,0xb3fdU,
		0x77f1U,0xefe2U,0xcfcfU,0x8f95U,0x0f21U,0x1e42U,0x3c84U,0x7908U,0xf210U,0xf42bU,
		0xf85dU,0xe0b1U,0xd169U,0xb2d9U,0x75b9U,0xeb72U,0xc6efU,0x9dd5U,0x2ba1U,0x5742U,
		0xae84U,0x4d03U,0x9a06U,0x2407U,0x480eU,0x901cU,0x3033U,0x6066U,0xc0ccU,0x9193U,
		0x332dU,0x665aU,0xccb4U,0x8963U,0x02cdU,0x059aU,0x0b34U,0x1668U,0x2cd0U,0x59a0U,
		0xb340U,0x768bU,0xed16U,0xca27U,0x8445U,0x1881U,0x3102U,0x6204U,0xc408U,0x981bU,
		0x203dU,0x407aU,0x80f4U,0x11e3U,0x23c6U,0x478cU,0x8f18U,0x0e3bU,0x1c76U,0x38ecU,
		0x71d8U,0xe3b0U,0xd76bU,0xbeddU,0x6db1U,0xdb62U,0xa6cfU,0x5d95U,0xbb2aU,0x665fU,
		0xccbeU,0x8977U,0x02e5U,0x05caU,0x0b94U,0x1728U,0x2e50U,0x5ca0U,0xb940U,0x628bU,
		0xc516U,0x9a27U,0x2445U,0x488aU,0x9114U,0x3223U,0x6446U,0xc88cU,0x8113U,0x122dU,
		0x245aU,0x48b4U,0x9168U,0x32dbU,0x65b6U,0xcb6cU,0x86d3U,0x1dadU,0x3b5aU,0x76b4U,
		0xed68U,0xcadbU,0x85bdU,0x1b71U,0x36e2U,0x6dc4U,0xdb88U,0xa71bU,0x5e3dU,0xbc7aU,
		0x68ffU,0xd1feU,0xb3f7U,0x77e5U,0xefcaU,0xcf9fU,0x8f35U,0x0e61U,0x1cc2U,0x3984U,
		0x7308U,0xe610U,0xdc2bU,0xa85dU,0x40b1U,0x8162U,0x12cfU,0x259eU,0x4b3cU,0x9678U,
		0x3cfbU,0x79f6U,0xf3ecU,0xf7d3U,0xffadU,0xef51U,0xcea9U,0x8d59U,0x0ab9U,0x1572U,
		0x2ae4U,0x55c8U,0xab90U,0x472bU,0x8e56U,0x0ca7U,0x194eU,0x329cU,0x6538U,0xca70U,
		0x84ebU,0x19ddU,0x33baU,0x6774U,0xcee8U,0x8ddbU,0x0bbdU,0x177aU,0x2ef4U,0x5de8U,
		0xbbd0U,0x67abU,0xcf56U,0x8ea7U,0x0d45U,0x1a8aU,0x3514U,0x6a28U,0xd450U,0xb8abU,
		0x615dU,0xc2baU,0x957fU,0x3af5U,0x75eaU,0xebd4U,0xc7a3U,0x9f4dU,0x2e91U,0x5d22U,
		0xba44U,0x6483U,0xc906U,0x8207U,0x1405U,0x280aU,0x5014U,0xa028U,0x505bU,0xa0b6U,
		0x5167U,0xa2ceU,0x5597U,0xab2eU,0x4657U,0x8caeU,0x0957U,0x12aeU,0x255cU,0x4ab8U,
		0x9570U,0x3aebU,0x75d6U,0xebacU,0xc753U,0x9eadU,0x2d51U,0x5aa2U,0xb544U,0x7a83U,
		0xf506U,0xfa07U,0xe405U,0xd801U,0xa009U,0x5019U,0xa032U,0x506fU,0xa0deU,0x51b7U,
		0xa36eU,0x56d7U,0xadaeU,0x4b57U,0x96aeU,0x3d57U,0x7aaeU,0xf55cU,0xfab3U,0xe56dU,
		0xdad1U,0xa5a9U,0x5b59U,0xb6b2U,0x7d6fU,0xfadeU,0xe5b7U,0xdb65U,0xa6c1U,0x5d89U,
		0xbb12U,0x662fU,0xcc5eU,0x88b7U,0x0165U,0x02caU,0x0594U,0x0b28U,0x1650U,0x2ca0U,
		0x5940U,0xb280U,0x750bU,0xea16U,0xc427U,0x9845U,0x2081U,0x4102U,0x8204U,0x1403U,
		0x2806U,0x500cU,0xa018U,0x503bU,0xa076U,0x50e7U,0xa1ceU,0x5397U,0xa72eU,0x5e57U,
		0xbcaeU,0x6957U,0xd2aeU,0xb557U,0x7aa5U,0xf54aU,0xfa9fU,0xe535U,0xda61U,0xa4c9U,
		0x5999U,0xb332U,0x766fU,0xecdeU,0xc9b7U,0x8365U,0x16c1U,0x2d82U,0x5b04U,0xb608U,
		0x7c1bU,0xf836U,0xe067U,0xd0c5U,0xb181U,0x7309U,0xe612U,0xdc2fU,0xa855U,0x40a1U,
		0x8142U,0x128fU,0x251eU,0x4a3cU,0x9478U,0x38fbU,0x71f6U,0xe3ecU,0xd7d3U,0xbfadU,
		0x6f51U,0xdea2U,0xad4fU,0x4a95U,0x952aU,0x3a5fU,0x74beU,0xe97cU,0xc2f3U,0x95edU,
		0x3bd1U,0x77a2U,0xef44U,0xce83U,0x8d0dU,0x0a11U,0x1422U,0x2844U,0x5088U,0xa110U,
		0x522bU,0xa456U,0x58a7U,0xb14eU,0x7297U,0xe52eU,0xda57U,0xa4a5U,0x5941U,0xb282U,
		0x750fU,0xea1eU,0xc437U,0x9865U,0x20c1U,0x4182U,0x8304U,0x1603U,0x2c06U,0x580cU,
		0xb018U,0x703bU,0xe076U,0xd0e7U,0xb1c5U,0x7381U,0xe702U,0xde0fU,0xac15U,0x4821U,
		0x9042U,0x308fU,0x611eU,0xc23cU,0x9473U,0x38edU,0x71daU,0xe3b4U,0xd763U,0xbecdU,
		0x6d91U,0xdb22U,0xa64fU,0x5c95U,0xb92aU,0x625fU,0xc4beU,0x9977U,0x22e5U,0x45caU,
		0x8b94U,0x0723U,0x0e46U,0x1c8cU,0x3918U,0x7230U,0xe460U,0xd8cbU,0xa19dU,0x5331U,
		0xa662U,0x5ccfU,0xb99eU,0x6337U,0xc66eU,0x9cd7U,0x29a5U,0x534aU,0xa694U,0x5d23U,
		0xba46U,0x6487U,0xc90eU,0x8217U,0x1425U,0x284aU,0x5094U,0xa128U,0x525bU,0xa4b6U,
		0x5967U,0xb2ceU,0x7597U,0xeb2eU,0xc657U,0x9ca5U,0x2941U,0x5282U,0xa504U,0x5a03U,
		0xb406U,0x7807U,0xf00eU,0xf017U,0xf025U,0xf041U,0xf089U,0xf119U,0xf239U,0xf479U,
		0xf8f9U,0xe1f9U,0xd3f9U,0xb7f9U,0x7ff9U,0xfff2U,0xefefU,0xcfd5U,0x8fa1U,0x0f49U,
		0x1e92U,0x3d24U,0x7a48U,0xf490U,0xf92bU,0xe25dU,0xd4b1U,0xb969U,0x62d9U,0xc5b2U,
		0x9b6fU,0x26d5U,0x4daaU,0x9b54U,0x26a3U,0x4d46U,0x9a8cU,0x2513U,0x4a26U,0x944cU,
		0x3893U,0x7126U,0xe24cU,0xd493U,0xb92dU,0x6251U,0xc4a2U,0x994fU,0x2295U,0x452aU,
		0x8a54U,0x04a3U,0x0946U,0x128cU,0x2518U,0x4a30U,0x9460U,0x38cbU,0x7196U,0xe32cU,
		0xd653U,0xbcadU,0x6951U,0xd2a2U,0xb54fU,0x7a95U,0xf52aU,0xfa5fU,0xe4b5U,0xd961U,
		0xa2c9U,0x5599U,0xab32U,0x466fU,0x8cdeU,0x09b7U,0x136eU,0x26dcU,0x4db8U,0x9b70U,
		0x26ebU,0x4dd6U,0x9bacU,0x2753U,0x4ea6U,0x9d4cU,0x2a93U,0x5526U,0xaa4cU,0x4493U,
		0x8926U,0x0247U,0x048eU,0x091cU,0x1238U,0x2470U,0x48e0U,0x91c0U,0x338bU,0x6716U,
		0xce2cU,0x8c53U,0x08adU,0x115aU,0x22b4U,0x4568U,0x8ad0U,0x05abU,0x0b56U,0x16acU,
		0x2d58U,0x5ab0U,0xb560U,0x7acbU,0xf596U,0xfb27U,0xe645U,0xdc81U,0xa909U,0x4219U,
		0x8432U,0x186fU,0x30deU,0x61bcU,0xc378U,0x96fbU,0x3dfdU,0x7bfaU,0xf7f4U,0xffe3U,
		0xefcdU,0xcf91U,0x8f29U,0x0e59U,0x1cb2U,0x3964U,0x72c8U,0xe590U,0xdb2bU,0xa65dU,
		0x5cb1U,0xb962U,0x62cfU,0xc59eU,0x9b37U,0x2665U,0x4ccaU,0x9994U,0x2323U,0x4646U,
		0x8c8cU,0x0913U,0x1226U,0x244cU,0x4898U,0x9130U,0x326bU,0x64d6U,0xc9acU,0x8353U,
		0x16adU,0x2d5aU,0x5ab4U,0xb568U,0x7adbU,0xf5b6U,0xfb67U,0xe6c5U,0xdd81U,0xab09U,
		0x4619U,0x8c32U,0x086fU,0x10deU,0x21bcU,0x4378U,0x86f0U,0x1debU,0x3bd6U,0x77acU,
		0xef58U,0xcebbU,0x8d7dU,0x0af1U,0x15e2U,0x2bc4U,0x5788U,0xaf10U,0x4e2bU,0x9c56U,
		0x28a7U,0x514eU,0xa29cU,0x5533U,0xaa66U,0x44c7U,0x898eU,0x0317U,0x062eU,0x0c5cU,
		0x18b8U,0x3170U,0x62e0U,0xc5c0U,0x9b8bU,0x271dU,0x4e3aU,0x9c74U,0x28e3U,0x51c6U,
		0xa38cU,0x5713U,0xae26U,0x4c47U,0x988eU,0x2117U,0x422eU,0x845cU,0x18b3U,0x3166U,
		0x62ccU,0xc598U,0x9b3bU,0x267dU,0x4cfaU,0x99f4U,0x23e3U,0x47c6U,0x8f8cU,0x0f13U,
		0x1e26U,0x3c4cU,0x7898U,0xf130U,0xf26bU,0xf4ddU,0xf9b1U,0xe369U,0xd6d9U,0xbdb9U,
		0x6b79U,0xd6f2U,0xbdefU,0x6bd5U,0xd7aaU,0xbf5fU,0x6eb5U,0xdd6aU,0xaadfU,0x45b5U,
		0x8b6aU,0x06dfU,0x0dbeU,0x1b7cU,0x36f8U,0x6df0U,0xdbe0U,0xa7cbU,0x5f9dU,0xbf3aU,
		0x6e7fU,0xdcfeU,0xa9f7U,0x43e5U,0x87caU,0x1f9fU,0x3f3eU,0x7e7cU,0xfcf8U,0xe9fbU,
		0xc3fdU,0x97f1U,0x3fe9U,0x7fd2U,0xffa4U,0xef43U,0xce8dU,0x8d11U,0x0a29U,0x1452U,
		0x28a4U,0x5148U,0xa290U,0x552bU,0xaa56U,0x44a7U,0x894eU,0x0297U,0x052eU,0x0a5cU,
		0x14b8U,0x2970U,0x52e0U,0xa5c0U,0x5b8bU,0xb716U,0x7e27U,0xfc4eU,0xe897U,0xc125U,
		0x9241U,0x3489U,0x6912U,0xd224U,0xb443U,0x788dU,0xf11aU,0xf23fU,0xf475U,0xf8e1U,
		0xe1c9U,0xd399U,0xb739U,0x7e79U,0xfcf2U,0xe9efU,0xc3d5U,0x97a1U,0x3f49U,0x7e92U,
		0xfd24U,0xea43U,0xc48dU,0x9911U,0x2229U,0x4452U,0x88a4U,0x0143U,0x0286U,0x050cU,
		0x0a18U,0x1430U,0x2860U,0x50c0U,0xa180U,0x530bU,0xa616U,0x5c27U,0xb84eU,0x6097U,
		0xc12eU,0x9257U,0x34a5U,0x694aU,0xd294U,0xb523U,0x7a4dU,0xf49aU,0xf93fU,0xe275U,
		0xd4e1U,0xb9c9U,0x6399U,0xc732U,0x9e6fU,0x2cd5U,0x59aaU,0xb354U,0x76a3U,0xed46U,
		0xca87U,0x8505U,0x1a01U,0x3402U,0x6804U,0xd008U,0xb01bU,0x703dU,0xe07aU,0xd0ffU,
		0xb1f5U,0x73e1U,0xe7c2U,0xdf8fU,0xaf15U,0x4e21U,0x9c42U,0x288fU,0x511eU,0xa23cU,
		0x5473U,0xa8e6U,0x41c7U,0x838eU,0x1717U,0x2e2eU,0x5c5cU,0xb8b8U,0x617bU,0xc2f6U,
		0x95e7U,0x3bc5U,0x778aU,0xef14U,0xce23U,0x8c4dU,0x0891U,0x1122U,0x2244U,0x4488U,
		0x8910U,0x022bU,0x0456U,0x08acU,0x1158U,0x22b0U,0x4560U,0x8ac0U,0x058bU,0x0b16U,
		0x162cU,0x2c58U,0x58b0U,0xb160U,0x72cbU,0xe596U,0xdb27U,0xa645U,0x5c81U,0xb902U,
		0x620fU,0xc41eU,0x9837U,0x2065U,0x40caU,0x8194U,0x1323U,0x2646U,0x4c8cU,0x9918U,
		0x223bU,0x4476U,0x88ecU,0x01d3U,0x03a6U,0x074cU,0x0e98U,0x1d30U,0x3a60U,0x74c0U,
		0xe980U,0xc30bU,0x961dU,0x3c31U,0x7862U,0xf0c4U,0xf183U,0xf30dU,0xf611U,0xfc29U,
		0xe859U,0xc0b9U,0x9179U,0x32f9U,0x65f2U,0xcbe4U,0x87c3U,0x1f8dU,0x3f1aU,0x7e34U,
		0xfc68U,0xe8dbU,0xc1bdU,0x9371U,0x36e9U,0x6dd2U,0xdba4U,0xa743U,0x5e8dU,0xbd1aU,
		0x6a3fU,0xd47eU,0xb8f7U,0x61e5U,0xc3caU,0x979fU,0x3f35U,0x7e6aU,0xfcd4U,0xe9a3U,
		0xc34dU,0x9691U,0x3d29U,0x7a52U,0xf4a4U,0xf943U,0xe28dU,0xd511U,0xba29U,0x6459U,
		0xc8b2U,0x816fU,0x12d5U,0x25aaU,0x4b54U,0x96a8U,0x3d5bU,0x7ab6U,0xf56cU,0xfad3U,
		0xe5adU,0xdb51U,0xa6a9U,0x5d59U,0xbab2U,0x656fU,0xcadeU,0x85b7U,0x1b65U,0x36caU,
		0x6d94U,0xdb28U,0xa65bU,0x5cbdU,0xb97aU,0x62ffU,0xc5feU,0x9bf7U,0x27e5U,0x4fcaU,
		0x9f94U,0x2f23U,0x5e46U,0xbc8cU,0x6913U,0xd226U,0xb447U,0x7885U,0xf10aU,0xf21fU,
		0xf435U,0xf861U,0xe0c9U,0xd199U,0xb339U,0x7679U,0xecf2U,0xc9efU,0x83d5U,0x17a1U,
		0x2f42U,0x5e84U,0xbd08U,0x6a1bU,0xd436U,0xb867U,0x60c5U,0xc18aU,0x931fU,0x3635U,
		0x6c6aU,0xd8d4U,0xa1a3U,0x534dU,0xa69aU,0x5d3fU,0xba7eU,0x64f7U,0xc9eeU,0x83d7U,
		0x17a5U,0x2f4aU,0x5e94U,0xbd28U,0x6a5bU,0xd4b6U,0xb967U,0x62c5U,0xc58aU,0x9b1fU,
		0x2635U,0x4c6aU,0x98d4U,0x21a3U,0x4346U,0x868cU,0x1d13U,0x3a26U,0x744cU,0xe898U,
		0xc13bU,0x927dU,0x34f1U,0x69e2U,0xd3c4U,0xb783U,0x7f0dU,0xfe1aU,0xec3fU,0xc875U,
		0x80e1U,0x11c9U,0x2392U,0x4724U,0x8e48U,0x0c9bU,0x1936U,0x326cU,0x64d8U,0xc9b0U,
		0x836bU,0x16ddU,0x2dbaU,0x5b74U,0xb6e8U,0x7ddbU,0xfbb6U,0xe767U,0xdec5U,0xad81U,
		0x4b09U,0x9612U,0x3c2fU,0x785eU,0xf0bcU,0xf173U,0xf2edU,0xf5d1U,0xfba9U,0xe759U,
		0xdeb9U,0xad79U,0x4af9U,0x95f2U,0x3befU,0x77deU,0xefbcU,0xcf73U,0x8eedU,0x0dd1U,
		0x1ba2U,0x3744U,0x6e88U,0xdd10U,0xaa2bU,0x445dU,0x88baU,0x017fU,0x02feU,0x05fcU,
		0x0bf8U,0x17f0U,0x2fe0U,0x5fc0U,0xbf80U,0x6f0bU,0xde16U,0xac27U,0x4845U,0x908aU,
		0x311fU,0x623eU,0xc47cU,0x98f3U,0x21edU,0x43daU,0x87b4U,0x1f63U,0x3ec6U,0x7d8cU,
		0xfb18U,0xe63bU,0xdc7dU,0xa8f1U,0x41e9U,0x83d2U,0x17afU,0x2f5eU,0x5ebcU,0xbd78U,
		0x6afbU,0xd5f6U,0xbbe7U,0x67c5U,0xcf8aU,0x8f1fU,0x0e35U,0x1c6aU,0x38d4U,0x71a8U,
		0xe350U,0xd6abU,0xbd5dU,0x6ab1U,0xd562U,0xbacfU,0x6595U,0xcb2aU,0x865fU,0x1cb5U,
		0x396aU,0x72d4U,0xe5a8U,0xdb5bU,0xa6bdU,0x5d71U,0xbae2U,0x65cfU,0xcb9eU,0x8737U,
		0x1e65U,0x3ccaU,0x7994U,0xf328U,0xf65bU,0xfcbdU,0xe971U,0xc2e9U,0x95d9U,0x3bb9U,
		0x7772U,0xeee4U,0xcdc3U,0x8b8dU,0x0711U,0x0e22U,0x1c44U,0x3888U,0x7110U,0xe220U,
		0xd44bU,0xb89dU,0x6131U,0xc262U,0x94cfU,0x3995U,0x732aU,0xe654U,0xdca3U,0xa94dU,
		0x4291U,0x8522U,0x1a4fU,0x349eU,0x693cU,0xd278U,0xb4fbU,0x79fdU,0xf3faU,0xf7ffU,
		0xfff5U,0xefe1U,0xcfc9U,0x8f99U,0x0f39U,0x1e72U,0x3ce4U,0x79c8U,0xf390U,0xf72bU,
		0xfe5dU,0xecb1U,0xc969U,0x82d9U,0x15b9U,0x2b72U,0x56e4U,0xadc8U,0x4b9bU,0x9736U,
		0x3e67U,0x7cceU,0xf99cU,0xe333U,0xd66dU,0xbcd1U,0x69a9U,0xd352U,0xb6afU,0x7d55U,
		0xfaaaU,0xe55fU,0xdab5U,0xa561U,0x5ac9U,0xb592U,0x7b2fU,0xf65eU,0xfcb7U,0xe965U,
		0xc2c1U,0x9589U,0x3b19U,0x7632U,0xec64U,0xc8c3U,0x818dU,0x1311U,0x2622U,0x4c44U,
		0x9888U,0x211bU,0x4236U,0x846cU,0x18d3U,0x31a6U,0x634cU,0xc698U,0x9d3bU,0x2a7dU,
		0x54faU,0xa9f4U,0x43e3U,0x87c6U,0x1f87U,0x3f0eU,0x7e1cU,0xfc38U,0xe87bU,0xc0fdU,
		0x91f1U,0x33e9U,0x67d2U,0xcfa4U,0x8f43U,0x0e8dU,0x1d1aU,0x3a34U,0x7468U,0xe8d0U,
		0xc1abU,0x935dU,0x36b1U,0x6d62U,0xdac4U,0xa583U,0x5b0dU,0xb61aU,0x7c3fU,0xf87eU,
		0xe0f7U,0xd1e5U,0xb3c1U,0x7789U,0xef12U,0xce2fU,0x8c55U,0x08a1U,0x1142U,0x2284U,
		0x4508U,0x8a10U,0x042bU,0x0856U,0x10acU,0x2158U,0x42b0U,0x8560U,0x1acbU,0x3596U,
		0x6b2cU,0xd658U,0xbcbbU,0x697dU,0xd2faU,0xb5ffU,0x7bf5U,0xf7eaU,0xffdfU,0xefb5U,
		0xcf61U,0x8ec9U,0x0d99U,0x1b32U,0x3664U,0x6cc8U,0xd990U,0xa32bU,0x565dU,0xacbaU,
		0x497fU,0x92feU,0x35f7U,0x6beeU,0xd7dcU,0xbfb3U,0x6f6dU,0xdedaU,0xadbfU,0x4b75U,
		0x96eaU,0x3ddfU,0x7bbeU,0xf77cU,0xfef3U,0xededU,0xcbd1U,0x87a9U,0x1f59U,0x3eb2U,
		0x7d64U,0xfac8U,0xe59bU,0xdb3dU,0xa671U,0x5ce9U,0xb9d2U,0x63afU,0xc75eU,0x9eb7U,
		0x2d65U,0x5acaU,0xb594U,0x7b23U,0xf646U,0xfc87U,0xe905U,0xc201U,0x9409U,0x3819U,
		0x7032U,0xe064U,0xd0c3U,0xb18dU,0x7311U,0xe622U,0xdc4fU,0xa895U,0x4121U,0x8242U,
		0x148fU,0x291eU,0x523cU,0xa478U,0x58fbU,0xb1f6U,0x73e7U,0xe7ceU,0xdf97U,0xaf25U,
		0x4e41U,0x9c82U,0x290fU,0x521eU,0xa43cU,0x5873U,0xb0e6U,0x71c7U,0xe38eU,0xd717U,
		0xbe25U,0x6c41U,0xd882U,0xa10fU,0x5215U,0xa42aU,0x585fU,0xb0beU,0x7177U,0xe2eeU,
		0xd5d7U,0xbba5U,0x6741U,0xce82U,0x8d0fU,0x0a15U,0x142aU,0x2854U,0x50a8U,0xa150U,
		0x52abU,0xa556U,0x5aa7U,0xb54eU,0x7a97U,0xf52eU,0xfa57U,0xe4a5U,0xd941U,0xa289U,
		0x5519U,0xaa32U,0x446fU,0x88deU,0x01b7U,0x036eU,0x06dcU,0x0db8U,0x1b70U,0x36e0U,
		0x6dc0U,0xdb80U,0xa70bU,0x5e1dU,0xbc3aU,0x687fU,0xd0feU,0xb1f7U,0x73e5U,0xe7caU,
		0xdf9fU,0xaf35U,0x4e61U,0x9cc2U,0x298fU,0x531eU,0xa63cU,0x5c73U,0xb8e6U,0x61c7U,
		0xc38eU,0x9717U,0x3e25U,0x7c4aU,0xf894U,0xe123U,0xd24dU,0xb491U,0x7929U,0xf252U,
		0xf4afU,0xf955U,0xe2a1U,0xd549U,0xba99U,0x6539U,0xca72U,0x84efU,0x19d5U,0x33aaU,
		0x6754U,0xcea8U,0x8d5bU,0x0abdU,0x157aU,0x2af4U,0x55e8U,0xabd0U,0x47abU,0x8f56U,
		0x0ea7U,0x1d4eU,0x3a9cU,0x7538U,0xea70U,0xc4ebU,0x99ddU,0x23b1U,0x4762U,0x8ec4U,
		0x0d83U,0x1b06U,0x360cU,0x6c18U,0xd830U,0xa06bU,0x50ddU,0xa1baU,0x537fU,0xa6feU,
		0x5df7U,0xbbeeU,0x67d7U,0xcfaeU,0x8f57U,0x0ea5U,0x1d4aU,0x3a94U,0x7528U,0xea50U,
		0xc4abU,0x995dU,0x22b1U,0x4562U,0x8ac4U,0x0583U,0x0b06U,0x160cU,0x2c18U,0x5830U,
		0xb060U,0x70cbU,0xe196U,0xd327U,0xb645U,0x7c81U,0xf902U,0xe20fU,0xd415U,0xb821U,
		0x6049U,0xc092U,0x912fU,0x3255U,0x64aaU,0xc954U,0x82a3U,0x154dU,0x2a9aU,0x5534U,
		0xaa68U,0x44dbU,0x89b6U,0x0367U,0x06ceU,0x0d9cU,0x1b38U,0x3670U,0x6ce0U,0xd9c0U,
		0xa38bU,0x571dU,0xae3aU,0x4c7fU,0x98feU,0x21f7U,0x43eeU,0x87dcU,0x1fb3U,0x3f66U,
		0x7eccU,0xfd98U,0xeb3bU,0xc67dU,0x9cf1U,0x29e9U,0x53d2U,0xa7a4U,0x5f43U,0xbe86U,
		0x6d07U,0xda0eU,0xa417U,0x5825U,0xb04aU,0x709fU,0xe13eU,0xd277U,0xb4e5U,0x79c1U,
		0xf382U,0xf70fU,0xfe15U,0xec21U,0xc849U,0x8099U,0x1139U,0x2272U,0x44e4U,0x89c8U,
		0x039bU,0x0736U,0x0e6cU,0x1cd8U,0x39b0U,0x7360U,0xe6c0U,0xdd8bU,0xab1dU,0x4631U,
		0x8c62U,0x08cfU,0x119eU,0x233cU,0x4678U,0x8cf0U,0x09ebU,0x13d6U,0x27acU,0x4f58U,
		0x9eb0U,0x2d6bU,0x5ad6U,0xb5acU,0x7b53U,0xf6a6U,0xfd47U,0xea85U,0xc501U,0x9a09U,
		0x2419U,0x4832U,0x9064U,0x30c3U,0x6186U,0xc30cU,0x9613U,0x3c2dU,0x785aU,0xf0b4U,
		0xf163U,0xf2cdU,0xf591U,0xfb29U,0xe659U,0xdcb9U,0xa979U,0x42f9U,0x85f2U,0x1befU,
		0x37deU,0x6fbcU,0xdf78U,0xaefbU,0x4dfdU,0x9bfaU,0x27ffU,0x4ffeU,0x9ffcU,0x2ff3U,
		0x5fe6U,0xbfccU,0x6f93U,0xdf26U,0xae47U,0x4c85U,0x990aU,0x221fU,0x443eU,0x887cU,
		0x00f3U,0x01e6U,0x03ccU,0x0798U,0x0f30U,0x1e60U,0x3cc0U,0x7980U,0xf300U,0xf60bU,
		0xfc1dU,0xe831U,0xc069U,0x90d9U,0x31b9U,0x6372U,0xc6e4U,0x9dc3U,0x2b8dU,0x571aU,
		0xae34U,0x4c63U,0x98c6U,0x2187U,0x430eU,0x861cU,0x1c33U,0x3866U,0x70ccU,0xe198U,
		0xd33bU,0xb67dU,0x7cf1U,0xf9e2U,0xe3cfU,0xd795U,0xbf21U,0x6e49U,0xdc92U,0xa92fU,
		0x4255U,0x84aaU,0x195fU,0x32beU,0x657cU,0xcaf8U,0x85fbU,0x1bfdU,0x37faU,0x6ff4U,
		0xdfe8U,0xafdbU,0x4fbdU,0x9f7aU,0x2effU,0x5dfeU,0xbbfcU,0x67f3U,0xcfe6U,0x8fc7U,
		0x0f85U,0x1f0aU,0x3e14U,0x7c28U,0xf850U,0xe0abU,0xd15dU,0xb2b1U,0x7569U,0xead2U,
		0xc5afU,0x9b55U,0x26a1U,0x4d42U,0x9a84U,0x2503U,0x4a06U,0x940cU,0x3813U,0x7026U,
		0xe04cU,0xd093U,0xb12dU,0x7251U,0xe4a2U,0xd94fU,0xa295U,0x5521U,0xaa42U,0x448fU,
		0x891eU,0x0237U,0x046eU,0x08dcU,0x11b8U,0x2370U,0x46e0U,0x8dc0U,0x0b8bU,0x1716U,
		0x2e2cU,0x5c58U,0xb8b0U,0x616bU,0xc2d6U,0x95a7U,0x3b45U,0x768aU,0xed14U,0xca23U,
		0x844dU,0x1891U,0x3122U,0x6244U,0xc488U,0x991bU,0x223dU,0x447aU,0x88f4U,0x01e3U,
		0x03c6U,0x078cU,0x0f18U,0x1e30U,0x3c60U,0x78c0U,0xf180U,0xf30bU,0xf61dU,0xfc31U,
		0xe869U,0xc0d9U,0x91b9U,0x3379U,0x66f2U,0xcde4U,0x8bc3U,0x078dU,0x0f1aU,0x1e34U,
		0x3c68U,0x78d0U,0xf1a0U,0xf34bU,0xf69dU,0xfd31U,0xea69U,0xc4d9U,0x99b9U,0x2379U,
		0x46f2U,0x8de4U,0x0bc3U,0x1786U,0x2f0cU,0x5e18U,0xbc30U,0x686bU,0xd0d6U,0xb1a7U,
		0x7345U,0xe68aU,0xdd1fU,0xaa35U,0x4461U,0x88c2U,0x018fU,0x031eU,0x063cU,0x0c78U,
		0x18f0U,0x31e0U,0x63c0U,0xc780U,0x9f0bU,0x2e1dU,0x5c3aU,0xb874U,0x60e3U,0xc1c6U,
		0x9387U,0x3705U,0x6e0aU,0xdc14U,0xa823U,0x404dU,0x809aU,0x113fU,0x227eU,0x44fcU,
		0x89f8U,0x03fbU,0x07f6U,0x0fecU,0x1fd8U,0x3fb0U,0x7f60U,0xfec0U,0xed8bU,0xcb1dU,
		0x8631U,0x1c69U,0x38d2U,0x71a4U,0xe348U,0xd69bU,0xbd3dU,0x6a71U,0xd4e2U,0xb9cfU,
		0x6395U,0xc72aU,0x9e5fU,0x2cb5U,0x596aU,0xb2d4U,0x75a3U,0xeb46U,0xc687U,0x9d05U,
		0x2a01U,0x5402U,0xa804U,0x4003U,0x8006U,0x1007U,0x200eU,0x401cU,0x8038U,0x107bU,
		0x20f6U,0x41ecU,0x83d8U,0x17bbU,0x2f76U,0x5eecU,0xbdd8U,0x6bbbU,0xd776U,0xbee7U,
		0x6dc5U,0xdb8aU,0xa71fU,0x5e35U,0xbc6aU,0x68dfU,0xd1beU,0xb377U,0x76e5U,0xedcaU,
		0xcb9fU,0x8735U,0x1e61U,0x3cc2U,0x7984U,0xf308U,0xf61bU,0xfc3dU,0xe871U,0xc0e9U,
		0x91d9U,0x33b9U,0x6772U,0xcee4U,0x8dc3U,0x0b8dU,0x171aU,0x2e34U,0x5c68U,0xb8d0U,
		0x61abU,0xc356U,0x96a7U,0x3d45U,0x7a8aU,0xf514U,0xfa23U,0xe44dU,0xd891U,0xa129U,
		0x5259U,0xa4b2U,0x596fU,0xb2deU,0x75b7U,0xeb6eU,0xc6d7U,0x9da5U,0x2b41U,0x5682U,
		0xad04U,0x4a03U,0x9406U,0x3807U,0x700eU,0xe01cU,0xd033U,0xb06dU,0x70d1U,0xe1a2U,
		0xd34fU,0xb695U,0x7d21U,0xfa42U,0xe48fU,0xd915U,0xa221U,0x5449U,0xa892U,0x412fU,
		0x825eU,0x14b7U,0x296eU,0x52dcU,0xa5b8U,0x5b7bU,0xb6f6U,0x7de7U,0xfbceU,0xe797U,
		0xdf25U,0xae41U,0x4c89U,0x9912U,0x222fU,0x445eU,0x88bcU,0x0173U,0x02e6U,0x05ccU,
		0x0b98U,0x1730U,0x2e60U,0x5cc0U,0xb980U,0x630bU,0xc616U,0x9c27U,0x2845U,0x508aU,
		0xa114U,0x5223U,0xa446U,0x5887U,0xb10eU,0x7217U,0xe42eU,0xd857U,0xa0a5U,0x5141U,
		0xa282U,0x550fU,0xaa1eU,0x4437U,0x886eU,0x00d7U,0x01aeU,0x035cU,0x06b8U,0x0d70U,
		0x1ae0U,0x35c0U,0x6b80U,0xd700U,0xbe0bU,0x6c1dU,0xd83aU,0xa07fU,0x50f5U,0xa1eaU,
		0x53dfU,0xa7beU,0x5f77U,0xbeeeU,0x6dd7U,0xdbaeU,0xa757U,0x5ea5U,0xbd4aU,0x6a9fU,
		0xd53eU,0xba77U,0x64e5U,0xc9caU,0x839fU,0x1735U,0x2e6aU,0x5cd4U,0xb9a8U,0x635bU,
		0xc6b6U,0x9d67U,0x2ac5U,0x558aU,0xab14U,0x4623U,0x8c46U,0x0887U,0x110eU,0x221cU,
		0x4438U,0x8870U,0x00ebU,0x01d6U,0x03acU,0x0758U,0x0eb0U,0x1d60U,0x3ac0U,0x7580U,
		0xeb00U,0xc60bU,0x9c1dU,0x2831U,0x5062U,0xa0c4U,0x5183U,0xa306U,0x5607U,0xac0eU,
		0x4817U,0x902eU,0x3057U,0x60aeU,0xc15cU,0x92b3U,0x356dU,0x6adaU,0xd5b4U,0xbb63U,
		0x66cdU,0xcd9aU,0x8b3fU,0x0675U,0x0ceaU,0x19d4U,0x33a8U,0x6750U,0xcea0U,0x8d4bU,
		0x0a9dU,0x153aU,0x2a74U,0x54e8U,0xa9d0U,0x43abU,0x8756U,0x1ea7U,0x3d4eU,0x7a9cU,
		0xf538U,0xfa7bU,0xe4fdU,0xd9f1U,0xa3e9U,0x57d9U,0xafb2U,0x4f6fU,0x9edeU,0x2db7U,
		0x5b6eU,0xb6dcU,0x7db3U,0xfb66U,0xe6c7U,0xdd85U,0xab01U,0x4609U,0x8c12U,0x082fU,
		0x105eU,0x20bcU,0x4178U,0x82f0U,0x15ebU,0x2bd6U,0x57acU,0xaf58U,0x4ebbU,0x9d76U,
		0x2ae7U,0x55ceU,0xab9cU,0x4733U,0x8e66U,0x0cc7U,0x198eU,0x331cU,0x6638U,0xcc70U,
		0x88ebU,0x01ddU,0x03baU,0x0774U,0x0ee8U,0x1dd0U,0x3ba0U,0x7740U,0xee80U,0xcd0bU,
		0x8a1dU,0x0431U,0x0862U,0x10c4U,0x2188U,0x4310U,0x8620U,0x1c4bU,0x3896U,0x712cU,
		0xe258U,0xd4bbU,0xb97dU,0x62f1U,0xc5e2U,0x9bcfU,0x2795U,0x4f2aU,0x9e54U,0x2ca3U,
		0x5946U,0xb28cU,0x7513U,0xea26U,0xc447U,0x9885U,0x2101U,0x4202U,0x8404U,0x1803U,
		0x3006U,0x600cU,0xc018U,0x903bU,0x307dU,0x60faU,0xc1f4U,0x93e3U,0x37cdU,0x6f9aU,
		0xdf34U,0xae63U,0x4ccdU,0x999aU,0x233fU,0x467eU,0x8cfcU,0x09f3U,0x13e6U,0x27ccU,
		0x4f98U,0x9f30U,0x2e6bU,0x5cd6U,0xb9acU,0x6353U,0xc6a6U,0x9d47U,0x2a85U,0x550aU,
		0xaa14U,0x4423U,0x8846U,0x0087U,0x010eU,0x021cU,0x0438U,0x0870U,0x10e0U,0x21c0U,
		0x4380U,0x8700U,0x1e0bU,0x3c16U,0x782cU,0xf058U,0xf0bbU,0xf17dU,0xf2f1U,0xf5e9U,
		0xfbd9U,0xe7b9U,0xdf79U,0xaef9U,0x4df9U,0x9bf2U,0x27efU,0x4fdeU,0x9fbcU,0x2f73U,
		0x5ee6U,0xbdccU,0x6b93U,0xd726U,0xbe47U,0x6c85U,0xd90aU,0xa21fU,0x5435U,0xa86aU,
		0x40dfU,0x81beU,0x1377U,0x26eeU,0x4ddcU,0x9bb8U,0x277bU,0x4ef6U,0x9decU,0x2bd3U,
		0x57a6U,0xaf4cU,0x4e93U,0x9d26U,0x2a47U,0x548eU,0xa91cU,0x4233U,0x8466U,0x18c7U,
		0x318eU,0x631cU,0xc638U,0x9c7bU,0x28fdU,0x51faU,0xa3f4U,0x57e3U,0xafc6U,0x4f87U,
		0x9f0eU,0x2e17U,0x5c2eU,0xb85cU,0x60b3U,0xc166U,0x92c7U,0x3585U,0x6b0aU,0xd614U,
		0xbc23U,0x684dU,0xd09aU,0xb13fU,0x7275U,0xe4eaU,0xd9dfU,0xa3b5U,0x5761U,0xaec2U,
		0x4d8fU,0x9b1eU,0x2637U,0x4c6eU,0x98dcU,0x21b3U,0x4366U,0x86ccU,0x1d93U,0x3b26U,
		0x764cU,0xec98U,0xc93bU,0x827dU,0x14f1U,0x29e2U,0x53c4U,0xa788U,0x5f1bU,0xbe36U,
		0x6c67U,0xd8ceU,0xa197U,0x5325U,0xa64aU,0x5c9fU,0xb93eU,0x6277U,0xc4eeU,0x99d7U,
		0x23a5U,0x474aU,0x8e94U,0x0d23U,0x1a46U,0x348cU,0x6918U,0xd230U,0xb46bU,0x78ddU,
		0xf1baU,0xf37fU,0xf6f5U,0xfde1U,0xebc9U,0xc799U,0x9f39U,0x2e79U,0x5cf2U,0xb9e4U,
		0x63c3U,0xc786U,0x9f07U,0x2e05U,0x5c0aU,0xb814U,0x6023U,0xc046U,0x9087U,0x3105U,
		0x620aU,0xc414U,0x9823U,0x204dU,0x409aU,0x8134U,0x1263U,0x24c6U,0x498cU,0x9318U,
		0x363bU,0x6c76U,0xd8ecU,0xa1d3U,0x53adU,0xa75aU,0x5ebfU,0xbd7eU,0x6af7U,0xd5eeU,
		0xbbd7U,0x67a5U,0xcf4aU,0x8e9fU,0x0d35U,0x1a6aU,0x34d4U,0x69a8U,0xd350U,0xb6abU,
		0x7d5dU,0xfabaU,0xe57fU,0xdaf5U,0xa5e1U,0x5bc9U,0xb792U,0x7f2fU,0xfe5eU,0xecb7U,
		0xc965U,0x82c1U,0x1589U,0x2b12U,0x5624U,0xac48U,0x489bU,0x9136U,0x3267U,0x64ceU,
		0xc99cU,0x8333U,0x166dU,0x2cdaU,0x59b4U,0xb368U,0x76dbU,0xedb6U,0xcb67U,0x86c5U,
		0x1d81U,0x3b02U,0x7604U,0xec08U,0xc81bU,0x803dU,0x1071U,0x20e2U,0x41c4U,0x8388U,
		0x171bU,0x2e36U,0x5c6cU,0xb8d8U,0x61bbU,0xc376U,0x96e7U,0x3dc5U,0x7b8aU,0xf714U,
		0xfe23U,0xec4dU,0xc891U,0x8129U,0x1259U,0x24b2U,0x4964U,0x92c8U,0x359bU,0x6b36U,
		0xd66cU,0xbcd3U,0x69adU,0xd35aU,0xb6bfU,0x7d75U,0xfaeaU,0xe5dfU,0xdbb5U,0xa761U,
		0x5ec9U,0xbd92U,0x6b2fU,0xd65eU,0xbcb7U,0x6965U,0xd2caU,0xb59fU,0x7b35U,0xf66aU,
		0xfcdfU,0xe9b5U,0xc361U,0x96c9U,0x3d99U,0x7b32U,0xf664U,0xfcc3U,0xe98dU,0xc311U,
		0x9629U,0x3c59U,0x78b2U,0xf164U,0xf2c3U,0xf58dU,0xfb11U,0xe629U,0xdc59U,0xa8b9U,
		0x4179U,0x82f2U,0x15efU,0x2bdeU,0x57bcU,0xaf78U,0x4efbU,0x9df6U,0x2be7U,0x57ceU,
		0xaf9cU,0x4f33U,0x9e66U,0x2cc7U,0x598eU,0xb31cU,0x7633U,0xec66U,0xc8c7U,0x8185U,
		0x1301U,0x2602U,0x4c04U,0x9808U,0x201bU,0x4036U,0x806cU,0x10d3U,0x21a6U,0x434cU,
		0x8698U,0x1d3bU,0x3a76U,0x74ecU,0xe9d8U,0xc3bbU,0x977dU,0x3ef1U,0x7de2U,0xfbc4U,
		0xe783U,0xdf0dU,0xae11U,0x4c29U,0x9852U,0x20afU,0x415eU,0x82bcU,0x1573U,0x2ae6U,
		0x55ccU,0xab98U,0x473bU,0x8e76U,0x0ce7U,0x19ceU,0x339cU,0x6738U,0xce70U,0x8cebU,
		0x09ddU,0x13baU,0x2774U,0x4ee8U,0x9dd0U,0x2babU,0x5756U,0xaeacU,0x4d53U,0x9aa6U,
		0x2547U,0x4a8eU,0x951cU,0x3a33U,0x7466U,0xe8ccU,0xc193U,0x932dU,0x3651U,0x6ca2U,
		0xd944U,0xa283U,0x550dU,0xaa1aU,0x443fU,0x887eU,0x00f7U,0x01eeU,0x03dcU,0x07b8U,
		0x0f70U,0x1ee0U,0x3dc0U,0x7b80U,0xf700U,0xfe0bU,0xec1dU,0xc831U,0x8069U,0x10d9U,
		0x21b2U,0x4364U,0x86c8U,0x1d9bU,0x3b36U,0x766cU,0xecd8U,0xc9bbU,0x837dU,0x16f1U,
		0x2de2U,0x5bc4U,0xb788U,0x7f1bU,0xfe36U,0xec67U,0xc8c5U,0x8181U,0x1309U,0x2612U,
		0x4c24U,0x9848U,0x209bU,0x4136U,0x826cU,0x14d3U,0x29a6U,0x534cU,0xa698U,0x5d3bU,
		0xba76U,0x64e7U,0xc9ceU,0x8397U,0x1725U,0x2e4aU,0x5c94U,0xb928U,0x625bU,0xc4b6U,
		0x9967U,0x22c5U,0x458aU,0x8b14U,0x0623U,0x0c46U,0x188cU,0x3118U,0x6230U,0xc460U,
		0x98cbU,0x219dU,0x433aU,0x8674U,0x1ce3U,0x39c6U,0x738cU,0xe718U,0xde3bU,0xac7dU,
		0x48f1U,0x91e2U,0x33cfU,0x679eU,0xcf3cU,0x8e73U,0x0cedU,0x19daU,0x33b4U,0x6768U,
		0xced0U,0x8dabU,0x0b5dU,0x16baU,0x2d74U,0x5ae8U,0xb5d0U,0x7babU,0xf756U,0xfea7U,
		0xed45U,0xca81U,0x8509U,0x1a19U,0x3432U,0x6864U,0xd0c8U,0xb19bU,0x733dU,0xe67aU,
		0xdcffU,0xa9f5U,0x43e1U,0x87c2U,0x1f8fU,0x3f1eU,0x7e3cU,0xfc78U,0xe8fbU,0xc1fdU,
		0x93f1U,0x37e9U,0x6fd2U,0xdfa4U,0xaf43U,0x4e8dU,0x9d1aU,0x2a3fU,0x547eU,0xa8fcU,
		0x41f3U,0x83e6U,0x17c7U,0x2f8eU,0x5f1cU,0xbe38U,0x6c7bU,0xd8f6U,0xa1e7U,0x53c5U,
		0xa78aU,0x5f1fU,0xbe3eU,0x6c77U,0xd8eeU,0xa1d7U,0x53a5U,0xa74aU,0x5e9fU,0xbd3eU,
		0x6a77U,0xd4eeU,0xb9d7U,0x63a5U,0xc74aU,0x9e9fU,0x2d35U,0x5a6aU,0xb4d4U,0x79a3U,
		0xf346U,0xf687U,0xfd05U,0xea01U,0xc409U,0x9819U,0x2039U,0x4072U,0x80e4U,0x11c3U,
		0x2386U,0x470cU,0x8e18U,0x0c3bU,0x1876U,0x30ecU,0x61d8U,0xc3b0U,0x976bU,0x3eddU,
		0x7dbaU,0xfb74U,0xe6e3U,0xddcdU,0xab91U,0x4729U,0x8e52U,0x0cafU,0x195eU,0x32bcU,
		0x6578U,0xcaf0U,0x85ebU,0x1bddU,0x37baU,0x6f74U,0xdee8U,0xaddbU,0x4bbdU,0x977aU,
		0x3effU,0x7dfeU,0xfbfcU,0xe7f3U,0xdfedU,0xafd1U,0x4fa9U,0x9f52U,0x2eafU,0x5d5eU,
		0xbabcU,0x6573U,0xcae6U,0x85c7U,0x1b85U,0x370aU,0x6e14U,0xdc28U,0xa85bU,0x40bdU,
		0x817aU,0x12ffU,0x25feU,0x4bfcU,0x97f8U,0x3ffbU,0x7ff6U,0xffecU,0xefd3U,0xcfadU,
		0x8f51U,0x0ea9U,0x1d52U,0x3aa4U,0x7548U,0xea90U,0xc52bU,0x9a5dU,0x24b1U,0x4962U,
		0x92c4U,0x3583U,0x6b06U,0xd60cU,0xbc13U,0x682dU,0xd05aU,0xb0bfU,0x7175U,0xe2eaU,
		0xd5dfU,0xbbb5U,0x6761U,0xcec2U,0x8d8fU,0x0b15U,0x162aU,0x2c54U,0x58a8U,0xb150U,
		0x72abU,0xe556U,0xdaa7U,0xa545U,0x5a81U,0xb502U,0x7a0fU,0xf41eU,0xf837U,0xe065U,
		0xd0c1U,0xb189U,0x7319U,0xe632U,0xdc6fU,0xa8d5U,0x41a1U,0x8342U,0x168fU,0x2d1eU,
		0x5a3cU,0xb478U,0x78fbU,0xf1f6U,0xf3e7U,0xf7c5U,0xff81U,0xef09U,0xce19U,0x8c39U,
		0x0879U,0x10f2U,0x21e4U,0x43c8U,0x8790U,0x1f2bU,0x3e56U,0x7cacU,0xf958U,0xe2bbU,
		0xd57dU,0xbaf1U,0x65e9U,0xcbd2U,0x87afU,0x1f55U,0x3eaaU,0x7d54U,0xfaa8U,0xe55bU,
		0xdabdU,0xa571U,0x5ae9U,0xb5d2U,0x7bafU,0xf75eU,0xfeb7U,0xed65U,0xcac1U,0x8589U,
		0x1b19U,0x3632U,0x6c64U,0xd8c8U,0xa19bU,0x533dU,0xa67aU,0x5cffU,0xb9feU,0x63f7U,
		0xc7eeU,0x9fd7U,0x2fa5U,0x5f4aU,0xbe94U,0x6d23U,0xda46U,0xa487U,0x5905U,0xb20aU,
		0x741fU,0xe83eU,0xc077U,0x90e5U,0x31c1U,0x6382U,0xc704U,0x9e03U,0x2c0dU,0x581aU,
		0xb034U,0x7063U,0xe0c6U,0xd187U,0xb305U,0x7601U,0xec02U,0xc80fU,0x8015U,0x1021U,
		0x2042U,0x4084U,0x8108U,0x121bU,0x2436U,0x486cU,0x90d8U,0x31bbU,0x6376U,0xc6ecU,
		0x9dd3U,0x2badU,0x575aU,0xaeb4U,0x4d63U,0x9ac6U,0x2587U,0x4b0eU,0x961cU,0x3c33U,
		0x7866U,0xf0ccU,0xf193U,0xf32dU,0xf651U,0xfca9U,0xe959U,0xc2b9U,0x9579U,0x3af9U,
		0x75f2U,0xebe4U,0xc7c3U,0x9f8dU,0x2f11U,0x5e22U,0xbc44U,0x6883U,0xd106U,0xb207U,
		0x7405U,0xe80aU,0xc01fU,0x9035U,0x3061U,0x60c2U,0xc184U,0x9303U,0x360dU,0x6c1aU,
		0xd834U,0xa063U,0x50cdU,0xa19aU,0x533fU,0xa67eU,0x5cf7U,0xb9eeU,0x63d7U,0xc7aeU,
		0x9f57U,0x2ea5U,0x5d4aU,0xba94U,0x6523U,0xca46U,0x8487U,0x1905U,0x320aU,0x6414U,
		0xc828U,0x805bU,0x10bdU,0x217aU,0x42f4U,0x85e8U,0x1bdbU,0x37b6U,0x6f6cU,0xded8U,
		0xadbbU,0x4b7dU,0x96faU,0x3dffU,0x7bfeU,0xf7fcU,0xfff3U,0xefedU,0xcfd1U,0x8fa9U,
		0x0f59U,0x1eb2U,0x3d64U,0x7ac8U,0xf590U,0xfb2bU,0xe65dU,0xdcb1U,0xa969U,0x42d9U,
		0x85b2U,0x1b6fU,0x36deU,0x6dbcU,0xdb78U,0xa6fbU,0x5dfdU,0xbbfaU,0x67ffU,0xcffeU,
		0x8ff7U,0x0fe5U,0x1fcaU,0x3f94U,0x7f28U,0xfe50U,0xecabU,0xc95dU,0x82b1U,0x1569U,
		0x2ad2U,0x55a4U,0xab48U,0x469bU,0x8d36U,0x0a67U,0x14ceU,0x299cU,0x5338U,0xa670U,
		0x5cebU,0xb9d6U,0x63a7U,0xc74eU,0x9e97U,0x2d25U,0x5a4aU,0xb494U,0x7923U,0xf246U,
		0xf487U,0xf905U,0xe201U,0xd409U,0xb819U,0x6039U,0xc072U,0x90efU,0x31d5U,0x63aaU,
		0xc754U,0x9ea3U,0x2d4dU,0x5a9aU,0xb534U,0x7a63U,0xf4c6U,0xf987U,0xe305U,0xd601U,
		0xbc09U,0x6819U,0xd032U,0xb06fU,0x70d5U,0xe1aaU,0xd35fU,0xb6b5U,0x7d61U,0xfac2U,
		0xe58fU,0xdb15U,0xa621U,0x5c49U,0xb892U,0x612fU,0xc25eU,0x94b7U,0x3965U,0x72caU,
		0xe594U,0xdb23U,0xa64dU,0x5c91U,0xb922U,0x624fU,0xc49eU,0x9937U,0x2265U,0x44caU,
		0x8994U,0x0323U,0x0646U,0x0c8cU,0x1918U,0x3230U,0x6460U,0xc8c0U,0x818bU,0x131dU,
		0x263aU,0x4c74U,0x98e8U,0x21dbU,0x43b6U,0x876cU,0x1ed3U,0x3da6U,0x7b4cU,0xf698U,
		0xfd3bU,0xea7dU,0xc4f1U,0x99e9U,0x23d9U,0x47b2U,0x8f64U,0x0ec3U,0x1d86U,0x3b0cU,
		0x7618U,0xec30U,0xc86bU,0x80ddU,0x11b1U,0x2362U,0x46c4U,0x8d88U,0x0b1bU,0x1636U,
		0x2c6cU,0x58d8U,0xb1b0U,0x736bU,0xe6d6U,0xdda7U,0xab45U,0x4681U,0x8d02U,0x0a0fU,
		0x141eU,0x283cU,0x5078U,0xa0f0U,0x51ebU,0xa3d6U,0x57a7U,0xaf4eU,0x4e97U,0x9d2eU,
		0x2a57U,0x54aeU,0xa95cU,0x42b3U,0x8566U,0x1ac7U,0x358eU,0x6b1cU,0xd638U,0xbc7bU,
		0x68fdU,0xd1faU,0xb3ffU,0x77f5U,0xefeaU,0xcfdfU,0x8fb5U,0x0f61U,0x1ec2U,0x3d84U,
		0x7b08U,0xf610U,0xfc2bU,0xe85dU,0xc0b1U,0x9169U,0x32d9U,0x65b2U,0xcb64U,0x86c3U,
		0x1d8dU,0x3b1aU,0x7634U,0xec68U,0xc8dbU,0x81bdU,0x1371U,0x26e2U,0x4dc4U,0x9b88U,
		0x271bU,0x4e36U,0x9c6cU,0x28d3U,0x51a6U,0xa34cU,0x5693U,0xad26U,0x4a47U,0x948eU,
		0x3917U,0x722eU,0xe45cU,0xd8b3U,0xa16dU,0x52d1U,0xa5a2U,0x5b4fU,0xb69eU,0x7d37U,
		0xfa6eU,0xe4d7U,0xd9a5U,0xa341U,0x5689U,0xad12U,0x4a2fU,0x945eU,0x38b7U,0x716eU,
		0xe2dcU,0xd5b3U,0xbb6dU,0x66d1U,0xcda2U,0x8b4fU,0x0695U,0x0d2aU,0x1a54U,0x34a8U,
		0x6950U,0xd2a0U,0xb54bU,0x7a9dU,0xf53aU,0xfa7fU,0xe4f5U,0xd9e1U,0xa3c9U,0x5799U,
		0xaf32U,0x4e6fU,0x9cdeU,0x29b7U,0x536eU,0xa6dcU,0x5db3U,0xbb66U,0x66c7U,0xcd8eU,
		0x8b17U,0x0625U,0x0c4aU,0x1894U,0x3128U,0x6250U,0xc4a0U,0x994bU,0x229dU,0x453aU,
		0x8a74U,0x04e3U,0x09c6U,0x138cU,0x2718U,0x4e30U,0x9c60U,0x28cbU,0x5196U,0xa32cU,
		0x5653U,0xaca6U,0x4947U,0x928eU,0x3517U,0x6a2eU,0xd45cU,0xb8b3U,0x616dU,0xc2daU,
		0x95bfU,0x3b75U,0x76eaU,0xedd4U,0xcba3U,0x874dU,0x1e91U,0x3d22U,0x7a44U,0xf488U,
		0xf91bU,0xe23dU,0xd471U,0xb8e9U,0x61d9U,0xc3b2U,0x976fU,0x3ed5U,0x7daaU,0xfb54U,
		0xe6a3U,0xdd4dU,0xaa91U,0x4529U,0x8a52U,0x04afU,0x095eU,0x12bcU,0x2578U,0x4af0U,
		0x95e0U,0x3bcbU,0x7796U,0xef2cU,0xce53U,0x8cadU,0x0951U,0x12a2U,0x2544U,0x4a88U,
		0x9510U,0x3a2bU,0x7456U,0xe8acU,0xc153U,0x92adU,0x3551U,0x6aa2U,0xd544U,0xba83U,
		0x650dU,0xca1aU,0x843fU,0x1875U,0x30eaU,0x61d4U,0xc3a8U,0x975bU,0x3ebdU,0x7d7aU,
		0xfaf4U,0xe5e3U,0xdbcdU,0xa791U,0x5f29U,0xbe52U,0x6cafU,0xd95eU,0xa2b7U,0x5565U,
		0xaacaU,0x459fU,0x8b3eU,0x0677U,0x0ceeU,0x19dcU,0x33b8U,0x6770U,0xcee0U,0x8dcbU,
		0x0b9dU,0x173aU,0x2e74U,0x5ce8U,0xb9d0U,0x63abU,0xc756U,0x9ea7U,0x2d45U,0x5a8aU,
		0xb514U,0x7a23U,0xf446U,0xf887U,0xe105U,0xd201U,0xb409U,0x7819U,0xf032U,0xf06fU,
		0xf0d5U,0xf1a1U,0xf349U,0xf699U,0xfd39U,0xea79U,0xc4f9U,0x99f9U,0x23f9U,0x47f2U,
		0x8fe4U,0x0fc3U,0x1f86U,0x3f0cU,0x7e18U,0xfc30U,0xe86bU,0xc0ddU,0x91b1U,0x3369U,
		0x66d2U,0xcda4U,0x8b43U,0x068dU,0x0d1aU,0x1a34U,0x3468U,0x68d0U,0xd1a0U,0xb34bU,
		0x769dU,0xed3aU,0xca7fU,0x84f5U,0x19e1U,0x33c2U,0x6784U,0xcf08U,0x8e1bU,0x0c3dU,
		0x187aU,0x30f4U,0x61e8U,0xc3d0U,0x97abU,0x3f5dU,0x7ebaU,0xfd74U,0xeae3U,0xc5cdU,
		0x9b91U,0x2729U,0x4e52U,0x9ca4U,0x2943U,0x5286U,0xa50cU,0x5a13U,0xb426U,0x7847U,
		0xf08eU,0xf117U,0xf225U,0xf441U,0xf889U,0xe119U,0xd239U,0xb479U,0x78f9U,0xf1f2U,
		0xf3efU,0xf7d5U,0xffa1U,0xef49U,0xce99U,0x8d39U,0x0a79U,0x14f2U,0x29e4U,0x53c8U,
		0xa790U,0x5f2bU,0xbe56U,0x6ca7U,0xd94eU,0xa297U,0x5525U,0xaa4aU,0x449fU,0x893eU,
		0x0277U,0x04eeU,0x09dcU,0x13b8U,0x2770U,0x4ee0U,0x9dc0U,0x2b8bU,0x5716U,0xae2cU,
		0x4c53U,0x98a6U,0x2147U,0x428eU,0x851cU,0x1a33U,0x3466U,0x68ccU,0xd198U,0xb33bU,
		0x767dU,0xecfaU,0xc9ffU,0x83f5U,0x17e1U,0x2fc2U,0x5f84U,0xbf08U,0x6e1bU,0xdc36U,
		0xa867U,0x40c5U,0x818aU,0x131fU,0x263eU,0x4c7cU,0x98f8U,0x21fbU,0x43f6U,0x87ecU,
		0x1fd3U,0x3fa6U,0x7f4cU,0xfe98U,0xed3bU,0xca7dU,0x84f1U,0x19e9U,0x33d2U,0x67a4U,
		0xcf48U,0x8e9bU,0x0d3dU,0x1a7aU,0x34f4U,0x69e8U,0xd3d0U,0xb7abU,0x7f5dU,0xfebaU,
		0xed7fU,0xcaf5U,0x85e1U,0x1bc9U,0x3792U,0x6f24U,0xde48U,0xac9bU,0x493dU,0x927aU,
		0x34ffU,0x69feU,0xd3fcU,0xb7f3U,0x7fedU,0xffdaU,0xefbfU,0xcf75U,0x8ee1U,0x0dc9U,
		0x1b92U,0x3724U,0x6e48U,0xdc90U,0xa92bU,0x425dU,0x84baU,0x197fU,0x32feU,0x65fcU,
		0xcbf8U,0x87fbU,0x1ffdU,0x3ffaU,0x7ff4U,0xffe8U,0xefdbU,0xcfbdU,0x8f71U,0x0ee9U,
		0x1dd2U,0x3ba4U,0x7748U,0xee90U,0xcd2bU,0x8a5dU,0x04b1U,0x0962U,0x12c4U,0x2588U,
		0x4b10U,0x9620U,0x3c4bU,0x7896U,0xf12cU,0xf253U,0xf4adU,0xf951U,0xe2a9U,0xd559U,
		0xbab9U,0x6579U,0xcaf2U,0x85efU,0x1bd5U,0x37aaU,0x6f54U,0xdea8U,0xad5bU,0x4abdU,
		0x957aU,0x3affU,0x75feU,0xebfcU,0xc7f3U,0x9fedU,0x2fd1U,0x5fa2U,0xbf44U,0x6e83U,
		0xdd06U,0xaa07U,0x4405U,0x880aU,0x001fU,0x003eU,0x007cU,0x00f8U,0x01f0U,0x03e0U,
		0x07c0U,0x0f80U,0x1f00U,0x3e00U,0x7c00U,0xf800U,0xe00bU,0xd01dU,0xb031U,0x7069U,
		0xe0d2U,0xd1afU,0xb355U,0x76a1U,0xed42U,0xca8fU,0x8515U,0x1a21U,0x3442U,0x6884U,
		0xd108U,0xb21bU,0x743dU,0xe87aU,0xc0ffU,0x91f5U,0x33e1U,0x67c2U,0xcf84U,0x8f03U,
		0x0e0dU,0x1c1aU,0x3834U,0x7068U,0xe0d0U,0xd1abU,0xb35dU,0x76b1U,0xed62U,0xcacfU,
		0x8595U,0x1b21U,0x3642U,0x6c84U,0xd908U,0xa21bU,0x543dU,0xa87aU,0x40ffU,0x81feU,
		0x13f7U,0x27eeU,0x4fdcU,0x9fb8U,0x2f7bU,0x5ef6U,0xbdecU,0x6bd3U,0xd7a6U,0xbf47U,
		0x6e85U,0xdd0aU,0xaa1fU,0x4435U,0x886aU,0x00dfU,0x01beU,0x037cU,0x06f8U,0x0df0U,
		0x1be0U,0x37c0U,0x6f80U,0xdf00U,0xae0bU,0x4c1dU,0x983aU,0x207fU,0x40feU,0x81fcU,
		0x13f3U,0x27e6U,0x4fccU,0x9f98U,0x2f3bU,0x5e76U,0xbcecU,0x69d3U,0xd3a6U,0xb747U,
		0x7e85U,0xfd0aU,0xea1fU,0xc435U,0x9861U,0x20c9U,0x4192U,0x8324U,0x1643U,0x2c86U,
		0x590cU,0xb218U,0x743bU,0xe876U,0xc0e7U,0x91c5U,0x3381U,0x6702U,0xce04U,0x8c03U,
		0x080dU,0x101aU,0x2034U,0x4068U,0x80d0U,0x11abU,0x2356U,0x46acU,0x8d58U,0x0abbU,
		0x1576U,0x2aecU,0x55d8U,0xabb0U,0x476bU,0x8ed6U,0x0da7U,0x1b4eU,0x369cU,0x6d38U,
		0xda70U,0xa4ebU,0x59ddU,0xb3baU,0x777fU,0xeefeU,0xcdf7U,0x8be5U,0x07c1U,0x0f82U,
		0x1f04U,0x3e08U,0x7c10U,0xf820U,0xe04bU,0xd09dU,0xb131U,0x7269U,0xe4d2U,0xd9afU,
		0xa355U,0x56a1U,0xad42U,0x4a8fU,0x951eU,0x3a37U,0x746eU,0xe8dcU,0xc1b3U,0x936dU,
		0x36d1U,0x6da2U,0xdb44U,0xa683U,0x5d0dU,0xba1aU,0x643fU,0xc87eU,0x80f7U,0x11e5U,
		0x23caU,0x4794U,0x8f28U,0x0e5bU,0x1cb6U,0x396cU,0x72d8U,0xe5b0U,0xdb6bU,0xa6ddU,
		0x5db1U,0xbb62U,0x66cfU,0xcd9eU,0x8b37U,0x0665U,0x0ccaU,0x1994U,0x3328U,0x6650U,
		0xcca0U,0x894bU,0x029dU,0x053aU,0x0a74U,0x14e8U,0x29d0U,0x53a0U,0xa740U,0x5e8bU,
		0xbd16U,0x6a27U,0xd44eU,0xb897U,0x6125U,0xc24aU,0x949fU,0x3935U,0x726aU,0xe4d4U,
		0xd9a3U,0xa34dU,0x5691U,0xad22U,0x4a4fU,0x949eU,0x3937U,0x726eU,0xe4dcU,0xd9b3U,
		0xa36dU,0x56d1U,0xada2U,0x4b4fU,0x969eU,0x3d37U,0x7a6eU,0xf4dcU,0xf9b3U,0xe36dU,
		0xd6d1U,0xbda9U,0x6b59U,0xd6b2U,0xbd6fU,0x6ad5U,0xd5aaU,0xbb5fU,0x66b5U,0xcd6aU,
		0x8adfU,0x05b5U,0x0b6aU,0x16d4U,0x2da8U,0x5b50U,0xb6a0U,0x7d4bU,0xfa96U,0xe527U,
		0xda45U,0xa481U,0x5909U,0xb212U,0x742fU,0xe85eU,0xc0b7U,0x9165U,0x32c1U,0x6582U,
		0xcb04U,0x8603U,0x1c0dU,0x381aU,0x7034U,0xe068U,0xd0dbU,0xb1bdU,0x7371U,0xe6e2U,
		0xddcfU,0xab95U,0x4721U,0x8e42U,0x0c8fU,0x191eU,0x323cU,0x6478U,0xc8f0U,0x81ebU,
		0x13ddU,0x27baU,0x4f74U,0x9ee8U,0x2ddbU,0x5bb6U,0xb76cU,0x7ed3U,0xfda6U,0xeb47U,
		0xc685U,0x9d01U,0x2a09U,0x5412U,0xa824U,0x4043U,0x8086U,0x1107U,0x220eU,0x441cU,
		0x8838U,0x007bU,0x00f6U,0x01ecU,0x03d8U,0x07b0U,0x0f60U,0x1ec0U,0x3d80U,0x7b00U,
		0xf600U,0xfc0bU,0xe81dU,0xc031U,0x9069U,0x30d9U,0x61b2U,0xc364U,0x96c3U,0x3d8dU,
		0x7b1aU,0xf634U,0xfc63U,0xe8cdU,0xc191U,0x9329U,0x3659U,0x6cb2U,0xd964U,0xa2c3U,
		0x558dU,0xab1aU,0x463fU,0x8c7eU,0x08f7U,0x11eeU,0x23dcU,0x47b8U,0x8f70U,0x0eebU,
		0x1dd6U,0x3bacU,0x7758U,0xeeb0U,0xcd6bU,0x8addU,0x05b1U,0x0b62U,0x16c4U,0x2d88U,
		0x5b10U,0xb620U,0x7c4bU,0xf896U,0xe127U,0xd245U,0xb481U,0x7909U,0xf212U,0xf42fU,
		0xf855U,0xe0a1U,0xd149U,0xb299U,0x7539U,0xea72U,0xc4efU,0x99d5U,0x23a1U,0x4742U,
		0x8e84U,0x0d03U,0x1a06U,0x340cU,0x6818U,0xd030U,0xb06bU,0x70ddU,0xe1baU,0xd37fU,
		0xb6f5U,0x7de1U,0xfbc2U,0xe78fU,0xdf15U,0xae21U,0x4c49U,0x9892U,0x212fU,0x425eU,
		0x84bcU,0x1973U,0x32e6U,0x65ccU,0xcb98U,0x873bU,0x1e7dU,0x3cfaU,0x79f4U,0xf3e8U,
		0xf7dbU,0xffbdU,0xef71U,0xcee9U,0x8dd9U,0x0bb9U,0x1772U,0x2ee4U,0x5dc8U,0xbb90U,
		0x672bU,0xce56U,0x8ca7U,0x0945U,0x128aU,0x2514U,0x4a28U,0x9450U,0x38abU,0x7156U,
		0xe2acU,0xd553U,0xbaadU,0x6551U,0xcaa2U,0x854fU,0x1a95U,0x352aU,0x6a54U,0xd4a8U,
		0xb95bU,0x62bdU,0xc57aU,0x9affU,0x25f5U,0x4beaU,0x97d4U,0x3fa3U,0x7f46U,0xfe8cU,
		0xed13U,0xca2dU,0x8451U,0x18a9U,0x3152U,0x62a4U,0xc548U,0x9a9bU,0x253dU,0x4a7aU,
		0x94f4U,0x39e3U,0x73c6U,0xe78cU,0xdf13U,0xae2dU,0x4c51U,0x98a2U,0x214fU,0x429eU,
		0x853cU,0x1a73U,0x34e6U,0x69ccU,0xd398U,0xb73bU,0x7e7dU,0xfcfaU,0xe9ffU,0xc3f5U,
		0x97e1U,0x3fc9U,0x7f92U,0xff24U,0xee43U,0xcc8dU,0x8911U,0x0229U,0x0452U,0x08a4U,
		0x1148U,0x2290U,0x4520U,0x8a40U,0x048bU,0x0916U,0x122cU,0x2458U,0x48b0U,0x9160U,
		0x32cbU,0x6596U,0xcb2cU,0x8653U,0x1cadU,0x395aU,0x72b4U,0xe568U,0xdadbU,0xa5bdU,
		0x5b71U,0xb6e2U,0x7dcfU,0xfb9eU,0xe737U,0xde65U,0xacc1U,0x4989U,0x9312U,0x362fU,
		0x6c5eU,0xd8bcU,0xa173U,0x52edU,0xa5daU,0x5bbfU,0xb77eU,0x7ef7U,0xfdeeU,0xebd7U,
		0xc7a5U,0x9f41U,0x2e89U,0x5d12U,0xba24U,0x6443U,0xc886U,0x8107U,0x1205U,0x240aU,
		0x4814U,0x9028U,0x305bU,0x60b6U,0xc16cU,0x92d3U,0x35adU,0x6b5aU,0xd6b4U,0xbd63U,
		0x6acdU,0xd59aU,0xbb3fU,0x6675U,0xcceaU,0x89dfU,0x03b5U,0x076aU,0x0ed4U,0x1da8U,
		0x3b50U,0x76a0U,0xed40U,0xca8bU,0x851dU,0x1a31U,0x3462U,0x68c4U,0xd188U,0xb31bU,
		0x763dU,0xec7aU,0xc8ffU,0x81f5U,0x13e1U,0x27c2U,0x4f84U,0x9f08U,0x2e1bU,0x5c36U,
		0xb86cU,0x60d3U,0xc1a6U,0x9347U,0x3685U,0x6d0aU,0xda14U,0xa423U,0x584dU,0xb09aU,
		0x713fU,0xe27eU,0xd4f7U,0xb9e5U,0x63c1U,0xc782U,0x9f0fU,0x2e15U,0x5c2aU,0xb854U,
		0x60a3U,0xc146U,0x9287U,0x3505U,0x6a0aU,0xd414U,0xb823U,0x604dU,0xc09aU,0x913fU,
		0x3275U,0x64eaU,0xc9d4U,0x83a3U,0x174dU,0x2e9aU,0x5d34U,0xba68U,0x64dbU,0xc9b6U,
		0x8367U,0x16c5U,0x2d8aU,0x5b14U,0xb628U,0x7c5bU,0xf8b6U,0xe167U,0xd2c5U,0xb581U,
		0x7b09U,0xf612U,0xfc2fU,0xe855U,0xc0a1U,0x9149U,0x3299U,0x6532U,0xca64U,0x84c3U,
		0x198dU,0x331aU,0x6634U,0xcc68U,0x88dbU,0x01bdU,0x037aU,0x06f4U,0x0de8U,0x1bd0U,
		0x37a0U,0x6f40U,0xde80U,0xad0bU,0x4a1dU,0x943aU,0x387fU,0x70feU,0xe1fcU,0xd3f3U,
		0xb7edU,0x7fd1U,0xffa2U,0xef4fU,0xce95U,0x8d21U,0x0a49U,0x1492U,0x2924U,0x5248U,
		0xa490U,0x592bU,0xb256U,0x74a7U,0xe94eU,0xc297U,0x9525U,0x3a41U,0x7482U,0xe904U,
		0xc203U,0x940dU,0x3811U,0x7022U,0xe044U,0xd083U,0xb10dU,0x7211U,0xe422U,0xd84fU,
		0xa095U,0x5121U,0xa242U,0x548fU,0xa91eU,0x4237U,0x846eU,0x18d7U,0x31aeU,0x635cU,
		0xc6b8U,0x9d7bU,0x2afdU,0x55faU,0xabf4U,0x47e3U,0x8fc6U,0x0f87U,0x1f0eU,0x3e1cU,
		0x7c38U,0xf870U,0xe0ebU,0xd1ddU,0xb3b1U,0x7769U,0xeed2U,0xcdafU,0x8b55U,0x06a1U,
		0x0d42U,0x1a84U,0x3508U,0x6a10U,0xd420U,0xb84bU,0x609dU,0xc13aU,0x927fU,0x34f5U,
		0x69eaU,0xd3d4U,0xb7a3U,0x7f4dU,0xfe9aU,0xed3fU,0xca75U,0x84e1U,0x19c9U,0x3392U,
		0x6724U,0xce48U,0x8c9bU,0x093dU,0x127aU,0x24f4U,0x49e8U,0x93d0U,0x37abU,0x6f56U,
		0xdeacU,0xad53U,0x4aadU,0x955aU,0x3abfU,0x757eU,0xeafcU,0xc5f3U,0x9bedU,0x27d1U,
		0x4fa2U,0x9f44U,0x2e83U,0x5d06U,0xba0cU,0x6413U,0xc826U,0x8047U,0x1085U,0x210aU,
		0x4214U,0x8428U,0x185bU,0x30b6U,0x616cU,0xc2d8U,0x95bbU,0x3b7dU,0x76faU,0xedf4U,
		0xcbe3U,0x87cdU,0x1f91U,0x3f22U,0x7e44U,0xfc88U,0xe91bU,0xc23dU,0x9471U,0x38e9U,
		0x71d2U,0xe3a4U,0xd743U,0xbe8dU,0x6d11U,0xda22U,0xa44fU,0x5895U,0xb12aU,0x725fU,
		0xe4beU,0xd977U,0xa2e5U,0x55c1U,0xab82U,0x470fU,0x8e1eU,0x0c37U,0x186eU,0x30dcU,
		0x61b8U,0xc370U,0x96ebU,0x3dddU,0x7bbaU,0xf774U,0xfee3U,0xedcdU,0xcb91U,0x8729U,
		0x1e59U,0x3cb2U,0x7964U,0xf2c8U,0xf59bU,0xfb3dU,0xe671U,0xdce9U,0xa9d9U,0x43b9U,
		0x8772U,0x1eefU,0x3ddeU,0x7bbcU,0xf778U,0xfefbU,0xedfdU,0xcbf1U,0x87e9U,0x1fd9U,
		0x3fb2U,0x7f64U,0xfec8U,0xed9bU,0xcb3dU,0x8671U,0x1ce9U,0x39d2U,0x73a4U,0xe748U,
		0xde9bU,0xad3dU,0x4a71U,0x94e2U,0x39cfU,0x739eU,0xe73cU,0xde73U,0xacedU,0x49d1U,
		0x93a2U,0x374fU,0x6e9eU,0xdd3cU,0xaa73U,0x44edU,0x89daU,0x03bfU,0x077eU,0x0efcU,
		0x1df8U,0x3bf0U,0x77e0U,0xefc0U,0xcf8bU,0x8f1dU,0x0e31U,0x1c62U,0x38c4U,0x7188U,
		0xe310U,0xd62bU,0xbc5dU,0x68b1U,0xd162U,0xb2cfU,0x7595U,0xeb2aU,0xc65fU,0x9cb5U,
		0x2961U,0x52c2U,0xa584U,0x5b03U,0xb606U,0x7c07U,0xf80eU,0xe017U,0xd025U,0xb041U,
		0x7089U,0xe112U,0xd22fU,0xb455U,0x78a1U,0xf142U,0xf28fU,0xf515U,0xfa21U,0xe449U,
		0xd899U,0xa139U,0x5279U,0xa4f2U,0x59efU,0xb3deU,0x77b7U,0xef6eU,0xced7U,0x8da5U,
		0x0b41U,0x1682U,0x2d04U,0x5a08U,0xb410U,0x782bU,0xf056U,0xf0a7U,0xf145U,0xf281U,
		0xf509U,0xfa19U,0xe439U,0xd879U,0xa0f9U,0x51f9U,0xa3f2U,0x57efU,0xafdeU,0x4fb7U,
		0x9f6eU,0x2ed7U,0x5daeU,0xbb5cU,0x66b3U,0xcd66U,0x8ac7U,0x0585U,0x0b0aU,0x1614U,
		0x2c28U,0x5850U,0xb0a0U,0x714bU,0xe296U,0xd527U,0xba45U,0x6481U,0xc902U,0x820fU,
		0x1415U,0x282aU,0x5054U,0xa0a8U,0x515bU,0xa2b6U,0x5567U,0xaaceU,0x4597U,0x8b2eU,
		0x0657U,0x0caeU,0x195cU,0x32b8U,0x6570U,0xcae0U,0x85cbU,0x1b9dU,0x373aU,0x6e74U,
		0xdce8U,0xa9dbU,0x43bdU,0x877aU,0x1effU,0x3dfeU,0x7bfcU,0xf7f8U,0xfffbU,0xeffdU,
		0xcff1U,0x8fe9U,0x0fd9U,0x1fb2U,0x3f64U,0x7ec8U,0xfd90U,0xeb2bU,0xc65dU,0x9cb1U,
		0x2969U,0x52d2U,0xa5a4U,0x5b43U,0xb686U,0x7d07U,0xfa0eU,0xe417U,0xd825U,0xa041U,
		0x5089U,0xa112U,0x522fU,0xa45eU,0x58b7U,0xb16eU,0x72d7U,0xe5aeU,0xdb57U,0xa6a5U,
		0x5d41U,0xba82U,0x650fU,0xca1eU,0x8437U,0x1865U,0x30caU,0x6194U,0xc328U,0x965bU,
		0x3cbdU,0x797aU,0xf2f4U,0xf5e3U,0xfbcdU,0xe791U,0xdf29U,0xae59U,0x4cb9U,0x9972U,
		0x22efU,0x45deU,0x8bbcU,0x0773U,0x0ee6U,0x1dccU,0x3b98U,0x7730U,0xee60U,0xcccbU,
		0x899dU,0x0331U,0x0662U,0x0cc4U,0x1988U,0x3310U,0x6620U,0xcc40U,0x888bU,0x011dU,
		0x023aU,0x0474U,0x08e8U,0x11d0U,0x23a0U,0x4740U,0x8e80U,0x0d0bU,0x1a16U,0x342cU,
		0x6858U,0xd0b0U,0xb16bU,0x72ddU,0xe5baU,0xdb7fU,0xa6f5U,0x5de1U,0xbbc2U,0x678fU,
		0xcf1eU,0x8e37U,0x0c65U,0x18caU,0x3194U,0x6328U,0xc650U,0x9cabU,0x295dU,0x52baU,
		0xa574U,0x5ae3U,0xb5c6U,0x7b87U,0xf70eU,0xfe17U,0xec25U,0xc841U,0x8089U,0x1119U,
		0x2232U,0x4464U,0x88c8U,0x019bU,0x0336U,0x066cU,0x0cd8U,0x19b0U,0x3360U,0x66c0U,
		0xcd80U,0x8b0bU,0x061dU,0x0c3aU,0x1874U,0x30e8U,0x61d0U,0xc3a0U,0x974bU,0x3e9dU,
		0x7d3aU,0xfa74U,0xe4e3U,0xd9cdU,0xa391U,0x5729U,0xae52U,0x4cafU,0x995eU,0x22b7U,
		0x456eU,0x8adcU,0x05b3U,0x0b66U,0x16ccU,0x2d98U,0x5b30U,0xb660U,0x7ccbU,0xf996U,
		0xe327U,0xd645U,0xbc81U,0x6909U,0xd212U,0xb42fU,0x7855U,0xf0aaU,0xf15fU,0xf2b5U,
		0xf561U,0xfac9U,0xe599U,0xdb39U,0xa679U,0x5cf9U,0xb9f2U,0x63efU,0xc7deU,0x9fb7U,
		0x2f65U,0x5ecaU,0xbd94U,0x6b23U,0xd646U,0xbc87U,0x6905U,0xd20aU,0xb41fU,0x7835U,
		0xf06aU,0xf0dfU,0xf1b5U,0xf361U,0xf6c9U,0xfd99U,0xeb39U,0xc679U,0x9cf9U,0x29f9U,
		0x53f2U,0xa7e4U,0x5fc3U,0xbf86U,0x6f07U,0xde0eU,0xac17U,0x4825U,0x904aU,0x309fU,
		0x613eU,0xc27cU,0x94f3U,0x39edU,0x73daU,0xe7b4U,0xdf63U,0xaecdU,0x4d91U,0x9b22U,
		0x264fU,0x4c9eU,0x993cU,0x2273U,0x44e6U,0x89ccU,0x0393U,0x0726U,0x0e4cU,0x1c98U,
		0x3930U,0x7260U,0xe4c0U,0xd98bU,0xa31dU,0x5631U,0xac62U,0x48cfU,0x919eU,0x3337U,
		0x666eU,0xccdcU,0x89b3U,0x036dU,0x06daU,0x0db4U,0x1b68U,0x36d0U,0x6da0U,0xdb40U,
		0xa68bU,0x5d1dU,0xba3aU,0x647fU,0xc8feU,0x81f7U,0x13e5U,0x27caU,0x4f94U,0x9f28U,
		0x2e5bU,0x5cb6U,0xb96cU,0x62d3U,0xc5a6U,0x9b47U,0x2685U,0x4d0aU,0x9a14U,0x2423U,
		0x4846U,0x908cU,0x3113U,0x6226U,0xc44cU,0x9893U,0x212dU,0x425aU,0x84b4U,0x1963U,
		0x32c6U,0x658cU,0xcb18U,0x863bU,0x1c7dU,0x38faU,0x71f4U,0xe3e8U,0xd7dbU,0xbfbdU,
		0x6f71U,0xdee2U,0xadcfU,0x4b95U,0x972aU,0x3e5fU,0x7cbeU,0xf97cU,0xe2f3U,0xd5edU,
		0xbbd1U,0x67a9U,0xcf52U,0x8eafU,0x0d55U,0x1aaaU,0x3554U,0x6aa8U,0xd550U,0xbaabU,
		0x655dU,0xcabaU,0x857fU,0x1af5U,0x35eaU,0x6bd4U,0xd7a8U,0xbf5bU,0x6ebdU,0xdd7aU,
		0xaaffU,0x45f5U,0x8beaU,0x07dfU,0x0fbeU,0x1f7cU,0x3ef8U,0x7df0U,0xfbe0U,0xe7cbU,
		0xdf9dU,0xaf31U,0x4e69U,0x9cd2U,0x29afU,0x535eU,0xa6bcU,0x5d73U,0xbae6U,0x65c7U,
		0xcb8eU,0x8717U,0x1e25U,0x3c4aU,0x7894U,0xf128U,0xf25bU,0xf4bdU,0xf971U,0xe2e9U,
		0xd5d9U,0xbbb9U,0x6779U,0xcef2U,0x8defU,0x0bd5U,0x17aaU,0x2f54U,0x5ea8U,0xbd50U,
		0x6aabU,0xd556U,0xbaa7U,0x6545U,0xca8aU,0x851fU,0x1a35U,0x346aU,0x68d4U,0xd1a8U,
		0xb35bU,0x76bdU,0xed7aU,0xcaffU,0x85f5U,0x1be1U,0x37c2U,0x6f84U,0xdf08U,0xae1bU,
		0x4c3dU,0x987aU,0x20ffU,0x41feU,0x83fcU,0x17f3U,0x2fe6U,0x5fccU,0xbf98U,0x6f3bU,
		0xde76U,0xace7U,0x49c5U,0x938aU,0x371fU,0x6e3eU,0xdc7cU,0xa8f3U,0x41edU,0x83daU,
		0x17bfU,0x2f7eU,0x5efcU,0xbdf8U,0x6bfbU,0xd7f6U,0xbfe7U,0x6fc5U,0xdf8aU,0xaf1fU,
		0x4e35U,0x9c6aU,0x28dfU,0x51beU,0xa37cU,0x56f3U,0xade6U,0x4bc7U,0x978eU,0x3f17U,
		0x7e2eU,0xfc5cU,0xe8b3U,0xc16dU,0x92d1U,0x35a9U,0x6b52U,0xd6a4U,0xbd43U,0x6a8dU,
		0xd51aU,0xba3fU,0x6475U,0xc8eaU,0x81dfU,0x13b5U,0x276aU,0x4ed4U,0x9da8U,0x2b5bU,
		0x56b6U,0xad6cU,0x4ad3U,0x95a6U,0x3b47U,0x768eU,0xed1cU,0xca33U,0x846dU,0x18d1U,
		0x31a2U,0x6344U,0xc688U,0x9d1bU,0x2a3dU,0x547aU,0xa8f4U,0x41e3U,0x83c6U,0x1787U,
		0x2f0eU,0x5e1cU,0xbc38U,0x687bU,0xd0f6U,0xb1e7U,0x73c5U,0xe78aU,0xdf1fU,0xae35U,
		0x4c61U,0x98c2U,0x218fU,0x431eU,0x863cU,0x1c73U,0x38e6U,0x71ccU,0xe398U,0xd73bU,
		0xbe7dU,0x6cf1U,0xd9e2U,0xa3cfU,0x5795U,0xaf2aU,0x4e5fU,0x9cbeU,0x2977U,0x52eeU,
		0xa5dcU,0x5bb3U,0xb766U,0x7ec7U,0xfd8eU,0xeb17U,0xc625U,0x9c41U,0x2889U,0x5112U,
		0xa224U,0x5443U,0xa886U,0x4107U,0x820eU,0x1417U,0x282eU,0x505cU,0xa0b8U,0x517bU,
		0xa2f6U,0x55e7U,0xabceU,0x4797U,0x8f2eU,0x0e57U,0x1caeU,0x395cU,0x72b8U,0xe570U,
		0xdaebU,0xa5ddU,0x5bb1U,0xb762U,0x7ecfU,0xfd9eU,0xeb37U,0xc665U,0x9cc1U,0x2989U,
		0x5312U,0xa624U,0x5c43U,0xb886U,0x6107U,0xc20eU,0x9417U,0x3825U,0x704aU,0xe094U,
		0xd123U,0xb24dU,0x7491U,0xe922U,0xc24fU,0x9495U,0x3921U,0x7242U,0xe484U,0xd903U,
		0xa20dU,0x5411U,0xa822U,0x404fU,0x809eU,0x1137U,0x226eU,0x44dcU,0x89b8U,0x037bU,
		0x06f6U,0x0decU,0x1bd8U,0x37b0U,0x6f60U,0xdec0U,0xad8bU,0x4b1dU,0x963aU,0x3c7fU,
		0x78feU,0xf1fcU,0xf3f3U,0xf7edU,0xffd1U,0xefa9U,0xcf59U,0x8eb9U,0x0d79U,0x1af2U,
		0x35e4U,0x6bc8U,0xd790U,0xbf2bU,0x6e5dU,0xdcbaU,0xa97fU,0x42f5U,0x85eaU,0x1bdfU,
		0x37beU,0x6f7cU,0xdef8U,0xadfbU,0x4bfdU,0x97faU,0x3fffU,0x7ffeU,0xfffcU,0xeff3U,
		0xcfedU,0x8fd1U,0x0fa9U,0x1f52U,0x3ea4U,0x7d48U,0xfa90U,0xe52bU,0xda5dU,0xa4b1U,
		0x5969U,0xb2d2U,0x75afU,0xeb5eU,0xc6b7U,0x9d65U,0x2ac1U,0x5582U,0xab04U,0x4603U,
		0x8c06U,0x0807U,0x100eU,0x201cU,0x4038U,0x8070U,0x10ebU,0x21d6U,0x43acU,0x8758U,
		0x1ebbU,0x3d76U,0x7aecU,0xf5d8U,0xfbbbU,0xe77dU,0xdef1U,0xade9U,0x4bd9U,0x97b2U,
		0x3f6fU,0x7edeU,0xfdbcU,0xeb73U,0xc6edU,0x9dd1U,0x2ba9U,0x5752U,0xaea4U,0x4d43U,
		0x9a86U,0x2507U,0x4a0eU,0x941cU,0x3833U,0x7066U,0xe0ccU,0xd193U,0xb32dU,0x7651U,
		0xeca2U,0xc94fU,0x8295U,0x1521U,0x2a42U,0x5484U,0xa908U,0x421bU,0x8436U,0x1867U,
		0x30ceU,0x619cU,0xc338U,0x967bU,0x3cfdU,0x79faU,0xf3f4U,0xf7e3U,0xffcdU,0xef91U,
		0xcf29U,0x8e59U,0x0cb9U,0x1972U,0x32e4U,0x65c8U,0xcb90U,0x872bU,0x1e5dU,0x3cbaU,
		0x7974U,0xf2e8U,0xf5dbU,0xfbbdU,0xe771U,0xdee9U,0xadd9U,0x4bb9U,0x9772U,0x3eefU,
		0x7ddeU,0xfbbcU,0xe773U,0xdeedU,0xadd1U,0x4ba9U,0x9752U,0x3eafU,0x7d5eU,0xfabcU,
		0xe573U,0xdaedU,0xa5d1U,0x5ba9U,0xb752U,0x7eafU,0xfd5eU,0xeab7U,0xc565U,0x9ac1U,
		0x2589U,0x4b12U,0x9624U,0x3c43U,0x7886U,0xf10cU,0xf213U,0xf42dU,0xf851U,0xe0a9U,
		0xd159U,0xb2b9U,0x7579U,0xeaf2U,0xc5efU,0x9bd5U,0x27a1U,0x4f42U,0x9e84U,0x2d03U,
		0x5a06U,0xb40cU,0x7813U,0xf026U,0xf047U,0xf085U,0xf101U,0xf209U,0xf419U,0xf839U,
		0xe079U,0xd0f9U,0xb1f9U,0x73f9U,0xe7f2U,0xdfefU,0xafd5U,0x4fa1U,0x9f42U,0x2e8fU,
		0x5d1eU,0xba3cU,0x6473U,0xc8e6U,0x81c7U,0x1385U,0x270aU,0x4e14U,0x9c28U,0x285bU,
		0x50b6U,0xa16cU,0x52d3U,0xa5a6U,0x5b47U,0xb68eU,0x7d17U,0xfa2eU,0xe457U,0xd8a5U,
		0xa141U,0x5289U,0xa512U,0x5a2fU,0xb45eU,0x78b7U,0xf16eU,0xf2d7U,0xf5a5U,0xfb41U,
		0xe689U,0xdd19U,0xaa39U,0x4479U,0x88f2U,0x01efU,0x03deU,0x07bcU,0x0f78U,0x1ef0U,
		0x3de0U,0x7bc0U,0xf780U,0xff0bU,0xee1dU,0xcc31U,0x8869U,0x00d9U,0x01b2U,0x0364U,
		0x06c8U,0x0d90U,0x1b20U,0x3640U,0x6c80U,0xd900U,0xa20bU,0x541dU,0xa83aU,0x407fU,
		0x80feU,0x11f7U,0x23eeU,0x47dcU,0x8fb8U,0x0f7bU,0x1ef6U,0x3decU,0x7bd8U,0xf7b0U,
		0xff6bU,0xeeddU,0xcdb1U,0x8b69U,0x06d9U,0x0db2U,0x1b64U,0x36c8U,0x6d90U,0xdb20U,
		0xa64bU,0x5c9dU,0xb93aU,0x627fU,0xc4feU,0x99f7U,0x23e5U,0x47caU,0x8f94U,0x0f23U,
		0x1e46U,0x3c8cU,0x7918U,0xf230U,0xf46bU,0xf8ddU,0xe1b1U,0xd369U,0xb6d9U,0x7db9U,
		0xfb72U,0xe6efU,0xddd5U,0xaba1U,0x4749U,0x8e92U,0x0d2fU,0x1a5eU,0x34bcU,0x6978U,
		0xd2f0U,0xb5ebU,0x7bddU,0xf7baU,0xff7fU,0xeef5U,0xcde1U,0x8bc9U,0x0799U,0x0f32U,
		0x1e64U,0x3cc8U,0x7990U,0xf320U,0xf64bU,0xfc9dU,0xe931U,0xc269U,0x94d9U,0x39b9U,
		0x7372U,0xe6e4U,0xddc3U,0xab8dU,0x4711U,0x8e22U,0x0c4fU,0x189eU,0x313cU,0x6278U,
		0xc4f0U,0x99ebU,0x23ddU,0x47baU,0x8f74U,0x0ee3U,0x1dc6U,0x3b8cU,0x7718U,0xee30U,
		0xcc6bU,0x88ddU,0x01b1U,0x0362U,0x06c4U,0x0d88U,0x1b10U,0x3620U,0x6c40U,0xd880U,
		0xa10bU,0x521dU,0xa43aU,0x587fU,0xb0feU,0x71f7U,0xe3eeU,0xd7d7U,0xbfa5U,0x6f41U,
		0xde82U,0xad0fU,0x4a15U,0x942aU,0x385fU,0x70beU,0xe17cU,0xd2f3U,0xb5edU,0x7bd1U,
		0xf7a2U,0xff4fU,0xee95U,0xcd21U,0x8a49U,0x0499U,0x0932U,0x1264U,0x24c8U,0x4990U,
		0x9320U,0x364bU,0x6c96U,0xd92cU,0xa253U,0x54adU,0xa95aU,0x42bfU,0x857eU,0x1af7U,
		0x35eeU,0x6bdcU,0xd7b8U,0xbf7bU,0x6efdU,0xddfaU,0xabffU,0x47f5U,0x8feaU,0x0fdfU,
		0x1fbeU,0x3f7cU,0x7ef8U,0xfdf0U,0xebebU,0xc7ddU,0x9fb1U,0x2f69U,0x5ed2U,0xbda4U,
		0x6b43U,0xd686U,0xbd07U,0x6a05U,0xd40aU,0xb81fU,0x6035U,0xc06aU,0x90dfU,0x31b5U,
		0x636aU,0xc6d4U,0x9da3U,0x2b4dU,0x569aU,0xad34U,0x4a63U,0x94c6U,0x3987U,0x730eU,
		0xe61cU,0xdc33U,0xa86dU,0x40d1U,0x81a2U,0x134fU,0x269eU,0x4d3cU,0x9a78U,0x24fbU,
		0x49f6U,0x93ecU,0x37d3U,0x6fa6U,0xdf4cU,0xae93U,0x4d2dU,0x9a5aU,0x24bfU,0x497eU,
		0x92fcU,0x35f3U,0x6be6U,0xd7ccU,0xbf93U,0x6f2dU,0xde5aU,0xacbfU,0x4975U,0x92eaU,
		0x35dfU,0x6bbeU,0xd77cU,0xbef3U,0x6dedU,0xdbdaU,0xa7bfU,0x5f75U,0xbeeaU,0x6ddfU,
		0xdbbeU,0xa777U,0x5ee5U,0xbdcaU,0x6b9fU,0xd73eU,0xbe77U,0x6ce5U,0xd9caU,0xa39fU,
		0x5735U,0xae6aU,0x4cdfU,0x99beU,0x2377U,0x46eeU,0x8ddcU,0x0bb3U,0x1766U,0x2eccU,
		0x5d98U,0xbb30U,0x666bU,0xccd6U,0x89a7U,0x0345U,0x068aU,0x0d14U,0x1a28U,0x3450U,
		0x68a0U,0xd140U,0xb28bU,0x751dU,0xea3aU,0xc47fU,0x98f5U,0x21e1U,0x43c2U,0x8784U,
		0x1f03U,0x3e06U,0x7c0cU,0xf818U,0xe03bU,0xd07dU,0xb0f1U,0x71e9U,0xe3d2U,0xd7afU,
		0xbf55U,0x6ea1U,0xdd42U,0xaa8fU,0x4515U,0x8a2aU,0x045fU,0x08beU,0x117cU,0x22f8U,
		0x45f0U,0x8be0U,0x07cbU,0x0f96U,0x1f2cU,0x3e58U,0x7cb0U,0xf960U,0xe2cbU,0xd59dU,
		0xbb31U,0x6669U,0xccd2U,0x89afU,0x0355U,0x06aaU,0x0d54U,0x1aa8U,0x3550U,0x6aa0U,
		0xd540U,0xba8bU,0x651dU,0xca3aU,0x847fU,0x18f5U,0x31eaU,0x63d4U,0xc7a8U,0x9f5bU,
		0x2ebdU,0x5d7aU,0xbaf4U,0x65e3U,0xcbc6U,0x8787U,0x1f05U,0x3e0aU,0x7c14U,0xf828U,
		0xe05bU,0xd0bdU,0xb171U,0x72e9U,0xe5d2U,0xdbafU,0xa755U,0x5ea1U,0xbd42U,0x6a8fU,
		0xd51eU,0xba37U,0x6465U,0xc8caU,0x819fU,0x1335U,0x266aU,0x4cd4U,0x99a8U,0x235bU,
		0x46b6U,0x8d6cU,0x0ad3U,0x15a6U,0x2b4cU,0x5698U,0xad30U,0x4a6bU,0x94d6U,0x39a7U,
		0x734eU,0xe69cU,0xdd33U,0xaa6dU,0x44d1U,0x89a2U,0x034fU,0x069eU,0x0d3cU,0x1a78U,
		0x34f0U,0x69e0U,0xd3c0U,0xb78bU,0x7f1dU,0xfe3aU,0xec7fU,0xc8f5U,0x81e1U,0x13c9U,
		0x2792U,0x4f24U,0x9e48U,0x2c9bU,0x5936U,0xb26cU,0x74d3U,0xe9a6U,0xc347U,0x9685U,
		0x3d01U,0x7a02U,0xf404U,0xf803U,0xe00dU,0xd011U,0xb029U,0x7059U,0xe0b2U,0xd16fU,
		0xb2d5U,0x75a1U,0xeb42U,0xc68fU,0x9d15U,0x2a21U,0x5442U,0xa884U,0x4103U,0x8206U,
		0x1407U,0x280eU,0x501cU,0xa038U,0x507bU,0xa0f6U,0x51e7U,0xa3ceU,0x5797U,0xaf2eU,
		0x4e57U,0x9caeU,0x2957U,0x52aeU,0xa55cU,0x5ab3U,0xb566U,0x7ac7U,0xf58eU,0xfb17U,
		0xe625U,0xdc41U,0xa889U,0x4119U,0x8232U,0x146fU,0x28deU,0x51bcU,0xa378U,0x56fbU,
		0xadf6U,0x4be7U,0x97ceU,0x3f97U,0x7f2eU,0xfe5cU,0xecb3U,0xc96dU,0x82d1U,0x15a9U,
		0x2b52U,0x56a4U,0xad48U,0x4a9bU,0x9536U,0x3a67U,0x74ceU,0xe99cU,0xc333U,0x966dU,
		0x3cd1U,0x79a2U,0xf344U,0xf683U,0xfd0dU,0xea11U,0xc429U,0x9859U,0x20b9U,0x4172U,
		0x82e4U,0x15c3U,0x2b86U,0x570cU,0xae18U,0x4c3bU,0x9876U,0x20e7U,0x41ceU,0x839cU,
		0x1733U,0x2e66U,0x5cccU,0xb998U,0x633bU,0xc676U,0x9ce7U,0x29c5U,0x538aU,0xa714U,
		0x5e23U,0xbc46U,0x6887U,0xd10eU,0xb217U,0x7425U,0xe84aU,0xc09fU,0x9135U,0x3261U,
		0x64c2U,0xc984U,0x8303U,0x160dU,0x2c1aU,0x5834U,0xb068U,0x70dbU,0xe1b6U,0xd367U,
		0xb6c5U,0x7d81U,0xfb02U,0xe60fU,0xdc15U,0xa821U,0x4049U,0x8092U,0x112fU,0x225eU,
		0x44bcU,0x8978U,0x02fbU,0x05f6U,0x0becU,0x17d8U,0x2fb0U,0x5f60U,0xbec0U,0x6d8bU,
		0xdb16U,0xa627U,0x5c45U,0xb88aU,0x611fU,0xc23eU,0x9477U,0x38e5U,0x71caU,0xe394U,
		0xd723U,0xbe4dU,0x6c91U,0xd922U,0xa24fU,0x5495U,0xa92aU,0x425fU,0x84beU,0x1977U,
		0x32eeU,0x65dcU,0xcbb8U,0x877bU,0x1efdU,0x3dfaU,0x7bf4U,0xf7e8U,0xffdbU,0xefbdU,
		0xcf71U,0x8ee9U,0x0dd9U,0x1bb2U,0x3764U,0x6ec8U,0xdd90U,0xab2bU,0x465dU,0x8cbaU,
		0x097fU,0x12feU,0x25fcU,0x4bf8U,0x97f0U,0x3febU,0x7fd6U,0xffacU,0xef53U,0xceadU,
		0x8d51U,0x0aa9U,0x1552U,0x2aa4U,0x5548U,0xaa90U,0x452bU,0x8a56U,0x04a7U,0x094eU,
		0x129cU,0x2538U,0x4a70U,0x94e0U,0x39cbU,0x7396U,0xe72cU,0xde53U,0xacadU,0x4951U,
		0x92a2U,0x354fU,0x6a9eU,0xd53cU,0xba73U,0x64edU,0xc9daU,0x83bfU,0x1775U,0x2eeaU,
		0x5dd4U,0xbba8U,0x675bU,0xceb6U,0x8d67U,0x0ac5U,0x158aU,0x2b14U,0x5628U,0xac50U,
		0x48abU,0x9156U,0x32a7U,0x654eU,0xca9cU,0x8533U,0x1a6dU,0x34daU,0x69b4U,0xd368U,
		0xb6dbU,0x7dbdU,0xfb7aU,0xe6ffU,0xddf5U,0xabe1U,0x47c9U,0x8f92U,0x0f2fU,0x1e5eU,
		0x3cbcU,0x7978U,0xf2f0U,0xf5ebU,0xfbddU,0xe7b1U,0xdf69U,0xaed9U,0x4db9U,0x9b72U,
		0x26efU,0x4ddeU,0x9bbcU,0x2773U,0x4ee6U,0x9dccU,0x2b93U,0x5726U,0xae4cU,0x4c93U,
		0x9926U,0x2247U,0x448eU,0x891cU,0x0233U,0x0466U,0x08ccU,0x1198U,0x2330U,0x4660U,
		0x8cc0U,0x098bU,0x1316U,0x262cU,0x4c58U,0x98b0U,0x216bU,0x42d6U,0x85acU,0x1b53U,
		0x36a6U,0x6d4cU,0xda98U,0xa53bU,0x5a7dU,0xb4faU,0x79ffU,0xf3feU,0xf7f7U,0xffe5U,
		0xefc1U,0xcf89U,0x8f19U,0x0e39U,0x1c72U,0x38e4U,0x71c8U,0xe390U,0xd72bU,0xbe5dU,
		0x6cb1U,0xd962U,0xa2cfU,0x5595U,0xab2aU,0x465fU,0x8cbeU,0x0977U,0x12eeU,0x25dcU,
		0x4bb8U,0x9770U,0x3eebU,0x7dd6U,0xfbacU,0xe753U,0xdeadU,0xad51U,0x4aa9U,0x9552U,
		0x3aafU,0x755eU,0xeabcU,0xc573U,0x9aedU,0x25d1U,0x4ba2U,0x9744U,0x3e83U,0x7d06U,
		0xfa0cU,0xe413U,0xd82dU,0xa051U,0x50a9U,0xa152U,0x52afU,0xa55eU,0x5ab7U,0xb56eU,
		0x7ad7U,0xf5aeU,0xfb57U,0xe6a5U,0xdd41U,0xaa89U,0x4519U,0x8a32U,0x046fU,0x08deU,
		0x11bcU,0x2378U,0x46f0U,0x8de0U,0x0bcbU,0x1796U,0x2f2cU,0x5e58U,0xbcb0U,0x696bU,
		0xd2d6U,0xb5a7U,0x7b45U,0xf68aU,0xfd1fU,0xea35U,0xc461U,0x98c9U,0x2199U,0x4332U,
		0x8664U,0x1cc3U,0x3986U,0x730cU,0xe618U,0xdc3bU,0xa87dU,0x40f1U,0x81e2U,0x13cfU,
		0x279eU,0x4f3cU,0x9e78U,0x2cfbU,0x59f6U,0xb3ecU,0x77d3U,0xefa6U,0xcf47U,0x8e85U,
		0x0d01U,0x1a02U,0x3404U,0x6808U,0xd010U,0xb02bU,0x705dU,0xe0baU,0xd17fU,0xb2f5U,
		0x75e1U,0xebc2U,0xc78fU,0x9f15U,0x2e21U,0x5c42U,0xb884U,0x6103U,0xc206U,0x9407U,
		0x3805U,0x700aU,0xe014U,0xd023U,0xb04dU,0x7091U,0xe122U,0xd24fU,0xb495U,0x7921U,
		0xf242U,0xf48fU,0xf915U,0xe221U,0xd449U,0xb899U,0x6139U,0xc272U,0x94efU,0x39d5U,
		0x73aaU,0xe754U,0xdea3U,0xad4dU,0x4a91U,0x9522U,0x3a4fU,0x749eU,0xe93cU,0xc273U,
		0x94edU,0x39d1U,0x73a2U,0xe744U,0xde83U,0xad0dU,0x4a11U,0x9422U,0x384fU,0x709eU,
		0xe13cU,0xd273U,0xb4edU,0x79d1U,0xf3a2U,0xf74fU,0xfe95U,0xed21U,0xca49U,0x8499U,
		0x1939U,0x3272U,0x64e4U,0xc9c8U,0x839bU,0x173dU,0x2e7aU,0x5cf4U,0xb9e8U,0x63dbU,
		0xc7b6U,0x9f67U,0x2ec5U,0x5d8aU,0xbb14U,0x6623U,0xcc46U,0x8887U,0x0105U,0x020aU,
		0x0414U,0x0828U,0x1050U,0x20a0U,0x4140U,0x8280U,0x150bU,0x2a16U,0x542cU,0xa858U,
		0x40bbU,0x8176U,0x12e7U,0x25ceU,0x4b9cU,0x9738U,0x3e7bU,0x7cf6U,0xf9ecU,0xe3d3U,
		0xd7adU,0xbf51U,0x6ea9U,0xdd52U,0xaaafU,0x4555U,0x8aaaU,0x055fU,0x0abeU,0x157cU,
		0x2af8U,0x55f0U,0xabe0U,0x47cbU,0x8f96U,0x0f27U,0x1e4eU,0x3c9cU,0x7938U,0xf270U,
		0xf4ebU,0xf9ddU,0xe3b1U,0xd769U,0xbed9U,0x6db9U,0xdb72U,0xa6efU,0x5dd5U,0xbbaaU,
		0x675fU,0xcebeU,0x8d77U,0x0ae5U,0x15caU,0x2b94U,0x5728U,0xae50U,0x4cabU,0x9956U,
		0x22a7U,0x454eU,0x8a9cU,0x0533U,0x0a66U,0x14ccU,0x2998U,0x5330U,0xa660U,0x5ccbU,
		0xb996U,0x6327U,0xc64eU,0x9c97U,0x2925U,0x524aU,0xa494U,0x5923U,0xb246U,0x7487U,
		0xe90eU,0xc217U,0x9425U,0x3841U,0x7082U,0xe104U,0xd203U,0xb40dU,0x7811U,0xf022U,
		0xf04fU,0xf095U,0xf121U,0xf249U,0xf499U,0xf939U,0xe279U,0xd4f9U,0xb9f9U,0x63f9U,
		0xc7f2U,0x9fefU,0x2fd5U,0x5faaU,0xbf54U,0x6ea3U,0xdd46U,0xaa87U,0x4505U,0x8a0aU,
		0x041fU,0x083eU,0x107cU,0x20f8U,0x41f0U,0x83e0U,0x17cbU,0x2f96U,0x5f2cU,0xbe58U,
		0x6cbbU,0xd976U,0xa2e7U,0x55c5U,0xab8aU,0x471fU,0x8e3eU,0x0c77U,0x18eeU,0x31dcU,
		0x63b8U,0xc770U,0x9eebU,0x2dddU,0x5bbaU,0xb774U,0x7ee3U,0xfdc6U,0xeb87U,0xc705U,
		0x9e01U,0x2c09U,0x5812U,0xb024U,0x7043U,0xe086U,0xd107U,0xb205U,0x7401U,0xe802U,
		0xc00fU,0x9015U,0x3021U,0x6042U,0xc084U,0x9103U,0x320dU,0x641aU,0xc834U,0x8063U,
		0x10cdU,0x219aU,0x4334U,0x8668U,0x1cdbU,0x39b6U,0x736cU,0xe6d8U,0xddbbU,0xab7dU,
		0x46f1U,0x8de2U,0x0bcfU,0x179eU,0x2f3cU,0x5e78U,0xbcf0U,0x69ebU,0xd3d6U,0xb7a7U,
		0x7f45U,0xfe8aU,0xed1fU,0xca35U,0x8461U,0x18c9U,0x3192U,0x6324U,0xc648U,0x9c9bU,
		0x293dU,0x527aU,0xa4f4U,0x59e3U,0xb3c6U,0x7787U,0xef0eU,0xce17U,0x8c25U,0x0841U,
		0x1082U,0x2104U,0x4208U,0x8410U,0x182bU,0x3056U,0x60acU,0xc158U,0x92bbU,0x357dU,
		0x6afaU,0xd5f4U,0xbbe3U,0x67cdU,0xcf9aU,0x8f3fU,0x0e75U,0x1ceaU,0x39d4U,0x73a8U,
		0xe750U,0xdeabU,0xad5dU,0x4ab1U,0x9562U,0x3acfU,0x759eU,0xeb3cU,0xc673U,0x9cedU,
		0x29d1U,0x53a2U,0xa744U,0x5e83U,0xbd06U,0x6a07U,0xd40eU,0xb817U,0x6025U,0xc04aU,
		0x909fU,0x3135U,0x626aU,0xc4d4U,0x99a3U,0x234dU,0x469aU,0x8d34U,0x0a63U,0x14c6U,
		0x298cU,0x5318U,0xa630U,0x5c6bU,0xb8d6U,0x61a7U,0xc34eU,0x9697U,0x3d25U,0x7a4aU,
		0xf494U,0xf923U,0xe24dU,0xd491U,0xb929U,0x6259U,0xc4b2U,0x996fU,0x22d5U,0x45aaU,
		0x8b54U,0x06a3U,0x0d46U,0x1a8cU,0x3518U,0x6a30U,0xd460U,0xb8cbU,0x619dU,0xc33aU,
		0x967fU,0x3cf5U,0x79eaU,0xf3d4U,0xf7a3U,0xff4dU,0xee91U,0xcd29U,0x8a59U,0x04b9U,
		0x0972U,0x12e4U,0x25c8U,0x4b90U,0x9720U,0x3e4bU,0x7c96U,0xf92cU,0xe253U,0xd4adU,
		0xb951U,0x62a9U,0xc552U,0x9aafU,0x2555U,0x4aaaU,0x9554U,0x3aa3U,0x7546U,0xea8cU,
		0xc513U,0x9a2dU,0x2451U,0x48a2U,0x9144U,0x3283U,0x6506U,0xca0cU,0x8413U,0x182dU,
		0x305aU,0x60b4U,0xc168U,0x92dbU,0x35bdU,0x6b7aU,0xd6f4U,0xbde3U,0x6bcdU,0xd79aU,
		0xbf3fU,0x6e75U,0xdceaU,0xa9dfU,0x43b5U,0x876aU,0x1edfU,0x3dbeU,0x7b7cU,0xf6f8U,
		0xfdfbU,0xebfdU,0xc7f1U,0x9fe9U,0x2fd9U,0x5fb2U,0xbf64U,0x6ec3U,0xdd86U,0xab07U,
		0x4605U,0x8c0aU,0x081fU,0x103eU,0x207cU,0x40f8U,0x81f0U,0x13ebU,0x27d6U,0x4facU,
		0x9f58U,0x2ebbU,0x5d76U,0xbaecU,0x65d3U,0xcba6U,0x8747U,0x1e85U,0x3d0aU,0x7a14U,
		0xf428U,0xf85bU,0xe0bdU,0xd171U,0xb2e9U,0x75d9U,0xebb2U,0xc76fU,0x9ed5U,0x2da1U,
		0x5b42U,0xb684U,0x7d03U,0xfa06U,0xe407U,0xd805U,0xa001U,0x5009U,0xa012U,0x502fU,
		0xa05eU,0x50b7U,0xa16eU,0x52d7U,0xa5aeU,0x5b57U,0xb6aeU,0x7d57U,0xfaaeU,0xe557U,
		0xdaa5U,0xa541U,0x5a89U,0xb512U,0x7a2fU,0xf45eU,0xf8b7U,0xe165U,0xd2c1U,0xb589U,
		0x7b19U,0xf632U,0xfc6fU,0xe8d5U,0xc1a1U,0x9349U,0x3699U,0x6d32U,0xda64U,0xa4c3U,
		0x598dU,0xb31aU,0x763fU,0xec7eU,0xc8f7U,0x81e5U,0x13c1U,0x2782U,0x4f04U,0x9e08U,
		0x2c1bU,0x5836U,0xb06cU,0x70d3U,0xe1a6U,0xd347U,0xb685U,0x7d01U,0xfa02U,0xe40fU,
		0xd815U,0xa021U,0x5049U,0xa092U,0x512fU,0xa25eU,0x54b7U,0xa96eU,0x42d7U,0x85aeU,
		0x1b57U,0x36aeU,0x6d5cU,0xdab8U,0xa57bU,0x5afdU,0xb5faU,0x7bffU,0xf7feU,0xfff7U,
		0xefe5U,0xcfc1U,0x8f89U,0x0f19U,0x1e32U,0x3c64U,0x78c8U,0xf190U,0xf32bU,0xf65dU,
		0xfcb1U,0xe969U,0xc2d9U,0x95b9U,0x3b79U,0x76f2U,0xede4U,0xcbc3U,0x878dU,0x1f11U,
		0x3e22U,0x7c44U,0xf888U,0xe11bU,0xd23dU,0xb471U,0x78e9U,0xf1d2U,0xf3afU,0xf755U,
		0xfea1U,0xed49U,0xca99U,0x8539U,0x1a79U,0x34f2U,0x69e4U,0xd3c8U,0xb79bU,0x7f3dU,
		0xfe7aU,0xecffU,0xc9f5U,0x83e1U,0x17c9U,0x2f92U,0x5f24U,0xbe48U,0x6c9bU,0xd936U,
		0xa267U,0x54c5U,0xa98aU,0x431fU,0x863eU,0x1c77U,0x38eeU,0x71dcU,0xe3b8U,0xd77bU,
		0xbefdU,0x6df1U,0xdbe2U,0xa7cfU,0x5f95U,0xbf2aU,0x6e5fU,0xdcbeU,0xa977U,0x42e5U,
		0x85caU,0x1b9fU,0x373eU,0x6e7cU,0xdcf8U,0xa9fbU,0x43fdU,0x87faU,0x1fffU,0x3ffeU,
		0x7ffcU,0xfff8U,0xeffbU,0xcffdU,0x8ff1U,0x0fe9U,0x1fd2U,0x3fa4U,0x7f48U,0xfe90U,
		0xed2bU,0xca5dU,0x84b1U,0x1969U,0x32d2U,0x65a4U,0xcb48U,0x869bU,0x1d3dU,0x3a7aU,
		0x74f4U,0xe9e8U,0xc3dbU,0x97bdU,0x3f71U,0x7ee2U,0xfdc4U,0xeb83U,0xc70dU,0x9e11U,
		0x2c29U,0x5852U,0xb0a4U,0x7143U,0xe286U,0xd507U,0xba05U,0x6401U,0xc802U,0x800fU,
		0x1015U,0x202aU,0x4054U,0x80a8U,0x115bU,0x22b6U,0x456cU,0x8ad8U,0x05bbU,0x0b76U,
		0x16ecU,0x2dd8U,0x5bb0U,0xb760U,0x7ecbU,0xfd96U,0xeb27U,0xc645U,0x9c81U,0x2909U,
		0x5212U,0xa424U,0x5843U,0xb086U,0x7107U,0xe20eU,0xd417U,0xb825U,0x6041U,0xc082U,
		0x910fU,0x3215U,0x642aU,0xc854U,0x80a3U,0x114dU,0x229aU,0x4534U,0x8a68U,0x04dbU,
		0x09b6U,0x136cU,0x26d8U,0x4db0U,0x9b60U,0x26cbU,0x4d96U,0x9b2cU,0x2653U,0x4ca6U,
		0x994cU,0x2293U,0x4526U,0x8a4cU,0x0493U,0x0926U,0x124cU,0x2498U,0x4930U,0x9260U,
		0x34cbU,0x6996U,0xd32cU,0xb653U,0x7cadU,0xf95aU,0xe2bfU,0xd575U,0xbae1U,0x65c9U,
		0xcb92U,0x872fU,0x1e55U,0x3caaU,0x7954U,0xf2a8U,0xf55bU,0xfabdU,0xe571U,0xdae9U,
		0xa5d9U,0x5bb9U,0xb772U,0x7eefU,0xfddeU,0xebb7U,0xc765U,0x9ec1U,0x2d89U,0x5b12U,
		0xb624U,0x7c43U,0xf886U,0xe107U,0xd205U,0xb401U,0x7809U,0xf012U,0xf02fU,0xf055U,
		0xf0a1U,0xf149U,0xf299U,0xf539U,0xfa79U,0xe4f9U,0xd9f9U,0xa3f9U,0x57f9U,0xaff2U,
		0x4fefU,0x9fdeU,0x2fb7U,0x5f6eU,0xbedcU,0x6db3U,0xdb66U,0xa6c7U,0x5d85U,0xbb0aU,
		0x661fU,0xcc3eU,0x8877U,0x00e5U,0x01caU,0x0394U,0x0728U,0x0e50U,0x1ca0U,0x3940U,
		0x7280U,0xe500U,0xda0bU,0xa41dU,0x5831U,0xb062U,0x70cfU,0xe19eU,0xd337U,0xb665U,
		0x7cc1U,0xf982U,0xe30fU,0xd615U,0xbc21U,0x6849U,0xd092U,0xb12fU,0x7255U,0xe4aaU,
		0xd95fU,0xa2b5U,0x5561U,0xaac2U,0x458fU,0x8b1eU,0x0637U,0x0c6eU,0x18dcU,0x31b8U,
		0x6370U,0xc6e0U,0x9dcbU,0x2b9dU,0x573aU,0xae74U,0x4ce3U,0x99c6U,0x2387U,0x470eU,
		0x8e1cU,0x0c33U,0x1866U,0x30ccU,0x6198U,0xc330U,0x966bU,0x3cddU,0x79baU,0xf374U,
		0xf6e3U,0xfdcdU,0xeb91U,0xc729U,0x9e59U,0x2cb9U,0x5972U,0xb2e4U,0x75c3U,0xeb86U,
		0xc707U,0x9e05U,0x2c01U,0x5802U,0xb004U,0x7003U,0xe006U,0xd007U,0xb005U,0x7001U,
		0xe002U,0xd00fU,0xb015U,0x7021U,0xe042U,0xd08fU,0xb115U,0x7221U,0xe442U,0xd88fU,
		0xa115U,0x5221U,0xa442U,0x588fU,0xb11eU,0x7237U,0xe46eU,0xd8d7U,0xa1a5U,0x5341U,
		0xa682U,0x5d0fU,0xba1eU,0x6437U,0xc86eU,0x80d7U,0x11a5U,0x234aU,0x4694U,0x8d28U,
		0x0a5bU,0x14b6U,0x296cU,0x52d8U,0xa5b0U,0x5b6bU,0xb6d6U,0x7da7U,0xfb4eU,0xe697U,
		0xdd25U,0xaa41U,0x4489U,0x8912U,0x022fU,0x045eU,0x08bcU,0x1178U,0x22f0U,0x45e0U,
		0x8bc0U,0x078bU,0x0f16U,0x1e2cU,0x3c58U,0x78b0U,0xf160U,0xf2cbU,0xf59dU,0xfb31U,
		0xe669U,0xdcd9U,0xa9b9U,0x4379U,0x86f2U,0x1defU,0x3bdeU,0x77bcU,0xef78U,0xcefbU,
		0x8dfdU,0x0bf1U,0x17e2U,0x2fc4U,0x5f88U,0xbf10U,0x6e2bU,0xdc56U,0xa8a7U,0x4145U,
		0x828aU,0x151fU,0x2a3eU,0x547cU,0xa8f8U,0x41fbU,0x83f6U,0x17e7U,0x2fceU,0x5f9cU,
		0xbf38U,0x6e7bU,0xdcf6U,0xa9e7U,0x43c5U,0x878aU,0x1f1fU,0x3e3eU,0x7c7cU,0xf8f8U,
		0xe1fbU,0xd3fdU,0xb7f1U,0x7fe9U,0xffd2U,0xefafU,0xcf55U,0x8ea1U,0x0d49U,0x1a92U,
		0x3524U,0x6a48U,0xd490U,0xb92bU,0x625dU,0xc4baU,0x997fU,0x22f5U,0x45eaU,0x8bd4U,
		0x07a3U,0x0f46U,0x1e8cU,0x3d18U,0x7a30U,0xf460U,0xf8cbU,0xe19dU,0xd331U,0xb669U,
		0x7cd9U,0xf9b2U,0xe36fU,0xd6d5U,0xbda1U,0x6b49U,0xd692U,0xbd2fU,0x6a55U,0xd4aaU,
		0xb95fU,0x62b5U,0xc56aU,0x9adfU,0x25b5U,0x4b6aU,0x96d4U,0x3da3U,0x7b46U,0xf68cU,
		0xfd13U,0xea2dU,0xc451U,0x98a9U,0x2159U,0x42b2U,0x8564U,0x1ac3U,0x3586U,0x6b0cU,
		0xd618U,0xbc3bU,0x687dU,0xd0faU,0xb1ffU,0x73f5U,0xe7eaU,0xdfdfU,0xafb5U,0x4f61U,
		0x9ec2U,0x2d8fU,0x5b1eU,0xb63cU,0x7c73U,0xf8e6U,0xe1c7U,0xd385U,0xb701U,0x7e09U,
		0xfc12U,0xe82fU,0xc055U,0x90a1U,0x3149U,0x6292U,0xc524U,0x9a43U,0x248dU,0x491aU,
		0x9234U,0x3463U,0x68c6U,0xd18cU,0xb313U,0x762dU,0xec5aU,0xc8bfU,0x8175U,0x12e1U,
		0x25c2U,0x4b84U,0x9708U,0x3e1bU,0x7c36U,0xf86cU,0xe0d3U,0xd1adU,0xb351U,0x76a9U,
		0xed52U,0xcaafU,0x8555U,0x1aa1U,0x3542U,0x6a84U,0xd508U,0xba1bU,0x643dU,0xc87aU,
		0x80ffU,0x11f5U,0x23eaU,0x47d4U,0x8fa8U,0x0f5bU,0x1eb6U,0x3d6cU,0x7ad8U,0xf5b0U,
		0xfb6bU,0xe6ddU,0xddb1U,0xab69U,0x46d9U,0x8db2U,0x0b6fU,0x16deU,0x2dbcU,0x5b78U,
		0xb6f0U,0x7debU,0xfbd6U,0xe7a7U,0xdf45U,0xae81U,0x4d09U,0x9a12U,0x242fU,0x485eU,
		0x90bcU,0x3173U,0x62e6U,0xc5ccU,0x9b93U,0x272dU,0x4e5aU,0x9cb4U,0x2963U,0x52c6U,
		0xa58cU,0x5b13U,0xb626U,0x7c47U,0xf88eU,0xe117U,0xd225U,0xb441U,0x7889U,0xf112U,
		0xf22fU,0xf455U,0xf8a1U,0xe149U,0xd299U,0xb539U,0x7a79U,0xf4f2U,0xf9efU,0xe3d5U,
		0xd7a1U,0xbf49U,0x6e99U,0xdd32U,0xaa6fU,0x44d5U,0x89aaU,0x035fU,0x06beU,0x0d7cU,
		0x1af8U,0x35f0U,0x6be0U,0xd7c0U,0xbf8bU,0x6f1dU,0xde3aU,0xac7fU,0x48f5U,0x91eaU,
		0x33dfU,0x67beU,0xcf7cU,0x8ef3U,0x0dedU,0x1bdaU,0x37b4U,0x6f68U,0xded0U,0xadabU,
		0x4b5dU,0x96baU,0x3d7fU,0x7afeU,0xf5fcU,0xfbf3U,0xe7edU,0xdfd1U,0xafa9U,0x4f59U,
		0x9eb2U,0x2d6fU,0x5adeU,0xb5bcU,0x7b73U,0xf6e6U,0xfdc7U,0xeb85U,0xc701U,0x9e09U,
		0x2c19U,0x5832U,0xb064U,0x70c3U,0xe186U,0xd307U,0xb605U,0x7c01U,0xf802U,0xe00fU,
		0xd015U,0xb021U,0x7049U,0xe092U,0xd12fU,0xb255U,0x74a1U,0xe942U,0xc28fU,0x9515U,
		0x3a21U,0x7442U,0xe884U,0xc103U,0x920dU,0x3411U,0x6822U,0xd044U,0xb083U,0x710dU,
		0xe21aU,0xd43fU,0xb875U,0x60e1U,0xc1c2U,0x938fU,0x3715U,0x6e2aU,0xdc54U,0xa8a3U,
		0x414dU,0x829aU,0x153fU,0x2a7eU,0x54fcU,0xa9f8U,0x43fbU,0x87f6U,0x1fe7U,0x3fceU,
		0x7f9cU,0xff38U,0xee7bU,0xccfdU,0x89f1U,0x03e9U,0x07d2U,0x0fa4U,0x1f48U,0x3e90U,
		0x7d20U,0xfa40U,0xe48bU,0xd91dU,0xa231U,0x5469U,0xa8d2U,0x41afU,0x835eU,0x16b7U,
		0x2d6eU,0x5adcU,0xb5b8U,0x7b7bU,0xf6f6U,0xfde7U,0xebc5U,0xc781U,0x9f09U,0x2e19U,
		0x5c32U,0xb864U,0x60c3U,0xc186U,0x9307U,0x3605U,0x6c0aU,0xd814U,0xa023U,0x504dU,
		0xa09aU,0x513fU,0xa27eU,0x54f7U,0xa9eeU,0x43d7U,0x87aeU,0x1f57U,0x3eaeU,0x7d5cU,
		0xfab8U,0xe57bU,0xdafdU,0xa5f1U,0x5be9U,0xb7d2U,0x7fafU,0xff5eU,0xeeb7U,0xcd65U,
		0x8ac1U,0x0589U,0x0b12U,0x1624U,0x2c48U,0x5890U,0xb120U,0x724bU,0xe496U,0xd927U,
		0xa245U,0x5481U,0xa902U,0x420fU,0x841eU,0x1837U,0x306eU,0x60dcU,0xc1b8U,0x937bU,
		0x36fdU,0x6dfaU,0xdbf4U,0xa7e3U,0x5fcdU,0xbf9aU,0x6f3fU,0xde7eU,0xacf7U,0x49e5U,
		0x93caU,0x379fU,0x6f3eU,0xde7cU,0xacf3U,0x49edU,0x93daU,0x37bfU,0x6f7eU,0xdefcU,
		0xadf3U,0x4bedU,0x97daU,0x3fbfU,0x7f7eU,0xfefcU,0xedf3U,0xcbedU,0x87d1U,0x1fa9U,
		0x3f52U,0x7ea4U,0xfd48U,0xea9bU,0xc53dU,0x9a71U,0x24e9U,0x49d2U,0x93a4U,0x3743U,
		0x6e86U,0xdd0cU,0xaa13U,0x442dU,0x885aU,0x00bfU,0x017eU,0x02fcU,0x05f8U,0x0bf0U,
		0x17e0U,0x2fc0U,0x5f80U,0xbf00U,0x6e0bU,0xdc16U,0xa827U,0x4045U,0x808aU,0x111fU,
		0x223eU,0x447cU,0x88f8U,0x01fbU,0x03f6U,0x07ecU,0x0fd8U,0x1fb0U,0x3f60U,0x7ec0U,
		0xfd80U,0xeb0bU,0xc61dU,0x9c31U,0x2869U,0x50d2U,0xa1a4U,0x5343U,0xa686U,0x5d07U,
		0xba0eU,0x6417U,0xc82eU,0x8057U,0x10a5U,0x214aU,0x4294U,0x8528U,0x1a5bU,0x34b6U,
		0x696cU,0xd2d8U,0xb5bbU,0x7b7dU,0xf6faU,0xfdffU,0xebf5U,0xc7e1U,0x9fc9U,0x2f99U,
		0x5f32U,0xbe64U,0x6cc3U,0xd986U,0xa307U,0x5605U,0xac0aU,0x481fU,0x903eU,0x3077U,
		0x60eeU,0xc1dcU,0x93b3U,0x376dU,0x6edaU,0xddb4U,0xab63U,0x46cdU,0x8d9aU,0x0b3fU,
		0x167eU,0x2cfcU,0x59f8U,0xb3f0U,0x77ebU,0xefd6U,0xcfa7U,0x8f45U,0x0e81U,0x1d02U,
		0x3a04U,0x7408U,0xe810U,0xc02bU,0x905dU,0x30b1U,0x6162U,0xc2c4U,0x9583U,0x3b0dU,
		0x761aU,0xec34U,0xc863U,0x80cdU,0x1191U,0x2322U,0x4644U,0x8c88U,0x091bU,0x1236U,
		0x246cU,0x48d8U,0x91b0U,0x336bU,0x66d6U,0xcdacU,0x8b53U,0x06adU,0x0d5aU,0x1ab4U,
		0x3568U,0x6ad0U,0xd5a0U,0xbb4bU,0x669dU,0xcd3aU,0x8a7fU,0x04f5U,0x09eaU,0x13d4U,
		0x27a8U,0x4f50U,0x9ea0U,0x2d4bU,0x5a96U,0xb52cU,0x7a53U,0xf4a6U,0xf947U,0xe285U,
		0xd501U,0xba09U,0x6419U,0xc832U,0x806fU,0x10d5U,0x21aaU,0x4354U,0x86a8U,0x1d5bU,
		0x3ab6U,0x756cU,0xead8U,0xc5bbU,0x9b7dU,0x26f1U,0x4de2U,0x9bc4U,0x2783U,0x4f06U,
		0x9e0cU,0x2c13U,0x5826U,0xb04cU,0x7093U,0xe126U,0xd247U,0xb485U,0x7901U,0xf202U,
		0xf40fU,0xf815U,0xe021U,0xd049U,0xb099U,0x7139U,0xe272U,0xd4efU,0xb9d5U,0x63a1U,
		0xc742U,0x9e8fU,0x2d15U,0x5a2aU,0xb454U,0x78a3U,0xf146U,0xf287U,0xf505U,0xfa01U,
		0xe409U,0xd819U,0xa039U,0x5079U,0xa0f2U,0x51efU,0xa3deU,0x57b7U,0xaf6eU,0x4ed7U,
		0x9daeU,0x2b57U,0x56aeU,0xad5cU,0x4ab3U,0x9566U,0x3ac7U,0x758eU,0xeb1cU,0xc633U,
		0x9c6dU,0x28d1U,0x51a2U,0xa344U,0x5683U,0xad06U,0x4a07U,0x940eU,0x3817U,0x702eU,
		0xe05cU,0xd0b3U,0xb16dU,0x72d1U,0xe5a2U,0xdb4fU,0xa695U,0x5d21U,0xba42U,0x648fU,
		0xc91eU,0x8237U,0x1465U,0x28caU,0x5194U,0xa328U,0x565bU,0xacb6U,0x4967U,0x92ceU,
		0x3597U,0x6b2eU,0xd65cU,0xbcb3U,0x696dU,0xd2daU,0xb5bfU,0x7b75U,0xf6eaU,0xfddfU,
		0xebb5U,0xc761U,0x9ec9U,0x2d99U,0x5b32U,0xb664U,0x7cc3U,0xf986U,0xe307U,0xd605U,
		0xbc01U,0x6809U,0xd012U,0xb02fU,0x7055U,0xe0aaU,0xd15fU,0xb2b5U,0x7561U,0xeac2U,
		0xc58fU,0x9b15U,0x2621U,0x4c42U,0x9884U,0x2103U,0x4206U,0x840cU,0x1813U,0x3026U,
		0x604cU,0xc098U,0x913bU,0x327dU,0x64faU,0xc9f4U,0x83e3U,0x17cdU,0x2f9aU,0x5f34U,
		0xbe68U,0x6cdbU,0xd9b6U,0xa367U,0x56c5U,0xad8aU,0x4b1fU,0x963eU,0x3c77U,0x78eeU,
		0xf1dcU,0xf3b3U,0xf76dU,0xfed1U,0xeda9U,0xcb59U,0x86b9U,0x1d79U,0x3af2U,0x75e4U,
		0xebc8U,0xc79bU,0x9f3dU,0x2e71U,0x5ce2U,0xb9c4U,0x6383U,0xc706U,0x9e07U,0x2c05U,
		0x580aU,0xb014U,0x7023U,0xe046U,0xd087U,0xb105U,0x7201U,0xe402U,0xd80fU,0xa015U,
		0x5021U,0xa042U,0x508fU,0xa11eU,0x5237U,0xa46eU,0x58d7U,0xb1aeU,0x7357U,0xe6aeU,
		0xdd57U,0xaaa5U,0x4541U,0x8a82U,0x050fU,0x0a1eU,0x143cU,0x2878U,0x50f0U,0xa1e0U,
		0x53cbU,0xa796U,0x5f27U,0xbe4eU,0x6c97U,0xd92eU,0xa257U,0x54a5U,0xa94aU,0x429fU,
		0x853eU,0x1a77U,0x34eeU,0x69dcU,0xd3b8U,0xb77bU,0x7efdU,0xfdfaU,0xebffU,0xc7f5U,
		0x9fe1U,0x2fc9U,0x5f92U,0xbf24U,0x6e43U,0xdc86U,0xa907U,0x4205U,0x840aU,0x181fU,
		0x303eU,0x607cU,0xc0f8U,0x91fbU,0x33fdU,0x67faU,0xcff4U,0x8fe3U,0x0fcdU,0x1f9aU,
		0x3f34U,0x7e68U,0xfcd0U,0xe9abU,0xc35dU,0x96b1U,0x3d69U,0x7ad2U,0xf5a4U,0xfb43U,
		0xe68dU,0xdd11U,0xaa29U,0x4459U,0x88b2U,0x016fU,0x02deU,0x05bcU,0x0b78U,0x16f0U,
		0x2de0U,0x5bc0U,0xb780U,0x7f0bU,0xfe16U,0xec27U,0xc845U,0x8081U,0x1109U,0x2212U,
		0x4424U,0x8848U,0x009bU,0x0136U,0x026cU,0x04d8U,0x09b0U,0x1360U,0x26c0U,0x4d80U,
		0x9b00U,0x260bU,0x4c16U,0x982cU,0x2053U,0x40a6U,0x814cU,0x1293U,0x2526U,0x4a4cU,
		0x9498U,0x393bU,0x7276U,0xe4ecU,0xd9d3U,0xa3adU,0x5751U,0xaea2U,0x4d4fU,0x9a9eU,
		0x2537U,0x4a6eU,0x94dcU,0x39b3U,0x7366U,0xe6ccU,0xdd93U,0xab2dU,0x4651U,0x8ca2U,
		0x094fU,0x129eU,0x253cU,0x4a78U,0x94f0U,0x39ebU,0x73d6U,0xe7acU,0xdf53U,0xaeadU,
		0x4d51U,0x9aa2U,0x254fU,0x4a9eU,0x953cU,0x3a73U,0x74e6U,0xe9ccU,0xc393U,0x972dU,
		0x3e51U,0x7ca2U,0xf944U,0xe283U,0xd50dU,0xba11U,0x6429U,0xc852U,0x80afU,0x1155U,
		0x22aaU,0x4554U,0x8aa8U,0x055bU,0x0ab6U,0x156cU,0x2ad8U,0x55b0U,0xab60U,0x46cbU,
		0x8d96U,0x0b27U,0x164eU,0x2c9cU,0x5938U,0xb270U,0x74ebU,0xe9d6U,0xc3a7U,0x9745U,
		0x3e81U,0x7d02U,0xfa04U,0xe403U,0xd80dU,0xa011U,0x5029U,0xa052U,0x50afU,0xa15eU,
		0x52b7U,0xa56eU,0x5ad7U,0xb5aeU,0x7b57U,0xf6aeU,0xfd57U,0xeaa5U,0xc541U,0x9a89U,
		0x2519U,0x4a32U,0x9464U,0x38c3U,0x7186U,0xe30cU,0xd613U,0xbc2dU,0x6851U,0xd0a2U,
		0xb14fU,0x7295U,0xe52aU,0xda5fU,0xa4b5U,0x5961U,0xb2c2U,0x758fU,0xeb1eU,0xc637U,
		0x9c65U,0x28c1U,0x5182U,0xa304U,0x5603U,0xac06U,0x4807U,0x900eU,0x3017U,0x602eU,
		0xc05cU,0x90b3U,0x316dU,0x62daU,0xc5b4U,0x9b63U,0x26cdU,0x4d9aU,0x9b34U,0x2663U,
		0x4cc6U,0x998cU,0x2313U,0x4626U,0x8c4cU,0x0893U,0x1126U,0x224cU,0x4498U,0x8930U,
		0x026bU,0x04d6U,0x09acU,0x1358U,0x26b0U,0x4d60U,0x9ac0U,0x258bU,0x4b16U,0x962cU,
		0x3c53U,0x78a6U,0xf14cU,0xf293U,0xf52dU,0xfa51U,0xe4a9U,0xd959U,0xa2b9U,0x5579U,
		0xaaf2U,0x45efU,0x8bdeU,0x07b7U,0x0f6eU,0x1edcU,0x3db8U,0x7b70U,0xf6e0U,0xfdcbU,
		0xeb9dU,0xc731U,0x9e69U,0x2cd9U,0x59b2U,0xb364U,0x76c3U,0xed86U,0xcb07U,0x8605U,
		0x1c01U,0x3802U,0x7004U,0xe008U,0xd01bU,0xb03dU,0x7071U,0xe0e2U,0xd1cfU,0xb395U,
		0x7721U,0xee42U,0xcc8fU,0x8915U,0x0221U,0x0442U,0x0884U,0x1108U,0x2210U,0x4420U,
		0x8840U,0x008bU,0x0116U,0x022cU,0x0458U,0x08b0U,0x1160U,0x22c0U,0x4580U,0x8b00U,
		0x060bU,0x0c16U,0x182cU,0x3058U,0x60b0U,0xc160U,0x92cbU,0x359dU,0x6b3aU,0xd674U,
		0xbce3U,0x69cdU,0xd39aU,0xb73fU,0x7e75U,0xfceaU,0xe9dfU,0xc3b5U,0x9761U,0x3ec9U,
		0x7d92U,0xfb24U,0xe643U,0xdc8dU,0xa911U,0x4229U,0x8452U,0x18afU,0x315eU,0x62bcU,
		0xc578U,0x9afbU,0x25fdU,0x4bfaU,0x97f4U,0x3fe3U,0x7fc6U,0xff8cU,0xef13U,0xce2dU,
		0x8c51U,0x08a9U,0x1152U,0x22a4U,0x4548U,0x8a90U,0x052bU,0x0a56U,0x14acU,0x2958U,
		0x52b0U,0xa560U,0x5acbU,0xb596U,0x7b27U,0xf64eU,0xfc97U,0xe925U,0xc241U,0x9489U,
		0x3919U,0x7232U,0xe464U,0xd8c3U,0xa18dU,0x5311U,0xa622U,0x5c4fU,0xb89eU,0x6137U,
		0xc26eU,0x94d7U,0x39a5U,0x734aU,0xe694U,0xdd23U,0xaa4dU,0x4491U,0x8922U,0x024fU,
		0x049eU,0x093cU,0x1278U,0x24f0U,0x49e0U,0x93c0U,0x378bU,0x6f16U,0xde2cU,0xac53U,
		0x48adU,0x915aU,0x32bfU,0x657eU,0xcafcU,0x85f3U,0x1bedU,0x37daU,0x6fb4U,0xdf68U,
		0xaedbU,0x4dbdU,0x9b7aU,0x26ffU,0x4dfeU,0x9bfcU,0x27f3U,0x4fe6U,0x9fccU,0x2f93U,
		0x5f26U,0xbe4cU,0x6c93U,0xd926U,0xa247U,0x5485U,0xa90aU,0x421fU,0x843eU,0x1877U,
		0x30eeU,0x61dcU,0xc3b8U,0x977bU,0x3efdU,0x7dfaU,0xfbf4U,0xe7e3U,0xdfcdU,0xaf91U,
		0x4f29U,0x9e52U,0x2cafU,0x595eU,0xb2bcU,0x7573U,0xeae6U,0xc5c7U,0x9b85U,0x2701U,
		0x4e02U,0x9c04U,0x2803U,0x5006U,0xa00cU,0x5013U,0xa026U,0x5047U,0xa08eU,0x5117U,
		0xa22eU,0x5457U,0xa8aeU,0x4157U,0x82aeU,0x1557U,0x2aaeU,0x555cU,0xaab8U,0x457bU,
		0x8af6U,0x05e7U,0x0bceU,0x179cU,0x2f38U,0x5e70U,0xbce0U,0x69cbU,0xd396U,0xb727U,
		0x7e45U,0xfc8aU,0xe91fU,0xc235U,0x9461U,0x38c9U,0x7192U,0xe324U,0xd643U,0xbc8dU,
		0x6911U,0xd222U,0xb44fU,0x7895U,0xf12aU,0xf25fU,0xf4b5U,0xf961U,0xe2c9U,0xd599U,
		0xbb39U,0x6679U,0xccf2U,0x89efU,0x03d5U,0x07aaU,0x0f54U,0x1ea8U,0x3d50U,0x7aa0U,
		0xf540U,0xfa8bU,0xe51dU,0xda31U,0xa469U,0x58d9U,0xb1b2U,0x736fU,0xe6deU,0xddb7U,
		0xab65U,0x46c1U,0x8d82U,0x0b0fU,0x161eU,0x2c3cU,0x5878U,0xb0f0U,0x71ebU,0xe3d6U,
		0xd7a7U,0xbf45U,0x6e81U,0xdd02U,0xaa0fU,0x4415U,0x882aU,0x005fU,0x00beU,0x017cU,
		0x02f8U,0x05f0U,0x0be0U,0x17c0U,0x2f80U,0x5f00U,0xbe00U,0x6c0bU,0xd816U,0xa027U,
		0x5045U,0xa08aU,0x511fU,0xa23eU,0x5477U,0xa8eeU,0x41d7U,0x83aeU,0x1757U,0x2eaeU,
		0x5d5cU,0xbab8U,0x657bU,0xcaf6U,0x85e7U,0x1bc5U,0x378aU,0x6f14U,0xde28U,0xac5bU,
		0x48bdU,0x917aU,0x32ffU,0x65feU,0xcbfcU,0x87f3U,0x1fedU,0x3fdaU,0x7fb4U,0xff68U,
		0xeedbU,0xcdbdU,0x8b71U,0x06e9U,0x0dd2U,0x1ba4U,0x3748U,0x6e90U,0xdd20U,0xaa4bU,
		0x449dU,0x893aU,0x027fU,0x04feU,0x09fcU,0x13f8U,0x27f0U,0x4fe0U,0x9fc0U,0x2f8bU,
		0x5f16U,0xbe2cU,0x6c53U,0xd8a6U,0xa147U,0x5285U,0xa50aU,0x5a1fU,0xb43eU,0x7877U,
		0xf0eeU,0xf1d7U,0xf3a5U,0xf741U,0xfe89U,0xed19U,0xca39U,0x8479U,0x18f9U,0x31f2U,
		0x63e4U,0xc7c8U,0x9f9bU,0x2f3dU,0x5e7aU,0xbcf4U,0x69e3U,0xd3c6U,0xb787U,0x7f05U,
		0xfe0aU,0xec1fU,0xc835U,0x8061U,0x10c9U,0x2192U,0x4324U,0x8648U,0x1c9bU,0x3936U,
		0x726cU,0xe4d8U,0xd9bbU,0xa37dU,0x56f1U,0xade2U,0x4bcfU,0x979eU,0x3f37U,0x7e6eU,
		0xfcdcU,0xe9b3U,0xc36dU,0x96d1U,0x3da9U,0x7b52U,0xf6a4U,0xfd43U,0xea8dU,0xc511U,
		0x9a29U,0x2459U,0x48b2U,0x9164U,0x32c3U,0x6586U,0xcb0cU,0x8613U,0x1c2dU,0x385aU,
		0x70b4U,0xe168U,0xd2dbU,0xb5bdU,0x7b71U,0xf6e2U,0xfdcfU,0xeb95U,0xc721U,0x9e49U,
		0x2c99U,0x5932U,0xb264U,0x74c3U,0xe986U,0xc307U,0x9605U,0x3c01U,0x7802U,0xf004U,
		0xf003U,0xf00dU,0xf011U,0xf029U,0xf059U,0xf0b9U,0xf179U,0xf2f9U,0xf5f9U,0xfbf9U,
		0xe7f9U,0xdff9U,0xaff9U,0x4ff9U,0x9ff2U,0x2fefU,0x5fdeU,0xbfbcU,0x6f73U,0xdee6U,
		0xadc7U,0x4b85U,0x970aU,0x3e1fU,0x7c3eU,0xf87cU,0xe0f3U,0xd1edU,0xb3d1U,0x77a9U,
		0xef52U,0xceafU,0x8d55U,0x0aa1U,0x1542U,0x2a84U,0x5508U,0xaa10U,0x442bU,0x8856U,
		0x00a7U,0x014eU,0x029cU,0x0538U,0x0a70U,0x14e0U,0x29c0U,0x5380U,0xa700U,0x5e0bU,
		0xbc16U,0x6827U,0xd04eU,0xb097U,0x7125U,0xe24aU,0xd49fU,0xb935U,0x6261U,0xc4c2U,
		0x998fU,0x2315U,0x462aU,0x8c54U,0x08a3U,0x1146U,0x228cU,0x4518U,0x8a30U,0x046bU,
		0x08d6U,0x11acU,0x2358U,0x46b0U,0x8d60U,0x0acbU,0x1596U,0x2b2cU,0x5658U,0xacb0U,
		0x496bU,0x92d6U,0x35a7U,0x6b4eU,0xd69cU,0xbd33U,0x6a6dU,0xd4daU,0xb9bfU,0x6375U,
		0xc6eaU,0x9ddfU,0x2bb5U,0x576aU,0xaed4U,0x4da3U,0x9b46U,0x2687U,0x4d0eU,0x9a1cU,
		0x2433U,0x4866U,0x90ccU,0x3193U,0x6326U,0xc64cU,0x9c93U,0x292dU,0x525aU,0xa4b4U,
		0x5963U,0xb2c6U,0x7587U,0xeb0eU,0xc617U,0x9c25U,0x2841U,0x5082U,0xa104U,0x5203U,
		0xa406U,0x5807U,0xb00eU,0x7017U,0xe02eU,0xd057U,0xb0a5U,0x7141U,0xe282U,0xd50fU,
		0xba15U,0x6421U,0xc842U,0x808fU,0x1115U,0x222aU,0x4454U,0x88a8U,0x015bU,0x02b6U,
		0x056cU,0x0ad8U,0x15b0U,0x2b60U,0x56c0U,0xad80U,0x4b0bU,0x9616U,0x3c27U,0x784eU,
		0xf09cU,0xf133U,0xf26dU,0xf4d1U,0xf9a9U,0xe359U,0xd6b9U,0xbd79U,0x6af9U,0xd5f2U,
		0xbbefU,0x67d5U,0xcfaaU,0x8f5fU,0x0eb5U,0x1d6aU,0x3ad4U,0x75a8U,0xeb50U,0xc6abU,
		0x9d5dU,0x2ab1U,0x5562U,0xaac4U,0x4583U,0x8b06U,0x0607U,0x0c0eU,0x181cU,0x3038U,
		0x6070U,0xc0e0U,0x91cbU,0x339dU,0x673aU,0xce74U,0x8ce3U,0x09cdU,0x139aU,0x2734U,
		0x4e68U,0x9cd0U,0x29abU,0x5356U,0xa6acU,0x5d53U,0xbaa6U,0x6547U,0xca8eU,0x8517U,
		0x1a25U,0x344aU,0x6894U,0xd128U,0xb25bU,0x74bdU,0xe97aU,0xc2ffU,0x95f5U,0x3be1U,
		0x77c2U,0xef84U,0xcf03U,0x8e0dU,0x0c11U,0x1822U,0x3044U,0x6088U,0xc110U,0x922bU,
		0x345dU,0x68baU,0xd174U,0xb2e3U,0x75cdU,0xeb9aU,0xc73fU,0x9e75U,0x2ce1U,0x59c2U,
		0xb384U,0x7703U,0xee06U,0xcc07U,0x8805U,0x0001U,0x0002U,0x0004U,0x0008U,0x0010U,0x0020U,0x0040U,0x0080U,0x0100U,0x0200U,
		0x0400U,0x0800U,0x1000U,0x2000U,0x4000U,0x8000U,0x100bU,0x2016U,0x402cU,0x8058U,
		0x10bbU,0x2176U,0x42ecU,0x85d8U,0x1bbbU,0x3776U,0x6eecU,0xddd8U,0xabbbU,0x477dU,
		0x8efaU,0x0dffU,0x1bfeU,0x37fcU,0x6ff8U,0xdff0U,0xafebU,0x4fddU,0x9fbaU,0x2f7fU,
		0x5efeU,0xbdfcU,0x6bf3U,0xd7e6U,0xbfc7U,0x6f85U,0xdf0aU,0xae1fU,0x4c35U,0x986aU,
		0x20dfU,0x41beU,0x837cU,0x16f3U,0x2de6U,0x5bccU,0xb798U,0x7f3bU,0xfe76U,0xece7U,
		0xc9c5U,0x8381U,0x1709U,0x2e12U,0x5c24U,0xb848U,0x609bU,0xc136U,0x9267U,0x34c5U,
		0x698aU,0xd314U,0xb623U,0x7c4dU,0xf89aU,0xe13fU,0xd275U,0xb4e1U,0x79c9U,0xf392U,
		0xf72fU,0xfe55U,0xeca1U,0xc949U,0x8299U,0x1539U,0x2a72U,0x54e4U,0xa9c8U,0x439bU,
		0x8736U,0x1e67U,0x3cceU,0x799cU,0xf338U,0xf67bU,0xfcfdU,0xe9f1U,0xc3e9U,0x97d9U,
		0x3fb9U,0x7f72U,0xfee4U,0xedc3U,0xcb8dU,0x8711U,0x1e29U,0x3c52U,0x78a4U,0xf148U,
		0xf29bU,0xf53dU,0xfa71U,0xe4e9U,0xd9d9U,0xa3b9U,0x5779U,0xaef2U,0x4defU,0x9bdeU,
		0x27b7U,0x4f6eU,0x9edcU,0x2db3U,0x5b66U,0xb6ccU,0x7d93U,0xfb26U,0xe647U,0xdc85U,
		0xa901U,0x4209U,0x8412U,0x182fU,0x305eU,0x60bcU,0xc178U,0x92fbU,0x35fdU,0x6bfaU,
		0xd7f4U,0xbfe3U,0x6fcdU,0xdf9aU,0xaf3fU,0x4e75U,0x9ceaU,0x29dfU,0x53beU,0xa77cU,
		0x5ef3U,0xbde6U,0x6bc7U,0xd78eU,0xbf17U,0x6e25U,0xdc4aU,0xa89fU,0x4135U,0x826aU,
		0x14dfU,0x29beU,0x537cU,0xa6f8U,0x5dfbU,0xbbf6U,0x67e7U,0xcfceU,0x8f97U,0x0f25U,
		0x1e4aU,0x3c94U,0x7928U,0xf250U,0xf4abU,0xf95dU,0xe2b1U,0xd569U,0xbad9U,0x65b9U,
		0xcb72U,0x86efU,0x1dd5U,0x3baaU,0x7754U,0xeea8U,0xcd5bU,0x8abdU,0x0571U,0x0ae2U,
		0x15c4U,0x2b88U,0x5710U,0xae20U,0x4c4bU,0x9896U,0x2127U,0x424eU,0x849cU,0x1933U,
		0x3266U,0x64ccU,0xc998U,0x833bU,0x167dU,0x2cfaU,0x59f4U,0xb3e8U,0x77dbU,0xefb6U,
		0xcf67U,0x8ec5U,0x0d81U,0x1b02U,0x3604U,0x6c08U,0xd810U,0xa02bU,0x505dU,0xa0baU,
		0x517fU,0xa2feU,0x55f7U,0xabeeU,0x47d7U,0x8faeU,0x0f57U,0x1eaeU,0x3d5cU,0x7ab8U,
		0xf570U,0xfaebU,0xe5ddU,0xdbb1U,0xa769U,0x5ed9U,0xbdb2U,0x6b6fU,0xd6deU,0xbdb7U,
		0x6b65U,0xd6caU,0xbd9fU,0x6b35U,0xd66aU,0xbcdfU,0x69b5U,0xd36aU,0xb6dfU,0x7db5U,
		0xfb6aU,0xe6dfU,0xddb5U,0xab61U,0x46c9U,0x8d92U,0x0b2fU,0x165eU,0x2cbcU,0x5978U,
		0xb2f0U,0x75ebU,0xebd6U,0xc7a7U,0x9f45U,0x2e81U,0x5d02U,0xba04U,0x6403U,0xc806U,
		0x8007U,0x1005U,0x200aU,0x4014U,0x8028U,0x105bU,0x20b6U,0x416cU,0x82d8U,0x15bbU,
		0x2b76U,0x56ecU,0xadd8U,0x4bbbU,0x9776U,0x3ee7U,0x7dceU,0xfb9cU,0xe733U,0xde6dU,
		0xacd1U,0x49a9U,0x9352U,0x36afU,0x6d5eU,0xdabcU,0xa573U,0x5aedU,0xb5daU,0x7bbfU,
		0xf77eU,0xfef7U,0xede5U,0xcbc1U,0x8789U,0x1f19U,0x3e32U,0x7c64U,0xf8c8U,0xe19bU,
		0xd33dU,0xb671U,0x7ce9U,0xf9d2U,0xe3afU,0xd755U,0xbea1U,0x6d49U,0xda92U,0xa52fU,
		0x5a55U,0xb4aaU,0x795fU,0xf2beU,0xf577U,0xfae5U,0xe5c1U,0xdb89U,0xa719U,0x5e39U,
		0xbc72U,0x68efU,0xd1deU,0xb3b7U,0x7765U,0xeecaU,0xcd9fU,0x8b35U,0x0661U,0x0cc2U,
		0x1984U,0x3308U,0x6610U,0xcc20U,0x884bU,0x009dU,0x013aU,0x0274U,0x04e8U,0x09d0U,
		0x13a0U,0x2740U,0x4e80U,0x9d00U,0x2a0bU,0x5416U,0xa82cU,0x4053U,0x80a6U,0x1147U,
		0x228eU,0x451cU,0x8a38U,0x047bU,0x08f6U,0x11ecU,0x23d8U,0x47b0U,0x8f60U,0x0ecbU,
		0x1d96U,0x3b2cU,0x7658U,0xecb0U,0xc96bU,0x82ddU,0x15b1U,0x2b62U,0x56c4U,0xad88U,
		0x4b1bU,0x9636U,0x3c67U,0x78ceU,0xf19cU,0xf333U,0xf66dU,0xfcd1U,0xe9a9U,0xc359U,
		0x96b9U,0x3d79U,0x7af2U,0xf5e4U,0xfbc3U,0xe78dU,0xdf11U,0xae29U,0x4c59U,0x98b2U,
		0x216fU,0x42deU,0x85bcU,0x1b73U,0x36e6U,0x6dccU,0xdb98U,0xa73bU,0x5e7dU,0xbcfaU,
		0x69ffU,0xd3feU,0xb7f7U,0x7fe5U,0xffcaU,0xef9fU,0xcf35U,0x8e61U,0x0cc9U,0x1992U,
		0x3324U,0x6648U,0xcc90U,0x892bU,0x025dU,0x04baU,0x0974U,0x12e8U,0x25d0U,0x4ba0U,
		0x9740U,0x3e8bU,0x7d16U,0xfa2cU,0xe453U,0xd8adU,0xa151U,0x52a9U,0xa552U,0x5aafU,
		0xb55eU,0x7ab7U,0xf56eU,0xfad7U,0xe5a5U,0xdb41U,0xa689U,0x5d19U,0xba32U,0x646fU,
		0xc8deU,0x81b7U,0x1365U,0x26caU,0x4d94U,0x9b28U,0x265bU,0x4cb6U,0x996cU,0x22d3U,
		0x45a6U,0x8b4cU,0x0693U,0x0d26U,0x1a4cU,0x3498U,0x6930U,0xd260U,0xb4cbU,0x799dU,
		0xf33aU,0xf67fU,0xfcf5U,0xe9e1U,0xc3c9U,0x9799U,0x3f39U,0x7e72U,0xfce4U,0xe9c3U,
		0xc38dU,0x9711U,0x3e29U,0x7c52U,0xf8a4U,0xe143U,0xd28dU,0xb511U,0x7a29U,0xf452U,
		0xf8afU,0xe155U,0xd2a1U,0xb549U,0x7a99U,0xf532U,0xfa6fU,0xe4d5U,0xd9a1U,0xa349U,
		0x5699U,0xad32U,0x4a6fU,0x94deU,0x39b7U,0x736eU,0xe6dcU,0xddb3U,0xab6dU,0x46d1U,
		0x8da2U,0x0b4fU,0x169eU,0x2d3cU,0x5a78U,0xb4f0U,0x79ebU,0xf3d6U,0xf7a7U,0xff45U,
		0xee81U,0xcd09U,0x8a19U,0x0439U,0x0872U,0x10e4U,0x21c8U,0x4390U,0x8720U,0x1e4bU,
		0x3c96U,0x792cU,0xf258U,0xf4bbU,0xf97dU,0xe2f1U,0xd5e9U,0xbbd9U,0x67b9U,0xcf72U,
		0x8eefU,0x0dd5U,0x1baaU,0x3754U,0x6ea8U,0xdd50U,0xaaabU,0x455dU,0x8abaU,0x057fU,
		0x0afeU,0x15fcU,0x2bf8U,0x57f0U,0xafe0U,0x4fcbU,0x9f96U,0x2f27U,0x5e4eU,0xbc9cU,
		0x6933U,0xd266U,0xb4c7U,0x7985U,0xf30aU,0xf61fU,0xfc35U,0xe861U,0xc0c9U,0x9199U,
		0x3339U,0x6672U,0xcce4U,0x89c3U,0x038dU,0x071aU,0x0e34U,0x1c68U,0x38d0U,0x71a0U,
		0xe340U,0xd68bU,0xbd1dU,0x6a31U,0xd462U,0xb8cfU,0x6195U,0xc32aU,0x965fU,0x3cb5U,
		0x796aU,0xf2d4U,0xf5a3U,0xfb4dU,0xe691U,0xdd29U,0xaa59U,0x44b9U,0x8972U,0x02efU,
		0x05deU,0x0bbcU,0x1778U,0x2ef0U,0x5de0U,0xbbc0U,0x678bU,0xcf16U,0x8e27U,0x0c45U,
		0x188aU,0x3114U,0x6228U,0xc450U,0x98abU,0x215dU,0x42baU,0x8574U,0x1ae3U,0x35c6U,
		0x6b8cU,0xd718U,0xbe3bU,0x6c7dU,0xd8faU,0xa1ffU,0x53f5U,0xa7eaU,0x5fdfU,0xbfbeU,
		0x6f77U,0xdeeeU,0xadd7U,0x4ba5U,0x974aU,0x3e9fU,0x7d3eU,0xfa7cU,0xe4f3U,0xd9edU,
		0xa3d1U,0x57a9U,0xaf52U,0x4eafU,0x9d5eU,0x2ab7U,0x556eU,0xaadcU,0x45b3U,0x8b66U,
		0x06c7U,0x0d8eU,0x1b1cU,0x3638U,0x6c70U,0xd8e0U,0xa1cbU,0x539dU,0xa73aU,0x5e7fU,
		0xbcfeU,0x69f7U,0xd3eeU,0xb7d7U,0x7fa5U,0xff4aU,0xee9fU,0xcd35U,0x8a61U,0x04c9U,
		0x0992U,0x1324U,0x2648U,0x4c90U,0x9920U,0x224bU,0x4496U,0x892cU,0x0253U,0x04a6U,
		0x094cU,0x1298U,0x2530U,0x4a60U,0x94c0U,0x398bU,0x7316U,0xe62cU,0xdc53U,0xa8adU,
		0x4151U,0x82a2U,0x154fU,0x2a9eU,0x553cU,0xaa78U,0x44fbU,0x89f6U,0x03e7U,0x07ceU,
		0x0f9cU,0x1f38U,0x3e70U,0x7ce0U,0xf9c0U,0xe38bU,0xd71dU,0xbe31U,0x6c69U,0xd8d2U,
		0xa1afU,0x5355U,0xa6aaU,0x5d5fU,0xbabeU,0x6577U,0xcaeeU,0x85d7U,0x1ba5U,0x374aU,
		0x6e94U,0xdd28U,0xaa5bU,0x44bdU,0x897aU,0x02ffU,0x05feU,0x0bfcU,0x17f8U,0x2ff0U,
		0x5fe0U,0xbfc0U,0x6f8bU,0xdf16U,0xae27U,0x4c45U,0x988aU,0x211fU,0x423eU,0x847cU,
		0x18f3U,0x31e6U,0x63ccU,0xc798U,0x9f3bU,0x2e7dU,0x5cfaU,0xb9f4U,0x63e3U,0xc7c6U,
		0x9f87U,0x2f05U,0x5e0aU,0xbc14U,0x6823U,0xd046U,0xb087U,0x7105U,0xe20aU,0xd41fU,
		0xb835U,0x6061U,0xc0c2U,0x918fU,0x3315U,0x662aU,0xcc54U,0x88a3U,0x014dU,0x029aU,
		0x0534U,0x0a68U,0x14d0U,0x29a0U,0x5340U,0xa680U,0x5d0bU,0xba16U,0x6427U,0xc84eU,
		0x8097U,0x1125U,0x224aU,0x4494U,0x8928U,0x025bU,0x04b6U,0x096cU,0x12d8U,0x25b0U,
		0x4b60U,0x96c0U,0x3d8bU,0x7b16U,0xf62cU,0xfc53U,0xe8adU,0xc151U,0x92a9U,0x3559U,
		0x6ab2U,0xd564U,0xbac3U,0x658dU,0xcb1aU,0x863fU,0x1c75U,0x38eaU,0x71d4U,0xe3a8U,
		0xd75bU,0xbebdU,0x6d71U,0xdae2U,0xa5cfU,0x5b95U,0xb72aU,0x7e5fU,0xfcbeU,0xe977U,
		0xc2e5U,0x95c1U,0x3b89U,0x7712U,0xee24U,0xcc43U,0x888dU,0x0111U,0x0222U,0x0444U,
		0x0888U,0x1110U,0x2220U,0x4440U,0x8880U,0x010bU,0x0216U,0x042cU,0x0858U,0x10b0U,
		0x2160U,0x42c0U,0x8580U,0x1b0bU,0x3616U,0x6c2cU,0xd858U,0xa0bbU,0x517dU,0xa2faU,
		0x55ffU,0xabfeU,0x47f7U,0x8feeU,0x0fd7U,0x1faeU,0x3f5cU,0x7eb8U,0xfd70U,0xeaebU,
		0xc5ddU,0x9bb1U,0x2769U,0x4ed2U,0x9da4U,0x2b43U,0x5686U,0xad0cU,0x4a13U,0x9426U,
		0x3847U,0x708eU,0xe11cU,0xd233U,0xb46dU,0x78d1U,0xf1a2U,0xf34fU,0xf695U,0xfd21U,
		0xea49U,0xc499U,0x9939U,0x2279U,0x44f2U,0x89e4U,0x03c3U,0x0786U,0x0f0cU,0x1e18U,
		0x3c30U,0x7860U,0xf0c0U,0xf18bU,0xf31dU,0xf631U,0xfc69U,0xe8d9U,0xc1b9U,0x9379U,
		0x36f9U,0x6df2U,0xdbe4U,0xa7c3U,0x5f8dU,0xbf1aU,0x6e3fU,0xdc7eU,0xa8f7U,0x41e5U,
		0x83caU,0x179fU,0x2f3eU,0x5e7cU,0xbcf8U,0x69fbU,0xd3f6U,0xb7e7U,0x7fc5U,0xff8aU,
		0xef1fU,0xce35U,0x8c61U,0x08c9U,0x1192U,0x2324U,0x4648U,0x8c90U,0x092bU,0x1256U,
		0x24acU,0x4958U,0x92b0U,0x356bU,0x6ad6U,0xd5acU,0xbb53U,0x66adU,0xcd5aU,0x8abfU,
		0x0575U,0x0aeaU,0x15d4U,0x2ba8U,0x5750U,0xaea0U,0x4d4bU,0x9a96U,0x2527U,0x4a4eU,
		0x949cU,0x3933U,0x7266U,0xe4ccU,0xd993U,0xa32dU,0x5651U,0xaca2U,0x494fU,0x929eU,
		0x3537U,0x6a6eU,0xd4dcU,0xb9b3U,0x636dU,0xc6daU,0x9dbfU,0x2b75U,0x56eaU,0xadd4U,
		0x4ba3U,0x9746U,0x3e87U,0x7d0eU,0xfa1cU,0xe433U,0xd86dU,0xa0d1U,0x51a9U,0xa352U,
		0x56afU,0xad5eU,0x4ab7U,0x956eU,0x3ad7U,0x75aeU,0xeb5cU,0xc6b3U,0x9d6dU,0x2ad1U,
		0x55a2U,0xab44U,0x4683U,0x8d06U,0x0a07U,0x140eU,0x281cU,0x5038U,0xa070U,0x50ebU,
		0xa1d6U,0x53a7U,0xa74eU,0x5e97U,0xbd2eU,0x6a57U,0xd4aeU,0xb957U,0x62a5U,0xc54aU,
		0x9a9fU,0x2535U,0x4a6aU,0x94d4U,0x39a3U,0x7346U,0xe68cU,0xdd13U,0xaa2dU,0x4451U,
		0x88a2U,0x014fU,0x029eU,0x053cU,0x0a78U,0x14f0U,0x29e0U,0x53c0U,0xa780U,0x5f0bU,
		0xbe16U,0x6c27U,0xd84eU,0xa097U,0x5125U,0xa24aU,0x549fU,0xa93eU,0x4277U,0x84eeU,
		0x19d7U,0x33aeU,0x675cU,0xceb8U,0x8d7bU,0x0afdU,0x15faU,0x2bf4U,0x57e8U,0xafd0U,
		0x4fabU,0x9f56U,0x2ea7U,0x5d4eU,0xba9cU,0x6533U,0xca66U,0x84c7U,0x1985U,0x330aU,
		0x6614U,0xcc28U,0x885bU,0x00bdU,0x017aU,0x02f4U,0x05e8U,0x0bd0U,0x17a0U,0x2f40U,
		0x5e80U,0xbd00U,0x6a0bU,0xd416U,0xb827U,0x6045U,0xc08aU,0x911fU,0x3235U,0x646aU,
		0xc8d4U,0x81a3U,0x134dU,0x269aU,0x4d34U,0x9a68U,0x24dbU,0x49b6U,0x936cU,0x36d3U,
		0x6da6U,0xdb4cU,0xa693U,0x5d2dU,0xba5aU,0x64bfU,0xc97eU,0x82f7U,0x15e5U,0x2bcaU,
		0x5794U,0xaf28U,0x4e5bU,0x9cb6U,0x2967U,0x52ceU,0xa59cU,0x5b33U,0xb666U,0x7cc7U,
		0xf98eU,0xe317U,0xd625U,0xbc41U,0x6889U,0xd112U,0xb22fU,0x7455U,0xe8aaU,0xc15fU,
		0x92b5U,0x3561U,0x6ac2U,0xd584U,0xbb03U,0x660dU,0xcc1aU,0x883fU,0x0075U,0x00eaU,
		0x01d4U,0x03a8U,0x0750U,0x0ea0U,0x1d40U,0x3a80U,0x7500U,0xea00U,0xc40bU,0x981dU,
		0x2031U,0x4062U,0x80c4U,0x1183U,0x2306U,0x460cU,0x8c18U,0x083bU,0x1076U,0x20ecU,
		0x41d8U,0x83b0U,0x176bU,0x2ed6U,0x5dacU,0xbb58U,0x66bbU,0xcd76U,0x8ae7U,0x05c5U,
		0x0b8aU,0x1714U,0x2e28U,0x5c50U,0xb8a0U,0x614bU,0xc296U,0x9527U,0x3a45U,0x748aU,
		0xe914U,0xc223U,0x944dU,0x3891U,0x7122U,0xe244U,0xd483U,0xb90dU,0x6211U,0xc422U,
		0x984fU,0x2095U,0x412aU,0x8254U,0x14a3U,0x2946U,0x528cU,0xa518U,0x5a3bU,0xb476U,
		0x78e7U,0xf1ceU,0xf397U,0xf725U,0xfe41U,0xec89U,0xc919U,0x8239U,0x1479U,0x28f2U,
		0x51e4U,0xa3c8U,0x579bU,0xaf36U,0x4e67U,0x9cceU,0x2997U,0x532eU,0xa65cU,0x5cb3U,
		0xb966U,0x62c7U,0xc58eU,0x9b17U,0x2625U,0x4c4aU,0x9894U,0x2123U,0x4246U,0x848cU,
		0x1913U,0x3226U,0x644cU,0xc898U,0x813bU,0x127dU,0x24faU,0x49f4U,0x93e8U,0x37dbU,
		0x6fb6U,0xdf6cU,0xaed3U,0x4dadU,0x9b5aU,0x26bfU,0x4d7eU,0x9afcU,0x25f3U,0x4be6U,
		0x97ccU,0x3f93U,0x7f26U,0xfe4cU,0xec93U,0xc92dU,0x8251U,0x14a9U,0x2952U,0x52a4U,
		0xa548U,0x5a9bU,0xb536U,0x7a67U,0xf4ceU,0xf997U,0xe325U,0xd641U,0xbc89U,0x6919U,
		0xd232U,0xb46fU,0x78d5U,0xf1aaU,0xf35fU,0xf6b5U,0xfd61U,0xeac9U,0xc599U,0x9b39U,
		0x2679U,0x4cf2U,0x99e4U,0x23c3U,0x4786U,0x8f0cU,0x0e13U,0x1c26U,0x384cU,0x7098U,
		0xe130U,0xd26bU,0xb4ddU,0x79b1U,0xf362U,0xf6cfU,0xfd95U,0xeb21U,0xc649U,0x9c99U,
		0x2939U,0x5272U,0xa4e4U,0x59c3U,0xb386U,0x7707U,0xee0eU,0xcc17U,0x8825U,0x0041U,
		0x0082U,0x0104U,0x0208U,0x0410U,0x0820U,0x1040U,0x2080U,0x4100U,0x8200U,0x140bU,
		0x2816U,0x502cU,0xa058U,0x50bbU,0xa176U,0x52e7U,0xa5ceU,0x5b97U,0xb72eU,0x7e57U,
		0xfcaeU,0xe957U,0xc2a5U,0x9541U,0x3a89U,0x7512U,0xea24U,0xc443U,0x988dU,0x2111U,
		0x4222U,0x8444U,0x1883U,0x3106U,0x620cU,0xc418U,0x983bU,0x207dU,0x40faU,0x81f4U,
		0x13e3U,0x27c6U,0x4f8cU,0x9f18U,0x2e3bU,0x5c76U,0xb8ecU,0x61d3U,0xc3a6U,0x9747U,
		0x3e85U,0x7d0aU,0xfa14U,0xe423U,0xd84dU,0xa091U,0x5129U,0xa252U,0x54afU,0xa95eU,
		0x42b7U,0x856eU,0x1ad7U,0x35aeU,0x6b5cU,0xd6b8U,0xbd7bU,0x6afdU,0xd5faU,0xbbffU,
		0x67f5U,0xcfeaU,0x8fdfU,0x0fb5U,0x1f6aU,0x3ed4U,0x7da8U,0xfb50U,0xe6abU,0xdd5dU,
		0xaab1U,0x4569U,0x8ad2U,0x05afU,0x0b5eU,0x16bcU,0x2d78U,0x5af0U,0xb5e0U,0x7bcbU,
		0xf796U,0xff27U,0xee45U,0xcc81U,0x8909U,0x0219U,0x0432U,0x0864U,0x10c8U,0x2190U,
		0x4320U,0x8640U,0x1c8bU,0x3916U,0x722cU,0xe458U,0xd8bbU,0xa17dU,0x52f1U,0xa5e2U,
		0x5bcfU,0xb79eU,0x7f37U,0xfe6eU,0xecd7U,0xc9a5U,0x8341U,0x1689U,0x2d12U,0x5a24U,
		0xb448U,0x789bU,0xf136U,0xf267U,0xf4c5U,0xf981U,0xe309U,0xd619U,0xbc39U,0x6879U,
		0xd0f2U,0xb1efU,0x73d5U,0xe7aaU,0xdf5fU,0xaeb5U,0x4d61U,0x9ac2U,0x258fU,0x4b1eU,
		0x963cU,0x3c73U,0x78e6U,0xf1ccU,0xf393U,0xf72dU,0xfe51U,0xeca9U,0xc959U,0x82b9U,
		0x1579U,0x2af2U,0x55e4U,0xabc8U,0x479bU,0x8f36U,0x0e67U,0x1cceU,0x399cU,0x7338U,
		0xe670U,0xdcebU,0xa9ddU,0x43b1U,0x8762U,0x1ecfU,0x3d9eU,0x7b3cU,0xf678U,0xfcfbU,
		0xe9fdU,0xc3f1U,0x97e9U,0x3fd9U,0x7fb2U,0xff64U,0xeec3U,0xcd8dU,0x8b11U,0x0629U,
		0x0c52U,0x18a4U,0x3148U,0x6290U,0xc520U,0x9a4bU,0x249dU,0x493aU,0x9274U,0x34e3U,
		0x69c6U,0xd38cU,0xb713U,0x7e2dU,0xfc5aU,0xe8bfU,0xc175U,0x92e1U,0x35c9U,0x6b92U,
		0xd724U,0xbe43U,0x6c8dU,0xd91aU,0xa23fU,0x5475U,0xa8eaU,0x41dfU,0x83beU,0x1777U,
		0x2eeeU,0x5ddcU,0xbbb8U,0x677bU,0xcef6U,0x8de7U,0x0bc5U,0x178aU,0x2f14U,0x5e28U,
		0xbc50U,0x68abU,0xd156U,0xb2a7U,0x7545U,0xea8aU,0xc51fU,0x9a35U,0x2461U,0x48c2U,
		0x9184U,0x3303U,0x6606U,0xcc0cU,0x8813U,0x002dU,0x005aU,0x00b4U,0x0168U,0x02d0U,
		0x05a0U,0x0b40U,0x1680U,0x2d00U,0x5a00U,0xb400U,0x780bU,0xf016U,0xf027U,0xf045U,
		0xf081U,0xf109U,0xf219U,0xf439U,0xf879U,0xe0f9U,0xd1f9U,0xb3f9U,0x77f9U,0xeff2U,
		0xcfefU,0x8fd5U,0x0fa1U,0x1f42U,0x3e84U,0x7d08U,0xfa10U,0xe42bU,0xd85dU,0xa0b1U,
		0x5169U,0xa2d2U,0x55afU,0xab5eU,0x46b7U,0x8d6eU,0x0ad7U,0x15aeU,0x2b5cU,0x56b8U,
		0xad70U,0x4aebU,0x95d6U,0x3ba7U,0x774eU,0xee9cU,0xcd33U,0x8a6dU,0x04d1U,0x09a2U,
		0x1344U,0x2688U,0x4d10U,0x9a20U,0x244bU,0x4896U,0x912cU,0x3253U,0x64a6U,0xc94cU,
		0x8293U,0x152dU,0x2a5aU,0x54b4U,0xa968U,0x42dbU,0x85b6U,0x1b67U,0x36ceU,0x6d9cU,
		0xdb38U,0xa67bU,0x5cfdU,0xb9faU,0x63ffU,0xc7feU,0x9ff7U,0x2fe5U,0x5fcaU,0xbf94U,
		0x6f23U,0xde46U,0xac87U,0x4905U,0x920aU,0x341fU,0x683eU,0xd07cU,0xb0f3U,0x71edU,
		0xe3daU,0xd7bfU,0xbf75U,0x6ee1U,0xddc2U,0xab8fU,0x4715U,0x8e2aU,0x0c5fU,0x18beU,
		0x317cU,0x62f8U,0xc5f0U,0x9bebU,0x27ddU,0x4fbaU,0x9f74U,0x2ee3U,0x5dc6U,0xbb8cU,
		0x6713U,0xce26U,0x8c47U,0x0885U,0x110aU,0x2214U,0x4428U,0x8850U,0x00abU,0x0156U,
		0x02acU,0x0558U,0x0ab0U,0x1560U,0x2ac0U,0x5580U,0xab00U,0x460bU,0x8c16U,0x0827U,
		0x104eU,0x209cU,0x4138U,0x8270U,0x14ebU,0x29d6U,0x53acU,0xa758U,0x5ebbU,0xbd76U,
		0x6ae7U,0xd5ceU,0xbb97U,0x6725U,0xce4aU,0x8c9fU,0x0935U,0x126aU,0x24d4U,0x49a8U,
		0x9350U,0x36abU,0x6d56U,0xdaacU,0xa553U,0x5aadU,0xb55aU,0x7abfU,0xf57eU,0xfaf7U,
		0xe5e5U,0xdbc1U,0xa789U,0x5f19U,0xbe32U,0x6c6fU,0xd8deU,0xa1b7U,0x5365U,0xa6caU,
		0x5d9fU,0xbb3eU,0x6677U,0xcceeU,0x89d7U,0x03a5U,0x074aU,0x0e94U,0x1d28U,0x3a50U,
		0x74a0U,0xe940U,0xc28bU,0x951dU,0x3a31U,0x7462U,0xe8c4U,0xc183U,0x930dU,0x3611U,
		0x6c22U,0xd844U,0xa083U,0x510dU,0xa21aU,0x543fU,0xa87eU,0x40f7U,0x81eeU,0x13d7U,
		0x27aeU,0x4f5cU,0x9eb8U,0x2d7bU,0x5af6U,0xb5ecU,0x7bd3U,0xf7a6U,0xff47U,0xee85U,
		0xcd01U,0x8a09U,0x0419U,0x0832U,0x1064U,0x20c8U,0x4190U,0x8320U,0x164bU,0x2c96U,
		0x592cU,0xb258U,0x74bbU,0xe976U,0xc2e7U,0x95c5U,0x3b81U,0x7702U,0xee04U,0xcc03U,
		0x880dU,0x0011U,0x0022U,0x0044U,0x0088U,0x0110U,0x0220U,0x0440U,0x0880U,0x1100U,
		0x2200U,0x4400U,0x8800U,0x000bU,0x0016U,0x002cU,0x0058U,0x00b0U,0x0160U,0x02c0U,
		0x0580U,0x0b00U,0x1600U,0x2c00U,0x5800U,0xb000U,0x700bU,0xe016U,0xd027U,0xb045U,
		0x7081U,0xe102U,0xd20fU,0xb415U,0x7821U,0xf042U,0xf08fU,0xf115U,0xf221U,0xf449U,
		0xf899U,0xe139U,0xd279U,0xb4f9U,0x79f9U,0xf3f2U,0xf7efU,0xffd5U,0xefa1U,0xcf49U,
		0x8e99U,0x0d39U,0x1a72U,0x34e4U,0x69c8U,0xd390U,0xb72bU,0x7e5dU,0xfcbaU,0xe97fU,
		0xc2f5U,0x95e1U,0x3bc9U,0x7792U,0xef24U,0xce43U,0x8c8dU,0x0911U,0x1222U,0x2444U,
		0x4888U,0x9110U,0x322bU,0x6456U,0xc8acU,0x8153U,0x12adU,0x255aU,0x4ab4U,0x9568U,
		0x3adbU,0x75b6U,0xeb6cU,0xc6d3U,0x9dadU,0x2b51U,0x56a2U,0xad44U,0x4a83U,0x9506U,
		0x3a07U,0x740eU,0xe81cU,0xc033U,0x906dU,0x30d1U,0x61a2U,0xc344U,0x9683U,0x3d0dU,
		0x7a1aU,0xf434U,0xf863U,0xe0cdU,0xd191U,0xb329U,0x7659U,0xecb2U,0xc96fU,0x82d5U,
		0x15a1U,0x2b42U,0x5684U,0xad08U,0x4a1bU,0x9436U,0x3867U,0x70ceU,0xe19cU,0xd333U,
		0xb66dU,0x7cd1U,0xf9a2U,0xe34fU,0xd695U,0xbd21U,0x6a49U,0xd492U,0xb92fU,0x6255U,
		0xc4aaU,0x995fU,0x22b5U,0x456aU,0x8ad4U,0x05a3U,0x0b46U,0x168cU,0x2d18U,0x5a30U,
		0xb460U,0x78cbU,0xf196U,0xf327U,0xf645U,0xfc81U,0xe909U,0xc219U,0x9439U,0x3879U,
		0x70f2U,0xe1e4U,0xd3c3U,0xb78dU,0x7f11U,0xfe22U,0xec4fU,0xc895U,0x8121U,0x1249U,
		0x2492U,0x4924U,0x9248U,0x349bU,0x6936U,0xd26cU,0xb4d3U,0x79adU,0xf35aU,0xf6bfU,
		0xfd75U,0xeae1U,0xc5c9U,0x9b99U,0x2739U,0x4e72U,0x9ce4U,0x29c3U,0x5386U,0xa70cU,
		0x5e13U,0xbc26U,0x6847U,0xd08eU,0xb117U,0x7225U,0xe44aU,0xd89fU,0xa135U,0x5261U,
		0xa4c2U,0x598fU,0xb31eU,0x7637U,0xec6eU,0xc8d7U,0x81a5U,0x1341U,0x2682U,0x4d04U,
		0x9a08U,0x241bU,0x4836U,0x906cU,0x30d3U,0x61a6U,0xc34cU,0x9693U,0x3d2dU,0x7a5aU,
		0xf4b4U,0xf963U,0xe2cdU,0xd591U,0xbb29U,0x6659U,0xccb2U,0x896fU,0x02d5U,0x05aaU,
		0x0b54U,0x16a8U,0x2d50U,0x5aa0U,0xb540U,0x7a8bU,0xf516U,0xfa27U,0xe445U,0xd881U,
		0xa109U,0x5219U,0xa432U,0x586fU,0xb0deU,0x71b7U,0xe36eU,0xd6d7U,0xbda5U,0x6b41U,
		0xd682U,0xbd0fU,0x6a15U,0xd42aU,0xb85fU,0x60b5U,0xc16aU,0x92dfU,0x35b5U,0x6b6aU,
		0xd6d4U,0xbda3U,0x6b4dU,0xd69aU,0xbd3fU,0x6a75U,0xd4eaU,0xb9dfU,0x63b5U,0xc76aU,
		0x9edfU,0x2db5U,0x5b6aU,0xb6d4U,0x7da3U,0xfb46U,0xe687U,0xdd05U,0xaa01U,0x4409U,
		0x8812U,0x002fU,0x005eU,0x00bcU,0x0178U,0x02f0U,0x05e0U,0x0bc0U,0x1780U,0x2f00U,
		0x5e00U,0xbc00U,0x680bU,0xd016U,0xb027U,0x7045U,0xe08aU,0xd11fU,0xb235U,0x7461U,
		0xe8c2U,0xc18fU,0x9315U,0x3621U,0x6c42U,0xd884U,0xa103U,0x520dU,0xa41aU,0x583fU,
		0xb07eU,0x70f7U,0xe1eeU,0xd3d7U,0xb7a5U,0x7f41U,0xfe82U,0xed0fU,0xca15U,0x8421U,
		0x1849U,0x3092U,0x6124U,0xc248U,0x949bU,0x393dU,0x727aU,0xe4f4U,0xd9e3U,0xa3cdU,
		0x5791U,0xaf22U,0x4e4fU,0x9c9eU,0x2937U,0x526eU,0xa4dcU,0x59b3U,0xb366U,0x76c7U,
		0xed8eU,0xcb17U,0x8625U,0x1c41U,0x3882U,0x7104U,0xe208U,0xd41bU,0xb83dU,0x6071U,
		0xc0e2U,0x91cfU,0x3395U,0x672aU,0xce54U,0x8ca3U,0x094dU,0x129aU,0x2534U,0x4a68U,
		0x94d0U,0x39abU,0x7356U,0xe6acU,0xdd53U,0xaaadU,0x4551U,0x8aa2U,0x054fU,0x0a9eU,
		0x153cU,0x2a78U,0x54f0U,0xa9e0U,0x43cbU,0x8796U,0x1f27U,0x3e4eU,0x7c9cU,0xf938U,
		0xe27bU,0xd4fdU,0xb9f1U,0x63e9U,0xc7d2U,0x9fafU,0x2f55U,0x5eaaU,0xbd54U,0x6aa3U,
		0xd546U,0xba87U,0x6505U,0xca0aU,0x841fU,0x1835U,0x306aU,0x60d4U,0xc1a8U,0x935bU,
		0x36bdU,0x6d7aU,0xdaf4U,0xa5e3U,0x5bcdU,0xb79aU,0x7f3fU,0xfe7eU,0xecf7U,0xc9e5U,
		0x83c1U,0x1789U,0x2f12U,0x5e24U,0xbc48U,0x689bU,0xd136U,0xb267U,0x74c5U,0xe98aU,
		0xc31fU,0x9635U,0x3c61U,0x78c2U,0xf184U,0xf303U,0xf60dU,0xfc11U,0xe829U,0xc059U,
		0x90b9U,0x3179U,0x62f2U,0xc5e4U,0x9bc3U,0x278dU,0x4f1aU,0x9e34U,0x2c63U,0x58c6U,
		0xb18cU,0x7313U,0xe626U,0xdc47U,0xa885U,0x4101U,0x8202U,0x140fU,0x281eU,0x503cU,
		0xa078U,0x50fbU,0xa1f6U,0x53e7U,0xa7ceU,0x5f97U,0xbf2eU,0x6e57U,0xdcaeU,0xa957U,
		0x42a5U,0x854aU,0x1a9fU,0x353eU,0x6a7cU,0xd4f8U,0xb9fbU,0x63fdU,0xc7faU,0x9fffU,
		0x2ff5U,0x5feaU,0xbfd4U,0x6fa3U,0xdf46U,0xae87U,0x4d05U,0x9a0aU,0x241fU,0x483eU,
		0x907cU,0x30f3U,0x61e6U,0xc3ccU,0x9793U,0x3f2dU,0x7e5aU,0xfcb4U,0xe963U,0xc2cdU,
		0x9591U,0x3b29U,0x7652U,0xeca4U,0xc943U,0x828dU,0x1511U,0x2a22U,0x5444U,0xa888U,
		0x411bU,0x8236U,0x1467U,0x28ceU,0x519cU,0xa338U,0x567bU,0xacf6U,0x49e7U,0x93ceU,
		0x3797U,0x6f2eU,0xde5cU,0xacb3U,0x496dU,0x92daU,0x35bfU,0x6b7eU,0xd6fcU,0xbdf3U,
		0x6bedU,0xd7daU,0xbfbfU,0x6f75U,0xdeeaU,0xaddfU,0x4bb5U,0x976aU,0x3edfU,0x7dbeU,
		0xfb7cU,0xe6f3U,0xddedU,0xabd1U,0x47a9U,0x8f52U,0x0eafU,0x1d5eU,0x3abcU,0x7578U,
		0xeaf0U,0xc5ebU,0x9bddU,0x27b1U,0x4f62U,0x9ec4U,0x2d83U,0x5b06U,0xb60cU,0x7c13U,
		0xf826U,0xe047U,0xd085U,0xb101U,0x7209U,0xe412U,0xd82fU,0xa055U,0x50a1U,0xa142U,
		0x528fU,0xa51eU,0x5a37U,0xb46eU,0x78d7U,0xf1aeU,0xf357U,0xf6a5U,0xfd41U,0xea89U,
		0xc519U,0x9a39U,0x2479U,0x48f2U,0x91e4U,0x33c3U,0x6786U,0xcf0cU,0x8e13U,0x0c2dU,
		0x185aU,0x30b4U,0x6168U,0xc2d0U,0x95abU,0x3b5dU,0x76baU,0xed74U,0xcae3U,0x85cdU,
		0x1b91U,0x3722U,0x6e44U,0xdc88U,0xa91bU,0x423dU,0x847aU,0x18ffU,0x31feU,0x63fcU,
		0xc7f8U,0x9ffbU,0x2ffdU,0x5ffaU,0xbff4U,0x6fe3U,0xdfc6U,0xaf87U,0x4f05U,0x9e0aU,
		0x2c1fU,0x583eU,0xb07cU,0x70f3U,0xe1e6U,0xd3c7U,0xb785U,0x7f01U,0xfe02U,0xec0fU,
		0xc815U,0x8021U,0x1049U,0x2092U,0x4124U,0x8248U,0x149bU,0x2936U,0x526cU,0xa4d8U,
		0x59bbU,0xb376U,0x76e7U,0xedceU,0xcb97U,0x8725U,0x1e41U,0x3c82U,0x7904U,0xf208U,
		0xf41bU,0xf83dU,0xe071U,0xd0e9U,0xb1d9U,0x73b9U,0xe772U,0xdeefU,0xadd5U,0x4ba1U,
		0x9742U,0x3e8fU,0x7d1eU,0xfa3cU,0xe473U,0xd8edU,0xa1d1U,0x53a9U,0xa752U,0x5eafU,
		0xbd5eU,0x6ab7U,0xd56eU,0xbad7U,0x65a5U,0xcb4aU,0x869fU,0x1d35U,0x3a6aU,0x74d4U,
		0xe9a8U,0xc35bU,0x96bdU,0x3d71U,0x7ae2U,0xf5c4U,0xfb83U,0xe70dU,0xde11U,0xac29U,
		0x4859U,0x90b2U,0x316fU,0x62deU,0xc5bcU,0x9b73U,0x26edU,0x4ddaU,0x9bb4U,0x2763U,
		0x4ec6U,0x9d8cU,0x2b13U,0x5626U,0xac4cU,0x4893U,0x9126U,0x3247U,0x648eU,0xc91cU,
		0x8233U,0x146dU,0x28daU,0x51b4U,0xa368U,0x56dbU,0xadb6U,0x4b67U,0x96ceU,0x3d97U,
		0x7b2eU,0xf65cU,0xfcb3U,0xe96dU,0xc2d1U,0x95a9U,0x3b59U,0x76b2U,0xed64U,0xcac3U,
		0x858dU,0x1b11U,0x3622U,0x6c44U,0xd888U,0xa11bU,0x523dU,0xa47aU,0x58ffU,0xb1feU,
		0x73f7U,0xe7eeU,0xdfd7U,0xafa5U,0x4f41U,0x9e82U,0x2d0fU,0x5a1eU,0xb43cU,0x7873U,
		0xf0e6U,0xf1c7U,0xf385U,0xf701U,0xfe09U,0xec19U,0xc839U,0x8079U,0x10f9U,0x21f2U,
		0x43e4U,0x87c8U,0x1f9bU,0x3f36U,0x7e6cU,0xfcd8U,0xe9bbU,0xc37dU,0x96f1U,0x3de9U,
		0x7bd2U,0xf7a4U,0xff43U,0xee8dU,0xcd11U,0x8a29U,0x0459U,0x08b2U,0x1164U,0x22c8U,
		0x4590U,0x8b20U,0x064bU,0x0c96U,0x192cU,0x3258U,0x64b0U,0xc960U,0x82cbU,0x159dU,
		0x2b3aU,0x5674U,0xace8U,0x49dbU,0x93b6U,0x3767U,0x6eceU,0xdd9cU,0xab33U,0x466dU,
		0x8cdaU,0x09bfU,0x137eU,0x26fcU,0x4df8U,0x9bf0U,0x27ebU,0x4fd6U,0x9facU,0x2f53U,
		0x5ea6U,0xbd4cU,0x6a93U,0xd526U,0xba47U,0x6485U,0xc90aU,0x821fU,0x1435U,0x286aU,
		0x50d4U,0xa1a8U,0x535bU,0xa6b6U,0x5d67U,0xbaceU,0x6597U,0xcb2eU,0x8657U,0x1ca5U,
		0x394aU,0x7294U,0xe528U,0xda5bU,0xa4bdU,0x5971U,0xb2e2U,0x75cfU,0xeb9eU,0xc737U,
		0x9e65U,0x2cc1U,0x5982U,0xb304U,0x7603U,0xec06U,0xc807U,0x8005U,0x1001U,0x2002U,
		0x4004U,0x8008U,0x101bU,0x2036U,0x406cU,0x80d8U,0x11bbU,0x2376U,0x46ecU,0x8dd8U,
		0x0bbbU,0x1776U,0x2eecU,0x5dd8U,0xbbb0U,0x676bU,0xced6U,0x8da7U,0x0b45U,0x168aU,
		0x2d14U,0x5a28U,0xb450U,0x78abU,0xf156U,0xf2a7U,0xf545U,0xfa81U,0xe509U,0xda19U,
		0xa439U,0x5879U,0xb0f2U,0x71efU,0xe3deU,0xd7b7U,0xbf65U,0x6ec1U,0xdd82U,0xab0fU,
		0x4615U,0x8c2aU,0x085fU,0x10beU,0x217cU,0x42f8U,0x85f0U,0x1bebU,0x37d6U,0x6facU,
		0xdf58U,0xaebbU,0x4d7dU,0x9afaU,0x25ffU,0x4bfeU,0x97fcU,0x3ff3U,0x7fe6U,0xffccU,
		0xef93U,0xcf2dU,0x8e51U,0x0ca9U,0x1952U,0x32a4U,0x6548U,0xca90U,0x852bU,0x1a5dU,
		0x34baU,0x6974U,0xd2e8U,0xb5dbU,0x7bbdU,0xf77aU,0xfeffU,0xedf5U,0xcbe1U,0x87c9U,
		0x1f99U,0x3f32U,0x7e64U,0xfcc8U,0xe99bU,0xc33dU,0x9671U,0x3ce9U,0x79d2U,0xf3a4U,
		0xf743U,0xfe8dU,0xed11U,0xca29U,0x8459U,0x18b9U,0x3172U,0x62e4U,0xc5c8U,0x9b9bU,
		0x273dU,0x4e7aU,0x9cf4U,0x29e3U,0x53c6U,0xa78cU,0x5f13U,0xbe26U,0x6c47U,0xd88eU,
		0xa117U,0x5225U,0xa44aU,0x589fU,0xb13eU,0x7277U,0xe4eeU,0xd9d7U,0xa3a5U,0x5741U,
		0xae82U,0x4d0fU,0x9a1eU,0x2437U,0x486eU,0x90dcU,0x31b3U,0x6366U,0xc6ccU,0x9d93U,
		0x2b2dU,0x565aU,0xacb4U,0x4963U,0x92c6U,0x3587U,0x6b0eU,0xd61cU,0xbc33U,0x686dU,
		0xd0daU,0xb1bfU,0x7375U,0xe6eaU,0xdddfU,0xabb5U,0x4761U,0x8ec2U,0x0d8fU,0x1b1eU,
		0x363cU,0x6c78U,0xd8f0U,0xa1ebU,0x53ddU,0xa7baU,0x5f7fU,0xbefeU,0x6df7U,0xdbeeU,
		0xa7d7U,0x5fa5U,0xbf4aU,0x6e9fU,0xdd3eU,0xaa77U,0x44e5U,0x89caU,0x039fU,0x073eU,
		0x0e7cU,0x1cf8U,0x39f0U,0x73e0U,0xe7c0U,0xdf8bU,0xaf1dU,0x4e31U,0x9c62U,0x28cfU,
		0x519eU,0xa33cU,0x5673U,0xace6U,0x49c7U,0x938eU,0x3717U,0x6e2eU,0xdc5cU,0xa8b3U,
		0x416dU,0x82daU,0x15bfU,0x2b7eU,0x56fcU,0xadf8U,0x4bfbU,0x97f6U,0x3fe7U,0x7fceU,
		0xff9cU,0xef33U,0xce6dU,0x8cd1U,0x09a9U,0x1352U,0x26a4U,0x4d48U,0x9a90U,0x252bU,
		0x4a56U,0x94acU,0x3953U,0x72a6U,0xe54cU,0xda93U,0xa52dU,0x5a51U,0xb4a2U,0x794fU,
		0xf29eU,0xf537U,0xfa65U,0xe4c1U,0xd989U,0xa319U,0x5639U,0xac72U,0x48efU,0x91deU,
		0x33b7U,0x676eU,0xcedcU,0x8db3U,0x0b6dU,0x16daU,0x2db4U,0x5b68U,0xb6d0U,0x7dabU,
		0xfb56U,0xe6a7U,0xdd45U,0xaa81U,0x4509U,0x8a12U,0x042fU,0x085eU,0x10bcU,0x2178U,
		0x42f0U,0x85e0U,0x1bcbU,0x3796U,0x6f2cU,0xde58U,0xacbbU,0x497dU,0x92faU,0x35ffU,
		0x6bfeU,0xd7fcU,0xbff3U,0x6fedU,0xdfdaU,0xafbfU,0x4f75U,0x9eeaU,0x2ddfU,0x5bbeU,
		0xb77cU,0x7ef3U,0xfde6U,0xebc7U,0xc785U,0x9f01U,0x2e09U,0x5c12U,0xb824U,0x6043U,
		0xc086U,0x9107U,0x3205U,0x640aU,0xc814U,0x8023U,0x104dU,0x209aU,0x4134U,0x8268U,
		0x14dbU,0x29b6U,0x536cU,0xa6d8U,0x5dbbU,0xbb76U,0x66e7U,0xcdceU,0x8b97U,0x0725U,
		0x0e4aU,0x1c94U,0x3928U,0x7250U,0xe4a0U,0xd94bU,0xa29dU,0x5531U,0xaa62U,0x44cfU,
		0x899eU,0x0337U,0x066eU,0x0cdcU,0x19b8U,0x3370U,0x66e0U,0xcdc0U,0x8b8bU,0x071dU,
		0x0e3aU,0x1c74U,0x38e8U,0x71d0U,0xe3a0U,0xd74bU,0xbe9dU,0x6d31U,0xda62U,0xa4cfU,
		0x5995U,0xb32aU,0x765fU,0xecbeU,0xc977U,0x82e5U,0x15c1U,0x2b82U,0x5704U,0xae08U,
		0x4c1bU,0x9836U,0x2067U,0x40ceU,0x819cU,0x1333U,0x2666U,0x4cccU,0x9998U,0x233bU,
		0x4676U,0x8cecU,0x09d3U,0x13a6U,0x274cU,0x4e98U,0x9d30U,0x2a6bU,0x54d6U,0xa9acU,
		0x4353U,0x86a6U,0x1d47U,0x3a8eU,0x751cU,0xea38U,0xc47bU,0x98fdU,0x21f1U,0x43e2U,
		0x87c4U,0x1f83U,0x3f06U,0x7e0cU,0xfc18U,0xe83bU,0xc07dU,0x90f1U,0x31e9U,0x63d2U,
		0xc7a4U,0x9f43U,0x2e8dU,0x5d1aU,0xba34U,0x6463U,0xc8c6U,0x8187U,0x1305U,0x260aU,
		0x4c14U,0x9828U,0x205bU,0x40b6U,0x816cU,0x12d3U,0x25a6U,0x4b4cU,0x9698U,0x3d3bU,
		0x7a76U,0xf4ecU,0xf9d3U,0xe3adU,0xd751U,0xbea9U,0x6d59U,0xdab2U,0xa56fU,0x5ad5U,
		0xb5aaU,0x7b5fU,0xf6beU,0xfd77U,0xeae5U,0xc5c1U,0x9b89U,0x2719U,0x4e32U,0x9c64U,
		0x28c3U,0x5186U,0xa30cU,0x5613U,0xac26U,0x4847U,0x908eU,0x3117U,0x622eU,0xc45cU,
		0x98b3U,0x216dU,0x42daU,0x85b4U,0x1b63U,0x36c6U,0x6d8cU,0xdb18U,0xa63bU,0x5c7dU,
		0xb8faU,0x61ffU,0xc3feU,0x97f7U,0x3fe5U,0x7fcaU,0xff94U,0xef23U,0xce4dU,0x8c91U,
		0x0929U,0x1252U,0x24a4U,0x4948U,0x9290U,0x352bU,0x6a56U,0xd4acU,0xb953U,0x62adU,
		0xc55aU,0x9abfU,0x2575U,0x4aeaU,0x95d4U,0x3ba3U,0x7746U,0xee8cU,0xcd13U,0x8a2dU,
		0x0451U,0x08a2U,0x1144U,0x2288U,0x4510U,0x8a20U,0x044bU,0x0896U,0x112cU,0x2258U,
		0x44b0U,0x8960U,0x02cbU,0x0596U,0x0b2cU,0x1658U,0x2cb0U,0x5960U,0xb2c0U,0x758bU,
		0xeb16U,0xc627U,0x9c45U,0x2881U,0x5102U,0xa204U,0x5403U,0xa806U,0x4007U,0x800eU,
		0x1017U,0x202eU,0x405cU,0x80b8U,0x117bU,0x22f6U,0x45ecU,0x8bd8U,0x07bbU,0x0f76U,
		0x1eecU,0x3dd8U,0x7bb0U,0xf760U,0xfecbU,0xed9dU,0xcb31U,0x8669U,0x1cd9U,0x39b2U,
		0x7364U,0xe6c8U,0xdd9bU,0xab3dU,0x4671U,0x8ce2U,0x09cfU,0x139eU,0x273cU,0x4e78U,
		0x9cf0U,0x29ebU,0x53d6U,0xa7acU,0x5f53U,0xbea6U,0x6d47U,0xda8eU,0xa517U,0x5a25U,
		0xb44aU,0x789fU,0xf13eU,0xf277U,0xf4e5U,0xf9c1U,0xe389U,0xd719U,0xbe39U,0x6c79U,
		0xd8f2U,0xa1efU,0x53d5U,0xa7aaU,0x5f5fU,0xbebeU,0x6d77U,0xdaeeU,0xa5d7U,0x5ba5U,
		0xb74aU,0x7e9fU,0xfd3eU,0xea77U,0xc4e5U,0x99c1U,0x2389U,0x4712U,0x8e24U,0x0c43U,
		0x1886U,0x310cU,0x6218U,0xc430U,0x986bU,0x20ddU,0x41baU,0x8374U,0x16e3U,0x2dc6U,
		0x5b8cU,0xb718U,0x7e3bU,0xfc76U,0xe8e7U,0xc1c5U,0x9381U,0x3709U,0x6e12U,0xdc24U,
		0xa843U,0x408dU,0x811aU,0x123fU,0x247eU,0x48fcU,0x91f8U,0x33fbU,0x67f6U,0xcfecU,
		0x8fd3U,0x0fadU,0x1f5aU,0x3eb4U,0x7d68U,0xfad0U,0xe5abU,0xdb5dU,0xa6b1U,0x5d69U,
		0xbad2U,0x65afU,0xcb5eU,0x86b7U,0x1d65U,0x3acaU,0x7594U,0xeb28U,0xc65bU,0x9cbdU,
		0x2971U,0x52e2U,0xa5c4U,0x5b83U,0xb706U,0x7e07U,0xfc0eU,0xe817U,0xc025U,0x9041U,
		0x3089U,0x6112U,0xc224U,0x9443U,0x388dU,0x711aU,0xe234U,0xd463U,0xb8cdU,0x6191U,
		0xc322U,0x964fU,0x3c95U,0x792aU,0xf254U,0xf4a3U,0xf94dU,0xe291U,0xd529U,0xba59U,
		0x64b9U,0xc972U,0x82efU,0x15d5U,0x2baaU,0x5754U,0xaea8U,0x4d5bU,0x9ab6U,0x2567U,
		0x4aceU,0x959cU,0x3b33U,0x7666U,0xecccU,0xc993U,0x832dU,0x1651U,0x2ca2U,0x5944U,
		0xb288U,0x751bU,0xea36U,0xc467U,0x98c5U,0x2181U,0x4302U,0x8604U,0x1c03U,0x3806U,
		0x700cU,0xe018U,0xd03bU,0xb07dU,0x70f1U,0xe1e2U,0xd3cfU,0xb795U,0x7f21U,0xfe42U,
		0xec8fU,0xc915U,0x8221U,0x1449U,0x2892U,0x5124U,0xa248U,0x549bU,0xa936U,0x4267U,
		0x84ceU,0x1997U,0x332eU,0x665cU,0xccb8U,0x897bU,0x02fdU,0x05faU,0x0bf4U,0x17e8U,
		0x2fd0U,0x5fa0U,0xbf40U,0x6e8bU,0xdd16U,0xaa27U,0x4445U,0x888aU,0x011fU,0x023eU,
		0x047cU,0x08f8U,0x11f0U,0x23e0U,0x47c0U,0x8f80U,0x0f0bU,0x1e16U,0x3c2cU,0x7858U,
		0xf0b0U,0xf16bU,0xf2ddU,0xf5b1U,0xfb69U,0xe6d9U,0xddb9U,0xab79U,0x46f9U,0x8df2U,
		0x0befU,0x17deU,0x2fbcU,0x5f78U,0xbef0U,0x6debU,0xdbd6U,0xa7a7U,0x5f45U,0xbe8aU,
		0x6d1fU,0xda3eU,0xa477U,0x58e5U,0xb1caU,0x739fU,0xe73eU,0xde77U,0xace5U,0x49c1U,
		0x9382U,0x370fU,0x6e1eU,0xdc3cU,0xa873U,0x40edU,0x81daU,0x13bfU,0x277eU,0x4efcU,
		0x9df8U,0x2bfbU,0x57f6U,0xafecU,0x4fd3U,0x9fa6U,0x2f47U,0x5e8eU,0xbd1cU,0x6a33U,
		0xd466U,0xb8c7U,0x6185U,0xc30aU,0x961fU,0x3c35U,0x786aU,0xf0d4U,0xf1a3U,0xf34dU,
		0xf691U,0xfd29U,0xea59U,0xc4b9U,0x9979U,0x22f9U,0x45f2U,0x8be4U,0x07c3U,0x0f86U,
		0x1f0cU,0x3e18U,0x7c30U,0xf860U,0xe0cbU,0xd19dU,0xb331U,0x7669U,0xecd2U,0xc9afU,
		0x8355U,0x16a1U,0x2d42U,0x5a84U,0xb508U,0x7a1bU,0xf436U,0xf867U,0xe0c5U,0xd181U,
		0xb309U,0x7619U,0xec32U,0xc86fU,0x80d5U,0x11a1U,0x2342U,0x4684U,0x8d08U,0x0a1bU,
		0x1436U,0x286cU,0x50d8U,0xa1b0U,0x536bU,0xa6d6U,0x5da7U,0xbb4eU,0x6697U,0xcd2eU,
		0x8a57U,0x04a5U,0x094aU,0x1294U,0x2528U,0x4a50U,0x94a0U,0x394bU,0x7296U,0xe52cU,
		0xda53U,0xa4adU,0x5951U,0xb2a2U,0x754fU,0xea9eU,0xc537U,0x9a65U,0x24c1U,0x4982U,
		0x9304U,0x3603U,0x6c06U,0xd80cU,0xa013U,0x502dU,0xa05aU,0x50bfU,0xa17eU,0x52f7U,
		0xa5eeU,0x5bd7U,0xb7aeU,0x7f57U,0xfeaeU,0xed57U,0xcaa5U,0x8541U,0x1a89U,0x3512U,
		0x6a24U,0xd448U,0xb89bU,0x613dU,0xc27aU,0x94ffU,0x39f5U,0x73eaU,0xe7d4U,0xdfa3U,
		0xaf4dU,0x4e91U,0x9d22U,0x2a4fU,0x549eU,0xa93cU,0x4273U,0x84e6U,0x19c7U,0x338eU,
		0x671cU,0xce38U,0x8c7bU,0x08fdU,0x11faU,0x23f4U,0x47e8U,0x8fd0U,0x0fabU,0x1f56U,
		0x3eacU,0x7d58U,0xfab0U,0xe56bU,0xdaddU,0xa5b1U,0x5b69U,0xb6d2U,0x7dafU,0xfb5eU,
		0xe6b7U,0xdd65U,0xaac1U,0x4589U,0x8b12U,0x062fU,0x0c5eU,0x18bcU,0x3178U,0x62f0U,
		0xc5e0U,0x9bcbU,0x279dU,0x4f3aU,0x9e74U,0x2ce3U,0x59c6U,0xb38cU,0x7713U,0xee26U,
		0xcc47U,0x8885U,0x0101U,0x0202U,0x0404U,0x0808U,0x1010U,0x2020U,0x4040U,0x8080U,
		0x110bU,0x2216U,0x442cU,0x8858U,0x00bbU,0x0176U,0x02ecU,0x05d8U,0x0bb0U,0x1760U,
		0x2ec0U,0x5d80U,0xbb00U,0x660bU,0xcc16U,0x8827U,0x0045U,0x008aU,0x0114U,0x0228U,
		0x0450U,0x08a0U,0x1140U,0x2280U,0x4500U,0x8a00U,0x040bU,0x0816U,0x102cU,0x2058U,
		0x40b0U,0x8160U,0x12cbU,0x2596U,0x4b2cU,0x9658U,0x3cbbU,0x7976U,0xf2ecU,0xf5d3U,
		0xfbadU,0xe751U,0xdea9U,0xad59U,0x4ab9U,0x9572U,0x3aefU,0x75deU,0xebbcU,0xc773U,
		0x9eedU,0x2dd1U,0x5ba2U,0xb744U,0x7e83U,0xfd06U,0xea07U,0xc405U,0x9801U,0x2009U,
		0x4012U,0x8024U,0x1043U,0x2086U,0x410cU,0x8218U,0x143bU,0x2876U,0x50ecU,0xa1d8U,
		0x53bbU,0xa776U,0x5ee7U,0xbdceU,0x6b97U,0xd72eU,0xbe57U,0x6ca5U,0xd94aU,0xa29fU,
		0x5535U,0xaa6aU,0x44dfU,0x89beU,0x0377U,0x06eeU,0x0ddcU,0x1bb8U,0x3770U,0x6ee0U,
		0xddc0U,0xab8bU,0x471dU,0x8e3aU,0x0c7fU,0x18feU,0x31fcU,0x63f8U,0xc7f0U,0x9febU,
		0x2fddU,0x5fbaU,0xbf74U,0x6ee3U,0xddc6U,0xab87U,0x4705U,0x8e0aU,0x0c1fU,0x183eU,
		0x307cU,0x60f8U,0xc1f0U,0x93ebU,0x37ddU,0x6fbaU,0xdf74U,0xaee3U,0x4dcdU,0x9b9aU,
		0x273fU,0x4e7eU,0x9cfcU,0x29f3U,0x53e6U,0xa7ccU,0x5f93U,0xbf26U,0x6e47U,0xdc8eU,
		0xa917U,0x4225U,0x844aU,0x189fU,0x313eU,0x627cU,0xc4f8U,0x99fbU,0x23fdU,0x47faU,
		0x8ff4U,0x0fe3U,0x1fc6U,0x3f8cU,0x7f18U,0xfe30U,0xec6bU,0xc8ddU,0x81b1U,0x1369U,
		0x26d2U,0x4da4U,0x9b48U,0x269bU,0x4d36U,0x9a6cU,0x24d3U,0x49a6U,0x934cU,0x3693U,
		0x6d26U,0xda4cU,0xa493U,0x592dU,0xb25aU,0x74bfU,0xe97eU,0xc2f7U,0x95e5U,0x3bc1U,
		0x7782U,0xef04U,0xce03U,0x8c0dU,0x0811U,0x1022U,0x2044U,0x4088U,0x8110U,0x122bU,
		0x2456U,0x48acU,0x9158U,0x32bbU,0x6576U,0xcaecU,0x85d3U,0x1badU,0x375aU,0x6eb4U,
		0xdd68U,0xaadbU,0x45bdU,0x8b7aU,0x06ffU,0x0dfeU,0x1bfcU,0x37f8U,0x6ff0U,0xdfe0U,
		0xafcbU,0x4f9dU,0x9f3aU,0x2e7fU,0x5cfeU,0xb9fcU,0x63f3U,0xc7e6U,0x9fc7U,0x2f85U,
		0x5f0aU,0xbe14U,0x6c23U,0xd846U,0xa087U,0x5105U,0xa20aU,0x541fU,0xa83eU,0x4077U,
		0x80eeU,0x11d7U,0x23aeU,0x475cU,0x8eb8U,0x0d7bU,0x1af6U,0x35ecU,0x6bd8U,0xd7b0U,
		0xbf6bU,0x6eddU,0xddbaU,0xab7fU,0x46f5U,0x8deaU,0x0bdfU,0x17beU,0x2f7cU,0x5ef8U,
		0xbdf0U,0x6bebU,0xd7d6U,0xbfa7U,0x6f45U,0xde8aU,0xad1fU,0x4a35U,0x946aU,0x38dfU,
		0x71beU,0xe37cU,0xd6f3U,0xbdedU,0x6bd1U,0xd7a2U,0xbf4fU,0x6e95U,0xdd2aU,0xaa5fU,
		0x44b5U,0x896aU,0x02dfU,0x05beU,0x0b7cU,0x16f8U,0x2df0U,0x5be0U,0xb7c0U,0x7f8bU,
		0xff16U,0xee27U,0xcc45U,0x8881U,0x0109U,0x0212U,0x0424U,0x0848U,0x1090U,0x2120U,
		0x4240U,0x8480U,0x190bU,0x3216U,0x642cU,0xc858U,0x80bbU,0x117dU,0x22faU,0x45f4U,
		0x8be8U,0x07dbU,0x0fb6U,0x1f6cU,0x3ed8U,0x7db0U,0xfb60U,0xe6cbU,0xdd9dU,0xab31U,
		0x4669U,0x8cd2U,0x09afU,0x135eU,0x26bcU,0x4d78U,0x9af0U,0x25ebU,0x4bd6U,0x97acU,
		0x3f53U,0x7ea6U,0xfd4cU,0xea93U,0xc52dU,0x9a51U,0x24a9U,0x4952U,0x92a4U,0x3543U,
		0x6a86U,0xd50cU,0xba13U,0x642dU,0xc85aU,0x80bfU,0x1175U,0x22eaU,0x45d4U,0x8ba8U,
		0x075bU,0x0eb6U,0x1d6cU,0x3ad8U,0x75b0U,0xeb60U,0xc6cbU,0x9d9dU,0x2b31U,0x5662U,
		0xacc4U,0x4983U,0x9306U,0x3607U,0x6c0eU,0xd81cU,0xa033U,0x506dU,0xa0daU,0x51bfU,
		0xa37eU,0x56f7U,0xadeeU,0x4bd7U,0x97aeU,0x3f57U,0x7eaeU,0xfd5cU,0xeab3U,0xc56dU,
		0x9ad1U,0x25a9U,0x4b52U,0x96a4U,0x3d43U,0x7a86U,0xf50cU,0xfa13U,0xe42dU,0xd851U,
		0xa0a9U,0x5159U,0xa2b2U,0x556fU,0xaadeU,0x45b7U,0x8b6eU,0x06d7U,0x0daeU,0x1b5cU,
		0x36b8U,0x6d70U,0xdae0U,0xa5cbU,0x5b9dU,0xb73aU,0x7e7fU,0xfcfeU,0xe9f7U,0xc3e5U,
		0x97c1U,0x3f89U,0x7f12U,0xfe24U,0xec43U,0xc88dU,0x8111U,0x1229U,0x2452U,0x48a4U,
		0x9148U,0x329bU,0x6536U,0xca6cU,0x84d3U,0x19adU,0x335aU,0x66b4U,0xcd68U,0x8adbU,
		0x05bdU,0x0b7aU,0x16f4U,0x2de8U,0x5bd0U,0xb7a0U,0x7f4bU,0xfe96U,0xed27U,0xca45U,
		0x8481U,0x1909U,0x3212U,0x6424U,0xc848U,0x809bU,0x113dU,0x227aU,0x44f4U,0x89e8U,
		0x03dbU,0x07b6U,0x0f6cU,0x1ed8U,0x3db0U,0x7b60U,0xf6c0U,0xfd8bU,0xeb1dU,0xc631U,
		0x9c69U,0x28d9U,0x51b2U,0xa364U,0x56c3U,0xad86U,0x4b07U,0x960eU,0x3c17U,0x782eU,
		0xf05cU,0xf0b3U,0xf16dU,0xf2d1U,0xf5a9U,0xfb59U,0xe6b9U,0xdd79U,0xaaf9U,0x45f9U,
		0x8bf2U,0x07efU,0x0fdeU,0x1fbcU,0x3f78U,0x7ef0U,0xfde0U,0xebcbU,0xc79dU,0x9f31U,
		0x2e69U,0x5cd2U,0xb9a4U,0x6343U,0xc686U,0x9d07U,0x2a05U,0x540aU,0xa814U,0x4023U,
		0x8046U,0x1087U,0x210eU,0x421cU,0x8438U,0x187bU,0x30f6U,0x61ecU,0xc3d8U,0x97bbU,
		0x3f7dU,0x7efaU,0xfdf4U,0xebe3U,0xc7cdU,0x9f91U,0x2f29U,0x5e52U,0xbca4U,0x6943U,
		0xd286U,0xb507U,0x7a05U,0xf40aU,0xf81fU,0xe035U,0xd061U,0xb0c9U,0x7199U,0xe332U,
		0xd66fU,0xbcd5U,0x69a1U,0xd342U,0xb68fU,0x7d15U,0xfa2aU,0xe45fU,0xd8b5U,0xa161U,
		0x52c9U,0xa592U,0x5b2fU,0xb65eU,0x7cb7U,0xf96eU,0xe2d7U,0xd5a5U,0xbb41U,0x6689U,
		0xcd12U,0x8a2fU,0x0455U,0x08aaU,0x1154U,0x22a8U,0x4550U,0x8aa0U,0x054bU,0x0a96U,
		0x152cU,0x2a58U,0x54b0U,0xa960U,0x42cbU,0x8596U,0x1b27U,0x364eU,0x6c9cU,0xd938U,
		0xa27bU,0x54fdU,0xa9faU,0x43ffU,0x87feU,0x1ff7U,0x3feeU,0x7fdcU,0xffb8U,0xef7bU,
		0xcefdU,0x8df1U,0x0be9U,0x17d2U,0x2fa4U,0x5f48U,0xbe90U,0x6d2bU,0xda56U,0xa4a7U,
		0x5945U,0xb28aU,0x751fU,0xea3eU,0xc477U,0x98e5U,0x21c1U,0x4382U,0x8704U,0x1e03U,
		0x3c06U,0x780cU,0xf018U,0xf03bU,0xf07dU,0xf0f1U,0xf1e9U,0xf3d9U,0xf7b9U,0xff79U,
		0xeef9U,0xcdf9U,0x8bf9U,0x07f9U,0x0ff2U,0x1fe4U,0x3fc8U,0x7f90U,0xff20U,0xee4bU,
		0xcc9dU,0x8931U,0x0269U,0x04d2U,0x09a4U,0x1348U,0x2690U,0x4d20U,0x9a40U,0x248bU,
		0x4916U,0x922cU,0x3453U,0x68a6U,0xd14cU,0xb293U,0x752dU,0xea5aU,0xc4bfU,0x9975U,
		0x22e1U,0x45c2U,0x8b84U,0x0703U,0x0e06U,0x1c0cU,0x3818U,0x7030U,0xe060U,0xd0cbU,
		0xb19dU,0x7331U,0xe662U,0xdccfU,0xa995U,0x4321U,0x8642U,0x1c8fU,0x391eU,0x723cU,
		0xe478U,0xd8fbU,0xa1fdU,0x53f1U,0xa7e2U,0x5fcfU,0xbf9eU,0x6f37U,0xde6eU,0xacd7U,
		0x49a5U,0x934aU,0x369fU,0x6d3eU,0xda7cU,0xa4f3U,0x59edU,0xb3daU,0x77bfU,0xef7eU,
		0xcef7U,0x8de5U,0x0bc1U,0x1782U,0x2f04U,0x5e08U,0xbc10U,0x682bU,0xd056U,0xb0a7U,
		0x7145U,0xe28aU,0xd51fU,0xba35U,0x6461U,0xc8c2U,0x818fU,0x1315U,0x262aU,0x4c54U,
		0x98a8U,0x215bU,0x42b6U,0x856cU,0x1ad3U,0x35a6U,0x6b4cU,0xd698U,0xbd3bU,0x6a7dU,
		0xd4faU,0xb9ffU,0x63f5U,0xc7eaU,0x9fdfU,0x2fb5U,0x5f6aU,0xbed4U,0x6da3U,0xdb46U,
		0xa687U,0x5d05U,0xba0aU,0x641fU,0xc83eU,0x8077U,0x10e5U,0x21caU,0x4394U,0x8728U,
		0x1e5bU,0x3cb6U,0x796cU,0xf2d8U,0xf5bbU,0xfb7dU,0xe6f1U,0xdde9U,0xabd9U,0x47b9U,
		0x8f72U,0x0eefU,0x1ddeU,0x3bbcU,0x7778U,0xeef0U,0xcdebU,0x8bddU,0x07b1U,0x0f62U,
		0x1ec4U,0x3d88U,0x7b10U,0xf620U,0xfc4bU,0xe89dU,0xc131U,0x9269U,0x34d9U,0x69b2U,
		0xd364U,0xb6c3U,0x7d8dU,0xfb1aU,0xe63fU,0xdc75U,0xa8e1U,0x41c9U,0x8392U,0x172fU,
		0x2e5eU,0x5cbcU,0xb978U,0x62fbU,0xc5f6U,0x9be7U,0x27c5U,0x4f8aU,0x9f14U,0x2e23U,
		0x5c46U,0xb88cU,0x6113U,0xc226U,0x9447U,0x3885U,0x710aU,0xe214U,0xd423U,0xb84dU,
		0x6091U,0xc122U,0x924fU,0x3495U,0x692aU,0xd254U,0xb4a3U,0x794dU,0xf29aU,0xf53fU,
		0xfa75U,0xe4e1U,0xd9c9U,0xa399U,0x5739U,0xae72U,0x4cefU,0x99deU,0x23b7U,0x476eU,
		0x8edcU,0x0db3U,0x1b66U,0x36ccU,0x6d98U,0xdb30U,0xa66bU,0x5cddU,0xb9baU,0x637fU,
		0xc6feU,0x9df7U,0x2be5U,0x57caU,0xaf94U,0x4f23U,0x9e46U,0x2c87U,0x590eU,0xb21cU,
		0x7433U,0xe866U,0xc0c7U,0x9185U,0x3301U,0x6602U,0xcc04U,0x8803U,0x000dU,0x001aU,
		0x0034U,0x0068U,0x00d0U,0x01a0U,0x0340U,0x0680U,0x0d00U,0x1a00U,0x3400U,0x6800U,
		0xd000U,0xb00bU,0x701dU,0xe03aU,0xd07fU,0xb0f5U,0x71e1U,0xe3c2U,0xd78fU,0xbf15U,
		0x6e21U,0xdc42U,0xa88fU,0x4115U,0x822aU,0x145fU,0x28beU,0x517cU,0xa2f8U,0x55fbU,
		0xabf6U,0x47e7U,0x8fceU,0x0f97U,0x1f2eU,0x3e5cU,0x7cb8U,0xf970U,0xe2ebU,0xd5ddU,
		0xbbb1U,0x6769U,0xced2U,0x8dafU,0x0b55U,0x16aaU,0x2d54U,0x5aa8U,0xb550U,0x7aabU,
		0xf556U,0xfaa7U,0xe545U,0xda81U,0xa509U,0x5a19U,0xb432U,0x786fU,0xf0deU,0xf1b7U,
		0xf365U,0xf6c1U,0xfd89U,0xeb19U,0xc639U,0x9c79U,0x28f9U,0x51f2U,0xa3e4U,0x57c3U,
		0xaf86U,0x4f07U,0x9e0eU,0x2c17U,0x582eU,0xb05cU,0x70b3U,0xe166U,0xd2c7U,0xb585U,
		0x7b01U,0xf602U,0xfc0fU,0xe815U,0xc021U,0x9049U,0x3099U,0x6132U,0xc264U,0x94c3U,
		0x398dU,0x731aU,0xe634U,0xdc63U,0xa8cdU,0x4191U,0x8322U,0x164fU,0x2c9eU,0x593cU,
		0xb278U,0x74fbU,0xe9f6U,0xc3e7U,0x97c5U,0x3f81U,0x7f02U,0xfe04U,0xec03U,0xc80dU,
		0x8011U,0x1029U,0x2052U,0x40a4U,0x8148U,0x129bU,0x2536U,0x4a6cU,0x94d8U,0x39bbU,
		0x7376U,0xe6ecU,0xddd3U,0xabadU,0x4751U,0x8ea2U,0x0d4fU,0x1a9eU,0x353cU,0x6a78U,
		0xd4f0U,0xb9ebU,0x63ddU,0xc7baU,0x9f7fU,0x2ef5U,0x5deaU,0xbbd4U,0x67a3U,0xcf46U,
		0x8e87U,0x0d05U,0x1a0aU,0x3414U,0x6828U,0xd050U,0xb0abU,0x715dU,0xe2baU,0xd57fU,
		0xbaf5U,0x65e1U,0xcbc2U,0x878fU,0x1f15U,0x3e2aU,0x7c54U,0xf8a8U,0xe15bU,0xd2bdU,
		0xb571U,0x7ae9U,0xf5d2U,0xfbafU,0xe755U,0xdea1U,0xad49U,0x4a99U,0x9532U,0x3a6fU,
		0x74deU,0xe9bcU,0xc373U,0x96edU,0x3dd1U,0x7ba2U,0xf744U,0xfe83U,0xed0dU,0xca11U,
		0x8429U,0x1859U,0x30b2U,0x6164U,0xc2c8U,0x959bU,0x3b3dU,0x767aU,0xecf4U,0xc9e3U,
		0x83cdU,0x1791U,0x2f22U,0x5e44U,0xbc88U,0x691bU,0xd236U,0xb467U,0x78c5U,0xf18aU,
		0xf31fU,0xf635U,0xfc61U,0xe8c9U,0xc199U,0x9339U,0x3679U,0x6cf2U,0xd9e4U,0xa3c3U,
		0x578dU,0xaf1aU,0x4e3fU,0x9c7eU,0x28f7U,0x51eeU,0xa3dcU,0x57b3U,0xaf66U,0x4ec7U,
		0x9d8eU,0x2b17U,0x562eU,0xac5cU,0x48b3U,0x9166U,0x32c7U,0x658eU,0xcb1cU,0x8633U,
		0x1c6dU,0x38daU,0x71b4U,0xe368U,0xd6dbU,0xbdbdU,0x6b71U,0xd6e2U,0xbdcfU,0x6b95U,
		0xd72aU,0xbe5fU,0x6cb5U,0xd96aU,0xa2dfU,0x55b5U,0xab6aU,0x46dfU,0x8dbeU,0x0b77U,
		0x16eeU,0x2ddcU,0x5bb8U,0xb770U,0x7eebU,0xfdd6U,0xeba7U,0xc745U,0x9e81U,0x2d09U,
		0x5a12U,0xb424U,0x7843U,0xf086U,0xf107U,0xf205U,0xf401U,0xf809U,0xe019U,0xd039U,
		0xb079U,0x70f9U,0xe1f2U,0xd3efU,0xb7d5U,0x7fa1U,0xff42U,0xee8fU,0xcd15U,0x8a21U,
		0x0449U,0x0892U,0x1124U,0x2248U,0x4490U,0x8920U,0x024bU,0x0496U,0x092cU,0x1258U,
		0x24b0U,0x4960U,0x92c0U,0x358bU,0x6b16U,0xd62cU,0xbc53U,0x68adU,0xd15aU,0xb2bfU,
		0x7575U,0xeaeaU,0xc5dfU,0x9bb5U,0x2761U,0x4ec2U,0x9d84U,0x2b03U,0x5606U,0xac0cU,
		0x4813U,0x9026U,0x3047U,0x608eU,0xc11cU,0x9233U,0x346dU,0x68daU,0xd1b4U,0xb363U,
		0x76cdU,0xed9aU,0xcb3fU,0x8675U,0x1ce1U,0x39c2U,0x7384U,0xe708U,0xde1bU,0xac3dU,
		0x4871U,0x90e2U,0x31cfU,0x639eU,0xc73cU,0x9e73U,0x2cedU,0x59daU,0xb3b4U,0x7763U,
		0xeec6U,0xcd87U,0x8b05U,0x0601U,0x0c02U,0x1804U,0x3008U,0x6010U,0xc020U,0x904bU,
		0x309dU,0x613aU,0xc274U,0x94e3U,0x39cdU,0x739aU,0xe734U,0xde63U,0xaccdU,0x4991U,
		0x9322U,0x364fU,0x6c9eU,0xd93cU,0xa273U,0x54edU,0xa9daU,0x43bfU,0x877eU,0x1ef7U,
		0x3deeU,0x7bdcU,0xf7b8U,0xff7bU,0xeefdU,0xcdf1U,0x8be9U,0x07d9U,0x0fb2U,0x1f64U,
		0x3ec8U,0x7d90U,0xfb20U,0xe64bU,0xdc9dU,0xa931U,0x4269U,0x84d2U,0x19afU,0x335eU,
		0x66bcU,0xcd78U,0x8afbU,0x05fdU,0x0bfaU,0x17f4U,0x2fe8U,0x5fd0U,0xbfa0U,0x6f4bU,
		0xde96U,0xad27U,0x4a45U,0x948aU,0x391fU,0x723eU,0xe47cU,0xd8f3U,0xa1edU,0x53d1U,
		0xa7a2U,0x5f4fU,0xbe9eU,0x6d37U,0xda6eU,0xa4d7U,0x59a5U,0xb34aU,0x769fU,0xed3eU,
		0xca77U,0x84e5U,0x19c1U,0x3382U,0x6704U,0xce08U,0x8c1bU,0x083dU,0x107aU,0x20f4U,
		0x41e8U,0x83d0U,0x17abU,0x2f56U,0x5eacU,0xbd58U,0x6abbU,0xd576U,0xbae7U,0x65c5U,
		0xcb8aU,0x871fU,0x1e35U,0x3c6aU,0x78d4U,0xf1a8U,0xf35bU,0xf6bdU,0xfd71U,0xeae9U,
		0xc5d9U,0x9bb9U,0x2779U,0x4ef2U,0x9de4U,0x2bc3U,0x5786U,0xaf0cU,0x4e13U,0x9c26U,
		0x2847U,0x508eU,0xa11cU,0x5233U,0xa466U,0x58c7U,0xb18eU,0x7317U,0xe62eU,0xdc57U,
		0xa8a5U,0x4141U,0x8282U,0x150fU,0x2a1eU,0x543cU,0xa878U,0x40fbU,0x81f6U,0x13e7U,
		0x27ceU,0x4f9cU,0x9f38U,0x2e7bU,0x5cf6U,0xb9ecU,0x63d3U,0xc7a6U,0x9f47U,0x2e85U,
		0x5d0aU,0xba14U,0x6423U,0xc846U,0x8087U,0x1105U,0x220aU,0x4414U,0x8828U,0x005bU,
		0x00b6U,0x016cU,0x02d8U,0x05b0U,0x0b60U,0x16c0U,0x2d80U,0x5b00U,0xb600U,0x7c0bU,
		0xf816U,0xe027U,0xd045U,0xb081U,0x7109U,0xe212U,0xd42fU,0xb855U,0x60a1U,0xc142U,
		0x928fU,0x3515U,0x6a2aU,0xd454U,0xb8a3U,0x614dU,0xc29aU,0x953fU,0x3a75U,0x74eaU,
		0xe9d4U,0xc3a3U,0x974dU,0x3e91U,0x7d22U,0xfa44U,0xe483U,0xd90dU,0xa211U,0x5429U,
		0xa852U,0x40afU,0x815eU,0x12b7U,0x256eU,0x4adcU,0x95b8U,0x3b7bU,0x76f6U,0xedecU,
		0xcbd3U,0x87adU,0x1f51U,0x3ea2U,0x7d44U,0xfa88U,0xe51bU,0xda3dU,0xa471U,0x58e9U,
		0xb1d2U,0x73afU,0xe75eU,0xdeb7U,0xad65U,0x4ac1U,0x9582U,0x3b0fU,0x761eU,0xec3cU,
		0xc873U,0x80edU,0x11d1U,0x23a2U,0x4744U,0x8e88U,0x0d1bU,0x1a36U,0x346cU,0x68d8U,
		0xd1b0U,0xb36bU,0x76ddU,0xedbaU,0xcb7fU,0x86f5U,0x1de1U,0x3bc2U,0x7784U,0xef08U,
		0xce1bU,0x8c3dU,0x0871U,0x10e2U,0x21c4U,0x4388U,0x8710U,0x1e2bU,0x3c56U,0x78acU,
		0xf158U,0xf2bbU,0xf57dU,0xfaf1U,0xe5e9U,0xdbd9U,0xa7b9U,0x5f79U,0xbef2U,0x6defU,
		0xdbdeU,0xa7b7U,0x5f65U,0xbecaU,0x6d9fU,0xdb3eU,0xa677U,0x5ce5U,0xb9caU,0x639fU,
		0xc73eU,0x9e77U,0x2ce5U,0x59caU,0xb394U,0x7723U,0xee46U,0xcc87U,0x8905U,0x0201U,
		0x0402U,0x0804U,0x1008U,0x2010U,0x4020U,0x8040U,0x108bU,0x2116U,0x422cU,0x8458U,
		0x18bbU,0x3176U,0x62ecU,0xc5d8U,0x9bbbU,0x277dU,0x4efaU,0x9df4U,0x2be3U,0x57c6U,
		0xaf8cU,0x4f13U,0x9e26U,0x2c47U,0x588eU,0xb11cU,0x7233U,0xe466U,0xd8c7U,0xa185U,
		0x5301U,0xa602U,0x5c0fU,0xb81eU,0x6037U,0xc06eU,0x90d7U,0x31a5U,0x634aU,0xc694U,
		0x9d23U,0x2a4dU,0x549aU,0xa934U,0x4263U,0x84c6U,0x1987U,0x330eU,0x661cU,0xcc38U,
		0x887bU,0x00fdU,0x01faU,0x03f4U,0x07e8U,0x0fd0U,0x1fa0U,0x3f40U,0x7e80U,0xfd00U,
		0xea0bU,0xc41dU,0x9831U,0x2069U,0x40d2U,0x81a4U,0x1343U,0x2686U,0x4d0cU,0x9a18U,
		0x243bU,0x4876U,0x90ecU,0x31d3U,0x63a6U,0xc74cU,0x9e93U,0x2d2dU,0x5a5aU,0xb4b4U,
		0x7963U,0xf2c6U,0xf587U,0xfb05U,0xe601U,0xdc09U,0xa819U,0x4039U,0x8072U,0x10efU,
		0x21deU,0x43bcU,0x8778U,0x1efbU,0x3df6U,0x7becU,0xf7d8U,0xffbbU,0xef7dU,0xcef1U,
		0x8de9U,0x0bd9U,0x17b2U,0x2f64U,0x5ec8U,0xbd90U,0x6b2bU,0xd656U,0xbca7U,0x6945U,
		0xd28aU,0xb51fU,0x7a35U,0xf46aU,0xf8dfU,0xe1b5U,0xd361U,0xb6c9U,0x7d99U,0xfb32U,
		0xe66fU,0xdcd5U,0xa9a1U,0x4349U,0x8692U,0x1d2fU,0x3a5eU,0x74bcU,0xe978U,0xc2fbU,
		0x95fdU,0x3bf1U,0x77e2U,0xefc4U,0xcf83U,0x8f0dU,0x0e11U,0x1c22U,0x3844U,0x7088U,
		0xe110U,0xd22bU,0xb45dU,0x78b1U,0xf162U,0xf2cfU,0xf595U,0xfb21U,0xe649U,0xdc99U,
		0xa939U,0x4279U,0x84f2U,0x19efU,0x33deU,0x67bcU,0xcf78U,0x8efbU,0x0dfdU,0x1bfaU,
		0x37f4U,0x6fe8U,0xdfd0U,0xafabU,0x4f5dU,0x9ebaU,0x2d7fU,0x5afeU,0xb5fcU,0x7bf3U,
		0xf7e6U,0xffc7U,0xef85U,0xcf01U,0x8e09U,0x0c19U,0x1832U,0x3064U,0x60c8U,0xc190U,
		0x932bU,0x365dU,0x6cbaU,0xd974U,0xa2e3U,0x55cdU,0xab9aU,0x473fU,0x8e7eU,0x0cf7U,
		0x19eeU,0x33dcU,0x67b8U,0xcf70U,0x8eebU,0x0dddU,0x1bbaU,0x3774U,0x6ee8U,0xddd0U,
		0xababU,0x475dU,0x8ebaU,0x0d7fU,0x1afeU,0x35fcU,0x6bf8U,0xd7f0U,0xbfebU,0x6fddU,
		0xdfbaU,0xaf7fU,0x4ef5U,0x9deaU,0x2bdfU,0x57beU,0xaf7cU,0x4ef3U,0x9de6U,0x2bc7U,
		0x578eU,0xaf1cU,0x4e33U,0x9c66U,0x28c7U,0x518eU,0xa31cU,0x5633U,0xac66U,0x48c7U,
		0x918eU,0x3317U,0x662eU,0xcc5cU,0x88b3U,0x016dU,0x02daU,0x05b4U,0x0b68U,0x16d0U,
		0x2da0U,0x5b40U,0xb680U,0x7d0bU,0xfa16U,0xe427U,0xd845U,0xa081U,0x5109U,0xa212U,
		0x542fU,0xa85eU,0x40b7U,0x816eU,0x12d7U,0x25aeU,0x4b5cU,0x96b8U,0x3d7bU,0x7af6U,
		0xf5ecU,0xfbd3U,0xe7adU,0xdf51U,0xaea9U,0x4d59U,0x9ab2U,0x256fU,0x4adeU,0x95bcU,
		0x3b73U,0x76e6U,0xedccU,0xcb93U,0x872dU,0x1e51U,0x3ca2U,0x7944U,0xf288U,0xf51bU,
		0xfa3dU,0xe471U,0xd8e9U,0xa1d9U,0x53b9U,0xa772U,0x5eefU,0xbddeU,0x6bb7U,0xd76eU,
		0xbed7U,0x6da5U,0xdb4aU,0xa69fU,0x5d35U,0xba6aU,0x64dfU,0xc9beU,0x8377U,0x16e5U,
		0x2dcaU,0x5b94U,0xb728U,0x7e5bU,0xfcb6U,0xe967U,0xc2c5U,0x9581U,0x3b09U,0x7612U,
		0xec24U,0xc843U,0x808dU,0x1111U,0x2222U,0x4444U,0x8888U,0x011bU,0x0236U,0x046cU,
		0x08d8U,0x11b0U,0x2360U,0x46c0U,0x8d80U,0x0b0bU,0x1616U,0x2c2cU,0x5858U,0xb0b0U,
		0x716bU,0xe2d6U,0xd5a7U,0xbb45U,0x6681U,0xcd02U,0x8a0fU,0x0415U,0x082aU,0x1054U,
		0x20a8U,0x4150U,0x82a0U,0x154bU,0x2a96U,0x552cU,0xaa58U,0x44bbU,0x8976U,0x02e7U,
		0x05ceU,0x0b9cU,0x1738U,0x2e70U,0x5ce0U,0xb9c0U,0x638bU,0xc716U,0x9e27U,0x2c45U,
		0x588aU,0xb114U,0x7223U,0xe446U,0xd887U,0xa105U,0x5201U,0xa402U,0x580fU,0xb01eU,
		0x7037U,0xe06eU,0xd0d7U,0xb1a5U,0x7341U,0xe682U,0xdd0fU,0xaa15U,0x4421U,0x8842U,
		0x008fU,0x011eU,0x023cU,0x0478U,0x08f0U,0x11e0U,0x23c0U,0x4780U,0x8f00U,0x0e0bU,
		0x1c16U,0x382cU,0x7058U,0xe0b0U,0xd16bU,0xb2ddU,0x75b1U,0xeb62U,0xc6cfU,0x9d95U,
		0x2b21U,0x5642U,0xac84U,0x4903U,0x9206U,0x3407U,0x680eU,0xd01cU,0xb033U,0x706dU,
		0xe0daU,0xd1bfU,0xb375U,0x76e1U,0xedc2U,0xcb8fU,0x8715U,0x1e21U,0x3c42U,0x7884U,
		0xf108U,0xf21bU,0xf43dU,0xf871U,0xe0e9U,0xd1d9U,0xb3b9U,0x7779U,0xeef2U,0xcdefU,
		0x8bd5U,0x07a1U,0x0f42U,0x1e84U,0x3d08U,0x7a10U,0xf420U,0xf84bU,0xe09dU,0xd131U,
		0xb269U,0x74d9U,0xe9b2U,0xc36fU,0x96d5U,0x3da1U,0x7b42U,0xf684U,0xfd03U,0xea0dU,
		0xc411U,0x9829U,0x2059U,0x40b2U,0x8164U,0x12c3U,0x2586U,0x4b0cU,0x9618U,0x3c3bU,
		0x7876U,0xf0ecU,0xf1d3U,0xf3adU,0xf751U,0xfea9U,0xed59U,0xcab9U,0x8579U,0x1af9U,
		0x35f2U,0x6be4U,0xd7c8U,0xbf9bU,0x6f3dU,0xde7aU,0xacffU,0x49f5U,0x93eaU,0x37dfU,
		0x6fbeU,0xdf7cU,0xaef3U,0x4dedU,0x9bdaU,0x27bfU,0x4f7eU,0x9efcU,0x2df3U,0x5be6U,
		0xb7ccU,0x7f93U,0xff26U,0xee47U,0xcc85U,0x8901U,0x0209U,0x0412U,0x0824U,0x1048U,
		0x2090U,0x4120U,0x8240U,0x148bU,0x2916U,0x522cU,0xa458U,0x58bbU,0xb176U,0x72e7U,
		0xe5ceU,0xdb97U,0xa725U,0x5e41U,0xbc82U,0x690fU,0xd21eU,0xb437U,0x7865U,0xf0caU,
		0xf19fU,0xf335U,0xf661U,0xfcc9U,0xe999U,0xc339U,0x9679U,0x3cf9U,0x79f2U,0xf3e4U,
		0xf7c3U,0xff8dU,0xef11U,0xce29U,0x8c59U,0x08b9U,0x1172U,0x22e4U,0x45c8U,0x8b90U,
		0x072bU,0x0e56U,0x1cacU,0x3958U,0x72b0U,0xe560U,0xdacbU,0xa59dU,0x5b31U,0xb662U,
		0x7ccfU,0xf99eU,0xe337U,0xd665U,0xbcc1U,0x6989U,0xd312U,0xb62fU,0x7c55U,0xf8aaU,
		0xe15fU,0xd2b5U,0xb561U,0x7ac9U,0xf592U,0xfb2fU,0xe655U,0xdca1U,0xa949U,0x4299U,
		0x8532U,0x1a6fU,0x34deU,0x69bcU,0xd378U,0xb6fbU,0x7dfdU,0xfbfaU,0xe7ffU,0xdff5U,
		0xafe1U,0x4fc9U,0x9f92U,0x2f2fU,0x5e5eU,0xbcbcU,0x6973U,0xd2e6U,0xb5c7U,0x7b85U,
		0xf70aU,0xfe1fU,0xec35U,0xc861U,0x80c9U,0x1199U,0x2332U,0x4664U,0x8cc8U,0x099bU,
		0x1336U,0x266cU,0x4cd8U,0x99b0U,0x236bU,0x46d6U,0x8dacU,0x0b53U,0x16a6U,0x2d4cU,
		0x5a98U,0xb530U,0x7a6bU,0xf4d6U,0xf9a7U,0xe345U,0xd681U,0xbd09U,0x6a19U,0xd432U,
		0xb86fU,0x60d5U,0xc1aaU,0x935fU,0x36b5U,0x6d6aU,0xdad4U,0xa5a3U,0x5b4dU,0xb69aU,
		0x7d3fU,0xfa7eU,0xe4f7U,0xd9e5U,0xa3c1U,0x5789U,0xaf12U,0x4e2fU,0x9c5eU,0x28b7U,
		0x516eU,0xa2dcU,0x55b3U,0xab66U,0x46c7U,0x8d8eU,0x0b17U,0x162eU,0x2c5cU,0x58b8U,
		0xb170U,0x72ebU,0xe5d6U,0xdba7U,0xa745U,0x5e81U,0xbd02U,0x6a0fU,0xd41eU,0xb837U,
		0x6065U,0xc0caU,0x919fU,0x3335U,0x666aU,0xccd4U,0x89a3U,0x034dU,0x069aU,0x0d34U,
		0x1a68U,0x34d0U,0x69a0U,0xd340U,0xb68bU,0x7d1dU,0xfa3aU,0xe47fU,0xd8f5U,0xa1e1U,
		0x53c9U,0xa792U,0x5f2fU,0xbe5eU,0x6cb7U,0xd96eU,0xa2d7U,0x55a5U,0xab4aU,0x469fU,
		0x8d3eU,0x0a77U,0x14eeU,0x29dcU,0x53b8U,0xa770U,0x5eebU,0xbdd6U,0x6ba7U,0xd74eU,
		0xbe97U,0x6d25U,0xda4aU,0xa49fU,0x5935U,0xb26aU,0x74dfU,0xe9beU,0xc377U,0x96e5U,
		0x3dc1U,0x7b82U,0xf704U,0xfe03U,0xec0dU,0xc811U,0x8029U,0x1059U,0x20b2U,0x4164U,
		0x82c8U,0x159bU,0x2b36U,0x566cU,0xacd8U,0x49bbU,0x9376U,0x36e7U,0x6dceU,0xdb9cU,
		0xa733U,0x5e6dU,0xbcdaU,0x69bfU,0xd37eU,0xb6f7U,0x7de5U,0xfbcaU,0xe79fU,0xdf35U,
		0xae61U,0x4cc9U,0x9992U,0x232fU,0x465eU,0x8cbcU,0x0973U,0x12e6U,0x25ccU,0x4b98U,
		0x9730U,0x3e6bU,0x7cd6U,0xf9acU,0xe353U,0xd6adU,0xbd51U,0x6aa9U,0xd552U,0xbaafU,
		0x6555U,0xcaaaU,0x855fU,0x1ab5U,0x356aU,0x6ad4U,0xd5a8U,0xbb5bU,0x66bdU,0xcd7aU,
		0x8affU,0x05f5U,0x0beaU,0x17d4U,0x2fa8U,0x5f50U,0xbea0U,0x6d4bU,0xda96U,0xa527U,
		0x5a45U,0xb48aU,0x791fU,0xf23eU,0xf477U,0xf8e5U,0xe1c1U,0xd389U,0xb719U,0x7e39U,
		0xfc72U,0xe8efU,0xc1d5U,0x93a1U,0x3749U,0x6e92U,0xdd24U,0xaa43U,0x448dU,0x891aU,
		0x023fU,0x047eU,0x08fcU,0x11f8U,0x23f0U,0x47e0U,0x8fc0U,0x0f8bU,0x1f16U,0x3e2cU,
		0x7c58U,0xf8b0U,0xe16bU,0xd2ddU,0xb5b1U,0x7b69U,0xf6d2U,0xfdafU,0xeb55U,0xc6a1U,
		0x9d49U,0x2a99U,0x5532U,0xaa64U,0x44c3U,0x8986U,0x0307U,0x060eU,0x0c1cU,0x1838U,
		0x3070U,0x60e0U,0xc1c0U,0x938bU,0x371dU,0x6e3aU,0xdc74U,0xa8e3U,0x41cdU,0x839aU,
		0x173fU,0x2e7eU,0x5cfcU,0xb9f8U,0x63fbU,0xc7f6U,0x9fe7U,0x2fc5U,0x5f8aU,0xbf14U,
		0x6e23U,0xdc46U,0xa887U,0x4105U,0x820aU,0x141fU,0x283eU,0x507cU,0xa0f8U,0x51fbU,
		0xa3f6U,0x57e7U,0xafceU,0x4f97U,0x9f2eU,0x2e57U,0x5caeU,0xb95cU,0x62b3U,0xc566U,
		0x9ac7U,0x2585U,0x4b0aU,0x9614U,0x3c23U,0x7846U,0xf08cU,0xf113U,0xf22dU,0xf451U,
		0xf8a9U,0xe159U,0xd2b9U,0xb579U,0x7af9U,0xf5f2U,0xfbefU,0xe7d5U,0xdfa1U,0xaf49U,
		0x4e99U,0x9d32U,0x2a6fU,0x54deU,0xa9bcU,0x4373U,0x86e6U,0x1dc7U,0x3b8eU,0x771cU,
		0xee38U,0xcc7bU,0x88fdU,0x01f1U,0x03e2U,0x07c4U,0x0f88U,0x1f10U,0x3e20U,0x7c40U,
		0xf880U,0xe10bU,0xd21dU,0xb431U,0x7869U,0xf0d2U,0xf1afU,0xf355U,0xf6a1U,0xfd49U,
		0xea99U,0xc539U,0x9a79U,0x24f9U,0x49f2U,0x93e4U,0x37c3U,0x6f86U,0xdf0cU,0xae13U,
		0x4c2dU,0x985aU,0x20bfU,0x417eU,0x82fcU,0x15f3U,0x2be6U,0x57ccU,0xaf98U,0x4f3bU,
		0x9e76U,0x2ce7U,0x59ceU,0xb39cU,0x7733U,0xee66U,0xccc7U,0x8985U,0x0301U,0x0602U,
		0x0c04U,0x1808U,0x3010U,0x6020U,0xc040U,0x908bU,0x311dU,0x623aU,0xc474U,0x98e3U,
		0x21cdU,0x439aU,0x8734U,0x1e63U,0x3cc6U,0x798cU,0xf318U,0xf63bU,0xfc7dU,0xe8f1U,
		0xc1e9U,0x93d9U,0x37b9U,0x6f72U,0xdee4U,0xadc3U,0x4b8dU,0x971aU,0x3e3fU,0x7c7eU,
		0xf8fcU,0xe1f3U,0xd3edU,0xb7d1U,0x7fa9U,0xff52U,0xeeafU,0xcd55U,0x8aa1U,0x0549U,
		0x0a92U,0x1524U,0x2a48U,0x5490U,0xa920U,0x424bU,0x8496U,0x1927U,0x324eU,0x649cU,
		0xc938U,0x827bU,0x14fdU,0x29faU,0x53f4U,0xa7e8U,0x5fdbU,0xbfb6U,0x6f67U,0xdeceU,
		0xad97U,0x4b25U,0x964aU,0x3c9fU,0x793eU,0xf27cU,0xf4f3U,0xf9edU,0xe3d1U,0xd7a9U,
		0xbf59U,0x6eb9U,0xdd72U,0xaaefU,0x45d5U,0x8baaU,0x075fU,0x0ebeU,0x1d7cU,0x3af8U,
		0x75f0U,0xebe0U,0xc7cbU,0x9f9dU,0x2f31U,0x5e62U,0xbcc4U,0x6983U,0xd306U,0xb607U,
		0x7c05U,0xf80aU,0xe01fU,0xd035U,0xb061U,0x70c9U,0xe192U,0xd32fU,0xb655U,0x7ca1U,
		0xf942U,0xe28fU,0xd515U,0xba21U,0x6449U,0xc892U,0x812fU,0x1255U,0x24aaU,0x4954U,
		0x92a8U,0x355bU,0x6ab6U,0xd56cU,0xbad3U,0x65adU,0xcb5aU,0x86bfU,0x1d75U,0x3aeaU,
		0x75d4U,0xeba8U,0xc75bU,0x9ebdU,0x2d71U,0x5ae2U,0xb5c4U,0x7b83U,0xf706U,0xfe07U,
		0xec05U,0xc801U,0x8009U,0x1019U,0x2032U,0x4064U,0x80c8U,0x119bU,0x2336U,0x466cU,
		0x8cd8U,0x09bbU,0x1376U,0x26ecU,0x4dd8U,0x9bb0U,0x276bU,0x4ed6U,0x9dacU,0x2b53U,
		0x56a6U,0xad4cU,0x4a93U,0x9526U,0x3a47U,0x748eU,0xe91cU,0xc233U,0x946dU,0x38d1U,
		0x71a2U,0xe344U,0xd683U,0xbd0dU,0x6a11U,0xd422U,0xb84fU,0x6095U,0xc12aU,0x925fU,
		0x34b5U,0x696aU,0xd2d4U,0xb5a3U,0x7b4dU,0xf69aU,0xfd3fU,0xea75U,0xc4e1U,0x99c9U,
		0x2399U,0x4732U,0x8e64U,0x0cc3U,0x1986U,0x330cU,0x6618U,0xcc30U,0x886bU,0x00ddU,
		0x01baU,0x0374U,0x06e8U,0x0dd0U,0x1ba0U,0x3740U,0x6e80U,0xdd00U,0xaa0bU,0x441dU,
		0x883aU,0x007fU,0x00feU,0x01fcU,0x03f8U,0x07f0U,0x0fe0U,0x1fc0U,0x3f80U,0x7f00U,
		0xfe00U,0xec0bU,0xc81dU,0x8031U,0x1069U,0x20d2U,0x41a4U,0x8348U,0x169bU,0x2d36U,
		0x5a6cU,0xb4d8U,0x79bbU,0xf376U,0xf6e7U,0xfdc5U,0xeb81U,0xc709U,0x9e19U,0x2c39U,
		0x5872U,0xb0e4U,0x71c3U,0xe386U,0xd707U,0xbe05U,0x6c01U,0xd802U,0xa00fU,0x5015U,
		0xa02aU,0x505fU,0xa0beU,0x5177U,0xa2eeU,0x55d7U,0xabaeU,0x4757U,0x8eaeU,0x0d57U,
		0x1aaeU,0x355cU,0x6ab8U,0xd570U,0xbaebU,0x65ddU,0xcbbaU,0x877fU,0x1ef5U,0x3deaU,
		0x7bd4U,0xf7a8U,0xff5bU,0xeebdU,0xcd71U,0x8ae9U,0x05d9U,0x0bb2U,0x1764U,0x2ec8U,
		0x5d90U,0xbb20U,0x664bU,0xcc96U,0x8927U,0x0245U,0x048aU,0x0914U,0x1228U,0x2450U,
		0x48a0U,0x9140U,0x328bU,0x6516U,0xca2cU,0x8453U,0x18adU,0x315aU,0x62b4U,0xc568U,
		0x9adbU,0x25bdU,0x4b7aU,0x96f4U,0x3de3U,0x7bc6U,0xf78cU,0xff13U,0xee2dU,0xcc51U,
		0x88a9U,0x0159U,0x02b2U,0x0564U,0x0ac8U,0x1590U,0x2b20U,0x5640U,0xac80U,0x490bU,
		0x9216U,0x3427U,0x684eU,0xd09cU,0xb133U,0x726dU,0xe4daU,0xd9bfU,0xa375U,0x56e1U,
		0xadc2U,0x4b8fU,0x971eU,0x3e37U,0x7c6eU,0xf8dcU,0xe1b3U,0xd36dU,0xb6d1U,0x7da9U,
		0xfb52U,0xe6afU,0xdd55U,0xaaa1U,0x4549U,0x8a92U,0x052fU,0x0a5eU,0x14bcU,0x2978U,
		0x52f0U,0xa5e0U,0x5bcbU,0xb796U,0x7f27U,0xfe4eU,0xec97U,0xc925U,0x8241U,0x1489U,
		0x2912U,0x5224U,0xa448U,0x589bU,0xb136U,0x7267U,0xe4ceU,0xd997U,0xa325U,0x5641U,
		0xac82U,0x490fU,0x921eU,0x3437U,0x686eU,0xd0dcU,0xb1b3U,0x736dU,0xe6daU,0xddbfU,
		0xab75U,0x46e1U,0x8dc2U,0x0b8fU,0x171eU,0x2e3cU,0x5c78U,0xb8f0U,0x61ebU,0xc3d6U,
		0x97a7U,0x3f45U,0x7e8aU,0xfd14U,0xea23U,0xc44dU,0x9891U,0x2129U,0x4252U,0x84a4U,
		0x1943U,0x3286U,0x650cU,0xca18U,0x843bU,0x187dU,0x30faU,0x61f4U,0xc3e8U,0x97dbU,
		0x3fbdU,0x7f7aU,0xfef4U,0xede3U,0xcbcdU,0x8791U,0x1f29U,0x3e52U,0x7ca4U,0xf948U,
		0xe29bU,0xd53dU,0xba71U,0x64e9U,0xc9d2U,0x83afU,0x1755U,0x2eaaU,0x5d54U,0xbaa8U,
		0x655bU,0xcab6U,0x8567U,0x1ac5U,0x358aU,0x6b14U,0xd628U,0xbc5bU,0x68bdU,0xd17aU,
		0xb2ffU,0x75f5U,0xebeaU,0xc7dfU,0x9fb5U,0x2f61U,0x5ec2U,0xbd84U,0x6b03U,0xd606U,
		0xbc07U,0x6805U,0xd00aU,0xb01fU,0x7035U,0xe06aU,0xd0dfU,0xb1b5U,0x7361U,0xe6c2U,
		0xdd8fU,0xab15U,0x4621U,0x8c42U,0x088fU,0x111eU,0x223cU,0x4478U,0x88f0U,0x01ebU,
		0x03d6U,0x07acU,0x0f58U,0x1eb0U,0x3d60U,0x7ac0U,0xf580U,0xfb0bU,0xe61dU,0xdc31U,
		0xa869U,0x40d9U,0x81b2U,0x136fU,0x26deU,0x4dbcU,0x9b78U,0x26fbU,0x4df6U,0x9becU,
		0x27d3U,0x4fa6U,0x9f4cU,0x2e93U,0x5d26U,0xba4cU,0x6493U,0xc926U,0x8247U,0x1485U,
		0x290aU,0x5214U,0xa428U,0x585bU,0xb0b6U,0x7167U,0xe2ceU,0xd597U,0xbb25U,0x6641U,
		0xcc82U,0x890fU,0x0215U,0x042aU,0x0854U,0x10a8U,0x2150U,0x42a0U,0x8540U,0x1a8bU,
		0x3516U,0x6a2cU,0xd458U,0xb8bbU,0x617dU,0xc2faU,0x95ffU,0x3bf5U,0x77eaU,0xefd4U,
		0xcfa3U,0x8f4dU,0x0e91U,0x1d22U,0x3a44U,0x7488U,0xe910U,0xc22bU,0x945dU,0x38b1U,
		0x7162U,0xe2c4U,0xd583U,0xbb0dU,0x6611U,0xcc22U,0x884fU,0x0095U,0x012aU,0x0254U,
		0x04a8U,0x0950U,0x12a0U,0x2540U,0x4a80U,0x9500U,0x3a0bU,0x7416U,0xe82cU,0xc053U,
		0x90adU,0x3151U,0x62a2U,0xc544U,0x9a83U,0x250dU,0x4a1aU,0x9434U,0x3863U,0x70c6U,
		0xe18cU,0xd313U,0xb62dU,0x7c51U,0xf8a2U,0xe14fU,0xd295U,0xb521U,0x7a49U,0xf492U,
		0xf92fU,0xe255U,0xd4a1U,0xb949U,0x6299U,0xc532U,0x9a6fU,0x24d5U,0x49aaU,0x9354U,
		0x36a3U,0x6d46U,0xda8cU,0xa513U,0x5a2dU,0xb45aU,0x78bfU,0xf17eU,0xf2f7U,0xf5e5U,
		0xfbc1U,0xe789U,0xdf19U,0xae39U,0x4c79U,0x98f2U,0x21efU,0x43deU,0x87bcU,0x1f73U,
		0x3ee6U,0x7dccU,0xfb98U,0xe73bU,0xde7dU,0xacf1U,0x49e9U,0x93d2U,0x37afU,0x6f5eU,
		0xdebcU,0xad73U,0x4aedU,0x95daU,0x3bbfU,0x777eU,0xeefcU,0xcdf3U,0x8bedU,0x07d1U,
		0x0fa2U,0x1f44U,0x3e88U,0x7d10U,0xfa20U,0xe44bU,0xd89dU,0xa131U,0x5269U,0xa4d2U,
		0x59afU,0xb35eU,0x76b7U,0xed6eU,0xcad7U,0x85a5U,0x1b41U,0x3682U,0x6d04U,0xda08U,
		0xa41bU,0x583dU,0xb07aU,0x70ffU,0xe1feU,0xd3f7U,0xb7e5U,0x7fc1U,0xff82U,0xef0fU,
		0xce15U,0x8c21U,0x0849U,0x1092U,0x2124U,0x4248U,0x8490U,0x192bU,0x3256U,0x64acU,
		0xc958U,0x82bbU,0x157dU,0x2afaU,0x55f4U,0xabe8U,0x47dbU,0x8fb6U,0x0f67U,0x1eceU,
		0x3d9cU,0x7b38U,0xf670U,0xfcebU,0xe9ddU,0xc3b1U,0x9769U,0x3ed9U,0x7db2U,0xfb64U,
		0xe6c3U,0xdd8dU,0xab11U,0x4629U,0x8c52U,0x08afU,0x115eU,0x22bcU,0x4578U,0x8af0U,
		0x05ebU,0x0bd6U,0x17acU,0x2f58U,0x5eb0U,0xbd60U,0x6acbU,0xd596U,0xbb27U,0x6645U,
		0xcc8aU,0x891fU,0x0235U,0x046aU,0x08d4U,0x11a8U,0x2350U,0x46a0U,0x8d40U,0x0a8bU,
		0x1516U,0x2a2cU,0x5458U,0xa8b0U,0x416bU,0x82d6U,0x15a7U,0x2b4eU,0x569cU,0xad38U,
		0x4a7bU,0x94f6U,0x39e7U,0x73ceU,0xe79cU,0xdf33U,0xae6dU,0x4cd1U,0x99a2U,0x234fU,
		0x469eU,0x8d3cU,0x0a73U,0x14e6U,0x29ccU,0x5398U,0xa730U,0x5e6bU,0xbcd6U,0x69a7U,
		0xd34eU,0xb697U,0x7d25U,0xfa4aU,0xe49fU,0xd935U,0xa261U,0x54c9U,0xa992U,0x432fU,
		0x865eU,0x1cb7U,0x396eU,0x72dcU,0xe5b8U,0xdb7bU,0xa6fdU,0x5df1U,0xbbe2U,0x67cfU,
		0xcf9eU,0x8f37U,0x0e65U,0x1ccaU,0x3994U,0x7328U,0xe650U,0xdcabU,0xa95dU,0x42b1U,
		0x8562U,0x1acfU,0x359eU,0x6b3cU,0xd678U,0xbcfbU,0x69fdU,0xd3faU,0xb7ffU,0x7ff5U,
		0xffeaU,0xefdfU,0xcfb5U,0x8f61U,0x0ec9U,0x1d92U,0x3b24U,0x7648U,0xec90U,0xc92bU,
		0x825dU,0x14b1U,0x2962U,0x52c4U,0xa588U,0x5b1bU,0xb636U,0x7c67U,0xf8ceU,0xe197U,
		0xd325U,0xb641U,0x7c89U,0xf912U,0xe22fU,0xd455U,0xb8a1U,0x6149U,0xc292U,0x952fU,
		0x3a55U,0x74aaU,0xe954U,0xc2a3U,0x954dU,0x3a91U,0x7522U,0xea44U,0xc483U,0x990dU,
		0x2211U,0x4422U,0x8844U,0x0083U,0x0106U,0x020cU,0x0418U,0x0830U,0x1060U,0x20c0U,
		0x4180U,0x8300U,0x160bU,0x2c16U,0x582cU,0xb058U,0x70bbU,0xe176U,0xd2e7U,0xb5c5U,
		0x7b81U,0xf702U,0xfe0fU,0xec15U,0xc821U,0x8049U,0x1099U,0x2132U,0x4264U,0x84c8U,
		0x199bU,0x3336U,0x666cU,0xccd8U,0x89bbU,0x037dU,0x06faU,0x0df4U,0x1be8U,0x37d0U,
		0x6fa0U,0xdf40U,0xae8bU,0x4d1dU,0x9a3aU,0x247fU,0x48feU,0x91fcU,0x33f3U,0x67e6U,
		0xcfccU,0x8f93U,0x0f2dU,0x1e5aU,0x3cb4U,0x7968U,0xf2d0U,0xf5abU,0xfb5dU,0xe6b1U,
		0xdd69U,0xaad9U,0x45b9U,0x8b72U,0x06efU,0x0ddeU,0x1bbcU,0x3778U,0x6ef0U,0xdde0U,
		0xabcbU,0x479dU,0x8f3aU,0x0e7fU,0x1cfeU,0x39fcU,0x73f8U,0xe7f0U,0xdfebU,0xafddU,
		0x4fb1U,0x9f62U,0x2ecfU,0x5d9eU,0xbb3cU,0x6673U,0xcce6U,0x89c7U,0x0385U,0x070aU,
		0x0e14U,0x1c28U,0x3850U,0x70a0U,0xe140U,0xd28bU,0xb51dU,0x7a31U,0xf462U,0xf8cfU,
		0xe195U,0xd321U,0xb649U,0x7c99U,0xf932U,0xe26fU,0xd4d5U,0xb9a1U,0x6349U,0xc692U,
		0x9d2fU,0x2a55U,0x54aaU,0xa954U,0x42a3U,0x8546U,0x1a87U,0x350eU,0x6a1cU,0xd438U,
		0xb87bU,0x60fdU,0xc1faU,0x93ffU,0x37f5U,0x6feaU,0xdfd4U,0xafa3U,0x4f4dU,0x9e9aU,
		0x2d3fU,0x5a7eU,0xb4fcU,0x79f3U,0xf3e6U,0xf7c7U,0xff85U,0xef01U,0xce09U,0x8c19U,
		0x0839U,0x1072U,0x20e4U,0x41c8U,0x8390U,0x172bU,0x2e56U,0x5cacU,0xb958U,0x62bbU,
		0xc576U,0x9ae7U,0x25c5U,0x4b8aU,0x9714U,0x3e23U,0x7c46U,0xf88cU,0xe113U,0xd22dU,
		0xb451U,0x78a9U,0xf152U,0xf2afU,0xf555U,0xfaa1U,0xe549U,0xda99U,0xa539U,0x5a79U,
		0xb4f2U,0x79efU,0xf3deU,0xf7b7U,0xff65U,0xeec1U,0xcd89U,0x8b19U,0x0639U,0x0c72U,
		0x18e4U,0x31c8U,0x6390U,0xc720U,0x9e4bU,0x2c9dU,0x593aU,0xb274U,0x74e3U,0xe9c6U,
		0xc387U,0x9705U,0x3e01U,0x7c02U,0xf804U,0xe003U,0xd00dU,0xb011U,0x7029U,0xe052U,
		0xd0afU,0xb155U,0x72a1U,0xe542U,0xda8fU,0xa515U,0x5a21U,0xb442U,0x788fU,0xf11eU,
		0xf237U,0xf465U,0xf8c1U,0xe189U,0xd319U,0xb639U,0x7c79U,0xf8f2U,0xe1efU,0xd3d5U,
		0xb7a1U,0x7f49U,0xfe92U,0xed2fU,0xca55U,0x84a1U,0x1949U,0x3292U,0x6524U,0xca48U,
		0x849bU,0x193dU,0x327aU,0x64f4U,0xc9e8U,0x83dbU,0x17bdU,0x2f7aU,0x5ef4U,0xbde8U,
		0x6bdbU,0xd7b6U,0xbf67U,0x6ec5U,0xdd8aU,0xab1fU,0x4635U,0x8c6aU,0x08dfU,0x11beU,
		0x237cU,0x46f8U,0x8df0U,0x0bebU,0x17d6U,0x2facU,0x5f58U,0xbeb0U,0x6d6bU,0xdad6U,
		0xa5a7U,0x5b45U,0xb68aU,0x7d1fU,0xfa3eU,0xe477U,0xd8e5U,0xa1c1U,0x5389U,0xa712U,
		0x5e2fU,0xbc5eU,0x68b7U,0xd16eU,0xb2d7U,0x75a5U,0xeb4aU,0xc69fU,0x9d35U,0x2a61U,
		0x54c2U,0xa984U,0x4303U,0x8606U,0x1c07U,0x380eU,0x701cU,0xe038U,0xd07bU,0xb0fdU,
		0x71f1U,0xe3e2U,0xd7cfU,0xbf95U,0x6f21U,0xde42U,0xac8fU,0x4915U,0x922aU,0x345fU,
		0x68beU,0xd17cU,0xb2f3U,0x75edU,0xebdaU,0xc7bfU,0x9f75U,0x2ee1U,0x5dc2U,0xbb84U,
		0x6703U,0xce06U,0x8c07U,0x0805U,0x100aU,0x2014U,0x4028U,0x8050U,0x10abU,0x2156U,
		0x42acU,0x8558U,0x1abbU,0x3576U,0x6aecU,0xd5d8U,0xbbbbU,0x677dU,0xcefaU,0x8dffU,
		0x0bf5U,0x17eaU,0x2fd4U,0x5fa8U,0xbf50U,0x6eabU,0xdd56U,0xaaa7U,0x4545U,0x8a8aU,
		0x051fU,0x0a3eU,0x147cU,0x28f8U,0x51f0U,0xa3e0U,0x57cbU,0xaf96U,0x4f27U,0x9e4eU,
		0x2c97U,0x592eU,0xb25cU,0x74b3U,0xe966U,0xc2c7U,0x9585U,0x3b01U,0x7602U,0xec04U,
		0xc803U,0x800dU,0x1011U,0x2022U,0x4044U,0x8088U,0x111bU,0x2236U,0x446cU,0x88d8U,
		0x01bbU,0x0376U,0x06ecU,0x0dd8U,0x1bb0U,0x3760U,0x6ec0U,0xdd80U,0xab0bU,0x461dU,
		0x8c3aU,0x087fU,0x10feU,0x21fcU,0x43f8U,0x87f0U,0x1febU,0x3fd6U,0x7facU,0xff58U,
		0xeebbU,0xcd7dU,0x8af1U,0x05e9U,0x0bd2U,0x17a4U,0x2f48U,0x5e90U,0xbd20U,0x6a4bU,
		0xd496U,0xb927U,0x6245U,0xc48aU,0x991fU,0x2235U,0x446aU,0x88d4U,0x01a3U,0x0346U,
		0x068cU,0x0d18U,0x1a30U,0x3460U,0x68c0U,0xd180U,0xb30bU,0x761dU,0xec3aU,0xc87fU,
		0x80f5U,0x11e1U,0x23c2U,0x4784U,0x8f08U,0x0e1bU,0x1c36U,0x386cU,0x70d8U,0xe1b0U,
		0xd36bU,0xb6ddU,0x7db1U,0xfb62U,0xe6cfU,0xdd95U,0xab21U,0x4649U,0x8c92U,0x092fU,
		0x125eU,0x24bcU,0x4978U,0x92f0U,0x35ebU,0x6bd6U,0xd7acU,0xbf53U,0x6eadU,0xdd5aU,
		0xaabfU,0x4575U,0x8aeaU,0x05dfU,0x0bbeU,0x177cU,0x2ef8U,0x5df0U,0xbbe0U,0x67cbU,
		0xcf96U,0x8f27U,0x0e45U,0x1c8aU,0x3914U,0x7228U,0xe450U,0xd8abU,0xa15dU,0x52b1U,
		0xa562U,0x5acfU,0xb59eU,0x7b37U,0xf66eU,0xfcd7U,0xe9a5U,0xc341U,0x9689U,0x3d19U,
		0x7a32U,0xf464U,0xf8c3U,0xe18dU,0xd311U,0xb629U,0x7c59U,0xf8b2U,0xe16fU,0xd2d5U,
		0xb5a1U,0x7b49U,0xf692U,0xfd2fU,0xea55U,0xc4a1U,0x9949U,0x2299U,0x4532U,0x8a64U,
		0x04c3U,0x0986U,0x130cU,0x2618U,0x4c30U,0x9860U,0x20cbU,0x4196U,0x832cU,0x1653U,
		0x2ca6U,0x594cU,0xb298U,0x753bU,0xea76U,0xc4e7U,0x99c5U,0x2381U,0x4702U,0x8e04U,
		0x0c03U,0x1806U,0x300cU,0x6018U,0xc030U,0x906bU,0x30ddU,0x61baU,0xc374U,0x96e3U,
		0x3dcdU,0x7b9aU,0xf734U,0xfe63U,0xeccdU,0xc991U,0x8329U,0x1659U,0x2cb2U,0x5964U,
		0xb2c8U,0x759bU,0xeb36U,0xc667U,0x9cc5U,0x2981U,0x5302U,0xa604U,0x5c03U,0xb806U,
		0x6007U,0xc00eU,0x9017U,0x3025U,0x604aU,0xc094U,0x9123U,0x324dU,0x649aU,0xc934U,
		0x8263U,0x14cdU,0x299aU,0x5334U,0xa668U,0x5cdbU,0xb9b6U,0x6367U,0xc6ceU,0x9d97U,
		0x2b25U,0x564aU,0xac94U,0x4923U,0x9246U,0x3487U,0x690eU,0xd21cU,0xb433U,0x786dU,
		0xf0daU,0xf1bfU,0xf375U,0xf6e1U,0xfdc9U,0xeb99U,0xc739U,0x9e79U,0x2cf9U,0x59f2U,
		0xb3e4U,0x77c3U,0xef86U,0xcf07U,0x8e05U,0x0c01U,0x1802U,0x3004U,0x6008U,0xc010U,
		0x902bU,0x305dU,0x60baU,0xc174U,0x92e3U,0x35cdU,0x6b9aU,0xd734U,0xbe63U,0x6ccdU,
		0xd99aU,0xa33fU,0x5675U,0xaceaU,0x49dfU,0x93beU,0x3777U,0x6eeeU,0xdddcU,0xabb3U,
		0x476dU,0x8edaU,0x0dbfU,0x1b7eU,0x36fcU,0x6df8U,0xdbf0U,0xa7ebU,0x5fddU,0xbfbaU,
		0x6f7fU,0xdefeU,0xadf7U,0x4be5U,0x97caU,0x3f9fU,0x7f3eU,0xfe7cU,0xecf3U,0xc9edU,
		0x83d1U,0x17a9U,0x2f52U,0x5ea4U,0xbd48U,0x6a9bU,0xd536U,0xba67U,0x64c5U,0xc98aU,
		0x831fU,0x1635U,0x2c6aU,0x58d4U,0xb1a8U,0x735bU,0xe6b6U,0xdd67U,0xaac5U,0x4581U,
		0x8b02U,0x060fU,0x0c1eU,0x183cU,0x3078U,0x60f0U,0xc1e0U,0x93cbU,0x379dU,0x6f3aU,
		0xde74U,0xace3U,0x49cdU,0x939aU,0x373fU,0x6e7eU,0xdcfcU,0xa9f3U,0x43edU,0x87daU,
		0x1fbfU,0x3f7eU,0x7efcU,0xfdf8U,0xebfbU,0xc7fdU,0x9ff1U,0x2fe9U,0x5fd2U,0xbfa4U,
		0x6f43U,0xde86U,0xad07U,0x4a05U,0x940aU,0x381fU,0x703eU,0xe07cU,0xd0f3U,0xb1edU,
		0x73d1U,0xe7a2U,0xdf4fU,0xae95U,0x4d21U,0x9a42U,0x248fU,0x491eU,0x923cU,0x3473U,
		0x68e6U,0xd1ccU,0xb393U,0x772dU,0xee5aU,0xccbfU,0x8975U,0x02e1U,0x05c2U,0x0b84U,
		0x1708U,0x2e10U,0x5c20U,0xb840U,0x608bU,0xc116U,0x9227U,0x3445U,0x688aU,0xd114U,
		0xb223U,0x744dU,0xe89aU,0xc13fU,0x9275U,0x34e1U,0x69c2U,0xd384U,0xb703U,0x7e0dU,
		0xfc1aU,0xe83fU,0xc075U,0x90e1U,0x31c9U,0x6392U,0xc724U,0x9e43U,0x2c8dU,0x591aU,
		0xb234U,0x7463U,0xe8c6U,0xc187U,0x9305U,0x3601U,0x6c02U,0xd804U,0xa003U,0x500dU,
		0xa01aU,0x503fU,0xa07eU,0x50f7U,0xa1eeU,0x53d7U,0xa7aeU,0x5f57U,0xbeaeU,0x6d57U,
		0xdaaeU,0xa557U,0x5aa5U,0xb54aU,0x7a9fU,0xf53eU,0xfa77U,0xe4e5U,0xd9c1U,0xa389U,
		0x5719U,0xae32U,0x4c6fU,0x98deU,0x21b7U,0x436eU,0x86dcU,0x1db3U,0x3b66U,0x76ccU,
		0xed98U,0xcb3bU,0x867dU,0x1cf1U,0x39e2U,0x73c4U,0xe788U,0xdf1bU,0xae3dU,0x4c71U,
		0x98e2U,0x21cfU,0x439eU,0x873cU,0x1e73U,0x3ce6U,0x79ccU,0xf398U,0xf73bU,0xfe7dU,
		0xecf1U,0xc9e9U,0x83d9U,0x17b9U,0x2f72U,0x5ee4U,0xbdc8U,0x6b9bU,0xd736U,0xbe67U,
		0x6cc5U,0xd98aU,0xa31fU,0x5635U,0xac6aU,0x48dfU,0x91beU,0x3377U,0x66eeU,0xcddcU,
		0x8bb3U,0x076dU,0x0edaU,0x1db4U,0x3b68U,0x76d0U,0xeda0U,0xcb4bU,0x869dU,0x1d31U,
		0x3a62U,0x74c4U,0xe988U,0xc31bU,0x963dU,0x3c71U,0x78e2U,0xf1c4U,0xf383U,0xf70dU,
		0xfe11U,0xec29U,0xc859U,0x80b9U,0x1179U,0x22f2U,0x45e4U,0x8bc8U,0x079bU,0x0f36U,
		0x1e6cU,0x3cd8U,0x79b0U,0xf360U,0xf6cbU,0xfd9dU,0xeb31U,0xc669U,0x9cd9U,0x29b9U,
		0x5372U,0xa6e4U,0x5dc3U,0xbb86U,0x6707U,0xce0eU,0x8c17U,0x0825U,0x104aU,0x2094U,
		0x4128U,0x8250U,0x14abU,0x2956U,0x52acU,0xa558U,0x5abbU,0xb576U,0x7ae7U,0xf5ceU,
		0xfb97U,0xe725U,0xde41U,0xac89U,0x4919U,0x9232U,0x346fU,0x68deU,0xd1bcU,0xb373U,
		0x76edU,0xeddaU,0xcbbfU,0x8775U,0x1ee1U,0x3dc2U,0x7b84U,0xf708U,0xfe1bU,0xec3dU,
		0xc871U,0x80e9U,0x11d9U,0x23b2U,0x4764U,0x8ec8U,0x0d9bU,0x1b36U,0x366cU,0x6cd8U,
		0xd9b0U,0xa36bU,0x56ddU,0xadbaU,0x4b7fU,0x96feU,0x3df7U,0x7beeU,0xf7dcU,0xffb3U,
		0xef6dU,0xced1U,0x8da9U,0x0b59U,0x16b2U,0x2d64U,0x5ac8U,0xb590U,0x7b2bU,0xf656U,
		0xfca7U,0xe945U,0xc281U,0x9509U,0x3a19U,0x7432U,0xe864U,0xc0c3U,0x918dU,0x3311U,
		0x6622U,0xcc44U,0x8883U,0x010dU,0x021aU,0x0434U,0x0868U,0x10d0U,0x21a0U,0x4340U,
		0x8680U,0x1d0bU,0x3a16U,0x742cU,0xe858U,0xc0bbU,0x917dU,0x32f1U,0x65e2U,0xcbc4U,
		0x8783U,0x1f0dU,0x3e1aU,0x7c34U,0xf868U,0xe0dbU,0xd1bdU,0xb371U,0x76e9U,0xedd2U,
		0xcbafU,0x8755U,0x1ea1U,0x3d42U,0x7a84U,0xf508U,0xfa1bU,0xe43dU,0xd871U,0xa0e9U,
		0x51d9U,0xa3b2U,0x576fU,0xaedeU,0x4db7U,0x9b6eU,0x26d7U,0x4daeU,0x9b5cU,0x26b3U,
		0x4d66U,0x9accU,0x2593U,0x4b26U,0x964cU,0x3c93U,0x7926U,0xf24cU,0xf493U,0xf92dU,
		0xe251U,0xd4a9U,0xb959U,0x62b9U,0xc572U,0x9aefU,0x25d5U,0x4baaU,0x9754U,0x3ea3U,
		0x7d46U,0xfa8cU,0xe513U,0xda2dU,0xa451U,0x58a9U,0xb152U,0x72afU,0xe55eU,0xdab7U,
		0xa565U,0x5ac1U,0xb582U,0x7b0fU,0xf61eU,0xfc37U,0xe865U,0xc0c1U,0x9189U,0x3319U,
		0x6632U,0xcc64U,0x88c3U,0x018dU,0x031aU,0x0634U,0x0c68U,0x18d0U,0x31a0U,0x6340U,
		0xc680U,0x9d0bU,0x2a1dU,0x543aU,0xa874U,0x40e3U,0x81c6U,0x1387U,0x270eU,0x4e1cU,
		0x9c38U,0x287bU,0x50f6U,0xa1ecU,0x53d3U,0xa7a6U,0x5f47U,0xbe8eU,0x6d17U,0xda2eU,
		0xa457U,0x58a5U,0xb14aU,0x729fU,0xe53eU,0xda77U,0xa4e5U,0x59c1U,0xb382U,0x770fU,
		0xee1eU,0xcc37U,0x8865U,0x00c1U,0x0182U,0x0304U,0x0608U,0x0c10U,0x1820U,0x3040U,
		0x6080U,0xc100U,0x920bU,0x341dU,0x683aU,0xd074U,0xb0e3U,0x71cdU,0xe39aU,0xd73fU,
		0xbe75U,0x6ce1U,0xd9c2U,0xa38fU,0x5715U,0xae2aU,0x4c5fU,0x98beU,0x2177U,0x42eeU,
		0x85dcU,0x1bb3U,0x3766U,0x6eccU,0xdd98U,0xab3bU,0x467dU,0x8cfaU,0x09ffU,0x13feU,
		0x27fcU,0x4ff8U,0x9ff0U,0x2febU,0x5fd6U,0xbfacU,0x6f53U,0xdea6U,0xad47U,0x4a85U,
		0x950aU,0x3a1fU,0x743eU,0xe87cU,0xc0f3U,0x91edU,0x33d1U,0x67a2U,0xcf44U,0x8e83U,
		0x0d0dU,0x1a1aU,0x3434U,0x6868U,0xd0d0U,0xb1abU,0x735dU,0xe6baU,0xdd7fU,0xaaf5U,
		0x45e1U,0x8bc2U,0x078fU,0x0f1eU,0x1e3cU,0x3c78U,0x78f0U,0xf1e0U,0xf3cbU,0xf79dU,
		0xff31U,0xee69U,0xccd9U,0x89b9U,0x0379U,0x06f2U,0x0de4U,0x1bc8U,0x3790U,0x6f20U,
		0xde40U,0xac8bU,0x491dU,0x923aU,0x347fU,0x68feU,0xd1fcU,0xb3f3U,0x77edU,0xefdaU,
		0xcfbfU,0x8f75U,0x0ee1U,0x1dc2U,0x3b84U,0x7708U,0xee10U,0xcc2bU,0x885dU,0x00b1U,
		0x0162U,0x02c4U,0x0588U,0x0b10U,0x1620U,0x2c40U,0x5880U,0xb100U,0x720bU,0xe416U,
		0xd827U,0xa045U,0x5081U,0xa102U,0x520fU,0xa41eU,0x5837U,0xb06eU,0x70d7U,0xe1aeU,
		0xd357U,0xb6a5U,0x7d41U,0xfa82U,0xe50fU,0xda15U,0xa421U,0x5849U,0xb092U,0x712fU,
		0xe25eU,0xd4b7U,0xb965U,0x62c1U,0xc582U,0x9b0fU,0x2615U,0x4c2aU,0x9854U,0x20a3U,
		0x4146U,0x828cU,0x1513U,0x2a26U,0x544cU,0xa898U,0x413bU,0x8276U,0x14e7U,0x29ceU,
		0x539cU,0xa738U,0x5e7bU,0xbcf6U,0x69e7U,0xd3ceU,0xb797U,0x7f25U,0xfe4aU,0xec9fU,
		0xc935U,0x8261U,0x14c9U,0x2992U,0x5324U,0xa648U,0x5c9bU,0xb936U,0x6267U,0xc4ceU,
		0x9997U,0x2325U,0x464aU,0x8c94U,0x0923U,0x1246U,0x248cU,0x4918U,0x9230U,0x346bU,
		0x68d6U,0xd1acU,0xb353U,0x76adU,0xed5aU,0xcabfU,0x8575U,0x1ae1U,0x35c2U,0x6b84U,
		0xd708U,0xbe1bU,0x6c3dU,0xd87aU,0xa0ffU,0x51f5U,0xa3eaU,0x57dfU,0xafbeU,0x4f77U,
		0x9eeeU,0x2dd7U,0x5baeU,0xb75cU,0x7eb3U,0xfd66U,0xeac7U,0xc585U,0x9b01U,0x2609U,
		0x4c12U,0x9824U,0x2043U,0x4086U,0x810cU,0x1213U,0x2426U,0x484cU,0x9098U,0x313bU,
		0x6276U,0xc4ecU,0x99d3U,0x23adU,0x475aU,0x8eb4U,0x0d63U,0x1ac6U,0x358cU,0x6b18U,
		0xd630U,0xbc6bU,0x68ddU,0xd1baU,0xb37fU,0x76f5U,0xedeaU,0xcbdfU,0x87b5U,0x1f61U,
		0x3ec2U,0x7d84U,0xfb08U,0xe61bU,0xdc3dU,0xa871U,0x40e9U,0x81d2U,0x13afU,0x275eU,
		0x4ebcU,0x9d78U,0x2afbU,0x55f6U,0xabecU,0x47d3U,0x8fa6U,0x0f47U,0x1e8eU,0x3d1cU,
		0x7a38U,0xf470U,0xf8ebU,0xe1ddU,0xd3b1U,0xb769U,0x7ed9U,0xfdb2U,0xeb6fU,0xc6d5U,
		0x9da1U,0x2b49U,0x5692U,0xad24U,0x4a43U,0x9486U,0x3907U,0x720eU,0xe41cU,0xd833U,
		0xa06dU,0x50d1U,0xa1a2U,0x534fU,0xa69eU,0x5d37U,0xba6eU,0x64d7U,0xc9aeU,0x8357U,
		0x16a5U,0x2d4aU,0x5a94U,0xb528U,0x7a5bU,0xf4b6U,0xf967U,0xe2c5U,0xd581U,0xbb09U,
		0x6619U,0xcc32U,0x886fU,0x00d5U,0x01aaU,0x0354U,0x06a8U,0x0d50U,0x1aa0U,0x3540U,
		0x6a80U,0xd500U,0xba0bU,0x641dU,0xc83aU,0x807fU,0x10f5U,0x21eaU,0x43d4U,0x87a8U,
		0x1f5bU,0x3eb6U,0x7d6cU,0xfad8U,0xe5bbU,0xdb7dU,0xa6f1U,0x5de9U,0xbbd2U,0x67afU,
		0xcf5eU,0x8eb7U,0x0d65U,0x1acaU,0x3594U,0x6b28U,0xd650U,0xbcabU,0x695dU,0xd2baU,
		0xb57fU,0x7af5U,0xf5eaU,0xfbdfU,0xe7b5U,0xdf61U,0xaec9U,0x4d99U,0x9b32U,0x266fU,
		0x4cdeU,0x99bcU,0x2373U,0x46e6U,0x8dccU,0x0b93U,0x1726U,0x2e4cU,0x5c98U,0xb930U,
		0x626bU,0xc4d6U,0x99a7U,0x2345U,0x468aU,0x8d14U,0x0a23U,0x1446U,0x288cU,0x5118U,
		0xa230U,0x546bU,0xa8d6U,0x41a7U,0x834eU,0x1697U,0x2d2eU,0x5a5cU,0xb4b8U,0x797bU,
		0xf2f6U,0xf5e7U,0xfbc5U,0xe781U,0xdf09U,0xae19U,0x4c39U,0x9872U,0x20efU,0x41deU,
		0x83bcU,0x1773U,0x2ee6U,0x5dccU,0xbb98U,0x673bU,0xce76U,0x8ce7U,0x09c5U,0x138aU,
		0x2714U,0x4e28U,0x9c50U,0x28abU,0x5156U,0xa2acU,0x5553U,0xaaa6U,0x4547U,0x8a8eU,
		0x0517U,0x0a2eU,0x145cU,0x28b8U,0x5170U,0xa2e0U,0x55cbU,0xab96U,0x4727U,0x8e4eU,
		0x0c97U,0x192eU,0x325cU,0x64b8U,0xc970U,0x82ebU,0x15ddU,0x2bbaU,0x5774U,0xaee8U,
		0x4ddbU,0x9bb6U,0x2767U,0x4eceU,0x9d9cU,0x2b33U,0x5666U,0xacccU,0x4993U,0x9326U,
		0x3647U,0x6c8eU,0xd91cU,0xa233U,0x546dU,0xa8daU,0x41bfU,0x837eU,0x16f7U,0x2deeU,
		0x5bdcU,0xb7b8U,0x7f7bU,0xfef6U,0xede7U,0xcbc5U,0x8781U,0x1f09U,0x3e12U,0x7c24U,
		0xf848U,0xe09bU,0xd13dU,0xb271U,0x74e9U,0xe9d2U,0xc3afU,0x9755U,0x3ea1U,0x7d42U,
		0xfa84U,0xe503U,0xda0dU,0xa411U,0x5829U,0xb052U,0x70afU,0xe15eU,0xd2b7U,0xb565U,
		0x7ac1U,0xf582U,0xfb0fU,0xe615U,0xdc21U,0xa849U,0x4099U,0x8132U,0x126fU,0x24deU,
		0x49bcU,0x9378U,0x36fbU,0x6df6U,0xdbecU,0xa7d3U,0x5fadU,0xbf5aU,0x6ebfU,0xdd7eU,
		0xaaf7U,0x45e5U,0x8bcaU,0x079fU,0x0f3eU,0x1e7cU,0x3cf8U,0x79f0U,0xf3e0U,0xf7cbU,
		0xff9dU,0xef31U,0xce69U,0x8cd9U,0x09b9U,0x1372U,0x26e4U,0x4dc8U,0x9b90U,0x272bU,
		0x4e56U,0x9cacU,0x2953U,0x52a6U,0xa54cU,0x5a93U,0xb526U,0x7a47U,0xf48eU,0xf917U,
		0xe225U,0xd441U,0xb889U,0x6119U,0xc232U,0x946fU,0x38d5U,0x71aaU,0xe354U,0xd6a3U,
		0xbd4dU,0x6a91U,0xd522U,0xba4fU,0x6495U,0xc92aU,0x825fU,0x14b5U,0x296aU,0x52d4U,
		0xa5a8U,0x5b5bU,0xb6b6U,0x7d67U,0xfaceU,0xe597U,0xdb25U,0xa641U,0x5c89U,0xb912U,
		0x622fU,0xc45eU,0x98b7U,0x2165U,0x42caU,0x8594U,0x1b23U,0x3646U,0x6c8cU,0xd918U,
		0xa23bU,0x547dU,0xa8faU,0x41ffU,0x83feU,0x17f7U,0x2feeU,0x5fdcU,0xbfb8U,0x6f7bU,
		0xdef6U,0xade7U,0x4bc5U,0x978aU,0x3f1fU,0x7e3eU,0xfc7cU,0xe8f3U,0xc1edU,0x93d1U,
		0x37a9U,0x6f52U,0xdea4U,0xad43U,0x4a8dU,0x951aU,0x3a3fU,0x747eU,0xe8fcU,0xc1f3U,
		0x93edU,0x37d1U,0x6fa2U,0xdf44U,0xae83U,0x4d0dU,0x9a1aU,0x243fU,0x487eU,0x90fcU,
		0x31f3U,0x63e6U,0xc7ccU,0x9f93U,0x2f2dU,0x5e5aU,0xbcb4U,0x6963U,0xd2c6U,0xb587U,
		0x7b05U,0xf60aU,0xfc1fU,0xe835U,0xc061U,0x90c9U,0x3199U,0x6332U,0xc664U,0x9cc3U,
		0x298dU,0x531aU,0xa634U,0x5c63U,0xb8c6U,0x6187U,0xc30eU,0x9617U,0x3c25U,0x784aU,
		0xf094U,0xf123U,0xf24dU,0xf491U,0xf929U,0xe259U,0xd4b9U,0xb979U,0x62f9U,0xc5f2U,
		0x9befU,0x27d5U,0x4faaU,0x9f54U,0x2ea3U,0x5d46U,0xba8cU,0x6513U,0xca26U,0x8447U,
		0x1885U,0x310aU,0x6214U,0xc428U,0x985bU,0x20bdU,0x417aU,0x82f4U,0x15e3U,0x2bc6U,
		0x578cU,0xaf18U,0x4e3bU,0x9c76U,0x28e7U,0x51ceU,0xa39cU,0x5733U,0xae66U,0x4cc7U,
		0x998eU,0x2317U,0x462eU,0x8c5cU,0x08b3U,0x1166U,0x22ccU,0x4598U,0x8b30U,0x066bU,
		0x0cd6U,0x19acU,0x3358U,0x66b0U,0xcd60U,0x8acbU,0x059dU,0x0b3aU,0x1674U,0x2ce8U,
		0x59d0U,0xb3a0U,0x774bU,0xee96U,0xcd27U,0x8a45U,0x0481U,0x0902U,0x1204U,0x2408U,
		0x4810U,0x9020U,0x304bU,0x6096U,0xc12cU,0x9253U,0x34adU,0x695aU,0xd2b4U,0xb563U,
		0x7acdU,0xf59aU,0xfb3fU,0xe675U,0xdce1U,0xa9c9U,0x4399U,0x8732U,0x1e6fU,0x3cdeU,
		0x79bcU,0xf378U,0xf6fbU,0xfdfdU,0xebf1U,0xc7e9U,0x9fd9U,0x2fb9U,0x5f72U,0xbee4U,
		0x6dc3U,0xdb86U,0xa707U,0x5e05U,0xbc0aU,0x681fU,0xd03eU,0xb077U,0x70e5U,0xe1caU,
		0xd39fU,0xb735U,0x7e61U,0xfcc2U,0xe98fU,0xc315U,0x9621U,0x3c49U,0x7892U,0xf124U,
		0xf243U,0xf48dU,0xf911U,0xe229U,0xd459U,0xb8b9U,0x6179U,0xc2f2U,0x95efU,0x3bd5U,
		0x77aaU,0xef54U,0xcea3U,0x8d4dU,0x0a91U,0x1522U,0x2a44U,0x5488U,0xa910U,0x422bU,
		0x8456U,0x18a7U,0x314eU,0x629cU,0xc538U,0x9a7bU,0x24fdU,0x49faU,0x93f4U,0x37e3U,
		0x6fc6U,0xdf8cU,0xaf13U,0x4e2dU,0x9c5aU,0x28bfU,0x517eU,0xa2fcU,0x55f3U,0xabe6U,
		0x47c7U,0x8f8eU,0x0f17U,0x1e2eU,0x3c5cU,0x78b8U,0xf170U,0xf2ebU,0xf5ddU,0xfbb1U,
		0xe769U,0xded9U,0xadb9U,0x4b79U,0x96f2U,0x3defU,0x7bdeU,0xf7bcU,0xff73U,0xeeedU,
		0xcdd1U,0x8ba9U,0x0759U,0x0eb2U,0x1d64U,0x3ac8U,0x7590U,0xeb20U,0xc64bU,0x9c9dU,
		0x2931U,0x5262U,0xa4c4U,0x5983U,0xb306U,0x7607U,0xec0eU,0xc817U,0x8025U,0x1041U,
		0x2082U,0x4104U,0x8208U,0x141bU,0x2836U,0x506cU,0xa0d8U,0x51bbU,0xa376U,0x56e7U,
		0xadceU,0x4b97U,0x972eU,0x3e57U,0x7caeU,0xf95cU,0xe2b3U,0xd56dU,0xbad1U,0x65a9U,
		0xcb52U,0x86afU,0x1d55U,0x3aaaU,0x7554U,0xeaa8U,0xc55bU,0x9abdU,0x2571U,0x4ae2U,
		0x95c4U,0x3b83U,0x7706U,0xee0cU,0xcc13U,0x882dU,0x0051U,0x00a2U,0x0144U,0x0288U,
		0x0510U,0x0a20U,0x1440U,0x2880U,0x5100U,0xa200U,0x540bU,0xa816U,0x4027U,0x804eU,
		0x1097U,0x212eU,0x425cU,0x84b8U,0x197bU,0x32f6U,0x65ecU,0xcbd8U,0x87bbU,0x1f7dU,
		0x3efaU,0x7df4U,0xfbe8U,0xe7dbU,0xdfbdU,0xaf71U,0x4ee9U,0x9dd2U,0x2bafU,0x575eU,
		0xaebcU,0x4d73U,0x9ae6U,0x25c7U,0x4b8eU,0x971cU,0x3e33U,0x7c66U,0xf8ccU,0xe193U,
		0xd32dU,0xb651U,0x7ca9U,0xf952U,0xe2afU,0xd555U,0xbaa1U,0x6549U,0xca92U,0x852fU,
		0x1a55U,0x34aaU,0x6954U,0xd2a8U,0xb55bU,0x7abdU,0xf57aU,0xfaffU,0xe5f5U,0xdbe1U,
		0xa7c9U,0x5f99U,0xbf32U,0x6e6fU,0xdcdeU,0xa9b7U,0x4365U,0x86caU,0x1d9fU,0x3b3eU,
		0x767cU,0xecf8U,0xc9fbU,0x83fdU,0x17f1U,0x2fe2U,0x5fc4U,0xbf88U,0x6f1bU,0xde36U,
		0xac67U,0x48c5U,0x918aU,0x331fU,0x663eU,0xcc7cU,0x88f3U,0x01edU,0x03daU,0x07b4U,
		0x0f68U,0x1ed0U,0x3da0U,0x7b40U,0xf680U,0xfd0bU,0xea1dU,0xc431U,0x9869U,0x20d9U,
		0x41b2U,0x8364U,0x16c3U,0x2d86U,0x5b0cU,0xb618U,0x7c3bU,0xf876U,0xe0e7U,0xd1c5U,
		0xb381U,0x7709U,0xee12U,0xcc2fU,0x8855U,0x00a1U,0x0142U,0x0284U,0x0508U,0x0a10U,
		0x1420U,0x2840U,0x5080U,0xa100U,0x520bU,0xa416U,0x5827U,0xb04eU,0x7097U,0xe12eU,
		0xd257U,0xb4a5U,0x7941U,0xf282U,0xf50fU,0xfa15U,0xe421U,0xd849U,0xa099U,0x5139U,
		0xa272U,0x54efU,0xa9deU,0x43b7U,0x876eU,0x1ed7U,0x3daeU,0x7b5cU,0xf6b8U,0xfd7bU,
		0xeafdU,0xc5f1U,0x9be9U,0x27d9U,0x4fb2U,0x9f64U,0x2ec3U,0x5d86U,0xbb0cU,0x6613U,
		0xcc26U,0x8847U,0x0085U,0x010aU,0x0214U,0x0428U,0x0850U,0x10a0U,0x2140U,0x4280U,
		0x8500U,0x1a0bU,0x3416U,0x682cU,0xd058U,0xb0bbU,0x717dU,0xe2faU,0xd5ffU,0xbbf5U,
		0x67e1U,0xcfc2U,0x8f8fU,0x0f15U,0x1e2aU,0x3c54U,0x78a8U,0xf150U,0xf2abU,0xf55dU,
		0xfab1U,0xe569U,0xdad9U,0xa5b9U,0x5b79U,0xb6f2U,0x7defU,0xfbdeU,0xe7b7U,0xdf65U,
		0xaec1U,0x4d89U,0x9b12U,0x262fU,0x4c5eU,0x98bcU,0x2173U,0x42e6U,0x85ccU,0x1b93U,
		0x3726U,0x6e4cU,0xdc98U,0xa93bU,0x427dU,0x84faU,0x19ffU,0x33feU,0x67fcU,0xcff8U,
		0x8ffbU,0x0ffdU,0x1ffaU,0x3ff4U,0x7fe8U,0xffd0U,0xefabU,0xcf5dU,0x8eb1U,0x0d69U,
		0x1ad2U,0x35a4U,0x6b48U,0xd690U,0xbd2bU,0x6a5dU,0xd4baU,0xb97fU,0x62f5U,0xc5eaU,
		0x9bdfU,0x27b5U,0x4f6aU,0x9ed4U,0x2da3U,0x5b46U,0xb68cU,0x7d13U,0xfa26U,0xe447U,
		0xd885U,0xa101U,0x5209U,0xa412U,0x582fU,0xb05eU,0x70b7U,0xe16eU,0xd2d7U,0xb5a5U,
		0x7b41U,0xf682U,0xfd0fU,0xea15U,0xc421U,0x9849U,0x2099U,0x4132U,0x8264U,0x14c3U,
		0x2986U,0x530cU,0xa618U,0x5c3bU,0xb876U,0x60e7U,0xc1ceU,0x9397U,0x3725U,0x6e4aU,
		0xdc94U,0xa923U,0x424dU,0x849aU,0x193fU,0x327eU,0x64fcU,0xc9f8U,0x83fbU,0x17fdU,
		0x2ffaU,0x5ff4U,0xbfe8U,0x6fdbU,0xdfb6U,0xaf67U,0x4ec5U,0x9d8aU,0x2b1fU,0x563eU,
		0xac7cU,0x48f3U,0x91e6U,0x33c7U,0x678eU,0xcf1cU,0x8e33U,0x0c6dU,0x18daU,0x31b4U,
		0x6368U,0xc6d0U,0x9dabU,0x2b5dU,0x56baU,0xad74U,0x4ae3U,0x95c6U,0x3b87U,0x770eU,
		0xee1cU,0xcc33U,0x886dU,0x00d1U,0x01a2U,0x0344U,0x0688U,0x0d10U,0x1a20U,0x3440U,
		0x6880U,0xd100U,0xb20bU,0x741dU,0xe83aU,0xc07fU,0x90f5U,0x31e1U,0x63c2U,0xc784U,
		0x9f03U,0x2e0dU,0x5c1aU,0xb834U,0x6063U,0xc0c6U,0x9187U,0x3305U,0x660aU,0xcc14U,
		0x8823U,0x004dU,0x009aU,0x0134U,0x0268U,0x04d0U,0x09a0U,0x1340U,0x2680U,0x4d00U,
		0x9a00U,0x240bU,0x4816U,0x902cU,0x3053U,0x60a6U,0xc14cU,0x9293U,0x352dU,0x6a5aU,
		0xd4b4U,0xb963U,0x62cdU,0xc59aU,0x9b3fU,0x2675U,0x4ceaU,0x99d4U,0x23a3U,0x4746U,
		0x8e8cU,0x0d13U,0x1a26U,0x344cU,0x6898U,0xd130U,0xb26bU,0x74ddU,0xe9baU,0xc37fU,
		0x96f5U,0x3de1U,0x7bc2U,0xf784U,0xff03U,0xee0dU,0xcc11U,0x8829U,0x0059U,0x00b2U,
		0x0164U,0x02c8U,0x0590U,0x0b20U,0x1640U,0x2c80U,0x5900U,0xb200U,0x740bU,0xe816U,
		0xc027U,0x9045U,0x3081U,0x6102U,0xc204U,0x9403U,0x380dU,0x701aU,0xe034U,0xd063U,
		0xb0cdU,0x7191U,0xe322U,0xd64fU,0xbc95U,0x6921U,0xd242U,0xb48fU,0x7915U,0xf22aU,
		0xf45fU,0xf8b5U,0xe161U,0xd2c9U,0xb599U,0x7b39U,0xf672U,0xfcefU,0xe9d5U,0xc3a1U,
		0x9749U,0x3e99U,0x7d32U,0xfa64U,0xe4c3U,0xd98dU,0xa311U,0x5629U,0xac52U,0x48afU,
		0x915eU,0x32b7U,0x656eU,0xcadcU,0x85b3U,0x1b6dU,0x36daU,0x6db4U,0xdb68U,0xa6dbU,
		0x5dbdU,0xbb7aU,0x66ffU,0xcdfeU,0x8bf7U,0x07e5U,0x0fcaU,0x1f94U,0x3f28U,0x7e50U,
		0xfca0U,0xe94bU,0xc29dU,0x9531U,0x3a69U,0x74d2U,0xe9a4U,0xc343U,0x968dU,0x3d11U,
		0x7a22U,0xf444U,0xf883U,0xe10dU,0xd211U,0xb429U,0x7859U,0xf0b2U,0xf16fU,0xf2d5U,
		0xf5a1U,0xfb49U,0xe699U,0xdd39U,0xaa79U,0x44f9U,0x89f2U,0x03efU,0x07deU,0x0fbcU,
		0x1f78U,0x3ef0U,0x7de0U,0xfbc0U,0xe78bU,0xdf1dU,0xae31U,0x4c69U,0x98d2U,0x21afU,
		0x435eU,0x86bcU,0x1d73U,0x3ae6U,0x75ccU,0xeb98U,0xc73bU,0x9e7dU,0x2cf1U,0x59e2U,
		0xb3c4U,0x7783U,0xef06U,0xce07U,0x8c05U,0x0801U,0x1002U,0x2004U,0x4008U,0x8010U,
		0x102bU,0x2056U,0x40acU,0x8158U,0x12bbU,0x2576U,0x4aecU,0x95d8U,0x3bbbU,0x7776U,
		0xeeecU,0xcdd3U,0x8badU,0x0751U,0x0ea2U,0x1d44U,0x3a88U,0x7510U,0xea20U,0xc44bU,
		0x989dU,0x2131U,0x4262U,0x84c4U,0x1983U,0x3306U,0x660cU,0xcc18U,0x883bU,0x007dU,
		0x00faU,0x01f4U,0x03e8U,0x07d0U,0x0fa0U,0x1f40U,0x3e80U,0x7d00U,0xfa00U,0xe40bU,
		0xd81dU,0xa031U,0x5069U,0xa0d2U,0x51afU,0xa35eU,0x56b7U,0xad6eU,0x4ad7U,0x95aeU,
		0x3b57U,0x76aeU,0xed5cU,0xcab3U,0x856dU,0x1ad1U,0x35a2U,0x6b44U,0xd688U,0xbd1bU,
		0x6a3dU,0xd47aU,0xb8ffU,0x61f5U,0xc3eaU,0x97dfU,0x3fb5U,0x7f6aU,0xfed4U,0xeda3U,
		0xcb4dU,0x8691U,0x1d29U,0x3a52U,0x74a4U,0xe948U,0xc29bU,0x953dU,0x3a71U,0x74e2U,
		0xe9c4U,0xc383U,0x970dU,0x3e11U,0x7c22U,0xf844U,0xe083U,0xd10dU,0xb211U,0x7429U,
		0xe852U,0xc0afU,0x9155U,0x32a1U,0x6542U,0xca84U,0x8503U,0x1a0dU,0x341aU,0x6834U,
		0xd068U,0xb0dbU,0x71bdU,0xe37aU,0xd6ffU,0xbdf5U,0x6be1U,0xd7c2U,0xbf8fU,0x6f15U,
		0xde2aU,0xac5fU,0x48b5U,0x916aU,0x32dfU,0x65beU,0xcb7cU,0x86f3U,0x1dedU,0x3bdaU,
		0x77b4U,0xef68U,0xcedbU,0x8dbdU,0x0b71U,0x16e2U,0x2dc4U,0x5b88U,0xb710U,0x7e2bU,
		0xfc56U,0xe8a7U,0xc145U,0x9281U,0x3509U,0x6a12U,0xd424U,0xb843U,0x608dU,0xc11aU,
		0x923fU,0x3475U,0x68eaU,0xd1d4U,0xb3a3U,0x774dU,0xee9aU,0xcd3fU,0x8a75U,0x04e1U,
		0x09c2U,0x1384U,0x2708U,0x4e10U,0x9c20U,0x284bU,0x5096U,0xa12cU,0x5253U,0xa4a6U,
		0x5947U,0xb28eU,0x7517U,0xea2eU,0xc457U,0x98a5U,0x2141U,0x4282U,0x8504U,0x1a03U,
		0x3406U,0x680cU,0xd018U,0xb03bU,0x707dU,0xe0faU,0xd1ffU,0xb3f5U,0x77e1U,0xefc2U,
		0xcf8fU,0x8f15U,0x0e21U,0x1c42U,0x3884U,0x7108U,0xe210U,0xd42bU,0xb85dU,0x60b1U,
		0xc162U,0x92cfU,0x3595U,0x6b2aU,0xd654U,0xbca3U,0x694dU,0xd29aU,0xb53fU,0x7a75U,
		0xf4eaU,0xf9dfU,0xe3b5U,0xd761U,0xbec9U,0x6d99U,0xdb32U,0xa66fU,0x5cd5U,0xb9aaU,
		0x635fU,0xc6beU,0x9d77U,0x2ae5U,0x55caU,0xab94U,0x4723U,0x8e46U,0x0c87U,0x190eU,
		0x321cU,0x6438U,0xc870U,0x80ebU,0x11ddU,0x23baU,0x4774U,0x8ee8U,0x0ddbU,0x1bb6U,
		0x376cU,0x6ed8U,0xddb0U,0xab6bU,0x46ddU,0x8dbaU,0x0b7fU,0x16feU,0x2dfcU,0x5bf8U,
		0xb7f0U,0x7febU,0xffd6U,0xefa7U,0xcf45U,0x8e81U,0x0d09U,0x1a12U,0x3424U,0x6848U,
		0xd090U,0xb12bU,0x725dU,0xe4baU,0xd97fU,0xa2f5U,0x55e1U,0xabc2U,0x478fU,0x8f1eU,
		0x0e37U,0x1c6eU,0x38dcU,0x71b8U,0xe370U,0xd6ebU,0xbdddU,0x6bb1U,0xd762U,0xbecfU,
		0x6d95U,0xdb2aU,0xa65fU,0x5cb5U,0xb96aU,0x62dfU,0xc5beU,0x9b77U,0x26e5U,0x4dcaU,
		0x9b94U,0x2723U,0x4e46U,0x9c8cU,0x2913U,0x5226U,0xa44cU,0x5893U,0xb126U,0x7247U,
		0xe48eU,0xd917U,0xa225U,0x5441U,0xa882U,0x410fU,0x821eU,0x1437U,0x286eU,0x50dcU,
		0xa1b8U,0x537bU,0xa6f6U,0x5de7U,0xbbceU,0x6797U,0xcf2eU,0x8e57U,0x0ca5U,0x194aU,
		0x3294U,0x6528U,0xca50U,0x84abU,0x195dU,0x32baU,0x6574U,0xcae8U,0x85dbU,0x1bbdU,
		0x377aU,0x6ef4U,0xdde8U,0xabdbU,0x47bdU,0x8f7aU,0x0effU,0x1dfeU,0x3bfcU,0x77f8U,
		0xeff0U,0xcfebU,0x8fddU,0x0fb1U,0x1f62U,0x3ec4U,0x7d88U,0xfb10U,0xe62bU,0xdc5dU,
		0xa8b1U,0x4169U,0x82d2U,0x15afU,0x2b5eU,0x56bcU,0xad78U,0x4afbU,0x95f6U,0x3be7U,
		0x77ceU,0xef9cU,0xcf33U,0x8e6dU,0x0cd1U,0x19a2U,0x3344U,0x6688U,0xcd10U,0x8a2bU,
		0x045dU,0x08baU,0x1174U,0x22e8U,0x45d0U,0x8ba0U,0x074bU,0x0e96U,0x1d2cU,0x3a58U,
		0x74b0U,0xe960U,0xc2cbU,0x959dU,0x3b31U,0x7662U,0xecc4U,0xc983U,0x830dU,0x1611U,
		0x2c22U,0x5844U,0xb088U,0x711bU,0xe236U,0xd467U,0xb8c5U,0x6181U,0xc302U,0x960fU,
		0x3c15U,0x782aU,0xf054U,0xf0a3U,0xf14dU,0xf291U,0xf529U,0xfa59U,0xe4b9U,0xd979U,
		0xa2f9U,0x55f9U,0xabf2U,0x47efU,0x8fdeU,0x0fb7U,0x1f6eU,0x3edcU,0x7db8U,0xfb70U,
		0xe6ebU,0xddddU,0xabb1U,0x4769U,0x8ed2U,0x0dafU,0x1b5eU,0x36bcU,0x6d78U,0xdaf0U,
		0xa5ebU,0x5bddU,0xb7baU,0x7f7fU,0xfefeU,0xedf7U,0xcbe5U,0x87c1U,0x1f89U,0x3f12U,
		0x7e24U,0xfc48U,0xe89bU,0xc13dU,0x9271U,0x34e9U,0x69d2U,0xd3a4U,0xb743U,0x7e8dU,
		0xfd1aU,0xea3fU,0xc475U,0x98e1U,0x21c9U,0x4392U,0x8724U,0x1e43U,0x3c86U,0x790cU,
		0xf218U,0xf43bU,0xf87dU,0xe0f1U,0xd1e9U,0xb3d9U,0x77b9U,0xef72U,0xceefU,0x8dd5U,
		0x0ba1U,0x1742U,0x2e84U,0x5d08U,0xba10U,0x642bU,0xc856U,0x80a7U,0x1145U,0x228aU,
		0x4514U,0x8a28U,0x045bU,0x08b6U,0x116cU,0x22d8U,0x45b0U,0x8b60U,0x06cbU,0x0d96U,
		0x1b2cU,0x3658U,0x6cb0U,0xd960U,0xa2cbU,0x559dU,0xab3aU,0x467fU,0x8cfeU,0x09f7U,
		0x13eeU,0x27dcU,0x4fb8U,0x9f70U,0x2eebU,0x5dd6U,0xbbacU,0x6753U,0xcea6U,0x8d47U,
		0x0a85U,0x150aU,0x2a14U,0x5428U,0xa850U,0x40abU,0x8156U,0x12a7U,0x254eU,0x4a9cU,
		0x9538U,0x3a7bU,0x74f6U,0xe9ecU,0xc3d3U,0x97adU,0x3f51U,0x7ea2U,0xfd44U,0xea83U,
		0xc50dU,0x9a11U,0x2429U,0x4852U,0x90a4U,0x3143U,0x6286U,0xc50cU,0x9a13U,0x242dU,
		0x485aU,0x90b4U,0x3163U,0x62c6U,0xc58cU,0x9b13U,0x262dU,0x4c5aU,0x98b4U,0x2163U,
		0x42c6U,0x858cU,0x1b13U,0x3626U,0x6c4cU,0xd898U,0xa13bU,0x527dU,0xa4faU,0x59ffU,
		0xb3feU,0x77f7U,0xefeeU,0xcfd7U,0x8fa5U,0x0f41U,0x1e82U,0x3d04U,0x7a08U,0xf410U,
		0xf82bU,0xe05dU,0xd0b1U,0xb169U,0x72d9U,0xe5b2U,0xdb6fU,0xa6d5U,0x5da1U,0xbb42U,
		0x668fU,0xcd1eU,0x8a37U,0x0465U,0x08caU,0x1194U,0x2328U,0x4650U,0x8ca0U,0x094bU,
		0x1296U,0x252cU,0x4a58U,0x94b0U,0x396bU,0x72d6U,0xe5acU,0xdb53U,0xa6adU,0x5d51U,
		0xbaa2U,0x654fU,0xca9eU,0x8537U,0x1a65U,0x34caU,0x6994U,0xd328U,0xb65bU,0x7cbdU,
		0xf97aU,0xe2ffU,0xd5f5U,0xbbe1U,0x67c9U,0xcf92U,0x8f2fU,0x0e55U,0x1caaU,0x3954U,
		0x72a8U,0xe550U,0xdaabU,0xa55dU,0x5ab1U,0xb562U,0x7acfU,0xf59eU,0xfb37U,0xe665U,
		0xdcc1U,0xa989U,0x4319U,0x8632U,0x1c6fU,0x38deU,0x71bcU,0xe378U,0xd6fbU,0xbdfdU,
		0x6bf1U,0xd7e2U,0xbfcfU,0x6f95U,0xdf2aU,0xae5fU,0x4cb5U,0x996aU,0x22dfU,0x45beU,
		0x8b7cU,0x06f3U,0x0de6U,0x1bccU,0x3798U,0x6f30U,0xde60U,0xaccbU,0x499dU,0x933aU,
		0x367fU,0x6cfeU,0xd9fcU,0xa3f3U,0x57edU,0xafdaU,0x4fbfU,0x9f7eU,0x2ef7U,0x5deeU,
		0xbbdcU,0x67b3U,0xcf66U,0x8ec7U,0x0d85U,0x1b0aU,0x3614U,0x6c28U,0xd850U,0xa0abU,
		0x515dU,0xa2baU,0x557fU,0xaafeU,0x45f7U,0x8beeU,0x07d7U,0x0faeU,0x1f5cU,0x3eb8U,
		0x7d70U,0xfae0U,0xe5cbU,0xdb9dU,0xa731U,0x5e69U,0xbcd2U,0x69afU,0xd35eU,0xb6b7U,
		0x7d65U,0xfacaU,0xe59fU,0xdb35U,0xa661U,0x5cc9U,0xb992U,0x632fU,0xc65eU,0x9cb7U,
		0x2965U,0x52caU,0xa594U,0x5b23U,0xb646U,0x7c87U,0xf90eU,0xe217U,0xd425U,0xb841U,
		0x6089U,0xc112U,0x922fU,0x3455U,0x68aaU,0xd154U,0xb2a3U,0x754dU,0xea9aU,0xc53fU,
		0x9a75U,0x24e1U,0x49c2U,0x9384U,0x3703U,0x6e06U,0xdc0cU,0xa813U,0x402dU,0x805aU,
		0x10bfU,0x217eU,0x42fcU,0x85f8U,0x1bfbU,0x37f6U,0x6fecU,0xdfd8U,0xafbbU,0x4f7dU,
		0x9efaU,0x2dffU,0x5bfeU,0xb7fcU,0x7ff3U,0xffe6U,0xefc7U,0xcf85U,0x8f01U,0x0e09U,
		0x1c12U,0x3824U,0x7048U,0xe090U,0xd12bU,0xb25dU,0x74b1U,0xe962U,0xc2cfU,0x9595U,
		0x3b21U,0x7642U,0xec84U,0xc903U,0x820dU,0x1411U,0x2822U,0x5044U,0xa088U,0x511bU,
		0xa236U,0x5467U,0xa8ceU,0x4197U,0x832eU,0x1657U,0x2caeU,0x595cU,0xb2b8U,0x757bU,
		0xeaf6U,0xc5e7U,0x9bc5U,0x2781U,0x4f02U,0x9e04U,0x2c03U,0x5806U,0xb00cU,0x7013U,
		0xe026U,0xd047U,0xb085U,0x7101U,0xe202U,0xd40fU,0xb815U,0x6021U,0xc042U,0x908fU,
		0x3115U,0x622aU,0xc454U,0x98a3U,0x214dU,0x429aU,0x8534U,0x1a63U,0x34c6U,0x698cU,
		0xd318U,0xb63bU,0x7c7dU,0xf8faU,0xe1ffU,0xd3f5U,0xb7e1U,0x7fc9U,0xff92U,0xef2fU,
		0xce55U,0x8ca1U,0x0949U,0x1292U,0x2524U,0x4a48U,0x9490U,0x392bU,0x7256U,0xe4acU,
		0xd953U,0xa2adU,0x5551U,0xaaa2U,0x454fU,0x8a9eU,0x0537U,0x0a6eU,0x14dcU,0x29b8U,
		0x5370U,0xa6e0U,0x5dcbU,0xbb96U,0x6727U,0xce4eU,0x8c97U,0x0925U,0x124aU,0x2494U,
		0x4928U,0x9250U,0x34abU,0x6956U,0xd2acU,0xb553U,0x7aadU,0xf55aU,0xfabfU,0xe575U,
		0xdae1U,0xa5c9U,0x5b99U,0xb732U,0x7e6fU,0xfcdeU,0xe9b7U,0xc365U,0x96c1U,0x3d89U,
		0x7b12U,0xf624U,0xfc43U,0xe88dU,0xc111U,0x9229U,0x3459U,0x68b2U,0xd164U,0xb2c3U,
		0x758dU,0xeb1aU,0xc63fU,0x9c75U,0x28e1U,0x51c2U,0xa384U,0x5703U,0xae06U,0x4c07U,
		0x980eU,0x2017U,0x402eU,0x805cU,0x10b3U,0x2166U,0x42ccU,0x8598U,0x1b3bU,0x3676U,
		0x6cecU,0xd9d8U,0xa3bbU,0x577dU,0xaefaU,0x4dffU,0x9bfeU,0x27f7U,0x4feeU,0x9fdcU,
		0x2fb3U,0x5f66U,0xbeccU,0x6d93U,0xdb26U,0xa647U,0x5c85U,0xb90aU,0x621fU,0xc43eU,
		0x9877U,0x20e5U,0x41caU,0x8394U,0x1723U,0x2e46U,0x5c8cU,0xb918U,0x623bU,0xc476U,
		0x98e7U,0x21c5U,0x438aU,0x8714U,0x1e23U,0x3c46U,0x788cU,0xf118U,0xf23bU,0xf47dU,
		0xf8f1U,0xe1e9U,0xd3d9U,0xb7b9U,0x7f79U,0xfef2U,0xedefU,0xcbd5U,0x87a1U,0x1f49U,
		0x3e92U,0x7d24U,0xfa48U,0xe49bU,0xd93dU,0xa271U,0x54e9U,0xa9d2U,0x43afU,0x875eU,
		0x1eb7U,0x3d6eU,0x7adcU,0xf5b8U,0xfb7bU,0xe6fdU,0xddf1U,0xabe9U,0x47d9U,0x8fb2U,
		0x0f6fU,0x1edeU,0x3dbcU,0x7b78U,0xf6f0U,0xfdebU,0xebddU,0xc7b1U,0x9f69U,0x2ed9U,
		0x5db2U,0xbb64U,0x66c3U,0xcd86U,0x8b07U,0x0605U,0x0c0aU,0x1814U,0x3028U,0x6050U,
		0xc0a0U,0x914bU,0x329dU,0x653aU,0xca74U,0x84e3U,0x19cdU,0x339aU,0x6734U,0xce68U,
		0x8cdbU,0x09bdU,0x137aU,0x26f4U,0x4de8U,0x9bd0U,0x27abU,0x4f56U,0x9eacU,0x2d53U,
		0x5aa6U,0xb54cU,0x7a93U,0xf526U,0xfa47U,0xe485U,0xd901U,0xa209U,0x5419U,0xa832U,
		0x406fU,0x80deU,0x11b7U,0x236eU,0x46dcU,0x8db8U,0x0b7bU,0x16f6U,0x2decU,0x5bd8U,
		0xb7b0U,0x7f6bU,0xfed6U,0xeda7U,0xcb45U,0x8681U,0x1d09U,0x3a12U,0x7424U,0xe848U,
		0xc09bU,0x913dU,0x3271U,0x64e2U,0xc9c4U,0x8383U,0x170dU,0x2e1aU,0x5c34U,0xb868U,
		0x60dbU,0xc1b6U,0x9367U,0x36c5U,0x6d8aU,0xdb14U,0xa623U,0x5c4dU,0xb89aU,0x613fU,
		0xc27eU,0x94f7U,0x39e5U,0x73caU,0xe794U,0xdf23U,0xae4dU,0x4c91U,0x9922U,0x224fU,
		0x449eU,0x893cU,0x0273U,0x04e6U,0x09ccU,0x1398U,0x2730U,0x4e60U,0x9cc0U,0x298bU,
		0x5316U,0xa62cU,0x5c53U,0xb8a6U,0x6147U,0xc28eU,0x9517U,0x3a25U,0x744aU,0xe894U,
		0xc123U,0x924dU,0x3491U,0x6922U,0xd244U,0xb483U,0x790dU,0xf21aU,0xf43fU,0xf875U,
		0xe0e1U,0xd1c9U,0xb399U,0x7739U,0xee72U,0xccefU,0x89d5U,0x03a1U,0x0742U,0x0e84U,
		0x1d08U,0x3a10U,0x7420U,0xe840U,0xc08bU,0x911dU,0x3231U,0x6462U,0xc8c4U,0x8183U,
		0x130dU,0x261aU,0x4c34U,0x9868U,0x20dbU,0x41b6U,0x836cU,0x16d3U,0x2da6U,0x5b4cU,
		0xb698U,0x7d3bU,0xfa76U,0xe4e7U,0xd9c5U,0xa381U,0x5709U,0xae12U,0x4c2fU,0x985eU,
		0x20b7U,0x416eU,0x82dcU,0x15b3U,0x2b66U,0x56ccU,0xad98U,0x4b3bU,0x9676U,0x3ce7U,
		0x79ceU,0xf39cU,0xf733U,0xfe6dU,0xecd1U,0xc9a9U,0x8359U,0x16b9U,0x2d72U,0x5ae4U,
		0xb5c8U,0x7b9bU,0xf736U,0xfe67U,0xecc5U,0xc981U,0x8309U,0x1619U,0x2c32U,0x5864U,
		0xb0c8U,0x719bU,0xe336U,0xd667U,0xbcc5U,0x6981U,0xd302U,0xb60fU,0x7c15U,0xf82aU,
		0xe05fU,0xd0b5U,0xb161U,0x72c9U,0xe592U,0xdb2fU,0xa655U,0x5ca1U,0xb942U,0x628fU,
		0xc51eU,0x9a37U,0x2465U,0x48caU,0x9194U,0x3323U,0x6646U,0xcc8cU,0x8913U,0x022dU,
		0x045aU,0x08b4U,0x1168U,0x22d0U,0x45a0U,0x8b40U,0x068bU,0x0d16U,0x1a2cU,0x3458U,
		0x68b0U,0xd160U,0xb2cbU,0x759dU,0xeb3aU,0xc67fU,0x9cf5U,0x29e1U,0x53c2U,0xa784U,
		0x5f03U,0xbe06U,0x6c07U,0xd80eU,0xa017U,0x5025U,0xa04aU,0x509fU,0xa13eU,0x5277U,
		0xa4eeU,0x59d7U,0xb3aeU,0x7757U,0xeeaeU,0xcd57U,0x8aa5U,0x0541U,0x0a82U,0x1504U,
		0x2a08U,0x5410U,0xa820U,0x404bU,0x8096U,0x1127U,0x224eU,0x449cU,0x8938U,0x027bU,
		0x04f6U,0x09ecU,0x13d8U,0x27b0U,0x4f60U,0x9ec0U,0x2d8bU,0x5b16U,0xb62cU,0x7c53U,
		0xf8a6U,0xe147U,0xd285U,0xb501U,0x7a09U,0xf412U,0xf82fU,0xe055U,0xd0a1U,0xb149U,
		0x7299U,0xe532U,0xda6fU,0xa4d5U,0x59a1U,0xb342U,0x768fU,0xed1eU,0xca37U,0x8465U,
		0x18c1U,0x3182U,0x6304U,0xc608U,0x9c1bU,0x283dU,0x507aU,0xa0f4U,0x51e3U,0xa3c6U,
		0x5787U,0xaf0eU,0x4e17U,0x9c2eU,0x2857U,0x50aeU,0xa15cU,0x52b3U,0xa566U,0x5ac7U,
		0xb58eU,0x7b17U,0xf62eU,0xfc57U,0xe8a5U,0xc141U,0x9289U,0x3519U,0x6a32U,0xd464U,
		0xb8c3U,0x618dU,0xc31aU,0x963fU,0x3c75U,0x78eaU,0xf1d4U,0xf3a3U,0xf74dU,0xfe91U,
		0xed29U,0xca59U,0x84b9U,0x1979U,0x32f2U,0x65e4U,0xcbc8U,0x879bU,0x1f3dU,0x3e7aU,
		0x7cf4U,0xf9e8U,0xe3dbU,0xd7bdU,0xbf71U,0x6ee9U,0xddd2U,0xabafU,0x4755U,0x8eaaU,
		0x0d5fU,0x1abeU,0x357cU,0x6af8U,0xd5f0U,0xbbebU,0x67ddU,0xcfbaU,0x8f7fU,0x0ef5U,
		0x1deaU,0x3bd4U,0x77a8U,0xef50U,0xceabU,0x8d5dU,0x0ab1U,0x1562U,0x2ac4U,0x5588U,
		0xab10U,0x462bU,0x8c56U,0x08a7U,0x114eU,0x229cU,0x4538U,0x8a70U,0x04ebU,0x09d6U,
		0x13acU,0x2758U,0x4eb0U,0x9d60U,0x2acbU,0x5596U,0xab2cU,0x4653U,0x8ca6U,0x0947U,
		0x128eU,0x251cU,0x4a38U,0x9470U,0x38ebU,0x71d6U,0xe3acU,0xd753U,0xbeadU,0x6d51U,
		0xdaa2U,0xa54fU,0x5a95U,0xb52aU,0x7a5fU,0xf4beU,0xf977U,0xe2e5U,0xd5c1U,0xbb89U,
		0x6719U,0xce32U,0x8c6fU,0x08d5U,0x11aaU,0x2354U,0x46a8U,0x8d50U,0x0aabU,0x1556U,
		0x2aacU,0x5558U,0xaab0U,0x456bU,0x8ad6U,0x05a7U,0x0b4eU,0x169cU,0x2d38U,0x5a70U,
		0xb4e0U,0x79cbU,0xf396U,0xf727U,0xfe45U,0xec81U,0xc909U,0x8219U,0x1439U,0x2872U,
		0x50e4U,0xa1c8U,0x539bU,0xa736U,0x5e67U,0xbcceU,0x6997U,0xd32eU,0xb657U,0x7ca5U,
		0xf94aU,0xe29fU,0xd535U,0xba61U,0x64c9U,0xc992U,0x832fU,0x1655U,0x2caaU,0x5954U,
		0xb2a8U,0x755bU,0xeab6U,0xc567U,0x9ac5U,0x2581U,0x4b02U,0x9604U,0x3c03U,0x7806U,
		0xf00cU,0xf013U,0xf02dU,0xf051U,0xf0a9U,0xf159U,0xf2b9U,0xf579U,0xfaf9U,0xe5f9U,
		0xdbf9U,0xa7f9U,0x5ff9U,0xbff2U,0x6fefU,0xdfdeU,0xafb7U,0x4f65U,0x9ecaU,0x2d9fU,
		0x5b3eU,0xb67cU,0x7cf3U,0xf9e6U,0xe3c7U,0xd785U,0xbf01U,0x6e09U,0xdc12U,0xa82fU,
		0x4055U,0x80aaU,0x115fU,0x22beU,0x457cU,0x8af8U,0x05fbU,0x0bf6U,0x17ecU,0x2fd8U,
		0x5fb0U,0xbf60U,0x6ecbU,0xdd96U,0xab27U,0x4645U,0x8c8aU,0x091fU,0x123eU,0x247cU,
		0x48f8U,0x91f0U,0x33ebU,0x67d6U,0xcfacU,0x8f53U,0x0eadU,0x1d5aU,0x3ab4U,0x7568U,
		0xead0U,0xc5abU,0x9b5dU,0x26b1U,0x4d62U,0x9ac4U,0x2583U,0x4b06U,0x960cU,0x3c13U,
		0x7826U,0xf04cU,0xf093U,0xf12dU,0xf251U,0xf4a9U,0xf959U,0xe2b9U,0xd579U,0xbaf9U,
		0x65f9U,0xcbf2U,0x87efU,0x1fd5U,0x3faaU,0x7f54U,0xfea8U,0xed5bU,0xcabdU,0x8571U,
		0x1ae9U,0x35d2U,0x6ba4U,0xd748U,0xbe9bU,0x6d3dU,0xda7aU,0xa4ffU,0x59f5U,0xb3eaU,
		0x77dfU,0xefbeU,0xcf77U,0x8ee5U,0x0dc1U,0x1b82U,0x3704U,0x6e08U,0xdc10U,0xa82bU,
		0x405dU,0x80baU,0x117fU,0x22feU,0x45fcU,0x8bf8U,0x07fbU,0x0ff6U,0x1fecU,0x3fd8U,
		0x7fb0U,0xff60U,0xeecbU,0xcd9dU,0x8b31U,0x0669U,0x0cd2U,0x19a4U,0x3348U,0x6690U,
		0xcd20U,0x8a4bU,0x049dU,0x093aU,0x1274U,0x24e8U,0x49d0U,0x93a0U,0x374bU,0x6e96U,
		0xdd2cU,0xaa53U,0x44adU,0x895aU,0x02bfU,0x057eU,0x0afcU,0x15f8U,0x2bf0U,0x57e0U,
		0xafc0U,0x4f8bU,0x9f16U,0x2e27U,0x5c4eU,0xb89cU,0x6133U,0xc266U,0x94c7U,0x3985U,
		0x730aU,0xe614U,0xdc23U,0xa84dU,0x4091U,0x8122U,0x124fU,0x249eU,0x493cU,0x9278U,
		0x34fbU,0x69f6U,0xd3ecU,0xb7d3U,0x7fadU,0xff5aU,0xeebfU,0xcd75U,0x8ae1U,0x05c9U,
		0x0b92U,0x1724U,0x2e48U,0x5c90U,0xb920U,0x624bU,0xc496U,0x9927U,0x2245U,0x448aU,
		0x8914U,0x0223U,0x0446U,0x088cU,0x1118U,0x2230U,0x4460U,0x88c0U,0x018bU,0x0316U,
		0x062cU,0x0c58U,0x18b0U,0x3160U,0x62c0U,0xc580U,0x9b0bU,0x261dU,0x4c3aU,0x9874U,
		0x20e3U,0x41c6U,0x838cU,0x1713U,0x2e26U,0x5c4cU,0xb898U,0x613bU,0xc276U,0x94e7U,
		0x39c5U,0x738aU,0xe714U,0xde23U,0xac4dU,0x4891U,0x9122U,0x324fU,0x649eU,0xc93cU,
		0x8273U,0x14edU,0x29daU,0x53b4U,0xa768U,0x5edbU,0xbdb6U,0x6b67U,0xd6ceU,0xbd97U,
		0x6b25U,0xd64aU,0xbc9fU,0x6935U,0xd26aU,0xb4dfU,0x79b5U,0xf36aU,0xf6dfU,0xfdb5U,
		0xeb61U,0xc6c9U,0x9d99U,0x2b39U,0x5672U,0xace4U,0x49c3U,0x9386U,0x3707U,0x6e0eU,
		0xdc1cU,0xa833U,0x406dU,0x80daU,0x11bfU,0x237eU,0x46fcU,0x8df8U,0x0bfbU,0x17f6U,
		0x2fecU,0x5fd8U,0xbfb0U,0x6f6bU,0xded6U,0xada7U,0x4b45U,0x968aU,0x3d1fU,0x7a3eU,
		0xf47cU,0xf8f3U,0xe1edU,0xd3d1U,0xb7a9U,0x7f59U,0xfeb2U,0xed6fU,0xcad5U,0x85a1U,
		0x1b49U,0x3692U,0x6d24U,0xda48U,0xa49bU,0x593dU,0xb27aU,0x74ffU,0xe9feU,0xc3f7U,
		0x97e5U,0x3fc1U,0x7f82U,0xff04U,0xee03U,0xcc0dU,0x8811U,0x0029U,0x0052U,0x00a4U,
		0x0148U,0x0290U,0x0520U,0x0a40U,0x1480U,0x2900U,0x5200U,0xa400U,0x580bU,0xb016U,
		0x7027U,0xe04eU,0xd097U,0xb125U,0x7241U,0xe482U,0xd90fU,0xa215U,0x5421U,0xa842U,
		0x408fU,0x811eU,0x1237U,0x246eU,0x48dcU,0x91b8U,0x337bU,0x66f6U,0xcdecU,0x8bd3U,
		0x07adU,0x0f5aU,0x1eb4U,0x3d68U,0x7ad0U,0xf5a0U,0xfb4bU,0xe69dU,0xdd31U,0xaa69U,
		0x44d9U,0x89b2U,0x036fU,0x06deU,0x0dbcU,0x1b78U,0x36f0U,0x6de0U,0xdbc0U,0xa78bU,
		0x5f1dU,0xbe3aU,0x6c7fU,0xd8feU,0xa1f7U,0x53e5U,0xa7caU,0x5f9fU,0xbf3eU,0x6e77U,
		0xdceeU,0xa9d7U,0x43a5U,0x874aU,0x1e9fU,0x3d3eU,0x7a7cU,0xf4f8U,0xf9fbU,0xe3fdU,
		0xd7f1U,0xbfe9U,0x6fd9U,0xdfb2U,0xaf6fU,0x4ed5U,0x9daaU,0x2b5fU,0x56beU,0xad7cU,
		0x4af3U,0x95e6U,0x3bc7U,0x778eU,0xef1cU,0xce33U,0x8c6dU,0x08d1U,0x11a2U,0x2344U,
		0x4688U,0x8d10U,0x0a2bU,0x1456U,0x28acU,0x5158U,0xa2b0U,0x556bU,0xaad6U,0x45a7U,
		0x8b4eU,0x0697U,0x0d2eU,0x1a5cU,0x34b8U,0x6970U,0xd2e0U,0xb5cbU,0x7b9dU,0xf73aU,
		0xfe7fU,0xecf5U,0xc9e1U,0x83c9U,0x1799U,0x2f32U,0x5e64U,0xbcc8U,0x699bU,0xd336U,
		0xb667U,0x7cc5U,0xf98aU,0xe31fU,0xd635U,0xbc61U,0x68c9U,0xd192U,0xb32fU,0x7655U,
		0xecaaU,0xc95fU,0x82b5U,0x1561U,0x2ac2U,0x5584U,0xab08U,0x461bU,0x8c36U,0x0867U,
		0x10ceU,0x219cU,0x4338U,0x8670U,0x1cebU,0x39d6U,0x73acU,0xe758U,0xdebbU,0xad7dU,
		0x4af1U,0x95e2U,0x3bcfU,0x779eU,0xef3cU,0xce73U,0x8cedU,0x09d1U,0x13a2U,0x2744U,
		0x4e88U,0x9d10U,0x2a2bU,0x5456U,0xa8acU,0x4153U,0x82a6U,0x1547U,0x2a8eU,0x551cU,
		0xaa38U,0x447bU,0x88f6U,0x01e7U,0x03ceU,0x079cU,0x0f38U,0x1e70U,0x3ce0U,0x79c0U,
		0xf380U,0xf70bU,0xfe1dU,0xec31U,0xc869U,0x80d9U,0x11b9U,0x2372U,0x46e4U,0x8dc8U,
		0x0b9bU,0x1736U,0x2e6cU,0x5cd8U,0xb9b0U,0x636bU,0xc6d6U,0x9da7U,0x2b45U,0x568aU,
		0xad14U,0x4a23U,0x9446U,0x3887U,0x710eU,0xe21cU,0xd433U,0xb86dU,0x60d1U,0xc1a2U,
		0x934fU,0x3695U,0x6d2aU,0xda54U,0xa4a3U,0x594dU,0xb29aU,0x753fU,0xea7eU,0xc4f7U,
		0x99e5U,0x23c1U,0x4782U,0x8f04U,0x0e03U,0x1c06U,0x380cU,0x7018U,0xe030U,0xd06bU,
		0xb0ddU,0x71b1U,0xe362U,0xd6cfU,0xbd95U,0x6b21U,0xd642U,0xbc8fU,0x6915U,0xd22aU,
		0xb45fU,0x78b5U,0xf16aU,0xf2dfU,0xf5b5U,0xfb61U,0xe6c9U,0xdd99U,0xab39U,0x4679U,
		0x8cf2U,0x09efU,0x13deU,0x27bcU,0x4f78U,0x9ef0U,0x2debU,0x5bd6U,0xb7acU,0x7f53U,
		0xfea6U,0xed47U,0xca85U,0x8501U,0x1a09U,0x3412U,0x6824U,0xd048U,0xb09bU,0x713dU,
		0xe27aU,0xd4ffU,0xb9f5U,0x63e1U,0xc7c2U,0x9f8fU,0x2f15U,0x5e2aU,0xbc54U,0x68a3U,
		0xd146U,0xb287U,0x7505U,0xea0aU,0xc41fU,0x9835U,0x2061U,0x40c2U,0x8184U,0x1303U,
		0x2606U,0x4c0cU,0x9818U,0x203bU,0x4076U,0x80ecU,0x11d3U,0x23a6U,0x474cU,0x8e98U,
		0x0d3bU,0x1a76U,0x34ecU,0x69d8U,0xd3b0U,0xb76bU,0x7eddU,0xfdbaU,0xeb7fU,0xc6f5U,
		0x9de1U,0x2bc9U,0x5792U,0xaf24U,0x4e43U,0x9c86U,0x2907U,0x520eU,0xa41cU,0x5833U,
		0xb066U,0x70c7U,0xe18eU,0xd317U,0xb625U,0x7c41U,0xf882U,0xe10fU,0xd215U,0xb421U,
		0x7849U,0xf092U,0xf12fU,0xf255U,0xf4a1U,0xf949U,0xe299U,0xd539U,0xba79U,0x64f9U,
		0xc9f2U,0x83efU,0x17d5U,0x2faaU,0x5f54U,0xbea8U,0x6d5bU,0xdab6U,0xa567U,0x5ac5U,
		0xb58aU,0x7b1fU,0xf63eU,0xfc77U,0xe8e5U,0xc1c1U,0x9389U,0x3719U,0x6e32U,0xdc64U,
		0xa8c3U,0x418dU,0x831aU,0x163fU,0x2c7eU,0x58fcU,0xb1f8U,0x73fbU,0xe7f6U,0xdfe7U,
		0xafc5U,0x4f81U,0x9f02U,0x2e0fU,0x5c1eU,0xb83cU,0x6073U,0xc0e6U,0x91c7U,0x3385U,
		0x670aU,0xce14U,0x8c23U,0x084dU,0x109aU,0x2134U,0x4268U,0x84d0U,0x19abU,0x3356U,
		0x66acU,0xcd58U,0x8abbU,0x057dU,0x0afaU,0x15f4U,0x2be8U,0x57d0U,0xafa0U,0x4f4bU,
		0x9e96U,0x2d27U,0x5a4eU,0xb49cU,0x7933U,0xf266U,0xf4c7U,0xf985U,0xe301U,0xd609U,
		0xbc19U,0x6839U,0xd072U,0xb0efU,0x71d5U,0xe3aaU,0xd75fU,0xbeb5U,0x6d61U,0xdac2U,
		0xa58fU,0x5b15U,0xb62aU,0x7c5fU,0xf8beU,0xe177U,0xd2e5U,0xb5c1U,0x7b89U,0xf712U,
		0xfe2fU,0xec55U,0xc8a1U,0x8149U,0x1299U,0x2532U,0x4a64U,0x94c8U,0x399bU,0x7336U,
		0xe66cU,0xdcd3U,0xa9adU,0x4351U,0x86a2U,0x1d4fU,0x3a9eU,0x753cU,0xea78U,0xc4fbU,
		0x99fdU,0x23f1U,0x47e2U,0x8fc4U,0x0f83U,0x1f06U,0x3e0cU,0x7c18U,0xf830U,0xe06bU,
		0xd0ddU,0xb1b1U,0x7369U,0xe6d2U,0xddafU,0xab55U,0x46a1U,0x8d42U,0x0a8fU,0x151eU,
		0x2a3cU,0x5478U,0xa8f0U,0x41ebU,0x83d6U,0x17a7U,0x2f4eU,0x5e9cU,0xbd38U,0x6a7bU,
		0xd4f6U,0xb9e7U,0x63c5U,0xc78aU,0x9f1fU,0x2e35U,0x5c6aU,0xb8d4U,0x61a3U,0xc346U,
		0x9687U,0x3d05U,0x7a0aU,0xf414U,0xf823U,0xe04dU,0xd091U,0xb129U,0x7259U,0xe4b2U,
		0xd96fU,0xa2d5U,0x55a1U,0xab42U,0x468fU,0x8d1eU,0x0a37U,0x146eU,0x28dcU,0x51b8U,
		0xa370U,0x56ebU,0xadd6U,0x4ba7U,0x974eU,0x3e97U,0x7d2eU,0xfa5cU,0xe4b3U,0xd96dU,
		0xa2d1U,0x55a9U,0xab52U,0x46afU,0x8d5eU,0x0ab7U,0x156eU,0x2adcU,0x55b8U,0xab70U,
		0x46ebU,0x8dd6U,0x0ba7U,0x174eU,0x2e9cU,0x5d38U,0xba70U,0x64ebU,0xc9d6U,0x83a7U,
		0x1745U,0x2e8aU,0x5d14U,0xba28U,0x645bU,0xc8b6U,0x8167U,0x12c5U,0x258aU,0x4b14U,
		0x9628U,0x3c5bU,0x78b6U,0xf16cU,0xf2d3U,0xf5adU,0xfb51U,0xe6a9U,0xdd59U,0xaab9U,
		0x4579U,0x8af2U,0x05efU,0x0bdeU,0x17bcU,0x2f78U,0x5ef0U,0xbde0U,0x6bcbU,0xd796U,
		0xbf27U,0x6e45U,0xdc8aU,0xa91fU,0x4235U,0x846aU,0x18dfU,0x31beU,0x637cU,0xc6f8U,
		0x9dfbU,0x2bfdU,0x57faU,0xaff4U,0x4fe3U,0x9fc6U,0x2f87U,0x5f0eU,0xbe1cU,0x6c33U,
		0xd866U,0xa0c7U,0x5185U,0xa30aU,0x561fU,0xac3eU,0x4877U,0x90eeU,0x31d7U,0x63aeU,
		0xc75cU,0x9eb3U,0x2d6dU,0x5adaU,0xb5b4U,0x7b63U,0xf6c6U,0xfd87U,0xeb05U,0xc601U,
		0x9c09U,0x2819U,0x5032U,0xa064U,0x50c3U,0xa186U,0x5307U,0xa60eU,0x5c17U,0xb82eU,
		0x6057U,0xc0aeU,0x9157U,0x32a5U,0x654aU,0xca94U,0x8523U,0x1a4dU,0x349aU,0x6934U,
		0xd268U,0xb4dbU,0x79bdU,0xf37aU,0xf6ffU,0xfdf5U,0xebe1U,0xc7c9U,0x9f99U,0x2f39U,
		0x5e72U,0xbce4U,0x69c3U,0xd386U,0xb707U,0x7e05U,0xfc0aU,0xe81fU,0xc035U,0x9061U,
		0x30c9U,0x6192U,0xc324U,0x9643U,0x3c8dU,0x791aU,0xf234U,0xf463U,0xf8cdU,0xe191U,
		0xd329U,0xb659U,0x7cb9U,0xf972U,0xe2efU,0xd5d5U,0xbba1U,0x6749U,0xce92U,0x8d2fU,
		0x0a55U,0x14aaU,0x2954U,0x52a8U,0xa550U,0x5aabU,0xb556U,0x7aa7U,0xf54eU,0xfa97U,
		0xe525U,0xda41U,0xa489U,0x5919U,0xb232U,0x746fU,0xe8deU,0xc1b7U,0x9365U,0x36c1U,
		0x6d82U,0xdb04U,0xa603U,0x5c0dU,0xb81aU,0x603fU,0xc07eU,0x90f7U,0x31e5U,0x63caU,
		0xc794U,0x9f23U,0x2e4dU,0x5c9aU,0xb934U,0x6263U,0xc4c6U,0x9987U,0x2305U,0x460aU,
		0x8c14U,0x0823U,0x1046U,0x208cU,0x4118U,0x8230U,0x146bU,0x28d6U,0x51acU,0xa358U,
		0x56bbU,0xad76U,0x4ae7U,0x95ceU,0x3b97U,0x772eU,0xee5cU,0xccb3U,0x896dU,0x02d1U,
		0x05a2U,0x0b44U,0x1688U,0x2d10U,0x5a20U,0xb440U,0x788bU,0xf116U,0xf227U,0xf445U,
		0xf881U,0xe109U,0xd219U,0xb439U,0x7879U,0xf0f2U,0xf1efU,0xf3d5U,0xf7a1U,0xff49U,
		0xee99U,0xcd39U,0x8a79U,0x04f9U,0x09f2U,0x13e4U,0x27c8U,0x4f90U,0x9f20U,0x2e4bU,
		0x5c96U,0xb92cU,0x6253U,0xc4a6U,0x9947U,0x2285U,0x450aU,0x8a14U,0x0423U,0x0846U,
		0x108cU,0x2118U,0x4230U,0x8460U,0x18cbU,0x3196U,0x632cU,0xc658U,0x9cbbU,0x297dU,
		0x52faU,0xa5f4U,0x5be3U,0xb7c6U,0x7f87U,0xff0eU,0xee17U,0xcc25U,0x8841U,0x0089U,
		0x0112U,0x0224U,0x0448U,0x0890U,0x1120U,0x2240U,0x4480U,0x8900U,0x020bU,0x0416U,
		0x082cU,0x1058U,0x20b0U,0x4160U,0x82c0U,0x158bU,0x2b16U,0x562cU,0xac58U,0x48bbU,
		0x9176U,0x32e7U,0x65ceU,0xcb9cU,0x8733U,0x1e6dU,0x3cdaU,0x79b4U,0xf368U,0xf6dbU,
		0xfdbdU,0xeb71U,0xc6e9U,0x9dd9U,0x2bb9U,0x5772U,0xaee4U,0x4dc3U,0x9b86U,0x2707U,
		0x4e0eU,0x9c1cU,0x2833U,0x5066U,0xa0ccU,0x5193U,0xa326U,0x5647U,0xac8eU,0x4917U,
		0x922eU,0x3457U,0x68aeU,0xd15cU,0xb2b3U,0x756dU,0xeadaU,0xc5bfU,0x9b75U,0x26e1U,
		0x4dc2U,0x9b84U,0x2703U,0x4e06U,0x9c0cU,0x2813U,0x5026U,0xa04cU,0x5093U,0xa126U,
		0x5247U,0xa48eU,0x5917U,0xb22eU,0x7457U,0xe8aeU,0xc157U,0x92a5U,0x3541U,0x6a82U,
		0xd504U,0xba03U,0x640dU,0xc81aU,0x803fU,0x1075U,0x20eaU,0x41d4U,0x83a8U,0x175bU,
		0x2eb6U,0x5d6cU,0xbad8U,0x65bbU,0xcb76U,0x86e7U,0x1dc5U,0x3b8aU,0x7714U,0xee28U,
		0xcc5bU,0x88bdU,0x0171U,0x02e2U,0x05c4U,0x0b88U,0x1710U,0x2e20U,0x5c40U,0xb880U,
		0x610bU,0xc216U,0x9427U,0x3845U,0x708aU,0xe114U,0xd223U,0xb44dU,0x7891U,0xf122U,
		0xf24fU,0xf495U,0xf921U,0xe249U,0xd499U,0xb939U,0x6279U,0xc4f2U,0x99efU,0x23d5U,
		0x47aaU,0x8f54U,0x0ea3U,0x1d46U,0x3a8cU,0x7518U,0xea30U,0xc46bU,0x98ddU,0x21b1U,
		0x4362U,0x86c4U,0x1d83U,0x3b06U,0x760cU,0xec18U,0xc83bU,0x807dU,0x10f1U,0x21e2U,
		0x43c4U,0x8788U,0x1f1bU,0x3e36U,0x7c6cU,0xf8d8U,0xe1bbU,0xd37dU,0xb6f1U,0x7de9U,
		0xfbd2U,0xe7afU,0xdf55U,0xaea1U,0x4d49U,0x9a92U,0x252fU,0x4a5eU,0x94bcU,0x3973U,
		0x72e6U,0xe5ccU,0xdb93U,0xa72dU,0x5e51U,0xbca2U,0x694fU,0xd29eU,0xb537U,0x7a65U,
		0xf4caU,0xf99fU,0xe335U,0xd661U,0xbcc9U,0x6999U,0xd332U,0xb66fU,0x7cd5U,0xf9aaU,
		0xe35fU,0xd6b5U,0xbd61U,0x6ac9U,0xd592U,0xbb2fU,0x6655U,0xccaaU,0x895fU,0x02b5U,
		0x056aU,0x0ad4U,0x15a8U,0x2b50U,0x56a0U,0xad40U,0x4a8bU,0x9516U,0x3a27U,0x744eU,
		0xe89cU,0xc133U,0x926dU,0x34d1U,0x69a2U,0xd344U,0xb683U,0x7d0dU,0xfa1aU,0xe43fU,
		0xd875U,0xa0e1U,0x51c9U,0xa392U,0x572fU,0xae5eU,0x4cb7U,0x996eU,0x22d7U,0x45aeU,
		0x8b5cU,0x06b3U,0x0d66U,0x1accU,0x3598U,0x6b30U,0xd660U,0xbccbU,0x699dU,0xd33aU,
		0xb67fU,0x7cf5U,0xf9eaU,0xe3dfU,0xd7b5U,0xbf61U,0x6ec9U,0xdd92U,0xab2fU,0x4655U,
		0x8caaU,0x095fU,0x12beU,0x257cU,0x4af8U,0x95f0U,0x3bebU,0x77d6U,0xefacU,0xcf53U,
		0x8eadU,0x0d51U,0x1aa2U,0x3544U,0x6a88U,0xd510U,0xba2bU,0x645dU,0xc8baU,0x817fU,
		0x12f5U,0x25eaU,0x4bd4U,0x97a8U,0x3f5bU,0x7eb6U,0xfd6cU,0xead3U,0xc5adU,0x9b51U,
		0x26a9U,0x4d52U,0x9aa4U,0x2543U,0x4a86U,0x950cU,0x3a13U,0x7426U,0xe84cU,0xc093U,
		0x912dU,0x3251U,0x64a2U,0xc944U,0x8283U,0x150dU,0x2a1aU,0x5434U,0xa868U,0x40dbU,
		0x81b6U,0x1367U,0x26ceU,0x4d9cU,0x9b38U,0x267bU,0x4cf6U,0x99ecU,0x23d3U,0x47a6U,
		0x8f4cU,0x0e93U,0x1d26U,0x3a4cU,0x7498U,0xe930U,0xc26bU,0x94ddU,0x39b1U,0x7362U,
		0xe6c4U,0xdd83U,0xab0dU,0x4611U,0x8c22U,0x084fU,0x109eU,0x213cU,0x4278U,0x84f0U,
		0x19ebU,0x33d6U,0x67acU,0xcf58U,0x8ebbU,0x0d7dU,0x1afaU,0x35f4U,0x6be8U,0xd7d0U,
		0xbfabU,0x6f5dU,0xdebaU,0xad7fU,0x4af5U,0x95eaU,0x3bdfU,0x77beU,0xef7cU,0xcef3U,
		0x8dedU,0x0bd1U,0x17a2U,0x2f44U,0x5e88U,0xbd10U,0x6a2bU,0xd456U,0xb8a7U,0x6145U,
		0xc28aU,0x951fU,0x3a35U,0x746aU,0xe8d4U,0xc1a3U,0x934dU,0x3691U,0x6d22U,0xda44U,
		0xa483U,0x590dU,0xb21aU,0x743fU,0xe87eU,0xc0f7U,0x91e5U,0x33c1U,0x6782U,0xcf04U,
		0x8e03U,0x0c0dU,0x181aU,0x3034U,0x6068U,0xc0d0U,0x91abU,0x335dU,0x66baU,0xcd74U,
		0x8ae3U,0x05cdU,0x0b9aU,0x1734U,0x2e68U,0x5cd0U,0xb9a0U,0x634bU,0xc696U,0x9d27U,
		0x2a45U,0x548aU,0xa914U,0x4223U,0x8446U,0x1887U,0x310eU,0x621cU,0xc438U,0x987bU,
		0x20fdU,0x41faU,0x83f4U,0x17e3U,0x2fc6U,0x5f8cU,0xbf18U,0x6e3bU,0xdc76U,0xa8e7U,
		0x41c5U,0x838aU,0x171fU,0x2e3eU,0x5c7cU,0xb8f8U,0x61fbU,0xc3f6U,0x97e7U,0x3fc5U,
		0x7f8aU,0xff14U,0xee23U,0xcc4dU,0x8891U,0x0129U,0x0252U,0x04a4U,0x0948U,0x1290U,
		0x2520U,0x4a40U,0x9480U,0x390bU,0x7216U,0xe42cU,0xd853U,0xa0adU,0x5151U,0xa2a2U,
		0x554fU,0xaa9eU,0x4537U,0x8a6eU,0x04d7U,0x09aeU,0x135cU,0x26b8U,0x4d70U,0x9ae0U,
		0x25cbU,0x4b96U,0x972cU,0x3e53U,0x7ca6U,0xf94cU,0xe293U,0xd52dU,0xba51U,0x64a9U,
		0xc952U,0x82afU,0x1555U,0x2aaaU,0x5554U,0xaaa8U,0x455bU,0x8ab6U,0x0567U,0x0aceU,
		0x159cU,0x2b38U,0x5670U,0xace0U,0x49cbU,0x9396U,0x3727U,0x6e4eU,0xdc9cU,0xa933U,
		0x426dU,0x84daU,0x19bfU,0x337eU,0x66fcU,0xcdf8U,0x8bfbU,0x07fdU,0x0ffaU,0x1ff4U,
		0x3fe8U,0x7fd0U,0xffa0U,0xef4bU,0xce9dU,0x8d31U,0x0a69U,0x14d2U,0x29a4U,0x5348U,
		0xa690U,0x5d2bU,0xba56U,0x64a7U,0xc94eU,0x8297U,0x1525U,0x2a4aU,0x5494U,0xa928U,
		0x425bU,0x84b6U,0x1967U,0x32ceU,0x659cU,0xcb38U,0x867bU,0x1cfdU,0x39faU,0x73f4U,
		0xe7e8U,0xdfdbU,0xafbdU,0x4f71U,0x9ee2U,0x2dcfU,0x5b9eU,0xb73cU,0x7e73U,0xfce6U,
		0xe9c7U,0xc385U,0x9701U,0x3e09U,0x7c12U,0xf824U,0xe043U,0xd08dU,0xb111U,0x7229U,
		0xe452U,0xd8afU,0xa155U,0x52a1U,0xa542U,0x5a8fU,0xb51eU,0x7a37U,0xf46eU,0xf8d7U,
		0xe1a5U,0xd341U,0xb689U,0x7d19U,0xfa32U,0xe46fU,0xd8d5U,0xa1a1U,0x5349U,0xa692U,
		0x5d2fU,0xba5eU,0x64b7U,0xc96eU,0x82d7U,0x15a5U,0x2b4aU,0x5694U,0xad28U,0x4a5bU,
		0x94b6U,0x3967U,0x72ceU,0xe59cU,0xdb33U,0xa66dU,0x5cd1U,0xb9a2U,0x634fU,0xc69eU,
		0x9d37U,0x2a65U,0x54caU,0xa994U,0x4323U,0x8646U,0x1c87U,0x390eU,0x721cU,0xe438U,
		0xd87bU,0xa0fdU,0x51f1U,0xa3e2U,0x57cfU,0xaf9eU,0x4f37U,0x9e6eU,0x2cd7U,0x59aeU,
		0xb35cU,0x76b3U,0xed66U,0xcac7U,0x8585U,0x1b01U,0x3602U,0x6c04U,0xd808U,0xa01bU,
		0x503dU,0xa07aU,0x50ffU,0xa1feU,0x53f7U,0xa7eeU,0x5fd7U,0xbfaeU,0x6f57U,0xdeaeU,
		0xad57U,0x4aa5U,0x954aU,0x3a9fU,0x753eU,0xea7cU,0xc4f3U,0x99edU,0x23d1U,0x47a2U,
		0x8f44U,0x0e83U,0x1d06U,0x3a0cU,0x7418U,0xe830U,0xc06bU,0x90ddU,0x31b1U,0x6362U,
		0xc6c4U,0x9d83U,0x2b0dU,0x561aU,0xac34U,0x4863U,0x90c6U,0x3187U,0x630eU,0xc61cU,
		0x9c33U,0x286dU,0x50daU,0xa1b4U,0x5363U,0xa6c6U,0x5d87U,0xbb0eU,0x6617U,0xcc2eU,
		0x8857U,0x00a5U,0x014aU,0x0294U,0x0528U,0x0a50U,0x14a0U,0x2940U,0x5280U,0xa500U,
		0x5a0bU,0xb416U,0x7827U,0xf04eU,0xf097U,0xf125U,0xf241U,0xf489U,0xf919U,0xe239U,
		0xd479U,0xb8f9U,0x61f9U,0xc3f2U,0x97efU,0x3fd5U,0x7faaU,0xff54U,0xeea3U,0xcd4dU,
		0x8a91U,0x0529U,0x0a52U,0x14a4U,0x2948U,0x5290U,0xa520U,0x5a4bU,0xb496U,0x7927U,
		0xf24eU,0xf497U,0xf925U,0xe241U,0xd489U,0xb919U,0x6239U,0xc472U,0x98efU,0x21d5U,
		0x43aaU,0x8754U,0x1ea3U,0x3d46U,0x7a8cU,0xf518U,0xfa3bU,0xe47dU,0xd8f1U,0xa1e9U,
		0x53d9U,0xa7b2U,0x5f6fU,0xbedeU,0x6db7U,0xdb6eU,0xa6d7U,0x5da5U,0xbb4aU,0x669fU,
		0xcd3eU,0x8a77U,0x04e5U,0x09caU,0x1394U,0x2728U,0x4e50U,0x9ca0U,0x294bU,0x5296U,
		0xa52cU,0x5a53U,0xb4a6U,0x7947U,0xf28eU,0xf517U,0xfa25U,0xe441U,0xd889U,0xa119U,
		0x5239U,0xa472U,0x58efU,0xb1deU,0x73b7U,0xe76eU,0xded7U,0xada5U,0x4b41U,0x9682U,
		0x3d0fU,0x7a1eU,0xf43cU,0xf873U,0xe0edU,0xd1d1U,0xb3a9U,0x7759U,0xeeb2U,0xcd6fU,
		0x8ad5U,0x05a1U,0x0b42U,0x1684U,0x2d08U,0x5a10U,0xb420U,0x784bU,0xf096U,0xf127U,
		0xf245U,0xf481U,0xf909U,0xe219U,0xd439U,0xb879U,0x60f9U,0xc1f2U,0x93efU,0x37d5U,
		0x6faaU,0xdf54U,0xaea3U,0x4d4dU,0x9a9aU,0x253fU,0x4a7eU,0x94fcU,0x39f3U,0x73e6U,
		0xe7ccU,0xdf93U,0xaf2dU,0x4e51U,0x9ca2U,0x294fU,0x529eU,0xa53cU,0x5a73U,0xb4e6U,
		0x79c7U,0xf38eU,0xf717U,0xfe25U,0xec41U,0xc889U,0x8119U,0x1239U,0x2472U,0x48e4U,
		0x91c8U,0x339bU,0x6736U,0xce6cU,0x8cd3U,0x09adU,0x135aU,0x26b4U,0x4d68U,0x9ad0U,
		0x25abU,0x4b56U,0x96acU,0x3d53U,0x7aa6U,0xf54cU,0xfa93U,0xe52dU,0xda51U,0xa4a9U,
		0x5959U,0xb2b2U,0x756fU,0xeadeU,0xc5b7U,0x9b65U,0x26c1U,0x4d82U,0x9b04U,0x2603U,
		0x4c06U,0x980cU,0x2013U,0x4026U,0x804cU,0x1093U,0x2126U,0x424cU,0x8498U,0x193bU,
		0x3276U,0x64ecU,0xc9d8U,0x83bbU,0x177dU,0x2efaU,0x5df4U,0xbbe8U,0x67dbU,0xcfb6U,
		0x8f67U,0x0ec5U,0x1d8aU,0x3b14U,0x7628U,0xec50U,0xc8abU,0x815dU,0x12b1U,0x2562U,
		0x4ac4U,0x9588U,0x3b1bU,0x7636U,0xec6cU,0xc8d3U,0x81adU,0x1351U,0x26a2U,0x4d44U,
		0x9a88U,0x251bU,0x4a36U,0x946cU,0x38d3U,0x71a6U,0xe34cU,0xd693U,0xbd2dU,0x6a51U,
		0xd4a2U,0xb94fU,0x6295U,0xc52aU,0x9a5fU,0x24b5U,0x496aU,0x92d4U,0x35a3U,0x6b46U,
		0xd68cU,0xbd13U,0x6a2dU,0xd45aU,0xb8bfU,0x6175U,0xc2eaU,0x95dfU,0x3bb5U,0x776aU,
		0xeed4U,0xcda3U,0x8b4dU,0x0691U,0x0d22U,0x1a44U,0x3488U,0x6910U,0xd220U,0xb44bU,
		0x789dU,0xf13aU,0xf27fU,0xf4f5U,0xf9e1U,0xe3c9U,0xd799U,0xbf39U,0x6e79U,0xdcf2U,
		0xa9efU,0x43d5U,0x87aaU,0x1f5fU,0x3ebeU,0x7d7cU,0xfaf8U,0xe5fbU,0xdbfdU,0xa7f1U,
		0x5fe9U,0xbfd2U,0x6fafU,0xdf5eU,0xaeb7U,0x4d65U,0x9acaU,0x259fU,0x4b3eU,0x967cU,
		0x3cf3U,0x79e6U,0xf3ccU,0xf793U,0xff2dU,0xee51U,0xcca9U,0x8959U,0x02b9U,0x0572U,
		0x0ae4U,0x15c8U,0x2b90U,0x5720U,0xae40U,0x4c8bU,0x9916U,0x2227U,0x444eU,0x889cU,
		0x0133U,0x0266U,0x04ccU,0x0998U,0x1330U,0x2660U,0x4cc0U,0x9980U,0x230bU,0x4616U,
		0x8c2cU,0x0853U,0x10a6U,0x214cU,0x4298U,0x8530U,0x1a6bU,0x34d6U,0x69acU,0xd358U,
		0xb6bbU,0x7d7dU,0xfafaU,0xe5ffU,0xdbf5U,0xa7e1U,0x5fc9U,0xbf92U,0x6f2fU,0xde5eU,
		0xacb7U,0x4965U,0x92caU,0x359fU,0x6b3eU,0xd67cU,0xbcf3U,0x69edU,0xd3daU,0xb7bfU,
		0x7f75U,0xfeeaU,0xeddfU,0xcbb5U,0x8761U,0x1ec9U,0x3d92U,0x7b24U,0xf648U,0xfc9bU,
		0xe93dU,0xc271U,0x94e9U,0x39d9U,0x73b2U,0xe764U,0xdec3U,0xad8dU,0x4b11U,0x9622U,
		0x3c4fU,0x789eU,0xf13cU,0xf273U,0xf4edU,0xf9d1U,0xe3a9U,0xd759U,0xbeb9U,0x6d79U,
		0xdaf2U,0xa5efU,0x5bd5U,0xb7aaU,0x7f5fU,0xfebeU,0xed77U,0xcae5U,0x85c1U,0x1b89U,
		0x3712U,0x6e24U,0xdc48U,0xa89bU,0x413dU,0x827aU,0x14ffU,0x29feU,0x53fcU,0xa7f8U,
		0x5ffbU,0xbff6U,0x6fe7U,0xdfceU,0xaf97U,0x4f25U,0x9e4aU,0x2c9fU,0x593eU,0xb27cU,
		0x74f3U,0xe9e6U,0xc3c7U,0x9785U,0x3f01U,0x7e02U,0xfc04U,0xe803U,0xc00dU,0x9011U,
		0x3029U,0x6052U,0xc0a4U,0x9143U,0x328dU,0x651aU,0xca34U,0x8463U,0x18cdU,0x319aU,
		0x6334U,0xc668U,0x9cdbU,0x29bdU,0x537aU,0xa6f4U,0x5de3U,0xbbc6U,0x6787U,0xcf0eU,
		0x8e17U,0x0c25U,0x184aU,0x3094U,0x6128U,0xc250U,0x94abU,0x395dU,0x72baU,0xe574U,
		0xdae3U,0xa5cdU,0x5b91U,0xb722U,0x7e4fU,0xfc9eU,0xe937U,0xc265U,0x94c1U,0x3989U,
		0x7312U,0xe624U,0xdc43U,0xa88dU,0x4111U,0x8222U,0x144fU,0x289eU,0x513cU,0xa278U,
		0x54fbU,0xa9f6U,0x43e7U,0x87ceU,0x1f97U,0x3f2eU,0x7e5cU,0xfcb8U,0xe97bU,0xc2fdU,
		0x95f1U,0x3be9U,0x77d2U,0xefa4U,0xcf43U,0x8e8dU,0x0d11U,0x1a22U,0x3444U,0x6888U,
		0xd110U,0xb22bU,0x745dU,0xe8baU,0xc17fU,0x92f5U,0x35e1U,0x6bc2U,0xd784U,0xbf03U,
		0x6e0dU,0xdc1aU,0xa83fU,0x4075U,0x80eaU,0x11dfU,0x23beU,0x477cU,0x8ef8U,0x0dfbU,
		0x1bf6U,0x37ecU,0x6fd8U,0xdfb0U,0xaf6bU,0x4eddU,0x9dbaU,0x2b7fU,0x56feU,0xadfcU,
		0x4bf3U,0x97e6U,0x3fc7U,0x7f8eU,0xff1cU,0xee33U,0xcc6dU,0x88d1U,0x01a9U,0x0352U,
		0x06a4U,0x0d48U,0x1a90U,0x3520U,0x6a40U,0xd480U,0xb90bU,0x621dU,0xc43aU,0x987fU,
		0x20f5U,0x41eaU,0x83d4U,0x17a3U,0x2f46U,0x5e8cU,0xbd18U,0x6a3bU,0xd476U,0xb8e7U,
		0x61c5U,0xc38aU,0x971fU,0x3e35U,0x7c6aU,0xf8d4U,0xe1a3U,0xd34dU,0xb691U,0x7d29U,
		0xfa52U,0xe4afU,0xd955U,0xa2a1U,0x5549U,0xaa92U,0x452fU,0x8a5eU,0x04b7U,0x096eU,
		0x12dcU,0x25b8U,0x4b70U,0x96e0U,0x3dcbU,0x7b96U,0xf72cU,0xfe53U,0xecadU,0xc951U,
		0x82a9U,0x1559U,0x2ab2U,0x5564U,0xaac8U,0x459bU,0x8b36U,0x0667U,0x0cceU,0x199cU,
		0x3338U,0x6670U,0xcce0U,0x89cbU,0x039dU,0x073aU,0x0e74U,0x1ce8U,0x39d0U,0x73a0U,
		0xe740U,0xde8bU,0xad1dU,0x4a31U,0x9462U,0x38cfU,0x719eU,0xe33cU,0xd673U,0xbcedU,
		0x69d1U,0xd3a2U,0xb74fU,0x7e95U,0xfd2aU,0xea5fU,0xc4b5U,0x9961U,0x22c9U,0x4592U,
		0x8b24U,0x0643U,0x0c86U,0x190cU,0x3218U,0x6430U,0xc860U,0x80cbU,0x119dU,0x233aU,
		0x4674U,0x8ce8U,0x09dbU,0x13b6U,0x276cU,0x4ed8U,0x9db0U,0x2b6bU,0x56d6U,0xadacU,
		0x4b53U,0x96a6U,0x3d47U,0x7a8eU,0xf51cU,0xfa33U,0xe46dU,0xd8d1U,0xa1a9U,0x5359U,
		0xa6b2U,0x5d6fU,0xbadeU,0x65b7U,0xcb6eU,0x86d7U,0x1da5U,0x3b4aU,0x7694U,0xed28U,
		0xca5bU,0x84bdU,0x1971U,0x32e2U,0x65c4U,0xcb88U,0x871bU,0x1e3dU,0x3c7aU,0x78f4U,
		0xf1e8U,0xf3dbU,0xf7bdU,0xff71U,0xeee9U,0xcdd9U,0x8bb9U,0x0779U,0x0ef2U,0x1de4U,
		0x3bc8U,0x7790U,0xef20U,0xce4bU,0x8c9dU,0x0931U,0x1262U,0x24c4U,0x4988U,0x9310U,
		0x362bU,0x6c56U,0xd8acU,0xa153U,0x52adU,0xa55aU,0x5abfU,0xb57eU,0x7af7U,0xf5eeU,
		0xfbd7U,0xe7a5U,0xdf41U,0xae89U,0x4d19U,0x9a32U,0x246fU,0x48deU,0x91bcU,0x3373U,
		0x66e6U,0xcdccU,0x8b93U,0x072dU,0x0e5aU,0x1cb4U,0x3968U,0x72d0U,0xe5a0U,0xdb4bU,
		0xa69dU,0x5d31U,0xba62U,0x64cfU,0xc99eU,0x8337U,0x1665U,0x2ccaU,0x5994U,0xb328U,
		0x765bU,0xecb6U,0xc967U,0x82c5U,0x1581U,0x2b02U,0x5604U,0xac08U,0x481bU,0x9036U,
		0x3067U,0x60ceU,0xc19cU,0x9333U,0x366dU,0x6cdaU,0xd9b4U,0xa363U,0x56cdU,0xad9aU,
		0x4b3fU,0x967eU,0x3cf7U,0x79eeU,0xf3dcU,0xf7b3U,0xff6dU,0xeed1U,0xcda9U,0x8b59U,
		0x06b9U,0x0d72U,0x1ae4U,0x35c8U,0x6b90U,0xd720U,0xbe4bU,0x6c9dU,0xd93aU,0xa27fU,
		0x54f5U,0xa9eaU,0x43dfU,0x87beU,0x1f77U,0x3eeeU,0x7ddcU,0xfbb8U,0xe77bU,0xdefdU,
		0xadf1U,0x4be9U,0x97d2U,0x3fafU,0x7f5eU,0xfebcU,0xed73U,0xcaedU,0x85d1U,0x1ba9U,
		0x3752U,0x6ea4U,0xdd48U,0xaa9bU,0x453dU,0x8a7aU,0x04ffU,0x09feU,0x13fcU,0x27f8U,
		0x4ff0U,0x9fe0U,0x2fcbU,0x5f96U,0xbf2cU,0x6e53U,0xdca6U,0xa947U,0x4285U,0x850aU,
		0x1a1fU,0x343eU,0x687cU,0xd0f8U,0xb1fbU,0x73fdU,0xe7faU,0xdfffU,0xaff5U,0x4fe1U,
		0x9fc2U,0x2f8fU,0x5f1eU,0xbe3cU,0x6c73U,0xd8e6U,0xa1c7U,0x5385U,0xa70aU,0x5e1fU,
		0xbc3eU,0x6877U,0xd0eeU,0xb1d7U,0x73a5U,0xe74aU,0xde9fU,0xad35U,0x4a61U,0x94c2U,
		0x398fU,0x731eU,0xe63cU,0xdc73U,0xa8edU,0x41d1U,0x83a2U,0x174fU,0x2e9eU,0x5d3cU,
		0xba78U,0x64fbU,0xc9f6U,0x83e7U,0x17c5U,0x2f8aU,0x5f14U,0xbe28U,0x6c5bU,0xd8b6U,
		0xa167U,0x52c5U,0xa58aU,0x5b1fU,0xb63eU,0x7c77U,0xf8eeU,0xe1d7U,0xd3a5U,0xb741U,
		0x7e89U,0xfd12U,0xea2fU,0xc455U,0x98a1U,0x2149U,0x4292U,0x8524U,0x1a43U,0x3486U,
		0x690cU,0xd218U,0xb43bU,0x787dU,0xf0faU,0xf1ffU,0xf3f5U,0xf7e1U,0xffc9U,0xef99U,
		0xcf39U,0x8e79U,0x0cf9U,0x19f2U,0x33e4U,0x67c8U,0xcf90U,0x8f2bU,0x0e5dU,0x1cbaU,
		0x3974U,0x72e8U,0xe5d0U,0xdbabU,0xa75dU,0x5eb1U,0xbd62U,0x6acfU,0xd59eU,0xbb37U,
		0x6665U,0xcccaU,0x899fU,0x0335U,0x066aU,0x0cd4U,0x19a8U,0x3350U,0x66a0U,0xcd40U,
		0x8a8bU,0x051dU,0x0a3aU,0x1474U,0x28e8U,0x51d0U,0xa3a0U,0x574bU,0xae96U,0x4d27U,
		0x9a4eU,0x2497U,0x492eU,0x925cU,0x34b3U,0x6966U,0xd2ccU,0xb593U,0x7b2dU,0xf65aU,
		0xfcbfU,0xe975U,0xc2e1U,0x95c9U,0x3b99U,0x7732U,0xee64U,0xccc3U,0x898dU,0x0311U,
		0x0622U,0x0c44U,0x1888U,0x3110U,0x6220U,0xc440U,0x988bU,0x211dU,0x423aU,0x8474U,
		0x18e3U,0x31c6U,0x638cU,0xc718U,0x9e3bU,0x2c7dU,0x58faU,0xb1f4U,0x73e3U,0xe7c6U,
		0xdf87U,0xaf05U,0x4e01U,0x9c02U,0x280fU,0x501eU,0xa03cU,0x5073U,0xa0e6U,0x51c7U,
		0xa38eU,0x5717U,0xae2eU,0x4c57U,0x98aeU,0x2157U,0x42aeU,0x855cU,0x1ab3U,0x3566U,
		0x6accU,0xd598U,0xbb3bU,0x667dU,0xccfaU,0x89ffU,0x03f5U,0x07eaU,0x0fd4U,0x1fa8U,
		0x3f50U,0x7ea0U,0xfd40U,0xea8bU,0xc51dU,0x9a31U,0x2469U,0x48d2U,0x91a4U,0x3343U,
		0x6686U,0xcd0cU,0x8a13U,0x042dU,0x085aU,0x10b4U,0x2168U,0x42d0U,0x85a0U,0x1b4bU,
		0x3696U,0x6d2cU,0xda58U,0xa4bbU,0x597dU,0xb2faU,0x75ffU,0xebfeU,0xc7f7U,0x9fe5U,
		0x2fc1U,0x5f82U,0xbf04U,0x6e03U,0xdc06U,0xa807U,0x4005U,0x800aU,0x101fU,0x203eU,
		0x407cU,0x80f8U,0x11fbU,0x23f6U,0x47ecU,0x8fd8U,0x0fbbU,0x1f76U,0x3eecU,0x7dd8U,
		0xfbb0U,0xe76bU,0xdeddU,0xadb1U,0x4b69U,0x96d2U,0x3dafU,0x7b5eU,0xf6bcU,0xfd73U,
		0xeaedU,0xc5d1U,0x9ba9U,0x2759U,0x4eb2U,0x9d64U,0x2ac3U,0x5586U,0xab0cU,0x4613U,
		0x8c26U,0x0847U,0x108eU,0x211cU,0x4238U,0x8470U,0x18ebU,0x31d6U,0x63acU,0xc758U,
		0x9ebbU,0x2d7dU,0x5afaU,0xb5f4U,0x7be3U,0xf7c6U,0xff87U,0xef05U,0xce01U,0x8c09U,
		0x0819U,0x1032U,0x2064U,0x40c8U,0x8190U,0x132bU,0x2656U,0x4cacU,0x9958U,0x22bbU,
		0x4576U,0x8aecU,0x05d3U,0x0ba6U,0x174cU,0x2e98U,0x5d30U,0xba60U,0x64cbU,0xc996U,
		0x8327U,0x1645U,0x2c8aU,0x5914U,0xb228U,0x745bU,0xe8b6U,0xc167U,0x92c5U,0x3581U,
		0x6b02U,0xd604U,0xbc03U,0x680dU,0xd01aU,0xb03fU,0x7075U,0xe0eaU,0xd1dfU,0xb3b5U,
		0x7761U,0xeec2U,0xcd8fU,0x8b15U,0x0621U,0x0c42U,0x1884U,0x3108U,0x6210U,0xc420U,
		0x984bU,0x209dU,0x413aU,0x8274U,0x14e3U,0x29c6U,0x538cU,0xa718U,0x5e3bU,0xbc76U,
		0x68e7U,0xd1ceU,0xb397U,0x7725U,0xee4aU,0xcc9fU,0x8935U,0x0261U,0x04c2U,0x0984U,
		0x1308U,0x2610U,0x4c20U,0x9840U,0x208bU,0x4116U,0x822cU,0x1453U,0x28a6U,0x514cU,
		0xa298U,0x553bU,0xaa76U,0x44e7U,0x89ceU,0x0397U,0x072eU,0x0e5cU,0x1cb8U,0x3970U,
		0x72e0U,0xe5c0U,0xdb8bU,0xa71dU,0x5e31U,0xbc62U,0x68cfU,0xd19eU,0xb337U,0x7665U,
		0xeccaU,0xc99fU,0x8335U,0x1661U,0x2cc2U,0x5984U,0xb308U,0x761bU,0xec36U,0xc867U,
		0x80c5U,0x1181U,0x2302U,0x4604U,0x8c08U,0x081bU,0x1036U,0x206cU,0x40d8U,0x81b0U,
		0x136bU,0x26d6U,0x4dacU,0x9b58U,0x26bbU,0x4d76U,0x9aecU,0x25d3U,0x4ba6U,0x974cU,
		0x3e93U,0x7d26U,0xfa4cU,0xe493U,0xd92dU,0xa251U,0x54a9U,0xa952U,0x42afU,0x855eU,
		0x1ab7U,0x356eU,0x6adcU,0xd5b8U,0xbb7bU,0x66fdU,0xcdfaU,0x8bffU,0x07f5U,0x0feaU,
		0x1fd4U,0x3fa8U,0x7f50U,0xfea0U,0xed4bU,0xca9dU,0x8531U,0x1a69U,0x34d2U,0x69a4U,
		0xd348U,0xb69bU,0x7d3dU,0xfa7aU,0xe4ffU,0xd9f5U,0xa3e1U,0x57c9U,0xaf92U,0x4f2fU,
		0x9e5eU,0x2cb7U,0x596eU,0xb2dcU,0x75b3U,0xeb66U,0xc6c7U,0x9d85U,0x2b01U,0x5602U,
		0xac04U,0x4803U,0x9006U,0x3007U,0x600eU,0xc01cU,0x9033U,0x306dU,0x60daU,0xc1b4U,
		0x9363U,0x36cdU,0x6d9aU,0xdb34U,0xa663U,0x5ccdU,0xb99aU,0x633fU,0xc67eU,0x9cf7U,
		0x29e5U,0x53caU,0xa794U,0x5f23U,0xbe46U,0x6c87U,0xd90eU,0xa217U,0x5425U,0xa84aU,
		0x409fU,0x813eU,0x1277U,0x24eeU,0x49dcU,0x93b8U,0x377bU,0x6ef6U,0xddecU,0xabd3U,
		0x47adU,0x8f5aU,0x0ebfU,0x1d7eU,0x3afcU,0x75f8U,0xebf0U,0xc7ebU,0x9fddU,0x2fb1U,
		0x5f62U,0xbec4U,0x6d83U,0xdb06U,0xa607U,0x5c05U,0xb80aU,0x601fU,0xc03eU,0x9077U,
		0x30e5U,0x61caU,0xc394U,0x9723U,0x3e4dU,0x7c9aU,0xf934U,0xe263U,0xd4cdU,0xb991U,
		0x6329U,0xc652U,0x9cafU,0x2955U,0x52aaU,0xa554U,0x5aa3U,0xb546U,0x7a87U,0xf50eU,
		0xfa17U,0xe425U,0xd841U,0xa089U,0x5119U,0xa232U,0x546fU,0xa8deU,0x41b7U,0x836eU,
		0x16d7U,0x2daeU,0x5b5cU,0xb6b8U,0x7d7bU,0xfaf6U,0xe5e7U,0xdbc5U,0xa781U,0x5f09U,
		0xbe12U,0x6c2fU,0xd85eU,0xa0b7U,0x5165U,0xa2caU,0x559fU,0xab3eU,0x4677U,0x8ceeU,
		0x09d7U,0x13aeU,0x275cU,0x4eb8U,0x9d70U,0x2aebU,0x55d6U,0xabacU,0x4753U,0x8ea6U,
		0x0d47U,0x1a8eU,0x351cU,0x6a38U,0xd470U,0xb8ebU,0x61ddU,0xc3baU,0x977fU,0x3ef5U,
		0x7deaU,0xfbd4U,0xe7a3U,0xdf4dU,0xae91U,0x4d29U,0x9a52U,0x24afU,0x495eU,0x92bcU,
		0x3573U,0x6ae6U,0xd5ccU,0xbb93U,0x672dU,0xce5aU,0x8cbfU,0x0975U,0x12eaU,0x25d4U,
		0x4ba8U,0x9750U,0x3eabU,0x7d56U,0xfaacU,0xe553U,0xdaadU,0xa551U,0x5aa9U,0xb552U,
		0x7aafU,0xf55eU,0xfab7U,0xe565U,0xdac1U,0xa589U,0x5b19U,0xb632U,0x7c6fU,0xf8deU,
		0xe1b7U,0xd365U,0xb6c1U,0x7d89U,0xfb12U,0xe62fU,0xdc55U,0xa8a1U,0x4149U,0x8292U,
		0x152fU,0x2a5eU,0x54bcU,0xa978U,0x42fbU,0x85f6U,0x1be7U,0x37ceU,0x6f9cU,0xdf38U,
		0xae7bU,0x4cfdU,0x99faU,0x23ffU,0x47feU,0x8ffcU,0x0ff3U,0x1fe6U,0x3fccU,0x7f98U,
		0xff30U,0xee6bU,0xccddU,0x89b1U,0x0369U,0x06d2U,0x0da4U,0x1b48U,0x3690U,0x6d20U,
		0xda40U,0xa48bU,0x591dU,0xb23aU,0x747fU,0xe8feU,0xc1f7U,0x93e5U,0x37c1U,0x6f82U,
		0xdf04U,0xae03U,0x4c0dU,0x981aU,0x203fU,0x407eU,0x80fcU,0x11f3U,0x23e6U,0x47ccU,
		0x8f98U,0x0f3bU,0x1e76U,0x3cecU,0x79d8U,0xf3b0U,0xf76bU,0xfeddU,0xedb1U,0xcb69U,
		0x86d9U,0x1db9U,0x3b72U,0x76e4U,0xedc8U,0xcb9bU,0x873dU,0x1e71U,0x3ce2U,0x79c4U,
		0xf388U,0xf71bU,0xfe3dU,0xec71U,0xc8e9U,0x81d9U,0x13b9U,0x2772U,0x4ee4U,0x9dc8U,
		0x2b9bU,0x5736U,0xae6cU,0x4cd3U,0x99a6U,0x2347U,0x468eU,0x8d1cU,0x0a33U,0x1466U,
		0x28ccU,0x5198U,0xa330U,0x566bU,0xacd6U,0x49a7U,0x934eU,0x3697U,0x6d2eU,0xda5cU,
		0xa4b3U,0x596dU,0xb2daU,0x75bfU,0xeb7eU,0xc6f7U,0x9de5U,0x2bc1U,0x5782U,0xaf04U,
		0x4e03U,0x9c06U,0x2807U,0x500eU,0xa01cU,0x5033U,0xa066U,0x50c7U,0xa18eU,0x5317U,
		0xa62eU,0x5c57U,0xb8aeU,0x6157U,0xc2aeU,0x9557U,0x3aa5U,0x754aU,0xea94U,0xc523U,
		0x9a4dU,0x2491U,0x4922U,0x9244U,0x3483U,0x6906U,0xd20cU,0xb413U,0x782dU,0xf05aU,
		0xf0bfU,0xf175U,0xf2e1U,0xf5c9U,0xfb99U,0xe739U,0xde79U,0xacf9U,0x49f9U,0x93f2U,
		0x37efU,0x6fdeU,0xdfbcU,0xaf73U,0x4eedU,0x9ddaU,0x2bbfU,0x577eU,0xaefcU,0x4df3U,
		0x9be6U,0x27c7U,0x4f8eU,0x9f1cU,0x2e33U,0x5c66U,0xb8ccU,0x6193U,0xc326U,0x9647U,
		0x3c85U,0x790aU,0xf214U,0xf423U,0xf84dU,0xe091U,0xd129U,0xb259U,0x74b9U,0xe972U,
		0xc2efU,0x95d5U,0x3ba1U,0x7742U,0xee84U,0xcd03U,0x8a0dU,0x0411U,0x0822U,0x1044U,
		0x2088U,0x4110U,0x8220U,0x144bU,0x2896U,0x512cU,0xa258U,0x54bbU,0xa976U,0x42e7U,
		0x85ceU,0x1b97U,0x372eU,0x6e5cU,0xdcb8U,0xa97bU,0x42fdU,0x85faU,0x1bffU,0x37feU,
		0x6ffcU,0xdff8U,0xaffbU,0x4ffdU,0x9ffaU,0x2fffU,0x5ffeU,0xbffcU,0x6ff3U,0xdfe6U,
		0xafc7U,0x4f85U,0x9f0aU,0x2e1fU,0x5c3eU,0xb87cU,0x60f3U,0xc1e6U,0x93c7U,0x3785U,
		0x6f0aU,0xde14U,0xac23U,0x484dU,0x909aU,0x313fU,0x627eU,0xc4fcU,0x99f3U,0x23edU,
		0x47daU,0x8fb4U,0x0f63U,0x1ec6U,0x3d8cU,0x7b18U,0xf630U,0xfc6bU,0xe8ddU,0xc1b1U,
		0x9369U,0x36d9U,0x6db2U,0xdb64U,0xa6c3U,0x5d8dU,0xbb1aU,0x663fU,0xcc7eU,0x88f7U,
		0x01e5U,0x03caU,0x0794U,0x0f28U,0x1e50U,0x3ca0U,0x7940U,0xf280U,0xf50bU,0xfa1dU,
		0xe431U,0xd869U,0xa0d9U,0x51b9U,0xa372U,0x56efU,0xaddeU,0x4bb7U,0x976eU,0x3ed7U,
		0x7daeU,0xfb5cU,0xe6b3U,0xdd6dU,0xaad1U,0x45a9U,0x8b52U,0x06afU,0x0d5eU,0x1abcU,
		0x3578U,0x6af0U,0xd5e0U,0xbbcbU,0x679dU,0xcf3aU,0x8e7fU,0x0cf5U,0x19eaU,0x33d4U,
		0x67a8U,0xcf50U,0x8eabU,0x0d5dU,0x1abaU,0x3574U,0x6ae8U,0xd5d0U,0xbbabU,0x675dU,
		0xcebaU,0x8d7fU,0x0af5U,0x15eaU,0x2bd4U,0x57a8U,0xaf50U,0x4eabU,0x9d56U,0x2aa7U,
		0x554eU,0xaa9cU,0x4533U,0x8a66U,0x04c7U,0x098eU,0x131cU,0x2638U,0x4c70U,0x98e0U,
		0x21cbU,0x4396U,0x872cU,0x1e53U,0x3ca6U,0x794cU,0xf298U,0xf53bU,0xfa7dU,0xe4f1U,
		0xd9e9U,0xa3d9U,0x57b9U,0xaf72U,0x4eefU,0x9ddeU,0x2bb7U,0x576eU,0xaedcU,0x4db3U,
		0x9b66U,0x26c7U,0x4d8eU,0x9b1cU,0x2633U,0x4c66U,0x98ccU,0x2193U,0x4326U,0x864cU,
		0x1c93U,0x3926U,0x724cU,0xe498U,0xd93bU,0xa27dU,0x54f1U,0xa9e2U,0x43cfU,0x879eU,
		0x1f37U,0x3e6eU,0x7cdcU,0xf9b8U,0xe37bU,0xd6fdU,0xbdf1U,0x6be9U,0xd7d2U,0xbfafU,
		0x6f55U,0xdeaaU,0xad5fU,0x4ab5U,0x956aU,0x3adfU,0x75beU,0xeb7cU,0xc6f3U,0x9dedU,
		0x2bd1U,0x57a2U,0xaf44U,0x4e83U,0x9d06U,0x2a07U,0x540eU,0xa81cU,0x4033U,0x8066U,
		0x10c7U,0x218eU,0x431cU,0x8638U,0x1c7bU,0x38f6U,0x71ecU,0xe3d8U,0xd7bbU,0xbf7dU,
		0x6ef1U,0xdde2U,0xabcfU,0x4795U,0x8f2aU,0x0e5fU,0x1cbeU,0x397cU,0x72f8U,0xe5f0U,
		0xdbebU,0xa7ddU,0x5fb1U,0xbf62U,0x6ecfU,0xdd9eU,0xab37U,0x4665U,0x8ccaU,0x099fU,
		0x133eU,0x267cU,0x4cf8U,0x99f0U,0x23ebU,0x47d6U,0x8facU,0x0f53U,0x1ea6U,0x3d4cU,
		0x7a98U,0xf530U,0xfa6bU,0xe4ddU,0xd9b1U,0xa369U,0x56d9U,0xadb2U,0x4b6fU,0x96deU,
		0x3db7U,0x7b6eU,0xf6dcU,0xfdb3U,0xeb6dU,0xc6d1U,0x9da9U,0x2b59U,0x56b2U,0xad64U,
		0x4ac3U,0x9586U,0x3b07U,0x760eU,0xec1cU,0xc833U,0x806dU,0x10d1U,0x21a2U,0x4344U,
		0x8688U,0x1d1bU,0x3a36U,0x746cU,0xe8d8U,0xc1bbU,0x937dU,0x36f1U,0x6de2U,0xdbc4U,
		0xa783U,0x5f0dU,0xbe1aU,0x6c3fU,0xd87eU,0xa0f7U,0x51e5U,0xa3caU,0x579fU,0xaf3eU,
		0x4e77U,0x9ceeU,0x29d7U,0x53aeU,0xa75cU,0x5eb3U,0xbd66U,0x6ac7U,0xd58eU,0xbb17U,
		0x6625U,0xcc4aU,0x889fU,0x0135U,0x026aU,0x04d4U,0x09a8U,0x1350U,0x26a0U,0x4d40U,
		0x9a80U,0x250bU,0x4a16U,0x942cU,0x3853U,0x70a6U,0xe14cU,0xd293U,0xb52dU,0x7a51U,
		0xf4a2U,0xf94fU,0xe295U,0xd521U,0xba49U,0x6499U,0xc932U,0x826fU,0x14d5U,0x29aaU,
		0x5354U,0xa6a8U,0x5d5bU,0xbab6U,0x6567U,0xcaceU,0x8597U,0x1b25U,0x364aU,0x6c94U,
		0xd928U,0xa25bU,0x54bdU,0xa97aU,0x42ffU,0x85feU,0x1bf7U,0x37eeU,0x6fdcU,0xdfb8U,
		0xaf7bU,0x4efdU,0x9dfaU,0x2bffU,0x57feU,0xaffcU,0x4ff3U,0x9fe6U,0x2fc7U,0x5f8eU,
		0xbf1cU,0x6e33U,0xdc66U,0xa8c7U,0x4185U,0x830aU,0x161fU,0x2c3eU,0x587cU,0xb0f8U,
		0x71fbU,0xe3f6U,0xd7e7U,0xbfc5U,0x6f81U,0xdf02U,0xae0fU,0x4c15U,0x982aU,0x205fU,
		0x40beU,0x817cU,0x12f3U,0x25e6U,0x4bccU,0x9798U,0x3f3bU,0x7e76U,0xfcecU,0xe9d3U,
		0xc3adU,0x9751U,0x3ea9U,0x7d52U,0xfaa4U,0xe543U,0xda8dU,0xa511U,0x5a29U,0xb452U,
		0x78afU,0xf15eU,0xf2b7U,0xf565U,0xfac1U,0xe589U,0xdb19U,0xa639U,0x5c79U,0xb8f2U,
		0x61efU,0xc3deU,0x97b7U,0x3f65U,0x7ecaU,0xfd94U,0xeb23U,0xc64dU,0x9c91U,0x2929U,
		0x5252U,0xa4a4U,0x5943U,0xb286U,0x7507U,0xea0eU,0xc417U,0x9825U,0x2041U,0x4082U,
		0x8104U,0x1203U,0x2406U,0x480cU,0x9018U,0x303bU,0x6076U,0xc0ecU,0x91d3U,0x33adU,
		0x675aU,0xceb4U,0x8d63U,0x0acdU,0x159aU,0x2b34U,0x5668U,0xacd0U,0x49abU,0x9356U,
		0x36a7U,0x6d4eU,0xda9cU,0xa533U,0x5a6dU,0xb4daU,0x79bfU,0xf37eU,0xf6f7U,0xfde5U,
		0xebc1U,0xc789U,0x9f19U,0x2e39U,0x5c72U,0xb8e4U,0x61c3U,0xc386U,0x9707U,0x3e05U,
		0x7c0aU,0xf814U,0xe023U,0xd04dU,0xb091U,0x7129U,0xe252U,0xd4afU,0xb955U,0x62a1U,
		0xc542U,0x9a8fU,0x2515U,0x4a2aU,0x9454U,0x38a3U,0x7146U,0xe28cU,0xd513U,0xba2dU,
		0x6451U,0xc8a2U,0x814fU,0x1295U,0x252aU,0x4a54U,0x94a8U,0x395bU,0x72b6U,0xe56cU,
		0xdad3U,0xa5adU,0x5b51U,0xb6a2U,0x7d4fU,0xfa9eU,0xe537U,0xda65U,0xa4c1U,0x5989U,
		0xb312U,0x762fU,0xec5eU,0xc8b7U,0x8165U,0x12c1U,0x2582U,0x4b04U,0x9608U,0x3c1bU,
		0x7836U,0xf06cU,0xf0d3U,0xf1adU,0xf351U,0xf6a9U,0xfd59U,0xeab9U,0xc579U,0x9af9U,
		0x25f9U,0x4bf2U,0x97e4U,0x3fc3U,0x7f86U,0xff0cU,0xee13U,0xcc2dU,0x8851U,0x00a9U,
		0x0152U,0x02a4U,0x0548U,0x0a90U,0x1520U,0x2a40U,0x5480U,0xa900U,0x420bU,0x8416U,
		0x1827U,0x304eU,0x609cU,0xc138U,0x927bU,0x34fdU,0x69faU,0xd3f4U,0xb7e3U,0x7fcdU,
		0xff9aU,0xef3fU,0xce75U,0x8ce1U,0x09c9U,0x1392U,0x2724U,0x4e48U,0x9c90U,0x292bU,
		0x5256U,0xa4acU,0x5953U,0xb2a6U,0x7547U,0xea8eU,0xc517U,0x9a25U,0x2441U,0x4882U,
		0x9104U,0x3203U,0x6406U,0xc80cU,0x8013U,0x102dU,0x205aU,0x40b4U,0x8168U,0x12dbU,
		0x25b6U,0x4b6cU,0x96d8U,0x3dbbU,0x7b76U,0xf6ecU,0xfdd3U,0xebadU,0xc751U,0x9ea9U,
		0x2d59U,0x5ab2U,0xb564U,0x7ac3U,0xf586U,0xfb07U,0xe605U,0xdc01U,0xa809U,0x4019U,
		0x8032U,0x106fU,0x20deU,0x41bcU,0x8378U,0x16fbU,0x2df6U,0x5becU,0xb7d8U,0x7fbbU,
		0xff76U,0xeee7U,0xcdc5U,0x8b81U,0x0709U,0x0e12U,0x1c24U,0x3848U,0x7090U,0xe120U,
		0xd24bU,0xb49dU,0x7931U,0xf262U,0xf4cfU,0xf995U,0xe321U,0xd649U,0xbc99U,0x6939U,
		0xd272U,0xb4efU,0x79d5U,0xf3aaU,0xf75fU,0xfeb5U,0xed61U,0xcac9U,0x8599U,0x1b39U,
		0x3672U,0x6ce4U,0xd9c8U,0xa39bU,0x573dU,0xae7aU,0x4cffU,0x99feU,0x23f7U,0x47eeU,
		0x8fdcU,0x0fb3U,0x1f66U,0x3eccU,0x7d98U,0xfb30U,0xe66bU,0xdcddU,0xa9b1U,0x4369U,
		0x86d2U,0x1dafU,0x3b5eU,0x76bcU,0xed78U,0xcafbU,0x85fdU,0x1bf1U,0x37e2U,0x6fc4U,
		0xdf88U,0xaf1bU,0x4e3dU,0x9c7aU,0x28ffU,0x51feU,0xa3fcU,0x57f3U,0xafe6U,0x4fc7U,
		0x9f8eU,0x2f17U,0x5e2eU,0xbc5cU,0x68b3U,0xd166U,0xb2c7U,0x7585U,0xeb0aU,0xc61fU,
		0x9c35U,0x2861U,0x50c2U,0xa184U,0x5303U,0xa606U,0x5c07U,0xb80eU,0x6017U,0xc02eU,
		0x9057U,0x30a5U,0x614aU,0xc294U,0x9523U,0x3a4dU,0x749aU,0xe934U,0xc263U,0x94cdU,
		0x3991U,0x7322U,0xe644U,0xdc83U,0xa90dU,0x4211U,0x8422U,0x184fU,0x309eU,0x613cU,
		0xc278U,0x94fbU,0x39fdU,0x73faU,0xe7f4U,0xdfe3U,0xafcdU,0x4f91U,0x9f22U,0x2e4fU,
		0x5c9eU,0xb93cU,0x6273U,0xc4e6U,0x99c7U,0x2385U,0x470aU,0x8e14U,0x0c23U,0x1846U,
		0x308cU,0x6118U,0xc230U,0x946bU,0x38ddU,0x71baU,0xe374U,0xd6e3U,0xbdcdU,0x6b91U,
		0xd722U,0xbe4fU,0x6c95U,0xd92aU,0xa25fU,0x54b5U,0xa96aU,0x42dfU,0x85beU,0x1b77U,
		0x36eeU,0x6ddcU,0xdbb8U,0xa77bU,0x5efdU,0xbdfaU,0x6bffU,0xd7feU,0xbff7U,0x6fe5U,
		0xdfcaU,0xaf9fU,0x4f35U,0x9e6aU,0x2cdfU,0x59beU,0xb37cU,0x76f3U,0xede6U,0xcbc7U,
		0x8785U,0x1f01U,0x3e02U,0x7c04U,0xf808U,0xe01bU,0xd03dU,0xb071U,0x70e9U,0xe1d2U,
		0xd3afU,0xb755U,0x7ea1U,0xfd42U,0xea8fU,0xc515U,0x9a21U,0x2449U,0x4892U,0x9124U,
		0x3243U,0x6486U,0xc90cU,0x8213U,0x142dU,0x285aU,0x50b4U,0xa168U,0x52dbU,0xa5b6U,
		0x5b67U,0xb6ceU,0x7d97U,0xfb2eU,0xe657U,0xdca5U,0xa941U,0x4289U,0x8512U,0x1a2fU,
		0x345eU,0x68bcU,0xd178U,0xb2fbU,0x75fdU,0xebfaU,0xc7ffU,0x9ff5U,0x2fe1U,0x5fc2U,
		0xbf84U,0x6f03U,0xde06U,0xac07U,0x4805U,0x900aU,0x301fU,0x603eU,0xc07cU,0x90f3U,
		0x31edU,0x63daU,0xc7b4U,0x9f63U,0x2ecdU,0x5d9aU,0xbb34U,0x6663U,0xccc6U,0x8987U,
		0x0305U,0x060aU,0x0c14U,0x1828U,0x3050U,0x60a0U,0xc140U,0x928bU,0x351dU,0x6a3aU,
		0xd474U,0xb8e3U,0x61cdU,0xc39aU,0x973fU,0x3e75U,0x7ceaU,0xf9d4U,0xe3a3U,0xd74dU,
		0xbe91U,0x6d29U,0xda52U,0xa4afU,0x5955U,0xb2aaU,0x755fU,0xeabeU,0xc577U,0x9ae5U,
		0x25c1U,0x4b82U,0x9704U,0x3e03U,0x7c06U,0xf80cU,0xe013U,0xd02dU,0xb051U,0x70a9U,
		0xe152U,0xd2afU,0xb555U,0x7aa1U,0xf542U,0xfa8fU,0xe515U,0xda21U,0xa449U,0x5899U,
		0xb132U,0x726fU,0xe4deU,0xd9b7U,0xa365U,0x56c1U,0xad82U,0x4b0fU,0x961eU,0x3c37U,
		0x786eU,0xf0dcU,0xf1b3U,0xf36dU,0xf6d1U,0xfda9U,0xeb59U,0xc6b9U,0x9d79U,0x2af9U,
		0x55f2U,0xabe4U,0x47c3U,0x8f86U,0x0f07U,0x1e0eU,0x3c1cU,0x7838U,0xf070U,0xf0ebU,
		0xf1ddU,0xf3b1U,0xf769U,0xfed9U,0xedb9U,0xcb79U,0x86f9U,0x1df9U,0x3bf2U,0x77e4U,
		0xefc8U,0xcf9bU,0x8f3dU,0x0e71U,0x1ce2U,0x39c4U,0x7388U,0xe710U,0xde2bU,0xac5dU,
		0x48b1U,0x9162U,0x32cfU,0x659eU,0xcb3cU,0x8673U,0x1cedU,0x39daU,0x73b4U,0xe768U,
		0xdedbU,0xadbdU,0x4b71U,0x96e2U,0x3dcfU,0x7b9eU,0xf73cU,0xfe73U,0xecedU,0xc9d1U,
		0x83a9U,0x1759U,0x2eb2U,0x5d64U,0xbac8U,0x659bU,0xcb36U,0x8667U,0x1cc5U,0x398aU,
		0x7314U,0xe628U,0xdc5bU,0xa8bdU,0x4171U,0x82e2U,0x15cfU,0x2b9eU,0x573cU,0xae78U,
		0x4cfbU,0x99f6U,0x23e7U,0x47ceU,0x8f9cU,0x0f33U,0x1e66U,0x3cccU,0x7998U,0xf330U,
		0xf66bU,0xfcddU,0xe9b1U,0xc369U,0x96d9U,0x3db9U,0x7b72U,0xf6e4U,0xfdc3U,0xeb8dU,
		0xc711U,0x9e29U,0x2c59U,0x58b2U,0xb164U,0x72c3U,0xe586U,0xdb07U,0xa605U,0x5c01U,
		0xb802U,0x600fU,0xc01eU,0x9037U,0x3065U,0x60caU,0xc194U,0x9323U,0x364dU,0x6c9aU,
		0xd934U,0xa263U,0x54cdU,0xa99aU,0x433fU,0x867eU,0x1cf7U,0x39eeU,0x73dcU,0xe7b8U,
		0xdf7bU,0xaefdU,0x4df1U,0x9be2U,0x27cfU,0x4f9eU,0x9f3cU,0x2e73U,0x5ce6U,0xb9ccU,
		0x6393U,0xc726U,0x9e47U,0x2c85U,0x590aU,0xb214U,0x7423U,0xe846U,0xc087U,0x9105U,
		0x3201U,0x6402U,0xc804U,0x8003U,0x100dU,0x201aU,0x4034U,0x8068U,0x10dbU,0x21b6U,
		0x436cU,0x86d8U,0x1dbbU,0x3b76U,0x76ecU,0xedd8U,0xcbbbU,0x877dU,0x1ef1U,0x3de2U,
		0x7bc4U,0xf788U,0xff1bU,0xee3dU,0xcc71U,0x88e9U,0x01d9U,0x03b2U,0x0764U,0x0ec8U,
		0x1d90U,0x3b20U,0x7640U,0xec80U,0xc90bU,0x821dU,0x1431U,0x2862U,0x50c4U,0xa188U,
		0x531bU,0xa636U,0x5c67U,0xb8ceU,0x6197U,0xc32eU,0x9657U,0x3ca5U,0x794aU,0xf294U,
		0xf523U,0xfa4dU,0xe491U,0xd929U,0xa259U,0x54b9U,0xa972U,0x42efU,0x85deU,0x1bb7U,
		0x376eU,0x6edcU,0xddb8U,0xab7bU,0x46fdU,0x8dfaU,0x0bffU,0x17feU,0x2ffcU,0x5ff8U,
		0xbff0U,0x6febU,0xdfd6U,0xafa7U,0x4f45U,0x9e8aU,0x2d1fU,0x5a3eU,0xb47cU,0x78f3U,
		0xf1e6U,0xf3c7U,0xf785U,0xff01U,0xee09U,0xcc19U,0x8839U,0x0079U,0x00f2U,0x01e4U,
		0x03c8U,0x0790U,0x0f20U,0x1e40U,0x3c80U,0x7900U,0xf200U,0xf40bU,0xf81dU,0xe031U,
		0xd069U,0xb0d9U,0x71b9U,0xe372U,0xd6efU,0xbdd5U,0x6ba1U,0xd742U,0xbe8fU,0x6d15U,
		0xda2aU,0xa45fU,0x58b5U,0xb16aU,0x72dfU,0xe5beU,0xdb77U,0xa6e5U,0x5dc1U,0xbb82U,
		0x670fU,0xce1eU,0x8c37U,0x0865U,0x10caU,0x2194U,0x4328U,0x8650U,0x1cabU,0x3956U,
		0x72acU,0xe558U,0xdabbU,0xa57dU,0x5af1U,0xb5e2U,0x7bcfU,0xf79eU,0xff37U,0xee65U,
		0xccc1U,0x8989U,0x0319U,0x0632U,0x0c64U,0x18c8U,0x3190U,0x6320U,0xc640U,0x9c8bU,
		0x291dU,0x523aU,0xa474U,0x58e3U,0xb1c6U,0x7387U,0xe70eU,0xde17U,0xac25U,0x4841U,
		0x9082U,0x310fU,0x621eU,0xc43cU,0x9873U,0x20edU,0x41daU,0x83b4U,0x1763U,0x2ec6U,
		0x5d8cU,0xbb18U,0x663bU,0xcc76U,0x88e7U,0x01c5U,0x038aU,0x0714U,0x0e28U,0x1c50U,
		0x38a0U,0x7140U,0xe280U,0xd50bU,0xba1dU,0x6431U,0xc862U,0x80cfU,0x1195U,0x232aU,
		0x4654U,0x8ca8U,0x095bU,0x12b6U,0x256cU,0x4ad8U,0x95b0U,0x3b6bU,0x76d6U,0xedacU,
		0xcb53U,0x86adU,0x1d51U,0x3aa2U,0x7544U,0xea88U,0xc51bU,0x9a3dU,0x2471U,0x48e2U,
		0x91c4U,0x3383U,0x6706U,0xce0cU,0x8c13U,0x082dU,0x105aU,0x20b4U,0x4168U,0x82d0U,
		0x15abU,0x2b56U,0x56acU,0xad58U,0x4abbU,0x9576U,0x3ae7U,0x75ceU,0xeb9cU,0xc733U,
		0x9e6dU,0x2cd1U,0x59a2U,0xb344U,0x7683U,0xed06U,0xca07U,0x8405U,0x1801U,0x3002U,
		0x6004U,0xc008U,0x901bU,0x303dU,0x607aU,0xc0f4U,0x91e3U,0x33cdU,0x679aU,0xcf34U,
		0x8e63U,0x0ccdU,0x199aU,0x3334U,0x6668U,0xccd0U,0x89abU,0x035dU,0x06baU,0x0d74U,
		0x1ae8U,0x35d0U,0x6ba0U,0xd740U,0xbe8bU,0x6d1dU,0xda3aU,0xa47fU,0x58f5U,0xb1eaU,
		0x73dfU,0xe7beU,0xdf77U,0xaee5U,0x4dc1U,0x9b82U,0x270fU,0x4e1eU,0x9c3cU,0x2873U,
		0x50e6U,0xa1ccU,0x5393U,0xa726U,0x5e47U,0xbc8eU,0x6917U,0xd22eU,0xb457U,0x78a5U,
		0xf14aU,0xf29fU,0xf535U,0xfa61U,0xe4c9U,0xd999U,0xa339U,0x5679U,0xacf2U,0x49efU,
		0x93deU,0x37b7U,0x6f6eU,0xdedcU,0xadb3U,0x4b6dU,0x96daU,0x3dbfU,0x7b7eU,0xf6fcU,
		0xfdf3U,0xebedU,0xc7d1U,0x9fa9U,0x2f59U,0x5eb2U,0xbd64U,0x6ac3U,0xd586U,0xbb07U,
		0x6605U,0xcc0aU,0x881fU,0x0035U,0x006aU,0x00d4U,0x01a8U,0x0350U,0x06a0U,0x0d40U,
		0x1a80U,0x3500U,0x6a00U,0xd400U,0xb80bU,0x601dU,0xc03aU,0x907fU,0x30f5U,0x61eaU,
		0xc3d4U,0x97a3U,0x3f4dU,0x7e9aU,0xfd34U,0xea63U,0xc4cdU,0x9991U,0x2329U,0x4652U,
		0x8ca4U,0x0943U,0x1286U,0x250cU,0x4a18U,0x9430U,0x386bU,0x70d6U,0xe1acU,0xd353U,
		0xb6adU,0x7d51U,0xfaa2U,0xe54fU,0xda95U,0xa521U,0x5a49U,0xb492U,0x792fU,0xf25eU,
		0xf4b7U,0xf965U,0xe2c1U,0xd589U,0xbb19U,0x6639U,0xcc72U,0x88efU,0x01d5U,0x03aaU,
		0x0754U,0x0ea8U,0x1d50U,0x3aa0U,0x7540U,0xea80U,0xc50bU,0x9a1dU,0x2431U,0x4862U,
		0x90c4U,0x3183U,0x6306U,0xc60cU,0x9c13U,0x282dU,0x505aU,0xa0b4U,0x5163U,0xa2c6U,
		0x5587U,0xab0eU,0x4617U,0x8c2eU,0x0857U,0x10aeU,0x215cU,0x42b8U,0x8570U,0x1aebU,
		0x35d6U,0x6bacU,0xd758U,0xbebbU,0x6d7dU,0xdafaU,0xa5ffU,0x5bf5U,0xb7eaU,0x7fdfU,
		0xffbeU,0xef77U,0xcee5U,0x8dc1U,0x0b89U,0x1712U,0x2e24U,0x5c48U,0xb890U,0x612bU,
		0xc256U,0x94a7U,0x3945U,0x728aU,0xe514U,0xda23U,0xa44dU,0x5891U,0xb122U,0x724fU,
		0xe49eU,0xd937U,0xa265U,0x54c1U,0xa982U,0x430fU,0x861eU,0x1c37U,0x386eU,0x70dcU,
		0xe1b8U,0xd37bU,0xb6fdU,0x7df1U,0xfbe2U,0xe7cfU,0xdf95U,0xaf21U,0x4e49U,0x9c92U,
		0x292fU,0x525eU,0xa4bcU,0x5973U,0xb2e6U,0x75c7U,0xeb8eU,0xc717U,0x9e25U,0x2c41U,
		0x5882U,0xb104U,0x7203U,0xe406U,0xd807U,0xa005U,0x5001U,0xa002U,0x500fU,0xa01eU,
		0x5037U,0xa06eU,0x50d7U,0xa1aeU,0x5357U,0xa6aeU,0x5d57U,0xbaaeU,0x6557U,0xcaaeU,
		0x8557U,0x1aa5U,0x354aU,0x6a94U,0xd528U,0xba5bU,0x64bdU,0xc97aU,0x82ffU,0x15f5U,
		0x2beaU,0x57d4U,0xafa8U,0x4f5bU,0x9eb6U,0x2d67U,0x5aceU,0xb59cU,0x7b33U,0xf666U,
		0xfcc7U,0xe985U,0xc301U,0x9609U,0x3c19U,0x7832U,0xf064U,0xf0c3U,0xf18dU,0xf311U,
		0xf629U,0xfc59U,0xe8b9U,0xc179U,0x92f9U,0x35f9U,0x6bf2U,0xd7e4U,0xbfc3U,0x6f8dU,
		0xdf1aU,0xae3fU,0x4c75U,0x98eaU,0x21dfU,0x43beU,0x877cU,0x1ef3U,0x3de6U,0x7bccU,
		0xf798U,0xff3bU,0xee7dU,0xccf1U,0x89e9U,0x03d9U,0x07b2U,0x0f64U,0x1ec8U,0x3d90U,
		0x7b20U,0xf640U,0xfc8bU,0xe91dU,0xc231U,0x9469U,0x38d9U,0x71b2U,0xe364U,0xd6c3U,
		0xbd8dU,0x6b11U,0xd622U,0xbc4fU,0x6895U,0xd12aU,0xb25fU,0x74b5U,0xe96aU,0xc2dfU,
		0x95b5U,0x3b61U,0x76c2U,0xed84U,0xcb03U,0x860dU,0x1c11U,0x3822U,0x7044U,0xe088U,
		0xd11bU,0xb23dU,0x7471U,0xe8e2U,0xc1cfU,0x9395U,0x3721U,0x6e42U,0xdc84U,0xa903U,
		0x420dU,0x841aU,0x183fU,0x307eU,0x60fcU,0xc1f8U,0x93fbU,0x37fdU,0x6ffaU,0xdff4U,
		0xafe3U,0x4fcdU,0x9f9aU,0x2f3fU,0x5e7eU,0xbcfcU,0x69f3U,0xd3e6U,0xb7c7U,0x7f85U,
		0xff0aU,0xee1fU,0xcc35U,0x8861U,0x00c9U,0x0192U,0x0324U,0x0648U,0x0c90U,0x1920U,
		0x3240U,0x6480U,0xc900U,0x820bU,0x141dU,0x283aU,0x5074U,0xa0e8U,0x51dbU,0xa3b6U,
		0x5767U,0xaeceU,0x4d97U,0x9b2eU,0x2657U,0x4caeU,0x995cU,0x22b3U,0x4566U,0x8accU,
		0x0593U,0x0b26U,0x164cU,0x2c98U,0x5930U,0xb260U,0x74cbU,0xe996U,0xc327U,0x9645U,
		0x3c81U,0x7902U,0xf204U,0xf403U,0xf80dU,0xe011U,0xd029U,0xb059U,0x70b9U,0xe172U,
		0xd2efU,0xb5d5U,0x7ba1U,0xf742U,0xfe8fU,0xed15U,0xca21U,0x8449U,0x1899U,0x3132U,
		0x6264U,0xc4c8U,0x999bU,0x233dU,0x467aU,0x8cf4U,0x09e3U,0x13c6U,0x278cU,0x4f18U,
		0x9e30U,0x2c6bU,0x58d6U,0xb1acU,0x7353U,0xe6a6U,0xdd47U,0xaa85U,0x4501U,0x8a02U,
		0x040fU,0x081eU,0x103cU,0x2078U,0x40f0U,0x81e0U,0x13cbU,0x2796U,0x4f2cU,0x9e58U,
		0x2cbbU,0x5976U,0xb2ecU,0x75d3U,0xeba6U,0xc747U,0x9e85U,0x2d01U,0x5a02U,0xb404U,
		0x7803U,0xf006U,0xf007U,0xf005U,0xf001U,0xf009U,0xf019U,0xf039U,0xf079U,0xf0f9U,
		0xf1f9U,0xf3f9U,0xf7f9U,0xfff9U,0xeff9U,0xcff9U,0x8ff9U,0x0ff9U,0x1ff2U,0x3fe4U,
		0x7fc8U,0xff90U,0xef2bU,0xce5dU,0x8cb1U,0x0969U,0x12d2U,0x25a4U,0x4b48U,0x9690U,
		0x3d2bU,0x7a56U,0xf4acU,0xf953U,0xe2adU,0xd551U,0xbaa9U,0x6559U,0xcab2U,0x856fU,
		0x1ad5U,0x35aaU,0x6b54U,0xd6a8U,0xbd5bU,0x6abdU,0xd57aU,0xbaffU,0x65f5U,0xcbeaU,
		0x87dfU,0x1fb5U,0x3f6aU,0x7ed4U,0xfda8U,0xeb5bU,0xc6bdU,0x9d71U,0x2ae9U,0x55d2U,
		0xaba4U,0x4743U,0x8e86U,0x0d07U,0x1a0eU,0x341cU,0x6838U,0xd070U,0xb0ebU,0x71ddU,
		0xe3baU,0xd77fU,0xbef5U,0x6de1U,0xdbc2U,0xa78fU,0x5f15U,0xbe2aU,0x6c5fU,0xd8beU,
		0xa177U,0x52e5U,0xa5caU,0x5b9fU,0xb73eU,0x7e77U,0xfceeU,0xe9d7U,0xc3a5U,0x9741U,
		0x3e89U,0x7d12U,0xfa24U,0xe443U,0xd88dU,0xa111U,0x5229U,0xa452U,0x58afU,0xb15eU,
		0x72b7U,0xe56eU,0xdad7U,0xa5a5U,0x5b41U,0xb682U,0x7d0fU,0xfa1eU,0xe437U,0xd865U,
		0xa0c1U,0x5189U,0xa312U,0x562fU,0xac5eU,0x48b7U,0x916eU,0x32d7U,0x65aeU,0xcb5cU,
		0x86b3U,0x1d6dU,0x3adaU,0x75b4U,0xeb68U,0xc6dbU,0x9dbdU,0x2b71U,0x56e2U,0xadc4U,
		0x4b83U,0x9706U,0x3e07U,0x7c0eU,0xf81cU,0xe033U,0xd06dU,0xb0d1U,0x71a9U,0xe352U,
		0xd6afU,0xbd55U,0x6aa1U,0xd542U,0xba8fU,0x6515U,0xca2aU,0x845fU,0x18b5U,0x316aU,
		0x62d4U,0xc5a8U,0x9b5bU,0x26bdU,0x4d7aU,0x9af4U,0x25e3U,0x4bc6U,0x978cU,0x3f13U,
		0x7e26U,0xfc4cU,0xe893U,0xc12dU,0x9251U,0x34a9U,0x6952U,0xd2a4U,0xb543U,0x7a8dU,
		0xf51aU,0xfa3fU,0xe475U,0xd8e1U,0xa1c9U,0x5399U,0xa732U,0x5e6fU,0xbcdeU,0x69b7U,
		0xd36eU,0xb6d7U,0x7da5U,0xfb4aU,0xe69fU,0xdd35U,0xaa61U,0x44c9U,0x8992U,0x032fU,
		0x065eU,0x0cbcU,0x1978U,0x32f0U,0x65e0U,0xcbc0U,0x878bU,0x1f1dU,0x3e3aU,0x7c74U,
		0xf8e8U,0xe1dbU,0xd3bdU,0xb771U,0x7ee9U,0xfdd2U,0xebafU,0xc755U,0x9ea1U,0x2d49U,
		0x5a92U,0xb524U,0x7a43U,0xf486U,0xf907U,0xe205U,0xd401U,0xb809U,0x6019U,0xc032U,
		0x906fU,0x30d5U,0x61aaU,0xc354U,0x96a3U,0x3d4dU,0x7a9aU,0xf534U,0xfa63U,0xe4cdU,
		0xd991U,0xa329U,0x5659U,0xacb2U,0x496fU,0x92deU,0x35b7U,0x6b6eU,0xd6dcU,0xbdb3U,
		0x6b6dU,0xd6daU,0xbdbfU,0x6b75U,0xd6eaU,0xbddfU,0x6bb5U,0xd76aU,0xbedfU,0x6db5U,
		0xdb6aU,0xa6dfU,0x5db5U,0xbb6aU,0x66dfU,0xcdbeU,0x8b77U,0x06e5U,0x0dcaU,0x1b94U,
		0x3728U,0x6e50U,0xdca0U,0xa94bU,0x429dU,0x853aU,0x1a7fU,0x34feU,0x69fcU,0xd3f8U,
		0xb7fbU,0x7ffdU,0xfffaU,0xefffU,0xcff5U,0x8fe1U,0x0fc9U,0x1f92U,0x3f24U,0x7e48U,
		0xfc90U,0xe92bU,0xc25dU,0x94b1U,0x3969U,0x72d2U,0xe5a4U,0xdb43U,0xa68dU,0x5d11U,
		0xba22U,0x644fU,0xc89eU,0x8137U,0x1265U,0x24caU,0x4994U,0x9328U,0x365bU,0x6cb6U,
		0xd96cU,0xa2d3U,0x55adU,0xab5aU,0x46bfU,0x8d7eU,0x0af7U,0x15eeU,0x2bdcU,0x57b8U,
		0xaf70U,0x4eebU,0x9dd6U,0x2ba7U,0x574eU,0xae9cU,0x4d33U,0x9a66U,0x24c7U,0x498eU,
		0x931cU,0x3633U,0x6c66U,0xd8ccU,0xa193U,0x532dU,0xa65aU,0x5cbfU,0xb97eU,0x62f7U,
		0xc5eeU,0x9bd7U,0x27a5U,0x4f4aU,0x9e94U,0x2d23U,0x5a46U,0xb48cU,0x7913U,0xf226U,
		0xf447U,0xf885U,0xe101U,0xd209U,0xb419U,0x7839U,0xf072U,0xf0efU,0xf1d5U,0xf3a1U,
		0xf749U,0xfe99U,0xed39U,0xca79U,0x84f9U,0x19f9U,0x33f2U,0x67e4U,0xcfc8U,0x8f9bU,
		0x0f3dU,0x1e7aU,0x3cf4U,0x79e8U,0xf3d0U,0xf7abU,0xff5dU,0xeeb1U,0xcd69U,0x8ad9U,
		0x05b9U,0x0b72U,0x16e4U,0x2dc8U,0x5b90U,0xb720U,0x7e4bU,0xfc96U,0xe927U,0xc245U,
		0x9481U,0x3909U,0x7212U,0xe424U,0xd843U,0xa08dU,0x5111U,0xa222U,0x544fU,0xa89eU,
		0x4137U,0x826eU,0x14d7U,0x29aeU,0x535cU,0xa6b8U,0x5d7bU,0xbaf6U,0x65e7U,0xcbceU,
		0x8797U,0x1f25U,0x3e4aU,0x7c94U,0xf928U,0xe25bU,0xd4bdU,0xb971U,0x62e9U,0xc5d2U,
		0x9bafU,0x2755U,0x4eaaU,0x9d54U,0x2aa3U,0x5546U,0xaa8cU,0x4513U,0x8a26U,0x0447U,
		0x088eU,0x111cU,0x2238U,0x4470U,0x88e0U,0x01cbU,0x0396U,0x072cU,0x0e58U,0x1cb0U,
		0x3960U,0x72c0U,0xe580U,0xdb0bU,0xa61dU,0x5c31U,0xb862U,0x60cfU,0xc19eU,0x9337U,
		0x3665U,0x6ccaU,0xd994U,0xa323U,0x564dU,0xac9aU,0x493fU,0x927eU,0x34f7U,0x69eeU,
		0xd3dcU,0xb7b3U,0x7f6dU,0xfedaU,0xedbfU,0xcb75U,0x86e1U,0x1dc9U,0x3b92U,0x7724U,
		0xee48U,0xcc9bU,0x893dU,0x0271U,0x04e2U,0x09c4U,0x1388U,0x2710U,0x4e20U,0x9c40U,
		0x288bU,0x5116U,0xa22cU,0x5453U,0xa8a6U,0x4147U,0x828eU,0x1517U,0x2a2eU,0x545cU,
		0xa8b8U,0x417bU,0x82f6U,0x15e7U,0x2bceU,0x579cU,0xaf38U,0x4e7bU,0x9cf6U,0x29e7U,
		0x53ceU,0xa79cU,0x5f33U,0xbe66U,0x6cc7U,0xd98eU,0xa317U,0x5625U,0xac4aU,0x489fU,
		0x913eU,0x3277U,0x64eeU,0xc9dcU,0x83b3U,0x176dU,0x2edaU,0x5db4U,0xbb68U,0x66dbU,
		0xcdb6U,0x8b67U,0x06c5U,0x0d8aU,0x1b14U,0x3628U,0x6c50U,0xd8a0U,0xa14bU,0x529dU,
		0xa53aU,0x5a7fU,0xb4feU,0x79f7U,0xf3eeU,0xf7d7U,0xffa5U,0xef41U,0xce89U,0x8d19U,
		0x0a39U,0x1472U,0x28e4U,0x51c8U,0xa390U,0x572bU,0xae56U,0x4ca7U,0x994eU,0x2297U,
		0x452eU,0x8a5cU,0x04b3U,0x0966U,0x12ccU,0x2598U,0x4b30U,0x9660U,0x3ccbU,0x7996U,
		0xf32cU,0xf653U,0xfcadU,0xe951U,0xc2a9U,0x9559U,0x3ab9U,0x7572U,0xeae4U,0xc5c3U,
		0x9b8dU,0x2711U,0x4e22U,0x9c44U,0x2883U,0x5106U,0xa20cU,0x5413U,0xa826U,0x4047U,
		0x808eU,0x1117U,0x222eU,0x445cU,0x88b8U,0x017bU,0x02f6U,0x05ecU,0x0bd8U,0x17b0U,
		0x2f60U,0x5ec0U,0xbd80U,0x6b0bU,0xd616U,0xbc27U,0x6845U,0xd08aU,0xb11fU,0x7235U,
		0xe46aU,0xd8dfU,0xa1b5U,0x5361U,0xa6c2U,0x5d8fU,0xbb1eU,0x6637U,0xcc6eU,0x88d7U,
		0x01a5U,0x034aU,0x0694U,0x0d28U,0x1a50U,0x34a0U,0x6940U,0xd280U,0xb50bU,0x7a1dU,
		0xf43aU,0xf87fU,0xe0f5U,0xd1e1U,0xb3c9U,0x7799U,0xef32U,0xce6fU,0x8cd5U,0x09a1U,
		0x1342U,0x2684U,0x4d08U,0x9a10U,0x242bU,0x4856U,0x90acU,0x3153U,0x62a6U,0xc54cU,
		0x9a93U,0x252dU,0x4a5aU,0x94b4U,0x3963U,0x72c6U,0xe58cU,0xdb13U,0xa62dU,0x5c51U,
		0xb8a2U,0x614fU,0xc29eU,0x9537U,0x3a65U,0x74caU,0xe994U,0xc323U,0x964dU,0x3c91U,
		0x7922U,0xf244U,0xf483U,0xf90dU,0xe211U,0xd429U,0xb859U,0x60b9U,0xc172U,0x92efU,
		0x35d5U,0x6baaU,0xd754U,0xbea3U,0x6d4dU,0xda9aU,0xa53fU,0x5a75U,0xb4eaU,0x79dfU,
		0xf3beU,0xf777U,0xfee5U,0xedc1U,0xcb89U,0x8719U,0x1e39U,0x3c72U,0x78e4U,0xf1c8U,
		0xf39bU,0xf73dU,0xfe71U,0xece9U,0xc9d9U,0x83b9U,0x1779U,0x2ef2U,0x5de4U,0xbbc8U,
		0x679bU,0xcf36U,0x8e67U,0x0cc5U,0x198aU,0x3314U,0x6628U,0xcc50U,0x88abU,0x015dU,
		0x02baU,0x0574U,0x0ae8U,0x15d0U,0x2ba0U,0x5740U,0xae80U,0x4d0bU,0x9a16U,0x2427U,
		0x484eU,0x909cU,0x3133U,0x6266U,0xc4ccU,0x9993U,0x232dU,0x465aU,0x8cb4U,0x0963U,
		0x12c6U,0x258cU,0x4b18U,0x9630U,0x3c6bU,0x78d6U,0xf1acU,0xf353U,0xf6adU,0xfd51U,
		0xeaa9U,0xc559U,0x9ab9U,0x2579U,0x4af2U,0x95e4U,0x3bc3U,0x7786U,0xef0cU,0xce13U,
		0x8c2dU,0x0851U,0x10a2U,0x2144U,0x4288U,0x8510U,0x1a2bU,0x3456U,0x68acU,0xd158U,
		0xb2bbU,0x757dU,0xeafaU,0xc5ffU,0x9bf5U,0x27e1U,0x4fc2U,0x9f84U,0x2f03U,0x5e06U,
		0xbc0cU,0x6813U,0xd026U,0xb047U,0x7085U,0xe10aU,0xd21fU,0xb435U,0x7861U,0xf0c2U,
		0xf18fU,0xf315U,0xf621U,0xfc49U,0xe899U,0xc139U,0x9279U,0x34f9U,0x69f2U,0xd3e4U,
		0xb7c3U,0x7f8dU,0xff1aU,0xee3fU,0xcc75U,0x88e1U,0x01c9U,0x0392U,0x0724U,0x0e48U,
		0x1c90U,0x3920U,0x7240U,0xe480U,0xd90bU,0xa21dU,0x5431U,0xa862U,0x40cfU,0x819eU,
		0x1337U,0x266eU,0x4cdcU,0x99b8U,0x237bU,0x46f6U,0x8decU,0x0bd3U,0x17a6U,0x2f4cU,
		0x5e98U,0xbd30U,0x6a6bU,0xd4d6U,0xb9a7U,0x6345U,0xc68aU,0x9d1fU,0x2a35U,0x546aU,
		0xa8d4U,0x41a3U,0x8346U,0x1687U,0x2d0eU,0x5a1cU,0xb438U,0x787bU,0xf0f6U,0xf1e7U,
		0xf3c5U,0xf781U,0xff09U,0xee19U,0xcc39U,0x8879U,0x00f9U,0x01f2U,0x03e4U,0x07c8U,
		0x0f90U,0x1f20U,0x3e40U,0x7c80U,0xf900U,0xe20bU,0xd41dU,0xb831U,0x6069U,0xc0d2U,
		0x91afU,0x3355U,0x66aaU,0xcd54U,0x8aa3U,0x054dU,0x0a9aU,0x1534U,0x2a68U,0x54d0U,
		0xa9a0U,0x434bU,0x8696U,0x1d27U,0x3a4eU,0x749cU,0xe938U,0xc27bU,0x94fdU,0x39f1U,
		0x73e2U,0xe7c4U,0xdf83U,0xaf0dU,0x4e11U,0x9c22U,0x284fU,0x509eU,0xa13cU,0x5273U,
		0xa4e6U,0x59c7U,0xb38eU,0x7717U,0xee2eU,0xcc57U,0x88a5U,0x0141U,0x0282U,0x0504U,
		0x0a08U,0x1410U,0x2820U,0x5040U,0xa080U,0x510bU,0xa216U,0x5427U,0xa84eU,0x4097U,
		0x812eU,0x1257U,0x24aeU,0x495cU,0x92b8U,0x357bU,0x6af6U,0xd5ecU,0xbbd3U,0x67adU,
		0xcf5aU,0x8ebfU,0x0d75U,0x1aeaU,0x35d4U,0x6ba8U,0xd750U,0xbeabU,0x6d5dU,0xdabaU,
		0xa57fU,0x5af5U,0xb5eaU,0x7bdfU,0xf7beU,0xff77U,0xeee5U,0xcdc1U,0x8b89U,0x0719U,
		0x0e32U,0x1c64U,0x38c8U,0x7190U,0xe320U,0xd64bU,0xbc9dU,0x6931U,0xd262U,0xb4cfU,
		0x7995U,0xf32aU,0xf65fU,0xfcb5U,0xe961U,0xc2c9U,0x9599U,0x3b39U,0x7672U,0xece4U,
		0xc9c3U,0x838dU,0x1711U,0x2e22U,0x5c44U,0xb888U,0x611bU,0xc236U,0x9467U,0x38c5U,
		0x718aU,0xe314U,0xd623U,0xbc4dU,0x6891U,0xd122U,0xb24fU,0x7495U,0xe92aU,0xc25fU,
		0x94b5U,0x3961U,0x72c2U,0xe584U,0xdb03U,0xa60dU,0x5c11U,0xb822U,0x604fU,0xc09eU,
		0x9137U,0x3265U,0x64caU,0xc994U,0x8323U,0x164dU,0x2c9aU,0x5934U,0xb268U,0x74dbU,
		0xe9b6U,0xc367U,0x96c5U,0x3d81U,0x7b02U,0xf604U,0xfc03U,0xe80dU,0xc011U,0x9029U,
		0x3059U,0x60b2U,0xc164U,0x92c3U,0x358dU,0x6b1aU,0xd634U,0xbc63U,0x68cdU,0xd19aU,
		0xb33fU,0x7675U,0xeceaU,0xc9dfU,0x83b5U,0x1761U,0x2ec2U,0x5d84U,0xbb08U,0x661bU,
		0xcc36U,0x8867U,0x00c5U,0x018aU,0x0314U,0x0628U,0x0c50U,0x18a0U,0x3140U,0x6280U,
		0xc500U,0x9a0bU,0x241dU,0x483aU,0x9074U,0x30e3U,0x61c6U,0xc38cU,0x9713U,0x3e2dU,
		0x7c5aU,0xf8b4U,0xe163U,0xd2cdU,0xb591U,0x7b29U,0xf652U,0xfcafU,0xe955U,0xc2a1U,
		0x9549U,0x3a99U,0x7532U,0xea64U,0xc4c3U,0x998dU,0x2311U,0x4622U,0x8c44U,0x0883U,
		0x1106U,0x220cU,0x4418U,0x8830U,0x006bU,0x00d6U,0x01acU,0x0358U,0x06b0U,0x0d60U,
		0x1ac0U,0x3580U,0x6b00U,0xd600U,0xbc0bU,0x681dU,0xd03aU,0xb07fU,0x70f5U,0xe1eaU,
		0xd3dfU,0xb7b5U,0x7f61U,0xfec2U,0xed8fU,0xcb15U,0x8621U,0x1c49U,0x3892U,0x7124U,
		0xe248U,0xd49bU,0xb93dU,0x6271U,0xc4e2U,0x99cfU,0x2395U,0x472aU,0x8e54U,0x0ca3U,
		0x1946U,0x328cU,0x6518U,0xca30U,0x846bU,0x18ddU,0x31baU,0x6374U,0xc6e8U,0x9ddbU,
		0x2bbdU,0x577aU,0xaef4U,0x4de3U,0x9bc6U,0x2787U,0x4f0eU,0x9e1cU,0x2c33U,0x5866U,
		0xb0ccU,0x7193U,0xe326U,0xd647U,0xbc85U,0x6901U,0xd202U,0xb40fU,0x7815U,0xf02aU,
		0xf05fU,0xf0b5U,0xf161U,0xf2c9U,0xf599U,0xfb39U,0xe679U,0xdcf9U,0xa9f9U,0x43f9U,
		0x87f2U,0x1fefU,0x3fdeU,0x7fbcU,0xff78U,0xeefbU,0xcdfdU,0x8bf1U,0x07e9U,0x0fd2U,
		0x1fa4U,0x3f48U,0x7e90U,0xfd20U,0xea4bU,0xc49dU,0x9931U,0x2269U,0x44d2U,0x89a4U,
		0x0343U,0x0686U,0x0d0cU,0x1a18U,0x3430U,0x6860U,0xd0c0U,0xb18bU,0x731dU,0xe63aU,
		0xdc7fU,0xa8f5U,0x41e1U,0x83c2U,0x178fU,0x2f1eU,0x5e3cU,0xbc78U,0x68fbU,0xd1f6U,
		0xb3e7U,0x77c5U,0xef8aU,0xcf1fU,0x8e35U,0x0c61U,0x18c2U,0x3184U,0x6308U,0xc610U,
		0x9c2bU,0x285dU,0x50baU,0xa174U,0x52e3U,0xa5c6U,0x5b87U,0xb70eU,0x7e17U,0xfc2eU,
		0xe857U,0xc0a5U,0x9141U,0x3289U,0x6512U,0xca24U,0x8443U,0x188dU,0x311aU,0x6234U,
		0xc468U,0x98dbU,0x21bdU,0x437aU,0x86f4U,0x1de3U,0x3bc6U,0x778cU,0xef18U,0xce3bU,
		0x8c7dU,0x08f1U,0x11e2U,0x23c4U,0x4788U,0x8f10U,0x0e2bU,0x1c56U,0x38acU,0x7158U,
		0xe2b0U,0xd56bU,0xbaddU,0x65b1U,0xcb62U,0x86cfU,0x1d95U,0x3b2aU,0x7654U,0xeca8U,
		0xc95bU,0x82bdU,0x1571U,0x2ae2U,0x55c4U,0xab88U,0x471bU,0x8e36U,0x0c67U,0x18ceU,
		0x319cU,0x6338U,0xc670U,0x9cebU,0x29ddU,0x53baU,0xa774U,0x5ee3U,0xbdc6U,0x6b87U,
		0xd70eU,0xbe17U,0x6c25U,0xd84aU,0xa09fU,0x5135U,0xa26aU,0x54dfU,0xa9beU,0x4377U,
		0x86eeU,0x1dd7U,0x3baeU,0x775cU,0xeeb8U,0xcd7bU,0x8afdU,0x05f1U,0x0be2U,0x17c4U,
		0x2f88U,0x5f10U,0xbe20U,0x6c4bU,0xd896U,0xa127U,0x5245U,0xa48aU,0x591fU,0xb23eU,
		0x7477U,0xe8eeU,0xc1d7U,0x93a5U,0x3741U,0x6e82U,0xdd04U,0xaa03U,0x440dU,0x881aU,
		0x003fU,0x007eU,0x00fcU,0x01f8U,0x03f0U,0x07e0U,0x0fc0U,0x1f80U,0x3f00U,0x7e00U,
		0xfc00U,0xe80bU,0xc01dU,0x9031U,0x3069U,0x60d2U,0xc1a4U,0x9343U,0x368dU,0x6d1aU,
		0xda34U,0xa463U,0x58cdU,0xb19aU,0x733fU,0xe67eU,0xdcf7U,0xa9e5U,0x43c1U,0x8782U,
		0x1f0fU,0x3e1eU,0x7c3cU,0xf878U,0xe0fbU,0xd1fdU,0xb3f1U,0x77e9U,0xefd2U,0xcfafU,
		0x8f55U,0x0ea1U,0x1d42U,0x3a84U,0x7508U,0xea10U,0xc42bU,0x985dU,0x20b1U,0x4162U,
		0x82c4U,0x1583U,0x2b06U,0x560cU,0xac18U,0x483bU,0x9076U,0x30e7U,0x61ceU,0xc39cU,
		0x9733U,0x3e6dU,0x7cdaU,0xf9b4U,0xe363U,0xd6cdU,0xbd91U,0x6b29U,0xd652U,0xbcafU,
		0x6955U,0xd2aaU,0xb55fU,0x7ab5U,0xf56aU,0xfadfU,0xe5b5U,0xdb61U,0xa6c9U,0x5d99U,
		0xbb32U,0x666fU,0xccdeU,0x89b7U,0x0365U,0x06caU,0x0d94U,0x1b28U,0x3650U,0x6ca0U,
		0xd940U,0xa28bU,0x551dU,0xaa3aU,0x447fU,0x88feU,0x01f7U,0x03eeU,0x07dcU,0x0fb8U,
		0x1f70U,0x3ee0U,0x7dc0U,0xfb80U,0xe70bU,0xde1dU,0xac31U,0x4869U,0x90d2U,0x31afU,
		0x635eU,0xc6bcU,0x9d73U,0x2aedU,0x55daU,0xabb4U,0x4763U,0x8ec6U,0x0d87U,0x1b0eU,
		0x361cU,0x6c38U,0xd870U,0xa0ebU,0x51ddU,0xa3baU,0x577fU,0xaefeU,0x4df7U,0x9beeU,
		0x27d7U,0x4faeU,0x9f5cU,0x2eb3U,0x5d66U,0xbaccU,0x6593U,0xcb26U,0x8647U,0x1c85U,
		0x390aU,0x7214U,0xe428U,0xd85bU,0xa0bdU,0x5171U,0xa2e2U,0x55cfU,0xab9eU,0x4737U,
		0x8e6eU,0x0cd7U,0x19aeU,0x335cU,0x66b8U,0xcd70U,0x8aebU,0x05ddU,0x0bbaU,0x1774U,
		0x2ee8U,0x5dd0U,0xbba0U,0x674bU,0xce96U,0x8d27U,0x0a45U,0x148aU,0x2914U,0x5228U,
		0xa450U,0x58abU,0xb156U,0x72a7U,0xe54eU,0xda97U,0xa525U,0x5a41U,0xb482U,0x790fU,
		0xf21eU,0xf437U,0xf865U,0xe0c1U,0xd189U,0xb319U,0x7639U,0xec72U,0xc8efU,0x81d5U,
		0x13a1U,0x2742U,0x4e84U,0x9d08U,0x2a1bU,0x5436U,0xa86cU,0x40d3U,0x81a6U,0x1347U,
		0x268eU,0x4d1cU,0x9a38U,0x247bU,0x48f6U,0x91ecU,0x33d3U,0x67a6U,0xcf4cU,0x8e93U,
		0x0d2dU,0x1a5aU,0x34b4U,0x6968U,0xd2d0U,0xb5abU,0x7b5dU,0xf6baU,0xfd7fU,0xeaf5U,
		0xc5e1U,0x9bc9U,0x2799U,0x4f32U,0x9e64U,0x2cc3U,0x5986U,0xb30cU,0x7613U,0xec26U,
		0xc847U,0x8085U,0x1101U,0x2202U,0x4404U,0x8808U,0x001bU,0x0036U,0x006cU,0x00d8U,
		0x01b0U,0x0360U,0x06c0U,0x0d80U,0x1b00U,0x3600U,0x6c00U,0xd800U,0xa00bU,0x501dU,
		0xa03aU,0x507fU,0xa0feU,0x51f7U,0xa3eeU,0x57d7U,0xafaeU,0x4f57U,0x9eaeU,0x2d57U,
		0x5aaeU,0xb55cU,0x7ab3U,0xf566U,0xfac7U,0xe585U,0xdb01U,0xa609U,0x5c19U,0xb832U,
		0x606fU,0xc0deU,0x91b7U,0x3365U,0x66caU,0xcd94U,0x8b23U,0x064dU,0x0c9aU,0x1934U,
		0x3268U,0x64d0U,0xc9a0U,0x834bU,0x169dU,0x2d3aU,0x5a74U,0xb4e8U,0x79dbU,0xf3b6U,
		0xf767U,0xfec5U,0xed81U,0xcb09U,0x8619U,0x1c39U,0x3872U,0x70e4U,0xe1c8U,0xd39bU,
		0xb73dU,0x7e71U,0xfce2U,0xe9cfU,0xc395U,0x9721U,0x3e49U,0x7c92U,0xf924U,0xe243U,
		0xd48dU,0xb911U,0x6229U,0xc452U,0x98afU,0x2155U,0x42aaU,0x8554U,0x1aa3U,0x3546U,
		0x6a8cU,0xd518U,0xba3bU,0x647dU,0xc8faU,0x81ffU,0x13f5U,0x27eaU,0x4fd4U,0x9fa8U,
		0x2f5bU,0x5eb6U,0xbd6cU,0x6ad3U,0xd5a6U,0xbb47U,0x6685U,0xcd0aU,0x8a1fU,0x0435U,
		0x086aU,0x10d4U,0x21a8U,0x4350U,0x86a0U,0x1d4bU,0x3a96U,0x752cU,0xea58U,0xc4bbU,
		0x997dU,0x22f1U,0x45e2U,0x8bc4U,0x0783U,0x0f06U,0x1e0cU,0x3c18U,0x7830U,0xf060U,
		0xf0cbU,0xf19dU,0xf331U,0xf669U,0xfcd9U,0xe9b9U,0xc379U,0x96f9U,0x3df9U,0x7bf2U,
		0xf7e4U,0xffc3U,0xef8dU,0xcf11U,0x8e29U,0x0c59U,0x18b2U,0x3164U,0x62c8U,0xc590U,
		0x9b2bU,0x265dU,0x4cbaU,0x9974U,0x22e3U,0x45c6U,0x8b8cU,0x0713U,0x0e26U,0x1c4cU,
		0x3898U,0x7130U,0xe260U,0xd4cbU,0xb99dU,0x6331U,0xc662U,0x9ccfU,0x2995U,0x532aU,
		0xa654U,0x5ca3U,0xb946U,0x6287U,0xc50eU,0x9a17U,0x2425U,0x484aU,0x9094U,0x3123U,
		0x6246U,0xc48cU,0x9913U,0x222dU,0x445aU,0x88b4U,0x0163U,0x02c6U,0x058cU,0x0b18U,
		0x1630U,0x2c60U,0x58c0U,0xb180U,0x730bU,0xe616U,0xdc27U,0xa845U,0x4081U,0x8102U,
		0x120fU,0x241eU,0x483cU,0x9078U,0x30fbU,0x61f6U,0xc3ecU,0x97d3U,0x3fadU,0x7f5aU,
		0xfeb4U,0xed63U,0xcacdU,0x8591U,0x1b29U,0x3652U,0x6ca4U,0xd948U,0xa29bU,0x553dU,
		0xaa7aU,0x44ffU,0x89feU,0x03f7U,0x07eeU,0x0fdcU,0x1fb8U,0x3f70U,0x7ee0U,0xfdc0U,
		0xeb8bU,0xc71dU,0x9e31U,0x2c69U,0x58d2U,0xb1a4U,0x7343U,0xe686U,0xdd07U,0xaa05U,
		0x4401U,0x8802U,0x000fU,0x001eU,0x003cU,0x0078U,0x00f0U,0x01e0U,0x03c0U,0x0780U,
		0x0f00U,0x1e00U,0x3c00U,0x7800U,0xf000U,0xf00bU,0xf01dU,0xf031U,0xf069U,0xf0d9U,
		0xf1b9U,0xf379U,0xf6f9U,0xfdf9U,0xebf9U,0xc7f9U,0x9ff9U,0x2ff9U,0x5ff2U,0xbfe4U,
		0x6fc3U,0xdf86U,0xaf07U,0x4e05U,0x9c0aU,0x281fU,0x503eU,0xa07cU,0x50f3U,0xa1e6U,
		0x53c7U,0xa78eU,0x5f17U,0xbe2eU,0x6c57U,0xd8aeU,0xa157U,0x52a5U,0xa54aU,0x5a9fU,
		0xb53eU,0x7a77U,0xf4eeU,0xf9d7U,0xe3a5U,0xd741U,0xbe89U,0x6d19U,0xda32U,0xa46fU,
		0x58d5U,0xb1aaU,0x735fU,0xe6beU,0xdd77U,0xaae5U,0x45c1U,0x8b82U,0x070fU,0x0e1eU,
		0x1c3cU,0x3878U,0x70f0U,0xe1e0U,0xd3cbU,0xb79dU,0x7f31U,0xfe62U,0xeccfU,0xc995U,
		0x8321U,0x1649U,0x2c92U,0x5924U,0xb248U,0x749bU,0xe936U,0xc267U,0x94c5U,0x3981U,
		0x7302U,0xe604U,0xdc03U,0xa80dU,0x4011U,0x8022U,0x104fU,0x209eU,0x413cU,0x8278U,
		0x14fbU,0x29f6U,0x53ecU,0xa7d8U,0x5fbbU,0xbf76U,0x6ee7U,0xddceU,0xab97U,0x4725U,
		0x8e4aU,0x0c9fU,0x193eU,0x327cU,0x64f8U,0xc9f0U,0x83ebU,0x17ddU,0x2fbaU,0x5f74U,
		0xbee8U,0x6ddbU,0xdbb6U,0xa767U,0x5ec5U,0xbd8aU,0x6b1fU,0xd63eU,0xbc77U,0x68e5U,
		0xd1caU,0xb39fU,0x7735U,0xee6aU,0xccdfU,0x89b5U,0x0361U,0x06c2U,0x0d84U,0x1b08U,
		0x3610U,0x6c20U,0xd840U,0xa08bU,0x511dU,0xa23aU,0x547fU,0xa8feU,0x41f7U,0x83eeU,
		0x17d7U,0x2faeU,0x5f5cU,0xbeb8U,0x6d7bU,0xdaf6U,0xa5e7U,0x5bc5U,0xb78aU,0x7f1fU,
		0xfe3eU,0xec77U,0xc8e5U,0x81c1U,0x1389U,0x2712U,0x4e24U,0x9c48U,0x289bU,0x5136U,
		0xa26cU,0x54d3U,0xa9a6U,0x4347U,0x868eU,0x1d17U,0x3a2eU,0x745cU,0xe8b8U,0xc17bU,
		0x92fdU,0x35f1U,0x6be2U,0xd7c4U,0xbf83U,0x6f0dU,0xde1aU,0xac3fU,0x4875U,0x90eaU,
		0x31dfU,0x63beU,0xc77cU,0x9ef3U,0x2dedU,0x5bdaU,0xb7b4U,0x7f63U,0xfec6U,0xed87U,
		0xcb05U,0x8601U,0x1c09U,0x3812U,0x7024U,0xe048U,0xd09bU,0xb13dU,0x7271U,0xe4e2U,
		0xd9cfU,0xa395U,0x5721U,0xae42U,0x4c8fU,0x991eU,0x2237U,0x446eU,0x88dcU,0x01b3U,
		0x0366U,0x06ccU,0x0d98U,0x1b30U,0x3660U,0x6cc0U,0xd980U,0xa30bU,0x561dU,0xac3aU,
		0x487fU,0x90feU,0x31f7U,0x63eeU,0xc7dcU,0x9fb3U,0x2f6dU,0x5edaU,0xbdb4U,0x6b63U,
		0xd6c6U,0xbd87U,0x6b05U,0xd60aU,0xbc1fU,0x6835U,0xd06aU,0xb0dfU,0x71b5U,0xe36aU,
		0xd6dfU,0xbdb5U,0x6b61U,0xd6c2U,0xbd8fU,0x6b15U,0xd62aU,0xbc5fU,0x68b5U,0xd16aU,
		0xb2dfU,0x75b5U,0xeb6aU,0xc6dfU,0x9db5U,0x2b61U,0x56c2U,0xad84U,0x4b03U,0x9606U,
		0x3c07U,0x780eU,0xf01cU,0xf033U,0xf06dU,0xf0d1U,0xf1a9U,0xf359U,0xf6b9U,0xfd79U,
		0xeaf9U,0xc5f9U,0x9bf9U,0x27f9U,0x4ff2U,0x9fe4U,0x2fc3U,0x5f86U,0xbf0cU,0x6e13U,
		0xdc26U,0xa847U,0x4085U,0x810aU,0x121fU,0x243eU,0x487cU,0x90f8U,0x31fbU,0x63f6U,
		0xc7ecU,0x9fd3U,0x2fadU,0x5f5aU,0xbeb4U,0x6d63U,0xdac6U,0xa587U,0x5b05U,0xb60aU,
		0x7c1fU,0xf83eU,0xe077U,0xd0e5U,0xb1c1U,0x7389U,0xe712U,0xde2fU,0xac55U,0x48a1U,
		0x9142U,0x328fU,0x651eU,0xca3cU,0x8473U,0x18edU,0x31daU,0x63b4U,0xc768U,0x9edbU,
		0x2dbdU,0x5b7aU,0xb6f4U,0x7de3U,0xfbc6U,0xe787U,0xdf05U,0xae01U,0x4c09U,0x9812U,
		0x202fU,0x405eU,0x80bcU,0x1173U,0x22e6U,0x45ccU,0x8b98U,0x073bU,0x0e76U,0x1cecU,
		0x39d8U,0x73b0U,0xe760U,0xdecbU,0xad9dU,0x4b31U,0x9662U,0x3ccfU,0x799eU,0xf33cU,
		0xf673U,0xfcedU,0xe9d1U,0xc3a9U,0x9759U,0x3eb9U,0x7d72U,0xfae4U,0xe5c3U,0xdb8dU,
		0xa711U,0x5e29U,0xbc52U,0x68afU,0xd15eU,0xb2b7U,0x7565U,0xeacaU,0xc59fU,0x9b35U,
		0x2661U,0x4cc2U,0x9984U,0x2303U,0x4606U,0x8c0cU,0x0813U,0x1026U,0x204cU,0x4098U,
		0x8130U,0x126bU,0x24d6U,0x49acU,0x9358U,0x36bbU,0x6d76U,0xdaecU,0xa5d3U,0x5badU,
		0xb75aU,0x7ebfU,0xfd7eU,0xeaf7U,0xc5e5U,0x9bc1U,0x2789U,0x4f12U,0x9e24U,0x2c43U,
		0x5886U,0xb10cU,0x7213U,0xe426U,0xd847U,0xa085U,0x5101U,0xa202U,0x540fU,0xa81eU,
		0x4037U,0x806eU,0x10d7U,0x21aeU,0x435cU,0x86b8U,0x1d7bU,0x3af6U,0x75ecU,0xebd8U,
		0xc7bbU,0x9f7dU,0x2ef1U,0x5de2U,0xbbc4U,0x6783U,0xcf06U,0x8e07U,0x0c05U,0x180aU,
		0x3014U,0x6028U,0xc050U,0x90abU,0x315dU,0x62baU,0xc574U,0x9ae3U,0x25cdU,0x4b9aU,
		0x9734U,0x3e63U,0x7cc6U,0xf98cU,0xe313U,0xd62dU,0xbc51U,0x68a9U,0xd152U,0xb2afU,
		0x7555U,0xeaaaU,0xc55fU,0x9ab5U,0x2561U,0x4ac2U,0x9584U,0x3b03U,0x7606U,0xec0cU,
		0xc813U,0x802dU,0x1051U,0x20a2U,0x4144U,0x8288U,0x151bU,0x2a36U,0x546cU,0xa8d8U,
		0x41bbU,0x8376U,0x16e7U,0x2dceU,0x5b9cU,0xb738U,0x7e7bU,0xfcf6U,0xe9e7U,0xc3c5U,
		0x9781U,0x3f09U,0x7e12U,0xfc24U,0xe843U,0xc08dU,0x9111U,0x3229U,0x6452U,0xc8a4U,
		0x8143U,0x128dU,0x251aU,0x4a34U,0x9468U,0x38dbU,0x71b6U,0xe36cU,0xd6d3U,0xbdadU,
		0x6b51U,0xd6a2U,0xbd4fU,0x6a95U,0xd52aU,0xba5fU,0x64b5U,0xc96aU,0x82dfU,0x15b5U,
		0x2b6aU,0x56d4U,0xada8U,0x4b5bU,0x96b6U,0x3d67U,0x7aceU,0xf59cU,0xfb33U,0xe66dU,
		0xdcd1U,0xa9a9U,0x4359U,0x86b2U,0x1d6fU,0x3adeU,0x75bcU,0xeb78U,0xc6fbU,0x9dfdU,
		0x2bf1U,0x57e2U,0xafc4U,0x4f83U,0x9f06U,0x2e07U,0x5c0eU,0xb81cU,0x6033U,0xc066U,
		0x90c7U,0x3185U,0x630aU,0xc614U,0x9c23U,0x284dU,0x509aU,0xa134U,0x5263U,0xa4c6U,
		0x5987U,0xb30eU,0x7617U,0xec2eU,0xc857U,0x80a5U,0x1141U,0x2282U,0x4504U,0x8a08U,
		0x041bU,0x0836U,0x106cU,0x20d8U,0x41b0U,0x8360U,0x16cbU,0x2d96U,0x5b2cU,0xb658U,
		0x7cbbU,0xf976U,0xe2e7U,0xd5c5U,0xbb81U,0x6709U,0xce12U,0x8c2fU,0x0855U,0x10aaU,
		0x2154U,0x42a8U,0x8550U,0x1aabU,0x3556U,0x6aacU,0xd558U,0xbabbU,0x657dU,0xcafaU,
		0x85ffU,0x1bf5U,0x37eaU,0x6fd4U,0xdfa8U,0xaf5bU,0x4ebdU,0x9d7aU,0x2affU,0x55feU,
		0xabfcU,0x47f3U,0x8fe6U,0x0fc7U,0x1f8eU,0x3f1cU,0x7e38U,0xfc70U,0xe8ebU,0xc1ddU,
		0x93b1U,0x3769U,0x6ed2U,0xdda4U,0xab43U,0x468dU,0x8d1aU,0x0a3fU,0x147eU,0x28fcU,
		0x51f8U,0xa3f0U,0x57ebU,0xafd6U,0x4fa7U,0x9f4eU,0x2e97U,0x5d2eU,0xba5cU,0x64b3U,
		0xc966U,0x82c7U,0x1585U,0x2b0aU,0x5614U,0xac28U,0x485bU,0x90b6U,0x3167U,0x62ceU,
		0xc59cU,0x9b33U,0x266dU,0x4cdaU,0x99b4U,0x2363U,0x46c6U,0x8d8cU,0x0b13U,0x1626U,
		0x2c4cU,0x5898U,0xb130U,0x726bU,0xe4d6U,0xd9a7U,0xa345U,0x5681U,0xad02U,0x4a0fU,
		0x941eU,0x3837U,0x706eU,0xe0dcU,0xd1b3U,0xb36dU,0x76d1U,0xeda2U,0xcb4fU,0x8695U,
		0x1d21U,0x3a42U,0x7484U,0xe908U,0xc21bU,0x943dU,0x3871U,0x70e2U,0xe1c4U,0xd383U,
		0xb70dU,0x7e11U,0xfc22U,0xe84fU,0xc095U,0x9121U,0x3249U,0x6492U,0xc924U,0x8243U,
		0x148dU,0x291aU,0x5234U,0xa468U,0x58dbU,0xb1b6U,0x7367U,0xe6ceU,0xdd97U,0xab25U,
		0x4641U,0x8c82U,0x090fU,0x121eU,0x243cU,0x4878U,0x90f0U,0x31ebU,0x63d6U,0xc7acU,
		0x9f53U,0x2eadU,0x5d5aU,0xbab4U,0x6563U,0xcac6U,0x8587U,0x1b05U,0x360aU,0x6c14U,
		0xd828U,0xa05bU,0x50bdU,0xa17aU,0x52ffU,0xa5feU,0x5bf7U,0xb7eeU,0x7fd7U,0xffaeU,
		0xef57U,0xcea5U,0x8d41U,0x0a89U,0x1512U,0x2a24U,0x5448U,0xa890U,0x412bU,0x8256U,
		0x14a7U,0x294eU,0x529cU,0xa538U,0x5a7bU,0xb4f6U,0x79e7U,0xf3ceU,0xf797U,0xff25U,
		0xee41U,0xcc89U,0x8919U,0x0239U,0x0472U,0x08e4U,0x11c8U,0x2390U,0x4720U,0x8e40U,
		0x0c8bU,0x1916U,0x322cU,0x6458U,0xc8b0U,0x816bU,0x12ddU,0x25baU,0x4b74U,0x96e8U,
		0x3ddbU,0x7bb6U,0xf76cU,0xfed3U,0xedadU,0xcb51U,0x86a9U,0x1d59U,0x3ab2U,0x7564U,
		0xeac8U,0xc59bU,0x9b3dU,0x2671U,0x4ce2U,0x99c4U,0x2383U,0x4706U,0x8e0cU,0x0c13U,
		0x1826U,0x304cU,0x6098U,0xc130U,0x926bU,0x34ddU,0x69baU,0xd374U,0xb6e3U,0x7dcdU,
		0xfb9aU,0xe73fU,0xde75U,0xace1U,0x49c9U,0x9392U,0x372fU,0x6e5eU,0xdcbcU,0xa973U,
		0x42edU,0x85daU,0x1bbfU,0x377eU,0x6efcU,0xddf8U,0xabfbU,0x47fdU,0x8ffaU,0x0fffU,
		0x1ffeU,0x3ffcU,0x7ff8U,0xfff0U,0xefebU,0xcfddU,0x8fb1U,0x0f69U,0x1ed2U,0x3da4U,
		0x7b48U,0xf690U,0xfd2bU,0xea5dU,0xc4b1U,0x9969U,0x22d9U,0x45b2U,0x8b64U,0x06c3U,
		0x0d86U,0x1b0cU,0x3618U,0x6c30U,0xd860U,0xa0cbU,0x519dU,0xa33aU,0x567fU,0xacfeU,
		0x49f7U,0x93eeU,0x37d7U,0x6faeU,0xdf5cU,0xaeb3U,0x4d6dU,0x9adaU,0x25bfU,0x4b7eU,
		0x96fcU,0x3df3U,0x7be6U,0xf7ccU,0xff93U,0xef2dU,0xce51U,0x8ca9U,0x0959U,0x12b2U,
		0x2564U,0x4ac8U,0x9590U,0x3b2bU,0x7656U,0xecacU,0xc953U,0x82adU,0x1551U,0x2aa2U,
		0x5544U,0xaa88U,0x451bU,0x8a36U,0x0467U,0x08ceU,0x119cU,0x2338U,0x4670U,0x8ce0U,
		0x09cbU,0x1396U,0x272cU,0x4e58U,0x9cb0U,0x296bU,0x52d6U,0xa5acU,0x5b53U,0xb6a6U,
		0x7d47U,0xfa8eU,0xe517U,0xda25U,0xa441U,0x5889U,0xb112U,0x722fU,0xe45eU,0xd8b7U,
		0xa165U,0x52c1U,0xa582U,0x5b0fU,0xb61eU,0x7c37U,0xf86eU,0xe0d7U,0xd1a5U,0xb341U,
		0x7689U,0xed12U,0xca2fU,0x8455U,0x18a1U,0x3142U,0x6284U,0xc508U,0x9a1bU,0x243dU,
		0x487aU,0x90f4U,0x31e3U,0x63c6U,0xc78cU,0x9f13U,0x2e2dU,0x5c5aU,0xb8b4U,0x6163U,
		0xc2c6U,0x9587U,0x3b05U,0x760aU,0xec14U,0xc823U,0x804dU,0x1091U,0x2122U,0x4244U,
		0x8488U,0x191bU,0x3236U,0x646cU,0xc8d8U,0x81bbU,0x137dU,0x26faU,0x4df4U,0x9be8U,
		0x27dbU,0x4fb6U,0x9f6cU,0x2ed3U,0x5da6U,0xbb4cU,0x6693U,0xcd26U,0x8a47U,0x0485U,
		0x090aU,0x1214U,0x2428U,0x4850U,0x90a0U,0x314bU,0x6296U,0xc52cU,0x9a53U,0x24adU,
		0x495aU,0x92b4U,0x3563U,0x6ac6U,0xd58cU,0xbb13U,0x662dU,0xcc5aU,0x88bfU,0x0175U,
		0x02eaU,0x05d4U,0x0ba8U,0x1750U,0x2ea0U,0x5d40U,0xba80U,0x650bU,0xca16U,0x8427U,
		0x1845U,0x308aU,0x6114U,0xc228U,0x945bU,0x38bdU,0x717aU,0xe2f4U,0xd5e3U,0xbbcdU,
		0x6791U,0xcf22U,0x8e4fU,0x0c95U,0x192aU,0x3254U,0x64a8U,0xc950U,0x82abU,0x155dU,
		0x2abaU,0x5574U,0xaae8U,0x45dbU,0x8bb6U,0x0767U,0x0eceU,0x1d9cU,0x3b38U,0x7670U,
		0xece0U,0xc9cbU,0x839dU,0x1731U,0x2e62U,0x5cc4U,0xb988U,0x631bU,0xc636U,0x9c67U,
		0x28c5U,0x518aU,0xa314U,0x5623U,0xac46U,0x4887U,0x910eU,0x3217U,0x642eU,0xc85cU,
		0x80b3U,0x116dU,0x22daU,0x45b4U,0x8b68U,0x06dbU,0x0db6U,0x1b6cU,0x36d8U,0x6db0U,
		0xdb60U,0xa6cbU,0x5d9dU,0xbb3aU,0x667fU,0xccfeU,0x89f7U,0x03e5U,0x07caU,0x0f94U,
		0x1f28U,0x3e50U,0x7ca0U,0xf940U,0xe28bU,0xd51dU,0xba31U,0x6469U,0xc8d2U,0x81afU,
		0x1355U,0x26aaU,0x4d54U,0x9aa8U,0x255bU,0x4ab6U,0x956cU,0x3ad3U,0x75a6U,0xeb4cU,
		0xc693U,0x9d2dU,0x2a51U,0x54a2U,0xa944U,0x4283U,0x8506U,0x1a07U,0x340eU,0x681cU,
		0xd038U,0xb07bU,0x70fdU,0xe1faU,0xd3ffU,0xb7f5U,0x7fe1U,0xffc2U,0xef8fU,0xcf15U,
		0x8e21U,0x0c49U,0x1892U,0x3124U,0x6248U,0xc490U,0x992bU,0x225dU,0x44baU,0x8974U,
		0x02e3U,0x05c6U,0x0b8cU,0x1718U,0x2e30U,0x5c60U,0xb8c0U,0x618bU,0xc316U,0x9627U,
		0x3c45U,0x788aU,0xf114U,0xf223U,0xf44dU,0xf891U,0xe129U,0xd259U,0xb4b9U,0x7979U,
		0xf2f2U,0xf5efU,0xfbd5U,0xe7a1U,0xdf49U,0xae99U,0x4d39U,0x9a72U,0x24efU,0x49deU,
		0x93bcU,0x3773U,0x6ee6U,0xddccU,0xab93U,0x472dU,0x8e5aU,0x0cbfU,0x197eU,0x32fcU,
		0x65f8U,0xcbf0U,0x87ebU,0x1fddU,0x3fbaU,0x7f74U,0xfee8U,0xeddbU,0xcbbdU,0x8771U,
		0x1ee9U,0x3dd2U,0x7ba4U,0xf748U,0xfe9bU,0xed3dU,0xca71U,0x84e9U,0x19d9U,0x33b2U,
		0x6764U,0xcec8U,0x8d9bU,0x0b3dU,0x167aU,0x2cf4U,0x59e8U,0xb3d0U,0x77abU,0xef56U,
		0xcea7U,0x8d45U,0x0a81U,0x1502U,0x2a04U,0x5408U,0xa810U,0x402bU,0x8056U,0x10a7U,
		0x214eU,0x429cU,0x8538U,0x1a7bU,0x34f6U,0x69ecU,0xd3d8U,0xb7bbU,0x7f7dU,0xfefaU,
		0xedffU,0xcbf5U,0x87e1U,0x1fc9U,0x3f92U,0x7f24U,0xfe48U,0xec9bU,0xc93dU,0x8271U,
		0x14e9U,0x29d2U,0x53a4U,0xa748U,0x5e9bU,0xbd36U,0x6a67U,0xd4ceU,0xb997U,0x6325U,
		0xc64aU,0x9c9fU,0x2935U,0x526aU,0xa4d4U,0x59a3U,0xb346U,0x7687U,0xed0eU,0xca17U,
		0x8425U,0x1841U,0x3082U,0x6104U,0xc208U,0x941bU,0x383dU,0x707aU,0xe0f4U,0xd1e3U,
		0xb3cdU,0x7791U,0xef22U,0xce4fU,0x8c95U,0x0921U,0x1242U,0x2484U,0x4908U,0x9210U,
		0x342bU,0x6856U,0xd0acU,0xb153U,0x72adU,0xe55aU,0xdabfU,0xa575U,0x5ae1U,0xb5c2U,
		0x7b8fU,0xf71eU,0xfe37U,0xec65U,0xc8c1U,0x8189U,0x1319U,0x2632U,0x4c64U,0x98c8U,
		0x219bU,0x4336U,0x866cU,0x1cd3U,0x39a6U,0x734cU,0xe698U,0xdd3bU,0xaa7dU,0x44f1U,
		0x89e2U,0x03cfU,0x079eU,0x0f3cU,0x1e78U,0x3cf0U,0x79e0U,0xf3c0U,0xf78bU,0xff1dU,
		0xee31U,0xcc69U,0x88d9U,0x01b9U,0x0372U,0x06e4U,0x0dc8U,0x1b90U,0x3720U,0x6e40U,
		0xdc80U,0xa90bU,0x421dU,0x843aU,0x187fU,0x30feU,0x61fcU,0xc3f8U,0x97fbU,0x3ffdU,
		0x7ffaU,0xfff4U,0xefe3U,0xcfcdU,0x8f91U,0x0f29U,0x1e52U,0x3ca4U,0x7948U,0xf290U,
		0xf52bU,0xfa5dU,0xe4b1U,0xd969U,0xa2d9U,0x55b9U,0xab72U,0x46efU,0x8ddeU,0x0bb7U,
		0x176eU,0x2edcU,0x5db8U,0xbb70U,0x66ebU,0xcdd6U,0x8ba7U,0x0745U,0x0e8aU,0x1d14U,
		0x3a28U,0x7450U,0xe8a0U,0xc14bU,0x929dU,0x3531U,0x6a62U,0xd4c4U,0xb983U,0x630dU,
		0xc61aU,0x9c3fU,0x2875U,0x50eaU,0xa1d4U,0x53a3U,0xa746U,0x5e87U,0xbd0eU,0x6a17U,
		0xd42eU,0xb857U,0x60a5U,0xc14aU,0x929fU,0x3535U,0x6a6aU,0xd4d4U,0xb9a3U,0x634dU,
		0xc69aU,0x9d3fU,0x2a75U,0x54eaU,0xa9d4U,0x43a3U,0x8746U,0x1e87U,0x3d0eU,0x7a1cU,
		0xf438U,0xf87bU,0xe0fdU,0xd1f1U,0xb3e9U,0x77d9U,0xefb2U,0xcf6fU,0x8ed5U,0x0da1U,
		0x1b42U,0x3684U,0x6d08U,0xda10U,0xa42bU,0x585dU,0xb0baU,0x717fU,0xe2feU,0xd5f7U,
		0xbbe5U,0x67c1U,0xcf82U,0x8f0fU,0x0e15U,0x1c2aU,0x3854U,0x70a8U,0xe150U,0xd2abU,
		0xb55dU,0x7ab1U,0xf562U,0xfacfU,0xe595U,0xdb21U,0xa649U,0x5c99U,0xb932U,0x626fU,
		0xc4deU,0x99b7U,0x2365U,0x46caU,0x8d94U,0x0b23U,0x1646U,0x2c8cU,0x5918U,0xb230U,
		0x746bU,0xe8d6U,0xc1a7U,0x9345U,0x3681U,0x6d02U,0xda04U,0xa403U,0x580dU,0xb01aU,
		0x703fU,0xe07eU,0xd0f7U,0xb1e5U,0x73c1U,0xe782U,0xdf0fU,0xae15U,0x4c21U,0x9842U,
		0x208fU,0x411eU,0x823cU,0x1473U,0x28e6U,0x51ccU,0xa398U,0x573bU,0xae76U,0x4ce7U,
		0x99ceU,0x2397U,0x472eU,0x8e5cU,0x0cb3U,0x1966U,0x32ccU,0x6598U,0xcb30U,0x866bU,
		0x1cddU,0x39baU,0x7374U,0xe6e8U,0xdddbU,0xabbdU,0x4771U,0x8ee2U,0x0dcfU,0x1b9eU,
		0x373cU,0x6e78U,0xdcf0U,0xa9ebU,0x43ddU,0x87baU,0x1f7fU,0x3efeU,0x7dfcU,0xfbf8U,
		0xe7fbU,0xdffdU,0xaff1U,0x4fe9U,0x9fd2U,0x2fafU,0x5f5eU,0xbebcU,0x6d73U,0xdae6U,
		0xa5c7U,0x5b85U,0xb70aU,0x7e1fU,0xfc3eU,0xe877U,0xc0e5U,0x91c1U,0x3389U,0x6712U,
		0xce24U,0x8c43U,0x088dU,0x111aU,0x2234U,0x4468U,0x88d0U,0x01abU,0x0356U,0x06acU,
		0x0d58U,0x1ab0U,0x3560U,0x6ac0U,0xd580U,0xbb0bU,0x661dU,0xcc3aU,0x887fU,0x00f5U,
		0x01eaU,0x03d4U,0x07a8U,0x0f50U,0x1ea0U,0x3d40U,0x7a80U,0xf500U,0xfa0bU,0xe41dU,
		0xd831U,0xa069U,0x50d9U,0xa1b2U,0x536fU,0xa6deU,0x5db7U,0xbb6eU,0x66d7U,0xcdaeU,
		0x8b57U,0x06a5U,0x0d4aU,0x1a94U,0x3528U,0x6a50U,0xd4a0U,0xb94bU,0x629dU,0xc53aU,
		0x9a7fU,0x24f5U,0x49eaU,0x93d4U,0x37a3U,0x6f46U,0xde8cU,0xad13U,0x4a2dU,0x945aU,
		0x38bfU,0x717eU,0xe2fcU,0xd5f3U,0xbbedU,0x67d1U,0xcfa2U,0x8f4fU,0x0e95U,0x1d2aU,
		0x3a54U,0x74a8U,0xe950U,0xc2abU,0x955dU,0x3ab1U,0x7562U,0xeac4U,0xc583U,0x9b0dU,
		0x2611U,0x4c22U,0x9844U,0x2083U,0x4106U,0x820cU,0x1413U,0x2826U,0x504cU,0xa098U,
		0x513bU,0xa276U,0x54e7U,0xa9ceU,0x4397U,0x872eU,0x1e57U,0x3caeU,0x795cU,0xf2b8U,
		0xf57bU,0xfafdU,0xe5f1U,0xdbe9U,0xa7d9U,0x5fb9U,0xbf72U,0x6eefU,0xdddeU,0xabb7U,
		0x4765U,0x8ecaU,0x0d9fU,0x1b3eU,0x367cU,0x6cf8U,0xd9f0U,0xa3ebU,0x57ddU,0xafbaU,
		0x4f7fU,0x9efeU,0x2df7U,0x5beeU,0xb7dcU,0x7fb3U,0xff66U,0xeec7U,0xcd85U,0x8b01U,
		0x0609U,0x0c12U,0x1824U,0x3048U,0x6090U,0xc120U,0x924bU,0x349dU,0x693aU,0xd274U,
		0xb4e3U,0x79cdU,0xf39aU,0xf73fU,0xfe75U,0xece1U,0xc9c9U,0x8399U,0x1739U,0x2e72U,
		0x5ce4U,0xb9c8U,0x639bU,0xc736U,0x9e67U,0x2cc5U,0x598aU,0xb314U,0x7623U,0xec46U,
		0xc887U,0x8105U,0x1201U,0x2402U,0x4804U,0x9008U,0x301bU,0x6036U,0xc06cU,0x90d3U,
		0x31adU,0x635aU,0xc6b4U,0x9d63U,0x2acdU,0x559aU,0xab34U,0x4663U,0x8cc6U,0x0987U,
		0x130eU,0x261cU,0x4c38U,0x9870U,0x20ebU,0x41d6U,0x83acU,0x1753U,0x2ea6U,0x5d4cU,
		0xba98U,0x653bU,0xca76U,0x84e7U,0x19c5U,0x338aU,0x6714U,0xce28U,0x8c5bU,0x08bdU,
		0x117aU,0x22f4U,0x45e8U,0x8bd0U,0x07abU,0x0f56U,0x1eacU,0x3d58U,0x7ab0U,0xf560U,
		0xfacbU,0xe59dU,0xdb31U,0xa669U,0x5cd9U,0xb9b2U,0x636fU,0xc6deU,0x9db7U,0x2b65U,
		0x56caU,0xad94U,0x4b23U,0x9646U,0x3c87U,0x790eU,0xf21cU,0xf433U,0xf86dU,0xe0d1U,
		0xd1a9U,0xb359U,0x76b9U,0xed72U,0xcaefU,0x85d5U,0x1ba1U,0x3742U,0x6e84U,0xdd08U,
		0xaa1bU,0x443dU,0x887aU,0x00ffU,0x01feU,0x03fcU,0x07f8U,0x0ff0U,0x1fe0U,0x3fc0U,
		0x7f80U,0xff00U,0xee0bU,0xcc1dU,0x8831U,0x0069U,0x00d2U,0x01a4U,0x0348U,0x0690U,
		0x0d20U,0x1a40U,0x3480U,0x6900U,0xd200U,0xb40bU,0x781dU,0xf03aU,0xf07fU,0xf0f5U,
		0xf1e1U,0xf3c9U,0xf799U,0xff39U,0xee79U,0xccf9U,0x89f9U,0x03f9U,0x07f2U,0x0fe4U,
		0x1fc8U,0x3f90U,0x7f20U,0xfe40U,0xec8bU,0xc91dU,0x8231U,0x1469U,0x28d2U,0x51a4U,
		0xa348U,0x569bU,0xad36U,0x4a67U,0x94ceU,0x3997U,0x732eU,0xe65cU,0xdcb3U,0xa96dU,
		0x42d1U,0x85a2U,0x1b4fU,0x369eU,0x6d3cU,0xda78U,0xa4fbU,0x59fdU,0xb3faU,0x77ffU,
		0xeffeU,0xcff7U,0x8fe5U,0x0fc1U,0x1f82U,0x3f04U,0x7e08U,0xfc10U,0xe82bU,0xc05dU,
		0x90b1U,0x3169U,0x62d2U,0xc5a4U,0x9b43U,0x268dU,0x4d1aU,0x9a34U,0x2463U,0x48c6U,
		0x918cU,0x3313U,0x6626U,0xcc4cU,0x8893U,0x012dU,0x025aU,0x04b4U,0x0968U,0x12d0U,
		0x25a0U,0x4b40U,0x9680U,0x3d0bU,0x7a16U,0xf42cU,0xf853U,0xe0adU,0xd151U,0xb2a9U,
		0x7559U,0xeab2U,0xc56fU,0x9ad5U,0x25a1U,0x4b42U,0x9684U,0x3d03U,0x7a06U,0xf40cU,
		0xf813U,0xe02dU,0xd051U,0xb0a9U,0x7159U,0xe2b2U,0xd56fU,0xbad5U,0x65a1U,0xcb42U,
		0x868fU,0x1d15U,0x3a2aU,0x7454U,0xe8a8U,0xc15bU,0x92bdU,0x3571U,0x6ae2U,0xd5c4U,
		0xbb83U,0x670dU,0xce1aU,0x8c3fU,0x0875U,0x10eaU,0x21d4U,0x43a8U,0x8750U,0x1eabU,
		0x3d56U,0x7aacU,0xf558U,0xfabbU,0xe57dU,0xdaf1U,0xa5e9U,0x5bd9U,0xb7b2U,0x7f6fU,
		0xfedeU,0xedb7U,0xcb65U,0x86c1U,0x1d89U,0x3b12U,0x7624U,0xec48U,0xc89bU,0x813dU,
		0x1271U,0x24e2U,0x49c4U,0x9388U,0x371bU,0x6e36U,0xdc6cU,0xa8d3U,0x41adU,0x835aU,
		0x16bfU,0x2d7eU,0x5afcU,0xb5f8U,0x7bfbU,0xf7f6U,0xffe7U,0xefc5U,0xcf81U,0x8f09U,
		0x0e19U,0x1c32U,0x3864U,0x70c8U,0xe190U,0xd32bU,0xb65dU,0x7cb1U,0xf962U,0xe2cfU,
		0xd595U,0xbb21U,0x6649U,0xcc92U,0x892fU,0x0255U,0x04aaU,0x0954U,0x12a8U,0x2550U,
		0x4aa0U,0x9540U,0x3a8bU,0x7516U,0xea2cU,0xc453U,0x98adU,0x2151U,0x42a2U,0x8544U,
		0x1a83U,0x3506U,0x6a0cU,0xd418U,0xb83bU,0x607dU,0xc0faU,0x91ffU,0x33f5U,0x67eaU,
		0xcfd4U,0x8fa3U,0x0f4dU,0x1e9aU,0x3d34U,0x7a68U,0xf4d0U,0xf9abU,0xe35dU,0xd6b1U,
		0xbd69U,0x6ad9U,0xd5b2U,0xbb6fU,0x66d5U,0xcdaaU,0x8b5fU,0x06b5U,0x0d6aU,0x1ad4U,
		0x35a8U,0x6b50U,0xd6a0U,0xbd4bU,0x6a9dU,0xd53aU,0xba7fU,0x64f5U,0xc9eaU,0x83dfU,
		0x17b5U,0x2f6aU,0x5ed4U,0xbda8U,0x6b5bU,0xd6b6U,0xbd67U,0x6ac5U,0xd58aU,0xbb1fU,
		0x6635U,0xcc6aU,0x88dfU,0x01b5U,0x036aU,0x06d4U,0x0da8U,0x1b50U,0x36a0U,0x6d40U,
		0xda80U,0xa50bU,0x5a1dU,0xb43aU,0x787fU,0xf0feU,0xf1f7U,0xf3e5U,0xf7c1U,0xff89U,
		0xef19U,0xce39U,0x8c79U,0x08f9U,0x11f2U,0x23e4U,0x47c8U,0x8f90U,0x0f2bU,0x1e56U,
		0x3cacU,0x7958U,0xf2b0U,0xf56bU,0xfaddU,0xe5b1U,0xdb69U,0xa6d9U,0x5db9U,0xbb72U,
		0x66efU,0xcddeU,0x8bb7U,0x0765U,0x0ecaU,0x1d94U,0x3b28U,0x7650U,0xeca0U,0xc94bU,
		0x829dU,0x1531U,0x2a62U,0x54c4U,0xa988U,0x431bU,0x8636U,0x1c67U,0x38ceU,0x719cU,
		0xe338U,0xd67bU,0xbcfdU,0x69f1U,0xd3e2U,0xb7cfU,0x7f95U,0xff2aU,0xee5fU,0xccb5U,
		0x8961U,0x02c9U,0x0592U,0x0b24U,0x1648U,0x2c90U,0x5920U,0xb240U,0x748bU,0xe916U,
		0xc227U,0x9445U,0x3881U,0x7102U,0xe204U,0xd403U,0xb80dU,0x6011U,0xc022U,0x904fU,
		0x3095U,0x612aU,0xc254U,0x94a3U,0x394dU,0x729aU,0xe534U,0xda63U,0xa4cdU,0x5991U,
		0xb322U,0x764fU,0xec9eU,0xc937U,0x8265U,0x14c1U,0x2982U,0x5304U,0xa608U,0x5c1bU,
		0xb836U,0x6067U,0xc0ceU,0x9197U,0x3325U,0x664aU,0xcc94U,0x8923U,0x024dU,0x049aU,
		0x0934U,0x1268U,0x24d0U,0x49a0U,0x9340U,0x368bU,0x6d16U,0xda2cU,0xa453U,0x58adU,
		0xb15aU,0x72bfU,0xe57eU,0xdaf7U,0xa5e5U,0x5bc1U,0xb782U,0x7f0fU,0xfe1eU,0xec37U,
		0xc865U,0x80c1U,0x1189U,0x2312U,0x4624U,0x8c48U,0x089bU,0x1136U,0x226cU,0x44d8U,
		0x89b0U,0x036bU,0x06d6U,0x0dacU,0x1b58U,0x36b0U,0x6d60U,0xdac0U,0xa58bU,0x5b1dU,
		0xb63aU,0x7c7fU,0xf8feU,0xe1f7U,0xd3e5U,0xb7c1U,0x7f89U,0xff12U,0xee2fU,0xcc55U,
		0x88a1U,0x0149U,0x0292U,0x0524U,0x0a48U,0x1490U,0x2920U,0x5240U,0xa480U,0x590bU,
		0xb216U,0x7427U,0xe84eU,0xc097U,0x9125U,0x3241U,0x6482U,0xc904U,0x8203U,0x140dU,
		0x281aU,0x5034U,0xa068U,0x50dbU,0xa1b6U,0x5367U,0xa6ceU,0x5d97U,0xbb2eU,0x6657U,
		0xccaeU,0x8957U,0x02a5U,0x054aU,0x0a94U,0x1528U,0x2a50U,0x54a0U,0xa940U,0x428bU,
		0x8516U,0x1a27U,0x344eU,0x689cU,0xd138U,0xb27bU,0x74fdU,0xe9faU,0xc3ffU,0x97f5U,
		0x3fe1U,0x7fc2U,0xff84U,0xef03U,0xce0dU,0x8c11U,0x0829U,0x1052U,0x20a4U,0x4148U,
		0x8290U,0x152bU,0x2a56U,0x54acU,0xa958U,0x42bbU,0x8576U,0x1ae7U,0x35ceU,0x6b9cU,
		0xd738U,0xbe7bU,0x6cfdU,0xd9faU,0xa3ffU,0x57f5U,0xafeaU,0x4fdfU,0x9fbeU,0x2f77U,
		0x5eeeU,0xbddcU,0x6bb3U,0xd766U,0xbec7U,0x6d85U,0xdb0aU,0xa61fU,0x5c35U,0xb86aU,
		0x60dfU,0xc1beU,0x9377U,0x36e5U,0x6dcaU,0xdb94U,0xa723U,0x5e4dU,0xbc9aU,0x693fU,
		0xd27eU,0xb4f7U,0x79e5U,0xf3caU,0xf79fU,0xff35U,0xee61U,0xccc9U,0x8999U,0x0339U,
		0x0672U,0x0ce4U,0x19c8U,0x3390U,0x6720U,0xce40U,0x8c8bU,0x091dU,0x123aU,0x2474U,
		0x48e8U,0x91d0U,0x33abU,0x6756U,0xceacU,0x8d53U,0x0aadU,0x155aU,0x2ab4U,0x5568U,
		0xaad0U,0x45abU,0x8b56U,0x06a7U,0x0d4eU,0x1a9cU,0x3538U,0x6a70U,0xd4e0U,0xb9cbU,
		0x639dU,0xc73aU,0x9e7fU,0x2cf5U,0x59eaU,0xb3d4U,0x77a3U,0xef46U,0xce87U,0x8d05U,
		0x0a01U,0x1402U,0x2804U,0x5008U,0xa010U,0x502bU,0xa056U,0x50a7U,0xa14eU,0x5297U,
		0xa52eU,0x5a57U,0xb4aeU,0x7957U,0xf2aeU,0xf557U,0xfaa5U,0xe541U,0xda89U,0xa519U,
		0x5a39U,0xb472U,0x78efU,0xf1deU,0xf3b7U,0xf765U,0xfec1U,0xed89U,0xcb19U,0x8639U,
		0x1c79U,0x38f2U,0x71e4U,0xe3c8U,0xd79bU,0xbf3dU,0x6e71U,0xdce2U,0xa9cfU,0x4395U,
		0x872aU,0x1e5fU,0x3cbeU,0x797cU,0xf2f8U,0xf5fbU,0xfbfdU,0xe7f1U,0xdfe9U,0xafd9U,
		0x4fb9U,0x9f72U,0x2eefU,0x5ddeU,0xbbbcU,0x6773U,0xcee6U,0x8dc7U,0x0b85U,0x170aU,
		0x2e14U,0x5c28U,0xb850U,0x60abU,0xc156U,0x92a7U,0x3545U,0x6a8aU,0xd514U,0xba23U,
		0x644dU,0xc89aU,0x813fU,0x1275U,0x24eaU,0x49d4U,0x93a8U,0x375bU,0x6eb6U,0xdd6cU,
		0xaad3U,0x45adU,0x8b5aU,0x06bfU,0x0d7eU,0x1afcU,0x35f8U,0x6bf0U,0xd7e0U,0xbfcbU,
		0x6f9dU,0xdf3aU,0xae7fU,0x4cf5U,0x99eaU,0x23dfU,0x47beU,0x8f7cU,0x0ef3U,0x1de6U,
		0x3bccU,0x7798U,0xef30U,0xce6bU,0x8cddU,0x09b1U,0x1362U,0x26c4U,0x4d88U,0x9b10U,
		0x262bU,0x4c56U,0x98acU,0x2153U,0x42a6U,0x854cU,0x1a93U,0x3526U,0x6a4cU,0xd498U,
		0xb93bU,0x627dU,0xc4faU,0x99ffU,0x23f5U,0x47eaU,0x8fd4U,0x0fa3U,0x1f46U,0x3e8cU,
		0x7d18U,0xfa30U,0xe46bU,0xd8ddU,0xa1b1U,0x5369U,0xa6d2U,0x5dafU,0xbb5eU,0x66b7U,
		0xcd6eU,0x8ad7U,0x05a5U,0x0b4aU,0x1694U,0x2d28U,0x5a50U,0xb4a0U,0x794bU,0xf296U,
		0xf527U,0xfa45U,0xe481U,0xd909U,0xa219U,0x5439U,0xa872U,0x40efU,0x81deU,0x13b7U,
		0x276eU,0x4edcU,0x9db8U,0x2b7bU,0x56f6U,0xadecU,0x4bd3U,0x97a6U,0x3f47U,0x7e8eU,
		0xfd1cU,0xea33U,0xc46dU,0x98d1U,0x21a9U,0x4352U,0x86a4U,0x1d43U,0x3a86U,0x750cU,
		0xea18U,0xc43bU,0x987dU,0x20f1U,0x41e2U,0x83c4U,0x1783U,0x2f06U,0x5e0cU,0xbc18U,
		0x683bU,0xd076U,0xb0e7U,0x71c5U,0xe38aU,0xd71fU,0xbe35U,0x6c61U,0xd8c2U,0xa18fU,
		0x5315U,0xa62aU,0x5c5fU,0xb8beU,0x6177U,0xc2eeU,0x95d7U,0x3ba5U,0x774aU,0xee94U,
		0xcd23U,0x8a4dU,0x0491U,0x0922U,0x1244U,0x2488U,0x4910U,0x9220U,0x344bU,0x6896U,
		0xd12cU,0xb253U,0x74adU,0xe95aU,0xc2bfU,0x9575U,0x3ae1U,0x75c2U,0xeb84U,0xc703U,
		0x9e0dU,0x2c11U,0x5822U,0xb044U,0x7083U,0xe106U,0xd207U,0xb405U,0x7801U,0xf002U,
		0xf00fU,0xf015U,0xf021U,0xf049U,0xf099U,0xf139U,0xf279U,0xf4f9U,0xf9f9U,0xe3f9U,
		0xd7f9U,0xbff9U,0x6ff9U,0xdff2U,0xafefU,0x4fd5U,0x9faaU,0x2f5fU,0x5ebeU,0xbd7cU,
		0x6af3U,0xd5e6U,0xbbc7U,0x6785U,0xcf0aU,0x8e1fU,0x0c35U,0x186aU,0x30d4U,0x61a8U,
		0xc350U,0x96abU,0x3d5dU,0x7abaU,0xf574U,0xfae3U,0xe5cdU,0xdb91U,0xa729U,0x5e59U,
		0xbcb2U,0x696fU,0xd2deU,0xb5b7U,0x7b65U,0xf6caU,0xfd9fU,0xeb35U,0xc661U,0x9cc9U,
		0x2999U,0x5332U,0xa664U,0x5cc3U,0xb986U,0x6307U,0xc60eU,0x9c17U,0x2825U,0x504aU,
		0xa094U,0x5123U,0xa246U,0x5487U,0xa90eU,0x4217U,0x842eU,0x1857U,0x30aeU,0x615cU,
		0xc2b8U,0x957bU,0x3afdU,0x75faU,0xebf4U,0xc7e3U,0x9fcdU,0x2f91U,0x5f22U,0xbe44U,
		0x6c83U,0xd906U,0xa207U,0x5405U,0xa80aU,0x401fU,0x803eU,0x1077U,0x20eeU,0x41dcU,
		0x83b8U,0x177bU,0x2ef6U,0x5decU,0xbbd8U,0x67bbU,0xcf76U,0x8ee7U,0x0dc5U,0x1b8aU,
		0x3714U,0x6e28U,0xdc50U,0xa8abU,0x415dU,0x82baU,0x157fU,0x2afeU,0x55fcU,0xabf8U,
		0x47fbU,0x8ff6U,0x0fe7U,0x1fceU,0x3f9cU,0x7f38U,0xfe70U,0xecebU,0xc9ddU,0x83b1U,
		0x1769U,0x2ed2U,0x5da4U,0xbb48U,0x669bU,0xcd36U,0x8a67U,0x04c5U,0x098aU,0x1314U,
		0x2628U,0x4c50U,0x98a0U,0x214bU,0x4296U,0x852cU,0x1a53U,0x34a6U,0x694cU,0xd298U,
		0xb53bU,0x7a7dU,0xf4faU,0xf9ffU,0xe3f5U,0xd7e1U,0xbfc9U,0x6f99U,0xdf32U,0xae6fU,
		0x4cd5U,0x99aaU,0x235fU,0x46beU,0x8d7cU,0x0af3U,0x15e6U,0x2bccU,0x5798U,0xaf30U,
		0x4e6bU,0x9cd6U,0x29a7U,0x534eU,0xa69cU,0x5d33U,0xba66U,0x64c7U,0xc98eU,0x8317U,
		0x1625U,0x2c4aU,0x5894U,0xb128U,0x725bU,0xe4b6U,0xd967U,0xa2c5U,0x5581U,0xab02U,
		0x460fU,0x8c1eU,0x0837U,0x106eU,0x20dcU,0x41b8U,0x8370U,0x16ebU,0x2dd6U,0x5bacU,
		0xb758U,0x7ebbU,0xfd76U,0xeae7U,0xc5c5U,0x9b81U,0x2709U,0x4e12U,0x9c24U,0x2843U,
		0x5086U,0xa10cU,0x5213U,0xa426U,0x5847U,0xb08eU,0x7117U,0xe22eU,0xd457U,0xb8a5U,
		0x6141U,0xc282U,0x950fU,0x3a15U,0x742aU,0xe854U,0xc0a3U,0x914dU,0x3291U,0x6522U,
		0xca44U,0x8483U,0x190dU,0x321aU,0x6434U,0xc868U,0x80dbU,0x11bdU,0x237aU,0x46f4U,
		0x8de8U,0x0bdbU,0x17b6U,0x2f6cU,0x5ed8U,0xbdb0U,0x6b6bU,0xd6d6U,0xbda7U,0x6b45U,
		0xd68aU,0xbd1fU,0x6a35U,0xd46aU,0xb8dfU,0x61b5U,0xc36aU,0x96dfU,0x3db5U,0x7b6aU,
		0xf6d4U,0xfda3U,0xeb4dU,0xc691U,0x9d29U,0x2a59U,0x54b2U,0xa964U,0x42c3U,0x8586U,
		0x1b07U,0x360eU,0x6c1cU,0xd838U,0xa07bU,0x50fdU,0xa1faU,0x53ffU,0xa7feU,0x5ff7U,
		0xbfeeU,0x6fd7U,0xdfaeU,0xaf57U,0x4ea5U,0x9d4aU,0x2a9fU,0x553eU,0xaa7cU,0x44f3U,
		0x89e6U,0x03c7U,0x078eU,0x0f1cU,0x1e38U,0x3c70U,0x78e0U,0xf1c0U,0xf38bU,0xf71dU,
		0xfe31U,0xec69U,0xc8d9U,0x81b9U,0x1379U,0x26f2U,0x4de4U,0x9bc8U,0x279bU,0x4f36U,
		0x9e6cU,0x2cd3U,0x59a6U,0xb34cU,0x7693U,0xed26U,0xca47U,0x8485U,0x1901U,0x3202U,
		0x6404U,0xc808U,0x801bU,0x103dU,0x207aU,0x40f4U,0x81e8U,0x13dbU,0x27b6U,0x4f6cU,
		0x9ed8U,0x2dbbU,0x5b76U,0xb6ecU,0x7dd3U,0xfba6U,0xe747U,0xde85U,0xad01U,0x4a09U,
		0x9412U,0x382fU,0x705eU,0xe0bcU,0xd173U,0xb2edU,0x75d1U,0xeba2U,0xc74fU,0x9e95U,
		0x2d21U,0x5a42U,0xb484U,0x7903U,0xf206U,0xf407U,0xf805U,0xe001U,0xd009U,0xb019U,
		0x7039U,0xe072U,0xd0efU,0xb1d5U,0x73a1U,0xe742U,0xde8fU,0xad15U,0x4a21U,0x9442U,
		0x388fU,0x711eU,0xe23cU,0xd473U,0xb8edU,0x61d1U,0xc3a2U,0x974fU,0x3e95U,0x7d2aU,
		0xfa54U,0xe4a3U,0xd94dU,0xa291U,0x5529U,0xaa52U,0x44afU,0x895eU,0x02b7U,0x056eU,
		0x0adcU,0x15b8U,0x2b70U,0x56e0U,0xadc0U,0x4b8bU,0x9716U,0x3e27U,0x7c4eU,0xf89cU,
		0xe133U,0xd26dU,0xb4d1U,0x79a9U,0xf352U,0xf6afU,0xfd55U,0xeaa1U,0xc549U,0x9a99U,
		0x2539U,0x4a72U,0x94e4U,0x39c3U,0x7386U,0xe70cU,0xde13U,0xac2dU,0x4851U,0x90a2U,
		0x314fU,0x629eU,0xc53cU,0x9a73U,0x24edU,0x49daU,0x93b4U,0x3763U,0x6ec6U,0xdd8cU,
		0xab13U,0x462dU,0x8c5aU,0x08bfU,0x117eU,0x22fcU,0x45f8U,0x8bf0U,0x07ebU,0x0fd6U,
		0x1facU,0x3f58U,0x7eb0U,0xfd60U,0xeacbU,0xc59dU,0x9b31U,0x2669U,0x4cd2U,0x99a4U,
		0x2343U,0x4686U,0x8d0cU,0x0a13U,0x1426U,0x284cU,0x5098U,0xa130U,0x526bU,0xa4d6U,
		0x59a7U,0xb34eU,0x7697U,0xed2eU,0xca57U,0x84a5U,0x1941U,0x3282U,0x6504U,0xca08U,
		0x841bU,0x183dU,0x307aU,0x60f4U,0xc1e8U,0x93dbU,0x37bdU,0x6f7aU,0xdef4U,0xade3U,
		0x4bcdU,0x979aU,0x3f3fU,0x7e7eU,0xfcfcU,0xe9f3U,0xc3edU,0x97d1U,0x3fa9U,0x7f52U,
		0xfea4U,0xed43U,0xca8dU,0x8511U,0x1a29U,0x3452U,0x68a4U,0xd148U,0xb29bU,0x753dU,
		0xea7aU,0xc4ffU,0x99f5U,0x23e1U,0x47c2U,0x8f84U,0x0f03U,0x1e06U,0x3c0cU,0x7818U,
		0xf030U,0xf06bU,0xf0ddU,0xf1b1U,0xf369U,0xf6d9U,0xfdb9U,0xeb79U,0xc6f9U,0x9df9U,
		0x2bf9U,0x57f2U,0xafe4U,0x4fc3U,0x9f86U,0x2f07U,0x5e0eU,0xbc1cU,0x6833U,0xd066U,
		0xb0c7U,0x7185U,0xe30aU,0xd61fU,0xbc35U,0x6861U,0xd0c2U,0xb18fU,0x7315U,0xe62aU,
		0xdc5fU,0xa8b5U,0x4161U,0x82c2U,0x158fU,0x2b1eU,0x563cU,0xac78U,0x48fbU,0x91f6U,
		0x33e7U,0x67ceU,0xcf9cU,0x8f33U,0x0e6dU,0x1cdaU,0x39b4U,0x7368U,0xe6d0U,0xddabU,
		0xab5dU,0x46b1U,0x8d62U,0x0acfU,0x159eU,0x2b3cU,0x5678U,0xacf0U,0x49ebU,0x93d6U,
		0x37a7U,0x6f4eU,0xde9cU,0xad33U,0x4a6dU,0x94daU,0x39bfU,0x737eU,0xe6fcU,0xddf3U,
		0xabedU,0x47d1U,0x8fa2U,0x0f4fU,0x1e9eU,0x3d3cU,0x7a78U,0xf4f0U,0xf9ebU,0xe3ddU,
		0xd7b1U,0xbf69U,0x6ed9U,0xddb2U,0xab6fU,0x46d5U,0x8daaU,0x0b5fU,0x16beU,0x2d7cU,
		0x5af8U,0xb5f0U,0x7bebU,0xf7d6U,0xffa7U,0xef45U,0xce81U,0x8d09U,0x0a19U,0x1432U,
		0x2864U,0x50c8U,0xa190U,0x532bU,0xa656U,0x5ca7U,0xb94eU,0x6297U,0xc52eU,0x9a57U,
		0x24a5U,0x494aU,0x9294U,0x3523U,0x6a46U,0xd48cU,0xb913U,0x622dU,0xc45aU,0x98bfU,
		0x2175U,0x42eaU,0x85d4U,0x1ba3U,0x3746U,0x6e8cU,0xdd18U,0xaa3bU,0x447dU,0x88faU,
		0x01ffU,0x03feU,0x07fcU,0x0ff8U,0x1ff0U,0x3fe0U,0x7fc0U,0xff80U,0xef0bU,0xce1dU,
		0x8c31U,0x0869U,0x10d2U,0x21a4U,0x4348U,0x8690U,0x1d2bU,0x3a56U,0x74acU,0xe958U,
		0xc2bbU,0x957dU,0x3af1U,0x75e2U,0xebc4U,0xc783U,0x9f0dU,0x2e11U,0x5c22U,0xb844U,
		0x6083U,0xc106U,0x9207U,0x3405U,0x680aU,0xd014U,0xb023U,0x704dU,0xe09aU,0xd13fU,
		0xb275U,0x74e1U,0xe9c2U,0xc38fU,0x9715U,0x3e21U,0x7c42U,0xf884U,0xe103U,0xd20dU,
		0xb411U,0x7829U,0xf052U,0xf0afU,0xf155U,0xf2a1U,0xf549U,0xfa99U,0xe539U,0xda79U,
		0xa4f9U,0x59f9U,0xb3f2U,0x77efU,0xefdeU,0xcfb7U,0x8f65U,0x0ec1U,0x1d82U,0x3b04U,
		0x7608U,0xec10U,0xc82bU,0x805dU,0x10b1U,0x2162U,0x42c4U,0x8588U,0x1b1bU,0x3636U,
		0x6c6cU,0xd8d8U,0xa1bbU,0x537dU,0xa6faU,0x5dffU,0xbbfeU,0x67f7U,0xcfeeU,0x8fd7U,
		0x0fa5U,0x1f4aU,0x3e94U,0x7d28U,0xfa50U,0xe4abU,0xd95dU,0xa2b1U,0x5569U,0xaad2U,
		0x45afU,0x8b5eU,0x06b7U,0x0d6eU,0x1adcU,0x35b8U,0x6b70U,0xd6e0U,0xbdcbU,0x6b9dU,
		0xd73aU,0xbe7fU,0x6cf5U,0xd9eaU,0xa3dfU,0x57b5U,0xaf6aU,0x4edfU,0x9dbeU,0x2b77U,
		0x56eeU,0xaddcU,0x4bb3U,0x9766U,0x3ec7U,0x7d8eU,0xfb1cU,0xe633U,0xdc6dU,0xa8d1U,
		0x41a9U,0x8352U,0x16afU,0x2d5eU,0x5abcU,0xb578U,0x7afbU,0xf5f6U,0xfbe7U,0xe7c5U,
		0xdf81U,0xaf09U,0x4e19U,0x9c32U,0x286fU,0x50deU,0xa1bcU,0x5373U,0xa6e6U,0x5dc7U,
		0xbb8eU,0x6717U,0xce2eU,0x8c57U,0x08a5U,0x114aU,0x2294U,0x4528U,0x8a50U,0x04abU,
		0x0956U,0x12acU,0x2558U,0x4ab0U,0x9560U,0x3acbU,0x7596U,0xeb2cU,0xc653U,0x9cadU,
		0x2951U,0x52a2U,0xa544U,0x5a83U,0xb506U,0x7a07U,0xf40eU,0xf817U,0xe025U,0xd041U,
		0xb089U,0x7119U,0xe232U,0xd46fU,0xb8d5U,0x61a1U,0xc342U,0x968fU,0x3d15U,0x7a2aU,
		0xf454U,0xf8a3U,0xe14dU,0xd291U,0xb529U,0x7a59U,0xf4b2U,0xf96fU,0xe2d5U,0xd5a1U,
		0xbb49U,0x6699U,0xcd32U,0x8a6fU,0x04d5U,0x09aaU,0x1354U,0x26a8U,0x4d50U,0x9aa0U,
		0x254bU,0x4a96U,0x952cU,0x3a53U,0x74a6U,0xe94cU,0xc293U,0x952dU,0x3a51U,0x74a2U,
		0xe944U,0xc283U,0x950dU,0x3a11U,0x7422U,0xe844U,0xc083U,0x910dU,0x3211U,0x6422U,
		0xc844U,0x8083U,0x110dU,0x221aU,0x4434U,0x8868U,0x00dbU,0x01b6U,0x036cU,0x06d8U,
		0x0db0U,0x1b60U,0x36c0U,0x6d80U,0xdb00U,0xa60bU,0x5c1dU,0xb83aU,0x607fU,0xc0feU,
		0x91f7U,0x33e5U,0x67caU,0xcf94U,0x8f23U,0x0e4dU,0x1c9aU,0x3934U,0x7268U,0xe4d0U,
		0xd9abU,0xa35dU,0x56b1U,0xad62U,0x4acfU,0x959eU,0x3b37U,0x766eU,0xecdcU,0xc9b3U,
		0x836dU,0x16d1U,0x2da2U,0x5b44U,0xb688U,0x7d1bU,0xfa36U,0xe467U,0xd8c5U,0xa181U,
		0x5309U,0xa612U,0x5c2fU,0xb85eU,0x60b7U,0xc16eU,0x92d7U,0x35a5U,0x6b4aU,0xd694U,
		0xbd23U,0x6a4dU,0xd49aU,0xb93fU,0x6275U,0xc4eaU,0x99dfU,0x23b5U,0x476aU,0x8ed4U,
		0x0da3U,0x1b46U,0x368cU,0x6d18U,0xda30U,0xa46bU,0x58ddU,0xb1baU,0x737fU,0xe6feU,
		0xddf7U,0xabe5U,0x47c1U,0x8f82U,0x0f0fU,0x1e1eU,0x3c3cU,0x7878U,0xf0f0U,0xf1ebU,
		0xf3ddU,0xf7b1U,0xff69U,0xeed9U,0xcdb9U,0x8b79U,0x06f9U,0x0df2U,0x1be4U,0x37c8U,
		0x6f90U,0xdf20U,0xae4bU,0x4c9dU,0x993aU,0x227fU,0x44feU,0x89fcU,0x03f3U,0x07e6U,
		0x0fccU,0x1f98U,0x3f30U,0x7e60U,0xfcc0U,0xe98bU,0xc31dU,0x9631U,0x3c69U,0x78d2U,
		0xf1a4U,0xf343U,0xf68dU,0xfd11U,0xea29U,0xc459U,0x98b9U,0x2179U,0x42f2U,0x85e4U,
		0x1bc3U,0x3786U,0x6f0cU,0xde18U,0xac3bU,0x487dU,0x90faU,0x31ffU,0x63feU,0xc7fcU,
		0x9ff3U,0x2fedU,0x5fdaU,0xbfb4U,0x6f63U,0xdec6U,0xad87U,0x4b05U,0x960aU,0x3c1fU,
		0x783eU,0xf07cU,0xf0f3U,0xf1edU,0xf3d1U,0xf7a9U,0xff59U,0xeeb9U,0xcd79U,0x8af9U,
		0x05f9U,0x0bf2U,0x17e4U,0x2fc8U,0x5f90U,0xbf20U,0x6e4bU,0xdc96U,0xa927U,0x4245U,
		0x848aU,0x191fU,0x323eU,0x647cU,0xc8f8U,0x81fbU,0x13fdU,0x27faU,0x4ff4U,0x9fe8U,
		0x2fdbU,0x5fb6U,0xbf6cU,0x6ed3U,0xdda6U,0xab47U,0x4685U,0x8d0aU,0x0a1fU,0x143eU,
		0x287cU,0x50f8U,0xa1f0U,0x53ebU,0xa7d6U,0x5fa7U,0xbf4eU,0x6e97U,0xdd2eU,0xaa57U,
		0x44a5U,0x894aU,0x029fU,0x053eU,0x0a7cU,0x14f8U,0x29f0U,0x53e0U,0xa7c0U,0x5f8bU,
		0xbf16U,0x6e27U,0xdc4eU,0xa897U,0x4125U,0x824aU,0x149fU,0x293eU,0x527cU,0xa4f8U,
		0x59fbU,0xb3f6U,0x77e7U,0xefceU,0xcf97U,0x8f25U,0x0e41U,0x1c82U,0x3904U,0x7208U,
		0xe410U,0xd82bU,0xa05dU,0x50b1U,0xa162U,0x52cfU,0xa59eU,0x5b37U,0xb66eU,0x7cd7U,
		0xf9aeU,0xe357U,0xd6a5U,0xbd41U,0x6a89U,0xd512U,0xba2fU,0x6455U,0xc8aaU,0x815fU,
		0x12b5U,0x256aU,0x4ad4U,0x95a8U,0x3b5bU,0x76b6U,0xed6cU,0xcad3U,0x85adU,0x1b51U,
		0x36a2U,0x6d44U,0xda88U,0xa51bU,0x5a3dU,0xb47aU,0x78ffU,0xf1feU,0xf3f7U,0xf7e5U,
		0xffc1U,0xef89U,0xcf19U,0x8e39U,0x0c79U,0x18f2U,0x31e4U,0x63c8U,0xc790U,0x9f2bU,
		0x2e5dU,0x5cbaU,0xb974U,0x62e3U,0xc5c6U,0x9b87U,0x2705U,0x4e0aU,0x9c14U,0x2823U,
		0x5046U,0xa08cU,0x5113U,0xa226U,0x5447U,0xa88eU,0x4117U,0x822eU,0x1457U,0x28aeU,
		0x515cU,0xa2b8U,0x557bU,0xaaf6U,0x45e7U,0x8bceU,0x0797U,0x0f2eU,0x1e5cU,0x3cb8U,
		0x7970U,0xf2e0U,0xf5cbU,0xfb9dU,0xe731U,0xde69U,0xacd9U,0x49b9U,0x9372U,0x36efU,
		0x6ddeU,0xdbbcU,0xa773U,0x5eedU,0xbddaU,0x6bbfU,0xd77eU,0xbef7U,0x6de5U,0xdbcaU,
		0xa79fU,0x5f35U,0xbe6aU,0x6cdfU,0xd9beU,0xa377U,0x56e5U,0xadcaU,0x4b9fU,0x973eU,
		0x3e77U,0x7ceeU,0xf9dcU,0xe3b3U,0xd76dU,0xbed1U,0x6da9U,0xdb52U,0xa6afU,0x5d55U,
		0xbaaaU,0x655fU,0xcabeU,0x8577U,0x1ae5U,0x35caU,0x6b94U,0xd728U,0xbe5bU,0x6cbdU,
		0xd97aU,0xa2ffU,0x55f5U,0xabeaU,0x47dfU,0x8fbeU,0x0f77U,0x1eeeU,0x3ddcU,0x7bb8U,
		0xf770U,0xfeebU,0xedddU,0xcbb1U,0x8769U,0x1ed9U,0x3db2U,0x7b64U,0xf6c8U,0xfd9bU,
		0xeb3dU,0xc671U,0x9ce9U,0x29d9U,0x53b2U,0xa764U,0x5ec3U,0xbd86U,0x6b07U,0xd60eU,
		0xbc17U,0x6825U,0xd04aU,0xb09fU,0x7135U,0xe26aU,0xd4dfU,0xb9b5U,0x6361U,0xc6c2U,
		0x9d8fU,0x2b15U,0x562aU,0xac54U,0x48a3U,0x9146U,0x3287U,0x650eU,0xca1cU,0x8433U,
		0x186dU,0x30daU,0x61b4U,0xc368U,0x96dbU,0x3dbdU,0x7b7aU,0xf6f4U,0xfde3U,0xebcdU,
		0xc791U,0x9f29U,0x2e59U,0x5cb2U,0xb964U,0x62c3U,0xc586U,0x9b07U,0x2605U,0x4c0aU,
		0x9814U,0x2023U,0x4046U,0x808cU,0x1113U,0x2226U,0x444cU,0x8898U,0x013bU,0x0276U,
		0x04ecU,0x09d8U,0x13b0U,0x2760U,0x4ec0U,0x9d80U,0x2b0bU,0x5616U,0xac2cU,0x4853U,
		0x90a6U,0x3147U,0x628eU,0xc51cU,0x9a33U,0x246dU,0x48daU,0x91b4U,0x3363U,0x66c6U,
		0xcd8cU,0x8b13U,0x062dU,0x0c5aU,0x18b4U,0x3168U,0x62d0U,0xc5a0U,0x9b4bU,0x269dU,
		0x4d3aU,0x9a74U,0x24e3U,0x49c6U,0x938cU,0x3713U,0x6e26U,0xdc4cU,0xa893U,0x412dU,
		0x825aU,0x14bfU,0x297eU,0x52fcU,0xa5f8U,0x5bfbU,0xb7f6U,0x7fe7U,0xffceU,0xef97U,
		0xcf25U,0x8e41U,0x0c89U,0x1912U,0x3224U,0x6448U,0xc890U,0x812bU,0x125dU,0x24baU,
		0x4974U,0x92e8U,0x35dbU,0x6bb6U,0xd76cU,0xbed3U,0x6dadU,0xdb5aU,0xa6bfU,0x5d75U,
		0xbaeaU,0x65dfU,0xcbbeU,0x8777U,0x1ee5U,0x3dcaU,0x7b94U,0xf728U,0xfe5bU,0xecbdU,
		0xc971U,0x82e9U,0x15d9U,0x2bb2U,0x5764U,0xaec8U,0x4d9bU,0x9b36U,0x2667U,0x4cceU,
		0x999cU,0x2333U,0x4666U,0x8cccU,0x0993U,0x1326U,0x264cU,0x4c98U,0x9930U,0x226bU,
		0x44d6U,0x89acU,0x0353U,0x06a6U,0x0d4cU,0x1a98U,0x3530U,0x6a60U,0xd4c0U,0xb98bU,
		0x631dU,0xc63aU,0x9c7fU,0x28f5U,0x51eaU,0xa3d4U,0x57a3U,0xaf46U,0x4e87U,0x9d0eU,
		0x2a17U,0x542eU,0xa85cU,0x40b3U,0x8166U,0x12c7U,0x258eU,0x4b1cU,0x9638U,0x3c7bU,
		0x78f6U,0xf1ecU,0xf3d3U,0xf7adU,0xff51U,0xeea9U,0xcd59U,0x8ab9U,0x0579U,0x0af2U,
		0x15e4U,0x2bc8U,0x5790U,0xaf20U,0x4e4bU,0x9c96U,0x2927U,0x524eU,0xa49cU,0x5933U,
		0xb266U,0x74c7U,0xe98eU,0xc317U,0x9625U,0x3c41U,0x7882U,0xf104U,0xf203U,0xf40dU,
		0xf811U,0xe029U,0xd059U,0xb0b9U,0x7179U,0xe2f2U,0xd5efU,0xbbd5U,0x67a1U,0xcf42U,
		0x8e8fU,0x0d15U,0x1a2aU,0x3454U,0x68a8U,0xd150U,0xb2abU,0x755dU,0xeabaU,0xc57fU,
		0x9af5U,0x25e1U,0x4bc2U,0x9784U,0x3f03U,0x7e06U,0xfc0cU,0xe813U,0xc02dU,0x9051U,
		0x30a9U,0x6152U,0xc2a4U,0x9543U,0x3a8dU,0x751aU,0xea34U,0xc463U,0x98cdU,0x2191U,
		0x4322U,0x8644U,0x1c83U,0x3906U,0x720cU,0xe418U,0xd83bU,0xa07dU,0x50f1U,0xa1e2U,
		0x53cfU,0xa79eU,0x5f37U,0xbe6eU,0x6cd7U,0xd9aeU,0xa357U,0x56a5U,0xad4aU,0x4a9fU,
		0x953eU,0x3a77U,0x74eeU,0xe9dcU,0xc3b3U,0x976dU,0x3ed1U,0x7da2U,0xfb44U,0xe683U,
		0xdd0dU,0xaa11U,0x4429U,0x8852U,0x00afU,0x015eU,0x02bcU,0x0578U,0x0af0U,0x15e0U,
		0x2bc0U,0x5780U,0xaf00U,0x4e0bU,0x9c16U,0x2827U,0x504eU,0xa09cU,0x5133U,0xa266U,
		0x54c7U,0xa98eU,0x4317U,0x862eU,0x1c57U,0x38aeU,0x715cU,0xe2b8U,0xd57bU,0xbafdU,
		0x65f1U,0xcbe2U,0x87cfU,0x1f95U,0x3f2aU,0x7e54U,0xfca8U,0xe95bU,0xc2bdU,0x9571U,
		0x3ae9U,0x75d2U,0xeba4U,0xc743U,0x9e8dU,0x2d11U,0x5a22U,0xb444U,0x7883U,0xf106U,
		0xf207U,0xf405U,0xf801U,0xe009U,0xd019U,0xb039U,0x7079U,0xe0f2U,0xd1efU,0xb3d5U,
		0x77a1U,0xef42U,0xce8fU,0x8d15U,0x0a21U,0x1442U,0x2884U,0x5108U,0xa210U,0x542bU,
		0xa856U,0x40a7U,0x814eU,0x1297U,0x252eU,0x4a5cU,0x94b8U,0x397bU,0x72f6U,0xe5ecU,
		0xdbd3U,0xa7adU,0x5f51U,0xbea2U,0x6d4fU,0xda9eU,0xa537U,0x5a65U,0xb4caU,0x799fU,
		0xf33eU,0xf677U,0xfce5U,0xe9c1U,0xc389U,0x9719U,0x3e39U,0x7c72U,0xf8e4U,0xe1c3U,
		0xd38dU,0xb711U,0x7e29U,0xfc52U,0xe8afU,0xc155U,0x92a1U,0x3549U,0x6a92U,0xd524U,
		0xba43U,0x648dU,0xc91aU,0x823fU,0x1475U,0x28eaU,0x51d4U,0xa3a8U,0x575bU,0xaeb6U,
		0x4d67U,0x9aceU,0x2597U,0x4b2eU,0x965cU,0x3cb3U,0x7966U,0xf2ccU,0xf593U,0xfb2dU,
		0xe651U,0xdca9U,0xa959U,0x42b9U,0x8572U,0x1aefU,0x35deU,0x6bbcU,0xd778U,0xbefbU,
		0x6dfdU,0xdbfaU,0xa7ffU,0x5ff5U,0xbfeaU,0x6fdfU,0xdfbeU,0xaf77U,0x4ee5U,0x9dcaU,
		0x2b9fU,0x573eU,0xae7cU,0x4cf3U,0x99e6U,0x23c7U,0x478eU,0x8f1cU,0x0e33U,0x1c66U,
		0x38ccU,0x7198U,0xe330U,0xd66bU,0xbcddU,0x69b1U,0xd362U,0xb6cfU,0x7d95U,0xfb2aU,
		0xe65fU,0xdcb5U,0xa961U,0x42c9U,0x8592U,0x1b2fU,0x365eU,0x6cbcU,0xd978U,0xa2fbU,
		0x55fdU,0xabfaU,0x47ffU,0x8ffeU,0x0ff7U,0x1feeU,0x3fdcU,0x7fb8U,0xff70U,0xeeebU,
		0xcdddU,0x8bb1U,0x0769U,0x0ed2U,0x1da4U,0x3b48U,0x7690U,0xed20U,0xca4bU,0x849dU,
		0x1931U,0x3262U,0x64c4U,0xc988U,0x831bU,0x163dU,0x2c7aU,0x58f4U,0xb1e8U,0x73dbU,
		0xe7b6U,0xdf67U,0xaec5U,0x4d81U,0x9b02U,0x260fU,0x4c1eU,0x983cU,0x2073U,0x40e6U,
		0x81ccU,0x1393U,0x2726U,0x4e4cU,0x9c98U,0x293bU,0x5276U,0xa4ecU,0x59d3U,0xb3a6U,
		0x7747U,0xee8eU,0xcd17U,0x8a25U,0x0441U,0x0882U,0x1104U,0x2208U,0x4410U,0x8820U,
		0x004bU,0x0096U,0x012cU,0x0258U,0x04b0U,0x0960U,0x12c0U,0x2580U,0x4b00U,0x9600U,
		0x3c0bU,0x7816U,0xf02cU,0xf053U,0xf0adU,0xf151U,0xf2a9U,0xf559U,0xfab9U,0xe579U,
		0xdaf9U,0xa5f9U,0x5bf9U,0xb7f2U,0x7fefU,0xffdeU,0xefb7U,0xcf65U,0x8ec1U,0x0d89U,
		0x1b12U,0x3624U,0x6c48U,0xd890U,0xa12bU,0x525dU,0xa4baU,0x597fU,0xb2feU,0x75f7U,
		0xebeeU,0xc7d7U,0x9fa5U,0x2f41U,0x5e82U,0xbd04U,0x6a03U,0xd406U,0xb807U,0x6005U,
		0xc00aU,0x901fU,0x3035U,0x606aU,0xc0d4U,0x91a3U,0x334dU,0x669aU,0xcd34U,0x8a63U,
		0x04cdU,0x099aU,0x1334U,0x2668U,0x4cd0U,0x99a0U,0x234bU,0x4696U,0x8d2cU,0x0a53U,
		0x14a6U,0x294cU,0x5298U,0xa530U,0x5a6bU,0xb4d6U,0x79a7U,0xf34eU,0xf697U,0xfd25U,
		0xea41U,0xc489U,0x9919U,0x2239U,0x4472U,0x88e4U,0x01c3U,0x0386U,0x070cU,0x0e18U,
		0x1c30U,0x3860U,0x70c0U,0xe180U,0xd30bU,0xb61dU,0x7c31U,0xf862U,0xe0cfU,0xd195U,
		0xb321U,0x7649U,0xec92U,0xc92fU,0x8255U,0x14a1U,0x2942U,0x5284U,0xa508U,0x5a1bU,
		0xb436U,0x7867U,0xf0ceU,0xf197U,0xf325U,0xf641U,0xfc89U,0xe919U,0xc239U,0x9479U,
		0x38f9U,0x71f2U,0xe3e4U,0xd7c3U,0xbf8dU,0x6f11U,0xde22U,0xac4fU,0x4895U,0x912aU,
		0x325fU,0x64beU,0xc97cU,0x82f3U,0x15edU,0x2bdaU,0x57b4U,0xaf68U,0x4edbU,0x9db6U,
		0x2b67U,0x56ceU,0xad9cU,0x4b33U,0x9666U,0x3cc7U,0x798eU,0xf31cU,0xf633U,0xfc6dU,
		0xe8d1U,0xc1a9U,0x9359U,0x36b9U,0x6d72U,0xdae4U,0xa5c3U,0x5b8dU,0xb71aU,0x7e3fU,
		0xfc7eU,0xe8f7U,0xc1e5U,0x93c1U,0x3789U,0x6f12U,0xde24U,0xac43U,0x488dU,0x911aU,
		0x323fU,0x647eU,0xc8fcU,0x81f3U,0x13edU,0x27daU,0x4fb4U,0x9f68U,0x2edbU,0x5db6U,
		0xbb6cU,0x66d3U,0xcda6U,0x8b47U,0x0685U,0x0d0aU,0x1a14U,0x3428U,0x6850U,0xd0a0U,
		0xb14bU,0x729dU,0xe53aU,0xda7fU,0xa4f5U,0x59e1U,0xb3c2U,0x778fU,0xef1eU,0xce37U,
		0x8c65U,0x08c1U,0x1182U,0x2304U,0x4608U,0x8c10U,0x082bU,0x1056U,0x20acU,0x4158U,
		0x82b0U,0x156bU,0x2ad6U,0x55acU,0xab58U,0x46bbU,0x8d76U,0x0ae7U,0x15ceU,0x2b9cU,
		0x5738U,0xae70U,0x4cebU,0x99d6U,0x23a7U,0x474eU,0x8e9cU,0x0d33U,0x1a66U,0x34ccU,
		0x6998U,0xd330U,0xb66bU,0x7cddU,0xf9baU,0xe37fU,0xd6f5U,0xbde1U,0x6bc9U,0xd792U,
		0xbf2fU,0x6e55U,0xdcaaU,0xa95fU,0x42b5U,0x856aU,0x1adfU,0x35beU,0x6b7cU,0xd6f8U,
		0xbdfbU,0x6bfdU,0xd7faU,0xbfffU,0x6ff5U,0xdfeaU,0xafdfU,0x4fb5U,0x9f6aU,0x2edfU,
		0x5dbeU,0xbb7cU,0x66f3U,0xcde6U,0x8bc7U,0x0785U,0x0f0aU,0x1e14U,0x3c28U,0x7850U,
		0xf0a0U,0xf14bU,0xf29dU,0xf531U,0xfa69U,0xe4d9U,0xd9b9U,0xa379U,0x56f9U,0xadf2U,
		0x4befU,0x97deU,0x3fb7U,0x7f6eU,0xfedcU,0xedb3U,0xcb6dU,0x86d1U,0x1da9U,0x3b52U,
		0x76a4U,0xed48U,0xca9bU,0x853dU,0x1a71U,0x34e2U,0x69c4U,0xd388U,0xb71bU,0x7e3dU,
		0xfc7aU,0xe8ffU,0xc1f5U,0x93e1U,0x37c9U,0x6f92U,0xdf24U,0xae43U,0x4c8dU,0x991aU,
		0x223fU,0x447eU,0x88fcU,0x01f3U,0x03e6U,0x07ccU,0x0f98U,0x1f30U,0x3e60U,0x7cc0U,
		0xf980U,0xe30bU,0xd61dU,0xbc31U,0x6869U,0xd0d2U,0xb1afU,0x7355U,0xe6aaU,0xdd5fU,
		0xaab5U,0x4561U,0x8ac2U,0x058fU,0x0b1eU,0x163cU,0x2c78U,0x58f0U,0xb1e0U,0x73cbU,
		0xe796U,0xdf27U,0xae45U,0x4c81U,0x9902U,0x220fU,0x441eU,0x883cU,0x0073U,0x00e6U,
		0x01ccU,0x0398U,0x0730U,0x0e60U,0x1cc0U,0x3980U,0x7300U,0xe600U,0xdc0bU,0xa81dU,
		0x4031U,0x8062U,0x10cfU,0x219eU,0x433cU,0x8678U,0x1cfbU,0x39f6U,0x73ecU,0xe7d8U,
		0xdfbbU,0xaf7dU,0x4ef1U,0x9de2U,0x2bcfU,0x579eU,0xaf3cU,0x4e73U,0x9ce6U,0x29c7U,
		0x538eU,0xa71cU,0x5e33U,0xbc66U,0x68c7U,0xd18eU,0xb317U,0x7625U,0xec4aU,0xc89fU,
		0x8135U,0x1261U,0x24c2U,0x4984U,0x9308U,0x361bU,0x6c36U,0xd86cU,0xa0d3U,0x51adU,
		0xa35aU,0x56bfU,0xad7eU,0x4af7U,0x95eeU,0x3bd7U,0x77aeU,0xef5cU,0xceb3U,0x8d6dU,
		0x0ad1U,0x15a2U,0x2b44U,0x5688U,0xad10U,0x4a2bU,0x9456U,0x38a7U,0x714eU,0xe29cU,
		0xd533U,0xba6dU,0x64d1U,0xc9a2U,0x834fU,0x1695U,0x2d2aU,0x5a54U,0xb4a8U,0x795bU,
		0xf2b6U,0xf567U,0xfac5U,0xe581U,0xdb09U,0xa619U,0x5c39U,0xb872U,0x60efU,0xc1deU,
		0x93b7U,0x3765U,0x6ecaU,0xdd94U,0xab23U,0x464dU,0x8c9aU,0x093fU,0x127eU,0x24fcU,
		0x49f8U,0x93f0U,0x37ebU,0x6fd6U,0xdfacU,0xaf53U,0x4eadU,0x9d5aU,0x2abfU,0x557eU,
		0xaafcU,0x45f3U,0x8be6U,0x07c7U,0x0f8eU,0x1f1cU,0x3e38U,0x7c70U,0xf8e0U,0xe1cbU,
		0xd39dU,0xb731U,0x7e69U,0xfcd2U,0xe9afU,0xc355U,0x96a1U,0x3d49U,0x7a92U,0xf524U,
		0xfa43U,0xe48dU,0xd911U,0xa229U,0x5459U,0xa8b2U,0x416fU,0x82deU,0x15b7U,0x2b6eU,
		0x56dcU,0xadb8U,0x4b7bU,0x96f6U,0x3de7U,0x7bceU,0xf79cU,0xff33U,0xee6dU,0xccd1U,
		0x89a9U,0x0359U,0x06b2U,0x0d64U,0x1ac8U,0x3590U,0x6b20U,0xd640U,0xbc8bU,0x691dU,
		0xd23aU,0xb47fU,0x78f5U,0xf1eaU,0xf3dfU,0xf7b5U,0xff61U,0xeec9U,0xcd99U,0x8b39U,
		0x0679U,0x0cf2U,0x19e4U,0x33c8U,0x6790U,0xcf20U,0x8e4bU,0x0c9dU,0x193aU,0x3274U,
		0x64e8U,0xc9d0U,0x83abU,0x175dU,0x2ebaU,0x5d74U,0xbae8U,0x65dbU,0xcbb6U,0x8767U,
		0x1ec5U,0x3d8aU,0x7b14U,0xf628U,0xfc5bU,0xe8bdU,0xc171U,0x92e9U,0x35d9U,0x6bb2U,
		0xd764U,0xbec3U,0x6d8dU,0xdb1aU,0xa63fU,0x5c75U,0xb8eaU,0x61dfU,0xc3beU,0x9777U,
		0x3ee5U,0x7dcaU,0xfb94U,0xe723U,0xde4dU,0xac91U,0x4929U,0x9252U,0x34afU,0x695eU,
		0xd2bcU,0xb573U,0x7aedU,0xf5daU,0xfbbfU,0xe775U,0xdee1U,0xadc9U,0x4b99U,0x9732U,
		0x3e6fU,0x7cdeU,0xf9bcU,0xe373U,0xd6edU,0xbdd1U,0x6ba9U,0xd752U,0xbeafU,0x6d55U,
		0xdaaaU,0xa55fU,0x5ab5U,0xb56aU,0x7adfU,0xf5beU,0xfb77U,0xe6e5U,0xddc1U,0xab89U,
		0x4719U,0x8e32U,0x0c6fU,0x18deU,0x31bcU,0x6378U,0xc6f0U,0x9debU,0x2bddU,0x57baU,
		0xaf74U,0x4ee3U,0x9dc6U,0x2b87U,0x570eU,0xae1cU,0x4c33U,0x9866U,0x20c7U,0x418eU,
		0x831cU,0x1633U,0x2c66U,0x58ccU,0xb198U,0x733bU,0xe676U,0xdce7U,0xa9c5U,0x4381U,
		0x8702U,0x1e0fU,0x3c1eU,0x783cU,0xf078U,0xf0fbU,0xf1fdU,0xf3f1U,0xf7e9U,0xffd9U,
		0xefb9U,0xcf79U,0x8ef9U,0x0df9U,0x1bf2U,0x37e4U,0x6fc8U,0xdf90U,0xaf2bU,0x4e5dU,
		0x9cbaU,0x297fU,0x52feU,0xa5fcU,0x5bf3U,0xb7e6U,0x7fc7U,0xff8eU,0xef17U,0xce25U,
		0x8c41U,0x0889U,0x1112U,0x2224U,0x4448U,0x8890U,0x012bU,0x0256U,0x04acU,0x0958U,
		0x12b0U,0x2560U,0x4ac0U,0x9580U,0x3b0bU,0x7616U,0xec2cU,0xc853U,0x80adU,0x1151U,
		0x22a2U,0x4544U,0x8a88U,0x051bU,0x0a36U,0x146cU,0x28d8U,0x51b0U,0xa360U,0x56cbU,
		0xad96U,0x4b27U,0x964eU,0x3c97U,0x792eU,0xf25cU,0xf4b3U,0xf96dU,0xe2d1U,0xd5a9U,
		0xbb59U,0x66b9U,0xcd72U,0x8aefU,0x05d5U,0x0baaU,0x1754U,0x2ea8U,0x5d50U,0xbaa0U,
		0x654bU,0xca96U,0x8527U,0x1a45U,0x348aU,0x6914U,0xd228U,0xb45bU,0x78bdU,0xf17aU,
		0xf2ffU,0xf5f5U,0xfbe1U,0xe7c9U,0xdf99U,0xaf39U,0x4e79U,0x9cf2U,0x29efU,0x53deU,
		0xa7bcU,0x5f73U,0xbee6U,0x6dc7U,0xdb8eU,0xa717U,0x5e25U,0xbc4aU,0x689fU,0xd13eU,
		0xb277U,0x74e5U,0xe9caU,0xc39fU,0x9735U,0x3e61U,0x7cc2U,0xf984U,0xe303U,0xd60dU,
		0xbc11U,0x6829U,0xd052U,0xb0afU,0x7155U,0xe2aaU,0xd55fU,0xbab5U,0x6561U,0xcac2U,
		0x858fU,0x1b15U,0x362aU,0x6c54U,0xd8a8U,0xa15bU,0x52bdU,0xa57aU,0x5affU,0xb5feU,
		0x7bf7U,0xf7eeU,0xffd7U,0xefa5U,0xcf41U,0x8e89U,0x0d19U,0x1a32U,0x3464U,0x68c8U,
		0xd190U,0xb32bU,0x765dU,0xecbaU,0xc97fU,0x82f5U,0x15e1U,0x2bc2U,0x5784U,0xaf08U,
		0x4e1bU,0x9c36U,0x2867U,0x50ceU,0xa19cU,0x5333U,0xa666U,0x5cc7U,0xb98eU,0x6317U,
		0xc62eU,0x9c57U,0x28a5U,0x514aU,0xa294U,0x5523U,0xaa46U,0x4487U,0x890eU,0x0217U,
		0x042eU,0x085cU,0x10b8U,0x2170U,0x42e0U,0x85c0U,0x1b8bU,0x3716U,0x6e2cU,0xdc58U,
		0xa8bbU,0x417dU,0x82faU,0x15ffU,0x2bfeU,0x57fcU,0xaff8U,0x4ffbU,0x9ff6U,0x2fe7U,
		0x5fceU,0xbf9cU,0x6f33U,0xde66U,0xacc7U,0x4985U,0x930aU,0x361fU,0x6c3eU,0xd87cU,
		0xa0f3U,0x51edU,0xa3daU,0x57bfU,0xaf7eU,0x4ef7U,0x9deeU,0x2bd7U,0x57aeU,0xaf5cU,
		0x4eb3U,0x9d66U,0x2ac7U,0x558eU,0xab1cU,0x4633U,0x8c66U,0x08c7U,0x118eU,0x231cU,
		0x4638U,0x8c70U,0x08ebU,0x11d6U,0x23acU,0x4758U,0x8eb0U,0x0d6bU,0x1ad6U,0x35acU,
		0x6b58U,0xd6b0U,0xbd6bU,0x6addU,0xd5baU,0xbb7fU,0x66f5U,0xcdeaU,0x8bdfU,0x07b5U,
		0x0f6aU,0x1ed4U,0x3da8U,0x7b50U,0xf6a0U,0xfd4bU,0xea9dU,0xc531U,0x9a69U,0x24d9U,
		0x49b2U,0x9364U,0x36c3U,0x6d86U,0xdb0cU,0xa613U,0x5c2dU,0xb85aU,0x60bfU,0xc17eU,
		0x92f7U,0x35e5U,0x6bcaU,0xd794U,0xbf23U,0x6e4dU,0xdc9aU,0xa93fU,0x4275U,0x84eaU,
		0x19dfU,0x33beU,0x677cU,0xcef8U,0x8dfbU,0x0bfdU,0x17faU,0x2ff4U,0x5fe8U,0xbfd0U,
		0x6fabU,0xdf56U,0xaea7U,0x4d45U,0x9a8aU,0x251fU,0x4a3eU,0x947cU,0x38f3U,0x71e6U,
		0xe3ccU,0xd793U,0xbf2dU,0x6e51U,0xdca2U,0xa94fU,0x4295U,0x852aU,0x1a5fU,0x34beU,
		0x697cU,0xd2f8U,0xb5fbU,0x7bfdU,0xf7faU,0xffffU,0xeff5U,0xcfe1U,0x8fc9U,0x0f99U,
		0x1f32U,0x3e64U,0x7cc8U,0xf990U,0xe32bU,0xd65dU,0xbcb1U,0x6969U,0xd2d2U,0xb5afU,
		0x7b55U,0xf6aaU,0xfd5fU,0xeab5U,0xc561U,0x9ac9U,0x2599U,0x4b32U,0x9664U,0x3cc3U,
		0x7986U,0xf30cU,0xf613U,0xfc2dU,0xe851U,0xc0a9U,0x9159U,0x32b9U,0x6572U,0xcae4U,
		0x85c3U,0x1b8dU,0x371aU,0x6e34U,0xdc68U,0xa8dbU,0x41bdU,0x837aU,0x16ffU,0x2dfeU,
		0x5bfcU,0xb7f8U,0x7ffbU,0xfff6U,0xefe7U,0xcfc5U,0x8f81U,0x0f09U,0x1e12U,0x3c24U,
		0x7848U,0xf090U,0xf12bU,0xf25dU,0xf4b1U,0xf969U,0xe2d9U,0xd5b9U,0xbb79U,0x66f9U,
		0xcdf2U,0x8befU,0x07d5U,0x0faaU,0x1f54U,0x3ea8U,0x7d50U,0xfaa0U,0xe54bU,0xda9dU,
		0xa531U,0x5a69U,0xb4d2U,0x79afU,0xf35eU,0xf6b7U,0xfd65U,0xeac1U,0xc589U,0x9b19U,
		0x2639U,0x4c72U,0x98e4U,0x21c3U,0x4386U,0x870cU,0x1e13U,0x3c26U,0x784cU,0xf098U,
		0xf13bU,0xf27dU,0xf4f1U,0xf9e9U,0xe3d9U,0xd7b9U,0xbf79U,0x6ef9U,0xddf2U,0xabefU,
		0x47d5U,0x8faaU,0x0f5fU,0x1ebeU,0x3d7cU,0x7af8U,0xf5f0U,0xfbebU,0xe7ddU,0xdfb1U,
		0xaf69U,0x4ed9U,0x9db2U,0x2b6fU,0x56deU,0xadbcU,0x4b73U,0x96e6U,0x3dc7U,0x7b8eU,
		0xf71cU,0xfe33U,0xec6dU,0xc8d1U,0x81a9U,0x1359U,0x26b2U,0x4d64U,0x9ac8U,0x259bU,
		0x4b36U,0x966cU,0x3cd3U,0x79a6U,0xf34cU,0xf693U,0xfd2dU,0xea51U,0xc4a9U,0x9959U,
		0x22b9U,0x4572U,0x8ae4U,0x05c3U,0x0b86U,0x170cU,0x2e18U,0x5c30U,0xb860U,0x60cbU,
		0xc196U,0x9327U,0x3645U,0x6c8aU,0xd914U,0xa223U,0x544dU,0xa89aU,0x413fU,0x827eU,
		0x14f7U,0x29eeU,0x53dcU,0xa7b8U,0x5f7bU,0xbef6U,0x6de7U,0xdbceU,0xa797U,0x5f25U,
		0xbe4aU,0x6c9fU,0xd93eU,0xa277U,0x54e5U,0xa9caU,0x439fU,0x873eU,0x1e77U,0x3ceeU,
		0x79dcU,0xf3b8U,0xf77bU,0xfefdU,0xedf1U,0xcbe9U,0x87d9U,0x1fb9U,0x3f72U,0x7ee4U,
		0xfdc8U,0xeb9bU,0xc73dU,0x9e71U,0x2ce9U,0x59d2U,0xb3a4U,0x7743U,0xee86U,0xcd07U,
		0x8a05U,0x0401U,0x0802U,0x1004U,0x2008U,0x4010U,0x8020U,0x104bU,0x2096U,0x412cU,
		0x8258U,0x14bbU,0x2976U,0x52ecU,0xa5d8U,0x5bbbU,0xb776U,0x7ee7U,0xfdceU,0xeb97U,
		0xc725U,0x9e41U,0x2c89U,0x5912U,0xb224U,0x7443U,0xe886U,0xc107U,0x9205U,0x3401U,
		0x6802U,0xd004U,0xb003U,0x700dU,0xe01aU,0xd03fU,0xb075U,0x70e1U,0xe1c2U,0xd38fU,
		0xb715U,0x7e21U,0xfc42U,0xe88fU,0xc115U,0x9221U,0x3449U,0x6892U,0xd124U,0xb243U,
		0x748dU,0xe91aU,0xc23fU,0x9475U,0x38e1U,0x71c2U,0xe384U,0xd703U,0xbe0dU,0x6c11U,
		0xd822U,0xa04fU,0x5095U,0xa12aU,0x525fU,0xa4beU,0x5977U,0xb2eeU,0x75d7U,0xebaeU,
		0xc757U,0x9ea5U,0x2d41U,0x5a82U,0xb504U,0x7a03U,0xf406U,0xf807U,0xe005U,0xd001U,
		0xb009U,0x7019U,0xe032U,0xd06fU,0xb0d5U,0x71a1U,0xe342U,0xd68fU,0xbd15U,0x6a21U,
		0xd442U,0xb88fU,0x6115U,0xc22aU,0x945fU,0x38b5U,0x716aU,0xe2d4U,0xd5a3U,0xbb4dU,
		0x6691U,0xcd22U,0x8a4fU,0x0495U,0x092aU,0x1254U,0x24a8U,0x4950U,0x92a0U,0x354bU,
		0x6a96U,0xd52cU,0xba53U,0x64adU,0xc95aU,0x82bfU,0x1575U,0x2aeaU,0x55d4U,0xaba8U,
		0x475bU,0x8eb6U,0x0d67U,0x1aceU,0x359cU,0x6b38U,0xd670U,0xbcebU,0x69ddU,0xd3baU,
		0xb77fU,0x7ef5U,0xfdeaU,0xebdfU,0xc7b5U,0x9f61U,0x2ec9U,0x5d92U,0xbb24U,0x6643U,
		0xcc86U,0x8907U,0x0205U,0x040aU,0x0814U,0x1028U,0x2050U,0x40a0U,0x8140U,0x128bU,
		0x2516U,0x4a2cU,0x9458U,0x38bbU,0x7176U,0xe2ecU,0xd5d3U,0xbbadU,0x6751U,0xcea2U,
		0x8d4fU,0x0a95U,0x152aU,0x2a54U,0x54a8U,0xa950U,0x42abU,0x8556U,0x1aa7U,0x354eU,
		0x6a9cU,0xd538U,0xba7bU,0x64fdU,0xc9faU,0x83ffU,0x17f5U,0x2feaU,0x5fd4U,0xbfa8U,
		0x6f5bU,0xdeb6U,0xad67U,0x4ac5U,0x958aU,0x3b1fU,0x763eU,0xec7cU,0xc8f3U,0x81edU,
		0x13d1U,0x27a2U,0x4f44U,0x9e88U,0x2d1bU,0x5a36U,0xb46cU,0x78d3U,0xf1a6U,0xf347U,
		0xf685U,0xfd01U,0xea09U,0xc419U,0x9839U,0x2079U,0x40f2U,0x81e4U,0x13c3U,0x2786U,
		0x4f0cU,0x9e18U,0x2c3bU,0x5876U,0xb0ecU,0x71d3U,0xe3a6U,0xd747U,0xbe85U,0x6d01U,
		0xda02U,0xa40fU,0x5815U,0xb02aU,0x705fU,0xe0beU,0xd177U,0xb2e5U,0x75c1U,0xeb82U,
		0xc70fU,0x9e15U,0x2c21U,0x5842U,0xb084U,0x7103U,0xe206U,0xd407U,0xb805U,0x6001U,
		0xc002U,0x900fU,0x3015U,0x602aU,0xc054U,0x90a3U,0x314dU,0x629aU,0xc534U,0x9a63U,
		0x24cdU,0x499aU,0x9334U,0x3663U,0x6cc6U,0xd98cU,0xa313U,0x562dU,0xac5aU,0x48bfU,
		0x917eU,0x32f7U,0x65eeU,0xcbdcU,0x87b3U,0x1f6dU,0x3edaU,0x7db4U,0xfb68U,0xe6dbU,
		0xddbdU,0xab71U,0x46e9U,0x8dd2U,0x0bafU,0x175eU,0x2ebcU,0x5d78U,0xbaf0U,0x65ebU,
		0xcbd6U,0x87a7U,0x1f45U,0x3e8aU,0x7d14U,0xfa28U,0xe45bU,0xd8bdU,0xa171U,0x52e9U,
		0xa5d2U,0x5bafU,0xb75eU,0x7eb7U,0xfd6eU,0xead7U,0xc5a5U,0x9b41U,0x2689U,0x4d12U,
		0x9a24U,0x2443U,0x4886U,0x910cU,0x3213U,0x6426U,0xc84cU,0x8093U,0x112dU,0x225aU,
		0x44b4U,0x8968U,0x02dbU,0x05b6U,0x0b6cU,0x16d8U,0x2db0U,0x5b60U,0xb6c0U,0x7d8bU,
		0xfb16U,0xe627U,0xdc45U,0xa881U,0x4109U,0x8212U,0x142fU,0x285eU,0x50bcU,0xa178U,
		0x52fbU,0xa5f6U,0x5be7U,0xb7ceU,0x7f97U,0xff2eU,0xee57U,0xcca5U,0x8941U,0x0289U,
		0x0512U,0x0a24U,0x1448U,0x2890U,0x5120U,0xa240U,0x548bU,0xa916U,0x4227U,0x844eU,
		0x1897U,0x312eU,0x625cU,0xc4b8U,0x997bU,0x22fdU,0x45faU,0x8bf4U,0x07e3U,0x0fc6U,
		0x1f8cU,0x3f18U,0x7e30U,0xfc60U,0xe8cbU,0xc19dU,0x9331U,0x3669U,0x6cd2U,0xd9a4U,
		0xa343U,0x568dU,0xad1aU,0x4a3fU,0x947eU,0x38f7U,0x71eeU,0xe3dcU,0xd7b3U,0xbf6dU,
		0x6ed1U,0xdda2U,0xab4fU,0x4695U,0x8d2aU,0x0a5fU,0x14beU,0x297cU,0x52f8U,0xa5f0U,
		0x5bebU,0xb7d6U,0x7fa7U,0xff4eU,0xee97U,0xcd25U,0x8a41U,0x0489U,0x0912U,0x1224U,
		0x2448U,0x4890U,0x9120U,0x324bU,0x6496U,0xc92cU,0x8253U,0x14adU,0x295aU,0x52b4U,
		0xa568U,0x5adbU,0xb5b6U,0x7b67U,0xf6ceU,0xfd97U,0xeb25U,0xc641U,0x9c89U,0x2919U,
		0x5232U,0xa464U,0x58c3U,0xb186U,0x7307U,0xe60eU,0xdc17U,0xa825U,0x4041U,0x8082U,
		0x110fU,0x221eU,0x443cU,0x8878U,0x00fbU,0x01f6U,0x03ecU,0x07d8U,0x0fb0U,0x1f60U,
		0x3ec0U,0x7d80U,0xfb00U,0xe60bU,0xdc1dU,0xa831U,0x4069U,0x80d2U,0x11afU,0x235eU,
		0x46bcU,0x8d78U,0x0afbU,0x15f6U,0x2becU,0x57d8U,0xafb0U,0x4f6bU,0x9ed6U,0x2da7U,
		0x5b4eU,0xb69cU,0x7d33U,0xfa66U,0xe4c7U,0xd985U,0xa301U,0x5609U,0xac12U,0x482fU,
		0x905eU,0x30b7U,0x616eU,0xc2dcU,0x95b3U,0x3b6dU,0x76daU,0xedb4U,0xcb63U,0x86cdU,
		0x1d91U,0x3b22U,0x7644U,0xec88U,0xc91bU,0x823dU,0x1471U,0x28e2U,0x51c4U,0xa388U,
		0x571bU,0xae36U,0x4c67U,0x98ceU,0x2197U,0x432eU,0x865cU,0x1cb3U,0x3966U,0x72ccU,
		0xe598U,0xdb3bU,0xa67dU,0x5cf1U,0xb9e2U,0x63cfU,0xc79eU,0x9f37U,0x2e65U,0x5ccaU,
		0xb994U,0x6323U,0xc646U,0x9c87U,0x2905U,0x520aU,0xa414U,0x5823U,0xb046U,0x7087U,
		0xe10eU,0xd217U,0xb425U,0x7841U,0xf082U,0xf10fU,0xf215U,0xf421U,0xf849U,0xe099U,
		0xd139U,0xb279U,0x74f9U,0xe9f2U,0xc3efU,0x97d5U,0x3fa1U,0x7f42U,0xfe84U,0xed03U,
		0xca0dU,0x8411U,0x1829U,0x3052U,0x60a4U,0xc148U,0x929bU,0x353dU,0x6a7aU,0xd4f4U,
		0xb9e3U,0x63cdU,0xc79aU,0x9f3fU,0x2e75U,0x5ceaU,0xb9d4U,0x63a3U,0xc746U,0x9e87U,
		0x2d05U,0x5a0aU,0xb414U,0x7823U,0xf046U,0xf087U,0xf105U,0xf201U,0xf409U,0xf819U,
		0xe039U,0xd079U,0xb0f9U,0x71f9U,0xe3f2U,0xd7efU,0xbfd5U,0x6fa1U,0xdf42U,0xae8fU,
		0x4d15U,0x9a2aU,0x245fU,0x48beU,0x917cU,0x32f3U,0x65e6U,0xcbccU,0x8793U,0x1f2dU,
		0x3e5aU,0x7cb4U,0xf968U,0xe2dbU,0xd5bdU,0xbb71U,0x66e9U,0xcdd2U,0x8bafU,0x0755U,
		0x0eaaU,0x1d54U,0x3aa8U,0x7550U,0xeaa0U,0xc54bU,0x9a9dU,0x2531U,0x4a62U,0x94c4U,
		0x3983U,0x7306U,0xe60cU,0xdc13U,0xa82dU,0x4051U,0x80a2U,0x114fU,0x229eU,0x453cU,
		0x8a78U,0x04fbU,0x09f6U,0x13ecU,0x27d8U,0x4fb0U,0x9f60U,0x2ecbU,0x5d96U,0xbb2cU,
		0x6653U,0xcca6U,0x8947U,0x0285U,0x050aU,0x0a14U,0x1428U,0x2850U,0x50a0U,0xa140U,
		0x528bU,0xa516U,0x5a27U,0xb44eU,0x7897U,0xf12eU,0xf257U,0xf4a5U,0xf941U,0xe289U,
		0xd519U,0xba39U,0x6479U,0xc8f2U,0x81efU,0x13d5U,0x27aaU,0x4f54U,0x9ea8U,0x2d5bU,
		0x5ab6U,0xb56cU,0x7ad3U,0xf5a6U,0xfb47U,0xe685U,0xdd01U,0xaa09U,0x4419U,0x8832U,
		0x006fU,0x00deU,0x01bcU,0x0378U,0x06f0U,0x0de0U,0x1bc0U,0x3780U,0x6f00U,0xde00U,
		0xac0bU,0x481dU,0x903aU,0x307fU,0x60feU,0xc1fcU,0x93f3U,0x37edU,0x6fdaU,0xdfb4U,
		0xaf63U,0x4ecdU,0x9d9aU,0x2b3fU,0x567eU,0xacfcU,0x49f3U,0x93e6U,0x37c7U,0x6f8eU,
		0xdf1cU,0xae33U,0x4c6dU,0x98daU,0x21bfU,0x437eU,0x86fcU,0x1df3U,0x3be6U,0x77ccU,
		0xef98U,0xcf3bU,0x8e7dU,0x0cf1U,0x19e2U,0x33c4U,0x6788U,0xcf10U,0x8e2bU,0x0c5dU,
		0x18baU,0x3174U,0x62e8U,0xc5d0U,0x9babU,0x275dU,0x4ebaU,0x9d74U,0x2ae3U,0x55c6U,
		0xab8cU,0x4713U,0x8e26U,0x0c47U,0x188eU,0x311cU,0x6238U,0xc470U,0x98ebU,0x21ddU,
		0x43baU,0x8774U,0x1ee3U,0x3dc6U,0x7b8cU,0xf718U,0xfe3bU,0xec7dU,0xc8f1U,0x81e9U,
		0x13d9U,0x27b2U,0x4f64U,0x9ec8U,0x2d9bU,0x5b36U,0xb66cU,0x7cd3U,0xf9a6U,0xe347U,
		0xd685U,0xbd01U,0x6a09U,0xd412U,0xb82fU,0x6055U,0xc0aaU,0x915fU,0x32b5U,0x656aU,
		0xcad4U,0x85a3U,0x1b4dU,0x369aU,0x6d34U,0xda68U,0xa4dbU,0x59bdU,0xb37aU,0x76ffU,
		0xedfeU,0xcbf7U,0x87e5U,0x1fc1U,0x3f82U,0x7f04U,0xfe08U,0xec1bU,0xc83dU,0x8071U,
		0x10e9U,0x21d2U,0x43a4U,0x8748U,0x1e9bU,0x3d36U,0x7a6cU,0xf4d8U,0xf9bbU,0xe37dU,
		0xd6f1U,0xbde9U,0x6bd9U,0xd7b2U,0xbf6fU,0x6ed5U,0xddaaU,0xab5fU,0x46b5U,0x8d6aU,
		0x0adfU,0x15beU,0x2b7cU,0x56f8U,0xadf0U,0x4bebU,0x97d6U,0x3fa7U,0x7f4eU,0xfe9cU,
		0xed33U,0xca6dU,0x84d1U,0x19a9U,0x3352U,0x66a4U,0xcd48U,0x8a9bU,0x053dU,0x0a7aU,
		0x14f4U,0x29e8U,0x53d0U,0xa7a0U,0x5f4bU,0xbe96U,0x6d27U,0xda4eU,0xa497U,0x5925U,
		0xb24aU,0x749fU,0xe93eU,0xc277U,0x94e5U,0x39c1U,0x7382U,0xe704U,0xde03U,0xac0dU,
		0x4811U,0x9022U,0x304fU,0x609eU,0xc13cU,0x9273U,0x34edU,0x69daU,0xd3b4U,0xb763U,
		0x7ecdU,0xfd9aU,0xeb3fU,0xc675U,0x9ce1U,0x29c9U,0x5392U,0xa724U,0x5e43U,0xbc86U,
		0x6907U,0xd20eU,0xb417U,0x7825U,0xf04aU,0xf09fU,0xf135U,0xf261U,0xf4c9U,0xf999U,
		0xe339U,0xd679U,0xbcf9U,0x69f9U,0xd3f2U,0xb7efU,0x7fd5U,0xffaaU,0xef5fU,0xceb5U,
		0x8d61U,0x0ac9U,0x1592U,0x2b24U,0x5648U,0xac90U,0x492bU,0x9256U,0x34a7U,0x694eU,
		0xd29cU,0xb533U,0x7a6dU,0xf4daU,0xf9bfU,0xe375U,0xd6e1U,0xbdc9U,0x6b99U,0xd732U,
		0xbe6fU,0x6cd5U,0xd9aaU,0xa35fU,0x56b5U,0xad6aU,0x4adfU,0x95beU,0x3b77U,0x76eeU,
		0xeddcU,0xcbb3U,0x876dU,0x1ed1U,0x3da2U,0x7b44U,0xf688U,0xfd1bU,0xea3dU,0xc471U,
		0x98e9U,0x21d9U,0x43b2U,0x8764U,0x1ec3U,0x3d86U,0x7b0cU,0xf618U,0xfc3bU,0xe87dU,
		0xc0f1U,0x91e9U,0x33d9U,0x67b2U,0xcf64U,0x8ec3U,0x0d8dU,0x1b1aU,0x3634U,0x6c68U,
		0xd8d0U,0xa1abU,0x535dU,0xa6baU,0x5d7fU,0xbafeU,0x65f7U,0xcbeeU,0x87d7U,0x1fa5U,
		0x3f4aU,0x7e94U,0xfd28U,0xea5bU,0xc4bdU,0x9971U,0x22e9U,0x45d2U,0x8ba4U,0x0743U,
		0x0e86U,0x1d0cU,0x3a18U,0x7430U,0xe860U,0xc0cbU,0x919dU,0x3331U,0x6662U,0xccc4U,
		0x8983U,0x030dU,0x061aU,0x0c34U,0x1868U,0x30d0U,0x61a0U,0xc340U,0x968bU,0x3d1dU,
		0x7a3aU,0xf474U,0xf8e3U,0xe1cdU,0xd391U,0xb729U,0x7e59U,0xfcb2U,0xe96fU,0xc2d5U,
		0x95a1U,0x3b49U,0x7692U,0xed24U,0xca43U,0x848dU,0x1911U,0x3222U,0x6444U,0xc888U,
		0x811bU,0x123dU,0x247aU,0x48f4U,0x91e8U,0x33dbU,0x67b6U,0xcf6cU,0x8ed3U,0x0dadU,
		0x1b5aU,0x36b4U,0x6d68U,0xdad0U,0xa5abU,0x5b5dU,0xb6baU,0x7d7fU,0xfafeU,0xe5f7U,
		0xdbe5U,0xa7c1U,0x5f89U,0xbf12U,0x6e2fU,0xdc5eU,0xa8b7U,0x4165U,0x82caU,0x159fU,
		0x2b3eU,0x567cU,0xacf8U,0x49fbU,0x93f6U,0x37e7U,0x6fceU,0xdf9cU,0xaf33U,0x4e6dU,
		0x9cdaU,0x29bfU,0x537eU,0xa6fcU,0x5df3U,0xbbe6U,0x67c7U,0xcf8eU,0x8f17U,0x0e25U,
		0x1c4aU,0x3894U,0x7128U,0xe250U,0xd4abU,0xb95dU,0x62b1U,0xc562U,0x9acfU,0x2595U,
		0x4b2aU,0x9654U,0x3ca3U,0x7946U,0xf28cU,0xf513U,0xfa2dU,0xe451U,0xd8a9U,0xa159U,
		0x52b9U,0xa572U,0x5aefU,0xb5deU,0x7bb7U,0xf76eU,0xfed7U,0xeda5U,0xcb41U,0x8689U,
		0x1d19U,0x3a32U,0x7464U,0xe8c8U,0xc19bU,0x933dU,0x3671U,0x6ce2U,0xd9c4U,0xa383U,
		0x570dU,0xae1aU,0x4c3fU,0x987eU,0x20f7U,0x41eeU,0x83dcU,0x17b3U,0x2f66U,0x5eccU,
		0xbd98U,0x6b3bU,0xd676U,0xbce7U,0x69c5U,0xd38aU,0xb71fU,0x7e35U,0xfc6aU,0xe8dfU,
		0xc1b5U,0x9361U,0x36c9U,0x6d92U,0xdb24U,0xa643U,0x5c8dU,0xb91aU,0x623fU,0xc47eU,
		0x98f7U,0x21e5U,0x43caU,0x8794U,0x1f23U,0x3e46U,0x7c8cU,0xf918U,0xe23bU,0xd47dU,
		0xb8f1U,0x61e9U,0xc3d2U,0x97afU,0x3f55U,0x7eaaU,0xfd54U,0xeaa3U,0xc54dU,0x9a91U,
		0x2529U,0x4a52U,0x94a4U,0x3943U,0x7286U,0xe50cU,0xda13U,0xa42dU,0x5851U,0xb0a2U,
		0x714fU,0xe29eU,0xd537U,0xba65U,0x64c1U,0xc982U,0x830fU,0x1615U,0x2c2aU,0x5854U,
		0xb0a8U,0x715bU,0xe2b6U,0xd567U,0xbac5U,0x6581U,0xcb02U,0x860fU,0x1c15U,0x382aU,
		0x7054U,0xe0a8U,0xd15bU,0xb2bdU,0x7571U,0xeae2U,0xc5cfU,0x9b95U,0x2721U,0x4e42U,
		0x9c84U,0x2903U,0x5206U,0xa40cU,0x5813U,0xb026U,0x7047U,0xe08eU,0xd117U,0xb225U,
		0x7441U,0xe882U,0xc10fU,0x9215U,0x3421U,0x6842U,0xd084U,0xb103U,0x720dU,0xe41aU,
		0xd83fU,0xa075U,0x50e1U,0xa1c2U,0x538fU,0xa71eU,0x5e37U,0xbc6eU,0x68d7U,0xd1aeU,
		0xb357U,0x76a5U,0xed4aU,0xca9fU,0x8535U,0x1a61U,0x34c2U,0x6984U,0xd308U,0xb61bU,
		0x7c3dU,0xf87aU,0xe0ffU,0xd1f5U,0xb3e1U,0x77c9U,0xef92U,0xcf2fU,0x8e55U,0x0ca1U,
		0x1942U,0x3284U,0x6508U,0xca10U,0x842bU,0x185dU,0x30baU,0x6174U,0xc2e8U,0x95dbU,
		0x3bbdU,0x777aU,0xeef4U,0xcde3U,0x8bcdU,0x0791U,0x0f22U,0x1e44U,0x3c88U,0x7910U,
		0xf220U,0xf44bU,0xf89dU,0xe131U,0xd269U,0xb4d9U,0x79b9U,0xf372U,0xf6efU,0xfdd5U,
		0xeba1U,0xc749U,0x9e99U,0x2d39U,0x5a72U,0xb4e4U,0x79c3U,0xf386U,0xf707U,0xfe05U,
		0xec01U,0xc809U,0x8019U,0x1039U,0x2072U,0x40e4U,0x81c8U,0x139bU,0x2736U,0x4e6cU,
		0x9cd8U,0x29bbU,0x5376U,0xa6ecU,0x5dd3U,0xbba6U,0x6747U,0xce8eU,0x8d17U,0x0a25U,
		0x144aU,0x2894U,0x5128U,0xa250U,0x54abU,0xa956U,0x42a7U,0x854eU,0x1a97U,0x352eU,
		0x6a5cU,0xd4b8U,0xb97bU,0x62fdU,0xc5faU,0x9bffU,0x27f5U,0x4feaU,0x9fd4U,0x2fa3U,
		0x5f46U,0xbe8cU,0x6d13U,0xda26U,0xa447U,0x5885U,0xb10aU,0x721fU,0xe43eU,0xd877U,
		0xa0e5U,0x51c1U,0xa382U,0x570fU,0xae1eU,0x4c37U,0x986eU,0x20d7U,0x41aeU,0x835cU,
		0x16b3U,0x2d66U,0x5accU,0xb598U,0x7b3bU,0xf676U,0xfce7U,0xe9c5U,0xc381U,0x9709U,
		0x3e19U,0x7c32U,0xf864U,0xe0c3U,0xd18dU,0xb311U,0x7629U,0xec52U,0xc8afU,0x8155U,
		0x12a1U,0x2542U,0x4a84U,0x9508U,0x3a1bU,0x7436U,0xe86cU,0xc0d3U,0x91adU,0x3351U,
		0x66a2U,0xcd44U,0x8a83U,0x050dU,0x0a1aU,0x1434U,0x2868U,0x50d0U,0xa1a0U,0x534bU,
		0xa696U,0x5d27U,0xba4eU,0x6497U,0xc92eU,0x8257U,0x14a5U,0x294aU,0x5294U,0xa528U,
		0x5a5bU,0xb4b6U,0x7967U,0xf2ceU,0xf597U,0xfb25U,0xe641U,0xdc89U,0xa919U,0x4239U,
		0x8472U,0x18efU,0x31deU,0x63bcU,0xc778U,0x9efbU,0x2dfdU,0x5bfaU,0xb7f4U,0x7fe3U,
		0xffc6U,0xef87U,0xcf05U,0x8e01U,0x0c09U,0x1812U,0x3024U,0x6048U,0xc090U,0x912bU,
		0x325dU,0x64baU,0xc974U,0x82e3U,0x15cdU,0x2b9aU,0x5734U,0xae68U,0x4cdbU,0x99b6U,
		0x2367U,0x46ceU,0x8d9cU,0x0b33U,0x1666U,0x2cccU,0x5998U,0xb330U,0x766bU,0xecd6U,
		0xc9a7U,0x8345U,0x1681U,0x2d02U,0x5a04U,0xb408U,0x781bU,0xf036U,0xf067U,0xf0c5U,
		0xf181U,0xf309U,0xf619U,0xfc39U,0xe879U,0xc0f9U,0x91f9U,0x33f9U,0x67f2U,0xcfe4U,
		0x8fc3U,0x0f8dU,0x1f1aU,0x3e34U,0x7c68U,0xf8d0U,0xe1abU,0xd35dU,0xb6b1U,0x7d69U,
		0xfad2U,0xe5afU,0xdb55U,0xa6a1U,0x5d49U,0xba92U,0x652fU,0xca5eU,0x84b7U,0x1965U,
		0x32caU,0x6594U,0xcb28U,0x865bU,0x1cbdU,0x397aU,0x72f4U,0xe5e8U,0xdbdbU,0xa7bdU,
		0x5f71U,0xbee2U,0x6dcfU,0xdb9eU,0xa737U,0x5e65U,0xbccaU,0x699fU,0xd33eU,0xb677U,
		0x7ce5U,0xf9caU,0xe39fU,0xd735U,0xbe61U,0x6cc9U,0xd992U,0xa32fU,0x5655U,0xacaaU,
		0x495fU,0x92beU,0x3577U,0x6aeeU,0xd5dcU,0xbbb3U,0x676dU,0xcedaU,0x8dbfU,0x0b75U,
		0x16eaU,0x2dd4U,0x5ba8U,0xb750U,0x7eabU,0xfd56U,0xeaa7U,0xc545U,0x9a81U,0x2509U,
		0x4a12U,0x9424U,0x3843U,0x7086U,0xe10cU,0xd213U,0xb42dU,0x7851U,0xf0a2U,0xf14fU,
		0xf295U,0xf521U,0xfa49U,0xe499U,0xd939U,0xa279U,0x54f9U,0xa9f2U,0x43efU,0x87deU,
		0x1fb7U,0x3f6eU,0x7edcU,0xfdb8U,0xeb7bU,0xc6fdU,0x9df1U,0x2be9U,0x57d2U,0xafa4U,
		0x4f43U,0x9e86U,0x2d07U,0x5a0eU,0xb41cU,0x7833U,0xf066U,0xf0c7U,0xf185U,0xf301U,
		0xf609U,0xfc19U,0xe839U,0xc079U,0x90f9U,0x31f9U,0x63f2U,0xc7e4U,0x9fc3U,0x2f8dU,
		0x5f1aU,0xbe34U,0x6c63U,0xd8c6U,0xa187U,0x5305U,0xa60aU,0x5c1fU,0xb83eU,0x6077U,
		0xc0eeU,0x91d7U,0x33a5U,0x674aU,0xce94U,0x8d23U,0x0a4dU,0x149aU,0x2934U,0x5268U,
		0xa4d0U,0x59abU,0xb356U,0x76a7U,0xed4eU,0xca97U,0x8525U,0x1a41U,0x3482U,0x6904U,
		0xd208U,0xb41bU,0x783dU,0xf07aU,0xf0ffU,0xf1f5U,0xf3e1U,0xf7c9U,0xff99U,0xef39U,
		0xce79U,0x8cf9U,0x09f9U,0x13f2U,0x27e4U,0x4fc8U,0x9f90U,0x2f2bU,0x5e56U,0xbcacU,
		0x6953U,0xd2a6U,0xb547U,0x7a85U,0xf50aU,0xfa1fU,0xe435U,0xd861U,0xa0c9U,0x5199U,
		0xa332U,0x566fU,0xacdeU,0x49b7U,0x936eU,0x36d7U,0x6daeU,0xdb5cU,0xa6b3U,0x5d6dU,
		0xbadaU,0x65bfU,0xcb7eU,0x86f7U,0x1de5U,0x3bcaU,0x7794U,0xef28U,0xce5bU,0x8cbdU,
		0x0971U,0x12e2U,0x25c4U,0x4b88U,0x9710U,0x3e2bU,0x7c56U,0xf8acU,0xe153U,0xd2adU,
		0xb551U,0x7aa9U,0xf552U,0xfaafU,0xe555U,0xdaa1U,0xa549U,0x5a99U,0xb532U,0x7a6fU,
		0xf4deU,0xf9b7U,0xe365U,0xd6c1U,0xbd89U,0x6b19U,0xd632U,0xbc6fU,0x68d5U,0xd1aaU,
		0xb35fU,0x76b5U,0xed6aU,0xcadfU,0x85b5U,0x1b61U,0x36c2U,0x6d84U,0xdb08U,0xa61bU,
		0x5c3dU,0xb87aU,0x60ffU,0xc1feU,0x93f7U,0x37e5U,0x6fcaU,0xdf94U,0xaf23U,0x4e4dU,
		0x9c9aU,0x293fU,0x527eU,0xa4fcU,0x59f3U,0xb3e6U,0x77c7U,0xef8eU,0xcf17U,0x8e25U,
		0x0c41U,0x1882U,0x3104U,0x6208U,0xc410U,0x982bU,0x205dU,0x40baU,0x8174U,0x12e3U,
		0x25c6U,0x4b8cU,0x9718U,0x3e3bU,0x7c76U,0xf8ecU,0xe1d3U,0xd3adU,0xb751U,0x7ea9U,
		0xfd52U,0xeaafU,0xc555U,0x9aa1U,0x2549U,0x4a92U,0x9524U,0x3a43U,0x7486U,0xe90cU,
		0xc213U,0x942dU,0x3851U,0x70a2U,0xe144U,0xd283U,0xb50dU,0x7a11U,0xf422U,0xf84fU,
		0xe095U,0xd121U,0xb249U,0x7499U,0xe932U,0xc26fU,0x94d5U,0x39a1U,0x7342U,0xe684U,
		0xdd03U,0xaa0dU,0x4411U,0x8822U,0x004fU,0x009eU,0x013cU,0x0278U,0x04f0U,0x09e0U,
		0x13c0U,0x2780U,0x4f00U,0x9e00U,0x2c0bU,0x5816U,0xb02cU,0x7053U,0xe0a6U,0xd147U,
		0xb285U,0x7501U,0xea02U,0xc40fU,0x9815U,0x2021U,0x4042U,0x8084U,0x1103U,0x2206U,
		0x440cU,0x8818U,0x003bU,0x0076U,0x00ecU,0x01d8U,0x03b0U,0x0760U,0x0ec0U,0x1d80U,
		0x3b00U,0x7600U,0xec00U,0xc80bU,0x801dU,0x1031U,0x2062U,0x40c4U,0x8188U,0x131bU,
		0x2636U,0x4c6cU,0x98d8U,0x21bbU,0x4376U,0x86ecU,0x1dd3U,0x3ba6U,0x774cU,0xee98U,
		0xcd3bU,0x8a7dU,0x04f1U,0x09e2U,0x13c4U,0x2788U,0x4f10U,0x9e20U,0x2c4bU,0x5896U,
		0xb12cU,0x7253U,0xe4a6U,0xd947U,0xa285U,0x5501U,0xaa02U,0x440fU,0x881eU,0x0037U,
		0x006eU,0x00dcU,0x01b8U,0x0370U,0x06e0U,0x0dc0U,0x1b80U,0x3700U,0x6e00U,0xdc00U,
		0xa80bU,0x401dU,0x803aU,0x107fU,0x20feU,0x41fcU,0x83f8U,0x17fbU,0x2ff6U,0x5fecU,
		0xbfd8U,0x6fbbU,0xdf76U,0xaee7U,0x4dc5U,0x9b8aU,0x271fU,0x4e3eU,0x9c7cU,0x28f3U,
		0x51e6U,0xa3ccU,0x5793U,0xaf26U,0x4e47U,0x9c8eU,0x2917U,0x522eU,0xa45cU,0x58b3U,
		0xb166U,0x72c7U,0xe58eU,0xdb17U,0xa625U,0x5c41U,0xb882U,0x610fU,0xc21eU,0x9437U,
		0x3865U,0x70caU,0xe194U,0xd323U,0xb64dU,0x7c91U,0xf922U,0xe24fU,0xd495U,0xb921U,
		0x6249U,0xc492U,0x992fU,0x2255U,0x44aaU,0x8954U,0x02a3U,0x0546U,0x0a8cU,0x1518U,
		0x2a30U,0x5460U,0xa8c0U,0x418bU,0x8316U,0x1627U,0x2c4eU,0x589cU,0xb138U,0x727bU,
		0xe4f6U,0xd9e7U,0xa3c5U,0x5781U,0xaf02U,0x4e0fU,0x9c1eU,0x2837U,0x506eU,0xa0dcU,
		0x51b3U,0xa366U,0x56c7U,0xad8eU,0x4b17U,0x962eU,0x3c57U,0x78aeU,0xf15cU,0xf2b3U,
		0xf56dU,0xfad1U,0xe5a9U,0xdb59U,0xa6b9U,0x5d79U,0xbaf2U,0x65efU,0xcbdeU,0x87b7U,
		0x1f65U,0x3ecaU,0x7d94U,0xfb28U,0xe65bU,0xdcbdU,0xa971U,0x42e9U,0x85d2U,0x1bafU,
		0x375eU,0x6ebcU,0xdd78U,0xaafbU,0x45fdU,0x8bfaU,0x07ffU,0x0ffeU,0x1ffcU,0x3ff8U,
		0x7ff0U,0xffe0U,0xefcbU,0xcf9dU,0x8f31U,0x0e69U,0x1cd2U,0x39a4U,0x7348U,0xe690U,
		0xdd2bU,0xaa5dU,0x44b1U,0x8962U,0x02cfU,0x059eU,0x0b3cU,0x1678U,0x2cf0U,0x59e0U,
		0xb3c0U,0x778bU,0xef16U,0xce27U,0x8c45U,0x0881U,0x1102U,0x2204U,0x4408U,0x8810U,
		0x002bU,0x0056U,0x00acU,0x0158U,0x02b0U,0x0560U,0x0ac0U,0x1580U,0x2b00U,0x5600U,
		0xac00U,0x480bU,0x9016U,0x3027U,0x604eU,0xc09cU,0x9133U,0x326dU,0x64daU,0xc9b4U,
		0x8363U,0x16cdU,0x2d9aU,0x5b34U,0xb668U,0x7cdbU,0xf9b6U,0xe367U,0xd6c5U,0xbd81U,
		0x6b09U,0xd612U,0xbc2fU,0x6855U,0xd0aaU,0xb15fU,0x72b5U,0xe56aU,0xdadfU,0xa5b5U,
		0x5b61U,0xb6c2U,0x7d8fU,0xfb1eU,0xe637U,0xdc65U,0xa8c1U,0x4189U,0x8312U,0x162fU,
		0x2c5eU,0x58bcU,0xb178U,0x72fbU,0xe5f6U,0xdbe7U,0xa7c5U,0x5f81U,0xbf02U,0x6e0fU,
		0xdc1eU,0xa837U,0x4065U,0x80caU,0x119fU,0x233eU,0x467cU,0x8cf8U,0x09fbU,0x13f6U,
		0x27ecU,0x4fd8U,0x9fb0U,0x2f6bU,0x5ed6U,0xbdacU,0x6b53U,0xd6a6U,0xbd47U,0x6a85U,
		0xd50aU,0xba1fU,0x6435U,0xc86aU,0x80dfU,0x11b5U,0x236aU,0x46d4U,0x8da8U,0x0b5bU,
		0x16b6U,0x2d6cU,0x5ad8U,0xb5b0U,0x7b6bU,0xf6d6U,0xfda7U,0xeb45U,0xc681U,0x9d09U,
		0x2a19U,0x5432U,0xa864U,0x40c3U,0x8186U,0x1307U,0x260eU,0x4c1cU,0x9838U,0x207bU,
		0x40f6U,0x81ecU,0x13d3U,0x27a6U,0x4f4cU,0x9e98U,0x2d3bU,0x5a76U,0xb4ecU,0x79d3U,
		0xf3a6U,0xf747U,0xfe85U,0xed01U,0xca09U,0x8419U,0x1839U,0x3072U,0x60e4U,0xc1c8U,
		0x939bU,0x373dU,0x6e7aU,0xdcf4U,0xa9e3U,0x43cdU,0x879aU,0x1f3fU,0x3e7eU,0x7cfcU,
		0xf9f8U,0xe3fbU,0xd7fdU,0xbff1U,0x6fe9U,0xdfd2U,0xafafU,0x4f55U,0x9eaaU,0x2d5fU,
		0x5abeU,0xb57cU,0x7af3U,0xf5e6U,0xfbc7U,0xe785U,0xdf01U,0xae09U,0x4c19U,0x9832U,
		0x206fU,0x40deU,0x81bcU,0x1373U,0x26e6U,0x4dccU,0x9b98U,0x273bU,0x4e76U,0x9cecU,
		0x29d3U,0x53a6U,0xa74cU,0x5e93U,0xbd26U,0x6a47U,0xd48eU,0xb917U,0x6225U,0xc44aU,
		0x989fU,0x2135U,0x426aU,0x84d4U,0x19a3U,0x3346U,0x668cU,0xcd18U,0x8a3bU,0x047dU,
		0x08faU,0x11f4U,0x23e8U,0x47d0U,0x8fa0U,0x0f4bU,0x1e96U,0x3d2cU,0x7a58U,0xf4b0U,
		0xf96bU,0xe2ddU,0xd5b1U,0xbb69U,0x66d9U,0xcdb2U,0x8b6fU,0x06d5U,0x0daaU,0x1b54U,
		0x36a8U,0x6d50U,0xdaa0U,0xa54bU,0x5a9dU,0xb53aU,0x7a7fU,0xf4feU,0xf9f7U,0xe3e5U,
		0xd7c1U,0xbf89U,0x6f19U,0xde32U,0xac6fU,0x48d5U,0x91aaU,0x335fU,0x66beU,0xcd7cU,
		0x8af3U,0x05edU,0x0bdaU,0x17b4U,0x2f68U,0x5ed0U,0xbda0U,0x6b4bU,0xd696U,0xbd27U,
		0x6a45U,0xd48aU,0xb91fU,0x6235U,0xc46aU,0x98dfU,0x21b5U,0x436aU,0x86d4U,0x1da3U,
		0x3b46U,0x768cU,0xed18U,0xca3bU,0x847dU,0x18f1U,0x31e2U,0x63c4U,0xc788U,0x9f1bU,
		0x2e3dU,0x5c7aU,0xb8f4U,0x61e3U,0xc3c6U,0x9787U,0x3f05U,0x7e0aU,0xfc14U,0xe823U,
		0xc04dU,0x9091U,0x3129U,0x6252U,0xc4a4U,0x9943U,0x228dU,0x451aU,0x8a34U,0x0463U,
		0x08c6U,0x118cU,0x2318U,0x4630U,0x8c60U,0x08cbU,0x1196U,0x232cU,0x4658U,0x8cb0U,
		0x096bU,0x12d6U,0x25acU,0x4b58U,0x96b0U,0x3d6bU,0x7ad6U,0xf5acU,0xfb53U,0xe6adU,
		0xdd51U,0xaaa9U,0x4559U,0x8ab2U,0x056fU,0x0adeU,0x15bcU,0x2b78U,0x56f0U,0xade0U,
		0x4bcbU,0x9796U,0x3f27U,0x7e4eU,0xfc9cU,0xe933U,0xc26dU,0x94d1U,0x39a9U,0x7352U,
		0xe6a4U,0xdd43U,0xaa8dU,0x4511U,0x8a22U,0x044fU,0x089eU,0x113cU,0x2278U,0x44f0U,
		0x89e0U,0x03cbU,0x0796U,0x0f2cU,0x1e58U,0x3cb0U,0x7960U,0xf2c0U,0xf58bU,0xfb1dU,
		0xe631U,0xdc69U,0xa8d9U,0x41b9U,0x8372U,0x16efU,0x2ddeU,0x5bbcU,0xb778U,0x7efbU,
		0xfdf6U,0xebe7U,0xc7c5U,0x9f81U,0x2f09U,0x5e12U,0xbc24U,0x6843U,0xd086U,0xb107U,
		0x7205U,0xe40aU,0xd81fU,0xa035U,0x5061U,0xa0c2U,0x518fU,0xa31eU,0x5637U,0xac6eU,
		0x48d7U,0x91aeU,0x3357U,0x66aeU,0xcd5cU,0x8ab3U,0x056dU,0x0adaU,0x15b4U,0x2b68U,
		0x56d0U,0xada0U,0x4b4bU,0x9696U,0x3d27U,0x7a4eU,0xf49cU,0xf933U,0xe26dU,0xd4d1U,
		0xb9a9U,0x6359U,0xc6b2U,0x9d6fU,0x2ad5U,0x55aaU,0xab54U,0x46a3U,0x8d46U,0x0a87U,
		0x150eU,0x2a1cU,0x5438U,0xa870U,0x40ebU,0x81d6U,0x13a7U,0x274eU,0x4e9cU,0x9d38U,
		0x2a7bU,0x54f6U,0xa9ecU,0x43d3U,0x87a6U,0x1f47U,0x3e8eU,0x7d1cU,0xfa38U,0xe47bU,
		0xd8fdU,0xa1f1U,0x53e9U,0xa7d2U,0x5fafU,0xbf5eU,0x6eb7U,0xdd6eU,0xaad7U,0x45a5U,
		0x8b4aU,0x069fU,0x0d3eU,0x1a7cU,0x34f8U,0x69f0U,0xd3e0U,0xb7cbU,0x7f9dU,0xff3aU,
		0xee7fU,0xccf5U,0x89e1U,0x03c9U,0x0792U,0x0f24U,0x1e48U,0x3c90U,0x7920U,0xf240U,
		0xf48bU,0xf91dU,0xe231U,0xd469U,0xb8d9U,0x61b9U,0xc372U,0x96efU,0x3dd5U,0x7baaU,
		0xf754U,0xfea3U,0xed4dU,0xca91U,0x8529U,0x1a59U,0x34b2U,0x6964U,0xd2c8U,0xb59bU,
		0x7b3dU,0xf67aU,0xfcffU,0xe9f5U,0xc3e1U,0x97c9U,0x3f99U,0x7f32U,0xfe64U,0xecc3U,
		0xc98dU,0x8311U,0x1629U,0x2c52U,0x58a4U,0xb148U,0x729bU,0xe536U,0xda67U,0xa4c5U,
		0x5981U,0xb302U,0x760fU,0xec1eU,0xc837U,0x8065U,0x10c1U,0x2182U,0x4304U,0x8608U,
		0x1c1bU,0x3836U,0x706cU,0xe0d8U,0xd1bbU,0xb37dU,0x76f1U,0xede2U,0xcbcfU,0x8795U,
		0x1f21U,0x3e42U,0x7c84U,0xf908U,0xe21bU,0xd43dU,0xb871U,0x60e9U,0xc1d2U,0x93afU,
		0x3755U,0x6eaaU,0xdd54U,0xaaa3U,0x454dU,0x8a9aU,0x053fU,0x0a7eU,0x14fcU,0x29f8U,
		0x53f0U,0xa7e0U,0x5fcbU,0xbf96U,0x6f27U,0xde4eU,0xac97U,0x4925U,0x924aU,0x349fU,
		0x693eU,0xd27cU,0xb4f3U,0x79edU,0xf3daU,0xf7bfU,0xff75U,0xeee1U,0xcdc9U,0x8b99U,
		0x0739U,0x0e72U,0x1ce4U,0x39c8U,0x7390U,0xe720U,0xde4bU,0xac9dU,0x4931U,0x9262U,
		0x34cfU,0x699eU,0xd33cU,0xb673U,0x7cedU,0xf9daU,0xe3bfU,0xd775U,0xbee1U,0x6dc9U,
		0xdb92U,0xa72fU,0x5e55U,0xbcaaU,0x695fU,0xd2beU,0xb577U,0x7ae5U,0xf5caU,0xfb9fU,
		0xe735U,0xde61U,0xacc9U,0x4999U,0x9332U,0x366fU,0x6cdeU,0xd9bcU,0xa373U,0x56edU,
		0xaddaU,0x4bbfU,0x977eU,0x3ef7U,0x7deeU,0xfbdcU,0xe7b3U,0xdf6dU,0xaed1U,0x4da9U,
		0x9b52U,0x26afU,0x4d5eU,0x9abcU,0x2573U,0x4ae6U,0x95ccU,0x3b93U,0x7726U,0xee4cU,
		0xcc93U,0x892dU,0x0251U,0x04a2U,0x0944U,0x1288U,0x2510U,0x4a20U,0x9440U,0x388bU,
		0x7116U,0xe22cU,0xd453U,0xb8adU,0x6151U,0xc2a2U,0x954fU,0x3a95U,0x752aU,0xea54U,
		0xc4a3U,0x994dU,0x2291U,0x4522U,0x8a44U,0x0483U,0x0906U,0x120cU,0x2418U,0x4830U,
		0x9060U,0x30cbU,0x6196U,0xc32cU,0x9653U,0x3cadU,0x795aU,0xf2b4U,0xf563U,0xfacdU,
		0xe591U,0xdb29U,0xa659U,0x5cb9U,0xb972U,0x62efU,0xc5deU,0x9bb7U,0x2765U,0x4ecaU,
		0x9d94U,0x2b23U,0x5646U,0xac8cU,0x4913U,0x9226U,0x3447U,0x688eU,0xd11cU,0xb233U,
		0x746dU,0xe8daU,0xc1bfU,0x9375U,0x36e1U,0x6dc2U,0xdb84U,0xa703U,0x5e0dU,0xbc1aU,
		0x683fU,0xd07eU,0xb0f7U,0x71e5U,0xe3caU,0xd79fU,0xbf35U,0x6e61U,0xdcc2U,0xa98fU,
		0x4315U,0x862aU,0x1c5fU,0x38beU,0x717cU,0xe2f8U,0xd5fbU,0xbbfdU,0x67f1U,0xcfe2U,
		0x8fcfU,0x0f95U,0x1f2aU,0x3e54U,0x7ca8U,0xf950U,0xe2abU,0xd55dU,0xbab1U,0x6569U,
		0xcad2U,0x85afU,0x1b55U,0x36aaU,0x6d54U,0xdaa8U,0xa55bU,0x5abdU,0xb57aU,0x7affU,
		0xf5feU,0xfbf7U,0xe7e5U,0xdfc1U,0xaf89U,0x4f19U,0x9e32U,0x2c6fU,0x58deU,0xb1bcU,
		0x7373U,0xe6e6U,0xddc7U,0xab85U,0x4701U,0x8e02U,0x0c0fU,0x181eU,0x303cU,0x6078U,
		0xc0f0U,0x91ebU,0x33ddU,0x67baU,0xcf74U,0x8ee3U,0x0dcdU,0x1b9aU,0x3734U,0x6e68U,
		0xdcd0U,0xa9abU,0x435dU,0x86baU,0x1d7fU,0x3afeU,0x75fcU,0xebf8U,0xc7fbU,0x9ffdU,
		0x2ff1U,0x5fe2U,0xbfc4U,0x6f83U,0xdf06U,0xae07U,0x4c05U,0x980aU,0x201fU,0x403eU,
		0x807cU,0x10f3U,0x21e6U,0x43ccU,0x8798U,0x1f3bU,0x3e76U,0x7cecU,0xf9d8U,0xe3bbU,
		0xd77dU,0xbef1U,0x6de9U,0xdbd2U,0xa7afU,0x5f55U,0xbeaaU,0x6d5fU,0xdabeU,0xa577U,
		0x5ae5U,0xb5caU,0x7b9fU,0xf73eU,0xfe77U,0xece5U,0xc9c1U,0x8389U,0x1719U,0x2e32U,
		0x5c64U,0xb8c8U,0x619bU,0xc336U,0x9667U,0x3cc5U,0x798aU,0xf314U,0xf623U,0xfc4dU,
		0xe891U,0xc129U,0x9259U,0x34b9U,0x6972U,0xd2e4U,0xb5c3U,0x7b8dU,0xf71aU,0xfe3fU,
		0xec75U,0xc8e1U,0x81c9U,0x1399U,0x2732U,0x4e64U,0x9cc8U,0x299bU,0x5336U,0xa66cU,
		0x5cd3U,0xb9a6U,0x6347U,0xc68eU,0x9d17U,0x2a25U,0x544aU,0xa894U,0x4123U,0x8246U,
		0x1487U,0x290eU,0x521cU,0xa438U,0x587bU,0xb0f6U,0x71e7U,0xe3ceU,0xd797U,0xbf25U,
		0x6e41U,0xdc82U,0xa90fU,0x4215U,0x842aU,0x185fU,0x30beU,0x617cU,0xc2f8U,0x95fbU,
		0x3bfdU,0x77faU,0xeff4U,0xcfe3U,0x8fcdU,0x0f91U,0x1f22U,0x3e44U,0x7c88U,0xf910U,
		0xe22bU,0xd45dU,0xb8b1U,0x6169U,0xc2d2U,0x95afU,0x3b55U,0x76aaU,0xed54U,0xcaa3U,
		0x854dU,0x1a91U,0x3522U,0x6a44U,0xd488U,0xb91bU,0x623dU,0xc47aU,0x98ffU,0x21f5U,
		0x43eaU,0x87d4U,0x1fa3U,0x3f46U,0x7e8cU,0xfd18U,0xea3bU,0xc47dU,0x98f1U,0x21e9U,
		0x43d2U,0x87a4U,0x1f43U,0x3e86U,0x7d0cU,0xfa18U,0xe43bU,0xd87dU,0xa0f1U,0x51e9U,
		0xa3d2U,0x57afU,0xaf5eU,0x4eb7U,0x9d6eU,0x2ad7U,0x55aeU,0xab5cU,0x46b3U,0x8d66U,
		0x0ac7U,0x158eU,0x2b1cU,0x5638U,0xac70U,0x48ebU,0x91d6U,0x33a7U,0x674eU,0xce9cU,
		0x8d33U,0x0a6dU,0x14daU,0x29b4U,0x5368U,0xa6d0U,0x5dabU,0xbb56U,0x66a7U,0xcd4eU,
		0x8a97U,0x0525U,0x0a4aU,0x1494U,0x2928U,0x5250U,0xa4a0U,0x594bU,0xb296U,0x7527U,
		0xea4eU,0xc497U,0x9925U,0x2241U,0x4482U,0x8904U,0x0203U,0x0406U,0x080cU,0x1018U,
		0x2030U,0x4060U,0x80c0U,0x118bU,0x2316U,0x462cU,0x8c58U,0x08bbU,0x1176U,0x22ecU,
		0x45d8U,0x8bb0U,0x076bU,0x0ed6U,0x1dacU,0x3b58U,0x76b0U,0xed60U,0xcacbU,0x859dU,
		0x1b31U,0x3662U,0x6cc4U,0xd988U,0xa31bU,0x563dU,0xac7aU,0x48ffU,0x91feU,0x33f7U,
		0x67eeU,0xcfdcU,0x8fb3U,0x0f6dU,0x1edaU,0x3db4U,0x7b68U,0xf6d0U,0xfdabU,0xeb5dU,
		0xc6b1U,0x9d69U,0x2ad9U,0x55b2U,0xab64U,0x46c3U,0x8d86U,0x0b07U,0x160eU,0x2c1cU,
		0x5838U,0xb070U,0x70ebU,0xe1d6U,0xd3a7U,0xb745U,0x7e81U,0xfd02U,0xea0fU,0xc415U,
		0x9821U,0x2049U,0x4092U,0x8124U,0x1243U,0x2486U,0x490cU,0x9218U,0x343bU,0x6876U,
		0xd0ecU,0xb1d3U,0x73adU,0xe75aU,0xdebfU,0xad75U,0x4ae1U,0x95c2U,0x3b8fU,0x771eU,
		0xee3cU,0xcc73U,0x88edU,0x01d1U,0x03a2U,0x0744U,0x0e88U,0x1d10U,0x3a20U,0x7440U,
		0xe880U,0xc10bU,0x921dU,0x3431U,0x6862U,0xd0c4U,0xb183U,0x730dU,0xe61aU,0xdc3fU,
		0xa875U,0x40e1U,0x81c2U,0x138fU,0x271eU,0x4e3cU,0x9c78U,0x28fbU,0x51f6U,0xa3ecU,
		0x57d3U,0xafa6U,0x4f47U,0x9e8eU,0x2d17U,0x5a2eU,0xb45cU,0x78b3U,0xf166U,0xf2c7U,
		0xf585U,0xfb01U,0xe609U,0xdc19U,0xa839U,0x4079U,0x80f2U,0x11efU,0x23deU,0x47bcU,
		0x8f78U,0x0efbU,0x1df6U,0x3becU,0x77d8U,0xefb0U,0xcf6bU,0x8eddU,0x0db1U,0x1b62U,
		0x36c4U,0x6d88U,0xdb10U,0xa62bU,0x5c5dU,0xb8baU,0x617fU,0xc2feU,0x95f7U,0x3be5U,
		0x77caU,0xef94U,0xcf23U,0x8e4dU,0x0c91U,0x1922U,0x3244U,0x6488U,0xc910U,0x822bU,
		0x145dU,0x28baU,0x5174U,0xa2e8U,0x55dbU,0xabb6U,0x4767U,0x8eceU,0x0d97U,0x1b2eU,
		0x365cU,0x6cb8U,0xd970U,0xa2ebU,0x55ddU,0xabbaU,0x477fU,0x8efeU,0x0df7U,0x1beeU,
		0x37dcU,0x6fb8U,0xdf70U,0xaeebU,0x4dddU,0x9bbaU,0x277fU,0x4efeU,0x9dfcU,0x2bf3U,
		0x57e6U,0xafccU,0x4f93U,0x9f26U,0x2e47U,0x5c8eU,0xb91cU,0x6233U,0xc466U,0x98c7U,
		0x2185U,0x430aU,0x8614U,0x1c23U,0x3846U,0x708cU,0xe118U,0xd23bU,0xb47dU,0x78f1U,
		0xf1e2U,0xf3cfU,0xf795U,0xff21U,0xee49U,0xcc99U,0x8939U,0x0279U,0x04f2U,0x09e4U,
		0x13c8U,0x2790U,0x4f20U,0x9e40U,0x2c8bU,0x5916U,0xb22cU,0x7453U,0xe8a6U,0xc147U,
		0x9285U,0x3501U,0x6a02U,0xd404U,0xb803U,0x600dU,0xc01aU,0x903fU,0x3075U,0x60eaU,
		0xc1d4U,0x93a3U,0x374dU,0x6e9aU,0xdd34U,0xaa63U,0x44cdU,0x899aU,0x033fU,0x067eU,
		0x0cfcU,0x19f8U,0x33f0U,0x67e0U,0xcfc0U,0x8f8bU,0x0f1dU,0x1e3aU,0x3c74U,0x78e8U,
		0xf1d0U,0xf3abU,0xf75dU,0xfeb1U,0xed69U,0xcad9U,0x85b9U,0x1b79U,0x36f2U,0x6de4U,
		0xdbc8U,0xa79bU,0x5f3dU,0xbe7aU,0x6cffU,0xd9feU,0xa3f7U,0x57e5U,0xafcaU,0x4f9fU,
		0x9f3eU,0x2e77U,0x5ceeU,0xb9dcU,0x63b3U,0xc766U,0x9ec7U,0x2d85U,0x5b0aU,0xb614U,
		0x7c23U,0xf846U,0xe087U,0xd105U,0xb201U,0x7409U,0xe812U,0xc02fU,0x9055U,0x30a1U,
		0x6142U,0xc284U,0x9503U,0x3a0dU,0x741aU,0xe834U,0xc063U,0x90cdU,0x3191U,0x6322U,
		0xc644U,0x9c83U,0x290dU,0x521aU,0xa434U,0x5863U,0xb0c6U,0x7187U,0xe30eU,0xd617U,
		0xbc25U,0x6841U,0xd082U,0xb10fU,0x7215U,0xe42aU,0xd85fU,0xa0b5U,0x5161U,0xa2c2U,
		0x558fU,0xab1eU,0x4637U,0x8c6eU,0x08d7U,0x11aeU,0x235cU,0x46b8U,0x8d70U,0x0aebU,
		0x15d6U,0x2bacU,0x5758U,0xaeb0U,0x4d6bU,0x9ad6U,0x25a7U,0x4b4eU,0x969cU,0x3d33U,
		0x7a66U,0xf4ccU,0xf993U,0xe32dU,0xd651U,0xbca9U,0x6959U,0xd2b2U,0xb56fU,0x7ad5U,
		0xf5aaU,0xfb5fU,0xe6b5U,0xdd61U,0xaac9U,0x4599U,0x8b32U,0x066fU,0x0cdeU,0x19bcU,
		0x3378U,0x66f0U,0xcde0U,0x8bcbU,0x079dU,0x0f3aU,0x1e74U,0x3ce8U,0x79d0U,0xf3a0U,
		0xf74bU,0xfe9dU,0xed31U,0xca69U,0x84d9U,0x19b9U,0x3372U,0x66e4U,0xcdc8U,0x8b9bU,
		0x073dU,0x0e7aU,0x1cf4U,0x39e8U,0x73d0U,0xe7a0U,0xdf4bU,0xae9dU,0x4d31U,0x9a62U,
		0x24cfU,0x499eU,0x933cU,0x3673U,0x6ce6U,0xd9ccU,0xa393U,0x572dU,0xae5aU,0x4cbfU,
		0x997eU,0x22f7U,0x45eeU,0x8bdcU,0x07b3U,0x0f66U,0x1eccU,0x3d98U,0x7b30U,0xf660U,
		0xfccbU,0xe99dU,0xc331U,0x9669U,0x3cd9U,0x79b2U,0xf364U,0xf6c3U,0xfd8dU,0xeb11U,
		0xc629U,0x9c59U,0x28b9U,0x5172U,0xa2e4U,0x55c3U,0xab86U,0x4707U,0x8e0eU,0x0c17U,
		0x182eU,0x305cU,0x60b8U,0xc170U,0x92ebU,0x35ddU,0x6bbaU,0xd774U,0xbee3U,0x6dcdU,
		0xdb9aU,0xa73fU,0x5e75U,0xbceaU,0x69dfU,0xd3beU,0xb777U,0x7ee5U,0xfdcaU,0xeb9fU,
		0xc735U,0x9e61U,0x2cc9U,0x5992U,0xb324U,0x7643U,0xec86U,0xc907U,0x8205U,0x1401U,
		0x2802U,0x5004U,0xa008U,0x501bU,0xa036U,0x5067U,0xa0ceU,0x5197U,0xa32eU,0x5657U,
		0xacaeU,0x4957U,0x92aeU,0x3557U,0x6aaeU,0xd55cU,0xbab3U,0x656dU,0xcadaU,0x85bfU,
		0x1b75U,0x36eaU,0x6dd4U,0xdba8U,0xa75bU,0x5ebdU,0xbd7aU,0x6affU,0xd5feU,0xbbf7U,
		0x67e5U,0xcfcaU,0x8f9fU,0x0f35U,0x1e6aU,0x3cd4U,0x79a8U,0xf350U,0xf6abU,0xfd5dU,
		0xeab1U,0xc569U,0x9ad9U,0x25b9U,0x4b72U,0x96e4U,0x3dc3U,0x7b86U,0xf70cU,0xfe13U,
		0xec2dU,0xc851U,0x80a9U,0x1159U,0x22b2U,0x4564U,0x8ac8U,0x059bU,0x0b36U,0x166cU,
		0x2cd8U,0x59b0U,0xb360U,0x76cbU,0xed96U,0xcb27U,0x8645U,0x1c81U,0x3902U,0x7204U,
		0xe408U,0xd81bU,0xa03dU,0x5071U,0xa0e2U,0x51cfU,0xa39eU,0x5737U,0xae6eU,0x4cd7U,
		0x99aeU,0x2357U,0x46aeU,0x8d5cU,0x0ab3U,0x1566U,0x2accU,0x5598U,0xab30U,0x466bU,
		0x8cd6U,0x09a7U,0x134eU,0x269cU,0x4d38U,0x9a70U,0x24ebU,0x49d6U,0x93acU,0x3753U,
		0x6ea6U,0xdd4cU,0xaa93U,0x452dU,0x8a5aU,0x04bfU,0x097eU,0x12fcU,0x25f8U,0x4bf0U,
		0x97e0U,0x3fcbU,0x7f96U,0xff2cU,0xee53U,0xccadU,0x8951U,0x02a9U,0x0552U,0x0aa4U,
		0x1548U,0x2a90U,0x5520U,0xaa40U,0x448bU,0x8916U,0x0227U,0x044eU,0x089cU,0x1138U,
		0x2270U,0x44e0U,0x89c0U,0x038bU,0x0716U,0x0e2cU,0x1c58U,0x38b0U,0x7160U,0xe2c0U,
		0xd58bU,0xbb1dU,0x6631U,0xcc62U,0x88cfU,0x0195U,0x032aU,0x0654U,0x0ca8U,0x1950U,
		0x32a0U,0x6540U,0xca80U,0x850bU,0x1a1dU,0x343aU,0x6874U,0xd0e8U,0xb1dbU,0x73bdU,
		0xe77aU,0xdeffU,0xadf5U,0x4be1U,0x97c2U,0x3f8fU,0x7f1eU,0xfe3cU,0xec73U,0xc8edU,
		0x81d1U,0x13a9U,0x2752U,0x4ea4U,0x9d48U,0x2a9bU,0x5536U,0xaa6cU,0x44d3U,0x89a6U,
		0x0347U,0x068eU,0x0d1cU,0x1a38U,0x3470U,0x68e0U,0xd1c0U,0xb38bU,0x771dU,0xee3aU,
		0xcc7fU,0x88f5U,0x01e1U,0x03c2U,0x0784U,0x0f08U,0x1e10U,0x3c20U,0x7840U,0xf080U,
		0xf10bU,0xf21dU,0xf431U,0xf869U,0xe0d9U,0xd1b9U,0xb379U,0x76f9U,0xedf2U,0xcbefU,
		0x87d5U,0x1fa1U,0x3f42U,0x7e84U,0xfd08U,0xea1bU,0xc43dU,0x9871U,0x20e9U,0x41d2U,
		0x83a4U,0x1743U,0x2e86U,0x5d0cU,0xba18U,0x643bU,0xc876U,0x80e7U,0x11c5U,0x238aU,
		0x4714U,0x8e28U,0x0c5bU,0x18b6U,0x316cU,0x62d8U,0xc5b0U,0x9b6bU,0x26ddU,0x4dbaU,
		0x9b74U,0x26e3U,0x4dc6U,0x9b8cU,0x2713U,0x4e26U,0x9c4cU,0x2893U,0x5126U,0xa24cU,
		0x5493U,0xa926U,0x4247U,0x848eU,0x1917U,0x322eU,0x645cU,0xc8b8U,0x817bU,0x12fdU,
		0x25faU,0x4bf4U,0x97e8U,0x3fdbU,0x7fb6U,0xff6cU,0xeed3U,0xcdadU,0x8b51U,0x06a9U,
		0x0d52U,0x1aa4U,0x3548U,0x6a90U,0xd520U,0xba4bU,0x649dU,0xc93aU,0x827fU,0x14f5U,
		0x29eaU,0x53d4U,0xa7a8U,0x5f5bU,0xbeb6U,0x6d67U,0xdaceU,0xa597U,0x5b25U,0xb64aU,
		0x7c9fU,0xf93eU,0xe277U,0xd4e5U,0xb9c1U,0x6389U,0xc712U,0x9e2fU,0x2c55U,0x58aaU,
		0xb154U,0x72a3U,0xe546U,0xda87U,0xa505U,0x5a01U,0xb402U,0x780fU,0xf01eU,0xf037U,
		0xf065U,0xf0c1U,0xf189U,0xf319U,0xf639U,0xfc79U,0xe8f9U,0xc1f9U,0x93f9U,0x37f9U,
		0x6ff2U,0xdfe4U,0xafc3U,0x4f8dU,0x9f1aU,0x2e3fU,0x5c7eU,0xb8fcU,0x61f3U,0xc3e6U,
		0x97c7U,0x3f85U,0x7f0aU,0xfe14U,0xec23U,0xc84dU,0x8091U,0x1129U,0x2252U,0x44a4U,
		0x8948U,0x029bU,0x0536U,0x0a6cU,0x14d8U,0x29b0U,0x5360U,0xa6c0U,0x5d8bU,0xbb16U,
		0x6627U,0xcc4eU,0x8897U,0x0125U,0x024aU,0x0494U,0x0928U,0x1250U,0x24a0U,0x4940U,
		0x9280U,0x350bU,0x6a16U,0xd42cU,0xb853U,0x60adU,0xc15aU,0x92bfU,0x3575U,0x6aeaU,
		0xd5d4U,0xbba3U,0x674dU,0xce9aU,0x8d3fU,0x0a75U,0x14eaU,0x29d4U,0x53a8U,0xa750U,
		0x5eabU,0xbd56U,0x6aa7U,0xd54eU,0xba97U,0x6525U,0xca4aU,0x849fU,0x1935U,0x326aU,
		0x64d4U,0xc9a8U,0x835bU,0x16bdU,0x2d7aU,0x5af4U,0xb5e8U,0x7bdbU,0xf7b6U,0xff67U,
		0xeec5U,0xcd81U,0x8b09U,0x0619U,0x0c32U,0x1864U,0x30c8U,0x6190U,0xc320U,0x964bU,
		0x3c9dU,0x793aU,0xf274U,0xf4e3U,0xf9cdU,0xe391U,0xd729U,0xbe59U,0x6cb9U,0xd972U,
		0xa2efU,0x55d5U,0xabaaU,0x475fU,0x8ebeU,0x0d77U,0x1aeeU,0x35dcU,0x6bb8U,0xd770U,
		0xbeebU,0x6dddU,0xdbbaU,0xa77fU,0x5ef5U,0xbdeaU,0x6bdfU,0xd7beU,0xbf77U,0x6ee5U,
		0xddcaU,0xab9fU,0x4735U,0x8e6aU,0x0cdfU,0x19beU,0x337cU,0x66f8U,0xcdf0U,0x8bebU,
		0x07ddU,0x0fbaU,0x1f74U,0x3ee8U,0x7dd0U,0xfba0U,0xe74bU,0xde9dU,0xad31U,0x4a69U,
		0x94d2U,0x39afU,0x735eU,0xe6bcU,0xdd73U,0xaaedU,0x45d1U,0x8ba2U,0x074fU,0x0e9eU,
		0x1d3cU,0x3a78U,0x74f0U,0xe9e0U,0xc3cbU,0x979dU,0x3f31U,0x7e62U,0xfcc4U,0xe983U,
		0xc30dU,0x9611U,0x3c29U,0x7852U,0xf0a4U,0xf143U,0xf28dU,0xf511U,0xfa29U,0xe459U,
		0xd8b9U,0xa179U,0x52f9U,0xa5f2U,0x5befU,0xb7deU,0x7fb7U,0xff6eU,0xeed7U,0xcda5U,
		0x8b41U,0x0689U,0x0d12U,0x1a24U,0x3448U,0x6890U,0xd120U,0xb24bU,0x749dU,0xe93aU,
		0xc27fU,0x94f5U,0x39e1U,0x73c2U,0xe784U,0xdf03U,0xae0dU,0x4c11U,0x9822U,0x204fU,
		0x409eU,0x813cU,0x1273U,0x24e6U,0x49ccU,0x9398U,0x373bU,0x6e76U,0xdcecU,0xa9d3U,
		0x43adU,0x875aU,0x1ebfU,0x3d7eU,0x7afcU,0xf5f8U,0xfbfbU,0xe7fdU,0xdff1U,0xafe9U,
		0x4fd9U,0x9fb2U,0x2f6fU,0x5edeU,0xbdbcU,0x6b73U,0xd6e6U,0xbdc7U,0x6b85U,0xd70aU,
		0xbe1fU,0x6c35U,0xd86aU,0xa0dfU,0x51b5U,0xa36aU,0x56dfU,0xadbeU,0x4b77U,0x96eeU,
		0x3dd7U,0x7baeU,0xf75cU,0xfeb3U,0xed6dU,0xcad1U,0x85a9U,0x1b59U,0x36b2U,0x6d64U,
		0xdac8U,0xa59bU,0x5b3dU,0xb67aU,0x7cffU,0xf9feU,0xe3f7U,0xd7e5U,0xbfc1U,0x6f89U,
		0xdf12U,0xae2fU,0x4c55U,0x98aaU,0x215fU,0x42beU,0x857cU,0x1af3U,0x35e6U,0x6bccU,
		0xd798U,0xbf3bU,0x6e7dU,0xdcfaU,0xa9ffU,0x43f5U,0x87eaU,0x1fdfU,0x3fbeU,0x7f7cU,
		0xfef8U,0xedfbU,0xcbfdU,0x87f1U,0x1fe9U,0x3fd2U,0x7fa4U,0xff48U,0xee9bU,0xcd3dU,
		0x8a71U,0x04e9U,0x09d2U,0x13a4U,0x2748U,0x4e90U,0x9d20U,0x2a4bU,0x5496U,0xa92cU,
		0x4253U,0x84a6U,0x1947U,0x328eU,0x651cU,0xca38U,0x847bU,0x18fdU,0x31faU,0x63f4U,
		0xc7e8U,0x9fdbU,0x2fbdU,0x5f7aU,0xbef4U,0x6de3U,0xdbc6U,0xa787U,0x5f05U,0xbe0aU,
		0x6c1fU,0xd83eU,0xa077U,0x50e5U,0xa1caU,0x539fU,0xa73eU,0x5e77U,0xbceeU,0x69d7U,
		0xd3aeU,0xb757U,0x7ea5U,0xfd4aU,0xea9fU,0xc535U,0x9a61U,0x24c9U,0x4992U,0x9324U,
		0x3643U,0x6c86U,0xd90cU,0xa213U,0x542dU,0xa85aU,0x40bfU,0x817eU,0x12f7U,0x25eeU,
		0x4bdcU,0x97b8U,0x3f7bU,0x7ef6U,0xfdecU,0xebd3U,0xc7adU,0x9f51U,0x2ea9U,0x5d52U,
		0xbaa4U,0x6543U,0xca86U,0x8507U,0x1a05U,0x340aU,0x6814U,0xd028U,0xb05bU,0x70bdU,
		0xe17aU,0xd2ffU,0xb5f5U,0x7be1U,0xf7c2U,0xff8fU,0xef15U,0xce21U,0x8c49U,0x0899U,
		0x1132U,0x2264U,0x44c8U,0x8990U,0x032bU,0x0656U,0x0cacU,0x1958U,0x32b0U,0x6560U,
		0xcac0U,0x858bU,0x1b1dU,0x363aU,0x6c74U,0xd8e8U,0xa1dbU,0x53bdU,0xa77aU,0x5effU,
		0xbdfeU,0x6bf7U,0xd7eeU,0xbfd7U,0x6fa5U,0xdf4aU,0xae9fU,0x4d35U,0x9a6aU,0x24dfU,
		0x49beU,0x937cU,0x36f3U,0x6de6U,0xdbccU,0xa793U,0x5f2dU,0xbe5aU,0x6cbfU,0xd97eU,
		0xa2f7U,0x55e5U,0xabcaU,0x479fU,0x8f3eU,0x0e77U,0x1ceeU,0x39dcU,0x73b8U,0xe770U,
		0xdeebU,0xadddU,0x4bb1U,0x9762U,0x3ecfU,0x7d9eU,0xfb3cU,0xe673U,0xdcedU,0xa9d1U,
		0x43a9U,0x8752U,0x1eafU,0x3d5eU,0x7abcU,0xf578U,0xfafbU,0xe5fdU,0xdbf1U,0xa7e9U,
		0x5fd9U,0xbfb2U,0x6f6fU,0xdedeU,0xadb7U,0x4b65U,0x96caU,0x3d9fU,0x7b3eU,0xf67cU,
		0xfcf3U,0xe9edU,0xc3d1U,0x97a9U,0x3f59U,0x7eb2U,0xfd64U,0xeac3U,0xc58dU,0x9b11U,
		0x2629U,0x4c52U,0x98a4U,0x2143U,0x4286U,0x850cU,0x1a13U,0x3426U,0x684cU,0xd098U,
		0xb13bU,0x727dU,0xe4faU,0xd9ffU,0xa3f5U,0x57e1U,0xafc2U,0x4f8fU,0x9f1eU,0x2e37U,
		0x5c6eU,0xb8dcU,0x61b3U,0xc366U,0x96c7U,0x3d85U,0x7b0aU,0xf614U,0xfc23U,0xe84dU,
		0xc091U,0x9129U,0x3259U,0x64b2U,0xc964U,0x82c3U,0x158dU,0x2b1aU,0x5634U,0xac68U,
		0x48dbU,0x91b6U,0x3367U,0x66ceU,0xcd9cU,0x8b33U,0x066dU,0x0cdaU,0x19b4U,0x3368U,
		0x66d0U,0xcda0U,0x8b4bU,0x069dU,0x0d3aU,0x1a74U,0x34e8U,0x69d0U,0xd3a0U,0xb74bU,
		0x7e9dU,0xfd3aU,0xea7fU,0xc4f5U,0x99e1U,0x23c9U,0x4792U,0x8f24U,0x0e43U,0x1c86U,
		0x390cU,0x7218U,0xe430U,0xd86bU,0xa0ddU,0x51b1U,0xa362U,0x56cfU,0xad9eU,0x4b37U,
		0x966eU,0x3cd7U,0x79aeU,0xf35cU,0xf6b3U,0xfd6dU,0xead1U,0xc5a9U,0x9b59U,0x26b9U,
		0x4d72U,0x9ae4U,0x25c3U,0x4b86U,0x970cU,0x3e13U,0x7c26U,0xf84cU,0xe093U,0xd12dU,
		0xb251U,0x74a9U,0xe952U,0xc2afU,0x9555U,0x3aa1U,0x7542U,0xea84U,0xc503U,0x9a0dU,
		0x2411U,0x4822U,0x9044U,0x3083U,0x6106U,0xc20cU,0x9413U,0x382dU,0x705aU,0xe0b4U,
		0xd163U,0xb2cdU,0x7591U,0xeb22U,0xc64fU,0x9c95U,0x2921U,0x5242U,0xa484U,0x5903U,
		0xb206U,0x7407U,0xe80eU,0xc017U,0x9025U,0x3041U,0x6082U,0xc104U,0x9203U,0x340dU,
		0x681aU,0xd034U,0xb063U,0x70cdU,0xe19aU,0xd33fU,0xb675U,0x7ce1U,0xf9c2U,0xe38fU,
		0xd715U,0xbe21U,0x6c49U,0xd892U,0xa12fU,0x5255U,0xa4aaU,0x595fU,0xb2beU,0x7577U,
		0xeaeeU,0xc5d7U,0x9ba5U,0x2741U,0x4e82U,0x9d04U,0x2a03U,0x5406U,0xa80cU,0x4013U,
		0x8026U,0x1047U,0x208eU,0x411cU,0x8238U,0x147bU,0x28f6U,0x51ecU,0xa3d8U,0x57bbU,
		0xaf76U,0x4ee7U,0x9dceU,0x2b97U,0x572eU,0xae5cU,0x4cb3U,0x9966U,0x22c7U,0x458eU,
		0x8b1cU,0x0633U,0x0c66U,0x18ccU,0x3198U,0x6330U,0xc660U,0x9ccbU,0x299dU,0x533aU,
		0xa674U,0x5ce3U,0xb9c6U,0x6387U,0xc70eU,0x9e17U,0x2c25U,0x584aU,0xb094U,0x7123U,
		0xe246U,0xd487U,0xb905U,0x6201U,0xc402U,0x980fU,0x2015U,0x402aU,0x8054U,0x10a3U,
		0x2146U,0x428cU,0x8518U,0x1a3bU,0x3476U,0x68ecU,0xd1d8U,0xb3bbU,0x777dU,0xeefaU,
		0xcdffU,0x8bf5U,0x07e1U,0x0fc2U,0x1f84U,0x3f08U,0x7e10U,0xfc20U,0xe84bU,0xc09dU,
		0x9131U,0x3269U,0x64d2U,0xc9a4U,0x8343U,0x168dU,0x2d1aU,0x5a34U,0xb468U,0x78dbU,
		0xf1b6U,0xf367U,0xf6c5U,0xfd81U,0xeb09U,0xc619U,0x9c39U,0x2879U,0x50f2U,0xa1e4U,
		0x53c3U,0xa786U,0x5f07U,0xbe0eU,0x6c17U,0xd82eU,0xa057U,0x50a5U,0xa14aU,0x529fU,
		0xa53eU,0x5a77U,0xb4eeU,0x79d7U,0xf3aeU,0xf757U,0xfea5U,0xed41U,0xca89U,0x8519U,
		0x1a39U,0x3472U,0x68e4U,0xd1c8U,0xb39bU,0x773dU,0xee7aU,0xccffU,0x89f5U,0x03e1U,
		0x07c2U,0x0f84U,0x1f08U,0x3e10U,0x7c20U,0xf840U,0xe08bU,0xd11dU,0xb231U,0x7469U,
		0xe8d2U,0xc1afU,0x9355U,0x36a1U,0x6d42U,0xda84U,0xa503U,0x5a0dU,0xb41aU,0x783fU,
		0xf07eU,0xf0f7U,0xf1e5U,0xf3c1U,0xf789U,0xff19U,0xee39U,0xcc79U,0x88f9U,0x01f9U,
		0x03f2U,0x07e4U,0x0fc8U,0x1f90U,0x3f20U,0x7e40U,0xfc80U,0xe90bU,0xc21dU,0x9431U,
		0x3869U,0x70d2U,0xe1a4U,0xd343U,0xb68dU,0x7d11U,0xfa22U,0xe44fU,0xd895U,0xa121U,
		0x5249U,0xa492U,0x592fU,0xb25eU,0x74b7U,0xe96eU,0xc2d7U,0x95a5U,0x3b41U,0x7682U,
		0xed04U,0xca03U,0x840dU,0x1811U,0x3022U,0x6044U,0xc088U,0x911bU,0x323dU,0x647aU,
		0xc8f4U,0x81e3U,0x13cdU,0x279aU,0x4f34U,0x9e68U,0x2cdbU,0x59b6U,0xb36cU,0x76d3U,
		0xeda6U,0xcb47U,0x8685U,0x1d01U,0x3a02U,0x7404U,0xe808U,0xc01bU,0x903dU,0x3071U,
		0x60e2U,0xc1c4U,0x9383U,0x370dU,0x6e1aU,0xdc34U,0xa863U,0x40cdU,0x819aU,0x133fU,
		0x267eU,0x4cfcU,0x99f8U,0x23fbU,0x47f6U,0x8fecU,0x0fd3U,0x1fa6U,0x3f4cU,0x7e98U,
		0xfd30U,0xea6bU,0xc4ddU,0x99b1U,0x2369U,0x46d2U,0x8da4U,0x0b43U,0x1686U,0x2d0cU,
		0x5a18U,0xb430U,0x786bU,0xf0d6U,0xf1a7U,0xf345U,0xf681U,0xfd09U,0xea19U,0xc439U,
		0x9879U,0x20f9U,0x41f2U,0x83e4U,0x17c3U,0x2f86U,0x5f0cU,0xbe18U,0x6c3bU,0xd876U,
		0xa0e7U,0x51c5U,0xa38aU,0x571fU,0xae3eU,0x4c77U,0x98eeU,0x21d7U,0x43aeU,0x875cU,
		0x1eb3U,0x3d66U,0x7accU,0xf598U,0xfb3bU,0xe67dU,0xdcf1U,0xa9e9U,0x43d9U,0x87b2U,
		0x1f6fU,0x3edeU,0x7dbcU,0xfb78U,0xe6fbU,0xddfdU,0xabf1U,0x47e9U,0x8fd2U,0x0fafU,
		0x1f5eU,0x3ebcU,0x7d78U,0xfaf0U,0xe5ebU,0xdbddU,0xa7b1U,0x5f69U,0xbed2U,0x6dafU,
		0xdb5eU,0xa6b7U,0x5d65U,0xbacaU,0x659fU,0xcb3eU,0x8677U,0x1ce5U,0x39caU,0x7394U,
		0xe728U,0xde5bU,0xacbdU,0x4971U,0x92e2U,0x35cfU,0x6b9eU,0xd73cU,0xbe73U,0x6cedU,
		0xd9daU,0xa3bfU,0x5775U,0xaeeaU,0x4ddfU,0x9bbeU,0x2777U,0x4eeeU,0x9ddcU,0x2bb3U,
		0x5766U,0xaeccU,0x4d93U,0x9b26U,0x2647U,0x4c8eU,0x991cU,0x2233U,0x4466U,0x88ccU,
		0x0193U,0x0326U,0x064cU,0x0c98U,0x1930U,0x3260U,0x64c0U,0xc980U,0x830bU,0x161dU,
		0x2c3aU,0x5874U,0xb0e8U,0x71dbU,0xe3b6U,0xd767U,0xbec5U,0x6d81U,0xdb02U,0xa60fU,
		0x5c15U,0xb82aU,0x605fU,0xc0beU,0x9177U,0x32e5U,0x65caU,0xcb94U,0x8723U,0x1e4dU,
		0x3c9aU,0x7934U,0xf268U,0xf4dbU,0xf9bdU,0xe371U,0xd6e9U,0xbdd9U,0x6bb9U,0xd772U,
		0xbeefU,0x6dd5U,0xdbaaU,0xa75fU,0x5eb5U,0xbd6aU,0x6adfU,0xd5beU,0xbb77U,0x66e5U,
		0xcdcaU,0x8b9fU,0x0735U,0x0e6aU,0x1cd4U,0x39a8U,0x7350U,0xe6a0U,0xdd4bU,0xaa9dU,
		0x4531U,0x8a62U,0x04cfU,0x099eU,0x133cU,0x2678U,0x4cf0U,0x99e0U,0x23cbU,0x4796U,
		0x8f2cU,0x0e53U,0x1ca6U,0x394cU,0x7298U,0xe530U,0xda6bU,0xa4ddU,0x59b1U,0xb362U,
		0x76cfU,0xed9eU,0xcb37U,0x8665U,0x1cc1U,0x3982U,0x7304U,0xe608U,0xdc1bU,0xa83dU,
		0x4071U,0x80e2U,0x11cfU,0x239eU,0x473cU,0x8e78U,0x0cfbU,0x19f6U,0x33ecU,0x67d8U,
		0xcfb0U,0x8f6bU,0x0eddU,0x1dbaU,0x3b74U,0x76e8U,0xedd0U,0xcbabU,0x875dU,0x1eb1U,
		0x3d62U,0x7ac4U,0xf588U,0xfb1bU,0xe63dU,0xdc71U,0xa8e9U,0x41d9U,0x83b2U,0x176fU,
		0x2edeU,0x5dbcU,0xbb78U,0x66fbU,0xcdf6U,0x8be7U,0x07c5U,0x0f8aU,0x1f14U,0x3e28U,
		0x7c50U,0xf8a0U,0xe14bU,0xd29dU,0xb531U,0x7a69U,0xf4d2U,0xf9afU,0xe355U,0xd6a1U,
		0xbd49U,0x6a99U,0xd532U,0xba6fU,0x64d5U,0xc9aaU,0x835fU,0x16b5U,0x2d6aU,0x5ad4U,
		0xb5a8U,0x7b5bU,0xf6b6U,0xfd67U,0xeac5U,0xc581U,0x9b09U,0x2619U,0x4c32U,0x9864U,
		0x20c3U,0x4186U,0x830cU,0x1613U,0x2c26U,0x584cU,0xb098U,0x713bU,0xe276U,0xd4e7U,
		0xb9c5U,0x6381U,0xc702U,0x9e0fU,0x2c15U,0x582aU,0xb054U,0x70a3U,0xe146U,0xd287U,
		0xb505U,0x7a01U,0xf402U,0xf80fU,0xe015U,0xd021U,0xb049U,0x7099U,0xe132U,0xd26fU,
		0xb4d5U,0x79a1U,0xf342U,0xf68fU,0xfd15U,0xea21U,0xc449U,0x9899U,0x2139U,0x4272U,
		0x84e4U,0x19c3U,0x3386U,0x670cU,0xce18U,0x8c3bU,0x087dU,0x10faU,0x21f4U,0x43e8U,
		0x87d0U,0x1fabU,0x3f56U,0x7eacU,0xfd58U,0xeabbU,0xc57dU,0x9af1U,0x25e9U,0x4bd2U,
		0x97a4U,0x3f43U,0x7e86U,0xfd0cU,0xea13U,0xc42dU,0x9851U,0x20a9U,0x4152U,0x82a4U,
		0x1543U,0x2a86U,0x550cU,0xaa18U,0x443bU,0x8876U,0x00e7U,0x01ceU,0x039cU,0x0738U,
		0x0e70U,0x1ce0U,0x39c0U,0x7380U,0xe700U,0xde0bU,0xac1dU,0x4831U,0x9062U,0x30cfU,
		0x619eU,0xc33cU,0x9673U,0x3cedU,0x79daU,0xf3b4U,0xf763U,0xfecdU,0xed91U,0xcb29U,
		0x8659U,0x1cb9U,0x3972U,0x72e4U,0xe5c8U,0xdb9bU,0xa73dU,0x5e71U,0xbce2U,0x69cfU,
		0xd39eU,0xb737U,0x7e65U,0xfccaU,0xe99fU,0xc335U,0x9661U,0x3cc9U,0x7992U,0xf324U,
		0xf643U,0xfc8dU,0xe911U,0xc229U,0x9459U,0x38b9U,0x7172U,0xe2e4U,0xd5c3U,0xbb8dU,
		0x6711U,0xce22U,0x8c4fU,0x0895U,0x112aU,0x2254U,0x44a8U,0x8950U,0x02abU,0x0556U,
		0x0aacU,0x1558U,0x2ab0U,0x5560U,0xaac0U,0x458bU,0x8b16U,0x0627U,0x0c4eU,0x189cU,
		0x3138U,0x6270U,0xc4e0U,0x99cbU,0x239dU,0x473aU,0x8e74U,0x0ce3U,0x19c6U,0x338cU,
		0x6718U,0xce30U,0x8c6bU,0x08ddU,0x11baU,0x2374U,0x46e8U,0x8dd0U,0x0babU,0x1756U,
		0x2eacU,0x5d58U,0xbab0U,0x656bU,0xcad6U,0x85a7U,0x1b45U,0x368aU,0x6d14U,0xda28U,
		0xa45bU,0x58bdU,0xb17aU,0x72ffU,0xe5feU,0xdbf7U,0xa7e5U,0x5fc1U,0xbf82U,0x6f0fU,
		0xde1eU,0xac37U,0x4865U,0x90caU,0x319fU,0x633eU,0xc67cU,0x9cf3U,0x29edU,0x53daU,
		0xa7b4U,0x5f63U,0xbec6U,0x6d87U,0xdb0eU,0xa617U,0x5c25U,0xb84aU,0x609fU,0xc13eU,
		0x9277U,0x34e5U,0x69caU,0xd394U,0xb723U,0x7e4dU,0xfc9aU,0xe93fU,0xc275U,0x94e1U,
		0x39c9U,0x7392U,0xe724U,0xde43U,0xac8dU,0x4911U,0x9222U,0x344fU,0x689eU,0xd13cU,
		0xb273U,0x74edU,0xe9daU,0xc3bfU,0x9775U,0x3ee1U,0x7dc2U,0xfb84U,0xe703U,0xde0dU,
		0xac11U,0x4829U,0x9052U,0x30afU,0x615eU,0xc2bcU,0x9573U,0x3aedU,0x75daU,0xebb4U,
		0xc763U,0x9ecdU,0x2d91U,0x5b22U,0xb644U,0x7c83U,0xf906U,0xe207U,0xd405U,0xb801U,
		0x6009U,0xc012U,0x902fU,0x3055U,0x60aaU,0xc154U,0x92a3U,0x354dU,0x6a9aU,0xd534U,
		0xba63U,0x64cdU,0xc99aU,0x833fU,0x1675U,0x2ceaU,0x59d4U,0xb3a8U,0x775bU,0xeeb6U,
		0xcd67U,0x8ac5U,0x0581U,0x0b02U,0x1604U,0x2c08U,0x5810U,0xb020U,0x704bU,0xe096U,
		0xd127U,0xb245U,0x7481U,0xe902U,0xc20fU,0x9415U,0x3821U,0x7042U,0xe084U,0xd103U,
		0xb20dU,0x7411U,0xe822U,0xc04fU,0x9095U,0x3121U,0x6242U,0xc484U,0x9903U,0x220dU,
		0x441aU,0x8834U,0x0063U,0x00c6U,0x018cU,0x0318U,0x0630U,0x0c60U,0x18c0U,0x3180U,
		0x6300U,0xc600U,0x9c0bU,0x281dU,0x503aU,0xa074U,0x50e3U,0xa1c6U,0x5387U,0xa70eU,
		0x5e17U,0xbc2eU,0x6857U,0xd0aeU,0xb157U,0x72a5U,0xe54aU,0xda9fU,0xa535U,0x5a61U,
		0xb4c2U,0x798fU,0xf31eU,0xf637U,0xfc65U,0xe8c1U,0xc189U,0x9319U,0x3639U,0x6c72U,
		0xd8e4U,0xa1c3U,0x538dU,0xa71aU,0x5e3fU,0xbc7eU,0x68f7U,0xd1eeU,0xb3d7U,0x77a5U,
		0xef4aU,0xce9fU,0x8d35U,0x0a61U,0x14c2U,0x2984U,0x5308U,0xa610U,0x5c2bU,0xb856U,
		0x60a7U,0xc14eU,0x9297U,0x3525U,0x6a4aU,0xd494U,0xb923U,0x624dU,0xc49aU,0x993fU,
		0x2275U,0x44eaU,0x89d4U,0x03a3U,0x0746U,0x0e8cU,0x1d18U,0x3a30U,0x7460U,0xe8c0U,
		0xc18bU,0x931dU,0x3631U,0x6c62U,0xd8c4U,0xa183U,0x530dU,0xa61aU,0x5c3fU,0xb87eU,
		0x60f7U,0xc1eeU,0x93d7U,0x37a5U,0x6f4aU,0xde94U,0xad23U,0x4a4dU,0x949aU,0x393fU,
		0x727eU,0xe4fcU,0xd9f3U,0xa3edU,0x57d1U,0xafa2U,0x4f4fU,0x9e9eU,0x2d37U,0x5a6eU,
		0xb4dcU,0x79b3U,0xf366U,0xf6c7U,0xfd85U,0xeb01U,0xc609U,0x9c19U,0x2839U,0x5072U,
		0xa0e4U,0x51c3U,0xa386U,0x5707U,0xae0eU,0x4c17U,0x982eU,0x2057U,0x40aeU,0x815cU,
		0x12b3U,0x2566U,0x4accU,0x9598U,0x3b3bU,0x7676U,0xececU,0xc9d3U,0x83adU,0x1751U,
		0x2ea2U,0x5d44U,0xba88U,0x651bU,0xca36U,0x8467U,0x18c5U,0x318aU,0x6314U,0xc628U,
		0x9c5bU,0x28bdU,0x517aU,0xa2f4U,0x55e3U,0xabc6U,0x4787U,0x8f0eU,0x0e17U,0x1c2eU,
		0x385cU,0x70b8U,0xe170U,0xd2ebU,0xb5ddU,0x7bb1U,0xf762U,0xfecfU,0xed95U,0xcb21U,
		0x8649U,0x1c99U,0x3932U,0x7264U,0xe4c8U,0xd99bU,0xa33dU,0x5671U,0xace2U,0x49cfU,
		0x939eU,0x3737U,0x6e6eU,0xdcdcU,0xa9b3U,0x436dU,0x86daU,0x1dbfU,0x3b7eU,0x76fcU,
		0xedf8U,0xcbfbU,0x87fdU,0x1ff1U,0x3fe2U,0x7fc4U,0xff88U,0xef1bU,0xce3dU,0x8c71U,
		0x08e9U,0x11d2U,0x23a4U,0x4748U,0x8e90U,0x0d2bU,0x1a56U,0x34acU,0x6958U,0xd2b0U,
		0xb56bU,0x7addU,0xf5baU,0xfb7fU,0xe6f5U,0xdde1U,0xabc9U,0x4799U,0x8f32U,0x0e6fU,
		0x1cdeU,0x39bcU,0x7378U,0xe6f0U,0xddebU,0xabddU,0x47b1U,0x8f62U,0x0ecfU,0x1d9eU,
		0x3b3cU,0x7678U,0xecf0U,0xc9ebU,0x83ddU,0x17b1U,0x2f62U,0x5ec4U,0xbd88U,0x6b1bU,
		0xd636U,0xbc67U,0x68c5U,0xd18aU,0xb31fU,0x7635U,0xec6aU,0xc8dfU,0x81b5U,0x1361U,
		0x26c2U,0x4d84U,0x9b08U,0x261bU,0x4c36U,0x986cU,0x20d3U,0x41a6U,0x834cU,0x1693U,
		0x2d26U,0x5a4cU,0xb498U,0x793bU,0xf276U,0xf4e7U,0xf9c5U,0xe381U,0xd709U,0xbe19U,
		0x6c39U,0xd872U,0xa0efU,0x51d5U,0xa3aaU,0x575fU,0xaebeU,0x4d77U,0x9aeeU,0x25d7U,
		0x4baeU,0x975cU,0x3eb3U,0x7d66U,0xfaccU,0xe593U,0xdb2dU,0xa651U,0x5ca9U,0xb952U,
		0x62afU,0xc55eU,0x9ab7U,0x2565U,0x4acaU,0x9594U,0x3b23U,0x7646U,0xec8cU,0xc913U,
		0x822dU,0x1451U,0x28a2U,0x5144U,0xa288U,0x551bU,0xaa36U,0x4467U,0x88ceU,0x0197U,
		0x032eU,0x065cU,0x0cb8U,0x1970U,0x32e0U,0x65c0U,0xcb80U,0x870bU,0x1e1dU,0x3c3aU,
		0x7874U,0xf0e8U,0xf1dbU,0xf3bdU,0xf771U,0xfee9U,0xedd9U,0xcbb9U,0x8779U,0x1ef9U,
		0x3df2U,0x7be4U,0xf7c8U,0xff9bU,0xef3dU,0xce71U,0x8ce9U,0x09d9U,0x13b2U,0x2764U,
		0x4ec8U,0x9d90U,0x2b2bU,0x5656U,0xacacU,0x4953U,0x92a6U,0x3547U,0x6a8eU,0xd51cU,
		0xba33U,0x646dU,0xc8daU,0x81bfU,0x1375U,0x26eaU,0x4dd4U,0x9ba8U,0x275bU,0x4eb6U,
		0x9d6cU,0x2ad3U,0x55a6U,0xab4cU,0x4693U,0x8d26U,0x0a47U,0x148eU,0x291cU,0x5238U,
		0xa470U,0x58ebU,0xb1d6U,0x73a7U,0xe74eU,0xde97U,0xad25U,0x4a41U,0x9482U,0x390fU,
		0x721eU,0xe43cU,0xd873U,0xa0edU,0x51d1U,0xa3a2U,0x574fU,0xae9eU,0x4d37U,0x9a6eU,
		0x24d7U,0x49aeU,0x935cU,0x36b3U,0x6d66U,0xdaccU,0xa593U,0x5b2dU,0xb65aU,0x7cbfU,
		0xf97eU,0xe2f7U,0xd5e5U,0xbbc1U,0x6789U,0xcf12U,0x8e2fU,0x0c55U,0x18aaU,0x3154U,
		0x62a8U,0xc550U,0x9aabU,0x255dU,0x4abaU,0x9574U,0x3ae3U,0x75c6U,0xeb8cU,0xc713U,
		0x9e2dU,0x2c51U,0x58a2U,0xb144U,0x7283U,0xe506U,0xda07U,0xa405U,0x5801U,0xb002U,
		0x700fU,0xe01eU,0xd037U,0xb065U,0x70c1U,0xe182U,0xd30fU,0xb615U,0x7c21U,0xf842U,
		0xe08fU,0xd115U,0xb221U,0x7449U,0xe892U,0xc12fU,0x9255U,0x34a1U,0x6942U,0xd284U,
		0xb503U,0x7a0dU,0xf41aU,0xf83fU,0xe075U,0xd0e1U,0xb1c9U,0x7399U,0xe732U,0xde6fU,
		0xacd5U,0x49a1U,0x9342U,0x368fU,0x6d1eU,0xda3cU,0xa473U,0x58edU,0xb1daU,0x73bfU,
		0xe77eU,0xdef7U,0xade5U,0x4bc1U,0x9782U,0x3f0fU,0x7e1eU,0xfc3cU,0xe873U,0xc0edU,
		0x91d1U,0x33a9U,0x6752U,0xcea4U,0x8d43U,0x0a8dU,0x151aU,0x2a34U,0x5468U,0xa8d0U,
		0x41abU,0x8356U,0x16a7U,0x2d4eU,0x5a9cU,0xb538U,0x7a7bU,0xf4f6U,0xf9e7U,0xe3c5U,
		0xd781U,0xbf09U,0x6e19U,0xdc32U,0xa86fU,0x40d5U,0x81aaU,0x135fU,0x26beU,0x4d7cU,
		0x9af8U,0x25fbU,0x4bf6U,0x97ecU,0x3fd3U,0x7fa6U,0xff4cU,0xee93U,0xcd2dU,0x8a51U,
		0x04a9U,0x0952U,0x12a4U,0x2548U,0x4a90U,0x9520U,0x3a4bU,0x7496U,0xe92cU,0xc253U,
		0x94adU,0x3951U,0x72a2U,0xe544U,0xda83U,0xa50dU,0x5a11U,0xb422U,0x784fU,0xf09eU,
		0xf137U,0xf265U,0xf4c1U,0xf989U,0xe319U,0xd639U,0xbc79U,0x68f9U,0xd1f2U,0xb3efU,
		0x77d5U,0xefaaU,0xcf5fU,0x8eb5U,0x0d61U,0x1ac2U,0x3584U,0x6b08U,0xd610U,0xbc2bU,
		0x685dU,0xd0baU,0xb17fU,0x72f5U,0xe5eaU,0xdbdfU,0xa7b5U,0x5f61U,0xbec2U,0x6d8fU,
		0xdb1eU,0xa637U,0x5c65U,0xb8caU,0x619fU,0xc33eU,0x9677U,0x3ce5U,0x79caU,0xf394U,
		0xf723U,0xfe4dU,0xec91U,0xc929U,0x8259U,0x14b9U,0x2972U,0x52e4U,0xa5c8U,0x5b9bU,
		0xb736U,0x7e67U,0xfcceU,0xe997U,0xc325U,0x9641U,0x3c89U,0x7912U,0xf224U,0xf443U,
		0xf88dU,0xe111U,0xd229U,0xb459U,0x78b9U,0xf172U,0xf2efU,0xf5d5U,0xfba1U,0xe749U,
		0xde99U,0xad39U,0x4a79U,0x94f2U,0x39efU,0x73deU,0xe7bcU,0xdf73U,0xaeedU,0x4dd1U,
		0x9ba2U,0x274fU,0x4e9eU,0x9d3cU,0x2a73U,0x54e6U,0xa9ccU,0x4393U,0x8726U,0x1e47U,
		0x3c8eU,0x791cU,0xf238U,0xf47bU,0xf8fdU,0xe1f1U,0xd3e9U,0xb7d9U,0x7fb9U,0xff72U,
		0xeeefU,0xcdd5U,0x8ba1U,0x0749U,0x0e92U,0x1d24U,0x3a48U,0x7490U,0xe920U,0xc24bU,
		0x949dU,0x3931U,0x7262U,0xe4c4U,0xd983U,0xa30dU,0x5611U,0xac22U,0x484fU,0x909eU,
		0x3137U,0x626eU,0xc4dcU,0x99b3U,0x236dU,0x46daU,0x8db4U,0x0b63U,0x16c6U,0x2d8cU,
		0x5b18U,0xb630U,0x7c6bU,0xf8d6U,0xe1a7U,0xd345U,0xb681U,0x7d09U,0xfa12U,0xe42fU,
		0xd855U,0xa0a1U,0x5149U,0xa292U,0x552fU,0xaa5eU,0x44b7U,0x896eU,0x02d7U,0x05aeU,
		0x0b5cU,0x16b8U,0x2d70U,0x5ae0U,0xb5c0U,0x7b8bU,0xf716U,0xfe27U,0xec45U,0xc881U,
		0x8109U,0x1219U,0x2432U,0x4864U,0x90c8U,0x319bU,0x6336U,0xc66cU,0x9cd3U,0x29adU,
		0x535aU,0xa6b4U,0x5d63U,0xbac6U,0x6587U,0xcb0eU,0x8617U,0x1c25U,0x384aU,0x7094U,
		0xe128U,0xd25bU,0xb4bdU,0x7971U,0xf2e2U,0xf5cfU,0xfb95U,0xe721U,0xde49U,0xac99U,
		0x4939U,0x9272U,0x34efU,0x69deU,0xd3bcU,0xb773U,0x7eedU,0xfddaU,0xebbfU,0xc775U,
		0x9ee1U,0x2dc9U,0x5b92U,0xb724U,0x7e43U,0xfc86U,0xe907U,0xc205U,0x9401U,0x3809U,
		0x7012U,0xe024U,0xd043U,0xb08dU,0x7111U,0xe222U,0xd44fU,0xb895U,0x6121U,0xc242U,
		0x948fU,0x3915U,0x722aU,0xe454U,0xd8a3U,0xa14dU,0x5291U,0xa522U,0x5a4fU,0xb49eU,
		0x7937U,0xf26eU,0xf4d7U,0xf9a5U,0xe341U,0xd689U,0xbd19U,0x6a39U,0xd472U,0xb8efU,
		0x61d5U,0xc3aaU,0x975fU,0x3eb5U,0x7d6aU,0xfad4U,0xe5a3U,0xdb4dU,0xa691U,0x5d29U,
		0xba52U,0x64afU,0xc95eU,0x82b7U,0x1565U,0x2acaU,0x5594U,0xab28U,0x465bU,0x8cb6U,
		0x0967U,0x12ceU,0x259cU,0x4b38U,0x9670U,0x3cebU,0x79d6U,0xf3acU,0xf753U,0xfeadU,
		0xed51U,0xcaa9U,0x8559U,0x1ab9U,0x3572U,0x6ae4U,0xd5c8U,0xbb9bU,0x673dU,0xce7aU,
		0x8cffU,0x09f5U,0x13eaU,0x27d4U,0x4fa8U,0x9f50U,0x2eabU,0x5d56U,0xbaacU,0x6553U,
		0xcaa6U,0x8547U,0x1a85U,0x350aU,0x6a14U,0xd428U,0xb85bU,0x60bdU,0xc17aU,0x92ffU,
		0x35f5U,0x6beaU,0xd7d4U,0xbfa3U,0x6f4dU,0xde9aU,0xad3fU,0x4a75U,0x94eaU,0x39dfU,
		0x73beU,0xe77cU,0xdef3U,0xadedU,0x4bd1U,0x97a2U,0x3f4fU,0x7e9eU,0xfd3cU,0xea73U,
		0xc4edU,0x99d1U,0x23a9U,0x4752U,0x8ea4U,0x0d43U,0x1a86U,0x350cU,0x6a18U,0xd430U,
		0xb86bU,0x60ddU,0xc1baU,0x937fU,0x36f5U,0x6deaU,0xdbd4U,0xa7a3U,0x5f4dU,0xbe9aU,
		0x6d3fU,0xda7eU,0xa4f7U,0x59e5U,0xb3caU,0x779fU,0xef3eU,0xce77U,0x8ce5U,0x09c1U,
		0x1382U,0x2704U,0x4e08U,0x9c10U,0x282bU,0x5056U,0xa0acU,0x5153U,0xa2a6U,0x5547U,
		0xaa8eU,0x4517U,0x8a2eU,0x0457U,0x08aeU,0x115cU,0x22b8U,0x4570U,0x8ae0U,0x05cbU,
		0x0b96U,0x172cU,0x2e58U,0x5cb0U,0xb960U,0x62cbU,0xc596U,0x9b27U,0x2645U,0x4c8aU,
		0x9914U,0x2223U,0x4446U,0x888cU,0x0113U,0x0226U,0x044cU,0x0898U,0x1130U,0x2260U,
		0x44c0U,0x8980U,0x030bU,0x0616U,0x0c2cU,0x1858U,0x30b0U,0x6160U,0xc2c0U,0x958bU,
		0x3b1dU,0x763aU,0xec74U,0xc8e3U,0x81cdU,0x1391U,0x2722U,0x4e44U,0x9c88U,0x291bU,
		0x5236U,0xa46cU,0x58d3U,0xb1a6U,0x7347U,0xe68eU,0xdd17U,0xaa25U,0x4441U,0x8882U,
		0x010fU,0x021eU,0x043cU,0x0878U,0x10f0U,0x21e0U,0x43c0U,0x8780U,0x1f0bU,0x3e16U,
		0x7c2cU,0xf858U,0xe0bbU,0xd17dU,0xb2f1U,0x75e9U,0xebd2U,0xc7afU,0x9f55U,0x2ea1U,
		0x5d42U,0xba84U,0x6503U,0xca06U,0x8407U,0x1805U,0x300aU,0x6014U,0xc028U,0x905bU,
		0x30bdU,0x617aU,0xc2f4U,0x95e3U,0x3bcdU,0x779aU,0xef34U,0xce63U,0x8ccdU,0x0991U,
		0x1322U,0x2644U,0x4c88U,0x9910U,0x222bU,0x4456U,0x88acU,0x0153U,0x02a6U,0x054cU,
		0x0a98U,0x1530U,0x2a60U,0x54c0U,0xa980U,0x430bU,0x8616U,0x1c27U,0x384eU,0x709cU,
		0xe138U,0xd27bU,0xb4fdU,0x79f1U,0xf3e2U,0xf7cfU,0xff95U,0xef21U,0xce49U,0x8c99U,
		0x0939U,0x1272U,0x24e4U,0x49c8U,0x9390U,0x372bU,0x6e56U,0xdcacU,0xa953U,0x42adU,
		0x855aU,0x1abfU,0x357eU,0x6afcU,0xd5f8U,0xbbfbU,0x67fdU,0xcffaU,0x8fffU,0x0ff5U,
		0x1feaU,0x3fd4U,0x7fa8U,0xff50U,0xeeabU,0xcd5dU,0x8ab1U,0x0569U,0x0ad2U,0x15a4U,
		0x2b48U,0x5690U,0xad20U,0x4a4bU,0x9496U,0x3927U,0x724eU,0xe49cU,0xd933U,0xa26dU,
		0x54d1U,0xa9a2U,0x434fU,0x869eU,0x1d37U,0x3a6eU,0x74dcU,0xe9b8U,0xc37bU,0x96fdU,
		0x3df1U,0x7be2U,0xf7c4U,0xff83U,0xef0dU,0xce11U,0x8c29U,0x0859U,0x10b2U,0x2164U,
		0x42c8U,0x8590U,0x1b2bU,0x3656U,0x6cacU,0xd958U,0xa2bbU,0x557dU,0xaafaU,0x45ffU,
		0x8bfeU,0x07f7U,0x0feeU,0x1fdcU,0x3fb8U,0x7f70U,0xfee0U,0xedcbU,0xcb9dU,0x8731U,
		0x1e69U,0x3cd2U,0x79a4U,0xf348U,0xf69bU,0xfd3dU,0xea71U,0xc4e9U,0x99d9U,0x23b9U,
		0x4772U,0x8ee4U,0x0dc3U,0x1b86U,0x370cU,0x6e18U,0xdc30U,0xa86bU,0x40ddU,0x81baU,
		0x137fU,0x26feU,0x4dfcU,0x9bf8U,0x27fbU,0x4ff6U,0x9fecU,0x2fd3U,0x5fa6U,0xbf4cU,
		0x6e93U,0xdd26U,0xaa47U,0x4485U,0x890aU,0x021fU,0x043eU,0x087cU,0x10f8U,0x21f0U,
		0x43e0U,0x87c0U,0x1f8bU,0x3f16U,0x7e2cU,0xfc58U,0xe8bbU,0xc17dU,0x92f1U,0x35e9U,
		0x6bd2U,0xd7a4U,0xbf43U,0x6e8dU,0xdd1aU,0xaa3fU,0x4475U,0x88eaU,0x01dfU,0x03beU,
		0x077cU,0x0ef8U,0x1df0U,0x3be0U,0x77c0U,0xef80U,0xcf0bU,0x8e1dU,0x0c31U,0x1862U,
		0x30c4U,0x6188U,0xc310U,0x962bU,0x3c5dU,0x78baU,0xf174U,0xf2e3U,0xf5cdU,0xfb91U,
		0xe729U,0xde59U,0xacb9U,0x4979U,0x92f2U,0x35efU,0x6bdeU,0xd7bcU,0xbf73U,0x6eedU,
		0xdddaU,0xabbfU,0x4775U,0x8eeaU,0x0ddfU,0x1bbeU,0x377cU,0x6ef8U,0xddf0U,0xabebU,
		0x47ddU,0x8fbaU,0x0f7fU,0x1efeU,0x3dfcU,0x7bf8U,0xf7f0U,0xffebU,0xefddU,0xcfb1U,
		0x8f69U,0x0ed9U,0x1db2U,0x3b64U,0x76c8U,0xed90U,0xcb2bU,0x865dU,0x1cb1U,0x3962U,
		0x72c4U,0xe588U,0xdb1bU,0xa63dU,0x5c71U,0xb8e2U,0x61cfU,0xc39eU,0x9737U,0x3e65U,
		0x7ccaU,0xf994U,0xe323U,0xd64dU,0xbc91U,0x6929U,0xd252U,0xb4afU,0x7955U,0xf2aaU,
		0xf55fU,0xfab5U,0xe561U,0xdac9U,0xa599U,0x5b39U,0xb672U,0x7cefU,0xf9deU,0xe3b7U,
		0xd765U,0xbec1U,0x6d89U,0xdb12U,0xa62fU,0x5c55U,0xb8aaU,0x615fU,0xc2beU,0x9577U,
		0x3ae5U,0x75caU,0xeb94U,0xc723U,0x9e4dU,0x2c91U,0x5922U,0xb244U,0x7483U,0xe906U,
		0xc207U,0x9405U,0x3801U,0x7002U,0xe004U,0xd003U,0xb00dU,0x7011U,0xe022U,0xd04fU,
		0xb095U,0x7121U,0xe242U,0xd48fU,0xb915U,0x6221U,0xc442U,0x988fU,0x2115U,0x422aU,
		0x8454U,0x18a3U,0x3146U,0x628cU,0xc518U,0x9a3bU,0x247dU,0x48faU,0x91f4U,0x33e3U,
		0x67c6U,0xcf8cU,0x8f13U,0x0e2dU,0x1c5aU,0x38b4U,0x7168U,0xe2d0U,0xd5abU,0xbb5dU,
		0x66b1U,0xcd62U,0x8acfU,0x0595U,0x0b2aU,0x1654U,0x2ca8U,0x5950U,0xb2a0U,0x754bU,
		0xea96U,0xc527U,0x9a45U,0x2481U,0x4902U,0x9204U,0x3403U,0x6806U,0xd00cU,0xb013U,
		0x702dU,0xe05aU,0xd0bfU,0xb175U,0x72e1U,0xe5c2U,0xdb8fU,0xa715U,0x5e21U,0xbc42U,
		0x688fU,0xd11eU,0xb237U,0x7465U,0xe8caU,0xc19fU,0x9335U,0x3661U,0x6cc2U,0xd984U,
		0xa303U,0x560dU,0xac1aU,0x483fU,0x907eU,0x30f7U,0x61eeU,0xc3dcU,0x97b3U,0x3f6dU,
		0x7edaU,0xfdb4U,0xeb63U,0xc6cdU,0x9d91U,0x2b29U,0x5652U,0xaca4U,0x4943U,0x9286U,
		0x3507U,0x6a0eU,0xd41cU,0xb833U,0x606dU,0xc0daU,0x91bfU,0x3375U,0x66eaU,0xcdd4U,
		0x8ba3U,0x074dU,0x0e9aU,0x1d34U,0x3a68U,0x74d0U,0xe9a0U,0xc34bU,0x969dU,0x3d31U,
		0x7a62U,0xf4c4U,0xf983U,0xe30dU,0xd611U,0xbc29U,0x6859U,0xd0b2U,0xb16fU,0x72d5U,
		0xe5aaU,0xdb5fU,0xa6b5U,0x5d61U,0xbac2U,0x658fU,0xcb1eU,0x8637U,0x1c65U,0x38caU,
		0x7194U,0xe328U,0xd65bU,0xbcbdU,0x6971U,0xd2e2U,0xb5cfU,0x7b95U,0xf72aU,0xfe5fU,
		0xecb5U,0xc961U,0x82c9U,0x1599U,0x2b32U,0x5664U,0xacc8U,0x499bU,0x9336U,0x3667U,
		0x6cceU,0xd99cU,0xa333U,0x566dU,0xacdaU,0x49bfU,0x937eU,0x36f7U,0x6deeU,0xdbdcU,
		0xa7b3U,0x5f6dU,0xbedaU,0x6dbfU,0xdb7eU,0xa6f7U,0x5de5U,0xbbcaU,0x679fU,0xcf3eU,
		0x8e77U,0x0ce5U,0x19caU,0x3394U,0x6728U,0xce50U,0x8cabU,0x095dU,0x12baU,0x2574U,
		0x4ae8U,0x95d0U,0x3babU,0x7756U,0xeeacU,0xcd53U,0x8aadU,0x0551U,0x0aa2U,0x1544U,
		0x2a88U,0x5510U,0xaa20U,0x444bU,0x8896U,0x0127U,0x024eU,0x049cU,0x0938U,0x1270U,
		0x24e0U,0x49c0U,0x9380U,0x370bU,0x6e16U,0xdc2cU,0xa853U,0x40adU,0x815aU,0x12bfU,
		0x257eU,0x4afcU,0x95f8U,0x3bfbU,0x77f6U,0xefecU,0xcfd3U,0x8fadU,0x0f51U,0x1ea2U,
		0x3d44U,0x7a88U,0xf510U,0xfa2bU,0xe45dU,0xd8b1U,0xa169U,0x52d9U,0xa5b2U,0x5b6fU,
		0xb6deU,0x7db7U,0xfb6eU,0xe6d7U,0xdda5U,0xab41U,0x4689U,0x8d12U,0x0a2fU,0x145eU,
		0x28bcU,0x5178U,0xa2f0U,0x55ebU,0xabd6U,0x47a7U,0x8f4eU,0x0e97U,0x1d2eU,0x3a5cU,
		0x74b8U,0xe970U,0xc2ebU,0x95ddU,0x3bb1U,0x7762U,0xeec4U,0xcd83U,0x8b0dU,0x0611U,
		0x0c22U,0x1844U,0x3088U,0x6110U,0xc220U,0x944bU,0x389dU,0x713aU,0xe274U,0xd4e3U,
		0xb9cdU,0x6391U,0xc722U,0x9e4fU,0x2c95U,0x592aU,0xb254U,0x74a3U,0xe946U,0xc287U,
		0x9505U,0x3a01U,0x7402U,0xe804U,0xc003U,0x900dU,0x3011U,0x6022U,0xc044U,0x9083U,
		0x310dU,0x621aU,0xc434U,0x9863U,0x20cdU,0x419aU,0x8334U,0x1663U,0x2cc6U,0x598cU,
		0xb318U,0x763bU,0xec76U,0xc8e7U,0x81c5U,0x1381U,0x2702U,0x4e04U,0x9c08U,0x281bU,
		0x5036U,0xa06cU,0x50d3U,0xa1a6U,0x5347U,0xa68eU,0x5d17U,0xba2eU,0x6457U,0xc8aeU,
		0x8157U,0x12a5U,0x254aU,0x4a94U,0x9528U,0x3a5bU,0x74b6U,0xe96cU,0xc2d3U,0x95adU,
		0x3b51U,0x76a2U,0xed44U,0xca83U,0x850dU,0x1a11U,0x3422U,0x6844U,0xd088U,0xb11bU,
		0x723dU,0xe47aU,0xd8ffU,0xa1f5U,0x53e1U,0xa7c2U,0x5f8fU,0xbf1eU,0x6e37U,0xdc6eU,
		0xa8d7U,0x41a5U,0x834aU,0x169fU,0x2d3eU,0x5a7cU,0xb4f8U,0x79fbU,0xf3f6U,0xf7e7U,
		0xffc5U,0xef81U,0xcf09U,0x8e19U,0x0c39U,0x1872U,0x30e4U,0x61c8U,0xc390U,0x972bU,
		0x3e5dU,0x7cbaU,0xf974U,0xe2e3U,0xd5cdU,0xbb91U,0x6729U,0xce52U,0x8cafU,0x0955U,
		0x12aaU,0x2554U,0x4aa8U,0x9550U,0x3aabU,0x7556U,0xeaacU,0xc553U,0x9aadU,0x2551U,
		0x4aa2U,0x9544U,0x3a83U,0x7506U,0xea0cU,0xc413U,0x982dU,0x2051U,0x40a2U,0x8144U,
		0x1283U,0x2506U,0x4a0cU,0x9418U,0x383bU,0x7076U,0xe0ecU,0xd1d3U,0xb3adU,0x7751U,
		0xeea2U,0xcd4fU,0x8a95U,0x0521U,0x0a42U,0x1484U,0x2908U,0x5210U,0xa420U,0x584bU,
		0xb096U,0x7127U,0xe24eU,0xd497U,0xb925U,0x6241U,0xc482U,0x990fU,0x2215U,0x442aU,
		0x8854U,0x00a3U,0x0146U,0x028cU,0x0518U,0x0a30U,0x1460U,0x28c0U,0x5180U,0xa300U,
		0x560bU,0xac16U,0x4827U,0x904eU,0x3097U,0x612eU,0xc25cU,0x94b3U,0x396dU,0x72daU,
		0xe5b4U,0xdb63U,0xa6cdU,0x5d91U,0xbb22U,0x664fU,0xcc9eU,0x8937U,0x0265U,0x04caU,
		0x0994U,0x1328U,0x2650U,0x4ca0U,0x9940U,0x228bU,0x4516U,0x8a2cU,0x0453U,0x08a6U,
		0x114cU,0x2298U,0x4530U,0x8a60U,0x04cbU,0x0996U,0x132cU,0x2658U,0x4cb0U,0x9960U,
		0x22cbU,0x4596U,0x8b2cU,0x0653U,0x0ca6U,0x194cU,0x3298U,0x6530U,0xca60U,0x84cbU,
		0x199dU,0x333aU,0x6674U,0xcce8U,0x89dbU,0x03bdU,0x077aU,0x0ef4U,0x1de8U,0x3bd0U,
		0x77a0U,0xef40U,0xce8bU,0x8d1dU,0x0a31U,0x1462U,0x28c4U,0x5188U,0xa310U,0x562bU,
		0xac56U,0x48a7U,0x914eU,0x3297U,0x652eU,0xca5cU,0x84b3U,0x196dU,0x32daU,0x65b4U,
		0xcb68U,0x86dbU,0x1dbdU,0x3b7aU,0x76f4U,0xede8U,0xcbdbU,0x87bdU,0x1f71U,0x3ee2U,
		0x7dc4U,0xfb88U,0xe71bU,0xde3dU,0xac71U,0x48e9U,0x91d2U,0x33afU,0x675eU,0xcebcU,
		0x8d73U,0x0aedU,0x15daU,0x2bb4U,0x5768U,0xaed0U,0x4dabU,0x9b56U,0x26a7U,0x4d4eU,
		0x9a9cU,0x2533U,0x4a66U,0x94ccU,0x3993U,0x7326U,0xe64cU,0xdc93U,0xa92dU,0x4251U,
		0x84a2U,0x194fU,0x329eU,0x653cU,0xca78U,0x84fbU,0x19fdU,0x33faU,0x67f4U,0xcfe8U,
		0x8fdbU,0x0fbdU,0x1f7aU,0x3ef4U,0x7de8U,0xfbd0U,0xe7abU,0xdf5dU,0xaeb1U,0x4d69U,
		0x9ad2U,0x25afU,0x4b5eU,0x96bcU,0x3d73U,0x7ae6U,0xf5ccU,0xfb93U,0xe72dU,0xde51U,
		0xaca9U,0x4959U,0x92b2U,0x356fU,0x6adeU,0xd5bcU,0xbb73U,0x66edU,0xcddaU,0x8bbfU,
		0x0775U,0x0eeaU,0x1dd4U,0x3ba8U,0x7750U,0xeea0U,0xcd4bU,0x8a9dU,0x0531U,0x0a62U,
		0x14c4U,0x2988U,0x5310U,0xa620U,0x5c4bU,0xb896U,0x6127U,0xc24eU,0x9497U,0x3925U,
		0x724aU,0xe494U,0xd923U,0xa24dU,0x5491U,0xa922U,0x424fU,0x849eU,0x1937U,0x326eU,
		0x64dcU,0xc9b8U,0x837bU,0x16fdU,0x2dfaU,0x5bf4U,0xb7e8U,0x7fdbU,0xffb6U,0xef67U,
		0xcec5U,0x8d81U,0x0b09U,0x1612U,0x2c24U,0x5848U,0xb090U,0x712bU,0xe256U,0xd4a7U,
		0xb945U,0x6281U,0xc502U,0x9a0fU,0x2415U,0x482aU,0x9054U,0x30a3U,0x6146U,0xc28cU,
		0x9513U,0x3a2dU,0x745aU,0xe8b4U,0xc163U,0x92cdU,0x3591U,0x6b22U,0xd644U,0xbc83U,
		0x690dU,0xd21aU,0xb43fU,0x7875U,0xf0eaU,0xf1dfU,0xf3b5U,0xf761U,0xfec9U,0xed99U,
		0xcb39U,0x8679U,0x1cf9U,0x39f2U,0x73e4U,0xe7c8U,0xdf9bU,0xaf3dU,0x4e71U,0x9ce2U,
		0x29cfU,0x539eU,0xa73cU,0x5e73U,0xbce6U,0x69c7U,0xd38eU,0xb717U,0x7e25U,0xfc4aU,
		0xe89fU,0xc135U,0x9261U,0x34c9U,0x6992U,0xd324U,0xb643U,0x7c8dU,0xf91aU,0xe23fU,
		0xd475U,0xb8e1U,0x61c9U,0xc392U,0x972fU,0x3e55U,0x7caaU,0xf954U,0xe2a3U,0xd54dU,
		0xba91U,0x6529U,0xca52U,0x84afU,0x1955U,0x32aaU,0x6554U,0xcaa8U,0x855bU,0x1abdU,
		0x357aU,0x6af4U,0xd5e8U,0xbbdbU,0x67bdU,0xcf7aU,0x8effU,0x0df5U,0x1beaU,0x37d4U,
		0x6fa8U,0xdf50U,0xaeabU,0x4d5dU,0x9abaU,0x257fU,0x4afeU,0x95fcU,0x3bf3U,0x77e6U,
		0xefccU,0xcf93U,0x8f2dU,0x0e51U,0x1ca2U,0x3944U,0x7288U,0xe510U,0xda2bU,0xa45dU,
		0x58b1U,0xb162U,0x72cfU,0xe59eU,0xdb37U,0xa665U,0x5cc1U,0xb982U,0x630fU,0xc61eU,
		0x9c37U,0x2865U,0x50caU,0xa194U,0x5323U,0xa646U,0x5c87U,0xb90eU,0x6217U,0xc42eU,
		0x9857U,0x20a5U,0x414aU,0x8294U,0x1523U,0x2a46U,0x548cU,0xa918U,0x423bU,0x8476U,
		0x18e7U,0x31ceU,0x639cU,0xc738U,0x9e7bU,0x2cfdU,0x59faU,0xb3f4U,0x77e3U,0xefc6U,
		0xcf87U,0x8f05U,0x0e01U,0x1c02U,0x3804U,0x7008U,0xe010U,0xd02bU,0xb05dU,0x70b1U,
		0xe162U,0xd2cfU,0xb595U,0x7b21U,0xf642U,0xfc8fU,0xe915U,0xc221U,0x9449U,0x3899U,
		0x7132U,0xe264U,0xd4c3U,0xb98dU,0x6311U,0xc622U,0x9c4fU,0x2895U,0x512aU,0xa254U,
		0x54a3U,0xa946U,0x4287U,0x850eU,0x1a17U,0x342eU,0x685cU,0xd0b8U,0xb17bU,0x72fdU,
		0xe5faU,0xdbffU,0xa7f5U,0x5fe1U,0xbfc2U,0x6f8fU,0xdf1eU,0xae37U,0x4c65U,0x98caU,
		0x219fU,0x433eU,0x867cU,0x1cf3U,0x39e6U,0x73ccU,0xe798U,0xdf3bU,0xae7dU,0x4cf1U,
		0x99e2U,0x23cfU,0x479eU,0x8f3cU,0x0e73U,0x1ce6U,0x39ccU,0x7398U,0xe730U,0xde6bU,
		0xacddU,0x49b1U,0x9362U,0x36cfU,0x6d9eU,0xdb3cU,0xa673U,0x5cedU,0xb9daU,0x63bfU,
		0xc77eU,0x9ef7U,0x2de5U,0x5bcaU,0xb794U,0x7f23U,0xfe46U,0xec87U,0xc905U,0x8201U,
		0x1409U,0x2812U,0x5024U,0xa048U,0x509bU,0xa136U,0x5267U,0xa4ceU,0x5997U,0xb32eU,
		0x7657U,0xecaeU,0xc957U,0x82a5U,0x1541U,0x2a82U,0x5504U,0xaa08U,0x441bU,0x8836U,
		0x0067U,0x00ceU,0x019cU,0x0338U,0x0670U,0x0ce0U,0x19c0U,0x3380U,0x6700U,0xce00U,
		0x8c0bU,0x081dU,0x103aU,0x2074U,0x40e8U,0x81d0U,0x13abU,0x2756U,0x4eacU,0x9d58U,
		0x2abbU,0x5576U,0xaaecU,0x45d3U,0x8ba6U,0x0747U,0x0e8eU,0x1d1cU,0x3a38U,0x7470U,
		0xe8e0U,0xc1cbU,0x939dU,0x3731U,0x6e62U,0xdcc4U,0xa983U,0x430dU,0x861aU,0x1c3fU,
		0x387eU,0x70fcU,0xe1f8U,0xd3fbU,0xb7fdU,0x7ff1U,0xffe2U,0xefcfU,0xcf95U,0x8f21U,
		0x0e49U,0x1c92U,0x3924U,0x7248U,0xe490U,0xd92bU,0xa25dU,0x54b1U,0xa962U,0x42cfU,
		0x859eU,0x1b37U,0x366eU,0x6cdcU,0xd9b8U,0xa37bU,0x56fdU,0xadfaU,0x4bffU,0x97feU,
		0x3ff7U,0x7feeU,0xffdcU,0xefb3U,0xcf6dU,0x8ed1U,0x0da9U,0x1b52U,0x36a4U,0x6d48U,
		0xda90U,0xa52bU,0x5a5dU,0xb4baU,0x797fU,0xf2feU,0xf5f7U,0xfbe5U,0xe7c1U,0xdf89U,
		0xaf19U,0x4e39U,0x9c72U,0x28efU,0x51deU,0xa3bcU,0x5773U,0xaee6U,0x4dc7U,0x9b8eU,
		0x2717U,0x4e2eU,0x9c5cU,0x28b3U,0x5166U,0xa2ccU,0x5593U,0xab26U,0x4647U,0x8c8eU,
		0x0917U,0x122eU,0x245cU,0x48b8U,0x9170U,0x32ebU,0x65d6U,0xcbacU,0x8753U,0x1eadU,
		0x3d5aU,0x7ab4U,0xf568U,0xfadbU,0xe5bdU,0xdb71U,0xa6e9U,0x5dd9U,0xbbb2U,0x676fU,
		0xcedeU,0x8db7U,0x0b65U,0x16caU,0x2d94U,0x5b28U,0xb650U,0x7cabU,0xf956U,0xe2a7U,
		0xd545U,0xba81U,0x6509U,0xca12U,0x842fU,0x1855U,0x30aaU,0x6154U,0xc2a8U,0x955bU,
		0x3abdU,0x757aU,0xeaf4U,0xc5e3U,0x9bcdU,0x2791U,0x4f22U,0x9e44U,0x2c83U,0x5906U,
		0xb20cU,0x7413U,0xe826U,0xc047U,0x9085U,0x3101U,0x6202U,0xc404U,0x9803U,0x200dU,
		0x401aU,0x8034U,0x1063U,0x20c6U,0x418cU,0x8318U,0x163bU,0x2c76U,0x58ecU,0xb1d8U,
		0x73bbU,0xe776U,0xdee7U,0xadc5U,0x4b81U,0x9702U,0x3e0fU,0x7c1eU,0xf83cU,0xe073U,
		0xd0edU,0xb1d1U,0x73a9U,0xe752U,0xdeafU,0xad55U,0x4aa1U,0x9542U,0x3a8fU,0x751eU,
		0xea3cU,0xc473U,0x98edU,0x21d1U,0x43a2U,0x8744U,0x1e83U,0x3d06U,0x7a0cU,0xf418U,
		0xf83bU,0xe07dU,0xd0f1U,0xb1e9U,0x73d9U,0xe7b2U,0xdf6fU,0xaed5U,0x4da1U,0x9b42U,
		0x268fU,0x4d1eU,0x9a3cU,0x2473U,0x48e6U,0x91ccU,0x3393U,0x6726U,0xce4cU,0x8c93U,
		0x092dU,0x125aU,0x24b4U,0x4968U,0x92d0U,0x35abU,0x6b56U,0xd6acU,0xbd53U,0x6aadU,
		0xd55aU,0xbabfU,0x6575U,0xcaeaU,0x85dfU,0x1bb5U,0x376aU,0x6ed4U,0xdda8U,0xab5bU,
		0x46bdU,0x8d7aU,0x0affU,0x15feU,0x2bfcU,0x57f8U,0xaff0U,0x4febU,0x9fd6U,0x2fa7U,
		0x5f4eU,0xbe9cU,0x6d33U,0xda66U,0xa4c7U,0x5985U,0xb30aU,0x761fU,0xec3eU,0xc877U,
		0x80e5U,0x11c1U,0x2382U,0x4704U,0x8e08U,0x0c1bU,0x1836U,0x306cU,0x60d8U,0xc1b0U,
		0x936bU,0x36ddU,0x6dbaU,0xdb74U,0xa6e3U,0x5dcdU,0xbb9aU,0x673fU,0xce7eU,0x8cf7U,
		0x09e5U,0x13caU,0x2794U,0x4f28U,0x9e50U,0x2cabU,0x5956U,0xb2acU,0x7553U,0xeaa6U,
		0xc547U,0x9a85U,0x2501U,0x4a02U,0x9404U,0x3803U,0x7006U,0xe00cU,0xd013U,0xb02dU,
		0x7051U,0xe0a2U,0xd14fU,0xb295U,0x7521U,0xea42U,0xc48fU,0x9915U,0x2221U,0x4442U,
		0x8884U,0x0103U,0x0206U,0x040cU,0x0818U,0x1030U,0x2060U,0x40c0U,0x8180U,0x130bU,
		0x2616U,0x4c2cU,0x9858U,0x20bbU,0x4176U,0x82ecU,0x15d3U,0x2ba6U,0x574cU,0xae98U,
		0x4d3bU,0x9a76U,0x24e7U,0x49ceU,0x939cU,0x3733U,0x6e66U,0xdcccU,0xa993U,0x432dU,
		0x865aU,0x1cbfU,0x397eU,0x72fcU,0xe5f8U,0xdbfbU,0xa7fdU,0x5ff1U,0xbfe2U,0x6fcfU,
		0xdf9eU,0xaf37U,0x4e65U,0x9ccaU,0x299fU,0x533eU,0xa67cU,0x5cf3U,0xb9e6U,0x63c7U,
		0xc78eU,0x9f17U,0x2e25U,0x5c4aU,0xb894U,0x6123U,0xc246U,0x9487U,0x3905U,0x720aU,
		0xe414U,0xd823U,0xa04dU,0x5091U,0xa122U,0x524fU,0xa49eU,0x5937U,0xb26eU,0x74d7U,
		0xe9aeU,0xc357U,0x96a5U,0x3d41U,0x7a82U,0xf504U,0xfa03U,0xe40dU,0xd811U,0xa029U,
		0x5059U,0xa0b2U,0x516fU,0xa2deU,0x55b7U,0xab6eU,0x46d7U,0x8daeU,0x0b57U,0x16aeU,
		0x2d5cU,0x5ab8U,0xb570U,0x7aebU,0xf5d6U,0xfba7U,0xe745U,0xde81U,0xad09U,0x4a19U,
		0x9432U,0x386fU,0x70deU,0xe1bcU,0xd373U,0xb6edU,0x7dd1U,0xfba2U,0xe74fU,0xde95U,
		0xad21U,0x4a49U,0x9492U,0x392fU,0x725eU,0xe4bcU,0xd973U,0xa2edU,0x55d1U,0xaba2U,
		0x474fU,0x8e9eU,0x0d37U,0x1a6eU,0x34dcU,0x69b8U,0xd370U,0xb6ebU,0x7dddU,0xfbbaU,
		0xe77fU,0xdef5U,0xade1U,0x4bc9U,0x9792U,0x3f2fU,0x7e5eU,0xfcbcU,0xe973U,0xc2edU,
		0x95d1U,0x3ba9U,0x7752U,0xeea4U,0xcd43U,0x8a8dU,0x0511U,0x0a22U,0x1444U,0x2888U,
		0x5110U,0xa220U,0x544bU,0xa896U,0x4127U,0x824eU,0x1497U,0x292eU,0x525cU,0xa4b8U,
		0x597bU,0xb2f6U,0x75e7U,0xebceU,0xc797U,0x9f25U,0x2e41U,0x5c82U,0xb904U,0x6203U,
		0xc406U,0x9807U,0x2005U,0x400aU,0x8014U,0x1023U,0x2046U,0x408cU,0x8118U,0x123bU,
		0x2476U,0x48ecU,0x91d8U,0x33bbU,0x6776U,0xceecU,0x8dd3U,0x0badU,0x175aU,0x2eb4U,
		0x5d68U,0xbad0U,0x65abU,0xcb56U,0x86a7U,0x1d45U,0x3a8aU,0x7514U,0xea28U,0xc45bU,
		0x98bdU,0x2171U,0x42e2U,0x85c4U,0x1b83U,0x3706U,0x6e0cU,0xdc18U,0xa83bU,0x407dU,
		0x80faU,0x11ffU,0x23feU,0x47fcU,0x8ff8U,0x0ffbU,0x1ff6U,0x3fecU,0x7fd8U,0xffb0U,
		0xef6bU,0xceddU,0x8db1U,0x0b69U,0x16d2U,0x2da4U,0x5b48U,0xb690U,0x7d2bU,0xfa56U,
		0xe4a7U,0xd945U,0xa281U,0x5509U,0xaa12U,0x442fU,0x885eU,0x00b7U,0x016eU,0x02dcU,
		0x05b8U,0x0b70U,0x16e0U,0x2dc0U,0x5b80U,0xb700U,0x7e0bU,0xfc16U,0xe827U,0xc045U,
		0x9081U,0x3109U,0x6212U,0xc424U,0x9843U,0x208dU,0x411aU,0x8234U,0x1463U,0x28c6U,
		0x518cU,0xa318U,0x563bU,0xac76U,0x48e7U,0x91ceU,0x3397U,0x672eU,0xce5cU,0x8cb3U,
		0x096dU,0x12daU,0x25b4U,0x4b68U,0x96d0U,0x3dabU,0x7b56U,0xf6acU,0xfd53U,0xeaadU,
		0xc551U,0x9aa9U,0x2559U,0x4ab2U,0x9564U,0x3ac3U,0x7586U,0xeb0cU,0xc613U,0x9c2dU,
		0x2851U,0x50a2U,0xa144U,0x5283U,0xa506U,0x5a07U,0xb40eU,0x7817U,0xf02eU,0xf057U,
		0xf0a5U,0xf141U,0xf289U,0xf519U,0xfa39U,0xe479U,0xd8f9U,0xa1f9U,0x53f9U,0xa7f2U,
		0x5fefU,0xbfdeU,0x6fb7U,0xdf6eU,0xaed7U,0x4da5U,0x9b4aU,0x269fU,0x4d3eU,0x9a7cU,
		0x24f3U,0x49e6U,0x93ccU,0x3793U,0x6f26U,0xde4cU,0xac93U,0x492dU,0x925aU,0x34bfU,
		0x697eU,0xd2fcU,0xb5f3U,0x7bedU,0xf7daU,0xffbfU,0xef75U,0xcee1U,0x8dc9U,0x0b99U,
		0x1732U,0x2e64U,0x5cc8U,0xb990U,0x632bU,0xc656U,0x9ca7U,0x2945U,0x528aU,0xa514U,
		0x5a23U,0xb446U,0x7887U,0xf10eU,0xf217U,0xf425U,0xf841U,0xe089U,0xd119U,0xb239U,
		0x7479U,0xe8f2U,0xc1efU,0x93d5U,0x37a1U,0x6f42U,0xde84U,0xad03U,0x4a0dU,0x941aU,
		0x383fU,0x707eU,0xe0fcU,0xd1f3U,0xb3edU,0x77d1U,0xefa2U,0xcf4fU,0x8e95U,0x0d21U,
		0x1a42U,0x3484U,0x6908U,0xd210U,0xb42bU,0x785dU,0xf0baU,0xf17fU,0xf2f5U,0xf5e1U,
		0xfbc9U,0xe799U,0xdf39U,0xae79U,0x4cf9U,0x99f2U,0x23efU,0x47deU,0x8fbcU,0x0f73U,
		0x1ee6U,0x3dccU,0x7b98U,0xf730U,0xfe6bU,0xecddU,0xc9b1U,0x8369U,0x16d9U,0x2db2U,
		0x5b64U,0xb6c8U,0x7d9bU,0xfb36U,0xe667U,0xdcc5U,0xa981U,0x4309U,0x8612U,0x1c2fU,
		0x385eU,0x70bcU,0xe178U,0xd2fbU,0xb5fdU,0x7bf1U,0xf7e2U,0xffcfU,0xef95U,0xcf21U,
		0x8e49U,0x0c99U,0x1932U,0x3264U,0x64c8U,0xc990U,0x832bU,0x165dU,0x2cbaU,0x5974U,
		0xb2e8U,0x75dbU,0xebb6U,0xc767U,0x9ec5U,0x2d81U,0x5b02U,0xb604U,0x7c03U,0xf806U,
		0xe007U,0xd005U,0xb001U,0x7009U,0xe012U,0xd02fU,0xb055U,0x70a1U,0xe142U,0xd28fU,
		0xb515U,0x7a21U,0xf442U,0xf88fU,0xe115U,0xd221U,0xb449U,0x7899U,0xf132U,0xf26fU,
		0xf4d5U,0xf9a1U,0xe349U,0xd699U,0xbd39U,0x6a79U,0xd4f2U,0xb9efU,0x63d5U,0xc7aaU,
		0x9f5fU,0x2eb5U,0x5d6aU,0xbad4U,0x65a3U,0xcb46U,0x8687U,0x1d05U,0x3a0aU,0x7414U,
		0xe828U,0xc05bU,0x90bdU,0x3171U,0x62e2U,0xc5c4U,0x9b83U,0x270dU,0x4e1aU,0x9c34U,
		0x2863U,0x50c6U,0xa18cU,0x5313U,0xa626U,0x5c47U,0xb88eU,0x6117U,0xc22eU,0x9457U,
		0x38a5U,0x714aU,0xe294U,0xd523U,0xba4dU,0x6491U,0xc922U,0x824fU,0x1495U,0x292aU,
		0x5254U,0xa4a8U,0x595bU,0xb2b6U,0x7567U,0xeaceU,0xc597U,0x9b25U,0x2641U,0x4c82U,
		0x9904U,0x2203U,0x4406U,0x880cU,0x0013U,0x0026U,0x004cU,0x0098U,0x0130U,0x0260U,
		0x04c0U,0x0980U,0x1300U,0x2600U,0x4c00U,0x9800U,0x200bU,0x4016U,0x802cU,0x1053U,
		0x20a6U,0x414cU,0x8298U,0x153bU,0x2a76U,0x54ecU,0xa9d8U,0x43bbU,0x8776U,0x1ee7U,
		0x3dceU,0x7b9cU,0xf738U,0xfe7bU,0xecfdU,0xc9f1U,0x83e9U,0x17d9U,0x2fb2U,0x5f64U,
		0xbec8U,0x6d9bU,0xdb36U,0xa667U,0x5cc5U,0xb98aU,0x631fU,0xc63eU,0x9c77U,0x28e5U,
		0x51caU,0xa394U,0x5723U,0xae46U,0x4c87U,0x990eU,0x2217U,0x442eU,0x885cU,0x00b3U,
		0x0166U,0x02ccU,0x0598U,0x0b30U,0x1660U,0x2cc0U,0x5980U,0xb300U,0x760bU,0xec16U,
		0xc827U,0x8045U,0x1081U,0x2102U,0x4204U,0x8408U,0x181bU,0x3036U,0x606cU,0xc0d8U,
		0x91bbU,0x337dU,0x66faU,0xcdf4U,0x8be3U,0x07cdU,0x0f9aU,0x1f34U,0x3e68U,0x7cd0U,
		0xf9a0U,0xe34bU,0xd69dU,0xbd31U,0x6a69U,0xd4d2U,0xb9afU,0x6355U,0xc6aaU,0x9d5fU,
		0x2ab5U,0x556aU,0xaad4U,0x45a3U,0x8b46U,0x0687U,0x0d0eU,0x1a1cU,0x3438U,0x6870U,
		0xd0e0U,0xb1cbU,0x739dU,0xe73aU,0xde7fU,0xacf5U,0x49e1U,0x93c2U,0x378fU,0x6f1eU,
		0xde3cU,0xac73U,0x48edU,0x91daU,0x33bfU,0x677eU,0xcefcU,0x8df3U,0x0bedU,0x17daU,
		0x2fb4U,0x5f68U,0xbed0U,0x6dabU,0xdb56U,0xa6a7U,0x5d45U,0xba8aU,0x651fU,0xca3eU,
		0x8477U,0x18e5U,0x31caU,0x6394U,0xc728U,0x9e5bU,0x2cbdU,0x597aU,0xb2f4U,0x75e3U,
		0xebc6U,0xc787U,0x9f05U,0x2e01U,0x5c02U,0xb804U,0x6003U,0xc006U,0x9007U,0x3005U,
		0x600aU,0xc014U,0x9023U,0x304dU,0x609aU,0xc134U,0x9263U,0x34cdU,0x699aU,0xd334U,
		0xb663U,0x7ccdU,0xf99aU,0xe33fU,0xd675U,0xbce1U,0x69c9U,0xd392U,0xb72fU,0x7e55U,
		0xfcaaU,0xe95fU,0xc2b5U,0x9561U,0x3ac9U,0x7592U,0xeb24U,0xc643U,0x9c8dU,0x2911U,
		0x5222U,0xa444U,0x5883U,0xb106U,0x7207U,0xe40eU,0xd817U,0xa025U,0x5041U,0xa082U,
		0x510fU,0xa21eU,0x5437U,0xa86eU,0x40d7U,0x81aeU,0x1357U,0x26aeU,0x4d5cU,0x9ab8U,
		0x257bU,0x4af6U,0x95ecU,0x3bd3U,0x77a6U,0xef4cU,0xce93U,0x8d2dU,0x0a51U,0x14a2U,
		0x2944U,0x5288U,0xa510U,0x5a2bU,0xb456U,0x78a7U,0xf14eU,0xf297U,0xf525U,0xfa41U,
		0xe489U,0xd919U,0xa239U,0x5479U,0xa8f2U,0x41efU,0x83deU,0x17b7U,0x2f6eU,0x5edcU,
		0xbdb8U,0x6b7bU,0xd6f6U,0xbde7U,0x6bc5U,0xd78aU,0xbf1fU,0x6e35U,0xdc6aU,0xa8dfU,
		0x41b5U,0x836aU,0x16dfU,0x2dbeU,0x5b7cU,0xb6f8U,0x7dfbU,0xfbf6U,0xe7e7U,0xdfc5U,
		0xaf81U,0x4f09U,0x9e12U,0x2c2fU,0x585eU,0xb0bcU,0x7173U,0xe2e6U,0xd5c7U,0xbb85U,
		0x6701U,0xce02U,0x8c0fU,0x0815U,0x102aU,0x2054U,0x40a8U,0x8150U,0x12abU,0x2556U,
		0x4aacU,0x9558U,0x3abbU,0x7576U,0xeaecU,0xc5d3U,0x9badU,0x2751U,0x4ea2U,0x9d44U,
		0x2a83U,0x5506U,0xaa0cU,0x4413U,0x8826U,0x0047U,0x008eU,0x011cU,0x0238U,0x0470U,
		0x08e0U,0x11c0U,0x2380U,0x4700U,0x8e00U,0x0c0bU,0x1816U,0x302cU,0x6058U,0xc0b0U,
		0x916bU,0x32ddU,0x65baU,0xcb74U,0x86e3U,0x1dcdU,0x3b9aU,0x7734U,0xee68U,0xccdbU,
		0x89bdU,0x0371U,0x06e2U,0x0dc4U,0x1b88U,0x3710U,0x6e20U,0xdc40U,0xa88bU,0x411dU,
		0x823aU,0x147fU,0x28feU,0x51fcU,0xa3f8U,0x57fbU,0xaff6U,0x4fe7U,0x9fceU,0x2f97U,
		0x5f2eU,0xbe5cU,0x6cb3U,0xd966U,0xa2c7U,0x5585U,0xab0aU,0x461fU,0x8c3eU,0x0877U,
		0x10eeU,0x21dcU,0x43b8U,0x8770U,0x1eebU,0x3dd6U,0x7bacU,0xf758U,0xfebbU,0xed7dU,
		0xcaf1U,0x85e9U,0x1bd9U,0x37b2U,0x6f64U,0xdec8U,0xad9bU,0x4b3dU,0x967aU,0x3cffU,
		0x79feU,0xf3fcU,0xf7f3U,0xffedU,0xefd1U,0xcfa9U,0x8f59U,0x0eb9U,0x1d72U,0x3ae4U,
		0x75c8U,0xeb90U,0xc72bU,0x9e5dU,0x2cb1U,0x5962U,0xb2c4U,0x7583U,0xeb06U,0xc607U,
		0x9c05U,0x2801U,0x5002U,0xa004U,0x5003U,0xa006U,0x5007U,0xa00eU,0x5017U,0xa02eU,
		0x5057U,0xa0aeU,0x5157U,0xa2aeU,0x5557U,0xaaaeU,0x4557U,0x8aaeU,0x0557U,0x0aaeU,
		0x155cU,0x2ab8U,0x5570U,0xaae0U,0x45cbU,0x8b96U,0x0727U,0x0e4eU,0x1c9cU,0x3938U,
		0x7270U,0xe4e0U,0xd9cbU,0xa39dU,0x5731U,0xae62U,0x4ccfU,0x999eU,0x2337U,0x466eU,
		0x8cdcU,0x09b3U,0x1366U,0x26ccU,0x4d98U,0x9b30U,0x266bU,0x4cd6U,0x99acU,0x2353U,
		0x46a6U,0x8d4cU,0x0a93U,0x1526U,0x2a4cU,0x5498U,0xa930U,0x426bU,0x84d6U,0x19a7U,
		0x334eU,0x669cU,0xcd38U,0x8a7bU,0x04fdU,0x09faU,0x13f4U,0x27e8U,0x4fd0U,0x9fa0U,
		0x2f4bU,0x5e96U,0xbd2cU,0x6a53U,0xd4a6U,0xb947U,0x6285U,0xc50aU,0x9a1fU,0x2435U,
		0x486aU,0x90d4U,0x31a3U,0x6346U,0xc68cU,0x9d13U,0x2a2dU,0x545aU,0xa8b4U,0x4163U,
		0x82c6U,0x1587U,0x2b0eU,0x561cU,0xac38U,0x487bU,0x90f6U,0x31e7U,0x63ceU,0xc79cU,
		0x9f33U,0x2e6dU,0x5cdaU,0xb9b4U,0x6363U,0xc6c6U,0x9d87U,0x2b05U,0x560aU,0xac14U,
		0x4823U,0x9046U,0x3087U,0x610eU,0xc21cU,0x9433U,0x386dU,0x70daU,0xe1b4U,0xd363U,
		0xb6cdU,0x7d91U,0xfb22U,0xe64fU,0xdc95U,0xa921U,0x4249U,0x8492U,0x192fU,0x325eU,
		0x64bcU,0xc978U,0x82fbU,0x15fdU,0x2bfaU,0x57f4U,0xafe8U,0x4fdbU,0x9fb6U,0x2f67U,
		0x5eceU,0xbd9cU,0x6b33U,0xd666U,0xbcc7U,0x6985U,0xd30aU,0xb61fU,0x7c35U,0xf86aU,
		0xe0dfU,0xd1b5U,0xb361U,0x76c9U,0xed92U,0xcb2fU,0x8655U,0x1ca1U,0x3942U,0x7284U,
		0xe508U,0xda1bU,0xa43dU,0x5871U,0xb0e2U,0x71cfU,0xe39eU,0xd737U,0xbe65U,0x6cc1U,
		0xd982U,0xa30fU,0x5615U,0xac2aU,0x485fU,0x90beU,0x3177U,0x62eeU,0xc5dcU,0x9bb3U,
		0x276dU,0x4edaU,0x9db4U,0x2b63U,0x56c6U,0xad8cU,0x4b13U,0x9626U,0x3c47U,0x788eU,
		0xf11cU,0xf233U,0xf46dU,0xf8d1U,0xe1a9U,0xd359U,0xb6b9U,0x7d79U,0xfaf2U,0xe5efU,
		0xdbd5U,0xa7a1U,0x5f49U,0xbe92U,0x6d2fU,0xda5eU,0xa4b7U,0x5965U,0xb2caU,0x759fU,
		0xeb3eU,0xc677U,0x9ce5U,0x29c1U,0x5382U,0xa704U,0x5e03U,0xbc06U,0x6807U,0xd00eU,
		0xb017U,0x7025U,0xe04aU,0xd09fU,0xb135U,0x7261U,0xe4c2U,0xd98fU,0xa315U,0x5621U,
		0xac42U,0x488fU,0x911eU,0x3237U,0x646eU,0xc8dcU,0x81b3U,0x136dU,0x26daU,0x4db4U,
		0x9b68U,0x26dbU,0x4db6U,0x9b6cU,0x26d3U,0x4da6U,0x9b4cU,0x2693U,0x4d26U,0x9a4cU,
		0x2493U,0x4926U,0x924cU,0x3493U,0x6926U,0xd24cU,0xb493U,0x792dU,0xf25aU,0xf4bfU,
		0xf975U,0xe2e1U,0xd5c9U,0xbb99U,0x6739U,0xce72U,0x8cefU,0x09d5U,0x13aaU,0x2754U,
		0x4ea8U,0x9d50U,0x2aabU,0x5556U,0xaaacU,0x4553U,0x8aa6U,0x0547U,0x0a8eU,0x151cU,
		0x2a38U,0x5470U,0xa8e0U,0x41cbU,0x8396U,0x1727U,0x2e4eU,0x5c9cU,0xb938U,0x627bU,
		0xc4f6U,0x99e7U,0x23c5U,0x478aU,0x8f14U,0x0e23U,0x1c46U,0x388cU,0x7118U,0xe230U,
		0xd46bU,0xb8ddU,0x61b1U,0xc362U,0x96cfU,0x3d95U,0x7b2aU,0xf654U,0xfca3U,0xe94dU,
		0xc291U,0x9529U,0x3a59U,0x74b2U,0xe964U,0xc2c3U,0x958dU,0x3b11U,0x7622U,0xec44U,
		0xc883U,0x810dU,0x1211U,0x2422U,0x4844U,0x9088U,0x311bU,0x6236U,0xc46cU,0x98d3U,
		0x21adU,0x435aU,0x86b4U,0x1d63U,0x3ac6U,0x758cU,0xeb18U,0xc63bU,0x9c7dU,0x28f1U,
		0x51e2U,0xa3c4U,0x5783U,0xaf06U,0x4e07U,0x9c0eU,0x2817U,0x502eU,0xa05cU,0x50b3U,
		0xa166U,0x52c7U,0xa58eU,0x5b17U,0xb62eU,0x7c57U,0xf8aeU,0xe157U,0xd2a5U,0xb541U,
		0x7a89U,0xf512U,0xfa2fU,0xe455U,0xd8a1U,0xa149U,0x5299U,0xa532U,0x5a6fU,0xb4deU,
		0x79b7U,0xf36eU,0xf6d7U,0xfda5U,0xeb41U,0xc689U,0x9d19U,0x2a39U,0x5472U,0xa8e4U,
		0x41c3U,0x8386U,0x1707U,0x2e0eU,0x5c1cU,0xb838U,0x607bU,0xc0f6U,0x91e7U,0x33c5U,
		0x678aU,0xcf14U,0x8e23U,0x0c4dU,0x189aU,0x3134U,0x6268U,0xc4d0U,0x99abU,0x235dU,
		0x46baU,0x8d74U,0x0ae3U,0x15c6U,0x2b8cU,0x5718U,0xae30U,0x4c6bU,0x98d6U,0x21a7U,
		0x434eU,0x869cU,0x1d33U,0x3a66U,0x74ccU,0xe998U,0xc33bU,0x967dU,0x3cf1U,0x79e2U,
		0xf3c4U,0xf783U,0xff0dU,0xee11U,0xcc29U,0x8859U,0x00b9U,0x0172U,0x02e4U,0x05c8U,
		0x0b90U,0x1720U,0x2e40U,0x5c80U,0xb900U,0x620bU,0xc416U,0x9827U,0x2045U,0x408aU,
		0x8114U,0x1223U,0x2446U,0x488cU,0x9118U,0x323bU,0x6476U,0xc8ecU,0x81d3U,0x13adU,
		0x275aU,0x4eb4U,0x9d68U,0x2adbU,0x55b6U,0xab6cU,0x46d3U,0x8da6U,0x0b47U,0x168eU,
		0x2d1cU,0x5a38U,0xb470U,0x78ebU,0xf1d6U,0xf3a7U,0xf745U,0xfe81U,0xed09U,0xca19U,
		0x8439U,0x1879U,0x30f2U,0x61e4U,0xc3c8U,0x979bU,0x3f3dU,0x7e7aU,0xfcf4U,0xe9e3U,
		0xc3cdU,0x9791U,0x3f29U,0x7e52U,0xfca4U,0xe943U,0xc28dU,0x9511U,0x3a29U,0x7452U,
		0xe8a4U,0xc143U,0x928dU,0x3511U,0x6a22U,0xd444U,0xb883U,0x610dU,0xc21aU,0x943fU,
		0x3875U,0x70eaU,0xe1d4U,0xd3a3U,0xb74dU,0x7e91U,0xfd22U,0xea4fU,0xc495U,0x9921U,
		0x2249U,0x4492U,0x8924U,0x0243U,0x0486U,0x090cU,0x1218U,0x2430U,0x4860U,0x90c0U,
		0x318bU,0x6316U,0xc62cU,0x9c53U,0x28adU,0x515aU,0xa2b4U,0x5563U,0xaac6U,0x4587U,
		0x8b0eU,0x0617U,0x0c2eU,0x185cU,0x30b8U,0x6170U,0xc2e0U,0x95cbU,0x3b9dU,0x773aU,
		0xee74U,0xcce3U,0x89cdU,0x0391U,0x0722U,0x0e44U,0x1c88U,0x3910U,0x7220U,0xe440U,
		0xd88bU,0xa11dU,0x5231U,0xa462U,0x58cfU,0xb19eU,0x7337U,0xe66eU,0xdcd7U,0xa9a5U,
		0x4341U,0x8682U,0x1d0fU,0x3a1eU,0x743cU,0xe878U,0xc0fbU,0x91fdU,0x33f1U,0x67e2U,
		0xcfc4U,0x8f83U,0x0f0dU,0x1e1aU,0x3c34U,0x7868U,0xf0d0U,0xf1abU,0xf35dU,0xf6b1U,
		0xfd69U,0xead9U,0xc5b9U,0x9b79U,0x26f9U,0x4df2U,0x9be4U,0x27c3U,0x4f86U,0x9f0cU,
		0x2e13U,0x5c26U,0xb84cU,0x6093U,0xc126U,0x9247U,0x3485U,0x690aU,0xd214U,0xb423U,
		0x784dU,0xf09aU,0xf13fU,0xf275U,0xf4e1U,0xf9c9U,0xe399U,0xd739U,0xbe79U,0x6cf9U,
		0xd9f2U,0xa3efU,0x57d5U,0xafaaU,0x4f5fU,0x9ebeU,0x2d77U,0x5aeeU,0xb5dcU,0x7bb3U,
		0xf766U,0xfec7U,0xed85U,0xcb01U,0x8609U,0x1c19U,0x3832U,0x7064U,0xe0c8U,0xd19bU,
		0xb33dU,0x7671U,0xece2U,0xc9cfU,0x8395U,0x1721U,0x2e42U,0x5c84U,0xb908U,0x621bU,
		0xc436U,0x9867U,0x20c5U,0x418aU,0x8314U,0x1623U,0x2c46U,0x588cU,0xb118U,0x723bU,
		0xe476U,0xd8e7U,0xa1c5U,0x5381U,0xa702U,0x5e0fU,0xbc1eU,0x6837U,0xd06eU,0xb0d7U,
		0x71a5U,0xe34aU,0xd69fU,0xbd35U,0x6a61U,0xd4c2U,0xb98fU,0x6315U,0xc62aU,0x9c5fU,
		0x28b5U,0x516aU,0xa2d4U,0x55a3U,0xab46U,0x4687U,0x8d0eU,0x0a17U,0x142eU,0x285cU,
		0x50b8U,0xa170U,0x52ebU,0xa5d6U,0x5ba7U,0xb74eU,0x7e97U,0xfd2eU,0xea57U,0xc4a5U,
		0x9941U,0x2289U,0x4512U,0x8a24U,0x0443U,0x0886U,0x110cU,0x2218U,0x4430U,0x8860U,
		0x00cbU,0x0196U,0x032cU,0x0658U,0x0cb0U,0x1960U,0x32c0U,0x6580U,0xcb00U,0x860bU,
		0x1c1dU,0x383aU,0x7074U,0xe0e8U,0xd1dbU,0xb3bdU,0x7771U,0xeee2U,0xcdcfU,0x8b95U,
		0x0721U,0x0e42U,0x1c84U,0x3908U,0x7210U,0xe420U,0xd84bU,0xa09dU,0x5131U,0xa262U,
		0x54cfU,0xa99eU,0x4337U,0x866eU,0x1cd7U,0x39aeU,0x735cU,0xe6b8U,0xdd7bU,0xaafdU,
		0x45f1U,0x8be2U,0x07cfU,0x0f9eU,0x1f3cU,0x3e78U,0x7cf0U,0xf9e0U,0xe3cbU,0xd79dU,
		0xbf31U,0x6e69U,0xdcd2U,0xa9afU,0x4355U,0x86aaU,0x1d5fU,0x3abeU,0x757cU,0xeaf8U,
		0xc5fbU,0x9bfdU,0x27f1U,0x4fe2U,0x9fc4U,0x2f83U,0x5f06U,0xbe0cU,0x6c13U,0xd826U,
		0xa047U,0x5085U,0xa10aU,0x521fU,0xa43eU,0x5877U,0xb0eeU,0x71d7U,0xe3aeU,0xd757U,
		0xbea5U,0x6d41U,0xda82U,0xa50fU,0x5a15U,0xb42aU,0x785fU,0xf0beU,0xf177U,0xf2e5U,
		0xf5c1U,0xfb89U,0xe719U,0xde39U,0xac79U,0x48f9U,0x91f2U,0x33efU,0x67deU,0xcfbcU,
		0x8f73U,0x0eedU,0x1ddaU,0x3bb4U,0x7768U,0xeed0U,0xcdabU,0x8b5dU,0x06b1U,0x0d62U,
		0x1ac4U,0x3588U,0x6b10U,0xd620U,0xbc4bU,0x689dU,0xd13aU,0xb27fU,0x74f5U,0xe9eaU,
		0xc3dfU,0x97b5U,0x3f61U,0x7ec2U,0xfd84U,0xeb03U,0xc60dU,0x9c11U,0x2829U,0x5052U,
		0xa0a4U,0x5143U,0xa286U,0x5507U,0xaa0eU,0x4417U,0x882eU,0x0057U,0x00aeU,0x015cU,
		0x02b8U,0x0570U,0x0ae0U,0x15c0U,0x2b80U,0x5700U,0xae00U,0x4c0bU,0x9816U,0x2027U,
		0x404eU,0x809cU,0x1133U,0x2266U,0x44ccU,0x8998U,0x033bU,0x0676U,0x0cecU,0x19d8U,
		0x33b0U,0x6760U,0xcec0U,0x8d8bU,0x0b1dU,0x163aU,0x2c74U,0x58e8U,0xb1d0U,0x73abU,
		0xe756U,0xdea7U,0xad45U,0x4a81U,0x9502U,0x3a0fU,0x741eU,0xe83cU,0xc073U,0x90edU,
		0x31d1U,0x63a2U,0xc744U,0x9e83U,0x2d0dU,0x5a1aU,0xb434U,0x7863U,0xf0c6U,0xf187U,
		0xf305U,0xf601U,0xfc09U,0xe819U,0xc039U,0x9079U,0x30f9U,0x61f2U,0xc3e4U,0x97c3U,
		0x3f8dU,0x7f1aU,0xfe34U,0xec63U,0xc8cdU,0x8191U,0x1329U,0x2652U,0x4ca4U,0x9948U,
		0x229bU,0x4536U,0x8a6cU,0x04d3U,0x09a6U,0x134cU,0x2698U,0x4d30U,0x9a60U,0x24cbU,
		0x4996U,0x932cU,0x3653U,0x6ca6U,0xd94cU,0xa293U,0x552dU,0xaa5aU,0x44bfU,0x897eU,
		0x02f7U,0x05eeU,0x0bdcU,0x17b8U,0x2f70U,0x5ee0U,0xbdc0U,0x6b8bU,0xd716U,0xbe27U,
		0x6c45U,0xd88aU,0xa11fU,0x5235U,0xa46aU,0x58dfU,0xb1beU,0x7377U,0xe6eeU,0xddd7U,
		0xaba5U,0x4741U,0x8e82U,0x0d0fU,0x1a1eU,0x343cU,0x6878U,0xd0f0U,0xb1ebU,0x73ddU,
		0xe7baU,0xdf7fU,0xaef5U,0x4de1U,0x9bc2U,0x278fU,0x4f1eU,0x9e3cU,0x2c73U,0x58e6U,
		0xb1ccU,0x7393U,0xe726U,0xde47U,0xac85U,0x4901U,0x9202U,0x340fU,0x681eU,0xd03cU,
		0xb073U,0x70edU,0xe1daU,0xd3bfU,0xb775U,0x7ee1U,0xfdc2U,0xeb8fU,0xc715U,0x9e21U,
		0x2c49U,0x5892U,0xb124U,0x7243U,0xe486U,0xd907U,0xa205U,0x5401U,0xa802U,0x400fU,
		0x801eU,0x1037U,0x206eU,0x40dcU,0x81b8U,0x137bU,0x26f6U,0x4decU,0x9bd8U,0x27bbU,
		0x4f76U,0x9eecU,0x2dd3U,0x5ba6U,0xb74cU,0x7e93U,0xfd26U,0xea47U,0xc485U,0x9901U,
		0x2209U,0x4412U,0x8824U,0x0043U,0x0086U,0x010cU,0x0218U,0x0430U,0x0860U,0x10c0U,
		0x2180U,0x4300U,0x8600U,0x1c0bU,0x3816U,0x702cU,0xe058U,0xd0bbU,0xb17dU,0x72f1U,
		0xe5e2U,0xdbcfU,0xa795U,0x5f21U,0xbe42U,0x6c8fU,0xd91eU,0xa237U,0x5465U,0xa8caU,
		0x419fU,0x833eU,0x1677U,0x2ceeU,0x59dcU,0xb3b8U,0x777bU,0xeef6U,0xcde7U,0x8bc5U,
		0x0781U,0x0f02U,0x1e04U,0x3c08U,0x7810U,0xf020U,0xf04bU,0xf09dU,0xf131U,0xf269U,
		0xf4d9U,0xf9b9U,0xe379U,0xd6f9U,0xbdf9U,0x6bf9U,0xd7f2U,0xbfefU,0x6fd5U,0xdfaaU,
		0xaf5fU,0x4eb5U,0x9d6aU,0x2adfU,0x55beU,0xab7cU,0x46f3U,0x8de6U,0x0bc7U,0x178eU,
		0x2f1cU,0x5e38U,0xbc70U,0x68ebU,0xd1d6U,0xb3a7U,0x7745U,0xee8aU,0xcd1fU,0x8a35U,
		0x0461U,0x08c2U,0x1184U,0x2308U,0x4610U,0x8c20U,0x084bU,0x1096U,0x212cU,0x4258U,
		0x84b0U,0x196bU,0x32d6U,0x65acU,0xcb58U,0x86bbU,0x1d7dU,0x3afaU,0x75f4U,0xebe8U,
		0xc7dbU,0x9fbdU,0x2f71U,0x5ee2U,0xbdc4U,0x6b83U,0xd706U,0xbe07U,0x6c05U,0xd80aU,
		0xa01fU,0x5035U,0xa06aU,0x50dfU,0xa1beU,0x5377U,0xa6eeU,0x5dd7U,0xbbaeU,0x6757U,
		0xceaeU,0x8d57U,0x0aa5U,0x154aU,0x2a94U,0x5528U,0xaa50U,0x44abU,0x8956U,0x02a7U,
		0x054eU,0x0a9cU,0x1538U,0x2a70U,0x54e0U,0xa9c0U,0x438bU,0x8716U,0x1e27U,0x3c4eU,
		0x789cU,0xf138U,0xf27bU,0xf4fdU,0xf9f1U,0xe3e9U,0xd7d9U,0xbfb9U,0x6f79U,0xdef2U,
		0xadefU,0x4bd5U,0x97aaU,0x3f5fU,0x7ebeU,0xfd7cU,0xeaf3U,0xc5edU,0x9bd1U,0x27a9U,
		0x4f52U,0x9ea4U,0x2d43U,0x5a86U,0xb50cU,0x7a13U,0xf426U,0xf847U,0xe085U,0xd101U,
		0xb209U,0x7419U,0xe832U,0xc06fU,0x90d5U,0x31a1U,0x6342U,0xc684U,0x9d03U,0x2a0dU,
		0x541aU,0xa834U,0x4063U,0x80c6U,0x1187U,0x230eU,0x461cU,0x8c38U,0x087bU,0x10f6U,
		0x21ecU,0x43d8U,0x87b0U,0x1f6bU,0x3ed6U,0x7dacU,0xfb58U,0xe6bbU,0xdd7dU,0xaaf1U,
		0x45e9U,0x8bd2U,0x07afU,0x0f5eU,0x1ebcU,0x3d78U,0x7af0U,0xf5e0U,0xfbcbU,0xe79dU,
		0xdf31U,0xae69U,0x4cd9U,0x99b2U,0x236fU,0x46deU,0x8dbcU,0x0b73U,0x16e6U,0x2dccU,
		0x5b98U,0xb730U,0x7e6bU,0xfcd6U,0xe9a7U,0xc345U,0x9681U,0x3d09U,0x7a12U,0xf424U,
		0xf843U,0xe08dU,0xd111U,0xb229U,0x7459U,0xe8b2U,0xc16fU,0x92d5U,0x35a1U,0x6b42U,
		0xd684U,0xbd03U,0x6a0dU,0xd41aU,0xb83fU,0x6075U,0xc0eaU,0x91dfU,0x33b5U,0x676aU,
		0xced4U,0x8da3U,0x0b4dU,0x169aU,0x2d34U,0x5a68U,0xb4d0U,0x79abU,0xf356U,0xf6a7U,
		0xfd45U,0xea81U,0xc509U,0x9a19U,0x2439U,0x4872U,0x90e4U,0x31c3U,0x6386U,0xc70cU,
		0x9e13U,0x2c2dU,0x585aU,0xb0b4U,0x7163U,0xe2c6U,0xd587U,0xbb05U,0x6601U,0xcc02U,
		0x880fU,0x0015U,0x002aU,0x0054U,0x00a8U,0x0150U,0x02a0U,0x0540U,0x0a80U,0x1500U,
		0x2a00U,0x5400U,0xa800U,0x400bU,0x8016U,0x1027U,0x204eU,0x409cU,0x8138U,0x127bU,
		0x24f6U,0x49ecU,0x93d8U,0x37bbU,0x6f76U,0xdeecU,0xadd3U,0x4badU,0x975aU,0x3ebfU,
		0x7d7eU,0xfafcU,0xe5f3U,0xdbedU,0xa7d1U,0x5fa9U,0xbf52U,0x6eafU,0xdd5eU,0xaab7U,
		0x4565U,0x8acaU,0x059fU,0x0b3eU,0x167cU,0x2cf8U,0x59f0U,0xb3e0U,0x77cbU,0xef96U,
		0xcf27U,0x8e45U,0x0c81U,0x1902U,0x3204U,0x6408U,0xc810U,0x802bU,0x105dU,0x20baU,
		0x4174U,0x82e8U,0x15dbU,0x2bb6U,0x576cU,0xaed8U,0x4dbbU,0x9b76U,0x26e7U,0x4dceU,
		0x9b9cU,0x2733U,0x4e66U,0x9cccU,0x2993U,0x5326U,0xa64cU,0x5c93U,0xb926U,0x6247U,
		0xc48eU,0x9917U,0x2225U,0x444aU,0x8894U,0x0123U,0x0246U,0x048cU,0x0918U,0x1230U,
		0x2460U,0x48c0U,0x9180U,0x330bU,0x6616U,0xcc2cU,0x8853U,0x00adU,0x015aU,0x02b4U,
		0x0568U,0x0ad0U,0x15a0U,0x2b40U,0x5680U,0xad00U,0x4a0bU,0x9416U,0x3827U,0x704eU,
		0xe09cU,0xd133U,0xb26dU,0x74d1U,0xe9a2U,0xc34fU,0x9695U,0x3d21U,0x7a42U,0xf484U,
		0xf903U,0xe20dU,0xd411U,0xb829U,0x6059U,0xc0b2U,0x916fU,0x32d5U,0x65aaU,0xcb54U,
		0x86a3U,0x1d4dU,0x3a9aU,0x7534U,0xea68U,0xc4dbU,0x99bdU,0x2371U,0x46e2U,0x8dc4U,
		0x0b83U,0x1706U,0x2e0cU,0x5c18U,0xb830U,0x606bU,0xc0d6U,0x91a7U,0x3345U,0x668aU,
		0xcd14U,0x8a23U,0x044dU,0x089aU,0x1134U,0x2268U,0x44d0U,0x89a0U,0x034bU,0x0696U,
		0x0d2cU,0x1a58U,0x34b0U,0x6960U,0xd2c0U,0xb58bU,0x7b1dU,0xf63aU,0xfc7fU,0xe8f5U,
		0xc1e1U,0x93c9U,0x3799U,0x6f32U,0xde64U,0xacc3U,0x498dU,0x931aU,0x363fU,0x6c7eU,
		0xd8fcU,0xa1f3U,0x53edU,0xa7daU,0x5fbfU,0xbf7eU,0x6ef7U,0xddeeU,0xabd7U,0x47a5U,
		0x8f4aU,0x0e9fU,0x1d3eU,0x3a7cU,0x74f8U,0xe9f0U,0xc3ebU,0x97ddU,0x3fb1U,0x7f62U,
		0xfec4U,0xed83U,0xcb0dU,0x8611U,0x1c29U,0x3852U,0x70a4U,0xe148U,0xd29bU,0xb53dU,
		0x7a71U,0xf4e2U,0xf9cfU,0xe395U,0xd721U,0xbe49U,0x6c99U,0xd932U,0xa26fU,0x54d5U,
		0xa9aaU,0x435fU,0x86beU,0x1d77U,0x3aeeU,0x75dcU,0xebb8U,0xc77bU,0x9efdU,0x2df1U,
		0x5be2U,0xb7c4U,0x7f83U,0xff06U,0xee07U,0xcc05U,0x8801U,0x0009U,0x0012U,0x0024U,
		0x0048U,0x0090U,0x0120U,0x0240U,0x0480U,0x0900U,0x1200U,0x2400U,0x4800U,0x9000U,
		0x300bU,0x6016U,0xc02cU,0x9053U,0x30adU,0x615aU,0xc2b4U,0x9563U,0x3acdU,0x759aU,
		0xeb34U,0xc663U,0x9ccdU,0x2991U,0x5322U,0xa644U,0x5c83U,0xb906U,0x6207U,0xc40eU,
		0x9817U,0x2025U,0x404aU,0x8094U,0x1123U,0x2246U,0x448cU,0x8918U,0x023bU,0x0476U,
		0x08ecU,0x11d8U,0x23b0U,0x4760U,0x8ec0U,0x0d8bU,0x1b16U,0x362cU,0x6c58U,0xd8b0U,
		0xa16bU,0x52ddU,0xa5baU,0x5b7fU,0xb6feU,0x7df7U,0xfbeeU,0xe7d7U,0xdfa5U,0xaf41U,
		0x4e89U,0x9d12U,0x2a2fU,0x545eU,0xa8bcU,0x4173U,0x82e6U,0x15c7U,0x2b8eU,0x571cU,
		0xae38U,0x4c7bU,0x98f6U,0x21e7U,0x43ceU,0x879cU,0x1f33U,0x3e66U,0x7cccU,0xf998U,
		0xe33bU,0xd67dU,0xbcf1U,0x69e9U,0xd3d2U,0xb7afU,0x7f55U,0xfeaaU,0xed5fU,0xcab5U,
		0x8561U,0x1ac9U,0x3592U,0x6b24U,0xd648U,0xbc9bU,0x693dU,0xd27aU,0xb4ffU,0x79f5U,
		0xf3eaU,0xf7dfU,0xffb5U,0xef61U,0xcec9U,0x8d99U,0x0b39U,0x1672U,0x2ce4U,0x59c8U,
		0xb390U,0x772bU,0xee56U,0xcca7U,0x8945U,0x0281U,0x0502U,0x0a04U,0x1408U,0x2810U,
		0x5020U,0xa040U,0x508bU,0xa116U,0x5227U,0xa44eU,0x5897U,0xb12eU,0x7257U,0xe4aeU,
		0xd957U,0xa2a5U,0x5541U,0xaa82U,0x450fU,0x8a1eU,0x0437U,0x086eU,0x10dcU,0x21b8U,
		0x4370U,0x86e0U,0x1dcbU,0x3b96U,0x772cU,0xee58U,0xccbbU,0x897dU,0x02f1U,0x05e2U,
		0x0bc4U,0x1788U,0x2f10U,0x5e20U,0xbc40U,0x688bU,0xd116U,0xb227U,0x7445U,0xe88aU,
		0xc11fU,0x9235U,0x3461U,0x68c2U,0xd184U,0xb303U,0x760dU,0xec1aU,0xc83fU,0x8075U,
		0x10e1U,0x21c2U,0x4384U,0x8708U,0x1e1bU,0x3c36U,0x786cU,0xf0d8U,0xf1bbU,0xf37dU,
		0xf6f1U,0xfde9U,0xebd9U,0xc7b9U,0x9f79U,0x2ef9U,0x5df2U,0xbbe4U,0x67c3U,0xcf86U,
		0x8f07U,0x0e05U,0x1c0aU,0x3814U,0x7028U,0xe050U,0xd0abU,0xb15dU,0x72b1U,0xe562U,
		0xdacfU,0xa595U,0x5b21U,0xb642U,0x7c8fU,0xf91eU,0xe237U,0xd465U,0xb8c1U,0x6189U,
		0xc312U,0x962fU,0x3c55U,0x78aaU,0xf154U,0xf2a3U,0xf54dU,0xfa91U,0xe529U,0xda59U,
		0xa4b9U,0x5979U,0xb2f2U,0x75efU,0xebdeU,0xc7b7U,0x9f65U,0x2ec1U,0x5d82U,0xbb04U,
		0x6603U,0xcc06U,0x8807U,0x0005U,0x000aU,0x0014U,0x0028U,0x0050U,0x00a0U,0x0140U,
		0x0280U,0x0500U,0x0a00U,0x1400U,0x2800U,0x5000U,0xa000U,0x500bU,0xa016U,0x5027U,
		0xa04eU,0x5097U,0xa12eU,0x5257U,0xa4aeU,0x5957U,0xb2aeU,0x7557U,0xeaaeU,0xc557U,
		0x9aa5U,0x2541U,0x4a82U,0x9504U,0x3a03U,0x7406U,0xe80cU,0xc013U,0x902dU,0x3051U,
		0x60a2U,0xc144U,0x9283U,0x350dU,0x6a1aU,0xd434U,0xb863U,0x60cdU,0xc19aU,0x933fU,
		0x3675U,0x6ceaU,0xd9d4U,0xa3a3U,0x574dU,0xae9aU,0x4d3fU,0x9a7eU,0x24f7U,0x49eeU,
		0x93dcU,0x37b3U,0x6f66U,0xdeccU,0xad93U,0x4b2dU,0x965aU,0x3cbfU,0x797eU,0xf2fcU,
		0xf5f3U,0xfbedU,0xe7d1U,0xdfa9U,0xaf59U,0x4eb9U,0x9d72U,0x2aefU,0x55deU,0xabbcU,
		0x4773U,0x8ee6U,0x0dc7U,0x1b8eU,0x371cU,0x6e38U,0xdc70U,0xa8ebU,0x41ddU,0x83baU,
		0x177fU,0x2efeU,0x5dfcU,0xbbf8U,0x67fbU,0xcff6U,0x8fe7U,0x0fc5U,0x1f8aU,0x3f14U,
		0x7e28U,0xfc50U,0xe8abU,0xc15dU,0x92b1U,0x3569U,0x6ad2U,0xd5a4U,0xbb43U,0x668dU,
		0xcd1aU,0x8a3fU,0x0475U,0x08eaU,0x11d4U,0x23a8U,0x4750U,0x8ea0U,0x0d4bU,0x1a96U,
		0x352cU,0x6a58U,0xd4b0U,0xb96bU,0x62ddU,0xc5baU,0x9b7fU,0x26f5U,0x4deaU,0x9bd4U,
		0x27a3U,0x4f46U,0x9e8cU,0x2d13U,0x5a26U,0xb44cU,0x7893U,0xf126U,0xf247U,0xf485U,
		0xf901U,0xe209U,0xd419U,0xb839U,0x6079U,0xc0f2U,0x91efU,0x33d5U,0x67aaU,0xcf54U,
		0x8ea3U,0x0d4dU,0x1a9aU,0x3534U,0x6a68U,0xd4d0U,0xb9abU,0x635dU,0xc6baU,0x9d7fU,
		0x2af5U,0x55eaU,0xabd4U,0x47a3U,0x8f46U,0x0e87U,0x1d0eU,0x3a1cU,0x7438U,0xe870U,
		0xc0ebU,0x91ddU,0x33b1U,0x6762U,0xcec4U,0x8d83U,0x0b0dU,0x161aU,0x2c34U,0x5868U,
		0xb0d0U,0x71abU,0xe356U,0xd6a7U,0xbd45U,0x6a81U,0xd502U,0xba0fU,0x6415U,0xc82aU,
		0x805fU,0x10b5U,0x216aU,0x42d4U,0x85a8U,0x1b5bU,0x36b6U,0x6d6cU,0xdad8U,0xa5bbU,
		0x5b7dU,0xb6faU,0x7dffU,0xfbfeU,0xe7f7U,0xdfe5U,0xafc1U,0x4f89U,0x9f12U,0x2e2fU,
		0x5c5eU,0xb8bcU,0x6173U,0xc2e6U,0x95c7U,0x3b85U,0x770aU,0xee14U,0xcc23U,0x884dU,
		0x0091U,0x0122U,0x0244U,0x0488U,0x0910U,0x1220U,0x2440U,0x4880U,0x9100U,0x320bU,
		0x6416U,0xc82cU,0x8053U,0x10adU,0x215aU,0x42b4U,0x8568U,0x1adbU,0x35b6U,0x6b6cU,
		0xd6d8U,0xbdbbU,0x6b7dU,0xd6faU,0xbdffU,0x6bf5U,0xd7eaU,0xbfdfU,0x6fb5U,0xdf6aU,
		0xaedfU,0x4db5U,0x9b6aU,0x26dfU,0x4dbeU,0x9b7cU,0x26f3U,0x4de6U,0x9bccU,0x2793U,
		0x4f26U,0x9e4cU,0x2c93U,0x5926U,0xb24cU,0x7493U,0xe926U,0xc247U,0x9485U,0x3901U,
		0x7202U,0xe404U,0xd803U,0xa00dU,0x5011U,0xa022U,0x504fU,0xa09eU,0x5137U,0xa26eU,
		0x54d7U,0xa9aeU,0x4357U,0x86aeU,0x1d57U,0x3aaeU,0x755cU,0xeab8U,0xc57bU,0x9afdU,
		0x25f1U,0x4be2U,0x97c4U,0x3f83U,0x7f06U,0xfe0cU,0xec13U,0xc82dU,0x8051U,0x10a9U,
		0x2152U,0x42a4U,0x8548U,0x1a9bU,0x3536U,0x6a6cU,0xd4d8U,0xb9bbU,0x637dU,0xc6faU,
		0x9dffU,0x2bf5U,0x57eaU,0xafd4U,0x4fa3U,0x9f46U,0x2e87U,0x5d0eU,0xba1cU,0x6433U,
		0xc866U,0x80c7U,0x1185U,0x230aU,0x4614U,0x8c28U,0x085bU,0x10b6U,0x216cU,0x42d8U,
		0x85b0U,0x1b6bU,0x36d6U,0x6dacU,0xdb58U,0xa6bbU,0x5d7dU,0xbafaU,0x65ffU,0xcbfeU,
		0x87f7U,0x1fe5U,0x3fcaU,0x7f94U,0xff28U,0xee5bU,0xccbdU,0x8971U,0x02e9U,0x05d2U,
		0x0ba4U,0x1748U,0x2e90U,0x5d20U,0xba40U,0x648bU,0xc916U,0x8227U,0x1445U,0x288aU,
		0x5114U,0xa228U,0x545bU,0xa8b6U,0x4167U,0x82ceU,0x1597U,0x2b2eU,0x565cU,0xacb8U,
		0x497bU,0x92f6U,0x35e7U,0x6bceU,0xd79cU,0xbf33U,0x6e6dU,0xdcdaU,0xa9bfU,0x4375U,
		0x86eaU,0x1ddfU,0x3bbeU,0x777cU,0xeef8U,0xcdfbU,0x8bfdU,0x07f1U,0x0fe2U,0x1fc4U,
		0x3f88U,0x7f10U,0xfe20U,0xec4bU,0xc89dU,0x8131U,0x1269U,0x24d2U,0x49a4U,0x9348U,
		0x369bU,0x6d36U,0xda6cU,0xa4d3U,0x59adU,0xb35aU,0x76bfU,0xed7eU,0xcaf7U,0x85e5U,
		0x1bc1U,0x3782U,0x6f04U,0xde08U,0xac1bU,0x483dU,0x907aU,0x30ffU,0x61feU,0xc3fcU,
		0x97f3U,0x3fedU,0x7fdaU,0xffb4U,0xef63U,0xcecdU,0x8d91U,0x0b29U,0x1652U,0x2ca4U,
		0x5948U,0xb290U,0x752bU,0xea56U,0xc4a7U,0x9945U,0x2281U,0x4502U,0x8a04U,0x0403U,
		0x0806U,0x100cU,0x2018U,0x4030U,0x8060U,0x10cbU,0x2196U,0x432cU,0x8658U,0x1cbbU,
		0x3976U,0x72ecU,0xe5d8U,0xdbbbU,0xa77dU,0x5ef1U,0xbde2U,0x6bcfU,0xd79eU,0xbf37U,
		0x6e65U,0xdccaU,0xa99fU,0x4335U,0x866aU,0x1cdfU,0x39beU,0x737cU,0xe6f8U,0xddfbU,
		0xabfdU,0x47f1U,0x8fe2U,0x0fcfU,0x1f9eU,0x3f3cU,0x7e78U,0xfcf0U,0xe9ebU,0xc3ddU,
		0x97b1U,0x3f69U,0x7ed2U,0xfda4U,0xeb43U,0xc68dU,0x9d11U,0x2a29U,0x5452U,0xa8a4U,
		0x4143U,0x8286U,0x1507U,0x2a0eU,0x541cU,0xa838U,0x407bU,0x80f6U,0x11e7U,0x23ceU,
		0x479cU,0x8f38U,0x0e7bU,0x1cf6U,0x39ecU,0x73d8U,0xe7b0U,0xdf6bU,0xaeddU,0x4db1U,
		0x9b62U,0x26cfU,0x4d9eU,0x9b3cU,0x2673U,0x4ce6U,0x99ccU,0x2393U,0x4726U,0x8e4cU,
		0x0c93U,0x1926U,0x324cU,0x6498U,0xc930U,0x826bU,0x14ddU,0x29baU,0x5374U,0xa6e8U,
		0x5ddbU,0xbbb6U,0x6767U,0xceceU,0x8d97U,0x0b25U,0x164aU,0x2c94U,0x5928U,0xb250U,
		0x74abU,0xe956U,0xc2a7U,0x9545U,0x3a81U,0x7502U,0xea04U,0xc403U,0x980dU,0x2011U,
		0x4022U,0x8044U,0x1083U,0x2106U,0x420cU,0x8418U,0x183bU,0x3076U,0x60ecU,0xc1d8U,
		0x93bbU,0x377dU,0x6efaU,0xddf4U,0xabe3U,0x47cdU,0x8f9aU,0x0f3fU,0x1e7eU,0x3cfcU,
		0x79f8U,0xf3f0U,0xf7ebU,0xffddU,0xefb1U,0xcf69U,0x8ed9U,0x0db9U,0x1b72U,0x36e4U,
		0x6dc8U,0xdb90U,0xa72bU,0x5e5dU,0xbcbaU,0x697fU,0xd2feU,0xb5f7U,0x7be5U,0xf7caU,
		0xff9fU,0xef35U,0xce61U,0x8cc9U,0x0999U,0x1332U,0x2664U,0x4cc8U,0x9990U,0x232bU,
		0x4656U,0x8cacU,0x0953U,0x12a6U,0x254cU,0x4a98U,0x9530U,0x3a6bU,0x74d6U,0xe9acU,
		0xc353U,0x96adU,0x3d51U,0x7aa2U,0xf544U,0xfa83U,0xe50dU,0xda11U,0xa429U,0x5859U,
		0xb0b2U,0x716fU,0xe2deU,0xd5b7U,0xbb65U,0x66c1U,0xcd82U,0x8b0fU,0x0615U,0x0c2aU,
		0x1854U,0x30a8U,0x6150U,0xc2a0U,0x954bU,0x3a9dU,0x753aU,0xea74U,0xc4e3U,0x99cdU,
		0x2391U,0x4722U,0x8e44U,0x0c83U,0x1906U,0x320cU,0x6418U,0xc830U,0x806bU,0x10ddU,
		0x21baU,0x4374U,0x86e8U,0x1ddbU,0x3bb6U,0x776cU,0xeed8U,0xcdbbU,0x8b7dU,0x06f1U,
		0x0de2U,0x1bc4U,0x3788U,0x6f10U,0xde20U,0xac4bU,0x489dU,0x913aU,0x327fU,0x64feU,
		0xc9fcU,0x83f3U,0x17edU,0x2fdaU,0x5fb4U,0xbf68U,0x6edbU,0xddb6U,0xab67U,0x46c5U,
		0x8d8aU,0x0b1fU,0x163eU,0x2c7cU,0x58f8U,0xb1f0U,0x73ebU,0xe7d6U,0xdfa7U,0xaf45U,
		0x4e81U,0x9d02U,0x2a0fU,0x541eU,0xa83cU,0x4073U,0x80e6U,0x11c7U,0x238eU,0x471cU,
		0x8e38U,0x0c7bU,0x18f6U,0x31ecU,0x63d8U,0xc7b0U,0x9f6bU,0x2eddU,0x5dbaU,0xbb74U,
		0x66e3U,0xcdc6U,0x8b87U,0x0705U,0x0e0aU,0x1c14U,0x3828U,0x7050U,0xe0a0U,0xd14bU,
		0xb29dU,0x7531U,0xea62U,0xc4cfU,0x9995U,0x2321U,0x4642U,0x8c84U,0x0903U,0x1206U,
		0x240cU,0x4818U,0x9030U,0x306bU,0x60d6U,0xc1acU,0x9353U,0x36adU,0x6d5aU,0xdab4U,
		0xa563U,0x5acdU,0xb59aU,0x7b3fU,0xf67eU,0xfcf7U,0xe9e5U,0xc3c1U,0x9789U,0x3f19U,
		0x7e32U,0xfc64U,0xe8c3U,0xc18dU,0x9311U,0x3629U,0x6c52U,0xd8a4U,0xa143U,0x528dU,
		0xa51aU,0x5a3fU,0xb47eU,0x78f7U,0xf1eeU,0xf3d7U,0xf7a5U,0xff41U,0xee89U,0xcd19U,
		0x8a39U,0x0479U,0x08f2U,0x11e4U,0x23c8U,0x4790U,0x8f20U,0x0e4bU,0x1c96U,0x392cU,
		0x7258U,0xe4b0U,0xd96bU,0xa2ddU,0x55b1U,0xab62U,0x46cfU,0x8d9eU,0x0b37U,0x166eU,
		0x2cdcU,0x59b8U,0xb370U,0x76ebU,0xedd6U,0xcba7U,0x8745U,0x1e81U,0x3d02U,0x7a04U,
		0xf408U,0xf81bU,0xe03dU,0xd071U,0xb0e9U,0x71d9U,0xe3b2U,0xd76fU,0xbed5U,0x6da1U,
		0xdb42U,0xa68fU,0x5d15U,0xba2aU,0x645fU,0xc8beU,0x8177U,0x12e5U,0x25caU,0x4b94U,
		0x9728U,0x3e5bU,0x7cb6U,0xf96cU,0xe2d3U,0xd5adU,0xbb51U,0x66a9U,0xcd52U,0x8aafU,
		0x0555U,0x0aaaU,0x1554U,0x2aa8U,0x5550U,0xaaa0U,0x454bU,0x8a96U,0x0527U,0x0a4eU,
		0x149cU,0x2938U,0x5270U,0xa4e0U,0x59cbU,0xb396U,0x7727U,0xee4eU,0xcc97U,0x8925U,
		0x0241U,0x0482U,0x0904U,0x1208U,0x2410U,0x4820U,0x9040U,0x308bU,0x6116U,0xc22cU,
		0x9453U,0x38adU,0x715aU,0xe2b4U,0xd563U,0xbacdU,0x6591U,0xcb22U,0x864fU,0x1c95U,
		0x392aU,0x7254U,0xe4a8U,0xd95bU,0xa2bdU,0x5571U,0xaae2U,0x45cfU,0x8b9eU,0x0737U,
		0x0e6eU,0x1cdcU,0x39b8U,0x7370U,0xe6e0U,0xddcbU,0xab9dU,0x4731U,0x8e62U,0x0ccfU,
		0x199eU,0x333cU,0x6678U,0xccf0U,0x89ebU,0x03ddU,0x07baU,0x0f74U,0x1ee8U,0x3dd0U,
		0x7ba0U,0xf740U,0xfe8bU,0xed1dU,0xca31U,0x8469U,0x18d9U,0x31b2U,0x6364U,0xc6c8U,
		0x9d9bU,0x2b3dU,0x567aU,0xacf4U,0x49e3U,0x93c6U,0x3787U,0x6f0eU,0xde1cU,0xac33U,
		0x486dU,0x90daU,0x31bfU,0x637eU,0xc6fcU,0x9df3U,0x2bedU,0x57daU,0xafb4U,0x4f63U,
		0x9ec6U,0x2d87U,0x5b0eU,0xb61cU,0x7c33U,0xf866U,0xe0c7U,0xd185U,0xb301U,0x7609U,
		0xec12U,0xc82fU,0x8055U,0x10a1U,0x2142U,0x4284U,0x8508U,0x1a1bU,0x3436U,0x686cU,
		0xd0d8U,0xb1bbU,0x737dU,0xe6faU,0xddffU,0xabf5U,0x47e1U,0x8fc2U,0x0f8fU,0x1f1eU,
		0x3e3cU,0x7c78U,0xf8f0U,0xe1ebU,0xd3ddU,0xb7b1U,0x7f69U,0xfed2U,0xedafU,0xcb55U,
		0x86a1U,0x1d49U,0x3a92U,0x7524U,0xea48U,0xc49bU,0x993dU,0x2271U,0x44e2U,0x89c4U,
		0x0383U,0x0706U,0x0e0cU,0x1c18U,0x3830U,0x7060U,0xe0c0U,0xd18bU,0xb31dU,0x7631U,
		0xec62U,0xc8cfU,0x8195U,0x1321U,0x2642U,0x4c84U,0x9908U,0x221bU,0x4436U,0x886cU,
		0x00d3U,0x01a6U,0x034cU,0x0698U,0x0d30U,0x1a60U,0x34c0U,0x6980U,0xd300U,0xb60bU,
		0x7c1dU,0xf83aU,0xe07fU,0xd0f5U,0xb1e1U,0x73c9U,0xe792U,0xdf2fU,0xae55U,0x4ca1U,
		0x9942U,0x228fU,0x451eU,0x8a3cU,0x0473U,0x08e6U,0x11ccU,0x2398U,0x4730U,0x8e60U,
		0x0ccbU,0x1996U,0x332cU,0x6658U,0xccb0U,0x896bU,0x02ddU,0x05baU,0x0b74U,0x16e8U,
		0x2dd0U,0x5ba0U,0xb740U,0x7e8bU,0xfd16U,0xea27U,0xc445U,0x9881U,0x2109U,0x4212U,
		0x8424U,0x1843U,0x3086U,0x610cU,0xc218U,0x943bU,0x387dU,0x70faU,0xe1f4U,0xd3e3U,
		0xb7cdU,0x7f91U,0xff22U,0xee4fU,0xcc95U,0x8921U,0x0249U,0x0492U,0x0924U,0x1248U,
		0x2490U,0x4920U,0x9240U,0x348bU,0x6916U,0xd22cU,0xb453U,0x78adU,0xf15aU,0xf2bfU,
		0xf575U,0xfae1U,0xe5c9U,0xdb99U,0xa739U,0x5e79U,0xbcf2U,0x69efU,0xd3deU,0xb7b7U,
		0x7f65U,0xfecaU,0xed9fU,0xcb35U,0x8661U,0x1cc9U,0x3992U,0x7324U,0xe648U,0xdc9bU,
		0xa93dU,0x4271U,0x84e2U,0x19cfU,0x339eU,0x673cU,0xce78U,0x8cfbU,0x09fdU,0x13faU,
		0x27f4U,0x4fe8U,0x9fd0U,0x2fabU,0x5f56U,0xbeacU,0x6d53U,0xdaa6U,0xa547U,0x5a85U,
		0xb50aU,0x7a1fU,0xf43eU,0xf877U,0xe0e5U,0xd1c1U,0xb389U,0x7719U,0xee32U,0xcc6fU,
		0x88d5U,0x01a1U,0x0342U,0x0684U,0x0d08U,0x1a10U,0x3420U,0x6840U,0xd080U,0xb10bU,
		0x721dU,0xe43aU,0xd87fU,0xa0f5U,0x51e1U,0xa3c2U,0x578fU,0xaf1eU,0x4e37U,0x9c6eU,
		0x28d7U,0x51aeU,0xa35cU,0x56b3U,0xad66U,0x4ac7U,0x958eU,0x3b17U,0x762eU,0xec5cU,
		0xc8b3U,0x816dU,0x12d1U,0x25a2U,0x4b44U,0x9688U,0x3d1bU,0x7a36U,0xf46cU,0xf8d3U,
		0xe1adU,0xd351U,0xb6a9U,0x7d59U,0xfab2U,0xe56fU,0xdad5U,0xa5a1U,0x5b49U,0xb692U,
		0x7d2fU,0xfa5eU,0xe4b7U,0xd965U,0xa2c1U,0x5589U,0xab12U,0x462fU,0x8c5eU,0x08b7U,
		0x116eU,0x22dcU,0x45b8U,0x8b70U,0x06ebU,0x0dd6U,0x1bacU,0x3758U,0x6eb0U,0xdd60U,
		0xaacbU,0x459dU,0x8b3aU,0x067fU,0x0cfeU,0x19fcU,0x33f8U,0x67f0U,0xcfe0U,0x8fcbU,
		0x0f9dU,0x1f3aU,0x3e74U,0x7ce8U,0xf9d0U,0xe3abU,0xd75dU,0xbeb1U,0x6d69U,0xdad2U,
		0xa5afU,0x5b55U,0xb6aaU,0x7d5fU,0xfabeU,0xe577U,0xdae5U,0xa5c1U,0x5b89U,0xb712U,
		0x7e2fU,0xfc5eU,0xe8b7U,0xc165U,0x92c1U,0x3589U,0x6b12U,0xd624U,0xbc43U,0x688dU,
		0xd11aU,0xb23fU,0x7475U,0xe8eaU,0xc1dfU,0x93b5U,0x3761U,0x6ec2U,0xdd84U,0xab03U,
		0x460dU,0x8c1aU,0x083fU,0x107eU,0x20fcU,0x41f8U,0x83f0U,0x17ebU,0x2fd6U,0x5facU,
		0xbf58U,0x6ebbU,0xdd76U,0xaae7U,0x45c5U,0x8b8aU,0x071fU,0x0e3eU,0x1c7cU,0x38f8U,
		0x71f0U,0xe3e0U,0xd7cbU,0xbf9dU,0x6f31U,0xde62U,0xaccfU,0x4995U,0x932aU,0x365fU,
		0x6cbeU,0xd97cU,0xa2f3U,0x55edU,0xabdaU,0x47bfU,0x8f7eU,0x0ef7U,0x1deeU,0x3bdcU,
		0x77b8U,0xef70U,0xceebU,0x8dddU,0x0bb1U,0x1762U,0x2ec4U,0x5d88U,0xbb10U,0x662bU,
		0xcc56U,0x88a7U,0x0145U,0x028aU,0x0514U,0x0a28U,0x1450U,0x28a0U,0x5140U,0xa280U,
		0x550bU,0xaa16U,0x4427U,0x884eU,0x0097U,0x012eU,0x025cU,0x04b8U,0x0970U,0x12e0U,
		0x25c0U,0x4b80U,0x9700U,0x3e0bU,0x7c16U,0xf82cU,0xe053U,0xd0adU,0xb151U,0x72a9U,
		0xe552U,0xdaafU,0xa555U,0x5aa1U,0xb542U,0x7a8fU,0xf51eU,0xfa37U,0xe465U,0xd8c1U,
		0xa189U,0x5319U,0xa632U,0x5c6fU,0xb8deU,0x61b7U,0xc36eU,0x96d7U,0x3da5U,0x7b4aU,
		0xf694U,0xfd23U,0xea4dU,0xc491U,0x9929U,0x2259U,0x44b2U,0x8964U,0x02c3U,0x0586U,
		0x0b0cU,0x1618U,0x2c30U,0x5860U,0xb0c0U,0x718bU,0xe316U,0xd627U,0xbc45U,0x6881U,
		0xd102U,0xb20fU,0x7415U,0xe82aU,0xc05fU,0x90b5U,0x3161U,0x62c2U,0xc584U,0x9b03U,
		0x260dU,0x4c1aU,0x9834U,0x2063U,0x40c6U,0x818cU,0x1313U,0x2626U,0x4c4cU,0x9898U,
		0x213bU,0x4276U,0x84ecU,0x19d3U,0x33a6U,0x674cU,0xce98U,0x8d3bU,0x0a7dU,0x14faU,
		0x29f4U,0x53e8U,0xa7d0U,0x5fabU,0xbf56U,0x6ea7U,0xdd4eU,0xaa97U,0x4525U,0x8a4aU,
		0x049fU,0x093eU,0x127cU,0x24f8U,0x49f0U,0x93e0U,0x37cbU,0x6f96U,0xdf2cU,0xae53U,
		0x4cadU,0x995aU,0x22bfU,0x457eU,0x8afcU,0x05f3U,0x0be6U,0x17ccU,0x2f98U,0x5f30U,
		0xbe60U,0x6ccbU,0xd996U,0xa327U,0x5645U,0xac8aU,0x491fU,0x923eU,0x3477U,0x68eeU,
		0xd1dcU,0xb3b3U,0x776dU,0xeedaU,0xcdbfU,0x8b75U,0x06e1U,0x0dc2U,0x1b84U,0x3708U,
		0x6e10U,0xdc20U,0xa84bU,0x409dU,0x813aU,0x127fU,0x24feU,0x49fcU,0x93f8U,0x37fbU,
		0x6ff6U,0xdfecU,0xafd3U,0x4fadU,0x9f5aU,0x2ebfU,0x5d7eU,0xbafcU,0x65f3U,0xcbe6U,
		0x87c7U,0x1f85U,0x3f0aU,0x7e14U,0xfc28U,0xe85bU,0xc0bdU,0x9171U,0x32e9U,0x65d2U,
		0xcba4U,0x8743U,0x1e8dU,0x3d1aU,0x7a34U,0xf468U,0xf8dbU,0xe1bdU,0xd371U,0xb6e9U,
		0x7dd9U,0xfbb2U,0xe76fU,0xded5U,0xada1U,0x4b49U,0x9692U,0x3d2fU,0x7a5eU,0xf4bcU,
		0xf973U,0xe2edU,0xd5d1U,0xbba9U,0x6759U,0xceb2U,0x8d6fU,0x0ad5U,0x15aaU,0x2b54U,
		0x56a8U,0xad50U,0x4aabU,0x9556U,0x3aa7U,0x754eU,0xea9cU,0xc533U,0x9a6dU,0x24d1U,
		0x49a2U,0x9344U,0x3683U,0x6d06U,0xda0cU,0xa413U,0x582dU,0xb05aU,0x70bfU,0xe17eU,
		0xd2f7U,0xb5e5U,0x7bc1U,0xf782U,0xff0fU,0xee15U,0xcc21U,0x8849U,0x0099U,0x0132U,
		0x0264U,0x04c8U,0x0990U,0x1320U,0x2640U,0x4c80U,0x9900U,0x220bU,0x4416U,0x882cU,
		0x0053U,0x00a6U,0x014cU,0x0298U,0x0530U,0x0a60U,0x14c0U,0x2980U,0x5300U,0xa600U,
		0x5c0bU,0xb816U,0x6027U,0xc04eU,0x9097U,0x3125U,0x624aU,0xc494U,0x9923U,0x224dU,
		0x449aU,0x8934U,0x0263U,0x04c6U,0x098cU,0x1318U,0x2630U,0x4c60U,0x98c0U,0x218bU,
		0x4316U,0x862cU,0x1c53U,0x38a6U,0x714cU,0xe298U,0xd53bU,0xba7dU,0x64f1U,0xc9e2U,
		0x83cfU,0x1795U,0x2f2aU,0x5e54U,0xbca8U,0x695bU,0xd2b6U,0xb567U,0x7ac5U,0xf58aU,
		0xfb1fU,0xe635U,0xdc61U,0xa8c9U,0x4199U,0x8332U,0x166fU,0x2cdeU,0x59bcU,0xb378U,
		0x76fbU,0xedf6U,0xcbe7U,0x87c5U,0x1f81U,0x3f02U,0x7e04U,0xfc08U,0xe81bU,0xc03dU,
		0x9071U,0x30e9U,0x61d2U,0xc3a4U,0x9743U,0x3e8dU,0x7d1aU,0xfa34U,0xe463U,0xd8cdU,
		0xa191U,0x5329U,0xa652U,0x5cafU,0xb95eU,0x62b7U,0xc56eU,0x9ad7U,0x25a5U,0x4b4aU,
		0x9694U,0x3d23U,0x7a46U,0xf48cU,0xf913U,0xe22dU,0xd451U,0xb8a9U,0x6159U,0xc2b2U,
		0x956fU,0x3ad5U,0x75aaU,0xeb54U,0xc6a3U,0x9d4dU,0x2a91U,0x5522U,0xaa44U,0x4483U,
		0x8906U,0x0207U,0x040eU,0x081cU,0x1038U,0x2070U,0x40e0U,0x81c0U,0x138bU,0x2716U,
		0x4e2cU,0x9c58U,0x28bbU,0x5176U,0xa2ecU,0x55d3U,0xaba6U,0x4747U,0x8e8eU,0x0d17U,
		0x1a2eU,0x345cU,0x68b8U,0xd170U,0xb2ebU,0x75ddU,0xebbaU,0xc77fU,0x9ef5U,0x2de1U,
		0x5bc2U,0xb784U,0x7f03U,0xfe06U,0xec07U,0xc805U,0x8001U,0x1009U,0x2012U,0x4024U,
		0x8048U,0x109bU,0x2136U,0x426cU,0x84d8U,0x19bbU,0x3376U,0x66ecU,0xcdd8U,0x8bbbU,
		0x077dU,0x0efaU,0x1df4U,0x3be8U,0x77d0U,0xefa0U,0xcf4bU,0x8e9dU,0x0d31U,0x1a62U,
		0x34c4U,0x6988U,0xd310U,0xb62bU,0x7c5dU,0xf8baU,0xe17fU,0xd2f5U,0xb5e1U,0x7bc9U,
		0xf792U,0xff2fU,0xee55U,0xcca1U,0x8949U,0x0299U,0x0532U,0x0a64U,0x14c8U,0x2990U,
		0x5320U,0xa640U,0x5c8bU,0xb916U,0x6227U,0xc44eU,0x9897U,0x2125U,0x424aU,0x8494U,
		0x1923U,0x3246U,0x648cU,0xc918U,0x823bU,0x147dU,0x28faU,0x51f4U,0xa3e8U,0x57dbU,
		0xafb6U,0x4f67U,0x9eceU,0x2d97U,0x5b2eU,0xb65cU,0x7cb3U,0xf966U,0xe2c7U,0xd585U,
		0xbb01U,0x6609U,0xcc12U,0x882fU,0x0055U,0x00aaU,0x0154U,0x02a8U,0x0550U,0x0aa0U,
		0x1540U,0x2a80U,0x5500U,0xaa00U,0x440bU,0x8816U,0x0027U,0x004eU,0x009cU,0x0138U,
		0x0270U,0x04e0U,0x09c0U,0x1380U,0x2700U,0x4e00U,0x9c00U,0x280bU,0x5016U,0xa02cU,
		0x5053U,0xa0a6U,0x5147U,0xa28eU,0x5517U,0xaa2eU,0x4457U,0x88aeU,0x0157U,0x02aeU,
		0x055cU,0x0ab8U,0x1570U,0x2ae0U,0x55c0U,0xab80U,0x470bU,0x8e16U,0x0c27U,0x184eU,
		0x309cU,0x6138U,0xc270U,0x94ebU,0x39ddU,0x73baU,0xe774U,0xdee3U,0xadcdU,0x4b91U,
		0x9722U,0x3e4fU,0x7c9eU,0xf93cU,0xe273U,0xd4edU,0xb9d1U,0x63a9U,0xc752U,0x9eafU,
		0x2d55U,0x5aaaU,0xb554U,0x7aa3U,0xf546U,0xfa87U,0xe505U,0xda01U,0xa409U,0x5819U,
		0xb032U,0x706fU,0xe0deU,0xd1b7U,0xb365U,0x76c1U,0xed82U,0xcb0fU,0x8615U,0x1c21U,
		0x3842U,0x7084U,0xe108U,0xd21bU,0xb43dU,0x7871U,0xf0e2U,0xf1cfU,0xf395U,0xf721U,
		0xfe49U,0xec99U,0xc939U,0x8279U,0x14f9U,0x29f2U,0x53e4U,0xa7c8U,0x5f9bU,0xbf36U,
		0x6e67U,0xdcceU,0xa997U,0x4325U,0x864aU,0x1c9fU,0x393eU,0x727cU,0xe4f8U,0xd9fbU,
		0xa3fdU,0x57f1U,0xafe2U,0x4fcfU,0x9f9eU,0x2f37U,0x5e6eU,0xbcdcU,0x69b3U,0xd366U,
		0xb6c7U,0x7d85U,0xfb0aU,0xe61fU,0xdc35U,0xa861U,0x40c9U,0x8192U,0x132fU,0x265eU,
		0x4cbcU,0x9978U,0x22fbU,0x45f6U,0x8becU,0x07d3U,0x0fa6U,0x1f4cU,0x3e98U,0x7d30U,
		0xfa60U,0xe4cbU,0xd99dU,0xa331U,0x5669U,0xacd2U,0x49afU,0x935eU,0x36b7U,0x6d6eU,
		0xdadcU,0xa5b3U,0x5b6dU,0xb6daU,0x7dbfU,0xfb7eU,0xe6f7U,0xdde5U,0xabc1U,0x4789U,
		0x8f12U,0x0e2fU,0x1c5eU,0x38bcU,0x7178U,0xe2f0U,0xd5ebU,0xbbddU,0x67b1U,0xcf62U,
		0x8ecfU,0x0d95U,0x1b2aU,0x3654U,0x6ca8U,0xd950U,0xa2abU,0x555dU,0xaabaU,0x457fU,
		0x8afeU,0x05f7U,0x0beeU,0x17dcU,0x2fb8U,0x5f70U,0xbee0U,0x6dcbU,0xdb96U,0xa727U,
		0x5e45U,0xbc8aU,0x691fU,0xd23eU,0xb477U,0x78e5U,0xf1caU,0xf39fU,0xf735U,0xfe61U,
		0xecc9U,0xc999U,0x8339U,0x1679U,0x2cf2U,0x59e4U,0xb3c8U,0x779bU,0xef36U,0xce67U,
		0x8cc5U,0x0981U,0x1302U,0x2604U,0x4c08U,0x9810U,0x202bU,0x4056U,0x80acU,0x1153U,
		0x22a6U,0x454cU,0x8a98U,0x053bU,0x0a76U,0x14ecU,0x29d8U,0x53b0U,0xa760U,0x5ecbU,
		0xbd96U,0x6b27U,0xd64eU,0xbc97U,0x6925U,0xd24aU,0xb49fU,0x7935U,0xf26aU,0xf4dfU,
		0xf9b5U,0xe361U,0xd6c9U,0xbd99U,0x6b39U,0xd672U,0xbcefU,0x69d5U,0xd3aaU,0xb75fU,
		0x7eb5U,0xfd6aU,0xeadfU,0xc5b5U,0x9b61U,0x26c9U,0x4d92U,0x9b24U,0x2643U,0x4c86U,
		0x990cU,0x2213U,0x4426U,0x884cU,0x0093U,0x0126U,0x024cU,0x0498U,0x0930U,0x1260U,
		0x24c0U,0x4980U,0x9300U,0x360bU,0x6c16U,0xd82cU,0xa053U,0x50adU,0xa15aU,0x52bfU,
		0xa57eU,0x5af7U,0xb5eeU,0x7bd7U,0xf7aeU,0xff57U,0xeea5U,0xcd41U,0x8a89U,0x0519U,
		0x0a32U,0x1464U,0x28c8U,0x5190U,0xa320U,0x564bU,0xac96U,0x4927U,0x924eU,0x3497U,
		0x692eU,0xd25cU,0xb4b3U,0x796dU,0xf2daU,0xf5bfU,0xfb75U,0xe6e1U,0xddc9U,0xab99U,
		0x4739U,0x8e72U,0x0cefU,0x19deU,0x33bcU,0x6778U,0xcef0U,0x8debU,0x0bddU,0x17baU,
		0x2f74U,0x5ee8U,0xbdd0U,0x6babU,0xd756U,0xbea7U,0x6d45U,0xda8aU,0xa51fU,0x5a35U,
		0xb46aU,0x78dfU,0xf1beU,0xf377U,0xf6e5U,0xfdc1U,0xeb89U,0xc719U,0x9e39U,0x2c79U,
		0x58f2U,0xb1e4U,0x73c3U,0xe786U,0xdf07U,0xae05U,0x4c01U,0x9802U,0x200fU,0x401eU,
		0x803cU,0x1073U,0x20e6U,0x41ccU,0x8398U,0x173bU,0x2e76U,0x5cecU,0xb9d8U,0x63bbU,
		0xc776U,0x9ee7U,0x2dc5U,0x5b8aU,0xb714U,0x7e23U,0xfc46U,0xe887U,0xc105U,0x9201U,
		0x3409U,0x6812U,0xd024U,0xb043U,0x708dU,0xe11aU,0xd23fU,0xb475U,0x78e1U,0xf1c2U,
		0xf38fU,0xf715U,0xfe21U,0xec49U,0xc899U,0x8139U,0x1279U,0x24f2U,0x49e4U,0x93c8U,
		0x379bU,0x6f36U,0xde6cU,0xacd3U,0x49adU,0x935aU,0x36bfU,0x6d7eU,0xdafcU,0xa5f3U,
		0x5bedU,0xb7daU,0x7fbfU,0xff7eU,0xeef7U,0xcde5U,0x8bc1U,0x0789U,0x0f12U,0x1e24U,
		0x3c48U,0x7890U,0xf120U,0xf24bU,0xf49dU,0xf931U,0xe269U,0xd4d9U,0xb9b9U,0x6379U,
		0xc6f2U,0x9defU,0x2bd5U,0x57aaU,0xaf54U,0x4ea3U,0x9d46U,0x2a87U,0x550eU,0xaa1cU,
		0x4433U,0x8866U,0x00c7U,0x018eU,0x031cU,0x0638U,0x0c70U,0x18e0U,0x31c0U,0x6380U,
		0xc700U,0x9e0bU,0x2c1dU,0x583aU,0xb074U,0x70e3U,0xe1c6U,0xd387U,0xb705U,0x7e01U,
		0xfc02U,0xe80fU,0xc015U,0x9021U,0x3049U,0x6092U,0xc124U,0x9243U,0x348dU,0x691aU,
		0xd234U,0xb463U,0x78cdU,0xf19aU,0xf33fU,0xf675U,0xfce1U,0xe9c9U,0xc399U,0x9739U,
		0x3e79U,0x7cf2U,0xf9e4U,0xe3c3U,0xd78dU,0xbf11U,0x6e29U,0xdc52U,0xa8afU,0x4155U,
		0x82aaU,0x155fU,0x2abeU,0x557cU,0xaaf8U,0x45fbU,0x8bf6U,0x07e7U,0x0fceU,0x1f9cU,
		0x3f38U,0x7e70U,0xfce0U,0xe9cbU,0xc39dU,0x9731U,0x3e69U,0x7cd2U,0xf9a4U,0xe343U,
		0xd68dU,0xbd11U,0x6a29U,0xd452U,0xb8afU,0x6155U,0xc2aaU,0x955fU,0x3ab5U,0x756aU,
		0xead4U,0xc5a3U,0x9b4dU,0x2691U,0x4d22U,0x9a44U,0x2483U,0x4906U,0x920cU,0x3413U,
		0x6826U,0xd04cU,0xb093U,0x712dU,0xe25aU,0xd4bfU,0xb975U,0x62e1U,0xc5c2U,0x9b8fU,
		0x2715U,0x4e2aU,0x9c54U,0x28a3U,0x5146U,0xa28cU,0x5513U,0xaa26U,0x4447U,0x888eU,
		0x0117U,0x022eU,0x045cU,0x08b8U,0x1170U,0x22e0U,0x45c0U,0x8b80U,0x070bU,0x0e16U,
		0x1c2cU,0x3858U,0x70b0U,0xe160U,0xd2cbU,0xb59dU,0x7b31U,0xf662U,0xfccfU,0xe995U,
		0xc321U,0x9649U,0x3c99U,0x7932U,0xf264U,0xf4c3U,0xf98dU,0xe311U,0xd629U,0xbc59U,
		0x68b9U,0xd172U,0xb2efU,0x75d5U,0xebaaU,0xc75fU,0x9eb5U,0x2d61U,0x5ac2U,0xb584U,
		0x7b03U,0xf606U,0xfc07U,0xe805U,0xc001U,0x9009U,0x3019U,0x6032U,0xc064U,0x90c3U,
		0x318dU,0x631aU,0xc634U,0x9c63U,0x28cdU,0x519aU,0xa334U,0x5663U,0xacc6U,0x4987U,
		0x930eU,0x3617U,0x6c2eU,0xd85cU,0xa0b3U,0x516dU,0xa2daU,0x55bfU,0xab7eU,0x46f7U,
		0x8deeU,0x0bd7U,0x17aeU,0x2f5cU,0x5eb8U,0xbd70U,0x6aebU,0xd5d6U,0xbba7U,0x6745U,
		0xce8aU,0x8d1fU,0x0a35U,0x146aU,0x28d4U,0x51a8U,0xa350U,0x56abU,0xad56U,0x4aa7U,
		0x954eU,0x3a97U,0x752eU,0xea5cU,0xc4b3U,0x996dU,0x22d1U,0x45a2U,0x8b44U,0x0683U,
		0x0d06U,0x1a0cU,0x3418U,0x6830U,0xd060U,0xb0cbU,0x719dU,0xe33aU,0xd67fU,0xbcf5U,
		0x69e1U,0xd3c2U,0xb78fU,0x7f15U,0xfe2aU,0xec5fU,0xc8b5U,0x8161U,0x12c9U,0x2592U,
		0x4b24U,0x9648U,0x3c9bU,0x7936U,0xf26cU,0xf4d3U,0xf9adU,0xe351U,0xd6a9U,0xbd59U,
		0x6ab9U,0xd572U,0xbaefU,0x65d5U,0xcbaaU,0x875fU,0x1eb5U,0x3d6aU,0x7ad4U,0xf5a8U,
		0xfb5bU,0xe6bdU,0xdd71U,0xaae9U,0x45d9U,0x8bb2U,0x076fU,0x0edeU,0x1dbcU,0x3b78U,
		0x76f0U,0xede0U,0xcbcbU,0x879dU,0x1f31U,0x3e62U,0x7cc4U,0xf988U,0xe31bU,0xd63dU,
		0xbc71U,0x68e9U,0xd1d2U,0xb3afU,0x7755U,0xeeaaU,0xcd5fU,0x8ab5U,0x0561U,0x0ac2U,
		0x1584U,0x2b08U,0x5610U,0xac20U,0x484bU,0x9096U,0x3127U,0x624eU,0xc49cU,0x9933U,
		0x226dU,0x44daU,0x89b4U,0x0363U,0x06c6U,0x0d8cU,0x1b18U,0x3630U,0x6c60U,0xd8c0U,
		0xa18bU,0x531dU,0xa63aU,0x5c7fU,0xb8feU,0x61f7U,0xc3eeU,0x97d7U,0x3fa5U,0x7f4aU,
		0xfe94U,0xed23U,0xca4dU,0x8491U,0x1929U,0x3252U,0x64a4U,0xc948U,0x829bU,0x153dU,
		0x2a7aU,0x54f4U,0xa9e8U,0x43dbU,0x87b6U,0x1f67U,0x3eceU,0x7d9cU,0xfb38U,0xe67bU,
		0xdcfdU,0xa9f1U,0x43e9U,0x87d2U,0x1fafU,0x3f5eU,0x7ebcU,0xfd78U,0xeafbU,0xc5fdU,
		0x9bf1U,0x27e9U,0x4fd2U,0x9fa4U,0x2f43U,0x5e86U,0xbd0cU,0x6a13U,0xd426U,0xb847U,
		0x6085U,0xc10aU,0x921fU,0x3435U,0x686aU,0xd0d4U,0xb1a3U,0x734dU,0xe69aU,0xdd3fU,
		0xaa75U,0x44e1U,0x89c2U,0x038fU,0x071eU,0x0e3cU,0x1c78U,0x38f0U,0x71e0U,0xe3c0U,
		0xd78bU,0xbf1dU,0x6e31U,0xdc62U,0xa8cfU,0x4195U,0x832aU,0x165fU,0x2cbeU,0x597cU,
		0xb2f8U,0x75fbU,0xebf6U,0xc7e7U,0x9fc5U,0x2f81U,0x5f02U,0xbe04U,0x6c03U,0xd806U,
		0xa007U,0x5005U,0xa00aU,0x501fU,0xa03eU,0x5077U,0xa0eeU,0x51d7U,0xa3aeU,0x5757U,
		0xaeaeU,0x4d57U,0x9aaeU,0x2557U,0x4aaeU,0x955cU,0x3ab3U,0x7566U,0xeaccU,0xc593U,
		0x9b2dU,0x2651U,0x4ca2U,0x9944U,0x2283U,0x4506U,0x8a0cU,0x0413U,0x0826U,0x104cU,
		0x2098U,0x4130U,0x8260U,0x14cbU,0x2996U,0x532cU,0xa658U,0x5cbbU,0xb976U,0x62e7U,
		0xc5ceU,0x9b97U,0x2725U,0x4e4aU,0x9c94U,0x2923U,0x5246U,0xa48cU,0x5913U,0xb226U,
		0x7447U,0xe88eU,0xc117U,0x9225U,0x3441U,0x6882U,0xd104U,0xb203U,0x740dU,0xe81aU,
		0xc03fU,0x9075U,0x30e1U,0x61c2U,0xc384U,0x9703U,0x3e0dU,0x7c1aU,0xf834U,0xe063U,
		0xd0cdU,0xb191U,0x7329U,0xe652U,0xdcafU,0xa955U,0x42a1U,0x8542U,0x1a8fU,0x351eU,
		0x6a3cU,0xd478U,0xb8fbU,0x61fdU,0xc3faU,0x97ffU,0x3ff5U,0x7feaU,0xffd4U,0xefa3U,
		0xcf4dU,0x8e91U,0x0d29U,0x1a52U,0x34a4U,0x6948U,0xd290U,0xb52bU,0x7a5dU,0xf4baU,
		0xf97fU,0xe2f5U,0xd5e1U,0xbbc9U,0x6799U,0xcf32U,0x8e6fU,0x0cd5U,0x19aaU,0x3354U,
		0x66a8U,0xcd50U,0x8aabU,0x055dU,0x0abaU,0x1574U,0x2ae8U,0x55d0U,0xaba0U,0x474bU,
		0x8e96U,0x0d27U,0x1a4eU,0x349cU,0x6938U,0xd270U,0xb4ebU,0x79ddU,0xf3baU,0xf77fU,
		0xfef5U,0xede1U,0xcbc9U,0x8799U,0x1f39U,0x3e72U,0x7ce4U,0xf9c8U,0xe39bU,0xd73dU,
		0xbe71U,0x6ce9U,0xd9d2U,0xa3afU,0x5755U,0xaeaaU,0x4d5fU,0x9abeU,0x2577U,0x4aeeU,
		0x95dcU,0x3bb3U,0x7766U,0xeeccU,0xcd93U,0x8b2dU,0x0651U,0x0ca2U,0x1944U,0x3288U,
		0x6510U,0xca20U,0x844bU,0x189dU,0x313aU,0x6274U,0xc4e8U,0x99dbU,0x23bdU,0x477aU,
		0x8ef4U,0x0de3U,0x1bc6U,0x378cU,0x6f18U,0xde30U,0xac6bU,0x48ddU,0x91baU,0x337fU,
		0x66feU,0xcdfcU,0x8bf3U,0x07edU,0x0fdaU,0x1fb4U,0x3f68U,0x7ed0U,0xfda0U,0xeb4bU,
		0xc69dU,0x9d31U,0x2a69U,0x54d2U,0xa9a4U,0x4343U,0x8686U,0x1d07U,0x3a0eU,0x741cU,
		0xe838U,0xc07bU,0x90fdU,0x31f1U,0x63e2U,0xc7c4U,0x9f83U,0x2f0dU,0x5e1aU,0xbc34U,
		0x6863U,0xd0c6U,0xb187U,0x7305U,0xe60aU,0xdc1fU,0xa835U,0x4061U,0x80c2U,0x118fU,
		0x231eU,0x463cU,0x8c78U,0x08fbU,0x11f6U,0x23ecU,0x47d8U,0x8fb0U,0x0f6bU,0x1ed6U,
		0x3dacU,0x7b58U,0xf6b0U,0xfd6bU,0xeaddU,0xc5b1U,0x9b69U,0x26d9U,0x4db2U,0x9b64U,
		0x26c3U,0x4d86U,0x9b0cU,0x2613U,0x4c26U,0x984cU,0x2093U,0x4126U,0x824cU,0x1493U,
		0x2926U,0x524cU,0xa498U,0x593bU,0xb276U,0x74e7U,0xe9ceU,0xc397U,0x9725U,0x3e41U,
		0x7c82U,0xf904U,0xe203U,0xd40dU,0xb811U,0x6029U,0xc052U,0x90afU,0x3155U,0x62aaU,
		0xc554U,0x9aa3U,0x254dU,0x4a9aU,0x9534U,0x3a63U,0x74c6U,0xe98cU,0xc313U,0x962dU,
		0x3c51U,0x78a2U,0xf144U,0xf283U,0xf50dU,0xfa11U,0xe429U,0xd859U,0xa0b9U,0x5179U,
		0xa2f2U,0x55efU,0xabdeU,0x47b7U,0x8f6eU,0x0ed7U,0x1daeU,0x3b5cU,0x76b8U,0xed70U,
		0xcaebU,0x85ddU,0x1bb1U,0x3762U,0x6ec4U,0xdd88U,0xab1bU,0x463dU,0x8c7aU,0x08ffU,
		0x11feU,0x23fcU,0x47f8U,0x8ff0U,0x0febU,0x1fd6U,0x3facU,0x7f58U,0xfeb0U,0xed6bU,
		0xcaddU,0x85b1U,0x1b69U,0x36d2U,0x6da4U,0xdb48U,0xa69bU,0x5d3dU,0xba7aU,0x64ffU,
		0xc9feU,0x83f7U,0x17e5U,0x2fcaU,0x5f94U,0xbf28U,0x6e5bU,0xdcb6U,0xa967U,0x42c5U,
		0x858aU,0x1b1fU,0x363eU,0x6c7cU,0xd8f8U,0xa1fbU,0x53fdU,0xa7faU,0x5fffU,0xbffeU,
		0x6ff7U,0xdfeeU,0xafd7U,0x4fa5U,0x9f4aU,0x2e9fU,0x5d3eU,0xba7cU,0x64f3U,0xc9e6U,
		0x83c7U,0x1785U,0x2f0aU,0x5e14U,0xbc28U,0x685bU,0xd0b6U,0xb167U,0x72c5U,0xe58aU,
		0xdb1fU,0xa635U,0x5c61U,0xb8c2U,0x618fU,0xc31eU,0x9637U,0x3c65U,0x78caU,0xf194U,
		0xf323U,0xf64dU,0xfc91U,0xe929U,0xc259U,0x94b9U,0x3979U,0x72f2U,0xe5e4U,0xdbc3U,
		0xa78dU,0x5f11U,0xbe22U,0x6c4fU,0xd89eU,0xa137U,0x5265U,0xa4caU,0x599fU,0xb33eU,
		0x7677U,0xeceeU,0xc9d7U,0x83a5U,0x1741U,0x2e82U,0x5d04U,0xba08U,0x641bU,0xc836U,
		0x8067U,0x10c5U,0x218aU,0x4314U,0x8628U,0x1c5bU,0x38b6U,0x716cU,0xe2d8U,0xd5bbU,
		0xbb7dU,0x66f1U,0xcde2U,0x8bcfU,0x0795U,0x0f2aU,0x1e54U,0x3ca8U,0x7950U,0xf2a0U,
		0xf54bU,0xfa9dU,0xe531U,0xda69U,0xa4d9U,0x59b9U,0xb372U,0x76efU,0xeddeU,0xcbb7U,
		0x8765U,0x1ec1U,0x3d82U,0x7b04U,0xf608U,0xfc1bU,0xe83dU,0xc071U,0x90e9U,0x31d9U,
		0x63b2U,0xc764U,0x9ec3U,0x2d8dU,0x5b1aU,0xb634U,0x7c63U,0xf8c6U,0xe187U,0xd305U,
		0xb601U,0x7c09U,0xf812U,0xe02fU,0xd055U,0xb0a1U,0x7149U,0xe292U,0xd52fU,0xba55U,
		0x64a1U,0xc942U,0x828fU,0x1515U,0x2a2aU,0x5454U,0xa8a8U,0x415bU,0x82b6U,0x1567U,
		0x2aceU,0x559cU,0xab38U,0x467bU,0x8cf6U,0x09e7U,0x13ceU,0x279cU,0x4f38U,0x9e70U,
		0x2cebU,0x59d6U,0xb3acU,0x7753U,0xeea6U,0xcd47U,0x8a85U,0x0501U,0x0a02U,0x1404U,
		0x2808U,0x5010U,0xa020U,0x504bU,0xa096U,0x5127U,0xa24eU,0x5497U,0xa92eU,0x4257U,
		0x84aeU,0x1957U,0x32aeU,0x655cU,0xcab8U,0x857bU,0x1afdU,0x35faU,0x6bf4U,0xd7e8U,
		0xbfdbU,0x6fbdU,0xdf7aU,0xaeffU,0x4df5U,0x9beaU,0x27dfU,0x4fbeU,0x9f7cU,0x2ef3U,
		0x5de6U,0xbbccU,0x6793U,0xcf26U,0x8e47U,0x0c85U,0x190aU,0x3214U,0x6428U,0xc850U,
		0x80abU,0x115dU,0x22baU,0x4574U,0x8ae8U,0x05dbU,0x0bb6U,0x176cU,0x2ed8U,0x5db0U,
		0xbb60U,0x66cbU,0xcd96U,0x8b27U,0x0645U,0x0c8aU,0x1914U,0x3228U,0x6450U,0xc8a0U,
		0x814bU,0x129dU,0x253aU,0x4a74U,0x94e8U,0x39dbU,0x73b6U,0xe76cU,0xded3U,0xadadU,
		0x4b51U,0x96a2U,0x3d4fU,0x7a9eU,0xf53cU,0xfa73U,0xe4edU,0xd9d1U,0xa3a9U,0x5759U,
		0xaeb2U,0x4d6fU,0x9adeU,0x25b7U,0x4b6eU,0x96dcU,0x3db3U,0x7b66U,0xf6ccU,0xfd93U,
		0xeb2dU,0xc651U,0x9ca9U,0x2959U,0x52b2U,0xa564U,0x5ac3U,0xb586U,0x7b07U,0xf60eU,
		0xfc17U,0xe825U,0xc041U,0x9089U,0x3119U,0x6232U,0xc464U,0x98c3U,0x218dU,0x431aU,
		0x8634U,0x1c63U,0x38c6U,0x718cU,0xe318U,0xd63bU,0xbc7dU,0x68f1U,0xd1e2U,0xb3cfU,
		0x7795U,0xef2aU,0xce5fU,0x8cb5U,0x0961U,0x12c2U,0x2584U,0x4b08U,0x9610U,0x3c2bU,
		0x7856U,0xf0acU,0xf153U,0xf2adU,0xf551U,0xfaa9U,0xe559U,0xdab9U,0xa579U,0x5af9U,
		0xb5f2U,0x7befU,0xf7deU,0xffb7U,0xef65U,0xcec1U,0x8d89U,0x0b19U,0x1632U,0x2c64U,
		0x58c8U,0xb190U,0x732bU,0xe656U,0xdca7U,0xa945U,0x4281U,0x8502U,0x1a0fU,0x341eU,
		0x683cU,0xd078U,0xb0fbU,0x71fdU,0xe3faU,0xd7ffU,0xbff5U,0x6fe1U,0xdfc2U,0xaf8fU,
		0x4f15U,0x9e2aU,0x2c5fU,0x58beU,0xb17cU,0x72f3U,0xe5e6U,0xdbc7U,0xa785U,0x5f01U,
		0xbe02U,0x6c0fU,0xd81eU,0xa037U,0x5065U,0xa0caU,0x519fU,0xa33eU,0x5677U,0xaceeU,
		0x49d7U,0x93aeU,0x3757U,0x6eaeU,0xdd5cU,0xaab3U,0x456dU,0x8adaU,0x05bfU,0x0b7eU,
		0x16fcU,0x2df8U,0x5bf0U,0xb7e0U,0x7fcbU,0xff96U,0xef27U,0xce45U,0x8c81U,0x0909U,
		0x1212U,0x2424U,0x4848U,0x9090U,0x312bU,0x6256U,0xc4acU,0x9953U,0x22adU,0x455aU,
		0x8ab4U,0x0563U,0x0ac6U,0x158cU,0x2b18U,0x5630U,0xac60U,0x48cbU,0x9196U,0x3327U,
		0x664eU,0xcc9cU,0x8933U,0x026dU,0x04daU,0x09b4U,0x1368U,0x26d0U,0x4da0U,0x9b40U,
		0x268bU,0x4d16U,0x9a2cU,0x2453U,0x48a6U,0x914cU,0x3293U,0x6526U,0xca4cU,0x8493U,
		0x192dU,0x325aU,0x64b4U,0xc968U,0x82dbU,0x15bdU,0x2b7aU,0x56f4U,0xade8U,0x4bdbU,
		0x97b6U,0x3f67U,0x7eceU,0xfd9cU,0xeb33U,0xc66dU,0x9cd1U,0x29a9U,0x5352U,0xa6a4U,
		0x5d43U,0xba86U,0x6507U,0xca0eU,0x8417U,0x1825U,0x304aU,0x6094U,0xc128U,0x925bU,
		0x34bdU,0x697aU,0xd2f4U,0xb5e3U,0x7bcdU,0xf79aU,0xff3fU,0xee75U,0xcce1U,0x89c9U,
		0x0399U,0x0732U,0x0e64U,0x1cc8U,0x3990U,0x7320U,0xe640U,0xdc8bU,0xa91dU,0x4231U,
		0x8462U,0x18cfU,0x319eU,0x633cU,0xc678U,0x9cfbU,0x29fdU,0x53faU,0xa7f4U,0x5fe3U,
		0xbfc6U,0x6f87U,0xdf0eU,0xae17U,0x4c25U,0x984aU,0x209fU,0x413eU,0x827cU,0x14f3U,
		0x29e6U,0x53ccU,0xa798U,0x5f3bU,0xbe76U,0x6ce7U,0xd9ceU,0xa397U,0x5725U,0xae4aU,
		0x4c9fU,0x993eU,0x2277U,0x44eeU,0x89dcU,0x03b3U,0x0766U,0x0eccU,0x1d98U,0x3b30U,
		0x7660U,0xecc0U,0xc98bU,0x831dU,0x1631U,0x2c62U,0x58c4U,0xb188U,0x731bU,0xe636U,
		0xdc67U,0xa8c5U,0x4181U,0x8302U,0x160fU,0x2c1eU,0x583cU,0xb078U,0x70fbU,0xe1f6U,
		0xd3e7U,0xb7c5U,0x7f81U,0xff02U,0xee0fU,0xcc15U,0x8821U,0x0049U,0x0092U,0x0124U,
		0x0248U,0x0490U,0x0920U,0x1240U,0x2480U,0x4900U,0x9200U,0x340bU,0x6816U,0xd02cU,
		0xb053U,0x70adU,0xe15aU,0xd2bfU,0xb575U,0x7ae1U,0xf5c2U,0xfb8fU,0xe715U,0xde21U,
		0xac49U,0x4899U,0x9132U,0x326fU,0x64deU,0xc9bcU,0x8373U,0x16edU,0x2ddaU,0x5bb4U,
		0xb768U,0x7edbU,0xfdb6U,0xeb67U,0xc6c5U,0x9d81U,0x2b09U,0x5612U,0xac24U,0x4843U,
		0x9086U,0x3107U,0x620eU,0xc41cU,0x9833U,0x206dU,0x40daU,0x81b4U,0x1363U,0x26c6U,
		0x4d8cU,0x9b18U,0x263bU,0x4c76U,0x98ecU,0x21d3U,0x43a6U,0x874cU,0x1e93U,0x3d26U,
		0x7a4cU,0xf498U,0xf93bU,0xe27dU,0xd4f1U,0xb9e9U,0x63d9U,0xc7b2U,0x9f6fU,0x2ed5U,
		0x5daaU,0xbb54U,0x66a3U,0xcd46U,0x8a87U,0x0505U,0x0a0aU,0x1414U,0x2828U,0x5050U,
		0xa0a0U,0x514bU,0xa296U,0x5527U,0xaa4eU,0x4497U,0x892eU,0x0257U,0x04aeU,0x095cU,
		0x12b8U,0x2570U,0x4ae0U,0x95c0U,0x3b8bU,0x7716U,0xee2cU,0xcc53U,0x88adU,0x0151U,
		0x02a2U,0x0544U,0x0a88U,0x1510U,0x2a20U,0x5440U,0xa880U,0x410bU,0x8216U,0x1427U,
		0x284eU,0x509cU,0xa138U,0x527bU,0xa4f6U,0x59e7U,0xb3ceU,0x7797U,0xef2eU,0xce57U,
		0x8ca5U,0x0941U,0x1282U,0x2504U,0x4a08U,0x9410U,0x382bU,0x7056U,0xe0acU,0xd153U,
		0xb2adU,0x7551U,0xeaa2U,0xc54fU,0x9a95U,0x2521U,0x4a42U,0x9484U,0x3903U,0x7206U,
		0xe40cU,0xd813U,0xa02dU,0x5051U,0xa0a2U,0x514fU,0xa29eU,0x5537U,0xaa6eU,0x44d7U,
		0x89aeU,0x0357U,0x06aeU,0x0d5cU,0x1ab8U,0x3570U,0x6ae0U,0xd5c0U,0xbb8bU,0x671dU,
		0xce3aU,0x8c7fU,0x08f5U,0x11eaU,0x23d4U,0x47a8U,0x8f50U,0x0eabU,0x1d56U,0x3aacU,
		0x7558U,0xeab0U,0xc56bU,0x9addU,0x25b1U,0x4b62U,0x96c4U,0x3d83U,0x7b06U,0xf60cU,
		0xfc13U,0xe82dU,0xc051U,0x90a9U,0x3159U,0x62b2U,0xc564U,0x9ac3U,0x258dU,0x4b1aU,
		0x9634U,0x3c63U,0x78c6U,0xf18cU,0xf313U,0xf62dU,0xfc51U,0xe8a9U,0xc159U,0x92b9U,
		0x3579U,0x6af2U,0xd5e4U,0xbbc3U,0x678dU,0xcf1aU,0x8e3fU,0x0c75U,0x18eaU,0x31d4U,
		0x63a8U,0xc750U,0x9eabU,0x2d5dU,0x5abaU,0xb574U,0x7ae3U,0xf5c6U,0xfb87U,0xe705U,
		0xde01U,0xac09U,0x4819U,0x9032U,0x306fU,0x60deU,0xc1bcU,0x9373U,0x36edU,0x6ddaU,
		0xdbb4U,0xa763U,0x5ecdU,0xbd9aU,0x6b3fU,0xd67eU,0xbcf7U,0x69e5U,0xd3caU,0xb79fU,
		0x7f35U,0xfe6aU,0xecdfU,0xc9b5U,0x8361U,0x16c9U,0x2d92U,0x5b24U,0xb648U,0x7c9bU,
		0xf936U,0xe267U,0xd4c5U,0xb981U,0x6309U,0xc612U,0x9c2fU,0x2855U,0x50aaU,0xa154U,
		0x52a3U,0xa546U,0x5a87U,0xb50eU,0x7a17U,0xf42eU,0xf857U,0xe0a5U,0xd141U,0xb289U,
		0x7519U,0xea32U,0xc46fU,0x98d5U,0x21a1U,0x4342U,0x8684U,0x1d03U,0x3a06U,0x740cU,
		0xe818U,0xc03bU,0x907dU,0x30f1U,0x61e2U,0xc3c4U,0x9783U,0x3f0dU,0x7e1aU,0xfc34U,
		0xe863U,0xc0cdU,0x9191U,0x3329U,0x6652U,0xcca4U,0x8943U,0x028dU,0x051aU,0x0a34U,
		0x1468U,0x28d0U,0x51a0U,0xa340U,0x568bU,0xad16U,0x4a27U,0x944eU,0x3897U,0x712eU,
		0xe25cU,0xd4b3U,0xb96dU,0x62d1U,0xc5a2U,0x9b4fU,0x2695U,0x4d2aU,0x9a54U,0x24a3U,
		0x4946U,0x928cU,0x3513U,0x6a26U,0xd44cU,0xb893U,0x612dU,0xc25aU,0x94bfU,0x3975U,
		0x72eaU,0xe5d4U,0xdba3U,0xa74dU,0x5e91U,0xbd22U,0x6a4fU,0xd49eU,0xb937U,0x6265U,
		0xc4caU,0x999fU,0x2335U,0x466aU,0x8cd4U,0x09a3U,0x1346U,0x268cU,0x4d18U,0x9a30U,
		0x246bU,0x48d6U,0x91acU,0x3353U,0x66a6U,0xcd4cU,0x8a93U,0x052dU,0x0a5aU,0x14b4U,
		0x2968U,0x52d0U,0xa5a0U,0x5b4bU,0xb696U,0x7d27U,0xfa4eU,0xe497U,0xd925U,0xa241U,
		0x5489U,0xa912U,0x422fU,0x845eU,0x18b7U,0x316eU,0x62dcU,0xc5b8U,0x9b7bU,0x26fdU,
		0x4dfaU,0x9bf4U,0x27e3U,0x4fc6U,0x9f8cU,0x2f13U,0x5e26U,0xbc4cU,0x6893U,0xd126U,
		0xb247U,0x7485U,0xe90aU,0xc21fU,0x9435U,0x3861U,0x70c2U,0xe184U,0xd303U,0xb60dU,
		0x7c11U,0xf822U,0xe04fU,0xd095U,0xb121U,0x7249U,0xe492U,0xd92fU,0xa255U,0x54a1U,
		0xa942U,0x428fU,0x851eU,0x1a37U,0x346eU,0x68dcU,0xd1b8U,0xb37bU,0x76fdU,0xedfaU,
		0xcbffU,0x87f5U,0x1fe1U,0x3fc2U,0x7f84U,0xff08U,0xee1bU,0xcc3dU,0x8871U,0x00e9U,
		0x01d2U,0x03a4U,0x0748U,0x0e90U,0x1d20U,0x3a40U,0x7480U,0xe900U,0xc20bU,0x941dU,
		0x3831U,0x7062U,0xe0c4U,0xd183U,0xb30dU,0x7611U,0xec22U,0xc84fU,0x8095U,0x1121U,
		0x2242U,0x4484U,0x8908U,0x021bU,0x0436U,0x086cU,0x10d8U,0x21b0U,0x4360U,0x86c0U,
		0x1d8bU,0x3b16U,0x762cU,0xec58U,0xc8bbU,0x817dU,0x12f1U,0x25e2U,0x4bc4U,0x9788U,
		0x3f1bU,0x7e36U,0xfc6cU,0xe8d3U,0xc1adU,0x9351U,0x36a9U,0x6d52U,0xdaa4U,0xa543U,
		0x5a8dU,0xb51aU,0x7a3fU,0xf47eU,0xf8f7U,0xe1e5U,0xd3c1U,0xb789U,0x7f19U,0xfe32U,
		0xec6fU,0xc8d5U,0x81a1U,0x1349U,0x2692U,0x4d24U,0x9a48U,0x249bU,0x4936U,0x926cU,
		0x34d3U,0x69a6U,0xd34cU,0xb693U,0x7d2dU,0xfa5aU,0xe4bfU,0xd975U,0xa2e1U,0x55c9U,
		0xab92U,0x472fU,0x8e5eU,0x0cb7U,0x196eU,0x32dcU,0x65b8U,0xcb70U,0x86ebU,0x1dddU,
		0x3bbaU,0x7774U,0xeee8U,0xcddbU,0x8bbdU,0x0771U,0x0ee2U,0x1dc4U,0x3b88U,0x7710U,
		0xee20U,0xcc4bU,0x889dU,0x0131U,0x0262U,0x04c4U,0x0988U,0x1310U,0x2620U,0x4c40U,
		0x9880U,0x210bU,0x4216U,0x842cU,0x1853U,0x30a6U,0x614cU,0xc298U,0x953bU,0x3a7dU,
		0x74faU,0xe9f4U,0xc3e3U,0x97cdU,0x3f91U,0x7f22U,0xfe44U,0xec83U,0xc90dU,0x8211U,
		0x1429U,0x2852U,0x50a4U,0xa148U,0x529bU,0xa536U,0x5a67U,0xb4ceU,0x7997U,0xf32eU,
		0xf657U,0xfca5U,0xe941U,0xc289U,0x9519U,0x3a39U,0x7472U,0xe8e4U,0xc1c3U,0x938dU,
		0x3711U,0x6e22U,0xdc44U,0xa883U,0x410dU,0x821aU,0x143fU,0x287eU,0x50fcU,0xa1f8U,
		0x53fbU,0xa7f6U,0x5fe7U,0xbfceU,0x6f97U,0xdf2eU,0xae57U,0x4ca5U,0x994aU,0x229fU,
		0x453eU,0x8a7cU,0x04f3U,0x09e6U,0x13ccU,0x2798U,0x4f30U,0x9e60U,0x2ccbU,0x5996U,
		0xb32cU,0x7653U,0xeca6U,0xc947U,0x8285U,0x1501U,0x2a02U,0x5404U,0xa808U,0x401bU,
		0x8036U,0x1067U,0x20ceU,0x419cU,0x8338U,0x167bU,0x2cf6U,0x59ecU,0xb3d8U,0x77bbU,
		0xef76U,0xcee7U,0x8dc5U,0x0b81U,0x1702U,0x2e04U,0x5c08U,0xb810U,0x602bU,0xc056U,
		0x90a7U,0x3145U,0x628aU,0xc514U,0x9a23U,0x244dU,0x489aU,0x9134U,0x3263U,0x64c6U,
		0xc98cU,0x8313U,0x162dU,0x2c5aU,0x58b4U,0xb168U,0x72dbU,0xe5b6U,0xdb67U,0xa6c5U,
		0x5d81U,0xbb02U,0x660fU,0xcc1eU,0x8837U,0x0065U,0x00caU,0x0194U,0x0328U,0x0650U,
		0x0ca0U,0x1940U,0x3280U,0x6500U,0xca00U,0x840bU,0x181dU,0x303aU,0x6074U,0xc0e8U,
		0x91dbU,0x33bdU,0x677aU,0xcef4U,0x8de3U,0x0bcdU,0x179aU,0x2f34U,0x5e68U,0xbcd0U,
		0x69abU,0xd356U,0xb6a7U,0x7d45U,0xfa8aU,0xe51fU,0xda35U,0xa461U,0x58c9U,0xb192U,
		0x732fU,0xe65eU,0xdcb7U,0xa965U,0x42c1U,0x8582U,0x1b0fU,0x361eU,0x6c3cU,0xd878U,
		0xa0fbU,0x51fdU,0xa3faU,0x57ffU,0xaffeU,0x4ff7U,0x9feeU,0x2fd7U,0x5faeU,0xbf5cU,
		0x6eb3U,0xdd66U,0xaac7U,0x4585U,0x8b0aU,0x061fU,0x0c3eU,0x187cU,0x30f8U,0x61f0U,
		0xc3e0U,0x97cbU,0x3f9dU,0x7f3aU,0xfe74U,0xece3U,0xc9cdU,0x8391U,0x1729U,0x2e52U,
		0x5ca4U,0xb948U,0x629bU,0xc536U,0x9a67U,0x24c5U,0x498aU,0x9314U,0x3623U,0x6c46U,
		0xd88cU,0xa113U,0x522dU,0xa45aU,0x58bfU,0xb17eU,0x72f7U,0xe5eeU,0xdbd7U,0xa7a5U,
		0x5f41U,0xbe82U,0x6d0fU,0xda1eU,0xa437U,0x5865U,0xb0caU,0x719fU,0xe33eU,0xd677U,
		0xbce5U,0x69c1U,0xd382U,0xb70fU,0x7e15U,0xfc2aU,0xe85fU,0xc0b5U,0x9161U,0x32c9U,
		0x6592U,0xcb24U,0x8643U,0x1c8dU,0x391aU,0x7234U,0xe468U,0xd8dbU,0xa1bdU,0x5371U,
		0xa6e2U,0x5dcfU,0xbb9eU,0x6737U,0xce6eU,0x8cd7U,0x09a5U,0x134aU,0x2694U,0x4d28U,
		0x9a50U,0x24abU,0x4956U,0x92acU,0x3553U,0x6aa6U,0xd54cU,0xba93U,0x652dU,0xca5aU,
		0x84bfU,0x1975U,0x32eaU,0x65d4U,0xcba8U,0x875bU,0x1ebdU,0x3d7aU,0x7af4U,0xf5e8U,
		0xfbdbU,0xe7bdU,0xdf71U,0xaee9U,0x4dd9U,0x9bb2U,0x276fU,0x4edeU,0x9dbcU,0x2b73U,
		0x56e6U,0xadccU,0x4b93U,0x9726U,0x3e47U,0x7c8eU,0xf91cU,0xe233U,0xd46dU,0xb8d1U,
		0x61a9U,0xc352U,0x96afU,0x3d55U,0x7aaaU,0xf554U,0xfaa3U,0xe54dU,0xda91U,0xa529U,
		0x5a59U,0xb4b2U,0x796fU,0xf2deU,0xf5b7U,0xfb65U,0xe6c1U,0xdd89U,0xab19U,0x4639U,
		0x8c72U,0x08efU,0x11deU,0x23bcU,0x4778U,0x8ef0U,0x0debU,0x1bd6U,0x37acU,0x6f58U,
		0xdeb0U,0xad6bU,0x4addU,0x95baU,0x3b7fU,0x76feU,0xedfcU,0xcbf3U,0x87edU,0x1fd1U,
		0x3fa2U,0x7f44U,0xfe88U,0xed1bU,0xca3dU,0x8471U,0x18e9U,0x31d2U,0x63a4U,0xc748U,
		0x9e9bU,0x2d3dU,0x5a7aU,0xb4f4U,0x79e3U,0xf3c6U,0xf787U,0xff05U,0xee01U,0xcc09U,
		0x8819U,0x0039U,0x0072U,0x00e4U,0x01c8U,0x0390U,0x0720U,0x0e40U,0x1c80U,0x3900U,
		0x7200U,0xe400U,0xd80bU,0xa01dU,0x5031U,0xa062U,0x50cfU,0xa19eU,0x5337U,0xa66eU,
		0x5cd7U,0xb9aeU,0x6357U,0xc6aeU,0x9d57U,0x2aa5U,0x554aU,0xaa94U,0x4523U,0x8a46U,
		0x0487U,0x090eU,0x121cU,0x2438U,0x4870U,0x90e0U,0x31cbU,0x6396U,0xc72cU,0x9e53U,
		0x2cadU,0x595aU,0xb2b4U,0x7563U,0xeac6U,0xc587U,0x9b05U,0x2601U,0x4c02U,0x9804U,
		0x2003U,0x4006U,0x800cU,0x1013U,0x2026U,0x404cU,0x8098U,0x113bU,0x2276U,0x44ecU,
		0x89d8U,0x03bbU,0x0776U,0x0eecU,0x1dd8U,0x3bb0U,0x7760U,0xeec0U,0xcd8bU,0x8b1dU,
		0x0631U,0x0c62U,0x18c4U,0x3188U,0x6310U,0xc620U,0x9c4bU,0x289dU,0x513aU,0xa274U,
		0x54e3U,0xa9c6U,0x4387U,0x870eU,0x1e17U,0x3c2eU,0x785cU,0xf0b8U,0xf17bU,0xf2fdU,
		0xf5f1U,0xfbe9U,0xe7d9U,0xdfb9U,0xaf79U,0x4ef9U,0x9df2U,0x2befU,0x57deU,0xafbcU,
		0x4f73U,0x9ee6U,0x2dc7U,0x5b8eU,0xb71cU,0x7e33U,0xfc66U,0xe8c7U,0xc185U,0x9301U,
		0x3609U,0x6c12U,0xd824U,0xa043U,0x508dU,0xa11aU,0x523fU,0xa47eU,0x58f7U,0xb1eeU,
		0x73d7U,0xe7aeU,0xdf57U,0xaea5U,0x4d41U,0x9a82U,0x250fU,0x4a1eU,0x943cU,0x3873U,
		0x70e6U,0xe1ccU,0xd393U,0xb72dU,0x7e51U,0xfca2U,0xe94fU,0xc295U,0x9521U,0x3a49U,
		0x7492U,0xe924U,0xc243U,0x948dU,0x3911U,0x7222U,0xe444U,0xd883U,0xa10dU,0x5211U,
		0xa422U,0x584fU,0xb09eU,0x7137U,0xe26eU,0xd4d7U,0xb9a5U,0x6341U,0xc682U,0x9d0fU,
		0x2a15U,0x542aU,0xa854U,0x40a3U,0x8146U,0x1287U,0x250eU,0x4a1cU,0x9438U,0x387bU,
		0x70f6U,0xe1ecU,0xd3d3U,0xb7adU,0x7f51U,0xfea2U,0xed4fU,0xca95U,0x8521U,0x1a49U,
		0x3492U,0x6924U,0xd248U,0xb49bU,0x793dU,0xf27aU,0xf4ffU,0xf9f5U,0xe3e1U,0xd7c9U,
		0xbf99U,0x6f39U,0xde72U,0xacefU,0x49d5U,0x93aaU,0x375fU,0x6ebeU,0xdd7cU,0xaaf3U,
		0x45edU,0x8bdaU,0x07bfU,0x0f7eU,0x1efcU,0x3df8U,0x7bf0U,0xf7e0U,0xffcbU,0xef9dU,
		0xcf31U,0x8e69U,0x0cd9U,0x19b2U,0x3364U,0x66c8U,0xcd90U,0x8b2bU,0x065dU,0x0cbaU,
		0x1974U,0x32e8U,0x65d0U,0xcba0U,0x874bU,0x1e9dU,0x3d3aU,0x7a74U,0xf4e8U,0xf9dbU,
		0xe3bdU,0xd771U,0xbee9U,0x6dd9U,0xdbb2U,0xa76fU,0x5ed5U,0xbdaaU,0x6b5fU,0xd6beU,
		0xbd77U,0x6ae5U,0xd5caU,0xbb9fU,0x6735U,0xce6aU,0x8cdfU,0x09b5U,0x136aU,0x26d4U,
		0x4da8U,0x9b50U,0x26abU,0x4d56U,0x9aacU,0x2553U,0x4aa6U,0x954cU,0x3a93U,0x7526U,
		0xea4cU,0xc493U,0x992dU,0x2251U,0x44a2U,0x8944U,0x0283U,0x0506U,0x0a0cU,0x1418U,
		0x2830U,0x5060U,0xa0c0U,0x518bU,0xa316U,0x5627U,0xac4eU,0x4897U,0x912eU,0x3257U,
		0x64aeU,0xc95cU,0x82b3U,0x156dU,0x2adaU,0x55b4U,0xab68U,0x46dbU,0x8db6U,0x0b67U,
		0x16ceU,0x2d9cU,0x5b38U,0xb670U,0x7cebU,0xf9d6U,0xe3a7U,0xd745U,0xbe81U,0x6d09U,
		0xda12U,0xa42fU,0x5855U,0xb0aaU,0x715fU,0xe2beU,0xd577U,0xbae5U,0x65c1U,0xcb82U,
		0x870fU,0x1e15U,0x3c2aU,0x7854U,0xf0a8U,0xf15bU,0xf2bdU,0xf571U,0xfae9U,0xe5d9U,
		0xdbb9U,0xa779U,0x5ef9U,0xbdf2U,0x6befU,0xd7deU,0xbfb7U,0x6f65U,0xdecaU,0xad9fU,
		0x4b35U,0x966aU,0x3cdfU,0x79beU,0xf37cU,0xf6f3U,0xfdedU,0xebd1U,0xc7a9U,0x9f59U,
		0x2eb9U,0x5d72U,0xbae4U,0x65c3U,0xcb86U,0x8707U,0x1e05U,0x3c0aU,0x7814U,0xf028U,
		0xf05bU,0xf0bdU,0xf171U,0xf2e9U,0xf5d9U,0xfbb9U,0xe779U,0xdef9U,0xadf9U,0x4bf9U,
		0x97f2U,0x3fefU,0x7fdeU,0xffbcU,0xef73U,0xceedU,0x8dd1U,0x0ba9U,0x1752U,0x2ea4U,
		0x5d48U,0xba90U,0x652bU,0xca56U,0x84a7U,0x1945U,0x328aU,0x6514U,0xca28U,0x845bU,
		0x18bdU,0x317aU,0x62f4U,0xc5e8U,0x9bdbU,0x27bdU,0x4f7aU,0x9ef4U,0x2de3U,0x5bc6U,
		0xb78cU,0x7f13U,0xfe26U,0xec47U,0xc885U,0x8101U,0x1209U,0x2412U,0x4824U,0x9048U,
		0x309bU,0x6136U,0xc26cU,0x94d3U,0x39adU,0x735aU,0xe6b4U,0xdd63U,0xaacdU,0x4591U,
		0x8b22U,0x064fU,0x0c9eU,0x193cU,0x3278U,0x64f0U,0xc9e0U,0x83cbU,0x179dU,0x2f3aU,
		0x5e74U,0xbce8U,0x69dbU,0xd3b6U,0xb767U,0x7ec5U,0xfd8aU,0xeb1fU,0xc635U,0x9c61U,
		0x28c9U,0x5192U,0xa324U,0x5643U,0xac86U,0x4907U,0x920eU,0x3417U,0x682eU,0xd05cU,
		0xb0b3U,0x716dU,0xe2daU,0xd5bfU,0xbb75U,0x66e1U,0xcdc2U,0x8b8fU,0x0715U,0x0e2aU,
		0x1c54U,0x38a8U,0x7150U,0xe2a0U,0xd54bU,0xba9dU,0x6531U,0xca62U,0x84cfU,0x1995U,
		0x332aU,0x6654U,0xcca8U,0x895bU,0x02bdU,0x057aU,0x0af4U,0x15e8U,0x2bd0U,0x57a0U,
		0xaf40U,0x4e8bU,0x9d16U,0x2a27U,0x544eU,0xa89cU,0x4133U,0x8266U,0x14c7U,0x298eU,
		0x531cU,0xa638U,0x5c7bU,0xb8f6U,0x61e7U,0xc3ceU,0x9797U,0x3f25U,0x7e4aU,0xfc94U,
		0xe923U,0xc24dU,0x9491U,0x3929U,0x7252U,0xe4a4U,0xd943U,0xa28dU,0x5511U,0xaa22U,
		0x444fU,0x889eU,0x0137U,0x026eU,0x04dcU,0x09b8U,0x1370U,0x26e0U,0x4dc0U,0x9b80U,
		0x270bU,0x4e16U,0x9c2cU,0x2853U,0x50a6U,0xa14cU,0x5293U,0xa526U,0x5a47U,0xb48eU,
		0x7917U,0xf22eU,0xf457U,0xf8a5U,0xe141U,0xd289U,0xb519U,0x7a39U,0xf472U,0xf8efU,
		0xe1d5U,0xd3a1U,0xb749U,0x7e99U,0xfd32U,0xea6fU,0xc4d5U,0x99a1U,0x2349U,0x4692U,
		0x8d24U,0x0a43U,0x1486U,0x290cU,0x5218U,0xa430U,0x586bU,0xb0d6U,0x71a7U,0xe34eU,
		0xd697U,0xbd25U,0x6a41U,0xd482U,0xb90fU,0x6215U,0xc42aU,0x985fU,0x20b5U,0x416aU,
		0x82d4U,0x15a3U,0x2b46U,0x568cU,0xad18U,0x4a3bU,0x9476U,0x38e7U,0x71ceU,0xe39cU,
		0xd733U,0xbe6dU,0x6cd1U,0xd9a2U,0xa34fU,0x5695U,0xad2aU,0x4a5fU,0x94beU,0x3977U,
		0x72eeU,0xe5dcU,0xdbb3U,0xa76dU,0x5ed1U,0xbda2U,0x6b4fU,0xd69eU,0xbd37U,0x6a65U,
		0xd4caU,0xb99fU,0x6335U,0xc66aU,0x9cdfU,0x29b5U,0x536aU,0xa6d4U,0x5da3U,0xbb46U,
		0x6687U,0xcd0eU,0x8a17U,0x0425U,0x084aU,0x1094U,0x2128U,0x4250U,0x84a0U,0x194bU,
		0x3296U,0x652cU,0xca58U,0x84bbU,0x197dU,0x32faU,0x65f4U,0xcbe8U,0x87dbU,0x1fbdU,
		0x3f7aU,0x7ef4U,0xfde8U,0xebdbU,0xc7bdU,0x9f71U,0x2ee9U,0x5dd2U,0xbba4U,0x6743U,
		0xce86U,0x8d07U,0x0a05U,0x140aU,0x2814U,0x5028U,0xa050U,0x50abU,0xa156U,0x52a7U,
		0xa54eU,0x5a97U,0xb52eU,0x7a57U,0xf4aeU,0xf957U,0xe2a5U,0xd541U,0xba89U,0x6519U,
		0xca32U,0x846fU,0x18d5U,0x31aaU,0x6354U,0xc6a8U,0x9d5bU,0x2abdU,0x557aU,0xaaf4U,
		0x45e3U,0x8bc6U,0x0787U,0x0f0eU,0x1e1cU,0x3c38U,0x7870U,0xf0e0U,0xf1cbU,0xf39dU,
		0xf731U,0xfe69U,0xecd9U,0xc9b9U,0x8379U,0x16f9U,0x2df2U,0x5be4U,0xb7c8U,0x7f9bU,
		0xff36U,0xee67U,0xccc5U,0x8981U,0x0309U,0x0612U,0x0c24U,0x1848U,0x3090U,0x6120U,
		0xc240U,0x948bU,0x391dU,0x723aU,0xe474U,0xd8e3U,0xa1cdU,0x5391U,0xa722U,0x5e4fU,
		0xbc9eU,0x6937U,0xd26eU,0xb4d7U,0x79a5U,0xf34aU,0xf69fU,0xfd35U,0xea61U,0xc4c9U,
		0x9999U,0x2339U,0x4672U,0x8ce4U,0x09c3U,0x1386U,0x270cU,0x4e18U,0x9c30U,0x286bU,
		0x50d6U,0xa1acU,0x5353U,0xa6a6U,0x5d47U,0xba8eU,0x6517U,0xca2eU,0x8457U,0x18a5U,
		0x314aU,0x6294U,0xc528U,0x9a5bU,0x24bdU,0x497aU,0x92f4U,0x35e3U,0x6bc6U,0xd78cU,
		0xbf13U,0x6e2dU,0xdc5aU,0xa8bfU,0x4175U,0x82eaU,0x15dfU,0x2bbeU,0x577cU,0xaef8U,
		0x4dfbU,0x9bf6U,0x27e7U,0x4fceU,0x9f9cU,0x2f33U,0x5e66U,0xbcccU,0x6993U,0xd326U,
		0xb647U,0x7c85U,0xf90aU,0xe21fU,0xd435U,0xb861U,0x60c9U,0xc192U,0x932fU,0x3655U,
		0x6caaU,0xd954U,0xa2a3U,0x554dU,0xaa9aU,0x453fU,0x8a7eU,0x04f7U,0x09eeU,0x13dcU,
		0x27b8U,0x4f70U,0x9ee0U,0x2dcbU,0x5b96U,0xb72cU,0x7e53U,0xfca6U,0xe947U,0xc285U,
		0x9501U,0x3a09U,0x7412U,0xe824U,0xc043U,0x908dU,0x3111U,0x6222U,0xc444U,0x9883U,
		0x210dU,0x421aU,0x8434U,0x1863U,0x30c6U,0x618cU,0xc318U,0x963bU,0x3c7dU,0x78faU,
		0xf1f4U,0xf3e3U,0xf7cdU,0xff91U,0xef29U,0xce59U,0x8cb9U,0x0979U,0x12f2U,0x25e4U,
		0x4bc8U,0x9790U,0x3f2bU,0x7e56U,0xfcacU,0xe953U,0xc2adU,0x9551U,0x3aa9U,0x7552U,
		0xeaa4U,0xc543U,0x9a8dU,0x2511U,0x4a22U,0x9444U,0x3883U,0x7106U,0xe20cU,0xd413U,
		0xb82dU,0x6051U,0xc0a2U,0x914fU,0x3295U,0x652aU,0xca54U,0x84a3U,0x194dU,0x329aU,
		0x6534U,0xca68U,0x84dbU,0x19bdU,0x337aU,0x66f4U,0xcde8U,0x8bdbU,0x07bdU,0x0f7aU,
		0x1ef4U,0x3de8U,0x7bd0U,0xf7a0U,0xff4bU,0xee9dU,0xcd31U,0x8a69U,0x04d9U,0x09b2U,
		0x1364U,0x26c8U,0x4d90U,0x9b20U,0x264bU,0x4c96U,0x992cU,0x2253U,0x44a6U,0x894cU,
		0x0293U,0x0526U,0x0a4cU,0x1498U,0x2930U,0x5260U,0xa4c0U,0x598bU,0xb316U,0x7627U,
		0xec4eU,0xc897U,0x8125U,0x1241U,0x2482U,0x4904U,0x9208U,0x341bU,0x6836U,0xd06cU,
		0xb0d3U,0x71adU,0xe35aU,0xd6bfU,0xbd75U,0x6ae1U,0xd5c2U,0xbb8fU,0x6715U,0xce2aU,
		0x8c5fU,0x08b5U,0x116aU,0x22d4U,0x45a8U,0x8b50U,0x06abU,0x0d56U,0x1aacU,0x3558U,
		0x6ab0U,0xd560U,0xbacbU,0x659dU,0xcb3aU,0x867fU,0x1cf5U,0x39eaU,0x73d4U,0xe7a8U,
		0xdf5bU,0xaebdU,0x4d71U,0x9ae2U,0x25cfU,0x4b9eU,0x973cU,0x3e73U,0x7ce6U,0xf9ccU,
		0xe393U,0xd72dU,0xbe51U,0x6ca9U,0xd952U,0xa2afU,0x5555U,0xaaaaU,0x455fU,0x8abeU,
		0x0577U,0x0aeeU,0x15dcU,0x2bb8U,0x5770U,0xaee0U,0x4dcbU,0x9b96U,0x2727U,0x4e4eU,
		0x9c9cU,0x2933U,0x5266U,0xa4ccU,0x5993U,0xb326U,0x7647U,0xec8eU,0xc917U,0x8225U,
		0x1441U,0x2882U,0x5104U,0xa208U,0x541bU,0xa836U,0x4067U,0x80ceU,0x1197U,0x232eU,
		0x465cU,0x8cb8U,0x097bU,0x12f6U,0x25ecU,0x4bd8U,0x97b0U,0x3f6bU,0x7ed6U,0xfdacU,
		0xeb53U,0xc6adU,0x9d51U,0x2aa9U,0x5552U,0xaaa4U,0x4543U,0x8a86U,0x0507U,0x0a0eU,
		0x141cU,0x2838U,0x5070U,0xa0e0U,0x51cbU,0xa396U,0x5727U,0xae4eU,0x4c97U,0x992eU,
		0x2257U,0x44aeU,0x895cU,0x02b3U,0x0566U,0x0accU,0x1598U,0x2b30U,0x5660U,0xacc0U,
		0x498bU,0x9316U,0x3627U,0x6c4eU,0xd89cU,0xa133U,0x526dU,0xa4daU,0x59bfU,0xb37eU,
		0x76f7U,0xedeeU,0xcbd7U,0x87a5U,0x1f41U,0x3e82U,0x7d04U,0xfa08U,0xe41bU,0xd83dU,
		0xa071U,0x50e9U,0xa1d2U,0x53afU,0xa75eU,0x5eb7U,0xbd6eU,0x6ad7U,0xd5aeU,0xbb57U,
		0x66a5U,0xcd4aU,0x8a9fU,0x0535U,0x0a6aU,0x14d4U,0x29a8U,0x5350U,0xa6a0U,0x5d4bU,
		0xba96U,0x6527U,0xca4eU,0x8497U,0x1925U,0x324aU,0x6494U,0xc928U,0x825bU,0x14bdU,
		0x297aU,0x52f4U,0xa5e8U,0x5bdbU,0xb7b6U,0x7f67U,0xfeceU,0xed97U,0xcb25U,0x8641U,
		0x1c89U,0x3912U,0x7224U,0xe448U,0xd89bU,0xa13dU,0x5271U,0xa4e2U,0x59cfU,0xb39eU,
		0x7737U,0xee6eU,0xccd7U,0x89a5U,0x0341U,0x0682U,0x0d04U,0x1a08U,0x3410U,0x6820U,
		0xd040U,0xb08bU,0x711dU,0xe23aU,0xd47fU,0xb8f5U,0x61e1U,0xc3c2U,0x978fU,0x3f15U,
		0x7e2aU,0xfc54U,0xe8a3U,0xc14dU,0x9291U,0x3529U,0x6a52U,0xd4a4U,0xb943U,0x628dU,
		0xc51aU,0x9a3fU,0x2475U,0x48eaU,0x91d4U,0x33a3U,0x6746U,0xce8cU,0x8d13U,0x0a2dU,
		0x145aU,0x28b4U,0x5168U,0xa2d0U,0x55abU,0xab56U,0x46a7U,0x8d4eU,0x0a97U,0x152eU,
		0x2a5cU,0x54b8U,0xa970U,0x42ebU,0x85d6U,0x1ba7U,0x374eU,0x6e9cU,0xdd38U,0xaa7bU,
		0x44fdU,0x89faU,0x03ffU,0x07feU,0x0ffcU,0x1ff8U,0x3ff0U,0x7fe0U,0xffc0U,0xef8bU,
		0xcf1dU,0x8e31U,0x0c69U,0x18d2U,0x31a4U,0x6348U,0xc690U,0x9d2bU,0x2a5dU,0x54baU,
		0xa974U,0x42e3U,0x85c6U,0x1b87U,0x370eU,0x6e1cU,0xdc38U,0xa87bU,0x40fdU,0x81faU,
		0x13ffU,0x27feU,0x4ffcU,0x9ff8U,0x2ffbU,0x5ff6U,0xbfecU,0x6fd3U,0xdfa6U,0xaf47U,
		0x4e85U,0x9d0aU,0x2a1fU,0x543eU,0xa87cU,0x40f3U,0x81e6U,0x13c7U,0x278eU,0x4f1cU,
		0x9e38U,0x2c7bU,0x58f6U,0xb1ecU,0x73d3U,0xe7a6U,0xdf47U,0xae85U,0x4d01U,0x9a02U,
		0x240fU,0x481eU,0x903cU,0x3073U,0x60e6U,0xc1ccU,0x9393U,0x372dU,0x6e5aU,0xdcb4U,
		0xa963U,0x42cdU,0x859aU,0x1b3fU,0x367eU,0x6cfcU,0xd9f8U,0xa3fbU,0x57fdU,0xaffaU,
		0x4fffU,0x9ffeU,0x2ff7U,0x5feeU,0xbfdcU,0x6fb3U,0xdf66U,0xaec7U,0x4d85U,0x9b0aU,
		0x261fU,0x4c3eU,0x987cU,0x20f3U,0x41e6U,0x83ccU,0x1793U,0x2f26U,0x5e4cU,0xbc98U,
		0x693bU,0xd276U,0xb4e7U,0x79c5U,0xf38aU,0xf71fU,0xfe35U,0xec61U,0xc8c9U,0x8199U,
		0x1339U,0x2672U,0x4ce4U,0x99c8U,0x239bU,0x4736U,0x8e6cU,0x0cd3U,0x19a6U,0x334cU,
		0x6698U,0xcd30U,0x8a6bU,0x04ddU,0x09baU,0x1374U,0x26e8U,0x4dd0U,0x9ba0U,0x274bU,
		0x4e96U,0x9d2cU,0x2a53U,0x54a6U,0xa94cU,0x4293U,0x8526U,0x1a47U,0x348eU,0x691cU,
		0xd238U,0xb47bU,0x78fdU,0xf1faU,0xf3ffU,0xf7f5U,0xffe1U,0xefc9U,0xcf99U,0x8f39U,
		0x0e79U,0x1cf2U,0x39e4U,0x73c8U,0xe790U,0xdf2bU,0xae5dU,0x4cb1U,0x9962U,0x22cfU,
		0x459eU,0x8b3cU,0x0673U,0x0ce6U,0x19ccU,0x3398U,0x6730U,0xce60U,0x8ccbU,0x099dU,
		0x133aU,0x2674U,0x4ce8U,0x99d0U,0x23abU,0x4756U,0x8eacU,0x0d53U,0x1aa6U,0x354cU,
		0x6a98U,0xd530U,0xba6bU,0x64ddU,0xc9baU,0x837fU,0x16f5U,0x2deaU,0x5bd4U,0xb7a8U,
		0x7f5bU,0xfeb6U,0xed67U,0xcac5U,0x8581U,0x1b09U,0x3612U,0x6c24U,0xd848U,0xa09bU,
		0x513dU,0xa27aU,0x54ffU,0xa9feU,0x43f7U,0x87eeU,0x1fd7U,0x3faeU,0x7f5cU,0xfeb8U,
		0xed7bU,0xcafdU,0x85f1U,0x1be9U,0x37d2U,0x6fa4U,0xdf48U,0xae9bU,0x4d3dU,0x9a7aU,
		0x24ffU,0x49feU,0x93fcU,0x37f3U,0x6fe6U,0xdfccU,0xaf93U,0x4f2dU,0x9e5aU,0x2cbfU,
		0x597eU,0xb2fcU,0x75f3U,0xebe6U,0xc7c7U,0x9f85U,0x2f01U,0x5e02U,0xbc04U,0x6803U,
		0xd006U,0xb007U,0x7005U,0xe00aU,0xd01fU,0xb035U,0x7061U,0xe0c2U,0xd18fU,0xb315U,
		0x7621U,0xec42U,0xc88fU,0x8115U,0x1221U,0x2442U,0x4884U,0x9108U,0x321bU,0x6436U,
		0xc86cU,0x80d3U,0x11adU,0x235aU,0x46b4U,0x8d68U,0x0adbU,0x15b6U,0x2b6cU,0x56d8U,
		0xadb0U,0x4b6bU,0x96d6U,0x3da7U,0x7b4eU,0xf69cU,0xfd33U,0xea6dU,0xc4d1U,0x99a9U,
		0x2359U,0x46b2U,0x8d64U,0x0ac3U,0x1586U,0x2b0cU,0x5618U,0xac30U,0x486bU,0x90d6U,
		0x31a7U,0x634eU,0xc69cU,0x9d33U,0x2a6dU,0x54daU,0xa9b4U,0x4363U,0x86c6U,0x1d87U,
		0x3b0eU,0x761cU,0xec38U,0xc87bU,0x80fdU,0x11f1U,0x23e2U,0x47c4U,0x8f88U,0x0f1bU,
		0x1e36U,0x3c6cU,0x78d8U,0xf1b0U,0xf36bU,0xf6ddU,0xfdb1U,0xeb69U,0xc6d9U,0x9db9U,
		0x2b79U,0x56f2U,0xade4U,0x4bc3U,0x9786U,0x3f07U,0x7e0eU,0xfc1cU,0xe833U,0xc06dU,
		0x90d1U,0x31a9U,0x6352U,0xc6a4U,0x9d43U,0x2a8dU,0x551aU,0xaa34U,0x4463U,0x88c6U,
		0x0187U,0x030eU,0x061cU,0x0c38U,0x1870U,0x30e0U,0x61c0U,0xc380U,0x970bU,0x3e1dU,
		0x7c3aU,0xf874U,0xe0e3U,0xd1cdU,0xb391U,0x7729U,0xee52U,0xccafU,0x8955U,0x02a1U,
		0x0542U,0x0a84U,0x1508U,0x2a10U,0x5420U,0xa840U,0x408bU,0x8116U,0x1227U,0x244eU,
		0x489cU,0x9138U,0x327bU,0x64f6U,0xc9ecU,0x83d3U,0x17adU,0x2f5aU,0x5eb4U,0xbd68U,
		0x6adbU,0xd5b6U,0xbb67U,0x66c5U,0xcd8aU,0x8b1fU,0x0635U,0x0c6aU,0x18d4U,0x31a8U,
		0x6350U,0xc6a0U,0x9d4bU,0x2a9dU,0x553aU,0xaa74U,0x44e3U,0x89c6U,0x0387U,0x070eU,
		0x0e1cU,0x1c38U,0x3870U,0x70e0U,0xe1c0U,0xd38bU,0xb71dU,0x7e31U,0xfc62U,0xe8cfU,
		0xc195U,0x9321U,0x3649U,0x6c92U,0xd924U,0xa243U,0x548dU,0xa91aU,0x423fU,0x847eU,
		0x18f7U,0x31eeU,0x63dcU,0xc7b8U,0x9f7bU,0x2efdU,0x5dfaU,0xbbf4U,0x67e3U,0xcfc6U,
		0x8f87U,0x0f05U,0x1e0aU,0x3c14U,0x7828U,0xf050U,0xf0abU,0xf15dU,0xf2b1U,0xf569U,
		0xfad9U,0xe5b9U,0xdb79U,0xa6f9U,0x5df9U,0xbbf2U,0x67efU,0xcfdeU,0x8fb7U,0x0f65U,
		0x1ecaU,0x3d94U,0x7b28U,0xf650U,0xfcabU,0xe95dU,0xc2b1U,0x9569U,0x3ad9U,0x75b2U,
		0xeb64U,0xc6c3U,0x9d8dU,0x2b11U,0x5622U,0xac44U,0x4883U,0x9106U,0x3207U,0x640eU,
		0xc81cU,0x8033U,0x106dU,0x20daU,0x41b4U,0x8368U,0x16dbU,0x2db6U,0x5b6cU,0xb6d8U,
		0x7dbbU,0xfb76U,0xe6e7U,0xddc5U,0xab81U,0x4709U,0x8e12U,0x0c2fU,0x185eU,0x30bcU,
		0x6178U,0xc2f0U,0x95ebU,0x3bddU,0x77baU,0xef74U,0xcee3U,0x8dcdU,0x0b91U,0x1722U,
		0x2e44U,0x5c88U,0xb910U,0x622bU,0xc456U,0x98a7U,0x2145U,0x428aU,0x8514U,0x1a23U,
		0x3446U,0x688cU,0xd118U,0xb23bU,0x747dU,0xe8faU,0xc1ffU,0x93f5U,0x37e1U,0x6fc2U,
		0xdf84U,0xaf03U,0x4e0dU,0x9c1aU,0x283fU,0x507eU,0xa0fcU,0x51f3U,0xa3e6U,0x57c7U,
		0xaf8eU,0x4f17U,0x9e2eU,0x2c57U,0x58aeU,0xb15cU,0x72b3U,0xe566U,0xdac7U,0xa585U,
		0x5b01U,0xb602U,0x7c0fU,0xf81eU,0xe037U,0xd065U,0xb0c1U,0x7189U,0xe312U,0xd62fU,
		0xbc55U,0x68a1U,0xd142U,0xb28fU,0x7515U,0xea2aU,0xc45fU,0x98b5U,0x2161U,0x42c2U,
		0x8584U,0x1b03U,0x3606U,0x6c0cU,0xd818U,0xa03bU,0x507dU,0xa0faU,0x51ffU,0xa3feU,
		0x57f7U,0xafeeU,0x4fd7U,0x9faeU,0x2f57U,0x5eaeU,0xbd5cU,0x6ab3U,0xd566U,0xbac7U,
		0x6585U,0xcb0aU,0x861fU,0x1c35U,0x386aU,0x70d4U,0xe1a8U,0xd35bU,0xb6bdU,0x7d71U,
		0xfae2U,0xe5cfU,0xdb95U,0xa721U,0x5e49U,0xbc92U,0x692fU,0xd25eU,0xb4b7U,0x7965U,
		0xf2caU,0xf59fU,0xfb35U,0xe661U,0xdcc9U,0xa999U,0x4339U,0x8672U,0x1cefU,0x39deU,
		0x73bcU,0xe778U,0xdefbU,0xadfdU,0x4bf1U,0x97e2U,0x3fcfU,0x7f9eU,0xff3cU,0xee73U,
		0xccedU,0x89d1U,0x03a9U,0x0752U,0x0ea4U,0x1d48U,0x3a90U,0x7520U,0xea40U,0xc48bU,
		0x991dU,0x2231U,0x4462U,0x88c4U,0x0183U,0x0306U,0x060cU,0x0c18U,0x1830U,0x3060U,
		0x60c0U,0xc180U,0x930bU,0x361dU,0x6c3aU,0xd874U,0xa0e3U,0x51cdU,0xa39aU,0x573fU,
		0xae7eU,0x4cf7U,0x99eeU,0x23d7U,0x47aeU,0x8f5cU,0x0eb3U,0x1d66U,0x3accU,0x7598U,
		0xeb30U,0xc66bU,0x9cddU,0x29b1U,0x5362U,0xa6c4U,0x5d83U,0xbb06U,0x6607U,0xcc0eU,
		0x8817U,0x0025U,0x004aU,0x0094U,0x0128U,0x0250U,0x04a0U,0x0940U,0x1280U,0x2500U,
		0x4a00U,0x9400U,0x380bU,0x7016U,0xe02cU,0xd053U,0xb0adU,0x7151U,0xe2a2U,0xd54fU,
		0xba95U,0x6521U,0xca42U,0x848fU,0x1915U,0x322aU,0x6454U,0xc8a8U,0x815bU,0x12bdU,
		0x257aU,0x4af4U,0x95e8U,0x3bdbU,0x77b6U,0xef6cU,0xced3U,0x8dadU,0x0b51U,0x16a2U,
		0x2d44U,0x5a88U,0xb510U,0x7a2bU,0xf456U,0xf8a7U,0xe145U,0xd281U,0xb509U,0x7a19U,
		0xf432U,0xf86fU,0xe0d5U,0xd1a1U,0xb349U,0x7699U,0xed32U,0xca6fU,0x84d5U,0x19a1U,
		0x3342U,0x6684U,0xcd08U,0x8a1bU,0x043dU,0x087aU,0x10f4U,0x21e8U,0x43d0U,0x87a0U,
		0x1f4bU,0x3e96U,0x7d2cU,0xfa58U,0xe4bbU,0xd97dU,0xa2f1U,0x55e9U,0xabd2U,0x47afU,
		0x8f5eU,0x0eb7U,0x1d6eU,0x3adcU,0x75b8U,0xeb70U,0xc6ebU,0x9dddU,0x2bb1U,0x5762U,
		0xaec4U,0x4d83U,0x9b06U,0x2607U,0x4c0eU,0x981cU,0x2033U,0x4066U,0x80ccU,0x1193U,
		0x2326U,0x464cU,0x8c98U,0x093bU,0x1276U,0x24ecU,0x49d8U,0x93b0U,0x376bU,0x6ed6U,
		0xddacU,0xab53U,0x46adU,0x8d5aU,0x0abfU,0x157eU,0x2afcU,0x55f8U,0xabf0U,0x47ebU,
		0x8fd6U,0x0fa7U,0x1f4eU,0x3e9cU,0x7d38U,0xfa70U,0xe4ebU,0xd9ddU,0xa3b1U,0x5769U,
		0xaed2U,0x4dafU,0x9b5eU,0x26b7U,0x4d6eU,0x9adcU,0x25b3U,0x4b66U,0x96ccU,0x3d93U,
		0x7b26U,0xf64cU,0xfc93U,0xe92dU,0xc251U,0x94a9U,0x3959U,0x72b2U,0xe564U,0xdac3U,
		0xa58dU,0x5b11U,0xb622U,0x7c4fU,0xf89eU,0xe137U,0xd265U,0xb4c1U,0x7989U,0xf312U,
		0xf62fU,0xfc55U,0xe8a1U,0xc149U,0x9299U,0x3539U,0x6a72U,0xd4e4U,0xb9c3U,0x638dU,
		0xc71aU,0x9e3fU,0x2c75U,0x58eaU,0xb1d4U,0x73a3U,0xe746U,0xde87U,0xad05U,0x4a01U,
		0x9402U,0x380fU,0x701eU,0xe03cU,0xd073U,0xb0edU,0x71d1U,0xe3a2U,0xd74fU,0xbe95U,
		0x6d21U,0xda42U,0xa48fU,0x5915U,0xb22aU,0x745fU,0xe8beU,0xc177U,0x92e5U,0x35c1U,
		0x6b82U,0xd704U,0xbe03U,0x6c0dU,0xd81aU,0xa03fU,0x5075U,0xa0eaU,0x51dfU,0xa3beU,
		0x5777U,0xaeeeU,0x4dd7U,0x9baeU,0x2757U,0x4eaeU,0x9d5cU,0x2ab3U,0x5566U,0xaaccU,
		0x4593U,0x8b26U,0x0647U,0x0c8eU,0x191cU,0x3238U,0x6470U,0xc8e0U,0x81cbU,0x139dU,
		0x273aU,0x4e74U,0x9ce8U,0x29dbU,0x53b6U,0xa76cU,0x5ed3U,0xbda6U,0x6b47U,0xd68eU,
		0xbd17U,0x6a25U,0xd44aU,0xb89fU,0x6135U,0xc26aU,0x94dfU,0x39b5U,0x736aU,0xe6d4U,
		0xdda3U,0xab4dU,0x4691U,0x8d22U,0x0a4fU,0x149eU,0x293cU,0x5278U,0xa4f0U,0x59ebU,
		0xb3d6U,0x77a7U,0xef4eU,0xce97U,0x8d25U,0x0a41U,0x1482U,0x2904U,0x5208U,0xa410U,
		0x582bU,0xb056U,0x70a7U,0xe14eU,0xd297U,0xb525U,0x7a41U,0xf482U,0xf90fU,0xe215U,
		0xd421U,0xb849U,0x6099U,0xc132U,0x926fU,0x34d5U,0x69aaU,0xd354U,0xb6a3U,0x7d4dU,
		0xfa9aU,0xe53fU,0xda75U,0xa4e1U,0x59c9U,0xb392U,0x772fU,0xee5eU,0xccb7U,0x8965U,
		0x02c1U,0x0582U,0x0b04U,0x1608U,0x2c10U,0x5820U,0xb040U,0x708bU,0xe116U,0xd227U,
		0xb445U,0x7881U,0xf102U,0xf20fU,0xf415U,0xf821U,0xe049U,0xd099U,0xb139U,0x7279U,
		0xe4f2U,0xd9efU,0xa3d5U,0x57a1U,0xaf42U,0x4e8fU,0x9d1eU,0x2a37U,0x546eU,0xa8dcU,
		0x41b3U,0x8366U,0x16c7U,0x2d8eU,0x5b1cU,0xb638U,0x7c7bU,0xf8f6U,0xe1e7U,0xd3c5U,
		0xb781U,0x7f09U,0xfe12U,0xec2fU,0xc855U,0x80a1U,0x1149U,0x2292U,0x4524U,0x8a48U,
		0x049bU,0x0936U,0x126cU,0x24d8U,0x49b0U,0x9360U,0x36cbU,0x6d96U,0xdb2cU,0xa653U,
		0x5cadU,0xb95aU,0x62bfU,0xc57eU,0x9af7U,0x25e5U,0x4bcaU,0x9794U,0x3f23U,0x7e46U,
		0xfc8cU,0xe913U,0xc22dU,0x9451U,0x38a9U,0x7152U,0xe2a4U,0xd543U,0xba8dU,0x6511U,
		0xca22U,0x844fU,0x1895U,0x312aU,0x6254U,0xc4a8U,0x995bU,0x22bdU,0x457aU,0x8af4U,
		0x05e3U,0x0bc6U,0x178cU,0x2f18U,0x5e30U,0xbc60U,0x68cbU,0xd196U,0xb327U,0x7645U,
		0xec8aU,0xc91fU,0x8235U,0x1461U,0x28c2U,0x5184U,0xa308U,0x561bU,0xac36U,0x4867U,
		0x90ceU,0x3197U,0x632eU,0xc65cU,0x9cb3U,0x296dU,0x52daU,0xa5b4U,0x5b63U,0xb6c6U,
		0x7d87U,0xfb0eU,0xe617U,0xdc25U,0xa841U,0x4089U,0x8112U,0x122fU,0x245eU,0x48bcU,
		0x9178U,0x32fbU,0x65f6U,0xcbecU,0x87d3U,0x1fadU,0x3f5aU,0x7eb4U,0xfd68U,0xeadbU,
		0xc5bdU,0x9b71U,0x26e9U,0x4dd2U,0x9ba4U,0x2743U,0x4e86U,0x9d0cU,0x2a13U,0x5426U,
		0xa84cU,0x4093U,0x8126U,0x1247U,0x248eU,0x491cU,0x9238U,0x347bU,0x68f6U,0xd1ecU,
		0xb3d3U,0x77adU,0xef5aU,0xcebfU,0x8d75U,0x0ae1U,0x15c2U,0x2b84U,0x5708U,0xae10U,
		0x4c2bU,0x9856U,0x20a7U,0x414eU,0x829cU,0x1533U,0x2a66U,0x54ccU,0xa998U,0x433bU,
		0x8676U,0x1ce7U,0x39ceU,0x739cU,0xe738U,0xde7bU,0xacfdU,0x49f1U,0x93e2U,0x37cfU,
		0x6f9eU,0xdf3cU,0xae73U,0x4cedU,0x99daU,0x23bfU,0x477eU,0x8efcU,0x0df3U,0x1be6U,
		0x37ccU,0x6f98U,0xdf30U,0xae6bU,0x4cddU,0x99baU,0x237fU,0x46feU,0x8dfcU,0x0bf3U,
		0x17e6U,0x2fccU,0x5f98U,0xbf30U,0x6e6bU,0xdcd6U,0xa9a7U,0x4345U,0x868aU,0x1d1fU,
		0x3a3eU,0x747cU,0xe8f8U,0xc1fbU,0x93fdU,0x37f1U,0x6fe2U,0xdfc4U,0xaf83U,0x4f0dU,
		0x9e1aU,0x2c3fU,0x587eU,0xb0fcU,0x71f3U,0xe3e6U,0xd7c7U,0xbf85U,0x6f01U,0xde02U,
		0xac0fU,0x4815U,0x902aU,0x305fU,0x60beU,0xc17cU,0x92f3U,0x35edU,0x6bdaU,0xd7b4U,
		0xbf63U,0x6ecdU,0xdd9aU,0xab3fU,0x4675U,0x8ceaU,0x09dfU,0x13beU,0x277cU,0x4ef8U,
		0x9df0U,0x2bebU,0x57d6U,0xafacU,0x4f53U,0x9ea6U,0x2d47U,0x5a8eU,0xb51cU,0x7a33U,
		0xf466U,0xf8c7U,0xe185U,0xd301U,0xb609U,0x7c19U,0xf832U,0xe06fU,0xd0d5U,0xb1a1U,
		0x7349U,0xe692U,0xdd2fU,0xaa55U,0x44a1U,0x8942U,0x028fU,0x051eU,0x0a3cU,0x1478U,
		0x28f0U,0x51e0U,0xa3c0U,0x578bU,0xaf16U,0x4e27U,0x9c4eU,0x2897U,0x512eU,0xa25cU,
		0x54b3U,0xa966U,0x42c7U,0x858eU,0x1b17U,0x362eU,0x6c5cU,0xd8b8U,0xa17bU,0x52fdU,
		0xa5faU,0x5bffU,0xb7feU,0x7ff7U,0xffeeU,0xefd7U,0xcfa5U,0x8f41U,0x0e89U,0x1d12U,
		0x3a24U,0x7448U,0xe890U,0xc12bU,0x925dU,0x34b1U,0x6962U,0xd2c4U,0xb583U,0x7b0dU,
		0xf61aU,0xfc3fU,0xe875U,0xc0e1U,0x91c9U,0x3399U,0x6732U,0xce64U,0x8cc3U,0x098dU,
		0x131aU,0x2634U,0x4c68U,0x98d0U,0x21abU,0x4356U,0x86acU,0x1d53U,0x3aa6U,0x754cU,
		0xea98U,0xc53bU,0x9a7dU,0x24f1U,0x49e2U,0x93c4U,0x3783U,0x6f06U,0xde0cU,0xac13U,
		0x482dU,0x905aU,0x30bfU,0x617eU,0xc2fcU,0x95f3U,0x3bedU,0x77daU,0xefb4U,0xcf63U,
		0x8ecdU,0x0d91U,0x1b22U,0x3644U,0x6c88U,0xd910U,0xa22bU,0x545dU,0xa8baU,0x417fU,
		0x82feU,0x15f7U,0x2beeU,0x57dcU,0xafb8U,0x4f7bU,0x9ef6U,0x2de7U,0x5bceU,0xb79cU,
		0x7f33U,0xfe66U,0xecc7U,0xc985U,0x8301U,0x1609U,0x2c12U,0x5824U,0xb048U,0x709bU,
		0xe136U,0xd267U,0xb4c5U,0x7981U,0xf302U,0xf60fU,0xfc15U,0xe821U,0xc049U,0x9099U,
		0x3139U,0x6272U,0xc4e4U,0x99c3U,0x238dU,0x471aU,0x8e34U,0x0c63U,0x18c6U,0x318cU,
		0x6318U,0xc630U,0x9c6bU,0x28ddU,0x51baU,0xa374U,0x56e3U,0xadc6U,0x4b87U,0x970eU,
		0x3e17U,0x7c2eU,0xf85cU,0xe0b3U,0xd16dU,0xb2d1U,0x75a9U,0xeb52U,0xc6afU,0x9d55U,
		0x2aa1U,0x5542U,0xaa84U,0x4503U,0x8a06U,0x0407U,0x080eU,0x101cU,0x2038U,0x4070U,
		0x80e0U,0x11cbU,0x2396U,0x472cU,0x8e58U,0x0cbbU,0x1976U,0x32ecU,0x65d8U,0xcbb0U,
		0x876bU,0x1eddU,0x3dbaU,0x7b74U,0xf6e8U,0xfddbU,0xebbdU,0xc771U,0x9ee9U,0x2dd9U,
		0x5bb2U,0xb764U,0x7ec3U,0xfd86U,0xeb07U,0xc605U,0x9c01U,0x2809U,0x5012U,0xa024U,
		0x5043U,0xa086U,0x5107U,0xa20eU,0x5417U,0xa82eU,0x4057U,0x80aeU,0x1157U,0x22aeU,
		0x455cU,0x8ab8U,0x057bU,0x0af6U,0x15ecU,0x2bd8U,0x57b0U,0xaf60U,0x4ecbU,0x9d96U,
		0x2b27U,0x564eU,0xac9cU,0x4933U,0x9266U,0x34c7U,0x698eU,0xd31cU,0xb633U,0x7c6dU,
		0xf8daU,0xe1bfU,0xd375U,0xb6e1U,0x7dc9U,0xfb92U,0xe72fU,0xde55U,0xaca1U,0x4949U,
		0x9292U,0x352fU,0x6a5eU,0xd4bcU,0xb973U,0x62edU,0xc5daU,0x9bbfU,0x2775U,0x4eeaU,
		0x9dd4U,0x2ba3U,0x5746U,0xae8cU,0x4d13U,0x9a26U,0x2447U,0x488eU,0x911cU,0x3233U,
		0x6466U,0xc8ccU,0x8193U,0x132dU,0x265aU,0x4cb4U,0x9968U,0x22dbU,0x45b6U,0x8b6cU,
		0x06d3U,0x0da6U,0x1b4cU,0x3698U,0x6d30U,0xda60U,0xa4cbU,0x599dU,0xb33aU,0x767fU,
		0xecfeU,0xc9f7U,0x83e5U,0x17c1U,0x2f82U,0x5f04U,0xbe08U,0x6c1bU,0xd836U,0xa067U,
		0x50c5U,0xa18aU,0x531fU,0xa63eU,0x5c77U,0xb8eeU,0x61d7U,0xc3aeU,0x9757U,0x3ea5U,
		0x7d4aU,0xfa94U,0xe523U,0xda4dU,0xa491U,0x5929U,0xb252U,0x74afU,0xe95eU,0xc2b7U,
		0x9565U,0x3ac1U,0x7582U,0xeb04U,0xc603U,0x9c0dU,0x2811U,0x5022U,0xa044U,0x5083U,
		0xa106U,0x5207U,0xa40eU,0x5817U,0xb02eU,0x7057U,0xe0aeU,0xd157U,0xb2a5U,0x7541U,
		0xea82U,0xc50fU,0x9a15U,0x2421U,0x4842U,0x9084U,0x3103U,0x6206U,0xc40cU,0x9813U,
		0x202dU,0x405aU,0x80b4U,0x1163U,0x22c6U,0x458cU,0x8b18U,0x063bU,0x0c76U,0x18ecU,
		0x31d8U,0x63b0U,0xc760U,0x9ecbU,0x2d9dU,0x5b3aU,0xb674U,0x7ce3U,0xf9c6U,0xe387U,
		0xd705U,0xbe01U,0x6c09U,0xd812U,0xa02fU,0x5055U,0xa0aaU,0x515fU,0xa2beU,0x5577U,
		0xaaeeU,0x45d7U,0x8baeU,0x0757U,0x0eaeU,0x1d5cU,0x3ab8U,0x7570U,0xeae0U,0xc5cbU,
		0x9b9dU,0x2731U,0x4e62U,0x9cc4U,0x2983U,0x5306U,0xa60cU,0x5c13U,0xb826U,0x6047U,
		0xc08eU,0x9117U,0x3225U,0x644aU,0xc894U,0x8123U,0x124dU,0x249aU,0x4934U,0x9268U,
		0x34dbU,0x69b6U,0xd36cU,0xb6d3U,0x7dadU,0xfb5aU,0xe6bfU,0xdd75U,0xaae1U,0x45c9U,
		0x8b92U,0x072fU,0x0e5eU,0x1cbcU,0x3978U,0x72f0U,0xe5e0U,0xdbcbU,0xa79dU,0x5f31U,
		0xbe62U,0x6ccfU,0xd99eU,0xa337U,0x5665U,0xaccaU,0x499fU,0x933eU,0x3677U,0x6ceeU,
		0xd9dcU,0xa3b3U,0x576dU,0xaedaU,0x4dbfU,0x9b7eU,0x26f7U,0x4deeU,0x9bdcU,0x27b3U,
		0x4f66U,0x9eccU,0x2d93U,0x5b26U,0xb64cU,0x7c93U,0xf926U,0xe247U,0xd485U,0xb901U,
		0x6209U,0xc412U,0x982fU,0x2055U,0x40aaU,0x8154U,0x12a3U,0x2546U,0x4a8cU,0x9518U,
		0x3a3bU,0x7476U,0xe8ecU,0xc1d3U,0x93adU,0x3751U,0x6ea2U,0xdd44U,0xaa83U,0x450dU,
		0x8a1aU,0x043fU,0x087eU,0x10fcU,0x21f8U,0x43f0U,0x87e0U,0x1fcbU,0x3f96U,0x7f2cU,
		0xfe58U,0xecbbU,0xc97dU,0x82f1U,0x15e9U,0x2bd2U,0x57a4U,0xaf48U,0x4e9bU,0x9d36U,
		0x2a67U,0x54ceU,0xa99cU,0x4333U,0x8666U,0x1cc7U,0x398eU,0x731cU,0xe638U,0xdc7bU,
		0xa8fdU,0x41f1U,0x83e2U,0x17cfU,0x2f9eU,0x5f3cU,0xbe78U,0x6cfbU,0xd9f6U,0xa3e7U,
		0x57c5U,0xaf8aU,0x4f1fU,0x9e3eU,0x2c77U,0x58eeU,0xb1dcU,0x73b3U,0xe766U,0xdec7U,
		0xad85U,0x4b01U,0x9602U,0x3c0fU,0x781eU,0xf03cU,0xf073U,0xf0edU,0xf1d1U,0xf3a9U,
		0xf759U,0xfeb9U,0xed79U,0xcaf9U,0x85f9U,0x1bf9U,0x37f2U,0x6fe4U,0xdfc8U,0xaf9bU,
		0x4f3dU,0x9e7aU,0x2cffU,0x59feU,0xb3fcU,0x77f3U,0xefe6U,0xcfc7U,0x8f85U,0x0f01U,
		0x1e02U,0x3c04U,0x7808U,0xf010U,0xf02bU,0xf05dU,0xf0b1U,0xf169U,0xf2d9U,0xf5b9U,
		0xfb79U,0xe6f9U,0xddf9U,0xabf9U,0x47f9U,0x8ff2U,0x0fefU,0x1fdeU,0x3fbcU,0x7f78U,
		0xfef0U,0xedebU,0xcbddU,0x87b1U,0x1f69U,0x3ed2U,0x7da4U,0xfb48U,0xe69bU,0xdd3dU,
		0xaa71U,0x44e9U,0x89d2U,0x03afU,0x075eU,0x0ebcU,0x1d78U,0x3af0U,0x75e0U,0xebc0U,
		0xc78bU,0x9f1dU,0x2e31U,0x5c62U,0xb8c4U,0x6183U,0xc306U,0x9607U,0x3c05U,0x780aU,
		0xf014U,0xf023U,0xf04dU,0xf091U,0xf129U,0xf259U,0xf4b9U,0xf979U,0xe2f9U,0xd5f9U,
		0xbbf9U,0x67f9U,0xcff2U,0x8fefU,0x0fd5U,0x1faaU,0x3f54U,0x7ea8U,0xfd50U,0xeaabU,
		0xc55dU,0x9ab1U,0x2569U,0x4ad2U,0x95a4U,0x3b43U,0x7686U,0xed0cU,0xca13U,0x842dU,
		0x1851U,0x30a2U,0x6144U,0xc288U,0x951bU,0x3a3dU,0x747aU,0xe8f4U,0xc1e3U,0x93cdU,
		0x3791U,0x6f22U,0xde44U,0xac83U,0x490dU,0x921aU,0x343fU,0x687eU,0xd0fcU,0xb1f3U,
		0x73edU,0xe7daU,0xdfbfU,0xaf75U,0x4ee1U,0x9dc2U,0x2b8fU,0x571eU,0xae3cU,0x4c73U,
		0x98e6U,0x21c7U,0x438eU,0x871cU,0x1e33U,0x3c66U,0x78ccU,0xf198U,0xf33bU,0xf67dU,
		0xfcf1U,0xe9e9U,0xc3d9U,0x97b9U,0x3f79U,0x7ef2U,0xfde4U,0xebc3U,0xc78dU,0x9f11U,
		0x2e29U,0x5c52U,0xb8a4U,0x6143U,0xc286U,0x9507U,0x3a05U,0x740aU,0xe814U,0xc023U,
		0x904dU,0x3091U,0x6122U,0xc244U,0x9483U,0x390dU,0x721aU,0xe434U,0xd863U,0xa0cdU,
		0x5191U,0xa322U,0x564fU,0xac9eU,0x4937U,0x926eU,0x34d7U,0x69aeU,0xd35cU,0xb6b3U,
		0x7d6dU,0xfadaU,0xe5bfU,0xdb75U,0xa6e1U,0x5dc9U,0xbb92U,0x672fU,0xce5eU,0x8cb7U,
		0x0965U,0x12caU,0x2594U,0x4b28U,0x9650U,0x3cabU,0x7956U,0xf2acU,0xf553U,0xfaadU,
		0xe551U,0xdaa9U,0xa559U,0x5ab9U,0xb572U,0x7aefU,0xf5deU,0xfbb7U,0xe765U,0xdec1U,
		0xad89U,0x4b19U,0x9632U,0x3c6fU,0x78deU,0xf1bcU,0xf373U,0xf6edU,0xfdd1U,0xeba9U,
		0xc759U,0x9eb9U,0x2d79U,0x5af2U,0xb5e4U,0x7bc3U,0xf786U,0xff07U,0xee05U,0xcc01U,
		0x8809U,0x0019U,0x0032U,0x0064U,0x00c8U,0x0190U,0x0320U,0x0640U,0x0c80U,0x1900U,
		0x3200U,0x6400U,0xc800U,0x800bU,0x101dU,0x203aU,0x4074U,0x80e8U,0x11dbU,0x23b6U,
		0x476cU,0x8ed8U,0x0dbbU,0x1b76U,0x36ecU,0x6dd8U,0xdbb0U,0xa76bU,0x5eddU,0xbdbaU,
		0x6b7fU,0xd6feU,0xbdf7U,0x6be5U,0xd7caU,0xbf9fU,0x6f35U,0xde6aU,0xacdfU,0x49b5U,
		0x936aU,0x36dfU,0x6dbeU,0xdb7cU,0xa6f3U,0x5dedU,0xbbdaU,0x67bfU,0xcf7eU,0x8ef7U,
		0x0de5U,0x1bcaU,0x3794U,0x6f28U,0xde50U,0xacabU,0x495dU,0x92baU,0x357fU,0x6afeU,
		0xd5fcU,0xbbf3U,0x67edU,0xcfdaU,0x8fbfU,0x0f75U,0x1eeaU,0x3dd4U,0x7ba8U,0xf750U,
		0xfeabU,0xed5dU,0xcab1U,0x8569U,0x1ad9U,0x35b2U,0x6b64U,0xd6c8U,0xbd9bU,0x6b3dU,
		0xd67aU,0xbcffU,0x69f5U,0xd3eaU,0xb7dfU,0x7fb5U,0xff6aU,0xeedfU,0xcdb5U,0x8b61U,
		0x06c9U,0x0d92U,0x1b24U,0x3648U,0x6c90U,0xd920U,0xa24bU,0x549dU,0xa93aU,0x427fU,
		0x84feU,0x19f7U,0x33eeU,0x67dcU,0xcfb8U,0x8f7bU,0x0efdU,0x1dfaU,0x3bf4U,0x77e8U,
		0xefd0U,0xcfabU,0x8f5dU,0x0eb1U,0x1d62U,0x3ac4U,0x7588U,0xeb10U,0xc62bU,0x9c5dU,
		0x28b1U,0x5162U,0xa2c4U,0x5583U,0xab06U,0x4607U,0x8c0eU,0x0817U,0x102eU,0x205cU,
		0x40b8U,0x8170U,0x12ebU,0x25d6U,0x4bacU,0x9758U,0x3ebbU,0x7d76U,0xfaecU,0xe5d3U,
		0xdbadU,0xa751U,0x5ea9U,0xbd52U,0x6aafU,0xd55eU,0xbab7U,0x6565U,0xcacaU,0x859fU,
		0x1b35U,0x366aU,0x6cd4U,0xd9a8U,0xa35bU,0x56bdU,0xad7aU,0x4affU,0x95feU,0x3bf7U,
		0x77eeU,0xefdcU,0xcfb3U,0x8f6dU,0x0ed1U,0x1da2U,0x3b44U,0x7688U,0xed10U,0xca2bU,
		0x845dU,0x18b1U,0x3162U,0x62c4U,0xc588U,0x9b1bU,0x263dU,0x4c7aU,0x98f4U,0x21e3U,
		0x43c6U,0x878cU,0x1f13U,0x3e26U,0x7c4cU,0xf898U,0xe13bU,0xd27dU,0xb4f1U,0x79e9U,
		0xf3d2U,0xf7afU,0xff55U,0xeea1U,0xcd49U,0x8a99U,0x0539U,0x0a72U,0x14e4U,0x29c8U,
		0x5390U,0xa720U,0x5e4bU,0xbc96U,0x6927U,0xd24eU,0xb497U,0x7925U,0xf24aU,0xf49fU,
		0xf935U,0xe261U,0xd4c9U,0xb999U,0x6339U,0xc672U,0x9cefU,0x29d5U,0x53aaU,0xa754U,
		0x5ea3U,0xbd46U,0x6a87U,0xd50eU,0xba17U,0x6425U,0xc84aU,0x809fU,0x1135U,0x226aU,
		0x44d4U,0x89a8U,0x035bU,0x06b6U,0x0d6cU,0x1ad8U,0x35b0U,0x6b60U,0xd6c0U,0xbd8bU,
		0x6b1dU,0xd63aU,0xbc7fU,0x68f5U,0xd1eaU,0xb3dfU,0x77b5U,0xef6aU,0xcedfU,0x8db5U,
		0x0b61U,0x16c2U,0x2d84U,0x5b08U,0xb610U,0x7c2bU,0xf856U,0xe0a7U,0xd145U,0xb281U,
		0x7509U,0xea12U,0xc42fU,0x9855U,0x20a1U,0x4142U,0x8284U,0x1503U,0x2a06U,0x540cU,
		0xa818U,0x403bU,0x8076U,0x10e7U,0x21ceU,0x439cU,0x8738U,0x1e7bU,0x3cf6U,0x79ecU,
		0xf3d8U,0xf7bbU,0xff7dU,0xeef1U,0xcde9U,0x8bd9U,0x07b9U,0x0f72U,0x1ee4U,0x3dc8U,
		0x7b90U,0xf720U,0xfe4bU,0xec9dU,0xc931U,0x8269U,0x14d9U,0x29b2U,0x5364U,0xa6c8U,
		0x5d9bU,0xbb36U,0x6667U,0xccceU,0x8997U,0x0325U,0x064aU,0x0c94U,0x1928U,0x3250U,
		0x64a0U,0xc940U,0x828bU,0x151dU,0x2a3aU,0x5474U,0xa8e8U,0x41dbU,0x83b6U,0x1767U,
		0x2eceU,0x5d9cU,0xbb38U,0x667bU,0xccf6U,0x89e7U,0x03c5U,0x078aU,0x0f14U,0x1e28U,
		0x3c50U,0x78a0U,0xf140U,0xf28bU,0xf51dU,0xfa31U,0xe469U,0xd8d9U,0xa1b9U,0x5379U,
		0xa6f2U,0x5defU,0xbbdeU,0x67b7U,0xcf6eU,0x8ed7U,0x0da5U,0x1b4aU,0x3694U,0x6d28U,
		0xda50U,0xa4abU,0x595dU,0xb2baU,0x757fU,0xeafeU,0xc5f7U,0x9be5U,0x27c1U,0x4f82U,
		0x9f04U,0x2e03U,0x5c06U,0xb80cU,0x6013U,0xc026U,0x9047U,0x3085U,0x610aU,0xc214U,
		0x9423U,0x384dU,0x709aU,0xe134U,0xd263U,0xb4cdU,0x7991U,0xf322U,0xf64fU,0xfc95U,
		0xe921U,0xc249U,0x9499U,0x3939U,0x7272U,0xe4e4U,0xd9c3U,0xa38dU,0x5711U,0xae22U,
		0x4c4fU,0x989eU,0x2137U,0x426eU,0x84dcU,0x19b3U,0x3366U,0x66ccU,0xcd98U,0x8b3bU,
		0x067dU,0x0cfaU,0x19f4U,0x33e8U,0x67d0U,0xcfa0U,0x8f4bU,0x0e9dU,0x1d3aU,0x3a74U,
		0x74e8U,0xe9d0U,0xc3abU,0x975dU,0x3eb1U,0x7d62U,0xfac4U,0xe583U,0xdb0dU,0xa611U,
		0x5c29U,0xb852U,0x60afU,0xc15eU,0x92b7U,0x3565U,0x6acaU,0xd594U,0xbb23U,0x664dU,
		0xcc9aU,0x893fU,0x0275U,0x04eaU,0x09d4U,0x13a8U,0x2750U,0x4ea0U,0x9d40U,0x2a8bU,
		0x5516U,0xaa2cU,0x4453U,0x88a6U,0x0147U,0x028eU,0x051cU,0x0a38U,0x1470U,0x28e0U,
		0x51c0U,0xa380U,0x570bU,0xae16U,0x4c27U,0x984eU,0x2097U,0x412eU,0x825cU,0x14b3U,
		0x2966U,0x52ccU,0xa598U,0x5b3bU,0xb676U,0x7ce7U,0xf9ceU,0xe397U,0xd725U,0xbe41U,
		0x6c89U,0xd912U,0xa22fU,0x5455U,0xa8aaU,0x415fU,0x82beU,0x1577U,0x2aeeU,0x55dcU,
		0xabb8U,0x477bU,0x8ef6U,0x0de7U,0x1bceU,0x379cU,0x6f38U,0xde70U,0xacebU,0x49ddU,
		0x93baU,0x377fU,0x6efeU,0xddfcU,0xabf3U,0x47edU,0x8fdaU,0x0fbfU,0x1f7eU,0x3efcU,
		0x7df8U,0xfbf0U,0xe7ebU,0xdfddU,0xafb1U,0x4f69U,0x9ed2U,0x2dafU,0x5b5eU,0xb6bcU,
		0x7d73U,0xfae6U,0xe5c7U,0xdb85U,0xa701U,0x5e09U,0xbc12U,0x682fU,0xd05eU,0xb0b7U,
		0x7165U,0xe2caU,0xd59fU,0xbb35U,0x6661U,0xccc2U,0x898fU,0x0315U,0x062aU,0x0c54U,
		0x18a8U,0x3150U,0x62a0U,0xc540U,0x9a8bU,0x251dU,0x4a3aU,0x9474U,0x38e3U,0x71c6U,
		0xe38cU,0xd713U,0xbe2dU,0x6c51U,0xd8a2U,0xa14fU,0x5295U,0xa52aU,0x5a5fU,0xb4beU,
		0x7977U,0xf2eeU,0xf5d7U,0xfba5U,0xe741U,0xde89U,0xad19U,0x4a39U,0x9472U,0x38efU,
		0x71deU,0xe3bcU,0xd773U,0xbeedU,0x6dd1U,0xdba2U,0xa74fU,0x5e95U,0xbd2aU,0x6a5fU,
		0xd4beU,0xb977U,0x62e5U,0xc5caU,0x9b9fU,0x2735U,0x4e6aU,0x9cd4U,0x29a3U,0x5346U,
		0xa68cU,0x5d13U,0xba26U,0x6447U,0xc88eU,0x8117U,0x1225U,0x244aU,0x4894U,0x9128U,
		0x325bU,0x64b6U,0xc96cU,0x82d3U,0x15adU,0x2b5aU,0x56b4U,0xad68U,0x4adbU,0x95b6U,
		0x3b67U,0x76ceU,0xed9cU,0xcb33U,0x866dU,0x1cd1U,0x39a2U,0x7344U,0xe688U,0xdd1bU,
		0xaa3dU,0x4471U,0x88e2U,0x01cfU,0x039eU,0x073cU,0x0e78U,0x1cf0U,0x39e0U,0x73c0U,
		0xe780U,0xdf0bU,0xae1dU,0x4c31U,0x9862U,0x20cfU,0x419eU,0x833cU,0x1673U,0x2ce6U,
		0x59ccU,0xb398U,0x773bU,0xee76U,0xcce7U,0x89c5U,0x0381U,0x0702U,0x0e04U,0x1c08U,
		0x3810U,0x7020U,0xe040U,0xd08bU,0xb11dU,0x7231U,0xe462U,0xd8cfU,0xa195U,0x5321U,
		0xa642U,0x5c8fU,0xb91eU,0x6237U,0xc46eU,0x98d7U,0x21a5U,0x434aU,0x8694U,0x1d23U,
		0x3a46U,0x748cU,0xe918U,0xc23bU,0x947dU,0x38f1U,0x71e2U,0xe3c4U,0xd783U,0xbf0dU,
		0x6e11U,0xdc22U,0xa84fU,0x4095U,0x812aU,0x125fU,0x24beU,0x497cU,0x92f8U,0x35fbU,
		0x6bf6U,0xd7ecU,0xbfd3U,0x6fadU,0xdf5aU,0xaebfU,0x4d75U,0x9aeaU,0x25dfU,0x4bbeU,
		0x977cU,0x3ef3U,0x7de6U,0xfbccU,0xe793U,0xdf2dU,0xae51U,0x4ca9U,0x9952U,0x22afU,
		0x455eU,0x8abcU,0x0573U,0x0ae6U,0x15ccU,0x2b98U,0x5730U,0xae60U,0x4ccbU,0x9996U,
		0x2327U,0x464eU,0x8c9cU,0x0933U,0x1266U,0x24ccU,0x4998U,0x9330U,0x366bU,0x6cd6U,
		0xd9acU,0xa353U,0x56adU,0xad5aU,0x4abfU,0x957eU,0x3af7U,0x75eeU,0xebdcU,0xc7b3U,
		0x9f6dU,0x2ed1U,0x5da2U,0xbb44U,0x6683U,0xcd06U,0x8a07U,0x0405U,0x080aU,0x1014U,
		0x2028U,0x4050U,0x80a0U,0x114bU,0x2296U,0x452cU,0x8a58U,0x04bbU,0x0976U,0x12ecU,
		0x25d8U,0x4bb0U,0x9760U,0x3ecbU,0x7d96U,0xfb2cU,0xe653U,0xdcadU,0xa951U,0x42a9U,
		0x8552U,0x1aafU,0x355eU,0x6abcU,0xd578U,0xbafbU,0x65fdU,0xcbfaU,0x87ffU,0x1ff5U,
		0x3feaU,0x7fd4U,0xffa8U,0xef5bU,0xcebdU,0x8d71U,0x0ae9U,0x15d2U,0x2ba4U,0x5748U,
		0xae90U,0x4d2bU,0x9a56U,0x24a7U,0x494eU,0x929cU,0x3533U,0x6a66U,0xd4ccU,0xb993U,
		0x632dU,0xc65aU,0x9cbfU,0x2975U,0x52eaU,0xa5d4U,0x5ba3U,0xb746U,0x7e87U,0xfd0eU,
		0xea17U,0xc425U,0x9841U,0x2089U,0x4112U,0x8224U,0x1443U,0x2886U,0x510cU,0xa218U,
		0x543bU,0xa876U,0x40e7U,0x81ceU,0x1397U,0x272eU,0x4e5cU,0x9cb8U,0x297bU,0x52f6U,
		0xa5ecU,0x5bd3U,0xb7a6U,0x7f47U,0xfe8eU,0xed17U,0xca25U,0x8441U,0x1889U,0x3112U,
		0x6224U,0xc448U,0x989bU,0x213dU,0x427aU,0x84f4U,0x19e3U,0x33c6U,0x678cU,0xcf18U,
		0x8e3bU,0x0c7dU,0x18faU,0x31f4U,0x63e8U,0xc7d0U,0x9fabU,0x2f5dU,0x5ebaU,0xbd74U,
		0x6ae3U,0xd5c6U,0xbb87U,0x6705U,0xce0aU,0x8c1fU,0x0835U,0x106aU,0x20d4U,0x41a8U,
		0x8350U,0x16abU,0x2d56U,0x5aacU,0xb558U,0x7abbU,0xf576U,0xfae7U,0xe5c5U,0xdb81U,
		0xa709U,0x5e19U,0xbc32U,0x686fU,0xd0deU,0xb1b7U,0x7365U,0xe6caU,0xdd9fU,0xab35U,
		0x4661U,0x8cc2U,0x098fU,0x131eU,0x263cU,0x4c78U,0x98f0U,0x21ebU,0x43d6U,0x87acU,
		0x1f53U,0x3ea6U,0x7d4cU,0xfa98U,0xe53bU,0xda7dU,0xa4f1U,0x59e9U,0xb3d2U,0x77afU,
		0xef5eU,0xceb7U,0x8d65U,0x0ac1U,0x1582U,0x2b04U,0x5608U,0xac10U,0x482bU,0x9056U,
		0x30a7U,0x614eU,0xc29cU,0x9533U,0x3a6dU,0x74daU,0xe9b4U,0xc363U,0x96cdU,0x3d91U,
		0x7b22U,0xf644U,0xfc83U,0xe90dU,0xc211U,0x9429U,0x3859U,0x70b2U,0xe164U,0xd2c3U,
		0xb58dU,0x7b11U,0xf622U,0xfc4fU,0xe895U,0xc121U,0x9249U,0x3499U,0x6932U,0xd264U,
		0xb4c3U,0x798dU,0xf31aU,0xf63fU,0xfc75U,0xe8e1U,0xc1c9U,0x9399U,0x3739U,0x6e72U,
		0xdce4U,0xa9c3U,0x438dU,0x871aU,0x1e3fU,0x3c7eU,0x78fcU,0xf1f8U,0xf3fbU,0xf7fdU,
		0xfff1U,0xefe9U,0xcfd9U,0x8fb9U,0x0f79U,0x1ef2U,0x3de4U,0x7bc8U,0xf790U,0xff2bU,
		0xee5dU,0xccb1U,0x8969U,0x02d9U,0x05b2U,0x0b64U,0x16c8U,0x2d90U,0x5b20U,0xb640U,
		0x7c8bU,0xf916U,0xe227U,0xd445U,0xb881U,0x6109U,0xc212U,0x942fU,0x3855U,0x70aaU,
		0xe154U,0xd2a3U,0xb54dU,0x7a91U,0xf522U,0xfa4fU,0xe495U,0xd921U,0xa249U,0x5499U,
		0xa932U,0x426fU,0x84deU,0x19b7U,0x336eU,0x66dcU,0xcdb8U,0x8b7bU,0x06fdU,0x0dfaU,
		0x1bf4U,0x37e8U,0x6fd0U,0xdfa0U,0xaf4bU,0x4e9dU,0x9d3aU,0x2a7fU,0x54feU,0xa9fcU,
		0x43f3U,0x87e6U,0x1fc7U,0x3f8eU,0x7f1cU,0xfe38U,0xec7bU,0xc8fdU,0x81f1U,0x13e9U,
		0x27d2U,0x4fa4U,0x9f48U,0x2e9bU,0x5d36U,0xba6cU,0x64d3U,0xc9a6U,0x8347U,0x1685U,
		0x2d0aU,0x5a14U,0xb428U,0x785bU,0xf0b6U,0xf167U,0xf2c5U,0xf581U,0xfb09U,0xe619U,
		0xdc39U,0xa879U,0x40f9U,0x81f2U,0x13efU,0x27deU,0x4fbcU,0x9f78U,0x2efbU,0x5df6U,
		0xbbecU,0x67d3U,0xcfa6U,0x8f47U,0x0e85U,0x1d0aU,0x3a14U,0x7428U,0xe850U,0xc0abU,
		0x915dU,0x32b1U,0x6562U,0xcac4U,0x8583U,0x1b0dU,0x361aU,0x6c34U,0xd868U,0xa0dbU,
		0x51bdU,0xa37aU,0x56ffU,0xadfeU,0x4bf7U,0x97eeU,0x3fd7U,0x7faeU,0xff5cU,0xeeb3U,
		0xcd6dU,0x8ad1U,0x05a9U,0x0b52U,0x16a4U,0x2d48U,0x5a90U,0xb520U,0x7a4bU,0xf496U,
		0xf927U,0xe245U,0xd481U,0xb909U,0x6219U,0xc432U,0x986fU,0x20d5U,0x41aaU,0x8354U,
		0x16a3U,0x2d46U,0x5a8cU,0xb518U,0x7a3bU,0xf476U,0xf8e7U,0xe1c5U,0xd381U,0xb709U,
		0x7e19U,0xfc32U,0xe86fU,0xc0d5U,0x91a1U,0x3349U,0x6692U,0xcd24U,0x8a43U,0x048dU,
		0x091aU,0x1234U,0x2468U,0x48d0U,0x91a0U,0x334bU,0x6696U,0xcd2cU,0x8a53U,0x04adU,
		0x095aU,0x12b4U,0x2568U,0x4ad0U,0x95a0U,0x3b4bU,0x7696U,0xed2cU,0xca53U,0x84adU,
		0x1951U,0x32a2U,0x6544U,0xca88U,0x851bU,0x1a3dU,0x347aU,0x68f4U,0xd1e8U,0xb3dbU,
		0x77bdU,0xef7aU,0xceffU,0x8df5U,0x0be1U,0x17c2U,0x2f84U,0x5f08U,0xbe10U,0x6c2bU,
		0xd856U,0xa0a7U,0x5145U,0xa28aU,0x551fU,0xaa3eU,0x4477U,0x88eeU,0x01d7U,0x03aeU,
		0x075cU,0x0eb8U,0x1d70U,0x3ae0U,0x75c0U,0xeb80U,0xc70bU,0x9e1dU,0x2c31U,0x5862U,
		0xb0c4U,0x7183U,0xe306U,0xd607U,0xbc05U,0x6801U,0xd002U,0xb00fU,0x7015U,0xe02aU,
		0xd05fU,0xb0b5U,0x7161U,0xe2c2U,0xd58fU,0xbb15U,0x6621U,0xcc42U,0x888fU,0x0115U,
		0x022aU,0x0454U,0x08a8U,0x1150U,0x22a0U,0x4540U,0x8a80U,0x050bU,0x0a16U,0x142cU,
		0x2858U,0x50b0U,0xa160U,0x52cbU,0xa596U,0x5b27U,0xb64eU,0x7c97U,0xf92eU,0xe257U,
		0xd4a5U,0xb941U,0x6289U,0xc512U,0x9a2fU,0x2455U,0x48aaU,0x9154U,0x32a3U,0x6546U,
		0xca8cU,0x8513U,0x1a2dU,0x345aU,0x68b4U,0xd168U,0xb2dbU,0x75bdU,0xeb7aU,0xc6ffU,
		0x9df5U,0x2be1U,0x57c2U,0xaf84U,0x4f03U,0x9e06U,0x2c07U,0x580eU,0xb01cU,0x7033U,
		0xe066U,0xd0c7U,0xb185U,0x7301U,0xe602U,0xdc0fU,0xa815U,0x4021U,0x8042U,0x108fU,
		0x211eU,0x423cU,0x8478U,0x18fbU,0x31f6U,0x63ecU,0xc7d8U,0x9fbbU,0x2f7dU,0x5efaU,
		0xbdf4U,0x6be3U,0xd7c6U,0xbf87U,0x6f05U,0xde0aU,0xac1fU,0x4835U,0x906aU,0x30dfU,
		0x61beU,0xc37cU,0x96f3U,0x3dedU,0x7bdaU,0xf7b4U,0xff63U,0xeecdU,0xcd91U,0x8b29U,
		0x0659U,0x0cb2U,0x1964U,0x32c8U,0x6590U,0xcb20U,0x864bU,0x1c9dU,0x393aU,0x7274U,
		0xe4e8U,0xd9dbU,0xa3bdU,0x5771U,0xaee2U,0x4dcfU,0x9b9eU,0x2737U,0x4e6eU,0x9cdcU,
		0x29b3U,0x5366U,0xa6ccU,0x5d93U,0xbb26U,0x6647U,0xcc8eU,0x8917U,0x0225U,0x044aU,
		0x0894U,0x1128U,0x2250U,0x44a0U,0x8940U,0x028bU,0x0516U,0x0a2cU,0x1458U,0x28b0U,
		0x5160U,0xa2c0U,0x558bU,0xab16U,0x4627U,0x8c4eU,0x0897U,0x112eU,0x225cU,0x44b8U,
		0x8970U,0x02ebU,0x05d6U,0x0bacU,0x1758U,0x2eb0U,0x5d60U,0xbac0U,0x658bU,0xcb16U,
		0x8627U,0x1c45U,0x388aU,0x7114U,0xe228U,0xd45bU,0xb8bdU,0x6171U,0xc2e2U,0x95cfU,
		0x3b95U,0x772aU,0xee54U,0xcca3U,0x894dU,0x0291U,0x0522U,0x0a44U,0x1488U,0x2910U,
		0x5220U,0xa440U,0x588bU,0xb116U,0x7227U,0xe44eU,0xd897U,0xa125U,0x5241U,0xa482U,
		0x590fU,0xb21eU,0x7437U,0xe86eU,0xc0d7U,0x91a5U,0x3341U,0x6682U,0xcd04U,0x8a03U,
		0x040dU,0x081aU,0x1034U,0x2068U,0x40d0U,0x81a0U,0x134bU,0x2696U,0x4d2cU,0x9a58U,
		0x24bbU,0x4976U,0x92ecU,0x35d3U,0x6ba6U,0xd74cU,0xbe93U,0x6d2dU,0xda5aU,0xa4bfU,
		0x5975U,0xb2eaU,0x75dfU,0xebbeU,0xc777U,0x9ee5U,0x2dc1U,0x5b82U,0xb704U,0x7e03U,
		0xfc06U,0xe807U,0xc005U,0x9001U,0x3009U,0x6012U,0xc024U,0x9043U,0x308dU,0x611aU,
		0xc234U,0x9463U,0x38cdU,0x719aU,0xe334U,0xd663U,0xbccdU,0x6991U,0xd322U,0xb64fU,
		0x7c95U,0xf92aU,0xe25fU,0xd4b5U,0xb961U,0x62c9U,0xc592U,0x9b2fU,0x2655U,0x4caaU,
		0x9954U,0x22a3U,0x4546U,0x8a8cU,0x0513U,0x0a26U,0x144cU,0x2898U,0x5130U,0xa260U,
		0x54cbU,0xa996U,0x4327U,0x864eU,0x1c97U,0x392eU,0x725cU,0xe4b8U,0xd97bU,0xa2fdU,
		0x55f1U,0xabe2U,0x47cfU,0x8f9eU,0x0f37U,0x1e6eU,0x3cdcU,0x79b8U,0xf370U,0xf6ebU,
		0xfdddU,0xebb1U,0xc769U,0x9ed9U,0x2db9U,0x5b72U,0xb6e4U,0x7dc3U,0xfb86U,0xe707U,
		0xde05U,0xac01U,0x4809U,0x9012U,0x302fU,0x605eU,0xc0bcU,0x9173U,0x32edU,0x65daU,
		0xcbb4U,0x8763U,0x1ecdU,0x3d9aU,0x7b34U,0xf668U,0xfcdbU,0xe9bdU,0xc371U,0x96e9U,
		0x3dd9U,0x7bb2U,0xf764U,0xfec3U,0xed8dU,0xcb11U,0x8629U,0x1c59U,0x38b2U,0x7164U,
		0xe2c8U,0xd59bU,0xbb3dU,0x6671U,0xcce2U,0x89cfU,0x0395U,0x072aU,0x0e54U,0x1ca8U,
		0x3950U,0x72a0U,0xe540U,0xda8bU,0xa51dU,0x5a31U,0xb462U,0x78cfU,0xf19eU,0xf337U,
		0xf665U,0xfcc1U,0xe989U,0xc319U,0x9639U,0x3c79U,0x78f2U,0xf1e4U,0xf3c3U,0xf78dU,
		0xff11U,0xee29U,0xcc59U,0x88b9U,0x0179U,0x02f2U,0x05e4U,0x0bc8U,0x1790U,0x2f20U,
		0x5e40U,0xbc80U,0x690bU,0xd216U,0xb427U,0x7845U,0xf08aU,0xf11fU,0xf235U,0xf461U,
		0xf8c9U,0xe199U,0xd339U,0xb679U,0x7cf9U,0xf9f2U,0xe3efU,0xd7d5U,0xbfa1U,0x6f49U,
		0xde92U,0xad2fU,0x4a55U,0x94aaU,0x395fU,0x72beU,0xe57cU,0xdaf3U,0xa5edU,0x5bd1U,
		0xb7a2U,0x7f4fU,0xfe9eU,0xed37U,0xca65U,0x84c1U,0x1989U,0x3312U,0x6624U,0xcc48U,
		0x889bU,0x013dU,0x027aU,0x04f4U,0x09e8U,0x13d0U,0x27a0U,0x4f40U,0x9e80U,0x2d0bU,
		0x5a16U,0xb42cU,0x7853U,0xf0a6U,0xf147U,0xf285U,0xf501U,0xfa09U,0xe419U,0xd839U,
		0xa079U,0x50f9U,0xa1f2U,0x53efU,0xa7deU,0x5fb7U,0xbf6eU,0x6ed7U,0xddaeU,0xab57U,
		0x46a5U,0x8d4aU,0x0a9fU,0x153eU,0x2a7cU,0x54f8U,0xa9f0U,0x43ebU,0x87d6U,0x1fa7U,
		0x3f4eU,0x7e9cU,0xfd38U,0xea7bU,0xc4fdU,0x99f1U,0x23e9U,0x47d2U,0x8fa4U,0x0f43U,
		0x1e86U,0x3d0cU,0x7a18U,0xf430U,0xf86bU,0xe0ddU,0xd1b1U,0xb369U,0x76d9U,0xedb2U,
		0xcb6fU,0x86d5U,0x1da1U,0x3b42U,0x7684U,0xed08U,0xca1bU,0x843dU,0x1871U,0x30e2U,
		0x61c4U,0xc388U,0x971bU,0x3e3dU,0x7c7aU,0xf8f4U,0xe1e3U,0xd3cdU,0xb791U,0x7f29U,
		0xfe52U,0xecafU,0xc955U,0x82a1U,0x1549U,0x2a92U,0x5524U,0xaa48U,0x449bU,0x8936U,
		0x0267U,0x04ceU,0x099cU,0x1338U,0x2670U,0x4ce0U,0x99c0U,0x238bU,0x4716U,0x8e2cU,
		0x0c53U,0x18a6U,0x314cU,0x6298U,0xc530U,0x9a6bU,0x24ddU,0x49baU,0x9374U,0x36e3U,
		0x6dc6U,0xdb8cU,0xa713U,0x5e2dU,0xbc5aU,0x68bfU,0xd17eU,0xb2f7U,0x75e5U,0xebcaU,
		0xc79fU,0x9f35U,0x2e61U,0x5cc2U,0xb984U,0x6303U,0xc606U,0x9c07U,0x2805U,0x500aU,
		0xa014U,0x5023U,0xa046U,0x5087U,0xa10eU,0x5217U,0xa42eU,0x5857U,0xb0aeU,0x7157U,
		0xe2aeU,0xd557U,0xbaa5U,0x6541U,0xca82U,0x850fU,0x1a15U,0x342aU,0x6854U,0xd0a8U,
		0xb15bU,0x72bdU,0xe57aU,0xdaffU,0xa5f5U,0x5be1U,0xb7c2U,0x7f8fU,0xff1eU,0xee37U,
		0xcc65U,0x88c1U,0x0189U,0x0312U,0x0624U,0x0c48U,0x1890U,0x3120U,0x6240U,0xc480U,
		0x990bU,0x221dU,0x443aU,0x8874U,0x00e3U,0x01c6U,0x038cU,0x0718U,0x0e30U,0x1c60U,
		0x38c0U,0x7180U,0xe300U,0xd60bU,0xbc1dU,0x6831U,0xd062U,0xb0cfU,0x7195U,0xe32aU,
		0xd65fU,0xbcb5U,0x6961U,0xd2c2U,0xb58fU,0x7b15U,0xf62aU,0xfc5fU,0xe8b5U,0xc161U,
		0x92c9U,0x3599U,0x6b32U,0xd664U,0xbcc3U,0x698dU,0xd31aU,0xb63fU,0x7c75U,0xf8eaU,
		0xe1dfU,0xd3b5U,0xb761U,0x7ec9U,0xfd92U,0xeb2fU,0xc655U,0x9ca1U,0x2949U,0x5292U,
		0xa524U,0x5a43U,0xb486U,0x7907U,0xf20eU,0xf417U,0xf825U,0xe041U,0xd089U,0xb119U,
		0x7239U,0xe472U,0xd8efU,0xa1d5U,0x53a1U,0xa742U,0x5e8fU,0xbd1eU,0x6a37U,0xd46eU,
		0xb8d7U,0x61a5U,0xc34aU,0x969fU,0x3d35U,0x7a6aU,0xf4d4U,0xf9a3U,0xe34dU,0xd691U,
		0xbd29U,0x6a59U,0xd4b2U,0xb96fU,0x62d5U,0xc5aaU,0x9b5fU,0x26b5U,0x4d6aU,0x9ad4U,
		0x25a3U,0x4b46U,0x968cU,0x3d13U,0x7a26U,0xf44cU,0xf893U,0xe12dU,0xd251U,0xb4a9U,
		0x7959U,0xf2b2U,0xf56fU,0xfad5U,0xe5a1U,0xdb49U,0xa699U,0x5d39U,0xba72U,0x64efU,
		0xc9deU,0x83b7U,0x1765U,0x2ecaU,0x5d94U,0xbb28U,0x665bU,0xccb6U,0x8967U,0x02c5U,
		0x058aU,0x0b14U,0x1628U,0x2c50U,0x58a0U,0xb140U,0x728bU,0xe516U,0xda27U,0xa445U,
		0x5881U,0xb102U,0x720fU,0xe41eU,0xd837U,0xa065U,0x50c1U,0xa182U,0x530fU,0xa61eU,
		0x5c37U,0xb86eU,0x60d7U,0xc1aeU,0x9357U,0x36a5U,0x6d4aU,0xda94U,0xa523U,0x5a4dU,
		0xb49aU,0x793fU,0xf27eU,0xf4f7U,0xf9e5U,0xe3c1U,0xd789U,0xbf19U,0x6e39U,0xdc72U,
		0xa8efU,0x41d5U,0x83aaU,0x175fU,0x2ebeU,0x5d7cU,0xbaf8U,0x65fbU,0xcbf6U,0x87e7U,
		0x1fc5U,0x3f8aU,0x7f14U,0xfe28U,0xec5bU,0xc8bdU,0x8171U,0x12e9U,0x25d2U,0x4ba4U,
		0x9748U,0x3e9bU,0x7d36U,0xfa6cU,0xe4d3U,0xd9adU,0xa351U,0x56a9U,0xad52U,0x4aafU,
		0x955eU,0x3ab7U,0x756eU,0xeadcU,0xc5b3U,0x9b6dU,0x26d1U,0x4da2U,0x9b44U,0x2683U,
		0x4d06U,0x9a0cU,0x2413U,0x4826U,0x904cU,0x3093U,0x6126U,0xc24cU,0x9493U,0x392dU,
		0x725aU,0xe4b4U,0xd963U,0xa2cdU,0x5591U,0xab22U,0x464fU,0x8c9eU,0x0937U,0x126eU,
		0x24dcU,0x49b8U,0x9370U,0x36ebU,0x6dd6U,0xdbacU,0xa753U,0x5eadU,0xbd5aU,0x6abfU,
		0xd57eU,0xbaf7U,0x65e5U,0xcbcaU,0x879fU,0x1f35U,0x3e6aU,0x7cd4U,0xf9a8U,0xe35bU,
		0xd6bdU,0xbd71U,0x6ae9U,0xd5d2U,0xbbafU,0x6755U,0xceaaU,0x8d5fU,0x0ab5U,0x156aU,
		0x2ad4U,0x55a8U,0xab50U,0x46abU,0x8d56U,0x0aa7U,0x154eU,0x2a9cU,0x5538U,0xaa70U,
		0x44ebU,0x89d6U,0x03a7U,0x074eU,0x0e9cU,0x1d38U,0x3a70U,0x74e0U,0xe9c0U,0xc38bU,
		0x971dU,0x3e31U,0x7c62U,0xf8c4U,0xe183U,0xd30dU,0xb611U,0x7c29U,0xf852U,0xe0afU,
		0xd155U,0xb2a1U,0x7549U,0xea92U,0xc52fU,0x9a55U,0x24a1U,0x4942U,0x9284U,0x3503U,
		0x6a06U,0xd40cU,0xb813U,0x602dU,0xc05aU,0x90bfU,0x3175U,0x62eaU,0xc5d4U,0x9ba3U,
		0x274dU,0x4e9aU,0x9d34U,0x2a63U,0x54c6U,0xa98cU,0x4313U,0x8626U,0x1c47U,0x388eU,
		0x711cU,0xe238U,0xd47bU,0xb8fdU,0x61f1U,0xc3e2U,0x97cfU,0x3f95U,0x7f2aU,0xfe54U,
		0xeca3U,0xc94dU,0x8291U,0x1529U,0x2a52U,0x54a4U,0xa948U,0x429bU,0x8536U,0x1a67U,
		0x34ceU,0x699cU,0xd338U,0xb67bU,0x7cfdU,0xf9faU,0xe3ffU,0xd7f5U,0xbfe1U,0x6fc9U,
		0xdf92U,0xaf2fU,0x4e55U,0x9caaU,0x295fU,0x52beU,0xa57cU,0x5af3U,0xb5e6U,0x7bc7U,
		0xf78eU,0xff17U,0xee25U,0xcc41U,0x8889U,0x0119U,0x0232U,0x0464U,0x08c8U,0x1190U,
		0x2320U,0x4640U,0x8c80U,0x090bU,0x1216U,0x242cU,0x4858U,0x90b0U,0x316bU,0x62d6U,
		0xc5acU,0x9b53U,0x26adU,0x4d5aU,0x9ab4U,0x2563U,0x4ac6U,0x958cU,0x3b13U,0x7626U,
		0xec4cU,0xc893U,0x812dU,0x1251U,0x24a2U,0x4944U,0x9288U,0x351bU,0x6a36U,0xd46cU,
		0xb8d3U,0x61adU,0xc35aU,0x96bfU,0x3d75U,0x7aeaU,0xf5d4U,0xfba3U,0xe74dU,0xde91U,
		0xad29U,0x4a59U,0x94b2U,0x396fU,0x72deU,0xe5bcU,0xdb73U,0xa6edU,0x5dd1U,0xbba2U,
		0x674fU,0xce9eU,0x8d37U,0x0a65U,0x14caU,0x2994U,0x5328U,0xa650U,0x5cabU,0xb956U,
		0x62a7U,0xc54eU,0x9a97U,0x2525U,0x4a4aU,0x9494U,0x3923U,0x7246U,0xe48cU,0xd913U,
		0xa22dU,0x5451U,0xa8a2U,0x414fU,0x829eU,0x1537U,0x2a6eU,0x54dcU,0xa9b8U,0x437bU,
		0x86f6U,0x1de7U,0x3bceU,0x779cU,0xef38U,0xce7bU,0x8cfdU,0x09f1U,0x13e2U,0x27c4U,
		0x4f88U,0x9f10U,0x2e2bU,0x5c56U,0xb8acU,0x6153U,0xc2a6U,0x9547U,0x3a85U,0x750aU,
		0xea14U,0xc423U,0x984dU,0x2091U,0x4122U,0x8244U,0x1483U,0x2906U,0x520cU,0xa418U,
		0x583bU,0xb076U,0x70e7U,0xe1ceU,0xd397U,0xb725U,0x7e41U,0xfc82U,0xe90fU,0xc215U,
		0x9421U,0x3849U,0x7092U,0xe124U,0xd243U,0xb48dU,0x7911U,0xf222U,0xf44fU,0xf895U,
		0xe121U,0xd249U,0xb499U,0x7939U,0xf272U,0xf4efU,0xf9d5U,0xe3a1U,0xd749U,0xbe99U,
		0x6d39U,0xda72U,0xa4efU,0x59d5U,0xb3aaU,0x775fU,0xeebeU,0xcd77U,0x8ae5U,0x05c1U,
		0x0b82U,0x1704U,0x2e08U,0x5c10U,0xb820U,0x604bU,0xc096U,0x9127U,0x3245U,0x648aU,
		0xc914U,0x8223U,0x144dU,0x289aU,0x5134U,0xa268U,0x54dbU,0xa9b6U,0x4367U,0x86ceU,
		0x1d97U,0x3b2eU,0x765cU,0xecb8U,0xc97bU,0x82fdU,0x15f1U,0x2be2U,0x57c4U,0xaf88U,
		0x4f1bU,0x9e36U,0x2c67U,0x58ceU,0xb19cU,0x7333U,0xe666U,0xdcc7U,0xa985U,0x4301U,
		0x8602U,0x1c0fU,0x381eU,0x703cU,0xe078U,0xd0fbU,0xb1fdU,0x73f1U,0xe7e2U,0xdfcfU,
		0xaf95U,0x4f21U,0x9e42U,0x2c8fU,0x591eU,0xb23cU,0x7473U,0xe8e6U,0xc1c7U,0x9385U,
		0x3701U,0x6e02U,0xdc04U,0xa803U,0x400dU,0x801aU,0x103fU,0x207eU,0x40fcU,0x81f8U,
		0x13fbU,0x27f6U,0x4fecU,0x9fd8U,0x2fbbU,0x5f76U,0xbeecU,0x6dd3U,0xdba6U,0xa747U,
		0x5e85U,0xbd0aU,0x6a1fU,0xd43eU,0xb877U,0x60e5U,0xc1caU,0x939fU,0x3735U,0x6e6aU,
		0xdcd4U,0xa9a3U,0x434dU,0x869aU,0x1d3fU,0x3a7eU,0x74fcU,0xe9f8U,0xc3fbU,0x97fdU,
		0x3ff1U,0x7fe2U,0xffc4U,0xef83U,0xcf0dU,0x8e11U,0x0c29U,0x1852U,0x30a4U,0x6148U,
		0xc290U,0x952bU,0x3a5dU,0x74baU,0xe974U,0xc2e3U,0x95cdU,0x3b91U,0x7722U,0xee44U,
		0xcc83U,0x890dU,0x0211U,0x0422U,0x0844U,0x1088U,0x2110U,0x4220U,0x8440U,0x188bU,
		0x3116U,0x622cU,0xc458U,0x98bbU,0x217dU,0x42faU,0x85f4U,0x1be3U,0x37c6U,0x6f8cU,
		0xdf18U,0xae3bU,0x4c7dU,0x98faU,0x21ffU,0x43feU,0x87fcU,0x1ff3U,0x3fe6U,0x7fccU,
		0xff98U,0xef3bU,0xce7dU,0x8cf1U,0x09e9U,0x13d2U,0x27a4U,0x4f48U,0x9e90U,0x2d2bU,
		0x5a56U,0xb4acU,0x7953U,0xf2a6U,0xf547U,0xfa85U,0xe501U,0xda09U,0xa419U,0x5839U,
		0xb072U,0x70efU,0xe1deU,0xd3b7U,0xb765U,0x7ec1U,0xfd82U,0xeb0fU,0xc615U,0x9c21U,
		0x2849U,0x5092U,0xa124U,0x5243U,0xa486U,0x5907U,0xb20eU,0x7417U,0xe82eU,0xc057U,
		0x90a5U,0x3141U,0x6282U,0xc504U,0x9a03U,0x240dU,0x481aU,0x9034U,0x3063U,0x60c6U,
		0xc18cU,0x9313U,0x362dU,0x6c5aU,0xd8b4U,0xa163U,0x52cdU,0xa59aU,0x5b3fU,0xb67eU,
		0x7cf7U,0xf9eeU,0xe3d7U,0xd7a5U,0xbf41U,0x6e89U,0xdd12U,0xaa2fU,0x4455U,0x88aaU,
		0x015fU,0x02beU,0x057cU,0x0af8U,0x15f0U,0x2be0U,0x57c0U,0xaf80U,0x4f0bU,0x9e16U,
		0x2c27U,0x584eU,0xb09cU,0x7133U,0xe266U,0xd4c7U,0xb985U,0x6301U,0xc602U,0x9c0fU,
		0x2815U,0x502aU,0xa054U,0x50a3U,0xa146U,0x5287U,0xa50eU,0x5a17U,0xb42eU,0x7857U,
		0xf0aeU,0xf157U,0xf2a5U,0xf541U,0xfa89U,0xe519U,0xda39U,0xa479U,0x58f9U,0xb1f2U,
		0x73efU,0xe7deU,0xdfb7U,0xaf65U,0x4ec1U,0x9d82U,0x2b0fU,0x561eU,0xac3cU,0x4873U,
		0x90e6U,0x31c7U,0x638eU,0xc71cU,0x9e33U,0x2c6dU,0x58daU,0xb1b4U,0x7363U,0xe6c6U,
		0xdd87U,0xab05U,0x4601U,0x8c02U,0x080fU,0x101eU,0x203cU,0x4078U,0x80f0U,0x11ebU,
		0x23d6U,0x47acU,0x8f58U,0x0ebbU,0x1d76U,0x3aecU,0x75d8U,0xebb0U,0xc76bU,0x9eddU,
		0x2db1U,0x5b62U,0xb6c4U,0x7d83U,0xfb06U,0xe607U,0xdc05U,0xa801U,0x4009U,0x8012U,
		0x102fU,0x205eU,0x40bcU,0x8178U,0x12fbU,0x25f6U,0x4becU,0x97d8U,0x3fbbU,0x7f76U,
		0xfeecU,0xedd3U,0xcbadU,0x8751U,0x1ea9U,0x3d52U,0x7aa4U,0xf548U,0xfa9bU,0xe53dU,
		0xda71U,0xa4e9U,0x59d9U,0xb3b2U,0x776fU,0xeedeU,0xcdb7U,0x8b65U,0x06c1U,0x0d82U,
		0x1b04U,0x3608U,0x6c10U,0xd820U,0xa04bU,0x509dU,0xa13aU,0x527fU,0xa4feU,0x59f7U,
		0xb3eeU,0x77d7U,0xefaeU,0xcf57U,0x8ea5U,0x0d41U,0x1a82U,0x3504U,0x6a08U,0xd410U,
		0xb82bU,0x605dU,0xc0baU,0x917fU,0x32f5U,0x65eaU,0xcbd4U,0x87a3U,0x1f4dU,0x3e9aU,
		0x7d34U,0xfa68U,0xe4dbU,0xd9bdU,0xa371U,0x56e9U,0xadd2U,0x4bafU,0x975eU,0x3eb7U,
		0x7d6eU,0xfadcU,0xe5b3U,0xdb6dU,0xa6d1U,0x5da9U,0xbb52U,0x66afU,0xcd5eU,0x8ab7U,
		0x0565U,0x0acaU,0x1594U,0x2b28U,0x5650U,0xaca0U,0x494bU,0x9296U,0x3527U,0x6a4eU,
		0xd49cU,0xb933U,0x626dU,0xc4daU,0x99bfU,0x2375U,0x46eaU,0x8dd4U,0x0ba3U,0x1746U,
		0x2e8cU,0x5d18U,0xba30U,0x646bU,0xc8d6U,0x81a7U,0x1345U,0x268aU,0x4d14U,0x9a28U,
		0x245bU,0x48b6U,0x916cU,0x32d3U,0x65a6U,0xcb4cU,0x8693U,0x1d2dU,0x3a5aU,0x74b4U,
		0xe968U,0xc2dbU,0x95bdU,0x3b71U,0x76e2U,0xedc4U,0xcb83U,0x870dU,0x1e11U,0x3c22U,
		0x7844U,0xf088U,0xf11bU,0xf23dU,0xf471U,0xf8e9U,0xe1d9U,0xd3b9U,0xb779U,0x7ef9U,
		0xfdf2U,0xebefU,0xc7d5U,0x9fa1U,0x2f49U,0x5e92U,0xbd24U,0x6a43U,0xd486U,0xb907U,
		0x6205U,0xc40aU,0x981fU,0x2035U,0x406aU,0x80d4U,0x11a3U,0x2346U,0x468cU,0x8d18U,
		0x0a3bU,0x1476U,0x28ecU,0x51d8U,0xa3b0U,0x576bU,0xaed6U,0x4da7U,0x9b4eU,0x2697U,
		0x4d2eU,0x9a5cU,0x24b3U,0x4966U,0x92ccU,0x3593U,0x6b26U,0xd64cU,0xbc93U,0x692dU,
		0xd25aU,0xb4bfU,0x7975U,0xf2eaU,0xf5dfU,0xfbb5U,0xe761U,0xdec9U,0xad99U,0x4b39U,
		0x9672U,0x3cefU,0x79deU,0xf3bcU,0xf773U,0xfeedU,0xedd1U,0xcba9U,0x8759U,0x1eb9U,
		0x3d72U,0x7ae4U,0xf5c8U,0xfb9bU,0xe73dU,0xde71U,0xace9U,0x49d9U,0x93b2U,0x376fU,
		0x6edeU,0xddbcU,0xab73U,0x46edU,0x8ddaU,0x0bbfU,0x177eU,0x2efcU,0x5df8U,0xbbf0U,
		0x67ebU,0xcfd6U,0x8fa7U,0x0f45U,0x1e8aU,0x3d14U,0x7a28U,0xf450U,0xf8abU,0xe15dU,
		0xd2b1U,0xb569U,0x7ad9U,0xf5b2U,0xfb6fU,0xe6d5U,0xdda1U,0xab49U,0x4699U,0x8d32U,
		0x0a6fU,0x14deU,0x29bcU,0x5378U,0xa6f0U,0x5debU,0xbbd6U,0x67a7U,0xcf4eU,0x8e97U,
		0x0d25U,0x1a4aU,0x3494U,0x6928U,0xd250U,0xb4abU,0x795dU,0xf2baU,0xf57fU,0xfaf5U,
		0xe5e1U,0xdbc9U,0xa799U,0x5f39U,0xbe72U,0x6cefU,0xd9deU,0xa3b7U,0x5765U,0xaecaU,
		0x4d9fU,0x9b3eU,0x2677U,0x4ceeU,0x99dcU,0x23b3U,0x4766U,0x8eccU,0x0d93U,0x1b26U,
		0x364cU,0x6c98U,0xd930U,0xa26bU,0x54ddU,0xa9baU,0x437fU,0x86feU,0x1df7U,0x3beeU,
		0x77dcU,0xefb8U,0xcf7bU,0x8efdU,0x0df1U,0x1be2U,0x37c4U,0x6f88U,0xdf10U,0xae2bU,
		0x4c5dU,0x98baU,0x217fU,0x42feU,0x85fcU,0x1bf3U,0x37e6U,0x6fccU,0xdf98U,0xaf3bU,
		0x4e7dU,0x9cfaU,0x29ffU,0x53feU,0xa7fcU,0x5ff3U,0xbfe6U,0x6fc7U,0xdf8eU,0xaf17U,
		0x4e25U,0x9c4aU,0x289fU,0x513eU,0xa27cU,0x54f3U,0xa9e6U,0x43c7U,0x878eU,0x1f17U,
		0x3e2eU,0x7c5cU,0xf8b8U,0xe17bU,0xd2fdU,0xb5f1U,0x7be9U,0xf7d2U,0xffafU,0xef55U,
		0xcea1U,0x8d49U,0x0a99U,0x1532U,0x2a64U,0x54c8U,0xa990U,0x432bU,0x8656U,0x1ca7U,
		0x394eU,0x729cU,0xe538U,0xda7bU,0xa4fdU,0x59f1U,0xb3e2U,0x77cfU,0xef9eU,0xcf37U,
		0x8e65U,0x0cc1U,0x1982U,0x3304U,0x6608U,0xcc10U,0x882bU,0x005dU,0x00baU,0x0174U,
		0x02e8U,0x05d0U,0x0ba0U,0x1740U,0x2e80U,0x5d00U,0xba00U,0x640bU,0xc816U,0x8027U,
		0x1045U,0x208aU,0x4114U,0x8228U,0x145bU,0x28b6U,0x516cU,0xa2d8U,0x55bbU,0xab76U,
		0x46e7U,0x8dceU,0x0b97U,0x172eU,0x2e5cU,0x5cb8U,0xb970U,0x62ebU,0xc5d6U,0x9ba7U,
		0x2745U,0x4e8aU,0x9d14U,0x2a23U,0x5446U,0xa88cU,0x4113U,0x8226U,0x1447U,0x288eU,
		0x511cU,0xa238U,0x547bU,0xa8f6U,0x41e7U,0x83ceU,0x1797U,0x2f2eU,0x5e5cU,0xbcb8U,
		0x697bU,0xd2f6U,0xb5e7U,0x7bc5U,0xf78aU,0xff1fU,0xee35U,0xcc61U,0x88c9U,0x0199U,
		0x0332U,0x0664U,0x0cc8U,0x1990U,0x3320U,0x6640U,0xcc80U,0x890bU,0x021dU,0x043aU,
		0x0874U,0x10e8U,0x21d0U,0x43a0U,0x8740U,0x1e8bU,0x3d16U,0x7a2cU,0xf458U,0xf8bbU,
		0xe17dU,0xd2f1U,0xb5e9U,0x7bd9U,0xf7b2U,0xff6fU,0xeed5U,0xcda1U,0x8b49U,0x0699U,
		0x0d32U,0x1a64U,0x34c8U,0x6990U,0xd320U,0xb64bU,0x7c9dU,0xf93aU,0xe27fU,0xd4f5U,
		0xb9e1U,0x63c9U,0xc792U,0x9f2fU,0x2e55U,0x5caaU,0xb954U,0x62a3U,0xc546U,0x9a87U,
		0x2505U,0x4a0aU,0x9414U,0x3823U,0x7046U,0xe08cU,0xd113U,0xb22dU,0x7451U,0xe8a2U,
		0xc14fU,0x9295U,0x3521U,0x6a42U,0xd484U,0xb903U,0x620dU,0xc41aU,0x983fU,0x2075U,
		0x40eaU,0x81d4U,0x13a3U,0x2746U,0x4e8cU,0x9d18U,0x2a3bU,0x5476U,0xa8ecU,0x41d3U,
		0x83a6U,0x1747U,0x2e8eU,0x5d1cU,0xba38U,0x647bU,0xc8f6U,0x81e7U,0x13c5U,0x278aU,
		0x4f14U,0x9e28U,0x2c5bU,0x58b6U,0xb16cU,0x72d3U,0xe5a6U,0xdb47U,0xa685U,0x5d01U,
		0xba02U,0x640fU,0xc81eU,0x8037U,0x1065U,0x20caU,0x4194U,0x8328U,0x165bU,0x2cb6U,
		0x596cU,0xb2d8U,0x75bbU,0xeb76U,0xc6e7U,0x9dc5U,0x2b81U,0x5702U,0xae04U,0x4c03U,
		0x9806U,0x2007U,0x400eU,0x801cU,0x1033U,0x2066U,0x40ccU,0x8198U,0x133bU,0x2676U,
		0x4cecU,0x99d8U,0x23bbU,0x4776U,0x8eecU,0x0dd3U,0x1ba6U,0x374cU,0x6e98U,0xdd30U,
		0xaa6bU,0x44ddU,0x89baU,0x037fU,0x06feU,0x0dfcU,0x1bf8U,0x37f0U,0x6fe0U,0xdfc0U,
		0xaf8bU,0x4f1dU,0x9e3aU,0x2c7fU,0x58feU,0xb1fcU,0x73f3U,0xe7e6U,0xdfc7U,0xaf85U,
		0x4f01U,0x9e02U,0x2c0fU,0x581eU,0xb03cU,0x7073U,0xe0e6U,0xd1c7U,0xb385U,0x7701U,
		0xee02U,0xcc0fU,0x8815U,0x0021U,0x0042U,0x0084U,0x0108U,0x0210U,0x0420U,0x0840U,
		0x1080U,0x2100U,0x4200U,0x8400U,0x180bU,0x3016U,0x602cU,0xc058U,0x90bbU,0x317dU,
		0x62faU,0xc5f4U,0x9be3U,0x27cdU,0x4f9aU,0x9f34U,0x2e63U,0x5cc6U,0xb98cU,0x6313U,
		0xc626U,0x9c47U,0x2885U,0x510aU,0xa214U,0x5423U,0xa846U,0x4087U,0x810eU,0x1217U,
		0x242eU,0x485cU,0x90b8U,0x317bU,0x62f6U,0xc5ecU,0x9bd3U,0x27adU,0x4f5aU,0x9eb4U,
		0x2d63U,0x5ac6U,0xb58cU,0x7b13U,0xf626U,0xfc47U,0xe885U,0xc101U,0x9209U,0x3419U,
		0x6832U,0xd064U,0xb0c3U,0x718dU,0xe31aU,0xd63fU,0xbc75U,0x68e1U,0xd1c2U,0xb38fU,
		0x7715U,0xee2aU,0xcc5fU,0x88b5U,0x0161U,0x02c2U,0x0584U,0x0b08U,0x1610U,0x2c20U,
		0x5840U,0xb080U,0x710bU,0xe216U,0xd427U,0xb845U,0x6081U,0xc102U,0x920fU,0x3415U,
		0x682aU,0xd054U,0xb0a3U,0x714dU,0xe29aU,0xd53fU,0xba75U,0x64e1U,0xc9c2U,0x838fU,
		0x1715U,0x2e2aU,0x5c54U,0xb8a8U,0x615bU,0xc2b6U,0x9567U,0x3ac5U,0x758aU,0xeb14U,
		0xc623U,0x9c4dU,0x2891U,0x5122U,0xa244U,0x5483U,0xa906U,0x4207U,0x840eU,0x1817U,
		0x302eU,0x605cU,0xc0b8U,0x917bU,0x32fdU,0x65faU,0xcbf4U,0x87e3U,0x1fcdU,0x3f9aU,
		0x7f34U,0xfe68U,0xecdbU,0xc9bdU,0x8371U,0x16e9U,0x2dd2U,0x5ba4U,0xb748U,0x7e9bU,
		0xfd36U,0xea67U,0xc4c5U,0x9981U,0x2309U,0x4612U,0x8c24U,0x0843U,0x1086U,0x210cU,
		0x4218U,0x8430U,0x186bU,0x30d6U,0x61acU,0xc358U,0x96bbU,0x3d7dU,0x7afaU,0xf5f4U,
		0xfbe3U,0xe7cdU,0xdf91U,0xaf29U,0x4e59U,0x9cb2U,0x296fU,0x52deU,0xa5bcU,0x5b73U,
		0xb6e6U,0x7dc7U,0xfb8eU,0xe717U,0xde25U,0xac41U,0x4889U,0x9112U,0x322fU,0x645eU,
		0xc8bcU,0x8173U,0x12edU,0x25daU,0x4bb4U,0x9768U,0x3edbU,0x7db6U,0xfb6cU,0xe6d3U,
		0xddadU,0xab51U,0x46a9U,0x8d52U,0x0aafU,0x155eU,0x2abcU,0x5578U,0xaaf0U,0x45ebU,
		0x8bd6U,0x07a7U,0x0f4eU,0x1e9cU,0x3d38U,0x7a70U,0xf4e0U,0xf9cbU,0xe39dU,0xd731U,
		0xbe69U,0x6cd9U,0xd9b2U,0xa36fU,0x56d5U,0xadaaU,0x4b5fU,0x96beU,0x3d77U,0x7aeeU,
		0xf5dcU,0xfbb3U,0xe76dU,0xded1U,0xada9U,0x4b59U,0x96b2U,0x3d6fU,0x7adeU,0xf5bcU,
		0xfb73U,0xe6edU,0xddd1U,0xaba9U,0x4759U,0x8eb2U,0x0d6fU,0x1adeU,0x35bcU,0x6b78U,
		0xd6f0U,0xbdebU,0x6bddU,0xd7baU,0xbf7fU,0x6ef5U,0xddeaU,0xabdfU,0x47b5U,0x8f6aU,
		0x0edfU,0x1dbeU,0x3b7cU,0x76f8U,0xedf0U,0xcbebU,0x87ddU,0x1fb1U,0x3f62U,0x7ec4U,
		0xfd88U,0xeb1bU,0xc63dU,0x9c71U,0x28e9U,0x51d2U,0xa3a4U,0x5743U,0xae86U,0x4d07U,
		0x9a0eU,0x2417U,0x482eU,0x905cU,0x30b3U,0x6166U,0xc2ccU,0x9593U,0x3b2dU,0x765aU,
		0xecb4U,0xc963U,0x82cdU,0x1591U,0x2b22U,0x5644U,0xac88U,0x491bU,0x9236U,0x3467U,
		0x68ceU,0xd19cU,0xb333U,0x766dU,0xecdaU,0xc9bfU,0x8375U,0x16e1U,0x2dc2U,0x5b84U,
		0xb708U,0x7e1bU,0xfc36U,0xe867U,0xc0c5U,0x9181U,0x3309U,0x6612U,0xcc24U,0x8843U,
		0x008dU,0x011aU,0x0234U,0x0468U,0x08d0U,0x11a0U,0x2340U,0x4680U,0x8d00U,0x0a0bU,
		0x1416U,0x282cU,0x5058U,0xa0b0U,0x516bU,0xa2d6U,0x55a7U,0xab4eU,0x4697U,0x8d2eU,
		0x0a57U,0x14aeU,0x295cU,0x52b8U,0xa570U,0x5aebU,0xb5d6U,0x7ba7U,0xf74eU,0xfe97U,
		0xed25U,0xca41U,0x8489U,0x1919U,0x3232U,0x6464U,0xc8c8U,0x819bU,0x133dU,0x267aU,
		0x4cf4U,0x99e8U,0x23dbU,0x47b6U,0x8f6cU,0x0ed3U,0x1da6U,0x3b4cU,0x7698U,0xed30U,
		0xca6bU,0x84ddU,0x19b1U,0x3362U,0x66c4U,0xcd88U,0x8b1bU,0x063dU,0x0c7aU,0x18f4U,
		0x31e8U,0x63d0U,0xc7a0U,0x9f4bU,0x2e9dU,0x5d3aU,0xba74U,0x64e3U,0xc9c6U,0x8387U,
		0x1705U,0x2e0aU,0x5c14U,0xb828U,0x605bU,0xc0b6U,0x9167U,0x32c5U,0x658aU,0xcb14U,
		0x8623U,0x1c4dU,0x389aU,0x7134U,0xe268U,0xd4dbU,0xb9bdU,0x6371U,0xc6e2U,0x9dcfU,
		0x2b95U,0x572aU,0xae54U,0x4ca3U,0x9946U,0x2287U,0x450eU,0x8a1cU,0x0433U,0x0866U,
		0x10ccU,0x2198U,0x4330U,0x8660U,0x1ccbU,0x3996U,0x732cU,0xe658U,0xdcbbU,0xa97dU,
		0x42f1U,0x85e2U,0x1bcfU,0x379eU,0x6f3cU,0xde78U,0xacfbU,0x49fdU,0x93faU,0x37ffU,
		0x6ffeU,0xdffcU,0xaff3U,0x4fedU,0x9fdaU,0x2fbfU,0x5f7eU,0xbefcU,0x6df3U,0xdbe6U,
		0xa7c7U,0x5f85U,0xbf0aU,0x6e1fU,0xdc3eU,0xa877U,0x40e5U,0x81caU,0x139fU,0x273eU,
		0x4e7cU,0x9cf8U,0x29fbU,0x53f6U,0xa7ecU,0x5fd3U,0xbfa6U,0x6f47U,0xde8eU,0xad17U,
		0x4a25U,0x944aU,0x389fU,0x713eU,0xe27cU,0xd4f3U,0xb9edU,0x63d1U,0xc7a2U,0x9f4fU,
		0x2e95U,0x5d2aU,0xba54U,0x64a3U,0xc946U,0x8287U,0x1505U,0x2a0aU,0x5414U,0xa828U,
		0x405bU,0x80b6U,0x1167U,0x22ceU,0x459cU,0x8b38U,0x067bU,0x0cf6U,0x19ecU,0x33d8U,
		0x67b0U,0xcf60U,0x8ecbU,0x0d9dU,0x1b3aU,0x3674U,0x6ce8U,0xd9d0U,0xa3abU,0x575dU,
		0xaebaU,0x4d7fU,0x9afeU,0x25f7U,0x4beeU,0x97dcU,0x3fb3U,0x7f66U,0xfeccU,0xed93U,
		0xcb2dU,0x8651U,0x1ca9U,0x3952U,0x72a4U,0xe548U,0xda9bU,0xa53dU,0x5a71U,0xb4e2U,
		0x79cfU,0xf39eU,0xf737U,0xfe65U,0xecc1U,0xc989U,0x8319U,0x1639U,0x2c72U,0x58e4U,
		0xb1c8U,0x739bU,0xe736U,0xde67U,0xacc5U,0x4981U,0x9302U,0x360fU,0x6c1eU,0xd83cU,
		0xa073U,0x50edU,0xa1daU,0x53bfU,0xa77eU,0x5ef7U,0xbdeeU,0x6bd7U,0xd7aeU,0xbf57U,
		0x6ea5U,0xdd4aU,0xaa9fU,0x4535U,0x8a6aU,0x04dfU,0x09beU,0x137cU,0x26f8U,0x4df0U,
		0x9be0U,0x27cbU,0x4f96U,0x9f2cU,0x2e53U,0x5ca6U,0xb94cU,0x6293U,0xc526U,0x9a47U,
		0x2485U,0x490aU,0x9214U,0x3423U,0x6846U,0xd08cU,0xb113U,0x722dU,0xe45aU,0xd8bfU,
		0xa175U,0x52e1U,0xa5c2U,0x5b8fU,0xb71eU,0x7e37U,0xfc6eU,0xe8d7U,0xc1a5U,0x9341U,
		0x3689U,0x6d12U,0xda24U,0xa443U,0x588dU,0xb11aU,0x723fU,0xe47eU,0xd8f7U,0xa1e5U,
		0x53c1U,0xa782U,0x5f0fU,0xbe1eU,0x6c37U,0xd86eU,0xa0d7U,0x51a5U,0xa34aU,0x569fU,
		0xad3eU,0x4a77U,0x94eeU,0x39d7U,0x73aeU,0xe75cU,0xdeb3U,0xad6dU,0x4ad1U,0x95a2U,
		0x3b4fU,0x769eU,0xed3cU,0xca73U,0x84edU,0x19d1U,0x33a2U,0x6744U,0xce88U,0x8d1bU,
		0x0a3dU,0x147aU,0x28f4U,0x51e8U,0xa3d0U,0x57abU,0xaf56U,0x4ea7U,0x9d4eU,0x2a97U,
		0x552eU,0xaa5cU,0x44b3U,0x8966U,0x02c7U,0x058eU,0x0b1cU,0x1638U,0x2c70U,0x58e0U,
		0xb1c0U,0x738bU,0xe716U,0xde27U,0xac45U,0x4881U,0x9102U,0x320fU,0x641eU,0xc83cU,
		0x8073U,0x10edU,0x21daU,0x43b4U,0x8768U,0x1edbU,0x3db6U,0x7b6cU,0xf6d8U,0xfdbbU,
		0xeb7dU,0xc6f1U,0x9de9U,0x2bd9U,0x57b2U,0xaf64U,0x4ec3U,0x9d86U,0x2b07U,0x560eU,
		0xac1cU,0x4833U,0x9066U,0x30c7U,0x618eU,0xc31cU,0x9633U,0x3c6dU,0x78daU,0xf1b4U,
		0xf363U,0xf6cdU,0xfd91U,0xeb29U,0xc659U,0x9cb9U,0x2979U,0x52f2U,0xa5e4U,0x5bc3U,
		0xb786U,0x7f07U,0xfe0eU,0xec17U,0xc825U,0x8041U,0x1089U,0x2112U,0x4224U,0x8448U,
		0x189bU,0x3136U,0x626cU,0xc4d8U,0x99bbU,0x237dU,0x46faU,0x8df4U,0x0be3U,0x17c6U,
		0x2f8cU,0x5f18U,0xbe30U,0x6c6bU,0xd8d6U,0xa1a7U,0x5345U,0xa68aU,0x5d1fU,0xba3eU,
		0x6477U,0xc8eeU,0x81d7U,0x13a5U,0x274aU,0x4e94U,0x9d28U,0x2a5bU,0x54b6U,0xa96cU,
		0x42d3U,0x85a6U,0x1b47U,0x368eU,0x6d1cU,0xda38U,0xa47bU,0x58fdU,0xb1faU,0x73ffU,
		0xe7feU,0xdff7U,0xafe5U,0x4fc1U,0x9f82U,0x2f0fU,0x5e1eU,0xbc3cU,0x6873U,0xd0e6U,
		0xb1c7U,0x7385U,0xe70aU,0xde1fU,0xac35U,0x4861U,0x90c2U,0x318fU,0x631eU,0xc63cU,
		0x9c73U,0x28edU,0x51daU,0xa3b4U,0x5763U,0xaec6U,0x4d87U,0x9b0eU,0x2617U,0x4c2eU,
		0x985cU,0x20b3U,0x4166U,0x82ccU,0x1593U,0x2b26U,0x564cU,0xac98U,0x493bU,0x9276U,
		0x34e7U,0x69ceU,0xd39cU,0xb733U,0x7e6dU,0xfcdaU,0xe9bfU,0xc375U,0x96e1U,0x3dc9U,
		0x7b92U,0xf724U,0xfe43U,0xec8dU,0xc911U,0x8229U,0x1459U,0x28b2U,0x5164U,0xa2c8U,
		0x559bU,0xab36U,0x4667U,0x8cceU,0x0997U,0x132eU,0x265cU,0x4cb8U,0x9970U,0x22ebU,
		0x45d6U,0x8bacU,0x0753U,0x0ea6U,0x1d4cU,0x3a98U,0x7530U,0xea60U,0xc4cbU,0x999dU,
		0x2331U,0x4662U,0x8cc4U,0x0983U,0x1306U,0x260cU,0x4c18U,0x9830U,0x206bU,0x40d6U,
		0x81acU,0x1353U,0x26a6U,0x4d4cU,0x9a98U,0x253bU,0x4a76U,0x94ecU,0x39d3U,0x73a6U,
		0xe74cU,0xde93U,0xad2dU,0x4a51U,0x94a2U,0x394fU,0x729eU,0xe53cU,0xda73U,0xa4edU,
		0x59d1U,0xb3a2U,0x774fU,0xee9eU,0xcd37U,0x8a65U,0x04c1U,0x0982U,0x1304U,0x2608U,
		0x4c10U,0x9820U,0x204bU,0x4096U,0x812cU,0x1253U,0x24a6U,0x494cU,0x9298U,0x353bU,
		0x6a76U,0xd4ecU,0xb9d3U,0x63adU,0xc75aU,0x9ebfU,0x2d75U,0x5aeaU,0xb5d4U,0x7ba3U,
		0xf746U,0xfe87U,0xed05U,0xca01U,0x8409U,0x1819U,0x3032U,0x6064U,0xc0c8U,0x919bU,
		0x333dU,0x667aU,0xccf4U,0x89e3U,0x03cdU,0x079aU,0x0f34U,0x1e68U,0x3cd0U,0x79a0U,
		0xf340U,0xf68bU,0xfd1dU,0xea31U,0xc469U,0x98d9U,0x21b9U,0x4372U,0x86e4U,0x1dc3U,
		0x3b86U,0x770cU,0xee18U,0xcc3bU,0x887dU,0x00f1U,0x01e2U,0x03c4U,0x0788U,0x0f10U,
		0x1e20U,0x3c40U,0x7880U,0xf100U,0xf20bU,0xf41dU,0xf831U,0xe069U,0xd0d9U,0xb1b9U,
		0x7379U,0xe6f2U,0xddefU,0xabd5U,0x47a1U,0x8f42U,0x0e8fU,0x1d1eU,0x3a3cU,0x7478U,
		0xe8f0U,0xc1ebU,0x93ddU,0x37b1U,0x6f62U,0xdec4U,0xad83U,0x4b0dU,0x961aU,0x3c3fU,
		0x787eU,0xf0fcU,0xf1f3U,0xf3edU,0xf7d1U,0xffa9U,0xef59U,0xceb9U,0x8d79U,0x0af9U,
		0x15f2U,0x2be4U,0x57c8U,0xaf90U,0x4f2bU,0x9e56U,0x2ca7U,0x594eU,0xb29cU,0x7533U,
		0xea66U,0xc4c7U,0x9985U,0x2301U,0x4602U,0x8c04U,0x0803U,0x1006U,0x200cU,0x4018U,
		0x8030U,0x106bU,0x20d6U,0x41acU,0x8358U,0x16bbU,0x2d76U,0x5aecU,0xb5d8U,0x7bbbU,
		0xf776U,0xfee7U,0xedc5U,0xcb81U,0x8709U,0x1e19U,0x3c32U,0x7864U,0xf0c8U,0xf19bU,
		0xf33dU,0xf671U,0xfce9U,0xe9d9U,0xc3b9U,0x9779U,0x3ef9U,0x7df2U,0xfbe4U,0xe7c3U,
		0xdf8dU,0xaf11U,0x4e29U,0x9c52U,0x28afU,0x515eU,0xa2bcU,0x5573U,0xaae6U,0x45c7U,
		0x8b8eU,0x0717U,0x0e2eU,0x1c5cU,0x38b8U,0x7170U,0xe2e0U,0xd5cbU,0xbb9dU,0x6731U,
		0xce62U,0x8ccfU,0x0995U,0x132aU,0x2654U,0x4ca8U,0x9950U,0x22abU,0x4556U,0x8aacU,
		0x0553U,0x0aa6U,0x154cU,0x2a98U,0x5530U,0xaa60U,0x44cbU,0x8996U,0x0327U,0x064eU,
		0x0c9cU,0x1938U,0x3270U,0x64e0U,0xc9c0U,0x838bU,0x171dU,0x2e3aU,0x5c74U,0xb8e8U,
		0x61dbU,0xc3b6U,0x9767U,0x3ec5U,0x7d8aU,0xfb14U,0xe623U,0xdc4dU,0xa891U,0x4129U,
		0x8252U,0x14afU,0x295eU,0x52bcU,0xa578U,0x5afbU,0xb5f6U,0x7be7U,0xf7ceU,0xff97U,
		0xef25U,0xce41U,0x8c89U,0x0919U,0x1232U,0x2464U,0x48c8U,0x9190U,0x332bU,0x6656U,
		0xccacU,0x8953U,0x02adU,0x055aU,0x0ab4U,0x1568U,0x2ad0U,0x55a0U,0xab40U,0x468bU,
		0x8d16U,0x0a27U,0x144eU,0x289cU,0x5138U,0xa270U,0x54ebU,0xa9d6U,0x43a7U,0x874eU,
		0x1e97U,0x3d2eU,0x7a5cU,0xf4b8U,0xf97bU,0xe2fdU,0xd5f1U,0xbbe9U,0x67d9U,0xcfb2U,
		0x8f6fU,0x0ed5U,0x1daaU,0x3b54U,0x76a8U,0xed50U,0xcaabU,0x855dU,0x1ab1U,0x3562U,
		0x6ac4U,0xd588U,0xbb1bU,0x663dU,0xcc7aU,0x88ffU,0x01f5U,0x03eaU,0x07d4U,0x0fa8U,
		0x1f50U,0x3ea0U,0x7d40U,0xfa80U,0xe50bU,0xda1dU,0xa431U,0x5869U,0xb0d2U,0x71afU,
		0xe35eU,0xd6b7U,0xbd65U,0x6ac1U,0xd582U,0xbb0fU,0x6615U,0xcc2aU,0x885fU,0x00b5U,
		0x016aU,0x02d4U,0x05a8U,0x0b50U,0x16a0U,0x2d40U,0x5a80U,0xb500U,0x7a0bU,0xf416U,
		0xf827U,0xe045U,0xd081U,0xb109U,0x7219U,0xe432U,0xd86fU,0xa0d5U,0x51a1U,0xa342U,
		0x568fU,0xad1eU,0x4a37U,0x946eU,0x38d7U,0x71aeU,0xe35cU,0xd6b3U,0xbd6dU,0x6ad1U,
		0xd5a2U,0xbb4fU,0x6695U,0xcd2aU,0x8a5fU,0x04b5U,0x096aU,0x12d4U,0x25a8U,0x4b50U,
		0x96a0U,0x3d4bU,0x7a96U,0xf52cU,0xfa53U,0xe4adU,0xd951U,0xa2a9U,0x5559U,0xaab2U,
		0x456fU,0x8adeU,0x05b7U,0x0b6eU,0x16dcU,0x2db8U,0x5b70U,0xb6e0U,0x7dcbU,0xfb96U,
		0xe727U,0xde45U,0xac81U,0x4909U,0x9212U,0x342fU,0x685eU,0xd0bcU,0xb173U,0x72edU,
		0xe5daU,0xdbbfU,0xa775U,0x5ee1U,0xbdc2U,0x6b8fU,0xd71eU,0xbe37U,0x6c65U,0xd8caU,
		0xa19fU,0x5335U,0xa66aU,0x5cdfU,0xb9beU,0x6377U,0xc6eeU,0x9dd7U,0x2ba5U,0x574aU,
		0xae94U,0x4d23U,0x9a46U,0x2487U,0x490eU,0x921cU,0x3433U,0x6866U,0xd0ccU,0xb193U,
		0x732dU,0xe65aU,0xdcbfU,0xa975U,0x42e1U,0x85c2U,0x1b8fU,0x371eU,0x6e3cU,0xdc78U,
		0xa8fbU,0x41fdU,0x83faU,0x17ffU,0x2ffeU,0x5ffcU,0xbff8U,0x6ffbU,0xdff6U,0xafe7U,
		0x4fc5U,0x9f8aU,0x2f1fU,0x5e3eU,0xbc7cU,0x68f3U,0xd1e6U,0xb3c7U,0x7785U,0xef0aU,
		0xce1fU,0x8c35U,0x0861U,0x10c2U,0x2184U,0x4308U,0x8610U,0x1c2bU,0x3856U,0x70acU,
		0xe158U,0xd2bbU,0xb57dU,0x7af1U,0xf5e2U,0xfbcfU,0xe795U,0xdf21U,0xae49U,0x4c99U,
		0x9932U,0x226fU,0x44deU,0x89bcU,0x0373U,0x06e6U,0x0dccU,0x1b98U,0x3730U,0x6e60U,
		0xdcc0U,0xa98bU,0x431dU,0x863aU,0x1c7fU,0x38feU,0x71fcU,0xe3f8U,0xd7fbU,0xbffdU,
		0x6ff1U,0xdfe2U,0xafcfU,0x4f95U,0x9f2aU,0x2e5fU,0x5cbeU,0xb97cU,0x62f3U,0xc5e6U,
		0x9bc7U,0x2785U,0x4f0aU,0x9e14U,0x2c23U,0x5846U,0xb08cU,0x7113U,0xe226U,0xd447U,
		0xb885U,0x6101U,0xc202U,0x940fU,0x3815U,0x702aU,0xe054U,0xd0a3U,0xb14dU,0x7291U,
		0xe522U,0xda4fU,0xa495U,0x5921U,0xb242U,0x748fU,0xe91eU,0xc237U,0x9465U,0x38c1U,
		0x7182U,0xe304U,0xd603U,0xbc0dU,0x6811U,0xd022U,0xb04fU,0x7095U,0xe12aU,0xd25fU,
		0xb4b5U,0x7961U,0xf2c2U,0xf58fU,0xfb15U,0xe621U,0xdc49U,0xa899U,0x4139U,0x8272U,
		0x14efU,0x29deU,0x53bcU,0xa778U,0x5efbU,0xbdf6U,0x6be7U,0xd7ceU,0xbf97U,0x6f25U,
		0xde4aU,0xac9fU,0x4935U,0x926aU,0x34dfU,0x69beU,0xd37cU,0xb6f3U,0x7dedU,0xfbdaU,
		0xe7bfU,0xdf75U,0xaee1U,0x4dc9U,0x9b92U,0x272fU,0x4e5eU,0x9cbcU,0x2973U,0x52e6U,
		0xa5ccU,0x5b93U,0xb726U,0x7e47U,0xfc8eU,0xe917U,0xc225U,0x9441U,0x3889U,0x7112U,
		0xe224U,0xd443U,0xb88dU,0x6111U,0xc222U,0x944fU,0x3895U,0x712aU,0xe254U,0xd4a3U,
		0xb94dU,0x6291U,0xc522U,0x9a4fU,0x2495U,0x492aU,0x9254U,0x34a3U,0x6946U,0xd28cU,
		0xb513U,0x7a2dU,0xf45aU,0xf8bfU,0xe175U,0xd2e1U,0xb5c9U,0x7b99U,0xf732U,0xfe6fU,
		0xecd5U,0xc9a1U,0x8349U,0x1699U,0x2d32U,0x5a64U,0xb4c8U,0x799bU,0xf336U,0xf667U,
		0xfcc5U,0xe981U,0xc309U,0x9619U,0x3c39U,0x7872U,0xf0e4U,0xf1c3U,0xf38dU,0xf711U,
		0xfe29U,0xec59U,0xc8b9U,0x8179U,0x12f9U,0x25f2U,0x4be4U,0x97c8U,0x3f9bU,0x7f36U,
		0xfe6cU,0xecd3U,0xc9adU,0x8351U,0x16a9U,0x2d52U,0x5aa4U,0xb548U,0x7a9bU,0xf536U,
		0xfa67U,0xe4c5U,0xd981U,0xa309U,0x5619U,0xac32U,0x486fU,0x90deU,0x31b7U,0x636eU,
		0xc6dcU,0x9db3U,0x2b6dU,0x56daU,0xadb4U,0x4b63U,0x96c6U,0x3d87U,0x7b0eU,0xf61cU,
		0xfc33U,0xe86dU,0xc0d1U,0x91a9U,0x3359U,0x66b2U,0xcd64U,0x8ac3U,0x058dU,0x0b1aU,
		0x1634U,0x2c68U,0x58d0U,0xb1a0U,0x734bU,0xe696U,0xdd27U,0xaa45U,0x4481U,0x8902U,
		0x020fU,0x041eU,0x083cU,0x1078U,0x20f0U,0x41e0U,0x83c0U,0x178bU,0x2f16U,0x5e2cU,
		0xbc58U,0x68bbU,0xd176U,0xb2e7U,0x75c5U,0xeb8aU,0xc71fU,0x9e35U,0x2c61U,0x58c2U,
		0xb184U,0x7303U,0xe606U,0xdc07U,0xa805U,0x4001U,0x8002U,0x100fU,0x201eU,0x403cU,
		0x8078U,0x10fbU,0x21f6U,0x43ecU,0x87d8U,0x1fbbU,0x3f76U,0x7eecU,0xfdd8U,0xebbbU,
		0xc77dU,0x9ef1U,0x2de9U,0x5bd2U,0xb7a4U,0x7f43U,0xfe86U,0xed07U,0xca05U,0x8401U,
		0x1809U,0x3012U,0x6024U,0xc048U,0x909bU,0x313dU,0x627aU,0xc4f4U,0x99e3U,0x23cdU,
		0x479aU,0x8f34U,0x0e63U,0x1cc6U,0x398cU,0x7318U,0xe630U,0xdc6bU,0xa8ddU,0x41b1U,
		0x8362U,0x16cfU,0x2d9eU,0x5b3cU,0xb678U,0x7cfbU,0xf9f6U,0xe3e7U,0xd7c5U,0xbf81U,
		0x6f09U,0xde12U,0xac2fU,0x4855U,0x90aaU,0x315fU,0x62beU,0xc57cU,0x9af3U,0x25edU,
		0x4bdaU,0x97b4U,0x3f63U,0x7ec6U,0xfd8cU,0xeb13U,0xc62dU,0x9c51U,0x28a9U,0x5152U,
		0xa2a4U,0x5543U,0xaa86U,0x4507U,0x8a0eU,0x0417U,0x082eU,0x105cU,0x20b8U,0x4170U,
		0x82e0U,0x15cbU,0x2b96U,0x572cU,0xae58U,0x4cbbU,0x9976U,0x22e7U,0x45ceU,0x8b9cU,
		0x0733U,0x0e66U,0x1cccU,0x3998U,0x7330U,0xe660U,0xdccbU,0xa99dU,0x4331U,0x8662U,
		0x1ccfU,0x399eU,0x733cU,0xe678U,0xdcfbU,0xa9fdU,0x43f1U,0x87e2U,0x1fcfU,0x3f9eU,
		0x7f3cU,0xfe78U,0xecfbU,0xc9fdU,0x83f1U,0x17e9U,0x2fd2U,0x5fa4U,0xbf48U,0x6e9bU,
		0xdd36U,0xaa67U,0x44c5U,0x898aU,0x031fU,0x063eU,0x0c7cU,0x18f8U,0x31f0U,0x63e0U,
		0xc7c0U,0x9f8bU,0x2f1dU,0x5e3aU,0xbc74U,0x68e3U,0xd1c6U,0xb387U,0x7705U,0xee0aU,
		0xcc1fU,0x8835U,0x0061U,0x00c2U,0x0184U,0x0308U,0x0610U,0x0c20U,0x1840U,0x3080U,
		0x6100U,0xc200U,0x940bU,0x381dU,0x703aU,0xe074U,0xd0e3U,0xb1cdU,0x7391U,0xe722U,
		0xde4fU,0xac95U,0x4921U,0x9242U,0x348fU,0x691eU,0xd23cU,0xb473U,0x78edU,0xf1daU,
		0xf3bfU,0xf775U,0xfee1U,0xedc9U,0xcb99U,0x8739U,0x1e79U,0x3cf2U,0x79e4U,0xf3c8U,
		0xf79bU,0xff3dU,0xee71U,0xcce9U,0x89d9U,0x03b9U,0x0772U,0x0ee4U,0x1dc8U,0x3b90U,
		0x7720U,0xee40U,0xcc8bU,0x891dU,0x0231U,0x0462U,0x08c4U,0x1188U,0x2310U,0x4620U,
		0x8c40U,0x088bU,0x1116U,0x222cU,0x4458U,0x88b0U,0x016bU,0x02d6U,0x05acU,0x0b58U,
		0x16b0U,0x2d60U,0x5ac0U,0xb580U,0x7b0bU,0xf616U,0xfc27U,0xe845U,0xc081U,0x9109U,
		0x3219U,0x6432U,0xc864U,0x80c3U,0x118dU,0x231aU,0x4634U,0x8c68U,0x08dbU,0x11b6U,
		0x236cU,0x46d8U,0x8db0U,0x0b6bU,0x16d6U,0x2dacU,0x5b58U,0xb6b0U,0x7d6bU,0xfad6U,
		0xe5a7U,0xdb45U,0xa681U,0x5d09U,0xba12U,0x642fU,0xc85eU,0x80b7U,0x1165U,0x22caU,
		0x4594U,0x8b28U,0x065bU,0x0cb6U,0x196cU,0x32d8U,0x65b0U,0xcb60U,0x86cbU,0x1d9dU,
		0x3b3aU,0x7674U,0xece8U,0xc9dbU,0x83bdU,0x1771U,0x2ee2U,0x5dc4U,0xbb88U,0x671bU,
		0xce36U,0x8c67U,0x08c5U,0x118aU,0x2314U,0x4628U,0x8c50U,0x08abU,0x1156U,0x22acU,
		0x4558U,0x8ab0U,0x056bU,0x0ad6U,0x15acU,0x2b58U,0x56b0U,0xad60U,0x4acbU,0x9596U,
		0x3b27U,0x764eU,0xec9cU,0xc933U,0x826dU,0x14d1U,0x29a2U,0x5344U,0xa688U,0x5d1bU,
		0xba36U,0x6467U,0xc8ceU,0x8197U,0x1325U,0x264aU,0x4c94U,0x9928U,0x225bU,0x44b6U,
		0x896cU,0x02d3U,0x05a6U,0x0b4cU,0x1698U,0x2d30U,0x5a60U,0xb4c0U,0x798bU,0xf316U,
		0xf627U,0xfc45U,0xe881U,0xc109U,0x9219U,0x3439U,0x6872U,0xd0e4U,0xb1c3U,0x738dU,
		0xe71aU,0xde3fU,0xac75U,0x48e1U,0x91c2U,0x338fU,0x671eU,0xce3cU,0x8c73U,0x08edU,
		0x11daU,0x23b4U,0x4768U,0x8ed0U,0x0dabU,0x1b56U,0x36acU,0x6d58U,0xdab0U,0xa56bU,
		0x5addU,0xb5baU,0x7b7fU,0xf6feU,0xfdf7U,0xebe5U,0xc7c1U,0x9f89U,0x2f19U,0x5e32U,
		0xbc64U,0x68c3U,0xd186U,0xb307U,0x7605U,0xec0aU,0xc81fU,0x8035U,0x1061U,0x20c2U,
		0x4184U,0x8308U,0x161bU,0x2c36U,0x586cU,0xb0d8U,0x71bbU,0xe376U,0xd6e7U,0xbdc5U,
		0x6b81U,0xd702U,0xbe0fU,0x6c15U,0xd82aU,0xa05fU,0x50b5U,0xa16aU,0x52dfU,0xa5beU,
		0x5b77U,0xb6eeU,0x7dd7U,0xfbaeU,0xe757U,0xdea5U,0xad41U,0x4a89U,0x9512U,0x3a2fU,
		0x745eU,0xe8bcU,0xc173U,0x92edU,0x35d1U,0x6ba2U,0xd744U,0xbe83U,0x6d0dU,0xda1aU,
		0xa43fU,0x5875U,0xb0eaU,0x71dfU,0xe3beU,0xd777U,0xbee5U,0x6dc1U,0xdb82U,0xa70fU,
		0x5e15U,0xbc2aU,0x685fU,0xd0beU,0xb177U,0x72e5U,0xe5caU,0xdb9fU,0xa735U,0x5e61U,
		0xbcc2U,0x698fU,0xd31eU,0xb637U,0x7c65U,0xf8caU,0xe19fU,0xd335U,0xb661U,0x7cc9U,
		0xf992U,0xe32fU,0xd655U,0xbca1U,0x6949U,0xd292U,0xb52fU,0x7a55U,0xf4aaU,0xf95fU,
		0xe2b5U,0xd561U,0xbac9U,0x6599U,0xcb32U,0x866fU,0x1cd5U,0x39aaU,0x7354U,0xe6a8U,
		0xdd5bU,0xaabdU,0x4571U,0x8ae2U,0x05cfU,0x0b9eU,0x173cU,0x2e78U,0x5cf0U,0xb9e0U,
		0x63cbU,0xc796U,0x9f27U,0x2e45U,0x5c8aU,0xb914U,0x6223U,0xc446U,0x9887U,0x2105U,
		0x420aU,0x8414U,0x1823U,0x3046U,0x608cU,0xc118U,0x923bU,0x347dU,0x68faU,0xd1f4U,
		0xb3e3U,0x77cdU,0xef9aU,0xcf3fU,0x8e75U,0x0ce1U,0x19c2U,0x3384U,0x6708U,0xce10U,
		0x8c2bU,0x085dU,0x10baU,0x2174U,0x42e8U,0x85d0U,0x1babU,0x3756U,0x6eacU,0xdd58U,
		0xaabbU,0x457dU,0x8afaU,0x05ffU,0x0bfeU,0x17fcU,0x2ff8U,0x5ff0U,0xbfe0U,0x6fcbU,
		0xdf96U,0xaf27U,0x4e45U,0x9c8aU,0x291fU,0x523eU,0xa47cU,0x58f3U,0xb1e6U,0x73c7U,
		0xe78eU,0xdf17U,0xae25U,0x4c41U,0x9882U,0x210fU,0x421eU,0x843cU,0x1873U,0x30e6U,
		0x61ccU,0xc398U,0x973bU,0x3e7dU,0x7cfaU,0xf9f4U,0xe3e3U,0xd7cdU,0xbf91U,0x6f29U,
		0xde52U,0xacafU,0x4955U,0x92aaU,0x355fU,0x6abeU,0xd57cU,0xbaf3U,0x65edU,0xcbdaU,
		0x87bfU,0x1f75U,0x3eeaU,0x7dd4U,0xfba8U,0xe75bU,0xdebdU,0xad71U,0x4ae9U,0x95d2U,
		0x3bafU,0x775eU,0xeebcU,0xcd73U,0x8aedU,0x05d1U,0x0ba2U,0x1744U,0x2e88U,0x5d10U,
		0xba20U,0x644bU,0xc896U,0x8127U,0x1245U,0x248aU,0x4914U,0x9228U,0x345bU,0x68b6U,
		0xd16cU,0xb2d3U,0x75adU,0xeb5aU,0xc6bfU,0x9d75U,0x2ae1U,0x55c2U,0xab84U,0x4703U,
		0x8e06U,0x0c07U,0x180eU,0x301cU,0x6038U,0xc070U,0x90ebU,0x31ddU,0x63baU,0xc774U,
		0x9ee3U,0x2dcdU,0x5b9aU,0xb734U,0x7e63U,0xfcc6U,0xe987U,0xc305U,0x9601U,0x3c09U,
		0x7812U,0xf024U,0xf043U,0xf08dU,0xf111U,0xf229U,0xf459U,0xf8b9U,0xe179U,0xd2f9U,
		0xb5f9U,0x7bf9U,0xf7f2U,0xffefU,0xefd5U,0xcfa1U,0x8f49U,0x0e99U,0x1d32U,0x3a64U,
		0x74c8U,0xe990U,0xc32bU,0x965dU,0x3cb1U,0x7962U,0xf2c4U,0xf583U,0xfb0dU,0xe611U,
		0xdc29U,0xa859U,0x40b9U,0x8172U,0x12efU,0x25deU,0x4bbcU,0x9778U,0x3efbU,0x7df6U,
		0xfbecU,0xe7d3U,0xdfadU,0xaf51U,0x4ea9U,0x9d52U,0x2aafU,0x555eU,0xaabcU,0x4573U,
		0x8ae6U,0x05c7U,0x0b8eU,0x171cU,0x2e38U,0x5c70U,0xb8e0U,0x61cbU,0xc396U,0x9727U,
		0x3e45U,0x7c8aU,0xf914U,0xe223U,0xd44dU,0xb891U,0x6129U,0xc252U,0x94afU,0x3955U,
		0x72aaU,0xe554U,0xdaa3U,0xa54dU,0x5a91U,0xb522U,0x7a4fU,0xf49eU,0xf937U,0xe265U,
		0xd4c1U,0xb989U,0x6319U,0xc632U,0x9c6fU,0x28d5U,0x51aaU,0xa354U,0x56a3U,0xad46U,
		0x4a87U,0x950eU,0x3a17U,0x742eU,0xe85cU,0xc0b3U,0x916dU,0x32d1U,0x65a2U,0xcb44U,
		0x8683U,0x1d0dU,0x3a1aU,0x7434U,0xe868U,0xc0dbU,0x91bdU,0x3371U,0x66e2U,0xcdc4U,
		0x8b83U,0x070dU,0x0e1aU,0x1c34U,0x3868U,0x70d0U,0xe1a0U,0xd34bU,0xb69dU,0x7d31U,
		0xfa62U,0xe4cfU,0xd995U,0xa321U,0x5649U,0xac92U,0x492fU,0x925eU,0x34b7U,0x696eU,
		0xd2dcU,0xb5b3U,0x7b6dU,0xf6daU,0xfdbfU,0xeb75U,0xc6e1U,0x9dc9U,0x2b99U,0x5732U,
		0xae64U,0x4cc3U,0x9986U,0x2307U,0x460eU,0x8c1cU,0x0833U,0x1066U,0x20ccU,0x4198U,
		0x8330U,0x166bU,0x2cd6U,0x59acU,0xb358U,0x76bbU,0xed76U,0xcae7U,0x85c5U,0x1b81U,
		0x3702U,0x6e04U,0xdc08U,0xa81bU,0x403dU,0x807aU,0x10ffU,0x21feU,0x43fcU,0x87f8U,
		0x1ffbU,0x3ff6U,0x7fecU,0xffd8U,0xefbbU,0xcf7dU,0x8ef1U,0x0de9U,0x1bd2U,0x37a4U,
		0x6f48U,0xde90U,0xad2bU,0x4a5dU,0x94baU,0x397fU,0x72feU,0xe5fcU,0xdbf3U,0xa7edU,
		0x5fd1U,0xbfa2U,0x6f4fU,0xde9eU,0xad37U,0x4a65U,0x94caU,0x399fU,0x733eU,0xe67cU,
		0xdcf3U,0xa9edU,0x43d1U,0x87a2U,0x1f4fU,0x3e9eU,0x7d3cU,0xfa78U,0xe4fbU,0xd9fdU,
		0xa3f1U,0x57e9U,0xafd2U,0x4fafU,0x9f5eU,0x2eb7U,0x5d6eU,0xbadcU,0x65b3U,0xcb66U,
		0x86c7U,0x1d85U,0x3b0aU,0x7614U,0xec28U,0xc85bU,0x80bdU,0x1171U,0x22e2U,0x45c4U,
		0x8b88U,0x071bU,0x0e36U,0x1c6cU,0x38d8U,0x71b0U,0xe360U,0xd6cbU,0xbd9dU,0x6b31U,
		0xd662U,0xbccfU,0x6995U,0xd32aU,0xb65fU,0x7cb5U,0xf96aU,0xe2dfU,0xd5b5U,0xbb61U,
		0x66c9U,0xcd92U,0x8b2fU,0x0655U,0x0caaU,0x1954U,0x32a8U,0x6550U,0xcaa0U,0x854bU,
		0x1a9dU,0x353aU,0x6a74U,0xd4e8U,0xb9dbU,0x63bdU,0xc77aU,0x9effU,0x2df5U,0x5beaU,
		0xb7d4U,0x7fa3U,0xff46U,0xee87U,0xcd05U,0x8a01U,0x0409U,0x0812U,0x1024U,0x2048U,
		0x4090U,0x8120U,0x124bU,0x2496U,0x492cU,0x9258U,0x34bbU,0x6976U,0xd2ecU,0xb5d3U,
		0x7badU,0xf75aU,0xfebfU,0xed75U,0xcae1U,0x85c9U,0x1b99U,0x3732U,0x6e64U,0xdcc8U,
		0xa99bU,0x433dU,0x867aU,0x1cffU,0x39feU,0x73fcU,0xe7f8U,0xdffbU,0xaffdU,0x4ff1U,
		0x9fe2U,0x2fcfU,0x5f9eU,0xbf3cU,0x6e73U,0xdce6U,0xa9c7U,0x4385U,0x870aU,0x1e1fU,
		0x3c3eU,0x787cU,0xf0f8U,0xf1fbU,0xf3fdU,0xf7f1U,0xffe9U,0xefd9U,0xcfb9U,0x8f79U,
		0x0ef9U,0x1df2U,0x3be4U,0x77c8U,0xef90U,0xcf2bU,0x8e5dU,0x0cb1U,0x1962U,0x32c4U,
		0x6588U,0xcb10U,0x862bU,0x1c5dU,0x38baU,0x7174U,0xe2e8U,0xd5dbU,0xbbbdU,0x6771U,
		0xcee2U,0x8dcfU,0x0b95U,0x172aU,0x2e54U,0x5ca8U,0xb950U,0x62abU,0xc556U,0x9aa7U,
		0x2545U,0x4a8aU,0x9514U,0x3a23U,0x7446U,0xe88cU,0xc113U,0x922dU,0x3451U,0x68a2U,
		0xd144U,0xb283U,0x750dU,0xea1aU,0xc43fU,0x9875U,0x20e1U,0x41c2U,0x8384U,0x1703U,
		0x2e06U,0x5c0cU,0xb818U,0x603bU,0xc076U,0x90e7U,0x31c5U,0x638aU,0xc714U,0x9e23U,
		0x2c4dU,0x589aU,0xb134U,0x7263U,0xe4c6U,0xd987U,0xa305U,0x5601U,0xac02U,0x480fU,
		0x901eU,0x3037U,0x606eU,0xc0dcU,0x91b3U,0x336dU,0x66daU,0xcdb4U,0x8b63U,0x06cdU,
		0x0d9aU,0x1b34U,0x3668U,0x6cd0U,0xd9a0U,0xa34bU,0x569dU,0xad3aU,0x4a7fU,0x94feU,
		0x39f7U,0x73eeU,0xe7dcU,0xdfb3U,0xaf6dU,0x4ed1U,0x9da2U,0x2b4fU,0x569eU,0xad3cU,
		0x4a73U,0x94e6U,0x39c7U,0x738eU,0xe71cU,0xde33U,0xac6dU,0x48d1U,0x91a2U,0x334fU,
		0x669eU,0xcd3cU,0x8a73U,0x04edU,0x09daU,0x13b4U,0x2768U,0x4ed0U,0x9da0U,0x2b4bU,
		0x5696U,0xad2cU,0x4a53U,0x94a6U,0x3947U,0x728eU,0xe51cU,0xda33U,0xa46dU,0x58d1U,
		0xb1a2U,0x734fU,0xe69eU,0xdd37U,0xaa65U,0x44c1U,0x8982U,0x030fU,0x061eU,0x0c3cU,
		0x1878U,0x30f0U,0x61e0U,0xc3c0U,0x978bU,0x3f1dU,0x7e3aU,0xfc74U,0xe8e3U,0xc1cdU,
		0x9391U,0x3729U,0x6e52U,0xdca4U,0xa943U,0x428dU,0x851aU,0x1a3fU,0x347eU,0x68fcU,
		0xd1f8U,0xb3fbU,0x77fdU,0xeffaU,0xcfffU,0x8ff5U,0x0fe1U,0x1fc2U,0x3f84U,0x7f08U,
		0xfe10U,0xec2bU,0xc85dU,0x80b1U,0x1169U,0x22d2U,0x45a4U,0x8b48U,0x069bU,0x0d36U,
		0x1a6cU,0x34d8U,0x69b0U,0xd360U,0xb6cbU,0x7d9dU,0xfb3aU,0xe67fU,0xdcf5U,0xa9e1U,
		0x43c9U,0x8792U,0x1f2fU,0x3e5eU,0x7cbcU,0xf978U,0xe2fbU,0xd5fdU,0xbbf1U,0x67e9U,
		0xcfd2U,0x8fafU,0x0f55U,0x1eaaU,0x3d54U,0x7aa8U,0xf550U,0xfaabU,0xe55dU,0xdab1U,
		0xa569U,0x5ad9U,0xb5b2U,0x7b6fU,0xf6deU,0xfdb7U,0xeb65U,0xc6c1U,0x9d89U,0x2b19U,
		0x5632U,0xac64U,0x48c3U,0x9186U,0x3307U,0x660eU,0xcc1cU,0x8833U,0x006dU,0x00daU,
		0x01b4U,0x0368U,0x06d0U,0x0da0U,0x1b40U,0x3680U,0x6d00U,0xda00U,0xa40bU,0x581dU,
		0xb03aU,0x707fU,0xe0feU,0xd1f7U,0xb3e5U,0x77c1U,0xef82U,0xcf0fU,0x8e15U,0x0c21U,
		0x1842U,0x3084U,0x6108U,0xc210U,0x942bU,0x385dU,0x70baU,0xe174U,0xd2e3U,0xb5cdU,
		0x7b91U,0xf722U,0xfe4fU,0xec95U,0xc921U,0x8249U,0x1499U,0x2932U,0x5264U,0xa4c8U,
		0x599bU,0xb336U,0x7667U,0xecceU,0xc997U,0x8325U,0x1641U,0x2c82U,0x5904U,0xb208U,
		0x741bU,0xe836U,0xc067U,0x90c5U,0x3181U,0x6302U,0xc604U,0x9c03U,0x280dU,0x501aU,
		0xa034U,0x5063U,0xa0c6U,0x5187U,0xa30eU,0x5617U,0xac2eU,0x4857U,0x90aeU,0x3157U,
		0x62aeU,0xc55cU,0x9ab3U,0x256dU,0x4adaU,0x95b4U,0x3b63U,0x76c6U,0xed8cU,0xcb13U,
		0x862dU,0x1c51U,0x38a2U,0x7144U,0xe288U,0xd51bU,0xba3dU,0x6471U,0xc8e2U,0x81cfU,
		0x1395U,0x272aU,0x4e54U,0x9ca8U,0x295bU,0x52b6U,0xa56cU,0x5ad3U,0xb5a6U,0x7b47U,
		0xf68eU,0xfd17U,0xea25U,0xc441U,0x9889U,0x2119U,0x4232U,0x8464U,0x18c3U,0x3186U,
		0x630cU,0xc618U,0x9c3bU,0x287dU,0x50faU,0xa1f4U,0x53e3U,0xa7c6U,0x5f87U,0xbf0eU,
		0x6e17U,0xdc2eU,0xa857U,0x40a5U,0x814aU,0x129fU,0x253eU,0x4a7cU,0x94f8U,0x39fbU,
		0x73f6U,0xe7ecU,0xdfd3U,0xafadU,0x4f51U,0x9ea2U,0x2d4fU,0x5a9eU,0xb53cU,0x7a73U,
		0xf4e6U,0xf9c7U,0xe385U,0xd701U,0xbe09U,0x6c19U,0xd832U,0xa06fU,0x50d5U,0xa1aaU,
		0x535fU,0xa6beU,0x5d77U,0xbaeeU,0x65d7U,0xcbaeU,0x8757U,0x1ea5U,0x3d4aU,0x7a94U,
		0xf528U,0xfa5bU,0xe4bdU,0xd971U,0xa2e9U,0x55d9U,0xabb2U,0x476fU,0x8edeU,0x0db7U,
		0x1b6eU,0x36dcU,0x6db8U,0xdb70U,0xa6ebU,0x5dddU,0xbbbaU,0x677fU,0xcefeU,0x8df7U,
		0x0be5U,0x17caU,0x2f94U,0x5f28U,0xbe50U,0x6cabU,0xd956U,0xa2a7U,0x5545U,0xaa8aU,
		0x451fU,0x8a3eU,0x0477U,0x08eeU,0x11dcU,0x23b8U,0x4770U,0x8ee0U,0x0dcbU,0x1b96U,
		0x372cU,0x6e58U,0xdcb0U,0xa96bU,0x42ddU,0x85baU,0x1b7fU,0x36feU,0x6dfcU,0xdbf8U,
		0xa7fbU,0x5ffdU,0xbffaU,0x6fffU,0xdffeU,0xaff7U,0x4fe5U,0x9fcaU,0x2f9fU,0x5f3eU,
		0xbe7cU,0x6cf3U,0xd9e6U,0xa3c7U,0x5785U,0xaf0aU,0x4e1fU,0x9c3eU,0x2877U,0x50eeU,
		0xa1dcU,0x53b3U,0xa766U,0x5ec7U,0xbd8eU,0x6b17U,0xd62eU,0xbc57U,0x68a5U,0xd14aU,
		0xb29fU,0x7535U,0xea6aU,0xc4dfU,0x99b5U,0x2361U,0x46c2U,0x8d84U,0x0b03U,0x1606U,
		0x2c0cU,0x5818U,0xb030U,0x706bU,0xe0d6U,0xd1a7U,0xb345U,0x7681U,0xed02U,0xca0fU,
		0x8415U,0x1821U,0x3042U,0x6084U,0xc108U,0x921bU,0x343dU,0x687aU,0xd0f4U,0xb1e3U,
		0x73cdU,0xe79aU,0xdf3fU,0xae75U,0x4ce1U,0x99c2U,0x238fU,0x471eU,0x8e3cU,0x0c73U,
		0x18e6U,0x31ccU,0x6398U,0xc730U,0x9e6bU,0x2cddU,0x59baU,0xb374U,0x76e3U,0xedc6U,
		0xcb87U,0x8705U,0x1e01U,0x3c02U,0x7804U,0xf008U,0xf01bU,0xf03dU,0xf071U,0xf0e9U,
		0xf1d9U,0xf3b9U,0xf779U,0xfef9U,0xedf9U,0xcbf9U,0x87f9U,0x1ff9U,0x3ff2U,0x7fe4U,
		0xffc8U,0xef9bU,0xcf3dU,0x8e71U,0x0ce9U,0x19d2U,0x33a4U,0x6748U,0xce90U,0x8d2bU,
		0x0a5dU,0x14baU,0x2974U,0x52e8U,0xa5d0U,0x5babU,0xb756U,0x7ea7U,0xfd4eU,0xea97U,
		0xc525U,0x9a41U,0x2489U,0x4912U,0x9224U,0x3443U,0x6886U,0xd10cU,0xb213U,0x742dU,
		0xe85aU,0xc0bfU,0x9175U,0x32e1U,0x65c2U,0xcb84U,0x8703U,0x1e0dU,0x3c1aU,0x7834U,
		0xf068U,0xf0dbU,0xf1bdU,0xf371U,0xf6e9U,0xfdd9U,0xebb9U,0xc779U,0x9ef9U,0x2df9U,
		0x5bf2U,0xb7e4U,0x7fc3U,0xff86U,0xef07U,0xce05U,0x8c01U,0x0809U,0x1012U,0x2024U,
		0x4048U,0x8090U,0x112bU,0x2256U,0x44acU,0x8958U,0x02bbU,0x0576U,0x0aecU,0x15d8U,
		0x2bb0U,0x5760U,0xaec0U,0x4d8bU,0x9b16U,0x2627U,0x4c4eU,0x989cU,0x2133U,0x4266U,
		0x84ccU,0x1993U,0x3326U,0x664cU,0xcc98U,0x893bU,0x027dU,0x04faU,0x09f4U,0x13e8U,
		0x27d0U,0x4fa0U,0x9f40U,0x2e8bU,0x5d16U,0xba2cU,0x6453U,0xc8a6U,0x8147U,0x1285U,
		0x250aU,0x4a14U,0x9428U,0x385bU,0x70b6U,0xe16cU,0xd2d3U,0xb5adU,0x7b51U,0xf6a2U,
		0xfd4fU,0xea95U,0xc521U,0x9a49U,0x2499U,0x4932U,0x9264U,0x34c3U,0x6986U,0xd30cU,
		0xb613U,0x7c2dU,0xf85aU,0xe0bfU,0xd175U,0xb2e1U,0x75c9U,0xeb92U,0xc72fU,0x9e55U,
		0x2ca1U,0x5942U,0xb284U,0x7503U,0xea06U,0xc407U,0x9805U,0x2001U,0x4002U,0x8004U,
		0x1003U,0x2006U,0x400cU,0x8018U,0x103bU,0x2076U,0x40ecU,0x81d8U,0x13bbU,0x2776U,
		0x4eecU,0x9dd8U,0x2bbbU,0x5776U,0xaeecU,0x4dd3U,0x9ba6U,0x2747U,0x4e8eU,0x9d1cU,
		0x2a33U,0x5466U,0xa8ccU,0x4193U,0x8326U,0x1647U,0x2c8eU,0x591cU,0xb238U,0x747bU,
		0xe8f6U,0xc1e7U,0x93c5U,0x3781U,0x6f02U,0xde04U,0xac03U,0x480dU,0x901aU,0x303fU,
		0x607eU,0xc0fcU,0x91f3U,0x33edU,0x67daU,0xcfb4U,0x8f63U,0x0ecdU,0x1d9aU,0x3b34U,
		0x7668U,0xecd0U,0xc9abU,0x835dU,0x16b1U,0x2d62U,0x5ac4U,0xb588U,0x7b1bU,0xf636U,
		0xfc67U,0xe8c5U,0xc181U,0x9309U,0x3619U,0x6c32U,0xd864U,0xa0c3U,0x518dU,0xa31aU,
		0x563fU,0xac7eU,0x48f7U,0x91eeU,0x33d7U,0x67aeU,0xcf5cU,0x8eb3U,0x0d6dU,0x1adaU,
		0x35b4U,0x6b68U,0xd6d0U,0xbdabU,0x6b5dU,0xd6baU,0xbd7fU,0x6af5U,0xd5eaU,0xbbdfU,
		0x67b5U,0xcf6aU,0x8edfU,0x0db5U,0x1b6aU,0x36d4U,0x6da8U,0xdb50U,0xa6abU,0x5d5dU,
		0xbabaU,0x657fU,0xcafeU,0x85f7U,0x1be5U,0x37caU,0x6f94U,0xdf28U,0xae5bU,0x4cbdU,
		0x997aU,0x22ffU,0x45feU,0x8bfcU,0x07f3U,0x0fe6U,0x1fccU,0x3f98U,0x7f30U,0xfe60U,
		0xeccbU,0xc99dU,0x8331U,0x1669U,0x2cd2U,0x59a4U,0xb348U,0x769bU,0xed36U,0xca67U,
		0x84c5U,0x1981U,0x3302U,0x6604U,0xcc08U,0x881bU,0x003dU,0x007aU,0x00f4U,0x01e8U,
		0x03d0U,0x07a0U,0x0f40U,0x1e80U,0x3d00U,0x7a00U,0xf400U,0xf80bU,0xe01dU,0xd031U,
		0xb069U,0x70d9U,0xe1b2U,0xd36fU,0xb6d5U,0x7da1U,0xfb42U,0xe68fU,0xdd15U,0xaa21U,
		0x4449U,0x8892U,0x012fU,0x025eU,0x04bcU,0x0978U,0x12f0U,0x25e0U,0x4bc0U,0x9780U,
		0x3f0bU,0x7e16U,0xfc2cU,0xe853U,0xc0adU,0x9151U,0x32a9U,0x6552U,0xcaa4U,0x8543U,
		0x1a8dU,0x351aU,0x6a34U,0xd468U,0xb8dbU,0x61bdU,0xc37aU,0x96ffU,0x3df5U,0x7beaU,
		0xf7d4U,0xffa3U,0xef4dU,0xce91U,0x8d29U,0x0a59U,0x14b2U,0x2964U,0x52c8U,0xa590U,
		0x5b2bU,0xb656U,0x7ca7U,0xf94eU,0xe297U,0xd525U,0xba41U,0x6489U,0xc912U,0x822fU,
		0x1455U,0x28aaU,0x5154U,0xa2a8U,0x555bU,0xaab6U,0x4567U,0x8aceU,0x0597U,0x0b2eU,
		0x165cU,0x2cb8U,0x5970U,0xb2e0U,0x75cbU,0xeb96U,0xc727U,0x9e45U,0x2c81U,0x5902U,
		0xb204U,0x7403U,0xe806U,0xc007U,0x9005U,0x3001U,0x6002U,0xc004U,0x9003U,0x300dU,
		0x601aU,0xc034U,0x9063U,0x30cdU,0x619aU,0xc334U,0x9663U,0x3ccdU,0x799aU,0xf334U,
		0xf663U,0xfccdU,0xe991U,0xc329U,0x9659U,0x3cb9U,0x7972U,0xf2e4U,0xf5c3U,0xfb8dU,
		0xe711U,0xde29U,0xac59U,0x48b9U,0x9172U,0x32efU,0x65deU,0xcbbcU,0x8773U,0x1eedU,
		0x3ddaU,0x7bb4U,0xf768U,0xfedbU,0xedbdU,0xcb71U,0x86e9U,0x1dd9U,0x3bb2U,0x7764U,
		0xeec8U,0xcd9bU,0x8b3dU,0x0671U,0x0ce2U,0x19c4U,0x3388U,0x6710U,0xce20U,0x8c4bU,
		0x089dU,0x113aU,0x2274U,0x44e8U,0x89d0U,0x03abU,0x0756U,0x0eacU,0x1d58U,0x3ab0U,
		0x7560U,0xeac0U,0xc58bU,0x9b1dU,0x2631U,0x4c62U,0x98c4U,0x2183U,0x4306U,0x860cU,
		0x1c13U,0x3826U,0x704cU,0xe098U,0xd13bU,0xb27dU,0x74f1U,0xe9e2U,0xc3cfU,0x9795U,
		0x3f21U,0x7e42U,0xfc84U,0xe903U,0xc20dU,0x9411U,0x3829U,0x7052U,0xe0a4U,0xd143U,
		0xb28dU,0x7511U,0xea22U,0xc44fU,0x9895U,0x2121U,0x4242U,0x8484U,0x1903U,0x3206U,
		0x640cU,0xc818U,0x803bU,0x107dU,0x20faU,0x41f4U,0x83e8U,0x17dbU,0x2fb6U,0x5f6cU,
		0xbed8U,0x6dbbU,0xdb76U,0xa6e7U,0x5dc5U,0xbb8aU,0x671fU,0xce3eU,0x8c77U,0x08e5U,
		0x11caU,0x2394U,0x4728U,0x8e50U,0x0cabU,0x1956U,0x32acU,0x6558U,0xcab0U,0x856bU,
		0x1addU,0x35baU,0x6b74U,0xd6e8U,0xbddbU,0x6bbdU,0xd77aU,0xbeffU,0x6df5U,0xdbeaU,
		0xa7dfU,0x5fb5U,0xbf6aU,0x6edfU,0xddbeU,0xab77U,0x46e5U,0x8dcaU,0x0b9fU,0x173eU,
		0x2e7cU,0x5cf8U,0xb9f0U,0x63ebU,0xc7d6U,0x9fa7U,0x2f45U,0x5e8aU,0xbd14U,0x6a23U,
		0xd446U,0xb887U,0x6105U,0xc20aU,0x941fU,0x3835U,0x706aU,0xe0d4U,0xd1a3U,0xb34dU,
		0x7691U,0xed22U,0xca4fU,0x8495U,0x1921U,0x3242U,0x6484U,0xc908U,0x821bU,0x143dU,
		0x287aU,0x50f4U,0xa1e8U,0x53dbU,0xa7b6U,0x5f67U,0xbeceU,0x6d97U,0xdb2eU,0xa657U,
		0x5ca5U,0xb94aU,0x629fU,0xc53eU,0x9a77U,0x24e5U,0x49caU,0x9394U,0x3723U,0x6e46U,
		0xdc8cU,0xa913U,0x422dU,0x845aU,0x18bfU,0x317eU,0x62fcU,0xc5f8U,0x9bfbU,0x27fdU,
		0x4ffaU,0x9ff4U,0x2fe3U,0x5fc6U,0xbf8cU,0x6f13U,0xde26U,0xac47U,0x4885U,0x910aU,
		0x321fU,0x643eU,0xc87cU,0x80f3U,0x11edU,0x23daU,0x47b4U,0x8f68U,0x0edbU,0x1db6U,
		0x3b6cU,0x76d8U,0xedb0U,0xcb6bU,0x86ddU,0x1db1U,0x3b62U,0x76c4U,0xed88U,0xcb1bU,
		0x863dU,0x1c71U,0x38e2U,0x71c4U,0xe388U,0xd71bU,0xbe3dU,0x6c71U,0xd8e2U,0xa1cfU,
		0x5395U,0xa72aU,0x5e5fU,0xbcbeU,0x6977U,0xd2eeU,0xb5d7U,0x7ba5U,0xf74aU,0xfe9fU,
		0xed35U,0xca61U,0x84c9U,0x1999U,0x3332U,0x6664U,0xccc8U,0x899bU,0x033dU,0x067aU,
		0x0cf4U,0x19e8U,0x33d0U,0x67a0U,0xcf40U,0x8e8bU,0x0d1dU,0x1a3aU,0x3474U,0x68e8U,
		0xd1d0U,0xb3abU,0x775dU,0xeebaU,0xcd7fU,0x8af5U,0x05e1U,0x0bc2U,0x1784U,0x2f08U,
		0x5e10U,0xbc20U,0x684bU,0xd096U,0xb127U,0x7245U,0xe48aU,0xd91fU,0xa235U,0x5461U,
		0xa8c2U,0x418fU,0x831eU,0x1637U,0x2c6eU,0x58dcU,0xb1b8U,0x737bU,0xe6f6U,0xdde7U,
		0xabc5U,0x4781U,0x8f02U,0x0e0fU,0x1c1eU,0x383cU,0x7078U,0xe0f0U,0xd1ebU,0xb3ddU,
		0x77b1U,0xef62U,0xcecfU,0x8d95U,0x0b21U,0x1642U,0x2c84U,0x5908U,0xb210U,0x742bU,
		0xe856U,0xc0a7U,0x9145U,0x3281U,0x6502U,0xca04U,0x8403U,0x180dU,0x301aU,0x6034U,
		0xc068U,0x90dbU,0x31bdU,0x637aU,0xc6f4U,0x9de3U,0x2bcdU,0x579aU,0xaf34U,0x4e63U,
		0x9cc6U,0x2987U,0x530eU,0xa61cU,0x5c33U,0xb866U,0x60c7U,0xc18eU,0x9317U,0x3625U,
		0x6c4aU,0xd894U,0xa123U,0x524dU,0xa49aU,0x593fU,0xb27eU,0x74f7U,0xe9eeU,0xc3d7U,
		0x97a5U,0x3f41U,0x7e82U,0xfd04U,0xea03U,0xc40dU,0x9811U,0x2029U,0x4052U,0x80a4U,
		0x1143U,0x2286U,0x450cU,0x8a18U,0x043bU,0x0876U,0x10ecU,0x21d8U,0x43b0U,0x8760U,
		0x1ecbU,0x3d96U,0x7b2cU,0xf658U,0xfcbbU,0xe97dU,0xc2f1U,0x95e9U,0x3bd9U,0x77b2U,
		0xef64U,0xcec3U,0x8d8dU,0x0b11U,0x1622U,0x2c44U,0x5888U,0xb110U,0x722bU,0xe456U,
		0xd8a7U,0xa145U,0x5281U,0xa502U,0x5a0fU,0xb41eU,0x7837U,0xf06eU,0xf0d7U,0xf1a5U,
		0xf341U,0xf689U,0xfd19U,0xea39U,0xc479U,0x98f9U,0x21f9U,0x43f2U,0x87e4U,0x1fc3U,
		0x3f86U,0x7f0cU,0xfe18U,0xec3bU,0xc87dU,0x80f1U,0x11e9U,0x23d2U,0x47a4U,0x8f48U,
		0x0e9bU,0x1d36U,0x3a6cU,0x74d8U,0xe9b0U,0xc36bU,0x96ddU,0x3db1U,0x7b62U,0xf6c4U,
		0xfd83U,0xeb0dU,0xc611U,0x9c29U,0x2859U,0x50b2U,0xa164U,0x52c3U,0xa586U,0x5b07U,
		0xb60eU,0x7c17U,0xf82eU,0xe057U,0xd0a5U,0xb141U,0x7289U,0xe512U,0xda2fU,0xa455U,
		0x58a1U,0xb142U,0x728fU,0xe51eU,0xda37U,0xa465U,0x58c1U,0xb182U,0x730fU,0xe61eU,
		0xdc37U,0xa865U,0x40c1U,0x8182U,0x130fU,0x261eU,0x4c3cU,0x9878U,0x20fbU,0x41f6U,
		0x83ecU,0x17d3U,0x2fa6U,0x5f4cU,0xbe98U,0x6d3bU,0xda76U,0xa4e7U,0x59c5U,0xb38aU,
		0x771fU,0xee3eU,0xcc77U,0x88e5U,0x01c1U,0x0382U,0x0704U,0x0e08U,0x1c10U,0x3820U,
		0x7040U,0xe080U,0xd10bU,0xb21dU,0x7431U,0xe862U,0xc0cfU,0x9195U,0x3321U,0x6642U,
		0xcc84U,0x8903U,0x020dU,0x041aU,0x0834U,0x1068U,0x20d0U,0x41a0U,0x8340U,0x168bU,
		0x2d16U,0x5a2cU,0xb458U,0x78bbU,0xf176U,0xf2e7U,0xf5c5U,0xfb81U,0xe709U,0xde19U,
		0xac39U,0x4879U,0x90f2U,0x31efU,0x63deU,0xc7bcU,0x9f73U,0x2eedU,0x5ddaU,0xbbb4U,
		0x6763U,0xcec6U,0x8d87U,0x0b05U,0x160aU,0x2c14U,0x5828U,0xb050U,0x70abU,0xe156U,
		0xd2a7U,0xb545U,0x7a81U,0xf502U,0xfa0fU,0xe415U,0xd821U,0xa049U,0x5099U,0xa132U,
		0x526fU,0xa4deU,0x59b7U,0xb36eU,0x76d7U,0xedaeU,0xcb57U,0x86a5U,0x1d41U,0x3a82U,
		0x7504U,0xea08U,0xc41bU,0x983dU,0x2071U,0x40e2U,0x81c4U,0x1383U,0x2706U,0x4e0cU,
		0x9c18U,0x283bU,0x5076U,0xa0ecU,0x51d3U,0xa3a6U,0x5747U,0xae8eU,0x4d17U,0x9a2eU,
		0x2457U,0x48aeU,0x915cU,0x32b3U,0x6566U,0xcaccU,0x8593U,0x1b2dU,0x365aU,0x6cb4U,
		0xd968U,0xa2dbU,0x55bdU,0xab7aU,0x46ffU,0x8dfeU,0x0bf7U,0x17eeU,0x2fdcU,0x5fb8U,
		0xbf70U,0x6eebU,0xddd6U,0xaba7U,0x4745U,0x8e8aU,0x0d1fU,0x1a3eU,0x347cU,0x68f8U,
		0xd1f0U,0xb3ebU,0x77ddU,0xefbaU,0xcf7fU,0x8ef5U,0x0de1U,0x1bc2U,0x3784U,0x6f08U,
		0xde10U,0xac2bU,0x485dU,0x90baU,0x317fU,0x62feU,0xc5fcU,0x9bf3U,0x27edU,0x4fdaU,
		0x9fb4U,0x2f63U,0x5ec6U,0xbd8cU,0x6b13U,0xd626U,0xbc47U,0x6885U,0xd10aU,0xb21fU,
		0x7435U,0xe86aU,0xc0dfU,0x91b5U,0x3361U,0x66c2U,0xcd84U,0x8b03U,0x060dU,0x0c1aU,
		0x1834U,0x3068U,0x60d0U,0xc1a0U,0x934bU,0x369dU,0x6d3aU,0xda74U,0xa4e3U,0x59cdU,
		0xb39aU,0x773fU,0xee7eU,0xccf7U,0x89e5U,0x03c1U,0x0782U,0x0f04U,0x1e08U,0x3c10U,
		0x7820U,0xf040U,0xf08bU,0xf11dU,0xf231U,0xf469U,0xf8d9U,0xe1b9U,0xd379U,0xb6f9U,
		0x7df9U,0xfbf2U,0xe7efU,0xdfd5U,0xafa1U,0x4f49U,0x9e92U,0x2d2fU,0x5a5eU,0xb4bcU,
		0x7973U,0xf2e6U,0xf5c7U,0xfb85U,0xe701U,0xde09U,0xac19U,0x4839U,0x9072U,0x30efU,
		0x61deU,0xc3bcU,0x9773U,0x3eedU,0x7ddaU,0xfbb4U,0xe763U,0xdecdU,0xad91U,0x4b29U,
		0x9652U,0x3cafU,0x795eU,0xf2bcU,0xf573U,0xfaedU,0xe5d1U,0xdba9U,0xa759U,0x5eb9U,
		0xbd72U,0x6aefU,0xd5deU,0xbbb7U,0x6765U,0xcecaU,0x8d9fU,0x0b35U,0x166aU,0x2cd4U,
		0x59a8U,0xb350U,0x76abU,0xed56U,0xcaa7U,0x8545U,0x1a81U,0x3502U,0x6a04U,0xd408U,
		0xb81bU,0x603dU,0xc07aU,0x90ffU,0x31f5U,0x63eaU,0xc7d4U,0x9fa3U,0x2f4dU,0x5e9aU,
		0xbd34U,0x6a63U,0xd4c6U,0xb987U,0x6305U,0xc60aU,0x9c1fU,0x2835U,0x506aU,0xa0d4U,
		0x51a3U,0xa346U,0x5687U,0xad0eU,0x4a17U,0x942eU,0x3857U,0x70aeU,0xe15cU,0xd2b3U,
		0xb56dU,0x7ad1U,0xf5a2U,0xfb4fU,0xe695U,0xdd21U,0xaa49U,0x4499U,0x8932U,0x026fU,
		0x04deU,0x09bcU,0x1378U,0x26f0U,0x4de0U,0x9bc0U,0x278bU,0x4f16U,0x9e2cU,0x2c53U,
		0x58a6U,0xb14cU,0x7293U,0xe526U,0xda47U,0xa485U,0x5901U,0xb202U,0x740fU,0xe81eU,
		0xc037U,0x9065U,0x30c1U,0x6182U,0xc304U,0x9603U,0x3c0dU,0x781aU,0xf034U,0xf063U,
		0xf0cdU,0xf191U,0xf329U,0xf659U,0xfcb9U,0xe979U,0xc2f9U,0x95f9U,0x3bf9U,0x77f2U,
		0xefe4U,0xcfc3U,0x8f8dU,0x0f11U,0x1e22U,0x3c44U,0x7888U,0xf110U,0xf22bU,0xf45dU,
		0xf8b1U,0xe169U,0xd2d9U,0xb5b9U,0x7b79U,0xf6f2U,0xfdefU,0xebd5U,0xc7a1U,0x9f49U,
		0x2e99U,0x5d32U,0xba64U,0x64c3U,0xc986U,0x8307U,0x1605U,0x2c0aU,0x5814U,0xb028U,
		0x705bU,0xe0b6U,0xd167U,0xb2c5U,0x7581U,0xeb02U,0xc60fU,0x9c15U,0x2821U,0x5042U,
		0xa084U,0x5103U,0xa206U,0x5407U,0xa80eU,0x4017U,0x802eU,0x1057U,0x20aeU,0x415cU,
		0x82b8U,0x157bU,0x2af6U,0x55ecU,0xabd8U,0x47bbU,0x8f76U,0x0ee7U,0x1dceU,0x3b9cU,
		0x7738U,0xee70U,0xccebU,0x89ddU,0x03b1U,0x0762U,0x0ec4U,0x1d88U,0x3b10U,0x7620U,
		0xec40U,0xc88bU,0x811dU,0x1231U,0x2462U,0x48c4U,0x9188U,0x331bU,0x6636U,0xcc6cU,
		0x88d3U,0x01adU,0x035aU,0x06b4U,0x0d68U,0x1ad0U,0x35a0U,0x6b40U,0xd680U,0xbd0bU,
		0x6a1dU,0xd43aU,0xb87fU,0x60f5U,0xc1eaU,0x93dfU,0x37b5U,0x6f6aU,0xded4U,0xada3U,
		0x4b4dU,0x969aU,0x3d3fU,0x7a7eU,0xf4fcU,0xf9f3U,0xe3edU,0xd7d1U,0xbfa9U,0x6f59U,
		0xdeb2U,0xad6fU,0x4ad5U,0x95aaU,0x3b5fU,0x76beU,0xed7cU,0xcaf3U,0x85edU,0x1bd1U,
		0x37a2U,0x6f44U,0xde88U,0xad1bU,0x4a3dU,0x947aU,0x38ffU,0x71feU,0xe3fcU,0xd7f3U,
		0xbfedU,0x6fd1U,0xdfa2U,0xaf4fU,0x4e95U,0x9d2aU,0x2a5fU,0x54beU,0xa97cU,0x42f3U,
		0x85e6U,0x1bc7U,0x378eU,0x6f1cU,0xde38U,0xac7bU,0x48fdU,0x91faU,0x33ffU,0x67feU,
		0xcffcU,0x8ff3U,0x0fedU,0x1fdaU,0x3fb4U,0x7f68U,0xfed0U,0xedabU,0xcb5dU,0x86b1U,
		0x1d69U,0x3ad2U,0x75a4U,0xeb48U,0xc69bU,0x9d3dU,0x2a71U,0x54e2U,0xa9c4U,0x4383U,
		0x8706U,0x1e07U,0x3c0eU,0x781cU,0xf038U,0xf07bU,0xf0fdU,0xf1f1U,0xf3e9U,0xf7d9U,
		0xffb9U,0xef79U,0xcef9U,0x8df9U,0x0bf9U,0x17f2U,0x2fe4U,0x5fc8U,0xbf90U,0x6f2bU,
		0xde56U,0xaca7U,0x4945U,0x928aU,0x351fU,0x6a3eU,0xd47cU,0xb8f3U,0x61edU,0xc3daU,
		0x97bfU,0x3f75U,0x7eeaU,0xfdd4U,0xeba3U,0xc74dU,0x9e91U,0x2d29U,0x5a52U,0xb4a4U,
		0x7943U,0xf286U,0xf507U,0xfa05U,0xe401U,0xd809U,0xa019U,0x5039U,0xa072U,0x50efU,
		0xa1deU,0x53b7U,0xa76eU,0x5ed7U,0xbdaeU,0x6b57U,0xd6aeU,0xbd57U,0x6aa5U,0xd54aU,
		0xba9fU,0x6535U,0xca6aU,0x84dfU,0x19b5U,0x336aU,0x66d4U,0xcda8U,0x8b5bU,0x06bdU,
		0x0d7aU,0x1af4U,0x35e8U,0x6bd0U,0xd7a0U,0xbf4bU,0x6e9dU,0xdd3aU,0xaa7fU,0x44f5U,
		0x89eaU,0x03dfU,0x07beU,0x0f7cU,0x1ef8U,0x3df0U,0x7be0U,0xf7c0U,0xff8bU,0xef1dU,
		0xce31U,0x8c69U,0x08d9U,0x11b2U,0x2364U,0x46c8U,0x8d90U,0x0b2bU,0x1656U,0x2cacU,
		0x5958U,0xb2b0U,0x756bU,0xead6U,0xc5a7U,0x9b45U,0x2681U,0x4d02U,0x9a04U,0x2403U,
		0x4806U,0x900cU,0x3013U,0x6026U,0xc04cU,0x9093U,0x312dU,0x625aU,0xc4b4U,0x9963U,
		0x22cdU,0x459aU,0x8b34U,0x0663U,0x0cc6U,0x198cU,0x3318U,0x6630U,0xcc60U,0x88cbU,
		0x019dU,0x033aU,0x0674U,0x0ce8U,0x19d0U,0x33a0U,0x6740U,0xce80U,0x8d0bU,0x0a1dU,
		0x143aU,0x2874U,0x50e8U,0xa1d0U,0x53abU,0xa756U,0x5ea7U,0xbd4eU,0x6a97U,0xd52eU,
		0xba57U,0x64a5U,0xc94aU,0x829fU,0x1535U,0x2a6aU,0x54d4U,0xa9a8U,0x435bU,0x86b6U,
		0x1d67U,0x3aceU,0x759cU,0xeb38U,0xc67bU,0x9cfdU,0x29f1U,0x53e2U,0xa7c4U,0x5f83U,
		0xbf06U,0x6e07U,0xdc0eU,0xa817U,0x4025U,0x804aU,0x109fU,0x213eU,0x427cU,0x84f8U,
		0x19fbU,0x33f6U,0x67ecU,0xcfd8U,0x8fbbU,0x0f7dU,0x1efaU,0x3df4U,0x7be8U,0xf7d0U,
		0xffabU,0xef5dU,0xceb1U,0x8d69U,0x0ad9U,0x15b2U,0x2b64U,0x56c8U,0xad90U,0x4b2bU,
		0x9656U,0x3ca7U,0x794eU,0xf29cU,0xf533U,0xfa6dU,0xe4d1U,0xd9a9U,0xa359U,0x56b9U,
		0xad72U,0x4aefU,0x95deU,0x3bb7U,0x776eU,0xeedcU,0xcdb3U,0x8b6dU,0x06d1U,0x0da2U,
		0x1b44U,0x3688U,0x6d10U,0xda20U,0xa44bU,0x589dU,0xb13aU,0x727fU,0xe4feU,0xd9f7U,
		0xa3e5U,0x57c1U,0xaf82U,0x4f0fU,0x9e1eU,0x2c37U,0x586eU,0xb0dcU,0x71b3U,0xe366U,
		0xd6c7U,0xbd85U,0x6b01U,0xd602U,0xbc0fU,0x6815U,0xd02aU,0xb05fU,0x70b5U,0xe16aU,
		0xd2dfU,0xb5b5U,0x7b61U,0xf6c2U,0xfd8fU,0xeb15U,0xc621U,0x9c49U,0x2899U,0x5132U,
		0xa264U,0x54c3U,0xa986U,0x4307U,0x860eU,0x1c17U,0x382eU,0x705cU,0xe0b8U,0xd17bU,
		0xb2fdU,0x75f1U,0xebe2U,0xc7cfU,0x9f95U,0x2f21U,0x5e42U,0xbc84U,0x6903U,0xd206U,
		0xb407U,0x7805U,0xf00aU,0xf01fU,0xf035U,0xf061U,0xf0c9U,0xf199U,0xf339U,0xf679U,
		0xfcf9U,0xe9f9U,0xc3f9U,0x97f9U,0x3ff9U,0x7ff2U,0xffe4U,0xefc3U,0xcf8dU,0x8f11U,
		0x0e29U,0x1c52U,0x38a4U,0x7148U,0xe290U,0xd52bU,0xba5dU,0x64b1U,0xc962U,0x82cfU,
		0x1595U,0x2b2aU,0x5654U,0xaca8U,0x495bU,0x92b6U,0x3567U,0x6aceU,0xd59cU,0xbb33U,
		0x666dU,0xccdaU,0x89bfU,0x0375U,0x06eaU,0x0dd4U,0x1ba8U,0x3750U,0x6ea0U,0xdd40U,
		0xaa8bU,0x451dU,0x8a3aU,0x047fU,0x08feU,0x11fcU,0x23f8U,0x47f0U,0x8fe0U,0x0fcbU,
		0x1f96U,0x3f2cU,0x7e58U,0xfcb0U,0xe96bU,0xc2ddU,0x95b1U,0x3b69U,0x76d2U,0xeda4U,
		0xcb43U,0x868dU,0x1d11U,0x3a22U,0x7444U,0xe888U,0xc11bU,0x923dU,0x3471U,0x68e2U,
		0xd1c4U,0xb383U,0x770dU,0xee1aU,0xcc3fU,0x8875U,0x00e1U,0x01c2U,0x0384U,0x0708U,
		0x0e10U,0x1c20U,0x3840U,0x7080U,0xe100U,0xd20bU,0xb41dU,0x7831U,0xf062U,0xf0cfU,
		0xf195U,0xf321U,0xf649U,0xfc99U,0xe939U,0xc279U,0x94f9U,0x39f9U,0x73f2U,0xe7e4U,
		0xdfc3U,0xaf8dU,0x4f11U,0x9e22U,0x2c4fU,0x589eU,0xb13cU,0x7273U,0xe4e6U,0xd9c7U,
		0xa385U,0x5701U,0xae02U,0x4c0fU,0x981eU,0x2037U,0x406eU,0x80dcU,0x11b3U,0x2366U,
		0x46ccU,0x8d98U,0x0b3bU,0x1676U,0x2cecU,0x59d8U,0xb3b0U,0x776bU,0xeed6U,0xcda7U,
		0x8b45U,0x0681U,0x0d02U,0x1a04U,0x3408U,0x6810U,0xd020U,0xb04bU,0x709dU,0xe13aU,
		0xd27fU,0xb4f5U,0x79e1U,0xf3c2U,0xf78fU,0xff15U,0xee21U,0xcc49U,0x8899U,0x0139U,
		0x0272U,0x04e4U,0x09c8U,0x1390U,0x2720U,0x4e40U,0x9c80U,0x290bU,0x5216U,0xa42cU,
		0x5853U,0xb0a6U,0x7147U,0xe28eU,0xd517U,0xba25U,0x6441U,0xc882U,0x810fU,0x1215U,
		0x242aU,0x4854U,0x90a8U,0x315bU,0x62b6U,0xc56cU,0x9ad3U,0x25adU,0x4b5aU,0x96b4U,
		0x3d63U,0x7ac6U,0xf58cU,0xfb13U,0xe62dU,0xdc51U,0xa8a9U,0x4159U,0x82b2U,0x156fU,
		0x2adeU,0x55bcU,0xab78U,0x46fbU,0x8df6U,0x0be7U,0x17ceU,0x2f9cU,0x5f38U,0xbe70U,
		0x6cebU,0xd9d6U,0xa3a7U,0x5745U,0xae8aU,0x4d1fU,0x9a3eU,0x2477U,0x48eeU,0x91dcU,
		0x33b3U,0x6766U,0xceccU,0x8d93U,0x0b2dU,0x165aU,0x2cb4U,0x5968U,0xb2d0U,0x75abU,
		0xeb56U,0xc6a7U,0x9d45U,0x2a81U,0x5502U,0xaa04U,0x4403U,0x8806U,0x0007U,0x000eU,
		0x001cU,0x0038U,0x0070U,0x00e0U,0x01c0U,0x0380U,0x0700U,0x0e00U,0x1c00U,0x3800U,
		0x7000U,0xe000U,0xd00bU,0xb01dU,0x7031U,0xe062U,0xd0cfU,0xb195U,0x7321U,0xe642U,
		0xdc8fU,0xa915U,0x4221U,0x8442U,0x188fU,0x311eU,0x623cU,0xc478U,0x98fbU,0x21fdU,
		0x43faU,0x87f4U,0x1fe3U,0x3fc6U,0x7f8cU,0xff18U,0xee3bU,0xcc7dU,0x88f1U,0x01e9U,
		0x03d2U,0x07a4U,0x0f48U,0x1e90U,0x3d20U,0x7a40U,0xf480U,0xf90bU,0xe21dU,0xd431U,
		0xb869U,0x60d9U,0xc1b2U,0x936fU,0x36d5U,0x6daaU,0xdb54U,0xa6a3U,0x5d4dU,0xba9aU,
		0x653fU,0xca7eU,0x84f7U,0x19e5U,0x33caU,0x6794U,0xcf28U,0x8e5bU,0x0cbdU,0x197aU,
		0x32f4U,0x65e8U,0xcbd0U,0x87abU,0x1f5dU,0x3ebaU,0x7d74U,0xfae8U,0xe5dbU,0xdbbdU,
		0xa771U,0x5ee9U,0xbdd2U,0x6bafU,0xd75eU,0xbeb7U,0x6d65U,0xdacaU,0xa59fU,0x5b35U,
		0xb66aU,0x7cdfU,0xf9beU,0xe377U,0xd6e5U,0xbdc1U,0x6b89U,0xd712U,0xbe2fU,0x6c55U,
		0xd8aaU,0xa15fU,0x52b5U,0xa56aU,0x5adfU,0xb5beU,0x7b77U,0xf6eeU,0xfdd7U,0xeba5U,
		0xc741U,0x9e89U,0x2d19U,0x5a32U,0xb464U,0x78c3U,0xf186U,0xf307U,0xf605U,0xfc01U,
		0xe809U,0xc019U,0x9039U,0x3079U,0x60f2U,0xc1e4U,0x93c3U,0x378dU,0x6f1aU,0xde34U,
		0xac63U,0x48cdU,0x919aU,0x333fU,0x667eU,0xccfcU,0x89f3U,0x03edU,0x07daU,0x0fb4U,
		0x1f68U,0x3ed0U,0x7da0U,0xfb40U,0xe68bU,0xdd1dU,0xaa31U,0x4469U,0x88d2U,0x01afU,
		0x035eU,0x06bcU,0x0d78U,0x1af0U,0x35e0U,0x6bc0U,0xd780U,0xbf0bU,0x6e1dU,0xdc3aU,
		0xa87fU,0x40f5U,0x81eaU,0x13dfU,0x27beU,0x4f7cU,0x9ef8U,0x2dfbU,0x5bf6U,0xb7ecU,
		0x7fd3U,0xffa6U,0xef47U,0xce85U,0x8d01U,0x0a09U,0x1412U,0x2824U,0x5048U,0xa090U,
		0x512bU,0xa256U,0x54a7U,0xa94eU,0x4297U,0x852eU,0x1a57U,0x34aeU,0x695cU,0xd2b8U,
		0xb57bU,0x7afdU,0xf5faU,0xfbffU,0xe7f5U,0xdfe1U,0xafc9U,0x4f99U,0x9f32U,0x2e6fU,
		0x5cdeU,0xb9bcU,0x6373U,0xc6e6U,0x9dc7U,0x2b85U,0x570aU,0xae14U,0x4c23U,0x9846U,
		0x2087U,0x410eU,0x821cU,0x1433U,0x2866U,0x50ccU,0xa198U,0x533bU,0xa676U,0x5ce7U,
		0xb9ceU,0x6397U,0xc72eU,0x9e57U,0x2ca5U,0x594aU,0xb294U,0x7523U,0xea46U,0xc487U,
		0x9905U,0x2201U,0x4402U,0x8804U,0x0003U,0x0006U,0x000cU,0x0018U,0x0030U,0x0060U,
		0x00c0U,0x0180U,0x0300U,0x0600U,0x0c00U,0x1800U,0x3000U,0x6000U,0xc000U,0x900bU,
		0x301dU,0x603aU,0xc074U,0x90e3U,0x31cdU,0x639aU,0xc734U,0x9e63U,0x2ccdU,0x599aU,
		0xb334U,0x7663U,0xecc6U,0xc987U,0x8305U,0x1601U,0x2c02U,0x5804U,0xb008U,0x701bU,
		0xe036U,0xd067U,0xb0c5U,0x7181U,0xe302U,0xd60fU,0xbc15U,0x6821U,0xd042U,0xb08fU,
		0x7115U,0xe22aU,0xd45fU,0xb8b5U,0x6161U,0xc2c2U,0x958fU,0x3b15U,0x762aU,0xec54U,
		0xc8a3U,0x814dU,0x1291U,0x2522U,0x4a44U,0x9488U,0x391bU,0x7236U,0xe46cU,0xd8d3U,
		0xa1adU,0x5351U,0xa6a2U,0x5d4fU,0xba9eU,0x6537U,0xca6eU,0x84d7U,0x19a5U,0x334aU,
		0x6694U,0xcd28U,0x8a5bU,0x04bdU,0x097aU,0x12f4U,0x25e8U,0x4bd0U,0x97a0U,0x3f4bU,
		0x7e96U,0xfd2cU,0xea53U,0xc4adU,0x9951U,0x22a9U,0x4552U,0x8aa4U,0x0543U,0x0a86U,
		0x150cU,0x2a18U,0x5430U,0xa860U,0x40cbU,0x8196U,0x1327U,0x264eU,0x4c9cU,0x9938U,
		0x227bU,0x44f6U,0x89ecU,0x03d3U,0x07a6U,0x0f4cU,0x1e98U,0x3d30U,0x7a60U,0xf4c0U,
		0xf98bU,0xe31dU,0xd631U,0xbc69U,0x68d9U,0xd1b2U,0xb36fU,0x76d5U,0xedaaU,0xcb5fU,
		0x86b5U,0x1d61U,0x3ac2U,0x7584U,0xeb08U,0xc61bU,0x9c3dU,0x2871U,0x50e2U,0xa1c4U,
		0x5383U,0xa706U,0x5e07U,0xbc0eU,0x6817U,0xd02eU,0xb057U,0x70a5U,0xe14aU,0xd29fU,
		0xb535U,0x7a61U,0xf4c2U,0xf98fU,0xe315U,0xd621U,0xbc49U,0x6899U,0xd132U,0xb26fU,
		0x74d5U,0xe9aaU,0xc35fU,0x96b5U,0x3d61U,0x7ac2U,0xf584U,0xfb03U,0xe60dU,0xdc11U,
		0xa829U,0x4059U,0x80b2U,0x116fU,0x22deU,0x45bcU,0x8b78U,0x06fbU,0x0df6U,0x1becU,
		0x37d8U,0x6fb0U,0xdf60U,0xaecbU,0x4d9dU,0x9b3aU,0x267fU,0x4cfeU,0x99fcU,0x23f3U,
		0x47e6U,0x8fccU,0x0f93U,0x1f26U,0x3e4cU,0x7c98U,0xf930U,0xe26bU,0xd4ddU,0xb9b1U,
		0x6369U,0xc6d2U,0x9dafU,0x2b55U,0x56aaU,0xad54U,0x4aa3U,0x9546U,0x3a87U,0x750eU,
		0xea1cU,0xc433U,0x986dU,0x20d1U,0x41a2U,0x8344U,0x1683U,0x2d06U,0x5a0cU,0xb418U,
		0x783bU,0xf076U,0xf0e7U,0xf1c5U,0xf381U,0xf709U,0xfe19U,0xec39U,0xc879U,0x80f9U,
		0x11f9U,0x23f2U,0x47e4U,0x8fc8U,0x0f9bU,0x1f36U,0x3e6cU,0x7cd8U,0xf9b0U,0xe36bU,
		0xd6ddU,0xbdb1U,0x6b69U,0xd6d2U,0xbdafU,0x6b55U,0xd6aaU,0xbd5fU,0x6ab5U,0xd56aU,
		0xbadfU,0x65b5U,0xcb6aU,0x86dfU,0x1db5U,0x3b6aU,0x76d4U,0xeda8U,0xcb5bU,0x86bdU,
		0x1d71U,0x3ae2U,0x75c4U,0xeb88U,0xc71bU,0x9e3dU,0x2c71U,0x58e2U,0xb1c4U,0x7383U,
		0xe706U,0xde07U,0xac05U,0x4801U,0x9002U,0x300fU,0x601eU,0xc03cU,0x9073U,0x30edU,
		0x61daU,0xc3b4U,0x9763U,0x3ecdU,0x7d9aU,0xfb34U,0xe663U,0xdccdU,0xa991U,0x4329U,
		0x8652U,0x1cafU,0x395eU,0x72bcU,0xe578U,0xdafbU,0xa5fdU,0x5bf1U,0xb7e2U,0x7fcfU,
		0xff9eU,0xef37U,0xce65U,0x8cc1U,0x0989U,0x1312U,0x2624U,0x4c48U,0x9890U,0x212bU,
		0x4256U,0x84acU,0x1953U,0x32a6U,0x654cU,0xca98U,0x853bU,0x1a7dU,0x34faU,0x69f4U,
		0xd3e8U,0xb7dbU,0x7fbdU,0xff7aU,0xeeffU,0xcdf5U,0x8be1U,0x07c9U,0x0f92U,0x1f24U,
		0x3e48U,0x7c90U,0xf920U,0xe24bU,0xd49dU,0xb931U,0x6269U,0xc4d2U,0x99afU,0x2355U,
		0x46aaU,0x8d54U,0x0aa3U,0x1546U,0x2a8cU,0x5518U,0xaa30U,0x446bU,0x88d6U,0x01a7U,
		0x034eU,0x069cU,0x0d38U,0x1a70U,0x34e0U,0x69c0U,0xd380U,0xb70bU,0x7e1dU,0xfc3aU,
		0xe87fU,0xc0f5U,0x91e1U,0x33c9U,0x6792U,0xcf24U,0x8e43U,0x0c8dU,0x191aU,0x3234U,
		0x6468U,0xc8d0U,0x81abU,0x135dU,0x26baU,0x4d74U,0x9ae8U,0x25dbU,0x4bb6U,0x976cU,
		0x3ed3U,0x7da6U,0xfb4cU,0xe693U,0xdd2dU,0xaa51U,0x44a9U,0x8952U,0x02afU,0x055eU,
		0x0abcU,0x1578U,0x2af0U,0x55e0U,0xabc0U,0x478bU,0x8f16U,0x0e27U,0x1c4eU,0x389cU,
		0x7138U,0xe270U,0xd4ebU,0xb9ddU,0x63b1U,0xc762U,0x9ecfU,0x2d95U,0x5b2aU,0xb654U,
		0x7ca3U,0xf946U,0xe287U,0xd505U,0xba01U,0x6409U,0xc812U,0x802fU,0x1055U,0x20aaU,
		0x4154U,0x82a8U,0x155bU,0x2ab6U,0x556cU,0xaad8U,0x45bbU,0x8b76U,0x06e7U,0x0dceU,
		0x1b9cU,0x3738U,0x6e70U,0xdce0U,0xa9cbU,0x439dU,0x873aU,0x1e7fU,0x3cfeU,0x79fcU,
		0xf3f8U,0xf7fbU,0xfffdU,0xeff1U,0xcfe9U,0x8fd9U,0x0fb9U,0x1f72U,0x3ee4U,0x7dc8U,
		0xfb90U,0xe72bU,0xde5dU,0xacb1U,0x4969U,0x92d2U,0x35afU,0x6b5eU,0xd6bcU,0xbd73U,
		0x6aedU,0xd5daU,0xbbbfU,0x6775U,0xceeaU,0x8ddfU,0x0bb5U,0x176aU,0x2ed4U,0x5da8U,
		0xbb50U,0x66abU,0xcd56U,0x8aa7U,0x0545U,0x0a8aU,0x1514U,0x2a28U,0x5450U,0xa8a0U,
		0x414bU,0x8296U,0x1527U,0x2a4eU,0x549cU,0xa938U,0x427bU,0x84f6U,0x19e7U,0x33ceU,
		0x679cU,0xcf38U,0x8e7bU,0x0cfdU,0x19faU,0x33f4U,0x67e8U,0xcfd0U,0x8fabU,0x0f5dU,
		0x1ebaU,0x3d74U,0x7ae8U,0xf5d0U,0xfbabU,0xe75dU,0xdeb1U,0xad69U,0x4ad9U,0x95b2U,
		0x3b6fU,0x76deU,0xedbcU,0xcb73U,0x86edU,0x1dd1U,0x3ba2U,0x7744U,0xee88U,0xcd1bU,
		0x8a3dU,0x0471U,0x08e2U,0x11c4U,0x2388U,0x4710U,0x8e20U,0x0c4bU,0x1896U,0x312cU,
		0x6258U,0xc4b0U,0x996bU,0x22ddU,0x45baU,0x8b74U,0x06e3U,0x0dc6U,0x1b8cU,0x3718U,
		0x6e30U,0xdc60U,0xa8cbU,0x419dU,0x833aU,0x167fU,0x2cfeU,0x59fcU,0xb3f8U,0x77fbU,
		0xeff6U,0xcfe7U,0x8fc5U,0x0f81U,0x1f02U,0x3e04U,0x7c08U,0xf810U,0xe02bU,0xd05dU,
		0xb0b1U,0x7169U,0xe2d2U,0xd5afU,0xbb55U,0x66a1U,0xcd42U,0x8a8fU,0x0515U,0x0a2aU,
		0x1454U,0x28a8U,0x5150U,0xa2a0U,0x554bU,0xaa96U,0x4527U,0x8a4eU,0x0497U,0x092eU,
		0x125cU,0x24b8U,0x4970U,0x92e0U,0x35cbU,0x6b96U,0xd72cU,0xbe53U,0x6cadU,0xd95aU,
		0xa2bfU,0x5575U,0xaaeaU,0x45dfU,0x8bbeU,0x0777U,0x0eeeU,0x1ddcU,0x3bb8U,0x7770U,
		0xeee0U,0xcdcbU,0x8b9dU,0x0731U,0x0e62U,0x1cc4U,0x3988U,0x7310U,0xe620U,0xdc4bU,
		0xa89dU,0x4131U,0x8262U,0x14cfU,0x299eU,0x533cU,0xa678U,0x5cfbU,0xb9f6U,0x63e7U,
		0xc7ceU,0x9f97U,0x2f25U,0x5e4aU,0xbc94U,0x6923U,0xd246U,0xb487U,0x7905U,0xf20aU,
		0xf41fU,0xf835U,0xe061U,0xd0c9U,0xb199U,0x7339U,0xe672U,0xdcefU,0xa9d5U,0x43a1U,
		0x8742U,0x1e8fU,0x3d1eU,0x7a3cU,0xf478U,0xf8fbU,0xe1fdU,0xd3f1U,0xb7e9U,0x7fd9U,
		0xffb2U,0xef6fU,0xced5U,0x8da1U,0x0b49U,0x1692U,0x2d24U,0x5a48U,0xb490U,0x792bU,
		0xf256U,0xf4a7U,0xf945U,0xe281U,0xd509U,0xba19U,0x6439U,0xc872U,0x80efU,0x11d5U,
		0x23aaU,0x4754U,0x8ea8U,0x0d5bU,0x1ab6U,0x356cU,0x6ad8U,0xd5b0U,0xbb6bU,0x66ddU,
		0xcdbaU,0x8b7fU,0x06f5U,0x0deaU,0x1bd4U,0x37a8U,0x6f50U,0xdea0U,0xad4bU,0x4a9dU,
		0x953aU,0x3a7fU,0x74feU,0xe9fcU,0xc3f3U,0x97edU,0x3fd1U,0x7fa2U,0xff44U,0xee83U,
		0xcd0dU,0x8a11U,0x0429U,0x0852U,0x10a4U,0x2148U,0x4290U,0x8520U,0x1a4bU,0x3496U,
		0x692cU,0xd258U,0xb4bbU,0x797dU,0xf2faU,0xf5ffU,0xfbf5U,0xe7e1U,0xdfc9U,0xaf99U,
		0x4f39U,0x9e72U,0x2cefU,0x59deU,0xb3bcU,0x7773U,0xeee6U,0xcdc7U,0x8b85U,0x0701U,
		0x0e02U,0x1c04U,0x3808U,0x7010U,0xe020U,0xd04bU,0xb09dU,0x7131U,0xe262U,0xd4cfU,
		0xb995U,0x6321U,0xc642U,0x9c8fU,0x2915U,0x522aU,0xa454U,0x58a3U,0xb146U,0x7287U,
		0xe50eU,0xda17U,0xa425U,0x5841U,0xb082U,0x710fU,0xe21eU,0xd437U,0xb865U,0x60c1U,
		0xc182U,0x930fU,0x3615U,0x6c2aU,0xd854U,0xa0a3U,0x514dU,0xa29aU,0x553fU,0xaa7eU,
		0x44f7U,0x89eeU,0x03d7U,0x07aeU,0x0f5cU,0x1eb8U,0x3d70U,0x7ae0U,0xf5c0U,0xfb8bU,
		0xe71dU,0xde31U,0xac69U,0x48d9U,0x91b2U,0x336fU,0x66deU,0xcdbcU,0x8b73U,0x06edU,
		0x0ddaU,0x1bb4U,0x3768U,0x6ed0U,0xdda0U,0xab4bU,0x469dU,0x8d3aU,0x0a7fU,0x14feU,
		0x29fcU,0x53f8U,0xa7f0U,0x5febU,0xbfd6U,0x6fa7U,0xdf4eU,0xae97U,0x4d25U,0x9a4aU,
		0x249fU,0x493eU,0x927cU,0x34f3U,0x69e6U,0xd3ccU,0xb793U,0x7f2dU,0xfe5aU,0xecbfU,
		0xc975U,0x82e1U,0x15c9U,0x2b92U,0x5724U,0xae48U,0x4c9bU,0x9936U,0x2267U,0x44ceU,
		0x899cU,0x0333U,0x0666U,0x0cccU,0x1998U,0x3330U,0x6660U,0xccc0U,0x898bU,0x031dU,
		0x063aU,0x0c74U,0x18e8U,0x31d0U,0x63a0U,0xc740U,0x9e8bU,0x2d1dU,0x5a3aU,0xb474U,
		0x78e3U,0xf1c6U,0xf387U,0xf705U,0xfe01U,0xec09U,0xc819U,0x8039U,0x1079U,0x20f2U,
		0x41e4U,0x83c8U,0x179bU,0x2f36U,0x5e6cU,0xbcd8U,0x69bbU,0xd376U,0xb6e7U,0x7dc5U,
		0xfb8aU,0xe71fU,0xde35U,0xac61U,0x48c9U,0x9192U,0x332fU,0x665eU,0xccbcU,0x8973U,
		0x02edU,0x05daU,0x0bb4U,0x1768U,0x2ed0U,0x5da0U,0xbb40U,0x668bU,0xcd16U,0x8a27U,
		0x0445U,0x088aU,0x1114U,0x2228U,0x4450U,0x88a0U,0x014bU,0x0296U,0x052cU,0x0a58U,
		0x14b0U,0x2960U,0x52c0U,0xa580U,0x5b0bU,0xb616U,0x7c27U,0xf84eU,0xe097U,0xd125U,
		0xb241U,0x7489U,0xe912U,0xc22fU,0x9455U,0x38a1U,0x7142U,0xe284U,0xd503U,0xba0dU,
		0x6411U,0xc822U,0x804fU,0x1095U,0x212aU,0x4254U,0x84a8U,0x195bU,0x32b6U,0x656cU,
		0xcad8U,0x85bbU,0x1b7dU,0x36faU,0x6df4U,0xdbe8U,0xa7dbU,0x5fbdU,0xbf7aU,0x6effU,
		0xddfeU,0xabf7U,0x47e5U,0x8fcaU,0x0f9fU,0x1f3eU,0x3e7cU,0x7cf8U,0xf9f0U,0xe3ebU,
		0xd7ddU,0xbfb1U,0x6f69U,0xded2U,0xadafU,0x4b55U,0x96aaU,0x3d5fU,0x7abeU,0xf57cU,
		0xfaf3U,0xe5edU,0xdbd1U,0xa7a9U,0x5f59U,0xbeb2U,0x6d6fU,0xdadeU,0xa5b7U,0x5b65U,
		0xb6caU,0x7d9fU,0xfb3eU,0xe677U,0xdce5U,0xa9c1U,0x4389U,0x8712U,0x1e2fU,0x3c5eU,
		0x78bcU,0xf178U,0xf2fbU,0xf5fdU,0xfbf1U,0xe7e9U,0xdfd9U,0xafb9U,0x4f79U,0x9ef2U,
		0x2defU,0x5bdeU,0xb7bcU,0x7f73U,0xfee6U,0xedc7U,0xcb85U,0x8701U,0x1e09U,0x3c12U,
		0x7824U,0xf048U,0xf09bU,0xf13dU,0xf271U,0xf4e9U,0xf9d9U,0xe3b9U,0xd779U,0xbef9U,
		0x6df9U,0xdbf2U,0xa7efU,0x5fd5U,0xbfaaU,0x6f5fU,0xdebeU,0xad77U,0x4ae5U,0x95caU,
		0x3b9fU,0x773eU,0xee7cU,0xccf3U,0x89edU,0x03d1U,0x07a2U,0x0f44U,0x1e88U,0x3d10U,
		0x7a20U,0xf440U,0xf88bU,0xe11dU,0xd231U,0xb469U,0x78d9U,0xf1b2U,0xf36fU,0xf6d5U,
		0xfda1U,0xeb49U,0xc699U,0x9d39U,0x2a79U,0x54f2U,0xa9e4U,0x43c3U,0x8786U,0x1f07U,
		0x3e0eU,0x7c1cU,0xf838U,0xe07bU,0xd0fdU,0xb1f1U,0x73e9U,0xe7d2U,0xdfafU,0xaf55U,
		0x4ea1U,0x9d42U,0x2a8fU,0x551eU,0xaa3cU,0x4473U,0x88e6U,0x01c7U,0x038eU,0x071cU,
		0x0e38U,0x1c70U,0x38e0U,0x71c0U,0xe380U,0xd70bU,0xbe1dU,0x6c31U,0xd862U,0xa0cfU,
		0x5195U,0xa32aU,0x565fU,0xacbeU,0x4977U,0x92eeU,0x35d7U,0x6baeU,0xd75cU,0xbeb3U,
		0x6d6dU,0xdadaU,0xa5bfU,0x5b75U,0xb6eaU,0x7ddfU,0xfbbeU,0xe777U,0xdee5U,0xadc1U,
		0x4b89U,0x9712U,0x3e2fU,0x7c5eU,0xf8bcU,0xe173U,0xd2edU,0xb5d1U,0x7ba9U,0xf752U,
		0xfeafU,0xed55U,0xcaa1U,0x8549U,0x1a99U,0x3532U,0x6a64U,0xd4c8U,0xb99bU,0x633dU,
		0xc67aU,0x9cffU,0x29f5U,0x53eaU,0xa7d4U,0x5fa3U,0xbf46U,0x6e87U,0xdd0eU,0xaa17U,
		0x4425U,0x884aU,0x009fU,0x013eU,0x027cU,0x04f8U,0x09f0U,0x13e0U,0x27c0U,0x4f80U,
		0x9f00U,0x2e0bU,0x5c16U,0xb82cU,0x6053U,0xc0a6U,0x9147U,0x3285U,0x650aU,0xca14U,
		0x8423U,0x184dU,0x309aU,0x6134U,0xc268U,0x94dbU,0x39bdU,0x737aU,0xe6f4U,0xdde3U,
		0xabcdU,0x4791U,0x8f22U,0x0e4fU,0x1c9eU,0x393cU,0x7278U,0xe4f0U,0xd9ebU,0xa3ddU,
		0x57b1U,0xaf62U,0x4ecfU,0x9d9eU,0x2b37U,0x566eU,0xacdcU,0x49b3U,0x9366U,0x36c7U,
		0x6d8eU,0xdb1cU,0xa633U,0x5c6dU,0xb8daU,0x61bfU,0xc37eU,0x96f7U,0x3de5U,0x7bcaU,
		0xf794U,0xff23U,0xee4dU,0xcc91U,0x8929U,0x0259U,0x04b2U,0x0964U,0x12c8U,0x2590U,
		0x4b20U,0x9640U,0x3c8bU,0x7916U,0xf22cU,0xf453U,0xf8adU,0xe151U,0xd2a9U,0xb559U,
		0x7ab9U,0xf572U,0xfaefU,0xe5d5U,0xdba1U,0xa749U,0x5e99U,0xbd32U,0x6a6fU,0xd4deU,
		0xb9b7U,0x6365U,0xc6caU,0x9d9fU,0x2b35U,0x566aU,0xacd4U,0x49a3U,0x9346U,0x3687U,
		0x6d0eU,0xda1cU,0xa433U,0x586dU,0xb0daU,0x71bfU,0xe37eU,0xd6f7U,0xbde5U,0x6bc1U,
		0xd782U,0xbf0fU,0x6e15U,0xdc2aU,0xa85fU,0x40b5U,0x816aU,0x12dfU,0x25beU,0x4b7cU,
		0x96f8U,0x3dfbU,0x7bf6U,0xf7ecU,0xffd3U,0xefadU,0xcf51U,0x8ea9U,0x0d59U,0x1ab2U,
		0x3564U,0x6ac8U,0xd590U,0xbb2bU,0x665dU,0xccbaU,0x897fU,0x02f5U,0x05eaU,0x0bd4U,
		0x17a8U,0x2f50U,0x5ea0U,0xbd40U,0x6a8bU,0xd516U,0xba27U,0x6445U,0xc88aU,0x811fU,
		0x1235U,0x246aU,0x48d4U,0x91a8U,0x335bU,0x66b6U,0xcd6cU,0x8ad3U,0x05adU,0x0b5aU,
		0x16b4U,0x2d68U,0x5ad0U,0xb5a0U,0x7b4bU,0xf696U,0xfd27U,0xea45U,0xc481U,0x9909U,
		0x2219U,0x4432U,0x8864U,0x00c3U,0x0186U,0x030cU,0x0618U,0x0c30U,0x1860U,0x30c0U,
		0x6180U,0xc300U,0x960bU,0x3c1dU,0x783aU,0xf074U,0xf0e3U,0xf1cdU,0xf391U,0xf729U,
		0xfe59U,0xecb9U,0xc979U,0x82f9U,0x15f9U,0x2bf2U,0x57e4U,0xafc8U,0x4f9bU,0x9f36U,
		0x2e67U,0x5cceU,0xb99cU,0x6333U,0xc666U,0x9cc7U,0x2985U,0x530aU,0xa614U,0x5c23U,
		0xb846U,0x6087U,0xc10eU,0x9217U,0x3425U,0x684aU,0xd094U,0xb123U,0x724dU,0xe49aU,
		0xd93fU,0xa275U,0x54e1U,0xa9c2U,0x438fU,0x871eU,0x1e37U,0x3c6eU,0x78dcU,0xf1b8U,
		0xf37bU,0xf6fdU,0xfdf1U,0xebe9U,0xc7d9U,0x9fb9U,0x2f79U,0x5ef2U,0xbde4U,0x6bc3U,
		0xd786U,0xbf07U,0x6e05U,0xdc0aU,0xa81fU,0x4035U,0x806aU,0x10dfU,0x21beU,0x437cU,
		0x86f8U,0x1dfbU,0x3bf6U,0x77ecU,0xefd8U,0xcfbbU,0x8f7dU,0x0ef1U,0x1de2U,0x3bc4U,
		0x7788U,0xef10U,0xce2bU,0x8c5dU,0x08b1U,0x1162U,0x22c4U,0x4588U,0x8b10U,0x062bU,
		0x0c56U,0x18acU,0x3158U,0x62b0U,0xc560U,0x9acbU,0x259dU,0x4b3aU,0x9674U,0x3ce3U,
		0x79c6U,0xf38cU,0xf713U,0xfe2dU,0xec51U,0xc8a9U,0x8159U,0x12b9U,0x2572U,0x4ae4U,
		0x95c8U,0x3b9bU,0x7736U,0xee6cU,0xccd3U,0x89adU,0x0351U,0x06a2U,0x0d44U,0x1a88U,
		0x3510U,0x6a20U,0xd440U,0xb88bU,0x611dU,0xc23aU,0x947fU,0x38f5U,0x71eaU,0xe3d4U,
		0xd7a3U,0xbf4dU,0x6e91U,0xdd22U,0xaa4fU,0x4495U,0x892aU,0x025fU,0x04beU,0x097cU,
		0x12f8U,0x25f0U,0x4be0U,0x97c0U,0x3f8bU,0x7f16U,0xfe2cU,0xec53U,0xc8adU,0x8151U,
		0x12a9U,0x2552U,0x4aa4U,0x9548U,0x3a9bU,0x7536U,0xea6cU,0xc4d3U,0x99adU,0x2351U,
		0x46a2U,0x8d44U,0x0a83U,0x1506U,0x2a0cU,0x5418U,0xa830U,0x406bU,0x80d6U,0x11a7U,
		0x234eU,0x469cU,0x8d38U,0x0a7bU,0x14f6U,0x29ecU,0x53d8U,0xa7b0U,0x5f6bU,0xbed6U,
		0x6da7U,0xdb4eU,0xa697U,0x5d25U,0xba4aU,0x649fU,0xc93eU,0x8277U,0x14e5U,0x29caU,
		0x5394U,0xa728U,0x5e5bU,0xbcb6U,0x6967U,0xd2ceU,0xb597U,0x7b25U,0xf64aU,0xfc9fU,
		0xe935U,0xc261U,0x94c9U,0x3999U,0x7332U,0xe664U,0xdcc3U,0xa98dU,0x4311U,0x8622U,
		0x1c4fU,0x389eU,0x713cU,0xe278U,0xd4fbU,0xb9fdU,0x63f1U,0xc7e2U,0x9fcfU,0x2f95U,
		0x5f2aU,0xbe54U,0x6ca3U,0xd946U,0xa287U,0x5505U,0xaa0aU,0x441fU,0x883eU,0x0077U,
		0x00eeU,0x01dcU,0x03b8U,0x0770U,0x0ee0U,0x1dc0U,0x3b80U,0x7700U,0xee00U,0xcc0bU,
		0x881dU,0x0031U,0x0062U,0x00c4U,0x0188U,0x0310U,0x0620U,0x0c40U,0x1880U,0x3100U,
		0x6200U,0xc400U,0x980bU,0x201dU,0x403aU,0x8074U,0x10e3U,0x21c6U,0x438cU,0x8718U,
		0x1e3bU,0x3c76U,0x78ecU,0xf1d8U,0xf3bbU,0xf77dU,0xfef1U,0xede9U,0xcbd9U,0x87b9U,
		0x1f79U,0x3ef2U,0x7de4U,0xfbc8U,0xe79bU,0xdf3dU,0xae71U,0x4ce9U,0x99d2U,0x23afU,
		0x475eU,0x8ebcU,0x0d73U,0x1ae6U,0x35ccU,0x6b98U,0xd730U,0xbe6bU,0x6cddU,0xd9baU,
		0xa37fU,0x56f5U,0xadeaU,0x4bdfU,0x97beU,0x3f77U,0x7eeeU,0xfddcU,0xebb3U,0xc76dU,
		0x9ed1U,0x2da9U,0x5b52U,0xb6a4U,0x7d43U,0xfa86U,0xe507U,0xda05U,0xa401U,0x5809U,
		0xb012U,0x702fU,0xe05eU,0xd0b7U,0xb165U,0x72c1U,0xe582U,0xdb0fU,0xa615U,0x5c21U,
		0xb842U,0x608fU,0xc11eU,0x9237U,0x3465U,0x68caU,0xd194U,0xb323U,0x764dU,0xec9aU,
		0xc93fU,0x8275U,0x14e1U,0x29c2U,0x5384U,0xa708U,0x5e1bU,0xbc36U,0x6867U,0xd0ceU,
		0xb197U,0x7325U,0xe64aU,0xdc9fU,0xa935U,0x4261U,0x84c2U,0x198fU,0x331eU,0x663cU,
		0xcc78U,0x88fbU,0x01fdU,0x03faU,0x07f4U,0x0fe8U,0x1fd0U,0x3fa0U,0x7f40U,0xfe80U,
		0xed0bU,0xca1dU,0x8431U,0x1869U,0x30d2U,0x61a4U,0xc348U,0x969bU,0x3d3dU,0x7a7aU,
		0xf4f4U,0xf9e3U,0xe3cdU,0xd791U,0xbf29U,0x6e59U,0xdcb2U,0xa96fU,0x42d5U,0x85aaU,
		0x1b5fU,0x36beU,0x6d7cU,0xdaf8U,0xa5fbU,0x5bfdU,0xb7faU,0x7fffU,0xfffeU,0xeff7U,
		0xcfe5U,0x8fc1U,0x0f89U,0x1f12U,0x3e24U,0x7c48U,0xf890U,0xe12bU,0xd25dU,0xb4b1U,
		0x7969U,0xf2d2U,0xf5afU,0xfb55U,0xe6a1U,0xdd49U,0xaa99U,0x4539U,0x8a72U,0x04efU,
		0x09deU,0x13bcU,0x2778U,0x4ef0U,0x9de0U,0x2bcbU,0x5796U,0xaf2cU,0x4e53U,0x9ca6U,
		0x2947U,0x528eU,0xa51cU,0x5a33U,0xb466U,0x78c7U,0xf18eU,0xf317U,0xf625U,0xfc41U,
		0xe889U,0xc119U,0x9239U,0x3479U,0x68f2U,0xd1e4U,0xb3c3U,0x778dU,0xef1aU,0xce3fU,
		0x8c75U,0x08e1U,0x11c2U,0x2384U,0x4708U,0x8e10U,0x0c2bU,0x1856U,0x30acU,0x6158U,
		0xc2b0U,0x956bU,0x3addU,0x75baU,0xeb74U,0xc6e3U,0x9dcdU,0x2b91U,0x5722U,0xae44U,
		0x4c83U,0x9906U,0x2207U,0x440eU,0x881cU,0x0033U,0x0066U,0x00ccU,0x0198U,0x0330U,
		0x0660U,0x0cc0U,0x1980U,0x3300U,0x6600U,0xcc00U,0x880bU,0x001dU,0x003aU,0x0074U,
		0x00e8U,0x01d0U,0x03a0U,0x0740U,0x0e80U,0x1d00U,0x3a00U,0x7400U,0xe800U,0xc00bU,
		0x901dU,0x3031U,0x6062U,0xc0c4U,0x9183U,0x330dU,0x661aU,0xcc34U,0x8863U,0x00cdU,
		0x019aU,0x0334U,0x0668U,0x0cd0U,0x19a0U,0x3340U,0x6680U,0xcd00U,0x8a0bU,0x041dU,
		0x083aU,0x1074U,0x20e8U,0x41d0U,0x83a0U,0x174bU,0x2e96U,0x5d2cU,0xba58U,0x64bbU,
		0xc976U,0x82e7U,0x15c5U,0x2b8aU,0x5714U,0xae28U,0x4c5bU,0x98b6U,0x2167U,0x42ceU,
		0x859cU,0x1b33U,0x3666U,0x6cccU,0xd998U,0xa33bU,0x567dU,0xacfaU,0x49ffU,0x93feU,
		0x37f7U,0x6feeU,0xdfdcU,0xafb3U,0x4f6dU,0x9edaU,0x2dbfU,0x5b7eU,0xb6fcU,0x7df3U,
		0xfbe6U,0xe7c7U,0xdf85U,0xaf01U,0x4e09U,0x9c12U,0x282fU,0x505eU,0xa0bcU,0x5173U,
		0xa2e6U,0x55c7U,0xab8eU,0x4717U,0x8e2eU,0x0c57U,0x18aeU,0x315cU,0x62b8U,0xc570U,
		0x9aebU,0x25ddU,0x4bbaU,0x9774U,0x3ee3U,0x7dc6U,0xfb8cU,0xe713U,0xde2dU,0xac51U,
		0x48a9U,0x9152U,0x32afU,0x655eU,0xcabcU,0x8573U,0x1aedU,0x35daU,0x6bb4U,0xd768U,
		0xbedbU,0x6dbdU,0xdb7aU,0xa6ffU,0x5df5U,0xbbeaU,0x67dfU,0xcfbeU,0x8f77U,0x0ee5U,
		0x1dcaU,0x3b94U,0x7728U,0xee50U,0xccabU,0x895dU,0x02b1U,0x0562U,0x0ac4U,0x1588U,
		0x2b10U,0x5620U,0xac40U,0x488bU,0x9116U,0x3227U,0x644eU,0xc89cU,0x8133U,0x126dU,
		0x24daU,0x49b4U,0x9368U,0x36dbU,0x6db6U,0xdb6cU,0xa6d3U,0x5dadU,0xbb5aU,0x66bfU,
		0xcd7eU,0x8af7U,0x05e5U,0x0bcaU,0x1794U,0x2f28U,0x5e50U,0xbca0U,0x694bU,0xd296U,
		0xb527U,0x7a45U,0xf48aU,0xf91fU,0xe235U,0xd461U,0xb8c9U,0x6199U,0xc332U,0x966fU,
		0x3cd5U,0x79aaU,0xf354U,0xf6a3U,0xfd4dU,0xea91U,0xc529U,0x9a59U,0x24b9U,0x4972U,
		0x92e4U,0x35c3U,0x6b86U,0xd70cU,0xbe13U,0x6c2dU,0xd85aU,0xa0bfU,0x5175U,0xa2eaU,
		0x55dfU,0xabbeU,0x4777U,0x8eeeU,0x0dd7U,0x1baeU,0x375cU,0x6eb8U,0xdd70U,0xaaebU,
		0x45ddU,0x8bbaU,0x077fU,0x0efeU,0x1dfcU,0x3bf8U,0x77f0U,0xefe0U,0xcfcbU,0x8f9dU,
		0x0f31U,0x1e62U,0x3cc4U,0x7988U,0xf310U,0xf62bU,0xfc5dU,0xe8b1U,0xc169U,0x92d9U,
		0x35b9U,0x6b72U,0xd6e4U,0xbdc3U,0x6b8dU,0xd71aU,0xbe3fU,0x6c75U,0xd8eaU,0xa1dfU,
		0x53b5U,0xa76aU,0x5edfU,0xbdbeU,0x6b77U,0xd6eeU,0xbdd7U,0x6ba5U,0xd74aU,0xbe9fU,
		0x6d35U,0xda6aU,0xa4dfU,0x59b5U,0xb36aU,0x76dfU,0xedbeU,0xcb77U,0x86e5U,0x1dc1U,
		0x3b82U,0x7704U,0xee08U,0xcc1bU,0x883dU,0x0071U,0x00e2U,0x01c4U,0x0388U,0x0710U,
		0x0e20U,0x1c40U,0x3880U,0x7100U,0xe200U,0xd40bU,0xb81dU,0x6031U,0xc062U,0x90cfU,
		0x3195U,0x632aU,0xc654U,0x9ca3U,0x294dU,0x529aU,0xa534U,0x5a63U,0xb4c6U,0x7987U,
		0xf30eU,0xf617U,0xfc25U,0xe841U,0xc089U,0x9119U,0x3239U,0x6472U,0xc8e4U,0x81c3U,
		0x138dU,0x271aU,0x4e34U,0x9c68U,0x28dbU,0x51b6U,0xa36cU,0x56d3U,0xada6U,0x4b47U,
		0x968eU,0x3d17U,0x7a2eU,0xf45cU,0xf8b3U,0xe16dU,0xd2d1U,0xb5a9U,0x7b59U,0xf6b2U,
		0xfd6fU,0xead5U,0xc5a1U,0x9b49U,0x2699U,0x4d32U,0x9a64U,0x24c3U,0x4986U,0x930cU,
		0x3613U,0x6c26U,0xd84cU,0xa093U,0x512dU,0xa25aU,0x54bfU,0xa97eU,0x42f7U,0x85eeU,
		0x1bd7U,0x37aeU,0x6f5cU,0xdeb8U,0xad7bU,0x4afdU,0x95faU,0x3bffU,0x77feU,0xeffcU,
		0xcff3U,0x8fedU,0x0fd1U,0x1fa2U,0x3f44U,0x7e88U,0xfd10U,0xea2bU,0xc45dU,0x98b1U,
		0x2169U,0x42d2U,0x85a4U,0x1b43U,0x3686U,0x6d0cU,0xda18U,0xa43bU,0x587dU,0xb0faU,
		0x71ffU,0xe3feU,0xd7f7U,0xbfe5U,0x6fc1U,0xdf82U,0xaf0fU,0x4e15U,0x9c2aU,0x285fU,
		0x50beU,0xa17cU,0x52f3U,0xa5e6U,0x5bc7U,0xb78eU,0x7f17U,0xfe2eU,0xec57U,0xc8a5U,
		0x8141U,0x1289U,0x2512U,0x4a24U,0x9448U,0x389bU,0x7136U,0xe26cU,0xd4d3U,0xb9adU,
		0x6351U,0xc6a2U,0x9d4fU,0x2a95U,0x552aU,0xaa54U,0x44a3U,0x8946U,0x0287U,0x050eU,
		0x0a1cU,0x1438U,0x2870U,0x50e0U,0xa1c0U,0x538bU,0xa716U,0x5e27U,0xbc4eU,0x6897U,
		0xd12eU,0xb257U,0x74a5U,0xe94aU,0xc29fU,0x9535U,0x3a61U,0x74c2U,0xe984U,0xc303U,
		0x960dU,0x3c11U,0x7822U,0xf044U,0xf083U,0xf10dU,0xf211U,0xf429U,0xf859U,0xe0b9U,
		0xd179U,0xb2f9U,0x75f9U,0xebf2U,0xc7efU,0x9fd5U,0x2fa1U,0x5f42U,0xbe84U,0x6d03U,
		0xda06U,0xa407U,0x5805U,0xb00aU,0x701fU,0xe03eU,0xd077U,0xb0e5U,0x71c1U,0xe382U,
		0xd70fU,0xbe15U,0x6c21U,0xd842U,0xa08fU,0x5115U,0xa22aU,0x545fU,0xa8beU,0x4177U,
		0x82eeU,0x15d7U,0x2baeU,0x575cU,0xaeb8U,0x4d7bU,0x9af6U,0x25e7U,0x4bceU,0x979cU,
		0x3f33U,0x7e66U,0xfcccU,0xe993U,0xc32dU,0x9651U,0x3ca9U,0x7952U,0xf2a4U,0xf543U,
		0xfa8dU,0xe511U,0xda29U,0xa459U,0x58b9U,0xb172U,0x72efU,0xe5deU,0xdbb7U,0xa765U,
		0x5ec1U,0xbd82U,0x6b0fU,0xd61eU,0xbc37U,0x6865U,0xd0caU,0xb19fU,0x7335U,0xe66aU,
		0xdcdfU,0xa9b5U,0x4361U,0x86c2U,0x1d8fU,0x3b1eU,0x763cU,0xec78U,0xc8fbU,0x81fdU,
		0x13f1U,0x27e2U,0x4fc4U,0x9f88U,0x2f1bU,0x5e36U,0xbc6cU,0x68d3U,0xd1a6U,0xb347U,
		0x7685U,0xed0aU,0xca1fU,0x8435U,0x1861U,0x30c2U,0x6184U,0xc308U,0x961bU,0x3c3dU,
		0x787aU,0xf0f4U,0xf1e3U,0xf3cdU,0xf791U,0xff29U,0xee59U,0xccb9U,0x8979U,0x02f9U,
		0x05f2U,0x0be4U,0x17c8U,0x2f90U,0x5f20U,0xbe40U,0x6c8bU,0xd916U,0xa227U,0x5445U,
		0xa88aU,0x411fU,0x823eU,0x1477U,0x28eeU,0x51dcU,0xa3b8U,0x577bU,0xaef6U,0x4de7U,
		0x9bceU,0x2797U,0x4f2eU,0x9e5cU,0x2cb3U,0x5966U,0xb2ccU,0x7593U,0xeb26U,0xc647U,
		0x9c85U,0x2901U,0x5202U,0xa404U,0x5803U,0xb006U,0x7007U,0xe00eU,0xd017U,0xb025U,
		0x7041U,0xe082U,0xd10fU,0xb215U,0x7421U,0xe842U,0xc08fU,0x9115U,0x3221U,0x6442U,
		0xc884U,0x8103U,0x120dU,0x241aU,0x4834U,0x9068U,0x30dbU,0x61b6U,0xc36cU,0x96d3U,
		0x3dadU,0x7b5aU,0xf6b4U,0xfd63U,0xeacdU,0xc591U,0x9b29U,0x2659U,0x4cb2U,0x9964U,
		0x22c3U,0x4586U,0x8b0cU,0x0613U,0x0c26U,0x184cU,0x3098U,0x6130U,0xc260U,0x94cbU,
		0x399dU,0x733aU,0xe674U,0xdce3U,0xa9cdU,0x4391U,0x8722U,0x1e4fU,0x3c9eU,0x793cU,
		0xf278U,0xf4fbU,0xf9fdU,0xe3f1U,0xd7e9U,0xbfd9U,0x6fb9U,0xdf72U,0xaeefU,0x4dd5U,
		0x9baaU,0x275fU,0x4ebeU,0x9d7cU,0x2af3U,0x55e6U,0xabccU,0x4793U,0x8f26U,0x0e47U,
		0x1c8eU,0x391cU,0x7238U,0xe470U,0xd8ebU,0xa1ddU,0x53b1U,0xa762U,0x5ecfU,0xbd9eU,
		0x6b37U,0xd66eU,0xbcd7U,0x69a5U,0xd34aU,0xb69fU,0x7d35U,0xfa6aU,0xe4dfU,0xd9b5U,
		0xa361U,0x56c9U,0xad92U,0x4b2fU,0x965eU,0x3cb7U,0x796eU,0xf2dcU,0xf5b3U,0xfb6dU,
		0xe6d1U,0xdda9U,0xab59U,0x46b9U,0x8d72U,0x0aefU,0x15deU,0x2bbcU,0x5778U,0xaef0U,
		0x4debU,0x9bd6U,0x27a7U,0x4f4eU,0x9e9cU,0x2d33U,0x5a66U,0xb4ccU,0x7993U,0xf326U,
		0xf647U,0xfc85U,0xe901U,0xc209U,0x9419U,0x3839U,0x7072U,0xe0e4U,0xd1c3U,0xb38dU,
		0x7711U,0xee22U,0xcc4fU,0x8895U,0x0121U,0x0242U,0x0484U,0x0908U,0x1210U,0x2420U,
		0x4840U,0x9080U,0x310bU,0x6216U,0xc42cU,0x9853U,0x20adU,0x415aU,0x82b4U,0x1563U,
		0x2ac6U,0x558cU,0xab18U,0x463bU,0x8c76U,0x08e7U,0x11ceU,0x239cU,0x4738U,0x8e70U,
		0x0cebU,0x19d6U,0x33acU,0x6758U,0xceb0U,0x8d6bU,0x0addU,0x15baU,0x2b74U,0x56e8U,
		0xadd0U,0x4babU,0x9756U,0x3ea7U,0x7d4eU,0xfa9cU,0xe533U,0xda6dU,0xa4d1U,0x59a9U,
		0xb352U,0x76afU,0xed5eU,0xcab7U,0x8565U,0x1ac1U,0x3582U,0x6b04U,0xd608U,0xbc1bU,
		0x683dU,0xd07aU,0xb0ffU,0x71f5U,0xe3eaU,0xd7dfU,0xbfb5U,0x6f61U,0xdec2U,0xad8fU,
		0x4b15U,0x962aU,0x3c5fU,0x78beU,0xf17cU,0xf2f3U,0xf5edU,0xfbd1U,0xe7a9U,0xdf59U,
		0xaeb9U,0x4d79U,0x9af2U,0x25efU,0x4bdeU,0x97bcU,0x3f73U,0x7ee6U,0xfdccU,0xeb93U,
		0xc72dU,0x9e51U,0x2ca9U,0x5952U,0xb2a4U,0x7543U,0xea86U,0xc507U,0x9a05U,0x2401U,
		0x4802U,0x9004U,0x3003U,0x6006U,0xc00cU,0x9013U,0x302dU,0x605aU,0xc0b4U,0x9163U,
		0x32cdU,0x659aU,0xcb34U,0x8663U,0x1ccdU,0x399aU,0x7334U,0xe668U,0xdcdbU,0xa9bdU,
		0x4371U,0x86e2U,0x1dcfU,0x3b9eU,0x773cU,0xee78U,0xccfbU,0x89fdU,0x03f1U,0x07e2U,
		0x0fc4U,0x1f88U,0x3f10U,0x7e20U,0xfc40U,0xe88bU,0xc11dU,0x9231U,0x3469U,0x68d2U,
		0xd1a4U,0xb343U,0x768dU,0xed1aU,0xca3fU,0x8475U,0x18e1U,0x31c2U,0x6384U,0xc708U,
		0x9e1bU,0x2c3dU,0x587aU,0xb0f4U,0x71e3U,0xe3c6U,0xd787U,0xbf05U,0x6e01U,0xdc02U,
		0xa80fU,0x4015U,0x802aU,0x105fU,0x20beU,0x417cU,0x82f8U,0x15fbU,0x2bf6U,0x57ecU,
		0xafd8U,0x4fbbU,0x9f76U,0x2ee7U,0x5dceU,0xbb9cU,0x6733U,0xce66U,0x8cc7U,0x0985U,
		0x130aU,0x2614U,0x4c28U,0x9850U,0x20abU,0x4156U,0x82acU,0x1553U,0x2aa6U,0x554cU,
		0xaa98U,0x453bU,0x8a76U,0x04e7U,0x09ceU,0x139cU,0x2738U,0x4e70U,0x9ce0U,0x29cbU,
		0x5396U,0xa72cU,0x5e53U,0xbca6U,0x6947U,0xd28eU,0xb517U,0x7a25U,0xf44aU,0xf89fU,
		0xe135U,0xd261U,0xb4c9U,0x7999U,0xf332U,0xf66fU,0xfcd5U,0xe9a1U,0xc349U,0x9699U,
		0x3d39U,0x7a72U,0xf4e4U,0xf9c3U,0xe38dU,0xd711U,0xbe29U,0x6c59U,0xd8b2U,0xa16fU,
		0x52d5U,0xa5aaU,0x5b5fU,0xb6beU,0x7d77U,0xfaeeU,0xe5d7U,0xdba5U,0xa741U,0x5e89U,
		0xbd12U,0x6a2fU,0xd45eU,0xb8b7U,0x6165U,0xc2caU,0x959fU,0x3b35U,0x766aU,0xecd4U,
		0xc9a3U,0x834dU,0x1691U,0x2d22U,0x5a44U,0xb488U,0x791bU,0xf236U,0xf467U,0xf8c5U,
		0xe181U,0xd309U,0xb619U,0x7c39U,0xf872U,0xe0efU,0xd1d5U,0xb3a1U,0x7749U,0xee92U,
		0xcd2fU,0x8a55U,0x04a1U,0x0942U,0x1284U,0x2508U,0x4a10U,0x9420U,0x384bU,0x7096U,
		0xe12cU,0xd253U,0xb4adU,0x7951U,0xf2a2U,0xf54fU,0xfa95U,0xe521U,0xda49U,0xa499U,
		0x5939U,0xb272U,0x74efU,0xe9deU,0xc3b7U,0x9765U,0x3ec1U,0x7d82U,0xfb04U,0xe603U,
		0xdc0dU,0xa811U,0x4029U,0x8052U,0x10afU,0x215eU,0x42bcU,0x8578U,0x1afbU,0x35f6U,
		0x6becU,0xd7d8U,0xbfbbU,0x6f7dU,0xdefaU,0xadffU,0x4bf5U,0x97eaU,0x3fdfU,0x7fbeU,
		0xff7cU,0xeef3U,0xcdedU,0x8bd1U,0x07a9U,0x0f52U,0x1ea4U,0x3d48U,0x7a90U,0xf520U,
		0xfa4bU,0xe49dU,0xd931U,0xa269U,0x54d9U,0xa9b2U,0x436fU,0x86deU,0x1db7U,0x3b6eU,
		0x76dcU,0xedb8U,0xcb7bU,0x86fdU,0x1df1U,0x3be2U,0x77c4U,0xef88U,0xcf1bU,0x8e3dU,
		0x0c71U,0x18e2U,0x31c4U,0x6388U,0xc710U,0x9e2bU,0x2c5dU,0x58baU,0xb174U,0x72e3U,
		0xe5c6U,0xdb87U,0xa705U,0x5e01U,0xbc02U,0x680fU,0xd01eU,0xb037U,0x7065U,0xe0caU,
		0xd19fU,0xb335U,0x7661U,0xecc2U,0xc98fU,0x8315U,0x1621U,0x2c42U,0x5884U,0xb108U,
		0x721bU,0xe436U,0xd867U,0xa0c5U,0x5181U,0xa302U,0x560fU,0xac1eU,0x4837U,0x906eU,
		0x30d7U,0x61aeU,0xc35cU,0x96b3U,0x3d6dU,0x7adaU,0xf5b4U,0xfb63U,0xe6cdU,0xdd91U,
		0xab29U,0x4659U,0x8cb2U,0x096fU,0x12deU,0x25bcU,0x4b78U,0x96f0U,0x3debU,0x7bd6U,
		0xf7acU,0xff53U,0xeeadU,0xcd51U,0x8aa9U,0x0559U,0x0ab2U,0x1564U,0x2ac8U,0x5590U,
		0xab20U,0x464bU,0x8c96U,0x0927U,0x124eU,0x249cU,0x4938U,0x9270U,0x34ebU,0x69d6U,
		0xd3acU,0xb753U,0x7eadU,0xfd5aU,0xeabfU,0xc575U,0x9ae1U,0x25c9U,0x4b92U,0x9724U,
		0x3e43U,0x7c86U,0xf90cU,0xe213U,0xd42dU,0xb851U,0x60a9U,0xc152U,0x92afU,0x3555U,
		0x6aaaU,0xd554U,0xbaa3U,0x654dU,0xca9aU,0x853fU,0x1a75U,0x34eaU,0x69d4U,0xd3a8U,
		0xb75bU,0x7ebdU,0xfd7aU,0xeaffU,0xc5f5U,0x9be1U,0x27c9U,0x4f92U,0x9f24U,0x2e43U,
		0x5c86U,0xb90cU,0x6213U,0xc426U,0x9847U,0x2085U,0x410aU,0x8214U,0x1423U,0x2846U,
		0x508cU,0xa118U,0x523bU,0xa476U,0x58e7U,0xb1ceU,0x7397U,0xe72eU,0xde57U,0xaca5U,
		0x4941U,0x9282U,0x350fU,0x6a1eU,0xd43cU,0xb873U,0x60edU,0xc1daU,0x93bfU,0x3775U,
		0x6eeaU,0xddd4U,0xaba3U,0x474dU,0x8e9aU,0x0d3fU,0x1a7eU,0x34fcU,0x69f8U,0xd3f0U,
		0xb7ebU,0x7fddU,0xffbaU,0xef7fU,0xcef5U,0x8de1U,0x0bc9U,0x1792U,0x2f24U,0x5e48U,
		0xbc90U,0x692bU,0xd256U,0xb4a7U,0x7945U,0xf28aU,0xf51fU,0xfa35U,0xe461U,0xd8c9U,
		0xa199U,0x5339U,0xa672U,0x5cefU,0xb9deU,0x63b7U,0xc76eU,0x9ed7U,0x2da5U,0x5b4aU,
		0xb694U,0x7d23U,0xfa46U,0xe487U,0xd905U,0xa201U,0x5409U,0xa812U,0x402fU,0x805eU,
		0x10b7U,0x216eU,0x42dcU,0x85b8U,0x1b7bU,0x36f6U,0x6decU,0xdbd8U,0xa7bbU,0x5f7dU,
		0xbefaU,0x6dffU,0xdbfeU,0xa7f7U,0x5fe5U,0xbfcaU,0x6f9fU,0xdf3eU,0xae77U,0x4ce5U,
		0x99caU,0x239fU,0x473eU,0x8e7cU,0x0cf3U,0x19e6U,0x33ccU,0x6798U,0xcf30U,0x8e6bU,
		0x0cddU,0x19baU,0x3374U,0x66e8U,0xcdd0U,0x8babU,0x075dU,0x0ebaU,0x1d74U,0x3ae8U,
		0x75d0U,0xeba0U,0xc74bU,0x9e9dU,0x2d31U,0x5a62U,0xb4c4U,0x7983U,0xf306U,0xf607U,
		0xfc05U,0xe801U,0xc009U,0x9019U,0x3039U,0x6072U,0xc0e4U,0x91c3U,0x338dU,0x671aU,
		0xce34U,0x8c63U,0x08cdU,0x119aU,0x2334U,0x4668U,0x8cd0U,0x09abU,0x1356U,0x26acU,
		0x4d58U,0x9ab0U,0x256bU,0x4ad6U,0x95acU,0x3b53U,0x76a6U,0xed4cU,0xca93U,0x852dU,
		0x1a51U,0x34a2U,0x6944U,0xd288U,0xb51bU,0x7a3dU,0xf47aU,0xf8ffU,0xe1f5U,0xd3e1U,
		0xb7c9U,0x7f99U,0xff32U,0xee6fU,0xccd5U,0x89a1U,0x0349U,0x0692U,0x0d24U,0x1a48U,
		0x3490U,0x6920U,0xd240U,0xb48bU,0x791dU,0xf23aU,0xf47fU,0xf8f5U,0xe1e1U,0xd3c9U,
		0xb799U,0x7f39U,0xfe72U,0xecefU,0xc9d5U,0x83a1U,0x1749U,0x2e92U,0x5d24U,0xba48U,
		0x649bU,0xc936U,0x8267U,0x14c5U,0x298aU,0x5314U,0xa628U,0x5c5bU,0xb8b6U,0x6167U,
		0xc2ceU,0x9597U,0x3b25U,0x764aU,0xec94U,0xc923U,0x824dU,0x1491U,0x2922U,0x5244U,
		0xa488U,0x591bU,0xb236U,0x7467U,0xe8ceU,0xc197U,0x9325U,0x3641U,0x6c82U,0xd904U,
		0xa203U,0x540dU,0xa81aU,0x403fU,0x807eU,0x10f7U,0x21eeU,0x43dcU,0x87b8U,0x1f7bU,
		0x3ef6U,0x7decU,0xfbd8U,0xe7bbU,0xdf7dU,0xaef1U,0x4de9U,0x9bd2U,0x27afU,0x4f5eU,
		0x9ebcU,0x2d73U,0x5ae6U,0xb5ccU,0x7b93U,0xf726U,0xfe47U,0xec85U,0xc901U,0x8209U,
		0x1419U,0x2832U,0x5064U,0xa0c8U,0x519bU,0xa336U,0x5667U,0xacceU,0x4997U,0x932eU,
		0x3657U,0x6caeU,0xd95cU,0xa2b3U,0x556dU,0xaadaU,0x45bfU,0x8b7eU,0x06f7U,0x0deeU,
		0x1bdcU,0x37b8U,0x6f70U,0xdee0U,0xadcbU,0x4b9dU,0x973aU,0x3e7fU,0x7cfeU,0xf9fcU,
		0xe3f3U,0xd7edU,0xbfd1U,0x6fa9U,0xdf52U,0xaeafU,0x4d55U,0x9aaaU,0x255fU,0x4abeU,
		0x957cU,0x3af3U,0x75e6U,0xebccU,0xc793U,0x9f2dU,0x2e51U,0x5ca2U,0xb944U,0x6283U,
		0xc506U,0x9a07U,0x2405U,0x480aU,0x9014U,0x3023U,0x6046U,0xc08cU,0x9113U,0x322dU,
		0x645aU,0xc8b4U,0x8163U,0x12cdU,0x259aU,0x4b34U,0x9668U,0x3cdbU,0x79b6U,0xf36cU,
		0xf6d3U,0xfdadU,0xeb51U,0xc6a9U,0x9d59U,0x2ab9U,0x5572U,0xaae4U,0x45c3U,0x8b86U,
		0x0707U,0x0e0eU,0x1c1cU,0x3838U,0x7070U,0xe0e0U,0xd1cbU,0xb39dU,0x7731U,0xee62U,
		0xcccfU,0x8995U,0x0321U,0x0642U,0x0c84U,0x1908U,0x3210U,0x6420U,0xc840U,0x808bU,
		0x111dU,0x223aU,0x4474U,0x88e8U,0x01dbU,0x03b6U,0x076cU,0x0ed8U,0x1db0U,0x3b60U,
		0x76c0U,0xed80U,0xcb0bU,0x861dU,0x1c31U,0x3862U,0x70c4U,0xe188U,0xd31bU,0xb63dU,
		0x7c71U,0xf8e2U,0xe1cfU,0xd395U,0xb721U,0x7e49U,0xfc92U,0xe92fU,0xc255U,0x94a1U,
		0x3949U,0x7292U,0xe524U,0xda43U,0xa48dU,0x5911U,0xb222U,0x744fU,0xe89eU,0xc137U,
		0x9265U,0x34c1U,0x6982U,0xd304U,0xb603U,0x7c0dU,0xf81aU,0xe03fU,0xd075U,0xb0e1U,
		0x71c9U,0xe392U,0xd72fU,0xbe55U,0x6ca1U,0xd942U,0xa28fU,0x5515U,0xaa2aU,0x445fU,
		0x88beU,0x0177U,0x02eeU,0x05dcU,0x0bb8U,0x1770U,0x2ee0U,0x5dc0U,0xbb80U,0x670bU,
		0xce16U,0x8c27U,0x0845U,0x108aU,0x2114U,0x4228U,0x8450U,0x18abU,0x3156U,0x62acU,
		0xc558U,0x9abbU,0x257dU,0x4afaU,0x95f4U,0x3be3U,0x77c6U,0xef8cU,0xcf13U,0x8e2dU,
		0x0c51U,0x18a2U,0x3144U,0x6288U,0xc510U,0x9a2bU,0x245dU,0x48baU,0x9174U,0x32e3U,
		0x65c6U,0xcb8cU,0x8713U,0x1e2dU,0x3c5aU,0x78b4U,0xf168U,0xf2dbU,0xf5bdU,0xfb71U,
		0xe6e9U,0xddd9U,0xabb9U,0x4779U,0x8ef2U,0x0defU,0x1bdeU,0x37bcU,0x6f78U,0xdef0U,
		0xadebU,0x4bddU,0x97baU,0x3f7fU,0x7efeU,0xfdfcU,0xebf3U,0xc7edU,0x9fd1U,0x2fa9U,
		0x5f52U,0xbea4U,0x6d43U,0xda86U,0xa507U,0x5a05U,0xb40aU,0x781fU,0xf03eU,0xf077U,
		0xf0e5U,0xf1c1U,0xf389U,0xf719U,0xfe39U,0xec79U,0xc8f9U,0x81f9U,0x13f9U,0x27f2U,
		0x4fe4U,0x9fc8U,0x2f9bU,0x5f36U,0xbe6cU,0x6cd3U,0xd9a6U,0xa347U,0x5685U,0xad0aU,
		0x4a1fU,0x943eU,0x3877U,0x70eeU,0xe1dcU,0xd3b3U,0xb76dU,0x7ed1U,0xfda2U,0xeb4fU,
		0xc695U,0x9d21U,0x2a49U,0x5492U,0xa924U,0x4243U,0x8486U,0x1907U,0x320eU,0x641cU,
		0xc838U,0x807bU,0x10fdU,0x21faU,0x43f4U,0x87e8U,0x1fdbU,0x3fb6U,0x7f6cU,0xfed8U,
		0xedbbU,0xcb7dU,0x86f1U,0x1de9U,0x3bd2U,0x77a4U,0xef48U,0xce9bU,0x8d3dU,0x0a71U,
		0x14e2U,0x29c4U,0x5388U,0xa710U,0x5e2bU,0xbc56U,0x68a7U,0xd14eU,0xb297U,0x7525U,
		0xea4aU,0xc49fU,0x9935U,0x2261U,0x44c2U,0x8984U,0x0303U,0x0606U,0x0c0cU,0x1818U,
		0x3030U,0x6060U,0xc0c0U,0x918bU,0x331dU,0x663aU,0xcc74U,0x88e3U,0x01cdU,0x039aU,
		0x0734U,0x0e68U,0x1cd0U,0x39a0U,0x7340U,0xe680U,0xdd0bU,0xaa1dU,0x4431U,0x8862U,
		0x00cfU,0x019eU,0x033cU,0x0678U,0x0cf0U,0x19e0U,0x33c0U,0x6780U,0xcf00U,0x8e0bU,
		0x0c1dU,0x183aU,0x3074U,0x60e8U,0xc1d0U,0x93abU,0x375dU,0x6ebaU,0xdd74U,0xaae3U,
		0x45cdU,0x8b9aU,0x073fU,0x0e7eU,0x1cfcU,0x39f8U,0x73f0U,0xe7e0U,0xdfcbU,0xaf9dU,
		0x4f31U,0x9e62U,0x2ccfU,0x599eU,0xb33cU,0x7673U,0xece6U,0xc9c7U,0x8385U,0x1701U,
		0x2e02U,0x5c04U,0xb808U,0x601bU,0xc036U,0x9067U,0x30c5U,0x618aU,0xc314U,0x9623U,
		0x3c4dU,0x789aU,0xf134U,0xf263U,0xf4cdU,0xf991U,0xe329U,0xd659U,0xbcb9U,0x6979U,
		0xd2f2U,0xb5efU,0x7bd5U,0xf7aaU,0xff5fU,0xeeb5U,0xcd61U,0x8ac9U,0x0599U,0x0b32U,
		0x1664U,0x2cc8U,0x5990U,0xb320U,0x764bU,0xec96U,0xc927U,0x8245U,0x1481U,0x2902U,
		0x5204U,0xa408U,0x581bU,0xb036U,0x7067U,0xe0ceU,0xd197U,0xb325U,0x7641U,0xec82U,
		0xc90fU,0x8215U,0x1421U,0x2842U,0x5084U,0xa108U,0x521bU,0xa436U,0x5867U,0xb0ceU,
		0x7197U,0xe32eU,0xd657U,0xbca5U,0x6941U,0xd282U,0xb50fU,0x7a15U,0xf42aU,0xf85fU,
		0xe0b5U,0xd161U,0xb2c9U,0x7599U,0xeb32U,0xc66fU,0x9cd5U,0x29a1U,0x5342U,0xa684U,
		0x5d03U,0xba06U,0x6407U,0xc80eU,0x8017U,0x1025U,0x204aU,0x4094U,0x8128U,0x125bU,
		0x24b6U,0x496cU,0x92d8U,0x35bbU,0x6b76U,0xd6ecU,0xbdd3U,0x6badU,0xd75aU,0xbebfU,
		0x6d75U,0xdaeaU,0xa5dfU,0x5bb5U,0xb76aU,0x7edfU,0xfdbeU,0xeb77U,0xc6e5U,0x9dc1U,
		0x2b89U,0x5712U,0xae24U,0x4c43U,0x9886U,0x2107U,0x420eU,0x841cU,0x1833U,0x3066U,
		0x60ccU,0xc198U,0x933bU,0x367dU,0x6cfaU,0xd9f4U,0xa3e3U,0x57cdU,0xaf9aU,0x4f3fU,
		0x9e7eU,0x2cf7U,0x59eeU,0xb3dcU,0x77b3U,0xef66U,0xcec7U,0x8d85U,0x0b01U,0x1602U,
		0x2c04U,0x5808U,0xb010U,0x702bU,0xe056U,0xd0a7U,0xb145U,0x7281U,0xe502U,0xda0fU,
		0xa415U,0x5821U,0xb042U,0x708fU,0xe11eU,0xd237U,0xb465U,0x78c1U,0xf182U,0xf30fU,
		0xf615U,0xfc21U,0xe849U,0xc099U,0x9139U,0x3279U,0x64f2U,0xc9e4U,0x83c3U,0x178dU,
		0x2f1aU,0x5e34U,0xbc68U,0x68dbU,0xd1b6U,0xb367U,0x76c5U,0xed8aU,0xcb1fU,0x8635U,
		0x1c61U,0x38c2U,0x7184U,0xe308U,0xd61bU,0xbc3dU,0x6871U,0xd0e2U,0xb1cfU,0x7395U,
		0xe72aU,0xde5fU,0xacb5U,0x4961U,0x92c2U,0x358fU,0x6b1eU,0xd63cU,0xbc73U,0x68edU,
		0xd1daU,0xb3bfU,0x7775U,0xeeeaU,0xcddfU,0x8bb5U,0x0761U,0x0ec2U,0x1d84U,0x3b08U,
		0x7610U,0xec20U,0xc84bU,0x809dU,0x1131U,0x2262U,0x44c4U,0x8988U,0x031bU,0x0636U,
		0x0c6cU,0x18d8U,0x31b0U,0x6360U,0xc6c0U,0x9d8bU,0x2b1dU,0x563aU,0xac74U,0x48e3U,
		0x91c6U,0x3387U,0x670eU,0xce1cU,0x8c33U,0x086dU,0x10daU,0x21b4U,0x4368U,0x86d0U,
		0x1dabU,0x3b56U,0x76acU,0xed58U,0xcabbU,0x857dU,0x1af1U,0x35e2U,0x6bc4U,0xd788U,
		0xbf1bU,0x6e3dU,0xdc7aU,0xa8ffU,0x41f5U,0x83eaU,0x17dfU,0x2fbeU,0x5f7cU,0xbef8U,
		0x6dfbU,0xdbf6U,0xa7e7U,0x5fc5U,0xbf8aU,0x6f1fU,0xde3eU,0xac77U,0x48e5U,0x91caU,
		0x339fU,0x673eU,0xce7cU,0x8cf3U,0x09edU,0x13daU,0x27b4U,0x4f68U,0x9ed0U,0x2dabU,
		0x5b56U,0xb6acU,0x7d53U,0xfaa6U,0xe547U,0xda85U,0xa501U,0x5a09U,0xb412U,0x782fU,
		0xf05eU,0xf0b7U,0xf165U,0xf2c1U,0xf589U,0xfb19U,0xe639U,0xdc79U,0xa8f9U,0x41f9U,
		0x83f2U,0x17efU,0x2fdeU,0x5fbcU,0xbf78U,0x6efbU,0xddf6U,0xabe7U,0x47c5U,0x8f8aU,
		0x0f1fU,0x1e3eU,0x3c7cU,0x78f8U,0xf1f0U,0xf3ebU,0xf7ddU,0xffb1U,0xef69U,0xced9U,
		0x8db9U,0x0b79U,0x16f2U,0x2de4U,0x5bc8U,0xb790U,0x7f2bU,0xfe56U,0xeca7U,0xc945U,
		0x8281U,0x1509U,0x2a12U,0x5424U,0xa848U,0x409bU,0x8136U,0x1267U,0x24ceU,0x499cU,
		0x9338U,0x367bU,0x6cf6U,0xd9ecU,0xa3d3U,0x57adU,0xaf5aU,0x4ebfU,0x9d7eU,0x2af7U,
		0x55eeU,0xabdcU,0x47b3U,0x8f66U,0x0ec7U,0x1d8eU,0x3b1cU,0x7638U,0xec70U,0xc8ebU,
		0x81ddU,0x13b1U,0x2762U,0x4ec4U,0x9d88U,0x2b1bU,0x5636U,0xac6cU,0x48d3U,0x91a6U,
		0x3347U,0x668eU,0xcd1cU,0x8a33U,0x046dU,0x08daU,0x11b4U,0x2368U,0x46d0U,0x8da0U,
		0x0b4bU,0x1696U,0x2d2cU,0x5a58U,0xb4b0U,0x796bU,0xf2d6U,0xf5a7U,0xfb45U,0xe681U,
		0xdd09U,0xaa19U,0x4439U,0x8872U,0x00efU,0x01deU,0x03bcU,0x0778U,0x0ef0U,0x1de0U,
		0x3bc0U,0x7780U,0xef00U,0xce0bU,0x8c1dU,0x0831U,0x1062U,0x20c4U,0x4188U,0x8310U,
		0x162bU,0x2c56U,0x58acU,0xb158U,0x72bbU,0xe576U,0xdae7U,0xa5c5U,0x5b81U,0xb702U,
		0x7e0fU,0xfc1eU,0xe837U,0xc065U,0x90c1U,0x3189U,0x6312U,0xc624U,0x9c43U,0x288dU,
		0x511aU,0xa234U,0x5463U,0xa8c6U,0x4187U,0x830eU,0x1617U,0x2c2eU,0x585cU,0xb0b8U,
		0x717bU,0xe2f6U,0xd5e7U,0xbbc5U,0x6781U,0xcf02U,0x8e0fU,0x0c15U,0x182aU,0x3054U,
		0x60a8U,0xc150U,0x92abU,0x355dU,0x6abaU,0xd574U,0xbae3U,0x65cdU,0xcb9aU,0x873fU,
		0x1e75U,0x3ceaU,0x79d4U,0xf3a8U,0xf75bU,0xfebdU,0xed71U,0xcae9U,0x85d9U,0x1bb9U,
		0x3772U,0x6ee4U,0xddc8U,0xab9bU,0x473dU,0x8e7aU,0x0cffU,0x19feU,0x33fcU,0x67f8U,
		0xcff0U,0x8febU,0x0fddU,0x1fbaU,0x3f74U,0x7ee8U,0xfdd0U,0xebabU,0xc75dU,0x9eb1U,
		0x2d69U,0x5ad2U,0xb5a4U,0x7b43U,0xf686U,0xfd07U,0xea05U,0xc401U,0x9809U,0x2019U,
		0x4032U,0x8064U,0x10c3U,0x2186U,0x430cU,0x8618U,0x1c3bU,0x3876U,0x70ecU,0xe1d8U,
		0xd3bbU,0xb77dU,0x7ef1U,0xfde2U,0xebcfU,0xc795U,0x9f21U,0x2e49U,0x5c92U,0xb924U,
		0x6243U,0xc486U,0x9907U,0x2205U,0x440aU,0x8814U,0x0023U,0x0046U,0x008cU,0x0118U,
		0x0230U,0x0460U,0x08c0U,0x1180U,0x2300U,0x4600U,0x8c00U,0x080bU,0x1016U,0x202cU,
		0x4058U,0x80b0U,0x116bU,0x22d6U,0x45acU,0x8b58U,0x06bbU,0x0d76U,0x1aecU,0x35d8U,
		0x6bb0U,0xd760U,0xbecbU,0x6d9dU,0xdb3aU,0xa67fU,0x5cf5U,0xb9eaU,0x63dfU,0xc7beU,
		0x9f77U,0x2ee5U,0x5dcaU,0xbb94U,0x6723U,0xce46U,0x8c87U,0x0905U,0x120aU,0x2414U,
		0x4828U,0x9050U,0x30abU,0x6156U,0xc2acU,0x9553U,0x3aadU,0x755aU,0xeab4U,0xc563U,
		0x9acdU,0x2591U,0x4b22U,0x9644U,0x3c83U,0x7906U,0xf20cU,0xf413U,0xf82dU,0xe051U,
		0xd0a9U,0xb159U,0x72b9U,0xe572U,0xdaefU,0xa5d5U,0x5ba1U,0xb742U,0x7e8fU,0xfd1eU,
		0xea37U,0xc465U,0x98c1U,0x2189U,0x4312U,0x8624U,0x1c43U,0x3886U,0x710cU,0xe218U,
		0xd43bU,0xb87dU,0x60f1U,0xc1e2U,0x93cfU,0x3795U,0x6f2aU,0xde54U,0xaca3U,0x494dU,
		0x929aU,0x353fU,0x6a7eU,0xd4fcU,0xb9f3U,0x63edU,0xc7daU,0x9fbfU,0x2f75U,0x5eeaU,
		0xbdd4U,0x6ba3U,0xd746U,0xbe87U,0x6d05U,0xda0aU,0xa41fU,0x5835U,0xb06aU,0x70dfU,
		0xe1beU,0xd377U,0xb6e5U,0x7dc1U,0xfb82U,0xe70fU,0xde15U,0xac21U,0x4849U,0x9092U,
		0x312fU,0x625eU,0xc4bcU,0x9973U,0x22edU,0x45daU,0x8bb4U,0x0763U,0x0ec6U,0x1d8cU,
		0x3b18U,0x7630U,0xec60U,0xc8cbU,0x819dU,0x1331U,0x2662U,0x4cc4U,0x9988U,0x231bU,
		0x4636U,0x8c6cU,0x08d3U,0x11a6U,0x234cU,0x4698U,0x8d30U,0x0a6bU,0x14d6U,0x29acU,
		0x5358U,0xa6b0U,0x5d6bU,0xbad6U,0x65a7U,0xcb4eU,0x8697U,0x1d25U,0x3a4aU,0x7494U,
		0xe928U,0xc25bU,0x94bdU,0x3971U,0x72e2U,0xe5c4U,0xdb83U,0xa70dU,0x5e11U,0xbc22U,
		0x684fU,0xd09eU,0xb137U,0x7265U,0xe4caU,0xd99fU,0xa335U,0x5661U,0xacc2U,0x498fU,
		0x931eU,0x3637U,0x6c6eU,0xd8dcU,0xa1b3U,0x536dU,0xa6daU,0x5dbfU,0xbb7eU,0x66f7U,
		0xcdeeU,0x8bd7U,0x07a5U,0x0f4aU,0x1e94U,0x3d28U,0x7a50U,0xf4a0U,0xf94bU,0xe29dU,
		0xd531U,0xba69U,0x64d9U,0xc9b2U,0x836fU,0x16d5U,0x2daaU,0x5b54U,0xb6a8U,0x7d5bU,
		0xfab6U,0xe567U,0xdac5U,0xa581U,0x5b09U,0xb612U,0x7c2fU,0xf85eU,0xe0b7U,0xd165U,
		0xb2c1U,0x7589U,0xeb12U,0xc62fU,0x9c55U,0x28a1U,0x5142U,0xa284U,0x5503U,0xaa06U,
		0x4407U,0x880eU,0x0017U,0x002eU,0x005cU,0x00b8U,0x0170U,0x02e0U,0x05c0U,0x0b80U,
		0x1700U,0x2e00U,0x5c00U,0xb800U,0x600bU,0xc016U,0x9027U,0x3045U,0x608aU,0xc114U,
		0x9223U,0x344dU,0x689aU,0xd134U,0xb263U,0x74cdU,0xe99aU,0xc33fU,0x9675U,0x3ce1U,
		0x79c2U,0xf384U,0xf703U,0xfe0dU,0xec11U,0xc829U,0x8059U,0x10b9U,0x2172U,0x42e4U,
		0x85c8U,0x1b9bU,0x3736U,0x6e6cU,0xdcd8U,0xa9bbU,0x437dU,0x86faU,0x1dffU,0x3bfeU,
		0x77fcU,0xeff8U,0xcffbU,0x8ffdU,0x0ff1U,0x1fe2U,0x3fc4U,0x7f88U,0xff10U,0xee2bU,
		0xcc5dU,0x88b1U,0x0169U,0x02d2U,0x05a4U,0x0b48U,0x1690U,0x2d20U,0x5a40U,0xb480U,
		0x790bU,0xf216U,0xf427U,0xf845U,0xe081U,0xd109U,0xb219U,0x7439U,0xe872U,0xc0efU,
		0x91d5U,0x33a1U,0x6742U,0xce84U,0x8d03U,0x0a0dU,0x141aU,0x2834U,0x5068U,0xa0d0U,
		0x51abU,0xa356U,0x56a7U,0xad4eU,0x4a97U,0x952eU,0x3a57U,0x74aeU,0xe95cU,0xc2b3U,
		0x956dU,0x3ad1U,0x75a2U,0xeb44U,0xc683U,0x9d0dU,0x2a11U,0x5422U,0xa844U,0x4083U,
		0x8106U,0x1207U,0x240eU,0x481cU,0x9038U,0x307bU,0x60f6U,0xc1ecU,0x93d3U,0x37adU,
		0x6f5aU,0xdeb4U,0xad63U,0x4acdU,0x959aU,0x3b3fU,0x767eU,0xecfcU,0xc9f3U,0x83edU,
		0x17d1U,0x2fa2U,0x5f44U,0xbe88U,0x6d1bU,0xda36U,0xa467U,0x58c5U,0xb18aU,0x731fU,
		0xe63eU,0xdc77U,0xa8e5U,0x41c1U,0x8382U,0x170fU,0x2e1eU,0x5c3cU,0xb878U,0x60fbU,
		0xc1f6U,0x93e7U,0x37c5U,0x6f8aU,0xdf14U,0xae23U,0x4c4dU,0x989aU,0x213fU,0x427eU,
		0x84fcU,0x19f3U,0x33e6U,0x67ccU,0xcf98U,0x8f3bU,0x0e7dU,0x1cfaU,0x39f4U,0x73e8U,
		0xe7d0U,0xdfabU,0xaf5dU,0x4eb1U,0x9d62U,0x2acfU,0x559eU,0xab3cU,0x4673U,0x8ce6U,
		0x09c7U,0x138eU,0x271cU,0x4e38U,0x9c70U,0x28ebU,0x51d6U,0xa3acU,0x5753U,0xaea6U,
		0x4d47U,0x9a8eU,0x2517U,0x4a2eU,0x945cU,0x38b3U,0x7166U,0xe2ccU,0xd593U,0xbb2dU,
		0x6651U,0xcca2U,0x894fU,0x0295U,0x052aU,0x0a54U,0x14a8U,0x2950U,0x52a0U,0xa540U,
		0x5a8bU,0xb516U,0x7a27U,0xf44eU,0xf897U,0xe125U,0xd241U,0xb489U,0x7919U,0xf232U,
		0xf46fU,0xf8d5U,0xe1a1U,0xd349U,0xb699U,0x7d39U,0xfa72U,0xe4efU,0xd9d5U,0xa3a1U,
		0x5749U,0xae92U,0x4d2fU,0x9a5eU,0x24b7U,0x496eU,0x92dcU,0x35b3U,0x6b66U,0xd6ccU,
		0xbd93U,0x6b2dU,0xd65aU,0xbcbfU,0x6975U,0xd2eaU,0xb5dfU,0x7bb5U,0xf76aU,0xfedfU,
		0xedb5U,0xcb61U,0x86c9U,0x1d99U,0x3b32U,0x7664U,0xecc8U,0xc99bU,0x833dU,0x1671U,
		0x2ce2U,0x59c4U,0xb388U,0x771bU,0xee36U,0xcc67U,0x88c5U,0x0181U,0x0302U,0x0604U,
		0x0c08U,0x1810U,0x3020U,0x6040U,0xc080U,0x910bU,0x321dU,0x643aU,0xc874U,0x80e3U,
		0x11cdU,0x239aU,0x4734U,0x8e68U,0x0cdbU,0x19b6U,0x336cU,0x66d8U,0xcdb0U,0x8b6bU,
		0x06ddU,0x0dbaU,0x1b74U,0x36e8U,0x6dd0U,0xdba0U,0xa74bU,0x5e9dU,0xbd3aU,0x6a7fU,
		0xd4feU,0xb9f7U,0x63e5U,0xc7caU,0x9f9fU,0x2f35U,0x5e6aU,0xbcd4U,0x69a3U,0xd346U,
		0xb687U,0x7d05U,0xfa0aU,0xe41fU,0xd835U,0xa061U,0x50c9U,0xa192U,0x532fU,0xa65eU,
		0x5cb7U,0xb96eU,0x62d7U,0xc5aeU,0x9b57U,0x26a5U,0x4d4aU,0x9a94U,0x2523U,0x4a46U,
		0x948cU,0x3913U,0x7226U,0xe44cU,0xd893U,0xa12dU,0x5251U,0xa4a2U,0x594fU,0xb29eU,
		0x7537U,0xea6eU,0xc4d7U,0x99a5U,0x2341U,0x4682U,0x8d04U,0x0a03U,0x1406U,0x280cU,
		0x5018U,0xa030U,0x506bU,0xa0d6U,0x51a7U,0xa34eU,0x5697U,0xad2eU,0x4a57U,0x94aeU,
		0x3957U,0x72aeU,0xe55cU,0xdab3U,0xa56dU,0x5ad1U,0xb5a2U,0x7b4fU,0xf69eU,0xfd37U,
		0xea65U,0xc4c1U,0x9989U,0x2319U,0x4632U,0x8c64U,0x08c3U,0x1186U,0x230cU,0x4618U,
		0x8c30U,0x086bU,0x10d6U,0x21acU,0x4358U,0x86b0U,0x1d6bU,0x3ad6U,0x75acU,0xeb58U,
		0xc6bbU,0x9d7dU,0x2af1U,0x55e2U,0xabc4U,0x4783U,0x8f06U,0x0e07U,0x1c0eU,0x381cU,
		0x7038U,0xe070U,0xd0ebU,0xb1ddU,0x73b1U,0xe762U,0xdecfU,0xad95U,0x4b21U,0x9642U,
		0x3c8fU,0x791eU,0xf23cU,0xf473U,0xf8edU,0xe1d1U,0xd3a9U,0xb759U,0x7eb9U,0xfd72U,
		0xeaefU,0xc5d5U,0x9ba1U,0x2749U,0x4e92U,0x9d24U,0x2a43U,0x5486U,0xa90cU,0x4213U,
		0x8426U,0x1847U,0x308eU,0x611cU,0xc238U,0x947bU,0x38fdU,0x71faU,0xe3f4U,0xd7e3U,
		0xbfcdU,0x6f91U,0xdf22U,0xae4fU,0x4c95U,0x992aU,0x225fU,0x44beU,0x897cU,0x02f3U,
		0x05e6U,0x0bccU,0x1798U,0x2f30U,0x5e60U,0xbcc0U,0x698bU,0xd316U,0xb627U,0x7c45U,
		0xf88aU,0xe11fU,0xd235U,0xb461U,0x78c9U,0xf192U,0xf32fU,0xf655U,0xfca1U,0xe949U,
		0xc299U,0x9539U,0x3a79U,0x74f2U,0xe9e4U,0xc3c3U,0x978dU,0x3f11U,0x7e22U,0xfc44U,
		0xe883U,0xc10dU,0x9211U,0x3429U,0x6852U,0xd0a4U,0xb143U,0x728dU,0xe51aU,0xda3fU,
		0xa475U,0x58e1U,0xb1c2U,0x738fU,0xe71eU,0xde37U,0xac65U,0x48c1U,0x9182U,0x330fU,
		0x661eU,0xcc3cU,0x8873U,0x00edU,0x01daU,0x03b4U,0x0768U,0x0ed0U,0x1da0U,0x3b40U,
		0x7680U,0xed00U,0xca0bU,0x841dU,0x1831U,0x3062U,0x60c4U,0xc188U,0x931bU,0x363dU,
		0x6c7aU,0xd8f4U,0xa1e3U,0x53cdU,0xa79aU,0x5f3fU,0xbe7eU,0x6cf7U,0xd9eeU,0xa3d7U,
		0x57a5U,0xaf4aU,0x4e9fU,0x9d3eU,0x2a77U,0x54eeU,0xa9dcU,0x43b3U,0x8766U,0x1ec7U,
		0x3d8eU,0x7b1cU,0xf638U,0xfc7bU,0xe8fdU,0xc1f1U,0x93e9U,0x37d9U,0x6fb2U,0xdf64U,
		0xaec3U,0x4d8dU,0x9b1aU,0x263fU,0x4c7eU,0x98fcU,0x21f3U,0x43e6U,0x87ccU,0x1f93U,
		0x3f26U,0x7e4cU,0xfc98U,0xe93bU,0xc27dU,0x94f1U,0x39e9U,0x73d2U,0xe7a4U,0xdf43U,
		0xae8dU,0x4d11U,0x9a22U,0x244fU,0x489eU,0x913cU,0x3273U,0x64e6U,0xc9ccU,0x8393U,
		0x172dU,0x2e5aU,0x5cb4U,0xb968U,0x62dbU,0xc5b6U,0x9b67U,0x26c5U,0x4d8aU,0x9b14U,
		0x2623U,0x4c46U,0x988cU,0x2113U,0x4226U,0x844cU,0x1893U,0x3126U,0x624cU,0xc498U,
		0x993bU,0x227dU,0x44faU,0x89f4U,0x03e3U,0x07c6U,0x0f8cU,0x1f18U,0x3e30U,0x7c60U,
		0xf8c0U,0xe18bU,0xd31dU,0xb631U,0x7c69U,0xf8d2U,0xe1afU,0xd355U,0xb6a1U,0x7d49U,
		0xfa92U,0xe52fU,0xda55U,0xa4a1U,0x5949U,0xb292U,0x752fU,0xea5eU,0xc4b7U,0x9965U,
		0x22c1U,0x4582U,0x8b04U,0x0603U,0x0c06U,0x180cU,0x3018U,0x6030U,0xc060U,0x90cbU,
		0x319dU,0x633aU,0xc674U,0x9ce3U,0x29cdU,0x539aU,0xa734U,0x5e63U,0xbcc6U,0x6987U,
		0xd30eU,0xb617U,0x7c25U,0xf84aU,0xe09fU,0xd135U,0xb261U,0x74c9U,0xe992U,0xc32fU,
		0x9655U,0x3ca1U,0x7942U,0xf284U,0xf503U,0xfa0dU,0xe411U,0xd829U,0xa059U,0x50b9U,
		0xa172U,0x52efU,0xa5deU,0x5bb7U,0xb76eU,0x7ed7U,0xfdaeU,0xeb57U,0xc6a5U,0x9d41U,
		0x2a89U,0x5512U,0xaa24U,0x4443U,0x8886U,0x0107U,0x020eU,0x041cU,0x0838U,0x1070U,
		0x20e0U,0x41c0U,0x8380U,0x170bU,0x2e16U,0x5c2cU,0xb858U,0x60bbU,0xc176U,0x92e7U,
		0x35c5U,0x6b8aU,0xd714U,0xbe23U,0x6c4dU,0xd89aU,0xa13fU,0x5275U,0xa4eaU,0x59dfU,
		0xb3beU,0x7777U,0xeeeeU,0xcdd7U,0x8ba5U,0x0741U,0x0e82U,0x1d04U,0x3a08U,0x7410U,
		0xe820U,0xc04bU,0x909dU,0x3131U,0x6262U,0xc4c4U,0x9983U,0x230dU,0x461aU,0x8c34U,
		0x0863U,0x10c6U,0x218cU,0x4318U,0x8630U,0x1c6bU,0x38d6U,0x71acU,0xe358U,0xd6bbU,
		0xbd7dU,0x6af1U,0xd5e2U,0xbbcfU,0x6795U,0xcf2aU,0x8e5fU,0x0cb5U,0x196aU,0x32d4U,
		0x65a8U,0xcb50U,0x86abU,0x1d5dU,0x3abaU,0x7574U,0xeae8U,0xc5dbU,0x9bbdU,0x2771U,
		0x4ee2U,0x9dc4U,0x2b83U,0x5706U,0xae0cU,0x4c13U,0x9826U,0x2047U,0x408eU,0x811cU,
		0x1233U,0x2466U,0x48ccU,0x9198U,0x333bU,0x6676U,0xccecU,0x89d3U,0x03adU,0x075aU,
		0x0eb4U,0x1d68U,0x3ad0U,0x75a0U,0xeb40U,0xc68bU,0x9d1dU,0x2a31U,0x5462U,0xa8c4U,
		0x4183U,0x8306U,0x1607U,0x2c0eU,0x581cU,0xb038U,0x707bU,0xe0f6U,0xd1e7U,0xb3c5U,
		0x7781U,0xef02U,0xce0fU,0x8c15U,0x0821U,0x1042U,0x2084U,0x4108U,0x8210U,0x142bU,
		0x2856U,0x50acU,0xa158U,0x52bbU,0xa576U,0x5ae7U,0xb5ceU,0x7b97U,0xf72eU,0xfe57U,
		0xeca5U,0xc941U,0x8289U,0x1519U,0x2a32U,0x5464U,0xa8c8U,0x419bU,0x8336U,0x1667U,
		0x2cceU,0x599cU,0xb338U,0x767bU,0xecf6U,0xc9e7U,0x83c5U,0x1781U,0x2f02U,0x5e04U,
		0xbc08U,0x681bU,0xd036U,0xb067U,0x70c5U,0xe18aU,0xd31fU,0xb635U,0x7c61U,0xf8c2U,
		0xe18fU,0xd315U,0xb621U,0x7c49U,0xf892U,0xe12fU,0xd255U,0xb4a1U,0x7949U,0xf292U,
		0xf52fU,0xfa55U,0xe4a1U,0xd949U,0xa299U,0x5539U,0xaa72U,0x44efU,0x89deU,0x03b7U,
		0x076eU,0x0edcU,0x1db8U,0x3b70U,0x76e0U,0xedc0U,0xcb8bU,0x871dU,0x1e31U,0x3c62U,
		0x78c4U,0xf188U,0xf31bU,0xf63dU,0xfc71U,0xe8e9U,0xc1d9U,0x93b9U,0x3779U,0x6ef2U,
		0xdde4U,0xabc3U,0x478dU,0x8f1aU,0x0e3fU,0x1c7eU,0x38fcU,0x71f8U,0xe3f0U,0xd7ebU,
		0xbfddU,0x6fb1U,0xdf62U,0xaecfU,0x4d95U,0x9b2aU,0x265fU,0x4cbeU,0x997cU,0x22f3U,
		0x45e6U,0x8bccU,0x0793U,0x0f26U,0x1e4cU,0x3c98U,0x7930U,0xf260U,0xf4cbU,0xf99dU,
		0xe331U,0xd669U,0xbcd9U,0x69b9U,0xd372U,0xb6efU,0x7dd5U,0xfbaaU,0xe75fU,0xdeb5U,
		0xad61U,0x4ac9U,0x9592U,0x3b2fU,0x765eU,0xecbcU,0xc973U,0x82edU,0x15d1U,0x2ba2U,
		0x5744U,0xae88U,0x4d1bU,0x9a36U,0x2467U,0x48ceU,0x919cU,0x3333U,0x6666U,0xccccU,
		0x8993U,0x032dU,0x065aU,0x0cb4U,0x1968U,0x32d0U,0x65a0U,0xcb40U,0x868bU,0x1d1dU,
		0x3a3aU,0x7474U,0xe8e8U,0xc1dbU,0x93bdU,0x3771U,0x6ee2U,0xddc4U,0xab83U,0x470dU,
		0x8e1aU,0x0c3fU,0x187eU,0x30fcU,0x61f8U,0xc3f0U,0x97ebU,0x3fddU,0x7fbaU,0xff74U,
		0xeee3U,0xcdcdU,0x8b91U,0x0729U,0x0e52U,0x1ca4U,0x3948U,0x7290U,0xe520U,0xda4bU,
		0xa49dU,0x5931U,0xb262U,0x74cfU,0xe99eU,0xc337U,0x9665U,0x3cc1U,0x7982U,0xf304U,
		0xf603U,0xfc0dU,0xe811U,0xc029U,0x9059U,0x30b9U,0x6172U,0xc2e4U,0x95c3U,0x3b8dU,
		0x771aU,0xee34U,0xcc63U,0x88cdU,0x0191U,0x0322U,0x0644U,0x0c88U,0x1910U,0x3220U,
		0x6440U,0xc880U,0x810bU,0x121dU,0x243aU,0x4874U,0x90e8U,0x31dbU,0x63b6U,0xc76cU,
		0x9ed3U,0x2dadU,0x5b5aU,0xb6b4U,0x7d63U,0xfac6U,0xe587U,0xdb05U,0xa601U,0x5c09U,
		0xb812U,0x602fU,0xc05eU,0x90b7U,0x3165U,0x62caU,0xc594U,0x9b23U,0x264dU,0x4c9aU,
		0x9934U,0x2263U,0x44c6U,0x898cU,0x0313U,0x0626U,0x0c4cU,0x1898U,0x3130U,0x6260U,
		0xc4c0U,0x998bU,0x231dU,0x463aU,0x8c74U,0x08e3U,0x11c6U,0x238cU,0x4718U,0x8e30U,
		0x0c6bU,0x18d6U,0x31acU,0x6358U,0xc6b0U,0x9d6bU,0x2addU,0x55baU,0xab74U,0x46e3U,
		0x8dc6U,0x0b87U,0x170eU,0x2e1cU,0x5c38U,0xb870U,0x60ebU,0xc1d6U,0x93a7U,0x3745U,
		0x6e8aU,0xdd14U,0xaa23U,0x444dU,0x889aU,0x013fU,0x027eU,0x04fcU,0x09f8U,0x13f0U,
		0x27e0U,0x4fc0U,0x9f80U,0x2f0bU,0x5e16U,0xbc2cU,0x6853U,0xd0a6U,0xb147U,0x7285U,
		0xe50aU,0xda1fU,0xa435U,0x5861U,0xb0c2U,0x718fU,0xe31eU,0xd637U,0xbc65U,0x68c1U,
		0xd182U,0xb30fU,0x7615U,0xec2aU,0xc85fU,0x80b5U,0x1161U,0x22c2U,0x4584U,0x8b08U,
		0x061bU,0x0c36U,0x186cU,0x30d8U,0x61b0U,0xc360U,0x96cbU,0x3d9dU,0x7b3aU,0xf674U,
		0xfce3U,0xe9cdU,0xc391U,0x9729U,0x3e59U,0x7cb2U,0xf964U,0xe2c3U,0xd58dU,0xbb11U,
		0x6629U,0xcc52U,0x88afU,0x0155U,0x02aaU,0x0554U,0x0aa8U,0x1550U,0x2aa0U,0x5540U,
		0xaa80U,0x450bU,0x8a16U,0x0427U,0x084eU,0x109cU,0x2138U,0x4270U,0x84e0U,0x19cbU,
		0x3396U,0x672cU,0xce58U,0x8cbbU,0x097dU,0x12faU,0x25f4U,0x4be8U,0x97d0U,0x3fabU,
		0x7f56U,0xfeacU,0xed53U,0xcaadU,0x8551U,0x1aa9U,0x3552U,0x6aa4U,0xd548U,0xba9bU,
		0x653dU,0xca7aU,0x84ffU,0x19f5U,0x33eaU,0x67d4U,0xcfa8U,0x8f5bU,0x0ebdU,0x1d7aU,
		0x3af4U,0x75e8U,0xebd0U,0xc7abU,0x9f5dU,0x2eb1U,0x5d62U,0xbac4U,0x6583U,0xcb06U,
		0x8607U,0x1c05U,0x380aU,0x7014U,0xe028U,0xd05bU,0xb0bdU,0x7171U,0xe2e2U,0xd5cfU,
		0xbb95U,0x6721U,0xce42U,0x8c8fU,0x0915U,0x122aU,0x2454U,0x48a8U,0x9150U,0x32abU,
		0x6556U,0xcaacU,0x8553U,0x1aadU,0x355aU,0x6ab4U,0xd568U,0xbadbU,0x65bdU,0xcb7aU,
		0x86ffU,0x1df5U,0x3beaU,0x77d4U,0xefa8U,0xcf5bU,0x8ebdU,0x0d71U,0x1ae2U,0x35c4U,
		0x6b88U,0xd710U,0xbe2bU,0x6c5dU,0xd8baU,0xa17fU,0x52f5U,0xa5eaU,0x5bdfU,0xb7beU,
		0x7f77U,0xfeeeU,0xedd7U,0xcba5U,0x8741U,0x1e89U,0x3d12U,0x7a24U,0xf448U,0xf89bU,
		0xe13dU,0xd271U,0xb4e9U,0x79d9U,0xf3b2U,0xf76fU,0xfed5U,0xeda1U,0xcb49U,0x8699U,
		0x1d39U,0x3a72U,0x74e4U,0xe9c8U,0xc39bU,0x973dU,0x3e71U,0x7ce2U,0xf9c4U,0xe383U,
		0xd70dU,0xbe11U,0x6c29U,0xd852U,0xa0afU,0x5155U,0xa2aaU,0x555fU,0xaabeU,0x4577U,
		0x8aeeU,0x05d7U,0x0baeU,0x175cU,0x2eb8U,0x5d70U,0xbae0U,0x65cbU,0xcb96U,0x8727U,
		0x1e45U,0x3c8aU,0x7914U,0xf228U,0xf45bU,0xf8bdU,0xe171U,0xd2e9U,0xb5d9U,0x7bb9U,
		0xf772U,0xfeefU,0xedd5U,0xcba1U,0x8749U,0x1e99U,0x3d32U,0x7a64U,0xf4c8U,0xf99bU,
		0xe33dU,0xd671U,0xbce9U,0x69d9U,0xd3b2U,0xb76fU,0x7ed5U,0xfdaaU,0xeb5fU,0xc6b5U,
		0x9d61U,0x2ac9U,0x5592U,0xab24U,0x4643U,0x8c86U,0x0907U,0x120eU,0x241cU,0x4838U,
		0x9070U,0x30ebU,0x61d6U,0xc3acU,0x9753U,0x3eadU,0x7d5aU,0xfab4U,0xe563U,0xdacdU,
		0xa591U,0x5b29U,0xb652U,0x7cafU,0xf95eU,0xe2b7U,0xd565U,0xbac1U,0x6589U,0xcb12U,
		0x862fU,0x1c55U,0x38aaU,0x7154U,0xe2a8U,0xd55bU,0xbabdU,0x6571U,0xcae2U,0x85cfU,
		0x1b95U,0x372aU,0x6e54U,0xdca8U,0xa95bU,0x42bdU,0x857aU,0x1affU,0x35feU,0x6bfcU,
		0xd7f8U,0xbffbU,0x6ffdU,0xdffaU,0xafffU,0x4ff5U,0x9feaU,0x2fdfU,0x5fbeU,0xbf7cU,
		0x6ef3U,0xdde6U,0xabc7U,0x4785U,0x8f0aU,0x0e1fU,0x1c3eU,0x387cU,0x70f8U,0xe1f0U,
		0xd3ebU,0xb7ddU,0x7fb1U,0xff62U,0xeecfU,0xcd95U,0x8b21U,0x0649U,0x0c92U,0x1924U,
		0x3248U,0x6490U,0xc920U,0x824bU,0x149dU,0x293aU,0x5274U,0xa4e8U,0x59dbU,0xb3b6U,
		0x7767U,0xeeceU,0xcd97U,0x8b25U,0x0641U,0x0c82U,0x1904U,0x3208U,0x6410U,0xc820U,
		0x804bU,0x109dU,0x213aU,0x4274U,0x84e8U,0x19dbU,0x33b6U,0x676cU,0xced8U,0x8dbbU,
		0x0b7dU,0x16faU,0x2df4U,0x5be8U,0xb7d0U,0x7fabU,0xff56U,0xeea7U,0xcd45U,0x8a81U,
		0x0509U,0x0a12U,0x1424U,0x2848U,0x5090U,0xa120U,0x524bU,0xa496U,0x5927U,0xb24eU,
		0x7497U,0xe92eU,0xc257U,0x94a5U,0x3941U,0x7282U,0xe504U,0xda03U,0xa40dU,0x5811U,
		0xb022U,0x704fU,0xe09eU,0xd137U,0xb265U,0x74c1U,0xe982U,0xc30fU,0x9615U,0x3c21U,
		0x7842U,0xf084U,0xf103U,0xf20dU,0xf411U,0xf829U,0xe059U,0xd0b9U,0xb179U,0x72f9U,
		0xe5f2U,0xdbefU,0xa7d5U,0x5fa1U,0xbf42U,0x6e8fU,0xdd1eU,0xaa37U,0x4465U,0x88caU,
		0x019fU,0x033eU,0x067cU,0x0cf8U,0x19f0U,0x33e0U,0x67c0U,0xcf80U,0x8f0bU,0x0e1dU,
		0x1c3aU,0x3874U,0x70e8U,0xe1d0U,0xd3abU,0xb75dU,0x7eb1U,0xfd62U,0xeacfU,0xc595U,
		0x9b21U,0x2649U,0x4c92U,0x9924U,0x2243U,0x4486U,0x890cU,0x0213U,0x0426U,0x084cU,
		0x1098U,0x2130U,0x4260U,0x84c0U,0x198bU,0x3316U,0x662cU,0xcc58U,0x88bbU,0x017dU,
		0x02faU,0x05f4U,0x0be8U,0x17d0U,0x2fa0U,0x5f40U,0xbe80U,0x6d0bU,0xda16U,0xa427U,
		0x5845U,0xb08aU,0x711fU,0xe23eU,0xd477U,0xb8e5U,0x61c1U,0xc382U,0x970fU,0x3e15U,
		0x7c2aU,0xf854U,0xe0a3U,0xd14dU,0xb291U,0x7529U,0xea52U,0xc4afU,0x9955U,0x22a1U,
		0x4542U,0x8a84U,0x0503U,0x0a06U,0x140cU,0x2818U,0x5030U,0xa060U,0x50cbU,0xa196U,
		0x5327U,0xa64eU,0x5c97U,0xb92eU,0x6257U,0xc4aeU,0x9957U,0x22a5U,0x454aU,0x8a94U,
		0x0523U,0x0a46U,0x148cU,0x2918U,0x5230U,0xa460U,0x58cbU,0xb196U,0x7327U,0xe64eU,
		0xdc97U,0xa925U,0x4241U,0x8482U,0x190fU,0x321eU,0x643cU,0xc878U,0x80fbU,0x11fdU,
		0x23faU,0x47f4U,0x8fe8U,0x0fdbU,0x1fb6U,0x3f6cU,0x7ed8U,0xfdb0U,0xeb6bU,0xc6ddU,
		0x9db1U,0x2b69U,0x56d2U,0xada4U,0x4b43U,0x9686U,0x3d07U,0x7a0eU,0xf41cU,0xf833U,
		0xe06dU,0xd0d1U,0xb1a9U,0x7359U,0xe6b2U,0xdd6fU,0xaad5U,0x45a1U,0x8b42U,0x068fU,
		0x0d1eU,0x1a3cU,0x3478U,0x68f0U,0xd1e0U,0xb3cbU,0x779dU,0xef3aU,0xce7fU,0x8cf5U,
		0x09e1U,0x13c2U,0x2784U,0x4f08U,0x9e10U,0x2c2bU,0x5856U,0xb0acU,0x7153U,0xe2a6U,
		0xd547U,0xba85U,0x6501U,0xca02U,0x840fU,0x1815U,0x302aU,0x6054U,0xc0a8U,0x915bU,
		0x32bdU,0x657aU,0xcaf4U,0x85e3U,0x1bcdU,0x379aU,0x6f34U,0xde68U,0xacdbU,0x49bdU,
		0x937aU,0x36ffU,0x6dfeU,0xdbfcU,0xa7f3U,0x5fedU,0xbfdaU,0x6fbfU,0xdf7eU,0xaef7U,
		0x4de5U,0x9bcaU,0x279fU,0x4f3eU,0x9e7cU,0x2cf3U,0x59e6U,0xb3ccU,0x7793U,0xef26U,
		0xce47U,0x8c85U,0x0901U,0x1202U,0x2404U,0x4808U,0x9010U,0x302bU,0x6056U,0xc0acU,
		0x9153U,0x32adU,0x655aU,0xcab4U,0x8563U,0x1acdU,0x359aU,0x6b34U,0xd668U,0xbcdbU,
		0x69bdU,0xd37aU,0xb6ffU,0x7df5U,0xfbeaU,0xe7dfU,0xdfb5U,0xaf61U,0x4ec9U,0x9d92U,
		0x2b2fU,0x565eU,0xacbcU,0x4973U,0x92e6U,0x35c7U,0x6b8eU,0xd71cU,0xbe33U,0x6c6dU,
		0xd8daU,0xa1bfU,0x5375U,0xa6eaU,0x5ddfU,0xbbbeU,0x6777U,0xceeeU,0x8dd7U,0x0ba5U,
		0x174aU,0x2e94U,0x5d28U,0xba50U,0x64abU,0xc956U,0x82a7U,0x1545U,0x2a8aU,0x5514U,
		0xaa28U,0x445bU,0x88b6U,0x0167U,0x02ceU,0x059cU,0x0b38U,0x1670U,0x2ce0U,0x59c0U,
		0xb380U,0x770bU,0xee16U,0xcc27U,0x8845U,0x0081U,0x0102U,0x0204U,0x0408U,0x0810U,
		0x1020U,0x2040U,0x4080U,0x8100U,0x120bU,0x2416U,0x482cU,0x9058U,0x30bbU,0x6176U,
		0xc2ecU,0x95d3U,0x3badU,0x775aU,0xeeb4U,0xcd63U,0x8acdU,0x0591U,0x0b22U,0x1644U,
		0x2c88U,0x5910U,0xb220U,0x744bU,0xe896U,0xc127U,0x9245U,0x3481U,0x6902U,0xd204U,
		0xb403U,0x780dU,0xf01aU,0xf03fU,0xf075U,0xf0e1U,0xf1c9U,0xf399U,0xf739U,0xfe79U,
		0xecf9U,0xc9f9U,0x83f9U,0x17f9U,0x2ff2U,0x5fe4U,0xbfc8U,0x6f9bU,0xdf36U,0xae67U,
		0x4cc5U,0x998aU,0x231fU,0x463eU,0x8c7cU,0x08f3U,0x11e6U,0x23ccU,0x4798U,0x8f30U,
		0x0e6bU,0x1cd6U,0x39acU,0x7358U,0xe6b0U,0xdd6bU,0xaaddU,0x45b1U,0x8b62U,0x06cfU,
		0x0d9eU,0x1b3cU,0x3678U,0x6cf0U,0xd9e0U,0xa3cbU,0x579dU,0xaf3aU,0x4e7fU,0x9cfeU,
		0x29f7U,0x53eeU,0xa7dcU,0x5fb3U,0xbf66U,0x6ec7U,0xdd8eU,0xab17U,0x4625U,0x8c4aU,
		0x089fU,0x113eU,0x227cU,0x44f8U,0x89f0U,0x03ebU,0x07d6U,0x0facU,0x1f58U,0x3eb0U,
		0x7d60U,0xfac0U,0xe58bU,0xdb1dU,0xa631U,0x5c69U,0xb8d2U,0x61afU,0xc35eU,0x96b7U,
		0x3d65U,0x7acaU,0xf594U,0xfb23U,0xe64dU,0xdc91U,0xa929U,0x4259U,0x84b2U,0x196fU,
		0x32deU,0x65bcU,0xcb78U,0x86fbU,0x1dfdU,0x3bfaU,0x77f4U,0xefe8U,0xcfdbU,0x8fbdU,
		0x0f71U,0x1ee2U,0x3dc4U,0x7b88U,0xf710U,0xfe2bU,0xec5dU,0xc8b1U,0x8169U,0x12d9U,
		0x25b2U,0x4b64U,0x96c8U,0x3d9bU,0x7b36U,0xf66cU,0xfcd3U,0xe9adU,0xc351U,0x96a9U,
		0x3d59U,0x7ab2U,0xf564U,0xfac3U,0xe58dU,0xdb11U,0xa629U,0x5c59U,0xb8b2U,0x616fU,
		0xc2deU,0x95b7U,0x3b65U,0x76caU,0xed94U,0xcb23U,0x864dU,0x1c91U,0x3922U,0x7244U,
		0xe488U,0xd91bU,0xa23dU,0x5471U,0xa8e2U,0x41cfU,0x839eU,0x1737U,0x2e6eU,0x5cdcU,
		0xb9b8U,0x637bU,0xc6f6U,0x9de7U,0x2bc5U,0x578aU,0xaf14U,0x4e23U,0x9c46U,0x2887U,
		0x510eU,0xa21cU,0x5433U,0xa866U,0x40c7U,0x818eU,0x1317U,0x262eU,0x4c5cU,0x98b8U,
		0x217bU,0x42f6U,0x85ecU,0x1bd3U,0x37a6U,0x6f4cU,0xde98U,0xad3bU,0x4a7dU,0x94faU,
		0x39ffU,0x73feU,0xe7fcU,0xdff3U,0xafedU,0x4fd1U,0x9fa2U,0x2f4fU,0x5e9eU,0xbd3cU,
		0x6a73U,0xd4e6U,0xb9c7U,0x6385U,0xc70aU,0x9e1fU,0x2c35U,0x586aU,0xb0d4U,0x71a3U,
		0xe346U,0xd687U,0xbd05U,0x6a01U,0xd402U,0xb80fU,0x6015U,0xc02aU,0x905fU,0x30b5U,
		0x616aU,0xc2d4U,0x95a3U,0x3b4dU,0x769aU,0xed34U,0xca63U,0x84cdU,0x1991U,0x3322U,
		0x6644U,0xcc88U,0x891bU,0x023dU,0x047aU,0x08f4U,0x11e8U,0x23d0U,0x47a0U,0x8f40U,
		0x0e8bU,0x1d16U,0x3a2cU,0x7458U,0xe8b0U,0xc16bU,0x92ddU,0x35b1U,0x6b62U,0xd6c4U,
		0xbd83U,0x6b0dU,0xd61aU,0xbc3fU,0x6875U,0xd0eaU,0xb1dfU,0x73b5U,0xe76aU,0xdedfU,
		0xadb5U,0x4b61U,0x96c2U,0x3d8fU,0x7b1eU,0xf63cU,0xfc73U,0xe8edU,0xc1d1U,0x93a9U,
		0x3759U,0x6eb2U,0xdd64U,0xaac3U,0x458dU,0x8b1aU,0x063fU,0x0c7eU,0x18fcU,0x31f8U,
		0x63f0U,0xc7e0U,0x9fcbU,0x2f9dU,0x5f3aU,0xbe74U,0x6ce3U,0xd9c6U,0xa387U,0x5705U,
		0xae0aU,0x4c1fU,0x983eU,0x2077U,0x40eeU,0x81dcU,0x13b3U,0x2766U,0x4eccU,0x9d98U,
		0x2b3bU,0x5676U,0xacecU,0x49d3U,0x93a6U,0x3747U,0x6e8eU,0xdd1cU,0xaa33U,0x446dU,
		0x88daU,0x01bfU,0x037eU,0x06fcU,0x0df8U,0x1bf0U,0x37e0U,0x6fc0U,0xdf80U,0xaf0bU,
		0x4e1dU,0x9c3aU,0x287fU,0x50feU,0xa1fcU,0x53f3U,0xa7e6U,0x5fc7U,0xbf8eU,0x6f17U,
		0xde2eU,0xac57U,0x48a5U,0x914aU,0x329fU,0x653eU,0xca7cU,0x84f3U,0x19edU,0x33daU,
		0x67b4U,0xcf68U,0x8edbU,0x0dbdU,0x1b7aU,0x36f4U,0x6de8U,0xdbd0U,0xa7abU,0x5f5dU,
		0xbebaU,0x6d7fU,0xdafeU,0xa5f7U,0x5be5U,0xb7caU,0x7f9fU,0xff3eU,0xee77U,0xcce5U,
		0x89c1U,0x0389U,0x0712U,0x0e24U,0x1c48U,0x3890U,0x7120U,0xe240U,0xd48bU,0xb91dU,
		0x6231U,0xc462U,0x98cfU,0x2195U,0x432aU,0x8654U,0x1ca3U,0x3946U,0x728cU,0xe518U,
		0xda3bU,0xa47dU,0x58f1U,0xb1e2U,0x73cfU,0xe79eU,0xdf37U,0xae65U,0x4cc1U,0x9982U,
		0x230fU,0x461eU,0x8c3cU,0x0873U,0x10e6U,0x21ccU,0x4398U,0x8730U,0x1e6bU,0x3cd6U,
		0x79acU,0xf358U,0xf6bbU,0xfd7dU,0xeaf1U,0xc5e9U,0x9bd9U,0x27b9U,0x4f72U,0x9ee4U,
		0x2dc3U,0x5b86U,0xb70cU,0x7e13U,0xfc26U,0xe847U,0xc085U,0x9101U,0x3209U,0x6412U,
		0xc824U,0x8043U,0x108dU,0x211aU,0x4234U,0x8468U,0x18dbU,0x31b6U,0x636cU,0xc6d8U,
		0x9dbbU,0x2b7dU,0x56faU,0xadf4U,0x4be3U,0x97c6U,0x3f87U,0x7f0eU,0xfe1cU,0xec33U,
		0xc86dU,0x80d1U,0x11a9U,0x2352U,0x46a4U,0x8d48U,0x0a9bU,0x1536U,0x2a6cU,0x54d8U,
		0xa9b0U,0x436bU,0x86d6U,0x1da7U,0x3b4eU,0x769cU,0xed38U,0xca7bU,0x84fdU,0x19f1U,
		0x33e2U,0x67c4U,0xcf88U,0x8f1bU,0x0e3dU,0x1c7aU,0x38f4U,0x71e8U,0xe3d0U,0xd7abU,
		0xbf5dU,0x6eb1U,0xdd62U,0xaacfU,0x4595U,0x8b2aU,0x065fU,0x0cbeU,0x197cU,0x32f8U,
		0x65f0U,0xcbe0U,0x87cbU,0x1f9dU,0x3f3aU,0x7e74U,0xfce8U,0xe9dbU,0xc3bdU,0x9771U,
		0x3ee9U,0x7dd2U,0xfba4U,0xe743U,0xde8dU,0xad11U,0x4a29U,0x9452U,0x38afU,0x715eU,
		0xe2bcU,0xd573U,0xbaedU,0x65d1U,0xcba2U,0x874fU,0x1e95U,0x3d2aU,0x7a54U,0xf4a8U,
		0xf95bU,0xe2bdU,0xd571U,0xbae9U,0x65d9U,0xcbb2U,0x876fU,0x1ed5U,0x3daaU,0x7b54U,
		0xf6a8U,0xfd5bU,0xeabdU,0xc571U,0x9ae9U,0x25d9U,0x4bb2U,0x9764U,0x3ec3U,0x7d86U,
		0xfb0cU,0xe613U,0xdc2dU,0xa851U,0x40a9U,0x8152U,0x12afU,0x255eU,0x4abcU,0x9578U,
		0x3afbU,0x75f6U,0xebecU,0xc7d3U,0x9fadU,0x2f51U,0x5ea2U,0xbd44U,0x6a83U,0xd506U,
		0xba07U,0x6405U,0xc80aU,0x801fU,0x1035U,0x206aU,0x40d4U,0x81a8U,0x135bU,0x26b6U,
		0x4d6cU,0x9ad8U,0x25bbU,0x4b76U,0x96ecU,0x3dd3U,0x7ba6U,0xf74cU,0xfe93U,0xed2dU,
		0xca51U,0x84a9U,0x1959U,0x32b2U,0x6564U,0xcac8U,0x859bU,0x1b3dU,0x367aU,0x6cf4U,
		0xd9e8U,0xa3dbU,0x57bdU,0xaf7aU,0x4effU,0x9dfeU,0x2bf7U,0x57eeU,0xafdcU,0x4fb3U,
		0x9f66U,0x2ec7U,0x5d8eU,0xbb1cU,0x6633U,0xcc66U,0x88c7U,0x0185U,0x030aU,0x0614U,
		0x0c28U,0x1850U,0x30a0U,0x6140U,0xc280U,0x950bU,0x3a1dU,0x743aU,0xe874U,0xc0e3U,
		0x91cdU,0x3391U,0x6722U,0xce44U,0x8c83U,0x090dU,0x121aU,0x2434U,0x4868U,0x90d0U,
		0x31abU,0x6356U,0xc6acU,0x9d53U,0x2aadU,0x555aU,0xaab4U,0x4563U,0x8ac6U,0x0587U,
		0x0b0eU,0x161cU,0x2c38U,0x5870U,0xb0e0U,0x71cbU,0xe396U,0xd727U,0xbe45U,0x6c81U,
		0xd902U,0xa20fU,0x5415U,0xa82aU,0x405fU,0x80beU,0x1177U,0x22eeU,0x45dcU,0x8bb8U,
		0x077bU,0x0ef6U,0x1decU,0x3bd8U,0x77b0U,0xef60U,0xcecbU,0x8d9dU,0x0b31U,0x1662U,
		0x2cc4U,0x5988U,0xb310U,0x762bU,0xec56U,0xc8a7U,0x8145U,0x1281U,0x2502U,0x4a04U,
		0x9408U,0x381bU,0x7036U,0xe06cU,0xd0d3U,0xb1adU,0x7351U,0xe6a2U,0xdd4fU,0xaa95U,
		0x4521U,0x8a42U,0x048fU,0x091eU,0x123cU,0x2478U,0x48f0U,0x91e0U,0x33cbU,0x6796U,
		0xcf2cU,0x8e53U,0x0cadU,0x195aU,0x32b4U,0x6568U,0xcad0U,0x85abU,0x1b5dU,0x36baU,
		0x6d74U,0xdae8U,0xa5dbU,0x5bbdU,0xb77aU,0x7effU,0xfdfeU,0xebf7U,0xc7e5U,0x9fc1U,
		0x2f89U,0x5f12U,0xbe24U,0x6c43U,0xd886U,0xa107U,0x5205U,0xa40aU,0x581fU,0xb03eU,
		0x7077U,0xe0eeU,0xd1d7U,0xb3a5U,0x7741U,0xee82U,0xcd0fU,0x8a15U,0x0421U,0x0842U,
		0x1084U,0x2108U,0x4210U,0x8420U,0x184bU,0x3096U,0x612cU,0xc258U,0x94bbU,0x397dU,
		0x72faU,0xe5f4U,0xdbe3U,0xa7cdU,0x5f91U,0xbf22U,0x6e4fU,0xdc9eU,0xa937U,0x4265U,
		0x84caU,0x199fU,0x333eU,0x667cU,0xccf8U,0x89fbU,0x03fdU,0x07faU,0x0ff4U,0x1fe8U,
		0x3fd0U,0x7fa0U,0xff40U,0xee8bU,0xcd1dU,0x8a31U,0x0469U,0x08d2U,0x11a4U,0x2348U,
		0x4690U,0x8d20U,0x0a4bU,0x1496U,0x292cU,0x5258U,0xa4b0U,0x596bU,0xb2d6U,0x75a7U,
		0xeb4eU,0xc697U,0x9d25U,0x2a41U,0x5482U,0xa904U,0x4203U,0x8406U,0x1807U,0x300eU,
		0x601cU,0xc038U,0x907bU,0x30fdU,0x61faU,0xc3f4U,0x97e3U,0x3fcdU,0x7f9aU,0xff34U,
		0xee63U,0xcccdU,0x8991U,0x0329U,0x0652U,0x0ca4U,0x1948U,0x3290U,0x6520U,0xca40U,
		0x848bU,0x191dU,0x323aU,0x6474U,0xc8e8U,0x81dbU,0x13bdU,0x277aU,0x4ef4U,0x9de8U,
		0x2bdbU,0x57b6U,0xaf6cU,0x4ed3U,0x9da6U,0x2b47U,0x568eU,0xad1cU,0x4a33U,0x9466U,
		0x38c7U,0x718eU,0xe31cU,0xd633U,0xbc6dU,0x68d1U,0xd1a2U,0xb34fU,0x7695U,0xed2aU,
		0xca5fU,0x84b5U,0x1961U,0x32c2U,0x6584U,0xcb08U,0x861bU,0x1c3dU,0x387aU,0x70f4U,
		0xe1e8U,0xd3dbU,0xb7bdU,0x7f71U,0xfee2U,0xedcfU,0xcb95U,0x8721U,0x1e49U,0x3c92U,
		0x7924U,0xf248U,0xf49bU,0xf93dU,0xe271U,0xd4e9U,0xb9d9U,0x63b9U,0xc772U,0x9eefU,
		0x2dd5U,0x5baaU,0xb754U,0x7ea3U,0xfd46U,0xea87U,0xc505U,0x9a01U,0x2409U,0x4812U,
		0x9024U,0x3043U,0x6086U,0xc10cU,0x9213U,0x342dU,0x685aU,0xd0b4U,0xb163U,0x72cdU,
		0xe59aU,0xdb3fU,0xa675U,0x5ce1U,0xb9c2U,0x638fU,0xc71eU,0x9e37U,0x2c65U,0x58caU,
		0xb194U,0x7323U,0xe646U,0xdc87U,0xa905U,0x4201U,0x8402U,0x180fU,0x301eU,0x603cU,
		0xc078U,0x90fbU,0x31fdU,0x63faU,0xc7f4U,0x9fe3U,0x2fcdU,0x5f9aU,0xbf34U,0x6e63U,
		0xdcc6U,0xa987U,0x4305U,0x860aU,0x1c1fU,0x383eU,0x707cU,0xe0f8U,0xd1fbU,0xb3fdU,
		0x77f1U,0xefe2U,0xcfcfU,0x8f95U,0x0f21U,0x1e42U,0x3c84U,0x7908U,0xf210U,0xf42bU,
		0xf85dU,0xe0b1U,0xd169U,0xb2d9U,0x75b9U,0xeb72U,0xc6efU,0x9dd5U,0x2ba1U,0x5742U,
		0xae84U,0x4d03U,0x9a06U,0x2407U,0x480eU,0x901cU,0x3033U,0x6066U,0xc0ccU,0x9193U,
		0x332dU,0x665aU,0xccb4U,0x8963U,0x02cdU,0x059aU,0x0b34U,0x1668U,0x2cd0U,0x59a0U,
		0xb340U,0x768bU,0xed16U,0xca27U,0x8445U,0x1881U,0x3102U,0x6204U,0xc408U,0x981bU,
		0x203dU,0x407aU,0x80f4U,0x11e3U,0x23c6U,0x478cU,0x8f18U,0x0e3bU,0x1c76U,0x38ecU,
		0x71d8U,0xe3b0U,0xd76bU,0xbeddU,0x6db1U,0xdb62U,0xa6cfU,0x5d95U,0xbb2aU,0x665fU,
		0xccbeU,0x8977U,0x02e5U,0x05caU,0x0b94U,0x1728U,0x2e50U,0x5ca0U,0xb940U,0x628bU,
		0xc516U,0x9a27U,0x2445U,0x488aU,0x9114U,0x3223U,0x6446U,0xc88cU,0x8113U,0x122dU,
		0x245aU,0x48b4U,0x9168U,0x32dbU,0x65b6U,0xcb6cU,0x86d3U,0x1dadU,0x3b5aU,0x76b4U,
		0xed68U,0xcadbU,0x85bdU,0x1b71U,0x36e2U,0x6dc4U,0xdb88U,0xa71bU,0x5e3dU,0xbc7aU,
		0x68ffU,0xd1feU,0xb3f7U,0x77e5U,0xefcaU,0xcf9fU,0x8f35U,0x0e61U,0x1cc2U,0x3984U,
		0x7308U,0xe610U,0xdc2bU,0xa85dU,0x40b1U,0x8162U,0x12cfU,0x259eU,0x4b3cU,0x9678U,
		0x3cfbU,0x79f6U,0xf3ecU,0xf7d3U,0xffadU,0xef51U,0xcea9U,0x8d59U,0x0ab9U,0x1572U,
		0x2ae4U,0x55c8U,0xab90U,0x472bU,0x8e56U,0x0ca7U,0x194eU,0x329cU,0x6538U,0xca70U,
		0x84ebU,0x19ddU,0x33baU,0x6774U,0xcee8U,0x8ddbU,0x0bbdU,0x177aU,0x2ef4U,0x5de8U,
		0xbbd0U,0x67abU,0xcf56U,0x8ea7U,0x0d45U,0x1a8aU,0x3514U,0x6a28U,0xd450U,0xb8abU,
		0x615dU,0xc2baU,0x957fU,0x3af5U,0x75eaU,0xebd4U,0xc7a3U,0x9f4dU,0x2e91U,0x5d22U,
		0xba44U,0x6483U,0xc906U,0x8207U,0x1405U,0x280aU,0x5014U,0xa028U,0x505bU,0xa0b6U,
		0x5167U,0xa2ceU,0x5597U,0xab2eU,0x4657U,0x8caeU,0x0957U,0x12aeU,0x255cU,0x4ab8U,
		0x9570U,0x3aebU,0x75d6U,0xebacU,0xc753U,0x9eadU,0x2d51U,0x5aa2U,0xb544U,0x7a83U,
		0xf506U,0xfa07U,0xe405U,0xd801U,0xa009U,0x5019U,0xa032U,0x506fU,0xa0deU,0x51b7U,
		0xa36eU,0x56d7U,0xadaeU,0x4b57U,0x96aeU,0x3d57U,0x7aaeU,0xf55cU,0xfab3U,0xe56dU,
		0xdad1U,0xa5a9U,0x5b59U,0xb6b2U,0x7d6fU,0xfadeU,0xe5b7U,0xdb65U,0xa6c1U,0x5d89U,
		0xbb12U,0x662fU,0xcc5eU,0x88b7U,0x0165U,0x02caU,0x0594U,0x0b28U,0x1650U,0x2ca0U,
		0x5940U,0xb280U,0x750bU,0xea16U,0xc427U,0x9845U,0x2081U,0x4102U,0x8204U,0x1403U,
		0x2806U,0x500cU,0xa018U,0x503bU,0xa076U,0x50e7U,0xa1ceU,0x5397U,0xa72eU,0x5e57U,
		0xbcaeU,0x6957U,0xd2aeU,0xb557U,0x7aa5U,0xf54aU,0xfa9fU,0xe535U,0xda61U,0xa4c9U,
		0x5999U,0xb332U,0x766fU,0xecdeU,0xc9b7U,0x8365U,0x16c1U,0x2d82U,0x5b04U,0xb608U,
		0x7c1bU,0xf836U,0xe067U,0xd0c5U,0xb181U,0x7309U,0xe612U,0xdc2fU,0xa855U,0x40a1U,
		0x8142U,0x128fU,0x251eU,0x4a3cU,0x9478U,0x38fbU,0x71f6U,0xe3ecU,0xd7d3U,0xbfadU,
		0x6f51U,0xdea2U,0xad4fU,0x4a95U,0x952aU,0x3a5fU,0x74beU,0xe97cU,0xc2f3U,0x95edU,
		0x3bd1U,0x77a2U,0xef44U,0xce83U,0x8d0dU,0x0a11U,0x1422U,0x2844U,0x5088U,0xa110U,
		0x522bU,0xa456U,0x58a7U,0xb14eU,0x7297U,0xe52eU,0xda57U,0xa4a5U,0x5941U,0xb282U,
		0x750fU,0xea1eU,0xc437U,0x9865U,0x20c1U,0x4182U,0x8304U,0x1603U,0x2c06U,0x580cU,
		0xb018U,0x703bU,0xe076U,0xd0e7U,0xb1c5U,0x7381U,0xe702U,0xde0fU,0xac15U,0x4821U,
		0x9042U,0x308fU,0x611eU,0xc23cU,0x9473U,0x38edU,0x71daU,0xe3b4U,0xd763U,0xbecdU,
		0x6d91U,0xdb22U,0xa64fU,0x5c95U,0xb92aU,0x625fU,0xc4beU,0x9977U,0x22e5U,0x45caU,
		0x8b94U,0x0723U,0x0e46U,0x1c8cU,0x3918U,0x7230U,0xe460U,0xd8cbU,0xa19dU,0x5331U,
		0xa662U,0x5ccfU,0xb99eU,0x6337U,0xc66eU,0x9cd7U,0x29a5U,0x534aU,0xa694U,0x5d23U,
		0xba46U,0x6487U,0xc90eU,0x8217U,0x1425U,0x284aU,0x5094U,0xa128U,0x525bU,0xa4b6U,
		0x5967U,0xb2ceU,0x7597U,0xeb2eU,0xc657U,0x9ca5U,0x2941U,0x5282U,0xa504U,0x5a03U,
		0xb406U,0x7807U,0xf00eU,0xf017U,0xf025U,0xf041U,0xf089U,0xf119U,0xf239U,0xf479U,
		0xf8f9U,0xe1f9U,0xd3f9U,0xb7f9U,0x7ff9U,0xfff2U,0xefefU,0xcfd5U,0x8fa1U,0x0f49U,
		0x1e92U,0x3d24U,0x7a48U,0xf490U,0xf92bU,0xe25dU,0xd4b1U,0xb969U,0x62d9U,0xc5b2U,
		0x9b6fU,0x26d5U,0x4daaU,0x9b54U,0x26a3U,0x4d46U,0x9a8cU,0x2513U,0x4a26U,0x944cU,
		0x3893U,0x7126U,0xe24cU,0xd493U,0xb92dU,0x6251U,0xc4a2U,0x994fU,0x2295U,0x452aU,
		0x8a54U,0x04a3U,0x0946U,0x128cU,0x2518U,0x4a30U,0x9460U,0x38cbU,0x7196U,0xe32cU,
		0xd653U,0xbcadU,0x6951U,0xd2a2U,0xb54fU,0x7a95U,0xf52aU,0xfa5fU,0xe4b5U,0xd961U,
		0xa2c9U,0x5599U,0xab32U,0x466fU,0x8cdeU,0x09b7U,0x136eU,0x26dcU,0x4db8U,0x9b70U,
		0x26ebU,0x4dd6U,0x9bacU,0x2753U,0x4ea6U,0x9d4cU,0x2a93U,0x5526U,0xaa4cU,0x4493U,
		0x8926U,0x0247U,0x048eU,0x091cU,0x1238U,0x2470U,0x48e0U,0x91c0U,0x338bU,0x6716U,
		0xce2cU,0x8c53U,0x08adU,0x115aU,0x22b4U,0x4568U,0x8ad0U,0x05abU,0x0b56U,0x16acU,
		0x2d58U,0x5ab0U,0xb560U,0x7acbU,0xf596U,0xfb27U,0xe645U,0xdc81U,0xa909U,0x4219U,
		0x8432U,0x186fU,0x30deU,0x61bcU,0xc378U,0x96fbU,0x3dfdU,0x7bfaU,0xf7f4U,0xffe3U,
		0xefcdU,0xcf91U,0x8f29U,0x0e59U,0x1cb2U,0x3964U,0x72c8U,0xe590U,0xdb2bU,0xa65dU,
		0x5cb1U,0xb962U,0x62cfU,0xc59eU,0x9b37U,0x2665U,0x4ccaU,0x9994U,0x2323U,0x4646U,
		0x8c8cU,0x0913U,0x1226U,0x244cU,0x4898U,0x9130U,0x326bU,0x64d6U,0xc9acU,0x8353U,
		0x16adU,0x2d5aU,0x5ab4U,0xb568U,0x7adbU,0xf5b6U,0xfb67U,0xe6c5U,0xdd81U,0xab09U,
		0x4619U,0x8c32U,0x086fU,0x10deU,0x21bcU,0x4378U,0x86f0U,0x1debU,0x3bd6U,0x77acU,
		0xef58U,0xcebbU,0x8d7dU,0x0af1U,0x15e2U,0x2bc4U,0x5788U,0xaf10U,0x4e2bU,0x9c56U,
		0x28a7U,0x514eU,0xa29cU,0x5533U,0xaa66U,0x44c7U,0x898eU,0x0317U,0x062eU,0x0c5cU,
		0x18b8U,0x3170U,0x62e0U,0xc5c0U,0x9b8bU,0x271dU,0x4e3aU,0x9c74U,0x28e3U,0x51c6U,
		0xa38cU,0x5713U,0xae26U,0x4c47U,0x988eU,0x2117U,0x422eU,0x845cU,0x18b3U,0x3166U,
		0x62ccU,0xc598U,0x9b3bU,0x267dU,0x4cfaU,0x99f4U,0x23e3U,0x47c6U,0x8f8cU,0x0f13U,
		0x1e26U,0x3c4cU,0x7898U,0xf130U,0xf26bU,0xf4ddU,0xf9b1U,0xe369U,0xd6d9U,0xbdb9U,
		0x6b79U,0xd6f2U,0xbdefU,0x6bd5U,0xd7aaU,0xbf5fU,0x6eb5U,0xdd6aU,0xaadfU,0x45b5U,
		0x8b6aU,0x06dfU,0x0dbeU,0x1b7cU,0x36f8U,0x6df0U,0xdbe0U,0xa7cbU,0x5f9dU,0xbf3aU,
		0x6e7fU,0xdcfeU,0xa9f7U,0x43e5U,0x87caU,0x1f9fU,0x3f3eU,0x7e7cU,0xfcf8U,0xe9fbU,
		0xc3fdU,0x97f1U,0x3fe9U,0x7fd2U,0xffa4U,0xef43U,0xce8dU,0x8d11U,0x0a29U,0x1452U,
		0x28a4U,0x5148U,0xa290U,0x552bU,0xaa56U,0x44a7U,0x894eU,0x0297U,0x052eU,0x0a5cU,
		0x14b8U,0x2970U,0x52e0U,0xa5c0U,0x5b8bU,0xb716U,0x7e27U,0xfc4eU,0xe897U,0xc125U,
		0x9241U,0x3489U,0x6912U,0xd224U,0xb443U,0x788dU,0xf11aU,0xf23fU,0xf475U,0xf8e1U,
		0xe1c9U,0xd399U,0xb739U,0x7e79U,0xfcf2U,0xe9efU,0xc3d5U,0x97a1U,0x3f49U,0x7e92U,
		0xfd24U,0xea43U,0xc48dU,0x9911U,0x2229U,0x4452U,0x88a4U,0x0143U,0x0286U,0x050cU,
		0x0a18U,0x1430U,0x2860U,0x50c0U,0xa180U,0x530bU,0xa616U,0x5c27U,0xb84eU,0x6097U,
		0xc12eU,0x9257U,0x34a5U,0x694aU,0xd294U,0xb523U,0x7a4dU,0xf49aU,0xf93fU,0xe275U,
		0xd4e1U,0xb9c9U,0x6399U,0xc732U,0x9e6fU,0x2cd5U,0x59aaU,0xb354U,0x76a3U,0xed46U,
		0xca87U,0x8505U,0x1a01U,0x3402U,0x6804U,0xd008U,0xb01bU,0x703dU,0xe07aU,0xd0ffU,
		0xb1f5U,0x73e1U,0xe7c2U,0xdf8fU,0xaf15U,0x4e21U,0x9c42U,0x288fU,0x511eU,0xa23cU,
		0x5473U,0xa8e6U,0x41c7U,0x838eU,0x1717U,0x2e2eU,0x5c5cU,0xb8b8U,0x617bU,0xc2f6U,
		0x95e7U,0x3bc5U,0x778aU,0xef14U,0xce23U,0x8c4dU,0x0891U,0x1122U,0x2244U,0x4488U,
		0x8910U,0x022bU,0x0456U,0x08acU,0x1158U,0x22b0U,0x4560U,0x8ac0U,0x058bU,0x0b16U,
		0x162cU,0x2c58U,0x58b0U,0xb160U,0x72cbU,0xe596U,0xdb27U,0xa645U,0x5c81U,0xb902U,
		0x620fU,0xc41eU,0x9837U,0x2065U,0x40caU,0x8194U,0x1323U,0x2646U,0x4c8cU,0x9918U,
		0x223bU,0x4476U,0x88ecU,0x01d3U,0x03a6U,0x074cU,0x0e98U,0x1d30U,0x3a60U,0x74c0U,
		0xe980U,0xc30bU,0x961dU,0x3c31U,0x7862U,0xf0c4U,0xf183U,0xf30dU,0xf611U,0xfc29U,
		0xe859U,0xc0b9U,0x9179U,0x32f9U,0x65f2U,0xcbe4U,0x87c3U,0x1f8dU,0x3f1aU,0x7e34U,
		0xfc68U,0xe8dbU,0xc1bdU,0x9371U,0x36e9U,0x6dd2U,0xdba4U,0xa743U,0x5e8dU,0xbd1aU,
		0x6a3fU,0xd47eU,0xb8f7U,0x61e5U,0xc3caU,0x979fU,0x3f35U,0x7e6aU,0xfcd4U,0xe9a3U,
		0xc34dU,0x9691U,0x3d29U,0x7a52U,0xf4a4U,0xf943U,0xe28dU,0xd511U,0xba29U,0x6459U,
		0xc8b2U,0x816fU,0x12d5U,0x25aaU,0x4b54U,0x96a8U,0x3d5bU,0x7ab6U,0xf56cU,0xfad3U,
		0xe5adU,0xdb51U,0xa6a9U,0x5d59U,0xbab2U,0x656fU,0xcadeU,0x85b7U,0x1b65U,0x36caU,
		0x6d94U,0xdb28U,0xa65bU,0x5cbdU,0xb97aU,0x62ffU,0xc5feU,0x9bf7U,0x27e5U,0x4fcaU,
		0x9f94U,0x2f23U,0x5e46U,0xbc8cU,0x6913U,0xd226U,0xb447U,0x7885U,0xf10aU,0xf21fU,
		0xf435U,0xf861U,0xe0c9U,0xd199U,0xb339U,0x7679U,0xecf2U,0xc9efU,0x83d5U,0x17a1U,
		0x2f42U,0x5e84U,0xbd08U,0x6a1bU,0xd436U,0xb867U,0x60c5U,0xc18aU,0x931fU,0x3635U,
		0x6c6aU,0xd8d4U,0xa1a3U,0x534dU,0xa69aU,0x5d3fU,0xba7eU,0x64f7U,0xc9eeU,0x83d7U,
		0x17a5U,0x2f4aU,0x5e94U,0xbd28U,0x6a5bU,0xd4b6U,0xb967U,0x62c5U,0xc58aU,0x9b1fU,
		0x2635U,0x4c6aU,0x98d4U,0x21a3U,0x4346U,0x868cU,0x1d13U,0x3a26U,0x744cU,0xe898U,
		0xc13bU,0x927dU,0x34f1U,0x69e2U,0xd3c4U,0xb783U,0x7f0dU,0xfe1aU,0xec3fU,0xc875U,
		0x80e1U,0x11c9U,0x2392U,0x4724U,0x8e48U,0x0c9bU,0x1936U,0x326cU,0x64d8U,0xc9b0U,
		0x836bU,0x16ddU,0x2dbaU,0x5b74U,0xb6e8U,0x7ddbU,0xfbb6U,0xe767U,0xdec5U,0xad81U,
		0x4b09U,0x9612U,0x3c2fU,0x785eU,0xf0bcU,0xf173U,0xf2edU,0xf5d1U,0xfba9U,0xe759U,
		0xdeb9U,0xad79U,0x4af9U,0x95f2U,0x3befU,0x77deU,0xefbcU,0xcf73U,0x8eedU,0x0dd1U,
		0x1ba2U,0x3744U,0x6e88U,0xdd10U,0xaa2bU,0x445dU,0x88baU,0x017fU,0x02feU,0x05fcU,
		0x0bf8U,0x17f0U,0x2fe0U,0x5fc0U,0xbf80U,0x6f0bU,0xde16U,0xac27U,0x4845U,0x908aU,
		0x311fU,0x623eU,0xc47cU,0x98f3U,0x21edU,0x43daU,0x87b4U,0x1f63U,0x3ec6U,0x7d8cU,
		0xfb18U,0xe63bU,0xdc7dU,0xa8f1U,0x41e9U,0x83d2U,0x17afU,0x2f5eU,0x5ebcU,0xbd78U,
		0x6afbU,0xd5f6U,0xbbe7U,0x67c5U,0xcf8aU,0x8f1fU,0x0e35U,0x1c6aU,0x38d4U,0x71a8U,
		0xe350U,0xd6abU,0xbd5dU,0x6ab1U,0xd562U,0xbacfU,0x6595U,0xcb2aU,0x865fU,0x1cb5U,
		0x396aU,0x72d4U,0xe5a8U,0xdb5bU,0xa6bdU,0x5d71U,0xbae2U,0x65cfU,0xcb9eU,0x8737U,
		0x1e65U,0x3ccaU,0x7994U,0xf328U,0xf65bU,0xfcbdU,0xe971U,0xc2e9U,0x95d9U,0x3bb9U,
		0x7772U,0xeee4U,0xcdc3U,0x8b8dU,0x0711U,0x0e22U,0x1c44U,0x3888U,0x7110U,0xe220U,
		0xd44bU,0xb89dU,0x6131U,0xc262U,0x94cfU,0x3995U,0x732aU,0xe654U,0xdca3U,0xa94dU,
		0x4291U,0x8522U,0x1a4fU,0x349eU,0x693cU,0xd278U,0xb4fbU,0x79fdU,0xf3faU,0xf7ffU,
		0xfff5U,0xefe1U,0xcfc9U,0x8f99U,0x0f39U,0x1e72U,0x3ce4U,0x79c8U,0xf390U,0xf72bU,
		0xfe5dU,0xecb1U,0xc969U,0x82d9U,0x15b9U,0x2b72U,0x56e4U,0xadc8U,0x4b9bU,0x9736U,
		0x3e67U,0x7cceU,0xf99cU,0xe333U,0xd66dU,0xbcd1U,0x69a9U,0xd352U,0xb6afU,0x7d55U,
		0xfaaaU,0xe55fU,0xdab5U,0xa561U,0x5ac9U,0xb592U,0x7b2fU,0xf65eU,0xfcb7U,0xe965U,
		0xc2c1U,0x9589U,0x3b19U,0x7632U,0xec64U,0xc8c3U,0x818dU,0x1311U,0x2622U,0x4c44U,
		0x9888U,0x211bU,0x4236U,0x846cU,0x18d3U,0x31a6U,0x634cU,0xc698U,0x9d3bU,0x2a7dU,
		0x54faU,0xa9f4U,0x43e3U,0x87c6U,0x1f87U,0x3f0eU,0x7e1cU,0xfc38U,0xe87bU,0xc0fdU,
		0x91f1U,0x33e9U,0x67d2U,0xcfa4U,0x8f43U,0x0e8dU,0x1d1aU,0x3a34U,0x7468U,0xe8d0U,
		0xc1abU,0x935dU,0x36b1U,0x6d62U,0xdac4U,0xa583U,0x5b0dU,0xb61aU,0x7c3fU,0xf87eU,
		0xe0f7U,0xd1e5U,0xb3c1U,0x7789U,0xef12U,0xce2fU,0x8c55U,0x08a1U,0x1142U,0x2284U,
		0x4508U,0x8a10U,0x042bU,0x0856U,0x10acU,0x2158U,0x42b0U,0x8560U,0x1acbU,0x3596U,
		0x6b2cU,0xd658U,0xbcbbU,0x697dU,0xd2faU,0xb5ffU,0x7bf5U,0xf7eaU,0xffdfU,0xefb5U,
		0xcf61U,0x8ec9U,0x0d99U,0x1b32U,0x3664U,0x6cc8U,0xd990U,0xa32bU,0x565dU,0xacbaU,
		0x497fU,0x92feU,0x35f7U,0x6beeU,0xd7dcU,0xbfb3U,0x6f6dU,0xdedaU,0xadbfU,0x4b75U,
		0x96eaU,0x3ddfU,0x7bbeU,0xf77cU,0xfef3U,0xededU,0xcbd1U,0x87a9U,0x1f59U,0x3eb2U,
		0x7d64U,0xfac8U,0xe59bU,0xdb3dU,0xa671U,0x5ce9U,0xb9d2U,0x63afU,0xc75eU,0x9eb7U,
		0x2d65U,0x5acaU,0xb594U,0x7b23U,0xf646U,0xfc87U,0xe905U,0xc201U,0x9409U,0x3819U,
		0x7032U,0xe064U,0xd0c3U,0xb18dU,0x7311U,0xe622U,0xdc4fU,0xa895U,0x4121U,0x8242U,
		0x148fU,0x291eU,0x523cU,0xa478U,0x58fbU,0xb1f6U,0x73e7U,0xe7ceU,0xdf97U,0xaf25U,
		0x4e41U,0x9c82U,0x290fU,0x521eU,0xa43cU,0x5873U,0xb0e6U,0x71c7U,0xe38eU,0xd717U,
		0xbe25U,0x6c41U,0xd882U,0xa10fU,0x5215U,0xa42aU,0x585fU,0xb0beU,0x7177U,0xe2eeU,
		0xd5d7U,0xbba5U,0x6741U,0xce82U,0x8d0fU,0x0a15U,0x142aU,0x2854U,0x50a8U,0xa150U,
		0x52abU,0xa556U,0x5aa7U,0xb54eU,0x7a97U,0xf52eU,0xfa57U,0xe4a5U,0xd941U,0xa289U,
		0x5519U,0xaa32U,0x446fU,0x88deU,0x01b7U,0x036eU,0x06dcU,0x0db8U,0x1b70U,0x36e0U,
		0x6dc0U,0xdb80U,0xa70bU,0x5e1dU,0xbc3aU,0x687fU,0xd0feU,0xb1f7U,0x73e5U,0xe7caU,
		0xdf9fU,0xaf35U,0x4e61U,0x9cc2U,0x298fU,0x531eU,0xa63cU,0x5c73U,0xb8e6U,0x61c7U,
		0xc38eU,0x9717U,0x3e25U,0x7c4aU,0xf894U,0xe123U,0xd24dU,0xb491U,0x7929U,0xf252U,
		0xf4afU,0xf955U,0xe2a1U,0xd549U,0xba99U,0x6539U,0xca72U,0x84efU,0x19d5U,0x33aaU,
		0x6754U,0xcea8U,0x8d5bU,0x0abdU,0x157aU,0x2af4U,0x55e8U,0xabd0U,0x47abU,0x8f56U,
		0x0ea7U,0x1d4eU,0x3a9cU,0x7538U,0xea70U,0xc4ebU,0x99ddU,0x23b1U,0x4762U,0x8ec4U,
		0x0d83U,0x1b06U,0x360cU,0x6c18U,0xd830U,0xa06bU,0x50ddU,0xa1baU,0x537fU,0xa6feU,
		0x5df7U,0xbbeeU,0x67d7U,0xcfaeU,0x8f57U,0x0ea5U,0x1d4aU,0x3a94U,0x7528U,0xea50U,
		0xc4abU,0x995dU,0x22b1U,0x4562U,0x8ac4U,0x0583U,0x0b06U,0x160cU,0x2c18U,0x5830U,
		0xb060U,0x70cbU,0xe196U,0xd327U,0xb645U,0x7c81U,0xf902U,0xe20fU,0xd415U,0xb821U,
		0x6049U,0xc092U,0x912fU,0x3255U,0x64aaU,0xc954U,0x82a3U,0x154dU,0x2a9aU,0x5534U,
		0xaa68U,0x44dbU,0x89b6U,0x0367U,0x06ceU,0x0d9cU,0x1b38U,0x3670U,0x6ce0U,0xd9c0U,
		0xa38bU,0x571dU,0xae3aU,0x4c7fU,0x98feU,0x21f7U,0x43eeU,0x87dcU,0x1fb3U,0x3f66U,
		0x7eccU,0xfd98U,0xeb3bU,0xc67dU,0x9cf1U,0x29e9U,0x53d2U,0xa7a4U,0x5f43U,0xbe86U,
		0x6d07U,0xda0eU,0xa417U,0x5825U,0xb04aU,0x709fU,0xe13eU,0xd277U,0xb4e5U,0x79c1U,
		0xf382U,0xf70fU,0xfe15U,0xec21U,0xc849U,0x8099U,0x1139U,0x2272U,0x44e4U,0x89c8U,
		0x039bU,0x0736U,0x0e6cU,0x1cd8U,0x39b0U,0x7360U,0xe6c0U,0xdd8bU,0xab1dU,0x4631U,
		0x8c62U,0x08cfU,0x119eU,0x233cU,0x4678U,0x8cf0U,0x09ebU,0x13d6U,0x27acU,0x4f58U,
		0x9eb0U,0x2d6bU,0x5ad6U,0xb5acU,0x7b53U,0xf6a6U,0xfd47U,0xea85U,0xc501U,0x9a09U,
		0x2419U,0x4832U,0x9064U,0x30c3U,0x6186U,0xc30cU,0x9613U,0x3c2dU,0x785aU,0xf0b4U,
		0xf163U,0xf2cdU,0xf591U,0xfb29U,0xe659U,0xdcb9U,0xa979U,0x42f9U,0x85f2U,0x1befU,
		0x37deU,0x6fbcU,0xdf78U,0xaefbU,0x4dfdU,0x9bfaU,0x27ffU,0x4ffeU,0x9ffcU,0x2ff3U,
		0x5fe6U,0xbfccU,0x6f93U,0xdf26U,0xae47U,0x4c85U,0x990aU,0x221fU,0x443eU,0x887cU,
		0x00f3U,0x01e6U,0x03ccU,0x0798U,0x0f30U,0x1e60U,0x3cc0U,0x7980U,0xf300U,0xf60bU,
		0xfc1dU,0xe831U,0xc069U,0x90d9U,0x31b9U,0x6372U,0xc6e4U,0x9dc3U,0x2b8dU,0x571aU,
		0xae34U,0x4c63U,0x98c6U,0x2187U,0x430eU,0x861cU,0x1c33U,0x3866U,0x70ccU,0xe198U,
		0xd33bU,0xb67dU,0x7cf1U,0xf9e2U,0xe3cfU,0xd795U,0xbf21U,0x6e49U,0xdc92U,0xa92fU,
		0x4255U,0x84aaU,0x195fU,0x32beU,0x657cU,0xcaf8U,0x85fbU,0x1bfdU,0x37faU,0x6ff4U,
		0xdfe8U,0xafdbU,0x4fbdU,0x9f7aU,0x2effU,0x5dfeU,0xbbfcU,0x67f3U,0xcfe6U,0x8fc7U,
		0x0f85U,0x1f0aU,0x3e14U,0x7c28U,0xf850U,0xe0abU,0xd15dU,0xb2b1U,0x7569U,0xead2U,
		0xc5afU,0x9b55U,0x26a1U,0x4d42U,0x9a84U,0x2503U,0x4a06U,0x940cU,0x3813U,0x7026U,
		0xe04cU,0xd093U,0xb12dU,0x7251U,0xe4a2U,0xd94fU,0xa295U,0x5521U,0xaa42U,0x448fU,
		0x891eU,0x0237U,0x046eU,0x08dcU,0x11b8U,0x2370U,0x46e0U,0x8dc0U,0x0b8bU,0x1716U,
		0x2e2cU,0x5c58U,0xb8b0U,0x616bU,0xc2d6U,0x95a7U,0x3b45U,0x768aU,0xed14U,0xca23U,
		0x844dU,0x1891U,0x3122U,0x6244U,0xc488U,0x991bU,0x223dU,0x447aU,0x88f4U,0x01e3U,
		0x03c6U,0x078cU,0x0f18U,0x1e30U,0x3c60U,0x78c0U,0xf180U,0xf30bU,0xf61dU,0xfc31U,
		0xe869U,0xc0d9U,0x91b9U,0x3379U,0x66f2U,0xcde4U,0x8bc3U,0x078dU,0x0f1aU,0x1e34U,
		0x3c68U,0x78d0U,0xf1a0U,0xf34bU,0xf69dU,0xfd31U,0xea69U,0xc4d9U,0x99b9U,0x2379U,
		0x46f2U,0x8de4U,0x0bc3U,0x1786U,0x2f0cU,0x5e18U,0xbc30U,0x686bU,0xd0d6U,0xb1a7U,
		0x7345U,0xe68aU,0xdd1fU,0xaa35U,0x4461U,0x88c2U,0x018fU,0x031eU,0x063cU,0x0c78U,
		0x18f0U,0x31e0U,0x63c0U,0xc780U,0x9f0bU,0x2e1dU,0x5c3aU,0xb874U,0x60e3U,0xc1c6U,
		0x9387U,0x3705U,0x6e0aU,0xdc14U,0xa823U,0x404dU,0x809aU,0x113fU,0x227eU,0x44fcU,
		0x89f8U,0x03fbU,0x07f6U,0x0fecU,0x1fd8U,0x3fb0U,0x7f60U,0xfec0U,0xed8bU,0xcb1dU,
		0x8631U,0x1c69U,0x38d2U,0x71a4U,0xe348U,0xd69bU,0xbd3dU,0x6a71U,0xd4e2U,0xb9cfU,
		0x6395U,0xc72aU,0x9e5fU,0x2cb5U,0x596aU,0xb2d4U,0x75a3U,0xeb46U,0xc687U,0x9d05U,
		0x2a01U,0x5402U,0xa804U,0x4003U,0x8006U,0x1007U,0x200eU,0x401cU,0x8038U,0x107bU,
		0x20f6U,0x41ecU,0x83d8U,0x17bbU,0x2f76U,0x5eecU,0xbdd8U,0x6bbbU,0xd776U,0xbee7U,
		0x6dc5U,0xdb8aU,0xa71fU,0x5e35U,0xbc6aU,0x68dfU,0xd1beU,0xb377U,0x76e5U,0xedcaU,
		0xcb9fU,0x8735U,0x1e61U,0x3cc2U,0x7984U,0xf308U,0xf61bU,0xfc3dU,0xe871U,0xc0e9U,
		0x91d9U,0x33b9U,0x6772U,0xcee4U,0x8dc3U,0x0b8dU,0x171aU,0x2e34U,0x5c68U,0xb8d0U,
		0x61abU,0xc356U,0x96a7U,0x3d45U,0x7a8aU,0xf514U,0xfa23U,0xe44dU,0xd891U,0xa129U,
		0x5259U,0xa4b2U,0x596fU,0xb2deU,0x75b7U,0xeb6eU,0xc6d7U,0x9da5U,0x2b41U,0x5682U,
		0xad04U,0x4a03U,0x9406U,0x3807U,0x700eU,0xe01cU,0xd033U,0xb06dU,0x70d1U,0xe1a2U,
		0xd34fU,0xb695U,0x7d21U,0xfa42U,0xe48fU,0xd915U,0xa221U,0x5449U,0xa892U,0x412fU,
		0x825eU,0x14b7U,0x296eU,0x52dcU,0xa5b8U,0x5b7bU,0xb6f6U,0x7de7U,0xfbceU,0xe797U,
		0xdf25U,0xae41U,0x4c89U,0x9912U,0x222fU,0x445eU,0x88bcU,0x0173U,0x02e6U,0x05ccU,
		0x0b98U,0x1730U,0x2e60U,0x5cc0U,0xb980U,0x630bU,0xc616U,0x9c27U,0x2845U,0x508aU,
		0xa114U,0x5223U,0xa446U,0x5887U,0xb10eU,0x7217U,0xe42eU,0xd857U,0xa0a5U,0x5141U,
		0xa282U,0x550fU,0xaa1eU,0x4437U,0x886eU,0x00d7U,0x01aeU,0x035cU,0x06b8U,0x0d70U,
		0x1ae0U,0x35c0U,0x6b80U,0xd700U,0xbe0bU,0x6c1dU,0xd83aU,0xa07fU,0x50f5U,0xa1eaU,
		0x53dfU,0xa7beU,0x5f77U,0xbeeeU,0x6dd7U,0xdbaeU,0xa757U,0x5ea5U,0xbd4aU,0x6a9fU,
		0xd53eU,0xba77U,0x64e5U,0xc9caU,0x839fU,0x1735U,0x2e6aU,0x5cd4U,0xb9a8U,0x635bU,
		0xc6b6U,0x9d67U,0x2ac5U,0x558aU,0xab14U,0x4623U,0x8c46U,0x0887U,0x110eU,0x221cU,
		0x4438U,0x8870U,0x00ebU,0x01d6U,0x03acU,0x0758U,0x0eb0U,0x1d60U,0x3ac0U,0x7580U,
		0xeb00U,0xc60bU,0x9c1dU,0x2831U,0x5062U,0xa0c4U,0x5183U,0xa306U,0x5607U,0xac0eU,
		0x4817U,0x902eU,0x3057U,0x60aeU,0xc15cU,0x92b3U,0x356dU,0x6adaU,0xd5b4U,0xbb63U,
		0x66cdU,0xcd9aU,0x8b3fU,0x0675U,0x0ceaU,0x19d4U,0x33a8U,0x6750U,0xcea0U,0x8d4bU,
		0x0a9dU,0x153aU,0x2a74U,0x54e8U,0xa9d0U,0x43abU,0x8756U,0x1ea7U,0x3d4eU,0x7a9cU,
		0xf538U,0xfa7bU,0xe4fdU,0xd9f1U,0xa3e9U,0x57d9U,0xafb2U,0x4f6fU,0x9edeU,0x2db7U,
		0x5b6eU,0xb6dcU,0x7db3U,0xfb66U,0xe6c7U,0xdd85U,0xab01U,0x4609U,0x8c12U,0x082fU,
		0x105eU,0x20bcU,0x4178U,0x82f0U,0x15ebU,0x2bd6U,0x57acU,0xaf58U,0x4ebbU,0x9d76U,
		0x2ae7U,0x55ceU,0xab9cU,0x4733U,0x8e66U,0x0cc7U,0x198eU,0x331cU,0x6638U,0xcc70U,
		0x88ebU,0x01ddU,0x03baU,0x0774U,0x0ee8U,0x1dd0U,0x3ba0U,0x7740U,0xee80U,0xcd0bU,
		0x8a1dU,0x0431U,0x0862U,0x10c4U,0x2188U,0x4310U,0x8620U,0x1c4bU,0x3896U,0x712cU,
		0xe258U,0xd4bbU,0xb97dU,0x62f1U,0xc5e2U,0x9bcfU,0x2795U,0x4f2aU,0x9e54U,0x2ca3U,
		0x5946U,0xb28cU,0x7513U,0xea26U,0xc447U,0x9885U,0x2101U,0x4202U,0x8404U,0x1803U,
		0x3006U,0x600cU,0xc018U,0x903bU,0x307dU,0x60faU,0xc1f4U,0x93e3U,0x37cdU,0x6f9aU,
		0xdf34U,0xae63U,0x4ccdU,0x999aU,0x233fU,0x467eU,0x8cfcU,0x09f3U,0x13e6U,0x27ccU,
		0x4f98U,0x9f30U,0x2e6bU,0x5cd6U,0xb9acU,0x6353U,0xc6a6U,0x9d47U,0x2a85U,0x550aU,
		0xaa14U,0x4423U,0x8846U,0x0087U,0x010eU,0x021cU,0x0438U,0x0870U,0x10e0U,0x21c0U,
		0x4380U,0x8700U,0x1e0bU,0x3c16U,0x782cU,0xf058U,0xf0bbU,0xf17dU,0xf2f1U,0xf5e9U,
		0xfbd9U,0xe7b9U,0xdf79U,0xaef9U,0x4df9U,0x9bf2U,0x27efU,0x4fdeU,0x9fbcU,0x2f73U,
		0x5ee6U,0xbdccU,0x6b93U,0xd726U,0xbe47U,0x6c85U,0xd90aU,0xa21fU,0x5435U,0xa86aU,
		0x40dfU,0x81beU,0x1377U,0x26eeU,0x4ddcU,0x9bb8U,0x277bU,0x4ef6U,0x9decU,0x2bd3U,
		0x57a6U,0xaf4cU,0x4e93U,0x9d26U,0x2a47U,0x548eU,0xa91cU,0x4233U,0x8466U,0x18c7U,
		0x318eU,0x631cU,0xc638U,0x9c7bU,0x28fdU,0x51faU,0xa3f4U,0x57e3U,0xafc6U,0x4f87U,
		0x9f0eU,0x2e17U,0x5c2eU,0xb85cU,0x60b3U,0xc166U,0x92c7U,0x3585U,0x6b0aU,0xd614U,
		0xbc23U,0x684dU,0xd09aU,0xb13fU,0x7275U,0xe4eaU,0xd9dfU,0xa3b5U,0x5761U,0xaec2U,
		0x4d8fU,0x9b1eU,0x2637U,0x4c6eU,0x98dcU,0x21b3U,0x4366U,0x86ccU,0x1d93U,0x3b26U,
		0x764cU,0xec98U,0xc93bU,0x827dU,0x14f1U,0x29e2U,0x53c4U,0xa788U,0x5f1bU,0xbe36U,
		0x6c67U,0xd8ceU,0xa197U,0x5325U,0xa64aU,0x5c9fU,0xb93eU,0x6277U,0xc4eeU,0x99d7U,
		0x23a5U,0x474aU,0x8e94U,0x0d23U,0x1a46U,0x348cU,0x6918U,0xd230U,0xb46bU,0x78ddU,
		0xf1baU,0xf37fU,0xf6f5U,0xfde1U,0xebc9U,0xc799U,0x9f39U,0x2e79U,0x5cf2U,0xb9e4U,
		0x63c3U,0xc786U,0x9f07U,0x2e05U,0x5c0aU,0xb814U,0x6023U,0xc046U,0x9087U,0x3105U,
		0x620aU,0xc414U,0x9823U,0x204dU,0x409aU,0x8134U,0x1263U,0x24c6U,0x498cU,0x9318U,
		0x363bU,0x6c76U,0xd8ecU,0xa1d3U,0x53adU,0xa75aU,0x5ebfU,0xbd7eU,0x6af7U,0xd5eeU,
		0xbbd7U,0x67a5U,0xcf4aU,0x8e9fU,0x0d35U,0x1a6aU,0x34d4U,0x69a8U,0xd350U,0xb6abU,
		0x7d5dU,0xfabaU,0xe57fU,0xdaf5U,0xa5e1U,0x5bc9U,0xb792U,0x7f2fU,0xfe5eU,0xecb7U,
		0xc965U,0x82c1U,0x1589U,0x2b12U,0x5624U,0xac48U,0x489bU,0x9136U,0x3267U,0x64ceU,
		0xc99cU,0x8333U,0x166dU,0x2cdaU,0x59b4U,0xb368U,0x76dbU,0xedb6U,0xcb67U,0x86c5U,
		0x1d81U,0x3b02U,0x7604U,0xec08U,0xc81bU,0x803dU,0x1071U,0x20e2U,0x41c4U,0x8388U,
		0x171bU,0x2e36U,0x5c6cU,0xb8d8U,0x61bbU,0xc376U,0x96e7U,0x3dc5U,0x7b8aU,0xf714U,
		0xfe23U,0xec4dU,0xc891U,0x8129U,0x1259U,0x24b2U,0x4964U,0x92c8U,0x359bU,0x6b36U,
		0xd66cU,0xbcd3U,0x69adU,0xd35aU,0xb6bfU,0x7d75U,0xfaeaU,0xe5dfU,0xdbb5U,0xa761U,
		0x5ec9U,0xbd92U,0x6b2fU,0xd65eU,0xbcb7U,0x6965U,0xd2caU,0xb59fU,0x7b35U,0xf66aU,
		0xfcdfU,0xe9b5U,0xc361U,0x96c9U,0x3d99U,0x7b32U,0xf664U,0xfcc3U,0xe98dU,0xc311U,
		0x9629U,0x3c59U,0x78b2U,0xf164U,0xf2c3U,0xf58dU,0xfb11U,0xe629U,0xdc59U,0xa8b9U,
		0x4179U,0x82f2U,0x15efU,0x2bdeU,0x57bcU,0xaf78U,0x4efbU,0x9df6U,0x2be7U,0x57ceU,
		0xaf9cU,0x4f33U,0x9e66U,0x2cc7U,0x598eU,0xb31cU,0x7633U,0xec66U,0xc8c7U,0x8185U,
		0x1301U,0x2602U,0x4c04U,0x9808U,0x201bU,0x4036U,0x806cU,0x10d3U,0x21a6U,0x434cU,
		0x8698U,0x1d3bU,0x3a76U,0x74ecU,0xe9d8U,0xc3bbU,0x977dU,0x3ef1U,0x7de2U,0xfbc4U,
		0xe783U,0xdf0dU,0xae11U,0x4c29U,0x9852U,0x20afU,0x415eU,0x82bcU,0x1573U,0x2ae6U,
		0x55ccU,0xab98U,0x473bU,0x8e76U,0x0ce7U,0x19ceU,0x339cU,0x6738U,0xce70U,0x8cebU,
		0x09ddU,0x13baU,0x2774U,0x4ee8U,0x9dd0U,0x2babU,0x5756U,0xaeacU,0x4d53U,0x9aa6U,
		0x2547U,0x4a8eU,0x951cU,0x3a33U,0x7466U,0xe8ccU,0xc193U,0x932dU,0x3651U,0x6ca2U,
		0xd944U,0xa283U,0x550dU,0xaa1aU,0x443fU,0x887eU,0x00f7U,0x01eeU,0x03dcU,0x07b8U,
		0x0f70U,0x1ee0U,0x3dc0U,0x7b80U,0xf700U,0xfe0bU,0xec1dU,0xc831U,0x8069U,0x10d9U,
		0x21b2U,0x4364U,0x86c8U,0x1d9bU,0x3b36U,0x766cU,0xecd8U,0xc9bbU,0x837dU,0x16f1U,
		0x2de2U,0x5bc4U,0xb788U,0x7f1bU,0xfe36U,0xec67U,0xc8c5U,0x8181U,0x1309U,0x2612U,
		0x4c24U,0x9848U,0x209bU,0x4136U,0x826cU,0x14d3U,0x29a6U,0x534cU,0xa698U,0x5d3bU,
		0xba76U,0x64e7U,0xc9ceU,0x8397U,0x1725U,0x2e4aU,0x5c94U,0xb928U,0x625bU,0xc4b6U,
		0x9967U,0x22c5U,0x458aU,0x8b14U,0x0623U,0x0c46U,0x188cU,0x3118U,0x6230U,0xc460U,
		0x98cbU,0x219dU,0x433aU,0x8674U,0x1ce3U,0x39c6U,0x738cU,0xe718U,0xde3bU,0xac7dU,
		0x48f1U,0x91e2U,0x33cfU,0x679eU,0xcf3cU,0x8e73U,0x0cedU,0x19daU,0x33b4U,0x6768U,
		0xced0U,0x8dabU,0x0b5dU,0x16baU,0x2d74U,0x5ae8U,0xb5d0U,0x7babU,0xf756U,0xfea7U,
		0xed45U,0xca81U,0x8509U,0x1a19U,0x3432U,0x6864U,0xd0c8U,0xb19bU,0x733dU,0xe67aU,
		0xdcffU,0xa9f5U,0x43e1U,0x87c2U,0x1f8fU,0x3f1eU,0x7e3cU,0xfc78U,0xe8fbU,0xc1fdU,
		0x93f1U,0x37e9U,0x6fd2U,0xdfa4U,0xaf43U,0x4e8dU,0x9d1aU,0x2a3fU,0x547eU,0xa8fcU,
		0x41f3U,0x83e6U,0x17c7U,0x2f8eU,0x5f1cU,0xbe38U,0x6c7bU,0xd8f6U,0xa1e7U,0x53c5U,
		0xa78aU,0x5f1fU,0xbe3eU,0x6c77U,0xd8eeU,0xa1d7U,0x53a5U,0xa74aU,0x5e9fU,0xbd3eU,
		0x6a77U,0xd4eeU,0xb9d7U,0x63a5U,0xc74aU,0x9e9fU,0x2d35U,0x5a6aU,0xb4d4U,0x79a3U,
		0xf346U,0xf687U,0xfd05U,0xea01U,0xc409U,0x9819U,0x2039U,0x4072U,0x80e4U,0x11c3U,
		0x2386U,0x470cU,0x8e18U,0x0c3bU,0x1876U,0x30ecU,0x61d8U,0xc3b0U,0x976bU,0x3eddU,
		0x7dbaU,0xfb74U,0xe6e3U,0xddcdU,0xab91U,0x4729U,0x8e52U,0x0cafU,0x195eU,0x32bcU,
		0x6578U,0xcaf0U,0x85ebU,0x1bddU,0x37baU,0x6f74U,0xdee8U,0xaddbU,0x4bbdU,0x977aU,
		0x3effU,0x7dfeU,0xfbfcU,0xe7f3U,0xdfedU,0xafd1U,0x4fa9U,0x9f52U,0x2eafU,0x5d5eU,
		0xbabcU,0x6573U,0xcae6U,0x85c7U,0x1b85U,0x370aU,0x6e14U,0xdc28U,0xa85bU,0x40bdU,
		0x817aU,0x12ffU,0x25feU,0x4bfcU,0x97f8U,0x3ffbU,0x7ff6U,0xffecU,0xefd3U,0xcfadU,
		0x8f51U,0x0ea9U,0x1d52U,0x3aa4U,0x7548U,0xea90U,0xc52bU,0x9a5dU,0x24b1U,0x4962U,
		0x92c4U,0x3583U,0x6b06U,0xd60cU,0xbc13U,0x682dU,0xd05aU,0xb0bfU,0x7175U,0xe2eaU,
		0xd5dfU,0xbbb5U,0x6761U,0xcec2U,0x8d8fU,0x0b15U,0x162aU,0x2c54U,0x58a8U,0xb150U,
		0x72abU,0xe556U,0xdaa7U,0xa545U,0x5a81U,0xb502U,0x7a0fU,0xf41eU,0xf837U,0xe065U,
		0xd0c1U,0xb189U,0x7319U,0xe632U,0xdc6fU,0xa8d5U,0x41a1U,0x8342U,0x168fU,0x2d1eU,
		0x5a3cU,0xb478U,0x78fbU,0xf1f6U,0xf3e7U,0xf7c5U,0xff81U,0xef09U,0xce19U,0x8c39U,
		0x0879U,0x10f2U,0x21e4U,0x43c8U,0x8790U,0x1f2bU,0x3e56U,0x7cacU,0xf958U,0xe2bbU,
		0xd57dU,0xbaf1U,0x65e9U,0xcbd2U,0x87afU,0x1f55U,0x3eaaU,0x7d54U,0xfaa8U,0xe55bU,
		0xdabdU,0xa571U,0x5ae9U,0xb5d2U,0x7bafU,0xf75eU,0xfeb7U,0xed65U,0xcac1U,0x8589U,
		0x1b19U,0x3632U,0x6c64U,0xd8c8U,0xa19bU,0x533dU,0xa67aU,0x5cffU,0xb9feU,0x63f7U,
		0xc7eeU,0x9fd7U,0x2fa5U,0x5f4aU,0xbe94U,0x6d23U,0xda46U,0xa487U,0x5905U,0xb20aU,
		0x741fU,0xe83eU,0xc077U,0x90e5U,0x31c1U,0x6382U,0xc704U,0x9e03U,0x2c0dU,0x581aU,
		0xb034U,0x7063U,0xe0c6U,0xd187U,0xb305U,0x7601U,0xec02U,0xc80fU,0x8015U,0x1021U,
		0x2042U,0x4084U,0x8108U,0x121bU,0x2436U,0x486cU,0x90d8U,0x31bbU,0x6376U,0xc6ecU,
		0x9dd3U,0x2badU,0x575aU,0xaeb4U,0x4d63U,0x9ac6U,0x2587U,0x4b0eU,0x961cU,0x3c33U,
		0x7866U,0xf0ccU,0xf193U,0xf32dU,0xf651U,0xfca9U,0xe959U,0xc2b9U,0x9579U,0x3af9U,
		0x75f2U,0xebe4U,0xc7c3U,0x9f8dU,0x2f11U,0x5e22U,0xbc44U,0x6883U,0xd106U,0xb207U,
		0x7405U,0xe80aU,0xc01fU,0x9035U,0x3061U,0x60c2U,0xc184U,0x9303U,0x360dU,0x6c1aU,
		0xd834U,0xa063U,0x50cdU,0xa19aU,0x533fU,0xa67eU,0x5cf7U,0xb9eeU,0x63d7U,0xc7aeU,
		0x9f57U,0x2ea5U,0x5d4aU,0xba94U,0x6523U,0xca46U,0x8487U,0x1905U,0x320aU,0x6414U,
		0xc828U,0x805bU,0x10bdU,0x217aU,0x42f4U,0x85e8U,0x1bdbU,0x37b6U,0x6f6cU,0xded8U,
		0xadbbU,0x4b7dU,0x96faU,0x3dffU,0x7bfeU,0xf7fcU,0xfff3U,0xefedU,0xcfd1U,0x8fa9U,
		0x0f59U,0x1eb2U,0x3d64U,0x7ac8U,0xf590U,0xfb2bU,0xe65dU,0xdcb1U,0xa969U,0x42d9U,
		0x85b2U,0x1b6fU,0x36deU,0x6dbcU,0xdb78U,0xa6fbU,0x5dfdU,0xbbfaU,0x67ffU,0xcffeU,
		0x8ff7U,0x0fe5U,0x1fcaU,0x3f94U,0x7f28U,0xfe50U,0xecabU,0xc95dU,0x82b1U,0x1569U,
		0x2ad2U,0x55a4U,0xab48U,0x469bU,0x8d36U,0x0a67U,0x14ceU,0x299cU,0x5338U,0xa670U,
		0x5cebU,0xb9d6U,0x63a7U,0xc74eU,0x9e97U,0x2d25U,0x5a4aU,0xb494U,0x7923U,0xf246U,
		0xf487U,0xf905U,0xe201U,0xd409U,0xb819U,0x6039U,0xc072U,0x90efU,0x31d5U,0x63aaU,
		0xc754U,0x9ea3U,0x2d4dU,0x5a9aU,0xb534U,0x7a63U,0xf4c6U,0xf987U,0xe305U,0xd601U,
		0xbc09U,0x6819U,0xd032U,0xb06fU,0x70d5U,0xe1aaU,0xd35fU,0xb6b5U,0x7d61U,0xfac2U,
		0xe58fU,0xdb15U,0xa621U,0x5c49U,0xb892U,0x612fU,0xc25eU,0x94b7U,0x3965U,0x72caU,
		0xe594U,0xdb23U,0xa64dU,0x5c91U,0xb922U,0x624fU,0xc49eU,0x9937U,0x2265U,0x44caU,
		0x8994U,0x0323U,0x0646U,0x0c8cU,0x1918U,0x3230U,0x6460U,0xc8c0U,0x818bU,0x131dU,
		0x263aU,0x4c74U,0x98e8U,0x21dbU,0x43b6U,0x876cU,0x1ed3U,0x3da6U,0x7b4cU,0xf698U,
		0xfd3bU,0xea7dU,0xc4f1U,0x99e9U,0x23d9U,0x47b2U,0x8f64U,0x0ec3U,0x1d86U,0x3b0cU,
		0x7618U,0xec30U,0xc86bU,0x80ddU,0x11b1U,0x2362U,0x46c4U,0x8d88U,0x0b1bU,0x1636U,
		0x2c6cU,0x58d8U,0xb1b0U,0x736bU,0xe6d6U,0xdda7U,0xab45U,0x4681U,0x8d02U,0x0a0fU,
		0x141eU,0x283cU,0x5078U,0xa0f0U,0x51ebU,0xa3d6U,0x57a7U,0xaf4eU,0x4e97U,0x9d2eU,
		0x2a57U,0x54aeU,0xa95cU,0x42b3U,0x8566U,0x1ac7U,0x358eU,0x6b1cU,0xd638U,0xbc7bU,
		0x68fdU,0xd1faU,0xb3ffU,0x77f5U,0xefeaU,0xcfdfU,0x8fb5U,0x0f61U,0x1ec2U,0x3d84U,
		0x7b08U,0xf610U,0xfc2bU,0xe85dU,0xc0b1U,0x9169U,0x32d9U,0x65b2U,0xcb64U,0x86c3U,
		0x1d8dU,0x3b1aU,0x7634U,0xec68U,0xc8dbU,0x81bdU,0x1371U,0x26e2U,0x4dc4U,0x9b88U,
		0x271bU,0x4e36U,0x9c6cU,0x28d3U,0x51a6U,0xa34cU,0x5693U,0xad26U,0x4a47U,0x948eU,
		0x3917U,0x722eU,0xe45cU,0xd8b3U,0xa16dU,0x52d1U,0xa5a2U,0x5b4fU,0xb69eU,0x7d37U,
		0xfa6eU,0xe4d7U,0xd9a5U,0xa341U,0x5689U,0xad12U,0x4a2fU,0x945eU,0x38b7U,0x716eU,
		0xe2dcU,0xd5b3U,0xbb6dU,0x66d1U,0xcda2U,0x8b4fU,0x0695U,0x0d2aU,0x1a54U,0x34a8U,
		0x6950U,0xd2a0U,0xb54bU,0x7a9dU,0xf53aU,0xfa7fU,0xe4f5U,0xd9e1U,0xa3c9U,0x5799U,
		0xaf32U,0x4e6fU,0x9cdeU,0x29b7U,0x536eU,0xa6dcU,0x5db3U,0xbb66U,0x66c7U,0xcd8eU,
		0x8b17U,0x0625U,0x0c4aU,0x1894U,0x3128U,0x6250U,0xc4a0U,0x994bU,0x229dU,0x453aU,
		0x8a74U,0x04e3U,0x09c6U,0x138cU,0x2718U,0x4e30U,0x9c60U,0x28cbU,0x5196U,0xa32cU,
		0x5653U,0xaca6U,0x4947U,0x928eU,0x3517U,0x6a2eU,0xd45cU,0xb8b3U,0x616dU,0xc2daU,
		0x95bfU,0x3b75U,0x76eaU,0xedd4U,0xcba3U,0x874dU,0x1e91U,0x3d22U,0x7a44U,0xf488U,
		0xf91bU,0xe23dU,0xd471U,0xb8e9U,0x61d9U,0xc3b2U,0x976fU,0x3ed5U,0x7daaU,0xfb54U,
		0xe6a3U,0xdd4dU,0xaa91U,0x4529U,0x8a52U,0x04afU,0x095eU,0x12bcU,0x2578U,0x4af0U,
		0x95e0U,0x3bcbU,0x7796U,0xef2cU,0xce53U,0x8cadU,0x0951U,0x12a2U,0x2544U,0x4a88U,
		0x9510U,0x3a2bU,0x7456U,0xe8acU,0xc153U,0x92adU,0x3551U,0x6aa2U,0xd544U,0xba83U,
		0x650dU,0xca1aU,0x843fU,0x1875U,0x30eaU,0x61d4U,0xc3a8U,0x975bU,0x3ebdU,0x7d7aU,
		0xfaf4U,0xe5e3U,0xdbcdU,0xa791U,0x5f29U,0xbe52U,0x6cafU,0xd95eU,0xa2b7U,0x5565U,
		0xaacaU,0x459fU,0x8b3eU,0x0677U,0x0ceeU,0x19dcU,0x33b8U,0x6770U,0xcee0U,0x8dcbU,
		0x0b9dU,0x173aU,0x2e74U,0x5ce8U,0xb9d0U,0x63abU,0xc756U,0x9ea7U,0x2d45U,0x5a8aU,
		0xb514U,0x7a23U,0xf446U,0xf887U,0xe105U,0xd201U,0xb409U,0x7819U,0xf032U,0xf06fU,
		0xf0d5U,0xf1a1U,0xf349U,0xf699U,0xfd39U,0xea79U,0xc4f9U,0x99f9U,0x23f9U,0x47f2U,
		0x8fe4U,0x0fc3U,0x1f86U,0x3f0cU,0x7e18U,0xfc30U,0xe86bU,0xc0ddU,0x91b1U,0x3369U,
		0x66d2U,0xcda4U,0x8b43U,0x068dU,0x0d1aU,0x1a34U,0x3468U,0x68d0U,0xd1a0U,0xb34bU,
		0x769dU,0xed3aU,0xca7fU,0x84f5U,0x19e1U,0x33c2U,0x6784U,0xcf08U,0x8e1bU,0x0c3dU,
		0x187aU,0x30f4U,0x61e8U,0xc3d0U,0x97abU,0x3f5dU,0x7ebaU,0xfd74U,0xeae3U,0xc5cdU,
		0x9b91U,0x2729U,0x4e52U,0x9ca4U,0x2943U,0x5286U,0xa50cU,0x5a13U,0xb426U,0x7847U,
		0xf08eU,0xf117U,0xf225U,0xf441U,0xf889U,0xe119U,0xd239U,0xb479U,0x78f9U,0xf1f2U,
		0xf3efU,0xf7d5U,0xffa1U,0xef49U,0xce99U,0x8d39U,0x0a79U,0x14f2U,0x29e4U,0x53c8U,
		0xa790U,0x5f2bU,0xbe56U,0x6ca7U,0xd94eU,0xa297U,0x5525U,0xaa4aU,0x449fU,0x893eU,
		0x0277U,0x04eeU,0x09dcU,0x13b8U,0x2770U,0x4ee0U,0x9dc0U,0x2b8bU,0x5716U,0xae2cU,
		0x4c53U,0x98a6U,0x2147U,0x428eU,0x851cU,0x1a33U,0x3466U,0x68ccU,0xd198U,0xb33bU,
		0x767dU,0xecfaU,0xc9ffU,0x83f5U,0x17e1U,0x2fc2U,0x5f84U,0xbf08U,0x6e1bU,0xdc36U,
		0xa867U,0x40c5U,0x818aU,0x131fU,0x263eU,0x4c7cU,0x98f8U,0x21fbU,0x43f6U,0x87ecU,
		0x1fd3U,0x3fa6U,0x7f4cU,0xfe98U,0xed3bU,0xca7dU,0x84f1U,0x19e9U,0x33d2U,0x67a4U,
		0xcf48U,0x8e9bU,0x0d3dU,0x1a7aU,0x34f4U,0x69e8U,0xd3d0U,0xb7abU,0x7f5dU,0xfebaU,
		0xed7fU,0xcaf5U,0x85e1U,0x1bc9U,0x3792U,0x6f24U,0xde48U,0xac9bU,0x493dU,0x927aU,
		0x34ffU,0x69feU,0xd3fcU,0xb7f3U,0x7fedU,0xffdaU,0xefbfU,0xcf75U,0x8ee1U,0x0dc9U,
		0x1b92U,0x3724U,0x6e48U,0xdc90U,0xa92bU,0x425dU,0x84baU,0x197fU,0x32feU,0x65fcU,
		0xcbf8U,0x87fbU,0x1ffdU,0x3ffaU,0x7ff4U,0xffe8U,0xefdbU,0xcfbdU,0x8f71U,0x0ee9U,
		0x1dd2U,0x3ba4U,0x7748U,0xee90U,0xcd2bU,0x8a5dU,0x04b1U,0x0962U,0x12c4U,0x2588U,
		0x4b10U,0x9620U,0x3c4bU,0x7896U,0xf12cU,0xf253U,0xf4adU,0xf951U,0xe2a9U,0xd559U,
		0xbab9U,0x6579U,0xcaf2U,0x85efU,0x1bd5U,0x37aaU,0x6f54U,0xdea8U,0xad5bU,0x4abdU,
		0x957aU,0x3affU,0x75feU,0xebfcU,0xc7f3U,0x9fedU,0x2fd1U,0x5fa2U,0xbf44U,0x6e83U,
		0xdd06U,0xaa07U,0x4405U,0x880aU,0x001fU,0x003eU,0x007cU,0x00f8U,0x01f0U,0x03e0U,
		0x07c0U,0x0f80U,0x1f00U,0x3e00U,0x7c00U,0xf800U,0xe00bU,0xd01dU,0xb031U,0x7069U,
		0xe0d2U,0xd1afU,0xb355U,0x76a1U,0xed42U,0xca8fU,0x8515U,0x1a21U,0x3442U,0x6884U,
		0xd108U,0xb21bU,0x743dU,0xe87aU,0xc0ffU,0x91f5U,0x33e1U,0x67c2U,0xcf84U,0x8f03U,
		0x0e0dU,0x1c1aU,0x3834U,0x7068U,0xe0d0U,0xd1abU,0xb35dU,0x76b1U,0xed62U,0xcacfU,
		0x8595U,0x1b21U,0x3642U,0x6c84U,0xd908U,0xa21bU,0x543dU,0xa87aU,0x40ffU,0x81feU,
		0x13f7U,0x27eeU,0x4fdcU,0x9fb8U,0x2f7bU,0x5ef6U,0xbdecU,0x6bd3U,0xd7a6U,0xbf47U,
		0x6e85U,0xdd0aU,0xaa1fU,0x4435U,0x886aU,0x00dfU,0x01beU,0x037cU,0x06f8U,0x0df0U,
		0x1be0U,0x37c0U,0x6f80U,0xdf00U,0xae0bU,0x4c1dU,0x983aU,0x207fU,0x40feU,0x81fcU,
		0x13f3U,0x27e6U,0x4fccU,0x9f98U,0x2f3bU,0x5e76U,0xbcecU,0x69d3U,0xd3a6U,0xb747U,
		0x7e85U,0xfd0aU,0xea1fU,0xc435U,0x9861U,0x20c9U,0x4192U,0x8324U,0x1643U,0x2c86U,
		0x590cU,0xb218U,0x743bU,0xe876U,0xc0e7U,0x91c5U,0x3381U,0x6702U,0xce04U,0x8c03U,
		0x080dU,0x101aU,0x2034U,0x4068U,0x80d0U,0x11abU,0x2356U,0x46acU,0x8d58U,0x0abbU,
		0x1576U,0x2aecU,0x55d8U,0xabb0U,0x476bU,0x8ed6U,0x0da7U,0x1b4eU,0x369cU,0x6d38U,
		0xda70U,0xa4ebU,0x59ddU,0xb3baU,0x777fU,0xeefeU,0xcdf7U,0x8be5U,0x07c1U,0x0f82U,
		0x1f04U,0x3e08U,0x7c10U,0xf820U,0xe04bU,0xd09dU,0xb131U,0x7269U,0xe4d2U,0xd9afU,
		0xa355U,0x56a1U,0xad42U,0x4a8fU,0x951eU,0x3a37U,0x746eU,0xe8dcU,0xc1b3U,0x936dU,
		0x36d1U,0x6da2U,0xdb44U,0xa683U,0x5d0dU,0xba1aU,0x643fU,0xc87eU,0x80f7U,0x11e5U,
		0x23caU,0x4794U,0x8f28U,0x0e5bU,0x1cb6U,0x396cU,0x72d8U,0xe5b0U,0xdb6bU,0xa6ddU,
		0x5db1U,0xbb62U,0x66cfU,0xcd9eU,0x8b37U,0x0665U,0x0ccaU,0x1994U,0x3328U,0x6650U,
		0xcca0U,0x894bU,0x029dU,0x053aU,0x0a74U,0x14e8U,0x29d0U,0x53a0U,0xa740U,0x5e8bU,
		0xbd16U,0x6a27U,0xd44eU,0xb897U,0x6125U,0xc24aU,0x949fU,0x3935U,0x726aU,0xe4d4U,
		0xd9a3U,0xa34dU,0x5691U,0xad22U,0x4a4fU,0x949eU,0x3937U,0x726eU,0xe4dcU,0xd9b3U,
		0xa36dU,0x56d1U,0xada2U,0x4b4fU,0x969eU,0x3d37U,0x7a6eU,0xf4dcU,0xf9b3U,0xe36dU,
		0xd6d1U,0xbda9U,0x6b59U,0xd6b2U,0xbd6fU,0x6ad5U,0xd5aaU,0xbb5fU,0x66b5U,0xcd6aU,
		0x8adfU,0x05b5U,0x0b6aU,0x16d4U,0x2da8U,0x5b50U,0xb6a0U,0x7d4bU,0xfa96U,0xe527U,
		0xda45U,0xa481U,0x5909U,0xb212U,0x742fU,0xe85eU,0xc0b7U,0x9165U,0x32c1U,0x6582U,
		0xcb04U,0x8603U,0x1c0dU,0x381aU,0x7034U,0xe068U,0xd0dbU,0xb1bdU,0x7371U,0xe6e2U,
		0xddcfU,0xab95U,0x4721U,0x8e42U,0x0c8fU,0x191eU,0x323cU,0x6478U,0xc8f0U,0x81ebU,
		0x13ddU,0x27baU,0x4f74U,0x9ee8U,0x2ddbU,0x5bb6U,0xb76cU,0x7ed3U,0xfda6U,0xeb47U,
		0xc685U,0x9d01U,0x2a09U,0x5412U,0xa824U,0x4043U,0x8086U,0x1107U,0x220eU,0x441cU,
		0x8838U,0x007bU,0x00f6U,0x01ecU,0x03d8U,0x07b0U,0x0f60U,0x1ec0U,0x3d80U,0x7b00U,
		0xf600U,0xfc0bU,0xe81dU,0xc031U,0x9069U,0x30d9U,0x61b2U,0xc364U,0x96c3U,0x3d8dU,
		0x7b1aU,0xf634U,0xfc63U,0xe8cdU,0xc191U,0x9329U,0x3659U,0x6cb2U,0xd964U,0xa2c3U,
		0x558dU,0xab1aU,0x463fU,0x8c7eU,0x08f7U,0x11eeU,0x23dcU,0x47b8U,0x8f70U,0x0eebU,
		0x1dd6U,0x3bacU,0x7758U,0xeeb0U,0xcd6bU,0x8addU,0x05b1U,0x0b62U,0x16c4U,0x2d88U,
		0x5b10U,0xb620U,0x7c4bU,0xf896U,0xe127U,0xd245U,0xb481U,0x7909U,0xf212U,0xf42fU,
		0xf855U,0xe0a1U,0xd149U,0xb299U,0x7539U,0xea72U,0xc4efU,0x99d5U,0x23a1U,0x4742U,
		0x8e84U,0x0d03U,0x1a06U,0x340cU,0x6818U,0xd030U,0xb06bU,0x70ddU,0xe1baU,0xd37fU,
		0xb6f5U,0x7de1U,0xfbc2U,0xe78fU,0xdf15U,0xae21U,0x4c49U,0x9892U,0x212fU,0x425eU,
		0x84bcU,0x1973U,0x32e6U,0x65ccU,0xcb98U,0x873bU,0x1e7dU,0x3cfaU,0x79f4U,0xf3e8U,
		0xf7dbU,0xffbdU,0xef71U,0xcee9U,0x8dd9U,0x0bb9U,0x1772U,0x2ee4U,0x5dc8U,0xbb90U,
		0x672bU,0xce56U,0x8ca7U,0x0945U,0x128aU,0x2514U,0x4a28U,0x9450U,0x38abU,0x7156U,
		0xe2acU,0xd553U,0xbaadU,0x6551U,0xcaa2U,0x854fU,0x1a95U,0x352aU,0x6a54U,0xd4a8U,
		0xb95bU,0x62bdU,0xc57aU,0x9affU,0x25f5U,0x4beaU,0x97d4U,0x3fa3U,0x7f46U,0xfe8cU,
		0xed13U,0xca2dU,0x8451U,0x18a9U,0x3152U,0x62a4U,0xc548U,0x9a9bU,0x253dU,0x4a7aU,
		0x94f4U,0x39e3U,0x73c6U,0xe78cU,0xdf13U,0xae2dU,0x4c51U,0x98a2U,0x214fU,0x429eU,
		0x853cU,0x1a73U,0x34e6U,0x69ccU,0xd398U,0xb73bU,0x7e7dU,0xfcfaU,0xe9ffU,0xc3f5U,
		0x97e1U,0x3fc9U,0x7f92U,0xff24U,0xee43U,0xcc8dU,0x8911U,0x0229U,0x0452U,0x08a4U,
		0x1148U,0x2290U,0x4520U,0x8a40U,0x048bU,0x0916U,0x122cU,0x2458U,0x48b0U,0x9160U,
		0x32cbU,0x6596U,0xcb2cU,0x8653U,0x1cadU,0x395aU,0x72b4U,0xe568U,0xdadbU,0xa5bdU,
		0x5b71U,0xb6e2U,0x7dcfU,0xfb9eU,0xe737U,0xde65U,0xacc1U,0x4989U,0x9312U,0x362fU,
		0x6c5eU,0xd8bcU,0xa173U,0x52edU,0xa5daU,0x5bbfU,0xb77eU,0x7ef7U,0xfdeeU,0xebd7U,
		0xc7a5U,0x9f41U,0x2e89U,0x5d12U,0xba24U,0x6443U,0xc886U,0x8107U,0x1205U,0x240aU,
		0x4814U,0x9028U,0x305bU,0x60b6U,0xc16cU,0x92d3U,0x35adU,0x6b5aU,0xd6b4U,0xbd63U,
		0x6acdU,0xd59aU,0xbb3fU,0x6675U,0xcceaU,0x89dfU,0x03b5U,0x076aU,0x0ed4U,0x1da8U,
		0x3b50U,0x76a0U,0xed40U,0xca8bU,0x851dU,0x1a31U,0x3462U,0x68c4U,0xd188U,0xb31bU,
		0x763dU,0xec7aU,0xc8ffU,0x81f5U,0x13e1U,0x27c2U,0x4f84U,0x9f08U,0x2e1bU,0x5c36U,
		0xb86cU,0x60d3U,0xc1a6U,0x9347U,0x3685U,0x6d0aU,0xda14U,0xa423U,0x584dU,0xb09aU,
		0x713fU,0xe27eU,0xd4f7U,0xb9e5U,0x63c1U,0xc782U,0x9f0fU,0x2e15U,0x5c2aU,0xb854U,
		0x60a3U,0xc146U,0x9287U,0x3505U,0x6a0aU,0xd414U,0xb823U,0x604dU,0xc09aU,0x913fU,
		0x3275U,0x64eaU,0xc9d4U,0x83a3U,0x174dU,0x2e9aU,0x5d34U,0xba68U,0x64dbU,0xc9b6U,
		0x8367U,0x16c5U,0x2d8aU,0x5b14U,0xb628U,0x7c5bU,0xf8b6U,0xe167U,0xd2c5U,0xb581U,
		0x7b09U,0xf612U,0xfc2fU,0xe855U,0xc0a1U,0x9149U,0x3299U,0x6532U,0xca64U,0x84c3U,
		0x198dU,0x331aU,0x6634U,0xcc68U,0x88dbU,0x01bdU,0x037aU,0x06f4U,0x0de8U,0x1bd0U,
		0x37a0U,0x6f40U,0xde80U,0xad0bU,0x4a1dU,0x943aU,0x387fU,0x70feU,0xe1fcU,0xd3f3U,
		0xb7edU,0x7fd1U,0xffa2U,0xef4fU,0xce95U,0x8d21U,0x0a49U,0x1492U,0x2924U,0x5248U,
		0xa490U,0x592bU,0xb256U,0x74a7U,0xe94eU,0xc297U,0x9525U,0x3a41U,0x7482U,0xe904U,
		0xc203U,0x940dU,0x3811U,0x7022U,0xe044U,0xd083U,0xb10dU,0x7211U,0xe422U,0xd84fU,
		0xa095U,0x5121U,0xa242U,0x548fU,0xa91eU,0x4237U,0x846eU,0x18d7U,0x31aeU,0x635cU,
		0xc6b8U,0x9d7bU,0x2afdU,0x55faU,0xabf4U,0x47e3U,0x8fc6U,0x0f87U,0x1f0eU,0x3e1cU,
		0x7c38U,0xf870U,0xe0ebU,0xd1ddU,0xb3b1U,0x7769U,0xeed2U,0xcdafU,0x8b55U,0x06a1U,
		0x0d42U,0x1a84U,0x3508U,0x6a10U,0xd420U,0xb84bU,0x609dU,0xc13aU,0x927fU,0x34f5U,
		0x69eaU,0xd3d4U,0xb7a3U,0x7f4dU,0xfe9aU,0xed3fU,0xca75U,0x84e1U,0x19c9U,0x3392U,
		0x6724U,0xce48U,0x8c9bU,0x093dU,0x127aU,0x24f4U,0x49e8U,0x93d0U,0x37abU,0x6f56U,
		0xdeacU,0xad53U,0x4aadU,0x955aU,0x3abfU,0x757eU,0xeafcU,0xc5f3U,0x9bedU,0x27d1U,
		0x4fa2U,0x9f44U,0x2e83U,0x5d06U,0xba0cU,0x6413U,0xc826U,0x8047U,0x1085U,0x210aU,
		0x4214U,0x8428U,0x185bU,0x30b6U,0x616cU,0xc2d8U,0x95bbU,0x3b7dU,0x76faU,0xedf4U,
		0xcbe3U,0x87cdU,0x1f91U,0x3f22U,0x7e44U,0xfc88U,0xe91bU,0xc23dU,0x9471U,0x38e9U,
		0x71d2U,0xe3a4U,0xd743U,0xbe8dU,0x6d11U,0xda22U,0xa44fU,0x5895U,0xb12aU,0x725fU,
		0xe4beU,0xd977U,0xa2e5U,0x55c1U,0xab82U,0x470fU,0x8e1eU,0x0c37U,0x186eU,0x30dcU,
		0x61b8U,0xc370U,0x96ebU,0x3dddU,0x7bbaU,0xf774U,0xfee3U,0xedcdU,0xcb91U,0x8729U,
		0x1e59U,0x3cb2U,0x7964U,0xf2c8U,0xf59bU,0xfb3dU,0xe671U,0xdce9U,0xa9d9U,0x43b9U,
		0x8772U,0x1eefU,0x3ddeU,0x7bbcU,0xf778U,0xfefbU,0xedfdU,0xcbf1U,0x87e9U,0x1fd9U,
		0x3fb2U,0x7f64U,0xfec8U,0xed9bU,0xcb3dU,0x8671U,0x1ce9U,0x39d2U,0x73a4U,0xe748U,
		0xde9bU,0xad3dU,0x4a71U,0x94e2U,0x39cfU,0x739eU,0xe73cU,0xde73U,0xacedU,0x49d1U,
		0x93a2U,0x374fU,0x6e9eU,0xdd3cU,0xaa73U,0x44edU,0x89daU,0x03bfU,0x077eU,0x0efcU,
		0x1df8U,0x3bf0U,0x77e0U,0xefc0U,0xcf8bU,0x8f1dU,0x0e31U,0x1c62U,0x38c4U,0x7188U,
		0xe310U,0xd62bU,0xbc5dU,0x68b1U,0xd162U,0xb2cfU,0x7595U,0xeb2aU,0xc65fU,0x9cb5U,
		0x2961U,0x52c2U,0xa584U,0x5b03U,0xb606U,0x7c07U,0xf80eU,0xe017U,0xd025U,0xb041U,
		0x7089U,0xe112U,0xd22fU,0xb455U,0x78a1U,0xf142U,0xf28fU,0xf515U,0xfa21U,0xe449U,
		0xd899U,0xa139U,0x5279U,0xa4f2U,0x59efU,0xb3deU,0x77b7U,0xef6eU,0xced7U,0x8da5U,
		0x0b41U,0x1682U,0x2d04U,0x5a08U,0xb410U,0x782bU,0xf056U,0xf0a7U,0xf145U,0xf281U,
		0xf509U,0xfa19U,0xe439U,0xd879U,0xa0f9U,0x51f9U,0xa3f2U,0x57efU,0xafdeU,0x4fb7U,
		0x9f6eU,0x2ed7U,0x5daeU,0xbb5cU,0x66b3U,0xcd66U,0x8ac7U,0x0585U,0x0b0aU,0x1614U,
		0x2c28U,0x5850U,0xb0a0U,0x714bU,0xe296U,0xd527U,0xba45U,0x6481U,0xc902U,0x820fU,
		0x1415U,0x282aU,0x5054U,0xa0a8U,0x515bU,0xa2b6U,0x5567U,0xaaceU,0x4597U,0x8b2eU,
		0x0657U,0x0caeU,0x195cU,0x32b8U,0x6570U,0xcae0U,0x85cbU,0x1b9dU,0x373aU,0x6e74U,
		0xdce8U,0xa9dbU,0x43bdU,0x877aU,0x1effU,0x3dfeU,0x7bfcU,0xf7f8U,0xfffbU,0xeffdU,
		0xcff1U,0x8fe9U,0x0fd9U,0x1fb2U,0x3f64U,0x7ec8U,0xfd90U,0xeb2bU,0xc65dU,0x9cb1U,
		0x2969U,0x52d2U,0xa5a4U,0x5b43U,0xb686U,0x7d07U,0xfa0eU,0xe417U,0xd825U,0xa041U,
		0x5089U,0xa112U,0x522fU,0xa45eU,0x58b7U,0xb16eU,0x72d7U,0xe5aeU,0xdb57U,0xa6a5U,
		0x5d41U,0xba82U,0x650fU,0xca1eU,0x8437U,0x1865U,0x30caU,0x6194U,0xc328U,0x965bU,
		0x3cbdU,0x797aU,0xf2f4U,0xf5e3U,0xfbcdU,0xe791U,0xdf29U,0xae59U,0x4cb9U,0x9972U,
		0x22efU,0x45deU,0x8bbcU,0x0773U,0x0ee6U,0x1dccU,0x3b98U,0x7730U,0xee60U,0xcccbU,
		0x899dU,0x0331U,0x0662U,0x0cc4U,0x1988U,0x3310U,0x6620U,0xcc40U,0x888bU,0x011dU,
		0x023aU,0x0474U,0x08e8U,0x11d0U,0x23a0U,0x4740U,0x8e80U,0x0d0bU,0x1a16U,0x342cU,
		0x6858U,0xd0b0U,0xb16bU,0x72ddU,0xe5baU,0xdb7fU,0xa6f5U,0x5de1U,0xbbc2U,0x678fU,
		0xcf1eU,0x8e37U,0x0c65U,0x18caU,0x3194U,0x6328U,0xc650U,0x9cabU,0x295dU,0x52baU,
		0xa574U,0x5ae3U,0xb5c6U,0x7b87U,0xf70eU,0xfe17U,0xec25U,0xc841U,0x8089U,0x1119U,
		0x2232U,0x4464U,0x88c8U,0x019bU,0x0336U,0x066cU,0x0cd8U,0x19b0U,0x3360U,0x66c0U,
		0xcd80U,0x8b0bU,0x061dU,0x0c3aU,0x1874U,0x30e8U,0x61d0U,0xc3a0U,0x974bU,0x3e9dU,
		0x7d3aU,0xfa74U,0xe4e3U,0xd9cdU,0xa391U,0x5729U,0xae52U,0x4cafU,0x995eU,0x22b7U,
		0x456eU,0x8adcU,0x05b3U,0x0b66U,0x16ccU,0x2d98U,0x5b30U,0xb660U,0x7ccbU,0xf996U,
		0xe327U,0xd645U,0xbc81U,0x6909U,0xd212U,0xb42fU,0x7855U,0xf0aaU,0xf15fU,0xf2b5U,
		0xf561U,0xfac9U,0xe599U,0xdb39U,0xa679U,0x5cf9U,0xb9f2U,0x63efU,0xc7deU,0x9fb7U,
		0x2f65U,0x5ecaU,0xbd94U,0x6b23U,0xd646U,0xbc87U,0x6905U,0xd20aU,0xb41fU,0x7835U,
		0xf06aU,0xf0dfU,0xf1b5U,0xf361U,0xf6c9U,0xfd99U,0xeb39U,0xc679U,0x9cf9U,0x29f9U,
		0x53f2U,0xa7e4U,0x5fc3U,0xbf86U,0x6f07U,0xde0eU,0xac17U,0x4825U,0x904aU,0x309fU,
		0x613eU,0xc27cU,0x94f3U,0x39edU,0x73daU,0xe7b4U,0xdf63U,0xaecdU,0x4d91U,0x9b22U,
		0x264fU,0x4c9eU,0x993cU,0x2273U,0x44e6U,0x89ccU,0x0393U,0x0726U,0x0e4cU,0x1c98U,
		0x3930U,0x7260U,0xe4c0U,0xd98bU,0xa31dU,0x5631U,0xac62U,0x48cfU,0x919eU,0x3337U,
		0x666eU,0xccdcU,0x89b3U,0x036dU,0x06daU,0x0db4U,0x1b68U,0x36d0U,0x6da0U,0xdb40U,
		0xa68bU,0x5d1dU,0xba3aU,0x647fU,0xc8feU,0x81f7U,0x13e5U,0x27caU,0x4f94U,0x9f28U,
		0x2e5bU,0x5cb6U,0xb96cU,0x62d3U,0xc5a6U,0x9b47U,0x2685U,0x4d0aU,0x9a14U,0x2423U,
		0x4846U,0x908cU,0x3113U,0x6226U,0xc44cU,0x9893U,0x212dU,0x425aU,0x84b4U,0x1963U,
		0x32c6U,0x658cU,0xcb18U,0x863bU,0x1c7dU,0x38faU,0x71f4U,0xe3e8U,0xd7dbU,0xbfbdU,
		0x6f71U,0xdee2U,0xadcfU,0x4b95U,0x972aU,0x3e5fU,0x7cbeU,0xf97cU,0xe2f3U,0xd5edU,
		0xbbd1U,0x67a9U,0xcf52U,0x8eafU,0x0d55U,0x1aaaU,0x3554U,0x6aa8U,0xd550U,0xbaabU,
		0x655dU,0xcabaU,0x857fU,0x1af5U,0x35eaU,0x6bd4U,0xd7a8U,0xbf5bU,0x6ebdU,0xdd7aU,
		0xaaffU,0x45f5U,0x8beaU,0x07dfU,0x0fbeU,0x1f7cU,0x3ef8U,0x7df0U,0xfbe0U,0xe7cbU,
		0xdf9dU,0xaf31U,0x4e69U,0x9cd2U,0x29afU,0x535eU,0xa6bcU,0x5d73U,0xbae6U,0x65c7U,
		0xcb8eU,0x8717U,0x1e25U,0x3c4aU,0x7894U,0xf128U,0xf25bU,0xf4bdU,0xf971U,0xe2e9U,
		0xd5d9U,0xbbb9U,0x6779U,0xcef2U,0x8defU,0x0bd5U,0x17aaU,0x2f54U,0x5ea8U,0xbd50U,
		0x6aabU,0xd556U,0xbaa7U,0x6545U,0xca8aU,0x851fU,0x1a35U,0x346aU,0x68d4U,0xd1a8U,
		0xb35bU,0x76bdU,0xed7aU,0xcaffU,0x85f5U,0x1be1U,0x37c2U,0x6f84U,0xdf08U,0xae1bU,
		0x4c3dU,0x987aU,0x20ffU,0x41feU,0x83fcU,0x17f3U,0x2fe6U,0x5fccU,0xbf98U,0x6f3bU,
		0xde76U,0xace7U,0x49c5U,0x938aU,0x371fU,0x6e3eU,0xdc7cU,0xa8f3U,0x41edU,0x83daU,
		0x17bfU,0x2f7eU,0x5efcU,0xbdf8U,0x6bfbU,0xd7f6U,0xbfe7U,0x6fc5U,0xdf8aU,0xaf1fU,
		0x4e35U,0x9c6aU,0x28dfU,0x51beU,0xa37cU,0x56f3U,0xade6U,0x4bc7U,0x978eU,0x3f17U,
		0x7e2eU,0xfc5cU,0xe8b3U,0xc16dU,0x92d1U,0x35a9U,0x6b52U,0xd6a4U,0xbd43U,0x6a8dU,
		0xd51aU,0xba3fU,0x6475U,0xc8eaU,0x81dfU,0x13b5U,0x276aU,0x4ed4U,0x9da8U,0x2b5bU,
		0x56b6U,0xad6cU,0x4ad3U,0x95a6U,0x3b47U,0x768eU,0xed1cU,0xca33U,0x846dU,0x18d1U,
		0x31a2U,0x6344U,0xc688U,0x9d1bU,0x2a3dU,0x547aU,0xa8f4U,0x41e3U,0x83c6U,0x1787U,
		0x2f0eU,0x5e1cU,0xbc38U,0x687bU,0xd0f6U,0xb1e7U,0x73c5U,0xe78aU,0xdf1fU,0xae35U,
		0x4c61U,0x98c2U,0x218fU,0x431eU,0x863cU,0x1c73U,0x38e6U,0x71ccU,0xe398U,0xd73bU,
		0xbe7dU,0x6cf1U,0xd9e2U,0xa3cfU,0x5795U,0xaf2aU,0x4e5fU,0x9cbeU,0x2977U,0x52eeU,
		0xa5dcU,0x5bb3U,0xb766U,0x7ec7U,0xfd8eU,0xeb17U,0xc625U,0x9c41U,0x2889U,0x5112U,
		0xa224U,0x5443U,0xa886U,0x4107U,0x820eU,0x1417U,0x282eU,0x505cU,0xa0b8U,0x517bU,
		0xa2f6U,0x55e7U,0xabceU,0x4797U,0x8f2eU,0x0e57U,0x1caeU,0x395cU,0x72b8U,0xe570U,
		0xdaebU,0xa5ddU,0x5bb1U,0xb762U,0x7ecfU,0xfd9eU,0xeb37U,0xc665U,0x9cc1U,0x2989U,
		0x5312U,0xa624U,0x5c43U,0xb886U,0x6107U,0xc20eU,0x9417U,0x3825U,0x704aU,0xe094U,
		0xd123U,0xb24dU,0x7491U,0xe922U,0xc24fU,0x9495U,0x3921U,0x7242U,0xe484U,0xd903U,
		0xa20dU,0x5411U,0xa822U,0x404fU,0x809eU,0x1137U,0x226eU,0x44dcU,0x89b8U,0x037bU,
		0x06f6U,0x0decU,0x1bd8U,0x37b0U,0x6f60U,0xdec0U,0xad8bU,0x4b1dU,0x963aU,0x3c7fU,
		0x78feU,0xf1fcU,0xf3f3U,0xf7edU,0xffd1U,0xefa9U,0xcf59U,0x8eb9U,0x0d79U,0x1af2U,
		0x35e4U,0x6bc8U,0xd790U,0xbf2bU,0x6e5dU,0xdcbaU,0xa97fU,0x42f5U,0x85eaU,0x1bdfU,
		0x37beU,0x6f7cU,0xdef8U,0xadfbU,0x4bfdU,0x97faU,0x3fffU,0x7ffeU,0xfffcU,0xeff3U,
		0xcfedU,0x8fd1U,0x0fa9U,0x1f52U,0x3ea4U,0x7d48U,0xfa90U,0xe52bU,0xda5dU,0xa4b1U,
		0x5969U,0xb2d2U,0x75afU,0xeb5eU,0xc6b7U,0x9d65U,0x2ac1U,0x5582U,0xab04U,0x4603U,
		0x8c06U,0x0807U,0x100eU,0x201cU,0x4038U,0x8070U,0x10ebU,0x21d6U,0x43acU,0x8758U,
		0x1ebbU,0x3d76U,0x7aecU,0xf5d8U,0xfbbbU,0xe77dU,0xdef1U,0xade9U,0x4bd9U,0x97b2U,
		0x3f6fU,0x7edeU,0xfdbcU,0xeb73U,0xc6edU,0x9dd1U,0x2ba9U,0x5752U,0xaea4U,0x4d43U,
		0x9a86U,0x2507U,0x4a0eU,0x941cU,0x3833U,0x7066U,0xe0ccU,0xd193U,0xb32dU,0x7651U,
		0xeca2U,0xc94fU,0x8295U,0x1521U,0x2a42U,0x5484U,0xa908U,0x421bU,0x8436U,0x1867U,
		0x30ceU,0x619cU,0xc338U,0x967bU,0x3cfdU,0x79faU,0xf3f4U,0xf7e3U,0xffcdU,0xef91U,
		0xcf29U,0x8e59U,0x0cb9U,0x1972U,0x32e4U,0x65c8U,0xcb90U,0x872bU,0x1e5dU,0x3cbaU,
		0x7974U,0xf2e8U,0xf5dbU,0xfbbdU,0xe771U,0xdee9U,0xadd9U,0x4bb9U,0x9772U,0x3eefU,
		0x7ddeU,0xfbbcU,0xe773U,0xdeedU,0xadd1U,0x4ba9U,0x9752U,0x3eafU,0x7d5eU,0xfabcU,
		0xe573U,0xdaedU,0xa5d1U,0x5ba9U,0xb752U,0x7eafU,0xfd5eU,0xeab7U,0xc565U,0x9ac1U,
		0x2589U,0x4b12U,0x9624U,0x3c43U,0x7886U,0xf10cU,0xf213U,0xf42dU,0xf851U,0xe0a9U,
		0xd159U,0xb2b9U,0x7579U,0xeaf2U,0xc5efU,0x9bd5U,0x27a1U,0x4f42U,0x9e84U,0x2d03U,
		0x5a06U,0xb40cU,0x7813U,0xf026U,0xf047U,0xf085U,0xf101U,0xf209U,0xf419U,0xf839U,
		0xe079U,0xd0f9U,0xb1f9U,0x73f9U,0xe7f2U,0xdfefU,0xafd5U,0x4fa1U,0x9f42U,0x2e8fU,
		0x5d1eU,0xba3cU,0x6473U,0xc8e6U,0x81c7U,0x1385U,0x270aU,0x4e14U,0x9c28U,0x285bU,
		0x50b6U,0xa16cU,0x52d3U,0xa5a6U,0x5b47U,0xb68eU,0x7d17U,0xfa2eU,0xe457U,0xd8a5U,
		0xa141U,0x5289U,0xa512U,0x5a2fU,0xb45eU,0x78b7U,0xf16eU,0xf2d7U,0xf5a5U,0xfb41U,
		0xe689U,0xdd19U,0xaa39U,0x4479U,0x88f2U,0x01efU,0x03deU,0x07bcU,0x0f78U,0x1ef0U,
		0x3de0U,0x7bc0U,0xf780U,0xff0bU,0xee1dU,0xcc31U,0x8869U,0x00d9U,0x01b2U,0x0364U,
		0x06c8U,0x0d90U,0x1b20U,0x3640U,0x6c80U,0xd900U,0xa20bU,0x541dU,0xa83aU,0x407fU,
		0x80feU,0x11f7U,0x23eeU,0x47dcU,0x8fb8U,0x0f7bU,0x1ef6U,0x3decU,0x7bd8U,0xf7b0U,
		0xff6bU,0xeeddU,0xcdb1U,0x8b69U,0x06d9U,0x0db2U,0x1b64U,0x36c8U,0x6d90U,0xdb20U,
		0xa64bU,0x5c9dU,0xb93aU,0x627fU,0xc4feU,0x99f7U,0x23e5U,0x47caU,0x8f94U,0x0f23U,
		0x1e46U,0x3c8cU,0x7918U,0xf230U,0xf46bU,0xf8ddU,0xe1b1U,0xd369U,0xb6d9U,0x7db9U,
		0xfb72U,0xe6efU,0xddd5U,0xaba1U,0x4749U,0x8e92U,0x0d2fU,0x1a5eU,0x34bcU,0x6978U,
		0xd2f0U,0xb5ebU,0x7bddU,0xf7baU,0xff7fU,0xeef5U,0xcde1U,0x8bc9U,0x0799U,0x0f32U,
		0x1e64U,0x3cc8U,0x7990U,0xf320U,0xf64bU,0xfc9dU,0xe931U,0xc269U,0x94d9U,0x39b9U,
		0x7372U,0xe6e4U,0xddc3U,0xab8dU,0x4711U,0x8e22U,0x0c4fU,0x189eU,0x313cU,0x6278U,
		0xc4f0U,0x99ebU,0x23ddU,0x47baU,0x8f74U,0x0ee3U,0x1dc6U,0x3b8cU,0x7718U,0xee30U,
		0xcc6bU,0x88ddU,0x01b1U,0x0362U,0x06c4U,0x0d88U,0x1b10U,0x3620U,0x6c40U,0xd880U,
		0xa10bU,0x521dU,0xa43aU,0x587fU,0xb0feU,0x71f7U,0xe3eeU,0xd7d7U,0xbfa5U,0x6f41U,
		0xde82U,0xad0fU,0x4a15U,0x942aU,0x385fU,0x70beU,0xe17cU,0xd2f3U,0xb5edU,0x7bd1U,
		0xf7a2U,0xff4fU,0xee95U,0xcd21U,0x8a49U,0x0499U,0x0932U,0x1264U,0x24c8U,0x4990U,
		0x9320U,0x364bU,0x6c96U,0xd92cU,0xa253U,0x54adU,0xa95aU,0x42bfU,0x857eU,0x1af7U,
		0x35eeU,0x6bdcU,0xd7b8U,0xbf7bU,0x6efdU,0xddfaU,0xabffU,0x47f5U,0x8feaU,0x0fdfU,
		0x1fbeU,0x3f7cU,0x7ef8U,0xfdf0U,0xebebU,0xc7ddU,0x9fb1U,0x2f69U,0x5ed2U,0xbda4U,
		0x6b43U,0xd686U,0xbd07U,0x6a05U,0xd40aU,0xb81fU,0x6035U,0xc06aU,0x90dfU,0x31b5U,
		0x636aU,0xc6d4U,0x9da3U,0x2b4dU,0x569aU,0xad34U,0x4a63U,0x94c6U,0x3987U,0x730eU,
		0xe61cU,0xdc33U,0xa86dU,0x40d1U,0x81a2U,0x134fU,0x269eU,0x4d3cU,0x9a78U,0x24fbU,
		0x49f6U,0x93ecU,0x37d3U,0x6fa6U,0xdf4cU,0xae93U,0x4d2dU,0x9a5aU,0x24bfU,0x497eU,
		0x92fcU,0x35f3U,0x6be6U,0xd7ccU,0xbf93U,0x6f2dU,0xde5aU,0xacbfU,0x4975U,0x92eaU,
		0x35dfU,0x6bbeU,0xd77cU,0xbef3U,0x6dedU,0xdbdaU,0xa7bfU,0x5f75U,0xbeeaU,0x6ddfU,
		0xdbbeU,0xa777U,0x5ee5U,0xbdcaU,0x6b9fU,0xd73eU,0xbe77U,0x6ce5U,0xd9caU,0xa39fU,
		0x5735U,0xae6aU,0x4cdfU,0x99beU,0x2377U,0x46eeU,0x8ddcU,0x0bb3U,0x1766U,0x2eccU,
		0x5d98U,0xbb30U,0x666bU,0xccd6U,0x89a7U,0x0345U,0x068aU,0x0d14U,0x1a28U,0x3450U,
		0x68a0U,0xd140U,0xb28bU,0x751dU,0xea3aU,0xc47fU,0x98f5U,0x21e1U,0x43c2U,0x8784U,
		0x1f03U,0x3e06U,0x7c0cU,0xf818U,0xe03bU,0xd07dU,0xb0f1U,0x71e9U,0xe3d2U,0xd7afU,
		0xbf55U,0x6ea1U,0xdd42U,0xaa8fU,0x4515U,0x8a2aU,0x045fU,0x08beU,0x117cU,0x22f8U,
		0x45f0U,0x8be0U,0x07cbU,0x0f96U,0x1f2cU,0x3e58U,0x7cb0U,0xf960U,0xe2cbU,0xd59dU,
		0xbb31U,0x6669U,0xccd2U,0x89afU,0x0355U,0x06aaU,0x0d54U,0x1aa8U,0x3550U,0x6aa0U,
		0xd540U,0xba8bU,0x651dU,0xca3aU,0x847fU,0x18f5U,0x31eaU,0x63d4U,0xc7a8U,0x9f5bU,
		0x2ebdU,0x5d7aU,0xbaf4U,0x65e3U,0xcbc6U,0x8787U,0x1f05U,0x3e0aU,0x7c14U,0xf828U,
		0xe05bU,0xd0bdU,0xb171U,0x72e9U,0xe5d2U,0xdbafU,0xa755U,0x5ea1U,0xbd42U,0x6a8fU,
		0xd51eU,0xba37U,0x6465U,0xc8caU,0x819fU,0x1335U,0x266aU,0x4cd4U,0x99a8U,0x235bU,
		0x46b6U,0x8d6cU,0x0ad3U,0x15a6U,0x2b4cU,0x5698U,0xad30U,0x4a6bU,0x94d6U,0x39a7U,
		0x734eU,0xe69cU,0xdd33U,0xaa6dU,0x44d1U,0x89a2U,0x034fU,0x069eU,0x0d3cU,0x1a78U,
		0x34f0U,0x69e0U,0xd3c0U,0xb78bU,0x7f1dU,0xfe3aU,0xec7fU,0xc8f5U,0x81e1U,0x13c9U,
		0x2792U,0x4f24U,0x9e48U,0x2c9bU,0x5936U,0xb26cU,0x74d3U,0xe9a6U,0xc347U,0x9685U,
		0x3d01U,0x7a02U,0xf404U,0xf803U,0xe00dU,0xd011U,0xb029U,0x7059U,0xe0b2U,0xd16fU,
		0xb2d5U,0x75a1U,0xeb42U,0xc68fU,0x9d15U,0x2a21U,0x5442U,0xa884U,0x4103U,0x8206U,
		0x1407U,0x280eU,0x501cU,0xa038U,0x507bU,0xa0f6U,0x51e7U,0xa3ceU,0x5797U,0xaf2eU,
		0x4e57U,0x9caeU,0x2957U,0x52aeU,0xa55cU,0x5ab3U,0xb566U,0x7ac7U,0xf58eU,0xfb17U,
		0xe625U,0xdc41U,0xa889U,0x4119U,0x8232U,0x146fU,0x28deU,0x51bcU,0xa378U,0x56fbU,
		0xadf6U,0x4be7U,0x97ceU,0x3f97U,0x7f2eU,0xfe5cU,0xecb3U,0xc96dU,0x82d1U,0x15a9U,
		0x2b52U,0x56a4U,0xad48U,0x4a9bU,0x9536U,0x3a67U,0x74ceU,0xe99cU,0xc333U,0x966dU,
		0x3cd1U,0x79a2U,0xf344U,0xf683U,0xfd0dU,0xea11U,0xc429U,0x9859U,0x20b9U,0x4172U,
		0x82e4U,0x15c3U,0x2b86U,0x570cU,0xae18U,0x4c3bU,0x9876U,0x20e7U,0x41ceU,0x839cU,
		0x1733U,0x2e66U,0x5cccU,0xb998U,0x633bU,0xc676U,0x9ce7U,0x29c5U,0x538aU,0xa714U,
		0x5e23U,0xbc46U,0x6887U,0xd10eU,0xb217U,0x7425U,0xe84aU,0xc09fU,0x9135U,0x3261U,
		0x64c2U,0xc984U,0x8303U,0x160dU,0x2c1aU,0x5834U,0xb068U,0x70dbU,0xe1b6U,0xd367U,
		0xb6c5U,0x7d81U,0xfb02U,0xe60fU,0xdc15U,0xa821U,0x4049U,0x8092U,0x112fU,0x225eU,
		0x44bcU,0x8978U,0x02fbU,0x05f6U,0x0becU,0x17d8U,0x2fb0U,0x5f60U,0xbec0U,0x6d8bU,
		0xdb16U,0xa627U,0x5c45U,0xb88aU,0x611fU,0xc23eU,0x9477U,0x38e5U,0x71caU,0xe394U,
		0xd723U,0xbe4dU,0x6c91U,0xd922U,0xa24fU,0x5495U,0xa92aU,0x425fU,0x84beU,0x1977U,
		0x32eeU,0x65dcU,0xcbb8U,0x877bU,0x1efdU,0x3dfaU,0x7bf4U,0xf7e8U,0xffdbU,0xefbdU,
		0xcf71U,0x8ee9U,0x0dd9U,0x1bb2U,0x3764U,0x6ec8U,0xdd90U,0xab2bU,0x465dU,0x8cbaU,
		0x097fU,0x12feU,0x25fcU,0x4bf8U,0x97f0U,0x3febU,0x7fd6U,0xffacU,0xef53U,0xceadU,
		0x8d51U,0x0aa9U,0x1552U,0x2aa4U,0x5548U,0xaa90U,0x452bU,0x8a56U,0x04a7U,0x094eU,
		0x129cU,0x2538U,0x4a70U,0x94e0U,0x39cbU,0x7396U,0xe72cU,0xde53U,0xacadU,0x4951U,
		0x92a2U,0x354fU,0x6a9eU,0xd53cU,0xba73U,0x64edU,0xc9daU,0x83bfU,0x1775U,0x2eeaU,
		0x5dd4U,0xbba8U,0x675bU,0xceb6U,0x8d67U,0x0ac5U,0x158aU,0x2b14U,0x5628U,0xac50U,
		0x48abU,0x9156U,0x32a7U,0x654eU,0xca9cU,0x8533U,0x1a6dU,0x34daU,0x69b4U,0xd368U,
		0xb6dbU,0x7dbdU,0xfb7aU,0xe6ffU,0xddf5U,0xabe1U,0x47c9U,0x8f92U,0x0f2fU,0x1e5eU,
		0x3cbcU,0x7978U,0xf2f0U,0xf5ebU,0xfbddU,0xe7b1U,0xdf69U,0xaed9U,0x4db9U,0x9b72U,
		0x26efU,0x4ddeU,0x9bbcU,0x2773U,0x4ee6U,0x9dccU,0x2b93U,0x5726U,0xae4cU,0x4c93U,
		0x9926U,0x2247U,0x448eU,0x891cU,0x0233U,0x0466U,0x08ccU,0x1198U,0x2330U,0x4660U,
		0x8cc0U,0x098bU,0x1316U,0x262cU,0x4c58U,0x98b0U,0x216bU,0x42d6U,0x85acU,0x1b53U,
		0x36a6U,0x6d4cU,0xda98U,0xa53bU,0x5a7dU,0xb4faU,0x79ffU,0xf3feU,0xf7f7U,0xffe5U,
		0xefc1U,0xcf89U,0x8f19U,0x0e39U,0x1c72U,0x38e4U,0x71c8U,0xe390U,0xd72bU,0xbe5dU,
		0x6cb1U,0xd962U,0xa2cfU,0x5595U,0xab2aU,0x465fU,0x8cbeU,0x0977U,0x12eeU,0x25dcU,
		0x4bb8U,0x9770U,0x3eebU,0x7dd6U,0xfbacU,0xe753U,0xdeadU,0xad51U,0x4aa9U,0x9552U,
		0x3aafU,0x755eU,0xeabcU,0xc573U,0x9aedU,0x25d1U,0x4ba2U,0x9744U,0x3e83U,0x7d06U,
		0xfa0cU,0xe413U,0xd82dU,0xa051U,0x50a9U,0xa152U,0x52afU,0xa55eU,0x5ab7U,0xb56eU,
		0x7ad7U,0xf5aeU,0xfb57U,0xe6a5U,0xdd41U,0xaa89U,0x4519U,0x8a32U,0x046fU,0x08deU,
		0x11bcU,0x2378U,0x46f0U,0x8de0U,0x0bcbU,0x1796U,0x2f2cU,0x5e58U,0xbcb0U,0x696bU,
		0xd2d6U,0xb5a7U,0x7b45U,0xf68aU,0xfd1fU,0xea35U,0xc461U,0x98c9U,0x2199U,0x4332U,
		0x8664U,0x1cc3U,0x3986U,0x730cU,0xe618U,0xdc3bU,0xa87dU,0x40f1U,0x81e2U,0x13cfU,
		0x279eU,0x4f3cU,0x9e78U,0x2cfbU,0x59f6U,0xb3ecU,0x77d3U,0xefa6U,0xcf47U,0x8e85U,
		0x0d01U,0x1a02U,0x3404U,0x6808U,0xd010U,0xb02bU,0x705dU,0xe0baU,0xd17fU,0xb2f5U,
		0x75e1U,0xebc2U,0xc78fU,0x9f15U,0x2e21U,0x5c42U,0xb884U,0x6103U,0xc206U,0x9407U,
		0x3805U,0x700aU,0xe014U,0xd023U,0xb04dU,0x7091U,0xe122U,0xd24fU,0xb495U,0x7921U,
		0xf242U,0xf48fU,0xf915U,0xe221U,0xd449U,0xb899U,0x6139U,0xc272U,0x94efU,0x39d5U,
		0x73aaU,0xe754U,0xdea3U,0xad4dU,0x4a91U,0x9522U,0x3a4fU,0x749eU,0xe93cU,0xc273U,
		0x94edU,0x39d1U,0x73a2U,0xe744U,0xde83U,0xad0dU,0x4a11U,0x9422U,0x384fU,0x709eU,
		0xe13cU,0xd273U,0xb4edU,0x79d1U,0xf3a2U,0xf74fU,0xfe95U,0xed21U,0xca49U,0x8499U,
		0x1939U,0x3272U,0x64e4U,0xc9c8U,0x839bU,0x173dU,0x2e7aU,0x5cf4U,0xb9e8U,0x63dbU,
		0xc7b6U,0x9f67U,0x2ec5U,0x5d8aU,0xbb14U,0x6623U,0xcc46U,0x8887U,0x0105U,0x020aU,
		0x0414U,0x0828U,0x1050U,0x20a0U,0x4140U,0x8280U,0x150bU,0x2a16U,0x542cU,0xa858U,
		0x40bbU,0x8176U,0x12e7U,0x25ceU,0x4b9cU,0x9738U,0x3e7bU,0x7cf6U,0xf9ecU,0xe3d3U,
		0xd7adU,0xbf51U,0x6ea9U,0xdd52U,0xaaafU,0x4555U,0x8aaaU,0x055fU,0x0abeU,0x157cU,
		0x2af8U,0x55f0U,0xabe0U,0x47cbU,0x8f96U,0x0f27U,0x1e4eU,0x3c9cU,0x7938U,0xf270U,
		0xf4ebU,0xf9ddU,0xe3b1U,0xd769U,0xbed9U,0x6db9U,0xdb72U,0xa6efU,0x5dd5U,0xbbaaU,
		0x675fU,0xcebeU,0x8d77U,0x0ae5U,0x15caU,0x2b94U,0x5728U,0xae50U,0x4cabU,0x9956U,
		0x22a7U,0x454eU,0x8a9cU,0x0533U,0x0a66U,0x14ccU,0x2998U,0x5330U,0xa660U,0x5ccbU,
		0xb996U,0x6327U,0xc64eU,0x9c97U,0x2925U,0x524aU,0xa494U,0x5923U,0xb246U,0x7487U,
		0xe90eU,0xc217U,0x9425U,0x3841U,0x7082U,0xe104U,0xd203U,0xb40dU,0x7811U,0xf022U,
		0xf04fU,0xf095U,0xf121U,0xf249U,0xf499U,0xf939U,0xe279U,0xd4f9U,0xb9f9U,0x63f9U,
		0xc7f2U,0x9fefU,0x2fd5U,0x5faaU,0xbf54U,0x6ea3U,0xdd46U,0xaa87U,0x4505U,0x8a0aU,
		0x041fU,0x083eU,0x107cU,0x20f8U,0x41f0U,0x83e0U,0x17cbU,0x2f96U,0x5f2cU,0xbe58U,
		0x6cbbU,0xd976U,0xa2e7U,0x55c5U,0xab8aU,0x471fU,0x8e3eU,0x0c77U,0x18eeU,0x31dcU,
		0x63b8U,0xc770U,0x9eebU,0x2dddU,0x5bbaU,0xb774U,0x7ee3U,0xfdc6U,0xeb87U,0xc705U,
		0x9e01U,0x2c09U,0x5812U,0xb024U,0x7043U,0xe086U,0xd107U,0xb205U,0x7401U,0xe802U,
		0xc00fU,0x9015U,0x3021U,0x6042U,0xc084U,0x9103U,0x320dU,0x641aU,0xc834U,0x8063U,
		0x10cdU,0x219aU,0x4334U,0x8668U,0x1cdbU,0x39b6U,0x736cU,0xe6d8U,0xddbbU,0xab7dU,
		0x46f1U,0x8de2U,0x0bcfU,0x179eU,0x2f3cU,0x5e78U,0xbcf0U,0x69ebU,0xd3d6U,0xb7a7U,
		0x7f45U,0xfe8aU,0xed1fU,0xca35U,0x8461U,0x18c9U,0x3192U,0x6324U,0xc648U,0x9c9bU,
		0x293dU,0x527aU,0xa4f4U,0x59e3U,0xb3c6U,0x7787U,0xef0eU,0xce17U,0x8c25U,0x0841U,
		0x1082U,0x2104U,0x4208U,0x8410U,0x182bU,0x3056U,0x60acU,0xc158U,0x92bbU,0x357dU,
		0x6afaU,0xd5f4U,0xbbe3U,0x67cdU,0xcf9aU,0x8f3fU,0x0e75U,0x1ceaU,0x39d4U,0x73a8U,
		0xe750U,0xdeabU,0xad5dU,0x4ab1U,0x9562U,0x3acfU,0x759eU,0xeb3cU,0xc673U,0x9cedU,
		0x29d1U,0x53a2U,0xa744U,0x5e83U,0xbd06U,0x6a07U,0xd40eU,0xb817U,0x6025U,0xc04aU,
		0x909fU,0x3135U,0x626aU,0xc4d4U,0x99a3U,0x234dU,0x469aU,0x8d34U,0x0a63U,0x14c6U,
		0x298cU,0x5318U,0xa630U,0x5c6bU,0xb8d6U,0x61a7U,0xc34eU,0x9697U,0x3d25U,0x7a4aU,
		0xf494U,0xf923U,0xe24dU,0xd491U,0xb929U,0x6259U,0xc4b2U,0x996fU,0x22d5U,0x45aaU,
		0x8b54U,0x06a3U,0x0d46U,0x1a8cU,0x3518U,0x6a30U,0xd460U,0xb8cbU,0x619dU,0xc33aU,
		0x967fU,0x3cf5U,0x79eaU,0xf3d4U,0xf7a3U,0xff4dU,0xee91U,0xcd29U,0x8a59U,0x04b9U,
		0x0972U,0x12e4U,0x25c8U,0x4b90U,0x9720U,0x3e4bU,0x7c96U,0xf92cU,0xe253U,0xd4adU,
		0xb951U,0x62a9U,0xc552U,0x9aafU,0x2555U,0x4aaaU,0x9554U,0x3aa3U,0x7546U,0xea8cU,
		0xc513U,0x9a2dU,0x2451U,0x48a2U,0x9144U,0x3283U,0x6506U,0xca0cU,0x8413U,0x182dU,
		0x305aU,0x60b4U,0xc168U,0x92dbU,0x35bdU,0x6b7aU,0xd6f4U,0xbde3U,0x6bcdU,0xd79aU,
		0xbf3fU,0x6e75U,0xdceaU,0xa9dfU,0x43b5U,0x876aU,0x1edfU,0x3dbeU,0x7b7cU,0xf6f8U,
		0xfdfbU,0xebfdU,0xc7f1U,0x9fe9U,0x2fd9U,0x5fb2U,0xbf64U,0x6ec3U,0xdd86U,0xab07U,
		0x4605U,0x8c0aU,0x081fU,0x103eU,0x207cU,0x40f8U,0x81f0U,0x13ebU,0x27d6U,0x4facU,
		0x9f58U,0x2ebbU,0x5d76U,0xbaecU,0x65d3U,0xcba6U,0x8747U,0x1e85U,0x3d0aU,0x7a14U,
		0xf428U,0xf85bU,0xe0bdU,0xd171U,0xb2e9U,0x75d9U,0xebb2U,0xc76fU,0x9ed5U,0x2da1U,
		0x5b42U,0xb684U,0x7d03U,0xfa06U,0xe407U,0xd805U,0xa001U,0x5009U,0xa012U,0x502fU,
		0xa05eU,0x50b7U,0xa16eU,0x52d7U,0xa5aeU,0x5b57U,0xb6aeU,0x7d57U,0xfaaeU,0xe557U,
		0xdaa5U,0xa541U,0x5a89U,0xb512U,0x7a2fU,0xf45eU,0xf8b7U,0xe165U,0xd2c1U,0xb589U,
		0x7b19U,0xf632U,0xfc6fU,0xe8d5U,0xc1a1U,0x9349U,0x3699U,0x6d32U,0xda64U,0xa4c3U,
		0x598dU,0xb31aU,0x763fU,0xec7eU,0xc8f7U,0x81e5U,0x13c1U,0x2782U,0x4f04U,0x9e08U,
		0x2c1bU,0x5836U,0xb06cU,0x70d3U,0xe1a6U,0xd347U,0xb685U,0x7d01U,0xfa02U,0xe40fU,
		0xd815U,0xa021U,0x5049U,0xa092U,0x512fU,0xa25eU,0x54b7U,0xa96eU,0x42d7U,0x85aeU,
		0x1b57U,0x36aeU,0x6d5cU,0xdab8U,0xa57bU,0x5afdU,0xb5faU,0x7bffU,0xf7feU,0xfff7U,
		0xefe5U,0xcfc1U,0x8f89U,0x0f19U,0x1e32U,0x3c64U,0x78c8U,0xf190U,0xf32bU,0xf65dU,
		0xfcb1U,0xe969U,0xc2d9U,0x95b9U,0x3b79U,0x76f2U,0xede4U,0xcbc3U,0x878dU,0x1f11U,
		0x3e22U,0x7c44U,0xf888U,0xe11bU,0xd23dU,0xb471U,0x78e9U,0xf1d2U,0xf3afU,0xf755U,
		0xfea1U,0xed49U,0xca99U,0x8539U,0x1a79U,0x34f2U,0x69e4U,0xd3c8U,0xb79bU,0x7f3dU,
		0xfe7aU,0xecffU,0xc9f5U,0x83e1U,0x17c9U,0x2f92U,0x5f24U,0xbe48U,0x6c9bU,0xd936U,
		0xa267U,0x54c5U,0xa98aU,0x431fU,0x863eU,0x1c77U,0x38eeU,0x71dcU,0xe3b8U,0xd77bU,
		0xbefdU,0x6df1U,0xdbe2U,0xa7cfU,0x5f95U,0xbf2aU,0x6e5fU,0xdcbeU,0xa977U,0x42e5U,
		0x85caU,0x1b9fU,0x373eU,0x6e7cU,0xdcf8U,0xa9fbU,0x43fdU,0x87faU,0x1fffU,0x3ffeU,
		0x7ffcU,0xfff8U,0xeffbU,0xcffdU,0x8ff1U,0x0fe9U,0x1fd2U,0x3fa4U,0x7f48U,0xfe90U,
		0xed2bU,0xca5dU,0x84b1U,0x1969U,0x32d2U,0x65a4U,0xcb48U,0x869bU,0x1d3dU,0x3a7aU,
		0x74f4U,0xe9e8U,0xc3dbU,0x97bdU,0x3f71U,0x7ee2U,0xfdc4U,0xeb83U,0xc70dU,0x9e11U,
		0x2c29U,0x5852U,0xb0a4U,0x7143U,0xe286U,0xd507U,0xba05U,0x6401U,0xc802U,0x800fU,
		0x1015U,0x202aU,0x4054U,0x80a8U,0x115bU,0x22b6U,0x456cU,0x8ad8U,0x05bbU,0x0b76U,
		0x16ecU,0x2dd8U,0x5bb0U,0xb760U,0x7ecbU,0xfd96U,0xeb27U,0xc645U,0x9c81U,0x2909U,
		0x5212U,0xa424U,0x5843U,0xb086U,0x7107U,0xe20eU,0xd417U,0xb825U,0x6041U,0xc082U,
		0x910fU,0x3215U,0x642aU,0xc854U,0x80a3U,0x114dU,0x229aU,0x4534U,0x8a68U,0x04dbU,
		0x09b6U,0x136cU,0x26d8U,0x4db0U,0x9b60U,0x26cbU,0x4d96U,0x9b2cU,0x2653U,0x4ca6U,
		0x994cU,0x2293U,0x4526U,0x8a4cU,0x0493U,0x0926U,0x124cU,0x2498U,0x4930U,0x9260U,
		0x34cbU,0x6996U,0xd32cU,0xb653U,0x7cadU,0xf95aU,0xe2bfU,0xd575U,0xbae1U,0x65c9U,
		0xcb92U,0x872fU,0x1e55U,0x3caaU,0x7954U,0xf2a8U,0xf55bU,0xfabdU,0xe571U,0xdae9U,
		0xa5d9U,0x5bb9U,0xb772U,0x7eefU,0xfddeU,0xebb7U,0xc765U,0x9ec1U,0x2d89U,0x5b12U,
		0xb624U,0x7c43U,0xf886U,0xe107U,0xd205U,0xb401U,0x7809U,0xf012U,0xf02fU,0xf055U,
		0xf0a1U,0xf149U,0xf299U,0xf539U,0xfa79U,0xe4f9U,0xd9f9U,0xa3f9U,0x57f9U,0xaff2U,
		0x4fefU,0x9fdeU,0x2fb7U,0x5f6eU,0xbedcU,0x6db3U,0xdb66U,0xa6c7U,0x5d85U,0xbb0aU,
		0x661fU,0xcc3eU,0x8877U,0x00e5U,0x01caU,0x0394U,0x0728U,0x0e50U,0x1ca0U,0x3940U,
		0x7280U,0xe500U,0xda0bU,0xa41dU,0x5831U,0xb062U,0x70cfU,0xe19eU,0xd337U,0xb665U,
		0x7cc1U,0xf982U,0xe30fU,0xd615U,0xbc21U,0x6849U,0xd092U,0xb12fU,0x7255U,0xe4aaU,
		0xd95fU,0xa2b5U,0x5561U,0xaac2U,0x458fU,0x8b1eU,0x0637U,0x0c6eU,0x18dcU,0x31b8U,
		0x6370U,0xc6e0U,0x9dcbU,0x2b9dU,0x573aU,0xae74U,0x4ce3U,0x99c6U,0x2387U,0x470eU,
		0x8e1cU,0x0c33U,0x1866U,0x30ccU,0x6198U,0xc330U,0x966bU,0x3cddU,0x79baU,0xf374U,
		0xf6e3U,0xfdcdU,0xeb91U,0xc729U,0x9e59U,0x2cb9U,0x5972U,0xb2e4U,0x75c3U,0xeb86U,
		0xc707U,0x9e05U,0x2c01U,0x5802U,0xb004U,0x7003U,0xe006U,0xd007U,0xb005U,0x7001U,
		0xe002U,0xd00fU,0xb015U,0x7021U,0xe042U,0xd08fU,0xb115U,0x7221U,0xe442U,0xd88fU,
		0xa115U,0x5221U,0xa442U,0x588fU,0xb11eU,0x7237U,0xe46eU,0xd8d7U,0xa1a5U,0x5341U,
		0xa682U,0x5d0fU,0xba1eU,0x6437U,0xc86eU,0x80d7U,0x11a5U,0x234aU,0x4694U,0x8d28U,
		0x0a5bU,0x14b6U,0x296cU,0x52d8U,0xa5b0U,0x5b6bU,0xb6d6U,0x7da7U,0xfb4eU,0xe697U,
		0xdd25U,0xaa41U,0x4489U,0x8912U,0x022fU,0x045eU,0x08bcU,0x1178U,0x22f0U,0x45e0U,
		0x8bc0U,0x078bU,0x0f16U,0x1e2cU,0x3c58U,0x78b0U,0xf160U,0xf2cbU,0xf59dU,0xfb31U,
		0xe669U,0xdcd9U,0xa9b9U,0x4379U,0x86f2U,0x1defU,0x3bdeU,0x77bcU,0xef78U,0xcefbU,
		0x8dfdU,0x0bf1U,0x17e2U,0x2fc4U,0x5f88U,0xbf10U,0x6e2bU,0xdc56U,0xa8a7U,0x4145U,
		0x828aU,0x151fU,0x2a3eU,0x547cU,0xa8f8U,0x41fbU,0x83f6U,0x17e7U,0x2fceU,0x5f9cU,
		0xbf38U,0x6e7bU,0xdcf6U,0xa9e7U,0x43c5U,0x878aU,0x1f1fU,0x3e3eU,0x7c7cU,0xf8f8U,
		0xe1fbU,0xd3fdU,0xb7f1U,0x7fe9U,0xffd2U,0xefafU,0xcf55U,0x8ea1U,0x0d49U,0x1a92U,
		0x3524U,0x6a48U,0xd490U,0xb92bU,0x625dU,0xc4baU,0x997fU,0x22f5U,0x45eaU,0x8bd4U,
		0x07a3U,0x0f46U,0x1e8cU,0x3d18U,0x7a30U,0xf460U,0xf8cbU,0xe19dU,0xd331U,0xb669U,
		0x7cd9U,0xf9b2U,0xe36fU,0xd6d5U,0xbda1U,0x6b49U,0xd692U,0xbd2fU,0x6a55U,0xd4aaU,
		0xb95fU,0x62b5U,0xc56aU,0x9adfU,0x25b5U,0x4b6aU,0x96d4U,0x3da3U,0x7b46U,0xf68cU,
		0xfd13U,0xea2dU,0xc451U,0x98a9U,0x2159U,0x42b2U,0x8564U,0x1ac3U,0x3586U,0x6b0cU,
		0xd618U,0xbc3bU,0x687dU,0xd0faU,0xb1ffU,0x73f5U,0xe7eaU,0xdfdfU,0xafb5U,0x4f61U,
		0x9ec2U,0x2d8fU,0x5b1eU,0xb63cU,0x7c73U,0xf8e6U,0xe1c7U,0xd385U,0xb701U,0x7e09U,
		0xfc12U,0xe82fU,0xc055U,0x90a1U,0x3149U,0x6292U,0xc524U,0x9a43U,0x248dU,0x491aU,
		0x9234U,0x3463U,0x68c6U,0xd18cU,0xb313U,0x762dU,0xec5aU,0xc8bfU,0x8175U,0x12e1U,
		0x25c2U,0x4b84U,0x9708U,0x3e1bU,0x7c36U,0xf86cU,0xe0d3U,0xd1adU,0xb351U,0x76a9U,
		0xed52U,0xcaafU,0x8555U,0x1aa1U,0x3542U,0x6a84U,0xd508U,0xba1bU,0x643dU,0xc87aU,
		0x80ffU,0x11f5U,0x23eaU,0x47d4U,0x8fa8U,0x0f5bU,0x1eb6U,0x3d6cU,0x7ad8U,0xf5b0U,
		0xfb6bU,0xe6ddU,0xddb1U,0xab69U,0x46d9U,0x8db2U,0x0b6fU,0x16deU,0x2dbcU,0x5b78U,
		0xb6f0U,0x7debU,0xfbd6U,0xe7a7U,0xdf45U,0xae81U,0x4d09U,0x9a12U,0x242fU,0x485eU,
		0x90bcU,0x3173U,0x62e6U,0xc5ccU,0x9b93U,0x272dU,0x4e5aU,0x9cb4U,0x2963U,0x52c6U,
		0xa58cU,0x5b13U,0xb626U,0x7c47U,0xf88eU,0xe117U,0xd225U,0xb441U,0x7889U,0xf112U,
		0xf22fU,0xf455U,0xf8a1U,0xe149U,0xd299U,0xb539U,0x7a79U,0xf4f2U,0xf9efU,0xe3d5U,
		0xd7a1U,0xbf49U,0x6e99U,0xdd32U,0xaa6fU,0x44d5U,0x89aaU,0x035fU,0x06beU,0x0d7cU,
		0x1af8U,0x35f0U,0x6be0U,0xd7c0U,0xbf8bU,0x6f1dU,0xde3aU,0xac7fU,0x48f5U,0x91eaU,
		0x33dfU,0x67beU,0xcf7cU,0x8ef3U,0x0dedU,0x1bdaU,0x37b4U,0x6f68U,0xded0U,0xadabU,
		0x4b5dU,0x96baU,0x3d7fU,0x7afeU,0xf5fcU,0xfbf3U,0xe7edU,0xdfd1U,0xafa9U,0x4f59U,
		0x9eb2U,0x2d6fU,0x5adeU,0xb5bcU,0x7b73U,0xf6e6U,0xfdc7U,0xeb85U,0xc701U,0x9e09U,
		0x2c19U,0x5832U,0xb064U,0x70c3U,0xe186U,0xd307U,0xb605U,0x7c01U,0xf802U,0xe00fU,
		0xd015U,0xb021U,0x7049U,0xe092U,0xd12fU,0xb255U,0x74a1U,0xe942U,0xc28fU,0x9515U,
		0x3a21U,0x7442U,0xe884U,0xc103U,0x920dU,0x3411U,0x6822U,0xd044U,0xb083U,0x710dU,
		0xe21aU,0xd43fU,0xb875U,0x60e1U,0xc1c2U,0x938fU,0x3715U,0x6e2aU,0xdc54U,0xa8a3U,
		0x414dU,0x829aU,0x153fU,0x2a7eU,0x54fcU,0xa9f8U,0x43fbU,0x87f6U,0x1fe7U,0x3fceU,
		0x7f9cU,0xff38U,0xee7bU,0xccfdU,0x89f1U,0x03e9U,0x07d2U,0x0fa4U,0x1f48U,0x3e90U,
		0x7d20U,0xfa40U,0xe48bU,0xd91dU,0xa231U,0x5469U,0xa8d2U,0x41afU,0x835eU,0x16b7U,
		0x2d6eU,0x5adcU,0xb5b8U,0x7b7bU,0xf6f6U,0xfde7U,0xebc5U,0xc781U,0x9f09U,0x2e19U,
		0x5c32U,0xb864U,0x60c3U,0xc186U,0x9307U,0x3605U,0x6c0aU,0xd814U,0xa023U,0x504dU,
		0xa09aU,0x513fU,0xa27eU,0x54f7U,0xa9eeU,0x43d7U,0x87aeU,0x1f57U,0x3eaeU,0x7d5cU,
		0xfab8U,0xe57bU,0xdafdU,0xa5f1U,0x5be9U,0xb7d2U,0x7fafU,0xff5eU,0xeeb7U,0xcd65U,
		0x8ac1U,0x0589U,0x0b12U,0x1624U,0x2c48U,0x5890U,0xb120U,0x724bU,0xe496U,0xd927U,
		0xa245U,0x5481U,0xa902U,0x420fU,0x841eU,0x1837U,0x306eU,0x60dcU,0xc1b8U,0x937bU,
		0x36fdU,0x6dfaU,0xdbf4U,0xa7e3U,0x5fcdU,0xbf9aU,0x6f3fU,0xde7eU,0xacf7U,0x49e5U,
		0x93caU,0x379fU,0x6f3eU,0xde7cU,0xacf3U,0x49edU,0x93daU,0x37bfU,0x6f7eU,0xdefcU,
		0xadf3U,0x4bedU,0x97daU,0x3fbfU,0x7f7eU,0xfefcU,0xedf3U,0xcbedU,0x87d1U,0x1fa9U,
		0x3f52U,0x7ea4U,0xfd48U,0xea9bU,0xc53dU,0x9a71U,0x24e9U,0x49d2U,0x93a4U,0x3743U,
		0x6e86U,0xdd0cU,0xaa13U,0x442dU,0x885aU,0x00bfU,0x017eU,0x02fcU,0x05f8U,0x0bf0U,
		0x17e0U,0x2fc0U,0x5f80U,0xbf00U,0x6e0bU,0xdc16U,0xa827U,0x4045U,0x808aU,0x111fU,
		0x223eU,0x447cU,0x88f8U,0x01fbU,0x03f6U,0x07ecU,0x0fd8U,0x1fb0U,0x3f60U,0x7ec0U,
		0xfd80U,0xeb0bU,0xc61dU,0x9c31U,0x2869U,0x50d2U,0xa1a4U,0x5343U,0xa686U,0x5d07U,
		0xba0eU,0x6417U,0xc82eU,0x8057U,0x10a5U,0x214aU,0x4294U,0x8528U,0x1a5bU,0x34b6U,
		0x696cU,0xd2d8U,0xb5bbU,0x7b7dU,0xf6faU,0xfdffU,0xebf5U,0xc7e1U,0x9fc9U,0x2f99U,
		0x5f32U,0xbe64U,0x6cc3U,0xd986U,0xa307U,0x5605U,0xac0aU,0x481fU,0x903eU,0x3077U,
		0x60eeU,0xc1dcU,0x93b3U,0x376dU,0x6edaU,0xddb4U,0xab63U,0x46cdU,0x8d9aU,0x0b3fU,
		0x167eU,0x2cfcU,0x59f8U,0xb3f0U,0x77ebU,0xefd6U,0xcfa7U,0x8f45U,0x0e81U,0x1d02U,
		0x3a04U,0x7408U,0xe810U,0xc02bU,0x905dU,0x30b1U,0x6162U,0xc2c4U,0x9583U,0x3b0dU,
		0x761aU,0xec34U,0xc863U,0x80cdU,0x1191U,0x2322U,0x4644U,0x8c88U,0x091bU,0x1236U,
		0x246cU,0x48d8U,0x91b0U,0x336bU,0x66d6U,0xcdacU,0x8b53U,0x06adU,0x0d5aU,0x1ab4U,
		0x3568U,0x6ad0U,0xd5a0U,0xbb4bU,0x669dU,0xcd3aU,0x8a7fU,0x04f5U,0x09eaU,0x13d4U,
		0x27a8U,0x4f50U,0x9ea0U,0x2d4bU,0x5a96U,0xb52cU,0x7a53U,0xf4a6U,0xf947U,0xe285U,
		0xd501U,0xba09U,0x6419U,0xc832U,0x806fU,0x10d5U,0x21aaU,0x4354U,0x86a8U,0x1d5bU,
		0x3ab6U,0x756cU,0xead8U,0xc5bbU,0x9b7dU,0x26f1U,0x4de2U,0x9bc4U,0x2783U,0x4f06U,
		0x9e0cU,0x2c13U,0x5826U,0xb04cU,0x7093U,0xe126U,0xd247U,0xb485U,0x7901U,0xf202U,
		0xf40fU,0xf815U,0xe021U,0xd049U,0xb099U,0x7139U,0xe272U,0xd4efU,0xb9d5U,0x63a1U,
		0xc742U,0x9e8fU,0x2d15U,0x5a2aU,0xb454U,0x78a3U,0xf146U,0xf287U,0xf505U,0xfa01U,
		0xe409U,0xd819U,0xa039U,0x5079U,0xa0f2U,0x51efU,0xa3deU,0x57b7U,0xaf6eU,0x4ed7U,
		0x9daeU,0x2b57U,0x56aeU,0xad5cU,0x4ab3U,0x9566U,0x3ac7U,0x758eU,0xeb1cU,0xc633U,
		0x9c6dU,0x28d1U,0x51a2U,0xa344U,0x5683U,0xad06U,0x4a07U,0x940eU,0x3817U,0x702eU,
		0xe05cU,0xd0b3U,0xb16dU,0x72d1U,0xe5a2U,0xdb4fU,0xa695U,0x5d21U,0xba42U,0x648fU,
		0xc91eU,0x8237U,0x1465U,0x28caU,0x5194U,0xa328U,0x565bU,0xacb6U,0x4967U,0x92ceU,
		0x3597U,0x6b2eU,0xd65cU,0xbcb3U,0x696dU,0xd2daU,0xb5bfU,0x7b75U,0xf6eaU,0xfddfU,
		0xebb5U,0xc761U,0x9ec9U,0x2d99U,0x5b32U,0xb664U,0x7cc3U,0xf986U,0xe307U,0xd605U,
		0xbc01U,0x6809U,0xd012U,0xb02fU,0x7055U,0xe0aaU,0xd15fU,0xb2b5U,0x7561U,0xeac2U,
		0xc58fU,0x9b15U,0x2621U,0x4c42U,0x9884U,0x2103U,0x4206U,0x840cU,0x1813U,0x3026U,
		0x604cU,0xc098U,0x913bU,0x327dU,0x64faU,0xc9f4U,0x83e3U,0x17cdU,0x2f9aU,0x5f34U,
		0xbe68U,0x6cdbU,0xd9b6U,0xa367U,0x56c5U,0xad8aU,0x4b1fU,0x963eU,0x3c77U,0x78eeU,
		0xf1dcU,0xf3b3U,0xf76dU,0xfed1U,0xeda9U,0xcb59U,0x86b9U,0x1d79U,0x3af2U,0x75e4U,
		0xebc8U,0xc79bU,0x9f3dU,0x2e71U,0x5ce2U,0xb9c4U,0x6383U,0xc706U,0x9e07U,0x2c05U,
		0x580aU,0xb014U,0x7023U,0xe046U,0xd087U,0xb105U,0x7201U,0xe402U,0xd80fU,0xa015U,
		0x5021U,0xa042U,0x508fU,0xa11eU,0x5237U,0xa46eU,0x58d7U,0xb1aeU,0x7357U,0xe6aeU,
		0xdd57U,0xaaa5U,0x4541U,0x8a82U,0x050fU,0x0a1eU,0x143cU,0x2878U,0x50f0U,0xa1e0U,
		0x53cbU,0xa796U,0x5f27U,0xbe4eU,0x6c97U,0xd92eU,0xa257U,0x54a5U,0xa94aU,0x429fU,
		0x853eU,0x1a77U,0x34eeU,0x69dcU,0xd3b8U,0xb77bU,0x7efdU,0xfdfaU,0xebffU,0xc7f5U,
		0x9fe1U,0x2fc9U,0x5f92U,0xbf24U,0x6e43U,0xdc86U,0xa907U,0x4205U,0x840aU,0x181fU,
		0x303eU,0x607cU,0xc0f8U,0x91fbU,0x33fdU,0x67faU,0xcff4U,0x8fe3U,0x0fcdU,0x1f9aU,
		0x3f34U,0x7e68U,0xfcd0U,0xe9abU,0xc35dU,0x96b1U,0x3d69U,0x7ad2U,0xf5a4U,0xfb43U,
		0xe68dU,0xdd11U,0xaa29U,0x4459U,0x88b2U,0x016fU,0x02deU,0x05bcU,0x0b78U,0x16f0U,
		0x2de0U,0x5bc0U,0xb780U,0x7f0bU,0xfe16U,0xec27U,0xc845U,0x8081U,0x1109U,0x2212U,
		0x4424U,0x8848U,0x009bU,0x0136U,0x026cU,0x04d8U,0x09b0U,0x1360U,0x26c0U,0x4d80U,
		0x9b00U,0x260bU,0x4c16U,0x982cU,0x2053U,0x40a6U,0x814cU,0x1293U,0x2526U,0x4a4cU,
		0x9498U,0x393bU,0x7276U,0xe4ecU,0xd9d3U,0xa3adU,0x5751U,0xaea2U,0x4d4fU,0x9a9eU,
		0x2537U,0x4a6eU,0x94dcU,0x39b3U,0x7366U,0xe6ccU,0xdd93U,0xab2dU,0x4651U,0x8ca2U,
		0x094fU,0x129eU,0x253cU,0x4a78U,0x94f0U,0x39ebU,0x73d6U,0xe7acU,0xdf53U,0xaeadU,
		0x4d51U,0x9aa2U,0x254fU,0x4a9eU,0x953cU,0x3a73U,0x74e6U,0xe9ccU,0xc393U,0x972dU,
		0x3e51U,0x7ca2U,0xf944U,0xe283U,0xd50dU,0xba11U,0x6429U,0xc852U,0x80afU,0x1155U,
		0x22aaU,0x4554U,0x8aa8U,0x055bU,0x0ab6U,0x156cU,0x2ad8U,0x55b0U,0xab60U,0x46cbU,
		0x8d96U,0x0b27U,0x164eU,0x2c9cU,0x5938U,0xb270U,0x74ebU,0xe9d6U,0xc3a7U,0x9745U,
		0x3e81U,0x7d02U,0xfa04U,0xe403U,0xd80dU,0xa011U,0x5029U,0xa052U,0x50afU,0xa15eU,
		0x52b7U,0xa56eU,0x5ad7U,0xb5aeU,0x7b57U,0xf6aeU,0xfd57U,0xeaa5U,0xc541U,0x9a89U,
		0x2519U,0x4a32U,0x9464U,0x38c3U,0x7186U,0xe30cU,0xd613U,0xbc2dU,0x6851U,0xd0a2U,
		0xb14fU,0x7295U,0xe52aU,0xda5fU,0xa4b5U,0x5961U,0xb2c2U,0x758fU,0xeb1eU,0xc637U,
		0x9c65U,0x28c1U,0x5182U,0xa304U,0x5603U,0xac06U,0x4807U,0x900eU,0x3017U,0x602eU,
		0xc05cU,0x90b3U,0x316dU,0x62daU,0xc5b4U,0x9b63U,0x26cdU,0x4d9aU,0x9b34U,0x2663U,
		0x4cc6U,0x998cU,0x2313U,0x4626U,0x8c4cU,0x0893U,0x1126U,0x224cU,0x4498U,0x8930U,
		0x026bU,0x04d6U,0x09acU,0x1358U,0x26b0U,0x4d60U,0x9ac0U,0x258bU,0x4b16U,0x962cU,
		0x3c53U,0x78a6U,0xf14cU,0xf293U,0xf52dU,0xfa51U,0xe4a9U,0xd959U,0xa2b9U,0x5579U,
		0xaaf2U,0x45efU,0x8bdeU,0x07b7U,0x0f6eU,0x1edcU,0x3db8U,0x7b70U,0xf6e0U,0xfdcbU,
		0xeb9dU,0xc731U,0x9e69U,0x2cd9U,0x59b2U,0xb364U,0x76c3U,0xed86U,0xcb07U,0x8605U,
		0x1c01U,0x3802U,0x7004U,0xe008U,0xd01bU,0xb03dU,0x7071U,0xe0e2U,0xd1cfU,0xb395U,
		0x7721U,0xee42U,0xcc8fU,0x8915U,0x0221U,0x0442U,0x0884U,0x1108U,0x2210U,0x4420U,
		0x8840U,0x008bU,0x0116U,0x022cU,0x0458U,0x08b0U,0x1160U,0x22c0U,0x4580U,0x8b00U,
		0x060bU,0x0c16U,0x182cU,0x3058U,0x60b0U,0xc160U,0x92cbU,0x359dU,0x6b3aU,0xd674U,
		0xbce3U,0x69cdU,0xd39aU,0xb73fU,0x7e75U,0xfceaU,0xe9dfU,0xc3b5U,0x9761U,0x3ec9U,
		0x7d92U,0xfb24U,0xe643U,0xdc8dU,0xa911U,0x4229U,0x8452U,0x18afU,0x315eU,0x62bcU,
		0xc578U,0x9afbU,0x25fdU,0x4bfaU,0x97f4U,0x3fe3U,0x7fc6U,0xff8cU,0xef13U,0xce2dU,
		0x8c51U,0x08a9U,0x1152U,0x22a4U,0x4548U,0x8a90U,0x052bU,0x0a56U,0x14acU,0x2958U,
		0x52b0U,0xa560U,0x5acbU,0xb596U,0x7b27U,0xf64eU,0xfc97U,0xe925U,0xc241U,0x9489U,
		0x3919U,0x7232U,0xe464U,0xd8c3U,0xa18dU,0x5311U,0xa622U,0x5c4fU,0xb89eU,0x6137U,
		0xc26eU,0x94d7U,0x39a5U,0x734aU,0xe694U,0xdd23U,0xaa4dU,0x4491U,0x8922U,0x024fU,
		0x049eU,0x093cU,0x1278U,0x24f0U,0x49e0U,0x93c0U,0x378bU,0x6f16U,0xde2cU,0xac53U,
		0x48adU,0x915aU,0x32bfU,0x657eU,0xcafcU,0x85f3U,0x1bedU,0x37daU,0x6fb4U,0xdf68U,
		0xaedbU,0x4dbdU,0x9b7aU,0x26ffU,0x4dfeU,0x9bfcU,0x27f3U,0x4fe6U,0x9fccU,0x2f93U,
		0x5f26U,0xbe4cU,0x6c93U,0xd926U,0xa247U,0x5485U,0xa90aU,0x421fU,0x843eU,0x1877U,
		0x30eeU,0x61dcU,0xc3b8U,0x977bU,0x3efdU,0x7dfaU,0xfbf4U,0xe7e3U,0xdfcdU,0xaf91U,
		0x4f29U,0x9e52U,0x2cafU,0x595eU,0xb2bcU,0x7573U,0xeae6U,0xc5c7U,0x9b85U,0x2701U,
		0x4e02U,0x9c04U,0x2803U,0x5006U,0xa00cU,0x5013U,0xa026U,0x5047U,0xa08eU,0x5117U,
		0xa22eU,0x5457U,0xa8aeU,0x4157U,0x82aeU,0x1557U,0x2aaeU,0x555cU,0xaab8U,0x457bU,
		0x8af6U,0x05e7U,0x0bceU,0x179cU,0x2f38U,0x5e70U,0xbce0U,0x69cbU,0xd396U,0xb727U,
		0x7e45U,0xfc8aU,0xe91fU,0xc235U,0x9461U,0x38c9U,0x7192U,0xe324U,0xd643U,0xbc8dU,
		0x6911U,0xd222U,0xb44fU,0x7895U,0xf12aU,0xf25fU,0xf4b5U,0xf961U,0xe2c9U,0xd599U,
		0xbb39U,0x6679U,0xccf2U,0x89efU,0x03d5U,0x07aaU,0x0f54U,0x1ea8U,0x3d50U,0x7aa0U,
		0xf540U,0xfa8bU,0xe51dU,0xda31U,0xa469U,0x58d9U,0xb1b2U,0x736fU,0xe6deU,0xddb7U,
		0xab65U,0x46c1U,0x8d82U,0x0b0fU,0x161eU,0x2c3cU,0x5878U,0xb0f0U,0x71ebU,0xe3d6U,
		0xd7a7U,0xbf45U,0x6e81U,0xdd02U,0xaa0fU,0x4415U,0x882aU,0x005fU,0x00beU,0x017cU,
		0x02f8U,0x05f0U,0x0be0U,0x17c0U,0x2f80U,0x5f00U,0xbe00U,0x6c0bU,0xd816U,0xa027U,
		0x5045U,0xa08aU,0x511fU,0xa23eU,0x5477U,0xa8eeU,0x41d7U,0x83aeU,0x1757U,0x2eaeU,
		0x5d5cU,0xbab8U,0x657bU,0xcaf6U,0x85e7U,0x1bc5U,0x378aU,0x6f14U,0xde28U,0xac5bU,
		0x48bdU,0x917aU,0x32ffU,0x65feU,0xcbfcU,0x87f3U,0x1fedU,0x3fdaU,0x7fb4U,0xff68U,
		0xeedbU,0xcdbdU,0x8b71U,0x06e9U,0x0dd2U,0x1ba4U,0x3748U,0x6e90U,0xdd20U,0xaa4bU,
		0x449dU,0x893aU,0x027fU,0x04feU,0x09fcU,0x13f8U,0x27f0U,0x4fe0U,0x9fc0U,0x2f8bU,
		0x5f16U,0xbe2cU,0x6c53U,0xd8a6U,0xa147U,0x5285U,0xa50aU,0x5a1fU,0xb43eU,0x7877U,
		0xf0eeU,0xf1d7U,0xf3a5U,0xf741U,0xfe89U,0xed19U,0xca39U,0x8479U,0x18f9U,0x31f2U,
		0x63e4U,0xc7c8U,0x9f9bU,0x2f3dU,0x5e7aU,0xbcf4U,0x69e3U,0xd3c6U,0xb787U,0x7f05U,
		0xfe0aU,0xec1fU,0xc835U,0x8061U,0x10c9U,0x2192U,0x4324U,0x8648U,0x1c9bU,0x3936U,
		0x726cU,0xe4d8U,0xd9bbU,0xa37dU,0x56f1U,0xade2U,0x4bcfU,0x979eU,0x3f37U,0x7e6eU,
		0xfcdcU,0xe9b3U,0xc36dU,0x96d1U,0x3da9U,0x7b52U,0xf6a4U,0xfd43U,0xea8dU,0xc511U,
		0x9a29U,0x2459U,0x48b2U,0x9164U,0x32c3U,0x6586U,0xcb0cU,0x8613U,0x1c2dU,0x385aU,
		0x70b4U,0xe168U,0xd2dbU,0xb5bdU,0x7b71U,0xf6e2U,0xfdcfU,0xeb95U,0xc721U,0x9e49U,
		0x2c99U,0x5932U,0xb264U,0x74c3U,0xe986U,0xc307U,0x9605U,0x3c01U,0x7802U,0xf004U,
		0xf003U,0xf00dU,0xf011U,0xf029U,0xf059U,0xf0b9U,0xf179U,0xf2f9U,0xf5f9U,0xfbf9U,
		0xe7f9U,0xdff9U,0xaff9U,0x4ff9U,0x9ff2U,0x2fefU,0x5fdeU,0xbfbcU,0x6f73U,0xdee6U,
		0xadc7U,0x4b85U,0x970aU,0x3e1fU,0x7c3eU,0xf87cU,0xe0f3U,0xd1edU,0xb3d1U,0x77a9U,
		0xef52U,0xceafU,0x8d55U,0x0aa1U,0x1542U,0x2a84U,0x5508U,0xaa10U,0x442bU,0x8856U,
		0x00a7U,0x014eU,0x029cU,0x0538U,0x0a70U,0x14e0U,0x29c0U,0x5380U,0xa700U,0x5e0bU,
		0xbc16U,0x6827U,0xd04eU,0xb097U,0x7125U,0xe24aU,0xd49fU,0xb935U,0x6261U,0xc4c2U,
		0x998fU,0x2315U,0x462aU,0x8c54U,0x08a3U,0x1146U,0x228cU,0x4518U,0x8a30U,0x046bU,
		0x08d6U,0x11acU,0x2358U,0x46b0U,0x8d60U,0x0acbU,0x1596U,0x2b2cU,0x5658U,0xacb0U,
		0x496bU,0x92d6U,0x35a7U,0x6b4eU,0xd69cU,0xbd33U,0x6a6dU,0xd4daU,0xb9bfU,0x6375U,
		0xc6eaU,0x9ddfU,0x2bb5U,0x576aU,0xaed4U,0x4da3U,0x9b46U,0x2687U,0x4d0eU,0x9a1cU,
		0x2433U,0x4866U,0x90ccU,0x3193U,0x6326U,0xc64cU,0x9c93U,0x292dU,0x525aU,0xa4b4U,
		0x5963U,0xb2c6U,0x7587U,0xeb0eU,0xc617U,0x9c25U,0x2841U,0x5082U,0xa104U,0x5203U,
		0xa406U,0x5807U,0xb00eU,0x7017U,0xe02eU,0xd057U,0xb0a5U,0x7141U,0xe282U,0xd50fU,
		0xba15U,0x6421U,0xc842U,0x808fU,0x1115U,0x222aU,0x4454U,0x88a8U,0x015bU,0x02b6U,
		0x056cU,0x0ad8U,0x15b0U,0x2b60U,0x56c0U,0xad80U,0x4b0bU,0x9616U,0x3c27U,0x784eU,
		0xf09cU,0xf133U,0xf26dU,0xf4d1U,0xf9a9U,0xe359U,0xd6b9U,0xbd79U,0x6af9U,0xd5f2U,
		0xbbefU,0x67d5U,0xcfaaU,0x8f5fU,0x0eb5U,0x1d6aU,0x3ad4U,0x75a8U,0xeb50U,0xc6abU,
		0x9d5dU,0x2ab1U,0x5562U,0xaac4U,0x4583U,0x8b06U,0x0607U,0x0c0eU,0x181cU,0x3038U,
		0x6070U,0xc0e0U,0x91cbU,0x339dU,0x673aU,0xce74U,0x8ce3U,0x09cdU,0x139aU,0x2734U,
		0x4e68U,0x9cd0U,0x29abU,0x5356U,0xa6acU,0x5d53U,0xbaa6U,0x6547U,0xca8eU,0x8517U,
		0x1a25U,0x344aU,0x6894U,0xd128U,0xb25bU,0x74bdU,0xe97aU,0xc2ffU,0x95f5U,0x3be1U,
		0x77c2U,0xef84U,0xcf03U,0x8e0dU,0x0c11U,0x1822U,0x3044U,0x6088U,0xc110U,0x922bU,
		0x345dU,0x68baU,0xd174U,0xb2e3U,0x75cdU,0xeb9aU,0xc73fU,0x9e75U,0x2ce1U,0x59c2U,
		0xb384U,0x7703U,0xee06U,0xcc07U,0x8805U
	};
	return f16AntilogTable[a];
}